/* Automatically @generated by tree-sitter */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 8356
#define LARGE_STATE_COUNT 1773
#define SYMBOL_COUNT 457
#define ALIAS_COUNT 13
#define TOKEN_COUNT 153
#define EXTERNAL_TOKEN_COUNT 7
#define FIELD_COUNT 39
#define MAX_ALIAS_SEQUENCE_LENGTH 11
#define MAX_RESERVED_WORD_SET_SIZE 55
#define PRODUCTION_ID_COUNT 231
#define SUPERTYPE_COUNT 32

enum ts_symbol_identifiers {
  sym__lowercase_identifier = 1,
  sym_shebang = 2,
  anon_sym_SEMI_SEMI = 3,
  anon_sym_let = 4,
  anon_sym_rec = 5,
  anon_sym_and = 6,
  anon_sym_EQ = 7,
  anon_sym_TILDE = 8,
  anon_sym_QMARK = 9,
  anon_sym_COLON = 10,
  anon_sym_LPAREN = 11,
  anon_sym_RPAREN = 12,
  anon_sym_external = 13,
  anon_sym_type = 14,
  anon_sym_nonrec = 15,
  anon_sym_COLON_EQ = 16,
  anon_sym_private = 17,
  anon_sym_DOT_DOT = 18,
  anon_sym_PLUS_EQ = 19,
  anon_sym_COMMA = 20,
  anon_sym_PLUS = 21,
  anon_sym_DASH = 22,
  anon_sym_BANG = 23,
  anon_sym__ = 24,
  anon_sym_PIPE = 25,
  anon_sym_of = 26,
  anon_sym_COLON2 = 27,
  anon_sym_DOT = 28,
  anon_sym_DASH_GT = 29,
  anon_sym_LBRACK = 30,
  anon_sym_RBRACK = 31,
  anon_sym_true = 32,
  anon_sym_false = 33,
  anon_sym_STAR = 34,
  anon_sym_LBRACE = 35,
  anon_sym_SEMI = 36,
  anon_sym_RBRACE = 37,
  anon_sym_mutable = 38,
  anon_sym_constraint = 39,
  anon_sym_exception = 40,
  anon_sym_module = 41,
  anon_sym_open = 42,
  anon_sym_include = 43,
  anon_sym_class = 44,
  anon_sym_virtual = 45,
  anon_sym_val = 46,
  anon_sym_sig = 47,
  anon_sym_end = 48,
  anon_sym_with = 49,
  anon_sym_functor = 50,
  anon_sym_struct = 51,
  anon_sym_COLON_GT = 52,
  anon_sym_object = 53,
  anon_sym_inherit = 54,
  anon_sym_method = 55,
  anon_sym_in = 56,
  anon_sym_fun = 57,
  anon_sym_as = 58,
  anon_sym_initializer = 59,
  anon_sym_LBRACK_GT = 60,
  anon_sym_LBRACK_LT = 61,
  anon_sym_GT = 62,
  anon_sym_AMP = 63,
  anon_sym_LT = 64,
  anon_sym_POUND = 65,
  anon_sym_COLON_COLON = 66,
  anon_sym_LBRACK_PIPE = 67,
  anon_sym_PIPE_RBRACK = 68,
  anon_sym_LT_DASH = 69,
  anon_sym_if = 70,
  anon_sym_then = 71,
  anon_sym_else = 72,
  anon_sym_while = 73,
  anon_sym_do = 74,
  anon_sym_done = 75,
  anon_sym_for = 76,
  anon_sym_to = 77,
  anon_sym_downto = 78,
  anon_sym_match = 79,
  anon_sym_when = 80,
  anon_sym_function = 81,
  anon_sym_try = 82,
  anon_sym_assert = 83,
  anon_sym_lazy = 84,
  anon_sym_new = 85,
  anon_sym_LBRACE_LT = 86,
  anon_sym_GT_RBRACE = 87,
  anon_sym_begin = 88,
  sym_ocamlyacc_value = 89,
  anon_sym_effect = 90,
  aux_sym_attribute_token1 = 91,
  anon_sym_LBRACK_AT_AT = 92,
  anon_sym_LBRACK_AT_AT_AT = 93,
  anon_sym_LBRACK_PERCENT = 94,
  anon_sym_LBRACE_PERCENT = 95,
  aux_sym_quoted_extension_token1 = 96,
  anon_sym_LBRACK_PERCENT_PERCENT = 97,
  anon_sym_LBRACE_PERCENT_PERCENT = 98,
  anon_sym_PERCENT = 99,
  aux_sym_number_token1 = 100,
  aux_sym_signed_number_token1 = 101,
  anon_sym_SQUOTE = 102,
  anon_sym_SQUOTE2 = 103,
  aux_sym_character_content_token1 = 104,
  aux_sym_character_content_token2 = 105,
  anon_sym_DQUOTE = 106,
  aux_sym_string_content_token1 = 107,
  aux_sym_string_content_token2 = 108,
  aux_sym_string_content_token3 = 109,
  aux_sym_string_content_token4 = 110,
  aux_sym_string_content_token5 = 111,
  aux_sym_quoted_string_content_token1 = 112,
  aux_sym_escape_sequence_token1 = 113,
  aux_sym_escape_sequence_token2 = 114,
  aux_sym_escape_sequence_token3 = 115,
  aux_sym_escape_sequence_token4 = 116,
  sym_conversion_specification = 117,
  sym_pretty_printing_indication = 118,
  sym_prefix_operator = 119,
  aux_sym_sign_operator_token1 = 120,
  sym_hash_operator = 121,
  anon_sym_lsl = 122,
  anon_sym_lsr = 123,
  anon_sym_asr = 124,
  aux_sym_pow_operator_token1 = 125,
  anon_sym_mod = 126,
  anon_sym_land = 127,
  anon_sym_lor = 128,
  anon_sym_lxor = 129,
  aux_sym_mult_operator_token1 = 130,
  aux_sym_add_operator_token1 = 131,
  sym_concat_operator = 132,
  sym_rel_operator = 133,
  anon_sym_AMP_AMP = 134,
  anon_sym_or = 135,
  anon_sym_PIPE_PIPE = 136,
  sym_assign_operator = 137,
  sym_indexing_operator = 138,
  sym_let_operator = 139,
  sym_let_and_operator = 140,
  sym_match_operator = 141,
  sym__uppercase_identifier = 142,
  aux_sym_directive_token1 = 143,
  aux_sym_type_variable_token1 = 144,
  aux_sym_tag_token1 = 145,
  aux_sym_attribute_id_token1 = 146,
  sym_comment = 147,
  sym__left_quoted_string_delimiter = 148,
  sym__right_quoted_string_delimiter = 149,
  sym_line_number_directive = 150,
  sym__null = 151,
  sym__error_sentinel = 152,
  sym_compilation_unit = 153,
  sym__structure = 154,
  sym_expression_item = 155,
  sym__signature = 156,
  sym_toplevel_directive = 157,
  sym__local_structure_item = 158,
  sym__structure_item = 159,
  sym_value_definition = 160,
  sym_let_binding = 161,
  sym_parameter = 162,
  sym_external = 163,
  sym_type_definition = 164,
  sym_type_binding = 165,
  sym__type_params = 166,
  sym__type_param = 167,
  sym_variant_declaration = 168,
  sym_constructor_declaration = 169,
  sym__constructor_declaration_name = 170,
  sym__constructor_argument = 171,
  sym_record_declaration = 172,
  sym_field_declaration = 173,
  sym_external_declaration = 174,
  sym_type_constraint = 175,
  sym_exception_definition = 176,
  sym_module_definition = 177,
  sym_module_binding = 178,
  sym_module_parameter = 179,
  sym_module_type_definition = 180,
  sym_open_module = 181,
  sym_include_module = 182,
  sym_class_definition = 183,
  sym_class_binding = 184,
  sym_class_type_definition = 185,
  sym_class_type_binding = 186,
  sym__signature_item = 187,
  sym_value_specification = 188,
  sym_include_module_type = 189,
  sym__module_typed = 190,
  sym__module_type = 191,
  sym_signature = 192,
  sym_module_type_constraint = 193,
  sym_constrain_type = 194,
  sym_constrain_module = 195,
  sym_constrain_module_type = 196,
  sym_module_type_of = 197,
  sym_functor_type = 198,
  sym_parenthesized_module_type = 199,
  sym__simple_module_expression = 200,
  sym__module_expression = 201,
  sym_structure = 202,
  sym_functor = 203,
  sym_module_application = 204,
  sym_typed_module_expression = 205,
  sym_packed_module = 206,
  sym_parenthesized_module_expression = 207,
  sym__class_typed = 208,
  sym__simple_class_type = 209,
  sym__class_type = 210,
  sym_instantiated_class_type = 211,
  sym_class_body_type = 212,
  sym__class_field_specification = 213,
  sym_inheritance_specification = 214,
  sym_instance_variable_specification = 215,
  sym_method_specification = 216,
  sym_type_parameter_constraint = 217,
  sym_let_open_class_type = 218,
  sym_class_function_type = 219,
  sym__simple_class_expression = 220,
  sym__class_expression = 221,
  sym_instantiated_class = 222,
  sym_typed_class_expression = 223,
  sym_class_function = 224,
  sym_class_application = 225,
  sym_let_class_expression = 226,
  sym__class_field = 227,
  sym_inheritance_definition = 228,
  sym_instance_variable_definition = 229,
  sym_method_definition = 230,
  sym_class_initializer = 231,
  sym_let_open_class_expression = 232,
  sym_parenthesized_class_expression = 233,
  sym__typed = 234,
  sym__simple_typed = 235,
  sym__coerced = 236,
  sym__type_constrained = 237,
  sym__maybe_polymorphic_typed = 238,
  sym__polymorphic_typed = 239,
  sym__polymorphic_type = 240,
  sym_polymorphic_type = 241,
  sym__abstract_type = 242,
  sym__parenthesized_abstract_type = 243,
  sym__delimited_type = 244,
  sym__simple_type = 245,
  sym__type = 246,
  sym_function_type = 247,
  sym_labeled_argument_type = 248,
  sym__proper_tuple_type = 249,
  sym__labeled_tuple_type = 250,
  sym_labeled_tuple_element_type = 251,
  sym__tuple_type_rhs = 252,
  sym_constructed_type = 253,
  sym_aliased_type = 254,
  sym_local_open_type = 255,
  sym_polymorphic_variant_type = 256,
  sym__tag_spec = 257,
  sym_tag_specification = 258,
  sym_package_type = 259,
  sym_object_type = 260,
  sym_method_type = 261,
  sym_hash_type = 262,
  sym_parenthesized_type = 263,
  sym__delimited_expression = 264,
  sym__simple_expression = 265,
  sym__expression = 266,
  sym__sequence_expression = 267,
  sym_typed_expression = 268,
  sym_labeled_tuple_element = 269,
  sym__tuple_expression = 270,
  sym_cons_expression = 271,
  sym_list_expression = 272,
  sym_array_expression = 273,
  sym__sequence_expression_content = 274,
  sym_record_expression = 275,
  sym_field_expression = 276,
  sym_application_expression = 277,
  sym_labeled_argument = 278,
  sym_prefix_expression = 279,
  sym_sign_expression = 280,
  sym_hash_expression = 281,
  sym_infix_expression = 282,
  sym_field_get_expression = 283,
  sym_array_get_expression = 284,
  sym_string_get_expression = 285,
  sym_bigarray_get_expression = 286,
  sym_set_expression = 287,
  sym_if_expression = 288,
  sym_then_clause = 289,
  sym_else_clause = 290,
  sym_while_expression = 291,
  sym_do_clause = 292,
  sym_for_expression = 293,
  sym__sequence_expression_anonymous = 294,
  sym_match_expression = 295,
  sym__match_cases = 296,
  sym_match_case = 297,
  sym_guard = 298,
  sym_refutation_case = 299,
  sym_function_expression = 300,
  sym_fun_expression = 301,
  sym_try_expression = 302,
  sym_let_expression = 303,
  sym_assert_expression = 304,
  sym_lazy_expression = 305,
  sym_local_open_expression = 306,
  sym_package_expression = 307,
  sym_new_expression = 308,
  sym_object_copy_expression = 309,
  sym_instance_variable_expression = 310,
  sym_method_invocation = 311,
  sym_object_expression = 312,
  sym_parenthesized_expression = 313,
  sym__delimited_pattern = 314,
  sym__simple_pattern = 315,
  sym__effect_pattern = 316,
  sym__pattern = 317,
  sym__delimited_binding_pattern = 318,
  sym__simple_binding_pattern = 319,
  sym__effect_binding_pattern = 320,
  sym__binding_pattern_no_exn = 321,
  sym__binding_pattern = 322,
  sym_alias_pattern = 323,
  sym_alias_binding_pattern_no_exn = 324,
  sym_alias_binding_pattern = 325,
  sym_typed_pattern = 326,
  sym_typed_binding_pattern = 327,
  sym__or_pattern_anonymous = 328,
  sym__or_binding_pattern_no_exn_anonymous = 329,
  sym__or_binding_pattern_anonymous = 330,
  sym_constructor_pattern = 331,
  sym_constructor_binding_pattern = 332,
  sym_tag_pattern = 333,
  sym_tag_binding_pattern = 334,
  sym_polymorphic_variant_pattern = 335,
  sym_labeled_tuple_element_pattern = 336,
  sym__tuple_pattern = 337,
  sym_labeled_tuple_element_binding_pattern = 338,
  sym__tuple_binding_pattern_no_exn = 339,
  sym__tuple_binding_pattern = 340,
  sym_record_pattern = 341,
  sym_field_pattern = 342,
  sym_record_binding_pattern = 343,
  sym_field_binding_pattern = 344,
  sym_list_pattern = 345,
  sym_list_binding_pattern = 346,
  sym_cons_pattern = 347,
  sym_cons_binding_pattern_no_exn = 348,
  sym_cons_binding_pattern = 349,
  sym_array_pattern = 350,
  sym_array_binding_pattern = 351,
  sym__sequence_pattern_content = 352,
  sym__sequence_binding_pattern_content = 353,
  sym_range_pattern = 354,
  sym_lazy_pattern = 355,
  sym_lazy_binding_pattern = 356,
  sym_local_open_pattern = 357,
  sym_local_open_binding_pattern = 358,
  sym_package_pattern = 359,
  sym_parenthesized_pattern = 360,
  sym_parenthesized_binding_pattern = 361,
  sym_exception_pattern = 362,
  sym_exception_binding_pattern = 363,
  sym_effect_pattern = 364,
  sym_effect_binding_pattern = 365,
  sym_attribute = 366,
  sym_item_attribute = 367,
  sym_floating_attribute = 368,
  sym_attribute_payload = 369,
  sym_extension = 370,
  sym_quoted_extension = 371,
  sym_item_extension = 372,
  sym_quoted_item_extension = 373,
  sym__attribute = 374,
  sym__constant = 375,
  sym__signed_constant = 376,
  sym_number = 377,
  sym_signed_number = 378,
  sym_character = 379,
  sym_character_content = 380,
  sym_string = 381,
  sym_string_content = 382,
  sym_quoted_string = 383,
  sym__quoted_string = 384,
  sym_quoted_string_content = 385,
  sym_escape_sequence = 386,
  sym_boolean = 387,
  sym_unit = 388,
  sym_sign_operator = 389,
  sym__infix_operator = 390,
  sym_pow_operator = 391,
  sym_mult_operator = 392,
  sym_add_operator = 393,
  sym_and_operator = 394,
  sym_or_operator = 395,
  sym_indexing_operator_path = 396,
  sym__value_name = 397,
  sym__simple_value_pattern = 398,
  sym_parenthesized_operator = 399,
  sym_value_path = 400,
  sym_module_path = 401,
  sym_extended_module_path = 402,
  sym_module_type_path = 403,
  sym_field_path = 404,
  sym_constructor_path = 405,
  sym_type_constructor_path = 406,
  sym_class_path = 407,
  sym_class_type_path = 408,
  sym__instance_variable_name = 409,
  sym__constructor_name = 410,
  sym_directive = 411,
  sym_type_variable = 412,
  sym_tag = 413,
  sym_attribute_id = 414,
  aux_sym__structure_repeat1 = 415,
  aux_sym__structure_repeat2 = 416,
  aux_sym_expression_item_repeat1 = 417,
  aux_sym__signature_repeat1 = 418,
  aux_sym_value_definition_repeat1 = 419,
  aux_sym_let_binding_repeat1 = 420,
  aux_sym_external_repeat1 = 421,
  aux_sym_type_definition_repeat1 = 422,
  aux_sym_type_binding_repeat1 = 423,
  aux_sym__type_params_repeat1 = 424,
  aux_sym__type_param_repeat1 = 425,
  aux_sym_variant_declaration_repeat1 = 426,
  aux_sym_constructor_declaration_repeat1 = 427,
  aux_sym__constructor_argument_repeat1 = 428,
  aux_sym_record_declaration_repeat1 = 429,
  aux_sym_module_definition_repeat1 = 430,
  aux_sym_module_binding_repeat1 = 431,
  aux_sym_class_definition_repeat1 = 432,
  aux_sym_class_type_definition_repeat1 = 433,
  aux_sym_module_type_constraint_repeat1 = 434,
  aux_sym_instantiated_class_type_repeat1 = 435,
  aux_sym_class_body_type_repeat1 = 436,
  aux_sym_instance_variable_specification_repeat1 = 437,
  aux_sym_method_specification_repeat1 = 438,
  aux_sym_class_application_repeat1 = 439,
  aux_sym__abstract_type_repeat1 = 440,
  aux_sym_polymorphic_variant_type_repeat1 = 441,
  aux_sym_polymorphic_variant_type_repeat2 = 442,
  aux_sym_tag_specification_repeat1 = 443,
  aux_sym_object_type_repeat1 = 444,
  aux_sym__sequence_expression_content_repeat1 = 445,
  aux_sym_record_expression_repeat1 = 446,
  aux_sym__match_cases_repeat1 = 447,
  aux_sym_object_copy_expression_repeat1 = 448,
  aux_sym_object_expression_repeat1 = 449,
  aux_sym_record_pattern_repeat1 = 450,
  aux_sym_record_binding_pattern_repeat1 = 451,
  aux_sym__sequence_pattern_content_repeat1 = 452,
  aux_sym__sequence_binding_pattern_content_repeat1 = 453,
  aux_sym_string_content_repeat1 = 454,
  aux_sym_quoted_string_content_repeat1 = 455,
  aux_sym_attribute_id_repeat1 = 456,
  alias_sym_class_name = 457,
  alias_sym_class_type_name = 458,
  alias_sym_field_name = 459,
  alias_sym_instance_variable_name = 460,
  alias_sym_label_name = 461,
  alias_sym_method_name = 462,
  alias_sym_module_name = 463,
  alias_sym_module_type_name = 464,
  alias_sym_sequence_expression = 465,
  alias_sym_tuple_expression = 466,
  alias_sym_type_constructor = 467,
  alias_sym_value_name = 468,
  alias_sym_value_pattern = 469,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym__lowercase_identifier] = "_lowercase_identifier",
  [sym_shebang] = "shebang",
  [anon_sym_SEMI_SEMI] = ";;",
  [anon_sym_let] = "let",
  [anon_sym_rec] = "rec",
  [anon_sym_and] = "and",
  [anon_sym_EQ] = "=",
  [anon_sym_TILDE] = "~",
  [anon_sym_QMARK] = "\?",
  [anon_sym_COLON] = ":",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_external] = "external",
  [anon_sym_type] = "type",
  [anon_sym_nonrec] = "nonrec",
  [anon_sym_COLON_EQ] = ":=",
  [anon_sym_private] = "private",
  [anon_sym_DOT_DOT] = "..",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_COMMA] = ",",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_BANG] = "!",
  [anon_sym__] = "_",
  [anon_sym_PIPE] = "|",
  [anon_sym_of] = "of",
  [anon_sym_COLON2] = ":",
  [anon_sym_DOT] = ".",
  [anon_sym_DASH_GT] = "->",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_true] = "true",
  [anon_sym_false] = "false",
  [anon_sym_STAR] = "*",
  [anon_sym_LBRACE] = "{",
  [anon_sym_SEMI] = ";",
  [anon_sym_RBRACE] = "}",
  [anon_sym_mutable] = "mutable",
  [anon_sym_constraint] = "constraint",
  [anon_sym_exception] = "exception",
  [anon_sym_module] = "module",
  [anon_sym_open] = "open",
  [anon_sym_include] = "include",
  [anon_sym_class] = "class",
  [anon_sym_virtual] = "virtual",
  [anon_sym_val] = "val",
  [anon_sym_sig] = "sig",
  [anon_sym_end] = "end",
  [anon_sym_with] = "with",
  [anon_sym_functor] = "functor",
  [anon_sym_struct] = "struct",
  [anon_sym_COLON_GT] = ":>",
  [anon_sym_object] = "object",
  [anon_sym_inherit] = "inherit",
  [anon_sym_method] = "method",
  [anon_sym_in] = "in",
  [anon_sym_fun] = "fun",
  [anon_sym_as] = "as",
  [anon_sym_initializer] = "initializer",
  [anon_sym_LBRACK_GT] = "[>",
  [anon_sym_LBRACK_LT] = "[<",
  [anon_sym_GT] = ">",
  [anon_sym_AMP] = "&",
  [anon_sym_LT] = "<",
  [anon_sym_POUND] = "#",
  [anon_sym_COLON_COLON] = "::",
  [anon_sym_LBRACK_PIPE] = "[|",
  [anon_sym_PIPE_RBRACK] = "|]",
  [anon_sym_LT_DASH] = "<-",
  [anon_sym_if] = "if",
  [anon_sym_then] = "then",
  [anon_sym_else] = "else",
  [anon_sym_while] = "while",
  [anon_sym_do] = "do",
  [anon_sym_done] = "done",
  [anon_sym_for] = "for",
  [anon_sym_to] = "to",
  [anon_sym_downto] = "downto",
  [anon_sym_match] = "match",
  [anon_sym_when] = "when",
  [anon_sym_function] = "function",
  [anon_sym_try] = "try",
  [anon_sym_assert] = "assert",
  [anon_sym_lazy] = "lazy",
  [anon_sym_new] = "new",
  [anon_sym_LBRACE_LT] = "{<",
  [anon_sym_GT_RBRACE] = ">}",
  [anon_sym_begin] = "begin",
  [sym_ocamlyacc_value] = "ocamlyacc_value",
  [anon_sym_effect] = "effect",
  [aux_sym_attribute_token1] = "[@",
  [anon_sym_LBRACK_AT_AT] = "[@@",
  [anon_sym_LBRACK_AT_AT_AT] = "[@@@",
  [anon_sym_LBRACK_PERCENT] = "[%",
  [anon_sym_LBRACE_PERCENT] = "{%",
  [aux_sym_quoted_extension_token1] = "quoted_extension_token1",
  [anon_sym_LBRACK_PERCENT_PERCENT] = "[%%",
  [anon_sym_LBRACE_PERCENT_PERCENT] = "{%%",
  [anon_sym_PERCENT] = "%",
  [aux_sym_number_token1] = "number_token1",
  [aux_sym_signed_number_token1] = "signed_number_token1",
  [anon_sym_SQUOTE] = "'",
  [anon_sym_SQUOTE2] = "'",
  [aux_sym_character_content_token1] = "character_content_token1",
  [aux_sym_character_content_token2] = "character_content_token2",
  [anon_sym_DQUOTE] = "\"",
  [aux_sym_string_content_token1] = "string_content_token1",
  [aux_sym_string_content_token2] = "string_content_token2",
  [aux_sym_string_content_token3] = "string_content_token3",
  [aux_sym_string_content_token4] = "escape_sequence",
  [aux_sym_string_content_token5] = "escape_sequence",
  [aux_sym_quoted_string_content_token1] = "quoted_string_content_token1",
  [aux_sym_escape_sequence_token1] = "escape_sequence_token1",
  [aux_sym_escape_sequence_token2] = "escape_sequence_token2",
  [aux_sym_escape_sequence_token3] = "escape_sequence_token3",
  [aux_sym_escape_sequence_token4] = "escape_sequence_token4",
  [sym_conversion_specification] = "conversion_specification",
  [sym_pretty_printing_indication] = "pretty_printing_indication",
  [sym_prefix_operator] = "prefix_operator",
  [aux_sym_sign_operator_token1] = "sign_operator_token1",
  [sym_hash_operator] = "hash_operator",
  [anon_sym_lsl] = "lsl",
  [anon_sym_lsr] = "lsr",
  [anon_sym_asr] = "asr",
  [aux_sym_pow_operator_token1] = "pow_operator_token1",
  [anon_sym_mod] = "mod",
  [anon_sym_land] = "land",
  [anon_sym_lor] = "lor",
  [anon_sym_lxor] = "lxor",
  [aux_sym_mult_operator_token1] = "mult_operator_token1",
  [aux_sym_add_operator_token1] = "add_operator_token1",
  [sym_concat_operator] = "concat_operator",
  [sym_rel_operator] = "rel_operator",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_or] = "or",
  [anon_sym_PIPE_PIPE] = "||",
  [sym_assign_operator] = "assign_operator",
  [sym_indexing_operator] = "indexing_operator",
  [sym_let_operator] = "let_operator",
  [sym_let_and_operator] = "let_and_operator",
  [sym_match_operator] = "match_operator",
  [sym__uppercase_identifier] = "_uppercase_identifier",
  [aux_sym_directive_token1] = "directive_token1",
  [aux_sym_type_variable_token1] = "type_variable_token1",
  [aux_sym_tag_token1] = "tag_token1",
  [aux_sym_attribute_id_token1] = "attribute_id_token1",
  [sym_comment] = "comment",
  [sym__left_quoted_string_delimiter] = "_left_quoted_string_delimiter",
  [sym__right_quoted_string_delimiter] = "_right_quoted_string_delimiter",
  [sym_line_number_directive] = "line_number_directive",
  [sym__null] = "_null",
  [sym__error_sentinel] = "_error_sentinel",
  [sym_compilation_unit] = "compilation_unit",
  [sym__structure] = "_structure",
  [sym_expression_item] = "expression_item",
  [sym__signature] = "_signature",
  [sym_toplevel_directive] = "toplevel_directive",
  [sym__local_structure_item] = "_local_structure_item",
  [sym__structure_item] = "_structure_item",
  [sym_value_definition] = "value_definition",
  [sym_let_binding] = "let_binding",
  [sym_parameter] = "parameter",
  [sym_external] = "external",
  [sym_type_definition] = "type_definition",
  [sym_type_binding] = "type_binding",
  [sym__type_params] = "_type_params",
  [sym__type_param] = "_type_param",
  [sym_variant_declaration] = "variant_declaration",
  [sym_constructor_declaration] = "constructor_declaration",
  [sym__constructor_declaration_name] = "constructor_name",
  [sym__constructor_argument] = "_constructor_argument",
  [sym_record_declaration] = "record_declaration",
  [sym_field_declaration] = "field_declaration",
  [sym_external_declaration] = "external_declaration",
  [sym_type_constraint] = "type_constraint",
  [sym_exception_definition] = "exception_definition",
  [sym_module_definition] = "module_definition",
  [sym_module_binding] = "module_binding",
  [sym_module_parameter] = "module_parameter",
  [sym_module_type_definition] = "module_type_definition",
  [sym_open_module] = "open_module",
  [sym_include_module] = "include_module",
  [sym_class_definition] = "class_definition",
  [sym_class_binding] = "class_binding",
  [sym_class_type_definition] = "class_type_definition",
  [sym_class_type_binding] = "class_type_binding",
  [sym__signature_item] = "_signature_item",
  [sym_value_specification] = "value_specification",
  [sym_include_module_type] = "include_module_type",
  [sym__module_typed] = "_module_typed",
  [sym__module_type] = "_module_type",
  [sym_signature] = "signature",
  [sym_module_type_constraint] = "module_type_constraint",
  [sym_constrain_type] = "constrain_type",
  [sym_constrain_module] = "constrain_module",
  [sym_constrain_module_type] = "constrain_module_type",
  [sym_module_type_of] = "module_type_of",
  [sym_functor_type] = "functor_type",
  [sym_parenthesized_module_type] = "parenthesized_module_type",
  [sym__simple_module_expression] = "_simple_module_expression",
  [sym__module_expression] = "_module_expression",
  [sym_structure] = "structure",
  [sym_functor] = "functor",
  [sym_module_application] = "module_application",
  [sym_typed_module_expression] = "typed_module_expression",
  [sym_packed_module] = "packed_module",
  [sym_parenthesized_module_expression] = "parenthesized_module_expression",
  [sym__class_typed] = "_class_typed",
  [sym__simple_class_type] = "_simple_class_type",
  [sym__class_type] = "_class_type",
  [sym_instantiated_class_type] = "instantiated_class_type",
  [sym_class_body_type] = "class_body_type",
  [sym__class_field_specification] = "_class_field_specification",
  [sym_inheritance_specification] = "inheritance_specification",
  [sym_instance_variable_specification] = "instance_variable_specification",
  [sym_method_specification] = "method_specification",
  [sym_type_parameter_constraint] = "type_parameter_constraint",
  [sym_let_open_class_type] = "let_open_class_type",
  [sym_class_function_type] = "class_function_type",
  [sym__simple_class_expression] = "_simple_class_expression",
  [sym__class_expression] = "_class_expression",
  [sym_instantiated_class] = "instantiated_class",
  [sym_typed_class_expression] = "typed_class_expression",
  [sym_class_function] = "class_function",
  [sym_class_application] = "class_application",
  [sym_let_class_expression] = "let_class_expression",
  [sym__class_field] = "_class_field",
  [sym_inheritance_definition] = "inheritance_definition",
  [sym_instance_variable_definition] = "instance_variable_definition",
  [sym_method_definition] = "method_definition",
  [sym_class_initializer] = "class_initializer",
  [sym_let_open_class_expression] = "let_open_class_expression",
  [sym_parenthesized_class_expression] = "parenthesized_class_expression",
  [sym__typed] = "_typed",
  [sym__simple_typed] = "_simple_typed",
  [sym__coerced] = "_coerced",
  [sym__type_constrained] = "_type_constrained",
  [sym__maybe_polymorphic_typed] = "_maybe_polymorphic_typed",
  [sym__polymorphic_typed] = "_polymorphic_typed",
  [sym__polymorphic_type] = "_polymorphic_type",
  [sym_polymorphic_type] = "polymorphic_type",
  [sym__abstract_type] = "_abstract_type",
  [sym__parenthesized_abstract_type] = "abstract_type",
  [sym__delimited_type] = "_delimited_type",
  [sym__simple_type] = "_simple_type",
  [sym__type] = "_type",
  [sym_function_type] = "function_type",
  [sym_labeled_argument_type] = "labeled_argument_type",
  [sym__proper_tuple_type] = "tuple_type",
  [sym__labeled_tuple_type] = "tuple_type",
  [sym_labeled_tuple_element_type] = "labeled_tuple_element_type",
  [sym__tuple_type_rhs] = "_tuple_type_rhs",
  [sym_constructed_type] = "constructed_type",
  [sym_aliased_type] = "aliased_type",
  [sym_local_open_type] = "local_open_type",
  [sym_polymorphic_variant_type] = "polymorphic_variant_type",
  [sym__tag_spec] = "_tag_spec",
  [sym_tag_specification] = "tag_specification",
  [sym_package_type] = "package_type",
  [sym_object_type] = "object_type",
  [sym_method_type] = "method_type",
  [sym_hash_type] = "hash_type",
  [sym_parenthesized_type] = "parenthesized_type",
  [sym__delimited_expression] = "_delimited_expression",
  [sym__simple_expression] = "_simple_expression",
  [sym__expression] = "_expression",
  [sym__sequence_expression] = "_sequence_expression",
  [sym_typed_expression] = "typed_expression",
  [sym_labeled_tuple_element] = "labeled_tuple_element",
  [sym__tuple_expression] = "_tuple_expression",
  [sym_cons_expression] = "cons_expression",
  [sym_list_expression] = "list_expression",
  [sym_array_expression] = "array_expression",
  [sym__sequence_expression_content] = "_sequence_expression_content",
  [sym_record_expression] = "record_expression",
  [sym_field_expression] = "field_expression",
  [sym_application_expression] = "application_expression",
  [sym_labeled_argument] = "labeled_argument",
  [sym_prefix_expression] = "prefix_expression",
  [sym_sign_expression] = "sign_expression",
  [sym_hash_expression] = "hash_expression",
  [sym_infix_expression] = "infix_expression",
  [sym_field_get_expression] = "field_get_expression",
  [sym_array_get_expression] = "array_get_expression",
  [sym_string_get_expression] = "string_get_expression",
  [sym_bigarray_get_expression] = "bigarray_get_expression",
  [sym_set_expression] = "set_expression",
  [sym_if_expression] = "if_expression",
  [sym_then_clause] = "then_clause",
  [sym_else_clause] = "else_clause",
  [sym_while_expression] = "while_expression",
  [sym_do_clause] = "do_clause",
  [sym_for_expression] = "for_expression",
  [sym__sequence_expression_anonymous] = "_sequence_expression_anonymous",
  [sym_match_expression] = "match_expression",
  [sym__match_cases] = "_match_cases",
  [sym_match_case] = "match_case",
  [sym_guard] = "guard",
  [sym_refutation_case] = "refutation_case",
  [sym_function_expression] = "function_expression",
  [sym_fun_expression] = "fun_expression",
  [sym_try_expression] = "try_expression",
  [sym_let_expression] = "let_expression",
  [sym_assert_expression] = "assert_expression",
  [sym_lazy_expression] = "lazy_expression",
  [sym_local_open_expression] = "local_open_expression",
  [sym_package_expression] = "package_expression",
  [sym_new_expression] = "new_expression",
  [sym_object_copy_expression] = "object_copy_expression",
  [sym_instance_variable_expression] = "instance_variable_expression",
  [sym_method_invocation] = "method_invocation",
  [sym_object_expression] = "object_expression",
  [sym_parenthesized_expression] = "parenthesized_expression",
  [sym__delimited_pattern] = "_delimited_pattern",
  [sym__simple_pattern] = "_simple_pattern",
  [sym__effect_pattern] = "_effect_pattern",
  [sym__pattern] = "_pattern",
  [sym__delimited_binding_pattern] = "_delimited_binding_pattern",
  [sym__simple_binding_pattern] = "_simple_binding_pattern",
  [sym__effect_binding_pattern] = "_effect_binding_pattern",
  [sym__binding_pattern_no_exn] = "_binding_pattern_no_exn",
  [sym__binding_pattern] = "_binding_pattern",
  [sym_alias_pattern] = "alias_pattern",
  [sym_alias_binding_pattern_no_exn] = "alias_pattern",
  [sym_alias_binding_pattern] = "alias_pattern",
  [sym_typed_pattern] = "typed_pattern",
  [sym_typed_binding_pattern] = "typed_pattern",
  [sym__or_pattern_anonymous] = "_or_pattern_anonymous",
  [sym__or_binding_pattern_no_exn_anonymous] = "or_pattern",
  [sym__or_binding_pattern_anonymous] = "_or_binding_pattern_anonymous",
  [sym_constructor_pattern] = "constructor_pattern",
  [sym_constructor_binding_pattern] = "constructor_pattern",
  [sym_tag_pattern] = "tag_pattern",
  [sym_tag_binding_pattern] = "tag_pattern",
  [sym_polymorphic_variant_pattern] = "polymorphic_variant_pattern",
  [sym_labeled_tuple_element_pattern] = "labeled_tuple_element_pattern",
  [sym__tuple_pattern] = "_tuple_pattern",
  [sym_labeled_tuple_element_binding_pattern] = "labeled_tuple_element_binding_pattern",
  [sym__tuple_binding_pattern_no_exn] = "tuple_pattern",
  [sym__tuple_binding_pattern] = "_tuple_binding_pattern",
  [sym_record_pattern] = "record_pattern",
  [sym_field_pattern] = "field_pattern",
  [sym_record_binding_pattern] = "record_pattern",
  [sym_field_binding_pattern] = "field_pattern",
  [sym_list_pattern] = "list_pattern",
  [sym_list_binding_pattern] = "list_pattern",
  [sym_cons_pattern] = "cons_pattern",
  [sym_cons_binding_pattern_no_exn] = "cons_pattern",
  [sym_cons_binding_pattern] = "cons_pattern",
  [sym_array_pattern] = "array_pattern",
  [sym_array_binding_pattern] = "array_pattern",
  [sym__sequence_pattern_content] = "_sequence_pattern_content",
  [sym__sequence_binding_pattern_content] = "_sequence_binding_pattern_content",
  [sym_range_pattern] = "range_pattern",
  [sym_lazy_pattern] = "lazy_pattern",
  [sym_lazy_binding_pattern] = "lazy_pattern",
  [sym_local_open_pattern] = "local_open_pattern",
  [sym_local_open_binding_pattern] = "local_open_pattern",
  [sym_package_pattern] = "package_pattern",
  [sym_parenthesized_pattern] = "parenthesized_pattern",
  [sym_parenthesized_binding_pattern] = "parenthesized_pattern",
  [sym_exception_pattern] = "exception_pattern",
  [sym_exception_binding_pattern] = "exception_pattern",
  [sym_effect_pattern] = "effect_pattern",
  [sym_effect_binding_pattern] = "effect_pattern",
  [sym_attribute] = "attribute",
  [sym_item_attribute] = "item_attribute",
  [sym_floating_attribute] = "floating_attribute",
  [sym_attribute_payload] = "attribute_payload",
  [sym_extension] = "extension",
  [sym_quoted_extension] = "quoted_extension",
  [sym_item_extension] = "item_extension",
  [sym_quoted_item_extension] = "quoted_item_extension",
  [sym__attribute] = "_attribute",
  [sym__constant] = "_constant",
  [sym__signed_constant] = "_signed_constant",
  [sym_number] = "number",
  [sym_signed_number] = "signed_number",
  [sym_character] = "character",
  [sym_character_content] = "character_content",
  [sym_string] = "string",
  [sym_string_content] = "string_content",
  [sym_quoted_string] = "quoted_string",
  [sym__quoted_string] = "_quoted_string",
  [sym_quoted_string_content] = "quoted_string_content",
  [sym_escape_sequence] = "escape_sequence",
  [sym_boolean] = "boolean",
  [sym_unit] = "unit",
  [sym_sign_operator] = "sign_operator",
  [sym__infix_operator] = "_infix_operator",
  [sym_pow_operator] = "pow_operator",
  [sym_mult_operator] = "mult_operator",
  [sym_add_operator] = "add_operator",
  [sym_and_operator] = "and_operator",
  [sym_or_operator] = "or_operator",
  [sym_indexing_operator_path] = "indexing_operator_path",
  [sym__value_name] = "_value_name",
  [sym__simple_value_pattern] = "_simple_value_pattern",
  [sym_parenthesized_operator] = "parenthesized_operator",
  [sym_value_path] = "value_path",
  [sym_module_path] = "module_path",
  [sym_extended_module_path] = "extended_module_path",
  [sym_module_type_path] = "module_type_path",
  [sym_field_path] = "field_path",
  [sym_constructor_path] = "constructor_path",
  [sym_type_constructor_path] = "type_constructor_path",
  [sym_class_path] = "class_path",
  [sym_class_type_path] = "class_type_path",
  [sym__instance_variable_name] = "_instance_variable_name",
  [sym__constructor_name] = "_constructor_name",
  [sym_directive] = "directive",
  [sym_type_variable] = "type_variable",
  [sym_tag] = "tag",
  [sym_attribute_id] = "attribute_id",
  [aux_sym__structure_repeat1] = "_structure_repeat1",
  [aux_sym__structure_repeat2] = "_structure_repeat2",
  [aux_sym_expression_item_repeat1] = "expression_item_repeat1",
  [aux_sym__signature_repeat1] = "_signature_repeat1",
  [aux_sym_value_definition_repeat1] = "value_definition_repeat1",
  [aux_sym_let_binding_repeat1] = "let_binding_repeat1",
  [aux_sym_external_repeat1] = "external_repeat1",
  [aux_sym_type_definition_repeat1] = "type_definition_repeat1",
  [aux_sym_type_binding_repeat1] = "type_binding_repeat1",
  [aux_sym__type_params_repeat1] = "_type_params_repeat1",
  [aux_sym__type_param_repeat1] = "_type_param_repeat1",
  [aux_sym_variant_declaration_repeat1] = "variant_declaration_repeat1",
  [aux_sym_constructor_declaration_repeat1] = "constructor_declaration_repeat1",
  [aux_sym__constructor_argument_repeat1] = "_constructor_argument_repeat1",
  [aux_sym_record_declaration_repeat1] = "record_declaration_repeat1",
  [aux_sym_module_definition_repeat1] = "module_definition_repeat1",
  [aux_sym_module_binding_repeat1] = "module_binding_repeat1",
  [aux_sym_class_definition_repeat1] = "class_definition_repeat1",
  [aux_sym_class_type_definition_repeat1] = "class_type_definition_repeat1",
  [aux_sym_module_type_constraint_repeat1] = "module_type_constraint_repeat1",
  [aux_sym_instantiated_class_type_repeat1] = "instantiated_class_type_repeat1",
  [aux_sym_class_body_type_repeat1] = "class_body_type_repeat1",
  [aux_sym_instance_variable_specification_repeat1] = "instance_variable_specification_repeat1",
  [aux_sym_method_specification_repeat1] = "method_specification_repeat1",
  [aux_sym_class_application_repeat1] = "class_application_repeat1",
  [aux_sym__abstract_type_repeat1] = "_abstract_type_repeat1",
  [aux_sym_polymorphic_variant_type_repeat1] = "polymorphic_variant_type_repeat1",
  [aux_sym_polymorphic_variant_type_repeat2] = "polymorphic_variant_type_repeat2",
  [aux_sym_tag_specification_repeat1] = "tag_specification_repeat1",
  [aux_sym_object_type_repeat1] = "object_type_repeat1",
  [aux_sym__sequence_expression_content_repeat1] = "_sequence_expression_content_repeat1",
  [aux_sym_record_expression_repeat1] = "record_expression_repeat1",
  [aux_sym__match_cases_repeat1] = "_match_cases_repeat1",
  [aux_sym_object_copy_expression_repeat1] = "object_copy_expression_repeat1",
  [aux_sym_object_expression_repeat1] = "object_expression_repeat1",
  [aux_sym_record_pattern_repeat1] = "record_pattern_repeat1",
  [aux_sym_record_binding_pattern_repeat1] = "record_binding_pattern_repeat1",
  [aux_sym__sequence_pattern_content_repeat1] = "_sequence_pattern_content_repeat1",
  [aux_sym__sequence_binding_pattern_content_repeat1] = "_sequence_binding_pattern_content_repeat1",
  [aux_sym_string_content_repeat1] = "string_content_repeat1",
  [aux_sym_quoted_string_content_repeat1] = "quoted_string_content_repeat1",
  [aux_sym_attribute_id_repeat1] = "attribute_id_repeat1",
  [alias_sym_class_name] = "class_name",
  [alias_sym_class_type_name] = "class_type_name",
  [alias_sym_field_name] = "field_name",
  [alias_sym_instance_variable_name] = "instance_variable_name",
  [alias_sym_label_name] = "label_name",
  [alias_sym_method_name] = "method_name",
  [alias_sym_module_name] = "module_name",
  [alias_sym_module_type_name] = "module_type_name",
  [alias_sym_sequence_expression] = "sequence_expression",
  [alias_sym_tuple_expression] = "tuple_expression",
  [alias_sym_type_constructor] = "type_constructor",
  [alias_sym_value_name] = "value_name",
  [alias_sym_value_pattern] = "value_pattern",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym__lowercase_identifier] = sym__lowercase_identifier,
  [sym_shebang] = sym_shebang,
  [anon_sym_SEMI_SEMI] = anon_sym_SEMI_SEMI,
  [anon_sym_let] = anon_sym_let,
  [anon_sym_rec] = anon_sym_rec,
  [anon_sym_and] = anon_sym_and,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_external] = anon_sym_external,
  [anon_sym_type] = anon_sym_type,
  [anon_sym_nonrec] = anon_sym_nonrec,
  [anon_sym_COLON_EQ] = anon_sym_COLON_EQ,
  [anon_sym_private] = anon_sym_private,
  [anon_sym_DOT_DOT] = anon_sym_DOT_DOT,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym__] = anon_sym__,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_of] = anon_sym_of,
  [anon_sym_COLON2] = anon_sym_COLON,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_true] = anon_sym_true,
  [anon_sym_false] = anon_sym_false,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_mutable] = anon_sym_mutable,
  [anon_sym_constraint] = anon_sym_constraint,
  [anon_sym_exception] = anon_sym_exception,
  [anon_sym_module] = anon_sym_module,
  [anon_sym_open] = anon_sym_open,
  [anon_sym_include] = anon_sym_include,
  [anon_sym_class] = anon_sym_class,
  [anon_sym_virtual] = anon_sym_virtual,
  [anon_sym_val] = anon_sym_val,
  [anon_sym_sig] = anon_sym_sig,
  [anon_sym_end] = anon_sym_end,
  [anon_sym_with] = anon_sym_with,
  [anon_sym_functor] = anon_sym_functor,
  [anon_sym_struct] = anon_sym_struct,
  [anon_sym_COLON_GT] = anon_sym_COLON_GT,
  [anon_sym_object] = anon_sym_object,
  [anon_sym_inherit] = anon_sym_inherit,
  [anon_sym_method] = anon_sym_method,
  [anon_sym_in] = anon_sym_in,
  [anon_sym_fun] = anon_sym_fun,
  [anon_sym_as] = anon_sym_as,
  [anon_sym_initializer] = anon_sym_initializer,
  [anon_sym_LBRACK_GT] = anon_sym_LBRACK_GT,
  [anon_sym_LBRACK_LT] = anon_sym_LBRACK_LT,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_POUND] = anon_sym_POUND,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [anon_sym_LBRACK_PIPE] = anon_sym_LBRACK_PIPE,
  [anon_sym_PIPE_RBRACK] = anon_sym_PIPE_RBRACK,
  [anon_sym_LT_DASH] = anon_sym_LT_DASH,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_then] = anon_sym_then,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_while] = anon_sym_while,
  [anon_sym_do] = anon_sym_do,
  [anon_sym_done] = anon_sym_done,
  [anon_sym_for] = anon_sym_for,
  [anon_sym_to] = anon_sym_to,
  [anon_sym_downto] = anon_sym_downto,
  [anon_sym_match] = anon_sym_match,
  [anon_sym_when] = anon_sym_when,
  [anon_sym_function] = anon_sym_function,
  [anon_sym_try] = anon_sym_try,
  [anon_sym_assert] = anon_sym_assert,
  [anon_sym_lazy] = anon_sym_lazy,
  [anon_sym_new] = anon_sym_new,
  [anon_sym_LBRACE_LT] = anon_sym_LBRACE_LT,
  [anon_sym_GT_RBRACE] = anon_sym_GT_RBRACE,
  [anon_sym_begin] = anon_sym_begin,
  [sym_ocamlyacc_value] = sym_ocamlyacc_value,
  [anon_sym_effect] = anon_sym_effect,
  [aux_sym_attribute_token1] = aux_sym_attribute_token1,
  [anon_sym_LBRACK_AT_AT] = anon_sym_LBRACK_AT_AT,
  [anon_sym_LBRACK_AT_AT_AT] = anon_sym_LBRACK_AT_AT_AT,
  [anon_sym_LBRACK_PERCENT] = anon_sym_LBRACK_PERCENT,
  [anon_sym_LBRACE_PERCENT] = anon_sym_LBRACE_PERCENT,
  [aux_sym_quoted_extension_token1] = aux_sym_quoted_extension_token1,
  [anon_sym_LBRACK_PERCENT_PERCENT] = anon_sym_LBRACK_PERCENT_PERCENT,
  [anon_sym_LBRACE_PERCENT_PERCENT] = anon_sym_LBRACE_PERCENT_PERCENT,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [aux_sym_number_token1] = aux_sym_number_token1,
  [aux_sym_signed_number_token1] = aux_sym_signed_number_token1,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [anon_sym_SQUOTE2] = anon_sym_SQUOTE,
  [aux_sym_character_content_token1] = aux_sym_character_content_token1,
  [aux_sym_character_content_token2] = aux_sym_character_content_token2,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [aux_sym_string_content_token1] = aux_sym_string_content_token1,
  [aux_sym_string_content_token2] = aux_sym_string_content_token2,
  [aux_sym_string_content_token3] = aux_sym_string_content_token3,
  [aux_sym_string_content_token4] = sym_escape_sequence,
  [aux_sym_string_content_token5] = sym_escape_sequence,
  [aux_sym_quoted_string_content_token1] = aux_sym_quoted_string_content_token1,
  [aux_sym_escape_sequence_token1] = aux_sym_escape_sequence_token1,
  [aux_sym_escape_sequence_token2] = aux_sym_escape_sequence_token2,
  [aux_sym_escape_sequence_token3] = aux_sym_escape_sequence_token3,
  [aux_sym_escape_sequence_token4] = aux_sym_escape_sequence_token4,
  [sym_conversion_specification] = sym_conversion_specification,
  [sym_pretty_printing_indication] = sym_pretty_printing_indication,
  [sym_prefix_operator] = sym_prefix_operator,
  [aux_sym_sign_operator_token1] = aux_sym_sign_operator_token1,
  [sym_hash_operator] = sym_hash_operator,
  [anon_sym_lsl] = anon_sym_lsl,
  [anon_sym_lsr] = anon_sym_lsr,
  [anon_sym_asr] = anon_sym_asr,
  [aux_sym_pow_operator_token1] = aux_sym_pow_operator_token1,
  [anon_sym_mod] = anon_sym_mod,
  [anon_sym_land] = anon_sym_land,
  [anon_sym_lor] = anon_sym_lor,
  [anon_sym_lxor] = anon_sym_lxor,
  [aux_sym_mult_operator_token1] = aux_sym_mult_operator_token1,
  [aux_sym_add_operator_token1] = aux_sym_add_operator_token1,
  [sym_concat_operator] = sym_concat_operator,
  [sym_rel_operator] = sym_rel_operator,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_or] = anon_sym_or,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [sym_assign_operator] = sym_assign_operator,
  [sym_indexing_operator] = sym_indexing_operator,
  [sym_let_operator] = sym_let_operator,
  [sym_let_and_operator] = sym_let_and_operator,
  [sym_match_operator] = sym_match_operator,
  [sym__uppercase_identifier] = sym__uppercase_identifier,
  [aux_sym_directive_token1] = aux_sym_directive_token1,
  [aux_sym_type_variable_token1] = aux_sym_type_variable_token1,
  [aux_sym_tag_token1] = aux_sym_tag_token1,
  [aux_sym_attribute_id_token1] = aux_sym_attribute_id_token1,
  [sym_comment] = sym_comment,
  [sym__left_quoted_string_delimiter] = sym__left_quoted_string_delimiter,
  [sym__right_quoted_string_delimiter] = sym__right_quoted_string_delimiter,
  [sym_line_number_directive] = sym_line_number_directive,
  [sym__null] = sym__null,
  [sym__error_sentinel] = sym__error_sentinel,
  [sym_compilation_unit] = sym_compilation_unit,
  [sym__structure] = sym__structure,
  [sym_expression_item] = sym_expression_item,
  [sym__signature] = sym__signature,
  [sym_toplevel_directive] = sym_toplevel_directive,
  [sym__local_structure_item] = sym__local_structure_item,
  [sym__structure_item] = sym__structure_item,
  [sym_value_definition] = sym_value_definition,
  [sym_let_binding] = sym_let_binding,
  [sym_parameter] = sym_parameter,
  [sym_external] = sym_external,
  [sym_type_definition] = sym_type_definition,
  [sym_type_binding] = sym_type_binding,
  [sym__type_params] = sym__type_params,
  [sym__type_param] = sym__type_param,
  [sym_variant_declaration] = sym_variant_declaration,
  [sym_constructor_declaration] = sym_constructor_declaration,
  [sym__constructor_declaration_name] = sym__constructor_declaration_name,
  [sym__constructor_argument] = sym__constructor_argument,
  [sym_record_declaration] = sym_record_declaration,
  [sym_field_declaration] = sym_field_declaration,
  [sym_external_declaration] = sym_external_declaration,
  [sym_type_constraint] = sym_type_constraint,
  [sym_exception_definition] = sym_exception_definition,
  [sym_module_definition] = sym_module_definition,
  [sym_module_binding] = sym_module_binding,
  [sym_module_parameter] = sym_module_parameter,
  [sym_module_type_definition] = sym_module_type_definition,
  [sym_open_module] = sym_open_module,
  [sym_include_module] = sym_include_module,
  [sym_class_definition] = sym_class_definition,
  [sym_class_binding] = sym_class_binding,
  [sym_class_type_definition] = sym_class_type_definition,
  [sym_class_type_binding] = sym_class_type_binding,
  [sym__signature_item] = sym__signature_item,
  [sym_value_specification] = sym_value_specification,
  [sym_include_module_type] = sym_include_module_type,
  [sym__module_typed] = sym__module_typed,
  [sym__module_type] = sym__module_type,
  [sym_signature] = sym_signature,
  [sym_module_type_constraint] = sym_module_type_constraint,
  [sym_constrain_type] = sym_constrain_type,
  [sym_constrain_module] = sym_constrain_module,
  [sym_constrain_module_type] = sym_constrain_module_type,
  [sym_module_type_of] = sym_module_type_of,
  [sym_functor_type] = sym_functor_type,
  [sym_parenthesized_module_type] = sym_parenthesized_module_type,
  [sym__simple_module_expression] = sym__simple_module_expression,
  [sym__module_expression] = sym__module_expression,
  [sym_structure] = sym_structure,
  [sym_functor] = sym_functor,
  [sym_module_application] = sym_module_application,
  [sym_typed_module_expression] = sym_typed_module_expression,
  [sym_packed_module] = sym_packed_module,
  [sym_parenthesized_module_expression] = sym_parenthesized_module_expression,
  [sym__class_typed] = sym__class_typed,
  [sym__simple_class_type] = sym__simple_class_type,
  [sym__class_type] = sym__class_type,
  [sym_instantiated_class_type] = sym_instantiated_class_type,
  [sym_class_body_type] = sym_class_body_type,
  [sym__class_field_specification] = sym__class_field_specification,
  [sym_inheritance_specification] = sym_inheritance_specification,
  [sym_instance_variable_specification] = sym_instance_variable_specification,
  [sym_method_specification] = sym_method_specification,
  [sym_type_parameter_constraint] = sym_type_parameter_constraint,
  [sym_let_open_class_type] = sym_let_open_class_type,
  [sym_class_function_type] = sym_class_function_type,
  [sym__simple_class_expression] = sym__simple_class_expression,
  [sym__class_expression] = sym__class_expression,
  [sym_instantiated_class] = sym_instantiated_class,
  [sym_typed_class_expression] = sym_typed_class_expression,
  [sym_class_function] = sym_class_function,
  [sym_class_application] = sym_class_application,
  [sym_let_class_expression] = sym_let_class_expression,
  [sym__class_field] = sym__class_field,
  [sym_inheritance_definition] = sym_inheritance_definition,
  [sym_instance_variable_definition] = sym_instance_variable_definition,
  [sym_method_definition] = sym_method_definition,
  [sym_class_initializer] = sym_class_initializer,
  [sym_let_open_class_expression] = sym_let_open_class_expression,
  [sym_parenthesized_class_expression] = sym_parenthesized_class_expression,
  [sym__typed] = sym__typed,
  [sym__simple_typed] = sym__simple_typed,
  [sym__coerced] = sym__coerced,
  [sym__type_constrained] = sym__type_constrained,
  [sym__maybe_polymorphic_typed] = sym__maybe_polymorphic_typed,
  [sym__polymorphic_typed] = sym__polymorphic_typed,
  [sym__polymorphic_type] = sym__polymorphic_type,
  [sym_polymorphic_type] = sym_polymorphic_type,
  [sym__abstract_type] = sym__abstract_type,
  [sym__parenthesized_abstract_type] = sym__parenthesized_abstract_type,
  [sym__delimited_type] = sym__delimited_type,
  [sym__simple_type] = sym__simple_type,
  [sym__type] = sym__type,
  [sym_function_type] = sym_function_type,
  [sym_labeled_argument_type] = sym_labeled_argument_type,
  [sym__proper_tuple_type] = sym__proper_tuple_type,
  [sym__labeled_tuple_type] = sym__proper_tuple_type,
  [sym_labeled_tuple_element_type] = sym_labeled_tuple_element_type,
  [sym__tuple_type_rhs] = sym__tuple_type_rhs,
  [sym_constructed_type] = sym_constructed_type,
  [sym_aliased_type] = sym_aliased_type,
  [sym_local_open_type] = sym_local_open_type,
  [sym_polymorphic_variant_type] = sym_polymorphic_variant_type,
  [sym__tag_spec] = sym__tag_spec,
  [sym_tag_specification] = sym_tag_specification,
  [sym_package_type] = sym_package_type,
  [sym_object_type] = sym_object_type,
  [sym_method_type] = sym_method_type,
  [sym_hash_type] = sym_hash_type,
  [sym_parenthesized_type] = sym_parenthesized_type,
  [sym__delimited_expression] = sym__delimited_expression,
  [sym__simple_expression] = sym__simple_expression,
  [sym__expression] = sym__expression,
  [sym__sequence_expression] = sym__sequence_expression,
  [sym_typed_expression] = sym_typed_expression,
  [sym_labeled_tuple_element] = sym_labeled_tuple_element,
  [sym__tuple_expression] = sym__tuple_expression,
  [sym_cons_expression] = sym_cons_expression,
  [sym_list_expression] = sym_list_expression,
  [sym_array_expression] = sym_array_expression,
  [sym__sequence_expression_content] = sym__sequence_expression_content,
  [sym_record_expression] = sym_record_expression,
  [sym_field_expression] = sym_field_expression,
  [sym_application_expression] = sym_application_expression,
  [sym_labeled_argument] = sym_labeled_argument,
  [sym_prefix_expression] = sym_prefix_expression,
  [sym_sign_expression] = sym_sign_expression,
  [sym_hash_expression] = sym_hash_expression,
  [sym_infix_expression] = sym_infix_expression,
  [sym_field_get_expression] = sym_field_get_expression,
  [sym_array_get_expression] = sym_array_get_expression,
  [sym_string_get_expression] = sym_string_get_expression,
  [sym_bigarray_get_expression] = sym_bigarray_get_expression,
  [sym_set_expression] = sym_set_expression,
  [sym_if_expression] = sym_if_expression,
  [sym_then_clause] = sym_then_clause,
  [sym_else_clause] = sym_else_clause,
  [sym_while_expression] = sym_while_expression,
  [sym_do_clause] = sym_do_clause,
  [sym_for_expression] = sym_for_expression,
  [sym__sequence_expression_anonymous] = sym__sequence_expression_anonymous,
  [sym_match_expression] = sym_match_expression,
  [sym__match_cases] = sym__match_cases,
  [sym_match_case] = sym_match_case,
  [sym_guard] = sym_guard,
  [sym_refutation_case] = sym_refutation_case,
  [sym_function_expression] = sym_function_expression,
  [sym_fun_expression] = sym_fun_expression,
  [sym_try_expression] = sym_try_expression,
  [sym_let_expression] = sym_let_expression,
  [sym_assert_expression] = sym_assert_expression,
  [sym_lazy_expression] = sym_lazy_expression,
  [sym_local_open_expression] = sym_local_open_expression,
  [sym_package_expression] = sym_package_expression,
  [sym_new_expression] = sym_new_expression,
  [sym_object_copy_expression] = sym_object_copy_expression,
  [sym_instance_variable_expression] = sym_instance_variable_expression,
  [sym_method_invocation] = sym_method_invocation,
  [sym_object_expression] = sym_object_expression,
  [sym_parenthesized_expression] = sym_parenthesized_expression,
  [sym__delimited_pattern] = sym__delimited_pattern,
  [sym__simple_pattern] = sym__simple_pattern,
  [sym__effect_pattern] = sym__effect_pattern,
  [sym__pattern] = sym__pattern,
  [sym__delimited_binding_pattern] = sym__delimited_binding_pattern,
  [sym__simple_binding_pattern] = sym__simple_binding_pattern,
  [sym__effect_binding_pattern] = sym__effect_binding_pattern,
  [sym__binding_pattern_no_exn] = sym__binding_pattern_no_exn,
  [sym__binding_pattern] = sym__binding_pattern,
  [sym_alias_pattern] = sym_alias_pattern,
  [sym_alias_binding_pattern_no_exn] = sym_alias_pattern,
  [sym_alias_binding_pattern] = sym_alias_pattern,
  [sym_typed_pattern] = sym_typed_pattern,
  [sym_typed_binding_pattern] = sym_typed_pattern,
  [sym__or_pattern_anonymous] = sym__or_pattern_anonymous,
  [sym__or_binding_pattern_no_exn_anonymous] = sym__or_binding_pattern_no_exn_anonymous,
  [sym__or_binding_pattern_anonymous] = sym__or_binding_pattern_anonymous,
  [sym_constructor_pattern] = sym_constructor_pattern,
  [sym_constructor_binding_pattern] = sym_constructor_pattern,
  [sym_tag_pattern] = sym_tag_pattern,
  [sym_tag_binding_pattern] = sym_tag_pattern,
  [sym_polymorphic_variant_pattern] = sym_polymorphic_variant_pattern,
  [sym_labeled_tuple_element_pattern] = sym_labeled_tuple_element_pattern,
  [sym__tuple_pattern] = sym__tuple_pattern,
  [sym_labeled_tuple_element_binding_pattern] = sym_labeled_tuple_element_binding_pattern,
  [sym__tuple_binding_pattern_no_exn] = sym__tuple_binding_pattern_no_exn,
  [sym__tuple_binding_pattern] = sym__tuple_binding_pattern,
  [sym_record_pattern] = sym_record_pattern,
  [sym_field_pattern] = sym_field_pattern,
  [sym_record_binding_pattern] = sym_record_pattern,
  [sym_field_binding_pattern] = sym_field_pattern,
  [sym_list_pattern] = sym_list_pattern,
  [sym_list_binding_pattern] = sym_list_pattern,
  [sym_cons_pattern] = sym_cons_pattern,
  [sym_cons_binding_pattern_no_exn] = sym_cons_pattern,
  [sym_cons_binding_pattern] = sym_cons_pattern,
  [sym_array_pattern] = sym_array_pattern,
  [sym_array_binding_pattern] = sym_array_pattern,
  [sym__sequence_pattern_content] = sym__sequence_pattern_content,
  [sym__sequence_binding_pattern_content] = sym__sequence_binding_pattern_content,
  [sym_range_pattern] = sym_range_pattern,
  [sym_lazy_pattern] = sym_lazy_pattern,
  [sym_lazy_binding_pattern] = sym_lazy_pattern,
  [sym_local_open_pattern] = sym_local_open_pattern,
  [sym_local_open_binding_pattern] = sym_local_open_pattern,
  [sym_package_pattern] = sym_package_pattern,
  [sym_parenthesized_pattern] = sym_parenthesized_pattern,
  [sym_parenthesized_binding_pattern] = sym_parenthesized_pattern,
  [sym_exception_pattern] = sym_exception_pattern,
  [sym_exception_binding_pattern] = sym_exception_pattern,
  [sym_effect_pattern] = sym_effect_pattern,
  [sym_effect_binding_pattern] = sym_effect_pattern,
  [sym_attribute] = sym_attribute,
  [sym_item_attribute] = sym_item_attribute,
  [sym_floating_attribute] = sym_floating_attribute,
  [sym_attribute_payload] = sym_attribute_payload,
  [sym_extension] = sym_extension,
  [sym_quoted_extension] = sym_quoted_extension,
  [sym_item_extension] = sym_item_extension,
  [sym_quoted_item_extension] = sym_quoted_item_extension,
  [sym__attribute] = sym__attribute,
  [sym__constant] = sym__constant,
  [sym__signed_constant] = sym__signed_constant,
  [sym_number] = sym_number,
  [sym_signed_number] = sym_signed_number,
  [sym_character] = sym_character,
  [sym_character_content] = sym_character_content,
  [sym_string] = sym_string,
  [sym_string_content] = sym_string_content,
  [sym_quoted_string] = sym_quoted_string,
  [sym__quoted_string] = sym__quoted_string,
  [sym_quoted_string_content] = sym_quoted_string_content,
  [sym_escape_sequence] = sym_escape_sequence,
  [sym_boolean] = sym_boolean,
  [sym_unit] = sym_unit,
  [sym_sign_operator] = sym_sign_operator,
  [sym__infix_operator] = sym__infix_operator,
  [sym_pow_operator] = sym_pow_operator,
  [sym_mult_operator] = sym_mult_operator,
  [sym_add_operator] = sym_add_operator,
  [sym_and_operator] = sym_and_operator,
  [sym_or_operator] = sym_or_operator,
  [sym_indexing_operator_path] = sym_indexing_operator_path,
  [sym__value_name] = sym__value_name,
  [sym__simple_value_pattern] = sym__simple_value_pattern,
  [sym_parenthesized_operator] = sym_parenthesized_operator,
  [sym_value_path] = sym_value_path,
  [sym_module_path] = sym_module_path,
  [sym_extended_module_path] = sym_extended_module_path,
  [sym_module_type_path] = sym_module_type_path,
  [sym_field_path] = sym_field_path,
  [sym_constructor_path] = sym_constructor_path,
  [sym_type_constructor_path] = sym_type_constructor_path,
  [sym_class_path] = sym_class_path,
  [sym_class_type_path] = sym_class_type_path,
  [sym__instance_variable_name] = sym__instance_variable_name,
  [sym__constructor_name] = sym__constructor_name,
  [sym_directive] = sym_directive,
  [sym_type_variable] = sym_type_variable,
  [sym_tag] = sym_tag,
  [sym_attribute_id] = sym_attribute_id,
  [aux_sym__structure_repeat1] = aux_sym__structure_repeat1,
  [aux_sym__structure_repeat2] = aux_sym__structure_repeat2,
  [aux_sym_expression_item_repeat1] = aux_sym_expression_item_repeat1,
  [aux_sym__signature_repeat1] = aux_sym__signature_repeat1,
  [aux_sym_value_definition_repeat1] = aux_sym_value_definition_repeat1,
  [aux_sym_let_binding_repeat1] = aux_sym_let_binding_repeat1,
  [aux_sym_external_repeat1] = aux_sym_external_repeat1,
  [aux_sym_type_definition_repeat1] = aux_sym_type_definition_repeat1,
  [aux_sym_type_binding_repeat1] = aux_sym_type_binding_repeat1,
  [aux_sym__type_params_repeat1] = aux_sym__type_params_repeat1,
  [aux_sym__type_param_repeat1] = aux_sym__type_param_repeat1,
  [aux_sym_variant_declaration_repeat1] = aux_sym_variant_declaration_repeat1,
  [aux_sym_constructor_declaration_repeat1] = aux_sym_constructor_declaration_repeat1,
  [aux_sym__constructor_argument_repeat1] = aux_sym__constructor_argument_repeat1,
  [aux_sym_record_declaration_repeat1] = aux_sym_record_declaration_repeat1,
  [aux_sym_module_definition_repeat1] = aux_sym_module_definition_repeat1,
  [aux_sym_module_binding_repeat1] = aux_sym_module_binding_repeat1,
  [aux_sym_class_definition_repeat1] = aux_sym_class_definition_repeat1,
  [aux_sym_class_type_definition_repeat1] = aux_sym_class_type_definition_repeat1,
  [aux_sym_module_type_constraint_repeat1] = aux_sym_module_type_constraint_repeat1,
  [aux_sym_instantiated_class_type_repeat1] = aux_sym_instantiated_class_type_repeat1,
  [aux_sym_class_body_type_repeat1] = aux_sym_class_body_type_repeat1,
  [aux_sym_instance_variable_specification_repeat1] = aux_sym_instance_variable_specification_repeat1,
  [aux_sym_method_specification_repeat1] = aux_sym_method_specification_repeat1,
  [aux_sym_class_application_repeat1] = aux_sym_class_application_repeat1,
  [aux_sym__abstract_type_repeat1] = aux_sym__abstract_type_repeat1,
  [aux_sym_polymorphic_variant_type_repeat1] = aux_sym_polymorphic_variant_type_repeat1,
  [aux_sym_polymorphic_variant_type_repeat2] = aux_sym_polymorphic_variant_type_repeat2,
  [aux_sym_tag_specification_repeat1] = aux_sym_tag_specification_repeat1,
  [aux_sym_object_type_repeat1] = aux_sym_object_type_repeat1,
  [aux_sym__sequence_expression_content_repeat1] = aux_sym__sequence_expression_content_repeat1,
  [aux_sym_record_expression_repeat1] = aux_sym_record_expression_repeat1,
  [aux_sym__match_cases_repeat1] = aux_sym__match_cases_repeat1,
  [aux_sym_object_copy_expression_repeat1] = aux_sym_object_copy_expression_repeat1,
  [aux_sym_object_expression_repeat1] = aux_sym_object_expression_repeat1,
  [aux_sym_record_pattern_repeat1] = aux_sym_record_pattern_repeat1,
  [aux_sym_record_binding_pattern_repeat1] = aux_sym_record_binding_pattern_repeat1,
  [aux_sym__sequence_pattern_content_repeat1] = aux_sym__sequence_pattern_content_repeat1,
  [aux_sym__sequence_binding_pattern_content_repeat1] = aux_sym__sequence_binding_pattern_content_repeat1,
  [aux_sym_string_content_repeat1] = aux_sym_string_content_repeat1,
  [aux_sym_quoted_string_content_repeat1] = aux_sym_quoted_string_content_repeat1,
  [aux_sym_attribute_id_repeat1] = aux_sym_attribute_id_repeat1,
  [alias_sym_class_name] = alias_sym_class_name,
  [alias_sym_class_type_name] = alias_sym_class_type_name,
  [alias_sym_field_name] = alias_sym_field_name,
  [alias_sym_instance_variable_name] = alias_sym_instance_variable_name,
  [alias_sym_label_name] = alias_sym_label_name,
  [alias_sym_method_name] = alias_sym_method_name,
  [alias_sym_module_name] = alias_sym_module_name,
  [alias_sym_module_type_name] = alias_sym_module_type_name,
  [alias_sym_sequence_expression] = alias_sym_sequence_expression,
  [alias_sym_tuple_expression] = alias_sym_tuple_expression,
  [alias_sym_type_constructor] = alias_sym_type_constructor,
  [alias_sym_value_name] = alias_sym_value_name,
  [alias_sym_value_pattern] = alias_sym_value_pattern,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym__lowercase_identifier] = {
    .visible = false,
    .named = true,
  },
  [sym_shebang] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_SEMI_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_let] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_rec] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_and] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_external] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_type] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_nonrec] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_private] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_of] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_true] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_false] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_mutable] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_constraint] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_exception] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_module] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_open] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_include] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_class] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_virtual] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_val] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_sig] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_end] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_with] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_functor] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_struct] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_object] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_inherit] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_method] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_in] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_fun] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_as] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_initializer] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUND] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_then] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_while] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_do] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_done] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_for] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_to] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_downto] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_match] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_when] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_function] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_try] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_assert] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lazy] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_new] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_begin] = {
    .visible = true,
    .named = false,
  },
  [sym_ocamlyacc_value] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_effect] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_attribute_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_AT_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_AT_AT_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_quoted_extension_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_LBRACK_PERCENT_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE_PERCENT_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_number_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_signed_number_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_character_content_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_character_content_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_string_content_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_content_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_content_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_content_token4] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_string_content_token5] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_quoted_string_content_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_escape_sequence_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_escape_sequence_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_escape_sequence_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_escape_sequence_token4] = {
    .visible = false,
    .named = false,
  },
  [sym_conversion_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_pretty_printing_indication] = {
    .visible = true,
    .named = true,
  },
  [sym_prefix_operator] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_sign_operator_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_hash_operator] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_lsl] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lsr] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_asr] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_pow_operator_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_mod] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_land] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lor] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lxor] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_mult_operator_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_add_operator_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_concat_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_rel_operator] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_or] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [sym_assign_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_indexing_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_let_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_let_and_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_match_operator] = {
    .visible = true,
    .named = true,
  },
  [sym__uppercase_identifier] = {
    .visible = false,
    .named = true,
  },
  [aux_sym_directive_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_variable_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_tag_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_id_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym__left_quoted_string_delimiter] = {
    .visible = false,
    .named = true,
  },
  [sym__right_quoted_string_delimiter] = {
    .visible = false,
    .named = true,
  },
  [sym_line_number_directive] = {
    .visible = true,
    .named = true,
  },
  [sym__null] = {
    .visible = false,
    .named = true,
  },
  [sym__error_sentinel] = {
    .visible = false,
    .named = true,
  },
  [sym_compilation_unit] = {
    .visible = true,
    .named = true,
  },
  [sym__structure] = {
    .visible = false,
    .named = true,
  },
  [sym_expression_item] = {
    .visible = true,
    .named = true,
  },
  [sym__signature] = {
    .visible = false,
    .named = true,
  },
  [sym_toplevel_directive] = {
    .visible = true,
    .named = true,
  },
  [sym__local_structure_item] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__structure_item] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_value_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_let_binding] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_external] = {
    .visible = true,
    .named = true,
  },
  [sym_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_type_binding] = {
    .visible = true,
    .named = true,
  },
  [sym__type_params] = {
    .visible = false,
    .named = true,
  },
  [sym__type_param] = {
    .visible = false,
    .named = true,
  },
  [sym_variant_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_constructor_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__constructor_declaration_name] = {
    .visible = true,
    .named = true,
  },
  [sym__constructor_argument] = {
    .visible = false,
    .named = true,
  },
  [sym_record_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_field_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_external_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_type_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_exception_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_module_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_module_binding] = {
    .visible = true,
    .named = true,
  },
  [sym_module_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_module_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_open_module] = {
    .visible = true,
    .named = true,
  },
  [sym_include_module] = {
    .visible = true,
    .named = true,
  },
  [sym_class_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_class_binding] = {
    .visible = true,
    .named = true,
  },
  [sym_class_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_class_type_binding] = {
    .visible = true,
    .named = true,
  },
  [sym__signature_item] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_value_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_include_module_type] = {
    .visible = true,
    .named = true,
  },
  [sym__module_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__module_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_signature] = {
    .visible = true,
    .named = true,
  },
  [sym_module_type_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_constrain_type] = {
    .visible = true,
    .named = true,
  },
  [sym_constrain_module] = {
    .visible = true,
    .named = true,
  },
  [sym_constrain_module_type] = {
    .visible = true,
    .named = true,
  },
  [sym_module_type_of] = {
    .visible = true,
    .named = true,
  },
  [sym_functor_type] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_module_type] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_module_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__module_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_structure] = {
    .visible = true,
    .named = true,
  },
  [sym_functor] = {
    .visible = true,
    .named = true,
  },
  [sym_module_application] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_module_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_packed_module] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_module_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__class_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_class_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__class_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_instantiated_class_type] = {
    .visible = true,
    .named = true,
  },
  [sym_class_body_type] = {
    .visible = true,
    .named = true,
  },
  [sym__class_field_specification] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_inheritance_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_instance_variable_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_method_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_type_parameter_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_let_open_class_type] = {
    .visible = true,
    .named = true,
  },
  [sym_class_function_type] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_class_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__class_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_instantiated_class] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_class_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_class_function] = {
    .visible = true,
    .named = true,
  },
  [sym_class_application] = {
    .visible = true,
    .named = true,
  },
  [sym_let_class_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__class_field] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_inheritance_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_instance_variable_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_method_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_class_initializer] = {
    .visible = true,
    .named = true,
  },
  [sym_let_open_class_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_class_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__typed] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__coerced] = {
    .visible = false,
    .named = true,
  },
  [sym__type_constrained] = {
    .visible = false,
    .named = true,
  },
  [sym__maybe_polymorphic_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__polymorphic_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__polymorphic_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_polymorphic_type] = {
    .visible = true,
    .named = true,
  },
  [sym__abstract_type] = {
    .visible = false,
    .named = true,
  },
  [sym__parenthesized_abstract_type] = {
    .visible = true,
    .named = true,
  },
  [sym__delimited_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__simple_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_function_type] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_argument_type] = {
    .visible = true,
    .named = true,
  },
  [sym__proper_tuple_type] = {
    .visible = true,
    .named = true,
  },
  [sym__labeled_tuple_type] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_tuple_element_type] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_type_rhs] = {
    .visible = false,
    .named = true,
  },
  [sym_constructed_type] = {
    .visible = true,
    .named = true,
  },
  [sym_aliased_type] = {
    .visible = true,
    .named = true,
  },
  [sym_local_open_type] = {
    .visible = true,
    .named = true,
  },
  [sym_polymorphic_variant_type] = {
    .visible = true,
    .named = true,
  },
  [sym__tag_spec] = {
    .visible = false,
    .named = true,
  },
  [sym_tag_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_package_type] = {
    .visible = true,
    .named = true,
  },
  [sym_object_type] = {
    .visible = true,
    .named = true,
  },
  [sym_method_type] = {
    .visible = true,
    .named = true,
  },
  [sym_hash_type] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_type] = {
    .visible = true,
    .named = true,
  },
  [sym__delimited_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__simple_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__sequence_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_typed_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_tuple_element] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_cons_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_list_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_array_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__sequence_expression_content] = {
    .visible = false,
    .named = true,
  },
  [sym_record_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_field_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_application_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_prefix_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_sign_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hash_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_infix_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_field_get_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_array_get_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_string_get_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_bigarray_get_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_set_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_if_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_then_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_else_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_while_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_do_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_for_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__sequence_expression_anonymous] = {
    .visible = false,
    .named = true,
  },
  [sym_match_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__match_cases] = {
    .visible = false,
    .named = true,
  },
  [sym_match_case] = {
    .visible = true,
    .named = true,
  },
  [sym_guard] = {
    .visible = true,
    .named = true,
  },
  [sym_refutation_case] = {
    .visible = true,
    .named = true,
  },
  [sym_function_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_fun_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_try_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_let_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_assert_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_lazy_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_local_open_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_package_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_new_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_object_copy_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_instance_variable_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_method_invocation] = {
    .visible = true,
    .named = true,
  },
  [sym_object_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__delimited_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__simple_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__effect_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__delimited_binding_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__simple_binding_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__effect_binding_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__binding_pattern_no_exn] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__binding_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_alias_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_alias_binding_pattern_no_exn] = {
    .visible = true,
    .named = true,
  },
  [sym_alias_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__or_pattern_anonymous] = {
    .visible = false,
    .named = true,
  },
  [sym__or_binding_pattern_no_exn_anonymous] = {
    .visible = true,
    .named = true,
  },
  [sym__or_binding_pattern_anonymous] = {
    .visible = false,
    .named = true,
  },
  [sym_constructor_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_constructor_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_tag_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_tag_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_polymorphic_variant_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_tuple_element_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_pattern] = {
    .visible = false,
    .named = true,
  },
  [sym_labeled_tuple_element_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_binding_pattern_no_exn] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_binding_pattern] = {
    .visible = false,
    .named = true,
  },
  [sym_record_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_field_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_record_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_field_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_list_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_list_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_cons_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_cons_binding_pattern_no_exn] = {
    .visible = true,
    .named = true,
  },
  [sym_cons_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_array_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_array_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__sequence_pattern_content] = {
    .visible = false,
    .named = true,
  },
  [sym__sequence_binding_pattern_content] = {
    .visible = false,
    .named = true,
  },
  [sym_range_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_lazy_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_lazy_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_local_open_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_local_open_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_package_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_exception_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_exception_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_effect_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_effect_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_item_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_floating_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_payload] = {
    .visible = true,
    .named = true,
  },
  [sym_extension] = {
    .visible = true,
    .named = true,
  },
  [sym_quoted_extension] = {
    .visible = true,
    .named = true,
  },
  [sym_item_extension] = {
    .visible = true,
    .named = true,
  },
  [sym_quoted_item_extension] = {
    .visible = true,
    .named = true,
  },
  [sym__attribute] = {
    .visible = false,
    .named = true,
  },
  [sym__constant] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__signed_constant] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_number] = {
    .visible = true,
    .named = true,
  },
  [sym_signed_number] = {
    .visible = true,
    .named = true,
  },
  [sym_character] = {
    .visible = true,
    .named = true,
  },
  [sym_character_content] = {
    .visible = true,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym_string_content] = {
    .visible = true,
    .named = true,
  },
  [sym_quoted_string] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_string] = {
    .visible = false,
    .named = true,
  },
  [sym_quoted_string_content] = {
    .visible = true,
    .named = true,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean] = {
    .visible = true,
    .named = true,
  },
  [sym_unit] = {
    .visible = true,
    .named = true,
  },
  [sym_sign_operator] = {
    .visible = true,
    .named = true,
  },
  [sym__infix_operator] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_pow_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_mult_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_add_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_and_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_or_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_indexing_operator_path] = {
    .visible = true,
    .named = true,
  },
  [sym__value_name] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_value_pattern] = {
    .visible = false,
    .named = true,
  },
  [sym_parenthesized_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_value_path] = {
    .visible = true,
    .named = true,
  },
  [sym_module_path] = {
    .visible = true,
    .named = true,
  },
  [sym_extended_module_path] = {
    .visible = true,
    .named = true,
  },
  [sym_module_type_path] = {
    .visible = true,
    .named = true,
  },
  [sym_field_path] = {
    .visible = true,
    .named = true,
  },
  [sym_constructor_path] = {
    .visible = true,
    .named = true,
  },
  [sym_type_constructor_path] = {
    .visible = true,
    .named = true,
  },
  [sym_class_path] = {
    .visible = true,
    .named = true,
  },
  [sym_class_type_path] = {
    .visible = true,
    .named = true,
  },
  [sym__instance_variable_name] = {
    .visible = false,
    .named = true,
  },
  [sym__constructor_name] = {
    .visible = false,
    .named = true,
  },
  [sym_directive] = {
    .visible = true,
    .named = true,
  },
  [sym_type_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_tag] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_id] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__structure_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__structure_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expression_item_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__signature_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_value_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_let_binding_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_external_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_binding_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__type_params_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__type_param_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_variant_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_constructor_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__constructor_argument_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_module_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_module_binding_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_type_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_module_type_constraint_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_instantiated_class_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_body_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_instance_variable_specification_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_method_specification_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_application_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__abstract_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_polymorphic_variant_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_polymorphic_variant_type_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_tag_specification_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_object_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__sequence_expression_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__match_cases_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_object_copy_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_object_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_binding_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__sequence_pattern_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__sequence_binding_pattern_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_quoted_string_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_id_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_class_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_class_type_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_field_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_instance_variable_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_label_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_method_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_module_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_module_type_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_sequence_expression] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_tuple_expression] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_type_constructor] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_value_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_value_pattern] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_alias = 1,
  field_argument = 2,
  field_array = 3,
  field_body = 4,
  field_class = 5,
  field_class_type = 6,
  field_codomain = 7,
  field_coercion = 8,
  field_condition = 9,
  field_constraint = 10,
  field_continuation = 11,
  field_default = 12,
  field_domain = 13,
  field_effect = 14,
  field_equation = 15,
  field_expression = 16,
  field_field = 17,
  field_from = 18,
  field_function = 19,
  field_functor = 20,
  field_index = 21,
  field_initializer = 22,
  field_left = 23,
  field_method = 24,
  field_module = 25,
  field_module_type = 26,
  field_name = 27,
  field_object = 28,
  field_operator = 29,
  field_pattern = 30,
  field_record = 31,
  field_right = 32,
  field_self = 33,
  field_self_type = 34,
  field_string = 35,
  field_synonym = 36,
  field_to = 37,
  field_type = 38,
  field_value = 39,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alias] = "alias",
  [field_argument] = "argument",
  [field_array] = "array",
  [field_body] = "body",
  [field_class] = "class",
  [field_class_type] = "class_type",
  [field_codomain] = "codomain",
  [field_coercion] = "coercion",
  [field_condition] = "condition",
  [field_constraint] = "constraint",
  [field_continuation] = "continuation",
  [field_default] = "default",
  [field_domain] = "domain",
  [field_effect] = "effect",
  [field_equation] = "equation",
  [field_expression] = "expression",
  [field_field] = "field",
  [field_from] = "from",
  [field_function] = "function",
  [field_functor] = "functor",
  [field_index] = "index",
  [field_initializer] = "initializer",
  [field_left] = "left",
  [field_method] = "method",
  [field_module] = "module",
  [field_module_type] = "module_type",
  [field_name] = "name",
  [field_object] = "object",
  [field_operator] = "operator",
  [field_pattern] = "pattern",
  [field_record] = "record",
  [field_right] = "right",
  [field_self] = "self",
  [field_self_type] = "self_type",
  [field_string] = "string",
  [field_synonym] = "synonym",
  [field_to] = "to",
  [field_type] = "type",
  [field_value] = "value",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [7] = {.index = 0, .length = 1},
  [10] = {.index = 1, .length = 1},
  [13] = {.index = 2, .length = 1},
  [18] = {.index = 3, .length = 1},
  [19] = {.index = 4, .length = 2},
  [20] = {.index = 6, .length = 1},
  [21] = {.index = 7, .length = 2},
  [22] = {.index = 9, .length = 1},
  [24] = {.index = 10, .length = 1},
  [25] = {.index = 11, .length = 1},
  [26] = {.index = 12, .length = 1},
  [28] = {.index = 13, .length = 2},
  [30] = {.index = 15, .length = 1},
  [31] = {.index = 16, .length = 1},
  [32] = {.index = 17, .length = 2},
  [34] = {.index = 19, .length = 1},
  [35] = {.index = 20, .length = 1},
  [36] = {.index = 21, .length = 1},
  [38] = {.index = 22, .length = 1},
  [39] = {.index = 23, .length = 1},
  [40] = {.index = 24, .length = 1},
  [41] = {.index = 25, .length = 1},
  [42] = {.index = 26, .length = 2},
  [43] = {.index = 28, .length = 2},
  [44] = {.index = 30, .length = 3},
  [45] = {.index = 33, .length = 2},
  [46] = {.index = 35, .length = 2},
  [48] = {.index = 37, .length = 1},
  [49] = {.index = 38, .length = 1},
  [50] = {.index = 39, .length = 1},
  [51] = {.index = 40, .length = 2},
  [52] = {.index = 42, .length = 1},
  [53] = {.index = 43, .length = 1},
  [54] = {.index = 44, .length = 2},
  [55] = {.index = 46, .length = 1},
  [56] = {.index = 47, .length = 2},
  [57] = {.index = 49, .length = 3},
  [58] = {.index = 52, .length = 2},
  [60] = {.index = 52, .length = 2},
  [63] = {.index = 25, .length = 1},
  [65] = {.index = 54, .length = 1},
  [66] = {.index = 55, .length = 1},
  [67] = {.index = 56, .length = 1},
  [68] = {.index = 57, .length = 1},
  [69] = {.index = 58, .length = 1},
  [70] = {.index = 25, .length = 1},
  [72] = {.index = 59, .length = 1},
  [73] = {.index = 58, .length = 1},
  [74] = {.index = 60, .length = 1},
  [75] = {.index = 61, .length = 2},
  [76] = {.index = 63, .length = 1},
  [77] = {.index = 64, .length = 2},
  [79] = {.index = 20, .length = 1},
  [80] = {.index = 66, .length = 1},
  [82] = {.index = 67, .length = 1},
  [83] = {.index = 68, .length = 2},
  [84] = {.index = 70, .length = 3},
  [85] = {.index = 73, .length = 3},
  [86] = {.index = 76, .length = 2},
  [87] = {.index = 78, .length = 1},
  [88] = {.index = 79, .length = 2},
  [89] = {.index = 81, .length = 2},
  [90] = {.index = 83, .length = 2},
  [91] = {.index = 85, .length = 2},
  [92] = {.index = 85, .length = 2},
  [93] = {.index = 83, .length = 2},
  [94] = {.index = 87, .length = 1},
  [95] = {.index = 88, .length = 3},
  [96] = {.index = 78, .length = 1},
  [97] = {.index = 91, .length = 1},
  [98] = {.index = 92, .length = 2},
  [99] = {.index = 39, .length = 1},
  [100] = {.index = 94, .length = 1},
  [101] = {.index = 95, .length = 2},
  [102] = {.index = 25, .length = 1},
  [104] = {.index = 39, .length = 1},
  [105] = {.index = 97, .length = 1},
  [106] = {.index = 98, .length = 2},
  [107] = {.index = 39, .length = 1},
  [108] = {.index = 100, .length = 1},
  [109] = {.index = 37, .length = 1},
  [110] = {.index = 101, .length = 1},
  [111] = {.index = 102, .length = 2},
  [112] = {.index = 104, .length = 2},
  [113] = {.index = 106, .length = 2},
  [114] = {.index = 101, .length = 1},
  [115] = {.index = 79, .length = 2},
  [116] = {.index = 108, .length = 2},
  [117] = {.index = 60, .length = 1},
  [119] = {.index = 39, .length = 1},
  [120] = {.index = 60, .length = 1},
  [121] = {.index = 110, .length = 2},
  [122] = {.index = 112, .length = 2},
  [123] = {.index = 114, .length = 2},
  [124] = {.index = 116, .length = 2},
  [125] = {.index = 118, .length = 1},
  [126] = {.index = 60, .length = 1},
  [127] = {.index = 119, .length = 1},
  [129] = {.index = 120, .length = 2},
  [130] = {.index = 101, .length = 1},
  [131] = {.index = 122, .length = 4},
  [132] = {.index = 126, .length = 3},
  [133] = {.index = 129, .length = 3},
  [134] = {.index = 132, .length = 2},
  [135] = {.index = 37, .length = 1},
  [136] = {.index = 134, .length = 3},
  [137] = {.index = 137, .length = 2},
  [138] = {.index = 137, .length = 2},
  [139] = {.index = 139, .length = 1},
  [140] = {.index = 55, .length = 1},
  [141] = {.index = 140, .length = 1},
  [142] = {.index = 141, .length = 1},
  [143] = {.index = 142, .length = 2},
  [144] = {.index = 144, .length = 2},
  [145] = {.index = 39, .length = 1},
  [146] = {.index = 146, .length = 1},
  [148] = {.index = 147, .length = 2},
  [149] = {.index = 78, .length = 1},
  [151] = {.index = 147, .length = 2},
  [152] = {.index = 149, .length = 2},
  [153] = {.index = 151, .length = 1},
  [154] = {.index = 152, .length = 2},
  [155] = {.index = 154, .length = 3},
  [156] = {.index = 157, .length = 2},
  [157] = {.index = 159, .length = 2},
  [158] = {.index = 78, .length = 1},
  [159] = {.index = 161, .length = 1},
  [160] = {.index = 161, .length = 1},
  [161] = {.index = 110, .length = 2},
  [162] = {.index = 78, .length = 1},
  [163] = {.index = 162, .length = 1},
  [164] = {.index = 163, .length = 2},
  [165] = {.index = 165, .length = 2},
  [166] = {.index = 167, .length = 3},
  [167] = {.index = 170, .length = 3},
  [168] = {.index = 173, .length = 1},
  [169] = {.index = 174, .length = 4},
  [170] = {.index = 20, .length = 1},
  [171] = {.index = 178, .length = 3},
  [172] = {.index = 181, .length = 3},
  [173] = {.index = 184, .length = 1},
  [174] = {.index = 185, .length = 2},
  [175] = {.index = 139, .length = 1},
  [176] = {.index = 187, .length = 1},
  [177] = {.index = 187, .length = 1},
  [179] = {.index = 188, .length = 2},
  [180] = {.index = 190, .length = 1},
  [181] = {.index = 191, .length = 4},
  [182] = {.index = 195, .length = 3},
  [183] = {.index = 198, .length = 3},
  [184] = {.index = 139, .length = 1},
  [185] = {.index = 165, .length = 2},
  [186] = {.index = 139, .length = 1},
  [187] = {.index = 139, .length = 1},
  [188] = {.index = 201, .length = 1},
  [189] = {.index = 165, .length = 2},
  [190] = {.index = 202, .length = 2},
  [191] = {.index = 204, .length = 2},
  [192] = {.index = 206, .length = 3},
  [193] = {.index = 209, .length = 3},
  [194] = {.index = 212, .length = 3},
  [195] = {.index = 139, .length = 1},
  [196] = {.index = 215, .length = 2},
  [197] = {.index = 217, .length = 1},
  [198] = {.index = 217, .length = 1},
  [199] = {.index = 218, .length = 1},
  [200] = {.index = 218, .length = 1},
  [201] = {.index = 219, .length = 1},
  [202] = {.index = 220, .length = 3},
  [203] = {.index = 223, .length = 2},
  [204] = {.index = 225, .length = 3},
  [205] = {.index = 228, .length = 2},
  [206] = {.index = 228, .length = 2},
  [207] = {.index = 217, .length = 1},
  [208] = {.index = 230, .length = 3},
  [209] = {.index = 233, .length = 2},
  [210] = {.index = 235, .length = 3},
  [211] = {.index = 238, .length = 1},
  [212] = {.index = 239, .length = 1},
  [213] = {.index = 217, .length = 1},
  [214] = {.index = 240, .length = 2},
  [215] = {.index = 240, .length = 2},
  [216] = {.index = 242, .length = 1},
  [217] = {.index = 242, .length = 1},
  [218] = {.index = 243, .length = 1},
  [219] = {.index = 244, .length = 2},
  [220] = {.index = 246, .length = 2},
  [221] = {.index = 248, .length = 2},
  [222] = {.index = 250, .length = 1},
  [223] = {.index = 251, .length = 1},
  [224] = {.index = 242, .length = 1},
  [225] = {.index = 252, .length = 2},
  [226] = {.index = 252, .length = 2},
  [227] = {.index = 254, .length = 1},
  [228] = {.index = 255, .length = 3},
  [229] = {.index = 258, .length = 1},
  [230] = {.index = 259, .length = 2},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_pattern, 0},
  [1] =
    {field_name, 0},
  [2] =
    {field_module, 1},
  [3] =
    {field_expression, 1},
  [4] =
    {field_expression, 1},
    {field_operator, 0},
  [6] =
    {field_argument, 0},
  [7] =
    {field_argument, 1, .inherited = true},
    {field_function, 0},
  [9] =
    {field_pattern, 1},
  [10] =
    {field_type, 0, .inherited = true},
  [11] =
    {field_coercion, 0, .inherited = true},
  [12] =
    {field_name, 1},
  [13] =
    {field_coercion, 1, .inherited = true},
    {field_type, 1, .inherited = true},
  [15] =
    {field_module_type, 1, .inherited = true},
  [16] =
    {field_module, 2},
  [17] =
    {field_argument, 1},
    {field_functor, 0},
  [19] =
    {field_class_type, 1, .inherited = true},
  [20] =
    {field_type, 2, .inherited = true},
  [21] =
    {field_class, 1},
  [22] =
    {field_initializer, 1},
  [23] =
    {field_condition, 1},
  [24] =
    {field_expression, 2},
  [25] =
    {field_body, 2},
  [26] =
    {field_field, 2},
    {field_record, 0},
  [28] =
    {field_method, 2},
    {field_object, 0},
  [30] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [33] =
    {field_argument, 0, .inherited = true},
    {field_argument, 1, .inherited = true},
  [35] =
    {field_left, 0},
    {field_right, 2},
  [37] =
    {field_type, 1, .inherited = true},
  [38] =
    {field_pattern, 2},
  [39] =
    {field_body, 3},
  [40] =
    {field_body, 2},
    {field_pattern, 0},
  [42] =
    {field_type, 1},
  [43] =
    {field_coercion, 1},
  [44] =
    {field_alias, 2},
    {field_pattern, 0},
  [46] =
    {field_expression, 3},
  [47] =
    {field_coercion, 1, .inherited = true},
    {field_type, 0, .inherited = true},
  [49] =
    {field_coercion, 2, .inherited = true},
    {field_expression, 1},
    {field_type, 2, .inherited = true},
  [52] =
    {field_body, 2},
    {field_name, 0},
  [54] =
    {field_module_type, 1},
  [55] =
    {field_module_type, 2, .inherited = true},
  [56] =
    {field_module, 3},
  [57] =
    {field_functor, 0},
  [58] =
    {field_class_type, 2, .inherited = true},
  [59] =
    {field_class_type, 1},
  [60] =
    {field_type, 3, .inherited = true},
  [61] =
    {field_coercion, 2, .inherited = true},
    {field_type, 2, .inherited = true},
  [63] =
    {field_class, 2},
  [64] =
    {field_argument, 1, .inherited = true},
    {field_class, 0},
  [66] =
    {field_condition, 2},
  [67] =
    {field_pattern, 3},
  [68] =
    {field_pattern, 1},
    {field_type, 2, .inherited = true},
  [70] =
    {field_body, 3},
    {field_coercion, 1, .inherited = true},
    {field_pattern, 0},
  [73] =
    {field_body, 3},
    {field_pattern, 0},
    {field_type, 1, .inherited = true},
  [76] =
    {field_body, 3},
    {field_pattern, 0},
  [78] =
    {field_body, 4},
  [79] =
    {field_coercion, 3, .inherited = true},
    {field_type, 3, .inherited = true},
  [81] =
    {field_module, 2},
    {field_module_type, 3, .inherited = true},
  [83] =
    {field_body, 3},
    {field_name, 0},
  [85] =
    {field_body, 3},
    {field_name, 1},
  [87] =
    {field_record, 1},
  [88] =
    {field_body, 3},
    {field_coercion, 1, .inherited = true},
    {field_type, 1, .inherited = true},
  [91] =
    {field_codomain, 1},
  [92] =
    {field_body, 3},
    {field_module_type, 1, .inherited = true},
  [94] =
    {field_value, 2},
  [95] =
    {field_module, 1},
    {field_module_type, 2, .inherited = true},
  [97] =
    {field_class_type, 3, .inherited = true},
  [98] =
    {field_body, 3},
    {field_class_type, 1, .inherited = true},
  [100] =
    {field_self, 2},
  [101] =
    {field_type, 2},
  [102] =
    {field_codomain, 2},
    {field_domain, 0},
  [104] =
    {field_constraint, 3},
    {field_type, 1},
  [106] =
    {field_alias, 2},
    {field_type, 0},
  [108] =
    {field_alias, 3},
    {field_class, 1},
  [110] =
    {field_body, 4},
    {field_type, 2, .inherited = true},
  [112] =
    {field_continuation, 3},
    {field_effect, 1},
  [114] =
    {field_array, 0},
    {field_index, 3},
  [116] =
    {field_index, 3},
    {field_string, 0},
  [118] =
    {field_module_type, 2},
  [119] =
    {field_module_type, 3, .inherited = true},
  [120] =
    {field_pattern, 3},
    {field_type, 1, .inherited = true},
  [122] =
    {field_body, 4},
    {field_coercion, 2, .inherited = true},
    {field_pattern, 0},
    {field_type, 1, .inherited = true},
  [126] =
    {field_body, 4},
    {field_coercion, 2, .inherited = true},
    {field_pattern, 0},
  [129] =
    {field_body, 4},
    {field_pattern, 0},
    {field_type, 2, .inherited = true},
  [132] =
    {field_module, 3},
    {field_module_type, 4, .inherited = true},
  [134] =
    {field_body, 4},
    {field_name, 0},
    {field_synonym, 2},
  [137] =
    {field_body, 4},
    {field_name, 1},
  [139] =
    {field_body, 5},
  [140] =
    {field_codomain, 2},
  [141] =
    {field_module_type, 0},
  [142] =
    {field_body, 4},
    {field_module_type, 2, .inherited = true},
  [144] =
    {field_module_type, 3, .inherited = true},
    {field_value, 2},
  [146] =
    {field_class_type, 4, .inherited = true},
  [147] =
    {field_body, 4},
    {field_class_type, 2, .inherited = true},
  [149] =
    {field_self, 2},
    {field_type, 3, .inherited = true},
  [151] =
    {field_type, 3},
  [152] =
    {field_coercion, 4, .inherited = true},
    {field_type, 4, .inherited = true},
  [154] =
    {field_body, 4},
    {field_coercion, 2, .inherited = true},
    {field_type, 2, .inherited = true},
  [157] =
    {field_class, 1},
    {field_class_type, 2, .inherited = true},
  [159] =
    {field_alias, 4},
    {field_class, 2},
  [161] =
    {field_type, 4, .inherited = true},
  [162] =
    {field_self, 3},
  [163] =
    {field_pattern, 2},
    {field_type, 3, .inherited = true},
  [165] =
    {field_body, 5},
    {field_type, 3, .inherited = true},
  [167] =
    {field_array, 0},
    {field_index, 4},
    {field_operator, 2},
  [170] =
    {field_index, 4},
    {field_operator, 2},
    {field_string, 0},
  [173] =
    {field_module_type, 4, .inherited = true},
  [174] =
    {field_body, 5},
    {field_coercion, 3, .inherited = true},
    {field_pattern, 0},
    {field_type, 2, .inherited = true},
  [178] =
    {field_body, 5},
    {field_name, 0},
    {field_synonym, 2},
  [181] =
    {field_body, 5},
    {field_name, 1},
    {field_synonym, 3},
  [184] =
    {field_codomain, 3},
  [185] =
    {field_coercion, 4},
    {field_value, 2},
  [187] =
    {field_class_type, 5, .inherited = true},
  [188] =
    {field_body, 5},
    {field_class_type, 3, .inherited = true},
  [190] =
    {field_module_type, 3},
  [191] =
    {field_codomain, 4},
    {field_domain, 0},
    {field_domain, 1},
    {field_domain, 2},
  [195] =
    {field_type, 2},
    {field_type, 3},
    {field_type, 4},
  [198] =
    {field_body, 5},
    {field_coercion, 3, .inherited = true},
    {field_type, 3, .inherited = true},
  [201] =
    {field_type, 5, .inherited = true},
  [202] =
    {field_self, 3},
    {field_type, 4, .inherited = true},
  [204] =
    {field_default, 4},
    {field_pattern, 2},
  [206] =
    {field_from, 3},
    {field_name, 1},
    {field_to, 5},
  [209] =
    {field_body, 6},
    {field_name, 1},
    {field_synonym, 3},
  [212] =
    {field_coercion, 5},
    {field_module_type, 3, .inherited = true},
    {field_value, 2},
  [215] =
    {field_body, 6},
    {field_class_type, 4, .inherited = true},
  [217] =
    {field_body, 6},
  [218] =
    {field_class_type, 6, .inherited = true},
  [219] =
    {field_self_type, 2},
  [220] =
    {field_type, 3},
    {field_type, 4},
    {field_type, 5},
  [223] =
    {field_module, 2},
    {field_module_type, 4},
  [225] =
    {field_body, 6},
    {field_coercion, 4, .inherited = true},
    {field_type, 4, .inherited = true},
  [228] =
    {field_body, 6},
    {field_type, 4, .inherited = true},
  [230] =
    {field_default, 5},
    {field_pattern, 2},
    {field_type, 3, .inherited = true},
  [233] =
    {field_pattern, 4},
    {field_type, 5, .inherited = true},
  [235] =
    {field_from, 4},
    {field_name, 2},
    {field_to, 6},
  [238] =
    {field_equation, 3},
  [239] =
    {field_constraint, 3},
  [240] =
    {field_body, 7},
    {field_class_type, 5, .inherited = true},
  [242] =
    {field_body, 7},
  [243] =
    {field_class_type, 7, .inherited = true},
  [244] =
    {field_module, 3},
    {field_module_type, 5},
  [246] =
    {field_body, 7},
    {field_type, 5, .inherited = true},
  [248] =
    {field_default, 6},
    {field_pattern, 4},
  [250] =
    {field_equation, 4},
  [251] =
    {field_constraint, 4},
  [252] =
    {field_body, 8},
    {field_class_type, 6, .inherited = true},
  [254] =
    {field_body, 8},
  [255] =
    {field_default, 7},
    {field_pattern, 4},
    {field_type, 5, .inherited = true},
  [258] =
    {field_equation, 5},
  [259] =
    {field_body, 9},
    {field_class_type, 7, .inherited = true},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = alias_sym_value_name,
  },
  [2] = {
    [0] = alias_sym_instance_variable_name,
  },
  [3] = {
    [0] = alias_sym_module_name,
  },
  [4] = {
    [0] = sym__constructor_declaration_name,
  },
  [5] = {
    [0] = alias_sym_tuple_expression,
  },
  [6] = {
    [0] = alias_sym_sequence_expression,
  },
  [8] = {
    [1] = alias_sym_label_name,
  },
  [9] = {
    [0] = sym_type_variable,
  },
  [10] = {
    [0] = alias_sym_type_constructor,
  },
  [11] = {
    [0] = alias_sym_type_constructor,
  },
  [12] = {
    [0] = alias_sym_field_name,
  },
  [14] = {
    [0] = alias_sym_class_name,
  },
  [15] = {
    [0] = alias_sym_value_pattern,
  },
  [16] = {
    [0] = sym__or_binding_pattern_no_exn_anonymous,
  },
  [17] = {
    [0] = sym__tuple_binding_pattern_no_exn,
  },
  [23] = {
    [1] = sym__constructor_declaration_name,
  },
  [26] = {
    [1] = alias_sym_type_constructor,
  },
  [27] = {
    [1] = sym_type_variable,
  },
  [29] = {
    [2] = alias_sym_module_type_name,
  },
  [30] = {
    [0] = alias_sym_module_name,
  },
  [33] = {
    [1] = alias_sym_class_name,
  },
  [34] = {
    [0] = alias_sym_class_name,
  },
  [37] = {
    [1] = alias_sym_method_name,
  },
  [43] = {
    [2] = alias_sym_method_name,
  },
  [47] = {
    [2] = alias_sym_module_name,
  },
  [55] = {
    [1] = alias_sym_label_name,
  },
  [58] = {
    [0] = alias_sym_type_constructor,
  },
  [59] = {
    [2] = alias_sym_type_constructor,
  },
  [61] = {
    [2] = alias_sym_field_name,
  },
  [62] = {
    [3] = alias_sym_module_type_name,
  },
  [63] = {
    [0] = alias_sym_module_name,
  },
  [64] = {
    [0] = alias_sym_module_type_name,
  },
  [66] = {
    [0] = alias_sym_module_name,
  },
  [69] = {
    [1] = alias_sym_class_name,
  },
  [70] = {
    [0] = alias_sym_class_name,
  },
  [71] = {
    [0] = alias_sym_class_type_name,
  },
  [73] = {
    [0] = alias_sym_class_name,
  },
  [78] = {
    [2] = alias_sym_method_name,
  },
  [79] = {
    [1] = alias_sym_method_name,
  },
  [81] = {
    [2] = alias_sym_class_name,
  },
  [82] = {
    [1] = alias_sym_label_name,
  },
  [88] = {
    [2] = alias_sym_label_name,
  },
  [90] = {
    [0] = alias_sym_type_constructor,
  },
  [91] = {
    [1] = alias_sym_type_constructor,
  },
  [96] = {
    [2] = alias_sym_module_type_name,
  },
  [98] = {
    [0] = alias_sym_module_name,
  },
  [99] = {
    [0] = alias_sym_module_name,
  },
  [102] = {
    [0] = alias_sym_class_type_name,
  },
  [103] = {
    [3] = alias_sym_class_name,
  },
  [104] = {
    [1] = alias_sym_class_name,
  },
  [105] = {
    [1] = alias_sym_class_name,
  },
  [106] = {
    [0] = alias_sym_class_name,
  },
  [107] = {
    [0] = alias_sym_class_name,
  },
  [109] = {
    [0] = alias_sym_method_name,
  },
  [110] = {
    [0] = alias_sym_label_name,
  },
  [117] = {
    [2] = alias_sym_method_name,
  },
  [118] = {
    [3] = alias_sym_method_name,
  },
  [119] = {
    [1] = alias_sym_method_name,
  },
  [120] = {
    [1] = alias_sym_method_name,
  },
  [126] = {
    [2] = alias_sym_label_name,
  },
  [127] = {
    [2] = alias_sym_module_name,
  },
  [128] = {
    [3] = alias_sym_module_name,
  },
  [130] = {
    [0] = sym__parenthesized_abstract_type,
  },
  [135] = {
    [0] = alias_sym_field_name,
  },
  [136] = {
    [0] = alias_sym_type_constructor,
  },
  [137] = {
    [1] = alias_sym_type_constructor,
  },
  [139] = {
    [3] = alias_sym_module_type_name,
  },
  [140] = {
    [1] = alias_sym_module_name,
  },
  [143] = {
    [0] = alias_sym_module_name,
  },
  [145] = {
    [1] = alias_sym_class_type_name,
  },
  [146] = {
    [3] = alias_sym_class_name,
  },
  [147] = {
    [4] = alias_sym_class_name,
  },
  [148] = {
    [1] = alias_sym_class_name,
  },
  [149] = {
    [1] = alias_sym_class_name,
  },
  [150] = {
    [2] = alias_sym_class_type_name,
  },
  [151] = {
    [0] = alias_sym_class_name,
  },
  [153] = {
    [1] = alias_sym_label_name,
  },
  [158] = {
    [2] = alias_sym_method_name,
  },
  [159] = {
    [2] = alias_sym_method_name,
  },
  [160] = {
    [3] = alias_sym_method_name,
  },
  [161] = {
    [1] = alias_sym_method_name,
  },
  [162] = {
    [1] = alias_sym_method_name,
  },
  [168] = {
    [3] = alias_sym_module_name,
  },
  [170] = {
    [1] = alias_sym_field_name,
  },
  [171] = {
    [0] = alias_sym_type_constructor,
  },
  [172] = {
    [1] = alias_sym_type_constructor,
  },
  [175] = {
    [3] = alias_sym_class_name,
  },
  [176] = {
    [3] = alias_sym_class_name,
  },
  [177] = {
    [4] = alias_sym_class_name,
  },
  [178] = {
    [5] = alias_sym_class_name,
  },
  [179] = {
    [1] = alias_sym_class_name,
  },
  [182] = {
    [0] = alias_sym_label_name,
  },
  [185] = {
    [2] = alias_sym_method_name,
  },
  [186] = {
    [2] = alias_sym_method_name,
  },
  [187] = {
    [3] = alias_sym_method_name,
  },
  [188] = {
    [3] = alias_sym_method_name,
  },
  [189] = {
    [1] = alias_sym_method_name,
  },
  [193] = {
    [1] = alias_sym_type_constructor,
  },
  [195] = {
    [3] = alias_sym_class_type_name,
  },
  [196] = {
    [3] = alias_sym_class_name,
  },
  [197] = {
    [3] = alias_sym_class_name,
  },
  [198] = {
    [4] = alias_sym_class_name,
  },
  [199] = {
    [4] = alias_sym_class_name,
  },
  [200] = {
    [5] = alias_sym_class_name,
  },
  [202] = {
    [1] = alias_sym_label_name,
  },
  [203] = {
    [2] = alias_sym_module_name,
  },
  [205] = {
    [2] = alias_sym_method_name,
  },
  [206] = {
    [3] = alias_sym_method_name,
  },
  [207] = {
    [3] = alias_sym_method_name,
  },
  [209] = {
    [1] = alias_sym_label_name,
  },
  [213] = {
    [4] = alias_sym_class_type_name,
  },
  [214] = {
    [3] = alias_sym_class_name,
  },
  [215] = {
    [4] = alias_sym_class_name,
  },
  [216] = {
    [4] = alias_sym_class_name,
  },
  [217] = {
    [5] = alias_sym_class_name,
  },
  [218] = {
    [5] = alias_sym_class_name,
  },
  [219] = {
    [3] = alias_sym_module_name,
  },
  [220] = {
    [3] = alias_sym_method_name,
  },
  [221] = {
    [1] = alias_sym_label_name,
  },
  [224] = {
    [5] = alias_sym_class_type_name,
  },
  [225] = {
    [4] = alias_sym_class_name,
  },
  [226] = {
    [5] = alias_sym_class_name,
  },
  [227] = {
    [5] = alias_sym_class_name,
  },
  [228] = {
    [1] = alias_sym_label_name,
  },
  [230] = {
    [5] = alias_sym_class_name,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym__abstract_type, 2,
    sym__abstract_type,
    sym__parenthesized_abstract_type,
  sym__tuple_expression, 2,
    sym__tuple_expression,
    alias_sym_tuple_expression,
  sym__sequence_expression_anonymous, 2,
    sym__sequence_expression_anonymous,
    alias_sym_sequence_expression,
  sym__or_pattern_anonymous, 2,
    sym__or_pattern_anonymous,
    sym__or_binding_pattern_no_exn_anonymous,
  sym__or_binding_pattern_anonymous, 2,
    sym__or_binding_pattern_anonymous,
    sym__or_binding_pattern_no_exn_anonymous,
  sym__tuple_pattern, 2,
    sym__tuple_pattern,
    sym__tuple_binding_pattern_no_exn,
  sym__tuple_binding_pattern, 2,
    sym__tuple_binding_pattern,
    sym__tuple_binding_pattern_no_exn,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 2,
  [4] = 2,
  [5] = 2,
  [6] = 2,
  [7] = 2,
  [8] = 8,
  [9] = 9,
  [10] = 8,
  [11] = 11,
  [12] = 8,
  [13] = 13,
  [14] = 9,
  [15] = 13,
  [16] = 11,
  [17] = 8,
  [18] = 9,
  [19] = 13,
  [20] = 9,
  [21] = 8,
  [22] = 8,
  [23] = 8,
  [24] = 24,
  [25] = 8,
  [26] = 8,
  [27] = 8,
  [28] = 8,
  [29] = 8,
  [30] = 8,
  [31] = 8,
  [32] = 8,
  [33] = 8,
  [34] = 8,
  [35] = 8,
  [36] = 8,
  [37] = 8,
  [38] = 8,
  [39] = 8,
  [40] = 8,
  [41] = 8,
  [42] = 42,
  [43] = 43,
  [44] = 42,
  [45] = 45,
  [46] = 46,
  [47] = 47,
  [48] = 2,
  [49] = 49,
  [50] = 47,
  [51] = 46,
  [52] = 2,
  [53] = 45,
  [54] = 43,
  [55] = 55,
  [56] = 55,
  [57] = 2,
  [58] = 2,
  [59] = 59,
  [60] = 60,
  [61] = 60,
  [62] = 60,
  [63] = 59,
  [64] = 60,
  [65] = 59,
  [66] = 60,
  [67] = 59,
  [68] = 59,
  [69] = 60,
  [70] = 60,
  [71] = 59,
  [72] = 60,
  [73] = 59,
  [74] = 60,
  [75] = 59,
  [76] = 60,
  [77] = 59,
  [78] = 60,
  [79] = 59,
  [80] = 60,
  [81] = 59,
  [82] = 60,
  [83] = 59,
  [84] = 59,
  [85] = 60,
  [86] = 59,
  [87] = 2,
  [88] = 2,
  [89] = 2,
  [90] = 2,
  [91] = 2,
  [92] = 2,
  [93] = 2,
  [94] = 2,
  [95] = 2,
  [96] = 2,
  [97] = 97,
  [98] = 2,
  [99] = 99,
  [100] = 2,
  [101] = 97,
  [102] = 102,
  [103] = 97,
  [104] = 97,
  [105] = 102,
  [106] = 99,
  [107] = 97,
  [108] = 102,
  [109] = 2,
  [110] = 99,
  [111] = 102,
  [112] = 97,
  [113] = 99,
  [114] = 2,
  [115] = 99,
  [116] = 99,
  [117] = 102,
  [118] = 102,
  [119] = 2,
  [120] = 2,
  [121] = 121,
  [122] = 121,
  [123] = 121,
  [124] = 121,
  [125] = 121,
  [126] = 121,
  [127] = 121,
  [128] = 128,
  [129] = 121,
  [130] = 121,
  [131] = 121,
  [132] = 132,
  [133] = 121,
  [134] = 121,
  [135] = 132,
  [136] = 121,
  [137] = 121,
  [138] = 132,
  [139] = 128,
  [140] = 128,
  [141] = 132,
  [142] = 142,
  [143] = 142,
  [144] = 144,
  [145] = 145,
  [146] = 146,
  [147] = 146,
  [148] = 146,
  [149] = 2,
  [150] = 146,
  [151] = 145,
  [152] = 145,
  [153] = 145,
  [154] = 145,
  [155] = 144,
  [156] = 145,
  [157] = 145,
  [158] = 146,
  [159] = 146,
  [160] = 146,
  [161] = 145,
  [162] = 162,
  [163] = 163,
  [164] = 164,
  [165] = 145,
  [166] = 146,
  [167] = 164,
  [168] = 145,
  [169] = 145,
  [170] = 163,
  [171] = 164,
  [172] = 163,
  [173] = 164,
  [174] = 163,
  [175] = 163,
  [176] = 163,
  [177] = 163,
  [178] = 163,
  [179] = 163,
  [180] = 146,
  [181] = 163,
  [182] = 163,
  [183] = 163,
  [184] = 163,
  [185] = 162,
  [186] = 144,
  [187] = 146,
  [188] = 2,
  [189] = 146,
  [190] = 163,
  [191] = 162,
  [192] = 144,
  [193] = 162,
  [194] = 144,
  [195] = 162,
  [196] = 144,
  [197] = 162,
  [198] = 144,
  [199] = 162,
  [200] = 144,
  [201] = 162,
  [202] = 163,
  [203] = 162,
  [204] = 144,
  [205] = 162,
  [206] = 144,
  [207] = 162,
  [208] = 144,
  [209] = 162,
  [210] = 144,
  [211] = 162,
  [212] = 144,
  [213] = 163,
  [214] = 214,
  [215] = 215,
  [216] = 216,
  [217] = 214,
  [218] = 218,
  [219] = 218,
  [220] = 215,
  [221] = 216,
  [222] = 214,
  [223] = 223,
  [224] = 218,
  [225] = 216,
  [226] = 214,
  [227] = 216,
  [228] = 214,
  [229] = 216,
  [230] = 214,
  [231] = 214,
  [232] = 216,
  [233] = 214,
  [234] = 234,
  [235] = 216,
  [236] = 214,
  [237] = 216,
  [238] = 216,
  [239] = 214,
  [240] = 216,
  [241] = 214,
  [242] = 216,
  [243] = 214,
  [244] = 216,
  [245] = 214,
  [246] = 216,
  [247] = 223,
  [248] = 234,
  [249] = 215,
  [250] = 223,
  [251] = 215,
  [252] = 215,
  [253] = 216,
  [254] = 234,
  [255] = 223,
  [256] = 214,
  [257] = 218,
  [258] = 215,
  [259] = 259,
  [260] = 260,
  [261] = 261,
  [262] = 262,
  [263] = 263,
  [264] = 264,
  [265] = 265,
  [266] = 266,
  [267] = 259,
  [268] = 268,
  [269] = 269,
  [270] = 270,
  [271] = 259,
  [272] = 266,
  [273] = 273,
  [274] = 260,
  [275] = 259,
  [276] = 260,
  [277] = 270,
  [278] = 278,
  [279] = 269,
  [280] = 260,
  [281] = 281,
  [282] = 282,
  [283] = 283,
  [284] = 284,
  [285] = 285,
  [286] = 286,
  [287] = 287,
  [288] = 288,
  [289] = 268,
  [290] = 259,
  [291] = 260,
  [292] = 292,
  [293] = 261,
  [294] = 261,
  [295] = 262,
  [296] = 263,
  [297] = 264,
  [298] = 268,
  [299] = 262,
  [300] = 259,
  [301] = 260,
  [302] = 261,
  [303] = 262,
  [304] = 263,
  [305] = 264,
  [306] = 263,
  [307] = 266,
  [308] = 273,
  [309] = 309,
  [310] = 261,
  [311] = 262,
  [312] = 270,
  [313] = 263,
  [314] = 314,
  [315] = 269,
  [316] = 316,
  [317] = 264,
  [318] = 264,
  [319] = 282,
  [320] = 283,
  [321] = 284,
  [322] = 285,
  [323] = 286,
  [324] = 287,
  [325] = 288,
  [326] = 326,
  [327] = 268,
  [328] = 328,
  [329] = 266,
  [330] = 273,
  [331] = 331,
  [332] = 270,
  [333] = 268,
  [334] = 259,
  [335] = 269,
  [336] = 282,
  [337] = 283,
  [338] = 284,
  [339] = 285,
  [340] = 286,
  [341] = 287,
  [342] = 288,
  [343] = 260,
  [344] = 344,
  [345] = 266,
  [346] = 273,
  [347] = 347,
  [348] = 348,
  [349] = 270,
  [350] = 259,
  [351] = 261,
  [352] = 262,
  [353] = 266,
  [354] = 273,
  [355] = 263,
  [356] = 260,
  [357] = 270,
  [358] = 358,
  [359] = 264,
  [360] = 266,
  [361] = 273,
  [362] = 268,
  [363] = 261,
  [364] = 270,
  [365] = 262,
  [366] = 366,
  [367] = 263,
  [368] = 259,
  [369] = 266,
  [370] = 273,
  [371] = 260,
  [372] = 264,
  [373] = 270,
  [374] = 268,
  [375] = 282,
  [376] = 283,
  [377] = 266,
  [378] = 273,
  [379] = 261,
  [380] = 380,
  [381] = 381,
  [382] = 262,
  [383] = 263,
  [384] = 284,
  [385] = 266,
  [386] = 273,
  [387] = 264,
  [388] = 259,
  [389] = 260,
  [390] = 268,
  [391] = 261,
  [392] = 259,
  [393] = 266,
  [394] = 273,
  [395] = 260,
  [396] = 396,
  [397] = 261,
  [398] = 266,
  [399] = 273,
  [400] = 262,
  [401] = 261,
  [402] = 262,
  [403] = 263,
  [404] = 404,
  [405] = 264,
  [406] = 406,
  [407] = 263,
  [408] = 408,
  [409] = 268,
  [410] = 262,
  [411] = 263,
  [412] = 259,
  [413] = 260,
  [414] = 381,
  [415] = 264,
  [416] = 261,
  [417] = 262,
  [418] = 263,
  [419] = 273,
  [420] = 264,
  [421] = 421,
  [422] = 292,
  [423] = 268,
  [424] = 268,
  [425] = 331,
  [426] = 344,
  [427] = 347,
  [428] = 421,
  [429] = 429,
  [430] = 430,
  [431] = 431,
  [432] = 432,
  [433] = 433,
  [434] = 281,
  [435] = 314,
  [436] = 408,
  [437] = 437,
  [438] = 268,
  [439] = 429,
  [440] = 259,
  [441] = 260,
  [442] = 261,
  [443] = 262,
  [444] = 263,
  [445] = 264,
  [446] = 268,
  [447] = 260,
  [448] = 448,
  [449] = 270,
  [450] = 260,
  [451] = 437,
  [452] = 270,
  [453] = 270,
  [454] = 270,
  [455] = 270,
  [456] = 261,
  [457] = 292,
  [458] = 331,
  [459] = 344,
  [460] = 347,
  [461] = 421,
  [462] = 429,
  [463] = 430,
  [464] = 433,
  [465] = 314,
  [466] = 270,
  [467] = 262,
  [468] = 263,
  [469] = 292,
  [470] = 331,
  [471] = 344,
  [472] = 347,
  [473] = 421,
  [474] = 429,
  [475] = 430,
  [476] = 433,
  [477] = 314,
  [478] = 430,
  [479] = 331,
  [480] = 344,
  [481] = 347,
  [482] = 421,
  [483] = 429,
  [484] = 430,
  [485] = 264,
  [486] = 331,
  [487] = 344,
  [488] = 347,
  [489] = 421,
  [490] = 429,
  [491] = 430,
  [492] = 268,
  [493] = 331,
  [494] = 344,
  [495] = 347,
  [496] = 421,
  [497] = 429,
  [498] = 430,
  [499] = 285,
  [500] = 331,
  [501] = 344,
  [502] = 347,
  [503] = 421,
  [504] = 429,
  [505] = 430,
  [506] = 286,
  [507] = 331,
  [508] = 344,
  [509] = 347,
  [510] = 421,
  [511] = 429,
  [512] = 430,
  [513] = 287,
  [514] = 331,
  [515] = 344,
  [516] = 347,
  [517] = 421,
  [518] = 429,
  [519] = 430,
  [520] = 520,
  [521] = 331,
  [522] = 344,
  [523] = 347,
  [524] = 421,
  [525] = 429,
  [526] = 430,
  [527] = 527,
  [528] = 331,
  [529] = 344,
  [530] = 347,
  [531] = 421,
  [532] = 429,
  [533] = 430,
  [534] = 534,
  [535] = 331,
  [536] = 344,
  [537] = 347,
  [538] = 421,
  [539] = 429,
  [540] = 430,
  [541] = 541,
  [542] = 542,
  [543] = 266,
  [544] = 273,
  [545] = 366,
  [546] = 316,
  [547] = 328,
  [548] = 268,
  [549] = 431,
  [550] = 259,
  [551] = 260,
  [552] = 261,
  [553] = 262,
  [554] = 263,
  [555] = 264,
  [556] = 261,
  [557] = 270,
  [558] = 262,
  [559] = 264,
  [560] = 270,
  [561] = 366,
  [562] = 316,
  [563] = 328,
  [564] = 263,
  [565] = 264,
  [566] = 366,
  [567] = 316,
  [568] = 328,
  [569] = 433,
  [570] = 288,
  [571] = 366,
  [572] = 316,
  [573] = 328,
  [574] = 574,
  [575] = 269,
  [576] = 366,
  [577] = 316,
  [578] = 328,
  [579] = 579,
  [580] = 268,
  [581] = 366,
  [582] = 316,
  [583] = 328,
  [584] = 584,
  [585] = 326,
  [586] = 366,
  [587] = 316,
  [588] = 328,
  [589] = 589,
  [590] = 366,
  [591] = 316,
  [592] = 328,
  [593] = 282,
  [594] = 283,
  [595] = 366,
  [596] = 316,
  [597] = 328,
  [598] = 284,
  [599] = 285,
  [600] = 366,
  [601] = 316,
  [602] = 328,
  [603] = 286,
  [604] = 287,
  [605] = 366,
  [606] = 316,
  [607] = 328,
  [608] = 259,
  [609] = 288,
  [610] = 366,
  [611] = 316,
  [612] = 328,
  [613] = 309,
  [614] = 381,
  [615] = 309,
  [616] = 381,
  [617] = 309,
  [618] = 259,
  [619] = 619,
  [620] = 619,
  [621] = 619,
  [622] = 619,
  [623] = 619,
  [624] = 619,
  [625] = 619,
  [626] = 626,
  [627] = 619,
  [628] = 619,
  [629] = 619,
  [630] = 619,
  [631] = 631,
  [632] = 626,
  [633] = 619,
  [634] = 619,
  [635] = 619,
  [636] = 619,
  [637] = 619,
  [638] = 619,
  [639] = 619,
  [640] = 619,
  [641] = 619,
  [642] = 619,
  [643] = 619,
  [644] = 619,
  [645] = 619,
  [646] = 619,
  [647] = 631,
  [648] = 619,
  [649] = 619,
  [650] = 619,
  [651] = 651,
  [652] = 652,
  [653] = 651,
  [654] = 654,
  [655] = 655,
  [656] = 656,
  [657] = 657,
  [658] = 658,
  [659] = 659,
  [660] = 660,
  [661] = 661,
  [662] = 662,
  [663] = 663,
  [664] = 664,
  [665] = 665,
  [666] = 652,
  [667] = 651,
  [668] = 654,
  [669] = 655,
  [670] = 656,
  [671] = 657,
  [672] = 658,
  [673] = 659,
  [674] = 660,
  [675] = 661,
  [676] = 662,
  [677] = 663,
  [678] = 678,
  [679] = 664,
  [680] = 680,
  [681] = 665,
  [682] = 652,
  [683] = 651,
  [684] = 654,
  [685] = 655,
  [686] = 656,
  [687] = 657,
  [688] = 658,
  [689] = 659,
  [690] = 660,
  [691] = 661,
  [692] = 662,
  [693] = 663,
  [694] = 664,
  [695] = 664,
  [696] = 665,
  [697] = 652,
  [698] = 651,
  [699] = 654,
  [700] = 655,
  [701] = 656,
  [702] = 657,
  [703] = 658,
  [704] = 659,
  [705] = 660,
  [706] = 661,
  [707] = 662,
  [708] = 663,
  [709] = 664,
  [710] = 665,
  [711] = 711,
  [712] = 652,
  [713] = 651,
  [714] = 654,
  [715] = 655,
  [716] = 656,
  [717] = 657,
  [718] = 658,
  [719] = 659,
  [720] = 660,
  [721] = 661,
  [722] = 662,
  [723] = 663,
  [724] = 664,
  [725] = 725,
  [726] = 665,
  [727] = 727,
  [728] = 665,
  [729] = 652,
  [730] = 651,
  [731] = 654,
  [732] = 655,
  [733] = 656,
  [734] = 657,
  [735] = 658,
  [736] = 659,
  [737] = 660,
  [738] = 661,
  [739] = 662,
  [740] = 663,
  [741] = 664,
  [742] = 665,
  [743] = 743,
  [744] = 652,
  [745] = 651,
  [746] = 652,
  [747] = 654,
  [748] = 655,
  [749] = 656,
  [750] = 657,
  [751] = 658,
  [752] = 659,
  [753] = 660,
  [754] = 661,
  [755] = 662,
  [756] = 663,
  [757] = 651,
  [758] = 664,
  [759] = 654,
  [760] = 655,
  [761] = 656,
  [762] = 657,
  [763] = 658,
  [764] = 680,
  [765] = 659,
  [766] = 743,
  [767] = 660,
  [768] = 661,
  [769] = 665,
  [770] = 662,
  [771] = 652,
  [772] = 651,
  [773] = 654,
  [774] = 655,
  [775] = 656,
  [776] = 657,
  [777] = 658,
  [778] = 659,
  [779] = 660,
  [780] = 661,
  [781] = 662,
  [782] = 663,
  [783] = 664,
  [784] = 663,
  [785] = 665,
  [786] = 652,
  [787] = 651,
  [788] = 655,
  [789] = 656,
  [790] = 657,
  [791] = 658,
  [792] = 659,
  [793] = 660,
  [794] = 661,
  [795] = 662,
  [796] = 663,
  [797] = 664,
  [798] = 665,
  [799] = 652,
  [800] = 651,
  [801] = 654,
  [802] = 655,
  [803] = 656,
  [804] = 657,
  [805] = 658,
  [806] = 659,
  [807] = 660,
  [808] = 661,
  [809] = 662,
  [810] = 663,
  [811] = 664,
  [812] = 665,
  [813] = 652,
  [814] = 651,
  [815] = 654,
  [816] = 655,
  [817] = 656,
  [818] = 657,
  [819] = 658,
  [820] = 659,
  [821] = 660,
  [822] = 661,
  [823] = 662,
  [824] = 663,
  [825] = 664,
  [826] = 665,
  [827] = 652,
  [828] = 651,
  [829] = 654,
  [830] = 655,
  [831] = 656,
  [832] = 657,
  [833] = 658,
  [834] = 659,
  [835] = 660,
  [836] = 661,
  [837] = 662,
  [838] = 663,
  [839] = 664,
  [840] = 665,
  [841] = 652,
  [842] = 651,
  [843] = 654,
  [844] = 655,
  [845] = 656,
  [846] = 657,
  [847] = 658,
  [848] = 659,
  [849] = 660,
  [850] = 661,
  [851] = 662,
  [852] = 663,
  [853] = 664,
  [854] = 665,
  [855] = 652,
  [856] = 651,
  [857] = 654,
  [858] = 655,
  [859] = 656,
  [860] = 657,
  [861] = 658,
  [862] = 659,
  [863] = 660,
  [864] = 661,
  [865] = 662,
  [866] = 663,
  [867] = 664,
  [868] = 652,
  [869] = 654,
  [870] = 655,
  [871] = 656,
  [872] = 657,
  [873] = 658,
  [874] = 659,
  [875] = 660,
  [876] = 661,
  [877] = 662,
  [878] = 663,
  [879] = 664,
  [880] = 665,
  [881] = 665,
  [882] = 654,
  [883] = 883,
  [884] = 97,
  [885] = 99,
  [886] = 102,
  [887] = 887,
  [888] = 887,
  [889] = 887,
  [890] = 887,
  [891] = 887,
  [892] = 887,
  [893] = 887,
  [894] = 887,
  [895] = 887,
  [896] = 887,
  [897] = 887,
  [898] = 887,
  [899] = 887,
  [900] = 887,
  [901] = 887,
  [902] = 887,
  [903] = 887,
  [904] = 887,
  [905] = 905,
  [906] = 906,
  [907] = 906,
  [908] = 906,
  [909] = 906,
  [910] = 905,
  [911] = 906,
  [912] = 906,
  [913] = 906,
  [914] = 906,
  [915] = 906,
  [916] = 906,
  [917] = 906,
  [918] = 905,
  [919] = 906,
  [920] = 906,
  [921] = 906,
  [922] = 906,
  [923] = 906,
  [924] = 906,
  [925] = 906,
  [926] = 905,
  [927] = 927,
  [928] = 928,
  [929] = 929,
  [930] = 930,
  [931] = 927,
  [932] = 932,
  [933] = 928,
  [934] = 934,
  [935] = 935,
  [936] = 936,
  [937] = 937,
  [938] = 938,
  [939] = 939,
  [940] = 940,
  [941] = 941,
  [942] = 942,
  [943] = 943,
  [944] = 944,
  [945] = 945,
  [946] = 946,
  [947] = 947,
  [948] = 948,
  [949] = 949,
  [950] = 950,
  [951] = 951,
  [952] = 952,
  [953] = 953,
  [954] = 954,
  [955] = 955,
  [956] = 956,
  [957] = 957,
  [958] = 958,
  [959] = 959,
  [960] = 960,
  [961] = 961,
  [962] = 962,
  [963] = 963,
  [964] = 964,
  [965] = 965,
  [966] = 966,
  [967] = 967,
  [968] = 968,
  [969] = 969,
  [970] = 970,
  [971] = 971,
  [972] = 972,
  [973] = 973,
  [974] = 974,
  [975] = 975,
  [976] = 976,
  [977] = 977,
  [978] = 978,
  [979] = 979,
  [980] = 980,
  [981] = 981,
  [982] = 982,
  [983] = 983,
  [984] = 984,
  [985] = 985,
  [986] = 986,
  [987] = 987,
  [988] = 988,
  [989] = 989,
  [990] = 990,
  [991] = 991,
  [992] = 992,
  [993] = 993,
  [994] = 994,
  [995] = 995,
  [996] = 996,
  [997] = 997,
  [998] = 998,
  [999] = 999,
  [1000] = 1000,
  [1001] = 1001,
  [1002] = 1002,
  [1003] = 1003,
  [1004] = 1004,
  [1005] = 1005,
  [1006] = 1006,
  [1007] = 1007,
  [1008] = 1008,
  [1009] = 1009,
  [1010] = 1010,
  [1011] = 1011,
  [1012] = 928,
  [1013] = 927,
  [1014] = 99,
  [1015] = 928,
  [1016] = 1016,
  [1017] = 1017,
  [1018] = 725,
  [1019] = 927,
  [1020] = 1020,
  [1021] = 1021,
  [1022] = 1022,
  [1023] = 1023,
  [1024] = 1024,
  [1025] = 1025,
  [1026] = 1017,
  [1027] = 1027,
  [1028] = 1028,
  [1029] = 1029,
  [1030] = 1030,
  [1031] = 883,
  [1032] = 932,
  [1033] = 935,
  [1034] = 934,
  [1035] = 936,
  [1036] = 937,
  [1037] = 938,
  [1038] = 939,
  [1039] = 1039,
  [1040] = 1040,
  [1041] = 930,
  [1042] = 929,
  [1043] = 975,
  [1044] = 969,
  [1045] = 946,
  [1046] = 958,
  [1047] = 970,
  [1048] = 977,
  [1049] = 979,
  [1050] = 994,
  [1051] = 943,
  [1052] = 944,
  [1053] = 945,
  [1054] = 948,
  [1055] = 949,
  [1056] = 950,
  [1057] = 951,
  [1058] = 952,
  [1059] = 953,
  [1060] = 955,
  [1061] = 956,
  [1062] = 957,
  [1063] = 961,
  [1064] = 962,
  [1065] = 963,
  [1066] = 964,
  [1067] = 965,
  [1068] = 966,
  [1069] = 968,
  [1070] = 972,
  [1071] = 973,
  [1072] = 980,
  [1073] = 1011,
  [1074] = 989,
  [1075] = 971,
  [1076] = 978,
  [1077] = 947,
  [1078] = 954,
  [1079] = 959,
  [1080] = 960,
  [1081] = 940,
  [1082] = 974,
  [1083] = 976,
  [1084] = 981,
  [1085] = 982,
  [1086] = 983,
  [1087] = 984,
  [1088] = 1022,
  [1089] = 1089,
  [1090] = 996,
  [1091] = 1008,
  [1092] = 1009,
  [1093] = 1010,
  [1094] = 967,
  [1095] = 991,
  [1096] = 993,
  [1097] = 1024,
  [1098] = 1021,
  [1099] = 999,
  [1100] = 1025,
  [1101] = 985,
  [1102] = 1027,
  [1103] = 986,
  [1104] = 987,
  [1105] = 1028,
  [1106] = 988,
  [1107] = 990,
  [1108] = 995,
  [1109] = 1039,
  [1110] = 1001,
  [1111] = 1003,
  [1112] = 1004,
  [1113] = 1040,
  [1114] = 998,
  [1115] = 992,
  [1116] = 997,
  [1117] = 1002,
  [1118] = 1000,
  [1119] = 1005,
  [1120] = 1006,
  [1121] = 1007,
  [1122] = 942,
  [1123] = 1020,
  [1124] = 1029,
  [1125] = 941,
  [1126] = 1126,
  [1127] = 930,
  [1128] = 929,
  [1129] = 937,
  [1130] = 938,
  [1131] = 939,
  [1132] = 934,
  [1133] = 935,
  [1134] = 1134,
  [1135] = 932,
  [1136] = 936,
  [1137] = 1137,
  [1138] = 1017,
  [1139] = 986,
  [1140] = 992,
  [1141] = 988,
  [1142] = 997,
  [1143] = 1137,
  [1144] = 966,
  [1145] = 990,
  [1146] = 1009,
  [1147] = 995,
  [1148] = 987,
  [1149] = 980,
  [1150] = 957,
  [1151] = 1002,
  [1152] = 961,
  [1153] = 1011,
  [1154] = 989,
  [1155] = 962,
  [1156] = 1010,
  [1157] = 929,
  [1158] = 967,
  [1159] = 998,
  [1160] = 991,
  [1161] = 993,
  [1162] = 1001,
  [1163] = 999,
  [1164] = 936,
  [1165] = 971,
  [1166] = 1000,
  [1167] = 1003,
  [1168] = 944,
  [1169] = 985,
  [1170] = 945,
  [1171] = 952,
  [1172] = 1005,
  [1173] = 1006,
  [1174] = 1007,
  [1175] = 948,
  [1176] = 949,
  [1177] = 950,
  [1178] = 1004,
  [1179] = 963,
  [1180] = 964,
  [1181] = 984,
  [1182] = 972,
  [1183] = 937,
  [1184] = 938,
  [1185] = 939,
  [1186] = 969,
  [1187] = 968,
  [1188] = 978,
  [1189] = 946,
  [1190] = 958,
  [1191] = 970,
  [1192] = 965,
  [1193] = 1040,
  [1194] = 947,
  [1195] = 942,
  [1196] = 977,
  [1197] = 951,
  [1198] = 979,
  [1199] = 996,
  [1200] = 973,
  [1201] = 954,
  [1202] = 959,
  [1203] = 994,
  [1204] = 941,
  [1205] = 1016,
  [1206] = 932,
  [1207] = 960,
  [1208] = 1008,
  [1209] = 1020,
  [1210] = 1029,
  [1211] = 1028,
  [1212] = 934,
  [1213] = 1039,
  [1214] = 930,
  [1215] = 943,
  [1216] = 935,
  [1217] = 1137,
  [1218] = 940,
  [1219] = 1022,
  [1220] = 974,
  [1221] = 1024,
  [1222] = 1021,
  [1223] = 955,
  [1224] = 975,
  [1225] = 1017,
  [1226] = 956,
  [1227] = 976,
  [1228] = 981,
  [1229] = 1025,
  [1230] = 1027,
  [1231] = 982,
  [1232] = 983,
  [1233] = 953,
  [1234] = 1023,
  [1235] = 990,
  [1236] = 991,
  [1237] = 992,
  [1238] = 951,
  [1239] = 952,
  [1240] = 953,
  [1241] = 993,
  [1242] = 1002,
  [1243] = 955,
  [1244] = 956,
  [1245] = 957,
  [1246] = 959,
  [1247] = 999,
  [1248] = 961,
  [1249] = 962,
  [1250] = 1003,
  [1251] = 963,
  [1252] = 1004,
  [1253] = 964,
  [1254] = 965,
  [1255] = 966,
  [1256] = 940,
  [1257] = 968,
  [1258] = 932,
  [1259] = 972,
  [1260] = 973,
  [1261] = 974,
  [1262] = 975,
  [1263] = 976,
  [1264] = 1030,
  [1265] = 985,
  [1266] = 986,
  [1267] = 988,
  [1268] = 980,
  [1269] = 981,
  [1270] = 982,
  [1271] = 983,
  [1272] = 992,
  [1273] = 995,
  [1274] = 997,
  [1275] = 984,
  [1276] = 1000,
  [1277] = 1001,
  [1278] = 1002,
  [1279] = 1005,
  [1280] = 1006,
  [1281] = 1007,
  [1282] = 987,
  [1283] = 989,
  [1284] = 998,
  [1285] = 978,
  [1286] = 929,
  [1287] = 954,
  [1288] = 960,
  [1289] = 934,
  [1290] = 935,
  [1291] = 936,
  [1292] = 937,
  [1293] = 938,
  [1294] = 939,
  [1295] = 946,
  [1296] = 969,
  [1297] = 725,
  [1298] = 958,
  [1299] = 942,
  [1300] = 970,
  [1301] = 971,
  [1302] = 977,
  [1303] = 979,
  [1304] = 994,
  [1305] = 941,
  [1306] = 930,
  [1307] = 997,
  [1308] = 996,
  [1309] = 943,
  [1310] = 944,
  [1311] = 1008,
  [1312] = 945,
  [1313] = 947,
  [1314] = 1009,
  [1315] = 948,
  [1316] = 1010,
  [1317] = 949,
  [1318] = 1089,
  [1319] = 967,
  [1320] = 1137,
  [1321] = 950,
  [1322] = 1011,
  [1323] = 1023,
  [1324] = 930,
  [1325] = 1004,
  [1326] = 934,
  [1327] = 929,
  [1328] = 1016,
  [1329] = 940,
  [1330] = 952,
  [1331] = 968,
  [1332] = 985,
  [1333] = 1008,
  [1334] = 986,
  [1335] = 937,
  [1336] = 988,
  [1337] = 1010,
  [1338] = 972,
  [1339] = 973,
  [1340] = 974,
  [1341] = 992,
  [1342] = 975,
  [1343] = 976,
  [1344] = 995,
  [1345] = 979,
  [1346] = 954,
  [1347] = 1089,
  [1348] = 997,
  [1349] = 980,
  [1350] = 981,
  [1351] = 982,
  [1352] = 1000,
  [1353] = 1001,
  [1354] = 932,
  [1355] = 935,
  [1356] = 1002,
  [1357] = 983,
  [1358] = 984,
  [1359] = 938,
  [1360] = 949,
  [1361] = 967,
  [1362] = 883,
  [1363] = 987,
  [1364] = 946,
  [1365] = 941,
  [1366] = 1011,
  [1367] = 998,
  [1368] = 1005,
  [1369] = 1006,
  [1370] = 953,
  [1371] = 948,
  [1372] = 1009,
  [1373] = 939,
  [1374] = 994,
  [1375] = 990,
  [1376] = 1007,
  [1377] = 958,
  [1378] = 989,
  [1379] = 970,
  [1380] = 971,
  [1381] = 977,
  [1382] = 950,
  [1383] = 1017,
  [1384] = 991,
  [1385] = 943,
  [1386] = 1137,
  [1387] = 944,
  [1388] = 725,
  [1389] = 725,
  [1390] = 945,
  [1391] = 996,
  [1392] = 993,
  [1393] = 942,
  [1394] = 956,
  [1395] = 957,
  [1396] = 936,
  [1397] = 959,
  [1398] = 960,
  [1399] = 961,
  [1400] = 962,
  [1401] = 947,
  [1402] = 978,
  [1403] = 951,
  [1404] = 963,
  [1405] = 964,
  [1406] = 1126,
  [1407] = 965,
  [1408] = 999,
  [1409] = 966,
  [1410] = 969,
  [1411] = 1003,
  [1412] = 1030,
  [1413] = 955,
  [1414] = 951,
  [1415] = 969,
  [1416] = 1137,
  [1417] = 1017,
  [1418] = 958,
  [1419] = 1000,
  [1420] = 982,
  [1421] = 983,
  [1422] = 950,
  [1423] = 941,
  [1424] = 1010,
  [1425] = 978,
  [1426] = 993,
  [1427] = 984,
  [1428] = 967,
  [1429] = 947,
  [1430] = 979,
  [1431] = 991,
  [1432] = 995,
  [1433] = 996,
  [1434] = 976,
  [1435] = 994,
  [1436] = 987,
  [1437] = 955,
  [1438] = 956,
  [1439] = 957,
  [1440] = 999,
  [1441] = 988,
  [1442] = 1011,
  [1443] = 970,
  [1444] = 959,
  [1445] = 973,
  [1446] = 998,
  [1447] = 883,
  [1448] = 883,
  [1449] = 1001,
  [1450] = 1008,
  [1451] = 943,
  [1452] = 945,
  [1453] = 944,
  [1454] = 972,
  [1455] = 1009,
  [1456] = 980,
  [1457] = 1126,
  [1458] = 952,
  [1459] = 948,
  [1460] = 974,
  [1461] = 1003,
  [1462] = 971,
  [1463] = 953,
  [1464] = 960,
  [1465] = 961,
  [1466] = 962,
  [1467] = 1005,
  [1468] = 1006,
  [1469] = 954,
  [1470] = 985,
  [1471] = 946,
  [1472] = 977,
  [1473] = 986,
  [1474] = 1004,
  [1475] = 963,
  [1476] = 964,
  [1477] = 1134,
  [1478] = 965,
  [1479] = 966,
  [1480] = 940,
  [1481] = 1016,
  [1482] = 975,
  [1483] = 989,
  [1484] = 990,
  [1485] = 981,
  [1486] = 942,
  [1487] = 968,
  [1488] = 1007,
  [1489] = 949,
  [1490] = 1023,
  [1491] = 1016,
  [1492] = 1492,
  [1493] = 1492,
  [1494] = 1492,
  [1495] = 1492,
  [1496] = 1030,
  [1497] = 1492,
  [1498] = 1492,
  [1499] = 1492,
  [1500] = 1492,
  [1501] = 1492,
  [1502] = 1492,
  [1503] = 1492,
  [1504] = 1492,
  [1505] = 1492,
  [1506] = 1134,
  [1507] = 1016,
  [1508] = 1508,
  [1509] = 1509,
  [1510] = 1508,
  [1511] = 1508,
  [1512] = 1512,
  [1513] = 1513,
  [1514] = 1514,
  [1515] = 1509,
  [1516] = 1508,
  [1517] = 1512,
  [1518] = 1513,
  [1519] = 1514,
  [1520] = 1509,
  [1521] = 1508,
  [1522] = 1514,
  [1523] = 1512,
  [1524] = 1509,
  [1525] = 1508,
  [1526] = 1512,
  [1527] = 1513,
  [1528] = 1513,
  [1529] = 1512,
  [1530] = 1513,
  [1531] = 1509,
  [1532] = 1514,
  [1533] = 1514,
  [1534] = 1534,
  [1535] = 1509,
  [1536] = 1508,
  [1537] = 1537,
  [1538] = 1512,
  [1539] = 1513,
  [1540] = 1514,
  [1541] = 1509,
  [1542] = 1542,
  [1543] = 1512,
  [1544] = 1513,
  [1545] = 1512,
  [1546] = 1513,
  [1547] = 905,
  [1548] = 1514,
  [1549] = 1514,
  [1550] = 1509,
  [1551] = 1508,
  [1552] = 1512,
  [1553] = 1513,
  [1554] = 1514,
  [1555] = 1509,
  [1556] = 1513,
  [1557] = 1508,
  [1558] = 1512,
  [1559] = 1513,
  [1560] = 1514,
  [1561] = 1561,
  [1562] = 1509,
  [1563] = 1508,
  [1564] = 1512,
  [1565] = 1513,
  [1566] = 1509,
  [1567] = 1514,
  [1568] = 1514,
  [1569] = 1569,
  [1570] = 1570,
  [1571] = 1509,
  [1572] = 1508,
  [1573] = 1512,
  [1574] = 1513,
  [1575] = 1514,
  [1576] = 1509,
  [1577] = 1508,
  [1578] = 1512,
  [1579] = 1508,
  [1580] = 1580,
  [1581] = 1581,
  [1582] = 1582,
  [1583] = 1583,
  [1584] = 1584,
  [1585] = 1582,
  [1586] = 1583,
  [1587] = 1587,
  [1588] = 1588,
  [1589] = 1589,
  [1590] = 1581,
  [1591] = 1591,
  [1592] = 1584,
  [1593] = 1582,
  [1594] = 1594,
  [1595] = 1584,
  [1596] = 1581,
  [1597] = 1584,
  [1598] = 1598,
  [1599] = 1583,
  [1600] = 1582,
  [1601] = 1582,
  [1602] = 1588,
  [1603] = 1589,
  [1604] = 1582,
  [1605] = 1591,
  [1606] = 1606,
  [1607] = 1594,
  [1608] = 1606,
  [1609] = 1594,
  [1610] = 1598,
  [1611] = 1583,
  [1612] = 1584,
  [1613] = 1588,
  [1614] = 1606,
  [1615] = 1594,
  [1616] = 1583,
  [1617] = 1589,
  [1618] = 1588,
  [1619] = 1589,
  [1620] = 1620,
  [1621] = 1580,
  [1622] = 1591,
  [1623] = 1620,
  [1624] = 1580,
  [1625] = 1089,
  [1626] = 1598,
  [1627] = 1598,
  [1628] = 1584,
  [1629] = 1598,
  [1630] = 1582,
  [1631] = 1591,
  [1632] = 1583,
  [1633] = 725,
  [1634] = 1620,
  [1635] = 1580,
  [1636] = 1591,
  [1637] = 1581,
  [1638] = 1606,
  [1639] = 1126,
  [1640] = 1640,
  [1641] = 1641,
  [1642] = 1642,
  [1643] = 1643,
  [1644] = 1642,
  [1645] = 1645,
  [1646] = 1646,
  [1647] = 1641,
  [1648] = 1646,
  [1649] = 1642,
  [1650] = 1645,
  [1651] = 1642,
  [1652] = 1641,
  [1653] = 1641,
  [1654] = 1641,
  [1655] = 1641,
  [1656] = 1645,
  [1657] = 1641,
  [1658] = 1641,
  [1659] = 1642,
  [1660] = 1642,
  [1661] = 1661,
  [1662] = 1662,
  [1663] = 1662,
  [1664] = 1664,
  [1665] = 1645,
  [1666] = 1642,
  [1667] = 1664,
  [1668] = 1661,
  [1669] = 1641,
  [1670] = 1642,
  [1671] = 1641,
  [1672] = 1645,
  [1673] = 1640,
  [1674] = 883,
  [1675] = 1641,
  [1676] = 1642,
  [1677] = 1642,
  [1678] = 1661,
  [1679] = 1642,
  [1680] = 1643,
  [1681] = 1681,
  [1682] = 1682,
  [1683] = 1683,
  [1684] = 1681,
  [1685] = 1685,
  [1686] = 1681,
  [1687] = 1687,
  [1688] = 1688,
  [1689] = 1685,
  [1690] = 1687,
  [1691] = 1691,
  [1692] = 1692,
  [1693] = 1693,
  [1694] = 1694,
  [1695] = 1695,
  [1696] = 1696,
  [1697] = 1697,
  [1698] = 1695,
  [1699] = 1697,
  [1700] = 1700,
  [1701] = 1700,
  [1702] = 1702,
  [1703] = 1695,
  [1704] = 1685,
  [1705] = 1697,
  [1706] = 1706,
  [1707] = 1707,
  [1708] = 1700,
  [1709] = 1681,
  [1710] = 1685,
  [1711] = 1694,
  [1712] = 1712,
  [1713] = 1713,
  [1714] = 1714,
  [1715] = 1691,
  [1716] = 1137,
  [1717] = 1717,
  [1718] = 1696,
  [1719] = 1692,
  [1720] = 1696,
  [1721] = 1695,
  [1722] = 1697,
  [1723] = 1700,
  [1724] = 1683,
  [1725] = 1692,
  [1726] = 1726,
  [1727] = 1727,
  [1728] = 1683,
  [1729] = 1685,
  [1730] = 1706,
  [1731] = 1714,
  [1732] = 1707,
  [1733] = 1682,
  [1734] = 1712,
  [1735] = 1691,
  [1736] = 1692,
  [1737] = 1693,
  [1738] = 1694,
  [1739] = 1696,
  [1740] = 1688,
  [1741] = 1695,
  [1742] = 1681,
  [1743] = 1714,
  [1744] = 1697,
  [1745] = 1712,
  [1746] = 1700,
  [1747] = 1714,
  [1748] = 1712,
  [1749] = 1714,
  [1750] = 1712,
  [1751] = 1714,
  [1752] = 1712,
  [1753] = 1714,
  [1754] = 1712,
  [1755] = 1714,
  [1756] = 1712,
  [1757] = 1714,
  [1758] = 1712,
  [1759] = 1714,
  [1760] = 1712,
  [1761] = 1714,
  [1762] = 1712,
  [1763] = 1714,
  [1764] = 1712,
  [1765] = 1714,
  [1766] = 1685,
  [1767] = 1712,
  [1768] = 1714,
  [1769] = 1712,
  [1770] = 1693,
  [1771] = 1683,
  [1772] = 1681,
  [1773] = 1137,
  [1774] = 1137,
  [1775] = 1775,
  [1776] = 1776,
  [1777] = 1777,
  [1778] = 1776,
  [1779] = 1776,
  [1780] = 1775,
  [1781] = 1776,
  [1782] = 1775,
  [1783] = 1776,
  [1784] = 1775,
  [1785] = 1776,
  [1786] = 1775,
  [1787] = 1775,
  [1788] = 1775,
  [1789] = 1776,
  [1790] = 1775,
  [1791] = 1775,
  [1792] = 1775,
  [1793] = 1775,
  [1794] = 1775,
  [1795] = 1775,
  [1796] = 1776,
  [1797] = 1776,
  [1798] = 1777,
  [1799] = 1776,
  [1800] = 1137,
  [1801] = 1776,
  [1802] = 1776,
  [1803] = 1776,
  [1804] = 1777,
  [1805] = 1776,
  [1806] = 1775,
  [1807] = 1807,
  [1808] = 1807,
  [1809] = 1809,
  [1810] = 1807,
  [1811] = 1811,
  [1812] = 1811,
  [1813] = 1809,
  [1814] = 1807,
  [1815] = 1815,
  [1816] = 1811,
  [1817] = 1815,
  [1818] = 1809,
  [1819] = 1815,
  [1820] = 1807,
  [1821] = 1811,
  [1822] = 1815,
  [1823] = 1811,
  [1824] = 1815,
  [1825] = 1809,
  [1826] = 1809,
  [1827] = 1827,
  [1828] = 1828,
  [1829] = 1022,
  [1830] = 1025,
  [1831] = 1027,
  [1832] = 1028,
  [1833] = 1827,
  [1834] = 1827,
  [1835] = 1827,
  [1836] = 1836,
  [1837] = 1040,
  [1838] = 1836,
  [1839] = 1828,
  [1840] = 1836,
  [1841] = 1024,
  [1842] = 1021,
  [1843] = 1827,
  [1844] = 1827,
  [1845] = 1039,
  [1846] = 1827,
  [1847] = 1827,
  [1848] = 1827,
  [1849] = 1828,
  [1850] = 1020,
  [1851] = 1828,
  [1852] = 1029,
  [1853] = 1827,
  [1854] = 1827,
  [1855] = 1828,
  [1856] = 1827,
  [1857] = 1827,
  [1858] = 1836,
  [1859] = 1836,
  [1860] = 1860,
  [1861] = 1860,
  [1862] = 1860,
  [1863] = 1860,
  [1864] = 1864,
  [1865] = 1860,
  [1866] = 1860,
  [1867] = 1860,
  [1868] = 1860,
  [1869] = 1860,
  [1870] = 1860,
  [1871] = 1860,
  [1872] = 1860,
  [1873] = 1873,
  [1874] = 1860,
  [1875] = 1875,
  [1876] = 1860,
  [1877] = 1877,
  [1878] = 1877,
  [1879] = 1879,
  [1880] = 1880,
  [1881] = 1881,
  [1882] = 1882,
  [1883] = 1880,
  [1884] = 1877,
  [1885] = 1877,
  [1886] = 1879,
  [1887] = 1881,
  [1888] = 1888,
  [1889] = 1882,
  [1890] = 1877,
  [1891] = 1881,
  [1892] = 1882,
  [1893] = 1879,
  [1894] = 1879,
  [1895] = 1879,
  [1896] = 1879,
  [1897] = 1879,
  [1898] = 1880,
  [1899] = 1880,
  [1900] = 1888,
  [1901] = 1879,
  [1902] = 1880,
  [1903] = 1888,
  [1904] = 1879,
  [1905] = 1888,
  [1906] = 1879,
  [1907] = 1879,
  [1908] = 1880,
  [1909] = 1880,
  [1910] = 1880,
  [1911] = 1877,
  [1912] = 1879,
  [1913] = 1882,
  [1914] = 1881,
  [1915] = 1880,
  [1916] = 1881,
  [1917] = 1880,
  [1918] = 1882,
  [1919] = 1880,
  [1920] = 1880,
  [1921] = 1888,
  [1922] = 1888,
  [1923] = 1888,
  [1924] = 1879,
  [1925] = 1880,
  [1926] = 1888,
  [1927] = 1879,
  [1928] = 1880,
  [1929] = 1929,
  [1930] = 1930,
  [1931] = 1931,
  [1932] = 1017,
  [1933] = 1930,
  [1934] = 1930,
  [1935] = 1930,
  [1936] = 1930,
  [1937] = 934,
  [1938] = 935,
  [1939] = 1939,
  [1940] = 936,
  [1941] = 937,
  [1942] = 938,
  [1943] = 939,
  [1944] = 1137,
  [1945] = 930,
  [1946] = 932,
  [1947] = 929,
  [1948] = 997,
  [1949] = 970,
  [1950] = 977,
  [1951] = 954,
  [1952] = 979,
  [1953] = 1137,
  [1954] = 994,
  [1955] = 953,
  [1956] = 959,
  [1957] = 978,
  [1958] = 941,
  [1959] = 960,
  [1960] = 942,
  [1961] = 1009,
  [1962] = 961,
  [1963] = 962,
  [1964] = 1010,
  [1965] = 985,
  [1966] = 986,
  [1967] = 987,
  [1968] = 988,
  [1969] = 990,
  [1970] = 996,
  [1971] = 963,
  [1972] = 964,
  [1973] = 992,
  [1974] = 965,
  [1975] = 966,
  [1976] = 967,
  [1977] = 946,
  [1978] = 940,
  [1979] = 995,
  [1980] = 1008,
  [1981] = 1001,
  [1982] = 1002,
  [1983] = 1003,
  [1984] = 1004,
  [1985] = 971,
  [1986] = 991,
  [1987] = 1987,
  [1988] = 974,
  [1989] = 975,
  [1990] = 976,
  [1991] = 943,
  [1992] = 944,
  [1993] = 981,
  [1994] = 993,
  [1995] = 982,
  [1996] = 983,
  [1997] = 972,
  [1998] = 973,
  [1999] = 947,
  [2000] = 984,
  [2001] = 945,
  [2002] = 958,
  [2003] = 948,
  [2004] = 1000,
  [2005] = 949,
  [2006] = 955,
  [2007] = 1005,
  [2008] = 1006,
  [2009] = 1007,
  [2010] = 956,
  [2011] = 950,
  [2012] = 980,
  [2013] = 957,
  [2014] = 999,
  [2015] = 951,
  [2016] = 1137,
  [2017] = 952,
  [2018] = 1011,
  [2019] = 989,
  [2020] = 998,
  [2021] = 969,
  [2022] = 968,
  [2023] = 2023,
  [2024] = 2024,
  [2025] = 2023,
  [2026] = 2024,
  [2027] = 2023,
  [2028] = 2023,
  [2029] = 2023,
  [2030] = 2023,
  [2031] = 2024,
  [2032] = 2023,
  [2033] = 2024,
  [2034] = 2023,
  [2035] = 2024,
  [2036] = 2036,
  [2037] = 2024,
  [2038] = 2024,
  [2039] = 2023,
  [2040] = 2023,
  [2041] = 2023,
  [2042] = 2024,
  [2043] = 1137,
  [2044] = 2023,
  [2045] = 2023,
  [2046] = 2024,
  [2047] = 2024,
  [2048] = 2023,
  [2049] = 2024,
  [2050] = 2024,
  [2051] = 2024,
  [2052] = 2023,
  [2053] = 2023,
  [2054] = 2024,
  [2055] = 2024,
  [2056] = 2024,
  [2057] = 2057,
  [2058] = 2057,
  [2059] = 2057,
  [2060] = 987,
  [2061] = 2057,
  [2062] = 2057,
  [2063] = 2057,
  [2064] = 2057,
  [2065] = 1016,
  [2066] = 2057,
  [2067] = 2057,
  [2068] = 2057,
  [2069] = 2057,
  [2070] = 2057,
  [2071] = 2057,
  [2072] = 1003,
  [2073] = 2057,
  [2074] = 1137,
  [2075] = 2057,
  [2076] = 2076,
  [2077] = 2057,
  [2078] = 1030,
  [2079] = 2079,
  [2080] = 2080,
  [2081] = 1023,
  [2082] = 2082,
  [2083] = 2080,
  [2084] = 2084,
  [2085] = 2085,
  [2086] = 2080,
  [2087] = 2087,
  [2088] = 990,
  [2089] = 2080,
  [2090] = 2090,
  [2091] = 2091,
  [2092] = 2092,
  [2093] = 2093,
  [2094] = 2094,
  [2095] = 2080,
  [2096] = 2096,
  [2097] = 2097,
  [2098] = 2098,
  [2099] = 2099,
  [2100] = 1004,
  [2101] = 2101,
  [2102] = 1137,
  [2103] = 2103,
  [2104] = 2104,
  [2105] = 993,
  [2106] = 995,
  [2107] = 993,
  [2108] = 2108,
  [2109] = 2109,
  [2110] = 2110,
  [2111] = 2087,
  [2112] = 2091,
  [2113] = 2113,
  [2114] = 2114,
  [2115] = 2109,
  [2116] = 2097,
  [2117] = 2108,
  [2118] = 2108,
  [2119] = 2108,
  [2120] = 2098,
  [2121] = 2097,
  [2122] = 990,
  [2123] = 988,
  [2124] = 2087,
  [2125] = 2091,
  [2126] = 987,
  [2127] = 997,
  [2128] = 2084,
  [2129] = 2092,
  [2130] = 2093,
  [2131] = 2094,
  [2132] = 2101,
  [2133] = 2133,
  [2134] = 2134,
  [2135] = 2108,
  [2136] = 2079,
  [2137] = 1873,
  [2138] = 2113,
  [2139] = 2096,
  [2140] = 2084,
  [2141] = 2090,
  [2142] = 2108,
  [2143] = 986,
  [2144] = 2108,
  [2145] = 2110,
  [2146] = 2092,
  [2147] = 2093,
  [2148] = 2110,
  [2149] = 2113,
  [2150] = 2109,
  [2151] = 2108,
  [2152] = 2108,
  [2153] = 1001,
  [2154] = 2154,
  [2155] = 2113,
  [2156] = 2094,
  [2157] = 2109,
  [2158] = 2085,
  [2159] = 1864,
  [2160] = 1003,
  [2161] = 1002,
  [2162] = 2109,
  [2163] = 2103,
  [2164] = 1875,
  [2165] = 2104,
  [2166] = 2108,
  [2167] = 2108,
  [2168] = 2099,
  [2169] = 2085,
  [2170] = 2101,
  [2171] = 2108,
  [2172] = 2103,
  [2173] = 985,
  [2174] = 1004,
  [2175] = 2108,
  [2176] = 2108,
  [2177] = 2104,
  [2178] = 992,
  [2179] = 2108,
  [2180] = 2082,
  [2181] = 2098,
  [2182] = 2082,
  [2183] = 2113,
  [2184] = 2079,
  [2185] = 2108,
  [2186] = 2096,
  [2187] = 2090,
  [2188] = 2099,
  [2189] = 975,
  [2190] = 960,
  [2191] = 1003,
  [2192] = 940,
  [2193] = 992,
  [2194] = 974,
  [2195] = 999,
  [2196] = 991,
  [2197] = 1005,
  [2198] = 970,
  [2199] = 2084,
  [2200] = 971,
  [2201] = 977,
  [2202] = 2202,
  [2203] = 2203,
  [2204] = 978,
  [2205] = 979,
  [2206] = 975,
  [2207] = 976,
  [2208] = 2208,
  [2209] = 2085,
  [2210] = 994,
  [2211] = 2090,
  [2212] = 981,
  [2213] = 941,
  [2214] = 2214,
  [2215] = 2215,
  [2216] = 995,
  [2217] = 2092,
  [2218] = 2218,
  [2219] = 982,
  [2220] = 996,
  [2221] = 2093,
  [2222] = 2084,
  [2223] = 2094,
  [2224] = 983,
  [2225] = 984,
  [2226] = 948,
  [2227] = 2099,
  [2228] = 1004,
  [2229] = 1007,
  [2230] = 2101,
  [2231] = 987,
  [2232] = 2103,
  [2233] = 2104,
  [2234] = 2098,
  [2235] = 2082,
  [2236] = 929,
  [2237] = 2079,
  [2238] = 2238,
  [2239] = 937,
  [2240] = 938,
  [2241] = 939,
  [2242] = 1003,
  [2243] = 930,
  [2244] = 2087,
  [2245] = 2091,
  [2246] = 2246,
  [2247] = 2247,
  [2248] = 2248,
  [2249] = 2249,
  [2250] = 980,
  [2251] = 2251,
  [2252] = 990,
  [2253] = 996,
  [2254] = 986,
  [2255] = 2097,
  [2256] = 1008,
  [2257] = 2257,
  [2258] = 943,
  [2259] = 949,
  [2260] = 944,
  [2261] = 945,
  [2262] = 950,
  [2263] = 1006,
  [2264] = 947,
  [2265] = 2265,
  [2266] = 2266,
  [2267] = 997,
  [2268] = 1001,
  [2269] = 2269,
  [2270] = 948,
  [2271] = 999,
  [2272] = 989,
  [2273] = 2096,
  [2274] = 2096,
  [2275] = 2275,
  [2276] = 2276,
  [2277] = 970,
  [2278] = 2278,
  [2279] = 988,
  [2280] = 991,
  [2281] = 932,
  [2282] = 1002,
  [2283] = 998,
  [2284] = 949,
  [2285] = 950,
  [2286] = 951,
  [2287] = 952,
  [2288] = 2288,
  [2289] = 2289,
  [2290] = 2290,
  [2291] = 2291,
  [2292] = 1008,
  [2293] = 953,
  [2294] = 1009,
  [2295] = 2295,
  [2296] = 977,
  [2297] = 1011,
  [2298] = 955,
  [2299] = 998,
  [2300] = 956,
  [2301] = 957,
  [2302] = 929,
  [2303] = 2276,
  [2304] = 2304,
  [2305] = 979,
  [2306] = 961,
  [2307] = 971,
  [2308] = 978,
  [2309] = 962,
  [2310] = 994,
  [2311] = 1007,
  [2312] = 2312,
  [2313] = 998,
  [2314] = 963,
  [2315] = 964,
  [2316] = 1005,
  [2317] = 932,
  [2318] = 951,
  [2319] = 2319,
  [2320] = 1009,
  [2321] = 965,
  [2322] = 2322,
  [2323] = 966,
  [2324] = 1006,
  [2325] = 959,
  [2326] = 2202,
  [2327] = 2203,
  [2328] = 968,
  [2329] = 953,
  [2330] = 2330,
  [2331] = 1008,
  [2332] = 954,
  [2333] = 1009,
  [2334] = 934,
  [2335] = 2335,
  [2336] = 2336,
  [2337] = 1010,
  [2338] = 2278,
  [2339] = 935,
  [2340] = 936,
  [2341] = 2304,
  [2342] = 955,
  [2343] = 1010,
  [2344] = 967,
  [2345] = 1004,
  [2346] = 992,
  [2347] = 941,
  [2348] = 956,
  [2349] = 957,
  [2350] = 999,
  [2351] = 1001,
  [2352] = 1000,
  [2353] = 2353,
  [2354] = 959,
  [2355] = 972,
  [2356] = 973,
  [2357] = 987,
  [2358] = 2358,
  [2359] = 1010,
  [2360] = 2360,
  [2361] = 967,
  [2362] = 2362,
  [2363] = 997,
  [2364] = 960,
  [2365] = 961,
  [2366] = 989,
  [2367] = 946,
  [2368] = 1000,
  [2369] = 962,
  [2370] = 996,
  [2371] = 998,
  [2372] = 2085,
  [2373] = 1002,
  [2374] = 946,
  [2375] = 1005,
  [2376] = 2376,
  [2377] = 2377,
  [2378] = 2378,
  [2379] = 2379,
  [2380] = 2380,
  [2381] = 2381,
  [2382] = 930,
  [2383] = 2383,
  [2384] = 2384,
  [2385] = 2385,
  [2386] = 967,
  [2387] = 2387,
  [2388] = 993,
  [2389] = 985,
  [2390] = 963,
  [2391] = 964,
  [2392] = 965,
  [2393] = 991,
  [2394] = 966,
  [2395] = 940,
  [2396] = 995,
  [2397] = 952,
  [2398] = 968,
  [2399] = 990,
  [2400] = 985,
  [2401] = 1003,
  [2402] = 934,
  [2403] = 958,
  [2404] = 935,
  [2405] = 936,
  [2406] = 2406,
  [2407] = 947,
  [2408] = 2408,
  [2409] = 972,
  [2410] = 943,
  [2411] = 973,
  [2412] = 974,
  [2413] = 989,
  [2414] = 1000,
  [2415] = 976,
  [2416] = 986,
  [2417] = 2417,
  [2418] = 937,
  [2419] = 938,
  [2420] = 939,
  [2421] = 1006,
  [2422] = 2422,
  [2423] = 980,
  [2424] = 958,
  [2425] = 2425,
  [2426] = 2426,
  [2427] = 2427,
  [2428] = 2428,
  [2429] = 981,
  [2430] = 982,
  [2431] = 954,
  [2432] = 988,
  [2433] = 1007,
  [2434] = 983,
  [2435] = 2435,
  [2436] = 984,
  [2437] = 944,
  [2438] = 2438,
  [2439] = 1011,
  [2440] = 987,
  [2441] = 2441,
  [2442] = 2442,
  [2443] = 945,
  [2444] = 2444,
  [2445] = 978,
  [2446] = 1007,
  [2447] = 2202,
  [2448] = 2203,
  [2449] = 946,
  [2450] = 958,
  [2451] = 989,
  [2452] = 970,
  [2453] = 977,
  [2454] = 979,
  [2455] = 994,
  [2456] = 941,
  [2457] = 943,
  [2458] = 944,
  [2459] = 945,
  [2460] = 948,
  [2461] = 949,
  [2462] = 950,
  [2463] = 932,
  [2464] = 951,
  [2465] = 952,
  [2466] = 953,
  [2467] = 955,
  [2468] = 956,
  [2469] = 957,
  [2470] = 961,
  [2471] = 962,
  [2472] = 963,
  [2473] = 964,
  [2474] = 965,
  [2475] = 966,
  [2476] = 968,
  [2477] = 934,
  [2478] = 935,
  [2479] = 936,
  [2480] = 972,
  [2481] = 973,
  [2482] = 937,
  [2483] = 938,
  [2484] = 939,
  [2485] = 980,
  [2486] = 1011,
  [2487] = 971,
  [2488] = 930,
  [2489] = 947,
  [2490] = 954,
  [2491] = 959,
  [2492] = 960,
  [2493] = 929,
  [2494] = 940,
  [2495] = 974,
  [2496] = 975,
  [2497] = 976,
  [2498] = 981,
  [2499] = 982,
  [2500] = 983,
  [2501] = 984,
  [2502] = 2304,
  [2503] = 996,
  [2504] = 1008,
  [2505] = 1009,
  [2506] = 1010,
  [2507] = 967,
  [2508] = 989,
  [2509] = 998,
  [2510] = 991,
  [2511] = 993,
  [2512] = 999,
  [2513] = 985,
  [2514] = 986,
  [2515] = 987,
  [2516] = 988,
  [2517] = 990,
  [2518] = 996,
  [2519] = 1001,
  [2520] = 1003,
  [2521] = 1004,
  [2522] = 1008,
  [2523] = 1009,
  [2524] = 1010,
  [2525] = 967,
  [2526] = 992,
  [2527] = 997,
  [2528] = 1002,
  [2529] = 995,
  [2530] = 1000,
  [2531] = 1005,
  [2532] = 1006,
  [2533] = 1007,
  [2534] = 2090,
  [2535] = 2092,
  [2536] = 2093,
  [2537] = 2094,
  [2538] = 2084,
  [2539] = 2085,
  [2540] = 2099,
  [2541] = 2101,
  [2542] = 2103,
  [2543] = 2104,
  [2544] = 2082,
  [2545] = 2079,
  [2546] = 2087,
  [2547] = 2091,
  [2548] = 2096,
  [2549] = 2097,
  [2550] = 2096,
  [2551] = 946,
  [2552] = 958,
  [2553] = 1011,
  [2554] = 970,
  [2555] = 971,
  [2556] = 977,
  [2557] = 2202,
  [2558] = 2203,
  [2559] = 978,
  [2560] = 979,
  [2561] = 994,
  [2562] = 941,
  [2563] = 930,
  [2564] = 2096,
  [2565] = 943,
  [2566] = 944,
  [2567] = 945,
  [2568] = 947,
  [2569] = 948,
  [2570] = 2276,
  [2571] = 2278,
  [2572] = 949,
  [2573] = 950,
  [2574] = 932,
  [2575] = 951,
  [2576] = 952,
  [2577] = 953,
  [2578] = 954,
  [2579] = 2304,
  [2580] = 955,
  [2581] = 956,
  [2582] = 957,
  [2583] = 959,
  [2584] = 960,
  [2585] = 961,
  [2586] = 962,
  [2587] = 929,
  [2588] = 963,
  [2589] = 964,
  [2590] = 965,
  [2591] = 966,
  [2592] = 940,
  [2593] = 968,
  [2594] = 934,
  [2595] = 935,
  [2596] = 936,
  [2597] = 972,
  [2598] = 973,
  [2599] = 974,
  [2600] = 975,
  [2601] = 976,
  [2602] = 937,
  [2603] = 938,
  [2604] = 939,
  [2605] = 980,
  [2606] = 981,
  [2607] = 982,
  [2608] = 983,
  [2609] = 984,
  [2610] = 990,
  [2611] = 993,
  [2612] = 2084,
  [2613] = 1004,
  [2614] = 2085,
  [2615] = 985,
  [2616] = 986,
  [2617] = 988,
  [2618] = 991,
  [2619] = 992,
  [2620] = 993,
  [2621] = 995,
  [2622] = 997,
  [2623] = 999,
  [2624] = 1001,
  [2625] = 1002,
  [2626] = 2276,
  [2627] = 995,
  [2628] = 2278,
  [2629] = 1000,
  [2630] = 2084,
  [2631] = 1005,
  [2632] = 1006,
  [2633] = 2085,
  [2634] = 2098,
  [2635] = 2635,
  [2636] = 2085,
  [2637] = 2637,
  [2638] = 2638,
  [2639] = 2091,
  [2640] = 2640,
  [2641] = 2641,
  [2642] = 2642,
  [2643] = 2643,
  [2644] = 2644,
  [2645] = 2096,
  [2646] = 2646,
  [2647] = 2647,
  [2648] = 2648,
  [2649] = 2649,
  [2650] = 2099,
  [2651] = 2651,
  [2652] = 2652,
  [2653] = 2653,
  [2654] = 2654,
  [2655] = 2655,
  [2656] = 2101,
  [2657] = 2103,
  [2658] = 2096,
  [2659] = 2104,
  [2660] = 2098,
  [2661] = 2661,
  [2662] = 2092,
  [2663] = 2093,
  [2664] = 2094,
  [2665] = 2665,
  [2666] = 2082,
  [2667] = 2079,
  [2668] = 2087,
  [2669] = 2084,
  [2670] = 2670,
  [2671] = 2084,
  [2672] = 2085,
  [2673] = 2673,
  [2674] = 2673,
  [2675] = 2661,
  [2676] = 2676,
  [2677] = 2677,
  [2678] = 2678,
  [2679] = 2679,
  [2680] = 2097,
  [2681] = 2681,
  [2682] = 2682,
  [2683] = 2683,
  [2684] = 2684,
  [2685] = 2685,
  [2686] = 2673,
  [2687] = 2665,
  [2688] = 2688,
  [2689] = 2689,
  [2690] = 2690,
  [2691] = 2691,
  [2692] = 2692,
  [2693] = 2693,
  [2694] = 2694,
  [2695] = 2695,
  [2696] = 2696,
  [2697] = 2697,
  [2698] = 2698,
  [2699] = 2661,
  [2700] = 2090,
  [2701] = 2701,
  [2702] = 2702,
  [2703] = 2703,
  [2704] = 2704,
  [2705] = 2692,
  [2706] = 2685,
  [2707] = 2641,
  [2708] = 2643,
  [2709] = 2644,
  [2710] = 2665,
  [2711] = 2642,
  [2712] = 2649,
  [2713] = 2651,
  [2714] = 2693,
  [2715] = 2085,
  [2716] = 2635,
  [2717] = 2702,
  [2718] = 2653,
  [2719] = 2688,
  [2720] = 2689,
  [2721] = 2690,
  [2722] = 2655,
  [2723] = 2654,
  [2724] = 2691,
  [2725] = 2703,
  [2726] = 2655,
  [2727] = 2702,
  [2728] = 2084,
  [2729] = 2670,
  [2730] = 2652,
  [2731] = 2688,
  [2732] = 2689,
  [2733] = 2690,
  [2734] = 2697,
  [2735] = 2691,
  [2736] = 2638,
  [2737] = 2665,
  [2738] = 2640,
  [2739] = 2694,
  [2740] = 2698,
  [2741] = 2704,
  [2742] = 2096,
  [2743] = 2677,
  [2744] = 2744,
  [2745] = 2679,
  [2746] = 2681,
  [2747] = 2682,
  [2748] = 2683,
  [2749] = 2692,
  [2750] = 2750,
  [2751] = 2751,
  [2752] = 2744,
  [2753] = 2751,
  [2754] = 1873,
  [2755] = 2755,
  [2756] = 2704,
  [2757] = 2750,
  [2758] = 2758,
  [2759] = 2085,
  [2760] = 2362,
  [2761] = 2638,
  [2762] = 2265,
  [2763] = 2084,
  [2764] = 1875,
  [2765] = 2275,
  [2766] = 2154,
  [2767] = 2640,
  [2768] = 2768,
  [2769] = 2641,
  [2770] = 2768,
  [2771] = 2643,
  [2772] = 2772,
  [2773] = 2758,
  [2774] = 2644,
  [2775] = 2649,
  [2776] = 2755,
  [2777] = 2750,
  [2778] = 2778,
  [2779] = 2779,
  [2780] = 2653,
  [2781] = 2654,
  [2782] = 2684,
  [2783] = 985,
  [2784] = 986,
  [2785] = 988,
  [2786] = 2786,
  [2787] = 991,
  [2788] = 992,
  [2789] = 2789,
  [2790] = 2758,
  [2791] = 2791,
  [2792] = 2792,
  [2793] = 996,
  [2794] = 2702,
  [2795] = 2795,
  [2796] = 2796,
  [2797] = 997,
  [2798] = 2798,
  [2799] = 2677,
  [2800] = 2679,
  [2801] = 2694,
  [2802] = 2751,
  [2803] = 2681,
  [2804] = 999,
  [2805] = 1001,
  [2806] = 1002,
  [2807] = 2682,
  [2808] = 2750,
  [2809] = 2683,
  [2810] = 2214,
  [2811] = 2685,
  [2812] = 2215,
  [2813] = 2744,
  [2814] = 2814,
  [2815] = 2815,
  [2816] = 2816,
  [2817] = 2697,
  [2818] = 2758,
  [2819] = 2703,
  [2820] = 2655,
  [2821] = 2693,
  [2822] = 1008,
  [2823] = 1009,
  [2824] = 2698,
  [2825] = 2642,
  [2826] = 1010,
  [2827] = 2635,
  [2828] = 2768,
  [2829] = 2688,
  [2830] = 2689,
  [2831] = 2690,
  [2832] = 2096,
  [2833] = 2691,
  [2834] = 2692,
  [2835] = 1864,
  [2836] = 2652,
  [2837] = 2758,
  [2838] = 2670,
  [2839] = 2772,
  [2840] = 2134,
  [2841] = 1000,
  [2842] = 2651,
  [2843] = 2843,
  [2844] = 1005,
  [2845] = 1006,
  [2846] = 2843,
  [2847] = 1007,
  [2848] = 2848,
  [2849] = 2849,
  [2850] = 2850,
  [2851] = 2646,
  [2852] = 2852,
  [2853] = 2772,
  [2854] = 2843,
  [2855] = 2855,
  [2856] = 2751,
  [2857] = 2647,
  [2858] = 2755,
  [2859] = 2648,
  [2860] = 2758,
  [2861] = 2861,
  [2862] = 2862,
  [2863] = 2312,
  [2864] = 2864,
  [2865] = 2426,
  [2866] = 2866,
  [2867] = 2867,
  [2868] = 2868,
  [2869] = 2869,
  [2870] = 2870,
  [2871] = 2871,
  [2872] = 2872,
  [2873] = 2750,
  [2874] = 967,
  [2875] = 2875,
  [2876] = 2876,
  [2877] = 2702,
  [2878] = 2692,
  [2879] = 2691,
  [2880] = 2875,
  [2881] = 2689,
  [2882] = 2690,
  [2883] = 2876,
  [2884] = 2744,
  [2885] = 2876,
  [2886] = 2688,
  [2887] = 2875,
  [2888] = 2655,
  [2889] = 975,
  [2890] = 1006,
  [2891] = 2678,
  [2892] = 2701,
  [2893] = 1007,
  [2894] = 2676,
  [2895] = 2895,
  [2896] = 2691,
  [2897] = 959,
  [2898] = 2689,
  [2899] = 2695,
  [2900] = 2696,
  [2901] = 1005,
  [2902] = 978,
  [2903] = 2692,
  [2904] = 2688,
  [2905] = 2744,
  [2906] = 1000,
  [2907] = 940,
  [2908] = 971,
  [2909] = 974,
  [2910] = 1003,
  [2911] = 976,
  [2912] = 981,
  [2913] = 982,
  [2914] = 983,
  [2915] = 984,
  [2916] = 987,
  [2917] = 2690,
  [2918] = 2702,
  [2919] = 2655,
  [2920] = 960,
  [2921] = 947,
  [2922] = 996,
  [2923] = 1008,
  [2924] = 979,
  [2925] = 994,
  [2926] = 2855,
  [2927] = 2866,
  [2928] = 2852,
  [2929] = 997,
  [2930] = 947,
  [2931] = 941,
  [2932] = 1009,
  [2933] = 993,
  [2934] = 1000,
  [2935] = 1010,
  [2936] = 967,
  [2937] = 1001,
  [2938] = 1002,
  [2939] = 991,
  [2940] = 943,
  [2941] = 2815,
  [2942] = 2816,
  [2943] = 993,
  [2944] = 944,
  [2945] = 945,
  [2946] = 948,
  [2947] = 999,
  [2948] = 954,
  [2949] = 2850,
  [2950] = 2870,
  [2951] = 2778,
  [2952] = 2779,
  [2953] = 2786,
  [2954] = 959,
  [2955] = 960,
  [2956] = 949,
  [2957] = 1005,
  [2958] = 1006,
  [2959] = 2862,
  [2960] = 2795,
  [2961] = 2798,
  [2962] = 2868,
  [2963] = 940,
  [2964] = 950,
  [2965] = 1007,
  [2966] = 999,
  [2967] = 985,
  [2968] = 986,
  [2969] = 987,
  [2970] = 2849,
  [2971] = 988,
  [2972] = 2864,
  [2973] = 990,
  [2974] = 2872,
  [2975] = 995,
  [2976] = 974,
  [2977] = 975,
  [2978] = 976,
  [2979] = 1001,
  [2980] = 1003,
  [2981] = 1004,
  [2982] = 981,
  [2983] = 982,
  [2984] = 983,
  [2985] = 992,
  [2986] = 997,
  [2987] = 1002,
  [2988] = 2637,
  [2989] = 984,
  [2990] = 932,
  [2991] = 1000,
  [2992] = 951,
  [2993] = 952,
  [2994] = 1005,
  [2995] = 1006,
  [2996] = 1007,
  [2997] = 1009,
  [2998] = 2867,
  [2999] = 953,
  [3000] = 1004,
  [3001] = 955,
  [3002] = 956,
  [3003] = 957,
  [3004] = 961,
  [3005] = 929,
  [3006] = 962,
  [3007] = 963,
  [3008] = 2814,
  [3009] = 969,
  [3010] = 964,
  [3011] = 965,
  [3012] = 966,
  [3013] = 968,
  [3014] = 934,
  [3015] = 935,
  [3016] = 2362,
  [3017] = 936,
  [3018] = 972,
  [3019] = 973,
  [3020] = 937,
  [3021] = 938,
  [3022] = 2134,
  [3023] = 939,
  [3024] = 2791,
  [3025] = 2792,
  [3026] = 2796,
  [3027] = 996,
  [3028] = 980,
  [3029] = 2265,
  [3030] = 2275,
  [3031] = 1010,
  [3032] = 985,
  [3033] = 2154,
  [3034] = 986,
  [3035] = 967,
  [3036] = 988,
  [3037] = 1011,
  [3038] = 989,
  [3039] = 998,
  [3040] = 2426,
  [3041] = 2214,
  [3042] = 2215,
  [3043] = 2312,
  [3044] = 946,
  [3045] = 971,
  [3046] = 2848,
  [3047] = 1008,
  [3048] = 978,
  [3049] = 958,
  [3050] = 2744,
  [3051] = 992,
  [3052] = 2789,
  [3053] = 990,
  [3054] = 991,
  [3055] = 970,
  [3056] = 2861,
  [3057] = 995,
  [3058] = 942,
  [3059] = 2869,
  [3060] = 2871,
  [3061] = 977,
  [3062] = 930,
  [3063] = 2093,
  [3064] = 3064,
  [3065] = 2099,
  [3066] = 2101,
  [3067] = 2103,
  [3068] = 2104,
  [3069] = 2098,
  [3070] = 2082,
  [3071] = 2079,
  [3072] = 2087,
  [3073] = 2091,
  [3074] = 2097,
  [3075] = 2084,
  [3076] = 2085,
  [3077] = 2096,
  [3078] = 3064,
  [3079] = 1875,
  [3080] = 1873,
  [3081] = 3064,
  [3082] = 2090,
  [3083] = 2092,
  [3084] = 2094,
  [3085] = 1864,
  [3086] = 3086,
  [3087] = 2092,
  [3088] = 2094,
  [3089] = 2084,
  [3090] = 2090,
  [3091] = 2093,
  [3092] = 2099,
  [3093] = 2101,
  [3094] = 2103,
  [3095] = 2104,
  [3096] = 2098,
  [3097] = 2082,
  [3098] = 2079,
  [3099] = 2087,
  [3100] = 2091,
  [3101] = 3101,
  [3102] = 2097,
  [3103] = 1016,
  [3104] = 3086,
  [3105] = 3086,
  [3106] = 2096,
  [3107] = 2085,
  [3108] = 3101,
  [3109] = 3101,
  [3110] = 3110,
  [3111] = 2103,
  [3112] = 2104,
  [3113] = 2098,
  [3114] = 2101,
  [3115] = 2093,
  [3116] = 2094,
  [3117] = 2082,
  [3118] = 2079,
  [3119] = 2087,
  [3120] = 2091,
  [3121] = 2097,
  [3122] = 2099,
  [3123] = 3123,
  [3124] = 3124,
  [3125] = 1003,
  [3126] = 2084,
  [3127] = 2097,
  [3128] = 2090,
  [3129] = 2085,
  [3130] = 3130,
  [3131] = 3131,
  [3132] = 3132,
  [3133] = 3124,
  [3134] = 2084,
  [3135] = 1030,
  [3136] = 3123,
  [3137] = 2091,
  [3138] = 2096,
  [3139] = 3130,
  [3140] = 3140,
  [3141] = 3131,
  [3142] = 3124,
  [3143] = 2895,
  [3144] = 2085,
  [3145] = 2079,
  [3146] = 2101,
  [3147] = 2103,
  [3148] = 2104,
  [3149] = 3149,
  [3150] = 3130,
  [3151] = 2098,
  [3152] = 3132,
  [3153] = 2092,
  [3154] = 3154,
  [3155] = 3149,
  [3156] = 3123,
  [3157] = 3132,
  [3158] = 2093,
  [3159] = 2087,
  [3160] = 987,
  [3161] = 2082,
  [3162] = 3132,
  [3163] = 3163,
  [3164] = 2094,
  [3165] = 3154,
  [3166] = 3132,
  [3167] = 3149,
  [3168] = 3154,
  [3169] = 3154,
  [3170] = 2096,
  [3171] = 3131,
  [3172] = 3130,
  [3173] = 1023,
  [3174] = 2090,
  [3175] = 2099,
  [3176] = 3130,
  [3177] = 2092,
  [3178] = 3178,
  [3179] = 3179,
  [3180] = 2096,
  [3181] = 3178,
  [3182] = 1004,
  [3183] = 3183,
  [3184] = 3184,
  [3185] = 3185,
  [3186] = 3186,
  [3187] = 3187,
  [3188] = 3184,
  [3189] = 3185,
  [3190] = 3190,
  [3191] = 990,
  [3192] = 3192,
  [3193] = 3193,
  [3194] = 3194,
  [3195] = 3185,
  [3196] = 3178,
  [3197] = 3194,
  [3198] = 993,
  [3199] = 3184,
  [3200] = 3200,
  [3201] = 995,
  [3202] = 3185,
  [3203] = 2084,
  [3204] = 3194,
  [3205] = 3205,
  [3206] = 2085,
  [3207] = 3192,
  [3208] = 3193,
  [3209] = 3194,
  [3210] = 3192,
  [3211] = 3184,
  [3212] = 3212,
  [3213] = 3193,
  [3214] = 2689,
  [3215] = 3179,
  [3216] = 2702,
  [3217] = 2895,
  [3218] = 3212,
  [3219] = 3200,
  [3220] = 3220,
  [3221] = 3183,
  [3222] = 2691,
  [3223] = 2692,
  [3224] = 2688,
  [3225] = 3225,
  [3226] = 2690,
  [3227] = 2655,
  [3228] = 3190,
  [3229] = 3229,
  [3230] = 3230,
  [3231] = 2084,
  [3232] = 2702,
  [3233] = 2691,
  [3234] = 2096,
  [3235] = 3235,
  [3236] = 3236,
  [3237] = 3237,
  [3238] = 2744,
  [3239] = 3239,
  [3240] = 3240,
  [3241] = 3241,
  [3242] = 3242,
  [3243] = 3243,
  [3244] = 3244,
  [3245] = 3245,
  [3246] = 3246,
  [3247] = 2689,
  [3248] = 3248,
  [3249] = 2690,
  [3250] = 2655,
  [3251] = 3251,
  [3252] = 2692,
  [3253] = 2085,
  [3254] = 2688,
  [3255] = 973,
  [3256] = 971,
  [3257] = 974,
  [3258] = 975,
  [3259] = 976,
  [3260] = 962,
  [3261] = 981,
  [3262] = 982,
  [3263] = 983,
  [3264] = 1000,
  [3265] = 1005,
  [3266] = 1006,
  [3267] = 3267,
  [3268] = 978,
  [3269] = 984,
  [3270] = 1007,
  [3271] = 3271,
  [3272] = 3267,
  [3273] = 1001,
  [3274] = 977,
  [3275] = 2702,
  [3276] = 979,
  [3277] = 2702,
  [3278] = 963,
  [3279] = 964,
  [3280] = 965,
  [3281] = 955,
  [3282] = 966,
  [3283] = 1929,
  [3284] = 994,
  [3285] = 968,
  [3286] = 941,
  [3287] = 930,
  [3288] = 934,
  [3289] = 935,
  [3290] = 956,
  [3291] = 936,
  [3292] = 1931,
  [3293] = 972,
  [3294] = 3294,
  [3295] = 957,
  [3296] = 989,
  [3297] = 998,
  [3298] = 2744,
  [3299] = 3299,
  [3300] = 3300,
  [3301] = 3267,
  [3302] = 943,
  [3303] = 944,
  [3304] = 945,
  [3305] = 947,
  [3306] = 1939,
  [3307] = 3294,
  [3308] = 937,
  [3309] = 3271,
  [3310] = 958,
  [3311] = 996,
  [3312] = 948,
  [3313] = 2304,
  [3314] = 938,
  [3315] = 939,
  [3316] = 980,
  [3317] = 1008,
  [3318] = 992,
  [3319] = 949,
  [3320] = 1009,
  [3321] = 991,
  [3322] = 993,
  [3323] = 954,
  [3324] = 961,
  [3325] = 2202,
  [3326] = 2203,
  [3327] = 1010,
  [3328] = 997,
  [3329] = 1002,
  [3330] = 3271,
  [3331] = 959,
  [3332] = 946,
  [3333] = 950,
  [3334] = 987,
  [3335] = 990,
  [3336] = 2655,
  [3337] = 960,
  [3338] = 929,
  [3339] = 999,
  [3340] = 940,
  [3341] = 995,
  [3342] = 1003,
  [3343] = 1004,
  [3344] = 932,
  [3345] = 951,
  [3346] = 952,
  [3347] = 985,
  [3348] = 2655,
  [3349] = 3299,
  [3350] = 3294,
  [3351] = 986,
  [3352] = 988,
  [3353] = 3299,
  [3354] = 953,
  [3355] = 2362,
  [3356] = 2276,
  [3357] = 3299,
  [3358] = 987,
  [3359] = 2278,
  [3360] = 1011,
  [3361] = 967,
  [3362] = 3294,
  [3363] = 1003,
  [3364] = 970,
  [3365] = 3365,
  [3366] = 3366,
  [3367] = 3367,
  [3368] = 3368,
  [3369] = 3369,
  [3370] = 3370,
  [3371] = 3368,
  [3372] = 3370,
  [3373] = 3373,
  [3374] = 3374,
  [3375] = 3368,
  [3376] = 3376,
  [3377] = 3377,
  [3378] = 3376,
  [3379] = 3377,
  [3380] = 3380,
  [3381] = 3381,
  [3382] = 3382,
  [3383] = 3383,
  [3384] = 1987,
  [3385] = 3370,
  [3386] = 3381,
  [3387] = 3376,
  [3388] = 3380,
  [3389] = 3382,
  [3390] = 2087,
  [3391] = 3367,
  [3392] = 3365,
  [3393] = 995,
  [3394] = 3394,
  [3395] = 2097,
  [3396] = 2091,
  [3397] = 3377,
  [3398] = 3383,
  [3399] = 3374,
  [3400] = 3366,
  [3401] = 3367,
  [3402] = 3402,
  [3403] = 3403,
  [3404] = 3380,
  [3405] = 3377,
  [3406] = 3380,
  [3407] = 2096,
  [3408] = 3408,
  [3409] = 2079,
  [3410] = 3383,
  [3411] = 3411,
  [3412] = 3374,
  [3413] = 3381,
  [3414] = 3376,
  [3415] = 2084,
  [3416] = 3383,
  [3417] = 3383,
  [3418] = 2744,
  [3419] = 2265,
  [3420] = 3365,
  [3421] = 3366,
  [3422] = 2085,
  [3423] = 3376,
  [3424] = 2090,
  [3425] = 3373,
  [3426] = 3426,
  [3427] = 3373,
  [3428] = 2744,
  [3429] = 990,
  [3430] = 993,
  [3431] = 3373,
  [3432] = 2092,
  [3433] = 2099,
  [3434] = 3377,
  [3435] = 2101,
  [3436] = 2103,
  [3437] = 3380,
  [3438] = 2082,
  [3439] = 2104,
  [3440] = 2093,
  [3441] = 3366,
  [3442] = 2098,
  [3443] = 1004,
  [3444] = 2094,
  [3445] = 3365,
  [3446] = 3367,
  [3447] = 2093,
  [3448] = 1929,
  [3449] = 2665,
  [3450] = 2090,
  [3451] = 1931,
  [3452] = 2099,
  [3453] = 2092,
  [3454] = 2094,
  [3455] = 1939,
  [3456] = 3403,
  [3457] = 2084,
  [3458] = 2101,
  [3459] = 3459,
  [3460] = 2085,
  [3461] = 2096,
  [3462] = 2103,
  [3463] = 2104,
  [3464] = 2098,
  [3465] = 2082,
  [3466] = 2079,
  [3467] = 2087,
  [3468] = 3468,
  [3469] = 1000,
  [3470] = 2099,
  [3471] = 2085,
  [3472] = 2103,
  [3473] = 3411,
  [3474] = 2104,
  [3475] = 2098,
  [3476] = 2082,
  [3477] = 2079,
  [3478] = 2087,
  [3479] = 2091,
  [3480] = 2091,
  [3481] = 1005,
  [3482] = 1006,
  [3483] = 2097,
  [3484] = 1007,
  [3485] = 2090,
  [3486] = 2092,
  [3487] = 2093,
  [3488] = 2094,
  [3489] = 2097,
  [3490] = 3225,
  [3491] = 2084,
  [3492] = 3492,
  [3493] = 2096,
  [3494] = 3459,
  [3495] = 2101,
  [3496] = 2093,
  [3497] = 3497,
  [3498] = 2653,
  [3499] = 2654,
  [3500] = 2677,
  [3501] = 2679,
  [3502] = 2681,
  [3503] = 2682,
  [3504] = 3504,
  [3505] = 2683,
  [3506] = 2684,
  [3507] = 2685,
  [3508] = 3508,
  [3509] = 2114,
  [3510] = 2697,
  [3511] = 3511,
  [3512] = 3512,
  [3513] = 3513,
  [3514] = 3514,
  [3515] = 3515,
  [3516] = 3516,
  [3517] = 3517,
  [3518] = 3518,
  [3519] = 3519,
  [3520] = 3520,
  [3521] = 2703,
  [3522] = 3522,
  [3523] = 2693,
  [3524] = 1003,
  [3525] = 3525,
  [3526] = 3526,
  [3527] = 3527,
  [3528] = 2134,
  [3529] = 3529,
  [3530] = 1002,
  [3531] = 3531,
  [3532] = 3532,
  [3533] = 2635,
  [3534] = 3534,
  [3535] = 3535,
  [3536] = 3536,
  [3537] = 2154,
  [3538] = 2090,
  [3539] = 3508,
  [3540] = 2092,
  [3541] = 2091,
  [3542] = 2094,
  [3543] = 2096,
  [3544] = 3544,
  [3545] = 985,
  [3546] = 3546,
  [3547] = 988,
  [3548] = 2084,
  [3549] = 3549,
  [3550] = 1001,
  [3551] = 3551,
  [3552] = 2694,
  [3553] = 2133,
  [3554] = 2698,
  [3555] = 3555,
  [3556] = 3556,
  [3557] = 2670,
  [3558] = 986,
  [3559] = 2651,
  [3560] = 3560,
  [3561] = 3561,
  [3562] = 2096,
  [3563] = 1987,
  [3564] = 2642,
  [3565] = 2704,
  [3566] = 2638,
  [3567] = 2640,
  [3568] = 2641,
  [3569] = 2085,
  [3570] = 3504,
  [3571] = 1864,
  [3572] = 992,
  [3573] = 2643,
  [3574] = 2644,
  [3575] = 3504,
  [3576] = 3576,
  [3577] = 3577,
  [3578] = 1000,
  [3579] = 3504,
  [3580] = 3580,
  [3581] = 2097,
  [3582] = 1873,
  [3583] = 2362,
  [3584] = 1875,
  [3585] = 2646,
  [3586] = 3586,
  [3587] = 3587,
  [3588] = 3504,
  [3589] = 3589,
  [3590] = 2647,
  [3591] = 3591,
  [3592] = 2648,
  [3593] = 3593,
  [3594] = 2265,
  [3595] = 997,
  [3596] = 1005,
  [3597] = 2652,
  [3598] = 1006,
  [3599] = 2649,
  [3600] = 987,
  [3601] = 1007,
  [3602] = 3602,
  [3603] = 2099,
  [3604] = 2084,
  [3605] = 2101,
  [3606] = 2103,
  [3607] = 2104,
  [3608] = 2098,
  [3609] = 2082,
  [3610] = 2079,
  [3611] = 2087,
  [3612] = 2085,
  [3613] = 3613,
  [3614] = 2248,
  [3615] = 2249,
  [3616] = 2251,
  [3617] = 2444,
  [3618] = 2319,
  [3619] = 2322,
  [3620] = 2358,
  [3621] = 2291,
  [3622] = 2442,
  [3623] = 3587,
  [3624] = 1004,
  [3625] = 3591,
  [3626] = 2090,
  [3627] = 2099,
  [3628] = 2101,
  [3629] = 2103,
  [3630] = 2104,
  [3631] = 2098,
  [3632] = 2092,
  [3633] = 2093,
  [3634] = 2094,
  [3635] = 2082,
  [3636] = 2079,
  [3637] = 2087,
  [3638] = 3593,
  [3639] = 2096,
  [3640] = 2091,
  [3641] = 3641,
  [3642] = 1000,
  [3643] = 2097,
  [3644] = 3546,
  [3645] = 2266,
  [3646] = 2265,
  [3647] = 2275,
  [3648] = 3648,
  [3649] = 2408,
  [3650] = 3526,
  [3651] = 3532,
  [3652] = 3534,
  [3653] = 3551,
  [3654] = 2335,
  [3655] = 3555,
  [3656] = 2295,
  [3657] = 2379,
  [3658] = 2417,
  [3659] = 2257,
  [3660] = 2084,
  [3661] = 2422,
  [3662] = 2427,
  [3663] = 2208,
  [3664] = 2330,
  [3665] = 2269,
  [3666] = 2085,
  [3667] = 2425,
  [3668] = 2435,
  [3669] = 2362,
  [3670] = 3670,
  [3671] = 3671,
  [3672] = 2288,
  [3673] = 2353,
  [3674] = 2426,
  [3675] = 3675,
  [3676] = 2096,
  [3677] = 998,
  [3678] = 2336,
  [3679] = 2238,
  [3680] = 2438,
  [3681] = 2214,
  [3682] = 2215,
  [3683] = 1000,
  [3684] = 2312,
  [3685] = 1005,
  [3686] = 1006,
  [3687] = 3687,
  [3688] = 990,
  [3689] = 989,
  [3690] = 3690,
  [3691] = 2289,
  [3692] = 3527,
  [3693] = 3529,
  [3694] = 3694,
  [3695] = 1007,
  [3696] = 2290,
  [3697] = 1005,
  [3698] = 1006,
  [3699] = 3699,
  [3700] = 3225,
  [3701] = 2376,
  [3702] = 3702,
  [3703] = 2377,
  [3704] = 2378,
  [3705] = 3705,
  [3706] = 2380,
  [3707] = 3707,
  [3708] = 2381,
  [3709] = 2383,
  [3710] = 2384,
  [3711] = 3556,
  [3712] = 3525,
  [3713] = 3549,
  [3714] = 3560,
  [3715] = 3561,
  [3716] = 2385,
  [3717] = 2387,
  [3718] = 2428,
  [3719] = 2441,
  [3720] = 3720,
  [3721] = 3721,
  [3722] = 996,
  [3723] = 3641,
  [3724] = 3497,
  [3725] = 2084,
  [3726] = 3511,
  [3727] = 3512,
  [3728] = 3513,
  [3729] = 3514,
  [3730] = 2360,
  [3731] = 3731,
  [3732] = 3731,
  [3733] = 3733,
  [3734] = 3516,
  [3735] = 3735,
  [3736] = 2085,
  [3737] = 3737,
  [3738] = 3738,
  [3739] = 1008,
  [3740] = 1009,
  [3741] = 3741,
  [3742] = 3742,
  [3743] = 1010,
  [3744] = 967,
  [3745] = 991,
  [3746] = 1007,
  [3747] = 3518,
  [3748] = 3519,
  [3749] = 3520,
  [3750] = 3522,
  [3751] = 999,
  [3752] = 3752,
  [3753] = 3753,
  [3754] = 3586,
  [3755] = 2406,
  [3756] = 2218,
  [3757] = 3544,
  [3758] = 2246,
  [3759] = 2247,
  [3760] = 2104,
  [3761] = 3761,
  [3762] = 2689,
  [3763] = 2690,
  [3764] = 3411,
  [3765] = 3765,
  [3766] = 2655,
  [3767] = 2702,
  [3768] = 3765,
  [3769] = 3769,
  [3770] = 3770,
  [3771] = 2090,
  [3772] = 2099,
  [3773] = 2101,
  [3774] = 2103,
  [3775] = 3775,
  [3776] = 3776,
  [3777] = 2092,
  [3778] = 2093,
  [3779] = 2094,
  [3780] = 2082,
  [3781] = 2079,
  [3782] = 2087,
  [3783] = 2091,
  [3784] = 2097,
  [3785] = 3785,
  [3786] = 2084,
  [3787] = 2085,
  [3788] = 2096,
  [3789] = 2691,
  [3790] = 3790,
  [3791] = 3403,
  [3792] = 3765,
  [3793] = 3769,
  [3794] = 2692,
  [3795] = 2688,
  [3796] = 2098,
  [3797] = 3797,
  [3798] = 3798,
  [3799] = 2655,
  [3800] = 3800,
  [3801] = 3801,
  [3802] = 2744,
  [3803] = 3803,
  [3804] = 3804,
  [3805] = 3805,
  [3806] = 3806,
  [3807] = 3805,
  [3808] = 3808,
  [3809] = 3805,
  [3810] = 3613,
  [3811] = 2691,
  [3812] = 3804,
  [3813] = 3813,
  [3814] = 3814,
  [3815] = 3815,
  [3816] = 2096,
  [3817] = 2688,
  [3818] = 3818,
  [3819] = 2084,
  [3820] = 2085,
  [3821] = 1003,
  [3822] = 3822,
  [3823] = 2689,
  [3824] = 3824,
  [3825] = 3818,
  [3826] = 3813,
  [3827] = 3827,
  [3828] = 2084,
  [3829] = 3803,
  [3830] = 2690,
  [3831] = 2692,
  [3832] = 3832,
  [3833] = 3833,
  [3834] = 3834,
  [3835] = 3492,
  [3836] = 3806,
  [3837] = 2702,
  [3838] = 3838,
  [3839] = 3839,
  [3840] = 3840,
  [3841] = 3804,
  [3842] = 3411,
  [3843] = 2702,
  [3844] = 3818,
  [3845] = 987,
  [3846] = 3806,
  [3847] = 3459,
  [3848] = 2655,
  [3849] = 3849,
  [3850] = 3803,
  [3851] = 2096,
  [3852] = 3468,
  [3853] = 3853,
  [3854] = 3854,
  [3855] = 2085,
  [3856] = 3856,
  [3857] = 3512,
  [3858] = 1000,
  [3859] = 3859,
  [3860] = 3860,
  [3861] = 3529,
  [3862] = 3862,
  [3863] = 3863,
  [3864] = 3864,
  [3865] = 3860,
  [3866] = 3866,
  [3867] = 3862,
  [3868] = 3546,
  [3869] = 3513,
  [3870] = 3870,
  [3871] = 3871,
  [3872] = 2744,
  [3873] = 3862,
  [3874] = 3864,
  [3875] = 2655,
  [3876] = 3556,
  [3877] = 3877,
  [3878] = 3864,
  [3879] = 990,
  [3880] = 1004,
  [3881] = 3860,
  [3882] = 3560,
  [3883] = 3862,
  [3884] = 993,
  [3885] = 995,
  [3886] = 3886,
  [3887] = 3527,
  [3888] = 3520,
  [3889] = 1005,
  [3890] = 3525,
  [3891] = 3514,
  [3892] = 3511,
  [3893] = 3893,
  [3894] = 1006,
  [3895] = 3895,
  [3896] = 1007,
  [3897] = 3586,
  [3898] = 3587,
  [3899] = 3589,
  [3900] = 3860,
  [3901] = 3534,
  [3902] = 3591,
  [3903] = 3903,
  [3904] = 3593,
  [3905] = 3549,
  [3906] = 3906,
  [3907] = 3907,
  [3908] = 3908,
  [3909] = 3909,
  [3910] = 3864,
  [3911] = 3561,
  [3912] = 3860,
  [3913] = 2265,
  [3914] = 3914,
  [3915] = 3862,
  [3916] = 3916,
  [3917] = 3917,
  [3918] = 3519,
  [3919] = 3919,
  [3920] = 3864,
  [3921] = 3526,
  [3922] = 3508,
  [3923] = 3532,
  [3924] = 3544,
  [3925] = 3516,
  [3926] = 3551,
  [3927] = 3555,
  [3928] = 3518,
  [3929] = 2744,
  [3930] = 2362,
  [3931] = 2702,
  [3932] = 3497,
  [3933] = 3522,
  [3934] = 3648,
  [3935] = 2691,
  [3936] = 3936,
  [3937] = 2689,
  [3938] = 3938,
  [3939] = 2655,
  [3940] = 3940,
  [3941] = 3941,
  [3942] = 2690,
  [3943] = 3943,
  [3944] = 1009,
  [3945] = 3945,
  [3946] = 3946,
  [3947] = 3947,
  [3948] = 3948,
  [3949] = 2688,
  [3950] = 3950,
  [3951] = 3951,
  [3952] = 3952,
  [3953] = 2744,
  [3954] = 996,
  [3955] = 3955,
  [3956] = 1008,
  [3957] = 967,
  [3958] = 3958,
  [3959] = 2702,
  [3960] = 3694,
  [3961] = 2692,
  [3962] = 3962,
  [3963] = 3613,
  [3964] = 3964,
  [3965] = 3965,
  [3966] = 3966,
  [3967] = 1010,
  [3968] = 3968,
  [3969] = 3969,
  [3970] = 3970,
  [3971] = 3971,
  [3972] = 2744,
  [3973] = 3973,
  [3974] = 3974,
  [3975] = 3975,
  [3976] = 3976,
  [3977] = 3977,
  [3978] = 3978,
  [3979] = 3979,
  [3980] = 3980,
  [3981] = 3981,
  [3982] = 3982,
  [3983] = 3983,
  [3984] = 3863,
  [3985] = 3985,
  [3986] = 3977,
  [3987] = 3978,
  [3988] = 3988,
  [3989] = 3989,
  [3990] = 3990,
  [3991] = 3991,
  [3992] = 3992,
  [3993] = 3993,
  [3994] = 3994,
  [3995] = 3982,
  [3996] = 3996,
  [3997] = 3997,
  [3998] = 3993,
  [3999] = 3999,
  [4000] = 3980,
  [4001] = 4001,
  [4002] = 4002,
  [4003] = 995,
  [4004] = 3994,
  [4005] = 4005,
  [4006] = 4006,
  [4007] = 4007,
  [4008] = 4008,
  [4009] = 4009,
  [4010] = 4010,
  [4011] = 3971,
  [4012] = 3983,
  [4013] = 4013,
  [4014] = 4014,
  [4015] = 3977,
  [4016] = 3978,
  [4017] = 3993,
  [4018] = 3994,
  [4019] = 4001,
  [4020] = 4010,
  [4021] = 3989,
  [4022] = 3989,
  [4023] = 3999,
  [4024] = 4002,
  [4025] = 4009,
  [4026] = 3999,
  [4027] = 4027,
  [4028] = 4028,
  [4029] = 3970,
  [4030] = 4002,
  [4031] = 4031,
  [4032] = 4001,
  [4033] = 3980,
  [4034] = 4014,
  [4035] = 4035,
  [4036] = 4036,
  [4037] = 4009,
  [4038] = 3981,
  [4039] = 3785,
  [4040] = 4027,
  [4041] = 4041,
  [4042] = 4042,
  [4043] = 4028,
  [4044] = 3970,
  [4045] = 4045,
  [4046] = 4046,
  [4047] = 4047,
  [4048] = 4048,
  [4049] = 4049,
  [4050] = 4010,
  [4051] = 4051,
  [4052] = 4052,
  [4053] = 4031,
  [4054] = 3980,
  [4055] = 3989,
  [4056] = 4056,
  [4057] = 3990,
  [4058] = 3775,
  [4059] = 4014,
  [4060] = 4035,
  [4061] = 4002,
  [4062] = 4036,
  [4063] = 4006,
  [4064] = 4036,
  [4065] = 2646,
  [4066] = 4001,
  [4067] = 3776,
  [4068] = 2647,
  [4069] = 2648,
  [4070] = 4070,
  [4071] = 4071,
  [4072] = 4072,
  [4073] = 4027,
  [4074] = 4013,
  [4075] = 4006,
  [4076] = 3977,
  [4077] = 3978,
  [4078] = 4078,
  [4079] = 3982,
  [4080] = 2702,
  [4081] = 4081,
  [4082] = 3996,
  [4083] = 3997,
  [4084] = 4008,
  [4085] = 3981,
  [4086] = 4035,
  [4087] = 4087,
  [4088] = 4028,
  [4089] = 4008,
  [4090] = 4090,
  [4091] = 4091,
  [4092] = 4092,
  [4093] = 3971,
  [4094] = 3970,
  [4095] = 3981,
  [4096] = 4096,
  [4097] = 3983,
  [4098] = 4098,
  [4099] = 4013,
  [4100] = 3982,
  [4101] = 3996,
  [4102] = 3996,
  [4103] = 3997,
  [4104] = 4009,
  [4105] = 4027,
  [4106] = 4028,
  [4107] = 3980,
  [4108] = 3994,
  [4109] = 4109,
  [4110] = 4110,
  [4111] = 4111,
  [4112] = 4031,
  [4113] = 4113,
  [4114] = 4031,
  [4115] = 2684,
  [4116] = 4014,
  [4117] = 4035,
  [4118] = 4036,
  [4119] = 3971,
  [4120] = 3983,
  [4121] = 4013,
  [4122] = 3977,
  [4123] = 4123,
  [4124] = 2655,
  [4125] = 3978,
  [4126] = 3997,
  [4127] = 4127,
  [4128] = 3761,
  [4129] = 3993,
  [4130] = 3999,
  [4131] = 4131,
  [4132] = 4132,
  [4133] = 4133,
  [4134] = 4134,
  [4135] = 3952,
  [4136] = 3853,
  [4137] = 3814,
  [4138] = 3822,
  [4139] = 4139,
  [4140] = 4140,
  [4141] = 4141,
  [4142] = 2744,
  [4143] = 3815,
  [4144] = 4144,
  [4145] = 4145,
  [4146] = 4146,
  [4147] = 4147,
  [4148] = 4148,
  [4149] = 3832,
  [4150] = 4150,
  [4151] = 4151,
  [4152] = 4152,
  [4153] = 4153,
  [4154] = 4154,
  [4155] = 4155,
  [4156] = 4156,
  [4157] = 4157,
  [4158] = 4158,
  [4159] = 4159,
  [4160] = 4160,
  [4161] = 4161,
  [4162] = 4162,
  [4163] = 998,
  [4164] = 4164,
  [4165] = 3838,
  [4166] = 3950,
  [4167] = 4167,
  [4168] = 4168,
  [4169] = 4169,
  [4170] = 4170,
  [4171] = 3941,
  [4172] = 3859,
  [4173] = 4173,
  [4174] = 3906,
  [4175] = 3871,
  [4176] = 3856,
  [4177] = 4177,
  [4178] = 4173,
  [4179] = 4177,
  [4180] = 4180,
  [4181] = 1007,
  [4182] = 3893,
  [4183] = 4183,
  [4184] = 3866,
  [4185] = 3870,
  [4186] = 3589,
  [4187] = 3903,
  [4188] = 1006,
  [4189] = 4189,
  [4190] = 1000,
  [4191] = 4191,
  [4192] = 4192,
  [4193] = 4193,
  [4194] = 4194,
  [4195] = 4195,
  [4196] = 1005,
  [4197] = 4197,
  [4198] = 4198,
  [4199] = 4177,
  [4200] = 3895,
  [4201] = 4201,
  [4202] = 4202,
  [4203] = 4203,
  [4204] = 4204,
  [4205] = 4205,
  [4206] = 4206,
  [4207] = 4207,
  [4208] = 4208,
  [4209] = 4209,
  [4210] = 4210,
  [4211] = 4211,
  [4212] = 4212,
  [4213] = 4213,
  [4214] = 4209,
  [4215] = 4215,
  [4216] = 4216,
  [4217] = 4159,
  [4218] = 4218,
  [4219] = 4219,
  [4220] = 4220,
  [4221] = 1939,
  [4222] = 4222,
  [4223] = 4223,
  [4224] = 4224,
  [4225] = 4225,
  [4226] = 4226,
  [4227] = 4227,
  [4228] = 4228,
  [4229] = 4162,
  [4230] = 4230,
  [4231] = 4231,
  [4232] = 4209,
  [4233] = 4209,
  [4234] = 4234,
  [4235] = 4209,
  [4236] = 4236,
  [4237] = 4237,
  [4238] = 4238,
  [4239] = 4209,
  [4240] = 4240,
  [4241] = 4241,
  [4242] = 4242,
  [4243] = 4243,
  [4244] = 4209,
  [4245] = 4245,
  [4246] = 4246,
  [4247] = 4247,
  [4248] = 4248,
  [4249] = 4249,
  [4250] = 4250,
  [4251] = 4251,
  [4252] = 4252,
  [4253] = 4253,
  [4254] = 4254,
  [4255] = 4255,
  [4256] = 4256,
  [4257] = 4257,
  [4258] = 4258,
  [4259] = 4259,
  [4260] = 4260,
  [4261] = 4261,
  [4262] = 4262,
  [4263] = 4263,
  [4264] = 4264,
  [4265] = 4265,
  [4266] = 4266,
  [4267] = 4267,
  [4268] = 4268,
  [4269] = 4269,
  [4270] = 4209,
  [4271] = 4271,
  [4272] = 4272,
  [4273] = 3966,
  [4274] = 3946,
  [4275] = 4275,
  [4276] = 4276,
  [4277] = 4277,
  [4278] = 4278,
  [4279] = 4279,
  [4280] = 4280,
  [4281] = 4281,
  [4282] = 3969,
  [4283] = 4209,
  [4284] = 2855,
  [4285] = 4285,
  [4286] = 4209,
  [4287] = 4287,
  [4288] = 4288,
  [4289] = 4289,
  [4290] = 4290,
  [4291] = 4291,
  [4292] = 4292,
  [4293] = 4293,
  [4294] = 4294,
  [4295] = 4295,
  [4296] = 4296,
  [4297] = 4297,
  [4298] = 4298,
  [4299] = 4299,
  [4300] = 4300,
  [4301] = 4301,
  [4302] = 4302,
  [4303] = 4209,
  [4304] = 4304,
  [4305] = 4305,
  [4306] = 4306,
  [4307] = 4307,
  [4308] = 1931,
  [4309] = 4309,
  [4310] = 3945,
  [4311] = 4311,
  [4312] = 4312,
  [4313] = 4313,
  [4314] = 4314,
  [4315] = 4209,
  [4316] = 4316,
  [4317] = 4317,
  [4318] = 4318,
  [4319] = 4319,
  [4320] = 1929,
  [4321] = 4321,
  [4322] = 4322,
  [4323] = 4323,
  [4324] = 4324,
  [4325] = 4325,
  [4326] = 4326,
  [4327] = 4327,
  [4328] = 4328,
  [4329] = 4329,
  [4330] = 4209,
  [4331] = 4209,
  [4332] = 4332,
  [4333] = 4333,
  [4334] = 4334,
  [4335] = 4335,
  [4336] = 4336,
  [4337] = 4160,
  [4338] = 4338,
  [4339] = 4339,
  [4340] = 4340,
  [4341] = 4341,
  [4342] = 4342,
  [4343] = 4343,
  [4344] = 4344,
  [4345] = 4345,
  [4346] = 4346,
  [4347] = 4347,
  [4348] = 4348,
  [4349] = 4349,
  [4350] = 4350,
  [4351] = 4351,
  [4352] = 4352,
  [4353] = 4353,
  [4354] = 4354,
  [4355] = 3948,
  [4356] = 4356,
  [4357] = 4357,
  [4358] = 4358,
  [4359] = 4359,
  [4360] = 1987,
  [4361] = 4361,
  [4362] = 995,
  [4363] = 4363,
  [4364] = 4364,
  [4365] = 4365,
  [4366] = 4366,
  [4367] = 997,
  [4368] = 992,
  [4369] = 4228,
  [4370] = 1002,
  [4371] = 4371,
  [4372] = 4206,
  [4373] = 4373,
  [4374] = 4333,
  [4375] = 4042,
  [4376] = 4376,
  [4377] = 4377,
  [4378] = 4378,
  [4379] = 4379,
  [4380] = 4380,
  [4381] = 4381,
  [4382] = 4382,
  [4383] = 4383,
  [4384] = 4384,
  [4385] = 4385,
  [4386] = 4386,
  [4387] = 4387,
  [4388] = 4388,
  [4389] = 4389,
  [4390] = 4390,
  [4391] = 4391,
  [4392] = 4382,
  [4393] = 4393,
  [4394] = 997,
  [4395] = 4240,
  [4396] = 4384,
  [4397] = 4397,
  [4398] = 1002,
  [4399] = 4399,
  [4400] = 4400,
  [4401] = 4401,
  [4402] = 992,
  [4403] = 4357,
  [4404] = 4404,
  [4405] = 4405,
  [4406] = 3561,
  [4407] = 4071,
  [4408] = 4389,
  [4409] = 4253,
  [4410] = 4410,
  [4411] = 4255,
  [4412] = 4412,
  [4413] = 997,
  [4414] = 4087,
  [4415] = 4415,
  [4416] = 4416,
  [4417] = 3979,
  [4418] = 4366,
  [4419] = 4098,
  [4420] = 3985,
  [4421] = 4005,
  [4422] = 4357,
  [4423] = 4405,
  [4424] = 4113,
  [4425] = 4127,
  [4426] = 3973,
  [4427] = 3976,
  [4428] = 4391,
  [4429] = 4046,
  [4430] = 4047,
  [4431] = 4048,
  [4432] = 4052,
  [4433] = 4090,
  [4434] = 4091,
  [4435] = 4092,
  [4436] = 4096,
  [4437] = 4123,
  [4438] = 4438,
  [4439] = 4439,
  [4440] = 4382,
  [4441] = 4441,
  [4442] = 4442,
  [4443] = 1002,
  [4444] = 4393,
  [4445] = 4445,
  [4446] = 4389,
  [4447] = 4384,
  [4448] = 4386,
  [4449] = 4386,
  [4450] = 4279,
  [4451] = 4357,
  [4452] = 4405,
  [4453] = 4391,
  [4454] = 4358,
  [4455] = 4401,
  [4456] = 4456,
  [4457] = 4366,
  [4458] = 4391,
  [4459] = 4459,
  [4460] = 4460,
  [4461] = 4461,
  [4462] = 4357,
  [4463] = 4405,
  [4464] = 4464,
  [4465] = 4210,
  [4466] = 4269,
  [4467] = 4271,
  [4468] = 992,
  [4469] = 4469,
  [4470] = 4365,
  [4471] = 4438,
  [4472] = 4472,
  [4473] = 4473,
  [4474] = 4474,
  [4475] = 4475,
  [4476] = 4476,
  [4477] = 4366,
  [4478] = 4357,
  [4479] = 4405,
  [4480] = 4391,
  [4481] = 3991,
  [4482] = 3534,
  [4483] = 4366,
  [4484] = 4401,
  [4485] = 4405,
  [4486] = 4391,
  [4487] = 4358,
  [4488] = 4358,
  [4489] = 4389,
  [4490] = 4365,
  [4491] = 4366,
  [4492] = 4049,
  [4493] = 4379,
  [4494] = 4494,
  [4495] = 4495,
  [4496] = 4496,
  [4497] = 4139,
  [4498] = 4496,
  [4499] = 4496,
  [4500] = 4460,
  [4501] = 4501,
  [4502] = 4502,
  [4503] = 4503,
  [4504] = 4504,
  [4505] = 4460,
  [4506] = 4461,
  [4507] = 4496,
  [4508] = 4508,
  [4509] = 4144,
  [4510] = 4510,
  [4511] = 4496,
  [4512] = 4512,
  [4513] = 4376,
  [4514] = 4496,
  [4515] = 4377,
  [4516] = 4516,
  [4517] = 4378,
  [4518] = 4518,
  [4519] = 4496,
  [4520] = 3225,
  [4521] = 4496,
  [4522] = 4522,
  [4523] = 4378,
  [4524] = 4524,
  [4525] = 4525,
  [4526] = 4496,
  [4527] = 4527,
  [4528] = 4461,
  [4529] = 3468,
  [4530] = 4379,
  [4531] = 4496,
  [4532] = 4496,
  [4533] = 3492,
  [4534] = 4148,
  [4535] = 4381,
  [4536] = 4152,
  [4537] = 4496,
  [4538] = 4133,
  [4539] = 4539,
  [4540] = 4540,
  [4541] = 4541,
  [4542] = 4542,
  [4543] = 4158,
  [4544] = 4544,
  [4545] = 4134,
  [4546] = 4376,
  [4547] = 4496,
  [4548] = 4164,
  [4549] = 4549,
  [4550] = 4377,
  [4551] = 4551,
  [4552] = 4504,
  [4553] = 4496,
  [4554] = 4554,
  [4555] = 4381,
  [4556] = 4556,
  [4557] = 4201,
  [4558] = 4558,
  [4559] = 4556,
  [4560] = 4560,
  [4561] = 4561,
  [4562] = 4562,
  [4563] = 4563,
  [4564] = 4564,
  [4565] = 4565,
  [4566] = 4566,
  [4567] = 4566,
  [4568] = 4556,
  [4569] = 4560,
  [4570] = 4565,
  [4571] = 4571,
  [4572] = 4572,
  [4573] = 4573,
  [4574] = 4574,
  [4575] = 4562,
  [4576] = 4558,
  [4577] = 4558,
  [4578] = 4562,
  [4579] = 4561,
  [4580] = 4563,
  [4581] = 4566,
  [4582] = 4556,
  [4583] = 4560,
  [4584] = 4562,
  [4585] = 4565,
  [4586] = 4571,
  [4587] = 4572,
  [4588] = 4573,
  [4589] = 4574,
  [4590] = 4571,
  [4591] = 4562,
  [4592] = 4556,
  [4593] = 4560,
  [4594] = 4571,
  [4595] = 4565,
  [4596] = 4562,
  [4597] = 4571,
  [4598] = 4572,
  [4599] = 4573,
  [4600] = 4572,
  [4601] = 4573,
  [4602] = 4574,
  [4603] = 4558,
  [4604] = 4562,
  [4605] = 4562,
  [4606] = 4562,
  [4607] = 4572,
  [4608] = 4573,
  [4609] = 4609,
  [4610] = 4561,
  [4611] = 4558,
  [4612] = 4564,
  [4613] = 4563,
  [4614] = 4574,
  [4615] = 4562,
  [4616] = 4376,
  [4617] = 4561,
  [4618] = 4618,
  [4619] = 4562,
  [4620] = 4563,
  [4621] = 4621,
  [4622] = 4622,
  [4623] = 4558,
  [4624] = 4566,
  [4625] = 4625,
  [4626] = 4626,
  [4627] = 4556,
  [4628] = 3468,
  [4629] = 4460,
  [4630] = 4630,
  [4631] = 4561,
  [4632] = 4563,
  [4633] = 4633,
  [4634] = 4377,
  [4635] = 4378,
  [4636] = 4461,
  [4637] = 4560,
  [4638] = 4379,
  [4639] = 4562,
  [4640] = 4565,
  [4641] = 4571,
  [4642] = 4558,
  [4643] = 4572,
  [4644] = 4573,
  [4645] = 4574,
  [4646] = 4197,
  [4647] = 4333,
  [4648] = 4561,
  [4649] = 4649,
  [4650] = 4650,
  [4651] = 4563,
  [4652] = 4183,
  [4653] = 4653,
  [4654] = 4558,
  [4655] = 4381,
  [4656] = 4561,
  [4657] = 4202,
  [4658] = 4563,
  [4659] = 4206,
  [4660] = 4210,
  [4661] = 4558,
  [4662] = 4561,
  [4663] = 4228,
  [4664] = 4563,
  [4665] = 4665,
  [4666] = 3492,
  [4667] = 4566,
  [4668] = 4240,
  [4669] = 4193,
  [4670] = 4195,
  [4671] = 4566,
  [4672] = 4253,
  [4673] = 4255,
  [4674] = 4618,
  [4675] = 4566,
  [4676] = 4269,
  [4677] = 4556,
  [4678] = 4560,
  [4679] = 4271,
  [4680] = 4565,
  [4681] = 4279,
  [4682] = 4566,
  [4683] = 4571,
  [4684] = 4572,
  [4685] = 4573,
  [4686] = 4574,
  [4687] = 4563,
  [4688] = 3492,
  [4689] = 4689,
  [4690] = 4566,
  [4691] = 4561,
  [4692] = 4556,
  [4693] = 4560,
  [4694] = 4565,
  [4695] = 4561,
  [4696] = 4558,
  [4697] = 4571,
  [4698] = 4572,
  [4699] = 4573,
  [4700] = 4700,
  [4701] = 4574,
  [4702] = 4574,
  [4703] = 4561,
  [4704] = 4563,
  [4705] = 4556,
  [4706] = 4560,
  [4707] = 4566,
  [4708] = 4556,
  [4709] = 4560,
  [4710] = 4565,
  [4711] = 4571,
  [4712] = 4572,
  [4713] = 4573,
  [4714] = 4574,
  [4715] = 4565,
  [4716] = 3468,
  [4717] = 4565,
  [4718] = 4571,
  [4719] = 4719,
  [4720] = 4563,
  [4721] = 4558,
  [4722] = 4572,
  [4723] = 4573,
  [4724] = 4561,
  [4725] = 4563,
  [4726] = 4558,
  [4727] = 4574,
  [4728] = 4566,
  [4729] = 4574,
  [4730] = 4560,
  [4731] = 4565,
  [4732] = 4571,
  [4733] = 4572,
  [4734] = 4573,
  [4735] = 4574,
  [4736] = 4561,
  [4737] = 4556,
  [4738] = 4563,
  [4739] = 4560,
  [4740] = 4556,
  [4741] = 4566,
  [4742] = 4565,
  [4743] = 4743,
  [4744] = 4566,
  [4745] = 4560,
  [4746] = 4558,
  [4747] = 4571,
  [4748] = 4562,
  [4749] = 4556,
  [4750] = 4561,
  [4751] = 4560,
  [4752] = 4565,
  [4753] = 3589,
  [4754] = 4572,
  [4755] = 4189,
  [4756] = 4571,
  [4757] = 4572,
  [4758] = 4573,
  [4759] = 4573,
  [4760] = 4563,
  [4761] = 4574,
  [4762] = 4574,
  [4763] = 4566,
  [4764] = 4192,
  [4765] = 4556,
  [4766] = 4198,
  [4767] = 4560,
  [4768] = 4565,
  [4769] = 4571,
  [4770] = 4572,
  [4771] = 4573,
  [4772] = 4558,
  [4773] = 4773,
  [4774] = 1010,
  [4775] = 4775,
  [4776] = 4776,
  [4777] = 4626,
  [4778] = 4776,
  [4779] = 967,
  [4780] = 1001,
  [4781] = 4781,
  [4782] = 985,
  [4783] = 986,
  [4784] = 4784,
  [4785] = 988,
  [4786] = 1011,
  [4787] = 4787,
  [4788] = 4042,
  [4789] = 1009,
  [4790] = 4784,
  [4791] = 4791,
  [4792] = 4787,
  [4793] = 4793,
  [4794] = 996,
  [4795] = 4649,
  [4796] = 1008,
  [4797] = 4650,
  [4798] = 4798,
  [4799] = 4799,
  [4800] = 4784,
  [4801] = 954,
  [4802] = 989,
  [4803] = 4803,
  [4804] = 4625,
  [4805] = 4630,
  [4806] = 4806,
  [4807] = 4807,
  [4808] = 4808,
  [4809] = 4809,
  [4810] = 4810,
  [4811] = 4809,
  [4812] = 985,
  [4813] = 1001,
  [4814] = 986,
  [4815] = 4815,
  [4816] = 4810,
  [4817] = 4810,
  [4818] = 4818,
  [4819] = 988,
  [4820] = 4820,
  [4821] = 4815,
  [4822] = 4822,
  [4823] = 4798,
  [4824] = 4820,
  [4825] = 4825,
  [4826] = 4810,
  [4827] = 4827,
  [4828] = 4828,
  [4829] = 4827,
  [4830] = 4810,
  [4831] = 4808,
  [4832] = 4822,
  [4833] = 4810,
  [4834] = 4810,
  [4835] = 4835,
  [4836] = 4807,
  [4837] = 4837,
  [4838] = 4818,
  [4839] = 4839,
  [4840] = 4840,
  [4841] = 4809,
  [4842] = 4810,
  [4843] = 4810,
  [4844] = 4844,
  [4845] = 4809,
  [4846] = 4844,
  [4847] = 4808,
  [4848] = 4810,
  [4849] = 4849,
  [4850] = 4810,
  [4851] = 4810,
  [4852] = 4852,
  [4853] = 4835,
  [4854] = 4849,
  [4855] = 4852,
  [4856] = 4837,
  [4857] = 4808,
  [4858] = 4810,
  [4859] = 4859,
  [4860] = 4860,
  [4861] = 4860,
  [4862] = 4539,
  [4863] = 999,
  [4864] = 4860,
  [4865] = 3236,
  [4866] = 4501,
  [4867] = 4524,
  [4868] = 4544,
  [4869] = 4860,
  [4870] = 4510,
  [4871] = 4512,
  [4872] = 4625,
  [4873] = 4860,
  [4874] = 4626,
  [4875] = 4860,
  [4876] = 4860,
  [4877] = 4549,
  [4878] = 4541,
  [4879] = 4542,
  [4880] = 4860,
  [4881] = 4860,
  [4882] = 4860,
  [4883] = 4860,
  [4884] = 4860,
  [4885] = 3241,
  [4886] = 4860,
  [4887] = 991,
  [4888] = 4649,
  [4889] = 4650,
  [4890] = 4859,
  [4891] = 4860,
  [4892] = 4860,
  [4893] = 4630,
  [4894] = 4860,
  [4895] = 4860,
  [4896] = 4860,
  [4897] = 4860,
  [4898] = 4516,
  [4899] = 4860,
  [4900] = 4859,
  [4901] = 4494,
  [4902] = 4495,
  [4903] = 4502,
  [4904] = 4503,
  [4905] = 4518,
  [4906] = 4522,
  [4907] = 4551,
  [4908] = 4860,
  [4909] = 4909,
  [4910] = 4910,
  [4911] = 4911,
  [4912] = 4909,
  [4913] = 4913,
  [4914] = 4914,
  [4915] = 4913,
  [4916] = 4916,
  [4917] = 4917,
  [4918] = 4918,
  [4919] = 4918,
  [4920] = 4920,
  [4921] = 4920,
  [4922] = 4917,
  [4923] = 4923,
  [4924] = 4917,
  [4925] = 4925,
  [4926] = 4923,
  [4927] = 4927,
  [4928] = 4914,
  [4929] = 4918,
  [4930] = 3871,
  [4931] = 2036,
  [4932] = 4923,
  [4933] = 4933,
  [4934] = 4913,
  [4935] = 4913,
  [4936] = 4923,
  [4937] = 4916,
  [4938] = 4938,
  [4939] = 4925,
  [4940] = 4938,
  [4941] = 4920,
  [4942] = 4942,
  [4943] = 3535,
  [4944] = 3948,
  [4945] = 4945,
  [4946] = 4516,
  [4947] = 4947,
  [4948] = 4948,
  [4949] = 3576,
  [4950] = 4950,
  [4951] = 4950,
  [4952] = 4948,
  [4953] = 4953,
  [4954] = 3515,
  [4955] = 3577,
  [4956] = 3517,
  [4957] = 3871,
  [4958] = 4950,
  [4959] = 4542,
  [4960] = 4960,
  [4961] = 4948,
  [4962] = 4962,
  [4963] = 993,
  [4964] = 4541,
  [4965] = 3236,
  [4966] = 4947,
  [4967] = 4948,
  [4968] = 3531,
  [4969] = 4947,
  [4970] = 2076,
  [4971] = 3871,
  [4972] = 4972,
  [4973] = 3248,
  [4974] = 4974,
  [4975] = 4975,
  [4976] = 4976,
  [4977] = 4975,
  [4978] = 4978,
  [4979] = 4979,
  [4980] = 4980,
  [4981] = 4981,
  [4982] = 4974,
  [4983] = 3948,
  [4984] = 4984,
  [4985] = 4985,
  [4986] = 3237,
  [4987] = 4975,
  [4988] = 4975,
  [4989] = 4989,
  [4990] = 4974,
  [4991] = 3687,
  [4992] = 4975,
  [4993] = 4975,
  [4994] = 4994,
  [4995] = 4974,
  [4996] = 3246,
  [4997] = 989,
  [4998] = 4998,
  [4999] = 4974,
  [5000] = 3671,
  [5001] = 5001,
  [5002] = 3239,
  [5003] = 4974,
  [5004] = 5004,
  [5005] = 2134,
  [5006] = 3648,
  [5007] = 5007,
  [5008] = 5008,
  [5009] = 2154,
  [5010] = 3251,
  [5011] = 4980,
  [5012] = 3737,
  [5013] = 3241,
  [5014] = 4975,
  [5015] = 3240,
  [5016] = 5016,
  [5017] = 5017,
  [5018] = 3242,
  [5019] = 5019,
  [5020] = 4975,
  [5021] = 4974,
  [5022] = 4974,
  [5023] = 4985,
  [5024] = 3236,
  [5025] = 5025,
  [5026] = 3948,
  [5027] = 4974,
  [5028] = 4985,
  [5029] = 3694,
  [5030] = 3241,
  [5031] = 4980,
  [5032] = 5032,
  [5033] = 5033,
  [5034] = 3832,
  [5035] = 2849,
  [5036] = 5036,
  [5037] = 5037,
  [5038] = 5038,
  [5039] = 5039,
  [5040] = 5040,
  [5041] = 5041,
  [5042] = 5042,
  [5043] = 5043,
  [5044] = 5044,
  [5045] = 5045,
  [5046] = 5046,
  [5047] = 5047,
  [5048] = 2864,
  [5049] = 5049,
  [5050] = 5050,
  [5051] = 5051,
  [5052] = 5052,
  [5053] = 5053,
  [5054] = 5054,
  [5055] = 2872,
  [5056] = 5056,
  [5057] = 5057,
  [5058] = 5058,
  [5059] = 5059,
  [5060] = 5038,
  [5061] = 5061,
  [5062] = 5062,
  [5063] = 5041,
  [5064] = 5064,
  [5065] = 5038,
  [5066] = 5041,
  [5067] = 5067,
  [5068] = 5068,
  [5069] = 5038,
  [5070] = 5070,
  [5071] = 2275,
  [5072] = 5041,
  [5073] = 5073,
  [5074] = 5074,
  [5075] = 5075,
  [5076] = 5038,
  [5077] = 5041,
  [5078] = 5078,
  [5079] = 5079,
  [5080] = 5080,
  [5081] = 5081,
  [5082] = 5082,
  [5083] = 5083,
  [5084] = 5084,
  [5085] = 5085,
  [5086] = 5086,
  [5087] = 5087,
  [5088] = 5088,
  [5089] = 5038,
  [5090] = 5090,
  [5091] = 5091,
  [5092] = 5041,
  [5093] = 5093,
  [5094] = 5094,
  [5095] = 5095,
  [5096] = 5096,
  [5097] = 5038,
  [5098] = 5098,
  [5099] = 5099,
  [5100] = 5100,
  [5101] = 5038,
  [5102] = 5041,
  [5103] = 5038,
  [5104] = 2426,
  [5105] = 5041,
  [5106] = 5106,
  [5107] = 5107,
  [5108] = 5108,
  [5109] = 5041,
  [5110] = 5110,
  [5111] = 5111,
  [5112] = 5038,
  [5113] = 5113,
  [5114] = 5114,
  [5115] = 5115,
  [5116] = 5041,
  [5117] = 5117,
  [5118] = 5118,
  [5119] = 2214,
  [5120] = 2215,
  [5121] = 5038,
  [5122] = 5041,
  [5123] = 2312,
  [5124] = 5124,
  [5125] = 5125,
  [5126] = 5126,
  [5127] = 5127,
  [5128] = 5128,
  [5129] = 5129,
  [5130] = 3775,
  [5131] = 5131,
  [5132] = 2866,
  [5133] = 2852,
  [5134] = 987,
  [5135] = 1003,
  [5136] = 5041,
  [5137] = 3761,
  [5138] = 5138,
  [5139] = 5139,
  [5140] = 5140,
  [5141] = 5141,
  [5142] = 2815,
  [5143] = 2816,
  [5144] = 5144,
  [5145] = 5145,
  [5146] = 5146,
  [5147] = 5147,
  [5148] = 5148,
  [5149] = 5038,
  [5150] = 5131,
  [5151] = 5151,
  [5152] = 5152,
  [5153] = 4459,
  [5154] = 5041,
  [5155] = 2330,
  [5156] = 4412,
  [5157] = 4415,
  [5158] = 4441,
  [5159] = 3776,
  [5160] = 4361,
  [5161] = 4363,
  [5162] = 4364,
  [5163] = 3785,
  [5164] = 4397,
  [5165] = 5165,
  [5166] = 2850,
  [5167] = 2870,
  [5168] = 2778,
  [5169] = 2779,
  [5170] = 2786,
  [5171] = 5038,
  [5172] = 2795,
  [5173] = 2798,
  [5174] = 5174,
  [5175] = 5175,
  [5176] = 5176,
  [5177] = 5177,
  [5178] = 5178,
  [5179] = 5179,
  [5180] = 990,
  [5181] = 1004,
  [5182] = 3853,
  [5183] = 3838,
  [5184] = 5184,
  [5185] = 5185,
  [5186] = 3814,
  [5187] = 3822,
  [5188] = 5188,
  [5189] = 5189,
  [5190] = 5190,
  [5191] = 5191,
  [5192] = 5192,
  [5193] = 998,
  [5194] = 5185,
  [5195] = 5184,
  [5196] = 3602,
  [5197] = 5197,
  [5198] = 3815,
  [5199] = 5185,
  [5200] = 3536,
  [5201] = 3580,
  [5202] = 5184,
  [5203] = 3741,
  [5204] = 930,
  [5205] = 5205,
  [5206] = 5206,
  [5207] = 1000,
  [5208] = 1005,
  [5209] = 1006,
  [5210] = 5210,
  [5211] = 5211,
  [5212] = 3866,
  [5213] = 3870,
  [5214] = 1007,
  [5215] = 3906,
  [5216] = 5206,
  [5217] = 5217,
  [5218] = 3903,
  [5219] = 3720,
  [5220] = 3877,
  [5221] = 3895,
  [5222] = 3856,
  [5223] = 5223,
  [5224] = 3690,
  [5225] = 5225,
  [5226] = 5226,
  [5227] = 3893,
  [5228] = 5228,
  [5229] = 3699,
  [5230] = 5210,
  [5231] = 2114,
  [5232] = 3733,
  [5233] = 3752,
  [5234] = 3753,
  [5235] = 5210,
  [5236] = 5236,
  [5237] = 5237,
  [5238] = 5225,
  [5239] = 3702,
  [5240] = 5225,
  [5241] = 929,
  [5242] = 5236,
  [5243] = 5236,
  [5244] = 3859,
  [5245] = 5211,
  [5246] = 3735,
  [5247] = 3738,
  [5248] = 5248,
  [5249] = 3670,
  [5250] = 3675,
  [5251] = 5223,
  [5252] = 5211,
  [5253] = 5253,
  [5254] = 5254,
  [5255] = 3946,
  [5256] = 5256,
  [5257] = 3824,
  [5258] = 3854,
  [5259] = 5254,
  [5260] = 5254,
  [5261] = 2295,
  [5262] = 5256,
  [5263] = 2379,
  [5264] = 2417,
  [5265] = 3853,
  [5266] = 2422,
  [5267] = 2427,
  [5268] = 5256,
  [5269] = 5256,
  [5270] = 3814,
  [5271] = 2425,
  [5272] = 2435,
  [5273] = 5256,
  [5274] = 3822,
  [5275] = 3969,
  [5276] = 2218,
  [5277] = 5256,
  [5278] = 2247,
  [5279] = 2249,
  [5280] = 5254,
  [5281] = 5256,
  [5282] = 2251,
  [5283] = 2319,
  [5284] = 5256,
  [5285] = 2322,
  [5286] = 5256,
  [5287] = 2358,
  [5288] = 2406,
  [5289] = 3849,
  [5290] = 5256,
  [5291] = 2291,
  [5292] = 2442,
  [5293] = 5254,
  [5294] = 2444,
  [5295] = 5256,
  [5296] = 5254,
  [5297] = 5297,
  [5298] = 5256,
  [5299] = 5256,
  [5300] = 5256,
  [5301] = 5256,
  [5302] = 5256,
  [5303] = 5256,
  [5304] = 5256,
  [5305] = 3834,
  [5306] = 5306,
  [5307] = 3886,
  [5308] = 5308,
  [5309] = 5254,
  [5310] = 2336,
  [5311] = 2238,
  [5312] = 2438,
  [5313] = 5254,
  [5314] = 3800,
  [5315] = 2360,
  [5316] = 3797,
  [5317] = 5254,
  [5318] = 2377,
  [5319] = 2381,
  [5320] = 2428,
  [5321] = 2441,
  [5322] = 3945,
  [5323] = 3966,
  [5324] = 2384,
  [5325] = 2385,
  [5326] = 3840,
  [5327] = 2248,
  [5328] = 5328,
  [5329] = 2246,
  [5330] = 5254,
  [5331] = 3411,
  [5332] = 5254,
  [5333] = 5254,
  [5334] = 5254,
  [5335] = 5256,
  [5336] = 5254,
  [5337] = 5256,
  [5338] = 5338,
  [5339] = 5339,
  [5340] = 5340,
  [5341] = 5341,
  [5342] = 5342,
  [5343] = 3856,
  [5344] = 4045,
  [5345] = 4051,
  [5346] = 4042,
  [5347] = 5341,
  [5348] = 4046,
  [5349] = 5339,
  [5350] = 4047,
  [5351] = 4048,
  [5352] = 4049,
  [5353] = 3893,
  [5354] = 3866,
  [5355] = 3870,
  [5356] = 5356,
  [5357] = 5357,
  [5358] = 5358,
  [5359] = 4052,
  [5360] = 5360,
  [5361] = 5361,
  [5362] = 5362,
  [5363] = 3914,
  [5364] = 4071,
  [5365] = 5341,
  [5366] = 5366,
  [5367] = 5367,
  [5368] = 4113,
  [5369] = 5369,
  [5370] = 4127,
  [5371] = 2133,
  [5372] = 3973,
  [5373] = 5197,
  [5374] = 5374,
  [5375] = 3976,
  [5376] = 4090,
  [5377] = 5360,
  [5378] = 4091,
  [5379] = 4092,
  [5380] = 4087,
  [5381] = 5338,
  [5382] = 4096,
  [5383] = 3979,
  [5384] = 5338,
  [5385] = 5362,
  [5386] = 4123,
  [5387] = 5366,
  [5388] = 5374,
  [5389] = 5362,
  [5390] = 3991,
  [5391] = 4098,
  [5392] = 3985,
  [5393] = 5339,
  [5394] = 4005,
  [5395] = 5360,
  [5396] = 5338,
  [5397] = 5361,
  [5398] = 5362,
  [5399] = 4056,
  [5400] = 5400,
  [5401] = 5401,
  [5402] = 2154,
  [5403] = 5403,
  [5404] = 5404,
  [5405] = 4078,
  [5406] = 4131,
  [5407] = 5407,
  [5408] = 5408,
  [5409] = 5409,
  [5410] = 5410,
  [5411] = 5407,
  [5412] = 4164,
  [5413] = 985,
  [5414] = 5414,
  [5415] = 5415,
  [5416] = 986,
  [5417] = 5407,
  [5418] = 5197,
  [5419] = 5419,
  [5420] = 5420,
  [5421] = 5421,
  [5422] = 5422,
  [5423] = 5407,
  [5424] = 5424,
  [5425] = 5425,
  [5426] = 5400,
  [5427] = 988,
  [5428] = 5408,
  [5429] = 5400,
  [5430] = 2114,
  [5431] = 5431,
  [5432] = 5432,
  [5433] = 5415,
  [5434] = 5409,
  [5435] = 5435,
  [5436] = 5432,
  [5437] = 5437,
  [5438] = 5422,
  [5439] = 5425,
  [5440] = 5408,
  [5441] = 1001,
  [5442] = 5400,
  [5443] = 5415,
  [5444] = 5422,
  [5445] = 5419,
  [5446] = 5400,
  [5447] = 5415,
  [5448] = 5025,
  [5449] = 5424,
  [5450] = 5407,
  [5451] = 2134,
  [5452] = 5415,
  [5453] = 5432,
  [5454] = 5400,
  [5455] = 5415,
  [5456] = 2376,
  [5457] = 992,
  [5458] = 2378,
  [5459] = 2380,
  [5460] = 2383,
  [5461] = 5400,
  [5462] = 2387,
  [5463] = 5415,
  [5464] = 5431,
  [5465] = 5465,
  [5466] = 997,
  [5467] = 1002,
  [5468] = 5400,
  [5469] = 5407,
  [5470] = 5415,
  [5471] = 3965,
  [5472] = 5421,
  [5473] = 5400,
  [5474] = 5415,
  [5475] = 5419,
  [5476] = 3968,
  [5477] = 5400,
  [5478] = 5415,
  [5479] = 4144,
  [5480] = 5400,
  [5481] = 5415,
  [5482] = 5432,
  [5483] = 5400,
  [5484] = 5415,
  [5485] = 5400,
  [5486] = 5415,
  [5487] = 5410,
  [5488] = 3988,
  [5489] = 5407,
  [5490] = 5410,
  [5491] = 5431,
  [5492] = 3974,
  [5493] = 3940,
  [5494] = 4158,
  [5495] = 3964,
  [5496] = 5425,
  [5497] = 3951,
  [5498] = 4161,
  [5499] = 5499,
  [5500] = 5228,
  [5501] = 5217,
  [5502] = 5237,
  [5503] = 5226,
  [5504] = 5205,
  [5505] = 3958,
  [5506] = 5431,
  [5507] = 2133,
  [5508] = 5422,
  [5509] = 4133,
  [5510] = 5400,
  [5511] = 5414,
  [5512] = 4148,
  [5513] = 4152,
  [5514] = 5419,
  [5515] = 5425,
  [5516] = 987,
  [5517] = 2335,
  [5518] = 3943,
  [5519] = 1003,
  [5520] = 2133,
  [5521] = 5465,
  [5522] = 5407,
  [5523] = 4139,
  [5524] = 4134,
  [5525] = 3947,
  [5526] = 5408,
  [5527] = 5410,
  [5528] = 5407,
  [5529] = 5407,
  [5530] = 5248,
  [5531] = 5421,
  [5532] = 5532,
  [5533] = 5533,
  [5534] = 5407,
  [5535] = 5424,
  [5536] = 5437,
  [5537] = 5415,
  [5538] = 5421,
  [5539] = 5424,
  [5540] = 5409,
  [5541] = 5541,
  [5542] = 5542,
  [5543] = 2872,
  [5544] = 5544,
  [5545] = 5248,
  [5546] = 5033,
  [5547] = 5036,
  [5548] = 5043,
  [5549] = 5549,
  [5550] = 4132,
  [5551] = 2362,
  [5552] = 5552,
  [5553] = 4156,
  [5554] = 3941,
  [5555] = 5549,
  [5556] = 5556,
  [5557] = 2336,
  [5558] = 2238,
  [5559] = 5542,
  [5560] = 2442,
  [5561] = 2438,
  [5562] = 5552,
  [5563] = 5556,
  [5564] = 2251,
  [5565] = 2815,
  [5566] = 5552,
  [5567] = 998,
  [5568] = 5568,
  [5569] = 2816,
  [5570] = 996,
  [5571] = 5544,
  [5572] = 4153,
  [5573] = 5573,
  [5574] = 4157,
  [5575] = 5575,
  [5576] = 5556,
  [5577] = 5544,
  [5578] = 4151,
  [5579] = 5579,
  [5580] = 2335,
  [5581] = 2295,
  [5582] = 2379,
  [5583] = 2319,
  [5584] = 5091,
  [5585] = 5098,
  [5586] = 2417,
  [5587] = 1008,
  [5588] = 1009,
  [5589] = 5552,
  [5590] = 2422,
  [5591] = 1010,
  [5592] = 2427,
  [5593] = 5552,
  [5594] = 4140,
  [5595] = 967,
  [5596] = 5099,
  [5597] = 5100,
  [5598] = 2322,
  [5599] = 5575,
  [5600] = 5544,
  [5601] = 5544,
  [5602] = 4052,
  [5603] = 2218,
  [5604] = 993,
  [5605] = 4096,
  [5606] = 2330,
  [5607] = 5542,
  [5608] = 2425,
  [5609] = 2435,
  [5610] = 5556,
  [5611] = 5152,
  [5612] = 5552,
  [5613] = 5613,
  [5614] = 5549,
  [5615] = 4155,
  [5616] = 4154,
  [5617] = 989,
  [5618] = 5542,
  [5619] = 4167,
  [5620] = 5620,
  [5621] = 5544,
  [5622] = 5556,
  [5623] = 5575,
  [5624] = 2358,
  [5625] = 5575,
  [5626] = 2291,
  [5627] = 5542,
  [5628] = 5549,
  [5629] = 4197,
  [5630] = 2855,
  [5631] = 990,
  [5632] = 5544,
  [5633] = 2866,
  [5634] = 995,
  [5635] = 5635,
  [5636] = 2852,
  [5637] = 1004,
  [5638] = 2246,
  [5639] = 5556,
  [5640] = 2248,
  [5641] = 4269,
  [5642] = 2249,
  [5643] = 5556,
  [5644] = 4271,
  [5645] = 5552,
  [5646] = 5544,
  [5647] = 5064,
  [5648] = 5542,
  [5649] = 2850,
  [5650] = 2870,
  [5651] = 2376,
  [5652] = 5542,
  [5653] = 2406,
  [5654] = 5575,
  [5655] = 2778,
  [5656] = 2779,
  [5657] = 2378,
  [5658] = 2380,
  [5659] = 5549,
  [5660] = 2786,
  [5661] = 5542,
  [5662] = 4189,
  [5663] = 2383,
  [5664] = 5552,
  [5665] = 5549,
  [5666] = 5666,
  [5667] = 2387,
  [5668] = 4169,
  [5669] = 5544,
  [5670] = 5613,
  [5671] = 4198,
  [5672] = 5549,
  [5673] = 5575,
  [5674] = 5575,
  [5675] = 5675,
  [5676] = 5552,
  [5677] = 5677,
  [5678] = 2360,
  [5679] = 5575,
  [5680] = 5680,
  [5681] = 2426,
  [5682] = 5556,
  [5683] = 5544,
  [5684] = 4253,
  [5685] = 5685,
  [5686] = 4201,
  [5687] = 5549,
  [5688] = 5556,
  [5689] = 5549,
  [5690] = 991,
  [5691] = 5635,
  [5692] = 5575,
  [5693] = 4206,
  [5694] = 3991,
  [5695] = 5695,
  [5696] = 4210,
  [5697] = 5552,
  [5698] = 5542,
  [5699] = 5699,
  [5700] = 5544,
  [5701] = 4255,
  [5702] = 2428,
  [5703] = 2441,
  [5704] = 5704,
  [5705] = 5542,
  [5706] = 5544,
  [5707] = 5228,
  [5708] = 5067,
  [5709] = 5068,
  [5710] = 5217,
  [5711] = 5237,
  [5712] = 5070,
  [5713] = 5226,
  [5714] = 5073,
  [5715] = 5074,
  [5716] = 5205,
  [5717] = 5075,
  [5718] = 5556,
  [5719] = 2795,
  [5720] = 5542,
  [5721] = 2376,
  [5722] = 2377,
  [5723] = 2378,
  [5724] = 2380,
  [5725] = 2381,
  [5726] = 2383,
  [5727] = 2384,
  [5728] = 2385,
  [5729] = 2387,
  [5730] = 2798,
  [5731] = 5675,
  [5732] = 4168,
  [5733] = 4141,
  [5734] = 5552,
  [5735] = 5151,
  [5736] = 4145,
  [5737] = 4146,
  [5738] = 4147,
  [5739] = 5739,
  [5740] = 5544,
  [5741] = 5635,
  [5742] = 5695,
  [5743] = 999,
  [5744] = 2335,
  [5745] = 4333,
  [5746] = 4183,
  [5747] = 5542,
  [5748] = 4202,
  [5749] = 5575,
  [5750] = 5575,
  [5751] = 5549,
  [5752] = 5549,
  [5753] = 2247,
  [5754] = 2214,
  [5755] = 4192,
  [5756] = 2215,
  [5757] = 5757,
  [5758] = 4170,
  [5759] = 5575,
  [5760] = 4279,
  [5761] = 5575,
  [5762] = 2265,
  [5763] = 5556,
  [5764] = 5549,
  [5765] = 4228,
  [5766] = 5579,
  [5767] = 5549,
  [5768] = 2849,
  [5769] = 4240,
  [5770] = 5620,
  [5771] = 5579,
  [5772] = 5552,
  [5773] = 2312,
  [5774] = 2444,
  [5775] = 4193,
  [5776] = 4195,
  [5777] = 5695,
  [5778] = 5556,
  [5779] = 5556,
  [5780] = 2275,
  [5781] = 5781,
  [5782] = 5568,
  [5783] = 4150,
  [5784] = 2864,
  [5785] = 5552,
  [5786] = 5542,
  [5787] = 5787,
  [5788] = 4351,
  [5789] = 4220,
  [5790] = 4321,
  [5791] = 5791,
  [5792] = 4254,
  [5793] = 4306,
  [5794] = 5794,
  [5795] = 4313,
  [5796] = 4247,
  [5797] = 4323,
  [5798] = 5798,
  [5799] = 4205,
  [5800] = 4350,
  [5801] = 5801,
  [5802] = 4234,
  [5803] = 4258,
  [5804] = 4224,
  [5805] = 4347,
  [5806] = 5806,
  [5807] = 5794,
  [5808] = 4349,
  [5809] = 4204,
  [5810] = 4277,
  [5811] = 4213,
  [5812] = 4231,
  [5813] = 4328,
  [5814] = 4263,
  [5815] = 4322,
  [5816] = 4356,
  [5817] = 5791,
  [5818] = 4208,
  [5819] = 4246,
  [5820] = 5794,
  [5821] = 5821,
  [5822] = 5822,
  [5823] = 5823,
  [5824] = 4285,
  [5825] = 4236,
  [5826] = 4294,
  [5827] = 4342,
  [5828] = 4227,
  [5829] = 4354,
  [5830] = 4207,
  [5831] = 4272,
  [5832] = 5801,
  [5833] = 4245,
  [5834] = 4989,
  [5835] = 5835,
  [5836] = 5836,
  [5837] = 4251,
  [5838] = 5838,
  [5839] = 5839,
  [5840] = 5840,
  [5841] = 5841,
  [5842] = 4261,
  [5843] = 5794,
  [5844] = 5801,
  [5845] = 5794,
  [5846] = 3941,
  [5847] = 5847,
  [5848] = 5794,
  [5849] = 5791,
  [5850] = 4311,
  [5851] = 4312,
  [5852] = 5835,
  [5853] = 5847,
  [5854] = 5836,
  [5855] = 4316,
  [5856] = 5856,
  [5857] = 5857,
  [5858] = 5787,
  [5859] = 4317,
  [5860] = 5841,
  [5861] = 4334,
  [5862] = 5862,
  [5863] = 5863,
  [5864] = 5836,
  [5865] = 4348,
  [5866] = 5841,
  [5867] = 5836,
  [5868] = 5868,
  [5869] = 4230,
  [5870] = 5870,
  [5871] = 5841,
  [5872] = 5870,
  [5873] = 5873,
  [5874] = 4319,
  [5875] = 5836,
  [5876] = 4241,
  [5877] = 5794,
  [5878] = 5878,
  [5879] = 4325,
  [5880] = 5836,
  [5881] = 4326,
  [5882] = 4329,
  [5883] = 5836,
  [5884] = 5884,
  [5885] = 5847,
  [5886] = 5836,
  [5887] = 4344,
  [5888] = 4194,
  [5889] = 4299,
  [5890] = 5856,
  [5891] = 4327,
  [5892] = 5838,
  [5893] = 5836,
  [5894] = 5894,
  [5895] = 5895,
  [5896] = 5894,
  [5897] = 4256,
  [5898] = 4276,
  [5899] = 5836,
  [5900] = 4278,
  [5901] = 5835,
  [5902] = 5836,
  [5903] = 4219,
  [5904] = 4248,
  [5905] = 4280,
  [5906] = 5836,
  [5907] = 5836,
  [5908] = 4180,
  [5909] = 4309,
  [5910] = 4318,
  [5911] = 4301,
  [5912] = 5912,
  [5913] = 4962,
  [5914] = 5794,
  [5915] = 4340,
  [5916] = 4289,
  [5917] = 5806,
  [5918] = 4292,
  [5919] = 5919,
  [5920] = 4223,
  [5921] = 5878,
  [5922] = 4304,
  [5923] = 4257,
  [5924] = 4314,
  [5925] = 4203,
  [5926] = 4262,
  [5927] = 5894,
  [5928] = 5928,
  [5929] = 5841,
  [5930] = 4265,
  [5931] = 5931,
  [5932] = 4249,
  [5933] = 4266,
  [5934] = 4267,
  [5935] = 4346,
  [5936] = 5878,
  [5937] = 4332,
  [5938] = 952,
  [5939] = 5856,
  [5940] = 5895,
  [5941] = 4268,
  [5942] = 5794,
  [5943] = 4341,
  [5944] = 4225,
  [5945] = 4226,
  [5946] = 5856,
  [5947] = 5947,
  [5948] = 4287,
  [5949] = 4288,
  [5950] = 5950,
  [5951] = 4290,
  [5952] = 5952,
  [5953] = 5878,
  [5954] = 4291,
  [5955] = 5895,
  [5956] = 4293,
  [5957] = 5835,
  [5958] = 5958,
  [5959] = 4295,
  [5960] = 4298,
  [5961] = 4300,
  [5962] = 5806,
  [5963] = 5841,
  [5964] = 4302,
  [5965] = 4305,
  [5966] = 5791,
  [5967] = 4148,
  [5968] = 4152,
  [5969] = 4133,
  [5970] = 4218,
  [5971] = 5870,
  [5972] = 5972,
  [5973] = 4243,
  [5974] = 4250,
  [5975] = 4259,
  [5976] = 4275,
  [5977] = 4297,
  [5978] = 4335,
  [5979] = 4338,
  [5980] = 5787,
  [5981] = 4238,
  [5982] = 5794,
  [5983] = 5794,
  [5984] = 4260,
  [5985] = 4296,
  [5986] = 4252,
  [5987] = 4343,
  [5988] = 4211,
  [5989] = 4237,
  [5990] = 4324,
  [5991] = 5794,
  [5992] = 4345,
  [5993] = 5993,
  [5994] = 4212,
  [5995] = 5794,
  [5996] = 5894,
  [5997] = 4134,
  [5998] = 4216,
  [5999] = 5993,
  [6000] = 5928,
  [6001] = 4352,
  [6002] = 4353,
  [6003] = 5993,
  [6004] = 5928,
  [6005] = 5993,
  [6006] = 5928,
  [6007] = 5838,
  [6008] = 6008,
  [6009] = 4339,
  [6010] = 5801,
  [6011] = 4264,
  [6012] = 5870,
  [6013] = 6013,
  [6014] = 4162,
  [6015] = 6015,
  [6016] = 4361,
  [6017] = 6017,
  [6018] = 4373,
  [6019] = 6019,
  [6020] = 6020,
  [6021] = 6021,
  [6022] = 6022,
  [6023] = 6023,
  [6024] = 6024,
  [6025] = 6025,
  [6026] = 6026,
  [6027] = 6027,
  [6028] = 6028,
  [6029] = 4387,
  [6030] = 4380,
  [6031] = 6031,
  [6032] = 6015,
  [6033] = 1929,
  [6034] = 6034,
  [6035] = 4404,
  [6036] = 4460,
  [6037] = 6013,
  [6038] = 6015,
  [6039] = 6023,
  [6040] = 6013,
  [6041] = 6041,
  [6042] = 6042,
  [6043] = 1939,
  [6044] = 4377,
  [6045] = 4416,
  [6046] = 4472,
  [6047] = 6047,
  [6048] = 6048,
  [6049] = 6024,
  [6050] = 2322,
  [6051] = 6015,
  [6052] = 6052,
  [6053] = 6027,
  [6054] = 6031,
  [6055] = 5017,
  [6056] = 6056,
  [6057] = 6057,
  [6058] = 6058,
  [6059] = 6059,
  [6060] = 2358,
  [6061] = 2291,
  [6062] = 4363,
  [6063] = 6063,
  [6064] = 6064,
  [6065] = 6026,
  [6066] = 6066,
  [6067] = 6067,
  [6068] = 6068,
  [6069] = 6031,
  [6070] = 6070,
  [6071] = 6071,
  [6072] = 6072,
  [6073] = 5033,
  [6074] = 5036,
  [6075] = 6057,
  [6076] = 6017,
  [6077] = 6077,
  [6078] = 4476,
  [6079] = 5043,
  [6080] = 6080,
  [6081] = 6034,
  [6082] = 6025,
  [6083] = 6031,
  [6084] = 6084,
  [6085] = 6072,
  [6086] = 6041,
  [6087] = 6056,
  [6088] = 6067,
  [6089] = 6042,
  [6090] = 6013,
  [6091] = 6015,
  [6092] = 6019,
  [6093] = 6022,
  [6094] = 6013,
  [6095] = 4410,
  [6096] = 4400,
  [6097] = 6097,
  [6098] = 5075,
  [6099] = 6097,
  [6100] = 6100,
  [6101] = 3975,
  [6102] = 6102,
  [6103] = 6103,
  [6104] = 4148,
  [6105] = 6105,
  [6106] = 4152,
  [6107] = 6015,
  [6108] = 6013,
  [6109] = 5067,
  [6110] = 6068,
  [6111] = 4215,
  [6112] = 4222,
  [6113] = 4081,
  [6114] = 6114,
  [6115] = 6067,
  [6116] = 4133,
  [6117] = 6117,
  [6118] = 6034,
  [6119] = 6105,
  [6120] = 6120,
  [6121] = 6121,
  [6122] = 6122,
  [6123] = 6123,
  [6124] = 6124,
  [6125] = 6063,
  [6126] = 4378,
  [6127] = 6105,
  [6128] = 6041,
  [6129] = 6015,
  [6130] = 6130,
  [6131] = 6027,
  [6132] = 6027,
  [6133] = 4007,
  [6134] = 6134,
  [6135] = 4281,
  [6136] = 2304,
  [6137] = 6130,
  [6138] = 6077,
  [6139] = 6139,
  [6140] = 6013,
  [6141] = 6013,
  [6142] = 6058,
  [6143] = 6063,
  [6144] = 6027,
  [6145] = 6025,
  [6146] = 6146,
  [6147] = 6047,
  [6148] = 6139,
  [6149] = 6013,
  [6150] = 6102,
  [6151] = 4412,
  [6152] = 4415,
  [6153] = 4148,
  [6154] = 4441,
  [6155] = 4152,
  [6156] = 6048,
  [6157] = 4133,
  [6158] = 6117,
  [6159] = 6015,
  [6160] = 6160,
  [6161] = 3561,
  [6162] = 4541,
  [6163] = 4516,
  [6164] = 6164,
  [6165] = 6165,
  [6166] = 4379,
  [6167] = 6080,
  [6168] = 6130,
  [6169] = 6134,
  [6170] = 6170,
  [6171] = 6015,
  [6172] = 6025,
  [6173] = 6072,
  [6174] = 6027,
  [6175] = 6175,
  [6176] = 6057,
  [6177] = 6025,
  [6178] = 6070,
  [6179] = 6026,
  [6180] = 4459,
  [6181] = 6181,
  [6182] = 6019,
  [6183] = 4307,
  [6184] = 4381,
  [6185] = 6028,
  [6186] = 6013,
  [6187] = 4388,
  [6188] = 6188,
  [6189] = 4110,
  [6190] = 6064,
  [6191] = 6027,
  [6192] = 4111,
  [6193] = 6134,
  [6194] = 4390,
  [6195] = 6077,
  [6196] = 6196,
  [6197] = 6059,
  [6198] = 6165,
  [6199] = 6027,
  [6200] = 6015,
  [6201] = 6026,
  [6202] = 6160,
  [6203] = 6015,
  [6204] = 6164,
  [6205] = 6013,
  [6206] = 6028,
  [6207] = 6027,
  [6208] = 6026,
  [6209] = 6103,
  [6210] = 6013,
  [6211] = 6071,
  [6212] = 6212,
  [6213] = 6070,
  [6214] = 1931,
  [6215] = 6025,
  [6216] = 6026,
  [6217] = 4456,
  [6218] = 4159,
  [6219] = 4461,
  [6220] = 4364,
  [6221] = 6117,
  [6222] = 6026,
  [6223] = 6160,
  [6224] = 6224,
  [6225] = 6072,
  [6226] = 6026,
  [6227] = 6227,
  [6228] = 6027,
  [6229] = 6023,
  [6230] = 3534,
  [6231] = 4473,
  [6232] = 6026,
  [6233] = 6165,
  [6234] = 6034,
  [6235] = 6026,
  [6236] = 6026,
  [6237] = 6017,
  [6238] = 6238,
  [6239] = 6026,
  [6240] = 4160,
  [6241] = 4442,
  [6242] = 6242,
  [6243] = 4371,
  [6244] = 6026,
  [6245] = 4041,
  [6246] = 4070,
  [6247] = 6097,
  [6248] = 6084,
  [6249] = 4072,
  [6250] = 6027,
  [6251] = 6165,
  [6252] = 6072,
  [6253] = 6102,
  [6254] = 6056,
  [6255] = 6031,
  [6256] = 6130,
  [6257] = 4158,
  [6258] = 6134,
  [6259] = 4134,
  [6260] = 6170,
  [6261] = 6022,
  [6262] = 6034,
  [6263] = 6058,
  [6264] = 6123,
  [6265] = 4134,
  [6266] = 6024,
  [6267] = 6084,
  [6268] = 6268,
  [6269] = 4542,
  [6270] = 6270,
  [6271] = 6123,
  [6272] = 4376,
  [6273] = 6052,
  [6274] = 6064,
  [6275] = 6130,
  [6276] = 6276,
  [6277] = 6134,
  [6278] = 4397,
  [6279] = 6071,
  [6280] = 6034,
  [6281] = 6020,
  [6282] = 6170,
  [6283] = 6027,
  [6284] = 6015,
  [6285] = 6021,
  [6286] = 2427,
  [6287] = 6170,
  [6288] = 6170,
  [6289] = 6242,
  [6290] = 6117,
  [6291] = 6020,
  [6292] = 6130,
  [6293] = 6134,
  [6294] = 6170,
  [6295] = 6015,
  [6296] = 6105,
  [6297] = 6042,
  [6298] = 6048,
  [6299] = 6041,
  [6300] = 6013,
  [6301] = 6160,
  [6302] = 2435,
  [6303] = 6059,
  [6304] = 6027,
  [6305] = 6305,
  [6306] = 6306,
  [6307] = 6307,
  [6308] = 6308,
  [6309] = 6309,
  [6310] = 6310,
  [6311] = 6311,
  [6312] = 6312,
  [6313] = 6313,
  [6314] = 6314,
  [6315] = 6315,
  [6316] = 6316,
  [6317] = 6308,
  [6318] = 6318,
  [6319] = 6312,
  [6320] = 6320,
  [6321] = 6311,
  [6322] = 6313,
  [6323] = 6323,
  [6324] = 6324,
  [6325] = 6325,
  [6326] = 6326,
  [6327] = 6327,
  [6328] = 6307,
  [6329] = 6329,
  [6330] = 6330,
  [6331] = 6331,
  [6332] = 6332,
  [6333] = 6333,
  [6334] = 6334,
  [6335] = 6335,
  [6336] = 6336,
  [6337] = 6337,
  [6338] = 6338,
  [6339] = 6339,
  [6340] = 6340,
  [6341] = 6341,
  [6342] = 6341,
  [6343] = 6343,
  [6344] = 6344,
  [6345] = 6345,
  [6346] = 6315,
  [6347] = 6305,
  [6348] = 6348,
  [6349] = 6349,
  [6350] = 6334,
  [6351] = 6351,
  [6352] = 6352,
  [6353] = 6353,
  [6354] = 6337,
  [6355] = 6355,
  [6356] = 6356,
  [6357] = 6320,
  [6358] = 6307,
  [6359] = 6359,
  [6360] = 6355,
  [6361] = 6308,
  [6362] = 4549,
  [6363] = 6339,
  [6364] = 6364,
  [6365] = 6308,
  [6366] = 6315,
  [6367] = 6367,
  [6368] = 6368,
  [6369] = 6313,
  [6370] = 6311,
  [6371] = 6323,
  [6372] = 4522,
  [6373] = 6325,
  [6374] = 6336,
  [6375] = 6375,
  [6376] = 6312,
  [6377] = 6377,
  [6378] = 6378,
  [6379] = 6379,
  [6380] = 6313,
  [6381] = 4630,
  [6382] = 6382,
  [6383] = 6383,
  [6384] = 6375,
  [6385] = 6385,
  [6386] = 6386,
  [6387] = 6387,
  [6388] = 6388,
  [6389] = 6389,
  [6390] = 6330,
  [6391] = 6391,
  [6392] = 6392,
  [6393] = 6311,
  [6394] = 6394,
  [6395] = 6395,
  [6396] = 6334,
  [6397] = 6397,
  [6398] = 6398,
  [6399] = 6399,
  [6400] = 6400,
  [6401] = 6341,
  [6402] = 6324,
  [6403] = 6385,
  [6404] = 4160,
  [6405] = 6405,
  [6406] = 6305,
  [6407] = 6407,
  [6408] = 6364,
  [6409] = 6323,
  [6410] = 6410,
  [6411] = 6386,
  [6412] = 6351,
  [6413] = 6307,
  [6414] = 6414,
  [6415] = 6400,
  [6416] = 6405,
  [6417] = 6308,
  [6418] = 4544,
  [6419] = 6349,
  [6420] = 6311,
  [6421] = 6421,
  [6422] = 4539,
  [6423] = 6312,
  [6424] = 4551,
  [6425] = 6425,
  [6426] = 6312,
  [6427] = 6313,
  [6428] = 6379,
  [6429] = 6410,
  [6430] = 6356,
  [6431] = 4494,
  [6432] = 6378,
  [6433] = 4554,
  [6434] = 6345,
  [6435] = 6414,
  [6436] = 6344,
  [6437] = 4625,
  [6438] = 4626,
  [6439] = 6331,
  [6440] = 6313,
  [6441] = 6324,
  [6442] = 6334,
  [6443] = 6421,
  [6444] = 3908,
  [6445] = 6387,
  [6446] = 6446,
  [6447] = 6447,
  [6448] = 6308,
  [6449] = 6334,
  [6450] = 6311,
  [6451] = 6341,
  [6452] = 6341,
  [6453] = 4162,
  [6454] = 6454,
  [6455] = 6455,
  [6456] = 6331,
  [6457] = 6457,
  [6458] = 6458,
  [6459] = 6459,
  [6460] = 6337,
  [6461] = 6340,
  [6462] = 6344,
  [6463] = 6307,
  [6464] = 6464,
  [6465] = 6349,
  [6466] = 6307,
  [6467] = 6467,
  [6468] = 6348,
  [6469] = 6355,
  [6470] = 6308,
  [6471] = 6367,
  [6472] = 6368,
  [6473] = 6310,
  [6474] = 6311,
  [6475] = 6335,
  [6476] = 6388,
  [6477] = 6375,
  [6478] = 6385,
  [6479] = 6386,
  [6480] = 6387,
  [6481] = 6481,
  [6482] = 6306,
  [6483] = 6392,
  [6484] = 6394,
  [6485] = 6485,
  [6486] = 6486,
  [6487] = 6312,
  [6488] = 6313,
  [6489] = 6394,
  [6490] = 3916,
  [6491] = 6308,
  [6492] = 1987,
  [6493] = 6315,
  [6494] = 6331,
  [6495] = 6495,
  [6496] = 6496,
  [6497] = 6323,
  [6498] = 6337,
  [6499] = 6340,
  [6500] = 6344,
  [6501] = 6355,
  [6502] = 3919,
  [6503] = 6392,
  [6504] = 6425,
  [6505] = 6311,
  [6506] = 4439,
  [6507] = 6334,
  [6508] = 6331,
  [6509] = 6337,
  [6510] = 6344,
  [6511] = 6511,
  [6512] = 6351,
  [6513] = 3938,
  [6514] = 6399,
  [6515] = 6355,
  [6516] = 6341,
  [6517] = 6345,
  [6518] = 3936,
  [6519] = 6383,
  [6520] = 6331,
  [6521] = 6391,
  [6522] = 6522,
  [6523] = 6337,
  [6524] = 6351,
  [6525] = 6307,
  [6526] = 6331,
  [6527] = 6455,
  [6528] = 6337,
  [6529] = 6308,
  [6530] = 6312,
  [6531] = 6379,
  [6532] = 4501,
  [6533] = 6331,
  [6534] = 6324,
  [6535] = 6337,
  [6536] = 6311,
  [6537] = 6447,
  [6538] = 6312,
  [6539] = 6331,
  [6540] = 6337,
  [6541] = 6314,
  [6542] = 6313,
  [6543] = 6313,
  [6544] = 6325,
  [6545] = 4139,
  [6546] = 6333,
  [6547] = 6331,
  [6548] = 6337,
  [6549] = 4510,
  [6550] = 6340,
  [6551] = 6551,
  [6552] = 6331,
  [6553] = 6334,
  [6554] = 6337,
  [6555] = 6312,
  [6556] = 6400,
  [6557] = 6389,
  [6558] = 6410,
  [6559] = 4512,
  [6560] = 6414,
  [6561] = 6331,
  [6562] = 6337,
  [6563] = 6315,
  [6564] = 6334,
  [6565] = 6324,
  [6566] = 6331,
  [6567] = 6337,
  [6568] = 6341,
  [6569] = 4495,
  [6570] = 6331,
  [6571] = 6571,
  [6572] = 6337,
  [6573] = 4144,
  [6574] = 6331,
  [6575] = 6575,
  [6576] = 6337,
  [6577] = 6331,
  [6578] = 6397,
  [6579] = 6337,
  [6580] = 6331,
  [6581] = 6337,
  [6582] = 6331,
  [6583] = 6337,
  [6584] = 6331,
  [6585] = 6307,
  [6586] = 6337,
  [6587] = 6331,
  [6588] = 6588,
  [6589] = 6337,
  [6590] = 6331,
  [6591] = 6337,
  [6592] = 6331,
  [6593] = 6337,
  [6594] = 6331,
  [6595] = 6337,
  [6596] = 6331,
  [6597] = 6337,
  [6598] = 4989,
  [6599] = 6599,
  [6600] = 6332,
  [6601] = 6327,
  [6602] = 6391,
  [6603] = 6397,
  [6604] = 6308,
  [6605] = 6447,
  [6606] = 6457,
  [6607] = 6459,
  [6608] = 6339,
  [6609] = 6457,
  [6610] = 6311,
  [6611] = 6551,
  [6612] = 6318,
  [6613] = 6395,
  [6614] = 6407,
  [6615] = 6481,
  [6616] = 6312,
  [6617] = 6334,
  [6618] = 6458,
  [6619] = 6446,
  [6620] = 6310,
  [6621] = 6313,
  [6622] = 4502,
  [6623] = 4503,
  [6624] = 6459,
  [6625] = 6338,
  [6626] = 4540,
  [6627] = 4508,
  [6628] = 6628,
  [6629] = 6379,
  [6630] = 6599,
  [6631] = 6341,
  [6632] = 4649,
  [6633] = 4650,
  [6634] = 6634,
  [6635] = 6334,
  [6636] = 6332,
  [6637] = 6341,
  [6638] = 6638,
  [6639] = 6341,
  [6640] = 6307,
  [6641] = 6377,
  [6642] = 6571,
  [6643] = 6367,
  [6644] = 6368,
  [6645] = 6308,
  [6646] = 6306,
  [6647] = 4159,
  [6648] = 6311,
  [6649] = 6649,
  [6650] = 6312,
  [6651] = 6335,
  [6652] = 6599,
  [6653] = 6313,
  [6654] = 6307,
  [6655] = 6312,
  [6656] = 6551,
  [6657] = 6481,
  [6658] = 6658,
  [6659] = 6338,
  [6660] = 6599,
  [6661] = 6308,
  [6662] = 6551,
  [6663] = 6481,
  [6664] = 6599,
  [6665] = 4524,
  [6666] = 6455,
  [6667] = 6551,
  [6668] = 6334,
  [6669] = 6481,
  [6670] = 6315,
  [6671] = 6599,
  [6672] = 6551,
  [6673] = 6341,
  [6674] = 6481,
  [6675] = 6389,
  [6676] = 6599,
  [6677] = 6551,
  [6678] = 6481,
  [6679] = 6599,
  [6680] = 6311,
  [6681] = 4518,
  [6682] = 6551,
  [6683] = 6481,
  [6684] = 6599,
  [6685] = 6551,
  [6686] = 6481,
  [6687] = 6307,
  [6688] = 6599,
  [6689] = 4525,
  [6690] = 6324,
  [6691] = 6305,
  [6692] = 6551,
  [6693] = 6481,
  [6694] = 6599,
  [6695] = 4527,
  [6696] = 6551,
  [6697] = 6481,
  [6698] = 6599,
  [6699] = 6334,
  [6700] = 6551,
  [6701] = 6481,
  [6702] = 6308,
  [6703] = 6599,
  [6704] = 6551,
  [6705] = 6575,
  [6706] = 6481,
  [6707] = 6599,
  [6708] = 6551,
  [6709] = 6599,
  [6710] = 6551,
  [6711] = 6599,
  [6712] = 6551,
  [6713] = 6599,
  [6714] = 6551,
  [6715] = 6599,
  [6716] = 6551,
  [6717] = 6599,
  [6718] = 6551,
  [6719] = 6599,
  [6720] = 6551,
  [6721] = 6599,
  [6722] = 6599,
  [6723] = 6599,
  [6724] = 6599,
  [6725] = 6327,
  [6726] = 6311,
  [6727] = 6349,
  [6728] = 6312,
  [6729] = 6379,
  [6730] = 6638,
  [6731] = 6367,
  [6732] = 6368,
  [6733] = 6341,
  [6734] = 6312,
  [6735] = 6352,
  [6736] = 6353,
  [6737] = 6375,
  [6738] = 6738,
  [6739] = 6738,
  [6740] = 6313,
  [6741] = 6455,
  [6742] = 6335,
  [6743] = 6743,
  [6744] = 6307,
  [6745] = 6313,
  [6746] = 6379,
  [6747] = 6307,
  [6748] = 3917,
  [6749] = 6399,
  [6750] = 6750,
  [6751] = 6458,
  [6752] = 6334,
  [6753] = 6306,
  [6754] = 6356,
  [6755] = 6378,
  [6756] = 6421,
  [6757] = 6389,
  [6758] = 6341,
  [6759] = 6405,
  [6760] = 6760,
  [6761] = 6761,
  [6762] = 6762,
  [6763] = 6763,
  [6764] = 6760,
  [6765] = 6765,
  [6766] = 6766,
  [6767] = 6767,
  [6768] = 6768,
  [6769] = 6769,
  [6770] = 6770,
  [6771] = 6771,
  [6772] = 6772,
  [6773] = 6773,
  [6774] = 6774,
  [6775] = 6775,
  [6776] = 6776,
  [6777] = 6760,
  [6778] = 6778,
  [6779] = 6760,
  [6780] = 6780,
  [6781] = 6781,
  [6782] = 6782,
  [6783] = 6783,
  [6784] = 6775,
  [6785] = 6768,
  [6786] = 6786,
  [6787] = 6766,
  [6788] = 6778,
  [6789] = 6760,
  [6790] = 6790,
  [6791] = 6775,
  [6792] = 6766,
  [6793] = 6793,
  [6794] = 6766,
  [6795] = 6795,
  [6796] = 6790,
  [6797] = 6778,
  [6798] = 6778,
  [6799] = 6775,
  [6800] = 6766,
  [6801] = 6760,
  [6802] = 6760,
  [6803] = 6773,
  [6804] = 6804,
  [6805] = 6805,
  [6806] = 6783,
  [6807] = 6760,
  [6808] = 6808,
  [6809] = 6775,
  [6810] = 6810,
  [6811] = 6769,
  [6812] = 6770,
  [6813] = 6813,
  [6814] = 6814,
  [6815] = 6773,
  [6816] = 6766,
  [6817] = 6817,
  [6818] = 6818,
  [6819] = 6773,
  [6820] = 6820,
  [6821] = 6821,
  [6822] = 6778,
  [6823] = 6820,
  [6824] = 6795,
  [6825] = 6775,
  [6826] = 6775,
  [6827] = 6773,
  [6828] = 6783,
  [6829] = 6760,
  [6830] = 6773,
  [6831] = 6795,
  [6832] = 6832,
  [6833] = 6833,
  [6834] = 6773,
  [6835] = 6835,
  [6836] = 6773,
  [6837] = 6773,
  [6838] = 6838,
  [6839] = 6839,
  [6840] = 6832,
  [6841] = 6835,
  [6842] = 6780,
  [6843] = 6766,
  [6844] = 6844,
  [6845] = 6845,
  [6846] = 6778,
  [6847] = 6780,
  [6848] = 6821,
  [6849] = 6768,
  [6850] = 6850,
  [6851] = 6760,
  [6852] = 6852,
  [6853] = 6853,
  [6854] = 6854,
  [6855] = 6768,
  [6856] = 6856,
  [6857] = 6762,
  [6858] = 6763,
  [6859] = 6859,
  [6860] = 6766,
  [6861] = 6832,
  [6862] = 6862,
  [6863] = 6863,
  [6864] = 6778,
  [6865] = 6760,
  [6866] = 6866,
  [6867] = 6867,
  [6868] = 6868,
  [6869] = 6766,
  [6870] = 6775,
  [6871] = 6871,
  [6872] = 6768,
  [6873] = 6873,
  [6874] = 6766,
  [6875] = 6773,
  [6876] = 6876,
  [6877] = 6790,
  [6878] = 6878,
  [6879] = 6879,
  [6880] = 6778,
  [6881] = 6778,
  [6882] = 6839,
  [6883] = 6883,
  [6884] = 6760,
  [6885] = 6835,
  [6886] = 6805,
  [6887] = 6887,
  [6888] = 6771,
  [6889] = 6778,
  [6890] = 6760,
  [6891] = 6850,
  [6892] = 6853,
  [6893] = 6854,
  [6894] = 6856,
  [6895] = 6762,
  [6896] = 6763,
  [6897] = 6859,
  [6898] = 6786,
  [6899] = 6775,
  [6900] = 6900,
  [6901] = 6862,
  [6902] = 6839,
  [6903] = 6835,
  [6904] = 6773,
  [6905] = 6863,
  [6906] = 6850,
  [6907] = 6767,
  [6908] = 6854,
  [6909] = 6856,
  [6910] = 6762,
  [6911] = 6763,
  [6912] = 6859,
  [6913] = 6913,
  [6914] = 6770,
  [6915] = 6839,
  [6916] = 6835,
  [6917] = 6775,
  [6918] = 6850,
  [6919] = 6854,
  [6920] = 6856,
  [6921] = 6762,
  [6922] = 6763,
  [6923] = 6780,
  [6924] = 6805,
  [6925] = 6925,
  [6926] = 6839,
  [6927] = 6768,
  [6928] = 6850,
  [6929] = 6856,
  [6930] = 6762,
  [6931] = 6763,
  [6932] = 6839,
  [6933] = 6850,
  [6934] = 6793,
  [6935] = 6856,
  [6936] = 6762,
  [6937] = 6763,
  [6938] = 6766,
  [6939] = 6839,
  [6940] = 6766,
  [6941] = 6941,
  [6942] = 6942,
  [6943] = 6850,
  [6944] = 6856,
  [6945] = 6762,
  [6946] = 6763,
  [6947] = 4633,
  [6948] = 6948,
  [6949] = 6810,
  [6950] = 6850,
  [6951] = 6778,
  [6952] = 6952,
  [6953] = 6953,
  [6954] = 6856,
  [6955] = 6762,
  [6956] = 6763,
  [6957] = 6773,
  [6958] = 6760,
  [6959] = 6850,
  [6960] = 6952,
  [6961] = 6856,
  [6962] = 6762,
  [6963] = 6763,
  [6964] = 6760,
  [6965] = 6783,
  [6966] = 6795,
  [6967] = 6850,
  [6968] = 6845,
  [6969] = 6813,
  [6970] = 6856,
  [6971] = 6762,
  [6972] = 6763,
  [6973] = 6835,
  [6974] = 6832,
  [6975] = 6900,
  [6976] = 6850,
  [6977] = 6856,
  [6978] = 6762,
  [6979] = 6763,
  [6980] = 6863,
  [6981] = 6766,
  [6982] = 6850,
  [6983] = 6856,
  [6984] = 6762,
  [6985] = 6763,
  [6986] = 6778,
  [6987] = 6852,
  [6988] = 6850,
  [6989] = 6856,
  [6990] = 6762,
  [6991] = 6763,
  [6992] = 6992,
  [6993] = 6856,
  [6994] = 6762,
  [6995] = 6763,
  [6996] = 6850,
  [6997] = 6856,
  [6998] = 6762,
  [6999] = 6763,
  [7000] = 6793,
  [7001] = 7001,
  [7002] = 6850,
  [7003] = 6856,
  [7004] = 6762,
  [7005] = 6763,
  [7006] = 6795,
  [7007] = 6766,
  [7008] = 6850,
  [7009] = 6856,
  [7010] = 6762,
  [7011] = 6763,
  [7012] = 6853,
  [7013] = 6850,
  [7014] = 6856,
  [7015] = 6762,
  [7016] = 6763,
  [7017] = 6953,
  [7018] = 6850,
  [7019] = 6850,
  [7020] = 6768,
  [7021] = 6850,
  [7022] = 6814,
  [7023] = 6850,
  [7024] = 6790,
  [7025] = 6805,
  [7026] = 6883,
  [7027] = 6768,
  [7028] = 6813,
  [7029] = 6821,
  [7030] = 7030,
  [7031] = 6810,
  [7032] = 7032,
  [7033] = 7033,
  [7034] = 6863,
  [7035] = 7033,
  [7036] = 6887,
  [7037] = 6782,
  [7038] = 6771,
  [7039] = 6778,
  [7040] = 6765,
  [7041] = 6850,
  [7042] = 6774,
  [7043] = 6774,
  [7044] = 6766,
  [7045] = 6780,
  [7046] = 4621,
  [7047] = 6867,
  [7048] = 7048,
  [7049] = 6853,
  [7050] = 6776,
  [7051] = 6992,
  [7052] = 6900,
  [7053] = 6992,
  [7054] = 7054,
  [7055] = 7055,
  [7056] = 6773,
  [7057] = 7055,
  [7058] = 6854,
  [7059] = 6778,
  [7060] = 6873,
  [7061] = 6879,
  [7062] = 6900,
  [7063] = 6778,
  [7064] = 6773,
  [7065] = 6820,
  [7066] = 6786,
  [7067] = 6760,
  [7068] = 6835,
  [7069] = 4719,
  [7070] = 6948,
  [7071] = 6778,
  [7072] = 6775,
  [7073] = 2665,
  [7074] = 6760,
  [7075] = 6856,
  [7076] = 6762,
  [7077] = 6795,
  [7078] = 6838,
  [7079] = 7079,
  [7080] = 6775,
  [7081] = 7081,
  [7082] = 6763,
  [7083] = 6810,
  [7084] = 6838,
  [7085] = 7030,
  [7086] = 7086,
  [7087] = 7087,
  [7088] = 6793,
  [7089] = 7089,
  [7090] = 6768,
  [7091] = 7048,
  [7092] = 6775,
  [7093] = 6786,
  [7094] = 7032,
  [7095] = 6773,
  [7096] = 6768,
  [7097] = 6766,
  [7098] = 7048,
  [7099] = 6790,
  [7100] = 6780,
  [7101] = 6805,
  [7102] = 6795,
  [7103] = 6883,
  [7104] = 6821,
  [7105] = 7033,
  [7106] = 7106,
  [7107] = 7107,
  [7108] = 6776,
  [7109] = 6778,
  [7110] = 6873,
  [7111] = 6838,
  [7112] = 6775,
  [7113] = 7113,
  [7114] = 6876,
  [7115] = 6768,
  [7116] = 6887,
  [7117] = 7117,
  [7118] = 6760,
  [7119] = 6925,
  [7120] = 6776,
  [7121] = 6873,
  [7122] = 6760,
  [7123] = 2133,
  [7124] = 7048,
  [7125] = 7106,
  [7126] = 6771,
  [7127] = 6790,
  [7128] = 6775,
  [7129] = 7129,
  [7130] = 7130,
  [7131] = 6867,
  [7132] = 6859,
  [7133] = 7032,
  [7134] = 6778,
  [7135] = 6760,
  [7136] = 6770,
  [7137] = 6805,
  [7138] = 7138,
  [7139] = 4743,
  [7140] = 6854,
  [7141] = 6887,
  [7142] = 6768,
  [7143] = 6766,
  [7144] = 6765,
  [7145] = 6810,
  [7146] = 6768,
  [7147] = 6813,
  [7148] = 6992,
  [7149] = 7149,
  [7150] = 6773,
  [7151] = 6820,
  [7152] = 6775,
  [7153] = 6768,
  [7154] = 6778,
  [7155] = 6766,
  [7156] = 6795,
  [7157] = 5197,
  [7158] = 7054,
  [7159] = 6838,
  [7160] = 6839,
  [7161] = 7161,
  [7162] = 6850,
  [7163] = 6780,
  [7164] = 6767,
  [7165] = 6778,
  [7166] = 6766,
  [7167] = 6770,
  [7168] = 6780,
  [7169] = 6765,
  [7170] = 6783,
  [7171] = 6760,
  [7172] = 7001,
  [7173] = 7055,
  [7174] = 6862,
  [7175] = 7175,
  [7176] = 7138,
  [7177] = 6852,
  [7178] = 6766,
  [7179] = 6856,
  [7180] = 7055,
  [7181] = 6867,
  [7182] = 6859,
  [7183] = 6850,
  [7184] = 7184,
  [7185] = 7185,
  [7186] = 7186,
  [7187] = 7187,
  [7188] = 7188,
  [7189] = 7189,
  [7190] = 7187,
  [7191] = 7191,
  [7192] = 7192,
  [7193] = 7193,
  [7194] = 7194,
  [7195] = 7195,
  [7196] = 7185,
  [7197] = 7197,
  [7198] = 7198,
  [7199] = 7199,
  [7200] = 7200,
  [7201] = 7201,
  [7202] = 7202,
  [7203] = 7203,
  [7204] = 7204,
  [7205] = 7205,
  [7206] = 7206,
  [7207] = 7192,
  [7208] = 7208,
  [7209] = 7209,
  [7210] = 7210,
  [7211] = 7211,
  [7212] = 7212,
  [7213] = 7213,
  [7214] = 7214,
  [7215] = 7198,
  [7216] = 7216,
  [7217] = 7194,
  [7218] = 7218,
  [7219] = 7219,
  [7220] = 7220,
  [7221] = 7221,
  [7222] = 7222,
  [7223] = 7223,
  [7224] = 7224,
  [7225] = 7225,
  [7226] = 7226,
  [7227] = 7204,
  [7228] = 7199,
  [7229] = 7229,
  [7230] = 7230,
  [7231] = 7210,
  [7232] = 7232,
  [7233] = 7233,
  [7234] = 7208,
  [7235] = 7202,
  [7236] = 7236,
  [7237] = 7214,
  [7238] = 7238,
  [7239] = 7239,
  [7240] = 7187,
  [7241] = 7241,
  [7242] = 7242,
  [7243] = 7243,
  [7244] = 7244,
  [7245] = 7192,
  [7246] = 7189,
  [7247] = 7185,
  [7248] = 7248,
  [7249] = 7202,
  [7250] = 7250,
  [7251] = 7251,
  [7252] = 7212,
  [7253] = 7253,
  [7254] = 7254,
  [7255] = 7255,
  [7256] = 7256,
  [7257] = 7229,
  [7258] = 7258,
  [7259] = 7238,
  [7260] = 7243,
  [7261] = 7187,
  [7262] = 7262,
  [7263] = 7203,
  [7264] = 7264,
  [7265] = 7265,
  [7266] = 7192,
  [7267] = 7187,
  [7268] = 7186,
  [7269] = 7189,
  [7270] = 7270,
  [7271] = 7209,
  [7272] = 7193,
  [7273] = 7273,
  [7274] = 7195,
  [7275] = 7197,
  [7276] = 7276,
  [7277] = 7264,
  [7278] = 7200,
  [7279] = 7201,
  [7280] = 7185,
  [7281] = 7204,
  [7282] = 7221,
  [7283] = 7209,
  [7284] = 7243,
  [7285] = 7243,
  [7286] = 7186,
  [7287] = 7287,
  [7288] = 7202,
  [7289] = 7218,
  [7290] = 7219,
  [7291] = 7192,
  [7292] = 7221,
  [7293] = 7224,
  [7294] = 7294,
  [7295] = 7185,
  [7296] = 7192,
  [7297] = 7226,
  [7298] = 7191,
  [7299] = 7232,
  [7300] = 7233,
  [7301] = 7243,
  [7302] = 7187,
  [7303] = 7225,
  [7304] = 7204,
  [7305] = 7239,
  [7306] = 7244,
  [7307] = 7241,
  [7308] = 7242,
  [7309] = 7192,
  [7310] = 7310,
  [7311] = 7216,
  [7312] = 7312,
  [7313] = 7185,
  [7314] = 7248,
  [7315] = 7251,
  [7316] = 7253,
  [7317] = 7254,
  [7318] = 7255,
  [7319] = 7202,
  [7320] = 7320,
  [7321] = 7321,
  [7322] = 7322,
  [7323] = 7238,
  [7324] = 7198,
  [7325] = 7212,
  [7326] = 7193,
  [7327] = 7208,
  [7328] = 7328,
  [7329] = 7208,
  [7330] = 7186,
  [7331] = 7189,
  [7332] = 7328,
  [7333] = 7333,
  [7334] = 7334,
  [7335] = 7195,
  [7336] = 7197,
  [7337] = 7337,
  [7338] = 7338,
  [7339] = 7200,
  [7340] = 7201,
  [7341] = 7244,
  [7342] = 7203,
  [7343] = 7338,
  [7344] = 7224,
  [7345] = 7345,
  [7346] = 7225,
  [7347] = 7219,
  [7348] = 7320,
  [7349] = 7221,
  [7350] = 7350,
  [7351] = 7351,
  [7352] = 7204,
  [7353] = 7353,
  [7354] = 7226,
  [7355] = 7273,
  [7356] = 7356,
  [7357] = 7357,
  [7358] = 7210,
  [7359] = 7239,
  [7360] = 7360,
  [7361] = 7202,
  [7362] = 7214,
  [7363] = 7363,
  [7364] = 1873,
  [7365] = 7251,
  [7366] = 7253,
  [7367] = 7254,
  [7368] = 7255,
  [7369] = 7369,
  [7370] = 7370,
  [7371] = 7371,
  [7372] = 7256,
  [7373] = 7185,
  [7374] = 7338,
  [7375] = 7369,
  [7376] = 7186,
  [7377] = 7189,
  [7378] = 7334,
  [7379] = 7244,
  [7380] = 7310,
  [7381] = 7195,
  [7382] = 7197,
  [7383] = 7200,
  [7384] = 7201,
  [7385] = 7186,
  [7386] = 7220,
  [7387] = 7387,
  [7388] = 3230,
  [7389] = 7219,
  [7390] = 7221,
  [7391] = 7391,
  [7392] = 7256,
  [7393] = 7371,
  [7394] = 7394,
  [7395] = 7226,
  [7396] = 7357,
  [7397] = 7229,
  [7398] = 7225,
  [7399] = 7239,
  [7400] = 7243,
  [7401] = 7401,
  [7402] = 7229,
  [7403] = 7403,
  [7404] = 7251,
  [7405] = 7253,
  [7406] = 7254,
  [7407] = 7255,
  [7408] = 7408,
  [7409] = 7192,
  [7410] = 7410,
  [7411] = 7411,
  [7412] = 7189,
  [7413] = 7201,
  [7414] = 7353,
  [7415] = 7186,
  [7416] = 7189,
  [7417] = 7224,
  [7418] = 7204,
  [7419] = 7419,
  [7420] = 7195,
  [7421] = 7197,
  [7422] = 7200,
  [7423] = 7201,
  [7424] = 7229,
  [7425] = 7192,
  [7426] = 7357,
  [7427] = 7294,
  [7428] = 7219,
  [7429] = 7221,
  [7430] = 7202,
  [7431] = 7320,
  [7432] = 7238,
  [7433] = 7210,
  [7434] = 7226,
  [7435] = 7435,
  [7436] = 7238,
  [7437] = 7193,
  [7438] = 7239,
  [7439] = 7210,
  [7440] = 7202,
  [7441] = 7185,
  [7442] = 7371,
  [7443] = 7251,
  [7444] = 7253,
  [7445] = 7254,
  [7446] = 7255,
  [7447] = 7394,
  [7448] = 7401,
  [7449] = 7449,
  [7450] = 7203,
  [7451] = 7321,
  [7452] = 7328,
  [7453] = 7453,
  [7454] = 7186,
  [7455] = 7321,
  [7456] = 7208,
  [7457] = 7195,
  [7458] = 7197,
  [7459] = 7200,
  [7460] = 7201,
  [7461] = 4781,
  [7462] = 7202,
  [7463] = 7203,
  [7464] = 7244,
  [7465] = 7219,
  [7466] = 7221,
  [7467] = 7193,
  [7468] = 7223,
  [7469] = 7469,
  [7470] = 7322,
  [7471] = 7226,
  [7472] = 7243,
  [7473] = 7244,
  [7474] = 7208,
  [7475] = 7239,
  [7476] = 7198,
  [7477] = 7320,
  [7478] = 7370,
  [7479] = 7212,
  [7480] = 7251,
  [7481] = 7253,
  [7482] = 7254,
  [7483] = 7255,
  [7484] = 3243,
  [7485] = 7214,
  [7486] = 7192,
  [7487] = 7337,
  [7488] = 7225,
  [7489] = 7242,
  [7490] = 7229,
  [7491] = 7186,
  [7492] = 7419,
  [7493] = 7449,
  [7494] = 7197,
  [7495] = 7200,
  [7496] = 7201,
  [7497] = 7241,
  [7498] = 7224,
  [7499] = 7204,
  [7500] = 7238,
  [7501] = 7219,
  [7502] = 7221,
  [7503] = 7244,
  [7504] = 7210,
  [7505] = 7185,
  [7506] = 7338,
  [7507] = 7226,
  [7508] = 7203,
  [7509] = 7509,
  [7510] = 7510,
  [7511] = 7239,
  [7512] = 7394,
  [7513] = 7187,
  [7514] = 7514,
  [7515] = 7264,
  [7516] = 7251,
  [7517] = 7253,
  [7518] = 7254,
  [7519] = 7255,
  [7520] = 7195,
  [7521] = 7521,
  [7522] = 7401,
  [7523] = 7273,
  [7524] = 7345,
  [7525] = 7265,
  [7526] = 7526,
  [7527] = 7186,
  [7528] = 7528,
  [7529] = 7229,
  [7530] = 7197,
  [7531] = 7200,
  [7532] = 7201,
  [7533] = 7197,
  [7534] = 7203,
  [7535] = 7225,
  [7536] = 7536,
  [7537] = 7219,
  [7538] = 7221,
  [7539] = 7370,
  [7540] = 7214,
  [7541] = 7236,
  [7542] = 7200,
  [7543] = 7226,
  [7544] = 7264,
  [7545] = 7320,
  [7546] = 7345,
  [7547] = 7239,
  [7548] = 7548,
  [7549] = 7201,
  [7550] = 7238,
  [7551] = 7551,
  [7552] = 7251,
  [7553] = 7253,
  [7554] = 7254,
  [7555] = 7255,
  [7556] = 7556,
  [7557] = 7557,
  [7558] = 7210,
  [7559] = 7202,
  [7560] = 7185,
  [7561] = 7204,
  [7562] = 7186,
  [7563] = 7350,
  [7564] = 7244,
  [7565] = 7197,
  [7566] = 7200,
  [7567] = 7201,
  [7568] = 7193,
  [7569] = 7258,
  [7570] = 7570,
  [7571] = 7229,
  [7572] = 7219,
  [7573] = 7221,
  [7574] = 7574,
  [7575] = 7371,
  [7576] = 7351,
  [7577] = 7203,
  [7578] = 7226,
  [7579] = 7203,
  [7580] = 7187,
  [7581] = 7320,
  [7582] = 7239,
  [7583] = 7238,
  [7584] = 7208,
  [7585] = 7345,
  [7586] = 7203,
  [7587] = 7251,
  [7588] = 7253,
  [7589] = 7254,
  [7590] = 7255,
  [7591] = 7591,
  [7592] = 7592,
  [7593] = 7242,
  [7594] = 7526,
  [7595] = 7270,
  [7596] = 7394,
  [7597] = 7264,
  [7598] = 7186,
  [7599] = 7322,
  [7600] = 7198,
  [7601] = 7197,
  [7602] = 7200,
  [7603] = 7201,
  [7604] = 7310,
  [7605] = 7212,
  [7606] = 7357,
  [7607] = 7403,
  [7608] = 7219,
  [7609] = 7221,
  [7610] = 7264,
  [7611] = 7273,
  [7612] = 7242,
  [7613] = 7287,
  [7614] = 7226,
  [7615] = 7202,
  [7616] = 7242,
  [7617] = 7617,
  [7618] = 7239,
  [7619] = 7206,
  [7620] = 7620,
  [7621] = 7310,
  [7622] = 7287,
  [7623] = 7251,
  [7624] = 7253,
  [7625] = 7254,
  [7626] = 7255,
  [7627] = 7294,
  [7628] = 7214,
  [7629] = 7401,
  [7630] = 7187,
  [7631] = 7193,
  [7632] = 7270,
  [7633] = 7208,
  [7634] = 7186,
  [7635] = 7209,
  [7636] = 7201,
  [7637] = 7276,
  [7638] = 7210,
  [7639] = 7570,
  [7640] = 7226,
  [7641] = 7310,
  [7642] = 7328,
  [7643] = 7353,
  [7644] = 7239,
  [7645] = 7338,
  [7646] = 7251,
  [7647] = 7253,
  [7648] = 7254,
  [7649] = 7255,
  [7650] = 7202,
  [7651] = 7199,
  [7652] = 7321,
  [7653] = 7345,
  [7654] = 7242,
  [7655] = 7210,
  [7656] = 7225,
  [7657] = 7204,
  [7658] = 7202,
  [7659] = 7193,
  [7660] = 7253,
  [7661] = 7254,
  [7662] = 7255,
  [7663] = 7310,
  [7664] = 7310,
  [7665] = 7193,
  [7666] = 7208,
  [7667] = 7310,
  [7668] = 7243,
  [7669] = 7370,
  [7670] = 7253,
  [7671] = 7254,
  [7672] = 7255,
  [7673] = 7208,
  [7674] = 7351,
  [7675] = 7345,
  [7676] = 7338,
  [7677] = 7191,
  [7678] = 7202,
  [7679] = 7225,
  [7680] = 7253,
  [7681] = 7254,
  [7682] = 7255,
  [7683] = 7204,
  [7684] = 7244,
  [7685] = 7685,
  [7686] = 7219,
  [7687] = 7225,
  [7688] = 7204,
  [7689] = 7192,
  [7690] = 7253,
  [7691] = 7254,
  [7692] = 7255,
  [7693] = 7193,
  [7694] = 7225,
  [7695] = 7338,
  [7696] = 7204,
  [7697] = 7369,
  [7698] = 7698,
  [7699] = 7244,
  [7700] = 7253,
  [7701] = 7254,
  [7702] = 7255,
  [7703] = 7213,
  [7704] = 7685,
  [7705] = 7208,
  [7706] = 7206,
  [7707] = 7707,
  [7708] = 7203,
  [7709] = 7709,
  [7710] = 7253,
  [7711] = 7254,
  [7712] = 7255,
  [7713] = 7617,
  [7714] = 7193,
  [7715] = 7186,
  [7716] = 7187,
  [7717] = 7345,
  [7718] = 7338,
  [7719] = 7719,
  [7720] = 7203,
  [7721] = 7189,
  [7722] = 7617,
  [7723] = 7322,
  [7724] = 7208,
  [7725] = 7229,
  [7726] = 7726,
  [7727] = 7370,
  [7728] = 7719,
  [7729] = 7729,
  [7730] = 1875,
  [7731] = 7469,
  [7732] = 7321,
  [7733] = 7733,
  [7734] = 7198,
  [7735] = 7294,
  [7736] = 7369,
  [7737] = 7570,
  [7738] = 7345,
  [7739] = 7243,
  [7740] = 7740,
  [7741] = 7229,
  [7742] = 7236,
  [7743] = 7743,
  [7744] = 7204,
  [7745] = 7251,
  [7746] = 7435,
  [7747] = 7194,
  [7748] = 7229,
  [7749] = 7212,
  [7750] = 7750,
  [7751] = 7256,
  [7752] = 7243,
  [7753] = 7192,
  [7754] = 7218,
  [7755] = 7345,
  [7756] = 7195,
  [7757] = 7214,
  [7758] = 7351,
  [7759] = 7224,
  [7760] = 7363,
  [7761] = 7351,
  [7762] = 7411,
  [7763] = 7219,
  [7764] = 7764,
  [7765] = 1864,
  [7766] = 7197,
  [7767] = 7243,
  [7768] = 7232,
  [7769] = 7320,
  [7770] = 4242,
  [7771] = 7200,
  [7772] = 7772,
  [7773] = 7184,
  [7774] = 7205,
  [7775] = 7775,
  [7776] = 7221,
  [7777] = 7201,
  [7778] = 7192,
  [7779] = 7408,
  [7780] = 7185,
  [7781] = 7238,
  [7782] = 7213,
  [7783] = 7783,
  [7784] = 7784,
  [7785] = 7185,
  [7786] = 7371,
  [7787] = 7351,
  [7788] = 7185,
  [7789] = 7394,
  [7790] = 7790,
  [7791] = 7206,
  [7792] = 7401,
  [7793] = 7356,
  [7794] = 7794,
  [7795] = 7222,
  [7796] = 7270,
  [7797] = 7321,
  [7798] = 7209,
  [7799] = 7244,
  [7800] = 7233,
  [7801] = 7214,
  [7802] = 7226,
  [7803] = 7369,
  [7804] = 7265,
  [7805] = 7225,
  [7806] = 7556,
  [7807] = 7322,
  [7808] = 7808,
  [7809] = 7809,
  [7810] = 7528,
  [7811] = 7551,
  [7812] = 7258,
  [7813] = 7203,
  [7814] = 7351,
  [7815] = 7184,
  [7816] = 7264,
  [7817] = 7322,
  [7818] = 7198,
  [7819] = 7264,
  [7820] = 7187,
  [7821] = 7821,
  [7822] = 7822,
  [7823] = 7823,
  [7824] = 7214,
  [7825] = 7187,
  [7826] = 7826,
  [7827] = 7212,
  [7828] = 7214,
  [7829] = 7218,
  [7830] = 7740,
  [7831] = 7370,
  [7832] = 7775,
  [7833] = 7225,
  [7834] = 7733,
  [7835] = 7230,
  [7836] = 7514,
  [7837] = 7264,
  [7838] = 7198,
  [7839] = 7194,
  [7840] = 7840,
  [7841] = 7218,
  [7842] = 7258,
  [7843] = 7256,
  [7844] = 7187,
  [7845] = 7199,
  [7846] = 7410,
  [7847] = 7847,
  [7848] = 7411,
  [7849] = 7391,
  [7850] = 7219,
  [7851] = 7851,
  [7852] = 7469,
  [7853] = 7243,
  [7854] = 7221,
  [7855] = 7847,
  [7856] = 7185,
  [7857] = 7192,
  [7858] = 7363,
  [7859] = 7369,
  [7860] = 7210,
  [7861] = 7202,
  [7862] = 7371,
  [7863] = 7244,
  [7864] = 7242,
  [7865] = 7276,
  [7866] = 7557,
  [7867] = 7548,
  [7868] = 7229,
  [7869] = 7869,
  [7870] = 7232,
  [7871] = 7233,
  [7872] = 7244,
  [7873] = 7262,
  [7874] = 7310,
  [7875] = 7185,
  [7876] = 7238,
  [7877] = 7370,
  [7878] = 7214,
  [7879] = 7203,
  [7880] = 7253,
  [7881] = 7254,
  [7882] = 7255,
  [7883] = 7394,
  [7884] = 7242,
  [7885] = 7226,
  [7886] = 7264,
  [7887] = 7193,
  [7888] = 7208,
  [7889] = 7294,
  [7890] = 7224,
  [7891] = 7242,
  [7892] = 7371,
  [7893] = 7230,
  [7894] = 7310,
  [7895] = 7202,
  [7896] = 7896,
  [7897] = 7294,
  [7898] = 7898,
  [7899] = 7232,
  [7900] = 7233,
  [7901] = 7244,
  [7902] = 7187,
  [7903] = 7193,
  [7904] = 7208,
  [7905] = 7370,
  [7906] = 7338,
  [7907] = 7514,
  [7908] = 7345,
  [7909] = 7338,
  [7910] = 7351,
  [7911] = 7310,
  [7912] = 7187,
  [7913] = 7783,
  [7914] = 7394,
  [7915] = 7370,
  [7916] = 7916,
  [7917] = 7783,
  [7918] = 7918,
  [7919] = 7265,
  [7920] = 7869,
  [7921] = 7514,
  [7922] = 7922,
  [7923] = 7212,
  [7924] = 7924,
  [7925] = 4336,
  [7926] = 7926,
  [7927] = 7197,
  [7928] = 7369,
  [7929] = 7203,
  [7930] = 7214,
  [7931] = 7617,
  [7932] = 7369,
  [7933] = 7193,
  [7934] = 7256,
  [7935] = 7243,
  [7936] = 7698,
  [7937] = 7937,
  [7938] = 7322,
  [7939] = 7256,
  [7940] = 7294,
  [7941] = 7192,
  [7942] = 7239,
  [7943] = 7224,
  [7944] = 7709,
  [7945] = 7345,
  [7946] = 7320,
  [7947] = 7225,
  [7948] = 7204,
  [7949] = 7210,
  [7950] = 7401,
  [7951] = 7312,
  [7952] = 7808,
  [7953] = 7345,
  [7954] = 7320,
  [7955] = 7193,
  [7956] = 7370,
  [7957] = 7208,
  [7958] = 7351,
  [7959] = 7185,
  [7960] = 7320,
  [7961] = 7419,
  [7962] = 7371,
  [7963] = 7521,
  [7964] = 7394,
  [7965] = 7401,
  [7966] = 7191,
  [7967] = 7205,
  [7968] = 7243,
  [7969] = 7211,
  [7970] = 7192,
  [7971] = 7401,
  [7972] = 7321,
  [7973] = 7241,
  [7974] = 7256,
  [7975] = 7975,
  [7976] = 7243,
  [7977] = 7977,
  [7978] = 7208,
  [7979] = 7224,
  [7980] = 7740,
  [7981] = 7273,
  [7982] = 7826,
  [7983] = 7869,
  [7984] = 7322,
  [7985] = 7243,
  [7986] = 7198,
  [7987] = 7212,
  [7988] = 7214,
  [7989] = 7276,
  [7990] = 7764,
  [7991] = 7338,
  [7992] = 7192,
  [7993] = 7192,
  [7994] = 7403,
  [7995] = 7995,
  [7996] = 7996,
  [7997] = 7185,
  [7998] = 7370,
  [7999] = 7356,
  [8000] = 7926,
  [8001] = 7287,
  [8002] = 7809,
  [8003] = 7248,
  [8004] = 7821,
  [8005] = 7822,
  [8006] = 7823,
  [8007] = 7743,
  [8008] = 7320,
  [8009] = 7210,
  [8010] = 7187,
  [8011] = 7369,
  [8012] = 7205,
  [8013] = 7250,
  [8014] = 8014,
  [8015] = 7202,
  [8016] = 7214,
  [8017] = 7369,
  [8018] = 8018,
  [8019] = 7403,
  [8020] = 7211,
  [8021] = 7320,
  [8022] = 7740,
  [8023] = 7187,
  [8024] = 7256,
  [8025] = 7294,
  [8026] = 3229,
  [8027] = 7764,
  [8028] = 7764,
  [8029] = 7370,
  [8030] = 7251,
  [8031] = 7210,
  [8032] = 7294,
  [8033] = 7202,
  [8034] = 7356,
  [8035] = 7898,
  [8036] = 7243,
  [8037] = 7924,
  [8038] = 7821,
  [8039] = 7822,
  [8040] = 7823,
  [8041] = 7253,
  [8042] = 7244,
  [8043] = 7192,
  [8044] = 7254,
  [8045] = 7210,
  [8046] = 7707,
  [8047] = 7229,
  [8048] = 7510,
  [8049] = 7224,
  [8050] = 7202,
  [8051] = 7238,
  [8052] = 7255,
  [8053] = 7975,
  [8054] = 7356,
  [8055] = 7351,
  [8056] = 7821,
  [8057] = 7822,
  [8058] = 7823,
  [8059] = 7203,
  [8060] = 7357,
  [8061] = 7244,
  [8062] = 7719,
  [8063] = 7617,
  [8064] = 7264,
  [8065] = 7937,
  [8066] = 7242,
  [8067] = 7198,
  [8068] = 8068,
  [8069] = 7356,
  [8070] = 7239,
  [8071] = 7821,
  [8072] = 7822,
  [8073] = 7823,
  [8074] = 7320,
  [8075] = 7310,
  [8076] = 7808,
  [8077] = 7408,
  [8078] = 7224,
  [8079] = 7202,
  [8080] = 7709,
  [8081] = 7213,
  [8082] = 7270,
  [8083] = 7193,
  [8084] = 7294,
  [8085] = 7821,
  [8086] = 7822,
  [8087] = 7823,
  [8088] = 7185,
  [8089] = 7239,
  [8090] = 7214,
  [8091] = 7208,
  [8092] = 7241,
  [8093] = 7370,
  [8094] = 7338,
  [8095] = 7345,
  [8096] = 7185,
  [8097] = 7345,
  [8098] = 7371,
  [8099] = 7821,
  [8100] = 7822,
  [8101] = 7823,
  [8102] = 7244,
  [8103] = 7203,
  [8104] = 7203,
  [8105] = 7351,
  [8106] = 7185,
  [8107] = 7225,
  [8108] = 7204,
  [8109] = 7926,
  [8110] = 7394,
  [8111] = 8111,
  [8112] = 7244,
  [8113] = 7821,
  [8114] = 7822,
  [8115] = 7823,
  [8116] = 7220,
  [8117] = 8117,
  [8118] = 7401,
  [8119] = 7371,
  [8120] = 7320,
  [8121] = 7371,
  [8122] = 7230,
  [8123] = 7410,
  [8124] = 7394,
  [8125] = 7357,
  [8126] = 7248,
  [8127] = 7821,
  [8128] = 7822,
  [8129] = 7823,
  [8130] = 7210,
  [8131] = 7202,
  [8132] = 7187,
  [8133] = 7401,
  [8134] = 7370,
  [8135] = 7244,
  [8136] = 7784,
  [8137] = 7408,
  [8138] = 7193,
  [8139] = 7229,
  [8140] = 7321,
  [8141] = 7821,
  [8142] = 7822,
  [8143] = 7823,
  [8144] = 7394,
  [8145] = 7360,
  [8146] = 7206,
  [8147] = 7238,
  [8148] = 7321,
  [8149] = 7193,
  [8150] = 7203,
  [8151] = 7294,
  [8152] = 8152,
  [8153] = 7294,
  [8154] = 7208,
  [8155] = 7821,
  [8156] = 7822,
  [8157] = 7823,
  [8158] = 7591,
  [8159] = 7256,
  [8160] = 7208,
  [8161] = 7243,
  [8162] = 7192,
  [8163] = 7185,
  [8164] = 7264,
  [8165] = 7322,
  [8166] = 7345,
  [8167] = 7707,
  [8168] = 7521,
  [8169] = 7821,
  [8170] = 7822,
  [8171] = 7823,
  [8172] = 7250,
  [8173] = 7243,
  [8174] = 7270,
  [8175] = 7224,
  [8176] = 7821,
  [8177] = 7822,
  [8178] = 7823,
  [8179] = 7369,
  [8180] = 7236,
  [8181] = 7821,
  [8182] = 7822,
  [8183] = 7823,
  [8184] = 7256,
  [8185] = 7322,
  [8186] = 7821,
  [8187] = 7822,
  [8188] = 7823,
  [8189] = 7243,
  [8190] = 7198,
  [8191] = 7821,
  [8192] = 7822,
  [8193] = 7823,
  [8194] = 5001,
  [8195] = 7370,
  [8196] = 7821,
  [8197] = 7822,
  [8198] = 7823,
  [8199] = 7251,
  [8200] = 7212,
  [8201] = 7821,
  [8202] = 7822,
  [8203] = 7823,
  [8204] = 7223,
  [8205] = 7229,
  [8206] = 7821,
  [8207] = 7822,
  [8208] = 7823,
  [8209] = 7273,
  [8210] = 7253,
  [8211] = 7254,
  [8212] = 7255,
  [8213] = 7371,
  [8214] = 7214,
  [8215] = 7918,
  [8216] = 7229,
  [8217] = 7223,
  [8218] = 7394,
  [8219] = 7401,
  [8220] = 7526,
  [8221] = 7320,
  [8222] = 3245,
  [8223] = 7224,
  [8224] = 7808,
  [8225] = 7743,
  [8226] = 7410,
  [8227] = 7194,
  [8228] = 7975,
  [8229] = 7242,
  [8230] = 7199,
  [8231] = 7185,
  [8232] = 7371,
  [8233] = 7394,
  [8234] = 7401,
  [8235] = 7256,
  [8236] = 7243,
  [8237] = 7203,
  [8238] = 8238,
  [8239] = 7192,
  [8240] = 7419,
  [8241] = 7310,
  [8242] = 7401,
  [8243] = 8238,
  [8244] = 7321,
  [8245] = 7187,
  [8246] = 8246,
  [8247] = 8247,
  [8248] = 7236,
  [8249] = 7851,
  [8250] = 7193,
  [8251] = 7208,
  [8252] = 7192,
  [8253] = 7351,
  [8254] = 7338,
  [8255] = 7321,
  [8256] = 7185,
  [8257] = 7256,
  [8258] = 8246,
  [8259] = 7243,
  [8260] = 7243,
  [8261] = 7707,
  [8262] = 7320,
  [8263] = 7840,
  [8264] = 7719,
  [8265] = 7192,
  [8266] = 7918,
  [8267] = 7322,
  [8268] = 7617,
  [8269] = 7198,
  [8270] = 7212,
  [8271] = 7918,
  [8272] = 7809,
  [8273] = 7345,
  [8274] = 7918,
  [8275] = 7918,
  [8276] = 7918,
  [8277] = 7918,
  [8278] = 7918,
  [8279] = 7918,
  [8280] = 7918,
  [8281] = 7918,
  [8282] = 7918,
  [8283] = 7918,
  [8284] = 7918,
  [8285] = 7918,
  [8286] = 7918,
  [8287] = 7918,
  [8288] = 7918,
  [8289] = 7918,
  [8290] = 7322,
  [8291] = 7214,
  [8292] = 7790,
  [8293] = 7521,
  [8294] = 7200,
  [8295] = 7775,
  [8296] = 7222,
  [8297] = 7211,
  [8298] = 7225,
  [8299] = 7410,
  [8300] = 7187,
  [8301] = 7198,
  [8302] = 7808,
  [8303] = 7321,
  [8304] = 7743,
  [8305] = 7192,
  [8306] = 7351,
  [8307] = 7248,
  [8308] = 7453,
  [8309] = 7212,
  [8310] = 7294,
  [8311] = 7772,
  [8312] = 7214,
  [8313] = 7435,
  [8314] = 7225,
  [8315] = 7210,
  [8316] = 7469,
  [8317] = 7353,
  [8318] = 7195,
  [8319] = 7509,
  [8320] = 7202,
  [8321] = 8321,
  [8322] = 7244,
  [8323] = 7250,
  [8324] = 7204,
  [8325] = 7937,
  [8326] = 7321,
  [8327] = 7743,
  [8328] = 7238,
  [8329] = 7410,
  [8330] = 7193,
  [8331] = 7937,
  [8332] = 7226,
  [8333] = 7203,
  [8334] = 7709,
  [8335] = 7208,
  [8336] = 7185,
  [8337] = 7243,
  [8338] = 7214,
  [8339] = 7216,
  [8340] = 7370,
  [8341] = 7821,
  [8342] = 7822,
  [8343] = 7264,
  [8344] = 7230,
  [8345] = 7823,
  [8346] = 7212,
  [8347] = 7242,
  [8348] = 7556,
  [8349] = 7570,
  [8350] = 7214,
  [8351] = 7322,
  [8352] = 7790,
  [8353] = 3244,
  [8354] = 8354,
  [8355] = 8355,
};

static const TSSymbol ts_supertype_symbols[SUPERTYPE_COUNT] = {
  sym__binding_pattern,
  sym__binding_pattern_no_exn,
  sym__class_expression,
  sym__class_field,
  sym__class_field_specification,
  sym__class_type,
  sym__constant,
  sym__delimited_binding_pattern,
  sym__delimited_expression,
  sym__delimited_pattern,
  sym__delimited_type,
  sym__effect_binding_pattern,
  sym__effect_pattern,
  sym__expression,
  sym__infix_operator,
  sym__local_structure_item,
  sym__module_expression,
  sym__module_type,
  sym__pattern,
  sym__polymorphic_type,
  sym__sequence_expression,
  sym__signature_item,
  sym__signed_constant,
  sym__simple_binding_pattern,
  sym__simple_class_expression,
  sym__simple_class_type,
  sym__simple_expression,
  sym__simple_module_expression,
  sym__simple_pattern,
  sym__simple_type,
  sym__structure_item,
  sym__type,
};

static const TSMapSlice ts_supertype_map_slices[] = {
  [sym__binding_pattern] = {.index = 0, .length = 14},
  [sym__binding_pattern_no_exn] = {.index = 14, .length = 10},
  [sym__class_expression] = {.index = 24, .length = 5},
  [sym__class_field] = {.index = 29, .length = 7},
  [sym__class_field_specification] = {.index = 36, .length = 6},
  [sym__class_type] = {.index = 42, .length = 2},
  [sym__constant] = {.index = 44, .length = 5},
  [sym__delimited_binding_pattern] = {.index = 49, .length = 9},
  [sym__delimited_expression] = {.index = 58, .length = 7},
  [sym__delimited_pattern] = {.index = 65, .length = 5},
  [sym__delimited_type] = {.index = 70, .length = 3},
  [sym__effect_binding_pattern] = {.index = 73, .length = 7},
  [sym__effect_pattern] = {.index = 80, .length = 4},
  [sym__expression] = {.index = 84, .length = 17},
  [sym__infix_operator] = {.index = 101, .length = 8},
  [sym__local_structure_item] = {.index = 109, .length = 11},
  [sym__module_expression] = {.index = 120, .length = 5},
  [sym__module_type] = {.index = 125, .length = 8},
  [sym__pattern] = {.index = 133, .length = 8},
  [sym__polymorphic_type] = {.index = 141, .length = 2},
  [sym__sequence_expression] = {.index = 143, .length = 2},
  [sym__signature_item] = {.index = 145, .length = 13},
  [sym__signed_constant] = {.index = 158, .length = 2},
  [sym__simple_binding_pattern] = {.index = 160, .length = 14},
  [sym__simple_class_expression] = {.index = 174, .length = 7},
  [sym__simple_class_type] = {.index = 181, .length = 6},
  [sym__simple_expression] = {.index = 187, .length = 19},
  [sym__simple_module_expression] = {.index = 206, .length = 5},
  [sym__simple_pattern] = {.index = 211, .length = 12},
  [sym__simple_type] = {.index = 223, .length = 9},
  [sym__structure_item] = {.index = 232, .length = 4},
  [sym__type] = {.index = 236, .length = 5},
};

static const TSSymbol ts_supertype_map_entries[] = {
  [0] =
    sym__effect_binding_pattern,
    sym__or_binding_pattern_no_exn_anonymous,
    sym__tuple_binding_pattern_no_exn,
    sym_alias_binding_pattern,
    sym_alias_binding_pattern_no_exn,
    sym_alias_pattern,
    sym_cons_binding_pattern,
    sym_cons_binding_pattern_no_exn,
    sym_cons_pattern,
    sym_effect_binding_pattern,
    sym_effect_pattern,
    sym_exception_binding_pattern,
    sym_exception_pattern,
    sym_range_pattern,
  [14] =
    sym__effect_binding_pattern,
    sym__or_binding_pattern_no_exn_anonymous,
    sym__tuple_binding_pattern_no_exn,
    sym_alias_binding_pattern,
    sym_alias_binding_pattern_no_exn,
    sym_alias_pattern,
    sym_cons_binding_pattern,
    sym_cons_binding_pattern_no_exn,
    sym_cons_pattern,
    sym_range_pattern,
  [24] =
    sym__simple_class_expression,
    sym_class_application,
    sym_class_function,
    sym_let_class_expression,
    sym_let_open_class_expression,
  [29] =
    sym_class_initializer,
    sym_inheritance_definition,
    sym_instance_variable_definition,
    sym_item_extension,
    sym_method_definition,
    sym_quoted_item_extension,
    sym_type_parameter_constraint,
  [36] =
    sym_inheritance_specification,
    sym_instance_variable_specification,
    sym_item_extension,
    sym_method_specification,
    sym_quoted_item_extension,
    sym_type_parameter_constraint,
  [42] =
    sym__simple_class_type,
    sym_class_function_type,
  [44] =
    sym_boolean,
    sym_character,
    sym_number,
    sym_quoted_string,
    sym_string,
  [49] =
    sym_array_binding_pattern,
    sym_array_pattern,
    sym_list_binding_pattern,
    sym_list_pattern,
    sym_parenthesized_binding_pattern,
    sym_parenthesized_pattern,
    sym_record_binding_pattern,
    sym_record_pattern,
    sym_unit,
  [58] =
    sym_array_expression,
    sym_list_expression,
    sym_object_copy_expression,
    sym_package_expression,
    sym_parenthesized_expression,
    sym_record_expression,
    sym_unit,
  [65] =
    sym_array_pattern,
    sym_list_pattern,
    sym_parenthesized_pattern,
    sym_record_pattern,
    sym_unit,
  [70] =
    sym_package_type,
    sym_parenthesized_type,
    sym_polymorphic_variant_type,
  [73] =
    sym__simple_binding_pattern,
    sym_constructor_binding_pattern,
    sym_constructor_pattern,
    sym_lazy_binding_pattern,
    sym_lazy_pattern,
    sym_tag_binding_pattern,
    sym_tag_pattern,
  [80] =
    sym__simple_pattern,
    sym_constructor_pattern,
    sym_lazy_pattern,
    sym_tag_pattern,
  [84] =
    alias_sym_tuple_expression,
    sym__simple_expression,
    sym_application_expression,
    sym_assert_expression,
    sym_cons_expression,
    sym_for_expression,
    sym_fun_expression,
    sym_function_expression,
    sym_if_expression,
    sym_infix_expression,
    sym_lazy_expression,
    sym_let_expression,
    sym_match_expression,
    sym_set_expression,
    sym_sign_expression,
    sym_try_expression,
    sym_while_expression,
  [101] =
    sym_add_operator,
    sym_and_operator,
    sym_assign_operator,
    sym_concat_operator,
    sym_mult_operator,
    sym_or_operator,
    sym_pow_operator,
    sym_rel_operator,
  [109] =
    sym_class_definition,
    sym_class_type_definition,
    sym_exception_definition,
    sym_external,
    sym_floating_attribute,
    sym_item_extension,
    sym_module_definition,
    sym_module_type_definition,
    sym_open_module,
    sym_quoted_item_extension,
    sym_type_definition,
  [120] =
    sym__simple_module_expression,
    sym_functor,
    sym_module_application,
    sym_module_path,
    sym_structure,
  [125] =
    sym_extension,
    sym_functor_type,
    sym_module_type_constraint,
    sym_module_type_of,
    sym_module_type_path,
    sym_parenthesized_module_type,
    sym_quoted_extension,
    sym_signature,
  [133] =
    sym__effect_pattern,
    sym__or_binding_pattern_no_exn_anonymous,
    sym__tuple_binding_pattern_no_exn,
    sym_alias_pattern,
    sym_cons_pattern,
    sym_effect_pattern,
    sym_exception_pattern,
    sym_range_pattern,
  [141] =
    sym__type,
    sym_polymorphic_type,
  [143] =
    alias_sym_sequence_expression,
    sym__expression,
  [145] =
    sym_class_definition,
    sym_class_type_definition,
    sym_exception_definition,
    sym_external,
    sym_floating_attribute,
    sym_include_module_type,
    sym_item_extension,
    sym_module_definition,
    sym_module_type_definition,
    sym_open_module,
    sym_quoted_item_extension,
    sym_type_definition,
    sym_value_specification,
  [158] =
    sym__constant,
    sym_signed_number,
  [160] =
    alias_sym_value_name,
    sym__delimited_binding_pattern,
    sym__signed_constant,
    sym_constructor_path,
    sym_extension,
    sym_local_open_binding_pattern,
    sym_local_open_pattern,
    sym_package_pattern,
    sym_parenthesized_operator,
    sym_polymorphic_variant_pattern,
    sym_quoted_extension,
    sym_tag,
    sym_typed_binding_pattern,
    sym_typed_pattern,
  [174] =
    sym_class_path,
    sym_extension,
    sym_instantiated_class,
    sym_object_expression,
    sym_parenthesized_class_expression,
    sym_quoted_extension,
    sym_typed_class_expression,
  [181] =
    sym_class_body_type,
    sym_class_type_path,
    sym_extension,
    sym_instantiated_class_type,
    sym_let_open_class_type,
    sym_quoted_extension,
  [187] =
    sym__constant,
    sym__delimited_expression,
    sym_array_get_expression,
    sym_bigarray_get_expression,
    sym_constructor_path,
    sym_extension,
    sym_field_get_expression,
    sym_hash_expression,
    sym_local_open_expression,
    sym_method_invocation,
    sym_new_expression,
    sym_object_expression,
    sym_ocamlyacc_value,
    sym_prefix_expression,
    sym_quoted_extension,
    sym_string_get_expression,
    sym_tag,
    sym_typed_expression,
    sym_value_path,
  [206] =
    sym_extension,
    sym_packed_module,
    sym_parenthesized_module_expression,
    sym_quoted_extension,
    sym_typed_module_expression,
  [211] =
    alias_sym_value_pattern,
    sym__delimited_pattern,
    sym__signed_constant,
    sym_constructor_path,
    sym_extension,
    sym_local_open_pattern,
    sym_package_pattern,
    sym_parenthesized_operator,
    sym_polymorphic_variant_pattern,
    sym_quoted_extension,
    sym_tag,
    sym_typed_pattern,
  [223] =
    sym__delimited_type,
    sym_constructed_type,
    sym_extension,
    sym_hash_type,
    sym_local_open_type,
    sym_object_type,
    sym_quoted_extension,
    sym_type_constructor_path,
    sym_type_variable,
  [232] =
    sym__local_structure_item,
    sym_include_module,
    sym_value_definition,
    sym_value_specification,
  [236] =
    sym__labeled_tuple_type,
    sym__proper_tuple_type,
    sym__simple_type,
    sym_aliased_type,
    sym_function_type,
};

static const TSCharacterRange sym_conversion_specification_character_set_6[] = {
  {'!', '!'}, {'%', '%'}, {',', ','}, {'@', '@'}, {'B', 'C'}, {'E', 'H'}, {'L', 'L'}, {'N', 'N'},
  {'S', 'S'}, {'X', 'X'}, {'a', 'i'}, {'l', 'l'}, {'n', 'o'}, {'s', 'u'}, {'x', 'x'},
};

static const TSCharacterRange sym_conversion_specification_character_set_7[] = {
  {'!', '!'}, {'%', '%'}, {',', ','}, {'0', '9'}, {'@', '@'}, {'B', 'C'}, {'E', 'H'}, {'L', 'L'},
  {'N', 'N'}, {'S', 'S'}, {'X', 'X'}, {'a', 'i'}, {'l', 'l'}, {'n', 'o'}, {'s', 'u'}, {'x', 'x'},
};

static const TSCharacterRange sym_prefix_operator_character_set_1[] = {
  {'!', '!'}, {'#', '&'}, {'*', '+'}, {'-', '/'}, {':', ':'}, {'<', '@'}, {'^', '^'}, {'|', '|'},
  {'~', '~'},
};

static const TSCharacterRange aux_sym_pow_operator_token1_character_set_1[] = {
  {'!', '!'}, {'$', '&'}, {'*', '+'}, {'-', '/'}, {':', ':'}, {'<', '@'}, {'^', '^'}, {'|', '|'},
  {'~', '~'},
};

static const TSCharacterRange sym_indexing_operator_character_set_1[] = {
  {'!', '!'}, {'$', '&'}, {'*', '+'}, {'-', '-'}, {'/', '/'}, {':', ':'}, {'=', '@'}, {'^', '^'},
  {'|', '|'},
};

static const TSCharacterRange sym_let_operator_character_set_1[] = {
  {'$', '$'}, {'&', '&'}, {'*', '+'}, {'-', '-'}, {'/', '/'}, {'<', '>'}, {'@', '@'}, {'^', '^'},
  {'|', '|'},
};

static const TSCharacterRange sym__lowercase_identifier_character_set_2[] = {
  {'\'', '\''}, {'0', '9'}, {'A', 'Z'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xb7, 0xb7},
  {0xba, 0xba}, {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee},
  {0x300, 0x374}, {0x376, 0x377}, {0x37b, 0x37d}, {0x37f, 0x37f}, {0x386, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5},
  {0x3f7, 0x481}, {0x483, 0x487}, {0x48a, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x591, 0x5bd}, {0x5bf, 0x5bf},
  {0x5c1, 0x5c2}, {0x5c4, 0x5c5}, {0x5c7, 0x5c7}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x610, 0x61a}, {0x620, 0x669}, {0x66e, 0x6d3},
  {0x6d5, 0x6dc}, {0x6df, 0x6e8}, {0x6ea, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x74a}, {0x74d, 0x7b1}, {0x7c0, 0x7f5}, {0x7fa, 0x7fa},
  {0x7fd, 0x7fd}, {0x800, 0x82d}, {0x840, 0x85b}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x897, 0x8e1}, {0x8e3, 0x963},
  {0x966, 0x96f}, {0x971, 0x983}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9},
  {0x9bc, 0x9c4}, {0x9c7, 0x9c8}, {0x9cb, 0x9ce}, {0x9d7, 0x9d7}, {0x9dc, 0x9dd}, {0x9df, 0x9e3}, {0x9e6, 0x9f1}, {0x9fc, 0x9fc},
  {0x9fe, 0x9fe}, {0xa01, 0xa03}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36},
  {0xa38, 0xa39}, {0xa3c, 0xa3c}, {0xa3e, 0xa42}, {0xa47, 0xa48}, {0xa4b, 0xa4d}, {0xa51, 0xa51}, {0xa59, 0xa5c}, {0xa5e, 0xa5e},
  {0xa66, 0xa75}, {0xa81, 0xa83}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9},
  {0xabc, 0xac5}, {0xac7, 0xac9}, {0xacb, 0xacd}, {0xad0, 0xad0}, {0xae0, 0xae3}, {0xae6, 0xaef}, {0xaf9, 0xaff}, {0xb01, 0xb03},
  {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3c, 0xb44}, {0xb47, 0xb48},
  {0xb4b, 0xb4d}, {0xb55, 0xb57}, {0xb5c, 0xb5d}, {0xb5f, 0xb63}, {0xb66, 0xb6f}, {0xb71, 0xb71}, {0xb82, 0xb83}, {0xb85, 0xb8a},
  {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9},
  {0xbbe, 0xbc2}, {0xbc6, 0xbc8}, {0xbca, 0xbcd}, {0xbd0, 0xbd0}, {0xbd7, 0xbd7}, {0xbe6, 0xbef}, {0xc00, 0xc0c}, {0xc0e, 0xc10},
  {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3c, 0xc44}, {0xc46, 0xc48}, {0xc4a, 0xc4d}, {0xc55, 0xc56}, {0xc58, 0xc5a}, {0xc5d, 0xc5d},
  {0xc60, 0xc63}, {0xc66, 0xc6f}, {0xc80, 0xc83}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9},
  {0xcbc, 0xcc4}, {0xcc6, 0xcc8}, {0xcca, 0xccd}, {0xcd5, 0xcd6}, {0xcdd, 0xcde}, {0xce0, 0xce3}, {0xce6, 0xcef}, {0xcf1, 0xcf3},
  {0xd00, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd44}, {0xd46, 0xd48}, {0xd4a, 0xd4e}, {0xd54, 0xd57}, {0xd5f, 0xd63}, {0xd66, 0xd6f},
  {0xd7a, 0xd7f}, {0xd81, 0xd83}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xdca, 0xdca},
  {0xdcf, 0xdd4}, {0xdd6, 0xdd6}, {0xdd8, 0xddf}, {0xde6, 0xdef}, {0xdf2, 0xdf3}, {0xe01, 0xe3a}, {0xe40, 0xe4e}, {0xe50, 0xe59},
  {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6},
  {0xec8, 0xece}, {0xed0, 0xed9}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf18, 0xf19}, {0xf20, 0xf29}, {0xf35, 0xf35}, {0xf37, 0xf37},
  {0xf39, 0xf39}, {0xf3e, 0xf47}, {0xf49, 0xf6c}, {0xf71, 0xf84}, {0xf86, 0xf97}, {0xf99, 0xfbc}, {0xfc6, 0xfc6}, {0x1000, 0x1049},
  {0x1050, 0x109d}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256},
  {0x1258, 0x1258}, {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0},
  {0x12c2, 0x12c5}, {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x135d, 0x135f}, {0x1369, 0x1371}, {0x1380, 0x138f},
  {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1715},
  {0x171f, 0x1734}, {0x1740, 0x1753}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1772, 0x1773}, {0x1780, 0x17d3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dd},
  {0x17e0, 0x17e9}, {0x180b, 0x180d}, {0x180f, 0x1819}, {0x1820, 0x1878}, {0x1880, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e}, {0x1920, 0x192b},
  {0x1930, 0x193b}, {0x1946, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x19d0, 0x19da}, {0x1a00, 0x1a1b}, {0x1a20, 0x1a5e},
  {0x1a60, 0x1a7c}, {0x1a7f, 0x1a89}, {0x1a90, 0x1a99}, {0x1aa7, 0x1aa7}, {0x1ab0, 0x1abd}, {0x1abf, 0x1ace}, {0x1b00, 0x1b4c}, {0x1b50, 0x1b59},
  {0x1b6b, 0x1b73}, {0x1b80, 0x1bf3}, {0x1c00, 0x1c37}, {0x1c40, 0x1c49}, {0x1c4d, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf},
  {0x1cd0, 0x1cd2}, {0x1cd4, 0x1cfa}, {0x1d00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59},
  {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc},
  {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x200c, 0x200d}, {0x203f, 0x2040}, {0x2054, 0x2054},
  {0x2071, 0x2071}, {0x207f, 0x207f}, {0x2090, 0x209c}, {0x20d0, 0x20dc}, {0x20e1, 0x20e1}, {0x20e5, 0x20f0}, {0x2102, 0x2102}, {0x2107, 0x2107},
  {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139}, {0x213c, 0x213f},
  {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d},
  {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d7f, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6},
  {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x2de0, 0x2dff}, {0x3005, 0x3007}, {0x3021, 0x302f}, {0x3031, 0x3035}, {0x3038, 0x303c},
  {0x3041, 0x3096}, {0x3099, 0x309a}, {0x309d, 0x309f}, {0x30a1, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff},
  {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa62b}, {0xa640, 0xa66f}, {0xa674, 0xa67d}, {0xa67f, 0xa6f1},
  {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa827}, {0xa82c, 0xa82c},
  {0xa840, 0xa873}, {0xa880, 0xa8c5}, {0xa8d0, 0xa8d9}, {0xa8e0, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa92d}, {0xa930, 0xa953}, {0xa960, 0xa97c},
  {0xa980, 0xa9c0}, {0xa9cf, 0xa9d9}, {0xa9e0, 0xa9fe}, {0xaa00, 0xaa36}, {0xaa40, 0xaa4d}, {0xaa50, 0xaa59}, {0xaa60, 0xaa76}, {0xaa7a, 0xaac2},
  {0xaadb, 0xaadd}, {0xaae0, 0xaaef}, {0xaaf2, 0xaaf6}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e},
  {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabea}, {0xabec, 0xabed}, {0xabf0, 0xabf9}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb},
  {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e},
  {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfc5d}, {0xfc64, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdf9},
  {0xfe00, 0xfe0f}, {0xfe20, 0xfe2f}, {0xfe33, 0xfe34}, {0xfe4d, 0xfe4f}, {0xfe71, 0xfe71}, {0xfe73, 0xfe73}, {0xfe77, 0xfe77}, {0xfe79, 0xfe79},
  {0xfe7b, 0xfe7b}, {0xfe7d, 0xfe7d}, {0xfe7f, 0xfefc}, {0xff10, 0xff19}, {0xff21, 0xff3a}, {0xff3f, 0xff3f}, {0xff41, 0xff5a}, {0xff65, 0xffbe},
  {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d},
  {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x101fd, 0x101fd}, {0x10280, 0x1029c}, {0x102a0, 0x102d0}, {0x102e0, 0x102e0},
  {0x10300, 0x1031f}, {0x1032d, 0x1034a}, {0x10350, 0x1037a}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5}, {0x10400, 0x1049d},
  {0x104a0, 0x104a9}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592},
  {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3}, {0x10600, 0x10736}, {0x10740, 0x10755},
  {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808}, {0x1080a, 0x10835}, {0x10837, 0x10838},
  {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5}, {0x10900, 0x10915}, {0x10920, 0x10939},
  {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a03}, {0x10a05, 0x10a06}, {0x10a0c, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35}, {0x10a38, 0x10a3a},
  {0x10a3f, 0x10a3f}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae6}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55}, {0x10b60, 0x10b72},
  {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d27}, {0x10d30, 0x10d39}, {0x10d40, 0x10d65}, {0x10d69, 0x10d6d},
  {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9}, {0x10eab, 0x10eac}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10efc, 0x10f1c}, {0x10f27, 0x10f27}, {0x10f30, 0x10f50},
  {0x10f70, 0x10f85}, {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11000, 0x11046}, {0x11066, 0x11075}, {0x1107f, 0x110ba}, {0x110c2, 0x110c2}, {0x110d0, 0x110e8},
  {0x110f0, 0x110f9}, {0x11100, 0x11134}, {0x11136, 0x1113f}, {0x11144, 0x11147}, {0x11150, 0x11173}, {0x11176, 0x11176}, {0x11180, 0x111c4}, {0x111c9, 0x111cc},
  {0x111ce, 0x111da}, {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x11237}, {0x1123e, 0x11241}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d},
  {0x1128f, 0x1129d}, {0x1129f, 0x112a8}, {0x112b0, 0x112ea}, {0x112f0, 0x112f9}, {0x11300, 0x11303}, {0x11305, 0x1130c}, {0x1130f, 0x11310}, {0x11313, 0x11328},
  {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133b, 0x11344}, {0x11347, 0x11348}, {0x1134b, 0x1134d}, {0x11350, 0x11350}, {0x11357, 0x11357},
  {0x1135d, 0x11363}, {0x11366, 0x1136c}, {0x11370, 0x11374}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113c0},
  {0x113c2, 0x113c2}, {0x113c5, 0x113c5}, {0x113c7, 0x113ca}, {0x113cc, 0x113d3}, {0x113e1, 0x113e2}, {0x11400, 0x1144a}, {0x11450, 0x11459}, {0x1145e, 0x11461},
  {0x11480, 0x114c5}, {0x114c7, 0x114c7}, {0x114d0, 0x114d9}, {0x11580, 0x115b5}, {0x115b8, 0x115c0}, {0x115d8, 0x115dd}, {0x11600, 0x11640}, {0x11644, 0x11644},
  {0x11650, 0x11659}, {0x11680, 0x116b8}, {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11700, 0x1171a}, {0x1171d, 0x1172b}, {0x11730, 0x11739}, {0x11740, 0x11746},
  {0x11800, 0x1183a}, {0x118a0, 0x118e9}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x11935}, {0x11937, 0x11938},
  {0x1193b, 0x11943}, {0x11950, 0x11959}, {0x119a0, 0x119a7}, {0x119aa, 0x119d7}, {0x119da, 0x119e1}, {0x119e3, 0x119e4}, {0x11a00, 0x11a3e}, {0x11a47, 0x11a47},
  {0x11a50, 0x11a99}, {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11bf0, 0x11bf9}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c36}, {0x11c38, 0x11c40},
  {0x11c50, 0x11c59}, {0x11c72, 0x11c8f}, {0x11c92, 0x11ca7}, {0x11ca9, 0x11cb6}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d36}, {0x11d3a, 0x11d3a},
  {0x11d3c, 0x11d3d}, {0x11d3f, 0x11d47}, {0x11d50, 0x11d59}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d8e}, {0x11d90, 0x11d91}, {0x11d93, 0x11d98},
  {0x11da0, 0x11da9}, {0x11ee0, 0x11ef6}, {0x11f00, 0x11f10}, {0x11f12, 0x11f3a}, {0x11f3e, 0x11f42}, {0x11f50, 0x11f5a}, {0x11fb0, 0x11fb0}, {0x12000, 0x12399},
  {0x12400, 0x1246e}, {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13440, 0x13455}, {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x16139},
  {0x16800, 0x16a38}, {0x16a40, 0x16a5e}, {0x16a60, 0x16a69}, {0x16a70, 0x16abe}, {0x16ac0, 0x16ac9}, {0x16ad0, 0x16aed}, {0x16af0, 0x16af4}, {0x16b00, 0x16b36},
  {0x16b40, 0x16b43}, {0x16b50, 0x16b59}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16d70, 0x16d79}, {0x16e40, 0x16e7f}, {0x16f00, 0x16f4a},
  {0x16f4f, 0x16f87}, {0x16f8f, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe4}, {0x16ff0, 0x16ff1}, {0x17000, 0x187f7}, {0x18800, 0x18cd5}, {0x18cff, 0x18d08},
  {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155}, {0x1b164, 0x1b167},
  {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1bc9d, 0x1bc9e}, {0x1ccf0, 0x1ccf9}, {0x1cf00, 0x1cf2d},
  {0x1cf30, 0x1cf46}, {0x1d165, 0x1d169}, {0x1d16d, 0x1d172}, {0x1d17b, 0x1d182}, {0x1d185, 0x1d18b}, {0x1d1aa, 0x1d1ad}, {0x1d242, 0x1d244}, {0x1d400, 0x1d454},
  {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3},
  {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546},
  {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e},
  {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1d7ce, 0x1d7ff}, {0x1da00, 0x1da36}, {0x1da3b, 0x1da6c},
  {0x1da75, 0x1da75}, {0x1da84, 0x1da84}, {0x1da9b, 0x1da9f}, {0x1daa1, 0x1daaf}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e000, 0x1e006}, {0x1e008, 0x1e018},
  {0x1e01b, 0x1e021}, {0x1e023, 0x1e024}, {0x1e026, 0x1e02a}, {0x1e030, 0x1e06d}, {0x1e08f, 0x1e08f}, {0x1e100, 0x1e12c}, {0x1e130, 0x1e13d}, {0x1e140, 0x1e149},
  {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ae}, {0x1e2c0, 0x1e2f9}, {0x1e4d0, 0x1e4f9}, {0x1e5d0, 0x1e5fa}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee},
  {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e8d0, 0x1e8d6}, {0x1e900, 0x1e94b}, {0x1e950, 0x1e959}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22},
  {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47},
  {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b},
  {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c},
  {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x1fbf0, 0x1fbf9}, {0x20000, 0x2a6df},
  {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a}, {0x31350, 0x323af},
  {0xe0100, 0xe01ef},
};

static const TSCharacterRange sym__lowercase_identifier_character_set_3[] = {
  {'_', '_'}, {'a', 'z'}, {0xb5, 0xb5}, {0xdf, 0xf6}, {0xf8, 0xff}, {0x101, 0x101}, {0x103, 0x103}, {0x105, 0x105},
  {0x107, 0x107}, {0x109, 0x109}, {0x10b, 0x10b}, {0x10d, 0x10d}, {0x10f, 0x10f}, {0x111, 0x111}, {0x113, 0x113}, {0x115, 0x115},
  {0x117, 0x117}, {0x119, 0x119}, {0x11b, 0x11b}, {0x11d, 0x11d}, {0x11f, 0x11f}, {0x121, 0x121}, {0x123, 0x123}, {0x125, 0x125},
  {0x127, 0x127}, {0x129, 0x129}, {0x12b, 0x12b}, {0x12d, 0x12d}, {0x12f, 0x12f}, {0x131, 0x131}, {0x133, 0x133}, {0x135, 0x135},
  {0x137, 0x138}, {0x13a, 0x13a}, {0x13c, 0x13c}, {0x13e, 0x13e}, {0x140, 0x140}, {0x142, 0x142}, {0x144, 0x144}, {0x146, 0x146},
  {0x148, 0x149}, {0x14b, 0x14b}, {0x14d, 0x14d}, {0x14f, 0x14f}, {0x151, 0x151}, {0x153, 0x153}, {0x155, 0x155}, {0x157, 0x157},
  {0x159, 0x159}, {0x15b, 0x15b}, {0x15d, 0x15d}, {0x15f, 0x15f}, {0x161, 0x161}, {0x163, 0x163}, {0x165, 0x165}, {0x167, 0x167},
  {0x169, 0x169}, {0x16b, 0x16b}, {0x16d, 0x16d}, {0x16f, 0x16f}, {0x171, 0x171}, {0x173, 0x173}, {0x175, 0x175}, {0x177, 0x177},
  {0x17a, 0x17a}, {0x17c, 0x17c}, {0x17e, 0x180}, {0x183, 0x183}, {0x185, 0x185}, {0x188, 0x188}, {0x18c, 0x18d}, {0x192, 0x192},
  {0x195, 0x195}, {0x199, 0x19b}, {0x19e, 0x19e}, {0x1a1, 0x1a1}, {0x1a3, 0x1a3}, {0x1a5, 0x1a5}, {0x1a8, 0x1a8}, {0x1aa, 0x1ab},
  {0x1ad, 0x1ad}, {0x1b0, 0x1b0}, {0x1b4, 0x1b4}, {0x1b6, 0x1b6}, {0x1b9, 0x1ba}, {0x1bd, 0x1bf}, {0x1c6, 0x1c6}, {0x1c9, 0x1c9},
  {0x1cc, 0x1cc}, {0x1ce, 0x1ce}, {0x1d0, 0x1d0}, {0x1d2, 0x1d2}, {0x1d4, 0x1d4}, {0x1d6, 0x1d6}, {0x1d8, 0x1d8}, {0x1da, 0x1da},
  {0x1dc, 0x1dd}, {0x1df, 0x1df}, {0x1e1, 0x1e1}, {0x1e3, 0x1e3}, {0x1e5, 0x1e5}, {0x1e7, 0x1e7}, {0x1e9, 0x1e9}, {0x1eb, 0x1eb},
  {0x1ed, 0x1ed}, {0x1ef, 0x1f0}, {0x1f3, 0x1f3}, {0x1f5, 0x1f5}, {0x1f9, 0x1f9}, {0x1fb, 0x1fb}, {0x1fd, 0x1fd}, {0x1ff, 0x1ff},
  {0x201, 0x201}, {0x203, 0x203}, {0x205, 0x205}, {0x207, 0x207}, {0x209, 0x209}, {0x20b, 0x20b}, {0x20d, 0x20d}, {0x20f, 0x20f},
  {0x211, 0x211}, {0x213, 0x213}, {0x215, 0x215}, {0x217, 0x217}, {0x219, 0x219}, {0x21b, 0x21b}, {0x21d, 0x21d}, {0x21f, 0x21f},
  {0x221, 0x221}, {0x223, 0x223}, {0x225, 0x225}, {0x227, 0x227}, {0x229, 0x229}, {0x22b, 0x22b}, {0x22d, 0x22d}, {0x22f, 0x22f},
  {0x231, 0x231}, {0x233, 0x239}, {0x23c, 0x23c}, {0x23f, 0x240}, {0x242, 0x242}, {0x247, 0x247}, {0x249, 0x249}, {0x24b, 0x24b},
  {0x24d, 0x24d}, {0x24f, 0x293}, {0x295, 0x2af}, {0x371, 0x371}, {0x373, 0x373}, {0x377, 0x377}, {0x37b, 0x37d}, {0x390, 0x390},
  {0x3ac, 0x3ce}, {0x3d0, 0x3d1}, {0x3d5, 0x3d7}, {0x3d9, 0x3d9}, {0x3db, 0x3db}, {0x3dd, 0x3dd}, {0x3df, 0x3df}, {0x3e1, 0x3e1},
  {0x3e3, 0x3e3}, {0x3e5, 0x3e5}, {0x3e7, 0x3e7}, {0x3e9, 0x3e9}, {0x3eb, 0x3eb}, {0x3ed, 0x3ed}, {0x3ef, 0x3f3}, {0x3f5, 0x3f5},
  {0x3f8, 0x3f8}, {0x3fb, 0x3fc}, {0x430, 0x45f}, {0x461, 0x461}, {0x463, 0x463}, {0x465, 0x465}, {0x467, 0x467}, {0x469, 0x469},
  {0x46b, 0x46b}, {0x46d, 0x46d}, {0x46f, 0x46f}, {0x471, 0x471}, {0x473, 0x473}, {0x475, 0x475}, {0x477, 0x477}, {0x479, 0x479},
  {0x47b, 0x47b}, {0x47d, 0x47d}, {0x47f, 0x47f}, {0x481, 0x481}, {0x48b, 0x48b}, {0x48d, 0x48d}, {0x48f, 0x48f}, {0x491, 0x491},
  {0x493, 0x493}, {0x495, 0x495}, {0x497, 0x497}, {0x499, 0x499}, {0x49b, 0x49b}, {0x49d, 0x49d}, {0x49f, 0x49f}, {0x4a1, 0x4a1},
  {0x4a3, 0x4a3}, {0x4a5, 0x4a5}, {0x4a7, 0x4a7}, {0x4a9, 0x4a9}, {0x4ab, 0x4ab}, {0x4ad, 0x4ad}, {0x4af, 0x4af}, {0x4b1, 0x4b1},
  {0x4b3, 0x4b3}, {0x4b5, 0x4b5}, {0x4b7, 0x4b7}, {0x4b9, 0x4b9}, {0x4bb, 0x4bb}, {0x4bd, 0x4bd}, {0x4bf, 0x4bf}, {0x4c2, 0x4c2},
  {0x4c4, 0x4c4}, {0x4c6, 0x4c6}, {0x4c8, 0x4c8}, {0x4ca, 0x4ca}, {0x4cc, 0x4cc}, {0x4ce, 0x4cf}, {0x4d1, 0x4d1}, {0x4d3, 0x4d3},
  {0x4d5, 0x4d5}, {0x4d7, 0x4d7}, {0x4d9, 0x4d9}, {0x4db, 0x4db}, {0x4dd, 0x4dd}, {0x4df, 0x4df}, {0x4e1, 0x4e1}, {0x4e3, 0x4e3},
  {0x4e5, 0x4e5}, {0x4e7, 0x4e7}, {0x4e9, 0x4e9}, {0x4eb, 0x4eb}, {0x4ed, 0x4ed}, {0x4ef, 0x4ef}, {0x4f1, 0x4f1}, {0x4f3, 0x4f3},
  {0x4f5, 0x4f5}, {0x4f7, 0x4f7}, {0x4f9, 0x4f9}, {0x4fb, 0x4fb}, {0x4fd, 0x4fd}, {0x4ff, 0x4ff}, {0x501, 0x501}, {0x503, 0x503},
  {0x505, 0x505}, {0x507, 0x507}, {0x509, 0x509}, {0x50b, 0x50b}, {0x50d, 0x50d}, {0x50f, 0x50f}, {0x511, 0x511}, {0x513, 0x513},
  {0x515, 0x515}, {0x517, 0x517}, {0x519, 0x519}, {0x51b, 0x51b}, {0x51d, 0x51d}, {0x51f, 0x51f}, {0x521, 0x521}, {0x523, 0x523},
  {0x525, 0x525}, {0x527, 0x527}, {0x529, 0x529}, {0x52b, 0x52b}, {0x52d, 0x52d}, {0x52f, 0x52f}, {0x560, 0x588}, {0x10d0, 0x10fa},
  {0x10fd, 0x10ff}, {0x13f8, 0x13fd}, {0x1c80, 0x1c88}, {0x1c8a, 0x1c8a}, {0x1d00, 0x1d2b}, {0x1d6b, 0x1d77}, {0x1d79, 0x1d9a}, {0x1e01, 0x1e01},
  {0x1e03, 0x1e03}, {0x1e05, 0x1e05}, {0x1e07, 0x1e07}, {0x1e09, 0x1e09}, {0x1e0b, 0x1e0b}, {0x1e0d, 0x1e0d}, {0x1e0f, 0x1e0f}, {0x1e11, 0x1e11},
  {0x1e13, 0x1e13}, {0x1e15, 0x1e15}, {0x1e17, 0x1e17}, {0x1e19, 0x1e19}, {0x1e1b, 0x1e1b}, {0x1e1d, 0x1e1d}, {0x1e1f, 0x1e1f}, {0x1e21, 0x1e21},
  {0x1e23, 0x1e23}, {0x1e25, 0x1e25}, {0x1e27, 0x1e27}, {0x1e29, 0x1e29}, {0x1e2b, 0x1e2b}, {0x1e2d, 0x1e2d}, {0x1e2f, 0x1e2f}, {0x1e31, 0x1e31},
  {0x1e33, 0x1e33}, {0x1e35, 0x1e35}, {0x1e37, 0x1e37}, {0x1e39, 0x1e39}, {0x1e3b, 0x1e3b}, {0x1e3d, 0x1e3d}, {0x1e3f, 0x1e3f}, {0x1e41, 0x1e41},
  {0x1e43, 0x1e43}, {0x1e45, 0x1e45}, {0x1e47, 0x1e47}, {0x1e49, 0x1e49}, {0x1e4b, 0x1e4b}, {0x1e4d, 0x1e4d}, {0x1e4f, 0x1e4f}, {0x1e51, 0x1e51},
  {0x1e53, 0x1e53}, {0x1e55, 0x1e55}, {0x1e57, 0x1e57}, {0x1e59, 0x1e59}, {0x1e5b, 0x1e5b}, {0x1e5d, 0x1e5d}, {0x1e5f, 0x1e5f}, {0x1e61, 0x1e61},
  {0x1e63, 0x1e63}, {0x1e65, 0x1e65}, {0x1e67, 0x1e67}, {0x1e69, 0x1e69}, {0x1e6b, 0x1e6b}, {0x1e6d, 0x1e6d}, {0x1e6f, 0x1e6f}, {0x1e71, 0x1e71},
  {0x1e73, 0x1e73}, {0x1e75, 0x1e75}, {0x1e77, 0x1e77}, {0x1e79, 0x1e79}, {0x1e7b, 0x1e7b}, {0x1e7d, 0x1e7d}, {0x1e7f, 0x1e7f}, {0x1e81, 0x1e81},
  {0x1e83, 0x1e83}, {0x1e85, 0x1e85}, {0x1e87, 0x1e87}, {0x1e89, 0x1e89}, {0x1e8b, 0x1e8b}, {0x1e8d, 0x1e8d}, {0x1e8f, 0x1e8f}, {0x1e91, 0x1e91},
  {0x1e93, 0x1e93}, {0x1e95, 0x1e9d}, {0x1e9f, 0x1e9f}, {0x1ea1, 0x1ea1}, {0x1ea3, 0x1ea3}, {0x1ea5, 0x1ea5}, {0x1ea7, 0x1ea7}, {0x1ea9, 0x1ea9},
  {0x1eab, 0x1eab}, {0x1ead, 0x1ead}, {0x1eaf, 0x1eaf}, {0x1eb1, 0x1eb1}, {0x1eb3, 0x1eb3}, {0x1eb5, 0x1eb5}, {0x1eb7, 0x1eb7}, {0x1eb9, 0x1eb9},
  {0x1ebb, 0x1ebb}, {0x1ebd, 0x1ebd}, {0x1ebf, 0x1ebf}, {0x1ec1, 0x1ec1}, {0x1ec3, 0x1ec3}, {0x1ec5, 0x1ec5}, {0x1ec7, 0x1ec7}, {0x1ec9, 0x1ec9},
  {0x1ecb, 0x1ecb}, {0x1ecd, 0x1ecd}, {0x1ecf, 0x1ecf}, {0x1ed1, 0x1ed1}, {0x1ed3, 0x1ed3}, {0x1ed5, 0x1ed5}, {0x1ed7, 0x1ed7}, {0x1ed9, 0x1ed9},
  {0x1edb, 0x1edb}, {0x1edd, 0x1edd}, {0x1edf, 0x1edf}, {0x1ee1, 0x1ee1}, {0x1ee3, 0x1ee3}, {0x1ee5, 0x1ee5}, {0x1ee7, 0x1ee7}, {0x1ee9, 0x1ee9},
  {0x1eeb, 0x1eeb}, {0x1eed, 0x1eed}, {0x1eef, 0x1eef}, {0x1ef1, 0x1ef1}, {0x1ef3, 0x1ef3}, {0x1ef5, 0x1ef5}, {0x1ef7, 0x1ef7}, {0x1ef9, 0x1ef9},
  {0x1efb, 0x1efb}, {0x1efd, 0x1efd}, {0x1eff, 0x1f07}, {0x1f10, 0x1f15}, {0x1f20, 0x1f27}, {0x1f30, 0x1f37}, {0x1f40, 0x1f45}, {0x1f50, 0x1f57},
  {0x1f60, 0x1f67}, {0x1f70, 0x1f7d}, {0x1f80, 0x1f87}, {0x1f90, 0x1f97}, {0x1fa0, 0x1fa7}, {0x1fb0, 0x1fb4}, {0x1fb6, 0x1fb7}, {0x1fbe, 0x1fbe},
  {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fc7}, {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fd7}, {0x1fe0, 0x1fe7}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ff7}, {0x210a, 0x210a},
  {0x210e, 0x210f}, {0x2113, 0x2113}, {0x212f, 0x212f}, {0x2134, 0x2134}, {0x2139, 0x2139}, {0x213c, 0x213d}, {0x2146, 0x2149}, {0x214e, 0x214e},
  {0x2184, 0x2184}, {0x2c30, 0x2c5f}, {0x2c61, 0x2c61}, {0x2c65, 0x2c66}, {0x2c68, 0x2c68}, {0x2c6a, 0x2c6a}, {0x2c6c, 0x2c6c}, {0x2c71, 0x2c71},
  {0x2c73, 0x2c74}, {0x2c76, 0x2c7b}, {0x2c81, 0x2c81}, {0x2c83, 0x2c83}, {0x2c85, 0x2c85}, {0x2c87, 0x2c87}, {0x2c89, 0x2c89}, {0x2c8b, 0x2c8b},
  {0x2c8d, 0x2c8d}, {0x2c8f, 0x2c8f}, {0x2c91, 0x2c91}, {0x2c93, 0x2c93}, {0x2c95, 0x2c95}, {0x2c97, 0x2c97}, {0x2c99, 0x2c99}, {0x2c9b, 0x2c9b},
  {0x2c9d, 0x2c9d}, {0x2c9f, 0x2c9f}, {0x2ca1, 0x2ca1}, {0x2ca3, 0x2ca3}, {0x2ca5, 0x2ca5}, {0x2ca7, 0x2ca7}, {0x2ca9, 0x2ca9}, {0x2cab, 0x2cab},
  {0x2cad, 0x2cad}, {0x2caf, 0x2caf}, {0x2cb1, 0x2cb1}, {0x2cb3, 0x2cb3}, {0x2cb5, 0x2cb5}, {0x2cb7, 0x2cb7}, {0x2cb9, 0x2cb9}, {0x2cbb, 0x2cbb},
  {0x2cbd, 0x2cbd}, {0x2cbf, 0x2cbf}, {0x2cc1, 0x2cc1}, {0x2cc3, 0x2cc3}, {0x2cc5, 0x2cc5}, {0x2cc7, 0x2cc7}, {0x2cc9, 0x2cc9}, {0x2ccb, 0x2ccb},
  {0x2ccd, 0x2ccd}, {0x2ccf, 0x2ccf}, {0x2cd1, 0x2cd1}, {0x2cd3, 0x2cd3}, {0x2cd5, 0x2cd5}, {0x2cd7, 0x2cd7}, {0x2cd9, 0x2cd9}, {0x2cdb, 0x2cdb},
  {0x2cdd, 0x2cdd}, {0x2cdf, 0x2cdf}, {0x2ce1, 0x2ce1}, {0x2ce3, 0x2ce4}, {0x2cec, 0x2cec}, {0x2cee, 0x2cee}, {0x2cf3, 0x2cf3}, {0x2d00, 0x2d25},
  {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d}, {0xa641, 0xa641}, {0xa643, 0xa643}, {0xa645, 0xa645}, {0xa647, 0xa647}, {0xa649, 0xa649}, {0xa64b, 0xa64b},
  {0xa64d, 0xa64d}, {0xa64f, 0xa64f}, {0xa651, 0xa651}, {0xa653, 0xa653}, {0xa655, 0xa655}, {0xa657, 0xa657}, {0xa659, 0xa659}, {0xa65b, 0xa65b},
  {0xa65d, 0xa65d}, {0xa65f, 0xa65f}, {0xa661, 0xa661}, {0xa663, 0xa663}, {0xa665, 0xa665}, {0xa667, 0xa667}, {0xa669, 0xa669}, {0xa66b, 0xa66b},
  {0xa66d, 0xa66d}, {0xa681, 0xa681}, {0xa683, 0xa683}, {0xa685, 0xa685}, {0xa687, 0xa687}, {0xa689, 0xa689}, {0xa68b, 0xa68b}, {0xa68d, 0xa68d},
  {0xa68f, 0xa68f}, {0xa691, 0xa691}, {0xa693, 0xa693}, {0xa695, 0xa695}, {0xa697, 0xa697}, {0xa699, 0xa699}, {0xa69b, 0xa69b}, {0xa723, 0xa723},
  {0xa725, 0xa725}, {0xa727, 0xa727}, {0xa729, 0xa729}, {0xa72b, 0xa72b}, {0xa72d, 0xa72d}, {0xa72f, 0xa731}, {0xa733, 0xa733}, {0xa735, 0xa735},
  {0xa737, 0xa737}, {0xa739, 0xa739}, {0xa73b, 0xa73b}, {0xa73d, 0xa73d}, {0xa73f, 0xa73f}, {0xa741, 0xa741}, {0xa743, 0xa743}, {0xa745, 0xa745},
  {0xa747, 0xa747}, {0xa749, 0xa749}, {0xa74b, 0xa74b}, {0xa74d, 0xa74d}, {0xa74f, 0xa74f}, {0xa751, 0xa751}, {0xa753, 0xa753}, {0xa755, 0xa755},
  {0xa757, 0xa757}, {0xa759, 0xa759}, {0xa75b, 0xa75b}, {0xa75d, 0xa75d}, {0xa75f, 0xa75f}, {0xa761, 0xa761}, {0xa763, 0xa763}, {0xa765, 0xa765},
  {0xa767, 0xa767}, {0xa769, 0xa769}, {0xa76b, 0xa76b}, {0xa76d, 0xa76d}, {0xa76f, 0xa76f}, {0xa771, 0xa778}, {0xa77a, 0xa77a}, {0xa77c, 0xa77c},
  {0xa77f, 0xa77f}, {0xa781, 0xa781}, {0xa783, 0xa783}, {0xa785, 0xa785}, {0xa787, 0xa787}, {0xa78c, 0xa78c}, {0xa78e, 0xa78e}, {0xa791, 0xa791},
  {0xa793, 0xa795}, {0xa797, 0xa797}, {0xa799, 0xa799}, {0xa79b, 0xa79b}, {0xa79d, 0xa79d}, {0xa79f, 0xa79f}, {0xa7a1, 0xa7a1}, {0xa7a3, 0xa7a3},
  {0xa7a5, 0xa7a5}, {0xa7a7, 0xa7a7}, {0xa7a9, 0xa7a9}, {0xa7af, 0xa7af}, {0xa7b5, 0xa7b5}, {0xa7b7, 0xa7b7}, {0xa7b9, 0xa7b9}, {0xa7bb, 0xa7bb},
  {0xa7bd, 0xa7bd}, {0xa7bf, 0xa7bf}, {0xa7c1, 0xa7c1}, {0xa7c3, 0xa7c3}, {0xa7c8, 0xa7c8}, {0xa7ca, 0xa7ca}, {0xa7cd, 0xa7cd}, {0xa7d1, 0xa7d1},
  {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7d5}, {0xa7d7, 0xa7d7}, {0xa7d9, 0xa7d9}, {0xa7db, 0xa7db}, {0xa7f6, 0xa7f6}, {0xa7fa, 0xa7fa}, {0xab30, 0xab5a},
  {0xab60, 0xab68}, {0xab70, 0xabbf}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xff41, 0xff5a}, {0x10428, 0x1044f}, {0x104d8, 0x104fb}, {0x10597, 0x105a1},
  {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x10cc0, 0x10cf2}, {0x10d70, 0x10d85}, {0x118c0, 0x118df}, {0x16e60, 0x16e7f}, {0x1d41a, 0x1d433},
  {0x1d44e, 0x1d454}, {0x1d456, 0x1d467}, {0x1d482, 0x1d49b}, {0x1d4b6, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d4cf}, {0x1d4ea, 0x1d503},
  {0x1d51e, 0x1d537}, {0x1d552, 0x1d56b}, {0x1d586, 0x1d59f}, {0x1d5ba, 0x1d5d3}, {0x1d5ee, 0x1d607}, {0x1d622, 0x1d63b}, {0x1d656, 0x1d66f}, {0x1d68a, 0x1d6a5},
  {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6e1}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d71b}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d755}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d78f},
  {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7c9}, {0x1d7cb, 0x1d7cb}, {0x1df00, 0x1df09}, {0x1df0b, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e922, 0x1e943},
};

static const TSCharacterRange sym__uppercase_identifier_character_set_1[] = {
  {'A', 'Z'}, {0xc0, 0xd6}, {0xd8, 0xde}, {0x100, 0x100}, {0x102, 0x102}, {0x104, 0x104}, {0x106, 0x106}, {0x108, 0x108},
  {0x10a, 0x10a}, {0x10c, 0x10c}, {0x10e, 0x10e}, {0x110, 0x110}, {0x112, 0x112}, {0x114, 0x114}, {0x116, 0x116}, {0x118, 0x118},
  {0x11a, 0x11a}, {0x11c, 0x11c}, {0x11e, 0x11e}, {0x120, 0x120}, {0x122, 0x122}, {0x124, 0x124}, {0x126, 0x126}, {0x128, 0x128},
  {0x12a, 0x12a}, {0x12c, 0x12c}, {0x12e, 0x12e}, {0x130, 0x130}, {0x132, 0x132}, {0x134, 0x134}, {0x136, 0x136}, {0x139, 0x139},
  {0x13b, 0x13b}, {0x13d, 0x13d}, {0x13f, 0x13f}, {0x141, 0x141}, {0x143, 0x143}, {0x145, 0x145}, {0x147, 0x147}, {0x14a, 0x14a},
  {0x14c, 0x14c}, {0x14e, 0x14e}, {0x150, 0x150}, {0x152, 0x152}, {0x154, 0x154}, {0x156, 0x156}, {0x158, 0x158}, {0x15a, 0x15a},
  {0x15c, 0x15c}, {0x15e, 0x15e}, {0x160, 0x160}, {0x162, 0x162}, {0x164, 0x164}, {0x166, 0x166}, {0x168, 0x168}, {0x16a, 0x16a},
  {0x16c, 0x16c}, {0x16e, 0x16e}, {0x170, 0x170}, {0x172, 0x172}, {0x174, 0x174}, {0x176, 0x176}, {0x178, 0x179}, {0x17b, 0x17b},
  {0x17d, 0x17d}, {0x181, 0x182}, {0x184, 0x184}, {0x186, 0x187}, {0x189, 0x18b}, {0x18e, 0x191}, {0x193, 0x194}, {0x196, 0x198},
  {0x19c, 0x19d}, {0x19f, 0x1a0}, {0x1a2, 0x1a2}, {0x1a4, 0x1a4}, {0x1a6, 0x1a7}, {0x1a9, 0x1a9}, {0x1ac, 0x1ac}, {0x1ae, 0x1af},
  {0x1b1, 0x1b3}, {0x1b5, 0x1b5}, {0x1b7, 0x1b8}, {0x1bc, 0x1bc}, {0x1c4, 0x1c4}, {0x1c7, 0x1c7}, {0x1ca, 0x1ca}, {0x1cd, 0x1cd},
  {0x1cf, 0x1cf}, {0x1d1, 0x1d1}, {0x1d3, 0x1d3}, {0x1d5, 0x1d5}, {0x1d7, 0x1d7}, {0x1d9, 0x1d9}, {0x1db, 0x1db}, {0x1de, 0x1de},
  {0x1e0, 0x1e0}, {0x1e2, 0x1e2}, {0x1e4, 0x1e4}, {0x1e6, 0x1e6}, {0x1e8, 0x1e8}, {0x1ea, 0x1ea}, {0x1ec, 0x1ec}, {0x1ee, 0x1ee},
  {0x1f1, 0x1f1}, {0x1f4, 0x1f4}, {0x1f6, 0x1f8}, {0x1fa, 0x1fa}, {0x1fc, 0x1fc}, {0x1fe, 0x1fe}, {0x200, 0x200}, {0x202, 0x202},
  {0x204, 0x204}, {0x206, 0x206}, {0x208, 0x208}, {0x20a, 0x20a}, {0x20c, 0x20c}, {0x20e, 0x20e}, {0x210, 0x210}, {0x212, 0x212},
  {0x214, 0x214}, {0x216, 0x216}, {0x218, 0x218}, {0x21a, 0x21a}, {0x21c, 0x21c}, {0x21e, 0x21e}, {0x220, 0x220}, {0x222, 0x222},
  {0x224, 0x224}, {0x226, 0x226}, {0x228, 0x228}, {0x22a, 0x22a}, {0x22c, 0x22c}, {0x22e, 0x22e}, {0x230, 0x230}, {0x232, 0x232},
  {0x23a, 0x23b}, {0x23d, 0x23e}, {0x241, 0x241}, {0x243, 0x246}, {0x248, 0x248}, {0x24a, 0x24a}, {0x24c, 0x24c}, {0x24e, 0x24e},
  {0x370, 0x370}, {0x372, 0x372}, {0x376, 0x376}, {0x37f, 0x37f}, {0x386, 0x386}, {0x388, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x38f},
  {0x391, 0x3a1}, {0x3a3, 0x3ab}, {0x3cf, 0x3cf}, {0x3d2, 0x3d4}, {0x3d8, 0x3d8}, {0x3da, 0x3da}, {0x3dc, 0x3dc}, {0x3de, 0x3de},
  {0x3e0, 0x3e0}, {0x3e2, 0x3e2}, {0x3e4, 0x3e4}, {0x3e6, 0x3e6}, {0x3e8, 0x3e8}, {0x3ea, 0x3ea}, {0x3ec, 0x3ec}, {0x3ee, 0x3ee},
  {0x3f4, 0x3f4}, {0x3f7, 0x3f7}, {0x3f9, 0x3fa}, {0x3fd, 0x42f}, {0x460, 0x460}, {0x462, 0x462}, {0x464, 0x464}, {0x466, 0x466},
  {0x468, 0x468}, {0x46a, 0x46a}, {0x46c, 0x46c}, {0x46e, 0x46e}, {0x470, 0x470}, {0x472, 0x472}, {0x474, 0x474}, {0x476, 0x476},
  {0x478, 0x478}, {0x47a, 0x47a}, {0x47c, 0x47c}, {0x47e, 0x47e}, {0x480, 0x480}, {0x48a, 0x48a}, {0x48c, 0x48c}, {0x48e, 0x48e},
  {0x490, 0x490}, {0x492, 0x492}, {0x494, 0x494}, {0x496, 0x496}, {0x498, 0x498}, {0x49a, 0x49a}, {0x49c, 0x49c}, {0x49e, 0x49e},
  {0x4a0, 0x4a0}, {0x4a2, 0x4a2}, {0x4a4, 0x4a4}, {0x4a6, 0x4a6}, {0x4a8, 0x4a8}, {0x4aa, 0x4aa}, {0x4ac, 0x4ac}, {0x4ae, 0x4ae},
  {0x4b0, 0x4b0}, {0x4b2, 0x4b2}, {0x4b4, 0x4b4}, {0x4b6, 0x4b6}, {0x4b8, 0x4b8}, {0x4ba, 0x4ba}, {0x4bc, 0x4bc}, {0x4be, 0x4be},
  {0x4c0, 0x4c1}, {0x4c3, 0x4c3}, {0x4c5, 0x4c5}, {0x4c7, 0x4c7}, {0x4c9, 0x4c9}, {0x4cb, 0x4cb}, {0x4cd, 0x4cd}, {0x4d0, 0x4d0},
  {0x4d2, 0x4d2}, {0x4d4, 0x4d4}, {0x4d6, 0x4d6}, {0x4d8, 0x4d8}, {0x4da, 0x4da}, {0x4dc, 0x4dc}, {0x4de, 0x4de}, {0x4e0, 0x4e0},
  {0x4e2, 0x4e2}, {0x4e4, 0x4e4}, {0x4e6, 0x4e6}, {0x4e8, 0x4e8}, {0x4ea, 0x4ea}, {0x4ec, 0x4ec}, {0x4ee, 0x4ee}, {0x4f0, 0x4f0},
  {0x4f2, 0x4f2}, {0x4f4, 0x4f4}, {0x4f6, 0x4f6}, {0x4f8, 0x4f8}, {0x4fa, 0x4fa}, {0x4fc, 0x4fc}, {0x4fe, 0x4fe}, {0x500, 0x500},
  {0x502, 0x502}, {0x504, 0x504}, {0x506, 0x506}, {0x508, 0x508}, {0x50a, 0x50a}, {0x50c, 0x50c}, {0x50e, 0x50e}, {0x510, 0x510},
  {0x512, 0x512}, {0x514, 0x514}, {0x516, 0x516}, {0x518, 0x518}, {0x51a, 0x51a}, {0x51c, 0x51c}, {0x51e, 0x51e}, {0x520, 0x520},
  {0x522, 0x522}, {0x524, 0x524}, {0x526, 0x526}, {0x528, 0x528}, {0x52a, 0x52a}, {0x52c, 0x52c}, {0x52e, 0x52e}, {0x531, 0x556},
  {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x13a0, 0x13f5}, {0x1c89, 0x1c89}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1e00, 0x1e00},
  {0x1e02, 0x1e02}, {0x1e04, 0x1e04}, {0x1e06, 0x1e06}, {0x1e08, 0x1e08}, {0x1e0a, 0x1e0a}, {0x1e0c, 0x1e0c}, {0x1e0e, 0x1e0e}, {0x1e10, 0x1e10},
  {0x1e12, 0x1e12}, {0x1e14, 0x1e14}, {0x1e16, 0x1e16}, {0x1e18, 0x1e18}, {0x1e1a, 0x1e1a}, {0x1e1c, 0x1e1c}, {0x1e1e, 0x1e1e}, {0x1e20, 0x1e20},
  {0x1e22, 0x1e22}, {0x1e24, 0x1e24}, {0x1e26, 0x1e26}, {0x1e28, 0x1e28}, {0x1e2a, 0x1e2a}, {0x1e2c, 0x1e2c}, {0x1e2e, 0x1e2e}, {0x1e30, 0x1e30},
  {0x1e32, 0x1e32}, {0x1e34, 0x1e34}, {0x1e36, 0x1e36}, {0x1e38, 0x1e38}, {0x1e3a, 0x1e3a}, {0x1e3c, 0x1e3c}, {0x1e3e, 0x1e3e}, {0x1e40, 0x1e40},
  {0x1e42, 0x1e42}, {0x1e44, 0x1e44}, {0x1e46, 0x1e46}, {0x1e48, 0x1e48}, {0x1e4a, 0x1e4a}, {0x1e4c, 0x1e4c}, {0x1e4e, 0x1e4e}, {0x1e50, 0x1e50},
  {0x1e52, 0x1e52}, {0x1e54, 0x1e54}, {0x1e56, 0x1e56}, {0x1e58, 0x1e58}, {0x1e5a, 0x1e5a}, {0x1e5c, 0x1e5c}, {0x1e5e, 0x1e5e}, {0x1e60, 0x1e60},
  {0x1e62, 0x1e62}, {0x1e64, 0x1e64}, {0x1e66, 0x1e66}, {0x1e68, 0x1e68}, {0x1e6a, 0x1e6a}, {0x1e6c, 0x1e6c}, {0x1e6e, 0x1e6e}, {0x1e70, 0x1e70},
  {0x1e72, 0x1e72}, {0x1e74, 0x1e74}, {0x1e76, 0x1e76}, {0x1e78, 0x1e78}, {0x1e7a, 0x1e7a}, {0x1e7c, 0x1e7c}, {0x1e7e, 0x1e7e}, {0x1e80, 0x1e80},
  {0x1e82, 0x1e82}, {0x1e84, 0x1e84}, {0x1e86, 0x1e86}, {0x1e88, 0x1e88}, {0x1e8a, 0x1e8a}, {0x1e8c, 0x1e8c}, {0x1e8e, 0x1e8e}, {0x1e90, 0x1e90},
  {0x1e92, 0x1e92}, {0x1e94, 0x1e94}, {0x1e9e, 0x1e9e}, {0x1ea0, 0x1ea0}, {0x1ea2, 0x1ea2}, {0x1ea4, 0x1ea4}, {0x1ea6, 0x1ea6}, {0x1ea8, 0x1ea8},
  {0x1eaa, 0x1eaa}, {0x1eac, 0x1eac}, {0x1eae, 0x1eae}, {0x1eb0, 0x1eb0}, {0x1eb2, 0x1eb2}, {0x1eb4, 0x1eb4}, {0x1eb6, 0x1eb6}, {0x1eb8, 0x1eb8},
  {0x1eba, 0x1eba}, {0x1ebc, 0x1ebc}, {0x1ebe, 0x1ebe}, {0x1ec0, 0x1ec0}, {0x1ec2, 0x1ec2}, {0x1ec4, 0x1ec4}, {0x1ec6, 0x1ec6}, {0x1ec8, 0x1ec8},
  {0x1eca, 0x1eca}, {0x1ecc, 0x1ecc}, {0x1ece, 0x1ece}, {0x1ed0, 0x1ed0}, {0x1ed2, 0x1ed2}, {0x1ed4, 0x1ed4}, {0x1ed6, 0x1ed6}, {0x1ed8, 0x1ed8},
  {0x1eda, 0x1eda}, {0x1edc, 0x1edc}, {0x1ede, 0x1ede}, {0x1ee0, 0x1ee0}, {0x1ee2, 0x1ee2}, {0x1ee4, 0x1ee4}, {0x1ee6, 0x1ee6}, {0x1ee8, 0x1ee8},
  {0x1eea, 0x1eea}, {0x1eec, 0x1eec}, {0x1eee, 0x1eee}, {0x1ef0, 0x1ef0}, {0x1ef2, 0x1ef2}, {0x1ef4, 0x1ef4}, {0x1ef6, 0x1ef6}, {0x1ef8, 0x1ef8},
  {0x1efa, 0x1efa}, {0x1efc, 0x1efc}, {0x1efe, 0x1efe}, {0x1f08, 0x1f0f}, {0x1f18, 0x1f1d}, {0x1f28, 0x1f2f}, {0x1f38, 0x1f3f}, {0x1f48, 0x1f4d},
  {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f5f}, {0x1f68, 0x1f6f}, {0x1fb8, 0x1fbb}, {0x1fc8, 0x1fcb}, {0x1fd8, 0x1fdb},
  {0x1fe8, 0x1fec}, {0x1ff8, 0x1ffb}, {0x2102, 0x2102}, {0x2107, 0x2107}, {0x210b, 0x210d}, {0x2110, 0x2112}, {0x2115, 0x2115}, {0x2119, 0x211d},
  {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x212d}, {0x2130, 0x2133}, {0x213e, 0x213f}, {0x2145, 0x2145}, {0x2183, 0x2183},
  {0x2c00, 0x2c2f}, {0x2c60, 0x2c60}, {0x2c62, 0x2c64}, {0x2c67, 0x2c67}, {0x2c69, 0x2c69}, {0x2c6b, 0x2c6b}, {0x2c6d, 0x2c70}, {0x2c72, 0x2c72},
  {0x2c75, 0x2c75}, {0x2c7e, 0x2c80}, {0x2c82, 0x2c82}, {0x2c84, 0x2c84}, {0x2c86, 0x2c86}, {0x2c88, 0x2c88}, {0x2c8a, 0x2c8a}, {0x2c8c, 0x2c8c},
  {0x2c8e, 0x2c8e}, {0x2c90, 0x2c90}, {0x2c92, 0x2c92}, {0x2c94, 0x2c94}, {0x2c96, 0x2c96}, {0x2c98, 0x2c98}, {0x2c9a, 0x2c9a}, {0x2c9c, 0x2c9c},
  {0x2c9e, 0x2c9e}, {0x2ca0, 0x2ca0}, {0x2ca2, 0x2ca2}, {0x2ca4, 0x2ca4}, {0x2ca6, 0x2ca6}, {0x2ca8, 0x2ca8}, {0x2caa, 0x2caa}, {0x2cac, 0x2cac},
  {0x2cae, 0x2cae}, {0x2cb0, 0x2cb0}, {0x2cb2, 0x2cb2}, {0x2cb4, 0x2cb4}, {0x2cb6, 0x2cb6}, {0x2cb8, 0x2cb8}, {0x2cba, 0x2cba}, {0x2cbc, 0x2cbc},
  {0x2cbe, 0x2cbe}, {0x2cc0, 0x2cc0}, {0x2cc2, 0x2cc2}, {0x2cc4, 0x2cc4}, {0x2cc6, 0x2cc6}, {0x2cc8, 0x2cc8}, {0x2cca, 0x2cca}, {0x2ccc, 0x2ccc},
  {0x2cce, 0x2cce}, {0x2cd0, 0x2cd0}, {0x2cd2, 0x2cd2}, {0x2cd4, 0x2cd4}, {0x2cd6, 0x2cd6}, {0x2cd8, 0x2cd8}, {0x2cda, 0x2cda}, {0x2cdc, 0x2cdc},
  {0x2cde, 0x2cde}, {0x2ce0, 0x2ce0}, {0x2ce2, 0x2ce2}, {0x2ceb, 0x2ceb}, {0x2ced, 0x2ced}, {0x2cf2, 0x2cf2}, {0xa640, 0xa640}, {0xa642, 0xa642},
  {0xa644, 0xa644}, {0xa646, 0xa646}, {0xa648, 0xa648}, {0xa64a, 0xa64a}, {0xa64c, 0xa64c}, {0xa64e, 0xa64e}, {0xa650, 0xa650}, {0xa652, 0xa652},
  {0xa654, 0xa654}, {0xa656, 0xa656}, {0xa658, 0xa658}, {0xa65a, 0xa65a}, {0xa65c, 0xa65c}, {0xa65e, 0xa65e}, {0xa660, 0xa660}, {0xa662, 0xa662},
  {0xa664, 0xa664}, {0xa666, 0xa666}, {0xa668, 0xa668}, {0xa66a, 0xa66a}, {0xa66c, 0xa66c}, {0xa680, 0xa680}, {0xa682, 0xa682}, {0xa684, 0xa684},
  {0xa686, 0xa686}, {0xa688, 0xa688}, {0xa68a, 0xa68a}, {0xa68c, 0xa68c}, {0xa68e, 0xa68e}, {0xa690, 0xa690}, {0xa692, 0xa692}, {0xa694, 0xa694},
  {0xa696, 0xa696}, {0xa698, 0xa698}, {0xa69a, 0xa69a}, {0xa722, 0xa722}, {0xa724, 0xa724}, {0xa726, 0xa726}, {0xa728, 0xa728}, {0xa72a, 0xa72a},
  {0xa72c, 0xa72c}, {0xa72e, 0xa72e}, {0xa732, 0xa732}, {0xa734, 0xa734}, {0xa736, 0xa736}, {0xa738, 0xa738}, {0xa73a, 0xa73a}, {0xa73c, 0xa73c},
  {0xa73e, 0xa73e}, {0xa740, 0xa740}, {0xa742, 0xa742}, {0xa744, 0xa744}, {0xa746, 0xa746}, {0xa748, 0xa748}, {0xa74a, 0xa74a}, {0xa74c, 0xa74c},
  {0xa74e, 0xa74e}, {0xa750, 0xa750}, {0xa752, 0xa752}, {0xa754, 0xa754}, {0xa756, 0xa756}, {0xa758, 0xa758}, {0xa75a, 0xa75a}, {0xa75c, 0xa75c},
  {0xa75e, 0xa75e}, {0xa760, 0xa760}, {0xa762, 0xa762}, {0xa764, 0xa764}, {0xa766, 0xa766}, {0xa768, 0xa768}, {0xa76a, 0xa76a}, {0xa76c, 0xa76c},
  {0xa76e, 0xa76e}, {0xa779, 0xa779}, {0xa77b, 0xa77b}, {0xa77d, 0xa77e}, {0xa780, 0xa780}, {0xa782, 0xa782}, {0xa784, 0xa784}, {0xa786, 0xa786},
  {0xa78b, 0xa78b}, {0xa78d, 0xa78d}, {0xa790, 0xa790}, {0xa792, 0xa792}, {0xa796, 0xa796}, {0xa798, 0xa798}, {0xa79a, 0xa79a}, {0xa79c, 0xa79c},
  {0xa79e, 0xa79e}, {0xa7a0, 0xa7a0}, {0xa7a2, 0xa7a2}, {0xa7a4, 0xa7a4}, {0xa7a6, 0xa7a6}, {0xa7a8, 0xa7a8}, {0xa7aa, 0xa7ae}, {0xa7b0, 0xa7b4},
  {0xa7b6, 0xa7b6}, {0xa7b8, 0xa7b8}, {0xa7ba, 0xa7ba}, {0xa7bc, 0xa7bc}, {0xa7be, 0xa7be}, {0xa7c0, 0xa7c0}, {0xa7c2, 0xa7c2}, {0xa7c4, 0xa7c7},
  {0xa7c9, 0xa7c9}, {0xa7cb, 0xa7cc}, {0xa7d0, 0xa7d0}, {0xa7d6, 0xa7d6}, {0xa7d8, 0xa7d8}, {0xa7da, 0xa7da}, {0xa7dc, 0xa7dc}, {0xa7f5, 0xa7f5},
  {0xff21, 0xff3a}, {0x10400, 0x10427}, {0x104b0, 0x104d3}, {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10c80, 0x10cb2},
  {0x10d50, 0x10d65}, {0x118a0, 0x118bf}, {0x16e40, 0x16e5f}, {0x1d400, 0x1d419}, {0x1d434, 0x1d44d}, {0x1d468, 0x1d481}, {0x1d49c, 0x1d49c}, {0x1d49e, 0x1d49f},
  {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b5}, {0x1d4d0, 0x1d4e9}, {0x1d504, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514},
  {0x1d516, 0x1d51c}, {0x1d538, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d56c, 0x1d585}, {0x1d5a0, 0x1d5b9},
  {0x1d5d4, 0x1d5ed}, {0x1d608, 0x1d621}, {0x1d63c, 0x1d655}, {0x1d670, 0x1d689}, {0x1d6a8, 0x1d6c0}, {0x1d6e2, 0x1d6fa}, {0x1d71c, 0x1d734}, {0x1d756, 0x1d76e},
  {0x1d790, 0x1d7a8}, {0x1d7ca, 0x1d7ca}, {0x1e900, 0x1e921},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 175,
        '"', 277,
        '#', 227,
        '$', 105,
        '%', 251,
        '&', 222,
        '\'', 273,
        '(', 164,
        ')', 165,
        '*', 207,
        '+', 171,
        ',', 169,
        '-', 173,
        '.', 191,
        '0', 254,
        ':', 160,
        ';', 215,
        '<', 225,
        '=', 153,
        '>', 220,
        '?', 156,
        '@', 311,
        '[', 194,
        '\\', 2,
        ']', 205,
        '^', 313,
        '`', 354,
        'a', 339,
        'l', 335,
        'm', 330,
        '{', 210,
        '|', 178,
        '}', 216,
        '~', 154,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(118);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 1:
      if (lookahead == '\n') ADVANCE(285);
      if (lookahead == '\r') ADVANCE(1);
      END_STATE();
    case 2:
      ADVANCE_MAP(
        '\n', 285,
        '\r', 1,
        '#', 117,
        'o', 102,
        'u', 94,
        'x', 116,
        ' ', 291,
        '"', 291,
        '\'', 291,
        '\\', 291,
        'b', 291,
        'n', 291,
        'r', 291,
        't', 291,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(109);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        '\n', 285,
        '\r', 1,
        'o', 102,
        'u', 94,
        'x', 116,
        ' ', 291,
        '"', 291,
        '\'', 291,
        '\\', 291,
        'b', 291,
        'n', 291,
        'r', 291,
        't', 291,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(109);
      END_STATE();
    case 4:
      if (lookahead == '\n') ADVANCE(274);
      if (lookahead == '\r') ADVANCE(5);
      if (lookahead == '[') ADVANCE(276);
      if (lookahead == '\\') ADVANCE(93);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(275);
      if (lookahead != 0 &&
          lookahead != '\'') ADVANCE(275);
      END_STATE();
    case 5:
      if (lookahead == '\n') ADVANCE(274);
      if (lookahead == '\r') ADVANCE(5);
      if (lookahead == '[') ADVANCE(86);
      if (lookahead == '\\') ADVANCE(93);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(89);
      END_STATE();
    case 6:
      ADVANCE_MAP(
        '!', 175,
        '"', 277,
        '#', 226,
        '%', 251,
        '\'', 272,
        '(', 164,
        '+', 265,
        '-', 271,
        '.', 190,
        '0', 254,
        ':', 189,
        ';', 79,
        '=', 153,
        '?', 156,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 209,
        '|', 90,
        '~', 154,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(6);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 7:
      ADVANCE_MAP(
        '!', 175,
        '"', 277,
        '%', 251,
        '&', 221,
        '\'', 272,
        '(', 164,
        ')', 165,
        '*', 206,
        '+', 265,
        ',', 169,
        '-', 271,
        '.', 73,
        '0', 254,
        ':', 183,
        ';', 79,
        '=', 153,
        '>', 220,
        '[', 197,
        '\\', 50,
        ']', 205,
        '{', 209,
        '|', 176,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(7);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 8:
      ADVANCE_MAP(
        '!', 175,
        '#', 226,
        '%', 251,
        '&', 221,
        '\'', 353,
        '(', 164,
        ')', 165,
        '+', 170,
        '-', 172,
        '.', 73,
        ':', 77,
        ';', 79,
        '<', 224,
        '>', 220,
        '?', 156,
        '[', 196,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 209,
        '|', 176,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(8);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 9:
      ADVANCE_MAP(
        '!', 175,
        '\'', 353,
        '(', 164,
        '+', 170,
        '-', 174,
        '.', 355,
        '[', 199,
        '\\', 50,
        'a', 340,
        'l', 336,
        '{', 58,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(9);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 10:
      ADVANCE_MAP(
        '!', 175,
        '\'', 353,
        '+', 170,
        ',', 169,
        '-', 172,
        '.', 73,
        ':', 77,
        ';', 214,
        '[', 86,
        '\\', 50,
        '|', 177,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(10);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 11:
      ADVANCE_MAP(
        '!', 175,
        '\'', 353,
        '+', 170,
        ',', 169,
        '-', 172,
        '.', 73,
        ':', 159,
        ';', 214,
        '[', 86,
        '\\', 50,
        '|', 177,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(10);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 12:
      ADVANCE_MAP(
        '!', 299,
        '"', 277,
        '#', 351,
        '$', 105,
        '\'', 272,
        '(', 164,
        '.', 355,
        '0', 254,
        ':', 181,
        ';', 79,
        '?', 157,
        '[', 200,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 335,
        'm', 330,
        '{', 210,
        '~', 155,
        '+', 266,
        '-', 266,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(12);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 13:
      ADVANCE_MAP(
        '!', 299,
        '"', 277,
        '#', 228,
        '$', 105,
        '\'', 272,
        '(', 164,
        '+', 80,
        '.', 190,
        '0', 254,
        ':', 158,
        ';', 79,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 210,
        '~', 155,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(14);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 14:
      ADVANCE_MAP(
        '!', 299,
        '"', 277,
        '#', 228,
        '$', 105,
        '\'', 272,
        '(', 164,
        '+', 80,
        '.', 190,
        '0', 254,
        ';', 79,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 210,
        '~', 155,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(14);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 15:
      ADVANCE_MAP(
        '!', 299,
        '"', 277,
        '#', 226,
        '$', 105,
        '%', 251,
        '&', 221,
        '\'', 272,
        '(', 164,
        ')', 165,
        '*', 206,
        '+', 265,
        ',', 169,
        '-', 271,
        '.', 73,
        '0', 254,
        ':', 183,
        ';', 79,
        '=', 153,
        '>', 220,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 210,
        '|', 176,
        '~', 155,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(15);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 16:
      ADVANCE_MAP(
        '!', 83,
        '"', 277,
        '#', 228,
        '&', 223,
        '\'', 273,
        '(', 164,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        '.', 190,
        ':', 185,
        ';', 214,
        '<', 316,
        '>', 317,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        '{', 209,
        '|', 179,
        '}', 216,
        '$', 319,
        '=', 319,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(17);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 17:
      ADVANCE_MAP(
        '!', 83,
        '"', 277,
        '#', 228,
        '&', 223,
        '(', 164,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        '.', 190,
        ':', 185,
        ';', 214,
        '<', 316,
        '>', 317,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        '{', 209,
        '|', 179,
        '}', 216,
        '$', 319,
        '=', 319,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(17);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 18:
      ADVANCE_MAP(
        '!', 83,
        '#', 228,
        '&', 223,
        '(', 164,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        '.', 190,
        ':', 185,
        ';', 214,
        '<', 316,
        '>', 317,
        '[', 203,
        '\\', 50,
        ']', 205,
        '{', 63,
        '|', 180,
        '}', 216,
        '$', 319,
        '=', 319,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(18);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 19:
      ADVANCE_MAP(
        '!', 83,
        '#', 228,
        '&', 223,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '.', 190,
        ':', 161,
        ';', 214,
        '<', 316,
        '[', 88,
        '\\', 50,
        'a', 339,
        '|', 180,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '$', 319,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(20);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 20:
      ADVANCE_MAP(
        '!', 83,
        '#', 228,
        '&', 223,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '.', 190,
        ':', 78,
        ';', 214,
        '<', 316,
        '[', 88,
        '\\', 50,
        'a', 339,
        '|', 180,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '$', 319,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(20);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 21:
      ADVANCE_MAP(
        '!', 83,
        '&', 223,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        ':', 162,
        ';', 214,
        '<', 316,
        '>', 317,
        '[', 61,
        '\\', 50,
        ']', 205,
        '{', 63,
        '|', 180,
        '}', 216,
        '$', 319,
        '=', 319,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(23);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 22:
      ADVANCE_MAP(
        '!', 83,
        '&', 223,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        ':', 162,
        ';', 214,
        '<', 316,
        '>', 317,
        '[', 61,
        '\\', 50,
        ']', 205,
        '{', 63,
        '|', 179,
        '}', 216,
        '$', 319,
        '=', 319,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(24);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 23:
      ADVANCE_MAP(
        '!', 83,
        '&', 223,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        ':', 185,
        ';', 214,
        '<', 316,
        '>', 317,
        '[', 61,
        '\\', 50,
        ']', 205,
        '{', 63,
        '|', 180,
        '}', 216,
        '$', 319,
        '=', 319,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(23);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 24:
      ADVANCE_MAP(
        '!', 83,
        '&', 223,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        ':', 185,
        ';', 214,
        '<', 316,
        '>', 317,
        '[', 61,
        '\\', 50,
        ']', 205,
        '{', 63,
        '|', 179,
        '}', 216,
        '$', 319,
        '=', 319,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(24);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 25:
      ADVANCE_MAP(
        '!', 83,
        '&', 223,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        ':', 185,
        ';', 214,
        '<', 316,
        '>', 317,
        '[', 88,
        '\\', 50,
        ']', 205,
        'a', 339,
        '|', 179,
        '}', 216,
        '$', 319,
        '=', 319,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(25);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 26:
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 228,
        '$', 318,
        '%', 252,
        '&', 223,
        '\'', 272,
        '(', 164,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        '.', 190,
        '/', 308,
        '0', 254,
        ':', 185,
        ';', 214,
        '<', 316,
        '=', 319,
        '>', 317,
        '?', 113,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 335,
        'm', 330,
        '{', 210,
        '|', 180,
        '}', 216,
        '~', 155,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(26);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 27:
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 228,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        '.', 190,
        '0', 254,
        ':', 185,
        ';', 214,
        '<', 315,
        '=', 319,
        '>', 317,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 210,
        '|', 180,
        '}', 216,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(27);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 28:
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 228,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        '.', 190,
        '0', 254,
        ':', 185,
        ';', 214,
        '<', 315,
        '=', 319,
        '>', 317,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 210,
        '|', 179,
        '}', 216,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(28);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 29:
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 228,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '.', 190,
        '0', 254,
        ':', 161,
        ';', 214,
        '<', 315,
        '?', 157,
        '[', 202,
        '\\', 50,
        '`', 354,
        'a', 339,
        'l', 336,
        '{', 212,
        '|', 180,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(30);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 30:
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 228,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '.', 190,
        '0', 254,
        ':', 78,
        ';', 214,
        '<', 315,
        '?', 157,
        '[', 202,
        '\\', 50,
        '`', 354,
        'a', 339,
        'l', 336,
        '{', 212,
        '|', 180,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(30);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 31:
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 228,
        '&', 223,
        '\'', 272,
        '(', 164,
        ')', 165,
        '*', 307,
        '+', 268,
        '-', 269,
        '.', 190,
        '0', 254,
        ':', 78,
        '<', 316,
        '?', 113,
        '[', 201,
        '\\', 50,
        '`', 354,
        'a', 339,
        'l', 335,
        'm', 330,
        '{', 211,
        '|', 95,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '$', 319,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(31);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 32:
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 112,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        ')', 165,
        '*', 307,
        '+', 268,
        '-', 269,
        '.', 190,
        '0', 254,
        ':', 78,
        ';', 214,
        '<', 316,
        '?', 113,
        '[', 201,
        '\\', 50,
        '`', 354,
        'a', 339,
        'l', 335,
        'm', 330,
        '{', 212,
        '|', 91,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(32);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 33:
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '$', 318,
        '%', 252,
        '&', 223,
        '\'', 272,
        '(', 164,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '/', 308,
        '0', 254,
        ':', 78,
        ';', 214,
        '<', 316,
        '?', 113,
        '[', 201,
        '\\', 50,
        '`', 354,
        'a', 340,
        'l', 335,
        'm', 330,
        '{', 212,
        '|', 179,
        '~', 155,
        '@', 313,
        '^', 313,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(33);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 34:
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '$', 318,
        '%', 252,
        '&', 223,
        '\'', 272,
        '(', 164,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '/', 308,
        '0', 254,
        ':', 78,
        ';', 214,
        '<', 316,
        '?', 113,
        '[', 202,
        '\\', 50,
        '`', 354,
        'a', 339,
        'l', 335,
        'm', 330,
        '{', 212,
        '|', 180,
        '~', 155,
        '@', 313,
        '^', 313,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(34);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 35:
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        '.', 355,
        '0', 254,
        ':', 162,
        ';', 214,
        '<', 316,
        '=', 319,
        '>', 317,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 210,
        '|', 179,
        '}', 216,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(36);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 36:
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        '.', 355,
        '0', 254,
        ':', 185,
        ';', 214,
        '<', 316,
        '=', 319,
        '>', 317,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 210,
        '|', 179,
        '}', 216,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(36);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 37:
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        '0', 254,
        ':', 162,
        ';', 214,
        '<', 316,
        '=', 319,
        '>', 317,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 210,
        '|', 180,
        '}', 216,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(38);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 38:
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 270,
        '0', 254,
        ':', 185,
        ';', 214,
        '<', 316,
        '=', 319,
        '>', 317,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 210,
        '|', 180,
        '}', 216,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(38);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 39:
      ADVANCE_MAP(
        '!', 300,
        '#', 112,
        '&', 223,
        '*', 307,
        '+', 268,
        '-', 269,
        '.', 190,
        ':', 82,
        '<', 316,
        '[', 86,
        '\\', 50,
        'a', 341,
        'l', 337,
        'm', 330,
        '|', 95,
        '%', 308,
        '/', 308,
        '?', 113,
        '~', 113,
        '@', 313,
        '^', 313,
        '$', 319,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(39);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 40:
      ADVANCE_MAP(
        '"', 277,
        '#', 228,
        '&', 221,
        '\'', 272,
        '(', 164,
        ')', 165,
        '*', 206,
        '+', 265,
        ',', 169,
        '-', 271,
        '.', 190,
        '0', 254,
        ':', 183,
        ';', 214,
        '=', 153,
        '>', 220,
        '?', 156,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 209,
        '|', 176,
        '}', 216,
        '~', 154,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(40);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 41:
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        '*', 206,
        '+', 265,
        ',', 169,
        '-', 271,
        '.', 190,
        '0', 254,
        ':', 186,
        ';', 214,
        '=', 153,
        '?', 156,
        '[', 198,
        '\\', 50,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 209,
        '|', 177,
        '~', 154,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(41);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 42:
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        '*', 206,
        '+', 265,
        ',', 169,
        '-', 271,
        '.', 190,
        '0', 254,
        ':', 163,
        ';', 214,
        '=', 153,
        '?', 156,
        '[', 198,
        '\\', 50,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 209,
        '|', 177,
        '~', 154,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(41);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 43:
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        ',', 169,
        '.', 190,
        '0', 254,
        ':', 186,
        ';', 214,
        '=', 153,
        '>', 97,
        '?', 156,
        '[', 202,
        '\\', 50,
        '`', 354,
        'a', 339,
        'l', 336,
        '{', 211,
        '|', 176,
        '}', 216,
        '~', 154,
        '+', 265,
        '-', 265,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(43);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 44:
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        ',', 169,
        '.', 73,
        '0', 254,
        ':', 186,
        '=', 153,
        '?', 156,
        '[', 202,
        '\\', 50,
        '`', 354,
        'a', 339,
        'l', 336,
        '{', 211,
        '|', 176,
        '~', 154,
        '+', 265,
        '-', 265,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(44);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 45:
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        ',', 169,
        '0', 254,
        ':', 186,
        '=', 153,
        '?', 156,
        '[', 202,
        '\\', 50,
        '`', 354,
        'a', 339,
        'l', 336,
        '{', 211,
        '|', 176,
        '~', 154,
        '+', 265,
        '-', 265,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(45);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 46:
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        ',', 169,
        '0', 254,
        ':', 163,
        '=', 153,
        '?', 156,
        '[', 202,
        '\\', 50,
        '`', 354,
        'a', 339,
        'l', 336,
        '{', 211,
        '|', 176,
        '~', 154,
        '+', 265,
        '-', 265,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(45);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 47:
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        '.', 355,
        '0', 254,
        ':', 181,
        ';', 79,
        '=', 153,
        '?', 156,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 209,
        '|', 176,
        '~', 154,
        '+', 265,
        '-', 265,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(47);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 48:
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        '.', 355,
        '0', 254,
        ':', 158,
        ';', 79,
        '=', 153,
        '?', 156,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 336,
        '{', 209,
        '|', 176,
        '~', 154,
        '+', 265,
        '-', 265,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(47);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 49:
      if (lookahead == '"') ADVANCE(277);
      if (lookahead == '%') ADVANCE(280);
      if (lookahead == '@') ADVANCE(281);
      if (lookahead == '[') ADVANCE(282);
      if (lookahead == '\\') ADVANCE(3);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(278);
      if (lookahead != 0) ADVANCE(283);
      END_STATE();
    case 50:
      if (lookahead == '#') ADVANCE(117);
      END_STATE();
    case 51:
      ADVANCE_MAP(
        '#', 228,
        '(', 164,
        ')', 165,
        ',', 169,
        '-', 84,
        '.', 190,
        ':', 186,
        ';', 214,
        '=', 153,
        '>', 97,
        '[', 61,
        '\\', 50,
        ']', 205,
        '{', 63,
        '|', 176,
        '}', 216,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(51);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 52:
      ADVANCE_MAP(
        '#', 228,
        ')', 165,
        '*', 206,
        ',', 169,
        '-', 84,
        '.', 190,
        ':', 183,
        ';', 214,
        '=', 153,
        '>', 220,
        '[', 88,
        '\\', 50,
        ']', 205,
        'a', 340,
        '|', 176,
        '}', 216,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(52);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 53:
      ADVANCE_MAP(
        '#', 228,
        ')', 165,
        ',', 169,
        '-', 84,
        '.', 191,
        ':', 182,
        ';', 214,
        '=', 153,
        '>', 220,
        '[', 86,
        '\\', 50,
        ']', 205,
        '|', 176,
        '}', 216,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(53);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 54:
      ADVANCE_MAP(
        '#', 228,
        ')', 165,
        ',', 169,
        '-', 84,
        '.', 191,
        ':', 182,
        ';', 214,
        '=', 153,
        '[', 88,
        '\\', 50,
        ']', 205,
        'a', 340,
        '|', 176,
        '}', 216,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(54);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 55:
      ADVANCE_MAP(
        '#', 226,
        '%', 251,
        '&', 221,
        '\'', 353,
        ')', 165,
        '*', 206,
        '+', 80,
        ',', 169,
        '-', 84,
        '.', 190,
        ':', 183,
        ';', 214,
        '=', 153,
        '>', 220,
        '[', 197,
        '\\', 50,
        ']', 205,
        'a', 340,
        'l', 336,
        '{', 58,
        '|', 176,
        '}', 216,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(55);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 56:
      ADVANCE_MAP(
        '#', 226,
        '\'', 353,
        '*', 206,
        '-', 84,
        '.', 190,
        ':', 81,
        ';', 79,
        '=', 153,
        '[', 61,
        '\\', 50,
        ']', 205,
        'a', 340,
        'l', 336,
        '{', 63,
        '|', 176,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(56);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 57:
      if (lookahead == '%') ADVANCE(245);
      if (lookahead == '@') ADVANCE(238);
      END_STATE();
    case 58:
      if (lookahead == '%') ADVANCE(247);
      END_STATE();
    case 59:
      if (lookahead == '%') ADVANCE(249);
      END_STATE();
    case 60:
      if (lookahead == '%') ADVANCE(250);
      END_STATE();
    case 61:
      if (lookahead == '%') ADVANCE(59);
      if (lookahead == '@') ADVANCE(238);
      END_STATE();
    case 62:
      if (lookahead == '%') ADVANCE(287);
      if (lookahead == '@') ADVANCE(288);
      if (lookahead == '[') ADVANCE(289);
      if (lookahead == '|') ADVANCE(286);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(278);
      if (lookahead != 0) ADVANCE(290);
      END_STATE();
    case 63:
      if (lookahead == '%') ADVANCE(60);
      END_STATE();
    case 64:
      ADVANCE_MAP(
        '&', 221,
        '(', 164,
        ')', 165,
        '*', 206,
        '+', 80,
        ',', 169,
        '-', 84,
        '.', 190,
        ':', 188,
        ';', 79,
        '=', 153,
        '>', 220,
        '[', 197,
        '\\', 50,
        ']', 205,
        'a', 340,
        '{', 58,
        '|', 176,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(64);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 65:
      if (lookahead == '(') ADVANCE(164);
      if (lookahead == '[') ADVANCE(204);
      if (lookahead == '\\') ADVANCE(50);
      if (lookahead == 'a') ADVANCE(340);
      if (lookahead == 'l') ADVANCE(336);
      if (lookahead == '{') ADVANCE(208);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(65);
      if (set_contains(sym_indexing_operator_character_set_1, 9, lookahead)) ADVANCE(326);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 66:
      ADVANCE_MAP(
        ')', 165,
        ',', 169,
        '-', 84,
        '.', 73,
        ':', 182,
        ';', 214,
        '=', 153,
        '[', 86,
        '\\', 50,
        ']', 205,
        '|', 176,
        '}', 216,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(66);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 67:
      ADVANCE_MAP(
        ')', 165,
        ',', 169,
        '-', 84,
        ':', 182,
        ';', 214,
        '=', 153,
        '[', 86,
        '\\', 50,
        ']', 205,
        '|', 176,
        '}', 216,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(67);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 68:
      ADVANCE_MAP(
        ')', 165,
        ',', 169,
        '-', 84,
        ':', 159,
        ';', 214,
        '=', 153,
        '[', 86,
        '\\', 50,
        ']', 205,
        '|', 176,
        '}', 216,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(67);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 69:
      if (lookahead == '*') ADVANCE(76);
      if (lookahead == '.') ADVANCE(70);
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(296);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(75);
      if (set_contains(sym_conversion_specification_character_set_6, 15, lookahead)) ADVANCE(295);
      END_STATE();
    case 70:
      if (lookahead == '*') ADVANCE(101);
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(296);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(100);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(295);
      END_STATE();
    case 71:
      if (lookahead == ',') ADVANCE(169);
      if (lookahead == '.') ADVANCE(190);
      if (lookahead == ':') ADVANCE(77);
      if (lookahead == ';') ADVANCE(214);
      if (lookahead == '[') ADVANCE(86);
      if (lookahead == '\\') ADVANCE(50);
      if (lookahead == '|') ADVANCE(177);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(71);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 72:
      if (lookahead == '-') ADVANCE(232);
      END_STATE();
    case 73:
      if (lookahead == '.') ADVANCE(167);
      END_STATE();
    case 74:
      if (lookahead == '.') ADVANCE(355);
      if (lookahead == '[') ADVANCE(86);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(248);
      END_STATE();
    case 75:
      if (lookahead == '.') ADVANCE(70);
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(296);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(75);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(295);
      END_STATE();
    case 76:
      if (lookahead == '.') ADVANCE(70);
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(296);
      if (set_contains(sym_conversion_specification_character_set_6, 15, lookahead)) ADVANCE(295);
      END_STATE();
    case 77:
      if (lookahead == ':') ADVANCE(229);
      END_STATE();
    case 78:
      if (lookahead == ':') ADVANCE(229);
      if (lookahead == '=') ADVANCE(325);
      END_STATE();
    case 79:
      if (lookahead == ';') ADVANCE(148);
      END_STATE();
    case 80:
      if (lookahead == '=') ADVANCE(168);
      END_STATE();
    case 81:
      if (lookahead == '=') ADVANCE(166);
      END_STATE();
    case 82:
      if (lookahead == '=') ADVANCE(325);
      END_STATE();
    case 83:
      if (lookahead == '=') ADVANCE(314);
      END_STATE();
    case 84:
      if (lookahead == '>') ADVANCE(192);
      END_STATE();
    case 85:
      if (lookahead == '>') ADVANCE(297);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(85);
      END_STATE();
    case 86:
      if (lookahead == '@') ADVANCE(237);
      END_STATE();
    case 87:
      if (lookahead == '@') ADVANCE(243);
      END_STATE();
    case 88:
      if (lookahead == '@') ADVANCE(239);
      END_STATE();
    case 89:
      if (lookahead == '[') ADVANCE(86);
      if (lookahead == '\\') ADVANCE(93);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(89);
      END_STATE();
    case 90:
      if (lookahead == ']') ADVANCE(231);
      END_STATE();
    case 91:
      if (lookahead == ']') ADVANCE(231);
      if (lookahead == '|') ADVANCE(324);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(319);
      END_STATE();
    case 92:
      if (lookahead == 'n') ADVANCE(297);
      END_STATE();
    case 93:
      ADVANCE_MAP(
        'o', 102,
        'x', 116,
        ' ', 291,
        '"', 291,
        '\'', 291,
        '\\', 291,
        'b', 291,
        'n', 291,
        'r', 291,
        't', 291,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(109);
      END_STATE();
    case 94:
      if (lookahead == '{') ADVANCE(114);
      END_STATE();
    case 95:
      if (lookahead == '|') ADVANCE(324);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(319);
      END_STATE();
    case 96:
      if (lookahead == '}') ADVANCE(284);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(96);
      END_STATE();
    case 97:
      if (lookahead == '}') ADVANCE(235);
      END_STATE();
    case 98:
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(106);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(263);
      END_STATE();
    case 99:
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(260);
      END_STATE();
    case 100:
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(296);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(100);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(295);
      END_STATE();
    case 101:
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(296);
      if (set_contains(sym_conversion_specification_character_set_6, 15, lookahead)) ADVANCE(295);
      END_STATE();
    case 102:
      if (('0' <= lookahead && lookahead <= '3')) ADVANCE(104);
      END_STATE();
    case 103:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(294);
      END_STATE();
    case 104:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(103);
      END_STATE();
    case 105:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(236);
      END_STATE();
    case 106:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(263);
      END_STATE();
    case 107:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(85);
      END_STATE();
    case 108:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(292);
      END_STATE();
    case 109:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(108);
      END_STATE();
    case 110:
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(319);
      END_STATE();
    case 111:
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(310);
      END_STATE();
    case 112:
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(305);
      END_STATE();
    case 113:
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(301);
      END_STATE();
    case 114:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(96);
      END_STATE();
    case 115:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(293);
      END_STATE();
    case 116:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(115);
      END_STATE();
    case 117:
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 118:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 175,
        '"', 277,
        '#', 227,
        '$', 105,
        '%', 251,
        '&', 222,
        '\'', 272,
        '(', 164,
        ')', 165,
        '*', 207,
        '+', 171,
        ',', 169,
        '-', 173,
        '.', 191,
        '0', 254,
        ':', 183,
        ';', 215,
        '<', 225,
        '=', 153,
        '>', 220,
        '?', 156,
        '@', 311,
        '[', 195,
        '\\', 2,
        ']', 205,
        '^', 313,
        '`', 354,
        'a', 339,
        'l', 335,
        'm', 330,
        '{', 210,
        '|', 178,
        '}', 216,
        '~', 154,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(118);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 119:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 175,
        '#', 351,
        '&', 221,
        '(', 164,
        ')', 165,
        '*', 206,
        '+', 80,
        ',', 169,
        '-', 84,
        '.', 190,
        ':', 183,
        ';', 79,
        '=', 153,
        '>', 220,
        '[', 197,
        '\\', 50,
        ']', 205,
        'a', 340,
        'l', 335,
        '{', 58,
        '|', 176,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(119);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 120:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 299,
        '"', 277,
        '#', 352,
        '$', 105,
        '%', 251,
        '\'', 272,
        '(', 164,
        ')', 165,
        '*', 206,
        '+', 266,
        ',', 169,
        '-', 267,
        '.', 190,
        '0', 254,
        ':', 183,
        ';', 79,
        '<', 72,
        '=', 153,
        '>', 220,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 335,
        'm', 330,
        '{', 210,
        '|', 90,
        '}', 216,
        '~', 155,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(120);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 121:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 299,
        '"', 277,
        '#', 351,
        '$', 105,
        '%', 251,
        '&', 221,
        '\'', 272,
        '(', 164,
        '0', 254,
        ';', 79,
        '>', 220,
        '?', 113,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 339,
        'l', 335,
        'm', 330,
        '{', 210,
        '|', 176,
        '~', 155,
        '+', 266,
        '-', 266,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(121);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 122:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 83,
        '"', 277,
        '#', 351,
        '&', 223,
        '(', 164,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '.', 190,
        ':', 78,
        ';', 215,
        '<', 316,
        '[', 57,
        '\\', 50,
        ']', 205,
        'l', 335,
        '{', 209,
        '|', 180,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '$', 319,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(122);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 123:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 83,
        '#', 351,
        '&', 223,
        '\'', 353,
        '(', 164,
        ')', 165,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '.', 190,
        ':', 185,
        ';', 215,
        '<', 316,
        '[', 203,
        '\\', 50,
        ']', 205,
        'a', 339,
        'l', 335,
        '{', 213,
        '|', 180,
        '}', 216,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '$', 319,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(123);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 124:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 83,
        '#', 351,
        '&', 223,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        ':', 161,
        ';', 215,
        '<', 316,
        '[', 61,
        '\\', 50,
        ']', 205,
        'a', 339,
        'l', 335,
        '{', 63,
        '|', 180,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '$', 319,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(125);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 125:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 83,
        '#', 351,
        '&', 223,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        ':', 78,
        ';', 215,
        '<', 316,
        '[', 61,
        '\\', 50,
        ']', 205,
        'a', 339,
        'l', 335,
        '{', 63,
        '|', 180,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '$', 319,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(125);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 126:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 83,
        '#', 228,
        '&', 223,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '.', 190,
        ':', 184,
        ';', 215,
        '<', 316,
        '[', 61,
        '\\', 50,
        ']', 205,
        'a', 339,
        'l', 335,
        '{', 63,
        '|', 180,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '$', 319,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(126);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 127:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 351,
        '$', 318,
        '%', 252,
        '&', 223,
        '\'', 272,
        '(', 164,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '.', 73,
        '/', 308,
        '0', 254,
        ':', 161,
        ';', 215,
        '<', 316,
        '=', 319,
        '>', 317,
        '?', 113,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 339,
        'l', 335,
        'm', 330,
        '{', 210,
        '|', 180,
        '~', 155,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(128);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 128:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 351,
        '$', 318,
        '%', 252,
        '&', 223,
        '\'', 272,
        '(', 164,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '.', 73,
        '/', 308,
        '0', 254,
        ':', 78,
        ';', 215,
        '<', 316,
        '=', 319,
        '>', 317,
        '?', 113,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 339,
        'l', 335,
        'm', 330,
        '{', 210,
        '|', 180,
        '~', 155,
        '@', 313,
        '^', 313,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(128);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 129:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 351,
        '$', 318,
        '%', 252,
        '&', 223,
        '\'', 272,
        '(', 164,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '/', 308,
        '0', 254,
        ':', 78,
        ';', 215,
        '<', 316,
        '?', 113,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 335,
        'm', 330,
        '{', 210,
        '|', 180,
        '~', 155,
        '@', 313,
        '^', 313,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(129);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 130:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 351,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '0', 254,
        ':', 161,
        ';', 215,
        '<', 316,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 339,
        'l', 335,
        '{', 210,
        '|', 180,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(132);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 131:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 351,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '0', 254,
        ':', 161,
        ';', 215,
        '<', 316,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 335,
        '{', 210,
        '|', 180,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(133);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 132:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 351,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '0', 254,
        ':', 78,
        ';', 215,
        '<', 316,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 339,
        'l', 335,
        '{', 210,
        '|', 180,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(132);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 133:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 351,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '0', 254,
        ':', 78,
        ';', 215,
        '<', 316,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 335,
        '{', 210,
        '|', 180,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(133);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 134:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 228,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '.', 190,
        '0', 254,
        ':', 78,
        ';', 215,
        '<', 315,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 339,
        'l', 335,
        '{', 210,
        '|', 180,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(134);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 135:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '!', 300,
        '"', 277,
        '#', 228,
        '$', 318,
        '&', 223,
        '\'', 272,
        '(', 164,
        '*', 307,
        '+', 268,
        ',', 169,
        '-', 269,
        '.', 190,
        '0', 254,
        ':', 78,
        ';', 215,
        '<', 315,
        '?', 157,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 335,
        '{', 210,
        '|', 180,
        '~', 155,
        '%', 308,
        '/', 308,
        '@', 313,
        '^', 313,
        '=', 319,
        '>', 319,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(135);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 136:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        '*', 206,
        '+', 265,
        '-', 271,
        '.', 190,
        '0', 254,
        ':', 187,
        ';', 79,
        '=', 153,
        '?', 156,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 335,
        '{', 209,
        '|', 176,
        '~', 154,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(136);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 137:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        ',', 169,
        '.', 190,
        '0', 254,
        ':', 186,
        ';', 79,
        '=', 153,
        '?', 156,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 339,
        'l', 335,
        '{', 209,
        '|', 176,
        '~', 154,
        '+', 265,
        '-', 265,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(137);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 138:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        ',', 169,
        '.', 73,
        '0', 254,
        ':', 186,
        ';', 79,
        '=', 153,
        '?', 156,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 339,
        'l', 335,
        '{', 209,
        '|', 176,
        '~', 154,
        '+', 265,
        '-', 265,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(138);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 139:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        ',', 169,
        '0', 254,
        ':', 186,
        ';', 79,
        '=', 153,
        '?', 156,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 339,
        'l', 335,
        '{', 209,
        '|', 176,
        '~', 154,
        '+', 265,
        '-', 265,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(139);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 140:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        ',', 169,
        '0', 254,
        ':', 163,
        ';', 79,
        '=', 153,
        '?', 156,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 339,
        'l', 335,
        '{', 209,
        '|', 176,
        '~', 154,
        '+', 265,
        '-', 265,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(139);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 141:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        '0', 254,
        ':', 181,
        ';', 79,
        '=', 153,
        '?', 156,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 335,
        '{', 209,
        '~', 154,
        '+', 265,
        '-', 265,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(141);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 142:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '"', 277,
        '#', 226,
        '\'', 272,
        '(', 164,
        '0', 254,
        ':', 158,
        ';', 79,
        '=', 153,
        '?', 156,
        '[', 198,
        '\\', 50,
        ']', 205,
        '`', 354,
        'a', 340,
        'l', 335,
        '{', 209,
        '~', 154,
        '+', 265,
        '-', 265,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(141);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 143:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '#', 351,
        ')', 165,
        ',', 169,
        '-', 84,
        ':', 183,
        ';', 79,
        '=', 153,
        '[', 61,
        '\\', 50,
        ']', 205,
        'l', 335,
        '{', 63,
        '|', 176,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(143);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 144:
      if (eof) ADVANCE(145);
      ADVANCE_MAP(
        '#', 226,
        '\'', 353,
        '*', 206,
        '-', 84,
        '.', 190,
        ':', 81,
        ';', 79,
        '=', 153,
        '[', 61,
        '\\', 50,
        ']', 205,
        'a', 340,
        'l', 335,
        '{', 63,
        '|', 176,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(144);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(349);
      END_STATE();
    case 145:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(sym_shebang);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(146);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(147);
      END_STATE();
    case 147:
      ACCEPT_TOKEN(sym_shebang);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(147);
      END_STATE();
    case 148:
      ACCEPT_TOKEN(anon_sym_SEMI_SEMI);
      END_STATE();
    case 149:
      ACCEPT_TOKEN(anon_sym_let);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(327);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 150:
      ACCEPT_TOKEN(anon_sym_let);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 151:
      ACCEPT_TOKEN(anon_sym_and);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(328);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 152:
      ACCEPT_TOKEN(anon_sym_and);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 153:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 154:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 155:
      ACCEPT_TOKEN(anon_sym_TILDE);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(301);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(301);
      END_STATE();
    case 158:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(229);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(229);
      if (lookahead == '=') ADVANCE(166);
      if (lookahead == '>') ADVANCE(217);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(229);
      if (lookahead == '=') ADVANCE(325);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(229);
      if (lookahead == '=') ADVANCE(325);
      if (lookahead == '>') ADVANCE(217);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(229);
      if (lookahead == '>') ADVANCE(217);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(anon_sym_COLON_EQ);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 169:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '.') ADVANCE(302);
      if (lookahead == '=') ADVANCE(168);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(302);
      if (lookahead == '>') ADVANCE(192);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '>') ADVANCE(192);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == ']') ADVANCE(231);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == ']') ADVANCE(231);
      if (lookahead == '|') ADVANCE(323);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == ']') ADVANCE(231);
      if (lookahead == '|') ADVANCE(324);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(319);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '|') ADVANCE(324);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(319);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(anon_sym_COLON2);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(229);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(229);
      if (lookahead == '=') ADVANCE(166);
      if (lookahead == '>') ADVANCE(217);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(229);
      if (lookahead == '=') ADVANCE(325);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(229);
      if (lookahead == '=') ADVANCE(325);
      if (lookahead == '>') ADVANCE(217);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(229);
      if (lookahead == '>') ADVANCE(217);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == '=') ADVANCE(166);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == '=') ADVANCE(166);
      if (lookahead == '>') ADVANCE(217);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == '>') ADVANCE(217);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 191:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(167);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(310);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(245);
      if (lookahead == '<') ADVANCE(219);
      if (lookahead == '>') ADVANCE(218);
      if (lookahead == '@') ADVANCE(279);
      if (lookahead == '|') ADVANCE(230);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(245);
      if (lookahead == '<') ADVANCE(219);
      if (lookahead == '>') ADVANCE(218);
      if (lookahead == '@') ADVANCE(237);
      if (lookahead == '|') ADVANCE(230);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(245);
      if (lookahead == '<') ADVANCE(219);
      if (lookahead == '>') ADVANCE(218);
      if (lookahead == '@') ADVANCE(240);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(245);
      if (lookahead == '@') ADVANCE(238);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(245);
      if (lookahead == '@') ADVANCE(238);
      if (lookahead == '|') ADVANCE(230);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(245);
      if (lookahead == '@') ADVANCE(240);
      END_STATE();
    case 200:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(245);
      if (lookahead == '@') ADVANCE(240);
      if (lookahead == '|') ADVANCE(230);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(244);
      if (lookahead == '@') ADVANCE(237);
      if (lookahead == '|') ADVANCE(230);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(244);
      if (lookahead == '@') ADVANCE(239);
      if (lookahead == '|') ADVANCE(230);
      END_STATE();
    case 203:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(59);
      if (lookahead == '@') ADVANCE(238);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '@') ADVANCE(237);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(306);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(247);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(247);
      if (lookahead == '<') ADVANCE(234);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(246);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(246);
      if (lookahead == '<') ADVANCE(234);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(60);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(anon_sym_SEMI);
      if (lookahead == ';') ADVANCE(148);
      END_STATE();
    case 216:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(anon_sym_COLON_GT);
      END_STATE();
    case 218:
      ACCEPT_TOKEN(anon_sym_LBRACK_GT);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(anon_sym_LBRACK_LT);
      END_STATE();
    case 220:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(321);
      END_STATE();
    case 223:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(322);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(319);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(anon_sym_LT);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '-') ADVANCE(232);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(anon_sym_POUND);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(anon_sym_POUND);
      if (lookahead == '!') ADVANCE(146);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(305);
      END_STATE();
    case 228:
      ACCEPT_TOKEN(anon_sym_POUND);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(305);
      END_STATE();
    case 229:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(anon_sym_LBRACK_PIPE);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(anon_sym_PIPE_RBRACK);
      END_STATE();
    case 232:
      ACCEPT_TOKEN(anon_sym_LT_DASH);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(anon_sym_LT_DASH);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(319);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(anon_sym_LBRACE_LT);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(anon_sym_GT_RBRACE);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(sym_ocamlyacc_value);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(236);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(aux_sym_attribute_token1);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(aux_sym_attribute_token1);
      if (lookahead == '@') ADVANCE(242);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(aux_sym_attribute_token1);
      if (lookahead == '@') ADVANCE(241);
      END_STATE();
    case 240:
      ACCEPT_TOKEN(aux_sym_attribute_token1);
      if (lookahead == '@') ADVANCE(87);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(anon_sym_LBRACK_AT_AT);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(anon_sym_LBRACK_AT_AT);
      if (lookahead == '@') ADVANCE(243);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(anon_sym_LBRACK_AT_AT_AT);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(anon_sym_LBRACK_PERCENT);
      END_STATE();
    case 245:
      ACCEPT_TOKEN(anon_sym_LBRACK_PERCENT);
      if (lookahead == '%') ADVANCE(249);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(anon_sym_LBRACE_PERCENT);
      END_STATE();
    case 247:
      ACCEPT_TOKEN(anon_sym_LBRACE_PERCENT);
      if (lookahead == '%') ADVANCE(250);
      END_STATE();
    case 248:
      ACCEPT_TOKEN(aux_sym_quoted_extension_token1);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(248);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(anon_sym_LBRACK_PERCENT_PERCENT);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(anon_sym_LBRACE_PERCENT_PERCENT);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(308);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(aux_sym_number_token1);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(aux_sym_number_token1);
      ADVANCE_MAP(
        '.', 258,
        'B', 99,
        'b', 99,
        'E', 98,
        'e', 98,
        'O', 261,
        'o', 261,
        'X', 264,
        'x', 264,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(255);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(253);
      END_STATE();
    case 255:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '.') ADVANCE(258);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(98);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(255);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(253);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '.') ADVANCE(259);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(257);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(256);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(253);
      END_STATE();
    case 257:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(106);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(263);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(98);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(258);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(253);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(257);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(259);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(253);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '0' ||
          lookahead == '1' ||
          lookahead == '_') ADVANCE(260);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(253);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(262);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (('0' <= lookahead && lookahead <= '7') ||
          lookahead == '_') ADVANCE(262);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(253);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(263);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(253);
      END_STATE();
    case 264:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(256);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      END_STATE();
    case 266:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(302);
      END_STATE();
    case 267:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(302);
      if (lookahead == '>') ADVANCE(192);
      END_STATE();
    case 268:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(304);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(310);
      END_STATE();
    case 269:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(303);
      if (lookahead == '>') ADVANCE(111);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(309);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(303);
      if (lookahead == '>') ADVANCE(193);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(309);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '>') ADVANCE(192);
      END_STATE();
    case 272:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(anon_sym_SQUOTE2);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(aux_sym_character_content_token1);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(aux_sym_character_content_token2);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(aux_sym_character_content_token2);
      if (lookahead == '@') ADVANCE(237);
      END_STATE();
    case 277:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 278:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      END_STATE();
    case 279:
      ACCEPT_TOKEN(aux_sym_string_content_token2);
      END_STATE();
    case 280:
      ACCEPT_TOKEN(aux_sym_string_content_token3);
      ADVANCE_MAP(
        '*', 76,
        '.', 70,
        'L', 296,
        'l', 296,
        'n', 296,
        ' ', 69,
        '#', 69,
        '+', 69,
        '-', 69,
        '0', 69,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(75);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(295);
      END_STATE();
    case 281:
      ACCEPT_TOKEN(aux_sym_string_content_token3);
      ADVANCE_MAP(
        '<', 107,
        '\\', 92,
        ' ', 297,
        ',', 297,
        '.', 297,
        ';', 297,
        '?', 297,
        '[', 297,
        ']', 297,
        '{', 297,
        '}', 297,
      );
      END_STATE();
    case 282:
      ACCEPT_TOKEN(aux_sym_string_content_token3);
      if (lookahead == '@') ADVANCE(279);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '%' &&
          lookahead != '\\') ADVANCE(283);
      END_STATE();
    case 283:
      ACCEPT_TOKEN(aux_sym_string_content_token3);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '%' &&
          lookahead != '@' &&
          lookahead != '\\') ADVANCE(283);
      END_STATE();
    case 284:
      ACCEPT_TOKEN(aux_sym_string_content_token4);
      END_STATE();
    case 285:
      ACCEPT_TOKEN(aux_sym_string_content_token5);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(285);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      END_STATE();
    case 287:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      ADVANCE_MAP(
        '*', 76,
        '.', 70,
        'L', 296,
        'l', 296,
        'n', 296,
        ' ', 69,
        '#', 69,
        '+', 69,
        '-', 69,
        '0', 69,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(75);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(295);
      END_STATE();
    case 288:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      ADVANCE_MAP(
        '<', 107,
        '\\', 92,
        ' ', 297,
        ',', 297,
        '.', 297,
        ';', 297,
        '?', 297,
        '[', 297,
        ']', 297,
        '{', 297,
        '}', 297,
      );
      END_STATE();
    case 289:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      if (lookahead == '@') ADVANCE(279);
      if (lookahead != 0 &&
          lookahead != '%' &&
          lookahead != '|') ADVANCE(290);
      END_STATE();
    case 290:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      if (lookahead != 0 &&
          lookahead != '%' &&
          lookahead != '@' &&
          lookahead != '|') ADVANCE(290);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(aux_sym_escape_sequence_token1);
      END_STATE();
    case 292:
      ACCEPT_TOKEN(aux_sym_escape_sequence_token2);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(aux_sym_escape_sequence_token3);
      END_STATE();
    case 294:
      ACCEPT_TOKEN(aux_sym_escape_sequence_token4);
      END_STATE();
    case 295:
      ACCEPT_TOKEN(sym_conversion_specification);
      END_STATE();
    case 296:
      ACCEPT_TOKEN(sym_conversion_specification);
      if (lookahead == 'X' ||
          lookahead == 'd' ||
          lookahead == 'i' ||
          lookahead == 'o' ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(295);
      END_STATE();
    case 297:
      ACCEPT_TOKEN(sym_pretty_printing_indication);
      END_STATE();
    case 298:
      ACCEPT_TOKEN(sym_pretty_printing_indication);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(313);
      END_STATE();
    case 299:
      ACCEPT_TOKEN(sym_prefix_operator);
      if (lookahead == '=') ADVANCE(113);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(301);
      END_STATE();
    case 300:
      ACCEPT_TOKEN(sym_prefix_operator);
      if (lookahead == '=') ADVANCE(320);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(301);
      END_STATE();
    case 301:
      ACCEPT_TOKEN(sym_prefix_operator);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(301);
      END_STATE();
    case 302:
      ACCEPT_TOKEN(aux_sym_sign_operator_token1);
      END_STATE();
    case 303:
      ACCEPT_TOKEN(aux_sym_sign_operator_token1);
      if (lookahead == '>') ADVANCE(310);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(309);
      END_STATE();
    case 304:
      ACCEPT_TOKEN(aux_sym_sign_operator_token1);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(310);
      END_STATE();
    case 305:
      ACCEPT_TOKEN(sym_hash_operator);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(305);
      END_STATE();
    case 306:
      ACCEPT_TOKEN(aux_sym_pow_operator_token1);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(306);
      END_STATE();
    case 307:
      ACCEPT_TOKEN(aux_sym_mult_operator_token1);
      if (lookahead == '*') ADVANCE(306);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(308);
      END_STATE();
    case 308:
      ACCEPT_TOKEN(aux_sym_mult_operator_token1);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(308);
      END_STATE();
    case 309:
      ACCEPT_TOKEN(aux_sym_add_operator_token1);
      if (lookahead == '>') ADVANCE(310);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(309);
      END_STATE();
    case 310:
      ACCEPT_TOKEN(aux_sym_add_operator_token1);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(310);
      END_STATE();
    case 311:
      ACCEPT_TOKEN(sym_concat_operator);
      ADVANCE_MAP(
        '<', 312,
        '\\', 92,
        '.', 298,
        '?', 298,
        ' ', 297,
        ',', 297,
        ';', 297,
        '[', 297,
        ']', 297,
        '{', 297,
        '}', 297,
      );
      if (lookahead == '!' ||
          ('$' <= lookahead && lookahead <= '&') ||
          ('*' <= lookahead && lookahead <= '/') ||
          (':' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          ('|' <= lookahead && lookahead <= '~')) ADVANCE(313);
      END_STATE();
    case 312:
      ACCEPT_TOKEN(sym_concat_operator);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(85);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(313);
      END_STATE();
    case 313:
      ACCEPT_TOKEN(sym_concat_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(313);
      END_STATE();
    case 314:
      ACCEPT_TOKEN(sym_rel_operator);
      END_STATE();
    case 315:
      ACCEPT_TOKEN(sym_rel_operator);
      if (lookahead == '-') ADVANCE(233);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(319);
      END_STATE();
    case 316:
      ACCEPT_TOKEN(sym_rel_operator);
      if (lookahead == '-') ADVANCE(110);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(319);
      END_STATE();
    case 317:
      ACCEPT_TOKEN(sym_rel_operator);
      if (lookahead == '}') ADVANCE(235);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(319);
      END_STATE();
    case 318:
      ACCEPT_TOKEN(sym_rel_operator);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(236);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(319);
      END_STATE();
    case 319:
      ACCEPT_TOKEN(sym_rel_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(319);
      END_STATE();
    case 320:
      ACCEPT_TOKEN(sym_rel_operator);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(301);
      END_STATE();
    case 321:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 322:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(319);
      END_STATE();
    case 323:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 324:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(319);
      END_STATE();
    case 325:
      ACCEPT_TOKEN(sym_assign_operator);
      END_STATE();
    case 326:
      ACCEPT_TOKEN(sym_indexing_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(326);
      END_STATE();
    case 327:
      ACCEPT_TOKEN(sym_let_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 328:
      ACCEPT_TOKEN(sym_let_and_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(328);
      END_STATE();
    case 329:
      ACCEPT_TOKEN(sym_match_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(329);
      END_STATE();
    case 330:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'a') ADVANCE(343);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 331:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'c') ADVANCE(338);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 332:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'd') ADVANCE(151);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 333:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'd') ADVANCE(152);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 334:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'd') ADVANCE(346);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 335:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'e') ADVANCE(342);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 336:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'e') ADVANCE(344);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 337:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'e') ADVANCE(345);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 338:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'h') ADVANCE(348);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 339:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'n') ADVANCE(332);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 340:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'n') ADVANCE(333);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 341:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'n') ADVANCE(334);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 342:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 't') ADVANCE(149);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 343:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 't') ADVANCE(331);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 344:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 't') ADVANCE(150);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 345:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 't') ADVANCE(347);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 346:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(328);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 347:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(327);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 348:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(329);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 349:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(349);
      END_STATE();
    case 350:
      ACCEPT_TOKEN(sym__uppercase_identifier);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(350);
      END_STATE();
    case 351:
      ACCEPT_TOKEN(aux_sym_directive_token1);
      END_STATE();
    case 352:
      ACCEPT_TOKEN(aux_sym_directive_token1);
      if (lookahead == '!') ADVANCE(147);
      END_STATE();
    case 353:
      ACCEPT_TOKEN(aux_sym_type_variable_token1);
      END_STATE();
    case 354:
      ACCEPT_TOKEN(aux_sym_tag_token1);
      END_STATE();
    case 355:
      ACCEPT_TOKEN(aux_sym_attribute_id_token1);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      ADVANCE_MAP(
        '_', 1,
        'a', 2,
        'b', 3,
        'c', 4,
        'd', 5,
        'e', 6,
        'f', 7,
        'i', 8,
        'l', 9,
        'm', 10,
        'n', 11,
        'o', 12,
        'p', 13,
        'r', 14,
        's', 15,
        't', 16,
        'v', 17,
        'w', 18,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(0);
      END_STATE();
    case 1:
      ACCEPT_TOKEN(anon_sym__);
      END_STATE();
    case 2:
      if (lookahead == 's') ADVANCE(19);
      END_STATE();
    case 3:
      if (lookahead == 'e') ADVANCE(20);
      END_STATE();
    case 4:
      if (lookahead == 'l') ADVANCE(21);
      if (lookahead == 'o') ADVANCE(22);
      END_STATE();
    case 5:
      if (lookahead == 'o') ADVANCE(23);
      END_STATE();
    case 6:
      if (lookahead == 'f') ADVANCE(24);
      if (lookahead == 'l') ADVANCE(25);
      if (lookahead == 'n') ADVANCE(26);
      if (lookahead == 'x') ADVANCE(27);
      END_STATE();
    case 7:
      if (lookahead == 'a') ADVANCE(28);
      if (lookahead == 'o') ADVANCE(29);
      if (lookahead == 'u') ADVANCE(30);
      END_STATE();
    case 8:
      if (lookahead == 'f') ADVANCE(31);
      if (lookahead == 'n') ADVANCE(32);
      END_STATE();
    case 9:
      if (lookahead == 'a') ADVANCE(33);
      if (lookahead == 'o') ADVANCE(34);
      if (lookahead == 's') ADVANCE(35);
      if (lookahead == 'x') ADVANCE(36);
      END_STATE();
    case 10:
      if (lookahead == 'a') ADVANCE(37);
      if (lookahead == 'e') ADVANCE(38);
      if (lookahead == 'o') ADVANCE(39);
      if (lookahead == 'u') ADVANCE(40);
      END_STATE();
    case 11:
      if (lookahead == 'e') ADVANCE(41);
      if (lookahead == 'o') ADVANCE(42);
      END_STATE();
    case 12:
      if (lookahead == 'b') ADVANCE(43);
      if (lookahead == 'f') ADVANCE(44);
      if (lookahead == 'p') ADVANCE(45);
      if (lookahead == 'r') ADVANCE(46);
      END_STATE();
    case 13:
      if (lookahead == 'r') ADVANCE(47);
      END_STATE();
    case 14:
      if (lookahead == 'e') ADVANCE(48);
      END_STATE();
    case 15:
      if (lookahead == 'i') ADVANCE(49);
      if (lookahead == 't') ADVANCE(50);
      END_STATE();
    case 16:
      if (lookahead == 'h') ADVANCE(51);
      if (lookahead == 'o') ADVANCE(52);
      if (lookahead == 'r') ADVANCE(53);
      if (lookahead == 'y') ADVANCE(54);
      END_STATE();
    case 17:
      if (lookahead == 'a') ADVANCE(55);
      if (lookahead == 'i') ADVANCE(56);
      END_STATE();
    case 18:
      if (lookahead == 'h') ADVANCE(57);
      if (lookahead == 'i') ADVANCE(58);
      END_STATE();
    case 19:
      ACCEPT_TOKEN(anon_sym_as);
      if (lookahead == 'r') ADVANCE(59);
      if (lookahead == 's') ADVANCE(60);
      END_STATE();
    case 20:
      if (lookahead == 'g') ADVANCE(61);
      END_STATE();
    case 21:
      if (lookahead == 'a') ADVANCE(62);
      END_STATE();
    case 22:
      if (lookahead == 'n') ADVANCE(63);
      END_STATE();
    case 23:
      ACCEPT_TOKEN(anon_sym_do);
      if (lookahead == 'n') ADVANCE(64);
      if (lookahead == 'w') ADVANCE(65);
      END_STATE();
    case 24:
      if (lookahead == 'f') ADVANCE(66);
      END_STATE();
    case 25:
      if (lookahead == 's') ADVANCE(67);
      END_STATE();
    case 26:
      if (lookahead == 'd') ADVANCE(68);
      END_STATE();
    case 27:
      if (lookahead == 'c') ADVANCE(69);
      if (lookahead == 't') ADVANCE(70);
      END_STATE();
    case 28:
      if (lookahead == 'l') ADVANCE(71);
      END_STATE();
    case 29:
      if (lookahead == 'r') ADVANCE(72);
      END_STATE();
    case 30:
      if (lookahead == 'n') ADVANCE(73);
      END_STATE();
    case 31:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 32:
      ACCEPT_TOKEN(anon_sym_in);
      if (lookahead == 'c') ADVANCE(74);
      if (lookahead == 'h') ADVANCE(75);
      if (lookahead == 'i') ADVANCE(76);
      END_STATE();
    case 33:
      if (lookahead == 'n') ADVANCE(77);
      if (lookahead == 'z') ADVANCE(78);
      END_STATE();
    case 34:
      if (lookahead == 'r') ADVANCE(79);
      END_STATE();
    case 35:
      if (lookahead == 'l') ADVANCE(80);
      if (lookahead == 'r') ADVANCE(81);
      END_STATE();
    case 36:
      if (lookahead == 'o') ADVANCE(82);
      END_STATE();
    case 37:
      if (lookahead == 't') ADVANCE(83);
      END_STATE();
    case 38:
      if (lookahead == 't') ADVANCE(84);
      END_STATE();
    case 39:
      if (lookahead == 'd') ADVANCE(85);
      END_STATE();
    case 40:
      if (lookahead == 't') ADVANCE(86);
      END_STATE();
    case 41:
      if (lookahead == 'w') ADVANCE(87);
      END_STATE();
    case 42:
      if (lookahead == 'n') ADVANCE(88);
      END_STATE();
    case 43:
      if (lookahead == 'j') ADVANCE(89);
      END_STATE();
    case 44:
      ACCEPT_TOKEN(anon_sym_of);
      END_STATE();
    case 45:
      if (lookahead == 'e') ADVANCE(90);
      END_STATE();
    case 46:
      ACCEPT_TOKEN(anon_sym_or);
      END_STATE();
    case 47:
      if (lookahead == 'i') ADVANCE(91);
      END_STATE();
    case 48:
      if (lookahead == 'c') ADVANCE(92);
      END_STATE();
    case 49:
      if (lookahead == 'g') ADVANCE(93);
      END_STATE();
    case 50:
      if (lookahead == 'r') ADVANCE(94);
      END_STATE();
    case 51:
      if (lookahead == 'e') ADVANCE(95);
      END_STATE();
    case 52:
      ACCEPT_TOKEN(anon_sym_to);
      END_STATE();
    case 53:
      if (lookahead == 'u') ADVANCE(96);
      if (lookahead == 'y') ADVANCE(97);
      END_STATE();
    case 54:
      if (lookahead == 'p') ADVANCE(98);
      END_STATE();
    case 55:
      if (lookahead == 'l') ADVANCE(99);
      END_STATE();
    case 56:
      if (lookahead == 'r') ADVANCE(100);
      END_STATE();
    case 57:
      if (lookahead == 'e') ADVANCE(101);
      if (lookahead == 'i') ADVANCE(102);
      END_STATE();
    case 58:
      if (lookahead == 't') ADVANCE(103);
      END_STATE();
    case 59:
      ACCEPT_TOKEN(anon_sym_asr);
      END_STATE();
    case 60:
      if (lookahead == 'e') ADVANCE(104);
      END_STATE();
    case 61:
      if (lookahead == 'i') ADVANCE(105);
      END_STATE();
    case 62:
      if (lookahead == 's') ADVANCE(106);
      END_STATE();
    case 63:
      if (lookahead == 's') ADVANCE(107);
      END_STATE();
    case 64:
      if (lookahead == 'e') ADVANCE(108);
      END_STATE();
    case 65:
      if (lookahead == 'n') ADVANCE(109);
      END_STATE();
    case 66:
      if (lookahead == 'e') ADVANCE(110);
      END_STATE();
    case 67:
      if (lookahead == 'e') ADVANCE(111);
      END_STATE();
    case 68:
      ACCEPT_TOKEN(anon_sym_end);
      END_STATE();
    case 69:
      if (lookahead == 'e') ADVANCE(112);
      END_STATE();
    case 70:
      if (lookahead == 'e') ADVANCE(113);
      END_STATE();
    case 71:
      if (lookahead == 's') ADVANCE(114);
      END_STATE();
    case 72:
      ACCEPT_TOKEN(anon_sym_for);
      END_STATE();
    case 73:
      ACCEPT_TOKEN(anon_sym_fun);
      if (lookahead == 'c') ADVANCE(115);
      END_STATE();
    case 74:
      if (lookahead == 'l') ADVANCE(116);
      END_STATE();
    case 75:
      if (lookahead == 'e') ADVANCE(117);
      END_STATE();
    case 76:
      if (lookahead == 't') ADVANCE(118);
      END_STATE();
    case 77:
      if (lookahead == 'd') ADVANCE(119);
      END_STATE();
    case 78:
      if (lookahead == 'y') ADVANCE(120);
      END_STATE();
    case 79:
      ACCEPT_TOKEN(anon_sym_lor);
      END_STATE();
    case 80:
      ACCEPT_TOKEN(anon_sym_lsl);
      END_STATE();
    case 81:
      ACCEPT_TOKEN(anon_sym_lsr);
      END_STATE();
    case 82:
      if (lookahead == 'r') ADVANCE(121);
      END_STATE();
    case 83:
      if (lookahead == 'c') ADVANCE(122);
      END_STATE();
    case 84:
      if (lookahead == 'h') ADVANCE(123);
      END_STATE();
    case 85:
      ACCEPT_TOKEN(anon_sym_mod);
      if (lookahead == 'u') ADVANCE(124);
      END_STATE();
    case 86:
      if (lookahead == 'a') ADVANCE(125);
      END_STATE();
    case 87:
      ACCEPT_TOKEN(anon_sym_new);
      END_STATE();
    case 88:
      if (lookahead == 'r') ADVANCE(126);
      END_STATE();
    case 89:
      if (lookahead == 'e') ADVANCE(127);
      END_STATE();
    case 90:
      if (lookahead == 'n') ADVANCE(128);
      END_STATE();
    case 91:
      if (lookahead == 'v') ADVANCE(129);
      END_STATE();
    case 92:
      ACCEPT_TOKEN(anon_sym_rec);
      END_STATE();
    case 93:
      ACCEPT_TOKEN(anon_sym_sig);
      END_STATE();
    case 94:
      if (lookahead == 'u') ADVANCE(130);
      END_STATE();
    case 95:
      if (lookahead == 'n') ADVANCE(131);
      END_STATE();
    case 96:
      if (lookahead == 'e') ADVANCE(132);
      END_STATE();
    case 97:
      ACCEPT_TOKEN(anon_sym_try);
      END_STATE();
    case 98:
      if (lookahead == 'e') ADVANCE(133);
      END_STATE();
    case 99:
      ACCEPT_TOKEN(anon_sym_val);
      END_STATE();
    case 100:
      if (lookahead == 't') ADVANCE(134);
      END_STATE();
    case 101:
      if (lookahead == 'n') ADVANCE(135);
      END_STATE();
    case 102:
      if (lookahead == 'l') ADVANCE(136);
      END_STATE();
    case 103:
      if (lookahead == 'h') ADVANCE(137);
      END_STATE();
    case 104:
      if (lookahead == 'r') ADVANCE(138);
      END_STATE();
    case 105:
      if (lookahead == 'n') ADVANCE(139);
      END_STATE();
    case 106:
      if (lookahead == 's') ADVANCE(140);
      END_STATE();
    case 107:
      if (lookahead == 't') ADVANCE(141);
      END_STATE();
    case 108:
      ACCEPT_TOKEN(anon_sym_done);
      END_STATE();
    case 109:
      if (lookahead == 't') ADVANCE(142);
      END_STATE();
    case 110:
      if (lookahead == 'c') ADVANCE(143);
      END_STATE();
    case 111:
      ACCEPT_TOKEN(anon_sym_else);
      END_STATE();
    case 112:
      if (lookahead == 'p') ADVANCE(144);
      END_STATE();
    case 113:
      if (lookahead == 'r') ADVANCE(145);
      END_STATE();
    case 114:
      if (lookahead == 'e') ADVANCE(146);
      END_STATE();
    case 115:
      if (lookahead == 't') ADVANCE(147);
      END_STATE();
    case 116:
      if (lookahead == 'u') ADVANCE(148);
      END_STATE();
    case 117:
      if (lookahead == 'r') ADVANCE(149);
      END_STATE();
    case 118:
      if (lookahead == 'i') ADVANCE(150);
      END_STATE();
    case 119:
      ACCEPT_TOKEN(anon_sym_land);
      END_STATE();
    case 120:
      ACCEPT_TOKEN(anon_sym_lazy);
      END_STATE();
    case 121:
      ACCEPT_TOKEN(anon_sym_lxor);
      END_STATE();
    case 122:
      if (lookahead == 'h') ADVANCE(151);
      END_STATE();
    case 123:
      if (lookahead == 'o') ADVANCE(152);
      END_STATE();
    case 124:
      if (lookahead == 'l') ADVANCE(153);
      END_STATE();
    case 125:
      if (lookahead == 'b') ADVANCE(154);
      END_STATE();
    case 126:
      if (lookahead == 'e') ADVANCE(155);
      END_STATE();
    case 127:
      if (lookahead == 'c') ADVANCE(156);
      END_STATE();
    case 128:
      ACCEPT_TOKEN(anon_sym_open);
      END_STATE();
    case 129:
      if (lookahead == 'a') ADVANCE(157);
      END_STATE();
    case 130:
      if (lookahead == 'c') ADVANCE(158);
      END_STATE();
    case 131:
      ACCEPT_TOKEN(anon_sym_then);
      END_STATE();
    case 132:
      ACCEPT_TOKEN(anon_sym_true);
      END_STATE();
    case 133:
      ACCEPT_TOKEN(anon_sym_type);
      END_STATE();
    case 134:
      if (lookahead == 'u') ADVANCE(159);
      END_STATE();
    case 135:
      ACCEPT_TOKEN(anon_sym_when);
      END_STATE();
    case 136:
      if (lookahead == 'e') ADVANCE(160);
      END_STATE();
    case 137:
      ACCEPT_TOKEN(anon_sym_with);
      END_STATE();
    case 138:
      if (lookahead == 't') ADVANCE(161);
      END_STATE();
    case 139:
      ACCEPT_TOKEN(anon_sym_begin);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(anon_sym_class);
      END_STATE();
    case 141:
      if (lookahead == 'r') ADVANCE(162);
      END_STATE();
    case 142:
      if (lookahead == 'o') ADVANCE(163);
      END_STATE();
    case 143:
      if (lookahead == 't') ADVANCE(164);
      END_STATE();
    case 144:
      if (lookahead == 't') ADVANCE(165);
      END_STATE();
    case 145:
      if (lookahead == 'n') ADVANCE(166);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(anon_sym_false);
      END_STATE();
    case 147:
      if (lookahead == 'i') ADVANCE(167);
      if (lookahead == 'o') ADVANCE(168);
      END_STATE();
    case 148:
      if (lookahead == 'd') ADVANCE(169);
      END_STATE();
    case 149:
      if (lookahead == 'i') ADVANCE(170);
      END_STATE();
    case 150:
      if (lookahead == 'a') ADVANCE(171);
      END_STATE();
    case 151:
      ACCEPT_TOKEN(anon_sym_match);
      END_STATE();
    case 152:
      if (lookahead == 'd') ADVANCE(172);
      END_STATE();
    case 153:
      if (lookahead == 'e') ADVANCE(173);
      END_STATE();
    case 154:
      if (lookahead == 'l') ADVANCE(174);
      END_STATE();
    case 155:
      if (lookahead == 'c') ADVANCE(175);
      END_STATE();
    case 156:
      if (lookahead == 't') ADVANCE(176);
      END_STATE();
    case 157:
      if (lookahead == 't') ADVANCE(177);
      END_STATE();
    case 158:
      if (lookahead == 't') ADVANCE(178);
      END_STATE();
    case 159:
      if (lookahead == 'a') ADVANCE(179);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(anon_sym_while);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(anon_sym_assert);
      END_STATE();
    case 162:
      if (lookahead == 'a') ADVANCE(180);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(anon_sym_downto);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(anon_sym_effect);
      END_STATE();
    case 165:
      if (lookahead == 'i') ADVANCE(181);
      END_STATE();
    case 166:
      if (lookahead == 'a') ADVANCE(182);
      END_STATE();
    case 167:
      if (lookahead == 'o') ADVANCE(183);
      END_STATE();
    case 168:
      if (lookahead == 'r') ADVANCE(184);
      END_STATE();
    case 169:
      if (lookahead == 'e') ADVANCE(185);
      END_STATE();
    case 170:
      if (lookahead == 't') ADVANCE(186);
      END_STATE();
    case 171:
      if (lookahead == 'l') ADVANCE(187);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_method);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(anon_sym_module);
      END_STATE();
    case 174:
      if (lookahead == 'e') ADVANCE(188);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(anon_sym_nonrec);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(anon_sym_object);
      END_STATE();
    case 177:
      if (lookahead == 'e') ADVANCE(189);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_struct);
      END_STATE();
    case 179:
      if (lookahead == 'l') ADVANCE(190);
      END_STATE();
    case 180:
      if (lookahead == 'i') ADVANCE(191);
      END_STATE();
    case 181:
      if (lookahead == 'o') ADVANCE(192);
      END_STATE();
    case 182:
      if (lookahead == 'l') ADVANCE(193);
      END_STATE();
    case 183:
      if (lookahead == 'n') ADVANCE(194);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_functor);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(anon_sym_include);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_inherit);
      END_STATE();
    case 187:
      if (lookahead == 'i') ADVANCE(195);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_mutable);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_private);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_virtual);
      END_STATE();
    case 191:
      if (lookahead == 'n') ADVANCE(196);
      END_STATE();
    case 192:
      if (lookahead == 'n') ADVANCE(197);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_external);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_function);
      END_STATE();
    case 195:
      if (lookahead == 'z') ADVANCE(198);
      END_STATE();
    case 196:
      if (lookahead == 't') ADVANCE(199);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_exception);
      END_STATE();
    case 198:
      if (lookahead == 'e') ADVANCE(200);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_constraint);
      END_STATE();
    case 200:
      if (lookahead == 'r') ADVANCE(201);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_initializer);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2] = {.lex_state = 127, .external_lex_state = 2, .reserved_word_set_id = 2},
  [3] = {.lex_state = 127, .external_lex_state = 2, .reserved_word_set_id = 3},
  [4] = {.lex_state = 127, .external_lex_state = 2, .reserved_word_set_id = 4},
  [5] = {.lex_state = 127, .external_lex_state = 2, .reserved_word_set_id = 5},
  [6] = {.lex_state = 129, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7] = {.lex_state = 129, .external_lex_state = 2, .reserved_word_set_id = 7},
  [8] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [9] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [10] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [11] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [12] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [13] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [14] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [15] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [16] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [17] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [18] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [19] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [20] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [21] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [22] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [23] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [24] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [25] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [26] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [27] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [28] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [29] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [30] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [31] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [32] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [33] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [34] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [35] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [36] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [37] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [38] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [39] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [40] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [41] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [42] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 9},
  [43] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 9},
  [44] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 9},
  [45] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 9},
  [46] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 9},
  [47] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 9},
  [48] = {.lex_state = 26, .external_lex_state = 2, .reserved_word_set_id = 10},
  [49] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [50] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [51] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [52] = {.lex_state = 26, .external_lex_state = 2, .reserved_word_set_id = 11},
  [53] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [54] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [55] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 8},
  [56] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [57] = {.lex_state = 26, .external_lex_state = 2, .reserved_word_set_id = 12},
  [58] = {.lex_state = 26, .external_lex_state = 2, .reserved_word_set_id = 11},
  [59] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [60] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [61] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [62] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [63] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [64] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [65] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [66] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [67] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [68] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [69] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [70] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [71] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [72] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [73] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [74] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [75] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [76] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [77] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [78] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [79] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [80] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [81] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [82] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [83] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [84] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [85] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [86] = {.lex_state = 32, .external_lex_state = 2, .reserved_word_set_id = 13},
  [87] = {.lex_state = 34, .external_lex_state = 2, .reserved_word_set_id = 14},
  [88] = {.lex_state = 34, .external_lex_state = 2, .reserved_word_set_id = 14},
  [89] = {.lex_state = 26, .external_lex_state = 2, .reserved_word_set_id = 15},
  [90] = {.lex_state = 26, .external_lex_state = 2, .reserved_word_set_id = 15},
  [91] = {.lex_state = 33, .external_lex_state = 2, .reserved_word_set_id = 16},
  [92] = {.lex_state = 26, .external_lex_state = 2, .reserved_word_set_id = 17},
  [93] = {.lex_state = 121, .external_lex_state = 2, .reserved_word_set_id = 18},
  [94] = {.lex_state = 121, .external_lex_state = 2, .reserved_word_set_id = 19},
  [95] = {.lex_state = 121, .external_lex_state = 2, .reserved_word_set_id = 1},
  [96] = {.lex_state = 121, .external_lex_state = 2, .reserved_word_set_id = 1},
  [97] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 20},
  [98] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 9},
  [99] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [100] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [101] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [102] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 20},
  [103] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [104] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [105] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [106] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [107] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 24},
  [108] = {.lex_state = 130, .external_lex_state = 2, .reserved_word_set_id = 22},
  [109] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [110] = {.lex_state = 130, .external_lex_state = 2, .reserved_word_set_id = 22},
  [111] = {.lex_state = 130, .external_lex_state = 2, .reserved_word_set_id = 23},
  [112] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 25},
  [113] = {.lex_state = 130, .external_lex_state = 2, .reserved_word_set_id = 23},
  [114] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 26},
  [115] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 27},
  [116] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 28},
  [117] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 24},
  [118] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 25},
  [119] = {.lex_state = 121, .external_lex_state = 2, .reserved_word_set_id = 1},
  [120] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 29},
  [121] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [122] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [123] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [124] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [125] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [126] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [127] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [128] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 31},
  [129] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [130] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [131] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [132] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [133] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [134] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [135] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 31},
  [136] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [137] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [138] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [139] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 31},
  [140] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 31},
  [141] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 31},
  [142] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [143] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 31},
  [144] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [145] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [146] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [147] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [148] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [149] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [150] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [151] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [152] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [153] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [154] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [155] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [156] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [157] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [158] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [159] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [160] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [161] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [162] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [163] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [164] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [165] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [166] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [167] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [168] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [169] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [170] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [171] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [172] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [173] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [174] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [175] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [176] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [177] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [178] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [179] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [180] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [181] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [182] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [183] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [184] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [185] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [186] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [187] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [188] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [189] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [190] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [191] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [192] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [193] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [194] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [195] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [196] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [197] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [198] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [199] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [200] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [201] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [202] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [203] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [204] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [205] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [206] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [207] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [208] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [209] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [210] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [211] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [212] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [213] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [214] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [215] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [216] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [217] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [218] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 33},
  [219] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 33},
  [220] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 34},
  [221] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [222] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [223] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 34},
  [224] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 33},
  [225] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [226] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [227] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [228] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [229] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [230] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [231] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [232] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [233] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [234] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 34},
  [235] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [236] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [237] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [238] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [239] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [240] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [241] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [242] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [243] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [244] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [245] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [246] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [247] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 34},
  [248] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 34},
  [249] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 34},
  [250] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [251] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [252] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 34},
  [253] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [254] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 34},
  [255] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [256] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [257] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 33},
  [258] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 34},
  [259] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [260] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [261] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [262] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [263] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [264] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [265] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [266] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [267] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [268] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [269] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [270] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [271] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [272] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [273] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [274] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [275] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [276] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [277] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [278] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [279] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [280] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [281] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [282] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [283] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [284] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [285] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [286] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [287] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [288] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [289] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [290] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [291] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [292] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [293] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [294] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [295] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [296] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [297] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [298] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [299] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [300] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [301] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [302] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [303] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [304] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [305] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [306] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [307] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [308] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [309] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [310] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [311] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [312] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [313] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [314] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [315] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [316] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [317] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [318] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [319] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [320] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [321] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [322] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [323] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [324] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [325] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [326] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [327] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [328] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [329] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [330] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [331] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [332] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [333] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [334] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [335] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [336] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [337] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [338] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [339] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [340] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [341] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [342] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [343] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [344] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [345] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [346] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [347] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [348] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [349] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [350] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [351] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [352] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [353] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [354] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [355] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [356] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [357] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [358] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [359] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [360] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [361] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [362] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [363] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [364] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [365] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [366] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [367] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [368] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [369] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [370] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [371] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [372] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [373] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [374] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [375] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [376] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [377] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [378] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [379] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [380] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [381] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [382] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [383] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [384] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [385] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [386] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [387] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [388] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [389] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [390] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [391] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [392] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [393] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [394] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [395] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [396] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [397] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [398] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [399] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [400] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [401] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [402] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [403] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [404] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [405] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [406] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [407] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [408] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [409] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [410] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [411] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [412] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [413] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [414] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [415] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [416] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [417] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [418] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [419] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [420] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [421] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [422] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [423] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [424] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [425] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [426] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [427] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [428] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [429] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [430] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [431] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [432] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [433] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [434] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [435] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [436] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [437] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [438] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [439] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [440] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [441] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [442] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [443] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [444] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [445] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [446] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [447] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [448] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [449] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [450] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [451] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [452] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [453] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [454] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [455] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [456] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [457] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [458] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [459] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [460] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [461] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [462] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [463] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [464] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [465] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [466] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [467] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [468] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [469] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [470] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [471] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [472] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [473] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [474] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [475] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [476] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [477] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [478] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [479] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [480] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [481] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [482] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [483] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [484] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [485] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [486] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [487] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [488] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [489] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [490] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [491] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [492] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [493] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [494] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [495] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [496] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [497] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [498] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [499] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [500] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [501] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [502] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [503] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [504] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [505] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [506] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [507] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [508] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [509] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [510] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [511] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [512] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [513] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [514] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [515] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [516] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [517] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [518] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [519] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [520] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [521] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [522] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [523] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [524] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [525] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [526] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [527] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [528] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [529] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [530] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [531] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [532] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [533] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [534] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [535] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [536] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [537] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [538] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [539] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [540] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [541] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [542] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [543] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [544] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [545] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [546] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [547] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [548] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [549] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [550] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [551] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [552] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [553] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [554] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [555] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [556] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [557] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [558] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [559] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [560] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [561] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [562] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [563] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [564] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [565] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [566] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [567] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [568] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [569] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [570] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [571] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [572] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [573] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [574] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [575] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [576] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [577] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [578] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [579] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [580] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [581] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [582] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [583] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [584] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [585] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [586] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [587] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [588] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [589] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [590] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [591] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [592] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [593] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [594] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [595] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [596] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [597] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [598] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [599] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [600] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [601] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [602] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [603] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [604] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [605] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [606] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [607] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [608] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [609] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [610] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [611] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [612] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [613] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [614] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [615] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [616] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [617] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [618] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [619] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [620] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [621] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [622] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [623] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [624] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [625] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [626] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [627] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [628] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [629] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [630] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [631] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [632] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [633] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [634] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [635] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [636] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [637] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [638] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [639] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [640] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [641] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [642] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [643] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [644] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [645] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [646] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [647] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [648] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [649] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [650] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [651] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [652] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [653] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [654] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [655] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [656] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [657] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [658] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [659] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [660] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [661] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [662] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [663] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [664] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [665] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [666] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [667] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [668] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [669] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [670] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [671] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [672] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [673] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [674] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [675] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [676] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [677] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [678] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [679] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [680] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [681] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [682] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [683] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [684] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [685] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [686] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [687] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [688] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [689] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [690] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [691] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [692] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [693] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [694] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [695] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [696] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [697] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [698] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [699] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [700] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [701] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [702] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [703] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [704] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [705] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [706] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [707] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [708] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [709] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [710] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [711] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [712] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [713] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [714] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [715] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [716] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [717] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [718] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [719] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [720] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [721] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [722] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [723] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [724] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [725] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 35},
  [726] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [727] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [728] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [729] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [730] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [731] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [732] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [733] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [734] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [735] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [736] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [737] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [738] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [739] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [740] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [741] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [742] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [743] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [744] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [745] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [746] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [747] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [748] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [749] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [750] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [751] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [752] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [753] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [754] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [755] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [756] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [757] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [758] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [759] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [760] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [761] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [762] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [763] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [764] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [765] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [766] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [767] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [768] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [769] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [770] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [771] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [772] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [773] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [774] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [775] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [776] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [777] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [778] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [779] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [780] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [781] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [782] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [783] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [784] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [785] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [786] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [787] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [788] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [789] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [790] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [791] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [792] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [793] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [794] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [795] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [796] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [797] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [798] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [799] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [800] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [801] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [802] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [803] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [804] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [805] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [806] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [807] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [808] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [809] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [810] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [811] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [812] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [813] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [814] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [815] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [816] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [817] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [818] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [819] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [820] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [821] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [822] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [823] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [824] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [825] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [826] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [827] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [828] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [829] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [830] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [831] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [832] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [833] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [834] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [835] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [836] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [837] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [838] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [839] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [840] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [841] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [842] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [843] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [844] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [845] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [846] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [847] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [848] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [849] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [850] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [851] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [852] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [853] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [854] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [855] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [856] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [857] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [858] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [859] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [860] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [861] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [862] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [863] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [864] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [865] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [866] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [867] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [868] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [869] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [870] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [871] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [872] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [873] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [874] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [875] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [876] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [877] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [878] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [879] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [880] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [881] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [882] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 32},
  [883] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 35},
  [884] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [885] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [886] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [887] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [888] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [889] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [890] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [891] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [892] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [893] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [894] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [895] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [896] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [897] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [898] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [899] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [900] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [901] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [902] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [903] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [904] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [905] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [906] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [907] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [908] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [909] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [910] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 1},
  [911] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [912] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [913] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [914] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [915] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [916] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [917] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [918] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 39},
  [919] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [920] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [921] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [922] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [923] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [924] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [925] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 37},
  [926] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 1},
  [927] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [928] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [929] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [930] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [931] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 1},
  [932] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [933] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 1},
  [934] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [935] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [936] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [937] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [938] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [939] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [940] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [941] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [942] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [943] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [944] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [945] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [946] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [947] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [948] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [949] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [950] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [951] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [952] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [953] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [954] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [955] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [956] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [957] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [958] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [959] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [960] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [961] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [962] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [963] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [964] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [965] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [966] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [967] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [968] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [969] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [970] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [971] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [972] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [973] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [974] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [975] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [976] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [977] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [978] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [979] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [980] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [981] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [982] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [983] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [984] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [985] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [986] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [987] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [988] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [989] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [990] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [991] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [992] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [993] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [994] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [995] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [996] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [997] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [998] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [999] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1000] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1001] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1002] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1003] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1004] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1005] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1006] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1007] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1008] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1009] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1010] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1011] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1012] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 41},
  [1013] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 41},
  [1014] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 41},
  [1015] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 42},
  [1016] = {.lex_state = 35, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1017] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1018] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1019] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 42},
  [1020] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [1021] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [1022] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [1023] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1024] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [1025] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [1026] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1027] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [1028] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [1029] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [1030] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 21},
  [1031] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1032] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1033] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1034] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1035] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1036] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1037] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1038] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1039] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [1040] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [1041] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1042] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1043] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1044] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1045] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1046] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1047] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1048] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1049] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1050] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1051] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1052] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1053] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1054] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1055] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1056] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1057] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1058] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1059] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1060] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1061] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1062] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1063] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1064] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1065] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1066] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1067] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1068] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1069] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1070] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1071] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1072] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1073] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1074] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1075] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1076] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1077] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1078] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1079] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1080] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1081] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1082] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1083] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1084] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1085] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1086] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1087] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1088] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1089] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 44},
  [1090] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1091] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1092] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1093] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1094] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1095] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1096] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1097] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1098] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1099] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1100] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1101] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1102] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1103] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1104] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1105] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1106] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1107] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1108] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1109] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1110] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1111] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1112] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1113] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1114] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1115] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1116] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1117] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1118] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1119] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1120] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1121] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1122] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1123] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1124] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1125] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1126] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 44},
  [1127] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1128] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1129] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1130] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1131] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1132] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1133] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1134] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [1135] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1136] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1137] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1138] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1139] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1140] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1141] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1142] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1143] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1144] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1145] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1146] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1147] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1148] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1149] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1150] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1151] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1152] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1153] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1154] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1155] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1156] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1157] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1158] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1159] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1160] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1161] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1162] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1163] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1164] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1165] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1166] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1167] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1168] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1169] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1170] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1171] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1172] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1173] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1174] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1175] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1176] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1177] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1178] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1179] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1180] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1181] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1182] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1183] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1184] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1185] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1186] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1187] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1188] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1189] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1190] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1191] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1192] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1193] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [1194] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1195] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1196] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1197] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1198] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1199] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1200] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1201] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1202] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1203] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1204] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1205] = {.lex_state = 37, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1206] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1207] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1208] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1209] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [1210] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [1211] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [1212] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1213] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [1214] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1215] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1216] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1217] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1218] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1219] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [1220] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1221] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [1222] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [1223] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1224] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1225] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1226] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1227] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1228] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1229] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [1230] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [1231] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1232] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1233] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1234] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1235] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1236] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1237] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1238] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1239] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1240] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1241] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1242] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1243] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1244] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1245] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1246] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1247] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1248] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1249] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1250] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1251] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1252] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1253] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1254] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1255] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1256] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1257] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1258] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1259] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1260] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1261] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1262] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1263] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1264] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 20},
  [1265] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1266] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1267] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1268] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1269] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1270] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1271] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1272] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1273] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1274] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1275] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1276] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1277] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1278] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1279] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1280] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1281] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1282] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1283] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1284] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1285] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1286] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1287] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1288] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1289] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1290] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1291] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1292] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1293] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1294] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1295] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1296] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1297] = {.lex_state = 40, .external_lex_state = 2, .reserved_word_set_id = 46},
  [1298] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1299] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1300] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1301] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1302] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1303] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1304] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1305] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1306] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1307] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1308] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1309] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1310] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1311] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1312] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1313] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1314] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1315] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1316] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1317] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1318] = {.lex_state = 40, .external_lex_state = 2, .reserved_word_set_id = 46},
  [1319] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1320] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1321] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1322] = {.lex_state = 134, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1323] = {.lex_state = 130, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1324] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1325] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1326] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1327] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1328] = {.lex_state = 130, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1329] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1330] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1331] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1332] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1333] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1334] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1335] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1336] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1337] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1338] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1339] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1340] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1341] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1342] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1343] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1344] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1345] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1346] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1347] = {.lex_state = 42, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1348] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1349] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1350] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1351] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1352] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1353] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1354] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1355] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1356] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1357] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1358] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1359] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1360] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1361] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1362] = {.lex_state = 40, .external_lex_state = 2, .reserved_word_set_id = 46},
  [1363] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1364] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1365] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1366] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1367] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1368] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1369] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1370] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1371] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1372] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1373] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1374] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1375] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1376] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1377] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1378] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1379] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1380] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1381] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1382] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1383] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 24},
  [1384] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1385] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1386] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 24},
  [1387] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1388] = {.lex_state = 42, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1389] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 46},
  [1390] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1391] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1392] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1393] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1394] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1395] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1396] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1397] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1398] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1399] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1400] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1401] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1402] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1403] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1404] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1405] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1406] = {.lex_state = 40, .external_lex_state = 2, .reserved_word_set_id = 46},
  [1407] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1408] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1409] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1410] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1411] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1412] = {.lex_state = 130, .external_lex_state = 2, .reserved_word_set_id = 22},
  [1413] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1414] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1415] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1416] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 25},
  [1417] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 25},
  [1418] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1419] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1420] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1421] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1422] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1423] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1424] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1425] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1426] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1427] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1428] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1429] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1430] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1431] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1432] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1433] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1434] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1435] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1436] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1437] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1438] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1439] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1440] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1441] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1442] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1443] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1444] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1445] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1446] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1447] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 46},
  [1448] = {.lex_state = 42, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1449] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1450] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1451] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1452] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1453] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1454] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1455] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1456] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1457] = {.lex_state = 42, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1458] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1459] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1460] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1461] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1462] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1463] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1464] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1465] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1466] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1467] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1468] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1469] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1470] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1471] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1472] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1473] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1474] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1475] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1476] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1477] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [1478] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1479] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1480] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1481] = {.lex_state = 130, .external_lex_state = 2, .reserved_word_set_id = 23},
  [1482] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1483] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1484] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1485] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1486] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1487] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1488] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1489] = {.lex_state = 135, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1490] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1491] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1492] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1493] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1494] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1495] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1496] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 27},
  [1497] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1498] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1499] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1500] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1501] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1502] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1503] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1504] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1505] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1506] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 48},
  [1507] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 28},
  [1508] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1509] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1510] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1511] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1512] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1513] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1514] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1515] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1516] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1517] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1518] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1519] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1520] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1521] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1522] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1523] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1524] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1525] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1526] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1527] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1528] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1529] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1530] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1531] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1532] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1533] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1534] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1535] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1536] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1537] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1538] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1539] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1540] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1541] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1542] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1543] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1544] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1545] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1546] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1547] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1548] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1549] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1550] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1551] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1552] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1553] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1554] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1555] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1556] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1557] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1558] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1559] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1560] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1561] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1562] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1563] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1564] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1565] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1566] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1567] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1568] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1569] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1570] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1571] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1572] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1573] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1574] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1575] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1576] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1577] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1578] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1579] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1580] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1581] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1582] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1583] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1584] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1585] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1586] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1587] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1588] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1589] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1590] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1591] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1592] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1593] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1594] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1595] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1596] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1597] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1598] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 49},
  [1599] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1600] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1601] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1602] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1603] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1604] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1605] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1606] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1607] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1608] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1609] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1610] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1611] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1612] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1613] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1614] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1615] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1616] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1617] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1618] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1619] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1620] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1621] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1622] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1623] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1624] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1625] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1626] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1627] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 49},
  [1628] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1629] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 49},
  [1630] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1631] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1632] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1633] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1634] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1635] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1636] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1637] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1638] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1639] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1640] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1641] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1642] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1643] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1644] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1645] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1646] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1647] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1648] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1649] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1650] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1651] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1652] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1653] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1654] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1655] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1656] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1657] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1658] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1659] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1660] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1661] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1662] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1663] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1664] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1665] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1666] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1667] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1668] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1669] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1670] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1671] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1672] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1673] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1674] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1675] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1676] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1677] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1678] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1679] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1680] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1681] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1682] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1683] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1684] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1685] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1686] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1687] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1688] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1689] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1690] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1691] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1692] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1693] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1694] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1695] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1696] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1697] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1698] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1699] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1700] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1701] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1702] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1703] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1704] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1705] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1706] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1707] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1708] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1709] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1710] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1711] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1712] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1713] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1714] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1715] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1716] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 50},
  [1717] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1718] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1719] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1720] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1721] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1722] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1723] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1724] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1725] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1726] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1727] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1728] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1729] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1730] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1731] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1732] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1733] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 51},
  [1734] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1735] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1736] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1737] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1738] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1739] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1740] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1741] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1742] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1743] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1744] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1745] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1746] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1747] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1748] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1749] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1750] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1751] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1752] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1753] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1754] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1755] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1756] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1757] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1758] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1759] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1760] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1761] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1762] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1763] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1764] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1765] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1766] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1767] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1768] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1769] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1770] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1771] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 47},
  [1772] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1773] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 52},
  [1774] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 52},
  [1775] = {.lex_state = 28, .external_lex_state = 4, .reserved_word_set_id = 53},
  [1776] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1777] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1778] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1779] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1780] = {.lex_state = 28, .external_lex_state = 4, .reserved_word_set_id = 53},
  [1781] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1782] = {.lex_state = 28, .external_lex_state = 4, .reserved_word_set_id = 53},
  [1783] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1784] = {.lex_state = 28, .external_lex_state = 4, .reserved_word_set_id = 53},
  [1785] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1786] = {.lex_state = 28, .external_lex_state = 4, .reserved_word_set_id = 53},
  [1787] = {.lex_state = 28, .external_lex_state = 4, .reserved_word_set_id = 53},
  [1788] = {.lex_state = 28, .external_lex_state = 4, .reserved_word_set_id = 53},
  [1789] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1790] = {.lex_state = 28, .external_lex_state = 4, .reserved_word_set_id = 53},
  [1791] = {.lex_state = 28, .external_lex_state = 4, .reserved_word_set_id = 53},
  [1792] = {.lex_state = 28, .external_lex_state = 4, .reserved_word_set_id = 53},
  [1793] = {.lex_state = 28, .external_lex_state = 4, .reserved_word_set_id = 53},
  [1794] = {.lex_state = 28, .external_lex_state = 4, .reserved_word_set_id = 53},
  [1795] = {.lex_state = 28, .external_lex_state = 4, .reserved_word_set_id = 53},
  [1796] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1797] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1798] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 51},
  [1799] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1800] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 55},
  [1801] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1802] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1803] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1804] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1805] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1806] = {.lex_state = 28, .external_lex_state = 4, .reserved_word_set_id = 53},
  [1807] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1808] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1809] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1810] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1811] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1812] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1813] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1814] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1815] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1816] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1817] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1818] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1819] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1820] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1821] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1822] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1823] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1824] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [1825] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1826] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1827] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1828] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1829] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1830] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1831] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1832] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1833] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1834] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1835] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1836] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1837] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1838] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1839] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1840] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1841] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1842] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1843] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1844] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1845] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1846] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1847] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1848] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1849] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1850] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1851] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1852] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1853] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1854] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1855] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1856] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1857] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1858] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1859] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1860] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1861] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1862] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1863] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1864] = {.lex_state = 123, .external_lex_state = 3},
  [1865] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1866] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1867] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1868] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1869] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1870] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1871] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1872] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1873] = {.lex_state = 123, .external_lex_state = 3},
  [1874] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1875] = {.lex_state = 123, .external_lex_state = 3},
  [1876] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1877] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1878] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1879] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1880] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1881] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1882] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1883] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1884] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1885] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1886] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1887] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1888] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1889] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1890] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1891] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1892] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1893] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1894] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1895] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1896] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1897] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1898] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1899] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1900] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1901] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1902] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1903] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1904] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1905] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1906] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1907] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1908] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1909] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1910] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1911] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1912] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1913] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1914] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1915] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1916] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1917] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1918] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1919] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1920] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1921] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1922] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1923] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1924] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1925] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1926] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1927] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1928] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 53},
  [1929] = {.lex_state = 12, .external_lex_state = 2, .reserved_word_set_id = 9},
  [1930] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1931] = {.lex_state = 12, .external_lex_state = 2, .reserved_word_set_id = 9},
  [1932] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1933] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1934] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [1935] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1936] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [1937] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1938] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1939] = {.lex_state = 12, .external_lex_state = 2, .reserved_word_set_id = 9},
  [1940] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1941] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1942] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1943] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1944] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1945] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1946] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1947] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1948] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1949] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1950] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1951] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1952] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1953] = {.lex_state = 27, .external_lex_state = 2, .reserved_word_set_id = 57},
  [1954] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1955] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1956] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1957] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1958] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1959] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1960] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1961] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1962] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1963] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1964] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1965] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1966] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1967] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1968] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1969] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1970] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1971] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1972] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1973] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1974] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1975] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1976] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1977] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1978] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1979] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1980] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1981] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1982] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1983] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1984] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1985] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1986] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1987] = {.lex_state = 12, .external_lex_state = 2, .reserved_word_set_id = 9},
  [1988] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1989] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1990] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1991] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1992] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1993] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1994] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1995] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1996] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1997] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1998] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [1999] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2000] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2001] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2002] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2003] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2004] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2005] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2006] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2007] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2008] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2009] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2010] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2011] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2012] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2013] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2014] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2015] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2016] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2017] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2018] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2019] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2020] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2021] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2022] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2023] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2024] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2025] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2026] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2027] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2028] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2029] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2030] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2031] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2032] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2033] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2034] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2035] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2036] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 9},
  [2037] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2038] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2039] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2040] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2041] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2042] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2043] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 57},
  [2044] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2045] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2046] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2047] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2048] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2049] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2050] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2051] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2052] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2053] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2054] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2055] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2056] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2057] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2058] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2059] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2060] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 35},
  [2061] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2062] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2063] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2064] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2065] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2066] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2067] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2068] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2069] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2070] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2071] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2072] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 35},
  [2073] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2074] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 58},
  [2075] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2076] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 9},
  [2077] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2078] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2079] = {.lex_state = 123, .external_lex_state = 3},
  [2080] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2081] = {.lex_state = 29, .external_lex_state = 2, .reserved_word_set_id = 36},
  [2082] = {.lex_state = 123, .external_lex_state = 3},
  [2083] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2084] = {.lex_state = 123, .external_lex_state = 3},
  [2085] = {.lex_state = 123, .external_lex_state = 3},
  [2086] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2087] = {.lex_state = 123, .external_lex_state = 3},
  [2088] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 35},
  [2089] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2090] = {.lex_state = 123, .external_lex_state = 3},
  [2091] = {.lex_state = 123, .external_lex_state = 3},
  [2092] = {.lex_state = 123, .external_lex_state = 3},
  [2093] = {.lex_state = 123, .external_lex_state = 3},
  [2094] = {.lex_state = 123, .external_lex_state = 3},
  [2095] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2096] = {.lex_state = 123, .external_lex_state = 3},
  [2097] = {.lex_state = 123, .external_lex_state = 3},
  [2098] = {.lex_state = 123, .external_lex_state = 3},
  [2099] = {.lex_state = 123, .external_lex_state = 3},
  [2100] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 35},
  [2101] = {.lex_state = 123, .external_lex_state = 3},
  [2102] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 59},
  [2103] = {.lex_state = 123, .external_lex_state = 3},
  [2104] = {.lex_state = 123, .external_lex_state = 3},
  [2105] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 35},
  [2106] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 35},
  [2107] = {.lex_state = 16, .external_lex_state = 3},
  [2108] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2109] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2110] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [2111] = {.lex_state = 123, .external_lex_state = 3},
  [2112] = {.lex_state = 123, .external_lex_state = 3},
  [2113] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [2114] = {.lex_state = 138, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2115] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2116] = {.lex_state = 123, .external_lex_state = 3},
  [2117] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2118] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2119] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2120] = {.lex_state = 123, .external_lex_state = 3},
  [2121] = {.lex_state = 123, .external_lex_state = 3},
  [2122] = {.lex_state = 126, .external_lex_state = 3},
  [2123] = {.lex_state = 138, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2124] = {.lex_state = 123, .external_lex_state = 3},
  [2125] = {.lex_state = 123, .external_lex_state = 3},
  [2126] = {.lex_state = 126, .external_lex_state = 3},
  [2127] = {.lex_state = 138, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2128] = {.lex_state = 123, .external_lex_state = 3},
  [2129] = {.lex_state = 123, .external_lex_state = 3},
  [2130] = {.lex_state = 123, .external_lex_state = 3},
  [2131] = {.lex_state = 123, .external_lex_state = 3},
  [2132] = {.lex_state = 123, .external_lex_state = 3},
  [2133] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2134] = {.lex_state = 138, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2135] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2136] = {.lex_state = 123, .external_lex_state = 3},
  [2137] = {.lex_state = 123, .external_lex_state = 3},
  [2138] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [2139] = {.lex_state = 123, .external_lex_state = 3},
  [2140] = {.lex_state = 123, .external_lex_state = 3},
  [2141] = {.lex_state = 123, .external_lex_state = 3},
  [2142] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2143] = {.lex_state = 138, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2144] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2145] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [2146] = {.lex_state = 123, .external_lex_state = 3},
  [2147] = {.lex_state = 123, .external_lex_state = 3},
  [2148] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2149] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2150] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2151] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2152] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2153] = {.lex_state = 138, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2154] = {.lex_state = 138, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2155] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2156] = {.lex_state = 123, .external_lex_state = 3},
  [2157] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2158] = {.lex_state = 123, .external_lex_state = 3},
  [2159] = {.lex_state = 123, .external_lex_state = 3},
  [2160] = {.lex_state = 126, .external_lex_state = 3},
  [2161] = {.lex_state = 138, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2162] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2163] = {.lex_state = 123, .external_lex_state = 3},
  [2164] = {.lex_state = 123, .external_lex_state = 3},
  [2165] = {.lex_state = 123, .external_lex_state = 3},
  [2166] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2167] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2168] = {.lex_state = 123, .external_lex_state = 3},
  [2169] = {.lex_state = 123, .external_lex_state = 3},
  [2170] = {.lex_state = 123, .external_lex_state = 3},
  [2171] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2172] = {.lex_state = 123, .external_lex_state = 3},
  [2173] = {.lex_state = 138, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2174] = {.lex_state = 126, .external_lex_state = 3},
  [2175] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2176] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2177] = {.lex_state = 123, .external_lex_state = 3},
  [2178] = {.lex_state = 138, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2179] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2180] = {.lex_state = 123, .external_lex_state = 3},
  [2181] = {.lex_state = 123, .external_lex_state = 3},
  [2182] = {.lex_state = 123, .external_lex_state = 3},
  [2183] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 54},
  [2184] = {.lex_state = 123, .external_lex_state = 3},
  [2185] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2186] = {.lex_state = 123, .external_lex_state = 3},
  [2187] = {.lex_state = 123, .external_lex_state = 3},
  [2188] = {.lex_state = 123, .external_lex_state = 3},
  [2189] = {.lex_state = 16, .external_lex_state = 3},
  [2190] = {.lex_state = 126, .external_lex_state = 3},
  [2191] = {.lex_state = 16, .external_lex_state = 3},
  [2192] = {.lex_state = 126, .external_lex_state = 3},
  [2193] = {.lex_state = 16, .external_lex_state = 3},
  [2194] = {.lex_state = 126, .external_lex_state = 3},
  [2195] = {.lex_state = 16, .external_lex_state = 3},
  [2196] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2197] = {.lex_state = 126, .external_lex_state = 3},
  [2198] = {.lex_state = 16, .external_lex_state = 3},
  [2199] = {.lex_state = 123, .external_lex_state = 3},
  [2200] = {.lex_state = 16, .external_lex_state = 3},
  [2201] = {.lex_state = 16, .external_lex_state = 3},
  [2202] = {.lex_state = 126, .external_lex_state = 3},
  [2203] = {.lex_state = 126, .external_lex_state = 3},
  [2204] = {.lex_state = 16, .external_lex_state = 3},
  [2205] = {.lex_state = 16, .external_lex_state = 3},
  [2206] = {.lex_state = 126, .external_lex_state = 3},
  [2207] = {.lex_state = 126, .external_lex_state = 3},
  [2208] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2209] = {.lex_state = 123, .external_lex_state = 3},
  [2210] = {.lex_state = 16, .external_lex_state = 3},
  [2211] = {.lex_state = 123, .external_lex_state = 3},
  [2212] = {.lex_state = 126, .external_lex_state = 3},
  [2213] = {.lex_state = 16, .external_lex_state = 3},
  [2214] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2215] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2216] = {.lex_state = 16, .external_lex_state = 3},
  [2217] = {.lex_state = 123, .external_lex_state = 3},
  [2218] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2219] = {.lex_state = 126, .external_lex_state = 3},
  [2220] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2221] = {.lex_state = 123, .external_lex_state = 3},
  [2222] = {.lex_state = 123, .external_lex_state = 3},
  [2223] = {.lex_state = 123, .external_lex_state = 3},
  [2224] = {.lex_state = 126, .external_lex_state = 3},
  [2225] = {.lex_state = 126, .external_lex_state = 3},
  [2226] = {.lex_state = 126, .external_lex_state = 3},
  [2227] = {.lex_state = 123, .external_lex_state = 3},
  [2228] = {.lex_state = 16, .external_lex_state = 3},
  [2229] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2230] = {.lex_state = 123, .external_lex_state = 3},
  [2231] = {.lex_state = 16, .external_lex_state = 3},
  [2232] = {.lex_state = 123, .external_lex_state = 3},
  [2233] = {.lex_state = 123, .external_lex_state = 3},
  [2234] = {.lex_state = 123, .external_lex_state = 3},
  [2235] = {.lex_state = 123, .external_lex_state = 3},
  [2236] = {.lex_state = 16, .external_lex_state = 3},
  [2237] = {.lex_state = 123, .external_lex_state = 3},
  [2238] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2239] = {.lex_state = 126, .external_lex_state = 3},
  [2240] = {.lex_state = 126, .external_lex_state = 3},
  [2241] = {.lex_state = 126, .external_lex_state = 3},
  [2242] = {.lex_state = 126, .external_lex_state = 3},
  [2243] = {.lex_state = 126, .external_lex_state = 3},
  [2244] = {.lex_state = 123, .external_lex_state = 3},
  [2245] = {.lex_state = 123, .external_lex_state = 3},
  [2246] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2247] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2248] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2249] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2250] = {.lex_state = 126, .external_lex_state = 3},
  [2251] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2252] = {.lex_state = 126, .external_lex_state = 3},
  [2253] = {.lex_state = 126, .external_lex_state = 3},
  [2254] = {.lex_state = 16, .external_lex_state = 3},
  [2255] = {.lex_state = 123, .external_lex_state = 3},
  [2256] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2257] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2258] = {.lex_state = 16, .external_lex_state = 3},
  [2259] = {.lex_state = 126, .external_lex_state = 3},
  [2260] = {.lex_state = 16, .external_lex_state = 3},
  [2261] = {.lex_state = 16, .external_lex_state = 3},
  [2262] = {.lex_state = 126, .external_lex_state = 3},
  [2263] = {.lex_state = 126, .external_lex_state = 3},
  [2264] = {.lex_state = 16, .external_lex_state = 3},
  [2265] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2266] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2267] = {.lex_state = 16, .external_lex_state = 3},
  [2268] = {.lex_state = 126, .external_lex_state = 3},
  [2269] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2270] = {.lex_state = 16, .external_lex_state = 3},
  [2271] = {.lex_state = 126, .external_lex_state = 3},
  [2272] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2273] = {.lex_state = 123, .external_lex_state = 3},
  [2274] = {.lex_state = 123, .external_lex_state = 3},
  [2275] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2276] = {.lex_state = 126, .external_lex_state = 3},
  [2277] = {.lex_state = 126, .external_lex_state = 3},
  [2278] = {.lex_state = 126, .external_lex_state = 3},
  [2279] = {.lex_state = 16, .external_lex_state = 3},
  [2280] = {.lex_state = 126, .external_lex_state = 3},
  [2281] = {.lex_state = 126, .external_lex_state = 3},
  [2282] = {.lex_state = 16, .external_lex_state = 3},
  [2283] = {.lex_state = 126, .external_lex_state = 3},
  [2284] = {.lex_state = 16, .external_lex_state = 3},
  [2285] = {.lex_state = 16, .external_lex_state = 3},
  [2286] = {.lex_state = 126, .external_lex_state = 3},
  [2287] = {.lex_state = 126, .external_lex_state = 3},
  [2288] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2289] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2290] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2291] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2292] = {.lex_state = 126, .external_lex_state = 3},
  [2293] = {.lex_state = 126, .external_lex_state = 3},
  [2294] = {.lex_state = 126, .external_lex_state = 3},
  [2295] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2296] = {.lex_state = 126, .external_lex_state = 3},
  [2297] = {.lex_state = 126, .external_lex_state = 3},
  [2298] = {.lex_state = 126, .external_lex_state = 3},
  [2299] = {.lex_state = 16, .external_lex_state = 3},
  [2300] = {.lex_state = 126, .external_lex_state = 3},
  [2301] = {.lex_state = 126, .external_lex_state = 3},
  [2302] = {.lex_state = 126, .external_lex_state = 3},
  [2303] = {.lex_state = 16, .external_lex_state = 3},
  [2304] = {.lex_state = 16, .external_lex_state = 3},
  [2305] = {.lex_state = 126, .external_lex_state = 3},
  [2306] = {.lex_state = 126, .external_lex_state = 3},
  [2307] = {.lex_state = 126, .external_lex_state = 3},
  [2308] = {.lex_state = 126, .external_lex_state = 3},
  [2309] = {.lex_state = 126, .external_lex_state = 3},
  [2310] = {.lex_state = 126, .external_lex_state = 3},
  [2311] = {.lex_state = 126, .external_lex_state = 3},
  [2312] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2313] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2314] = {.lex_state = 126, .external_lex_state = 3},
  [2315] = {.lex_state = 126, .external_lex_state = 3},
  [2316] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2317] = {.lex_state = 16, .external_lex_state = 3},
  [2318] = {.lex_state = 16, .external_lex_state = 3},
  [2319] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2320] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2321] = {.lex_state = 126, .external_lex_state = 3},
  [2322] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2323] = {.lex_state = 126, .external_lex_state = 3},
  [2324] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2325] = {.lex_state = 126, .external_lex_state = 3},
  [2326] = {.lex_state = 16, .external_lex_state = 3},
  [2327] = {.lex_state = 16, .external_lex_state = 3},
  [2328] = {.lex_state = 126, .external_lex_state = 3},
  [2329] = {.lex_state = 16, .external_lex_state = 3},
  [2330] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2331] = {.lex_state = 16, .external_lex_state = 3},
  [2332] = {.lex_state = 16, .external_lex_state = 3},
  [2333] = {.lex_state = 16, .external_lex_state = 3},
  [2334] = {.lex_state = 126, .external_lex_state = 3},
  [2335] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2336] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2337] = {.lex_state = 126, .external_lex_state = 3},
  [2338] = {.lex_state = 16, .external_lex_state = 3},
  [2339] = {.lex_state = 126, .external_lex_state = 3},
  [2340] = {.lex_state = 126, .external_lex_state = 3},
  [2341] = {.lex_state = 126, .external_lex_state = 3},
  [2342] = {.lex_state = 16, .external_lex_state = 3},
  [2343] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2344] = {.lex_state = 126, .external_lex_state = 3},
  [2345] = {.lex_state = 126, .external_lex_state = 3},
  [2346] = {.lex_state = 126, .external_lex_state = 3},
  [2347] = {.lex_state = 126, .external_lex_state = 3},
  [2348] = {.lex_state = 16, .external_lex_state = 3},
  [2349] = {.lex_state = 16, .external_lex_state = 3},
  [2350] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2351] = {.lex_state = 16, .external_lex_state = 3},
  [2352] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2353] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2354] = {.lex_state = 16, .external_lex_state = 3},
  [2355] = {.lex_state = 126, .external_lex_state = 3},
  [2356] = {.lex_state = 126, .external_lex_state = 3},
  [2357] = {.lex_state = 40, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2358] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2359] = {.lex_state = 16, .external_lex_state = 3},
  [2360] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2361] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2362] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2363] = {.lex_state = 126, .external_lex_state = 3},
  [2364] = {.lex_state = 16, .external_lex_state = 3},
  [2365] = {.lex_state = 16, .external_lex_state = 3},
  [2366] = {.lex_state = 16, .external_lex_state = 3},
  [2367] = {.lex_state = 16, .external_lex_state = 3},
  [2368] = {.lex_state = 16, .external_lex_state = 3},
  [2369] = {.lex_state = 16, .external_lex_state = 3},
  [2370] = {.lex_state = 16, .external_lex_state = 3},
  [2371] = {.lex_state = 126, .external_lex_state = 3},
  [2372] = {.lex_state = 123, .external_lex_state = 3},
  [2373] = {.lex_state = 126, .external_lex_state = 3},
  [2374] = {.lex_state = 126, .external_lex_state = 3},
  [2375] = {.lex_state = 16, .external_lex_state = 3},
  [2376] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2377] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2378] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2379] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2380] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2381] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2382] = {.lex_state = 16, .external_lex_state = 3},
  [2383] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2384] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2385] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2386] = {.lex_state = 16, .external_lex_state = 3},
  [2387] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2388] = {.lex_state = 126, .external_lex_state = 3},
  [2389] = {.lex_state = 126, .external_lex_state = 3},
  [2390] = {.lex_state = 16, .external_lex_state = 3},
  [2391] = {.lex_state = 16, .external_lex_state = 3},
  [2392] = {.lex_state = 16, .external_lex_state = 3},
  [2393] = {.lex_state = 16, .external_lex_state = 3},
  [2394] = {.lex_state = 16, .external_lex_state = 3},
  [2395] = {.lex_state = 16, .external_lex_state = 3},
  [2396] = {.lex_state = 126, .external_lex_state = 3},
  [2397] = {.lex_state = 16, .external_lex_state = 3},
  [2398] = {.lex_state = 16, .external_lex_state = 3},
  [2399] = {.lex_state = 16, .external_lex_state = 3},
  [2400] = {.lex_state = 16, .external_lex_state = 3},
  [2401] = {.lex_state = 40, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2402] = {.lex_state = 16, .external_lex_state = 3},
  [2403] = {.lex_state = 126, .external_lex_state = 3},
  [2404] = {.lex_state = 16, .external_lex_state = 3},
  [2405] = {.lex_state = 16, .external_lex_state = 3},
  [2406] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2407] = {.lex_state = 126, .external_lex_state = 3},
  [2408] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2409] = {.lex_state = 16, .external_lex_state = 3},
  [2410] = {.lex_state = 126, .external_lex_state = 3},
  [2411] = {.lex_state = 16, .external_lex_state = 3},
  [2412] = {.lex_state = 16, .external_lex_state = 3},
  [2413] = {.lex_state = 126, .external_lex_state = 3},
  [2414] = {.lex_state = 126, .external_lex_state = 3},
  [2415] = {.lex_state = 16, .external_lex_state = 3},
  [2416] = {.lex_state = 126, .external_lex_state = 3},
  [2417] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2418] = {.lex_state = 16, .external_lex_state = 3},
  [2419] = {.lex_state = 16, .external_lex_state = 3},
  [2420] = {.lex_state = 16, .external_lex_state = 3},
  [2421] = {.lex_state = 16, .external_lex_state = 3},
  [2422] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2423] = {.lex_state = 16, .external_lex_state = 3},
  [2424] = {.lex_state = 16, .external_lex_state = 3},
  [2425] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2426] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2427] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2428] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2429] = {.lex_state = 16, .external_lex_state = 3},
  [2430] = {.lex_state = 16, .external_lex_state = 3},
  [2431] = {.lex_state = 126, .external_lex_state = 3},
  [2432] = {.lex_state = 126, .external_lex_state = 3},
  [2433] = {.lex_state = 16, .external_lex_state = 3},
  [2434] = {.lex_state = 16, .external_lex_state = 3},
  [2435] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2436] = {.lex_state = 16, .external_lex_state = 3},
  [2437] = {.lex_state = 126, .external_lex_state = 3},
  [2438] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2439] = {.lex_state = 16, .external_lex_state = 3},
  [2440] = {.lex_state = 126, .external_lex_state = 3},
  [2441] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2442] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2443] = {.lex_state = 126, .external_lex_state = 3},
  [2444] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 38},
  [2445] = {.lex_state = 18, .external_lex_state = 3},
  [2446] = {.lex_state = 126, .external_lex_state = 3},
  [2447] = {.lex_state = 126, .external_lex_state = 3},
  [2448] = {.lex_state = 126, .external_lex_state = 3},
  [2449] = {.lex_state = 18, .external_lex_state = 3},
  [2450] = {.lex_state = 18, .external_lex_state = 3},
  [2451] = {.lex_state = 126, .external_lex_state = 3},
  [2452] = {.lex_state = 18, .external_lex_state = 3},
  [2453] = {.lex_state = 18, .external_lex_state = 3},
  [2454] = {.lex_state = 18, .external_lex_state = 3},
  [2455] = {.lex_state = 18, .external_lex_state = 3},
  [2456] = {.lex_state = 18, .external_lex_state = 3},
  [2457] = {.lex_state = 18, .external_lex_state = 3},
  [2458] = {.lex_state = 18, .external_lex_state = 3},
  [2459] = {.lex_state = 18, .external_lex_state = 3},
  [2460] = {.lex_state = 18, .external_lex_state = 3},
  [2461] = {.lex_state = 18, .external_lex_state = 3},
  [2462] = {.lex_state = 18, .external_lex_state = 3},
  [2463] = {.lex_state = 18, .external_lex_state = 3},
  [2464] = {.lex_state = 18, .external_lex_state = 3},
  [2465] = {.lex_state = 18, .external_lex_state = 3},
  [2466] = {.lex_state = 18, .external_lex_state = 3},
  [2467] = {.lex_state = 18, .external_lex_state = 3},
  [2468] = {.lex_state = 18, .external_lex_state = 3},
  [2469] = {.lex_state = 18, .external_lex_state = 3},
  [2470] = {.lex_state = 18, .external_lex_state = 3},
  [2471] = {.lex_state = 18, .external_lex_state = 3},
  [2472] = {.lex_state = 18, .external_lex_state = 3},
  [2473] = {.lex_state = 18, .external_lex_state = 3},
  [2474] = {.lex_state = 18, .external_lex_state = 3},
  [2475] = {.lex_state = 18, .external_lex_state = 3},
  [2476] = {.lex_state = 18, .external_lex_state = 3},
  [2477] = {.lex_state = 18, .external_lex_state = 3},
  [2478] = {.lex_state = 18, .external_lex_state = 3},
  [2479] = {.lex_state = 18, .external_lex_state = 3},
  [2480] = {.lex_state = 18, .external_lex_state = 3},
  [2481] = {.lex_state = 18, .external_lex_state = 3},
  [2482] = {.lex_state = 18, .external_lex_state = 3},
  [2483] = {.lex_state = 18, .external_lex_state = 3},
  [2484] = {.lex_state = 18, .external_lex_state = 3},
  [2485] = {.lex_state = 18, .external_lex_state = 3},
  [2486] = {.lex_state = 18, .external_lex_state = 3},
  [2487] = {.lex_state = 18, .external_lex_state = 3},
  [2488] = {.lex_state = 18, .external_lex_state = 3},
  [2489] = {.lex_state = 18, .external_lex_state = 3},
  [2490] = {.lex_state = 18, .external_lex_state = 3},
  [2491] = {.lex_state = 18, .external_lex_state = 3},
  [2492] = {.lex_state = 18, .external_lex_state = 3},
  [2493] = {.lex_state = 18, .external_lex_state = 3},
  [2494] = {.lex_state = 18, .external_lex_state = 3},
  [2495] = {.lex_state = 18, .external_lex_state = 3},
  [2496] = {.lex_state = 18, .external_lex_state = 3},
  [2497] = {.lex_state = 18, .external_lex_state = 3},
  [2498] = {.lex_state = 18, .external_lex_state = 3},
  [2499] = {.lex_state = 18, .external_lex_state = 3},
  [2500] = {.lex_state = 18, .external_lex_state = 3},
  [2501] = {.lex_state = 18, .external_lex_state = 3},
  [2502] = {.lex_state = 126, .external_lex_state = 3},
  [2503] = {.lex_state = 126, .external_lex_state = 3},
  [2504] = {.lex_state = 126, .external_lex_state = 3},
  [2505] = {.lex_state = 126, .external_lex_state = 3},
  [2506] = {.lex_state = 126, .external_lex_state = 3},
  [2507] = {.lex_state = 126, .external_lex_state = 3},
  [2508] = {.lex_state = 18, .external_lex_state = 3},
  [2509] = {.lex_state = 18, .external_lex_state = 3},
  [2510] = {.lex_state = 18, .external_lex_state = 3},
  [2511] = {.lex_state = 18, .external_lex_state = 3},
  [2512] = {.lex_state = 18, .external_lex_state = 3},
  [2513] = {.lex_state = 18, .external_lex_state = 3},
  [2514] = {.lex_state = 18, .external_lex_state = 3},
  [2515] = {.lex_state = 18, .external_lex_state = 3},
  [2516] = {.lex_state = 18, .external_lex_state = 3},
  [2517] = {.lex_state = 18, .external_lex_state = 3},
  [2518] = {.lex_state = 18, .external_lex_state = 3},
  [2519] = {.lex_state = 18, .external_lex_state = 3},
  [2520] = {.lex_state = 18, .external_lex_state = 3},
  [2521] = {.lex_state = 18, .external_lex_state = 3},
  [2522] = {.lex_state = 18, .external_lex_state = 3},
  [2523] = {.lex_state = 18, .external_lex_state = 3},
  [2524] = {.lex_state = 18, .external_lex_state = 3},
  [2525] = {.lex_state = 18, .external_lex_state = 3},
  [2526] = {.lex_state = 18, .external_lex_state = 3},
  [2527] = {.lex_state = 18, .external_lex_state = 3},
  [2528] = {.lex_state = 18, .external_lex_state = 3},
  [2529] = {.lex_state = 18, .external_lex_state = 3},
  [2530] = {.lex_state = 18, .external_lex_state = 3},
  [2531] = {.lex_state = 18, .external_lex_state = 3},
  [2532] = {.lex_state = 18, .external_lex_state = 3},
  [2533] = {.lex_state = 18, .external_lex_state = 3},
  [2534] = {.lex_state = 122, .external_lex_state = 3},
  [2535] = {.lex_state = 122, .external_lex_state = 3},
  [2536] = {.lex_state = 122, .external_lex_state = 3},
  [2537] = {.lex_state = 122, .external_lex_state = 3},
  [2538] = {.lex_state = 123, .external_lex_state = 3},
  [2539] = {.lex_state = 123, .external_lex_state = 3},
  [2540] = {.lex_state = 122, .external_lex_state = 3},
  [2541] = {.lex_state = 122, .external_lex_state = 3},
  [2542] = {.lex_state = 122, .external_lex_state = 3},
  [2543] = {.lex_state = 122, .external_lex_state = 3},
  [2544] = {.lex_state = 122, .external_lex_state = 3},
  [2545] = {.lex_state = 122, .external_lex_state = 3},
  [2546] = {.lex_state = 122, .external_lex_state = 3},
  [2547] = {.lex_state = 122, .external_lex_state = 3},
  [2548] = {.lex_state = 123, .external_lex_state = 3},
  [2549] = {.lex_state = 122, .external_lex_state = 3},
  [2550] = {.lex_state = 122, .external_lex_state = 3},
  [2551] = {.lex_state = 126, .external_lex_state = 3},
  [2552] = {.lex_state = 126, .external_lex_state = 3},
  [2553] = {.lex_state = 126, .external_lex_state = 3},
  [2554] = {.lex_state = 126, .external_lex_state = 3},
  [2555] = {.lex_state = 126, .external_lex_state = 3},
  [2556] = {.lex_state = 126, .external_lex_state = 3},
  [2557] = {.lex_state = 18, .external_lex_state = 3},
  [2558] = {.lex_state = 18, .external_lex_state = 3},
  [2559] = {.lex_state = 126, .external_lex_state = 3},
  [2560] = {.lex_state = 126, .external_lex_state = 3},
  [2561] = {.lex_state = 126, .external_lex_state = 3},
  [2562] = {.lex_state = 126, .external_lex_state = 3},
  [2563] = {.lex_state = 126, .external_lex_state = 3},
  [2564] = {.lex_state = 123, .external_lex_state = 3},
  [2565] = {.lex_state = 126, .external_lex_state = 3},
  [2566] = {.lex_state = 126, .external_lex_state = 3},
  [2567] = {.lex_state = 126, .external_lex_state = 3},
  [2568] = {.lex_state = 126, .external_lex_state = 3},
  [2569] = {.lex_state = 126, .external_lex_state = 3},
  [2570] = {.lex_state = 18, .external_lex_state = 3},
  [2571] = {.lex_state = 18, .external_lex_state = 3},
  [2572] = {.lex_state = 126, .external_lex_state = 3},
  [2573] = {.lex_state = 126, .external_lex_state = 3},
  [2574] = {.lex_state = 126, .external_lex_state = 3},
  [2575] = {.lex_state = 126, .external_lex_state = 3},
  [2576] = {.lex_state = 126, .external_lex_state = 3},
  [2577] = {.lex_state = 126, .external_lex_state = 3},
  [2578] = {.lex_state = 126, .external_lex_state = 3},
  [2579] = {.lex_state = 18, .external_lex_state = 3},
  [2580] = {.lex_state = 126, .external_lex_state = 3},
  [2581] = {.lex_state = 126, .external_lex_state = 3},
  [2582] = {.lex_state = 126, .external_lex_state = 3},
  [2583] = {.lex_state = 126, .external_lex_state = 3},
  [2584] = {.lex_state = 126, .external_lex_state = 3},
  [2585] = {.lex_state = 126, .external_lex_state = 3},
  [2586] = {.lex_state = 126, .external_lex_state = 3},
  [2587] = {.lex_state = 126, .external_lex_state = 3},
  [2588] = {.lex_state = 126, .external_lex_state = 3},
  [2589] = {.lex_state = 126, .external_lex_state = 3},
  [2590] = {.lex_state = 126, .external_lex_state = 3},
  [2591] = {.lex_state = 126, .external_lex_state = 3},
  [2592] = {.lex_state = 126, .external_lex_state = 3},
  [2593] = {.lex_state = 126, .external_lex_state = 3},
  [2594] = {.lex_state = 126, .external_lex_state = 3},
  [2595] = {.lex_state = 126, .external_lex_state = 3},
  [2596] = {.lex_state = 126, .external_lex_state = 3},
  [2597] = {.lex_state = 126, .external_lex_state = 3},
  [2598] = {.lex_state = 126, .external_lex_state = 3},
  [2599] = {.lex_state = 126, .external_lex_state = 3},
  [2600] = {.lex_state = 126, .external_lex_state = 3},
  [2601] = {.lex_state = 126, .external_lex_state = 3},
  [2602] = {.lex_state = 126, .external_lex_state = 3},
  [2603] = {.lex_state = 126, .external_lex_state = 3},
  [2604] = {.lex_state = 126, .external_lex_state = 3},
  [2605] = {.lex_state = 126, .external_lex_state = 3},
  [2606] = {.lex_state = 126, .external_lex_state = 3},
  [2607] = {.lex_state = 126, .external_lex_state = 3},
  [2608] = {.lex_state = 126, .external_lex_state = 3},
  [2609] = {.lex_state = 126, .external_lex_state = 3},
  [2610] = {.lex_state = 40, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2611] = {.lex_state = 40, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2612] = {.lex_state = 123, .external_lex_state = 3},
  [2613] = {.lex_state = 40, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2614] = {.lex_state = 123, .external_lex_state = 3},
  [2615] = {.lex_state = 126, .external_lex_state = 3},
  [2616] = {.lex_state = 126, .external_lex_state = 3},
  [2617] = {.lex_state = 126, .external_lex_state = 3},
  [2618] = {.lex_state = 126, .external_lex_state = 3},
  [2619] = {.lex_state = 126, .external_lex_state = 3},
  [2620] = {.lex_state = 126, .external_lex_state = 3},
  [2621] = {.lex_state = 40, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2622] = {.lex_state = 126, .external_lex_state = 3},
  [2623] = {.lex_state = 126, .external_lex_state = 3},
  [2624] = {.lex_state = 126, .external_lex_state = 3},
  [2625] = {.lex_state = 126, .external_lex_state = 3},
  [2626] = {.lex_state = 126, .external_lex_state = 3},
  [2627] = {.lex_state = 126, .external_lex_state = 3},
  [2628] = {.lex_state = 126, .external_lex_state = 3},
  [2629] = {.lex_state = 126, .external_lex_state = 3},
  [2630] = {.lex_state = 122, .external_lex_state = 3},
  [2631] = {.lex_state = 126, .external_lex_state = 3},
  [2632] = {.lex_state = 126, .external_lex_state = 3},
  [2633] = {.lex_state = 122, .external_lex_state = 3},
  [2634] = {.lex_state = 122, .external_lex_state = 3},
  [2635] = {.lex_state = 123, .external_lex_state = 3},
  [2636] = {.lex_state = 123, .external_lex_state = 3},
  [2637] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 61},
  [2638] = {.lex_state = 123, .external_lex_state = 3},
  [2639] = {.lex_state = 122, .external_lex_state = 3},
  [2640] = {.lex_state = 123, .external_lex_state = 3},
  [2641] = {.lex_state = 123, .external_lex_state = 3},
  [2642] = {.lex_state = 123, .external_lex_state = 3},
  [2643] = {.lex_state = 123, .external_lex_state = 3},
  [2644] = {.lex_state = 123, .external_lex_state = 3},
  [2645] = {.lex_state = 122, .external_lex_state = 3},
  [2646] = {.lex_state = 123, .external_lex_state = 3},
  [2647] = {.lex_state = 123, .external_lex_state = 3},
  [2648] = {.lex_state = 123, .external_lex_state = 3},
  [2649] = {.lex_state = 123, .external_lex_state = 3},
  [2650] = {.lex_state = 122, .external_lex_state = 3},
  [2651] = {.lex_state = 123, .external_lex_state = 3},
  [2652] = {.lex_state = 123, .external_lex_state = 3},
  [2653] = {.lex_state = 123, .external_lex_state = 3},
  [2654] = {.lex_state = 123, .external_lex_state = 3},
  [2655] = {.lex_state = 123, .external_lex_state = 3},
  [2656] = {.lex_state = 122, .external_lex_state = 3},
  [2657] = {.lex_state = 122, .external_lex_state = 3},
  [2658] = {.lex_state = 123, .external_lex_state = 3},
  [2659] = {.lex_state = 122, .external_lex_state = 3},
  [2660] = {.lex_state = 122, .external_lex_state = 3},
  [2661] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2662] = {.lex_state = 122, .external_lex_state = 3},
  [2663] = {.lex_state = 122, .external_lex_state = 3},
  [2664] = {.lex_state = 122, .external_lex_state = 3},
  [2665] = {.lex_state = 22, .external_lex_state = 3},
  [2666] = {.lex_state = 122, .external_lex_state = 3},
  [2667] = {.lex_state = 122, .external_lex_state = 3},
  [2668] = {.lex_state = 122, .external_lex_state = 3},
  [2669] = {.lex_state = 122, .external_lex_state = 3},
  [2670] = {.lex_state = 123, .external_lex_state = 3},
  [2671] = {.lex_state = 123, .external_lex_state = 3},
  [2672] = {.lex_state = 122, .external_lex_state = 3},
  [2673] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2674] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 62},
  [2675] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 62},
  [2676] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2677] = {.lex_state = 123, .external_lex_state = 3},
  [2678] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2679] = {.lex_state = 123, .external_lex_state = 3},
  [2680] = {.lex_state = 122, .external_lex_state = 3},
  [2681] = {.lex_state = 123, .external_lex_state = 3},
  [2682] = {.lex_state = 123, .external_lex_state = 3},
  [2683] = {.lex_state = 123, .external_lex_state = 3},
  [2684] = {.lex_state = 123, .external_lex_state = 3},
  [2685] = {.lex_state = 123, .external_lex_state = 3},
  [2686] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 62},
  [2687] = {.lex_state = 124, .external_lex_state = 3},
  [2688] = {.lex_state = 123, .external_lex_state = 3},
  [2689] = {.lex_state = 123, .external_lex_state = 3},
  [2690] = {.lex_state = 123, .external_lex_state = 3},
  [2691] = {.lex_state = 123, .external_lex_state = 3},
  [2692] = {.lex_state = 123, .external_lex_state = 3},
  [2693] = {.lex_state = 123, .external_lex_state = 3},
  [2694] = {.lex_state = 123, .external_lex_state = 3},
  [2695] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2696] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2697] = {.lex_state = 123, .external_lex_state = 3},
  [2698] = {.lex_state = 123, .external_lex_state = 3},
  [2699] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 62},
  [2700] = {.lex_state = 122, .external_lex_state = 3},
  [2701] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2702] = {.lex_state = 123, .external_lex_state = 3},
  [2703] = {.lex_state = 123, .external_lex_state = 3},
  [2704] = {.lex_state = 123, .external_lex_state = 3},
  [2705] = {.lex_state = 123, .external_lex_state = 3},
  [2706] = {.lex_state = 16, .external_lex_state = 3},
  [2707] = {.lex_state = 16, .external_lex_state = 3},
  [2708] = {.lex_state = 16, .external_lex_state = 3},
  [2709] = {.lex_state = 16, .external_lex_state = 3},
  [2710] = {.lex_state = 21, .external_lex_state = 3},
  [2711] = {.lex_state = 16, .external_lex_state = 3},
  [2712] = {.lex_state = 16, .external_lex_state = 3},
  [2713] = {.lex_state = 16, .external_lex_state = 3},
  [2714] = {.lex_state = 16, .external_lex_state = 3},
  [2715] = {.lex_state = 122, .external_lex_state = 3},
  [2716] = {.lex_state = 16, .external_lex_state = 3},
  [2717] = {.lex_state = 123, .external_lex_state = 3},
  [2718] = {.lex_state = 16, .external_lex_state = 3},
  [2719] = {.lex_state = 123, .external_lex_state = 3},
  [2720] = {.lex_state = 123, .external_lex_state = 3},
  [2721] = {.lex_state = 123, .external_lex_state = 3},
  [2722] = {.lex_state = 123, .external_lex_state = 3},
  [2723] = {.lex_state = 16, .external_lex_state = 3},
  [2724] = {.lex_state = 123, .external_lex_state = 3},
  [2725] = {.lex_state = 16, .external_lex_state = 3},
  [2726] = {.lex_state = 123, .external_lex_state = 3},
  [2727] = {.lex_state = 123, .external_lex_state = 3},
  [2728] = {.lex_state = 122, .external_lex_state = 3},
  [2729] = {.lex_state = 16, .external_lex_state = 3},
  [2730] = {.lex_state = 16, .external_lex_state = 3},
  [2731] = {.lex_state = 123, .external_lex_state = 3},
  [2732] = {.lex_state = 123, .external_lex_state = 3},
  [2733] = {.lex_state = 123, .external_lex_state = 3},
  [2734] = {.lex_state = 16, .external_lex_state = 3},
  [2735] = {.lex_state = 123, .external_lex_state = 3},
  [2736] = {.lex_state = 16, .external_lex_state = 3},
  [2737] = {.lex_state = 124, .external_lex_state = 3},
  [2738] = {.lex_state = 16, .external_lex_state = 3},
  [2739] = {.lex_state = 16, .external_lex_state = 3},
  [2740] = {.lex_state = 16, .external_lex_state = 3},
  [2741] = {.lex_state = 16, .external_lex_state = 3},
  [2742] = {.lex_state = 122, .external_lex_state = 3},
  [2743] = {.lex_state = 16, .external_lex_state = 3},
  [2744] = {.lex_state = 123, .external_lex_state = 3},
  [2745] = {.lex_state = 16, .external_lex_state = 3},
  [2746] = {.lex_state = 16, .external_lex_state = 3},
  [2747] = {.lex_state = 16, .external_lex_state = 3},
  [2748] = {.lex_state = 16, .external_lex_state = 3},
  [2749] = {.lex_state = 123, .external_lex_state = 3},
  [2750] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2751] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2752] = {.lex_state = 123, .external_lex_state = 3},
  [2753] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2754] = {.lex_state = 18, .external_lex_state = 3},
  [2755] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2756] = {.lex_state = 18, .external_lex_state = 3},
  [2757] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2758] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2759] = {.lex_state = 122, .external_lex_state = 3},
  [2760] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2761] = {.lex_state = 18, .external_lex_state = 3},
  [2762] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2763] = {.lex_state = 122, .external_lex_state = 3},
  [2764] = {.lex_state = 18, .external_lex_state = 3},
  [2765] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2766] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2767] = {.lex_state = 18, .external_lex_state = 3},
  [2768] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2769] = {.lex_state = 18, .external_lex_state = 3},
  [2770] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 65},
  [2771] = {.lex_state = 18, .external_lex_state = 3},
  [2772] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 65},
  [2773] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2774] = {.lex_state = 18, .external_lex_state = 3},
  [2775] = {.lex_state = 18, .external_lex_state = 3},
  [2776] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2777] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2778] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2779] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2780] = {.lex_state = 18, .external_lex_state = 3},
  [2781] = {.lex_state = 18, .external_lex_state = 3},
  [2782] = {.lex_state = 18, .external_lex_state = 3},
  [2783] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2784] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2785] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2786] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2787] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2788] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2789] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2790] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2791] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2792] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2793] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2794] = {.lex_state = 123, .external_lex_state = 3},
  [2795] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2796] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2797] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2798] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2799] = {.lex_state = 18, .external_lex_state = 3},
  [2800] = {.lex_state = 18, .external_lex_state = 3},
  [2801] = {.lex_state = 18, .external_lex_state = 3},
  [2802] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2803] = {.lex_state = 18, .external_lex_state = 3},
  [2804] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2805] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2806] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2807] = {.lex_state = 18, .external_lex_state = 3},
  [2808] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2809] = {.lex_state = 18, .external_lex_state = 3},
  [2810] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2811] = {.lex_state = 18, .external_lex_state = 3},
  [2812] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2813] = {.lex_state = 123, .external_lex_state = 3},
  [2814] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2815] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2816] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2817] = {.lex_state = 18, .external_lex_state = 3},
  [2818] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2819] = {.lex_state = 18, .external_lex_state = 3},
  [2820] = {.lex_state = 123, .external_lex_state = 3},
  [2821] = {.lex_state = 18, .external_lex_state = 3},
  [2822] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2823] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2824] = {.lex_state = 18, .external_lex_state = 3},
  [2825] = {.lex_state = 18, .external_lex_state = 3},
  [2826] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2827] = {.lex_state = 18, .external_lex_state = 3},
  [2828] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 65},
  [2829] = {.lex_state = 123, .external_lex_state = 3},
  [2830] = {.lex_state = 123, .external_lex_state = 3},
  [2831] = {.lex_state = 123, .external_lex_state = 3},
  [2832] = {.lex_state = 122, .external_lex_state = 3},
  [2833] = {.lex_state = 123, .external_lex_state = 3},
  [2834] = {.lex_state = 123, .external_lex_state = 3},
  [2835] = {.lex_state = 18, .external_lex_state = 3},
  [2836] = {.lex_state = 18, .external_lex_state = 3},
  [2837] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2838] = {.lex_state = 18, .external_lex_state = 3},
  [2839] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2840] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2841] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2842] = {.lex_state = 18, .external_lex_state = 3},
  [2843] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2844] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2845] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2846] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2847] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2848] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2849] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2850] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2851] = {.lex_state = 18, .external_lex_state = 3},
  [2852] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2853] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 65},
  [2854] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2855] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2856] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2857] = {.lex_state = 18, .external_lex_state = 3},
  [2858] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2859] = {.lex_state = 18, .external_lex_state = 3},
  [2860] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2861] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2862] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2863] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2864] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2865] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2866] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2867] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2868] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2869] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2870] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2871] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2872] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2873] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 56},
  [2874] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2875] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 66},
  [2876] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2877] = {.lex_state = 122, .external_lex_state = 3},
  [2878] = {.lex_state = 122, .external_lex_state = 3},
  [2879] = {.lex_state = 122, .external_lex_state = 3},
  [2880] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2881] = {.lex_state = 122, .external_lex_state = 3},
  [2882] = {.lex_state = 122, .external_lex_state = 3},
  [2883] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 66},
  [2884] = {.lex_state = 123, .external_lex_state = 3},
  [2885] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 66},
  [2886] = {.lex_state = 122, .external_lex_state = 3},
  [2887] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 66},
  [2888] = {.lex_state = 122, .external_lex_state = 3},
  [2889] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2890] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2891] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2892] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2893] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2894] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2895] = {.lex_state = 142, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2896] = {.lex_state = 122, .external_lex_state = 3},
  [2897] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2898] = {.lex_state = 122, .external_lex_state = 3},
  [2899] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2900] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2901] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2902] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2903] = {.lex_state = 122, .external_lex_state = 3},
  [2904] = {.lex_state = 122, .external_lex_state = 3},
  [2905] = {.lex_state = 122, .external_lex_state = 3},
  [2906] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2907] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2908] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2909] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2910] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2911] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2912] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2913] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2914] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2915] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2916] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2917] = {.lex_state = 122, .external_lex_state = 3},
  [2918] = {.lex_state = 122, .external_lex_state = 3},
  [2919] = {.lex_state = 122, .external_lex_state = 3},
  [2920] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2921] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 40},
  [2922] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2923] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2924] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2925] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2926] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2927] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2928] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2929] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2930] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2931] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2932] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2933] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2934] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2935] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2936] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2937] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2938] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2939] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2940] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2941] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2942] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2943] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2944] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2945] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2946] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2947] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2948] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2949] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2950] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2951] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2952] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2953] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2954] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2955] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2956] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2957] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2958] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2959] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2960] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2961] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2962] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2963] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2964] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2965] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2966] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2967] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2968] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2969] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2970] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2971] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2972] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2973] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2974] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2975] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2976] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2977] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2978] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2979] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2980] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2981] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2982] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2983] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2984] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2985] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2986] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2987] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2988] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2989] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2990] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2991] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2992] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2993] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2994] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2995] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2996] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [2997] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2998] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [2999] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3000] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3001] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3002] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3003] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3004] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3005] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3006] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3007] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3008] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3009] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3010] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3011] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3012] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3013] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3014] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3015] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3016] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3017] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3018] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3019] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3020] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3021] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3022] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3023] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3024] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3025] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3026] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3027] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3028] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3029] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3030] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3031] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3032] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3033] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3034] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3035] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3036] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3037] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3038] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3039] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3040] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3041] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3042] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3043] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3044] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3045] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3046] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3047] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3048] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3049] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3050] = {.lex_state = 122, .external_lex_state = 3},
  [3051] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3052] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3053] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3054] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3055] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3056] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3057] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3058] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3059] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3060] = {.lex_state = 136, .external_lex_state = 2, .reserved_word_set_id = 43},
  [3061] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3062] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3063] = {.lex_state = 18, .external_lex_state = 3},
  [3064] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 67},
  [3065] = {.lex_state = 18, .external_lex_state = 3},
  [3066] = {.lex_state = 18, .external_lex_state = 3},
  [3067] = {.lex_state = 18, .external_lex_state = 3},
  [3068] = {.lex_state = 18, .external_lex_state = 3},
  [3069] = {.lex_state = 18, .external_lex_state = 3},
  [3070] = {.lex_state = 18, .external_lex_state = 3},
  [3071] = {.lex_state = 18, .external_lex_state = 3},
  [3072] = {.lex_state = 18, .external_lex_state = 3},
  [3073] = {.lex_state = 18, .external_lex_state = 3},
  [3074] = {.lex_state = 18, .external_lex_state = 3},
  [3075] = {.lex_state = 18, .external_lex_state = 3},
  [3076] = {.lex_state = 18, .external_lex_state = 3},
  [3077] = {.lex_state = 18, .external_lex_state = 3},
  [3078] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 67},
  [3079] = {.lex_state = 25, .external_lex_state = 3},
  [3080] = {.lex_state = 25, .external_lex_state = 3},
  [3081] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 67},
  [3082] = {.lex_state = 18, .external_lex_state = 3},
  [3083] = {.lex_state = 18, .external_lex_state = 3},
  [3084] = {.lex_state = 18, .external_lex_state = 3},
  [3085] = {.lex_state = 25, .external_lex_state = 3},
  [3086] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3087] = {.lex_state = 18, .external_lex_state = 3},
  [3088] = {.lex_state = 18, .external_lex_state = 3},
  [3089] = {.lex_state = 18, .external_lex_state = 3},
  [3090] = {.lex_state = 18, .external_lex_state = 3},
  [3091] = {.lex_state = 18, .external_lex_state = 3},
  [3092] = {.lex_state = 18, .external_lex_state = 3},
  [3093] = {.lex_state = 18, .external_lex_state = 3},
  [3094] = {.lex_state = 18, .external_lex_state = 3},
  [3095] = {.lex_state = 18, .external_lex_state = 3},
  [3096] = {.lex_state = 18, .external_lex_state = 3},
  [3097] = {.lex_state = 18, .external_lex_state = 3},
  [3098] = {.lex_state = 18, .external_lex_state = 3},
  [3099] = {.lex_state = 18, .external_lex_state = 3},
  [3100] = {.lex_state = 18, .external_lex_state = 3},
  [3101] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3102] = {.lex_state = 18, .external_lex_state = 3},
  [3103] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3104] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3105] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3106] = {.lex_state = 18, .external_lex_state = 3},
  [3107] = {.lex_state = 18, .external_lex_state = 3},
  [3108] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3109] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3110] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3111] = {.lex_state = 16, .external_lex_state = 3},
  [3112] = {.lex_state = 16, .external_lex_state = 3},
  [3113] = {.lex_state = 16, .external_lex_state = 3},
  [3114] = {.lex_state = 16, .external_lex_state = 3},
  [3115] = {.lex_state = 16, .external_lex_state = 3},
  [3116] = {.lex_state = 16, .external_lex_state = 3},
  [3117] = {.lex_state = 16, .external_lex_state = 3},
  [3118] = {.lex_state = 16, .external_lex_state = 3},
  [3119] = {.lex_state = 16, .external_lex_state = 3},
  [3120] = {.lex_state = 16, .external_lex_state = 3},
  [3121] = {.lex_state = 16, .external_lex_state = 3},
  [3122] = {.lex_state = 16, .external_lex_state = 3},
  [3123] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3124] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3125] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [3126] = {.lex_state = 16, .external_lex_state = 3},
  [3127] = {.lex_state = 16, .external_lex_state = 3},
  [3128] = {.lex_state = 16, .external_lex_state = 3},
  [3129] = {.lex_state = 16, .external_lex_state = 3},
  [3130] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 67},
  [3131] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3132] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 68},
  [3133] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3134] = {.lex_state = 16, .external_lex_state = 3},
  [3135] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3136] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3137] = {.lex_state = 16, .external_lex_state = 3},
  [3138] = {.lex_state = 16, .external_lex_state = 3},
  [3139] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 67},
  [3140] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3141] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3142] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3143] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 45},
  [3144] = {.lex_state = 16, .external_lex_state = 3},
  [3145] = {.lex_state = 16, .external_lex_state = 3},
  [3146] = {.lex_state = 16, .external_lex_state = 3},
  [3147] = {.lex_state = 16, .external_lex_state = 3},
  [3148] = {.lex_state = 16, .external_lex_state = 3},
  [3149] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3150] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 67},
  [3151] = {.lex_state = 16, .external_lex_state = 3},
  [3152] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 68},
  [3153] = {.lex_state = 16, .external_lex_state = 3},
  [3154] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3155] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3156] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3157] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 68},
  [3158] = {.lex_state = 16, .external_lex_state = 3},
  [3159] = {.lex_state = 16, .external_lex_state = 3},
  [3160] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [3161] = {.lex_state = 16, .external_lex_state = 3},
  [3162] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 68},
  [3163] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3164] = {.lex_state = 16, .external_lex_state = 3},
  [3165] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 67},
  [3166] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 68},
  [3167] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3168] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3169] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 67},
  [3170] = {.lex_state = 16, .external_lex_state = 3},
  [3171] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3172] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 67},
  [3173] = {.lex_state = 15, .external_lex_state = 2, .reserved_word_set_id = 41},
  [3174] = {.lex_state = 16, .external_lex_state = 3},
  [3175] = {.lex_state = 16, .external_lex_state = 3},
  [3176] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 67},
  [3177] = {.lex_state = 16, .external_lex_state = 3},
  [3178] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3179] = {.lex_state = 122, .external_lex_state = 3},
  [3180] = {.lex_state = 16, .external_lex_state = 3},
  [3181] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3182] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [3183] = {.lex_state = 122, .external_lex_state = 3},
  [3184] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 69},
  [3185] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3186] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3187] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3188] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 69},
  [3189] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3190] = {.lex_state = 122, .external_lex_state = 3},
  [3191] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [3192] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3193] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3194] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3195] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 69},
  [3196] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3197] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 69},
  [3198] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [3199] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3200] = {.lex_state = 122, .external_lex_state = 3},
  [3201] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 45},
  [3202] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 69},
  [3203] = {.lex_state = 16, .external_lex_state = 3},
  [3204] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3205] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3206] = {.lex_state = 16, .external_lex_state = 3},
  [3207] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3208] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3209] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 69},
  [3210] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3211] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3212] = {.lex_state = 122, .external_lex_state = 3},
  [3213] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3214] = {.lex_state = 18, .external_lex_state = 3},
  [3215] = {.lex_state = 122, .external_lex_state = 3},
  [3216] = {.lex_state = 18, .external_lex_state = 3},
  [3217] = {.lex_state = 42, .external_lex_state = 2, .reserved_word_set_id = 48},
  [3218] = {.lex_state = 122, .external_lex_state = 3},
  [3219] = {.lex_state = 122, .external_lex_state = 3},
  [3220] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3221] = {.lex_state = 122, .external_lex_state = 3},
  [3222] = {.lex_state = 18, .external_lex_state = 3},
  [3223] = {.lex_state = 18, .external_lex_state = 3},
  [3224] = {.lex_state = 18, .external_lex_state = 3},
  [3225] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 30},
  [3226] = {.lex_state = 18, .external_lex_state = 3},
  [3227] = {.lex_state = 18, .external_lex_state = 3},
  [3228] = {.lex_state = 122, .external_lex_state = 3},
  [3229] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3230] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3231] = {.lex_state = 16, .external_lex_state = 3},
  [3232] = {.lex_state = 18, .external_lex_state = 3},
  [3233] = {.lex_state = 18, .external_lex_state = 3},
  [3234] = {.lex_state = 16, .external_lex_state = 3},
  [3235] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3236] = {.lex_state = 119, .external_lex_state = 3},
  [3237] = {.lex_state = 119, .external_lex_state = 3},
  [3238] = {.lex_state = 18, .external_lex_state = 3},
  [3239] = {.lex_state = 119, .external_lex_state = 3},
  [3240] = {.lex_state = 119, .external_lex_state = 3},
  [3241] = {.lex_state = 119, .external_lex_state = 3},
  [3242] = {.lex_state = 119, .external_lex_state = 3},
  [3243] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3244] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3245] = {.lex_state = 120, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3246] = {.lex_state = 119, .external_lex_state = 3},
  [3247] = {.lex_state = 18, .external_lex_state = 3},
  [3248] = {.lex_state = 119, .external_lex_state = 3},
  [3249] = {.lex_state = 18, .external_lex_state = 3},
  [3250] = {.lex_state = 18, .external_lex_state = 3},
  [3251] = {.lex_state = 119, .external_lex_state = 3},
  [3252] = {.lex_state = 18, .external_lex_state = 3},
  [3253] = {.lex_state = 16, .external_lex_state = 3},
  [3254] = {.lex_state = 18, .external_lex_state = 3},
  [3255] = {.lex_state = 19, .external_lex_state = 3},
  [3256] = {.lex_state = 19, .external_lex_state = 3},
  [3257] = {.lex_state = 19, .external_lex_state = 3},
  [3258] = {.lex_state = 19, .external_lex_state = 3},
  [3259] = {.lex_state = 19, .external_lex_state = 3},
  [3260] = {.lex_state = 19, .external_lex_state = 3},
  [3261] = {.lex_state = 19, .external_lex_state = 3},
  [3262] = {.lex_state = 19, .external_lex_state = 3},
  [3263] = {.lex_state = 19, .external_lex_state = 3},
  [3264] = {.lex_state = 19, .external_lex_state = 3},
  [3265] = {.lex_state = 19, .external_lex_state = 3},
  [3266] = {.lex_state = 19, .external_lex_state = 3},
  [3267] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 71},
  [3268] = {.lex_state = 19, .external_lex_state = 3},
  [3269] = {.lex_state = 19, .external_lex_state = 3},
  [3270] = {.lex_state = 19, .external_lex_state = 3},
  [3271] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 71},
  [3272] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 71},
  [3273] = {.lex_state = 19, .external_lex_state = 3},
  [3274] = {.lex_state = 19, .external_lex_state = 3},
  [3275] = {.lex_state = 16, .external_lex_state = 3},
  [3276] = {.lex_state = 19, .external_lex_state = 3},
  [3277] = {.lex_state = 16, .external_lex_state = 3},
  [3278] = {.lex_state = 19, .external_lex_state = 3},
  [3279] = {.lex_state = 19, .external_lex_state = 3},
  [3280] = {.lex_state = 19, .external_lex_state = 3},
  [3281] = {.lex_state = 19, .external_lex_state = 3},
  [3282] = {.lex_state = 19, .external_lex_state = 3},
  [3283] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 37},
  [3284] = {.lex_state = 19, .external_lex_state = 3},
  [3285] = {.lex_state = 19, .external_lex_state = 3},
  [3286] = {.lex_state = 19, .external_lex_state = 3},
  [3287] = {.lex_state = 19, .external_lex_state = 3},
  [3288] = {.lex_state = 19, .external_lex_state = 3},
  [3289] = {.lex_state = 19, .external_lex_state = 3},
  [3290] = {.lex_state = 19, .external_lex_state = 3},
  [3291] = {.lex_state = 19, .external_lex_state = 3},
  [3292] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 37},
  [3293] = {.lex_state = 19, .external_lex_state = 3},
  [3294] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 72},
  [3295] = {.lex_state = 19, .external_lex_state = 3},
  [3296] = {.lex_state = 19, .external_lex_state = 3},
  [3297] = {.lex_state = 19, .external_lex_state = 3},
  [3298] = {.lex_state = 18, .external_lex_state = 3},
  [3299] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 72},
  [3300] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3301] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 71},
  [3302] = {.lex_state = 19, .external_lex_state = 3},
  [3303] = {.lex_state = 19, .external_lex_state = 3},
  [3304] = {.lex_state = 19, .external_lex_state = 3},
  [3305] = {.lex_state = 19, .external_lex_state = 3},
  [3306] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 37},
  [3307] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 72},
  [3308] = {.lex_state = 19, .external_lex_state = 3},
  [3309] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 71},
  [3310] = {.lex_state = 19, .external_lex_state = 3},
  [3311] = {.lex_state = 19, .external_lex_state = 3},
  [3312] = {.lex_state = 19, .external_lex_state = 3},
  [3313] = {.lex_state = 19, .external_lex_state = 3},
  [3314] = {.lex_state = 19, .external_lex_state = 3},
  [3315] = {.lex_state = 19, .external_lex_state = 3},
  [3316] = {.lex_state = 19, .external_lex_state = 3},
  [3317] = {.lex_state = 19, .external_lex_state = 3},
  [3318] = {.lex_state = 19, .external_lex_state = 3},
  [3319] = {.lex_state = 19, .external_lex_state = 3},
  [3320] = {.lex_state = 19, .external_lex_state = 3},
  [3321] = {.lex_state = 19, .external_lex_state = 3},
  [3322] = {.lex_state = 19, .external_lex_state = 3},
  [3323] = {.lex_state = 19, .external_lex_state = 3},
  [3324] = {.lex_state = 19, .external_lex_state = 3},
  [3325] = {.lex_state = 19, .external_lex_state = 3},
  [3326] = {.lex_state = 19, .external_lex_state = 3},
  [3327] = {.lex_state = 19, .external_lex_state = 3},
  [3328] = {.lex_state = 19, .external_lex_state = 3},
  [3329] = {.lex_state = 19, .external_lex_state = 3},
  [3330] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 71},
  [3331] = {.lex_state = 19, .external_lex_state = 3},
  [3332] = {.lex_state = 19, .external_lex_state = 3},
  [3333] = {.lex_state = 19, .external_lex_state = 3},
  [3334] = {.lex_state = 19, .external_lex_state = 3},
  [3335] = {.lex_state = 19, .external_lex_state = 3},
  [3336] = {.lex_state = 16, .external_lex_state = 3},
  [3337] = {.lex_state = 19, .external_lex_state = 3},
  [3338] = {.lex_state = 19, .external_lex_state = 3},
  [3339] = {.lex_state = 19, .external_lex_state = 3},
  [3340] = {.lex_state = 19, .external_lex_state = 3},
  [3341] = {.lex_state = 19, .external_lex_state = 3},
  [3342] = {.lex_state = 19, .external_lex_state = 3},
  [3343] = {.lex_state = 19, .external_lex_state = 3},
  [3344] = {.lex_state = 19, .external_lex_state = 3},
  [3345] = {.lex_state = 19, .external_lex_state = 3},
  [3346] = {.lex_state = 19, .external_lex_state = 3},
  [3347] = {.lex_state = 19, .external_lex_state = 3},
  [3348] = {.lex_state = 16, .external_lex_state = 3},
  [3349] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 72},
  [3350] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 72},
  [3351] = {.lex_state = 19, .external_lex_state = 3},
  [3352] = {.lex_state = 19, .external_lex_state = 3},
  [3353] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 72},
  [3354] = {.lex_state = 19, .external_lex_state = 3},
  [3355] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 73},
  [3356] = {.lex_state = 19, .external_lex_state = 3},
  [3357] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 72},
  [3358] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 74},
  [3359] = {.lex_state = 19, .external_lex_state = 3},
  [3360] = {.lex_state = 19, .external_lex_state = 3},
  [3361] = {.lex_state = 19, .external_lex_state = 3},
  [3362] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 72},
  [3363] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 74},
  [3364] = {.lex_state = 19, .external_lex_state = 3},
  [3365] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3366] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3367] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3368] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3369] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3370] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3371] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3372] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3373] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3374] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3375] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3376] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3377] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3378] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3379] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3380] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3381] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3382] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3383] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3384] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 37},
  [3385] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3386] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3387] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3388] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3389] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3390] = {.lex_state = 19, .external_lex_state = 3},
  [3391] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3392] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3393] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 74},
  [3394] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3395] = {.lex_state = 19, .external_lex_state = 3},
  [3396] = {.lex_state = 19, .external_lex_state = 3},
  [3397] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3398] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3399] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3400] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3401] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3402] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3403] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3404] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3405] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3406] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3407] = {.lex_state = 19, .external_lex_state = 3},
  [3408] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3409] = {.lex_state = 19, .external_lex_state = 3},
  [3410] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3411] = {.lex_state = 144, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3412] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3413] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3414] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3415] = {.lex_state = 19, .external_lex_state = 3},
  [3416] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3417] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3418] = {.lex_state = 16, .external_lex_state = 3},
  [3419] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 73},
  [3420] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3421] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3422] = {.lex_state = 19, .external_lex_state = 3},
  [3423] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3424] = {.lex_state = 19, .external_lex_state = 3},
  [3425] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3426] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3427] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3428] = {.lex_state = 16, .external_lex_state = 3},
  [3429] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 74},
  [3430] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 74},
  [3431] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3432] = {.lex_state = 19, .external_lex_state = 3},
  [3433] = {.lex_state = 19, .external_lex_state = 3},
  [3434] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3435] = {.lex_state = 19, .external_lex_state = 3},
  [3436] = {.lex_state = 19, .external_lex_state = 3},
  [3437] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3438] = {.lex_state = 19, .external_lex_state = 3},
  [3439] = {.lex_state = 19, .external_lex_state = 3},
  [3440] = {.lex_state = 19, .external_lex_state = 3},
  [3441] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3442] = {.lex_state = 19, .external_lex_state = 3},
  [3443] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 74},
  [3444] = {.lex_state = 19, .external_lex_state = 3},
  [3445] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3446] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3447] = {.lex_state = 18, .external_lex_state = 3},
  [3448] = {.lex_state = 9, .external_lex_state = 3, .reserved_word_set_id = 76},
  [3449] = {.lex_state = 19, .external_lex_state = 3},
  [3450] = {.lex_state = 18, .external_lex_state = 3},
  [3451] = {.lex_state = 9, .external_lex_state = 3, .reserved_word_set_id = 76},
  [3452] = {.lex_state = 18, .external_lex_state = 3},
  [3453] = {.lex_state = 18, .external_lex_state = 3},
  [3454] = {.lex_state = 18, .external_lex_state = 3},
  [3455] = {.lex_state = 9, .external_lex_state = 3, .reserved_word_set_id = 76},
  [3456] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3457] = {.lex_state = 18, .external_lex_state = 3},
  [3458] = {.lex_state = 18, .external_lex_state = 3},
  [3459] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3460] = {.lex_state = 18, .external_lex_state = 3},
  [3461] = {.lex_state = 18, .external_lex_state = 3},
  [3462] = {.lex_state = 18, .external_lex_state = 3},
  [3463] = {.lex_state = 18, .external_lex_state = 3},
  [3464] = {.lex_state = 18, .external_lex_state = 3},
  [3465] = {.lex_state = 18, .external_lex_state = 3},
  [3466] = {.lex_state = 18, .external_lex_state = 3},
  [3467] = {.lex_state = 18, .external_lex_state = 3},
  [3468] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3469] = {.lex_state = 119, .external_lex_state = 3},
  [3470] = {.lex_state = 25, .external_lex_state = 3},
  [3471] = {.lex_state = 25, .external_lex_state = 3},
  [3472] = {.lex_state = 25, .external_lex_state = 3},
  [3473] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3474] = {.lex_state = 25, .external_lex_state = 3},
  [3475] = {.lex_state = 25, .external_lex_state = 3},
  [3476] = {.lex_state = 25, .external_lex_state = 3},
  [3477] = {.lex_state = 25, .external_lex_state = 3},
  [3478] = {.lex_state = 25, .external_lex_state = 3},
  [3479] = {.lex_state = 25, .external_lex_state = 3},
  [3480] = {.lex_state = 18, .external_lex_state = 3},
  [3481] = {.lex_state = 119, .external_lex_state = 3},
  [3482] = {.lex_state = 119, .external_lex_state = 3},
  [3483] = {.lex_state = 25, .external_lex_state = 3},
  [3484] = {.lex_state = 119, .external_lex_state = 3},
  [3485] = {.lex_state = 25, .external_lex_state = 3},
  [3486] = {.lex_state = 25, .external_lex_state = 3},
  [3487] = {.lex_state = 25, .external_lex_state = 3},
  [3488] = {.lex_state = 25, .external_lex_state = 3},
  [3489] = {.lex_state = 18, .external_lex_state = 3},
  [3490] = {.lex_state = 40, .external_lex_state = 2, .reserved_word_set_id = 37},
  [3491] = {.lex_state = 25, .external_lex_state = 3},
  [3492] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3493] = {.lex_state = 25, .external_lex_state = 3},
  [3494] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3495] = {.lex_state = 25, .external_lex_state = 3},
  [3496] = {.lex_state = 16, .external_lex_state = 3},
  [3497] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3498] = {.lex_state = 19, .external_lex_state = 3},
  [3499] = {.lex_state = 19, .external_lex_state = 3},
  [3500] = {.lex_state = 19, .external_lex_state = 3},
  [3501] = {.lex_state = 19, .external_lex_state = 3},
  [3502] = {.lex_state = 19, .external_lex_state = 3},
  [3503] = {.lex_state = 19, .external_lex_state = 3},
  [3504] = {.lex_state = 39, .external_lex_state = 3},
  [3505] = {.lex_state = 19, .external_lex_state = 3},
  [3506] = {.lex_state = 19, .external_lex_state = 3},
  [3507] = {.lex_state = 19, .external_lex_state = 3},
  [3508] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3509] = {.lex_state = 44, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3510] = {.lex_state = 19, .external_lex_state = 3},
  [3511] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3512] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3513] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3514] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3515] = {.lex_state = 131, .external_lex_state = 3},
  [3516] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3517] = {.lex_state = 131, .external_lex_state = 3},
  [3518] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3519] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3520] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3521] = {.lex_state = 19, .external_lex_state = 3},
  [3522] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3523] = {.lex_state = 19, .external_lex_state = 3},
  [3524] = {.lex_state = 119, .external_lex_state = 3},
  [3525] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3526] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3527] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3528] = {.lex_state = 44, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3529] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3530] = {.lex_state = 44, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3531] = {.lex_state = 131, .external_lex_state = 3},
  [3532] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3533] = {.lex_state = 19, .external_lex_state = 3},
  [3534] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3535] = {.lex_state = 131, .external_lex_state = 3},
  [3536] = {.lex_state = 119, .external_lex_state = 3},
  [3537] = {.lex_state = 44, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3538] = {.lex_state = 16, .external_lex_state = 3},
  [3539] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3540] = {.lex_state = 16, .external_lex_state = 3},
  [3541] = {.lex_state = 16, .external_lex_state = 3},
  [3542] = {.lex_state = 16, .external_lex_state = 3},
  [3543] = {.lex_state = 25, .external_lex_state = 3},
  [3544] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3545] = {.lex_state = 44, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3546] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3547] = {.lex_state = 44, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3548] = {.lex_state = 16, .external_lex_state = 3},
  [3549] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3550] = {.lex_state = 44, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3551] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3552] = {.lex_state = 19, .external_lex_state = 3},
  [3553] = {.lex_state = 46, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3554] = {.lex_state = 19, .external_lex_state = 3},
  [3555] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3556] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3557] = {.lex_state = 19, .external_lex_state = 3},
  [3558] = {.lex_state = 44, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3559] = {.lex_state = 19, .external_lex_state = 3},
  [3560] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3561] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3562] = {.lex_state = 16, .external_lex_state = 3},
  [3563] = {.lex_state = 9, .external_lex_state = 3, .reserved_word_set_id = 76},
  [3564] = {.lex_state = 19, .external_lex_state = 3},
  [3565] = {.lex_state = 19, .external_lex_state = 3},
  [3566] = {.lex_state = 19, .external_lex_state = 3},
  [3567] = {.lex_state = 19, .external_lex_state = 3},
  [3568] = {.lex_state = 19, .external_lex_state = 3},
  [3569] = {.lex_state = 25, .external_lex_state = 3},
  [3570] = {.lex_state = 39, .external_lex_state = 3},
  [3571] = {.lex_state = 19, .external_lex_state = 3},
  [3572] = {.lex_state = 44, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3573] = {.lex_state = 19, .external_lex_state = 3},
  [3574] = {.lex_state = 19, .external_lex_state = 3},
  [3575] = {.lex_state = 39, .external_lex_state = 3},
  [3576] = {.lex_state = 131, .external_lex_state = 3},
  [3577] = {.lex_state = 131, .external_lex_state = 3},
  [3578] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3579] = {.lex_state = 39, .external_lex_state = 3},
  [3580] = {.lex_state = 119, .external_lex_state = 3},
  [3581] = {.lex_state = 16, .external_lex_state = 3},
  [3582] = {.lex_state = 19, .external_lex_state = 3},
  [3583] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3584] = {.lex_state = 19, .external_lex_state = 3},
  [3585] = {.lex_state = 19, .external_lex_state = 3},
  [3586] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3587] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3588] = {.lex_state = 39, .external_lex_state = 3},
  [3589] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3590] = {.lex_state = 19, .external_lex_state = 3},
  [3591] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3592] = {.lex_state = 19, .external_lex_state = 3},
  [3593] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3594] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3595] = {.lex_state = 44, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3596] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3597] = {.lex_state = 19, .external_lex_state = 3},
  [3598] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3599] = {.lex_state = 19, .external_lex_state = 3},
  [3600] = {.lex_state = 119, .external_lex_state = 3},
  [3601] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 75},
  [3602] = {.lex_state = 119, .external_lex_state = 3},
  [3603] = {.lex_state = 16, .external_lex_state = 3},
  [3604] = {.lex_state = 25, .external_lex_state = 3},
  [3605] = {.lex_state = 16, .external_lex_state = 3},
  [3606] = {.lex_state = 16, .external_lex_state = 3},
  [3607] = {.lex_state = 16, .external_lex_state = 3},
  [3608] = {.lex_state = 16, .external_lex_state = 3},
  [3609] = {.lex_state = 16, .external_lex_state = 3},
  [3610] = {.lex_state = 16, .external_lex_state = 3},
  [3611] = {.lex_state = 16, .external_lex_state = 3},
  [3612] = {.lex_state = 16, .external_lex_state = 3},
  [3613] = {.lex_state = 144, .external_lex_state = 3, .reserved_word_set_id = 79},
  [3614] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3615] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3616] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3617] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3618] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3619] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3620] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3621] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3622] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3623] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3624] = {.lex_state = 119, .external_lex_state = 3},
  [3625] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3626] = {.lex_state = 16, .external_lex_state = 3},
  [3627] = {.lex_state = 16, .external_lex_state = 3},
  [3628] = {.lex_state = 16, .external_lex_state = 3},
  [3629] = {.lex_state = 16, .external_lex_state = 3},
  [3630] = {.lex_state = 16, .external_lex_state = 3},
  [3631] = {.lex_state = 16, .external_lex_state = 3},
  [3632] = {.lex_state = 16, .external_lex_state = 3},
  [3633] = {.lex_state = 16, .external_lex_state = 3},
  [3634] = {.lex_state = 16, .external_lex_state = 3},
  [3635] = {.lex_state = 16, .external_lex_state = 3},
  [3636] = {.lex_state = 16, .external_lex_state = 3},
  [3637] = {.lex_state = 16, .external_lex_state = 3},
  [3638] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3639] = {.lex_state = 16, .external_lex_state = 3},
  [3640] = {.lex_state = 16, .external_lex_state = 3},
  [3641] = {.lex_state = 7, .external_lex_state = 3},
  [3642] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 80},
  [3643] = {.lex_state = 16, .external_lex_state = 3},
  [3644] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3645] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3646] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3647] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3648] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 81},
  [3649] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3650] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3651] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3652] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3653] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3654] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3655] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3656] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3657] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3658] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3659] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3660] = {.lex_state = 16, .external_lex_state = 3},
  [3661] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3662] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3663] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3664] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3665] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3666] = {.lex_state = 16, .external_lex_state = 3},
  [3667] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3668] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3669] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3670] = {.lex_state = 119, .external_lex_state = 3},
  [3671] = {.lex_state = 122, .external_lex_state = 3},
  [3672] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3673] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3674] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3675] = {.lex_state = 119, .external_lex_state = 3},
  [3676] = {.lex_state = 16, .external_lex_state = 3},
  [3677] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3678] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3679] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3680] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3681] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3682] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3683] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3684] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3685] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3686] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3687] = {.lex_state = 122, .external_lex_state = 3},
  [3688] = {.lex_state = 119, .external_lex_state = 3},
  [3689] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3690] = {.lex_state = 119, .external_lex_state = 3},
  [3691] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3692] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3693] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3694] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 81},
  [3695] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3696] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3697] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 80},
  [3698] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 80},
  [3699] = {.lex_state = 119, .external_lex_state = 3},
  [3700] = {.lex_state = 9, .external_lex_state = 3, .reserved_word_set_id = 76},
  [3701] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3702] = {.lex_state = 119, .external_lex_state = 3},
  [3703] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3704] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3705] = {.lex_state = 122, .external_lex_state = 3},
  [3706] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3707] = {.lex_state = 122, .external_lex_state = 3},
  [3708] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3709] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3710] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3711] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3712] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3713] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3714] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3715] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3716] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3717] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3718] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3719] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3720] = {.lex_state = 119, .external_lex_state = 3},
  [3721] = {.lex_state = 7, .external_lex_state = 3},
  [3722] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3723] = {.lex_state = 7, .external_lex_state = 3},
  [3724] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3725] = {.lex_state = 16, .external_lex_state = 3},
  [3726] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3727] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3728] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3729] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3730] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3731] = {.lex_state = 16, .external_lex_state = 3},
  [3732] = {.lex_state = 16, .external_lex_state = 3},
  [3733] = {.lex_state = 119, .external_lex_state = 3},
  [3734] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3735] = {.lex_state = 119, .external_lex_state = 3},
  [3736] = {.lex_state = 16, .external_lex_state = 3},
  [3737] = {.lex_state = 122, .external_lex_state = 3},
  [3738] = {.lex_state = 119, .external_lex_state = 3},
  [3739] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3740] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3741] = {.lex_state = 119, .external_lex_state = 3},
  [3742] = {.lex_state = 7, .external_lex_state = 3},
  [3743] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3744] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3745] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3746] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 80},
  [3747] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3748] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3749] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3750] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3751] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3752] = {.lex_state = 119, .external_lex_state = 3},
  [3753] = {.lex_state = 119, .external_lex_state = 3},
  [3754] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3755] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3756] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3757] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 77},
  [3758] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3759] = {.lex_state = 43, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3760] = {.lex_state = 16, .external_lex_state = 3},
  [3761] = {.lex_state = 119, .external_lex_state = 3},
  [3762] = {.lex_state = 19, .external_lex_state = 3},
  [3763] = {.lex_state = 19, .external_lex_state = 3},
  [3764] = {.lex_state = 56, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3765] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3766] = {.lex_state = 19, .external_lex_state = 3},
  [3767] = {.lex_state = 19, .external_lex_state = 3},
  [3768] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3769] = {.lex_state = 16, .external_lex_state = 3},
  [3770] = {.lex_state = 7, .external_lex_state = 3},
  [3771] = {.lex_state = 16, .external_lex_state = 3},
  [3772] = {.lex_state = 16, .external_lex_state = 3},
  [3773] = {.lex_state = 16, .external_lex_state = 3},
  [3774] = {.lex_state = 16, .external_lex_state = 3},
  [3775] = {.lex_state = 119, .external_lex_state = 3},
  [3776] = {.lex_state = 119, .external_lex_state = 3},
  [3777] = {.lex_state = 16, .external_lex_state = 3},
  [3778] = {.lex_state = 16, .external_lex_state = 3},
  [3779] = {.lex_state = 16, .external_lex_state = 3},
  [3780] = {.lex_state = 16, .external_lex_state = 3},
  [3781] = {.lex_state = 16, .external_lex_state = 3},
  [3782] = {.lex_state = 16, .external_lex_state = 3},
  [3783] = {.lex_state = 16, .external_lex_state = 3},
  [3784] = {.lex_state = 16, .external_lex_state = 3},
  [3785] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 81},
  [3786] = {.lex_state = 16, .external_lex_state = 3},
  [3787] = {.lex_state = 16, .external_lex_state = 3},
  [3788] = {.lex_state = 16, .external_lex_state = 3},
  [3789] = {.lex_state = 19, .external_lex_state = 3},
  [3790] = {.lex_state = 7, .external_lex_state = 3},
  [3791] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3792] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3793] = {.lex_state = 16, .external_lex_state = 3},
  [3794] = {.lex_state = 19, .external_lex_state = 3},
  [3795] = {.lex_state = 19, .external_lex_state = 3},
  [3796] = {.lex_state = 16, .external_lex_state = 3},
  [3797] = {.lex_state = 119, .external_lex_state = 3},
  [3798] = {.lex_state = 16, .external_lex_state = 3},
  [3799] = {.lex_state = 18, .external_lex_state = 3},
  [3800] = {.lex_state = 119, .external_lex_state = 3},
  [3801] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3802] = {.lex_state = 19, .external_lex_state = 3},
  [3803] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3804] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3805] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3806] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3807] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3808] = {.lex_state = 16, .external_lex_state = 3},
  [3809] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3810] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 83},
  [3811] = {.lex_state = 18, .external_lex_state = 3},
  [3812] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3813] = {.lex_state = 16, .external_lex_state = 3},
  [3814] = {.lex_state = 119, .external_lex_state = 3},
  [3815] = {.lex_state = 119, .external_lex_state = 3},
  [3816] = {.lex_state = 16, .external_lex_state = 3},
  [3817] = {.lex_state = 18, .external_lex_state = 3},
  [3818] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3819] = {.lex_state = 16, .external_lex_state = 3},
  [3820] = {.lex_state = 16, .external_lex_state = 3},
  [3821] = {.lex_state = 42, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3822] = {.lex_state = 119, .external_lex_state = 3},
  [3823] = {.lex_state = 18, .external_lex_state = 3},
  [3824] = {.lex_state = 119, .external_lex_state = 3},
  [3825] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3826] = {.lex_state = 16, .external_lex_state = 3},
  [3827] = {.lex_state = 16, .external_lex_state = 3},
  [3828] = {.lex_state = 16, .external_lex_state = 3},
  [3829] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3830] = {.lex_state = 18, .external_lex_state = 3},
  [3831] = {.lex_state = 18, .external_lex_state = 3},
  [3832] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 81},
  [3833] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3834] = {.lex_state = 119, .external_lex_state = 3},
  [3835] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3836] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3837] = {.lex_state = 25, .external_lex_state = 3},
  [3838] = {.lex_state = 119, .external_lex_state = 3},
  [3839] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3840] = {.lex_state = 119, .external_lex_state = 3},
  [3841] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3842] = {.lex_state = 119, .external_lex_state = 3},
  [3843] = {.lex_state = 18, .external_lex_state = 3},
  [3844] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3845] = {.lex_state = 42, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3846] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3847] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3848] = {.lex_state = 25, .external_lex_state = 3},
  [3849] = {.lex_state = 119, .external_lex_state = 3},
  [3850] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3851] = {.lex_state = 16, .external_lex_state = 3},
  [3852] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3853] = {.lex_state = 119, .external_lex_state = 3},
  [3854] = {.lex_state = 119, .external_lex_state = 3},
  [3855] = {.lex_state = 16, .external_lex_state = 3},
  [3856] = {.lex_state = 119, .external_lex_state = 3},
  [3857] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3858] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3859] = {.lex_state = 119, .external_lex_state = 3},
  [3860] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3861] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3862] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3863] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 84},
  [3864] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3865] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3866] = {.lex_state = 119, .external_lex_state = 3},
  [3867] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3868] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3869] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3870] = {.lex_state = 119, .external_lex_state = 3},
  [3871] = {.lex_state = 119, .external_lex_state = 3},
  [3872] = {.lex_state = 18, .external_lex_state = 3},
  [3873] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3874] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3875] = {.lex_state = 16, .external_lex_state = 3},
  [3876] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3877] = {.lex_state = 119, .external_lex_state = 3},
  [3878] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3879] = {.lex_state = 42, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3880] = {.lex_state = 42, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3881] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3882] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3883] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3884] = {.lex_state = 42, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3885] = {.lex_state = 42, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3886] = {.lex_state = 119, .external_lex_state = 3},
  [3887] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3888] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3889] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3890] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3891] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3892] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3893] = {.lex_state = 119, .external_lex_state = 3},
  [3894] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3895] = {.lex_state = 119, .external_lex_state = 3},
  [3896] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3897] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3898] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3899] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3900] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3901] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3902] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3903] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 81},
  [3904] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3905] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3906] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 81},
  [3907] = {.lex_state = 131, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3908] = {.lex_state = 119, .external_lex_state = 3},
  [3909] = {.lex_state = 7, .external_lex_state = 3},
  [3910] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3911] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3912] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3913] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3914] = {.lex_state = 119, .external_lex_state = 3},
  [3915] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3916] = {.lex_state = 119, .external_lex_state = 3},
  [3917] = {.lex_state = 119, .external_lex_state = 3},
  [3918] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3919] = {.lex_state = 119, .external_lex_state = 3},
  [3920] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3921] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3922] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3923] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3924] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3925] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3926] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3927] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3928] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3929] = {.lex_state = 25, .external_lex_state = 3},
  [3930] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3931] = {.lex_state = 16, .external_lex_state = 3},
  [3932] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3933] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 82},
  [3934] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 85},
  [3935] = {.lex_state = 16, .external_lex_state = 3},
  [3936] = {.lex_state = 143, .external_lex_state = 3},
  [3937] = {.lex_state = 16, .external_lex_state = 3},
  [3938] = {.lex_state = 143, .external_lex_state = 3},
  [3939] = {.lex_state = 16, .external_lex_state = 3},
  [3940] = {.lex_state = 119, .external_lex_state = 3},
  [3941] = {.lex_state = 131, .external_lex_state = 3},
  [3942] = {.lex_state = 16, .external_lex_state = 3},
  [3943] = {.lex_state = 119, .external_lex_state = 3},
  [3944] = {.lex_state = 51, .external_lex_state = 3},
  [3945] = {.lex_state = 122, .external_lex_state = 2},
  [3946] = {.lex_state = 122, .external_lex_state = 2},
  [3947] = {.lex_state = 119, .external_lex_state = 3},
  [3948] = {.lex_state = 119, .external_lex_state = 3},
  [3949] = {.lex_state = 16, .external_lex_state = 3},
  [3950] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 84},
  [3951] = {.lex_state = 119, .external_lex_state = 3},
  [3952] = {.lex_state = 136, .external_lex_state = 3, .reserved_word_set_id = 84},
  [3953] = {.lex_state = 16, .external_lex_state = 3},
  [3954] = {.lex_state = 51, .external_lex_state = 3},
  [3955] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3956] = {.lex_state = 51, .external_lex_state = 3},
  [3957] = {.lex_state = 51, .external_lex_state = 3},
  [3958] = {.lex_state = 119, .external_lex_state = 3},
  [3959] = {.lex_state = 16, .external_lex_state = 3},
  [3960] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 85},
  [3961] = {.lex_state = 16, .external_lex_state = 3},
  [3962] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3963] = {.lex_state = 56, .external_lex_state = 3, .reserved_word_set_id = 86},
  [3964] = {.lex_state = 119, .external_lex_state = 3},
  [3965] = {.lex_state = 119, .external_lex_state = 3},
  [3966] = {.lex_state = 131, .external_lex_state = 3},
  [3967] = {.lex_state = 51, .external_lex_state = 3},
  [3968] = {.lex_state = 119, .external_lex_state = 3},
  [3969] = {.lex_state = 131, .external_lex_state = 3},
  [3970] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3971] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3972] = {.lex_state = 16, .external_lex_state = 3},
  [3973] = {.lex_state = 131, .external_lex_state = 3},
  [3974] = {.lex_state = 119, .external_lex_state = 3},
  [3975] = {.lex_state = 119, .external_lex_state = 3},
  [3976] = {.lex_state = 131, .external_lex_state = 3},
  [3977] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3978] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3979] = {.lex_state = 131, .external_lex_state = 3},
  [3980] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3981] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [3982] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3983] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3984] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 88},
  [3985] = {.lex_state = 131, .external_lex_state = 3},
  [3986] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [3987] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [3988] = {.lex_state = 119, .external_lex_state = 3},
  [3989] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [3990] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [3991] = {.lex_state = 119, .external_lex_state = 3},
  [3992] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3993] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3994] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3995] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [3996] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [3997] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [3998] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3999] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4000] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4001] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4002] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4003] = {.lex_state = 119, .external_lex_state = 3},
  [4004] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4005] = {.lex_state = 131, .external_lex_state = 3},
  [4006] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4007] = {.lex_state = 119, .external_lex_state = 3},
  [4008] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4009] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4010] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4011] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4012] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4013] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4014] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4015] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4016] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4017] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4018] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4019] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4020] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4021] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4022] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4023] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4024] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4025] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4026] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4027] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4028] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4029] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4030] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4031] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4032] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4033] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4034] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4035] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4036] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4037] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4038] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4039] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 85},
  [4040] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4041] = {.lex_state = 119, .external_lex_state = 3},
  [4042] = {.lex_state = 119, .external_lex_state = 3},
  [4043] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4044] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4045] = {.lex_state = 119, .external_lex_state = 3},
  [4046] = {.lex_state = 131, .external_lex_state = 3},
  [4047] = {.lex_state = 131, .external_lex_state = 3},
  [4048] = {.lex_state = 131, .external_lex_state = 3},
  [4049] = {.lex_state = 131, .external_lex_state = 3},
  [4050] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4051] = {.lex_state = 119, .external_lex_state = 3},
  [4052] = {.lex_state = 119, .external_lex_state = 3},
  [4053] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4054] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4055] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4056] = {.lex_state = 119, .external_lex_state = 3},
  [4057] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4058] = {.lex_state = 64, .external_lex_state = 3},
  [4059] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4060] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4061] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4062] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4063] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4064] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4065] = {.lex_state = 16, .external_lex_state = 3},
  [4066] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4067] = {.lex_state = 64, .external_lex_state = 3},
  [4068] = {.lex_state = 16, .external_lex_state = 3},
  [4069] = {.lex_state = 16, .external_lex_state = 3},
  [4070] = {.lex_state = 119, .external_lex_state = 3},
  [4071] = {.lex_state = 131, .external_lex_state = 3},
  [4072] = {.lex_state = 119, .external_lex_state = 3},
  [4073] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4074] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4075] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4076] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4077] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4078] = {.lex_state = 119, .external_lex_state = 3},
  [4079] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4080] = {.lex_state = 16, .external_lex_state = 3},
  [4081] = {.lex_state = 119, .external_lex_state = 3},
  [4082] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4083] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4084] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4085] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4086] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4087] = {.lex_state = 131, .external_lex_state = 3},
  [4088] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4089] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4090] = {.lex_state = 131, .external_lex_state = 3},
  [4091] = {.lex_state = 131, .external_lex_state = 3},
  [4092] = {.lex_state = 131, .external_lex_state = 3},
  [4093] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4094] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4095] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4096] = {.lex_state = 119, .external_lex_state = 3},
  [4097] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4098] = {.lex_state = 131, .external_lex_state = 3},
  [4099] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4100] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4101] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4102] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4103] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4104] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4105] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4106] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4107] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4108] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4109] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4110] = {.lex_state = 119, .external_lex_state = 3},
  [4111] = {.lex_state = 119, .external_lex_state = 3},
  [4112] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4113] = {.lex_state = 131, .external_lex_state = 3},
  [4114] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4115] = {.lex_state = 16, .external_lex_state = 3},
  [4116] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4117] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4118] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4119] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4120] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4121] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4122] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4123] = {.lex_state = 131, .external_lex_state = 3},
  [4124] = {.lex_state = 16, .external_lex_state = 3},
  [4125] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4126] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4127] = {.lex_state = 131, .external_lex_state = 3},
  [4128] = {.lex_state = 64, .external_lex_state = 3},
  [4129] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 87},
  [4130] = {.lex_state = 119, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4131] = {.lex_state = 119, .external_lex_state = 3},
  [4132] = {.lex_state = 130, .external_lex_state = 3},
  [4133] = {.lex_state = 119, .external_lex_state = 3},
  [4134] = {.lex_state = 119, .external_lex_state = 3},
  [4135] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 88},
  [4136] = {.lex_state = 64, .external_lex_state = 3},
  [4137] = {.lex_state = 64, .external_lex_state = 3},
  [4138] = {.lex_state = 64, .external_lex_state = 3},
  [4139] = {.lex_state = 119, .external_lex_state = 3},
  [4140] = {.lex_state = 130, .external_lex_state = 3},
  [4141] = {.lex_state = 130, .external_lex_state = 3},
  [4142] = {.lex_state = 16, .external_lex_state = 3},
  [4143] = {.lex_state = 64, .external_lex_state = 3},
  [4144] = {.lex_state = 119, .external_lex_state = 3},
  [4145] = {.lex_state = 130, .external_lex_state = 3},
  [4146] = {.lex_state = 130, .external_lex_state = 3},
  [4147] = {.lex_state = 130, .external_lex_state = 3},
  [4148] = {.lex_state = 119, .external_lex_state = 3},
  [4149] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 85},
  [4150] = {.lex_state = 130, .external_lex_state = 3},
  [4151] = {.lex_state = 130, .external_lex_state = 3},
  [4152] = {.lex_state = 119, .external_lex_state = 3},
  [4153] = {.lex_state = 130, .external_lex_state = 3},
  [4154] = {.lex_state = 130, .external_lex_state = 3},
  [4155] = {.lex_state = 130, .external_lex_state = 3},
  [4156] = {.lex_state = 130, .external_lex_state = 3},
  [4157] = {.lex_state = 130, .external_lex_state = 3},
  [4158] = {.lex_state = 119, .external_lex_state = 3},
  [4159] = {.lex_state = 131, .external_lex_state = 3},
  [4160] = {.lex_state = 131, .external_lex_state = 3},
  [4161] = {.lex_state = 119, .external_lex_state = 3},
  [4162] = {.lex_state = 131, .external_lex_state = 3},
  [4163] = {.lex_state = 143, .external_lex_state = 3},
  [4164] = {.lex_state = 122, .external_lex_state = 2},
  [4165] = {.lex_state = 64, .external_lex_state = 3},
  [4166] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 88},
  [4167] = {.lex_state = 130, .external_lex_state = 3},
  [4168] = {.lex_state = 130, .external_lex_state = 3},
  [4169] = {.lex_state = 130, .external_lex_state = 3},
  [4170] = {.lex_state = 130, .external_lex_state = 3},
  [4171] = {.lex_state = 130, .external_lex_state = 3},
  [4172] = {.lex_state = 64, .external_lex_state = 3},
  [4173] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4174] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 85},
  [4175] = {.lex_state = 64, .external_lex_state = 3},
  [4176] = {.lex_state = 64, .external_lex_state = 3},
  [4177] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4178] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4179] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4180] = {.lex_state = 119, .external_lex_state = 3},
  [4181] = {.lex_state = 42, .external_lex_state = 3, .reserved_word_set_id = 90},
  [4182] = {.lex_state = 64, .external_lex_state = 3},
  [4183] = {.lex_state = 119, .external_lex_state = 3},
  [4184] = {.lex_state = 64, .external_lex_state = 3},
  [4185] = {.lex_state = 64, .external_lex_state = 3},
  [4186] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 91},
  [4187] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 85},
  [4188] = {.lex_state = 42, .external_lex_state = 3, .reserved_word_set_id = 90},
  [4189] = {.lex_state = 143, .external_lex_state = 3},
  [4190] = {.lex_state = 42, .external_lex_state = 3, .reserved_word_set_id = 90},
  [4191] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4192] = {.lex_state = 143, .external_lex_state = 3},
  [4193] = {.lex_state = 119, .external_lex_state = 3},
  [4194] = {.lex_state = 119, .external_lex_state = 3},
  [4195] = {.lex_state = 119, .external_lex_state = 3},
  [4196] = {.lex_state = 42, .external_lex_state = 3, .reserved_word_set_id = 90},
  [4197] = {.lex_state = 119, .external_lex_state = 3},
  [4198] = {.lex_state = 143, .external_lex_state = 3},
  [4199] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4200] = {.lex_state = 64, .external_lex_state = 3},
  [4201] = {.lex_state = 143, .external_lex_state = 3},
  [4202] = {.lex_state = 119, .external_lex_state = 3},
  [4203] = {.lex_state = 119, .external_lex_state = 3},
  [4204] = {.lex_state = 119, .external_lex_state = 3},
  [4205] = {.lex_state = 131, .external_lex_state = 3},
  [4206] = {.lex_state = 119, .external_lex_state = 3},
  [4207] = {.lex_state = 131, .external_lex_state = 3},
  [4208] = {.lex_state = 131, .external_lex_state = 3},
  [4209] = {.lex_state = 7, .external_lex_state = 3},
  [4210] = {.lex_state = 119, .external_lex_state = 3},
  [4211] = {.lex_state = 131, .external_lex_state = 3},
  [4212] = {.lex_state = 131, .external_lex_state = 3},
  [4213] = {.lex_state = 131, .external_lex_state = 3},
  [4214] = {.lex_state = 7, .external_lex_state = 3},
  [4215] = {.lex_state = 119, .external_lex_state = 3},
  [4216] = {.lex_state = 131, .external_lex_state = 3},
  [4217] = {.lex_state = 130, .external_lex_state = 3},
  [4218] = {.lex_state = 131, .external_lex_state = 3},
  [4219] = {.lex_state = 131, .external_lex_state = 3},
  [4220] = {.lex_state = 131, .external_lex_state = 3},
  [4221] = {.lex_state = 35, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4222] = {.lex_state = 119, .external_lex_state = 3},
  [4223] = {.lex_state = 131, .external_lex_state = 3},
  [4224] = {.lex_state = 131, .external_lex_state = 3},
  [4225] = {.lex_state = 131, .external_lex_state = 3},
  [4226] = {.lex_state = 131, .external_lex_state = 3},
  [4227] = {.lex_state = 131, .external_lex_state = 3},
  [4228] = {.lex_state = 119, .external_lex_state = 3},
  [4229] = {.lex_state = 130, .external_lex_state = 3},
  [4230] = {.lex_state = 131, .external_lex_state = 3},
  [4231] = {.lex_state = 131, .external_lex_state = 3},
  [4232] = {.lex_state = 7, .external_lex_state = 3},
  [4233] = {.lex_state = 7, .external_lex_state = 3},
  [4234] = {.lex_state = 131, .external_lex_state = 3},
  [4235] = {.lex_state = 7, .external_lex_state = 3},
  [4236] = {.lex_state = 131, .external_lex_state = 3},
  [4237] = {.lex_state = 131, .external_lex_state = 3},
  [4238] = {.lex_state = 131, .external_lex_state = 3},
  [4239] = {.lex_state = 7, .external_lex_state = 3},
  [4240] = {.lex_state = 119, .external_lex_state = 3},
  [4241] = {.lex_state = 131, .external_lex_state = 3},
  [4242] = {.lex_state = 122, .external_lex_state = 3},
  [4243] = {.lex_state = 131, .external_lex_state = 3},
  [4244] = {.lex_state = 7, .external_lex_state = 3},
  [4245] = {.lex_state = 131, .external_lex_state = 3},
  [4246] = {.lex_state = 131, .external_lex_state = 3},
  [4247] = {.lex_state = 131, .external_lex_state = 3},
  [4248] = {.lex_state = 119, .external_lex_state = 3},
  [4249] = {.lex_state = 131, .external_lex_state = 3},
  [4250] = {.lex_state = 131, .external_lex_state = 3},
  [4251] = {.lex_state = 131, .external_lex_state = 3},
  [4252] = {.lex_state = 131, .external_lex_state = 3},
  [4253] = {.lex_state = 119, .external_lex_state = 3},
  [4254] = {.lex_state = 131, .external_lex_state = 3},
  [4255] = {.lex_state = 119, .external_lex_state = 3},
  [4256] = {.lex_state = 131, .external_lex_state = 3},
  [4257] = {.lex_state = 131, .external_lex_state = 3},
  [4258] = {.lex_state = 131, .external_lex_state = 3},
  [4259] = {.lex_state = 131, .external_lex_state = 3},
  [4260] = {.lex_state = 131, .external_lex_state = 3},
  [4261] = {.lex_state = 131, .external_lex_state = 3},
  [4262] = {.lex_state = 131, .external_lex_state = 3},
  [4263] = {.lex_state = 131, .external_lex_state = 3},
  [4264] = {.lex_state = 131, .external_lex_state = 3},
  [4265] = {.lex_state = 131, .external_lex_state = 3},
  [4266] = {.lex_state = 131, .external_lex_state = 3},
  [4267] = {.lex_state = 131, .external_lex_state = 3},
  [4268] = {.lex_state = 131, .external_lex_state = 3},
  [4269] = {.lex_state = 119, .external_lex_state = 3},
  [4270] = {.lex_state = 7, .external_lex_state = 3},
  [4271] = {.lex_state = 119, .external_lex_state = 3},
  [4272] = {.lex_state = 131, .external_lex_state = 3},
  [4273] = {.lex_state = 64, .external_lex_state = 3},
  [4274] = {.lex_state = 7, .external_lex_state = 2},
  [4275] = {.lex_state = 131, .external_lex_state = 3},
  [4276] = {.lex_state = 131, .external_lex_state = 3},
  [4277] = {.lex_state = 131, .external_lex_state = 3},
  [4278] = {.lex_state = 131, .external_lex_state = 3},
  [4279] = {.lex_state = 119, .external_lex_state = 3},
  [4280] = {.lex_state = 131, .external_lex_state = 3},
  [4281] = {.lex_state = 119, .external_lex_state = 3},
  [4282] = {.lex_state = 64, .external_lex_state = 3},
  [4283] = {.lex_state = 7, .external_lex_state = 3},
  [4284] = {.lex_state = 51, .external_lex_state = 3},
  [4285] = {.lex_state = 131, .external_lex_state = 3},
  [4286] = {.lex_state = 7, .external_lex_state = 3},
  [4287] = {.lex_state = 131, .external_lex_state = 3},
  [4288] = {.lex_state = 131, .external_lex_state = 3},
  [4289] = {.lex_state = 131, .external_lex_state = 3},
  [4290] = {.lex_state = 131, .external_lex_state = 3},
  [4291] = {.lex_state = 131, .external_lex_state = 3},
  [4292] = {.lex_state = 131, .external_lex_state = 3},
  [4293] = {.lex_state = 131, .external_lex_state = 3},
  [4294] = {.lex_state = 131, .external_lex_state = 3},
  [4295] = {.lex_state = 131, .external_lex_state = 3},
  [4296] = {.lex_state = 131, .external_lex_state = 3},
  [4297] = {.lex_state = 131, .external_lex_state = 3},
  [4298] = {.lex_state = 131, .external_lex_state = 3},
  [4299] = {.lex_state = 131, .external_lex_state = 3},
  [4300] = {.lex_state = 131, .external_lex_state = 3},
  [4301] = {.lex_state = 131, .external_lex_state = 3},
  [4302] = {.lex_state = 131, .external_lex_state = 3},
  [4303] = {.lex_state = 7, .external_lex_state = 3},
  [4304] = {.lex_state = 131, .external_lex_state = 3},
  [4305] = {.lex_state = 131, .external_lex_state = 3},
  [4306] = {.lex_state = 131, .external_lex_state = 3},
  [4307] = {.lex_state = 119, .external_lex_state = 3},
  [4308] = {.lex_state = 35, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4309] = {.lex_state = 131, .external_lex_state = 3},
  [4310] = {.lex_state = 7, .external_lex_state = 2},
  [4311] = {.lex_state = 131, .external_lex_state = 3},
  [4312] = {.lex_state = 131, .external_lex_state = 3},
  [4313] = {.lex_state = 131, .external_lex_state = 3},
  [4314] = {.lex_state = 131, .external_lex_state = 3},
  [4315] = {.lex_state = 7, .external_lex_state = 3},
  [4316] = {.lex_state = 131, .external_lex_state = 3},
  [4317] = {.lex_state = 131, .external_lex_state = 3},
  [4318] = {.lex_state = 131, .external_lex_state = 3},
  [4319] = {.lex_state = 131, .external_lex_state = 3},
  [4320] = {.lex_state = 35, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4321] = {.lex_state = 131, .external_lex_state = 3},
  [4322] = {.lex_state = 131, .external_lex_state = 3},
  [4323] = {.lex_state = 131, .external_lex_state = 3},
  [4324] = {.lex_state = 131, .external_lex_state = 3},
  [4325] = {.lex_state = 131, .external_lex_state = 3},
  [4326] = {.lex_state = 131, .external_lex_state = 3},
  [4327] = {.lex_state = 119, .external_lex_state = 3},
  [4328] = {.lex_state = 131, .external_lex_state = 3},
  [4329] = {.lex_state = 131, .external_lex_state = 3},
  [4330] = {.lex_state = 7, .external_lex_state = 3},
  [4331] = {.lex_state = 7, .external_lex_state = 3},
  [4332] = {.lex_state = 131, .external_lex_state = 3},
  [4333] = {.lex_state = 119, .external_lex_state = 3},
  [4334] = {.lex_state = 131, .external_lex_state = 3},
  [4335] = {.lex_state = 131, .external_lex_state = 3},
  [4336] = {.lex_state = 122, .external_lex_state = 3},
  [4337] = {.lex_state = 130, .external_lex_state = 3},
  [4338] = {.lex_state = 131, .external_lex_state = 3},
  [4339] = {.lex_state = 131, .external_lex_state = 3},
  [4340] = {.lex_state = 131, .external_lex_state = 3},
  [4341] = {.lex_state = 131, .external_lex_state = 3},
  [4342] = {.lex_state = 131, .external_lex_state = 3},
  [4343] = {.lex_state = 131, .external_lex_state = 3},
  [4344] = {.lex_state = 131, .external_lex_state = 3},
  [4345] = {.lex_state = 131, .external_lex_state = 3},
  [4346] = {.lex_state = 131, .external_lex_state = 3},
  [4347] = {.lex_state = 131, .external_lex_state = 3},
  [4348] = {.lex_state = 131, .external_lex_state = 3},
  [4349] = {.lex_state = 131, .external_lex_state = 3},
  [4350] = {.lex_state = 131, .external_lex_state = 3},
  [4351] = {.lex_state = 131, .external_lex_state = 3},
  [4352] = {.lex_state = 131, .external_lex_state = 3},
  [4353] = {.lex_state = 131, .external_lex_state = 3},
  [4354] = {.lex_state = 131, .external_lex_state = 3},
  [4355] = {.lex_state = 64, .external_lex_state = 3},
  [4356] = {.lex_state = 131, .external_lex_state = 3},
  [4357] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4358] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4359] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4360] = {.lex_state = 35, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4361] = {.lex_state = 122, .external_lex_state = 3},
  [4362] = {.lex_state = 52, .external_lex_state = 3},
  [4363] = {.lex_state = 122, .external_lex_state = 3},
  [4364] = {.lex_state = 122, .external_lex_state = 3},
  [4365] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4366] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4367] = {.lex_state = 54, .external_lex_state = 3},
  [4368] = {.lex_state = 122, .external_lex_state = 2},
  [4369] = {.lex_state = 119, .external_lex_state = 3},
  [4370] = {.lex_state = 54, .external_lex_state = 3},
  [4371] = {.lex_state = 122, .external_lex_state = 3},
  [4372] = {.lex_state = 119, .external_lex_state = 3},
  [4373] = {.lex_state = 122, .external_lex_state = 3},
  [4374] = {.lex_state = 119, .external_lex_state = 3},
  [4375] = {.lex_state = 64, .external_lex_state = 3},
  [4376] = {.lex_state = 130, .external_lex_state = 3},
  [4377] = {.lex_state = 130, .external_lex_state = 3},
  [4378] = {.lex_state = 130, .external_lex_state = 3},
  [4379] = {.lex_state = 130, .external_lex_state = 3},
  [4380] = {.lex_state = 122, .external_lex_state = 3},
  [4381] = {.lex_state = 130, .external_lex_state = 3},
  [4382] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4383] = {.lex_state = 122, .external_lex_state = 3},
  [4384] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4385] = {.lex_state = 122, .external_lex_state = 3},
  [4386] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4387] = {.lex_state = 122, .external_lex_state = 3},
  [4388] = {.lex_state = 122, .external_lex_state = 2},
  [4389] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4390] = {.lex_state = 131, .external_lex_state = 3},
  [4391] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4392] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4393] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4394] = {.lex_state = 122, .external_lex_state = 2},
  [4395] = {.lex_state = 119, .external_lex_state = 3},
  [4396] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4397] = {.lex_state = 122, .external_lex_state = 3},
  [4398] = {.lex_state = 122, .external_lex_state = 2},
  [4399] = {.lex_state = 122, .external_lex_state = 3},
  [4400] = {.lex_state = 122, .external_lex_state = 3},
  [4401] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4402] = {.lex_state = 131, .external_lex_state = 3},
  [4403] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4404] = {.lex_state = 122, .external_lex_state = 3},
  [4405] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4406] = {.lex_state = 119, .external_lex_state = 3},
  [4407] = {.lex_state = 64, .external_lex_state = 3},
  [4408] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4409] = {.lex_state = 119, .external_lex_state = 3},
  [4410] = {.lex_state = 122, .external_lex_state = 3},
  [4411] = {.lex_state = 119, .external_lex_state = 3},
  [4412] = {.lex_state = 122, .external_lex_state = 3},
  [4413] = {.lex_state = 131, .external_lex_state = 3},
  [4414] = {.lex_state = 64, .external_lex_state = 3},
  [4415] = {.lex_state = 122, .external_lex_state = 3},
  [4416] = {.lex_state = 122, .external_lex_state = 3},
  [4417] = {.lex_state = 64, .external_lex_state = 3},
  [4418] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4419] = {.lex_state = 64, .external_lex_state = 3},
  [4420] = {.lex_state = 64, .external_lex_state = 3},
  [4421] = {.lex_state = 64, .external_lex_state = 3},
  [4422] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4423] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4424] = {.lex_state = 64, .external_lex_state = 3},
  [4425] = {.lex_state = 64, .external_lex_state = 3},
  [4426] = {.lex_state = 64, .external_lex_state = 3},
  [4427] = {.lex_state = 64, .external_lex_state = 3},
  [4428] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4429] = {.lex_state = 64, .external_lex_state = 3},
  [4430] = {.lex_state = 64, .external_lex_state = 3},
  [4431] = {.lex_state = 64, .external_lex_state = 3},
  [4432] = {.lex_state = 64, .external_lex_state = 3},
  [4433] = {.lex_state = 64, .external_lex_state = 3},
  [4434] = {.lex_state = 64, .external_lex_state = 3},
  [4435] = {.lex_state = 64, .external_lex_state = 3},
  [4436] = {.lex_state = 64, .external_lex_state = 3},
  [4437] = {.lex_state = 64, .external_lex_state = 3},
  [4438] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4439] = {.lex_state = 130, .external_lex_state = 3},
  [4440] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4441] = {.lex_state = 122, .external_lex_state = 3},
  [4442] = {.lex_state = 122, .external_lex_state = 3},
  [4443] = {.lex_state = 131, .external_lex_state = 3},
  [4444] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4445] = {.lex_state = 122, .external_lex_state = 3},
  [4446] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4447] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4448] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4449] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4450] = {.lex_state = 119, .external_lex_state = 3},
  [4451] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4452] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4453] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4454] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4455] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4456] = {.lex_state = 122, .external_lex_state = 3},
  [4457] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4458] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4459] = {.lex_state = 122, .external_lex_state = 3},
  [4460] = {.lex_state = 130, .external_lex_state = 3},
  [4461] = {.lex_state = 130, .external_lex_state = 3},
  [4462] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4463] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4464] = {.lex_state = 122, .external_lex_state = 3},
  [4465] = {.lex_state = 119, .external_lex_state = 3},
  [4466] = {.lex_state = 119, .external_lex_state = 3},
  [4467] = {.lex_state = 119, .external_lex_state = 3},
  [4468] = {.lex_state = 54, .external_lex_state = 3},
  [4469] = {.lex_state = 122, .external_lex_state = 3},
  [4470] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4471] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4472] = {.lex_state = 122, .external_lex_state = 3},
  [4473] = {.lex_state = 122, .external_lex_state = 3},
  [4474] = {.lex_state = 122, .external_lex_state = 3},
  [4475] = {.lex_state = 122, .external_lex_state = 3},
  [4476] = {.lex_state = 122, .external_lex_state = 3},
  [4477] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4478] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4479] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4480] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4481] = {.lex_state = 64, .external_lex_state = 3},
  [4482] = {.lex_state = 119, .external_lex_state = 3},
  [4483] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4484] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4485] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4486] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4487] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4488] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4489] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4490] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4491] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 89},
  [4492] = {.lex_state = 64, .external_lex_state = 3},
  [4493] = {.lex_state = 130, .external_lex_state = 3},
  [4494] = {.lex_state = 131, .external_lex_state = 3},
  [4495] = {.lex_state = 131, .external_lex_state = 3},
  [4496] = {.lex_state = 16, .external_lex_state = 3},
  [4497] = {.lex_state = 64, .external_lex_state = 3},
  [4498] = {.lex_state = 16, .external_lex_state = 3},
  [4499] = {.lex_state = 16, .external_lex_state = 3},
  [4500] = {.lex_state = 130, .external_lex_state = 3},
  [4501] = {.lex_state = 131, .external_lex_state = 3},
  [4502] = {.lex_state = 131, .external_lex_state = 3},
  [4503] = {.lex_state = 131, .external_lex_state = 3},
  [4504] = {.lex_state = 7, .external_lex_state = 3},
  [4505] = {.lex_state = 130, .external_lex_state = 3},
  [4506] = {.lex_state = 130, .external_lex_state = 3},
  [4507] = {.lex_state = 16, .external_lex_state = 3},
  [4508] = {.lex_state = 131, .external_lex_state = 3},
  [4509] = {.lex_state = 64, .external_lex_state = 3},
  [4510] = {.lex_state = 131, .external_lex_state = 3},
  [4511] = {.lex_state = 16, .external_lex_state = 3},
  [4512] = {.lex_state = 131, .external_lex_state = 3},
  [4513] = {.lex_state = 130, .external_lex_state = 3},
  [4514] = {.lex_state = 16, .external_lex_state = 3},
  [4515] = {.lex_state = 130, .external_lex_state = 3},
  [4516] = {.lex_state = 122, .external_lex_state = 3},
  [4517] = {.lex_state = 130, .external_lex_state = 3},
  [4518] = {.lex_state = 131, .external_lex_state = 3},
  [4519] = {.lex_state = 16, .external_lex_state = 3},
  [4520] = {.lex_state = 28, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4521] = {.lex_state = 16, .external_lex_state = 3},
  [4522] = {.lex_state = 131, .external_lex_state = 3},
  [4523] = {.lex_state = 130, .external_lex_state = 3},
  [4524] = {.lex_state = 131, .external_lex_state = 3},
  [4525] = {.lex_state = 131, .external_lex_state = 3},
  [4526] = {.lex_state = 16, .external_lex_state = 3},
  [4527] = {.lex_state = 131, .external_lex_state = 3},
  [4528] = {.lex_state = 130, .external_lex_state = 3},
  [4529] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 90},
  [4530] = {.lex_state = 130, .external_lex_state = 3},
  [4531] = {.lex_state = 16, .external_lex_state = 3},
  [4532] = {.lex_state = 16, .external_lex_state = 3},
  [4533] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 90},
  [4534] = {.lex_state = 64, .external_lex_state = 3},
  [4535] = {.lex_state = 130, .external_lex_state = 3},
  [4536] = {.lex_state = 64, .external_lex_state = 3},
  [4537] = {.lex_state = 16, .external_lex_state = 3},
  [4538] = {.lex_state = 64, .external_lex_state = 3},
  [4539] = {.lex_state = 131, .external_lex_state = 3},
  [4540] = {.lex_state = 131, .external_lex_state = 3},
  [4541] = {.lex_state = 122, .external_lex_state = 3},
  [4542] = {.lex_state = 122, .external_lex_state = 3},
  [4543] = {.lex_state = 64, .external_lex_state = 3},
  [4544] = {.lex_state = 131, .external_lex_state = 3},
  [4545] = {.lex_state = 64, .external_lex_state = 3},
  [4546] = {.lex_state = 130, .external_lex_state = 3},
  [4547] = {.lex_state = 16, .external_lex_state = 3},
  [4548] = {.lex_state = 7, .external_lex_state = 2},
  [4549] = {.lex_state = 131, .external_lex_state = 3},
  [4550] = {.lex_state = 130, .external_lex_state = 3},
  [4551] = {.lex_state = 131, .external_lex_state = 3},
  [4552] = {.lex_state = 7, .external_lex_state = 3},
  [4553] = {.lex_state = 16, .external_lex_state = 3},
  [4554] = {.lex_state = 131, .external_lex_state = 3},
  [4555] = {.lex_state = 130, .external_lex_state = 3},
  [4556] = {.lex_state = 16, .external_lex_state = 3},
  [4557] = {.lex_state = 7, .external_lex_state = 3},
  [4558] = {.lex_state = 16, .external_lex_state = 3},
  [4559] = {.lex_state = 16, .external_lex_state = 3},
  [4560] = {.lex_state = 16, .external_lex_state = 3},
  [4561] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 92},
  [4562] = {.lex_state = 7, .external_lex_state = 3},
  [4563] = {.lex_state = 16, .external_lex_state = 3},
  [4564] = {.lex_state = 7, .external_lex_state = 3},
  [4565] = {.lex_state = 16, .external_lex_state = 3},
  [4566] = {.lex_state = 16, .external_lex_state = 3},
  [4567] = {.lex_state = 16, .external_lex_state = 3},
  [4568] = {.lex_state = 16, .external_lex_state = 3},
  [4569] = {.lex_state = 16, .external_lex_state = 3},
  [4570] = {.lex_state = 16, .external_lex_state = 3},
  [4571] = {.lex_state = 16, .external_lex_state = 3},
  [4572] = {.lex_state = 16, .external_lex_state = 3},
  [4573] = {.lex_state = 16, .external_lex_state = 3},
  [4574] = {.lex_state = 16, .external_lex_state = 3},
  [4575] = {.lex_state = 7, .external_lex_state = 3},
  [4576] = {.lex_state = 16, .external_lex_state = 3},
  [4577] = {.lex_state = 16, .external_lex_state = 3},
  [4578] = {.lex_state = 7, .external_lex_state = 3},
  [4579] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 92},
  [4580] = {.lex_state = 16, .external_lex_state = 3},
  [4581] = {.lex_state = 16, .external_lex_state = 3},
  [4582] = {.lex_state = 16, .external_lex_state = 3},
  [4583] = {.lex_state = 16, .external_lex_state = 3},
  [4584] = {.lex_state = 7, .external_lex_state = 3},
  [4585] = {.lex_state = 16, .external_lex_state = 3},
  [4586] = {.lex_state = 16, .external_lex_state = 3},
  [4587] = {.lex_state = 16, .external_lex_state = 3},
  [4588] = {.lex_state = 16, .external_lex_state = 3},
  [4589] = {.lex_state = 16, .external_lex_state = 3},
  [4590] = {.lex_state = 16, .external_lex_state = 3},
  [4591] = {.lex_state = 7, .external_lex_state = 3},
  [4592] = {.lex_state = 16, .external_lex_state = 3},
  [4593] = {.lex_state = 16, .external_lex_state = 3},
  [4594] = {.lex_state = 16, .external_lex_state = 3},
  [4595] = {.lex_state = 16, .external_lex_state = 3},
  [4596] = {.lex_state = 7, .external_lex_state = 3},
  [4597] = {.lex_state = 16, .external_lex_state = 3},
  [4598] = {.lex_state = 16, .external_lex_state = 3},
  [4599] = {.lex_state = 16, .external_lex_state = 3},
  [4600] = {.lex_state = 16, .external_lex_state = 3},
  [4601] = {.lex_state = 16, .external_lex_state = 3},
  [4602] = {.lex_state = 16, .external_lex_state = 3},
  [4603] = {.lex_state = 16, .external_lex_state = 3},
  [4604] = {.lex_state = 7, .external_lex_state = 3},
  [4605] = {.lex_state = 7, .external_lex_state = 3},
  [4606] = {.lex_state = 7, .external_lex_state = 3},
  [4607] = {.lex_state = 16, .external_lex_state = 3},
  [4608] = {.lex_state = 16, .external_lex_state = 3},
  [4609] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4610] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 92},
  [4611] = {.lex_state = 16, .external_lex_state = 3},
  [4612] = {.lex_state = 7, .external_lex_state = 3},
  [4613] = {.lex_state = 16, .external_lex_state = 3},
  [4614] = {.lex_state = 16, .external_lex_state = 3},
  [4615] = {.lex_state = 7, .external_lex_state = 3},
  [4616] = {.lex_state = 130, .external_lex_state = 3},
  [4617] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 92},
  [4618] = {.lex_state = 16, .external_lex_state = 3},
  [4619] = {.lex_state = 7, .external_lex_state = 3},
  [4620] = {.lex_state = 16, .external_lex_state = 3},
  [4621] = {.lex_state = 131, .external_lex_state = 3},
  [4622] = {.lex_state = 16, .external_lex_state = 3},
  [4623] = {.lex_state = 16, .external_lex_state = 3},
  [4624] = {.lex_state = 16, .external_lex_state = 3},
  [4625] = {.lex_state = 131, .external_lex_state = 3},
  [4626] = {.lex_state = 131, .external_lex_state = 3},
  [4627] = {.lex_state = 16, .external_lex_state = 3},
  [4628] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 91},
  [4629] = {.lex_state = 130, .external_lex_state = 3},
  [4630] = {.lex_state = 131, .external_lex_state = 3},
  [4631] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 92},
  [4632] = {.lex_state = 16, .external_lex_state = 3},
  [4633] = {.lex_state = 131, .external_lex_state = 3},
  [4634] = {.lex_state = 130, .external_lex_state = 3},
  [4635] = {.lex_state = 130, .external_lex_state = 3},
  [4636] = {.lex_state = 130, .external_lex_state = 3},
  [4637] = {.lex_state = 16, .external_lex_state = 3},
  [4638] = {.lex_state = 130, .external_lex_state = 3},
  [4639] = {.lex_state = 7, .external_lex_state = 3},
  [4640] = {.lex_state = 16, .external_lex_state = 3},
  [4641] = {.lex_state = 16, .external_lex_state = 3},
  [4642] = {.lex_state = 16, .external_lex_state = 3},
  [4643] = {.lex_state = 16, .external_lex_state = 3},
  [4644] = {.lex_state = 16, .external_lex_state = 3},
  [4645] = {.lex_state = 16, .external_lex_state = 3},
  [4646] = {.lex_state = 64, .external_lex_state = 3},
  [4647] = {.lex_state = 64, .external_lex_state = 3},
  [4648] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 92},
  [4649] = {.lex_state = 131, .external_lex_state = 3},
  [4650] = {.lex_state = 131, .external_lex_state = 3},
  [4651] = {.lex_state = 16, .external_lex_state = 3},
  [4652] = {.lex_state = 64, .external_lex_state = 3},
  [4653] = {.lex_state = 7, .external_lex_state = 3},
  [4654] = {.lex_state = 16, .external_lex_state = 3},
  [4655] = {.lex_state = 130, .external_lex_state = 3},
  [4656] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 92},
  [4657] = {.lex_state = 64, .external_lex_state = 3},
  [4658] = {.lex_state = 16, .external_lex_state = 3},
  [4659] = {.lex_state = 64, .external_lex_state = 3},
  [4660] = {.lex_state = 64, .external_lex_state = 3},
  [4661] = {.lex_state = 16, .external_lex_state = 3},
  [4662] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 92},
  [4663] = {.lex_state = 64, .external_lex_state = 3},
  [4664] = {.lex_state = 16, .external_lex_state = 3},
  [4665] = {.lex_state = 7, .external_lex_state = 3},
  [4666] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 91},
  [4667] = {.lex_state = 16, .external_lex_state = 3},
  [4668] = {.lex_state = 64, .external_lex_state = 3},
  [4669] = {.lex_state = 64, .external_lex_state = 3},
  [4670] = {.lex_state = 64, .external_lex_state = 3},
  [4671] = {.lex_state = 16, .external_lex_state = 3},
  [4672] = {.lex_state = 64, .external_lex_state = 3},
  [4673] = {.lex_state = 64, .external_lex_state = 3},
  [4674] = {.lex_state = 16, .external_lex_state = 3},
  [4675] = {.lex_state = 16, .external_lex_state = 3},
  [4676] = {.lex_state = 64, .external_lex_state = 3},
  [4677] = {.lex_state = 16, .external_lex_state = 3},
  [4678] = {.lex_state = 16, .external_lex_state = 3},
  [4679] = {.lex_state = 64, .external_lex_state = 3},
  [4680] = {.lex_state = 16, .external_lex_state = 3},
  [4681] = {.lex_state = 64, .external_lex_state = 3},
  [4682] = {.lex_state = 16, .external_lex_state = 3},
  [4683] = {.lex_state = 16, .external_lex_state = 3},
  [4684] = {.lex_state = 16, .external_lex_state = 3},
  [4685] = {.lex_state = 16, .external_lex_state = 3},
  [4686] = {.lex_state = 16, .external_lex_state = 3},
  [4687] = {.lex_state = 16, .external_lex_state = 3},
  [4688] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 93},
  [4689] = {.lex_state = 7, .external_lex_state = 3},
  [4690] = {.lex_state = 16, .external_lex_state = 3},
  [4691] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 92},
  [4692] = {.lex_state = 16, .external_lex_state = 3},
  [4693] = {.lex_state = 16, .external_lex_state = 3},
  [4694] = {.lex_state = 16, .external_lex_state = 3},
  [4695] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 92},
  [4696] = {.lex_state = 16, .external_lex_state = 3},
  [4697] = {.lex_state = 16, .external_lex_state = 3},
  [4698] = {.lex_state = 16, .external_lex_state = 3},
  [4699] = {.lex_state = 16, .external_lex_state = 3},
  [4700] = {.lex_state = 122, .external_lex_state = 3},
  [4701] = {.lex_state = 16, .external_lex_state = 3},
  [4702] = {.lex_state = 16, .external_lex_state = 3},
  [4703] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 92},
  [4704] = {.lex_state = 16, .external_lex_state = 3},
  [4705] = {.lex_state = 16, .external_lex_state = 3},
  [4706] = {.lex_state = 16, .external_lex_state = 3},
  [4707] = {.lex_state = 16, .external_lex_state = 3},
  [4708] = {.lex_state = 16, .external_lex_state = 3},
  [4709] = {.lex_state = 16, .external_lex_state = 3},
  [4710] = {.lex_state = 16, .external_lex_state = 3},
  [4711] = {.lex_state = 16, .external_lex_state = 3},
  [4712] = {.lex_state = 16, .external_lex_state = 3},
  [4713] = {.lex_state = 16, .external_lex_state = 3},
  [4714] = {.lex_state = 16, .external_lex_state = 3},
  [4715] = {.lex_state = 16, .external_lex_state = 3},
  [4716] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 93},
  [4717] = {.lex_state = 16, .external_lex_state = 3},
  [4718] = {.lex_state = 16, .external_lex_state = 3},
  [4719] = {.lex_state = 131, .external_lex_state = 3},
  [4720] = {.lex_state = 16, .external_lex_state = 3},
  [4721] = {.lex_state = 16, .external_lex_state = 3},
  [4722] = {.lex_state = 16, .external_lex_state = 3},
  [4723] = {.lex_state = 16, .external_lex_state = 3},
  [4724] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 92},
  [4725] = {.lex_state = 16, .external_lex_state = 3},
  [4726] = {.lex_state = 16, .external_lex_state = 3},
  [4727] = {.lex_state = 16, .external_lex_state = 3},
  [4728] = {.lex_state = 16, .external_lex_state = 3},
  [4729] = {.lex_state = 16, .external_lex_state = 3},
  [4730] = {.lex_state = 16, .external_lex_state = 3},
  [4731] = {.lex_state = 16, .external_lex_state = 3},
  [4732] = {.lex_state = 16, .external_lex_state = 3},
  [4733] = {.lex_state = 16, .external_lex_state = 3},
  [4734] = {.lex_state = 16, .external_lex_state = 3},
  [4735] = {.lex_state = 16, .external_lex_state = 3},
  [4736] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 92},
  [4737] = {.lex_state = 16, .external_lex_state = 3},
  [4738] = {.lex_state = 16, .external_lex_state = 3},
  [4739] = {.lex_state = 16, .external_lex_state = 3},
  [4740] = {.lex_state = 16, .external_lex_state = 3},
  [4741] = {.lex_state = 16, .external_lex_state = 3},
  [4742] = {.lex_state = 16, .external_lex_state = 3},
  [4743] = {.lex_state = 131, .external_lex_state = 3},
  [4744] = {.lex_state = 16, .external_lex_state = 3},
  [4745] = {.lex_state = 16, .external_lex_state = 3},
  [4746] = {.lex_state = 16, .external_lex_state = 3},
  [4747] = {.lex_state = 16, .external_lex_state = 3},
  [4748] = {.lex_state = 7, .external_lex_state = 3},
  [4749] = {.lex_state = 16, .external_lex_state = 3},
  [4750] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 92},
  [4751] = {.lex_state = 16, .external_lex_state = 3},
  [4752] = {.lex_state = 16, .external_lex_state = 3},
  [4753] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 90},
  [4754] = {.lex_state = 16, .external_lex_state = 3},
  [4755] = {.lex_state = 7, .external_lex_state = 3},
  [4756] = {.lex_state = 16, .external_lex_state = 3},
  [4757] = {.lex_state = 16, .external_lex_state = 3},
  [4758] = {.lex_state = 16, .external_lex_state = 3},
  [4759] = {.lex_state = 16, .external_lex_state = 3},
  [4760] = {.lex_state = 16, .external_lex_state = 3},
  [4761] = {.lex_state = 16, .external_lex_state = 3},
  [4762] = {.lex_state = 16, .external_lex_state = 3},
  [4763] = {.lex_state = 16, .external_lex_state = 3},
  [4764] = {.lex_state = 7, .external_lex_state = 3},
  [4765] = {.lex_state = 16, .external_lex_state = 3},
  [4766] = {.lex_state = 7, .external_lex_state = 3},
  [4767] = {.lex_state = 16, .external_lex_state = 3},
  [4768] = {.lex_state = 16, .external_lex_state = 3},
  [4769] = {.lex_state = 16, .external_lex_state = 3},
  [4770] = {.lex_state = 16, .external_lex_state = 3},
  [4771] = {.lex_state = 16, .external_lex_state = 3},
  [4772] = {.lex_state = 16, .external_lex_state = 3},
  [4773] = {.lex_state = 122, .external_lex_state = 3},
  [4774] = {.lex_state = 122, .external_lex_state = 3},
  [4775] = {.lex_state = 122, .external_lex_state = 3},
  [4776] = {.lex_state = 16, .external_lex_state = 3},
  [4777] = {.lex_state = 131, .external_lex_state = 3},
  [4778] = {.lex_state = 16, .external_lex_state = 3},
  [4779] = {.lex_state = 122, .external_lex_state = 3},
  [4780] = {.lex_state = 122, .external_lex_state = 3},
  [4781] = {.lex_state = 122, .external_lex_state = 3},
  [4782] = {.lex_state = 122, .external_lex_state = 3},
  [4783] = {.lex_state = 122, .external_lex_state = 3},
  [4784] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4785] = {.lex_state = 122, .external_lex_state = 3},
  [4786] = {.lex_state = 122, .external_lex_state = 3},
  [4787] = {.lex_state = 16, .external_lex_state = 3},
  [4788] = {.lex_state = 7, .external_lex_state = 3},
  [4789] = {.lex_state = 122, .external_lex_state = 3},
  [4790] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4791] = {.lex_state = 7, .external_lex_state = 3},
  [4792] = {.lex_state = 16, .external_lex_state = 3},
  [4793] = {.lex_state = 7, .external_lex_state = 3},
  [4794] = {.lex_state = 122, .external_lex_state = 3},
  [4795] = {.lex_state = 131, .external_lex_state = 3},
  [4796] = {.lex_state = 122, .external_lex_state = 3},
  [4797] = {.lex_state = 131, .external_lex_state = 3},
  [4798] = {.lex_state = 122, .external_lex_state = 3},
  [4799] = {.lex_state = 122, .external_lex_state = 3},
  [4800] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4801] = {.lex_state = 122, .external_lex_state = 3},
  [4802] = {.lex_state = 122, .external_lex_state = 3},
  [4803] = {.lex_state = 7, .external_lex_state = 3},
  [4804] = {.lex_state = 131, .external_lex_state = 3},
  [4805] = {.lex_state = 131, .external_lex_state = 3},
  [4806] = {.lex_state = 122, .external_lex_state = 3},
  [4807] = {.lex_state = 16, .external_lex_state = 3},
  [4808] = {.lex_state = 16, .external_lex_state = 3},
  [4809] = {.lex_state = 16, .external_lex_state = 3},
  [4810] = {.lex_state = 16, .external_lex_state = 3},
  [4811] = {.lex_state = 16, .external_lex_state = 3},
  [4812] = {.lex_state = 53, .external_lex_state = 3},
  [4813] = {.lex_state = 53, .external_lex_state = 3},
  [4814] = {.lex_state = 53, .external_lex_state = 3},
  [4815] = {.lex_state = 16, .external_lex_state = 3},
  [4816] = {.lex_state = 16, .external_lex_state = 3},
  [4817] = {.lex_state = 16, .external_lex_state = 3},
  [4818] = {.lex_state = 16, .external_lex_state = 3},
  [4819] = {.lex_state = 53, .external_lex_state = 3},
  [4820] = {.lex_state = 16, .external_lex_state = 3},
  [4821] = {.lex_state = 16, .external_lex_state = 3},
  [4822] = {.lex_state = 16, .external_lex_state = 3},
  [4823] = {.lex_state = 122, .external_lex_state = 3},
  [4824] = {.lex_state = 16, .external_lex_state = 3},
  [4825] = {.lex_state = 16, .external_lex_state = 3},
  [4826] = {.lex_state = 16, .external_lex_state = 3},
  [4827] = {.lex_state = 16, .external_lex_state = 3},
  [4828] = {.lex_state = 16, .external_lex_state = 3},
  [4829] = {.lex_state = 16, .external_lex_state = 3},
  [4830] = {.lex_state = 16, .external_lex_state = 3},
  [4831] = {.lex_state = 16, .external_lex_state = 3},
  [4832] = {.lex_state = 16, .external_lex_state = 3},
  [4833] = {.lex_state = 16, .external_lex_state = 3},
  [4834] = {.lex_state = 16, .external_lex_state = 3},
  [4835] = {.lex_state = 16, .external_lex_state = 3},
  [4836] = {.lex_state = 16, .external_lex_state = 3},
  [4837] = {.lex_state = 16, .external_lex_state = 3},
  [4838] = {.lex_state = 16, .external_lex_state = 3},
  [4839] = {.lex_state = 7, .external_lex_state = 3},
  [4840] = {.lex_state = 7, .external_lex_state = 3},
  [4841] = {.lex_state = 16, .external_lex_state = 3},
  [4842] = {.lex_state = 16, .external_lex_state = 3},
  [4843] = {.lex_state = 16, .external_lex_state = 3},
  [4844] = {.lex_state = 16, .external_lex_state = 3},
  [4845] = {.lex_state = 16, .external_lex_state = 3},
  [4846] = {.lex_state = 16, .external_lex_state = 3},
  [4847] = {.lex_state = 16, .external_lex_state = 3},
  [4848] = {.lex_state = 16, .external_lex_state = 3},
  [4849] = {.lex_state = 16, .external_lex_state = 3},
  [4850] = {.lex_state = 16, .external_lex_state = 3},
  [4851] = {.lex_state = 16, .external_lex_state = 3},
  [4852] = {.lex_state = 16, .external_lex_state = 3},
  [4853] = {.lex_state = 16, .external_lex_state = 3},
  [4854] = {.lex_state = 16, .external_lex_state = 3},
  [4855] = {.lex_state = 16, .external_lex_state = 3},
  [4856] = {.lex_state = 16, .external_lex_state = 3},
  [4857] = {.lex_state = 16, .external_lex_state = 3},
  [4858] = {.lex_state = 16, .external_lex_state = 3},
  [4859] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4860] = {.lex_state = 49, .external_lex_state = 5},
  [4861] = {.lex_state = 49, .external_lex_state = 5},
  [4862] = {.lex_state = 64, .external_lex_state = 3},
  [4863] = {.lex_state = 51, .external_lex_state = 3},
  [4864] = {.lex_state = 49, .external_lex_state = 5},
  [4865] = {.lex_state = 64, .external_lex_state = 3},
  [4866] = {.lex_state = 64, .external_lex_state = 3},
  [4867] = {.lex_state = 64, .external_lex_state = 3},
  [4868] = {.lex_state = 64, .external_lex_state = 3},
  [4869] = {.lex_state = 49, .external_lex_state = 5},
  [4870] = {.lex_state = 64, .external_lex_state = 3},
  [4871] = {.lex_state = 64, .external_lex_state = 3},
  [4872] = {.lex_state = 64, .external_lex_state = 3},
  [4873] = {.lex_state = 49, .external_lex_state = 5},
  [4874] = {.lex_state = 64, .external_lex_state = 3},
  [4875] = {.lex_state = 49, .external_lex_state = 5},
  [4876] = {.lex_state = 49, .external_lex_state = 5},
  [4877] = {.lex_state = 64, .external_lex_state = 3},
  [4878] = {.lex_state = 7, .external_lex_state = 3},
  [4879] = {.lex_state = 7, .external_lex_state = 3},
  [4880] = {.lex_state = 49, .external_lex_state = 5},
  [4881] = {.lex_state = 49, .external_lex_state = 5},
  [4882] = {.lex_state = 49, .external_lex_state = 5},
  [4883] = {.lex_state = 49, .external_lex_state = 5},
  [4884] = {.lex_state = 49, .external_lex_state = 5},
  [4885] = {.lex_state = 64, .external_lex_state = 3},
  [4886] = {.lex_state = 49, .external_lex_state = 5},
  [4887] = {.lex_state = 51, .external_lex_state = 3},
  [4888] = {.lex_state = 64, .external_lex_state = 3},
  [4889] = {.lex_state = 64, .external_lex_state = 3},
  [4890] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4891] = {.lex_state = 49, .external_lex_state = 5},
  [4892] = {.lex_state = 49, .external_lex_state = 5},
  [4893] = {.lex_state = 64, .external_lex_state = 3},
  [4894] = {.lex_state = 49, .external_lex_state = 5},
  [4895] = {.lex_state = 49, .external_lex_state = 5},
  [4896] = {.lex_state = 49, .external_lex_state = 5},
  [4897] = {.lex_state = 49, .external_lex_state = 5},
  [4898] = {.lex_state = 7, .external_lex_state = 3},
  [4899] = {.lex_state = 49, .external_lex_state = 5},
  [4900] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4901] = {.lex_state = 64, .external_lex_state = 3},
  [4902] = {.lex_state = 64, .external_lex_state = 3},
  [4903] = {.lex_state = 64, .external_lex_state = 3},
  [4904] = {.lex_state = 64, .external_lex_state = 3},
  [4905] = {.lex_state = 64, .external_lex_state = 3},
  [4906] = {.lex_state = 64, .external_lex_state = 3},
  [4907] = {.lex_state = 64, .external_lex_state = 3},
  [4908] = {.lex_state = 49, .external_lex_state = 5},
  [4909] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4910] = {.lex_state = 49, .external_lex_state = 5},
  [4911] = {.lex_state = 51, .external_lex_state = 3},
  [4912] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 69},
  [4913] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4914] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 69},
  [4915] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 69},
  [4916] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 69},
  [4917] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4918] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4919] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4920] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4921] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4922] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4923] = {.lex_state = 7, .external_lex_state = 2},
  [4924] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4925] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 69},
  [4926] = {.lex_state = 7, .external_lex_state = 2},
  [4927] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4928] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4929] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4930] = {.lex_state = 7, .external_lex_state = 3},
  [4931] = {.lex_state = 7, .external_lex_state = 3},
  [4932] = {.lex_state = 7, .external_lex_state = 2},
  [4933] = {.lex_state = 49, .external_lex_state = 5},
  [4934] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4935] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 69},
  [4936] = {.lex_state = 7, .external_lex_state = 2},
  [4937] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4938] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 69},
  [4939] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4940] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4941] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4942] = {.lex_state = 7, .external_lex_state = 3},
  [4943] = {.lex_state = 64, .external_lex_state = 3},
  [4944] = {.lex_state = 7, .external_lex_state = 3},
  [4945] = {.lex_state = 7, .external_lex_state = 3},
  [4946] = {.lex_state = 7, .external_lex_state = 3},
  [4947] = {.lex_state = 7, .external_lex_state = 3},
  [4948] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4949] = {.lex_state = 64, .external_lex_state = 3},
  [4950] = {.lex_state = 7, .external_lex_state = 3},
  [4951] = {.lex_state = 7, .external_lex_state = 3},
  [4952] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4953] = {.lex_state = 51, .external_lex_state = 3},
  [4954] = {.lex_state = 64, .external_lex_state = 3},
  [4955] = {.lex_state = 64, .external_lex_state = 3},
  [4956] = {.lex_state = 64, .external_lex_state = 3},
  [4957] = {.lex_state = 52, .external_lex_state = 3},
  [4958] = {.lex_state = 7, .external_lex_state = 3},
  [4959] = {.lex_state = 7, .external_lex_state = 3},
  [4960] = {.lex_state = 51, .external_lex_state = 3},
  [4961] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4962] = {.lex_state = 7, .external_lex_state = 3},
  [4963] = {.lex_state = 53, .external_lex_state = 3},
  [4964] = {.lex_state = 7, .external_lex_state = 3},
  [4965] = {.lex_state = 7, .external_lex_state = 3},
  [4966] = {.lex_state = 7, .external_lex_state = 3},
  [4967] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [4968] = {.lex_state = 64, .external_lex_state = 3},
  [4969] = {.lex_state = 7, .external_lex_state = 3},
  [4970] = {.lex_state = 7, .external_lex_state = 3},
  [4971] = {.lex_state = 64, .external_lex_state = 3},
  [4972] = {.lex_state = 7, .external_lex_state = 3},
  [4973] = {.lex_state = 52, .external_lex_state = 3},
  [4974] = {.lex_state = 16, .external_lex_state = 3},
  [4975] = {.lex_state = 16, .external_lex_state = 3},
  [4976] = {.lex_state = 7, .external_lex_state = 3},
  [4977] = {.lex_state = 16, .external_lex_state = 3},
  [4978] = {.lex_state = 7, .external_lex_state = 3},
  [4979] = {.lex_state = 16, .external_lex_state = 3},
  [4980] = {.lex_state = 7, .external_lex_state = 3},
  [4981] = {.lex_state = 7, .external_lex_state = 3},
  [4982] = {.lex_state = 16, .external_lex_state = 3},
  [4983] = {.lex_state = 52, .external_lex_state = 3},
  [4984] = {.lex_state = 7, .external_lex_state = 3},
  [4985] = {.lex_state = 7, .external_lex_state = 3},
  [4986] = {.lex_state = 52, .external_lex_state = 3},
  [4987] = {.lex_state = 16, .external_lex_state = 3},
  [4988] = {.lex_state = 16, .external_lex_state = 3},
  [4989] = {.lex_state = 7, .external_lex_state = 3},
  [4990] = {.lex_state = 16, .external_lex_state = 3},
  [4991] = {.lex_state = 16, .external_lex_state = 3},
  [4992] = {.lex_state = 16, .external_lex_state = 3},
  [4993] = {.lex_state = 16, .external_lex_state = 3},
  [4994] = {.lex_state = 16, .external_lex_state = 3},
  [4995] = {.lex_state = 16, .external_lex_state = 3},
  [4996] = {.lex_state = 52, .external_lex_state = 3},
  [4997] = {.lex_state = 51, .external_lex_state = 3},
  [4998] = {.lex_state = 7, .external_lex_state = 3},
  [4999] = {.lex_state = 16, .external_lex_state = 3},
  [5000] = {.lex_state = 16, .external_lex_state = 3},
  [5001] = {.lex_state = 49, .external_lex_state = 5},
  [5002] = {.lex_state = 52, .external_lex_state = 3},
  [5003] = {.lex_state = 16, .external_lex_state = 3},
  [5004] = {.lex_state = 7, .external_lex_state = 3},
  [5005] = {.lex_state = 66, .external_lex_state = 3},
  [5006] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 94},
  [5007] = {.lex_state = 7, .external_lex_state = 3},
  [5008] = {.lex_state = 16, .external_lex_state = 3},
  [5009] = {.lex_state = 66, .external_lex_state = 3},
  [5010] = {.lex_state = 52, .external_lex_state = 3},
  [5011] = {.lex_state = 7, .external_lex_state = 3},
  [5012] = {.lex_state = 16, .external_lex_state = 3},
  [5013] = {.lex_state = 7, .external_lex_state = 3},
  [5014] = {.lex_state = 16, .external_lex_state = 3},
  [5015] = {.lex_state = 52, .external_lex_state = 3},
  [5016] = {.lex_state = 7, .external_lex_state = 3},
  [5017] = {.lex_state = 7, .external_lex_state = 3},
  [5018] = {.lex_state = 52, .external_lex_state = 3},
  [5019] = {.lex_state = 7, .external_lex_state = 3},
  [5020] = {.lex_state = 16, .external_lex_state = 3},
  [5021] = {.lex_state = 16, .external_lex_state = 3},
  [5022] = {.lex_state = 16, .external_lex_state = 3},
  [5023] = {.lex_state = 7, .external_lex_state = 3},
  [5024] = {.lex_state = 7, .external_lex_state = 3},
  [5025] = {.lex_state = 66, .external_lex_state = 3},
  [5026] = {.lex_state = 64, .external_lex_state = 3},
  [5027] = {.lex_state = 16, .external_lex_state = 3},
  [5028] = {.lex_state = 7, .external_lex_state = 3},
  [5029] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 94},
  [5030] = {.lex_state = 7, .external_lex_state = 3},
  [5031] = {.lex_state = 7, .external_lex_state = 3},
  [5032] = {.lex_state = 49, .external_lex_state = 5},
  [5033] = {.lex_state = 51, .external_lex_state = 3},
  [5034] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 94},
  [5035] = {.lex_state = 51, .external_lex_state = 3},
  [5036] = {.lex_state = 51, .external_lex_state = 3},
  [5037] = {.lex_state = 16, .external_lex_state = 3},
  [5038] = {.lex_state = 120, .external_lex_state = 3},
  [5039] = {.lex_state = 16, .external_lex_state = 3},
  [5040] = {.lex_state = 16, .external_lex_state = 3},
  [5041] = {.lex_state = 120, .external_lex_state = 3},
  [5042] = {.lex_state = 16, .external_lex_state = 3},
  [5043] = {.lex_state = 51, .external_lex_state = 3},
  [5044] = {.lex_state = 16, .external_lex_state = 3},
  [5045] = {.lex_state = 16, .external_lex_state = 3},
  [5046] = {.lex_state = 16, .external_lex_state = 3},
  [5047] = {.lex_state = 16, .external_lex_state = 3},
  [5048] = {.lex_state = 51, .external_lex_state = 3},
  [5049] = {.lex_state = 16, .external_lex_state = 3},
  [5050] = {.lex_state = 16, .external_lex_state = 3},
  [5051] = {.lex_state = 16, .external_lex_state = 3},
  [5052] = {.lex_state = 16, .external_lex_state = 3},
  [5053] = {.lex_state = 16, .external_lex_state = 3},
  [5054] = {.lex_state = 16, .external_lex_state = 3},
  [5055] = {.lex_state = 51, .external_lex_state = 3},
  [5056] = {.lex_state = 16, .external_lex_state = 3},
  [5057] = {.lex_state = 16, .external_lex_state = 3},
  [5058] = {.lex_state = 16, .external_lex_state = 3},
  [5059] = {.lex_state = 16, .external_lex_state = 3},
  [5060] = {.lex_state = 120, .external_lex_state = 3},
  [5061] = {.lex_state = 16, .external_lex_state = 3},
  [5062] = {.lex_state = 16, .external_lex_state = 3},
  [5063] = {.lex_state = 120, .external_lex_state = 3},
  [5064] = {.lex_state = 51, .external_lex_state = 3},
  [5065] = {.lex_state = 120, .external_lex_state = 3},
  [5066] = {.lex_state = 120, .external_lex_state = 3},
  [5067] = {.lex_state = 51, .external_lex_state = 3},
  [5068] = {.lex_state = 51, .external_lex_state = 3},
  [5069] = {.lex_state = 120, .external_lex_state = 3},
  [5070] = {.lex_state = 51, .external_lex_state = 3},
  [5071] = {.lex_state = 51, .external_lex_state = 3},
  [5072] = {.lex_state = 120, .external_lex_state = 3},
  [5073] = {.lex_state = 51, .external_lex_state = 3},
  [5074] = {.lex_state = 51, .external_lex_state = 3},
  [5075] = {.lex_state = 51, .external_lex_state = 3},
  [5076] = {.lex_state = 120, .external_lex_state = 3},
  [5077] = {.lex_state = 120, .external_lex_state = 3},
  [5078] = {.lex_state = 16, .external_lex_state = 3},
  [5079] = {.lex_state = 16, .external_lex_state = 3},
  [5080] = {.lex_state = 16, .external_lex_state = 3},
  [5081] = {.lex_state = 16, .external_lex_state = 3},
  [5082] = {.lex_state = 16, .external_lex_state = 3},
  [5083] = {.lex_state = 16, .external_lex_state = 3},
  [5084] = {.lex_state = 16, .external_lex_state = 3},
  [5085] = {.lex_state = 16, .external_lex_state = 3},
  [5086] = {.lex_state = 16, .external_lex_state = 3},
  [5087] = {.lex_state = 16, .external_lex_state = 3},
  [5088] = {.lex_state = 16, .external_lex_state = 3},
  [5089] = {.lex_state = 120, .external_lex_state = 3},
  [5090] = {.lex_state = 16, .external_lex_state = 3},
  [5091] = {.lex_state = 51, .external_lex_state = 3},
  [5092] = {.lex_state = 120, .external_lex_state = 3},
  [5093] = {.lex_state = 16, .external_lex_state = 3},
  [5094] = {.lex_state = 16, .external_lex_state = 3},
  [5095] = {.lex_state = 16, .external_lex_state = 3},
  [5096] = {.lex_state = 16, .external_lex_state = 3},
  [5097] = {.lex_state = 120, .external_lex_state = 3},
  [5098] = {.lex_state = 51, .external_lex_state = 3},
  [5099] = {.lex_state = 51, .external_lex_state = 3},
  [5100] = {.lex_state = 51, .external_lex_state = 3},
  [5101] = {.lex_state = 120, .external_lex_state = 3},
  [5102] = {.lex_state = 120, .external_lex_state = 3},
  [5103] = {.lex_state = 120, .external_lex_state = 3},
  [5104] = {.lex_state = 51, .external_lex_state = 3},
  [5105] = {.lex_state = 120, .external_lex_state = 3},
  [5106] = {.lex_state = 16, .external_lex_state = 3},
  [5107] = {.lex_state = 16, .external_lex_state = 3},
  [5108] = {.lex_state = 16, .external_lex_state = 3},
  [5109] = {.lex_state = 120, .external_lex_state = 3},
  [5110] = {.lex_state = 16, .external_lex_state = 3},
  [5111] = {.lex_state = 16, .external_lex_state = 3},
  [5112] = {.lex_state = 120, .external_lex_state = 3},
  [5113] = {.lex_state = 16, .external_lex_state = 3},
  [5114] = {.lex_state = 16, .external_lex_state = 3},
  [5115] = {.lex_state = 16, .external_lex_state = 3},
  [5116] = {.lex_state = 120, .external_lex_state = 3},
  [5117] = {.lex_state = 16, .external_lex_state = 3},
  [5118] = {.lex_state = 16, .external_lex_state = 3},
  [5119] = {.lex_state = 51, .external_lex_state = 3},
  [5120] = {.lex_state = 51, .external_lex_state = 3},
  [5121] = {.lex_state = 120, .external_lex_state = 3},
  [5122] = {.lex_state = 120, .external_lex_state = 3},
  [5123] = {.lex_state = 51, .external_lex_state = 3},
  [5124] = {.lex_state = 16, .external_lex_state = 3},
  [5125] = {.lex_state = 16, .external_lex_state = 3},
  [5126] = {.lex_state = 16, .external_lex_state = 3},
  [5127] = {.lex_state = 16, .external_lex_state = 3},
  [5128] = {.lex_state = 16, .external_lex_state = 3},
  [5129] = {.lex_state = 16, .external_lex_state = 3},
  [5130] = {.lex_state = 64, .external_lex_state = 3},
  [5131] = {.lex_state = 120, .external_lex_state = 3},
  [5132] = {.lex_state = 51, .external_lex_state = 3},
  [5133] = {.lex_state = 51, .external_lex_state = 3},
  [5134] = {.lex_state = 52, .external_lex_state = 3},
  [5135] = {.lex_state = 52, .external_lex_state = 3},
  [5136] = {.lex_state = 120, .external_lex_state = 3},
  [5137] = {.lex_state = 64, .external_lex_state = 3},
  [5138] = {.lex_state = 16, .external_lex_state = 3},
  [5139] = {.lex_state = 16, .external_lex_state = 3},
  [5140] = {.lex_state = 16, .external_lex_state = 3},
  [5141] = {.lex_state = 16, .external_lex_state = 3},
  [5142] = {.lex_state = 51, .external_lex_state = 3},
  [5143] = {.lex_state = 51, .external_lex_state = 3},
  [5144] = {.lex_state = 16, .external_lex_state = 3},
  [5145] = {.lex_state = 16, .external_lex_state = 3},
  [5146] = {.lex_state = 16, .external_lex_state = 3},
  [5147] = {.lex_state = 16, .external_lex_state = 3},
  [5148] = {.lex_state = 16, .external_lex_state = 3},
  [5149] = {.lex_state = 120, .external_lex_state = 3},
  [5150] = {.lex_state = 120, .external_lex_state = 3},
  [5151] = {.lex_state = 51, .external_lex_state = 3},
  [5152] = {.lex_state = 51, .external_lex_state = 3},
  [5153] = {.lex_state = 16, .external_lex_state = 3},
  [5154] = {.lex_state = 120, .external_lex_state = 3},
  [5155] = {.lex_state = 51, .external_lex_state = 3},
  [5156] = {.lex_state = 16, .external_lex_state = 3},
  [5157] = {.lex_state = 16, .external_lex_state = 3},
  [5158] = {.lex_state = 16, .external_lex_state = 3},
  [5159] = {.lex_state = 64, .external_lex_state = 3},
  [5160] = {.lex_state = 16, .external_lex_state = 3},
  [5161] = {.lex_state = 16, .external_lex_state = 3},
  [5162] = {.lex_state = 16, .external_lex_state = 3},
  [5163] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 94},
  [5164] = {.lex_state = 16, .external_lex_state = 3},
  [5165] = {.lex_state = 16, .external_lex_state = 3},
  [5166] = {.lex_state = 51, .external_lex_state = 3},
  [5167] = {.lex_state = 51, .external_lex_state = 3},
  [5168] = {.lex_state = 51, .external_lex_state = 3},
  [5169] = {.lex_state = 51, .external_lex_state = 3},
  [5170] = {.lex_state = 51, .external_lex_state = 3},
  [5171] = {.lex_state = 120, .external_lex_state = 3},
  [5172] = {.lex_state = 51, .external_lex_state = 3},
  [5173] = {.lex_state = 51, .external_lex_state = 3},
  [5174] = {.lex_state = 16, .external_lex_state = 3},
  [5175] = {.lex_state = 16, .external_lex_state = 3},
  [5176] = {.lex_state = 16, .external_lex_state = 3},
  [5177] = {.lex_state = 16, .external_lex_state = 3},
  [5178] = {.lex_state = 16, .external_lex_state = 3},
  [5179] = {.lex_state = 16, .external_lex_state = 3},
  [5180] = {.lex_state = 52, .external_lex_state = 3},
  [5181] = {.lex_state = 52, .external_lex_state = 3},
  [5182] = {.lex_state = 64, .external_lex_state = 3},
  [5183] = {.lex_state = 64, .external_lex_state = 3},
  [5184] = {.lex_state = 7, .external_lex_state = 3},
  [5185] = {.lex_state = 7, .external_lex_state = 3},
  [5186] = {.lex_state = 64, .external_lex_state = 3},
  [5187] = {.lex_state = 64, .external_lex_state = 3},
  [5188] = {.lex_state = 16, .external_lex_state = 3},
  [5189] = {.lex_state = 16, .external_lex_state = 3},
  [5190] = {.lex_state = 16, .external_lex_state = 3},
  [5191] = {.lex_state = 16, .external_lex_state = 3},
  [5192] = {.lex_state = 16, .external_lex_state = 3},
  [5193] = {.lex_state = 51, .external_lex_state = 3},
  [5194] = {.lex_state = 7, .external_lex_state = 3},
  [5195] = {.lex_state = 7, .external_lex_state = 3},
  [5196] = {.lex_state = 64, .external_lex_state = 3},
  [5197] = {.lex_state = 68, .external_lex_state = 3},
  [5198] = {.lex_state = 64, .external_lex_state = 3},
  [5199] = {.lex_state = 7, .external_lex_state = 3},
  [5200] = {.lex_state = 64, .external_lex_state = 3},
  [5201] = {.lex_state = 64, .external_lex_state = 3},
  [5202] = {.lex_state = 7, .external_lex_state = 3},
  [5203] = {.lex_state = 64, .external_lex_state = 3},
  [5204] = {.lex_state = 51, .external_lex_state = 3},
  [5205] = {.lex_state = 7, .external_lex_state = 3},
  [5206] = {.lex_state = 7, .external_lex_state = 3},
  [5207] = {.lex_state = 64, .external_lex_state = 3},
  [5208] = {.lex_state = 64, .external_lex_state = 3},
  [5209] = {.lex_state = 64, .external_lex_state = 3},
  [5210] = {.lex_state = 7, .external_lex_state = 3},
  [5211] = {.lex_state = 7, .external_lex_state = 3},
  [5212] = {.lex_state = 64, .external_lex_state = 3},
  [5213] = {.lex_state = 64, .external_lex_state = 3},
  [5214] = {.lex_state = 64, .external_lex_state = 3},
  [5215] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 94},
  [5216] = {.lex_state = 7, .external_lex_state = 3},
  [5217] = {.lex_state = 7, .external_lex_state = 3},
  [5218] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 94},
  [5219] = {.lex_state = 64, .external_lex_state = 3},
  [5220] = {.lex_state = 64, .external_lex_state = 3},
  [5221] = {.lex_state = 64, .external_lex_state = 3},
  [5222] = {.lex_state = 64, .external_lex_state = 3},
  [5223] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5224] = {.lex_state = 64, .external_lex_state = 3},
  [5225] = {.lex_state = 7, .external_lex_state = 3},
  [5226] = {.lex_state = 7, .external_lex_state = 3},
  [5227] = {.lex_state = 64, .external_lex_state = 3},
  [5228] = {.lex_state = 7, .external_lex_state = 3},
  [5229] = {.lex_state = 64, .external_lex_state = 3},
  [5230] = {.lex_state = 7, .external_lex_state = 3},
  [5231] = {.lex_state = 66, .external_lex_state = 3},
  [5232] = {.lex_state = 64, .external_lex_state = 3},
  [5233] = {.lex_state = 64, .external_lex_state = 3},
  [5234] = {.lex_state = 64, .external_lex_state = 3},
  [5235] = {.lex_state = 7, .external_lex_state = 3},
  [5236] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5237] = {.lex_state = 7, .external_lex_state = 3},
  [5238] = {.lex_state = 7, .external_lex_state = 3},
  [5239] = {.lex_state = 64, .external_lex_state = 3},
  [5240] = {.lex_state = 7, .external_lex_state = 3},
  [5241] = {.lex_state = 51, .external_lex_state = 3},
  [5242] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5243] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5244] = {.lex_state = 64, .external_lex_state = 3},
  [5245] = {.lex_state = 7, .external_lex_state = 3},
  [5246] = {.lex_state = 64, .external_lex_state = 3},
  [5247] = {.lex_state = 64, .external_lex_state = 3},
  [5248] = {.lex_state = 7, .external_lex_state = 3},
  [5249] = {.lex_state = 64, .external_lex_state = 3},
  [5250] = {.lex_state = 64, .external_lex_state = 3},
  [5251] = {.lex_state = 8, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5252] = {.lex_state = 7, .external_lex_state = 3},
  [5253] = {.lex_state = 16, .external_lex_state = 3},
  [5254] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5255] = {.lex_state = 16, .external_lex_state = 2},
  [5256] = {.lex_state = 4, .external_lex_state = 6},
  [5257] = {.lex_state = 64, .external_lex_state = 3},
  [5258] = {.lex_state = 64, .external_lex_state = 3},
  [5259] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5260] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5261] = {.lex_state = 51, .external_lex_state = 3},
  [5262] = {.lex_state = 4, .external_lex_state = 6},
  [5263] = {.lex_state = 51, .external_lex_state = 3},
  [5264] = {.lex_state = 51, .external_lex_state = 3},
  [5265] = {.lex_state = 64, .external_lex_state = 3},
  [5266] = {.lex_state = 51, .external_lex_state = 3},
  [5267] = {.lex_state = 51, .external_lex_state = 3},
  [5268] = {.lex_state = 4, .external_lex_state = 6},
  [5269] = {.lex_state = 4, .external_lex_state = 6},
  [5270] = {.lex_state = 64, .external_lex_state = 3},
  [5271] = {.lex_state = 51, .external_lex_state = 3},
  [5272] = {.lex_state = 51, .external_lex_state = 3},
  [5273] = {.lex_state = 4, .external_lex_state = 6},
  [5274] = {.lex_state = 64, .external_lex_state = 3},
  [5275] = {.lex_state = 64, .external_lex_state = 3},
  [5276] = {.lex_state = 51, .external_lex_state = 3},
  [5277] = {.lex_state = 4, .external_lex_state = 6},
  [5278] = {.lex_state = 51, .external_lex_state = 3},
  [5279] = {.lex_state = 51, .external_lex_state = 3},
  [5280] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5281] = {.lex_state = 4, .external_lex_state = 6},
  [5282] = {.lex_state = 51, .external_lex_state = 3},
  [5283] = {.lex_state = 51, .external_lex_state = 3},
  [5284] = {.lex_state = 4, .external_lex_state = 6},
  [5285] = {.lex_state = 51, .external_lex_state = 3},
  [5286] = {.lex_state = 4, .external_lex_state = 6},
  [5287] = {.lex_state = 51, .external_lex_state = 3},
  [5288] = {.lex_state = 51, .external_lex_state = 3},
  [5289] = {.lex_state = 64, .external_lex_state = 3},
  [5290] = {.lex_state = 4, .external_lex_state = 6},
  [5291] = {.lex_state = 51, .external_lex_state = 3},
  [5292] = {.lex_state = 51, .external_lex_state = 3},
  [5293] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5294] = {.lex_state = 51, .external_lex_state = 3},
  [5295] = {.lex_state = 4, .external_lex_state = 6},
  [5296] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5297] = {.lex_state = 51, .external_lex_state = 3},
  [5298] = {.lex_state = 4, .external_lex_state = 6},
  [5299] = {.lex_state = 4, .external_lex_state = 6},
  [5300] = {.lex_state = 4, .external_lex_state = 6},
  [5301] = {.lex_state = 4, .external_lex_state = 6},
  [5302] = {.lex_state = 4, .external_lex_state = 6},
  [5303] = {.lex_state = 4, .external_lex_state = 6},
  [5304] = {.lex_state = 4, .external_lex_state = 6},
  [5305] = {.lex_state = 64, .external_lex_state = 3},
  [5306] = {.lex_state = 62, .external_lex_state = 7},
  [5307] = {.lex_state = 64, .external_lex_state = 3},
  [5308] = {.lex_state = 51, .external_lex_state = 3},
  [5309] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5310] = {.lex_state = 51, .external_lex_state = 3},
  [5311] = {.lex_state = 51, .external_lex_state = 3},
  [5312] = {.lex_state = 51, .external_lex_state = 3},
  [5313] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5314] = {.lex_state = 64, .external_lex_state = 3},
  [5315] = {.lex_state = 51, .external_lex_state = 3},
  [5316] = {.lex_state = 64, .external_lex_state = 3},
  [5317] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5318] = {.lex_state = 51, .external_lex_state = 3},
  [5319] = {.lex_state = 51, .external_lex_state = 3},
  [5320] = {.lex_state = 51, .external_lex_state = 3},
  [5321] = {.lex_state = 51, .external_lex_state = 3},
  [5322] = {.lex_state = 16, .external_lex_state = 2},
  [5323] = {.lex_state = 64, .external_lex_state = 3},
  [5324] = {.lex_state = 51, .external_lex_state = 3},
  [5325] = {.lex_state = 51, .external_lex_state = 3},
  [5326] = {.lex_state = 64, .external_lex_state = 3},
  [5327] = {.lex_state = 51, .external_lex_state = 3},
  [5328] = {.lex_state = 16, .external_lex_state = 3},
  [5329] = {.lex_state = 51, .external_lex_state = 3},
  [5330] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5331] = {.lex_state = 64, .external_lex_state = 3},
  [5332] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5333] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5334] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5335] = {.lex_state = 4, .external_lex_state = 6},
  [5336] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5337] = {.lex_state = 4, .external_lex_state = 6},
  [5338] = {.lex_state = 11, .external_lex_state = 3},
  [5339] = {.lex_state = 18, .external_lex_state = 3},
  [5340] = {.lex_state = 62, .external_lex_state = 7},
  [5341] = {.lex_state = 18, .external_lex_state = 3},
  [5342] = {.lex_state = 43, .external_lex_state = 3},
  [5343] = {.lex_state = 64, .external_lex_state = 3},
  [5344] = {.lex_state = 64, .external_lex_state = 3},
  [5345] = {.lex_state = 64, .external_lex_state = 3},
  [5346] = {.lex_state = 64, .external_lex_state = 3},
  [5347] = {.lex_state = 18, .external_lex_state = 3},
  [5348] = {.lex_state = 64, .external_lex_state = 3},
  [5349] = {.lex_state = 18, .external_lex_state = 3},
  [5350] = {.lex_state = 64, .external_lex_state = 3},
  [5351] = {.lex_state = 64, .external_lex_state = 3},
  [5352] = {.lex_state = 64, .external_lex_state = 3},
  [5353] = {.lex_state = 64, .external_lex_state = 3},
  [5354] = {.lex_state = 64, .external_lex_state = 3},
  [5355] = {.lex_state = 64, .external_lex_state = 3},
  [5356] = {.lex_state = 16, .external_lex_state = 3},
  [5357] = {.lex_state = 16, .external_lex_state = 3},
  [5358] = {.lex_state = 11, .external_lex_state = 3},
  [5359] = {.lex_state = 64, .external_lex_state = 3},
  [5360] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5361] = {.lex_state = 11, .external_lex_state = 3},
  [5362] = {.lex_state = 11, .external_lex_state = 3},
  [5363] = {.lex_state = 64, .external_lex_state = 3},
  [5364] = {.lex_state = 64, .external_lex_state = 3},
  [5365] = {.lex_state = 18, .external_lex_state = 3},
  [5366] = {.lex_state = 11, .external_lex_state = 3},
  [5367] = {.lex_state = 11, .external_lex_state = 3},
  [5368] = {.lex_state = 64, .external_lex_state = 3},
  [5369] = {.lex_state = 62, .external_lex_state = 7},
  [5370] = {.lex_state = 64, .external_lex_state = 3},
  [5371] = {.lex_state = 68, .external_lex_state = 3},
  [5372] = {.lex_state = 64, .external_lex_state = 3},
  [5373] = {.lex_state = 68, .external_lex_state = 3},
  [5374] = {.lex_state = 7, .external_lex_state = 3},
  [5375] = {.lex_state = 64, .external_lex_state = 3},
  [5376] = {.lex_state = 64, .external_lex_state = 3},
  [5377] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5378] = {.lex_state = 64, .external_lex_state = 3},
  [5379] = {.lex_state = 64, .external_lex_state = 3},
  [5380] = {.lex_state = 64, .external_lex_state = 3},
  [5381] = {.lex_state = 11, .external_lex_state = 3},
  [5382] = {.lex_state = 64, .external_lex_state = 3},
  [5383] = {.lex_state = 64, .external_lex_state = 3},
  [5384] = {.lex_state = 11, .external_lex_state = 3},
  [5385] = {.lex_state = 11, .external_lex_state = 3},
  [5386] = {.lex_state = 64, .external_lex_state = 3},
  [5387] = {.lex_state = 11, .external_lex_state = 3},
  [5388] = {.lex_state = 7, .external_lex_state = 3},
  [5389] = {.lex_state = 11, .external_lex_state = 3},
  [5390] = {.lex_state = 64, .external_lex_state = 3},
  [5391] = {.lex_state = 64, .external_lex_state = 3},
  [5392] = {.lex_state = 64, .external_lex_state = 3},
  [5393] = {.lex_state = 18, .external_lex_state = 3},
  [5394] = {.lex_state = 64, .external_lex_state = 3},
  [5395] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5396] = {.lex_state = 11, .external_lex_state = 3},
  [5397] = {.lex_state = 11, .external_lex_state = 3},
  [5398] = {.lex_state = 11, .external_lex_state = 3},
  [5399] = {.lex_state = 64, .external_lex_state = 3},
  [5400] = {.lex_state = 7, .external_lex_state = 3},
  [5401] = {.lex_state = 62, .external_lex_state = 7},
  [5402] = {.lex_state = 11, .external_lex_state = 3},
  [5403] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5404] = {.lex_state = 7, .external_lex_state = 3},
  [5405] = {.lex_state = 64, .external_lex_state = 3},
  [5406] = {.lex_state = 64, .external_lex_state = 3},
  [5407] = {.lex_state = 7, .external_lex_state = 3},
  [5408] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5409] = {.lex_state = 7, .external_lex_state = 3},
  [5410] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5411] = {.lex_state = 7, .external_lex_state = 3},
  [5412] = {.lex_state = 16, .external_lex_state = 2},
  [5413] = {.lex_state = 11, .external_lex_state = 3},
  [5414] = {.lex_state = 11, .external_lex_state = 3},
  [5415] = {.lex_state = 7, .external_lex_state = 3},
  [5416] = {.lex_state = 11, .external_lex_state = 3},
  [5417] = {.lex_state = 7, .external_lex_state = 3},
  [5418] = {.lex_state = 11, .external_lex_state = 3},
  [5419] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5420] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5421] = {.lex_state = 7, .external_lex_state = 3},
  [5422] = {.lex_state = 28, .external_lex_state = 8, .reserved_word_set_id = 1},
  [5423] = {.lex_state = 7, .external_lex_state = 3},
  [5424] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5425] = {.lex_state = 28, .external_lex_state = 8, .reserved_word_set_id = 1},
  [5426] = {.lex_state = 7, .external_lex_state = 3},
  [5427] = {.lex_state = 11, .external_lex_state = 3},
  [5428] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5429] = {.lex_state = 7, .external_lex_state = 3},
  [5430] = {.lex_state = 11, .external_lex_state = 3},
  [5431] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5432] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5433] = {.lex_state = 7, .external_lex_state = 3},
  [5434] = {.lex_state = 7, .external_lex_state = 3},
  [5435] = {.lex_state = 51, .external_lex_state = 3},
  [5436] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5437] = {.lex_state = 7, .external_lex_state = 3},
  [5438] = {.lex_state = 28, .external_lex_state = 8, .reserved_word_set_id = 1},
  [5439] = {.lex_state = 28, .external_lex_state = 8, .reserved_word_set_id = 1},
  [5440] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5441] = {.lex_state = 11, .external_lex_state = 3},
  [5442] = {.lex_state = 7, .external_lex_state = 3},
  [5443] = {.lex_state = 7, .external_lex_state = 3},
  [5444] = {.lex_state = 28, .external_lex_state = 8, .reserved_word_set_id = 1},
  [5445] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5446] = {.lex_state = 7, .external_lex_state = 3},
  [5447] = {.lex_state = 7, .external_lex_state = 3},
  [5448] = {.lex_state = 11, .external_lex_state = 3},
  [5449] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5450] = {.lex_state = 7, .external_lex_state = 3},
  [5451] = {.lex_state = 11, .external_lex_state = 3},
  [5452] = {.lex_state = 7, .external_lex_state = 3},
  [5453] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5454] = {.lex_state = 7, .external_lex_state = 3},
  [5455] = {.lex_state = 7, .external_lex_state = 3},
  [5456] = {.lex_state = 51, .external_lex_state = 3},
  [5457] = {.lex_state = 11, .external_lex_state = 3},
  [5458] = {.lex_state = 51, .external_lex_state = 3},
  [5459] = {.lex_state = 51, .external_lex_state = 3},
  [5460] = {.lex_state = 51, .external_lex_state = 3},
  [5461] = {.lex_state = 7, .external_lex_state = 3},
  [5462] = {.lex_state = 51, .external_lex_state = 3},
  [5463] = {.lex_state = 7, .external_lex_state = 3},
  [5464] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5465] = {.lex_state = 51, .external_lex_state = 3},
  [5466] = {.lex_state = 11, .external_lex_state = 3},
  [5467] = {.lex_state = 11, .external_lex_state = 3},
  [5468] = {.lex_state = 7, .external_lex_state = 3},
  [5469] = {.lex_state = 7, .external_lex_state = 3},
  [5470] = {.lex_state = 7, .external_lex_state = 3},
  [5471] = {.lex_state = 64, .external_lex_state = 3},
  [5472] = {.lex_state = 7, .external_lex_state = 3},
  [5473] = {.lex_state = 7, .external_lex_state = 3},
  [5474] = {.lex_state = 7, .external_lex_state = 3},
  [5475] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5476] = {.lex_state = 64, .external_lex_state = 3},
  [5477] = {.lex_state = 7, .external_lex_state = 3},
  [5478] = {.lex_state = 7, .external_lex_state = 3},
  [5479] = {.lex_state = 64, .external_lex_state = 3},
  [5480] = {.lex_state = 7, .external_lex_state = 3},
  [5481] = {.lex_state = 7, .external_lex_state = 3},
  [5482] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5483] = {.lex_state = 7, .external_lex_state = 3},
  [5484] = {.lex_state = 7, .external_lex_state = 3},
  [5485] = {.lex_state = 7, .external_lex_state = 3},
  [5486] = {.lex_state = 7, .external_lex_state = 3},
  [5487] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5488] = {.lex_state = 64, .external_lex_state = 3},
  [5489] = {.lex_state = 7, .external_lex_state = 3},
  [5490] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 95},
  [5491] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5492] = {.lex_state = 64, .external_lex_state = 3},
  [5493] = {.lex_state = 64, .external_lex_state = 3},
  [5494] = {.lex_state = 64, .external_lex_state = 3},
  [5495] = {.lex_state = 64, .external_lex_state = 3},
  [5496] = {.lex_state = 28, .external_lex_state = 8, .reserved_word_set_id = 1},
  [5497] = {.lex_state = 64, .external_lex_state = 3},
  [5498] = {.lex_state = 64, .external_lex_state = 3},
  [5499] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5500] = {.lex_state = 51, .external_lex_state = 3},
  [5501] = {.lex_state = 51, .external_lex_state = 3},
  [5502] = {.lex_state = 51, .external_lex_state = 3},
  [5503] = {.lex_state = 51, .external_lex_state = 3},
  [5504] = {.lex_state = 51, .external_lex_state = 3},
  [5505] = {.lex_state = 64, .external_lex_state = 3},
  [5506] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5507] = {.lex_state = 68, .external_lex_state = 3},
  [5508] = {.lex_state = 28, .external_lex_state = 8, .reserved_word_set_id = 1},
  [5509] = {.lex_state = 64, .external_lex_state = 3},
  [5510] = {.lex_state = 7, .external_lex_state = 3},
  [5511] = {.lex_state = 51, .external_lex_state = 3},
  [5512] = {.lex_state = 64, .external_lex_state = 3},
  [5513] = {.lex_state = 64, .external_lex_state = 3},
  [5514] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5515] = {.lex_state = 28, .external_lex_state = 8, .reserved_word_set_id = 1},
  [5516] = {.lex_state = 71, .external_lex_state = 3},
  [5517] = {.lex_state = 51, .external_lex_state = 3},
  [5518] = {.lex_state = 64, .external_lex_state = 3},
  [5519] = {.lex_state = 71, .external_lex_state = 3},
  [5520] = {.lex_state = 11, .external_lex_state = 3},
  [5521] = {.lex_state = 11, .external_lex_state = 3},
  [5522] = {.lex_state = 7, .external_lex_state = 3},
  [5523] = {.lex_state = 64, .external_lex_state = 3},
  [5524] = {.lex_state = 64, .external_lex_state = 3},
  [5525] = {.lex_state = 64, .external_lex_state = 3},
  [5526] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5527] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 95},
  [5528] = {.lex_state = 7, .external_lex_state = 3},
  [5529] = {.lex_state = 7, .external_lex_state = 3},
  [5530] = {.lex_state = 51, .external_lex_state = 3},
  [5531] = {.lex_state = 7, .external_lex_state = 3},
  [5532] = {.lex_state = 11, .external_lex_state = 3},
  [5533] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5534] = {.lex_state = 7, .external_lex_state = 3},
  [5535] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5536] = {.lex_state = 7, .external_lex_state = 3},
  [5537] = {.lex_state = 7, .external_lex_state = 3},
  [5538] = {.lex_state = 7, .external_lex_state = 3},
  [5539] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5540] = {.lex_state = 7, .external_lex_state = 3},
  [5541] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5542] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5543] = {.lex_state = 11, .external_lex_state = 3},
  [5544] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5545] = {.lex_state = 11, .external_lex_state = 3},
  [5546] = {.lex_state = 11, .external_lex_state = 3},
  [5547] = {.lex_state = 11, .external_lex_state = 3},
  [5548] = {.lex_state = 11, .external_lex_state = 3},
  [5549] = {.lex_state = 120, .external_lex_state = 3},
  [5550] = {.lex_state = 25, .external_lex_state = 3},
  [5551] = {.lex_state = 11, .external_lex_state = 3},
  [5552] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5553] = {.lex_state = 25, .external_lex_state = 3},
  [5554] = {.lex_state = 25, .external_lex_state = 3},
  [5555] = {.lex_state = 120, .external_lex_state = 3},
  [5556] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5557] = {.lex_state = 11, .external_lex_state = 3},
  [5558] = {.lex_state = 11, .external_lex_state = 3},
  [5559] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5560] = {.lex_state = 11, .external_lex_state = 3},
  [5561] = {.lex_state = 11, .external_lex_state = 3},
  [5562] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5563] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5564] = {.lex_state = 11, .external_lex_state = 3},
  [5565] = {.lex_state = 11, .external_lex_state = 3},
  [5566] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5567] = {.lex_state = 11, .external_lex_state = 3},
  [5568] = {.lex_state = 51, .external_lex_state = 3},
  [5569] = {.lex_state = 11, .external_lex_state = 3},
  [5570] = {.lex_state = 11, .external_lex_state = 3},
  [5571] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5572] = {.lex_state = 25, .external_lex_state = 3},
  [5573] = {.lex_state = 51, .external_lex_state = 3},
  [5574] = {.lex_state = 25, .external_lex_state = 3},
  [5575] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5576] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5577] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5578] = {.lex_state = 25, .external_lex_state = 3},
  [5579] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5580] = {.lex_state = 11, .external_lex_state = 3},
  [5581] = {.lex_state = 11, .external_lex_state = 3},
  [5582] = {.lex_state = 11, .external_lex_state = 3},
  [5583] = {.lex_state = 11, .external_lex_state = 3},
  [5584] = {.lex_state = 11, .external_lex_state = 3},
  [5585] = {.lex_state = 11, .external_lex_state = 3},
  [5586] = {.lex_state = 11, .external_lex_state = 3},
  [5587] = {.lex_state = 11, .external_lex_state = 3},
  [5588] = {.lex_state = 11, .external_lex_state = 3},
  [5589] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5590] = {.lex_state = 11, .external_lex_state = 3},
  [5591] = {.lex_state = 11, .external_lex_state = 3},
  [5592] = {.lex_state = 11, .external_lex_state = 3},
  [5593] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5594] = {.lex_state = 25, .external_lex_state = 3},
  [5595] = {.lex_state = 11, .external_lex_state = 3},
  [5596] = {.lex_state = 11, .external_lex_state = 3},
  [5597] = {.lex_state = 11, .external_lex_state = 3},
  [5598] = {.lex_state = 11, .external_lex_state = 3},
  [5599] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5600] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5601] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5602] = {.lex_state = 64, .external_lex_state = 3},
  [5603] = {.lex_state = 11, .external_lex_state = 3},
  [5604] = {.lex_state = 11, .external_lex_state = 3},
  [5605] = {.lex_state = 64, .external_lex_state = 3},
  [5606] = {.lex_state = 11, .external_lex_state = 3},
  [5607] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5608] = {.lex_state = 11, .external_lex_state = 3},
  [5609] = {.lex_state = 11, .external_lex_state = 3},
  [5610] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5611] = {.lex_state = 11, .external_lex_state = 3},
  [5612] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5613] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5614] = {.lex_state = 120, .external_lex_state = 3},
  [5615] = {.lex_state = 25, .external_lex_state = 3},
  [5616] = {.lex_state = 25, .external_lex_state = 3},
  [5617] = {.lex_state = 11, .external_lex_state = 3},
  [5618] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5619] = {.lex_state = 25, .external_lex_state = 3},
  [5620] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 96},
  [5621] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5622] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5623] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5624] = {.lex_state = 11, .external_lex_state = 3},
  [5625] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5626] = {.lex_state = 11, .external_lex_state = 3},
  [5627] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5628] = {.lex_state = 120, .external_lex_state = 3},
  [5629] = {.lex_state = 64, .external_lex_state = 3},
  [5630] = {.lex_state = 11, .external_lex_state = 3},
  [5631] = {.lex_state = 11, .external_lex_state = 3},
  [5632] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5633] = {.lex_state = 11, .external_lex_state = 3},
  [5634] = {.lex_state = 11, .external_lex_state = 3},
  [5635] = {.lex_state = 51, .external_lex_state = 3},
  [5636] = {.lex_state = 11, .external_lex_state = 3},
  [5637] = {.lex_state = 11, .external_lex_state = 3},
  [5638] = {.lex_state = 11, .external_lex_state = 3},
  [5639] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5640] = {.lex_state = 11, .external_lex_state = 3},
  [5641] = {.lex_state = 64, .external_lex_state = 3},
  [5642] = {.lex_state = 11, .external_lex_state = 3},
  [5643] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5644] = {.lex_state = 64, .external_lex_state = 3},
  [5645] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5646] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5647] = {.lex_state = 11, .external_lex_state = 3},
  [5648] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5649] = {.lex_state = 11, .external_lex_state = 3},
  [5650] = {.lex_state = 11, .external_lex_state = 3},
  [5651] = {.lex_state = 7, .external_lex_state = 3},
  [5652] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5653] = {.lex_state = 11, .external_lex_state = 3},
  [5654] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5655] = {.lex_state = 11, .external_lex_state = 3},
  [5656] = {.lex_state = 11, .external_lex_state = 3},
  [5657] = {.lex_state = 7, .external_lex_state = 3},
  [5658] = {.lex_state = 7, .external_lex_state = 3},
  [5659] = {.lex_state = 120, .external_lex_state = 3},
  [5660] = {.lex_state = 11, .external_lex_state = 3},
  [5661] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5662] = {.lex_state = 7, .external_lex_state = 3},
  [5663] = {.lex_state = 7, .external_lex_state = 3},
  [5664] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5665] = {.lex_state = 120, .external_lex_state = 3},
  [5666] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5667] = {.lex_state = 7, .external_lex_state = 3},
  [5668] = {.lex_state = 25, .external_lex_state = 3},
  [5669] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5670] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5671] = {.lex_state = 7, .external_lex_state = 3},
  [5672] = {.lex_state = 120, .external_lex_state = 3},
  [5673] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5674] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5675] = {.lex_state = 51, .external_lex_state = 3},
  [5676] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5677] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5678] = {.lex_state = 11, .external_lex_state = 3},
  [5679] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5680] = {.lex_state = 51, .external_lex_state = 3},
  [5681] = {.lex_state = 11, .external_lex_state = 3},
  [5682] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5683] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5684] = {.lex_state = 64, .external_lex_state = 3},
  [5685] = {.lex_state = 11, .external_lex_state = 3},
  [5686] = {.lex_state = 7, .external_lex_state = 3},
  [5687] = {.lex_state = 120, .external_lex_state = 3},
  [5688] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5689] = {.lex_state = 120, .external_lex_state = 3},
  [5690] = {.lex_state = 11, .external_lex_state = 3},
  [5691] = {.lex_state = 51, .external_lex_state = 3},
  [5692] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5693] = {.lex_state = 64, .external_lex_state = 3},
  [5694] = {.lex_state = 64, .external_lex_state = 3},
  [5695] = {.lex_state = 51, .external_lex_state = 3},
  [5696] = {.lex_state = 64, .external_lex_state = 3},
  [5697] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5698] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5699] = {.lex_state = 51, .external_lex_state = 3},
  [5700] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5701] = {.lex_state = 64, .external_lex_state = 3},
  [5702] = {.lex_state = 11, .external_lex_state = 3},
  [5703] = {.lex_state = 11, .external_lex_state = 3},
  [5704] = {.lex_state = 6, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5705] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5706] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5707] = {.lex_state = 11, .external_lex_state = 3},
  [5708] = {.lex_state = 11, .external_lex_state = 3},
  [5709] = {.lex_state = 11, .external_lex_state = 3},
  [5710] = {.lex_state = 11, .external_lex_state = 3},
  [5711] = {.lex_state = 11, .external_lex_state = 3},
  [5712] = {.lex_state = 11, .external_lex_state = 3},
  [5713] = {.lex_state = 11, .external_lex_state = 3},
  [5714] = {.lex_state = 11, .external_lex_state = 3},
  [5715] = {.lex_state = 11, .external_lex_state = 3},
  [5716] = {.lex_state = 11, .external_lex_state = 3},
  [5717] = {.lex_state = 11, .external_lex_state = 3},
  [5718] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5719] = {.lex_state = 11, .external_lex_state = 3},
  [5720] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5721] = {.lex_state = 11, .external_lex_state = 3},
  [5722] = {.lex_state = 11, .external_lex_state = 3},
  [5723] = {.lex_state = 11, .external_lex_state = 3},
  [5724] = {.lex_state = 11, .external_lex_state = 3},
  [5725] = {.lex_state = 11, .external_lex_state = 3},
  [5726] = {.lex_state = 11, .external_lex_state = 3},
  [5727] = {.lex_state = 11, .external_lex_state = 3},
  [5728] = {.lex_state = 11, .external_lex_state = 3},
  [5729] = {.lex_state = 11, .external_lex_state = 3},
  [5730] = {.lex_state = 11, .external_lex_state = 3},
  [5731] = {.lex_state = 11, .external_lex_state = 3},
  [5732] = {.lex_state = 25, .external_lex_state = 3},
  [5733] = {.lex_state = 25, .external_lex_state = 3},
  [5734] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5735] = {.lex_state = 11, .external_lex_state = 3},
  [5736] = {.lex_state = 25, .external_lex_state = 3},
  [5737] = {.lex_state = 25, .external_lex_state = 3},
  [5738] = {.lex_state = 25, .external_lex_state = 3},
  [5739] = {.lex_state = 7, .external_lex_state = 3},
  [5740] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5741] = {.lex_state = 51, .external_lex_state = 3},
  [5742] = {.lex_state = 51, .external_lex_state = 3},
  [5743] = {.lex_state = 11, .external_lex_state = 3},
  [5744] = {.lex_state = 7, .external_lex_state = 3},
  [5745] = {.lex_state = 64, .external_lex_state = 3},
  [5746] = {.lex_state = 64, .external_lex_state = 3},
  [5747] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5748] = {.lex_state = 64, .external_lex_state = 3},
  [5749] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5750] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5751] = {.lex_state = 120, .external_lex_state = 3},
  [5752] = {.lex_state = 120, .external_lex_state = 3},
  [5753] = {.lex_state = 11, .external_lex_state = 3},
  [5754] = {.lex_state = 11, .external_lex_state = 3},
  [5755] = {.lex_state = 7, .external_lex_state = 3},
  [5756] = {.lex_state = 11, .external_lex_state = 3},
  [5757] = {.lex_state = 51, .external_lex_state = 3},
  [5758] = {.lex_state = 25, .external_lex_state = 3},
  [5759] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5760] = {.lex_state = 64, .external_lex_state = 3},
  [5761] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5762] = {.lex_state = 11, .external_lex_state = 3},
  [5763] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5764] = {.lex_state = 120, .external_lex_state = 3},
  [5765] = {.lex_state = 64, .external_lex_state = 3},
  [5766] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5767] = {.lex_state = 120, .external_lex_state = 3},
  [5768] = {.lex_state = 11, .external_lex_state = 3},
  [5769] = {.lex_state = 64, .external_lex_state = 3},
  [5770] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5771] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5772] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5773] = {.lex_state = 11, .external_lex_state = 3},
  [5774] = {.lex_state = 11, .external_lex_state = 3},
  [5775] = {.lex_state = 64, .external_lex_state = 3},
  [5776] = {.lex_state = 64, .external_lex_state = 3},
  [5777] = {.lex_state = 51, .external_lex_state = 3},
  [5778] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5779] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5780] = {.lex_state = 11, .external_lex_state = 3},
  [5781] = {.lex_state = 7, .external_lex_state = 3},
  [5782] = {.lex_state = 11, .external_lex_state = 3},
  [5783] = {.lex_state = 25, .external_lex_state = 3},
  [5784] = {.lex_state = 11, .external_lex_state = 3},
  [5785] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5786] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5787] = {.lex_state = 120, .external_lex_state = 2},
  [5788] = {.lex_state = 64, .external_lex_state = 3},
  [5789] = {.lex_state = 64, .external_lex_state = 3},
  [5790] = {.lex_state = 64, .external_lex_state = 3},
  [5791] = {.lex_state = 7, .external_lex_state = 3},
  [5792] = {.lex_state = 64, .external_lex_state = 3},
  [5793] = {.lex_state = 64, .external_lex_state = 3},
  [5794] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5795] = {.lex_state = 64, .external_lex_state = 3},
  [5796] = {.lex_state = 64, .external_lex_state = 3},
  [5797] = {.lex_state = 64, .external_lex_state = 3},
  [5798] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5799] = {.lex_state = 64, .external_lex_state = 3},
  [5800] = {.lex_state = 64, .external_lex_state = 3},
  [5801] = {.lex_state = 7, .external_lex_state = 3},
  [5802] = {.lex_state = 64, .external_lex_state = 3},
  [5803] = {.lex_state = 64, .external_lex_state = 3},
  [5804] = {.lex_state = 64, .external_lex_state = 3},
  [5805] = {.lex_state = 64, .external_lex_state = 3},
  [5806] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5807] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5808] = {.lex_state = 64, .external_lex_state = 3},
  [5809] = {.lex_state = 64, .external_lex_state = 3},
  [5810] = {.lex_state = 64, .external_lex_state = 3},
  [5811] = {.lex_state = 64, .external_lex_state = 3},
  [5812] = {.lex_state = 64, .external_lex_state = 3},
  [5813] = {.lex_state = 64, .external_lex_state = 3},
  [5814] = {.lex_state = 64, .external_lex_state = 3},
  [5815] = {.lex_state = 64, .external_lex_state = 3},
  [5816] = {.lex_state = 64, .external_lex_state = 3},
  [5817] = {.lex_state = 7, .external_lex_state = 3},
  [5818] = {.lex_state = 64, .external_lex_state = 3},
  [5819] = {.lex_state = 64, .external_lex_state = 3},
  [5820] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5821] = {.lex_state = 121, .external_lex_state = 3},
  [5822] = {.lex_state = 7, .external_lex_state = 3},
  [5823] = {.lex_state = 121, .external_lex_state = 3},
  [5824] = {.lex_state = 64, .external_lex_state = 3},
  [5825] = {.lex_state = 64, .external_lex_state = 3},
  [5826] = {.lex_state = 64, .external_lex_state = 3},
  [5827] = {.lex_state = 64, .external_lex_state = 3},
  [5828] = {.lex_state = 64, .external_lex_state = 3},
  [5829] = {.lex_state = 64, .external_lex_state = 3},
  [5830] = {.lex_state = 64, .external_lex_state = 3},
  [5831] = {.lex_state = 64, .external_lex_state = 3},
  [5832] = {.lex_state = 7, .external_lex_state = 3},
  [5833] = {.lex_state = 64, .external_lex_state = 3},
  [5834] = {.lex_state = 43, .external_lex_state = 3},
  [5835] = {.lex_state = 120, .external_lex_state = 3},
  [5836] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5837] = {.lex_state = 64, .external_lex_state = 3},
  [5838] = {.lex_state = 120, .external_lex_state = 3},
  [5839] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5840] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5841] = {.lex_state = 120, .external_lex_state = 3},
  [5842] = {.lex_state = 64, .external_lex_state = 3},
  [5843] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5844] = {.lex_state = 7, .external_lex_state = 3},
  [5845] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5846] = {.lex_state = 64, .external_lex_state = 3},
  [5847] = {.lex_state = 120, .external_lex_state = 2},
  [5848] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5849] = {.lex_state = 7, .external_lex_state = 3},
  [5850] = {.lex_state = 64, .external_lex_state = 3},
  [5851] = {.lex_state = 64, .external_lex_state = 3},
  [5852] = {.lex_state = 120, .external_lex_state = 3},
  [5853] = {.lex_state = 120, .external_lex_state = 2},
  [5854] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5855] = {.lex_state = 64, .external_lex_state = 3},
  [5856] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5857] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5858] = {.lex_state = 120, .external_lex_state = 2},
  [5859] = {.lex_state = 64, .external_lex_state = 3},
  [5860] = {.lex_state = 120, .external_lex_state = 3},
  [5861] = {.lex_state = 64, .external_lex_state = 3},
  [5862] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5863] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5864] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5865] = {.lex_state = 64, .external_lex_state = 3},
  [5866] = {.lex_state = 120, .external_lex_state = 3},
  [5867] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5868] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5869] = {.lex_state = 64, .external_lex_state = 3},
  [5870] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5871] = {.lex_state = 120, .external_lex_state = 3},
  [5872] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5873] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5874] = {.lex_state = 64, .external_lex_state = 3},
  [5875] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5876] = {.lex_state = 64, .external_lex_state = 3},
  [5877] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5878] = {.lex_state = 16, .external_lex_state = 3},
  [5879] = {.lex_state = 64, .external_lex_state = 3},
  [5880] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5881] = {.lex_state = 64, .external_lex_state = 3},
  [5882] = {.lex_state = 64, .external_lex_state = 3},
  [5883] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5884] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5885] = {.lex_state = 120, .external_lex_state = 2},
  [5886] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5887] = {.lex_state = 64, .external_lex_state = 3},
  [5888] = {.lex_state = 64, .external_lex_state = 3},
  [5889] = {.lex_state = 64, .external_lex_state = 3},
  [5890] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5891] = {.lex_state = 64, .external_lex_state = 3},
  [5892] = {.lex_state = 120, .external_lex_state = 3},
  [5893] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5894] = {.lex_state = 16, .external_lex_state = 3},
  [5895] = {.lex_state = 7, .external_lex_state = 3},
  [5896] = {.lex_state = 16, .external_lex_state = 3},
  [5897] = {.lex_state = 64, .external_lex_state = 3},
  [5898] = {.lex_state = 64, .external_lex_state = 3},
  [5899] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5900] = {.lex_state = 64, .external_lex_state = 3},
  [5901] = {.lex_state = 120, .external_lex_state = 3},
  [5902] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5903] = {.lex_state = 64, .external_lex_state = 3},
  [5904] = {.lex_state = 64, .external_lex_state = 3},
  [5905] = {.lex_state = 64, .external_lex_state = 3},
  [5906] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5907] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5908] = {.lex_state = 64, .external_lex_state = 3},
  [5909] = {.lex_state = 64, .external_lex_state = 3},
  [5910] = {.lex_state = 64, .external_lex_state = 3},
  [5911] = {.lex_state = 64, .external_lex_state = 3},
  [5912] = {.lex_state = 43, .external_lex_state = 3},
  [5913] = {.lex_state = 51, .external_lex_state = 3},
  [5914] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5915] = {.lex_state = 64, .external_lex_state = 3},
  [5916] = {.lex_state = 64, .external_lex_state = 3},
  [5917] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5918] = {.lex_state = 64, .external_lex_state = 3},
  [5919] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5920] = {.lex_state = 64, .external_lex_state = 3},
  [5921] = {.lex_state = 16, .external_lex_state = 3},
  [5922] = {.lex_state = 64, .external_lex_state = 3},
  [5923] = {.lex_state = 64, .external_lex_state = 3},
  [5924] = {.lex_state = 64, .external_lex_state = 3},
  [5925] = {.lex_state = 64, .external_lex_state = 3},
  [5926] = {.lex_state = 64, .external_lex_state = 3},
  [5927] = {.lex_state = 16, .external_lex_state = 3},
  [5928] = {.lex_state = 7, .external_lex_state = 3},
  [5929] = {.lex_state = 120, .external_lex_state = 3},
  [5930] = {.lex_state = 64, .external_lex_state = 3},
  [5931] = {.lex_state = 121, .external_lex_state = 3},
  [5932] = {.lex_state = 64, .external_lex_state = 3},
  [5933] = {.lex_state = 64, .external_lex_state = 3},
  [5934] = {.lex_state = 64, .external_lex_state = 3},
  [5935] = {.lex_state = 64, .external_lex_state = 3},
  [5936] = {.lex_state = 16, .external_lex_state = 3},
  [5937] = {.lex_state = 64, .external_lex_state = 3},
  [5938] = {.lex_state = 16, .external_lex_state = 3},
  [5939] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5940] = {.lex_state = 7, .external_lex_state = 3},
  [5941] = {.lex_state = 64, .external_lex_state = 3},
  [5942] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5943] = {.lex_state = 64, .external_lex_state = 3},
  [5944] = {.lex_state = 64, .external_lex_state = 3},
  [5945] = {.lex_state = 64, .external_lex_state = 3},
  [5946] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5947] = {.lex_state = 11, .external_lex_state = 3},
  [5948] = {.lex_state = 64, .external_lex_state = 3},
  [5949] = {.lex_state = 64, .external_lex_state = 3},
  [5950] = {.lex_state = 6, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5951] = {.lex_state = 64, .external_lex_state = 3},
  [5952] = {.lex_state = 43, .external_lex_state = 3},
  [5953] = {.lex_state = 16, .external_lex_state = 3},
  [5954] = {.lex_state = 64, .external_lex_state = 3},
  [5955] = {.lex_state = 7, .external_lex_state = 3},
  [5956] = {.lex_state = 64, .external_lex_state = 3},
  [5957] = {.lex_state = 120, .external_lex_state = 3},
  [5958] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5959] = {.lex_state = 64, .external_lex_state = 3},
  [5960] = {.lex_state = 64, .external_lex_state = 3},
  [5961] = {.lex_state = 64, .external_lex_state = 3},
  [5962] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5963] = {.lex_state = 120, .external_lex_state = 3},
  [5964] = {.lex_state = 64, .external_lex_state = 3},
  [5965] = {.lex_state = 64, .external_lex_state = 3},
  [5966] = {.lex_state = 7, .external_lex_state = 3},
  [5967] = {.lex_state = 64, .external_lex_state = 3},
  [5968] = {.lex_state = 64, .external_lex_state = 3},
  [5969] = {.lex_state = 64, .external_lex_state = 3},
  [5970] = {.lex_state = 64, .external_lex_state = 3},
  [5971] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5972] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5973] = {.lex_state = 64, .external_lex_state = 3},
  [5974] = {.lex_state = 64, .external_lex_state = 3},
  [5975] = {.lex_state = 64, .external_lex_state = 3},
  [5976] = {.lex_state = 64, .external_lex_state = 3},
  [5977] = {.lex_state = 64, .external_lex_state = 3},
  [5978] = {.lex_state = 64, .external_lex_state = 3},
  [5979] = {.lex_state = 64, .external_lex_state = 3},
  [5980] = {.lex_state = 120, .external_lex_state = 2},
  [5981] = {.lex_state = 64, .external_lex_state = 3},
  [5982] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5983] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5984] = {.lex_state = 64, .external_lex_state = 3},
  [5985] = {.lex_state = 64, .external_lex_state = 3},
  [5986] = {.lex_state = 64, .external_lex_state = 3},
  [5987] = {.lex_state = 64, .external_lex_state = 3},
  [5988] = {.lex_state = 64, .external_lex_state = 3},
  [5989] = {.lex_state = 64, .external_lex_state = 3},
  [5990] = {.lex_state = 64, .external_lex_state = 3},
  [5991] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5992] = {.lex_state = 64, .external_lex_state = 3},
  [5993] = {.lex_state = 7, .external_lex_state = 3},
  [5994] = {.lex_state = 64, .external_lex_state = 3},
  [5995] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [5996] = {.lex_state = 16, .external_lex_state = 3},
  [5997] = {.lex_state = 64, .external_lex_state = 3},
  [5998] = {.lex_state = 64, .external_lex_state = 3},
  [5999] = {.lex_state = 7, .external_lex_state = 3},
  [6000] = {.lex_state = 7, .external_lex_state = 3},
  [6001] = {.lex_state = 64, .external_lex_state = 3},
  [6002] = {.lex_state = 64, .external_lex_state = 3},
  [6003] = {.lex_state = 7, .external_lex_state = 3},
  [6004] = {.lex_state = 7, .external_lex_state = 3},
  [6005] = {.lex_state = 7, .external_lex_state = 3},
  [6006] = {.lex_state = 7, .external_lex_state = 3},
  [6007] = {.lex_state = 120, .external_lex_state = 3},
  [6008] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6009] = {.lex_state = 64, .external_lex_state = 3},
  [6010] = {.lex_state = 7, .external_lex_state = 3},
  [6011] = {.lex_state = 64, .external_lex_state = 3},
  [6012] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6013] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6014] = {.lex_state = 25, .external_lex_state = 3},
  [6015] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6016] = {.lex_state = 16, .external_lex_state = 3},
  [6017] = {.lex_state = 123, .external_lex_state = 3},
  [6018] = {.lex_state = 16, .external_lex_state = 3},
  [6019] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6020] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6021] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6022] = {.lex_state = 16, .external_lex_state = 3},
  [6023] = {.lex_state = 121, .external_lex_state = 3},
  [6024] = {.lex_state = 121, .external_lex_state = 3},
  [6025] = {.lex_state = 7, .external_lex_state = 3},
  [6026] = {.lex_state = 16, .external_lex_state = 3},
  [6027] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6028] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6029] = {.lex_state = 16, .external_lex_state = 3},
  [6030] = {.lex_state = 16, .external_lex_state = 3},
  [6031] = {.lex_state = 123, .external_lex_state = 3},
  [6032] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6033] = {.lex_state = 74, .external_lex_state = 8},
  [6034] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6035] = {.lex_state = 16, .external_lex_state = 3},
  [6036] = {.lex_state = 25, .external_lex_state = 3},
  [6037] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6038] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6039] = {.lex_state = 121, .external_lex_state = 3},
  [6040] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6041] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 96},
  [6042] = {.lex_state = 16, .external_lex_state = 3},
  [6043] = {.lex_state = 74, .external_lex_state = 8},
  [6044] = {.lex_state = 25, .external_lex_state = 3},
  [6045] = {.lex_state = 16, .external_lex_state = 3},
  [6046] = {.lex_state = 16, .external_lex_state = 3},
  [6047] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 97},
  [6048] = {.lex_state = 121, .external_lex_state = 3},
  [6049] = {.lex_state = 121, .external_lex_state = 3},
  [6050] = {.lex_state = 7, .external_lex_state = 3},
  [6051] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6052] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6053] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6054] = {.lex_state = 123, .external_lex_state = 3},
  [6055] = {.lex_state = 51, .external_lex_state = 3},
  [6056] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6057] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6058] = {.lex_state = 16, .external_lex_state = 3},
  [6059] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6060] = {.lex_state = 7, .external_lex_state = 3},
  [6061] = {.lex_state = 7, .external_lex_state = 3},
  [6062] = {.lex_state = 16, .external_lex_state = 3},
  [6063] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6064] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6065] = {.lex_state = 16, .external_lex_state = 3},
  [6066] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6067] = {.lex_state = 120, .external_lex_state = 3},
  [6068] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6069] = {.lex_state = 123, .external_lex_state = 3},
  [6070] = {.lex_state = 121, .external_lex_state = 3},
  [6071] = {.lex_state = 120, .external_lex_state = 3},
  [6072] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6073] = {.lex_state = 7, .external_lex_state = 3},
  [6074] = {.lex_state = 7, .external_lex_state = 3},
  [6075] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6076] = {.lex_state = 123, .external_lex_state = 3},
  [6077] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6078] = {.lex_state = 16, .external_lex_state = 3},
  [6079] = {.lex_state = 7, .external_lex_state = 3},
  [6080] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 97},
  [6081] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6082] = {.lex_state = 7, .external_lex_state = 3},
  [6083] = {.lex_state = 123, .external_lex_state = 3},
  [6084] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6085] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6086] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6087] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6088] = {.lex_state = 120, .external_lex_state = 3},
  [6089] = {.lex_state = 16, .external_lex_state = 3},
  [6090] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6091] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6092] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6093] = {.lex_state = 16, .external_lex_state = 3},
  [6094] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6095] = {.lex_state = 16, .external_lex_state = 3},
  [6096] = {.lex_state = 16, .external_lex_state = 3},
  [6097] = {.lex_state = 120, .external_lex_state = 3},
  [6098] = {.lex_state = 7, .external_lex_state = 3},
  [6099] = {.lex_state = 120, .external_lex_state = 3},
  [6100] = {.lex_state = 123, .external_lex_state = 3},
  [6101] = {.lex_state = 64, .external_lex_state = 3},
  [6102] = {.lex_state = 16, .external_lex_state = 3},
  [6103] = {.lex_state = 120, .external_lex_state = 2},
  [6104] = {.lex_state = 7, .external_lex_state = 3},
  [6105] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6106] = {.lex_state = 7, .external_lex_state = 3},
  [6107] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6108] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6109] = {.lex_state = 7, .external_lex_state = 3},
  [6110] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6111] = {.lex_state = 64, .external_lex_state = 3},
  [6112] = {.lex_state = 64, .external_lex_state = 3},
  [6113] = {.lex_state = 64, .external_lex_state = 3},
  [6114] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6115] = {.lex_state = 120, .external_lex_state = 3},
  [6116] = {.lex_state = 7, .external_lex_state = 3},
  [6117] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6118] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6119] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 96},
  [6120] = {.lex_state = 120, .external_lex_state = 3},
  [6121] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6122] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6123] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6124] = {.lex_state = 120, .external_lex_state = 3},
  [6125] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6126] = {.lex_state = 25, .external_lex_state = 3},
  [6127] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6128] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 96},
  [6129] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6130] = {.lex_state = 120, .external_lex_state = 3},
  [6131] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6132] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6133] = {.lex_state = 64, .external_lex_state = 3},
  [6134] = {.lex_state = 120, .external_lex_state = 3},
  [6135] = {.lex_state = 64, .external_lex_state = 3},
  [6136] = {.lex_state = 26, .external_lex_state = 3},
  [6137] = {.lex_state = 120, .external_lex_state = 3},
  [6138] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6139] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6140] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6141] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6142] = {.lex_state = 16, .external_lex_state = 3},
  [6143] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6144] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6145] = {.lex_state = 7, .external_lex_state = 3},
  [6146] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6147] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 97},
  [6148] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 96},
  [6149] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6150] = {.lex_state = 16, .external_lex_state = 3},
  [6151] = {.lex_state = 16, .external_lex_state = 3},
  [6152] = {.lex_state = 16, .external_lex_state = 3},
  [6153] = {.lex_state = 7, .external_lex_state = 3},
  [6154] = {.lex_state = 16, .external_lex_state = 3},
  [6155] = {.lex_state = 7, .external_lex_state = 3},
  [6156] = {.lex_state = 121, .external_lex_state = 3},
  [6157] = {.lex_state = 7, .external_lex_state = 3},
  [6158] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6159] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6160] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6161] = {.lex_state = 64, .external_lex_state = 3},
  [6162] = {.lex_state = 53, .external_lex_state = 3},
  [6163] = {.lex_state = 53, .external_lex_state = 3},
  [6164] = {.lex_state = 120, .external_lex_state = 3},
  [6165] = {.lex_state = 120, .external_lex_state = 3},
  [6166] = {.lex_state = 25, .external_lex_state = 3},
  [6167] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 97},
  [6168] = {.lex_state = 120, .external_lex_state = 3},
  [6169] = {.lex_state = 120, .external_lex_state = 3},
  [6170] = {.lex_state = 120, .external_lex_state = 3},
  [6171] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6172] = {.lex_state = 7, .external_lex_state = 3},
  [6173] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6174] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6175] = {.lex_state = 16, .external_lex_state = 3},
  [6176] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6177] = {.lex_state = 7, .external_lex_state = 3},
  [6178] = {.lex_state = 121, .external_lex_state = 3},
  [6179] = {.lex_state = 16, .external_lex_state = 3},
  [6180] = {.lex_state = 16, .external_lex_state = 3},
  [6181] = {.lex_state = 120, .external_lex_state = 3},
  [6182] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6183] = {.lex_state = 64, .external_lex_state = 3},
  [6184] = {.lex_state = 25, .external_lex_state = 3},
  [6185] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6186] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6187] = {.lex_state = 16, .external_lex_state = 2},
  [6188] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6189] = {.lex_state = 64, .external_lex_state = 3},
  [6190] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6191] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6192] = {.lex_state = 64, .external_lex_state = 3},
  [6193] = {.lex_state = 120, .external_lex_state = 3},
  [6194] = {.lex_state = 64, .external_lex_state = 3},
  [6195] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6196] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6197] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6198] = {.lex_state = 120, .external_lex_state = 3},
  [6199] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6200] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6201] = {.lex_state = 16, .external_lex_state = 3},
  [6202] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6203] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6204] = {.lex_state = 120, .external_lex_state = 3},
  [6205] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6206] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6207] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6208] = {.lex_state = 16, .external_lex_state = 3},
  [6209] = {.lex_state = 120, .external_lex_state = 2},
  [6210] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6211] = {.lex_state = 120, .external_lex_state = 3},
  [6212] = {.lex_state = 121, .external_lex_state = 3},
  [6213] = {.lex_state = 121, .external_lex_state = 3},
  [6214] = {.lex_state = 74, .external_lex_state = 8},
  [6215] = {.lex_state = 7, .external_lex_state = 3},
  [6216] = {.lex_state = 16, .external_lex_state = 3},
  [6217] = {.lex_state = 16, .external_lex_state = 3},
  [6218] = {.lex_state = 25, .external_lex_state = 3},
  [6219] = {.lex_state = 25, .external_lex_state = 3},
  [6220] = {.lex_state = 16, .external_lex_state = 3},
  [6221] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6222] = {.lex_state = 16, .external_lex_state = 3},
  [6223] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6224] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6225] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6226] = {.lex_state = 16, .external_lex_state = 3},
  [6227] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6228] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6229] = {.lex_state = 121, .external_lex_state = 3},
  [6230] = {.lex_state = 64, .external_lex_state = 3},
  [6231] = {.lex_state = 16, .external_lex_state = 3},
  [6232] = {.lex_state = 16, .external_lex_state = 3},
  [6233] = {.lex_state = 120, .external_lex_state = 3},
  [6234] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6235] = {.lex_state = 16, .external_lex_state = 3},
  [6236] = {.lex_state = 16, .external_lex_state = 3},
  [6237] = {.lex_state = 123, .external_lex_state = 3},
  [6238] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6239] = {.lex_state = 16, .external_lex_state = 3},
  [6240] = {.lex_state = 25, .external_lex_state = 3},
  [6241] = {.lex_state = 16, .external_lex_state = 3},
  [6242] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6243] = {.lex_state = 16, .external_lex_state = 3},
  [6244] = {.lex_state = 16, .external_lex_state = 3},
  [6245] = {.lex_state = 64, .external_lex_state = 3},
  [6246] = {.lex_state = 64, .external_lex_state = 3},
  [6247] = {.lex_state = 120, .external_lex_state = 3},
  [6248] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6249] = {.lex_state = 64, .external_lex_state = 3},
  [6250] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6251] = {.lex_state = 120, .external_lex_state = 3},
  [6252] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6253] = {.lex_state = 16, .external_lex_state = 3},
  [6254] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6255] = {.lex_state = 123, .external_lex_state = 3},
  [6256] = {.lex_state = 120, .external_lex_state = 3},
  [6257] = {.lex_state = 7, .external_lex_state = 3},
  [6258] = {.lex_state = 120, .external_lex_state = 3},
  [6259] = {.lex_state = 7, .external_lex_state = 3},
  [6260] = {.lex_state = 120, .external_lex_state = 3},
  [6261] = {.lex_state = 16, .external_lex_state = 3},
  [6262] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6263] = {.lex_state = 16, .external_lex_state = 3},
  [6264] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6265] = {.lex_state = 7, .external_lex_state = 3},
  [6266] = {.lex_state = 121, .external_lex_state = 3},
  [6267] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6268] = {.lex_state = 7, .external_lex_state = 3},
  [6269] = {.lex_state = 53, .external_lex_state = 3},
  [6270] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6271] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6272] = {.lex_state = 25, .external_lex_state = 3},
  [6273] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6274] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6275] = {.lex_state = 120, .external_lex_state = 3},
  [6276] = {.lex_state = 7, .external_lex_state = 3},
  [6277] = {.lex_state = 120, .external_lex_state = 3},
  [6278] = {.lex_state = 16, .external_lex_state = 3},
  [6279] = {.lex_state = 120, .external_lex_state = 3},
  [6280] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6281] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6282] = {.lex_state = 120, .external_lex_state = 3},
  [6283] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6284] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6285] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6286] = {.lex_state = 7, .external_lex_state = 3},
  [6287] = {.lex_state = 120, .external_lex_state = 3},
  [6288] = {.lex_state = 120, .external_lex_state = 3},
  [6289] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 96},
  [6290] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6291] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6292] = {.lex_state = 120, .external_lex_state = 3},
  [6293] = {.lex_state = 120, .external_lex_state = 3},
  [6294] = {.lex_state = 120, .external_lex_state = 3},
  [6295] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6296] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 96},
  [6297] = {.lex_state = 16, .external_lex_state = 3},
  [6298] = {.lex_state = 121, .external_lex_state = 3},
  [6299] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6300] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6301] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6302] = {.lex_state = 7, .external_lex_state = 3},
  [6303] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6304] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6305] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6306] = {.lex_state = 26, .external_lex_state = 3},
  [6307] = {.lex_state = 26, .external_lex_state = 3},
  [6308] = {.lex_state = 26, .external_lex_state = 3},
  [6309] = {.lex_state = 120, .external_lex_state = 3},
  [6310] = {.lex_state = 121, .external_lex_state = 3},
  [6311] = {.lex_state = 26, .external_lex_state = 3},
  [6312] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6313] = {.lex_state = 26, .external_lex_state = 3},
  [6314] = {.lex_state = 26, .external_lex_state = 3},
  [6315] = {.lex_state = 120, .external_lex_state = 3},
  [6316] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6317] = {.lex_state = 26, .external_lex_state = 3},
  [6318] = {.lex_state = 26, .external_lex_state = 3},
  [6319] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6320] = {.lex_state = 26, .external_lex_state = 3},
  [6321] = {.lex_state = 26, .external_lex_state = 3},
  [6322] = {.lex_state = 26, .external_lex_state = 3},
  [6323] = {.lex_state = 26, .external_lex_state = 3},
  [6324] = {.lex_state = 7, .external_lex_state = 3},
  [6325] = {.lex_state = 120, .external_lex_state = 3},
  [6326] = {.lex_state = 43, .external_lex_state = 3},
  [6327] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6328] = {.lex_state = 26, .external_lex_state = 3},
  [6329] = {.lex_state = 43, .external_lex_state = 3},
  [6330] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6331] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6332] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6333] = {.lex_state = 7, .external_lex_state = 3},
  [6334] = {.lex_state = 26, .external_lex_state = 3},
  [6335] = {.lex_state = 120, .external_lex_state = 3},
  [6336] = {.lex_state = 26, .external_lex_state = 3},
  [6337] = {.lex_state = 74, .external_lex_state = 8},
  [6338] = {.lex_state = 7, .external_lex_state = 3},
  [6339] = {.lex_state = 74, .external_lex_state = 8},
  [6340] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6341] = {.lex_state = 26, .external_lex_state = 3},
  [6342] = {.lex_state = 26, .external_lex_state = 3},
  [6343] = {.lex_state = 43, .external_lex_state = 3},
  [6344] = {.lex_state = 120, .external_lex_state = 3},
  [6345] = {.lex_state = 40, .external_lex_state = 3},
  [6346] = {.lex_state = 120, .external_lex_state = 3},
  [6347] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6348] = {.lex_state = 16, .external_lex_state = 3},
  [6349] = {.lex_state = 120, .external_lex_state = 3},
  [6350] = {.lex_state = 26, .external_lex_state = 3},
  [6351] = {.lex_state = 26, .external_lex_state = 3},
  [6352] = {.lex_state = 120, .external_lex_state = 3},
  [6353] = {.lex_state = 120, .external_lex_state = 3},
  [6354] = {.lex_state = 74, .external_lex_state = 8},
  [6355] = {.lex_state = 120, .external_lex_state = 3},
  [6356] = {.lex_state = 16, .external_lex_state = 3},
  [6357] = {.lex_state = 32, .external_lex_state = 3},
  [6358] = {.lex_state = 26, .external_lex_state = 3},
  [6359] = {.lex_state = 43, .external_lex_state = 3},
  [6360] = {.lex_state = 120, .external_lex_state = 3},
  [6361] = {.lex_state = 26, .external_lex_state = 3},
  [6362] = {.lex_state = 64, .external_lex_state = 3},
  [6363] = {.lex_state = 74, .external_lex_state = 8},
  [6364] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 97},
  [6365] = {.lex_state = 26, .external_lex_state = 3},
  [6366] = {.lex_state = 120, .external_lex_state = 3},
  [6367] = {.lex_state = 120, .external_lex_state = 3},
  [6368] = {.lex_state = 120, .external_lex_state = 3},
  [6369] = {.lex_state = 26, .external_lex_state = 3},
  [6370] = {.lex_state = 26, .external_lex_state = 3},
  [6371] = {.lex_state = 26, .external_lex_state = 3},
  [6372] = {.lex_state = 64, .external_lex_state = 3},
  [6373] = {.lex_state = 120, .external_lex_state = 3},
  [6374] = {.lex_state = 32, .external_lex_state = 3},
  [6375] = {.lex_state = 120, .external_lex_state = 3},
  [6376] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6377] = {.lex_state = 32, .external_lex_state = 3},
  [6378] = {.lex_state = 16, .external_lex_state = 3},
  [6379] = {.lex_state = 120, .external_lex_state = 3},
  [6380] = {.lex_state = 26, .external_lex_state = 3},
  [6381] = {.lex_state = 64, .external_lex_state = 3},
  [6382] = {.lex_state = 26, .external_lex_state = 3},
  [6383] = {.lex_state = 120, .external_lex_state = 3},
  [6384] = {.lex_state = 120, .external_lex_state = 3},
  [6385] = {.lex_state = 120, .external_lex_state = 3},
  [6386] = {.lex_state = 120, .external_lex_state = 3},
  [6387] = {.lex_state = 120, .external_lex_state = 3},
  [6388] = {.lex_state = 32, .external_lex_state = 3},
  [6389] = {.lex_state = 16, .external_lex_state = 3},
  [6390] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6391] = {.lex_state = 120, .external_lex_state = 3},
  [6392] = {.lex_state = 120, .external_lex_state = 3},
  [6393] = {.lex_state = 26, .external_lex_state = 3},
  [6394] = {.lex_state = 120, .external_lex_state = 3},
  [6395] = {.lex_state = 120, .external_lex_state = 3},
  [6396] = {.lex_state = 26, .external_lex_state = 3},
  [6397] = {.lex_state = 7, .external_lex_state = 3},
  [6398] = {.lex_state = 120, .external_lex_state = 3},
  [6399] = {.lex_state = 121, .external_lex_state = 3},
  [6400] = {.lex_state = 7, .external_lex_state = 3},
  [6401] = {.lex_state = 26, .external_lex_state = 3},
  [6402] = {.lex_state = 7, .external_lex_state = 3},
  [6403] = {.lex_state = 120, .external_lex_state = 3},
  [6404] = {.lex_state = 64, .external_lex_state = 3},
  [6405] = {.lex_state = 121, .external_lex_state = 3},
  [6406] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6407] = {.lex_state = 7, .external_lex_state = 3},
  [6408] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 97},
  [6409] = {.lex_state = 26, .external_lex_state = 3},
  [6410] = {.lex_state = 121, .external_lex_state = 3},
  [6411] = {.lex_state = 120, .external_lex_state = 3},
  [6412] = {.lex_state = 26, .external_lex_state = 3},
  [6413] = {.lex_state = 26, .external_lex_state = 3},
  [6414] = {.lex_state = 40, .external_lex_state = 3},
  [6415] = {.lex_state = 7, .external_lex_state = 3},
  [6416] = {.lex_state = 121, .external_lex_state = 3},
  [6417] = {.lex_state = 26, .external_lex_state = 3},
  [6418] = {.lex_state = 64, .external_lex_state = 3},
  [6419] = {.lex_state = 120, .external_lex_state = 3},
  [6420] = {.lex_state = 26, .external_lex_state = 3},
  [6421] = {.lex_state = 16, .external_lex_state = 3},
  [6422] = {.lex_state = 64, .external_lex_state = 3},
  [6423] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6424] = {.lex_state = 64, .external_lex_state = 3},
  [6425] = {.lex_state = 120, .external_lex_state = 3},
  [6426] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6427] = {.lex_state = 26, .external_lex_state = 3},
  [6428] = {.lex_state = 120, .external_lex_state = 3},
  [6429] = {.lex_state = 121, .external_lex_state = 3},
  [6430] = {.lex_state = 16, .external_lex_state = 3},
  [6431] = {.lex_state = 64, .external_lex_state = 3},
  [6432] = {.lex_state = 16, .external_lex_state = 3},
  [6433] = {.lex_state = 64, .external_lex_state = 3},
  [6434] = {.lex_state = 40, .external_lex_state = 3},
  [6435] = {.lex_state = 40, .external_lex_state = 3},
  [6436] = {.lex_state = 120, .external_lex_state = 3},
  [6437] = {.lex_state = 64, .external_lex_state = 3},
  [6438] = {.lex_state = 64, .external_lex_state = 3},
  [6439] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6440] = {.lex_state = 26, .external_lex_state = 3},
  [6441] = {.lex_state = 7, .external_lex_state = 3},
  [6442] = {.lex_state = 26, .external_lex_state = 3},
  [6443] = {.lex_state = 16, .external_lex_state = 3},
  [6444] = {.lex_state = 64, .external_lex_state = 3},
  [6445] = {.lex_state = 120, .external_lex_state = 3},
  [6446] = {.lex_state = 26, .external_lex_state = 3},
  [6447] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6448] = {.lex_state = 26, .external_lex_state = 3},
  [6449] = {.lex_state = 26, .external_lex_state = 3},
  [6450] = {.lex_state = 26, .external_lex_state = 3},
  [6451] = {.lex_state = 26, .external_lex_state = 3},
  [6452] = {.lex_state = 26, .external_lex_state = 3},
  [6453] = {.lex_state = 64, .external_lex_state = 3},
  [6454] = {.lex_state = 26, .external_lex_state = 3},
  [6455] = {.lex_state = 26, .external_lex_state = 3},
  [6456] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6457] = {.lex_state = 121, .external_lex_state = 3},
  [6458] = {.lex_state = 7, .external_lex_state = 3},
  [6459] = {.lex_state = 121, .external_lex_state = 3},
  [6460] = {.lex_state = 74, .external_lex_state = 8},
  [6461] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6462] = {.lex_state = 120, .external_lex_state = 3},
  [6463] = {.lex_state = 26, .external_lex_state = 3},
  [6464] = {.lex_state = 26, .external_lex_state = 3},
  [6465] = {.lex_state = 120, .external_lex_state = 3},
  [6466] = {.lex_state = 26, .external_lex_state = 3},
  [6467] = {.lex_state = 123, .external_lex_state = 3},
  [6468] = {.lex_state = 16, .external_lex_state = 3},
  [6469] = {.lex_state = 120, .external_lex_state = 3},
  [6470] = {.lex_state = 26, .external_lex_state = 3},
  [6471] = {.lex_state = 120, .external_lex_state = 3},
  [6472] = {.lex_state = 120, .external_lex_state = 3},
  [6473] = {.lex_state = 121, .external_lex_state = 3},
  [6474] = {.lex_state = 26, .external_lex_state = 3},
  [6475] = {.lex_state = 120, .external_lex_state = 3},
  [6476] = {.lex_state = 26, .external_lex_state = 3},
  [6477] = {.lex_state = 120, .external_lex_state = 3},
  [6478] = {.lex_state = 120, .external_lex_state = 3},
  [6479] = {.lex_state = 120, .external_lex_state = 3},
  [6480] = {.lex_state = 120, .external_lex_state = 3},
  [6481] = {.lex_state = 123, .external_lex_state = 3},
  [6482] = {.lex_state = 26, .external_lex_state = 3},
  [6483] = {.lex_state = 120, .external_lex_state = 3},
  [6484] = {.lex_state = 120, .external_lex_state = 3},
  [6485] = {.lex_state = 26, .external_lex_state = 3},
  [6486] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6487] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6488] = {.lex_state = 26, .external_lex_state = 3},
  [6489] = {.lex_state = 120, .external_lex_state = 3},
  [6490] = {.lex_state = 64, .external_lex_state = 3},
  [6491] = {.lex_state = 26, .external_lex_state = 3},
  [6492] = {.lex_state = 74, .external_lex_state = 8},
  [6493] = {.lex_state = 120, .external_lex_state = 3},
  [6494] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6495] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6496] = {.lex_state = 121, .external_lex_state = 3},
  [6497] = {.lex_state = 26, .external_lex_state = 3},
  [6498] = {.lex_state = 74, .external_lex_state = 8},
  [6499] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6500] = {.lex_state = 120, .external_lex_state = 3},
  [6501] = {.lex_state = 120, .external_lex_state = 3},
  [6502] = {.lex_state = 64, .external_lex_state = 3},
  [6503] = {.lex_state = 120, .external_lex_state = 3},
  [6504] = {.lex_state = 120, .external_lex_state = 3},
  [6505] = {.lex_state = 26, .external_lex_state = 3},
  [6506] = {.lex_state = 25, .external_lex_state = 3},
  [6507] = {.lex_state = 26, .external_lex_state = 3},
  [6508] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6509] = {.lex_state = 74, .external_lex_state = 8},
  [6510] = {.lex_state = 120, .external_lex_state = 3},
  [6511] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6512] = {.lex_state = 26, .external_lex_state = 3},
  [6513] = {.lex_state = 40, .external_lex_state = 3},
  [6514] = {.lex_state = 121, .external_lex_state = 3},
  [6515] = {.lex_state = 120, .external_lex_state = 3},
  [6516] = {.lex_state = 26, .external_lex_state = 3},
  [6517] = {.lex_state = 40, .external_lex_state = 3},
  [6518] = {.lex_state = 40, .external_lex_state = 3},
  [6519] = {.lex_state = 120, .external_lex_state = 3},
  [6520] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6521] = {.lex_state = 120, .external_lex_state = 3},
  [6522] = {.lex_state = 123, .external_lex_state = 3},
  [6523] = {.lex_state = 74, .external_lex_state = 8},
  [6524] = {.lex_state = 26, .external_lex_state = 3},
  [6525] = {.lex_state = 26, .external_lex_state = 3},
  [6526] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6527] = {.lex_state = 26, .external_lex_state = 3},
  [6528] = {.lex_state = 74, .external_lex_state = 8},
  [6529] = {.lex_state = 26, .external_lex_state = 3},
  [6530] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6531] = {.lex_state = 120, .external_lex_state = 3},
  [6532] = {.lex_state = 64, .external_lex_state = 3},
  [6533] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6534] = {.lex_state = 7, .external_lex_state = 3},
  [6535] = {.lex_state = 74, .external_lex_state = 8},
  [6536] = {.lex_state = 26, .external_lex_state = 3},
  [6537] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6538] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6539] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6540] = {.lex_state = 74, .external_lex_state = 8},
  [6541] = {.lex_state = 26, .external_lex_state = 3},
  [6542] = {.lex_state = 26, .external_lex_state = 3},
  [6543] = {.lex_state = 26, .external_lex_state = 3},
  [6544] = {.lex_state = 120, .external_lex_state = 3},
  [6545] = {.lex_state = 120, .external_lex_state = 3},
  [6546] = {.lex_state = 7, .external_lex_state = 3},
  [6547] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6548] = {.lex_state = 74, .external_lex_state = 8},
  [6549] = {.lex_state = 64, .external_lex_state = 3},
  [6550] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6551] = {.lex_state = 123, .external_lex_state = 3},
  [6552] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6553] = {.lex_state = 26, .external_lex_state = 3},
  [6554] = {.lex_state = 74, .external_lex_state = 8},
  [6555] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6556] = {.lex_state = 7, .external_lex_state = 3},
  [6557] = {.lex_state = 16, .external_lex_state = 3},
  [6558] = {.lex_state = 121, .external_lex_state = 3},
  [6559] = {.lex_state = 64, .external_lex_state = 3},
  [6560] = {.lex_state = 40, .external_lex_state = 3},
  [6561] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6562] = {.lex_state = 74, .external_lex_state = 8},
  [6563] = {.lex_state = 120, .external_lex_state = 3},
  [6564] = {.lex_state = 26, .external_lex_state = 3},
  [6565] = {.lex_state = 7, .external_lex_state = 3},
  [6566] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6567] = {.lex_state = 74, .external_lex_state = 8},
  [6568] = {.lex_state = 26, .external_lex_state = 3},
  [6569] = {.lex_state = 64, .external_lex_state = 3},
  [6570] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6571] = {.lex_state = 26, .external_lex_state = 3},
  [6572] = {.lex_state = 74, .external_lex_state = 8},
  [6573] = {.lex_state = 120, .external_lex_state = 3},
  [6574] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6575] = {.lex_state = 7, .external_lex_state = 3},
  [6576] = {.lex_state = 74, .external_lex_state = 8},
  [6577] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6578] = {.lex_state = 7, .external_lex_state = 3},
  [6579] = {.lex_state = 74, .external_lex_state = 8},
  [6580] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6581] = {.lex_state = 74, .external_lex_state = 8},
  [6582] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6583] = {.lex_state = 74, .external_lex_state = 8},
  [6584] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6585] = {.lex_state = 26, .external_lex_state = 3},
  [6586] = {.lex_state = 74, .external_lex_state = 8},
  [6587] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6588] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6589] = {.lex_state = 74, .external_lex_state = 8},
  [6590] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6591] = {.lex_state = 74, .external_lex_state = 8},
  [6592] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6593] = {.lex_state = 74, .external_lex_state = 8},
  [6594] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6595] = {.lex_state = 74, .external_lex_state = 8},
  [6596] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6597] = {.lex_state = 74, .external_lex_state = 8},
  [6598] = {.lex_state = 120, .external_lex_state = 3},
  [6599] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6600] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6601] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6602] = {.lex_state = 120, .external_lex_state = 3},
  [6603] = {.lex_state = 7, .external_lex_state = 3},
  [6604] = {.lex_state = 26, .external_lex_state = 3},
  [6605] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6606] = {.lex_state = 121, .external_lex_state = 3},
  [6607] = {.lex_state = 121, .external_lex_state = 3},
  [6608] = {.lex_state = 74, .external_lex_state = 8},
  [6609] = {.lex_state = 121, .external_lex_state = 3},
  [6610] = {.lex_state = 26, .external_lex_state = 3},
  [6611] = {.lex_state = 123, .external_lex_state = 3},
  [6612] = {.lex_state = 32, .external_lex_state = 3},
  [6613] = {.lex_state = 120, .external_lex_state = 3},
  [6614] = {.lex_state = 7, .external_lex_state = 3},
  [6615] = {.lex_state = 123, .external_lex_state = 3},
  [6616] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6617] = {.lex_state = 26, .external_lex_state = 3},
  [6618] = {.lex_state = 7, .external_lex_state = 3},
  [6619] = {.lex_state = 32, .external_lex_state = 3},
  [6620] = {.lex_state = 121, .external_lex_state = 3},
  [6621] = {.lex_state = 26, .external_lex_state = 3},
  [6622] = {.lex_state = 64, .external_lex_state = 3},
  [6623] = {.lex_state = 64, .external_lex_state = 3},
  [6624] = {.lex_state = 121, .external_lex_state = 3},
  [6625] = {.lex_state = 7, .external_lex_state = 3},
  [6626] = {.lex_state = 64, .external_lex_state = 3},
  [6627] = {.lex_state = 64, .external_lex_state = 3},
  [6628] = {.lex_state = 120, .external_lex_state = 3},
  [6629] = {.lex_state = 120, .external_lex_state = 3},
  [6630] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6631] = {.lex_state = 26, .external_lex_state = 3},
  [6632] = {.lex_state = 64, .external_lex_state = 3},
  [6633] = {.lex_state = 64, .external_lex_state = 3},
  [6634] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6635] = {.lex_state = 26, .external_lex_state = 3},
  [6636] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6637] = {.lex_state = 26, .external_lex_state = 3},
  [6638] = {.lex_state = 120, .external_lex_state = 3},
  [6639] = {.lex_state = 26, .external_lex_state = 3},
  [6640] = {.lex_state = 26, .external_lex_state = 3},
  [6641] = {.lex_state = 26, .external_lex_state = 3},
  [6642] = {.lex_state = 26, .external_lex_state = 3},
  [6643] = {.lex_state = 120, .external_lex_state = 3},
  [6644] = {.lex_state = 120, .external_lex_state = 3},
  [6645] = {.lex_state = 26, .external_lex_state = 3},
  [6646] = {.lex_state = 26, .external_lex_state = 3},
  [6647] = {.lex_state = 64, .external_lex_state = 3},
  [6648] = {.lex_state = 26, .external_lex_state = 3},
  [6649] = {.lex_state = 26, .external_lex_state = 3},
  [6650] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6651] = {.lex_state = 120, .external_lex_state = 3},
  [6652] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6653] = {.lex_state = 26, .external_lex_state = 3},
  [6654] = {.lex_state = 26, .external_lex_state = 3},
  [6655] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6656] = {.lex_state = 123, .external_lex_state = 3},
  [6657] = {.lex_state = 123, .external_lex_state = 3},
  [6658] = {.lex_state = 40, .external_lex_state = 3},
  [6659] = {.lex_state = 7, .external_lex_state = 3},
  [6660] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6661] = {.lex_state = 26, .external_lex_state = 3},
  [6662] = {.lex_state = 123, .external_lex_state = 3},
  [6663] = {.lex_state = 123, .external_lex_state = 3},
  [6664] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6665] = {.lex_state = 64, .external_lex_state = 3},
  [6666] = {.lex_state = 26, .external_lex_state = 3},
  [6667] = {.lex_state = 123, .external_lex_state = 3},
  [6668] = {.lex_state = 26, .external_lex_state = 3},
  [6669] = {.lex_state = 123, .external_lex_state = 3},
  [6670] = {.lex_state = 120, .external_lex_state = 3},
  [6671] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6672] = {.lex_state = 123, .external_lex_state = 3},
  [6673] = {.lex_state = 26, .external_lex_state = 3},
  [6674] = {.lex_state = 123, .external_lex_state = 3},
  [6675] = {.lex_state = 16, .external_lex_state = 3},
  [6676] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6677] = {.lex_state = 123, .external_lex_state = 3},
  [6678] = {.lex_state = 123, .external_lex_state = 3},
  [6679] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6680] = {.lex_state = 26, .external_lex_state = 3},
  [6681] = {.lex_state = 64, .external_lex_state = 3},
  [6682] = {.lex_state = 123, .external_lex_state = 3},
  [6683] = {.lex_state = 123, .external_lex_state = 3},
  [6684] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6685] = {.lex_state = 123, .external_lex_state = 3},
  [6686] = {.lex_state = 123, .external_lex_state = 3},
  [6687] = {.lex_state = 26, .external_lex_state = 3},
  [6688] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6689] = {.lex_state = 64, .external_lex_state = 3},
  [6690] = {.lex_state = 7, .external_lex_state = 3},
  [6691] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6692] = {.lex_state = 123, .external_lex_state = 3},
  [6693] = {.lex_state = 123, .external_lex_state = 3},
  [6694] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6695] = {.lex_state = 64, .external_lex_state = 3},
  [6696] = {.lex_state = 123, .external_lex_state = 3},
  [6697] = {.lex_state = 123, .external_lex_state = 3},
  [6698] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6699] = {.lex_state = 26, .external_lex_state = 3},
  [6700] = {.lex_state = 123, .external_lex_state = 3},
  [6701] = {.lex_state = 123, .external_lex_state = 3},
  [6702] = {.lex_state = 26, .external_lex_state = 3},
  [6703] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6704] = {.lex_state = 123, .external_lex_state = 3},
  [6705] = {.lex_state = 7, .external_lex_state = 3},
  [6706] = {.lex_state = 123, .external_lex_state = 3},
  [6707] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6708] = {.lex_state = 123, .external_lex_state = 3},
  [6709] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6710] = {.lex_state = 123, .external_lex_state = 3},
  [6711] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6712] = {.lex_state = 123, .external_lex_state = 3},
  [6713] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6714] = {.lex_state = 123, .external_lex_state = 3},
  [6715] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6716] = {.lex_state = 123, .external_lex_state = 3},
  [6717] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6718] = {.lex_state = 123, .external_lex_state = 3},
  [6719] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6720] = {.lex_state = 123, .external_lex_state = 3},
  [6721] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6722] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6723] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6724] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6725] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6726] = {.lex_state = 26, .external_lex_state = 3},
  [6727] = {.lex_state = 120, .external_lex_state = 3},
  [6728] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6729] = {.lex_state = 120, .external_lex_state = 3},
  [6730] = {.lex_state = 120, .external_lex_state = 3},
  [6731] = {.lex_state = 120, .external_lex_state = 3},
  [6732] = {.lex_state = 120, .external_lex_state = 3},
  [6733] = {.lex_state = 26, .external_lex_state = 3},
  [6734] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6735] = {.lex_state = 120, .external_lex_state = 3},
  [6736] = {.lex_state = 120, .external_lex_state = 3},
  [6737] = {.lex_state = 120, .external_lex_state = 3},
  [6738] = {.lex_state = 120, .external_lex_state = 3},
  [6739] = {.lex_state = 120, .external_lex_state = 3},
  [6740] = {.lex_state = 26, .external_lex_state = 3},
  [6741] = {.lex_state = 26, .external_lex_state = 3},
  [6742] = {.lex_state = 120, .external_lex_state = 3},
  [6743] = {.lex_state = 7, .external_lex_state = 3},
  [6744] = {.lex_state = 26, .external_lex_state = 3},
  [6745] = {.lex_state = 26, .external_lex_state = 3},
  [6746] = {.lex_state = 120, .external_lex_state = 3},
  [6747] = {.lex_state = 26, .external_lex_state = 3},
  [6748] = {.lex_state = 64, .external_lex_state = 3},
  [6749] = {.lex_state = 121, .external_lex_state = 3},
  [6750] = {.lex_state = 121, .external_lex_state = 3},
  [6751] = {.lex_state = 7, .external_lex_state = 3},
  [6752] = {.lex_state = 26, .external_lex_state = 3},
  [6753] = {.lex_state = 26, .external_lex_state = 3},
  [6754] = {.lex_state = 16, .external_lex_state = 3},
  [6755] = {.lex_state = 16, .external_lex_state = 3},
  [6756] = {.lex_state = 16, .external_lex_state = 3},
  [6757] = {.lex_state = 16, .external_lex_state = 3},
  [6758] = {.lex_state = 26, .external_lex_state = 3},
  [6759] = {.lex_state = 121, .external_lex_state = 3},
  [6760] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6761] = {.lex_state = 120, .external_lex_state = 3},
  [6762] = {.lex_state = 26, .external_lex_state = 3},
  [6763] = {.lex_state = 26, .external_lex_state = 3},
  [6764] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6765] = {.lex_state = 120, .external_lex_state = 3},
  [6766] = {.lex_state = 120, .external_lex_state = 3},
  [6767] = {.lex_state = 121, .external_lex_state = 3},
  [6768] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6769] = {.lex_state = 120, .external_lex_state = 3},
  [6770] = {.lex_state = 8, .external_lex_state = 3},
  [6771] = {.lex_state = 16, .external_lex_state = 3},
  [6772] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6773] = {.lex_state = 16, .external_lex_state = 3},
  [6774] = {.lex_state = 120, .external_lex_state = 8},
  [6775] = {.lex_state = 16, .external_lex_state = 3},
  [6776] = {.lex_state = 16, .external_lex_state = 3},
  [6777] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6778] = {.lex_state = 120, .external_lex_state = 3},
  [6779] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6780] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6781] = {.lex_state = 40, .external_lex_state = 3},
  [6782] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6783] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6784] = {.lex_state = 16, .external_lex_state = 3},
  [6785] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6786] = {.lex_state = 7, .external_lex_state = 3},
  [6787] = {.lex_state = 120, .external_lex_state = 3},
  [6788] = {.lex_state = 120, .external_lex_state = 3},
  [6789] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6790] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6791] = {.lex_state = 16, .external_lex_state = 3},
  [6792] = {.lex_state = 120, .external_lex_state = 3},
  [6793] = {.lex_state = 26, .external_lex_state = 3},
  [6794] = {.lex_state = 120, .external_lex_state = 3},
  [6795] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6796] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6797] = {.lex_state = 120, .external_lex_state = 3},
  [6798] = {.lex_state = 120, .external_lex_state = 3},
  [6799] = {.lex_state = 16, .external_lex_state = 3},
  [6800] = {.lex_state = 120, .external_lex_state = 3},
  [6801] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6802] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6803] = {.lex_state = 16, .external_lex_state = 3},
  [6804] = {.lex_state = 120, .external_lex_state = 3},
  [6805] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6806] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6807] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6808] = {.lex_state = 120, .external_lex_state = 3},
  [6809] = {.lex_state = 16, .external_lex_state = 3},
  [6810] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6811] = {.lex_state = 120, .external_lex_state = 3},
  [6812] = {.lex_state = 8, .external_lex_state = 3},
  [6813] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6814] = {.lex_state = 120, .external_lex_state = 3},
  [6815] = {.lex_state = 16, .external_lex_state = 3},
  [6816] = {.lex_state = 120, .external_lex_state = 3},
  [6817] = {.lex_state = 120, .external_lex_state = 3},
  [6818] = {.lex_state = 120, .external_lex_state = 3},
  [6819] = {.lex_state = 16, .external_lex_state = 3},
  [6820] = {.lex_state = 16, .external_lex_state = 3},
  [6821] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6822] = {.lex_state = 120, .external_lex_state = 3},
  [6823] = {.lex_state = 16, .external_lex_state = 3},
  [6824] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6825] = {.lex_state = 16, .external_lex_state = 3},
  [6826] = {.lex_state = 16, .external_lex_state = 3},
  [6827] = {.lex_state = 16, .external_lex_state = 3},
  [6828] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6829] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6830] = {.lex_state = 16, .external_lex_state = 3},
  [6831] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6832] = {.lex_state = 26, .external_lex_state = 3},
  [6833] = {.lex_state = 26, .external_lex_state = 3},
  [6834] = {.lex_state = 16, .external_lex_state = 3},
  [6835] = {.lex_state = 120, .external_lex_state = 8},
  [6836] = {.lex_state = 16, .external_lex_state = 3},
  [6837] = {.lex_state = 16, .external_lex_state = 3},
  [6838] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [6839] = {.lex_state = 120, .external_lex_state = 3},
  [6840] = {.lex_state = 26, .external_lex_state = 3},
  [6841] = {.lex_state = 120, .external_lex_state = 8},
  [6842] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6843] = {.lex_state = 120, .external_lex_state = 3},
  [6844] = {.lex_state = 26, .external_lex_state = 3},
  [6845] = {.lex_state = 120, .external_lex_state = 3},
  [6846] = {.lex_state = 120, .external_lex_state = 3},
  [6847] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6848] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6849] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6850] = {.lex_state = 120, .external_lex_state = 8},
  [6851] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6852] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6853] = {.lex_state = 120, .external_lex_state = 3},
  [6854] = {.lex_state = 16, .external_lex_state = 3},
  [6855] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6856] = {.lex_state = 26, .external_lex_state = 3},
  [6857] = {.lex_state = 26, .external_lex_state = 3},
  [6858] = {.lex_state = 26, .external_lex_state = 3},
  [6859] = {.lex_state = 120, .external_lex_state = 3},
  [6860] = {.lex_state = 120, .external_lex_state = 3},
  [6861] = {.lex_state = 26, .external_lex_state = 3},
  [6862] = {.lex_state = 120, .external_lex_state = 3},
  [6863] = {.lex_state = 26, .external_lex_state = 3},
  [6864] = {.lex_state = 120, .external_lex_state = 3},
  [6865] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6866] = {.lex_state = 26, .external_lex_state = 3},
  [6867] = {.lex_state = 16, .external_lex_state = 3},
  [6868] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6869] = {.lex_state = 120, .external_lex_state = 3},
  [6870] = {.lex_state = 16, .external_lex_state = 3},
  [6871] = {.lex_state = 7, .external_lex_state = 3},
  [6872] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6873] = {.lex_state = 16, .external_lex_state = 3},
  [6874] = {.lex_state = 120, .external_lex_state = 3},
  [6875] = {.lex_state = 16, .external_lex_state = 3},
  [6876] = {.lex_state = 120, .external_lex_state = 3},
  [6877] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6878] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6879] = {.lex_state = 120, .external_lex_state = 3},
  [6880] = {.lex_state = 120, .external_lex_state = 3},
  [6881] = {.lex_state = 120, .external_lex_state = 3},
  [6882] = {.lex_state = 120, .external_lex_state = 3},
  [6883] = {.lex_state = 120, .external_lex_state = 3},
  [6884] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6885] = {.lex_state = 120, .external_lex_state = 8},
  [6886] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6887] = {.lex_state = 26, .external_lex_state = 3},
  [6888] = {.lex_state = 16, .external_lex_state = 3},
  [6889] = {.lex_state = 120, .external_lex_state = 3},
  [6890] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6891] = {.lex_state = 120, .external_lex_state = 8},
  [6892] = {.lex_state = 120, .external_lex_state = 3},
  [6893] = {.lex_state = 16, .external_lex_state = 3},
  [6894] = {.lex_state = 26, .external_lex_state = 3},
  [6895] = {.lex_state = 26, .external_lex_state = 3},
  [6896] = {.lex_state = 26, .external_lex_state = 3},
  [6897] = {.lex_state = 120, .external_lex_state = 3},
  [6898] = {.lex_state = 7, .external_lex_state = 3},
  [6899] = {.lex_state = 16, .external_lex_state = 3},
  [6900] = {.lex_state = 16, .external_lex_state = 3},
  [6901] = {.lex_state = 120, .external_lex_state = 3},
  [6902] = {.lex_state = 120, .external_lex_state = 3},
  [6903] = {.lex_state = 120, .external_lex_state = 8},
  [6904] = {.lex_state = 16, .external_lex_state = 3},
  [6905] = {.lex_state = 26, .external_lex_state = 3},
  [6906] = {.lex_state = 120, .external_lex_state = 8},
  [6907] = {.lex_state = 121, .external_lex_state = 3},
  [6908] = {.lex_state = 16, .external_lex_state = 3},
  [6909] = {.lex_state = 26, .external_lex_state = 3},
  [6910] = {.lex_state = 26, .external_lex_state = 3},
  [6911] = {.lex_state = 26, .external_lex_state = 3},
  [6912] = {.lex_state = 120, .external_lex_state = 3},
  [6913] = {.lex_state = 120, .external_lex_state = 3},
  [6914] = {.lex_state = 8, .external_lex_state = 3},
  [6915] = {.lex_state = 120, .external_lex_state = 3},
  [6916] = {.lex_state = 120, .external_lex_state = 8},
  [6917] = {.lex_state = 16, .external_lex_state = 3},
  [6918] = {.lex_state = 120, .external_lex_state = 8},
  [6919] = {.lex_state = 16, .external_lex_state = 3},
  [6920] = {.lex_state = 26, .external_lex_state = 3},
  [6921] = {.lex_state = 26, .external_lex_state = 3},
  [6922] = {.lex_state = 26, .external_lex_state = 3},
  [6923] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6924] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6925] = {.lex_state = 120, .external_lex_state = 3},
  [6926] = {.lex_state = 120, .external_lex_state = 3},
  [6927] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6928] = {.lex_state = 120, .external_lex_state = 8},
  [6929] = {.lex_state = 26, .external_lex_state = 3},
  [6930] = {.lex_state = 26, .external_lex_state = 3},
  [6931] = {.lex_state = 26, .external_lex_state = 3},
  [6932] = {.lex_state = 120, .external_lex_state = 3},
  [6933] = {.lex_state = 120, .external_lex_state = 8},
  [6934] = {.lex_state = 26, .external_lex_state = 3},
  [6935] = {.lex_state = 26, .external_lex_state = 3},
  [6936] = {.lex_state = 26, .external_lex_state = 3},
  [6937] = {.lex_state = 26, .external_lex_state = 3},
  [6938] = {.lex_state = 120, .external_lex_state = 3},
  [6939] = {.lex_state = 120, .external_lex_state = 3},
  [6940] = {.lex_state = 120, .external_lex_state = 3},
  [6941] = {.lex_state = 40, .external_lex_state = 3},
  [6942] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6943] = {.lex_state = 120, .external_lex_state = 8},
  [6944] = {.lex_state = 26, .external_lex_state = 3},
  [6945] = {.lex_state = 26, .external_lex_state = 3},
  [6946] = {.lex_state = 26, .external_lex_state = 3},
  [6947] = {.lex_state = 64, .external_lex_state = 3},
  [6948] = {.lex_state = 120, .external_lex_state = 3},
  [6949] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6950] = {.lex_state = 120, .external_lex_state = 8},
  [6951] = {.lex_state = 120, .external_lex_state = 3},
  [6952] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6953] = {.lex_state = 120, .external_lex_state = 3},
  [6954] = {.lex_state = 26, .external_lex_state = 3},
  [6955] = {.lex_state = 26, .external_lex_state = 3},
  [6956] = {.lex_state = 26, .external_lex_state = 3},
  [6957] = {.lex_state = 16, .external_lex_state = 3},
  [6958] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6959] = {.lex_state = 120, .external_lex_state = 8},
  [6960] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6961] = {.lex_state = 26, .external_lex_state = 3},
  [6962] = {.lex_state = 26, .external_lex_state = 3},
  [6963] = {.lex_state = 26, .external_lex_state = 3},
  [6964] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6965] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6966] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6967] = {.lex_state = 120, .external_lex_state = 8},
  [6968] = {.lex_state = 120, .external_lex_state = 3},
  [6969] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6970] = {.lex_state = 26, .external_lex_state = 3},
  [6971] = {.lex_state = 26, .external_lex_state = 3},
  [6972] = {.lex_state = 26, .external_lex_state = 3},
  [6973] = {.lex_state = 120, .external_lex_state = 8},
  [6974] = {.lex_state = 26, .external_lex_state = 3},
  [6975] = {.lex_state = 16, .external_lex_state = 3},
  [6976] = {.lex_state = 120, .external_lex_state = 8},
  [6977] = {.lex_state = 26, .external_lex_state = 3},
  [6978] = {.lex_state = 26, .external_lex_state = 3},
  [6979] = {.lex_state = 26, .external_lex_state = 3},
  [6980] = {.lex_state = 26, .external_lex_state = 3},
  [6981] = {.lex_state = 120, .external_lex_state = 3},
  [6982] = {.lex_state = 120, .external_lex_state = 8},
  [6983] = {.lex_state = 26, .external_lex_state = 3},
  [6984] = {.lex_state = 26, .external_lex_state = 3},
  [6985] = {.lex_state = 26, .external_lex_state = 3},
  [6986] = {.lex_state = 120, .external_lex_state = 3},
  [6987] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [6988] = {.lex_state = 120, .external_lex_state = 8},
  [6989] = {.lex_state = 26, .external_lex_state = 3},
  [6990] = {.lex_state = 26, .external_lex_state = 3},
  [6991] = {.lex_state = 26, .external_lex_state = 3},
  [6992] = {.lex_state = 120, .external_lex_state = 3},
  [6993] = {.lex_state = 26, .external_lex_state = 3},
  [6994] = {.lex_state = 26, .external_lex_state = 3},
  [6995] = {.lex_state = 26, .external_lex_state = 3},
  [6996] = {.lex_state = 120, .external_lex_state = 8},
  [6997] = {.lex_state = 26, .external_lex_state = 3},
  [6998] = {.lex_state = 26, .external_lex_state = 3},
  [6999] = {.lex_state = 26, .external_lex_state = 3},
  [7000] = {.lex_state = 26, .external_lex_state = 3},
  [7001] = {.lex_state = 120, .external_lex_state = 3},
  [7002] = {.lex_state = 120, .external_lex_state = 8},
  [7003] = {.lex_state = 26, .external_lex_state = 3},
  [7004] = {.lex_state = 26, .external_lex_state = 3},
  [7005] = {.lex_state = 26, .external_lex_state = 3},
  [7006] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7007] = {.lex_state = 120, .external_lex_state = 3},
  [7008] = {.lex_state = 120, .external_lex_state = 8},
  [7009] = {.lex_state = 26, .external_lex_state = 3},
  [7010] = {.lex_state = 26, .external_lex_state = 3},
  [7011] = {.lex_state = 26, .external_lex_state = 3},
  [7012] = {.lex_state = 120, .external_lex_state = 3},
  [7013] = {.lex_state = 120, .external_lex_state = 8},
  [7014] = {.lex_state = 26, .external_lex_state = 3},
  [7015] = {.lex_state = 26, .external_lex_state = 3},
  [7016] = {.lex_state = 26, .external_lex_state = 3},
  [7017] = {.lex_state = 120, .external_lex_state = 3},
  [7018] = {.lex_state = 120, .external_lex_state = 8},
  [7019] = {.lex_state = 120, .external_lex_state = 8},
  [7020] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7021] = {.lex_state = 120, .external_lex_state = 8},
  [7022] = {.lex_state = 120, .external_lex_state = 3},
  [7023] = {.lex_state = 120, .external_lex_state = 8},
  [7024] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7025] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7026] = {.lex_state = 120, .external_lex_state = 3},
  [7027] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7028] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7029] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7030] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7031] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7032] = {.lex_state = 16, .external_lex_state = 3},
  [7033] = {.lex_state = 120, .external_lex_state = 3},
  [7034] = {.lex_state = 26, .external_lex_state = 3},
  [7035] = {.lex_state = 120, .external_lex_state = 3},
  [7036] = {.lex_state = 26, .external_lex_state = 3},
  [7037] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7038] = {.lex_state = 16, .external_lex_state = 3},
  [7039] = {.lex_state = 120, .external_lex_state = 3},
  [7040] = {.lex_state = 120, .external_lex_state = 3},
  [7041] = {.lex_state = 120, .external_lex_state = 8},
  [7042] = {.lex_state = 120, .external_lex_state = 8},
  [7043] = {.lex_state = 120, .external_lex_state = 8},
  [7044] = {.lex_state = 120, .external_lex_state = 3},
  [7045] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7046] = {.lex_state = 64, .external_lex_state = 3},
  [7047] = {.lex_state = 16, .external_lex_state = 3},
  [7048] = {.lex_state = 16, .external_lex_state = 3},
  [7049] = {.lex_state = 120, .external_lex_state = 3},
  [7050] = {.lex_state = 16, .external_lex_state = 3},
  [7051] = {.lex_state = 120, .external_lex_state = 3},
  [7052] = {.lex_state = 16, .external_lex_state = 3},
  [7053] = {.lex_state = 120, .external_lex_state = 3},
  [7054] = {.lex_state = 16, .external_lex_state = 3},
  [7055] = {.lex_state = 120, .external_lex_state = 3},
  [7056] = {.lex_state = 16, .external_lex_state = 3},
  [7057] = {.lex_state = 120, .external_lex_state = 3},
  [7058] = {.lex_state = 16, .external_lex_state = 3},
  [7059] = {.lex_state = 120, .external_lex_state = 3},
  [7060] = {.lex_state = 16, .external_lex_state = 3},
  [7061] = {.lex_state = 120, .external_lex_state = 3},
  [7062] = {.lex_state = 16, .external_lex_state = 3},
  [7063] = {.lex_state = 120, .external_lex_state = 3},
  [7064] = {.lex_state = 16, .external_lex_state = 3},
  [7065] = {.lex_state = 16, .external_lex_state = 3},
  [7066] = {.lex_state = 7, .external_lex_state = 3},
  [7067] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7068] = {.lex_state = 120, .external_lex_state = 8},
  [7069] = {.lex_state = 64, .external_lex_state = 3},
  [7070] = {.lex_state = 120, .external_lex_state = 3},
  [7071] = {.lex_state = 120, .external_lex_state = 3},
  [7072] = {.lex_state = 16, .external_lex_state = 3},
  [7073] = {.lex_state = 127, .external_lex_state = 3},
  [7074] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7075] = {.lex_state = 26, .external_lex_state = 3},
  [7076] = {.lex_state = 26, .external_lex_state = 3},
  [7077] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7078] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [7079] = {.lex_state = 120, .external_lex_state = 3},
  [7080] = {.lex_state = 16, .external_lex_state = 3},
  [7081] = {.lex_state = 26, .external_lex_state = 3},
  [7082] = {.lex_state = 26, .external_lex_state = 3},
  [7083] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7084] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [7085] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7086] = {.lex_state = 120, .external_lex_state = 3},
  [7087] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7088] = {.lex_state = 26, .external_lex_state = 3},
  [7089] = {.lex_state = 26, .external_lex_state = 3},
  [7090] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7091] = {.lex_state = 16, .external_lex_state = 3},
  [7092] = {.lex_state = 16, .external_lex_state = 3},
  [7093] = {.lex_state = 7, .external_lex_state = 3},
  [7094] = {.lex_state = 16, .external_lex_state = 3},
  [7095] = {.lex_state = 16, .external_lex_state = 3},
  [7096] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7097] = {.lex_state = 120, .external_lex_state = 3},
  [7098] = {.lex_state = 16, .external_lex_state = 3},
  [7099] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7100] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7101] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7102] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7103] = {.lex_state = 120, .external_lex_state = 3},
  [7104] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7105] = {.lex_state = 120, .external_lex_state = 3},
  [7106] = {.lex_state = 120, .external_lex_state = 3},
  [7107] = {.lex_state = 26, .external_lex_state = 3},
  [7108] = {.lex_state = 16, .external_lex_state = 3},
  [7109] = {.lex_state = 120, .external_lex_state = 3},
  [7110] = {.lex_state = 16, .external_lex_state = 3},
  [7111] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [7112] = {.lex_state = 16, .external_lex_state = 3},
  [7113] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7114] = {.lex_state = 120, .external_lex_state = 3},
  [7115] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7116] = {.lex_state = 26, .external_lex_state = 3},
  [7117] = {.lex_state = 123, .external_lex_state = 3},
  [7118] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7119] = {.lex_state = 120, .external_lex_state = 3},
  [7120] = {.lex_state = 16, .external_lex_state = 3},
  [7121] = {.lex_state = 16, .external_lex_state = 3},
  [7122] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7123] = {.lex_state = 127, .external_lex_state = 3},
  [7124] = {.lex_state = 16, .external_lex_state = 3},
  [7125] = {.lex_state = 120, .external_lex_state = 3},
  [7126] = {.lex_state = 16, .external_lex_state = 3},
  [7127] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7128] = {.lex_state = 16, .external_lex_state = 3},
  [7129] = {.lex_state = 120, .external_lex_state = 3},
  [7130] = {.lex_state = 26, .external_lex_state = 3},
  [7131] = {.lex_state = 16, .external_lex_state = 3},
  [7132] = {.lex_state = 120, .external_lex_state = 3},
  [7133] = {.lex_state = 16, .external_lex_state = 3},
  [7134] = {.lex_state = 120, .external_lex_state = 3},
  [7135] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7136] = {.lex_state = 8, .external_lex_state = 3},
  [7137] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7138] = {.lex_state = 120, .external_lex_state = 3},
  [7139] = {.lex_state = 64, .external_lex_state = 3},
  [7140] = {.lex_state = 16, .external_lex_state = 3},
  [7141] = {.lex_state = 26, .external_lex_state = 3},
  [7142] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7143] = {.lex_state = 120, .external_lex_state = 3},
  [7144] = {.lex_state = 120, .external_lex_state = 3},
  [7145] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7146] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7147] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7148] = {.lex_state = 120, .external_lex_state = 3},
  [7149] = {.lex_state = 16, .external_lex_state = 3},
  [7150] = {.lex_state = 16, .external_lex_state = 3},
  [7151] = {.lex_state = 16, .external_lex_state = 3},
  [7152] = {.lex_state = 16, .external_lex_state = 3},
  [7153] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7154] = {.lex_state = 120, .external_lex_state = 3},
  [7155] = {.lex_state = 120, .external_lex_state = 3},
  [7156] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7157] = {.lex_state = 127, .external_lex_state = 3},
  [7158] = {.lex_state = 16, .external_lex_state = 3},
  [7159] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 98},
  [7160] = {.lex_state = 120, .external_lex_state = 3},
  [7161] = {.lex_state = 120, .external_lex_state = 3},
  [7162] = {.lex_state = 120, .external_lex_state = 8},
  [7163] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7164] = {.lex_state = 121, .external_lex_state = 3},
  [7165] = {.lex_state = 120, .external_lex_state = 3},
  [7166] = {.lex_state = 120, .external_lex_state = 3},
  [7167] = {.lex_state = 8, .external_lex_state = 3},
  [7168] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7169] = {.lex_state = 120, .external_lex_state = 3},
  [7170] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7171] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7172] = {.lex_state = 120, .external_lex_state = 3},
  [7173] = {.lex_state = 120, .external_lex_state = 3},
  [7174] = {.lex_state = 120, .external_lex_state = 3},
  [7175] = {.lex_state = 120, .external_lex_state = 3},
  [7176] = {.lex_state = 120, .external_lex_state = 3},
  [7177] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7178] = {.lex_state = 120, .external_lex_state = 3},
  [7179] = {.lex_state = 26, .external_lex_state = 3},
  [7180] = {.lex_state = 120, .external_lex_state = 3},
  [7181] = {.lex_state = 16, .external_lex_state = 3},
  [7182] = {.lex_state = 120, .external_lex_state = 3},
  [7183] = {.lex_state = 120, .external_lex_state = 8},
  [7184] = {.lex_state = 120, .external_lex_state = 3},
  [7185] = {.lex_state = 120, .external_lex_state = 3},
  [7186] = {.lex_state = 120, .external_lex_state = 3},
  [7187] = {.lex_state = 120, .external_lex_state = 3},
  [7188] = {.lex_state = 120, .external_lex_state = 3},
  [7189] = {.lex_state = 120, .external_lex_state = 3},
  [7190] = {.lex_state = 120, .external_lex_state = 3},
  [7191] = {.lex_state = 120, .external_lex_state = 3},
  [7192] = {.lex_state = 120, .external_lex_state = 3},
  [7193] = {.lex_state = 16, .external_lex_state = 3},
  [7194] = {.lex_state = 120, .external_lex_state = 3},
  [7195] = {.lex_state = 120, .external_lex_state = 3},
  [7196] = {.lex_state = 120, .external_lex_state = 3},
  [7197] = {.lex_state = 16, .external_lex_state = 3},
  [7198] = {.lex_state = 120, .external_lex_state = 3},
  [7199] = {.lex_state = 120, .external_lex_state = 3},
  [7200] = {.lex_state = 16, .external_lex_state = 3},
  [7201] = {.lex_state = 120, .external_lex_state = 3},
  [7202] = {.lex_state = 120, .external_lex_state = 3},
  [7203] = {.lex_state = 120, .external_lex_state = 3},
  [7204] = {.lex_state = 120, .external_lex_state = 3},
  [7205] = {.lex_state = 120, .external_lex_state = 3},
  [7206] = {.lex_state = 120, .external_lex_state = 3},
  [7207] = {.lex_state = 120, .external_lex_state = 3},
  [7208] = {.lex_state = 120, .external_lex_state = 2},
  [7209] = {.lex_state = 120, .external_lex_state = 3},
  [7210] = {.lex_state = 16, .external_lex_state = 3},
  [7211] = {.lex_state = 16, .external_lex_state = 3},
  [7212] = {.lex_state = 120, .external_lex_state = 3},
  [7213] = {.lex_state = 120, .external_lex_state = 3},
  [7214] = {.lex_state = 120, .external_lex_state = 3},
  [7215] = {.lex_state = 120, .external_lex_state = 3},
  [7216] = {.lex_state = 120, .external_lex_state = 3},
  [7217] = {.lex_state = 120, .external_lex_state = 3},
  [7218] = {.lex_state = 16, .external_lex_state = 3},
  [7219] = {.lex_state = 16, .external_lex_state = 3},
  [7220] = {.lex_state = 120, .external_lex_state = 3},
  [7221] = {.lex_state = 16, .external_lex_state = 3},
  [7222] = {.lex_state = 120, .external_lex_state = 3},
  [7223] = {.lex_state = 120, .external_lex_state = 3},
  [7224] = {.lex_state = 120, .external_lex_state = 3},
  [7225] = {.lex_state = 120, .external_lex_state = 3},
  [7226] = {.lex_state = 120, .external_lex_state = 3},
  [7227] = {.lex_state = 120, .external_lex_state = 3},
  [7228] = {.lex_state = 120, .external_lex_state = 3},
  [7229] = {.lex_state = 16, .external_lex_state = 3},
  [7230] = {.lex_state = 120, .external_lex_state = 3},
  [7231] = {.lex_state = 16, .external_lex_state = 3},
  [7232] = {.lex_state = 120, .external_lex_state = 3},
  [7233] = {.lex_state = 120, .external_lex_state = 3},
  [7234] = {.lex_state = 120, .external_lex_state = 2},
  [7235] = {.lex_state = 120, .external_lex_state = 3},
  [7236] = {.lex_state = 120, .external_lex_state = 3},
  [7237] = {.lex_state = 120, .external_lex_state = 3},
  [7238] = {.lex_state = 120, .external_lex_state = 3},
  [7239] = {.lex_state = 120, .external_lex_state = 3},
  [7240] = {.lex_state = 120, .external_lex_state = 3},
  [7241] = {.lex_state = 16, .external_lex_state = 3},
  [7242] = {.lex_state = 127, .external_lex_state = 3},
  [7243] = {.lex_state = 120, .external_lex_state = 3},
  [7244] = {.lex_state = 120, .external_lex_state = 3},
  [7245] = {.lex_state = 120, .external_lex_state = 3},
  [7246] = {.lex_state = 120, .external_lex_state = 3},
  [7247] = {.lex_state = 120, .external_lex_state = 3},
  [7248] = {.lex_state = 120, .external_lex_state = 3},
  [7249] = {.lex_state = 120, .external_lex_state = 3},
  [7250] = {.lex_state = 120, .external_lex_state = 3},
  [7251] = {.lex_state = 120, .external_lex_state = 3},
  [7252] = {.lex_state = 120, .external_lex_state = 3},
  [7253] = {.lex_state = 120, .external_lex_state = 3},
  [7254] = {.lex_state = 120, .external_lex_state = 3},
  [7255] = {.lex_state = 120, .external_lex_state = 3},
  [7256] = {.lex_state = 16, .external_lex_state = 3},
  [7257] = {.lex_state = 16, .external_lex_state = 3},
  [7258] = {.lex_state = 120, .external_lex_state = 3},
  [7259] = {.lex_state = 120, .external_lex_state = 3},
  [7260] = {.lex_state = 120, .external_lex_state = 3},
  [7261] = {.lex_state = 120, .external_lex_state = 3},
  [7262] = {.lex_state = 120, .external_lex_state = 3},
  [7263] = {.lex_state = 120, .external_lex_state = 3},
  [7264] = {.lex_state = 120, .external_lex_state = 3},
  [7265] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7266] = {.lex_state = 120, .external_lex_state = 3},
  [7267] = {.lex_state = 120, .external_lex_state = 3},
  [7268] = {.lex_state = 120, .external_lex_state = 3},
  [7269] = {.lex_state = 120, .external_lex_state = 3},
  [7270] = {.lex_state = 120, .external_lex_state = 3},
  [7271] = {.lex_state = 120, .external_lex_state = 3},
  [7272] = {.lex_state = 16, .external_lex_state = 3},
  [7273] = {.lex_state = 120, .external_lex_state = 3},
  [7274] = {.lex_state = 120, .external_lex_state = 3},
  [7275] = {.lex_state = 16, .external_lex_state = 3},
  [7276] = {.lex_state = 120, .external_lex_state = 3},
  [7277] = {.lex_state = 120, .external_lex_state = 3},
  [7278] = {.lex_state = 16, .external_lex_state = 3},
  [7279] = {.lex_state = 120, .external_lex_state = 3},
  [7280] = {.lex_state = 120, .external_lex_state = 3},
  [7281] = {.lex_state = 120, .external_lex_state = 3},
  [7282] = {.lex_state = 16, .external_lex_state = 3},
  [7283] = {.lex_state = 120, .external_lex_state = 3},
  [7284] = {.lex_state = 120, .external_lex_state = 3},
  [7285] = {.lex_state = 120, .external_lex_state = 3},
  [7286] = {.lex_state = 120, .external_lex_state = 3},
  [7287] = {.lex_state = 120, .external_lex_state = 3},
  [7288] = {.lex_state = 120, .external_lex_state = 3},
  [7289] = {.lex_state = 16, .external_lex_state = 3},
  [7290] = {.lex_state = 16, .external_lex_state = 3},
  [7291] = {.lex_state = 120, .external_lex_state = 3},
  [7292] = {.lex_state = 16, .external_lex_state = 3},
  [7293] = {.lex_state = 120, .external_lex_state = 3},
  [7294] = {.lex_state = 120, .external_lex_state = 3},
  [7295] = {.lex_state = 120, .external_lex_state = 3},
  [7296] = {.lex_state = 120, .external_lex_state = 3},
  [7297] = {.lex_state = 120, .external_lex_state = 3},
  [7298] = {.lex_state = 120, .external_lex_state = 3},
  [7299] = {.lex_state = 120, .external_lex_state = 3},
  [7300] = {.lex_state = 120, .external_lex_state = 3},
  [7301] = {.lex_state = 120, .external_lex_state = 3},
  [7302] = {.lex_state = 120, .external_lex_state = 3},
  [7303] = {.lex_state = 120, .external_lex_state = 3},
  [7304] = {.lex_state = 120, .external_lex_state = 3},
  [7305] = {.lex_state = 120, .external_lex_state = 3},
  [7306] = {.lex_state = 120, .external_lex_state = 3},
  [7307] = {.lex_state = 16, .external_lex_state = 3},
  [7308] = {.lex_state = 127, .external_lex_state = 3},
  [7309] = {.lex_state = 120, .external_lex_state = 3},
  [7310] = {.lex_state = 16, .external_lex_state = 3},
  [7311] = {.lex_state = 120, .external_lex_state = 3},
  [7312] = {.lex_state = 120, .external_lex_state = 3},
  [7313] = {.lex_state = 120, .external_lex_state = 3},
  [7314] = {.lex_state = 120, .external_lex_state = 3},
  [7315] = {.lex_state = 120, .external_lex_state = 3},
  [7316] = {.lex_state = 120, .external_lex_state = 3},
  [7317] = {.lex_state = 120, .external_lex_state = 3},
  [7318] = {.lex_state = 120, .external_lex_state = 3},
  [7319] = {.lex_state = 120, .external_lex_state = 3},
  [7320] = {.lex_state = 120, .external_lex_state = 3},
  [7321] = {.lex_state = 120, .external_lex_state = 3},
  [7322] = {.lex_state = 120, .external_lex_state = 3},
  [7323] = {.lex_state = 120, .external_lex_state = 3},
  [7324] = {.lex_state = 120, .external_lex_state = 3},
  [7325] = {.lex_state = 120, .external_lex_state = 3},
  [7326] = {.lex_state = 16, .external_lex_state = 3},
  [7327] = {.lex_state = 120, .external_lex_state = 2},
  [7328] = {.lex_state = 120, .external_lex_state = 3},
  [7329] = {.lex_state = 120, .external_lex_state = 2},
  [7330] = {.lex_state = 120, .external_lex_state = 3},
  [7331] = {.lex_state = 120, .external_lex_state = 3},
  [7332] = {.lex_state = 120, .external_lex_state = 3},
  [7333] = {.lex_state = 120, .external_lex_state = 9},
  [7334] = {.lex_state = 120, .external_lex_state = 3},
  [7335] = {.lex_state = 120, .external_lex_state = 3},
  [7336] = {.lex_state = 16, .external_lex_state = 3},
  [7337] = {.lex_state = 120, .external_lex_state = 3},
  [7338] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7339] = {.lex_state = 16, .external_lex_state = 3},
  [7340] = {.lex_state = 120, .external_lex_state = 3},
  [7341] = {.lex_state = 120, .external_lex_state = 3},
  [7342] = {.lex_state = 120, .external_lex_state = 3},
  [7343] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7344] = {.lex_state = 120, .external_lex_state = 3},
  [7345] = {.lex_state = 16, .external_lex_state = 3},
  [7346] = {.lex_state = 120, .external_lex_state = 3},
  [7347] = {.lex_state = 16, .external_lex_state = 3},
  [7348] = {.lex_state = 120, .external_lex_state = 3},
  [7349] = {.lex_state = 16, .external_lex_state = 3},
  [7350] = {.lex_state = 120, .external_lex_state = 3},
  [7351] = {.lex_state = 120, .external_lex_state = 3},
  [7352] = {.lex_state = 120, .external_lex_state = 3},
  [7353] = {.lex_state = 16, .external_lex_state = 3},
  [7354] = {.lex_state = 120, .external_lex_state = 3},
  [7355] = {.lex_state = 120, .external_lex_state = 3},
  [7356] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7357] = {.lex_state = 120, .external_lex_state = 3},
  [7358] = {.lex_state = 16, .external_lex_state = 3},
  [7359] = {.lex_state = 120, .external_lex_state = 3},
  [7360] = {.lex_state = 120, .external_lex_state = 3},
  [7361] = {.lex_state = 120, .external_lex_state = 3},
  [7362] = {.lex_state = 120, .external_lex_state = 3},
  [7363] = {.lex_state = 120, .external_lex_state = 3},
  [7364] = {.lex_state = 120, .external_lex_state = 3},
  [7365] = {.lex_state = 120, .external_lex_state = 3},
  [7366] = {.lex_state = 120, .external_lex_state = 3},
  [7367] = {.lex_state = 120, .external_lex_state = 3},
  [7368] = {.lex_state = 120, .external_lex_state = 3},
  [7369] = {.lex_state = 120, .external_lex_state = 3},
  [7370] = {.lex_state = 120, .external_lex_state = 3},
  [7371] = {.lex_state = 120, .external_lex_state = 3},
  [7372] = {.lex_state = 16, .external_lex_state = 3},
  [7373] = {.lex_state = 120, .external_lex_state = 3},
  [7374] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7375] = {.lex_state = 120, .external_lex_state = 3},
  [7376] = {.lex_state = 120, .external_lex_state = 3},
  [7377] = {.lex_state = 120, .external_lex_state = 3},
  [7378] = {.lex_state = 120, .external_lex_state = 3},
  [7379] = {.lex_state = 120, .external_lex_state = 3},
  [7380] = {.lex_state = 16, .external_lex_state = 3},
  [7381] = {.lex_state = 120, .external_lex_state = 3},
  [7382] = {.lex_state = 16, .external_lex_state = 3},
  [7383] = {.lex_state = 16, .external_lex_state = 3},
  [7384] = {.lex_state = 120, .external_lex_state = 3},
  [7385] = {.lex_state = 120, .external_lex_state = 3},
  [7386] = {.lex_state = 120, .external_lex_state = 3},
  [7387] = {.lex_state = 120, .external_lex_state = 3},
  [7388] = {.lex_state = 120, .external_lex_state = 3},
  [7389] = {.lex_state = 16, .external_lex_state = 3},
  [7390] = {.lex_state = 16, .external_lex_state = 3},
  [7391] = {.lex_state = 120, .external_lex_state = 3},
  [7392] = {.lex_state = 16, .external_lex_state = 3},
  [7393] = {.lex_state = 120, .external_lex_state = 3},
  [7394] = {.lex_state = 120, .external_lex_state = 3},
  [7395] = {.lex_state = 120, .external_lex_state = 3},
  [7396] = {.lex_state = 120, .external_lex_state = 3},
  [7397] = {.lex_state = 16, .external_lex_state = 3},
  [7398] = {.lex_state = 120, .external_lex_state = 3},
  [7399] = {.lex_state = 120, .external_lex_state = 3},
  [7400] = {.lex_state = 120, .external_lex_state = 3},
  [7401] = {.lex_state = 120, .external_lex_state = 3},
  [7402] = {.lex_state = 16, .external_lex_state = 3},
  [7403] = {.lex_state = 120, .external_lex_state = 3},
  [7404] = {.lex_state = 120, .external_lex_state = 3},
  [7405] = {.lex_state = 120, .external_lex_state = 3},
  [7406] = {.lex_state = 120, .external_lex_state = 3},
  [7407] = {.lex_state = 120, .external_lex_state = 3},
  [7408] = {.lex_state = 120, .external_lex_state = 3},
  [7409] = {.lex_state = 120, .external_lex_state = 3},
  [7410] = {.lex_state = 120, .external_lex_state = 3},
  [7411] = {.lex_state = 120, .external_lex_state = 3},
  [7412] = {.lex_state = 120, .external_lex_state = 3},
  [7413] = {.lex_state = 120, .external_lex_state = 3},
  [7414] = {.lex_state = 16, .external_lex_state = 3},
  [7415] = {.lex_state = 120, .external_lex_state = 3},
  [7416] = {.lex_state = 120, .external_lex_state = 3},
  [7417] = {.lex_state = 120, .external_lex_state = 3},
  [7418] = {.lex_state = 120, .external_lex_state = 3},
  [7419] = {.lex_state = 16, .external_lex_state = 3},
  [7420] = {.lex_state = 120, .external_lex_state = 3},
  [7421] = {.lex_state = 16, .external_lex_state = 3},
  [7422] = {.lex_state = 16, .external_lex_state = 3},
  [7423] = {.lex_state = 120, .external_lex_state = 3},
  [7424] = {.lex_state = 16, .external_lex_state = 3},
  [7425] = {.lex_state = 120, .external_lex_state = 3},
  [7426] = {.lex_state = 120, .external_lex_state = 3},
  [7427] = {.lex_state = 120, .external_lex_state = 3},
  [7428] = {.lex_state = 16, .external_lex_state = 3},
  [7429] = {.lex_state = 16, .external_lex_state = 3},
  [7430] = {.lex_state = 120, .external_lex_state = 3},
  [7431] = {.lex_state = 120, .external_lex_state = 3},
  [7432] = {.lex_state = 120, .external_lex_state = 3},
  [7433] = {.lex_state = 16, .external_lex_state = 3},
  [7434] = {.lex_state = 120, .external_lex_state = 3},
  [7435] = {.lex_state = 13, .external_lex_state = 3},
  [7436] = {.lex_state = 120, .external_lex_state = 3},
  [7437] = {.lex_state = 16, .external_lex_state = 3},
  [7438] = {.lex_state = 120, .external_lex_state = 3},
  [7439] = {.lex_state = 16, .external_lex_state = 3},
  [7440] = {.lex_state = 120, .external_lex_state = 3},
  [7441] = {.lex_state = 120, .external_lex_state = 3},
  [7442] = {.lex_state = 120, .external_lex_state = 3},
  [7443] = {.lex_state = 120, .external_lex_state = 3},
  [7444] = {.lex_state = 120, .external_lex_state = 3},
  [7445] = {.lex_state = 120, .external_lex_state = 3},
  [7446] = {.lex_state = 120, .external_lex_state = 3},
  [7447] = {.lex_state = 120, .external_lex_state = 3},
  [7448] = {.lex_state = 120, .external_lex_state = 3},
  [7449] = {.lex_state = 16, .external_lex_state = 3},
  [7450] = {.lex_state = 120, .external_lex_state = 3},
  [7451] = {.lex_state = 120, .external_lex_state = 3},
  [7452] = {.lex_state = 120, .external_lex_state = 3},
  [7453] = {.lex_state = 120, .external_lex_state = 3},
  [7454] = {.lex_state = 120, .external_lex_state = 3},
  [7455] = {.lex_state = 120, .external_lex_state = 3},
  [7456] = {.lex_state = 120, .external_lex_state = 2},
  [7457] = {.lex_state = 120, .external_lex_state = 3},
  [7458] = {.lex_state = 16, .external_lex_state = 3},
  [7459] = {.lex_state = 16, .external_lex_state = 3},
  [7460] = {.lex_state = 120, .external_lex_state = 3},
  [7461] = {.lex_state = 16, .external_lex_state = 3},
  [7462] = {.lex_state = 120, .external_lex_state = 3},
  [7463] = {.lex_state = 120, .external_lex_state = 3},
  [7464] = {.lex_state = 120, .external_lex_state = 3},
  [7465] = {.lex_state = 16, .external_lex_state = 3},
  [7466] = {.lex_state = 16, .external_lex_state = 3},
  [7467] = {.lex_state = 16, .external_lex_state = 3},
  [7468] = {.lex_state = 120, .external_lex_state = 3},
  [7469] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7470] = {.lex_state = 120, .external_lex_state = 3},
  [7471] = {.lex_state = 120, .external_lex_state = 3},
  [7472] = {.lex_state = 120, .external_lex_state = 3},
  [7473] = {.lex_state = 120, .external_lex_state = 3},
  [7474] = {.lex_state = 120, .external_lex_state = 2},
  [7475] = {.lex_state = 120, .external_lex_state = 3},
  [7476] = {.lex_state = 120, .external_lex_state = 3},
  [7477] = {.lex_state = 120, .external_lex_state = 3},
  [7478] = {.lex_state = 120, .external_lex_state = 3},
  [7479] = {.lex_state = 120, .external_lex_state = 3},
  [7480] = {.lex_state = 120, .external_lex_state = 3},
  [7481] = {.lex_state = 120, .external_lex_state = 3},
  [7482] = {.lex_state = 120, .external_lex_state = 3},
  [7483] = {.lex_state = 120, .external_lex_state = 3},
  [7484] = {.lex_state = 120, .external_lex_state = 3},
  [7485] = {.lex_state = 120, .external_lex_state = 3},
  [7486] = {.lex_state = 120, .external_lex_state = 3},
  [7487] = {.lex_state = 120, .external_lex_state = 3},
  [7488] = {.lex_state = 120, .external_lex_state = 3},
  [7489] = {.lex_state = 127, .external_lex_state = 3},
  [7490] = {.lex_state = 16, .external_lex_state = 3},
  [7491] = {.lex_state = 120, .external_lex_state = 3},
  [7492] = {.lex_state = 16, .external_lex_state = 3},
  [7493] = {.lex_state = 16, .external_lex_state = 3},
  [7494] = {.lex_state = 16, .external_lex_state = 3},
  [7495] = {.lex_state = 16, .external_lex_state = 3},
  [7496] = {.lex_state = 120, .external_lex_state = 3},
  [7497] = {.lex_state = 16, .external_lex_state = 3},
  [7498] = {.lex_state = 120, .external_lex_state = 3},
  [7499] = {.lex_state = 120, .external_lex_state = 3},
  [7500] = {.lex_state = 120, .external_lex_state = 3},
  [7501] = {.lex_state = 16, .external_lex_state = 3},
  [7502] = {.lex_state = 16, .external_lex_state = 3},
  [7503] = {.lex_state = 120, .external_lex_state = 3},
  [7504] = {.lex_state = 16, .external_lex_state = 3},
  [7505] = {.lex_state = 120, .external_lex_state = 3},
  [7506] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7507] = {.lex_state = 120, .external_lex_state = 3},
  [7508] = {.lex_state = 120, .external_lex_state = 3},
  [7509] = {.lex_state = 120, .external_lex_state = 3},
  [7510] = {.lex_state = 120, .external_lex_state = 3},
  [7511] = {.lex_state = 120, .external_lex_state = 3},
  [7512] = {.lex_state = 120, .external_lex_state = 3},
  [7513] = {.lex_state = 120, .external_lex_state = 3},
  [7514] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7515] = {.lex_state = 120, .external_lex_state = 3},
  [7516] = {.lex_state = 120, .external_lex_state = 3},
  [7517] = {.lex_state = 120, .external_lex_state = 3},
  [7518] = {.lex_state = 120, .external_lex_state = 3},
  [7519] = {.lex_state = 120, .external_lex_state = 3},
  [7520] = {.lex_state = 120, .external_lex_state = 3},
  [7521] = {.lex_state = 16, .external_lex_state = 3},
  [7522] = {.lex_state = 120, .external_lex_state = 3},
  [7523] = {.lex_state = 120, .external_lex_state = 3},
  [7524] = {.lex_state = 16, .external_lex_state = 3},
  [7525] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7526] = {.lex_state = 120, .external_lex_state = 3},
  [7527] = {.lex_state = 120, .external_lex_state = 3},
  [7528] = {.lex_state = 120, .external_lex_state = 3},
  [7529] = {.lex_state = 16, .external_lex_state = 3},
  [7530] = {.lex_state = 16, .external_lex_state = 3},
  [7531] = {.lex_state = 16, .external_lex_state = 3},
  [7532] = {.lex_state = 120, .external_lex_state = 3},
  [7533] = {.lex_state = 16, .external_lex_state = 3},
  [7534] = {.lex_state = 120, .external_lex_state = 3},
  [7535] = {.lex_state = 120, .external_lex_state = 3},
  [7536] = {.lex_state = 120, .external_lex_state = 3},
  [7537] = {.lex_state = 16, .external_lex_state = 3},
  [7538] = {.lex_state = 16, .external_lex_state = 3},
  [7539] = {.lex_state = 120, .external_lex_state = 3},
  [7540] = {.lex_state = 120, .external_lex_state = 3},
  [7541] = {.lex_state = 120, .external_lex_state = 3},
  [7542] = {.lex_state = 16, .external_lex_state = 3},
  [7543] = {.lex_state = 120, .external_lex_state = 3},
  [7544] = {.lex_state = 120, .external_lex_state = 3},
  [7545] = {.lex_state = 120, .external_lex_state = 3},
  [7546] = {.lex_state = 16, .external_lex_state = 3},
  [7547] = {.lex_state = 120, .external_lex_state = 3},
  [7548] = {.lex_state = 120, .external_lex_state = 3},
  [7549] = {.lex_state = 120, .external_lex_state = 3},
  [7550] = {.lex_state = 120, .external_lex_state = 3},
  [7551] = {.lex_state = 120, .external_lex_state = 3},
  [7552] = {.lex_state = 120, .external_lex_state = 3},
  [7553] = {.lex_state = 120, .external_lex_state = 3},
  [7554] = {.lex_state = 120, .external_lex_state = 3},
  [7555] = {.lex_state = 120, .external_lex_state = 3},
  [7556] = {.lex_state = 120, .external_lex_state = 3},
  [7557] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7558] = {.lex_state = 16, .external_lex_state = 3},
  [7559] = {.lex_state = 120, .external_lex_state = 3},
  [7560] = {.lex_state = 120, .external_lex_state = 3},
  [7561] = {.lex_state = 120, .external_lex_state = 3},
  [7562] = {.lex_state = 120, .external_lex_state = 3},
  [7563] = {.lex_state = 120, .external_lex_state = 3},
  [7564] = {.lex_state = 120, .external_lex_state = 3},
  [7565] = {.lex_state = 16, .external_lex_state = 3},
  [7566] = {.lex_state = 16, .external_lex_state = 3},
  [7567] = {.lex_state = 120, .external_lex_state = 3},
  [7568] = {.lex_state = 16, .external_lex_state = 3},
  [7569] = {.lex_state = 120, .external_lex_state = 3},
  [7570] = {.lex_state = 120, .external_lex_state = 3},
  [7571] = {.lex_state = 16, .external_lex_state = 3},
  [7572] = {.lex_state = 16, .external_lex_state = 3},
  [7573] = {.lex_state = 16, .external_lex_state = 3},
  [7574] = {.lex_state = 120, .external_lex_state = 3},
  [7575] = {.lex_state = 120, .external_lex_state = 3},
  [7576] = {.lex_state = 120, .external_lex_state = 3},
  [7577] = {.lex_state = 120, .external_lex_state = 3},
  [7578] = {.lex_state = 120, .external_lex_state = 3},
  [7579] = {.lex_state = 120, .external_lex_state = 3},
  [7580] = {.lex_state = 120, .external_lex_state = 3},
  [7581] = {.lex_state = 120, .external_lex_state = 3},
  [7582] = {.lex_state = 120, .external_lex_state = 3},
  [7583] = {.lex_state = 120, .external_lex_state = 3},
  [7584] = {.lex_state = 120, .external_lex_state = 2},
  [7585] = {.lex_state = 16, .external_lex_state = 3},
  [7586] = {.lex_state = 120, .external_lex_state = 3},
  [7587] = {.lex_state = 120, .external_lex_state = 3},
  [7588] = {.lex_state = 120, .external_lex_state = 3},
  [7589] = {.lex_state = 120, .external_lex_state = 3},
  [7590] = {.lex_state = 120, .external_lex_state = 3},
  [7591] = {.lex_state = 120, .external_lex_state = 3},
  [7592] = {.lex_state = 120, .external_lex_state = 3},
  [7593] = {.lex_state = 127, .external_lex_state = 3},
  [7594] = {.lex_state = 120, .external_lex_state = 3},
  [7595] = {.lex_state = 120, .external_lex_state = 3},
  [7596] = {.lex_state = 120, .external_lex_state = 3},
  [7597] = {.lex_state = 120, .external_lex_state = 3},
  [7598] = {.lex_state = 120, .external_lex_state = 3},
  [7599] = {.lex_state = 120, .external_lex_state = 3},
  [7600] = {.lex_state = 120, .external_lex_state = 3},
  [7601] = {.lex_state = 16, .external_lex_state = 3},
  [7602] = {.lex_state = 16, .external_lex_state = 3},
  [7603] = {.lex_state = 120, .external_lex_state = 3},
  [7604] = {.lex_state = 16, .external_lex_state = 3},
  [7605] = {.lex_state = 120, .external_lex_state = 3},
  [7606] = {.lex_state = 120, .external_lex_state = 3},
  [7607] = {.lex_state = 120, .external_lex_state = 3},
  [7608] = {.lex_state = 16, .external_lex_state = 3},
  [7609] = {.lex_state = 16, .external_lex_state = 3},
  [7610] = {.lex_state = 120, .external_lex_state = 3},
  [7611] = {.lex_state = 120, .external_lex_state = 3},
  [7612] = {.lex_state = 127, .external_lex_state = 3},
  [7613] = {.lex_state = 120, .external_lex_state = 3},
  [7614] = {.lex_state = 120, .external_lex_state = 3},
  [7615] = {.lex_state = 120, .external_lex_state = 3},
  [7616] = {.lex_state = 127, .external_lex_state = 3},
  [7617] = {.lex_state = 16, .external_lex_state = 3},
  [7618] = {.lex_state = 120, .external_lex_state = 3},
  [7619] = {.lex_state = 120, .external_lex_state = 3},
  [7620] = {.lex_state = 120, .external_lex_state = 3},
  [7621] = {.lex_state = 16, .external_lex_state = 3},
  [7622] = {.lex_state = 120, .external_lex_state = 3},
  [7623] = {.lex_state = 120, .external_lex_state = 3},
  [7624] = {.lex_state = 120, .external_lex_state = 3},
  [7625] = {.lex_state = 120, .external_lex_state = 3},
  [7626] = {.lex_state = 120, .external_lex_state = 3},
  [7627] = {.lex_state = 120, .external_lex_state = 3},
  [7628] = {.lex_state = 120, .external_lex_state = 3},
  [7629] = {.lex_state = 120, .external_lex_state = 3},
  [7630] = {.lex_state = 120, .external_lex_state = 3},
  [7631] = {.lex_state = 16, .external_lex_state = 3},
  [7632] = {.lex_state = 120, .external_lex_state = 3},
  [7633] = {.lex_state = 120, .external_lex_state = 2},
  [7634] = {.lex_state = 120, .external_lex_state = 3},
  [7635] = {.lex_state = 120, .external_lex_state = 3},
  [7636] = {.lex_state = 120, .external_lex_state = 3},
  [7637] = {.lex_state = 120, .external_lex_state = 3},
  [7638] = {.lex_state = 16, .external_lex_state = 3},
  [7639] = {.lex_state = 120, .external_lex_state = 3},
  [7640] = {.lex_state = 120, .external_lex_state = 3},
  [7641] = {.lex_state = 16, .external_lex_state = 3},
  [7642] = {.lex_state = 120, .external_lex_state = 3},
  [7643] = {.lex_state = 16, .external_lex_state = 3},
  [7644] = {.lex_state = 120, .external_lex_state = 3},
  [7645] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7646] = {.lex_state = 120, .external_lex_state = 3},
  [7647] = {.lex_state = 120, .external_lex_state = 3},
  [7648] = {.lex_state = 120, .external_lex_state = 3},
  [7649] = {.lex_state = 120, .external_lex_state = 3},
  [7650] = {.lex_state = 120, .external_lex_state = 3},
  [7651] = {.lex_state = 120, .external_lex_state = 3},
  [7652] = {.lex_state = 120, .external_lex_state = 3},
  [7653] = {.lex_state = 16, .external_lex_state = 3},
  [7654] = {.lex_state = 127, .external_lex_state = 3},
  [7655] = {.lex_state = 16, .external_lex_state = 3},
  [7656] = {.lex_state = 120, .external_lex_state = 3},
  [7657] = {.lex_state = 120, .external_lex_state = 3},
  [7658] = {.lex_state = 120, .external_lex_state = 3},
  [7659] = {.lex_state = 16, .external_lex_state = 3},
  [7660] = {.lex_state = 120, .external_lex_state = 3},
  [7661] = {.lex_state = 120, .external_lex_state = 3},
  [7662] = {.lex_state = 120, .external_lex_state = 3},
  [7663] = {.lex_state = 16, .external_lex_state = 3},
  [7664] = {.lex_state = 16, .external_lex_state = 3},
  [7665] = {.lex_state = 16, .external_lex_state = 3},
  [7666] = {.lex_state = 120, .external_lex_state = 2},
  [7667] = {.lex_state = 16, .external_lex_state = 3},
  [7668] = {.lex_state = 120, .external_lex_state = 3},
  [7669] = {.lex_state = 120, .external_lex_state = 3},
  [7670] = {.lex_state = 120, .external_lex_state = 3},
  [7671] = {.lex_state = 120, .external_lex_state = 3},
  [7672] = {.lex_state = 120, .external_lex_state = 3},
  [7673] = {.lex_state = 120, .external_lex_state = 2},
  [7674] = {.lex_state = 120, .external_lex_state = 3},
  [7675] = {.lex_state = 16, .external_lex_state = 3},
  [7676] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7677] = {.lex_state = 120, .external_lex_state = 3},
  [7678] = {.lex_state = 120, .external_lex_state = 3},
  [7679] = {.lex_state = 120, .external_lex_state = 3},
  [7680] = {.lex_state = 120, .external_lex_state = 3},
  [7681] = {.lex_state = 120, .external_lex_state = 3},
  [7682] = {.lex_state = 120, .external_lex_state = 3},
  [7683] = {.lex_state = 120, .external_lex_state = 3},
  [7684] = {.lex_state = 120, .external_lex_state = 3},
  [7685] = {.lex_state = 120, .external_lex_state = 3},
  [7686] = {.lex_state = 16, .external_lex_state = 3},
  [7687] = {.lex_state = 120, .external_lex_state = 3},
  [7688] = {.lex_state = 120, .external_lex_state = 3},
  [7689] = {.lex_state = 120, .external_lex_state = 3},
  [7690] = {.lex_state = 120, .external_lex_state = 3},
  [7691] = {.lex_state = 120, .external_lex_state = 3},
  [7692] = {.lex_state = 120, .external_lex_state = 3},
  [7693] = {.lex_state = 16, .external_lex_state = 3},
  [7694] = {.lex_state = 120, .external_lex_state = 3},
  [7695] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7696] = {.lex_state = 120, .external_lex_state = 3},
  [7697] = {.lex_state = 120, .external_lex_state = 3},
  [7698] = {.lex_state = 16, .external_lex_state = 3},
  [7699] = {.lex_state = 120, .external_lex_state = 3},
  [7700] = {.lex_state = 120, .external_lex_state = 3},
  [7701] = {.lex_state = 120, .external_lex_state = 3},
  [7702] = {.lex_state = 120, .external_lex_state = 3},
  [7703] = {.lex_state = 120, .external_lex_state = 3},
  [7704] = {.lex_state = 120, .external_lex_state = 3},
  [7705] = {.lex_state = 120, .external_lex_state = 2},
  [7706] = {.lex_state = 120, .external_lex_state = 3},
  [7707] = {.lex_state = 120, .external_lex_state = 3},
  [7708] = {.lex_state = 120, .external_lex_state = 3},
  [7709] = {.lex_state = 120, .external_lex_state = 3},
  [7710] = {.lex_state = 120, .external_lex_state = 3},
  [7711] = {.lex_state = 120, .external_lex_state = 3},
  [7712] = {.lex_state = 120, .external_lex_state = 3},
  [7713] = {.lex_state = 16, .external_lex_state = 3},
  [7714] = {.lex_state = 16, .external_lex_state = 3},
  [7715] = {.lex_state = 120, .external_lex_state = 3},
  [7716] = {.lex_state = 120, .external_lex_state = 3},
  [7717] = {.lex_state = 16, .external_lex_state = 3},
  [7718] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7719] = {.lex_state = 120, .external_lex_state = 3},
  [7720] = {.lex_state = 120, .external_lex_state = 3},
  [7721] = {.lex_state = 120, .external_lex_state = 3},
  [7722] = {.lex_state = 16, .external_lex_state = 3},
  [7723] = {.lex_state = 120, .external_lex_state = 3},
  [7724] = {.lex_state = 120, .external_lex_state = 2},
  [7725] = {.lex_state = 16, .external_lex_state = 3},
  [7726] = {.lex_state = 120, .external_lex_state = 3},
  [7727] = {.lex_state = 120, .external_lex_state = 3},
  [7728] = {.lex_state = 120, .external_lex_state = 3},
  [7729] = {.lex_state = 120, .external_lex_state = 3},
  [7730] = {.lex_state = 120, .external_lex_state = 3},
  [7731] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7732] = {.lex_state = 120, .external_lex_state = 3},
  [7733] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7734] = {.lex_state = 120, .external_lex_state = 3},
  [7735] = {.lex_state = 120, .external_lex_state = 3},
  [7736] = {.lex_state = 120, .external_lex_state = 3},
  [7737] = {.lex_state = 120, .external_lex_state = 3},
  [7738] = {.lex_state = 16, .external_lex_state = 3},
  [7739] = {.lex_state = 120, .external_lex_state = 3},
  [7740] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7741] = {.lex_state = 16, .external_lex_state = 3},
  [7742] = {.lex_state = 120, .external_lex_state = 3},
  [7743] = {.lex_state = 120, .external_lex_state = 3},
  [7744] = {.lex_state = 120, .external_lex_state = 3},
  [7745] = {.lex_state = 120, .external_lex_state = 3},
  [7746] = {.lex_state = 13, .external_lex_state = 3},
  [7747] = {.lex_state = 120, .external_lex_state = 3},
  [7748] = {.lex_state = 16, .external_lex_state = 3},
  [7749] = {.lex_state = 120, .external_lex_state = 3},
  [7750] = {.lex_state = 120, .external_lex_state = 3},
  [7751] = {.lex_state = 16, .external_lex_state = 3},
  [7752] = {.lex_state = 120, .external_lex_state = 3},
  [7753] = {.lex_state = 120, .external_lex_state = 3},
  [7754] = {.lex_state = 16, .external_lex_state = 3},
  [7755] = {.lex_state = 16, .external_lex_state = 3},
  [7756] = {.lex_state = 120, .external_lex_state = 3},
  [7757] = {.lex_state = 120, .external_lex_state = 3},
  [7758] = {.lex_state = 120, .external_lex_state = 3},
  [7759] = {.lex_state = 120, .external_lex_state = 3},
  [7760] = {.lex_state = 120, .external_lex_state = 3},
  [7761] = {.lex_state = 120, .external_lex_state = 3},
  [7762] = {.lex_state = 120, .external_lex_state = 3},
  [7763] = {.lex_state = 16, .external_lex_state = 3},
  [7764] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7765] = {.lex_state = 120, .external_lex_state = 3},
  [7766] = {.lex_state = 16, .external_lex_state = 3},
  [7767] = {.lex_state = 120, .external_lex_state = 3},
  [7768] = {.lex_state = 120, .external_lex_state = 3},
  [7769] = {.lex_state = 120, .external_lex_state = 3},
  [7770] = {.lex_state = 16, .external_lex_state = 3},
  [7771] = {.lex_state = 16, .external_lex_state = 3},
  [7772] = {.lex_state = 120, .external_lex_state = 3},
  [7773] = {.lex_state = 120, .external_lex_state = 3},
  [7774] = {.lex_state = 120, .external_lex_state = 3},
  [7775] = {.lex_state = 13, .external_lex_state = 3},
  [7776] = {.lex_state = 16, .external_lex_state = 3},
  [7777] = {.lex_state = 120, .external_lex_state = 3},
  [7778] = {.lex_state = 120, .external_lex_state = 3},
  [7779] = {.lex_state = 120, .external_lex_state = 3},
  [7780] = {.lex_state = 120, .external_lex_state = 3},
  [7781] = {.lex_state = 120, .external_lex_state = 3},
  [7782] = {.lex_state = 120, .external_lex_state = 3},
  [7783] = {.lex_state = 120, .external_lex_state = 3},
  [7784] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7785] = {.lex_state = 120, .external_lex_state = 3},
  [7786] = {.lex_state = 120, .external_lex_state = 3},
  [7787] = {.lex_state = 120, .external_lex_state = 3},
  [7788] = {.lex_state = 120, .external_lex_state = 3},
  [7789] = {.lex_state = 120, .external_lex_state = 3},
  [7790] = {.lex_state = 120, .external_lex_state = 3},
  [7791] = {.lex_state = 120, .external_lex_state = 3},
  [7792] = {.lex_state = 120, .external_lex_state = 3},
  [7793] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7794] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7795] = {.lex_state = 120, .external_lex_state = 3},
  [7796] = {.lex_state = 120, .external_lex_state = 3},
  [7797] = {.lex_state = 120, .external_lex_state = 3},
  [7798] = {.lex_state = 120, .external_lex_state = 3},
  [7799] = {.lex_state = 120, .external_lex_state = 3},
  [7800] = {.lex_state = 120, .external_lex_state = 3},
  [7801] = {.lex_state = 120, .external_lex_state = 3},
  [7802] = {.lex_state = 120, .external_lex_state = 3},
  [7803] = {.lex_state = 120, .external_lex_state = 3},
  [7804] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7805] = {.lex_state = 120, .external_lex_state = 3},
  [7806] = {.lex_state = 120, .external_lex_state = 3},
  [7807] = {.lex_state = 120, .external_lex_state = 3},
  [7808] = {.lex_state = 120, .external_lex_state = 3},
  [7809] = {.lex_state = 121, .external_lex_state = 3},
  [7810] = {.lex_state = 120, .external_lex_state = 3},
  [7811] = {.lex_state = 120, .external_lex_state = 3},
  [7812] = {.lex_state = 120, .external_lex_state = 3},
  [7813] = {.lex_state = 120, .external_lex_state = 3},
  [7814] = {.lex_state = 120, .external_lex_state = 3},
  [7815] = {.lex_state = 120, .external_lex_state = 3},
  [7816] = {.lex_state = 120, .external_lex_state = 3},
  [7817] = {.lex_state = 120, .external_lex_state = 3},
  [7818] = {.lex_state = 120, .external_lex_state = 3},
  [7819] = {.lex_state = 120, .external_lex_state = 3},
  [7820] = {.lex_state = 120, .external_lex_state = 3},
  [7821] = {.lex_state = 127, .external_lex_state = 3},
  [7822] = {.lex_state = 127, .external_lex_state = 3},
  [7823] = {.lex_state = 127, .external_lex_state = 3},
  [7824] = {.lex_state = 120, .external_lex_state = 3},
  [7825] = {.lex_state = 120, .external_lex_state = 3},
  [7826] = {.lex_state = 120, .external_lex_state = 3},
  [7827] = {.lex_state = 120, .external_lex_state = 3},
  [7828] = {.lex_state = 120, .external_lex_state = 3},
  [7829] = {.lex_state = 16, .external_lex_state = 3},
  [7830] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7831] = {.lex_state = 120, .external_lex_state = 3},
  [7832] = {.lex_state = 13, .external_lex_state = 3},
  [7833] = {.lex_state = 120, .external_lex_state = 3},
  [7834] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7835] = {.lex_state = 120, .external_lex_state = 3},
  [7836] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7837] = {.lex_state = 120, .external_lex_state = 3},
  [7838] = {.lex_state = 120, .external_lex_state = 3},
  [7839] = {.lex_state = 120, .external_lex_state = 3},
  [7840] = {.lex_state = 120, .external_lex_state = 3},
  [7841] = {.lex_state = 16, .external_lex_state = 3},
  [7842] = {.lex_state = 120, .external_lex_state = 3},
  [7843] = {.lex_state = 16, .external_lex_state = 3},
  [7844] = {.lex_state = 120, .external_lex_state = 3},
  [7845] = {.lex_state = 120, .external_lex_state = 3},
  [7846] = {.lex_state = 120, .external_lex_state = 3},
  [7847] = {.lex_state = 120, .external_lex_state = 3},
  [7848] = {.lex_state = 120, .external_lex_state = 3},
  [7849] = {.lex_state = 120, .external_lex_state = 3},
  [7850] = {.lex_state = 16, .external_lex_state = 3},
  [7851] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7852] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7853] = {.lex_state = 120, .external_lex_state = 3},
  [7854] = {.lex_state = 16, .external_lex_state = 3},
  [7855] = {.lex_state = 120, .external_lex_state = 3},
  [7856] = {.lex_state = 120, .external_lex_state = 3},
  [7857] = {.lex_state = 120, .external_lex_state = 3},
  [7858] = {.lex_state = 120, .external_lex_state = 3},
  [7859] = {.lex_state = 120, .external_lex_state = 3},
  [7860] = {.lex_state = 16, .external_lex_state = 3},
  [7861] = {.lex_state = 120, .external_lex_state = 3},
  [7862] = {.lex_state = 120, .external_lex_state = 3},
  [7863] = {.lex_state = 120, .external_lex_state = 3},
  [7864] = {.lex_state = 127, .external_lex_state = 3},
  [7865] = {.lex_state = 120, .external_lex_state = 3},
  [7866] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7867] = {.lex_state = 120, .external_lex_state = 3},
  [7868] = {.lex_state = 16, .external_lex_state = 3},
  [7869] = {.lex_state = 120, .external_lex_state = 3},
  [7870] = {.lex_state = 120, .external_lex_state = 3},
  [7871] = {.lex_state = 120, .external_lex_state = 3},
  [7872] = {.lex_state = 120, .external_lex_state = 3},
  [7873] = {.lex_state = 120, .external_lex_state = 3},
  [7874] = {.lex_state = 16, .external_lex_state = 3},
  [7875] = {.lex_state = 120, .external_lex_state = 3},
  [7876] = {.lex_state = 120, .external_lex_state = 3},
  [7877] = {.lex_state = 120, .external_lex_state = 3},
  [7878] = {.lex_state = 120, .external_lex_state = 3},
  [7879] = {.lex_state = 120, .external_lex_state = 3},
  [7880] = {.lex_state = 120, .external_lex_state = 3},
  [7881] = {.lex_state = 120, .external_lex_state = 3},
  [7882] = {.lex_state = 120, .external_lex_state = 3},
  [7883] = {.lex_state = 120, .external_lex_state = 3},
  [7884] = {.lex_state = 127, .external_lex_state = 3},
  [7885] = {.lex_state = 120, .external_lex_state = 3},
  [7886] = {.lex_state = 120, .external_lex_state = 3},
  [7887] = {.lex_state = 16, .external_lex_state = 3},
  [7888] = {.lex_state = 120, .external_lex_state = 2},
  [7889] = {.lex_state = 120, .external_lex_state = 3},
  [7890] = {.lex_state = 120, .external_lex_state = 3},
  [7891] = {.lex_state = 127, .external_lex_state = 3},
  [7892] = {.lex_state = 120, .external_lex_state = 3},
  [7893] = {.lex_state = 120, .external_lex_state = 3},
  [7894] = {.lex_state = 16, .external_lex_state = 3},
  [7895] = {.lex_state = 120, .external_lex_state = 3},
  [7896] = {.lex_state = 120, .external_lex_state = 3},
  [7897] = {.lex_state = 120, .external_lex_state = 3},
  [7898] = {.lex_state = 120, .external_lex_state = 3},
  [7899] = {.lex_state = 120, .external_lex_state = 3},
  [7900] = {.lex_state = 120, .external_lex_state = 3},
  [7901] = {.lex_state = 120, .external_lex_state = 3},
  [7902] = {.lex_state = 120, .external_lex_state = 3},
  [7903] = {.lex_state = 16, .external_lex_state = 3},
  [7904] = {.lex_state = 120, .external_lex_state = 2},
  [7905] = {.lex_state = 120, .external_lex_state = 3},
  [7906] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7907] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7908] = {.lex_state = 16, .external_lex_state = 3},
  [7909] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7910] = {.lex_state = 120, .external_lex_state = 3},
  [7911] = {.lex_state = 16, .external_lex_state = 3},
  [7912] = {.lex_state = 120, .external_lex_state = 3},
  [7913] = {.lex_state = 120, .external_lex_state = 3},
  [7914] = {.lex_state = 120, .external_lex_state = 3},
  [7915] = {.lex_state = 120, .external_lex_state = 3},
  [7916] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7917] = {.lex_state = 120, .external_lex_state = 3},
  [7918] = {.lex_state = 65, .external_lex_state = 3},
  [7919] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7920] = {.lex_state = 120, .external_lex_state = 3},
  [7921] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7922] = {.lex_state = 120, .external_lex_state = 3},
  [7923] = {.lex_state = 120, .external_lex_state = 3},
  [7924] = {.lex_state = 120, .external_lex_state = 3},
  [7925] = {.lex_state = 16, .external_lex_state = 3},
  [7926] = {.lex_state = 120, .external_lex_state = 3},
  [7927] = {.lex_state = 16, .external_lex_state = 3},
  [7928] = {.lex_state = 120, .external_lex_state = 3},
  [7929] = {.lex_state = 120, .external_lex_state = 3},
  [7930] = {.lex_state = 120, .external_lex_state = 3},
  [7931] = {.lex_state = 16, .external_lex_state = 3},
  [7932] = {.lex_state = 120, .external_lex_state = 3},
  [7933] = {.lex_state = 16, .external_lex_state = 3},
  [7934] = {.lex_state = 16, .external_lex_state = 3},
  [7935] = {.lex_state = 120, .external_lex_state = 3},
  [7936] = {.lex_state = 16, .external_lex_state = 3},
  [7937] = {.lex_state = 120, .external_lex_state = 3},
  [7938] = {.lex_state = 120, .external_lex_state = 3},
  [7939] = {.lex_state = 16, .external_lex_state = 3},
  [7940] = {.lex_state = 120, .external_lex_state = 3},
  [7941] = {.lex_state = 120, .external_lex_state = 3},
  [7942] = {.lex_state = 120, .external_lex_state = 3},
  [7943] = {.lex_state = 120, .external_lex_state = 3},
  [7944] = {.lex_state = 120, .external_lex_state = 3},
  [7945] = {.lex_state = 16, .external_lex_state = 3},
  [7946] = {.lex_state = 120, .external_lex_state = 3},
  [7947] = {.lex_state = 120, .external_lex_state = 3},
  [7948] = {.lex_state = 120, .external_lex_state = 3},
  [7949] = {.lex_state = 16, .external_lex_state = 3},
  [7950] = {.lex_state = 120, .external_lex_state = 3},
  [7951] = {.lex_state = 120, .external_lex_state = 3},
  [7952] = {.lex_state = 120, .external_lex_state = 3},
  [7953] = {.lex_state = 16, .external_lex_state = 3},
  [7954] = {.lex_state = 120, .external_lex_state = 3},
  [7955] = {.lex_state = 16, .external_lex_state = 3},
  [7956] = {.lex_state = 120, .external_lex_state = 3},
  [7957] = {.lex_state = 120, .external_lex_state = 2},
  [7958] = {.lex_state = 120, .external_lex_state = 3},
  [7959] = {.lex_state = 120, .external_lex_state = 3},
  [7960] = {.lex_state = 120, .external_lex_state = 3},
  [7961] = {.lex_state = 16, .external_lex_state = 3},
  [7962] = {.lex_state = 120, .external_lex_state = 3},
  [7963] = {.lex_state = 16, .external_lex_state = 3},
  [7964] = {.lex_state = 120, .external_lex_state = 3},
  [7965] = {.lex_state = 120, .external_lex_state = 3},
  [7966] = {.lex_state = 120, .external_lex_state = 3},
  [7967] = {.lex_state = 120, .external_lex_state = 3},
  [7968] = {.lex_state = 120, .external_lex_state = 3},
  [7969] = {.lex_state = 16, .external_lex_state = 3},
  [7970] = {.lex_state = 120, .external_lex_state = 3},
  [7971] = {.lex_state = 120, .external_lex_state = 3},
  [7972] = {.lex_state = 120, .external_lex_state = 3},
  [7973] = {.lex_state = 16, .external_lex_state = 3},
  [7974] = {.lex_state = 16, .external_lex_state = 3},
  [7975] = {.lex_state = 120, .external_lex_state = 3},
  [7976] = {.lex_state = 120, .external_lex_state = 3},
  [7977] = {.lex_state = 120, .external_lex_state = 3},
  [7978] = {.lex_state = 120, .external_lex_state = 2},
  [7979] = {.lex_state = 120, .external_lex_state = 3},
  [7980] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7981] = {.lex_state = 120, .external_lex_state = 3},
  [7982] = {.lex_state = 120, .external_lex_state = 3},
  [7983] = {.lex_state = 120, .external_lex_state = 3},
  [7984] = {.lex_state = 120, .external_lex_state = 3},
  [7985] = {.lex_state = 120, .external_lex_state = 3},
  [7986] = {.lex_state = 120, .external_lex_state = 3},
  [7987] = {.lex_state = 120, .external_lex_state = 3},
  [7988] = {.lex_state = 120, .external_lex_state = 3},
  [7989] = {.lex_state = 120, .external_lex_state = 3},
  [7990] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7991] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [7992] = {.lex_state = 120, .external_lex_state = 3},
  [7993] = {.lex_state = 120, .external_lex_state = 3},
  [7994] = {.lex_state = 120, .external_lex_state = 3},
  [7995] = {.lex_state = 120, .external_lex_state = 3},
  [7996] = {.lex_state = 120, .external_lex_state = 3},
  [7997] = {.lex_state = 120, .external_lex_state = 3},
  [7998] = {.lex_state = 120, .external_lex_state = 3},
  [7999] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [8000] = {.lex_state = 120, .external_lex_state = 3},
  [8001] = {.lex_state = 120, .external_lex_state = 3},
  [8002] = {.lex_state = 121, .external_lex_state = 3},
  [8003] = {.lex_state = 120, .external_lex_state = 3},
  [8004] = {.lex_state = 127, .external_lex_state = 3},
  [8005] = {.lex_state = 127, .external_lex_state = 3},
  [8006] = {.lex_state = 127, .external_lex_state = 3},
  [8007] = {.lex_state = 120, .external_lex_state = 3},
  [8008] = {.lex_state = 120, .external_lex_state = 3},
  [8009] = {.lex_state = 16, .external_lex_state = 3},
  [8010] = {.lex_state = 120, .external_lex_state = 3},
  [8011] = {.lex_state = 120, .external_lex_state = 3},
  [8012] = {.lex_state = 120, .external_lex_state = 3},
  [8013] = {.lex_state = 120, .external_lex_state = 3},
  [8014] = {.lex_state = 120, .external_lex_state = 3},
  [8015] = {.lex_state = 120, .external_lex_state = 3},
  [8016] = {.lex_state = 120, .external_lex_state = 3},
  [8017] = {.lex_state = 120, .external_lex_state = 3},
  [8018] = {.lex_state = 120, .external_lex_state = 3},
  [8019] = {.lex_state = 120, .external_lex_state = 3},
  [8020] = {.lex_state = 16, .external_lex_state = 3},
  [8021] = {.lex_state = 120, .external_lex_state = 3},
  [8022] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [8023] = {.lex_state = 120, .external_lex_state = 3},
  [8024] = {.lex_state = 16, .external_lex_state = 3},
  [8025] = {.lex_state = 120, .external_lex_state = 3},
  [8026] = {.lex_state = 120, .external_lex_state = 3},
  [8027] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [8028] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [8029] = {.lex_state = 120, .external_lex_state = 3},
  [8030] = {.lex_state = 120, .external_lex_state = 3},
  [8031] = {.lex_state = 16, .external_lex_state = 3},
  [8032] = {.lex_state = 120, .external_lex_state = 3},
  [8033] = {.lex_state = 120, .external_lex_state = 3},
  [8034] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [8035] = {.lex_state = 120, .external_lex_state = 3},
  [8036] = {.lex_state = 120, .external_lex_state = 3},
  [8037] = {.lex_state = 120, .external_lex_state = 3},
  [8038] = {.lex_state = 127, .external_lex_state = 3},
  [8039] = {.lex_state = 127, .external_lex_state = 3},
  [8040] = {.lex_state = 127, .external_lex_state = 3},
  [8041] = {.lex_state = 120, .external_lex_state = 3},
  [8042] = {.lex_state = 120, .external_lex_state = 3},
  [8043] = {.lex_state = 120, .external_lex_state = 3},
  [8044] = {.lex_state = 120, .external_lex_state = 3},
  [8045] = {.lex_state = 16, .external_lex_state = 3},
  [8046] = {.lex_state = 120, .external_lex_state = 3},
  [8047] = {.lex_state = 16, .external_lex_state = 3},
  [8048] = {.lex_state = 120, .external_lex_state = 3},
  [8049] = {.lex_state = 120, .external_lex_state = 3},
  [8050] = {.lex_state = 120, .external_lex_state = 3},
  [8051] = {.lex_state = 120, .external_lex_state = 3},
  [8052] = {.lex_state = 120, .external_lex_state = 3},
  [8053] = {.lex_state = 120, .external_lex_state = 3},
  [8054] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [8055] = {.lex_state = 120, .external_lex_state = 3},
  [8056] = {.lex_state = 127, .external_lex_state = 3},
  [8057] = {.lex_state = 127, .external_lex_state = 3},
  [8058] = {.lex_state = 127, .external_lex_state = 3},
  [8059] = {.lex_state = 120, .external_lex_state = 3},
  [8060] = {.lex_state = 120, .external_lex_state = 3},
  [8061] = {.lex_state = 120, .external_lex_state = 3},
  [8062] = {.lex_state = 120, .external_lex_state = 3},
  [8063] = {.lex_state = 16, .external_lex_state = 3},
  [8064] = {.lex_state = 120, .external_lex_state = 3},
  [8065] = {.lex_state = 120, .external_lex_state = 3},
  [8066] = {.lex_state = 127, .external_lex_state = 3},
  [8067] = {.lex_state = 120, .external_lex_state = 3},
  [8068] = {.lex_state = 120, .external_lex_state = 3},
  [8069] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [8070] = {.lex_state = 120, .external_lex_state = 3},
  [8071] = {.lex_state = 127, .external_lex_state = 3},
  [8072] = {.lex_state = 127, .external_lex_state = 3},
  [8073] = {.lex_state = 127, .external_lex_state = 3},
  [8074] = {.lex_state = 120, .external_lex_state = 3},
  [8075] = {.lex_state = 16, .external_lex_state = 3},
  [8076] = {.lex_state = 120, .external_lex_state = 3},
  [8077] = {.lex_state = 120, .external_lex_state = 3},
  [8078] = {.lex_state = 120, .external_lex_state = 3},
  [8079] = {.lex_state = 120, .external_lex_state = 3},
  [8080] = {.lex_state = 120, .external_lex_state = 3},
  [8081] = {.lex_state = 120, .external_lex_state = 3},
  [8082] = {.lex_state = 120, .external_lex_state = 3},
  [8083] = {.lex_state = 16, .external_lex_state = 3},
  [8084] = {.lex_state = 120, .external_lex_state = 3},
  [8085] = {.lex_state = 127, .external_lex_state = 3},
  [8086] = {.lex_state = 127, .external_lex_state = 3},
  [8087] = {.lex_state = 127, .external_lex_state = 3},
  [8088] = {.lex_state = 120, .external_lex_state = 3},
  [8089] = {.lex_state = 120, .external_lex_state = 3},
  [8090] = {.lex_state = 120, .external_lex_state = 3},
  [8091] = {.lex_state = 120, .external_lex_state = 2},
  [8092] = {.lex_state = 16, .external_lex_state = 3},
  [8093] = {.lex_state = 120, .external_lex_state = 3},
  [8094] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [8095] = {.lex_state = 16, .external_lex_state = 3},
  [8096] = {.lex_state = 120, .external_lex_state = 3},
  [8097] = {.lex_state = 16, .external_lex_state = 3},
  [8098] = {.lex_state = 120, .external_lex_state = 3},
  [8099] = {.lex_state = 127, .external_lex_state = 3},
  [8100] = {.lex_state = 127, .external_lex_state = 3},
  [8101] = {.lex_state = 127, .external_lex_state = 3},
  [8102] = {.lex_state = 120, .external_lex_state = 3},
  [8103] = {.lex_state = 120, .external_lex_state = 3},
  [8104] = {.lex_state = 120, .external_lex_state = 3},
  [8105] = {.lex_state = 120, .external_lex_state = 3},
  [8106] = {.lex_state = 120, .external_lex_state = 3},
  [8107] = {.lex_state = 120, .external_lex_state = 3},
  [8108] = {.lex_state = 120, .external_lex_state = 3},
  [8109] = {.lex_state = 120, .external_lex_state = 3},
  [8110] = {.lex_state = 120, .external_lex_state = 3},
  [8111] = {.lex_state = 120, .external_lex_state = 3},
  [8112] = {.lex_state = 120, .external_lex_state = 3},
  [8113] = {.lex_state = 127, .external_lex_state = 3},
  [8114] = {.lex_state = 127, .external_lex_state = 3},
  [8115] = {.lex_state = 127, .external_lex_state = 3},
  [8116] = {.lex_state = 120, .external_lex_state = 3},
  [8117] = {.lex_state = 120, .external_lex_state = 3},
  [8118] = {.lex_state = 120, .external_lex_state = 3},
  [8119] = {.lex_state = 120, .external_lex_state = 3},
  [8120] = {.lex_state = 120, .external_lex_state = 3},
  [8121] = {.lex_state = 120, .external_lex_state = 3},
  [8122] = {.lex_state = 120, .external_lex_state = 3},
  [8123] = {.lex_state = 120, .external_lex_state = 3},
  [8124] = {.lex_state = 120, .external_lex_state = 3},
  [8125] = {.lex_state = 120, .external_lex_state = 3},
  [8126] = {.lex_state = 120, .external_lex_state = 3},
  [8127] = {.lex_state = 127, .external_lex_state = 3},
  [8128] = {.lex_state = 127, .external_lex_state = 3},
  [8129] = {.lex_state = 127, .external_lex_state = 3},
  [8130] = {.lex_state = 16, .external_lex_state = 3},
  [8131] = {.lex_state = 120, .external_lex_state = 3},
  [8132] = {.lex_state = 120, .external_lex_state = 3},
  [8133] = {.lex_state = 120, .external_lex_state = 3},
  [8134] = {.lex_state = 120, .external_lex_state = 3},
  [8135] = {.lex_state = 120, .external_lex_state = 3},
  [8136] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [8137] = {.lex_state = 120, .external_lex_state = 3},
  [8138] = {.lex_state = 16, .external_lex_state = 3},
  [8139] = {.lex_state = 16, .external_lex_state = 3},
  [8140] = {.lex_state = 120, .external_lex_state = 3},
  [8141] = {.lex_state = 127, .external_lex_state = 3},
  [8142] = {.lex_state = 127, .external_lex_state = 3},
  [8143] = {.lex_state = 127, .external_lex_state = 3},
  [8144] = {.lex_state = 120, .external_lex_state = 3},
  [8145] = {.lex_state = 120, .external_lex_state = 3},
  [8146] = {.lex_state = 120, .external_lex_state = 3},
  [8147] = {.lex_state = 120, .external_lex_state = 3},
  [8148] = {.lex_state = 120, .external_lex_state = 3},
  [8149] = {.lex_state = 16, .external_lex_state = 3},
  [8150] = {.lex_state = 120, .external_lex_state = 3},
  [8151] = {.lex_state = 120, .external_lex_state = 3},
  [8152] = {.lex_state = 16, .external_lex_state = 3},
  [8153] = {.lex_state = 120, .external_lex_state = 3},
  [8154] = {.lex_state = 120, .external_lex_state = 2},
  [8155] = {.lex_state = 127, .external_lex_state = 3},
  [8156] = {.lex_state = 127, .external_lex_state = 3},
  [8157] = {.lex_state = 127, .external_lex_state = 3},
  [8158] = {.lex_state = 120, .external_lex_state = 3},
  [8159] = {.lex_state = 16, .external_lex_state = 3},
  [8160] = {.lex_state = 120, .external_lex_state = 2},
  [8161] = {.lex_state = 120, .external_lex_state = 3},
  [8162] = {.lex_state = 120, .external_lex_state = 3},
  [8163] = {.lex_state = 120, .external_lex_state = 3},
  [8164] = {.lex_state = 120, .external_lex_state = 3},
  [8165] = {.lex_state = 120, .external_lex_state = 3},
  [8166] = {.lex_state = 16, .external_lex_state = 3},
  [8167] = {.lex_state = 120, .external_lex_state = 3},
  [8168] = {.lex_state = 16, .external_lex_state = 3},
  [8169] = {.lex_state = 127, .external_lex_state = 3},
  [8170] = {.lex_state = 127, .external_lex_state = 3},
  [8171] = {.lex_state = 127, .external_lex_state = 3},
  [8172] = {.lex_state = 120, .external_lex_state = 3},
  [8173] = {.lex_state = 120, .external_lex_state = 3},
  [8174] = {.lex_state = 120, .external_lex_state = 3},
  [8175] = {.lex_state = 120, .external_lex_state = 3},
  [8176] = {.lex_state = 127, .external_lex_state = 3},
  [8177] = {.lex_state = 127, .external_lex_state = 3},
  [8178] = {.lex_state = 127, .external_lex_state = 3},
  [8179] = {.lex_state = 120, .external_lex_state = 3},
  [8180] = {.lex_state = 120, .external_lex_state = 3},
  [8181] = {.lex_state = 127, .external_lex_state = 3},
  [8182] = {.lex_state = 127, .external_lex_state = 3},
  [8183] = {.lex_state = 127, .external_lex_state = 3},
  [8184] = {.lex_state = 16, .external_lex_state = 3},
  [8185] = {.lex_state = 120, .external_lex_state = 3},
  [8186] = {.lex_state = 127, .external_lex_state = 3},
  [8187] = {.lex_state = 127, .external_lex_state = 3},
  [8188] = {.lex_state = 127, .external_lex_state = 3},
  [8189] = {.lex_state = 120, .external_lex_state = 3},
  [8190] = {.lex_state = 120, .external_lex_state = 3},
  [8191] = {.lex_state = 127, .external_lex_state = 3},
  [8192] = {.lex_state = 127, .external_lex_state = 3},
  [8193] = {.lex_state = 127, .external_lex_state = 3},
  [8194] = {.lex_state = 16, .external_lex_state = 3},
  [8195] = {.lex_state = 120, .external_lex_state = 3},
  [8196] = {.lex_state = 127, .external_lex_state = 3},
  [8197] = {.lex_state = 127, .external_lex_state = 3},
  [8198] = {.lex_state = 127, .external_lex_state = 3},
  [8199] = {.lex_state = 120, .external_lex_state = 3},
  [8200] = {.lex_state = 120, .external_lex_state = 3},
  [8201] = {.lex_state = 127, .external_lex_state = 3},
  [8202] = {.lex_state = 127, .external_lex_state = 3},
  [8203] = {.lex_state = 127, .external_lex_state = 3},
  [8204] = {.lex_state = 120, .external_lex_state = 3},
  [8205] = {.lex_state = 16, .external_lex_state = 3},
  [8206] = {.lex_state = 127, .external_lex_state = 3},
  [8207] = {.lex_state = 127, .external_lex_state = 3},
  [8208] = {.lex_state = 127, .external_lex_state = 3},
  [8209] = {.lex_state = 120, .external_lex_state = 3},
  [8210] = {.lex_state = 120, .external_lex_state = 3},
  [8211] = {.lex_state = 120, .external_lex_state = 3},
  [8212] = {.lex_state = 120, .external_lex_state = 3},
  [8213] = {.lex_state = 120, .external_lex_state = 3},
  [8214] = {.lex_state = 120, .external_lex_state = 3},
  [8215] = {.lex_state = 65, .external_lex_state = 3},
  [8216] = {.lex_state = 16, .external_lex_state = 3},
  [8217] = {.lex_state = 120, .external_lex_state = 3},
  [8218] = {.lex_state = 120, .external_lex_state = 3},
  [8219] = {.lex_state = 120, .external_lex_state = 3},
  [8220] = {.lex_state = 120, .external_lex_state = 3},
  [8221] = {.lex_state = 120, .external_lex_state = 3},
  [8222] = {.lex_state = 120, .external_lex_state = 3},
  [8223] = {.lex_state = 120, .external_lex_state = 3},
  [8224] = {.lex_state = 120, .external_lex_state = 3},
  [8225] = {.lex_state = 120, .external_lex_state = 3},
  [8226] = {.lex_state = 120, .external_lex_state = 3},
  [8227] = {.lex_state = 120, .external_lex_state = 3},
  [8228] = {.lex_state = 120, .external_lex_state = 3},
  [8229] = {.lex_state = 127, .external_lex_state = 3},
  [8230] = {.lex_state = 120, .external_lex_state = 3},
  [8231] = {.lex_state = 120, .external_lex_state = 3},
  [8232] = {.lex_state = 120, .external_lex_state = 3},
  [8233] = {.lex_state = 120, .external_lex_state = 3},
  [8234] = {.lex_state = 120, .external_lex_state = 3},
  [8235] = {.lex_state = 16, .external_lex_state = 3},
  [8236] = {.lex_state = 120, .external_lex_state = 3},
  [8237] = {.lex_state = 120, .external_lex_state = 3},
  [8238] = {.lex_state = 120, .external_lex_state = 3},
  [8239] = {.lex_state = 120, .external_lex_state = 3},
  [8240] = {.lex_state = 16, .external_lex_state = 3},
  [8241] = {.lex_state = 16, .external_lex_state = 3},
  [8242] = {.lex_state = 120, .external_lex_state = 3},
  [8243] = {.lex_state = 120, .external_lex_state = 3},
  [8244] = {.lex_state = 120, .external_lex_state = 3},
  [8245] = {.lex_state = 120, .external_lex_state = 3},
  [8246] = {.lex_state = 120, .external_lex_state = 3},
  [8247] = {.lex_state = 120, .external_lex_state = 3},
  [8248] = {.lex_state = 120, .external_lex_state = 3},
  [8249] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [8250] = {.lex_state = 16, .external_lex_state = 3},
  [8251] = {.lex_state = 120, .external_lex_state = 2},
  [8252] = {.lex_state = 120, .external_lex_state = 3},
  [8253] = {.lex_state = 120, .external_lex_state = 3},
  [8254] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [8255] = {.lex_state = 120, .external_lex_state = 3},
  [8256] = {.lex_state = 120, .external_lex_state = 3},
  [8257] = {.lex_state = 16, .external_lex_state = 3},
  [8258] = {.lex_state = 120, .external_lex_state = 3},
  [8259] = {.lex_state = 120, .external_lex_state = 3},
  [8260] = {.lex_state = 120, .external_lex_state = 3},
  [8261] = {.lex_state = 120, .external_lex_state = 3},
  [8262] = {.lex_state = 120, .external_lex_state = 3},
  [8263] = {.lex_state = 120, .external_lex_state = 3},
  [8264] = {.lex_state = 120, .external_lex_state = 3},
  [8265] = {.lex_state = 120, .external_lex_state = 3},
  [8266] = {.lex_state = 65, .external_lex_state = 3},
  [8267] = {.lex_state = 120, .external_lex_state = 3},
  [8268] = {.lex_state = 16, .external_lex_state = 3},
  [8269] = {.lex_state = 120, .external_lex_state = 3},
  [8270] = {.lex_state = 120, .external_lex_state = 3},
  [8271] = {.lex_state = 65, .external_lex_state = 3},
  [8272] = {.lex_state = 121, .external_lex_state = 3},
  [8273] = {.lex_state = 16, .external_lex_state = 3},
  [8274] = {.lex_state = 65, .external_lex_state = 3},
  [8275] = {.lex_state = 65, .external_lex_state = 3},
  [8276] = {.lex_state = 65, .external_lex_state = 3},
  [8277] = {.lex_state = 65, .external_lex_state = 3},
  [8278] = {.lex_state = 65, .external_lex_state = 3},
  [8279] = {.lex_state = 65, .external_lex_state = 3},
  [8280] = {.lex_state = 65, .external_lex_state = 3},
  [8281] = {.lex_state = 65, .external_lex_state = 3},
  [8282] = {.lex_state = 65, .external_lex_state = 3},
  [8283] = {.lex_state = 65, .external_lex_state = 3},
  [8284] = {.lex_state = 65, .external_lex_state = 3},
  [8285] = {.lex_state = 65, .external_lex_state = 3},
  [8286] = {.lex_state = 65, .external_lex_state = 3},
  [8287] = {.lex_state = 65, .external_lex_state = 3},
  [8288] = {.lex_state = 65, .external_lex_state = 3},
  [8289] = {.lex_state = 65, .external_lex_state = 3},
  [8290] = {.lex_state = 120, .external_lex_state = 3},
  [8291] = {.lex_state = 120, .external_lex_state = 3},
  [8292] = {.lex_state = 120, .external_lex_state = 3},
  [8293] = {.lex_state = 16, .external_lex_state = 3},
  [8294] = {.lex_state = 16, .external_lex_state = 3},
  [8295] = {.lex_state = 13, .external_lex_state = 3},
  [8296] = {.lex_state = 120, .external_lex_state = 3},
  [8297] = {.lex_state = 16, .external_lex_state = 3},
  [8298] = {.lex_state = 120, .external_lex_state = 3},
  [8299] = {.lex_state = 120, .external_lex_state = 3},
  [8300] = {.lex_state = 120, .external_lex_state = 3},
  [8301] = {.lex_state = 120, .external_lex_state = 3},
  [8302] = {.lex_state = 120, .external_lex_state = 3},
  [8303] = {.lex_state = 120, .external_lex_state = 3},
  [8304] = {.lex_state = 120, .external_lex_state = 3},
  [8305] = {.lex_state = 120, .external_lex_state = 3},
  [8306] = {.lex_state = 120, .external_lex_state = 3},
  [8307] = {.lex_state = 120, .external_lex_state = 3},
  [8308] = {.lex_state = 120, .external_lex_state = 3},
  [8309] = {.lex_state = 120, .external_lex_state = 3},
  [8310] = {.lex_state = 120, .external_lex_state = 3},
  [8311] = {.lex_state = 120, .external_lex_state = 3},
  [8312] = {.lex_state = 120, .external_lex_state = 3},
  [8313] = {.lex_state = 13, .external_lex_state = 3},
  [8314] = {.lex_state = 120, .external_lex_state = 3},
  [8315] = {.lex_state = 16, .external_lex_state = 3},
  [8316] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 1},
  [8317] = {.lex_state = 16, .external_lex_state = 3},
  [8318] = {.lex_state = 120, .external_lex_state = 3},
  [8319] = {.lex_state = 120, .external_lex_state = 3},
  [8320] = {.lex_state = 120, .external_lex_state = 3},
  [8321] = {.lex_state = 120, .external_lex_state = 3},
  [8322] = {.lex_state = 120, .external_lex_state = 3},
  [8323] = {.lex_state = 120, .external_lex_state = 3},
  [8324] = {.lex_state = 120, .external_lex_state = 3},
  [8325] = {.lex_state = 120, .external_lex_state = 3},
  [8326] = {.lex_state = 120, .external_lex_state = 3},
  [8327] = {.lex_state = 120, .external_lex_state = 3},
  [8328] = {.lex_state = 120, .external_lex_state = 3},
  [8329] = {.lex_state = 120, .external_lex_state = 3},
  [8330] = {.lex_state = 16, .external_lex_state = 3},
  [8331] = {.lex_state = 120, .external_lex_state = 3},
  [8332] = {.lex_state = 120, .external_lex_state = 3},
  [8333] = {.lex_state = 120, .external_lex_state = 3},
  [8334] = {.lex_state = 120, .external_lex_state = 3},
  [8335] = {.lex_state = 120, .external_lex_state = 2},
  [8336] = {.lex_state = 120, .external_lex_state = 3},
  [8337] = {.lex_state = 120, .external_lex_state = 3},
  [8338] = {.lex_state = 120, .external_lex_state = 3},
  [8339] = {.lex_state = 120, .external_lex_state = 3},
  [8340] = {.lex_state = 120, .external_lex_state = 3},
  [8341] = {.lex_state = 127, .external_lex_state = 3},
  [8342] = {.lex_state = 127, .external_lex_state = 3},
  [8343] = {.lex_state = 120, .external_lex_state = 3},
  [8344] = {.lex_state = 120, .external_lex_state = 3},
  [8345] = {.lex_state = 127, .external_lex_state = 3},
  [8346] = {.lex_state = 120, .external_lex_state = 3},
  [8347] = {.lex_state = 127, .external_lex_state = 3},
  [8348] = {.lex_state = 120, .external_lex_state = 3},
  [8349] = {.lex_state = 120, .external_lex_state = 3},
  [8350] = {.lex_state = 120, .external_lex_state = 3},
  [8351] = {.lex_state = 120, .external_lex_state = 3},
  [8352] = {.lex_state = 120, .external_lex_state = 3},
  [8353] = {.lex_state = 120, .external_lex_state = 3},
  [8354] = {(TSStateId)(-1),},
  [8355] = {(TSStateId)(-1),},
};

static const TSSymbol ts_reserved_words[99][MAX_RESERVED_WORD_SET_SIZE] = {
  [1] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [2] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [3] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [4] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [5] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [6] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [7] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [8] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [9] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [10] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_as,
    anon_sym_then,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [11] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [12] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_as,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [13] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [14] = {
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [15] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_when,
  },
  [16] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [17] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [18] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [19] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [20] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_while,
    anon_sym_for,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [21] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_while,
    anon_sym_for,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [22] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [23] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [24] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [25] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [26] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [27] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [28] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [29] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [30] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [31] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
  },
  [32] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [33] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [34] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
  },
  [35] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [36] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [37] = {
    anon_sym_let,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_include,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [38] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [39] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [40] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [41] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [42] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [43] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [44] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [45] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [46] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [47] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [48] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [49] = {
    anon_sym_let,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [50] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [51] = {
    anon_sym_let,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [52] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [53] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [54] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [55] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [56] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [57] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_for,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [58] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [59] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [60] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [61] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [62] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_type,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [63] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [64] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [65] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [66] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [67] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [68] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [69] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [70] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [71] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [72] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [73] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [74] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [75] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [76] = {
    anon_sym_let,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_with,
    anon_sym_object,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [77] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [78] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [79] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [80] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [81] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [82] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [83] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [84] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [85] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [86] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [87] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [88] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [89] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_end,
    anon_sym_with,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [90] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [91] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [92] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [93] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [94] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [95] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [96] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [97] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [98] = {
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
  },
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [sym_attribute] = STATE(0),
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym__lowercase_identifier] = ACTIONS(1),
    [sym_shebang] = ACTIONS(1),
    [anon_sym_SEMI_SEMI] = ACTIONS(1),
    [anon_sym_let] = ACTIONS(1),
    [anon_sym_rec] = ACTIONS(1),
    [anon_sym_and] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_external] = ACTIONS(1),
    [anon_sym_type] = ACTIONS(1),
    [anon_sym_nonrec] = ACTIONS(1),
    [anon_sym_COLON_EQ] = ACTIONS(1),
    [anon_sym_private] = ACTIONS(1),
    [anon_sym_DOT_DOT] = ACTIONS(1),
    [anon_sym_PLUS_EQ] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym__] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_of] = ACTIONS(1),
    [anon_sym_COLON2] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_true] = ACTIONS(1),
    [anon_sym_false] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_mutable] = ACTIONS(1),
    [anon_sym_constraint] = ACTIONS(1),
    [anon_sym_exception] = ACTIONS(1),
    [anon_sym_module] = ACTIONS(1),
    [anon_sym_open] = ACTIONS(1),
    [anon_sym_include] = ACTIONS(1),
    [anon_sym_class] = ACTIONS(1),
    [anon_sym_virtual] = ACTIONS(1),
    [anon_sym_val] = ACTIONS(1),
    [anon_sym_sig] = ACTIONS(1),
    [anon_sym_end] = ACTIONS(1),
    [anon_sym_with] = ACTIONS(1),
    [anon_sym_functor] = ACTIONS(1),
    [anon_sym_struct] = ACTIONS(1),
    [anon_sym_COLON_GT] = ACTIONS(1),
    [anon_sym_object] = ACTIONS(1),
    [anon_sym_inherit] = ACTIONS(1),
    [anon_sym_method] = ACTIONS(1),
    [anon_sym_in] = ACTIONS(1),
    [anon_sym_fun] = ACTIONS(1),
    [anon_sym_as] = ACTIONS(1),
    [anon_sym_initializer] = ACTIONS(1),
    [anon_sym_LBRACK_GT] = ACTIONS(1),
    [anon_sym_LBRACK_LT] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_POUND] = ACTIONS(1),
    [anon_sym_COLON_COLON] = ACTIONS(1),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1),
    [anon_sym_LT_DASH] = ACTIONS(1),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_then] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_while] = ACTIONS(1),
    [anon_sym_do] = ACTIONS(1),
    [anon_sym_done] = ACTIONS(1),
    [anon_sym_for] = ACTIONS(1),
    [anon_sym_to] = ACTIONS(1),
    [anon_sym_downto] = ACTIONS(1),
    [anon_sym_match] = ACTIONS(1),
    [anon_sym_when] = ACTIONS(1),
    [anon_sym_function] = ACTIONS(1),
    [anon_sym_try] = ACTIONS(1),
    [anon_sym_assert] = ACTIONS(1),
    [anon_sym_lazy] = ACTIONS(1),
    [anon_sym_new] = ACTIONS(1),
    [anon_sym_LBRACE_LT] = ACTIONS(1),
    [anon_sym_begin] = ACTIONS(1),
    [sym_ocamlyacc_value] = ACTIONS(1),
    [anon_sym_effect] = ACTIONS(1),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [aux_sym_number_token1] = ACTIONS(1),
    [aux_sym_signed_number_token1] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_SQUOTE2] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [aux_sym_string_content_token2] = ACTIONS(1),
    [aux_sym_string_content_token4] = ACTIONS(1),
    [aux_sym_string_content_token5] = ACTIONS(1),
    [aux_sym_escape_sequence_token1] = ACTIONS(1),
    [aux_sym_escape_sequence_token2] = ACTIONS(1),
    [aux_sym_escape_sequence_token3] = ACTIONS(1),
    [aux_sym_escape_sequence_token4] = ACTIONS(1),
    [sym_pretty_printing_indication] = ACTIONS(1),
    [aux_sym_sign_operator_token1] = ACTIONS(1),
    [sym_hash_operator] = ACTIONS(1),
    [anon_sym_lsl] = ACTIONS(1),
    [anon_sym_lsr] = ACTIONS(1),
    [anon_sym_asr] = ACTIONS(1),
    [aux_sym_pow_operator_token1] = ACTIONS(1),
    [anon_sym_mod] = ACTIONS(1),
    [anon_sym_land] = ACTIONS(1),
    [anon_sym_lor] = ACTIONS(1),
    [anon_sym_lxor] = ACTIONS(1),
    [sym_concat_operator] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_or] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [sym_assign_operator] = ACTIONS(1),
    [sym_let_operator] = ACTIONS(1),
    [sym_let_and_operator] = ACTIONS(1),
    [sym_match_operator] = ACTIONS(1),
    [sym__uppercase_identifier] = ACTIONS(1),
    [aux_sym_directive_token1] = ACTIONS(1),
    [aux_sym_type_variable_token1] = ACTIONS(1),
    [aux_sym_tag_token1] = ACTIONS(1),
    [aux_sym_attribute_id_token1] = ACTIONS(1),
    [sym_comment] = ACTIONS(5),
    [sym__left_quoted_string_delimiter] = ACTIONS(1),
    [sym__right_quoted_string_delimiter] = ACTIONS(1),
    [sym_line_number_directive] = ACTIONS(5),
    [sym__null] = ACTIONS(1),
    [sym__error_sentinel] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_compilation_unit] = STATE(7922),
    [sym__structure] = STATE(7387),
    [sym_expression_item] = STATE(3221),
    [sym_toplevel_directive] = STATE(3221),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3221),
    [sym_value_definition] = STATE(4823),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2832),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(1),
    [sym_floating_attribute] = STATE(4781),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1471),
    [aux_sym__structure_repeat1] = STATE(51),
    [ts_builtin_sym_end] = ACTIONS(7),
    [sym__lowercase_identifier] = ACTIONS(9),
    [sym_shebang] = ACTIONS(11),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(15),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(75),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(77),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(95),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_directive_token1] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(2)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2084),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2137),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(2),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__attribute] = STATE(628),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_SEMI_SEMI] = ACTIONS(107),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_and] = ACTIONS(111),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_external] = ACTIONS(111),
    [anon_sym_type] = ACTIONS(111),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_PIPE] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_RBRACK] = ACTIONS(107),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_SEMI] = ACTIONS(111),
    [anon_sym_exception] = ACTIONS(111),
    [anon_sym_module] = ACTIONS(111),
    [anon_sym_open] = ACTIONS(111),
    [anon_sym_include] = ACTIONS(111),
    [anon_sym_class] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_end] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_in] = ACTIONS(111),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_let_and_operator] = ACTIONS(107),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_directive_token1] = ACTIONS(107),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(3)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2140),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1873),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(3),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__attribute] = STATE(622),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_SEMI_SEMI] = ACTIONS(107),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_and] = ACTIONS(111),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_external] = ACTIONS(111),
    [anon_sym_type] = ACTIONS(111),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_PIPE] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_RBRACK] = ACTIONS(107),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_SEMI] = ACTIONS(111),
    [anon_sym_exception] = ACTIONS(111),
    [anon_sym_module] = ACTIONS(111),
    [anon_sym_open] = ACTIONS(111),
    [anon_sym_include] = ACTIONS(111),
    [anon_sym_class] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_end] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_let_and_operator] = ACTIONS(107),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_directive_token1] = ACTIONS(107),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(4)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2128),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2137),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(4),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__attribute] = STATE(640),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [ts_builtin_sym_end] = ACTIONS(107),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_SEMI_SEMI] = ACTIONS(107),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_and] = ACTIONS(111),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_external] = ACTIONS(111),
    [anon_sym_type] = ACTIONS(111),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_PIPE] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_SEMI] = ACTIONS(111),
    [anon_sym_exception] = ACTIONS(111),
    [anon_sym_module] = ACTIONS(111),
    [anon_sym_open] = ACTIONS(111),
    [anon_sym_include] = ACTIONS(111),
    [anon_sym_class] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_in] = ACTIONS(111),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(217),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(219),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_let_and_operator] = ACTIONS(107),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_directive_token1] = ACTIONS(107),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(5)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2199),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1873),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(5),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__attribute] = STATE(620),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [ts_builtin_sym_end] = ACTIONS(107),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_SEMI_SEMI] = ACTIONS(107),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_and] = ACTIONS(111),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_external] = ACTIONS(111),
    [anon_sym_type] = ACTIONS(111),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_PIPE] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_SEMI] = ACTIONS(111),
    [anon_sym_exception] = ACTIONS(111),
    [anon_sym_module] = ACTIONS(111),
    [anon_sym_open] = ACTIONS(111),
    [anon_sym_include] = ACTIONS(111),
    [anon_sym_class] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(217),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(219),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_let_and_operator] = ACTIONS(107),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_directive_token1] = ACTIONS(107),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(6)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2630),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1873),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(6),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__attribute] = STATE(634),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(107),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(111),
    [anon_sym_type] = ACTIONS(111),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_PIPE] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(107),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_SEMI] = ACTIONS(111),
    [anon_sym_exception] = ACTIONS(111),
    [anon_sym_module] = ACTIONS(111),
    [anon_sym_open] = ACTIONS(111),
    [anon_sym_include] = ACTIONS(111),
    [anon_sym_class] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_end] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(107),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(7)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2669),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1873),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(7),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__attribute] = STATE(621),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [ts_builtin_sym_end] = ACTIONS(107),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_SEMI_SEMI] = ACTIONS(107),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_external] = ACTIONS(111),
    [anon_sym_type] = ACTIONS(111),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_PIPE] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_SEMI] = ACTIONS(111),
    [anon_sym_exception] = ACTIONS(111),
    [anon_sym_module] = ACTIONS(111),
    [anon_sym_open] = ACTIONS(111),
    [anon_sym_include] = ACTIONS(111),
    [anon_sym_class] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(75),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(77),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_directive_token1] = ACTIONS(107),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(8)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(8),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(8189),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(309),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(9)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(9),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7607),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(317),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(10)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(10),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(8259),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(319),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(11)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(11),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(8258),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(321),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(12)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(12),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(8260),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(323),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(13)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(13),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(8292),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(325),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(14)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(14),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7403),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(327),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(15)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(15),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(8352),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(329),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(16)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(16),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(8246),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(331),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(17)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(17),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7739),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(333),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(18)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(18),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7994),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(335),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(19)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(19),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7790),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(337),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(20)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(20),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(8019),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(339),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(21)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(21),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(8036),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(341),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(22)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(22),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7976),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(343),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(23)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(23),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7853),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(24)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(24),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7995),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(347),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(25)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(25),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7400),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(349),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(26)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(26),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7752),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(351),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(27)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(27),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7935),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(353),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(28)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(28),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(8161),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(29)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(29),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7284),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(357),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(30)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(30),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(8236),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(359),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(31)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(31),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7668),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(361),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(32)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(32),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(8337),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(33)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(33),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7472),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(365),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(34)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(34),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7767),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(367),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(35)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(35),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7985),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(369),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(36)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(36),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(8173),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(371),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(37)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(37),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7243),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(373),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(38)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(38),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7260),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(375),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(39)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(39),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7285),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(377),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(40)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(40),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7301),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(379),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(41)] = {
    [sym__structure] = STATE(7977),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(41),
    [sym_floating_attribute] = STATE(4781),
    [sym_attribute_payload] = STATE(7968),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_QMARK] = ACTIONS(305),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_COLON2] = ACTIONS(307),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(381),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(42)] = {
    [sym__structure] = STATE(7698),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(42),
    [sym_floating_attribute] = STATE(4781),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_end] = ACTIONS(383),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(43)] = {
    [sym_expression_item] = STATE(4799),
    [sym_toplevel_directive] = STATE(4799),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(4799),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(43),
    [sym_floating_attribute] = STATE(4781),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(2036),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(385),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_end] = ACTIONS(387),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(44)] = {
    [sym__structure] = STATE(7936),
    [sym_expression_item] = STATE(3183),
    [sym_toplevel_directive] = STATE(3183),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3183),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(44),
    [sym_floating_attribute] = STATE(4781),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(46),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_end] = ACTIONS(389),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(45)] = {
    [sym_expression_item] = STATE(4799),
    [sym_toplevel_directive] = STATE(4799),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(4799),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(45),
    [sym_floating_attribute] = STATE(4781),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(2036),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(391),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_end] = ACTIONS(393),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(46)] = {
    [sym_expression_item] = STATE(3179),
    [sym_toplevel_directive] = STATE(3179),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3179),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(46),
    [sym_floating_attribute] = STATE(4781),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(2036),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(395),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_end] = ACTIONS(397),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(47)] = {
    [sym_expression_item] = STATE(4799),
    [sym_toplevel_directive] = STATE(4799),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(4799),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(47),
    [sym_floating_attribute] = STATE(4781),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(2036),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(399),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_end] = ACTIONS(401),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(48)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3075),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2754),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(48),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__attribute] = STATE(624),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_PIPE] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_SEMI] = ACTIONS(107),
    [anon_sym_constraint] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_end] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_inherit] = ACTIONS(111),
    [anon_sym_method] = ACTIONS(111),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_initializer] = ACTIONS(111),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_do] = ACTIONS(111),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(445),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(447),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(49)] = {
    [sym__structure] = STATE(8247),
    [sym_expression_item] = STATE(3221),
    [sym_toplevel_directive] = STATE(3221),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3221),
    [sym_value_definition] = STATE(4823),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2832),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(49),
    [sym_floating_attribute] = STATE(4781),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1471),
    [aux_sym__structure_repeat1] = STATE(51),
    [ts_builtin_sym_end] = ACTIONS(463),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(15),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(75),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(77),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(95),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(50)] = {
    [sym_expression_item] = STATE(4799),
    [sym_toplevel_directive] = STATE(4799),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(4799),
    [sym_value_definition] = STATE(4823),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2832),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(50),
    [sym_floating_attribute] = STATE(4781),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1471),
    [aux_sym__structure_repeat1] = STATE(2036),
    [ts_builtin_sym_end] = ACTIONS(399),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(15),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(75),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(77),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(95),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(51)] = {
    [sym_expression_item] = STATE(3215),
    [sym_toplevel_directive] = STATE(3215),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(3215),
    [sym_value_definition] = STATE(4823),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2832),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(51),
    [sym_floating_attribute] = STATE(4781),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1471),
    [aux_sym__structure_repeat1] = STATE(2036),
    [ts_builtin_sym_end] = ACTIONS(395),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(15),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(75),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(77),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(95),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(52)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3089),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2754),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(52),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__attribute] = STATE(644),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_RPAREN] = ACTIONS(107),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_PIPE] = ACTIONS(111),
    [anon_sym_COLON2] = ACTIONS(111),
    [anon_sym_DASH_GT] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_RBRACK] = ACTIONS(107),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_SEMI] = ACTIONS(107),
    [anon_sym_RBRACE] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(111),
    [anon_sym_with] = ACTIONS(111),
    [anon_sym_COLON_GT] = ACTIONS(107),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_then] = ACTIONS(111),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_GT_RBRACE] = ACTIONS(107),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(53)] = {
    [sym_expression_item] = STATE(4799),
    [sym_toplevel_directive] = STATE(4799),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(4799),
    [sym_value_definition] = STATE(4823),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2832),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(53),
    [sym_floating_attribute] = STATE(4781),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1471),
    [aux_sym__structure_repeat1] = STATE(2036),
    [ts_builtin_sym_end] = ACTIONS(391),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(15),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(75),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(77),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(95),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(54)] = {
    [sym_expression_item] = STATE(4799),
    [sym_toplevel_directive] = STATE(4799),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(4799),
    [sym_value_definition] = STATE(4823),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2832),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(54),
    [sym_floating_attribute] = STATE(4781),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1471),
    [aux_sym__structure_repeat1] = STATE(2036),
    [ts_builtin_sym_end] = ACTIONS(385),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(15),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(75),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(77),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(95),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(55)] = {
    [sym_expression_item] = STATE(4799),
    [sym_toplevel_directive] = STATE(4799),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(4799),
    [sym_value_definition] = STATE(4798),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2742),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(55),
    [sym_floating_attribute] = STATE(4781),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1364),
    [aux_sym__structure_repeat1] = STATE(2036),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(303),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(311),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(313),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(56)] = {
    [sym_expression_item] = STATE(4799),
    [sym_toplevel_directive] = STATE(4799),
    [sym__local_structure_item] = STATE(4775),
    [sym__structure_item] = STATE(4799),
    [sym_value_definition] = STATE(4823),
    [sym_external] = STATE(4781),
    [sym_type_definition] = STATE(4781),
    [sym_exception_definition] = STATE(4781),
    [sym_module_definition] = STATE(4781),
    [sym_module_type_definition] = STATE(4781),
    [sym_open_module] = STATE(4781),
    [sym_include_module] = STATE(4775),
    [sym_class_definition] = STATE(4781),
    [sym_class_type_definition] = STATE(4781),
    [sym_value_specification] = STATE(4775),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2832),
    [sym__sequence_expression] = STATE(4469),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(56),
    [sym_floating_attribute] = STATE(4781),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym_item_extension] = STATE(4781),
    [sym_quoted_item_extension] = STATE(4781),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_directive] = STATE(3205),
    [sym_tag] = STATE(1471),
    [aux_sym__structure_repeat1] = STATE(2036),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_SEMI_SEMI] = ACTIONS(13),
    [anon_sym_let] = ACTIONS(15),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_external] = ACTIONS(21),
    [anon_sym_type] = ACTIONS(23),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_exception] = ACTIONS(31),
    [anon_sym_module] = ACTIONS(33),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(37),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_val] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(75),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(77),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(95),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_directive_token1] = ACTIONS(315),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(57)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3126),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(2754),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(57),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(625),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_SEMI] = ACTIONS(107),
    [anon_sym_constraint] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_end] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_inherit] = ACTIONS(111),
    [anon_sym_method] = ACTIONS(111),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_initializer] = ACTIONS(111),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(527),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(529),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(58)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3134),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3080),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(58),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(636),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(107),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_COLON2] = ACTIONS(111),
    [anon_sym_DASH_GT] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(107),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_SEMI] = ACTIONS(107),
    [anon_sym_RBRACE] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(111),
    [anon_sym_with] = ACTIONS(111),
    [anon_sym_COLON_GT] = ACTIONS(107),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_then] = ACTIONS(111),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_GT_RBRACE] = ACTIONS(107),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(59)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8084),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(59),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7699),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(563),
    [anon_sym_DOT] = ACTIONS(565),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(567),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(571),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(575),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(577),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(593),
    [sym_let_and_operator] = ACTIONS(577),
    [sym_match_operator] = ACTIONS(595),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(60)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(5752),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(60),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7473),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(597),
    [anon_sym_DOT] = ACTIONS(599),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(601),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(603),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(605),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(607),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(609),
    [sym_let_and_operator] = ACTIONS(607),
    [sym_match_operator] = ACTIONS(611),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(61)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(5672),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(61),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7872),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(613),
    [anon_sym_DOT] = ACTIONS(615),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(617),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(619),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(621),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(623),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(625),
    [sym_let_and_operator] = ACTIONS(623),
    [sym_match_operator] = ACTIONS(627),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(62)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(5767),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(62),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(8135),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(629),
    [anon_sym_DOT] = ACTIONS(631),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(633),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(635),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(637),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(639),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(641),
    [sym_let_and_operator] = ACTIONS(639),
    [sym_match_operator] = ACTIONS(643),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(63)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7897),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(63),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7863),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(645),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(649),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(651),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(653),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(655),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(657),
    [sym_let_and_operator] = ACTIONS(655),
    [sym_match_operator] = ACTIONS(659),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(64)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(5764),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(64),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7863),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(645),
    [anon_sym_DOT] = ACTIONS(647),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(649),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(651),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(653),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(655),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(657),
    [sym_let_and_operator] = ACTIONS(655),
    [sym_match_operator] = ACTIONS(659),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(65)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8032),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(65),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(8322),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(661),
    [anon_sym_DOT] = ACTIONS(663),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(665),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(667),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(669),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(671),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(673),
    [sym_let_and_operator] = ACTIONS(671),
    [sym_match_operator] = ACTIONS(675),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(66)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(5555),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(66),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(8322),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(661),
    [anon_sym_DOT] = ACTIONS(663),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(665),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(667),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(669),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(671),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(673),
    [sym_let_and_operator] = ACTIONS(671),
    [sym_match_operator] = ACTIONS(675),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(67)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7940),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(67),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(8135),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(629),
    [anon_sym_DOT] = ACTIONS(631),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(633),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(635),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(637),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(639),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(641),
    [sym_let_and_operator] = ACTIONS(639),
    [sym_match_operator] = ACTIONS(643),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(68)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8153),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(68),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(8061),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(677),
    [anon_sym_DOT] = ACTIONS(679),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(681),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(683),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(685),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(687),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(689),
    [sym_let_and_operator] = ACTIONS(687),
    [sym_match_operator] = ACTIONS(691),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(69)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(5689),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(69),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(8061),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(677),
    [anon_sym_DOT] = ACTIONS(679),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(681),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(683),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(685),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(687),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(689),
    [sym_let_and_operator] = ACTIONS(687),
    [sym_match_operator] = ACTIONS(691),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(70)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(5687),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(70),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7503),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(693),
    [anon_sym_DOT] = ACTIONS(695),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(697),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(699),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(701),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(703),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(705),
    [sym_let_and_operator] = ACTIONS(703),
    [sym_match_operator] = ACTIONS(707),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(71)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8025),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(71),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7503),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(693),
    [anon_sym_DOT] = ACTIONS(695),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(697),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(699),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(701),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(703),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(705),
    [sym_let_and_operator] = ACTIONS(703),
    [sym_match_operator] = ACTIONS(707),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(72)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(5764),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(72),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7464),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(709),
    [anon_sym_DOT] = ACTIONS(711),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(649),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(713),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(715),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(717),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(719),
    [sym_let_and_operator] = ACTIONS(717),
    [sym_match_operator] = ACTIONS(721),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(73)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8151),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(73),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(8112),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(723),
    [anon_sym_DOT] = ACTIONS(725),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(727),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(731),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(733),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(735),
    [sym_let_and_operator] = ACTIONS(733),
    [sym_match_operator] = ACTIONS(737),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(74)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(5549),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(74),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(8112),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(723),
    [anon_sym_DOT] = ACTIONS(725),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(727),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(731),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(733),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(735),
    [sym_let_and_operator] = ACTIONS(733),
    [sym_match_operator] = ACTIONS(737),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(75)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7427),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(75),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7306),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(739),
    [anon_sym_DOT] = ACTIONS(741),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(743),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(745),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(747),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(749),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(751),
    [sym_let_and_operator] = ACTIONS(749),
    [sym_match_operator] = ACTIONS(753),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(76)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(5628),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(76),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7306),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(739),
    [anon_sym_DOT] = ACTIONS(741),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(743),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(745),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(747),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(749),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(751),
    [sym_let_and_operator] = ACTIONS(749),
    [sym_match_operator] = ACTIONS(753),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(77)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7897),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(77),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7464),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(709),
    [anon_sym_DOT] = ACTIONS(711),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(649),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(713),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(715),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(717),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(719),
    [sym_let_and_operator] = ACTIONS(717),
    [sym_match_operator] = ACTIONS(721),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(78)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(5665),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(78),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7699),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(563),
    [anon_sym_DOT] = ACTIONS(565),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(567),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(571),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(575),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(577),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(593),
    [sym_let_and_operator] = ACTIONS(577),
    [sym_match_operator] = ACTIONS(595),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(79)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8310),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(79),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7901),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(755),
    [anon_sym_DOT] = ACTIONS(757),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(759),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(761),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(763),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(765),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(767),
    [sym_let_and_operator] = ACTIONS(765),
    [sym_match_operator] = ACTIONS(769),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(80)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(5751),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(80),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7901),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(755),
    [anon_sym_DOT] = ACTIONS(757),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(759),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(761),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(763),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(765),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(767),
    [sym_let_and_operator] = ACTIONS(765),
    [sym_match_operator] = ACTIONS(769),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(81)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7294),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(81),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(8102),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(771),
    [anon_sym_DOT] = ACTIONS(773),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(775),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(777),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(779),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(781),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(783),
    [sym_let_and_operator] = ACTIONS(781),
    [sym_match_operator] = ACTIONS(785),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(82)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(5614),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(82),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(8102),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(771),
    [anon_sym_DOT] = ACTIONS(773),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(775),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(777),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(779),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(781),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(783),
    [sym_let_and_operator] = ACTIONS(781),
    [sym_match_operator] = ACTIONS(785),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(83)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7627),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(83),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7872),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(613),
    [anon_sym_DOT] = ACTIONS(615),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(617),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(619),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(621),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(623),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(625),
    [sym_let_and_operator] = ACTIONS(623),
    [sym_match_operator] = ACTIONS(627),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(84)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7889),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(84),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7341),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(787),
    [anon_sym_DOT] = ACTIONS(789),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(791),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(793),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(795),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(797),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(799),
    [sym_let_and_operator] = ACTIONS(797),
    [sym_match_operator] = ACTIONS(801),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(85)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(5659),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(85),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7341),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(787),
    [anon_sym_DOT] = ACTIONS(789),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(791),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(793),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(795),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(797),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(799),
    [sym_let_and_operator] = ACTIONS(797),
    [sym_match_operator] = ACTIONS(801),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(86)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7735),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(86),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__infix_operator] = STATE(7473),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(597),
    [anon_sym_DOT] = ACTIONS(599),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_module] = ACTIONS(601),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_COLON_COLON] = ACTIONS(603),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(573),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(605),
    [aux_sym_sign_operator_token1] = ACTIONS(573),
    [sym_hash_operator] = ACTIONS(607),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(609),
    [sym_let_and_operator] = ACTIONS(607),
    [sym_match_operator] = ACTIONS(611),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(87)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3415),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3582),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(87),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__attribute] = STATE(619),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_and] = ACTIONS(111),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_PIPE] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_SEMI] = ACTIONS(107),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_in] = ACTIONS(111),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_let_and_operator] = ACTIONS(107),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(88)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3491),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3080),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(88),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__attribute] = STATE(645),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_and] = ACTIONS(111),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_SEMI] = ACTIONS(107),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_in] = ACTIONS(111),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_let_and_operator] = ACTIONS(107),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(89)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3457),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3582),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(89),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__attribute] = STATE(643),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_PIPE] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_SEMI] = ACTIONS(107),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_done] = ACTIONS(111),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_to] = ACTIONS(111),
    [anon_sym_downto] = ACTIONS(111),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(90)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3548),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3582),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(90),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(639),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_SEMI] = ACTIONS(107),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_done] = ACTIONS(111),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_to] = ACTIONS(111),
    [anon_sym_downto] = ACTIONS(111),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(91)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3660),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3080),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(91),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(637),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_PIPE] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_SEMI] = ACTIONS(107),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(107),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(92)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3786),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3080),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(92),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(629),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_COMMA] = ACTIONS(107),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_SEMI] = ACTIONS(107),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_AMP] = ACTIONS(111),
    [anon_sym_COLON_COLON] = ACTIONS(107),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_else] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_do] = ACTIONS(111),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(151),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(85),
    [anon_sym_lsl] = ACTIONS(111),
    [anon_sym_lsr] = ACTIONS(111),
    [anon_sym_asr] = ACTIONS(111),
    [aux_sym_pow_operator_token1] = ACTIONS(107),
    [anon_sym_mod] = ACTIONS(111),
    [anon_sym_land] = ACTIONS(111),
    [anon_sym_lor] = ACTIONS(111),
    [anon_sym_lxor] = ACTIONS(111),
    [aux_sym_mult_operator_token1] = ACTIONS(111),
    [aux_sym_add_operator_token1] = ACTIONS(111),
    [sym_concat_operator] = ACTIONS(107),
    [sym_rel_operator] = ACTIONS(111),
    [anon_sym_AMP_AMP] = ACTIONS(111),
    [anon_sym_or] = ACTIONS(111),
    [anon_sym_PIPE_PIPE] = ACTIONS(111),
    [sym_assign_operator] = ACTIONS(107),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(93)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2222),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2137),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(93),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__attribute] = STATE(635),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_SEMI_SEMI] = ACTIONS(107),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_and] = ACTIONS(111),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_external] = ACTIONS(111),
    [anon_sym_type] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_RBRACK] = ACTIONS(107),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_exception] = ACTIONS(111),
    [anon_sym_module] = ACTIONS(111),
    [anon_sym_open] = ACTIONS(111),
    [anon_sym_include] = ACTIONS(111),
    [anon_sym_class] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_end] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_in] = ACTIONS(111),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_let_and_operator] = ACTIONS(107),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_directive_token1] = ACTIONS(107),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(94)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2538),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1873),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(94),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__attribute] = STATE(630),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_SEMI_SEMI] = ACTIONS(107),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_and] = ACTIONS(111),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_external] = ACTIONS(111),
    [anon_sym_type] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_RBRACK] = ACTIONS(107),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_exception] = ACTIONS(111),
    [anon_sym_module] = ACTIONS(111),
    [anon_sym_open] = ACTIONS(111),
    [anon_sym_include] = ACTIONS(111),
    [anon_sym_class] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_end] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_let_and_operator] = ACTIONS(107),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_directive_token1] = ACTIONS(107),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(95)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2612),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2137),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(95),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__attribute] = STATE(641),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [ts_builtin_sym_end] = ACTIONS(107),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_SEMI_SEMI] = ACTIONS(107),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_and] = ACTIONS(111),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_external] = ACTIONS(111),
    [anon_sym_type] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_exception] = ACTIONS(111),
    [anon_sym_module] = ACTIONS(111),
    [anon_sym_open] = ACTIONS(111),
    [anon_sym_include] = ACTIONS(111),
    [anon_sym_class] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_in] = ACTIONS(111),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(217),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(219),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_let_and_operator] = ACTIONS(107),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_directive_token1] = ACTIONS(107),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(96)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2671),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1873),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(96),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__attribute] = STATE(638),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [ts_builtin_sym_end] = ACTIONS(107),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_SEMI_SEMI] = ACTIONS(107),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_and] = ACTIONS(111),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_external] = ACTIONS(111),
    [anon_sym_type] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_exception] = ACTIONS(111),
    [anon_sym_module] = ACTIONS(111),
    [anon_sym_open] = ACTIONS(111),
    [anon_sym_include] = ACTIONS(111),
    [anon_sym_class] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(217),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(219),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_let_and_operator] = ACTIONS(107),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_directive_token1] = ACTIONS(107),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(97)] = {
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(942),
    [sym_typed_expression] = STATE(946),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_labeled_argument] = STATE(1023),
    [sym_prefix_expression] = STATE(946),
    [sym_hash_expression] = STATE(946),
    [sym_field_get_expression] = STATE(946),
    [sym_array_get_expression] = STATE(946),
    [sym_string_get_expression] = STATE(946),
    [sym_bigarray_get_expression] = STATE(946),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(97),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [aux_sym_class_application_repeat1] = STATE(102),
    [sym__lowercase_identifier] = ACTIONS(911),
    [anon_sym_TILDE] = ACTIONS(913),
    [anon_sym_QMARK] = ACTIONS(913),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(915),
    [anon_sym_COMMA] = ACTIONS(915),
    [anon_sym_PIPE] = ACTIONS(917),
    [anon_sym_COLON2] = ACTIONS(917),
    [anon_sym_DOT] = ACTIONS(919),
    [anon_sym_DASH_GT] = ACTIONS(917),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(915),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_SEMI] = ACTIONS(915),
    [anon_sym_RBRACE] = ACTIONS(915),
    [anon_sym_constraint] = ACTIONS(917),
    [anon_sym_val] = ACTIONS(917),
    [anon_sym_end] = ACTIONS(917),
    [anon_sym_with] = ACTIONS(917),
    [anon_sym_COLON_GT] = ACTIONS(915),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_inherit] = ACTIONS(917),
    [anon_sym_method] = ACTIONS(917),
    [anon_sym_initializer] = ACTIONS(917),
    [anon_sym_AMP] = ACTIONS(917),
    [anon_sym_POUND] = ACTIONS(921),
    [anon_sym_COLON_COLON] = ACTIONS(915),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(915),
    [anon_sym_then] = ACTIONS(917),
    [anon_sym_else] = ACTIONS(917),
    [anon_sym_do] = ACTIONS(917),
    [anon_sym_done] = ACTIONS(917),
    [anon_sym_to] = ACTIONS(917),
    [anon_sym_downto] = ACTIONS(917),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_GT_RBRACE] = ACTIONS(915),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(917),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(915),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(527),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(529),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(915),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(915),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(917),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(917),
    [sym_hash_operator] = ACTIONS(923),
    [anon_sym_lsl] = ACTIONS(917),
    [anon_sym_lsr] = ACTIONS(917),
    [anon_sym_asr] = ACTIONS(917),
    [aux_sym_pow_operator_token1] = ACTIONS(915),
    [anon_sym_mod] = ACTIONS(917),
    [anon_sym_land] = ACTIONS(917),
    [anon_sym_lor] = ACTIONS(917),
    [anon_sym_lxor] = ACTIONS(917),
    [aux_sym_mult_operator_token1] = ACTIONS(917),
    [aux_sym_add_operator_token1] = ACTIONS(917),
    [sym_concat_operator] = ACTIONS(915),
    [sym_rel_operator] = ACTIONS(917),
    [anon_sym_AMP_AMP] = ACTIONS(917),
    [anon_sym_or] = ACTIONS(917),
    [anon_sym_PIPE_PIPE] = ACTIONS(917),
    [sym_assign_operator] = ACTIONS(915),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(98)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2728),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1873),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(98),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__attribute] = STATE(623),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_SEMI_SEMI] = ACTIONS(107),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(111),
    [anon_sym_type] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(107),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(111),
    [anon_sym_module] = ACTIONS(111),
    [anon_sym_open] = ACTIONS(111),
    [anon_sym_include] = ACTIONS(111),
    [anon_sym_class] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_end] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(107),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(99)] = {
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(942),
    [sym_typed_expression] = STATE(946),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_labeled_argument] = STATE(1023),
    [sym_prefix_expression] = STATE(946),
    [sym_hash_expression] = STATE(946),
    [sym_field_get_expression] = STATE(946),
    [sym_array_get_expression] = STATE(946),
    [sym_string_get_expression] = STATE(946),
    [sym_bigarray_get_expression] = STATE(946),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(99),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [aux_sym_class_application_repeat1] = STATE(99),
    [sym__lowercase_identifier] = ACTIONS(925),
    [anon_sym_and] = ACTIONS(928),
    [anon_sym_TILDE] = ACTIONS(930),
    [anon_sym_QMARK] = ACTIONS(930),
    [anon_sym_LPAREN] = ACTIONS(933),
    [anon_sym_RPAREN] = ACTIONS(936),
    [anon_sym_COMMA] = ACTIONS(936),
    [anon_sym_PIPE] = ACTIONS(928),
    [anon_sym_COLON2] = ACTIONS(928),
    [anon_sym_DASH_GT] = ACTIONS(928),
    [anon_sym_LBRACK] = ACTIONS(938),
    [anon_sym_RBRACK] = ACTIONS(936),
    [anon_sym_true] = ACTIONS(941),
    [anon_sym_false] = ACTIONS(941),
    [anon_sym_LBRACE] = ACTIONS(944),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym_RBRACE] = ACTIONS(936),
    [anon_sym_constraint] = ACTIONS(928),
    [anon_sym_val] = ACTIONS(928),
    [anon_sym_end] = ACTIONS(928),
    [anon_sym_with] = ACTIONS(928),
    [anon_sym_COLON_GT] = ACTIONS(936),
    [anon_sym_object] = ACTIONS(947),
    [anon_sym_inherit] = ACTIONS(928),
    [anon_sym_method] = ACTIONS(928),
    [anon_sym_in] = ACTIONS(928),
    [anon_sym_as] = ACTIONS(928),
    [anon_sym_initializer] = ACTIONS(928),
    [anon_sym_AMP] = ACTIONS(928),
    [anon_sym_COLON_COLON] = ACTIONS(936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(950),
    [anon_sym_PIPE_RBRACK] = ACTIONS(936),
    [anon_sym_then] = ACTIONS(928),
    [anon_sym_else] = ACTIONS(928),
    [anon_sym_do] = ACTIONS(928),
    [anon_sym_done] = ACTIONS(928),
    [anon_sym_to] = ACTIONS(928),
    [anon_sym_downto] = ACTIONS(928),
    [anon_sym_new] = ACTIONS(953),
    [anon_sym_LBRACE_LT] = ACTIONS(956),
    [anon_sym_GT_RBRACE] = ACTIONS(936),
    [anon_sym_begin] = ACTIONS(959),
    [sym_ocamlyacc_value] = ACTIONS(962),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(928),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(936),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(965),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(968),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(936),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(936),
    [aux_sym_number_token1] = ACTIONS(971),
    [aux_sym_signed_number_token1] = ACTIONS(928),
    [anon_sym_SQUOTE] = ACTIONS(974),
    [anon_sym_DQUOTE] = ACTIONS(977),
    [sym_prefix_operator] = ACTIONS(980),
    [aux_sym_sign_operator_token1] = ACTIONS(928),
    [anon_sym_lsl] = ACTIONS(928),
    [anon_sym_lsr] = ACTIONS(928),
    [anon_sym_asr] = ACTIONS(928),
    [aux_sym_pow_operator_token1] = ACTIONS(936),
    [anon_sym_mod] = ACTIONS(928),
    [anon_sym_land] = ACTIONS(928),
    [anon_sym_lor] = ACTIONS(928),
    [anon_sym_lxor] = ACTIONS(928),
    [aux_sym_mult_operator_token1] = ACTIONS(928),
    [aux_sym_add_operator_token1] = ACTIONS(928),
    [sym_concat_operator] = ACTIONS(936),
    [sym_rel_operator] = ACTIONS(928),
    [anon_sym_AMP_AMP] = ACTIONS(928),
    [anon_sym_or] = ACTIONS(928),
    [anon_sym_PIPE_PIPE] = ACTIONS(928),
    [sym_assign_operator] = ACTIONS(936),
    [sym__uppercase_identifier] = ACTIONS(983),
    [aux_sym_tag_token1] = ACTIONS(986),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(100)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2763),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1873),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(100),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__attribute] = STATE(642),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [ts_builtin_sym_end] = ACTIONS(107),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_SEMI_SEMI] = ACTIONS(107),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_external] = ACTIONS(111),
    [anon_sym_type] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_exception] = ACTIONS(111),
    [anon_sym_module] = ACTIONS(111),
    [anon_sym_open] = ACTIONS(111),
    [anon_sym_include] = ACTIONS(111),
    [anon_sym_class] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(75),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(77),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_directive_token1] = ACTIONS(107),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(101)] = {
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(1122),
    [sym_typed_expression] = STATE(1045),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_labeled_argument] = STATE(1234),
    [sym_prefix_expression] = STATE(1045),
    [sym_hash_expression] = STATE(1045),
    [sym_field_get_expression] = STATE(1045),
    [sym_array_get_expression] = STATE(1045),
    [sym_string_get_expression] = STATE(1045),
    [sym_bigarray_get_expression] = STATE(1045),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(101),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [aux_sym_class_application_repeat1] = STATE(105),
    [sym__lowercase_identifier] = ACTIONS(989),
    [anon_sym_TILDE] = ACTIONS(991),
    [anon_sym_QMARK] = ACTIONS(991),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_RPAREN] = ACTIONS(915),
    [anon_sym_COMMA] = ACTIONS(915),
    [anon_sym_PIPE] = ACTIONS(917),
    [anon_sym_COLON2] = ACTIONS(917),
    [anon_sym_DOT] = ACTIONS(993),
    [anon_sym_DASH_GT] = ACTIONS(917),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_RBRACK] = ACTIONS(915),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_SEMI] = ACTIONS(915),
    [anon_sym_RBRACE] = ACTIONS(915),
    [anon_sym_constraint] = ACTIONS(917),
    [anon_sym_val] = ACTIONS(917),
    [anon_sym_end] = ACTIONS(917),
    [anon_sym_with] = ACTIONS(917),
    [anon_sym_COLON_GT] = ACTIONS(915),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_inherit] = ACTIONS(917),
    [anon_sym_method] = ACTIONS(917),
    [anon_sym_initializer] = ACTIONS(917),
    [anon_sym_AMP] = ACTIONS(917),
    [anon_sym_POUND] = ACTIONS(995),
    [anon_sym_COLON_COLON] = ACTIONS(915),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_then] = ACTIONS(917),
    [anon_sym_else] = ACTIONS(917),
    [anon_sym_do] = ACTIONS(917),
    [anon_sym_done] = ACTIONS(917),
    [anon_sym_to] = ACTIONS(917),
    [anon_sym_downto] = ACTIONS(917),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_GT_RBRACE] = ACTIONS(915),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(917),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(915),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(445),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(447),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(915),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(915),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(917),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(917),
    [sym_hash_operator] = ACTIONS(997),
    [anon_sym_lsl] = ACTIONS(917),
    [anon_sym_lsr] = ACTIONS(917),
    [anon_sym_asr] = ACTIONS(917),
    [aux_sym_pow_operator_token1] = ACTIONS(915),
    [anon_sym_mod] = ACTIONS(917),
    [anon_sym_land] = ACTIONS(917),
    [anon_sym_lor] = ACTIONS(917),
    [anon_sym_lxor] = ACTIONS(917),
    [aux_sym_mult_operator_token1] = ACTIONS(917),
    [aux_sym_add_operator_token1] = ACTIONS(917),
    [sym_concat_operator] = ACTIONS(915),
    [sym_rel_operator] = ACTIONS(917),
    [anon_sym_AMP_AMP] = ACTIONS(917),
    [anon_sym_or] = ACTIONS(917),
    [anon_sym_PIPE_PIPE] = ACTIONS(917),
    [sym_assign_operator] = ACTIONS(915),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(102)] = {
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(942),
    [sym_typed_expression] = STATE(946),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_labeled_argument] = STATE(1023),
    [sym_prefix_expression] = STATE(946),
    [sym_hash_expression] = STATE(946),
    [sym_field_get_expression] = STATE(946),
    [sym_array_get_expression] = STATE(946),
    [sym_string_get_expression] = STATE(946),
    [sym_bigarray_get_expression] = STATE(946),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(102),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [aux_sym_class_application_repeat1] = STATE(99),
    [sym__lowercase_identifier] = ACTIONS(911),
    [anon_sym_TILDE] = ACTIONS(913),
    [anon_sym_QMARK] = ACTIONS(913),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(999),
    [anon_sym_COMMA] = ACTIONS(999),
    [anon_sym_PIPE] = ACTIONS(1001),
    [anon_sym_COLON2] = ACTIONS(1001),
    [anon_sym_DASH_GT] = ACTIONS(1001),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(999),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_SEMI] = ACTIONS(999),
    [anon_sym_RBRACE] = ACTIONS(999),
    [anon_sym_constraint] = ACTIONS(1001),
    [anon_sym_val] = ACTIONS(1001),
    [anon_sym_end] = ACTIONS(1001),
    [anon_sym_with] = ACTIONS(1001),
    [anon_sym_COLON_GT] = ACTIONS(999),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_inherit] = ACTIONS(1001),
    [anon_sym_method] = ACTIONS(1001),
    [anon_sym_initializer] = ACTIONS(1001),
    [anon_sym_AMP] = ACTIONS(1001),
    [anon_sym_COLON_COLON] = ACTIONS(999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(999),
    [anon_sym_then] = ACTIONS(1001),
    [anon_sym_else] = ACTIONS(1001),
    [anon_sym_do] = ACTIONS(1001),
    [anon_sym_done] = ACTIONS(1001),
    [anon_sym_to] = ACTIONS(1001),
    [anon_sym_downto] = ACTIONS(1001),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_GT_RBRACE] = ACTIONS(999),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1001),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(999),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(527),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(529),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(999),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(999),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(1001),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(1001),
    [anon_sym_lsl] = ACTIONS(1001),
    [anon_sym_lsr] = ACTIONS(1001),
    [anon_sym_asr] = ACTIONS(1001),
    [aux_sym_pow_operator_token1] = ACTIONS(999),
    [anon_sym_mod] = ACTIONS(1001),
    [anon_sym_land] = ACTIONS(1001),
    [anon_sym_lor] = ACTIONS(1001),
    [anon_sym_lxor] = ACTIONS(1001),
    [aux_sym_mult_operator_token1] = ACTIONS(1001),
    [aux_sym_add_operator_token1] = ACTIONS(1001),
    [sym_concat_operator] = ACTIONS(999),
    [sym_rel_operator] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(1001),
    [anon_sym_or] = ACTIONS(1001),
    [anon_sym_PIPE_PIPE] = ACTIONS(1001),
    [sym_assign_operator] = ACTIONS(999),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(103)] = {
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(1195),
    [sym_typed_expression] = STATE(1189),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_labeled_argument] = STATE(1323),
    [sym_prefix_expression] = STATE(1189),
    [sym_hash_expression] = STATE(1189),
    [sym_field_get_expression] = STATE(1189),
    [sym_array_get_expression] = STATE(1189),
    [sym_string_get_expression] = STATE(1189),
    [sym_bigarray_get_expression] = STATE(1189),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(103),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [aux_sym_class_application_repeat1] = STATE(108),
    [sym__lowercase_identifier] = ACTIONS(1003),
    [anon_sym_SEMI_SEMI] = ACTIONS(915),
    [anon_sym_let] = ACTIONS(917),
    [anon_sym_and] = ACTIONS(917),
    [anon_sym_TILDE] = ACTIONS(1005),
    [anon_sym_QMARK] = ACTIONS(1005),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_external] = ACTIONS(917),
    [anon_sym_type] = ACTIONS(917),
    [anon_sym_COMMA] = ACTIONS(915),
    [anon_sym_PIPE] = ACTIONS(917),
    [anon_sym_DOT] = ACTIONS(1007),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_RBRACK] = ACTIONS(915),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_SEMI] = ACTIONS(917),
    [anon_sym_exception] = ACTIONS(917),
    [anon_sym_module] = ACTIONS(917),
    [anon_sym_open] = ACTIONS(917),
    [anon_sym_include] = ACTIONS(917),
    [anon_sym_class] = ACTIONS(917),
    [anon_sym_val] = ACTIONS(917),
    [anon_sym_end] = ACTIONS(917),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_in] = ACTIONS(917),
    [anon_sym_AMP] = ACTIONS(917),
    [anon_sym_POUND] = ACTIONS(1009),
    [anon_sym_COLON_COLON] = ACTIONS(915),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_else] = ACTIONS(917),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(917),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(915),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(915),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(915),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(917),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(917),
    [sym_hash_operator] = ACTIONS(1011),
    [anon_sym_lsl] = ACTIONS(917),
    [anon_sym_lsr] = ACTIONS(917),
    [anon_sym_asr] = ACTIONS(917),
    [aux_sym_pow_operator_token1] = ACTIONS(915),
    [anon_sym_mod] = ACTIONS(917),
    [anon_sym_land] = ACTIONS(917),
    [anon_sym_lor] = ACTIONS(917),
    [anon_sym_lxor] = ACTIONS(917),
    [aux_sym_mult_operator_token1] = ACTIONS(917),
    [aux_sym_add_operator_token1] = ACTIONS(917),
    [sym_concat_operator] = ACTIONS(915),
    [sym_rel_operator] = ACTIONS(917),
    [anon_sym_AMP_AMP] = ACTIONS(917),
    [anon_sym_or] = ACTIONS(917),
    [anon_sym_PIPE_PIPE] = ACTIONS(917),
    [sym_assign_operator] = ACTIONS(915),
    [sym_let_operator] = ACTIONS(915),
    [sym_let_and_operator] = ACTIONS(915),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_directive_token1] = ACTIONS(917),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(104)] = {
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(1299),
    [sym_typed_expression] = STATE(1295),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_labeled_argument] = STATE(1323),
    [sym_prefix_expression] = STATE(1295),
    [sym_hash_expression] = STATE(1295),
    [sym_field_get_expression] = STATE(1295),
    [sym_array_get_expression] = STATE(1295),
    [sym_string_get_expression] = STATE(1295),
    [sym_bigarray_get_expression] = STATE(1295),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(104),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [aux_sym_class_application_repeat1] = STATE(111),
    [ts_builtin_sym_end] = ACTIONS(915),
    [sym__lowercase_identifier] = ACTIONS(1013),
    [anon_sym_SEMI_SEMI] = ACTIONS(915),
    [anon_sym_let] = ACTIONS(917),
    [anon_sym_and] = ACTIONS(917),
    [anon_sym_TILDE] = ACTIONS(1015),
    [anon_sym_QMARK] = ACTIONS(1015),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_external] = ACTIONS(917),
    [anon_sym_type] = ACTIONS(917),
    [anon_sym_COMMA] = ACTIONS(915),
    [anon_sym_PIPE] = ACTIONS(917),
    [anon_sym_DOT] = ACTIONS(1017),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_SEMI] = ACTIONS(917),
    [anon_sym_exception] = ACTIONS(917),
    [anon_sym_module] = ACTIONS(917),
    [anon_sym_open] = ACTIONS(917),
    [anon_sym_include] = ACTIONS(917),
    [anon_sym_class] = ACTIONS(917),
    [anon_sym_val] = ACTIONS(917),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_in] = ACTIONS(917),
    [anon_sym_AMP] = ACTIONS(917),
    [anon_sym_POUND] = ACTIONS(1019),
    [anon_sym_COLON_COLON] = ACTIONS(915),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_else] = ACTIONS(917),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(917),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(915),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(217),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(219),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(915),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(915),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(917),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(917),
    [sym_hash_operator] = ACTIONS(1021),
    [anon_sym_lsl] = ACTIONS(917),
    [anon_sym_lsr] = ACTIONS(917),
    [anon_sym_asr] = ACTIONS(917),
    [aux_sym_pow_operator_token1] = ACTIONS(915),
    [anon_sym_mod] = ACTIONS(917),
    [anon_sym_land] = ACTIONS(917),
    [anon_sym_lor] = ACTIONS(917),
    [anon_sym_lxor] = ACTIONS(917),
    [aux_sym_mult_operator_token1] = ACTIONS(917),
    [aux_sym_add_operator_token1] = ACTIONS(917),
    [sym_concat_operator] = ACTIONS(915),
    [sym_rel_operator] = ACTIONS(917),
    [anon_sym_AMP_AMP] = ACTIONS(917),
    [anon_sym_or] = ACTIONS(917),
    [anon_sym_PIPE_PIPE] = ACTIONS(917),
    [sym_assign_operator] = ACTIONS(915),
    [sym_let_operator] = ACTIONS(915),
    [sym_let_and_operator] = ACTIONS(915),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_directive_token1] = ACTIONS(917),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(105)] = {
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(1122),
    [sym_typed_expression] = STATE(1045),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_labeled_argument] = STATE(1234),
    [sym_prefix_expression] = STATE(1045),
    [sym_hash_expression] = STATE(1045),
    [sym_field_get_expression] = STATE(1045),
    [sym_array_get_expression] = STATE(1045),
    [sym_string_get_expression] = STATE(1045),
    [sym_bigarray_get_expression] = STATE(1045),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(105),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [aux_sym_class_application_repeat1] = STATE(106),
    [sym__lowercase_identifier] = ACTIONS(989),
    [anon_sym_TILDE] = ACTIONS(991),
    [anon_sym_QMARK] = ACTIONS(991),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_RPAREN] = ACTIONS(999),
    [anon_sym_COMMA] = ACTIONS(999),
    [anon_sym_PIPE] = ACTIONS(1001),
    [anon_sym_COLON2] = ACTIONS(1001),
    [anon_sym_DASH_GT] = ACTIONS(1001),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_RBRACK] = ACTIONS(999),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_SEMI] = ACTIONS(999),
    [anon_sym_RBRACE] = ACTIONS(999),
    [anon_sym_constraint] = ACTIONS(1001),
    [anon_sym_val] = ACTIONS(1001),
    [anon_sym_end] = ACTIONS(1001),
    [anon_sym_with] = ACTIONS(1001),
    [anon_sym_COLON_GT] = ACTIONS(999),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_inherit] = ACTIONS(1001),
    [anon_sym_method] = ACTIONS(1001),
    [anon_sym_initializer] = ACTIONS(1001),
    [anon_sym_AMP] = ACTIONS(1001),
    [anon_sym_COLON_COLON] = ACTIONS(999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_then] = ACTIONS(1001),
    [anon_sym_else] = ACTIONS(1001),
    [anon_sym_do] = ACTIONS(1001),
    [anon_sym_done] = ACTIONS(1001),
    [anon_sym_to] = ACTIONS(1001),
    [anon_sym_downto] = ACTIONS(1001),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_GT_RBRACE] = ACTIONS(999),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1001),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(999),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(445),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(447),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(999),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(999),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(1001),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(1001),
    [anon_sym_lsl] = ACTIONS(1001),
    [anon_sym_lsr] = ACTIONS(1001),
    [anon_sym_asr] = ACTIONS(1001),
    [aux_sym_pow_operator_token1] = ACTIONS(999),
    [anon_sym_mod] = ACTIONS(1001),
    [anon_sym_land] = ACTIONS(1001),
    [anon_sym_lor] = ACTIONS(1001),
    [anon_sym_lxor] = ACTIONS(1001),
    [aux_sym_mult_operator_token1] = ACTIONS(1001),
    [aux_sym_add_operator_token1] = ACTIONS(1001),
    [sym_concat_operator] = ACTIONS(999),
    [sym_rel_operator] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(1001),
    [anon_sym_or] = ACTIONS(1001),
    [anon_sym_PIPE_PIPE] = ACTIONS(1001),
    [sym_assign_operator] = ACTIONS(999),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(106)] = {
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(1122),
    [sym_typed_expression] = STATE(1045),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_labeled_argument] = STATE(1234),
    [sym_prefix_expression] = STATE(1045),
    [sym_hash_expression] = STATE(1045),
    [sym_field_get_expression] = STATE(1045),
    [sym_array_get_expression] = STATE(1045),
    [sym_string_get_expression] = STATE(1045),
    [sym_bigarray_get_expression] = STATE(1045),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(106),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [aux_sym_class_application_repeat1] = STATE(106),
    [sym__lowercase_identifier] = ACTIONS(1023),
    [anon_sym_TILDE] = ACTIONS(1026),
    [anon_sym_QMARK] = ACTIONS(1026),
    [anon_sym_LPAREN] = ACTIONS(1029),
    [anon_sym_RPAREN] = ACTIONS(936),
    [anon_sym_COMMA] = ACTIONS(936),
    [anon_sym_PIPE] = ACTIONS(928),
    [anon_sym_COLON2] = ACTIONS(928),
    [anon_sym_DASH_GT] = ACTIONS(928),
    [anon_sym_LBRACK] = ACTIONS(1032),
    [anon_sym_RBRACK] = ACTIONS(936),
    [anon_sym_true] = ACTIONS(1035),
    [anon_sym_false] = ACTIONS(1035),
    [anon_sym_LBRACE] = ACTIONS(1038),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym_RBRACE] = ACTIONS(936),
    [anon_sym_constraint] = ACTIONS(928),
    [anon_sym_val] = ACTIONS(928),
    [anon_sym_end] = ACTIONS(928),
    [anon_sym_with] = ACTIONS(928),
    [anon_sym_COLON_GT] = ACTIONS(936),
    [anon_sym_object] = ACTIONS(1041),
    [anon_sym_inherit] = ACTIONS(928),
    [anon_sym_method] = ACTIONS(928),
    [anon_sym_initializer] = ACTIONS(928),
    [anon_sym_AMP] = ACTIONS(928),
    [anon_sym_COLON_COLON] = ACTIONS(936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1044),
    [anon_sym_then] = ACTIONS(928),
    [anon_sym_else] = ACTIONS(928),
    [anon_sym_do] = ACTIONS(928),
    [anon_sym_done] = ACTIONS(928),
    [anon_sym_to] = ACTIONS(928),
    [anon_sym_downto] = ACTIONS(928),
    [anon_sym_new] = ACTIONS(1047),
    [anon_sym_LBRACE_LT] = ACTIONS(1050),
    [anon_sym_GT_RBRACE] = ACTIONS(936),
    [anon_sym_begin] = ACTIONS(1053),
    [sym_ocamlyacc_value] = ACTIONS(1056),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(928),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(936),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1059),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1062),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(936),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(936),
    [aux_sym_number_token1] = ACTIONS(1065),
    [aux_sym_signed_number_token1] = ACTIONS(928),
    [anon_sym_SQUOTE] = ACTIONS(1068),
    [anon_sym_DQUOTE] = ACTIONS(1071),
    [sym_prefix_operator] = ACTIONS(1074),
    [aux_sym_sign_operator_token1] = ACTIONS(928),
    [anon_sym_lsl] = ACTIONS(928),
    [anon_sym_lsr] = ACTIONS(928),
    [anon_sym_asr] = ACTIONS(928),
    [aux_sym_pow_operator_token1] = ACTIONS(936),
    [anon_sym_mod] = ACTIONS(928),
    [anon_sym_land] = ACTIONS(928),
    [anon_sym_lor] = ACTIONS(928),
    [anon_sym_lxor] = ACTIONS(928),
    [aux_sym_mult_operator_token1] = ACTIONS(928),
    [aux_sym_add_operator_token1] = ACTIONS(928),
    [sym_concat_operator] = ACTIONS(936),
    [sym_rel_operator] = ACTIONS(928),
    [anon_sym_AMP_AMP] = ACTIONS(928),
    [anon_sym_or] = ACTIONS(928),
    [anon_sym_PIPE_PIPE] = ACTIONS(928),
    [sym_assign_operator] = ACTIONS(936),
    [sym__uppercase_identifier] = ACTIONS(1077),
    [aux_sym_tag_token1] = ACTIONS(1080),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(107)] = {
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(1393),
    [sym_typed_expression] = STATE(1364),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_labeled_argument] = STATE(1490),
    [sym_prefix_expression] = STATE(1364),
    [sym_hash_expression] = STATE(1364),
    [sym_field_get_expression] = STATE(1364),
    [sym_array_get_expression] = STATE(1364),
    [sym_string_get_expression] = STATE(1364),
    [sym_bigarray_get_expression] = STATE(1364),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(107),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [aux_sym_class_application_repeat1] = STATE(117),
    [sym__lowercase_identifier] = ACTIONS(1083),
    [anon_sym_SEMI_SEMI] = ACTIONS(915),
    [anon_sym_let] = ACTIONS(917),
    [anon_sym_TILDE] = ACTIONS(1085),
    [anon_sym_QMARK] = ACTIONS(1085),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(917),
    [anon_sym_type] = ACTIONS(917),
    [anon_sym_COMMA] = ACTIONS(915),
    [anon_sym_PIPE] = ACTIONS(917),
    [anon_sym_DOT] = ACTIONS(1087),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(915),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_SEMI] = ACTIONS(917),
    [anon_sym_exception] = ACTIONS(917),
    [anon_sym_module] = ACTIONS(917),
    [anon_sym_open] = ACTIONS(917),
    [anon_sym_include] = ACTIONS(917),
    [anon_sym_class] = ACTIONS(917),
    [anon_sym_val] = ACTIONS(917),
    [anon_sym_end] = ACTIONS(917),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_AMP] = ACTIONS(917),
    [anon_sym_POUND] = ACTIONS(1089),
    [anon_sym_COLON_COLON] = ACTIONS(915),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_else] = ACTIONS(917),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(917),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(915),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(915),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(915),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(917),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(917),
    [sym_hash_operator] = ACTIONS(1091),
    [anon_sym_lsl] = ACTIONS(917),
    [anon_sym_lsr] = ACTIONS(917),
    [anon_sym_asr] = ACTIONS(917),
    [aux_sym_pow_operator_token1] = ACTIONS(915),
    [anon_sym_mod] = ACTIONS(917),
    [anon_sym_land] = ACTIONS(917),
    [anon_sym_lor] = ACTIONS(917),
    [anon_sym_lxor] = ACTIONS(917),
    [aux_sym_mult_operator_token1] = ACTIONS(917),
    [aux_sym_add_operator_token1] = ACTIONS(917),
    [sym_concat_operator] = ACTIONS(915),
    [sym_rel_operator] = ACTIONS(917),
    [anon_sym_AMP_AMP] = ACTIONS(917),
    [anon_sym_or] = ACTIONS(917),
    [anon_sym_PIPE_PIPE] = ACTIONS(917),
    [sym_assign_operator] = ACTIONS(915),
    [sym_let_operator] = ACTIONS(915),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(917),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(108)] = {
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(1195),
    [sym_typed_expression] = STATE(1189),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_labeled_argument] = STATE(1323),
    [sym_prefix_expression] = STATE(1189),
    [sym_hash_expression] = STATE(1189),
    [sym_field_get_expression] = STATE(1189),
    [sym_array_get_expression] = STATE(1189),
    [sym_string_get_expression] = STATE(1189),
    [sym_bigarray_get_expression] = STATE(1189),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(108),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [aux_sym_class_application_repeat1] = STATE(110),
    [sym__lowercase_identifier] = ACTIONS(1003),
    [anon_sym_SEMI_SEMI] = ACTIONS(999),
    [anon_sym_let] = ACTIONS(1001),
    [anon_sym_and] = ACTIONS(1001),
    [anon_sym_TILDE] = ACTIONS(1005),
    [anon_sym_QMARK] = ACTIONS(1005),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_external] = ACTIONS(1001),
    [anon_sym_type] = ACTIONS(1001),
    [anon_sym_COMMA] = ACTIONS(999),
    [anon_sym_PIPE] = ACTIONS(1001),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_RBRACK] = ACTIONS(999),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_SEMI] = ACTIONS(1001),
    [anon_sym_exception] = ACTIONS(1001),
    [anon_sym_module] = ACTIONS(1001),
    [anon_sym_open] = ACTIONS(1001),
    [anon_sym_include] = ACTIONS(1001),
    [anon_sym_class] = ACTIONS(1001),
    [anon_sym_val] = ACTIONS(1001),
    [anon_sym_end] = ACTIONS(1001),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_in] = ACTIONS(1001),
    [anon_sym_AMP] = ACTIONS(1001),
    [anon_sym_COLON_COLON] = ACTIONS(999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_else] = ACTIONS(1001),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1001),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(999),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(999),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(999),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(1001),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(1001),
    [anon_sym_lsl] = ACTIONS(1001),
    [anon_sym_lsr] = ACTIONS(1001),
    [anon_sym_asr] = ACTIONS(1001),
    [aux_sym_pow_operator_token1] = ACTIONS(999),
    [anon_sym_mod] = ACTIONS(1001),
    [anon_sym_land] = ACTIONS(1001),
    [anon_sym_lor] = ACTIONS(1001),
    [anon_sym_lxor] = ACTIONS(1001),
    [aux_sym_mult_operator_token1] = ACTIONS(1001),
    [aux_sym_add_operator_token1] = ACTIONS(1001),
    [sym_concat_operator] = ACTIONS(999),
    [sym_rel_operator] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(1001),
    [anon_sym_or] = ACTIONS(1001),
    [anon_sym_PIPE_PIPE] = ACTIONS(1001),
    [sym_assign_operator] = ACTIONS(999),
    [sym_let_operator] = ACTIONS(999),
    [sym_let_and_operator] = ACTIONS(999),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_directive_token1] = ACTIONS(999),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(109)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3203),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1873),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(109),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(649),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_constraint] = ACTIONS(111),
    [anon_sym_val] = ACTIONS(111),
    [anon_sym_end] = ACTIONS(111),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_inherit] = ACTIONS(111),
    [anon_sym_method] = ACTIONS(111),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_initializer] = ACTIONS(111),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(527),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(529),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(107),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(110)] = {
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(1195),
    [sym_typed_expression] = STATE(1189),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_labeled_argument] = STATE(1323),
    [sym_prefix_expression] = STATE(1189),
    [sym_hash_expression] = STATE(1189),
    [sym_field_get_expression] = STATE(1189),
    [sym_array_get_expression] = STATE(1189),
    [sym_string_get_expression] = STATE(1189),
    [sym_bigarray_get_expression] = STATE(1189),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(110),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [aux_sym_class_application_repeat1] = STATE(110),
    [sym__lowercase_identifier] = ACTIONS(1093),
    [anon_sym_SEMI_SEMI] = ACTIONS(936),
    [anon_sym_let] = ACTIONS(928),
    [anon_sym_and] = ACTIONS(928),
    [anon_sym_TILDE] = ACTIONS(1096),
    [anon_sym_QMARK] = ACTIONS(1096),
    [anon_sym_LPAREN] = ACTIONS(1099),
    [anon_sym_external] = ACTIONS(928),
    [anon_sym_type] = ACTIONS(928),
    [anon_sym_COMMA] = ACTIONS(936),
    [anon_sym_PIPE] = ACTIONS(928),
    [anon_sym_LBRACK] = ACTIONS(1102),
    [anon_sym_RBRACK] = ACTIONS(936),
    [anon_sym_true] = ACTIONS(1105),
    [anon_sym_false] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_SEMI] = ACTIONS(928),
    [anon_sym_exception] = ACTIONS(928),
    [anon_sym_module] = ACTIONS(928),
    [anon_sym_open] = ACTIONS(928),
    [anon_sym_include] = ACTIONS(928),
    [anon_sym_class] = ACTIONS(928),
    [anon_sym_val] = ACTIONS(928),
    [anon_sym_end] = ACTIONS(928),
    [anon_sym_object] = ACTIONS(1111),
    [anon_sym_in] = ACTIONS(928),
    [anon_sym_AMP] = ACTIONS(928),
    [anon_sym_COLON_COLON] = ACTIONS(936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1114),
    [anon_sym_else] = ACTIONS(928),
    [anon_sym_new] = ACTIONS(1117),
    [anon_sym_LBRACE_LT] = ACTIONS(1120),
    [anon_sym_begin] = ACTIONS(1123),
    [sym_ocamlyacc_value] = ACTIONS(1126),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(928),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(936),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1129),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1132),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(936),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(936),
    [aux_sym_number_token1] = ACTIONS(1135),
    [aux_sym_signed_number_token1] = ACTIONS(928),
    [anon_sym_SQUOTE] = ACTIONS(1138),
    [anon_sym_DQUOTE] = ACTIONS(1141),
    [sym_prefix_operator] = ACTIONS(1144),
    [aux_sym_sign_operator_token1] = ACTIONS(928),
    [anon_sym_lsl] = ACTIONS(928),
    [anon_sym_lsr] = ACTIONS(928),
    [anon_sym_asr] = ACTIONS(928),
    [aux_sym_pow_operator_token1] = ACTIONS(936),
    [anon_sym_mod] = ACTIONS(928),
    [anon_sym_land] = ACTIONS(928),
    [anon_sym_lor] = ACTIONS(928),
    [anon_sym_lxor] = ACTIONS(928),
    [aux_sym_mult_operator_token1] = ACTIONS(928),
    [aux_sym_add_operator_token1] = ACTIONS(928),
    [sym_concat_operator] = ACTIONS(936),
    [sym_rel_operator] = ACTIONS(928),
    [anon_sym_AMP_AMP] = ACTIONS(928),
    [anon_sym_or] = ACTIONS(928),
    [anon_sym_PIPE_PIPE] = ACTIONS(928),
    [sym_assign_operator] = ACTIONS(936),
    [sym_let_operator] = ACTIONS(936),
    [sym_let_and_operator] = ACTIONS(936),
    [sym__uppercase_identifier] = ACTIONS(1147),
    [aux_sym_directive_token1] = ACTIONS(936),
    [aux_sym_tag_token1] = ACTIONS(1150),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(111)] = {
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(1299),
    [sym_typed_expression] = STATE(1295),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_labeled_argument] = STATE(1323),
    [sym_prefix_expression] = STATE(1295),
    [sym_hash_expression] = STATE(1295),
    [sym_field_get_expression] = STATE(1295),
    [sym_array_get_expression] = STATE(1295),
    [sym_string_get_expression] = STATE(1295),
    [sym_bigarray_get_expression] = STATE(1295),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(111),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [aux_sym_class_application_repeat1] = STATE(113),
    [ts_builtin_sym_end] = ACTIONS(999),
    [sym__lowercase_identifier] = ACTIONS(1013),
    [anon_sym_SEMI_SEMI] = ACTIONS(999),
    [anon_sym_let] = ACTIONS(1001),
    [anon_sym_and] = ACTIONS(1001),
    [anon_sym_TILDE] = ACTIONS(1015),
    [anon_sym_QMARK] = ACTIONS(1015),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_external] = ACTIONS(1001),
    [anon_sym_type] = ACTIONS(1001),
    [anon_sym_COMMA] = ACTIONS(999),
    [anon_sym_PIPE] = ACTIONS(1001),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_SEMI] = ACTIONS(1001),
    [anon_sym_exception] = ACTIONS(1001),
    [anon_sym_module] = ACTIONS(1001),
    [anon_sym_open] = ACTIONS(1001),
    [anon_sym_include] = ACTIONS(1001),
    [anon_sym_class] = ACTIONS(1001),
    [anon_sym_val] = ACTIONS(1001),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_in] = ACTIONS(1001),
    [anon_sym_AMP] = ACTIONS(1001),
    [anon_sym_COLON_COLON] = ACTIONS(999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_else] = ACTIONS(1001),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1001),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(999),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(217),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(219),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(999),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(999),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(1001),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(1001),
    [anon_sym_lsl] = ACTIONS(1001),
    [anon_sym_lsr] = ACTIONS(1001),
    [anon_sym_asr] = ACTIONS(1001),
    [aux_sym_pow_operator_token1] = ACTIONS(999),
    [anon_sym_mod] = ACTIONS(1001),
    [anon_sym_land] = ACTIONS(1001),
    [anon_sym_lor] = ACTIONS(1001),
    [anon_sym_lxor] = ACTIONS(1001),
    [aux_sym_mult_operator_token1] = ACTIONS(1001),
    [aux_sym_add_operator_token1] = ACTIONS(1001),
    [sym_concat_operator] = ACTIONS(999),
    [sym_rel_operator] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(1001),
    [anon_sym_or] = ACTIONS(1001),
    [anon_sym_PIPE_PIPE] = ACTIONS(1001),
    [sym_assign_operator] = ACTIONS(999),
    [sym_let_operator] = ACTIONS(999),
    [sym_let_and_operator] = ACTIONS(999),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_directive_token1] = ACTIONS(999),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(112)] = {
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(1486),
    [sym_typed_expression] = STATE(1471),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_labeled_argument] = STATE(1490),
    [sym_prefix_expression] = STATE(1471),
    [sym_hash_expression] = STATE(1471),
    [sym_field_get_expression] = STATE(1471),
    [sym_array_get_expression] = STATE(1471),
    [sym_string_get_expression] = STATE(1471),
    [sym_bigarray_get_expression] = STATE(1471),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(112),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [aux_sym_class_application_repeat1] = STATE(118),
    [ts_builtin_sym_end] = ACTIONS(915),
    [sym__lowercase_identifier] = ACTIONS(1153),
    [anon_sym_SEMI_SEMI] = ACTIONS(915),
    [anon_sym_let] = ACTIONS(917),
    [anon_sym_TILDE] = ACTIONS(1155),
    [anon_sym_QMARK] = ACTIONS(1155),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_external] = ACTIONS(917),
    [anon_sym_type] = ACTIONS(917),
    [anon_sym_COMMA] = ACTIONS(915),
    [anon_sym_PIPE] = ACTIONS(917),
    [anon_sym_DOT] = ACTIONS(1157),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_SEMI] = ACTIONS(917),
    [anon_sym_exception] = ACTIONS(917),
    [anon_sym_module] = ACTIONS(917),
    [anon_sym_open] = ACTIONS(917),
    [anon_sym_include] = ACTIONS(917),
    [anon_sym_class] = ACTIONS(917),
    [anon_sym_val] = ACTIONS(917),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(917),
    [anon_sym_POUND] = ACTIONS(1159),
    [anon_sym_COLON_COLON] = ACTIONS(915),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_else] = ACTIONS(917),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(917),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(915),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(75),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(77),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(915),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(915),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(917),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(917),
    [sym_hash_operator] = ACTIONS(1161),
    [anon_sym_lsl] = ACTIONS(917),
    [anon_sym_lsr] = ACTIONS(917),
    [anon_sym_asr] = ACTIONS(917),
    [aux_sym_pow_operator_token1] = ACTIONS(915),
    [anon_sym_mod] = ACTIONS(917),
    [anon_sym_land] = ACTIONS(917),
    [anon_sym_lor] = ACTIONS(917),
    [anon_sym_lxor] = ACTIONS(917),
    [aux_sym_mult_operator_token1] = ACTIONS(917),
    [aux_sym_add_operator_token1] = ACTIONS(917),
    [sym_concat_operator] = ACTIONS(915),
    [sym_rel_operator] = ACTIONS(917),
    [anon_sym_AMP_AMP] = ACTIONS(917),
    [anon_sym_or] = ACTIONS(917),
    [anon_sym_PIPE_PIPE] = ACTIONS(917),
    [sym_assign_operator] = ACTIONS(915),
    [sym_let_operator] = ACTIONS(915),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_directive_token1] = ACTIONS(917),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(113)] = {
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(1299),
    [sym_typed_expression] = STATE(1295),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_labeled_argument] = STATE(1323),
    [sym_prefix_expression] = STATE(1295),
    [sym_hash_expression] = STATE(1295),
    [sym_field_get_expression] = STATE(1295),
    [sym_array_get_expression] = STATE(1295),
    [sym_string_get_expression] = STATE(1295),
    [sym_bigarray_get_expression] = STATE(1295),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(113),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [aux_sym_class_application_repeat1] = STATE(113),
    [ts_builtin_sym_end] = ACTIONS(936),
    [sym__lowercase_identifier] = ACTIONS(1163),
    [anon_sym_SEMI_SEMI] = ACTIONS(936),
    [anon_sym_let] = ACTIONS(928),
    [anon_sym_and] = ACTIONS(928),
    [anon_sym_TILDE] = ACTIONS(1166),
    [anon_sym_QMARK] = ACTIONS(1166),
    [anon_sym_LPAREN] = ACTIONS(1169),
    [anon_sym_external] = ACTIONS(928),
    [anon_sym_type] = ACTIONS(928),
    [anon_sym_COMMA] = ACTIONS(936),
    [anon_sym_PIPE] = ACTIONS(928),
    [anon_sym_LBRACK] = ACTIONS(1172),
    [anon_sym_true] = ACTIONS(1175),
    [anon_sym_false] = ACTIONS(1175),
    [anon_sym_LBRACE] = ACTIONS(1178),
    [anon_sym_SEMI] = ACTIONS(928),
    [anon_sym_exception] = ACTIONS(928),
    [anon_sym_module] = ACTIONS(928),
    [anon_sym_open] = ACTIONS(928),
    [anon_sym_include] = ACTIONS(928),
    [anon_sym_class] = ACTIONS(928),
    [anon_sym_val] = ACTIONS(928),
    [anon_sym_object] = ACTIONS(1181),
    [anon_sym_in] = ACTIONS(928),
    [anon_sym_AMP] = ACTIONS(928),
    [anon_sym_COLON_COLON] = ACTIONS(936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1184),
    [anon_sym_else] = ACTIONS(928),
    [anon_sym_new] = ACTIONS(1187),
    [anon_sym_LBRACE_LT] = ACTIONS(1190),
    [anon_sym_begin] = ACTIONS(1193),
    [sym_ocamlyacc_value] = ACTIONS(1196),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(928),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(936),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1199),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1202),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(936),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(936),
    [aux_sym_number_token1] = ACTIONS(1205),
    [aux_sym_signed_number_token1] = ACTIONS(928),
    [anon_sym_SQUOTE] = ACTIONS(1208),
    [anon_sym_DQUOTE] = ACTIONS(1211),
    [sym_prefix_operator] = ACTIONS(1214),
    [aux_sym_sign_operator_token1] = ACTIONS(928),
    [anon_sym_lsl] = ACTIONS(928),
    [anon_sym_lsr] = ACTIONS(928),
    [anon_sym_asr] = ACTIONS(928),
    [aux_sym_pow_operator_token1] = ACTIONS(936),
    [anon_sym_mod] = ACTIONS(928),
    [anon_sym_land] = ACTIONS(928),
    [anon_sym_lor] = ACTIONS(928),
    [anon_sym_lxor] = ACTIONS(928),
    [aux_sym_mult_operator_token1] = ACTIONS(928),
    [aux_sym_add_operator_token1] = ACTIONS(928),
    [sym_concat_operator] = ACTIONS(936),
    [sym_rel_operator] = ACTIONS(928),
    [anon_sym_AMP_AMP] = ACTIONS(928),
    [anon_sym_or] = ACTIONS(928),
    [anon_sym_PIPE_PIPE] = ACTIONS(928),
    [sym_assign_operator] = ACTIONS(936),
    [sym_let_operator] = ACTIONS(936),
    [sym_let_and_operator] = ACTIONS(936),
    [sym__uppercase_identifier] = ACTIONS(1217),
    [aux_sym_directive_token1] = ACTIONS(936),
    [aux_sym_tag_token1] = ACTIONS(1220),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(114)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3231),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1873),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(114),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(627),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(107),
    [anon_sym_COLON2] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(107),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_RBRACE] = ACTIONS(107),
    [anon_sym_end] = ACTIONS(111),
    [anon_sym_with] = ACTIONS(111),
    [anon_sym_COLON_GT] = ACTIONS(107),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_then] = ACTIONS(111),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(115)] = {
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(1393),
    [sym_typed_expression] = STATE(1364),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_labeled_argument] = STATE(1490),
    [sym_prefix_expression] = STATE(1364),
    [sym_hash_expression] = STATE(1364),
    [sym_field_get_expression] = STATE(1364),
    [sym_array_get_expression] = STATE(1364),
    [sym_string_get_expression] = STATE(1364),
    [sym_bigarray_get_expression] = STATE(1364),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(115),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [aux_sym_class_application_repeat1] = STATE(115),
    [sym__lowercase_identifier] = ACTIONS(1223),
    [anon_sym_SEMI_SEMI] = ACTIONS(936),
    [anon_sym_let] = ACTIONS(928),
    [anon_sym_and] = ACTIONS(928),
    [anon_sym_TILDE] = ACTIONS(1226),
    [anon_sym_QMARK] = ACTIONS(1226),
    [anon_sym_LPAREN] = ACTIONS(1229),
    [anon_sym_external] = ACTIONS(928),
    [anon_sym_type] = ACTIONS(928),
    [anon_sym_COMMA] = ACTIONS(936),
    [anon_sym_PIPE] = ACTIONS(928),
    [anon_sym_LBRACK] = ACTIONS(1232),
    [anon_sym_RBRACK] = ACTIONS(936),
    [anon_sym_true] = ACTIONS(1235),
    [anon_sym_false] = ACTIONS(1235),
    [anon_sym_LBRACE] = ACTIONS(1238),
    [anon_sym_SEMI] = ACTIONS(928),
    [anon_sym_exception] = ACTIONS(928),
    [anon_sym_module] = ACTIONS(928),
    [anon_sym_open] = ACTIONS(928),
    [anon_sym_include] = ACTIONS(928),
    [anon_sym_class] = ACTIONS(928),
    [anon_sym_val] = ACTIONS(928),
    [anon_sym_end] = ACTIONS(928),
    [anon_sym_object] = ACTIONS(1241),
    [anon_sym_AMP] = ACTIONS(928),
    [anon_sym_COLON_COLON] = ACTIONS(936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1244),
    [anon_sym_else] = ACTIONS(928),
    [anon_sym_new] = ACTIONS(1247),
    [anon_sym_LBRACE_LT] = ACTIONS(1250),
    [anon_sym_begin] = ACTIONS(1253),
    [sym_ocamlyacc_value] = ACTIONS(1256),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(928),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(936),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1259),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1262),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(936),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(936),
    [aux_sym_number_token1] = ACTIONS(1265),
    [aux_sym_signed_number_token1] = ACTIONS(928),
    [anon_sym_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1271),
    [sym_prefix_operator] = ACTIONS(1274),
    [aux_sym_sign_operator_token1] = ACTIONS(928),
    [anon_sym_lsl] = ACTIONS(928),
    [anon_sym_lsr] = ACTIONS(928),
    [anon_sym_asr] = ACTIONS(928),
    [aux_sym_pow_operator_token1] = ACTIONS(936),
    [anon_sym_mod] = ACTIONS(928),
    [anon_sym_land] = ACTIONS(928),
    [anon_sym_lor] = ACTIONS(928),
    [anon_sym_lxor] = ACTIONS(928),
    [aux_sym_mult_operator_token1] = ACTIONS(928),
    [aux_sym_add_operator_token1] = ACTIONS(928),
    [sym_concat_operator] = ACTIONS(936),
    [sym_rel_operator] = ACTIONS(928),
    [anon_sym_AMP_AMP] = ACTIONS(928),
    [anon_sym_or] = ACTIONS(928),
    [anon_sym_PIPE_PIPE] = ACTIONS(928),
    [sym_assign_operator] = ACTIONS(936),
    [sym_let_operator] = ACTIONS(936),
    [sym__uppercase_identifier] = ACTIONS(1277),
    [aux_sym_directive_token1] = ACTIONS(936),
    [aux_sym_tag_token1] = ACTIONS(1280),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(116)] = {
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(1486),
    [sym_typed_expression] = STATE(1471),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_labeled_argument] = STATE(1490),
    [sym_prefix_expression] = STATE(1471),
    [sym_hash_expression] = STATE(1471),
    [sym_field_get_expression] = STATE(1471),
    [sym_array_get_expression] = STATE(1471),
    [sym_string_get_expression] = STATE(1471),
    [sym_bigarray_get_expression] = STATE(1471),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(116),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [aux_sym_class_application_repeat1] = STATE(116),
    [ts_builtin_sym_end] = ACTIONS(936),
    [sym__lowercase_identifier] = ACTIONS(1283),
    [anon_sym_SEMI_SEMI] = ACTIONS(936),
    [anon_sym_let] = ACTIONS(928),
    [anon_sym_and] = ACTIONS(928),
    [anon_sym_TILDE] = ACTIONS(1286),
    [anon_sym_QMARK] = ACTIONS(1286),
    [anon_sym_LPAREN] = ACTIONS(1289),
    [anon_sym_external] = ACTIONS(928),
    [anon_sym_type] = ACTIONS(928),
    [anon_sym_COMMA] = ACTIONS(936),
    [anon_sym_PIPE] = ACTIONS(928),
    [anon_sym_LBRACK] = ACTIONS(1292),
    [anon_sym_true] = ACTIONS(1295),
    [anon_sym_false] = ACTIONS(1295),
    [anon_sym_LBRACE] = ACTIONS(1298),
    [anon_sym_SEMI] = ACTIONS(928),
    [anon_sym_exception] = ACTIONS(928),
    [anon_sym_module] = ACTIONS(928),
    [anon_sym_open] = ACTIONS(928),
    [anon_sym_include] = ACTIONS(928),
    [anon_sym_class] = ACTIONS(928),
    [anon_sym_val] = ACTIONS(928),
    [anon_sym_object] = ACTIONS(1301),
    [anon_sym_AMP] = ACTIONS(928),
    [anon_sym_COLON_COLON] = ACTIONS(936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1304),
    [anon_sym_else] = ACTIONS(928),
    [anon_sym_new] = ACTIONS(1307),
    [anon_sym_LBRACE_LT] = ACTIONS(1310),
    [anon_sym_begin] = ACTIONS(1313),
    [sym_ocamlyacc_value] = ACTIONS(1316),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(928),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(936),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1319),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1322),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(936),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(936),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(928),
    [anon_sym_SQUOTE] = ACTIONS(1328),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym_prefix_operator] = ACTIONS(1334),
    [aux_sym_sign_operator_token1] = ACTIONS(928),
    [anon_sym_lsl] = ACTIONS(928),
    [anon_sym_lsr] = ACTIONS(928),
    [anon_sym_asr] = ACTIONS(928),
    [aux_sym_pow_operator_token1] = ACTIONS(936),
    [anon_sym_mod] = ACTIONS(928),
    [anon_sym_land] = ACTIONS(928),
    [anon_sym_lor] = ACTIONS(928),
    [anon_sym_lxor] = ACTIONS(928),
    [aux_sym_mult_operator_token1] = ACTIONS(928),
    [aux_sym_add_operator_token1] = ACTIONS(928),
    [sym_concat_operator] = ACTIONS(936),
    [sym_rel_operator] = ACTIONS(928),
    [anon_sym_AMP_AMP] = ACTIONS(928),
    [anon_sym_or] = ACTIONS(928),
    [anon_sym_PIPE_PIPE] = ACTIONS(928),
    [sym_assign_operator] = ACTIONS(936),
    [sym_let_operator] = ACTIONS(936),
    [sym__uppercase_identifier] = ACTIONS(1337),
    [aux_sym_directive_token1] = ACTIONS(936),
    [aux_sym_tag_token1] = ACTIONS(1340),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(117)] = {
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(1393),
    [sym_typed_expression] = STATE(1364),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_labeled_argument] = STATE(1490),
    [sym_prefix_expression] = STATE(1364),
    [sym_hash_expression] = STATE(1364),
    [sym_field_get_expression] = STATE(1364),
    [sym_array_get_expression] = STATE(1364),
    [sym_string_get_expression] = STATE(1364),
    [sym_bigarray_get_expression] = STATE(1364),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(117),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [aux_sym_class_application_repeat1] = STATE(115),
    [sym__lowercase_identifier] = ACTIONS(1083),
    [anon_sym_SEMI_SEMI] = ACTIONS(999),
    [anon_sym_let] = ACTIONS(1001),
    [anon_sym_TILDE] = ACTIONS(1085),
    [anon_sym_QMARK] = ACTIONS(1085),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(1001),
    [anon_sym_type] = ACTIONS(1001),
    [anon_sym_COMMA] = ACTIONS(999),
    [anon_sym_PIPE] = ACTIONS(1001),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(999),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_SEMI] = ACTIONS(1001),
    [anon_sym_exception] = ACTIONS(1001),
    [anon_sym_module] = ACTIONS(1001),
    [anon_sym_open] = ACTIONS(1001),
    [anon_sym_include] = ACTIONS(1001),
    [anon_sym_class] = ACTIONS(1001),
    [anon_sym_val] = ACTIONS(1001),
    [anon_sym_end] = ACTIONS(1001),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_AMP] = ACTIONS(1001),
    [anon_sym_COLON_COLON] = ACTIONS(999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_else] = ACTIONS(1001),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1001),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(999),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(999),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(999),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(1001),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(1001),
    [anon_sym_lsl] = ACTIONS(1001),
    [anon_sym_lsr] = ACTIONS(1001),
    [anon_sym_asr] = ACTIONS(1001),
    [aux_sym_pow_operator_token1] = ACTIONS(999),
    [anon_sym_mod] = ACTIONS(1001),
    [anon_sym_land] = ACTIONS(1001),
    [anon_sym_lor] = ACTIONS(1001),
    [anon_sym_lxor] = ACTIONS(1001),
    [aux_sym_mult_operator_token1] = ACTIONS(1001),
    [aux_sym_add_operator_token1] = ACTIONS(1001),
    [sym_concat_operator] = ACTIONS(999),
    [sym_rel_operator] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(1001),
    [anon_sym_or] = ACTIONS(1001),
    [anon_sym_PIPE_PIPE] = ACTIONS(1001),
    [sym_assign_operator] = ACTIONS(999),
    [sym_let_operator] = ACTIONS(999),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(999),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(118)] = {
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(1486),
    [sym_typed_expression] = STATE(1471),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_labeled_argument] = STATE(1490),
    [sym_prefix_expression] = STATE(1471),
    [sym_hash_expression] = STATE(1471),
    [sym_field_get_expression] = STATE(1471),
    [sym_array_get_expression] = STATE(1471),
    [sym_string_get_expression] = STATE(1471),
    [sym_bigarray_get_expression] = STATE(1471),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(118),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [aux_sym_class_application_repeat1] = STATE(116),
    [ts_builtin_sym_end] = ACTIONS(999),
    [sym__lowercase_identifier] = ACTIONS(1153),
    [anon_sym_SEMI_SEMI] = ACTIONS(999),
    [anon_sym_let] = ACTIONS(1001),
    [anon_sym_TILDE] = ACTIONS(1155),
    [anon_sym_QMARK] = ACTIONS(1155),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_external] = ACTIONS(1001),
    [anon_sym_type] = ACTIONS(1001),
    [anon_sym_COMMA] = ACTIONS(999),
    [anon_sym_PIPE] = ACTIONS(1001),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_SEMI] = ACTIONS(1001),
    [anon_sym_exception] = ACTIONS(1001),
    [anon_sym_module] = ACTIONS(1001),
    [anon_sym_open] = ACTIONS(1001),
    [anon_sym_include] = ACTIONS(1001),
    [anon_sym_class] = ACTIONS(1001),
    [anon_sym_val] = ACTIONS(1001),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(1001),
    [anon_sym_COLON_COLON] = ACTIONS(999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_else] = ACTIONS(1001),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1001),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(999),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(75),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(77),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(999),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(999),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(1001),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(1001),
    [anon_sym_lsl] = ACTIONS(1001),
    [anon_sym_lsr] = ACTIONS(1001),
    [anon_sym_asr] = ACTIONS(1001),
    [aux_sym_pow_operator_token1] = ACTIONS(999),
    [anon_sym_mod] = ACTIONS(1001),
    [anon_sym_land] = ACTIONS(1001),
    [anon_sym_lor] = ACTIONS(1001),
    [anon_sym_lxor] = ACTIONS(1001),
    [aux_sym_mult_operator_token1] = ACTIONS(1001),
    [aux_sym_add_operator_token1] = ACTIONS(1001),
    [sym_concat_operator] = ACTIONS(999),
    [sym_rel_operator] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(1001),
    [anon_sym_or] = ACTIONS(1001),
    [anon_sym_PIPE_PIPE] = ACTIONS(1001),
    [sym_assign_operator] = ACTIONS(999),
    [sym_let_operator] = ACTIONS(999),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_directive_token1] = ACTIONS(999),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(119)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3604),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3080),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(119),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__attribute] = STATE(648),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_and] = ACTIONS(111),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_in] = ACTIONS(111),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(107),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_let_and_operator] = ACTIONS(107),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(120)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3725),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(2137),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(120),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(650),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_done] = ACTIONS(111),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_to] = ACTIONS(111),
    [anon_sym_downto] = ACTIONS(111),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(121)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7874),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(121),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(226),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1343),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(122)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8241),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(122),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(228),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1345),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(123)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7604),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(123),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(230),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1347),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(124)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7380),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(124),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(233),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1349),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(125)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7664),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(125),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(231),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1351),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(126)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7667),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(126),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(236),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1353),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(127)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7310),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(127),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(214),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1355),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(128)] = {
    [sym__abstract_type] = STATE(7924),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5388),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(128),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__infix_operator] = STATE(7684),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1363),
    [anon_sym_type] = ACTIONS(1365),
    [anon_sym_DOT] = ACTIONS(1367),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_module] = ACTIONS(1377),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1381),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1403),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1403),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1403),
    [sym_let_and_operator] = ACTIONS(1403),
    [sym_match_operator] = ACTIONS(1403),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(129)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7621),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(129),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(239),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1409),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(130)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7894),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(130),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(241),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1411),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(131)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8241),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(131),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(217),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1413),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(132)] = {
    [sym__abstract_type] = STATE(7924),
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5531),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(132),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__infix_operator] = STATE(7244),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1417),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(1421),
    [anon_sym_type] = ACTIONS(1365),
    [anon_sym_DOT] = ACTIONS(1423),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_module] = ACTIONS(1431),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1441),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1441),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1441),
    [sym_let_and_operator] = ACTIONS(1441),
    [sym_match_operator] = ACTIONS(1441),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(133)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8075),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(133),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(243),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1443),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(134)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7663),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(134),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(245),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1445),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(135)] = {
    [sym__abstract_type] = STATE(7924),
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5538),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(135),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__infix_operator] = STATE(7379),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1417),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(1447),
    [anon_sym_type] = ACTIONS(1365),
    [anon_sym_DOT] = ACTIONS(1449),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_module] = ACTIONS(1451),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1453),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1455),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1455),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1455),
    [sym_let_and_operator] = ACTIONS(1455),
    [sym_match_operator] = ACTIONS(1455),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(136)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7641),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(136),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(256),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1457),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(137)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7911),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(137),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(222),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1459),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(138)] = {
    [sym__abstract_type] = STATE(7924),
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5472),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(138),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__infix_operator] = STATE(7564),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1417),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(1461),
    [anon_sym_type] = ACTIONS(1365),
    [anon_sym_DOT] = ACTIONS(1463),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_module] = ACTIONS(1465),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1469),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1469),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1469),
    [sym_let_and_operator] = ACTIONS(1469),
    [sym_match_operator] = ACTIONS(1469),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(139)] = {
    [sym__abstract_type] = STATE(8037),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5374),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(139),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__infix_operator] = STATE(8042),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1471),
    [anon_sym_type] = ACTIONS(1365),
    [anon_sym_DOT] = ACTIONS(1473),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_module] = ACTIONS(1475),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1477),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1479),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1479),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1479),
    [sym_let_and_operator] = ACTIONS(1479),
    [sym_match_operator] = ACTIONS(1479),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(140)] = {
    [sym__abstract_type] = STATE(7924),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5388),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(140),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__infix_operator] = STATE(7684),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1363),
    [anon_sym_type] = ACTIONS(1365),
    [anon_sym_DOT] = ACTIONS(1367),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_module] = ACTIONS(1377),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1481),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1403),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1403),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1403),
    [sym_let_and_operator] = ACTIONS(1403),
    [sym_match_operator] = ACTIONS(1403),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(141)] = {
    [sym__abstract_type] = STATE(7924),
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5421),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(141),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__infix_operator] = STATE(7464),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1417),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(709),
    [anon_sym_type] = ACTIONS(1365),
    [anon_sym_DOT] = ACTIONS(711),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_module] = ACTIONS(1483),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1381),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(717),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(717),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(717),
    [sym_let_and_operator] = ACTIONS(717),
    [sym_match_operator] = ACTIONS(717),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(142)] = {
    [sym__abstract_type] = STATE(7924),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5437),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(142),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__infix_operator] = STATE(7244),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1421),
    [anon_sym_type] = ACTIONS(1365),
    [anon_sym_DOT] = ACTIONS(1423),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_module] = ACTIONS(1431),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1441),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1441),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1441),
    [sym_let_and_operator] = ACTIONS(1441),
    [sym_match_operator] = ACTIONS(1441),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(143)] = {
    [sym__abstract_type] = STATE(7924),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5536),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(143),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__infix_operator] = STATE(7464),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(709),
    [anon_sym_type] = ACTIONS(1365),
    [anon_sym_DOT] = ACTIONS(711),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_module] = ACTIONS(1483),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1381),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(717),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(717),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(717),
    [sym_let_and_operator] = ACTIONS(717),
    [sym_match_operator] = ACTIONS(717),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(144)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7422),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(144),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(588),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(145)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2645),
    [sym__sequence_expression] = STATE(2684),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_refutation_case] = STATE(2684),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(145),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_DOT] = ACTIONS(1485),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(146)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3676),
    [sym__sequence_expression] = STATE(4069),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_refutation_case] = STATE(4069),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(146),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_DOT] = ACTIONS(1491),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(147)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2139),
    [sym__sequence_expression] = STATE(2648),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_refutation_case] = STATE(2648),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(147),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_DOT] = ACTIONS(1485),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(148)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2550),
    [sym__sequence_expression] = STATE(2648),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_refutation_case] = STATE(2648),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(148),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_DOT] = ACTIONS(1485),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(149)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3828),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(7364),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(149),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(633),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_DASH_GT] = ACTIONS(107),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(150)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3461),
    [sym__sequence_expression] = STATE(3592),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_refutation_case] = STATE(3592),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(150),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_DOT] = ACTIONS(1501),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(151)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2139),
    [sym__sequence_expression] = STATE(2684),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_refutation_case] = STATE(2684),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(151),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_DOT] = ACTIONS(1485),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(152)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3077),
    [sym__sequence_expression] = STATE(2782),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_refutation_case] = STATE(2782),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(152),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_DOT] = ACTIONS(1503),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(153)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2550),
    [sym__sequence_expression] = STATE(2684),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_refutation_case] = STATE(2684),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(153),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_DOT] = ACTIONS(1485),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(154)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3407),
    [sym__sequence_expression] = STATE(3506),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_refutation_case] = STATE(3506),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(154),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_DOT] = ACTIONS(1501),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(155)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8294),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(155),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(328),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(156)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3106),
    [sym__sequence_expression] = STATE(2782),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_refutation_case] = STATE(2782),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(156),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_DOT] = ACTIONS(1503),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(157)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2096),
    [sym__sequence_expression] = STATE(2684),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_refutation_case] = STATE(2684),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(157),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_DOT] = ACTIONS(1485),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(158)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2274),
    [sym__sequence_expression] = STATE(2648),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_refutation_case] = STATE(2648),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(158),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1485),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(159)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3407),
    [sym__sequence_expression] = STATE(3592),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_refutation_case] = STATE(3592),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(159),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_DOT] = ACTIONS(1501),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(160)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2186),
    [sym__sequence_expression] = STATE(2648),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_refutation_case] = STATE(2648),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(160),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1485),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(161)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2186),
    [sym__sequence_expression] = STATE(2684),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_refutation_case] = STATE(2684),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(161),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1485),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(162)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7927),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(162),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(316),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(163)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6819),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(163),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(560),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(164)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(6820),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(164),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(422),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(165)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3676),
    [sym__sequence_expression] = STATE(4115),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_refutation_case] = STATE(4115),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(165),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_DOT] = ACTIONS(1491),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(166)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3106),
    [sym__sequence_expression] = STATE(2859),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_refutation_case] = STATE(2859),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(166),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_DOT] = ACTIONS(1503),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(167)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(6823),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(167),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(292),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(168)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2274),
    [sym__sequence_expression] = STATE(2684),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_refutation_case] = STATE(2684),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(168),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_DOT] = ACTIONS(1485),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(169)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3461),
    [sym__sequence_expression] = STATE(3506),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_refutation_case] = STATE(3506),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(169),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_DOT] = ACTIONS(1501),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(170)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6815),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(170),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(449),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(171)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(7151),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(171),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(457),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(172)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7150),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(172),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(277),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(173)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(7065),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(173),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(469),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(174)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6875),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(174),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(466),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(175)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7056),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(175),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(312),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(176)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7064),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(176),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(332),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(177)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6957),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(177),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(349),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(178)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6773),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(178),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(357),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(179)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6836),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(179),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(454),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(180)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2096),
    [sym__sequence_expression] = STATE(2648),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_refutation_case] = STATE(2648),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(180),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_DOT] = ACTIONS(1485),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(181)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6827),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(181),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(364),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(182)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6837),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(182),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(455),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(183)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6904),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(183),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(373),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(184)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6834),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(184),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(453),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(185)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(185),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(546),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(186)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7542),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(186),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(547),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(187)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2645),
    [sym__sequence_expression] = STATE(2648),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_refutation_case] = STATE(2648),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(187),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_DOT] = ACTIONS(1485),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(188)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3819),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1873),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(188),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(646),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_do] = ACTIONS(111),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(189)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3077),
    [sym__sequence_expression] = STATE(2859),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_refutation_case] = STATE(2859),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(189),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_DOT] = ACTIONS(1503),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(190)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7095),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(190),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(557),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(191)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7766),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(191),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(562),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(192)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7771),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(192),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(563),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(193)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7197),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(193),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(567),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(194)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7200),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(194),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(568),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(195)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7275),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(195),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(572),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(196)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7278),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(196),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(573),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(197)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7336),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(197),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(577),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(198)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7339),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(198),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(578),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(199)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7382),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(199),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(582),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(200)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7383),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(200),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(583),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(201)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7421),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(201),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(587),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(202)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6803),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(202),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(270),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(203)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7458),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(203),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(591),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(204)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7459),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(204),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(592),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(205)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7494),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(205),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(596),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(206)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7495),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(206),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(597),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(207)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7530),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(207),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(601),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(208)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7531),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(208),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(602),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(209)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7565),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(209),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(606),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(210)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7566),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(210),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(607),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(211)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7601),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(211),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(611),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(212)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7602),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(212),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(612),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(213)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6830),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(213),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__attribute] = STATE(452),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [anon_sym_PERCENT] = ACTIONS(909),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(214)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7392),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(214),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1509),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(215)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5531),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(215),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__infix_operator] = STATE(7244),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1417),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(1421),
    [anon_sym_DOT] = ACTIONS(1423),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_module] = ACTIONS(1431),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1441),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1441),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1441),
    [sym_let_and_operator] = ACTIONS(1441),
    [sym_match_operator] = ACTIONS(1441),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(216)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(216),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1511),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(217)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7372),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(217),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1513),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(218)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3639),
    [sym__sequence_expression] = STATE(7414),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(218),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_done] = ACTIONS(1515),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(219)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3639),
    [sym__sequence_expression] = STATE(8317),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(219),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_done] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(220)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5472),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(220),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__infix_operator] = STATE(7564),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1417),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(1461),
    [anon_sym_DOT] = ACTIONS(1463),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_module] = ACTIONS(1465),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1469),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1469),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1469),
    [sym_let_and_operator] = ACTIONS(1469),
    [sym_match_operator] = ACTIONS(1469),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(221)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(221),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1519),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(222)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8184),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(222),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1521),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(223)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5536),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(223),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__infix_operator] = STATE(7464),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(709),
    [anon_sym_DOT] = ACTIONS(711),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_module] = ACTIONS(1483),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1381),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(717),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(717),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(717),
    [sym_let_and_operator] = ACTIONS(717),
    [sym_match_operator] = ACTIONS(717),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(224)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3639),
    [sym__sequence_expression] = STATE(7643),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(224),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_done] = ACTIONS(1523),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(225)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(225),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1525),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(226)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8024),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(226),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1527),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(227)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(227),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1529),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(228)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7372),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(228),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1531),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(229)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(229),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1533),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(230)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7974),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(230),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1535),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(231)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8235),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(231),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1537),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(232)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(232),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1539),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(233)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7843),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(233),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1541),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(234)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5206),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(234),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__infix_operator] = STATE(8042),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1471),
    [anon_sym_DOT] = ACTIONS(1473),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_module] = ACTIONS(1475),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1477),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1479),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1479),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1479),
    [sym_let_and_operator] = ACTIONS(1479),
    [sym_match_operator] = ACTIONS(1479),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(235)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(235),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1543),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(236)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7939),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(236),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1545),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(237)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(237),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1547),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(238)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(238),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1549),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(239)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7751),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(239),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1551),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(240)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(240),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1553),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(241)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7934),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(241),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1555),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(242)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(242),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1557),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(243)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8159),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(243),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1559),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(244)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(244),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1561),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(245)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7256),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(245),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1563),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(246)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(246),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1565),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(247)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5536),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(247),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__infix_operator] = STATE(7464),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(709),
    [anon_sym_DOT] = ACTIONS(711),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_module] = ACTIONS(1483),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(713),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(717),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(717),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(717),
    [sym_let_and_operator] = ACTIONS(717),
    [sym_match_operator] = ACTIONS(717),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(248)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5216),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(248),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__infix_operator] = STATE(7684),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1363),
    [anon_sym_DOT] = ACTIONS(1367),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_module] = ACTIONS(1377),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1481),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1403),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1403),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1403),
    [sym_let_and_operator] = ACTIONS(1403),
    [sym_match_operator] = ACTIONS(1403),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(249)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5421),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(249),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__infix_operator] = STATE(7464),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1417),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(709),
    [anon_sym_DOT] = ACTIONS(711),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_module] = ACTIONS(1483),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(713),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(717),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(717),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(717),
    [sym_let_and_operator] = ACTIONS(717),
    [sym_match_operator] = ACTIONS(717),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(250)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5437),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(250),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__infix_operator] = STATE(7244),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1421),
    [anon_sym_DOT] = ACTIONS(1423),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_module] = ACTIONS(1431),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1567),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1441),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1441),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1441),
    [sym_let_and_operator] = ACTIONS(1441),
    [sym_match_operator] = ACTIONS(1441),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(251)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5531),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(251),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__infix_operator] = STATE(7244),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1417),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(1421),
    [anon_sym_DOT] = ACTIONS(1423),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_module] = ACTIONS(1431),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1567),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1441),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1441),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1441),
    [sym_let_and_operator] = ACTIONS(1441),
    [sym_match_operator] = ACTIONS(1441),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(252)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5421),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(252),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__infix_operator] = STATE(7464),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1417),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(709),
    [anon_sym_DOT] = ACTIONS(711),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_module] = ACTIONS(1483),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1381),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(717),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(717),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(717),
    [sym_let_and_operator] = ACTIONS(717),
    [sym_match_operator] = ACTIONS(717),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(253)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(253),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1569),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(254)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5216),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(254),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__infix_operator] = STATE(7684),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1363),
    [anon_sym_DOT] = ACTIONS(1367),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_module] = ACTIONS(1377),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1381),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1403),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1403),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1403),
    [sym_let_and_operator] = ACTIONS(1403),
    [sym_match_operator] = ACTIONS(1403),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(255)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5437),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(255),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__infix_operator] = STATE(7244),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1421),
    [anon_sym_DOT] = ACTIONS(1423),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_module] = ACTIONS(1431),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1441),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1441),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1441),
    [sym_let_and_operator] = ACTIONS(1441),
    [sym_match_operator] = ACTIONS(1441),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(256)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8257),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(256),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_end] = ACTIONS(1571),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(257)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3639),
    [sym__sequence_expression] = STATE(7353),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(257),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_done] = ACTIONS(1573),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(258)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5538),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(258),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__infix_operator] = STATE(7379),
    [sym_pow_operator] = STATE(8068),
    [sym_mult_operator] = STATE(8068),
    [sym_add_operator] = STATE(8068),
    [sym_and_operator] = STATE(8068),
    [sym_or_operator] = STATE(8068),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1417),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(1447),
    [anon_sym_DOT] = ACTIONS(1449),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_module] = ACTIONS(1451),
    [anon_sym_AMP] = ACTIONS(569),
    [anon_sym_POUND] = ACTIONS(1379),
    [anon_sym_COLON_COLON] = ACTIONS(1453),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(1397),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym_prefix_operator] = ACTIONS(1455),
    [aux_sym_sign_operator_token1] = ACTIONS(585),
    [sym_hash_operator] = ACTIONS(1455),
    [anon_sym_lsl] = ACTIONS(579),
    [anon_sym_lsr] = ACTIONS(579),
    [anon_sym_asr] = ACTIONS(579),
    [aux_sym_pow_operator_token1] = ACTIONS(581),
    [anon_sym_mod] = ACTIONS(583),
    [anon_sym_land] = ACTIONS(583),
    [anon_sym_lor] = ACTIONS(583),
    [anon_sym_lxor] = ACTIONS(583),
    [aux_sym_mult_operator_token1] = ACTIONS(583),
    [aux_sym_add_operator_token1] = ACTIONS(585),
    [sym_concat_operator] = ACTIONS(587),
    [sym_rel_operator] = ACTIONS(589),
    [anon_sym_AMP_AMP] = ACTIONS(569),
    [anon_sym_or] = ACTIONS(591),
    [anon_sym_PIPE_PIPE] = ACTIONS(591),
    [sym_assign_operator] = ACTIONS(587),
    [sym_let_operator] = ACTIONS(1455),
    [sym_let_and_operator] = ACTIONS(1455),
    [sym_match_operator] = ACTIONS(1455),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(259)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2096),
    [sym__sequence_expression] = STATE(2704),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(259),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(260)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2139),
    [sym__sequence_expression] = STATE(2638),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(260),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(261)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2139),
    [sym__sequence_expression] = STATE(2653),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(261),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(262)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2139),
    [sym__sequence_expression] = STATE(2677),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(262),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(263)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2139),
    [sym__sequence_expression] = STATE(2679),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(263),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(264)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2139),
    [sym__sequence_expression] = STATE(2697),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(264),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(265)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5165),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(265),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(266)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3793),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7323),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(266),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1575),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(267)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2645),
    [sym__sequence_expression] = STATE(2704),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(267),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(268)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3170),
    [sym__sequence_expression] = STATE(2729),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(268),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(269)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2548),
    [sym__sequence_expression] = STATE(4140),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(269),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(270)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7128),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(270),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(271)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3170),
    [sym__sequence_expression] = STATE(2741),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(271),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(272)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3793),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(8147),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(272),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(273)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3769),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(8164),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(273),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1579),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(274)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3170),
    [sym__sequence_expression] = STATE(2736),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(274),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(275)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3138),
    [sym__sequence_expression] = STATE(2741),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(275),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(276)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3138),
    [sym__sequence_expression] = STATE(2736),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(276),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(277)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6826),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(277),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(278)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5096),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(278),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(279)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2658),
    [sym__sequence_expression] = STATE(4140),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(279),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(280)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2645),
    [sym__sequence_expression] = STATE(2638),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(280),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(281)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7528),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(281),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(282)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2658),
    [sym__sequence_expression] = STATE(4145),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(282),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(283)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2658),
    [sym__sequence_expression] = STATE(4146),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(283),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(284)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2658),
    [sym__sequence_expression] = STATE(4147),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(284),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(285)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2658),
    [sym__sequence_expression] = STATE(4154),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(285),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(286)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2658),
    [sym__sequence_expression] = STATE(4156),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(286),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(287)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2658),
    [sym__sequence_expression] = STATE(4157),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(287),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(288)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2658),
    [sym__sequence_expression] = STATE(4132),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(288),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(289)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3562),
    [sym__sequence_expression] = STATE(3557),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(289),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(290)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3562),
    [sym__sequence_expression] = STATE(3565),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(290),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(291)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3562),
    [sym__sequence_expression] = STATE(3566),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(291),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(292)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(7131),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(292),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(293)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3170),
    [sym__sequence_expression] = STATE(2718),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(293),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(294)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3562),
    [sym__sequence_expression] = STATE(3498),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(294),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(295)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3562),
    [sym__sequence_expression] = STATE(3500),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(295),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(296)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3562),
    [sym__sequence_expression] = STATE(3501),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(296),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(297)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3562),
    [sym__sequence_expression] = STATE(3510),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(297),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(298)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3461),
    [sym__sequence_expression] = STATE(3557),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(298),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(299)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3170),
    [sym__sequence_expression] = STATE(2743),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(299),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(300)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3461),
    [sym__sequence_expression] = STATE(3565),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(300),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(301)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3461),
    [sym__sequence_expression] = STATE(3566),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(301),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(302)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3461),
    [sym__sequence_expression] = STATE(3498),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(302),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(303)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3461),
    [sym__sequence_expression] = STATE(3500),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(303),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(304)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3461),
    [sym__sequence_expression] = STATE(3501),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(304),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(305)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3461),
    [sym__sequence_expression] = STATE(3510),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(305),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(306)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3170),
    [sym__sequence_expression] = STATE(2745),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(306),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(307)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3793),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7500),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(307),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1581),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(308)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3769),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7816),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(308),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1583),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(309)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3639),
    [sym__sequence_expression] = STATE(6776),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(309),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(310)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3138),
    [sym__sequence_expression] = STATE(2718),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(310),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(311)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3138),
    [sym__sequence_expression] = STATE(2743),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(311),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(312)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7152),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(312),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(313)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3138),
    [sym__sequence_expression] = STATE(2745),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(313),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(314)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(6975),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(314),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(315)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2273),
    [sym__sequence_expression] = STATE(4140),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(315),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(316)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7686),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(316),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(317)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3138),
    [sym__sequence_expression] = STATE(2734),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(317),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(318)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3170),
    [sym__sequence_expression] = STATE(2734),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(318),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(319)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2273),
    [sym__sequence_expression] = STATE(4145),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(319),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(320)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2273),
    [sym__sequence_expression] = STATE(4146),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(320),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(321)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2273),
    [sym__sequence_expression] = STATE(4147),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(321),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(322)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2273),
    [sym__sequence_expression] = STATE(4154),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(322),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(323)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2273),
    [sym__sequence_expression] = STATE(4156),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(323),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(324)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2273),
    [sym__sequence_expression] = STATE(4157),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(324),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(325)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2273),
    [sym__sequence_expression] = STATE(4132),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(325),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(326)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3816),
    [sym__sequence_expression] = STATE(6804),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(7765),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(326),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(327)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2550),
    [sym__sequence_expression] = STATE(2670),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(327),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(328)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7282),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(328),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(329)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3793),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7781),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(329),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1585),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(330)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3769),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7837),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(330),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1587),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(331)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7371),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(331),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(332)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7072),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(332),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(333)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2274),
    [sym__sequence_expression] = STATE(2670),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(333),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(334)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2550),
    [sym__sequence_expression] = STATE(2704),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(334),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(335)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2564),
    [sym__sequence_expression] = STATE(4140),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(335),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(336)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2564),
    [sym__sequence_expression] = STATE(4145),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(336),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(337)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2564),
    [sym__sequence_expression] = STATE(4146),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(337),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(338)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2564),
    [sym__sequence_expression] = STATE(4147),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(338),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(339)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2564),
    [sym__sequence_expression] = STATE(4154),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(339),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(340)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2564),
    [sym__sequence_expression] = STATE(4156),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(340),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(341)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2564),
    [sym__sequence_expression] = STATE(4157),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(341),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(342)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2564),
    [sym__sequence_expression] = STATE(4132),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(342),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(343)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2550),
    [sym__sequence_expression] = STATE(2638),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(343),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(344)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7512),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(344),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(345)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3793),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7432),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(345),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1589),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(346)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3769),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7515),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(346),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1591),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(347)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7522),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(347),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(348)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5108),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(348),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(349)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6899),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(349),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(350)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2274),
    [sym__sequence_expression] = STATE(2704),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(350),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(351)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2550),
    [sym__sequence_expression] = STATE(2653),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(351),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(352)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2550),
    [sym__sequence_expression] = STATE(2677),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(352),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(353)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3793),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7238),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(353),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1593),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(354)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3769),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7819),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(354),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1595),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(355)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2550),
    [sym__sequence_expression] = STATE(2679),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(355),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(356)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2274),
    [sym__sequence_expression] = STATE(2638),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(356),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(357)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7092),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(357),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(358)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5111),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(358),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(359)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2550),
    [sym__sequence_expression] = STATE(2697),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(359),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(360)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3793),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7550),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(360),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1597),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(361)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3769),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7610),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(361),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1599),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(362)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2186),
    [sym__sequence_expression] = STATE(2670),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(362),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(363)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2274),
    [sym__sequence_expression] = STATE(2653),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(363),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(364)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6917),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(364),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(365)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2274),
    [sym__sequence_expression] = STATE(2677),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(365),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(366)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7927),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(366),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(367)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2274),
    [sym__sequence_expression] = STATE(2679),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(367),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(368)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2186),
    [sym__sequence_expression] = STATE(2704),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(368),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(369)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3793),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7259),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(369),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1601),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(370)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3769),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7277),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(370),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1603),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(371)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2186),
    [sym__sequence_expression] = STATE(2638),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(371),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(372)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2274),
    [sym__sequence_expression] = STATE(2697),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(372),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(373)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7112),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(373),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(374)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3407),
    [sym__sequence_expression] = STATE(3557),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(374),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(375)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2548),
    [sym__sequence_expression] = STATE(4145),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(375),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(376)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2548),
    [sym__sequence_expression] = STATE(4146),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(376),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(377)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3793),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7583),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(377),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1605),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(378)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3769),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7597),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(378),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1607),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(379)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2186),
    [sym__sequence_expression] = STATE(2653),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(379),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(380)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5114),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(380),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(381)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3639),
    [sym__sequence_expression] = STATE(7121),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(381),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(382)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2186),
    [sym__sequence_expression] = STATE(2677),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(382),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(383)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2186),
    [sym__sequence_expression] = STATE(2679),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(383),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(384)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2548),
    [sym__sequence_expression] = STATE(4147),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(384),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(385)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3793),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7876),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(385),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1609),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(386)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3769),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7886),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(386),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1611),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(387)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2186),
    [sym__sequence_expression] = STATE(2697),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(387),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(388)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3407),
    [sym__sequence_expression] = STATE(3565),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(388),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(389)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3407),
    [sym__sequence_expression] = STATE(3566),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(389),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(390)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3493),
    [sym__sequence_expression] = STATE(3557),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(390),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(391)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2645),
    [sym__sequence_expression] = STATE(2653),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(391),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(392)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3493),
    [sym__sequence_expression] = STATE(3565),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(392),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(393)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3793),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(8051),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(393),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1613),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(394)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3769),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(8064),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(394),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1615),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(395)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3493),
    [sym__sequence_expression] = STATE(3566),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(395),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(396)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5115),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(396),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(397)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3493),
    [sym__sequence_expression] = STATE(3498),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(397),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(398)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3793),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(8328),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(398),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1617),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(399)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3769),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(8343),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(399),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1619),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(400)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3493),
    [sym__sequence_expression] = STATE(3500),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(400),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(401)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3407),
    [sym__sequence_expression] = STATE(3498),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(401),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(402)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3407),
    [sym__sequence_expression] = STATE(3500),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(402),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(403)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3493),
    [sym__sequence_expression] = STATE(3501),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(403),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(404)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5139),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(404),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(405)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3493),
    [sym__sequence_expression] = STATE(3510),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(405),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(406)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5146),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(406),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(407)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3407),
    [sym__sequence_expression] = STATE(3501),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(407),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(408)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8263),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(408),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(409)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3106),
    [sym__sequence_expression] = STATE(2838),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(409),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(410)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2645),
    [sym__sequence_expression] = STATE(2677),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(410),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(411)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2645),
    [sym__sequence_expression] = STATE(2679),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(411),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(412)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3106),
    [sym__sequence_expression] = STATE(2756),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(412),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(413)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3106),
    [sym__sequence_expression] = STATE(2761),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(413),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(414)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3639),
    [sym__sequence_expression] = STATE(6873),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(414),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(415)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3407),
    [sym__sequence_expression] = STATE(3510),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3571),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(415),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(416)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3106),
    [sym__sequence_expression] = STATE(2780),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(416),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(417)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3106),
    [sym__sequence_expression] = STATE(2799),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(417),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(418)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3106),
    [sym__sequence_expression] = STATE(2800),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(418),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(419)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3769),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7264),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(419),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1621),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(420)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3106),
    [sym__sequence_expression] = STATE(2817),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(420),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(421)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7322),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(421),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(422)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(6867),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(422),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(423)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3077),
    [sym__sequence_expression] = STATE(2838),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(423),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(424)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2096),
    [sym__sequence_expression] = STATE(2670),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(424),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(425)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7393),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(425),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(426)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7394),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(426),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(427)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7401),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(427),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(428)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7599),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(428),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(429)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7600),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(429),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(430)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7605),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(430),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(431)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7772),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(431),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(432)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5128),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(432),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(433)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(7038),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(433),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(434)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7810),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(434),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(435)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(7052),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(435),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(436)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7840),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(436),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(437)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7849),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(437),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(438)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3676),
    [sym__sequence_expression] = STATE(2729),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(438),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(439)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7324),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(439),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(440)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3676),
    [sym__sequence_expression] = STATE(2741),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(440),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(441)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3676),
    [sym__sequence_expression] = STATE(2736),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(441),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(442)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3676),
    [sym__sequence_expression] = STATE(2718),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(442),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(443)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3676),
    [sym__sequence_expression] = STATE(2743),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(443),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(444)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3676),
    [sym__sequence_expression] = STATE(2745),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(444),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(445)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3676),
    [sym__sequence_expression] = STATE(2734),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(445),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(446)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3138),
    [sym__sequence_expression] = STATE(2729),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(446),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(447)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2096),
    [sym__sequence_expression] = STATE(2638),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(447),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(448)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5088),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(448),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(449)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6775),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(449),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(450)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3077),
    [sym__sequence_expression] = STATE(2761),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(450),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(451)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7391),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(451),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(452)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6784),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(452),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(453)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6791),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(453),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(454)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6799),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(454),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(455)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6809),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(455),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(456)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2096),
    [sym__sequence_expression] = STATE(2653),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(456),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(457)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(7181),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(457),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(458)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7862),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(458),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(459)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7883),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(459),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(460)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7950),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(460),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(461)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8165),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(461),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(462)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7215),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(462),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(463)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7252),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(463),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(464)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(7126),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(464),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(465)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(7062),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(465),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(466)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6870),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(466),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(467)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2096),
    [sym__sequence_expression] = STATE(2677),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(467),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(468)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2096),
    [sym__sequence_expression] = STATE(2679),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(468),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(469)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(7047),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(469),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(470)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7892),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(470),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(471)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7914),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(471),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(472)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7971),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(472),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(473)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7807),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(473),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(474)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7838),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(474),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(475)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7923),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(475),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(476)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(6888),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(476),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(477)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(6900),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(477),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(478)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7325),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(478),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(479)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8213),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(479),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(480)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8218),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(480),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(481)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8219),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(481),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(482)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7938),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(482),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(483)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8067),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(483),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(484)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8346),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(484),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(485)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2096),
    [sym__sequence_expression] = STATE(2697),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(485),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(486)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8098),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(486),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(487)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8110),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(487),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(488)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8118),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(488),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(489)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8185),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(489),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(490)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8190),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(490),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(491)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8200),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(491),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(492)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2645),
    [sym__sequence_expression] = STATE(2670),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(492),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(493)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8121),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(493),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(494)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8144),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(494),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(495)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8242),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(495),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(496)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8351),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(496),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(497)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7198),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(497),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(498)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7212),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(498),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(499)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2548),
    [sym__sequence_expression] = STATE(4154),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(499),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(500)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8119),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(500),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(501)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8124),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(501),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(502)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8133),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(502),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(503)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8290),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(503),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(504)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8301),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(504),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(505)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8309),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(505),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(506)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2548),
    [sym__sequence_expression] = STATE(4156),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(506),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(507)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7442),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(507),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(508)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7447),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(508),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(509)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7448),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(509),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(510)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7470),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(510),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(511)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7476),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(511),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(512)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7479),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(512),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(513)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2548),
    [sym__sequence_expression] = STATE(4157),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(513),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(514)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7786),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(514),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(515)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7789),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(515),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(516)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7792),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(516),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(517)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7817),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(517),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(518)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7818),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(518),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(519)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7827),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(519),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(520)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5040),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(520),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(521)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7962),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(521),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(522)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7964),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(522),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(523)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7965),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(523),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(524)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7984),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(524),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(525)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7986),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(525),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(526)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7987),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(526),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(527)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5046),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(527),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(528)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8232),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(528),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(529)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8233),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(529),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(530)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8234),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(530),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(531)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8267),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(531),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(532)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8269),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(532),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(533)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8270),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(533),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(534)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5050),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(534),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(535)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7575),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(535),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(536)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7596),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(536),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(537)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7629),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(537),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(538)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7723),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(538),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(539)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7734),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(539),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(540)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7749),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(540),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(541)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5058),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(541),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(542)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5059),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(542),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(543)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3793),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7436),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(543),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1623),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(544)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3769),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym__sequence_expression_content] = STATE(7544),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(544),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1625),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(545)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7533),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(545),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(546)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7763),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(546),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(547)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7776),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(547),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(548)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3788),
    [sym__sequence_expression] = STATE(2729),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(548),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(549)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(8311),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(549),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(550)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3788),
    [sym__sequence_expression] = STATE(2741),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(550),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(551)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3788),
    [sym__sequence_expression] = STATE(2736),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(551),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(552)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3788),
    [sym__sequence_expression] = STATE(2718),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(552),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(553)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3788),
    [sym__sequence_expression] = STATE(2743),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(553),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(554)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3788),
    [sym__sequence_expression] = STATE(2745),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(554),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(555)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3788),
    [sym__sequence_expression] = STATE(2734),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(555),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(556)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3077),
    [sym__sequence_expression] = STATE(2780),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(556),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(557)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7080),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(557),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(558)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3077),
    [sym__sequence_expression] = STATE(2799),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(558),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(559)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2645),
    [sym__sequence_expression] = STATE(2697),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(559),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(560)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6825),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(560),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(561)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7766),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(561),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(562)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7850),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(562),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(563)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7854),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(563),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(564)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3077),
    [sym__sequence_expression] = STATE(2800),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(564),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(565)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3077),
    [sym__sequence_expression] = STATE(2817),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(565),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(566)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7197),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(566),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(567)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7219),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(567),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(568)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7221),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(568),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(569)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3851),
    [sym__sequence_expression] = STATE(6771),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(569),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(570)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2548),
    [sym__sequence_expression] = STATE(4132),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(570),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(571)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7275),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(571),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(572)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7290),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(572),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(573)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7292),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(573),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(574)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5079),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(574),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(575)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3543),
    [sym__sequence_expression] = STATE(5594),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(575),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(576)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7336),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(576),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(577)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7347),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(577),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(578)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7349),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(578),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(579)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5080),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(579),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(580)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2139),
    [sym__sequence_expression] = STATE(2670),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(580),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(581)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7382),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(581),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(582)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7389),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(582),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(583)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7390),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(583),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(584)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5085),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(584),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(585)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(6804),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(585),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(586)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7421),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(586),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(587)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7428),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(587),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(588)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7429),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(588),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(589)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3180),
    [sym__sequence_expression] = STATE(5086),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(589),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(590)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7458),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(590),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(591)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7465),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(591),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(592)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7466),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(592),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(593)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3543),
    [sym__sequence_expression] = STATE(5736),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(593),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(594)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3543),
    [sym__sequence_expression] = STATE(5737),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(594),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(595)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7494),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(595),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(596)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7501),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(596),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(597)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7502),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(597),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(598)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3543),
    [sym__sequence_expression] = STATE(5738),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(598),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(599)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3543),
    [sym__sequence_expression] = STATE(5616),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(599),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(600)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7530),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(600),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(601)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7537),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(601),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(602)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7538),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(602),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(603)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3543),
    [sym__sequence_expression] = STATE(5553),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(603),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(604)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3543),
    [sym__sequence_expression] = STATE(5574),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(604),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(605)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7565),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(605),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(606)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7572),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(606),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(607)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7573),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(607),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(608)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2139),
    [sym__sequence_expression] = STATE(2704),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(608),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(609)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3543),
    [sym__sequence_expression] = STATE(5550),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3085),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(609),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(610)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7601),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(610),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(611)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7608),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(611),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(612)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3234),
    [sym__sequence_expression] = STATE(7609),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1864),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(612),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(613)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3639),
    [sym__sequence_expression] = STATE(7050),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(613),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(614)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3639),
    [sym__sequence_expression] = STATE(7060),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(614),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(615)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3639),
    [sym__sequence_expression] = STATE(7108),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(615),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(616)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3639),
    [sym__sequence_expression] = STATE(7110),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(616),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(617)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3639),
    [sym__sequence_expression] = STATE(7120),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(2159),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(617),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(618)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3077),
    [sym__sequence_expression] = STATE(2756),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2835),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(618),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(619)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3422),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3584),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(619),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(620)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2209),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1875),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(620),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(621)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2672),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1875),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(621),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(622)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2158),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1875),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(622),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(623)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2715),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1875),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(623),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(624)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3076),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2764),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(624),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(625)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3129),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(2764),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(625),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(626)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3813),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(626),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1627),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(627)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3253),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1875),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(627),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(628)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2085),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2164),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(628),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(629)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3787),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3079),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(629),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(630)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2539),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1875),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(630),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(631)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3813),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(631),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1629),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(632)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3826),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(632),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1627),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(633)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3820),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(7730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(633),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(634)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2633),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1875),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(634),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(635)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2372),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2164),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(635),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(636)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3144),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3079),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(636),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(637)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3666),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(3079),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(637),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(638)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2636),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1875),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(638),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(639)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3612),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3584),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(639),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(640)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2169),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2164),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(640),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(641)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2614),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(2164),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(641),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(642)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2759),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym__sequence_expression_anonymous] = STATE(1875),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(642),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(643)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3460),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3584),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(643),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(644)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3107),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym__sequence_expression_anonymous] = STATE(2764),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(644),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(645)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3471),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3079),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(645),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(646)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3855),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1875),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(646),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(647)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3826),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(647),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_RBRACK] = ACTIONS(1629),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(648)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3569),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(3079),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(648),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(649)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3206),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym__sequence_expression_anonymous] = STATE(1875),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(649),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(650)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3736),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym__sequence_expression_anonymous] = STATE(2164),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(650),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(651)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3772),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(4142),
    [sym__tuple_expression] = STATE(2713),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(651),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(1631),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(652)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2187),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(652),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(653)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2168),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(2752),
    [sym__tuple_expression] = STATE(2651),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(653),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(1633),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(654)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2132),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(654),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(655)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2163),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(655),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(656)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2165),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(656),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(657)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2120),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(657),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(658)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2129),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(658),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(659)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2130),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(659),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(660)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2131),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(660),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(661)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2180),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(661),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(662)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2136),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(662),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(663)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2124),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(663),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(664)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2125),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(664),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(665)] = {
    [sym_value_definition] = STATE(7490),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2121),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7346),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1143),
    [sym_array_get_expression] = STATE(1143),
    [sym_string_get_expression] = STATE(1143),
    [sym_bigarray_get_expression] = STATE(1143),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(665),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(652),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7352),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(185),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(201),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(203),
    [anon_sym_function] = ACTIONS(205),
    [anon_sym_try] = ACTIONS(207),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(229),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(666)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3485),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(666),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(667)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3470),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(3929),
    [sym__tuple_expression] = STATE(3559),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(667),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(1635),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(668)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3495),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(668),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(669)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3472),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(669),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(670)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3474),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(670),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(671)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3475),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(671),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(672)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3486),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(672),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(673)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3487),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(673),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(674)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3488),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(674),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(675)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3476),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(675),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(676)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3477),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(676),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(677)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3478),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(677),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(678)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3808),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(678),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(679)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3479),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(679),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(680)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3731),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(680),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(681)] = {
    [sym_value_definition] = STATE(7748),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3483),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(8298),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2016),
    [sym_array_get_expression] = STATE(2016),
    [sym_string_get_expression] = STATE(2016),
    [sym_bigarray_get_expression] = STATE(2016),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(681),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(666),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7281),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(863),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(865),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(867),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(682)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3090),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(682),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(683)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3092),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(3298),
    [sym__tuple_expression] = STATE(2842),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(683),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(1637),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(684)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3093),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(684),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(685)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3094),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(685),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(686)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3095),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(686),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(687)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3096),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(687),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(688)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3087),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(688),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(689)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3091),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(689),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(690)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3088),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(690),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(691)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3097),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(691),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(692)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3098),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(692),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(693)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3099),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(693),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(694)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3100),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(694),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(695)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2639),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(695),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(696)] = {
    [sym_value_definition] = STATE(7402),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3102),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7833),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1773),
    [sym_array_get_expression] = STATE(1773),
    [sym_string_get_expression] = STATE(1773),
    [sym_bigarray_get_expression] = STATE(1773),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(696),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(682),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7204),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(465),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(469),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(471),
    [anon_sym_function] = ACTIONS(473),
    [anon_sym_try] = ACTIONS(475),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(697)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2090),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(697),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(698)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2099),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(2744),
    [sym__tuple_expression] = STATE(2651),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(698),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(1639),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(699)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2101),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(699),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(700)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2103),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(700),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(701)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2104),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(701),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(702)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2098),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(702),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(703)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2092),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(703),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(704)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2093),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(704),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(705)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2094),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(705),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(706)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2082),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(706),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(707)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2079),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(707),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(708)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2087),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(708),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(709)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2091),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(709),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(710)] = {
    [sym_value_definition] = STATE(7257),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2097),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7398),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1137),
    [sym_array_get_expression] = STATE(1137),
    [sym_string_get_expression] = STATE(1137),
    [sym_bigarray_get_expression] = STATE(1137),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(710),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(697),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7418),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(109),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(127),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(163),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(711)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3827),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(711),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(712)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3538),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(712),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(713)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3603),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(3953),
    [sym__tuple_expression] = STATE(3559),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(713),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(1635),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(714)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3605),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(714),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(715)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3606),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(715),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(716)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3607),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(716),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(717)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3608),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(717),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(718)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3540),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(718),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(719)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3496),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(719),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(720)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3542),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(720),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(721)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3609),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(721),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(722)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3610),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(722),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(723)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3611),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(723),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(724)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3541),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(724),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(725)] = {
    [sym__parenthesized_abstract_type] = STATE(1704),
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2246),
    [sym__binding_pattern] = STATE(2358),
    [sym_alias_binding_pattern] = STATE(2246),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_anonymous] = STATE(2248),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7706),
    [sym__tuple_binding_pattern] = STATE(2249),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern] = STATE(2246),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2246),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_exception_binding_pattern] = STATE(2246),
    [sym_effect_binding_pattern] = STATE(2246),
    [sym_attribute] = STATE(725),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [ts_builtin_sym_end] = ACTIONS(1641),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_SEMI_SEMI] = ACTIONS(1641),
    [anon_sym_let] = ACTIONS(1645),
    [anon_sym_and] = ACTIONS(1645),
    [anon_sym_EQ] = ACTIONS(1641),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_QMARK] = ACTIONS(1641),
    [anon_sym_LPAREN] = ACTIONS(1649),
    [anon_sym_external] = ACTIONS(1645),
    [anon_sym_type] = ACTIONS(1645),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_PIPE] = ACTIONS(1641),
    [anon_sym_COLON2] = ACTIONS(1645),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_RBRACK] = ACTIONS(1641),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1657),
    [anon_sym_module] = ACTIONS(1645),
    [anon_sym_open] = ACTIONS(1645),
    [anon_sym_include] = ACTIONS(1645),
    [anon_sym_class] = ACTIONS(1645),
    [anon_sym_val] = ACTIONS(1645),
    [anon_sym_end] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1641),
    [anon_sym_in] = ACTIONS(1645),
    [anon_sym_as] = ACTIONS(1645),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_COLON_COLON] = ACTIONS(1641),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [anon_sym_effect] = ACTIONS(1667),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1645),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1641),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1669),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1671),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1641),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1641),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym_let_operator] = ACTIONS(1641),
    [sym_let_and_operator] = ACTIONS(1641),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_directive_token1] = ACTIONS(1645),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(726)] = {
    [sym_value_definition] = STATE(7725),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3581),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7535),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(2043),
    [sym_array_get_expression] = STATE(2043),
    [sym_string_get_expression] = STATE(2043),
    [sym_bigarray_get_expression] = STATE(2043),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(726),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(712),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7561),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(883),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(887),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(727)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3798),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(727),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(728)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2680),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(728),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(729)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3450),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(729),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(730)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3452),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(3872),
    [sym__tuple_expression] = STATE(3559),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(730),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(1635),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(731)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3458),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(731),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(732)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3462),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(732),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(733)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3463),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(733),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(734)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3464),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(734),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(735)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3453),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(735),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(736)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3447),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(736),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(737)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3454),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(737),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(738)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3465),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(738),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(739)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3466),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(739),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(740)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3467),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(740),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(741)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3480),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(741),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(742)] = {
    [sym_value_definition] = STATE(7868),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3489),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7225),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1953),
    [sym_array_get_expression] = STATE(1953),
    [sym_string_get_expression] = STATE(1953),
    [sym_bigarray_get_expression] = STATE(1953),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(742),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(729),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7227),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(871),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(873),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(875),
    [anon_sym_function] = ACTIONS(877),
    [anon_sym_try] = ACTIONS(879),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(881),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(743)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3826),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(743),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(744)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3128),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(744),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(745)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3122),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(3428),
    [sym__tuple_expression] = STATE(2713),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(745),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(1631),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(746)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2700),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(746),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(747)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3146),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(747),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(748)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3147),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(748),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(749)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3148),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(749),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(750)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3151),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(750),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(751)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3153),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(751),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(752)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3158),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(752),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(753)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3164),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(753),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(754)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3161),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(754),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(755)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3145),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(755),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(756)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3159),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(756),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(757)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2650),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(3050),
    [sym__tuple_expression] = STATE(2651),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(757),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(1633),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(758)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3137),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(758),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(759)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2656),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(759),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(760)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2657),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(760),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(761)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2659),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(761),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(762)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2660),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(762),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(763)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2662),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(763),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(764)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3732),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(764),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(765)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2663),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(765),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(766)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3813),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(766),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(767)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2664),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(767),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(768)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2666),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(768),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(769)] = {
    [sym_value_definition] = STATE(7229),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3127),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7303),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1774),
    [sym_array_get_expression] = STATE(1774),
    [sym_string_get_expression] = STATE(1774),
    [sym_bigarray_get_expression] = STATE(1774),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(769),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(744),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7304),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(545),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(547),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(551),
    [anon_sym_function] = ACTIONS(553),
    [anon_sym_try] = ACTIONS(555),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(561),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(770)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2667),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(770),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(771)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3626),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(771),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(772)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3627),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(3972),
    [sym__tuple_expression] = STATE(2713),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(772),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(1631),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(773)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3628),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(773),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(774)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3629),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(774),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(775)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3630),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(775),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(776)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3631),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(776),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(777)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3632),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(777),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(778)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3633),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(778),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(779)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3634),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(779),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(780)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3635),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(780),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(781)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3636),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(781),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(782)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3637),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(782),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(783)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3640),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(783),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(784)] = {
    [sym_value_definition] = STATE(7397),
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(112),
    [sym__expression] = STATE(2668),
    [sym_typed_expression] = STATE(1471),
    [sym_labeled_tuple_element] = STATE(7805),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1471),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1471),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1416),
    [sym_array_get_expression] = STATE(1416),
    [sym_string_get_expression] = STATE(1416),
    [sym_bigarray_get_expression] = STATE(1416),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(784),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym_sign_operator] = STATE(746),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__instance_variable_name] = STATE(7744),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [sym__lowercase_identifier] = ACTIONS(9),
    [anon_sym_let] = ACTIONS(301),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(45),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_if] = ACTIONS(49),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(55),
    [anon_sym_function] = ACTIONS(57),
    [anon_sym_try] = ACTIONS(59),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1487),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1489),
    [aux_sym_number_token1] = ACTIONS(83),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(97),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(785)] = {
    [sym_value_definition] = STATE(7741),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3643),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7488),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2074),
    [sym_array_get_expression] = STATE(2074),
    [sym_string_get_expression] = STATE(2074),
    [sym_bigarray_get_expression] = STATE(2074),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(785),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(771),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7499),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(891),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(893),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(895),
    [anon_sym_function] = ACTIONS(897),
    [anon_sym_try] = ACTIONS(899),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(901),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(786)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2211),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(786),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(787)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2227),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(2884),
    [sym__tuple_expression] = STATE(2651),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(787),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(1633),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(788)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2232),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(788),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(789)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2233),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(789),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(790)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2234),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(790),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(791)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2217),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(791),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(792)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2221),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(792),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(793)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2223),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(793),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(794)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2235),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(794),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(795)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2237),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(795),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(796)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2244),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(796),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(797)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2245),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(797),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(798)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2255),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(798),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(799)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3424),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(799),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(800)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3433),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(3802),
    [sym__tuple_expression] = STATE(3559),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(800),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(1635),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(801)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3435),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(801),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(802)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3436),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(802),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(803)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3439),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(803),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(804)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3442),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(804),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(805)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3432),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(805),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(806)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3440),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(806),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(807)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3444),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(807),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(808)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3438),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(808),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(809)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3409),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(809),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(810)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3390),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(810),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(811)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3396),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(811),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(812)] = {
    [sym_value_definition] = STATE(8047),
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(884),
    [sym__expression] = STATE(3395),
    [sym_typed_expression] = STATE(1977),
    [sym_labeled_tuple_element] = STATE(7679),
    [sym__tuple_expression] = STATE(3564),
    [sym_cons_expression] = STATE(3597),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_application_expression] = STATE(3597),
    [sym_prefix_expression] = STATE(1977),
    [sym_sign_expression] = STATE(3597),
    [sym_hash_expression] = STATE(1977),
    [sym_infix_expression] = STATE(3597),
    [sym_field_get_expression] = STATE(1944),
    [sym_array_get_expression] = STATE(1944),
    [sym_string_get_expression] = STATE(1944),
    [sym_bigarray_get_expression] = STATE(1944),
    [sym_set_expression] = STATE(3597),
    [sym_if_expression] = STATE(3597),
    [sym_while_expression] = STATE(3597),
    [sym_for_expression] = STATE(3597),
    [sym_match_expression] = STATE(3597),
    [sym_function_expression] = STATE(3597),
    [sym_fun_expression] = STATE(3597),
    [sym_try_expression] = STATE(3597),
    [sym_let_expression] = STATE(3597),
    [sym_assert_expression] = STATE(3597),
    [sym_lazy_expression] = STATE(3597),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(812),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym_sign_operator] = STATE(799),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__instance_variable_name] = STATE(7683),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [sym__lowercase_identifier] = ACTIONS(803),
    [anon_sym_let] = ACTIONS(805),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_fun] = ACTIONS(817),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_if] = ACTIONS(821),
    [anon_sym_while] = ACTIONS(823),
    [anon_sym_for] = ACTIONS(825),
    [anon_sym_match] = ACTIONS(827),
    [anon_sym_function] = ACTIONS(829),
    [anon_sym_try] = ACTIONS(831),
    [anon_sym_assert] = ACTIONS(833),
    [anon_sym_lazy] = ACTIONS(835),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(857),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(813)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3082),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(813),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(814)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3065),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(3238),
    [sym__tuple_expression] = STATE(2842),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(814),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(1637),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(815)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3066),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(815),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(816)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3067),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(816),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(817)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3068),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(817),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(818)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3069),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(818),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(819)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3083),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(819),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(820)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3063),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(820),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(821)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3084),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(821),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(822)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3070),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(822),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(823)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3071),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(823),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(824)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3072),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(824),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(825)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3073),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(825),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(826)] = {
    [sym_value_definition] = STATE(8205),
    [sym__delimited_expression] = STATE(1045),
    [sym__simple_expression] = STATE(101),
    [sym__expression] = STATE(3074),
    [sym_typed_expression] = STATE(1045),
    [sym_labeled_tuple_element] = STATE(7687),
    [sym__tuple_expression] = STATE(2825),
    [sym_cons_expression] = STATE(2836),
    [sym_list_expression] = STATE(1046),
    [sym_array_expression] = STATE(1046),
    [sym_record_expression] = STATE(1046),
    [sym_application_expression] = STATE(2836),
    [sym_prefix_expression] = STATE(1045),
    [sym_sign_expression] = STATE(2836),
    [sym_hash_expression] = STATE(1045),
    [sym_infix_expression] = STATE(2836),
    [sym_field_get_expression] = STATE(1716),
    [sym_array_get_expression] = STATE(1716),
    [sym_string_get_expression] = STATE(1716),
    [sym_bigarray_get_expression] = STATE(1716),
    [sym_set_expression] = STATE(2836),
    [sym_if_expression] = STATE(2836),
    [sym_while_expression] = STATE(2836),
    [sym_for_expression] = STATE(2836),
    [sym_match_expression] = STATE(2836),
    [sym_function_expression] = STATE(2836),
    [sym_fun_expression] = STATE(2836),
    [sym_try_expression] = STATE(2836),
    [sym_let_expression] = STATE(2836),
    [sym_assert_expression] = STATE(2836),
    [sym_lazy_expression] = STATE(2836),
    [sym_local_open_expression] = STATE(1045),
    [sym_package_expression] = STATE(1046),
    [sym_new_expression] = STATE(1045),
    [sym_object_copy_expression] = STATE(1046),
    [sym_method_invocation] = STATE(1045),
    [sym_object_expression] = STATE(1045),
    [sym_parenthesized_expression] = STATE(1046),
    [sym_attribute] = STATE(826),
    [sym_extension] = STATE(1045),
    [sym_quoted_extension] = STATE(1045),
    [sym__constant] = STATE(1045),
    [sym_number] = STATE(1106),
    [sym_character] = STATE(1106),
    [sym_string] = STATE(1106),
    [sym_quoted_string] = STATE(1106),
    [sym_boolean] = STATE(1106),
    [sym_unit] = STATE(1046),
    [sym_sign_operator] = STATE(813),
    [sym__value_name] = STATE(1073),
    [sym_parenthesized_operator] = STATE(1074),
    [sym_value_path] = STATE(1045),
    [sym_module_path] = STATE(7491),
    [sym_constructor_path] = STATE(1045),
    [sym__instance_variable_name] = STATE(7688),
    [sym__constructor_name] = STATE(1107),
    [sym_tag] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(403),
    [anon_sym_let] = ACTIONS(405),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(407),
    [anon_sym_LBRACK] = ACTIONS(409),
    [anon_sym_true] = ACTIONS(411),
    [anon_sym_false] = ACTIONS(411),
    [anon_sym_LBRACE] = ACTIONS(413),
    [anon_sym_object] = ACTIONS(415),
    [anon_sym_fun] = ACTIONS(417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(419),
    [anon_sym_if] = ACTIONS(421),
    [anon_sym_while] = ACTIONS(423),
    [anon_sym_for] = ACTIONS(425),
    [anon_sym_match] = ACTIONS(427),
    [anon_sym_function] = ACTIONS(429),
    [anon_sym_try] = ACTIONS(431),
    [anon_sym_assert] = ACTIONS(433),
    [anon_sym_lazy] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(437),
    [anon_sym_LBRACE_LT] = ACTIONS(439),
    [anon_sym_begin] = ACTIONS(441),
    [sym_ocamlyacc_value] = ACTIONS(443),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(479),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(481),
    [aux_sym_number_token1] = ACTIONS(449),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(451),
    [anon_sym_DQUOTE] = ACTIONS(453),
    [sym_prefix_operator] = ACTIONS(455),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(457),
    [sym__uppercase_identifier] = ACTIONS(459),
    [aux_sym_tag_token1] = ACTIONS(461),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(827)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2141),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(827),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(828)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2188),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(2813),
    [sym__tuple_expression] = STATE(2651),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(828),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(1639),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(829)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2170),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(829),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(830)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2172),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(830),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(831)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2177),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(831),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(832)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2181),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(832),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(833)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2146),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(833),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(834)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2147),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(834),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(835)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2156),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(835),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(836)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2182),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(836),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(837)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2184),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(837),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(838)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2111),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(838),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(839)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2112),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(839),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(840)] = {
    [sym_value_definition] = STATE(7571),
    [sym__delimited_expression] = STATE(1189),
    [sym__simple_expression] = STATE(103),
    [sym__expression] = STATE(2116),
    [sym_typed_expression] = STATE(1189),
    [sym_labeled_tuple_element] = STATE(7694),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1190),
    [sym_array_expression] = STATE(1190),
    [sym_record_expression] = STATE(1190),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1189),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1189),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1217),
    [sym_array_get_expression] = STATE(1217),
    [sym_string_get_expression] = STATE(1217),
    [sym_bigarray_get_expression] = STATE(1217),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1189),
    [sym_package_expression] = STATE(1190),
    [sym_new_expression] = STATE(1189),
    [sym_object_copy_expression] = STATE(1190),
    [sym_method_invocation] = STATE(1189),
    [sym_object_expression] = STATE(1189),
    [sym_parenthesized_expression] = STATE(1190),
    [sym_attribute] = STATE(840),
    [sym_extension] = STATE(1189),
    [sym_quoted_extension] = STATE(1189),
    [sym__constant] = STATE(1189),
    [sym_number] = STATE(1141),
    [sym_character] = STATE(1141),
    [sym_string] = STATE(1141),
    [sym_quoted_string] = STATE(1141),
    [sym_boolean] = STATE(1141),
    [sym_unit] = STATE(1190),
    [sym_sign_operator] = STATE(827),
    [sym__value_name] = STATE(1153),
    [sym_parenthesized_operator] = STATE(1154),
    [sym_value_path] = STATE(1189),
    [sym_module_path] = STATE(7562),
    [sym_constructor_path] = STATE(1189),
    [sym__instance_variable_name] = STATE(7696),
    [sym__constructor_name] = STATE(1145),
    [sym_tag] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(169),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(115),
    [anon_sym_true] = ACTIONS(117),
    [anon_sym_false] = ACTIONS(117),
    [anon_sym_LBRACE] = ACTIONS(119),
    [anon_sym_object] = ACTIONS(121),
    [anon_sym_fun] = ACTIONS(171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(125),
    [anon_sym_if] = ACTIONS(173),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(175),
    [anon_sym_function] = ACTIONS(177),
    [anon_sym_try] = ACTIONS(179),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1493),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1495),
    [aux_sym_number_token1] = ACTIONS(153),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(155),
    [anon_sym_DQUOTE] = ACTIONS(157),
    [sym_prefix_operator] = ACTIONS(159),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(181),
    [sym__uppercase_identifier] = ACTIONS(165),
    [aux_sym_tag_token1] = ACTIONS(167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(841)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3174),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(841),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(842)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3175),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(3418),
    [sym__tuple_expression] = STATE(2713),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(842),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(1631),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(843)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3114),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(843),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(844)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3111),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(844),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(845)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3112),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(845),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(846)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3113),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(846),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(847)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3177),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(847),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(848)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3115),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(848),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(849)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3116),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(849),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(850)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3117),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(850),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(851)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3118),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(851),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(852)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3119),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(852),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(853)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3120),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(853),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(854)] = {
    [sym_value_definition] = STATE(7424),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3121),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(8314),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(1800),
    [sym_array_get_expression] = STATE(1800),
    [sym_string_get_expression] = STATE(1800),
    [sym_bigarray_get_expression] = STATE(1800),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(854),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(841),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(8324),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(487),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(499),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(503),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(855)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2534),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(855),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(856)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2540),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(2905),
    [sym__tuple_expression] = STATE(2651),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(856),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(1639),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(857)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2541),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(857),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(858)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2542),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(858),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(859)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2543),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(859),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(860)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2634),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(860),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(861)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2535),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(861),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(862)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2536),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(862),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(863)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2537),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(863),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(864)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2544),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(864),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(865)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2545),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(865),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(866)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2546),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(866),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(867)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2547),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(867),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(868)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3771),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(868),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(869)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3773),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(869),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(870)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3774),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(870),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(871)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3760),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(871),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(872)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3796),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(872),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(873)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3777),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(873),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(874)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3778),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(874),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(875)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3779),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(875),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(876)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3780),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(876),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(877)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3781),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(877),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(878)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3782),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(878),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(879)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3783),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(879),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(880)] = {
    [sym_value_definition] = STATE(8216),
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(97),
    [sym__expression] = STATE(3784),
    [sym_typed_expression] = STATE(946),
    [sym_labeled_tuple_element] = STATE(7947),
    [sym__tuple_expression] = STATE(2711),
    [sym_cons_expression] = STATE(2730),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_application_expression] = STATE(2730),
    [sym_prefix_expression] = STATE(946),
    [sym_sign_expression] = STATE(2730),
    [sym_hash_expression] = STATE(946),
    [sym_infix_expression] = STATE(2730),
    [sym_field_get_expression] = STATE(2102),
    [sym_array_get_expression] = STATE(2102),
    [sym_string_get_expression] = STATE(2102),
    [sym_bigarray_get_expression] = STATE(2102),
    [sym_set_expression] = STATE(2730),
    [sym_if_expression] = STATE(2730),
    [sym_while_expression] = STATE(2730),
    [sym_for_expression] = STATE(2730),
    [sym_match_expression] = STATE(2730),
    [sym_function_expression] = STATE(2730),
    [sym_fun_expression] = STATE(2730),
    [sym_try_expression] = STATE(2730),
    [sym_let_expression] = STATE(2730),
    [sym_assert_expression] = STATE(2730),
    [sym_lazy_expression] = STATE(2730),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(880),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym_sign_operator] = STATE(868),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__instance_variable_name] = STATE(7948),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(485),
    [anon_sym_let] = ACTIONS(903),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_fun] = ACTIONS(905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_if] = ACTIONS(907),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(507),
    [anon_sym_match] = ACTIONS(509),
    [anon_sym_function] = ACTIONS(511),
    [anon_sym_try] = ACTIONS(513),
    [anon_sym_assert] = ACTIONS(515),
    [anon_sym_lazy] = ACTIONS(517),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(557),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(559),
    [aux_sym_number_token1] = ACTIONS(531),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(539),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(881)] = {
    [sym_value_definition] = STATE(8139),
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(107),
    [sym__expression] = STATE(2549),
    [sym_typed_expression] = STATE(1364),
    [sym_labeled_tuple_element] = STATE(8107),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1364),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1364),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1386),
    [sym_array_get_expression] = STATE(1386),
    [sym_string_get_expression] = STATE(1386),
    [sym_bigarray_get_expression] = STATE(1386),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(881),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym_sign_operator] = STATE(855),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__instance_variable_name] = STATE(8108),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(249),
    [anon_sym_let] = ACTIONS(251),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_fun] = ACTIONS(263),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_if] = ACTIONS(267),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(269),
    [anon_sym_function] = ACTIONS(271),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1497),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1499),
    [aux_sym_number_token1] = ACTIONS(287),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(295),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(882)] = {
    [sym_value_definition] = STATE(7529),
    [sym__delimited_expression] = STATE(1295),
    [sym__simple_expression] = STATE(104),
    [sym__expression] = STATE(2230),
    [sym_typed_expression] = STATE(1295),
    [sym_labeled_tuple_element] = STATE(7656),
    [sym__tuple_expression] = STATE(2642),
    [sym_cons_expression] = STATE(2652),
    [sym_list_expression] = STATE(1298),
    [sym_array_expression] = STATE(1298),
    [sym_record_expression] = STATE(1298),
    [sym_application_expression] = STATE(2652),
    [sym_prefix_expression] = STATE(1295),
    [sym_sign_expression] = STATE(2652),
    [sym_hash_expression] = STATE(1295),
    [sym_infix_expression] = STATE(2652),
    [sym_field_get_expression] = STATE(1320),
    [sym_array_get_expression] = STATE(1320),
    [sym_string_get_expression] = STATE(1320),
    [sym_bigarray_get_expression] = STATE(1320),
    [sym_set_expression] = STATE(2652),
    [sym_if_expression] = STATE(2652),
    [sym_while_expression] = STATE(2652),
    [sym_for_expression] = STATE(2652),
    [sym_match_expression] = STATE(2652),
    [sym_function_expression] = STATE(2652),
    [sym_fun_expression] = STATE(2652),
    [sym_try_expression] = STATE(2652),
    [sym_let_expression] = STATE(2652),
    [sym_assert_expression] = STATE(2652),
    [sym_lazy_expression] = STATE(2652),
    [sym_local_open_expression] = STATE(1295),
    [sym_package_expression] = STATE(1298),
    [sym_new_expression] = STATE(1295),
    [sym_object_copy_expression] = STATE(1298),
    [sym_method_invocation] = STATE(1295),
    [sym_object_expression] = STATE(1295),
    [sym_parenthesized_expression] = STATE(1298),
    [sym_attribute] = STATE(882),
    [sym_extension] = STATE(1295),
    [sym_quoted_extension] = STATE(1295),
    [sym__constant] = STATE(1295),
    [sym_number] = STATE(1267),
    [sym_character] = STATE(1267),
    [sym_string] = STATE(1267),
    [sym_quoted_string] = STATE(1267),
    [sym_boolean] = STATE(1267),
    [sym_unit] = STATE(1298),
    [sym_sign_operator] = STATE(786),
    [sym__value_name] = STATE(1322),
    [sym_parenthesized_operator] = STATE(1283),
    [sym_value_path] = STATE(1295),
    [sym_module_path] = STATE(7415),
    [sym_constructor_path] = STATE(1295),
    [sym__instance_variable_name] = STATE(7657),
    [sym__constructor_name] = STATE(1235),
    [sym_tag] = STATE(1295),
    [sym__lowercase_identifier] = ACTIONS(183),
    [anon_sym_let] = ACTIONS(235),
    [anon_sym_TILDE] = ACTIONS(17),
    [anon_sym_LPAREN] = ACTIONS(187),
    [anon_sym_LBRACK] = ACTIONS(189),
    [anon_sym_true] = ACTIONS(191),
    [anon_sym_false] = ACTIONS(191),
    [anon_sym_LBRACE] = ACTIONS(193),
    [anon_sym_object] = ACTIONS(195),
    [anon_sym_fun] = ACTIONS(237),
    [anon_sym_LBRACK_PIPE] = ACTIONS(199),
    [anon_sym_if] = ACTIONS(239),
    [anon_sym_while] = ACTIONS(51),
    [anon_sym_for] = ACTIONS(53),
    [anon_sym_match] = ACTIONS(241),
    [anon_sym_function] = ACTIONS(243),
    [anon_sym_try] = ACTIONS(245),
    [anon_sym_assert] = ACTIONS(61),
    [anon_sym_lazy] = ACTIONS(63),
    [anon_sym_new] = ACTIONS(209),
    [anon_sym_LBRACE_LT] = ACTIONS(211),
    [anon_sym_begin] = ACTIONS(213),
    [sym_ocamlyacc_value] = ACTIONS(215),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [aux_sym_number_token1] = ACTIONS(221),
    [aux_sym_signed_number_token1] = ACTIONS(85),
    [anon_sym_SQUOTE] = ACTIONS(223),
    [anon_sym_DQUOTE] = ACTIONS(225),
    [sym_prefix_operator] = ACTIONS(227),
    [aux_sym_sign_operator_token1] = ACTIONS(93),
    [sym_let_operator] = ACTIONS(161),
    [sym_match_operator] = ACTIONS(247),
    [sym__uppercase_identifier] = ACTIONS(231),
    [aux_sym_tag_token1] = ACTIONS(233),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(883)] = {
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2246),
    [sym__binding_pattern] = STATE(2291),
    [sym_alias_binding_pattern] = STATE(2246),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_anonymous] = STATE(2248),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7706),
    [sym__tuple_binding_pattern] = STATE(2249),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern] = STATE(2246),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2246),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_exception_binding_pattern] = STATE(2246),
    [sym_effect_binding_pattern] = STATE(2246),
    [sym_attribute] = STATE(883),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [ts_builtin_sym_end] = ACTIONS(1641),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_SEMI_SEMI] = ACTIONS(1641),
    [anon_sym_let] = ACTIONS(1645),
    [anon_sym_and] = ACTIONS(1645),
    [anon_sym_EQ] = ACTIONS(1641),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_QMARK] = ACTIONS(1641),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_external] = ACTIONS(1645),
    [anon_sym_type] = ACTIONS(1645),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_PIPE] = ACTIONS(1641),
    [anon_sym_COLON2] = ACTIONS(1645),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_RBRACK] = ACTIONS(1641),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1657),
    [anon_sym_module] = ACTIONS(1645),
    [anon_sym_open] = ACTIONS(1645),
    [anon_sym_include] = ACTIONS(1645),
    [anon_sym_class] = ACTIONS(1645),
    [anon_sym_val] = ACTIONS(1645),
    [anon_sym_end] = ACTIONS(1645),
    [anon_sym_COLON_GT] = ACTIONS(1641),
    [anon_sym_in] = ACTIONS(1645),
    [anon_sym_as] = ACTIONS(1645),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_COLON_COLON] = ACTIONS(1641),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [anon_sym_effect] = ACTIONS(1667),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1645),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1641),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1669),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1671),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1641),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1641),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym_let_operator] = ACTIONS(1641),
    [sym_let_and_operator] = ACTIONS(1641),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_directive_token1] = ACTIONS(1645),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(884)] = {
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(1960),
    [sym_typed_expression] = STATE(1977),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_labeled_argument] = STATE(2081),
    [sym_prefix_expression] = STATE(1977),
    [sym_hash_expression] = STATE(1977),
    [sym_field_get_expression] = STATE(1977),
    [sym_array_get_expression] = STATE(1977),
    [sym_string_get_expression] = STATE(1977),
    [sym_bigarray_get_expression] = STATE(1977),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(884),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [aux_sym_class_application_repeat1] = STATE(886),
    [sym__lowercase_identifier] = ACTIONS(1687),
    [anon_sym_and] = ACTIONS(917),
    [anon_sym_TILDE] = ACTIONS(1689),
    [anon_sym_QMARK] = ACTIONS(1689),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_COMMA] = ACTIONS(915),
    [anon_sym_PIPE] = ACTIONS(917),
    [anon_sym_DOT] = ACTIONS(1691),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_SEMI] = ACTIONS(915),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_in] = ACTIONS(917),
    [anon_sym_AMP] = ACTIONS(917),
    [anon_sym_POUND] = ACTIONS(1693),
    [anon_sym_COLON_COLON] = ACTIONS(915),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_else] = ACTIONS(917),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(915),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(917),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(917),
    [sym_hash_operator] = ACTIONS(1695),
    [anon_sym_lsl] = ACTIONS(917),
    [anon_sym_lsr] = ACTIONS(917),
    [anon_sym_asr] = ACTIONS(917),
    [aux_sym_pow_operator_token1] = ACTIONS(915),
    [anon_sym_mod] = ACTIONS(917),
    [anon_sym_land] = ACTIONS(917),
    [anon_sym_lor] = ACTIONS(917),
    [anon_sym_lxor] = ACTIONS(917),
    [aux_sym_mult_operator_token1] = ACTIONS(917),
    [aux_sym_add_operator_token1] = ACTIONS(917),
    [sym_concat_operator] = ACTIONS(915),
    [sym_rel_operator] = ACTIONS(917),
    [anon_sym_AMP_AMP] = ACTIONS(917),
    [anon_sym_or] = ACTIONS(917),
    [anon_sym_PIPE_PIPE] = ACTIONS(917),
    [sym_assign_operator] = ACTIONS(915),
    [sym_let_and_operator] = ACTIONS(915),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(885)] = {
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(1960),
    [sym_typed_expression] = STATE(1977),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_labeled_argument] = STATE(2081),
    [sym_prefix_expression] = STATE(1977),
    [sym_hash_expression] = STATE(1977),
    [sym_field_get_expression] = STATE(1977),
    [sym_array_get_expression] = STATE(1977),
    [sym_string_get_expression] = STATE(1977),
    [sym_bigarray_get_expression] = STATE(1977),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(885),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [aux_sym_class_application_repeat1] = STATE(885),
    [sym__lowercase_identifier] = ACTIONS(1697),
    [anon_sym_and] = ACTIONS(928),
    [anon_sym_TILDE] = ACTIONS(1700),
    [anon_sym_QMARK] = ACTIONS(1700),
    [anon_sym_LPAREN] = ACTIONS(1703),
    [anon_sym_COMMA] = ACTIONS(936),
    [anon_sym_PIPE] = ACTIONS(928),
    [anon_sym_LBRACK] = ACTIONS(1706),
    [anon_sym_true] = ACTIONS(1709),
    [anon_sym_false] = ACTIONS(1709),
    [anon_sym_LBRACE] = ACTIONS(1712),
    [anon_sym_SEMI] = ACTIONS(936),
    [anon_sym_object] = ACTIONS(1715),
    [anon_sym_in] = ACTIONS(928),
    [anon_sym_AMP] = ACTIONS(928),
    [anon_sym_COLON_COLON] = ACTIONS(936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1718),
    [anon_sym_else] = ACTIONS(928),
    [anon_sym_new] = ACTIONS(1721),
    [anon_sym_LBRACE_LT] = ACTIONS(1724),
    [anon_sym_begin] = ACTIONS(1727),
    [sym_ocamlyacc_value] = ACTIONS(1730),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(936),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1733),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1736),
    [aux_sym_number_token1] = ACTIONS(1739),
    [aux_sym_signed_number_token1] = ACTIONS(928),
    [anon_sym_SQUOTE] = ACTIONS(1742),
    [anon_sym_DQUOTE] = ACTIONS(1745),
    [sym_prefix_operator] = ACTIONS(1748),
    [aux_sym_sign_operator_token1] = ACTIONS(928),
    [anon_sym_lsl] = ACTIONS(928),
    [anon_sym_lsr] = ACTIONS(928),
    [anon_sym_asr] = ACTIONS(928),
    [aux_sym_pow_operator_token1] = ACTIONS(936),
    [anon_sym_mod] = ACTIONS(928),
    [anon_sym_land] = ACTIONS(928),
    [anon_sym_lor] = ACTIONS(928),
    [anon_sym_lxor] = ACTIONS(928),
    [aux_sym_mult_operator_token1] = ACTIONS(928),
    [aux_sym_add_operator_token1] = ACTIONS(928),
    [sym_concat_operator] = ACTIONS(936),
    [sym_rel_operator] = ACTIONS(928),
    [anon_sym_AMP_AMP] = ACTIONS(928),
    [anon_sym_or] = ACTIONS(928),
    [anon_sym_PIPE_PIPE] = ACTIONS(928),
    [sym_assign_operator] = ACTIONS(936),
    [sym_let_and_operator] = ACTIONS(936),
    [sym__uppercase_identifier] = ACTIONS(1751),
    [aux_sym_tag_token1] = ACTIONS(1754),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(886)] = {
    [sym__delimited_expression] = STATE(1977),
    [sym__simple_expression] = STATE(1960),
    [sym_typed_expression] = STATE(1977),
    [sym_list_expression] = STATE(2002),
    [sym_array_expression] = STATE(2002),
    [sym_record_expression] = STATE(2002),
    [sym_labeled_argument] = STATE(2081),
    [sym_prefix_expression] = STATE(1977),
    [sym_hash_expression] = STATE(1977),
    [sym_field_get_expression] = STATE(1977),
    [sym_array_get_expression] = STATE(1977),
    [sym_string_get_expression] = STATE(1977),
    [sym_bigarray_get_expression] = STATE(1977),
    [sym_local_open_expression] = STATE(1977),
    [sym_package_expression] = STATE(2002),
    [sym_new_expression] = STATE(1977),
    [sym_object_copy_expression] = STATE(2002),
    [sym_method_invocation] = STATE(1977),
    [sym_object_expression] = STATE(1977),
    [sym_parenthesized_expression] = STATE(2002),
    [sym_attribute] = STATE(886),
    [sym_extension] = STATE(1977),
    [sym_quoted_extension] = STATE(1977),
    [sym__constant] = STATE(1977),
    [sym_number] = STATE(1968),
    [sym_character] = STATE(1968),
    [sym_string] = STATE(1968),
    [sym_quoted_string] = STATE(1968),
    [sym_boolean] = STATE(1968),
    [sym_unit] = STATE(2002),
    [sym__value_name] = STATE(2018),
    [sym_parenthesized_operator] = STATE(2019),
    [sym_value_path] = STATE(1977),
    [sym_module_path] = STATE(7454),
    [sym_constructor_path] = STATE(1977),
    [sym__constructor_name] = STATE(1969),
    [sym_tag] = STATE(1977),
    [aux_sym_class_application_repeat1] = STATE(885),
    [sym__lowercase_identifier] = ACTIONS(1687),
    [anon_sym_and] = ACTIONS(1001),
    [anon_sym_TILDE] = ACTIONS(1689),
    [anon_sym_QMARK] = ACTIONS(1689),
    [anon_sym_LPAREN] = ACTIONS(807),
    [anon_sym_COMMA] = ACTIONS(999),
    [anon_sym_PIPE] = ACTIONS(1001),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_true] = ACTIONS(811),
    [anon_sym_false] = ACTIONS(811),
    [anon_sym_LBRACE] = ACTIONS(813),
    [anon_sym_SEMI] = ACTIONS(999),
    [anon_sym_object] = ACTIONS(815),
    [anon_sym_in] = ACTIONS(1001),
    [anon_sym_AMP] = ACTIONS(1001),
    [anon_sym_COLON_COLON] = ACTIONS(999),
    [anon_sym_LBRACK_PIPE] = ACTIONS(819),
    [anon_sym_else] = ACTIONS(1001),
    [anon_sym_new] = ACTIONS(837),
    [anon_sym_LBRACE_LT] = ACTIONS(839),
    [anon_sym_begin] = ACTIONS(841),
    [sym_ocamlyacc_value] = ACTIONS(843),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(999),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(845),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(847),
    [aux_sym_number_token1] = ACTIONS(849),
    [aux_sym_signed_number_token1] = ACTIONS(1001),
    [anon_sym_SQUOTE] = ACTIONS(851),
    [anon_sym_DQUOTE] = ACTIONS(853),
    [sym_prefix_operator] = ACTIONS(855),
    [aux_sym_sign_operator_token1] = ACTIONS(1001),
    [anon_sym_lsl] = ACTIONS(1001),
    [anon_sym_lsr] = ACTIONS(1001),
    [anon_sym_asr] = ACTIONS(1001),
    [aux_sym_pow_operator_token1] = ACTIONS(999),
    [anon_sym_mod] = ACTIONS(1001),
    [anon_sym_land] = ACTIONS(1001),
    [anon_sym_lor] = ACTIONS(1001),
    [anon_sym_lxor] = ACTIONS(1001),
    [aux_sym_mult_operator_token1] = ACTIONS(1001),
    [aux_sym_add_operator_token1] = ACTIONS(1001),
    [sym_concat_operator] = ACTIONS(999),
    [sym_rel_operator] = ACTIONS(1001),
    [anon_sym_AMP_AMP] = ACTIONS(1001),
    [anon_sym_or] = ACTIONS(1001),
    [anon_sym_PIPE_PIPE] = ACTIONS(1001),
    [sym_assign_operator] = ACTIONS(999),
    [sym_let_and_operator] = ACTIONS(999),
    [sym__uppercase_identifier] = ACTIONS(859),
    [aux_sym_tag_token1] = ACTIONS(861),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(887)] = {
    [sym__local_structure_item] = STATE(7655),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(887),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(924),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(888)] = {
    [sym__local_structure_item] = STATE(7558),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(888),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(916),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(889)] = {
    [sym__local_structure_item] = STATE(7439),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(889),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(911),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(890)] = {
    [sym__local_structure_item] = STATE(7231),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(890),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(913),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(891)] = {
    [sym__local_structure_item] = STATE(7433),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(891),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(908),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(892)] = {
    [sym__local_structure_item] = STATE(7860),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(892),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(915),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(893)] = {
    [sym__local_structure_item] = STATE(8031),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(893),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(917),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(894)] = {
    [sym__local_structure_item] = STATE(8045),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(894),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(906),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(895)] = {
    [sym__local_structure_item] = STATE(8315),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(895),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(922),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(896)] = {
    [sym__local_structure_item] = STATE(8009),
    [sym_let_binding] = STATE(4500),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2353),
    [sym__binding_pattern_no_exn] = STATE(910),
    [sym_alias_binding_pattern_no_exn] = STATE(2353),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2353),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8145),
    [sym__tuple_binding_pattern_no_exn] = STATE(2353),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern_no_exn] = STATE(2353),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2353),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_attribute] = STATE(896),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(909),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_rec] = ACTIONS(1813),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1669),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1671),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(897)] = {
    [sym__local_structure_item] = STATE(7358),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(897),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(907),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(898)] = {
    [sym__local_structure_item] = STATE(7949),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(898),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(925),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(899)] = {
    [sym__local_structure_item] = STATE(8009),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(899),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(914),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(900)] = {
    [sym__local_structure_item] = STATE(8130),
    [sym_let_binding] = STATE(4460),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2353),
    [sym__binding_pattern_no_exn] = STATE(905),
    [sym_alias_binding_pattern_no_exn] = STATE(2353),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2353),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8145),
    [sym__tuple_binding_pattern_no_exn] = STATE(2353),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern_no_exn] = STATE(2353),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2353),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_attribute] = STATE(900),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(920),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_rec] = ACTIONS(1815),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1669),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1671),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(901)] = {
    [sym__local_structure_item] = STATE(8130),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(901),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(923),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(902)] = {
    [sym__local_structure_item] = STATE(7638),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(902),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(919),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(903)] = {
    [sym__local_structure_item] = STATE(7210),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(903),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(921),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(904)] = {
    [sym__local_structure_item] = STATE(7504),
    [sym_let_binding] = STATE(6036),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(904),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__attribute] = STATE(912),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(905)] = {
    [sym_parameter] = STATE(2862),
    [sym__coerced] = STATE(7209),
    [sym__maybe_polymorphic_typed] = STATE(6500),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(905),
    [sym_item_attribute] = STATE(4337),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(8318),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(4169),
    [aux_sym_let_binding_repeat1] = STATE(1934),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(1819),
    [anon_sym_let] = ACTIONS(1821),
    [anon_sym_and] = ACTIONS(1821),
    [anon_sym_EQ] = ACTIONS(1823),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_external] = ACTIONS(1821),
    [anon_sym_type] = ACTIONS(1821),
    [anon_sym_COMMA] = ACTIONS(1829),
    [anon_sym_PIPE] = ACTIONS(1831),
    [anon_sym_COLON2] = ACTIONS(1833),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_RBRACK] = ACTIONS(1819),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_exception] = ACTIONS(1821),
    [anon_sym_module] = ACTIONS(1821),
    [anon_sym_open] = ACTIONS(1821),
    [anon_sym_include] = ACTIONS(1821),
    [anon_sym_class] = ACTIONS(1821),
    [anon_sym_val] = ACTIONS(1821),
    [anon_sym_end] = ACTIONS(1821),
    [anon_sym_COLON_GT] = ACTIONS(1841),
    [anon_sym_in] = ACTIONS(1821),
    [anon_sym_as] = ACTIONS(1843),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_COLON_COLON] = ACTIONS(1847),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1853),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1819),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1819),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1819),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym_let_operator] = ACTIONS(1819),
    [sym_let_and_operator] = ACTIONS(1819),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_directive_token1] = ACTIONS(1821),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(906)] = {
    [sym__local_structure_item] = STATE(8166),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(906),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(907)] = {
    [sym__local_structure_item] = STATE(7755),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(907),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(908)] = {
    [sym__local_structure_item] = STATE(7738),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(908),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(909)] = {
    [sym__local_structure_item] = STATE(7585),
    [sym_let_binding] = STATE(4528),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2353),
    [sym__binding_pattern_no_exn] = STATE(910),
    [sym_alias_binding_pattern_no_exn] = STATE(2353),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2353),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8145),
    [sym__tuple_binding_pattern_no_exn] = STATE(2353),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern_no_exn] = STATE(2353),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2353),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_attribute] = STATE(909),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_rec] = ACTIONS(1869),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1669),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1671),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(910)] = {
    [sym_parameter] = STATE(2862),
    [sym__coerced] = STATE(7271),
    [sym__maybe_polymorphic_typed] = STATE(6436),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(910),
    [sym_item_attribute] = STATE(4337),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(8318),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(4169),
    [aux_sym_let_binding_repeat1] = STATE(1935),
    [ts_builtin_sym_end] = ACTIONS(1819),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(1819),
    [anon_sym_let] = ACTIONS(1821),
    [anon_sym_and] = ACTIONS(1821),
    [anon_sym_EQ] = ACTIONS(1871),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_external] = ACTIONS(1821),
    [anon_sym_type] = ACTIONS(1821),
    [anon_sym_COMMA] = ACTIONS(1829),
    [anon_sym_PIPE] = ACTIONS(1831),
    [anon_sym_COLON2] = ACTIONS(1833),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_exception] = ACTIONS(1821),
    [anon_sym_module] = ACTIONS(1821),
    [anon_sym_open] = ACTIONS(1821),
    [anon_sym_include] = ACTIONS(1821),
    [anon_sym_class] = ACTIONS(1821),
    [anon_sym_val] = ACTIONS(1821),
    [anon_sym_COLON_GT] = ACTIONS(1841),
    [anon_sym_in] = ACTIONS(1821),
    [anon_sym_as] = ACTIONS(1843),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_COLON_COLON] = ACTIONS(1847),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1853),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1819),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1819),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1819),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym_let_operator] = ACTIONS(1819),
    [sym_let_and_operator] = ACTIONS(1819),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_directive_token1] = ACTIONS(1821),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(911)] = {
    [sym__local_structure_item] = STATE(8095),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(911),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(912)] = {
    [sym__local_structure_item] = STATE(7524),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(912),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(913)] = {
    [sym__local_structure_item] = STATE(7345),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(913),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(914)] = {
    [sym__local_structure_item] = STATE(7585),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(914),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(915)] = {
    [sym__local_structure_item] = STATE(7908),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(915),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(916)] = {
    [sym__local_structure_item] = STATE(7653),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(916),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(917)] = {
    [sym__local_structure_item] = STATE(8097),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(917),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(918)] = {
    [sym_parameter] = STATE(2862),
    [sym__coerced] = STATE(7283),
    [sym__maybe_polymorphic_typed] = STATE(6510),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(918),
    [sym_item_attribute] = STATE(4337),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(8318),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(4169),
    [aux_sym_let_binding_repeat1] = STATE(1936),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(1819),
    [anon_sym_let] = ACTIONS(1821),
    [anon_sym_and] = ACTIONS(1821),
    [anon_sym_EQ] = ACTIONS(1873),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_external] = ACTIONS(1821),
    [anon_sym_type] = ACTIONS(1821),
    [anon_sym_COMMA] = ACTIONS(1829),
    [anon_sym_PIPE] = ACTIONS(1831),
    [anon_sym_COLON2] = ACTIONS(1833),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_RBRACK] = ACTIONS(1819),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_exception] = ACTIONS(1821),
    [anon_sym_module] = ACTIONS(1821),
    [anon_sym_open] = ACTIONS(1821),
    [anon_sym_include] = ACTIONS(1821),
    [anon_sym_class] = ACTIONS(1821),
    [anon_sym_val] = ACTIONS(1821),
    [anon_sym_end] = ACTIONS(1821),
    [anon_sym_COLON_GT] = ACTIONS(1841),
    [anon_sym_as] = ACTIONS(1843),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_COLON_COLON] = ACTIONS(1847),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1853),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1819),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1819),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1819),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym_let_operator] = ACTIONS(1819),
    [sym_let_and_operator] = ACTIONS(1819),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_directive_token1] = ACTIONS(1821),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(919)] = {
    [sym__local_structure_item] = STATE(7945),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(919),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(920)] = {
    [sym__local_structure_item] = STATE(8273),
    [sym_let_binding] = STATE(4461),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2353),
    [sym__binding_pattern_no_exn] = STATE(905),
    [sym_alias_binding_pattern_no_exn] = STATE(2353),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2353),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8145),
    [sym__tuple_binding_pattern_no_exn] = STATE(2353),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern_no_exn] = STATE(2353),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2353),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_attribute] = STATE(920),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_rec] = ACTIONS(1875),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1669),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1671),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(921)] = {
    [sym__local_structure_item] = STATE(7717),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(921),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(922)] = {
    [sym__local_structure_item] = STATE(7675),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(922),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(923)] = {
    [sym__local_structure_item] = STATE(8273),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(923),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(924)] = {
    [sym__local_structure_item] = STATE(7546),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(924),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(925)] = {
    [sym__local_structure_item] = STATE(7953),
    [sym_let_binding] = STATE(6219),
    [sym_external] = STATE(7461),
    [sym_type_definition] = STATE(7461),
    [sym_exception_definition] = STATE(7461),
    [sym_module_definition] = STATE(7461),
    [sym_module_type_definition] = STATE(7461),
    [sym_open_module] = STATE(7461),
    [sym_class_definition] = STATE(7461),
    [sym_class_type_definition] = STATE(7461),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(925),
    [sym_floating_attribute] = STATE(7461),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym_item_extension] = STATE(7461),
    [sym_quoted_item_extension] = STATE(7461),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1867),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_external] = ACTIONS(1763),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1775),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1779),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1797),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(926)] = {
    [sym_parameter] = STATE(2862),
    [sym__coerced] = STATE(7798),
    [sym__maybe_polymorphic_typed] = STATE(6462),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(926),
    [sym_item_attribute] = STATE(4337),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(8318),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(4169),
    [aux_sym_let_binding_repeat1] = STATE(1933),
    [ts_builtin_sym_end] = ACTIONS(1819),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(1819),
    [anon_sym_let] = ACTIONS(1821),
    [anon_sym_and] = ACTIONS(1821),
    [anon_sym_EQ] = ACTIONS(1877),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_external] = ACTIONS(1821),
    [anon_sym_type] = ACTIONS(1821),
    [anon_sym_COMMA] = ACTIONS(1829),
    [anon_sym_PIPE] = ACTIONS(1831),
    [anon_sym_COLON2] = ACTIONS(1833),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_exception] = ACTIONS(1821),
    [anon_sym_module] = ACTIONS(1821),
    [anon_sym_open] = ACTIONS(1821),
    [anon_sym_include] = ACTIONS(1821),
    [anon_sym_class] = ACTIONS(1821),
    [anon_sym_val] = ACTIONS(1821),
    [anon_sym_COLON_GT] = ACTIONS(1841),
    [anon_sym_as] = ACTIONS(1843),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_COLON_COLON] = ACTIONS(1847),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1853),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1819),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1819),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1819),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym_let_operator] = ACTIONS(1819),
    [sym_let_and_operator] = ACTIONS(1819),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_directive_token1] = ACTIONS(1821),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(927)] = {
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(1393),
    [sym_typed_expression] = STATE(1364),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_labeled_argument] = STATE(1490),
    [sym_prefix_expression] = STATE(1364),
    [sym_hash_expression] = STATE(1364),
    [sym_field_get_expression] = STATE(1364),
    [sym_array_get_expression] = STATE(1364),
    [sym_string_get_expression] = STATE(1364),
    [sym_bigarray_get_expression] = STATE(1364),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(927),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [aux_sym_class_application_repeat1] = STATE(928),
    [sym__lowercase_identifier] = ACTIONS(1083),
    [anon_sym_SEMI_SEMI] = ACTIONS(1879),
    [anon_sym_let] = ACTIONS(1881),
    [anon_sym_and] = ACTIONS(1881),
    [anon_sym_TILDE] = ACTIONS(1085),
    [anon_sym_QMARK] = ACTIONS(1085),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(1881),
    [anon_sym_type] = ACTIONS(1881),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(1879),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(1881),
    [anon_sym_module] = ACTIONS(1881),
    [anon_sym_open] = ACTIONS(1881),
    [anon_sym_include] = ACTIONS(1881),
    [anon_sym_class] = ACTIONS(1881),
    [anon_sym_val] = ACTIONS(1881),
    [anon_sym_end] = ACTIONS(1881),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1881),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1879),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1879),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1879),
    [aux_sym_number_token1] = ACTIONS(287),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [sym_let_operator] = ACTIONS(1879),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(1879),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(928)] = {
    [sym__delimited_expression] = STATE(1364),
    [sym__simple_expression] = STATE(1393),
    [sym_typed_expression] = STATE(1364),
    [sym_list_expression] = STATE(1377),
    [sym_array_expression] = STATE(1377),
    [sym_record_expression] = STATE(1377),
    [sym_labeled_argument] = STATE(1490),
    [sym_prefix_expression] = STATE(1364),
    [sym_hash_expression] = STATE(1364),
    [sym_field_get_expression] = STATE(1364),
    [sym_array_get_expression] = STATE(1364),
    [sym_string_get_expression] = STATE(1364),
    [sym_bigarray_get_expression] = STATE(1364),
    [sym_local_open_expression] = STATE(1364),
    [sym_package_expression] = STATE(1377),
    [sym_new_expression] = STATE(1364),
    [sym_object_copy_expression] = STATE(1377),
    [sym_method_invocation] = STATE(1364),
    [sym_object_expression] = STATE(1364),
    [sym_parenthesized_expression] = STATE(1377),
    [sym_attribute] = STATE(928),
    [sym_extension] = STATE(1364),
    [sym_quoted_extension] = STATE(1364),
    [sym__constant] = STATE(1364),
    [sym_number] = STATE(1336),
    [sym_character] = STATE(1336),
    [sym_string] = STATE(1336),
    [sym_quoted_string] = STATE(1336),
    [sym_boolean] = STATE(1336),
    [sym_unit] = STATE(1377),
    [sym__value_name] = STATE(1366),
    [sym_parenthesized_operator] = STATE(1378),
    [sym_value_path] = STATE(1364),
    [sym_module_path] = STATE(7186),
    [sym_constructor_path] = STATE(1364),
    [sym__constructor_name] = STATE(1375),
    [sym_tag] = STATE(1364),
    [aux_sym_class_application_repeat1] = STATE(115),
    [sym__lowercase_identifier] = ACTIONS(1083),
    [anon_sym_SEMI_SEMI] = ACTIONS(1883),
    [anon_sym_let] = ACTIONS(1885),
    [anon_sym_and] = ACTIONS(1885),
    [anon_sym_TILDE] = ACTIONS(1085),
    [anon_sym_QMARK] = ACTIONS(1085),
    [anon_sym_LPAREN] = ACTIONS(253),
    [anon_sym_external] = ACTIONS(1885),
    [anon_sym_type] = ACTIONS(1885),
    [anon_sym_LBRACK] = ACTIONS(255),
    [anon_sym_RBRACK] = ACTIONS(1883),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_LBRACE] = ACTIONS(259),
    [anon_sym_exception] = ACTIONS(1885),
    [anon_sym_module] = ACTIONS(1885),
    [anon_sym_open] = ACTIONS(1885),
    [anon_sym_include] = ACTIONS(1885),
    [anon_sym_class] = ACTIONS(1885),
    [anon_sym_val] = ACTIONS(1885),
    [anon_sym_end] = ACTIONS(1885),
    [anon_sym_object] = ACTIONS(261),
    [anon_sym_LBRACK_PIPE] = ACTIONS(265),
    [anon_sym_new] = ACTIONS(275),
    [anon_sym_LBRACE_LT] = ACTIONS(277),
    [anon_sym_begin] = ACTIONS(279),
    [sym_ocamlyacc_value] = ACTIONS(281),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1885),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1883),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(283),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(285),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1883),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1883),
    [aux_sym_number_token1] = ACTIONS(287),
    [anon_sym_SQUOTE] = ACTIONS(289),
    [anon_sym_DQUOTE] = ACTIONS(291),
    [sym_prefix_operator] = ACTIONS(293),
    [sym_let_operator] = ACTIONS(1883),
    [sym__uppercase_identifier] = ACTIONS(297),
    [aux_sym_directive_token1] = ACTIONS(1883),
    [aux_sym_tag_token1] = ACTIONS(299),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(929)] = {
    [sym_attribute] = STATE(929),
    [sym__lowercase_identifier] = ACTIONS(1887),
    [anon_sym_and] = ACTIONS(1887),
    [anon_sym_TILDE] = ACTIONS(1887),
    [anon_sym_QMARK] = ACTIONS(1887),
    [anon_sym_LPAREN] = ACTIONS(1889),
    [anon_sym_RPAREN] = ACTIONS(1889),
    [anon_sym_COMMA] = ACTIONS(1889),
    [anon_sym_PIPE] = ACTIONS(1887),
    [anon_sym_COLON2] = ACTIONS(1887),
    [anon_sym_DOT] = ACTIONS(1889),
    [anon_sym_DASH_GT] = ACTIONS(1887),
    [anon_sym_LBRACK] = ACTIONS(1887),
    [anon_sym_RBRACK] = ACTIONS(1889),
    [anon_sym_true] = ACTIONS(1887),
    [anon_sym_false] = ACTIONS(1887),
    [anon_sym_LBRACE] = ACTIONS(1887),
    [anon_sym_SEMI] = ACTIONS(1889),
    [anon_sym_RBRACE] = ACTIONS(1889),
    [anon_sym_constraint] = ACTIONS(1887),
    [anon_sym_val] = ACTIONS(1887),
    [anon_sym_end] = ACTIONS(1887),
    [anon_sym_with] = ACTIONS(1887),
    [anon_sym_COLON_GT] = ACTIONS(1889),
    [anon_sym_object] = ACTIONS(1887),
    [anon_sym_inherit] = ACTIONS(1887),
    [anon_sym_method] = ACTIONS(1887),
    [anon_sym_in] = ACTIONS(1887),
    [anon_sym_as] = ACTIONS(1887),
    [anon_sym_initializer] = ACTIONS(1887),
    [anon_sym_AMP] = ACTIONS(1887),
    [anon_sym_POUND] = ACTIONS(1887),
    [anon_sym_COLON_COLON] = ACTIONS(1889),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1889),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1889),
    [anon_sym_LT_DASH] = ACTIONS(1887),
    [anon_sym_then] = ACTIONS(1887),
    [anon_sym_else] = ACTIONS(1887),
    [anon_sym_do] = ACTIONS(1887),
    [anon_sym_done] = ACTIONS(1887),
    [anon_sym_to] = ACTIONS(1887),
    [anon_sym_downto] = ACTIONS(1887),
    [anon_sym_new] = ACTIONS(1887),
    [anon_sym_LBRACE_LT] = ACTIONS(1889),
    [anon_sym_GT_RBRACE] = ACTIONS(1889),
    [anon_sym_begin] = ACTIONS(1887),
    [sym_ocamlyacc_value] = ACTIONS(1889),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1887),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1889),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1887),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1887),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1889),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1889),
    [aux_sym_number_token1] = ACTIONS(1889),
    [aux_sym_signed_number_token1] = ACTIONS(1887),
    [anon_sym_SQUOTE] = ACTIONS(1889),
    [anon_sym_DQUOTE] = ACTIONS(1889),
    [sym_prefix_operator] = ACTIONS(1889),
    [aux_sym_sign_operator_token1] = ACTIONS(1887),
    [sym_hash_operator] = ACTIONS(1889),
    [anon_sym_lsl] = ACTIONS(1887),
    [anon_sym_lsr] = ACTIONS(1887),
    [anon_sym_asr] = ACTIONS(1887),
    [aux_sym_pow_operator_token1] = ACTIONS(1889),
    [anon_sym_mod] = ACTIONS(1887),
    [anon_sym_land] = ACTIONS(1887),
    [anon_sym_lor] = ACTIONS(1887),
    [anon_sym_lxor] = ACTIONS(1887),
    [aux_sym_mult_operator_token1] = ACTIONS(1887),
    [aux_sym_add_operator_token1] = ACTIONS(1887),
    [sym_concat_operator] = ACTIONS(1889),
    [sym_rel_operator] = ACTIONS(1887),
    [anon_sym_AMP_AMP] = ACTIONS(1887),
    [anon_sym_or] = ACTIONS(1887),
    [anon_sym_PIPE_PIPE] = ACTIONS(1887),
    [sym_assign_operator] = ACTIONS(1889),
    [sym__uppercase_identifier] = ACTIONS(1889),
    [aux_sym_tag_token1] = ACTIONS(1889),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(930)] = {
    [sym_attribute] = STATE(930),
    [sym__lowercase_identifier] = ACTIONS(1891),
    [anon_sym_and] = ACTIONS(1891),
    [anon_sym_TILDE] = ACTIONS(1891),
    [anon_sym_QMARK] = ACTIONS(1891),
    [anon_sym_LPAREN] = ACTIONS(1893),
    [anon_sym_RPAREN] = ACTIONS(1893),
    [anon_sym_COMMA] = ACTIONS(1893),
    [anon_sym_PIPE] = ACTIONS(1891),
    [anon_sym_COLON2] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_DASH_GT] = ACTIONS(1891),
    [anon_sym_LBRACK] = ACTIONS(1891),
    [anon_sym_RBRACK] = ACTIONS(1893),
    [anon_sym_true] = ACTIONS(1891),
    [anon_sym_false] = ACTIONS(1891),
    [anon_sym_LBRACE] = ACTIONS(1891),
    [anon_sym_SEMI] = ACTIONS(1893),
    [anon_sym_RBRACE] = ACTIONS(1893),
    [anon_sym_constraint] = ACTIONS(1891),
    [anon_sym_val] = ACTIONS(1891),
    [anon_sym_end] = ACTIONS(1891),
    [anon_sym_with] = ACTIONS(1891),
    [anon_sym_COLON_GT] = ACTIONS(1893),
    [anon_sym_object] = ACTIONS(1891),
    [anon_sym_inherit] = ACTIONS(1891),
    [anon_sym_method] = ACTIONS(1891),
    [anon_sym_in] = ACTIONS(1891),
    [anon_sym_as] = ACTIONS(1891),
    [anon_sym_initializer] = ACTIONS(1891),
    [anon_sym_AMP] = ACTIONS(1891),
    [anon_sym_POUND] = ACTIONS(1891),
    [anon_sym_COLON_COLON] = ACTIONS(1893),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1893),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1893),
    [anon_sym_LT_DASH] = ACTIONS(1891),
    [anon_sym_then] = ACTIONS(1891),
    [anon_sym_else] = ACTIONS(1891),
    [anon_sym_do] = ACTIONS(1891),
    [anon_sym_done] = ACTIONS(1891),
    [anon_sym_to] = ACTIONS(1891),
    [anon_sym_downto] = ACTIONS(1891),
    [anon_sym_new] = ACTIONS(1891),
    [anon_sym_LBRACE_LT] = ACTIONS(1893),
    [anon_sym_GT_RBRACE] = ACTIONS(1893),
    [anon_sym_begin] = ACTIONS(1891),
    [sym_ocamlyacc_value] = ACTIONS(1893),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1891),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1893),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1891),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1891),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1893),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1893),
    [aux_sym_number_token1] = ACTIONS(1893),
    [aux_sym_signed_number_token1] = ACTIONS(1891),
    [anon_sym_SQUOTE] = ACTIONS(1893),
    [anon_sym_DQUOTE] = ACTIONS(1893),
    [sym_prefix_operator] = ACTIONS(1893),
    [aux_sym_sign_operator_token1] = ACTIONS(1891),
    [sym_hash_operator] = ACTIONS(1893),
    [anon_sym_lsl] = ACTIONS(1891),
    [anon_sym_lsr] = ACTIONS(1891),
    [anon_sym_asr] = ACTIONS(1891),
    [aux_sym_pow_operator_token1] = ACTIONS(1893),
    [anon_sym_mod] = ACTIONS(1891),
    [anon_sym_land] = ACTIONS(1891),
    [anon_sym_lor] = ACTIONS(1891),
    [anon_sym_lxor] = ACTIONS(1891),
    [aux_sym_mult_operator_token1] = ACTIONS(1891),
    [aux_sym_add_operator_token1] = ACTIONS(1891),
    [sym_concat_operator] = ACTIONS(1893),
    [sym_rel_operator] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1891),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [sym_assign_operator] = ACTIONS(1893),
    [sym__uppercase_identifier] = ACTIONS(1893),
    [aux_sym_tag_token1] = ACTIONS(1893),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(931)] = {
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(1486),
    [sym_typed_expression] = STATE(1471),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_labeled_argument] = STATE(1490),
    [sym_prefix_expression] = STATE(1471),
    [sym_hash_expression] = STATE(1471),
    [sym_field_get_expression] = STATE(1471),
    [sym_array_get_expression] = STATE(1471),
    [sym_string_get_expression] = STATE(1471),
    [sym_bigarray_get_expression] = STATE(1471),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(931),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [aux_sym_class_application_repeat1] = STATE(933),
    [ts_builtin_sym_end] = ACTIONS(1879),
    [sym__lowercase_identifier] = ACTIONS(1153),
    [anon_sym_SEMI_SEMI] = ACTIONS(1879),
    [anon_sym_let] = ACTIONS(1881),
    [anon_sym_and] = ACTIONS(1881),
    [anon_sym_TILDE] = ACTIONS(1155),
    [anon_sym_QMARK] = ACTIONS(1155),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_external] = ACTIONS(1881),
    [anon_sym_type] = ACTIONS(1881),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_exception] = ACTIONS(1881),
    [anon_sym_module] = ACTIONS(1881),
    [anon_sym_open] = ACTIONS(1881),
    [anon_sym_include] = ACTIONS(1881),
    [anon_sym_class] = ACTIONS(1881),
    [anon_sym_val] = ACTIONS(1881),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1881),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1879),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(75),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(77),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1879),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1879),
    [aux_sym_number_token1] = ACTIONS(83),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [sym_let_operator] = ACTIONS(1879),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_directive_token1] = ACTIONS(1879),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(932)] = {
    [sym_attribute] = STATE(932),
    [sym__lowercase_identifier] = ACTIONS(1895),
    [anon_sym_and] = ACTIONS(1895),
    [anon_sym_TILDE] = ACTIONS(1895),
    [anon_sym_QMARK] = ACTIONS(1895),
    [anon_sym_LPAREN] = ACTIONS(1897),
    [anon_sym_RPAREN] = ACTIONS(1897),
    [anon_sym_COMMA] = ACTIONS(1897),
    [anon_sym_PIPE] = ACTIONS(1895),
    [anon_sym_COLON2] = ACTIONS(1895),
    [anon_sym_DOT] = ACTIONS(1897),
    [anon_sym_DASH_GT] = ACTIONS(1895),
    [anon_sym_LBRACK] = ACTIONS(1895),
    [anon_sym_RBRACK] = ACTIONS(1897),
    [anon_sym_true] = ACTIONS(1895),
    [anon_sym_false] = ACTIONS(1895),
    [anon_sym_LBRACE] = ACTIONS(1895),
    [anon_sym_SEMI] = ACTIONS(1897),
    [anon_sym_RBRACE] = ACTIONS(1897),
    [anon_sym_constraint] = ACTIONS(1895),
    [anon_sym_val] = ACTIONS(1895),
    [anon_sym_end] = ACTIONS(1895),
    [anon_sym_with] = ACTIONS(1895),
    [anon_sym_COLON_GT] = ACTIONS(1897),
    [anon_sym_object] = ACTIONS(1895),
    [anon_sym_inherit] = ACTIONS(1895),
    [anon_sym_method] = ACTIONS(1895),
    [anon_sym_in] = ACTIONS(1895),
    [anon_sym_as] = ACTIONS(1895),
    [anon_sym_initializer] = ACTIONS(1895),
    [anon_sym_AMP] = ACTIONS(1895),
    [anon_sym_POUND] = ACTIONS(1895),
    [anon_sym_COLON_COLON] = ACTIONS(1897),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1897),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1897),
    [anon_sym_LT_DASH] = ACTIONS(1895),
    [anon_sym_then] = ACTIONS(1895),
    [anon_sym_else] = ACTIONS(1895),
    [anon_sym_do] = ACTIONS(1895),
    [anon_sym_done] = ACTIONS(1895),
    [anon_sym_to] = ACTIONS(1895),
    [anon_sym_downto] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1895),
    [anon_sym_LBRACE_LT] = ACTIONS(1897),
    [anon_sym_GT_RBRACE] = ACTIONS(1897),
    [anon_sym_begin] = ACTIONS(1895),
    [sym_ocamlyacc_value] = ACTIONS(1897),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1895),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1897),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1895),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1895),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1897),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1897),
    [aux_sym_number_token1] = ACTIONS(1897),
    [aux_sym_signed_number_token1] = ACTIONS(1895),
    [anon_sym_SQUOTE] = ACTIONS(1897),
    [anon_sym_DQUOTE] = ACTIONS(1897),
    [sym_prefix_operator] = ACTIONS(1897),
    [aux_sym_sign_operator_token1] = ACTIONS(1895),
    [sym_hash_operator] = ACTIONS(1897),
    [anon_sym_lsl] = ACTIONS(1895),
    [anon_sym_lsr] = ACTIONS(1895),
    [anon_sym_asr] = ACTIONS(1895),
    [aux_sym_pow_operator_token1] = ACTIONS(1897),
    [anon_sym_mod] = ACTIONS(1895),
    [anon_sym_land] = ACTIONS(1895),
    [anon_sym_lor] = ACTIONS(1895),
    [anon_sym_lxor] = ACTIONS(1895),
    [aux_sym_mult_operator_token1] = ACTIONS(1895),
    [aux_sym_add_operator_token1] = ACTIONS(1895),
    [sym_concat_operator] = ACTIONS(1897),
    [sym_rel_operator] = ACTIONS(1895),
    [anon_sym_AMP_AMP] = ACTIONS(1895),
    [anon_sym_or] = ACTIONS(1895),
    [anon_sym_PIPE_PIPE] = ACTIONS(1895),
    [sym_assign_operator] = ACTIONS(1897),
    [sym__uppercase_identifier] = ACTIONS(1897),
    [aux_sym_tag_token1] = ACTIONS(1897),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(933)] = {
    [sym__delimited_expression] = STATE(1471),
    [sym__simple_expression] = STATE(1486),
    [sym_typed_expression] = STATE(1471),
    [sym_list_expression] = STATE(1418),
    [sym_array_expression] = STATE(1418),
    [sym_record_expression] = STATE(1418),
    [sym_labeled_argument] = STATE(1490),
    [sym_prefix_expression] = STATE(1471),
    [sym_hash_expression] = STATE(1471),
    [sym_field_get_expression] = STATE(1471),
    [sym_array_get_expression] = STATE(1471),
    [sym_string_get_expression] = STATE(1471),
    [sym_bigarray_get_expression] = STATE(1471),
    [sym_local_open_expression] = STATE(1471),
    [sym_package_expression] = STATE(1418),
    [sym_new_expression] = STATE(1471),
    [sym_object_copy_expression] = STATE(1418),
    [sym_method_invocation] = STATE(1471),
    [sym_object_expression] = STATE(1471),
    [sym_parenthesized_expression] = STATE(1418),
    [sym_attribute] = STATE(933),
    [sym_extension] = STATE(1471),
    [sym_quoted_extension] = STATE(1471),
    [sym__constant] = STATE(1471),
    [sym_number] = STATE(1441),
    [sym_character] = STATE(1441),
    [sym_string] = STATE(1441),
    [sym_quoted_string] = STATE(1441),
    [sym_boolean] = STATE(1441),
    [sym_unit] = STATE(1418),
    [sym__value_name] = STATE(1442),
    [sym_parenthesized_operator] = STATE(1483),
    [sym_value_path] = STATE(1471),
    [sym_module_path] = STATE(7286),
    [sym_constructor_path] = STATE(1471),
    [sym__constructor_name] = STATE(1484),
    [sym_tag] = STATE(1471),
    [aux_sym_class_application_repeat1] = STATE(116),
    [ts_builtin_sym_end] = ACTIONS(1883),
    [sym__lowercase_identifier] = ACTIONS(1153),
    [anon_sym_SEMI_SEMI] = ACTIONS(1883),
    [anon_sym_let] = ACTIONS(1885),
    [anon_sym_and] = ACTIONS(1885),
    [anon_sym_TILDE] = ACTIONS(1155),
    [anon_sym_QMARK] = ACTIONS(1155),
    [anon_sym_LPAREN] = ACTIONS(19),
    [anon_sym_external] = ACTIONS(1885),
    [anon_sym_type] = ACTIONS(1885),
    [anon_sym_LBRACK] = ACTIONS(25),
    [anon_sym_true] = ACTIONS(27),
    [anon_sym_false] = ACTIONS(27),
    [anon_sym_LBRACE] = ACTIONS(29),
    [anon_sym_exception] = ACTIONS(1885),
    [anon_sym_module] = ACTIONS(1885),
    [anon_sym_open] = ACTIONS(1885),
    [anon_sym_include] = ACTIONS(1885),
    [anon_sym_class] = ACTIONS(1885),
    [anon_sym_val] = ACTIONS(1885),
    [anon_sym_object] = ACTIONS(43),
    [anon_sym_LBRACK_PIPE] = ACTIONS(47),
    [anon_sym_new] = ACTIONS(65),
    [anon_sym_LBRACE_LT] = ACTIONS(67),
    [anon_sym_begin] = ACTIONS(69),
    [sym_ocamlyacc_value] = ACTIONS(71),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1885),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1883),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(75),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(77),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1883),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1883),
    [aux_sym_number_token1] = ACTIONS(83),
    [anon_sym_SQUOTE] = ACTIONS(87),
    [anon_sym_DQUOTE] = ACTIONS(89),
    [sym_prefix_operator] = ACTIONS(91),
    [sym_let_operator] = ACTIONS(1883),
    [sym__uppercase_identifier] = ACTIONS(99),
    [aux_sym_directive_token1] = ACTIONS(1883),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(934)] = {
    [sym_attribute] = STATE(934),
    [sym__lowercase_identifier] = ACTIONS(1899),
    [anon_sym_and] = ACTIONS(1899),
    [anon_sym_TILDE] = ACTIONS(1899),
    [anon_sym_QMARK] = ACTIONS(1899),
    [anon_sym_LPAREN] = ACTIONS(1901),
    [anon_sym_RPAREN] = ACTIONS(1901),
    [anon_sym_COMMA] = ACTIONS(1901),
    [anon_sym_PIPE] = ACTIONS(1899),
    [anon_sym_COLON2] = ACTIONS(1899),
    [anon_sym_DOT] = ACTIONS(1901),
    [anon_sym_DASH_GT] = ACTIONS(1899),
    [anon_sym_LBRACK] = ACTIONS(1899),
    [anon_sym_RBRACK] = ACTIONS(1901),
    [anon_sym_true] = ACTIONS(1899),
    [anon_sym_false] = ACTIONS(1899),
    [anon_sym_LBRACE] = ACTIONS(1899),
    [anon_sym_SEMI] = ACTIONS(1901),
    [anon_sym_RBRACE] = ACTIONS(1901),
    [anon_sym_constraint] = ACTIONS(1899),
    [anon_sym_val] = ACTIONS(1899),
    [anon_sym_end] = ACTIONS(1899),
    [anon_sym_with] = ACTIONS(1899),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_object] = ACTIONS(1899),
    [anon_sym_inherit] = ACTIONS(1899),
    [anon_sym_method] = ACTIONS(1899),
    [anon_sym_in] = ACTIONS(1899),
    [anon_sym_as] = ACTIONS(1899),
    [anon_sym_initializer] = ACTIONS(1899),
    [anon_sym_AMP] = ACTIONS(1899),
    [anon_sym_POUND] = ACTIONS(1899),
    [anon_sym_COLON_COLON] = ACTIONS(1901),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1901),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1901),
    [anon_sym_LT_DASH] = ACTIONS(1899),
    [anon_sym_then] = ACTIONS(1899),
    [anon_sym_else] = ACTIONS(1899),
    [anon_sym_do] = ACTIONS(1899),
    [anon_sym_done] = ACTIONS(1899),
    [anon_sym_to] = ACTIONS(1899),
    [anon_sym_downto] = ACTIONS(1899),
    [anon_sym_new] = ACTIONS(1899),
    [anon_sym_LBRACE_LT] = ACTIONS(1901),
    [anon_sym_GT_RBRACE] = ACTIONS(1901),
    [anon_sym_begin] = ACTIONS(1899),
    [sym_ocamlyacc_value] = ACTIONS(1901),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1899),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1901),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1899),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1899),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1901),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1901),
    [aux_sym_number_token1] = ACTIONS(1901),
    [aux_sym_signed_number_token1] = ACTIONS(1899),
    [anon_sym_SQUOTE] = ACTIONS(1901),
    [anon_sym_DQUOTE] = ACTIONS(1901),
    [sym_prefix_operator] = ACTIONS(1901),
    [aux_sym_sign_operator_token1] = ACTIONS(1899),
    [sym_hash_operator] = ACTIONS(1901),
    [anon_sym_lsl] = ACTIONS(1899),
    [anon_sym_lsr] = ACTIONS(1899),
    [anon_sym_asr] = ACTIONS(1899),
    [aux_sym_pow_operator_token1] = ACTIONS(1901),
    [anon_sym_mod] = ACTIONS(1899),
    [anon_sym_land] = ACTIONS(1899),
    [anon_sym_lor] = ACTIONS(1899),
    [anon_sym_lxor] = ACTIONS(1899),
    [aux_sym_mult_operator_token1] = ACTIONS(1899),
    [aux_sym_add_operator_token1] = ACTIONS(1899),
    [sym_concat_operator] = ACTIONS(1901),
    [sym_rel_operator] = ACTIONS(1899),
    [anon_sym_AMP_AMP] = ACTIONS(1899),
    [anon_sym_or] = ACTIONS(1899),
    [anon_sym_PIPE_PIPE] = ACTIONS(1899),
    [sym_assign_operator] = ACTIONS(1901),
    [sym__uppercase_identifier] = ACTIONS(1901),
    [aux_sym_tag_token1] = ACTIONS(1901),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(935)] = {
    [sym_attribute] = STATE(935),
    [sym__lowercase_identifier] = ACTIONS(1903),
    [anon_sym_and] = ACTIONS(1903),
    [anon_sym_TILDE] = ACTIONS(1903),
    [anon_sym_QMARK] = ACTIONS(1903),
    [anon_sym_LPAREN] = ACTIONS(1905),
    [anon_sym_RPAREN] = ACTIONS(1905),
    [anon_sym_COMMA] = ACTIONS(1905),
    [anon_sym_PIPE] = ACTIONS(1903),
    [anon_sym_COLON2] = ACTIONS(1903),
    [anon_sym_DOT] = ACTIONS(1905),
    [anon_sym_DASH_GT] = ACTIONS(1903),
    [anon_sym_LBRACK] = ACTIONS(1903),
    [anon_sym_RBRACK] = ACTIONS(1905),
    [anon_sym_true] = ACTIONS(1903),
    [anon_sym_false] = ACTIONS(1903),
    [anon_sym_LBRACE] = ACTIONS(1903),
    [anon_sym_SEMI] = ACTIONS(1905),
    [anon_sym_RBRACE] = ACTIONS(1905),
    [anon_sym_constraint] = ACTIONS(1903),
    [anon_sym_val] = ACTIONS(1903),
    [anon_sym_end] = ACTIONS(1903),
    [anon_sym_with] = ACTIONS(1903),
    [anon_sym_COLON_GT] = ACTIONS(1905),
    [anon_sym_object] = ACTIONS(1903),
    [anon_sym_inherit] = ACTIONS(1903),
    [anon_sym_method] = ACTIONS(1903),
    [anon_sym_in] = ACTIONS(1903),
    [anon_sym_as] = ACTIONS(1903),
    [anon_sym_initializer] = ACTIONS(1903),
    [anon_sym_AMP] = ACTIONS(1903),
    [anon_sym_POUND] = ACTIONS(1903),
    [anon_sym_COLON_COLON] = ACTIONS(1905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1905),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1905),
    [anon_sym_LT_DASH] = ACTIONS(1903),
    [anon_sym_then] = ACTIONS(1903),
    [anon_sym_else] = ACTIONS(1903),
    [anon_sym_do] = ACTIONS(1903),
    [anon_sym_done] = ACTIONS(1903),
    [anon_sym_to] = ACTIONS(1903),
    [anon_sym_downto] = ACTIONS(1903),
    [anon_sym_new] = ACTIONS(1903),
    [anon_sym_LBRACE_LT] = ACTIONS(1905),
    [anon_sym_GT_RBRACE] = ACTIONS(1905),
    [anon_sym_begin] = ACTIONS(1903),
    [sym_ocamlyacc_value] = ACTIONS(1905),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1903),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1905),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1903),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1903),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1905),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1905),
    [aux_sym_number_token1] = ACTIONS(1905),
    [aux_sym_signed_number_token1] = ACTIONS(1903),
    [anon_sym_SQUOTE] = ACTIONS(1905),
    [anon_sym_DQUOTE] = ACTIONS(1905),
    [sym_prefix_operator] = ACTIONS(1905),
    [aux_sym_sign_operator_token1] = ACTIONS(1903),
    [sym_hash_operator] = ACTIONS(1905),
    [anon_sym_lsl] = ACTIONS(1903),
    [anon_sym_lsr] = ACTIONS(1903),
    [anon_sym_asr] = ACTIONS(1903),
    [aux_sym_pow_operator_token1] = ACTIONS(1905),
    [anon_sym_mod] = ACTIONS(1903),
    [anon_sym_land] = ACTIONS(1903),
    [anon_sym_lor] = ACTIONS(1903),
    [anon_sym_lxor] = ACTIONS(1903),
    [aux_sym_mult_operator_token1] = ACTIONS(1903),
    [aux_sym_add_operator_token1] = ACTIONS(1903),
    [sym_concat_operator] = ACTIONS(1905),
    [sym_rel_operator] = ACTIONS(1903),
    [anon_sym_AMP_AMP] = ACTIONS(1903),
    [anon_sym_or] = ACTIONS(1903),
    [anon_sym_PIPE_PIPE] = ACTIONS(1903),
    [sym_assign_operator] = ACTIONS(1905),
    [sym__uppercase_identifier] = ACTIONS(1905),
    [aux_sym_tag_token1] = ACTIONS(1905),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(936)] = {
    [sym_attribute] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1907),
    [anon_sym_and] = ACTIONS(1907),
    [anon_sym_TILDE] = ACTIONS(1907),
    [anon_sym_QMARK] = ACTIONS(1907),
    [anon_sym_LPAREN] = ACTIONS(1909),
    [anon_sym_RPAREN] = ACTIONS(1909),
    [anon_sym_COMMA] = ACTIONS(1909),
    [anon_sym_PIPE] = ACTIONS(1907),
    [anon_sym_COLON2] = ACTIONS(1907),
    [anon_sym_DOT] = ACTIONS(1909),
    [anon_sym_DASH_GT] = ACTIONS(1907),
    [anon_sym_LBRACK] = ACTIONS(1907),
    [anon_sym_RBRACK] = ACTIONS(1909),
    [anon_sym_true] = ACTIONS(1907),
    [anon_sym_false] = ACTIONS(1907),
    [anon_sym_LBRACE] = ACTIONS(1907),
    [anon_sym_SEMI] = ACTIONS(1909),
    [anon_sym_RBRACE] = ACTIONS(1909),
    [anon_sym_constraint] = ACTIONS(1907),
    [anon_sym_val] = ACTIONS(1907),
    [anon_sym_end] = ACTIONS(1907),
    [anon_sym_with] = ACTIONS(1907),
    [anon_sym_COLON_GT] = ACTIONS(1909),
    [anon_sym_object] = ACTIONS(1907),
    [anon_sym_inherit] = ACTIONS(1907),
    [anon_sym_method] = ACTIONS(1907),
    [anon_sym_in] = ACTIONS(1907),
    [anon_sym_as] = ACTIONS(1907),
    [anon_sym_initializer] = ACTIONS(1907),
    [anon_sym_AMP] = ACTIONS(1907),
    [anon_sym_POUND] = ACTIONS(1907),
    [anon_sym_COLON_COLON] = ACTIONS(1909),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1909),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1909),
    [anon_sym_LT_DASH] = ACTIONS(1907),
    [anon_sym_then] = ACTIONS(1907),
    [anon_sym_else] = ACTIONS(1907),
    [anon_sym_do] = ACTIONS(1907),
    [anon_sym_done] = ACTIONS(1907),
    [anon_sym_to] = ACTIONS(1907),
    [anon_sym_downto] = ACTIONS(1907),
    [anon_sym_new] = ACTIONS(1907),
    [anon_sym_LBRACE_LT] = ACTIONS(1909),
    [anon_sym_GT_RBRACE] = ACTIONS(1909),
    [anon_sym_begin] = ACTIONS(1907),
    [sym_ocamlyacc_value] = ACTIONS(1909),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1907),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1909),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1907),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1907),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1909),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1909),
    [aux_sym_number_token1] = ACTIONS(1909),
    [aux_sym_signed_number_token1] = ACTIONS(1907),
    [anon_sym_SQUOTE] = ACTIONS(1909),
    [anon_sym_DQUOTE] = ACTIONS(1909),
    [sym_prefix_operator] = ACTIONS(1909),
    [aux_sym_sign_operator_token1] = ACTIONS(1907),
    [sym_hash_operator] = ACTIONS(1909),
    [anon_sym_lsl] = ACTIONS(1907),
    [anon_sym_lsr] = ACTIONS(1907),
    [anon_sym_asr] = ACTIONS(1907),
    [aux_sym_pow_operator_token1] = ACTIONS(1909),
    [anon_sym_mod] = ACTIONS(1907),
    [anon_sym_land] = ACTIONS(1907),
    [anon_sym_lor] = ACTIONS(1907),
    [anon_sym_lxor] = ACTIONS(1907),
    [aux_sym_mult_operator_token1] = ACTIONS(1907),
    [aux_sym_add_operator_token1] = ACTIONS(1907),
    [sym_concat_operator] = ACTIONS(1909),
    [sym_rel_operator] = ACTIONS(1907),
    [anon_sym_AMP_AMP] = ACTIONS(1907),
    [anon_sym_or] = ACTIONS(1907),
    [anon_sym_PIPE_PIPE] = ACTIONS(1907),
    [sym_assign_operator] = ACTIONS(1909),
    [sym__uppercase_identifier] = ACTIONS(1909),
    [aux_sym_tag_token1] = ACTIONS(1909),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(937)] = {
    [sym_attribute] = STATE(937),
    [sym__lowercase_identifier] = ACTIONS(1911),
    [anon_sym_and] = ACTIONS(1911),
    [anon_sym_TILDE] = ACTIONS(1911),
    [anon_sym_QMARK] = ACTIONS(1911),
    [anon_sym_LPAREN] = ACTIONS(1913),
    [anon_sym_RPAREN] = ACTIONS(1913),
    [anon_sym_COMMA] = ACTIONS(1913),
    [anon_sym_PIPE] = ACTIONS(1911),
    [anon_sym_COLON2] = ACTIONS(1911),
    [anon_sym_DOT] = ACTIONS(1913),
    [anon_sym_DASH_GT] = ACTIONS(1911),
    [anon_sym_LBRACK] = ACTIONS(1911),
    [anon_sym_RBRACK] = ACTIONS(1913),
    [anon_sym_true] = ACTIONS(1911),
    [anon_sym_false] = ACTIONS(1911),
    [anon_sym_LBRACE] = ACTIONS(1911),
    [anon_sym_SEMI] = ACTIONS(1913),
    [anon_sym_RBRACE] = ACTIONS(1913),
    [anon_sym_constraint] = ACTIONS(1911),
    [anon_sym_val] = ACTIONS(1911),
    [anon_sym_end] = ACTIONS(1911),
    [anon_sym_with] = ACTIONS(1911),
    [anon_sym_COLON_GT] = ACTIONS(1913),
    [anon_sym_object] = ACTIONS(1911),
    [anon_sym_inherit] = ACTIONS(1911),
    [anon_sym_method] = ACTIONS(1911),
    [anon_sym_in] = ACTIONS(1911),
    [anon_sym_as] = ACTIONS(1911),
    [anon_sym_initializer] = ACTIONS(1911),
    [anon_sym_AMP] = ACTIONS(1911),
    [anon_sym_POUND] = ACTIONS(1911),
    [anon_sym_COLON_COLON] = ACTIONS(1913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1913),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1913),
    [anon_sym_LT_DASH] = ACTIONS(1911),
    [anon_sym_then] = ACTIONS(1911),
    [anon_sym_else] = ACTIONS(1911),
    [anon_sym_do] = ACTIONS(1911),
    [anon_sym_done] = ACTIONS(1911),
    [anon_sym_to] = ACTIONS(1911),
    [anon_sym_downto] = ACTIONS(1911),
    [anon_sym_new] = ACTIONS(1911),
    [anon_sym_LBRACE_LT] = ACTIONS(1913),
    [anon_sym_GT_RBRACE] = ACTIONS(1913),
    [anon_sym_begin] = ACTIONS(1911),
    [sym_ocamlyacc_value] = ACTIONS(1913),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1911),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1913),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1911),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1911),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1913),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1913),
    [aux_sym_number_token1] = ACTIONS(1913),
    [aux_sym_signed_number_token1] = ACTIONS(1911),
    [anon_sym_SQUOTE] = ACTIONS(1913),
    [anon_sym_DQUOTE] = ACTIONS(1913),
    [sym_prefix_operator] = ACTIONS(1913),
    [aux_sym_sign_operator_token1] = ACTIONS(1911),
    [sym_hash_operator] = ACTIONS(1913),
    [anon_sym_lsl] = ACTIONS(1911),
    [anon_sym_lsr] = ACTIONS(1911),
    [anon_sym_asr] = ACTIONS(1911),
    [aux_sym_pow_operator_token1] = ACTIONS(1913),
    [anon_sym_mod] = ACTIONS(1911),
    [anon_sym_land] = ACTIONS(1911),
    [anon_sym_lor] = ACTIONS(1911),
    [anon_sym_lxor] = ACTIONS(1911),
    [aux_sym_mult_operator_token1] = ACTIONS(1911),
    [aux_sym_add_operator_token1] = ACTIONS(1911),
    [sym_concat_operator] = ACTIONS(1913),
    [sym_rel_operator] = ACTIONS(1911),
    [anon_sym_AMP_AMP] = ACTIONS(1911),
    [anon_sym_or] = ACTIONS(1911),
    [anon_sym_PIPE_PIPE] = ACTIONS(1911),
    [sym_assign_operator] = ACTIONS(1913),
    [sym__uppercase_identifier] = ACTIONS(1913),
    [aux_sym_tag_token1] = ACTIONS(1913),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(938)] = {
    [sym_attribute] = STATE(938),
    [sym__lowercase_identifier] = ACTIONS(1915),
    [anon_sym_and] = ACTIONS(1915),
    [anon_sym_TILDE] = ACTIONS(1915),
    [anon_sym_QMARK] = ACTIONS(1915),
    [anon_sym_LPAREN] = ACTIONS(1917),
    [anon_sym_RPAREN] = ACTIONS(1917),
    [anon_sym_COMMA] = ACTIONS(1917),
    [anon_sym_PIPE] = ACTIONS(1915),
    [anon_sym_COLON2] = ACTIONS(1915),
    [anon_sym_DOT] = ACTIONS(1917),
    [anon_sym_DASH_GT] = ACTIONS(1915),
    [anon_sym_LBRACK] = ACTIONS(1915),
    [anon_sym_RBRACK] = ACTIONS(1917),
    [anon_sym_true] = ACTIONS(1915),
    [anon_sym_false] = ACTIONS(1915),
    [anon_sym_LBRACE] = ACTIONS(1915),
    [anon_sym_SEMI] = ACTIONS(1917),
    [anon_sym_RBRACE] = ACTIONS(1917),
    [anon_sym_constraint] = ACTIONS(1915),
    [anon_sym_val] = ACTIONS(1915),
    [anon_sym_end] = ACTIONS(1915),
    [anon_sym_with] = ACTIONS(1915),
    [anon_sym_COLON_GT] = ACTIONS(1917),
    [anon_sym_object] = ACTIONS(1915),
    [anon_sym_inherit] = ACTIONS(1915),
    [anon_sym_method] = ACTIONS(1915),
    [anon_sym_in] = ACTIONS(1915),
    [anon_sym_as] = ACTIONS(1915),
    [anon_sym_initializer] = ACTIONS(1915),
    [anon_sym_AMP] = ACTIONS(1915),
    [anon_sym_POUND] = ACTIONS(1915),
    [anon_sym_COLON_COLON] = ACTIONS(1917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1917),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1917),
    [anon_sym_LT_DASH] = ACTIONS(1915),
    [anon_sym_then] = ACTIONS(1915),
    [anon_sym_else] = ACTIONS(1915),
    [anon_sym_do] = ACTIONS(1915),
    [anon_sym_done] = ACTIONS(1915),
    [anon_sym_to] = ACTIONS(1915),
    [anon_sym_downto] = ACTIONS(1915),
    [anon_sym_new] = ACTIONS(1915),
    [anon_sym_LBRACE_LT] = ACTIONS(1917),
    [anon_sym_GT_RBRACE] = ACTIONS(1917),
    [anon_sym_begin] = ACTIONS(1915),
    [sym_ocamlyacc_value] = ACTIONS(1917),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1915),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1917),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1915),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1915),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1917),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1917),
    [aux_sym_number_token1] = ACTIONS(1917),
    [aux_sym_signed_number_token1] = ACTIONS(1915),
    [anon_sym_SQUOTE] = ACTIONS(1917),
    [anon_sym_DQUOTE] = ACTIONS(1917),
    [sym_prefix_operator] = ACTIONS(1917),
    [aux_sym_sign_operator_token1] = ACTIONS(1915),
    [sym_hash_operator] = ACTIONS(1917),
    [anon_sym_lsl] = ACTIONS(1915),
    [anon_sym_lsr] = ACTIONS(1915),
    [anon_sym_asr] = ACTIONS(1915),
    [aux_sym_pow_operator_token1] = ACTIONS(1917),
    [anon_sym_mod] = ACTIONS(1915),
    [anon_sym_land] = ACTIONS(1915),
    [anon_sym_lor] = ACTIONS(1915),
    [anon_sym_lxor] = ACTIONS(1915),
    [aux_sym_mult_operator_token1] = ACTIONS(1915),
    [aux_sym_add_operator_token1] = ACTIONS(1915),
    [sym_concat_operator] = ACTIONS(1917),
    [sym_rel_operator] = ACTIONS(1915),
    [anon_sym_AMP_AMP] = ACTIONS(1915),
    [anon_sym_or] = ACTIONS(1915),
    [anon_sym_PIPE_PIPE] = ACTIONS(1915),
    [sym_assign_operator] = ACTIONS(1917),
    [sym__uppercase_identifier] = ACTIONS(1917),
    [aux_sym_tag_token1] = ACTIONS(1917),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(939)] = {
    [sym_attribute] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(1919),
    [anon_sym_and] = ACTIONS(1919),
    [anon_sym_TILDE] = ACTIONS(1919),
    [anon_sym_QMARK] = ACTIONS(1919),
    [anon_sym_LPAREN] = ACTIONS(1921),
    [anon_sym_RPAREN] = ACTIONS(1921),
    [anon_sym_COMMA] = ACTIONS(1921),
    [anon_sym_PIPE] = ACTIONS(1919),
    [anon_sym_COLON2] = ACTIONS(1919),
    [anon_sym_DOT] = ACTIONS(1921),
    [anon_sym_DASH_GT] = ACTIONS(1919),
    [anon_sym_LBRACK] = ACTIONS(1919),
    [anon_sym_RBRACK] = ACTIONS(1921),
    [anon_sym_true] = ACTIONS(1919),
    [anon_sym_false] = ACTIONS(1919),
    [anon_sym_LBRACE] = ACTIONS(1919),
    [anon_sym_SEMI] = ACTIONS(1921),
    [anon_sym_RBRACE] = ACTIONS(1921),
    [anon_sym_constraint] = ACTIONS(1919),
    [anon_sym_val] = ACTIONS(1919),
    [anon_sym_end] = ACTIONS(1919),
    [anon_sym_with] = ACTIONS(1919),
    [anon_sym_COLON_GT] = ACTIONS(1921),
    [anon_sym_object] = ACTIONS(1919),
    [anon_sym_inherit] = ACTIONS(1919),
    [anon_sym_method] = ACTIONS(1919),
    [anon_sym_in] = ACTIONS(1919),
    [anon_sym_as] = ACTIONS(1919),
    [anon_sym_initializer] = ACTIONS(1919),
    [anon_sym_AMP] = ACTIONS(1919),
    [anon_sym_POUND] = ACTIONS(1919),
    [anon_sym_COLON_COLON] = ACTIONS(1921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1921),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1921),
    [anon_sym_LT_DASH] = ACTIONS(1919),
    [anon_sym_then] = ACTIONS(1919),
    [anon_sym_else] = ACTIONS(1919),
    [anon_sym_do] = ACTIONS(1919),
    [anon_sym_done] = ACTIONS(1919),
    [anon_sym_to] = ACTIONS(1919),
    [anon_sym_downto] = ACTIONS(1919),
    [anon_sym_new] = ACTIONS(1919),
    [anon_sym_LBRACE_LT] = ACTIONS(1921),
    [anon_sym_GT_RBRACE] = ACTIONS(1921),
    [anon_sym_begin] = ACTIONS(1919),
    [sym_ocamlyacc_value] = ACTIONS(1921),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1919),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1921),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1919),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1919),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1921),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1921),
    [aux_sym_number_token1] = ACTIONS(1921),
    [aux_sym_signed_number_token1] = ACTIONS(1919),
    [anon_sym_SQUOTE] = ACTIONS(1921),
    [anon_sym_DQUOTE] = ACTIONS(1921),
    [sym_prefix_operator] = ACTIONS(1921),
    [aux_sym_sign_operator_token1] = ACTIONS(1919),
    [sym_hash_operator] = ACTIONS(1921),
    [anon_sym_lsl] = ACTIONS(1919),
    [anon_sym_lsr] = ACTIONS(1919),
    [anon_sym_asr] = ACTIONS(1919),
    [aux_sym_pow_operator_token1] = ACTIONS(1921),
    [anon_sym_mod] = ACTIONS(1919),
    [anon_sym_land] = ACTIONS(1919),
    [anon_sym_lor] = ACTIONS(1919),
    [anon_sym_lxor] = ACTIONS(1919),
    [aux_sym_mult_operator_token1] = ACTIONS(1919),
    [aux_sym_add_operator_token1] = ACTIONS(1919),
    [sym_concat_operator] = ACTIONS(1921),
    [sym_rel_operator] = ACTIONS(1919),
    [anon_sym_AMP_AMP] = ACTIONS(1919),
    [anon_sym_or] = ACTIONS(1919),
    [anon_sym_PIPE_PIPE] = ACTIONS(1919),
    [sym_assign_operator] = ACTIONS(1921),
    [sym__uppercase_identifier] = ACTIONS(1921),
    [aux_sym_tag_token1] = ACTIONS(1921),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(940)] = {
    [sym_attribute] = STATE(940),
    [sym__lowercase_identifier] = ACTIONS(1923),
    [anon_sym_and] = ACTIONS(1923),
    [anon_sym_TILDE] = ACTIONS(1923),
    [anon_sym_QMARK] = ACTIONS(1923),
    [anon_sym_LPAREN] = ACTIONS(1925),
    [anon_sym_RPAREN] = ACTIONS(1925),
    [anon_sym_COMMA] = ACTIONS(1925),
    [anon_sym_PIPE] = ACTIONS(1923),
    [anon_sym_COLON2] = ACTIONS(1923),
    [anon_sym_DOT] = ACTIONS(1925),
    [anon_sym_DASH_GT] = ACTIONS(1923),
    [anon_sym_LBRACK] = ACTIONS(1923),
    [anon_sym_RBRACK] = ACTIONS(1925),
    [anon_sym_true] = ACTIONS(1923),
    [anon_sym_false] = ACTIONS(1923),
    [anon_sym_LBRACE] = ACTIONS(1923),
    [anon_sym_SEMI] = ACTIONS(1925),
    [anon_sym_RBRACE] = ACTIONS(1925),
    [anon_sym_constraint] = ACTIONS(1923),
    [anon_sym_val] = ACTIONS(1923),
    [anon_sym_end] = ACTIONS(1923),
    [anon_sym_with] = ACTIONS(1923),
    [anon_sym_COLON_GT] = ACTIONS(1925),
    [anon_sym_object] = ACTIONS(1923),
    [anon_sym_inherit] = ACTIONS(1923),
    [anon_sym_method] = ACTIONS(1923),
    [anon_sym_in] = ACTIONS(1923),
    [anon_sym_as] = ACTIONS(1923),
    [anon_sym_initializer] = ACTIONS(1923),
    [anon_sym_AMP] = ACTIONS(1923),
    [anon_sym_POUND] = ACTIONS(1923),
    [anon_sym_COLON_COLON] = ACTIONS(1925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1925),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1925),
    [anon_sym_then] = ACTIONS(1923),
    [anon_sym_else] = ACTIONS(1923),
    [anon_sym_do] = ACTIONS(1923),
    [anon_sym_done] = ACTIONS(1923),
    [anon_sym_to] = ACTIONS(1923),
    [anon_sym_downto] = ACTIONS(1923),
    [anon_sym_new] = ACTIONS(1923),
    [anon_sym_LBRACE_LT] = ACTIONS(1925),
    [anon_sym_GT_RBRACE] = ACTIONS(1925),
    [anon_sym_begin] = ACTIONS(1923),
    [sym_ocamlyacc_value] = ACTIONS(1925),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1923),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1925),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1923),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1923),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1925),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1925),
    [aux_sym_number_token1] = ACTIONS(1925),
    [aux_sym_signed_number_token1] = ACTIONS(1923),
    [anon_sym_SQUOTE] = ACTIONS(1925),
    [anon_sym_DQUOTE] = ACTIONS(1925),
    [sym_prefix_operator] = ACTIONS(1925),
    [aux_sym_sign_operator_token1] = ACTIONS(1923),
    [sym_hash_operator] = ACTIONS(1925),
    [anon_sym_lsl] = ACTIONS(1923),
    [anon_sym_lsr] = ACTIONS(1923),
    [anon_sym_asr] = ACTIONS(1923),
    [aux_sym_pow_operator_token1] = ACTIONS(1925),
    [anon_sym_mod] = ACTIONS(1923),
    [anon_sym_land] = ACTIONS(1923),
    [anon_sym_lor] = ACTIONS(1923),
    [anon_sym_lxor] = ACTIONS(1923),
    [aux_sym_mult_operator_token1] = ACTIONS(1923),
    [aux_sym_add_operator_token1] = ACTIONS(1923),
    [sym_concat_operator] = ACTIONS(1925),
    [sym_rel_operator] = ACTIONS(1923),
    [anon_sym_AMP_AMP] = ACTIONS(1923),
    [anon_sym_or] = ACTIONS(1923),
    [anon_sym_PIPE_PIPE] = ACTIONS(1923),
    [sym_assign_operator] = ACTIONS(1925),
    [sym__uppercase_identifier] = ACTIONS(1925),
    [aux_sym_tag_token1] = ACTIONS(1925),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(941)] = {
    [sym_attribute] = STATE(941),
    [sym__lowercase_identifier] = ACTIONS(1927),
    [anon_sym_and] = ACTIONS(1927),
    [anon_sym_TILDE] = ACTIONS(1927),
    [anon_sym_QMARK] = ACTIONS(1927),
    [anon_sym_LPAREN] = ACTIONS(1929),
    [anon_sym_RPAREN] = ACTIONS(1929),
    [anon_sym_COMMA] = ACTIONS(1929),
    [anon_sym_PIPE] = ACTIONS(1927),
    [anon_sym_COLON2] = ACTIONS(1927),
    [anon_sym_DOT] = ACTIONS(1929),
    [anon_sym_DASH_GT] = ACTIONS(1927),
    [anon_sym_LBRACK] = ACTIONS(1927),
    [anon_sym_RBRACK] = ACTIONS(1929),
    [anon_sym_true] = ACTIONS(1927),
    [anon_sym_false] = ACTIONS(1927),
    [anon_sym_LBRACE] = ACTIONS(1927),
    [anon_sym_SEMI] = ACTIONS(1929),
    [anon_sym_RBRACE] = ACTIONS(1929),
    [anon_sym_constraint] = ACTIONS(1927),
    [anon_sym_val] = ACTIONS(1927),
    [anon_sym_end] = ACTIONS(1927),
    [anon_sym_with] = ACTIONS(1927),
    [anon_sym_COLON_GT] = ACTIONS(1929),
    [anon_sym_object] = ACTIONS(1927),
    [anon_sym_inherit] = ACTIONS(1927),
    [anon_sym_method] = ACTIONS(1927),
    [anon_sym_in] = ACTIONS(1927),
    [anon_sym_as] = ACTIONS(1927),
    [anon_sym_initializer] = ACTIONS(1927),
    [anon_sym_AMP] = ACTIONS(1927),
    [anon_sym_POUND] = ACTIONS(1927),
    [anon_sym_COLON_COLON] = ACTIONS(1929),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1929),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1929),
    [anon_sym_then] = ACTIONS(1927),
    [anon_sym_else] = ACTIONS(1927),
    [anon_sym_do] = ACTIONS(1927),
    [anon_sym_done] = ACTIONS(1927),
    [anon_sym_to] = ACTIONS(1927),
    [anon_sym_downto] = ACTIONS(1927),
    [anon_sym_new] = ACTIONS(1927),
    [anon_sym_LBRACE_LT] = ACTIONS(1929),
    [anon_sym_GT_RBRACE] = ACTIONS(1929),
    [anon_sym_begin] = ACTIONS(1927),
    [sym_ocamlyacc_value] = ACTIONS(1929),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1927),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1929),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1927),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1927),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1929),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1929),
    [aux_sym_number_token1] = ACTIONS(1929),
    [aux_sym_signed_number_token1] = ACTIONS(1927),
    [anon_sym_SQUOTE] = ACTIONS(1929),
    [anon_sym_DQUOTE] = ACTIONS(1929),
    [sym_prefix_operator] = ACTIONS(1929),
    [aux_sym_sign_operator_token1] = ACTIONS(1927),
    [sym_hash_operator] = ACTIONS(1929),
    [anon_sym_lsl] = ACTIONS(1927),
    [anon_sym_lsr] = ACTIONS(1927),
    [anon_sym_asr] = ACTIONS(1927),
    [aux_sym_pow_operator_token1] = ACTIONS(1929),
    [anon_sym_mod] = ACTIONS(1927),
    [anon_sym_land] = ACTIONS(1927),
    [anon_sym_lor] = ACTIONS(1927),
    [anon_sym_lxor] = ACTIONS(1927),
    [aux_sym_mult_operator_token1] = ACTIONS(1927),
    [aux_sym_add_operator_token1] = ACTIONS(1927),
    [sym_concat_operator] = ACTIONS(1929),
    [sym_rel_operator] = ACTIONS(1927),
    [anon_sym_AMP_AMP] = ACTIONS(1927),
    [anon_sym_or] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1927),
    [sym_assign_operator] = ACTIONS(1929),
    [sym__uppercase_identifier] = ACTIONS(1929),
    [aux_sym_tag_token1] = ACTIONS(1929),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(942)] = {
    [sym_attribute] = STATE(942),
    [sym__lowercase_identifier] = ACTIONS(1931),
    [anon_sym_and] = ACTIONS(1931),
    [anon_sym_TILDE] = ACTIONS(1931),
    [anon_sym_QMARK] = ACTIONS(1931),
    [anon_sym_LPAREN] = ACTIONS(1933),
    [anon_sym_RPAREN] = ACTIONS(1933),
    [anon_sym_COMMA] = ACTIONS(1933),
    [anon_sym_PIPE] = ACTIONS(1931),
    [anon_sym_COLON2] = ACTIONS(1931),
    [anon_sym_DOT] = ACTIONS(919),
    [anon_sym_DASH_GT] = ACTIONS(1931),
    [anon_sym_LBRACK] = ACTIONS(1931),
    [anon_sym_RBRACK] = ACTIONS(1933),
    [anon_sym_true] = ACTIONS(1931),
    [anon_sym_false] = ACTIONS(1931),
    [anon_sym_LBRACE] = ACTIONS(1931),
    [anon_sym_SEMI] = ACTIONS(1933),
    [anon_sym_RBRACE] = ACTIONS(1933),
    [anon_sym_constraint] = ACTIONS(1931),
    [anon_sym_val] = ACTIONS(1931),
    [anon_sym_end] = ACTIONS(1931),
    [anon_sym_with] = ACTIONS(1931),
    [anon_sym_COLON_GT] = ACTIONS(1933),
    [anon_sym_object] = ACTIONS(1931),
    [anon_sym_inherit] = ACTIONS(1931),
    [anon_sym_method] = ACTIONS(1931),
    [anon_sym_in] = ACTIONS(1931),
    [anon_sym_as] = ACTIONS(1931),
    [anon_sym_initializer] = ACTIONS(1931),
    [anon_sym_AMP] = ACTIONS(1931),
    [anon_sym_POUND] = ACTIONS(921),
    [anon_sym_COLON_COLON] = ACTIONS(1933),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1933),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1933),
    [anon_sym_then] = ACTIONS(1931),
    [anon_sym_else] = ACTIONS(1931),
    [anon_sym_do] = ACTIONS(1931),
    [anon_sym_done] = ACTIONS(1931),
    [anon_sym_to] = ACTIONS(1931),
    [anon_sym_downto] = ACTIONS(1931),
    [anon_sym_new] = ACTIONS(1931),
    [anon_sym_LBRACE_LT] = ACTIONS(1933),
    [anon_sym_GT_RBRACE] = ACTIONS(1933),
    [anon_sym_begin] = ACTIONS(1931),
    [sym_ocamlyacc_value] = ACTIONS(1933),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1931),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1933),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1933),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1933),
    [aux_sym_number_token1] = ACTIONS(1933),
    [aux_sym_signed_number_token1] = ACTIONS(1931),
    [anon_sym_SQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1933),
    [sym_prefix_operator] = ACTIONS(1933),
    [aux_sym_sign_operator_token1] = ACTIONS(1931),
    [sym_hash_operator] = ACTIONS(923),
    [anon_sym_lsl] = ACTIONS(1931),
    [anon_sym_lsr] = ACTIONS(1931),
    [anon_sym_asr] = ACTIONS(1931),
    [aux_sym_pow_operator_token1] = ACTIONS(1933),
    [anon_sym_mod] = ACTIONS(1931),
    [anon_sym_land] = ACTIONS(1931),
    [anon_sym_lor] = ACTIONS(1931),
    [anon_sym_lxor] = ACTIONS(1931),
    [aux_sym_mult_operator_token1] = ACTIONS(1931),
    [aux_sym_add_operator_token1] = ACTIONS(1931),
    [sym_concat_operator] = ACTIONS(1933),
    [sym_rel_operator] = ACTIONS(1931),
    [anon_sym_AMP_AMP] = ACTIONS(1931),
    [anon_sym_or] = ACTIONS(1931),
    [anon_sym_PIPE_PIPE] = ACTIONS(1931),
    [sym_assign_operator] = ACTIONS(1933),
    [sym__uppercase_identifier] = ACTIONS(1933),
    [aux_sym_tag_token1] = ACTIONS(1933),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(943)] = {
    [sym_attribute] = STATE(943),
    [sym__lowercase_identifier] = ACTIONS(1935),
    [anon_sym_and] = ACTIONS(1935),
    [anon_sym_TILDE] = ACTIONS(1935),
    [anon_sym_QMARK] = ACTIONS(1935),
    [anon_sym_LPAREN] = ACTIONS(1937),
    [anon_sym_RPAREN] = ACTIONS(1937),
    [anon_sym_COMMA] = ACTIONS(1937),
    [anon_sym_PIPE] = ACTIONS(1935),
    [anon_sym_COLON2] = ACTIONS(1935),
    [anon_sym_DOT] = ACTIONS(1937),
    [anon_sym_DASH_GT] = ACTIONS(1935),
    [anon_sym_LBRACK] = ACTIONS(1935),
    [anon_sym_RBRACK] = ACTIONS(1937),
    [anon_sym_true] = ACTIONS(1935),
    [anon_sym_false] = ACTIONS(1935),
    [anon_sym_LBRACE] = ACTIONS(1935),
    [anon_sym_SEMI] = ACTIONS(1937),
    [anon_sym_RBRACE] = ACTIONS(1937),
    [anon_sym_constraint] = ACTIONS(1935),
    [anon_sym_val] = ACTIONS(1935),
    [anon_sym_end] = ACTIONS(1935),
    [anon_sym_with] = ACTIONS(1935),
    [anon_sym_COLON_GT] = ACTIONS(1937),
    [anon_sym_object] = ACTIONS(1935),
    [anon_sym_inherit] = ACTIONS(1935),
    [anon_sym_method] = ACTIONS(1935),
    [anon_sym_in] = ACTIONS(1935),
    [anon_sym_as] = ACTIONS(1935),
    [anon_sym_initializer] = ACTIONS(1935),
    [anon_sym_AMP] = ACTIONS(1935),
    [anon_sym_POUND] = ACTIONS(1935),
    [anon_sym_COLON_COLON] = ACTIONS(1937),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1937),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1937),
    [anon_sym_then] = ACTIONS(1935),
    [anon_sym_else] = ACTIONS(1935),
    [anon_sym_do] = ACTIONS(1935),
    [anon_sym_done] = ACTIONS(1935),
    [anon_sym_to] = ACTIONS(1935),
    [anon_sym_downto] = ACTIONS(1935),
    [anon_sym_new] = ACTIONS(1935),
    [anon_sym_LBRACE_LT] = ACTIONS(1937),
    [anon_sym_GT_RBRACE] = ACTIONS(1937),
    [anon_sym_begin] = ACTIONS(1935),
    [sym_ocamlyacc_value] = ACTIONS(1937),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1935),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1937),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1935),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1935),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1937),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1937),
    [aux_sym_number_token1] = ACTIONS(1937),
    [aux_sym_signed_number_token1] = ACTIONS(1935),
    [anon_sym_SQUOTE] = ACTIONS(1937),
    [anon_sym_DQUOTE] = ACTIONS(1937),
    [sym_prefix_operator] = ACTIONS(1937),
    [aux_sym_sign_operator_token1] = ACTIONS(1935),
    [sym_hash_operator] = ACTIONS(1937),
    [anon_sym_lsl] = ACTIONS(1935),
    [anon_sym_lsr] = ACTIONS(1935),
    [anon_sym_asr] = ACTIONS(1935),
    [aux_sym_pow_operator_token1] = ACTIONS(1937),
    [anon_sym_mod] = ACTIONS(1935),
    [anon_sym_land] = ACTIONS(1935),
    [anon_sym_lor] = ACTIONS(1935),
    [anon_sym_lxor] = ACTIONS(1935),
    [aux_sym_mult_operator_token1] = ACTIONS(1935),
    [aux_sym_add_operator_token1] = ACTIONS(1935),
    [sym_concat_operator] = ACTIONS(1937),
    [sym_rel_operator] = ACTIONS(1935),
    [anon_sym_AMP_AMP] = ACTIONS(1935),
    [anon_sym_or] = ACTIONS(1935),
    [anon_sym_PIPE_PIPE] = ACTIONS(1935),
    [sym_assign_operator] = ACTIONS(1937),
    [sym__uppercase_identifier] = ACTIONS(1937),
    [aux_sym_tag_token1] = ACTIONS(1937),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(944)] = {
    [sym_attribute] = STATE(944),
    [sym__lowercase_identifier] = ACTIONS(1939),
    [anon_sym_and] = ACTIONS(1939),
    [anon_sym_TILDE] = ACTIONS(1939),
    [anon_sym_QMARK] = ACTIONS(1939),
    [anon_sym_LPAREN] = ACTIONS(1941),
    [anon_sym_RPAREN] = ACTIONS(1941),
    [anon_sym_COMMA] = ACTIONS(1941),
    [anon_sym_PIPE] = ACTIONS(1939),
    [anon_sym_COLON2] = ACTIONS(1939),
    [anon_sym_DOT] = ACTIONS(1941),
    [anon_sym_DASH_GT] = ACTIONS(1939),
    [anon_sym_LBRACK] = ACTIONS(1939),
    [anon_sym_RBRACK] = ACTIONS(1941),
    [anon_sym_true] = ACTIONS(1939),
    [anon_sym_false] = ACTIONS(1939),
    [anon_sym_LBRACE] = ACTIONS(1939),
    [anon_sym_SEMI] = ACTIONS(1941),
    [anon_sym_RBRACE] = ACTIONS(1941),
    [anon_sym_constraint] = ACTIONS(1939),
    [anon_sym_val] = ACTIONS(1939),
    [anon_sym_end] = ACTIONS(1939),
    [anon_sym_with] = ACTIONS(1939),
    [anon_sym_COLON_GT] = ACTIONS(1941),
    [anon_sym_object] = ACTIONS(1939),
    [anon_sym_inherit] = ACTIONS(1939),
    [anon_sym_method] = ACTIONS(1939),
    [anon_sym_in] = ACTIONS(1939),
    [anon_sym_as] = ACTIONS(1939),
    [anon_sym_initializer] = ACTIONS(1939),
    [anon_sym_AMP] = ACTIONS(1939),
    [anon_sym_POUND] = ACTIONS(1939),
    [anon_sym_COLON_COLON] = ACTIONS(1941),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1941),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1941),
    [anon_sym_then] = ACTIONS(1939),
    [anon_sym_else] = ACTIONS(1939),
    [anon_sym_do] = ACTIONS(1939),
    [anon_sym_done] = ACTIONS(1939),
    [anon_sym_to] = ACTIONS(1939),
    [anon_sym_downto] = ACTIONS(1939),
    [anon_sym_new] = ACTIONS(1939),
    [anon_sym_LBRACE_LT] = ACTIONS(1941),
    [anon_sym_GT_RBRACE] = ACTIONS(1941),
    [anon_sym_begin] = ACTIONS(1939),
    [sym_ocamlyacc_value] = ACTIONS(1941),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1939),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1941),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1939),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1939),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1941),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1941),
    [aux_sym_number_token1] = ACTIONS(1941),
    [aux_sym_signed_number_token1] = ACTIONS(1939),
    [anon_sym_SQUOTE] = ACTIONS(1941),
    [anon_sym_DQUOTE] = ACTIONS(1941),
    [sym_prefix_operator] = ACTIONS(1941),
    [aux_sym_sign_operator_token1] = ACTIONS(1939),
    [sym_hash_operator] = ACTIONS(1941),
    [anon_sym_lsl] = ACTIONS(1939),
    [anon_sym_lsr] = ACTIONS(1939),
    [anon_sym_asr] = ACTIONS(1939),
    [aux_sym_pow_operator_token1] = ACTIONS(1941),
    [anon_sym_mod] = ACTIONS(1939),
    [anon_sym_land] = ACTIONS(1939),
    [anon_sym_lor] = ACTIONS(1939),
    [anon_sym_lxor] = ACTIONS(1939),
    [aux_sym_mult_operator_token1] = ACTIONS(1939),
    [aux_sym_add_operator_token1] = ACTIONS(1939),
    [sym_concat_operator] = ACTIONS(1941),
    [sym_rel_operator] = ACTIONS(1939),
    [anon_sym_AMP_AMP] = ACTIONS(1939),
    [anon_sym_or] = ACTIONS(1939),
    [anon_sym_PIPE_PIPE] = ACTIONS(1939),
    [sym_assign_operator] = ACTIONS(1941),
    [sym__uppercase_identifier] = ACTIONS(1941),
    [aux_sym_tag_token1] = ACTIONS(1941),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(945)] = {
    [sym_attribute] = STATE(945),
    [sym__lowercase_identifier] = ACTIONS(1943),
    [anon_sym_and] = ACTIONS(1943),
    [anon_sym_TILDE] = ACTIONS(1943),
    [anon_sym_QMARK] = ACTIONS(1943),
    [anon_sym_LPAREN] = ACTIONS(1945),
    [anon_sym_RPAREN] = ACTIONS(1945),
    [anon_sym_COMMA] = ACTIONS(1945),
    [anon_sym_PIPE] = ACTIONS(1943),
    [anon_sym_COLON2] = ACTIONS(1943),
    [anon_sym_DOT] = ACTIONS(1945),
    [anon_sym_DASH_GT] = ACTIONS(1943),
    [anon_sym_LBRACK] = ACTIONS(1943),
    [anon_sym_RBRACK] = ACTIONS(1945),
    [anon_sym_true] = ACTIONS(1943),
    [anon_sym_false] = ACTIONS(1943),
    [anon_sym_LBRACE] = ACTIONS(1943),
    [anon_sym_SEMI] = ACTIONS(1945),
    [anon_sym_RBRACE] = ACTIONS(1945),
    [anon_sym_constraint] = ACTIONS(1943),
    [anon_sym_val] = ACTIONS(1943),
    [anon_sym_end] = ACTIONS(1943),
    [anon_sym_with] = ACTIONS(1943),
    [anon_sym_COLON_GT] = ACTIONS(1945),
    [anon_sym_object] = ACTIONS(1943),
    [anon_sym_inherit] = ACTIONS(1943),
    [anon_sym_method] = ACTIONS(1943),
    [anon_sym_in] = ACTIONS(1943),
    [anon_sym_as] = ACTIONS(1943),
    [anon_sym_initializer] = ACTIONS(1943),
    [anon_sym_AMP] = ACTIONS(1943),
    [anon_sym_POUND] = ACTIONS(1943),
    [anon_sym_COLON_COLON] = ACTIONS(1945),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1945),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1945),
    [anon_sym_then] = ACTIONS(1943),
    [anon_sym_else] = ACTIONS(1943),
    [anon_sym_do] = ACTIONS(1943),
    [anon_sym_done] = ACTIONS(1943),
    [anon_sym_to] = ACTIONS(1943),
    [anon_sym_downto] = ACTIONS(1943),
    [anon_sym_new] = ACTIONS(1943),
    [anon_sym_LBRACE_LT] = ACTIONS(1945),
    [anon_sym_GT_RBRACE] = ACTIONS(1945),
    [anon_sym_begin] = ACTIONS(1943),
    [sym_ocamlyacc_value] = ACTIONS(1945),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1943),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1945),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1943),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1943),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1945),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1945),
    [aux_sym_number_token1] = ACTIONS(1945),
    [aux_sym_signed_number_token1] = ACTIONS(1943),
    [anon_sym_SQUOTE] = ACTIONS(1945),
    [anon_sym_DQUOTE] = ACTIONS(1945),
    [sym_prefix_operator] = ACTIONS(1945),
    [aux_sym_sign_operator_token1] = ACTIONS(1943),
    [sym_hash_operator] = ACTIONS(1945),
    [anon_sym_lsl] = ACTIONS(1943),
    [anon_sym_lsr] = ACTIONS(1943),
    [anon_sym_asr] = ACTIONS(1943),
    [aux_sym_pow_operator_token1] = ACTIONS(1945),
    [anon_sym_mod] = ACTIONS(1943),
    [anon_sym_land] = ACTIONS(1943),
    [anon_sym_lor] = ACTIONS(1943),
    [anon_sym_lxor] = ACTIONS(1943),
    [aux_sym_mult_operator_token1] = ACTIONS(1943),
    [aux_sym_add_operator_token1] = ACTIONS(1943),
    [sym_concat_operator] = ACTIONS(1945),
    [sym_rel_operator] = ACTIONS(1943),
    [anon_sym_AMP_AMP] = ACTIONS(1943),
    [anon_sym_or] = ACTIONS(1943),
    [anon_sym_PIPE_PIPE] = ACTIONS(1943),
    [sym_assign_operator] = ACTIONS(1945),
    [sym__uppercase_identifier] = ACTIONS(1945),
    [aux_sym_tag_token1] = ACTIONS(1945),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(946)] = {
    [sym_attribute] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(1947),
    [anon_sym_and] = ACTIONS(1947),
    [anon_sym_TILDE] = ACTIONS(1947),
    [anon_sym_QMARK] = ACTIONS(1947),
    [anon_sym_LPAREN] = ACTIONS(1949),
    [anon_sym_RPAREN] = ACTIONS(1949),
    [anon_sym_COMMA] = ACTIONS(1949),
    [anon_sym_PIPE] = ACTIONS(1947),
    [anon_sym_COLON2] = ACTIONS(1947),
    [anon_sym_DOT] = ACTIONS(1949),
    [anon_sym_DASH_GT] = ACTIONS(1947),
    [anon_sym_LBRACK] = ACTIONS(1947),
    [anon_sym_RBRACK] = ACTIONS(1949),
    [anon_sym_true] = ACTIONS(1947),
    [anon_sym_false] = ACTIONS(1947),
    [anon_sym_LBRACE] = ACTIONS(1947),
    [anon_sym_SEMI] = ACTIONS(1949),
    [anon_sym_RBRACE] = ACTIONS(1949),
    [anon_sym_constraint] = ACTIONS(1947),
    [anon_sym_val] = ACTIONS(1947),
    [anon_sym_end] = ACTIONS(1947),
    [anon_sym_with] = ACTIONS(1947),
    [anon_sym_COLON_GT] = ACTIONS(1949),
    [anon_sym_object] = ACTIONS(1947),
    [anon_sym_inherit] = ACTIONS(1947),
    [anon_sym_method] = ACTIONS(1947),
    [anon_sym_in] = ACTIONS(1947),
    [anon_sym_as] = ACTIONS(1947),
    [anon_sym_initializer] = ACTIONS(1947),
    [anon_sym_AMP] = ACTIONS(1947),
    [anon_sym_POUND] = ACTIONS(1947),
    [anon_sym_COLON_COLON] = ACTIONS(1949),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1949),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1949),
    [anon_sym_then] = ACTIONS(1947),
    [anon_sym_else] = ACTIONS(1947),
    [anon_sym_do] = ACTIONS(1947),
    [anon_sym_done] = ACTIONS(1947),
    [anon_sym_to] = ACTIONS(1947),
    [anon_sym_downto] = ACTIONS(1947),
    [anon_sym_new] = ACTIONS(1947),
    [anon_sym_LBRACE_LT] = ACTIONS(1949),
    [anon_sym_GT_RBRACE] = ACTIONS(1949),
    [anon_sym_begin] = ACTIONS(1947),
    [sym_ocamlyacc_value] = ACTIONS(1949),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1947),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1949),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1949),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1949),
    [aux_sym_number_token1] = ACTIONS(1949),
    [aux_sym_signed_number_token1] = ACTIONS(1947),
    [anon_sym_SQUOTE] = ACTIONS(1949),
    [anon_sym_DQUOTE] = ACTIONS(1949),
    [sym_prefix_operator] = ACTIONS(1949),
    [aux_sym_sign_operator_token1] = ACTIONS(1947),
    [sym_hash_operator] = ACTIONS(1949),
    [anon_sym_lsl] = ACTIONS(1947),
    [anon_sym_lsr] = ACTIONS(1947),
    [anon_sym_asr] = ACTIONS(1947),
    [aux_sym_pow_operator_token1] = ACTIONS(1949),
    [anon_sym_mod] = ACTIONS(1947),
    [anon_sym_land] = ACTIONS(1947),
    [anon_sym_lor] = ACTIONS(1947),
    [anon_sym_lxor] = ACTIONS(1947),
    [aux_sym_mult_operator_token1] = ACTIONS(1947),
    [aux_sym_add_operator_token1] = ACTIONS(1947),
    [sym_concat_operator] = ACTIONS(1949),
    [sym_rel_operator] = ACTIONS(1947),
    [anon_sym_AMP_AMP] = ACTIONS(1947),
    [anon_sym_or] = ACTIONS(1947),
    [anon_sym_PIPE_PIPE] = ACTIONS(1947),
    [sym_assign_operator] = ACTIONS(1949),
    [sym__uppercase_identifier] = ACTIONS(1949),
    [aux_sym_tag_token1] = ACTIONS(1949),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(947)] = {
    [sym_attribute] = STATE(947),
    [sym__lowercase_identifier] = ACTIONS(1951),
    [anon_sym_and] = ACTIONS(1951),
    [anon_sym_TILDE] = ACTIONS(1951),
    [anon_sym_QMARK] = ACTIONS(1951),
    [anon_sym_LPAREN] = ACTIONS(1953),
    [anon_sym_RPAREN] = ACTIONS(1953),
    [anon_sym_COMMA] = ACTIONS(1953),
    [anon_sym_PIPE] = ACTIONS(1951),
    [anon_sym_COLON2] = ACTIONS(1951),
    [anon_sym_DOT] = ACTIONS(1953),
    [anon_sym_DASH_GT] = ACTIONS(1951),
    [anon_sym_LBRACK] = ACTIONS(1951),
    [anon_sym_RBRACK] = ACTIONS(1953),
    [anon_sym_true] = ACTIONS(1951),
    [anon_sym_false] = ACTIONS(1951),
    [anon_sym_LBRACE] = ACTIONS(1951),
    [anon_sym_SEMI] = ACTIONS(1953),
    [anon_sym_RBRACE] = ACTIONS(1953),
    [anon_sym_constraint] = ACTIONS(1951),
    [anon_sym_val] = ACTIONS(1951),
    [anon_sym_end] = ACTIONS(1951),
    [anon_sym_with] = ACTIONS(1951),
    [anon_sym_COLON_GT] = ACTIONS(1953),
    [anon_sym_object] = ACTIONS(1951),
    [anon_sym_inherit] = ACTIONS(1951),
    [anon_sym_method] = ACTIONS(1951),
    [anon_sym_in] = ACTIONS(1951),
    [anon_sym_as] = ACTIONS(1951),
    [anon_sym_initializer] = ACTIONS(1951),
    [anon_sym_AMP] = ACTIONS(1951),
    [anon_sym_POUND] = ACTIONS(1951),
    [anon_sym_COLON_COLON] = ACTIONS(1953),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1953),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1953),
    [anon_sym_then] = ACTIONS(1951),
    [anon_sym_else] = ACTIONS(1951),
    [anon_sym_do] = ACTIONS(1951),
    [anon_sym_done] = ACTIONS(1951),
    [anon_sym_to] = ACTIONS(1951),
    [anon_sym_downto] = ACTIONS(1951),
    [anon_sym_new] = ACTIONS(1951),
    [anon_sym_LBRACE_LT] = ACTIONS(1953),
    [anon_sym_GT_RBRACE] = ACTIONS(1953),
    [anon_sym_begin] = ACTIONS(1951),
    [sym_ocamlyacc_value] = ACTIONS(1953),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1951),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1953),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1951),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1951),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1953),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1953),
    [aux_sym_number_token1] = ACTIONS(1953),
    [aux_sym_signed_number_token1] = ACTIONS(1951),
    [anon_sym_SQUOTE] = ACTIONS(1953),
    [anon_sym_DQUOTE] = ACTIONS(1953),
    [sym_prefix_operator] = ACTIONS(1953),
    [aux_sym_sign_operator_token1] = ACTIONS(1951),
    [sym_hash_operator] = ACTIONS(1953),
    [anon_sym_lsl] = ACTIONS(1951),
    [anon_sym_lsr] = ACTIONS(1951),
    [anon_sym_asr] = ACTIONS(1951),
    [aux_sym_pow_operator_token1] = ACTIONS(1953),
    [anon_sym_mod] = ACTIONS(1951),
    [anon_sym_land] = ACTIONS(1951),
    [anon_sym_lor] = ACTIONS(1951),
    [anon_sym_lxor] = ACTIONS(1951),
    [aux_sym_mult_operator_token1] = ACTIONS(1951),
    [aux_sym_add_operator_token1] = ACTIONS(1951),
    [sym_concat_operator] = ACTIONS(1953),
    [sym_rel_operator] = ACTIONS(1951),
    [anon_sym_AMP_AMP] = ACTIONS(1951),
    [anon_sym_or] = ACTIONS(1951),
    [anon_sym_PIPE_PIPE] = ACTIONS(1951),
    [sym_assign_operator] = ACTIONS(1953),
    [sym__uppercase_identifier] = ACTIONS(1953),
    [aux_sym_tag_token1] = ACTIONS(1953),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(948)] = {
    [sym_attribute] = STATE(948),
    [sym__lowercase_identifier] = ACTIONS(1955),
    [anon_sym_and] = ACTIONS(1955),
    [anon_sym_TILDE] = ACTIONS(1955),
    [anon_sym_QMARK] = ACTIONS(1955),
    [anon_sym_LPAREN] = ACTIONS(1957),
    [anon_sym_RPAREN] = ACTIONS(1957),
    [anon_sym_COMMA] = ACTIONS(1957),
    [anon_sym_PIPE] = ACTIONS(1955),
    [anon_sym_COLON2] = ACTIONS(1955),
    [anon_sym_DOT] = ACTIONS(1957),
    [anon_sym_DASH_GT] = ACTIONS(1955),
    [anon_sym_LBRACK] = ACTIONS(1955),
    [anon_sym_RBRACK] = ACTIONS(1957),
    [anon_sym_true] = ACTIONS(1955),
    [anon_sym_false] = ACTIONS(1955),
    [anon_sym_LBRACE] = ACTIONS(1955),
    [anon_sym_SEMI] = ACTIONS(1957),
    [anon_sym_RBRACE] = ACTIONS(1957),
    [anon_sym_constraint] = ACTIONS(1955),
    [anon_sym_val] = ACTIONS(1955),
    [anon_sym_end] = ACTIONS(1955),
    [anon_sym_with] = ACTIONS(1955),
    [anon_sym_COLON_GT] = ACTIONS(1957),
    [anon_sym_object] = ACTIONS(1955),
    [anon_sym_inherit] = ACTIONS(1955),
    [anon_sym_method] = ACTIONS(1955),
    [anon_sym_in] = ACTIONS(1955),
    [anon_sym_as] = ACTIONS(1955),
    [anon_sym_initializer] = ACTIONS(1955),
    [anon_sym_AMP] = ACTIONS(1955),
    [anon_sym_POUND] = ACTIONS(1955),
    [anon_sym_COLON_COLON] = ACTIONS(1957),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1957),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1957),
    [anon_sym_then] = ACTIONS(1955),
    [anon_sym_else] = ACTIONS(1955),
    [anon_sym_do] = ACTIONS(1955),
    [anon_sym_done] = ACTIONS(1955),
    [anon_sym_to] = ACTIONS(1955),
    [anon_sym_downto] = ACTIONS(1955),
    [anon_sym_new] = ACTIONS(1955),
    [anon_sym_LBRACE_LT] = ACTIONS(1957),
    [anon_sym_GT_RBRACE] = ACTIONS(1957),
    [anon_sym_begin] = ACTIONS(1955),
    [sym_ocamlyacc_value] = ACTIONS(1957),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1955),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1957),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1955),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1955),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1957),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1957),
    [aux_sym_number_token1] = ACTIONS(1957),
    [aux_sym_signed_number_token1] = ACTIONS(1955),
    [anon_sym_SQUOTE] = ACTIONS(1957),
    [anon_sym_DQUOTE] = ACTIONS(1957),
    [sym_prefix_operator] = ACTIONS(1957),
    [aux_sym_sign_operator_token1] = ACTIONS(1955),
    [sym_hash_operator] = ACTIONS(1957),
    [anon_sym_lsl] = ACTIONS(1955),
    [anon_sym_lsr] = ACTIONS(1955),
    [anon_sym_asr] = ACTIONS(1955),
    [aux_sym_pow_operator_token1] = ACTIONS(1957),
    [anon_sym_mod] = ACTIONS(1955),
    [anon_sym_land] = ACTIONS(1955),
    [anon_sym_lor] = ACTIONS(1955),
    [anon_sym_lxor] = ACTIONS(1955),
    [aux_sym_mult_operator_token1] = ACTIONS(1955),
    [aux_sym_add_operator_token1] = ACTIONS(1955),
    [sym_concat_operator] = ACTIONS(1957),
    [sym_rel_operator] = ACTIONS(1955),
    [anon_sym_AMP_AMP] = ACTIONS(1955),
    [anon_sym_or] = ACTIONS(1955),
    [anon_sym_PIPE_PIPE] = ACTIONS(1955),
    [sym_assign_operator] = ACTIONS(1957),
    [sym__uppercase_identifier] = ACTIONS(1957),
    [aux_sym_tag_token1] = ACTIONS(1957),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(949)] = {
    [sym_attribute] = STATE(949),
    [sym__lowercase_identifier] = ACTIONS(1959),
    [anon_sym_and] = ACTIONS(1959),
    [anon_sym_TILDE] = ACTIONS(1959),
    [anon_sym_QMARK] = ACTIONS(1959),
    [anon_sym_LPAREN] = ACTIONS(1961),
    [anon_sym_RPAREN] = ACTIONS(1961),
    [anon_sym_COMMA] = ACTIONS(1961),
    [anon_sym_PIPE] = ACTIONS(1959),
    [anon_sym_COLON2] = ACTIONS(1959),
    [anon_sym_DOT] = ACTIONS(1961),
    [anon_sym_DASH_GT] = ACTIONS(1959),
    [anon_sym_LBRACK] = ACTIONS(1959),
    [anon_sym_RBRACK] = ACTIONS(1961),
    [anon_sym_true] = ACTIONS(1959),
    [anon_sym_false] = ACTIONS(1959),
    [anon_sym_LBRACE] = ACTIONS(1959),
    [anon_sym_SEMI] = ACTIONS(1961),
    [anon_sym_RBRACE] = ACTIONS(1961),
    [anon_sym_constraint] = ACTIONS(1959),
    [anon_sym_val] = ACTIONS(1959),
    [anon_sym_end] = ACTIONS(1959),
    [anon_sym_with] = ACTIONS(1959),
    [anon_sym_COLON_GT] = ACTIONS(1961),
    [anon_sym_object] = ACTIONS(1959),
    [anon_sym_inherit] = ACTIONS(1959),
    [anon_sym_method] = ACTIONS(1959),
    [anon_sym_in] = ACTIONS(1959),
    [anon_sym_as] = ACTIONS(1959),
    [anon_sym_initializer] = ACTIONS(1959),
    [anon_sym_AMP] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1959),
    [anon_sym_COLON_COLON] = ACTIONS(1961),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1961),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1961),
    [anon_sym_then] = ACTIONS(1959),
    [anon_sym_else] = ACTIONS(1959),
    [anon_sym_do] = ACTIONS(1959),
    [anon_sym_done] = ACTIONS(1959),
    [anon_sym_to] = ACTIONS(1959),
    [anon_sym_downto] = ACTIONS(1959),
    [anon_sym_new] = ACTIONS(1959),
    [anon_sym_LBRACE_LT] = ACTIONS(1961),
    [anon_sym_GT_RBRACE] = ACTIONS(1961),
    [anon_sym_begin] = ACTIONS(1959),
    [sym_ocamlyacc_value] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1959),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1961),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1959),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1959),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1961),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1961),
    [aux_sym_number_token1] = ACTIONS(1961),
    [aux_sym_signed_number_token1] = ACTIONS(1959),
    [anon_sym_SQUOTE] = ACTIONS(1961),
    [anon_sym_DQUOTE] = ACTIONS(1961),
    [sym_prefix_operator] = ACTIONS(1961),
    [aux_sym_sign_operator_token1] = ACTIONS(1959),
    [sym_hash_operator] = ACTIONS(1961),
    [anon_sym_lsl] = ACTIONS(1959),
    [anon_sym_lsr] = ACTIONS(1959),
    [anon_sym_asr] = ACTIONS(1959),
    [aux_sym_pow_operator_token1] = ACTIONS(1961),
    [anon_sym_mod] = ACTIONS(1959),
    [anon_sym_land] = ACTIONS(1959),
    [anon_sym_lor] = ACTIONS(1959),
    [anon_sym_lxor] = ACTIONS(1959),
    [aux_sym_mult_operator_token1] = ACTIONS(1959),
    [aux_sym_add_operator_token1] = ACTIONS(1959),
    [sym_concat_operator] = ACTIONS(1961),
    [sym_rel_operator] = ACTIONS(1959),
    [anon_sym_AMP_AMP] = ACTIONS(1959),
    [anon_sym_or] = ACTIONS(1959),
    [anon_sym_PIPE_PIPE] = ACTIONS(1959),
    [sym_assign_operator] = ACTIONS(1961),
    [sym__uppercase_identifier] = ACTIONS(1961),
    [aux_sym_tag_token1] = ACTIONS(1961),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(950)] = {
    [sym_attribute] = STATE(950),
    [sym__lowercase_identifier] = ACTIONS(1963),
    [anon_sym_and] = ACTIONS(1963),
    [anon_sym_TILDE] = ACTIONS(1963),
    [anon_sym_QMARK] = ACTIONS(1963),
    [anon_sym_LPAREN] = ACTIONS(1965),
    [anon_sym_RPAREN] = ACTIONS(1965),
    [anon_sym_COMMA] = ACTIONS(1965),
    [anon_sym_PIPE] = ACTIONS(1963),
    [anon_sym_COLON2] = ACTIONS(1963),
    [anon_sym_DOT] = ACTIONS(1965),
    [anon_sym_DASH_GT] = ACTIONS(1963),
    [anon_sym_LBRACK] = ACTIONS(1963),
    [anon_sym_RBRACK] = ACTIONS(1965),
    [anon_sym_true] = ACTIONS(1963),
    [anon_sym_false] = ACTIONS(1963),
    [anon_sym_LBRACE] = ACTIONS(1963),
    [anon_sym_SEMI] = ACTIONS(1965),
    [anon_sym_RBRACE] = ACTIONS(1965),
    [anon_sym_constraint] = ACTIONS(1963),
    [anon_sym_val] = ACTIONS(1963),
    [anon_sym_end] = ACTIONS(1963),
    [anon_sym_with] = ACTIONS(1963),
    [anon_sym_COLON_GT] = ACTIONS(1965),
    [anon_sym_object] = ACTIONS(1963),
    [anon_sym_inherit] = ACTIONS(1963),
    [anon_sym_method] = ACTIONS(1963),
    [anon_sym_in] = ACTIONS(1963),
    [anon_sym_as] = ACTIONS(1963),
    [anon_sym_initializer] = ACTIONS(1963),
    [anon_sym_AMP] = ACTIONS(1963),
    [anon_sym_POUND] = ACTIONS(1963),
    [anon_sym_COLON_COLON] = ACTIONS(1965),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1965),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1965),
    [anon_sym_then] = ACTIONS(1963),
    [anon_sym_else] = ACTIONS(1963),
    [anon_sym_do] = ACTIONS(1963),
    [anon_sym_done] = ACTIONS(1963),
    [anon_sym_to] = ACTIONS(1963),
    [anon_sym_downto] = ACTIONS(1963),
    [anon_sym_new] = ACTIONS(1963),
    [anon_sym_LBRACE_LT] = ACTIONS(1965),
    [anon_sym_GT_RBRACE] = ACTIONS(1965),
    [anon_sym_begin] = ACTIONS(1963),
    [sym_ocamlyacc_value] = ACTIONS(1965),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1963),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1965),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1965),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1965),
    [aux_sym_signed_number_token1] = ACTIONS(1963),
    [anon_sym_SQUOTE] = ACTIONS(1965),
    [anon_sym_DQUOTE] = ACTIONS(1965),
    [sym_prefix_operator] = ACTIONS(1965),
    [aux_sym_sign_operator_token1] = ACTIONS(1963),
    [sym_hash_operator] = ACTIONS(1965),
    [anon_sym_lsl] = ACTIONS(1963),
    [anon_sym_lsr] = ACTIONS(1963),
    [anon_sym_asr] = ACTIONS(1963),
    [aux_sym_pow_operator_token1] = ACTIONS(1965),
    [anon_sym_mod] = ACTIONS(1963),
    [anon_sym_land] = ACTIONS(1963),
    [anon_sym_lor] = ACTIONS(1963),
    [anon_sym_lxor] = ACTIONS(1963),
    [aux_sym_mult_operator_token1] = ACTIONS(1963),
    [aux_sym_add_operator_token1] = ACTIONS(1963),
    [sym_concat_operator] = ACTIONS(1965),
    [sym_rel_operator] = ACTIONS(1963),
    [anon_sym_AMP_AMP] = ACTIONS(1963),
    [anon_sym_or] = ACTIONS(1963),
    [anon_sym_PIPE_PIPE] = ACTIONS(1963),
    [sym_assign_operator] = ACTIONS(1965),
    [sym__uppercase_identifier] = ACTIONS(1965),
    [aux_sym_tag_token1] = ACTIONS(1965),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(951)] = {
    [sym_attribute] = STATE(951),
    [sym__lowercase_identifier] = ACTIONS(1967),
    [anon_sym_and] = ACTIONS(1967),
    [anon_sym_TILDE] = ACTIONS(1967),
    [anon_sym_QMARK] = ACTIONS(1967),
    [anon_sym_LPAREN] = ACTIONS(1969),
    [anon_sym_RPAREN] = ACTIONS(1969),
    [anon_sym_COMMA] = ACTIONS(1969),
    [anon_sym_PIPE] = ACTIONS(1967),
    [anon_sym_COLON2] = ACTIONS(1967),
    [anon_sym_DOT] = ACTIONS(1969),
    [anon_sym_DASH_GT] = ACTIONS(1967),
    [anon_sym_LBRACK] = ACTIONS(1967),
    [anon_sym_RBRACK] = ACTIONS(1969),
    [anon_sym_true] = ACTIONS(1967),
    [anon_sym_false] = ACTIONS(1967),
    [anon_sym_LBRACE] = ACTIONS(1967),
    [anon_sym_SEMI] = ACTIONS(1969),
    [anon_sym_RBRACE] = ACTIONS(1969),
    [anon_sym_constraint] = ACTIONS(1967),
    [anon_sym_val] = ACTIONS(1967),
    [anon_sym_end] = ACTIONS(1967),
    [anon_sym_with] = ACTIONS(1967),
    [anon_sym_COLON_GT] = ACTIONS(1969),
    [anon_sym_object] = ACTIONS(1967),
    [anon_sym_inherit] = ACTIONS(1967),
    [anon_sym_method] = ACTIONS(1967),
    [anon_sym_in] = ACTIONS(1967),
    [anon_sym_as] = ACTIONS(1967),
    [anon_sym_initializer] = ACTIONS(1967),
    [anon_sym_AMP] = ACTIONS(1967),
    [anon_sym_POUND] = ACTIONS(1967),
    [anon_sym_COLON_COLON] = ACTIONS(1969),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1969),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1969),
    [anon_sym_then] = ACTIONS(1967),
    [anon_sym_else] = ACTIONS(1967),
    [anon_sym_do] = ACTIONS(1967),
    [anon_sym_done] = ACTIONS(1967),
    [anon_sym_to] = ACTIONS(1967),
    [anon_sym_downto] = ACTIONS(1967),
    [anon_sym_new] = ACTIONS(1967),
    [anon_sym_LBRACE_LT] = ACTIONS(1969),
    [anon_sym_GT_RBRACE] = ACTIONS(1969),
    [anon_sym_begin] = ACTIONS(1967),
    [sym_ocamlyacc_value] = ACTIONS(1969),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1967),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1969),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1969),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1969),
    [aux_sym_number_token1] = ACTIONS(1969),
    [aux_sym_signed_number_token1] = ACTIONS(1967),
    [anon_sym_SQUOTE] = ACTIONS(1969),
    [anon_sym_DQUOTE] = ACTIONS(1969),
    [sym_prefix_operator] = ACTIONS(1969),
    [aux_sym_sign_operator_token1] = ACTIONS(1967),
    [sym_hash_operator] = ACTIONS(1969),
    [anon_sym_lsl] = ACTIONS(1967),
    [anon_sym_lsr] = ACTIONS(1967),
    [anon_sym_asr] = ACTIONS(1967),
    [aux_sym_pow_operator_token1] = ACTIONS(1969),
    [anon_sym_mod] = ACTIONS(1967),
    [anon_sym_land] = ACTIONS(1967),
    [anon_sym_lor] = ACTIONS(1967),
    [anon_sym_lxor] = ACTIONS(1967),
    [aux_sym_mult_operator_token1] = ACTIONS(1967),
    [aux_sym_add_operator_token1] = ACTIONS(1967),
    [sym_concat_operator] = ACTIONS(1969),
    [sym_rel_operator] = ACTIONS(1967),
    [anon_sym_AMP_AMP] = ACTIONS(1967),
    [anon_sym_or] = ACTIONS(1967),
    [anon_sym_PIPE_PIPE] = ACTIONS(1967),
    [sym_assign_operator] = ACTIONS(1969),
    [sym__uppercase_identifier] = ACTIONS(1969),
    [aux_sym_tag_token1] = ACTIONS(1969),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(952)] = {
    [sym_attribute] = STATE(952),
    [sym__lowercase_identifier] = ACTIONS(1971),
    [anon_sym_and] = ACTIONS(1971),
    [anon_sym_TILDE] = ACTIONS(1971),
    [anon_sym_QMARK] = ACTIONS(1971),
    [anon_sym_LPAREN] = ACTIONS(1973),
    [anon_sym_RPAREN] = ACTIONS(1973),
    [anon_sym_COMMA] = ACTIONS(1973),
    [anon_sym_PIPE] = ACTIONS(1971),
    [anon_sym_COLON2] = ACTIONS(1971),
    [anon_sym_DOT] = ACTIONS(919),
    [anon_sym_DASH_GT] = ACTIONS(1971),
    [anon_sym_LBRACK] = ACTIONS(1971),
    [anon_sym_RBRACK] = ACTIONS(1973),
    [anon_sym_true] = ACTIONS(1971),
    [anon_sym_false] = ACTIONS(1971),
    [anon_sym_LBRACE] = ACTIONS(1971),
    [anon_sym_SEMI] = ACTIONS(1973),
    [anon_sym_RBRACE] = ACTIONS(1973),
    [anon_sym_constraint] = ACTIONS(1971),
    [anon_sym_val] = ACTIONS(1971),
    [anon_sym_end] = ACTIONS(1971),
    [anon_sym_with] = ACTIONS(1971),
    [anon_sym_COLON_GT] = ACTIONS(1973),
    [anon_sym_object] = ACTIONS(1971),
    [anon_sym_inherit] = ACTIONS(1971),
    [anon_sym_method] = ACTIONS(1971),
    [anon_sym_in] = ACTIONS(1971),
    [anon_sym_as] = ACTIONS(1971),
    [anon_sym_initializer] = ACTIONS(1971),
    [anon_sym_AMP] = ACTIONS(1971),
    [anon_sym_POUND] = ACTIONS(1971),
    [anon_sym_COLON_COLON] = ACTIONS(1973),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1973),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1973),
    [anon_sym_then] = ACTIONS(1971),
    [anon_sym_else] = ACTIONS(1971),
    [anon_sym_do] = ACTIONS(1971),
    [anon_sym_done] = ACTIONS(1971),
    [anon_sym_to] = ACTIONS(1971),
    [anon_sym_downto] = ACTIONS(1971),
    [anon_sym_new] = ACTIONS(1971),
    [anon_sym_LBRACE_LT] = ACTIONS(1973),
    [anon_sym_GT_RBRACE] = ACTIONS(1973),
    [anon_sym_begin] = ACTIONS(1971),
    [sym_ocamlyacc_value] = ACTIONS(1973),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1971),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1973),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1973),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1973),
    [aux_sym_number_token1] = ACTIONS(1973),
    [aux_sym_signed_number_token1] = ACTIONS(1971),
    [anon_sym_SQUOTE] = ACTIONS(1973),
    [anon_sym_DQUOTE] = ACTIONS(1973),
    [sym_prefix_operator] = ACTIONS(1973),
    [aux_sym_sign_operator_token1] = ACTIONS(1971),
    [sym_hash_operator] = ACTIONS(1973),
    [anon_sym_lsl] = ACTIONS(1971),
    [anon_sym_lsr] = ACTIONS(1971),
    [anon_sym_asr] = ACTIONS(1971),
    [aux_sym_pow_operator_token1] = ACTIONS(1973),
    [anon_sym_mod] = ACTIONS(1971),
    [anon_sym_land] = ACTIONS(1971),
    [anon_sym_lor] = ACTIONS(1971),
    [anon_sym_lxor] = ACTIONS(1971),
    [aux_sym_mult_operator_token1] = ACTIONS(1971),
    [aux_sym_add_operator_token1] = ACTIONS(1971),
    [sym_concat_operator] = ACTIONS(1973),
    [sym_rel_operator] = ACTIONS(1971),
    [anon_sym_AMP_AMP] = ACTIONS(1971),
    [anon_sym_or] = ACTIONS(1971),
    [anon_sym_PIPE_PIPE] = ACTIONS(1971),
    [sym_assign_operator] = ACTIONS(1973),
    [sym__uppercase_identifier] = ACTIONS(1973),
    [aux_sym_tag_token1] = ACTIONS(1973),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(953)] = {
    [sym_attribute] = STATE(953),
    [sym__lowercase_identifier] = ACTIONS(1975),
    [anon_sym_and] = ACTIONS(1975),
    [anon_sym_TILDE] = ACTIONS(1975),
    [anon_sym_QMARK] = ACTIONS(1975),
    [anon_sym_LPAREN] = ACTIONS(1977),
    [anon_sym_RPAREN] = ACTIONS(1977),
    [anon_sym_COMMA] = ACTIONS(1977),
    [anon_sym_PIPE] = ACTIONS(1975),
    [anon_sym_COLON2] = ACTIONS(1975),
    [anon_sym_DOT] = ACTIONS(1977),
    [anon_sym_DASH_GT] = ACTIONS(1975),
    [anon_sym_LBRACK] = ACTIONS(1975),
    [anon_sym_RBRACK] = ACTIONS(1977),
    [anon_sym_true] = ACTIONS(1975),
    [anon_sym_false] = ACTIONS(1975),
    [anon_sym_LBRACE] = ACTIONS(1975),
    [anon_sym_SEMI] = ACTIONS(1977),
    [anon_sym_RBRACE] = ACTIONS(1977),
    [anon_sym_constraint] = ACTIONS(1975),
    [anon_sym_val] = ACTIONS(1975),
    [anon_sym_end] = ACTIONS(1975),
    [anon_sym_with] = ACTIONS(1975),
    [anon_sym_COLON_GT] = ACTIONS(1977),
    [anon_sym_object] = ACTIONS(1975),
    [anon_sym_inherit] = ACTIONS(1975),
    [anon_sym_method] = ACTIONS(1975),
    [anon_sym_in] = ACTIONS(1975),
    [anon_sym_as] = ACTIONS(1975),
    [anon_sym_initializer] = ACTIONS(1975),
    [anon_sym_AMP] = ACTIONS(1975),
    [anon_sym_POUND] = ACTIONS(1975),
    [anon_sym_COLON_COLON] = ACTIONS(1977),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1977),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1977),
    [anon_sym_then] = ACTIONS(1975),
    [anon_sym_else] = ACTIONS(1975),
    [anon_sym_do] = ACTIONS(1975),
    [anon_sym_done] = ACTIONS(1975),
    [anon_sym_to] = ACTIONS(1975),
    [anon_sym_downto] = ACTIONS(1975),
    [anon_sym_new] = ACTIONS(1975),
    [anon_sym_LBRACE_LT] = ACTIONS(1977),
    [anon_sym_GT_RBRACE] = ACTIONS(1977),
    [anon_sym_begin] = ACTIONS(1975),
    [sym_ocamlyacc_value] = ACTIONS(1977),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1975),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1977),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1975),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1975),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1977),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1977),
    [aux_sym_number_token1] = ACTIONS(1977),
    [aux_sym_signed_number_token1] = ACTIONS(1975),
    [anon_sym_SQUOTE] = ACTIONS(1977),
    [anon_sym_DQUOTE] = ACTIONS(1977),
    [sym_prefix_operator] = ACTIONS(1977),
    [aux_sym_sign_operator_token1] = ACTIONS(1975),
    [sym_hash_operator] = ACTIONS(1977),
    [anon_sym_lsl] = ACTIONS(1975),
    [anon_sym_lsr] = ACTIONS(1975),
    [anon_sym_asr] = ACTIONS(1975),
    [aux_sym_pow_operator_token1] = ACTIONS(1977),
    [anon_sym_mod] = ACTIONS(1975),
    [anon_sym_land] = ACTIONS(1975),
    [anon_sym_lor] = ACTIONS(1975),
    [anon_sym_lxor] = ACTIONS(1975),
    [aux_sym_mult_operator_token1] = ACTIONS(1975),
    [aux_sym_add_operator_token1] = ACTIONS(1975),
    [sym_concat_operator] = ACTIONS(1977),
    [sym_rel_operator] = ACTIONS(1975),
    [anon_sym_AMP_AMP] = ACTIONS(1975),
    [anon_sym_or] = ACTIONS(1975),
    [anon_sym_PIPE_PIPE] = ACTIONS(1975),
    [sym_assign_operator] = ACTIONS(1977),
    [sym__uppercase_identifier] = ACTIONS(1977),
    [aux_sym_tag_token1] = ACTIONS(1977),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(954)] = {
    [sym_attribute] = STATE(954),
    [sym__lowercase_identifier] = ACTIONS(1979),
    [anon_sym_and] = ACTIONS(1979),
    [anon_sym_TILDE] = ACTIONS(1979),
    [anon_sym_QMARK] = ACTIONS(1979),
    [anon_sym_LPAREN] = ACTIONS(1981),
    [anon_sym_RPAREN] = ACTIONS(1981),
    [anon_sym_COMMA] = ACTIONS(1981),
    [anon_sym_PIPE] = ACTIONS(1979),
    [anon_sym_COLON2] = ACTIONS(1979),
    [anon_sym_DOT] = ACTIONS(1981),
    [anon_sym_DASH_GT] = ACTIONS(1979),
    [anon_sym_LBRACK] = ACTIONS(1979),
    [anon_sym_RBRACK] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(1979),
    [anon_sym_false] = ACTIONS(1979),
    [anon_sym_LBRACE] = ACTIONS(1979),
    [anon_sym_SEMI] = ACTIONS(1981),
    [anon_sym_RBRACE] = ACTIONS(1981),
    [anon_sym_constraint] = ACTIONS(1979),
    [anon_sym_val] = ACTIONS(1979),
    [anon_sym_end] = ACTIONS(1979),
    [anon_sym_with] = ACTIONS(1979),
    [anon_sym_COLON_GT] = ACTIONS(1981),
    [anon_sym_object] = ACTIONS(1979),
    [anon_sym_inherit] = ACTIONS(1979),
    [anon_sym_method] = ACTIONS(1979),
    [anon_sym_in] = ACTIONS(1979),
    [anon_sym_as] = ACTIONS(1979),
    [anon_sym_initializer] = ACTIONS(1979),
    [anon_sym_AMP] = ACTIONS(1979),
    [anon_sym_POUND] = ACTIONS(1979),
    [anon_sym_COLON_COLON] = ACTIONS(1981),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1981),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1981),
    [anon_sym_then] = ACTIONS(1979),
    [anon_sym_else] = ACTIONS(1979),
    [anon_sym_do] = ACTIONS(1979),
    [anon_sym_done] = ACTIONS(1979),
    [anon_sym_to] = ACTIONS(1979),
    [anon_sym_downto] = ACTIONS(1979),
    [anon_sym_new] = ACTIONS(1979),
    [anon_sym_LBRACE_LT] = ACTIONS(1981),
    [anon_sym_GT_RBRACE] = ACTIONS(1981),
    [anon_sym_begin] = ACTIONS(1979),
    [sym_ocamlyacc_value] = ACTIONS(1981),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1979),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1981),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1979),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1979),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1981),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1981),
    [aux_sym_number_token1] = ACTIONS(1981),
    [aux_sym_signed_number_token1] = ACTIONS(1979),
    [anon_sym_SQUOTE] = ACTIONS(1981),
    [anon_sym_DQUOTE] = ACTIONS(1981),
    [sym_prefix_operator] = ACTIONS(1981),
    [aux_sym_sign_operator_token1] = ACTIONS(1979),
    [sym_hash_operator] = ACTIONS(1981),
    [anon_sym_lsl] = ACTIONS(1979),
    [anon_sym_lsr] = ACTIONS(1979),
    [anon_sym_asr] = ACTIONS(1979),
    [aux_sym_pow_operator_token1] = ACTIONS(1981),
    [anon_sym_mod] = ACTIONS(1979),
    [anon_sym_land] = ACTIONS(1979),
    [anon_sym_lor] = ACTIONS(1979),
    [anon_sym_lxor] = ACTIONS(1979),
    [aux_sym_mult_operator_token1] = ACTIONS(1979),
    [aux_sym_add_operator_token1] = ACTIONS(1979),
    [sym_concat_operator] = ACTIONS(1981),
    [sym_rel_operator] = ACTIONS(1979),
    [anon_sym_AMP_AMP] = ACTIONS(1979),
    [anon_sym_or] = ACTIONS(1979),
    [anon_sym_PIPE_PIPE] = ACTIONS(1979),
    [sym_assign_operator] = ACTIONS(1981),
    [sym__uppercase_identifier] = ACTIONS(1981),
    [aux_sym_tag_token1] = ACTIONS(1981),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(955)] = {
    [sym_attribute] = STATE(955),
    [sym__lowercase_identifier] = ACTIONS(1983),
    [anon_sym_and] = ACTIONS(1983),
    [anon_sym_TILDE] = ACTIONS(1983),
    [anon_sym_QMARK] = ACTIONS(1983),
    [anon_sym_LPAREN] = ACTIONS(1985),
    [anon_sym_RPAREN] = ACTIONS(1985),
    [anon_sym_COMMA] = ACTIONS(1985),
    [anon_sym_PIPE] = ACTIONS(1983),
    [anon_sym_COLON2] = ACTIONS(1983),
    [anon_sym_DOT] = ACTIONS(1985),
    [anon_sym_DASH_GT] = ACTIONS(1983),
    [anon_sym_LBRACK] = ACTIONS(1983),
    [anon_sym_RBRACK] = ACTIONS(1985),
    [anon_sym_true] = ACTIONS(1983),
    [anon_sym_false] = ACTIONS(1983),
    [anon_sym_LBRACE] = ACTIONS(1983),
    [anon_sym_SEMI] = ACTIONS(1985),
    [anon_sym_RBRACE] = ACTIONS(1985),
    [anon_sym_constraint] = ACTIONS(1983),
    [anon_sym_val] = ACTIONS(1983),
    [anon_sym_end] = ACTIONS(1983),
    [anon_sym_with] = ACTIONS(1983),
    [anon_sym_COLON_GT] = ACTIONS(1985),
    [anon_sym_object] = ACTIONS(1983),
    [anon_sym_inherit] = ACTIONS(1983),
    [anon_sym_method] = ACTIONS(1983),
    [anon_sym_in] = ACTIONS(1983),
    [anon_sym_as] = ACTIONS(1983),
    [anon_sym_initializer] = ACTIONS(1983),
    [anon_sym_AMP] = ACTIONS(1983),
    [anon_sym_POUND] = ACTIONS(1983),
    [anon_sym_COLON_COLON] = ACTIONS(1985),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1985),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1985),
    [anon_sym_then] = ACTIONS(1983),
    [anon_sym_else] = ACTIONS(1983),
    [anon_sym_do] = ACTIONS(1983),
    [anon_sym_done] = ACTIONS(1983),
    [anon_sym_to] = ACTIONS(1983),
    [anon_sym_downto] = ACTIONS(1983),
    [anon_sym_new] = ACTIONS(1983),
    [anon_sym_LBRACE_LT] = ACTIONS(1985),
    [anon_sym_GT_RBRACE] = ACTIONS(1985),
    [anon_sym_begin] = ACTIONS(1983),
    [sym_ocamlyacc_value] = ACTIONS(1985),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1983),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1985),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1983),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1983),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1985),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1985),
    [aux_sym_number_token1] = ACTIONS(1985),
    [aux_sym_signed_number_token1] = ACTIONS(1983),
    [anon_sym_SQUOTE] = ACTIONS(1985),
    [anon_sym_DQUOTE] = ACTIONS(1985),
    [sym_prefix_operator] = ACTIONS(1985),
    [aux_sym_sign_operator_token1] = ACTIONS(1983),
    [sym_hash_operator] = ACTIONS(1985),
    [anon_sym_lsl] = ACTIONS(1983),
    [anon_sym_lsr] = ACTIONS(1983),
    [anon_sym_asr] = ACTIONS(1983),
    [aux_sym_pow_operator_token1] = ACTIONS(1985),
    [anon_sym_mod] = ACTIONS(1983),
    [anon_sym_land] = ACTIONS(1983),
    [anon_sym_lor] = ACTIONS(1983),
    [anon_sym_lxor] = ACTIONS(1983),
    [aux_sym_mult_operator_token1] = ACTIONS(1983),
    [aux_sym_add_operator_token1] = ACTIONS(1983),
    [sym_concat_operator] = ACTIONS(1985),
    [sym_rel_operator] = ACTIONS(1983),
    [anon_sym_AMP_AMP] = ACTIONS(1983),
    [anon_sym_or] = ACTIONS(1983),
    [anon_sym_PIPE_PIPE] = ACTIONS(1983),
    [sym_assign_operator] = ACTIONS(1985),
    [sym__uppercase_identifier] = ACTIONS(1985),
    [aux_sym_tag_token1] = ACTIONS(1985),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(956)] = {
    [sym_attribute] = STATE(956),
    [sym__lowercase_identifier] = ACTIONS(1987),
    [anon_sym_and] = ACTIONS(1987),
    [anon_sym_TILDE] = ACTIONS(1987),
    [anon_sym_QMARK] = ACTIONS(1987),
    [anon_sym_LPAREN] = ACTIONS(1989),
    [anon_sym_RPAREN] = ACTIONS(1989),
    [anon_sym_COMMA] = ACTIONS(1989),
    [anon_sym_PIPE] = ACTIONS(1987),
    [anon_sym_COLON2] = ACTIONS(1987),
    [anon_sym_DOT] = ACTIONS(1989),
    [anon_sym_DASH_GT] = ACTIONS(1987),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_RBRACK] = ACTIONS(1989),
    [anon_sym_true] = ACTIONS(1987),
    [anon_sym_false] = ACTIONS(1987),
    [anon_sym_LBRACE] = ACTIONS(1987),
    [anon_sym_SEMI] = ACTIONS(1989),
    [anon_sym_RBRACE] = ACTIONS(1989),
    [anon_sym_constraint] = ACTIONS(1987),
    [anon_sym_val] = ACTIONS(1987),
    [anon_sym_end] = ACTIONS(1987),
    [anon_sym_with] = ACTIONS(1987),
    [anon_sym_COLON_GT] = ACTIONS(1989),
    [anon_sym_object] = ACTIONS(1987),
    [anon_sym_inherit] = ACTIONS(1987),
    [anon_sym_method] = ACTIONS(1987),
    [anon_sym_in] = ACTIONS(1987),
    [anon_sym_as] = ACTIONS(1987),
    [anon_sym_initializer] = ACTIONS(1987),
    [anon_sym_AMP] = ACTIONS(1987),
    [anon_sym_POUND] = ACTIONS(1987),
    [anon_sym_COLON_COLON] = ACTIONS(1989),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1989),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1989),
    [anon_sym_then] = ACTIONS(1987),
    [anon_sym_else] = ACTIONS(1987),
    [anon_sym_do] = ACTIONS(1987),
    [anon_sym_done] = ACTIONS(1987),
    [anon_sym_to] = ACTIONS(1987),
    [anon_sym_downto] = ACTIONS(1987),
    [anon_sym_new] = ACTIONS(1987),
    [anon_sym_LBRACE_LT] = ACTIONS(1989),
    [anon_sym_GT_RBRACE] = ACTIONS(1989),
    [anon_sym_begin] = ACTIONS(1987),
    [sym_ocamlyacc_value] = ACTIONS(1989),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1987),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1989),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1987),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1987),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1989),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1989),
    [aux_sym_number_token1] = ACTIONS(1989),
    [aux_sym_signed_number_token1] = ACTIONS(1987),
    [anon_sym_SQUOTE] = ACTIONS(1989),
    [anon_sym_DQUOTE] = ACTIONS(1989),
    [sym_prefix_operator] = ACTIONS(1989),
    [aux_sym_sign_operator_token1] = ACTIONS(1987),
    [sym_hash_operator] = ACTIONS(1989),
    [anon_sym_lsl] = ACTIONS(1987),
    [anon_sym_lsr] = ACTIONS(1987),
    [anon_sym_asr] = ACTIONS(1987),
    [aux_sym_pow_operator_token1] = ACTIONS(1989),
    [anon_sym_mod] = ACTIONS(1987),
    [anon_sym_land] = ACTIONS(1987),
    [anon_sym_lor] = ACTIONS(1987),
    [anon_sym_lxor] = ACTIONS(1987),
    [aux_sym_mult_operator_token1] = ACTIONS(1987),
    [aux_sym_add_operator_token1] = ACTIONS(1987),
    [sym_concat_operator] = ACTIONS(1989),
    [sym_rel_operator] = ACTIONS(1987),
    [anon_sym_AMP_AMP] = ACTIONS(1987),
    [anon_sym_or] = ACTIONS(1987),
    [anon_sym_PIPE_PIPE] = ACTIONS(1987),
    [sym_assign_operator] = ACTIONS(1989),
    [sym__uppercase_identifier] = ACTIONS(1989),
    [aux_sym_tag_token1] = ACTIONS(1989),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(957)] = {
    [sym_attribute] = STATE(957),
    [sym__lowercase_identifier] = ACTIONS(1991),
    [anon_sym_and] = ACTIONS(1991),
    [anon_sym_TILDE] = ACTIONS(1991),
    [anon_sym_QMARK] = ACTIONS(1991),
    [anon_sym_LPAREN] = ACTIONS(1993),
    [anon_sym_RPAREN] = ACTIONS(1993),
    [anon_sym_COMMA] = ACTIONS(1993),
    [anon_sym_PIPE] = ACTIONS(1991),
    [anon_sym_COLON2] = ACTIONS(1991),
    [anon_sym_DOT] = ACTIONS(1993),
    [anon_sym_DASH_GT] = ACTIONS(1991),
    [anon_sym_LBRACK] = ACTIONS(1991),
    [anon_sym_RBRACK] = ACTIONS(1993),
    [anon_sym_true] = ACTIONS(1991),
    [anon_sym_false] = ACTIONS(1991),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_SEMI] = ACTIONS(1993),
    [anon_sym_RBRACE] = ACTIONS(1993),
    [anon_sym_constraint] = ACTIONS(1991),
    [anon_sym_val] = ACTIONS(1991),
    [anon_sym_end] = ACTIONS(1991),
    [anon_sym_with] = ACTIONS(1991),
    [anon_sym_COLON_GT] = ACTIONS(1993),
    [anon_sym_object] = ACTIONS(1991),
    [anon_sym_inherit] = ACTIONS(1991),
    [anon_sym_method] = ACTIONS(1991),
    [anon_sym_in] = ACTIONS(1991),
    [anon_sym_as] = ACTIONS(1991),
    [anon_sym_initializer] = ACTIONS(1991),
    [anon_sym_AMP] = ACTIONS(1991),
    [anon_sym_POUND] = ACTIONS(1991),
    [anon_sym_COLON_COLON] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1993),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1993),
    [anon_sym_then] = ACTIONS(1991),
    [anon_sym_else] = ACTIONS(1991),
    [anon_sym_do] = ACTIONS(1991),
    [anon_sym_done] = ACTIONS(1991),
    [anon_sym_to] = ACTIONS(1991),
    [anon_sym_downto] = ACTIONS(1991),
    [anon_sym_new] = ACTIONS(1991),
    [anon_sym_LBRACE_LT] = ACTIONS(1993),
    [anon_sym_GT_RBRACE] = ACTIONS(1993),
    [anon_sym_begin] = ACTIONS(1991),
    [sym_ocamlyacc_value] = ACTIONS(1993),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1991),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1993),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1991),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1991),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1993),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1993),
    [aux_sym_number_token1] = ACTIONS(1993),
    [aux_sym_signed_number_token1] = ACTIONS(1991),
    [anon_sym_SQUOTE] = ACTIONS(1993),
    [anon_sym_DQUOTE] = ACTIONS(1993),
    [sym_prefix_operator] = ACTIONS(1993),
    [aux_sym_sign_operator_token1] = ACTIONS(1991),
    [sym_hash_operator] = ACTIONS(1993),
    [anon_sym_lsl] = ACTIONS(1991),
    [anon_sym_lsr] = ACTIONS(1991),
    [anon_sym_asr] = ACTIONS(1991),
    [aux_sym_pow_operator_token1] = ACTIONS(1993),
    [anon_sym_mod] = ACTIONS(1991),
    [anon_sym_land] = ACTIONS(1991),
    [anon_sym_lor] = ACTIONS(1991),
    [anon_sym_lxor] = ACTIONS(1991),
    [aux_sym_mult_operator_token1] = ACTIONS(1991),
    [aux_sym_add_operator_token1] = ACTIONS(1991),
    [sym_concat_operator] = ACTIONS(1993),
    [sym_rel_operator] = ACTIONS(1991),
    [anon_sym_AMP_AMP] = ACTIONS(1991),
    [anon_sym_or] = ACTIONS(1991),
    [anon_sym_PIPE_PIPE] = ACTIONS(1991),
    [sym_assign_operator] = ACTIONS(1993),
    [sym__uppercase_identifier] = ACTIONS(1993),
    [aux_sym_tag_token1] = ACTIONS(1993),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(958)] = {
    [sym_attribute] = STATE(958),
    [sym__lowercase_identifier] = ACTIONS(1995),
    [anon_sym_and] = ACTIONS(1995),
    [anon_sym_TILDE] = ACTIONS(1995),
    [anon_sym_QMARK] = ACTIONS(1995),
    [anon_sym_LPAREN] = ACTIONS(1997),
    [anon_sym_RPAREN] = ACTIONS(1997),
    [anon_sym_COMMA] = ACTIONS(1997),
    [anon_sym_PIPE] = ACTIONS(1995),
    [anon_sym_COLON2] = ACTIONS(1995),
    [anon_sym_DOT] = ACTIONS(1997),
    [anon_sym_DASH_GT] = ACTIONS(1995),
    [anon_sym_LBRACK] = ACTIONS(1995),
    [anon_sym_RBRACK] = ACTIONS(1997),
    [anon_sym_true] = ACTIONS(1995),
    [anon_sym_false] = ACTIONS(1995),
    [anon_sym_LBRACE] = ACTIONS(1995),
    [anon_sym_SEMI] = ACTIONS(1997),
    [anon_sym_RBRACE] = ACTIONS(1997),
    [anon_sym_constraint] = ACTIONS(1995),
    [anon_sym_val] = ACTIONS(1995),
    [anon_sym_end] = ACTIONS(1995),
    [anon_sym_with] = ACTIONS(1995),
    [anon_sym_COLON_GT] = ACTIONS(1997),
    [anon_sym_object] = ACTIONS(1995),
    [anon_sym_inherit] = ACTIONS(1995),
    [anon_sym_method] = ACTIONS(1995),
    [anon_sym_in] = ACTIONS(1995),
    [anon_sym_as] = ACTIONS(1995),
    [anon_sym_initializer] = ACTIONS(1995),
    [anon_sym_AMP] = ACTIONS(1995),
    [anon_sym_POUND] = ACTIONS(1995),
    [anon_sym_COLON_COLON] = ACTIONS(1997),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1997),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1997),
    [anon_sym_then] = ACTIONS(1995),
    [anon_sym_else] = ACTIONS(1995),
    [anon_sym_do] = ACTIONS(1995),
    [anon_sym_done] = ACTIONS(1995),
    [anon_sym_to] = ACTIONS(1995),
    [anon_sym_downto] = ACTIONS(1995),
    [anon_sym_new] = ACTIONS(1995),
    [anon_sym_LBRACE_LT] = ACTIONS(1997),
    [anon_sym_GT_RBRACE] = ACTIONS(1997),
    [anon_sym_begin] = ACTIONS(1995),
    [sym_ocamlyacc_value] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1995),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1997),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1995),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1995),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1997),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1997),
    [aux_sym_number_token1] = ACTIONS(1997),
    [aux_sym_signed_number_token1] = ACTIONS(1995),
    [anon_sym_SQUOTE] = ACTIONS(1997),
    [anon_sym_DQUOTE] = ACTIONS(1997),
    [sym_prefix_operator] = ACTIONS(1997),
    [aux_sym_sign_operator_token1] = ACTIONS(1995),
    [sym_hash_operator] = ACTIONS(1997),
    [anon_sym_lsl] = ACTIONS(1995),
    [anon_sym_lsr] = ACTIONS(1995),
    [anon_sym_asr] = ACTIONS(1995),
    [aux_sym_pow_operator_token1] = ACTIONS(1997),
    [anon_sym_mod] = ACTIONS(1995),
    [anon_sym_land] = ACTIONS(1995),
    [anon_sym_lor] = ACTIONS(1995),
    [anon_sym_lxor] = ACTIONS(1995),
    [aux_sym_mult_operator_token1] = ACTIONS(1995),
    [aux_sym_add_operator_token1] = ACTIONS(1995),
    [sym_concat_operator] = ACTIONS(1997),
    [sym_rel_operator] = ACTIONS(1995),
    [anon_sym_AMP_AMP] = ACTIONS(1995),
    [anon_sym_or] = ACTIONS(1995),
    [anon_sym_PIPE_PIPE] = ACTIONS(1995),
    [sym_assign_operator] = ACTIONS(1997),
    [sym__uppercase_identifier] = ACTIONS(1997),
    [aux_sym_tag_token1] = ACTIONS(1997),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(959)] = {
    [sym_attribute] = STATE(959),
    [sym__lowercase_identifier] = ACTIONS(1999),
    [anon_sym_and] = ACTIONS(1999),
    [anon_sym_TILDE] = ACTIONS(1999),
    [anon_sym_QMARK] = ACTIONS(1999),
    [anon_sym_LPAREN] = ACTIONS(2001),
    [anon_sym_RPAREN] = ACTIONS(2001),
    [anon_sym_COMMA] = ACTIONS(2001),
    [anon_sym_PIPE] = ACTIONS(1999),
    [anon_sym_COLON2] = ACTIONS(1999),
    [anon_sym_DOT] = ACTIONS(2001),
    [anon_sym_DASH_GT] = ACTIONS(1999),
    [anon_sym_LBRACK] = ACTIONS(1999),
    [anon_sym_RBRACK] = ACTIONS(2001),
    [anon_sym_true] = ACTIONS(1999),
    [anon_sym_false] = ACTIONS(1999),
    [anon_sym_LBRACE] = ACTIONS(1999),
    [anon_sym_SEMI] = ACTIONS(2001),
    [anon_sym_RBRACE] = ACTIONS(2001),
    [anon_sym_constraint] = ACTIONS(1999),
    [anon_sym_val] = ACTIONS(1999),
    [anon_sym_end] = ACTIONS(1999),
    [anon_sym_with] = ACTIONS(1999),
    [anon_sym_COLON_GT] = ACTIONS(2001),
    [anon_sym_object] = ACTIONS(1999),
    [anon_sym_inherit] = ACTIONS(1999),
    [anon_sym_method] = ACTIONS(1999),
    [anon_sym_in] = ACTIONS(1999),
    [anon_sym_as] = ACTIONS(1999),
    [anon_sym_initializer] = ACTIONS(1999),
    [anon_sym_AMP] = ACTIONS(1999),
    [anon_sym_POUND] = ACTIONS(1999),
    [anon_sym_COLON_COLON] = ACTIONS(2001),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2001),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2001),
    [anon_sym_then] = ACTIONS(1999),
    [anon_sym_else] = ACTIONS(1999),
    [anon_sym_do] = ACTIONS(1999),
    [anon_sym_done] = ACTIONS(1999),
    [anon_sym_to] = ACTIONS(1999),
    [anon_sym_downto] = ACTIONS(1999),
    [anon_sym_new] = ACTIONS(1999),
    [anon_sym_LBRACE_LT] = ACTIONS(2001),
    [anon_sym_GT_RBRACE] = ACTIONS(2001),
    [anon_sym_begin] = ACTIONS(1999),
    [sym_ocamlyacc_value] = ACTIONS(2001),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2001),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1999),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1999),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2001),
    [aux_sym_number_token1] = ACTIONS(2001),
    [aux_sym_signed_number_token1] = ACTIONS(1999),
    [anon_sym_SQUOTE] = ACTIONS(2001),
    [anon_sym_DQUOTE] = ACTIONS(2001),
    [sym_prefix_operator] = ACTIONS(2001),
    [aux_sym_sign_operator_token1] = ACTIONS(1999),
    [sym_hash_operator] = ACTIONS(2001),
    [anon_sym_lsl] = ACTIONS(1999),
    [anon_sym_lsr] = ACTIONS(1999),
    [anon_sym_asr] = ACTIONS(1999),
    [aux_sym_pow_operator_token1] = ACTIONS(2001),
    [anon_sym_mod] = ACTIONS(1999),
    [anon_sym_land] = ACTIONS(1999),
    [anon_sym_lor] = ACTIONS(1999),
    [anon_sym_lxor] = ACTIONS(1999),
    [aux_sym_mult_operator_token1] = ACTIONS(1999),
    [aux_sym_add_operator_token1] = ACTIONS(1999),
    [sym_concat_operator] = ACTIONS(2001),
    [sym_rel_operator] = ACTIONS(1999),
    [anon_sym_AMP_AMP] = ACTIONS(1999),
    [anon_sym_or] = ACTIONS(1999),
    [anon_sym_PIPE_PIPE] = ACTIONS(1999),
    [sym_assign_operator] = ACTIONS(2001),
    [sym__uppercase_identifier] = ACTIONS(2001),
    [aux_sym_tag_token1] = ACTIONS(2001),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(960)] = {
    [sym_attribute] = STATE(960),
    [sym__lowercase_identifier] = ACTIONS(2003),
    [anon_sym_and] = ACTIONS(2003),
    [anon_sym_TILDE] = ACTIONS(2003),
    [anon_sym_QMARK] = ACTIONS(2003),
    [anon_sym_LPAREN] = ACTIONS(2005),
    [anon_sym_RPAREN] = ACTIONS(2005),
    [anon_sym_COMMA] = ACTIONS(2005),
    [anon_sym_PIPE] = ACTIONS(2003),
    [anon_sym_COLON2] = ACTIONS(2003),
    [anon_sym_DOT] = ACTIONS(2005),
    [anon_sym_DASH_GT] = ACTIONS(2003),
    [anon_sym_LBRACK] = ACTIONS(2003),
    [anon_sym_RBRACK] = ACTIONS(2005),
    [anon_sym_true] = ACTIONS(2003),
    [anon_sym_false] = ACTIONS(2003),
    [anon_sym_LBRACE] = ACTIONS(2003),
    [anon_sym_SEMI] = ACTIONS(2005),
    [anon_sym_RBRACE] = ACTIONS(2005),
    [anon_sym_constraint] = ACTIONS(2003),
    [anon_sym_val] = ACTIONS(2003),
    [anon_sym_end] = ACTIONS(2003),
    [anon_sym_with] = ACTIONS(2003),
    [anon_sym_COLON_GT] = ACTIONS(2005),
    [anon_sym_object] = ACTIONS(2003),
    [anon_sym_inherit] = ACTIONS(2003),
    [anon_sym_method] = ACTIONS(2003),
    [anon_sym_in] = ACTIONS(2003),
    [anon_sym_as] = ACTIONS(2003),
    [anon_sym_initializer] = ACTIONS(2003),
    [anon_sym_AMP] = ACTIONS(2003),
    [anon_sym_POUND] = ACTIONS(2003),
    [anon_sym_COLON_COLON] = ACTIONS(2005),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2005),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2005),
    [anon_sym_then] = ACTIONS(2003),
    [anon_sym_else] = ACTIONS(2003),
    [anon_sym_do] = ACTIONS(2003),
    [anon_sym_done] = ACTIONS(2003),
    [anon_sym_to] = ACTIONS(2003),
    [anon_sym_downto] = ACTIONS(2003),
    [anon_sym_new] = ACTIONS(2003),
    [anon_sym_LBRACE_LT] = ACTIONS(2005),
    [anon_sym_GT_RBRACE] = ACTIONS(2005),
    [anon_sym_begin] = ACTIONS(2003),
    [sym_ocamlyacc_value] = ACTIONS(2005),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2003),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2005),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2005),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2005),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2003),
    [anon_sym_SQUOTE] = ACTIONS(2005),
    [anon_sym_DQUOTE] = ACTIONS(2005),
    [sym_prefix_operator] = ACTIONS(2005),
    [aux_sym_sign_operator_token1] = ACTIONS(2003),
    [sym_hash_operator] = ACTIONS(2005),
    [anon_sym_lsl] = ACTIONS(2003),
    [anon_sym_lsr] = ACTIONS(2003),
    [anon_sym_asr] = ACTIONS(2003),
    [aux_sym_pow_operator_token1] = ACTIONS(2005),
    [anon_sym_mod] = ACTIONS(2003),
    [anon_sym_land] = ACTIONS(2003),
    [anon_sym_lor] = ACTIONS(2003),
    [anon_sym_lxor] = ACTIONS(2003),
    [aux_sym_mult_operator_token1] = ACTIONS(2003),
    [aux_sym_add_operator_token1] = ACTIONS(2003),
    [sym_concat_operator] = ACTIONS(2005),
    [sym_rel_operator] = ACTIONS(2003),
    [anon_sym_AMP_AMP] = ACTIONS(2003),
    [anon_sym_or] = ACTIONS(2003),
    [anon_sym_PIPE_PIPE] = ACTIONS(2003),
    [sym_assign_operator] = ACTIONS(2005),
    [sym__uppercase_identifier] = ACTIONS(2005),
    [aux_sym_tag_token1] = ACTIONS(2005),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(961)] = {
    [sym_attribute] = STATE(961),
    [sym__lowercase_identifier] = ACTIONS(2007),
    [anon_sym_and] = ACTIONS(2007),
    [anon_sym_TILDE] = ACTIONS(2007),
    [anon_sym_QMARK] = ACTIONS(2007),
    [anon_sym_LPAREN] = ACTIONS(2009),
    [anon_sym_RPAREN] = ACTIONS(2009),
    [anon_sym_COMMA] = ACTIONS(2009),
    [anon_sym_PIPE] = ACTIONS(2007),
    [anon_sym_COLON2] = ACTIONS(2007),
    [anon_sym_DOT] = ACTIONS(2009),
    [anon_sym_DASH_GT] = ACTIONS(2007),
    [anon_sym_LBRACK] = ACTIONS(2007),
    [anon_sym_RBRACK] = ACTIONS(2009),
    [anon_sym_true] = ACTIONS(2007),
    [anon_sym_false] = ACTIONS(2007),
    [anon_sym_LBRACE] = ACTIONS(2007),
    [anon_sym_SEMI] = ACTIONS(2009),
    [anon_sym_RBRACE] = ACTIONS(2009),
    [anon_sym_constraint] = ACTIONS(2007),
    [anon_sym_val] = ACTIONS(2007),
    [anon_sym_end] = ACTIONS(2007),
    [anon_sym_with] = ACTIONS(2007),
    [anon_sym_COLON_GT] = ACTIONS(2009),
    [anon_sym_object] = ACTIONS(2007),
    [anon_sym_inherit] = ACTIONS(2007),
    [anon_sym_method] = ACTIONS(2007),
    [anon_sym_in] = ACTIONS(2007),
    [anon_sym_as] = ACTIONS(2007),
    [anon_sym_initializer] = ACTIONS(2007),
    [anon_sym_AMP] = ACTIONS(2007),
    [anon_sym_POUND] = ACTIONS(2007),
    [anon_sym_COLON_COLON] = ACTIONS(2009),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2009),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2009),
    [anon_sym_then] = ACTIONS(2007),
    [anon_sym_else] = ACTIONS(2007),
    [anon_sym_do] = ACTIONS(2007),
    [anon_sym_done] = ACTIONS(2007),
    [anon_sym_to] = ACTIONS(2007),
    [anon_sym_downto] = ACTIONS(2007),
    [anon_sym_new] = ACTIONS(2007),
    [anon_sym_LBRACE_LT] = ACTIONS(2009),
    [anon_sym_GT_RBRACE] = ACTIONS(2009),
    [anon_sym_begin] = ACTIONS(2007),
    [sym_ocamlyacc_value] = ACTIONS(2009),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2007),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2009),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2007),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2007),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2009),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2009),
    [aux_sym_number_token1] = ACTIONS(2009),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2009),
    [sym_prefix_operator] = ACTIONS(2009),
    [aux_sym_sign_operator_token1] = ACTIONS(2007),
    [sym_hash_operator] = ACTIONS(2009),
    [anon_sym_lsl] = ACTIONS(2007),
    [anon_sym_lsr] = ACTIONS(2007),
    [anon_sym_asr] = ACTIONS(2007),
    [aux_sym_pow_operator_token1] = ACTIONS(2009),
    [anon_sym_mod] = ACTIONS(2007),
    [anon_sym_land] = ACTIONS(2007),
    [anon_sym_lor] = ACTIONS(2007),
    [anon_sym_lxor] = ACTIONS(2007),
    [aux_sym_mult_operator_token1] = ACTIONS(2007),
    [aux_sym_add_operator_token1] = ACTIONS(2007),
    [sym_concat_operator] = ACTIONS(2009),
    [sym_rel_operator] = ACTIONS(2007),
    [anon_sym_AMP_AMP] = ACTIONS(2007),
    [anon_sym_or] = ACTIONS(2007),
    [anon_sym_PIPE_PIPE] = ACTIONS(2007),
    [sym_assign_operator] = ACTIONS(2009),
    [sym__uppercase_identifier] = ACTIONS(2009),
    [aux_sym_tag_token1] = ACTIONS(2009),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(962)] = {
    [sym_attribute] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(2011),
    [anon_sym_and] = ACTIONS(2011),
    [anon_sym_TILDE] = ACTIONS(2011),
    [anon_sym_QMARK] = ACTIONS(2011),
    [anon_sym_LPAREN] = ACTIONS(2013),
    [anon_sym_RPAREN] = ACTIONS(2013),
    [anon_sym_COMMA] = ACTIONS(2013),
    [anon_sym_PIPE] = ACTIONS(2011),
    [anon_sym_COLON2] = ACTIONS(2011),
    [anon_sym_DOT] = ACTIONS(2013),
    [anon_sym_DASH_GT] = ACTIONS(2011),
    [anon_sym_LBRACK] = ACTIONS(2011),
    [anon_sym_RBRACK] = ACTIONS(2013),
    [anon_sym_true] = ACTIONS(2011),
    [anon_sym_false] = ACTIONS(2011),
    [anon_sym_LBRACE] = ACTIONS(2011),
    [anon_sym_SEMI] = ACTIONS(2013),
    [anon_sym_RBRACE] = ACTIONS(2013),
    [anon_sym_constraint] = ACTIONS(2011),
    [anon_sym_val] = ACTIONS(2011),
    [anon_sym_end] = ACTIONS(2011),
    [anon_sym_with] = ACTIONS(2011),
    [anon_sym_COLON_GT] = ACTIONS(2013),
    [anon_sym_object] = ACTIONS(2011),
    [anon_sym_inherit] = ACTIONS(2011),
    [anon_sym_method] = ACTIONS(2011),
    [anon_sym_in] = ACTIONS(2011),
    [anon_sym_as] = ACTIONS(2011),
    [anon_sym_initializer] = ACTIONS(2011),
    [anon_sym_AMP] = ACTIONS(2011),
    [anon_sym_POUND] = ACTIONS(2011),
    [anon_sym_COLON_COLON] = ACTIONS(2013),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2013),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2013),
    [anon_sym_then] = ACTIONS(2011),
    [anon_sym_else] = ACTIONS(2011),
    [anon_sym_do] = ACTIONS(2011),
    [anon_sym_done] = ACTIONS(2011),
    [anon_sym_to] = ACTIONS(2011),
    [anon_sym_downto] = ACTIONS(2011),
    [anon_sym_new] = ACTIONS(2011),
    [anon_sym_LBRACE_LT] = ACTIONS(2013),
    [anon_sym_GT_RBRACE] = ACTIONS(2013),
    [anon_sym_begin] = ACTIONS(2011),
    [sym_ocamlyacc_value] = ACTIONS(2013),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2011),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2013),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2011),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2011),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2013),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2013),
    [aux_sym_number_token1] = ACTIONS(2013),
    [aux_sym_signed_number_token1] = ACTIONS(2011),
    [anon_sym_SQUOTE] = ACTIONS(2013),
    [anon_sym_DQUOTE] = ACTIONS(2013),
    [sym_prefix_operator] = ACTIONS(2013),
    [aux_sym_sign_operator_token1] = ACTIONS(2011),
    [sym_hash_operator] = ACTIONS(2013),
    [anon_sym_lsl] = ACTIONS(2011),
    [anon_sym_lsr] = ACTIONS(2011),
    [anon_sym_asr] = ACTIONS(2011),
    [aux_sym_pow_operator_token1] = ACTIONS(2013),
    [anon_sym_mod] = ACTIONS(2011),
    [anon_sym_land] = ACTIONS(2011),
    [anon_sym_lor] = ACTIONS(2011),
    [anon_sym_lxor] = ACTIONS(2011),
    [aux_sym_mult_operator_token1] = ACTIONS(2011),
    [aux_sym_add_operator_token1] = ACTIONS(2011),
    [sym_concat_operator] = ACTIONS(2013),
    [sym_rel_operator] = ACTIONS(2011),
    [anon_sym_AMP_AMP] = ACTIONS(2011),
    [anon_sym_or] = ACTIONS(2011),
    [anon_sym_PIPE_PIPE] = ACTIONS(2011),
    [sym_assign_operator] = ACTIONS(2013),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_tag_token1] = ACTIONS(2013),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(963)] = {
    [sym_attribute] = STATE(963),
    [sym__lowercase_identifier] = ACTIONS(2015),
    [anon_sym_and] = ACTIONS(2015),
    [anon_sym_TILDE] = ACTIONS(2015),
    [anon_sym_QMARK] = ACTIONS(2015),
    [anon_sym_LPAREN] = ACTIONS(2017),
    [anon_sym_RPAREN] = ACTIONS(2017),
    [anon_sym_COMMA] = ACTIONS(2017),
    [anon_sym_PIPE] = ACTIONS(2015),
    [anon_sym_COLON2] = ACTIONS(2015),
    [anon_sym_DOT] = ACTIONS(2017),
    [anon_sym_DASH_GT] = ACTIONS(2015),
    [anon_sym_LBRACK] = ACTIONS(2015),
    [anon_sym_RBRACK] = ACTIONS(2017),
    [anon_sym_true] = ACTIONS(2015),
    [anon_sym_false] = ACTIONS(2015),
    [anon_sym_LBRACE] = ACTIONS(2015),
    [anon_sym_SEMI] = ACTIONS(2017),
    [anon_sym_RBRACE] = ACTIONS(2017),
    [anon_sym_constraint] = ACTIONS(2015),
    [anon_sym_val] = ACTIONS(2015),
    [anon_sym_end] = ACTIONS(2015),
    [anon_sym_with] = ACTIONS(2015),
    [anon_sym_COLON_GT] = ACTIONS(2017),
    [anon_sym_object] = ACTIONS(2015),
    [anon_sym_inherit] = ACTIONS(2015),
    [anon_sym_method] = ACTIONS(2015),
    [anon_sym_in] = ACTIONS(2015),
    [anon_sym_as] = ACTIONS(2015),
    [anon_sym_initializer] = ACTIONS(2015),
    [anon_sym_AMP] = ACTIONS(2015),
    [anon_sym_POUND] = ACTIONS(2015),
    [anon_sym_COLON_COLON] = ACTIONS(2017),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2017),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2017),
    [anon_sym_then] = ACTIONS(2015),
    [anon_sym_else] = ACTIONS(2015),
    [anon_sym_do] = ACTIONS(2015),
    [anon_sym_done] = ACTIONS(2015),
    [anon_sym_to] = ACTIONS(2015),
    [anon_sym_downto] = ACTIONS(2015),
    [anon_sym_new] = ACTIONS(2015),
    [anon_sym_LBRACE_LT] = ACTIONS(2017),
    [anon_sym_GT_RBRACE] = ACTIONS(2017),
    [anon_sym_begin] = ACTIONS(2015),
    [sym_ocamlyacc_value] = ACTIONS(2017),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2015),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2017),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2015),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2015),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2017),
    [aux_sym_number_token1] = ACTIONS(2017),
    [aux_sym_signed_number_token1] = ACTIONS(2015),
    [anon_sym_SQUOTE] = ACTIONS(2017),
    [anon_sym_DQUOTE] = ACTIONS(2017),
    [sym_prefix_operator] = ACTIONS(2017),
    [aux_sym_sign_operator_token1] = ACTIONS(2015),
    [sym_hash_operator] = ACTIONS(2017),
    [anon_sym_lsl] = ACTIONS(2015),
    [anon_sym_lsr] = ACTIONS(2015),
    [anon_sym_asr] = ACTIONS(2015),
    [aux_sym_pow_operator_token1] = ACTIONS(2017),
    [anon_sym_mod] = ACTIONS(2015),
    [anon_sym_land] = ACTIONS(2015),
    [anon_sym_lor] = ACTIONS(2015),
    [anon_sym_lxor] = ACTIONS(2015),
    [aux_sym_mult_operator_token1] = ACTIONS(2015),
    [aux_sym_add_operator_token1] = ACTIONS(2015),
    [sym_concat_operator] = ACTIONS(2017),
    [sym_rel_operator] = ACTIONS(2015),
    [anon_sym_AMP_AMP] = ACTIONS(2015),
    [anon_sym_or] = ACTIONS(2015),
    [anon_sym_PIPE_PIPE] = ACTIONS(2015),
    [sym_assign_operator] = ACTIONS(2017),
    [sym__uppercase_identifier] = ACTIONS(2017),
    [aux_sym_tag_token1] = ACTIONS(2017),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(964)] = {
    [sym_attribute] = STATE(964),
    [sym__lowercase_identifier] = ACTIONS(2019),
    [anon_sym_and] = ACTIONS(2019),
    [anon_sym_TILDE] = ACTIONS(2019),
    [anon_sym_QMARK] = ACTIONS(2019),
    [anon_sym_LPAREN] = ACTIONS(2021),
    [anon_sym_RPAREN] = ACTIONS(2021),
    [anon_sym_COMMA] = ACTIONS(2021),
    [anon_sym_PIPE] = ACTIONS(2019),
    [anon_sym_COLON2] = ACTIONS(2019),
    [anon_sym_DOT] = ACTIONS(2021),
    [anon_sym_DASH_GT] = ACTIONS(2019),
    [anon_sym_LBRACK] = ACTIONS(2019),
    [anon_sym_RBRACK] = ACTIONS(2021),
    [anon_sym_true] = ACTIONS(2019),
    [anon_sym_false] = ACTIONS(2019),
    [anon_sym_LBRACE] = ACTIONS(2019),
    [anon_sym_SEMI] = ACTIONS(2021),
    [anon_sym_RBRACE] = ACTIONS(2021),
    [anon_sym_constraint] = ACTIONS(2019),
    [anon_sym_val] = ACTIONS(2019),
    [anon_sym_end] = ACTIONS(2019),
    [anon_sym_with] = ACTIONS(2019),
    [anon_sym_COLON_GT] = ACTIONS(2021),
    [anon_sym_object] = ACTIONS(2019),
    [anon_sym_inherit] = ACTIONS(2019),
    [anon_sym_method] = ACTIONS(2019),
    [anon_sym_in] = ACTIONS(2019),
    [anon_sym_as] = ACTIONS(2019),
    [anon_sym_initializer] = ACTIONS(2019),
    [anon_sym_AMP] = ACTIONS(2019),
    [anon_sym_POUND] = ACTIONS(2019),
    [anon_sym_COLON_COLON] = ACTIONS(2021),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2021),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2021),
    [anon_sym_then] = ACTIONS(2019),
    [anon_sym_else] = ACTIONS(2019),
    [anon_sym_do] = ACTIONS(2019),
    [anon_sym_done] = ACTIONS(2019),
    [anon_sym_to] = ACTIONS(2019),
    [anon_sym_downto] = ACTIONS(2019),
    [anon_sym_new] = ACTIONS(2019),
    [anon_sym_LBRACE_LT] = ACTIONS(2021),
    [anon_sym_GT_RBRACE] = ACTIONS(2021),
    [anon_sym_begin] = ACTIONS(2019),
    [sym_ocamlyacc_value] = ACTIONS(2021),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2019),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2021),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2021),
    [aux_sym_number_token1] = ACTIONS(2021),
    [aux_sym_signed_number_token1] = ACTIONS(2019),
    [anon_sym_SQUOTE] = ACTIONS(2021),
    [anon_sym_DQUOTE] = ACTIONS(2021),
    [sym_prefix_operator] = ACTIONS(2021),
    [aux_sym_sign_operator_token1] = ACTIONS(2019),
    [sym_hash_operator] = ACTIONS(2021),
    [anon_sym_lsl] = ACTIONS(2019),
    [anon_sym_lsr] = ACTIONS(2019),
    [anon_sym_asr] = ACTIONS(2019),
    [aux_sym_pow_operator_token1] = ACTIONS(2021),
    [anon_sym_mod] = ACTIONS(2019),
    [anon_sym_land] = ACTIONS(2019),
    [anon_sym_lor] = ACTIONS(2019),
    [anon_sym_lxor] = ACTIONS(2019),
    [aux_sym_mult_operator_token1] = ACTIONS(2019),
    [aux_sym_add_operator_token1] = ACTIONS(2019),
    [sym_concat_operator] = ACTIONS(2021),
    [sym_rel_operator] = ACTIONS(2019),
    [anon_sym_AMP_AMP] = ACTIONS(2019),
    [anon_sym_or] = ACTIONS(2019),
    [anon_sym_PIPE_PIPE] = ACTIONS(2019),
    [sym_assign_operator] = ACTIONS(2021),
    [sym__uppercase_identifier] = ACTIONS(2021),
    [aux_sym_tag_token1] = ACTIONS(2021),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(965)] = {
    [sym_attribute] = STATE(965),
    [sym__lowercase_identifier] = ACTIONS(2023),
    [anon_sym_and] = ACTIONS(2023),
    [anon_sym_TILDE] = ACTIONS(2023),
    [anon_sym_QMARK] = ACTIONS(2023),
    [anon_sym_LPAREN] = ACTIONS(2025),
    [anon_sym_RPAREN] = ACTIONS(2025),
    [anon_sym_COMMA] = ACTIONS(2025),
    [anon_sym_PIPE] = ACTIONS(2023),
    [anon_sym_COLON2] = ACTIONS(2023),
    [anon_sym_DOT] = ACTIONS(2025),
    [anon_sym_DASH_GT] = ACTIONS(2023),
    [anon_sym_LBRACK] = ACTIONS(2023),
    [anon_sym_RBRACK] = ACTIONS(2025),
    [anon_sym_true] = ACTIONS(2023),
    [anon_sym_false] = ACTIONS(2023),
    [anon_sym_LBRACE] = ACTIONS(2023),
    [anon_sym_SEMI] = ACTIONS(2025),
    [anon_sym_RBRACE] = ACTIONS(2025),
    [anon_sym_constraint] = ACTIONS(2023),
    [anon_sym_val] = ACTIONS(2023),
    [anon_sym_end] = ACTIONS(2023),
    [anon_sym_with] = ACTIONS(2023),
    [anon_sym_COLON_GT] = ACTIONS(2025),
    [anon_sym_object] = ACTIONS(2023),
    [anon_sym_inherit] = ACTIONS(2023),
    [anon_sym_method] = ACTIONS(2023),
    [anon_sym_in] = ACTIONS(2023),
    [anon_sym_as] = ACTIONS(2023),
    [anon_sym_initializer] = ACTIONS(2023),
    [anon_sym_AMP] = ACTIONS(2023),
    [anon_sym_POUND] = ACTIONS(2023),
    [anon_sym_COLON_COLON] = ACTIONS(2025),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2025),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2025),
    [anon_sym_then] = ACTIONS(2023),
    [anon_sym_else] = ACTIONS(2023),
    [anon_sym_do] = ACTIONS(2023),
    [anon_sym_done] = ACTIONS(2023),
    [anon_sym_to] = ACTIONS(2023),
    [anon_sym_downto] = ACTIONS(2023),
    [anon_sym_new] = ACTIONS(2023),
    [anon_sym_LBRACE_LT] = ACTIONS(2025),
    [anon_sym_GT_RBRACE] = ACTIONS(2025),
    [anon_sym_begin] = ACTIONS(2023),
    [sym_ocamlyacc_value] = ACTIONS(2025),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2023),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2025),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2025),
    [aux_sym_number_token1] = ACTIONS(2025),
    [aux_sym_signed_number_token1] = ACTIONS(2023),
    [anon_sym_SQUOTE] = ACTIONS(2025),
    [anon_sym_DQUOTE] = ACTIONS(2025),
    [sym_prefix_operator] = ACTIONS(2025),
    [aux_sym_sign_operator_token1] = ACTIONS(2023),
    [sym_hash_operator] = ACTIONS(2025),
    [anon_sym_lsl] = ACTIONS(2023),
    [anon_sym_lsr] = ACTIONS(2023),
    [anon_sym_asr] = ACTIONS(2023),
    [aux_sym_pow_operator_token1] = ACTIONS(2025),
    [anon_sym_mod] = ACTIONS(2023),
    [anon_sym_land] = ACTIONS(2023),
    [anon_sym_lor] = ACTIONS(2023),
    [anon_sym_lxor] = ACTIONS(2023),
    [aux_sym_mult_operator_token1] = ACTIONS(2023),
    [aux_sym_add_operator_token1] = ACTIONS(2023),
    [sym_concat_operator] = ACTIONS(2025),
    [sym_rel_operator] = ACTIONS(2023),
    [anon_sym_AMP_AMP] = ACTIONS(2023),
    [anon_sym_or] = ACTIONS(2023),
    [anon_sym_PIPE_PIPE] = ACTIONS(2023),
    [sym_assign_operator] = ACTIONS(2025),
    [sym__uppercase_identifier] = ACTIONS(2025),
    [aux_sym_tag_token1] = ACTIONS(2025),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(966)] = {
    [sym_attribute] = STATE(966),
    [sym__lowercase_identifier] = ACTIONS(2027),
    [anon_sym_and] = ACTIONS(2027),
    [anon_sym_TILDE] = ACTIONS(2027),
    [anon_sym_QMARK] = ACTIONS(2027),
    [anon_sym_LPAREN] = ACTIONS(2029),
    [anon_sym_RPAREN] = ACTIONS(2029),
    [anon_sym_COMMA] = ACTIONS(2029),
    [anon_sym_PIPE] = ACTIONS(2027),
    [anon_sym_COLON2] = ACTIONS(2027),
    [anon_sym_DOT] = ACTIONS(2029),
    [anon_sym_DASH_GT] = ACTIONS(2027),
    [anon_sym_LBRACK] = ACTIONS(2027),
    [anon_sym_RBRACK] = ACTIONS(2029),
    [anon_sym_true] = ACTIONS(2027),
    [anon_sym_false] = ACTIONS(2027),
    [anon_sym_LBRACE] = ACTIONS(2027),
    [anon_sym_SEMI] = ACTIONS(2029),
    [anon_sym_RBRACE] = ACTIONS(2029),
    [anon_sym_constraint] = ACTIONS(2027),
    [anon_sym_val] = ACTIONS(2027),
    [anon_sym_end] = ACTIONS(2027),
    [anon_sym_with] = ACTIONS(2027),
    [anon_sym_COLON_GT] = ACTIONS(2029),
    [anon_sym_object] = ACTIONS(2027),
    [anon_sym_inherit] = ACTIONS(2027),
    [anon_sym_method] = ACTIONS(2027),
    [anon_sym_in] = ACTIONS(2027),
    [anon_sym_as] = ACTIONS(2027),
    [anon_sym_initializer] = ACTIONS(2027),
    [anon_sym_AMP] = ACTIONS(2027),
    [anon_sym_POUND] = ACTIONS(2027),
    [anon_sym_COLON_COLON] = ACTIONS(2029),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2029),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2029),
    [anon_sym_then] = ACTIONS(2027),
    [anon_sym_else] = ACTIONS(2027),
    [anon_sym_do] = ACTIONS(2027),
    [anon_sym_done] = ACTIONS(2027),
    [anon_sym_to] = ACTIONS(2027),
    [anon_sym_downto] = ACTIONS(2027),
    [anon_sym_new] = ACTIONS(2027),
    [anon_sym_LBRACE_LT] = ACTIONS(2029),
    [anon_sym_GT_RBRACE] = ACTIONS(2029),
    [anon_sym_begin] = ACTIONS(2027),
    [sym_ocamlyacc_value] = ACTIONS(2029),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2027),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2029),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2029),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2029),
    [aux_sym_number_token1] = ACTIONS(2029),
    [aux_sym_signed_number_token1] = ACTIONS(2027),
    [anon_sym_SQUOTE] = ACTIONS(2029),
    [anon_sym_DQUOTE] = ACTIONS(2029),
    [sym_prefix_operator] = ACTIONS(2029),
    [aux_sym_sign_operator_token1] = ACTIONS(2027),
    [sym_hash_operator] = ACTIONS(2029),
    [anon_sym_lsl] = ACTIONS(2027),
    [anon_sym_lsr] = ACTIONS(2027),
    [anon_sym_asr] = ACTIONS(2027),
    [aux_sym_pow_operator_token1] = ACTIONS(2029),
    [anon_sym_mod] = ACTIONS(2027),
    [anon_sym_land] = ACTIONS(2027),
    [anon_sym_lor] = ACTIONS(2027),
    [anon_sym_lxor] = ACTIONS(2027),
    [aux_sym_mult_operator_token1] = ACTIONS(2027),
    [aux_sym_add_operator_token1] = ACTIONS(2027),
    [sym_concat_operator] = ACTIONS(2029),
    [sym_rel_operator] = ACTIONS(2027),
    [anon_sym_AMP_AMP] = ACTIONS(2027),
    [anon_sym_or] = ACTIONS(2027),
    [anon_sym_PIPE_PIPE] = ACTIONS(2027),
    [sym_assign_operator] = ACTIONS(2029),
    [sym__uppercase_identifier] = ACTIONS(2029),
    [aux_sym_tag_token1] = ACTIONS(2029),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(967)] = {
    [sym_attribute] = STATE(967),
    [sym__lowercase_identifier] = ACTIONS(2031),
    [anon_sym_and] = ACTIONS(2031),
    [anon_sym_TILDE] = ACTIONS(2031),
    [anon_sym_QMARK] = ACTIONS(2031),
    [anon_sym_LPAREN] = ACTIONS(2033),
    [anon_sym_RPAREN] = ACTIONS(2033),
    [anon_sym_COMMA] = ACTIONS(2033),
    [anon_sym_PIPE] = ACTIONS(2031),
    [anon_sym_COLON2] = ACTIONS(2031),
    [anon_sym_DOT] = ACTIONS(2033),
    [anon_sym_DASH_GT] = ACTIONS(2031),
    [anon_sym_LBRACK] = ACTIONS(2031),
    [anon_sym_RBRACK] = ACTIONS(2033),
    [anon_sym_true] = ACTIONS(2031),
    [anon_sym_false] = ACTIONS(2031),
    [anon_sym_LBRACE] = ACTIONS(2031),
    [anon_sym_SEMI] = ACTIONS(2033),
    [anon_sym_RBRACE] = ACTIONS(2033),
    [anon_sym_constraint] = ACTIONS(2031),
    [anon_sym_val] = ACTIONS(2031),
    [anon_sym_end] = ACTIONS(2031),
    [anon_sym_with] = ACTIONS(2031),
    [anon_sym_COLON_GT] = ACTIONS(2033),
    [anon_sym_object] = ACTIONS(2031),
    [anon_sym_inherit] = ACTIONS(2031),
    [anon_sym_method] = ACTIONS(2031),
    [anon_sym_in] = ACTIONS(2031),
    [anon_sym_as] = ACTIONS(2031),
    [anon_sym_initializer] = ACTIONS(2031),
    [anon_sym_AMP] = ACTIONS(2031),
    [anon_sym_POUND] = ACTIONS(2031),
    [anon_sym_COLON_COLON] = ACTIONS(2033),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2033),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2033),
    [anon_sym_then] = ACTIONS(2031),
    [anon_sym_else] = ACTIONS(2031),
    [anon_sym_do] = ACTIONS(2031),
    [anon_sym_done] = ACTIONS(2031),
    [anon_sym_to] = ACTIONS(2031),
    [anon_sym_downto] = ACTIONS(2031),
    [anon_sym_new] = ACTIONS(2031),
    [anon_sym_LBRACE_LT] = ACTIONS(2033),
    [anon_sym_GT_RBRACE] = ACTIONS(2033),
    [anon_sym_begin] = ACTIONS(2031),
    [sym_ocamlyacc_value] = ACTIONS(2033),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2031),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2033),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2031),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2031),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2033),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2033),
    [aux_sym_number_token1] = ACTIONS(2033),
    [aux_sym_signed_number_token1] = ACTIONS(2031),
    [anon_sym_SQUOTE] = ACTIONS(2033),
    [anon_sym_DQUOTE] = ACTIONS(2033),
    [sym_prefix_operator] = ACTIONS(2033),
    [aux_sym_sign_operator_token1] = ACTIONS(2031),
    [sym_hash_operator] = ACTIONS(2033),
    [anon_sym_lsl] = ACTIONS(2031),
    [anon_sym_lsr] = ACTIONS(2031),
    [anon_sym_asr] = ACTIONS(2031),
    [aux_sym_pow_operator_token1] = ACTIONS(2033),
    [anon_sym_mod] = ACTIONS(2031),
    [anon_sym_land] = ACTIONS(2031),
    [anon_sym_lor] = ACTIONS(2031),
    [anon_sym_lxor] = ACTIONS(2031),
    [aux_sym_mult_operator_token1] = ACTIONS(2031),
    [aux_sym_add_operator_token1] = ACTIONS(2031),
    [sym_concat_operator] = ACTIONS(2033),
    [sym_rel_operator] = ACTIONS(2031),
    [anon_sym_AMP_AMP] = ACTIONS(2031),
    [anon_sym_or] = ACTIONS(2031),
    [anon_sym_PIPE_PIPE] = ACTIONS(2031),
    [sym_assign_operator] = ACTIONS(2033),
    [sym__uppercase_identifier] = ACTIONS(2033),
    [aux_sym_tag_token1] = ACTIONS(2033),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(968)] = {
    [sym_attribute] = STATE(968),
    [sym__lowercase_identifier] = ACTIONS(2035),
    [anon_sym_and] = ACTIONS(2035),
    [anon_sym_TILDE] = ACTIONS(2035),
    [anon_sym_QMARK] = ACTIONS(2035),
    [anon_sym_LPAREN] = ACTIONS(2037),
    [anon_sym_RPAREN] = ACTIONS(2037),
    [anon_sym_COMMA] = ACTIONS(2037),
    [anon_sym_PIPE] = ACTIONS(2035),
    [anon_sym_COLON2] = ACTIONS(2035),
    [anon_sym_DOT] = ACTIONS(2037),
    [anon_sym_DASH_GT] = ACTIONS(2035),
    [anon_sym_LBRACK] = ACTIONS(2035),
    [anon_sym_RBRACK] = ACTIONS(2037),
    [anon_sym_true] = ACTIONS(2035),
    [anon_sym_false] = ACTIONS(2035),
    [anon_sym_LBRACE] = ACTIONS(2035),
    [anon_sym_SEMI] = ACTIONS(2037),
    [anon_sym_RBRACE] = ACTIONS(2037),
    [anon_sym_constraint] = ACTIONS(2035),
    [anon_sym_val] = ACTIONS(2035),
    [anon_sym_end] = ACTIONS(2035),
    [anon_sym_with] = ACTIONS(2035),
    [anon_sym_COLON_GT] = ACTIONS(2037),
    [anon_sym_object] = ACTIONS(2035),
    [anon_sym_inherit] = ACTIONS(2035),
    [anon_sym_method] = ACTIONS(2035),
    [anon_sym_in] = ACTIONS(2035),
    [anon_sym_as] = ACTIONS(2035),
    [anon_sym_initializer] = ACTIONS(2035),
    [anon_sym_AMP] = ACTIONS(2035),
    [anon_sym_POUND] = ACTIONS(2035),
    [anon_sym_COLON_COLON] = ACTIONS(2037),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2037),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2037),
    [anon_sym_then] = ACTIONS(2035),
    [anon_sym_else] = ACTIONS(2035),
    [anon_sym_do] = ACTIONS(2035),
    [anon_sym_done] = ACTIONS(2035),
    [anon_sym_to] = ACTIONS(2035),
    [anon_sym_downto] = ACTIONS(2035),
    [anon_sym_new] = ACTIONS(2035),
    [anon_sym_LBRACE_LT] = ACTIONS(2037),
    [anon_sym_GT_RBRACE] = ACTIONS(2037),
    [anon_sym_begin] = ACTIONS(2035),
    [sym_ocamlyacc_value] = ACTIONS(2037),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2035),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2037),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2035),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2035),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2037),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2037),
    [aux_sym_number_token1] = ACTIONS(2037),
    [aux_sym_signed_number_token1] = ACTIONS(2035),
    [anon_sym_SQUOTE] = ACTIONS(2037),
    [anon_sym_DQUOTE] = ACTIONS(2037),
    [sym_prefix_operator] = ACTIONS(2037),
    [aux_sym_sign_operator_token1] = ACTIONS(2035),
    [sym_hash_operator] = ACTIONS(2037),
    [anon_sym_lsl] = ACTIONS(2035),
    [anon_sym_lsr] = ACTIONS(2035),
    [anon_sym_asr] = ACTIONS(2035),
    [aux_sym_pow_operator_token1] = ACTIONS(2037),
    [anon_sym_mod] = ACTIONS(2035),
    [anon_sym_land] = ACTIONS(2035),
    [anon_sym_lor] = ACTIONS(2035),
    [anon_sym_lxor] = ACTIONS(2035),
    [aux_sym_mult_operator_token1] = ACTIONS(2035),
    [aux_sym_add_operator_token1] = ACTIONS(2035),
    [sym_concat_operator] = ACTIONS(2037),
    [sym_rel_operator] = ACTIONS(2035),
    [anon_sym_AMP_AMP] = ACTIONS(2035),
    [anon_sym_or] = ACTIONS(2035),
    [anon_sym_PIPE_PIPE] = ACTIONS(2035),
    [sym_assign_operator] = ACTIONS(2037),
    [sym__uppercase_identifier] = ACTIONS(2037),
    [aux_sym_tag_token1] = ACTIONS(2037),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(969)] = {
    [sym_attribute] = STATE(969),
    [sym__lowercase_identifier] = ACTIONS(2039),
    [anon_sym_and] = ACTIONS(2039),
    [anon_sym_TILDE] = ACTIONS(2039),
    [anon_sym_QMARK] = ACTIONS(2039),
    [anon_sym_LPAREN] = ACTIONS(2041),
    [anon_sym_RPAREN] = ACTIONS(2041),
    [anon_sym_COMMA] = ACTIONS(2041),
    [anon_sym_PIPE] = ACTIONS(2039),
    [anon_sym_COLON2] = ACTIONS(2039),
    [anon_sym_DOT] = ACTIONS(919),
    [anon_sym_DASH_GT] = ACTIONS(2039),
    [anon_sym_LBRACK] = ACTIONS(2039),
    [anon_sym_RBRACK] = ACTIONS(2041),
    [anon_sym_true] = ACTIONS(2039),
    [anon_sym_false] = ACTIONS(2039),
    [anon_sym_LBRACE] = ACTIONS(2039),
    [anon_sym_SEMI] = ACTIONS(2041),
    [anon_sym_RBRACE] = ACTIONS(2041),
    [anon_sym_constraint] = ACTIONS(2039),
    [anon_sym_val] = ACTIONS(2039),
    [anon_sym_end] = ACTIONS(2039),
    [anon_sym_with] = ACTIONS(2039),
    [anon_sym_COLON_GT] = ACTIONS(2041),
    [anon_sym_object] = ACTIONS(2039),
    [anon_sym_inherit] = ACTIONS(2039),
    [anon_sym_method] = ACTIONS(2039),
    [anon_sym_in] = ACTIONS(2039),
    [anon_sym_as] = ACTIONS(2039),
    [anon_sym_initializer] = ACTIONS(2039),
    [anon_sym_AMP] = ACTIONS(2039),
    [anon_sym_POUND] = ACTIONS(921),
    [anon_sym_COLON_COLON] = ACTIONS(2041),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2041),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2041),
    [anon_sym_then] = ACTIONS(2039),
    [anon_sym_else] = ACTIONS(2039),
    [anon_sym_do] = ACTIONS(2039),
    [anon_sym_done] = ACTIONS(2039),
    [anon_sym_to] = ACTIONS(2039),
    [anon_sym_downto] = ACTIONS(2039),
    [anon_sym_new] = ACTIONS(2039),
    [anon_sym_LBRACE_LT] = ACTIONS(2041),
    [anon_sym_GT_RBRACE] = ACTIONS(2041),
    [anon_sym_begin] = ACTIONS(2039),
    [sym_ocamlyacc_value] = ACTIONS(2041),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2039),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2041),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2039),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2039),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2041),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2041),
    [aux_sym_number_token1] = ACTIONS(2041),
    [aux_sym_signed_number_token1] = ACTIONS(2039),
    [anon_sym_SQUOTE] = ACTIONS(2041),
    [anon_sym_DQUOTE] = ACTIONS(2041),
    [sym_prefix_operator] = ACTIONS(2041),
    [aux_sym_sign_operator_token1] = ACTIONS(2039),
    [sym_hash_operator] = ACTIONS(923),
    [anon_sym_lsl] = ACTIONS(2039),
    [anon_sym_lsr] = ACTIONS(2039),
    [anon_sym_asr] = ACTIONS(2039),
    [aux_sym_pow_operator_token1] = ACTIONS(2041),
    [anon_sym_mod] = ACTIONS(2039),
    [anon_sym_land] = ACTIONS(2039),
    [anon_sym_lor] = ACTIONS(2039),
    [anon_sym_lxor] = ACTIONS(2039),
    [aux_sym_mult_operator_token1] = ACTIONS(2039),
    [aux_sym_add_operator_token1] = ACTIONS(2039),
    [sym_concat_operator] = ACTIONS(2041),
    [sym_rel_operator] = ACTIONS(2039),
    [anon_sym_AMP_AMP] = ACTIONS(2039),
    [anon_sym_or] = ACTIONS(2039),
    [anon_sym_PIPE_PIPE] = ACTIONS(2039),
    [sym_assign_operator] = ACTIONS(2041),
    [sym__uppercase_identifier] = ACTIONS(2041),
    [aux_sym_tag_token1] = ACTIONS(2041),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(970)] = {
    [sym_attribute] = STATE(970),
    [sym__lowercase_identifier] = ACTIONS(2043),
    [anon_sym_and] = ACTIONS(2043),
    [anon_sym_TILDE] = ACTIONS(2043),
    [anon_sym_QMARK] = ACTIONS(2043),
    [anon_sym_LPAREN] = ACTIONS(2045),
    [anon_sym_RPAREN] = ACTIONS(2045),
    [anon_sym_COMMA] = ACTIONS(2045),
    [anon_sym_PIPE] = ACTIONS(2043),
    [anon_sym_COLON2] = ACTIONS(2043),
    [anon_sym_DOT] = ACTIONS(2045),
    [anon_sym_DASH_GT] = ACTIONS(2043),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_RBRACK] = ACTIONS(2045),
    [anon_sym_true] = ACTIONS(2043),
    [anon_sym_false] = ACTIONS(2043),
    [anon_sym_LBRACE] = ACTIONS(2043),
    [anon_sym_SEMI] = ACTIONS(2045),
    [anon_sym_RBRACE] = ACTIONS(2045),
    [anon_sym_constraint] = ACTIONS(2043),
    [anon_sym_val] = ACTIONS(2043),
    [anon_sym_end] = ACTIONS(2043),
    [anon_sym_with] = ACTIONS(2043),
    [anon_sym_COLON_GT] = ACTIONS(2045),
    [anon_sym_object] = ACTIONS(2043),
    [anon_sym_inherit] = ACTIONS(2043),
    [anon_sym_method] = ACTIONS(2043),
    [anon_sym_in] = ACTIONS(2043),
    [anon_sym_as] = ACTIONS(2043),
    [anon_sym_initializer] = ACTIONS(2043),
    [anon_sym_AMP] = ACTIONS(2043),
    [anon_sym_POUND] = ACTIONS(2043),
    [anon_sym_COLON_COLON] = ACTIONS(2045),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2045),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2045),
    [anon_sym_then] = ACTIONS(2043),
    [anon_sym_else] = ACTIONS(2043),
    [anon_sym_do] = ACTIONS(2043),
    [anon_sym_done] = ACTIONS(2043),
    [anon_sym_to] = ACTIONS(2043),
    [anon_sym_downto] = ACTIONS(2043),
    [anon_sym_new] = ACTIONS(2043),
    [anon_sym_LBRACE_LT] = ACTIONS(2045),
    [anon_sym_GT_RBRACE] = ACTIONS(2045),
    [anon_sym_begin] = ACTIONS(2043),
    [sym_ocamlyacc_value] = ACTIONS(2045),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2043),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2045),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2043),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2043),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2045),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2045),
    [aux_sym_number_token1] = ACTIONS(2045),
    [aux_sym_signed_number_token1] = ACTIONS(2043),
    [anon_sym_SQUOTE] = ACTIONS(2045),
    [anon_sym_DQUOTE] = ACTIONS(2045),
    [sym_prefix_operator] = ACTIONS(2045),
    [aux_sym_sign_operator_token1] = ACTIONS(2043),
    [sym_hash_operator] = ACTIONS(2045),
    [anon_sym_lsl] = ACTIONS(2043),
    [anon_sym_lsr] = ACTIONS(2043),
    [anon_sym_asr] = ACTIONS(2043),
    [aux_sym_pow_operator_token1] = ACTIONS(2045),
    [anon_sym_mod] = ACTIONS(2043),
    [anon_sym_land] = ACTIONS(2043),
    [anon_sym_lor] = ACTIONS(2043),
    [anon_sym_lxor] = ACTIONS(2043),
    [aux_sym_mult_operator_token1] = ACTIONS(2043),
    [aux_sym_add_operator_token1] = ACTIONS(2043),
    [sym_concat_operator] = ACTIONS(2045),
    [sym_rel_operator] = ACTIONS(2043),
    [anon_sym_AMP_AMP] = ACTIONS(2043),
    [anon_sym_or] = ACTIONS(2043),
    [anon_sym_PIPE_PIPE] = ACTIONS(2043),
    [sym_assign_operator] = ACTIONS(2045),
    [sym__uppercase_identifier] = ACTIONS(2045),
    [aux_sym_tag_token1] = ACTIONS(2045),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(971)] = {
    [sym_attribute] = STATE(971),
    [sym__lowercase_identifier] = ACTIONS(2047),
    [anon_sym_and] = ACTIONS(2047),
    [anon_sym_TILDE] = ACTIONS(2047),
    [anon_sym_QMARK] = ACTIONS(2047),
    [anon_sym_LPAREN] = ACTIONS(2049),
    [anon_sym_RPAREN] = ACTIONS(2049),
    [anon_sym_COMMA] = ACTIONS(2049),
    [anon_sym_PIPE] = ACTIONS(2047),
    [anon_sym_COLON2] = ACTIONS(2047),
    [anon_sym_DOT] = ACTIONS(2049),
    [anon_sym_DASH_GT] = ACTIONS(2047),
    [anon_sym_LBRACK] = ACTIONS(2047),
    [anon_sym_RBRACK] = ACTIONS(2049),
    [anon_sym_true] = ACTIONS(2047),
    [anon_sym_false] = ACTIONS(2047),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_SEMI] = ACTIONS(2049),
    [anon_sym_RBRACE] = ACTIONS(2049),
    [anon_sym_constraint] = ACTIONS(2047),
    [anon_sym_val] = ACTIONS(2047),
    [anon_sym_end] = ACTIONS(2047),
    [anon_sym_with] = ACTIONS(2047),
    [anon_sym_COLON_GT] = ACTIONS(2049),
    [anon_sym_object] = ACTIONS(2047),
    [anon_sym_inherit] = ACTIONS(2047),
    [anon_sym_method] = ACTIONS(2047),
    [anon_sym_in] = ACTIONS(2047),
    [anon_sym_as] = ACTIONS(2047),
    [anon_sym_initializer] = ACTIONS(2047),
    [anon_sym_AMP] = ACTIONS(2047),
    [anon_sym_POUND] = ACTIONS(2047),
    [anon_sym_COLON_COLON] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2049),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2049),
    [anon_sym_then] = ACTIONS(2047),
    [anon_sym_else] = ACTIONS(2047),
    [anon_sym_do] = ACTIONS(2047),
    [anon_sym_done] = ACTIONS(2047),
    [anon_sym_to] = ACTIONS(2047),
    [anon_sym_downto] = ACTIONS(2047),
    [anon_sym_new] = ACTIONS(2047),
    [anon_sym_LBRACE_LT] = ACTIONS(2049),
    [anon_sym_GT_RBRACE] = ACTIONS(2049),
    [anon_sym_begin] = ACTIONS(2047),
    [sym_ocamlyacc_value] = ACTIONS(2049),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2047),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2049),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2047),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2047),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2049),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2049),
    [aux_sym_number_token1] = ACTIONS(2049),
    [aux_sym_signed_number_token1] = ACTIONS(2047),
    [anon_sym_SQUOTE] = ACTIONS(2049),
    [anon_sym_DQUOTE] = ACTIONS(2049),
    [sym_prefix_operator] = ACTIONS(2049),
    [aux_sym_sign_operator_token1] = ACTIONS(2047),
    [sym_hash_operator] = ACTIONS(2049),
    [anon_sym_lsl] = ACTIONS(2047),
    [anon_sym_lsr] = ACTIONS(2047),
    [anon_sym_asr] = ACTIONS(2047),
    [aux_sym_pow_operator_token1] = ACTIONS(2049),
    [anon_sym_mod] = ACTIONS(2047),
    [anon_sym_land] = ACTIONS(2047),
    [anon_sym_lor] = ACTIONS(2047),
    [anon_sym_lxor] = ACTIONS(2047),
    [aux_sym_mult_operator_token1] = ACTIONS(2047),
    [aux_sym_add_operator_token1] = ACTIONS(2047),
    [sym_concat_operator] = ACTIONS(2049),
    [sym_rel_operator] = ACTIONS(2047),
    [anon_sym_AMP_AMP] = ACTIONS(2047),
    [anon_sym_or] = ACTIONS(2047),
    [anon_sym_PIPE_PIPE] = ACTIONS(2047),
    [sym_assign_operator] = ACTIONS(2049),
    [sym__uppercase_identifier] = ACTIONS(2049),
    [aux_sym_tag_token1] = ACTIONS(2049),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(972)] = {
    [sym_attribute] = STATE(972),
    [sym__lowercase_identifier] = ACTIONS(2051),
    [anon_sym_and] = ACTIONS(2051),
    [anon_sym_TILDE] = ACTIONS(2051),
    [anon_sym_QMARK] = ACTIONS(2051),
    [anon_sym_LPAREN] = ACTIONS(2053),
    [anon_sym_RPAREN] = ACTIONS(2053),
    [anon_sym_COMMA] = ACTIONS(2053),
    [anon_sym_PIPE] = ACTIONS(2051),
    [anon_sym_COLON2] = ACTIONS(2051),
    [anon_sym_DOT] = ACTIONS(2053),
    [anon_sym_DASH_GT] = ACTIONS(2051),
    [anon_sym_LBRACK] = ACTIONS(2051),
    [anon_sym_RBRACK] = ACTIONS(2053),
    [anon_sym_true] = ACTIONS(2051),
    [anon_sym_false] = ACTIONS(2051),
    [anon_sym_LBRACE] = ACTIONS(2051),
    [anon_sym_SEMI] = ACTIONS(2053),
    [anon_sym_RBRACE] = ACTIONS(2053),
    [anon_sym_constraint] = ACTIONS(2051),
    [anon_sym_val] = ACTIONS(2051),
    [anon_sym_end] = ACTIONS(2051),
    [anon_sym_with] = ACTIONS(2051),
    [anon_sym_COLON_GT] = ACTIONS(2053),
    [anon_sym_object] = ACTIONS(2051),
    [anon_sym_inherit] = ACTIONS(2051),
    [anon_sym_method] = ACTIONS(2051),
    [anon_sym_in] = ACTIONS(2051),
    [anon_sym_as] = ACTIONS(2051),
    [anon_sym_initializer] = ACTIONS(2051),
    [anon_sym_AMP] = ACTIONS(2051),
    [anon_sym_POUND] = ACTIONS(2051),
    [anon_sym_COLON_COLON] = ACTIONS(2053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2053),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2053),
    [anon_sym_then] = ACTIONS(2051),
    [anon_sym_else] = ACTIONS(2051),
    [anon_sym_do] = ACTIONS(2051),
    [anon_sym_done] = ACTIONS(2051),
    [anon_sym_to] = ACTIONS(2051),
    [anon_sym_downto] = ACTIONS(2051),
    [anon_sym_new] = ACTIONS(2051),
    [anon_sym_LBRACE_LT] = ACTIONS(2053),
    [anon_sym_GT_RBRACE] = ACTIONS(2053),
    [anon_sym_begin] = ACTIONS(2051),
    [sym_ocamlyacc_value] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2051),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2053),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2051),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2051),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2053),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2053),
    [aux_sym_number_token1] = ACTIONS(2053),
    [aux_sym_signed_number_token1] = ACTIONS(2051),
    [anon_sym_SQUOTE] = ACTIONS(2053),
    [anon_sym_DQUOTE] = ACTIONS(2053),
    [sym_prefix_operator] = ACTIONS(2053),
    [aux_sym_sign_operator_token1] = ACTIONS(2051),
    [sym_hash_operator] = ACTIONS(2053),
    [anon_sym_lsl] = ACTIONS(2051),
    [anon_sym_lsr] = ACTIONS(2051),
    [anon_sym_asr] = ACTIONS(2051),
    [aux_sym_pow_operator_token1] = ACTIONS(2053),
    [anon_sym_mod] = ACTIONS(2051),
    [anon_sym_land] = ACTIONS(2051),
    [anon_sym_lor] = ACTIONS(2051),
    [anon_sym_lxor] = ACTIONS(2051),
    [aux_sym_mult_operator_token1] = ACTIONS(2051),
    [aux_sym_add_operator_token1] = ACTIONS(2051),
    [sym_concat_operator] = ACTIONS(2053),
    [sym_rel_operator] = ACTIONS(2051),
    [anon_sym_AMP_AMP] = ACTIONS(2051),
    [anon_sym_or] = ACTIONS(2051),
    [anon_sym_PIPE_PIPE] = ACTIONS(2051),
    [sym_assign_operator] = ACTIONS(2053),
    [sym__uppercase_identifier] = ACTIONS(2053),
    [aux_sym_tag_token1] = ACTIONS(2053),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(973)] = {
    [sym_attribute] = STATE(973),
    [sym__lowercase_identifier] = ACTIONS(2055),
    [anon_sym_and] = ACTIONS(2055),
    [anon_sym_TILDE] = ACTIONS(2055),
    [anon_sym_QMARK] = ACTIONS(2055),
    [anon_sym_LPAREN] = ACTIONS(2057),
    [anon_sym_RPAREN] = ACTIONS(2057),
    [anon_sym_COMMA] = ACTIONS(2057),
    [anon_sym_PIPE] = ACTIONS(2055),
    [anon_sym_COLON2] = ACTIONS(2055),
    [anon_sym_DOT] = ACTIONS(2057),
    [anon_sym_DASH_GT] = ACTIONS(2055),
    [anon_sym_LBRACK] = ACTIONS(2055),
    [anon_sym_RBRACK] = ACTIONS(2057),
    [anon_sym_true] = ACTIONS(2055),
    [anon_sym_false] = ACTIONS(2055),
    [anon_sym_LBRACE] = ACTIONS(2055),
    [anon_sym_SEMI] = ACTIONS(2057),
    [anon_sym_RBRACE] = ACTIONS(2057),
    [anon_sym_constraint] = ACTIONS(2055),
    [anon_sym_val] = ACTIONS(2055),
    [anon_sym_end] = ACTIONS(2055),
    [anon_sym_with] = ACTIONS(2055),
    [anon_sym_COLON_GT] = ACTIONS(2057),
    [anon_sym_object] = ACTIONS(2055),
    [anon_sym_inherit] = ACTIONS(2055),
    [anon_sym_method] = ACTIONS(2055),
    [anon_sym_in] = ACTIONS(2055),
    [anon_sym_as] = ACTIONS(2055),
    [anon_sym_initializer] = ACTIONS(2055),
    [anon_sym_AMP] = ACTIONS(2055),
    [anon_sym_POUND] = ACTIONS(2055),
    [anon_sym_COLON_COLON] = ACTIONS(2057),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2057),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2057),
    [anon_sym_then] = ACTIONS(2055),
    [anon_sym_else] = ACTIONS(2055),
    [anon_sym_do] = ACTIONS(2055),
    [anon_sym_done] = ACTIONS(2055),
    [anon_sym_to] = ACTIONS(2055),
    [anon_sym_downto] = ACTIONS(2055),
    [anon_sym_new] = ACTIONS(2055),
    [anon_sym_LBRACE_LT] = ACTIONS(2057),
    [anon_sym_GT_RBRACE] = ACTIONS(2057),
    [anon_sym_begin] = ACTIONS(2055),
    [sym_ocamlyacc_value] = ACTIONS(2057),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2057),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2055),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2055),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2057),
    [aux_sym_number_token1] = ACTIONS(2057),
    [aux_sym_signed_number_token1] = ACTIONS(2055),
    [anon_sym_SQUOTE] = ACTIONS(2057),
    [anon_sym_DQUOTE] = ACTIONS(2057),
    [sym_prefix_operator] = ACTIONS(2057),
    [aux_sym_sign_operator_token1] = ACTIONS(2055),
    [sym_hash_operator] = ACTIONS(2057),
    [anon_sym_lsl] = ACTIONS(2055),
    [anon_sym_lsr] = ACTIONS(2055),
    [anon_sym_asr] = ACTIONS(2055),
    [aux_sym_pow_operator_token1] = ACTIONS(2057),
    [anon_sym_mod] = ACTIONS(2055),
    [anon_sym_land] = ACTIONS(2055),
    [anon_sym_lor] = ACTIONS(2055),
    [anon_sym_lxor] = ACTIONS(2055),
    [aux_sym_mult_operator_token1] = ACTIONS(2055),
    [aux_sym_add_operator_token1] = ACTIONS(2055),
    [sym_concat_operator] = ACTIONS(2057),
    [sym_rel_operator] = ACTIONS(2055),
    [anon_sym_AMP_AMP] = ACTIONS(2055),
    [anon_sym_or] = ACTIONS(2055),
    [anon_sym_PIPE_PIPE] = ACTIONS(2055),
    [sym_assign_operator] = ACTIONS(2057),
    [sym__uppercase_identifier] = ACTIONS(2057),
    [aux_sym_tag_token1] = ACTIONS(2057),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(974)] = {
    [sym_attribute] = STATE(974),
    [sym__lowercase_identifier] = ACTIONS(2059),
    [anon_sym_and] = ACTIONS(2059),
    [anon_sym_TILDE] = ACTIONS(2059),
    [anon_sym_QMARK] = ACTIONS(2059),
    [anon_sym_LPAREN] = ACTIONS(2061),
    [anon_sym_RPAREN] = ACTIONS(2061),
    [anon_sym_COMMA] = ACTIONS(2061),
    [anon_sym_PIPE] = ACTIONS(2059),
    [anon_sym_COLON2] = ACTIONS(2059),
    [anon_sym_DOT] = ACTIONS(2061),
    [anon_sym_DASH_GT] = ACTIONS(2059),
    [anon_sym_LBRACK] = ACTIONS(2059),
    [anon_sym_RBRACK] = ACTIONS(2061),
    [anon_sym_true] = ACTIONS(2059),
    [anon_sym_false] = ACTIONS(2059),
    [anon_sym_LBRACE] = ACTIONS(2059),
    [anon_sym_SEMI] = ACTIONS(2061),
    [anon_sym_RBRACE] = ACTIONS(2061),
    [anon_sym_constraint] = ACTIONS(2059),
    [anon_sym_val] = ACTIONS(2059),
    [anon_sym_end] = ACTIONS(2059),
    [anon_sym_with] = ACTIONS(2059),
    [anon_sym_COLON_GT] = ACTIONS(2061),
    [anon_sym_object] = ACTIONS(2059),
    [anon_sym_inherit] = ACTIONS(2059),
    [anon_sym_method] = ACTIONS(2059),
    [anon_sym_in] = ACTIONS(2059),
    [anon_sym_as] = ACTIONS(2059),
    [anon_sym_initializer] = ACTIONS(2059),
    [anon_sym_AMP] = ACTIONS(2059),
    [anon_sym_POUND] = ACTIONS(2059),
    [anon_sym_COLON_COLON] = ACTIONS(2061),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2061),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2061),
    [anon_sym_then] = ACTIONS(2059),
    [anon_sym_else] = ACTIONS(2059),
    [anon_sym_do] = ACTIONS(2059),
    [anon_sym_done] = ACTIONS(2059),
    [anon_sym_to] = ACTIONS(2059),
    [anon_sym_downto] = ACTIONS(2059),
    [anon_sym_new] = ACTIONS(2059),
    [anon_sym_LBRACE_LT] = ACTIONS(2061),
    [anon_sym_GT_RBRACE] = ACTIONS(2061),
    [anon_sym_begin] = ACTIONS(2059),
    [sym_ocamlyacc_value] = ACTIONS(2061),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2059),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2061),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2061),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2061),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2059),
    [anon_sym_SQUOTE] = ACTIONS(2061),
    [anon_sym_DQUOTE] = ACTIONS(2061),
    [sym_prefix_operator] = ACTIONS(2061),
    [aux_sym_sign_operator_token1] = ACTIONS(2059),
    [sym_hash_operator] = ACTIONS(2061),
    [anon_sym_lsl] = ACTIONS(2059),
    [anon_sym_lsr] = ACTIONS(2059),
    [anon_sym_asr] = ACTIONS(2059),
    [aux_sym_pow_operator_token1] = ACTIONS(2061),
    [anon_sym_mod] = ACTIONS(2059),
    [anon_sym_land] = ACTIONS(2059),
    [anon_sym_lor] = ACTIONS(2059),
    [anon_sym_lxor] = ACTIONS(2059),
    [aux_sym_mult_operator_token1] = ACTIONS(2059),
    [aux_sym_add_operator_token1] = ACTIONS(2059),
    [sym_concat_operator] = ACTIONS(2061),
    [sym_rel_operator] = ACTIONS(2059),
    [anon_sym_AMP_AMP] = ACTIONS(2059),
    [anon_sym_or] = ACTIONS(2059),
    [anon_sym_PIPE_PIPE] = ACTIONS(2059),
    [sym_assign_operator] = ACTIONS(2061),
    [sym__uppercase_identifier] = ACTIONS(2061),
    [aux_sym_tag_token1] = ACTIONS(2061),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(975)] = {
    [sym_attribute] = STATE(975),
    [sym__lowercase_identifier] = ACTIONS(2063),
    [anon_sym_and] = ACTIONS(2063),
    [anon_sym_TILDE] = ACTIONS(2063),
    [anon_sym_QMARK] = ACTIONS(2063),
    [anon_sym_LPAREN] = ACTIONS(2065),
    [anon_sym_RPAREN] = ACTIONS(2065),
    [anon_sym_COMMA] = ACTIONS(2065),
    [anon_sym_PIPE] = ACTIONS(2063),
    [anon_sym_COLON2] = ACTIONS(2063),
    [anon_sym_DOT] = ACTIONS(2065),
    [anon_sym_DASH_GT] = ACTIONS(2063),
    [anon_sym_LBRACK] = ACTIONS(2063),
    [anon_sym_RBRACK] = ACTIONS(2065),
    [anon_sym_true] = ACTIONS(2063),
    [anon_sym_false] = ACTIONS(2063),
    [anon_sym_LBRACE] = ACTIONS(2063),
    [anon_sym_SEMI] = ACTIONS(2065),
    [anon_sym_RBRACE] = ACTIONS(2065),
    [anon_sym_constraint] = ACTIONS(2063),
    [anon_sym_val] = ACTIONS(2063),
    [anon_sym_end] = ACTIONS(2063),
    [anon_sym_with] = ACTIONS(2063),
    [anon_sym_COLON_GT] = ACTIONS(2065),
    [anon_sym_object] = ACTIONS(2063),
    [anon_sym_inherit] = ACTIONS(2063),
    [anon_sym_method] = ACTIONS(2063),
    [anon_sym_in] = ACTIONS(2063),
    [anon_sym_as] = ACTIONS(2063),
    [anon_sym_initializer] = ACTIONS(2063),
    [anon_sym_AMP] = ACTIONS(2063),
    [anon_sym_POUND] = ACTIONS(2063),
    [anon_sym_COLON_COLON] = ACTIONS(2065),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2065),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2065),
    [anon_sym_then] = ACTIONS(2063),
    [anon_sym_else] = ACTIONS(2063),
    [anon_sym_do] = ACTIONS(2063),
    [anon_sym_done] = ACTIONS(2063),
    [anon_sym_to] = ACTIONS(2063),
    [anon_sym_downto] = ACTIONS(2063),
    [anon_sym_new] = ACTIONS(2063),
    [anon_sym_LBRACE_LT] = ACTIONS(2065),
    [anon_sym_GT_RBRACE] = ACTIONS(2065),
    [anon_sym_begin] = ACTIONS(2063),
    [sym_ocamlyacc_value] = ACTIONS(2065),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2063),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2065),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2063),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2063),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2065),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2065),
    [aux_sym_number_token1] = ACTIONS(2065),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2065),
    [sym_prefix_operator] = ACTIONS(2065),
    [aux_sym_sign_operator_token1] = ACTIONS(2063),
    [sym_hash_operator] = ACTIONS(2065),
    [anon_sym_lsl] = ACTIONS(2063),
    [anon_sym_lsr] = ACTIONS(2063),
    [anon_sym_asr] = ACTIONS(2063),
    [aux_sym_pow_operator_token1] = ACTIONS(2065),
    [anon_sym_mod] = ACTIONS(2063),
    [anon_sym_land] = ACTIONS(2063),
    [anon_sym_lor] = ACTIONS(2063),
    [anon_sym_lxor] = ACTIONS(2063),
    [aux_sym_mult_operator_token1] = ACTIONS(2063),
    [aux_sym_add_operator_token1] = ACTIONS(2063),
    [sym_concat_operator] = ACTIONS(2065),
    [sym_rel_operator] = ACTIONS(2063),
    [anon_sym_AMP_AMP] = ACTIONS(2063),
    [anon_sym_or] = ACTIONS(2063),
    [anon_sym_PIPE_PIPE] = ACTIONS(2063),
    [sym_assign_operator] = ACTIONS(2065),
    [sym__uppercase_identifier] = ACTIONS(2065),
    [aux_sym_tag_token1] = ACTIONS(2065),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(976)] = {
    [sym_attribute] = STATE(976),
    [sym__lowercase_identifier] = ACTIONS(2067),
    [anon_sym_and] = ACTIONS(2067),
    [anon_sym_TILDE] = ACTIONS(2067),
    [anon_sym_QMARK] = ACTIONS(2067),
    [anon_sym_LPAREN] = ACTIONS(2069),
    [anon_sym_RPAREN] = ACTIONS(2069),
    [anon_sym_COMMA] = ACTIONS(2069),
    [anon_sym_PIPE] = ACTIONS(2067),
    [anon_sym_COLON2] = ACTIONS(2067),
    [anon_sym_DOT] = ACTIONS(2069),
    [anon_sym_DASH_GT] = ACTIONS(2067),
    [anon_sym_LBRACK] = ACTIONS(2067),
    [anon_sym_RBRACK] = ACTIONS(2069),
    [anon_sym_true] = ACTIONS(2067),
    [anon_sym_false] = ACTIONS(2067),
    [anon_sym_LBRACE] = ACTIONS(2067),
    [anon_sym_SEMI] = ACTIONS(2069),
    [anon_sym_RBRACE] = ACTIONS(2069),
    [anon_sym_constraint] = ACTIONS(2067),
    [anon_sym_val] = ACTIONS(2067),
    [anon_sym_end] = ACTIONS(2067),
    [anon_sym_with] = ACTIONS(2067),
    [anon_sym_COLON_GT] = ACTIONS(2069),
    [anon_sym_object] = ACTIONS(2067),
    [anon_sym_inherit] = ACTIONS(2067),
    [anon_sym_method] = ACTIONS(2067),
    [anon_sym_in] = ACTIONS(2067),
    [anon_sym_as] = ACTIONS(2067),
    [anon_sym_initializer] = ACTIONS(2067),
    [anon_sym_AMP] = ACTIONS(2067),
    [anon_sym_POUND] = ACTIONS(2067),
    [anon_sym_COLON_COLON] = ACTIONS(2069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2069),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2069),
    [anon_sym_then] = ACTIONS(2067),
    [anon_sym_else] = ACTIONS(2067),
    [anon_sym_do] = ACTIONS(2067),
    [anon_sym_done] = ACTIONS(2067),
    [anon_sym_to] = ACTIONS(2067),
    [anon_sym_downto] = ACTIONS(2067),
    [anon_sym_new] = ACTIONS(2067),
    [anon_sym_LBRACE_LT] = ACTIONS(2069),
    [anon_sym_GT_RBRACE] = ACTIONS(2069),
    [anon_sym_begin] = ACTIONS(2067),
    [sym_ocamlyacc_value] = ACTIONS(2069),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2067),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2069),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2067),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2067),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2069),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2069),
    [aux_sym_number_token1] = ACTIONS(2069),
    [aux_sym_signed_number_token1] = ACTIONS(2067),
    [anon_sym_SQUOTE] = ACTIONS(2069),
    [anon_sym_DQUOTE] = ACTIONS(2069),
    [sym_prefix_operator] = ACTIONS(2069),
    [aux_sym_sign_operator_token1] = ACTIONS(2067),
    [sym_hash_operator] = ACTIONS(2069),
    [anon_sym_lsl] = ACTIONS(2067),
    [anon_sym_lsr] = ACTIONS(2067),
    [anon_sym_asr] = ACTIONS(2067),
    [aux_sym_pow_operator_token1] = ACTIONS(2069),
    [anon_sym_mod] = ACTIONS(2067),
    [anon_sym_land] = ACTIONS(2067),
    [anon_sym_lor] = ACTIONS(2067),
    [anon_sym_lxor] = ACTIONS(2067),
    [aux_sym_mult_operator_token1] = ACTIONS(2067),
    [aux_sym_add_operator_token1] = ACTIONS(2067),
    [sym_concat_operator] = ACTIONS(2069),
    [sym_rel_operator] = ACTIONS(2067),
    [anon_sym_AMP_AMP] = ACTIONS(2067),
    [anon_sym_or] = ACTIONS(2067),
    [anon_sym_PIPE_PIPE] = ACTIONS(2067),
    [sym_assign_operator] = ACTIONS(2069),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_tag_token1] = ACTIONS(2069),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(977)] = {
    [sym_attribute] = STATE(977),
    [sym__lowercase_identifier] = ACTIONS(2071),
    [anon_sym_and] = ACTIONS(2071),
    [anon_sym_TILDE] = ACTIONS(2071),
    [anon_sym_QMARK] = ACTIONS(2071),
    [anon_sym_LPAREN] = ACTIONS(2073),
    [anon_sym_RPAREN] = ACTIONS(2073),
    [anon_sym_COMMA] = ACTIONS(2073),
    [anon_sym_PIPE] = ACTIONS(2071),
    [anon_sym_COLON2] = ACTIONS(2071),
    [anon_sym_DOT] = ACTIONS(2073),
    [anon_sym_DASH_GT] = ACTIONS(2071),
    [anon_sym_LBRACK] = ACTIONS(2071),
    [anon_sym_RBRACK] = ACTIONS(2073),
    [anon_sym_true] = ACTIONS(2071),
    [anon_sym_false] = ACTIONS(2071),
    [anon_sym_LBRACE] = ACTIONS(2071),
    [anon_sym_SEMI] = ACTIONS(2073),
    [anon_sym_RBRACE] = ACTIONS(2073),
    [anon_sym_constraint] = ACTIONS(2071),
    [anon_sym_val] = ACTIONS(2071),
    [anon_sym_end] = ACTIONS(2071),
    [anon_sym_with] = ACTIONS(2071),
    [anon_sym_COLON_GT] = ACTIONS(2073),
    [anon_sym_object] = ACTIONS(2071),
    [anon_sym_inherit] = ACTIONS(2071),
    [anon_sym_method] = ACTIONS(2071),
    [anon_sym_in] = ACTIONS(2071),
    [anon_sym_as] = ACTIONS(2071),
    [anon_sym_initializer] = ACTIONS(2071),
    [anon_sym_AMP] = ACTIONS(2071),
    [anon_sym_POUND] = ACTIONS(2071),
    [anon_sym_COLON_COLON] = ACTIONS(2073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2073),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2073),
    [anon_sym_then] = ACTIONS(2071),
    [anon_sym_else] = ACTIONS(2071),
    [anon_sym_do] = ACTIONS(2071),
    [anon_sym_done] = ACTIONS(2071),
    [anon_sym_to] = ACTIONS(2071),
    [anon_sym_downto] = ACTIONS(2071),
    [anon_sym_new] = ACTIONS(2071),
    [anon_sym_LBRACE_LT] = ACTIONS(2073),
    [anon_sym_GT_RBRACE] = ACTIONS(2073),
    [anon_sym_begin] = ACTIONS(2071),
    [sym_ocamlyacc_value] = ACTIONS(2073),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2071),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2073),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2071),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2071),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2073),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2073),
    [aux_sym_number_token1] = ACTIONS(2073),
    [aux_sym_signed_number_token1] = ACTIONS(2071),
    [anon_sym_SQUOTE] = ACTIONS(2073),
    [anon_sym_DQUOTE] = ACTIONS(2073),
    [sym_prefix_operator] = ACTIONS(2073),
    [aux_sym_sign_operator_token1] = ACTIONS(2071),
    [sym_hash_operator] = ACTIONS(2073),
    [anon_sym_lsl] = ACTIONS(2071),
    [anon_sym_lsr] = ACTIONS(2071),
    [anon_sym_asr] = ACTIONS(2071),
    [aux_sym_pow_operator_token1] = ACTIONS(2073),
    [anon_sym_mod] = ACTIONS(2071),
    [anon_sym_land] = ACTIONS(2071),
    [anon_sym_lor] = ACTIONS(2071),
    [anon_sym_lxor] = ACTIONS(2071),
    [aux_sym_mult_operator_token1] = ACTIONS(2071),
    [aux_sym_add_operator_token1] = ACTIONS(2071),
    [sym_concat_operator] = ACTIONS(2073),
    [sym_rel_operator] = ACTIONS(2071),
    [anon_sym_AMP_AMP] = ACTIONS(2071),
    [anon_sym_or] = ACTIONS(2071),
    [anon_sym_PIPE_PIPE] = ACTIONS(2071),
    [sym_assign_operator] = ACTIONS(2073),
    [sym__uppercase_identifier] = ACTIONS(2073),
    [aux_sym_tag_token1] = ACTIONS(2073),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(978)] = {
    [sym_attribute] = STATE(978),
    [sym__lowercase_identifier] = ACTIONS(2075),
    [anon_sym_and] = ACTIONS(2075),
    [anon_sym_TILDE] = ACTIONS(2075),
    [anon_sym_QMARK] = ACTIONS(2075),
    [anon_sym_LPAREN] = ACTIONS(2077),
    [anon_sym_RPAREN] = ACTIONS(2077),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_PIPE] = ACTIONS(2075),
    [anon_sym_COLON2] = ACTIONS(2075),
    [anon_sym_DOT] = ACTIONS(2077),
    [anon_sym_DASH_GT] = ACTIONS(2075),
    [anon_sym_LBRACK] = ACTIONS(2075),
    [anon_sym_RBRACK] = ACTIONS(2077),
    [anon_sym_true] = ACTIONS(2075),
    [anon_sym_false] = ACTIONS(2075),
    [anon_sym_LBRACE] = ACTIONS(2075),
    [anon_sym_SEMI] = ACTIONS(2077),
    [anon_sym_RBRACE] = ACTIONS(2077),
    [anon_sym_constraint] = ACTIONS(2075),
    [anon_sym_val] = ACTIONS(2075),
    [anon_sym_end] = ACTIONS(2075),
    [anon_sym_with] = ACTIONS(2075),
    [anon_sym_COLON_GT] = ACTIONS(2077),
    [anon_sym_object] = ACTIONS(2075),
    [anon_sym_inherit] = ACTIONS(2075),
    [anon_sym_method] = ACTIONS(2075),
    [anon_sym_in] = ACTIONS(2075),
    [anon_sym_as] = ACTIONS(2075),
    [anon_sym_initializer] = ACTIONS(2075),
    [anon_sym_AMP] = ACTIONS(2075),
    [anon_sym_POUND] = ACTIONS(2075),
    [anon_sym_COLON_COLON] = ACTIONS(2077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2077),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2077),
    [anon_sym_then] = ACTIONS(2075),
    [anon_sym_else] = ACTIONS(2075),
    [anon_sym_do] = ACTIONS(2075),
    [anon_sym_done] = ACTIONS(2075),
    [anon_sym_to] = ACTIONS(2075),
    [anon_sym_downto] = ACTIONS(2075),
    [anon_sym_new] = ACTIONS(2075),
    [anon_sym_LBRACE_LT] = ACTIONS(2077),
    [anon_sym_GT_RBRACE] = ACTIONS(2077),
    [anon_sym_begin] = ACTIONS(2075),
    [sym_ocamlyacc_value] = ACTIONS(2077),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2075),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2077),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2075),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2075),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2077),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2077),
    [aux_sym_number_token1] = ACTIONS(2077),
    [aux_sym_signed_number_token1] = ACTIONS(2075),
    [anon_sym_SQUOTE] = ACTIONS(2077),
    [anon_sym_DQUOTE] = ACTIONS(2077),
    [sym_prefix_operator] = ACTIONS(2077),
    [aux_sym_sign_operator_token1] = ACTIONS(2075),
    [sym_hash_operator] = ACTIONS(2077),
    [anon_sym_lsl] = ACTIONS(2075),
    [anon_sym_lsr] = ACTIONS(2075),
    [anon_sym_asr] = ACTIONS(2075),
    [aux_sym_pow_operator_token1] = ACTIONS(2077),
    [anon_sym_mod] = ACTIONS(2075),
    [anon_sym_land] = ACTIONS(2075),
    [anon_sym_lor] = ACTIONS(2075),
    [anon_sym_lxor] = ACTIONS(2075),
    [aux_sym_mult_operator_token1] = ACTIONS(2075),
    [aux_sym_add_operator_token1] = ACTIONS(2075),
    [sym_concat_operator] = ACTIONS(2077),
    [sym_rel_operator] = ACTIONS(2075),
    [anon_sym_AMP_AMP] = ACTIONS(2075),
    [anon_sym_or] = ACTIONS(2075),
    [anon_sym_PIPE_PIPE] = ACTIONS(2075),
    [sym_assign_operator] = ACTIONS(2077),
    [sym__uppercase_identifier] = ACTIONS(2077),
    [aux_sym_tag_token1] = ACTIONS(2077),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(979)] = {
    [sym_attribute] = STATE(979),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_and] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_RPAREN] = ACTIONS(2081),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_PIPE] = ACTIONS(2079),
    [anon_sym_COLON2] = ACTIONS(2079),
    [anon_sym_DOT] = ACTIONS(2081),
    [anon_sym_DASH_GT] = ACTIONS(2079),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_RBRACK] = ACTIONS(2081),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_SEMI] = ACTIONS(2081),
    [anon_sym_RBRACE] = ACTIONS(2081),
    [anon_sym_constraint] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_end] = ACTIONS(2079),
    [anon_sym_with] = ACTIONS(2079),
    [anon_sym_COLON_GT] = ACTIONS(2081),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_inherit] = ACTIONS(2079),
    [anon_sym_method] = ACTIONS(2079),
    [anon_sym_in] = ACTIONS(2079),
    [anon_sym_as] = ACTIONS(2079),
    [anon_sym_initializer] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_POUND] = ACTIONS(2079),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2081),
    [anon_sym_then] = ACTIONS(2079),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_do] = ACTIONS(2079),
    [anon_sym_done] = ACTIONS(2079),
    [anon_sym_to] = ACTIONS(2079),
    [anon_sym_downto] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_GT_RBRACE] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [sym_hash_operator] = ACTIONS(2081),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(980)] = {
    [sym_attribute] = STATE(980),
    [sym__lowercase_identifier] = ACTIONS(2083),
    [anon_sym_and] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(2083),
    [anon_sym_QMARK] = ACTIONS(2083),
    [anon_sym_LPAREN] = ACTIONS(2085),
    [anon_sym_RPAREN] = ACTIONS(2085),
    [anon_sym_COMMA] = ACTIONS(2085),
    [anon_sym_PIPE] = ACTIONS(2083),
    [anon_sym_COLON2] = ACTIONS(2083),
    [anon_sym_DOT] = ACTIONS(2085),
    [anon_sym_DASH_GT] = ACTIONS(2083),
    [anon_sym_LBRACK] = ACTIONS(2083),
    [anon_sym_RBRACK] = ACTIONS(2085),
    [anon_sym_true] = ACTIONS(2083),
    [anon_sym_false] = ACTIONS(2083),
    [anon_sym_LBRACE] = ACTIONS(2083),
    [anon_sym_SEMI] = ACTIONS(2085),
    [anon_sym_RBRACE] = ACTIONS(2085),
    [anon_sym_constraint] = ACTIONS(2083),
    [anon_sym_val] = ACTIONS(2083),
    [anon_sym_end] = ACTIONS(2083),
    [anon_sym_with] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(2085),
    [anon_sym_object] = ACTIONS(2083),
    [anon_sym_inherit] = ACTIONS(2083),
    [anon_sym_method] = ACTIONS(2083),
    [anon_sym_in] = ACTIONS(2083),
    [anon_sym_as] = ACTIONS(2083),
    [anon_sym_initializer] = ACTIONS(2083),
    [anon_sym_AMP] = ACTIONS(2083),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(2085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2085),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2085),
    [anon_sym_then] = ACTIONS(2083),
    [anon_sym_else] = ACTIONS(2083),
    [anon_sym_do] = ACTIONS(2083),
    [anon_sym_done] = ACTIONS(2083),
    [anon_sym_to] = ACTIONS(2083),
    [anon_sym_downto] = ACTIONS(2083),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_LBRACE_LT] = ACTIONS(2085),
    [anon_sym_GT_RBRACE] = ACTIONS(2085),
    [anon_sym_begin] = ACTIONS(2083),
    [sym_ocamlyacc_value] = ACTIONS(2085),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2083),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2085),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2083),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2083),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2085),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2085),
    [aux_sym_number_token1] = ACTIONS(2085),
    [aux_sym_signed_number_token1] = ACTIONS(2083),
    [anon_sym_SQUOTE] = ACTIONS(2085),
    [anon_sym_DQUOTE] = ACTIONS(2085),
    [sym_prefix_operator] = ACTIONS(2085),
    [aux_sym_sign_operator_token1] = ACTIONS(2083),
    [sym_hash_operator] = ACTIONS(2085),
    [anon_sym_lsl] = ACTIONS(2083),
    [anon_sym_lsr] = ACTIONS(2083),
    [anon_sym_asr] = ACTIONS(2083),
    [aux_sym_pow_operator_token1] = ACTIONS(2085),
    [anon_sym_mod] = ACTIONS(2083),
    [anon_sym_land] = ACTIONS(2083),
    [anon_sym_lor] = ACTIONS(2083),
    [anon_sym_lxor] = ACTIONS(2083),
    [aux_sym_mult_operator_token1] = ACTIONS(2083),
    [aux_sym_add_operator_token1] = ACTIONS(2083),
    [sym_concat_operator] = ACTIONS(2085),
    [sym_rel_operator] = ACTIONS(2083),
    [anon_sym_AMP_AMP] = ACTIONS(2083),
    [anon_sym_or] = ACTIONS(2083),
    [anon_sym_PIPE_PIPE] = ACTIONS(2083),
    [sym_assign_operator] = ACTIONS(2085),
    [sym__uppercase_identifier] = ACTIONS(2085),
    [aux_sym_tag_token1] = ACTIONS(2085),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(981)] = {
    [sym_attribute] = STATE(981),
    [sym__lowercase_identifier] = ACTIONS(2087),
    [anon_sym_and] = ACTIONS(2087),
    [anon_sym_TILDE] = ACTIONS(2087),
    [anon_sym_QMARK] = ACTIONS(2087),
    [anon_sym_LPAREN] = ACTIONS(2089),
    [anon_sym_RPAREN] = ACTIONS(2089),
    [anon_sym_COMMA] = ACTIONS(2089),
    [anon_sym_PIPE] = ACTIONS(2087),
    [anon_sym_COLON2] = ACTIONS(2087),
    [anon_sym_DOT] = ACTIONS(2089),
    [anon_sym_DASH_GT] = ACTIONS(2087),
    [anon_sym_LBRACK] = ACTIONS(2087),
    [anon_sym_RBRACK] = ACTIONS(2089),
    [anon_sym_true] = ACTIONS(2087),
    [anon_sym_false] = ACTIONS(2087),
    [anon_sym_LBRACE] = ACTIONS(2087),
    [anon_sym_SEMI] = ACTIONS(2089),
    [anon_sym_RBRACE] = ACTIONS(2089),
    [anon_sym_constraint] = ACTIONS(2087),
    [anon_sym_val] = ACTIONS(2087),
    [anon_sym_end] = ACTIONS(2087),
    [anon_sym_with] = ACTIONS(2087),
    [anon_sym_COLON_GT] = ACTIONS(2089),
    [anon_sym_object] = ACTIONS(2087),
    [anon_sym_inherit] = ACTIONS(2087),
    [anon_sym_method] = ACTIONS(2087),
    [anon_sym_in] = ACTIONS(2087),
    [anon_sym_as] = ACTIONS(2087),
    [anon_sym_initializer] = ACTIONS(2087),
    [anon_sym_AMP] = ACTIONS(2087),
    [anon_sym_POUND] = ACTIONS(2087),
    [anon_sym_COLON_COLON] = ACTIONS(2089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2089),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2089),
    [anon_sym_then] = ACTIONS(2087),
    [anon_sym_else] = ACTIONS(2087),
    [anon_sym_do] = ACTIONS(2087),
    [anon_sym_done] = ACTIONS(2087),
    [anon_sym_to] = ACTIONS(2087),
    [anon_sym_downto] = ACTIONS(2087),
    [anon_sym_new] = ACTIONS(2087),
    [anon_sym_LBRACE_LT] = ACTIONS(2089),
    [anon_sym_GT_RBRACE] = ACTIONS(2089),
    [anon_sym_begin] = ACTIONS(2087),
    [sym_ocamlyacc_value] = ACTIONS(2089),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2087),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2089),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2087),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2087),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2089),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2089),
    [aux_sym_number_token1] = ACTIONS(2089),
    [aux_sym_signed_number_token1] = ACTIONS(2087),
    [anon_sym_SQUOTE] = ACTIONS(2089),
    [anon_sym_DQUOTE] = ACTIONS(2089),
    [sym_prefix_operator] = ACTIONS(2089),
    [aux_sym_sign_operator_token1] = ACTIONS(2087),
    [sym_hash_operator] = ACTIONS(2089),
    [anon_sym_lsl] = ACTIONS(2087),
    [anon_sym_lsr] = ACTIONS(2087),
    [anon_sym_asr] = ACTIONS(2087),
    [aux_sym_pow_operator_token1] = ACTIONS(2089),
    [anon_sym_mod] = ACTIONS(2087),
    [anon_sym_land] = ACTIONS(2087),
    [anon_sym_lor] = ACTIONS(2087),
    [anon_sym_lxor] = ACTIONS(2087),
    [aux_sym_mult_operator_token1] = ACTIONS(2087),
    [aux_sym_add_operator_token1] = ACTIONS(2087),
    [sym_concat_operator] = ACTIONS(2089),
    [sym_rel_operator] = ACTIONS(2087),
    [anon_sym_AMP_AMP] = ACTIONS(2087),
    [anon_sym_or] = ACTIONS(2087),
    [anon_sym_PIPE_PIPE] = ACTIONS(2087),
    [sym_assign_operator] = ACTIONS(2089),
    [sym__uppercase_identifier] = ACTIONS(2089),
    [aux_sym_tag_token1] = ACTIONS(2089),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(982)] = {
    [sym_attribute] = STATE(982),
    [sym__lowercase_identifier] = ACTIONS(2091),
    [anon_sym_and] = ACTIONS(2091),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_RPAREN] = ACTIONS(2093),
    [anon_sym_COMMA] = ACTIONS(2093),
    [anon_sym_PIPE] = ACTIONS(2091),
    [anon_sym_COLON2] = ACTIONS(2091),
    [anon_sym_DOT] = ACTIONS(2093),
    [anon_sym_DASH_GT] = ACTIONS(2091),
    [anon_sym_LBRACK] = ACTIONS(2091),
    [anon_sym_RBRACK] = ACTIONS(2093),
    [anon_sym_true] = ACTIONS(2091),
    [anon_sym_false] = ACTIONS(2091),
    [anon_sym_LBRACE] = ACTIONS(2091),
    [anon_sym_SEMI] = ACTIONS(2093),
    [anon_sym_RBRACE] = ACTIONS(2093),
    [anon_sym_constraint] = ACTIONS(2091),
    [anon_sym_val] = ACTIONS(2091),
    [anon_sym_end] = ACTIONS(2091),
    [anon_sym_with] = ACTIONS(2091),
    [anon_sym_COLON_GT] = ACTIONS(2093),
    [anon_sym_object] = ACTIONS(2091),
    [anon_sym_inherit] = ACTIONS(2091),
    [anon_sym_method] = ACTIONS(2091),
    [anon_sym_in] = ACTIONS(2091),
    [anon_sym_as] = ACTIONS(2091),
    [anon_sym_initializer] = ACTIONS(2091),
    [anon_sym_AMP] = ACTIONS(2091),
    [anon_sym_POUND] = ACTIONS(2091),
    [anon_sym_COLON_COLON] = ACTIONS(2093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2093),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2093),
    [anon_sym_then] = ACTIONS(2091),
    [anon_sym_else] = ACTIONS(2091),
    [anon_sym_do] = ACTIONS(2091),
    [anon_sym_done] = ACTIONS(2091),
    [anon_sym_to] = ACTIONS(2091),
    [anon_sym_downto] = ACTIONS(2091),
    [anon_sym_new] = ACTIONS(2091),
    [anon_sym_LBRACE_LT] = ACTIONS(2093),
    [anon_sym_GT_RBRACE] = ACTIONS(2093),
    [anon_sym_begin] = ACTIONS(2091),
    [sym_ocamlyacc_value] = ACTIONS(2093),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2091),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2093),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2091),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2091),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2093),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2093),
    [aux_sym_number_token1] = ACTIONS(2093),
    [aux_sym_signed_number_token1] = ACTIONS(2091),
    [anon_sym_SQUOTE] = ACTIONS(2093),
    [anon_sym_DQUOTE] = ACTIONS(2093),
    [sym_prefix_operator] = ACTIONS(2093),
    [aux_sym_sign_operator_token1] = ACTIONS(2091),
    [sym_hash_operator] = ACTIONS(2093),
    [anon_sym_lsl] = ACTIONS(2091),
    [anon_sym_lsr] = ACTIONS(2091),
    [anon_sym_asr] = ACTIONS(2091),
    [aux_sym_pow_operator_token1] = ACTIONS(2093),
    [anon_sym_mod] = ACTIONS(2091),
    [anon_sym_land] = ACTIONS(2091),
    [anon_sym_lor] = ACTIONS(2091),
    [anon_sym_lxor] = ACTIONS(2091),
    [aux_sym_mult_operator_token1] = ACTIONS(2091),
    [aux_sym_add_operator_token1] = ACTIONS(2091),
    [sym_concat_operator] = ACTIONS(2093),
    [sym_rel_operator] = ACTIONS(2091),
    [anon_sym_AMP_AMP] = ACTIONS(2091),
    [anon_sym_or] = ACTIONS(2091),
    [anon_sym_PIPE_PIPE] = ACTIONS(2091),
    [sym_assign_operator] = ACTIONS(2093),
    [sym__uppercase_identifier] = ACTIONS(2093),
    [aux_sym_tag_token1] = ACTIONS(2093),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(983)] = {
    [sym_attribute] = STATE(983),
    [sym__lowercase_identifier] = ACTIONS(2095),
    [anon_sym_and] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(2095),
    [anon_sym_QMARK] = ACTIONS(2095),
    [anon_sym_LPAREN] = ACTIONS(2097),
    [anon_sym_RPAREN] = ACTIONS(2097),
    [anon_sym_COMMA] = ACTIONS(2097),
    [anon_sym_PIPE] = ACTIONS(2095),
    [anon_sym_COLON2] = ACTIONS(2095),
    [anon_sym_DOT] = ACTIONS(2097),
    [anon_sym_DASH_GT] = ACTIONS(2095),
    [anon_sym_LBRACK] = ACTIONS(2095),
    [anon_sym_RBRACK] = ACTIONS(2097),
    [anon_sym_true] = ACTIONS(2095),
    [anon_sym_false] = ACTIONS(2095),
    [anon_sym_LBRACE] = ACTIONS(2095),
    [anon_sym_SEMI] = ACTIONS(2097),
    [anon_sym_RBRACE] = ACTIONS(2097),
    [anon_sym_constraint] = ACTIONS(2095),
    [anon_sym_val] = ACTIONS(2095),
    [anon_sym_end] = ACTIONS(2095),
    [anon_sym_with] = ACTIONS(2095),
    [anon_sym_COLON_GT] = ACTIONS(2097),
    [anon_sym_object] = ACTIONS(2095),
    [anon_sym_inherit] = ACTIONS(2095),
    [anon_sym_method] = ACTIONS(2095),
    [anon_sym_in] = ACTIONS(2095),
    [anon_sym_as] = ACTIONS(2095),
    [anon_sym_initializer] = ACTIONS(2095),
    [anon_sym_AMP] = ACTIONS(2095),
    [anon_sym_POUND] = ACTIONS(2095),
    [anon_sym_COLON_COLON] = ACTIONS(2097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2097),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2097),
    [anon_sym_then] = ACTIONS(2095),
    [anon_sym_else] = ACTIONS(2095),
    [anon_sym_do] = ACTIONS(2095),
    [anon_sym_done] = ACTIONS(2095),
    [anon_sym_to] = ACTIONS(2095),
    [anon_sym_downto] = ACTIONS(2095),
    [anon_sym_new] = ACTIONS(2095),
    [anon_sym_LBRACE_LT] = ACTIONS(2097),
    [anon_sym_GT_RBRACE] = ACTIONS(2097),
    [anon_sym_begin] = ACTIONS(2095),
    [sym_ocamlyacc_value] = ACTIONS(2097),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2095),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2097),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2095),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2095),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2097),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2097),
    [aux_sym_number_token1] = ACTIONS(2097),
    [aux_sym_signed_number_token1] = ACTIONS(2095),
    [anon_sym_SQUOTE] = ACTIONS(2097),
    [anon_sym_DQUOTE] = ACTIONS(2097),
    [sym_prefix_operator] = ACTIONS(2097),
    [aux_sym_sign_operator_token1] = ACTIONS(2095),
    [sym_hash_operator] = ACTIONS(2097),
    [anon_sym_lsl] = ACTIONS(2095),
    [anon_sym_lsr] = ACTIONS(2095),
    [anon_sym_asr] = ACTIONS(2095),
    [aux_sym_pow_operator_token1] = ACTIONS(2097),
    [anon_sym_mod] = ACTIONS(2095),
    [anon_sym_land] = ACTIONS(2095),
    [anon_sym_lor] = ACTIONS(2095),
    [anon_sym_lxor] = ACTIONS(2095),
    [aux_sym_mult_operator_token1] = ACTIONS(2095),
    [aux_sym_add_operator_token1] = ACTIONS(2095),
    [sym_concat_operator] = ACTIONS(2097),
    [sym_rel_operator] = ACTIONS(2095),
    [anon_sym_AMP_AMP] = ACTIONS(2095),
    [anon_sym_or] = ACTIONS(2095),
    [anon_sym_PIPE_PIPE] = ACTIONS(2095),
    [sym_assign_operator] = ACTIONS(2097),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_tag_token1] = ACTIONS(2097),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(984)] = {
    [sym_attribute] = STATE(984),
    [sym__lowercase_identifier] = ACTIONS(2099),
    [anon_sym_and] = ACTIONS(2099),
    [anon_sym_TILDE] = ACTIONS(2099),
    [anon_sym_QMARK] = ACTIONS(2099),
    [anon_sym_LPAREN] = ACTIONS(2101),
    [anon_sym_RPAREN] = ACTIONS(2101),
    [anon_sym_COMMA] = ACTIONS(2101),
    [anon_sym_PIPE] = ACTIONS(2099),
    [anon_sym_COLON2] = ACTIONS(2099),
    [anon_sym_DOT] = ACTIONS(2101),
    [anon_sym_DASH_GT] = ACTIONS(2099),
    [anon_sym_LBRACK] = ACTIONS(2099),
    [anon_sym_RBRACK] = ACTIONS(2101),
    [anon_sym_true] = ACTIONS(2099),
    [anon_sym_false] = ACTIONS(2099),
    [anon_sym_LBRACE] = ACTIONS(2099),
    [anon_sym_SEMI] = ACTIONS(2101),
    [anon_sym_RBRACE] = ACTIONS(2101),
    [anon_sym_constraint] = ACTIONS(2099),
    [anon_sym_val] = ACTIONS(2099),
    [anon_sym_end] = ACTIONS(2099),
    [anon_sym_with] = ACTIONS(2099),
    [anon_sym_COLON_GT] = ACTIONS(2101),
    [anon_sym_object] = ACTIONS(2099),
    [anon_sym_inherit] = ACTIONS(2099),
    [anon_sym_method] = ACTIONS(2099),
    [anon_sym_in] = ACTIONS(2099),
    [anon_sym_as] = ACTIONS(2099),
    [anon_sym_initializer] = ACTIONS(2099),
    [anon_sym_AMP] = ACTIONS(2099),
    [anon_sym_POUND] = ACTIONS(2099),
    [anon_sym_COLON_COLON] = ACTIONS(2101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2101),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2101),
    [anon_sym_then] = ACTIONS(2099),
    [anon_sym_else] = ACTIONS(2099),
    [anon_sym_do] = ACTIONS(2099),
    [anon_sym_done] = ACTIONS(2099),
    [anon_sym_to] = ACTIONS(2099),
    [anon_sym_downto] = ACTIONS(2099),
    [anon_sym_new] = ACTIONS(2099),
    [anon_sym_LBRACE_LT] = ACTIONS(2101),
    [anon_sym_GT_RBRACE] = ACTIONS(2101),
    [anon_sym_begin] = ACTIONS(2099),
    [sym_ocamlyacc_value] = ACTIONS(2101),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2099),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2101),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2099),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2099),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2101),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2101),
    [aux_sym_number_token1] = ACTIONS(2101),
    [aux_sym_signed_number_token1] = ACTIONS(2099),
    [anon_sym_SQUOTE] = ACTIONS(2101),
    [anon_sym_DQUOTE] = ACTIONS(2101),
    [sym_prefix_operator] = ACTIONS(2101),
    [aux_sym_sign_operator_token1] = ACTIONS(2099),
    [sym_hash_operator] = ACTIONS(2101),
    [anon_sym_lsl] = ACTIONS(2099),
    [anon_sym_lsr] = ACTIONS(2099),
    [anon_sym_asr] = ACTIONS(2099),
    [aux_sym_pow_operator_token1] = ACTIONS(2101),
    [anon_sym_mod] = ACTIONS(2099),
    [anon_sym_land] = ACTIONS(2099),
    [anon_sym_lor] = ACTIONS(2099),
    [anon_sym_lxor] = ACTIONS(2099),
    [aux_sym_mult_operator_token1] = ACTIONS(2099),
    [aux_sym_add_operator_token1] = ACTIONS(2099),
    [sym_concat_operator] = ACTIONS(2101),
    [sym_rel_operator] = ACTIONS(2099),
    [anon_sym_AMP_AMP] = ACTIONS(2099),
    [anon_sym_or] = ACTIONS(2099),
    [anon_sym_PIPE_PIPE] = ACTIONS(2099),
    [sym_assign_operator] = ACTIONS(2101),
    [sym__uppercase_identifier] = ACTIONS(2101),
    [aux_sym_tag_token1] = ACTIONS(2101),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(985)] = {
    [sym_attribute] = STATE(985),
    [sym__lowercase_identifier] = ACTIONS(2103),
    [anon_sym_and] = ACTIONS(2103),
    [anon_sym_TILDE] = ACTIONS(2103),
    [anon_sym_QMARK] = ACTIONS(2103),
    [anon_sym_LPAREN] = ACTIONS(2105),
    [anon_sym_RPAREN] = ACTIONS(2105),
    [anon_sym_COMMA] = ACTIONS(2105),
    [anon_sym_PIPE] = ACTIONS(2103),
    [anon_sym_COLON2] = ACTIONS(2103),
    [anon_sym_DOT] = ACTIONS(2105),
    [anon_sym_DASH_GT] = ACTIONS(2103),
    [anon_sym_LBRACK] = ACTIONS(2103),
    [anon_sym_RBRACK] = ACTIONS(2105),
    [anon_sym_true] = ACTIONS(2103),
    [anon_sym_false] = ACTIONS(2103),
    [anon_sym_LBRACE] = ACTIONS(2103),
    [anon_sym_SEMI] = ACTIONS(2105),
    [anon_sym_RBRACE] = ACTIONS(2105),
    [anon_sym_constraint] = ACTIONS(2103),
    [anon_sym_val] = ACTIONS(2103),
    [anon_sym_end] = ACTIONS(2103),
    [anon_sym_with] = ACTIONS(2103),
    [anon_sym_COLON_GT] = ACTIONS(2105),
    [anon_sym_object] = ACTIONS(2103),
    [anon_sym_inherit] = ACTIONS(2103),
    [anon_sym_method] = ACTIONS(2103),
    [anon_sym_in] = ACTIONS(2103),
    [anon_sym_as] = ACTIONS(2103),
    [anon_sym_initializer] = ACTIONS(2103),
    [anon_sym_AMP] = ACTIONS(2103),
    [anon_sym_POUND] = ACTIONS(2103),
    [anon_sym_COLON_COLON] = ACTIONS(2105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2105),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2105),
    [anon_sym_then] = ACTIONS(2103),
    [anon_sym_else] = ACTIONS(2103),
    [anon_sym_do] = ACTIONS(2103),
    [anon_sym_done] = ACTIONS(2103),
    [anon_sym_to] = ACTIONS(2103),
    [anon_sym_downto] = ACTIONS(2103),
    [anon_sym_new] = ACTIONS(2103),
    [anon_sym_LBRACE_LT] = ACTIONS(2105),
    [anon_sym_GT_RBRACE] = ACTIONS(2105),
    [anon_sym_begin] = ACTIONS(2103),
    [sym_ocamlyacc_value] = ACTIONS(2105),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2103),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2105),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2103),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2103),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2105),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2105),
    [aux_sym_number_token1] = ACTIONS(2105),
    [aux_sym_signed_number_token1] = ACTIONS(2103),
    [anon_sym_SQUOTE] = ACTIONS(2105),
    [anon_sym_DQUOTE] = ACTIONS(2105),
    [sym_prefix_operator] = ACTIONS(2105),
    [aux_sym_sign_operator_token1] = ACTIONS(2103),
    [sym_hash_operator] = ACTIONS(2105),
    [anon_sym_lsl] = ACTIONS(2103),
    [anon_sym_lsr] = ACTIONS(2103),
    [anon_sym_asr] = ACTIONS(2103),
    [aux_sym_pow_operator_token1] = ACTIONS(2105),
    [anon_sym_mod] = ACTIONS(2103),
    [anon_sym_land] = ACTIONS(2103),
    [anon_sym_lor] = ACTIONS(2103),
    [anon_sym_lxor] = ACTIONS(2103),
    [aux_sym_mult_operator_token1] = ACTIONS(2103),
    [aux_sym_add_operator_token1] = ACTIONS(2103),
    [sym_concat_operator] = ACTIONS(2105),
    [sym_rel_operator] = ACTIONS(2103),
    [anon_sym_AMP_AMP] = ACTIONS(2103),
    [anon_sym_or] = ACTIONS(2103),
    [anon_sym_PIPE_PIPE] = ACTIONS(2103),
    [sym_assign_operator] = ACTIONS(2105),
    [sym__uppercase_identifier] = ACTIONS(2105),
    [aux_sym_tag_token1] = ACTIONS(2105),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(986)] = {
    [sym_attribute] = STATE(986),
    [sym__lowercase_identifier] = ACTIONS(2107),
    [anon_sym_and] = ACTIONS(2107),
    [anon_sym_TILDE] = ACTIONS(2107),
    [anon_sym_QMARK] = ACTIONS(2107),
    [anon_sym_LPAREN] = ACTIONS(2109),
    [anon_sym_RPAREN] = ACTIONS(2109),
    [anon_sym_COMMA] = ACTIONS(2109),
    [anon_sym_PIPE] = ACTIONS(2107),
    [anon_sym_COLON2] = ACTIONS(2107),
    [anon_sym_DOT] = ACTIONS(2109),
    [anon_sym_DASH_GT] = ACTIONS(2107),
    [anon_sym_LBRACK] = ACTIONS(2107),
    [anon_sym_RBRACK] = ACTIONS(2109),
    [anon_sym_true] = ACTIONS(2107),
    [anon_sym_false] = ACTIONS(2107),
    [anon_sym_LBRACE] = ACTIONS(2107),
    [anon_sym_SEMI] = ACTIONS(2109),
    [anon_sym_RBRACE] = ACTIONS(2109),
    [anon_sym_constraint] = ACTIONS(2107),
    [anon_sym_val] = ACTIONS(2107),
    [anon_sym_end] = ACTIONS(2107),
    [anon_sym_with] = ACTIONS(2107),
    [anon_sym_COLON_GT] = ACTIONS(2109),
    [anon_sym_object] = ACTIONS(2107),
    [anon_sym_inherit] = ACTIONS(2107),
    [anon_sym_method] = ACTIONS(2107),
    [anon_sym_in] = ACTIONS(2107),
    [anon_sym_as] = ACTIONS(2107),
    [anon_sym_initializer] = ACTIONS(2107),
    [anon_sym_AMP] = ACTIONS(2107),
    [anon_sym_POUND] = ACTIONS(2107),
    [anon_sym_COLON_COLON] = ACTIONS(2109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2109),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2109),
    [anon_sym_then] = ACTIONS(2107),
    [anon_sym_else] = ACTIONS(2107),
    [anon_sym_do] = ACTIONS(2107),
    [anon_sym_done] = ACTIONS(2107),
    [anon_sym_to] = ACTIONS(2107),
    [anon_sym_downto] = ACTIONS(2107),
    [anon_sym_new] = ACTIONS(2107),
    [anon_sym_LBRACE_LT] = ACTIONS(2109),
    [anon_sym_GT_RBRACE] = ACTIONS(2109),
    [anon_sym_begin] = ACTIONS(2107),
    [sym_ocamlyacc_value] = ACTIONS(2109),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2107),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2109),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2109),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2109),
    [aux_sym_number_token1] = ACTIONS(2109),
    [aux_sym_signed_number_token1] = ACTIONS(2107),
    [anon_sym_SQUOTE] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(2109),
    [sym_prefix_operator] = ACTIONS(2109),
    [aux_sym_sign_operator_token1] = ACTIONS(2107),
    [sym_hash_operator] = ACTIONS(2109),
    [anon_sym_lsl] = ACTIONS(2107),
    [anon_sym_lsr] = ACTIONS(2107),
    [anon_sym_asr] = ACTIONS(2107),
    [aux_sym_pow_operator_token1] = ACTIONS(2109),
    [anon_sym_mod] = ACTIONS(2107),
    [anon_sym_land] = ACTIONS(2107),
    [anon_sym_lor] = ACTIONS(2107),
    [anon_sym_lxor] = ACTIONS(2107),
    [aux_sym_mult_operator_token1] = ACTIONS(2107),
    [aux_sym_add_operator_token1] = ACTIONS(2107),
    [sym_concat_operator] = ACTIONS(2109),
    [sym_rel_operator] = ACTIONS(2107),
    [anon_sym_AMP_AMP] = ACTIONS(2107),
    [anon_sym_or] = ACTIONS(2107),
    [anon_sym_PIPE_PIPE] = ACTIONS(2107),
    [sym_assign_operator] = ACTIONS(2109),
    [sym__uppercase_identifier] = ACTIONS(2109),
    [aux_sym_tag_token1] = ACTIONS(2109),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(987)] = {
    [sym_attribute] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(2111),
    [anon_sym_and] = ACTIONS(2111),
    [anon_sym_TILDE] = ACTIONS(2111),
    [anon_sym_QMARK] = ACTIONS(2111),
    [anon_sym_LPAREN] = ACTIONS(2113),
    [anon_sym_RPAREN] = ACTIONS(2113),
    [anon_sym_COMMA] = ACTIONS(2113),
    [anon_sym_PIPE] = ACTIONS(2111),
    [anon_sym_COLON2] = ACTIONS(2111),
    [anon_sym_DOT] = ACTIONS(2115),
    [anon_sym_DASH_GT] = ACTIONS(2111),
    [anon_sym_LBRACK] = ACTIONS(2111),
    [anon_sym_RBRACK] = ACTIONS(2113),
    [anon_sym_true] = ACTIONS(2111),
    [anon_sym_false] = ACTIONS(2111),
    [anon_sym_LBRACE] = ACTIONS(2111),
    [anon_sym_SEMI] = ACTIONS(2113),
    [anon_sym_RBRACE] = ACTIONS(2113),
    [anon_sym_constraint] = ACTIONS(2111),
    [anon_sym_val] = ACTIONS(2111),
    [anon_sym_end] = ACTIONS(2111),
    [anon_sym_with] = ACTIONS(2111),
    [anon_sym_COLON_GT] = ACTIONS(2113),
    [anon_sym_object] = ACTIONS(2111),
    [anon_sym_inherit] = ACTIONS(2111),
    [anon_sym_method] = ACTIONS(2111),
    [anon_sym_in] = ACTIONS(2111),
    [anon_sym_as] = ACTIONS(2111),
    [anon_sym_initializer] = ACTIONS(2111),
    [anon_sym_AMP] = ACTIONS(2111),
    [anon_sym_POUND] = ACTIONS(2111),
    [anon_sym_COLON_COLON] = ACTIONS(2113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2113),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2113),
    [anon_sym_then] = ACTIONS(2111),
    [anon_sym_else] = ACTIONS(2111),
    [anon_sym_do] = ACTIONS(2111),
    [anon_sym_done] = ACTIONS(2111),
    [anon_sym_to] = ACTIONS(2111),
    [anon_sym_downto] = ACTIONS(2111),
    [anon_sym_new] = ACTIONS(2111),
    [anon_sym_LBRACE_LT] = ACTIONS(2113),
    [anon_sym_GT_RBRACE] = ACTIONS(2113),
    [anon_sym_begin] = ACTIONS(2111),
    [sym_ocamlyacc_value] = ACTIONS(2113),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2113),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2113),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2113),
    [aux_sym_number_token1] = ACTIONS(2113),
    [aux_sym_signed_number_token1] = ACTIONS(2111),
    [anon_sym_SQUOTE] = ACTIONS(2113),
    [anon_sym_DQUOTE] = ACTIONS(2113),
    [sym_prefix_operator] = ACTIONS(2113),
    [aux_sym_sign_operator_token1] = ACTIONS(2111),
    [sym_hash_operator] = ACTIONS(2113),
    [anon_sym_lsl] = ACTIONS(2111),
    [anon_sym_lsr] = ACTIONS(2111),
    [anon_sym_asr] = ACTIONS(2111),
    [aux_sym_pow_operator_token1] = ACTIONS(2113),
    [anon_sym_mod] = ACTIONS(2111),
    [anon_sym_land] = ACTIONS(2111),
    [anon_sym_lor] = ACTIONS(2111),
    [anon_sym_lxor] = ACTIONS(2111),
    [aux_sym_mult_operator_token1] = ACTIONS(2111),
    [aux_sym_add_operator_token1] = ACTIONS(2111),
    [sym_concat_operator] = ACTIONS(2113),
    [sym_rel_operator] = ACTIONS(2111),
    [anon_sym_AMP_AMP] = ACTIONS(2111),
    [anon_sym_or] = ACTIONS(2111),
    [anon_sym_PIPE_PIPE] = ACTIONS(2111),
    [sym_assign_operator] = ACTIONS(2113),
    [sym__uppercase_identifier] = ACTIONS(2113),
    [aux_sym_tag_token1] = ACTIONS(2113),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(988)] = {
    [sym_attribute] = STATE(988),
    [sym__lowercase_identifier] = ACTIONS(2117),
    [anon_sym_and] = ACTIONS(2117),
    [anon_sym_TILDE] = ACTIONS(2117),
    [anon_sym_QMARK] = ACTIONS(2117),
    [anon_sym_LPAREN] = ACTIONS(2119),
    [anon_sym_RPAREN] = ACTIONS(2119),
    [anon_sym_COMMA] = ACTIONS(2119),
    [anon_sym_PIPE] = ACTIONS(2117),
    [anon_sym_COLON2] = ACTIONS(2117),
    [anon_sym_DOT] = ACTIONS(2119),
    [anon_sym_DASH_GT] = ACTIONS(2117),
    [anon_sym_LBRACK] = ACTIONS(2117),
    [anon_sym_RBRACK] = ACTIONS(2119),
    [anon_sym_true] = ACTIONS(2117),
    [anon_sym_false] = ACTIONS(2117),
    [anon_sym_LBRACE] = ACTIONS(2117),
    [anon_sym_SEMI] = ACTIONS(2119),
    [anon_sym_RBRACE] = ACTIONS(2119),
    [anon_sym_constraint] = ACTIONS(2117),
    [anon_sym_val] = ACTIONS(2117),
    [anon_sym_end] = ACTIONS(2117),
    [anon_sym_with] = ACTIONS(2117),
    [anon_sym_COLON_GT] = ACTIONS(2119),
    [anon_sym_object] = ACTIONS(2117),
    [anon_sym_inherit] = ACTIONS(2117),
    [anon_sym_method] = ACTIONS(2117),
    [anon_sym_in] = ACTIONS(2117),
    [anon_sym_as] = ACTIONS(2117),
    [anon_sym_initializer] = ACTIONS(2117),
    [anon_sym_AMP] = ACTIONS(2117),
    [anon_sym_POUND] = ACTIONS(2117),
    [anon_sym_COLON_COLON] = ACTIONS(2119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2119),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2119),
    [anon_sym_then] = ACTIONS(2117),
    [anon_sym_else] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(2117),
    [anon_sym_done] = ACTIONS(2117),
    [anon_sym_to] = ACTIONS(2117),
    [anon_sym_downto] = ACTIONS(2117),
    [anon_sym_new] = ACTIONS(2117),
    [anon_sym_LBRACE_LT] = ACTIONS(2119),
    [anon_sym_GT_RBRACE] = ACTIONS(2119),
    [anon_sym_begin] = ACTIONS(2117),
    [sym_ocamlyacc_value] = ACTIONS(2119),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2117),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2119),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2117),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2117),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2119),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2119),
    [aux_sym_number_token1] = ACTIONS(2119),
    [aux_sym_signed_number_token1] = ACTIONS(2117),
    [anon_sym_SQUOTE] = ACTIONS(2119),
    [anon_sym_DQUOTE] = ACTIONS(2119),
    [sym_prefix_operator] = ACTIONS(2119),
    [aux_sym_sign_operator_token1] = ACTIONS(2117),
    [sym_hash_operator] = ACTIONS(2119),
    [anon_sym_lsl] = ACTIONS(2117),
    [anon_sym_lsr] = ACTIONS(2117),
    [anon_sym_asr] = ACTIONS(2117),
    [aux_sym_pow_operator_token1] = ACTIONS(2119),
    [anon_sym_mod] = ACTIONS(2117),
    [anon_sym_land] = ACTIONS(2117),
    [anon_sym_lor] = ACTIONS(2117),
    [anon_sym_lxor] = ACTIONS(2117),
    [aux_sym_mult_operator_token1] = ACTIONS(2117),
    [aux_sym_add_operator_token1] = ACTIONS(2117),
    [sym_concat_operator] = ACTIONS(2119),
    [sym_rel_operator] = ACTIONS(2117),
    [anon_sym_AMP_AMP] = ACTIONS(2117),
    [anon_sym_or] = ACTIONS(2117),
    [anon_sym_PIPE_PIPE] = ACTIONS(2117),
    [sym_assign_operator] = ACTIONS(2119),
    [sym__uppercase_identifier] = ACTIONS(2119),
    [aux_sym_tag_token1] = ACTIONS(2119),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(989)] = {
    [sym_attribute] = STATE(989),
    [sym__lowercase_identifier] = ACTIONS(2121),
    [anon_sym_and] = ACTIONS(2121),
    [anon_sym_TILDE] = ACTIONS(2121),
    [anon_sym_QMARK] = ACTIONS(2121),
    [anon_sym_LPAREN] = ACTIONS(2123),
    [anon_sym_RPAREN] = ACTIONS(2123),
    [anon_sym_COMMA] = ACTIONS(2123),
    [anon_sym_PIPE] = ACTIONS(2121),
    [anon_sym_COLON2] = ACTIONS(2121),
    [anon_sym_DOT] = ACTIONS(2123),
    [anon_sym_DASH_GT] = ACTIONS(2121),
    [anon_sym_LBRACK] = ACTIONS(2121),
    [anon_sym_RBRACK] = ACTIONS(2123),
    [anon_sym_true] = ACTIONS(2121),
    [anon_sym_false] = ACTIONS(2121),
    [anon_sym_LBRACE] = ACTIONS(2121),
    [anon_sym_SEMI] = ACTIONS(2123),
    [anon_sym_RBRACE] = ACTIONS(2123),
    [anon_sym_constraint] = ACTIONS(2121),
    [anon_sym_val] = ACTIONS(2121),
    [anon_sym_end] = ACTIONS(2121),
    [anon_sym_with] = ACTIONS(2121),
    [anon_sym_COLON_GT] = ACTIONS(2123),
    [anon_sym_object] = ACTIONS(2121),
    [anon_sym_inherit] = ACTIONS(2121),
    [anon_sym_method] = ACTIONS(2121),
    [anon_sym_in] = ACTIONS(2121),
    [anon_sym_as] = ACTIONS(2121),
    [anon_sym_initializer] = ACTIONS(2121),
    [anon_sym_AMP] = ACTIONS(2121),
    [anon_sym_POUND] = ACTIONS(2121),
    [anon_sym_COLON_COLON] = ACTIONS(2123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2123),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2123),
    [anon_sym_then] = ACTIONS(2121),
    [anon_sym_else] = ACTIONS(2121),
    [anon_sym_do] = ACTIONS(2121),
    [anon_sym_done] = ACTIONS(2121),
    [anon_sym_to] = ACTIONS(2121),
    [anon_sym_downto] = ACTIONS(2121),
    [anon_sym_new] = ACTIONS(2121),
    [anon_sym_LBRACE_LT] = ACTIONS(2123),
    [anon_sym_GT_RBRACE] = ACTIONS(2123),
    [anon_sym_begin] = ACTIONS(2121),
    [sym_ocamlyacc_value] = ACTIONS(2123),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2121),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2123),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2121),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2121),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2123),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2123),
    [aux_sym_number_token1] = ACTIONS(2123),
    [aux_sym_signed_number_token1] = ACTIONS(2121),
    [anon_sym_SQUOTE] = ACTIONS(2123),
    [anon_sym_DQUOTE] = ACTIONS(2123),
    [sym_prefix_operator] = ACTIONS(2123),
    [aux_sym_sign_operator_token1] = ACTIONS(2121),
    [sym_hash_operator] = ACTIONS(2123),
    [anon_sym_lsl] = ACTIONS(2121),
    [anon_sym_lsr] = ACTIONS(2121),
    [anon_sym_asr] = ACTIONS(2121),
    [aux_sym_pow_operator_token1] = ACTIONS(2123),
    [anon_sym_mod] = ACTIONS(2121),
    [anon_sym_land] = ACTIONS(2121),
    [anon_sym_lor] = ACTIONS(2121),
    [anon_sym_lxor] = ACTIONS(2121),
    [aux_sym_mult_operator_token1] = ACTIONS(2121),
    [aux_sym_add_operator_token1] = ACTIONS(2121),
    [sym_concat_operator] = ACTIONS(2123),
    [sym_rel_operator] = ACTIONS(2121),
    [anon_sym_AMP_AMP] = ACTIONS(2121),
    [anon_sym_or] = ACTIONS(2121),
    [anon_sym_PIPE_PIPE] = ACTIONS(2121),
    [sym_assign_operator] = ACTIONS(2123),
    [sym__uppercase_identifier] = ACTIONS(2123),
    [aux_sym_tag_token1] = ACTIONS(2123),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(990)] = {
    [sym_attribute] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(2125),
    [anon_sym_and] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(2125),
    [anon_sym_QMARK] = ACTIONS(2125),
    [anon_sym_LPAREN] = ACTIONS(2127),
    [anon_sym_RPAREN] = ACTIONS(2127),
    [anon_sym_COMMA] = ACTIONS(2127),
    [anon_sym_PIPE] = ACTIONS(2125),
    [anon_sym_COLON2] = ACTIONS(2125),
    [anon_sym_DOT] = ACTIONS(2127),
    [anon_sym_DASH_GT] = ACTIONS(2125),
    [anon_sym_LBRACK] = ACTIONS(2125),
    [anon_sym_RBRACK] = ACTIONS(2127),
    [anon_sym_true] = ACTIONS(2125),
    [anon_sym_false] = ACTIONS(2125),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_SEMI] = ACTIONS(2127),
    [anon_sym_RBRACE] = ACTIONS(2127),
    [anon_sym_constraint] = ACTIONS(2125),
    [anon_sym_val] = ACTIONS(2125),
    [anon_sym_end] = ACTIONS(2125),
    [anon_sym_with] = ACTIONS(2125),
    [anon_sym_COLON_GT] = ACTIONS(2127),
    [anon_sym_object] = ACTIONS(2125),
    [anon_sym_inherit] = ACTIONS(2125),
    [anon_sym_method] = ACTIONS(2125),
    [anon_sym_in] = ACTIONS(2125),
    [anon_sym_as] = ACTIONS(2125),
    [anon_sym_initializer] = ACTIONS(2125),
    [anon_sym_AMP] = ACTIONS(2125),
    [anon_sym_POUND] = ACTIONS(2125),
    [anon_sym_COLON_COLON] = ACTIONS(2127),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2127),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2127),
    [anon_sym_then] = ACTIONS(2125),
    [anon_sym_else] = ACTIONS(2125),
    [anon_sym_do] = ACTIONS(2125),
    [anon_sym_done] = ACTIONS(2125),
    [anon_sym_to] = ACTIONS(2125),
    [anon_sym_downto] = ACTIONS(2125),
    [anon_sym_new] = ACTIONS(2125),
    [anon_sym_LBRACE_LT] = ACTIONS(2127),
    [anon_sym_GT_RBRACE] = ACTIONS(2127),
    [anon_sym_begin] = ACTIONS(2125),
    [sym_ocamlyacc_value] = ACTIONS(2127),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2125),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2127),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2127),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2127),
    [aux_sym_number_token1] = ACTIONS(2127),
    [aux_sym_signed_number_token1] = ACTIONS(2125),
    [anon_sym_SQUOTE] = ACTIONS(2127),
    [anon_sym_DQUOTE] = ACTIONS(2127),
    [sym_prefix_operator] = ACTIONS(2127),
    [aux_sym_sign_operator_token1] = ACTIONS(2125),
    [sym_hash_operator] = ACTIONS(2127),
    [anon_sym_lsl] = ACTIONS(2125),
    [anon_sym_lsr] = ACTIONS(2125),
    [anon_sym_asr] = ACTIONS(2125),
    [aux_sym_pow_operator_token1] = ACTIONS(2127),
    [anon_sym_mod] = ACTIONS(2125),
    [anon_sym_land] = ACTIONS(2125),
    [anon_sym_lor] = ACTIONS(2125),
    [anon_sym_lxor] = ACTIONS(2125),
    [aux_sym_mult_operator_token1] = ACTIONS(2125),
    [aux_sym_add_operator_token1] = ACTIONS(2125),
    [sym_concat_operator] = ACTIONS(2127),
    [sym_rel_operator] = ACTIONS(2125),
    [anon_sym_AMP_AMP] = ACTIONS(2125),
    [anon_sym_or] = ACTIONS(2125),
    [anon_sym_PIPE_PIPE] = ACTIONS(2125),
    [sym_assign_operator] = ACTIONS(2127),
    [sym__uppercase_identifier] = ACTIONS(2127),
    [aux_sym_tag_token1] = ACTIONS(2127),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(991)] = {
    [sym_attribute] = STATE(991),
    [sym__lowercase_identifier] = ACTIONS(2129),
    [anon_sym_and] = ACTIONS(2129),
    [anon_sym_TILDE] = ACTIONS(2129),
    [anon_sym_QMARK] = ACTIONS(2129),
    [anon_sym_LPAREN] = ACTIONS(2131),
    [anon_sym_RPAREN] = ACTIONS(2131),
    [anon_sym_COMMA] = ACTIONS(2131),
    [anon_sym_PIPE] = ACTIONS(2129),
    [anon_sym_COLON2] = ACTIONS(2129),
    [anon_sym_DOT] = ACTIONS(2131),
    [anon_sym_DASH_GT] = ACTIONS(2129),
    [anon_sym_LBRACK] = ACTIONS(2129),
    [anon_sym_RBRACK] = ACTIONS(2131),
    [anon_sym_true] = ACTIONS(2129),
    [anon_sym_false] = ACTIONS(2129),
    [anon_sym_LBRACE] = ACTIONS(2129),
    [anon_sym_SEMI] = ACTIONS(2131),
    [anon_sym_RBRACE] = ACTIONS(2131),
    [anon_sym_constraint] = ACTIONS(2129),
    [anon_sym_val] = ACTIONS(2129),
    [anon_sym_end] = ACTIONS(2129),
    [anon_sym_with] = ACTIONS(2129),
    [anon_sym_COLON_GT] = ACTIONS(2131),
    [anon_sym_object] = ACTIONS(2129),
    [anon_sym_inherit] = ACTIONS(2129),
    [anon_sym_method] = ACTIONS(2129),
    [anon_sym_in] = ACTIONS(2129),
    [anon_sym_as] = ACTIONS(2129),
    [anon_sym_initializer] = ACTIONS(2129),
    [anon_sym_AMP] = ACTIONS(2129),
    [anon_sym_POUND] = ACTIONS(2129),
    [anon_sym_COLON_COLON] = ACTIONS(2131),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2131),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2131),
    [anon_sym_then] = ACTIONS(2129),
    [anon_sym_else] = ACTIONS(2129),
    [anon_sym_do] = ACTIONS(2129),
    [anon_sym_done] = ACTIONS(2129),
    [anon_sym_to] = ACTIONS(2129),
    [anon_sym_downto] = ACTIONS(2129),
    [anon_sym_new] = ACTIONS(2129),
    [anon_sym_LBRACE_LT] = ACTIONS(2131),
    [anon_sym_GT_RBRACE] = ACTIONS(2131),
    [anon_sym_begin] = ACTIONS(2129),
    [sym_ocamlyacc_value] = ACTIONS(2131),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2129),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2131),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2129),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2129),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2131),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2131),
    [aux_sym_number_token1] = ACTIONS(2131),
    [aux_sym_signed_number_token1] = ACTIONS(2129),
    [anon_sym_SQUOTE] = ACTIONS(2131),
    [anon_sym_DQUOTE] = ACTIONS(2131),
    [sym_prefix_operator] = ACTIONS(2131),
    [aux_sym_sign_operator_token1] = ACTIONS(2129),
    [sym_hash_operator] = ACTIONS(2131),
    [anon_sym_lsl] = ACTIONS(2129),
    [anon_sym_lsr] = ACTIONS(2129),
    [anon_sym_asr] = ACTIONS(2129),
    [aux_sym_pow_operator_token1] = ACTIONS(2131),
    [anon_sym_mod] = ACTIONS(2129),
    [anon_sym_land] = ACTIONS(2129),
    [anon_sym_lor] = ACTIONS(2129),
    [anon_sym_lxor] = ACTIONS(2129),
    [aux_sym_mult_operator_token1] = ACTIONS(2129),
    [aux_sym_add_operator_token1] = ACTIONS(2129),
    [sym_concat_operator] = ACTIONS(2131),
    [sym_rel_operator] = ACTIONS(2129),
    [anon_sym_AMP_AMP] = ACTIONS(2129),
    [anon_sym_or] = ACTIONS(2129),
    [anon_sym_PIPE_PIPE] = ACTIONS(2129),
    [sym_assign_operator] = ACTIONS(2131),
    [sym__uppercase_identifier] = ACTIONS(2131),
    [aux_sym_tag_token1] = ACTIONS(2131),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(992)] = {
    [sym_attribute] = STATE(992),
    [sym__lowercase_identifier] = ACTIONS(2133),
    [anon_sym_and] = ACTIONS(2133),
    [anon_sym_TILDE] = ACTIONS(2133),
    [anon_sym_QMARK] = ACTIONS(2133),
    [anon_sym_LPAREN] = ACTIONS(2135),
    [anon_sym_RPAREN] = ACTIONS(2135),
    [anon_sym_COMMA] = ACTIONS(2135),
    [anon_sym_PIPE] = ACTIONS(2133),
    [anon_sym_COLON2] = ACTIONS(2133),
    [anon_sym_DOT] = ACTIONS(2135),
    [anon_sym_DASH_GT] = ACTIONS(2133),
    [anon_sym_LBRACK] = ACTIONS(2133),
    [anon_sym_RBRACK] = ACTIONS(2135),
    [anon_sym_true] = ACTIONS(2133),
    [anon_sym_false] = ACTIONS(2133),
    [anon_sym_LBRACE] = ACTIONS(2133),
    [anon_sym_SEMI] = ACTIONS(2135),
    [anon_sym_RBRACE] = ACTIONS(2135),
    [anon_sym_constraint] = ACTIONS(2133),
    [anon_sym_val] = ACTIONS(2133),
    [anon_sym_end] = ACTIONS(2133),
    [anon_sym_with] = ACTIONS(2133),
    [anon_sym_COLON_GT] = ACTIONS(2135),
    [anon_sym_object] = ACTIONS(2133),
    [anon_sym_inherit] = ACTIONS(2133),
    [anon_sym_method] = ACTIONS(2133),
    [anon_sym_in] = ACTIONS(2133),
    [anon_sym_as] = ACTIONS(2133),
    [anon_sym_initializer] = ACTIONS(2133),
    [anon_sym_AMP] = ACTIONS(2133),
    [anon_sym_POUND] = ACTIONS(2133),
    [anon_sym_COLON_COLON] = ACTIONS(2135),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2135),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2135),
    [anon_sym_then] = ACTIONS(2133),
    [anon_sym_else] = ACTIONS(2133),
    [anon_sym_do] = ACTIONS(2133),
    [anon_sym_done] = ACTIONS(2133),
    [anon_sym_to] = ACTIONS(2133),
    [anon_sym_downto] = ACTIONS(2133),
    [anon_sym_new] = ACTIONS(2133),
    [anon_sym_LBRACE_LT] = ACTIONS(2135),
    [anon_sym_GT_RBRACE] = ACTIONS(2135),
    [anon_sym_begin] = ACTIONS(2133),
    [sym_ocamlyacc_value] = ACTIONS(2135),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2133),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2135),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2133),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2133),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2135),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2135),
    [aux_sym_number_token1] = ACTIONS(2135),
    [aux_sym_signed_number_token1] = ACTIONS(2133),
    [anon_sym_SQUOTE] = ACTIONS(2135),
    [anon_sym_DQUOTE] = ACTIONS(2135),
    [sym_prefix_operator] = ACTIONS(2135),
    [aux_sym_sign_operator_token1] = ACTIONS(2133),
    [sym_hash_operator] = ACTIONS(2135),
    [anon_sym_lsl] = ACTIONS(2133),
    [anon_sym_lsr] = ACTIONS(2133),
    [anon_sym_asr] = ACTIONS(2133),
    [aux_sym_pow_operator_token1] = ACTIONS(2135),
    [anon_sym_mod] = ACTIONS(2133),
    [anon_sym_land] = ACTIONS(2133),
    [anon_sym_lor] = ACTIONS(2133),
    [anon_sym_lxor] = ACTIONS(2133),
    [aux_sym_mult_operator_token1] = ACTIONS(2133),
    [aux_sym_add_operator_token1] = ACTIONS(2133),
    [sym_concat_operator] = ACTIONS(2135),
    [sym_rel_operator] = ACTIONS(2133),
    [anon_sym_AMP_AMP] = ACTIONS(2133),
    [anon_sym_or] = ACTIONS(2133),
    [anon_sym_PIPE_PIPE] = ACTIONS(2133),
    [sym_assign_operator] = ACTIONS(2135),
    [sym__uppercase_identifier] = ACTIONS(2135),
    [aux_sym_tag_token1] = ACTIONS(2135),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(993)] = {
    [sym_attribute] = STATE(993),
    [sym__lowercase_identifier] = ACTIONS(2137),
    [anon_sym_and] = ACTIONS(2137),
    [anon_sym_TILDE] = ACTIONS(2137),
    [anon_sym_QMARK] = ACTIONS(2137),
    [anon_sym_LPAREN] = ACTIONS(2139),
    [anon_sym_RPAREN] = ACTIONS(2139),
    [anon_sym_COMMA] = ACTIONS(2139),
    [anon_sym_PIPE] = ACTIONS(2137),
    [anon_sym_COLON2] = ACTIONS(2137),
    [anon_sym_DOT] = ACTIONS(2139),
    [anon_sym_DASH_GT] = ACTIONS(2137),
    [anon_sym_LBRACK] = ACTIONS(2137),
    [anon_sym_RBRACK] = ACTIONS(2139),
    [anon_sym_true] = ACTIONS(2137),
    [anon_sym_false] = ACTIONS(2137),
    [anon_sym_LBRACE] = ACTIONS(2137),
    [anon_sym_SEMI] = ACTIONS(2139),
    [anon_sym_RBRACE] = ACTIONS(2139),
    [anon_sym_constraint] = ACTIONS(2137),
    [anon_sym_val] = ACTIONS(2137),
    [anon_sym_end] = ACTIONS(2137),
    [anon_sym_with] = ACTIONS(2137),
    [anon_sym_COLON_GT] = ACTIONS(2139),
    [anon_sym_object] = ACTIONS(2137),
    [anon_sym_inherit] = ACTIONS(2137),
    [anon_sym_method] = ACTIONS(2137),
    [anon_sym_in] = ACTIONS(2137),
    [anon_sym_as] = ACTIONS(2137),
    [anon_sym_initializer] = ACTIONS(2137),
    [anon_sym_AMP] = ACTIONS(2137),
    [anon_sym_POUND] = ACTIONS(2137),
    [anon_sym_COLON_COLON] = ACTIONS(2139),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2139),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2139),
    [anon_sym_then] = ACTIONS(2137),
    [anon_sym_else] = ACTIONS(2137),
    [anon_sym_do] = ACTIONS(2137),
    [anon_sym_done] = ACTIONS(2137),
    [anon_sym_to] = ACTIONS(2137),
    [anon_sym_downto] = ACTIONS(2137),
    [anon_sym_new] = ACTIONS(2137),
    [anon_sym_LBRACE_LT] = ACTIONS(2139),
    [anon_sym_GT_RBRACE] = ACTIONS(2139),
    [anon_sym_begin] = ACTIONS(2137),
    [sym_ocamlyacc_value] = ACTIONS(2139),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2137),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2139),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2137),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2137),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2139),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2139),
    [aux_sym_number_token1] = ACTIONS(2139),
    [aux_sym_signed_number_token1] = ACTIONS(2137),
    [anon_sym_SQUOTE] = ACTIONS(2139),
    [anon_sym_DQUOTE] = ACTIONS(2139),
    [sym_prefix_operator] = ACTIONS(2139),
    [aux_sym_sign_operator_token1] = ACTIONS(2137),
    [sym_hash_operator] = ACTIONS(2139),
    [anon_sym_lsl] = ACTIONS(2137),
    [anon_sym_lsr] = ACTIONS(2137),
    [anon_sym_asr] = ACTIONS(2137),
    [aux_sym_pow_operator_token1] = ACTIONS(2139),
    [anon_sym_mod] = ACTIONS(2137),
    [anon_sym_land] = ACTIONS(2137),
    [anon_sym_lor] = ACTIONS(2137),
    [anon_sym_lxor] = ACTIONS(2137),
    [aux_sym_mult_operator_token1] = ACTIONS(2137),
    [aux_sym_add_operator_token1] = ACTIONS(2137),
    [sym_concat_operator] = ACTIONS(2139),
    [sym_rel_operator] = ACTIONS(2137),
    [anon_sym_AMP_AMP] = ACTIONS(2137),
    [anon_sym_or] = ACTIONS(2137),
    [anon_sym_PIPE_PIPE] = ACTIONS(2137),
    [sym_assign_operator] = ACTIONS(2139),
    [sym__uppercase_identifier] = ACTIONS(2139),
    [aux_sym_tag_token1] = ACTIONS(2139),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(994)] = {
    [sym_attribute] = STATE(994),
    [sym__lowercase_identifier] = ACTIONS(2141),
    [anon_sym_and] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_QMARK] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2143),
    [anon_sym_RPAREN] = ACTIONS(2143),
    [anon_sym_COMMA] = ACTIONS(2143),
    [anon_sym_PIPE] = ACTIONS(2141),
    [anon_sym_COLON2] = ACTIONS(2141),
    [anon_sym_DOT] = ACTIONS(2143),
    [anon_sym_DASH_GT] = ACTIONS(2141),
    [anon_sym_LBRACK] = ACTIONS(2141),
    [anon_sym_RBRACK] = ACTIONS(2143),
    [anon_sym_true] = ACTIONS(2141),
    [anon_sym_false] = ACTIONS(2141),
    [anon_sym_LBRACE] = ACTIONS(2141),
    [anon_sym_SEMI] = ACTIONS(2143),
    [anon_sym_RBRACE] = ACTIONS(2143),
    [anon_sym_constraint] = ACTIONS(2141),
    [anon_sym_val] = ACTIONS(2141),
    [anon_sym_end] = ACTIONS(2141),
    [anon_sym_with] = ACTIONS(2141),
    [anon_sym_COLON_GT] = ACTIONS(2143),
    [anon_sym_object] = ACTIONS(2141),
    [anon_sym_inherit] = ACTIONS(2141),
    [anon_sym_method] = ACTIONS(2141),
    [anon_sym_in] = ACTIONS(2141),
    [anon_sym_as] = ACTIONS(2141),
    [anon_sym_initializer] = ACTIONS(2141),
    [anon_sym_AMP] = ACTIONS(2141),
    [anon_sym_POUND] = ACTIONS(2141),
    [anon_sym_COLON_COLON] = ACTIONS(2143),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2143),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2143),
    [anon_sym_then] = ACTIONS(2141),
    [anon_sym_else] = ACTIONS(2141),
    [anon_sym_do] = ACTIONS(2141),
    [anon_sym_done] = ACTIONS(2141),
    [anon_sym_to] = ACTIONS(2141),
    [anon_sym_downto] = ACTIONS(2141),
    [anon_sym_new] = ACTIONS(2141),
    [anon_sym_LBRACE_LT] = ACTIONS(2143),
    [anon_sym_GT_RBRACE] = ACTIONS(2143),
    [anon_sym_begin] = ACTIONS(2141),
    [sym_ocamlyacc_value] = ACTIONS(2143),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2141),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2143),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2141),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2141),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2143),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2143),
    [aux_sym_number_token1] = ACTIONS(2143),
    [aux_sym_signed_number_token1] = ACTIONS(2141),
    [anon_sym_SQUOTE] = ACTIONS(2143),
    [anon_sym_DQUOTE] = ACTIONS(2143),
    [sym_prefix_operator] = ACTIONS(2143),
    [aux_sym_sign_operator_token1] = ACTIONS(2141),
    [sym_hash_operator] = ACTIONS(2143),
    [anon_sym_lsl] = ACTIONS(2141),
    [anon_sym_lsr] = ACTIONS(2141),
    [anon_sym_asr] = ACTIONS(2141),
    [aux_sym_pow_operator_token1] = ACTIONS(2143),
    [anon_sym_mod] = ACTIONS(2141),
    [anon_sym_land] = ACTIONS(2141),
    [anon_sym_lor] = ACTIONS(2141),
    [anon_sym_lxor] = ACTIONS(2141),
    [aux_sym_mult_operator_token1] = ACTIONS(2141),
    [aux_sym_add_operator_token1] = ACTIONS(2141),
    [sym_concat_operator] = ACTIONS(2143),
    [sym_rel_operator] = ACTIONS(2141),
    [anon_sym_AMP_AMP] = ACTIONS(2141),
    [anon_sym_or] = ACTIONS(2141),
    [anon_sym_PIPE_PIPE] = ACTIONS(2141),
    [sym_assign_operator] = ACTIONS(2143),
    [sym__uppercase_identifier] = ACTIONS(2143),
    [aux_sym_tag_token1] = ACTIONS(2143),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(995)] = {
    [sym_attribute] = STATE(995),
    [sym__lowercase_identifier] = ACTIONS(2145),
    [anon_sym_and] = ACTIONS(2145),
    [anon_sym_TILDE] = ACTIONS(2145),
    [anon_sym_QMARK] = ACTIONS(2145),
    [anon_sym_LPAREN] = ACTIONS(2147),
    [anon_sym_RPAREN] = ACTIONS(2147),
    [anon_sym_COMMA] = ACTIONS(2147),
    [anon_sym_PIPE] = ACTIONS(2145),
    [anon_sym_COLON2] = ACTIONS(2145),
    [anon_sym_DOT] = ACTIONS(2147),
    [anon_sym_DASH_GT] = ACTIONS(2145),
    [anon_sym_LBRACK] = ACTIONS(2145),
    [anon_sym_RBRACK] = ACTIONS(2147),
    [anon_sym_true] = ACTIONS(2145),
    [anon_sym_false] = ACTIONS(2145),
    [anon_sym_LBRACE] = ACTIONS(2145),
    [anon_sym_SEMI] = ACTIONS(2147),
    [anon_sym_RBRACE] = ACTIONS(2147),
    [anon_sym_constraint] = ACTIONS(2145),
    [anon_sym_val] = ACTIONS(2145),
    [anon_sym_end] = ACTIONS(2145),
    [anon_sym_with] = ACTIONS(2145),
    [anon_sym_COLON_GT] = ACTIONS(2147),
    [anon_sym_object] = ACTIONS(2145),
    [anon_sym_inherit] = ACTIONS(2145),
    [anon_sym_method] = ACTIONS(2145),
    [anon_sym_in] = ACTIONS(2145),
    [anon_sym_as] = ACTIONS(2145),
    [anon_sym_initializer] = ACTIONS(2145),
    [anon_sym_AMP] = ACTIONS(2145),
    [anon_sym_POUND] = ACTIONS(2145),
    [anon_sym_COLON_COLON] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2147),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2147),
    [anon_sym_then] = ACTIONS(2145),
    [anon_sym_else] = ACTIONS(2145),
    [anon_sym_do] = ACTIONS(2145),
    [anon_sym_done] = ACTIONS(2145),
    [anon_sym_to] = ACTIONS(2145),
    [anon_sym_downto] = ACTIONS(2145),
    [anon_sym_new] = ACTIONS(2145),
    [anon_sym_LBRACE_LT] = ACTIONS(2147),
    [anon_sym_GT_RBRACE] = ACTIONS(2147),
    [anon_sym_begin] = ACTIONS(2145),
    [sym_ocamlyacc_value] = ACTIONS(2147),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2145),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2147),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2145),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2145),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2147),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2147),
    [aux_sym_number_token1] = ACTIONS(2147),
    [aux_sym_signed_number_token1] = ACTIONS(2145),
    [anon_sym_SQUOTE] = ACTIONS(2147),
    [anon_sym_DQUOTE] = ACTIONS(2147),
    [sym_prefix_operator] = ACTIONS(2147),
    [aux_sym_sign_operator_token1] = ACTIONS(2145),
    [sym_hash_operator] = ACTIONS(2147),
    [anon_sym_lsl] = ACTIONS(2145),
    [anon_sym_lsr] = ACTIONS(2145),
    [anon_sym_asr] = ACTIONS(2145),
    [aux_sym_pow_operator_token1] = ACTIONS(2147),
    [anon_sym_mod] = ACTIONS(2145),
    [anon_sym_land] = ACTIONS(2145),
    [anon_sym_lor] = ACTIONS(2145),
    [anon_sym_lxor] = ACTIONS(2145),
    [aux_sym_mult_operator_token1] = ACTIONS(2145),
    [aux_sym_add_operator_token1] = ACTIONS(2145),
    [sym_concat_operator] = ACTIONS(2147),
    [sym_rel_operator] = ACTIONS(2145),
    [anon_sym_AMP_AMP] = ACTIONS(2145),
    [anon_sym_or] = ACTIONS(2145),
    [anon_sym_PIPE_PIPE] = ACTIONS(2145),
    [sym_assign_operator] = ACTIONS(2147),
    [sym__uppercase_identifier] = ACTIONS(2147),
    [aux_sym_tag_token1] = ACTIONS(2147),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(996)] = {
    [sym_attribute] = STATE(996),
    [sym__lowercase_identifier] = ACTIONS(2149),
    [anon_sym_and] = ACTIONS(2149),
    [anon_sym_TILDE] = ACTIONS(2149),
    [anon_sym_QMARK] = ACTIONS(2149),
    [anon_sym_LPAREN] = ACTIONS(2151),
    [anon_sym_RPAREN] = ACTIONS(2151),
    [anon_sym_COMMA] = ACTIONS(2151),
    [anon_sym_PIPE] = ACTIONS(2149),
    [anon_sym_COLON2] = ACTIONS(2149),
    [anon_sym_DOT] = ACTIONS(2151),
    [anon_sym_DASH_GT] = ACTIONS(2149),
    [anon_sym_LBRACK] = ACTIONS(2149),
    [anon_sym_RBRACK] = ACTIONS(2151),
    [anon_sym_true] = ACTIONS(2149),
    [anon_sym_false] = ACTIONS(2149),
    [anon_sym_LBRACE] = ACTIONS(2149),
    [anon_sym_SEMI] = ACTIONS(2151),
    [anon_sym_RBRACE] = ACTIONS(2151),
    [anon_sym_constraint] = ACTIONS(2149),
    [anon_sym_val] = ACTIONS(2149),
    [anon_sym_end] = ACTIONS(2149),
    [anon_sym_with] = ACTIONS(2149),
    [anon_sym_COLON_GT] = ACTIONS(2151),
    [anon_sym_object] = ACTIONS(2149),
    [anon_sym_inherit] = ACTIONS(2149),
    [anon_sym_method] = ACTIONS(2149),
    [anon_sym_in] = ACTIONS(2149),
    [anon_sym_as] = ACTIONS(2149),
    [anon_sym_initializer] = ACTIONS(2149),
    [anon_sym_AMP] = ACTIONS(2149),
    [anon_sym_POUND] = ACTIONS(2149),
    [anon_sym_COLON_COLON] = ACTIONS(2151),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2151),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2151),
    [anon_sym_then] = ACTIONS(2149),
    [anon_sym_else] = ACTIONS(2149),
    [anon_sym_do] = ACTIONS(2149),
    [anon_sym_done] = ACTIONS(2149),
    [anon_sym_to] = ACTIONS(2149),
    [anon_sym_downto] = ACTIONS(2149),
    [anon_sym_new] = ACTIONS(2149),
    [anon_sym_LBRACE_LT] = ACTIONS(2151),
    [anon_sym_GT_RBRACE] = ACTIONS(2151),
    [anon_sym_begin] = ACTIONS(2149),
    [sym_ocamlyacc_value] = ACTIONS(2151),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2149),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2151),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2149),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2151),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2151),
    [aux_sym_number_token1] = ACTIONS(2151),
    [aux_sym_signed_number_token1] = ACTIONS(2149),
    [anon_sym_SQUOTE] = ACTIONS(2151),
    [anon_sym_DQUOTE] = ACTIONS(2151),
    [sym_prefix_operator] = ACTIONS(2151),
    [aux_sym_sign_operator_token1] = ACTIONS(2149),
    [sym_hash_operator] = ACTIONS(2151),
    [anon_sym_lsl] = ACTIONS(2149),
    [anon_sym_lsr] = ACTIONS(2149),
    [anon_sym_asr] = ACTIONS(2149),
    [aux_sym_pow_operator_token1] = ACTIONS(2151),
    [anon_sym_mod] = ACTIONS(2149),
    [anon_sym_land] = ACTIONS(2149),
    [anon_sym_lor] = ACTIONS(2149),
    [anon_sym_lxor] = ACTIONS(2149),
    [aux_sym_mult_operator_token1] = ACTIONS(2149),
    [aux_sym_add_operator_token1] = ACTIONS(2149),
    [sym_concat_operator] = ACTIONS(2151),
    [sym_rel_operator] = ACTIONS(2149),
    [anon_sym_AMP_AMP] = ACTIONS(2149),
    [anon_sym_or] = ACTIONS(2149),
    [anon_sym_PIPE_PIPE] = ACTIONS(2149),
    [sym_assign_operator] = ACTIONS(2151),
    [sym__uppercase_identifier] = ACTIONS(2151),
    [aux_sym_tag_token1] = ACTIONS(2151),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(997)] = {
    [sym_attribute] = STATE(997),
    [sym__lowercase_identifier] = ACTIONS(2153),
    [anon_sym_and] = ACTIONS(2153),
    [anon_sym_TILDE] = ACTIONS(2153),
    [anon_sym_QMARK] = ACTIONS(2153),
    [anon_sym_LPAREN] = ACTIONS(2155),
    [anon_sym_RPAREN] = ACTIONS(2155),
    [anon_sym_COMMA] = ACTIONS(2155),
    [anon_sym_PIPE] = ACTIONS(2153),
    [anon_sym_COLON2] = ACTIONS(2153),
    [anon_sym_DOT] = ACTIONS(2155),
    [anon_sym_DASH_GT] = ACTIONS(2153),
    [anon_sym_LBRACK] = ACTIONS(2153),
    [anon_sym_RBRACK] = ACTIONS(2155),
    [anon_sym_true] = ACTIONS(2153),
    [anon_sym_false] = ACTIONS(2153),
    [anon_sym_LBRACE] = ACTIONS(2153),
    [anon_sym_SEMI] = ACTIONS(2155),
    [anon_sym_RBRACE] = ACTIONS(2155),
    [anon_sym_constraint] = ACTIONS(2153),
    [anon_sym_val] = ACTIONS(2153),
    [anon_sym_end] = ACTIONS(2153),
    [anon_sym_with] = ACTIONS(2153),
    [anon_sym_COLON_GT] = ACTIONS(2155),
    [anon_sym_object] = ACTIONS(2153),
    [anon_sym_inherit] = ACTIONS(2153),
    [anon_sym_method] = ACTIONS(2153),
    [anon_sym_in] = ACTIONS(2153),
    [anon_sym_as] = ACTIONS(2153),
    [anon_sym_initializer] = ACTIONS(2153),
    [anon_sym_AMP] = ACTIONS(2153),
    [anon_sym_POUND] = ACTIONS(2153),
    [anon_sym_COLON_COLON] = ACTIONS(2155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2155),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2155),
    [anon_sym_then] = ACTIONS(2153),
    [anon_sym_else] = ACTIONS(2153),
    [anon_sym_do] = ACTIONS(2153),
    [anon_sym_done] = ACTIONS(2153),
    [anon_sym_to] = ACTIONS(2153),
    [anon_sym_downto] = ACTIONS(2153),
    [anon_sym_new] = ACTIONS(2153),
    [anon_sym_LBRACE_LT] = ACTIONS(2155),
    [anon_sym_GT_RBRACE] = ACTIONS(2155),
    [anon_sym_begin] = ACTIONS(2153),
    [sym_ocamlyacc_value] = ACTIONS(2155),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2153),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2155),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2153),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2153),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2155),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2155),
    [aux_sym_number_token1] = ACTIONS(2155),
    [aux_sym_signed_number_token1] = ACTIONS(2153),
    [anon_sym_SQUOTE] = ACTIONS(2155),
    [anon_sym_DQUOTE] = ACTIONS(2155),
    [sym_prefix_operator] = ACTIONS(2155),
    [aux_sym_sign_operator_token1] = ACTIONS(2153),
    [sym_hash_operator] = ACTIONS(2155),
    [anon_sym_lsl] = ACTIONS(2153),
    [anon_sym_lsr] = ACTIONS(2153),
    [anon_sym_asr] = ACTIONS(2153),
    [aux_sym_pow_operator_token1] = ACTIONS(2155),
    [anon_sym_mod] = ACTIONS(2153),
    [anon_sym_land] = ACTIONS(2153),
    [anon_sym_lor] = ACTIONS(2153),
    [anon_sym_lxor] = ACTIONS(2153),
    [aux_sym_mult_operator_token1] = ACTIONS(2153),
    [aux_sym_add_operator_token1] = ACTIONS(2153),
    [sym_concat_operator] = ACTIONS(2155),
    [sym_rel_operator] = ACTIONS(2153),
    [anon_sym_AMP_AMP] = ACTIONS(2153),
    [anon_sym_or] = ACTIONS(2153),
    [anon_sym_PIPE_PIPE] = ACTIONS(2153),
    [sym_assign_operator] = ACTIONS(2155),
    [sym__uppercase_identifier] = ACTIONS(2155),
    [aux_sym_tag_token1] = ACTIONS(2155),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(998)] = {
    [sym_attribute] = STATE(998),
    [sym__lowercase_identifier] = ACTIONS(2157),
    [anon_sym_and] = ACTIONS(2157),
    [anon_sym_TILDE] = ACTIONS(2157),
    [anon_sym_QMARK] = ACTIONS(2157),
    [anon_sym_LPAREN] = ACTIONS(2159),
    [anon_sym_RPAREN] = ACTIONS(2159),
    [anon_sym_COMMA] = ACTIONS(2159),
    [anon_sym_PIPE] = ACTIONS(2157),
    [anon_sym_COLON2] = ACTIONS(2157),
    [anon_sym_DOT] = ACTIONS(2159),
    [anon_sym_DASH_GT] = ACTIONS(2157),
    [anon_sym_LBRACK] = ACTIONS(2157),
    [anon_sym_RBRACK] = ACTIONS(2159),
    [anon_sym_true] = ACTIONS(2157),
    [anon_sym_false] = ACTIONS(2157),
    [anon_sym_LBRACE] = ACTIONS(2157),
    [anon_sym_SEMI] = ACTIONS(2159),
    [anon_sym_RBRACE] = ACTIONS(2159),
    [anon_sym_constraint] = ACTIONS(2157),
    [anon_sym_val] = ACTIONS(2157),
    [anon_sym_end] = ACTIONS(2157),
    [anon_sym_with] = ACTIONS(2157),
    [anon_sym_COLON_GT] = ACTIONS(2159),
    [anon_sym_object] = ACTIONS(2157),
    [anon_sym_inherit] = ACTIONS(2157),
    [anon_sym_method] = ACTIONS(2157),
    [anon_sym_in] = ACTIONS(2157),
    [anon_sym_as] = ACTIONS(2157),
    [anon_sym_initializer] = ACTIONS(2157),
    [anon_sym_AMP] = ACTIONS(2157),
    [anon_sym_POUND] = ACTIONS(2157),
    [anon_sym_COLON_COLON] = ACTIONS(2159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2159),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2159),
    [anon_sym_then] = ACTIONS(2157),
    [anon_sym_else] = ACTIONS(2157),
    [anon_sym_do] = ACTIONS(2157),
    [anon_sym_done] = ACTIONS(2157),
    [anon_sym_to] = ACTIONS(2157),
    [anon_sym_downto] = ACTIONS(2157),
    [anon_sym_new] = ACTIONS(2157),
    [anon_sym_LBRACE_LT] = ACTIONS(2159),
    [anon_sym_GT_RBRACE] = ACTIONS(2159),
    [anon_sym_begin] = ACTIONS(2157),
    [sym_ocamlyacc_value] = ACTIONS(2159),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2157),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2159),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2159),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2159),
    [aux_sym_number_token1] = ACTIONS(2159),
    [aux_sym_signed_number_token1] = ACTIONS(2157),
    [anon_sym_SQUOTE] = ACTIONS(2159),
    [anon_sym_DQUOTE] = ACTIONS(2159),
    [sym_prefix_operator] = ACTIONS(2159),
    [aux_sym_sign_operator_token1] = ACTIONS(2157),
    [sym_hash_operator] = ACTIONS(2159),
    [anon_sym_lsl] = ACTIONS(2157),
    [anon_sym_lsr] = ACTIONS(2157),
    [anon_sym_asr] = ACTIONS(2157),
    [aux_sym_pow_operator_token1] = ACTIONS(2159),
    [anon_sym_mod] = ACTIONS(2157),
    [anon_sym_land] = ACTIONS(2157),
    [anon_sym_lor] = ACTIONS(2157),
    [anon_sym_lxor] = ACTIONS(2157),
    [aux_sym_mult_operator_token1] = ACTIONS(2157),
    [aux_sym_add_operator_token1] = ACTIONS(2157),
    [sym_concat_operator] = ACTIONS(2159),
    [sym_rel_operator] = ACTIONS(2157),
    [anon_sym_AMP_AMP] = ACTIONS(2157),
    [anon_sym_or] = ACTIONS(2157),
    [anon_sym_PIPE_PIPE] = ACTIONS(2157),
    [sym_assign_operator] = ACTIONS(2159),
    [sym__uppercase_identifier] = ACTIONS(2159),
    [aux_sym_tag_token1] = ACTIONS(2159),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(999)] = {
    [sym_attribute] = STATE(999),
    [sym__lowercase_identifier] = ACTIONS(2161),
    [anon_sym_and] = ACTIONS(2161),
    [anon_sym_TILDE] = ACTIONS(2161),
    [anon_sym_QMARK] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_RPAREN] = ACTIONS(2163),
    [anon_sym_COMMA] = ACTIONS(2163),
    [anon_sym_PIPE] = ACTIONS(2161),
    [anon_sym_COLON2] = ACTIONS(2161),
    [anon_sym_DOT] = ACTIONS(2163),
    [anon_sym_DASH_GT] = ACTIONS(2161),
    [anon_sym_LBRACK] = ACTIONS(2161),
    [anon_sym_RBRACK] = ACTIONS(2163),
    [anon_sym_true] = ACTIONS(2161),
    [anon_sym_false] = ACTIONS(2161),
    [anon_sym_LBRACE] = ACTIONS(2161),
    [anon_sym_SEMI] = ACTIONS(2163),
    [anon_sym_RBRACE] = ACTIONS(2163),
    [anon_sym_constraint] = ACTIONS(2161),
    [anon_sym_val] = ACTIONS(2161),
    [anon_sym_end] = ACTIONS(2161),
    [anon_sym_with] = ACTIONS(2161),
    [anon_sym_COLON_GT] = ACTIONS(2163),
    [anon_sym_object] = ACTIONS(2161),
    [anon_sym_inherit] = ACTIONS(2161),
    [anon_sym_method] = ACTIONS(2161),
    [anon_sym_in] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_initializer] = ACTIONS(2161),
    [anon_sym_AMP] = ACTIONS(2161),
    [anon_sym_POUND] = ACTIONS(2161),
    [anon_sym_COLON_COLON] = ACTIONS(2163),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2163),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2163),
    [anon_sym_then] = ACTIONS(2161),
    [anon_sym_else] = ACTIONS(2161),
    [anon_sym_do] = ACTIONS(2161),
    [anon_sym_done] = ACTIONS(2161),
    [anon_sym_to] = ACTIONS(2161),
    [anon_sym_downto] = ACTIONS(2161),
    [anon_sym_new] = ACTIONS(2161),
    [anon_sym_LBRACE_LT] = ACTIONS(2163),
    [anon_sym_GT_RBRACE] = ACTIONS(2163),
    [anon_sym_begin] = ACTIONS(2161),
    [sym_ocamlyacc_value] = ACTIONS(2163),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2161),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2163),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2163),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2163),
    [aux_sym_number_token1] = ACTIONS(2163),
    [aux_sym_signed_number_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2163),
    [anon_sym_DQUOTE] = ACTIONS(2163),
    [sym_prefix_operator] = ACTIONS(2163),
    [aux_sym_sign_operator_token1] = ACTIONS(2161),
    [sym_hash_operator] = ACTIONS(2163),
    [anon_sym_lsl] = ACTIONS(2161),
    [anon_sym_lsr] = ACTIONS(2161),
    [anon_sym_asr] = ACTIONS(2161),
    [aux_sym_pow_operator_token1] = ACTIONS(2163),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_land] = ACTIONS(2161),
    [anon_sym_lor] = ACTIONS(2161),
    [anon_sym_lxor] = ACTIONS(2161),
    [aux_sym_mult_operator_token1] = ACTIONS(2161),
    [aux_sym_add_operator_token1] = ACTIONS(2161),
    [sym_concat_operator] = ACTIONS(2163),
    [sym_rel_operator] = ACTIONS(2161),
    [anon_sym_AMP_AMP] = ACTIONS(2161),
    [anon_sym_or] = ACTIONS(2161),
    [anon_sym_PIPE_PIPE] = ACTIONS(2161),
    [sym_assign_operator] = ACTIONS(2163),
    [sym__uppercase_identifier] = ACTIONS(2163),
    [aux_sym_tag_token1] = ACTIONS(2163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1000)] = {
    [sym_attribute] = STATE(1000),
    [sym__lowercase_identifier] = ACTIONS(2165),
    [anon_sym_and] = ACTIONS(2165),
    [anon_sym_TILDE] = ACTIONS(2165),
    [anon_sym_QMARK] = ACTIONS(2165),
    [anon_sym_LPAREN] = ACTIONS(2167),
    [anon_sym_RPAREN] = ACTIONS(2167),
    [anon_sym_COMMA] = ACTIONS(2167),
    [anon_sym_PIPE] = ACTIONS(2165),
    [anon_sym_COLON2] = ACTIONS(2165),
    [anon_sym_DOT] = ACTIONS(2167),
    [anon_sym_DASH_GT] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2165),
    [anon_sym_RBRACK] = ACTIONS(2167),
    [anon_sym_true] = ACTIONS(2165),
    [anon_sym_false] = ACTIONS(2165),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_SEMI] = ACTIONS(2167),
    [anon_sym_RBRACE] = ACTIONS(2167),
    [anon_sym_constraint] = ACTIONS(2165),
    [anon_sym_val] = ACTIONS(2165),
    [anon_sym_end] = ACTIONS(2165),
    [anon_sym_with] = ACTIONS(2165),
    [anon_sym_COLON_GT] = ACTIONS(2167),
    [anon_sym_object] = ACTIONS(2165),
    [anon_sym_inherit] = ACTIONS(2165),
    [anon_sym_method] = ACTIONS(2165),
    [anon_sym_in] = ACTIONS(2165),
    [anon_sym_as] = ACTIONS(2165),
    [anon_sym_initializer] = ACTIONS(2165),
    [anon_sym_AMP] = ACTIONS(2165),
    [anon_sym_POUND] = ACTIONS(2165),
    [anon_sym_COLON_COLON] = ACTIONS(2167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2167),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2167),
    [anon_sym_then] = ACTIONS(2165),
    [anon_sym_else] = ACTIONS(2165),
    [anon_sym_do] = ACTIONS(2165),
    [anon_sym_done] = ACTIONS(2165),
    [anon_sym_to] = ACTIONS(2165),
    [anon_sym_downto] = ACTIONS(2165),
    [anon_sym_new] = ACTIONS(2165),
    [anon_sym_LBRACE_LT] = ACTIONS(2167),
    [anon_sym_GT_RBRACE] = ACTIONS(2167),
    [anon_sym_begin] = ACTIONS(2165),
    [sym_ocamlyacc_value] = ACTIONS(2167),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2165),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2167),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2167),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2167),
    [aux_sym_number_token1] = ACTIONS(2167),
    [aux_sym_signed_number_token1] = ACTIONS(2165),
    [anon_sym_SQUOTE] = ACTIONS(2167),
    [anon_sym_DQUOTE] = ACTIONS(2167),
    [sym_prefix_operator] = ACTIONS(2167),
    [aux_sym_sign_operator_token1] = ACTIONS(2165),
    [sym_hash_operator] = ACTIONS(2167),
    [anon_sym_lsl] = ACTIONS(2165),
    [anon_sym_lsr] = ACTIONS(2165),
    [anon_sym_asr] = ACTIONS(2165),
    [aux_sym_pow_operator_token1] = ACTIONS(2167),
    [anon_sym_mod] = ACTIONS(2165),
    [anon_sym_land] = ACTIONS(2165),
    [anon_sym_lor] = ACTIONS(2165),
    [anon_sym_lxor] = ACTIONS(2165),
    [aux_sym_mult_operator_token1] = ACTIONS(2165),
    [aux_sym_add_operator_token1] = ACTIONS(2165),
    [sym_concat_operator] = ACTIONS(2167),
    [sym_rel_operator] = ACTIONS(2165),
    [anon_sym_AMP_AMP] = ACTIONS(2165),
    [anon_sym_or] = ACTIONS(2165),
    [anon_sym_PIPE_PIPE] = ACTIONS(2165),
    [sym_assign_operator] = ACTIONS(2167),
    [sym__uppercase_identifier] = ACTIONS(2167),
    [aux_sym_tag_token1] = ACTIONS(2167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1001)] = {
    [sym_attribute] = STATE(1001),
    [sym__lowercase_identifier] = ACTIONS(2169),
    [anon_sym_and] = ACTIONS(2169),
    [anon_sym_TILDE] = ACTIONS(2169),
    [anon_sym_QMARK] = ACTIONS(2169),
    [anon_sym_LPAREN] = ACTIONS(2171),
    [anon_sym_RPAREN] = ACTIONS(2171),
    [anon_sym_COMMA] = ACTIONS(2171),
    [anon_sym_PIPE] = ACTIONS(2169),
    [anon_sym_COLON2] = ACTIONS(2169),
    [anon_sym_DOT] = ACTIONS(2171),
    [anon_sym_DASH_GT] = ACTIONS(2169),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_RBRACK] = ACTIONS(2171),
    [anon_sym_true] = ACTIONS(2169),
    [anon_sym_false] = ACTIONS(2169),
    [anon_sym_LBRACE] = ACTIONS(2169),
    [anon_sym_SEMI] = ACTIONS(2171),
    [anon_sym_RBRACE] = ACTIONS(2171),
    [anon_sym_constraint] = ACTIONS(2169),
    [anon_sym_val] = ACTIONS(2169),
    [anon_sym_end] = ACTIONS(2169),
    [anon_sym_with] = ACTIONS(2169),
    [anon_sym_COLON_GT] = ACTIONS(2171),
    [anon_sym_object] = ACTIONS(2169),
    [anon_sym_inherit] = ACTIONS(2169),
    [anon_sym_method] = ACTIONS(2169),
    [anon_sym_in] = ACTIONS(2169),
    [anon_sym_as] = ACTIONS(2169),
    [anon_sym_initializer] = ACTIONS(2169),
    [anon_sym_AMP] = ACTIONS(2169),
    [anon_sym_POUND] = ACTIONS(2169),
    [anon_sym_COLON_COLON] = ACTIONS(2171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2171),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2171),
    [anon_sym_then] = ACTIONS(2169),
    [anon_sym_else] = ACTIONS(2169),
    [anon_sym_do] = ACTIONS(2169),
    [anon_sym_done] = ACTIONS(2169),
    [anon_sym_to] = ACTIONS(2169),
    [anon_sym_downto] = ACTIONS(2169),
    [anon_sym_new] = ACTIONS(2169),
    [anon_sym_LBRACE_LT] = ACTIONS(2171),
    [anon_sym_GT_RBRACE] = ACTIONS(2171),
    [anon_sym_begin] = ACTIONS(2169),
    [sym_ocamlyacc_value] = ACTIONS(2171),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2169),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2171),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2169),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2169),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2171),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2171),
    [aux_sym_number_token1] = ACTIONS(2171),
    [aux_sym_signed_number_token1] = ACTIONS(2169),
    [anon_sym_SQUOTE] = ACTIONS(2171),
    [anon_sym_DQUOTE] = ACTIONS(2171),
    [sym_prefix_operator] = ACTIONS(2171),
    [aux_sym_sign_operator_token1] = ACTIONS(2169),
    [sym_hash_operator] = ACTIONS(2171),
    [anon_sym_lsl] = ACTIONS(2169),
    [anon_sym_lsr] = ACTIONS(2169),
    [anon_sym_asr] = ACTIONS(2169),
    [aux_sym_pow_operator_token1] = ACTIONS(2171),
    [anon_sym_mod] = ACTIONS(2169),
    [anon_sym_land] = ACTIONS(2169),
    [anon_sym_lor] = ACTIONS(2169),
    [anon_sym_lxor] = ACTIONS(2169),
    [aux_sym_mult_operator_token1] = ACTIONS(2169),
    [aux_sym_add_operator_token1] = ACTIONS(2169),
    [sym_concat_operator] = ACTIONS(2171),
    [sym_rel_operator] = ACTIONS(2169),
    [anon_sym_AMP_AMP] = ACTIONS(2169),
    [anon_sym_or] = ACTIONS(2169),
    [anon_sym_PIPE_PIPE] = ACTIONS(2169),
    [sym_assign_operator] = ACTIONS(2171),
    [sym__uppercase_identifier] = ACTIONS(2171),
    [aux_sym_tag_token1] = ACTIONS(2171),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1002)] = {
    [sym_attribute] = STATE(1002),
    [sym__lowercase_identifier] = ACTIONS(2173),
    [anon_sym_and] = ACTIONS(2173),
    [anon_sym_TILDE] = ACTIONS(2173),
    [anon_sym_QMARK] = ACTIONS(2173),
    [anon_sym_LPAREN] = ACTIONS(2175),
    [anon_sym_RPAREN] = ACTIONS(2175),
    [anon_sym_COMMA] = ACTIONS(2175),
    [anon_sym_PIPE] = ACTIONS(2173),
    [anon_sym_COLON2] = ACTIONS(2173),
    [anon_sym_DOT] = ACTIONS(2175),
    [anon_sym_DASH_GT] = ACTIONS(2173),
    [anon_sym_LBRACK] = ACTIONS(2173),
    [anon_sym_RBRACK] = ACTIONS(2175),
    [anon_sym_true] = ACTIONS(2173),
    [anon_sym_false] = ACTIONS(2173),
    [anon_sym_LBRACE] = ACTIONS(2173),
    [anon_sym_SEMI] = ACTIONS(2175),
    [anon_sym_RBRACE] = ACTIONS(2175),
    [anon_sym_constraint] = ACTIONS(2173),
    [anon_sym_val] = ACTIONS(2173),
    [anon_sym_end] = ACTIONS(2173),
    [anon_sym_with] = ACTIONS(2173),
    [anon_sym_COLON_GT] = ACTIONS(2175),
    [anon_sym_object] = ACTIONS(2173),
    [anon_sym_inherit] = ACTIONS(2173),
    [anon_sym_method] = ACTIONS(2173),
    [anon_sym_in] = ACTIONS(2173),
    [anon_sym_as] = ACTIONS(2173),
    [anon_sym_initializer] = ACTIONS(2173),
    [anon_sym_AMP] = ACTIONS(2173),
    [anon_sym_POUND] = ACTIONS(2173),
    [anon_sym_COLON_COLON] = ACTIONS(2175),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2175),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2175),
    [anon_sym_then] = ACTIONS(2173),
    [anon_sym_else] = ACTIONS(2173),
    [anon_sym_do] = ACTIONS(2173),
    [anon_sym_done] = ACTIONS(2173),
    [anon_sym_to] = ACTIONS(2173),
    [anon_sym_downto] = ACTIONS(2173),
    [anon_sym_new] = ACTIONS(2173),
    [anon_sym_LBRACE_LT] = ACTIONS(2175),
    [anon_sym_GT_RBRACE] = ACTIONS(2175),
    [anon_sym_begin] = ACTIONS(2173),
    [sym_ocamlyacc_value] = ACTIONS(2175),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2173),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2175),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2173),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2175),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2175),
    [aux_sym_number_token1] = ACTIONS(2175),
    [aux_sym_signed_number_token1] = ACTIONS(2173),
    [anon_sym_SQUOTE] = ACTIONS(2175),
    [anon_sym_DQUOTE] = ACTIONS(2175),
    [sym_prefix_operator] = ACTIONS(2175),
    [aux_sym_sign_operator_token1] = ACTIONS(2173),
    [sym_hash_operator] = ACTIONS(2175),
    [anon_sym_lsl] = ACTIONS(2173),
    [anon_sym_lsr] = ACTIONS(2173),
    [anon_sym_asr] = ACTIONS(2173),
    [aux_sym_pow_operator_token1] = ACTIONS(2175),
    [anon_sym_mod] = ACTIONS(2173),
    [anon_sym_land] = ACTIONS(2173),
    [anon_sym_lor] = ACTIONS(2173),
    [anon_sym_lxor] = ACTIONS(2173),
    [aux_sym_mult_operator_token1] = ACTIONS(2173),
    [aux_sym_add_operator_token1] = ACTIONS(2173),
    [sym_concat_operator] = ACTIONS(2175),
    [sym_rel_operator] = ACTIONS(2173),
    [anon_sym_AMP_AMP] = ACTIONS(2173),
    [anon_sym_or] = ACTIONS(2173),
    [anon_sym_PIPE_PIPE] = ACTIONS(2173),
    [sym_assign_operator] = ACTIONS(2175),
    [sym__uppercase_identifier] = ACTIONS(2175),
    [aux_sym_tag_token1] = ACTIONS(2175),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1003)] = {
    [sym_attribute] = STATE(1003),
    [sym__lowercase_identifier] = ACTIONS(2111),
    [anon_sym_and] = ACTIONS(2111),
    [anon_sym_TILDE] = ACTIONS(2111),
    [anon_sym_QMARK] = ACTIONS(2111),
    [anon_sym_LPAREN] = ACTIONS(2113),
    [anon_sym_RPAREN] = ACTIONS(2113),
    [anon_sym_COMMA] = ACTIONS(2113),
    [anon_sym_PIPE] = ACTIONS(2111),
    [anon_sym_COLON2] = ACTIONS(2111),
    [anon_sym_DOT] = ACTIONS(2177),
    [anon_sym_DASH_GT] = ACTIONS(2111),
    [anon_sym_LBRACK] = ACTIONS(2111),
    [anon_sym_RBRACK] = ACTIONS(2113),
    [anon_sym_true] = ACTIONS(2111),
    [anon_sym_false] = ACTIONS(2111),
    [anon_sym_LBRACE] = ACTIONS(2111),
    [anon_sym_SEMI] = ACTIONS(2113),
    [anon_sym_RBRACE] = ACTIONS(2113),
    [anon_sym_constraint] = ACTIONS(2111),
    [anon_sym_val] = ACTIONS(2111),
    [anon_sym_end] = ACTIONS(2111),
    [anon_sym_with] = ACTIONS(2111),
    [anon_sym_COLON_GT] = ACTIONS(2113),
    [anon_sym_object] = ACTIONS(2111),
    [anon_sym_inherit] = ACTIONS(2111),
    [anon_sym_method] = ACTIONS(2111),
    [anon_sym_in] = ACTIONS(2111),
    [anon_sym_as] = ACTIONS(2111),
    [anon_sym_initializer] = ACTIONS(2111),
    [anon_sym_AMP] = ACTIONS(2111),
    [anon_sym_POUND] = ACTIONS(2111),
    [anon_sym_COLON_COLON] = ACTIONS(2113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2113),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2113),
    [anon_sym_then] = ACTIONS(2111),
    [anon_sym_else] = ACTIONS(2111),
    [anon_sym_do] = ACTIONS(2111),
    [anon_sym_done] = ACTIONS(2111),
    [anon_sym_to] = ACTIONS(2111),
    [anon_sym_downto] = ACTIONS(2111),
    [anon_sym_new] = ACTIONS(2111),
    [anon_sym_LBRACE_LT] = ACTIONS(2113),
    [anon_sym_GT_RBRACE] = ACTIONS(2113),
    [anon_sym_begin] = ACTIONS(2111),
    [sym_ocamlyacc_value] = ACTIONS(2113),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2113),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2113),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2113),
    [aux_sym_number_token1] = ACTIONS(2113),
    [aux_sym_signed_number_token1] = ACTIONS(2111),
    [anon_sym_SQUOTE] = ACTIONS(2113),
    [anon_sym_DQUOTE] = ACTIONS(2113),
    [sym_prefix_operator] = ACTIONS(2113),
    [aux_sym_sign_operator_token1] = ACTIONS(2111),
    [sym_hash_operator] = ACTIONS(2113),
    [anon_sym_lsl] = ACTIONS(2111),
    [anon_sym_lsr] = ACTIONS(2111),
    [anon_sym_asr] = ACTIONS(2111),
    [aux_sym_pow_operator_token1] = ACTIONS(2113),
    [anon_sym_mod] = ACTIONS(2111),
    [anon_sym_land] = ACTIONS(2111),
    [anon_sym_lor] = ACTIONS(2111),
    [anon_sym_lxor] = ACTIONS(2111),
    [aux_sym_mult_operator_token1] = ACTIONS(2111),
    [aux_sym_add_operator_token1] = ACTIONS(2111),
    [sym_concat_operator] = ACTIONS(2113),
    [sym_rel_operator] = ACTIONS(2111),
    [anon_sym_AMP_AMP] = ACTIONS(2111),
    [anon_sym_or] = ACTIONS(2111),
    [anon_sym_PIPE_PIPE] = ACTIONS(2111),
    [sym_assign_operator] = ACTIONS(2113),
    [sym__uppercase_identifier] = ACTIONS(2113),
    [aux_sym_tag_token1] = ACTIONS(2113),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1004)] = {
    [sym_attribute] = STATE(1004),
    [sym__lowercase_identifier] = ACTIONS(2179),
    [anon_sym_and] = ACTIONS(2179),
    [anon_sym_TILDE] = ACTIONS(2179),
    [anon_sym_QMARK] = ACTIONS(2179),
    [anon_sym_LPAREN] = ACTIONS(2181),
    [anon_sym_RPAREN] = ACTIONS(2181),
    [anon_sym_COMMA] = ACTIONS(2181),
    [anon_sym_PIPE] = ACTIONS(2179),
    [anon_sym_COLON2] = ACTIONS(2179),
    [anon_sym_DOT] = ACTIONS(2181),
    [anon_sym_DASH_GT] = ACTIONS(2179),
    [anon_sym_LBRACK] = ACTIONS(2179),
    [anon_sym_RBRACK] = ACTIONS(2181),
    [anon_sym_true] = ACTIONS(2179),
    [anon_sym_false] = ACTIONS(2179),
    [anon_sym_LBRACE] = ACTIONS(2179),
    [anon_sym_SEMI] = ACTIONS(2181),
    [anon_sym_RBRACE] = ACTIONS(2181),
    [anon_sym_constraint] = ACTIONS(2179),
    [anon_sym_val] = ACTIONS(2179),
    [anon_sym_end] = ACTIONS(2179),
    [anon_sym_with] = ACTIONS(2179),
    [anon_sym_COLON_GT] = ACTIONS(2181),
    [anon_sym_object] = ACTIONS(2179),
    [anon_sym_inherit] = ACTIONS(2179),
    [anon_sym_method] = ACTIONS(2179),
    [anon_sym_in] = ACTIONS(2179),
    [anon_sym_as] = ACTIONS(2179),
    [anon_sym_initializer] = ACTIONS(2179),
    [anon_sym_AMP] = ACTIONS(2179),
    [anon_sym_POUND] = ACTIONS(2179),
    [anon_sym_COLON_COLON] = ACTIONS(2181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2181),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2181),
    [anon_sym_then] = ACTIONS(2179),
    [anon_sym_else] = ACTIONS(2179),
    [anon_sym_do] = ACTIONS(2179),
    [anon_sym_done] = ACTIONS(2179),
    [anon_sym_to] = ACTIONS(2179),
    [anon_sym_downto] = ACTIONS(2179),
    [anon_sym_new] = ACTIONS(2179),
    [anon_sym_LBRACE_LT] = ACTIONS(2181),
    [anon_sym_GT_RBRACE] = ACTIONS(2181),
    [anon_sym_begin] = ACTIONS(2179),
    [sym_ocamlyacc_value] = ACTIONS(2181),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2179),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2181),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2179),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2181),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2181),
    [aux_sym_number_token1] = ACTIONS(2181),
    [aux_sym_signed_number_token1] = ACTIONS(2179),
    [anon_sym_SQUOTE] = ACTIONS(2181),
    [anon_sym_DQUOTE] = ACTIONS(2181),
    [sym_prefix_operator] = ACTIONS(2181),
    [aux_sym_sign_operator_token1] = ACTIONS(2179),
    [sym_hash_operator] = ACTIONS(2181),
    [anon_sym_lsl] = ACTIONS(2179),
    [anon_sym_lsr] = ACTIONS(2179),
    [anon_sym_asr] = ACTIONS(2179),
    [aux_sym_pow_operator_token1] = ACTIONS(2181),
    [anon_sym_mod] = ACTIONS(2179),
    [anon_sym_land] = ACTIONS(2179),
    [anon_sym_lor] = ACTIONS(2179),
    [anon_sym_lxor] = ACTIONS(2179),
    [aux_sym_mult_operator_token1] = ACTIONS(2179),
    [aux_sym_add_operator_token1] = ACTIONS(2179),
    [sym_concat_operator] = ACTIONS(2181),
    [sym_rel_operator] = ACTIONS(2179),
    [anon_sym_AMP_AMP] = ACTIONS(2179),
    [anon_sym_or] = ACTIONS(2179),
    [anon_sym_PIPE_PIPE] = ACTIONS(2179),
    [sym_assign_operator] = ACTIONS(2181),
    [sym__uppercase_identifier] = ACTIONS(2181),
    [aux_sym_tag_token1] = ACTIONS(2181),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1005)] = {
    [sym_attribute] = STATE(1005),
    [sym__lowercase_identifier] = ACTIONS(2183),
    [anon_sym_and] = ACTIONS(2183),
    [anon_sym_TILDE] = ACTIONS(2183),
    [anon_sym_QMARK] = ACTIONS(2183),
    [anon_sym_LPAREN] = ACTIONS(2185),
    [anon_sym_RPAREN] = ACTIONS(2185),
    [anon_sym_COMMA] = ACTIONS(2185),
    [anon_sym_PIPE] = ACTIONS(2183),
    [anon_sym_COLON2] = ACTIONS(2183),
    [anon_sym_DOT] = ACTIONS(2185),
    [anon_sym_DASH_GT] = ACTIONS(2183),
    [anon_sym_LBRACK] = ACTIONS(2183),
    [anon_sym_RBRACK] = ACTIONS(2185),
    [anon_sym_true] = ACTIONS(2183),
    [anon_sym_false] = ACTIONS(2183),
    [anon_sym_LBRACE] = ACTIONS(2183),
    [anon_sym_SEMI] = ACTIONS(2185),
    [anon_sym_RBRACE] = ACTIONS(2185),
    [anon_sym_constraint] = ACTIONS(2183),
    [anon_sym_val] = ACTIONS(2183),
    [anon_sym_end] = ACTIONS(2183),
    [anon_sym_with] = ACTIONS(2183),
    [anon_sym_COLON_GT] = ACTIONS(2185),
    [anon_sym_object] = ACTIONS(2183),
    [anon_sym_inherit] = ACTIONS(2183),
    [anon_sym_method] = ACTIONS(2183),
    [anon_sym_in] = ACTIONS(2183),
    [anon_sym_as] = ACTIONS(2183),
    [anon_sym_initializer] = ACTIONS(2183),
    [anon_sym_AMP] = ACTIONS(2183),
    [anon_sym_POUND] = ACTIONS(2183),
    [anon_sym_COLON_COLON] = ACTIONS(2185),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2185),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2185),
    [anon_sym_then] = ACTIONS(2183),
    [anon_sym_else] = ACTIONS(2183),
    [anon_sym_do] = ACTIONS(2183),
    [anon_sym_done] = ACTIONS(2183),
    [anon_sym_to] = ACTIONS(2183),
    [anon_sym_downto] = ACTIONS(2183),
    [anon_sym_new] = ACTIONS(2183),
    [anon_sym_LBRACE_LT] = ACTIONS(2185),
    [anon_sym_GT_RBRACE] = ACTIONS(2185),
    [anon_sym_begin] = ACTIONS(2183),
    [sym_ocamlyacc_value] = ACTIONS(2185),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2183),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2185),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2183),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2183),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2185),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2185),
    [aux_sym_number_token1] = ACTIONS(2185),
    [aux_sym_signed_number_token1] = ACTIONS(2183),
    [anon_sym_SQUOTE] = ACTIONS(2185),
    [anon_sym_DQUOTE] = ACTIONS(2185),
    [sym_prefix_operator] = ACTIONS(2185),
    [aux_sym_sign_operator_token1] = ACTIONS(2183),
    [sym_hash_operator] = ACTIONS(2185),
    [anon_sym_lsl] = ACTIONS(2183),
    [anon_sym_lsr] = ACTIONS(2183),
    [anon_sym_asr] = ACTIONS(2183),
    [aux_sym_pow_operator_token1] = ACTIONS(2185),
    [anon_sym_mod] = ACTIONS(2183),
    [anon_sym_land] = ACTIONS(2183),
    [anon_sym_lor] = ACTIONS(2183),
    [anon_sym_lxor] = ACTIONS(2183),
    [aux_sym_mult_operator_token1] = ACTIONS(2183),
    [aux_sym_add_operator_token1] = ACTIONS(2183),
    [sym_concat_operator] = ACTIONS(2185),
    [sym_rel_operator] = ACTIONS(2183),
    [anon_sym_AMP_AMP] = ACTIONS(2183),
    [anon_sym_or] = ACTIONS(2183),
    [anon_sym_PIPE_PIPE] = ACTIONS(2183),
    [sym_assign_operator] = ACTIONS(2185),
    [sym__uppercase_identifier] = ACTIONS(2185),
    [aux_sym_tag_token1] = ACTIONS(2185),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1006)] = {
    [sym_attribute] = STATE(1006),
    [sym__lowercase_identifier] = ACTIONS(2187),
    [anon_sym_and] = ACTIONS(2187),
    [anon_sym_TILDE] = ACTIONS(2187),
    [anon_sym_QMARK] = ACTIONS(2187),
    [anon_sym_LPAREN] = ACTIONS(2189),
    [anon_sym_RPAREN] = ACTIONS(2189),
    [anon_sym_COMMA] = ACTIONS(2189),
    [anon_sym_PIPE] = ACTIONS(2187),
    [anon_sym_COLON2] = ACTIONS(2187),
    [anon_sym_DOT] = ACTIONS(2189),
    [anon_sym_DASH_GT] = ACTIONS(2187),
    [anon_sym_LBRACK] = ACTIONS(2187),
    [anon_sym_RBRACK] = ACTIONS(2189),
    [anon_sym_true] = ACTIONS(2187),
    [anon_sym_false] = ACTIONS(2187),
    [anon_sym_LBRACE] = ACTIONS(2187),
    [anon_sym_SEMI] = ACTIONS(2189),
    [anon_sym_RBRACE] = ACTIONS(2189),
    [anon_sym_constraint] = ACTIONS(2187),
    [anon_sym_val] = ACTIONS(2187),
    [anon_sym_end] = ACTIONS(2187),
    [anon_sym_with] = ACTIONS(2187),
    [anon_sym_COLON_GT] = ACTIONS(2189),
    [anon_sym_object] = ACTIONS(2187),
    [anon_sym_inherit] = ACTIONS(2187),
    [anon_sym_method] = ACTIONS(2187),
    [anon_sym_in] = ACTIONS(2187),
    [anon_sym_as] = ACTIONS(2187),
    [anon_sym_initializer] = ACTIONS(2187),
    [anon_sym_AMP] = ACTIONS(2187),
    [anon_sym_POUND] = ACTIONS(2187),
    [anon_sym_COLON_COLON] = ACTIONS(2189),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2189),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2189),
    [anon_sym_then] = ACTIONS(2187),
    [anon_sym_else] = ACTIONS(2187),
    [anon_sym_do] = ACTIONS(2187),
    [anon_sym_done] = ACTIONS(2187),
    [anon_sym_to] = ACTIONS(2187),
    [anon_sym_downto] = ACTIONS(2187),
    [anon_sym_new] = ACTIONS(2187),
    [anon_sym_LBRACE_LT] = ACTIONS(2189),
    [anon_sym_GT_RBRACE] = ACTIONS(2189),
    [anon_sym_begin] = ACTIONS(2187),
    [sym_ocamlyacc_value] = ACTIONS(2189),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2187),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2189),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2187),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2187),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2189),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2189),
    [aux_sym_number_token1] = ACTIONS(2189),
    [aux_sym_signed_number_token1] = ACTIONS(2187),
    [anon_sym_SQUOTE] = ACTIONS(2189),
    [anon_sym_DQUOTE] = ACTIONS(2189),
    [sym_prefix_operator] = ACTIONS(2189),
    [aux_sym_sign_operator_token1] = ACTIONS(2187),
    [sym_hash_operator] = ACTIONS(2189),
    [anon_sym_lsl] = ACTIONS(2187),
    [anon_sym_lsr] = ACTIONS(2187),
    [anon_sym_asr] = ACTIONS(2187),
    [aux_sym_pow_operator_token1] = ACTIONS(2189),
    [anon_sym_mod] = ACTIONS(2187),
    [anon_sym_land] = ACTIONS(2187),
    [anon_sym_lor] = ACTIONS(2187),
    [anon_sym_lxor] = ACTIONS(2187),
    [aux_sym_mult_operator_token1] = ACTIONS(2187),
    [aux_sym_add_operator_token1] = ACTIONS(2187),
    [sym_concat_operator] = ACTIONS(2189),
    [sym_rel_operator] = ACTIONS(2187),
    [anon_sym_AMP_AMP] = ACTIONS(2187),
    [anon_sym_or] = ACTIONS(2187),
    [anon_sym_PIPE_PIPE] = ACTIONS(2187),
    [sym_assign_operator] = ACTIONS(2189),
    [sym__uppercase_identifier] = ACTIONS(2189),
    [aux_sym_tag_token1] = ACTIONS(2189),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1007)] = {
    [sym_attribute] = STATE(1007),
    [sym__lowercase_identifier] = ACTIONS(2191),
    [anon_sym_and] = ACTIONS(2191),
    [anon_sym_TILDE] = ACTIONS(2191),
    [anon_sym_QMARK] = ACTIONS(2191),
    [anon_sym_LPAREN] = ACTIONS(2193),
    [anon_sym_RPAREN] = ACTIONS(2193),
    [anon_sym_COMMA] = ACTIONS(2193),
    [anon_sym_PIPE] = ACTIONS(2191),
    [anon_sym_COLON2] = ACTIONS(2191),
    [anon_sym_DOT] = ACTIONS(2193),
    [anon_sym_DASH_GT] = ACTIONS(2191),
    [anon_sym_LBRACK] = ACTIONS(2191),
    [anon_sym_RBRACK] = ACTIONS(2193),
    [anon_sym_true] = ACTIONS(2191),
    [anon_sym_false] = ACTIONS(2191),
    [anon_sym_LBRACE] = ACTIONS(2191),
    [anon_sym_SEMI] = ACTIONS(2193),
    [anon_sym_RBRACE] = ACTIONS(2193),
    [anon_sym_constraint] = ACTIONS(2191),
    [anon_sym_val] = ACTIONS(2191),
    [anon_sym_end] = ACTIONS(2191),
    [anon_sym_with] = ACTIONS(2191),
    [anon_sym_COLON_GT] = ACTIONS(2193),
    [anon_sym_object] = ACTIONS(2191),
    [anon_sym_inherit] = ACTIONS(2191),
    [anon_sym_method] = ACTIONS(2191),
    [anon_sym_in] = ACTIONS(2191),
    [anon_sym_as] = ACTIONS(2191),
    [anon_sym_initializer] = ACTIONS(2191),
    [anon_sym_AMP] = ACTIONS(2191),
    [anon_sym_POUND] = ACTIONS(2191),
    [anon_sym_COLON_COLON] = ACTIONS(2193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2193),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2193),
    [anon_sym_then] = ACTIONS(2191),
    [anon_sym_else] = ACTIONS(2191),
    [anon_sym_do] = ACTIONS(2191),
    [anon_sym_done] = ACTIONS(2191),
    [anon_sym_to] = ACTIONS(2191),
    [anon_sym_downto] = ACTIONS(2191),
    [anon_sym_new] = ACTIONS(2191),
    [anon_sym_LBRACE_LT] = ACTIONS(2193),
    [anon_sym_GT_RBRACE] = ACTIONS(2193),
    [anon_sym_begin] = ACTIONS(2191),
    [sym_ocamlyacc_value] = ACTIONS(2193),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2191),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2193),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2191),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2191),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2193),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2193),
    [aux_sym_number_token1] = ACTIONS(2193),
    [aux_sym_signed_number_token1] = ACTIONS(2191),
    [anon_sym_SQUOTE] = ACTIONS(2193),
    [anon_sym_DQUOTE] = ACTIONS(2193),
    [sym_prefix_operator] = ACTIONS(2193),
    [aux_sym_sign_operator_token1] = ACTIONS(2191),
    [sym_hash_operator] = ACTIONS(2193),
    [anon_sym_lsl] = ACTIONS(2191),
    [anon_sym_lsr] = ACTIONS(2191),
    [anon_sym_asr] = ACTIONS(2191),
    [aux_sym_pow_operator_token1] = ACTIONS(2193),
    [anon_sym_mod] = ACTIONS(2191),
    [anon_sym_land] = ACTIONS(2191),
    [anon_sym_lor] = ACTIONS(2191),
    [anon_sym_lxor] = ACTIONS(2191),
    [aux_sym_mult_operator_token1] = ACTIONS(2191),
    [aux_sym_add_operator_token1] = ACTIONS(2191),
    [sym_concat_operator] = ACTIONS(2193),
    [sym_rel_operator] = ACTIONS(2191),
    [anon_sym_AMP_AMP] = ACTIONS(2191),
    [anon_sym_or] = ACTIONS(2191),
    [anon_sym_PIPE_PIPE] = ACTIONS(2191),
    [sym_assign_operator] = ACTIONS(2193),
    [sym__uppercase_identifier] = ACTIONS(2193),
    [aux_sym_tag_token1] = ACTIONS(2193),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1008)] = {
    [sym_attribute] = STATE(1008),
    [sym__lowercase_identifier] = ACTIONS(2195),
    [anon_sym_and] = ACTIONS(2195),
    [anon_sym_TILDE] = ACTIONS(2195),
    [anon_sym_QMARK] = ACTIONS(2195),
    [anon_sym_LPAREN] = ACTIONS(2197),
    [anon_sym_RPAREN] = ACTIONS(2197),
    [anon_sym_COMMA] = ACTIONS(2197),
    [anon_sym_PIPE] = ACTIONS(2195),
    [anon_sym_COLON2] = ACTIONS(2195),
    [anon_sym_DOT] = ACTIONS(2197),
    [anon_sym_DASH_GT] = ACTIONS(2195),
    [anon_sym_LBRACK] = ACTIONS(2195),
    [anon_sym_RBRACK] = ACTIONS(2197),
    [anon_sym_true] = ACTIONS(2195),
    [anon_sym_false] = ACTIONS(2195),
    [anon_sym_LBRACE] = ACTIONS(2195),
    [anon_sym_SEMI] = ACTIONS(2197),
    [anon_sym_RBRACE] = ACTIONS(2197),
    [anon_sym_constraint] = ACTIONS(2195),
    [anon_sym_val] = ACTIONS(2195),
    [anon_sym_end] = ACTIONS(2195),
    [anon_sym_with] = ACTIONS(2195),
    [anon_sym_COLON_GT] = ACTIONS(2197),
    [anon_sym_object] = ACTIONS(2195),
    [anon_sym_inherit] = ACTIONS(2195),
    [anon_sym_method] = ACTIONS(2195),
    [anon_sym_in] = ACTIONS(2195),
    [anon_sym_as] = ACTIONS(2195),
    [anon_sym_initializer] = ACTIONS(2195),
    [anon_sym_AMP] = ACTIONS(2195),
    [anon_sym_POUND] = ACTIONS(2195),
    [anon_sym_COLON_COLON] = ACTIONS(2197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2197),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2197),
    [anon_sym_then] = ACTIONS(2195),
    [anon_sym_else] = ACTIONS(2195),
    [anon_sym_do] = ACTIONS(2195),
    [anon_sym_done] = ACTIONS(2195),
    [anon_sym_to] = ACTIONS(2195),
    [anon_sym_downto] = ACTIONS(2195),
    [anon_sym_new] = ACTIONS(2195),
    [anon_sym_LBRACE_LT] = ACTIONS(2197),
    [anon_sym_GT_RBRACE] = ACTIONS(2197),
    [anon_sym_begin] = ACTIONS(2195),
    [sym_ocamlyacc_value] = ACTIONS(2197),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2195),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2197),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2195),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2195),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2197),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2197),
    [aux_sym_number_token1] = ACTIONS(2197),
    [aux_sym_signed_number_token1] = ACTIONS(2195),
    [anon_sym_SQUOTE] = ACTIONS(2197),
    [anon_sym_DQUOTE] = ACTIONS(2197),
    [sym_prefix_operator] = ACTIONS(2197),
    [aux_sym_sign_operator_token1] = ACTIONS(2195),
    [sym_hash_operator] = ACTIONS(2197),
    [anon_sym_lsl] = ACTIONS(2195),
    [anon_sym_lsr] = ACTIONS(2195),
    [anon_sym_asr] = ACTIONS(2195),
    [aux_sym_pow_operator_token1] = ACTIONS(2197),
    [anon_sym_mod] = ACTIONS(2195),
    [anon_sym_land] = ACTIONS(2195),
    [anon_sym_lor] = ACTIONS(2195),
    [anon_sym_lxor] = ACTIONS(2195),
    [aux_sym_mult_operator_token1] = ACTIONS(2195),
    [aux_sym_add_operator_token1] = ACTIONS(2195),
    [sym_concat_operator] = ACTIONS(2197),
    [sym_rel_operator] = ACTIONS(2195),
    [anon_sym_AMP_AMP] = ACTIONS(2195),
    [anon_sym_or] = ACTIONS(2195),
    [anon_sym_PIPE_PIPE] = ACTIONS(2195),
    [sym_assign_operator] = ACTIONS(2197),
    [sym__uppercase_identifier] = ACTIONS(2197),
    [aux_sym_tag_token1] = ACTIONS(2197),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1009)] = {
    [sym_attribute] = STATE(1009),
    [sym__lowercase_identifier] = ACTIONS(2199),
    [anon_sym_and] = ACTIONS(2199),
    [anon_sym_TILDE] = ACTIONS(2199),
    [anon_sym_QMARK] = ACTIONS(2199),
    [anon_sym_LPAREN] = ACTIONS(2201),
    [anon_sym_RPAREN] = ACTIONS(2201),
    [anon_sym_COMMA] = ACTIONS(2201),
    [anon_sym_PIPE] = ACTIONS(2199),
    [anon_sym_COLON2] = ACTIONS(2199),
    [anon_sym_DOT] = ACTIONS(2201),
    [anon_sym_DASH_GT] = ACTIONS(2199),
    [anon_sym_LBRACK] = ACTIONS(2199),
    [anon_sym_RBRACK] = ACTIONS(2201),
    [anon_sym_true] = ACTIONS(2199),
    [anon_sym_false] = ACTIONS(2199),
    [anon_sym_LBRACE] = ACTIONS(2199),
    [anon_sym_SEMI] = ACTIONS(2201),
    [anon_sym_RBRACE] = ACTIONS(2201),
    [anon_sym_constraint] = ACTIONS(2199),
    [anon_sym_val] = ACTIONS(2199),
    [anon_sym_end] = ACTIONS(2199),
    [anon_sym_with] = ACTIONS(2199),
    [anon_sym_COLON_GT] = ACTIONS(2201),
    [anon_sym_object] = ACTIONS(2199),
    [anon_sym_inherit] = ACTIONS(2199),
    [anon_sym_method] = ACTIONS(2199),
    [anon_sym_in] = ACTIONS(2199),
    [anon_sym_as] = ACTIONS(2199),
    [anon_sym_initializer] = ACTIONS(2199),
    [anon_sym_AMP] = ACTIONS(2199),
    [anon_sym_POUND] = ACTIONS(2199),
    [anon_sym_COLON_COLON] = ACTIONS(2201),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2201),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2201),
    [anon_sym_then] = ACTIONS(2199),
    [anon_sym_else] = ACTIONS(2199),
    [anon_sym_do] = ACTIONS(2199),
    [anon_sym_done] = ACTIONS(2199),
    [anon_sym_to] = ACTIONS(2199),
    [anon_sym_downto] = ACTIONS(2199),
    [anon_sym_new] = ACTIONS(2199),
    [anon_sym_LBRACE_LT] = ACTIONS(2201),
    [anon_sym_GT_RBRACE] = ACTIONS(2201),
    [anon_sym_begin] = ACTIONS(2199),
    [sym_ocamlyacc_value] = ACTIONS(2201),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2199),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2201),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2199),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2199),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2201),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2201),
    [aux_sym_number_token1] = ACTIONS(2201),
    [aux_sym_signed_number_token1] = ACTIONS(2199),
    [anon_sym_SQUOTE] = ACTIONS(2201),
    [anon_sym_DQUOTE] = ACTIONS(2201),
    [sym_prefix_operator] = ACTIONS(2201),
    [aux_sym_sign_operator_token1] = ACTIONS(2199),
    [sym_hash_operator] = ACTIONS(2201),
    [anon_sym_lsl] = ACTIONS(2199),
    [anon_sym_lsr] = ACTIONS(2199),
    [anon_sym_asr] = ACTIONS(2199),
    [aux_sym_pow_operator_token1] = ACTIONS(2201),
    [anon_sym_mod] = ACTIONS(2199),
    [anon_sym_land] = ACTIONS(2199),
    [anon_sym_lor] = ACTIONS(2199),
    [anon_sym_lxor] = ACTIONS(2199),
    [aux_sym_mult_operator_token1] = ACTIONS(2199),
    [aux_sym_add_operator_token1] = ACTIONS(2199),
    [sym_concat_operator] = ACTIONS(2201),
    [sym_rel_operator] = ACTIONS(2199),
    [anon_sym_AMP_AMP] = ACTIONS(2199),
    [anon_sym_or] = ACTIONS(2199),
    [anon_sym_PIPE_PIPE] = ACTIONS(2199),
    [sym_assign_operator] = ACTIONS(2201),
    [sym__uppercase_identifier] = ACTIONS(2201),
    [aux_sym_tag_token1] = ACTIONS(2201),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1010)] = {
    [sym_attribute] = STATE(1010),
    [sym__lowercase_identifier] = ACTIONS(2203),
    [anon_sym_and] = ACTIONS(2203),
    [anon_sym_TILDE] = ACTIONS(2203),
    [anon_sym_QMARK] = ACTIONS(2203),
    [anon_sym_LPAREN] = ACTIONS(2205),
    [anon_sym_RPAREN] = ACTIONS(2205),
    [anon_sym_COMMA] = ACTIONS(2205),
    [anon_sym_PIPE] = ACTIONS(2203),
    [anon_sym_COLON2] = ACTIONS(2203),
    [anon_sym_DOT] = ACTIONS(2205),
    [anon_sym_DASH_GT] = ACTIONS(2203),
    [anon_sym_LBRACK] = ACTIONS(2203),
    [anon_sym_RBRACK] = ACTIONS(2205),
    [anon_sym_true] = ACTIONS(2203),
    [anon_sym_false] = ACTIONS(2203),
    [anon_sym_LBRACE] = ACTIONS(2203),
    [anon_sym_SEMI] = ACTIONS(2205),
    [anon_sym_RBRACE] = ACTIONS(2205),
    [anon_sym_constraint] = ACTIONS(2203),
    [anon_sym_val] = ACTIONS(2203),
    [anon_sym_end] = ACTIONS(2203),
    [anon_sym_with] = ACTIONS(2203),
    [anon_sym_COLON_GT] = ACTIONS(2205),
    [anon_sym_object] = ACTIONS(2203),
    [anon_sym_inherit] = ACTIONS(2203),
    [anon_sym_method] = ACTIONS(2203),
    [anon_sym_in] = ACTIONS(2203),
    [anon_sym_as] = ACTIONS(2203),
    [anon_sym_initializer] = ACTIONS(2203),
    [anon_sym_AMP] = ACTIONS(2203),
    [anon_sym_POUND] = ACTIONS(2203),
    [anon_sym_COLON_COLON] = ACTIONS(2205),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2205),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2205),
    [anon_sym_then] = ACTIONS(2203),
    [anon_sym_else] = ACTIONS(2203),
    [anon_sym_do] = ACTIONS(2203),
    [anon_sym_done] = ACTIONS(2203),
    [anon_sym_to] = ACTIONS(2203),
    [anon_sym_downto] = ACTIONS(2203),
    [anon_sym_new] = ACTIONS(2203),
    [anon_sym_LBRACE_LT] = ACTIONS(2205),
    [anon_sym_GT_RBRACE] = ACTIONS(2205),
    [anon_sym_begin] = ACTIONS(2203),
    [sym_ocamlyacc_value] = ACTIONS(2205),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2203),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2205),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2203),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2203),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2205),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2205),
    [aux_sym_number_token1] = ACTIONS(2205),
    [aux_sym_signed_number_token1] = ACTIONS(2203),
    [anon_sym_SQUOTE] = ACTIONS(2205),
    [anon_sym_DQUOTE] = ACTIONS(2205),
    [sym_prefix_operator] = ACTIONS(2205),
    [aux_sym_sign_operator_token1] = ACTIONS(2203),
    [sym_hash_operator] = ACTIONS(2205),
    [anon_sym_lsl] = ACTIONS(2203),
    [anon_sym_lsr] = ACTIONS(2203),
    [anon_sym_asr] = ACTIONS(2203),
    [aux_sym_pow_operator_token1] = ACTIONS(2205),
    [anon_sym_mod] = ACTIONS(2203),
    [anon_sym_land] = ACTIONS(2203),
    [anon_sym_lor] = ACTIONS(2203),
    [anon_sym_lxor] = ACTIONS(2203),
    [aux_sym_mult_operator_token1] = ACTIONS(2203),
    [aux_sym_add_operator_token1] = ACTIONS(2203),
    [sym_concat_operator] = ACTIONS(2205),
    [sym_rel_operator] = ACTIONS(2203),
    [anon_sym_AMP_AMP] = ACTIONS(2203),
    [anon_sym_or] = ACTIONS(2203),
    [anon_sym_PIPE_PIPE] = ACTIONS(2203),
    [sym_assign_operator] = ACTIONS(2205),
    [sym__uppercase_identifier] = ACTIONS(2205),
    [aux_sym_tag_token1] = ACTIONS(2205),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1011)] = {
    [sym_attribute] = STATE(1011),
    [sym__lowercase_identifier] = ACTIONS(2207),
    [anon_sym_and] = ACTIONS(2207),
    [anon_sym_TILDE] = ACTIONS(2207),
    [anon_sym_QMARK] = ACTIONS(2207),
    [anon_sym_LPAREN] = ACTIONS(2209),
    [anon_sym_RPAREN] = ACTIONS(2209),
    [anon_sym_COMMA] = ACTIONS(2209),
    [anon_sym_PIPE] = ACTIONS(2207),
    [anon_sym_COLON2] = ACTIONS(2207),
    [anon_sym_DOT] = ACTIONS(2209),
    [anon_sym_DASH_GT] = ACTIONS(2207),
    [anon_sym_LBRACK] = ACTIONS(2207),
    [anon_sym_RBRACK] = ACTIONS(2209),
    [anon_sym_true] = ACTIONS(2207),
    [anon_sym_false] = ACTIONS(2207),
    [anon_sym_LBRACE] = ACTIONS(2207),
    [anon_sym_SEMI] = ACTIONS(2209),
    [anon_sym_RBRACE] = ACTIONS(2209),
    [anon_sym_constraint] = ACTIONS(2207),
    [anon_sym_val] = ACTIONS(2207),
    [anon_sym_end] = ACTIONS(2207),
    [anon_sym_with] = ACTIONS(2207),
    [anon_sym_COLON_GT] = ACTIONS(2209),
    [anon_sym_object] = ACTIONS(2207),
    [anon_sym_inherit] = ACTIONS(2207),
    [anon_sym_method] = ACTIONS(2207),
    [anon_sym_in] = ACTIONS(2207),
    [anon_sym_as] = ACTIONS(2207),
    [anon_sym_initializer] = ACTIONS(2207),
    [anon_sym_AMP] = ACTIONS(2207),
    [anon_sym_POUND] = ACTIONS(2207),
    [anon_sym_COLON_COLON] = ACTIONS(2209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2209),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2209),
    [anon_sym_then] = ACTIONS(2207),
    [anon_sym_else] = ACTIONS(2207),
    [anon_sym_do] = ACTIONS(2207),
    [anon_sym_done] = ACTIONS(2207),
    [anon_sym_to] = ACTIONS(2207),
    [anon_sym_downto] = ACTIONS(2207),
    [anon_sym_new] = ACTIONS(2207),
    [anon_sym_LBRACE_LT] = ACTIONS(2209),
    [anon_sym_GT_RBRACE] = ACTIONS(2209),
    [anon_sym_begin] = ACTIONS(2207),
    [sym_ocamlyacc_value] = ACTIONS(2209),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2207),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2209),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2207),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2207),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2209),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2209),
    [aux_sym_number_token1] = ACTIONS(2209),
    [aux_sym_signed_number_token1] = ACTIONS(2207),
    [anon_sym_SQUOTE] = ACTIONS(2209),
    [anon_sym_DQUOTE] = ACTIONS(2209),
    [sym_prefix_operator] = ACTIONS(2209),
    [aux_sym_sign_operator_token1] = ACTIONS(2207),
    [sym_hash_operator] = ACTIONS(2209),
    [anon_sym_lsl] = ACTIONS(2207),
    [anon_sym_lsr] = ACTIONS(2207),
    [anon_sym_asr] = ACTIONS(2207),
    [aux_sym_pow_operator_token1] = ACTIONS(2209),
    [anon_sym_mod] = ACTIONS(2207),
    [anon_sym_land] = ACTIONS(2207),
    [anon_sym_lor] = ACTIONS(2207),
    [anon_sym_lxor] = ACTIONS(2207),
    [aux_sym_mult_operator_token1] = ACTIONS(2207),
    [aux_sym_add_operator_token1] = ACTIONS(2207),
    [sym_concat_operator] = ACTIONS(2209),
    [sym_rel_operator] = ACTIONS(2207),
    [anon_sym_AMP_AMP] = ACTIONS(2207),
    [anon_sym_or] = ACTIONS(2207),
    [anon_sym_PIPE_PIPE] = ACTIONS(2207),
    [sym_assign_operator] = ACTIONS(2209),
    [sym__uppercase_identifier] = ACTIONS(2209),
    [aux_sym_tag_token1] = ACTIONS(2209),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1012)] = {
    [sym__delimited_expression] = STATE(3044),
    [sym__simple_expression] = STATE(3058),
    [sym_typed_expression] = STATE(3044),
    [sym_list_expression] = STATE(3049),
    [sym_array_expression] = STATE(3049),
    [sym_record_expression] = STATE(3049),
    [sym_labeled_argument] = STATE(3173),
    [sym_prefix_expression] = STATE(3044),
    [sym_hash_expression] = STATE(3044),
    [sym_field_get_expression] = STATE(3044),
    [sym_array_get_expression] = STATE(3044),
    [sym_string_get_expression] = STATE(3044),
    [sym_bigarray_get_expression] = STATE(3044),
    [sym_local_open_expression] = STATE(3044),
    [sym_package_expression] = STATE(3049),
    [sym_new_expression] = STATE(3044),
    [sym_object_copy_expression] = STATE(3049),
    [sym_method_invocation] = STATE(3044),
    [sym_object_expression] = STATE(3044),
    [sym_parenthesized_expression] = STATE(3049),
    [sym_attribute] = STATE(1012),
    [sym_extension] = STATE(3044),
    [sym_quoted_extension] = STATE(3044),
    [sym__constant] = STATE(3044),
    [sym_number] = STATE(2971),
    [sym_character] = STATE(2971),
    [sym_string] = STATE(2971),
    [sym_quoted_string] = STATE(2971),
    [sym_boolean] = STATE(2971),
    [sym_unit] = STATE(3049),
    [sym__value_name] = STATE(3037),
    [sym_parenthesized_operator] = STATE(3038),
    [sym_value_path] = STATE(3044),
    [sym_module_path] = STATE(7634),
    [sym_constructor_path] = STATE(3044),
    [sym__constructor_name] = STATE(2973),
    [sym_tag] = STATE(3044),
    [aux_sym_class_application_repeat1] = STATE(1014),
    [sym__lowercase_identifier] = ACTIONS(2211),
    [anon_sym_SEMI_SEMI] = ACTIONS(1883),
    [anon_sym_and] = ACTIONS(1885),
    [anon_sym_TILDE] = ACTIONS(2213),
    [anon_sym_QMARK] = ACTIONS(2213),
    [anon_sym_LPAREN] = ACTIONS(2215),
    [anon_sym_external] = ACTIONS(1885),
    [anon_sym_type] = ACTIONS(1885),
    [anon_sym_LBRACK] = ACTIONS(2217),
    [anon_sym_RBRACK] = ACTIONS(1883),
    [anon_sym_true] = ACTIONS(2219),
    [anon_sym_false] = ACTIONS(2219),
    [anon_sym_LBRACE] = ACTIONS(2221),
    [anon_sym_exception] = ACTIONS(1885),
    [anon_sym_module] = ACTIONS(1885),
    [anon_sym_open] = ACTIONS(1885),
    [anon_sym_include] = ACTIONS(1885),
    [anon_sym_class] = ACTIONS(1885),
    [anon_sym_val] = ACTIONS(1885),
    [anon_sym_end] = ACTIONS(1885),
    [anon_sym_object] = ACTIONS(2223),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2225),
    [anon_sym_new] = ACTIONS(2227),
    [anon_sym_LBRACE_LT] = ACTIONS(2229),
    [anon_sym_begin] = ACTIONS(2231),
    [sym_ocamlyacc_value] = ACTIONS(2233),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1885),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1883),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2235),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2237),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1883),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1883),
    [aux_sym_number_token1] = ACTIONS(2239),
    [anon_sym_SQUOTE] = ACTIONS(2241),
    [anon_sym_DQUOTE] = ACTIONS(2243),
    [sym_prefix_operator] = ACTIONS(2245),
    [sym__uppercase_identifier] = ACTIONS(2247),
    [aux_sym_tag_token1] = ACTIONS(2249),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1013)] = {
    [sym__delimited_expression] = STATE(3044),
    [sym__simple_expression] = STATE(3058),
    [sym_typed_expression] = STATE(3044),
    [sym_list_expression] = STATE(3049),
    [sym_array_expression] = STATE(3049),
    [sym_record_expression] = STATE(3049),
    [sym_labeled_argument] = STATE(3173),
    [sym_prefix_expression] = STATE(3044),
    [sym_hash_expression] = STATE(3044),
    [sym_field_get_expression] = STATE(3044),
    [sym_array_get_expression] = STATE(3044),
    [sym_string_get_expression] = STATE(3044),
    [sym_bigarray_get_expression] = STATE(3044),
    [sym_local_open_expression] = STATE(3044),
    [sym_package_expression] = STATE(3049),
    [sym_new_expression] = STATE(3044),
    [sym_object_copy_expression] = STATE(3049),
    [sym_method_invocation] = STATE(3044),
    [sym_object_expression] = STATE(3044),
    [sym_parenthesized_expression] = STATE(3049),
    [sym_attribute] = STATE(1013),
    [sym_extension] = STATE(3044),
    [sym_quoted_extension] = STATE(3044),
    [sym__constant] = STATE(3044),
    [sym_number] = STATE(2971),
    [sym_character] = STATE(2971),
    [sym_string] = STATE(2971),
    [sym_quoted_string] = STATE(2971),
    [sym_boolean] = STATE(2971),
    [sym_unit] = STATE(3049),
    [sym__value_name] = STATE(3037),
    [sym_parenthesized_operator] = STATE(3038),
    [sym_value_path] = STATE(3044),
    [sym_module_path] = STATE(7634),
    [sym_constructor_path] = STATE(3044),
    [sym__constructor_name] = STATE(2973),
    [sym_tag] = STATE(3044),
    [aux_sym_class_application_repeat1] = STATE(1012),
    [sym__lowercase_identifier] = ACTIONS(2211),
    [anon_sym_SEMI_SEMI] = ACTIONS(1879),
    [anon_sym_and] = ACTIONS(1881),
    [anon_sym_TILDE] = ACTIONS(2213),
    [anon_sym_QMARK] = ACTIONS(2213),
    [anon_sym_LPAREN] = ACTIONS(2215),
    [anon_sym_external] = ACTIONS(1881),
    [anon_sym_type] = ACTIONS(1881),
    [anon_sym_LBRACK] = ACTIONS(2217),
    [anon_sym_RBRACK] = ACTIONS(1879),
    [anon_sym_true] = ACTIONS(2219),
    [anon_sym_false] = ACTIONS(2219),
    [anon_sym_LBRACE] = ACTIONS(2221),
    [anon_sym_exception] = ACTIONS(1881),
    [anon_sym_module] = ACTIONS(1881),
    [anon_sym_open] = ACTIONS(1881),
    [anon_sym_include] = ACTIONS(1881),
    [anon_sym_class] = ACTIONS(1881),
    [anon_sym_val] = ACTIONS(1881),
    [anon_sym_end] = ACTIONS(1881),
    [anon_sym_object] = ACTIONS(2223),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2225),
    [anon_sym_new] = ACTIONS(2227),
    [anon_sym_LBRACE_LT] = ACTIONS(2229),
    [anon_sym_begin] = ACTIONS(2231),
    [sym_ocamlyacc_value] = ACTIONS(2233),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1881),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1879),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2235),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2237),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1879),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1879),
    [aux_sym_number_token1] = ACTIONS(2239),
    [anon_sym_SQUOTE] = ACTIONS(2241),
    [anon_sym_DQUOTE] = ACTIONS(2243),
    [sym_prefix_operator] = ACTIONS(2245),
    [sym__uppercase_identifier] = ACTIONS(2247),
    [aux_sym_tag_token1] = ACTIONS(2249),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1014)] = {
    [sym__delimited_expression] = STATE(3044),
    [sym__simple_expression] = STATE(3058),
    [sym_typed_expression] = STATE(3044),
    [sym_list_expression] = STATE(3049),
    [sym_array_expression] = STATE(3049),
    [sym_record_expression] = STATE(3049),
    [sym_labeled_argument] = STATE(3173),
    [sym_prefix_expression] = STATE(3044),
    [sym_hash_expression] = STATE(3044),
    [sym_field_get_expression] = STATE(3044),
    [sym_array_get_expression] = STATE(3044),
    [sym_string_get_expression] = STATE(3044),
    [sym_bigarray_get_expression] = STATE(3044),
    [sym_local_open_expression] = STATE(3044),
    [sym_package_expression] = STATE(3049),
    [sym_new_expression] = STATE(3044),
    [sym_object_copy_expression] = STATE(3049),
    [sym_method_invocation] = STATE(3044),
    [sym_object_expression] = STATE(3044),
    [sym_parenthesized_expression] = STATE(3049),
    [sym_attribute] = STATE(1014),
    [sym_extension] = STATE(3044),
    [sym_quoted_extension] = STATE(3044),
    [sym__constant] = STATE(3044),
    [sym_number] = STATE(2971),
    [sym_character] = STATE(2971),
    [sym_string] = STATE(2971),
    [sym_quoted_string] = STATE(2971),
    [sym_boolean] = STATE(2971),
    [sym_unit] = STATE(3049),
    [sym__value_name] = STATE(3037),
    [sym_parenthesized_operator] = STATE(3038),
    [sym_value_path] = STATE(3044),
    [sym_module_path] = STATE(7634),
    [sym_constructor_path] = STATE(3044),
    [sym__constructor_name] = STATE(2973),
    [sym_tag] = STATE(3044),
    [aux_sym_class_application_repeat1] = STATE(1014),
    [sym__lowercase_identifier] = ACTIONS(2251),
    [anon_sym_SEMI_SEMI] = ACTIONS(936),
    [anon_sym_and] = ACTIONS(928),
    [anon_sym_TILDE] = ACTIONS(2254),
    [anon_sym_QMARK] = ACTIONS(2254),
    [anon_sym_LPAREN] = ACTIONS(2257),
    [anon_sym_external] = ACTIONS(928),
    [anon_sym_type] = ACTIONS(928),
    [anon_sym_LBRACK] = ACTIONS(2260),
    [anon_sym_RBRACK] = ACTIONS(936),
    [anon_sym_true] = ACTIONS(2263),
    [anon_sym_false] = ACTIONS(2263),
    [anon_sym_LBRACE] = ACTIONS(2266),
    [anon_sym_exception] = ACTIONS(928),
    [anon_sym_module] = ACTIONS(928),
    [anon_sym_open] = ACTIONS(928),
    [anon_sym_include] = ACTIONS(928),
    [anon_sym_class] = ACTIONS(928),
    [anon_sym_val] = ACTIONS(928),
    [anon_sym_end] = ACTIONS(928),
    [anon_sym_object] = ACTIONS(2269),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2272),
    [anon_sym_new] = ACTIONS(2275),
    [anon_sym_LBRACE_LT] = ACTIONS(2278),
    [anon_sym_begin] = ACTIONS(2281),
    [sym_ocamlyacc_value] = ACTIONS(2284),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(928),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(936),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2287),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2290),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(936),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(936),
    [aux_sym_number_token1] = ACTIONS(2293),
    [anon_sym_SQUOTE] = ACTIONS(2296),
    [anon_sym_DQUOTE] = ACTIONS(2299),
    [sym_prefix_operator] = ACTIONS(2302),
    [sym__uppercase_identifier] = ACTIONS(2305),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1015)] = {
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(942),
    [sym_typed_expression] = STATE(946),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_labeled_argument] = STATE(1023),
    [sym_prefix_expression] = STATE(946),
    [sym_hash_expression] = STATE(946),
    [sym_field_get_expression] = STATE(946),
    [sym_array_get_expression] = STATE(946),
    [sym_string_get_expression] = STATE(946),
    [sym_bigarray_get_expression] = STATE(946),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(1015),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [aux_sym_class_application_repeat1] = STATE(99),
    [sym__lowercase_identifier] = ACTIONS(911),
    [anon_sym_and] = ACTIONS(1885),
    [anon_sym_TILDE] = ACTIONS(913),
    [anon_sym_QMARK] = ACTIONS(913),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1883),
    [anon_sym_COLON2] = ACTIONS(1883),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_constraint] = ACTIONS(1885),
    [anon_sym_val] = ACTIONS(1885),
    [anon_sym_end] = ACTIONS(1885),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_inherit] = ACTIONS(1885),
    [anon_sym_method] = ACTIONS(1885),
    [anon_sym_in] = ACTIONS(1885),
    [anon_sym_as] = ACTIONS(1885),
    [anon_sym_initializer] = ACTIONS(1885),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1885),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1883),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(527),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(529),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1883),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1883),
    [aux_sym_number_token1] = ACTIONS(531),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1016)] = {
    [sym_attribute] = STATE(1016),
    [sym__lowercase_identifier] = ACTIONS(2311),
    [anon_sym_and] = ACTIONS(2311),
    [anon_sym_TILDE] = ACTIONS(2311),
    [anon_sym_QMARK] = ACTIONS(2311),
    [anon_sym_COLON] = ACTIONS(2313),
    [anon_sym_LPAREN] = ACTIONS(2315),
    [anon_sym_RPAREN] = ACTIONS(2315),
    [anon_sym_COMMA] = ACTIONS(2315),
    [anon_sym_PIPE] = ACTIONS(2311),
    [anon_sym_COLON2] = ACTIONS(2311),
    [anon_sym_DASH_GT] = ACTIONS(2311),
    [anon_sym_LBRACK] = ACTIONS(2311),
    [anon_sym_RBRACK] = ACTIONS(2315),
    [anon_sym_true] = ACTIONS(2311),
    [anon_sym_false] = ACTIONS(2311),
    [anon_sym_LBRACE] = ACTIONS(2311),
    [anon_sym_SEMI] = ACTIONS(2315),
    [anon_sym_RBRACE] = ACTIONS(2315),
    [anon_sym_constraint] = ACTIONS(2311),
    [anon_sym_val] = ACTIONS(2311),
    [anon_sym_end] = ACTIONS(2311),
    [anon_sym_with] = ACTIONS(2311),
    [anon_sym_COLON_GT] = ACTIONS(2315),
    [anon_sym_object] = ACTIONS(2311),
    [anon_sym_inherit] = ACTIONS(2311),
    [anon_sym_method] = ACTIONS(2311),
    [anon_sym_in] = ACTIONS(2311),
    [anon_sym_as] = ACTIONS(2311),
    [anon_sym_initializer] = ACTIONS(2311),
    [anon_sym_AMP] = ACTIONS(2311),
    [anon_sym_COLON_COLON] = ACTIONS(2315),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2315),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2315),
    [anon_sym_then] = ACTIONS(2311),
    [anon_sym_else] = ACTIONS(2311),
    [anon_sym_do] = ACTIONS(2311),
    [anon_sym_done] = ACTIONS(2311),
    [anon_sym_to] = ACTIONS(2311),
    [anon_sym_downto] = ACTIONS(2311),
    [anon_sym_new] = ACTIONS(2311),
    [anon_sym_LBRACE_LT] = ACTIONS(2315),
    [anon_sym_GT_RBRACE] = ACTIONS(2315),
    [anon_sym_begin] = ACTIONS(2311),
    [sym_ocamlyacc_value] = ACTIONS(2315),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2311),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2315),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2311),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2311),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2315),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2315),
    [aux_sym_number_token1] = ACTIONS(2315),
    [aux_sym_signed_number_token1] = ACTIONS(2311),
    [anon_sym_SQUOTE] = ACTIONS(2315),
    [anon_sym_DQUOTE] = ACTIONS(2315),
    [sym_prefix_operator] = ACTIONS(2315),
    [aux_sym_sign_operator_token1] = ACTIONS(2311),
    [anon_sym_lsl] = ACTIONS(2311),
    [anon_sym_lsr] = ACTIONS(2311),
    [anon_sym_asr] = ACTIONS(2311),
    [aux_sym_pow_operator_token1] = ACTIONS(2315),
    [anon_sym_mod] = ACTIONS(2311),
    [anon_sym_land] = ACTIONS(2311),
    [anon_sym_lor] = ACTIONS(2311),
    [anon_sym_lxor] = ACTIONS(2311),
    [aux_sym_mult_operator_token1] = ACTIONS(2311),
    [aux_sym_add_operator_token1] = ACTIONS(2311),
    [sym_concat_operator] = ACTIONS(2315),
    [sym_rel_operator] = ACTIONS(2311),
    [anon_sym_AMP_AMP] = ACTIONS(2311),
    [anon_sym_or] = ACTIONS(2311),
    [anon_sym_PIPE_PIPE] = ACTIONS(2311),
    [sym_assign_operator] = ACTIONS(2315),
    [sym__uppercase_identifier] = ACTIONS(2315),
    [aux_sym_tag_token1] = ACTIONS(2315),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1017)] = {
    [sym_attribute] = STATE(1017),
    [sym__lowercase_identifier] = ACTIONS(2157),
    [anon_sym_TILDE] = ACTIONS(2157),
    [anon_sym_QMARK] = ACTIONS(2157),
    [anon_sym_LPAREN] = ACTIONS(2159),
    [anon_sym_RPAREN] = ACTIONS(2159),
    [anon_sym_COMMA] = ACTIONS(2159),
    [anon_sym_PIPE] = ACTIONS(2157),
    [anon_sym_COLON2] = ACTIONS(2157),
    [anon_sym_DOT] = ACTIONS(2159),
    [anon_sym_DASH_GT] = ACTIONS(2157),
    [anon_sym_LBRACK] = ACTIONS(2157),
    [anon_sym_RBRACK] = ACTIONS(2159),
    [anon_sym_true] = ACTIONS(2157),
    [anon_sym_false] = ACTIONS(2157),
    [anon_sym_LBRACE] = ACTIONS(2157),
    [anon_sym_SEMI] = ACTIONS(2159),
    [anon_sym_RBRACE] = ACTIONS(2159),
    [anon_sym_constraint] = ACTIONS(2157),
    [anon_sym_val] = ACTIONS(2157),
    [anon_sym_end] = ACTIONS(2157),
    [anon_sym_with] = ACTIONS(2157),
    [anon_sym_COLON_GT] = ACTIONS(2159),
    [anon_sym_object] = ACTIONS(2157),
    [anon_sym_inherit] = ACTIONS(2157),
    [anon_sym_method] = ACTIONS(2157),
    [anon_sym_initializer] = ACTIONS(2157),
    [anon_sym_AMP] = ACTIONS(2157),
    [anon_sym_POUND] = ACTIONS(2157),
    [anon_sym_COLON_COLON] = ACTIONS(2159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2159),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2159),
    [anon_sym_LT_DASH] = ACTIONS(2317),
    [anon_sym_then] = ACTIONS(2157),
    [anon_sym_else] = ACTIONS(2157),
    [anon_sym_do] = ACTIONS(2157),
    [anon_sym_done] = ACTIONS(2157),
    [anon_sym_to] = ACTIONS(2157),
    [anon_sym_downto] = ACTIONS(2157),
    [anon_sym_new] = ACTIONS(2157),
    [anon_sym_LBRACE_LT] = ACTIONS(2159),
    [anon_sym_GT_RBRACE] = ACTIONS(2159),
    [anon_sym_begin] = ACTIONS(2157),
    [sym_ocamlyacc_value] = ACTIONS(2159),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2157),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2159),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2159),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2159),
    [aux_sym_number_token1] = ACTIONS(2159),
    [aux_sym_signed_number_token1] = ACTIONS(2157),
    [anon_sym_SQUOTE] = ACTIONS(2159),
    [anon_sym_DQUOTE] = ACTIONS(2159),
    [sym_prefix_operator] = ACTIONS(2159),
    [aux_sym_sign_operator_token1] = ACTIONS(2157),
    [sym_hash_operator] = ACTIONS(2159),
    [anon_sym_lsl] = ACTIONS(2157),
    [anon_sym_lsr] = ACTIONS(2157),
    [anon_sym_asr] = ACTIONS(2157),
    [aux_sym_pow_operator_token1] = ACTIONS(2159),
    [anon_sym_mod] = ACTIONS(2157),
    [anon_sym_land] = ACTIONS(2157),
    [anon_sym_lor] = ACTIONS(2157),
    [anon_sym_lxor] = ACTIONS(2157),
    [aux_sym_mult_operator_token1] = ACTIONS(2157),
    [aux_sym_add_operator_token1] = ACTIONS(2157),
    [sym_concat_operator] = ACTIONS(2159),
    [sym_rel_operator] = ACTIONS(2157),
    [anon_sym_AMP_AMP] = ACTIONS(2157),
    [anon_sym_or] = ACTIONS(2157),
    [anon_sym_PIPE_PIPE] = ACTIONS(2157),
    [sym_assign_operator] = ACTIONS(2159),
    [sym__uppercase_identifier] = ACTIONS(2159),
    [aux_sym_tag_token1] = ACTIONS(2159),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1018)] = {
    [sym__parenthesized_abstract_type] = STATE(1766),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3758),
    [sym__binding_pattern] = STATE(3620),
    [sym_alias_binding_pattern] = STATE(3758),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_anonymous] = STATE(3614),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7206),
    [sym__tuple_binding_pattern] = STATE(3615),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern] = STATE(3758),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3758),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_exception_binding_pattern] = STATE(3758),
    [sym_effect_binding_pattern] = STATE(3758),
    [sym_attribute] = STATE(1018),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_and] = ACTIONS(1645),
    [anon_sym_EQ] = ACTIONS(1641),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_QMARK] = ACTIONS(1641),
    [anon_sym_LPAREN] = ACTIONS(2319),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_PIPE] = ACTIONS(1641),
    [anon_sym_COLON2] = ACTIONS(1645),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(2321),
    [anon_sym_COLON_GT] = ACTIONS(1641),
    [anon_sym_in] = ACTIONS(1645),
    [anon_sym_as] = ACTIONS(1645),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_COLON_COLON] = ACTIONS(1641),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [anon_sym_effect] = ACTIONS(2323),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1641),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym_let_and_operator] = ACTIONS(1641),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1019)] = {
    [sym__delimited_expression] = STATE(946),
    [sym__simple_expression] = STATE(942),
    [sym_typed_expression] = STATE(946),
    [sym_list_expression] = STATE(958),
    [sym_array_expression] = STATE(958),
    [sym_record_expression] = STATE(958),
    [sym_labeled_argument] = STATE(1023),
    [sym_prefix_expression] = STATE(946),
    [sym_hash_expression] = STATE(946),
    [sym_field_get_expression] = STATE(946),
    [sym_array_get_expression] = STATE(946),
    [sym_string_get_expression] = STATE(946),
    [sym_bigarray_get_expression] = STATE(946),
    [sym_local_open_expression] = STATE(946),
    [sym_package_expression] = STATE(958),
    [sym_new_expression] = STATE(946),
    [sym_object_copy_expression] = STATE(958),
    [sym_method_invocation] = STATE(946),
    [sym_object_expression] = STATE(946),
    [sym_parenthesized_expression] = STATE(958),
    [sym_attribute] = STATE(1019),
    [sym_extension] = STATE(946),
    [sym_quoted_extension] = STATE(946),
    [sym__constant] = STATE(946),
    [sym_number] = STATE(988),
    [sym_character] = STATE(988),
    [sym_string] = STATE(988),
    [sym_quoted_string] = STATE(988),
    [sym_boolean] = STATE(988),
    [sym_unit] = STATE(958),
    [sym__value_name] = STATE(1011),
    [sym_parenthesized_operator] = STATE(989),
    [sym_value_path] = STATE(946),
    [sym_module_path] = STATE(7385),
    [sym_constructor_path] = STATE(946),
    [sym__constructor_name] = STATE(990),
    [sym_tag] = STATE(946),
    [aux_sym_class_application_repeat1] = STATE(1015),
    [sym__lowercase_identifier] = ACTIONS(911),
    [anon_sym_and] = ACTIONS(1881),
    [anon_sym_TILDE] = ACTIONS(913),
    [anon_sym_QMARK] = ACTIONS(913),
    [anon_sym_LPAREN] = ACTIONS(489),
    [anon_sym_RPAREN] = ACTIONS(1879),
    [anon_sym_COLON2] = ACTIONS(1879),
    [anon_sym_LBRACK] = ACTIONS(491),
    [anon_sym_true] = ACTIONS(493),
    [anon_sym_false] = ACTIONS(493),
    [anon_sym_LBRACE] = ACTIONS(495),
    [anon_sym_constraint] = ACTIONS(1881),
    [anon_sym_val] = ACTIONS(1881),
    [anon_sym_end] = ACTIONS(1881),
    [anon_sym_object] = ACTIONS(497),
    [anon_sym_inherit] = ACTIONS(1881),
    [anon_sym_method] = ACTIONS(1881),
    [anon_sym_in] = ACTIONS(1881),
    [anon_sym_as] = ACTIONS(1881),
    [anon_sym_initializer] = ACTIONS(1881),
    [anon_sym_LBRACK_PIPE] = ACTIONS(501),
    [anon_sym_new] = ACTIONS(519),
    [anon_sym_LBRACE_LT] = ACTIONS(521),
    [anon_sym_begin] = ACTIONS(523),
    [sym_ocamlyacc_value] = ACTIONS(525),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1881),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1879),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(527),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(529),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1879),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1879),
    [aux_sym_number_token1] = ACTIONS(531),
    [anon_sym_SQUOTE] = ACTIONS(533),
    [anon_sym_DQUOTE] = ACTIONS(535),
    [sym_prefix_operator] = ACTIONS(537),
    [sym__uppercase_identifier] = ACTIONS(541),
    [aux_sym_tag_token1] = ACTIONS(543),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1020)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4206),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1020),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4249),
    [aux_sym_let_binding_repeat1] = STATE(1039),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2331),
    [anon_sym_let] = ACTIONS(2333),
    [anon_sym_and] = ACTIONS(2333),
    [anon_sym_EQ] = ACTIONS(2335),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2333),
    [anon_sym_type] = ACTIONS(2333),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_RBRACK] = ACTIONS(2331),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2333),
    [anon_sym_module] = ACTIONS(2333),
    [anon_sym_open] = ACTIONS(2333),
    [anon_sym_include] = ACTIONS(2333),
    [anon_sym_class] = ACTIONS(2333),
    [anon_sym_val] = ACTIONS(2333),
    [anon_sym_end] = ACTIONS(2333),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2331),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2331),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2331),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2331),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2333),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1021)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4255),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1021),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4241),
    [aux_sym_let_binding_repeat1] = STATE(1025),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2373),
    [anon_sym_let] = ACTIONS(2375),
    [anon_sym_and] = ACTIONS(2375),
    [anon_sym_EQ] = ACTIONS(2377),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2375),
    [anon_sym_type] = ACTIONS(2375),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_RBRACK] = ACTIONS(2373),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2375),
    [anon_sym_module] = ACTIONS(2375),
    [anon_sym_open] = ACTIONS(2375),
    [anon_sym_include] = ACTIONS(2375),
    [anon_sym_class] = ACTIONS(2375),
    [anon_sym_val] = ACTIONS(2375),
    [anon_sym_end] = ACTIONS(2375),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2373),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2373),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2373),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2373),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2375),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1022)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4240),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1022),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4340),
    [aux_sym_let_binding_repeat1] = STATE(1024),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2379),
    [anon_sym_let] = ACTIONS(2381),
    [anon_sym_and] = ACTIONS(2381),
    [anon_sym_EQ] = ACTIONS(2383),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2381),
    [anon_sym_type] = ACTIONS(2381),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_RBRACK] = ACTIONS(2379),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2381),
    [anon_sym_module] = ACTIONS(2381),
    [anon_sym_open] = ACTIONS(2381),
    [anon_sym_include] = ACTIONS(2381),
    [anon_sym_class] = ACTIONS(2381),
    [anon_sym_val] = ACTIONS(2381),
    [anon_sym_end] = ACTIONS(2381),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2379),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2379),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2379),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2379),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2381),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1023)] = {
    [sym_attribute] = STATE(1023),
    [sym__lowercase_identifier] = ACTIONS(1931),
    [anon_sym_and] = ACTIONS(1931),
    [anon_sym_TILDE] = ACTIONS(1931),
    [anon_sym_QMARK] = ACTIONS(1931),
    [anon_sym_LPAREN] = ACTIONS(1933),
    [anon_sym_RPAREN] = ACTIONS(1933),
    [anon_sym_COMMA] = ACTIONS(1933),
    [anon_sym_PIPE] = ACTIONS(1931),
    [anon_sym_COLON2] = ACTIONS(1931),
    [anon_sym_DASH_GT] = ACTIONS(1931),
    [anon_sym_LBRACK] = ACTIONS(1931),
    [anon_sym_RBRACK] = ACTIONS(1933),
    [anon_sym_true] = ACTIONS(1931),
    [anon_sym_false] = ACTIONS(1931),
    [anon_sym_LBRACE] = ACTIONS(1931),
    [anon_sym_SEMI] = ACTIONS(1933),
    [anon_sym_RBRACE] = ACTIONS(1933),
    [anon_sym_constraint] = ACTIONS(1931),
    [anon_sym_val] = ACTIONS(1931),
    [anon_sym_end] = ACTIONS(1931),
    [anon_sym_with] = ACTIONS(1931),
    [anon_sym_COLON_GT] = ACTIONS(1933),
    [anon_sym_object] = ACTIONS(1931),
    [anon_sym_inherit] = ACTIONS(1931),
    [anon_sym_method] = ACTIONS(1931),
    [anon_sym_in] = ACTIONS(1931),
    [anon_sym_as] = ACTIONS(1931),
    [anon_sym_initializer] = ACTIONS(1931),
    [anon_sym_AMP] = ACTIONS(1931),
    [anon_sym_COLON_COLON] = ACTIONS(1933),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1933),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1933),
    [anon_sym_then] = ACTIONS(1931),
    [anon_sym_else] = ACTIONS(1931),
    [anon_sym_do] = ACTIONS(1931),
    [anon_sym_done] = ACTIONS(1931),
    [anon_sym_to] = ACTIONS(1931),
    [anon_sym_downto] = ACTIONS(1931),
    [anon_sym_new] = ACTIONS(1931),
    [anon_sym_LBRACE_LT] = ACTIONS(1933),
    [anon_sym_GT_RBRACE] = ACTIONS(1933),
    [anon_sym_begin] = ACTIONS(1931),
    [sym_ocamlyacc_value] = ACTIONS(1933),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1931),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1933),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1933),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1933),
    [aux_sym_number_token1] = ACTIONS(1933),
    [aux_sym_signed_number_token1] = ACTIONS(1931),
    [anon_sym_SQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1933),
    [sym_prefix_operator] = ACTIONS(1933),
    [aux_sym_sign_operator_token1] = ACTIONS(1931),
    [anon_sym_lsl] = ACTIONS(1931),
    [anon_sym_lsr] = ACTIONS(1931),
    [anon_sym_asr] = ACTIONS(1931),
    [aux_sym_pow_operator_token1] = ACTIONS(1933),
    [anon_sym_mod] = ACTIONS(1931),
    [anon_sym_land] = ACTIONS(1931),
    [anon_sym_lor] = ACTIONS(1931),
    [anon_sym_lxor] = ACTIONS(1931),
    [aux_sym_mult_operator_token1] = ACTIONS(1931),
    [aux_sym_add_operator_token1] = ACTIONS(1931),
    [sym_concat_operator] = ACTIONS(1933),
    [sym_rel_operator] = ACTIONS(1931),
    [anon_sym_AMP_AMP] = ACTIONS(1931),
    [anon_sym_or] = ACTIONS(1931),
    [anon_sym_PIPE_PIPE] = ACTIONS(1931),
    [sym_assign_operator] = ACTIONS(1933),
    [sym__uppercase_identifier] = ACTIONS(1933),
    [aux_sym_tag_token1] = ACTIONS(1933),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1024)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4253),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1024),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4230),
    [aux_sym_let_binding_repeat1] = STATE(1134),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2385),
    [anon_sym_let] = ACTIONS(2387),
    [anon_sym_and] = ACTIONS(2387),
    [anon_sym_EQ] = ACTIONS(2389),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2387),
    [anon_sym_type] = ACTIONS(2387),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_RBRACK] = ACTIONS(2385),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2387),
    [anon_sym_module] = ACTIONS(2387),
    [anon_sym_open] = ACTIONS(2387),
    [anon_sym_include] = ACTIONS(2387),
    [anon_sym_class] = ACTIONS(2387),
    [anon_sym_val] = ACTIONS(2387),
    [anon_sym_end] = ACTIONS(2387),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2385),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2385),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2385),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2385),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2387),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1025)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4269),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1025),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4300),
    [aux_sym_let_binding_repeat1] = STATE(1134),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2391),
    [anon_sym_let] = ACTIONS(2393),
    [anon_sym_and] = ACTIONS(2393),
    [anon_sym_EQ] = ACTIONS(2395),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2393),
    [anon_sym_type] = ACTIONS(2393),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_RBRACK] = ACTIONS(2391),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2393),
    [anon_sym_module] = ACTIONS(2393),
    [anon_sym_open] = ACTIONS(2393),
    [anon_sym_include] = ACTIONS(2393),
    [anon_sym_class] = ACTIONS(2393),
    [anon_sym_val] = ACTIONS(2393),
    [anon_sym_end] = ACTIONS(2393),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2391),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2391),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2391),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2391),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2393),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1026)] = {
    [sym_attribute] = STATE(1026),
    [sym__lowercase_identifier] = ACTIONS(2157),
    [anon_sym_TILDE] = ACTIONS(2157),
    [anon_sym_QMARK] = ACTIONS(2157),
    [anon_sym_LPAREN] = ACTIONS(2159),
    [anon_sym_RPAREN] = ACTIONS(2159),
    [anon_sym_COMMA] = ACTIONS(2159),
    [anon_sym_PIPE] = ACTIONS(2157),
    [anon_sym_COLON2] = ACTIONS(2157),
    [anon_sym_DOT] = ACTIONS(2159),
    [anon_sym_DASH_GT] = ACTIONS(2157),
    [anon_sym_LBRACK] = ACTIONS(2157),
    [anon_sym_RBRACK] = ACTIONS(2159),
    [anon_sym_true] = ACTIONS(2157),
    [anon_sym_false] = ACTIONS(2157),
    [anon_sym_LBRACE] = ACTIONS(2157),
    [anon_sym_SEMI] = ACTIONS(2159),
    [anon_sym_RBRACE] = ACTIONS(2159),
    [anon_sym_constraint] = ACTIONS(2157),
    [anon_sym_val] = ACTIONS(2157),
    [anon_sym_end] = ACTIONS(2157),
    [anon_sym_with] = ACTIONS(2157),
    [anon_sym_COLON_GT] = ACTIONS(2159),
    [anon_sym_object] = ACTIONS(2157),
    [anon_sym_inherit] = ACTIONS(2157),
    [anon_sym_method] = ACTIONS(2157),
    [anon_sym_initializer] = ACTIONS(2157),
    [anon_sym_AMP] = ACTIONS(2157),
    [anon_sym_POUND] = ACTIONS(2157),
    [anon_sym_COLON_COLON] = ACTIONS(2159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2159),
    [anon_sym_LT_DASH] = ACTIONS(2317),
    [anon_sym_then] = ACTIONS(2157),
    [anon_sym_else] = ACTIONS(2157),
    [anon_sym_do] = ACTIONS(2157),
    [anon_sym_done] = ACTIONS(2157),
    [anon_sym_to] = ACTIONS(2157),
    [anon_sym_downto] = ACTIONS(2157),
    [anon_sym_new] = ACTIONS(2157),
    [anon_sym_LBRACE_LT] = ACTIONS(2159),
    [anon_sym_GT_RBRACE] = ACTIONS(2159),
    [anon_sym_begin] = ACTIONS(2157),
    [sym_ocamlyacc_value] = ACTIONS(2159),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2157),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2159),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2159),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2159),
    [aux_sym_number_token1] = ACTIONS(2159),
    [aux_sym_signed_number_token1] = ACTIONS(2157),
    [anon_sym_SQUOTE] = ACTIONS(2159),
    [anon_sym_DQUOTE] = ACTIONS(2159),
    [sym_prefix_operator] = ACTIONS(2159),
    [aux_sym_sign_operator_token1] = ACTIONS(2157),
    [sym_hash_operator] = ACTIONS(2159),
    [anon_sym_lsl] = ACTIONS(2157),
    [anon_sym_lsr] = ACTIONS(2157),
    [anon_sym_asr] = ACTIONS(2157),
    [aux_sym_pow_operator_token1] = ACTIONS(2159),
    [anon_sym_mod] = ACTIONS(2157),
    [anon_sym_land] = ACTIONS(2157),
    [anon_sym_lor] = ACTIONS(2157),
    [anon_sym_lxor] = ACTIONS(2157),
    [aux_sym_mult_operator_token1] = ACTIONS(2157),
    [aux_sym_add_operator_token1] = ACTIONS(2157),
    [sym_concat_operator] = ACTIONS(2159),
    [sym_rel_operator] = ACTIONS(2157),
    [anon_sym_AMP_AMP] = ACTIONS(2157),
    [anon_sym_or] = ACTIONS(2157),
    [anon_sym_PIPE_PIPE] = ACTIONS(2157),
    [sym_assign_operator] = ACTIONS(2159),
    [sym__uppercase_identifier] = ACTIONS(2159),
    [aux_sym_tag_token1] = ACTIONS(2159),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1027)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4271),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1027),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4302),
    [aux_sym_let_binding_repeat1] = STATE(1028),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2397),
    [anon_sym_let] = ACTIONS(2399),
    [anon_sym_and] = ACTIONS(2399),
    [anon_sym_EQ] = ACTIONS(2401),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2399),
    [anon_sym_type] = ACTIONS(2399),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_RBRACK] = ACTIONS(2397),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2399),
    [anon_sym_module] = ACTIONS(2399),
    [anon_sym_open] = ACTIONS(2399),
    [anon_sym_include] = ACTIONS(2399),
    [anon_sym_class] = ACTIONS(2399),
    [anon_sym_val] = ACTIONS(2399),
    [anon_sym_end] = ACTIONS(2399),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2397),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2397),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2397),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2397),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2399),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1028)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4279),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1028),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4205),
    [aux_sym_let_binding_repeat1] = STATE(1134),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2403),
    [anon_sym_let] = ACTIONS(2405),
    [anon_sym_and] = ACTIONS(2405),
    [anon_sym_EQ] = ACTIONS(2407),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2405),
    [anon_sym_type] = ACTIONS(2405),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_RBRACK] = ACTIONS(2403),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2405),
    [anon_sym_module] = ACTIONS(2405),
    [anon_sym_open] = ACTIONS(2405),
    [anon_sym_include] = ACTIONS(2405),
    [anon_sym_class] = ACTIONS(2405),
    [anon_sym_val] = ACTIONS(2405),
    [anon_sym_end] = ACTIONS(2405),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2403),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2403),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2403),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2403),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2405),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1029)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4210),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1029),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4332),
    [aux_sym_let_binding_repeat1] = STATE(1134),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2409),
    [anon_sym_let] = ACTIONS(2411),
    [anon_sym_and] = ACTIONS(2411),
    [anon_sym_EQ] = ACTIONS(2413),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2411),
    [anon_sym_type] = ACTIONS(2411),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_RBRACK] = ACTIONS(2409),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2411),
    [anon_sym_module] = ACTIONS(2411),
    [anon_sym_open] = ACTIONS(2411),
    [anon_sym_include] = ACTIONS(2411),
    [anon_sym_class] = ACTIONS(2411),
    [anon_sym_val] = ACTIONS(2411),
    [anon_sym_end] = ACTIONS(2411),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2409),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2409),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2409),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2409),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2411),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1030)] = {
    [sym_attribute] = STATE(1030),
    [sym__lowercase_identifier] = ACTIONS(2415),
    [anon_sym_and] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(2415),
    [anon_sym_QMARK] = ACTIONS(2415),
    [anon_sym_LPAREN] = ACTIONS(2417),
    [anon_sym_RPAREN] = ACTIONS(2417),
    [anon_sym_COMMA] = ACTIONS(2417),
    [anon_sym_PIPE] = ACTIONS(2415),
    [anon_sym_COLON2] = ACTIONS(2415),
    [anon_sym_DASH_GT] = ACTIONS(2415),
    [anon_sym_LBRACK] = ACTIONS(2415),
    [anon_sym_RBRACK] = ACTIONS(2417),
    [anon_sym_true] = ACTIONS(2415),
    [anon_sym_false] = ACTIONS(2415),
    [anon_sym_LBRACE] = ACTIONS(2415),
    [anon_sym_SEMI] = ACTIONS(2417),
    [anon_sym_RBRACE] = ACTIONS(2417),
    [anon_sym_constraint] = ACTIONS(2415),
    [anon_sym_val] = ACTIONS(2415),
    [anon_sym_end] = ACTIONS(2415),
    [anon_sym_with] = ACTIONS(2415),
    [anon_sym_COLON_GT] = ACTIONS(2417),
    [anon_sym_object] = ACTIONS(2415),
    [anon_sym_inherit] = ACTIONS(2415),
    [anon_sym_method] = ACTIONS(2415),
    [anon_sym_in] = ACTIONS(2415),
    [anon_sym_as] = ACTIONS(2415),
    [anon_sym_initializer] = ACTIONS(2415),
    [anon_sym_AMP] = ACTIONS(2415),
    [anon_sym_COLON_COLON] = ACTIONS(2417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2417),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2417),
    [anon_sym_then] = ACTIONS(2415),
    [anon_sym_else] = ACTIONS(2415),
    [anon_sym_do] = ACTIONS(2415),
    [anon_sym_done] = ACTIONS(2415),
    [anon_sym_to] = ACTIONS(2415),
    [anon_sym_downto] = ACTIONS(2415),
    [anon_sym_new] = ACTIONS(2415),
    [anon_sym_LBRACE_LT] = ACTIONS(2417),
    [anon_sym_GT_RBRACE] = ACTIONS(2417),
    [anon_sym_begin] = ACTIONS(2415),
    [sym_ocamlyacc_value] = ACTIONS(2417),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2415),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2417),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2415),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2417),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2417),
    [aux_sym_number_token1] = ACTIONS(2417),
    [aux_sym_signed_number_token1] = ACTIONS(2415),
    [anon_sym_SQUOTE] = ACTIONS(2417),
    [anon_sym_DQUOTE] = ACTIONS(2417),
    [sym_prefix_operator] = ACTIONS(2417),
    [aux_sym_sign_operator_token1] = ACTIONS(2415),
    [anon_sym_lsl] = ACTIONS(2415),
    [anon_sym_lsr] = ACTIONS(2415),
    [anon_sym_asr] = ACTIONS(2415),
    [aux_sym_pow_operator_token1] = ACTIONS(2417),
    [anon_sym_mod] = ACTIONS(2415),
    [anon_sym_land] = ACTIONS(2415),
    [anon_sym_lor] = ACTIONS(2415),
    [anon_sym_lxor] = ACTIONS(2415),
    [aux_sym_mult_operator_token1] = ACTIONS(2415),
    [aux_sym_add_operator_token1] = ACTIONS(2415),
    [sym_concat_operator] = ACTIONS(2417),
    [sym_rel_operator] = ACTIONS(2415),
    [anon_sym_AMP_AMP] = ACTIONS(2415),
    [anon_sym_or] = ACTIONS(2415),
    [anon_sym_PIPE_PIPE] = ACTIONS(2415),
    [sym_assign_operator] = ACTIONS(2417),
    [sym__uppercase_identifier] = ACTIONS(2417),
    [aux_sym_tag_token1] = ACTIONS(2417),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1031)] = {
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3758),
    [sym__binding_pattern] = STATE(3621),
    [sym_alias_binding_pattern] = STATE(3758),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_anonymous] = STATE(3614),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7206),
    [sym__tuple_binding_pattern] = STATE(3615),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern] = STATE(3758),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3758),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_exception_binding_pattern] = STATE(3758),
    [sym_effect_binding_pattern] = STATE(3758),
    [sym_attribute] = STATE(1031),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_and] = ACTIONS(1645),
    [anon_sym_EQ] = ACTIONS(1641),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_QMARK] = ACTIONS(1641),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_PIPE] = ACTIONS(1641),
    [anon_sym_COLON2] = ACTIONS(1645),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(2321),
    [anon_sym_COLON_GT] = ACTIONS(1641),
    [anon_sym_in] = ACTIONS(1645),
    [anon_sym_as] = ACTIONS(1645),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_COLON_COLON] = ACTIONS(1641),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [anon_sym_effect] = ACTIONS(2323),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1641),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym_let_and_operator] = ACTIONS(1641),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1032)] = {
    [sym_attribute] = STATE(1032),
    [sym__lowercase_identifier] = ACTIONS(1895),
    [anon_sym_TILDE] = ACTIONS(1895),
    [anon_sym_QMARK] = ACTIONS(1895),
    [anon_sym_LPAREN] = ACTIONS(1897),
    [anon_sym_RPAREN] = ACTIONS(1897),
    [anon_sym_COMMA] = ACTIONS(1897),
    [anon_sym_PIPE] = ACTIONS(1895),
    [anon_sym_COLON2] = ACTIONS(1895),
    [anon_sym_DOT] = ACTIONS(1897),
    [anon_sym_DASH_GT] = ACTIONS(1895),
    [anon_sym_LBRACK] = ACTIONS(1895),
    [anon_sym_RBRACK] = ACTIONS(1897),
    [anon_sym_true] = ACTIONS(1895),
    [anon_sym_false] = ACTIONS(1895),
    [anon_sym_LBRACE] = ACTIONS(1895),
    [anon_sym_SEMI] = ACTIONS(1897),
    [anon_sym_RBRACE] = ACTIONS(1897),
    [anon_sym_constraint] = ACTIONS(1895),
    [anon_sym_val] = ACTIONS(1895),
    [anon_sym_end] = ACTIONS(1895),
    [anon_sym_with] = ACTIONS(1895),
    [anon_sym_COLON_GT] = ACTIONS(1897),
    [anon_sym_object] = ACTIONS(1895),
    [anon_sym_inherit] = ACTIONS(1895),
    [anon_sym_method] = ACTIONS(1895),
    [anon_sym_initializer] = ACTIONS(1895),
    [anon_sym_AMP] = ACTIONS(1895),
    [anon_sym_POUND] = ACTIONS(1895),
    [anon_sym_COLON_COLON] = ACTIONS(1897),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1897),
    [anon_sym_LT_DASH] = ACTIONS(1895),
    [anon_sym_then] = ACTIONS(1895),
    [anon_sym_else] = ACTIONS(1895),
    [anon_sym_do] = ACTIONS(1895),
    [anon_sym_done] = ACTIONS(1895),
    [anon_sym_to] = ACTIONS(1895),
    [anon_sym_downto] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1895),
    [anon_sym_LBRACE_LT] = ACTIONS(1897),
    [anon_sym_GT_RBRACE] = ACTIONS(1897),
    [anon_sym_begin] = ACTIONS(1895),
    [sym_ocamlyacc_value] = ACTIONS(1897),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1895),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1897),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1895),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1895),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1897),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1897),
    [aux_sym_number_token1] = ACTIONS(1897),
    [aux_sym_signed_number_token1] = ACTIONS(1895),
    [anon_sym_SQUOTE] = ACTIONS(1897),
    [anon_sym_DQUOTE] = ACTIONS(1897),
    [sym_prefix_operator] = ACTIONS(1897),
    [aux_sym_sign_operator_token1] = ACTIONS(1895),
    [sym_hash_operator] = ACTIONS(1897),
    [anon_sym_lsl] = ACTIONS(1895),
    [anon_sym_lsr] = ACTIONS(1895),
    [anon_sym_asr] = ACTIONS(1895),
    [aux_sym_pow_operator_token1] = ACTIONS(1897),
    [anon_sym_mod] = ACTIONS(1895),
    [anon_sym_land] = ACTIONS(1895),
    [anon_sym_lor] = ACTIONS(1895),
    [anon_sym_lxor] = ACTIONS(1895),
    [aux_sym_mult_operator_token1] = ACTIONS(1895),
    [aux_sym_add_operator_token1] = ACTIONS(1895),
    [sym_concat_operator] = ACTIONS(1897),
    [sym_rel_operator] = ACTIONS(1895),
    [anon_sym_AMP_AMP] = ACTIONS(1895),
    [anon_sym_or] = ACTIONS(1895),
    [anon_sym_PIPE_PIPE] = ACTIONS(1895),
    [sym_assign_operator] = ACTIONS(1897),
    [sym__uppercase_identifier] = ACTIONS(1897),
    [aux_sym_tag_token1] = ACTIONS(1897),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1033)] = {
    [sym_attribute] = STATE(1033),
    [sym__lowercase_identifier] = ACTIONS(1903),
    [anon_sym_TILDE] = ACTIONS(1903),
    [anon_sym_QMARK] = ACTIONS(1903),
    [anon_sym_LPAREN] = ACTIONS(1905),
    [anon_sym_RPAREN] = ACTIONS(1905),
    [anon_sym_COMMA] = ACTIONS(1905),
    [anon_sym_PIPE] = ACTIONS(1903),
    [anon_sym_COLON2] = ACTIONS(1903),
    [anon_sym_DOT] = ACTIONS(1905),
    [anon_sym_DASH_GT] = ACTIONS(1903),
    [anon_sym_LBRACK] = ACTIONS(1903),
    [anon_sym_RBRACK] = ACTIONS(1905),
    [anon_sym_true] = ACTIONS(1903),
    [anon_sym_false] = ACTIONS(1903),
    [anon_sym_LBRACE] = ACTIONS(1903),
    [anon_sym_SEMI] = ACTIONS(1905),
    [anon_sym_RBRACE] = ACTIONS(1905),
    [anon_sym_constraint] = ACTIONS(1903),
    [anon_sym_val] = ACTIONS(1903),
    [anon_sym_end] = ACTIONS(1903),
    [anon_sym_with] = ACTIONS(1903),
    [anon_sym_COLON_GT] = ACTIONS(1905),
    [anon_sym_object] = ACTIONS(1903),
    [anon_sym_inherit] = ACTIONS(1903),
    [anon_sym_method] = ACTIONS(1903),
    [anon_sym_initializer] = ACTIONS(1903),
    [anon_sym_AMP] = ACTIONS(1903),
    [anon_sym_POUND] = ACTIONS(1903),
    [anon_sym_COLON_COLON] = ACTIONS(1905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1905),
    [anon_sym_LT_DASH] = ACTIONS(1903),
    [anon_sym_then] = ACTIONS(1903),
    [anon_sym_else] = ACTIONS(1903),
    [anon_sym_do] = ACTIONS(1903),
    [anon_sym_done] = ACTIONS(1903),
    [anon_sym_to] = ACTIONS(1903),
    [anon_sym_downto] = ACTIONS(1903),
    [anon_sym_new] = ACTIONS(1903),
    [anon_sym_LBRACE_LT] = ACTIONS(1905),
    [anon_sym_GT_RBRACE] = ACTIONS(1905),
    [anon_sym_begin] = ACTIONS(1903),
    [sym_ocamlyacc_value] = ACTIONS(1905),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1903),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1905),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1903),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1903),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1905),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1905),
    [aux_sym_number_token1] = ACTIONS(1905),
    [aux_sym_signed_number_token1] = ACTIONS(1903),
    [anon_sym_SQUOTE] = ACTIONS(1905),
    [anon_sym_DQUOTE] = ACTIONS(1905),
    [sym_prefix_operator] = ACTIONS(1905),
    [aux_sym_sign_operator_token1] = ACTIONS(1903),
    [sym_hash_operator] = ACTIONS(1905),
    [anon_sym_lsl] = ACTIONS(1903),
    [anon_sym_lsr] = ACTIONS(1903),
    [anon_sym_asr] = ACTIONS(1903),
    [aux_sym_pow_operator_token1] = ACTIONS(1905),
    [anon_sym_mod] = ACTIONS(1903),
    [anon_sym_land] = ACTIONS(1903),
    [anon_sym_lor] = ACTIONS(1903),
    [anon_sym_lxor] = ACTIONS(1903),
    [aux_sym_mult_operator_token1] = ACTIONS(1903),
    [aux_sym_add_operator_token1] = ACTIONS(1903),
    [sym_concat_operator] = ACTIONS(1905),
    [sym_rel_operator] = ACTIONS(1903),
    [anon_sym_AMP_AMP] = ACTIONS(1903),
    [anon_sym_or] = ACTIONS(1903),
    [anon_sym_PIPE_PIPE] = ACTIONS(1903),
    [sym_assign_operator] = ACTIONS(1905),
    [sym__uppercase_identifier] = ACTIONS(1905),
    [aux_sym_tag_token1] = ACTIONS(1905),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1034)] = {
    [sym_attribute] = STATE(1034),
    [sym__lowercase_identifier] = ACTIONS(1899),
    [anon_sym_TILDE] = ACTIONS(1899),
    [anon_sym_QMARK] = ACTIONS(1899),
    [anon_sym_LPAREN] = ACTIONS(1901),
    [anon_sym_RPAREN] = ACTIONS(1901),
    [anon_sym_COMMA] = ACTIONS(1901),
    [anon_sym_PIPE] = ACTIONS(1899),
    [anon_sym_COLON2] = ACTIONS(1899),
    [anon_sym_DOT] = ACTIONS(1901),
    [anon_sym_DASH_GT] = ACTIONS(1899),
    [anon_sym_LBRACK] = ACTIONS(1899),
    [anon_sym_RBRACK] = ACTIONS(1901),
    [anon_sym_true] = ACTIONS(1899),
    [anon_sym_false] = ACTIONS(1899),
    [anon_sym_LBRACE] = ACTIONS(1899),
    [anon_sym_SEMI] = ACTIONS(1901),
    [anon_sym_RBRACE] = ACTIONS(1901),
    [anon_sym_constraint] = ACTIONS(1899),
    [anon_sym_val] = ACTIONS(1899),
    [anon_sym_end] = ACTIONS(1899),
    [anon_sym_with] = ACTIONS(1899),
    [anon_sym_COLON_GT] = ACTIONS(1901),
    [anon_sym_object] = ACTIONS(1899),
    [anon_sym_inherit] = ACTIONS(1899),
    [anon_sym_method] = ACTIONS(1899),
    [anon_sym_initializer] = ACTIONS(1899),
    [anon_sym_AMP] = ACTIONS(1899),
    [anon_sym_POUND] = ACTIONS(1899),
    [anon_sym_COLON_COLON] = ACTIONS(1901),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1901),
    [anon_sym_LT_DASH] = ACTIONS(1899),
    [anon_sym_then] = ACTIONS(1899),
    [anon_sym_else] = ACTIONS(1899),
    [anon_sym_do] = ACTIONS(1899),
    [anon_sym_done] = ACTIONS(1899),
    [anon_sym_to] = ACTIONS(1899),
    [anon_sym_downto] = ACTIONS(1899),
    [anon_sym_new] = ACTIONS(1899),
    [anon_sym_LBRACE_LT] = ACTIONS(1901),
    [anon_sym_GT_RBRACE] = ACTIONS(1901),
    [anon_sym_begin] = ACTIONS(1899),
    [sym_ocamlyacc_value] = ACTIONS(1901),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1899),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1901),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1899),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1899),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1901),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1901),
    [aux_sym_number_token1] = ACTIONS(1901),
    [aux_sym_signed_number_token1] = ACTIONS(1899),
    [anon_sym_SQUOTE] = ACTIONS(1901),
    [anon_sym_DQUOTE] = ACTIONS(1901),
    [sym_prefix_operator] = ACTIONS(1901),
    [aux_sym_sign_operator_token1] = ACTIONS(1899),
    [sym_hash_operator] = ACTIONS(1901),
    [anon_sym_lsl] = ACTIONS(1899),
    [anon_sym_lsr] = ACTIONS(1899),
    [anon_sym_asr] = ACTIONS(1899),
    [aux_sym_pow_operator_token1] = ACTIONS(1901),
    [anon_sym_mod] = ACTIONS(1899),
    [anon_sym_land] = ACTIONS(1899),
    [anon_sym_lor] = ACTIONS(1899),
    [anon_sym_lxor] = ACTIONS(1899),
    [aux_sym_mult_operator_token1] = ACTIONS(1899),
    [aux_sym_add_operator_token1] = ACTIONS(1899),
    [sym_concat_operator] = ACTIONS(1901),
    [sym_rel_operator] = ACTIONS(1899),
    [anon_sym_AMP_AMP] = ACTIONS(1899),
    [anon_sym_or] = ACTIONS(1899),
    [anon_sym_PIPE_PIPE] = ACTIONS(1899),
    [sym_assign_operator] = ACTIONS(1901),
    [sym__uppercase_identifier] = ACTIONS(1901),
    [aux_sym_tag_token1] = ACTIONS(1901),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1035)] = {
    [sym_attribute] = STATE(1035),
    [sym__lowercase_identifier] = ACTIONS(1907),
    [anon_sym_TILDE] = ACTIONS(1907),
    [anon_sym_QMARK] = ACTIONS(1907),
    [anon_sym_LPAREN] = ACTIONS(1909),
    [anon_sym_RPAREN] = ACTIONS(1909),
    [anon_sym_COMMA] = ACTIONS(1909),
    [anon_sym_PIPE] = ACTIONS(1907),
    [anon_sym_COLON2] = ACTIONS(1907),
    [anon_sym_DOT] = ACTIONS(1909),
    [anon_sym_DASH_GT] = ACTIONS(1907),
    [anon_sym_LBRACK] = ACTIONS(1907),
    [anon_sym_RBRACK] = ACTIONS(1909),
    [anon_sym_true] = ACTIONS(1907),
    [anon_sym_false] = ACTIONS(1907),
    [anon_sym_LBRACE] = ACTIONS(1907),
    [anon_sym_SEMI] = ACTIONS(1909),
    [anon_sym_RBRACE] = ACTIONS(1909),
    [anon_sym_constraint] = ACTIONS(1907),
    [anon_sym_val] = ACTIONS(1907),
    [anon_sym_end] = ACTIONS(1907),
    [anon_sym_with] = ACTIONS(1907),
    [anon_sym_COLON_GT] = ACTIONS(1909),
    [anon_sym_object] = ACTIONS(1907),
    [anon_sym_inherit] = ACTIONS(1907),
    [anon_sym_method] = ACTIONS(1907),
    [anon_sym_initializer] = ACTIONS(1907),
    [anon_sym_AMP] = ACTIONS(1907),
    [anon_sym_POUND] = ACTIONS(1907),
    [anon_sym_COLON_COLON] = ACTIONS(1909),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1909),
    [anon_sym_LT_DASH] = ACTIONS(1907),
    [anon_sym_then] = ACTIONS(1907),
    [anon_sym_else] = ACTIONS(1907),
    [anon_sym_do] = ACTIONS(1907),
    [anon_sym_done] = ACTIONS(1907),
    [anon_sym_to] = ACTIONS(1907),
    [anon_sym_downto] = ACTIONS(1907),
    [anon_sym_new] = ACTIONS(1907),
    [anon_sym_LBRACE_LT] = ACTIONS(1909),
    [anon_sym_GT_RBRACE] = ACTIONS(1909),
    [anon_sym_begin] = ACTIONS(1907),
    [sym_ocamlyacc_value] = ACTIONS(1909),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1907),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1909),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1907),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1907),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1909),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1909),
    [aux_sym_number_token1] = ACTIONS(1909),
    [aux_sym_signed_number_token1] = ACTIONS(1907),
    [anon_sym_SQUOTE] = ACTIONS(1909),
    [anon_sym_DQUOTE] = ACTIONS(1909),
    [sym_prefix_operator] = ACTIONS(1909),
    [aux_sym_sign_operator_token1] = ACTIONS(1907),
    [sym_hash_operator] = ACTIONS(1909),
    [anon_sym_lsl] = ACTIONS(1907),
    [anon_sym_lsr] = ACTIONS(1907),
    [anon_sym_asr] = ACTIONS(1907),
    [aux_sym_pow_operator_token1] = ACTIONS(1909),
    [anon_sym_mod] = ACTIONS(1907),
    [anon_sym_land] = ACTIONS(1907),
    [anon_sym_lor] = ACTIONS(1907),
    [anon_sym_lxor] = ACTIONS(1907),
    [aux_sym_mult_operator_token1] = ACTIONS(1907),
    [aux_sym_add_operator_token1] = ACTIONS(1907),
    [sym_concat_operator] = ACTIONS(1909),
    [sym_rel_operator] = ACTIONS(1907),
    [anon_sym_AMP_AMP] = ACTIONS(1907),
    [anon_sym_or] = ACTIONS(1907),
    [anon_sym_PIPE_PIPE] = ACTIONS(1907),
    [sym_assign_operator] = ACTIONS(1909),
    [sym__uppercase_identifier] = ACTIONS(1909),
    [aux_sym_tag_token1] = ACTIONS(1909),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1036)] = {
    [sym_attribute] = STATE(1036),
    [sym__lowercase_identifier] = ACTIONS(1911),
    [anon_sym_TILDE] = ACTIONS(1911),
    [anon_sym_QMARK] = ACTIONS(1911),
    [anon_sym_LPAREN] = ACTIONS(1913),
    [anon_sym_RPAREN] = ACTIONS(1913),
    [anon_sym_COMMA] = ACTIONS(1913),
    [anon_sym_PIPE] = ACTIONS(1911),
    [anon_sym_COLON2] = ACTIONS(1911),
    [anon_sym_DOT] = ACTIONS(1913),
    [anon_sym_DASH_GT] = ACTIONS(1911),
    [anon_sym_LBRACK] = ACTIONS(1911),
    [anon_sym_RBRACK] = ACTIONS(1913),
    [anon_sym_true] = ACTIONS(1911),
    [anon_sym_false] = ACTIONS(1911),
    [anon_sym_LBRACE] = ACTIONS(1911),
    [anon_sym_SEMI] = ACTIONS(1913),
    [anon_sym_RBRACE] = ACTIONS(1913),
    [anon_sym_constraint] = ACTIONS(1911),
    [anon_sym_val] = ACTIONS(1911),
    [anon_sym_end] = ACTIONS(1911),
    [anon_sym_with] = ACTIONS(1911),
    [anon_sym_COLON_GT] = ACTIONS(1913),
    [anon_sym_object] = ACTIONS(1911),
    [anon_sym_inherit] = ACTIONS(1911),
    [anon_sym_method] = ACTIONS(1911),
    [anon_sym_initializer] = ACTIONS(1911),
    [anon_sym_AMP] = ACTIONS(1911),
    [anon_sym_POUND] = ACTIONS(1911),
    [anon_sym_COLON_COLON] = ACTIONS(1913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1913),
    [anon_sym_LT_DASH] = ACTIONS(1911),
    [anon_sym_then] = ACTIONS(1911),
    [anon_sym_else] = ACTIONS(1911),
    [anon_sym_do] = ACTIONS(1911),
    [anon_sym_done] = ACTIONS(1911),
    [anon_sym_to] = ACTIONS(1911),
    [anon_sym_downto] = ACTIONS(1911),
    [anon_sym_new] = ACTIONS(1911),
    [anon_sym_LBRACE_LT] = ACTIONS(1913),
    [anon_sym_GT_RBRACE] = ACTIONS(1913),
    [anon_sym_begin] = ACTIONS(1911),
    [sym_ocamlyacc_value] = ACTIONS(1913),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1911),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1913),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1911),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1911),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1913),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1913),
    [aux_sym_number_token1] = ACTIONS(1913),
    [aux_sym_signed_number_token1] = ACTIONS(1911),
    [anon_sym_SQUOTE] = ACTIONS(1913),
    [anon_sym_DQUOTE] = ACTIONS(1913),
    [sym_prefix_operator] = ACTIONS(1913),
    [aux_sym_sign_operator_token1] = ACTIONS(1911),
    [sym_hash_operator] = ACTIONS(1913),
    [anon_sym_lsl] = ACTIONS(1911),
    [anon_sym_lsr] = ACTIONS(1911),
    [anon_sym_asr] = ACTIONS(1911),
    [aux_sym_pow_operator_token1] = ACTIONS(1913),
    [anon_sym_mod] = ACTIONS(1911),
    [anon_sym_land] = ACTIONS(1911),
    [anon_sym_lor] = ACTIONS(1911),
    [anon_sym_lxor] = ACTIONS(1911),
    [aux_sym_mult_operator_token1] = ACTIONS(1911),
    [aux_sym_add_operator_token1] = ACTIONS(1911),
    [sym_concat_operator] = ACTIONS(1913),
    [sym_rel_operator] = ACTIONS(1911),
    [anon_sym_AMP_AMP] = ACTIONS(1911),
    [anon_sym_or] = ACTIONS(1911),
    [anon_sym_PIPE_PIPE] = ACTIONS(1911),
    [sym_assign_operator] = ACTIONS(1913),
    [sym__uppercase_identifier] = ACTIONS(1913),
    [aux_sym_tag_token1] = ACTIONS(1913),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1037)] = {
    [sym_attribute] = STATE(1037),
    [sym__lowercase_identifier] = ACTIONS(1915),
    [anon_sym_TILDE] = ACTIONS(1915),
    [anon_sym_QMARK] = ACTIONS(1915),
    [anon_sym_LPAREN] = ACTIONS(1917),
    [anon_sym_RPAREN] = ACTIONS(1917),
    [anon_sym_COMMA] = ACTIONS(1917),
    [anon_sym_PIPE] = ACTIONS(1915),
    [anon_sym_COLON2] = ACTIONS(1915),
    [anon_sym_DOT] = ACTIONS(1917),
    [anon_sym_DASH_GT] = ACTIONS(1915),
    [anon_sym_LBRACK] = ACTIONS(1915),
    [anon_sym_RBRACK] = ACTIONS(1917),
    [anon_sym_true] = ACTIONS(1915),
    [anon_sym_false] = ACTIONS(1915),
    [anon_sym_LBRACE] = ACTIONS(1915),
    [anon_sym_SEMI] = ACTIONS(1917),
    [anon_sym_RBRACE] = ACTIONS(1917),
    [anon_sym_constraint] = ACTIONS(1915),
    [anon_sym_val] = ACTIONS(1915),
    [anon_sym_end] = ACTIONS(1915),
    [anon_sym_with] = ACTIONS(1915),
    [anon_sym_COLON_GT] = ACTIONS(1917),
    [anon_sym_object] = ACTIONS(1915),
    [anon_sym_inherit] = ACTIONS(1915),
    [anon_sym_method] = ACTIONS(1915),
    [anon_sym_initializer] = ACTIONS(1915),
    [anon_sym_AMP] = ACTIONS(1915),
    [anon_sym_POUND] = ACTIONS(1915),
    [anon_sym_COLON_COLON] = ACTIONS(1917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1917),
    [anon_sym_LT_DASH] = ACTIONS(1915),
    [anon_sym_then] = ACTIONS(1915),
    [anon_sym_else] = ACTIONS(1915),
    [anon_sym_do] = ACTIONS(1915),
    [anon_sym_done] = ACTIONS(1915),
    [anon_sym_to] = ACTIONS(1915),
    [anon_sym_downto] = ACTIONS(1915),
    [anon_sym_new] = ACTIONS(1915),
    [anon_sym_LBRACE_LT] = ACTIONS(1917),
    [anon_sym_GT_RBRACE] = ACTIONS(1917),
    [anon_sym_begin] = ACTIONS(1915),
    [sym_ocamlyacc_value] = ACTIONS(1917),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1915),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1917),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1915),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1915),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1917),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1917),
    [aux_sym_number_token1] = ACTIONS(1917),
    [aux_sym_signed_number_token1] = ACTIONS(1915),
    [anon_sym_SQUOTE] = ACTIONS(1917),
    [anon_sym_DQUOTE] = ACTIONS(1917),
    [sym_prefix_operator] = ACTIONS(1917),
    [aux_sym_sign_operator_token1] = ACTIONS(1915),
    [sym_hash_operator] = ACTIONS(1917),
    [anon_sym_lsl] = ACTIONS(1915),
    [anon_sym_lsr] = ACTIONS(1915),
    [anon_sym_asr] = ACTIONS(1915),
    [aux_sym_pow_operator_token1] = ACTIONS(1917),
    [anon_sym_mod] = ACTIONS(1915),
    [anon_sym_land] = ACTIONS(1915),
    [anon_sym_lor] = ACTIONS(1915),
    [anon_sym_lxor] = ACTIONS(1915),
    [aux_sym_mult_operator_token1] = ACTIONS(1915),
    [aux_sym_add_operator_token1] = ACTIONS(1915),
    [sym_concat_operator] = ACTIONS(1917),
    [sym_rel_operator] = ACTIONS(1915),
    [anon_sym_AMP_AMP] = ACTIONS(1915),
    [anon_sym_or] = ACTIONS(1915),
    [anon_sym_PIPE_PIPE] = ACTIONS(1915),
    [sym_assign_operator] = ACTIONS(1917),
    [sym__uppercase_identifier] = ACTIONS(1917),
    [aux_sym_tag_token1] = ACTIONS(1917),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1038)] = {
    [sym_attribute] = STATE(1038),
    [sym__lowercase_identifier] = ACTIONS(1919),
    [anon_sym_TILDE] = ACTIONS(1919),
    [anon_sym_QMARK] = ACTIONS(1919),
    [anon_sym_LPAREN] = ACTIONS(1921),
    [anon_sym_RPAREN] = ACTIONS(1921),
    [anon_sym_COMMA] = ACTIONS(1921),
    [anon_sym_PIPE] = ACTIONS(1919),
    [anon_sym_COLON2] = ACTIONS(1919),
    [anon_sym_DOT] = ACTIONS(1921),
    [anon_sym_DASH_GT] = ACTIONS(1919),
    [anon_sym_LBRACK] = ACTIONS(1919),
    [anon_sym_RBRACK] = ACTIONS(1921),
    [anon_sym_true] = ACTIONS(1919),
    [anon_sym_false] = ACTIONS(1919),
    [anon_sym_LBRACE] = ACTIONS(1919),
    [anon_sym_SEMI] = ACTIONS(1921),
    [anon_sym_RBRACE] = ACTIONS(1921),
    [anon_sym_constraint] = ACTIONS(1919),
    [anon_sym_val] = ACTIONS(1919),
    [anon_sym_end] = ACTIONS(1919),
    [anon_sym_with] = ACTIONS(1919),
    [anon_sym_COLON_GT] = ACTIONS(1921),
    [anon_sym_object] = ACTIONS(1919),
    [anon_sym_inherit] = ACTIONS(1919),
    [anon_sym_method] = ACTIONS(1919),
    [anon_sym_initializer] = ACTIONS(1919),
    [anon_sym_AMP] = ACTIONS(1919),
    [anon_sym_POUND] = ACTIONS(1919),
    [anon_sym_COLON_COLON] = ACTIONS(1921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1921),
    [anon_sym_LT_DASH] = ACTIONS(1919),
    [anon_sym_then] = ACTIONS(1919),
    [anon_sym_else] = ACTIONS(1919),
    [anon_sym_do] = ACTIONS(1919),
    [anon_sym_done] = ACTIONS(1919),
    [anon_sym_to] = ACTIONS(1919),
    [anon_sym_downto] = ACTIONS(1919),
    [anon_sym_new] = ACTIONS(1919),
    [anon_sym_LBRACE_LT] = ACTIONS(1921),
    [anon_sym_GT_RBRACE] = ACTIONS(1921),
    [anon_sym_begin] = ACTIONS(1919),
    [sym_ocamlyacc_value] = ACTIONS(1921),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1919),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1921),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1919),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1919),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1921),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1921),
    [aux_sym_number_token1] = ACTIONS(1921),
    [aux_sym_signed_number_token1] = ACTIONS(1919),
    [anon_sym_SQUOTE] = ACTIONS(1921),
    [anon_sym_DQUOTE] = ACTIONS(1921),
    [sym_prefix_operator] = ACTIONS(1921),
    [aux_sym_sign_operator_token1] = ACTIONS(1919),
    [sym_hash_operator] = ACTIONS(1921),
    [anon_sym_lsl] = ACTIONS(1919),
    [anon_sym_lsr] = ACTIONS(1919),
    [anon_sym_asr] = ACTIONS(1919),
    [aux_sym_pow_operator_token1] = ACTIONS(1921),
    [anon_sym_mod] = ACTIONS(1919),
    [anon_sym_land] = ACTIONS(1919),
    [anon_sym_lor] = ACTIONS(1919),
    [anon_sym_lxor] = ACTIONS(1919),
    [aux_sym_mult_operator_token1] = ACTIONS(1919),
    [aux_sym_add_operator_token1] = ACTIONS(1919),
    [sym_concat_operator] = ACTIONS(1921),
    [sym_rel_operator] = ACTIONS(1919),
    [anon_sym_AMP_AMP] = ACTIONS(1919),
    [anon_sym_or] = ACTIONS(1919),
    [anon_sym_PIPE_PIPE] = ACTIONS(1919),
    [sym_assign_operator] = ACTIONS(1921),
    [sym__uppercase_identifier] = ACTIONS(1921),
    [aux_sym_tag_token1] = ACTIONS(1921),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1039)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4228),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1039),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4211),
    [aux_sym_let_binding_repeat1] = STATE(1134),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2419),
    [anon_sym_let] = ACTIONS(2421),
    [anon_sym_and] = ACTIONS(2421),
    [anon_sym_EQ] = ACTIONS(2423),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2421),
    [anon_sym_type] = ACTIONS(2421),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_RBRACK] = ACTIONS(2419),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2421),
    [anon_sym_module] = ACTIONS(2421),
    [anon_sym_open] = ACTIONS(2421),
    [anon_sym_include] = ACTIONS(2421),
    [anon_sym_class] = ACTIONS(2421),
    [anon_sym_val] = ACTIONS(2421),
    [anon_sym_end] = ACTIONS(2421),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2419),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2419),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2419),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2419),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2421),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1040)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4333),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1040),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4347),
    [aux_sym_let_binding_repeat1] = STATE(1029),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2425),
    [anon_sym_let] = ACTIONS(2427),
    [anon_sym_and] = ACTIONS(2427),
    [anon_sym_EQ] = ACTIONS(2429),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2427),
    [anon_sym_type] = ACTIONS(2427),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_RBRACK] = ACTIONS(2425),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2427),
    [anon_sym_module] = ACTIONS(2427),
    [anon_sym_open] = ACTIONS(2427),
    [anon_sym_include] = ACTIONS(2427),
    [anon_sym_class] = ACTIONS(2427),
    [anon_sym_val] = ACTIONS(2427),
    [anon_sym_end] = ACTIONS(2427),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2425),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2425),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2425),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2425),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2427),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1041)] = {
    [sym_attribute] = STATE(1041),
    [sym__lowercase_identifier] = ACTIONS(1891),
    [anon_sym_TILDE] = ACTIONS(1891),
    [anon_sym_QMARK] = ACTIONS(1891),
    [anon_sym_LPAREN] = ACTIONS(1893),
    [anon_sym_RPAREN] = ACTIONS(1893),
    [anon_sym_COMMA] = ACTIONS(1893),
    [anon_sym_PIPE] = ACTIONS(1891),
    [anon_sym_COLON2] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_DASH_GT] = ACTIONS(1891),
    [anon_sym_LBRACK] = ACTIONS(1891),
    [anon_sym_RBRACK] = ACTIONS(1893),
    [anon_sym_true] = ACTIONS(1891),
    [anon_sym_false] = ACTIONS(1891),
    [anon_sym_LBRACE] = ACTIONS(1891),
    [anon_sym_SEMI] = ACTIONS(1893),
    [anon_sym_RBRACE] = ACTIONS(1893),
    [anon_sym_constraint] = ACTIONS(1891),
    [anon_sym_val] = ACTIONS(1891),
    [anon_sym_end] = ACTIONS(1891),
    [anon_sym_with] = ACTIONS(1891),
    [anon_sym_COLON_GT] = ACTIONS(1893),
    [anon_sym_object] = ACTIONS(1891),
    [anon_sym_inherit] = ACTIONS(1891),
    [anon_sym_method] = ACTIONS(1891),
    [anon_sym_initializer] = ACTIONS(1891),
    [anon_sym_AMP] = ACTIONS(1891),
    [anon_sym_POUND] = ACTIONS(1891),
    [anon_sym_COLON_COLON] = ACTIONS(1893),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1893),
    [anon_sym_LT_DASH] = ACTIONS(1891),
    [anon_sym_then] = ACTIONS(1891),
    [anon_sym_else] = ACTIONS(1891),
    [anon_sym_do] = ACTIONS(1891),
    [anon_sym_done] = ACTIONS(1891),
    [anon_sym_to] = ACTIONS(1891),
    [anon_sym_downto] = ACTIONS(1891),
    [anon_sym_new] = ACTIONS(1891),
    [anon_sym_LBRACE_LT] = ACTIONS(1893),
    [anon_sym_GT_RBRACE] = ACTIONS(1893),
    [anon_sym_begin] = ACTIONS(1891),
    [sym_ocamlyacc_value] = ACTIONS(1893),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1891),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1893),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1891),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1891),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1893),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1893),
    [aux_sym_number_token1] = ACTIONS(1893),
    [aux_sym_signed_number_token1] = ACTIONS(1891),
    [anon_sym_SQUOTE] = ACTIONS(1893),
    [anon_sym_DQUOTE] = ACTIONS(1893),
    [sym_prefix_operator] = ACTIONS(1893),
    [aux_sym_sign_operator_token1] = ACTIONS(1891),
    [sym_hash_operator] = ACTIONS(1893),
    [anon_sym_lsl] = ACTIONS(1891),
    [anon_sym_lsr] = ACTIONS(1891),
    [anon_sym_asr] = ACTIONS(1891),
    [aux_sym_pow_operator_token1] = ACTIONS(1893),
    [anon_sym_mod] = ACTIONS(1891),
    [anon_sym_land] = ACTIONS(1891),
    [anon_sym_lor] = ACTIONS(1891),
    [anon_sym_lxor] = ACTIONS(1891),
    [aux_sym_mult_operator_token1] = ACTIONS(1891),
    [aux_sym_add_operator_token1] = ACTIONS(1891),
    [sym_concat_operator] = ACTIONS(1893),
    [sym_rel_operator] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1891),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [sym_assign_operator] = ACTIONS(1893),
    [sym__uppercase_identifier] = ACTIONS(1893),
    [aux_sym_tag_token1] = ACTIONS(1893),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1042)] = {
    [sym_attribute] = STATE(1042),
    [sym__lowercase_identifier] = ACTIONS(1887),
    [anon_sym_TILDE] = ACTIONS(1887),
    [anon_sym_QMARK] = ACTIONS(1887),
    [anon_sym_LPAREN] = ACTIONS(1889),
    [anon_sym_RPAREN] = ACTIONS(1889),
    [anon_sym_COMMA] = ACTIONS(1889),
    [anon_sym_PIPE] = ACTIONS(1887),
    [anon_sym_COLON2] = ACTIONS(1887),
    [anon_sym_DOT] = ACTIONS(1889),
    [anon_sym_DASH_GT] = ACTIONS(1887),
    [anon_sym_LBRACK] = ACTIONS(1887),
    [anon_sym_RBRACK] = ACTIONS(1889),
    [anon_sym_true] = ACTIONS(1887),
    [anon_sym_false] = ACTIONS(1887),
    [anon_sym_LBRACE] = ACTIONS(1887),
    [anon_sym_SEMI] = ACTIONS(1889),
    [anon_sym_RBRACE] = ACTIONS(1889),
    [anon_sym_constraint] = ACTIONS(1887),
    [anon_sym_val] = ACTIONS(1887),
    [anon_sym_end] = ACTIONS(1887),
    [anon_sym_with] = ACTIONS(1887),
    [anon_sym_COLON_GT] = ACTIONS(1889),
    [anon_sym_object] = ACTIONS(1887),
    [anon_sym_inherit] = ACTIONS(1887),
    [anon_sym_method] = ACTIONS(1887),
    [anon_sym_initializer] = ACTIONS(1887),
    [anon_sym_AMP] = ACTIONS(1887),
    [anon_sym_POUND] = ACTIONS(1887),
    [anon_sym_COLON_COLON] = ACTIONS(1889),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1889),
    [anon_sym_LT_DASH] = ACTIONS(1887),
    [anon_sym_then] = ACTIONS(1887),
    [anon_sym_else] = ACTIONS(1887),
    [anon_sym_do] = ACTIONS(1887),
    [anon_sym_done] = ACTIONS(1887),
    [anon_sym_to] = ACTIONS(1887),
    [anon_sym_downto] = ACTIONS(1887),
    [anon_sym_new] = ACTIONS(1887),
    [anon_sym_LBRACE_LT] = ACTIONS(1889),
    [anon_sym_GT_RBRACE] = ACTIONS(1889),
    [anon_sym_begin] = ACTIONS(1887),
    [sym_ocamlyacc_value] = ACTIONS(1889),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1887),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1889),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1887),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1887),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1889),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1889),
    [aux_sym_number_token1] = ACTIONS(1889),
    [aux_sym_signed_number_token1] = ACTIONS(1887),
    [anon_sym_SQUOTE] = ACTIONS(1889),
    [anon_sym_DQUOTE] = ACTIONS(1889),
    [sym_prefix_operator] = ACTIONS(1889),
    [aux_sym_sign_operator_token1] = ACTIONS(1887),
    [sym_hash_operator] = ACTIONS(1889),
    [anon_sym_lsl] = ACTIONS(1887),
    [anon_sym_lsr] = ACTIONS(1887),
    [anon_sym_asr] = ACTIONS(1887),
    [aux_sym_pow_operator_token1] = ACTIONS(1889),
    [anon_sym_mod] = ACTIONS(1887),
    [anon_sym_land] = ACTIONS(1887),
    [anon_sym_lor] = ACTIONS(1887),
    [anon_sym_lxor] = ACTIONS(1887),
    [aux_sym_mult_operator_token1] = ACTIONS(1887),
    [aux_sym_add_operator_token1] = ACTIONS(1887),
    [sym_concat_operator] = ACTIONS(1889),
    [sym_rel_operator] = ACTIONS(1887),
    [anon_sym_AMP_AMP] = ACTIONS(1887),
    [anon_sym_or] = ACTIONS(1887),
    [anon_sym_PIPE_PIPE] = ACTIONS(1887),
    [sym_assign_operator] = ACTIONS(1889),
    [sym__uppercase_identifier] = ACTIONS(1889),
    [aux_sym_tag_token1] = ACTIONS(1889),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1043)] = {
    [sym_attribute] = STATE(1043),
    [sym__lowercase_identifier] = ACTIONS(2063),
    [anon_sym_TILDE] = ACTIONS(2063),
    [anon_sym_QMARK] = ACTIONS(2063),
    [anon_sym_LPAREN] = ACTIONS(2065),
    [anon_sym_RPAREN] = ACTIONS(2065),
    [anon_sym_COMMA] = ACTIONS(2065),
    [anon_sym_PIPE] = ACTIONS(2063),
    [anon_sym_COLON2] = ACTIONS(2063),
    [anon_sym_DOT] = ACTIONS(2065),
    [anon_sym_DASH_GT] = ACTIONS(2063),
    [anon_sym_LBRACK] = ACTIONS(2063),
    [anon_sym_RBRACK] = ACTIONS(2065),
    [anon_sym_true] = ACTIONS(2063),
    [anon_sym_false] = ACTIONS(2063),
    [anon_sym_LBRACE] = ACTIONS(2063),
    [anon_sym_SEMI] = ACTIONS(2065),
    [anon_sym_RBRACE] = ACTIONS(2065),
    [anon_sym_constraint] = ACTIONS(2063),
    [anon_sym_val] = ACTIONS(2063),
    [anon_sym_end] = ACTIONS(2063),
    [anon_sym_with] = ACTIONS(2063),
    [anon_sym_COLON_GT] = ACTIONS(2065),
    [anon_sym_object] = ACTIONS(2063),
    [anon_sym_inherit] = ACTIONS(2063),
    [anon_sym_method] = ACTIONS(2063),
    [anon_sym_initializer] = ACTIONS(2063),
    [anon_sym_AMP] = ACTIONS(2063),
    [anon_sym_POUND] = ACTIONS(2063),
    [anon_sym_COLON_COLON] = ACTIONS(2065),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2065),
    [anon_sym_then] = ACTIONS(2063),
    [anon_sym_else] = ACTIONS(2063),
    [anon_sym_do] = ACTIONS(2063),
    [anon_sym_done] = ACTIONS(2063),
    [anon_sym_to] = ACTIONS(2063),
    [anon_sym_downto] = ACTIONS(2063),
    [anon_sym_new] = ACTIONS(2063),
    [anon_sym_LBRACE_LT] = ACTIONS(2065),
    [anon_sym_GT_RBRACE] = ACTIONS(2065),
    [anon_sym_begin] = ACTIONS(2063),
    [sym_ocamlyacc_value] = ACTIONS(2065),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2063),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2065),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2063),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2063),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2065),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2065),
    [aux_sym_number_token1] = ACTIONS(2065),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2065),
    [sym_prefix_operator] = ACTIONS(2065),
    [aux_sym_sign_operator_token1] = ACTIONS(2063),
    [sym_hash_operator] = ACTIONS(2065),
    [anon_sym_lsl] = ACTIONS(2063),
    [anon_sym_lsr] = ACTIONS(2063),
    [anon_sym_asr] = ACTIONS(2063),
    [aux_sym_pow_operator_token1] = ACTIONS(2065),
    [anon_sym_mod] = ACTIONS(2063),
    [anon_sym_land] = ACTIONS(2063),
    [anon_sym_lor] = ACTIONS(2063),
    [anon_sym_lxor] = ACTIONS(2063),
    [aux_sym_mult_operator_token1] = ACTIONS(2063),
    [aux_sym_add_operator_token1] = ACTIONS(2063),
    [sym_concat_operator] = ACTIONS(2065),
    [sym_rel_operator] = ACTIONS(2063),
    [anon_sym_AMP_AMP] = ACTIONS(2063),
    [anon_sym_or] = ACTIONS(2063),
    [anon_sym_PIPE_PIPE] = ACTIONS(2063),
    [sym_assign_operator] = ACTIONS(2065),
    [sym__uppercase_identifier] = ACTIONS(2065),
    [aux_sym_tag_token1] = ACTIONS(2065),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1044)] = {
    [sym_attribute] = STATE(1044),
    [sym__lowercase_identifier] = ACTIONS(2039),
    [anon_sym_TILDE] = ACTIONS(2039),
    [anon_sym_QMARK] = ACTIONS(2039),
    [anon_sym_LPAREN] = ACTIONS(2041),
    [anon_sym_RPAREN] = ACTIONS(2041),
    [anon_sym_COMMA] = ACTIONS(2041),
    [anon_sym_PIPE] = ACTIONS(2039),
    [anon_sym_COLON2] = ACTIONS(2039),
    [anon_sym_DOT] = ACTIONS(993),
    [anon_sym_DASH_GT] = ACTIONS(2039),
    [anon_sym_LBRACK] = ACTIONS(2039),
    [anon_sym_RBRACK] = ACTIONS(2041),
    [anon_sym_true] = ACTIONS(2039),
    [anon_sym_false] = ACTIONS(2039),
    [anon_sym_LBRACE] = ACTIONS(2039),
    [anon_sym_SEMI] = ACTIONS(2041),
    [anon_sym_RBRACE] = ACTIONS(2041),
    [anon_sym_constraint] = ACTIONS(2039),
    [anon_sym_val] = ACTIONS(2039),
    [anon_sym_end] = ACTIONS(2039),
    [anon_sym_with] = ACTIONS(2039),
    [anon_sym_COLON_GT] = ACTIONS(2041),
    [anon_sym_object] = ACTIONS(2039),
    [anon_sym_inherit] = ACTIONS(2039),
    [anon_sym_method] = ACTIONS(2039),
    [anon_sym_initializer] = ACTIONS(2039),
    [anon_sym_AMP] = ACTIONS(2039),
    [anon_sym_POUND] = ACTIONS(995),
    [anon_sym_COLON_COLON] = ACTIONS(2041),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2041),
    [anon_sym_then] = ACTIONS(2039),
    [anon_sym_else] = ACTIONS(2039),
    [anon_sym_do] = ACTIONS(2039),
    [anon_sym_done] = ACTIONS(2039),
    [anon_sym_to] = ACTIONS(2039),
    [anon_sym_downto] = ACTIONS(2039),
    [anon_sym_new] = ACTIONS(2039),
    [anon_sym_LBRACE_LT] = ACTIONS(2041),
    [anon_sym_GT_RBRACE] = ACTIONS(2041),
    [anon_sym_begin] = ACTIONS(2039),
    [sym_ocamlyacc_value] = ACTIONS(2041),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2039),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2041),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2039),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2039),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2041),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2041),
    [aux_sym_number_token1] = ACTIONS(2041),
    [aux_sym_signed_number_token1] = ACTIONS(2039),
    [anon_sym_SQUOTE] = ACTIONS(2041),
    [anon_sym_DQUOTE] = ACTIONS(2041),
    [sym_prefix_operator] = ACTIONS(2041),
    [aux_sym_sign_operator_token1] = ACTIONS(2039),
    [sym_hash_operator] = ACTIONS(997),
    [anon_sym_lsl] = ACTIONS(2039),
    [anon_sym_lsr] = ACTIONS(2039),
    [anon_sym_asr] = ACTIONS(2039),
    [aux_sym_pow_operator_token1] = ACTIONS(2041),
    [anon_sym_mod] = ACTIONS(2039),
    [anon_sym_land] = ACTIONS(2039),
    [anon_sym_lor] = ACTIONS(2039),
    [anon_sym_lxor] = ACTIONS(2039),
    [aux_sym_mult_operator_token1] = ACTIONS(2039),
    [aux_sym_add_operator_token1] = ACTIONS(2039),
    [sym_concat_operator] = ACTIONS(2041),
    [sym_rel_operator] = ACTIONS(2039),
    [anon_sym_AMP_AMP] = ACTIONS(2039),
    [anon_sym_or] = ACTIONS(2039),
    [anon_sym_PIPE_PIPE] = ACTIONS(2039),
    [sym_assign_operator] = ACTIONS(2041),
    [sym__uppercase_identifier] = ACTIONS(2041),
    [aux_sym_tag_token1] = ACTIONS(2041),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1045)] = {
    [sym_attribute] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(1947),
    [anon_sym_TILDE] = ACTIONS(1947),
    [anon_sym_QMARK] = ACTIONS(1947),
    [anon_sym_LPAREN] = ACTIONS(1949),
    [anon_sym_RPAREN] = ACTIONS(1949),
    [anon_sym_COMMA] = ACTIONS(1949),
    [anon_sym_PIPE] = ACTIONS(1947),
    [anon_sym_COLON2] = ACTIONS(1947),
    [anon_sym_DOT] = ACTIONS(1949),
    [anon_sym_DASH_GT] = ACTIONS(1947),
    [anon_sym_LBRACK] = ACTIONS(1947),
    [anon_sym_RBRACK] = ACTIONS(1949),
    [anon_sym_true] = ACTIONS(1947),
    [anon_sym_false] = ACTIONS(1947),
    [anon_sym_LBRACE] = ACTIONS(1947),
    [anon_sym_SEMI] = ACTIONS(1949),
    [anon_sym_RBRACE] = ACTIONS(1949),
    [anon_sym_constraint] = ACTIONS(1947),
    [anon_sym_val] = ACTIONS(1947),
    [anon_sym_end] = ACTIONS(1947),
    [anon_sym_with] = ACTIONS(1947),
    [anon_sym_COLON_GT] = ACTIONS(1949),
    [anon_sym_object] = ACTIONS(1947),
    [anon_sym_inherit] = ACTIONS(1947),
    [anon_sym_method] = ACTIONS(1947),
    [anon_sym_initializer] = ACTIONS(1947),
    [anon_sym_AMP] = ACTIONS(1947),
    [anon_sym_POUND] = ACTIONS(1947),
    [anon_sym_COLON_COLON] = ACTIONS(1949),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1949),
    [anon_sym_then] = ACTIONS(1947),
    [anon_sym_else] = ACTIONS(1947),
    [anon_sym_do] = ACTIONS(1947),
    [anon_sym_done] = ACTIONS(1947),
    [anon_sym_to] = ACTIONS(1947),
    [anon_sym_downto] = ACTIONS(1947),
    [anon_sym_new] = ACTIONS(1947),
    [anon_sym_LBRACE_LT] = ACTIONS(1949),
    [anon_sym_GT_RBRACE] = ACTIONS(1949),
    [anon_sym_begin] = ACTIONS(1947),
    [sym_ocamlyacc_value] = ACTIONS(1949),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1947),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1949),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1949),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1949),
    [aux_sym_number_token1] = ACTIONS(1949),
    [aux_sym_signed_number_token1] = ACTIONS(1947),
    [anon_sym_SQUOTE] = ACTIONS(1949),
    [anon_sym_DQUOTE] = ACTIONS(1949),
    [sym_prefix_operator] = ACTIONS(1949),
    [aux_sym_sign_operator_token1] = ACTIONS(1947),
    [sym_hash_operator] = ACTIONS(1949),
    [anon_sym_lsl] = ACTIONS(1947),
    [anon_sym_lsr] = ACTIONS(1947),
    [anon_sym_asr] = ACTIONS(1947),
    [aux_sym_pow_operator_token1] = ACTIONS(1949),
    [anon_sym_mod] = ACTIONS(1947),
    [anon_sym_land] = ACTIONS(1947),
    [anon_sym_lor] = ACTIONS(1947),
    [anon_sym_lxor] = ACTIONS(1947),
    [aux_sym_mult_operator_token1] = ACTIONS(1947),
    [aux_sym_add_operator_token1] = ACTIONS(1947),
    [sym_concat_operator] = ACTIONS(1949),
    [sym_rel_operator] = ACTIONS(1947),
    [anon_sym_AMP_AMP] = ACTIONS(1947),
    [anon_sym_or] = ACTIONS(1947),
    [anon_sym_PIPE_PIPE] = ACTIONS(1947),
    [sym_assign_operator] = ACTIONS(1949),
    [sym__uppercase_identifier] = ACTIONS(1949),
    [aux_sym_tag_token1] = ACTIONS(1949),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1046)] = {
    [sym_attribute] = STATE(1046),
    [sym__lowercase_identifier] = ACTIONS(1995),
    [anon_sym_TILDE] = ACTIONS(1995),
    [anon_sym_QMARK] = ACTIONS(1995),
    [anon_sym_LPAREN] = ACTIONS(1997),
    [anon_sym_RPAREN] = ACTIONS(1997),
    [anon_sym_COMMA] = ACTIONS(1997),
    [anon_sym_PIPE] = ACTIONS(1995),
    [anon_sym_COLON2] = ACTIONS(1995),
    [anon_sym_DOT] = ACTIONS(1997),
    [anon_sym_DASH_GT] = ACTIONS(1995),
    [anon_sym_LBRACK] = ACTIONS(1995),
    [anon_sym_RBRACK] = ACTIONS(1997),
    [anon_sym_true] = ACTIONS(1995),
    [anon_sym_false] = ACTIONS(1995),
    [anon_sym_LBRACE] = ACTIONS(1995),
    [anon_sym_SEMI] = ACTIONS(1997),
    [anon_sym_RBRACE] = ACTIONS(1997),
    [anon_sym_constraint] = ACTIONS(1995),
    [anon_sym_val] = ACTIONS(1995),
    [anon_sym_end] = ACTIONS(1995),
    [anon_sym_with] = ACTIONS(1995),
    [anon_sym_COLON_GT] = ACTIONS(1997),
    [anon_sym_object] = ACTIONS(1995),
    [anon_sym_inherit] = ACTIONS(1995),
    [anon_sym_method] = ACTIONS(1995),
    [anon_sym_initializer] = ACTIONS(1995),
    [anon_sym_AMP] = ACTIONS(1995),
    [anon_sym_POUND] = ACTIONS(1995),
    [anon_sym_COLON_COLON] = ACTIONS(1997),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1997),
    [anon_sym_then] = ACTIONS(1995),
    [anon_sym_else] = ACTIONS(1995),
    [anon_sym_do] = ACTIONS(1995),
    [anon_sym_done] = ACTIONS(1995),
    [anon_sym_to] = ACTIONS(1995),
    [anon_sym_downto] = ACTIONS(1995),
    [anon_sym_new] = ACTIONS(1995),
    [anon_sym_LBRACE_LT] = ACTIONS(1997),
    [anon_sym_GT_RBRACE] = ACTIONS(1997),
    [anon_sym_begin] = ACTIONS(1995),
    [sym_ocamlyacc_value] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1995),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1997),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1995),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1995),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1997),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1997),
    [aux_sym_number_token1] = ACTIONS(1997),
    [aux_sym_signed_number_token1] = ACTIONS(1995),
    [anon_sym_SQUOTE] = ACTIONS(1997),
    [anon_sym_DQUOTE] = ACTIONS(1997),
    [sym_prefix_operator] = ACTIONS(1997),
    [aux_sym_sign_operator_token1] = ACTIONS(1995),
    [sym_hash_operator] = ACTIONS(1997),
    [anon_sym_lsl] = ACTIONS(1995),
    [anon_sym_lsr] = ACTIONS(1995),
    [anon_sym_asr] = ACTIONS(1995),
    [aux_sym_pow_operator_token1] = ACTIONS(1997),
    [anon_sym_mod] = ACTIONS(1995),
    [anon_sym_land] = ACTIONS(1995),
    [anon_sym_lor] = ACTIONS(1995),
    [anon_sym_lxor] = ACTIONS(1995),
    [aux_sym_mult_operator_token1] = ACTIONS(1995),
    [aux_sym_add_operator_token1] = ACTIONS(1995),
    [sym_concat_operator] = ACTIONS(1997),
    [sym_rel_operator] = ACTIONS(1995),
    [anon_sym_AMP_AMP] = ACTIONS(1995),
    [anon_sym_or] = ACTIONS(1995),
    [anon_sym_PIPE_PIPE] = ACTIONS(1995),
    [sym_assign_operator] = ACTIONS(1997),
    [sym__uppercase_identifier] = ACTIONS(1997),
    [aux_sym_tag_token1] = ACTIONS(1997),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1047)] = {
    [sym_attribute] = STATE(1047),
    [sym__lowercase_identifier] = ACTIONS(2043),
    [anon_sym_TILDE] = ACTIONS(2043),
    [anon_sym_QMARK] = ACTIONS(2043),
    [anon_sym_LPAREN] = ACTIONS(2045),
    [anon_sym_RPAREN] = ACTIONS(2045),
    [anon_sym_COMMA] = ACTIONS(2045),
    [anon_sym_PIPE] = ACTIONS(2043),
    [anon_sym_COLON2] = ACTIONS(2043),
    [anon_sym_DOT] = ACTIONS(2045),
    [anon_sym_DASH_GT] = ACTIONS(2043),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_RBRACK] = ACTIONS(2045),
    [anon_sym_true] = ACTIONS(2043),
    [anon_sym_false] = ACTIONS(2043),
    [anon_sym_LBRACE] = ACTIONS(2043),
    [anon_sym_SEMI] = ACTIONS(2045),
    [anon_sym_RBRACE] = ACTIONS(2045),
    [anon_sym_constraint] = ACTIONS(2043),
    [anon_sym_val] = ACTIONS(2043),
    [anon_sym_end] = ACTIONS(2043),
    [anon_sym_with] = ACTIONS(2043),
    [anon_sym_COLON_GT] = ACTIONS(2045),
    [anon_sym_object] = ACTIONS(2043),
    [anon_sym_inherit] = ACTIONS(2043),
    [anon_sym_method] = ACTIONS(2043),
    [anon_sym_initializer] = ACTIONS(2043),
    [anon_sym_AMP] = ACTIONS(2043),
    [anon_sym_POUND] = ACTIONS(2043),
    [anon_sym_COLON_COLON] = ACTIONS(2045),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2045),
    [anon_sym_then] = ACTIONS(2043),
    [anon_sym_else] = ACTIONS(2043),
    [anon_sym_do] = ACTIONS(2043),
    [anon_sym_done] = ACTIONS(2043),
    [anon_sym_to] = ACTIONS(2043),
    [anon_sym_downto] = ACTIONS(2043),
    [anon_sym_new] = ACTIONS(2043),
    [anon_sym_LBRACE_LT] = ACTIONS(2045),
    [anon_sym_GT_RBRACE] = ACTIONS(2045),
    [anon_sym_begin] = ACTIONS(2043),
    [sym_ocamlyacc_value] = ACTIONS(2045),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2043),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2045),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2043),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2043),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2045),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2045),
    [aux_sym_number_token1] = ACTIONS(2045),
    [aux_sym_signed_number_token1] = ACTIONS(2043),
    [anon_sym_SQUOTE] = ACTIONS(2045),
    [anon_sym_DQUOTE] = ACTIONS(2045),
    [sym_prefix_operator] = ACTIONS(2045),
    [aux_sym_sign_operator_token1] = ACTIONS(2043),
    [sym_hash_operator] = ACTIONS(2045),
    [anon_sym_lsl] = ACTIONS(2043),
    [anon_sym_lsr] = ACTIONS(2043),
    [anon_sym_asr] = ACTIONS(2043),
    [aux_sym_pow_operator_token1] = ACTIONS(2045),
    [anon_sym_mod] = ACTIONS(2043),
    [anon_sym_land] = ACTIONS(2043),
    [anon_sym_lor] = ACTIONS(2043),
    [anon_sym_lxor] = ACTIONS(2043),
    [aux_sym_mult_operator_token1] = ACTIONS(2043),
    [aux_sym_add_operator_token1] = ACTIONS(2043),
    [sym_concat_operator] = ACTIONS(2045),
    [sym_rel_operator] = ACTIONS(2043),
    [anon_sym_AMP_AMP] = ACTIONS(2043),
    [anon_sym_or] = ACTIONS(2043),
    [anon_sym_PIPE_PIPE] = ACTIONS(2043),
    [sym_assign_operator] = ACTIONS(2045),
    [sym__uppercase_identifier] = ACTIONS(2045),
    [aux_sym_tag_token1] = ACTIONS(2045),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1048)] = {
    [sym_attribute] = STATE(1048),
    [sym__lowercase_identifier] = ACTIONS(2071),
    [anon_sym_TILDE] = ACTIONS(2071),
    [anon_sym_QMARK] = ACTIONS(2071),
    [anon_sym_LPAREN] = ACTIONS(2073),
    [anon_sym_RPAREN] = ACTIONS(2073),
    [anon_sym_COMMA] = ACTIONS(2073),
    [anon_sym_PIPE] = ACTIONS(2071),
    [anon_sym_COLON2] = ACTIONS(2071),
    [anon_sym_DOT] = ACTIONS(2073),
    [anon_sym_DASH_GT] = ACTIONS(2071),
    [anon_sym_LBRACK] = ACTIONS(2071),
    [anon_sym_RBRACK] = ACTIONS(2073),
    [anon_sym_true] = ACTIONS(2071),
    [anon_sym_false] = ACTIONS(2071),
    [anon_sym_LBRACE] = ACTIONS(2071),
    [anon_sym_SEMI] = ACTIONS(2073),
    [anon_sym_RBRACE] = ACTIONS(2073),
    [anon_sym_constraint] = ACTIONS(2071),
    [anon_sym_val] = ACTIONS(2071),
    [anon_sym_end] = ACTIONS(2071),
    [anon_sym_with] = ACTIONS(2071),
    [anon_sym_COLON_GT] = ACTIONS(2073),
    [anon_sym_object] = ACTIONS(2071),
    [anon_sym_inherit] = ACTIONS(2071),
    [anon_sym_method] = ACTIONS(2071),
    [anon_sym_initializer] = ACTIONS(2071),
    [anon_sym_AMP] = ACTIONS(2071),
    [anon_sym_POUND] = ACTIONS(2071),
    [anon_sym_COLON_COLON] = ACTIONS(2073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2073),
    [anon_sym_then] = ACTIONS(2071),
    [anon_sym_else] = ACTIONS(2071),
    [anon_sym_do] = ACTIONS(2071),
    [anon_sym_done] = ACTIONS(2071),
    [anon_sym_to] = ACTIONS(2071),
    [anon_sym_downto] = ACTIONS(2071),
    [anon_sym_new] = ACTIONS(2071),
    [anon_sym_LBRACE_LT] = ACTIONS(2073),
    [anon_sym_GT_RBRACE] = ACTIONS(2073),
    [anon_sym_begin] = ACTIONS(2071),
    [sym_ocamlyacc_value] = ACTIONS(2073),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2071),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2073),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2071),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2071),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2073),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2073),
    [aux_sym_number_token1] = ACTIONS(2073),
    [aux_sym_signed_number_token1] = ACTIONS(2071),
    [anon_sym_SQUOTE] = ACTIONS(2073),
    [anon_sym_DQUOTE] = ACTIONS(2073),
    [sym_prefix_operator] = ACTIONS(2073),
    [aux_sym_sign_operator_token1] = ACTIONS(2071),
    [sym_hash_operator] = ACTIONS(2073),
    [anon_sym_lsl] = ACTIONS(2071),
    [anon_sym_lsr] = ACTIONS(2071),
    [anon_sym_asr] = ACTIONS(2071),
    [aux_sym_pow_operator_token1] = ACTIONS(2073),
    [anon_sym_mod] = ACTIONS(2071),
    [anon_sym_land] = ACTIONS(2071),
    [anon_sym_lor] = ACTIONS(2071),
    [anon_sym_lxor] = ACTIONS(2071),
    [aux_sym_mult_operator_token1] = ACTIONS(2071),
    [aux_sym_add_operator_token1] = ACTIONS(2071),
    [sym_concat_operator] = ACTIONS(2073),
    [sym_rel_operator] = ACTIONS(2071),
    [anon_sym_AMP_AMP] = ACTIONS(2071),
    [anon_sym_or] = ACTIONS(2071),
    [anon_sym_PIPE_PIPE] = ACTIONS(2071),
    [sym_assign_operator] = ACTIONS(2073),
    [sym__uppercase_identifier] = ACTIONS(2073),
    [aux_sym_tag_token1] = ACTIONS(2073),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1049)] = {
    [sym_attribute] = STATE(1049),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_RPAREN] = ACTIONS(2081),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_PIPE] = ACTIONS(2079),
    [anon_sym_COLON2] = ACTIONS(2079),
    [anon_sym_DOT] = ACTIONS(2081),
    [anon_sym_DASH_GT] = ACTIONS(2079),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_RBRACK] = ACTIONS(2081),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_SEMI] = ACTIONS(2081),
    [anon_sym_RBRACE] = ACTIONS(2081),
    [anon_sym_constraint] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_end] = ACTIONS(2079),
    [anon_sym_with] = ACTIONS(2079),
    [anon_sym_COLON_GT] = ACTIONS(2081),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_inherit] = ACTIONS(2079),
    [anon_sym_method] = ACTIONS(2079),
    [anon_sym_initializer] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_POUND] = ACTIONS(2079),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_then] = ACTIONS(2079),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_do] = ACTIONS(2079),
    [anon_sym_done] = ACTIONS(2079),
    [anon_sym_to] = ACTIONS(2079),
    [anon_sym_downto] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_GT_RBRACE] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [sym_hash_operator] = ACTIONS(2081),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1050)] = {
    [sym_attribute] = STATE(1050),
    [sym__lowercase_identifier] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_QMARK] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2143),
    [anon_sym_RPAREN] = ACTIONS(2143),
    [anon_sym_COMMA] = ACTIONS(2143),
    [anon_sym_PIPE] = ACTIONS(2141),
    [anon_sym_COLON2] = ACTIONS(2141),
    [anon_sym_DOT] = ACTIONS(2143),
    [anon_sym_DASH_GT] = ACTIONS(2141),
    [anon_sym_LBRACK] = ACTIONS(2141),
    [anon_sym_RBRACK] = ACTIONS(2143),
    [anon_sym_true] = ACTIONS(2141),
    [anon_sym_false] = ACTIONS(2141),
    [anon_sym_LBRACE] = ACTIONS(2141),
    [anon_sym_SEMI] = ACTIONS(2143),
    [anon_sym_RBRACE] = ACTIONS(2143),
    [anon_sym_constraint] = ACTIONS(2141),
    [anon_sym_val] = ACTIONS(2141),
    [anon_sym_end] = ACTIONS(2141),
    [anon_sym_with] = ACTIONS(2141),
    [anon_sym_COLON_GT] = ACTIONS(2143),
    [anon_sym_object] = ACTIONS(2141),
    [anon_sym_inherit] = ACTIONS(2141),
    [anon_sym_method] = ACTIONS(2141),
    [anon_sym_initializer] = ACTIONS(2141),
    [anon_sym_AMP] = ACTIONS(2141),
    [anon_sym_POUND] = ACTIONS(2141),
    [anon_sym_COLON_COLON] = ACTIONS(2143),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2143),
    [anon_sym_then] = ACTIONS(2141),
    [anon_sym_else] = ACTIONS(2141),
    [anon_sym_do] = ACTIONS(2141),
    [anon_sym_done] = ACTIONS(2141),
    [anon_sym_to] = ACTIONS(2141),
    [anon_sym_downto] = ACTIONS(2141),
    [anon_sym_new] = ACTIONS(2141),
    [anon_sym_LBRACE_LT] = ACTIONS(2143),
    [anon_sym_GT_RBRACE] = ACTIONS(2143),
    [anon_sym_begin] = ACTIONS(2141),
    [sym_ocamlyacc_value] = ACTIONS(2143),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2141),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2143),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2141),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2141),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2143),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2143),
    [aux_sym_number_token1] = ACTIONS(2143),
    [aux_sym_signed_number_token1] = ACTIONS(2141),
    [anon_sym_SQUOTE] = ACTIONS(2143),
    [anon_sym_DQUOTE] = ACTIONS(2143),
    [sym_prefix_operator] = ACTIONS(2143),
    [aux_sym_sign_operator_token1] = ACTIONS(2141),
    [sym_hash_operator] = ACTIONS(2143),
    [anon_sym_lsl] = ACTIONS(2141),
    [anon_sym_lsr] = ACTIONS(2141),
    [anon_sym_asr] = ACTIONS(2141),
    [aux_sym_pow_operator_token1] = ACTIONS(2143),
    [anon_sym_mod] = ACTIONS(2141),
    [anon_sym_land] = ACTIONS(2141),
    [anon_sym_lor] = ACTIONS(2141),
    [anon_sym_lxor] = ACTIONS(2141),
    [aux_sym_mult_operator_token1] = ACTIONS(2141),
    [aux_sym_add_operator_token1] = ACTIONS(2141),
    [sym_concat_operator] = ACTIONS(2143),
    [sym_rel_operator] = ACTIONS(2141),
    [anon_sym_AMP_AMP] = ACTIONS(2141),
    [anon_sym_or] = ACTIONS(2141),
    [anon_sym_PIPE_PIPE] = ACTIONS(2141),
    [sym_assign_operator] = ACTIONS(2143),
    [sym__uppercase_identifier] = ACTIONS(2143),
    [aux_sym_tag_token1] = ACTIONS(2143),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1051)] = {
    [sym_attribute] = STATE(1051),
    [sym__lowercase_identifier] = ACTIONS(1935),
    [anon_sym_TILDE] = ACTIONS(1935),
    [anon_sym_QMARK] = ACTIONS(1935),
    [anon_sym_LPAREN] = ACTIONS(1937),
    [anon_sym_RPAREN] = ACTIONS(1937),
    [anon_sym_COMMA] = ACTIONS(1937),
    [anon_sym_PIPE] = ACTIONS(1935),
    [anon_sym_COLON2] = ACTIONS(1935),
    [anon_sym_DOT] = ACTIONS(1937),
    [anon_sym_DASH_GT] = ACTIONS(1935),
    [anon_sym_LBRACK] = ACTIONS(1935),
    [anon_sym_RBRACK] = ACTIONS(1937),
    [anon_sym_true] = ACTIONS(1935),
    [anon_sym_false] = ACTIONS(1935),
    [anon_sym_LBRACE] = ACTIONS(1935),
    [anon_sym_SEMI] = ACTIONS(1937),
    [anon_sym_RBRACE] = ACTIONS(1937),
    [anon_sym_constraint] = ACTIONS(1935),
    [anon_sym_val] = ACTIONS(1935),
    [anon_sym_end] = ACTIONS(1935),
    [anon_sym_with] = ACTIONS(1935),
    [anon_sym_COLON_GT] = ACTIONS(1937),
    [anon_sym_object] = ACTIONS(1935),
    [anon_sym_inherit] = ACTIONS(1935),
    [anon_sym_method] = ACTIONS(1935),
    [anon_sym_initializer] = ACTIONS(1935),
    [anon_sym_AMP] = ACTIONS(1935),
    [anon_sym_POUND] = ACTIONS(1935),
    [anon_sym_COLON_COLON] = ACTIONS(1937),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1937),
    [anon_sym_then] = ACTIONS(1935),
    [anon_sym_else] = ACTIONS(1935),
    [anon_sym_do] = ACTIONS(1935),
    [anon_sym_done] = ACTIONS(1935),
    [anon_sym_to] = ACTIONS(1935),
    [anon_sym_downto] = ACTIONS(1935),
    [anon_sym_new] = ACTIONS(1935),
    [anon_sym_LBRACE_LT] = ACTIONS(1937),
    [anon_sym_GT_RBRACE] = ACTIONS(1937),
    [anon_sym_begin] = ACTIONS(1935),
    [sym_ocamlyacc_value] = ACTIONS(1937),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1935),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1937),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1935),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1935),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1937),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1937),
    [aux_sym_number_token1] = ACTIONS(1937),
    [aux_sym_signed_number_token1] = ACTIONS(1935),
    [anon_sym_SQUOTE] = ACTIONS(1937),
    [anon_sym_DQUOTE] = ACTIONS(1937),
    [sym_prefix_operator] = ACTIONS(1937),
    [aux_sym_sign_operator_token1] = ACTIONS(1935),
    [sym_hash_operator] = ACTIONS(1937),
    [anon_sym_lsl] = ACTIONS(1935),
    [anon_sym_lsr] = ACTIONS(1935),
    [anon_sym_asr] = ACTIONS(1935),
    [aux_sym_pow_operator_token1] = ACTIONS(1937),
    [anon_sym_mod] = ACTIONS(1935),
    [anon_sym_land] = ACTIONS(1935),
    [anon_sym_lor] = ACTIONS(1935),
    [anon_sym_lxor] = ACTIONS(1935),
    [aux_sym_mult_operator_token1] = ACTIONS(1935),
    [aux_sym_add_operator_token1] = ACTIONS(1935),
    [sym_concat_operator] = ACTIONS(1937),
    [sym_rel_operator] = ACTIONS(1935),
    [anon_sym_AMP_AMP] = ACTIONS(1935),
    [anon_sym_or] = ACTIONS(1935),
    [anon_sym_PIPE_PIPE] = ACTIONS(1935),
    [sym_assign_operator] = ACTIONS(1937),
    [sym__uppercase_identifier] = ACTIONS(1937),
    [aux_sym_tag_token1] = ACTIONS(1937),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1052)] = {
    [sym_attribute] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(1939),
    [anon_sym_TILDE] = ACTIONS(1939),
    [anon_sym_QMARK] = ACTIONS(1939),
    [anon_sym_LPAREN] = ACTIONS(1941),
    [anon_sym_RPAREN] = ACTIONS(1941),
    [anon_sym_COMMA] = ACTIONS(1941),
    [anon_sym_PIPE] = ACTIONS(1939),
    [anon_sym_COLON2] = ACTIONS(1939),
    [anon_sym_DOT] = ACTIONS(1941),
    [anon_sym_DASH_GT] = ACTIONS(1939),
    [anon_sym_LBRACK] = ACTIONS(1939),
    [anon_sym_RBRACK] = ACTIONS(1941),
    [anon_sym_true] = ACTIONS(1939),
    [anon_sym_false] = ACTIONS(1939),
    [anon_sym_LBRACE] = ACTIONS(1939),
    [anon_sym_SEMI] = ACTIONS(1941),
    [anon_sym_RBRACE] = ACTIONS(1941),
    [anon_sym_constraint] = ACTIONS(1939),
    [anon_sym_val] = ACTIONS(1939),
    [anon_sym_end] = ACTIONS(1939),
    [anon_sym_with] = ACTIONS(1939),
    [anon_sym_COLON_GT] = ACTIONS(1941),
    [anon_sym_object] = ACTIONS(1939),
    [anon_sym_inherit] = ACTIONS(1939),
    [anon_sym_method] = ACTIONS(1939),
    [anon_sym_initializer] = ACTIONS(1939),
    [anon_sym_AMP] = ACTIONS(1939),
    [anon_sym_POUND] = ACTIONS(1939),
    [anon_sym_COLON_COLON] = ACTIONS(1941),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1941),
    [anon_sym_then] = ACTIONS(1939),
    [anon_sym_else] = ACTIONS(1939),
    [anon_sym_do] = ACTIONS(1939),
    [anon_sym_done] = ACTIONS(1939),
    [anon_sym_to] = ACTIONS(1939),
    [anon_sym_downto] = ACTIONS(1939),
    [anon_sym_new] = ACTIONS(1939),
    [anon_sym_LBRACE_LT] = ACTIONS(1941),
    [anon_sym_GT_RBRACE] = ACTIONS(1941),
    [anon_sym_begin] = ACTIONS(1939),
    [sym_ocamlyacc_value] = ACTIONS(1941),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1939),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1941),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1939),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1939),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1941),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1941),
    [aux_sym_number_token1] = ACTIONS(1941),
    [aux_sym_signed_number_token1] = ACTIONS(1939),
    [anon_sym_SQUOTE] = ACTIONS(1941),
    [anon_sym_DQUOTE] = ACTIONS(1941),
    [sym_prefix_operator] = ACTIONS(1941),
    [aux_sym_sign_operator_token1] = ACTIONS(1939),
    [sym_hash_operator] = ACTIONS(1941),
    [anon_sym_lsl] = ACTIONS(1939),
    [anon_sym_lsr] = ACTIONS(1939),
    [anon_sym_asr] = ACTIONS(1939),
    [aux_sym_pow_operator_token1] = ACTIONS(1941),
    [anon_sym_mod] = ACTIONS(1939),
    [anon_sym_land] = ACTIONS(1939),
    [anon_sym_lor] = ACTIONS(1939),
    [anon_sym_lxor] = ACTIONS(1939),
    [aux_sym_mult_operator_token1] = ACTIONS(1939),
    [aux_sym_add_operator_token1] = ACTIONS(1939),
    [sym_concat_operator] = ACTIONS(1941),
    [sym_rel_operator] = ACTIONS(1939),
    [anon_sym_AMP_AMP] = ACTIONS(1939),
    [anon_sym_or] = ACTIONS(1939),
    [anon_sym_PIPE_PIPE] = ACTIONS(1939),
    [sym_assign_operator] = ACTIONS(1941),
    [sym__uppercase_identifier] = ACTIONS(1941),
    [aux_sym_tag_token1] = ACTIONS(1941),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1053)] = {
    [sym_attribute] = STATE(1053),
    [sym__lowercase_identifier] = ACTIONS(1943),
    [anon_sym_TILDE] = ACTIONS(1943),
    [anon_sym_QMARK] = ACTIONS(1943),
    [anon_sym_LPAREN] = ACTIONS(1945),
    [anon_sym_RPAREN] = ACTIONS(1945),
    [anon_sym_COMMA] = ACTIONS(1945),
    [anon_sym_PIPE] = ACTIONS(1943),
    [anon_sym_COLON2] = ACTIONS(1943),
    [anon_sym_DOT] = ACTIONS(1945),
    [anon_sym_DASH_GT] = ACTIONS(1943),
    [anon_sym_LBRACK] = ACTIONS(1943),
    [anon_sym_RBRACK] = ACTIONS(1945),
    [anon_sym_true] = ACTIONS(1943),
    [anon_sym_false] = ACTIONS(1943),
    [anon_sym_LBRACE] = ACTIONS(1943),
    [anon_sym_SEMI] = ACTIONS(1945),
    [anon_sym_RBRACE] = ACTIONS(1945),
    [anon_sym_constraint] = ACTIONS(1943),
    [anon_sym_val] = ACTIONS(1943),
    [anon_sym_end] = ACTIONS(1943),
    [anon_sym_with] = ACTIONS(1943),
    [anon_sym_COLON_GT] = ACTIONS(1945),
    [anon_sym_object] = ACTIONS(1943),
    [anon_sym_inherit] = ACTIONS(1943),
    [anon_sym_method] = ACTIONS(1943),
    [anon_sym_initializer] = ACTIONS(1943),
    [anon_sym_AMP] = ACTIONS(1943),
    [anon_sym_POUND] = ACTIONS(1943),
    [anon_sym_COLON_COLON] = ACTIONS(1945),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1945),
    [anon_sym_then] = ACTIONS(1943),
    [anon_sym_else] = ACTIONS(1943),
    [anon_sym_do] = ACTIONS(1943),
    [anon_sym_done] = ACTIONS(1943),
    [anon_sym_to] = ACTIONS(1943),
    [anon_sym_downto] = ACTIONS(1943),
    [anon_sym_new] = ACTIONS(1943),
    [anon_sym_LBRACE_LT] = ACTIONS(1945),
    [anon_sym_GT_RBRACE] = ACTIONS(1945),
    [anon_sym_begin] = ACTIONS(1943),
    [sym_ocamlyacc_value] = ACTIONS(1945),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1943),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1945),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1943),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1943),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1945),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1945),
    [aux_sym_number_token1] = ACTIONS(1945),
    [aux_sym_signed_number_token1] = ACTIONS(1943),
    [anon_sym_SQUOTE] = ACTIONS(1945),
    [anon_sym_DQUOTE] = ACTIONS(1945),
    [sym_prefix_operator] = ACTIONS(1945),
    [aux_sym_sign_operator_token1] = ACTIONS(1943),
    [sym_hash_operator] = ACTIONS(1945),
    [anon_sym_lsl] = ACTIONS(1943),
    [anon_sym_lsr] = ACTIONS(1943),
    [anon_sym_asr] = ACTIONS(1943),
    [aux_sym_pow_operator_token1] = ACTIONS(1945),
    [anon_sym_mod] = ACTIONS(1943),
    [anon_sym_land] = ACTIONS(1943),
    [anon_sym_lor] = ACTIONS(1943),
    [anon_sym_lxor] = ACTIONS(1943),
    [aux_sym_mult_operator_token1] = ACTIONS(1943),
    [aux_sym_add_operator_token1] = ACTIONS(1943),
    [sym_concat_operator] = ACTIONS(1945),
    [sym_rel_operator] = ACTIONS(1943),
    [anon_sym_AMP_AMP] = ACTIONS(1943),
    [anon_sym_or] = ACTIONS(1943),
    [anon_sym_PIPE_PIPE] = ACTIONS(1943),
    [sym_assign_operator] = ACTIONS(1945),
    [sym__uppercase_identifier] = ACTIONS(1945),
    [aux_sym_tag_token1] = ACTIONS(1945),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1054)] = {
    [sym_attribute] = STATE(1054),
    [sym__lowercase_identifier] = ACTIONS(1955),
    [anon_sym_TILDE] = ACTIONS(1955),
    [anon_sym_QMARK] = ACTIONS(1955),
    [anon_sym_LPAREN] = ACTIONS(1957),
    [anon_sym_RPAREN] = ACTIONS(1957),
    [anon_sym_COMMA] = ACTIONS(1957),
    [anon_sym_PIPE] = ACTIONS(1955),
    [anon_sym_COLON2] = ACTIONS(1955),
    [anon_sym_DOT] = ACTIONS(1957),
    [anon_sym_DASH_GT] = ACTIONS(1955),
    [anon_sym_LBRACK] = ACTIONS(1955),
    [anon_sym_RBRACK] = ACTIONS(1957),
    [anon_sym_true] = ACTIONS(1955),
    [anon_sym_false] = ACTIONS(1955),
    [anon_sym_LBRACE] = ACTIONS(1955),
    [anon_sym_SEMI] = ACTIONS(1957),
    [anon_sym_RBRACE] = ACTIONS(1957),
    [anon_sym_constraint] = ACTIONS(1955),
    [anon_sym_val] = ACTIONS(1955),
    [anon_sym_end] = ACTIONS(1955),
    [anon_sym_with] = ACTIONS(1955),
    [anon_sym_COLON_GT] = ACTIONS(1957),
    [anon_sym_object] = ACTIONS(1955),
    [anon_sym_inherit] = ACTIONS(1955),
    [anon_sym_method] = ACTIONS(1955),
    [anon_sym_initializer] = ACTIONS(1955),
    [anon_sym_AMP] = ACTIONS(1955),
    [anon_sym_POUND] = ACTIONS(1955),
    [anon_sym_COLON_COLON] = ACTIONS(1957),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1957),
    [anon_sym_then] = ACTIONS(1955),
    [anon_sym_else] = ACTIONS(1955),
    [anon_sym_do] = ACTIONS(1955),
    [anon_sym_done] = ACTIONS(1955),
    [anon_sym_to] = ACTIONS(1955),
    [anon_sym_downto] = ACTIONS(1955),
    [anon_sym_new] = ACTIONS(1955),
    [anon_sym_LBRACE_LT] = ACTIONS(1957),
    [anon_sym_GT_RBRACE] = ACTIONS(1957),
    [anon_sym_begin] = ACTIONS(1955),
    [sym_ocamlyacc_value] = ACTIONS(1957),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1955),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1957),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1955),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1955),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1957),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1957),
    [aux_sym_number_token1] = ACTIONS(1957),
    [aux_sym_signed_number_token1] = ACTIONS(1955),
    [anon_sym_SQUOTE] = ACTIONS(1957),
    [anon_sym_DQUOTE] = ACTIONS(1957),
    [sym_prefix_operator] = ACTIONS(1957),
    [aux_sym_sign_operator_token1] = ACTIONS(1955),
    [sym_hash_operator] = ACTIONS(1957),
    [anon_sym_lsl] = ACTIONS(1955),
    [anon_sym_lsr] = ACTIONS(1955),
    [anon_sym_asr] = ACTIONS(1955),
    [aux_sym_pow_operator_token1] = ACTIONS(1957),
    [anon_sym_mod] = ACTIONS(1955),
    [anon_sym_land] = ACTIONS(1955),
    [anon_sym_lor] = ACTIONS(1955),
    [anon_sym_lxor] = ACTIONS(1955),
    [aux_sym_mult_operator_token1] = ACTIONS(1955),
    [aux_sym_add_operator_token1] = ACTIONS(1955),
    [sym_concat_operator] = ACTIONS(1957),
    [sym_rel_operator] = ACTIONS(1955),
    [anon_sym_AMP_AMP] = ACTIONS(1955),
    [anon_sym_or] = ACTIONS(1955),
    [anon_sym_PIPE_PIPE] = ACTIONS(1955),
    [sym_assign_operator] = ACTIONS(1957),
    [sym__uppercase_identifier] = ACTIONS(1957),
    [aux_sym_tag_token1] = ACTIONS(1957),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1055)] = {
    [sym_attribute] = STATE(1055),
    [sym__lowercase_identifier] = ACTIONS(1959),
    [anon_sym_TILDE] = ACTIONS(1959),
    [anon_sym_QMARK] = ACTIONS(1959),
    [anon_sym_LPAREN] = ACTIONS(1961),
    [anon_sym_RPAREN] = ACTIONS(1961),
    [anon_sym_COMMA] = ACTIONS(1961),
    [anon_sym_PIPE] = ACTIONS(1959),
    [anon_sym_COLON2] = ACTIONS(1959),
    [anon_sym_DOT] = ACTIONS(1961),
    [anon_sym_DASH_GT] = ACTIONS(1959),
    [anon_sym_LBRACK] = ACTIONS(1959),
    [anon_sym_RBRACK] = ACTIONS(1961),
    [anon_sym_true] = ACTIONS(1959),
    [anon_sym_false] = ACTIONS(1959),
    [anon_sym_LBRACE] = ACTIONS(1959),
    [anon_sym_SEMI] = ACTIONS(1961),
    [anon_sym_RBRACE] = ACTIONS(1961),
    [anon_sym_constraint] = ACTIONS(1959),
    [anon_sym_val] = ACTIONS(1959),
    [anon_sym_end] = ACTIONS(1959),
    [anon_sym_with] = ACTIONS(1959),
    [anon_sym_COLON_GT] = ACTIONS(1961),
    [anon_sym_object] = ACTIONS(1959),
    [anon_sym_inherit] = ACTIONS(1959),
    [anon_sym_method] = ACTIONS(1959),
    [anon_sym_initializer] = ACTIONS(1959),
    [anon_sym_AMP] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1959),
    [anon_sym_COLON_COLON] = ACTIONS(1961),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1961),
    [anon_sym_then] = ACTIONS(1959),
    [anon_sym_else] = ACTIONS(1959),
    [anon_sym_do] = ACTIONS(1959),
    [anon_sym_done] = ACTIONS(1959),
    [anon_sym_to] = ACTIONS(1959),
    [anon_sym_downto] = ACTIONS(1959),
    [anon_sym_new] = ACTIONS(1959),
    [anon_sym_LBRACE_LT] = ACTIONS(1961),
    [anon_sym_GT_RBRACE] = ACTIONS(1961),
    [anon_sym_begin] = ACTIONS(1959),
    [sym_ocamlyacc_value] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1959),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1961),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1959),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1959),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1961),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1961),
    [aux_sym_number_token1] = ACTIONS(1961),
    [aux_sym_signed_number_token1] = ACTIONS(1959),
    [anon_sym_SQUOTE] = ACTIONS(1961),
    [anon_sym_DQUOTE] = ACTIONS(1961),
    [sym_prefix_operator] = ACTIONS(1961),
    [aux_sym_sign_operator_token1] = ACTIONS(1959),
    [sym_hash_operator] = ACTIONS(1961),
    [anon_sym_lsl] = ACTIONS(1959),
    [anon_sym_lsr] = ACTIONS(1959),
    [anon_sym_asr] = ACTIONS(1959),
    [aux_sym_pow_operator_token1] = ACTIONS(1961),
    [anon_sym_mod] = ACTIONS(1959),
    [anon_sym_land] = ACTIONS(1959),
    [anon_sym_lor] = ACTIONS(1959),
    [anon_sym_lxor] = ACTIONS(1959),
    [aux_sym_mult_operator_token1] = ACTIONS(1959),
    [aux_sym_add_operator_token1] = ACTIONS(1959),
    [sym_concat_operator] = ACTIONS(1961),
    [sym_rel_operator] = ACTIONS(1959),
    [anon_sym_AMP_AMP] = ACTIONS(1959),
    [anon_sym_or] = ACTIONS(1959),
    [anon_sym_PIPE_PIPE] = ACTIONS(1959),
    [sym_assign_operator] = ACTIONS(1961),
    [sym__uppercase_identifier] = ACTIONS(1961),
    [aux_sym_tag_token1] = ACTIONS(1961),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1056)] = {
    [sym_attribute] = STATE(1056),
    [sym__lowercase_identifier] = ACTIONS(1963),
    [anon_sym_TILDE] = ACTIONS(1963),
    [anon_sym_QMARK] = ACTIONS(1963),
    [anon_sym_LPAREN] = ACTIONS(1965),
    [anon_sym_RPAREN] = ACTIONS(1965),
    [anon_sym_COMMA] = ACTIONS(1965),
    [anon_sym_PIPE] = ACTIONS(1963),
    [anon_sym_COLON2] = ACTIONS(1963),
    [anon_sym_DOT] = ACTIONS(1965),
    [anon_sym_DASH_GT] = ACTIONS(1963),
    [anon_sym_LBRACK] = ACTIONS(1963),
    [anon_sym_RBRACK] = ACTIONS(1965),
    [anon_sym_true] = ACTIONS(1963),
    [anon_sym_false] = ACTIONS(1963),
    [anon_sym_LBRACE] = ACTIONS(1963),
    [anon_sym_SEMI] = ACTIONS(1965),
    [anon_sym_RBRACE] = ACTIONS(1965),
    [anon_sym_constraint] = ACTIONS(1963),
    [anon_sym_val] = ACTIONS(1963),
    [anon_sym_end] = ACTIONS(1963),
    [anon_sym_with] = ACTIONS(1963),
    [anon_sym_COLON_GT] = ACTIONS(1965),
    [anon_sym_object] = ACTIONS(1963),
    [anon_sym_inherit] = ACTIONS(1963),
    [anon_sym_method] = ACTIONS(1963),
    [anon_sym_initializer] = ACTIONS(1963),
    [anon_sym_AMP] = ACTIONS(1963),
    [anon_sym_POUND] = ACTIONS(1963),
    [anon_sym_COLON_COLON] = ACTIONS(1965),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1965),
    [anon_sym_then] = ACTIONS(1963),
    [anon_sym_else] = ACTIONS(1963),
    [anon_sym_do] = ACTIONS(1963),
    [anon_sym_done] = ACTIONS(1963),
    [anon_sym_to] = ACTIONS(1963),
    [anon_sym_downto] = ACTIONS(1963),
    [anon_sym_new] = ACTIONS(1963),
    [anon_sym_LBRACE_LT] = ACTIONS(1965),
    [anon_sym_GT_RBRACE] = ACTIONS(1965),
    [anon_sym_begin] = ACTIONS(1963),
    [sym_ocamlyacc_value] = ACTIONS(1965),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1963),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1965),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1965),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1965),
    [aux_sym_signed_number_token1] = ACTIONS(1963),
    [anon_sym_SQUOTE] = ACTIONS(1965),
    [anon_sym_DQUOTE] = ACTIONS(1965),
    [sym_prefix_operator] = ACTIONS(1965),
    [aux_sym_sign_operator_token1] = ACTIONS(1963),
    [sym_hash_operator] = ACTIONS(1965),
    [anon_sym_lsl] = ACTIONS(1963),
    [anon_sym_lsr] = ACTIONS(1963),
    [anon_sym_asr] = ACTIONS(1963),
    [aux_sym_pow_operator_token1] = ACTIONS(1965),
    [anon_sym_mod] = ACTIONS(1963),
    [anon_sym_land] = ACTIONS(1963),
    [anon_sym_lor] = ACTIONS(1963),
    [anon_sym_lxor] = ACTIONS(1963),
    [aux_sym_mult_operator_token1] = ACTIONS(1963),
    [aux_sym_add_operator_token1] = ACTIONS(1963),
    [sym_concat_operator] = ACTIONS(1965),
    [sym_rel_operator] = ACTIONS(1963),
    [anon_sym_AMP_AMP] = ACTIONS(1963),
    [anon_sym_or] = ACTIONS(1963),
    [anon_sym_PIPE_PIPE] = ACTIONS(1963),
    [sym_assign_operator] = ACTIONS(1965),
    [sym__uppercase_identifier] = ACTIONS(1965),
    [aux_sym_tag_token1] = ACTIONS(1965),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1057)] = {
    [sym_attribute] = STATE(1057),
    [sym__lowercase_identifier] = ACTIONS(1967),
    [anon_sym_TILDE] = ACTIONS(1967),
    [anon_sym_QMARK] = ACTIONS(1967),
    [anon_sym_LPAREN] = ACTIONS(1969),
    [anon_sym_RPAREN] = ACTIONS(1969),
    [anon_sym_COMMA] = ACTIONS(1969),
    [anon_sym_PIPE] = ACTIONS(1967),
    [anon_sym_COLON2] = ACTIONS(1967),
    [anon_sym_DOT] = ACTIONS(1969),
    [anon_sym_DASH_GT] = ACTIONS(1967),
    [anon_sym_LBRACK] = ACTIONS(1967),
    [anon_sym_RBRACK] = ACTIONS(1969),
    [anon_sym_true] = ACTIONS(1967),
    [anon_sym_false] = ACTIONS(1967),
    [anon_sym_LBRACE] = ACTIONS(1967),
    [anon_sym_SEMI] = ACTIONS(1969),
    [anon_sym_RBRACE] = ACTIONS(1969),
    [anon_sym_constraint] = ACTIONS(1967),
    [anon_sym_val] = ACTIONS(1967),
    [anon_sym_end] = ACTIONS(1967),
    [anon_sym_with] = ACTIONS(1967),
    [anon_sym_COLON_GT] = ACTIONS(1969),
    [anon_sym_object] = ACTIONS(1967),
    [anon_sym_inherit] = ACTIONS(1967),
    [anon_sym_method] = ACTIONS(1967),
    [anon_sym_initializer] = ACTIONS(1967),
    [anon_sym_AMP] = ACTIONS(1967),
    [anon_sym_POUND] = ACTIONS(1967),
    [anon_sym_COLON_COLON] = ACTIONS(1969),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1969),
    [anon_sym_then] = ACTIONS(1967),
    [anon_sym_else] = ACTIONS(1967),
    [anon_sym_do] = ACTIONS(1967),
    [anon_sym_done] = ACTIONS(1967),
    [anon_sym_to] = ACTIONS(1967),
    [anon_sym_downto] = ACTIONS(1967),
    [anon_sym_new] = ACTIONS(1967),
    [anon_sym_LBRACE_LT] = ACTIONS(1969),
    [anon_sym_GT_RBRACE] = ACTIONS(1969),
    [anon_sym_begin] = ACTIONS(1967),
    [sym_ocamlyacc_value] = ACTIONS(1969),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1967),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1969),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1969),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1969),
    [aux_sym_number_token1] = ACTIONS(1969),
    [aux_sym_signed_number_token1] = ACTIONS(1967),
    [anon_sym_SQUOTE] = ACTIONS(1969),
    [anon_sym_DQUOTE] = ACTIONS(1969),
    [sym_prefix_operator] = ACTIONS(1969),
    [aux_sym_sign_operator_token1] = ACTIONS(1967),
    [sym_hash_operator] = ACTIONS(1969),
    [anon_sym_lsl] = ACTIONS(1967),
    [anon_sym_lsr] = ACTIONS(1967),
    [anon_sym_asr] = ACTIONS(1967),
    [aux_sym_pow_operator_token1] = ACTIONS(1969),
    [anon_sym_mod] = ACTIONS(1967),
    [anon_sym_land] = ACTIONS(1967),
    [anon_sym_lor] = ACTIONS(1967),
    [anon_sym_lxor] = ACTIONS(1967),
    [aux_sym_mult_operator_token1] = ACTIONS(1967),
    [aux_sym_add_operator_token1] = ACTIONS(1967),
    [sym_concat_operator] = ACTIONS(1969),
    [sym_rel_operator] = ACTIONS(1967),
    [anon_sym_AMP_AMP] = ACTIONS(1967),
    [anon_sym_or] = ACTIONS(1967),
    [anon_sym_PIPE_PIPE] = ACTIONS(1967),
    [sym_assign_operator] = ACTIONS(1969),
    [sym__uppercase_identifier] = ACTIONS(1969),
    [aux_sym_tag_token1] = ACTIONS(1969),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1058)] = {
    [sym_attribute] = STATE(1058),
    [sym__lowercase_identifier] = ACTIONS(1971),
    [anon_sym_TILDE] = ACTIONS(1971),
    [anon_sym_QMARK] = ACTIONS(1971),
    [anon_sym_LPAREN] = ACTIONS(1973),
    [anon_sym_RPAREN] = ACTIONS(1973),
    [anon_sym_COMMA] = ACTIONS(1973),
    [anon_sym_PIPE] = ACTIONS(1971),
    [anon_sym_COLON2] = ACTIONS(1971),
    [anon_sym_DOT] = ACTIONS(993),
    [anon_sym_DASH_GT] = ACTIONS(1971),
    [anon_sym_LBRACK] = ACTIONS(1971),
    [anon_sym_RBRACK] = ACTIONS(1973),
    [anon_sym_true] = ACTIONS(1971),
    [anon_sym_false] = ACTIONS(1971),
    [anon_sym_LBRACE] = ACTIONS(1971),
    [anon_sym_SEMI] = ACTIONS(1973),
    [anon_sym_RBRACE] = ACTIONS(1973),
    [anon_sym_constraint] = ACTIONS(1971),
    [anon_sym_val] = ACTIONS(1971),
    [anon_sym_end] = ACTIONS(1971),
    [anon_sym_with] = ACTIONS(1971),
    [anon_sym_COLON_GT] = ACTIONS(1973),
    [anon_sym_object] = ACTIONS(1971),
    [anon_sym_inherit] = ACTIONS(1971),
    [anon_sym_method] = ACTIONS(1971),
    [anon_sym_initializer] = ACTIONS(1971),
    [anon_sym_AMP] = ACTIONS(1971),
    [anon_sym_POUND] = ACTIONS(1971),
    [anon_sym_COLON_COLON] = ACTIONS(1973),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1973),
    [anon_sym_then] = ACTIONS(1971),
    [anon_sym_else] = ACTIONS(1971),
    [anon_sym_do] = ACTIONS(1971),
    [anon_sym_done] = ACTIONS(1971),
    [anon_sym_to] = ACTIONS(1971),
    [anon_sym_downto] = ACTIONS(1971),
    [anon_sym_new] = ACTIONS(1971),
    [anon_sym_LBRACE_LT] = ACTIONS(1973),
    [anon_sym_GT_RBRACE] = ACTIONS(1973),
    [anon_sym_begin] = ACTIONS(1971),
    [sym_ocamlyacc_value] = ACTIONS(1973),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1971),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1973),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1973),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1973),
    [aux_sym_number_token1] = ACTIONS(1973),
    [aux_sym_signed_number_token1] = ACTIONS(1971),
    [anon_sym_SQUOTE] = ACTIONS(1973),
    [anon_sym_DQUOTE] = ACTIONS(1973),
    [sym_prefix_operator] = ACTIONS(1973),
    [aux_sym_sign_operator_token1] = ACTIONS(1971),
    [sym_hash_operator] = ACTIONS(1973),
    [anon_sym_lsl] = ACTIONS(1971),
    [anon_sym_lsr] = ACTIONS(1971),
    [anon_sym_asr] = ACTIONS(1971),
    [aux_sym_pow_operator_token1] = ACTIONS(1973),
    [anon_sym_mod] = ACTIONS(1971),
    [anon_sym_land] = ACTIONS(1971),
    [anon_sym_lor] = ACTIONS(1971),
    [anon_sym_lxor] = ACTIONS(1971),
    [aux_sym_mult_operator_token1] = ACTIONS(1971),
    [aux_sym_add_operator_token1] = ACTIONS(1971),
    [sym_concat_operator] = ACTIONS(1973),
    [sym_rel_operator] = ACTIONS(1971),
    [anon_sym_AMP_AMP] = ACTIONS(1971),
    [anon_sym_or] = ACTIONS(1971),
    [anon_sym_PIPE_PIPE] = ACTIONS(1971),
    [sym_assign_operator] = ACTIONS(1973),
    [sym__uppercase_identifier] = ACTIONS(1973),
    [aux_sym_tag_token1] = ACTIONS(1973),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1059)] = {
    [sym_attribute] = STATE(1059),
    [sym__lowercase_identifier] = ACTIONS(1975),
    [anon_sym_TILDE] = ACTIONS(1975),
    [anon_sym_QMARK] = ACTIONS(1975),
    [anon_sym_LPAREN] = ACTIONS(1977),
    [anon_sym_RPAREN] = ACTIONS(1977),
    [anon_sym_COMMA] = ACTIONS(1977),
    [anon_sym_PIPE] = ACTIONS(1975),
    [anon_sym_COLON2] = ACTIONS(1975),
    [anon_sym_DOT] = ACTIONS(1977),
    [anon_sym_DASH_GT] = ACTIONS(1975),
    [anon_sym_LBRACK] = ACTIONS(1975),
    [anon_sym_RBRACK] = ACTIONS(1977),
    [anon_sym_true] = ACTIONS(1975),
    [anon_sym_false] = ACTIONS(1975),
    [anon_sym_LBRACE] = ACTIONS(1975),
    [anon_sym_SEMI] = ACTIONS(1977),
    [anon_sym_RBRACE] = ACTIONS(1977),
    [anon_sym_constraint] = ACTIONS(1975),
    [anon_sym_val] = ACTIONS(1975),
    [anon_sym_end] = ACTIONS(1975),
    [anon_sym_with] = ACTIONS(1975),
    [anon_sym_COLON_GT] = ACTIONS(1977),
    [anon_sym_object] = ACTIONS(1975),
    [anon_sym_inherit] = ACTIONS(1975),
    [anon_sym_method] = ACTIONS(1975),
    [anon_sym_initializer] = ACTIONS(1975),
    [anon_sym_AMP] = ACTIONS(1975),
    [anon_sym_POUND] = ACTIONS(1975),
    [anon_sym_COLON_COLON] = ACTIONS(1977),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1977),
    [anon_sym_then] = ACTIONS(1975),
    [anon_sym_else] = ACTIONS(1975),
    [anon_sym_do] = ACTIONS(1975),
    [anon_sym_done] = ACTIONS(1975),
    [anon_sym_to] = ACTIONS(1975),
    [anon_sym_downto] = ACTIONS(1975),
    [anon_sym_new] = ACTIONS(1975),
    [anon_sym_LBRACE_LT] = ACTIONS(1977),
    [anon_sym_GT_RBRACE] = ACTIONS(1977),
    [anon_sym_begin] = ACTIONS(1975),
    [sym_ocamlyacc_value] = ACTIONS(1977),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1975),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1977),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1975),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1975),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1977),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1977),
    [aux_sym_number_token1] = ACTIONS(1977),
    [aux_sym_signed_number_token1] = ACTIONS(1975),
    [anon_sym_SQUOTE] = ACTIONS(1977),
    [anon_sym_DQUOTE] = ACTIONS(1977),
    [sym_prefix_operator] = ACTIONS(1977),
    [aux_sym_sign_operator_token1] = ACTIONS(1975),
    [sym_hash_operator] = ACTIONS(1977),
    [anon_sym_lsl] = ACTIONS(1975),
    [anon_sym_lsr] = ACTIONS(1975),
    [anon_sym_asr] = ACTIONS(1975),
    [aux_sym_pow_operator_token1] = ACTIONS(1977),
    [anon_sym_mod] = ACTIONS(1975),
    [anon_sym_land] = ACTIONS(1975),
    [anon_sym_lor] = ACTIONS(1975),
    [anon_sym_lxor] = ACTIONS(1975),
    [aux_sym_mult_operator_token1] = ACTIONS(1975),
    [aux_sym_add_operator_token1] = ACTIONS(1975),
    [sym_concat_operator] = ACTIONS(1977),
    [sym_rel_operator] = ACTIONS(1975),
    [anon_sym_AMP_AMP] = ACTIONS(1975),
    [anon_sym_or] = ACTIONS(1975),
    [anon_sym_PIPE_PIPE] = ACTIONS(1975),
    [sym_assign_operator] = ACTIONS(1977),
    [sym__uppercase_identifier] = ACTIONS(1977),
    [aux_sym_tag_token1] = ACTIONS(1977),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1060)] = {
    [sym_attribute] = STATE(1060),
    [sym__lowercase_identifier] = ACTIONS(1983),
    [anon_sym_TILDE] = ACTIONS(1983),
    [anon_sym_QMARK] = ACTIONS(1983),
    [anon_sym_LPAREN] = ACTIONS(1985),
    [anon_sym_RPAREN] = ACTIONS(1985),
    [anon_sym_COMMA] = ACTIONS(1985),
    [anon_sym_PIPE] = ACTIONS(1983),
    [anon_sym_COLON2] = ACTIONS(1983),
    [anon_sym_DOT] = ACTIONS(1985),
    [anon_sym_DASH_GT] = ACTIONS(1983),
    [anon_sym_LBRACK] = ACTIONS(1983),
    [anon_sym_RBRACK] = ACTIONS(1985),
    [anon_sym_true] = ACTIONS(1983),
    [anon_sym_false] = ACTIONS(1983),
    [anon_sym_LBRACE] = ACTIONS(1983),
    [anon_sym_SEMI] = ACTIONS(1985),
    [anon_sym_RBRACE] = ACTIONS(1985),
    [anon_sym_constraint] = ACTIONS(1983),
    [anon_sym_val] = ACTIONS(1983),
    [anon_sym_end] = ACTIONS(1983),
    [anon_sym_with] = ACTIONS(1983),
    [anon_sym_COLON_GT] = ACTIONS(1985),
    [anon_sym_object] = ACTIONS(1983),
    [anon_sym_inherit] = ACTIONS(1983),
    [anon_sym_method] = ACTIONS(1983),
    [anon_sym_initializer] = ACTIONS(1983),
    [anon_sym_AMP] = ACTIONS(1983),
    [anon_sym_POUND] = ACTIONS(1983),
    [anon_sym_COLON_COLON] = ACTIONS(1985),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1985),
    [anon_sym_then] = ACTIONS(1983),
    [anon_sym_else] = ACTIONS(1983),
    [anon_sym_do] = ACTIONS(1983),
    [anon_sym_done] = ACTIONS(1983),
    [anon_sym_to] = ACTIONS(1983),
    [anon_sym_downto] = ACTIONS(1983),
    [anon_sym_new] = ACTIONS(1983),
    [anon_sym_LBRACE_LT] = ACTIONS(1985),
    [anon_sym_GT_RBRACE] = ACTIONS(1985),
    [anon_sym_begin] = ACTIONS(1983),
    [sym_ocamlyacc_value] = ACTIONS(1985),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1983),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1985),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1983),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1983),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1985),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1985),
    [aux_sym_number_token1] = ACTIONS(1985),
    [aux_sym_signed_number_token1] = ACTIONS(1983),
    [anon_sym_SQUOTE] = ACTIONS(1985),
    [anon_sym_DQUOTE] = ACTIONS(1985),
    [sym_prefix_operator] = ACTIONS(1985),
    [aux_sym_sign_operator_token1] = ACTIONS(1983),
    [sym_hash_operator] = ACTIONS(1985),
    [anon_sym_lsl] = ACTIONS(1983),
    [anon_sym_lsr] = ACTIONS(1983),
    [anon_sym_asr] = ACTIONS(1983),
    [aux_sym_pow_operator_token1] = ACTIONS(1985),
    [anon_sym_mod] = ACTIONS(1983),
    [anon_sym_land] = ACTIONS(1983),
    [anon_sym_lor] = ACTIONS(1983),
    [anon_sym_lxor] = ACTIONS(1983),
    [aux_sym_mult_operator_token1] = ACTIONS(1983),
    [aux_sym_add_operator_token1] = ACTIONS(1983),
    [sym_concat_operator] = ACTIONS(1985),
    [sym_rel_operator] = ACTIONS(1983),
    [anon_sym_AMP_AMP] = ACTIONS(1983),
    [anon_sym_or] = ACTIONS(1983),
    [anon_sym_PIPE_PIPE] = ACTIONS(1983),
    [sym_assign_operator] = ACTIONS(1985),
    [sym__uppercase_identifier] = ACTIONS(1985),
    [aux_sym_tag_token1] = ACTIONS(1985),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1061)] = {
    [sym_attribute] = STATE(1061),
    [sym__lowercase_identifier] = ACTIONS(1987),
    [anon_sym_TILDE] = ACTIONS(1987),
    [anon_sym_QMARK] = ACTIONS(1987),
    [anon_sym_LPAREN] = ACTIONS(1989),
    [anon_sym_RPAREN] = ACTIONS(1989),
    [anon_sym_COMMA] = ACTIONS(1989),
    [anon_sym_PIPE] = ACTIONS(1987),
    [anon_sym_COLON2] = ACTIONS(1987),
    [anon_sym_DOT] = ACTIONS(1989),
    [anon_sym_DASH_GT] = ACTIONS(1987),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_RBRACK] = ACTIONS(1989),
    [anon_sym_true] = ACTIONS(1987),
    [anon_sym_false] = ACTIONS(1987),
    [anon_sym_LBRACE] = ACTIONS(1987),
    [anon_sym_SEMI] = ACTIONS(1989),
    [anon_sym_RBRACE] = ACTIONS(1989),
    [anon_sym_constraint] = ACTIONS(1987),
    [anon_sym_val] = ACTIONS(1987),
    [anon_sym_end] = ACTIONS(1987),
    [anon_sym_with] = ACTIONS(1987),
    [anon_sym_COLON_GT] = ACTIONS(1989),
    [anon_sym_object] = ACTIONS(1987),
    [anon_sym_inherit] = ACTIONS(1987),
    [anon_sym_method] = ACTIONS(1987),
    [anon_sym_initializer] = ACTIONS(1987),
    [anon_sym_AMP] = ACTIONS(1987),
    [anon_sym_POUND] = ACTIONS(1987),
    [anon_sym_COLON_COLON] = ACTIONS(1989),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1989),
    [anon_sym_then] = ACTIONS(1987),
    [anon_sym_else] = ACTIONS(1987),
    [anon_sym_do] = ACTIONS(1987),
    [anon_sym_done] = ACTIONS(1987),
    [anon_sym_to] = ACTIONS(1987),
    [anon_sym_downto] = ACTIONS(1987),
    [anon_sym_new] = ACTIONS(1987),
    [anon_sym_LBRACE_LT] = ACTIONS(1989),
    [anon_sym_GT_RBRACE] = ACTIONS(1989),
    [anon_sym_begin] = ACTIONS(1987),
    [sym_ocamlyacc_value] = ACTIONS(1989),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1987),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1989),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1987),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1987),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1989),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1989),
    [aux_sym_number_token1] = ACTIONS(1989),
    [aux_sym_signed_number_token1] = ACTIONS(1987),
    [anon_sym_SQUOTE] = ACTIONS(1989),
    [anon_sym_DQUOTE] = ACTIONS(1989),
    [sym_prefix_operator] = ACTIONS(1989),
    [aux_sym_sign_operator_token1] = ACTIONS(1987),
    [sym_hash_operator] = ACTIONS(1989),
    [anon_sym_lsl] = ACTIONS(1987),
    [anon_sym_lsr] = ACTIONS(1987),
    [anon_sym_asr] = ACTIONS(1987),
    [aux_sym_pow_operator_token1] = ACTIONS(1989),
    [anon_sym_mod] = ACTIONS(1987),
    [anon_sym_land] = ACTIONS(1987),
    [anon_sym_lor] = ACTIONS(1987),
    [anon_sym_lxor] = ACTIONS(1987),
    [aux_sym_mult_operator_token1] = ACTIONS(1987),
    [aux_sym_add_operator_token1] = ACTIONS(1987),
    [sym_concat_operator] = ACTIONS(1989),
    [sym_rel_operator] = ACTIONS(1987),
    [anon_sym_AMP_AMP] = ACTIONS(1987),
    [anon_sym_or] = ACTIONS(1987),
    [anon_sym_PIPE_PIPE] = ACTIONS(1987),
    [sym_assign_operator] = ACTIONS(1989),
    [sym__uppercase_identifier] = ACTIONS(1989),
    [aux_sym_tag_token1] = ACTIONS(1989),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1062)] = {
    [sym_attribute] = STATE(1062),
    [sym__lowercase_identifier] = ACTIONS(1991),
    [anon_sym_TILDE] = ACTIONS(1991),
    [anon_sym_QMARK] = ACTIONS(1991),
    [anon_sym_LPAREN] = ACTIONS(1993),
    [anon_sym_RPAREN] = ACTIONS(1993),
    [anon_sym_COMMA] = ACTIONS(1993),
    [anon_sym_PIPE] = ACTIONS(1991),
    [anon_sym_COLON2] = ACTIONS(1991),
    [anon_sym_DOT] = ACTIONS(1993),
    [anon_sym_DASH_GT] = ACTIONS(1991),
    [anon_sym_LBRACK] = ACTIONS(1991),
    [anon_sym_RBRACK] = ACTIONS(1993),
    [anon_sym_true] = ACTIONS(1991),
    [anon_sym_false] = ACTIONS(1991),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_SEMI] = ACTIONS(1993),
    [anon_sym_RBRACE] = ACTIONS(1993),
    [anon_sym_constraint] = ACTIONS(1991),
    [anon_sym_val] = ACTIONS(1991),
    [anon_sym_end] = ACTIONS(1991),
    [anon_sym_with] = ACTIONS(1991),
    [anon_sym_COLON_GT] = ACTIONS(1993),
    [anon_sym_object] = ACTIONS(1991),
    [anon_sym_inherit] = ACTIONS(1991),
    [anon_sym_method] = ACTIONS(1991),
    [anon_sym_initializer] = ACTIONS(1991),
    [anon_sym_AMP] = ACTIONS(1991),
    [anon_sym_POUND] = ACTIONS(1991),
    [anon_sym_COLON_COLON] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1993),
    [anon_sym_then] = ACTIONS(1991),
    [anon_sym_else] = ACTIONS(1991),
    [anon_sym_do] = ACTIONS(1991),
    [anon_sym_done] = ACTIONS(1991),
    [anon_sym_to] = ACTIONS(1991),
    [anon_sym_downto] = ACTIONS(1991),
    [anon_sym_new] = ACTIONS(1991),
    [anon_sym_LBRACE_LT] = ACTIONS(1993),
    [anon_sym_GT_RBRACE] = ACTIONS(1993),
    [anon_sym_begin] = ACTIONS(1991),
    [sym_ocamlyacc_value] = ACTIONS(1993),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1991),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1993),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1991),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1991),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1993),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1993),
    [aux_sym_number_token1] = ACTIONS(1993),
    [aux_sym_signed_number_token1] = ACTIONS(1991),
    [anon_sym_SQUOTE] = ACTIONS(1993),
    [anon_sym_DQUOTE] = ACTIONS(1993),
    [sym_prefix_operator] = ACTIONS(1993),
    [aux_sym_sign_operator_token1] = ACTIONS(1991),
    [sym_hash_operator] = ACTIONS(1993),
    [anon_sym_lsl] = ACTIONS(1991),
    [anon_sym_lsr] = ACTIONS(1991),
    [anon_sym_asr] = ACTIONS(1991),
    [aux_sym_pow_operator_token1] = ACTIONS(1993),
    [anon_sym_mod] = ACTIONS(1991),
    [anon_sym_land] = ACTIONS(1991),
    [anon_sym_lor] = ACTIONS(1991),
    [anon_sym_lxor] = ACTIONS(1991),
    [aux_sym_mult_operator_token1] = ACTIONS(1991),
    [aux_sym_add_operator_token1] = ACTIONS(1991),
    [sym_concat_operator] = ACTIONS(1993),
    [sym_rel_operator] = ACTIONS(1991),
    [anon_sym_AMP_AMP] = ACTIONS(1991),
    [anon_sym_or] = ACTIONS(1991),
    [anon_sym_PIPE_PIPE] = ACTIONS(1991),
    [sym_assign_operator] = ACTIONS(1993),
    [sym__uppercase_identifier] = ACTIONS(1993),
    [aux_sym_tag_token1] = ACTIONS(1993),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1063)] = {
    [sym_attribute] = STATE(1063),
    [sym__lowercase_identifier] = ACTIONS(2007),
    [anon_sym_TILDE] = ACTIONS(2007),
    [anon_sym_QMARK] = ACTIONS(2007),
    [anon_sym_LPAREN] = ACTIONS(2009),
    [anon_sym_RPAREN] = ACTIONS(2009),
    [anon_sym_COMMA] = ACTIONS(2009),
    [anon_sym_PIPE] = ACTIONS(2007),
    [anon_sym_COLON2] = ACTIONS(2007),
    [anon_sym_DOT] = ACTIONS(2009),
    [anon_sym_DASH_GT] = ACTIONS(2007),
    [anon_sym_LBRACK] = ACTIONS(2007),
    [anon_sym_RBRACK] = ACTIONS(2009),
    [anon_sym_true] = ACTIONS(2007),
    [anon_sym_false] = ACTIONS(2007),
    [anon_sym_LBRACE] = ACTIONS(2007),
    [anon_sym_SEMI] = ACTIONS(2009),
    [anon_sym_RBRACE] = ACTIONS(2009),
    [anon_sym_constraint] = ACTIONS(2007),
    [anon_sym_val] = ACTIONS(2007),
    [anon_sym_end] = ACTIONS(2007),
    [anon_sym_with] = ACTIONS(2007),
    [anon_sym_COLON_GT] = ACTIONS(2009),
    [anon_sym_object] = ACTIONS(2007),
    [anon_sym_inherit] = ACTIONS(2007),
    [anon_sym_method] = ACTIONS(2007),
    [anon_sym_initializer] = ACTIONS(2007),
    [anon_sym_AMP] = ACTIONS(2007),
    [anon_sym_POUND] = ACTIONS(2007),
    [anon_sym_COLON_COLON] = ACTIONS(2009),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2009),
    [anon_sym_then] = ACTIONS(2007),
    [anon_sym_else] = ACTIONS(2007),
    [anon_sym_do] = ACTIONS(2007),
    [anon_sym_done] = ACTIONS(2007),
    [anon_sym_to] = ACTIONS(2007),
    [anon_sym_downto] = ACTIONS(2007),
    [anon_sym_new] = ACTIONS(2007),
    [anon_sym_LBRACE_LT] = ACTIONS(2009),
    [anon_sym_GT_RBRACE] = ACTIONS(2009),
    [anon_sym_begin] = ACTIONS(2007),
    [sym_ocamlyacc_value] = ACTIONS(2009),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2007),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2009),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2007),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2007),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2009),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2009),
    [aux_sym_number_token1] = ACTIONS(2009),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2009),
    [sym_prefix_operator] = ACTIONS(2009),
    [aux_sym_sign_operator_token1] = ACTIONS(2007),
    [sym_hash_operator] = ACTIONS(2009),
    [anon_sym_lsl] = ACTIONS(2007),
    [anon_sym_lsr] = ACTIONS(2007),
    [anon_sym_asr] = ACTIONS(2007),
    [aux_sym_pow_operator_token1] = ACTIONS(2009),
    [anon_sym_mod] = ACTIONS(2007),
    [anon_sym_land] = ACTIONS(2007),
    [anon_sym_lor] = ACTIONS(2007),
    [anon_sym_lxor] = ACTIONS(2007),
    [aux_sym_mult_operator_token1] = ACTIONS(2007),
    [aux_sym_add_operator_token1] = ACTIONS(2007),
    [sym_concat_operator] = ACTIONS(2009),
    [sym_rel_operator] = ACTIONS(2007),
    [anon_sym_AMP_AMP] = ACTIONS(2007),
    [anon_sym_or] = ACTIONS(2007),
    [anon_sym_PIPE_PIPE] = ACTIONS(2007),
    [sym_assign_operator] = ACTIONS(2009),
    [sym__uppercase_identifier] = ACTIONS(2009),
    [aux_sym_tag_token1] = ACTIONS(2009),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1064)] = {
    [sym_attribute] = STATE(1064),
    [sym__lowercase_identifier] = ACTIONS(2011),
    [anon_sym_TILDE] = ACTIONS(2011),
    [anon_sym_QMARK] = ACTIONS(2011),
    [anon_sym_LPAREN] = ACTIONS(2013),
    [anon_sym_RPAREN] = ACTIONS(2013),
    [anon_sym_COMMA] = ACTIONS(2013),
    [anon_sym_PIPE] = ACTIONS(2011),
    [anon_sym_COLON2] = ACTIONS(2011),
    [anon_sym_DOT] = ACTIONS(2013),
    [anon_sym_DASH_GT] = ACTIONS(2011),
    [anon_sym_LBRACK] = ACTIONS(2011),
    [anon_sym_RBRACK] = ACTIONS(2013),
    [anon_sym_true] = ACTIONS(2011),
    [anon_sym_false] = ACTIONS(2011),
    [anon_sym_LBRACE] = ACTIONS(2011),
    [anon_sym_SEMI] = ACTIONS(2013),
    [anon_sym_RBRACE] = ACTIONS(2013),
    [anon_sym_constraint] = ACTIONS(2011),
    [anon_sym_val] = ACTIONS(2011),
    [anon_sym_end] = ACTIONS(2011),
    [anon_sym_with] = ACTIONS(2011),
    [anon_sym_COLON_GT] = ACTIONS(2013),
    [anon_sym_object] = ACTIONS(2011),
    [anon_sym_inherit] = ACTIONS(2011),
    [anon_sym_method] = ACTIONS(2011),
    [anon_sym_initializer] = ACTIONS(2011),
    [anon_sym_AMP] = ACTIONS(2011),
    [anon_sym_POUND] = ACTIONS(2011),
    [anon_sym_COLON_COLON] = ACTIONS(2013),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2013),
    [anon_sym_then] = ACTIONS(2011),
    [anon_sym_else] = ACTIONS(2011),
    [anon_sym_do] = ACTIONS(2011),
    [anon_sym_done] = ACTIONS(2011),
    [anon_sym_to] = ACTIONS(2011),
    [anon_sym_downto] = ACTIONS(2011),
    [anon_sym_new] = ACTIONS(2011),
    [anon_sym_LBRACE_LT] = ACTIONS(2013),
    [anon_sym_GT_RBRACE] = ACTIONS(2013),
    [anon_sym_begin] = ACTIONS(2011),
    [sym_ocamlyacc_value] = ACTIONS(2013),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2011),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2013),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2011),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2011),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2013),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2013),
    [aux_sym_number_token1] = ACTIONS(2013),
    [aux_sym_signed_number_token1] = ACTIONS(2011),
    [anon_sym_SQUOTE] = ACTIONS(2013),
    [anon_sym_DQUOTE] = ACTIONS(2013),
    [sym_prefix_operator] = ACTIONS(2013),
    [aux_sym_sign_operator_token1] = ACTIONS(2011),
    [sym_hash_operator] = ACTIONS(2013),
    [anon_sym_lsl] = ACTIONS(2011),
    [anon_sym_lsr] = ACTIONS(2011),
    [anon_sym_asr] = ACTIONS(2011),
    [aux_sym_pow_operator_token1] = ACTIONS(2013),
    [anon_sym_mod] = ACTIONS(2011),
    [anon_sym_land] = ACTIONS(2011),
    [anon_sym_lor] = ACTIONS(2011),
    [anon_sym_lxor] = ACTIONS(2011),
    [aux_sym_mult_operator_token1] = ACTIONS(2011),
    [aux_sym_add_operator_token1] = ACTIONS(2011),
    [sym_concat_operator] = ACTIONS(2013),
    [sym_rel_operator] = ACTIONS(2011),
    [anon_sym_AMP_AMP] = ACTIONS(2011),
    [anon_sym_or] = ACTIONS(2011),
    [anon_sym_PIPE_PIPE] = ACTIONS(2011),
    [sym_assign_operator] = ACTIONS(2013),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_tag_token1] = ACTIONS(2013),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1065)] = {
    [sym_attribute] = STATE(1065),
    [sym__lowercase_identifier] = ACTIONS(2015),
    [anon_sym_TILDE] = ACTIONS(2015),
    [anon_sym_QMARK] = ACTIONS(2015),
    [anon_sym_LPAREN] = ACTIONS(2017),
    [anon_sym_RPAREN] = ACTIONS(2017),
    [anon_sym_COMMA] = ACTIONS(2017),
    [anon_sym_PIPE] = ACTIONS(2015),
    [anon_sym_COLON2] = ACTIONS(2015),
    [anon_sym_DOT] = ACTIONS(2017),
    [anon_sym_DASH_GT] = ACTIONS(2015),
    [anon_sym_LBRACK] = ACTIONS(2015),
    [anon_sym_RBRACK] = ACTIONS(2017),
    [anon_sym_true] = ACTIONS(2015),
    [anon_sym_false] = ACTIONS(2015),
    [anon_sym_LBRACE] = ACTIONS(2015),
    [anon_sym_SEMI] = ACTIONS(2017),
    [anon_sym_RBRACE] = ACTIONS(2017),
    [anon_sym_constraint] = ACTIONS(2015),
    [anon_sym_val] = ACTIONS(2015),
    [anon_sym_end] = ACTIONS(2015),
    [anon_sym_with] = ACTIONS(2015),
    [anon_sym_COLON_GT] = ACTIONS(2017),
    [anon_sym_object] = ACTIONS(2015),
    [anon_sym_inherit] = ACTIONS(2015),
    [anon_sym_method] = ACTIONS(2015),
    [anon_sym_initializer] = ACTIONS(2015),
    [anon_sym_AMP] = ACTIONS(2015),
    [anon_sym_POUND] = ACTIONS(2015),
    [anon_sym_COLON_COLON] = ACTIONS(2017),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2017),
    [anon_sym_then] = ACTIONS(2015),
    [anon_sym_else] = ACTIONS(2015),
    [anon_sym_do] = ACTIONS(2015),
    [anon_sym_done] = ACTIONS(2015),
    [anon_sym_to] = ACTIONS(2015),
    [anon_sym_downto] = ACTIONS(2015),
    [anon_sym_new] = ACTIONS(2015),
    [anon_sym_LBRACE_LT] = ACTIONS(2017),
    [anon_sym_GT_RBRACE] = ACTIONS(2017),
    [anon_sym_begin] = ACTIONS(2015),
    [sym_ocamlyacc_value] = ACTIONS(2017),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2015),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2017),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2015),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2015),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2017),
    [aux_sym_number_token1] = ACTIONS(2017),
    [aux_sym_signed_number_token1] = ACTIONS(2015),
    [anon_sym_SQUOTE] = ACTIONS(2017),
    [anon_sym_DQUOTE] = ACTIONS(2017),
    [sym_prefix_operator] = ACTIONS(2017),
    [aux_sym_sign_operator_token1] = ACTIONS(2015),
    [sym_hash_operator] = ACTIONS(2017),
    [anon_sym_lsl] = ACTIONS(2015),
    [anon_sym_lsr] = ACTIONS(2015),
    [anon_sym_asr] = ACTIONS(2015),
    [aux_sym_pow_operator_token1] = ACTIONS(2017),
    [anon_sym_mod] = ACTIONS(2015),
    [anon_sym_land] = ACTIONS(2015),
    [anon_sym_lor] = ACTIONS(2015),
    [anon_sym_lxor] = ACTIONS(2015),
    [aux_sym_mult_operator_token1] = ACTIONS(2015),
    [aux_sym_add_operator_token1] = ACTIONS(2015),
    [sym_concat_operator] = ACTIONS(2017),
    [sym_rel_operator] = ACTIONS(2015),
    [anon_sym_AMP_AMP] = ACTIONS(2015),
    [anon_sym_or] = ACTIONS(2015),
    [anon_sym_PIPE_PIPE] = ACTIONS(2015),
    [sym_assign_operator] = ACTIONS(2017),
    [sym__uppercase_identifier] = ACTIONS(2017),
    [aux_sym_tag_token1] = ACTIONS(2017),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1066)] = {
    [sym_attribute] = STATE(1066),
    [sym__lowercase_identifier] = ACTIONS(2019),
    [anon_sym_TILDE] = ACTIONS(2019),
    [anon_sym_QMARK] = ACTIONS(2019),
    [anon_sym_LPAREN] = ACTIONS(2021),
    [anon_sym_RPAREN] = ACTIONS(2021),
    [anon_sym_COMMA] = ACTIONS(2021),
    [anon_sym_PIPE] = ACTIONS(2019),
    [anon_sym_COLON2] = ACTIONS(2019),
    [anon_sym_DOT] = ACTIONS(2021),
    [anon_sym_DASH_GT] = ACTIONS(2019),
    [anon_sym_LBRACK] = ACTIONS(2019),
    [anon_sym_RBRACK] = ACTIONS(2021),
    [anon_sym_true] = ACTIONS(2019),
    [anon_sym_false] = ACTIONS(2019),
    [anon_sym_LBRACE] = ACTIONS(2019),
    [anon_sym_SEMI] = ACTIONS(2021),
    [anon_sym_RBRACE] = ACTIONS(2021),
    [anon_sym_constraint] = ACTIONS(2019),
    [anon_sym_val] = ACTIONS(2019),
    [anon_sym_end] = ACTIONS(2019),
    [anon_sym_with] = ACTIONS(2019),
    [anon_sym_COLON_GT] = ACTIONS(2021),
    [anon_sym_object] = ACTIONS(2019),
    [anon_sym_inherit] = ACTIONS(2019),
    [anon_sym_method] = ACTIONS(2019),
    [anon_sym_initializer] = ACTIONS(2019),
    [anon_sym_AMP] = ACTIONS(2019),
    [anon_sym_POUND] = ACTIONS(2019),
    [anon_sym_COLON_COLON] = ACTIONS(2021),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2021),
    [anon_sym_then] = ACTIONS(2019),
    [anon_sym_else] = ACTIONS(2019),
    [anon_sym_do] = ACTIONS(2019),
    [anon_sym_done] = ACTIONS(2019),
    [anon_sym_to] = ACTIONS(2019),
    [anon_sym_downto] = ACTIONS(2019),
    [anon_sym_new] = ACTIONS(2019),
    [anon_sym_LBRACE_LT] = ACTIONS(2021),
    [anon_sym_GT_RBRACE] = ACTIONS(2021),
    [anon_sym_begin] = ACTIONS(2019),
    [sym_ocamlyacc_value] = ACTIONS(2021),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2019),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2021),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2021),
    [aux_sym_number_token1] = ACTIONS(2021),
    [aux_sym_signed_number_token1] = ACTIONS(2019),
    [anon_sym_SQUOTE] = ACTIONS(2021),
    [anon_sym_DQUOTE] = ACTIONS(2021),
    [sym_prefix_operator] = ACTIONS(2021),
    [aux_sym_sign_operator_token1] = ACTIONS(2019),
    [sym_hash_operator] = ACTIONS(2021),
    [anon_sym_lsl] = ACTIONS(2019),
    [anon_sym_lsr] = ACTIONS(2019),
    [anon_sym_asr] = ACTIONS(2019),
    [aux_sym_pow_operator_token1] = ACTIONS(2021),
    [anon_sym_mod] = ACTIONS(2019),
    [anon_sym_land] = ACTIONS(2019),
    [anon_sym_lor] = ACTIONS(2019),
    [anon_sym_lxor] = ACTIONS(2019),
    [aux_sym_mult_operator_token1] = ACTIONS(2019),
    [aux_sym_add_operator_token1] = ACTIONS(2019),
    [sym_concat_operator] = ACTIONS(2021),
    [sym_rel_operator] = ACTIONS(2019),
    [anon_sym_AMP_AMP] = ACTIONS(2019),
    [anon_sym_or] = ACTIONS(2019),
    [anon_sym_PIPE_PIPE] = ACTIONS(2019),
    [sym_assign_operator] = ACTIONS(2021),
    [sym__uppercase_identifier] = ACTIONS(2021),
    [aux_sym_tag_token1] = ACTIONS(2021),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1067)] = {
    [sym_attribute] = STATE(1067),
    [sym__lowercase_identifier] = ACTIONS(2023),
    [anon_sym_TILDE] = ACTIONS(2023),
    [anon_sym_QMARK] = ACTIONS(2023),
    [anon_sym_LPAREN] = ACTIONS(2025),
    [anon_sym_RPAREN] = ACTIONS(2025),
    [anon_sym_COMMA] = ACTIONS(2025),
    [anon_sym_PIPE] = ACTIONS(2023),
    [anon_sym_COLON2] = ACTIONS(2023),
    [anon_sym_DOT] = ACTIONS(2025),
    [anon_sym_DASH_GT] = ACTIONS(2023),
    [anon_sym_LBRACK] = ACTIONS(2023),
    [anon_sym_RBRACK] = ACTIONS(2025),
    [anon_sym_true] = ACTIONS(2023),
    [anon_sym_false] = ACTIONS(2023),
    [anon_sym_LBRACE] = ACTIONS(2023),
    [anon_sym_SEMI] = ACTIONS(2025),
    [anon_sym_RBRACE] = ACTIONS(2025),
    [anon_sym_constraint] = ACTIONS(2023),
    [anon_sym_val] = ACTIONS(2023),
    [anon_sym_end] = ACTIONS(2023),
    [anon_sym_with] = ACTIONS(2023),
    [anon_sym_COLON_GT] = ACTIONS(2025),
    [anon_sym_object] = ACTIONS(2023),
    [anon_sym_inherit] = ACTIONS(2023),
    [anon_sym_method] = ACTIONS(2023),
    [anon_sym_initializer] = ACTIONS(2023),
    [anon_sym_AMP] = ACTIONS(2023),
    [anon_sym_POUND] = ACTIONS(2023),
    [anon_sym_COLON_COLON] = ACTIONS(2025),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2025),
    [anon_sym_then] = ACTIONS(2023),
    [anon_sym_else] = ACTIONS(2023),
    [anon_sym_do] = ACTIONS(2023),
    [anon_sym_done] = ACTIONS(2023),
    [anon_sym_to] = ACTIONS(2023),
    [anon_sym_downto] = ACTIONS(2023),
    [anon_sym_new] = ACTIONS(2023),
    [anon_sym_LBRACE_LT] = ACTIONS(2025),
    [anon_sym_GT_RBRACE] = ACTIONS(2025),
    [anon_sym_begin] = ACTIONS(2023),
    [sym_ocamlyacc_value] = ACTIONS(2025),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2023),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2025),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2025),
    [aux_sym_number_token1] = ACTIONS(2025),
    [aux_sym_signed_number_token1] = ACTIONS(2023),
    [anon_sym_SQUOTE] = ACTIONS(2025),
    [anon_sym_DQUOTE] = ACTIONS(2025),
    [sym_prefix_operator] = ACTIONS(2025),
    [aux_sym_sign_operator_token1] = ACTIONS(2023),
    [sym_hash_operator] = ACTIONS(2025),
    [anon_sym_lsl] = ACTIONS(2023),
    [anon_sym_lsr] = ACTIONS(2023),
    [anon_sym_asr] = ACTIONS(2023),
    [aux_sym_pow_operator_token1] = ACTIONS(2025),
    [anon_sym_mod] = ACTIONS(2023),
    [anon_sym_land] = ACTIONS(2023),
    [anon_sym_lor] = ACTIONS(2023),
    [anon_sym_lxor] = ACTIONS(2023),
    [aux_sym_mult_operator_token1] = ACTIONS(2023),
    [aux_sym_add_operator_token1] = ACTIONS(2023),
    [sym_concat_operator] = ACTIONS(2025),
    [sym_rel_operator] = ACTIONS(2023),
    [anon_sym_AMP_AMP] = ACTIONS(2023),
    [anon_sym_or] = ACTIONS(2023),
    [anon_sym_PIPE_PIPE] = ACTIONS(2023),
    [sym_assign_operator] = ACTIONS(2025),
    [sym__uppercase_identifier] = ACTIONS(2025),
    [aux_sym_tag_token1] = ACTIONS(2025),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1068)] = {
    [sym_attribute] = STATE(1068),
    [sym__lowercase_identifier] = ACTIONS(2027),
    [anon_sym_TILDE] = ACTIONS(2027),
    [anon_sym_QMARK] = ACTIONS(2027),
    [anon_sym_LPAREN] = ACTIONS(2029),
    [anon_sym_RPAREN] = ACTIONS(2029),
    [anon_sym_COMMA] = ACTIONS(2029),
    [anon_sym_PIPE] = ACTIONS(2027),
    [anon_sym_COLON2] = ACTIONS(2027),
    [anon_sym_DOT] = ACTIONS(2029),
    [anon_sym_DASH_GT] = ACTIONS(2027),
    [anon_sym_LBRACK] = ACTIONS(2027),
    [anon_sym_RBRACK] = ACTIONS(2029),
    [anon_sym_true] = ACTIONS(2027),
    [anon_sym_false] = ACTIONS(2027),
    [anon_sym_LBRACE] = ACTIONS(2027),
    [anon_sym_SEMI] = ACTIONS(2029),
    [anon_sym_RBRACE] = ACTIONS(2029),
    [anon_sym_constraint] = ACTIONS(2027),
    [anon_sym_val] = ACTIONS(2027),
    [anon_sym_end] = ACTIONS(2027),
    [anon_sym_with] = ACTIONS(2027),
    [anon_sym_COLON_GT] = ACTIONS(2029),
    [anon_sym_object] = ACTIONS(2027),
    [anon_sym_inherit] = ACTIONS(2027),
    [anon_sym_method] = ACTIONS(2027),
    [anon_sym_initializer] = ACTIONS(2027),
    [anon_sym_AMP] = ACTIONS(2027),
    [anon_sym_POUND] = ACTIONS(2027),
    [anon_sym_COLON_COLON] = ACTIONS(2029),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2029),
    [anon_sym_then] = ACTIONS(2027),
    [anon_sym_else] = ACTIONS(2027),
    [anon_sym_do] = ACTIONS(2027),
    [anon_sym_done] = ACTIONS(2027),
    [anon_sym_to] = ACTIONS(2027),
    [anon_sym_downto] = ACTIONS(2027),
    [anon_sym_new] = ACTIONS(2027),
    [anon_sym_LBRACE_LT] = ACTIONS(2029),
    [anon_sym_GT_RBRACE] = ACTIONS(2029),
    [anon_sym_begin] = ACTIONS(2027),
    [sym_ocamlyacc_value] = ACTIONS(2029),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2027),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2029),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2029),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2029),
    [aux_sym_number_token1] = ACTIONS(2029),
    [aux_sym_signed_number_token1] = ACTIONS(2027),
    [anon_sym_SQUOTE] = ACTIONS(2029),
    [anon_sym_DQUOTE] = ACTIONS(2029),
    [sym_prefix_operator] = ACTIONS(2029),
    [aux_sym_sign_operator_token1] = ACTIONS(2027),
    [sym_hash_operator] = ACTIONS(2029),
    [anon_sym_lsl] = ACTIONS(2027),
    [anon_sym_lsr] = ACTIONS(2027),
    [anon_sym_asr] = ACTIONS(2027),
    [aux_sym_pow_operator_token1] = ACTIONS(2029),
    [anon_sym_mod] = ACTIONS(2027),
    [anon_sym_land] = ACTIONS(2027),
    [anon_sym_lor] = ACTIONS(2027),
    [anon_sym_lxor] = ACTIONS(2027),
    [aux_sym_mult_operator_token1] = ACTIONS(2027),
    [aux_sym_add_operator_token1] = ACTIONS(2027),
    [sym_concat_operator] = ACTIONS(2029),
    [sym_rel_operator] = ACTIONS(2027),
    [anon_sym_AMP_AMP] = ACTIONS(2027),
    [anon_sym_or] = ACTIONS(2027),
    [anon_sym_PIPE_PIPE] = ACTIONS(2027),
    [sym_assign_operator] = ACTIONS(2029),
    [sym__uppercase_identifier] = ACTIONS(2029),
    [aux_sym_tag_token1] = ACTIONS(2029),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1069)] = {
    [sym_attribute] = STATE(1069),
    [sym__lowercase_identifier] = ACTIONS(2035),
    [anon_sym_TILDE] = ACTIONS(2035),
    [anon_sym_QMARK] = ACTIONS(2035),
    [anon_sym_LPAREN] = ACTIONS(2037),
    [anon_sym_RPAREN] = ACTIONS(2037),
    [anon_sym_COMMA] = ACTIONS(2037),
    [anon_sym_PIPE] = ACTIONS(2035),
    [anon_sym_COLON2] = ACTIONS(2035),
    [anon_sym_DOT] = ACTIONS(2037),
    [anon_sym_DASH_GT] = ACTIONS(2035),
    [anon_sym_LBRACK] = ACTIONS(2035),
    [anon_sym_RBRACK] = ACTIONS(2037),
    [anon_sym_true] = ACTIONS(2035),
    [anon_sym_false] = ACTIONS(2035),
    [anon_sym_LBRACE] = ACTIONS(2035),
    [anon_sym_SEMI] = ACTIONS(2037),
    [anon_sym_RBRACE] = ACTIONS(2037),
    [anon_sym_constraint] = ACTIONS(2035),
    [anon_sym_val] = ACTIONS(2035),
    [anon_sym_end] = ACTIONS(2035),
    [anon_sym_with] = ACTIONS(2035),
    [anon_sym_COLON_GT] = ACTIONS(2037),
    [anon_sym_object] = ACTIONS(2035),
    [anon_sym_inherit] = ACTIONS(2035),
    [anon_sym_method] = ACTIONS(2035),
    [anon_sym_initializer] = ACTIONS(2035),
    [anon_sym_AMP] = ACTIONS(2035),
    [anon_sym_POUND] = ACTIONS(2035),
    [anon_sym_COLON_COLON] = ACTIONS(2037),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2037),
    [anon_sym_then] = ACTIONS(2035),
    [anon_sym_else] = ACTIONS(2035),
    [anon_sym_do] = ACTIONS(2035),
    [anon_sym_done] = ACTIONS(2035),
    [anon_sym_to] = ACTIONS(2035),
    [anon_sym_downto] = ACTIONS(2035),
    [anon_sym_new] = ACTIONS(2035),
    [anon_sym_LBRACE_LT] = ACTIONS(2037),
    [anon_sym_GT_RBRACE] = ACTIONS(2037),
    [anon_sym_begin] = ACTIONS(2035),
    [sym_ocamlyacc_value] = ACTIONS(2037),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2035),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2037),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2035),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2035),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2037),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2037),
    [aux_sym_number_token1] = ACTIONS(2037),
    [aux_sym_signed_number_token1] = ACTIONS(2035),
    [anon_sym_SQUOTE] = ACTIONS(2037),
    [anon_sym_DQUOTE] = ACTIONS(2037),
    [sym_prefix_operator] = ACTIONS(2037),
    [aux_sym_sign_operator_token1] = ACTIONS(2035),
    [sym_hash_operator] = ACTIONS(2037),
    [anon_sym_lsl] = ACTIONS(2035),
    [anon_sym_lsr] = ACTIONS(2035),
    [anon_sym_asr] = ACTIONS(2035),
    [aux_sym_pow_operator_token1] = ACTIONS(2037),
    [anon_sym_mod] = ACTIONS(2035),
    [anon_sym_land] = ACTIONS(2035),
    [anon_sym_lor] = ACTIONS(2035),
    [anon_sym_lxor] = ACTIONS(2035),
    [aux_sym_mult_operator_token1] = ACTIONS(2035),
    [aux_sym_add_operator_token1] = ACTIONS(2035),
    [sym_concat_operator] = ACTIONS(2037),
    [sym_rel_operator] = ACTIONS(2035),
    [anon_sym_AMP_AMP] = ACTIONS(2035),
    [anon_sym_or] = ACTIONS(2035),
    [anon_sym_PIPE_PIPE] = ACTIONS(2035),
    [sym_assign_operator] = ACTIONS(2037),
    [sym__uppercase_identifier] = ACTIONS(2037),
    [aux_sym_tag_token1] = ACTIONS(2037),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1070)] = {
    [sym_attribute] = STATE(1070),
    [sym__lowercase_identifier] = ACTIONS(2051),
    [anon_sym_TILDE] = ACTIONS(2051),
    [anon_sym_QMARK] = ACTIONS(2051),
    [anon_sym_LPAREN] = ACTIONS(2053),
    [anon_sym_RPAREN] = ACTIONS(2053),
    [anon_sym_COMMA] = ACTIONS(2053),
    [anon_sym_PIPE] = ACTIONS(2051),
    [anon_sym_COLON2] = ACTIONS(2051),
    [anon_sym_DOT] = ACTIONS(2053),
    [anon_sym_DASH_GT] = ACTIONS(2051),
    [anon_sym_LBRACK] = ACTIONS(2051),
    [anon_sym_RBRACK] = ACTIONS(2053),
    [anon_sym_true] = ACTIONS(2051),
    [anon_sym_false] = ACTIONS(2051),
    [anon_sym_LBRACE] = ACTIONS(2051),
    [anon_sym_SEMI] = ACTIONS(2053),
    [anon_sym_RBRACE] = ACTIONS(2053),
    [anon_sym_constraint] = ACTIONS(2051),
    [anon_sym_val] = ACTIONS(2051),
    [anon_sym_end] = ACTIONS(2051),
    [anon_sym_with] = ACTIONS(2051),
    [anon_sym_COLON_GT] = ACTIONS(2053),
    [anon_sym_object] = ACTIONS(2051),
    [anon_sym_inherit] = ACTIONS(2051),
    [anon_sym_method] = ACTIONS(2051),
    [anon_sym_initializer] = ACTIONS(2051),
    [anon_sym_AMP] = ACTIONS(2051),
    [anon_sym_POUND] = ACTIONS(2051),
    [anon_sym_COLON_COLON] = ACTIONS(2053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2053),
    [anon_sym_then] = ACTIONS(2051),
    [anon_sym_else] = ACTIONS(2051),
    [anon_sym_do] = ACTIONS(2051),
    [anon_sym_done] = ACTIONS(2051),
    [anon_sym_to] = ACTIONS(2051),
    [anon_sym_downto] = ACTIONS(2051),
    [anon_sym_new] = ACTIONS(2051),
    [anon_sym_LBRACE_LT] = ACTIONS(2053),
    [anon_sym_GT_RBRACE] = ACTIONS(2053),
    [anon_sym_begin] = ACTIONS(2051),
    [sym_ocamlyacc_value] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2051),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2053),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2051),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2051),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2053),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2053),
    [aux_sym_number_token1] = ACTIONS(2053),
    [aux_sym_signed_number_token1] = ACTIONS(2051),
    [anon_sym_SQUOTE] = ACTIONS(2053),
    [anon_sym_DQUOTE] = ACTIONS(2053),
    [sym_prefix_operator] = ACTIONS(2053),
    [aux_sym_sign_operator_token1] = ACTIONS(2051),
    [sym_hash_operator] = ACTIONS(2053),
    [anon_sym_lsl] = ACTIONS(2051),
    [anon_sym_lsr] = ACTIONS(2051),
    [anon_sym_asr] = ACTIONS(2051),
    [aux_sym_pow_operator_token1] = ACTIONS(2053),
    [anon_sym_mod] = ACTIONS(2051),
    [anon_sym_land] = ACTIONS(2051),
    [anon_sym_lor] = ACTIONS(2051),
    [anon_sym_lxor] = ACTIONS(2051),
    [aux_sym_mult_operator_token1] = ACTIONS(2051),
    [aux_sym_add_operator_token1] = ACTIONS(2051),
    [sym_concat_operator] = ACTIONS(2053),
    [sym_rel_operator] = ACTIONS(2051),
    [anon_sym_AMP_AMP] = ACTIONS(2051),
    [anon_sym_or] = ACTIONS(2051),
    [anon_sym_PIPE_PIPE] = ACTIONS(2051),
    [sym_assign_operator] = ACTIONS(2053),
    [sym__uppercase_identifier] = ACTIONS(2053),
    [aux_sym_tag_token1] = ACTIONS(2053),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1071)] = {
    [sym_attribute] = STATE(1071),
    [sym__lowercase_identifier] = ACTIONS(2055),
    [anon_sym_TILDE] = ACTIONS(2055),
    [anon_sym_QMARK] = ACTIONS(2055),
    [anon_sym_LPAREN] = ACTIONS(2057),
    [anon_sym_RPAREN] = ACTIONS(2057),
    [anon_sym_COMMA] = ACTIONS(2057),
    [anon_sym_PIPE] = ACTIONS(2055),
    [anon_sym_COLON2] = ACTIONS(2055),
    [anon_sym_DOT] = ACTIONS(2057),
    [anon_sym_DASH_GT] = ACTIONS(2055),
    [anon_sym_LBRACK] = ACTIONS(2055),
    [anon_sym_RBRACK] = ACTIONS(2057),
    [anon_sym_true] = ACTIONS(2055),
    [anon_sym_false] = ACTIONS(2055),
    [anon_sym_LBRACE] = ACTIONS(2055),
    [anon_sym_SEMI] = ACTIONS(2057),
    [anon_sym_RBRACE] = ACTIONS(2057),
    [anon_sym_constraint] = ACTIONS(2055),
    [anon_sym_val] = ACTIONS(2055),
    [anon_sym_end] = ACTIONS(2055),
    [anon_sym_with] = ACTIONS(2055),
    [anon_sym_COLON_GT] = ACTIONS(2057),
    [anon_sym_object] = ACTIONS(2055),
    [anon_sym_inherit] = ACTIONS(2055),
    [anon_sym_method] = ACTIONS(2055),
    [anon_sym_initializer] = ACTIONS(2055),
    [anon_sym_AMP] = ACTIONS(2055),
    [anon_sym_POUND] = ACTIONS(2055),
    [anon_sym_COLON_COLON] = ACTIONS(2057),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2057),
    [anon_sym_then] = ACTIONS(2055),
    [anon_sym_else] = ACTIONS(2055),
    [anon_sym_do] = ACTIONS(2055),
    [anon_sym_done] = ACTIONS(2055),
    [anon_sym_to] = ACTIONS(2055),
    [anon_sym_downto] = ACTIONS(2055),
    [anon_sym_new] = ACTIONS(2055),
    [anon_sym_LBRACE_LT] = ACTIONS(2057),
    [anon_sym_GT_RBRACE] = ACTIONS(2057),
    [anon_sym_begin] = ACTIONS(2055),
    [sym_ocamlyacc_value] = ACTIONS(2057),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2057),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2055),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2055),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2057),
    [aux_sym_number_token1] = ACTIONS(2057),
    [aux_sym_signed_number_token1] = ACTIONS(2055),
    [anon_sym_SQUOTE] = ACTIONS(2057),
    [anon_sym_DQUOTE] = ACTIONS(2057),
    [sym_prefix_operator] = ACTIONS(2057),
    [aux_sym_sign_operator_token1] = ACTIONS(2055),
    [sym_hash_operator] = ACTIONS(2057),
    [anon_sym_lsl] = ACTIONS(2055),
    [anon_sym_lsr] = ACTIONS(2055),
    [anon_sym_asr] = ACTIONS(2055),
    [aux_sym_pow_operator_token1] = ACTIONS(2057),
    [anon_sym_mod] = ACTIONS(2055),
    [anon_sym_land] = ACTIONS(2055),
    [anon_sym_lor] = ACTIONS(2055),
    [anon_sym_lxor] = ACTIONS(2055),
    [aux_sym_mult_operator_token1] = ACTIONS(2055),
    [aux_sym_add_operator_token1] = ACTIONS(2055),
    [sym_concat_operator] = ACTIONS(2057),
    [sym_rel_operator] = ACTIONS(2055),
    [anon_sym_AMP_AMP] = ACTIONS(2055),
    [anon_sym_or] = ACTIONS(2055),
    [anon_sym_PIPE_PIPE] = ACTIONS(2055),
    [sym_assign_operator] = ACTIONS(2057),
    [sym__uppercase_identifier] = ACTIONS(2057),
    [aux_sym_tag_token1] = ACTIONS(2057),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1072)] = {
    [sym_attribute] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(2083),
    [anon_sym_QMARK] = ACTIONS(2083),
    [anon_sym_LPAREN] = ACTIONS(2085),
    [anon_sym_RPAREN] = ACTIONS(2085),
    [anon_sym_COMMA] = ACTIONS(2085),
    [anon_sym_PIPE] = ACTIONS(2083),
    [anon_sym_COLON2] = ACTIONS(2083),
    [anon_sym_DOT] = ACTIONS(2085),
    [anon_sym_DASH_GT] = ACTIONS(2083),
    [anon_sym_LBRACK] = ACTIONS(2083),
    [anon_sym_RBRACK] = ACTIONS(2085),
    [anon_sym_true] = ACTIONS(2083),
    [anon_sym_false] = ACTIONS(2083),
    [anon_sym_LBRACE] = ACTIONS(2083),
    [anon_sym_SEMI] = ACTIONS(2085),
    [anon_sym_RBRACE] = ACTIONS(2085),
    [anon_sym_constraint] = ACTIONS(2083),
    [anon_sym_val] = ACTIONS(2083),
    [anon_sym_end] = ACTIONS(2083),
    [anon_sym_with] = ACTIONS(2083),
    [anon_sym_COLON_GT] = ACTIONS(2085),
    [anon_sym_object] = ACTIONS(2083),
    [anon_sym_inherit] = ACTIONS(2083),
    [anon_sym_method] = ACTIONS(2083),
    [anon_sym_initializer] = ACTIONS(2083),
    [anon_sym_AMP] = ACTIONS(2083),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(2085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2085),
    [anon_sym_then] = ACTIONS(2083),
    [anon_sym_else] = ACTIONS(2083),
    [anon_sym_do] = ACTIONS(2083),
    [anon_sym_done] = ACTIONS(2083),
    [anon_sym_to] = ACTIONS(2083),
    [anon_sym_downto] = ACTIONS(2083),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_LBRACE_LT] = ACTIONS(2085),
    [anon_sym_GT_RBRACE] = ACTIONS(2085),
    [anon_sym_begin] = ACTIONS(2083),
    [sym_ocamlyacc_value] = ACTIONS(2085),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2083),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2085),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2083),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2083),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2085),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2085),
    [aux_sym_number_token1] = ACTIONS(2085),
    [aux_sym_signed_number_token1] = ACTIONS(2083),
    [anon_sym_SQUOTE] = ACTIONS(2085),
    [anon_sym_DQUOTE] = ACTIONS(2085),
    [sym_prefix_operator] = ACTIONS(2085),
    [aux_sym_sign_operator_token1] = ACTIONS(2083),
    [sym_hash_operator] = ACTIONS(2085),
    [anon_sym_lsl] = ACTIONS(2083),
    [anon_sym_lsr] = ACTIONS(2083),
    [anon_sym_asr] = ACTIONS(2083),
    [aux_sym_pow_operator_token1] = ACTIONS(2085),
    [anon_sym_mod] = ACTIONS(2083),
    [anon_sym_land] = ACTIONS(2083),
    [anon_sym_lor] = ACTIONS(2083),
    [anon_sym_lxor] = ACTIONS(2083),
    [aux_sym_mult_operator_token1] = ACTIONS(2083),
    [aux_sym_add_operator_token1] = ACTIONS(2083),
    [sym_concat_operator] = ACTIONS(2085),
    [sym_rel_operator] = ACTIONS(2083),
    [anon_sym_AMP_AMP] = ACTIONS(2083),
    [anon_sym_or] = ACTIONS(2083),
    [anon_sym_PIPE_PIPE] = ACTIONS(2083),
    [sym_assign_operator] = ACTIONS(2085),
    [sym__uppercase_identifier] = ACTIONS(2085),
    [aux_sym_tag_token1] = ACTIONS(2085),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1073)] = {
    [sym_attribute] = STATE(1073),
    [sym__lowercase_identifier] = ACTIONS(2207),
    [anon_sym_TILDE] = ACTIONS(2207),
    [anon_sym_QMARK] = ACTIONS(2207),
    [anon_sym_LPAREN] = ACTIONS(2209),
    [anon_sym_RPAREN] = ACTIONS(2209),
    [anon_sym_COMMA] = ACTIONS(2209),
    [anon_sym_PIPE] = ACTIONS(2207),
    [anon_sym_COLON2] = ACTIONS(2207),
    [anon_sym_DOT] = ACTIONS(2209),
    [anon_sym_DASH_GT] = ACTIONS(2207),
    [anon_sym_LBRACK] = ACTIONS(2207),
    [anon_sym_RBRACK] = ACTIONS(2209),
    [anon_sym_true] = ACTIONS(2207),
    [anon_sym_false] = ACTIONS(2207),
    [anon_sym_LBRACE] = ACTIONS(2207),
    [anon_sym_SEMI] = ACTIONS(2209),
    [anon_sym_RBRACE] = ACTIONS(2209),
    [anon_sym_constraint] = ACTIONS(2207),
    [anon_sym_val] = ACTIONS(2207),
    [anon_sym_end] = ACTIONS(2207),
    [anon_sym_with] = ACTIONS(2207),
    [anon_sym_COLON_GT] = ACTIONS(2209),
    [anon_sym_object] = ACTIONS(2207),
    [anon_sym_inherit] = ACTIONS(2207),
    [anon_sym_method] = ACTIONS(2207),
    [anon_sym_initializer] = ACTIONS(2207),
    [anon_sym_AMP] = ACTIONS(2207),
    [anon_sym_POUND] = ACTIONS(2207),
    [anon_sym_COLON_COLON] = ACTIONS(2209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2209),
    [anon_sym_then] = ACTIONS(2207),
    [anon_sym_else] = ACTIONS(2207),
    [anon_sym_do] = ACTIONS(2207),
    [anon_sym_done] = ACTIONS(2207),
    [anon_sym_to] = ACTIONS(2207),
    [anon_sym_downto] = ACTIONS(2207),
    [anon_sym_new] = ACTIONS(2207),
    [anon_sym_LBRACE_LT] = ACTIONS(2209),
    [anon_sym_GT_RBRACE] = ACTIONS(2209),
    [anon_sym_begin] = ACTIONS(2207),
    [sym_ocamlyacc_value] = ACTIONS(2209),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2207),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2209),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2207),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2207),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2209),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2209),
    [aux_sym_number_token1] = ACTIONS(2209),
    [aux_sym_signed_number_token1] = ACTIONS(2207),
    [anon_sym_SQUOTE] = ACTIONS(2209),
    [anon_sym_DQUOTE] = ACTIONS(2209),
    [sym_prefix_operator] = ACTIONS(2209),
    [aux_sym_sign_operator_token1] = ACTIONS(2207),
    [sym_hash_operator] = ACTIONS(2209),
    [anon_sym_lsl] = ACTIONS(2207),
    [anon_sym_lsr] = ACTIONS(2207),
    [anon_sym_asr] = ACTIONS(2207),
    [aux_sym_pow_operator_token1] = ACTIONS(2209),
    [anon_sym_mod] = ACTIONS(2207),
    [anon_sym_land] = ACTIONS(2207),
    [anon_sym_lor] = ACTIONS(2207),
    [anon_sym_lxor] = ACTIONS(2207),
    [aux_sym_mult_operator_token1] = ACTIONS(2207),
    [aux_sym_add_operator_token1] = ACTIONS(2207),
    [sym_concat_operator] = ACTIONS(2209),
    [sym_rel_operator] = ACTIONS(2207),
    [anon_sym_AMP_AMP] = ACTIONS(2207),
    [anon_sym_or] = ACTIONS(2207),
    [anon_sym_PIPE_PIPE] = ACTIONS(2207),
    [sym_assign_operator] = ACTIONS(2209),
    [sym__uppercase_identifier] = ACTIONS(2209),
    [aux_sym_tag_token1] = ACTIONS(2209),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1074)] = {
    [sym_attribute] = STATE(1074),
    [sym__lowercase_identifier] = ACTIONS(2121),
    [anon_sym_TILDE] = ACTIONS(2121),
    [anon_sym_QMARK] = ACTIONS(2121),
    [anon_sym_LPAREN] = ACTIONS(2123),
    [anon_sym_RPAREN] = ACTIONS(2123),
    [anon_sym_COMMA] = ACTIONS(2123),
    [anon_sym_PIPE] = ACTIONS(2121),
    [anon_sym_COLON2] = ACTIONS(2121),
    [anon_sym_DOT] = ACTIONS(2123),
    [anon_sym_DASH_GT] = ACTIONS(2121),
    [anon_sym_LBRACK] = ACTIONS(2121),
    [anon_sym_RBRACK] = ACTIONS(2123),
    [anon_sym_true] = ACTIONS(2121),
    [anon_sym_false] = ACTIONS(2121),
    [anon_sym_LBRACE] = ACTIONS(2121),
    [anon_sym_SEMI] = ACTIONS(2123),
    [anon_sym_RBRACE] = ACTIONS(2123),
    [anon_sym_constraint] = ACTIONS(2121),
    [anon_sym_val] = ACTIONS(2121),
    [anon_sym_end] = ACTIONS(2121),
    [anon_sym_with] = ACTIONS(2121),
    [anon_sym_COLON_GT] = ACTIONS(2123),
    [anon_sym_object] = ACTIONS(2121),
    [anon_sym_inherit] = ACTIONS(2121),
    [anon_sym_method] = ACTIONS(2121),
    [anon_sym_initializer] = ACTIONS(2121),
    [anon_sym_AMP] = ACTIONS(2121),
    [anon_sym_POUND] = ACTIONS(2121),
    [anon_sym_COLON_COLON] = ACTIONS(2123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2123),
    [anon_sym_then] = ACTIONS(2121),
    [anon_sym_else] = ACTIONS(2121),
    [anon_sym_do] = ACTIONS(2121),
    [anon_sym_done] = ACTIONS(2121),
    [anon_sym_to] = ACTIONS(2121),
    [anon_sym_downto] = ACTIONS(2121),
    [anon_sym_new] = ACTIONS(2121),
    [anon_sym_LBRACE_LT] = ACTIONS(2123),
    [anon_sym_GT_RBRACE] = ACTIONS(2123),
    [anon_sym_begin] = ACTIONS(2121),
    [sym_ocamlyacc_value] = ACTIONS(2123),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2121),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2123),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2121),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2121),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2123),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2123),
    [aux_sym_number_token1] = ACTIONS(2123),
    [aux_sym_signed_number_token1] = ACTIONS(2121),
    [anon_sym_SQUOTE] = ACTIONS(2123),
    [anon_sym_DQUOTE] = ACTIONS(2123),
    [sym_prefix_operator] = ACTIONS(2123),
    [aux_sym_sign_operator_token1] = ACTIONS(2121),
    [sym_hash_operator] = ACTIONS(2123),
    [anon_sym_lsl] = ACTIONS(2121),
    [anon_sym_lsr] = ACTIONS(2121),
    [anon_sym_asr] = ACTIONS(2121),
    [aux_sym_pow_operator_token1] = ACTIONS(2123),
    [anon_sym_mod] = ACTIONS(2121),
    [anon_sym_land] = ACTIONS(2121),
    [anon_sym_lor] = ACTIONS(2121),
    [anon_sym_lxor] = ACTIONS(2121),
    [aux_sym_mult_operator_token1] = ACTIONS(2121),
    [aux_sym_add_operator_token1] = ACTIONS(2121),
    [sym_concat_operator] = ACTIONS(2123),
    [sym_rel_operator] = ACTIONS(2121),
    [anon_sym_AMP_AMP] = ACTIONS(2121),
    [anon_sym_or] = ACTIONS(2121),
    [anon_sym_PIPE_PIPE] = ACTIONS(2121),
    [sym_assign_operator] = ACTIONS(2123),
    [sym__uppercase_identifier] = ACTIONS(2123),
    [aux_sym_tag_token1] = ACTIONS(2123),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1075)] = {
    [sym_attribute] = STATE(1075),
    [sym__lowercase_identifier] = ACTIONS(2047),
    [anon_sym_TILDE] = ACTIONS(2047),
    [anon_sym_QMARK] = ACTIONS(2047),
    [anon_sym_LPAREN] = ACTIONS(2049),
    [anon_sym_RPAREN] = ACTIONS(2049),
    [anon_sym_COMMA] = ACTIONS(2049),
    [anon_sym_PIPE] = ACTIONS(2047),
    [anon_sym_COLON2] = ACTIONS(2047),
    [anon_sym_DOT] = ACTIONS(2049),
    [anon_sym_DASH_GT] = ACTIONS(2047),
    [anon_sym_LBRACK] = ACTIONS(2047),
    [anon_sym_RBRACK] = ACTIONS(2049),
    [anon_sym_true] = ACTIONS(2047),
    [anon_sym_false] = ACTIONS(2047),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_SEMI] = ACTIONS(2049),
    [anon_sym_RBRACE] = ACTIONS(2049),
    [anon_sym_constraint] = ACTIONS(2047),
    [anon_sym_val] = ACTIONS(2047),
    [anon_sym_end] = ACTIONS(2047),
    [anon_sym_with] = ACTIONS(2047),
    [anon_sym_COLON_GT] = ACTIONS(2049),
    [anon_sym_object] = ACTIONS(2047),
    [anon_sym_inherit] = ACTIONS(2047),
    [anon_sym_method] = ACTIONS(2047),
    [anon_sym_initializer] = ACTIONS(2047),
    [anon_sym_AMP] = ACTIONS(2047),
    [anon_sym_POUND] = ACTIONS(2047),
    [anon_sym_COLON_COLON] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2049),
    [anon_sym_then] = ACTIONS(2047),
    [anon_sym_else] = ACTIONS(2047),
    [anon_sym_do] = ACTIONS(2047),
    [anon_sym_done] = ACTIONS(2047),
    [anon_sym_to] = ACTIONS(2047),
    [anon_sym_downto] = ACTIONS(2047),
    [anon_sym_new] = ACTIONS(2047),
    [anon_sym_LBRACE_LT] = ACTIONS(2049),
    [anon_sym_GT_RBRACE] = ACTIONS(2049),
    [anon_sym_begin] = ACTIONS(2047),
    [sym_ocamlyacc_value] = ACTIONS(2049),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2047),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2049),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2047),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2047),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2049),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2049),
    [aux_sym_number_token1] = ACTIONS(2049),
    [aux_sym_signed_number_token1] = ACTIONS(2047),
    [anon_sym_SQUOTE] = ACTIONS(2049),
    [anon_sym_DQUOTE] = ACTIONS(2049),
    [sym_prefix_operator] = ACTIONS(2049),
    [aux_sym_sign_operator_token1] = ACTIONS(2047),
    [sym_hash_operator] = ACTIONS(2049),
    [anon_sym_lsl] = ACTIONS(2047),
    [anon_sym_lsr] = ACTIONS(2047),
    [anon_sym_asr] = ACTIONS(2047),
    [aux_sym_pow_operator_token1] = ACTIONS(2049),
    [anon_sym_mod] = ACTIONS(2047),
    [anon_sym_land] = ACTIONS(2047),
    [anon_sym_lor] = ACTIONS(2047),
    [anon_sym_lxor] = ACTIONS(2047),
    [aux_sym_mult_operator_token1] = ACTIONS(2047),
    [aux_sym_add_operator_token1] = ACTIONS(2047),
    [sym_concat_operator] = ACTIONS(2049),
    [sym_rel_operator] = ACTIONS(2047),
    [anon_sym_AMP_AMP] = ACTIONS(2047),
    [anon_sym_or] = ACTIONS(2047),
    [anon_sym_PIPE_PIPE] = ACTIONS(2047),
    [sym_assign_operator] = ACTIONS(2049),
    [sym__uppercase_identifier] = ACTIONS(2049),
    [aux_sym_tag_token1] = ACTIONS(2049),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1076)] = {
    [sym_attribute] = STATE(1076),
    [sym__lowercase_identifier] = ACTIONS(2075),
    [anon_sym_TILDE] = ACTIONS(2075),
    [anon_sym_QMARK] = ACTIONS(2075),
    [anon_sym_LPAREN] = ACTIONS(2077),
    [anon_sym_RPAREN] = ACTIONS(2077),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_PIPE] = ACTIONS(2075),
    [anon_sym_COLON2] = ACTIONS(2075),
    [anon_sym_DOT] = ACTIONS(2077),
    [anon_sym_DASH_GT] = ACTIONS(2075),
    [anon_sym_LBRACK] = ACTIONS(2075),
    [anon_sym_RBRACK] = ACTIONS(2077),
    [anon_sym_true] = ACTIONS(2075),
    [anon_sym_false] = ACTIONS(2075),
    [anon_sym_LBRACE] = ACTIONS(2075),
    [anon_sym_SEMI] = ACTIONS(2077),
    [anon_sym_RBRACE] = ACTIONS(2077),
    [anon_sym_constraint] = ACTIONS(2075),
    [anon_sym_val] = ACTIONS(2075),
    [anon_sym_end] = ACTIONS(2075),
    [anon_sym_with] = ACTIONS(2075),
    [anon_sym_COLON_GT] = ACTIONS(2077),
    [anon_sym_object] = ACTIONS(2075),
    [anon_sym_inherit] = ACTIONS(2075),
    [anon_sym_method] = ACTIONS(2075),
    [anon_sym_initializer] = ACTIONS(2075),
    [anon_sym_AMP] = ACTIONS(2075),
    [anon_sym_POUND] = ACTIONS(2075),
    [anon_sym_COLON_COLON] = ACTIONS(2077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2077),
    [anon_sym_then] = ACTIONS(2075),
    [anon_sym_else] = ACTIONS(2075),
    [anon_sym_do] = ACTIONS(2075),
    [anon_sym_done] = ACTIONS(2075),
    [anon_sym_to] = ACTIONS(2075),
    [anon_sym_downto] = ACTIONS(2075),
    [anon_sym_new] = ACTIONS(2075),
    [anon_sym_LBRACE_LT] = ACTIONS(2077),
    [anon_sym_GT_RBRACE] = ACTIONS(2077),
    [anon_sym_begin] = ACTIONS(2075),
    [sym_ocamlyacc_value] = ACTIONS(2077),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2075),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2077),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2075),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2075),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2077),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2077),
    [aux_sym_number_token1] = ACTIONS(2077),
    [aux_sym_signed_number_token1] = ACTIONS(2075),
    [anon_sym_SQUOTE] = ACTIONS(2077),
    [anon_sym_DQUOTE] = ACTIONS(2077),
    [sym_prefix_operator] = ACTIONS(2077),
    [aux_sym_sign_operator_token1] = ACTIONS(2075),
    [sym_hash_operator] = ACTIONS(2077),
    [anon_sym_lsl] = ACTIONS(2075),
    [anon_sym_lsr] = ACTIONS(2075),
    [anon_sym_asr] = ACTIONS(2075),
    [aux_sym_pow_operator_token1] = ACTIONS(2077),
    [anon_sym_mod] = ACTIONS(2075),
    [anon_sym_land] = ACTIONS(2075),
    [anon_sym_lor] = ACTIONS(2075),
    [anon_sym_lxor] = ACTIONS(2075),
    [aux_sym_mult_operator_token1] = ACTIONS(2075),
    [aux_sym_add_operator_token1] = ACTIONS(2075),
    [sym_concat_operator] = ACTIONS(2077),
    [sym_rel_operator] = ACTIONS(2075),
    [anon_sym_AMP_AMP] = ACTIONS(2075),
    [anon_sym_or] = ACTIONS(2075),
    [anon_sym_PIPE_PIPE] = ACTIONS(2075),
    [sym_assign_operator] = ACTIONS(2077),
    [sym__uppercase_identifier] = ACTIONS(2077),
    [aux_sym_tag_token1] = ACTIONS(2077),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1077)] = {
    [sym_attribute] = STATE(1077),
    [sym__lowercase_identifier] = ACTIONS(1951),
    [anon_sym_TILDE] = ACTIONS(1951),
    [anon_sym_QMARK] = ACTIONS(1951),
    [anon_sym_LPAREN] = ACTIONS(1953),
    [anon_sym_RPAREN] = ACTIONS(1953),
    [anon_sym_COMMA] = ACTIONS(1953),
    [anon_sym_PIPE] = ACTIONS(1951),
    [anon_sym_COLON2] = ACTIONS(1951),
    [anon_sym_DOT] = ACTIONS(1953),
    [anon_sym_DASH_GT] = ACTIONS(1951),
    [anon_sym_LBRACK] = ACTIONS(1951),
    [anon_sym_RBRACK] = ACTIONS(1953),
    [anon_sym_true] = ACTIONS(1951),
    [anon_sym_false] = ACTIONS(1951),
    [anon_sym_LBRACE] = ACTIONS(1951),
    [anon_sym_SEMI] = ACTIONS(1953),
    [anon_sym_RBRACE] = ACTIONS(1953),
    [anon_sym_constraint] = ACTIONS(1951),
    [anon_sym_val] = ACTIONS(1951),
    [anon_sym_end] = ACTIONS(1951),
    [anon_sym_with] = ACTIONS(1951),
    [anon_sym_COLON_GT] = ACTIONS(1953),
    [anon_sym_object] = ACTIONS(1951),
    [anon_sym_inherit] = ACTIONS(1951),
    [anon_sym_method] = ACTIONS(1951),
    [anon_sym_initializer] = ACTIONS(1951),
    [anon_sym_AMP] = ACTIONS(1951),
    [anon_sym_POUND] = ACTIONS(1951),
    [anon_sym_COLON_COLON] = ACTIONS(1953),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1953),
    [anon_sym_then] = ACTIONS(1951),
    [anon_sym_else] = ACTIONS(1951),
    [anon_sym_do] = ACTIONS(1951),
    [anon_sym_done] = ACTIONS(1951),
    [anon_sym_to] = ACTIONS(1951),
    [anon_sym_downto] = ACTIONS(1951),
    [anon_sym_new] = ACTIONS(1951),
    [anon_sym_LBRACE_LT] = ACTIONS(1953),
    [anon_sym_GT_RBRACE] = ACTIONS(1953),
    [anon_sym_begin] = ACTIONS(1951),
    [sym_ocamlyacc_value] = ACTIONS(1953),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1951),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1953),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1951),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1951),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1953),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1953),
    [aux_sym_number_token1] = ACTIONS(1953),
    [aux_sym_signed_number_token1] = ACTIONS(1951),
    [anon_sym_SQUOTE] = ACTIONS(1953),
    [anon_sym_DQUOTE] = ACTIONS(1953),
    [sym_prefix_operator] = ACTIONS(1953),
    [aux_sym_sign_operator_token1] = ACTIONS(1951),
    [sym_hash_operator] = ACTIONS(1953),
    [anon_sym_lsl] = ACTIONS(1951),
    [anon_sym_lsr] = ACTIONS(1951),
    [anon_sym_asr] = ACTIONS(1951),
    [aux_sym_pow_operator_token1] = ACTIONS(1953),
    [anon_sym_mod] = ACTIONS(1951),
    [anon_sym_land] = ACTIONS(1951),
    [anon_sym_lor] = ACTIONS(1951),
    [anon_sym_lxor] = ACTIONS(1951),
    [aux_sym_mult_operator_token1] = ACTIONS(1951),
    [aux_sym_add_operator_token1] = ACTIONS(1951),
    [sym_concat_operator] = ACTIONS(1953),
    [sym_rel_operator] = ACTIONS(1951),
    [anon_sym_AMP_AMP] = ACTIONS(1951),
    [anon_sym_or] = ACTIONS(1951),
    [anon_sym_PIPE_PIPE] = ACTIONS(1951),
    [sym_assign_operator] = ACTIONS(1953),
    [sym__uppercase_identifier] = ACTIONS(1953),
    [aux_sym_tag_token1] = ACTIONS(1953),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1078)] = {
    [sym_attribute] = STATE(1078),
    [sym__lowercase_identifier] = ACTIONS(1979),
    [anon_sym_TILDE] = ACTIONS(1979),
    [anon_sym_QMARK] = ACTIONS(1979),
    [anon_sym_LPAREN] = ACTIONS(1981),
    [anon_sym_RPAREN] = ACTIONS(1981),
    [anon_sym_COMMA] = ACTIONS(1981),
    [anon_sym_PIPE] = ACTIONS(1979),
    [anon_sym_COLON2] = ACTIONS(1979),
    [anon_sym_DOT] = ACTIONS(1981),
    [anon_sym_DASH_GT] = ACTIONS(1979),
    [anon_sym_LBRACK] = ACTIONS(1979),
    [anon_sym_RBRACK] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(1979),
    [anon_sym_false] = ACTIONS(1979),
    [anon_sym_LBRACE] = ACTIONS(1979),
    [anon_sym_SEMI] = ACTIONS(1981),
    [anon_sym_RBRACE] = ACTIONS(1981),
    [anon_sym_constraint] = ACTIONS(1979),
    [anon_sym_val] = ACTIONS(1979),
    [anon_sym_end] = ACTIONS(1979),
    [anon_sym_with] = ACTIONS(1979),
    [anon_sym_COLON_GT] = ACTIONS(1981),
    [anon_sym_object] = ACTIONS(1979),
    [anon_sym_inherit] = ACTIONS(1979),
    [anon_sym_method] = ACTIONS(1979),
    [anon_sym_initializer] = ACTIONS(1979),
    [anon_sym_AMP] = ACTIONS(1979),
    [anon_sym_POUND] = ACTIONS(1979),
    [anon_sym_COLON_COLON] = ACTIONS(1981),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1981),
    [anon_sym_then] = ACTIONS(1979),
    [anon_sym_else] = ACTIONS(1979),
    [anon_sym_do] = ACTIONS(1979),
    [anon_sym_done] = ACTIONS(1979),
    [anon_sym_to] = ACTIONS(1979),
    [anon_sym_downto] = ACTIONS(1979),
    [anon_sym_new] = ACTIONS(1979),
    [anon_sym_LBRACE_LT] = ACTIONS(1981),
    [anon_sym_GT_RBRACE] = ACTIONS(1981),
    [anon_sym_begin] = ACTIONS(1979),
    [sym_ocamlyacc_value] = ACTIONS(1981),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1979),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1981),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1979),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1979),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1981),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1981),
    [aux_sym_number_token1] = ACTIONS(1981),
    [aux_sym_signed_number_token1] = ACTIONS(1979),
    [anon_sym_SQUOTE] = ACTIONS(1981),
    [anon_sym_DQUOTE] = ACTIONS(1981),
    [sym_prefix_operator] = ACTIONS(1981),
    [aux_sym_sign_operator_token1] = ACTIONS(1979),
    [sym_hash_operator] = ACTIONS(1981),
    [anon_sym_lsl] = ACTIONS(1979),
    [anon_sym_lsr] = ACTIONS(1979),
    [anon_sym_asr] = ACTIONS(1979),
    [aux_sym_pow_operator_token1] = ACTIONS(1981),
    [anon_sym_mod] = ACTIONS(1979),
    [anon_sym_land] = ACTIONS(1979),
    [anon_sym_lor] = ACTIONS(1979),
    [anon_sym_lxor] = ACTIONS(1979),
    [aux_sym_mult_operator_token1] = ACTIONS(1979),
    [aux_sym_add_operator_token1] = ACTIONS(1979),
    [sym_concat_operator] = ACTIONS(1981),
    [sym_rel_operator] = ACTIONS(1979),
    [anon_sym_AMP_AMP] = ACTIONS(1979),
    [anon_sym_or] = ACTIONS(1979),
    [anon_sym_PIPE_PIPE] = ACTIONS(1979),
    [sym_assign_operator] = ACTIONS(1981),
    [sym__uppercase_identifier] = ACTIONS(1981),
    [aux_sym_tag_token1] = ACTIONS(1981),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1079)] = {
    [sym_attribute] = STATE(1079),
    [sym__lowercase_identifier] = ACTIONS(1999),
    [anon_sym_TILDE] = ACTIONS(1999),
    [anon_sym_QMARK] = ACTIONS(1999),
    [anon_sym_LPAREN] = ACTIONS(2001),
    [anon_sym_RPAREN] = ACTIONS(2001),
    [anon_sym_COMMA] = ACTIONS(2001),
    [anon_sym_PIPE] = ACTIONS(1999),
    [anon_sym_COLON2] = ACTIONS(1999),
    [anon_sym_DOT] = ACTIONS(2001),
    [anon_sym_DASH_GT] = ACTIONS(1999),
    [anon_sym_LBRACK] = ACTIONS(1999),
    [anon_sym_RBRACK] = ACTIONS(2001),
    [anon_sym_true] = ACTIONS(1999),
    [anon_sym_false] = ACTIONS(1999),
    [anon_sym_LBRACE] = ACTIONS(1999),
    [anon_sym_SEMI] = ACTIONS(2001),
    [anon_sym_RBRACE] = ACTIONS(2001),
    [anon_sym_constraint] = ACTIONS(1999),
    [anon_sym_val] = ACTIONS(1999),
    [anon_sym_end] = ACTIONS(1999),
    [anon_sym_with] = ACTIONS(1999),
    [anon_sym_COLON_GT] = ACTIONS(2001),
    [anon_sym_object] = ACTIONS(1999),
    [anon_sym_inherit] = ACTIONS(1999),
    [anon_sym_method] = ACTIONS(1999),
    [anon_sym_initializer] = ACTIONS(1999),
    [anon_sym_AMP] = ACTIONS(1999),
    [anon_sym_POUND] = ACTIONS(1999),
    [anon_sym_COLON_COLON] = ACTIONS(2001),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2001),
    [anon_sym_then] = ACTIONS(1999),
    [anon_sym_else] = ACTIONS(1999),
    [anon_sym_do] = ACTIONS(1999),
    [anon_sym_done] = ACTIONS(1999),
    [anon_sym_to] = ACTIONS(1999),
    [anon_sym_downto] = ACTIONS(1999),
    [anon_sym_new] = ACTIONS(1999),
    [anon_sym_LBRACE_LT] = ACTIONS(2001),
    [anon_sym_GT_RBRACE] = ACTIONS(2001),
    [anon_sym_begin] = ACTIONS(1999),
    [sym_ocamlyacc_value] = ACTIONS(2001),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2001),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1999),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1999),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2001),
    [aux_sym_number_token1] = ACTIONS(2001),
    [aux_sym_signed_number_token1] = ACTIONS(1999),
    [anon_sym_SQUOTE] = ACTIONS(2001),
    [anon_sym_DQUOTE] = ACTIONS(2001),
    [sym_prefix_operator] = ACTIONS(2001),
    [aux_sym_sign_operator_token1] = ACTIONS(1999),
    [sym_hash_operator] = ACTIONS(2001),
    [anon_sym_lsl] = ACTIONS(1999),
    [anon_sym_lsr] = ACTIONS(1999),
    [anon_sym_asr] = ACTIONS(1999),
    [aux_sym_pow_operator_token1] = ACTIONS(2001),
    [anon_sym_mod] = ACTIONS(1999),
    [anon_sym_land] = ACTIONS(1999),
    [anon_sym_lor] = ACTIONS(1999),
    [anon_sym_lxor] = ACTIONS(1999),
    [aux_sym_mult_operator_token1] = ACTIONS(1999),
    [aux_sym_add_operator_token1] = ACTIONS(1999),
    [sym_concat_operator] = ACTIONS(2001),
    [sym_rel_operator] = ACTIONS(1999),
    [anon_sym_AMP_AMP] = ACTIONS(1999),
    [anon_sym_or] = ACTIONS(1999),
    [anon_sym_PIPE_PIPE] = ACTIONS(1999),
    [sym_assign_operator] = ACTIONS(2001),
    [sym__uppercase_identifier] = ACTIONS(2001),
    [aux_sym_tag_token1] = ACTIONS(2001),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1080)] = {
    [sym_attribute] = STATE(1080),
    [sym__lowercase_identifier] = ACTIONS(2003),
    [anon_sym_TILDE] = ACTIONS(2003),
    [anon_sym_QMARK] = ACTIONS(2003),
    [anon_sym_LPAREN] = ACTIONS(2005),
    [anon_sym_RPAREN] = ACTIONS(2005),
    [anon_sym_COMMA] = ACTIONS(2005),
    [anon_sym_PIPE] = ACTIONS(2003),
    [anon_sym_COLON2] = ACTIONS(2003),
    [anon_sym_DOT] = ACTIONS(2005),
    [anon_sym_DASH_GT] = ACTIONS(2003),
    [anon_sym_LBRACK] = ACTIONS(2003),
    [anon_sym_RBRACK] = ACTIONS(2005),
    [anon_sym_true] = ACTIONS(2003),
    [anon_sym_false] = ACTIONS(2003),
    [anon_sym_LBRACE] = ACTIONS(2003),
    [anon_sym_SEMI] = ACTIONS(2005),
    [anon_sym_RBRACE] = ACTIONS(2005),
    [anon_sym_constraint] = ACTIONS(2003),
    [anon_sym_val] = ACTIONS(2003),
    [anon_sym_end] = ACTIONS(2003),
    [anon_sym_with] = ACTIONS(2003),
    [anon_sym_COLON_GT] = ACTIONS(2005),
    [anon_sym_object] = ACTIONS(2003),
    [anon_sym_inherit] = ACTIONS(2003),
    [anon_sym_method] = ACTIONS(2003),
    [anon_sym_initializer] = ACTIONS(2003),
    [anon_sym_AMP] = ACTIONS(2003),
    [anon_sym_POUND] = ACTIONS(2003),
    [anon_sym_COLON_COLON] = ACTIONS(2005),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2005),
    [anon_sym_then] = ACTIONS(2003),
    [anon_sym_else] = ACTIONS(2003),
    [anon_sym_do] = ACTIONS(2003),
    [anon_sym_done] = ACTIONS(2003),
    [anon_sym_to] = ACTIONS(2003),
    [anon_sym_downto] = ACTIONS(2003),
    [anon_sym_new] = ACTIONS(2003),
    [anon_sym_LBRACE_LT] = ACTIONS(2005),
    [anon_sym_GT_RBRACE] = ACTIONS(2005),
    [anon_sym_begin] = ACTIONS(2003),
    [sym_ocamlyacc_value] = ACTIONS(2005),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2003),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2005),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2005),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2005),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2003),
    [anon_sym_SQUOTE] = ACTIONS(2005),
    [anon_sym_DQUOTE] = ACTIONS(2005),
    [sym_prefix_operator] = ACTIONS(2005),
    [aux_sym_sign_operator_token1] = ACTIONS(2003),
    [sym_hash_operator] = ACTIONS(2005),
    [anon_sym_lsl] = ACTIONS(2003),
    [anon_sym_lsr] = ACTIONS(2003),
    [anon_sym_asr] = ACTIONS(2003),
    [aux_sym_pow_operator_token1] = ACTIONS(2005),
    [anon_sym_mod] = ACTIONS(2003),
    [anon_sym_land] = ACTIONS(2003),
    [anon_sym_lor] = ACTIONS(2003),
    [anon_sym_lxor] = ACTIONS(2003),
    [aux_sym_mult_operator_token1] = ACTIONS(2003),
    [aux_sym_add_operator_token1] = ACTIONS(2003),
    [sym_concat_operator] = ACTIONS(2005),
    [sym_rel_operator] = ACTIONS(2003),
    [anon_sym_AMP_AMP] = ACTIONS(2003),
    [anon_sym_or] = ACTIONS(2003),
    [anon_sym_PIPE_PIPE] = ACTIONS(2003),
    [sym_assign_operator] = ACTIONS(2005),
    [sym__uppercase_identifier] = ACTIONS(2005),
    [aux_sym_tag_token1] = ACTIONS(2005),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1081)] = {
    [sym_attribute] = STATE(1081),
    [sym__lowercase_identifier] = ACTIONS(1923),
    [anon_sym_TILDE] = ACTIONS(1923),
    [anon_sym_QMARK] = ACTIONS(1923),
    [anon_sym_LPAREN] = ACTIONS(1925),
    [anon_sym_RPAREN] = ACTIONS(1925),
    [anon_sym_COMMA] = ACTIONS(1925),
    [anon_sym_PIPE] = ACTIONS(1923),
    [anon_sym_COLON2] = ACTIONS(1923),
    [anon_sym_DOT] = ACTIONS(1925),
    [anon_sym_DASH_GT] = ACTIONS(1923),
    [anon_sym_LBRACK] = ACTIONS(1923),
    [anon_sym_RBRACK] = ACTIONS(1925),
    [anon_sym_true] = ACTIONS(1923),
    [anon_sym_false] = ACTIONS(1923),
    [anon_sym_LBRACE] = ACTIONS(1923),
    [anon_sym_SEMI] = ACTIONS(1925),
    [anon_sym_RBRACE] = ACTIONS(1925),
    [anon_sym_constraint] = ACTIONS(1923),
    [anon_sym_val] = ACTIONS(1923),
    [anon_sym_end] = ACTIONS(1923),
    [anon_sym_with] = ACTIONS(1923),
    [anon_sym_COLON_GT] = ACTIONS(1925),
    [anon_sym_object] = ACTIONS(1923),
    [anon_sym_inherit] = ACTIONS(1923),
    [anon_sym_method] = ACTIONS(1923),
    [anon_sym_initializer] = ACTIONS(1923),
    [anon_sym_AMP] = ACTIONS(1923),
    [anon_sym_POUND] = ACTIONS(1923),
    [anon_sym_COLON_COLON] = ACTIONS(1925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1925),
    [anon_sym_then] = ACTIONS(1923),
    [anon_sym_else] = ACTIONS(1923),
    [anon_sym_do] = ACTIONS(1923),
    [anon_sym_done] = ACTIONS(1923),
    [anon_sym_to] = ACTIONS(1923),
    [anon_sym_downto] = ACTIONS(1923),
    [anon_sym_new] = ACTIONS(1923),
    [anon_sym_LBRACE_LT] = ACTIONS(1925),
    [anon_sym_GT_RBRACE] = ACTIONS(1925),
    [anon_sym_begin] = ACTIONS(1923),
    [sym_ocamlyacc_value] = ACTIONS(1925),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1923),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1925),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1923),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1923),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1925),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1925),
    [aux_sym_number_token1] = ACTIONS(1925),
    [aux_sym_signed_number_token1] = ACTIONS(1923),
    [anon_sym_SQUOTE] = ACTIONS(1925),
    [anon_sym_DQUOTE] = ACTIONS(1925),
    [sym_prefix_operator] = ACTIONS(1925),
    [aux_sym_sign_operator_token1] = ACTIONS(1923),
    [sym_hash_operator] = ACTIONS(1925),
    [anon_sym_lsl] = ACTIONS(1923),
    [anon_sym_lsr] = ACTIONS(1923),
    [anon_sym_asr] = ACTIONS(1923),
    [aux_sym_pow_operator_token1] = ACTIONS(1925),
    [anon_sym_mod] = ACTIONS(1923),
    [anon_sym_land] = ACTIONS(1923),
    [anon_sym_lor] = ACTIONS(1923),
    [anon_sym_lxor] = ACTIONS(1923),
    [aux_sym_mult_operator_token1] = ACTIONS(1923),
    [aux_sym_add_operator_token1] = ACTIONS(1923),
    [sym_concat_operator] = ACTIONS(1925),
    [sym_rel_operator] = ACTIONS(1923),
    [anon_sym_AMP_AMP] = ACTIONS(1923),
    [anon_sym_or] = ACTIONS(1923),
    [anon_sym_PIPE_PIPE] = ACTIONS(1923),
    [sym_assign_operator] = ACTIONS(1925),
    [sym__uppercase_identifier] = ACTIONS(1925),
    [aux_sym_tag_token1] = ACTIONS(1925),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1082)] = {
    [sym_attribute] = STATE(1082),
    [sym__lowercase_identifier] = ACTIONS(2059),
    [anon_sym_TILDE] = ACTIONS(2059),
    [anon_sym_QMARK] = ACTIONS(2059),
    [anon_sym_LPAREN] = ACTIONS(2061),
    [anon_sym_RPAREN] = ACTIONS(2061),
    [anon_sym_COMMA] = ACTIONS(2061),
    [anon_sym_PIPE] = ACTIONS(2059),
    [anon_sym_COLON2] = ACTIONS(2059),
    [anon_sym_DOT] = ACTIONS(2061),
    [anon_sym_DASH_GT] = ACTIONS(2059),
    [anon_sym_LBRACK] = ACTIONS(2059),
    [anon_sym_RBRACK] = ACTIONS(2061),
    [anon_sym_true] = ACTIONS(2059),
    [anon_sym_false] = ACTIONS(2059),
    [anon_sym_LBRACE] = ACTIONS(2059),
    [anon_sym_SEMI] = ACTIONS(2061),
    [anon_sym_RBRACE] = ACTIONS(2061),
    [anon_sym_constraint] = ACTIONS(2059),
    [anon_sym_val] = ACTIONS(2059),
    [anon_sym_end] = ACTIONS(2059),
    [anon_sym_with] = ACTIONS(2059),
    [anon_sym_COLON_GT] = ACTIONS(2061),
    [anon_sym_object] = ACTIONS(2059),
    [anon_sym_inherit] = ACTIONS(2059),
    [anon_sym_method] = ACTIONS(2059),
    [anon_sym_initializer] = ACTIONS(2059),
    [anon_sym_AMP] = ACTIONS(2059),
    [anon_sym_POUND] = ACTIONS(2059),
    [anon_sym_COLON_COLON] = ACTIONS(2061),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2061),
    [anon_sym_then] = ACTIONS(2059),
    [anon_sym_else] = ACTIONS(2059),
    [anon_sym_do] = ACTIONS(2059),
    [anon_sym_done] = ACTIONS(2059),
    [anon_sym_to] = ACTIONS(2059),
    [anon_sym_downto] = ACTIONS(2059),
    [anon_sym_new] = ACTIONS(2059),
    [anon_sym_LBRACE_LT] = ACTIONS(2061),
    [anon_sym_GT_RBRACE] = ACTIONS(2061),
    [anon_sym_begin] = ACTIONS(2059),
    [sym_ocamlyacc_value] = ACTIONS(2061),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2059),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2061),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2061),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2061),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2059),
    [anon_sym_SQUOTE] = ACTIONS(2061),
    [anon_sym_DQUOTE] = ACTIONS(2061),
    [sym_prefix_operator] = ACTIONS(2061),
    [aux_sym_sign_operator_token1] = ACTIONS(2059),
    [sym_hash_operator] = ACTIONS(2061),
    [anon_sym_lsl] = ACTIONS(2059),
    [anon_sym_lsr] = ACTIONS(2059),
    [anon_sym_asr] = ACTIONS(2059),
    [aux_sym_pow_operator_token1] = ACTIONS(2061),
    [anon_sym_mod] = ACTIONS(2059),
    [anon_sym_land] = ACTIONS(2059),
    [anon_sym_lor] = ACTIONS(2059),
    [anon_sym_lxor] = ACTIONS(2059),
    [aux_sym_mult_operator_token1] = ACTIONS(2059),
    [aux_sym_add_operator_token1] = ACTIONS(2059),
    [sym_concat_operator] = ACTIONS(2061),
    [sym_rel_operator] = ACTIONS(2059),
    [anon_sym_AMP_AMP] = ACTIONS(2059),
    [anon_sym_or] = ACTIONS(2059),
    [anon_sym_PIPE_PIPE] = ACTIONS(2059),
    [sym_assign_operator] = ACTIONS(2061),
    [sym__uppercase_identifier] = ACTIONS(2061),
    [aux_sym_tag_token1] = ACTIONS(2061),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1083)] = {
    [sym_attribute] = STATE(1083),
    [sym__lowercase_identifier] = ACTIONS(2067),
    [anon_sym_TILDE] = ACTIONS(2067),
    [anon_sym_QMARK] = ACTIONS(2067),
    [anon_sym_LPAREN] = ACTIONS(2069),
    [anon_sym_RPAREN] = ACTIONS(2069),
    [anon_sym_COMMA] = ACTIONS(2069),
    [anon_sym_PIPE] = ACTIONS(2067),
    [anon_sym_COLON2] = ACTIONS(2067),
    [anon_sym_DOT] = ACTIONS(2069),
    [anon_sym_DASH_GT] = ACTIONS(2067),
    [anon_sym_LBRACK] = ACTIONS(2067),
    [anon_sym_RBRACK] = ACTIONS(2069),
    [anon_sym_true] = ACTIONS(2067),
    [anon_sym_false] = ACTIONS(2067),
    [anon_sym_LBRACE] = ACTIONS(2067),
    [anon_sym_SEMI] = ACTIONS(2069),
    [anon_sym_RBRACE] = ACTIONS(2069),
    [anon_sym_constraint] = ACTIONS(2067),
    [anon_sym_val] = ACTIONS(2067),
    [anon_sym_end] = ACTIONS(2067),
    [anon_sym_with] = ACTIONS(2067),
    [anon_sym_COLON_GT] = ACTIONS(2069),
    [anon_sym_object] = ACTIONS(2067),
    [anon_sym_inherit] = ACTIONS(2067),
    [anon_sym_method] = ACTIONS(2067),
    [anon_sym_initializer] = ACTIONS(2067),
    [anon_sym_AMP] = ACTIONS(2067),
    [anon_sym_POUND] = ACTIONS(2067),
    [anon_sym_COLON_COLON] = ACTIONS(2069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2069),
    [anon_sym_then] = ACTIONS(2067),
    [anon_sym_else] = ACTIONS(2067),
    [anon_sym_do] = ACTIONS(2067),
    [anon_sym_done] = ACTIONS(2067),
    [anon_sym_to] = ACTIONS(2067),
    [anon_sym_downto] = ACTIONS(2067),
    [anon_sym_new] = ACTIONS(2067),
    [anon_sym_LBRACE_LT] = ACTIONS(2069),
    [anon_sym_GT_RBRACE] = ACTIONS(2069),
    [anon_sym_begin] = ACTIONS(2067),
    [sym_ocamlyacc_value] = ACTIONS(2069),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2067),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2069),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2067),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2067),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2069),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2069),
    [aux_sym_number_token1] = ACTIONS(2069),
    [aux_sym_signed_number_token1] = ACTIONS(2067),
    [anon_sym_SQUOTE] = ACTIONS(2069),
    [anon_sym_DQUOTE] = ACTIONS(2069),
    [sym_prefix_operator] = ACTIONS(2069),
    [aux_sym_sign_operator_token1] = ACTIONS(2067),
    [sym_hash_operator] = ACTIONS(2069),
    [anon_sym_lsl] = ACTIONS(2067),
    [anon_sym_lsr] = ACTIONS(2067),
    [anon_sym_asr] = ACTIONS(2067),
    [aux_sym_pow_operator_token1] = ACTIONS(2069),
    [anon_sym_mod] = ACTIONS(2067),
    [anon_sym_land] = ACTIONS(2067),
    [anon_sym_lor] = ACTIONS(2067),
    [anon_sym_lxor] = ACTIONS(2067),
    [aux_sym_mult_operator_token1] = ACTIONS(2067),
    [aux_sym_add_operator_token1] = ACTIONS(2067),
    [sym_concat_operator] = ACTIONS(2069),
    [sym_rel_operator] = ACTIONS(2067),
    [anon_sym_AMP_AMP] = ACTIONS(2067),
    [anon_sym_or] = ACTIONS(2067),
    [anon_sym_PIPE_PIPE] = ACTIONS(2067),
    [sym_assign_operator] = ACTIONS(2069),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_tag_token1] = ACTIONS(2069),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1084)] = {
    [sym_attribute] = STATE(1084),
    [sym__lowercase_identifier] = ACTIONS(2087),
    [anon_sym_TILDE] = ACTIONS(2087),
    [anon_sym_QMARK] = ACTIONS(2087),
    [anon_sym_LPAREN] = ACTIONS(2089),
    [anon_sym_RPAREN] = ACTIONS(2089),
    [anon_sym_COMMA] = ACTIONS(2089),
    [anon_sym_PIPE] = ACTIONS(2087),
    [anon_sym_COLON2] = ACTIONS(2087),
    [anon_sym_DOT] = ACTIONS(2089),
    [anon_sym_DASH_GT] = ACTIONS(2087),
    [anon_sym_LBRACK] = ACTIONS(2087),
    [anon_sym_RBRACK] = ACTIONS(2089),
    [anon_sym_true] = ACTIONS(2087),
    [anon_sym_false] = ACTIONS(2087),
    [anon_sym_LBRACE] = ACTIONS(2087),
    [anon_sym_SEMI] = ACTIONS(2089),
    [anon_sym_RBRACE] = ACTIONS(2089),
    [anon_sym_constraint] = ACTIONS(2087),
    [anon_sym_val] = ACTIONS(2087),
    [anon_sym_end] = ACTIONS(2087),
    [anon_sym_with] = ACTIONS(2087),
    [anon_sym_COLON_GT] = ACTIONS(2089),
    [anon_sym_object] = ACTIONS(2087),
    [anon_sym_inherit] = ACTIONS(2087),
    [anon_sym_method] = ACTIONS(2087),
    [anon_sym_initializer] = ACTIONS(2087),
    [anon_sym_AMP] = ACTIONS(2087),
    [anon_sym_POUND] = ACTIONS(2087),
    [anon_sym_COLON_COLON] = ACTIONS(2089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2089),
    [anon_sym_then] = ACTIONS(2087),
    [anon_sym_else] = ACTIONS(2087),
    [anon_sym_do] = ACTIONS(2087),
    [anon_sym_done] = ACTIONS(2087),
    [anon_sym_to] = ACTIONS(2087),
    [anon_sym_downto] = ACTIONS(2087),
    [anon_sym_new] = ACTIONS(2087),
    [anon_sym_LBRACE_LT] = ACTIONS(2089),
    [anon_sym_GT_RBRACE] = ACTIONS(2089),
    [anon_sym_begin] = ACTIONS(2087),
    [sym_ocamlyacc_value] = ACTIONS(2089),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2087),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2089),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2087),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2087),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2089),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2089),
    [aux_sym_number_token1] = ACTIONS(2089),
    [aux_sym_signed_number_token1] = ACTIONS(2087),
    [anon_sym_SQUOTE] = ACTIONS(2089),
    [anon_sym_DQUOTE] = ACTIONS(2089),
    [sym_prefix_operator] = ACTIONS(2089),
    [aux_sym_sign_operator_token1] = ACTIONS(2087),
    [sym_hash_operator] = ACTIONS(2089),
    [anon_sym_lsl] = ACTIONS(2087),
    [anon_sym_lsr] = ACTIONS(2087),
    [anon_sym_asr] = ACTIONS(2087),
    [aux_sym_pow_operator_token1] = ACTIONS(2089),
    [anon_sym_mod] = ACTIONS(2087),
    [anon_sym_land] = ACTIONS(2087),
    [anon_sym_lor] = ACTIONS(2087),
    [anon_sym_lxor] = ACTIONS(2087),
    [aux_sym_mult_operator_token1] = ACTIONS(2087),
    [aux_sym_add_operator_token1] = ACTIONS(2087),
    [sym_concat_operator] = ACTIONS(2089),
    [sym_rel_operator] = ACTIONS(2087),
    [anon_sym_AMP_AMP] = ACTIONS(2087),
    [anon_sym_or] = ACTIONS(2087),
    [anon_sym_PIPE_PIPE] = ACTIONS(2087),
    [sym_assign_operator] = ACTIONS(2089),
    [sym__uppercase_identifier] = ACTIONS(2089),
    [aux_sym_tag_token1] = ACTIONS(2089),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1085)] = {
    [sym_attribute] = STATE(1085),
    [sym__lowercase_identifier] = ACTIONS(2091),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_RPAREN] = ACTIONS(2093),
    [anon_sym_COMMA] = ACTIONS(2093),
    [anon_sym_PIPE] = ACTIONS(2091),
    [anon_sym_COLON2] = ACTIONS(2091),
    [anon_sym_DOT] = ACTIONS(2093),
    [anon_sym_DASH_GT] = ACTIONS(2091),
    [anon_sym_LBRACK] = ACTIONS(2091),
    [anon_sym_RBRACK] = ACTIONS(2093),
    [anon_sym_true] = ACTIONS(2091),
    [anon_sym_false] = ACTIONS(2091),
    [anon_sym_LBRACE] = ACTIONS(2091),
    [anon_sym_SEMI] = ACTIONS(2093),
    [anon_sym_RBRACE] = ACTIONS(2093),
    [anon_sym_constraint] = ACTIONS(2091),
    [anon_sym_val] = ACTIONS(2091),
    [anon_sym_end] = ACTIONS(2091),
    [anon_sym_with] = ACTIONS(2091),
    [anon_sym_COLON_GT] = ACTIONS(2093),
    [anon_sym_object] = ACTIONS(2091),
    [anon_sym_inherit] = ACTIONS(2091),
    [anon_sym_method] = ACTIONS(2091),
    [anon_sym_initializer] = ACTIONS(2091),
    [anon_sym_AMP] = ACTIONS(2091),
    [anon_sym_POUND] = ACTIONS(2091),
    [anon_sym_COLON_COLON] = ACTIONS(2093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2093),
    [anon_sym_then] = ACTIONS(2091),
    [anon_sym_else] = ACTIONS(2091),
    [anon_sym_do] = ACTIONS(2091),
    [anon_sym_done] = ACTIONS(2091),
    [anon_sym_to] = ACTIONS(2091),
    [anon_sym_downto] = ACTIONS(2091),
    [anon_sym_new] = ACTIONS(2091),
    [anon_sym_LBRACE_LT] = ACTIONS(2093),
    [anon_sym_GT_RBRACE] = ACTIONS(2093),
    [anon_sym_begin] = ACTIONS(2091),
    [sym_ocamlyacc_value] = ACTIONS(2093),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2091),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2093),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2091),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2091),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2093),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2093),
    [aux_sym_number_token1] = ACTIONS(2093),
    [aux_sym_signed_number_token1] = ACTIONS(2091),
    [anon_sym_SQUOTE] = ACTIONS(2093),
    [anon_sym_DQUOTE] = ACTIONS(2093),
    [sym_prefix_operator] = ACTIONS(2093),
    [aux_sym_sign_operator_token1] = ACTIONS(2091),
    [sym_hash_operator] = ACTIONS(2093),
    [anon_sym_lsl] = ACTIONS(2091),
    [anon_sym_lsr] = ACTIONS(2091),
    [anon_sym_asr] = ACTIONS(2091),
    [aux_sym_pow_operator_token1] = ACTIONS(2093),
    [anon_sym_mod] = ACTIONS(2091),
    [anon_sym_land] = ACTIONS(2091),
    [anon_sym_lor] = ACTIONS(2091),
    [anon_sym_lxor] = ACTIONS(2091),
    [aux_sym_mult_operator_token1] = ACTIONS(2091),
    [aux_sym_add_operator_token1] = ACTIONS(2091),
    [sym_concat_operator] = ACTIONS(2093),
    [sym_rel_operator] = ACTIONS(2091),
    [anon_sym_AMP_AMP] = ACTIONS(2091),
    [anon_sym_or] = ACTIONS(2091),
    [anon_sym_PIPE_PIPE] = ACTIONS(2091),
    [sym_assign_operator] = ACTIONS(2093),
    [sym__uppercase_identifier] = ACTIONS(2093),
    [aux_sym_tag_token1] = ACTIONS(2093),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1086)] = {
    [sym_attribute] = STATE(1086),
    [sym__lowercase_identifier] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(2095),
    [anon_sym_QMARK] = ACTIONS(2095),
    [anon_sym_LPAREN] = ACTIONS(2097),
    [anon_sym_RPAREN] = ACTIONS(2097),
    [anon_sym_COMMA] = ACTIONS(2097),
    [anon_sym_PIPE] = ACTIONS(2095),
    [anon_sym_COLON2] = ACTIONS(2095),
    [anon_sym_DOT] = ACTIONS(2097),
    [anon_sym_DASH_GT] = ACTIONS(2095),
    [anon_sym_LBRACK] = ACTIONS(2095),
    [anon_sym_RBRACK] = ACTIONS(2097),
    [anon_sym_true] = ACTIONS(2095),
    [anon_sym_false] = ACTIONS(2095),
    [anon_sym_LBRACE] = ACTIONS(2095),
    [anon_sym_SEMI] = ACTIONS(2097),
    [anon_sym_RBRACE] = ACTIONS(2097),
    [anon_sym_constraint] = ACTIONS(2095),
    [anon_sym_val] = ACTIONS(2095),
    [anon_sym_end] = ACTIONS(2095),
    [anon_sym_with] = ACTIONS(2095),
    [anon_sym_COLON_GT] = ACTIONS(2097),
    [anon_sym_object] = ACTIONS(2095),
    [anon_sym_inherit] = ACTIONS(2095),
    [anon_sym_method] = ACTIONS(2095),
    [anon_sym_initializer] = ACTIONS(2095),
    [anon_sym_AMP] = ACTIONS(2095),
    [anon_sym_POUND] = ACTIONS(2095),
    [anon_sym_COLON_COLON] = ACTIONS(2097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2097),
    [anon_sym_then] = ACTIONS(2095),
    [anon_sym_else] = ACTIONS(2095),
    [anon_sym_do] = ACTIONS(2095),
    [anon_sym_done] = ACTIONS(2095),
    [anon_sym_to] = ACTIONS(2095),
    [anon_sym_downto] = ACTIONS(2095),
    [anon_sym_new] = ACTIONS(2095),
    [anon_sym_LBRACE_LT] = ACTIONS(2097),
    [anon_sym_GT_RBRACE] = ACTIONS(2097),
    [anon_sym_begin] = ACTIONS(2095),
    [sym_ocamlyacc_value] = ACTIONS(2097),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2095),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2097),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2095),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2095),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2097),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2097),
    [aux_sym_number_token1] = ACTIONS(2097),
    [aux_sym_signed_number_token1] = ACTIONS(2095),
    [anon_sym_SQUOTE] = ACTIONS(2097),
    [anon_sym_DQUOTE] = ACTIONS(2097),
    [sym_prefix_operator] = ACTIONS(2097),
    [aux_sym_sign_operator_token1] = ACTIONS(2095),
    [sym_hash_operator] = ACTIONS(2097),
    [anon_sym_lsl] = ACTIONS(2095),
    [anon_sym_lsr] = ACTIONS(2095),
    [anon_sym_asr] = ACTIONS(2095),
    [aux_sym_pow_operator_token1] = ACTIONS(2097),
    [anon_sym_mod] = ACTIONS(2095),
    [anon_sym_land] = ACTIONS(2095),
    [anon_sym_lor] = ACTIONS(2095),
    [anon_sym_lxor] = ACTIONS(2095),
    [aux_sym_mult_operator_token1] = ACTIONS(2095),
    [aux_sym_add_operator_token1] = ACTIONS(2095),
    [sym_concat_operator] = ACTIONS(2097),
    [sym_rel_operator] = ACTIONS(2095),
    [anon_sym_AMP_AMP] = ACTIONS(2095),
    [anon_sym_or] = ACTIONS(2095),
    [anon_sym_PIPE_PIPE] = ACTIONS(2095),
    [sym_assign_operator] = ACTIONS(2097),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_tag_token1] = ACTIONS(2097),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1087)] = {
    [sym_attribute] = STATE(1087),
    [sym__lowercase_identifier] = ACTIONS(2099),
    [anon_sym_TILDE] = ACTIONS(2099),
    [anon_sym_QMARK] = ACTIONS(2099),
    [anon_sym_LPAREN] = ACTIONS(2101),
    [anon_sym_RPAREN] = ACTIONS(2101),
    [anon_sym_COMMA] = ACTIONS(2101),
    [anon_sym_PIPE] = ACTIONS(2099),
    [anon_sym_COLON2] = ACTIONS(2099),
    [anon_sym_DOT] = ACTIONS(2101),
    [anon_sym_DASH_GT] = ACTIONS(2099),
    [anon_sym_LBRACK] = ACTIONS(2099),
    [anon_sym_RBRACK] = ACTIONS(2101),
    [anon_sym_true] = ACTIONS(2099),
    [anon_sym_false] = ACTIONS(2099),
    [anon_sym_LBRACE] = ACTIONS(2099),
    [anon_sym_SEMI] = ACTIONS(2101),
    [anon_sym_RBRACE] = ACTIONS(2101),
    [anon_sym_constraint] = ACTIONS(2099),
    [anon_sym_val] = ACTIONS(2099),
    [anon_sym_end] = ACTIONS(2099),
    [anon_sym_with] = ACTIONS(2099),
    [anon_sym_COLON_GT] = ACTIONS(2101),
    [anon_sym_object] = ACTIONS(2099),
    [anon_sym_inherit] = ACTIONS(2099),
    [anon_sym_method] = ACTIONS(2099),
    [anon_sym_initializer] = ACTIONS(2099),
    [anon_sym_AMP] = ACTIONS(2099),
    [anon_sym_POUND] = ACTIONS(2099),
    [anon_sym_COLON_COLON] = ACTIONS(2101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2101),
    [anon_sym_then] = ACTIONS(2099),
    [anon_sym_else] = ACTIONS(2099),
    [anon_sym_do] = ACTIONS(2099),
    [anon_sym_done] = ACTIONS(2099),
    [anon_sym_to] = ACTIONS(2099),
    [anon_sym_downto] = ACTIONS(2099),
    [anon_sym_new] = ACTIONS(2099),
    [anon_sym_LBRACE_LT] = ACTIONS(2101),
    [anon_sym_GT_RBRACE] = ACTIONS(2101),
    [anon_sym_begin] = ACTIONS(2099),
    [sym_ocamlyacc_value] = ACTIONS(2101),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2099),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2101),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2099),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2099),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2101),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2101),
    [aux_sym_number_token1] = ACTIONS(2101),
    [aux_sym_signed_number_token1] = ACTIONS(2099),
    [anon_sym_SQUOTE] = ACTIONS(2101),
    [anon_sym_DQUOTE] = ACTIONS(2101),
    [sym_prefix_operator] = ACTIONS(2101),
    [aux_sym_sign_operator_token1] = ACTIONS(2099),
    [sym_hash_operator] = ACTIONS(2101),
    [anon_sym_lsl] = ACTIONS(2099),
    [anon_sym_lsr] = ACTIONS(2099),
    [anon_sym_asr] = ACTIONS(2099),
    [aux_sym_pow_operator_token1] = ACTIONS(2101),
    [anon_sym_mod] = ACTIONS(2099),
    [anon_sym_land] = ACTIONS(2099),
    [anon_sym_lor] = ACTIONS(2099),
    [anon_sym_lxor] = ACTIONS(2099),
    [aux_sym_mult_operator_token1] = ACTIONS(2099),
    [aux_sym_add_operator_token1] = ACTIONS(2099),
    [sym_concat_operator] = ACTIONS(2101),
    [sym_rel_operator] = ACTIONS(2099),
    [anon_sym_AMP_AMP] = ACTIONS(2099),
    [anon_sym_or] = ACTIONS(2099),
    [anon_sym_PIPE_PIPE] = ACTIONS(2099),
    [sym_assign_operator] = ACTIONS(2101),
    [sym__uppercase_identifier] = ACTIONS(2101),
    [aux_sym_tag_token1] = ACTIONS(2101),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1088)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4395),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1088),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4340),
    [aux_sym_let_binding_repeat1] = STATE(1097),
    [ts_builtin_sym_end] = ACTIONS(2379),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2379),
    [anon_sym_let] = ACTIONS(2381),
    [anon_sym_and] = ACTIONS(2381),
    [anon_sym_EQ] = ACTIONS(2431),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2381),
    [anon_sym_type] = ACTIONS(2381),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2381),
    [anon_sym_module] = ACTIONS(2381),
    [anon_sym_open] = ACTIONS(2381),
    [anon_sym_include] = ACTIONS(2381),
    [anon_sym_class] = ACTIONS(2381),
    [anon_sym_val] = ACTIONS(2381),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2379),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2379),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2379),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2379),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2381),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1089)] = {
    [sym__parenthesized_abstract_type] = STATE(1718),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5036),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1089),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_EQ] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2437),
    [anon_sym_RPAREN] = ACTIONS(2433),
    [anon_sym_COMMA] = ACTIONS(2433),
    [anon_sym_PIPE] = ACTIONS(2433),
    [anon_sym_COLON2] = ACTIONS(2439),
    [anon_sym_DASH_GT] = ACTIONS(2433),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_RBRACK] = ACTIONS(2433),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_as] = ACTIONS(2439),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(2433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_when] = ACTIONS(2439),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2443),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1090)] = {
    [sym_attribute] = STATE(1090),
    [sym__lowercase_identifier] = ACTIONS(2149),
    [anon_sym_TILDE] = ACTIONS(2149),
    [anon_sym_QMARK] = ACTIONS(2149),
    [anon_sym_LPAREN] = ACTIONS(2151),
    [anon_sym_RPAREN] = ACTIONS(2151),
    [anon_sym_COMMA] = ACTIONS(2151),
    [anon_sym_PIPE] = ACTIONS(2149),
    [anon_sym_COLON2] = ACTIONS(2149),
    [anon_sym_DOT] = ACTIONS(2151),
    [anon_sym_DASH_GT] = ACTIONS(2149),
    [anon_sym_LBRACK] = ACTIONS(2149),
    [anon_sym_RBRACK] = ACTIONS(2151),
    [anon_sym_true] = ACTIONS(2149),
    [anon_sym_false] = ACTIONS(2149),
    [anon_sym_LBRACE] = ACTIONS(2149),
    [anon_sym_SEMI] = ACTIONS(2151),
    [anon_sym_RBRACE] = ACTIONS(2151),
    [anon_sym_constraint] = ACTIONS(2149),
    [anon_sym_val] = ACTIONS(2149),
    [anon_sym_end] = ACTIONS(2149),
    [anon_sym_with] = ACTIONS(2149),
    [anon_sym_COLON_GT] = ACTIONS(2151),
    [anon_sym_object] = ACTIONS(2149),
    [anon_sym_inherit] = ACTIONS(2149),
    [anon_sym_method] = ACTIONS(2149),
    [anon_sym_initializer] = ACTIONS(2149),
    [anon_sym_AMP] = ACTIONS(2149),
    [anon_sym_POUND] = ACTIONS(2149),
    [anon_sym_COLON_COLON] = ACTIONS(2151),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2151),
    [anon_sym_then] = ACTIONS(2149),
    [anon_sym_else] = ACTIONS(2149),
    [anon_sym_do] = ACTIONS(2149),
    [anon_sym_done] = ACTIONS(2149),
    [anon_sym_to] = ACTIONS(2149),
    [anon_sym_downto] = ACTIONS(2149),
    [anon_sym_new] = ACTIONS(2149),
    [anon_sym_LBRACE_LT] = ACTIONS(2151),
    [anon_sym_GT_RBRACE] = ACTIONS(2151),
    [anon_sym_begin] = ACTIONS(2149),
    [sym_ocamlyacc_value] = ACTIONS(2151),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2149),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2151),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2149),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2151),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2151),
    [aux_sym_number_token1] = ACTIONS(2151),
    [aux_sym_signed_number_token1] = ACTIONS(2149),
    [anon_sym_SQUOTE] = ACTIONS(2151),
    [anon_sym_DQUOTE] = ACTIONS(2151),
    [sym_prefix_operator] = ACTIONS(2151),
    [aux_sym_sign_operator_token1] = ACTIONS(2149),
    [sym_hash_operator] = ACTIONS(2151),
    [anon_sym_lsl] = ACTIONS(2149),
    [anon_sym_lsr] = ACTIONS(2149),
    [anon_sym_asr] = ACTIONS(2149),
    [aux_sym_pow_operator_token1] = ACTIONS(2151),
    [anon_sym_mod] = ACTIONS(2149),
    [anon_sym_land] = ACTIONS(2149),
    [anon_sym_lor] = ACTIONS(2149),
    [anon_sym_lxor] = ACTIONS(2149),
    [aux_sym_mult_operator_token1] = ACTIONS(2149),
    [aux_sym_add_operator_token1] = ACTIONS(2149),
    [sym_concat_operator] = ACTIONS(2151),
    [sym_rel_operator] = ACTIONS(2149),
    [anon_sym_AMP_AMP] = ACTIONS(2149),
    [anon_sym_or] = ACTIONS(2149),
    [anon_sym_PIPE_PIPE] = ACTIONS(2149),
    [sym_assign_operator] = ACTIONS(2151),
    [sym__uppercase_identifier] = ACTIONS(2151),
    [aux_sym_tag_token1] = ACTIONS(2151),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1091)] = {
    [sym_attribute] = STATE(1091),
    [sym__lowercase_identifier] = ACTIONS(2195),
    [anon_sym_TILDE] = ACTIONS(2195),
    [anon_sym_QMARK] = ACTIONS(2195),
    [anon_sym_LPAREN] = ACTIONS(2197),
    [anon_sym_RPAREN] = ACTIONS(2197),
    [anon_sym_COMMA] = ACTIONS(2197),
    [anon_sym_PIPE] = ACTIONS(2195),
    [anon_sym_COLON2] = ACTIONS(2195),
    [anon_sym_DOT] = ACTIONS(2197),
    [anon_sym_DASH_GT] = ACTIONS(2195),
    [anon_sym_LBRACK] = ACTIONS(2195),
    [anon_sym_RBRACK] = ACTIONS(2197),
    [anon_sym_true] = ACTIONS(2195),
    [anon_sym_false] = ACTIONS(2195),
    [anon_sym_LBRACE] = ACTIONS(2195),
    [anon_sym_SEMI] = ACTIONS(2197),
    [anon_sym_RBRACE] = ACTIONS(2197),
    [anon_sym_constraint] = ACTIONS(2195),
    [anon_sym_val] = ACTIONS(2195),
    [anon_sym_end] = ACTIONS(2195),
    [anon_sym_with] = ACTIONS(2195),
    [anon_sym_COLON_GT] = ACTIONS(2197),
    [anon_sym_object] = ACTIONS(2195),
    [anon_sym_inherit] = ACTIONS(2195),
    [anon_sym_method] = ACTIONS(2195),
    [anon_sym_initializer] = ACTIONS(2195),
    [anon_sym_AMP] = ACTIONS(2195),
    [anon_sym_POUND] = ACTIONS(2195),
    [anon_sym_COLON_COLON] = ACTIONS(2197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2197),
    [anon_sym_then] = ACTIONS(2195),
    [anon_sym_else] = ACTIONS(2195),
    [anon_sym_do] = ACTIONS(2195),
    [anon_sym_done] = ACTIONS(2195),
    [anon_sym_to] = ACTIONS(2195),
    [anon_sym_downto] = ACTIONS(2195),
    [anon_sym_new] = ACTIONS(2195),
    [anon_sym_LBRACE_LT] = ACTIONS(2197),
    [anon_sym_GT_RBRACE] = ACTIONS(2197),
    [anon_sym_begin] = ACTIONS(2195),
    [sym_ocamlyacc_value] = ACTIONS(2197),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2195),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2197),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2195),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2195),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2197),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2197),
    [aux_sym_number_token1] = ACTIONS(2197),
    [aux_sym_signed_number_token1] = ACTIONS(2195),
    [anon_sym_SQUOTE] = ACTIONS(2197),
    [anon_sym_DQUOTE] = ACTIONS(2197),
    [sym_prefix_operator] = ACTIONS(2197),
    [aux_sym_sign_operator_token1] = ACTIONS(2195),
    [sym_hash_operator] = ACTIONS(2197),
    [anon_sym_lsl] = ACTIONS(2195),
    [anon_sym_lsr] = ACTIONS(2195),
    [anon_sym_asr] = ACTIONS(2195),
    [aux_sym_pow_operator_token1] = ACTIONS(2197),
    [anon_sym_mod] = ACTIONS(2195),
    [anon_sym_land] = ACTIONS(2195),
    [anon_sym_lor] = ACTIONS(2195),
    [anon_sym_lxor] = ACTIONS(2195),
    [aux_sym_mult_operator_token1] = ACTIONS(2195),
    [aux_sym_add_operator_token1] = ACTIONS(2195),
    [sym_concat_operator] = ACTIONS(2197),
    [sym_rel_operator] = ACTIONS(2195),
    [anon_sym_AMP_AMP] = ACTIONS(2195),
    [anon_sym_or] = ACTIONS(2195),
    [anon_sym_PIPE_PIPE] = ACTIONS(2195),
    [sym_assign_operator] = ACTIONS(2197),
    [sym__uppercase_identifier] = ACTIONS(2197),
    [aux_sym_tag_token1] = ACTIONS(2197),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1092)] = {
    [sym_attribute] = STATE(1092),
    [sym__lowercase_identifier] = ACTIONS(2199),
    [anon_sym_TILDE] = ACTIONS(2199),
    [anon_sym_QMARK] = ACTIONS(2199),
    [anon_sym_LPAREN] = ACTIONS(2201),
    [anon_sym_RPAREN] = ACTIONS(2201),
    [anon_sym_COMMA] = ACTIONS(2201),
    [anon_sym_PIPE] = ACTIONS(2199),
    [anon_sym_COLON2] = ACTIONS(2199),
    [anon_sym_DOT] = ACTIONS(2201),
    [anon_sym_DASH_GT] = ACTIONS(2199),
    [anon_sym_LBRACK] = ACTIONS(2199),
    [anon_sym_RBRACK] = ACTIONS(2201),
    [anon_sym_true] = ACTIONS(2199),
    [anon_sym_false] = ACTIONS(2199),
    [anon_sym_LBRACE] = ACTIONS(2199),
    [anon_sym_SEMI] = ACTIONS(2201),
    [anon_sym_RBRACE] = ACTIONS(2201),
    [anon_sym_constraint] = ACTIONS(2199),
    [anon_sym_val] = ACTIONS(2199),
    [anon_sym_end] = ACTIONS(2199),
    [anon_sym_with] = ACTIONS(2199),
    [anon_sym_COLON_GT] = ACTIONS(2201),
    [anon_sym_object] = ACTIONS(2199),
    [anon_sym_inherit] = ACTIONS(2199),
    [anon_sym_method] = ACTIONS(2199),
    [anon_sym_initializer] = ACTIONS(2199),
    [anon_sym_AMP] = ACTIONS(2199),
    [anon_sym_POUND] = ACTIONS(2199),
    [anon_sym_COLON_COLON] = ACTIONS(2201),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2201),
    [anon_sym_then] = ACTIONS(2199),
    [anon_sym_else] = ACTIONS(2199),
    [anon_sym_do] = ACTIONS(2199),
    [anon_sym_done] = ACTIONS(2199),
    [anon_sym_to] = ACTIONS(2199),
    [anon_sym_downto] = ACTIONS(2199),
    [anon_sym_new] = ACTIONS(2199),
    [anon_sym_LBRACE_LT] = ACTIONS(2201),
    [anon_sym_GT_RBRACE] = ACTIONS(2201),
    [anon_sym_begin] = ACTIONS(2199),
    [sym_ocamlyacc_value] = ACTIONS(2201),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2199),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2201),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2199),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2199),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2201),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2201),
    [aux_sym_number_token1] = ACTIONS(2201),
    [aux_sym_signed_number_token1] = ACTIONS(2199),
    [anon_sym_SQUOTE] = ACTIONS(2201),
    [anon_sym_DQUOTE] = ACTIONS(2201),
    [sym_prefix_operator] = ACTIONS(2201),
    [aux_sym_sign_operator_token1] = ACTIONS(2199),
    [sym_hash_operator] = ACTIONS(2201),
    [anon_sym_lsl] = ACTIONS(2199),
    [anon_sym_lsr] = ACTIONS(2199),
    [anon_sym_asr] = ACTIONS(2199),
    [aux_sym_pow_operator_token1] = ACTIONS(2201),
    [anon_sym_mod] = ACTIONS(2199),
    [anon_sym_land] = ACTIONS(2199),
    [anon_sym_lor] = ACTIONS(2199),
    [anon_sym_lxor] = ACTIONS(2199),
    [aux_sym_mult_operator_token1] = ACTIONS(2199),
    [aux_sym_add_operator_token1] = ACTIONS(2199),
    [sym_concat_operator] = ACTIONS(2201),
    [sym_rel_operator] = ACTIONS(2199),
    [anon_sym_AMP_AMP] = ACTIONS(2199),
    [anon_sym_or] = ACTIONS(2199),
    [anon_sym_PIPE_PIPE] = ACTIONS(2199),
    [sym_assign_operator] = ACTIONS(2201),
    [sym__uppercase_identifier] = ACTIONS(2201),
    [aux_sym_tag_token1] = ACTIONS(2201),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1093)] = {
    [sym_attribute] = STATE(1093),
    [sym__lowercase_identifier] = ACTIONS(2203),
    [anon_sym_TILDE] = ACTIONS(2203),
    [anon_sym_QMARK] = ACTIONS(2203),
    [anon_sym_LPAREN] = ACTIONS(2205),
    [anon_sym_RPAREN] = ACTIONS(2205),
    [anon_sym_COMMA] = ACTIONS(2205),
    [anon_sym_PIPE] = ACTIONS(2203),
    [anon_sym_COLON2] = ACTIONS(2203),
    [anon_sym_DOT] = ACTIONS(2205),
    [anon_sym_DASH_GT] = ACTIONS(2203),
    [anon_sym_LBRACK] = ACTIONS(2203),
    [anon_sym_RBRACK] = ACTIONS(2205),
    [anon_sym_true] = ACTIONS(2203),
    [anon_sym_false] = ACTIONS(2203),
    [anon_sym_LBRACE] = ACTIONS(2203),
    [anon_sym_SEMI] = ACTIONS(2205),
    [anon_sym_RBRACE] = ACTIONS(2205),
    [anon_sym_constraint] = ACTIONS(2203),
    [anon_sym_val] = ACTIONS(2203),
    [anon_sym_end] = ACTIONS(2203),
    [anon_sym_with] = ACTIONS(2203),
    [anon_sym_COLON_GT] = ACTIONS(2205),
    [anon_sym_object] = ACTIONS(2203),
    [anon_sym_inherit] = ACTIONS(2203),
    [anon_sym_method] = ACTIONS(2203),
    [anon_sym_initializer] = ACTIONS(2203),
    [anon_sym_AMP] = ACTIONS(2203),
    [anon_sym_POUND] = ACTIONS(2203),
    [anon_sym_COLON_COLON] = ACTIONS(2205),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2205),
    [anon_sym_then] = ACTIONS(2203),
    [anon_sym_else] = ACTIONS(2203),
    [anon_sym_do] = ACTIONS(2203),
    [anon_sym_done] = ACTIONS(2203),
    [anon_sym_to] = ACTIONS(2203),
    [anon_sym_downto] = ACTIONS(2203),
    [anon_sym_new] = ACTIONS(2203),
    [anon_sym_LBRACE_LT] = ACTIONS(2205),
    [anon_sym_GT_RBRACE] = ACTIONS(2205),
    [anon_sym_begin] = ACTIONS(2203),
    [sym_ocamlyacc_value] = ACTIONS(2205),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2203),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2205),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2203),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2203),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2205),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2205),
    [aux_sym_number_token1] = ACTIONS(2205),
    [aux_sym_signed_number_token1] = ACTIONS(2203),
    [anon_sym_SQUOTE] = ACTIONS(2205),
    [anon_sym_DQUOTE] = ACTIONS(2205),
    [sym_prefix_operator] = ACTIONS(2205),
    [aux_sym_sign_operator_token1] = ACTIONS(2203),
    [sym_hash_operator] = ACTIONS(2205),
    [anon_sym_lsl] = ACTIONS(2203),
    [anon_sym_lsr] = ACTIONS(2203),
    [anon_sym_asr] = ACTIONS(2203),
    [aux_sym_pow_operator_token1] = ACTIONS(2205),
    [anon_sym_mod] = ACTIONS(2203),
    [anon_sym_land] = ACTIONS(2203),
    [anon_sym_lor] = ACTIONS(2203),
    [anon_sym_lxor] = ACTIONS(2203),
    [aux_sym_mult_operator_token1] = ACTIONS(2203),
    [aux_sym_add_operator_token1] = ACTIONS(2203),
    [sym_concat_operator] = ACTIONS(2205),
    [sym_rel_operator] = ACTIONS(2203),
    [anon_sym_AMP_AMP] = ACTIONS(2203),
    [anon_sym_or] = ACTIONS(2203),
    [anon_sym_PIPE_PIPE] = ACTIONS(2203),
    [sym_assign_operator] = ACTIONS(2205),
    [sym__uppercase_identifier] = ACTIONS(2205),
    [aux_sym_tag_token1] = ACTIONS(2205),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1094)] = {
    [sym_attribute] = STATE(1094),
    [sym__lowercase_identifier] = ACTIONS(2031),
    [anon_sym_TILDE] = ACTIONS(2031),
    [anon_sym_QMARK] = ACTIONS(2031),
    [anon_sym_LPAREN] = ACTIONS(2033),
    [anon_sym_RPAREN] = ACTIONS(2033),
    [anon_sym_COMMA] = ACTIONS(2033),
    [anon_sym_PIPE] = ACTIONS(2031),
    [anon_sym_COLON2] = ACTIONS(2031),
    [anon_sym_DOT] = ACTIONS(2033),
    [anon_sym_DASH_GT] = ACTIONS(2031),
    [anon_sym_LBRACK] = ACTIONS(2031),
    [anon_sym_RBRACK] = ACTIONS(2033),
    [anon_sym_true] = ACTIONS(2031),
    [anon_sym_false] = ACTIONS(2031),
    [anon_sym_LBRACE] = ACTIONS(2031),
    [anon_sym_SEMI] = ACTIONS(2033),
    [anon_sym_RBRACE] = ACTIONS(2033),
    [anon_sym_constraint] = ACTIONS(2031),
    [anon_sym_val] = ACTIONS(2031),
    [anon_sym_end] = ACTIONS(2031),
    [anon_sym_with] = ACTIONS(2031),
    [anon_sym_COLON_GT] = ACTIONS(2033),
    [anon_sym_object] = ACTIONS(2031),
    [anon_sym_inherit] = ACTIONS(2031),
    [anon_sym_method] = ACTIONS(2031),
    [anon_sym_initializer] = ACTIONS(2031),
    [anon_sym_AMP] = ACTIONS(2031),
    [anon_sym_POUND] = ACTIONS(2031),
    [anon_sym_COLON_COLON] = ACTIONS(2033),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2033),
    [anon_sym_then] = ACTIONS(2031),
    [anon_sym_else] = ACTIONS(2031),
    [anon_sym_do] = ACTIONS(2031),
    [anon_sym_done] = ACTIONS(2031),
    [anon_sym_to] = ACTIONS(2031),
    [anon_sym_downto] = ACTIONS(2031),
    [anon_sym_new] = ACTIONS(2031),
    [anon_sym_LBRACE_LT] = ACTIONS(2033),
    [anon_sym_GT_RBRACE] = ACTIONS(2033),
    [anon_sym_begin] = ACTIONS(2031),
    [sym_ocamlyacc_value] = ACTIONS(2033),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2031),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2033),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2031),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2031),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2033),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2033),
    [aux_sym_number_token1] = ACTIONS(2033),
    [aux_sym_signed_number_token1] = ACTIONS(2031),
    [anon_sym_SQUOTE] = ACTIONS(2033),
    [anon_sym_DQUOTE] = ACTIONS(2033),
    [sym_prefix_operator] = ACTIONS(2033),
    [aux_sym_sign_operator_token1] = ACTIONS(2031),
    [sym_hash_operator] = ACTIONS(2033),
    [anon_sym_lsl] = ACTIONS(2031),
    [anon_sym_lsr] = ACTIONS(2031),
    [anon_sym_asr] = ACTIONS(2031),
    [aux_sym_pow_operator_token1] = ACTIONS(2033),
    [anon_sym_mod] = ACTIONS(2031),
    [anon_sym_land] = ACTIONS(2031),
    [anon_sym_lor] = ACTIONS(2031),
    [anon_sym_lxor] = ACTIONS(2031),
    [aux_sym_mult_operator_token1] = ACTIONS(2031),
    [aux_sym_add_operator_token1] = ACTIONS(2031),
    [sym_concat_operator] = ACTIONS(2033),
    [sym_rel_operator] = ACTIONS(2031),
    [anon_sym_AMP_AMP] = ACTIONS(2031),
    [anon_sym_or] = ACTIONS(2031),
    [anon_sym_PIPE_PIPE] = ACTIONS(2031),
    [sym_assign_operator] = ACTIONS(2033),
    [sym__uppercase_identifier] = ACTIONS(2033),
    [aux_sym_tag_token1] = ACTIONS(2033),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1095)] = {
    [sym_attribute] = STATE(1095),
    [sym__lowercase_identifier] = ACTIONS(2129),
    [anon_sym_TILDE] = ACTIONS(2129),
    [anon_sym_QMARK] = ACTIONS(2129),
    [anon_sym_LPAREN] = ACTIONS(2131),
    [anon_sym_RPAREN] = ACTIONS(2131),
    [anon_sym_COMMA] = ACTIONS(2131),
    [anon_sym_PIPE] = ACTIONS(2129),
    [anon_sym_COLON2] = ACTIONS(2129),
    [anon_sym_DOT] = ACTIONS(2131),
    [anon_sym_DASH_GT] = ACTIONS(2129),
    [anon_sym_LBRACK] = ACTIONS(2129),
    [anon_sym_RBRACK] = ACTIONS(2131),
    [anon_sym_true] = ACTIONS(2129),
    [anon_sym_false] = ACTIONS(2129),
    [anon_sym_LBRACE] = ACTIONS(2129),
    [anon_sym_SEMI] = ACTIONS(2131),
    [anon_sym_RBRACE] = ACTIONS(2131),
    [anon_sym_constraint] = ACTIONS(2129),
    [anon_sym_val] = ACTIONS(2129),
    [anon_sym_end] = ACTIONS(2129),
    [anon_sym_with] = ACTIONS(2129),
    [anon_sym_COLON_GT] = ACTIONS(2131),
    [anon_sym_object] = ACTIONS(2129),
    [anon_sym_inherit] = ACTIONS(2129),
    [anon_sym_method] = ACTIONS(2129),
    [anon_sym_initializer] = ACTIONS(2129),
    [anon_sym_AMP] = ACTIONS(2129),
    [anon_sym_POUND] = ACTIONS(2129),
    [anon_sym_COLON_COLON] = ACTIONS(2131),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2131),
    [anon_sym_then] = ACTIONS(2129),
    [anon_sym_else] = ACTIONS(2129),
    [anon_sym_do] = ACTIONS(2129),
    [anon_sym_done] = ACTIONS(2129),
    [anon_sym_to] = ACTIONS(2129),
    [anon_sym_downto] = ACTIONS(2129),
    [anon_sym_new] = ACTIONS(2129),
    [anon_sym_LBRACE_LT] = ACTIONS(2131),
    [anon_sym_GT_RBRACE] = ACTIONS(2131),
    [anon_sym_begin] = ACTIONS(2129),
    [sym_ocamlyacc_value] = ACTIONS(2131),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2129),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2131),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2129),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2129),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2131),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2131),
    [aux_sym_number_token1] = ACTIONS(2131),
    [aux_sym_signed_number_token1] = ACTIONS(2129),
    [anon_sym_SQUOTE] = ACTIONS(2131),
    [anon_sym_DQUOTE] = ACTIONS(2131),
    [sym_prefix_operator] = ACTIONS(2131),
    [aux_sym_sign_operator_token1] = ACTIONS(2129),
    [sym_hash_operator] = ACTIONS(2131),
    [anon_sym_lsl] = ACTIONS(2129),
    [anon_sym_lsr] = ACTIONS(2129),
    [anon_sym_asr] = ACTIONS(2129),
    [aux_sym_pow_operator_token1] = ACTIONS(2131),
    [anon_sym_mod] = ACTIONS(2129),
    [anon_sym_land] = ACTIONS(2129),
    [anon_sym_lor] = ACTIONS(2129),
    [anon_sym_lxor] = ACTIONS(2129),
    [aux_sym_mult_operator_token1] = ACTIONS(2129),
    [aux_sym_add_operator_token1] = ACTIONS(2129),
    [sym_concat_operator] = ACTIONS(2131),
    [sym_rel_operator] = ACTIONS(2129),
    [anon_sym_AMP_AMP] = ACTIONS(2129),
    [anon_sym_or] = ACTIONS(2129),
    [anon_sym_PIPE_PIPE] = ACTIONS(2129),
    [sym_assign_operator] = ACTIONS(2131),
    [sym__uppercase_identifier] = ACTIONS(2131),
    [aux_sym_tag_token1] = ACTIONS(2131),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1096)] = {
    [sym_attribute] = STATE(1096),
    [sym__lowercase_identifier] = ACTIONS(2137),
    [anon_sym_TILDE] = ACTIONS(2137),
    [anon_sym_QMARK] = ACTIONS(2137),
    [anon_sym_LPAREN] = ACTIONS(2139),
    [anon_sym_RPAREN] = ACTIONS(2139),
    [anon_sym_COMMA] = ACTIONS(2139),
    [anon_sym_PIPE] = ACTIONS(2137),
    [anon_sym_COLON2] = ACTIONS(2137),
    [anon_sym_DOT] = ACTIONS(2139),
    [anon_sym_DASH_GT] = ACTIONS(2137),
    [anon_sym_LBRACK] = ACTIONS(2137),
    [anon_sym_RBRACK] = ACTIONS(2139),
    [anon_sym_true] = ACTIONS(2137),
    [anon_sym_false] = ACTIONS(2137),
    [anon_sym_LBRACE] = ACTIONS(2137),
    [anon_sym_SEMI] = ACTIONS(2139),
    [anon_sym_RBRACE] = ACTIONS(2139),
    [anon_sym_constraint] = ACTIONS(2137),
    [anon_sym_val] = ACTIONS(2137),
    [anon_sym_end] = ACTIONS(2137),
    [anon_sym_with] = ACTIONS(2137),
    [anon_sym_COLON_GT] = ACTIONS(2139),
    [anon_sym_object] = ACTIONS(2137),
    [anon_sym_inherit] = ACTIONS(2137),
    [anon_sym_method] = ACTIONS(2137),
    [anon_sym_initializer] = ACTIONS(2137),
    [anon_sym_AMP] = ACTIONS(2137),
    [anon_sym_POUND] = ACTIONS(2137),
    [anon_sym_COLON_COLON] = ACTIONS(2139),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2139),
    [anon_sym_then] = ACTIONS(2137),
    [anon_sym_else] = ACTIONS(2137),
    [anon_sym_do] = ACTIONS(2137),
    [anon_sym_done] = ACTIONS(2137),
    [anon_sym_to] = ACTIONS(2137),
    [anon_sym_downto] = ACTIONS(2137),
    [anon_sym_new] = ACTIONS(2137),
    [anon_sym_LBRACE_LT] = ACTIONS(2139),
    [anon_sym_GT_RBRACE] = ACTIONS(2139),
    [anon_sym_begin] = ACTIONS(2137),
    [sym_ocamlyacc_value] = ACTIONS(2139),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2137),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2139),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2137),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2137),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2139),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2139),
    [aux_sym_number_token1] = ACTIONS(2139),
    [aux_sym_signed_number_token1] = ACTIONS(2137),
    [anon_sym_SQUOTE] = ACTIONS(2139),
    [anon_sym_DQUOTE] = ACTIONS(2139),
    [sym_prefix_operator] = ACTIONS(2139),
    [aux_sym_sign_operator_token1] = ACTIONS(2137),
    [sym_hash_operator] = ACTIONS(2139),
    [anon_sym_lsl] = ACTIONS(2137),
    [anon_sym_lsr] = ACTIONS(2137),
    [anon_sym_asr] = ACTIONS(2137),
    [aux_sym_pow_operator_token1] = ACTIONS(2139),
    [anon_sym_mod] = ACTIONS(2137),
    [anon_sym_land] = ACTIONS(2137),
    [anon_sym_lor] = ACTIONS(2137),
    [anon_sym_lxor] = ACTIONS(2137),
    [aux_sym_mult_operator_token1] = ACTIONS(2137),
    [aux_sym_add_operator_token1] = ACTIONS(2137),
    [sym_concat_operator] = ACTIONS(2139),
    [sym_rel_operator] = ACTIONS(2137),
    [anon_sym_AMP_AMP] = ACTIONS(2137),
    [anon_sym_or] = ACTIONS(2137),
    [anon_sym_PIPE_PIPE] = ACTIONS(2137),
    [sym_assign_operator] = ACTIONS(2139),
    [sym__uppercase_identifier] = ACTIONS(2139),
    [aux_sym_tag_token1] = ACTIONS(2139),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1097)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4409),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1097),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4230),
    [aux_sym_let_binding_repeat1] = STATE(1134),
    [ts_builtin_sym_end] = ACTIONS(2385),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2385),
    [anon_sym_let] = ACTIONS(2387),
    [anon_sym_and] = ACTIONS(2387),
    [anon_sym_EQ] = ACTIONS(2445),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2387),
    [anon_sym_type] = ACTIONS(2387),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2387),
    [anon_sym_module] = ACTIONS(2387),
    [anon_sym_open] = ACTIONS(2387),
    [anon_sym_include] = ACTIONS(2387),
    [anon_sym_class] = ACTIONS(2387),
    [anon_sym_val] = ACTIONS(2387),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2385),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2385),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2385),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2385),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2387),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1098)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4411),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1098),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4241),
    [aux_sym_let_binding_repeat1] = STATE(1100),
    [ts_builtin_sym_end] = ACTIONS(2373),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2373),
    [anon_sym_let] = ACTIONS(2375),
    [anon_sym_and] = ACTIONS(2375),
    [anon_sym_EQ] = ACTIONS(2447),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2375),
    [anon_sym_type] = ACTIONS(2375),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2375),
    [anon_sym_module] = ACTIONS(2375),
    [anon_sym_open] = ACTIONS(2375),
    [anon_sym_include] = ACTIONS(2375),
    [anon_sym_class] = ACTIONS(2375),
    [anon_sym_val] = ACTIONS(2375),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2373),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2373),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2373),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2373),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2375),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1099)] = {
    [sym_attribute] = STATE(1099),
    [sym__lowercase_identifier] = ACTIONS(2161),
    [anon_sym_TILDE] = ACTIONS(2161),
    [anon_sym_QMARK] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_RPAREN] = ACTIONS(2163),
    [anon_sym_COMMA] = ACTIONS(2163),
    [anon_sym_PIPE] = ACTIONS(2161),
    [anon_sym_COLON2] = ACTIONS(2161),
    [anon_sym_DOT] = ACTIONS(2163),
    [anon_sym_DASH_GT] = ACTIONS(2161),
    [anon_sym_LBRACK] = ACTIONS(2161),
    [anon_sym_RBRACK] = ACTIONS(2163),
    [anon_sym_true] = ACTIONS(2161),
    [anon_sym_false] = ACTIONS(2161),
    [anon_sym_LBRACE] = ACTIONS(2161),
    [anon_sym_SEMI] = ACTIONS(2163),
    [anon_sym_RBRACE] = ACTIONS(2163),
    [anon_sym_constraint] = ACTIONS(2161),
    [anon_sym_val] = ACTIONS(2161),
    [anon_sym_end] = ACTIONS(2161),
    [anon_sym_with] = ACTIONS(2161),
    [anon_sym_COLON_GT] = ACTIONS(2163),
    [anon_sym_object] = ACTIONS(2161),
    [anon_sym_inherit] = ACTIONS(2161),
    [anon_sym_method] = ACTIONS(2161),
    [anon_sym_initializer] = ACTIONS(2161),
    [anon_sym_AMP] = ACTIONS(2161),
    [anon_sym_POUND] = ACTIONS(2161),
    [anon_sym_COLON_COLON] = ACTIONS(2163),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2163),
    [anon_sym_then] = ACTIONS(2161),
    [anon_sym_else] = ACTIONS(2161),
    [anon_sym_do] = ACTIONS(2161),
    [anon_sym_done] = ACTIONS(2161),
    [anon_sym_to] = ACTIONS(2161),
    [anon_sym_downto] = ACTIONS(2161),
    [anon_sym_new] = ACTIONS(2161),
    [anon_sym_LBRACE_LT] = ACTIONS(2163),
    [anon_sym_GT_RBRACE] = ACTIONS(2163),
    [anon_sym_begin] = ACTIONS(2161),
    [sym_ocamlyacc_value] = ACTIONS(2163),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2161),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2163),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2163),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2163),
    [aux_sym_number_token1] = ACTIONS(2163),
    [aux_sym_signed_number_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2163),
    [anon_sym_DQUOTE] = ACTIONS(2163),
    [sym_prefix_operator] = ACTIONS(2163),
    [aux_sym_sign_operator_token1] = ACTIONS(2161),
    [sym_hash_operator] = ACTIONS(2163),
    [anon_sym_lsl] = ACTIONS(2161),
    [anon_sym_lsr] = ACTIONS(2161),
    [anon_sym_asr] = ACTIONS(2161),
    [aux_sym_pow_operator_token1] = ACTIONS(2163),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_land] = ACTIONS(2161),
    [anon_sym_lor] = ACTIONS(2161),
    [anon_sym_lxor] = ACTIONS(2161),
    [aux_sym_mult_operator_token1] = ACTIONS(2161),
    [aux_sym_add_operator_token1] = ACTIONS(2161),
    [sym_concat_operator] = ACTIONS(2163),
    [sym_rel_operator] = ACTIONS(2161),
    [anon_sym_AMP_AMP] = ACTIONS(2161),
    [anon_sym_or] = ACTIONS(2161),
    [anon_sym_PIPE_PIPE] = ACTIONS(2161),
    [sym_assign_operator] = ACTIONS(2163),
    [sym__uppercase_identifier] = ACTIONS(2163),
    [aux_sym_tag_token1] = ACTIONS(2163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1100)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4466),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1100),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4300),
    [aux_sym_let_binding_repeat1] = STATE(1134),
    [ts_builtin_sym_end] = ACTIONS(2391),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2391),
    [anon_sym_let] = ACTIONS(2393),
    [anon_sym_and] = ACTIONS(2393),
    [anon_sym_EQ] = ACTIONS(2449),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2393),
    [anon_sym_type] = ACTIONS(2393),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2393),
    [anon_sym_module] = ACTIONS(2393),
    [anon_sym_open] = ACTIONS(2393),
    [anon_sym_include] = ACTIONS(2393),
    [anon_sym_class] = ACTIONS(2393),
    [anon_sym_val] = ACTIONS(2393),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2391),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2391),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2391),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2391),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2393),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1101)] = {
    [sym_attribute] = STATE(1101),
    [sym__lowercase_identifier] = ACTIONS(2103),
    [anon_sym_TILDE] = ACTIONS(2103),
    [anon_sym_QMARK] = ACTIONS(2103),
    [anon_sym_LPAREN] = ACTIONS(2105),
    [anon_sym_RPAREN] = ACTIONS(2105),
    [anon_sym_COMMA] = ACTIONS(2105),
    [anon_sym_PIPE] = ACTIONS(2103),
    [anon_sym_COLON2] = ACTIONS(2103),
    [anon_sym_DOT] = ACTIONS(2105),
    [anon_sym_DASH_GT] = ACTIONS(2103),
    [anon_sym_LBRACK] = ACTIONS(2103),
    [anon_sym_RBRACK] = ACTIONS(2105),
    [anon_sym_true] = ACTIONS(2103),
    [anon_sym_false] = ACTIONS(2103),
    [anon_sym_LBRACE] = ACTIONS(2103),
    [anon_sym_SEMI] = ACTIONS(2105),
    [anon_sym_RBRACE] = ACTIONS(2105),
    [anon_sym_constraint] = ACTIONS(2103),
    [anon_sym_val] = ACTIONS(2103),
    [anon_sym_end] = ACTIONS(2103),
    [anon_sym_with] = ACTIONS(2103),
    [anon_sym_COLON_GT] = ACTIONS(2105),
    [anon_sym_object] = ACTIONS(2103),
    [anon_sym_inherit] = ACTIONS(2103),
    [anon_sym_method] = ACTIONS(2103),
    [anon_sym_initializer] = ACTIONS(2103),
    [anon_sym_AMP] = ACTIONS(2103),
    [anon_sym_POUND] = ACTIONS(2103),
    [anon_sym_COLON_COLON] = ACTIONS(2105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2105),
    [anon_sym_then] = ACTIONS(2103),
    [anon_sym_else] = ACTIONS(2103),
    [anon_sym_do] = ACTIONS(2103),
    [anon_sym_done] = ACTIONS(2103),
    [anon_sym_to] = ACTIONS(2103),
    [anon_sym_downto] = ACTIONS(2103),
    [anon_sym_new] = ACTIONS(2103),
    [anon_sym_LBRACE_LT] = ACTIONS(2105),
    [anon_sym_GT_RBRACE] = ACTIONS(2105),
    [anon_sym_begin] = ACTIONS(2103),
    [sym_ocamlyacc_value] = ACTIONS(2105),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2103),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2105),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2103),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2103),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2105),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2105),
    [aux_sym_number_token1] = ACTIONS(2105),
    [aux_sym_signed_number_token1] = ACTIONS(2103),
    [anon_sym_SQUOTE] = ACTIONS(2105),
    [anon_sym_DQUOTE] = ACTIONS(2105),
    [sym_prefix_operator] = ACTIONS(2105),
    [aux_sym_sign_operator_token1] = ACTIONS(2103),
    [sym_hash_operator] = ACTIONS(2105),
    [anon_sym_lsl] = ACTIONS(2103),
    [anon_sym_lsr] = ACTIONS(2103),
    [anon_sym_asr] = ACTIONS(2103),
    [aux_sym_pow_operator_token1] = ACTIONS(2105),
    [anon_sym_mod] = ACTIONS(2103),
    [anon_sym_land] = ACTIONS(2103),
    [anon_sym_lor] = ACTIONS(2103),
    [anon_sym_lxor] = ACTIONS(2103),
    [aux_sym_mult_operator_token1] = ACTIONS(2103),
    [aux_sym_add_operator_token1] = ACTIONS(2103),
    [sym_concat_operator] = ACTIONS(2105),
    [sym_rel_operator] = ACTIONS(2103),
    [anon_sym_AMP_AMP] = ACTIONS(2103),
    [anon_sym_or] = ACTIONS(2103),
    [anon_sym_PIPE_PIPE] = ACTIONS(2103),
    [sym_assign_operator] = ACTIONS(2105),
    [sym__uppercase_identifier] = ACTIONS(2105),
    [aux_sym_tag_token1] = ACTIONS(2105),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1102)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4467),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1102),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4302),
    [aux_sym_let_binding_repeat1] = STATE(1105),
    [ts_builtin_sym_end] = ACTIONS(2397),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2397),
    [anon_sym_let] = ACTIONS(2399),
    [anon_sym_and] = ACTIONS(2399),
    [anon_sym_EQ] = ACTIONS(2451),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2399),
    [anon_sym_type] = ACTIONS(2399),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2399),
    [anon_sym_module] = ACTIONS(2399),
    [anon_sym_open] = ACTIONS(2399),
    [anon_sym_include] = ACTIONS(2399),
    [anon_sym_class] = ACTIONS(2399),
    [anon_sym_val] = ACTIONS(2399),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2397),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2397),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2397),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2397),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2399),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1103)] = {
    [sym_attribute] = STATE(1103),
    [sym__lowercase_identifier] = ACTIONS(2107),
    [anon_sym_TILDE] = ACTIONS(2107),
    [anon_sym_QMARK] = ACTIONS(2107),
    [anon_sym_LPAREN] = ACTIONS(2109),
    [anon_sym_RPAREN] = ACTIONS(2109),
    [anon_sym_COMMA] = ACTIONS(2109),
    [anon_sym_PIPE] = ACTIONS(2107),
    [anon_sym_COLON2] = ACTIONS(2107),
    [anon_sym_DOT] = ACTIONS(2109),
    [anon_sym_DASH_GT] = ACTIONS(2107),
    [anon_sym_LBRACK] = ACTIONS(2107),
    [anon_sym_RBRACK] = ACTIONS(2109),
    [anon_sym_true] = ACTIONS(2107),
    [anon_sym_false] = ACTIONS(2107),
    [anon_sym_LBRACE] = ACTIONS(2107),
    [anon_sym_SEMI] = ACTIONS(2109),
    [anon_sym_RBRACE] = ACTIONS(2109),
    [anon_sym_constraint] = ACTIONS(2107),
    [anon_sym_val] = ACTIONS(2107),
    [anon_sym_end] = ACTIONS(2107),
    [anon_sym_with] = ACTIONS(2107),
    [anon_sym_COLON_GT] = ACTIONS(2109),
    [anon_sym_object] = ACTIONS(2107),
    [anon_sym_inherit] = ACTIONS(2107),
    [anon_sym_method] = ACTIONS(2107),
    [anon_sym_initializer] = ACTIONS(2107),
    [anon_sym_AMP] = ACTIONS(2107),
    [anon_sym_POUND] = ACTIONS(2107),
    [anon_sym_COLON_COLON] = ACTIONS(2109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2109),
    [anon_sym_then] = ACTIONS(2107),
    [anon_sym_else] = ACTIONS(2107),
    [anon_sym_do] = ACTIONS(2107),
    [anon_sym_done] = ACTIONS(2107),
    [anon_sym_to] = ACTIONS(2107),
    [anon_sym_downto] = ACTIONS(2107),
    [anon_sym_new] = ACTIONS(2107),
    [anon_sym_LBRACE_LT] = ACTIONS(2109),
    [anon_sym_GT_RBRACE] = ACTIONS(2109),
    [anon_sym_begin] = ACTIONS(2107),
    [sym_ocamlyacc_value] = ACTIONS(2109),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2107),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2109),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2109),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2109),
    [aux_sym_number_token1] = ACTIONS(2109),
    [aux_sym_signed_number_token1] = ACTIONS(2107),
    [anon_sym_SQUOTE] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(2109),
    [sym_prefix_operator] = ACTIONS(2109),
    [aux_sym_sign_operator_token1] = ACTIONS(2107),
    [sym_hash_operator] = ACTIONS(2109),
    [anon_sym_lsl] = ACTIONS(2107),
    [anon_sym_lsr] = ACTIONS(2107),
    [anon_sym_asr] = ACTIONS(2107),
    [aux_sym_pow_operator_token1] = ACTIONS(2109),
    [anon_sym_mod] = ACTIONS(2107),
    [anon_sym_land] = ACTIONS(2107),
    [anon_sym_lor] = ACTIONS(2107),
    [anon_sym_lxor] = ACTIONS(2107),
    [aux_sym_mult_operator_token1] = ACTIONS(2107),
    [aux_sym_add_operator_token1] = ACTIONS(2107),
    [sym_concat_operator] = ACTIONS(2109),
    [sym_rel_operator] = ACTIONS(2107),
    [anon_sym_AMP_AMP] = ACTIONS(2107),
    [anon_sym_or] = ACTIONS(2107),
    [anon_sym_PIPE_PIPE] = ACTIONS(2107),
    [sym_assign_operator] = ACTIONS(2109),
    [sym__uppercase_identifier] = ACTIONS(2109),
    [aux_sym_tag_token1] = ACTIONS(2109),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1104)] = {
    [sym_attribute] = STATE(1104),
    [sym__lowercase_identifier] = ACTIONS(2111),
    [anon_sym_TILDE] = ACTIONS(2111),
    [anon_sym_QMARK] = ACTIONS(2111),
    [anon_sym_LPAREN] = ACTIONS(2113),
    [anon_sym_RPAREN] = ACTIONS(2113),
    [anon_sym_COMMA] = ACTIONS(2113),
    [anon_sym_PIPE] = ACTIONS(2111),
    [anon_sym_COLON2] = ACTIONS(2111),
    [anon_sym_DOT] = ACTIONS(2115),
    [anon_sym_DASH_GT] = ACTIONS(2111),
    [anon_sym_LBRACK] = ACTIONS(2111),
    [anon_sym_RBRACK] = ACTIONS(2113),
    [anon_sym_true] = ACTIONS(2111),
    [anon_sym_false] = ACTIONS(2111),
    [anon_sym_LBRACE] = ACTIONS(2111),
    [anon_sym_SEMI] = ACTIONS(2113),
    [anon_sym_RBRACE] = ACTIONS(2113),
    [anon_sym_constraint] = ACTIONS(2111),
    [anon_sym_val] = ACTIONS(2111),
    [anon_sym_end] = ACTIONS(2111),
    [anon_sym_with] = ACTIONS(2111),
    [anon_sym_COLON_GT] = ACTIONS(2113),
    [anon_sym_object] = ACTIONS(2111),
    [anon_sym_inherit] = ACTIONS(2111),
    [anon_sym_method] = ACTIONS(2111),
    [anon_sym_initializer] = ACTIONS(2111),
    [anon_sym_AMP] = ACTIONS(2111),
    [anon_sym_POUND] = ACTIONS(2111),
    [anon_sym_COLON_COLON] = ACTIONS(2113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2113),
    [anon_sym_then] = ACTIONS(2111),
    [anon_sym_else] = ACTIONS(2111),
    [anon_sym_do] = ACTIONS(2111),
    [anon_sym_done] = ACTIONS(2111),
    [anon_sym_to] = ACTIONS(2111),
    [anon_sym_downto] = ACTIONS(2111),
    [anon_sym_new] = ACTIONS(2111),
    [anon_sym_LBRACE_LT] = ACTIONS(2113),
    [anon_sym_GT_RBRACE] = ACTIONS(2113),
    [anon_sym_begin] = ACTIONS(2111),
    [sym_ocamlyacc_value] = ACTIONS(2113),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2113),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2113),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2113),
    [aux_sym_number_token1] = ACTIONS(2113),
    [aux_sym_signed_number_token1] = ACTIONS(2111),
    [anon_sym_SQUOTE] = ACTIONS(2113),
    [anon_sym_DQUOTE] = ACTIONS(2113),
    [sym_prefix_operator] = ACTIONS(2113),
    [aux_sym_sign_operator_token1] = ACTIONS(2111),
    [sym_hash_operator] = ACTIONS(2113),
    [anon_sym_lsl] = ACTIONS(2111),
    [anon_sym_lsr] = ACTIONS(2111),
    [anon_sym_asr] = ACTIONS(2111),
    [aux_sym_pow_operator_token1] = ACTIONS(2113),
    [anon_sym_mod] = ACTIONS(2111),
    [anon_sym_land] = ACTIONS(2111),
    [anon_sym_lor] = ACTIONS(2111),
    [anon_sym_lxor] = ACTIONS(2111),
    [aux_sym_mult_operator_token1] = ACTIONS(2111),
    [aux_sym_add_operator_token1] = ACTIONS(2111),
    [sym_concat_operator] = ACTIONS(2113),
    [sym_rel_operator] = ACTIONS(2111),
    [anon_sym_AMP_AMP] = ACTIONS(2111),
    [anon_sym_or] = ACTIONS(2111),
    [anon_sym_PIPE_PIPE] = ACTIONS(2111),
    [sym_assign_operator] = ACTIONS(2113),
    [sym__uppercase_identifier] = ACTIONS(2113),
    [aux_sym_tag_token1] = ACTIONS(2113),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1105)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4450),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1105),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4205),
    [aux_sym_let_binding_repeat1] = STATE(1134),
    [ts_builtin_sym_end] = ACTIONS(2403),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2403),
    [anon_sym_let] = ACTIONS(2405),
    [anon_sym_and] = ACTIONS(2405),
    [anon_sym_EQ] = ACTIONS(2453),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2405),
    [anon_sym_type] = ACTIONS(2405),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2405),
    [anon_sym_module] = ACTIONS(2405),
    [anon_sym_open] = ACTIONS(2405),
    [anon_sym_include] = ACTIONS(2405),
    [anon_sym_class] = ACTIONS(2405),
    [anon_sym_val] = ACTIONS(2405),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2403),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2403),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2403),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2403),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2405),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1106)] = {
    [sym_attribute] = STATE(1106),
    [sym__lowercase_identifier] = ACTIONS(2117),
    [anon_sym_TILDE] = ACTIONS(2117),
    [anon_sym_QMARK] = ACTIONS(2117),
    [anon_sym_LPAREN] = ACTIONS(2119),
    [anon_sym_RPAREN] = ACTIONS(2119),
    [anon_sym_COMMA] = ACTIONS(2119),
    [anon_sym_PIPE] = ACTIONS(2117),
    [anon_sym_COLON2] = ACTIONS(2117),
    [anon_sym_DOT] = ACTIONS(2119),
    [anon_sym_DASH_GT] = ACTIONS(2117),
    [anon_sym_LBRACK] = ACTIONS(2117),
    [anon_sym_RBRACK] = ACTIONS(2119),
    [anon_sym_true] = ACTIONS(2117),
    [anon_sym_false] = ACTIONS(2117),
    [anon_sym_LBRACE] = ACTIONS(2117),
    [anon_sym_SEMI] = ACTIONS(2119),
    [anon_sym_RBRACE] = ACTIONS(2119),
    [anon_sym_constraint] = ACTIONS(2117),
    [anon_sym_val] = ACTIONS(2117),
    [anon_sym_end] = ACTIONS(2117),
    [anon_sym_with] = ACTIONS(2117),
    [anon_sym_COLON_GT] = ACTIONS(2119),
    [anon_sym_object] = ACTIONS(2117),
    [anon_sym_inherit] = ACTIONS(2117),
    [anon_sym_method] = ACTIONS(2117),
    [anon_sym_initializer] = ACTIONS(2117),
    [anon_sym_AMP] = ACTIONS(2117),
    [anon_sym_POUND] = ACTIONS(2117),
    [anon_sym_COLON_COLON] = ACTIONS(2119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2119),
    [anon_sym_then] = ACTIONS(2117),
    [anon_sym_else] = ACTIONS(2117),
    [anon_sym_do] = ACTIONS(2117),
    [anon_sym_done] = ACTIONS(2117),
    [anon_sym_to] = ACTIONS(2117),
    [anon_sym_downto] = ACTIONS(2117),
    [anon_sym_new] = ACTIONS(2117),
    [anon_sym_LBRACE_LT] = ACTIONS(2119),
    [anon_sym_GT_RBRACE] = ACTIONS(2119),
    [anon_sym_begin] = ACTIONS(2117),
    [sym_ocamlyacc_value] = ACTIONS(2119),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2117),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2119),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2117),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2117),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2119),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2119),
    [aux_sym_number_token1] = ACTIONS(2119),
    [aux_sym_signed_number_token1] = ACTIONS(2117),
    [anon_sym_SQUOTE] = ACTIONS(2119),
    [anon_sym_DQUOTE] = ACTIONS(2119),
    [sym_prefix_operator] = ACTIONS(2119),
    [aux_sym_sign_operator_token1] = ACTIONS(2117),
    [sym_hash_operator] = ACTIONS(2119),
    [anon_sym_lsl] = ACTIONS(2117),
    [anon_sym_lsr] = ACTIONS(2117),
    [anon_sym_asr] = ACTIONS(2117),
    [aux_sym_pow_operator_token1] = ACTIONS(2119),
    [anon_sym_mod] = ACTIONS(2117),
    [anon_sym_land] = ACTIONS(2117),
    [anon_sym_lor] = ACTIONS(2117),
    [anon_sym_lxor] = ACTIONS(2117),
    [aux_sym_mult_operator_token1] = ACTIONS(2117),
    [aux_sym_add_operator_token1] = ACTIONS(2117),
    [sym_concat_operator] = ACTIONS(2119),
    [sym_rel_operator] = ACTIONS(2117),
    [anon_sym_AMP_AMP] = ACTIONS(2117),
    [anon_sym_or] = ACTIONS(2117),
    [anon_sym_PIPE_PIPE] = ACTIONS(2117),
    [sym_assign_operator] = ACTIONS(2119),
    [sym__uppercase_identifier] = ACTIONS(2119),
    [aux_sym_tag_token1] = ACTIONS(2119),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1107)] = {
    [sym_attribute] = STATE(1107),
    [sym__lowercase_identifier] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(2125),
    [anon_sym_QMARK] = ACTIONS(2125),
    [anon_sym_LPAREN] = ACTIONS(2127),
    [anon_sym_RPAREN] = ACTIONS(2127),
    [anon_sym_COMMA] = ACTIONS(2127),
    [anon_sym_PIPE] = ACTIONS(2125),
    [anon_sym_COLON2] = ACTIONS(2125),
    [anon_sym_DOT] = ACTIONS(2127),
    [anon_sym_DASH_GT] = ACTIONS(2125),
    [anon_sym_LBRACK] = ACTIONS(2125),
    [anon_sym_RBRACK] = ACTIONS(2127),
    [anon_sym_true] = ACTIONS(2125),
    [anon_sym_false] = ACTIONS(2125),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_SEMI] = ACTIONS(2127),
    [anon_sym_RBRACE] = ACTIONS(2127),
    [anon_sym_constraint] = ACTIONS(2125),
    [anon_sym_val] = ACTIONS(2125),
    [anon_sym_end] = ACTIONS(2125),
    [anon_sym_with] = ACTIONS(2125),
    [anon_sym_COLON_GT] = ACTIONS(2127),
    [anon_sym_object] = ACTIONS(2125),
    [anon_sym_inherit] = ACTIONS(2125),
    [anon_sym_method] = ACTIONS(2125),
    [anon_sym_initializer] = ACTIONS(2125),
    [anon_sym_AMP] = ACTIONS(2125),
    [anon_sym_POUND] = ACTIONS(2125),
    [anon_sym_COLON_COLON] = ACTIONS(2127),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2127),
    [anon_sym_then] = ACTIONS(2125),
    [anon_sym_else] = ACTIONS(2125),
    [anon_sym_do] = ACTIONS(2125),
    [anon_sym_done] = ACTIONS(2125),
    [anon_sym_to] = ACTIONS(2125),
    [anon_sym_downto] = ACTIONS(2125),
    [anon_sym_new] = ACTIONS(2125),
    [anon_sym_LBRACE_LT] = ACTIONS(2127),
    [anon_sym_GT_RBRACE] = ACTIONS(2127),
    [anon_sym_begin] = ACTIONS(2125),
    [sym_ocamlyacc_value] = ACTIONS(2127),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2125),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2127),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2127),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2127),
    [aux_sym_number_token1] = ACTIONS(2127),
    [aux_sym_signed_number_token1] = ACTIONS(2125),
    [anon_sym_SQUOTE] = ACTIONS(2127),
    [anon_sym_DQUOTE] = ACTIONS(2127),
    [sym_prefix_operator] = ACTIONS(2127),
    [aux_sym_sign_operator_token1] = ACTIONS(2125),
    [sym_hash_operator] = ACTIONS(2127),
    [anon_sym_lsl] = ACTIONS(2125),
    [anon_sym_lsr] = ACTIONS(2125),
    [anon_sym_asr] = ACTIONS(2125),
    [aux_sym_pow_operator_token1] = ACTIONS(2127),
    [anon_sym_mod] = ACTIONS(2125),
    [anon_sym_land] = ACTIONS(2125),
    [anon_sym_lor] = ACTIONS(2125),
    [anon_sym_lxor] = ACTIONS(2125),
    [aux_sym_mult_operator_token1] = ACTIONS(2125),
    [aux_sym_add_operator_token1] = ACTIONS(2125),
    [sym_concat_operator] = ACTIONS(2127),
    [sym_rel_operator] = ACTIONS(2125),
    [anon_sym_AMP_AMP] = ACTIONS(2125),
    [anon_sym_or] = ACTIONS(2125),
    [anon_sym_PIPE_PIPE] = ACTIONS(2125),
    [sym_assign_operator] = ACTIONS(2127),
    [sym__uppercase_identifier] = ACTIONS(2127),
    [aux_sym_tag_token1] = ACTIONS(2127),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1108)] = {
    [sym_attribute] = STATE(1108),
    [sym__lowercase_identifier] = ACTIONS(2145),
    [anon_sym_TILDE] = ACTIONS(2145),
    [anon_sym_QMARK] = ACTIONS(2145),
    [anon_sym_LPAREN] = ACTIONS(2147),
    [anon_sym_RPAREN] = ACTIONS(2147),
    [anon_sym_COMMA] = ACTIONS(2147),
    [anon_sym_PIPE] = ACTIONS(2145),
    [anon_sym_COLON2] = ACTIONS(2145),
    [anon_sym_DOT] = ACTIONS(2147),
    [anon_sym_DASH_GT] = ACTIONS(2145),
    [anon_sym_LBRACK] = ACTIONS(2145),
    [anon_sym_RBRACK] = ACTIONS(2147),
    [anon_sym_true] = ACTIONS(2145),
    [anon_sym_false] = ACTIONS(2145),
    [anon_sym_LBRACE] = ACTIONS(2145),
    [anon_sym_SEMI] = ACTIONS(2147),
    [anon_sym_RBRACE] = ACTIONS(2147),
    [anon_sym_constraint] = ACTIONS(2145),
    [anon_sym_val] = ACTIONS(2145),
    [anon_sym_end] = ACTIONS(2145),
    [anon_sym_with] = ACTIONS(2145),
    [anon_sym_COLON_GT] = ACTIONS(2147),
    [anon_sym_object] = ACTIONS(2145),
    [anon_sym_inherit] = ACTIONS(2145),
    [anon_sym_method] = ACTIONS(2145),
    [anon_sym_initializer] = ACTIONS(2145),
    [anon_sym_AMP] = ACTIONS(2145),
    [anon_sym_POUND] = ACTIONS(2145),
    [anon_sym_COLON_COLON] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2147),
    [anon_sym_then] = ACTIONS(2145),
    [anon_sym_else] = ACTIONS(2145),
    [anon_sym_do] = ACTIONS(2145),
    [anon_sym_done] = ACTIONS(2145),
    [anon_sym_to] = ACTIONS(2145),
    [anon_sym_downto] = ACTIONS(2145),
    [anon_sym_new] = ACTIONS(2145),
    [anon_sym_LBRACE_LT] = ACTIONS(2147),
    [anon_sym_GT_RBRACE] = ACTIONS(2147),
    [anon_sym_begin] = ACTIONS(2145),
    [sym_ocamlyacc_value] = ACTIONS(2147),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2145),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2147),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2145),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2145),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2147),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2147),
    [aux_sym_number_token1] = ACTIONS(2147),
    [aux_sym_signed_number_token1] = ACTIONS(2145),
    [anon_sym_SQUOTE] = ACTIONS(2147),
    [anon_sym_DQUOTE] = ACTIONS(2147),
    [sym_prefix_operator] = ACTIONS(2147),
    [aux_sym_sign_operator_token1] = ACTIONS(2145),
    [sym_hash_operator] = ACTIONS(2147),
    [anon_sym_lsl] = ACTIONS(2145),
    [anon_sym_lsr] = ACTIONS(2145),
    [anon_sym_asr] = ACTIONS(2145),
    [aux_sym_pow_operator_token1] = ACTIONS(2147),
    [anon_sym_mod] = ACTIONS(2145),
    [anon_sym_land] = ACTIONS(2145),
    [anon_sym_lor] = ACTIONS(2145),
    [anon_sym_lxor] = ACTIONS(2145),
    [aux_sym_mult_operator_token1] = ACTIONS(2145),
    [aux_sym_add_operator_token1] = ACTIONS(2145),
    [sym_concat_operator] = ACTIONS(2147),
    [sym_rel_operator] = ACTIONS(2145),
    [anon_sym_AMP_AMP] = ACTIONS(2145),
    [anon_sym_or] = ACTIONS(2145),
    [anon_sym_PIPE_PIPE] = ACTIONS(2145),
    [sym_assign_operator] = ACTIONS(2147),
    [sym__uppercase_identifier] = ACTIONS(2147),
    [aux_sym_tag_token1] = ACTIONS(2147),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1109)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4369),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1109),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4211),
    [aux_sym_let_binding_repeat1] = STATE(1134),
    [ts_builtin_sym_end] = ACTIONS(2419),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2419),
    [anon_sym_let] = ACTIONS(2421),
    [anon_sym_and] = ACTIONS(2421),
    [anon_sym_EQ] = ACTIONS(2455),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2421),
    [anon_sym_type] = ACTIONS(2421),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2421),
    [anon_sym_module] = ACTIONS(2421),
    [anon_sym_open] = ACTIONS(2421),
    [anon_sym_include] = ACTIONS(2421),
    [anon_sym_class] = ACTIONS(2421),
    [anon_sym_val] = ACTIONS(2421),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2419),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2419),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2419),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2419),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2421),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1110)] = {
    [sym_attribute] = STATE(1110),
    [sym__lowercase_identifier] = ACTIONS(2169),
    [anon_sym_TILDE] = ACTIONS(2169),
    [anon_sym_QMARK] = ACTIONS(2169),
    [anon_sym_LPAREN] = ACTIONS(2171),
    [anon_sym_RPAREN] = ACTIONS(2171),
    [anon_sym_COMMA] = ACTIONS(2171),
    [anon_sym_PIPE] = ACTIONS(2169),
    [anon_sym_COLON2] = ACTIONS(2169),
    [anon_sym_DOT] = ACTIONS(2171),
    [anon_sym_DASH_GT] = ACTIONS(2169),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_RBRACK] = ACTIONS(2171),
    [anon_sym_true] = ACTIONS(2169),
    [anon_sym_false] = ACTIONS(2169),
    [anon_sym_LBRACE] = ACTIONS(2169),
    [anon_sym_SEMI] = ACTIONS(2171),
    [anon_sym_RBRACE] = ACTIONS(2171),
    [anon_sym_constraint] = ACTIONS(2169),
    [anon_sym_val] = ACTIONS(2169),
    [anon_sym_end] = ACTIONS(2169),
    [anon_sym_with] = ACTIONS(2169),
    [anon_sym_COLON_GT] = ACTIONS(2171),
    [anon_sym_object] = ACTIONS(2169),
    [anon_sym_inherit] = ACTIONS(2169),
    [anon_sym_method] = ACTIONS(2169),
    [anon_sym_initializer] = ACTIONS(2169),
    [anon_sym_AMP] = ACTIONS(2169),
    [anon_sym_POUND] = ACTIONS(2169),
    [anon_sym_COLON_COLON] = ACTIONS(2171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2171),
    [anon_sym_then] = ACTIONS(2169),
    [anon_sym_else] = ACTIONS(2169),
    [anon_sym_do] = ACTIONS(2169),
    [anon_sym_done] = ACTIONS(2169),
    [anon_sym_to] = ACTIONS(2169),
    [anon_sym_downto] = ACTIONS(2169),
    [anon_sym_new] = ACTIONS(2169),
    [anon_sym_LBRACE_LT] = ACTIONS(2171),
    [anon_sym_GT_RBRACE] = ACTIONS(2171),
    [anon_sym_begin] = ACTIONS(2169),
    [sym_ocamlyacc_value] = ACTIONS(2171),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2169),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2171),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2169),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2169),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2171),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2171),
    [aux_sym_number_token1] = ACTIONS(2171),
    [aux_sym_signed_number_token1] = ACTIONS(2169),
    [anon_sym_SQUOTE] = ACTIONS(2171),
    [anon_sym_DQUOTE] = ACTIONS(2171),
    [sym_prefix_operator] = ACTIONS(2171),
    [aux_sym_sign_operator_token1] = ACTIONS(2169),
    [sym_hash_operator] = ACTIONS(2171),
    [anon_sym_lsl] = ACTIONS(2169),
    [anon_sym_lsr] = ACTIONS(2169),
    [anon_sym_asr] = ACTIONS(2169),
    [aux_sym_pow_operator_token1] = ACTIONS(2171),
    [anon_sym_mod] = ACTIONS(2169),
    [anon_sym_land] = ACTIONS(2169),
    [anon_sym_lor] = ACTIONS(2169),
    [anon_sym_lxor] = ACTIONS(2169),
    [aux_sym_mult_operator_token1] = ACTIONS(2169),
    [aux_sym_add_operator_token1] = ACTIONS(2169),
    [sym_concat_operator] = ACTIONS(2171),
    [sym_rel_operator] = ACTIONS(2169),
    [anon_sym_AMP_AMP] = ACTIONS(2169),
    [anon_sym_or] = ACTIONS(2169),
    [anon_sym_PIPE_PIPE] = ACTIONS(2169),
    [sym_assign_operator] = ACTIONS(2171),
    [sym__uppercase_identifier] = ACTIONS(2171),
    [aux_sym_tag_token1] = ACTIONS(2171),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1111)] = {
    [sym_attribute] = STATE(1111),
    [sym__lowercase_identifier] = ACTIONS(2111),
    [anon_sym_TILDE] = ACTIONS(2111),
    [anon_sym_QMARK] = ACTIONS(2111),
    [anon_sym_LPAREN] = ACTIONS(2113),
    [anon_sym_RPAREN] = ACTIONS(2113),
    [anon_sym_COMMA] = ACTIONS(2113),
    [anon_sym_PIPE] = ACTIONS(2111),
    [anon_sym_COLON2] = ACTIONS(2111),
    [anon_sym_DOT] = ACTIONS(2177),
    [anon_sym_DASH_GT] = ACTIONS(2111),
    [anon_sym_LBRACK] = ACTIONS(2111),
    [anon_sym_RBRACK] = ACTIONS(2113),
    [anon_sym_true] = ACTIONS(2111),
    [anon_sym_false] = ACTIONS(2111),
    [anon_sym_LBRACE] = ACTIONS(2111),
    [anon_sym_SEMI] = ACTIONS(2113),
    [anon_sym_RBRACE] = ACTIONS(2113),
    [anon_sym_constraint] = ACTIONS(2111),
    [anon_sym_val] = ACTIONS(2111),
    [anon_sym_end] = ACTIONS(2111),
    [anon_sym_with] = ACTIONS(2111),
    [anon_sym_COLON_GT] = ACTIONS(2113),
    [anon_sym_object] = ACTIONS(2111),
    [anon_sym_inherit] = ACTIONS(2111),
    [anon_sym_method] = ACTIONS(2111),
    [anon_sym_initializer] = ACTIONS(2111),
    [anon_sym_AMP] = ACTIONS(2111),
    [anon_sym_POUND] = ACTIONS(2111),
    [anon_sym_COLON_COLON] = ACTIONS(2113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2113),
    [anon_sym_then] = ACTIONS(2111),
    [anon_sym_else] = ACTIONS(2111),
    [anon_sym_do] = ACTIONS(2111),
    [anon_sym_done] = ACTIONS(2111),
    [anon_sym_to] = ACTIONS(2111),
    [anon_sym_downto] = ACTIONS(2111),
    [anon_sym_new] = ACTIONS(2111),
    [anon_sym_LBRACE_LT] = ACTIONS(2113),
    [anon_sym_GT_RBRACE] = ACTIONS(2113),
    [anon_sym_begin] = ACTIONS(2111),
    [sym_ocamlyacc_value] = ACTIONS(2113),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2113),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2113),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2113),
    [aux_sym_number_token1] = ACTIONS(2113),
    [aux_sym_signed_number_token1] = ACTIONS(2111),
    [anon_sym_SQUOTE] = ACTIONS(2113),
    [anon_sym_DQUOTE] = ACTIONS(2113),
    [sym_prefix_operator] = ACTIONS(2113),
    [aux_sym_sign_operator_token1] = ACTIONS(2111),
    [sym_hash_operator] = ACTIONS(2113),
    [anon_sym_lsl] = ACTIONS(2111),
    [anon_sym_lsr] = ACTIONS(2111),
    [anon_sym_asr] = ACTIONS(2111),
    [aux_sym_pow_operator_token1] = ACTIONS(2113),
    [anon_sym_mod] = ACTIONS(2111),
    [anon_sym_land] = ACTIONS(2111),
    [anon_sym_lor] = ACTIONS(2111),
    [anon_sym_lxor] = ACTIONS(2111),
    [aux_sym_mult_operator_token1] = ACTIONS(2111),
    [aux_sym_add_operator_token1] = ACTIONS(2111),
    [sym_concat_operator] = ACTIONS(2113),
    [sym_rel_operator] = ACTIONS(2111),
    [anon_sym_AMP_AMP] = ACTIONS(2111),
    [anon_sym_or] = ACTIONS(2111),
    [anon_sym_PIPE_PIPE] = ACTIONS(2111),
    [sym_assign_operator] = ACTIONS(2113),
    [sym__uppercase_identifier] = ACTIONS(2113),
    [aux_sym_tag_token1] = ACTIONS(2113),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1112)] = {
    [sym_attribute] = STATE(1112),
    [sym__lowercase_identifier] = ACTIONS(2179),
    [anon_sym_TILDE] = ACTIONS(2179),
    [anon_sym_QMARK] = ACTIONS(2179),
    [anon_sym_LPAREN] = ACTIONS(2181),
    [anon_sym_RPAREN] = ACTIONS(2181),
    [anon_sym_COMMA] = ACTIONS(2181),
    [anon_sym_PIPE] = ACTIONS(2179),
    [anon_sym_COLON2] = ACTIONS(2179),
    [anon_sym_DOT] = ACTIONS(2181),
    [anon_sym_DASH_GT] = ACTIONS(2179),
    [anon_sym_LBRACK] = ACTIONS(2179),
    [anon_sym_RBRACK] = ACTIONS(2181),
    [anon_sym_true] = ACTIONS(2179),
    [anon_sym_false] = ACTIONS(2179),
    [anon_sym_LBRACE] = ACTIONS(2179),
    [anon_sym_SEMI] = ACTIONS(2181),
    [anon_sym_RBRACE] = ACTIONS(2181),
    [anon_sym_constraint] = ACTIONS(2179),
    [anon_sym_val] = ACTIONS(2179),
    [anon_sym_end] = ACTIONS(2179),
    [anon_sym_with] = ACTIONS(2179),
    [anon_sym_COLON_GT] = ACTIONS(2181),
    [anon_sym_object] = ACTIONS(2179),
    [anon_sym_inherit] = ACTIONS(2179),
    [anon_sym_method] = ACTIONS(2179),
    [anon_sym_initializer] = ACTIONS(2179),
    [anon_sym_AMP] = ACTIONS(2179),
    [anon_sym_POUND] = ACTIONS(2179),
    [anon_sym_COLON_COLON] = ACTIONS(2181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2181),
    [anon_sym_then] = ACTIONS(2179),
    [anon_sym_else] = ACTIONS(2179),
    [anon_sym_do] = ACTIONS(2179),
    [anon_sym_done] = ACTIONS(2179),
    [anon_sym_to] = ACTIONS(2179),
    [anon_sym_downto] = ACTIONS(2179),
    [anon_sym_new] = ACTIONS(2179),
    [anon_sym_LBRACE_LT] = ACTIONS(2181),
    [anon_sym_GT_RBRACE] = ACTIONS(2181),
    [anon_sym_begin] = ACTIONS(2179),
    [sym_ocamlyacc_value] = ACTIONS(2181),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2179),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2181),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2179),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2181),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2181),
    [aux_sym_number_token1] = ACTIONS(2181),
    [aux_sym_signed_number_token1] = ACTIONS(2179),
    [anon_sym_SQUOTE] = ACTIONS(2181),
    [anon_sym_DQUOTE] = ACTIONS(2181),
    [sym_prefix_operator] = ACTIONS(2181),
    [aux_sym_sign_operator_token1] = ACTIONS(2179),
    [sym_hash_operator] = ACTIONS(2181),
    [anon_sym_lsl] = ACTIONS(2179),
    [anon_sym_lsr] = ACTIONS(2179),
    [anon_sym_asr] = ACTIONS(2179),
    [aux_sym_pow_operator_token1] = ACTIONS(2181),
    [anon_sym_mod] = ACTIONS(2179),
    [anon_sym_land] = ACTIONS(2179),
    [anon_sym_lor] = ACTIONS(2179),
    [anon_sym_lxor] = ACTIONS(2179),
    [aux_sym_mult_operator_token1] = ACTIONS(2179),
    [aux_sym_add_operator_token1] = ACTIONS(2179),
    [sym_concat_operator] = ACTIONS(2181),
    [sym_rel_operator] = ACTIONS(2179),
    [anon_sym_AMP_AMP] = ACTIONS(2179),
    [anon_sym_or] = ACTIONS(2179),
    [anon_sym_PIPE_PIPE] = ACTIONS(2179),
    [sym_assign_operator] = ACTIONS(2181),
    [sym__uppercase_identifier] = ACTIONS(2181),
    [aux_sym_tag_token1] = ACTIONS(2181),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1113)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4374),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1113),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4347),
    [aux_sym_let_binding_repeat1] = STATE(1124),
    [ts_builtin_sym_end] = ACTIONS(2425),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2425),
    [anon_sym_let] = ACTIONS(2427),
    [anon_sym_and] = ACTIONS(2427),
    [anon_sym_EQ] = ACTIONS(2457),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2427),
    [anon_sym_type] = ACTIONS(2427),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2427),
    [anon_sym_module] = ACTIONS(2427),
    [anon_sym_open] = ACTIONS(2427),
    [anon_sym_include] = ACTIONS(2427),
    [anon_sym_class] = ACTIONS(2427),
    [anon_sym_val] = ACTIONS(2427),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2425),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2425),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2425),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2425),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2427),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1114)] = {
    [sym_attribute] = STATE(1114),
    [sym__lowercase_identifier] = ACTIONS(2157),
    [anon_sym_TILDE] = ACTIONS(2157),
    [anon_sym_QMARK] = ACTIONS(2157),
    [anon_sym_LPAREN] = ACTIONS(2159),
    [anon_sym_RPAREN] = ACTIONS(2159),
    [anon_sym_COMMA] = ACTIONS(2159),
    [anon_sym_PIPE] = ACTIONS(2157),
    [anon_sym_COLON2] = ACTIONS(2157),
    [anon_sym_DOT] = ACTIONS(2159),
    [anon_sym_DASH_GT] = ACTIONS(2157),
    [anon_sym_LBRACK] = ACTIONS(2157),
    [anon_sym_RBRACK] = ACTIONS(2159),
    [anon_sym_true] = ACTIONS(2157),
    [anon_sym_false] = ACTIONS(2157),
    [anon_sym_LBRACE] = ACTIONS(2157),
    [anon_sym_SEMI] = ACTIONS(2159),
    [anon_sym_RBRACE] = ACTIONS(2159),
    [anon_sym_constraint] = ACTIONS(2157),
    [anon_sym_val] = ACTIONS(2157),
    [anon_sym_end] = ACTIONS(2157),
    [anon_sym_with] = ACTIONS(2157),
    [anon_sym_COLON_GT] = ACTIONS(2159),
    [anon_sym_object] = ACTIONS(2157),
    [anon_sym_inherit] = ACTIONS(2157),
    [anon_sym_method] = ACTIONS(2157),
    [anon_sym_initializer] = ACTIONS(2157),
    [anon_sym_AMP] = ACTIONS(2157),
    [anon_sym_POUND] = ACTIONS(2157),
    [anon_sym_COLON_COLON] = ACTIONS(2159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2159),
    [anon_sym_then] = ACTIONS(2157),
    [anon_sym_else] = ACTIONS(2157),
    [anon_sym_do] = ACTIONS(2157),
    [anon_sym_done] = ACTIONS(2157),
    [anon_sym_to] = ACTIONS(2157),
    [anon_sym_downto] = ACTIONS(2157),
    [anon_sym_new] = ACTIONS(2157),
    [anon_sym_LBRACE_LT] = ACTIONS(2159),
    [anon_sym_GT_RBRACE] = ACTIONS(2159),
    [anon_sym_begin] = ACTIONS(2157),
    [sym_ocamlyacc_value] = ACTIONS(2159),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2157),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2159),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2159),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2159),
    [aux_sym_number_token1] = ACTIONS(2159),
    [aux_sym_signed_number_token1] = ACTIONS(2157),
    [anon_sym_SQUOTE] = ACTIONS(2159),
    [anon_sym_DQUOTE] = ACTIONS(2159),
    [sym_prefix_operator] = ACTIONS(2159),
    [aux_sym_sign_operator_token1] = ACTIONS(2157),
    [sym_hash_operator] = ACTIONS(2159),
    [anon_sym_lsl] = ACTIONS(2157),
    [anon_sym_lsr] = ACTIONS(2157),
    [anon_sym_asr] = ACTIONS(2157),
    [aux_sym_pow_operator_token1] = ACTIONS(2159),
    [anon_sym_mod] = ACTIONS(2157),
    [anon_sym_land] = ACTIONS(2157),
    [anon_sym_lor] = ACTIONS(2157),
    [anon_sym_lxor] = ACTIONS(2157),
    [aux_sym_mult_operator_token1] = ACTIONS(2157),
    [aux_sym_add_operator_token1] = ACTIONS(2157),
    [sym_concat_operator] = ACTIONS(2159),
    [sym_rel_operator] = ACTIONS(2157),
    [anon_sym_AMP_AMP] = ACTIONS(2157),
    [anon_sym_or] = ACTIONS(2157),
    [anon_sym_PIPE_PIPE] = ACTIONS(2157),
    [sym_assign_operator] = ACTIONS(2159),
    [sym__uppercase_identifier] = ACTIONS(2159),
    [aux_sym_tag_token1] = ACTIONS(2159),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1115)] = {
    [sym_attribute] = STATE(1115),
    [sym__lowercase_identifier] = ACTIONS(2133),
    [anon_sym_TILDE] = ACTIONS(2133),
    [anon_sym_QMARK] = ACTIONS(2133),
    [anon_sym_LPAREN] = ACTIONS(2135),
    [anon_sym_RPAREN] = ACTIONS(2135),
    [anon_sym_COMMA] = ACTIONS(2135),
    [anon_sym_PIPE] = ACTIONS(2133),
    [anon_sym_COLON2] = ACTIONS(2133),
    [anon_sym_DOT] = ACTIONS(2135),
    [anon_sym_DASH_GT] = ACTIONS(2133),
    [anon_sym_LBRACK] = ACTIONS(2133),
    [anon_sym_RBRACK] = ACTIONS(2135),
    [anon_sym_true] = ACTIONS(2133),
    [anon_sym_false] = ACTIONS(2133),
    [anon_sym_LBRACE] = ACTIONS(2133),
    [anon_sym_SEMI] = ACTIONS(2135),
    [anon_sym_RBRACE] = ACTIONS(2135),
    [anon_sym_constraint] = ACTIONS(2133),
    [anon_sym_val] = ACTIONS(2133),
    [anon_sym_end] = ACTIONS(2133),
    [anon_sym_with] = ACTIONS(2133),
    [anon_sym_COLON_GT] = ACTIONS(2135),
    [anon_sym_object] = ACTIONS(2133),
    [anon_sym_inherit] = ACTIONS(2133),
    [anon_sym_method] = ACTIONS(2133),
    [anon_sym_initializer] = ACTIONS(2133),
    [anon_sym_AMP] = ACTIONS(2133),
    [anon_sym_POUND] = ACTIONS(2133),
    [anon_sym_COLON_COLON] = ACTIONS(2135),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2135),
    [anon_sym_then] = ACTIONS(2133),
    [anon_sym_else] = ACTIONS(2133),
    [anon_sym_do] = ACTIONS(2133),
    [anon_sym_done] = ACTIONS(2133),
    [anon_sym_to] = ACTIONS(2133),
    [anon_sym_downto] = ACTIONS(2133),
    [anon_sym_new] = ACTIONS(2133),
    [anon_sym_LBRACE_LT] = ACTIONS(2135),
    [anon_sym_GT_RBRACE] = ACTIONS(2135),
    [anon_sym_begin] = ACTIONS(2133),
    [sym_ocamlyacc_value] = ACTIONS(2135),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2133),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2135),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2133),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2133),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2135),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2135),
    [aux_sym_number_token1] = ACTIONS(2135),
    [aux_sym_signed_number_token1] = ACTIONS(2133),
    [anon_sym_SQUOTE] = ACTIONS(2135),
    [anon_sym_DQUOTE] = ACTIONS(2135),
    [sym_prefix_operator] = ACTIONS(2135),
    [aux_sym_sign_operator_token1] = ACTIONS(2133),
    [sym_hash_operator] = ACTIONS(2135),
    [anon_sym_lsl] = ACTIONS(2133),
    [anon_sym_lsr] = ACTIONS(2133),
    [anon_sym_asr] = ACTIONS(2133),
    [aux_sym_pow_operator_token1] = ACTIONS(2135),
    [anon_sym_mod] = ACTIONS(2133),
    [anon_sym_land] = ACTIONS(2133),
    [anon_sym_lor] = ACTIONS(2133),
    [anon_sym_lxor] = ACTIONS(2133),
    [aux_sym_mult_operator_token1] = ACTIONS(2133),
    [aux_sym_add_operator_token1] = ACTIONS(2133),
    [sym_concat_operator] = ACTIONS(2135),
    [sym_rel_operator] = ACTIONS(2133),
    [anon_sym_AMP_AMP] = ACTIONS(2133),
    [anon_sym_or] = ACTIONS(2133),
    [anon_sym_PIPE_PIPE] = ACTIONS(2133),
    [sym_assign_operator] = ACTIONS(2135),
    [sym__uppercase_identifier] = ACTIONS(2135),
    [aux_sym_tag_token1] = ACTIONS(2135),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1116)] = {
    [sym_attribute] = STATE(1116),
    [sym__lowercase_identifier] = ACTIONS(2153),
    [anon_sym_TILDE] = ACTIONS(2153),
    [anon_sym_QMARK] = ACTIONS(2153),
    [anon_sym_LPAREN] = ACTIONS(2155),
    [anon_sym_RPAREN] = ACTIONS(2155),
    [anon_sym_COMMA] = ACTIONS(2155),
    [anon_sym_PIPE] = ACTIONS(2153),
    [anon_sym_COLON2] = ACTIONS(2153),
    [anon_sym_DOT] = ACTIONS(2155),
    [anon_sym_DASH_GT] = ACTIONS(2153),
    [anon_sym_LBRACK] = ACTIONS(2153),
    [anon_sym_RBRACK] = ACTIONS(2155),
    [anon_sym_true] = ACTIONS(2153),
    [anon_sym_false] = ACTIONS(2153),
    [anon_sym_LBRACE] = ACTIONS(2153),
    [anon_sym_SEMI] = ACTIONS(2155),
    [anon_sym_RBRACE] = ACTIONS(2155),
    [anon_sym_constraint] = ACTIONS(2153),
    [anon_sym_val] = ACTIONS(2153),
    [anon_sym_end] = ACTIONS(2153),
    [anon_sym_with] = ACTIONS(2153),
    [anon_sym_COLON_GT] = ACTIONS(2155),
    [anon_sym_object] = ACTIONS(2153),
    [anon_sym_inherit] = ACTIONS(2153),
    [anon_sym_method] = ACTIONS(2153),
    [anon_sym_initializer] = ACTIONS(2153),
    [anon_sym_AMP] = ACTIONS(2153),
    [anon_sym_POUND] = ACTIONS(2153),
    [anon_sym_COLON_COLON] = ACTIONS(2155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2155),
    [anon_sym_then] = ACTIONS(2153),
    [anon_sym_else] = ACTIONS(2153),
    [anon_sym_do] = ACTIONS(2153),
    [anon_sym_done] = ACTIONS(2153),
    [anon_sym_to] = ACTIONS(2153),
    [anon_sym_downto] = ACTIONS(2153),
    [anon_sym_new] = ACTIONS(2153),
    [anon_sym_LBRACE_LT] = ACTIONS(2155),
    [anon_sym_GT_RBRACE] = ACTIONS(2155),
    [anon_sym_begin] = ACTIONS(2153),
    [sym_ocamlyacc_value] = ACTIONS(2155),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2153),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2155),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2153),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2153),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2155),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2155),
    [aux_sym_number_token1] = ACTIONS(2155),
    [aux_sym_signed_number_token1] = ACTIONS(2153),
    [anon_sym_SQUOTE] = ACTIONS(2155),
    [anon_sym_DQUOTE] = ACTIONS(2155),
    [sym_prefix_operator] = ACTIONS(2155),
    [aux_sym_sign_operator_token1] = ACTIONS(2153),
    [sym_hash_operator] = ACTIONS(2155),
    [anon_sym_lsl] = ACTIONS(2153),
    [anon_sym_lsr] = ACTIONS(2153),
    [anon_sym_asr] = ACTIONS(2153),
    [aux_sym_pow_operator_token1] = ACTIONS(2155),
    [anon_sym_mod] = ACTIONS(2153),
    [anon_sym_land] = ACTIONS(2153),
    [anon_sym_lor] = ACTIONS(2153),
    [anon_sym_lxor] = ACTIONS(2153),
    [aux_sym_mult_operator_token1] = ACTIONS(2153),
    [aux_sym_add_operator_token1] = ACTIONS(2153),
    [sym_concat_operator] = ACTIONS(2155),
    [sym_rel_operator] = ACTIONS(2153),
    [anon_sym_AMP_AMP] = ACTIONS(2153),
    [anon_sym_or] = ACTIONS(2153),
    [anon_sym_PIPE_PIPE] = ACTIONS(2153),
    [sym_assign_operator] = ACTIONS(2155),
    [sym__uppercase_identifier] = ACTIONS(2155),
    [aux_sym_tag_token1] = ACTIONS(2155),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1117)] = {
    [sym_attribute] = STATE(1117),
    [sym__lowercase_identifier] = ACTIONS(2173),
    [anon_sym_TILDE] = ACTIONS(2173),
    [anon_sym_QMARK] = ACTIONS(2173),
    [anon_sym_LPAREN] = ACTIONS(2175),
    [anon_sym_RPAREN] = ACTIONS(2175),
    [anon_sym_COMMA] = ACTIONS(2175),
    [anon_sym_PIPE] = ACTIONS(2173),
    [anon_sym_COLON2] = ACTIONS(2173),
    [anon_sym_DOT] = ACTIONS(2175),
    [anon_sym_DASH_GT] = ACTIONS(2173),
    [anon_sym_LBRACK] = ACTIONS(2173),
    [anon_sym_RBRACK] = ACTIONS(2175),
    [anon_sym_true] = ACTIONS(2173),
    [anon_sym_false] = ACTIONS(2173),
    [anon_sym_LBRACE] = ACTIONS(2173),
    [anon_sym_SEMI] = ACTIONS(2175),
    [anon_sym_RBRACE] = ACTIONS(2175),
    [anon_sym_constraint] = ACTIONS(2173),
    [anon_sym_val] = ACTIONS(2173),
    [anon_sym_end] = ACTIONS(2173),
    [anon_sym_with] = ACTIONS(2173),
    [anon_sym_COLON_GT] = ACTIONS(2175),
    [anon_sym_object] = ACTIONS(2173),
    [anon_sym_inherit] = ACTIONS(2173),
    [anon_sym_method] = ACTIONS(2173),
    [anon_sym_initializer] = ACTIONS(2173),
    [anon_sym_AMP] = ACTIONS(2173),
    [anon_sym_POUND] = ACTIONS(2173),
    [anon_sym_COLON_COLON] = ACTIONS(2175),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2175),
    [anon_sym_then] = ACTIONS(2173),
    [anon_sym_else] = ACTIONS(2173),
    [anon_sym_do] = ACTIONS(2173),
    [anon_sym_done] = ACTIONS(2173),
    [anon_sym_to] = ACTIONS(2173),
    [anon_sym_downto] = ACTIONS(2173),
    [anon_sym_new] = ACTIONS(2173),
    [anon_sym_LBRACE_LT] = ACTIONS(2175),
    [anon_sym_GT_RBRACE] = ACTIONS(2175),
    [anon_sym_begin] = ACTIONS(2173),
    [sym_ocamlyacc_value] = ACTIONS(2175),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2173),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2175),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2173),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2175),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2175),
    [aux_sym_number_token1] = ACTIONS(2175),
    [aux_sym_signed_number_token1] = ACTIONS(2173),
    [anon_sym_SQUOTE] = ACTIONS(2175),
    [anon_sym_DQUOTE] = ACTIONS(2175),
    [sym_prefix_operator] = ACTIONS(2175),
    [aux_sym_sign_operator_token1] = ACTIONS(2173),
    [sym_hash_operator] = ACTIONS(2175),
    [anon_sym_lsl] = ACTIONS(2173),
    [anon_sym_lsr] = ACTIONS(2173),
    [anon_sym_asr] = ACTIONS(2173),
    [aux_sym_pow_operator_token1] = ACTIONS(2175),
    [anon_sym_mod] = ACTIONS(2173),
    [anon_sym_land] = ACTIONS(2173),
    [anon_sym_lor] = ACTIONS(2173),
    [anon_sym_lxor] = ACTIONS(2173),
    [aux_sym_mult_operator_token1] = ACTIONS(2173),
    [aux_sym_add_operator_token1] = ACTIONS(2173),
    [sym_concat_operator] = ACTIONS(2175),
    [sym_rel_operator] = ACTIONS(2173),
    [anon_sym_AMP_AMP] = ACTIONS(2173),
    [anon_sym_or] = ACTIONS(2173),
    [anon_sym_PIPE_PIPE] = ACTIONS(2173),
    [sym_assign_operator] = ACTIONS(2175),
    [sym__uppercase_identifier] = ACTIONS(2175),
    [aux_sym_tag_token1] = ACTIONS(2175),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1118)] = {
    [sym_attribute] = STATE(1118),
    [sym__lowercase_identifier] = ACTIONS(2165),
    [anon_sym_TILDE] = ACTIONS(2165),
    [anon_sym_QMARK] = ACTIONS(2165),
    [anon_sym_LPAREN] = ACTIONS(2167),
    [anon_sym_RPAREN] = ACTIONS(2167),
    [anon_sym_COMMA] = ACTIONS(2167),
    [anon_sym_PIPE] = ACTIONS(2165),
    [anon_sym_COLON2] = ACTIONS(2165),
    [anon_sym_DOT] = ACTIONS(2167),
    [anon_sym_DASH_GT] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2165),
    [anon_sym_RBRACK] = ACTIONS(2167),
    [anon_sym_true] = ACTIONS(2165),
    [anon_sym_false] = ACTIONS(2165),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_SEMI] = ACTIONS(2167),
    [anon_sym_RBRACE] = ACTIONS(2167),
    [anon_sym_constraint] = ACTIONS(2165),
    [anon_sym_val] = ACTIONS(2165),
    [anon_sym_end] = ACTIONS(2165),
    [anon_sym_with] = ACTIONS(2165),
    [anon_sym_COLON_GT] = ACTIONS(2167),
    [anon_sym_object] = ACTIONS(2165),
    [anon_sym_inherit] = ACTIONS(2165),
    [anon_sym_method] = ACTIONS(2165),
    [anon_sym_initializer] = ACTIONS(2165),
    [anon_sym_AMP] = ACTIONS(2165),
    [anon_sym_POUND] = ACTIONS(2165),
    [anon_sym_COLON_COLON] = ACTIONS(2167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2167),
    [anon_sym_then] = ACTIONS(2165),
    [anon_sym_else] = ACTIONS(2165),
    [anon_sym_do] = ACTIONS(2165),
    [anon_sym_done] = ACTIONS(2165),
    [anon_sym_to] = ACTIONS(2165),
    [anon_sym_downto] = ACTIONS(2165),
    [anon_sym_new] = ACTIONS(2165),
    [anon_sym_LBRACE_LT] = ACTIONS(2167),
    [anon_sym_GT_RBRACE] = ACTIONS(2167),
    [anon_sym_begin] = ACTIONS(2165),
    [sym_ocamlyacc_value] = ACTIONS(2167),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2165),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2167),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2167),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2167),
    [aux_sym_number_token1] = ACTIONS(2167),
    [aux_sym_signed_number_token1] = ACTIONS(2165),
    [anon_sym_SQUOTE] = ACTIONS(2167),
    [anon_sym_DQUOTE] = ACTIONS(2167),
    [sym_prefix_operator] = ACTIONS(2167),
    [aux_sym_sign_operator_token1] = ACTIONS(2165),
    [sym_hash_operator] = ACTIONS(2167),
    [anon_sym_lsl] = ACTIONS(2165),
    [anon_sym_lsr] = ACTIONS(2165),
    [anon_sym_asr] = ACTIONS(2165),
    [aux_sym_pow_operator_token1] = ACTIONS(2167),
    [anon_sym_mod] = ACTIONS(2165),
    [anon_sym_land] = ACTIONS(2165),
    [anon_sym_lor] = ACTIONS(2165),
    [anon_sym_lxor] = ACTIONS(2165),
    [aux_sym_mult_operator_token1] = ACTIONS(2165),
    [aux_sym_add_operator_token1] = ACTIONS(2165),
    [sym_concat_operator] = ACTIONS(2167),
    [sym_rel_operator] = ACTIONS(2165),
    [anon_sym_AMP_AMP] = ACTIONS(2165),
    [anon_sym_or] = ACTIONS(2165),
    [anon_sym_PIPE_PIPE] = ACTIONS(2165),
    [sym_assign_operator] = ACTIONS(2167),
    [sym__uppercase_identifier] = ACTIONS(2167),
    [aux_sym_tag_token1] = ACTIONS(2167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1119)] = {
    [sym_attribute] = STATE(1119),
    [sym__lowercase_identifier] = ACTIONS(2183),
    [anon_sym_TILDE] = ACTIONS(2183),
    [anon_sym_QMARK] = ACTIONS(2183),
    [anon_sym_LPAREN] = ACTIONS(2185),
    [anon_sym_RPAREN] = ACTIONS(2185),
    [anon_sym_COMMA] = ACTIONS(2185),
    [anon_sym_PIPE] = ACTIONS(2183),
    [anon_sym_COLON2] = ACTIONS(2183),
    [anon_sym_DOT] = ACTIONS(2185),
    [anon_sym_DASH_GT] = ACTIONS(2183),
    [anon_sym_LBRACK] = ACTIONS(2183),
    [anon_sym_RBRACK] = ACTIONS(2185),
    [anon_sym_true] = ACTIONS(2183),
    [anon_sym_false] = ACTIONS(2183),
    [anon_sym_LBRACE] = ACTIONS(2183),
    [anon_sym_SEMI] = ACTIONS(2185),
    [anon_sym_RBRACE] = ACTIONS(2185),
    [anon_sym_constraint] = ACTIONS(2183),
    [anon_sym_val] = ACTIONS(2183),
    [anon_sym_end] = ACTIONS(2183),
    [anon_sym_with] = ACTIONS(2183),
    [anon_sym_COLON_GT] = ACTIONS(2185),
    [anon_sym_object] = ACTIONS(2183),
    [anon_sym_inherit] = ACTIONS(2183),
    [anon_sym_method] = ACTIONS(2183),
    [anon_sym_initializer] = ACTIONS(2183),
    [anon_sym_AMP] = ACTIONS(2183),
    [anon_sym_POUND] = ACTIONS(2183),
    [anon_sym_COLON_COLON] = ACTIONS(2185),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2185),
    [anon_sym_then] = ACTIONS(2183),
    [anon_sym_else] = ACTIONS(2183),
    [anon_sym_do] = ACTIONS(2183),
    [anon_sym_done] = ACTIONS(2183),
    [anon_sym_to] = ACTIONS(2183),
    [anon_sym_downto] = ACTIONS(2183),
    [anon_sym_new] = ACTIONS(2183),
    [anon_sym_LBRACE_LT] = ACTIONS(2185),
    [anon_sym_GT_RBRACE] = ACTIONS(2185),
    [anon_sym_begin] = ACTIONS(2183),
    [sym_ocamlyacc_value] = ACTIONS(2185),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2183),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2185),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2183),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2183),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2185),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2185),
    [aux_sym_number_token1] = ACTIONS(2185),
    [aux_sym_signed_number_token1] = ACTIONS(2183),
    [anon_sym_SQUOTE] = ACTIONS(2185),
    [anon_sym_DQUOTE] = ACTIONS(2185),
    [sym_prefix_operator] = ACTIONS(2185),
    [aux_sym_sign_operator_token1] = ACTIONS(2183),
    [sym_hash_operator] = ACTIONS(2185),
    [anon_sym_lsl] = ACTIONS(2183),
    [anon_sym_lsr] = ACTIONS(2183),
    [anon_sym_asr] = ACTIONS(2183),
    [aux_sym_pow_operator_token1] = ACTIONS(2185),
    [anon_sym_mod] = ACTIONS(2183),
    [anon_sym_land] = ACTIONS(2183),
    [anon_sym_lor] = ACTIONS(2183),
    [anon_sym_lxor] = ACTIONS(2183),
    [aux_sym_mult_operator_token1] = ACTIONS(2183),
    [aux_sym_add_operator_token1] = ACTIONS(2183),
    [sym_concat_operator] = ACTIONS(2185),
    [sym_rel_operator] = ACTIONS(2183),
    [anon_sym_AMP_AMP] = ACTIONS(2183),
    [anon_sym_or] = ACTIONS(2183),
    [anon_sym_PIPE_PIPE] = ACTIONS(2183),
    [sym_assign_operator] = ACTIONS(2185),
    [sym__uppercase_identifier] = ACTIONS(2185),
    [aux_sym_tag_token1] = ACTIONS(2185),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1120)] = {
    [sym_attribute] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2187),
    [anon_sym_TILDE] = ACTIONS(2187),
    [anon_sym_QMARK] = ACTIONS(2187),
    [anon_sym_LPAREN] = ACTIONS(2189),
    [anon_sym_RPAREN] = ACTIONS(2189),
    [anon_sym_COMMA] = ACTIONS(2189),
    [anon_sym_PIPE] = ACTIONS(2187),
    [anon_sym_COLON2] = ACTIONS(2187),
    [anon_sym_DOT] = ACTIONS(2189),
    [anon_sym_DASH_GT] = ACTIONS(2187),
    [anon_sym_LBRACK] = ACTIONS(2187),
    [anon_sym_RBRACK] = ACTIONS(2189),
    [anon_sym_true] = ACTIONS(2187),
    [anon_sym_false] = ACTIONS(2187),
    [anon_sym_LBRACE] = ACTIONS(2187),
    [anon_sym_SEMI] = ACTIONS(2189),
    [anon_sym_RBRACE] = ACTIONS(2189),
    [anon_sym_constraint] = ACTIONS(2187),
    [anon_sym_val] = ACTIONS(2187),
    [anon_sym_end] = ACTIONS(2187),
    [anon_sym_with] = ACTIONS(2187),
    [anon_sym_COLON_GT] = ACTIONS(2189),
    [anon_sym_object] = ACTIONS(2187),
    [anon_sym_inherit] = ACTIONS(2187),
    [anon_sym_method] = ACTIONS(2187),
    [anon_sym_initializer] = ACTIONS(2187),
    [anon_sym_AMP] = ACTIONS(2187),
    [anon_sym_POUND] = ACTIONS(2187),
    [anon_sym_COLON_COLON] = ACTIONS(2189),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2189),
    [anon_sym_then] = ACTIONS(2187),
    [anon_sym_else] = ACTIONS(2187),
    [anon_sym_do] = ACTIONS(2187),
    [anon_sym_done] = ACTIONS(2187),
    [anon_sym_to] = ACTIONS(2187),
    [anon_sym_downto] = ACTIONS(2187),
    [anon_sym_new] = ACTIONS(2187),
    [anon_sym_LBRACE_LT] = ACTIONS(2189),
    [anon_sym_GT_RBRACE] = ACTIONS(2189),
    [anon_sym_begin] = ACTIONS(2187),
    [sym_ocamlyacc_value] = ACTIONS(2189),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2187),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2189),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2187),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2187),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2189),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2189),
    [aux_sym_number_token1] = ACTIONS(2189),
    [aux_sym_signed_number_token1] = ACTIONS(2187),
    [anon_sym_SQUOTE] = ACTIONS(2189),
    [anon_sym_DQUOTE] = ACTIONS(2189),
    [sym_prefix_operator] = ACTIONS(2189),
    [aux_sym_sign_operator_token1] = ACTIONS(2187),
    [sym_hash_operator] = ACTIONS(2189),
    [anon_sym_lsl] = ACTIONS(2187),
    [anon_sym_lsr] = ACTIONS(2187),
    [anon_sym_asr] = ACTIONS(2187),
    [aux_sym_pow_operator_token1] = ACTIONS(2189),
    [anon_sym_mod] = ACTIONS(2187),
    [anon_sym_land] = ACTIONS(2187),
    [anon_sym_lor] = ACTIONS(2187),
    [anon_sym_lxor] = ACTIONS(2187),
    [aux_sym_mult_operator_token1] = ACTIONS(2187),
    [aux_sym_add_operator_token1] = ACTIONS(2187),
    [sym_concat_operator] = ACTIONS(2189),
    [sym_rel_operator] = ACTIONS(2187),
    [anon_sym_AMP_AMP] = ACTIONS(2187),
    [anon_sym_or] = ACTIONS(2187),
    [anon_sym_PIPE_PIPE] = ACTIONS(2187),
    [sym_assign_operator] = ACTIONS(2189),
    [sym__uppercase_identifier] = ACTIONS(2189),
    [aux_sym_tag_token1] = ACTIONS(2189),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1121)] = {
    [sym_attribute] = STATE(1121),
    [sym__lowercase_identifier] = ACTIONS(2191),
    [anon_sym_TILDE] = ACTIONS(2191),
    [anon_sym_QMARK] = ACTIONS(2191),
    [anon_sym_LPAREN] = ACTIONS(2193),
    [anon_sym_RPAREN] = ACTIONS(2193),
    [anon_sym_COMMA] = ACTIONS(2193),
    [anon_sym_PIPE] = ACTIONS(2191),
    [anon_sym_COLON2] = ACTIONS(2191),
    [anon_sym_DOT] = ACTIONS(2193),
    [anon_sym_DASH_GT] = ACTIONS(2191),
    [anon_sym_LBRACK] = ACTIONS(2191),
    [anon_sym_RBRACK] = ACTIONS(2193),
    [anon_sym_true] = ACTIONS(2191),
    [anon_sym_false] = ACTIONS(2191),
    [anon_sym_LBRACE] = ACTIONS(2191),
    [anon_sym_SEMI] = ACTIONS(2193),
    [anon_sym_RBRACE] = ACTIONS(2193),
    [anon_sym_constraint] = ACTIONS(2191),
    [anon_sym_val] = ACTIONS(2191),
    [anon_sym_end] = ACTIONS(2191),
    [anon_sym_with] = ACTIONS(2191),
    [anon_sym_COLON_GT] = ACTIONS(2193),
    [anon_sym_object] = ACTIONS(2191),
    [anon_sym_inherit] = ACTIONS(2191),
    [anon_sym_method] = ACTIONS(2191),
    [anon_sym_initializer] = ACTIONS(2191),
    [anon_sym_AMP] = ACTIONS(2191),
    [anon_sym_POUND] = ACTIONS(2191),
    [anon_sym_COLON_COLON] = ACTIONS(2193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2193),
    [anon_sym_then] = ACTIONS(2191),
    [anon_sym_else] = ACTIONS(2191),
    [anon_sym_do] = ACTIONS(2191),
    [anon_sym_done] = ACTIONS(2191),
    [anon_sym_to] = ACTIONS(2191),
    [anon_sym_downto] = ACTIONS(2191),
    [anon_sym_new] = ACTIONS(2191),
    [anon_sym_LBRACE_LT] = ACTIONS(2193),
    [anon_sym_GT_RBRACE] = ACTIONS(2193),
    [anon_sym_begin] = ACTIONS(2191),
    [sym_ocamlyacc_value] = ACTIONS(2193),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2191),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2193),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2191),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2191),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2193),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2193),
    [aux_sym_number_token1] = ACTIONS(2193),
    [aux_sym_signed_number_token1] = ACTIONS(2191),
    [anon_sym_SQUOTE] = ACTIONS(2193),
    [anon_sym_DQUOTE] = ACTIONS(2193),
    [sym_prefix_operator] = ACTIONS(2193),
    [aux_sym_sign_operator_token1] = ACTIONS(2191),
    [sym_hash_operator] = ACTIONS(2193),
    [anon_sym_lsl] = ACTIONS(2191),
    [anon_sym_lsr] = ACTIONS(2191),
    [anon_sym_asr] = ACTIONS(2191),
    [aux_sym_pow_operator_token1] = ACTIONS(2193),
    [anon_sym_mod] = ACTIONS(2191),
    [anon_sym_land] = ACTIONS(2191),
    [anon_sym_lor] = ACTIONS(2191),
    [anon_sym_lxor] = ACTIONS(2191),
    [aux_sym_mult_operator_token1] = ACTIONS(2191),
    [aux_sym_add_operator_token1] = ACTIONS(2191),
    [sym_concat_operator] = ACTIONS(2193),
    [sym_rel_operator] = ACTIONS(2191),
    [anon_sym_AMP_AMP] = ACTIONS(2191),
    [anon_sym_or] = ACTIONS(2191),
    [anon_sym_PIPE_PIPE] = ACTIONS(2191),
    [sym_assign_operator] = ACTIONS(2193),
    [sym__uppercase_identifier] = ACTIONS(2193),
    [aux_sym_tag_token1] = ACTIONS(2193),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1122)] = {
    [sym_attribute] = STATE(1122),
    [sym__lowercase_identifier] = ACTIONS(1931),
    [anon_sym_TILDE] = ACTIONS(1931),
    [anon_sym_QMARK] = ACTIONS(1931),
    [anon_sym_LPAREN] = ACTIONS(1933),
    [anon_sym_RPAREN] = ACTIONS(1933),
    [anon_sym_COMMA] = ACTIONS(1933),
    [anon_sym_PIPE] = ACTIONS(1931),
    [anon_sym_COLON2] = ACTIONS(1931),
    [anon_sym_DOT] = ACTIONS(993),
    [anon_sym_DASH_GT] = ACTIONS(1931),
    [anon_sym_LBRACK] = ACTIONS(1931),
    [anon_sym_RBRACK] = ACTIONS(1933),
    [anon_sym_true] = ACTIONS(1931),
    [anon_sym_false] = ACTIONS(1931),
    [anon_sym_LBRACE] = ACTIONS(1931),
    [anon_sym_SEMI] = ACTIONS(1933),
    [anon_sym_RBRACE] = ACTIONS(1933),
    [anon_sym_constraint] = ACTIONS(1931),
    [anon_sym_val] = ACTIONS(1931),
    [anon_sym_end] = ACTIONS(1931),
    [anon_sym_with] = ACTIONS(1931),
    [anon_sym_COLON_GT] = ACTIONS(1933),
    [anon_sym_object] = ACTIONS(1931),
    [anon_sym_inherit] = ACTIONS(1931),
    [anon_sym_method] = ACTIONS(1931),
    [anon_sym_initializer] = ACTIONS(1931),
    [anon_sym_AMP] = ACTIONS(1931),
    [anon_sym_POUND] = ACTIONS(995),
    [anon_sym_COLON_COLON] = ACTIONS(1933),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1933),
    [anon_sym_then] = ACTIONS(1931),
    [anon_sym_else] = ACTIONS(1931),
    [anon_sym_do] = ACTIONS(1931),
    [anon_sym_done] = ACTIONS(1931),
    [anon_sym_to] = ACTIONS(1931),
    [anon_sym_downto] = ACTIONS(1931),
    [anon_sym_new] = ACTIONS(1931),
    [anon_sym_LBRACE_LT] = ACTIONS(1933),
    [anon_sym_GT_RBRACE] = ACTIONS(1933),
    [anon_sym_begin] = ACTIONS(1931),
    [sym_ocamlyacc_value] = ACTIONS(1933),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1931),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1933),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1933),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1933),
    [aux_sym_number_token1] = ACTIONS(1933),
    [aux_sym_signed_number_token1] = ACTIONS(1931),
    [anon_sym_SQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1933),
    [sym_prefix_operator] = ACTIONS(1933),
    [aux_sym_sign_operator_token1] = ACTIONS(1931),
    [sym_hash_operator] = ACTIONS(997),
    [anon_sym_lsl] = ACTIONS(1931),
    [anon_sym_lsr] = ACTIONS(1931),
    [anon_sym_asr] = ACTIONS(1931),
    [aux_sym_pow_operator_token1] = ACTIONS(1933),
    [anon_sym_mod] = ACTIONS(1931),
    [anon_sym_land] = ACTIONS(1931),
    [anon_sym_lor] = ACTIONS(1931),
    [anon_sym_lxor] = ACTIONS(1931),
    [aux_sym_mult_operator_token1] = ACTIONS(1931),
    [aux_sym_add_operator_token1] = ACTIONS(1931),
    [sym_concat_operator] = ACTIONS(1933),
    [sym_rel_operator] = ACTIONS(1931),
    [anon_sym_AMP_AMP] = ACTIONS(1931),
    [anon_sym_or] = ACTIONS(1931),
    [anon_sym_PIPE_PIPE] = ACTIONS(1931),
    [sym_assign_operator] = ACTIONS(1933),
    [sym__uppercase_identifier] = ACTIONS(1933),
    [aux_sym_tag_token1] = ACTIONS(1933),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1123)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4372),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1123),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4249),
    [aux_sym_let_binding_repeat1] = STATE(1109),
    [ts_builtin_sym_end] = ACTIONS(2331),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2331),
    [anon_sym_let] = ACTIONS(2333),
    [anon_sym_and] = ACTIONS(2333),
    [anon_sym_EQ] = ACTIONS(2459),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2333),
    [anon_sym_type] = ACTIONS(2333),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2333),
    [anon_sym_module] = ACTIONS(2333),
    [anon_sym_open] = ACTIONS(2333),
    [anon_sym_include] = ACTIONS(2333),
    [anon_sym_class] = ACTIONS(2333),
    [anon_sym_val] = ACTIONS(2333),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2331),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2331),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2331),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2331),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2333),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1124)] = {
    [sym_parameter] = STATE(2959),
    [sym__class_typed] = STATE(4465),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1124),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_expression_item_repeat1] = STATE(4332),
    [aux_sym_let_binding_repeat1] = STATE(1134),
    [ts_builtin_sym_end] = ACTIONS(2409),
    [sym__lowercase_identifier] = ACTIONS(2329),
    [anon_sym_SEMI_SEMI] = ACTIONS(2409),
    [anon_sym_let] = ACTIONS(2411),
    [anon_sym_and] = ACTIONS(2411),
    [anon_sym_EQ] = ACTIONS(2461),
    [anon_sym_TILDE] = ACTIONS(2337),
    [anon_sym_QMARK] = ACTIONS(2337),
    [anon_sym_LPAREN] = ACTIONS(2339),
    [anon_sym_external] = ACTIONS(2411),
    [anon_sym_type] = ACTIONS(2411),
    [anon_sym_COLON2] = ACTIONS(2341),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_true] = ACTIONS(2345),
    [anon_sym_false] = ACTIONS(2345),
    [anon_sym_LBRACE] = ACTIONS(2347),
    [anon_sym_exception] = ACTIONS(2411),
    [anon_sym_module] = ACTIONS(2411),
    [anon_sym_open] = ACTIONS(2411),
    [anon_sym_include] = ACTIONS(2411),
    [anon_sym_class] = ACTIONS(2411),
    [anon_sym_val] = ACTIONS(2411),
    [anon_sym_POUND] = ACTIONS(2349),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2351),
    [anon_sym_begin] = ACTIONS(2353),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2409),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2357),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2359),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2409),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2409),
    [aux_sym_number_token1] = ACTIONS(2361),
    [aux_sym_signed_number_token1] = ACTIONS(2363),
    [anon_sym_SQUOTE] = ACTIONS(2365),
    [anon_sym_DQUOTE] = ACTIONS(2367),
    [sym_let_operator] = ACTIONS(2409),
    [sym__uppercase_identifier] = ACTIONS(2369),
    [aux_sym_directive_token1] = ACTIONS(2411),
    [aux_sym_tag_token1] = ACTIONS(2371),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1125)] = {
    [sym_attribute] = STATE(1125),
    [sym__lowercase_identifier] = ACTIONS(1927),
    [anon_sym_TILDE] = ACTIONS(1927),
    [anon_sym_QMARK] = ACTIONS(1927),
    [anon_sym_LPAREN] = ACTIONS(1929),
    [anon_sym_RPAREN] = ACTIONS(1929),
    [anon_sym_COMMA] = ACTIONS(1929),
    [anon_sym_PIPE] = ACTIONS(1927),
    [anon_sym_COLON2] = ACTIONS(1927),
    [anon_sym_DOT] = ACTIONS(1929),
    [anon_sym_DASH_GT] = ACTIONS(1927),
    [anon_sym_LBRACK] = ACTIONS(1927),
    [anon_sym_RBRACK] = ACTIONS(1929),
    [anon_sym_true] = ACTIONS(1927),
    [anon_sym_false] = ACTIONS(1927),
    [anon_sym_LBRACE] = ACTIONS(1927),
    [anon_sym_SEMI] = ACTIONS(1929),
    [anon_sym_RBRACE] = ACTIONS(1929),
    [anon_sym_constraint] = ACTIONS(1927),
    [anon_sym_val] = ACTIONS(1927),
    [anon_sym_end] = ACTIONS(1927),
    [anon_sym_with] = ACTIONS(1927),
    [anon_sym_COLON_GT] = ACTIONS(1929),
    [anon_sym_object] = ACTIONS(1927),
    [anon_sym_inherit] = ACTIONS(1927),
    [anon_sym_method] = ACTIONS(1927),
    [anon_sym_initializer] = ACTIONS(1927),
    [anon_sym_AMP] = ACTIONS(1927),
    [anon_sym_POUND] = ACTIONS(1927),
    [anon_sym_COLON_COLON] = ACTIONS(1929),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1929),
    [anon_sym_then] = ACTIONS(1927),
    [anon_sym_else] = ACTIONS(1927),
    [anon_sym_do] = ACTIONS(1927),
    [anon_sym_done] = ACTIONS(1927),
    [anon_sym_to] = ACTIONS(1927),
    [anon_sym_downto] = ACTIONS(1927),
    [anon_sym_new] = ACTIONS(1927),
    [anon_sym_LBRACE_LT] = ACTIONS(1929),
    [anon_sym_GT_RBRACE] = ACTIONS(1929),
    [anon_sym_begin] = ACTIONS(1927),
    [sym_ocamlyacc_value] = ACTIONS(1929),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1927),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1929),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1927),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1927),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1929),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1929),
    [aux_sym_number_token1] = ACTIONS(1929),
    [aux_sym_signed_number_token1] = ACTIONS(1927),
    [anon_sym_SQUOTE] = ACTIONS(1929),
    [anon_sym_DQUOTE] = ACTIONS(1929),
    [sym_prefix_operator] = ACTIONS(1929),
    [aux_sym_sign_operator_token1] = ACTIONS(1927),
    [sym_hash_operator] = ACTIONS(1929),
    [anon_sym_lsl] = ACTIONS(1927),
    [anon_sym_lsr] = ACTIONS(1927),
    [anon_sym_asr] = ACTIONS(1927),
    [aux_sym_pow_operator_token1] = ACTIONS(1929),
    [anon_sym_mod] = ACTIONS(1927),
    [anon_sym_land] = ACTIONS(1927),
    [anon_sym_lor] = ACTIONS(1927),
    [anon_sym_lxor] = ACTIONS(1927),
    [aux_sym_mult_operator_token1] = ACTIONS(1927),
    [aux_sym_add_operator_token1] = ACTIONS(1927),
    [sym_concat_operator] = ACTIONS(1929),
    [sym_rel_operator] = ACTIONS(1927),
    [anon_sym_AMP_AMP] = ACTIONS(1927),
    [anon_sym_or] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1927),
    [sym_assign_operator] = ACTIONS(1929),
    [sym__uppercase_identifier] = ACTIONS(1929),
    [aux_sym_tag_token1] = ACTIONS(1929),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1126)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5043),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1126),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_EQ] = ACTIONS(2433),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(2433),
    [anon_sym_COMMA] = ACTIONS(2433),
    [anon_sym_PIPE] = ACTIONS(2433),
    [anon_sym_COLON2] = ACTIONS(2439),
    [anon_sym_DASH_GT] = ACTIONS(2433),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_RBRACK] = ACTIONS(2433),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_as] = ACTIONS(2439),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(2433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_when] = ACTIONS(2439),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2443),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1127)] = {
    [sym_attribute] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(1891),
    [anon_sym_SEMI_SEMI] = ACTIONS(1893),
    [anon_sym_let] = ACTIONS(1891),
    [anon_sym_and] = ACTIONS(1891),
    [anon_sym_TILDE] = ACTIONS(1891),
    [anon_sym_QMARK] = ACTIONS(1891),
    [anon_sym_LPAREN] = ACTIONS(1893),
    [anon_sym_external] = ACTIONS(1891),
    [anon_sym_type] = ACTIONS(1891),
    [anon_sym_COMMA] = ACTIONS(1893),
    [anon_sym_PIPE] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACK] = ACTIONS(1891),
    [anon_sym_RBRACK] = ACTIONS(1893),
    [anon_sym_true] = ACTIONS(1891),
    [anon_sym_false] = ACTIONS(1891),
    [anon_sym_LBRACE] = ACTIONS(1891),
    [anon_sym_SEMI] = ACTIONS(1891),
    [anon_sym_exception] = ACTIONS(1891),
    [anon_sym_module] = ACTIONS(1891),
    [anon_sym_open] = ACTIONS(1891),
    [anon_sym_include] = ACTIONS(1891),
    [anon_sym_class] = ACTIONS(1891),
    [anon_sym_val] = ACTIONS(1891),
    [anon_sym_end] = ACTIONS(1891),
    [anon_sym_object] = ACTIONS(1891),
    [anon_sym_in] = ACTIONS(1891),
    [anon_sym_AMP] = ACTIONS(1891),
    [anon_sym_POUND] = ACTIONS(1891),
    [anon_sym_COLON_COLON] = ACTIONS(1893),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1893),
    [anon_sym_LT_DASH] = ACTIONS(1891),
    [anon_sym_else] = ACTIONS(1891),
    [anon_sym_new] = ACTIONS(1891),
    [anon_sym_LBRACE_LT] = ACTIONS(1893),
    [anon_sym_begin] = ACTIONS(1891),
    [sym_ocamlyacc_value] = ACTIONS(1893),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1891),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1893),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1891),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1891),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1893),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1893),
    [aux_sym_number_token1] = ACTIONS(1893),
    [aux_sym_signed_number_token1] = ACTIONS(1891),
    [anon_sym_SQUOTE] = ACTIONS(1893),
    [anon_sym_DQUOTE] = ACTIONS(1893),
    [sym_prefix_operator] = ACTIONS(1893),
    [aux_sym_sign_operator_token1] = ACTIONS(1891),
    [sym_hash_operator] = ACTIONS(1893),
    [anon_sym_lsl] = ACTIONS(1891),
    [anon_sym_lsr] = ACTIONS(1891),
    [anon_sym_asr] = ACTIONS(1891),
    [aux_sym_pow_operator_token1] = ACTIONS(1893),
    [anon_sym_mod] = ACTIONS(1891),
    [anon_sym_land] = ACTIONS(1891),
    [anon_sym_lor] = ACTIONS(1891),
    [anon_sym_lxor] = ACTIONS(1891),
    [aux_sym_mult_operator_token1] = ACTIONS(1891),
    [aux_sym_add_operator_token1] = ACTIONS(1891),
    [sym_concat_operator] = ACTIONS(1893),
    [sym_rel_operator] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1891),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [sym_assign_operator] = ACTIONS(1893),
    [sym_let_operator] = ACTIONS(1893),
    [sym_let_and_operator] = ACTIONS(1893),
    [sym__uppercase_identifier] = ACTIONS(1893),
    [aux_sym_directive_token1] = ACTIONS(1891),
    [aux_sym_tag_token1] = ACTIONS(1893),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1128)] = {
    [sym_attribute] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(1887),
    [anon_sym_SEMI_SEMI] = ACTIONS(1889),
    [anon_sym_let] = ACTIONS(1887),
    [anon_sym_and] = ACTIONS(1887),
    [anon_sym_TILDE] = ACTIONS(1887),
    [anon_sym_QMARK] = ACTIONS(1887),
    [anon_sym_LPAREN] = ACTIONS(1889),
    [anon_sym_external] = ACTIONS(1887),
    [anon_sym_type] = ACTIONS(1887),
    [anon_sym_COMMA] = ACTIONS(1889),
    [anon_sym_PIPE] = ACTIONS(1887),
    [anon_sym_DOT] = ACTIONS(1889),
    [anon_sym_LBRACK] = ACTIONS(1887),
    [anon_sym_RBRACK] = ACTIONS(1889),
    [anon_sym_true] = ACTIONS(1887),
    [anon_sym_false] = ACTIONS(1887),
    [anon_sym_LBRACE] = ACTIONS(1887),
    [anon_sym_SEMI] = ACTIONS(1887),
    [anon_sym_exception] = ACTIONS(1887),
    [anon_sym_module] = ACTIONS(1887),
    [anon_sym_open] = ACTIONS(1887),
    [anon_sym_include] = ACTIONS(1887),
    [anon_sym_class] = ACTIONS(1887),
    [anon_sym_val] = ACTIONS(1887),
    [anon_sym_end] = ACTIONS(1887),
    [anon_sym_object] = ACTIONS(1887),
    [anon_sym_in] = ACTIONS(1887),
    [anon_sym_AMP] = ACTIONS(1887),
    [anon_sym_POUND] = ACTIONS(1887),
    [anon_sym_COLON_COLON] = ACTIONS(1889),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1889),
    [anon_sym_LT_DASH] = ACTIONS(1887),
    [anon_sym_else] = ACTIONS(1887),
    [anon_sym_new] = ACTIONS(1887),
    [anon_sym_LBRACE_LT] = ACTIONS(1889),
    [anon_sym_begin] = ACTIONS(1887),
    [sym_ocamlyacc_value] = ACTIONS(1889),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1887),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1889),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1887),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1887),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1889),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1889),
    [aux_sym_number_token1] = ACTIONS(1889),
    [aux_sym_signed_number_token1] = ACTIONS(1887),
    [anon_sym_SQUOTE] = ACTIONS(1889),
    [anon_sym_DQUOTE] = ACTIONS(1889),
    [sym_prefix_operator] = ACTIONS(1889),
    [aux_sym_sign_operator_token1] = ACTIONS(1887),
    [sym_hash_operator] = ACTIONS(1889),
    [anon_sym_lsl] = ACTIONS(1887),
    [anon_sym_lsr] = ACTIONS(1887),
    [anon_sym_asr] = ACTIONS(1887),
    [aux_sym_pow_operator_token1] = ACTIONS(1889),
    [anon_sym_mod] = ACTIONS(1887),
    [anon_sym_land] = ACTIONS(1887),
    [anon_sym_lor] = ACTIONS(1887),
    [anon_sym_lxor] = ACTIONS(1887),
    [aux_sym_mult_operator_token1] = ACTIONS(1887),
    [aux_sym_add_operator_token1] = ACTIONS(1887),
    [sym_concat_operator] = ACTIONS(1889),
    [sym_rel_operator] = ACTIONS(1887),
    [anon_sym_AMP_AMP] = ACTIONS(1887),
    [anon_sym_or] = ACTIONS(1887),
    [anon_sym_PIPE_PIPE] = ACTIONS(1887),
    [sym_assign_operator] = ACTIONS(1889),
    [sym_let_operator] = ACTIONS(1889),
    [sym_let_and_operator] = ACTIONS(1889),
    [sym__uppercase_identifier] = ACTIONS(1889),
    [aux_sym_directive_token1] = ACTIONS(1887),
    [aux_sym_tag_token1] = ACTIONS(1889),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1129)] = {
    [sym_attribute] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1911),
    [anon_sym_SEMI_SEMI] = ACTIONS(1913),
    [anon_sym_let] = ACTIONS(1911),
    [anon_sym_and] = ACTIONS(1911),
    [anon_sym_TILDE] = ACTIONS(1911),
    [anon_sym_QMARK] = ACTIONS(1911),
    [anon_sym_LPAREN] = ACTIONS(1913),
    [anon_sym_external] = ACTIONS(1911),
    [anon_sym_type] = ACTIONS(1911),
    [anon_sym_COMMA] = ACTIONS(1913),
    [anon_sym_PIPE] = ACTIONS(1911),
    [anon_sym_DOT] = ACTIONS(1913),
    [anon_sym_LBRACK] = ACTIONS(1911),
    [anon_sym_RBRACK] = ACTIONS(1913),
    [anon_sym_true] = ACTIONS(1911),
    [anon_sym_false] = ACTIONS(1911),
    [anon_sym_LBRACE] = ACTIONS(1911),
    [anon_sym_SEMI] = ACTIONS(1911),
    [anon_sym_exception] = ACTIONS(1911),
    [anon_sym_module] = ACTIONS(1911),
    [anon_sym_open] = ACTIONS(1911),
    [anon_sym_include] = ACTIONS(1911),
    [anon_sym_class] = ACTIONS(1911),
    [anon_sym_val] = ACTIONS(1911),
    [anon_sym_end] = ACTIONS(1911),
    [anon_sym_object] = ACTIONS(1911),
    [anon_sym_in] = ACTIONS(1911),
    [anon_sym_AMP] = ACTIONS(1911),
    [anon_sym_POUND] = ACTIONS(1911),
    [anon_sym_COLON_COLON] = ACTIONS(1913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1913),
    [anon_sym_LT_DASH] = ACTIONS(1911),
    [anon_sym_else] = ACTIONS(1911),
    [anon_sym_new] = ACTIONS(1911),
    [anon_sym_LBRACE_LT] = ACTIONS(1913),
    [anon_sym_begin] = ACTIONS(1911),
    [sym_ocamlyacc_value] = ACTIONS(1913),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1911),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1913),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1911),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1911),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1913),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1913),
    [aux_sym_number_token1] = ACTIONS(1913),
    [aux_sym_signed_number_token1] = ACTIONS(1911),
    [anon_sym_SQUOTE] = ACTIONS(1913),
    [anon_sym_DQUOTE] = ACTIONS(1913),
    [sym_prefix_operator] = ACTIONS(1913),
    [aux_sym_sign_operator_token1] = ACTIONS(1911),
    [sym_hash_operator] = ACTIONS(1913),
    [anon_sym_lsl] = ACTIONS(1911),
    [anon_sym_lsr] = ACTIONS(1911),
    [anon_sym_asr] = ACTIONS(1911),
    [aux_sym_pow_operator_token1] = ACTIONS(1913),
    [anon_sym_mod] = ACTIONS(1911),
    [anon_sym_land] = ACTIONS(1911),
    [anon_sym_lor] = ACTIONS(1911),
    [anon_sym_lxor] = ACTIONS(1911),
    [aux_sym_mult_operator_token1] = ACTIONS(1911),
    [aux_sym_add_operator_token1] = ACTIONS(1911),
    [sym_concat_operator] = ACTIONS(1913),
    [sym_rel_operator] = ACTIONS(1911),
    [anon_sym_AMP_AMP] = ACTIONS(1911),
    [anon_sym_or] = ACTIONS(1911),
    [anon_sym_PIPE_PIPE] = ACTIONS(1911),
    [sym_assign_operator] = ACTIONS(1913),
    [sym_let_operator] = ACTIONS(1913),
    [sym_let_and_operator] = ACTIONS(1913),
    [sym__uppercase_identifier] = ACTIONS(1913),
    [aux_sym_directive_token1] = ACTIONS(1911),
    [aux_sym_tag_token1] = ACTIONS(1913),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1130)] = {
    [sym_attribute] = STATE(1130),
    [sym__lowercase_identifier] = ACTIONS(1915),
    [anon_sym_SEMI_SEMI] = ACTIONS(1917),
    [anon_sym_let] = ACTIONS(1915),
    [anon_sym_and] = ACTIONS(1915),
    [anon_sym_TILDE] = ACTIONS(1915),
    [anon_sym_QMARK] = ACTIONS(1915),
    [anon_sym_LPAREN] = ACTIONS(1917),
    [anon_sym_external] = ACTIONS(1915),
    [anon_sym_type] = ACTIONS(1915),
    [anon_sym_COMMA] = ACTIONS(1917),
    [anon_sym_PIPE] = ACTIONS(1915),
    [anon_sym_DOT] = ACTIONS(1917),
    [anon_sym_LBRACK] = ACTIONS(1915),
    [anon_sym_RBRACK] = ACTIONS(1917),
    [anon_sym_true] = ACTIONS(1915),
    [anon_sym_false] = ACTIONS(1915),
    [anon_sym_LBRACE] = ACTIONS(1915),
    [anon_sym_SEMI] = ACTIONS(1915),
    [anon_sym_exception] = ACTIONS(1915),
    [anon_sym_module] = ACTIONS(1915),
    [anon_sym_open] = ACTIONS(1915),
    [anon_sym_include] = ACTIONS(1915),
    [anon_sym_class] = ACTIONS(1915),
    [anon_sym_val] = ACTIONS(1915),
    [anon_sym_end] = ACTIONS(1915),
    [anon_sym_object] = ACTIONS(1915),
    [anon_sym_in] = ACTIONS(1915),
    [anon_sym_AMP] = ACTIONS(1915),
    [anon_sym_POUND] = ACTIONS(1915),
    [anon_sym_COLON_COLON] = ACTIONS(1917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1917),
    [anon_sym_LT_DASH] = ACTIONS(1915),
    [anon_sym_else] = ACTIONS(1915),
    [anon_sym_new] = ACTIONS(1915),
    [anon_sym_LBRACE_LT] = ACTIONS(1917),
    [anon_sym_begin] = ACTIONS(1915),
    [sym_ocamlyacc_value] = ACTIONS(1917),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1915),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1917),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1915),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1915),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1917),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1917),
    [aux_sym_number_token1] = ACTIONS(1917),
    [aux_sym_signed_number_token1] = ACTIONS(1915),
    [anon_sym_SQUOTE] = ACTIONS(1917),
    [anon_sym_DQUOTE] = ACTIONS(1917),
    [sym_prefix_operator] = ACTIONS(1917),
    [aux_sym_sign_operator_token1] = ACTIONS(1915),
    [sym_hash_operator] = ACTIONS(1917),
    [anon_sym_lsl] = ACTIONS(1915),
    [anon_sym_lsr] = ACTIONS(1915),
    [anon_sym_asr] = ACTIONS(1915),
    [aux_sym_pow_operator_token1] = ACTIONS(1917),
    [anon_sym_mod] = ACTIONS(1915),
    [anon_sym_land] = ACTIONS(1915),
    [anon_sym_lor] = ACTIONS(1915),
    [anon_sym_lxor] = ACTIONS(1915),
    [aux_sym_mult_operator_token1] = ACTIONS(1915),
    [aux_sym_add_operator_token1] = ACTIONS(1915),
    [sym_concat_operator] = ACTIONS(1917),
    [sym_rel_operator] = ACTIONS(1915),
    [anon_sym_AMP_AMP] = ACTIONS(1915),
    [anon_sym_or] = ACTIONS(1915),
    [anon_sym_PIPE_PIPE] = ACTIONS(1915),
    [sym_assign_operator] = ACTIONS(1917),
    [sym_let_operator] = ACTIONS(1917),
    [sym_let_and_operator] = ACTIONS(1917),
    [sym__uppercase_identifier] = ACTIONS(1917),
    [aux_sym_directive_token1] = ACTIONS(1915),
    [aux_sym_tag_token1] = ACTIONS(1917),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1131)] = {
    [sym_attribute] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(1919),
    [anon_sym_SEMI_SEMI] = ACTIONS(1921),
    [anon_sym_let] = ACTIONS(1919),
    [anon_sym_and] = ACTIONS(1919),
    [anon_sym_TILDE] = ACTIONS(1919),
    [anon_sym_QMARK] = ACTIONS(1919),
    [anon_sym_LPAREN] = ACTIONS(1921),
    [anon_sym_external] = ACTIONS(1919),
    [anon_sym_type] = ACTIONS(1919),
    [anon_sym_COMMA] = ACTIONS(1921),
    [anon_sym_PIPE] = ACTIONS(1919),
    [anon_sym_DOT] = ACTIONS(1921),
    [anon_sym_LBRACK] = ACTIONS(1919),
    [anon_sym_RBRACK] = ACTIONS(1921),
    [anon_sym_true] = ACTIONS(1919),
    [anon_sym_false] = ACTIONS(1919),
    [anon_sym_LBRACE] = ACTIONS(1919),
    [anon_sym_SEMI] = ACTIONS(1919),
    [anon_sym_exception] = ACTIONS(1919),
    [anon_sym_module] = ACTIONS(1919),
    [anon_sym_open] = ACTIONS(1919),
    [anon_sym_include] = ACTIONS(1919),
    [anon_sym_class] = ACTIONS(1919),
    [anon_sym_val] = ACTIONS(1919),
    [anon_sym_end] = ACTIONS(1919),
    [anon_sym_object] = ACTIONS(1919),
    [anon_sym_in] = ACTIONS(1919),
    [anon_sym_AMP] = ACTIONS(1919),
    [anon_sym_POUND] = ACTIONS(1919),
    [anon_sym_COLON_COLON] = ACTIONS(1921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1921),
    [anon_sym_LT_DASH] = ACTIONS(1919),
    [anon_sym_else] = ACTIONS(1919),
    [anon_sym_new] = ACTIONS(1919),
    [anon_sym_LBRACE_LT] = ACTIONS(1921),
    [anon_sym_begin] = ACTIONS(1919),
    [sym_ocamlyacc_value] = ACTIONS(1921),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1919),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1921),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1919),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1919),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1921),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1921),
    [aux_sym_number_token1] = ACTIONS(1921),
    [aux_sym_signed_number_token1] = ACTIONS(1919),
    [anon_sym_SQUOTE] = ACTIONS(1921),
    [anon_sym_DQUOTE] = ACTIONS(1921),
    [sym_prefix_operator] = ACTIONS(1921),
    [aux_sym_sign_operator_token1] = ACTIONS(1919),
    [sym_hash_operator] = ACTIONS(1921),
    [anon_sym_lsl] = ACTIONS(1919),
    [anon_sym_lsr] = ACTIONS(1919),
    [anon_sym_asr] = ACTIONS(1919),
    [aux_sym_pow_operator_token1] = ACTIONS(1921),
    [anon_sym_mod] = ACTIONS(1919),
    [anon_sym_land] = ACTIONS(1919),
    [anon_sym_lor] = ACTIONS(1919),
    [anon_sym_lxor] = ACTIONS(1919),
    [aux_sym_mult_operator_token1] = ACTIONS(1919),
    [aux_sym_add_operator_token1] = ACTIONS(1919),
    [sym_concat_operator] = ACTIONS(1921),
    [sym_rel_operator] = ACTIONS(1919),
    [anon_sym_AMP_AMP] = ACTIONS(1919),
    [anon_sym_or] = ACTIONS(1919),
    [anon_sym_PIPE_PIPE] = ACTIONS(1919),
    [sym_assign_operator] = ACTIONS(1921),
    [sym_let_operator] = ACTIONS(1921),
    [sym_let_and_operator] = ACTIONS(1921),
    [sym__uppercase_identifier] = ACTIONS(1921),
    [aux_sym_directive_token1] = ACTIONS(1919),
    [aux_sym_tag_token1] = ACTIONS(1921),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1132)] = {
    [sym_attribute] = STATE(1132),
    [sym__lowercase_identifier] = ACTIONS(1899),
    [anon_sym_SEMI_SEMI] = ACTIONS(1901),
    [anon_sym_let] = ACTIONS(1899),
    [anon_sym_and] = ACTIONS(1899),
    [anon_sym_TILDE] = ACTIONS(1899),
    [anon_sym_QMARK] = ACTIONS(1899),
    [anon_sym_LPAREN] = ACTIONS(1901),
    [anon_sym_external] = ACTIONS(1899),
    [anon_sym_type] = ACTIONS(1899),
    [anon_sym_COMMA] = ACTIONS(1901),
    [anon_sym_PIPE] = ACTIONS(1899),
    [anon_sym_DOT] = ACTIONS(1901),
    [anon_sym_LBRACK] = ACTIONS(1899),
    [anon_sym_RBRACK] = ACTIONS(1901),
    [anon_sym_true] = ACTIONS(1899),
    [anon_sym_false] = ACTIONS(1899),
    [anon_sym_LBRACE] = ACTIONS(1899),
    [anon_sym_SEMI] = ACTIONS(1899),
    [anon_sym_exception] = ACTIONS(1899),
    [anon_sym_module] = ACTIONS(1899),
    [anon_sym_open] = ACTIONS(1899),
    [anon_sym_include] = ACTIONS(1899),
    [anon_sym_class] = ACTIONS(1899),
    [anon_sym_val] = ACTIONS(1899),
    [anon_sym_end] = ACTIONS(1899),
    [anon_sym_object] = ACTIONS(1899),
    [anon_sym_in] = ACTIONS(1899),
    [anon_sym_AMP] = ACTIONS(1899),
    [anon_sym_POUND] = ACTIONS(1899),
    [anon_sym_COLON_COLON] = ACTIONS(1901),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1901),
    [anon_sym_LT_DASH] = ACTIONS(1899),
    [anon_sym_else] = ACTIONS(1899),
    [anon_sym_new] = ACTIONS(1899),
    [anon_sym_LBRACE_LT] = ACTIONS(1901),
    [anon_sym_begin] = ACTIONS(1899),
    [sym_ocamlyacc_value] = ACTIONS(1901),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1899),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1901),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1899),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1899),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1901),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1901),
    [aux_sym_number_token1] = ACTIONS(1901),
    [aux_sym_signed_number_token1] = ACTIONS(1899),
    [anon_sym_SQUOTE] = ACTIONS(1901),
    [anon_sym_DQUOTE] = ACTIONS(1901),
    [sym_prefix_operator] = ACTIONS(1901),
    [aux_sym_sign_operator_token1] = ACTIONS(1899),
    [sym_hash_operator] = ACTIONS(1901),
    [anon_sym_lsl] = ACTIONS(1899),
    [anon_sym_lsr] = ACTIONS(1899),
    [anon_sym_asr] = ACTIONS(1899),
    [aux_sym_pow_operator_token1] = ACTIONS(1901),
    [anon_sym_mod] = ACTIONS(1899),
    [anon_sym_land] = ACTIONS(1899),
    [anon_sym_lor] = ACTIONS(1899),
    [anon_sym_lxor] = ACTIONS(1899),
    [aux_sym_mult_operator_token1] = ACTIONS(1899),
    [aux_sym_add_operator_token1] = ACTIONS(1899),
    [sym_concat_operator] = ACTIONS(1901),
    [sym_rel_operator] = ACTIONS(1899),
    [anon_sym_AMP_AMP] = ACTIONS(1899),
    [anon_sym_or] = ACTIONS(1899),
    [anon_sym_PIPE_PIPE] = ACTIONS(1899),
    [sym_assign_operator] = ACTIONS(1901),
    [sym_let_operator] = ACTIONS(1901),
    [sym_let_and_operator] = ACTIONS(1901),
    [sym__uppercase_identifier] = ACTIONS(1901),
    [aux_sym_directive_token1] = ACTIONS(1899),
    [aux_sym_tag_token1] = ACTIONS(1901),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1133)] = {
    [sym_attribute] = STATE(1133),
    [sym__lowercase_identifier] = ACTIONS(1903),
    [anon_sym_SEMI_SEMI] = ACTIONS(1905),
    [anon_sym_let] = ACTIONS(1903),
    [anon_sym_and] = ACTIONS(1903),
    [anon_sym_TILDE] = ACTIONS(1903),
    [anon_sym_QMARK] = ACTIONS(1903),
    [anon_sym_LPAREN] = ACTIONS(1905),
    [anon_sym_external] = ACTIONS(1903),
    [anon_sym_type] = ACTIONS(1903),
    [anon_sym_COMMA] = ACTIONS(1905),
    [anon_sym_PIPE] = ACTIONS(1903),
    [anon_sym_DOT] = ACTIONS(1905),
    [anon_sym_LBRACK] = ACTIONS(1903),
    [anon_sym_RBRACK] = ACTIONS(1905),
    [anon_sym_true] = ACTIONS(1903),
    [anon_sym_false] = ACTIONS(1903),
    [anon_sym_LBRACE] = ACTIONS(1903),
    [anon_sym_SEMI] = ACTIONS(1903),
    [anon_sym_exception] = ACTIONS(1903),
    [anon_sym_module] = ACTIONS(1903),
    [anon_sym_open] = ACTIONS(1903),
    [anon_sym_include] = ACTIONS(1903),
    [anon_sym_class] = ACTIONS(1903),
    [anon_sym_val] = ACTIONS(1903),
    [anon_sym_end] = ACTIONS(1903),
    [anon_sym_object] = ACTIONS(1903),
    [anon_sym_in] = ACTIONS(1903),
    [anon_sym_AMP] = ACTIONS(1903),
    [anon_sym_POUND] = ACTIONS(1903),
    [anon_sym_COLON_COLON] = ACTIONS(1905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1905),
    [anon_sym_LT_DASH] = ACTIONS(1903),
    [anon_sym_else] = ACTIONS(1903),
    [anon_sym_new] = ACTIONS(1903),
    [anon_sym_LBRACE_LT] = ACTIONS(1905),
    [anon_sym_begin] = ACTIONS(1903),
    [sym_ocamlyacc_value] = ACTIONS(1905),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1903),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1905),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1903),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1903),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1905),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1905),
    [aux_sym_number_token1] = ACTIONS(1905),
    [aux_sym_signed_number_token1] = ACTIONS(1903),
    [anon_sym_SQUOTE] = ACTIONS(1905),
    [anon_sym_DQUOTE] = ACTIONS(1905),
    [sym_prefix_operator] = ACTIONS(1905),
    [aux_sym_sign_operator_token1] = ACTIONS(1903),
    [sym_hash_operator] = ACTIONS(1905),
    [anon_sym_lsl] = ACTIONS(1903),
    [anon_sym_lsr] = ACTIONS(1903),
    [anon_sym_asr] = ACTIONS(1903),
    [aux_sym_pow_operator_token1] = ACTIONS(1905),
    [anon_sym_mod] = ACTIONS(1903),
    [anon_sym_land] = ACTIONS(1903),
    [anon_sym_lor] = ACTIONS(1903),
    [anon_sym_lxor] = ACTIONS(1903),
    [aux_sym_mult_operator_token1] = ACTIONS(1903),
    [aux_sym_add_operator_token1] = ACTIONS(1903),
    [sym_concat_operator] = ACTIONS(1905),
    [sym_rel_operator] = ACTIONS(1903),
    [anon_sym_AMP_AMP] = ACTIONS(1903),
    [anon_sym_or] = ACTIONS(1903),
    [anon_sym_PIPE_PIPE] = ACTIONS(1903),
    [sym_assign_operator] = ACTIONS(1905),
    [sym_let_operator] = ACTIONS(1905),
    [sym_let_and_operator] = ACTIONS(1905),
    [sym__uppercase_identifier] = ACTIONS(1905),
    [aux_sym_directive_token1] = ACTIONS(1903),
    [aux_sym_tag_token1] = ACTIONS(1905),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1134)] = {
    [sym_parameter] = STATE(2959),
    [sym__parenthesized_abstract_type] = STATE(2959),
    [sym__delimited_pattern] = STATE(2927),
    [sym__simple_pattern] = STATE(2962),
    [sym_typed_pattern] = STATE(2927),
    [sym_polymorphic_variant_pattern] = STATE(2927),
    [sym_record_pattern] = STATE(2928),
    [sym_list_pattern] = STATE(2928),
    [sym_array_pattern] = STATE(2928),
    [sym_local_open_pattern] = STATE(2927),
    [sym_package_pattern] = STATE(2927),
    [sym_parenthesized_pattern] = STATE(2928),
    [sym_attribute] = STATE(1134),
    [sym_extension] = STATE(2927),
    [sym_quoted_extension] = STATE(2927),
    [sym__constant] = STATE(3022),
    [sym__signed_constant] = STATE(2927),
    [sym_number] = STATE(3036),
    [sym_signed_number] = STATE(3022),
    [sym_character] = STATE(3036),
    [sym_string] = STATE(3036),
    [sym_quoted_string] = STATE(3036),
    [sym_boolean] = STATE(3036),
    [sym_unit] = STATE(2928),
    [sym__simple_value_pattern] = STATE(2927),
    [sym_parenthesized_operator] = STATE(2927),
    [sym_module_path] = STATE(7756),
    [sym_constructor_path] = STATE(2927),
    [sym__constructor_name] = STATE(3053),
    [sym_tag] = STATE(2927),
    [aux_sym_let_binding_repeat1] = STATE(1134),
    [ts_builtin_sym_end] = ACTIONS(2463),
    [sym__lowercase_identifier] = ACTIONS(2465),
    [anon_sym_SEMI_SEMI] = ACTIONS(2463),
    [anon_sym_let] = ACTIONS(2468),
    [anon_sym_and] = ACTIONS(2468),
    [anon_sym_EQ] = ACTIONS(2463),
    [anon_sym_TILDE] = ACTIONS(2470),
    [anon_sym_QMARK] = ACTIONS(2470),
    [anon_sym_LPAREN] = ACTIONS(2473),
    [anon_sym_external] = ACTIONS(2468),
    [anon_sym_type] = ACTIONS(2468),
    [anon_sym_COLON2] = ACTIONS(2463),
    [anon_sym_LBRACK] = ACTIONS(2476),
    [anon_sym_RBRACK] = ACTIONS(2463),
    [anon_sym_true] = ACTIONS(2479),
    [anon_sym_false] = ACTIONS(2479),
    [anon_sym_LBRACE] = ACTIONS(2482),
    [anon_sym_exception] = ACTIONS(2468),
    [anon_sym_module] = ACTIONS(2468),
    [anon_sym_open] = ACTIONS(2468),
    [anon_sym_include] = ACTIONS(2468),
    [anon_sym_class] = ACTIONS(2468),
    [anon_sym_val] = ACTIONS(2468),
    [anon_sym_end] = ACTIONS(2468),
    [anon_sym_POUND] = ACTIONS(2485),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2488),
    [anon_sym_begin] = ACTIONS(2491),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2468),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2463),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2494),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2497),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2463),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2463),
    [aux_sym_number_token1] = ACTIONS(2500),
    [aux_sym_signed_number_token1] = ACTIONS(2503),
    [anon_sym_SQUOTE] = ACTIONS(2506),
    [anon_sym_DQUOTE] = ACTIONS(2509),
    [sym_let_operator] = ACTIONS(2463),
    [sym__uppercase_identifier] = ACTIONS(2512),
    [aux_sym_directive_token1] = ACTIONS(2468),
    [aux_sym_tag_token1] = ACTIONS(2515),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1135)] = {
    [sym_attribute] = STATE(1135),
    [sym__lowercase_identifier] = ACTIONS(1895),
    [anon_sym_SEMI_SEMI] = ACTIONS(1897),
    [anon_sym_let] = ACTIONS(1895),
    [anon_sym_and] = ACTIONS(1895),
    [anon_sym_TILDE] = ACTIONS(1895),
    [anon_sym_QMARK] = ACTIONS(1895),
    [anon_sym_LPAREN] = ACTIONS(1897),
    [anon_sym_external] = ACTIONS(1895),
    [anon_sym_type] = ACTIONS(1895),
    [anon_sym_COMMA] = ACTIONS(1897),
    [anon_sym_PIPE] = ACTIONS(1895),
    [anon_sym_DOT] = ACTIONS(1897),
    [anon_sym_LBRACK] = ACTIONS(1895),
    [anon_sym_RBRACK] = ACTIONS(1897),
    [anon_sym_true] = ACTIONS(1895),
    [anon_sym_false] = ACTIONS(1895),
    [anon_sym_LBRACE] = ACTIONS(1895),
    [anon_sym_SEMI] = ACTIONS(1895),
    [anon_sym_exception] = ACTIONS(1895),
    [anon_sym_module] = ACTIONS(1895),
    [anon_sym_open] = ACTIONS(1895),
    [anon_sym_include] = ACTIONS(1895),
    [anon_sym_class] = ACTIONS(1895),
    [anon_sym_val] = ACTIONS(1895),
    [anon_sym_end] = ACTIONS(1895),
    [anon_sym_object] = ACTIONS(1895),
    [anon_sym_in] = ACTIONS(1895),
    [anon_sym_AMP] = ACTIONS(1895),
    [anon_sym_POUND] = ACTIONS(1895),
    [anon_sym_COLON_COLON] = ACTIONS(1897),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1897),
    [anon_sym_LT_DASH] = ACTIONS(1895),
    [anon_sym_else] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1895),
    [anon_sym_LBRACE_LT] = ACTIONS(1897),
    [anon_sym_begin] = ACTIONS(1895),
    [sym_ocamlyacc_value] = ACTIONS(1897),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1895),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1897),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1895),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1895),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1897),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1897),
    [aux_sym_number_token1] = ACTIONS(1897),
    [aux_sym_signed_number_token1] = ACTIONS(1895),
    [anon_sym_SQUOTE] = ACTIONS(1897),
    [anon_sym_DQUOTE] = ACTIONS(1897),
    [sym_prefix_operator] = ACTIONS(1897),
    [aux_sym_sign_operator_token1] = ACTIONS(1895),
    [sym_hash_operator] = ACTIONS(1897),
    [anon_sym_lsl] = ACTIONS(1895),
    [anon_sym_lsr] = ACTIONS(1895),
    [anon_sym_asr] = ACTIONS(1895),
    [aux_sym_pow_operator_token1] = ACTIONS(1897),
    [anon_sym_mod] = ACTIONS(1895),
    [anon_sym_land] = ACTIONS(1895),
    [anon_sym_lor] = ACTIONS(1895),
    [anon_sym_lxor] = ACTIONS(1895),
    [aux_sym_mult_operator_token1] = ACTIONS(1895),
    [aux_sym_add_operator_token1] = ACTIONS(1895),
    [sym_concat_operator] = ACTIONS(1897),
    [sym_rel_operator] = ACTIONS(1895),
    [anon_sym_AMP_AMP] = ACTIONS(1895),
    [anon_sym_or] = ACTIONS(1895),
    [anon_sym_PIPE_PIPE] = ACTIONS(1895),
    [sym_assign_operator] = ACTIONS(1897),
    [sym_let_operator] = ACTIONS(1897),
    [sym_let_and_operator] = ACTIONS(1897),
    [sym__uppercase_identifier] = ACTIONS(1897),
    [aux_sym_directive_token1] = ACTIONS(1895),
    [aux_sym_tag_token1] = ACTIONS(1897),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1136)] = {
    [sym_attribute] = STATE(1136),
    [sym__lowercase_identifier] = ACTIONS(1907),
    [anon_sym_SEMI_SEMI] = ACTIONS(1909),
    [anon_sym_let] = ACTIONS(1907),
    [anon_sym_and] = ACTIONS(1907),
    [anon_sym_TILDE] = ACTIONS(1907),
    [anon_sym_QMARK] = ACTIONS(1907),
    [anon_sym_LPAREN] = ACTIONS(1909),
    [anon_sym_external] = ACTIONS(1907),
    [anon_sym_type] = ACTIONS(1907),
    [anon_sym_COMMA] = ACTIONS(1909),
    [anon_sym_PIPE] = ACTIONS(1907),
    [anon_sym_DOT] = ACTIONS(1909),
    [anon_sym_LBRACK] = ACTIONS(1907),
    [anon_sym_RBRACK] = ACTIONS(1909),
    [anon_sym_true] = ACTIONS(1907),
    [anon_sym_false] = ACTIONS(1907),
    [anon_sym_LBRACE] = ACTIONS(1907),
    [anon_sym_SEMI] = ACTIONS(1907),
    [anon_sym_exception] = ACTIONS(1907),
    [anon_sym_module] = ACTIONS(1907),
    [anon_sym_open] = ACTIONS(1907),
    [anon_sym_include] = ACTIONS(1907),
    [anon_sym_class] = ACTIONS(1907),
    [anon_sym_val] = ACTIONS(1907),
    [anon_sym_end] = ACTIONS(1907),
    [anon_sym_object] = ACTIONS(1907),
    [anon_sym_in] = ACTIONS(1907),
    [anon_sym_AMP] = ACTIONS(1907),
    [anon_sym_POUND] = ACTIONS(1907),
    [anon_sym_COLON_COLON] = ACTIONS(1909),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1909),
    [anon_sym_LT_DASH] = ACTIONS(1907),
    [anon_sym_else] = ACTIONS(1907),
    [anon_sym_new] = ACTIONS(1907),
    [anon_sym_LBRACE_LT] = ACTIONS(1909),
    [anon_sym_begin] = ACTIONS(1907),
    [sym_ocamlyacc_value] = ACTIONS(1909),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1907),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1909),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1907),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1907),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1909),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1909),
    [aux_sym_number_token1] = ACTIONS(1909),
    [aux_sym_signed_number_token1] = ACTIONS(1907),
    [anon_sym_SQUOTE] = ACTIONS(1909),
    [anon_sym_DQUOTE] = ACTIONS(1909),
    [sym_prefix_operator] = ACTIONS(1909),
    [aux_sym_sign_operator_token1] = ACTIONS(1907),
    [sym_hash_operator] = ACTIONS(1909),
    [anon_sym_lsl] = ACTIONS(1907),
    [anon_sym_lsr] = ACTIONS(1907),
    [anon_sym_asr] = ACTIONS(1907),
    [aux_sym_pow_operator_token1] = ACTIONS(1909),
    [anon_sym_mod] = ACTIONS(1907),
    [anon_sym_land] = ACTIONS(1907),
    [anon_sym_lor] = ACTIONS(1907),
    [anon_sym_lxor] = ACTIONS(1907),
    [aux_sym_mult_operator_token1] = ACTIONS(1907),
    [aux_sym_add_operator_token1] = ACTIONS(1907),
    [sym_concat_operator] = ACTIONS(1909),
    [sym_rel_operator] = ACTIONS(1907),
    [anon_sym_AMP_AMP] = ACTIONS(1907),
    [anon_sym_or] = ACTIONS(1907),
    [anon_sym_PIPE_PIPE] = ACTIONS(1907),
    [sym_assign_operator] = ACTIONS(1909),
    [sym_let_operator] = ACTIONS(1909),
    [sym_let_and_operator] = ACTIONS(1909),
    [sym__uppercase_identifier] = ACTIONS(1909),
    [aux_sym_directive_token1] = ACTIONS(1907),
    [aux_sym_tag_token1] = ACTIONS(1909),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1137)] = {
    [sym_attribute] = STATE(1137),
    [sym__lowercase_identifier] = ACTIONS(1947),
    [anon_sym_SEMI_SEMI] = ACTIONS(1949),
    [anon_sym_let] = ACTIONS(1947),
    [anon_sym_and] = ACTIONS(1947),
    [anon_sym_TILDE] = ACTIONS(1947),
    [anon_sym_QMARK] = ACTIONS(1947),
    [anon_sym_LPAREN] = ACTIONS(1949),
    [anon_sym_external] = ACTIONS(1947),
    [anon_sym_type] = ACTIONS(1947),
    [anon_sym_COMMA] = ACTIONS(1949),
    [anon_sym_PIPE] = ACTIONS(1947),
    [anon_sym_DOT] = ACTIONS(1949),
    [anon_sym_LBRACK] = ACTIONS(1947),
    [anon_sym_RBRACK] = ACTIONS(1949),
    [anon_sym_true] = ACTIONS(1947),
    [anon_sym_false] = ACTIONS(1947),
    [anon_sym_LBRACE] = ACTIONS(1947),
    [anon_sym_SEMI] = ACTIONS(1947),
    [anon_sym_exception] = ACTIONS(1947),
    [anon_sym_module] = ACTIONS(1947),
    [anon_sym_open] = ACTIONS(1947),
    [anon_sym_include] = ACTIONS(1947),
    [anon_sym_class] = ACTIONS(1947),
    [anon_sym_val] = ACTIONS(1947),
    [anon_sym_end] = ACTIONS(1947),
    [anon_sym_object] = ACTIONS(1947),
    [anon_sym_in] = ACTIONS(1947),
    [anon_sym_AMP] = ACTIONS(1947),
    [anon_sym_POUND] = ACTIONS(1947),
    [anon_sym_COLON_COLON] = ACTIONS(1949),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1949),
    [anon_sym_LT_DASH] = ACTIONS(2518),
    [anon_sym_else] = ACTIONS(1947),
    [anon_sym_new] = ACTIONS(1947),
    [anon_sym_LBRACE_LT] = ACTIONS(1949),
    [anon_sym_begin] = ACTIONS(1947),
    [sym_ocamlyacc_value] = ACTIONS(1949),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1947),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1949),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1949),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1949),
    [aux_sym_number_token1] = ACTIONS(1949),
    [aux_sym_signed_number_token1] = ACTIONS(1947),
    [anon_sym_SQUOTE] = ACTIONS(1949),
    [anon_sym_DQUOTE] = ACTIONS(1949),
    [sym_prefix_operator] = ACTIONS(1949),
    [aux_sym_sign_operator_token1] = ACTIONS(1947),
    [sym_hash_operator] = ACTIONS(1949),
    [anon_sym_lsl] = ACTIONS(1947),
    [anon_sym_lsr] = ACTIONS(1947),
    [anon_sym_asr] = ACTIONS(1947),
    [aux_sym_pow_operator_token1] = ACTIONS(1949),
    [anon_sym_mod] = ACTIONS(1947),
    [anon_sym_land] = ACTIONS(1947),
    [anon_sym_lor] = ACTIONS(1947),
    [anon_sym_lxor] = ACTIONS(1947),
    [aux_sym_mult_operator_token1] = ACTIONS(1947),
    [aux_sym_add_operator_token1] = ACTIONS(1947),
    [sym_concat_operator] = ACTIONS(1949),
    [sym_rel_operator] = ACTIONS(1947),
    [anon_sym_AMP_AMP] = ACTIONS(1947),
    [anon_sym_or] = ACTIONS(1947),
    [anon_sym_PIPE_PIPE] = ACTIONS(1947),
    [sym_assign_operator] = ACTIONS(1949),
    [sym_let_operator] = ACTIONS(1949),
    [sym_let_and_operator] = ACTIONS(1949),
    [sym__uppercase_identifier] = ACTIONS(1949),
    [aux_sym_directive_token1] = ACTIONS(1947),
    [aux_sym_tag_token1] = ACTIONS(1949),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1138)] = {
    [sym_attribute] = STATE(1138),
    [sym__lowercase_identifier] = ACTIONS(2157),
    [anon_sym_SEMI_SEMI] = ACTIONS(2159),
    [anon_sym_let] = ACTIONS(2157),
    [anon_sym_and] = ACTIONS(2157),
    [anon_sym_TILDE] = ACTIONS(2157),
    [anon_sym_QMARK] = ACTIONS(2157),
    [anon_sym_LPAREN] = ACTIONS(2159),
    [anon_sym_external] = ACTIONS(2157),
    [anon_sym_type] = ACTIONS(2157),
    [anon_sym_COMMA] = ACTIONS(2159),
    [anon_sym_PIPE] = ACTIONS(2157),
    [anon_sym_DOT] = ACTIONS(2159),
    [anon_sym_LBRACK] = ACTIONS(2157),
    [anon_sym_RBRACK] = ACTIONS(2159),
    [anon_sym_true] = ACTIONS(2157),
    [anon_sym_false] = ACTIONS(2157),
    [anon_sym_LBRACE] = ACTIONS(2157),
    [anon_sym_SEMI] = ACTIONS(2157),
    [anon_sym_exception] = ACTIONS(2157),
    [anon_sym_module] = ACTIONS(2157),
    [anon_sym_open] = ACTIONS(2157),
    [anon_sym_include] = ACTIONS(2157),
    [anon_sym_class] = ACTIONS(2157),
    [anon_sym_val] = ACTIONS(2157),
    [anon_sym_end] = ACTIONS(2157),
    [anon_sym_object] = ACTIONS(2157),
    [anon_sym_in] = ACTIONS(2157),
    [anon_sym_AMP] = ACTIONS(2157),
    [anon_sym_POUND] = ACTIONS(2157),
    [anon_sym_COLON_COLON] = ACTIONS(2159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2159),
    [anon_sym_LT_DASH] = ACTIONS(2317),
    [anon_sym_else] = ACTIONS(2157),
    [anon_sym_new] = ACTIONS(2157),
    [anon_sym_LBRACE_LT] = ACTIONS(2159),
    [anon_sym_begin] = ACTIONS(2157),
    [sym_ocamlyacc_value] = ACTIONS(2159),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2157),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2159),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2159),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2159),
    [aux_sym_number_token1] = ACTIONS(2159),
    [aux_sym_signed_number_token1] = ACTIONS(2157),
    [anon_sym_SQUOTE] = ACTIONS(2159),
    [anon_sym_DQUOTE] = ACTIONS(2159),
    [sym_prefix_operator] = ACTIONS(2159),
    [aux_sym_sign_operator_token1] = ACTIONS(2157),
    [sym_hash_operator] = ACTIONS(2159),
    [anon_sym_lsl] = ACTIONS(2157),
    [anon_sym_lsr] = ACTIONS(2157),
    [anon_sym_asr] = ACTIONS(2157),
    [aux_sym_pow_operator_token1] = ACTIONS(2159),
    [anon_sym_mod] = ACTIONS(2157),
    [anon_sym_land] = ACTIONS(2157),
    [anon_sym_lor] = ACTIONS(2157),
    [anon_sym_lxor] = ACTIONS(2157),
    [aux_sym_mult_operator_token1] = ACTIONS(2157),
    [aux_sym_add_operator_token1] = ACTIONS(2157),
    [sym_concat_operator] = ACTIONS(2159),
    [sym_rel_operator] = ACTIONS(2157),
    [anon_sym_AMP_AMP] = ACTIONS(2157),
    [anon_sym_or] = ACTIONS(2157),
    [anon_sym_PIPE_PIPE] = ACTIONS(2157),
    [sym_assign_operator] = ACTIONS(2159),
    [sym_let_operator] = ACTIONS(2159),
    [sym_let_and_operator] = ACTIONS(2159),
    [sym__uppercase_identifier] = ACTIONS(2159),
    [aux_sym_directive_token1] = ACTIONS(2157),
    [aux_sym_tag_token1] = ACTIONS(2159),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1139)] = {
    [sym_attribute] = STATE(1139),
    [sym__lowercase_identifier] = ACTIONS(2107),
    [anon_sym_SEMI_SEMI] = ACTIONS(2109),
    [anon_sym_let] = ACTIONS(2107),
    [anon_sym_and] = ACTIONS(2107),
    [anon_sym_TILDE] = ACTIONS(2107),
    [anon_sym_QMARK] = ACTIONS(2107),
    [anon_sym_LPAREN] = ACTIONS(2109),
    [anon_sym_external] = ACTIONS(2107),
    [anon_sym_type] = ACTIONS(2107),
    [anon_sym_COMMA] = ACTIONS(2109),
    [anon_sym_PIPE] = ACTIONS(2107),
    [anon_sym_DOT] = ACTIONS(2109),
    [anon_sym_LBRACK] = ACTIONS(2107),
    [anon_sym_RBRACK] = ACTIONS(2109),
    [anon_sym_true] = ACTIONS(2107),
    [anon_sym_false] = ACTIONS(2107),
    [anon_sym_LBRACE] = ACTIONS(2107),
    [anon_sym_SEMI] = ACTIONS(2107),
    [anon_sym_exception] = ACTIONS(2107),
    [anon_sym_module] = ACTIONS(2107),
    [anon_sym_open] = ACTIONS(2107),
    [anon_sym_include] = ACTIONS(2107),
    [anon_sym_class] = ACTIONS(2107),
    [anon_sym_val] = ACTIONS(2107),
    [anon_sym_end] = ACTIONS(2107),
    [anon_sym_object] = ACTIONS(2107),
    [anon_sym_in] = ACTIONS(2107),
    [anon_sym_AMP] = ACTIONS(2107),
    [anon_sym_POUND] = ACTIONS(2107),
    [anon_sym_COLON_COLON] = ACTIONS(2109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2109),
    [anon_sym_else] = ACTIONS(2107),
    [anon_sym_new] = ACTIONS(2107),
    [anon_sym_LBRACE_LT] = ACTIONS(2109),
    [anon_sym_begin] = ACTIONS(2107),
    [sym_ocamlyacc_value] = ACTIONS(2109),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2107),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2109),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2109),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2109),
    [aux_sym_number_token1] = ACTIONS(2109),
    [aux_sym_signed_number_token1] = ACTIONS(2107),
    [anon_sym_SQUOTE] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(2109),
    [sym_prefix_operator] = ACTIONS(2109),
    [aux_sym_sign_operator_token1] = ACTIONS(2107),
    [sym_hash_operator] = ACTIONS(2109),
    [anon_sym_lsl] = ACTIONS(2107),
    [anon_sym_lsr] = ACTIONS(2107),
    [anon_sym_asr] = ACTIONS(2107),
    [aux_sym_pow_operator_token1] = ACTIONS(2109),
    [anon_sym_mod] = ACTIONS(2107),
    [anon_sym_land] = ACTIONS(2107),
    [anon_sym_lor] = ACTIONS(2107),
    [anon_sym_lxor] = ACTIONS(2107),
    [aux_sym_mult_operator_token1] = ACTIONS(2107),
    [aux_sym_add_operator_token1] = ACTIONS(2107),
    [sym_concat_operator] = ACTIONS(2109),
    [sym_rel_operator] = ACTIONS(2107),
    [anon_sym_AMP_AMP] = ACTIONS(2107),
    [anon_sym_or] = ACTIONS(2107),
    [anon_sym_PIPE_PIPE] = ACTIONS(2107),
    [sym_assign_operator] = ACTIONS(2109),
    [sym_let_operator] = ACTIONS(2109),
    [sym_let_and_operator] = ACTIONS(2109),
    [sym__uppercase_identifier] = ACTIONS(2109),
    [aux_sym_directive_token1] = ACTIONS(2107),
    [aux_sym_tag_token1] = ACTIONS(2109),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1140)] = {
    [sym_attribute] = STATE(1140),
    [sym__lowercase_identifier] = ACTIONS(2133),
    [anon_sym_SEMI_SEMI] = ACTIONS(2135),
    [anon_sym_let] = ACTIONS(2133),
    [anon_sym_and] = ACTIONS(2133),
    [anon_sym_TILDE] = ACTIONS(2133),
    [anon_sym_QMARK] = ACTIONS(2133),
    [anon_sym_LPAREN] = ACTIONS(2135),
    [anon_sym_external] = ACTIONS(2133),
    [anon_sym_type] = ACTIONS(2133),
    [anon_sym_COMMA] = ACTIONS(2135),
    [anon_sym_PIPE] = ACTIONS(2133),
    [anon_sym_DOT] = ACTIONS(2135),
    [anon_sym_LBRACK] = ACTIONS(2133),
    [anon_sym_RBRACK] = ACTIONS(2135),
    [anon_sym_true] = ACTIONS(2133),
    [anon_sym_false] = ACTIONS(2133),
    [anon_sym_LBRACE] = ACTIONS(2133),
    [anon_sym_SEMI] = ACTIONS(2133),
    [anon_sym_exception] = ACTIONS(2133),
    [anon_sym_module] = ACTIONS(2133),
    [anon_sym_open] = ACTIONS(2133),
    [anon_sym_include] = ACTIONS(2133),
    [anon_sym_class] = ACTIONS(2133),
    [anon_sym_val] = ACTIONS(2133),
    [anon_sym_end] = ACTIONS(2133),
    [anon_sym_object] = ACTIONS(2133),
    [anon_sym_in] = ACTIONS(2133),
    [anon_sym_AMP] = ACTIONS(2133),
    [anon_sym_POUND] = ACTIONS(2133),
    [anon_sym_COLON_COLON] = ACTIONS(2135),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2135),
    [anon_sym_else] = ACTIONS(2133),
    [anon_sym_new] = ACTIONS(2133),
    [anon_sym_LBRACE_LT] = ACTIONS(2135),
    [anon_sym_begin] = ACTIONS(2133),
    [sym_ocamlyacc_value] = ACTIONS(2135),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2133),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2135),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2133),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2133),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2135),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2135),
    [aux_sym_number_token1] = ACTIONS(2135),
    [aux_sym_signed_number_token1] = ACTIONS(2133),
    [anon_sym_SQUOTE] = ACTIONS(2135),
    [anon_sym_DQUOTE] = ACTIONS(2135),
    [sym_prefix_operator] = ACTIONS(2135),
    [aux_sym_sign_operator_token1] = ACTIONS(2133),
    [sym_hash_operator] = ACTIONS(2135),
    [anon_sym_lsl] = ACTIONS(2133),
    [anon_sym_lsr] = ACTIONS(2133),
    [anon_sym_asr] = ACTIONS(2133),
    [aux_sym_pow_operator_token1] = ACTIONS(2135),
    [anon_sym_mod] = ACTIONS(2133),
    [anon_sym_land] = ACTIONS(2133),
    [anon_sym_lor] = ACTIONS(2133),
    [anon_sym_lxor] = ACTIONS(2133),
    [aux_sym_mult_operator_token1] = ACTIONS(2133),
    [aux_sym_add_operator_token1] = ACTIONS(2133),
    [sym_concat_operator] = ACTIONS(2135),
    [sym_rel_operator] = ACTIONS(2133),
    [anon_sym_AMP_AMP] = ACTIONS(2133),
    [anon_sym_or] = ACTIONS(2133),
    [anon_sym_PIPE_PIPE] = ACTIONS(2133),
    [sym_assign_operator] = ACTIONS(2135),
    [sym_let_operator] = ACTIONS(2135),
    [sym_let_and_operator] = ACTIONS(2135),
    [sym__uppercase_identifier] = ACTIONS(2135),
    [aux_sym_directive_token1] = ACTIONS(2133),
    [aux_sym_tag_token1] = ACTIONS(2135),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1141)] = {
    [sym_attribute] = STATE(1141),
    [sym__lowercase_identifier] = ACTIONS(2117),
    [anon_sym_SEMI_SEMI] = ACTIONS(2119),
    [anon_sym_let] = ACTIONS(2117),
    [anon_sym_and] = ACTIONS(2117),
    [anon_sym_TILDE] = ACTIONS(2117),
    [anon_sym_QMARK] = ACTIONS(2117),
    [anon_sym_LPAREN] = ACTIONS(2119),
    [anon_sym_external] = ACTIONS(2117),
    [anon_sym_type] = ACTIONS(2117),
    [anon_sym_COMMA] = ACTIONS(2119),
    [anon_sym_PIPE] = ACTIONS(2117),
    [anon_sym_DOT] = ACTIONS(2119),
    [anon_sym_LBRACK] = ACTIONS(2117),
    [anon_sym_RBRACK] = ACTIONS(2119),
    [anon_sym_true] = ACTIONS(2117),
    [anon_sym_false] = ACTIONS(2117),
    [anon_sym_LBRACE] = ACTIONS(2117),
    [anon_sym_SEMI] = ACTIONS(2117),
    [anon_sym_exception] = ACTIONS(2117),
    [anon_sym_module] = ACTIONS(2117),
    [anon_sym_open] = ACTIONS(2117),
    [anon_sym_include] = ACTIONS(2117),
    [anon_sym_class] = ACTIONS(2117),
    [anon_sym_val] = ACTIONS(2117),
    [anon_sym_end] = ACTIONS(2117),
    [anon_sym_object] = ACTIONS(2117),
    [anon_sym_in] = ACTIONS(2117),
    [anon_sym_AMP] = ACTIONS(2117),
    [anon_sym_POUND] = ACTIONS(2117),
    [anon_sym_COLON_COLON] = ACTIONS(2119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2119),
    [anon_sym_else] = ACTIONS(2117),
    [anon_sym_new] = ACTIONS(2117),
    [anon_sym_LBRACE_LT] = ACTIONS(2119),
    [anon_sym_begin] = ACTIONS(2117),
    [sym_ocamlyacc_value] = ACTIONS(2119),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2117),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2119),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2117),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2117),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2119),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2119),
    [aux_sym_number_token1] = ACTIONS(2119),
    [aux_sym_signed_number_token1] = ACTIONS(2117),
    [anon_sym_SQUOTE] = ACTIONS(2119),
    [anon_sym_DQUOTE] = ACTIONS(2119),
    [sym_prefix_operator] = ACTIONS(2119),
    [aux_sym_sign_operator_token1] = ACTIONS(2117),
    [sym_hash_operator] = ACTIONS(2119),
    [anon_sym_lsl] = ACTIONS(2117),
    [anon_sym_lsr] = ACTIONS(2117),
    [anon_sym_asr] = ACTIONS(2117),
    [aux_sym_pow_operator_token1] = ACTIONS(2119),
    [anon_sym_mod] = ACTIONS(2117),
    [anon_sym_land] = ACTIONS(2117),
    [anon_sym_lor] = ACTIONS(2117),
    [anon_sym_lxor] = ACTIONS(2117),
    [aux_sym_mult_operator_token1] = ACTIONS(2117),
    [aux_sym_add_operator_token1] = ACTIONS(2117),
    [sym_concat_operator] = ACTIONS(2119),
    [sym_rel_operator] = ACTIONS(2117),
    [anon_sym_AMP_AMP] = ACTIONS(2117),
    [anon_sym_or] = ACTIONS(2117),
    [anon_sym_PIPE_PIPE] = ACTIONS(2117),
    [sym_assign_operator] = ACTIONS(2119),
    [sym_let_operator] = ACTIONS(2119),
    [sym_let_and_operator] = ACTIONS(2119),
    [sym__uppercase_identifier] = ACTIONS(2119),
    [aux_sym_directive_token1] = ACTIONS(2117),
    [aux_sym_tag_token1] = ACTIONS(2119),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1142)] = {
    [sym_attribute] = STATE(1142),
    [sym__lowercase_identifier] = ACTIONS(2153),
    [anon_sym_SEMI_SEMI] = ACTIONS(2155),
    [anon_sym_let] = ACTIONS(2153),
    [anon_sym_and] = ACTIONS(2153),
    [anon_sym_TILDE] = ACTIONS(2153),
    [anon_sym_QMARK] = ACTIONS(2153),
    [anon_sym_LPAREN] = ACTIONS(2155),
    [anon_sym_external] = ACTIONS(2153),
    [anon_sym_type] = ACTIONS(2153),
    [anon_sym_COMMA] = ACTIONS(2155),
    [anon_sym_PIPE] = ACTIONS(2153),
    [anon_sym_DOT] = ACTIONS(2155),
    [anon_sym_LBRACK] = ACTIONS(2153),
    [anon_sym_RBRACK] = ACTIONS(2155),
    [anon_sym_true] = ACTIONS(2153),
    [anon_sym_false] = ACTIONS(2153),
    [anon_sym_LBRACE] = ACTIONS(2153),
    [anon_sym_SEMI] = ACTIONS(2153),
    [anon_sym_exception] = ACTIONS(2153),
    [anon_sym_module] = ACTIONS(2153),
    [anon_sym_open] = ACTIONS(2153),
    [anon_sym_include] = ACTIONS(2153),
    [anon_sym_class] = ACTIONS(2153),
    [anon_sym_val] = ACTIONS(2153),
    [anon_sym_end] = ACTIONS(2153),
    [anon_sym_object] = ACTIONS(2153),
    [anon_sym_in] = ACTIONS(2153),
    [anon_sym_AMP] = ACTIONS(2153),
    [anon_sym_POUND] = ACTIONS(2153),
    [anon_sym_COLON_COLON] = ACTIONS(2155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2155),
    [anon_sym_else] = ACTIONS(2153),
    [anon_sym_new] = ACTIONS(2153),
    [anon_sym_LBRACE_LT] = ACTIONS(2155),
    [anon_sym_begin] = ACTIONS(2153),
    [sym_ocamlyacc_value] = ACTIONS(2155),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2153),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2155),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2153),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2153),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2155),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2155),
    [aux_sym_number_token1] = ACTIONS(2155),
    [aux_sym_signed_number_token1] = ACTIONS(2153),
    [anon_sym_SQUOTE] = ACTIONS(2155),
    [anon_sym_DQUOTE] = ACTIONS(2155),
    [sym_prefix_operator] = ACTIONS(2155),
    [aux_sym_sign_operator_token1] = ACTIONS(2153),
    [sym_hash_operator] = ACTIONS(2155),
    [anon_sym_lsl] = ACTIONS(2153),
    [anon_sym_lsr] = ACTIONS(2153),
    [anon_sym_asr] = ACTIONS(2153),
    [aux_sym_pow_operator_token1] = ACTIONS(2155),
    [anon_sym_mod] = ACTIONS(2153),
    [anon_sym_land] = ACTIONS(2153),
    [anon_sym_lor] = ACTIONS(2153),
    [anon_sym_lxor] = ACTIONS(2153),
    [aux_sym_mult_operator_token1] = ACTIONS(2153),
    [aux_sym_add_operator_token1] = ACTIONS(2153),
    [sym_concat_operator] = ACTIONS(2155),
    [sym_rel_operator] = ACTIONS(2153),
    [anon_sym_AMP_AMP] = ACTIONS(2153),
    [anon_sym_or] = ACTIONS(2153),
    [anon_sym_PIPE_PIPE] = ACTIONS(2153),
    [sym_assign_operator] = ACTIONS(2155),
    [sym_let_operator] = ACTIONS(2155),
    [sym_let_and_operator] = ACTIONS(2155),
    [sym__uppercase_identifier] = ACTIONS(2155),
    [aux_sym_directive_token1] = ACTIONS(2153),
    [aux_sym_tag_token1] = ACTIONS(2155),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1143)] = {
    [sym_attribute] = STATE(1143),
    [ts_builtin_sym_end] = ACTIONS(1949),
    [sym__lowercase_identifier] = ACTIONS(1947),
    [anon_sym_SEMI_SEMI] = ACTIONS(1949),
    [anon_sym_let] = ACTIONS(1947),
    [anon_sym_and] = ACTIONS(1947),
    [anon_sym_TILDE] = ACTIONS(1947),
    [anon_sym_QMARK] = ACTIONS(1947),
    [anon_sym_LPAREN] = ACTIONS(1949),
    [anon_sym_external] = ACTIONS(1947),
    [anon_sym_type] = ACTIONS(1947),
    [anon_sym_COMMA] = ACTIONS(1949),
    [anon_sym_PIPE] = ACTIONS(1947),
    [anon_sym_DOT] = ACTIONS(1949),
    [anon_sym_LBRACK] = ACTIONS(1947),
    [anon_sym_true] = ACTIONS(1947),
    [anon_sym_false] = ACTIONS(1947),
    [anon_sym_LBRACE] = ACTIONS(1947),
    [anon_sym_SEMI] = ACTIONS(1947),
    [anon_sym_exception] = ACTIONS(1947),
    [anon_sym_module] = ACTIONS(1947),
    [anon_sym_open] = ACTIONS(1947),
    [anon_sym_include] = ACTIONS(1947),
    [anon_sym_class] = ACTIONS(1947),
    [anon_sym_val] = ACTIONS(1947),
    [anon_sym_object] = ACTIONS(1947),
    [anon_sym_in] = ACTIONS(1947),
    [anon_sym_AMP] = ACTIONS(1947),
    [anon_sym_POUND] = ACTIONS(1947),
    [anon_sym_COLON_COLON] = ACTIONS(1949),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1949),
    [anon_sym_LT_DASH] = ACTIONS(2520),
    [anon_sym_else] = ACTIONS(1947),
    [anon_sym_new] = ACTIONS(1947),
    [anon_sym_LBRACE_LT] = ACTIONS(1949),
    [anon_sym_begin] = ACTIONS(1947),
    [sym_ocamlyacc_value] = ACTIONS(1949),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1947),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1949),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1949),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1949),
    [aux_sym_number_token1] = ACTIONS(1949),
    [aux_sym_signed_number_token1] = ACTIONS(1947),
    [anon_sym_SQUOTE] = ACTIONS(1949),
    [anon_sym_DQUOTE] = ACTIONS(1949),
    [sym_prefix_operator] = ACTIONS(1949),
    [aux_sym_sign_operator_token1] = ACTIONS(1947),
    [sym_hash_operator] = ACTIONS(1949),
    [anon_sym_lsl] = ACTIONS(1947),
    [anon_sym_lsr] = ACTIONS(1947),
    [anon_sym_asr] = ACTIONS(1947),
    [aux_sym_pow_operator_token1] = ACTIONS(1949),
    [anon_sym_mod] = ACTIONS(1947),
    [anon_sym_land] = ACTIONS(1947),
    [anon_sym_lor] = ACTIONS(1947),
    [anon_sym_lxor] = ACTIONS(1947),
    [aux_sym_mult_operator_token1] = ACTIONS(1947),
    [aux_sym_add_operator_token1] = ACTIONS(1947),
    [sym_concat_operator] = ACTIONS(1949),
    [sym_rel_operator] = ACTIONS(1947),
    [anon_sym_AMP_AMP] = ACTIONS(1947),
    [anon_sym_or] = ACTIONS(1947),
    [anon_sym_PIPE_PIPE] = ACTIONS(1947),
    [sym_assign_operator] = ACTIONS(1949),
    [sym_let_operator] = ACTIONS(1949),
    [sym_let_and_operator] = ACTIONS(1949),
    [sym__uppercase_identifier] = ACTIONS(1949),
    [aux_sym_directive_token1] = ACTIONS(1947),
    [aux_sym_tag_token1] = ACTIONS(1949),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1144)] = {
    [sym_attribute] = STATE(1144),
    [sym__lowercase_identifier] = ACTIONS(2027),
    [anon_sym_SEMI_SEMI] = ACTIONS(2029),
    [anon_sym_let] = ACTIONS(2027),
    [anon_sym_and] = ACTIONS(2027),
    [anon_sym_TILDE] = ACTIONS(2027),
    [anon_sym_QMARK] = ACTIONS(2027),
    [anon_sym_LPAREN] = ACTIONS(2029),
    [anon_sym_external] = ACTIONS(2027),
    [anon_sym_type] = ACTIONS(2027),
    [anon_sym_COMMA] = ACTIONS(2029),
    [anon_sym_PIPE] = ACTIONS(2027),
    [anon_sym_DOT] = ACTIONS(2029),
    [anon_sym_LBRACK] = ACTIONS(2027),
    [anon_sym_RBRACK] = ACTIONS(2029),
    [anon_sym_true] = ACTIONS(2027),
    [anon_sym_false] = ACTIONS(2027),
    [anon_sym_LBRACE] = ACTIONS(2027),
    [anon_sym_SEMI] = ACTIONS(2027),
    [anon_sym_exception] = ACTIONS(2027),
    [anon_sym_module] = ACTIONS(2027),
    [anon_sym_open] = ACTIONS(2027),
    [anon_sym_include] = ACTIONS(2027),
    [anon_sym_class] = ACTIONS(2027),
    [anon_sym_val] = ACTIONS(2027),
    [anon_sym_end] = ACTIONS(2027),
    [anon_sym_object] = ACTIONS(2027),
    [anon_sym_in] = ACTIONS(2027),
    [anon_sym_AMP] = ACTIONS(2027),
    [anon_sym_POUND] = ACTIONS(2027),
    [anon_sym_COLON_COLON] = ACTIONS(2029),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2029),
    [anon_sym_else] = ACTIONS(2027),
    [anon_sym_new] = ACTIONS(2027),
    [anon_sym_LBRACE_LT] = ACTIONS(2029),
    [anon_sym_begin] = ACTIONS(2027),
    [sym_ocamlyacc_value] = ACTIONS(2029),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2027),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2029),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2029),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2029),
    [aux_sym_number_token1] = ACTIONS(2029),
    [aux_sym_signed_number_token1] = ACTIONS(2027),
    [anon_sym_SQUOTE] = ACTIONS(2029),
    [anon_sym_DQUOTE] = ACTIONS(2029),
    [sym_prefix_operator] = ACTIONS(2029),
    [aux_sym_sign_operator_token1] = ACTIONS(2027),
    [sym_hash_operator] = ACTIONS(2029),
    [anon_sym_lsl] = ACTIONS(2027),
    [anon_sym_lsr] = ACTIONS(2027),
    [anon_sym_asr] = ACTIONS(2027),
    [aux_sym_pow_operator_token1] = ACTIONS(2029),
    [anon_sym_mod] = ACTIONS(2027),
    [anon_sym_land] = ACTIONS(2027),
    [anon_sym_lor] = ACTIONS(2027),
    [anon_sym_lxor] = ACTIONS(2027),
    [aux_sym_mult_operator_token1] = ACTIONS(2027),
    [aux_sym_add_operator_token1] = ACTIONS(2027),
    [sym_concat_operator] = ACTIONS(2029),
    [sym_rel_operator] = ACTIONS(2027),
    [anon_sym_AMP_AMP] = ACTIONS(2027),
    [anon_sym_or] = ACTIONS(2027),
    [anon_sym_PIPE_PIPE] = ACTIONS(2027),
    [sym_assign_operator] = ACTIONS(2029),
    [sym_let_operator] = ACTIONS(2029),
    [sym_let_and_operator] = ACTIONS(2029),
    [sym__uppercase_identifier] = ACTIONS(2029),
    [aux_sym_directive_token1] = ACTIONS(2027),
    [aux_sym_tag_token1] = ACTIONS(2029),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1145)] = {
    [sym_attribute] = STATE(1145),
    [sym__lowercase_identifier] = ACTIONS(2125),
    [anon_sym_SEMI_SEMI] = ACTIONS(2127),
    [anon_sym_let] = ACTIONS(2125),
    [anon_sym_and] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(2125),
    [anon_sym_QMARK] = ACTIONS(2125),
    [anon_sym_LPAREN] = ACTIONS(2127),
    [anon_sym_external] = ACTIONS(2125),
    [anon_sym_type] = ACTIONS(2125),
    [anon_sym_COMMA] = ACTIONS(2127),
    [anon_sym_PIPE] = ACTIONS(2125),
    [anon_sym_DOT] = ACTIONS(2127),
    [anon_sym_LBRACK] = ACTIONS(2125),
    [anon_sym_RBRACK] = ACTIONS(2127),
    [anon_sym_true] = ACTIONS(2125),
    [anon_sym_false] = ACTIONS(2125),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_SEMI] = ACTIONS(2125),
    [anon_sym_exception] = ACTIONS(2125),
    [anon_sym_module] = ACTIONS(2125),
    [anon_sym_open] = ACTIONS(2125),
    [anon_sym_include] = ACTIONS(2125),
    [anon_sym_class] = ACTIONS(2125),
    [anon_sym_val] = ACTIONS(2125),
    [anon_sym_end] = ACTIONS(2125),
    [anon_sym_object] = ACTIONS(2125),
    [anon_sym_in] = ACTIONS(2125),
    [anon_sym_AMP] = ACTIONS(2125),
    [anon_sym_POUND] = ACTIONS(2125),
    [anon_sym_COLON_COLON] = ACTIONS(2127),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2127),
    [anon_sym_else] = ACTIONS(2125),
    [anon_sym_new] = ACTIONS(2125),
    [anon_sym_LBRACE_LT] = ACTIONS(2127),
    [anon_sym_begin] = ACTIONS(2125),
    [sym_ocamlyacc_value] = ACTIONS(2127),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2125),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2127),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2127),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2127),
    [aux_sym_number_token1] = ACTIONS(2127),
    [aux_sym_signed_number_token1] = ACTIONS(2125),
    [anon_sym_SQUOTE] = ACTIONS(2127),
    [anon_sym_DQUOTE] = ACTIONS(2127),
    [sym_prefix_operator] = ACTIONS(2127),
    [aux_sym_sign_operator_token1] = ACTIONS(2125),
    [sym_hash_operator] = ACTIONS(2127),
    [anon_sym_lsl] = ACTIONS(2125),
    [anon_sym_lsr] = ACTIONS(2125),
    [anon_sym_asr] = ACTIONS(2125),
    [aux_sym_pow_operator_token1] = ACTIONS(2127),
    [anon_sym_mod] = ACTIONS(2125),
    [anon_sym_land] = ACTIONS(2125),
    [anon_sym_lor] = ACTIONS(2125),
    [anon_sym_lxor] = ACTIONS(2125),
    [aux_sym_mult_operator_token1] = ACTIONS(2125),
    [aux_sym_add_operator_token1] = ACTIONS(2125),
    [sym_concat_operator] = ACTIONS(2127),
    [sym_rel_operator] = ACTIONS(2125),
    [anon_sym_AMP_AMP] = ACTIONS(2125),
    [anon_sym_or] = ACTIONS(2125),
    [anon_sym_PIPE_PIPE] = ACTIONS(2125),
    [sym_assign_operator] = ACTIONS(2127),
    [sym_let_operator] = ACTIONS(2127),
    [sym_let_and_operator] = ACTIONS(2127),
    [sym__uppercase_identifier] = ACTIONS(2127),
    [aux_sym_directive_token1] = ACTIONS(2125),
    [aux_sym_tag_token1] = ACTIONS(2127),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1146)] = {
    [sym_attribute] = STATE(1146),
    [sym__lowercase_identifier] = ACTIONS(2199),
    [anon_sym_SEMI_SEMI] = ACTIONS(2201),
    [anon_sym_let] = ACTIONS(2199),
    [anon_sym_and] = ACTIONS(2199),
    [anon_sym_TILDE] = ACTIONS(2199),
    [anon_sym_QMARK] = ACTIONS(2199),
    [anon_sym_LPAREN] = ACTIONS(2201),
    [anon_sym_external] = ACTIONS(2199),
    [anon_sym_type] = ACTIONS(2199),
    [anon_sym_COMMA] = ACTIONS(2201),
    [anon_sym_PIPE] = ACTIONS(2199),
    [anon_sym_DOT] = ACTIONS(2201),
    [anon_sym_LBRACK] = ACTIONS(2199),
    [anon_sym_RBRACK] = ACTIONS(2201),
    [anon_sym_true] = ACTIONS(2199),
    [anon_sym_false] = ACTIONS(2199),
    [anon_sym_LBRACE] = ACTIONS(2199),
    [anon_sym_SEMI] = ACTIONS(2199),
    [anon_sym_exception] = ACTIONS(2199),
    [anon_sym_module] = ACTIONS(2199),
    [anon_sym_open] = ACTIONS(2199),
    [anon_sym_include] = ACTIONS(2199),
    [anon_sym_class] = ACTIONS(2199),
    [anon_sym_val] = ACTIONS(2199),
    [anon_sym_end] = ACTIONS(2199),
    [anon_sym_object] = ACTIONS(2199),
    [anon_sym_in] = ACTIONS(2199),
    [anon_sym_AMP] = ACTIONS(2199),
    [anon_sym_POUND] = ACTIONS(2199),
    [anon_sym_COLON_COLON] = ACTIONS(2201),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2201),
    [anon_sym_else] = ACTIONS(2199),
    [anon_sym_new] = ACTIONS(2199),
    [anon_sym_LBRACE_LT] = ACTIONS(2201),
    [anon_sym_begin] = ACTIONS(2199),
    [sym_ocamlyacc_value] = ACTIONS(2201),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2199),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2201),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2199),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2199),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2201),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2201),
    [aux_sym_number_token1] = ACTIONS(2201),
    [aux_sym_signed_number_token1] = ACTIONS(2199),
    [anon_sym_SQUOTE] = ACTIONS(2201),
    [anon_sym_DQUOTE] = ACTIONS(2201),
    [sym_prefix_operator] = ACTIONS(2201),
    [aux_sym_sign_operator_token1] = ACTIONS(2199),
    [sym_hash_operator] = ACTIONS(2201),
    [anon_sym_lsl] = ACTIONS(2199),
    [anon_sym_lsr] = ACTIONS(2199),
    [anon_sym_asr] = ACTIONS(2199),
    [aux_sym_pow_operator_token1] = ACTIONS(2201),
    [anon_sym_mod] = ACTIONS(2199),
    [anon_sym_land] = ACTIONS(2199),
    [anon_sym_lor] = ACTIONS(2199),
    [anon_sym_lxor] = ACTIONS(2199),
    [aux_sym_mult_operator_token1] = ACTIONS(2199),
    [aux_sym_add_operator_token1] = ACTIONS(2199),
    [sym_concat_operator] = ACTIONS(2201),
    [sym_rel_operator] = ACTIONS(2199),
    [anon_sym_AMP_AMP] = ACTIONS(2199),
    [anon_sym_or] = ACTIONS(2199),
    [anon_sym_PIPE_PIPE] = ACTIONS(2199),
    [sym_assign_operator] = ACTIONS(2201),
    [sym_let_operator] = ACTIONS(2201),
    [sym_let_and_operator] = ACTIONS(2201),
    [sym__uppercase_identifier] = ACTIONS(2201),
    [aux_sym_directive_token1] = ACTIONS(2199),
    [aux_sym_tag_token1] = ACTIONS(2201),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1147)] = {
    [sym_attribute] = STATE(1147),
    [sym__lowercase_identifier] = ACTIONS(2145),
    [anon_sym_SEMI_SEMI] = ACTIONS(2147),
    [anon_sym_let] = ACTIONS(2145),
    [anon_sym_and] = ACTIONS(2145),
    [anon_sym_TILDE] = ACTIONS(2145),
    [anon_sym_QMARK] = ACTIONS(2145),
    [anon_sym_LPAREN] = ACTIONS(2147),
    [anon_sym_external] = ACTIONS(2145),
    [anon_sym_type] = ACTIONS(2145),
    [anon_sym_COMMA] = ACTIONS(2147),
    [anon_sym_PIPE] = ACTIONS(2145),
    [anon_sym_DOT] = ACTIONS(2147),
    [anon_sym_LBRACK] = ACTIONS(2145),
    [anon_sym_RBRACK] = ACTIONS(2147),
    [anon_sym_true] = ACTIONS(2145),
    [anon_sym_false] = ACTIONS(2145),
    [anon_sym_LBRACE] = ACTIONS(2145),
    [anon_sym_SEMI] = ACTIONS(2145),
    [anon_sym_exception] = ACTIONS(2145),
    [anon_sym_module] = ACTIONS(2145),
    [anon_sym_open] = ACTIONS(2145),
    [anon_sym_include] = ACTIONS(2145),
    [anon_sym_class] = ACTIONS(2145),
    [anon_sym_val] = ACTIONS(2145),
    [anon_sym_end] = ACTIONS(2145),
    [anon_sym_object] = ACTIONS(2145),
    [anon_sym_in] = ACTIONS(2145),
    [anon_sym_AMP] = ACTIONS(2145),
    [anon_sym_POUND] = ACTIONS(2145),
    [anon_sym_COLON_COLON] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2147),
    [anon_sym_else] = ACTIONS(2145),
    [anon_sym_new] = ACTIONS(2145),
    [anon_sym_LBRACE_LT] = ACTIONS(2147),
    [anon_sym_begin] = ACTIONS(2145),
    [sym_ocamlyacc_value] = ACTIONS(2147),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2145),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2147),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2145),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2145),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2147),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2147),
    [aux_sym_number_token1] = ACTIONS(2147),
    [aux_sym_signed_number_token1] = ACTIONS(2145),
    [anon_sym_SQUOTE] = ACTIONS(2147),
    [anon_sym_DQUOTE] = ACTIONS(2147),
    [sym_prefix_operator] = ACTIONS(2147),
    [aux_sym_sign_operator_token1] = ACTIONS(2145),
    [sym_hash_operator] = ACTIONS(2147),
    [anon_sym_lsl] = ACTIONS(2145),
    [anon_sym_lsr] = ACTIONS(2145),
    [anon_sym_asr] = ACTIONS(2145),
    [aux_sym_pow_operator_token1] = ACTIONS(2147),
    [anon_sym_mod] = ACTIONS(2145),
    [anon_sym_land] = ACTIONS(2145),
    [anon_sym_lor] = ACTIONS(2145),
    [anon_sym_lxor] = ACTIONS(2145),
    [aux_sym_mult_operator_token1] = ACTIONS(2145),
    [aux_sym_add_operator_token1] = ACTIONS(2145),
    [sym_concat_operator] = ACTIONS(2147),
    [sym_rel_operator] = ACTIONS(2145),
    [anon_sym_AMP_AMP] = ACTIONS(2145),
    [anon_sym_or] = ACTIONS(2145),
    [anon_sym_PIPE_PIPE] = ACTIONS(2145),
    [sym_assign_operator] = ACTIONS(2147),
    [sym_let_operator] = ACTIONS(2147),
    [sym_let_and_operator] = ACTIONS(2147),
    [sym__uppercase_identifier] = ACTIONS(2147),
    [aux_sym_directive_token1] = ACTIONS(2145),
    [aux_sym_tag_token1] = ACTIONS(2147),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1148)] = {
    [sym_attribute] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(2111),
    [anon_sym_SEMI_SEMI] = ACTIONS(2113),
    [anon_sym_let] = ACTIONS(2111),
    [anon_sym_and] = ACTIONS(2111),
    [anon_sym_TILDE] = ACTIONS(2111),
    [anon_sym_QMARK] = ACTIONS(2111),
    [anon_sym_LPAREN] = ACTIONS(2113),
    [anon_sym_external] = ACTIONS(2111),
    [anon_sym_type] = ACTIONS(2111),
    [anon_sym_COMMA] = ACTIONS(2113),
    [anon_sym_PIPE] = ACTIONS(2111),
    [anon_sym_DOT] = ACTIONS(2115),
    [anon_sym_LBRACK] = ACTIONS(2111),
    [anon_sym_RBRACK] = ACTIONS(2113),
    [anon_sym_true] = ACTIONS(2111),
    [anon_sym_false] = ACTIONS(2111),
    [anon_sym_LBRACE] = ACTIONS(2111),
    [anon_sym_SEMI] = ACTIONS(2111),
    [anon_sym_exception] = ACTIONS(2111),
    [anon_sym_module] = ACTIONS(2111),
    [anon_sym_open] = ACTIONS(2111),
    [anon_sym_include] = ACTIONS(2111),
    [anon_sym_class] = ACTIONS(2111),
    [anon_sym_val] = ACTIONS(2111),
    [anon_sym_end] = ACTIONS(2111),
    [anon_sym_object] = ACTIONS(2111),
    [anon_sym_in] = ACTIONS(2111),
    [anon_sym_AMP] = ACTIONS(2111),
    [anon_sym_POUND] = ACTIONS(2111),
    [anon_sym_COLON_COLON] = ACTIONS(2113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2113),
    [anon_sym_else] = ACTIONS(2111),
    [anon_sym_new] = ACTIONS(2111),
    [anon_sym_LBRACE_LT] = ACTIONS(2113),
    [anon_sym_begin] = ACTIONS(2111),
    [sym_ocamlyacc_value] = ACTIONS(2113),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2113),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2113),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2113),
    [aux_sym_number_token1] = ACTIONS(2113),
    [aux_sym_signed_number_token1] = ACTIONS(2111),
    [anon_sym_SQUOTE] = ACTIONS(2113),
    [anon_sym_DQUOTE] = ACTIONS(2113),
    [sym_prefix_operator] = ACTIONS(2113),
    [aux_sym_sign_operator_token1] = ACTIONS(2111),
    [sym_hash_operator] = ACTIONS(2113),
    [anon_sym_lsl] = ACTIONS(2111),
    [anon_sym_lsr] = ACTIONS(2111),
    [anon_sym_asr] = ACTIONS(2111),
    [aux_sym_pow_operator_token1] = ACTIONS(2113),
    [anon_sym_mod] = ACTIONS(2111),
    [anon_sym_land] = ACTIONS(2111),
    [anon_sym_lor] = ACTIONS(2111),
    [anon_sym_lxor] = ACTIONS(2111),
    [aux_sym_mult_operator_token1] = ACTIONS(2111),
    [aux_sym_add_operator_token1] = ACTIONS(2111),
    [sym_concat_operator] = ACTIONS(2113),
    [sym_rel_operator] = ACTIONS(2111),
    [anon_sym_AMP_AMP] = ACTIONS(2111),
    [anon_sym_or] = ACTIONS(2111),
    [anon_sym_PIPE_PIPE] = ACTIONS(2111),
    [sym_assign_operator] = ACTIONS(2113),
    [sym_let_operator] = ACTIONS(2113),
    [sym_let_and_operator] = ACTIONS(2113),
    [sym__uppercase_identifier] = ACTIONS(2113),
    [aux_sym_directive_token1] = ACTIONS(2111),
    [aux_sym_tag_token1] = ACTIONS(2113),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1149)] = {
    [sym_attribute] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2083),
    [anon_sym_SEMI_SEMI] = ACTIONS(2085),
    [anon_sym_let] = ACTIONS(2083),
    [anon_sym_and] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(2083),
    [anon_sym_QMARK] = ACTIONS(2083),
    [anon_sym_LPAREN] = ACTIONS(2085),
    [anon_sym_external] = ACTIONS(2083),
    [anon_sym_type] = ACTIONS(2083),
    [anon_sym_COMMA] = ACTIONS(2085),
    [anon_sym_PIPE] = ACTIONS(2083),
    [anon_sym_DOT] = ACTIONS(2085),
    [anon_sym_LBRACK] = ACTIONS(2083),
    [anon_sym_RBRACK] = ACTIONS(2085),
    [anon_sym_true] = ACTIONS(2083),
    [anon_sym_false] = ACTIONS(2083),
    [anon_sym_LBRACE] = ACTIONS(2083),
    [anon_sym_SEMI] = ACTIONS(2083),
    [anon_sym_exception] = ACTIONS(2083),
    [anon_sym_module] = ACTIONS(2083),
    [anon_sym_open] = ACTIONS(2083),
    [anon_sym_include] = ACTIONS(2083),
    [anon_sym_class] = ACTIONS(2083),
    [anon_sym_val] = ACTIONS(2083),
    [anon_sym_end] = ACTIONS(2083),
    [anon_sym_object] = ACTIONS(2083),
    [anon_sym_in] = ACTIONS(2083),
    [anon_sym_AMP] = ACTIONS(2083),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(2085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2085),
    [anon_sym_else] = ACTIONS(2083),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_LBRACE_LT] = ACTIONS(2085),
    [anon_sym_begin] = ACTIONS(2083),
    [sym_ocamlyacc_value] = ACTIONS(2085),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2083),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2085),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2083),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2083),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2085),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2085),
    [aux_sym_number_token1] = ACTIONS(2085),
    [aux_sym_signed_number_token1] = ACTIONS(2083),
    [anon_sym_SQUOTE] = ACTIONS(2085),
    [anon_sym_DQUOTE] = ACTIONS(2085),
    [sym_prefix_operator] = ACTIONS(2085),
    [aux_sym_sign_operator_token1] = ACTIONS(2083),
    [sym_hash_operator] = ACTIONS(2085),
    [anon_sym_lsl] = ACTIONS(2083),
    [anon_sym_lsr] = ACTIONS(2083),
    [anon_sym_asr] = ACTIONS(2083),
    [aux_sym_pow_operator_token1] = ACTIONS(2085),
    [anon_sym_mod] = ACTIONS(2083),
    [anon_sym_land] = ACTIONS(2083),
    [anon_sym_lor] = ACTIONS(2083),
    [anon_sym_lxor] = ACTIONS(2083),
    [aux_sym_mult_operator_token1] = ACTIONS(2083),
    [aux_sym_add_operator_token1] = ACTIONS(2083),
    [sym_concat_operator] = ACTIONS(2085),
    [sym_rel_operator] = ACTIONS(2083),
    [anon_sym_AMP_AMP] = ACTIONS(2083),
    [anon_sym_or] = ACTIONS(2083),
    [anon_sym_PIPE_PIPE] = ACTIONS(2083),
    [sym_assign_operator] = ACTIONS(2085),
    [sym_let_operator] = ACTIONS(2085),
    [sym_let_and_operator] = ACTIONS(2085),
    [sym__uppercase_identifier] = ACTIONS(2085),
    [aux_sym_directive_token1] = ACTIONS(2083),
    [aux_sym_tag_token1] = ACTIONS(2085),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1150)] = {
    [sym_attribute] = STATE(1150),
    [sym__lowercase_identifier] = ACTIONS(1991),
    [anon_sym_SEMI_SEMI] = ACTIONS(1993),
    [anon_sym_let] = ACTIONS(1991),
    [anon_sym_and] = ACTIONS(1991),
    [anon_sym_TILDE] = ACTIONS(1991),
    [anon_sym_QMARK] = ACTIONS(1991),
    [anon_sym_LPAREN] = ACTIONS(1993),
    [anon_sym_external] = ACTIONS(1991),
    [anon_sym_type] = ACTIONS(1991),
    [anon_sym_COMMA] = ACTIONS(1993),
    [anon_sym_PIPE] = ACTIONS(1991),
    [anon_sym_DOT] = ACTIONS(1993),
    [anon_sym_LBRACK] = ACTIONS(1991),
    [anon_sym_RBRACK] = ACTIONS(1993),
    [anon_sym_true] = ACTIONS(1991),
    [anon_sym_false] = ACTIONS(1991),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_SEMI] = ACTIONS(1991),
    [anon_sym_exception] = ACTIONS(1991),
    [anon_sym_module] = ACTIONS(1991),
    [anon_sym_open] = ACTIONS(1991),
    [anon_sym_include] = ACTIONS(1991),
    [anon_sym_class] = ACTIONS(1991),
    [anon_sym_val] = ACTIONS(1991),
    [anon_sym_end] = ACTIONS(1991),
    [anon_sym_object] = ACTIONS(1991),
    [anon_sym_in] = ACTIONS(1991),
    [anon_sym_AMP] = ACTIONS(1991),
    [anon_sym_POUND] = ACTIONS(1991),
    [anon_sym_COLON_COLON] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1993),
    [anon_sym_else] = ACTIONS(1991),
    [anon_sym_new] = ACTIONS(1991),
    [anon_sym_LBRACE_LT] = ACTIONS(1993),
    [anon_sym_begin] = ACTIONS(1991),
    [sym_ocamlyacc_value] = ACTIONS(1993),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1991),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1993),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1991),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1991),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1993),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1993),
    [aux_sym_number_token1] = ACTIONS(1993),
    [aux_sym_signed_number_token1] = ACTIONS(1991),
    [anon_sym_SQUOTE] = ACTIONS(1993),
    [anon_sym_DQUOTE] = ACTIONS(1993),
    [sym_prefix_operator] = ACTIONS(1993),
    [aux_sym_sign_operator_token1] = ACTIONS(1991),
    [sym_hash_operator] = ACTIONS(1993),
    [anon_sym_lsl] = ACTIONS(1991),
    [anon_sym_lsr] = ACTIONS(1991),
    [anon_sym_asr] = ACTIONS(1991),
    [aux_sym_pow_operator_token1] = ACTIONS(1993),
    [anon_sym_mod] = ACTIONS(1991),
    [anon_sym_land] = ACTIONS(1991),
    [anon_sym_lor] = ACTIONS(1991),
    [anon_sym_lxor] = ACTIONS(1991),
    [aux_sym_mult_operator_token1] = ACTIONS(1991),
    [aux_sym_add_operator_token1] = ACTIONS(1991),
    [sym_concat_operator] = ACTIONS(1993),
    [sym_rel_operator] = ACTIONS(1991),
    [anon_sym_AMP_AMP] = ACTIONS(1991),
    [anon_sym_or] = ACTIONS(1991),
    [anon_sym_PIPE_PIPE] = ACTIONS(1991),
    [sym_assign_operator] = ACTIONS(1993),
    [sym_let_operator] = ACTIONS(1993),
    [sym_let_and_operator] = ACTIONS(1993),
    [sym__uppercase_identifier] = ACTIONS(1993),
    [aux_sym_directive_token1] = ACTIONS(1991),
    [aux_sym_tag_token1] = ACTIONS(1993),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1151)] = {
    [sym_attribute] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2173),
    [anon_sym_SEMI_SEMI] = ACTIONS(2175),
    [anon_sym_let] = ACTIONS(2173),
    [anon_sym_and] = ACTIONS(2173),
    [anon_sym_TILDE] = ACTIONS(2173),
    [anon_sym_QMARK] = ACTIONS(2173),
    [anon_sym_LPAREN] = ACTIONS(2175),
    [anon_sym_external] = ACTIONS(2173),
    [anon_sym_type] = ACTIONS(2173),
    [anon_sym_COMMA] = ACTIONS(2175),
    [anon_sym_PIPE] = ACTIONS(2173),
    [anon_sym_DOT] = ACTIONS(2175),
    [anon_sym_LBRACK] = ACTIONS(2173),
    [anon_sym_RBRACK] = ACTIONS(2175),
    [anon_sym_true] = ACTIONS(2173),
    [anon_sym_false] = ACTIONS(2173),
    [anon_sym_LBRACE] = ACTIONS(2173),
    [anon_sym_SEMI] = ACTIONS(2173),
    [anon_sym_exception] = ACTIONS(2173),
    [anon_sym_module] = ACTIONS(2173),
    [anon_sym_open] = ACTIONS(2173),
    [anon_sym_include] = ACTIONS(2173),
    [anon_sym_class] = ACTIONS(2173),
    [anon_sym_val] = ACTIONS(2173),
    [anon_sym_end] = ACTIONS(2173),
    [anon_sym_object] = ACTIONS(2173),
    [anon_sym_in] = ACTIONS(2173),
    [anon_sym_AMP] = ACTIONS(2173),
    [anon_sym_POUND] = ACTIONS(2173),
    [anon_sym_COLON_COLON] = ACTIONS(2175),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2175),
    [anon_sym_else] = ACTIONS(2173),
    [anon_sym_new] = ACTIONS(2173),
    [anon_sym_LBRACE_LT] = ACTIONS(2175),
    [anon_sym_begin] = ACTIONS(2173),
    [sym_ocamlyacc_value] = ACTIONS(2175),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2173),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2175),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2173),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2175),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2175),
    [aux_sym_number_token1] = ACTIONS(2175),
    [aux_sym_signed_number_token1] = ACTIONS(2173),
    [anon_sym_SQUOTE] = ACTIONS(2175),
    [anon_sym_DQUOTE] = ACTIONS(2175),
    [sym_prefix_operator] = ACTIONS(2175),
    [aux_sym_sign_operator_token1] = ACTIONS(2173),
    [sym_hash_operator] = ACTIONS(2175),
    [anon_sym_lsl] = ACTIONS(2173),
    [anon_sym_lsr] = ACTIONS(2173),
    [anon_sym_asr] = ACTIONS(2173),
    [aux_sym_pow_operator_token1] = ACTIONS(2175),
    [anon_sym_mod] = ACTIONS(2173),
    [anon_sym_land] = ACTIONS(2173),
    [anon_sym_lor] = ACTIONS(2173),
    [anon_sym_lxor] = ACTIONS(2173),
    [aux_sym_mult_operator_token1] = ACTIONS(2173),
    [aux_sym_add_operator_token1] = ACTIONS(2173),
    [sym_concat_operator] = ACTIONS(2175),
    [sym_rel_operator] = ACTIONS(2173),
    [anon_sym_AMP_AMP] = ACTIONS(2173),
    [anon_sym_or] = ACTIONS(2173),
    [anon_sym_PIPE_PIPE] = ACTIONS(2173),
    [sym_assign_operator] = ACTIONS(2175),
    [sym_let_operator] = ACTIONS(2175),
    [sym_let_and_operator] = ACTIONS(2175),
    [sym__uppercase_identifier] = ACTIONS(2175),
    [aux_sym_directive_token1] = ACTIONS(2173),
    [aux_sym_tag_token1] = ACTIONS(2175),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1152)] = {
    [sym_attribute] = STATE(1152),
    [sym__lowercase_identifier] = ACTIONS(2007),
    [anon_sym_SEMI_SEMI] = ACTIONS(2009),
    [anon_sym_let] = ACTIONS(2007),
    [anon_sym_and] = ACTIONS(2007),
    [anon_sym_TILDE] = ACTIONS(2007),
    [anon_sym_QMARK] = ACTIONS(2007),
    [anon_sym_LPAREN] = ACTIONS(2009),
    [anon_sym_external] = ACTIONS(2007),
    [anon_sym_type] = ACTIONS(2007),
    [anon_sym_COMMA] = ACTIONS(2009),
    [anon_sym_PIPE] = ACTIONS(2007),
    [anon_sym_DOT] = ACTIONS(2009),
    [anon_sym_LBRACK] = ACTIONS(2007),
    [anon_sym_RBRACK] = ACTIONS(2009),
    [anon_sym_true] = ACTIONS(2007),
    [anon_sym_false] = ACTIONS(2007),
    [anon_sym_LBRACE] = ACTIONS(2007),
    [anon_sym_SEMI] = ACTIONS(2007),
    [anon_sym_exception] = ACTIONS(2007),
    [anon_sym_module] = ACTIONS(2007),
    [anon_sym_open] = ACTIONS(2007),
    [anon_sym_include] = ACTIONS(2007),
    [anon_sym_class] = ACTIONS(2007),
    [anon_sym_val] = ACTIONS(2007),
    [anon_sym_end] = ACTIONS(2007),
    [anon_sym_object] = ACTIONS(2007),
    [anon_sym_in] = ACTIONS(2007),
    [anon_sym_AMP] = ACTIONS(2007),
    [anon_sym_POUND] = ACTIONS(2007),
    [anon_sym_COLON_COLON] = ACTIONS(2009),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2009),
    [anon_sym_else] = ACTIONS(2007),
    [anon_sym_new] = ACTIONS(2007),
    [anon_sym_LBRACE_LT] = ACTIONS(2009),
    [anon_sym_begin] = ACTIONS(2007),
    [sym_ocamlyacc_value] = ACTIONS(2009),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2007),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2009),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2007),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2007),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2009),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2009),
    [aux_sym_number_token1] = ACTIONS(2009),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2009),
    [sym_prefix_operator] = ACTIONS(2009),
    [aux_sym_sign_operator_token1] = ACTIONS(2007),
    [sym_hash_operator] = ACTIONS(2009),
    [anon_sym_lsl] = ACTIONS(2007),
    [anon_sym_lsr] = ACTIONS(2007),
    [anon_sym_asr] = ACTIONS(2007),
    [aux_sym_pow_operator_token1] = ACTIONS(2009),
    [anon_sym_mod] = ACTIONS(2007),
    [anon_sym_land] = ACTIONS(2007),
    [anon_sym_lor] = ACTIONS(2007),
    [anon_sym_lxor] = ACTIONS(2007),
    [aux_sym_mult_operator_token1] = ACTIONS(2007),
    [aux_sym_add_operator_token1] = ACTIONS(2007),
    [sym_concat_operator] = ACTIONS(2009),
    [sym_rel_operator] = ACTIONS(2007),
    [anon_sym_AMP_AMP] = ACTIONS(2007),
    [anon_sym_or] = ACTIONS(2007),
    [anon_sym_PIPE_PIPE] = ACTIONS(2007),
    [sym_assign_operator] = ACTIONS(2009),
    [sym_let_operator] = ACTIONS(2009),
    [sym_let_and_operator] = ACTIONS(2009),
    [sym__uppercase_identifier] = ACTIONS(2009),
    [aux_sym_directive_token1] = ACTIONS(2007),
    [aux_sym_tag_token1] = ACTIONS(2009),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1153)] = {
    [sym_attribute] = STATE(1153),
    [sym__lowercase_identifier] = ACTIONS(2207),
    [anon_sym_SEMI_SEMI] = ACTIONS(2209),
    [anon_sym_let] = ACTIONS(2207),
    [anon_sym_and] = ACTIONS(2207),
    [anon_sym_TILDE] = ACTIONS(2207),
    [anon_sym_QMARK] = ACTIONS(2207),
    [anon_sym_LPAREN] = ACTIONS(2209),
    [anon_sym_external] = ACTIONS(2207),
    [anon_sym_type] = ACTIONS(2207),
    [anon_sym_COMMA] = ACTIONS(2209),
    [anon_sym_PIPE] = ACTIONS(2207),
    [anon_sym_DOT] = ACTIONS(2209),
    [anon_sym_LBRACK] = ACTIONS(2207),
    [anon_sym_RBRACK] = ACTIONS(2209),
    [anon_sym_true] = ACTIONS(2207),
    [anon_sym_false] = ACTIONS(2207),
    [anon_sym_LBRACE] = ACTIONS(2207),
    [anon_sym_SEMI] = ACTIONS(2207),
    [anon_sym_exception] = ACTIONS(2207),
    [anon_sym_module] = ACTIONS(2207),
    [anon_sym_open] = ACTIONS(2207),
    [anon_sym_include] = ACTIONS(2207),
    [anon_sym_class] = ACTIONS(2207),
    [anon_sym_val] = ACTIONS(2207),
    [anon_sym_end] = ACTIONS(2207),
    [anon_sym_object] = ACTIONS(2207),
    [anon_sym_in] = ACTIONS(2207),
    [anon_sym_AMP] = ACTIONS(2207),
    [anon_sym_POUND] = ACTIONS(2207),
    [anon_sym_COLON_COLON] = ACTIONS(2209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2209),
    [anon_sym_else] = ACTIONS(2207),
    [anon_sym_new] = ACTIONS(2207),
    [anon_sym_LBRACE_LT] = ACTIONS(2209),
    [anon_sym_begin] = ACTIONS(2207),
    [sym_ocamlyacc_value] = ACTIONS(2209),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2207),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2209),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2207),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2207),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2209),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2209),
    [aux_sym_number_token1] = ACTIONS(2209),
    [aux_sym_signed_number_token1] = ACTIONS(2207),
    [anon_sym_SQUOTE] = ACTIONS(2209),
    [anon_sym_DQUOTE] = ACTIONS(2209),
    [sym_prefix_operator] = ACTIONS(2209),
    [aux_sym_sign_operator_token1] = ACTIONS(2207),
    [sym_hash_operator] = ACTIONS(2209),
    [anon_sym_lsl] = ACTIONS(2207),
    [anon_sym_lsr] = ACTIONS(2207),
    [anon_sym_asr] = ACTIONS(2207),
    [aux_sym_pow_operator_token1] = ACTIONS(2209),
    [anon_sym_mod] = ACTIONS(2207),
    [anon_sym_land] = ACTIONS(2207),
    [anon_sym_lor] = ACTIONS(2207),
    [anon_sym_lxor] = ACTIONS(2207),
    [aux_sym_mult_operator_token1] = ACTIONS(2207),
    [aux_sym_add_operator_token1] = ACTIONS(2207),
    [sym_concat_operator] = ACTIONS(2209),
    [sym_rel_operator] = ACTIONS(2207),
    [anon_sym_AMP_AMP] = ACTIONS(2207),
    [anon_sym_or] = ACTIONS(2207),
    [anon_sym_PIPE_PIPE] = ACTIONS(2207),
    [sym_assign_operator] = ACTIONS(2209),
    [sym_let_operator] = ACTIONS(2209),
    [sym_let_and_operator] = ACTIONS(2209),
    [sym__uppercase_identifier] = ACTIONS(2209),
    [aux_sym_directive_token1] = ACTIONS(2207),
    [aux_sym_tag_token1] = ACTIONS(2209),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1154)] = {
    [sym_attribute] = STATE(1154),
    [sym__lowercase_identifier] = ACTIONS(2121),
    [anon_sym_SEMI_SEMI] = ACTIONS(2123),
    [anon_sym_let] = ACTIONS(2121),
    [anon_sym_and] = ACTIONS(2121),
    [anon_sym_TILDE] = ACTIONS(2121),
    [anon_sym_QMARK] = ACTIONS(2121),
    [anon_sym_LPAREN] = ACTIONS(2123),
    [anon_sym_external] = ACTIONS(2121),
    [anon_sym_type] = ACTIONS(2121),
    [anon_sym_COMMA] = ACTIONS(2123),
    [anon_sym_PIPE] = ACTIONS(2121),
    [anon_sym_DOT] = ACTIONS(2123),
    [anon_sym_LBRACK] = ACTIONS(2121),
    [anon_sym_RBRACK] = ACTIONS(2123),
    [anon_sym_true] = ACTIONS(2121),
    [anon_sym_false] = ACTIONS(2121),
    [anon_sym_LBRACE] = ACTIONS(2121),
    [anon_sym_SEMI] = ACTIONS(2121),
    [anon_sym_exception] = ACTIONS(2121),
    [anon_sym_module] = ACTIONS(2121),
    [anon_sym_open] = ACTIONS(2121),
    [anon_sym_include] = ACTIONS(2121),
    [anon_sym_class] = ACTIONS(2121),
    [anon_sym_val] = ACTIONS(2121),
    [anon_sym_end] = ACTIONS(2121),
    [anon_sym_object] = ACTIONS(2121),
    [anon_sym_in] = ACTIONS(2121),
    [anon_sym_AMP] = ACTIONS(2121),
    [anon_sym_POUND] = ACTIONS(2121),
    [anon_sym_COLON_COLON] = ACTIONS(2123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2123),
    [anon_sym_else] = ACTIONS(2121),
    [anon_sym_new] = ACTIONS(2121),
    [anon_sym_LBRACE_LT] = ACTIONS(2123),
    [anon_sym_begin] = ACTIONS(2121),
    [sym_ocamlyacc_value] = ACTIONS(2123),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2121),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2123),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2121),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2121),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2123),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2123),
    [aux_sym_number_token1] = ACTIONS(2123),
    [aux_sym_signed_number_token1] = ACTIONS(2121),
    [anon_sym_SQUOTE] = ACTIONS(2123),
    [anon_sym_DQUOTE] = ACTIONS(2123),
    [sym_prefix_operator] = ACTIONS(2123),
    [aux_sym_sign_operator_token1] = ACTIONS(2121),
    [sym_hash_operator] = ACTIONS(2123),
    [anon_sym_lsl] = ACTIONS(2121),
    [anon_sym_lsr] = ACTIONS(2121),
    [anon_sym_asr] = ACTIONS(2121),
    [aux_sym_pow_operator_token1] = ACTIONS(2123),
    [anon_sym_mod] = ACTIONS(2121),
    [anon_sym_land] = ACTIONS(2121),
    [anon_sym_lor] = ACTIONS(2121),
    [anon_sym_lxor] = ACTIONS(2121),
    [aux_sym_mult_operator_token1] = ACTIONS(2121),
    [aux_sym_add_operator_token1] = ACTIONS(2121),
    [sym_concat_operator] = ACTIONS(2123),
    [sym_rel_operator] = ACTIONS(2121),
    [anon_sym_AMP_AMP] = ACTIONS(2121),
    [anon_sym_or] = ACTIONS(2121),
    [anon_sym_PIPE_PIPE] = ACTIONS(2121),
    [sym_assign_operator] = ACTIONS(2123),
    [sym_let_operator] = ACTIONS(2123),
    [sym_let_and_operator] = ACTIONS(2123),
    [sym__uppercase_identifier] = ACTIONS(2123),
    [aux_sym_directive_token1] = ACTIONS(2121),
    [aux_sym_tag_token1] = ACTIONS(2123),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1155)] = {
    [sym_attribute] = STATE(1155),
    [sym__lowercase_identifier] = ACTIONS(2011),
    [anon_sym_SEMI_SEMI] = ACTIONS(2013),
    [anon_sym_let] = ACTIONS(2011),
    [anon_sym_and] = ACTIONS(2011),
    [anon_sym_TILDE] = ACTIONS(2011),
    [anon_sym_QMARK] = ACTIONS(2011),
    [anon_sym_LPAREN] = ACTIONS(2013),
    [anon_sym_external] = ACTIONS(2011),
    [anon_sym_type] = ACTIONS(2011),
    [anon_sym_COMMA] = ACTIONS(2013),
    [anon_sym_PIPE] = ACTIONS(2011),
    [anon_sym_DOT] = ACTIONS(2013),
    [anon_sym_LBRACK] = ACTIONS(2011),
    [anon_sym_RBRACK] = ACTIONS(2013),
    [anon_sym_true] = ACTIONS(2011),
    [anon_sym_false] = ACTIONS(2011),
    [anon_sym_LBRACE] = ACTIONS(2011),
    [anon_sym_SEMI] = ACTIONS(2011),
    [anon_sym_exception] = ACTIONS(2011),
    [anon_sym_module] = ACTIONS(2011),
    [anon_sym_open] = ACTIONS(2011),
    [anon_sym_include] = ACTIONS(2011),
    [anon_sym_class] = ACTIONS(2011),
    [anon_sym_val] = ACTIONS(2011),
    [anon_sym_end] = ACTIONS(2011),
    [anon_sym_object] = ACTIONS(2011),
    [anon_sym_in] = ACTIONS(2011),
    [anon_sym_AMP] = ACTIONS(2011),
    [anon_sym_POUND] = ACTIONS(2011),
    [anon_sym_COLON_COLON] = ACTIONS(2013),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2013),
    [anon_sym_else] = ACTIONS(2011),
    [anon_sym_new] = ACTIONS(2011),
    [anon_sym_LBRACE_LT] = ACTIONS(2013),
    [anon_sym_begin] = ACTIONS(2011),
    [sym_ocamlyacc_value] = ACTIONS(2013),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2011),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2013),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2011),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2011),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2013),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2013),
    [aux_sym_number_token1] = ACTIONS(2013),
    [aux_sym_signed_number_token1] = ACTIONS(2011),
    [anon_sym_SQUOTE] = ACTIONS(2013),
    [anon_sym_DQUOTE] = ACTIONS(2013),
    [sym_prefix_operator] = ACTIONS(2013),
    [aux_sym_sign_operator_token1] = ACTIONS(2011),
    [sym_hash_operator] = ACTIONS(2013),
    [anon_sym_lsl] = ACTIONS(2011),
    [anon_sym_lsr] = ACTIONS(2011),
    [anon_sym_asr] = ACTIONS(2011),
    [aux_sym_pow_operator_token1] = ACTIONS(2013),
    [anon_sym_mod] = ACTIONS(2011),
    [anon_sym_land] = ACTIONS(2011),
    [anon_sym_lor] = ACTIONS(2011),
    [anon_sym_lxor] = ACTIONS(2011),
    [aux_sym_mult_operator_token1] = ACTIONS(2011),
    [aux_sym_add_operator_token1] = ACTIONS(2011),
    [sym_concat_operator] = ACTIONS(2013),
    [sym_rel_operator] = ACTIONS(2011),
    [anon_sym_AMP_AMP] = ACTIONS(2011),
    [anon_sym_or] = ACTIONS(2011),
    [anon_sym_PIPE_PIPE] = ACTIONS(2011),
    [sym_assign_operator] = ACTIONS(2013),
    [sym_let_operator] = ACTIONS(2013),
    [sym_let_and_operator] = ACTIONS(2013),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_directive_token1] = ACTIONS(2011),
    [aux_sym_tag_token1] = ACTIONS(2013),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1156)] = {
    [sym_attribute] = STATE(1156),
    [sym__lowercase_identifier] = ACTIONS(2203),
    [anon_sym_SEMI_SEMI] = ACTIONS(2205),
    [anon_sym_let] = ACTIONS(2203),
    [anon_sym_and] = ACTIONS(2203),
    [anon_sym_TILDE] = ACTIONS(2203),
    [anon_sym_QMARK] = ACTIONS(2203),
    [anon_sym_LPAREN] = ACTIONS(2205),
    [anon_sym_external] = ACTIONS(2203),
    [anon_sym_type] = ACTIONS(2203),
    [anon_sym_COMMA] = ACTIONS(2205),
    [anon_sym_PIPE] = ACTIONS(2203),
    [anon_sym_DOT] = ACTIONS(2205),
    [anon_sym_LBRACK] = ACTIONS(2203),
    [anon_sym_RBRACK] = ACTIONS(2205),
    [anon_sym_true] = ACTIONS(2203),
    [anon_sym_false] = ACTIONS(2203),
    [anon_sym_LBRACE] = ACTIONS(2203),
    [anon_sym_SEMI] = ACTIONS(2203),
    [anon_sym_exception] = ACTIONS(2203),
    [anon_sym_module] = ACTIONS(2203),
    [anon_sym_open] = ACTIONS(2203),
    [anon_sym_include] = ACTIONS(2203),
    [anon_sym_class] = ACTIONS(2203),
    [anon_sym_val] = ACTIONS(2203),
    [anon_sym_end] = ACTIONS(2203),
    [anon_sym_object] = ACTIONS(2203),
    [anon_sym_in] = ACTIONS(2203),
    [anon_sym_AMP] = ACTIONS(2203),
    [anon_sym_POUND] = ACTIONS(2203),
    [anon_sym_COLON_COLON] = ACTIONS(2205),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2205),
    [anon_sym_else] = ACTIONS(2203),
    [anon_sym_new] = ACTIONS(2203),
    [anon_sym_LBRACE_LT] = ACTIONS(2205),
    [anon_sym_begin] = ACTIONS(2203),
    [sym_ocamlyacc_value] = ACTIONS(2205),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2203),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2205),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2203),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2203),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2205),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2205),
    [aux_sym_number_token1] = ACTIONS(2205),
    [aux_sym_signed_number_token1] = ACTIONS(2203),
    [anon_sym_SQUOTE] = ACTIONS(2205),
    [anon_sym_DQUOTE] = ACTIONS(2205),
    [sym_prefix_operator] = ACTIONS(2205),
    [aux_sym_sign_operator_token1] = ACTIONS(2203),
    [sym_hash_operator] = ACTIONS(2205),
    [anon_sym_lsl] = ACTIONS(2203),
    [anon_sym_lsr] = ACTIONS(2203),
    [anon_sym_asr] = ACTIONS(2203),
    [aux_sym_pow_operator_token1] = ACTIONS(2205),
    [anon_sym_mod] = ACTIONS(2203),
    [anon_sym_land] = ACTIONS(2203),
    [anon_sym_lor] = ACTIONS(2203),
    [anon_sym_lxor] = ACTIONS(2203),
    [aux_sym_mult_operator_token1] = ACTIONS(2203),
    [aux_sym_add_operator_token1] = ACTIONS(2203),
    [sym_concat_operator] = ACTIONS(2205),
    [sym_rel_operator] = ACTIONS(2203),
    [anon_sym_AMP_AMP] = ACTIONS(2203),
    [anon_sym_or] = ACTIONS(2203),
    [anon_sym_PIPE_PIPE] = ACTIONS(2203),
    [sym_assign_operator] = ACTIONS(2205),
    [sym_let_operator] = ACTIONS(2205),
    [sym_let_and_operator] = ACTIONS(2205),
    [sym__uppercase_identifier] = ACTIONS(2205),
    [aux_sym_directive_token1] = ACTIONS(2203),
    [aux_sym_tag_token1] = ACTIONS(2205),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1157)] = {
    [sym_attribute] = STATE(1157),
    [ts_builtin_sym_end] = ACTIONS(1889),
    [sym__lowercase_identifier] = ACTIONS(1887),
    [anon_sym_SEMI_SEMI] = ACTIONS(1889),
    [anon_sym_let] = ACTIONS(1887),
    [anon_sym_and] = ACTIONS(1887),
    [anon_sym_TILDE] = ACTIONS(1887),
    [anon_sym_QMARK] = ACTIONS(1887),
    [anon_sym_LPAREN] = ACTIONS(1889),
    [anon_sym_external] = ACTIONS(1887),
    [anon_sym_type] = ACTIONS(1887),
    [anon_sym_COMMA] = ACTIONS(1889),
    [anon_sym_PIPE] = ACTIONS(1887),
    [anon_sym_DOT] = ACTIONS(1889),
    [anon_sym_LBRACK] = ACTIONS(1887),
    [anon_sym_true] = ACTIONS(1887),
    [anon_sym_false] = ACTIONS(1887),
    [anon_sym_LBRACE] = ACTIONS(1887),
    [anon_sym_SEMI] = ACTIONS(1887),
    [anon_sym_exception] = ACTIONS(1887),
    [anon_sym_module] = ACTIONS(1887),
    [anon_sym_open] = ACTIONS(1887),
    [anon_sym_include] = ACTIONS(1887),
    [anon_sym_class] = ACTIONS(1887),
    [anon_sym_val] = ACTIONS(1887),
    [anon_sym_object] = ACTIONS(1887),
    [anon_sym_in] = ACTIONS(1887),
    [anon_sym_AMP] = ACTIONS(1887),
    [anon_sym_POUND] = ACTIONS(1887),
    [anon_sym_COLON_COLON] = ACTIONS(1889),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1889),
    [anon_sym_LT_DASH] = ACTIONS(1887),
    [anon_sym_else] = ACTIONS(1887),
    [anon_sym_new] = ACTIONS(1887),
    [anon_sym_LBRACE_LT] = ACTIONS(1889),
    [anon_sym_begin] = ACTIONS(1887),
    [sym_ocamlyacc_value] = ACTIONS(1889),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1887),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1889),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1887),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1887),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1889),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1889),
    [aux_sym_number_token1] = ACTIONS(1889),
    [aux_sym_signed_number_token1] = ACTIONS(1887),
    [anon_sym_SQUOTE] = ACTIONS(1889),
    [anon_sym_DQUOTE] = ACTIONS(1889),
    [sym_prefix_operator] = ACTIONS(1889),
    [aux_sym_sign_operator_token1] = ACTIONS(1887),
    [sym_hash_operator] = ACTIONS(1889),
    [anon_sym_lsl] = ACTIONS(1887),
    [anon_sym_lsr] = ACTIONS(1887),
    [anon_sym_asr] = ACTIONS(1887),
    [aux_sym_pow_operator_token1] = ACTIONS(1889),
    [anon_sym_mod] = ACTIONS(1887),
    [anon_sym_land] = ACTIONS(1887),
    [anon_sym_lor] = ACTIONS(1887),
    [anon_sym_lxor] = ACTIONS(1887),
    [aux_sym_mult_operator_token1] = ACTIONS(1887),
    [aux_sym_add_operator_token1] = ACTIONS(1887),
    [sym_concat_operator] = ACTIONS(1889),
    [sym_rel_operator] = ACTIONS(1887),
    [anon_sym_AMP_AMP] = ACTIONS(1887),
    [anon_sym_or] = ACTIONS(1887),
    [anon_sym_PIPE_PIPE] = ACTIONS(1887),
    [sym_assign_operator] = ACTIONS(1889),
    [sym_let_operator] = ACTIONS(1889),
    [sym_let_and_operator] = ACTIONS(1889),
    [sym__uppercase_identifier] = ACTIONS(1889),
    [aux_sym_directive_token1] = ACTIONS(1887),
    [aux_sym_tag_token1] = ACTIONS(1889),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1158)] = {
    [sym_attribute] = STATE(1158),
    [sym__lowercase_identifier] = ACTIONS(2031),
    [anon_sym_SEMI_SEMI] = ACTIONS(2033),
    [anon_sym_let] = ACTIONS(2031),
    [anon_sym_and] = ACTIONS(2031),
    [anon_sym_TILDE] = ACTIONS(2031),
    [anon_sym_QMARK] = ACTIONS(2031),
    [anon_sym_LPAREN] = ACTIONS(2033),
    [anon_sym_external] = ACTIONS(2031),
    [anon_sym_type] = ACTIONS(2031),
    [anon_sym_COMMA] = ACTIONS(2033),
    [anon_sym_PIPE] = ACTIONS(2031),
    [anon_sym_DOT] = ACTIONS(2033),
    [anon_sym_LBRACK] = ACTIONS(2031),
    [anon_sym_RBRACK] = ACTIONS(2033),
    [anon_sym_true] = ACTIONS(2031),
    [anon_sym_false] = ACTIONS(2031),
    [anon_sym_LBRACE] = ACTIONS(2031),
    [anon_sym_SEMI] = ACTIONS(2031),
    [anon_sym_exception] = ACTIONS(2031),
    [anon_sym_module] = ACTIONS(2031),
    [anon_sym_open] = ACTIONS(2031),
    [anon_sym_include] = ACTIONS(2031),
    [anon_sym_class] = ACTIONS(2031),
    [anon_sym_val] = ACTIONS(2031),
    [anon_sym_end] = ACTIONS(2031),
    [anon_sym_object] = ACTIONS(2031),
    [anon_sym_in] = ACTIONS(2031),
    [anon_sym_AMP] = ACTIONS(2031),
    [anon_sym_POUND] = ACTIONS(2031),
    [anon_sym_COLON_COLON] = ACTIONS(2033),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2033),
    [anon_sym_else] = ACTIONS(2031),
    [anon_sym_new] = ACTIONS(2031),
    [anon_sym_LBRACE_LT] = ACTIONS(2033),
    [anon_sym_begin] = ACTIONS(2031),
    [sym_ocamlyacc_value] = ACTIONS(2033),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2031),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2033),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2031),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2031),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2033),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2033),
    [aux_sym_number_token1] = ACTIONS(2033),
    [aux_sym_signed_number_token1] = ACTIONS(2031),
    [anon_sym_SQUOTE] = ACTIONS(2033),
    [anon_sym_DQUOTE] = ACTIONS(2033),
    [sym_prefix_operator] = ACTIONS(2033),
    [aux_sym_sign_operator_token1] = ACTIONS(2031),
    [sym_hash_operator] = ACTIONS(2033),
    [anon_sym_lsl] = ACTIONS(2031),
    [anon_sym_lsr] = ACTIONS(2031),
    [anon_sym_asr] = ACTIONS(2031),
    [aux_sym_pow_operator_token1] = ACTIONS(2033),
    [anon_sym_mod] = ACTIONS(2031),
    [anon_sym_land] = ACTIONS(2031),
    [anon_sym_lor] = ACTIONS(2031),
    [anon_sym_lxor] = ACTIONS(2031),
    [aux_sym_mult_operator_token1] = ACTIONS(2031),
    [aux_sym_add_operator_token1] = ACTIONS(2031),
    [sym_concat_operator] = ACTIONS(2033),
    [sym_rel_operator] = ACTIONS(2031),
    [anon_sym_AMP_AMP] = ACTIONS(2031),
    [anon_sym_or] = ACTIONS(2031),
    [anon_sym_PIPE_PIPE] = ACTIONS(2031),
    [sym_assign_operator] = ACTIONS(2033),
    [sym_let_operator] = ACTIONS(2033),
    [sym_let_and_operator] = ACTIONS(2033),
    [sym__uppercase_identifier] = ACTIONS(2033),
    [aux_sym_directive_token1] = ACTIONS(2031),
    [aux_sym_tag_token1] = ACTIONS(2033),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1159)] = {
    [sym_attribute] = STATE(1159),
    [sym__lowercase_identifier] = ACTIONS(2157),
    [anon_sym_SEMI_SEMI] = ACTIONS(2159),
    [anon_sym_let] = ACTIONS(2157),
    [anon_sym_and] = ACTIONS(2157),
    [anon_sym_TILDE] = ACTIONS(2157),
    [anon_sym_QMARK] = ACTIONS(2157),
    [anon_sym_LPAREN] = ACTIONS(2159),
    [anon_sym_external] = ACTIONS(2157),
    [anon_sym_type] = ACTIONS(2157),
    [anon_sym_COMMA] = ACTIONS(2159),
    [anon_sym_PIPE] = ACTIONS(2157),
    [anon_sym_DOT] = ACTIONS(2159),
    [anon_sym_LBRACK] = ACTIONS(2157),
    [anon_sym_RBRACK] = ACTIONS(2159),
    [anon_sym_true] = ACTIONS(2157),
    [anon_sym_false] = ACTIONS(2157),
    [anon_sym_LBRACE] = ACTIONS(2157),
    [anon_sym_SEMI] = ACTIONS(2157),
    [anon_sym_exception] = ACTIONS(2157),
    [anon_sym_module] = ACTIONS(2157),
    [anon_sym_open] = ACTIONS(2157),
    [anon_sym_include] = ACTIONS(2157),
    [anon_sym_class] = ACTIONS(2157),
    [anon_sym_val] = ACTIONS(2157),
    [anon_sym_end] = ACTIONS(2157),
    [anon_sym_object] = ACTIONS(2157),
    [anon_sym_in] = ACTIONS(2157),
    [anon_sym_AMP] = ACTIONS(2157),
    [anon_sym_POUND] = ACTIONS(2157),
    [anon_sym_COLON_COLON] = ACTIONS(2159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2159),
    [anon_sym_else] = ACTIONS(2157),
    [anon_sym_new] = ACTIONS(2157),
    [anon_sym_LBRACE_LT] = ACTIONS(2159),
    [anon_sym_begin] = ACTIONS(2157),
    [sym_ocamlyacc_value] = ACTIONS(2159),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2157),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2159),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2159),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2159),
    [aux_sym_number_token1] = ACTIONS(2159),
    [aux_sym_signed_number_token1] = ACTIONS(2157),
    [anon_sym_SQUOTE] = ACTIONS(2159),
    [anon_sym_DQUOTE] = ACTIONS(2159),
    [sym_prefix_operator] = ACTIONS(2159),
    [aux_sym_sign_operator_token1] = ACTIONS(2157),
    [sym_hash_operator] = ACTIONS(2159),
    [anon_sym_lsl] = ACTIONS(2157),
    [anon_sym_lsr] = ACTIONS(2157),
    [anon_sym_asr] = ACTIONS(2157),
    [aux_sym_pow_operator_token1] = ACTIONS(2159),
    [anon_sym_mod] = ACTIONS(2157),
    [anon_sym_land] = ACTIONS(2157),
    [anon_sym_lor] = ACTIONS(2157),
    [anon_sym_lxor] = ACTIONS(2157),
    [aux_sym_mult_operator_token1] = ACTIONS(2157),
    [aux_sym_add_operator_token1] = ACTIONS(2157),
    [sym_concat_operator] = ACTIONS(2159),
    [sym_rel_operator] = ACTIONS(2157),
    [anon_sym_AMP_AMP] = ACTIONS(2157),
    [anon_sym_or] = ACTIONS(2157),
    [anon_sym_PIPE_PIPE] = ACTIONS(2157),
    [sym_assign_operator] = ACTIONS(2159),
    [sym_let_operator] = ACTIONS(2159),
    [sym_let_and_operator] = ACTIONS(2159),
    [sym__uppercase_identifier] = ACTIONS(2159),
    [aux_sym_directive_token1] = ACTIONS(2157),
    [aux_sym_tag_token1] = ACTIONS(2159),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1160)] = {
    [sym_attribute] = STATE(1160),
    [sym__lowercase_identifier] = ACTIONS(2129),
    [anon_sym_SEMI_SEMI] = ACTIONS(2131),
    [anon_sym_let] = ACTIONS(2129),
    [anon_sym_and] = ACTIONS(2129),
    [anon_sym_TILDE] = ACTIONS(2129),
    [anon_sym_QMARK] = ACTIONS(2129),
    [anon_sym_LPAREN] = ACTIONS(2131),
    [anon_sym_external] = ACTIONS(2129),
    [anon_sym_type] = ACTIONS(2129),
    [anon_sym_COMMA] = ACTIONS(2131),
    [anon_sym_PIPE] = ACTIONS(2129),
    [anon_sym_DOT] = ACTIONS(2131),
    [anon_sym_LBRACK] = ACTIONS(2129),
    [anon_sym_RBRACK] = ACTIONS(2131),
    [anon_sym_true] = ACTIONS(2129),
    [anon_sym_false] = ACTIONS(2129),
    [anon_sym_LBRACE] = ACTIONS(2129),
    [anon_sym_SEMI] = ACTIONS(2129),
    [anon_sym_exception] = ACTIONS(2129),
    [anon_sym_module] = ACTIONS(2129),
    [anon_sym_open] = ACTIONS(2129),
    [anon_sym_include] = ACTIONS(2129),
    [anon_sym_class] = ACTIONS(2129),
    [anon_sym_val] = ACTIONS(2129),
    [anon_sym_end] = ACTIONS(2129),
    [anon_sym_object] = ACTIONS(2129),
    [anon_sym_in] = ACTIONS(2129),
    [anon_sym_AMP] = ACTIONS(2129),
    [anon_sym_POUND] = ACTIONS(2129),
    [anon_sym_COLON_COLON] = ACTIONS(2131),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2131),
    [anon_sym_else] = ACTIONS(2129),
    [anon_sym_new] = ACTIONS(2129),
    [anon_sym_LBRACE_LT] = ACTIONS(2131),
    [anon_sym_begin] = ACTIONS(2129),
    [sym_ocamlyacc_value] = ACTIONS(2131),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2129),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2131),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2129),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2129),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2131),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2131),
    [aux_sym_number_token1] = ACTIONS(2131),
    [aux_sym_signed_number_token1] = ACTIONS(2129),
    [anon_sym_SQUOTE] = ACTIONS(2131),
    [anon_sym_DQUOTE] = ACTIONS(2131),
    [sym_prefix_operator] = ACTIONS(2131),
    [aux_sym_sign_operator_token1] = ACTIONS(2129),
    [sym_hash_operator] = ACTIONS(2131),
    [anon_sym_lsl] = ACTIONS(2129),
    [anon_sym_lsr] = ACTIONS(2129),
    [anon_sym_asr] = ACTIONS(2129),
    [aux_sym_pow_operator_token1] = ACTIONS(2131),
    [anon_sym_mod] = ACTIONS(2129),
    [anon_sym_land] = ACTIONS(2129),
    [anon_sym_lor] = ACTIONS(2129),
    [anon_sym_lxor] = ACTIONS(2129),
    [aux_sym_mult_operator_token1] = ACTIONS(2129),
    [aux_sym_add_operator_token1] = ACTIONS(2129),
    [sym_concat_operator] = ACTIONS(2131),
    [sym_rel_operator] = ACTIONS(2129),
    [anon_sym_AMP_AMP] = ACTIONS(2129),
    [anon_sym_or] = ACTIONS(2129),
    [anon_sym_PIPE_PIPE] = ACTIONS(2129),
    [sym_assign_operator] = ACTIONS(2131),
    [sym_let_operator] = ACTIONS(2131),
    [sym_let_and_operator] = ACTIONS(2131),
    [sym__uppercase_identifier] = ACTIONS(2131),
    [aux_sym_directive_token1] = ACTIONS(2129),
    [aux_sym_tag_token1] = ACTIONS(2131),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1161)] = {
    [sym_attribute] = STATE(1161),
    [sym__lowercase_identifier] = ACTIONS(2137),
    [anon_sym_SEMI_SEMI] = ACTIONS(2139),
    [anon_sym_let] = ACTIONS(2137),
    [anon_sym_and] = ACTIONS(2137),
    [anon_sym_TILDE] = ACTIONS(2137),
    [anon_sym_QMARK] = ACTIONS(2137),
    [anon_sym_LPAREN] = ACTIONS(2139),
    [anon_sym_external] = ACTIONS(2137),
    [anon_sym_type] = ACTIONS(2137),
    [anon_sym_COMMA] = ACTIONS(2139),
    [anon_sym_PIPE] = ACTIONS(2137),
    [anon_sym_DOT] = ACTIONS(2139),
    [anon_sym_LBRACK] = ACTIONS(2137),
    [anon_sym_RBRACK] = ACTIONS(2139),
    [anon_sym_true] = ACTIONS(2137),
    [anon_sym_false] = ACTIONS(2137),
    [anon_sym_LBRACE] = ACTIONS(2137),
    [anon_sym_SEMI] = ACTIONS(2137),
    [anon_sym_exception] = ACTIONS(2137),
    [anon_sym_module] = ACTIONS(2137),
    [anon_sym_open] = ACTIONS(2137),
    [anon_sym_include] = ACTIONS(2137),
    [anon_sym_class] = ACTIONS(2137),
    [anon_sym_val] = ACTIONS(2137),
    [anon_sym_end] = ACTIONS(2137),
    [anon_sym_object] = ACTIONS(2137),
    [anon_sym_in] = ACTIONS(2137),
    [anon_sym_AMP] = ACTIONS(2137),
    [anon_sym_POUND] = ACTIONS(2137),
    [anon_sym_COLON_COLON] = ACTIONS(2139),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2139),
    [anon_sym_else] = ACTIONS(2137),
    [anon_sym_new] = ACTIONS(2137),
    [anon_sym_LBRACE_LT] = ACTIONS(2139),
    [anon_sym_begin] = ACTIONS(2137),
    [sym_ocamlyacc_value] = ACTIONS(2139),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2137),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2139),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2137),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2137),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2139),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2139),
    [aux_sym_number_token1] = ACTIONS(2139),
    [aux_sym_signed_number_token1] = ACTIONS(2137),
    [anon_sym_SQUOTE] = ACTIONS(2139),
    [anon_sym_DQUOTE] = ACTIONS(2139),
    [sym_prefix_operator] = ACTIONS(2139),
    [aux_sym_sign_operator_token1] = ACTIONS(2137),
    [sym_hash_operator] = ACTIONS(2139),
    [anon_sym_lsl] = ACTIONS(2137),
    [anon_sym_lsr] = ACTIONS(2137),
    [anon_sym_asr] = ACTIONS(2137),
    [aux_sym_pow_operator_token1] = ACTIONS(2139),
    [anon_sym_mod] = ACTIONS(2137),
    [anon_sym_land] = ACTIONS(2137),
    [anon_sym_lor] = ACTIONS(2137),
    [anon_sym_lxor] = ACTIONS(2137),
    [aux_sym_mult_operator_token1] = ACTIONS(2137),
    [aux_sym_add_operator_token1] = ACTIONS(2137),
    [sym_concat_operator] = ACTIONS(2139),
    [sym_rel_operator] = ACTIONS(2137),
    [anon_sym_AMP_AMP] = ACTIONS(2137),
    [anon_sym_or] = ACTIONS(2137),
    [anon_sym_PIPE_PIPE] = ACTIONS(2137),
    [sym_assign_operator] = ACTIONS(2139),
    [sym_let_operator] = ACTIONS(2139),
    [sym_let_and_operator] = ACTIONS(2139),
    [sym__uppercase_identifier] = ACTIONS(2139),
    [aux_sym_directive_token1] = ACTIONS(2137),
    [aux_sym_tag_token1] = ACTIONS(2139),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1162)] = {
    [sym_attribute] = STATE(1162),
    [sym__lowercase_identifier] = ACTIONS(2169),
    [anon_sym_SEMI_SEMI] = ACTIONS(2171),
    [anon_sym_let] = ACTIONS(2169),
    [anon_sym_and] = ACTIONS(2169),
    [anon_sym_TILDE] = ACTIONS(2169),
    [anon_sym_QMARK] = ACTIONS(2169),
    [anon_sym_LPAREN] = ACTIONS(2171),
    [anon_sym_external] = ACTIONS(2169),
    [anon_sym_type] = ACTIONS(2169),
    [anon_sym_COMMA] = ACTIONS(2171),
    [anon_sym_PIPE] = ACTIONS(2169),
    [anon_sym_DOT] = ACTIONS(2171),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_RBRACK] = ACTIONS(2171),
    [anon_sym_true] = ACTIONS(2169),
    [anon_sym_false] = ACTIONS(2169),
    [anon_sym_LBRACE] = ACTIONS(2169),
    [anon_sym_SEMI] = ACTIONS(2169),
    [anon_sym_exception] = ACTIONS(2169),
    [anon_sym_module] = ACTIONS(2169),
    [anon_sym_open] = ACTIONS(2169),
    [anon_sym_include] = ACTIONS(2169),
    [anon_sym_class] = ACTIONS(2169),
    [anon_sym_val] = ACTIONS(2169),
    [anon_sym_end] = ACTIONS(2169),
    [anon_sym_object] = ACTIONS(2169),
    [anon_sym_in] = ACTIONS(2169),
    [anon_sym_AMP] = ACTIONS(2169),
    [anon_sym_POUND] = ACTIONS(2169),
    [anon_sym_COLON_COLON] = ACTIONS(2171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2171),
    [anon_sym_else] = ACTIONS(2169),
    [anon_sym_new] = ACTIONS(2169),
    [anon_sym_LBRACE_LT] = ACTIONS(2171),
    [anon_sym_begin] = ACTIONS(2169),
    [sym_ocamlyacc_value] = ACTIONS(2171),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2169),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2171),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2169),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2169),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2171),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2171),
    [aux_sym_number_token1] = ACTIONS(2171),
    [aux_sym_signed_number_token1] = ACTIONS(2169),
    [anon_sym_SQUOTE] = ACTIONS(2171),
    [anon_sym_DQUOTE] = ACTIONS(2171),
    [sym_prefix_operator] = ACTIONS(2171),
    [aux_sym_sign_operator_token1] = ACTIONS(2169),
    [sym_hash_operator] = ACTIONS(2171),
    [anon_sym_lsl] = ACTIONS(2169),
    [anon_sym_lsr] = ACTIONS(2169),
    [anon_sym_asr] = ACTIONS(2169),
    [aux_sym_pow_operator_token1] = ACTIONS(2171),
    [anon_sym_mod] = ACTIONS(2169),
    [anon_sym_land] = ACTIONS(2169),
    [anon_sym_lor] = ACTIONS(2169),
    [anon_sym_lxor] = ACTIONS(2169),
    [aux_sym_mult_operator_token1] = ACTIONS(2169),
    [aux_sym_add_operator_token1] = ACTIONS(2169),
    [sym_concat_operator] = ACTIONS(2171),
    [sym_rel_operator] = ACTIONS(2169),
    [anon_sym_AMP_AMP] = ACTIONS(2169),
    [anon_sym_or] = ACTIONS(2169),
    [anon_sym_PIPE_PIPE] = ACTIONS(2169),
    [sym_assign_operator] = ACTIONS(2171),
    [sym_let_operator] = ACTIONS(2171),
    [sym_let_and_operator] = ACTIONS(2171),
    [sym__uppercase_identifier] = ACTIONS(2171),
    [aux_sym_directive_token1] = ACTIONS(2169),
    [aux_sym_tag_token1] = ACTIONS(2171),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1163)] = {
    [sym_attribute] = STATE(1163),
    [sym__lowercase_identifier] = ACTIONS(2161),
    [anon_sym_SEMI_SEMI] = ACTIONS(2163),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_and] = ACTIONS(2161),
    [anon_sym_TILDE] = ACTIONS(2161),
    [anon_sym_QMARK] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_external] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_COMMA] = ACTIONS(2163),
    [anon_sym_PIPE] = ACTIONS(2161),
    [anon_sym_DOT] = ACTIONS(2163),
    [anon_sym_LBRACK] = ACTIONS(2161),
    [anon_sym_RBRACK] = ACTIONS(2163),
    [anon_sym_true] = ACTIONS(2161),
    [anon_sym_false] = ACTIONS(2161),
    [anon_sym_LBRACE] = ACTIONS(2161),
    [anon_sym_SEMI] = ACTIONS(2161),
    [anon_sym_exception] = ACTIONS(2161),
    [anon_sym_module] = ACTIONS(2161),
    [anon_sym_open] = ACTIONS(2161),
    [anon_sym_include] = ACTIONS(2161),
    [anon_sym_class] = ACTIONS(2161),
    [anon_sym_val] = ACTIONS(2161),
    [anon_sym_end] = ACTIONS(2161),
    [anon_sym_object] = ACTIONS(2161),
    [anon_sym_in] = ACTIONS(2161),
    [anon_sym_AMP] = ACTIONS(2161),
    [anon_sym_POUND] = ACTIONS(2161),
    [anon_sym_COLON_COLON] = ACTIONS(2163),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2163),
    [anon_sym_else] = ACTIONS(2161),
    [anon_sym_new] = ACTIONS(2161),
    [anon_sym_LBRACE_LT] = ACTIONS(2163),
    [anon_sym_begin] = ACTIONS(2161),
    [sym_ocamlyacc_value] = ACTIONS(2163),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2161),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2163),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2163),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2163),
    [aux_sym_number_token1] = ACTIONS(2163),
    [aux_sym_signed_number_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2163),
    [anon_sym_DQUOTE] = ACTIONS(2163),
    [sym_prefix_operator] = ACTIONS(2163),
    [aux_sym_sign_operator_token1] = ACTIONS(2161),
    [sym_hash_operator] = ACTIONS(2163),
    [anon_sym_lsl] = ACTIONS(2161),
    [anon_sym_lsr] = ACTIONS(2161),
    [anon_sym_asr] = ACTIONS(2161),
    [aux_sym_pow_operator_token1] = ACTIONS(2163),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_land] = ACTIONS(2161),
    [anon_sym_lor] = ACTIONS(2161),
    [anon_sym_lxor] = ACTIONS(2161),
    [aux_sym_mult_operator_token1] = ACTIONS(2161),
    [aux_sym_add_operator_token1] = ACTIONS(2161),
    [sym_concat_operator] = ACTIONS(2163),
    [sym_rel_operator] = ACTIONS(2161),
    [anon_sym_AMP_AMP] = ACTIONS(2161),
    [anon_sym_or] = ACTIONS(2161),
    [anon_sym_PIPE_PIPE] = ACTIONS(2161),
    [sym_assign_operator] = ACTIONS(2163),
    [sym_let_operator] = ACTIONS(2163),
    [sym_let_and_operator] = ACTIONS(2163),
    [sym__uppercase_identifier] = ACTIONS(2163),
    [aux_sym_directive_token1] = ACTIONS(2161),
    [aux_sym_tag_token1] = ACTIONS(2163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1164)] = {
    [sym_attribute] = STATE(1164),
    [ts_builtin_sym_end] = ACTIONS(1909),
    [sym__lowercase_identifier] = ACTIONS(1907),
    [anon_sym_SEMI_SEMI] = ACTIONS(1909),
    [anon_sym_let] = ACTIONS(1907),
    [anon_sym_and] = ACTIONS(1907),
    [anon_sym_TILDE] = ACTIONS(1907),
    [anon_sym_QMARK] = ACTIONS(1907),
    [anon_sym_LPAREN] = ACTIONS(1909),
    [anon_sym_external] = ACTIONS(1907),
    [anon_sym_type] = ACTIONS(1907),
    [anon_sym_COMMA] = ACTIONS(1909),
    [anon_sym_PIPE] = ACTIONS(1907),
    [anon_sym_DOT] = ACTIONS(1909),
    [anon_sym_LBRACK] = ACTIONS(1907),
    [anon_sym_true] = ACTIONS(1907),
    [anon_sym_false] = ACTIONS(1907),
    [anon_sym_LBRACE] = ACTIONS(1907),
    [anon_sym_SEMI] = ACTIONS(1907),
    [anon_sym_exception] = ACTIONS(1907),
    [anon_sym_module] = ACTIONS(1907),
    [anon_sym_open] = ACTIONS(1907),
    [anon_sym_include] = ACTIONS(1907),
    [anon_sym_class] = ACTIONS(1907),
    [anon_sym_val] = ACTIONS(1907),
    [anon_sym_object] = ACTIONS(1907),
    [anon_sym_in] = ACTIONS(1907),
    [anon_sym_AMP] = ACTIONS(1907),
    [anon_sym_POUND] = ACTIONS(1907),
    [anon_sym_COLON_COLON] = ACTIONS(1909),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1909),
    [anon_sym_LT_DASH] = ACTIONS(1907),
    [anon_sym_else] = ACTIONS(1907),
    [anon_sym_new] = ACTIONS(1907),
    [anon_sym_LBRACE_LT] = ACTIONS(1909),
    [anon_sym_begin] = ACTIONS(1907),
    [sym_ocamlyacc_value] = ACTIONS(1909),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1907),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1909),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1907),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1907),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1909),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1909),
    [aux_sym_number_token1] = ACTIONS(1909),
    [aux_sym_signed_number_token1] = ACTIONS(1907),
    [anon_sym_SQUOTE] = ACTIONS(1909),
    [anon_sym_DQUOTE] = ACTIONS(1909),
    [sym_prefix_operator] = ACTIONS(1909),
    [aux_sym_sign_operator_token1] = ACTIONS(1907),
    [sym_hash_operator] = ACTIONS(1909),
    [anon_sym_lsl] = ACTIONS(1907),
    [anon_sym_lsr] = ACTIONS(1907),
    [anon_sym_asr] = ACTIONS(1907),
    [aux_sym_pow_operator_token1] = ACTIONS(1909),
    [anon_sym_mod] = ACTIONS(1907),
    [anon_sym_land] = ACTIONS(1907),
    [anon_sym_lor] = ACTIONS(1907),
    [anon_sym_lxor] = ACTIONS(1907),
    [aux_sym_mult_operator_token1] = ACTIONS(1907),
    [aux_sym_add_operator_token1] = ACTIONS(1907),
    [sym_concat_operator] = ACTIONS(1909),
    [sym_rel_operator] = ACTIONS(1907),
    [anon_sym_AMP_AMP] = ACTIONS(1907),
    [anon_sym_or] = ACTIONS(1907),
    [anon_sym_PIPE_PIPE] = ACTIONS(1907),
    [sym_assign_operator] = ACTIONS(1909),
    [sym_let_operator] = ACTIONS(1909),
    [sym_let_and_operator] = ACTIONS(1909),
    [sym__uppercase_identifier] = ACTIONS(1909),
    [aux_sym_directive_token1] = ACTIONS(1907),
    [aux_sym_tag_token1] = ACTIONS(1909),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1165)] = {
    [sym_attribute] = STATE(1165),
    [sym__lowercase_identifier] = ACTIONS(2047),
    [anon_sym_SEMI_SEMI] = ACTIONS(2049),
    [anon_sym_let] = ACTIONS(2047),
    [anon_sym_and] = ACTIONS(2047),
    [anon_sym_TILDE] = ACTIONS(2047),
    [anon_sym_QMARK] = ACTIONS(2047),
    [anon_sym_LPAREN] = ACTIONS(2049),
    [anon_sym_external] = ACTIONS(2047),
    [anon_sym_type] = ACTIONS(2047),
    [anon_sym_COMMA] = ACTIONS(2049),
    [anon_sym_PIPE] = ACTIONS(2047),
    [anon_sym_DOT] = ACTIONS(2049),
    [anon_sym_LBRACK] = ACTIONS(2047),
    [anon_sym_RBRACK] = ACTIONS(2049),
    [anon_sym_true] = ACTIONS(2047),
    [anon_sym_false] = ACTIONS(2047),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_SEMI] = ACTIONS(2047),
    [anon_sym_exception] = ACTIONS(2047),
    [anon_sym_module] = ACTIONS(2047),
    [anon_sym_open] = ACTIONS(2047),
    [anon_sym_include] = ACTIONS(2047),
    [anon_sym_class] = ACTIONS(2047),
    [anon_sym_val] = ACTIONS(2047),
    [anon_sym_end] = ACTIONS(2047),
    [anon_sym_object] = ACTIONS(2047),
    [anon_sym_in] = ACTIONS(2047),
    [anon_sym_AMP] = ACTIONS(2047),
    [anon_sym_POUND] = ACTIONS(2047),
    [anon_sym_COLON_COLON] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2049),
    [anon_sym_else] = ACTIONS(2047),
    [anon_sym_new] = ACTIONS(2047),
    [anon_sym_LBRACE_LT] = ACTIONS(2049),
    [anon_sym_begin] = ACTIONS(2047),
    [sym_ocamlyacc_value] = ACTIONS(2049),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2047),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2049),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2047),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2047),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2049),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2049),
    [aux_sym_number_token1] = ACTIONS(2049),
    [aux_sym_signed_number_token1] = ACTIONS(2047),
    [anon_sym_SQUOTE] = ACTIONS(2049),
    [anon_sym_DQUOTE] = ACTIONS(2049),
    [sym_prefix_operator] = ACTIONS(2049),
    [aux_sym_sign_operator_token1] = ACTIONS(2047),
    [sym_hash_operator] = ACTIONS(2049),
    [anon_sym_lsl] = ACTIONS(2047),
    [anon_sym_lsr] = ACTIONS(2047),
    [anon_sym_asr] = ACTIONS(2047),
    [aux_sym_pow_operator_token1] = ACTIONS(2049),
    [anon_sym_mod] = ACTIONS(2047),
    [anon_sym_land] = ACTIONS(2047),
    [anon_sym_lor] = ACTIONS(2047),
    [anon_sym_lxor] = ACTIONS(2047),
    [aux_sym_mult_operator_token1] = ACTIONS(2047),
    [aux_sym_add_operator_token1] = ACTIONS(2047),
    [sym_concat_operator] = ACTIONS(2049),
    [sym_rel_operator] = ACTIONS(2047),
    [anon_sym_AMP_AMP] = ACTIONS(2047),
    [anon_sym_or] = ACTIONS(2047),
    [anon_sym_PIPE_PIPE] = ACTIONS(2047),
    [sym_assign_operator] = ACTIONS(2049),
    [sym_let_operator] = ACTIONS(2049),
    [sym_let_and_operator] = ACTIONS(2049),
    [sym__uppercase_identifier] = ACTIONS(2049),
    [aux_sym_directive_token1] = ACTIONS(2047),
    [aux_sym_tag_token1] = ACTIONS(2049),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1166)] = {
    [sym_attribute] = STATE(1166),
    [sym__lowercase_identifier] = ACTIONS(2165),
    [anon_sym_SEMI_SEMI] = ACTIONS(2167),
    [anon_sym_let] = ACTIONS(2165),
    [anon_sym_and] = ACTIONS(2165),
    [anon_sym_TILDE] = ACTIONS(2165),
    [anon_sym_QMARK] = ACTIONS(2165),
    [anon_sym_LPAREN] = ACTIONS(2167),
    [anon_sym_external] = ACTIONS(2165),
    [anon_sym_type] = ACTIONS(2165),
    [anon_sym_COMMA] = ACTIONS(2167),
    [anon_sym_PIPE] = ACTIONS(2165),
    [anon_sym_DOT] = ACTIONS(2167),
    [anon_sym_LBRACK] = ACTIONS(2165),
    [anon_sym_RBRACK] = ACTIONS(2167),
    [anon_sym_true] = ACTIONS(2165),
    [anon_sym_false] = ACTIONS(2165),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_SEMI] = ACTIONS(2165),
    [anon_sym_exception] = ACTIONS(2165),
    [anon_sym_module] = ACTIONS(2165),
    [anon_sym_open] = ACTIONS(2165),
    [anon_sym_include] = ACTIONS(2165),
    [anon_sym_class] = ACTIONS(2165),
    [anon_sym_val] = ACTIONS(2165),
    [anon_sym_end] = ACTIONS(2165),
    [anon_sym_object] = ACTIONS(2165),
    [anon_sym_in] = ACTIONS(2165),
    [anon_sym_AMP] = ACTIONS(2165),
    [anon_sym_POUND] = ACTIONS(2165),
    [anon_sym_COLON_COLON] = ACTIONS(2167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2167),
    [anon_sym_else] = ACTIONS(2165),
    [anon_sym_new] = ACTIONS(2165),
    [anon_sym_LBRACE_LT] = ACTIONS(2167),
    [anon_sym_begin] = ACTIONS(2165),
    [sym_ocamlyacc_value] = ACTIONS(2167),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2165),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2167),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2167),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2167),
    [aux_sym_number_token1] = ACTIONS(2167),
    [aux_sym_signed_number_token1] = ACTIONS(2165),
    [anon_sym_SQUOTE] = ACTIONS(2167),
    [anon_sym_DQUOTE] = ACTIONS(2167),
    [sym_prefix_operator] = ACTIONS(2167),
    [aux_sym_sign_operator_token1] = ACTIONS(2165),
    [sym_hash_operator] = ACTIONS(2167),
    [anon_sym_lsl] = ACTIONS(2165),
    [anon_sym_lsr] = ACTIONS(2165),
    [anon_sym_asr] = ACTIONS(2165),
    [aux_sym_pow_operator_token1] = ACTIONS(2167),
    [anon_sym_mod] = ACTIONS(2165),
    [anon_sym_land] = ACTIONS(2165),
    [anon_sym_lor] = ACTIONS(2165),
    [anon_sym_lxor] = ACTIONS(2165),
    [aux_sym_mult_operator_token1] = ACTIONS(2165),
    [aux_sym_add_operator_token1] = ACTIONS(2165),
    [sym_concat_operator] = ACTIONS(2167),
    [sym_rel_operator] = ACTIONS(2165),
    [anon_sym_AMP_AMP] = ACTIONS(2165),
    [anon_sym_or] = ACTIONS(2165),
    [anon_sym_PIPE_PIPE] = ACTIONS(2165),
    [sym_assign_operator] = ACTIONS(2167),
    [sym_let_operator] = ACTIONS(2167),
    [sym_let_and_operator] = ACTIONS(2167),
    [sym__uppercase_identifier] = ACTIONS(2167),
    [aux_sym_directive_token1] = ACTIONS(2165),
    [aux_sym_tag_token1] = ACTIONS(2167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1167)] = {
    [sym_attribute] = STATE(1167),
    [sym__lowercase_identifier] = ACTIONS(2111),
    [anon_sym_SEMI_SEMI] = ACTIONS(2113),
    [anon_sym_let] = ACTIONS(2111),
    [anon_sym_and] = ACTIONS(2111),
    [anon_sym_TILDE] = ACTIONS(2111),
    [anon_sym_QMARK] = ACTIONS(2111),
    [anon_sym_LPAREN] = ACTIONS(2113),
    [anon_sym_external] = ACTIONS(2111),
    [anon_sym_type] = ACTIONS(2111),
    [anon_sym_COMMA] = ACTIONS(2113),
    [anon_sym_PIPE] = ACTIONS(2111),
    [anon_sym_DOT] = ACTIONS(2177),
    [anon_sym_LBRACK] = ACTIONS(2111),
    [anon_sym_RBRACK] = ACTIONS(2113),
    [anon_sym_true] = ACTIONS(2111),
    [anon_sym_false] = ACTIONS(2111),
    [anon_sym_LBRACE] = ACTIONS(2111),
    [anon_sym_SEMI] = ACTIONS(2111),
    [anon_sym_exception] = ACTIONS(2111),
    [anon_sym_module] = ACTIONS(2111),
    [anon_sym_open] = ACTIONS(2111),
    [anon_sym_include] = ACTIONS(2111),
    [anon_sym_class] = ACTIONS(2111),
    [anon_sym_val] = ACTIONS(2111),
    [anon_sym_end] = ACTIONS(2111),
    [anon_sym_object] = ACTIONS(2111),
    [anon_sym_in] = ACTIONS(2111),
    [anon_sym_AMP] = ACTIONS(2111),
    [anon_sym_POUND] = ACTIONS(2111),
    [anon_sym_COLON_COLON] = ACTIONS(2113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2113),
    [anon_sym_else] = ACTIONS(2111),
    [anon_sym_new] = ACTIONS(2111),
    [anon_sym_LBRACE_LT] = ACTIONS(2113),
    [anon_sym_begin] = ACTIONS(2111),
    [sym_ocamlyacc_value] = ACTIONS(2113),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2113),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2113),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2113),
    [aux_sym_number_token1] = ACTIONS(2113),
    [aux_sym_signed_number_token1] = ACTIONS(2111),
    [anon_sym_SQUOTE] = ACTIONS(2113),
    [anon_sym_DQUOTE] = ACTIONS(2113),
    [sym_prefix_operator] = ACTIONS(2113),
    [aux_sym_sign_operator_token1] = ACTIONS(2111),
    [sym_hash_operator] = ACTIONS(2113),
    [anon_sym_lsl] = ACTIONS(2111),
    [anon_sym_lsr] = ACTIONS(2111),
    [anon_sym_asr] = ACTIONS(2111),
    [aux_sym_pow_operator_token1] = ACTIONS(2113),
    [anon_sym_mod] = ACTIONS(2111),
    [anon_sym_land] = ACTIONS(2111),
    [anon_sym_lor] = ACTIONS(2111),
    [anon_sym_lxor] = ACTIONS(2111),
    [aux_sym_mult_operator_token1] = ACTIONS(2111),
    [aux_sym_add_operator_token1] = ACTIONS(2111),
    [sym_concat_operator] = ACTIONS(2113),
    [sym_rel_operator] = ACTIONS(2111),
    [anon_sym_AMP_AMP] = ACTIONS(2111),
    [anon_sym_or] = ACTIONS(2111),
    [anon_sym_PIPE_PIPE] = ACTIONS(2111),
    [sym_assign_operator] = ACTIONS(2113),
    [sym_let_operator] = ACTIONS(2113),
    [sym_let_and_operator] = ACTIONS(2113),
    [sym__uppercase_identifier] = ACTIONS(2113),
    [aux_sym_directive_token1] = ACTIONS(2111),
    [aux_sym_tag_token1] = ACTIONS(2113),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1168)] = {
    [sym_attribute] = STATE(1168),
    [sym__lowercase_identifier] = ACTIONS(1939),
    [anon_sym_SEMI_SEMI] = ACTIONS(1941),
    [anon_sym_let] = ACTIONS(1939),
    [anon_sym_and] = ACTIONS(1939),
    [anon_sym_TILDE] = ACTIONS(1939),
    [anon_sym_QMARK] = ACTIONS(1939),
    [anon_sym_LPAREN] = ACTIONS(1941),
    [anon_sym_external] = ACTIONS(1939),
    [anon_sym_type] = ACTIONS(1939),
    [anon_sym_COMMA] = ACTIONS(1941),
    [anon_sym_PIPE] = ACTIONS(1939),
    [anon_sym_DOT] = ACTIONS(1941),
    [anon_sym_LBRACK] = ACTIONS(1939),
    [anon_sym_RBRACK] = ACTIONS(1941),
    [anon_sym_true] = ACTIONS(1939),
    [anon_sym_false] = ACTIONS(1939),
    [anon_sym_LBRACE] = ACTIONS(1939),
    [anon_sym_SEMI] = ACTIONS(1939),
    [anon_sym_exception] = ACTIONS(1939),
    [anon_sym_module] = ACTIONS(1939),
    [anon_sym_open] = ACTIONS(1939),
    [anon_sym_include] = ACTIONS(1939),
    [anon_sym_class] = ACTIONS(1939),
    [anon_sym_val] = ACTIONS(1939),
    [anon_sym_end] = ACTIONS(1939),
    [anon_sym_object] = ACTIONS(1939),
    [anon_sym_in] = ACTIONS(1939),
    [anon_sym_AMP] = ACTIONS(1939),
    [anon_sym_POUND] = ACTIONS(1939),
    [anon_sym_COLON_COLON] = ACTIONS(1941),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1941),
    [anon_sym_else] = ACTIONS(1939),
    [anon_sym_new] = ACTIONS(1939),
    [anon_sym_LBRACE_LT] = ACTIONS(1941),
    [anon_sym_begin] = ACTIONS(1939),
    [sym_ocamlyacc_value] = ACTIONS(1941),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1939),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1941),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1939),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1939),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1941),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1941),
    [aux_sym_number_token1] = ACTIONS(1941),
    [aux_sym_signed_number_token1] = ACTIONS(1939),
    [anon_sym_SQUOTE] = ACTIONS(1941),
    [anon_sym_DQUOTE] = ACTIONS(1941),
    [sym_prefix_operator] = ACTIONS(1941),
    [aux_sym_sign_operator_token1] = ACTIONS(1939),
    [sym_hash_operator] = ACTIONS(1941),
    [anon_sym_lsl] = ACTIONS(1939),
    [anon_sym_lsr] = ACTIONS(1939),
    [anon_sym_asr] = ACTIONS(1939),
    [aux_sym_pow_operator_token1] = ACTIONS(1941),
    [anon_sym_mod] = ACTIONS(1939),
    [anon_sym_land] = ACTIONS(1939),
    [anon_sym_lor] = ACTIONS(1939),
    [anon_sym_lxor] = ACTIONS(1939),
    [aux_sym_mult_operator_token1] = ACTIONS(1939),
    [aux_sym_add_operator_token1] = ACTIONS(1939),
    [sym_concat_operator] = ACTIONS(1941),
    [sym_rel_operator] = ACTIONS(1939),
    [anon_sym_AMP_AMP] = ACTIONS(1939),
    [anon_sym_or] = ACTIONS(1939),
    [anon_sym_PIPE_PIPE] = ACTIONS(1939),
    [sym_assign_operator] = ACTIONS(1941),
    [sym_let_operator] = ACTIONS(1941),
    [sym_let_and_operator] = ACTIONS(1941),
    [sym__uppercase_identifier] = ACTIONS(1941),
    [aux_sym_directive_token1] = ACTIONS(1939),
    [aux_sym_tag_token1] = ACTIONS(1941),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1169)] = {
    [sym_attribute] = STATE(1169),
    [sym__lowercase_identifier] = ACTIONS(2103),
    [anon_sym_SEMI_SEMI] = ACTIONS(2105),
    [anon_sym_let] = ACTIONS(2103),
    [anon_sym_and] = ACTIONS(2103),
    [anon_sym_TILDE] = ACTIONS(2103),
    [anon_sym_QMARK] = ACTIONS(2103),
    [anon_sym_LPAREN] = ACTIONS(2105),
    [anon_sym_external] = ACTIONS(2103),
    [anon_sym_type] = ACTIONS(2103),
    [anon_sym_COMMA] = ACTIONS(2105),
    [anon_sym_PIPE] = ACTIONS(2103),
    [anon_sym_DOT] = ACTIONS(2105),
    [anon_sym_LBRACK] = ACTIONS(2103),
    [anon_sym_RBRACK] = ACTIONS(2105),
    [anon_sym_true] = ACTIONS(2103),
    [anon_sym_false] = ACTIONS(2103),
    [anon_sym_LBRACE] = ACTIONS(2103),
    [anon_sym_SEMI] = ACTIONS(2103),
    [anon_sym_exception] = ACTIONS(2103),
    [anon_sym_module] = ACTIONS(2103),
    [anon_sym_open] = ACTIONS(2103),
    [anon_sym_include] = ACTIONS(2103),
    [anon_sym_class] = ACTIONS(2103),
    [anon_sym_val] = ACTIONS(2103),
    [anon_sym_end] = ACTIONS(2103),
    [anon_sym_object] = ACTIONS(2103),
    [anon_sym_in] = ACTIONS(2103),
    [anon_sym_AMP] = ACTIONS(2103),
    [anon_sym_POUND] = ACTIONS(2103),
    [anon_sym_COLON_COLON] = ACTIONS(2105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2105),
    [anon_sym_else] = ACTIONS(2103),
    [anon_sym_new] = ACTIONS(2103),
    [anon_sym_LBRACE_LT] = ACTIONS(2105),
    [anon_sym_begin] = ACTIONS(2103),
    [sym_ocamlyacc_value] = ACTIONS(2105),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2103),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2105),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2103),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2103),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2105),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2105),
    [aux_sym_number_token1] = ACTIONS(2105),
    [aux_sym_signed_number_token1] = ACTIONS(2103),
    [anon_sym_SQUOTE] = ACTIONS(2105),
    [anon_sym_DQUOTE] = ACTIONS(2105),
    [sym_prefix_operator] = ACTIONS(2105),
    [aux_sym_sign_operator_token1] = ACTIONS(2103),
    [sym_hash_operator] = ACTIONS(2105),
    [anon_sym_lsl] = ACTIONS(2103),
    [anon_sym_lsr] = ACTIONS(2103),
    [anon_sym_asr] = ACTIONS(2103),
    [aux_sym_pow_operator_token1] = ACTIONS(2105),
    [anon_sym_mod] = ACTIONS(2103),
    [anon_sym_land] = ACTIONS(2103),
    [anon_sym_lor] = ACTIONS(2103),
    [anon_sym_lxor] = ACTIONS(2103),
    [aux_sym_mult_operator_token1] = ACTIONS(2103),
    [aux_sym_add_operator_token1] = ACTIONS(2103),
    [sym_concat_operator] = ACTIONS(2105),
    [sym_rel_operator] = ACTIONS(2103),
    [anon_sym_AMP_AMP] = ACTIONS(2103),
    [anon_sym_or] = ACTIONS(2103),
    [anon_sym_PIPE_PIPE] = ACTIONS(2103),
    [sym_assign_operator] = ACTIONS(2105),
    [sym_let_operator] = ACTIONS(2105),
    [sym_let_and_operator] = ACTIONS(2105),
    [sym__uppercase_identifier] = ACTIONS(2105),
    [aux_sym_directive_token1] = ACTIONS(2103),
    [aux_sym_tag_token1] = ACTIONS(2105),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1170)] = {
    [sym_attribute] = STATE(1170),
    [sym__lowercase_identifier] = ACTIONS(1943),
    [anon_sym_SEMI_SEMI] = ACTIONS(1945),
    [anon_sym_let] = ACTIONS(1943),
    [anon_sym_and] = ACTIONS(1943),
    [anon_sym_TILDE] = ACTIONS(1943),
    [anon_sym_QMARK] = ACTIONS(1943),
    [anon_sym_LPAREN] = ACTIONS(1945),
    [anon_sym_external] = ACTIONS(1943),
    [anon_sym_type] = ACTIONS(1943),
    [anon_sym_COMMA] = ACTIONS(1945),
    [anon_sym_PIPE] = ACTIONS(1943),
    [anon_sym_DOT] = ACTIONS(1945),
    [anon_sym_LBRACK] = ACTIONS(1943),
    [anon_sym_RBRACK] = ACTIONS(1945),
    [anon_sym_true] = ACTIONS(1943),
    [anon_sym_false] = ACTIONS(1943),
    [anon_sym_LBRACE] = ACTIONS(1943),
    [anon_sym_SEMI] = ACTIONS(1943),
    [anon_sym_exception] = ACTIONS(1943),
    [anon_sym_module] = ACTIONS(1943),
    [anon_sym_open] = ACTIONS(1943),
    [anon_sym_include] = ACTIONS(1943),
    [anon_sym_class] = ACTIONS(1943),
    [anon_sym_val] = ACTIONS(1943),
    [anon_sym_end] = ACTIONS(1943),
    [anon_sym_object] = ACTIONS(1943),
    [anon_sym_in] = ACTIONS(1943),
    [anon_sym_AMP] = ACTIONS(1943),
    [anon_sym_POUND] = ACTIONS(1943),
    [anon_sym_COLON_COLON] = ACTIONS(1945),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1945),
    [anon_sym_else] = ACTIONS(1943),
    [anon_sym_new] = ACTIONS(1943),
    [anon_sym_LBRACE_LT] = ACTIONS(1945),
    [anon_sym_begin] = ACTIONS(1943),
    [sym_ocamlyacc_value] = ACTIONS(1945),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1943),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1945),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1943),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1943),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1945),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1945),
    [aux_sym_number_token1] = ACTIONS(1945),
    [aux_sym_signed_number_token1] = ACTIONS(1943),
    [anon_sym_SQUOTE] = ACTIONS(1945),
    [anon_sym_DQUOTE] = ACTIONS(1945),
    [sym_prefix_operator] = ACTIONS(1945),
    [aux_sym_sign_operator_token1] = ACTIONS(1943),
    [sym_hash_operator] = ACTIONS(1945),
    [anon_sym_lsl] = ACTIONS(1943),
    [anon_sym_lsr] = ACTIONS(1943),
    [anon_sym_asr] = ACTIONS(1943),
    [aux_sym_pow_operator_token1] = ACTIONS(1945),
    [anon_sym_mod] = ACTIONS(1943),
    [anon_sym_land] = ACTIONS(1943),
    [anon_sym_lor] = ACTIONS(1943),
    [anon_sym_lxor] = ACTIONS(1943),
    [aux_sym_mult_operator_token1] = ACTIONS(1943),
    [aux_sym_add_operator_token1] = ACTIONS(1943),
    [sym_concat_operator] = ACTIONS(1945),
    [sym_rel_operator] = ACTIONS(1943),
    [anon_sym_AMP_AMP] = ACTIONS(1943),
    [anon_sym_or] = ACTIONS(1943),
    [anon_sym_PIPE_PIPE] = ACTIONS(1943),
    [sym_assign_operator] = ACTIONS(1945),
    [sym_let_operator] = ACTIONS(1945),
    [sym_let_and_operator] = ACTIONS(1945),
    [sym__uppercase_identifier] = ACTIONS(1945),
    [aux_sym_directive_token1] = ACTIONS(1943),
    [aux_sym_tag_token1] = ACTIONS(1945),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1171)] = {
    [sym_attribute] = STATE(1171),
    [sym__lowercase_identifier] = ACTIONS(1971),
    [anon_sym_SEMI_SEMI] = ACTIONS(1973),
    [anon_sym_let] = ACTIONS(1971),
    [anon_sym_and] = ACTIONS(1971),
    [anon_sym_TILDE] = ACTIONS(1971),
    [anon_sym_QMARK] = ACTIONS(1971),
    [anon_sym_LPAREN] = ACTIONS(1973),
    [anon_sym_external] = ACTIONS(1971),
    [anon_sym_type] = ACTIONS(1971),
    [anon_sym_COMMA] = ACTIONS(1973),
    [anon_sym_PIPE] = ACTIONS(1971),
    [anon_sym_DOT] = ACTIONS(1007),
    [anon_sym_LBRACK] = ACTIONS(1971),
    [anon_sym_RBRACK] = ACTIONS(1973),
    [anon_sym_true] = ACTIONS(1971),
    [anon_sym_false] = ACTIONS(1971),
    [anon_sym_LBRACE] = ACTIONS(1971),
    [anon_sym_SEMI] = ACTIONS(1971),
    [anon_sym_exception] = ACTIONS(1971),
    [anon_sym_module] = ACTIONS(1971),
    [anon_sym_open] = ACTIONS(1971),
    [anon_sym_include] = ACTIONS(1971),
    [anon_sym_class] = ACTIONS(1971),
    [anon_sym_val] = ACTIONS(1971),
    [anon_sym_end] = ACTIONS(1971),
    [anon_sym_object] = ACTIONS(1971),
    [anon_sym_in] = ACTIONS(1971),
    [anon_sym_AMP] = ACTIONS(1971),
    [anon_sym_POUND] = ACTIONS(1971),
    [anon_sym_COLON_COLON] = ACTIONS(1973),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1973),
    [anon_sym_else] = ACTIONS(1971),
    [anon_sym_new] = ACTIONS(1971),
    [anon_sym_LBRACE_LT] = ACTIONS(1973),
    [anon_sym_begin] = ACTIONS(1971),
    [sym_ocamlyacc_value] = ACTIONS(1973),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1971),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1973),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1973),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1973),
    [aux_sym_number_token1] = ACTIONS(1973),
    [aux_sym_signed_number_token1] = ACTIONS(1971),
    [anon_sym_SQUOTE] = ACTIONS(1973),
    [anon_sym_DQUOTE] = ACTIONS(1973),
    [sym_prefix_operator] = ACTIONS(1973),
    [aux_sym_sign_operator_token1] = ACTIONS(1971),
    [sym_hash_operator] = ACTIONS(1973),
    [anon_sym_lsl] = ACTIONS(1971),
    [anon_sym_lsr] = ACTIONS(1971),
    [anon_sym_asr] = ACTIONS(1971),
    [aux_sym_pow_operator_token1] = ACTIONS(1973),
    [anon_sym_mod] = ACTIONS(1971),
    [anon_sym_land] = ACTIONS(1971),
    [anon_sym_lor] = ACTIONS(1971),
    [anon_sym_lxor] = ACTIONS(1971),
    [aux_sym_mult_operator_token1] = ACTIONS(1971),
    [aux_sym_add_operator_token1] = ACTIONS(1971),
    [sym_concat_operator] = ACTIONS(1973),
    [sym_rel_operator] = ACTIONS(1971),
    [anon_sym_AMP_AMP] = ACTIONS(1971),
    [anon_sym_or] = ACTIONS(1971),
    [anon_sym_PIPE_PIPE] = ACTIONS(1971),
    [sym_assign_operator] = ACTIONS(1973),
    [sym_let_operator] = ACTIONS(1973),
    [sym_let_and_operator] = ACTIONS(1973),
    [sym__uppercase_identifier] = ACTIONS(1973),
    [aux_sym_directive_token1] = ACTIONS(1971),
    [aux_sym_tag_token1] = ACTIONS(1973),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1172)] = {
    [sym_attribute] = STATE(1172),
    [sym__lowercase_identifier] = ACTIONS(2183),
    [anon_sym_SEMI_SEMI] = ACTIONS(2185),
    [anon_sym_let] = ACTIONS(2183),
    [anon_sym_and] = ACTIONS(2183),
    [anon_sym_TILDE] = ACTIONS(2183),
    [anon_sym_QMARK] = ACTIONS(2183),
    [anon_sym_LPAREN] = ACTIONS(2185),
    [anon_sym_external] = ACTIONS(2183),
    [anon_sym_type] = ACTIONS(2183),
    [anon_sym_COMMA] = ACTIONS(2185),
    [anon_sym_PIPE] = ACTIONS(2183),
    [anon_sym_DOT] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2183),
    [anon_sym_RBRACK] = ACTIONS(2185),
    [anon_sym_true] = ACTIONS(2183),
    [anon_sym_false] = ACTIONS(2183),
    [anon_sym_LBRACE] = ACTIONS(2183),
    [anon_sym_SEMI] = ACTIONS(2183),
    [anon_sym_exception] = ACTIONS(2183),
    [anon_sym_module] = ACTIONS(2183),
    [anon_sym_open] = ACTIONS(2183),
    [anon_sym_include] = ACTIONS(2183),
    [anon_sym_class] = ACTIONS(2183),
    [anon_sym_val] = ACTIONS(2183),
    [anon_sym_end] = ACTIONS(2183),
    [anon_sym_object] = ACTIONS(2183),
    [anon_sym_in] = ACTIONS(2183),
    [anon_sym_AMP] = ACTIONS(2183),
    [anon_sym_POUND] = ACTIONS(2183),
    [anon_sym_COLON_COLON] = ACTIONS(2185),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2185),
    [anon_sym_else] = ACTIONS(2183),
    [anon_sym_new] = ACTIONS(2183),
    [anon_sym_LBRACE_LT] = ACTIONS(2185),
    [anon_sym_begin] = ACTIONS(2183),
    [sym_ocamlyacc_value] = ACTIONS(2185),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2183),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2185),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2183),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2183),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2185),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2185),
    [aux_sym_number_token1] = ACTIONS(2185),
    [aux_sym_signed_number_token1] = ACTIONS(2183),
    [anon_sym_SQUOTE] = ACTIONS(2185),
    [anon_sym_DQUOTE] = ACTIONS(2185),
    [sym_prefix_operator] = ACTIONS(2185),
    [aux_sym_sign_operator_token1] = ACTIONS(2183),
    [sym_hash_operator] = ACTIONS(2185),
    [anon_sym_lsl] = ACTIONS(2183),
    [anon_sym_lsr] = ACTIONS(2183),
    [anon_sym_asr] = ACTIONS(2183),
    [aux_sym_pow_operator_token1] = ACTIONS(2185),
    [anon_sym_mod] = ACTIONS(2183),
    [anon_sym_land] = ACTIONS(2183),
    [anon_sym_lor] = ACTIONS(2183),
    [anon_sym_lxor] = ACTIONS(2183),
    [aux_sym_mult_operator_token1] = ACTIONS(2183),
    [aux_sym_add_operator_token1] = ACTIONS(2183),
    [sym_concat_operator] = ACTIONS(2185),
    [sym_rel_operator] = ACTIONS(2183),
    [anon_sym_AMP_AMP] = ACTIONS(2183),
    [anon_sym_or] = ACTIONS(2183),
    [anon_sym_PIPE_PIPE] = ACTIONS(2183),
    [sym_assign_operator] = ACTIONS(2185),
    [sym_let_operator] = ACTIONS(2185),
    [sym_let_and_operator] = ACTIONS(2185),
    [sym__uppercase_identifier] = ACTIONS(2185),
    [aux_sym_directive_token1] = ACTIONS(2183),
    [aux_sym_tag_token1] = ACTIONS(2185),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1173)] = {
    [sym_attribute] = STATE(1173),
    [sym__lowercase_identifier] = ACTIONS(2187),
    [anon_sym_SEMI_SEMI] = ACTIONS(2189),
    [anon_sym_let] = ACTIONS(2187),
    [anon_sym_and] = ACTIONS(2187),
    [anon_sym_TILDE] = ACTIONS(2187),
    [anon_sym_QMARK] = ACTIONS(2187),
    [anon_sym_LPAREN] = ACTIONS(2189),
    [anon_sym_external] = ACTIONS(2187),
    [anon_sym_type] = ACTIONS(2187),
    [anon_sym_COMMA] = ACTIONS(2189),
    [anon_sym_PIPE] = ACTIONS(2187),
    [anon_sym_DOT] = ACTIONS(2189),
    [anon_sym_LBRACK] = ACTIONS(2187),
    [anon_sym_RBRACK] = ACTIONS(2189),
    [anon_sym_true] = ACTIONS(2187),
    [anon_sym_false] = ACTIONS(2187),
    [anon_sym_LBRACE] = ACTIONS(2187),
    [anon_sym_SEMI] = ACTIONS(2187),
    [anon_sym_exception] = ACTIONS(2187),
    [anon_sym_module] = ACTIONS(2187),
    [anon_sym_open] = ACTIONS(2187),
    [anon_sym_include] = ACTIONS(2187),
    [anon_sym_class] = ACTIONS(2187),
    [anon_sym_val] = ACTIONS(2187),
    [anon_sym_end] = ACTIONS(2187),
    [anon_sym_object] = ACTIONS(2187),
    [anon_sym_in] = ACTIONS(2187),
    [anon_sym_AMP] = ACTIONS(2187),
    [anon_sym_POUND] = ACTIONS(2187),
    [anon_sym_COLON_COLON] = ACTIONS(2189),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2189),
    [anon_sym_else] = ACTIONS(2187),
    [anon_sym_new] = ACTIONS(2187),
    [anon_sym_LBRACE_LT] = ACTIONS(2189),
    [anon_sym_begin] = ACTIONS(2187),
    [sym_ocamlyacc_value] = ACTIONS(2189),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2187),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2189),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2187),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2187),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2189),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2189),
    [aux_sym_number_token1] = ACTIONS(2189),
    [aux_sym_signed_number_token1] = ACTIONS(2187),
    [anon_sym_SQUOTE] = ACTIONS(2189),
    [anon_sym_DQUOTE] = ACTIONS(2189),
    [sym_prefix_operator] = ACTIONS(2189),
    [aux_sym_sign_operator_token1] = ACTIONS(2187),
    [sym_hash_operator] = ACTIONS(2189),
    [anon_sym_lsl] = ACTIONS(2187),
    [anon_sym_lsr] = ACTIONS(2187),
    [anon_sym_asr] = ACTIONS(2187),
    [aux_sym_pow_operator_token1] = ACTIONS(2189),
    [anon_sym_mod] = ACTIONS(2187),
    [anon_sym_land] = ACTIONS(2187),
    [anon_sym_lor] = ACTIONS(2187),
    [anon_sym_lxor] = ACTIONS(2187),
    [aux_sym_mult_operator_token1] = ACTIONS(2187),
    [aux_sym_add_operator_token1] = ACTIONS(2187),
    [sym_concat_operator] = ACTIONS(2189),
    [sym_rel_operator] = ACTIONS(2187),
    [anon_sym_AMP_AMP] = ACTIONS(2187),
    [anon_sym_or] = ACTIONS(2187),
    [anon_sym_PIPE_PIPE] = ACTIONS(2187),
    [sym_assign_operator] = ACTIONS(2189),
    [sym_let_operator] = ACTIONS(2189),
    [sym_let_and_operator] = ACTIONS(2189),
    [sym__uppercase_identifier] = ACTIONS(2189),
    [aux_sym_directive_token1] = ACTIONS(2187),
    [aux_sym_tag_token1] = ACTIONS(2189),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1174)] = {
    [sym_attribute] = STATE(1174),
    [sym__lowercase_identifier] = ACTIONS(2191),
    [anon_sym_SEMI_SEMI] = ACTIONS(2193),
    [anon_sym_let] = ACTIONS(2191),
    [anon_sym_and] = ACTIONS(2191),
    [anon_sym_TILDE] = ACTIONS(2191),
    [anon_sym_QMARK] = ACTIONS(2191),
    [anon_sym_LPAREN] = ACTIONS(2193),
    [anon_sym_external] = ACTIONS(2191),
    [anon_sym_type] = ACTIONS(2191),
    [anon_sym_COMMA] = ACTIONS(2193),
    [anon_sym_PIPE] = ACTIONS(2191),
    [anon_sym_DOT] = ACTIONS(2193),
    [anon_sym_LBRACK] = ACTIONS(2191),
    [anon_sym_RBRACK] = ACTIONS(2193),
    [anon_sym_true] = ACTIONS(2191),
    [anon_sym_false] = ACTIONS(2191),
    [anon_sym_LBRACE] = ACTIONS(2191),
    [anon_sym_SEMI] = ACTIONS(2191),
    [anon_sym_exception] = ACTIONS(2191),
    [anon_sym_module] = ACTIONS(2191),
    [anon_sym_open] = ACTIONS(2191),
    [anon_sym_include] = ACTIONS(2191),
    [anon_sym_class] = ACTIONS(2191),
    [anon_sym_val] = ACTIONS(2191),
    [anon_sym_end] = ACTIONS(2191),
    [anon_sym_object] = ACTIONS(2191),
    [anon_sym_in] = ACTIONS(2191),
    [anon_sym_AMP] = ACTIONS(2191),
    [anon_sym_POUND] = ACTIONS(2191),
    [anon_sym_COLON_COLON] = ACTIONS(2193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2193),
    [anon_sym_else] = ACTIONS(2191),
    [anon_sym_new] = ACTIONS(2191),
    [anon_sym_LBRACE_LT] = ACTIONS(2193),
    [anon_sym_begin] = ACTIONS(2191),
    [sym_ocamlyacc_value] = ACTIONS(2193),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2191),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2193),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2191),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2191),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2193),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2193),
    [aux_sym_number_token1] = ACTIONS(2193),
    [aux_sym_signed_number_token1] = ACTIONS(2191),
    [anon_sym_SQUOTE] = ACTIONS(2193),
    [anon_sym_DQUOTE] = ACTIONS(2193),
    [sym_prefix_operator] = ACTIONS(2193),
    [aux_sym_sign_operator_token1] = ACTIONS(2191),
    [sym_hash_operator] = ACTIONS(2193),
    [anon_sym_lsl] = ACTIONS(2191),
    [anon_sym_lsr] = ACTIONS(2191),
    [anon_sym_asr] = ACTIONS(2191),
    [aux_sym_pow_operator_token1] = ACTIONS(2193),
    [anon_sym_mod] = ACTIONS(2191),
    [anon_sym_land] = ACTIONS(2191),
    [anon_sym_lor] = ACTIONS(2191),
    [anon_sym_lxor] = ACTIONS(2191),
    [aux_sym_mult_operator_token1] = ACTIONS(2191),
    [aux_sym_add_operator_token1] = ACTIONS(2191),
    [sym_concat_operator] = ACTIONS(2193),
    [sym_rel_operator] = ACTIONS(2191),
    [anon_sym_AMP_AMP] = ACTIONS(2191),
    [anon_sym_or] = ACTIONS(2191),
    [anon_sym_PIPE_PIPE] = ACTIONS(2191),
    [sym_assign_operator] = ACTIONS(2193),
    [sym_let_operator] = ACTIONS(2193),
    [sym_let_and_operator] = ACTIONS(2193),
    [sym__uppercase_identifier] = ACTIONS(2193),
    [aux_sym_directive_token1] = ACTIONS(2191),
    [aux_sym_tag_token1] = ACTIONS(2193),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1175)] = {
    [sym_attribute] = STATE(1175),
    [sym__lowercase_identifier] = ACTIONS(1955),
    [anon_sym_SEMI_SEMI] = ACTIONS(1957),
    [anon_sym_let] = ACTIONS(1955),
    [anon_sym_and] = ACTIONS(1955),
    [anon_sym_TILDE] = ACTIONS(1955),
    [anon_sym_QMARK] = ACTIONS(1955),
    [anon_sym_LPAREN] = ACTIONS(1957),
    [anon_sym_external] = ACTIONS(1955),
    [anon_sym_type] = ACTIONS(1955),
    [anon_sym_COMMA] = ACTIONS(1957),
    [anon_sym_PIPE] = ACTIONS(1955),
    [anon_sym_DOT] = ACTIONS(1957),
    [anon_sym_LBRACK] = ACTIONS(1955),
    [anon_sym_RBRACK] = ACTIONS(1957),
    [anon_sym_true] = ACTIONS(1955),
    [anon_sym_false] = ACTIONS(1955),
    [anon_sym_LBRACE] = ACTIONS(1955),
    [anon_sym_SEMI] = ACTIONS(1955),
    [anon_sym_exception] = ACTIONS(1955),
    [anon_sym_module] = ACTIONS(1955),
    [anon_sym_open] = ACTIONS(1955),
    [anon_sym_include] = ACTIONS(1955),
    [anon_sym_class] = ACTIONS(1955),
    [anon_sym_val] = ACTIONS(1955),
    [anon_sym_end] = ACTIONS(1955),
    [anon_sym_object] = ACTIONS(1955),
    [anon_sym_in] = ACTIONS(1955),
    [anon_sym_AMP] = ACTIONS(1955),
    [anon_sym_POUND] = ACTIONS(1955),
    [anon_sym_COLON_COLON] = ACTIONS(1957),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1957),
    [anon_sym_else] = ACTIONS(1955),
    [anon_sym_new] = ACTIONS(1955),
    [anon_sym_LBRACE_LT] = ACTIONS(1957),
    [anon_sym_begin] = ACTIONS(1955),
    [sym_ocamlyacc_value] = ACTIONS(1957),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1955),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1957),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1955),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1955),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1957),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1957),
    [aux_sym_number_token1] = ACTIONS(1957),
    [aux_sym_signed_number_token1] = ACTIONS(1955),
    [anon_sym_SQUOTE] = ACTIONS(1957),
    [anon_sym_DQUOTE] = ACTIONS(1957),
    [sym_prefix_operator] = ACTIONS(1957),
    [aux_sym_sign_operator_token1] = ACTIONS(1955),
    [sym_hash_operator] = ACTIONS(1957),
    [anon_sym_lsl] = ACTIONS(1955),
    [anon_sym_lsr] = ACTIONS(1955),
    [anon_sym_asr] = ACTIONS(1955),
    [aux_sym_pow_operator_token1] = ACTIONS(1957),
    [anon_sym_mod] = ACTIONS(1955),
    [anon_sym_land] = ACTIONS(1955),
    [anon_sym_lor] = ACTIONS(1955),
    [anon_sym_lxor] = ACTIONS(1955),
    [aux_sym_mult_operator_token1] = ACTIONS(1955),
    [aux_sym_add_operator_token1] = ACTIONS(1955),
    [sym_concat_operator] = ACTIONS(1957),
    [sym_rel_operator] = ACTIONS(1955),
    [anon_sym_AMP_AMP] = ACTIONS(1955),
    [anon_sym_or] = ACTIONS(1955),
    [anon_sym_PIPE_PIPE] = ACTIONS(1955),
    [sym_assign_operator] = ACTIONS(1957),
    [sym_let_operator] = ACTIONS(1957),
    [sym_let_and_operator] = ACTIONS(1957),
    [sym__uppercase_identifier] = ACTIONS(1957),
    [aux_sym_directive_token1] = ACTIONS(1955),
    [aux_sym_tag_token1] = ACTIONS(1957),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1176)] = {
    [sym_attribute] = STATE(1176),
    [sym__lowercase_identifier] = ACTIONS(1959),
    [anon_sym_SEMI_SEMI] = ACTIONS(1961),
    [anon_sym_let] = ACTIONS(1959),
    [anon_sym_and] = ACTIONS(1959),
    [anon_sym_TILDE] = ACTIONS(1959),
    [anon_sym_QMARK] = ACTIONS(1959),
    [anon_sym_LPAREN] = ACTIONS(1961),
    [anon_sym_external] = ACTIONS(1959),
    [anon_sym_type] = ACTIONS(1959),
    [anon_sym_COMMA] = ACTIONS(1961),
    [anon_sym_PIPE] = ACTIONS(1959),
    [anon_sym_DOT] = ACTIONS(1961),
    [anon_sym_LBRACK] = ACTIONS(1959),
    [anon_sym_RBRACK] = ACTIONS(1961),
    [anon_sym_true] = ACTIONS(1959),
    [anon_sym_false] = ACTIONS(1959),
    [anon_sym_LBRACE] = ACTIONS(1959),
    [anon_sym_SEMI] = ACTIONS(1959),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_module] = ACTIONS(1959),
    [anon_sym_open] = ACTIONS(1959),
    [anon_sym_include] = ACTIONS(1959),
    [anon_sym_class] = ACTIONS(1959),
    [anon_sym_val] = ACTIONS(1959),
    [anon_sym_end] = ACTIONS(1959),
    [anon_sym_object] = ACTIONS(1959),
    [anon_sym_in] = ACTIONS(1959),
    [anon_sym_AMP] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1959),
    [anon_sym_COLON_COLON] = ACTIONS(1961),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1961),
    [anon_sym_else] = ACTIONS(1959),
    [anon_sym_new] = ACTIONS(1959),
    [anon_sym_LBRACE_LT] = ACTIONS(1961),
    [anon_sym_begin] = ACTIONS(1959),
    [sym_ocamlyacc_value] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1959),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1961),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1959),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1959),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1961),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1961),
    [aux_sym_number_token1] = ACTIONS(1961),
    [aux_sym_signed_number_token1] = ACTIONS(1959),
    [anon_sym_SQUOTE] = ACTIONS(1961),
    [anon_sym_DQUOTE] = ACTIONS(1961),
    [sym_prefix_operator] = ACTIONS(1961),
    [aux_sym_sign_operator_token1] = ACTIONS(1959),
    [sym_hash_operator] = ACTIONS(1961),
    [anon_sym_lsl] = ACTIONS(1959),
    [anon_sym_lsr] = ACTIONS(1959),
    [anon_sym_asr] = ACTIONS(1959),
    [aux_sym_pow_operator_token1] = ACTIONS(1961),
    [anon_sym_mod] = ACTIONS(1959),
    [anon_sym_land] = ACTIONS(1959),
    [anon_sym_lor] = ACTIONS(1959),
    [anon_sym_lxor] = ACTIONS(1959),
    [aux_sym_mult_operator_token1] = ACTIONS(1959),
    [aux_sym_add_operator_token1] = ACTIONS(1959),
    [sym_concat_operator] = ACTIONS(1961),
    [sym_rel_operator] = ACTIONS(1959),
    [anon_sym_AMP_AMP] = ACTIONS(1959),
    [anon_sym_or] = ACTIONS(1959),
    [anon_sym_PIPE_PIPE] = ACTIONS(1959),
    [sym_assign_operator] = ACTIONS(1961),
    [sym_let_operator] = ACTIONS(1961),
    [sym_let_and_operator] = ACTIONS(1961),
    [sym__uppercase_identifier] = ACTIONS(1961),
    [aux_sym_directive_token1] = ACTIONS(1959),
    [aux_sym_tag_token1] = ACTIONS(1961),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1177)] = {
    [sym_attribute] = STATE(1177),
    [sym__lowercase_identifier] = ACTIONS(1963),
    [anon_sym_SEMI_SEMI] = ACTIONS(1965),
    [anon_sym_let] = ACTIONS(1963),
    [anon_sym_and] = ACTIONS(1963),
    [anon_sym_TILDE] = ACTIONS(1963),
    [anon_sym_QMARK] = ACTIONS(1963),
    [anon_sym_LPAREN] = ACTIONS(1965),
    [anon_sym_external] = ACTIONS(1963),
    [anon_sym_type] = ACTIONS(1963),
    [anon_sym_COMMA] = ACTIONS(1965),
    [anon_sym_PIPE] = ACTIONS(1963),
    [anon_sym_DOT] = ACTIONS(1965),
    [anon_sym_LBRACK] = ACTIONS(1963),
    [anon_sym_RBRACK] = ACTIONS(1965),
    [anon_sym_true] = ACTIONS(1963),
    [anon_sym_false] = ACTIONS(1963),
    [anon_sym_LBRACE] = ACTIONS(1963),
    [anon_sym_SEMI] = ACTIONS(1963),
    [anon_sym_exception] = ACTIONS(1963),
    [anon_sym_module] = ACTIONS(1963),
    [anon_sym_open] = ACTIONS(1963),
    [anon_sym_include] = ACTIONS(1963),
    [anon_sym_class] = ACTIONS(1963),
    [anon_sym_val] = ACTIONS(1963),
    [anon_sym_end] = ACTIONS(1963),
    [anon_sym_object] = ACTIONS(1963),
    [anon_sym_in] = ACTIONS(1963),
    [anon_sym_AMP] = ACTIONS(1963),
    [anon_sym_POUND] = ACTIONS(1963),
    [anon_sym_COLON_COLON] = ACTIONS(1965),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1965),
    [anon_sym_else] = ACTIONS(1963),
    [anon_sym_new] = ACTIONS(1963),
    [anon_sym_LBRACE_LT] = ACTIONS(1965),
    [anon_sym_begin] = ACTIONS(1963),
    [sym_ocamlyacc_value] = ACTIONS(1965),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1963),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1965),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1965),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1965),
    [aux_sym_signed_number_token1] = ACTIONS(1963),
    [anon_sym_SQUOTE] = ACTIONS(1965),
    [anon_sym_DQUOTE] = ACTIONS(1965),
    [sym_prefix_operator] = ACTIONS(1965),
    [aux_sym_sign_operator_token1] = ACTIONS(1963),
    [sym_hash_operator] = ACTIONS(1965),
    [anon_sym_lsl] = ACTIONS(1963),
    [anon_sym_lsr] = ACTIONS(1963),
    [anon_sym_asr] = ACTIONS(1963),
    [aux_sym_pow_operator_token1] = ACTIONS(1965),
    [anon_sym_mod] = ACTIONS(1963),
    [anon_sym_land] = ACTIONS(1963),
    [anon_sym_lor] = ACTIONS(1963),
    [anon_sym_lxor] = ACTIONS(1963),
    [aux_sym_mult_operator_token1] = ACTIONS(1963),
    [aux_sym_add_operator_token1] = ACTIONS(1963),
    [sym_concat_operator] = ACTIONS(1965),
    [sym_rel_operator] = ACTIONS(1963),
    [anon_sym_AMP_AMP] = ACTIONS(1963),
    [anon_sym_or] = ACTIONS(1963),
    [anon_sym_PIPE_PIPE] = ACTIONS(1963),
    [sym_assign_operator] = ACTIONS(1965),
    [sym_let_operator] = ACTIONS(1965),
    [sym_let_and_operator] = ACTIONS(1965),
    [sym__uppercase_identifier] = ACTIONS(1965),
    [aux_sym_directive_token1] = ACTIONS(1963),
    [aux_sym_tag_token1] = ACTIONS(1965),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1178)] = {
    [sym_attribute] = STATE(1178),
    [sym__lowercase_identifier] = ACTIONS(2179),
    [anon_sym_SEMI_SEMI] = ACTIONS(2181),
    [anon_sym_let] = ACTIONS(2179),
    [anon_sym_and] = ACTIONS(2179),
    [anon_sym_TILDE] = ACTIONS(2179),
    [anon_sym_QMARK] = ACTIONS(2179),
    [anon_sym_LPAREN] = ACTIONS(2181),
    [anon_sym_external] = ACTIONS(2179),
    [anon_sym_type] = ACTIONS(2179),
    [anon_sym_COMMA] = ACTIONS(2181),
    [anon_sym_PIPE] = ACTIONS(2179),
    [anon_sym_DOT] = ACTIONS(2181),
    [anon_sym_LBRACK] = ACTIONS(2179),
    [anon_sym_RBRACK] = ACTIONS(2181),
    [anon_sym_true] = ACTIONS(2179),
    [anon_sym_false] = ACTIONS(2179),
    [anon_sym_LBRACE] = ACTIONS(2179),
    [anon_sym_SEMI] = ACTIONS(2179),
    [anon_sym_exception] = ACTIONS(2179),
    [anon_sym_module] = ACTIONS(2179),
    [anon_sym_open] = ACTIONS(2179),
    [anon_sym_include] = ACTIONS(2179),
    [anon_sym_class] = ACTIONS(2179),
    [anon_sym_val] = ACTIONS(2179),
    [anon_sym_end] = ACTIONS(2179),
    [anon_sym_object] = ACTIONS(2179),
    [anon_sym_in] = ACTIONS(2179),
    [anon_sym_AMP] = ACTIONS(2179),
    [anon_sym_POUND] = ACTIONS(2179),
    [anon_sym_COLON_COLON] = ACTIONS(2181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2181),
    [anon_sym_else] = ACTIONS(2179),
    [anon_sym_new] = ACTIONS(2179),
    [anon_sym_LBRACE_LT] = ACTIONS(2181),
    [anon_sym_begin] = ACTIONS(2179),
    [sym_ocamlyacc_value] = ACTIONS(2181),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2179),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2181),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2179),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2181),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2181),
    [aux_sym_number_token1] = ACTIONS(2181),
    [aux_sym_signed_number_token1] = ACTIONS(2179),
    [anon_sym_SQUOTE] = ACTIONS(2181),
    [anon_sym_DQUOTE] = ACTIONS(2181),
    [sym_prefix_operator] = ACTIONS(2181),
    [aux_sym_sign_operator_token1] = ACTIONS(2179),
    [sym_hash_operator] = ACTIONS(2181),
    [anon_sym_lsl] = ACTIONS(2179),
    [anon_sym_lsr] = ACTIONS(2179),
    [anon_sym_asr] = ACTIONS(2179),
    [aux_sym_pow_operator_token1] = ACTIONS(2181),
    [anon_sym_mod] = ACTIONS(2179),
    [anon_sym_land] = ACTIONS(2179),
    [anon_sym_lor] = ACTIONS(2179),
    [anon_sym_lxor] = ACTIONS(2179),
    [aux_sym_mult_operator_token1] = ACTIONS(2179),
    [aux_sym_add_operator_token1] = ACTIONS(2179),
    [sym_concat_operator] = ACTIONS(2181),
    [sym_rel_operator] = ACTIONS(2179),
    [anon_sym_AMP_AMP] = ACTIONS(2179),
    [anon_sym_or] = ACTIONS(2179),
    [anon_sym_PIPE_PIPE] = ACTIONS(2179),
    [sym_assign_operator] = ACTIONS(2181),
    [sym_let_operator] = ACTIONS(2181),
    [sym_let_and_operator] = ACTIONS(2181),
    [sym__uppercase_identifier] = ACTIONS(2181),
    [aux_sym_directive_token1] = ACTIONS(2179),
    [aux_sym_tag_token1] = ACTIONS(2181),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1179)] = {
    [sym_attribute] = STATE(1179),
    [sym__lowercase_identifier] = ACTIONS(2015),
    [anon_sym_SEMI_SEMI] = ACTIONS(2017),
    [anon_sym_let] = ACTIONS(2015),
    [anon_sym_and] = ACTIONS(2015),
    [anon_sym_TILDE] = ACTIONS(2015),
    [anon_sym_QMARK] = ACTIONS(2015),
    [anon_sym_LPAREN] = ACTIONS(2017),
    [anon_sym_external] = ACTIONS(2015),
    [anon_sym_type] = ACTIONS(2015),
    [anon_sym_COMMA] = ACTIONS(2017),
    [anon_sym_PIPE] = ACTIONS(2015),
    [anon_sym_DOT] = ACTIONS(2017),
    [anon_sym_LBRACK] = ACTIONS(2015),
    [anon_sym_RBRACK] = ACTIONS(2017),
    [anon_sym_true] = ACTIONS(2015),
    [anon_sym_false] = ACTIONS(2015),
    [anon_sym_LBRACE] = ACTIONS(2015),
    [anon_sym_SEMI] = ACTIONS(2015),
    [anon_sym_exception] = ACTIONS(2015),
    [anon_sym_module] = ACTIONS(2015),
    [anon_sym_open] = ACTIONS(2015),
    [anon_sym_include] = ACTIONS(2015),
    [anon_sym_class] = ACTIONS(2015),
    [anon_sym_val] = ACTIONS(2015),
    [anon_sym_end] = ACTIONS(2015),
    [anon_sym_object] = ACTIONS(2015),
    [anon_sym_in] = ACTIONS(2015),
    [anon_sym_AMP] = ACTIONS(2015),
    [anon_sym_POUND] = ACTIONS(2015),
    [anon_sym_COLON_COLON] = ACTIONS(2017),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2017),
    [anon_sym_else] = ACTIONS(2015),
    [anon_sym_new] = ACTIONS(2015),
    [anon_sym_LBRACE_LT] = ACTIONS(2017),
    [anon_sym_begin] = ACTIONS(2015),
    [sym_ocamlyacc_value] = ACTIONS(2017),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2015),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2017),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2015),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2015),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2017),
    [aux_sym_number_token1] = ACTIONS(2017),
    [aux_sym_signed_number_token1] = ACTIONS(2015),
    [anon_sym_SQUOTE] = ACTIONS(2017),
    [anon_sym_DQUOTE] = ACTIONS(2017),
    [sym_prefix_operator] = ACTIONS(2017),
    [aux_sym_sign_operator_token1] = ACTIONS(2015),
    [sym_hash_operator] = ACTIONS(2017),
    [anon_sym_lsl] = ACTIONS(2015),
    [anon_sym_lsr] = ACTIONS(2015),
    [anon_sym_asr] = ACTIONS(2015),
    [aux_sym_pow_operator_token1] = ACTIONS(2017),
    [anon_sym_mod] = ACTIONS(2015),
    [anon_sym_land] = ACTIONS(2015),
    [anon_sym_lor] = ACTIONS(2015),
    [anon_sym_lxor] = ACTIONS(2015),
    [aux_sym_mult_operator_token1] = ACTIONS(2015),
    [aux_sym_add_operator_token1] = ACTIONS(2015),
    [sym_concat_operator] = ACTIONS(2017),
    [sym_rel_operator] = ACTIONS(2015),
    [anon_sym_AMP_AMP] = ACTIONS(2015),
    [anon_sym_or] = ACTIONS(2015),
    [anon_sym_PIPE_PIPE] = ACTIONS(2015),
    [sym_assign_operator] = ACTIONS(2017),
    [sym_let_operator] = ACTIONS(2017),
    [sym_let_and_operator] = ACTIONS(2017),
    [sym__uppercase_identifier] = ACTIONS(2017),
    [aux_sym_directive_token1] = ACTIONS(2015),
    [aux_sym_tag_token1] = ACTIONS(2017),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1180)] = {
    [sym_attribute] = STATE(1180),
    [sym__lowercase_identifier] = ACTIONS(2019),
    [anon_sym_SEMI_SEMI] = ACTIONS(2021),
    [anon_sym_let] = ACTIONS(2019),
    [anon_sym_and] = ACTIONS(2019),
    [anon_sym_TILDE] = ACTIONS(2019),
    [anon_sym_QMARK] = ACTIONS(2019),
    [anon_sym_LPAREN] = ACTIONS(2021),
    [anon_sym_external] = ACTIONS(2019),
    [anon_sym_type] = ACTIONS(2019),
    [anon_sym_COMMA] = ACTIONS(2021),
    [anon_sym_PIPE] = ACTIONS(2019),
    [anon_sym_DOT] = ACTIONS(2021),
    [anon_sym_LBRACK] = ACTIONS(2019),
    [anon_sym_RBRACK] = ACTIONS(2021),
    [anon_sym_true] = ACTIONS(2019),
    [anon_sym_false] = ACTIONS(2019),
    [anon_sym_LBRACE] = ACTIONS(2019),
    [anon_sym_SEMI] = ACTIONS(2019),
    [anon_sym_exception] = ACTIONS(2019),
    [anon_sym_module] = ACTIONS(2019),
    [anon_sym_open] = ACTIONS(2019),
    [anon_sym_include] = ACTIONS(2019),
    [anon_sym_class] = ACTIONS(2019),
    [anon_sym_val] = ACTIONS(2019),
    [anon_sym_end] = ACTIONS(2019),
    [anon_sym_object] = ACTIONS(2019),
    [anon_sym_in] = ACTIONS(2019),
    [anon_sym_AMP] = ACTIONS(2019),
    [anon_sym_POUND] = ACTIONS(2019),
    [anon_sym_COLON_COLON] = ACTIONS(2021),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2021),
    [anon_sym_else] = ACTIONS(2019),
    [anon_sym_new] = ACTIONS(2019),
    [anon_sym_LBRACE_LT] = ACTIONS(2021),
    [anon_sym_begin] = ACTIONS(2019),
    [sym_ocamlyacc_value] = ACTIONS(2021),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2019),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2021),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2021),
    [aux_sym_number_token1] = ACTIONS(2021),
    [aux_sym_signed_number_token1] = ACTIONS(2019),
    [anon_sym_SQUOTE] = ACTIONS(2021),
    [anon_sym_DQUOTE] = ACTIONS(2021),
    [sym_prefix_operator] = ACTIONS(2021),
    [aux_sym_sign_operator_token1] = ACTIONS(2019),
    [sym_hash_operator] = ACTIONS(2021),
    [anon_sym_lsl] = ACTIONS(2019),
    [anon_sym_lsr] = ACTIONS(2019),
    [anon_sym_asr] = ACTIONS(2019),
    [aux_sym_pow_operator_token1] = ACTIONS(2021),
    [anon_sym_mod] = ACTIONS(2019),
    [anon_sym_land] = ACTIONS(2019),
    [anon_sym_lor] = ACTIONS(2019),
    [anon_sym_lxor] = ACTIONS(2019),
    [aux_sym_mult_operator_token1] = ACTIONS(2019),
    [aux_sym_add_operator_token1] = ACTIONS(2019),
    [sym_concat_operator] = ACTIONS(2021),
    [sym_rel_operator] = ACTIONS(2019),
    [anon_sym_AMP_AMP] = ACTIONS(2019),
    [anon_sym_or] = ACTIONS(2019),
    [anon_sym_PIPE_PIPE] = ACTIONS(2019),
    [sym_assign_operator] = ACTIONS(2021),
    [sym_let_operator] = ACTIONS(2021),
    [sym_let_and_operator] = ACTIONS(2021),
    [sym__uppercase_identifier] = ACTIONS(2021),
    [aux_sym_directive_token1] = ACTIONS(2019),
    [aux_sym_tag_token1] = ACTIONS(2021),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1181)] = {
    [sym_attribute] = STATE(1181),
    [sym__lowercase_identifier] = ACTIONS(2099),
    [anon_sym_SEMI_SEMI] = ACTIONS(2101),
    [anon_sym_let] = ACTIONS(2099),
    [anon_sym_and] = ACTIONS(2099),
    [anon_sym_TILDE] = ACTIONS(2099),
    [anon_sym_QMARK] = ACTIONS(2099),
    [anon_sym_LPAREN] = ACTIONS(2101),
    [anon_sym_external] = ACTIONS(2099),
    [anon_sym_type] = ACTIONS(2099),
    [anon_sym_COMMA] = ACTIONS(2101),
    [anon_sym_PIPE] = ACTIONS(2099),
    [anon_sym_DOT] = ACTIONS(2101),
    [anon_sym_LBRACK] = ACTIONS(2099),
    [anon_sym_RBRACK] = ACTIONS(2101),
    [anon_sym_true] = ACTIONS(2099),
    [anon_sym_false] = ACTIONS(2099),
    [anon_sym_LBRACE] = ACTIONS(2099),
    [anon_sym_SEMI] = ACTIONS(2099),
    [anon_sym_exception] = ACTIONS(2099),
    [anon_sym_module] = ACTIONS(2099),
    [anon_sym_open] = ACTIONS(2099),
    [anon_sym_include] = ACTIONS(2099),
    [anon_sym_class] = ACTIONS(2099),
    [anon_sym_val] = ACTIONS(2099),
    [anon_sym_end] = ACTIONS(2099),
    [anon_sym_object] = ACTIONS(2099),
    [anon_sym_in] = ACTIONS(2099),
    [anon_sym_AMP] = ACTIONS(2099),
    [anon_sym_POUND] = ACTIONS(2099),
    [anon_sym_COLON_COLON] = ACTIONS(2101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2101),
    [anon_sym_else] = ACTIONS(2099),
    [anon_sym_new] = ACTIONS(2099),
    [anon_sym_LBRACE_LT] = ACTIONS(2101),
    [anon_sym_begin] = ACTIONS(2099),
    [sym_ocamlyacc_value] = ACTIONS(2101),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2099),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2101),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2099),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2099),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2101),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2101),
    [aux_sym_number_token1] = ACTIONS(2101),
    [aux_sym_signed_number_token1] = ACTIONS(2099),
    [anon_sym_SQUOTE] = ACTIONS(2101),
    [anon_sym_DQUOTE] = ACTIONS(2101),
    [sym_prefix_operator] = ACTIONS(2101),
    [aux_sym_sign_operator_token1] = ACTIONS(2099),
    [sym_hash_operator] = ACTIONS(2101),
    [anon_sym_lsl] = ACTIONS(2099),
    [anon_sym_lsr] = ACTIONS(2099),
    [anon_sym_asr] = ACTIONS(2099),
    [aux_sym_pow_operator_token1] = ACTIONS(2101),
    [anon_sym_mod] = ACTIONS(2099),
    [anon_sym_land] = ACTIONS(2099),
    [anon_sym_lor] = ACTIONS(2099),
    [anon_sym_lxor] = ACTIONS(2099),
    [aux_sym_mult_operator_token1] = ACTIONS(2099),
    [aux_sym_add_operator_token1] = ACTIONS(2099),
    [sym_concat_operator] = ACTIONS(2101),
    [sym_rel_operator] = ACTIONS(2099),
    [anon_sym_AMP_AMP] = ACTIONS(2099),
    [anon_sym_or] = ACTIONS(2099),
    [anon_sym_PIPE_PIPE] = ACTIONS(2099),
    [sym_assign_operator] = ACTIONS(2101),
    [sym_let_operator] = ACTIONS(2101),
    [sym_let_and_operator] = ACTIONS(2101),
    [sym__uppercase_identifier] = ACTIONS(2101),
    [aux_sym_directive_token1] = ACTIONS(2099),
    [aux_sym_tag_token1] = ACTIONS(2101),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1182)] = {
    [sym_attribute] = STATE(1182),
    [sym__lowercase_identifier] = ACTIONS(2051),
    [anon_sym_SEMI_SEMI] = ACTIONS(2053),
    [anon_sym_let] = ACTIONS(2051),
    [anon_sym_and] = ACTIONS(2051),
    [anon_sym_TILDE] = ACTIONS(2051),
    [anon_sym_QMARK] = ACTIONS(2051),
    [anon_sym_LPAREN] = ACTIONS(2053),
    [anon_sym_external] = ACTIONS(2051),
    [anon_sym_type] = ACTIONS(2051),
    [anon_sym_COMMA] = ACTIONS(2053),
    [anon_sym_PIPE] = ACTIONS(2051),
    [anon_sym_DOT] = ACTIONS(2053),
    [anon_sym_LBRACK] = ACTIONS(2051),
    [anon_sym_RBRACK] = ACTIONS(2053),
    [anon_sym_true] = ACTIONS(2051),
    [anon_sym_false] = ACTIONS(2051),
    [anon_sym_LBRACE] = ACTIONS(2051),
    [anon_sym_SEMI] = ACTIONS(2051),
    [anon_sym_exception] = ACTIONS(2051),
    [anon_sym_module] = ACTIONS(2051),
    [anon_sym_open] = ACTIONS(2051),
    [anon_sym_include] = ACTIONS(2051),
    [anon_sym_class] = ACTIONS(2051),
    [anon_sym_val] = ACTIONS(2051),
    [anon_sym_end] = ACTIONS(2051),
    [anon_sym_object] = ACTIONS(2051),
    [anon_sym_in] = ACTIONS(2051),
    [anon_sym_AMP] = ACTIONS(2051),
    [anon_sym_POUND] = ACTIONS(2051),
    [anon_sym_COLON_COLON] = ACTIONS(2053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2053),
    [anon_sym_else] = ACTIONS(2051),
    [anon_sym_new] = ACTIONS(2051),
    [anon_sym_LBRACE_LT] = ACTIONS(2053),
    [anon_sym_begin] = ACTIONS(2051),
    [sym_ocamlyacc_value] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2051),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2053),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2051),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2051),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2053),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2053),
    [aux_sym_number_token1] = ACTIONS(2053),
    [aux_sym_signed_number_token1] = ACTIONS(2051),
    [anon_sym_SQUOTE] = ACTIONS(2053),
    [anon_sym_DQUOTE] = ACTIONS(2053),
    [sym_prefix_operator] = ACTIONS(2053),
    [aux_sym_sign_operator_token1] = ACTIONS(2051),
    [sym_hash_operator] = ACTIONS(2053),
    [anon_sym_lsl] = ACTIONS(2051),
    [anon_sym_lsr] = ACTIONS(2051),
    [anon_sym_asr] = ACTIONS(2051),
    [aux_sym_pow_operator_token1] = ACTIONS(2053),
    [anon_sym_mod] = ACTIONS(2051),
    [anon_sym_land] = ACTIONS(2051),
    [anon_sym_lor] = ACTIONS(2051),
    [anon_sym_lxor] = ACTIONS(2051),
    [aux_sym_mult_operator_token1] = ACTIONS(2051),
    [aux_sym_add_operator_token1] = ACTIONS(2051),
    [sym_concat_operator] = ACTIONS(2053),
    [sym_rel_operator] = ACTIONS(2051),
    [anon_sym_AMP_AMP] = ACTIONS(2051),
    [anon_sym_or] = ACTIONS(2051),
    [anon_sym_PIPE_PIPE] = ACTIONS(2051),
    [sym_assign_operator] = ACTIONS(2053),
    [sym_let_operator] = ACTIONS(2053),
    [sym_let_and_operator] = ACTIONS(2053),
    [sym__uppercase_identifier] = ACTIONS(2053),
    [aux_sym_directive_token1] = ACTIONS(2051),
    [aux_sym_tag_token1] = ACTIONS(2053),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1183)] = {
    [sym_attribute] = STATE(1183),
    [ts_builtin_sym_end] = ACTIONS(1913),
    [sym__lowercase_identifier] = ACTIONS(1911),
    [anon_sym_SEMI_SEMI] = ACTIONS(1913),
    [anon_sym_let] = ACTIONS(1911),
    [anon_sym_and] = ACTIONS(1911),
    [anon_sym_TILDE] = ACTIONS(1911),
    [anon_sym_QMARK] = ACTIONS(1911),
    [anon_sym_LPAREN] = ACTIONS(1913),
    [anon_sym_external] = ACTIONS(1911),
    [anon_sym_type] = ACTIONS(1911),
    [anon_sym_COMMA] = ACTIONS(1913),
    [anon_sym_PIPE] = ACTIONS(1911),
    [anon_sym_DOT] = ACTIONS(1913),
    [anon_sym_LBRACK] = ACTIONS(1911),
    [anon_sym_true] = ACTIONS(1911),
    [anon_sym_false] = ACTIONS(1911),
    [anon_sym_LBRACE] = ACTIONS(1911),
    [anon_sym_SEMI] = ACTIONS(1911),
    [anon_sym_exception] = ACTIONS(1911),
    [anon_sym_module] = ACTIONS(1911),
    [anon_sym_open] = ACTIONS(1911),
    [anon_sym_include] = ACTIONS(1911),
    [anon_sym_class] = ACTIONS(1911),
    [anon_sym_val] = ACTIONS(1911),
    [anon_sym_object] = ACTIONS(1911),
    [anon_sym_in] = ACTIONS(1911),
    [anon_sym_AMP] = ACTIONS(1911),
    [anon_sym_POUND] = ACTIONS(1911),
    [anon_sym_COLON_COLON] = ACTIONS(1913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1913),
    [anon_sym_LT_DASH] = ACTIONS(1911),
    [anon_sym_else] = ACTIONS(1911),
    [anon_sym_new] = ACTIONS(1911),
    [anon_sym_LBRACE_LT] = ACTIONS(1913),
    [anon_sym_begin] = ACTIONS(1911),
    [sym_ocamlyacc_value] = ACTIONS(1913),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1911),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1913),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1911),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1911),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1913),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1913),
    [aux_sym_number_token1] = ACTIONS(1913),
    [aux_sym_signed_number_token1] = ACTIONS(1911),
    [anon_sym_SQUOTE] = ACTIONS(1913),
    [anon_sym_DQUOTE] = ACTIONS(1913),
    [sym_prefix_operator] = ACTIONS(1913),
    [aux_sym_sign_operator_token1] = ACTIONS(1911),
    [sym_hash_operator] = ACTIONS(1913),
    [anon_sym_lsl] = ACTIONS(1911),
    [anon_sym_lsr] = ACTIONS(1911),
    [anon_sym_asr] = ACTIONS(1911),
    [aux_sym_pow_operator_token1] = ACTIONS(1913),
    [anon_sym_mod] = ACTIONS(1911),
    [anon_sym_land] = ACTIONS(1911),
    [anon_sym_lor] = ACTIONS(1911),
    [anon_sym_lxor] = ACTIONS(1911),
    [aux_sym_mult_operator_token1] = ACTIONS(1911),
    [aux_sym_add_operator_token1] = ACTIONS(1911),
    [sym_concat_operator] = ACTIONS(1913),
    [sym_rel_operator] = ACTIONS(1911),
    [anon_sym_AMP_AMP] = ACTIONS(1911),
    [anon_sym_or] = ACTIONS(1911),
    [anon_sym_PIPE_PIPE] = ACTIONS(1911),
    [sym_assign_operator] = ACTIONS(1913),
    [sym_let_operator] = ACTIONS(1913),
    [sym_let_and_operator] = ACTIONS(1913),
    [sym__uppercase_identifier] = ACTIONS(1913),
    [aux_sym_directive_token1] = ACTIONS(1911),
    [aux_sym_tag_token1] = ACTIONS(1913),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1184)] = {
    [sym_attribute] = STATE(1184),
    [ts_builtin_sym_end] = ACTIONS(1917),
    [sym__lowercase_identifier] = ACTIONS(1915),
    [anon_sym_SEMI_SEMI] = ACTIONS(1917),
    [anon_sym_let] = ACTIONS(1915),
    [anon_sym_and] = ACTIONS(1915),
    [anon_sym_TILDE] = ACTIONS(1915),
    [anon_sym_QMARK] = ACTIONS(1915),
    [anon_sym_LPAREN] = ACTIONS(1917),
    [anon_sym_external] = ACTIONS(1915),
    [anon_sym_type] = ACTIONS(1915),
    [anon_sym_COMMA] = ACTIONS(1917),
    [anon_sym_PIPE] = ACTIONS(1915),
    [anon_sym_DOT] = ACTIONS(1917),
    [anon_sym_LBRACK] = ACTIONS(1915),
    [anon_sym_true] = ACTIONS(1915),
    [anon_sym_false] = ACTIONS(1915),
    [anon_sym_LBRACE] = ACTIONS(1915),
    [anon_sym_SEMI] = ACTIONS(1915),
    [anon_sym_exception] = ACTIONS(1915),
    [anon_sym_module] = ACTIONS(1915),
    [anon_sym_open] = ACTIONS(1915),
    [anon_sym_include] = ACTIONS(1915),
    [anon_sym_class] = ACTIONS(1915),
    [anon_sym_val] = ACTIONS(1915),
    [anon_sym_object] = ACTIONS(1915),
    [anon_sym_in] = ACTIONS(1915),
    [anon_sym_AMP] = ACTIONS(1915),
    [anon_sym_POUND] = ACTIONS(1915),
    [anon_sym_COLON_COLON] = ACTIONS(1917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1917),
    [anon_sym_LT_DASH] = ACTIONS(1915),
    [anon_sym_else] = ACTIONS(1915),
    [anon_sym_new] = ACTIONS(1915),
    [anon_sym_LBRACE_LT] = ACTIONS(1917),
    [anon_sym_begin] = ACTIONS(1915),
    [sym_ocamlyacc_value] = ACTIONS(1917),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1915),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1917),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1915),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1915),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1917),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1917),
    [aux_sym_number_token1] = ACTIONS(1917),
    [aux_sym_signed_number_token1] = ACTIONS(1915),
    [anon_sym_SQUOTE] = ACTIONS(1917),
    [anon_sym_DQUOTE] = ACTIONS(1917),
    [sym_prefix_operator] = ACTIONS(1917),
    [aux_sym_sign_operator_token1] = ACTIONS(1915),
    [sym_hash_operator] = ACTIONS(1917),
    [anon_sym_lsl] = ACTIONS(1915),
    [anon_sym_lsr] = ACTIONS(1915),
    [anon_sym_asr] = ACTIONS(1915),
    [aux_sym_pow_operator_token1] = ACTIONS(1917),
    [anon_sym_mod] = ACTIONS(1915),
    [anon_sym_land] = ACTIONS(1915),
    [anon_sym_lor] = ACTIONS(1915),
    [anon_sym_lxor] = ACTIONS(1915),
    [aux_sym_mult_operator_token1] = ACTIONS(1915),
    [aux_sym_add_operator_token1] = ACTIONS(1915),
    [sym_concat_operator] = ACTIONS(1917),
    [sym_rel_operator] = ACTIONS(1915),
    [anon_sym_AMP_AMP] = ACTIONS(1915),
    [anon_sym_or] = ACTIONS(1915),
    [anon_sym_PIPE_PIPE] = ACTIONS(1915),
    [sym_assign_operator] = ACTIONS(1917),
    [sym_let_operator] = ACTIONS(1917),
    [sym_let_and_operator] = ACTIONS(1917),
    [sym__uppercase_identifier] = ACTIONS(1917),
    [aux_sym_directive_token1] = ACTIONS(1915),
    [aux_sym_tag_token1] = ACTIONS(1917),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1185)] = {
    [sym_attribute] = STATE(1185),
    [ts_builtin_sym_end] = ACTIONS(1921),
    [sym__lowercase_identifier] = ACTIONS(1919),
    [anon_sym_SEMI_SEMI] = ACTIONS(1921),
    [anon_sym_let] = ACTIONS(1919),
    [anon_sym_and] = ACTIONS(1919),
    [anon_sym_TILDE] = ACTIONS(1919),
    [anon_sym_QMARK] = ACTIONS(1919),
    [anon_sym_LPAREN] = ACTIONS(1921),
    [anon_sym_external] = ACTIONS(1919),
    [anon_sym_type] = ACTIONS(1919),
    [anon_sym_COMMA] = ACTIONS(1921),
    [anon_sym_PIPE] = ACTIONS(1919),
    [anon_sym_DOT] = ACTIONS(1921),
    [anon_sym_LBRACK] = ACTIONS(1919),
    [anon_sym_true] = ACTIONS(1919),
    [anon_sym_false] = ACTIONS(1919),
    [anon_sym_LBRACE] = ACTIONS(1919),
    [anon_sym_SEMI] = ACTIONS(1919),
    [anon_sym_exception] = ACTIONS(1919),
    [anon_sym_module] = ACTIONS(1919),
    [anon_sym_open] = ACTIONS(1919),
    [anon_sym_include] = ACTIONS(1919),
    [anon_sym_class] = ACTIONS(1919),
    [anon_sym_val] = ACTIONS(1919),
    [anon_sym_object] = ACTIONS(1919),
    [anon_sym_in] = ACTIONS(1919),
    [anon_sym_AMP] = ACTIONS(1919),
    [anon_sym_POUND] = ACTIONS(1919),
    [anon_sym_COLON_COLON] = ACTIONS(1921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1921),
    [anon_sym_LT_DASH] = ACTIONS(1919),
    [anon_sym_else] = ACTIONS(1919),
    [anon_sym_new] = ACTIONS(1919),
    [anon_sym_LBRACE_LT] = ACTIONS(1921),
    [anon_sym_begin] = ACTIONS(1919),
    [sym_ocamlyacc_value] = ACTIONS(1921),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1919),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1921),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1919),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1919),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1921),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1921),
    [aux_sym_number_token1] = ACTIONS(1921),
    [aux_sym_signed_number_token1] = ACTIONS(1919),
    [anon_sym_SQUOTE] = ACTIONS(1921),
    [anon_sym_DQUOTE] = ACTIONS(1921),
    [sym_prefix_operator] = ACTIONS(1921),
    [aux_sym_sign_operator_token1] = ACTIONS(1919),
    [sym_hash_operator] = ACTIONS(1921),
    [anon_sym_lsl] = ACTIONS(1919),
    [anon_sym_lsr] = ACTIONS(1919),
    [anon_sym_asr] = ACTIONS(1919),
    [aux_sym_pow_operator_token1] = ACTIONS(1921),
    [anon_sym_mod] = ACTIONS(1919),
    [anon_sym_land] = ACTIONS(1919),
    [anon_sym_lor] = ACTIONS(1919),
    [anon_sym_lxor] = ACTIONS(1919),
    [aux_sym_mult_operator_token1] = ACTIONS(1919),
    [aux_sym_add_operator_token1] = ACTIONS(1919),
    [sym_concat_operator] = ACTIONS(1921),
    [sym_rel_operator] = ACTIONS(1919),
    [anon_sym_AMP_AMP] = ACTIONS(1919),
    [anon_sym_or] = ACTIONS(1919),
    [anon_sym_PIPE_PIPE] = ACTIONS(1919),
    [sym_assign_operator] = ACTIONS(1921),
    [sym_let_operator] = ACTIONS(1921),
    [sym_let_and_operator] = ACTIONS(1921),
    [sym__uppercase_identifier] = ACTIONS(1921),
    [aux_sym_directive_token1] = ACTIONS(1919),
    [aux_sym_tag_token1] = ACTIONS(1921),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1186)] = {
    [sym_attribute] = STATE(1186),
    [sym__lowercase_identifier] = ACTIONS(2039),
    [anon_sym_SEMI_SEMI] = ACTIONS(2041),
    [anon_sym_let] = ACTIONS(2039),
    [anon_sym_and] = ACTIONS(2039),
    [anon_sym_TILDE] = ACTIONS(2039),
    [anon_sym_QMARK] = ACTIONS(2039),
    [anon_sym_LPAREN] = ACTIONS(2041),
    [anon_sym_external] = ACTIONS(2039),
    [anon_sym_type] = ACTIONS(2039),
    [anon_sym_COMMA] = ACTIONS(2041),
    [anon_sym_PIPE] = ACTIONS(2039),
    [anon_sym_DOT] = ACTIONS(1007),
    [anon_sym_LBRACK] = ACTIONS(2039),
    [anon_sym_RBRACK] = ACTIONS(2041),
    [anon_sym_true] = ACTIONS(2039),
    [anon_sym_false] = ACTIONS(2039),
    [anon_sym_LBRACE] = ACTIONS(2039),
    [anon_sym_SEMI] = ACTIONS(2039),
    [anon_sym_exception] = ACTIONS(2039),
    [anon_sym_module] = ACTIONS(2039),
    [anon_sym_open] = ACTIONS(2039),
    [anon_sym_include] = ACTIONS(2039),
    [anon_sym_class] = ACTIONS(2039),
    [anon_sym_val] = ACTIONS(2039),
    [anon_sym_end] = ACTIONS(2039),
    [anon_sym_object] = ACTIONS(2039),
    [anon_sym_in] = ACTIONS(2039),
    [anon_sym_AMP] = ACTIONS(2039),
    [anon_sym_POUND] = ACTIONS(1009),
    [anon_sym_COLON_COLON] = ACTIONS(2041),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2041),
    [anon_sym_else] = ACTIONS(2039),
    [anon_sym_new] = ACTIONS(2039),
    [anon_sym_LBRACE_LT] = ACTIONS(2041),
    [anon_sym_begin] = ACTIONS(2039),
    [sym_ocamlyacc_value] = ACTIONS(2041),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2039),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2041),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2039),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2039),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2041),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2041),
    [aux_sym_number_token1] = ACTIONS(2041),
    [aux_sym_signed_number_token1] = ACTIONS(2039),
    [anon_sym_SQUOTE] = ACTIONS(2041),
    [anon_sym_DQUOTE] = ACTIONS(2041),
    [sym_prefix_operator] = ACTIONS(2041),
    [aux_sym_sign_operator_token1] = ACTIONS(2039),
    [sym_hash_operator] = ACTIONS(1011),
    [anon_sym_lsl] = ACTIONS(2039),
    [anon_sym_lsr] = ACTIONS(2039),
    [anon_sym_asr] = ACTIONS(2039),
    [aux_sym_pow_operator_token1] = ACTIONS(2041),
    [anon_sym_mod] = ACTIONS(2039),
    [anon_sym_land] = ACTIONS(2039),
    [anon_sym_lor] = ACTIONS(2039),
    [anon_sym_lxor] = ACTIONS(2039),
    [aux_sym_mult_operator_token1] = ACTIONS(2039),
    [aux_sym_add_operator_token1] = ACTIONS(2039),
    [sym_concat_operator] = ACTIONS(2041),
    [sym_rel_operator] = ACTIONS(2039),
    [anon_sym_AMP_AMP] = ACTIONS(2039),
    [anon_sym_or] = ACTIONS(2039),
    [anon_sym_PIPE_PIPE] = ACTIONS(2039),
    [sym_assign_operator] = ACTIONS(2041),
    [sym_let_operator] = ACTIONS(2041),
    [sym_let_and_operator] = ACTIONS(2041),
    [sym__uppercase_identifier] = ACTIONS(2041),
    [aux_sym_directive_token1] = ACTIONS(2039),
    [aux_sym_tag_token1] = ACTIONS(2041),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1187)] = {
    [sym_attribute] = STATE(1187),
    [sym__lowercase_identifier] = ACTIONS(2035),
    [anon_sym_SEMI_SEMI] = ACTIONS(2037),
    [anon_sym_let] = ACTIONS(2035),
    [anon_sym_and] = ACTIONS(2035),
    [anon_sym_TILDE] = ACTIONS(2035),
    [anon_sym_QMARK] = ACTIONS(2035),
    [anon_sym_LPAREN] = ACTIONS(2037),
    [anon_sym_external] = ACTIONS(2035),
    [anon_sym_type] = ACTIONS(2035),
    [anon_sym_COMMA] = ACTIONS(2037),
    [anon_sym_PIPE] = ACTIONS(2035),
    [anon_sym_DOT] = ACTIONS(2037),
    [anon_sym_LBRACK] = ACTIONS(2035),
    [anon_sym_RBRACK] = ACTIONS(2037),
    [anon_sym_true] = ACTIONS(2035),
    [anon_sym_false] = ACTIONS(2035),
    [anon_sym_LBRACE] = ACTIONS(2035),
    [anon_sym_SEMI] = ACTIONS(2035),
    [anon_sym_exception] = ACTIONS(2035),
    [anon_sym_module] = ACTIONS(2035),
    [anon_sym_open] = ACTIONS(2035),
    [anon_sym_include] = ACTIONS(2035),
    [anon_sym_class] = ACTIONS(2035),
    [anon_sym_val] = ACTIONS(2035),
    [anon_sym_end] = ACTIONS(2035),
    [anon_sym_object] = ACTIONS(2035),
    [anon_sym_in] = ACTIONS(2035),
    [anon_sym_AMP] = ACTIONS(2035),
    [anon_sym_POUND] = ACTIONS(2035),
    [anon_sym_COLON_COLON] = ACTIONS(2037),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2037),
    [anon_sym_else] = ACTIONS(2035),
    [anon_sym_new] = ACTIONS(2035),
    [anon_sym_LBRACE_LT] = ACTIONS(2037),
    [anon_sym_begin] = ACTIONS(2035),
    [sym_ocamlyacc_value] = ACTIONS(2037),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2035),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2037),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2035),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2035),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2037),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2037),
    [aux_sym_number_token1] = ACTIONS(2037),
    [aux_sym_signed_number_token1] = ACTIONS(2035),
    [anon_sym_SQUOTE] = ACTIONS(2037),
    [anon_sym_DQUOTE] = ACTIONS(2037),
    [sym_prefix_operator] = ACTIONS(2037),
    [aux_sym_sign_operator_token1] = ACTIONS(2035),
    [sym_hash_operator] = ACTIONS(2037),
    [anon_sym_lsl] = ACTIONS(2035),
    [anon_sym_lsr] = ACTIONS(2035),
    [anon_sym_asr] = ACTIONS(2035),
    [aux_sym_pow_operator_token1] = ACTIONS(2037),
    [anon_sym_mod] = ACTIONS(2035),
    [anon_sym_land] = ACTIONS(2035),
    [anon_sym_lor] = ACTIONS(2035),
    [anon_sym_lxor] = ACTIONS(2035),
    [aux_sym_mult_operator_token1] = ACTIONS(2035),
    [aux_sym_add_operator_token1] = ACTIONS(2035),
    [sym_concat_operator] = ACTIONS(2037),
    [sym_rel_operator] = ACTIONS(2035),
    [anon_sym_AMP_AMP] = ACTIONS(2035),
    [anon_sym_or] = ACTIONS(2035),
    [anon_sym_PIPE_PIPE] = ACTIONS(2035),
    [sym_assign_operator] = ACTIONS(2037),
    [sym_let_operator] = ACTIONS(2037),
    [sym_let_and_operator] = ACTIONS(2037),
    [sym__uppercase_identifier] = ACTIONS(2037),
    [aux_sym_directive_token1] = ACTIONS(2035),
    [aux_sym_tag_token1] = ACTIONS(2037),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1188)] = {
    [sym_attribute] = STATE(1188),
    [sym__lowercase_identifier] = ACTIONS(2075),
    [anon_sym_SEMI_SEMI] = ACTIONS(2077),
    [anon_sym_let] = ACTIONS(2075),
    [anon_sym_and] = ACTIONS(2075),
    [anon_sym_TILDE] = ACTIONS(2075),
    [anon_sym_QMARK] = ACTIONS(2075),
    [anon_sym_LPAREN] = ACTIONS(2077),
    [anon_sym_external] = ACTIONS(2075),
    [anon_sym_type] = ACTIONS(2075),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_PIPE] = ACTIONS(2075),
    [anon_sym_DOT] = ACTIONS(2077),
    [anon_sym_LBRACK] = ACTIONS(2075),
    [anon_sym_RBRACK] = ACTIONS(2077),
    [anon_sym_true] = ACTIONS(2075),
    [anon_sym_false] = ACTIONS(2075),
    [anon_sym_LBRACE] = ACTIONS(2075),
    [anon_sym_SEMI] = ACTIONS(2075),
    [anon_sym_exception] = ACTIONS(2075),
    [anon_sym_module] = ACTIONS(2075),
    [anon_sym_open] = ACTIONS(2075),
    [anon_sym_include] = ACTIONS(2075),
    [anon_sym_class] = ACTIONS(2075),
    [anon_sym_val] = ACTIONS(2075),
    [anon_sym_end] = ACTIONS(2075),
    [anon_sym_object] = ACTIONS(2075),
    [anon_sym_in] = ACTIONS(2075),
    [anon_sym_AMP] = ACTIONS(2075),
    [anon_sym_POUND] = ACTIONS(2075),
    [anon_sym_COLON_COLON] = ACTIONS(2077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2077),
    [anon_sym_else] = ACTIONS(2075),
    [anon_sym_new] = ACTIONS(2075),
    [anon_sym_LBRACE_LT] = ACTIONS(2077),
    [anon_sym_begin] = ACTIONS(2075),
    [sym_ocamlyacc_value] = ACTIONS(2077),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2075),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2077),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2075),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2075),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2077),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2077),
    [aux_sym_number_token1] = ACTIONS(2077),
    [aux_sym_signed_number_token1] = ACTIONS(2075),
    [anon_sym_SQUOTE] = ACTIONS(2077),
    [anon_sym_DQUOTE] = ACTIONS(2077),
    [sym_prefix_operator] = ACTIONS(2077),
    [aux_sym_sign_operator_token1] = ACTIONS(2075),
    [sym_hash_operator] = ACTIONS(2077),
    [anon_sym_lsl] = ACTIONS(2075),
    [anon_sym_lsr] = ACTIONS(2075),
    [anon_sym_asr] = ACTIONS(2075),
    [aux_sym_pow_operator_token1] = ACTIONS(2077),
    [anon_sym_mod] = ACTIONS(2075),
    [anon_sym_land] = ACTIONS(2075),
    [anon_sym_lor] = ACTIONS(2075),
    [anon_sym_lxor] = ACTIONS(2075),
    [aux_sym_mult_operator_token1] = ACTIONS(2075),
    [aux_sym_add_operator_token1] = ACTIONS(2075),
    [sym_concat_operator] = ACTIONS(2077),
    [sym_rel_operator] = ACTIONS(2075),
    [anon_sym_AMP_AMP] = ACTIONS(2075),
    [anon_sym_or] = ACTIONS(2075),
    [anon_sym_PIPE_PIPE] = ACTIONS(2075),
    [sym_assign_operator] = ACTIONS(2077),
    [sym_let_operator] = ACTIONS(2077),
    [sym_let_and_operator] = ACTIONS(2077),
    [sym__uppercase_identifier] = ACTIONS(2077),
    [aux_sym_directive_token1] = ACTIONS(2075),
    [aux_sym_tag_token1] = ACTIONS(2077),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1189)] = {
    [sym_attribute] = STATE(1189),
    [sym__lowercase_identifier] = ACTIONS(1947),
    [anon_sym_SEMI_SEMI] = ACTIONS(1949),
    [anon_sym_let] = ACTIONS(1947),
    [anon_sym_and] = ACTIONS(1947),
    [anon_sym_TILDE] = ACTIONS(1947),
    [anon_sym_QMARK] = ACTIONS(1947),
    [anon_sym_LPAREN] = ACTIONS(1949),
    [anon_sym_external] = ACTIONS(1947),
    [anon_sym_type] = ACTIONS(1947),
    [anon_sym_COMMA] = ACTIONS(1949),
    [anon_sym_PIPE] = ACTIONS(1947),
    [anon_sym_DOT] = ACTIONS(1949),
    [anon_sym_LBRACK] = ACTIONS(1947),
    [anon_sym_RBRACK] = ACTIONS(1949),
    [anon_sym_true] = ACTIONS(1947),
    [anon_sym_false] = ACTIONS(1947),
    [anon_sym_LBRACE] = ACTIONS(1947),
    [anon_sym_SEMI] = ACTIONS(1947),
    [anon_sym_exception] = ACTIONS(1947),
    [anon_sym_module] = ACTIONS(1947),
    [anon_sym_open] = ACTIONS(1947),
    [anon_sym_include] = ACTIONS(1947),
    [anon_sym_class] = ACTIONS(1947),
    [anon_sym_val] = ACTIONS(1947),
    [anon_sym_end] = ACTIONS(1947),
    [anon_sym_object] = ACTIONS(1947),
    [anon_sym_in] = ACTIONS(1947),
    [anon_sym_AMP] = ACTIONS(1947),
    [anon_sym_POUND] = ACTIONS(1947),
    [anon_sym_COLON_COLON] = ACTIONS(1949),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1949),
    [anon_sym_else] = ACTIONS(1947),
    [anon_sym_new] = ACTIONS(1947),
    [anon_sym_LBRACE_LT] = ACTIONS(1949),
    [anon_sym_begin] = ACTIONS(1947),
    [sym_ocamlyacc_value] = ACTIONS(1949),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1947),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1949),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1949),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1949),
    [aux_sym_number_token1] = ACTIONS(1949),
    [aux_sym_signed_number_token1] = ACTIONS(1947),
    [anon_sym_SQUOTE] = ACTIONS(1949),
    [anon_sym_DQUOTE] = ACTIONS(1949),
    [sym_prefix_operator] = ACTIONS(1949),
    [aux_sym_sign_operator_token1] = ACTIONS(1947),
    [sym_hash_operator] = ACTIONS(1949),
    [anon_sym_lsl] = ACTIONS(1947),
    [anon_sym_lsr] = ACTIONS(1947),
    [anon_sym_asr] = ACTIONS(1947),
    [aux_sym_pow_operator_token1] = ACTIONS(1949),
    [anon_sym_mod] = ACTIONS(1947),
    [anon_sym_land] = ACTIONS(1947),
    [anon_sym_lor] = ACTIONS(1947),
    [anon_sym_lxor] = ACTIONS(1947),
    [aux_sym_mult_operator_token1] = ACTIONS(1947),
    [aux_sym_add_operator_token1] = ACTIONS(1947),
    [sym_concat_operator] = ACTIONS(1949),
    [sym_rel_operator] = ACTIONS(1947),
    [anon_sym_AMP_AMP] = ACTIONS(1947),
    [anon_sym_or] = ACTIONS(1947),
    [anon_sym_PIPE_PIPE] = ACTIONS(1947),
    [sym_assign_operator] = ACTIONS(1949),
    [sym_let_operator] = ACTIONS(1949),
    [sym_let_and_operator] = ACTIONS(1949),
    [sym__uppercase_identifier] = ACTIONS(1949),
    [aux_sym_directive_token1] = ACTIONS(1947),
    [aux_sym_tag_token1] = ACTIONS(1949),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1190)] = {
    [sym_attribute] = STATE(1190),
    [sym__lowercase_identifier] = ACTIONS(1995),
    [anon_sym_SEMI_SEMI] = ACTIONS(1997),
    [anon_sym_let] = ACTIONS(1995),
    [anon_sym_and] = ACTIONS(1995),
    [anon_sym_TILDE] = ACTIONS(1995),
    [anon_sym_QMARK] = ACTIONS(1995),
    [anon_sym_LPAREN] = ACTIONS(1997),
    [anon_sym_external] = ACTIONS(1995),
    [anon_sym_type] = ACTIONS(1995),
    [anon_sym_COMMA] = ACTIONS(1997),
    [anon_sym_PIPE] = ACTIONS(1995),
    [anon_sym_DOT] = ACTIONS(1997),
    [anon_sym_LBRACK] = ACTIONS(1995),
    [anon_sym_RBRACK] = ACTIONS(1997),
    [anon_sym_true] = ACTIONS(1995),
    [anon_sym_false] = ACTIONS(1995),
    [anon_sym_LBRACE] = ACTIONS(1995),
    [anon_sym_SEMI] = ACTIONS(1995),
    [anon_sym_exception] = ACTIONS(1995),
    [anon_sym_module] = ACTIONS(1995),
    [anon_sym_open] = ACTIONS(1995),
    [anon_sym_include] = ACTIONS(1995),
    [anon_sym_class] = ACTIONS(1995),
    [anon_sym_val] = ACTIONS(1995),
    [anon_sym_end] = ACTIONS(1995),
    [anon_sym_object] = ACTIONS(1995),
    [anon_sym_in] = ACTIONS(1995),
    [anon_sym_AMP] = ACTIONS(1995),
    [anon_sym_POUND] = ACTIONS(1995),
    [anon_sym_COLON_COLON] = ACTIONS(1997),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1997),
    [anon_sym_else] = ACTIONS(1995),
    [anon_sym_new] = ACTIONS(1995),
    [anon_sym_LBRACE_LT] = ACTIONS(1997),
    [anon_sym_begin] = ACTIONS(1995),
    [sym_ocamlyacc_value] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1995),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1997),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1995),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1995),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1997),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1997),
    [aux_sym_number_token1] = ACTIONS(1997),
    [aux_sym_signed_number_token1] = ACTIONS(1995),
    [anon_sym_SQUOTE] = ACTIONS(1997),
    [anon_sym_DQUOTE] = ACTIONS(1997),
    [sym_prefix_operator] = ACTIONS(1997),
    [aux_sym_sign_operator_token1] = ACTIONS(1995),
    [sym_hash_operator] = ACTIONS(1997),
    [anon_sym_lsl] = ACTIONS(1995),
    [anon_sym_lsr] = ACTIONS(1995),
    [anon_sym_asr] = ACTIONS(1995),
    [aux_sym_pow_operator_token1] = ACTIONS(1997),
    [anon_sym_mod] = ACTIONS(1995),
    [anon_sym_land] = ACTIONS(1995),
    [anon_sym_lor] = ACTIONS(1995),
    [anon_sym_lxor] = ACTIONS(1995),
    [aux_sym_mult_operator_token1] = ACTIONS(1995),
    [aux_sym_add_operator_token1] = ACTIONS(1995),
    [sym_concat_operator] = ACTIONS(1997),
    [sym_rel_operator] = ACTIONS(1995),
    [anon_sym_AMP_AMP] = ACTIONS(1995),
    [anon_sym_or] = ACTIONS(1995),
    [anon_sym_PIPE_PIPE] = ACTIONS(1995),
    [sym_assign_operator] = ACTIONS(1997),
    [sym_let_operator] = ACTIONS(1997),
    [sym_let_and_operator] = ACTIONS(1997),
    [sym__uppercase_identifier] = ACTIONS(1997),
    [aux_sym_directive_token1] = ACTIONS(1995),
    [aux_sym_tag_token1] = ACTIONS(1997),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1191)] = {
    [sym_attribute] = STATE(1191),
    [sym__lowercase_identifier] = ACTIONS(2043),
    [anon_sym_SEMI_SEMI] = ACTIONS(2045),
    [anon_sym_let] = ACTIONS(2043),
    [anon_sym_and] = ACTIONS(2043),
    [anon_sym_TILDE] = ACTIONS(2043),
    [anon_sym_QMARK] = ACTIONS(2043),
    [anon_sym_LPAREN] = ACTIONS(2045),
    [anon_sym_external] = ACTIONS(2043),
    [anon_sym_type] = ACTIONS(2043),
    [anon_sym_COMMA] = ACTIONS(2045),
    [anon_sym_PIPE] = ACTIONS(2043),
    [anon_sym_DOT] = ACTIONS(2045),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_RBRACK] = ACTIONS(2045),
    [anon_sym_true] = ACTIONS(2043),
    [anon_sym_false] = ACTIONS(2043),
    [anon_sym_LBRACE] = ACTIONS(2043),
    [anon_sym_SEMI] = ACTIONS(2043),
    [anon_sym_exception] = ACTIONS(2043),
    [anon_sym_module] = ACTIONS(2043),
    [anon_sym_open] = ACTIONS(2043),
    [anon_sym_include] = ACTIONS(2043),
    [anon_sym_class] = ACTIONS(2043),
    [anon_sym_val] = ACTIONS(2043),
    [anon_sym_end] = ACTIONS(2043),
    [anon_sym_object] = ACTIONS(2043),
    [anon_sym_in] = ACTIONS(2043),
    [anon_sym_AMP] = ACTIONS(2043),
    [anon_sym_POUND] = ACTIONS(2043),
    [anon_sym_COLON_COLON] = ACTIONS(2045),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2045),
    [anon_sym_else] = ACTIONS(2043),
    [anon_sym_new] = ACTIONS(2043),
    [anon_sym_LBRACE_LT] = ACTIONS(2045),
    [anon_sym_begin] = ACTIONS(2043),
    [sym_ocamlyacc_value] = ACTIONS(2045),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2043),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2045),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2043),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2043),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2045),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2045),
    [aux_sym_number_token1] = ACTIONS(2045),
    [aux_sym_signed_number_token1] = ACTIONS(2043),
    [anon_sym_SQUOTE] = ACTIONS(2045),
    [anon_sym_DQUOTE] = ACTIONS(2045),
    [sym_prefix_operator] = ACTIONS(2045),
    [aux_sym_sign_operator_token1] = ACTIONS(2043),
    [sym_hash_operator] = ACTIONS(2045),
    [anon_sym_lsl] = ACTIONS(2043),
    [anon_sym_lsr] = ACTIONS(2043),
    [anon_sym_asr] = ACTIONS(2043),
    [aux_sym_pow_operator_token1] = ACTIONS(2045),
    [anon_sym_mod] = ACTIONS(2043),
    [anon_sym_land] = ACTIONS(2043),
    [anon_sym_lor] = ACTIONS(2043),
    [anon_sym_lxor] = ACTIONS(2043),
    [aux_sym_mult_operator_token1] = ACTIONS(2043),
    [aux_sym_add_operator_token1] = ACTIONS(2043),
    [sym_concat_operator] = ACTIONS(2045),
    [sym_rel_operator] = ACTIONS(2043),
    [anon_sym_AMP_AMP] = ACTIONS(2043),
    [anon_sym_or] = ACTIONS(2043),
    [anon_sym_PIPE_PIPE] = ACTIONS(2043),
    [sym_assign_operator] = ACTIONS(2045),
    [sym_let_operator] = ACTIONS(2045),
    [sym_let_and_operator] = ACTIONS(2045),
    [sym__uppercase_identifier] = ACTIONS(2045),
    [aux_sym_directive_token1] = ACTIONS(2043),
    [aux_sym_tag_token1] = ACTIONS(2045),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1192)] = {
    [sym_attribute] = STATE(1192),
    [sym__lowercase_identifier] = ACTIONS(2023),
    [anon_sym_SEMI_SEMI] = ACTIONS(2025),
    [anon_sym_let] = ACTIONS(2023),
    [anon_sym_and] = ACTIONS(2023),
    [anon_sym_TILDE] = ACTIONS(2023),
    [anon_sym_QMARK] = ACTIONS(2023),
    [anon_sym_LPAREN] = ACTIONS(2025),
    [anon_sym_external] = ACTIONS(2023),
    [anon_sym_type] = ACTIONS(2023),
    [anon_sym_COMMA] = ACTIONS(2025),
    [anon_sym_PIPE] = ACTIONS(2023),
    [anon_sym_DOT] = ACTIONS(2025),
    [anon_sym_LBRACK] = ACTIONS(2023),
    [anon_sym_RBRACK] = ACTIONS(2025),
    [anon_sym_true] = ACTIONS(2023),
    [anon_sym_false] = ACTIONS(2023),
    [anon_sym_LBRACE] = ACTIONS(2023),
    [anon_sym_SEMI] = ACTIONS(2023),
    [anon_sym_exception] = ACTIONS(2023),
    [anon_sym_module] = ACTIONS(2023),
    [anon_sym_open] = ACTIONS(2023),
    [anon_sym_include] = ACTIONS(2023),
    [anon_sym_class] = ACTIONS(2023),
    [anon_sym_val] = ACTIONS(2023),
    [anon_sym_end] = ACTIONS(2023),
    [anon_sym_object] = ACTIONS(2023),
    [anon_sym_in] = ACTIONS(2023),
    [anon_sym_AMP] = ACTIONS(2023),
    [anon_sym_POUND] = ACTIONS(2023),
    [anon_sym_COLON_COLON] = ACTIONS(2025),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2025),
    [anon_sym_else] = ACTIONS(2023),
    [anon_sym_new] = ACTIONS(2023),
    [anon_sym_LBRACE_LT] = ACTIONS(2025),
    [anon_sym_begin] = ACTIONS(2023),
    [sym_ocamlyacc_value] = ACTIONS(2025),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2023),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2025),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2025),
    [aux_sym_number_token1] = ACTIONS(2025),
    [aux_sym_signed_number_token1] = ACTIONS(2023),
    [anon_sym_SQUOTE] = ACTIONS(2025),
    [anon_sym_DQUOTE] = ACTIONS(2025),
    [sym_prefix_operator] = ACTIONS(2025),
    [aux_sym_sign_operator_token1] = ACTIONS(2023),
    [sym_hash_operator] = ACTIONS(2025),
    [anon_sym_lsl] = ACTIONS(2023),
    [anon_sym_lsr] = ACTIONS(2023),
    [anon_sym_asr] = ACTIONS(2023),
    [aux_sym_pow_operator_token1] = ACTIONS(2025),
    [anon_sym_mod] = ACTIONS(2023),
    [anon_sym_land] = ACTIONS(2023),
    [anon_sym_lor] = ACTIONS(2023),
    [anon_sym_lxor] = ACTIONS(2023),
    [aux_sym_mult_operator_token1] = ACTIONS(2023),
    [aux_sym_add_operator_token1] = ACTIONS(2023),
    [sym_concat_operator] = ACTIONS(2025),
    [sym_rel_operator] = ACTIONS(2023),
    [anon_sym_AMP_AMP] = ACTIONS(2023),
    [anon_sym_or] = ACTIONS(2023),
    [anon_sym_PIPE_PIPE] = ACTIONS(2023),
    [sym_assign_operator] = ACTIONS(2025),
    [sym_let_operator] = ACTIONS(2025),
    [sym_let_and_operator] = ACTIONS(2025),
    [sym__uppercase_identifier] = ACTIONS(2025),
    [aux_sym_directive_token1] = ACTIONS(2023),
    [aux_sym_tag_token1] = ACTIONS(2025),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1193)] = {
    [sym_parameter] = STATE(2862),
    [sym__class_typed] = STATE(4647),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1193),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(7381),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(3191),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(4347),
    [aux_sym_let_binding_repeat1] = STATE(1210),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(2425),
    [anon_sym_and] = ACTIONS(2427),
    [anon_sym_EQ] = ACTIONS(2522),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_QMARK] = ACTIONS(2524),
    [anon_sym_LPAREN] = ACTIONS(2526),
    [anon_sym_external] = ACTIONS(2427),
    [anon_sym_type] = ACTIONS(2427),
    [anon_sym_COLON2] = ACTIONS(2528),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_RBRACK] = ACTIONS(2425),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_exception] = ACTIONS(2427),
    [anon_sym_module] = ACTIONS(2427),
    [anon_sym_open] = ACTIONS(2427),
    [anon_sym_include] = ACTIONS(2427),
    [anon_sym_class] = ACTIONS(2427),
    [anon_sym_val] = ACTIONS(2427),
    [anon_sym_end] = ACTIONS(2427),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2425),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2425),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2425),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(2530),
    [aux_sym_tag_token1] = ACTIONS(2532),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1194)] = {
    [sym_attribute] = STATE(1194),
    [sym__lowercase_identifier] = ACTIONS(1951),
    [anon_sym_SEMI_SEMI] = ACTIONS(1953),
    [anon_sym_let] = ACTIONS(1951),
    [anon_sym_and] = ACTIONS(1951),
    [anon_sym_TILDE] = ACTIONS(1951),
    [anon_sym_QMARK] = ACTIONS(1951),
    [anon_sym_LPAREN] = ACTIONS(1953),
    [anon_sym_external] = ACTIONS(1951),
    [anon_sym_type] = ACTIONS(1951),
    [anon_sym_COMMA] = ACTIONS(1953),
    [anon_sym_PIPE] = ACTIONS(1951),
    [anon_sym_DOT] = ACTIONS(1953),
    [anon_sym_LBRACK] = ACTIONS(1951),
    [anon_sym_RBRACK] = ACTIONS(1953),
    [anon_sym_true] = ACTIONS(1951),
    [anon_sym_false] = ACTIONS(1951),
    [anon_sym_LBRACE] = ACTIONS(1951),
    [anon_sym_SEMI] = ACTIONS(1951),
    [anon_sym_exception] = ACTIONS(1951),
    [anon_sym_module] = ACTIONS(1951),
    [anon_sym_open] = ACTIONS(1951),
    [anon_sym_include] = ACTIONS(1951),
    [anon_sym_class] = ACTIONS(1951),
    [anon_sym_val] = ACTIONS(1951),
    [anon_sym_end] = ACTIONS(1951),
    [anon_sym_object] = ACTIONS(1951),
    [anon_sym_in] = ACTIONS(1951),
    [anon_sym_AMP] = ACTIONS(1951),
    [anon_sym_POUND] = ACTIONS(1951),
    [anon_sym_COLON_COLON] = ACTIONS(1953),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1953),
    [anon_sym_else] = ACTIONS(1951),
    [anon_sym_new] = ACTIONS(1951),
    [anon_sym_LBRACE_LT] = ACTIONS(1953),
    [anon_sym_begin] = ACTIONS(1951),
    [sym_ocamlyacc_value] = ACTIONS(1953),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1951),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1953),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1951),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1951),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1953),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1953),
    [aux_sym_number_token1] = ACTIONS(1953),
    [aux_sym_signed_number_token1] = ACTIONS(1951),
    [anon_sym_SQUOTE] = ACTIONS(1953),
    [anon_sym_DQUOTE] = ACTIONS(1953),
    [sym_prefix_operator] = ACTIONS(1953),
    [aux_sym_sign_operator_token1] = ACTIONS(1951),
    [sym_hash_operator] = ACTIONS(1953),
    [anon_sym_lsl] = ACTIONS(1951),
    [anon_sym_lsr] = ACTIONS(1951),
    [anon_sym_asr] = ACTIONS(1951),
    [aux_sym_pow_operator_token1] = ACTIONS(1953),
    [anon_sym_mod] = ACTIONS(1951),
    [anon_sym_land] = ACTIONS(1951),
    [anon_sym_lor] = ACTIONS(1951),
    [anon_sym_lxor] = ACTIONS(1951),
    [aux_sym_mult_operator_token1] = ACTIONS(1951),
    [aux_sym_add_operator_token1] = ACTIONS(1951),
    [sym_concat_operator] = ACTIONS(1953),
    [sym_rel_operator] = ACTIONS(1951),
    [anon_sym_AMP_AMP] = ACTIONS(1951),
    [anon_sym_or] = ACTIONS(1951),
    [anon_sym_PIPE_PIPE] = ACTIONS(1951),
    [sym_assign_operator] = ACTIONS(1953),
    [sym_let_operator] = ACTIONS(1953),
    [sym_let_and_operator] = ACTIONS(1953),
    [sym__uppercase_identifier] = ACTIONS(1953),
    [aux_sym_directive_token1] = ACTIONS(1951),
    [aux_sym_tag_token1] = ACTIONS(1953),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1195)] = {
    [sym_attribute] = STATE(1195),
    [sym__lowercase_identifier] = ACTIONS(1931),
    [anon_sym_SEMI_SEMI] = ACTIONS(1933),
    [anon_sym_let] = ACTIONS(1931),
    [anon_sym_and] = ACTIONS(1931),
    [anon_sym_TILDE] = ACTIONS(1931),
    [anon_sym_QMARK] = ACTIONS(1931),
    [anon_sym_LPAREN] = ACTIONS(1933),
    [anon_sym_external] = ACTIONS(1931),
    [anon_sym_type] = ACTIONS(1931),
    [anon_sym_COMMA] = ACTIONS(1933),
    [anon_sym_PIPE] = ACTIONS(1931),
    [anon_sym_DOT] = ACTIONS(1007),
    [anon_sym_LBRACK] = ACTIONS(1931),
    [anon_sym_RBRACK] = ACTIONS(1933),
    [anon_sym_true] = ACTIONS(1931),
    [anon_sym_false] = ACTIONS(1931),
    [anon_sym_LBRACE] = ACTIONS(1931),
    [anon_sym_SEMI] = ACTIONS(1931),
    [anon_sym_exception] = ACTIONS(1931),
    [anon_sym_module] = ACTIONS(1931),
    [anon_sym_open] = ACTIONS(1931),
    [anon_sym_include] = ACTIONS(1931),
    [anon_sym_class] = ACTIONS(1931),
    [anon_sym_val] = ACTIONS(1931),
    [anon_sym_end] = ACTIONS(1931),
    [anon_sym_object] = ACTIONS(1931),
    [anon_sym_in] = ACTIONS(1931),
    [anon_sym_AMP] = ACTIONS(1931),
    [anon_sym_POUND] = ACTIONS(1009),
    [anon_sym_COLON_COLON] = ACTIONS(1933),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1933),
    [anon_sym_else] = ACTIONS(1931),
    [anon_sym_new] = ACTIONS(1931),
    [anon_sym_LBRACE_LT] = ACTIONS(1933),
    [anon_sym_begin] = ACTIONS(1931),
    [sym_ocamlyacc_value] = ACTIONS(1933),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1931),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1933),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1933),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1933),
    [aux_sym_number_token1] = ACTIONS(1933),
    [aux_sym_signed_number_token1] = ACTIONS(1931),
    [anon_sym_SQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1933),
    [sym_prefix_operator] = ACTIONS(1933),
    [aux_sym_sign_operator_token1] = ACTIONS(1931),
    [sym_hash_operator] = ACTIONS(1011),
    [anon_sym_lsl] = ACTIONS(1931),
    [anon_sym_lsr] = ACTIONS(1931),
    [anon_sym_asr] = ACTIONS(1931),
    [aux_sym_pow_operator_token1] = ACTIONS(1933),
    [anon_sym_mod] = ACTIONS(1931),
    [anon_sym_land] = ACTIONS(1931),
    [anon_sym_lor] = ACTIONS(1931),
    [anon_sym_lxor] = ACTIONS(1931),
    [aux_sym_mult_operator_token1] = ACTIONS(1931),
    [aux_sym_add_operator_token1] = ACTIONS(1931),
    [sym_concat_operator] = ACTIONS(1933),
    [sym_rel_operator] = ACTIONS(1931),
    [anon_sym_AMP_AMP] = ACTIONS(1931),
    [anon_sym_or] = ACTIONS(1931),
    [anon_sym_PIPE_PIPE] = ACTIONS(1931),
    [sym_assign_operator] = ACTIONS(1933),
    [sym_let_operator] = ACTIONS(1933),
    [sym_let_and_operator] = ACTIONS(1933),
    [sym__uppercase_identifier] = ACTIONS(1933),
    [aux_sym_directive_token1] = ACTIONS(1931),
    [aux_sym_tag_token1] = ACTIONS(1933),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1196)] = {
    [sym_attribute] = STATE(1196),
    [sym__lowercase_identifier] = ACTIONS(2071),
    [anon_sym_SEMI_SEMI] = ACTIONS(2073),
    [anon_sym_let] = ACTIONS(2071),
    [anon_sym_and] = ACTIONS(2071),
    [anon_sym_TILDE] = ACTIONS(2071),
    [anon_sym_QMARK] = ACTIONS(2071),
    [anon_sym_LPAREN] = ACTIONS(2073),
    [anon_sym_external] = ACTIONS(2071),
    [anon_sym_type] = ACTIONS(2071),
    [anon_sym_COMMA] = ACTIONS(2073),
    [anon_sym_PIPE] = ACTIONS(2071),
    [anon_sym_DOT] = ACTIONS(2073),
    [anon_sym_LBRACK] = ACTIONS(2071),
    [anon_sym_RBRACK] = ACTIONS(2073),
    [anon_sym_true] = ACTIONS(2071),
    [anon_sym_false] = ACTIONS(2071),
    [anon_sym_LBRACE] = ACTIONS(2071),
    [anon_sym_SEMI] = ACTIONS(2071),
    [anon_sym_exception] = ACTIONS(2071),
    [anon_sym_module] = ACTIONS(2071),
    [anon_sym_open] = ACTIONS(2071),
    [anon_sym_include] = ACTIONS(2071),
    [anon_sym_class] = ACTIONS(2071),
    [anon_sym_val] = ACTIONS(2071),
    [anon_sym_end] = ACTIONS(2071),
    [anon_sym_object] = ACTIONS(2071),
    [anon_sym_in] = ACTIONS(2071),
    [anon_sym_AMP] = ACTIONS(2071),
    [anon_sym_POUND] = ACTIONS(2071),
    [anon_sym_COLON_COLON] = ACTIONS(2073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2073),
    [anon_sym_else] = ACTIONS(2071),
    [anon_sym_new] = ACTIONS(2071),
    [anon_sym_LBRACE_LT] = ACTIONS(2073),
    [anon_sym_begin] = ACTIONS(2071),
    [sym_ocamlyacc_value] = ACTIONS(2073),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2071),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2073),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2071),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2071),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2073),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2073),
    [aux_sym_number_token1] = ACTIONS(2073),
    [aux_sym_signed_number_token1] = ACTIONS(2071),
    [anon_sym_SQUOTE] = ACTIONS(2073),
    [anon_sym_DQUOTE] = ACTIONS(2073),
    [sym_prefix_operator] = ACTIONS(2073),
    [aux_sym_sign_operator_token1] = ACTIONS(2071),
    [sym_hash_operator] = ACTIONS(2073),
    [anon_sym_lsl] = ACTIONS(2071),
    [anon_sym_lsr] = ACTIONS(2071),
    [anon_sym_asr] = ACTIONS(2071),
    [aux_sym_pow_operator_token1] = ACTIONS(2073),
    [anon_sym_mod] = ACTIONS(2071),
    [anon_sym_land] = ACTIONS(2071),
    [anon_sym_lor] = ACTIONS(2071),
    [anon_sym_lxor] = ACTIONS(2071),
    [aux_sym_mult_operator_token1] = ACTIONS(2071),
    [aux_sym_add_operator_token1] = ACTIONS(2071),
    [sym_concat_operator] = ACTIONS(2073),
    [sym_rel_operator] = ACTIONS(2071),
    [anon_sym_AMP_AMP] = ACTIONS(2071),
    [anon_sym_or] = ACTIONS(2071),
    [anon_sym_PIPE_PIPE] = ACTIONS(2071),
    [sym_assign_operator] = ACTIONS(2073),
    [sym_let_operator] = ACTIONS(2073),
    [sym_let_and_operator] = ACTIONS(2073),
    [sym__uppercase_identifier] = ACTIONS(2073),
    [aux_sym_directive_token1] = ACTIONS(2071),
    [aux_sym_tag_token1] = ACTIONS(2073),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1197)] = {
    [sym_attribute] = STATE(1197),
    [sym__lowercase_identifier] = ACTIONS(1967),
    [anon_sym_SEMI_SEMI] = ACTIONS(1969),
    [anon_sym_let] = ACTIONS(1967),
    [anon_sym_and] = ACTIONS(1967),
    [anon_sym_TILDE] = ACTIONS(1967),
    [anon_sym_QMARK] = ACTIONS(1967),
    [anon_sym_LPAREN] = ACTIONS(1969),
    [anon_sym_external] = ACTIONS(1967),
    [anon_sym_type] = ACTIONS(1967),
    [anon_sym_COMMA] = ACTIONS(1969),
    [anon_sym_PIPE] = ACTIONS(1967),
    [anon_sym_DOT] = ACTIONS(1969),
    [anon_sym_LBRACK] = ACTIONS(1967),
    [anon_sym_RBRACK] = ACTIONS(1969),
    [anon_sym_true] = ACTIONS(1967),
    [anon_sym_false] = ACTIONS(1967),
    [anon_sym_LBRACE] = ACTIONS(1967),
    [anon_sym_SEMI] = ACTIONS(1967),
    [anon_sym_exception] = ACTIONS(1967),
    [anon_sym_module] = ACTIONS(1967),
    [anon_sym_open] = ACTIONS(1967),
    [anon_sym_include] = ACTIONS(1967),
    [anon_sym_class] = ACTIONS(1967),
    [anon_sym_val] = ACTIONS(1967),
    [anon_sym_end] = ACTIONS(1967),
    [anon_sym_object] = ACTIONS(1967),
    [anon_sym_in] = ACTIONS(1967),
    [anon_sym_AMP] = ACTIONS(1967),
    [anon_sym_POUND] = ACTIONS(1967),
    [anon_sym_COLON_COLON] = ACTIONS(1969),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1969),
    [anon_sym_else] = ACTIONS(1967),
    [anon_sym_new] = ACTIONS(1967),
    [anon_sym_LBRACE_LT] = ACTIONS(1969),
    [anon_sym_begin] = ACTIONS(1967),
    [sym_ocamlyacc_value] = ACTIONS(1969),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1967),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1969),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1969),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1969),
    [aux_sym_number_token1] = ACTIONS(1969),
    [aux_sym_signed_number_token1] = ACTIONS(1967),
    [anon_sym_SQUOTE] = ACTIONS(1969),
    [anon_sym_DQUOTE] = ACTIONS(1969),
    [sym_prefix_operator] = ACTIONS(1969),
    [aux_sym_sign_operator_token1] = ACTIONS(1967),
    [sym_hash_operator] = ACTIONS(1969),
    [anon_sym_lsl] = ACTIONS(1967),
    [anon_sym_lsr] = ACTIONS(1967),
    [anon_sym_asr] = ACTIONS(1967),
    [aux_sym_pow_operator_token1] = ACTIONS(1969),
    [anon_sym_mod] = ACTIONS(1967),
    [anon_sym_land] = ACTIONS(1967),
    [anon_sym_lor] = ACTIONS(1967),
    [anon_sym_lxor] = ACTIONS(1967),
    [aux_sym_mult_operator_token1] = ACTIONS(1967),
    [aux_sym_add_operator_token1] = ACTIONS(1967),
    [sym_concat_operator] = ACTIONS(1969),
    [sym_rel_operator] = ACTIONS(1967),
    [anon_sym_AMP_AMP] = ACTIONS(1967),
    [anon_sym_or] = ACTIONS(1967),
    [anon_sym_PIPE_PIPE] = ACTIONS(1967),
    [sym_assign_operator] = ACTIONS(1969),
    [sym_let_operator] = ACTIONS(1969),
    [sym_let_and_operator] = ACTIONS(1969),
    [sym__uppercase_identifier] = ACTIONS(1969),
    [aux_sym_directive_token1] = ACTIONS(1967),
    [aux_sym_tag_token1] = ACTIONS(1969),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1198)] = {
    [sym_attribute] = STATE(1198),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_SEMI_SEMI] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2079),
    [anon_sym_and] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_external] = ACTIONS(2079),
    [anon_sym_type] = ACTIONS(2079),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_PIPE] = ACTIONS(2079),
    [anon_sym_DOT] = ACTIONS(2081),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_RBRACK] = ACTIONS(2081),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_SEMI] = ACTIONS(2079),
    [anon_sym_exception] = ACTIONS(2079),
    [anon_sym_module] = ACTIONS(2079),
    [anon_sym_open] = ACTIONS(2079),
    [anon_sym_include] = ACTIONS(2079),
    [anon_sym_class] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_end] = ACTIONS(2079),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_in] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_POUND] = ACTIONS(2079),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [sym_hash_operator] = ACTIONS(2081),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym_let_operator] = ACTIONS(2081),
    [sym_let_and_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_directive_token1] = ACTIONS(2079),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1199)] = {
    [sym_attribute] = STATE(1199),
    [sym__lowercase_identifier] = ACTIONS(2149),
    [anon_sym_SEMI_SEMI] = ACTIONS(2151),
    [anon_sym_let] = ACTIONS(2149),
    [anon_sym_and] = ACTIONS(2149),
    [anon_sym_TILDE] = ACTIONS(2149),
    [anon_sym_QMARK] = ACTIONS(2149),
    [anon_sym_LPAREN] = ACTIONS(2151),
    [anon_sym_external] = ACTIONS(2149),
    [anon_sym_type] = ACTIONS(2149),
    [anon_sym_COMMA] = ACTIONS(2151),
    [anon_sym_PIPE] = ACTIONS(2149),
    [anon_sym_DOT] = ACTIONS(2151),
    [anon_sym_LBRACK] = ACTIONS(2149),
    [anon_sym_RBRACK] = ACTIONS(2151),
    [anon_sym_true] = ACTIONS(2149),
    [anon_sym_false] = ACTIONS(2149),
    [anon_sym_LBRACE] = ACTIONS(2149),
    [anon_sym_SEMI] = ACTIONS(2149),
    [anon_sym_exception] = ACTIONS(2149),
    [anon_sym_module] = ACTIONS(2149),
    [anon_sym_open] = ACTIONS(2149),
    [anon_sym_include] = ACTIONS(2149),
    [anon_sym_class] = ACTIONS(2149),
    [anon_sym_val] = ACTIONS(2149),
    [anon_sym_end] = ACTIONS(2149),
    [anon_sym_object] = ACTIONS(2149),
    [anon_sym_in] = ACTIONS(2149),
    [anon_sym_AMP] = ACTIONS(2149),
    [anon_sym_POUND] = ACTIONS(2149),
    [anon_sym_COLON_COLON] = ACTIONS(2151),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2151),
    [anon_sym_else] = ACTIONS(2149),
    [anon_sym_new] = ACTIONS(2149),
    [anon_sym_LBRACE_LT] = ACTIONS(2151),
    [anon_sym_begin] = ACTIONS(2149),
    [sym_ocamlyacc_value] = ACTIONS(2151),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2149),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2151),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2149),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2151),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2151),
    [aux_sym_number_token1] = ACTIONS(2151),
    [aux_sym_signed_number_token1] = ACTIONS(2149),
    [anon_sym_SQUOTE] = ACTIONS(2151),
    [anon_sym_DQUOTE] = ACTIONS(2151),
    [sym_prefix_operator] = ACTIONS(2151),
    [aux_sym_sign_operator_token1] = ACTIONS(2149),
    [sym_hash_operator] = ACTIONS(2151),
    [anon_sym_lsl] = ACTIONS(2149),
    [anon_sym_lsr] = ACTIONS(2149),
    [anon_sym_asr] = ACTIONS(2149),
    [aux_sym_pow_operator_token1] = ACTIONS(2151),
    [anon_sym_mod] = ACTIONS(2149),
    [anon_sym_land] = ACTIONS(2149),
    [anon_sym_lor] = ACTIONS(2149),
    [anon_sym_lxor] = ACTIONS(2149),
    [aux_sym_mult_operator_token1] = ACTIONS(2149),
    [aux_sym_add_operator_token1] = ACTIONS(2149),
    [sym_concat_operator] = ACTIONS(2151),
    [sym_rel_operator] = ACTIONS(2149),
    [anon_sym_AMP_AMP] = ACTIONS(2149),
    [anon_sym_or] = ACTIONS(2149),
    [anon_sym_PIPE_PIPE] = ACTIONS(2149),
    [sym_assign_operator] = ACTIONS(2151),
    [sym_let_operator] = ACTIONS(2151),
    [sym_let_and_operator] = ACTIONS(2151),
    [sym__uppercase_identifier] = ACTIONS(2151),
    [aux_sym_directive_token1] = ACTIONS(2149),
    [aux_sym_tag_token1] = ACTIONS(2151),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1200)] = {
    [sym_attribute] = STATE(1200),
    [sym__lowercase_identifier] = ACTIONS(2055),
    [anon_sym_SEMI_SEMI] = ACTIONS(2057),
    [anon_sym_let] = ACTIONS(2055),
    [anon_sym_and] = ACTIONS(2055),
    [anon_sym_TILDE] = ACTIONS(2055),
    [anon_sym_QMARK] = ACTIONS(2055),
    [anon_sym_LPAREN] = ACTIONS(2057),
    [anon_sym_external] = ACTIONS(2055),
    [anon_sym_type] = ACTIONS(2055),
    [anon_sym_COMMA] = ACTIONS(2057),
    [anon_sym_PIPE] = ACTIONS(2055),
    [anon_sym_DOT] = ACTIONS(2057),
    [anon_sym_LBRACK] = ACTIONS(2055),
    [anon_sym_RBRACK] = ACTIONS(2057),
    [anon_sym_true] = ACTIONS(2055),
    [anon_sym_false] = ACTIONS(2055),
    [anon_sym_LBRACE] = ACTIONS(2055),
    [anon_sym_SEMI] = ACTIONS(2055),
    [anon_sym_exception] = ACTIONS(2055),
    [anon_sym_module] = ACTIONS(2055),
    [anon_sym_open] = ACTIONS(2055),
    [anon_sym_include] = ACTIONS(2055),
    [anon_sym_class] = ACTIONS(2055),
    [anon_sym_val] = ACTIONS(2055),
    [anon_sym_end] = ACTIONS(2055),
    [anon_sym_object] = ACTIONS(2055),
    [anon_sym_in] = ACTIONS(2055),
    [anon_sym_AMP] = ACTIONS(2055),
    [anon_sym_POUND] = ACTIONS(2055),
    [anon_sym_COLON_COLON] = ACTIONS(2057),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2057),
    [anon_sym_else] = ACTIONS(2055),
    [anon_sym_new] = ACTIONS(2055),
    [anon_sym_LBRACE_LT] = ACTIONS(2057),
    [anon_sym_begin] = ACTIONS(2055),
    [sym_ocamlyacc_value] = ACTIONS(2057),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2057),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2055),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2055),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2057),
    [aux_sym_number_token1] = ACTIONS(2057),
    [aux_sym_signed_number_token1] = ACTIONS(2055),
    [anon_sym_SQUOTE] = ACTIONS(2057),
    [anon_sym_DQUOTE] = ACTIONS(2057),
    [sym_prefix_operator] = ACTIONS(2057),
    [aux_sym_sign_operator_token1] = ACTIONS(2055),
    [sym_hash_operator] = ACTIONS(2057),
    [anon_sym_lsl] = ACTIONS(2055),
    [anon_sym_lsr] = ACTIONS(2055),
    [anon_sym_asr] = ACTIONS(2055),
    [aux_sym_pow_operator_token1] = ACTIONS(2057),
    [anon_sym_mod] = ACTIONS(2055),
    [anon_sym_land] = ACTIONS(2055),
    [anon_sym_lor] = ACTIONS(2055),
    [anon_sym_lxor] = ACTIONS(2055),
    [aux_sym_mult_operator_token1] = ACTIONS(2055),
    [aux_sym_add_operator_token1] = ACTIONS(2055),
    [sym_concat_operator] = ACTIONS(2057),
    [sym_rel_operator] = ACTIONS(2055),
    [anon_sym_AMP_AMP] = ACTIONS(2055),
    [anon_sym_or] = ACTIONS(2055),
    [anon_sym_PIPE_PIPE] = ACTIONS(2055),
    [sym_assign_operator] = ACTIONS(2057),
    [sym_let_operator] = ACTIONS(2057),
    [sym_let_and_operator] = ACTIONS(2057),
    [sym__uppercase_identifier] = ACTIONS(2057),
    [aux_sym_directive_token1] = ACTIONS(2055),
    [aux_sym_tag_token1] = ACTIONS(2057),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1201)] = {
    [sym_attribute] = STATE(1201),
    [sym__lowercase_identifier] = ACTIONS(1979),
    [anon_sym_SEMI_SEMI] = ACTIONS(1981),
    [anon_sym_let] = ACTIONS(1979),
    [anon_sym_and] = ACTIONS(1979),
    [anon_sym_TILDE] = ACTIONS(1979),
    [anon_sym_QMARK] = ACTIONS(1979),
    [anon_sym_LPAREN] = ACTIONS(1981),
    [anon_sym_external] = ACTIONS(1979),
    [anon_sym_type] = ACTIONS(1979),
    [anon_sym_COMMA] = ACTIONS(1981),
    [anon_sym_PIPE] = ACTIONS(1979),
    [anon_sym_DOT] = ACTIONS(1981),
    [anon_sym_LBRACK] = ACTIONS(1979),
    [anon_sym_RBRACK] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(1979),
    [anon_sym_false] = ACTIONS(1979),
    [anon_sym_LBRACE] = ACTIONS(1979),
    [anon_sym_SEMI] = ACTIONS(1979),
    [anon_sym_exception] = ACTIONS(1979),
    [anon_sym_module] = ACTIONS(1979),
    [anon_sym_open] = ACTIONS(1979),
    [anon_sym_include] = ACTIONS(1979),
    [anon_sym_class] = ACTIONS(1979),
    [anon_sym_val] = ACTIONS(1979),
    [anon_sym_end] = ACTIONS(1979),
    [anon_sym_object] = ACTIONS(1979),
    [anon_sym_in] = ACTIONS(1979),
    [anon_sym_AMP] = ACTIONS(1979),
    [anon_sym_POUND] = ACTIONS(1979),
    [anon_sym_COLON_COLON] = ACTIONS(1981),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1981),
    [anon_sym_else] = ACTIONS(1979),
    [anon_sym_new] = ACTIONS(1979),
    [anon_sym_LBRACE_LT] = ACTIONS(1981),
    [anon_sym_begin] = ACTIONS(1979),
    [sym_ocamlyacc_value] = ACTIONS(1981),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1979),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1981),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1979),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1979),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1981),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1981),
    [aux_sym_number_token1] = ACTIONS(1981),
    [aux_sym_signed_number_token1] = ACTIONS(1979),
    [anon_sym_SQUOTE] = ACTIONS(1981),
    [anon_sym_DQUOTE] = ACTIONS(1981),
    [sym_prefix_operator] = ACTIONS(1981),
    [aux_sym_sign_operator_token1] = ACTIONS(1979),
    [sym_hash_operator] = ACTIONS(1981),
    [anon_sym_lsl] = ACTIONS(1979),
    [anon_sym_lsr] = ACTIONS(1979),
    [anon_sym_asr] = ACTIONS(1979),
    [aux_sym_pow_operator_token1] = ACTIONS(1981),
    [anon_sym_mod] = ACTIONS(1979),
    [anon_sym_land] = ACTIONS(1979),
    [anon_sym_lor] = ACTIONS(1979),
    [anon_sym_lxor] = ACTIONS(1979),
    [aux_sym_mult_operator_token1] = ACTIONS(1979),
    [aux_sym_add_operator_token1] = ACTIONS(1979),
    [sym_concat_operator] = ACTIONS(1981),
    [sym_rel_operator] = ACTIONS(1979),
    [anon_sym_AMP_AMP] = ACTIONS(1979),
    [anon_sym_or] = ACTIONS(1979),
    [anon_sym_PIPE_PIPE] = ACTIONS(1979),
    [sym_assign_operator] = ACTIONS(1981),
    [sym_let_operator] = ACTIONS(1981),
    [sym_let_and_operator] = ACTIONS(1981),
    [sym__uppercase_identifier] = ACTIONS(1981),
    [aux_sym_directive_token1] = ACTIONS(1979),
    [aux_sym_tag_token1] = ACTIONS(1981),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1202)] = {
    [sym_attribute] = STATE(1202),
    [sym__lowercase_identifier] = ACTIONS(1999),
    [anon_sym_SEMI_SEMI] = ACTIONS(2001),
    [anon_sym_let] = ACTIONS(1999),
    [anon_sym_and] = ACTIONS(1999),
    [anon_sym_TILDE] = ACTIONS(1999),
    [anon_sym_QMARK] = ACTIONS(1999),
    [anon_sym_LPAREN] = ACTIONS(2001),
    [anon_sym_external] = ACTIONS(1999),
    [anon_sym_type] = ACTIONS(1999),
    [anon_sym_COMMA] = ACTIONS(2001),
    [anon_sym_PIPE] = ACTIONS(1999),
    [anon_sym_DOT] = ACTIONS(2001),
    [anon_sym_LBRACK] = ACTIONS(1999),
    [anon_sym_RBRACK] = ACTIONS(2001),
    [anon_sym_true] = ACTIONS(1999),
    [anon_sym_false] = ACTIONS(1999),
    [anon_sym_LBRACE] = ACTIONS(1999),
    [anon_sym_SEMI] = ACTIONS(1999),
    [anon_sym_exception] = ACTIONS(1999),
    [anon_sym_module] = ACTIONS(1999),
    [anon_sym_open] = ACTIONS(1999),
    [anon_sym_include] = ACTIONS(1999),
    [anon_sym_class] = ACTIONS(1999),
    [anon_sym_val] = ACTIONS(1999),
    [anon_sym_end] = ACTIONS(1999),
    [anon_sym_object] = ACTIONS(1999),
    [anon_sym_in] = ACTIONS(1999),
    [anon_sym_AMP] = ACTIONS(1999),
    [anon_sym_POUND] = ACTIONS(1999),
    [anon_sym_COLON_COLON] = ACTIONS(2001),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2001),
    [anon_sym_else] = ACTIONS(1999),
    [anon_sym_new] = ACTIONS(1999),
    [anon_sym_LBRACE_LT] = ACTIONS(2001),
    [anon_sym_begin] = ACTIONS(1999),
    [sym_ocamlyacc_value] = ACTIONS(2001),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2001),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1999),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1999),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2001),
    [aux_sym_number_token1] = ACTIONS(2001),
    [aux_sym_signed_number_token1] = ACTIONS(1999),
    [anon_sym_SQUOTE] = ACTIONS(2001),
    [anon_sym_DQUOTE] = ACTIONS(2001),
    [sym_prefix_operator] = ACTIONS(2001),
    [aux_sym_sign_operator_token1] = ACTIONS(1999),
    [sym_hash_operator] = ACTIONS(2001),
    [anon_sym_lsl] = ACTIONS(1999),
    [anon_sym_lsr] = ACTIONS(1999),
    [anon_sym_asr] = ACTIONS(1999),
    [aux_sym_pow_operator_token1] = ACTIONS(2001),
    [anon_sym_mod] = ACTIONS(1999),
    [anon_sym_land] = ACTIONS(1999),
    [anon_sym_lor] = ACTIONS(1999),
    [anon_sym_lxor] = ACTIONS(1999),
    [aux_sym_mult_operator_token1] = ACTIONS(1999),
    [aux_sym_add_operator_token1] = ACTIONS(1999),
    [sym_concat_operator] = ACTIONS(2001),
    [sym_rel_operator] = ACTIONS(1999),
    [anon_sym_AMP_AMP] = ACTIONS(1999),
    [anon_sym_or] = ACTIONS(1999),
    [anon_sym_PIPE_PIPE] = ACTIONS(1999),
    [sym_assign_operator] = ACTIONS(2001),
    [sym_let_operator] = ACTIONS(2001),
    [sym_let_and_operator] = ACTIONS(2001),
    [sym__uppercase_identifier] = ACTIONS(2001),
    [aux_sym_directive_token1] = ACTIONS(1999),
    [aux_sym_tag_token1] = ACTIONS(2001),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1203)] = {
    [sym_attribute] = STATE(1203),
    [sym__lowercase_identifier] = ACTIONS(2141),
    [anon_sym_SEMI_SEMI] = ACTIONS(2143),
    [anon_sym_let] = ACTIONS(2141),
    [anon_sym_and] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_QMARK] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2143),
    [anon_sym_external] = ACTIONS(2141),
    [anon_sym_type] = ACTIONS(2141),
    [anon_sym_COMMA] = ACTIONS(2143),
    [anon_sym_PIPE] = ACTIONS(2141),
    [anon_sym_DOT] = ACTIONS(2143),
    [anon_sym_LBRACK] = ACTIONS(2141),
    [anon_sym_RBRACK] = ACTIONS(2143),
    [anon_sym_true] = ACTIONS(2141),
    [anon_sym_false] = ACTIONS(2141),
    [anon_sym_LBRACE] = ACTIONS(2141),
    [anon_sym_SEMI] = ACTIONS(2141),
    [anon_sym_exception] = ACTIONS(2141),
    [anon_sym_module] = ACTIONS(2141),
    [anon_sym_open] = ACTIONS(2141),
    [anon_sym_include] = ACTIONS(2141),
    [anon_sym_class] = ACTIONS(2141),
    [anon_sym_val] = ACTIONS(2141),
    [anon_sym_end] = ACTIONS(2141),
    [anon_sym_object] = ACTIONS(2141),
    [anon_sym_in] = ACTIONS(2141),
    [anon_sym_AMP] = ACTIONS(2141),
    [anon_sym_POUND] = ACTIONS(2141),
    [anon_sym_COLON_COLON] = ACTIONS(2143),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2143),
    [anon_sym_else] = ACTIONS(2141),
    [anon_sym_new] = ACTIONS(2141),
    [anon_sym_LBRACE_LT] = ACTIONS(2143),
    [anon_sym_begin] = ACTIONS(2141),
    [sym_ocamlyacc_value] = ACTIONS(2143),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2141),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2143),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2141),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2141),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2143),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2143),
    [aux_sym_number_token1] = ACTIONS(2143),
    [aux_sym_signed_number_token1] = ACTIONS(2141),
    [anon_sym_SQUOTE] = ACTIONS(2143),
    [anon_sym_DQUOTE] = ACTIONS(2143),
    [sym_prefix_operator] = ACTIONS(2143),
    [aux_sym_sign_operator_token1] = ACTIONS(2141),
    [sym_hash_operator] = ACTIONS(2143),
    [anon_sym_lsl] = ACTIONS(2141),
    [anon_sym_lsr] = ACTIONS(2141),
    [anon_sym_asr] = ACTIONS(2141),
    [aux_sym_pow_operator_token1] = ACTIONS(2143),
    [anon_sym_mod] = ACTIONS(2141),
    [anon_sym_land] = ACTIONS(2141),
    [anon_sym_lor] = ACTIONS(2141),
    [anon_sym_lxor] = ACTIONS(2141),
    [aux_sym_mult_operator_token1] = ACTIONS(2141),
    [aux_sym_add_operator_token1] = ACTIONS(2141),
    [sym_concat_operator] = ACTIONS(2143),
    [sym_rel_operator] = ACTIONS(2141),
    [anon_sym_AMP_AMP] = ACTIONS(2141),
    [anon_sym_or] = ACTIONS(2141),
    [anon_sym_PIPE_PIPE] = ACTIONS(2141),
    [sym_assign_operator] = ACTIONS(2143),
    [sym_let_operator] = ACTIONS(2143),
    [sym_let_and_operator] = ACTIONS(2143),
    [sym__uppercase_identifier] = ACTIONS(2143),
    [aux_sym_directive_token1] = ACTIONS(2141),
    [aux_sym_tag_token1] = ACTIONS(2143),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1204)] = {
    [sym_attribute] = STATE(1204),
    [sym__lowercase_identifier] = ACTIONS(1927),
    [anon_sym_SEMI_SEMI] = ACTIONS(1929),
    [anon_sym_let] = ACTIONS(1927),
    [anon_sym_and] = ACTIONS(1927),
    [anon_sym_TILDE] = ACTIONS(1927),
    [anon_sym_QMARK] = ACTIONS(1927),
    [anon_sym_LPAREN] = ACTIONS(1929),
    [anon_sym_external] = ACTIONS(1927),
    [anon_sym_type] = ACTIONS(1927),
    [anon_sym_COMMA] = ACTIONS(1929),
    [anon_sym_PIPE] = ACTIONS(1927),
    [anon_sym_DOT] = ACTIONS(1929),
    [anon_sym_LBRACK] = ACTIONS(1927),
    [anon_sym_RBRACK] = ACTIONS(1929),
    [anon_sym_true] = ACTIONS(1927),
    [anon_sym_false] = ACTIONS(1927),
    [anon_sym_LBRACE] = ACTIONS(1927),
    [anon_sym_SEMI] = ACTIONS(1927),
    [anon_sym_exception] = ACTIONS(1927),
    [anon_sym_module] = ACTIONS(1927),
    [anon_sym_open] = ACTIONS(1927),
    [anon_sym_include] = ACTIONS(1927),
    [anon_sym_class] = ACTIONS(1927),
    [anon_sym_val] = ACTIONS(1927),
    [anon_sym_end] = ACTIONS(1927),
    [anon_sym_object] = ACTIONS(1927),
    [anon_sym_in] = ACTIONS(1927),
    [anon_sym_AMP] = ACTIONS(1927),
    [anon_sym_POUND] = ACTIONS(1927),
    [anon_sym_COLON_COLON] = ACTIONS(1929),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1929),
    [anon_sym_else] = ACTIONS(1927),
    [anon_sym_new] = ACTIONS(1927),
    [anon_sym_LBRACE_LT] = ACTIONS(1929),
    [anon_sym_begin] = ACTIONS(1927),
    [sym_ocamlyacc_value] = ACTIONS(1929),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1927),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1929),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1927),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1927),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1929),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1929),
    [aux_sym_number_token1] = ACTIONS(1929),
    [aux_sym_signed_number_token1] = ACTIONS(1927),
    [anon_sym_SQUOTE] = ACTIONS(1929),
    [anon_sym_DQUOTE] = ACTIONS(1929),
    [sym_prefix_operator] = ACTIONS(1929),
    [aux_sym_sign_operator_token1] = ACTIONS(1927),
    [sym_hash_operator] = ACTIONS(1929),
    [anon_sym_lsl] = ACTIONS(1927),
    [anon_sym_lsr] = ACTIONS(1927),
    [anon_sym_asr] = ACTIONS(1927),
    [aux_sym_pow_operator_token1] = ACTIONS(1929),
    [anon_sym_mod] = ACTIONS(1927),
    [anon_sym_land] = ACTIONS(1927),
    [anon_sym_lor] = ACTIONS(1927),
    [anon_sym_lxor] = ACTIONS(1927),
    [aux_sym_mult_operator_token1] = ACTIONS(1927),
    [aux_sym_add_operator_token1] = ACTIONS(1927),
    [sym_concat_operator] = ACTIONS(1929),
    [sym_rel_operator] = ACTIONS(1927),
    [anon_sym_AMP_AMP] = ACTIONS(1927),
    [anon_sym_or] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1927),
    [sym_assign_operator] = ACTIONS(1929),
    [sym_let_operator] = ACTIONS(1929),
    [sym_let_and_operator] = ACTIONS(1929),
    [sym__uppercase_identifier] = ACTIONS(1929),
    [aux_sym_directive_token1] = ACTIONS(1927),
    [aux_sym_tag_token1] = ACTIONS(1929),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1205)] = {
    [sym_attribute] = STATE(1205),
    [sym__lowercase_identifier] = ACTIONS(2311),
    [anon_sym_TILDE] = ACTIONS(2311),
    [anon_sym_QMARK] = ACTIONS(2311),
    [anon_sym_COLON] = ACTIONS(2534),
    [anon_sym_LPAREN] = ACTIONS(2315),
    [anon_sym_RPAREN] = ACTIONS(2315),
    [anon_sym_COMMA] = ACTIONS(2315),
    [anon_sym_PIPE] = ACTIONS(2311),
    [anon_sym_COLON2] = ACTIONS(2311),
    [anon_sym_DASH_GT] = ACTIONS(2311),
    [anon_sym_LBRACK] = ACTIONS(2311),
    [anon_sym_RBRACK] = ACTIONS(2315),
    [anon_sym_true] = ACTIONS(2311),
    [anon_sym_false] = ACTIONS(2311),
    [anon_sym_LBRACE] = ACTIONS(2311),
    [anon_sym_SEMI] = ACTIONS(2315),
    [anon_sym_RBRACE] = ACTIONS(2315),
    [anon_sym_constraint] = ACTIONS(2311),
    [anon_sym_val] = ACTIONS(2311),
    [anon_sym_end] = ACTIONS(2311),
    [anon_sym_with] = ACTIONS(2311),
    [anon_sym_COLON_GT] = ACTIONS(2315),
    [anon_sym_object] = ACTIONS(2311),
    [anon_sym_inherit] = ACTIONS(2311),
    [anon_sym_method] = ACTIONS(2311),
    [anon_sym_initializer] = ACTIONS(2311),
    [anon_sym_AMP] = ACTIONS(2311),
    [anon_sym_COLON_COLON] = ACTIONS(2315),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2315),
    [anon_sym_then] = ACTIONS(2311),
    [anon_sym_else] = ACTIONS(2311),
    [anon_sym_do] = ACTIONS(2311),
    [anon_sym_done] = ACTIONS(2311),
    [anon_sym_to] = ACTIONS(2311),
    [anon_sym_downto] = ACTIONS(2311),
    [anon_sym_new] = ACTIONS(2311),
    [anon_sym_LBRACE_LT] = ACTIONS(2315),
    [anon_sym_GT_RBRACE] = ACTIONS(2315),
    [anon_sym_begin] = ACTIONS(2311),
    [sym_ocamlyacc_value] = ACTIONS(2315),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2311),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2315),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2311),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2311),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2315),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2315),
    [aux_sym_number_token1] = ACTIONS(2315),
    [aux_sym_signed_number_token1] = ACTIONS(2311),
    [anon_sym_SQUOTE] = ACTIONS(2315),
    [anon_sym_DQUOTE] = ACTIONS(2315),
    [sym_prefix_operator] = ACTIONS(2315),
    [aux_sym_sign_operator_token1] = ACTIONS(2311),
    [anon_sym_lsl] = ACTIONS(2311),
    [anon_sym_lsr] = ACTIONS(2311),
    [anon_sym_asr] = ACTIONS(2311),
    [aux_sym_pow_operator_token1] = ACTIONS(2315),
    [anon_sym_mod] = ACTIONS(2311),
    [anon_sym_land] = ACTIONS(2311),
    [anon_sym_lor] = ACTIONS(2311),
    [anon_sym_lxor] = ACTIONS(2311),
    [aux_sym_mult_operator_token1] = ACTIONS(2311),
    [aux_sym_add_operator_token1] = ACTIONS(2311),
    [sym_concat_operator] = ACTIONS(2315),
    [sym_rel_operator] = ACTIONS(2311),
    [anon_sym_AMP_AMP] = ACTIONS(2311),
    [anon_sym_or] = ACTIONS(2311),
    [anon_sym_PIPE_PIPE] = ACTIONS(2311),
    [sym_assign_operator] = ACTIONS(2315),
    [sym__uppercase_identifier] = ACTIONS(2315),
    [aux_sym_tag_token1] = ACTIONS(2315),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1206)] = {
    [sym_attribute] = STATE(1206),
    [ts_builtin_sym_end] = ACTIONS(1897),
    [sym__lowercase_identifier] = ACTIONS(1895),
    [anon_sym_SEMI_SEMI] = ACTIONS(1897),
    [anon_sym_let] = ACTIONS(1895),
    [anon_sym_and] = ACTIONS(1895),
    [anon_sym_TILDE] = ACTIONS(1895),
    [anon_sym_QMARK] = ACTIONS(1895),
    [anon_sym_LPAREN] = ACTIONS(1897),
    [anon_sym_external] = ACTIONS(1895),
    [anon_sym_type] = ACTIONS(1895),
    [anon_sym_COMMA] = ACTIONS(1897),
    [anon_sym_PIPE] = ACTIONS(1895),
    [anon_sym_DOT] = ACTIONS(1897),
    [anon_sym_LBRACK] = ACTIONS(1895),
    [anon_sym_true] = ACTIONS(1895),
    [anon_sym_false] = ACTIONS(1895),
    [anon_sym_LBRACE] = ACTIONS(1895),
    [anon_sym_SEMI] = ACTIONS(1895),
    [anon_sym_exception] = ACTIONS(1895),
    [anon_sym_module] = ACTIONS(1895),
    [anon_sym_open] = ACTIONS(1895),
    [anon_sym_include] = ACTIONS(1895),
    [anon_sym_class] = ACTIONS(1895),
    [anon_sym_val] = ACTIONS(1895),
    [anon_sym_object] = ACTIONS(1895),
    [anon_sym_in] = ACTIONS(1895),
    [anon_sym_AMP] = ACTIONS(1895),
    [anon_sym_POUND] = ACTIONS(1895),
    [anon_sym_COLON_COLON] = ACTIONS(1897),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1897),
    [anon_sym_LT_DASH] = ACTIONS(1895),
    [anon_sym_else] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1895),
    [anon_sym_LBRACE_LT] = ACTIONS(1897),
    [anon_sym_begin] = ACTIONS(1895),
    [sym_ocamlyacc_value] = ACTIONS(1897),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1895),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1897),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1895),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1895),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1897),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1897),
    [aux_sym_number_token1] = ACTIONS(1897),
    [aux_sym_signed_number_token1] = ACTIONS(1895),
    [anon_sym_SQUOTE] = ACTIONS(1897),
    [anon_sym_DQUOTE] = ACTIONS(1897),
    [sym_prefix_operator] = ACTIONS(1897),
    [aux_sym_sign_operator_token1] = ACTIONS(1895),
    [sym_hash_operator] = ACTIONS(1897),
    [anon_sym_lsl] = ACTIONS(1895),
    [anon_sym_lsr] = ACTIONS(1895),
    [anon_sym_asr] = ACTIONS(1895),
    [aux_sym_pow_operator_token1] = ACTIONS(1897),
    [anon_sym_mod] = ACTIONS(1895),
    [anon_sym_land] = ACTIONS(1895),
    [anon_sym_lor] = ACTIONS(1895),
    [anon_sym_lxor] = ACTIONS(1895),
    [aux_sym_mult_operator_token1] = ACTIONS(1895),
    [aux_sym_add_operator_token1] = ACTIONS(1895),
    [sym_concat_operator] = ACTIONS(1897),
    [sym_rel_operator] = ACTIONS(1895),
    [anon_sym_AMP_AMP] = ACTIONS(1895),
    [anon_sym_or] = ACTIONS(1895),
    [anon_sym_PIPE_PIPE] = ACTIONS(1895),
    [sym_assign_operator] = ACTIONS(1897),
    [sym_let_operator] = ACTIONS(1897),
    [sym_let_and_operator] = ACTIONS(1897),
    [sym__uppercase_identifier] = ACTIONS(1897),
    [aux_sym_directive_token1] = ACTIONS(1895),
    [aux_sym_tag_token1] = ACTIONS(1897),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1207)] = {
    [sym_attribute] = STATE(1207),
    [sym__lowercase_identifier] = ACTIONS(2003),
    [anon_sym_SEMI_SEMI] = ACTIONS(2005),
    [anon_sym_let] = ACTIONS(2003),
    [anon_sym_and] = ACTIONS(2003),
    [anon_sym_TILDE] = ACTIONS(2003),
    [anon_sym_QMARK] = ACTIONS(2003),
    [anon_sym_LPAREN] = ACTIONS(2005),
    [anon_sym_external] = ACTIONS(2003),
    [anon_sym_type] = ACTIONS(2003),
    [anon_sym_COMMA] = ACTIONS(2005),
    [anon_sym_PIPE] = ACTIONS(2003),
    [anon_sym_DOT] = ACTIONS(2005),
    [anon_sym_LBRACK] = ACTIONS(2003),
    [anon_sym_RBRACK] = ACTIONS(2005),
    [anon_sym_true] = ACTIONS(2003),
    [anon_sym_false] = ACTIONS(2003),
    [anon_sym_LBRACE] = ACTIONS(2003),
    [anon_sym_SEMI] = ACTIONS(2003),
    [anon_sym_exception] = ACTIONS(2003),
    [anon_sym_module] = ACTIONS(2003),
    [anon_sym_open] = ACTIONS(2003),
    [anon_sym_include] = ACTIONS(2003),
    [anon_sym_class] = ACTIONS(2003),
    [anon_sym_val] = ACTIONS(2003),
    [anon_sym_end] = ACTIONS(2003),
    [anon_sym_object] = ACTIONS(2003),
    [anon_sym_in] = ACTIONS(2003),
    [anon_sym_AMP] = ACTIONS(2003),
    [anon_sym_POUND] = ACTIONS(2003),
    [anon_sym_COLON_COLON] = ACTIONS(2005),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2005),
    [anon_sym_else] = ACTIONS(2003),
    [anon_sym_new] = ACTIONS(2003),
    [anon_sym_LBRACE_LT] = ACTIONS(2005),
    [anon_sym_begin] = ACTIONS(2003),
    [sym_ocamlyacc_value] = ACTIONS(2005),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2003),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2005),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2005),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2005),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2003),
    [anon_sym_SQUOTE] = ACTIONS(2005),
    [anon_sym_DQUOTE] = ACTIONS(2005),
    [sym_prefix_operator] = ACTIONS(2005),
    [aux_sym_sign_operator_token1] = ACTIONS(2003),
    [sym_hash_operator] = ACTIONS(2005),
    [anon_sym_lsl] = ACTIONS(2003),
    [anon_sym_lsr] = ACTIONS(2003),
    [anon_sym_asr] = ACTIONS(2003),
    [aux_sym_pow_operator_token1] = ACTIONS(2005),
    [anon_sym_mod] = ACTIONS(2003),
    [anon_sym_land] = ACTIONS(2003),
    [anon_sym_lor] = ACTIONS(2003),
    [anon_sym_lxor] = ACTIONS(2003),
    [aux_sym_mult_operator_token1] = ACTIONS(2003),
    [aux_sym_add_operator_token1] = ACTIONS(2003),
    [sym_concat_operator] = ACTIONS(2005),
    [sym_rel_operator] = ACTIONS(2003),
    [anon_sym_AMP_AMP] = ACTIONS(2003),
    [anon_sym_or] = ACTIONS(2003),
    [anon_sym_PIPE_PIPE] = ACTIONS(2003),
    [sym_assign_operator] = ACTIONS(2005),
    [sym_let_operator] = ACTIONS(2005),
    [sym_let_and_operator] = ACTIONS(2005),
    [sym__uppercase_identifier] = ACTIONS(2005),
    [aux_sym_directive_token1] = ACTIONS(2003),
    [aux_sym_tag_token1] = ACTIONS(2005),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1208)] = {
    [sym_attribute] = STATE(1208),
    [sym__lowercase_identifier] = ACTIONS(2195),
    [anon_sym_SEMI_SEMI] = ACTIONS(2197),
    [anon_sym_let] = ACTIONS(2195),
    [anon_sym_and] = ACTIONS(2195),
    [anon_sym_TILDE] = ACTIONS(2195),
    [anon_sym_QMARK] = ACTIONS(2195),
    [anon_sym_LPAREN] = ACTIONS(2197),
    [anon_sym_external] = ACTIONS(2195),
    [anon_sym_type] = ACTIONS(2195),
    [anon_sym_COMMA] = ACTIONS(2197),
    [anon_sym_PIPE] = ACTIONS(2195),
    [anon_sym_DOT] = ACTIONS(2197),
    [anon_sym_LBRACK] = ACTIONS(2195),
    [anon_sym_RBRACK] = ACTIONS(2197),
    [anon_sym_true] = ACTIONS(2195),
    [anon_sym_false] = ACTIONS(2195),
    [anon_sym_LBRACE] = ACTIONS(2195),
    [anon_sym_SEMI] = ACTIONS(2195),
    [anon_sym_exception] = ACTIONS(2195),
    [anon_sym_module] = ACTIONS(2195),
    [anon_sym_open] = ACTIONS(2195),
    [anon_sym_include] = ACTIONS(2195),
    [anon_sym_class] = ACTIONS(2195),
    [anon_sym_val] = ACTIONS(2195),
    [anon_sym_end] = ACTIONS(2195),
    [anon_sym_object] = ACTIONS(2195),
    [anon_sym_in] = ACTIONS(2195),
    [anon_sym_AMP] = ACTIONS(2195),
    [anon_sym_POUND] = ACTIONS(2195),
    [anon_sym_COLON_COLON] = ACTIONS(2197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2197),
    [anon_sym_else] = ACTIONS(2195),
    [anon_sym_new] = ACTIONS(2195),
    [anon_sym_LBRACE_LT] = ACTIONS(2197),
    [anon_sym_begin] = ACTIONS(2195),
    [sym_ocamlyacc_value] = ACTIONS(2197),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2195),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2197),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2195),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2195),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2197),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2197),
    [aux_sym_number_token1] = ACTIONS(2197),
    [aux_sym_signed_number_token1] = ACTIONS(2195),
    [anon_sym_SQUOTE] = ACTIONS(2197),
    [anon_sym_DQUOTE] = ACTIONS(2197),
    [sym_prefix_operator] = ACTIONS(2197),
    [aux_sym_sign_operator_token1] = ACTIONS(2195),
    [sym_hash_operator] = ACTIONS(2197),
    [anon_sym_lsl] = ACTIONS(2195),
    [anon_sym_lsr] = ACTIONS(2195),
    [anon_sym_asr] = ACTIONS(2195),
    [aux_sym_pow_operator_token1] = ACTIONS(2197),
    [anon_sym_mod] = ACTIONS(2195),
    [anon_sym_land] = ACTIONS(2195),
    [anon_sym_lor] = ACTIONS(2195),
    [anon_sym_lxor] = ACTIONS(2195),
    [aux_sym_mult_operator_token1] = ACTIONS(2195),
    [aux_sym_add_operator_token1] = ACTIONS(2195),
    [sym_concat_operator] = ACTIONS(2197),
    [sym_rel_operator] = ACTIONS(2195),
    [anon_sym_AMP_AMP] = ACTIONS(2195),
    [anon_sym_or] = ACTIONS(2195),
    [anon_sym_PIPE_PIPE] = ACTIONS(2195),
    [sym_assign_operator] = ACTIONS(2197),
    [sym_let_operator] = ACTIONS(2197),
    [sym_let_and_operator] = ACTIONS(2197),
    [sym__uppercase_identifier] = ACTIONS(2197),
    [aux_sym_directive_token1] = ACTIONS(2195),
    [aux_sym_tag_token1] = ACTIONS(2197),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1209)] = {
    [sym_parameter] = STATE(2862),
    [sym__class_typed] = STATE(4659),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1209),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(7381),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(3191),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(4249),
    [aux_sym_let_binding_repeat1] = STATE(1213),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(2331),
    [anon_sym_and] = ACTIONS(2333),
    [anon_sym_EQ] = ACTIONS(2536),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_QMARK] = ACTIONS(2524),
    [anon_sym_LPAREN] = ACTIONS(2526),
    [anon_sym_external] = ACTIONS(2333),
    [anon_sym_type] = ACTIONS(2333),
    [anon_sym_COLON2] = ACTIONS(2528),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_RBRACK] = ACTIONS(2331),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_exception] = ACTIONS(2333),
    [anon_sym_module] = ACTIONS(2333),
    [anon_sym_open] = ACTIONS(2333),
    [anon_sym_include] = ACTIONS(2333),
    [anon_sym_class] = ACTIONS(2333),
    [anon_sym_val] = ACTIONS(2333),
    [anon_sym_end] = ACTIONS(2333),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2331),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2331),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2331),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(2530),
    [aux_sym_tag_token1] = ACTIONS(2532),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1210)] = {
    [sym_parameter] = STATE(2862),
    [sym__class_typed] = STATE(4660),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1210),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(7381),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(3191),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(4332),
    [aux_sym_let_binding_repeat1] = STATE(1477),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(2409),
    [anon_sym_and] = ACTIONS(2411),
    [anon_sym_EQ] = ACTIONS(2538),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_QMARK] = ACTIONS(2524),
    [anon_sym_LPAREN] = ACTIONS(2526),
    [anon_sym_external] = ACTIONS(2411),
    [anon_sym_type] = ACTIONS(2411),
    [anon_sym_COLON2] = ACTIONS(2528),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_RBRACK] = ACTIONS(2409),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_exception] = ACTIONS(2411),
    [anon_sym_module] = ACTIONS(2411),
    [anon_sym_open] = ACTIONS(2411),
    [anon_sym_include] = ACTIONS(2411),
    [anon_sym_class] = ACTIONS(2411),
    [anon_sym_val] = ACTIONS(2411),
    [anon_sym_end] = ACTIONS(2411),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2409),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2409),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2409),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(2530),
    [aux_sym_tag_token1] = ACTIONS(2532),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1211)] = {
    [sym_parameter] = STATE(2862),
    [sym__class_typed] = STATE(4681),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1211),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(7381),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(3191),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(4205),
    [aux_sym_let_binding_repeat1] = STATE(1477),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(2403),
    [anon_sym_and] = ACTIONS(2405),
    [anon_sym_EQ] = ACTIONS(2540),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_QMARK] = ACTIONS(2524),
    [anon_sym_LPAREN] = ACTIONS(2526),
    [anon_sym_external] = ACTIONS(2405),
    [anon_sym_type] = ACTIONS(2405),
    [anon_sym_COLON2] = ACTIONS(2528),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_RBRACK] = ACTIONS(2403),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_exception] = ACTIONS(2405),
    [anon_sym_module] = ACTIONS(2405),
    [anon_sym_open] = ACTIONS(2405),
    [anon_sym_include] = ACTIONS(2405),
    [anon_sym_class] = ACTIONS(2405),
    [anon_sym_val] = ACTIONS(2405),
    [anon_sym_end] = ACTIONS(2405),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2403),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2403),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2403),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(2530),
    [aux_sym_tag_token1] = ACTIONS(2532),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1212)] = {
    [sym_attribute] = STATE(1212),
    [ts_builtin_sym_end] = ACTIONS(1901),
    [sym__lowercase_identifier] = ACTIONS(1899),
    [anon_sym_SEMI_SEMI] = ACTIONS(1901),
    [anon_sym_let] = ACTIONS(1899),
    [anon_sym_and] = ACTIONS(1899),
    [anon_sym_TILDE] = ACTIONS(1899),
    [anon_sym_QMARK] = ACTIONS(1899),
    [anon_sym_LPAREN] = ACTIONS(1901),
    [anon_sym_external] = ACTIONS(1899),
    [anon_sym_type] = ACTIONS(1899),
    [anon_sym_COMMA] = ACTIONS(1901),
    [anon_sym_PIPE] = ACTIONS(1899),
    [anon_sym_DOT] = ACTIONS(1901),
    [anon_sym_LBRACK] = ACTIONS(1899),
    [anon_sym_true] = ACTIONS(1899),
    [anon_sym_false] = ACTIONS(1899),
    [anon_sym_LBRACE] = ACTIONS(1899),
    [anon_sym_SEMI] = ACTIONS(1899),
    [anon_sym_exception] = ACTIONS(1899),
    [anon_sym_module] = ACTIONS(1899),
    [anon_sym_open] = ACTIONS(1899),
    [anon_sym_include] = ACTIONS(1899),
    [anon_sym_class] = ACTIONS(1899),
    [anon_sym_val] = ACTIONS(1899),
    [anon_sym_object] = ACTIONS(1899),
    [anon_sym_in] = ACTIONS(1899),
    [anon_sym_AMP] = ACTIONS(1899),
    [anon_sym_POUND] = ACTIONS(1899),
    [anon_sym_COLON_COLON] = ACTIONS(1901),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1901),
    [anon_sym_LT_DASH] = ACTIONS(1899),
    [anon_sym_else] = ACTIONS(1899),
    [anon_sym_new] = ACTIONS(1899),
    [anon_sym_LBRACE_LT] = ACTIONS(1901),
    [anon_sym_begin] = ACTIONS(1899),
    [sym_ocamlyacc_value] = ACTIONS(1901),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1899),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1901),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1899),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1899),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1901),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1901),
    [aux_sym_number_token1] = ACTIONS(1901),
    [aux_sym_signed_number_token1] = ACTIONS(1899),
    [anon_sym_SQUOTE] = ACTIONS(1901),
    [anon_sym_DQUOTE] = ACTIONS(1901),
    [sym_prefix_operator] = ACTIONS(1901),
    [aux_sym_sign_operator_token1] = ACTIONS(1899),
    [sym_hash_operator] = ACTIONS(1901),
    [anon_sym_lsl] = ACTIONS(1899),
    [anon_sym_lsr] = ACTIONS(1899),
    [anon_sym_asr] = ACTIONS(1899),
    [aux_sym_pow_operator_token1] = ACTIONS(1901),
    [anon_sym_mod] = ACTIONS(1899),
    [anon_sym_land] = ACTIONS(1899),
    [anon_sym_lor] = ACTIONS(1899),
    [anon_sym_lxor] = ACTIONS(1899),
    [aux_sym_mult_operator_token1] = ACTIONS(1899),
    [aux_sym_add_operator_token1] = ACTIONS(1899),
    [sym_concat_operator] = ACTIONS(1901),
    [sym_rel_operator] = ACTIONS(1899),
    [anon_sym_AMP_AMP] = ACTIONS(1899),
    [anon_sym_or] = ACTIONS(1899),
    [anon_sym_PIPE_PIPE] = ACTIONS(1899),
    [sym_assign_operator] = ACTIONS(1901),
    [sym_let_operator] = ACTIONS(1901),
    [sym_let_and_operator] = ACTIONS(1901),
    [sym__uppercase_identifier] = ACTIONS(1901),
    [aux_sym_directive_token1] = ACTIONS(1899),
    [aux_sym_tag_token1] = ACTIONS(1901),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1213)] = {
    [sym_parameter] = STATE(2862),
    [sym__class_typed] = STATE(4663),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1213),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(7381),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(3191),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(4211),
    [aux_sym_let_binding_repeat1] = STATE(1477),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(2419),
    [anon_sym_and] = ACTIONS(2421),
    [anon_sym_EQ] = ACTIONS(2542),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_QMARK] = ACTIONS(2524),
    [anon_sym_LPAREN] = ACTIONS(2526),
    [anon_sym_external] = ACTIONS(2421),
    [anon_sym_type] = ACTIONS(2421),
    [anon_sym_COLON2] = ACTIONS(2528),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_RBRACK] = ACTIONS(2419),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_exception] = ACTIONS(2421),
    [anon_sym_module] = ACTIONS(2421),
    [anon_sym_open] = ACTIONS(2421),
    [anon_sym_include] = ACTIONS(2421),
    [anon_sym_class] = ACTIONS(2421),
    [anon_sym_val] = ACTIONS(2421),
    [anon_sym_end] = ACTIONS(2421),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2419),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2419),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2419),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(2530),
    [aux_sym_tag_token1] = ACTIONS(2532),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1214)] = {
    [sym_attribute] = STATE(1214),
    [ts_builtin_sym_end] = ACTIONS(1893),
    [sym__lowercase_identifier] = ACTIONS(1891),
    [anon_sym_SEMI_SEMI] = ACTIONS(1893),
    [anon_sym_let] = ACTIONS(1891),
    [anon_sym_and] = ACTIONS(1891),
    [anon_sym_TILDE] = ACTIONS(1891),
    [anon_sym_QMARK] = ACTIONS(1891),
    [anon_sym_LPAREN] = ACTIONS(1893),
    [anon_sym_external] = ACTIONS(1891),
    [anon_sym_type] = ACTIONS(1891),
    [anon_sym_COMMA] = ACTIONS(1893),
    [anon_sym_PIPE] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACK] = ACTIONS(1891),
    [anon_sym_true] = ACTIONS(1891),
    [anon_sym_false] = ACTIONS(1891),
    [anon_sym_LBRACE] = ACTIONS(1891),
    [anon_sym_SEMI] = ACTIONS(1891),
    [anon_sym_exception] = ACTIONS(1891),
    [anon_sym_module] = ACTIONS(1891),
    [anon_sym_open] = ACTIONS(1891),
    [anon_sym_include] = ACTIONS(1891),
    [anon_sym_class] = ACTIONS(1891),
    [anon_sym_val] = ACTIONS(1891),
    [anon_sym_object] = ACTIONS(1891),
    [anon_sym_in] = ACTIONS(1891),
    [anon_sym_AMP] = ACTIONS(1891),
    [anon_sym_POUND] = ACTIONS(1891),
    [anon_sym_COLON_COLON] = ACTIONS(1893),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1893),
    [anon_sym_LT_DASH] = ACTIONS(1891),
    [anon_sym_else] = ACTIONS(1891),
    [anon_sym_new] = ACTIONS(1891),
    [anon_sym_LBRACE_LT] = ACTIONS(1893),
    [anon_sym_begin] = ACTIONS(1891),
    [sym_ocamlyacc_value] = ACTIONS(1893),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1891),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1893),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1891),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1891),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1893),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1893),
    [aux_sym_number_token1] = ACTIONS(1893),
    [aux_sym_signed_number_token1] = ACTIONS(1891),
    [anon_sym_SQUOTE] = ACTIONS(1893),
    [anon_sym_DQUOTE] = ACTIONS(1893),
    [sym_prefix_operator] = ACTIONS(1893),
    [aux_sym_sign_operator_token1] = ACTIONS(1891),
    [sym_hash_operator] = ACTIONS(1893),
    [anon_sym_lsl] = ACTIONS(1891),
    [anon_sym_lsr] = ACTIONS(1891),
    [anon_sym_asr] = ACTIONS(1891),
    [aux_sym_pow_operator_token1] = ACTIONS(1893),
    [anon_sym_mod] = ACTIONS(1891),
    [anon_sym_land] = ACTIONS(1891),
    [anon_sym_lor] = ACTIONS(1891),
    [anon_sym_lxor] = ACTIONS(1891),
    [aux_sym_mult_operator_token1] = ACTIONS(1891),
    [aux_sym_add_operator_token1] = ACTIONS(1891),
    [sym_concat_operator] = ACTIONS(1893),
    [sym_rel_operator] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1891),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [sym_assign_operator] = ACTIONS(1893),
    [sym_let_operator] = ACTIONS(1893),
    [sym_let_and_operator] = ACTIONS(1893),
    [sym__uppercase_identifier] = ACTIONS(1893),
    [aux_sym_directive_token1] = ACTIONS(1891),
    [aux_sym_tag_token1] = ACTIONS(1893),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1215)] = {
    [sym_attribute] = STATE(1215),
    [sym__lowercase_identifier] = ACTIONS(1935),
    [anon_sym_SEMI_SEMI] = ACTIONS(1937),
    [anon_sym_let] = ACTIONS(1935),
    [anon_sym_and] = ACTIONS(1935),
    [anon_sym_TILDE] = ACTIONS(1935),
    [anon_sym_QMARK] = ACTIONS(1935),
    [anon_sym_LPAREN] = ACTIONS(1937),
    [anon_sym_external] = ACTIONS(1935),
    [anon_sym_type] = ACTIONS(1935),
    [anon_sym_COMMA] = ACTIONS(1937),
    [anon_sym_PIPE] = ACTIONS(1935),
    [anon_sym_DOT] = ACTIONS(1937),
    [anon_sym_LBRACK] = ACTIONS(1935),
    [anon_sym_RBRACK] = ACTIONS(1937),
    [anon_sym_true] = ACTIONS(1935),
    [anon_sym_false] = ACTIONS(1935),
    [anon_sym_LBRACE] = ACTIONS(1935),
    [anon_sym_SEMI] = ACTIONS(1935),
    [anon_sym_exception] = ACTIONS(1935),
    [anon_sym_module] = ACTIONS(1935),
    [anon_sym_open] = ACTIONS(1935),
    [anon_sym_include] = ACTIONS(1935),
    [anon_sym_class] = ACTIONS(1935),
    [anon_sym_val] = ACTIONS(1935),
    [anon_sym_end] = ACTIONS(1935),
    [anon_sym_object] = ACTIONS(1935),
    [anon_sym_in] = ACTIONS(1935),
    [anon_sym_AMP] = ACTIONS(1935),
    [anon_sym_POUND] = ACTIONS(1935),
    [anon_sym_COLON_COLON] = ACTIONS(1937),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1937),
    [anon_sym_else] = ACTIONS(1935),
    [anon_sym_new] = ACTIONS(1935),
    [anon_sym_LBRACE_LT] = ACTIONS(1937),
    [anon_sym_begin] = ACTIONS(1935),
    [sym_ocamlyacc_value] = ACTIONS(1937),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1935),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1937),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1935),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1935),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1937),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1937),
    [aux_sym_number_token1] = ACTIONS(1937),
    [aux_sym_signed_number_token1] = ACTIONS(1935),
    [anon_sym_SQUOTE] = ACTIONS(1937),
    [anon_sym_DQUOTE] = ACTIONS(1937),
    [sym_prefix_operator] = ACTIONS(1937),
    [aux_sym_sign_operator_token1] = ACTIONS(1935),
    [sym_hash_operator] = ACTIONS(1937),
    [anon_sym_lsl] = ACTIONS(1935),
    [anon_sym_lsr] = ACTIONS(1935),
    [anon_sym_asr] = ACTIONS(1935),
    [aux_sym_pow_operator_token1] = ACTIONS(1937),
    [anon_sym_mod] = ACTIONS(1935),
    [anon_sym_land] = ACTIONS(1935),
    [anon_sym_lor] = ACTIONS(1935),
    [anon_sym_lxor] = ACTIONS(1935),
    [aux_sym_mult_operator_token1] = ACTIONS(1935),
    [aux_sym_add_operator_token1] = ACTIONS(1935),
    [sym_concat_operator] = ACTIONS(1937),
    [sym_rel_operator] = ACTIONS(1935),
    [anon_sym_AMP_AMP] = ACTIONS(1935),
    [anon_sym_or] = ACTIONS(1935),
    [anon_sym_PIPE_PIPE] = ACTIONS(1935),
    [sym_assign_operator] = ACTIONS(1937),
    [sym_let_operator] = ACTIONS(1937),
    [sym_let_and_operator] = ACTIONS(1937),
    [sym__uppercase_identifier] = ACTIONS(1937),
    [aux_sym_directive_token1] = ACTIONS(1935),
    [aux_sym_tag_token1] = ACTIONS(1937),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1216)] = {
    [sym_attribute] = STATE(1216),
    [ts_builtin_sym_end] = ACTIONS(1905),
    [sym__lowercase_identifier] = ACTIONS(1903),
    [anon_sym_SEMI_SEMI] = ACTIONS(1905),
    [anon_sym_let] = ACTIONS(1903),
    [anon_sym_and] = ACTIONS(1903),
    [anon_sym_TILDE] = ACTIONS(1903),
    [anon_sym_QMARK] = ACTIONS(1903),
    [anon_sym_LPAREN] = ACTIONS(1905),
    [anon_sym_external] = ACTIONS(1903),
    [anon_sym_type] = ACTIONS(1903),
    [anon_sym_COMMA] = ACTIONS(1905),
    [anon_sym_PIPE] = ACTIONS(1903),
    [anon_sym_DOT] = ACTIONS(1905),
    [anon_sym_LBRACK] = ACTIONS(1903),
    [anon_sym_true] = ACTIONS(1903),
    [anon_sym_false] = ACTIONS(1903),
    [anon_sym_LBRACE] = ACTIONS(1903),
    [anon_sym_SEMI] = ACTIONS(1903),
    [anon_sym_exception] = ACTIONS(1903),
    [anon_sym_module] = ACTIONS(1903),
    [anon_sym_open] = ACTIONS(1903),
    [anon_sym_include] = ACTIONS(1903),
    [anon_sym_class] = ACTIONS(1903),
    [anon_sym_val] = ACTIONS(1903),
    [anon_sym_object] = ACTIONS(1903),
    [anon_sym_in] = ACTIONS(1903),
    [anon_sym_AMP] = ACTIONS(1903),
    [anon_sym_POUND] = ACTIONS(1903),
    [anon_sym_COLON_COLON] = ACTIONS(1905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1905),
    [anon_sym_LT_DASH] = ACTIONS(1903),
    [anon_sym_else] = ACTIONS(1903),
    [anon_sym_new] = ACTIONS(1903),
    [anon_sym_LBRACE_LT] = ACTIONS(1905),
    [anon_sym_begin] = ACTIONS(1903),
    [sym_ocamlyacc_value] = ACTIONS(1905),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1903),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1905),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1903),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1903),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1905),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1905),
    [aux_sym_number_token1] = ACTIONS(1905),
    [aux_sym_signed_number_token1] = ACTIONS(1903),
    [anon_sym_SQUOTE] = ACTIONS(1905),
    [anon_sym_DQUOTE] = ACTIONS(1905),
    [sym_prefix_operator] = ACTIONS(1905),
    [aux_sym_sign_operator_token1] = ACTIONS(1903),
    [sym_hash_operator] = ACTIONS(1905),
    [anon_sym_lsl] = ACTIONS(1903),
    [anon_sym_lsr] = ACTIONS(1903),
    [anon_sym_asr] = ACTIONS(1903),
    [aux_sym_pow_operator_token1] = ACTIONS(1905),
    [anon_sym_mod] = ACTIONS(1903),
    [anon_sym_land] = ACTIONS(1903),
    [anon_sym_lor] = ACTIONS(1903),
    [anon_sym_lxor] = ACTIONS(1903),
    [aux_sym_mult_operator_token1] = ACTIONS(1903),
    [aux_sym_add_operator_token1] = ACTIONS(1903),
    [sym_concat_operator] = ACTIONS(1905),
    [sym_rel_operator] = ACTIONS(1903),
    [anon_sym_AMP_AMP] = ACTIONS(1903),
    [anon_sym_or] = ACTIONS(1903),
    [anon_sym_PIPE_PIPE] = ACTIONS(1903),
    [sym_assign_operator] = ACTIONS(1905),
    [sym_let_operator] = ACTIONS(1905),
    [sym_let_and_operator] = ACTIONS(1905),
    [sym__uppercase_identifier] = ACTIONS(1905),
    [aux_sym_directive_token1] = ACTIONS(1903),
    [aux_sym_tag_token1] = ACTIONS(1905),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1217)] = {
    [sym_attribute] = STATE(1217),
    [sym__lowercase_identifier] = ACTIONS(1947),
    [anon_sym_SEMI_SEMI] = ACTIONS(1949),
    [anon_sym_let] = ACTIONS(1947),
    [anon_sym_and] = ACTIONS(1947),
    [anon_sym_TILDE] = ACTIONS(1947),
    [anon_sym_QMARK] = ACTIONS(1947),
    [anon_sym_LPAREN] = ACTIONS(1949),
    [anon_sym_external] = ACTIONS(1947),
    [anon_sym_type] = ACTIONS(1947),
    [anon_sym_COMMA] = ACTIONS(1949),
    [anon_sym_PIPE] = ACTIONS(1947),
    [anon_sym_DOT] = ACTIONS(1949),
    [anon_sym_LBRACK] = ACTIONS(1947),
    [anon_sym_RBRACK] = ACTIONS(1949),
    [anon_sym_true] = ACTIONS(1947),
    [anon_sym_false] = ACTIONS(1947),
    [anon_sym_LBRACE] = ACTIONS(1947),
    [anon_sym_SEMI] = ACTIONS(1947),
    [anon_sym_exception] = ACTIONS(1947),
    [anon_sym_module] = ACTIONS(1947),
    [anon_sym_open] = ACTIONS(1947),
    [anon_sym_include] = ACTIONS(1947),
    [anon_sym_class] = ACTIONS(1947),
    [anon_sym_val] = ACTIONS(1947),
    [anon_sym_end] = ACTIONS(1947),
    [anon_sym_object] = ACTIONS(1947),
    [anon_sym_AMP] = ACTIONS(1947),
    [anon_sym_POUND] = ACTIONS(1947),
    [anon_sym_COLON_COLON] = ACTIONS(1949),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1949),
    [anon_sym_LT_DASH] = ACTIONS(2544),
    [anon_sym_else] = ACTIONS(1947),
    [anon_sym_new] = ACTIONS(1947),
    [anon_sym_LBRACE_LT] = ACTIONS(1949),
    [anon_sym_begin] = ACTIONS(1947),
    [sym_ocamlyacc_value] = ACTIONS(1949),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1947),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1949),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1949),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1949),
    [aux_sym_number_token1] = ACTIONS(1949),
    [aux_sym_signed_number_token1] = ACTIONS(1947),
    [anon_sym_SQUOTE] = ACTIONS(1949),
    [anon_sym_DQUOTE] = ACTIONS(1949),
    [sym_prefix_operator] = ACTIONS(1949),
    [aux_sym_sign_operator_token1] = ACTIONS(1947),
    [sym_hash_operator] = ACTIONS(1949),
    [anon_sym_lsl] = ACTIONS(1947),
    [anon_sym_lsr] = ACTIONS(1947),
    [anon_sym_asr] = ACTIONS(1947),
    [aux_sym_pow_operator_token1] = ACTIONS(1949),
    [anon_sym_mod] = ACTIONS(1947),
    [anon_sym_land] = ACTIONS(1947),
    [anon_sym_lor] = ACTIONS(1947),
    [anon_sym_lxor] = ACTIONS(1947),
    [aux_sym_mult_operator_token1] = ACTIONS(1947),
    [aux_sym_add_operator_token1] = ACTIONS(1947),
    [sym_concat_operator] = ACTIONS(1949),
    [sym_rel_operator] = ACTIONS(1947),
    [anon_sym_AMP_AMP] = ACTIONS(1947),
    [anon_sym_or] = ACTIONS(1947),
    [anon_sym_PIPE_PIPE] = ACTIONS(1947),
    [sym_assign_operator] = ACTIONS(1949),
    [sym_let_operator] = ACTIONS(1949),
    [sym_let_and_operator] = ACTIONS(1949),
    [sym__uppercase_identifier] = ACTIONS(1949),
    [aux_sym_directive_token1] = ACTIONS(1947),
    [aux_sym_tag_token1] = ACTIONS(1949),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1218)] = {
    [sym_attribute] = STATE(1218),
    [sym__lowercase_identifier] = ACTIONS(1923),
    [anon_sym_SEMI_SEMI] = ACTIONS(1925),
    [anon_sym_let] = ACTIONS(1923),
    [anon_sym_and] = ACTIONS(1923),
    [anon_sym_TILDE] = ACTIONS(1923),
    [anon_sym_QMARK] = ACTIONS(1923),
    [anon_sym_LPAREN] = ACTIONS(1925),
    [anon_sym_external] = ACTIONS(1923),
    [anon_sym_type] = ACTIONS(1923),
    [anon_sym_COMMA] = ACTIONS(1925),
    [anon_sym_PIPE] = ACTIONS(1923),
    [anon_sym_DOT] = ACTIONS(1925),
    [anon_sym_LBRACK] = ACTIONS(1923),
    [anon_sym_RBRACK] = ACTIONS(1925),
    [anon_sym_true] = ACTIONS(1923),
    [anon_sym_false] = ACTIONS(1923),
    [anon_sym_LBRACE] = ACTIONS(1923),
    [anon_sym_SEMI] = ACTIONS(1923),
    [anon_sym_exception] = ACTIONS(1923),
    [anon_sym_module] = ACTIONS(1923),
    [anon_sym_open] = ACTIONS(1923),
    [anon_sym_include] = ACTIONS(1923),
    [anon_sym_class] = ACTIONS(1923),
    [anon_sym_val] = ACTIONS(1923),
    [anon_sym_end] = ACTIONS(1923),
    [anon_sym_object] = ACTIONS(1923),
    [anon_sym_in] = ACTIONS(1923),
    [anon_sym_AMP] = ACTIONS(1923),
    [anon_sym_POUND] = ACTIONS(1923),
    [anon_sym_COLON_COLON] = ACTIONS(1925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1925),
    [anon_sym_else] = ACTIONS(1923),
    [anon_sym_new] = ACTIONS(1923),
    [anon_sym_LBRACE_LT] = ACTIONS(1925),
    [anon_sym_begin] = ACTIONS(1923),
    [sym_ocamlyacc_value] = ACTIONS(1925),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1923),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1925),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1923),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1923),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1925),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1925),
    [aux_sym_number_token1] = ACTIONS(1925),
    [aux_sym_signed_number_token1] = ACTIONS(1923),
    [anon_sym_SQUOTE] = ACTIONS(1925),
    [anon_sym_DQUOTE] = ACTIONS(1925),
    [sym_prefix_operator] = ACTIONS(1925),
    [aux_sym_sign_operator_token1] = ACTIONS(1923),
    [sym_hash_operator] = ACTIONS(1925),
    [anon_sym_lsl] = ACTIONS(1923),
    [anon_sym_lsr] = ACTIONS(1923),
    [anon_sym_asr] = ACTIONS(1923),
    [aux_sym_pow_operator_token1] = ACTIONS(1925),
    [anon_sym_mod] = ACTIONS(1923),
    [anon_sym_land] = ACTIONS(1923),
    [anon_sym_lor] = ACTIONS(1923),
    [anon_sym_lxor] = ACTIONS(1923),
    [aux_sym_mult_operator_token1] = ACTIONS(1923),
    [aux_sym_add_operator_token1] = ACTIONS(1923),
    [sym_concat_operator] = ACTIONS(1925),
    [sym_rel_operator] = ACTIONS(1923),
    [anon_sym_AMP_AMP] = ACTIONS(1923),
    [anon_sym_or] = ACTIONS(1923),
    [anon_sym_PIPE_PIPE] = ACTIONS(1923),
    [sym_assign_operator] = ACTIONS(1925),
    [sym_let_operator] = ACTIONS(1925),
    [sym_let_and_operator] = ACTIONS(1925),
    [sym__uppercase_identifier] = ACTIONS(1925),
    [aux_sym_directive_token1] = ACTIONS(1923),
    [aux_sym_tag_token1] = ACTIONS(1925),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1219)] = {
    [sym_parameter] = STATE(2862),
    [sym__class_typed] = STATE(4668),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1219),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(7381),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(3191),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(4340),
    [aux_sym_let_binding_repeat1] = STATE(1221),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(2379),
    [anon_sym_and] = ACTIONS(2381),
    [anon_sym_EQ] = ACTIONS(2546),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_QMARK] = ACTIONS(2524),
    [anon_sym_LPAREN] = ACTIONS(2526),
    [anon_sym_external] = ACTIONS(2381),
    [anon_sym_type] = ACTIONS(2381),
    [anon_sym_COLON2] = ACTIONS(2528),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_RBRACK] = ACTIONS(2379),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_exception] = ACTIONS(2381),
    [anon_sym_module] = ACTIONS(2381),
    [anon_sym_open] = ACTIONS(2381),
    [anon_sym_include] = ACTIONS(2381),
    [anon_sym_class] = ACTIONS(2381),
    [anon_sym_val] = ACTIONS(2381),
    [anon_sym_end] = ACTIONS(2381),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2379),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2379),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2379),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(2530),
    [aux_sym_tag_token1] = ACTIONS(2532),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1220)] = {
    [sym_attribute] = STATE(1220),
    [sym__lowercase_identifier] = ACTIONS(2059),
    [anon_sym_SEMI_SEMI] = ACTIONS(2061),
    [anon_sym_let] = ACTIONS(2059),
    [anon_sym_and] = ACTIONS(2059),
    [anon_sym_TILDE] = ACTIONS(2059),
    [anon_sym_QMARK] = ACTIONS(2059),
    [anon_sym_LPAREN] = ACTIONS(2061),
    [anon_sym_external] = ACTIONS(2059),
    [anon_sym_type] = ACTIONS(2059),
    [anon_sym_COMMA] = ACTIONS(2061),
    [anon_sym_PIPE] = ACTIONS(2059),
    [anon_sym_DOT] = ACTIONS(2061),
    [anon_sym_LBRACK] = ACTIONS(2059),
    [anon_sym_RBRACK] = ACTIONS(2061),
    [anon_sym_true] = ACTIONS(2059),
    [anon_sym_false] = ACTIONS(2059),
    [anon_sym_LBRACE] = ACTIONS(2059),
    [anon_sym_SEMI] = ACTIONS(2059),
    [anon_sym_exception] = ACTIONS(2059),
    [anon_sym_module] = ACTIONS(2059),
    [anon_sym_open] = ACTIONS(2059),
    [anon_sym_include] = ACTIONS(2059),
    [anon_sym_class] = ACTIONS(2059),
    [anon_sym_val] = ACTIONS(2059),
    [anon_sym_end] = ACTIONS(2059),
    [anon_sym_object] = ACTIONS(2059),
    [anon_sym_in] = ACTIONS(2059),
    [anon_sym_AMP] = ACTIONS(2059),
    [anon_sym_POUND] = ACTIONS(2059),
    [anon_sym_COLON_COLON] = ACTIONS(2061),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2061),
    [anon_sym_else] = ACTIONS(2059),
    [anon_sym_new] = ACTIONS(2059),
    [anon_sym_LBRACE_LT] = ACTIONS(2061),
    [anon_sym_begin] = ACTIONS(2059),
    [sym_ocamlyacc_value] = ACTIONS(2061),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2059),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2061),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2061),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2061),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2059),
    [anon_sym_SQUOTE] = ACTIONS(2061),
    [anon_sym_DQUOTE] = ACTIONS(2061),
    [sym_prefix_operator] = ACTIONS(2061),
    [aux_sym_sign_operator_token1] = ACTIONS(2059),
    [sym_hash_operator] = ACTIONS(2061),
    [anon_sym_lsl] = ACTIONS(2059),
    [anon_sym_lsr] = ACTIONS(2059),
    [anon_sym_asr] = ACTIONS(2059),
    [aux_sym_pow_operator_token1] = ACTIONS(2061),
    [anon_sym_mod] = ACTIONS(2059),
    [anon_sym_land] = ACTIONS(2059),
    [anon_sym_lor] = ACTIONS(2059),
    [anon_sym_lxor] = ACTIONS(2059),
    [aux_sym_mult_operator_token1] = ACTIONS(2059),
    [aux_sym_add_operator_token1] = ACTIONS(2059),
    [sym_concat_operator] = ACTIONS(2061),
    [sym_rel_operator] = ACTIONS(2059),
    [anon_sym_AMP_AMP] = ACTIONS(2059),
    [anon_sym_or] = ACTIONS(2059),
    [anon_sym_PIPE_PIPE] = ACTIONS(2059),
    [sym_assign_operator] = ACTIONS(2061),
    [sym_let_operator] = ACTIONS(2061),
    [sym_let_and_operator] = ACTIONS(2061),
    [sym__uppercase_identifier] = ACTIONS(2061),
    [aux_sym_directive_token1] = ACTIONS(2059),
    [aux_sym_tag_token1] = ACTIONS(2061),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1221)] = {
    [sym_parameter] = STATE(2862),
    [sym__class_typed] = STATE(4672),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1221),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(7381),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(3191),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(4230),
    [aux_sym_let_binding_repeat1] = STATE(1477),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(2385),
    [anon_sym_and] = ACTIONS(2387),
    [anon_sym_EQ] = ACTIONS(2548),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_QMARK] = ACTIONS(2524),
    [anon_sym_LPAREN] = ACTIONS(2526),
    [anon_sym_external] = ACTIONS(2387),
    [anon_sym_type] = ACTIONS(2387),
    [anon_sym_COLON2] = ACTIONS(2528),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_RBRACK] = ACTIONS(2385),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_exception] = ACTIONS(2387),
    [anon_sym_module] = ACTIONS(2387),
    [anon_sym_open] = ACTIONS(2387),
    [anon_sym_include] = ACTIONS(2387),
    [anon_sym_class] = ACTIONS(2387),
    [anon_sym_val] = ACTIONS(2387),
    [anon_sym_end] = ACTIONS(2387),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2385),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2385),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2385),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(2530),
    [aux_sym_tag_token1] = ACTIONS(2532),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1222)] = {
    [sym_parameter] = STATE(2862),
    [sym__class_typed] = STATE(4673),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1222),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(7381),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(3191),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(4241),
    [aux_sym_let_binding_repeat1] = STATE(1229),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(2373),
    [anon_sym_and] = ACTIONS(2375),
    [anon_sym_EQ] = ACTIONS(2550),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_QMARK] = ACTIONS(2524),
    [anon_sym_LPAREN] = ACTIONS(2526),
    [anon_sym_external] = ACTIONS(2375),
    [anon_sym_type] = ACTIONS(2375),
    [anon_sym_COLON2] = ACTIONS(2528),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_RBRACK] = ACTIONS(2373),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_exception] = ACTIONS(2375),
    [anon_sym_module] = ACTIONS(2375),
    [anon_sym_open] = ACTIONS(2375),
    [anon_sym_include] = ACTIONS(2375),
    [anon_sym_class] = ACTIONS(2375),
    [anon_sym_val] = ACTIONS(2375),
    [anon_sym_end] = ACTIONS(2375),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2373),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2373),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2373),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(2530),
    [aux_sym_tag_token1] = ACTIONS(2532),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1223)] = {
    [sym_attribute] = STATE(1223),
    [sym__lowercase_identifier] = ACTIONS(1983),
    [anon_sym_SEMI_SEMI] = ACTIONS(1985),
    [anon_sym_let] = ACTIONS(1983),
    [anon_sym_and] = ACTIONS(1983),
    [anon_sym_TILDE] = ACTIONS(1983),
    [anon_sym_QMARK] = ACTIONS(1983),
    [anon_sym_LPAREN] = ACTIONS(1985),
    [anon_sym_external] = ACTIONS(1983),
    [anon_sym_type] = ACTIONS(1983),
    [anon_sym_COMMA] = ACTIONS(1985),
    [anon_sym_PIPE] = ACTIONS(1983),
    [anon_sym_DOT] = ACTIONS(1985),
    [anon_sym_LBRACK] = ACTIONS(1983),
    [anon_sym_RBRACK] = ACTIONS(1985),
    [anon_sym_true] = ACTIONS(1983),
    [anon_sym_false] = ACTIONS(1983),
    [anon_sym_LBRACE] = ACTIONS(1983),
    [anon_sym_SEMI] = ACTIONS(1983),
    [anon_sym_exception] = ACTIONS(1983),
    [anon_sym_module] = ACTIONS(1983),
    [anon_sym_open] = ACTIONS(1983),
    [anon_sym_include] = ACTIONS(1983),
    [anon_sym_class] = ACTIONS(1983),
    [anon_sym_val] = ACTIONS(1983),
    [anon_sym_end] = ACTIONS(1983),
    [anon_sym_object] = ACTIONS(1983),
    [anon_sym_in] = ACTIONS(1983),
    [anon_sym_AMP] = ACTIONS(1983),
    [anon_sym_POUND] = ACTIONS(1983),
    [anon_sym_COLON_COLON] = ACTIONS(1985),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1985),
    [anon_sym_else] = ACTIONS(1983),
    [anon_sym_new] = ACTIONS(1983),
    [anon_sym_LBRACE_LT] = ACTIONS(1985),
    [anon_sym_begin] = ACTIONS(1983),
    [sym_ocamlyacc_value] = ACTIONS(1985),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1983),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1985),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1983),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1983),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1985),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1985),
    [aux_sym_number_token1] = ACTIONS(1985),
    [aux_sym_signed_number_token1] = ACTIONS(1983),
    [anon_sym_SQUOTE] = ACTIONS(1985),
    [anon_sym_DQUOTE] = ACTIONS(1985),
    [sym_prefix_operator] = ACTIONS(1985),
    [aux_sym_sign_operator_token1] = ACTIONS(1983),
    [sym_hash_operator] = ACTIONS(1985),
    [anon_sym_lsl] = ACTIONS(1983),
    [anon_sym_lsr] = ACTIONS(1983),
    [anon_sym_asr] = ACTIONS(1983),
    [aux_sym_pow_operator_token1] = ACTIONS(1985),
    [anon_sym_mod] = ACTIONS(1983),
    [anon_sym_land] = ACTIONS(1983),
    [anon_sym_lor] = ACTIONS(1983),
    [anon_sym_lxor] = ACTIONS(1983),
    [aux_sym_mult_operator_token1] = ACTIONS(1983),
    [aux_sym_add_operator_token1] = ACTIONS(1983),
    [sym_concat_operator] = ACTIONS(1985),
    [sym_rel_operator] = ACTIONS(1983),
    [anon_sym_AMP_AMP] = ACTIONS(1983),
    [anon_sym_or] = ACTIONS(1983),
    [anon_sym_PIPE_PIPE] = ACTIONS(1983),
    [sym_assign_operator] = ACTIONS(1985),
    [sym_let_operator] = ACTIONS(1985),
    [sym_let_and_operator] = ACTIONS(1985),
    [sym__uppercase_identifier] = ACTIONS(1985),
    [aux_sym_directive_token1] = ACTIONS(1983),
    [aux_sym_tag_token1] = ACTIONS(1985),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1224)] = {
    [sym_attribute] = STATE(1224),
    [sym__lowercase_identifier] = ACTIONS(2063),
    [anon_sym_SEMI_SEMI] = ACTIONS(2065),
    [anon_sym_let] = ACTIONS(2063),
    [anon_sym_and] = ACTIONS(2063),
    [anon_sym_TILDE] = ACTIONS(2063),
    [anon_sym_QMARK] = ACTIONS(2063),
    [anon_sym_LPAREN] = ACTIONS(2065),
    [anon_sym_external] = ACTIONS(2063),
    [anon_sym_type] = ACTIONS(2063),
    [anon_sym_COMMA] = ACTIONS(2065),
    [anon_sym_PIPE] = ACTIONS(2063),
    [anon_sym_DOT] = ACTIONS(2065),
    [anon_sym_LBRACK] = ACTIONS(2063),
    [anon_sym_RBRACK] = ACTIONS(2065),
    [anon_sym_true] = ACTIONS(2063),
    [anon_sym_false] = ACTIONS(2063),
    [anon_sym_LBRACE] = ACTIONS(2063),
    [anon_sym_SEMI] = ACTIONS(2063),
    [anon_sym_exception] = ACTIONS(2063),
    [anon_sym_module] = ACTIONS(2063),
    [anon_sym_open] = ACTIONS(2063),
    [anon_sym_include] = ACTIONS(2063),
    [anon_sym_class] = ACTIONS(2063),
    [anon_sym_val] = ACTIONS(2063),
    [anon_sym_end] = ACTIONS(2063),
    [anon_sym_object] = ACTIONS(2063),
    [anon_sym_in] = ACTIONS(2063),
    [anon_sym_AMP] = ACTIONS(2063),
    [anon_sym_POUND] = ACTIONS(2063),
    [anon_sym_COLON_COLON] = ACTIONS(2065),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2065),
    [anon_sym_else] = ACTIONS(2063),
    [anon_sym_new] = ACTIONS(2063),
    [anon_sym_LBRACE_LT] = ACTIONS(2065),
    [anon_sym_begin] = ACTIONS(2063),
    [sym_ocamlyacc_value] = ACTIONS(2065),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2063),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2065),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2063),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2063),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2065),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2065),
    [aux_sym_number_token1] = ACTIONS(2065),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2065),
    [sym_prefix_operator] = ACTIONS(2065),
    [aux_sym_sign_operator_token1] = ACTIONS(2063),
    [sym_hash_operator] = ACTIONS(2065),
    [anon_sym_lsl] = ACTIONS(2063),
    [anon_sym_lsr] = ACTIONS(2063),
    [anon_sym_asr] = ACTIONS(2063),
    [aux_sym_pow_operator_token1] = ACTIONS(2065),
    [anon_sym_mod] = ACTIONS(2063),
    [anon_sym_land] = ACTIONS(2063),
    [anon_sym_lor] = ACTIONS(2063),
    [anon_sym_lxor] = ACTIONS(2063),
    [aux_sym_mult_operator_token1] = ACTIONS(2063),
    [aux_sym_add_operator_token1] = ACTIONS(2063),
    [sym_concat_operator] = ACTIONS(2065),
    [sym_rel_operator] = ACTIONS(2063),
    [anon_sym_AMP_AMP] = ACTIONS(2063),
    [anon_sym_or] = ACTIONS(2063),
    [anon_sym_PIPE_PIPE] = ACTIONS(2063),
    [sym_assign_operator] = ACTIONS(2065),
    [sym_let_operator] = ACTIONS(2065),
    [sym_let_and_operator] = ACTIONS(2065),
    [sym__uppercase_identifier] = ACTIONS(2065),
    [aux_sym_directive_token1] = ACTIONS(2063),
    [aux_sym_tag_token1] = ACTIONS(2065),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1225)] = {
    [sym_attribute] = STATE(1225),
    [ts_builtin_sym_end] = ACTIONS(2159),
    [sym__lowercase_identifier] = ACTIONS(2157),
    [anon_sym_SEMI_SEMI] = ACTIONS(2159),
    [anon_sym_let] = ACTIONS(2157),
    [anon_sym_and] = ACTIONS(2157),
    [anon_sym_TILDE] = ACTIONS(2157),
    [anon_sym_QMARK] = ACTIONS(2157),
    [anon_sym_LPAREN] = ACTIONS(2159),
    [anon_sym_external] = ACTIONS(2157),
    [anon_sym_type] = ACTIONS(2157),
    [anon_sym_COMMA] = ACTIONS(2159),
    [anon_sym_PIPE] = ACTIONS(2157),
    [anon_sym_DOT] = ACTIONS(2159),
    [anon_sym_LBRACK] = ACTIONS(2157),
    [anon_sym_true] = ACTIONS(2157),
    [anon_sym_false] = ACTIONS(2157),
    [anon_sym_LBRACE] = ACTIONS(2157),
    [anon_sym_SEMI] = ACTIONS(2157),
    [anon_sym_exception] = ACTIONS(2157),
    [anon_sym_module] = ACTIONS(2157),
    [anon_sym_open] = ACTIONS(2157),
    [anon_sym_include] = ACTIONS(2157),
    [anon_sym_class] = ACTIONS(2157),
    [anon_sym_val] = ACTIONS(2157),
    [anon_sym_object] = ACTIONS(2157),
    [anon_sym_in] = ACTIONS(2157),
    [anon_sym_AMP] = ACTIONS(2157),
    [anon_sym_POUND] = ACTIONS(2157),
    [anon_sym_COLON_COLON] = ACTIONS(2159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2159),
    [anon_sym_LT_DASH] = ACTIONS(2317),
    [anon_sym_else] = ACTIONS(2157),
    [anon_sym_new] = ACTIONS(2157),
    [anon_sym_LBRACE_LT] = ACTIONS(2159),
    [anon_sym_begin] = ACTIONS(2157),
    [sym_ocamlyacc_value] = ACTIONS(2159),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2157),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2159),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2159),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2159),
    [aux_sym_number_token1] = ACTIONS(2159),
    [aux_sym_signed_number_token1] = ACTIONS(2157),
    [anon_sym_SQUOTE] = ACTIONS(2159),
    [anon_sym_DQUOTE] = ACTIONS(2159),
    [sym_prefix_operator] = ACTIONS(2159),
    [aux_sym_sign_operator_token1] = ACTIONS(2157),
    [sym_hash_operator] = ACTIONS(2159),
    [anon_sym_lsl] = ACTIONS(2157),
    [anon_sym_lsr] = ACTIONS(2157),
    [anon_sym_asr] = ACTIONS(2157),
    [aux_sym_pow_operator_token1] = ACTIONS(2159),
    [anon_sym_mod] = ACTIONS(2157),
    [anon_sym_land] = ACTIONS(2157),
    [anon_sym_lor] = ACTIONS(2157),
    [anon_sym_lxor] = ACTIONS(2157),
    [aux_sym_mult_operator_token1] = ACTIONS(2157),
    [aux_sym_add_operator_token1] = ACTIONS(2157),
    [sym_concat_operator] = ACTIONS(2159),
    [sym_rel_operator] = ACTIONS(2157),
    [anon_sym_AMP_AMP] = ACTIONS(2157),
    [anon_sym_or] = ACTIONS(2157),
    [anon_sym_PIPE_PIPE] = ACTIONS(2157),
    [sym_assign_operator] = ACTIONS(2159),
    [sym_let_operator] = ACTIONS(2159),
    [sym_let_and_operator] = ACTIONS(2159),
    [sym__uppercase_identifier] = ACTIONS(2159),
    [aux_sym_directive_token1] = ACTIONS(2157),
    [aux_sym_tag_token1] = ACTIONS(2159),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1226)] = {
    [sym_attribute] = STATE(1226),
    [sym__lowercase_identifier] = ACTIONS(1987),
    [anon_sym_SEMI_SEMI] = ACTIONS(1989),
    [anon_sym_let] = ACTIONS(1987),
    [anon_sym_and] = ACTIONS(1987),
    [anon_sym_TILDE] = ACTIONS(1987),
    [anon_sym_QMARK] = ACTIONS(1987),
    [anon_sym_LPAREN] = ACTIONS(1989),
    [anon_sym_external] = ACTIONS(1987),
    [anon_sym_type] = ACTIONS(1987),
    [anon_sym_COMMA] = ACTIONS(1989),
    [anon_sym_PIPE] = ACTIONS(1987),
    [anon_sym_DOT] = ACTIONS(1989),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_RBRACK] = ACTIONS(1989),
    [anon_sym_true] = ACTIONS(1987),
    [anon_sym_false] = ACTIONS(1987),
    [anon_sym_LBRACE] = ACTIONS(1987),
    [anon_sym_SEMI] = ACTIONS(1987),
    [anon_sym_exception] = ACTIONS(1987),
    [anon_sym_module] = ACTIONS(1987),
    [anon_sym_open] = ACTIONS(1987),
    [anon_sym_include] = ACTIONS(1987),
    [anon_sym_class] = ACTIONS(1987),
    [anon_sym_val] = ACTIONS(1987),
    [anon_sym_end] = ACTIONS(1987),
    [anon_sym_object] = ACTIONS(1987),
    [anon_sym_in] = ACTIONS(1987),
    [anon_sym_AMP] = ACTIONS(1987),
    [anon_sym_POUND] = ACTIONS(1987),
    [anon_sym_COLON_COLON] = ACTIONS(1989),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1989),
    [anon_sym_else] = ACTIONS(1987),
    [anon_sym_new] = ACTIONS(1987),
    [anon_sym_LBRACE_LT] = ACTIONS(1989),
    [anon_sym_begin] = ACTIONS(1987),
    [sym_ocamlyacc_value] = ACTIONS(1989),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1987),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1989),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1987),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1987),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1989),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1989),
    [aux_sym_number_token1] = ACTIONS(1989),
    [aux_sym_signed_number_token1] = ACTIONS(1987),
    [anon_sym_SQUOTE] = ACTIONS(1989),
    [anon_sym_DQUOTE] = ACTIONS(1989),
    [sym_prefix_operator] = ACTIONS(1989),
    [aux_sym_sign_operator_token1] = ACTIONS(1987),
    [sym_hash_operator] = ACTIONS(1989),
    [anon_sym_lsl] = ACTIONS(1987),
    [anon_sym_lsr] = ACTIONS(1987),
    [anon_sym_asr] = ACTIONS(1987),
    [aux_sym_pow_operator_token1] = ACTIONS(1989),
    [anon_sym_mod] = ACTIONS(1987),
    [anon_sym_land] = ACTIONS(1987),
    [anon_sym_lor] = ACTIONS(1987),
    [anon_sym_lxor] = ACTIONS(1987),
    [aux_sym_mult_operator_token1] = ACTIONS(1987),
    [aux_sym_add_operator_token1] = ACTIONS(1987),
    [sym_concat_operator] = ACTIONS(1989),
    [sym_rel_operator] = ACTIONS(1987),
    [anon_sym_AMP_AMP] = ACTIONS(1987),
    [anon_sym_or] = ACTIONS(1987),
    [anon_sym_PIPE_PIPE] = ACTIONS(1987),
    [sym_assign_operator] = ACTIONS(1989),
    [sym_let_operator] = ACTIONS(1989),
    [sym_let_and_operator] = ACTIONS(1989),
    [sym__uppercase_identifier] = ACTIONS(1989),
    [aux_sym_directive_token1] = ACTIONS(1987),
    [aux_sym_tag_token1] = ACTIONS(1989),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1227)] = {
    [sym_attribute] = STATE(1227),
    [sym__lowercase_identifier] = ACTIONS(2067),
    [anon_sym_SEMI_SEMI] = ACTIONS(2069),
    [anon_sym_let] = ACTIONS(2067),
    [anon_sym_and] = ACTIONS(2067),
    [anon_sym_TILDE] = ACTIONS(2067),
    [anon_sym_QMARK] = ACTIONS(2067),
    [anon_sym_LPAREN] = ACTIONS(2069),
    [anon_sym_external] = ACTIONS(2067),
    [anon_sym_type] = ACTIONS(2067),
    [anon_sym_COMMA] = ACTIONS(2069),
    [anon_sym_PIPE] = ACTIONS(2067),
    [anon_sym_DOT] = ACTIONS(2069),
    [anon_sym_LBRACK] = ACTIONS(2067),
    [anon_sym_RBRACK] = ACTIONS(2069),
    [anon_sym_true] = ACTIONS(2067),
    [anon_sym_false] = ACTIONS(2067),
    [anon_sym_LBRACE] = ACTIONS(2067),
    [anon_sym_SEMI] = ACTIONS(2067),
    [anon_sym_exception] = ACTIONS(2067),
    [anon_sym_module] = ACTIONS(2067),
    [anon_sym_open] = ACTIONS(2067),
    [anon_sym_include] = ACTIONS(2067),
    [anon_sym_class] = ACTIONS(2067),
    [anon_sym_val] = ACTIONS(2067),
    [anon_sym_end] = ACTIONS(2067),
    [anon_sym_object] = ACTIONS(2067),
    [anon_sym_in] = ACTIONS(2067),
    [anon_sym_AMP] = ACTIONS(2067),
    [anon_sym_POUND] = ACTIONS(2067),
    [anon_sym_COLON_COLON] = ACTIONS(2069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2069),
    [anon_sym_else] = ACTIONS(2067),
    [anon_sym_new] = ACTIONS(2067),
    [anon_sym_LBRACE_LT] = ACTIONS(2069),
    [anon_sym_begin] = ACTIONS(2067),
    [sym_ocamlyacc_value] = ACTIONS(2069),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2067),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2069),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2067),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2067),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2069),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2069),
    [aux_sym_number_token1] = ACTIONS(2069),
    [aux_sym_signed_number_token1] = ACTIONS(2067),
    [anon_sym_SQUOTE] = ACTIONS(2069),
    [anon_sym_DQUOTE] = ACTIONS(2069),
    [sym_prefix_operator] = ACTIONS(2069),
    [aux_sym_sign_operator_token1] = ACTIONS(2067),
    [sym_hash_operator] = ACTIONS(2069),
    [anon_sym_lsl] = ACTIONS(2067),
    [anon_sym_lsr] = ACTIONS(2067),
    [anon_sym_asr] = ACTIONS(2067),
    [aux_sym_pow_operator_token1] = ACTIONS(2069),
    [anon_sym_mod] = ACTIONS(2067),
    [anon_sym_land] = ACTIONS(2067),
    [anon_sym_lor] = ACTIONS(2067),
    [anon_sym_lxor] = ACTIONS(2067),
    [aux_sym_mult_operator_token1] = ACTIONS(2067),
    [aux_sym_add_operator_token1] = ACTIONS(2067),
    [sym_concat_operator] = ACTIONS(2069),
    [sym_rel_operator] = ACTIONS(2067),
    [anon_sym_AMP_AMP] = ACTIONS(2067),
    [anon_sym_or] = ACTIONS(2067),
    [anon_sym_PIPE_PIPE] = ACTIONS(2067),
    [sym_assign_operator] = ACTIONS(2069),
    [sym_let_operator] = ACTIONS(2069),
    [sym_let_and_operator] = ACTIONS(2069),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_directive_token1] = ACTIONS(2067),
    [aux_sym_tag_token1] = ACTIONS(2069),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1228)] = {
    [sym_attribute] = STATE(1228),
    [sym__lowercase_identifier] = ACTIONS(2087),
    [anon_sym_SEMI_SEMI] = ACTIONS(2089),
    [anon_sym_let] = ACTIONS(2087),
    [anon_sym_and] = ACTIONS(2087),
    [anon_sym_TILDE] = ACTIONS(2087),
    [anon_sym_QMARK] = ACTIONS(2087),
    [anon_sym_LPAREN] = ACTIONS(2089),
    [anon_sym_external] = ACTIONS(2087),
    [anon_sym_type] = ACTIONS(2087),
    [anon_sym_COMMA] = ACTIONS(2089),
    [anon_sym_PIPE] = ACTIONS(2087),
    [anon_sym_DOT] = ACTIONS(2089),
    [anon_sym_LBRACK] = ACTIONS(2087),
    [anon_sym_RBRACK] = ACTIONS(2089),
    [anon_sym_true] = ACTIONS(2087),
    [anon_sym_false] = ACTIONS(2087),
    [anon_sym_LBRACE] = ACTIONS(2087),
    [anon_sym_SEMI] = ACTIONS(2087),
    [anon_sym_exception] = ACTIONS(2087),
    [anon_sym_module] = ACTIONS(2087),
    [anon_sym_open] = ACTIONS(2087),
    [anon_sym_include] = ACTIONS(2087),
    [anon_sym_class] = ACTIONS(2087),
    [anon_sym_val] = ACTIONS(2087),
    [anon_sym_end] = ACTIONS(2087),
    [anon_sym_object] = ACTIONS(2087),
    [anon_sym_in] = ACTIONS(2087),
    [anon_sym_AMP] = ACTIONS(2087),
    [anon_sym_POUND] = ACTIONS(2087),
    [anon_sym_COLON_COLON] = ACTIONS(2089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2089),
    [anon_sym_else] = ACTIONS(2087),
    [anon_sym_new] = ACTIONS(2087),
    [anon_sym_LBRACE_LT] = ACTIONS(2089),
    [anon_sym_begin] = ACTIONS(2087),
    [sym_ocamlyacc_value] = ACTIONS(2089),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2087),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2089),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2087),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2087),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2089),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2089),
    [aux_sym_number_token1] = ACTIONS(2089),
    [aux_sym_signed_number_token1] = ACTIONS(2087),
    [anon_sym_SQUOTE] = ACTIONS(2089),
    [anon_sym_DQUOTE] = ACTIONS(2089),
    [sym_prefix_operator] = ACTIONS(2089),
    [aux_sym_sign_operator_token1] = ACTIONS(2087),
    [sym_hash_operator] = ACTIONS(2089),
    [anon_sym_lsl] = ACTIONS(2087),
    [anon_sym_lsr] = ACTIONS(2087),
    [anon_sym_asr] = ACTIONS(2087),
    [aux_sym_pow_operator_token1] = ACTIONS(2089),
    [anon_sym_mod] = ACTIONS(2087),
    [anon_sym_land] = ACTIONS(2087),
    [anon_sym_lor] = ACTIONS(2087),
    [anon_sym_lxor] = ACTIONS(2087),
    [aux_sym_mult_operator_token1] = ACTIONS(2087),
    [aux_sym_add_operator_token1] = ACTIONS(2087),
    [sym_concat_operator] = ACTIONS(2089),
    [sym_rel_operator] = ACTIONS(2087),
    [anon_sym_AMP_AMP] = ACTIONS(2087),
    [anon_sym_or] = ACTIONS(2087),
    [anon_sym_PIPE_PIPE] = ACTIONS(2087),
    [sym_assign_operator] = ACTIONS(2089),
    [sym_let_operator] = ACTIONS(2089),
    [sym_let_and_operator] = ACTIONS(2089),
    [sym__uppercase_identifier] = ACTIONS(2089),
    [aux_sym_directive_token1] = ACTIONS(2087),
    [aux_sym_tag_token1] = ACTIONS(2089),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1229)] = {
    [sym_parameter] = STATE(2862),
    [sym__class_typed] = STATE(4676),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1229),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(7381),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(3191),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(4300),
    [aux_sym_let_binding_repeat1] = STATE(1477),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(2391),
    [anon_sym_and] = ACTIONS(2393),
    [anon_sym_EQ] = ACTIONS(2552),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_QMARK] = ACTIONS(2524),
    [anon_sym_LPAREN] = ACTIONS(2526),
    [anon_sym_external] = ACTIONS(2393),
    [anon_sym_type] = ACTIONS(2393),
    [anon_sym_COLON2] = ACTIONS(2528),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_RBRACK] = ACTIONS(2391),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_exception] = ACTIONS(2393),
    [anon_sym_module] = ACTIONS(2393),
    [anon_sym_open] = ACTIONS(2393),
    [anon_sym_include] = ACTIONS(2393),
    [anon_sym_class] = ACTIONS(2393),
    [anon_sym_val] = ACTIONS(2393),
    [anon_sym_end] = ACTIONS(2393),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2391),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2391),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2391),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(2530),
    [aux_sym_tag_token1] = ACTIONS(2532),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1230)] = {
    [sym_parameter] = STATE(2862),
    [sym__class_typed] = STATE(4679),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1230),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(7381),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(3191),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(4302),
    [aux_sym_let_binding_repeat1] = STATE(1211),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(2397),
    [anon_sym_and] = ACTIONS(2399),
    [anon_sym_EQ] = ACTIONS(2554),
    [anon_sym_TILDE] = ACTIONS(2524),
    [anon_sym_QMARK] = ACTIONS(2524),
    [anon_sym_LPAREN] = ACTIONS(2526),
    [anon_sym_external] = ACTIONS(2399),
    [anon_sym_type] = ACTIONS(2399),
    [anon_sym_COLON2] = ACTIONS(2528),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_RBRACK] = ACTIONS(2397),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_exception] = ACTIONS(2399),
    [anon_sym_module] = ACTIONS(2399),
    [anon_sym_open] = ACTIONS(2399),
    [anon_sym_include] = ACTIONS(2399),
    [anon_sym_class] = ACTIONS(2399),
    [anon_sym_val] = ACTIONS(2399),
    [anon_sym_end] = ACTIONS(2399),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2397),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2397),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2397),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(2530),
    [aux_sym_tag_token1] = ACTIONS(2532),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1231)] = {
    [sym_attribute] = STATE(1231),
    [sym__lowercase_identifier] = ACTIONS(2091),
    [anon_sym_SEMI_SEMI] = ACTIONS(2093),
    [anon_sym_let] = ACTIONS(2091),
    [anon_sym_and] = ACTIONS(2091),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_external] = ACTIONS(2091),
    [anon_sym_type] = ACTIONS(2091),
    [anon_sym_COMMA] = ACTIONS(2093),
    [anon_sym_PIPE] = ACTIONS(2091),
    [anon_sym_DOT] = ACTIONS(2093),
    [anon_sym_LBRACK] = ACTIONS(2091),
    [anon_sym_RBRACK] = ACTIONS(2093),
    [anon_sym_true] = ACTIONS(2091),
    [anon_sym_false] = ACTIONS(2091),
    [anon_sym_LBRACE] = ACTIONS(2091),
    [anon_sym_SEMI] = ACTIONS(2091),
    [anon_sym_exception] = ACTIONS(2091),
    [anon_sym_module] = ACTIONS(2091),
    [anon_sym_open] = ACTIONS(2091),
    [anon_sym_include] = ACTIONS(2091),
    [anon_sym_class] = ACTIONS(2091),
    [anon_sym_val] = ACTIONS(2091),
    [anon_sym_end] = ACTIONS(2091),
    [anon_sym_object] = ACTIONS(2091),
    [anon_sym_in] = ACTIONS(2091),
    [anon_sym_AMP] = ACTIONS(2091),
    [anon_sym_POUND] = ACTIONS(2091),
    [anon_sym_COLON_COLON] = ACTIONS(2093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2093),
    [anon_sym_else] = ACTIONS(2091),
    [anon_sym_new] = ACTIONS(2091),
    [anon_sym_LBRACE_LT] = ACTIONS(2093),
    [anon_sym_begin] = ACTIONS(2091),
    [sym_ocamlyacc_value] = ACTIONS(2093),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2091),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2093),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2091),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2091),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2093),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2093),
    [aux_sym_number_token1] = ACTIONS(2093),
    [aux_sym_signed_number_token1] = ACTIONS(2091),
    [anon_sym_SQUOTE] = ACTIONS(2093),
    [anon_sym_DQUOTE] = ACTIONS(2093),
    [sym_prefix_operator] = ACTIONS(2093),
    [aux_sym_sign_operator_token1] = ACTIONS(2091),
    [sym_hash_operator] = ACTIONS(2093),
    [anon_sym_lsl] = ACTIONS(2091),
    [anon_sym_lsr] = ACTIONS(2091),
    [anon_sym_asr] = ACTIONS(2091),
    [aux_sym_pow_operator_token1] = ACTIONS(2093),
    [anon_sym_mod] = ACTIONS(2091),
    [anon_sym_land] = ACTIONS(2091),
    [anon_sym_lor] = ACTIONS(2091),
    [anon_sym_lxor] = ACTIONS(2091),
    [aux_sym_mult_operator_token1] = ACTIONS(2091),
    [aux_sym_add_operator_token1] = ACTIONS(2091),
    [sym_concat_operator] = ACTIONS(2093),
    [sym_rel_operator] = ACTIONS(2091),
    [anon_sym_AMP_AMP] = ACTIONS(2091),
    [anon_sym_or] = ACTIONS(2091),
    [anon_sym_PIPE_PIPE] = ACTIONS(2091),
    [sym_assign_operator] = ACTIONS(2093),
    [sym_let_operator] = ACTIONS(2093),
    [sym_let_and_operator] = ACTIONS(2093),
    [sym__uppercase_identifier] = ACTIONS(2093),
    [aux_sym_directive_token1] = ACTIONS(2091),
    [aux_sym_tag_token1] = ACTIONS(2093),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1232)] = {
    [sym_attribute] = STATE(1232),
    [sym__lowercase_identifier] = ACTIONS(2095),
    [anon_sym_SEMI_SEMI] = ACTIONS(2097),
    [anon_sym_let] = ACTIONS(2095),
    [anon_sym_and] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(2095),
    [anon_sym_QMARK] = ACTIONS(2095),
    [anon_sym_LPAREN] = ACTIONS(2097),
    [anon_sym_external] = ACTIONS(2095),
    [anon_sym_type] = ACTIONS(2095),
    [anon_sym_COMMA] = ACTIONS(2097),
    [anon_sym_PIPE] = ACTIONS(2095),
    [anon_sym_DOT] = ACTIONS(2097),
    [anon_sym_LBRACK] = ACTIONS(2095),
    [anon_sym_RBRACK] = ACTIONS(2097),
    [anon_sym_true] = ACTIONS(2095),
    [anon_sym_false] = ACTIONS(2095),
    [anon_sym_LBRACE] = ACTIONS(2095),
    [anon_sym_SEMI] = ACTIONS(2095),
    [anon_sym_exception] = ACTIONS(2095),
    [anon_sym_module] = ACTIONS(2095),
    [anon_sym_open] = ACTIONS(2095),
    [anon_sym_include] = ACTIONS(2095),
    [anon_sym_class] = ACTIONS(2095),
    [anon_sym_val] = ACTIONS(2095),
    [anon_sym_end] = ACTIONS(2095),
    [anon_sym_object] = ACTIONS(2095),
    [anon_sym_in] = ACTIONS(2095),
    [anon_sym_AMP] = ACTIONS(2095),
    [anon_sym_POUND] = ACTIONS(2095),
    [anon_sym_COLON_COLON] = ACTIONS(2097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2097),
    [anon_sym_else] = ACTIONS(2095),
    [anon_sym_new] = ACTIONS(2095),
    [anon_sym_LBRACE_LT] = ACTIONS(2097),
    [anon_sym_begin] = ACTIONS(2095),
    [sym_ocamlyacc_value] = ACTIONS(2097),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2095),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2097),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2095),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2095),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2097),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2097),
    [aux_sym_number_token1] = ACTIONS(2097),
    [aux_sym_signed_number_token1] = ACTIONS(2095),
    [anon_sym_SQUOTE] = ACTIONS(2097),
    [anon_sym_DQUOTE] = ACTIONS(2097),
    [sym_prefix_operator] = ACTIONS(2097),
    [aux_sym_sign_operator_token1] = ACTIONS(2095),
    [sym_hash_operator] = ACTIONS(2097),
    [anon_sym_lsl] = ACTIONS(2095),
    [anon_sym_lsr] = ACTIONS(2095),
    [anon_sym_asr] = ACTIONS(2095),
    [aux_sym_pow_operator_token1] = ACTIONS(2097),
    [anon_sym_mod] = ACTIONS(2095),
    [anon_sym_land] = ACTIONS(2095),
    [anon_sym_lor] = ACTIONS(2095),
    [anon_sym_lxor] = ACTIONS(2095),
    [aux_sym_mult_operator_token1] = ACTIONS(2095),
    [aux_sym_add_operator_token1] = ACTIONS(2095),
    [sym_concat_operator] = ACTIONS(2097),
    [sym_rel_operator] = ACTIONS(2095),
    [anon_sym_AMP_AMP] = ACTIONS(2095),
    [anon_sym_or] = ACTIONS(2095),
    [anon_sym_PIPE_PIPE] = ACTIONS(2095),
    [sym_assign_operator] = ACTIONS(2097),
    [sym_let_operator] = ACTIONS(2097),
    [sym_let_and_operator] = ACTIONS(2097),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_directive_token1] = ACTIONS(2095),
    [aux_sym_tag_token1] = ACTIONS(2097),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1233)] = {
    [sym_attribute] = STATE(1233),
    [sym__lowercase_identifier] = ACTIONS(1975),
    [anon_sym_SEMI_SEMI] = ACTIONS(1977),
    [anon_sym_let] = ACTIONS(1975),
    [anon_sym_and] = ACTIONS(1975),
    [anon_sym_TILDE] = ACTIONS(1975),
    [anon_sym_QMARK] = ACTIONS(1975),
    [anon_sym_LPAREN] = ACTIONS(1977),
    [anon_sym_external] = ACTIONS(1975),
    [anon_sym_type] = ACTIONS(1975),
    [anon_sym_COMMA] = ACTIONS(1977),
    [anon_sym_PIPE] = ACTIONS(1975),
    [anon_sym_DOT] = ACTIONS(1977),
    [anon_sym_LBRACK] = ACTIONS(1975),
    [anon_sym_RBRACK] = ACTIONS(1977),
    [anon_sym_true] = ACTIONS(1975),
    [anon_sym_false] = ACTIONS(1975),
    [anon_sym_LBRACE] = ACTIONS(1975),
    [anon_sym_SEMI] = ACTIONS(1975),
    [anon_sym_exception] = ACTIONS(1975),
    [anon_sym_module] = ACTIONS(1975),
    [anon_sym_open] = ACTIONS(1975),
    [anon_sym_include] = ACTIONS(1975),
    [anon_sym_class] = ACTIONS(1975),
    [anon_sym_val] = ACTIONS(1975),
    [anon_sym_end] = ACTIONS(1975),
    [anon_sym_object] = ACTIONS(1975),
    [anon_sym_in] = ACTIONS(1975),
    [anon_sym_AMP] = ACTIONS(1975),
    [anon_sym_POUND] = ACTIONS(1975),
    [anon_sym_COLON_COLON] = ACTIONS(1977),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1977),
    [anon_sym_else] = ACTIONS(1975),
    [anon_sym_new] = ACTIONS(1975),
    [anon_sym_LBRACE_LT] = ACTIONS(1977),
    [anon_sym_begin] = ACTIONS(1975),
    [sym_ocamlyacc_value] = ACTIONS(1977),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1975),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1977),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1975),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1975),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1977),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1977),
    [aux_sym_number_token1] = ACTIONS(1977),
    [aux_sym_signed_number_token1] = ACTIONS(1975),
    [anon_sym_SQUOTE] = ACTIONS(1977),
    [anon_sym_DQUOTE] = ACTIONS(1977),
    [sym_prefix_operator] = ACTIONS(1977),
    [aux_sym_sign_operator_token1] = ACTIONS(1975),
    [sym_hash_operator] = ACTIONS(1977),
    [anon_sym_lsl] = ACTIONS(1975),
    [anon_sym_lsr] = ACTIONS(1975),
    [anon_sym_asr] = ACTIONS(1975),
    [aux_sym_pow_operator_token1] = ACTIONS(1977),
    [anon_sym_mod] = ACTIONS(1975),
    [anon_sym_land] = ACTIONS(1975),
    [anon_sym_lor] = ACTIONS(1975),
    [anon_sym_lxor] = ACTIONS(1975),
    [aux_sym_mult_operator_token1] = ACTIONS(1975),
    [aux_sym_add_operator_token1] = ACTIONS(1975),
    [sym_concat_operator] = ACTIONS(1977),
    [sym_rel_operator] = ACTIONS(1975),
    [anon_sym_AMP_AMP] = ACTIONS(1975),
    [anon_sym_or] = ACTIONS(1975),
    [anon_sym_PIPE_PIPE] = ACTIONS(1975),
    [sym_assign_operator] = ACTIONS(1977),
    [sym_let_operator] = ACTIONS(1977),
    [sym_let_and_operator] = ACTIONS(1977),
    [sym__uppercase_identifier] = ACTIONS(1977),
    [aux_sym_directive_token1] = ACTIONS(1975),
    [aux_sym_tag_token1] = ACTIONS(1977),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1234)] = {
    [sym_attribute] = STATE(1234),
    [sym__lowercase_identifier] = ACTIONS(1931),
    [anon_sym_TILDE] = ACTIONS(1931),
    [anon_sym_QMARK] = ACTIONS(1931),
    [anon_sym_LPAREN] = ACTIONS(1933),
    [anon_sym_RPAREN] = ACTIONS(1933),
    [anon_sym_COMMA] = ACTIONS(1933),
    [anon_sym_PIPE] = ACTIONS(1931),
    [anon_sym_COLON2] = ACTIONS(1931),
    [anon_sym_DASH_GT] = ACTIONS(1931),
    [anon_sym_LBRACK] = ACTIONS(1931),
    [anon_sym_RBRACK] = ACTIONS(1933),
    [anon_sym_true] = ACTIONS(1931),
    [anon_sym_false] = ACTIONS(1931),
    [anon_sym_LBRACE] = ACTIONS(1931),
    [anon_sym_SEMI] = ACTIONS(1933),
    [anon_sym_RBRACE] = ACTIONS(1933),
    [anon_sym_constraint] = ACTIONS(1931),
    [anon_sym_val] = ACTIONS(1931),
    [anon_sym_end] = ACTIONS(1931),
    [anon_sym_with] = ACTIONS(1931),
    [anon_sym_COLON_GT] = ACTIONS(1933),
    [anon_sym_object] = ACTIONS(1931),
    [anon_sym_inherit] = ACTIONS(1931),
    [anon_sym_method] = ACTIONS(1931),
    [anon_sym_initializer] = ACTIONS(1931),
    [anon_sym_AMP] = ACTIONS(1931),
    [anon_sym_COLON_COLON] = ACTIONS(1933),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1933),
    [anon_sym_then] = ACTIONS(1931),
    [anon_sym_else] = ACTIONS(1931),
    [anon_sym_do] = ACTIONS(1931),
    [anon_sym_done] = ACTIONS(1931),
    [anon_sym_to] = ACTIONS(1931),
    [anon_sym_downto] = ACTIONS(1931),
    [anon_sym_new] = ACTIONS(1931),
    [anon_sym_LBRACE_LT] = ACTIONS(1933),
    [anon_sym_GT_RBRACE] = ACTIONS(1933),
    [anon_sym_begin] = ACTIONS(1931),
    [sym_ocamlyacc_value] = ACTIONS(1933),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1931),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1933),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1933),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1933),
    [aux_sym_number_token1] = ACTIONS(1933),
    [aux_sym_signed_number_token1] = ACTIONS(1931),
    [anon_sym_SQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1933),
    [sym_prefix_operator] = ACTIONS(1933),
    [aux_sym_sign_operator_token1] = ACTIONS(1931),
    [anon_sym_lsl] = ACTIONS(1931),
    [anon_sym_lsr] = ACTIONS(1931),
    [anon_sym_asr] = ACTIONS(1931),
    [aux_sym_pow_operator_token1] = ACTIONS(1933),
    [anon_sym_mod] = ACTIONS(1931),
    [anon_sym_land] = ACTIONS(1931),
    [anon_sym_lor] = ACTIONS(1931),
    [anon_sym_lxor] = ACTIONS(1931),
    [aux_sym_mult_operator_token1] = ACTIONS(1931),
    [aux_sym_add_operator_token1] = ACTIONS(1931),
    [sym_concat_operator] = ACTIONS(1933),
    [sym_rel_operator] = ACTIONS(1931),
    [anon_sym_AMP_AMP] = ACTIONS(1931),
    [anon_sym_or] = ACTIONS(1931),
    [anon_sym_PIPE_PIPE] = ACTIONS(1931),
    [sym_assign_operator] = ACTIONS(1933),
    [sym__uppercase_identifier] = ACTIONS(1933),
    [aux_sym_tag_token1] = ACTIONS(1933),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1235)] = {
    [sym_attribute] = STATE(1235),
    [ts_builtin_sym_end] = ACTIONS(2127),
    [sym__lowercase_identifier] = ACTIONS(2125),
    [anon_sym_SEMI_SEMI] = ACTIONS(2127),
    [anon_sym_let] = ACTIONS(2125),
    [anon_sym_and] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(2125),
    [anon_sym_QMARK] = ACTIONS(2125),
    [anon_sym_LPAREN] = ACTIONS(2127),
    [anon_sym_external] = ACTIONS(2125),
    [anon_sym_type] = ACTIONS(2125),
    [anon_sym_COMMA] = ACTIONS(2127),
    [anon_sym_PIPE] = ACTIONS(2125),
    [anon_sym_DOT] = ACTIONS(2127),
    [anon_sym_LBRACK] = ACTIONS(2125),
    [anon_sym_true] = ACTIONS(2125),
    [anon_sym_false] = ACTIONS(2125),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_SEMI] = ACTIONS(2125),
    [anon_sym_exception] = ACTIONS(2125),
    [anon_sym_module] = ACTIONS(2125),
    [anon_sym_open] = ACTIONS(2125),
    [anon_sym_include] = ACTIONS(2125),
    [anon_sym_class] = ACTIONS(2125),
    [anon_sym_val] = ACTIONS(2125),
    [anon_sym_object] = ACTIONS(2125),
    [anon_sym_in] = ACTIONS(2125),
    [anon_sym_AMP] = ACTIONS(2125),
    [anon_sym_POUND] = ACTIONS(2125),
    [anon_sym_COLON_COLON] = ACTIONS(2127),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2127),
    [anon_sym_else] = ACTIONS(2125),
    [anon_sym_new] = ACTIONS(2125),
    [anon_sym_LBRACE_LT] = ACTIONS(2127),
    [anon_sym_begin] = ACTIONS(2125),
    [sym_ocamlyacc_value] = ACTIONS(2127),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2125),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2127),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2127),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2127),
    [aux_sym_number_token1] = ACTIONS(2127),
    [aux_sym_signed_number_token1] = ACTIONS(2125),
    [anon_sym_SQUOTE] = ACTIONS(2127),
    [anon_sym_DQUOTE] = ACTIONS(2127),
    [sym_prefix_operator] = ACTIONS(2127),
    [aux_sym_sign_operator_token1] = ACTIONS(2125),
    [sym_hash_operator] = ACTIONS(2127),
    [anon_sym_lsl] = ACTIONS(2125),
    [anon_sym_lsr] = ACTIONS(2125),
    [anon_sym_asr] = ACTIONS(2125),
    [aux_sym_pow_operator_token1] = ACTIONS(2127),
    [anon_sym_mod] = ACTIONS(2125),
    [anon_sym_land] = ACTIONS(2125),
    [anon_sym_lor] = ACTIONS(2125),
    [anon_sym_lxor] = ACTIONS(2125),
    [aux_sym_mult_operator_token1] = ACTIONS(2125),
    [aux_sym_add_operator_token1] = ACTIONS(2125),
    [sym_concat_operator] = ACTIONS(2127),
    [sym_rel_operator] = ACTIONS(2125),
    [anon_sym_AMP_AMP] = ACTIONS(2125),
    [anon_sym_or] = ACTIONS(2125),
    [anon_sym_PIPE_PIPE] = ACTIONS(2125),
    [sym_assign_operator] = ACTIONS(2127),
    [sym_let_operator] = ACTIONS(2127),
    [sym_let_and_operator] = ACTIONS(2127),
    [sym__uppercase_identifier] = ACTIONS(2127),
    [aux_sym_directive_token1] = ACTIONS(2125),
    [aux_sym_tag_token1] = ACTIONS(2127),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1236)] = {
    [sym_attribute] = STATE(1236),
    [ts_builtin_sym_end] = ACTIONS(2131),
    [sym__lowercase_identifier] = ACTIONS(2129),
    [anon_sym_SEMI_SEMI] = ACTIONS(2131),
    [anon_sym_let] = ACTIONS(2129),
    [anon_sym_and] = ACTIONS(2129),
    [anon_sym_TILDE] = ACTIONS(2129),
    [anon_sym_QMARK] = ACTIONS(2129),
    [anon_sym_LPAREN] = ACTIONS(2131),
    [anon_sym_external] = ACTIONS(2129),
    [anon_sym_type] = ACTIONS(2129),
    [anon_sym_COMMA] = ACTIONS(2131),
    [anon_sym_PIPE] = ACTIONS(2129),
    [anon_sym_DOT] = ACTIONS(2131),
    [anon_sym_LBRACK] = ACTIONS(2129),
    [anon_sym_true] = ACTIONS(2129),
    [anon_sym_false] = ACTIONS(2129),
    [anon_sym_LBRACE] = ACTIONS(2129),
    [anon_sym_SEMI] = ACTIONS(2129),
    [anon_sym_exception] = ACTIONS(2129),
    [anon_sym_module] = ACTIONS(2129),
    [anon_sym_open] = ACTIONS(2129),
    [anon_sym_include] = ACTIONS(2129),
    [anon_sym_class] = ACTIONS(2129),
    [anon_sym_val] = ACTIONS(2129),
    [anon_sym_object] = ACTIONS(2129),
    [anon_sym_in] = ACTIONS(2129),
    [anon_sym_AMP] = ACTIONS(2129),
    [anon_sym_POUND] = ACTIONS(2129),
    [anon_sym_COLON_COLON] = ACTIONS(2131),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2131),
    [anon_sym_else] = ACTIONS(2129),
    [anon_sym_new] = ACTIONS(2129),
    [anon_sym_LBRACE_LT] = ACTIONS(2131),
    [anon_sym_begin] = ACTIONS(2129),
    [sym_ocamlyacc_value] = ACTIONS(2131),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2129),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2131),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2129),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2129),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2131),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2131),
    [aux_sym_number_token1] = ACTIONS(2131),
    [aux_sym_signed_number_token1] = ACTIONS(2129),
    [anon_sym_SQUOTE] = ACTIONS(2131),
    [anon_sym_DQUOTE] = ACTIONS(2131),
    [sym_prefix_operator] = ACTIONS(2131),
    [aux_sym_sign_operator_token1] = ACTIONS(2129),
    [sym_hash_operator] = ACTIONS(2131),
    [anon_sym_lsl] = ACTIONS(2129),
    [anon_sym_lsr] = ACTIONS(2129),
    [anon_sym_asr] = ACTIONS(2129),
    [aux_sym_pow_operator_token1] = ACTIONS(2131),
    [anon_sym_mod] = ACTIONS(2129),
    [anon_sym_land] = ACTIONS(2129),
    [anon_sym_lor] = ACTIONS(2129),
    [anon_sym_lxor] = ACTIONS(2129),
    [aux_sym_mult_operator_token1] = ACTIONS(2129),
    [aux_sym_add_operator_token1] = ACTIONS(2129),
    [sym_concat_operator] = ACTIONS(2131),
    [sym_rel_operator] = ACTIONS(2129),
    [anon_sym_AMP_AMP] = ACTIONS(2129),
    [anon_sym_or] = ACTIONS(2129),
    [anon_sym_PIPE_PIPE] = ACTIONS(2129),
    [sym_assign_operator] = ACTIONS(2131),
    [sym_let_operator] = ACTIONS(2131),
    [sym_let_and_operator] = ACTIONS(2131),
    [sym__uppercase_identifier] = ACTIONS(2131),
    [aux_sym_directive_token1] = ACTIONS(2129),
    [aux_sym_tag_token1] = ACTIONS(2131),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1237)] = {
    [sym_attribute] = STATE(1237),
    [ts_builtin_sym_end] = ACTIONS(2135),
    [sym__lowercase_identifier] = ACTIONS(2133),
    [anon_sym_SEMI_SEMI] = ACTIONS(2135),
    [anon_sym_let] = ACTIONS(2133),
    [anon_sym_and] = ACTIONS(2133),
    [anon_sym_TILDE] = ACTIONS(2133),
    [anon_sym_QMARK] = ACTIONS(2133),
    [anon_sym_LPAREN] = ACTIONS(2135),
    [anon_sym_external] = ACTIONS(2133),
    [anon_sym_type] = ACTIONS(2133),
    [anon_sym_COMMA] = ACTIONS(2135),
    [anon_sym_PIPE] = ACTIONS(2133),
    [anon_sym_DOT] = ACTIONS(2135),
    [anon_sym_LBRACK] = ACTIONS(2133),
    [anon_sym_RBRACK] = ACTIONS(2135),
    [anon_sym_true] = ACTIONS(2133),
    [anon_sym_false] = ACTIONS(2133),
    [anon_sym_LBRACE] = ACTIONS(2133),
    [anon_sym_SEMI] = ACTIONS(2133),
    [anon_sym_exception] = ACTIONS(2133),
    [anon_sym_module] = ACTIONS(2133),
    [anon_sym_open] = ACTIONS(2133),
    [anon_sym_include] = ACTIONS(2133),
    [anon_sym_class] = ACTIONS(2133),
    [anon_sym_val] = ACTIONS(2133),
    [anon_sym_end] = ACTIONS(2133),
    [anon_sym_object] = ACTIONS(2133),
    [anon_sym_AMP] = ACTIONS(2133),
    [anon_sym_POUND] = ACTIONS(2133),
    [anon_sym_COLON_COLON] = ACTIONS(2135),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2135),
    [anon_sym_else] = ACTIONS(2133),
    [anon_sym_new] = ACTIONS(2133),
    [anon_sym_LBRACE_LT] = ACTIONS(2135),
    [anon_sym_begin] = ACTIONS(2133),
    [sym_ocamlyacc_value] = ACTIONS(2135),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2133),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2135),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2133),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2133),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2135),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2135),
    [aux_sym_number_token1] = ACTIONS(2135),
    [aux_sym_signed_number_token1] = ACTIONS(2133),
    [anon_sym_SQUOTE] = ACTIONS(2135),
    [anon_sym_DQUOTE] = ACTIONS(2135),
    [sym_prefix_operator] = ACTIONS(2135),
    [aux_sym_sign_operator_token1] = ACTIONS(2133),
    [sym_hash_operator] = ACTIONS(2135),
    [anon_sym_lsl] = ACTIONS(2133),
    [anon_sym_lsr] = ACTIONS(2133),
    [anon_sym_asr] = ACTIONS(2133),
    [aux_sym_pow_operator_token1] = ACTIONS(2135),
    [anon_sym_mod] = ACTIONS(2133),
    [anon_sym_land] = ACTIONS(2133),
    [anon_sym_lor] = ACTIONS(2133),
    [anon_sym_lxor] = ACTIONS(2133),
    [aux_sym_mult_operator_token1] = ACTIONS(2133),
    [aux_sym_add_operator_token1] = ACTIONS(2133),
    [sym_concat_operator] = ACTIONS(2135),
    [sym_rel_operator] = ACTIONS(2133),
    [anon_sym_AMP_AMP] = ACTIONS(2133),
    [anon_sym_or] = ACTIONS(2133),
    [anon_sym_PIPE_PIPE] = ACTIONS(2133),
    [sym_assign_operator] = ACTIONS(2135),
    [sym_let_operator] = ACTIONS(2135),
    [sym__uppercase_identifier] = ACTIONS(2135),
    [aux_sym_directive_token1] = ACTIONS(2133),
    [aux_sym_tag_token1] = ACTIONS(2135),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1238)] = {
    [sym_attribute] = STATE(1238),
    [ts_builtin_sym_end] = ACTIONS(1969),
    [sym__lowercase_identifier] = ACTIONS(1967),
    [anon_sym_SEMI_SEMI] = ACTIONS(1969),
    [anon_sym_let] = ACTIONS(1967),
    [anon_sym_and] = ACTIONS(1967),
    [anon_sym_TILDE] = ACTIONS(1967),
    [anon_sym_QMARK] = ACTIONS(1967),
    [anon_sym_LPAREN] = ACTIONS(1969),
    [anon_sym_external] = ACTIONS(1967),
    [anon_sym_type] = ACTIONS(1967),
    [anon_sym_COMMA] = ACTIONS(1969),
    [anon_sym_PIPE] = ACTIONS(1967),
    [anon_sym_DOT] = ACTIONS(1969),
    [anon_sym_LBRACK] = ACTIONS(1967),
    [anon_sym_true] = ACTIONS(1967),
    [anon_sym_false] = ACTIONS(1967),
    [anon_sym_LBRACE] = ACTIONS(1967),
    [anon_sym_SEMI] = ACTIONS(1967),
    [anon_sym_exception] = ACTIONS(1967),
    [anon_sym_module] = ACTIONS(1967),
    [anon_sym_open] = ACTIONS(1967),
    [anon_sym_include] = ACTIONS(1967),
    [anon_sym_class] = ACTIONS(1967),
    [anon_sym_val] = ACTIONS(1967),
    [anon_sym_object] = ACTIONS(1967),
    [anon_sym_in] = ACTIONS(1967),
    [anon_sym_AMP] = ACTIONS(1967),
    [anon_sym_POUND] = ACTIONS(1967),
    [anon_sym_COLON_COLON] = ACTIONS(1969),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1969),
    [anon_sym_else] = ACTIONS(1967),
    [anon_sym_new] = ACTIONS(1967),
    [anon_sym_LBRACE_LT] = ACTIONS(1969),
    [anon_sym_begin] = ACTIONS(1967),
    [sym_ocamlyacc_value] = ACTIONS(1969),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1967),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1969),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1969),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1969),
    [aux_sym_number_token1] = ACTIONS(1969),
    [aux_sym_signed_number_token1] = ACTIONS(1967),
    [anon_sym_SQUOTE] = ACTIONS(1969),
    [anon_sym_DQUOTE] = ACTIONS(1969),
    [sym_prefix_operator] = ACTIONS(1969),
    [aux_sym_sign_operator_token1] = ACTIONS(1967),
    [sym_hash_operator] = ACTIONS(1969),
    [anon_sym_lsl] = ACTIONS(1967),
    [anon_sym_lsr] = ACTIONS(1967),
    [anon_sym_asr] = ACTIONS(1967),
    [aux_sym_pow_operator_token1] = ACTIONS(1969),
    [anon_sym_mod] = ACTIONS(1967),
    [anon_sym_land] = ACTIONS(1967),
    [anon_sym_lor] = ACTIONS(1967),
    [anon_sym_lxor] = ACTIONS(1967),
    [aux_sym_mult_operator_token1] = ACTIONS(1967),
    [aux_sym_add_operator_token1] = ACTIONS(1967),
    [sym_concat_operator] = ACTIONS(1969),
    [sym_rel_operator] = ACTIONS(1967),
    [anon_sym_AMP_AMP] = ACTIONS(1967),
    [anon_sym_or] = ACTIONS(1967),
    [anon_sym_PIPE_PIPE] = ACTIONS(1967),
    [sym_assign_operator] = ACTIONS(1969),
    [sym_let_operator] = ACTIONS(1969),
    [sym_let_and_operator] = ACTIONS(1969),
    [sym__uppercase_identifier] = ACTIONS(1969),
    [aux_sym_directive_token1] = ACTIONS(1967),
    [aux_sym_tag_token1] = ACTIONS(1969),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1239)] = {
    [sym_attribute] = STATE(1239),
    [ts_builtin_sym_end] = ACTIONS(1973),
    [sym__lowercase_identifier] = ACTIONS(1971),
    [anon_sym_SEMI_SEMI] = ACTIONS(1973),
    [anon_sym_let] = ACTIONS(1971),
    [anon_sym_and] = ACTIONS(1971),
    [anon_sym_TILDE] = ACTIONS(1971),
    [anon_sym_QMARK] = ACTIONS(1971),
    [anon_sym_LPAREN] = ACTIONS(1973),
    [anon_sym_external] = ACTIONS(1971),
    [anon_sym_type] = ACTIONS(1971),
    [anon_sym_COMMA] = ACTIONS(1973),
    [anon_sym_PIPE] = ACTIONS(1971),
    [anon_sym_DOT] = ACTIONS(1017),
    [anon_sym_LBRACK] = ACTIONS(1971),
    [anon_sym_true] = ACTIONS(1971),
    [anon_sym_false] = ACTIONS(1971),
    [anon_sym_LBRACE] = ACTIONS(1971),
    [anon_sym_SEMI] = ACTIONS(1971),
    [anon_sym_exception] = ACTIONS(1971),
    [anon_sym_module] = ACTIONS(1971),
    [anon_sym_open] = ACTIONS(1971),
    [anon_sym_include] = ACTIONS(1971),
    [anon_sym_class] = ACTIONS(1971),
    [anon_sym_val] = ACTIONS(1971),
    [anon_sym_object] = ACTIONS(1971),
    [anon_sym_in] = ACTIONS(1971),
    [anon_sym_AMP] = ACTIONS(1971),
    [anon_sym_POUND] = ACTIONS(1971),
    [anon_sym_COLON_COLON] = ACTIONS(1973),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1973),
    [anon_sym_else] = ACTIONS(1971),
    [anon_sym_new] = ACTIONS(1971),
    [anon_sym_LBRACE_LT] = ACTIONS(1973),
    [anon_sym_begin] = ACTIONS(1971),
    [sym_ocamlyacc_value] = ACTIONS(1973),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1971),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1973),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1973),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1973),
    [aux_sym_number_token1] = ACTIONS(1973),
    [aux_sym_signed_number_token1] = ACTIONS(1971),
    [anon_sym_SQUOTE] = ACTIONS(1973),
    [anon_sym_DQUOTE] = ACTIONS(1973),
    [sym_prefix_operator] = ACTIONS(1973),
    [aux_sym_sign_operator_token1] = ACTIONS(1971),
    [sym_hash_operator] = ACTIONS(1973),
    [anon_sym_lsl] = ACTIONS(1971),
    [anon_sym_lsr] = ACTIONS(1971),
    [anon_sym_asr] = ACTIONS(1971),
    [aux_sym_pow_operator_token1] = ACTIONS(1973),
    [anon_sym_mod] = ACTIONS(1971),
    [anon_sym_land] = ACTIONS(1971),
    [anon_sym_lor] = ACTIONS(1971),
    [anon_sym_lxor] = ACTIONS(1971),
    [aux_sym_mult_operator_token1] = ACTIONS(1971),
    [aux_sym_add_operator_token1] = ACTIONS(1971),
    [sym_concat_operator] = ACTIONS(1973),
    [sym_rel_operator] = ACTIONS(1971),
    [anon_sym_AMP_AMP] = ACTIONS(1971),
    [anon_sym_or] = ACTIONS(1971),
    [anon_sym_PIPE_PIPE] = ACTIONS(1971),
    [sym_assign_operator] = ACTIONS(1973),
    [sym_let_operator] = ACTIONS(1973),
    [sym_let_and_operator] = ACTIONS(1973),
    [sym__uppercase_identifier] = ACTIONS(1973),
    [aux_sym_directive_token1] = ACTIONS(1971),
    [aux_sym_tag_token1] = ACTIONS(1973),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1240)] = {
    [sym_attribute] = STATE(1240),
    [ts_builtin_sym_end] = ACTIONS(1977),
    [sym__lowercase_identifier] = ACTIONS(1975),
    [anon_sym_SEMI_SEMI] = ACTIONS(1977),
    [anon_sym_let] = ACTIONS(1975),
    [anon_sym_and] = ACTIONS(1975),
    [anon_sym_TILDE] = ACTIONS(1975),
    [anon_sym_QMARK] = ACTIONS(1975),
    [anon_sym_LPAREN] = ACTIONS(1977),
    [anon_sym_external] = ACTIONS(1975),
    [anon_sym_type] = ACTIONS(1975),
    [anon_sym_COMMA] = ACTIONS(1977),
    [anon_sym_PIPE] = ACTIONS(1975),
    [anon_sym_DOT] = ACTIONS(1977),
    [anon_sym_LBRACK] = ACTIONS(1975),
    [anon_sym_true] = ACTIONS(1975),
    [anon_sym_false] = ACTIONS(1975),
    [anon_sym_LBRACE] = ACTIONS(1975),
    [anon_sym_SEMI] = ACTIONS(1975),
    [anon_sym_exception] = ACTIONS(1975),
    [anon_sym_module] = ACTIONS(1975),
    [anon_sym_open] = ACTIONS(1975),
    [anon_sym_include] = ACTIONS(1975),
    [anon_sym_class] = ACTIONS(1975),
    [anon_sym_val] = ACTIONS(1975),
    [anon_sym_object] = ACTIONS(1975),
    [anon_sym_in] = ACTIONS(1975),
    [anon_sym_AMP] = ACTIONS(1975),
    [anon_sym_POUND] = ACTIONS(1975),
    [anon_sym_COLON_COLON] = ACTIONS(1977),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1977),
    [anon_sym_else] = ACTIONS(1975),
    [anon_sym_new] = ACTIONS(1975),
    [anon_sym_LBRACE_LT] = ACTIONS(1977),
    [anon_sym_begin] = ACTIONS(1975),
    [sym_ocamlyacc_value] = ACTIONS(1977),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1975),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1977),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1975),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1975),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1977),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1977),
    [aux_sym_number_token1] = ACTIONS(1977),
    [aux_sym_signed_number_token1] = ACTIONS(1975),
    [anon_sym_SQUOTE] = ACTIONS(1977),
    [anon_sym_DQUOTE] = ACTIONS(1977),
    [sym_prefix_operator] = ACTIONS(1977),
    [aux_sym_sign_operator_token1] = ACTIONS(1975),
    [sym_hash_operator] = ACTIONS(1977),
    [anon_sym_lsl] = ACTIONS(1975),
    [anon_sym_lsr] = ACTIONS(1975),
    [anon_sym_asr] = ACTIONS(1975),
    [aux_sym_pow_operator_token1] = ACTIONS(1977),
    [anon_sym_mod] = ACTIONS(1975),
    [anon_sym_land] = ACTIONS(1975),
    [anon_sym_lor] = ACTIONS(1975),
    [anon_sym_lxor] = ACTIONS(1975),
    [aux_sym_mult_operator_token1] = ACTIONS(1975),
    [aux_sym_add_operator_token1] = ACTIONS(1975),
    [sym_concat_operator] = ACTIONS(1977),
    [sym_rel_operator] = ACTIONS(1975),
    [anon_sym_AMP_AMP] = ACTIONS(1975),
    [anon_sym_or] = ACTIONS(1975),
    [anon_sym_PIPE_PIPE] = ACTIONS(1975),
    [sym_assign_operator] = ACTIONS(1977),
    [sym_let_operator] = ACTIONS(1977),
    [sym_let_and_operator] = ACTIONS(1977),
    [sym__uppercase_identifier] = ACTIONS(1977),
    [aux_sym_directive_token1] = ACTIONS(1975),
    [aux_sym_tag_token1] = ACTIONS(1977),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1241)] = {
    [sym_attribute] = STATE(1241),
    [ts_builtin_sym_end] = ACTIONS(2139),
    [sym__lowercase_identifier] = ACTIONS(2137),
    [anon_sym_SEMI_SEMI] = ACTIONS(2139),
    [anon_sym_let] = ACTIONS(2137),
    [anon_sym_and] = ACTIONS(2137),
    [anon_sym_TILDE] = ACTIONS(2137),
    [anon_sym_QMARK] = ACTIONS(2137),
    [anon_sym_LPAREN] = ACTIONS(2139),
    [anon_sym_external] = ACTIONS(2137),
    [anon_sym_type] = ACTIONS(2137),
    [anon_sym_COMMA] = ACTIONS(2139),
    [anon_sym_PIPE] = ACTIONS(2137),
    [anon_sym_DOT] = ACTIONS(2139),
    [anon_sym_LBRACK] = ACTIONS(2137),
    [anon_sym_true] = ACTIONS(2137),
    [anon_sym_false] = ACTIONS(2137),
    [anon_sym_LBRACE] = ACTIONS(2137),
    [anon_sym_SEMI] = ACTIONS(2137),
    [anon_sym_exception] = ACTIONS(2137),
    [anon_sym_module] = ACTIONS(2137),
    [anon_sym_open] = ACTIONS(2137),
    [anon_sym_include] = ACTIONS(2137),
    [anon_sym_class] = ACTIONS(2137),
    [anon_sym_val] = ACTIONS(2137),
    [anon_sym_object] = ACTIONS(2137),
    [anon_sym_in] = ACTIONS(2137),
    [anon_sym_AMP] = ACTIONS(2137),
    [anon_sym_POUND] = ACTIONS(2137),
    [anon_sym_COLON_COLON] = ACTIONS(2139),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2139),
    [anon_sym_else] = ACTIONS(2137),
    [anon_sym_new] = ACTIONS(2137),
    [anon_sym_LBRACE_LT] = ACTIONS(2139),
    [anon_sym_begin] = ACTIONS(2137),
    [sym_ocamlyacc_value] = ACTIONS(2139),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2137),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2139),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2137),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2137),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2139),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2139),
    [aux_sym_number_token1] = ACTIONS(2139),
    [aux_sym_signed_number_token1] = ACTIONS(2137),
    [anon_sym_SQUOTE] = ACTIONS(2139),
    [anon_sym_DQUOTE] = ACTIONS(2139),
    [sym_prefix_operator] = ACTIONS(2139),
    [aux_sym_sign_operator_token1] = ACTIONS(2137),
    [sym_hash_operator] = ACTIONS(2139),
    [anon_sym_lsl] = ACTIONS(2137),
    [anon_sym_lsr] = ACTIONS(2137),
    [anon_sym_asr] = ACTIONS(2137),
    [aux_sym_pow_operator_token1] = ACTIONS(2139),
    [anon_sym_mod] = ACTIONS(2137),
    [anon_sym_land] = ACTIONS(2137),
    [anon_sym_lor] = ACTIONS(2137),
    [anon_sym_lxor] = ACTIONS(2137),
    [aux_sym_mult_operator_token1] = ACTIONS(2137),
    [aux_sym_add_operator_token1] = ACTIONS(2137),
    [sym_concat_operator] = ACTIONS(2139),
    [sym_rel_operator] = ACTIONS(2137),
    [anon_sym_AMP_AMP] = ACTIONS(2137),
    [anon_sym_or] = ACTIONS(2137),
    [anon_sym_PIPE_PIPE] = ACTIONS(2137),
    [sym_assign_operator] = ACTIONS(2139),
    [sym_let_operator] = ACTIONS(2139),
    [sym_let_and_operator] = ACTIONS(2139),
    [sym__uppercase_identifier] = ACTIONS(2139),
    [aux_sym_directive_token1] = ACTIONS(2137),
    [aux_sym_tag_token1] = ACTIONS(2139),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1242)] = {
    [sym_attribute] = STATE(1242),
    [ts_builtin_sym_end] = ACTIONS(2175),
    [sym__lowercase_identifier] = ACTIONS(2173),
    [anon_sym_SEMI_SEMI] = ACTIONS(2175),
    [anon_sym_let] = ACTIONS(2173),
    [anon_sym_and] = ACTIONS(2173),
    [anon_sym_TILDE] = ACTIONS(2173),
    [anon_sym_QMARK] = ACTIONS(2173),
    [anon_sym_LPAREN] = ACTIONS(2175),
    [anon_sym_external] = ACTIONS(2173),
    [anon_sym_type] = ACTIONS(2173),
    [anon_sym_COMMA] = ACTIONS(2175),
    [anon_sym_PIPE] = ACTIONS(2173),
    [anon_sym_DOT] = ACTIONS(2175),
    [anon_sym_LBRACK] = ACTIONS(2173),
    [anon_sym_RBRACK] = ACTIONS(2175),
    [anon_sym_true] = ACTIONS(2173),
    [anon_sym_false] = ACTIONS(2173),
    [anon_sym_LBRACE] = ACTIONS(2173),
    [anon_sym_SEMI] = ACTIONS(2173),
    [anon_sym_exception] = ACTIONS(2173),
    [anon_sym_module] = ACTIONS(2173),
    [anon_sym_open] = ACTIONS(2173),
    [anon_sym_include] = ACTIONS(2173),
    [anon_sym_class] = ACTIONS(2173),
    [anon_sym_val] = ACTIONS(2173),
    [anon_sym_end] = ACTIONS(2173),
    [anon_sym_object] = ACTIONS(2173),
    [anon_sym_AMP] = ACTIONS(2173),
    [anon_sym_POUND] = ACTIONS(2173),
    [anon_sym_COLON_COLON] = ACTIONS(2175),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2175),
    [anon_sym_else] = ACTIONS(2173),
    [anon_sym_new] = ACTIONS(2173),
    [anon_sym_LBRACE_LT] = ACTIONS(2175),
    [anon_sym_begin] = ACTIONS(2173),
    [sym_ocamlyacc_value] = ACTIONS(2175),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2173),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2175),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2173),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2175),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2175),
    [aux_sym_number_token1] = ACTIONS(2175),
    [aux_sym_signed_number_token1] = ACTIONS(2173),
    [anon_sym_SQUOTE] = ACTIONS(2175),
    [anon_sym_DQUOTE] = ACTIONS(2175),
    [sym_prefix_operator] = ACTIONS(2175),
    [aux_sym_sign_operator_token1] = ACTIONS(2173),
    [sym_hash_operator] = ACTIONS(2175),
    [anon_sym_lsl] = ACTIONS(2173),
    [anon_sym_lsr] = ACTIONS(2173),
    [anon_sym_asr] = ACTIONS(2173),
    [aux_sym_pow_operator_token1] = ACTIONS(2175),
    [anon_sym_mod] = ACTIONS(2173),
    [anon_sym_land] = ACTIONS(2173),
    [anon_sym_lor] = ACTIONS(2173),
    [anon_sym_lxor] = ACTIONS(2173),
    [aux_sym_mult_operator_token1] = ACTIONS(2173),
    [aux_sym_add_operator_token1] = ACTIONS(2173),
    [sym_concat_operator] = ACTIONS(2175),
    [sym_rel_operator] = ACTIONS(2173),
    [anon_sym_AMP_AMP] = ACTIONS(2173),
    [anon_sym_or] = ACTIONS(2173),
    [anon_sym_PIPE_PIPE] = ACTIONS(2173),
    [sym_assign_operator] = ACTIONS(2175),
    [sym_let_operator] = ACTIONS(2175),
    [sym__uppercase_identifier] = ACTIONS(2175),
    [aux_sym_directive_token1] = ACTIONS(2173),
    [aux_sym_tag_token1] = ACTIONS(2175),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1243)] = {
    [sym_attribute] = STATE(1243),
    [ts_builtin_sym_end] = ACTIONS(1985),
    [sym__lowercase_identifier] = ACTIONS(1983),
    [anon_sym_SEMI_SEMI] = ACTIONS(1985),
    [anon_sym_let] = ACTIONS(1983),
    [anon_sym_and] = ACTIONS(1983),
    [anon_sym_TILDE] = ACTIONS(1983),
    [anon_sym_QMARK] = ACTIONS(1983),
    [anon_sym_LPAREN] = ACTIONS(1985),
    [anon_sym_external] = ACTIONS(1983),
    [anon_sym_type] = ACTIONS(1983),
    [anon_sym_COMMA] = ACTIONS(1985),
    [anon_sym_PIPE] = ACTIONS(1983),
    [anon_sym_DOT] = ACTIONS(1985),
    [anon_sym_LBRACK] = ACTIONS(1983),
    [anon_sym_true] = ACTIONS(1983),
    [anon_sym_false] = ACTIONS(1983),
    [anon_sym_LBRACE] = ACTIONS(1983),
    [anon_sym_SEMI] = ACTIONS(1983),
    [anon_sym_exception] = ACTIONS(1983),
    [anon_sym_module] = ACTIONS(1983),
    [anon_sym_open] = ACTIONS(1983),
    [anon_sym_include] = ACTIONS(1983),
    [anon_sym_class] = ACTIONS(1983),
    [anon_sym_val] = ACTIONS(1983),
    [anon_sym_object] = ACTIONS(1983),
    [anon_sym_in] = ACTIONS(1983),
    [anon_sym_AMP] = ACTIONS(1983),
    [anon_sym_POUND] = ACTIONS(1983),
    [anon_sym_COLON_COLON] = ACTIONS(1985),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1985),
    [anon_sym_else] = ACTIONS(1983),
    [anon_sym_new] = ACTIONS(1983),
    [anon_sym_LBRACE_LT] = ACTIONS(1985),
    [anon_sym_begin] = ACTIONS(1983),
    [sym_ocamlyacc_value] = ACTIONS(1985),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1983),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1985),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1983),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1983),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1985),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1985),
    [aux_sym_number_token1] = ACTIONS(1985),
    [aux_sym_signed_number_token1] = ACTIONS(1983),
    [anon_sym_SQUOTE] = ACTIONS(1985),
    [anon_sym_DQUOTE] = ACTIONS(1985),
    [sym_prefix_operator] = ACTIONS(1985),
    [aux_sym_sign_operator_token1] = ACTIONS(1983),
    [sym_hash_operator] = ACTIONS(1985),
    [anon_sym_lsl] = ACTIONS(1983),
    [anon_sym_lsr] = ACTIONS(1983),
    [anon_sym_asr] = ACTIONS(1983),
    [aux_sym_pow_operator_token1] = ACTIONS(1985),
    [anon_sym_mod] = ACTIONS(1983),
    [anon_sym_land] = ACTIONS(1983),
    [anon_sym_lor] = ACTIONS(1983),
    [anon_sym_lxor] = ACTIONS(1983),
    [aux_sym_mult_operator_token1] = ACTIONS(1983),
    [aux_sym_add_operator_token1] = ACTIONS(1983),
    [sym_concat_operator] = ACTIONS(1985),
    [sym_rel_operator] = ACTIONS(1983),
    [anon_sym_AMP_AMP] = ACTIONS(1983),
    [anon_sym_or] = ACTIONS(1983),
    [anon_sym_PIPE_PIPE] = ACTIONS(1983),
    [sym_assign_operator] = ACTIONS(1985),
    [sym_let_operator] = ACTIONS(1985),
    [sym_let_and_operator] = ACTIONS(1985),
    [sym__uppercase_identifier] = ACTIONS(1985),
    [aux_sym_directive_token1] = ACTIONS(1983),
    [aux_sym_tag_token1] = ACTIONS(1985),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1244)] = {
    [sym_attribute] = STATE(1244),
    [ts_builtin_sym_end] = ACTIONS(1989),
    [sym__lowercase_identifier] = ACTIONS(1987),
    [anon_sym_SEMI_SEMI] = ACTIONS(1989),
    [anon_sym_let] = ACTIONS(1987),
    [anon_sym_and] = ACTIONS(1987),
    [anon_sym_TILDE] = ACTIONS(1987),
    [anon_sym_QMARK] = ACTIONS(1987),
    [anon_sym_LPAREN] = ACTIONS(1989),
    [anon_sym_external] = ACTIONS(1987),
    [anon_sym_type] = ACTIONS(1987),
    [anon_sym_COMMA] = ACTIONS(1989),
    [anon_sym_PIPE] = ACTIONS(1987),
    [anon_sym_DOT] = ACTIONS(1989),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_true] = ACTIONS(1987),
    [anon_sym_false] = ACTIONS(1987),
    [anon_sym_LBRACE] = ACTIONS(1987),
    [anon_sym_SEMI] = ACTIONS(1987),
    [anon_sym_exception] = ACTIONS(1987),
    [anon_sym_module] = ACTIONS(1987),
    [anon_sym_open] = ACTIONS(1987),
    [anon_sym_include] = ACTIONS(1987),
    [anon_sym_class] = ACTIONS(1987),
    [anon_sym_val] = ACTIONS(1987),
    [anon_sym_object] = ACTIONS(1987),
    [anon_sym_in] = ACTIONS(1987),
    [anon_sym_AMP] = ACTIONS(1987),
    [anon_sym_POUND] = ACTIONS(1987),
    [anon_sym_COLON_COLON] = ACTIONS(1989),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1989),
    [anon_sym_else] = ACTIONS(1987),
    [anon_sym_new] = ACTIONS(1987),
    [anon_sym_LBRACE_LT] = ACTIONS(1989),
    [anon_sym_begin] = ACTIONS(1987),
    [sym_ocamlyacc_value] = ACTIONS(1989),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1987),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1989),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1987),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1987),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1989),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1989),
    [aux_sym_number_token1] = ACTIONS(1989),
    [aux_sym_signed_number_token1] = ACTIONS(1987),
    [anon_sym_SQUOTE] = ACTIONS(1989),
    [anon_sym_DQUOTE] = ACTIONS(1989),
    [sym_prefix_operator] = ACTIONS(1989),
    [aux_sym_sign_operator_token1] = ACTIONS(1987),
    [sym_hash_operator] = ACTIONS(1989),
    [anon_sym_lsl] = ACTIONS(1987),
    [anon_sym_lsr] = ACTIONS(1987),
    [anon_sym_asr] = ACTIONS(1987),
    [aux_sym_pow_operator_token1] = ACTIONS(1989),
    [anon_sym_mod] = ACTIONS(1987),
    [anon_sym_land] = ACTIONS(1987),
    [anon_sym_lor] = ACTIONS(1987),
    [anon_sym_lxor] = ACTIONS(1987),
    [aux_sym_mult_operator_token1] = ACTIONS(1987),
    [aux_sym_add_operator_token1] = ACTIONS(1987),
    [sym_concat_operator] = ACTIONS(1989),
    [sym_rel_operator] = ACTIONS(1987),
    [anon_sym_AMP_AMP] = ACTIONS(1987),
    [anon_sym_or] = ACTIONS(1987),
    [anon_sym_PIPE_PIPE] = ACTIONS(1987),
    [sym_assign_operator] = ACTIONS(1989),
    [sym_let_operator] = ACTIONS(1989),
    [sym_let_and_operator] = ACTIONS(1989),
    [sym__uppercase_identifier] = ACTIONS(1989),
    [aux_sym_directive_token1] = ACTIONS(1987),
    [aux_sym_tag_token1] = ACTIONS(1989),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1245)] = {
    [sym_attribute] = STATE(1245),
    [ts_builtin_sym_end] = ACTIONS(1993),
    [sym__lowercase_identifier] = ACTIONS(1991),
    [anon_sym_SEMI_SEMI] = ACTIONS(1993),
    [anon_sym_let] = ACTIONS(1991),
    [anon_sym_and] = ACTIONS(1991),
    [anon_sym_TILDE] = ACTIONS(1991),
    [anon_sym_QMARK] = ACTIONS(1991),
    [anon_sym_LPAREN] = ACTIONS(1993),
    [anon_sym_external] = ACTIONS(1991),
    [anon_sym_type] = ACTIONS(1991),
    [anon_sym_COMMA] = ACTIONS(1993),
    [anon_sym_PIPE] = ACTIONS(1991),
    [anon_sym_DOT] = ACTIONS(1993),
    [anon_sym_LBRACK] = ACTIONS(1991),
    [anon_sym_true] = ACTIONS(1991),
    [anon_sym_false] = ACTIONS(1991),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_SEMI] = ACTIONS(1991),
    [anon_sym_exception] = ACTIONS(1991),
    [anon_sym_module] = ACTIONS(1991),
    [anon_sym_open] = ACTIONS(1991),
    [anon_sym_include] = ACTIONS(1991),
    [anon_sym_class] = ACTIONS(1991),
    [anon_sym_val] = ACTIONS(1991),
    [anon_sym_object] = ACTIONS(1991),
    [anon_sym_in] = ACTIONS(1991),
    [anon_sym_AMP] = ACTIONS(1991),
    [anon_sym_POUND] = ACTIONS(1991),
    [anon_sym_COLON_COLON] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1993),
    [anon_sym_else] = ACTIONS(1991),
    [anon_sym_new] = ACTIONS(1991),
    [anon_sym_LBRACE_LT] = ACTIONS(1993),
    [anon_sym_begin] = ACTIONS(1991),
    [sym_ocamlyacc_value] = ACTIONS(1993),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1991),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1993),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1991),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1991),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1993),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1993),
    [aux_sym_number_token1] = ACTIONS(1993),
    [aux_sym_signed_number_token1] = ACTIONS(1991),
    [anon_sym_SQUOTE] = ACTIONS(1993),
    [anon_sym_DQUOTE] = ACTIONS(1993),
    [sym_prefix_operator] = ACTIONS(1993),
    [aux_sym_sign_operator_token1] = ACTIONS(1991),
    [sym_hash_operator] = ACTIONS(1993),
    [anon_sym_lsl] = ACTIONS(1991),
    [anon_sym_lsr] = ACTIONS(1991),
    [anon_sym_asr] = ACTIONS(1991),
    [aux_sym_pow_operator_token1] = ACTIONS(1993),
    [anon_sym_mod] = ACTIONS(1991),
    [anon_sym_land] = ACTIONS(1991),
    [anon_sym_lor] = ACTIONS(1991),
    [anon_sym_lxor] = ACTIONS(1991),
    [aux_sym_mult_operator_token1] = ACTIONS(1991),
    [aux_sym_add_operator_token1] = ACTIONS(1991),
    [sym_concat_operator] = ACTIONS(1993),
    [sym_rel_operator] = ACTIONS(1991),
    [anon_sym_AMP_AMP] = ACTIONS(1991),
    [anon_sym_or] = ACTIONS(1991),
    [anon_sym_PIPE_PIPE] = ACTIONS(1991),
    [sym_assign_operator] = ACTIONS(1993),
    [sym_let_operator] = ACTIONS(1993),
    [sym_let_and_operator] = ACTIONS(1993),
    [sym__uppercase_identifier] = ACTIONS(1993),
    [aux_sym_directive_token1] = ACTIONS(1991),
    [aux_sym_tag_token1] = ACTIONS(1993),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1246)] = {
    [sym_attribute] = STATE(1246),
    [ts_builtin_sym_end] = ACTIONS(2001),
    [sym__lowercase_identifier] = ACTIONS(1999),
    [anon_sym_SEMI_SEMI] = ACTIONS(2001),
    [anon_sym_let] = ACTIONS(1999),
    [anon_sym_and] = ACTIONS(1999),
    [anon_sym_TILDE] = ACTIONS(1999),
    [anon_sym_QMARK] = ACTIONS(1999),
    [anon_sym_LPAREN] = ACTIONS(2001),
    [anon_sym_external] = ACTIONS(1999),
    [anon_sym_type] = ACTIONS(1999),
    [anon_sym_COMMA] = ACTIONS(2001),
    [anon_sym_PIPE] = ACTIONS(1999),
    [anon_sym_DOT] = ACTIONS(2001),
    [anon_sym_LBRACK] = ACTIONS(1999),
    [anon_sym_true] = ACTIONS(1999),
    [anon_sym_false] = ACTIONS(1999),
    [anon_sym_LBRACE] = ACTIONS(1999),
    [anon_sym_SEMI] = ACTIONS(1999),
    [anon_sym_exception] = ACTIONS(1999),
    [anon_sym_module] = ACTIONS(1999),
    [anon_sym_open] = ACTIONS(1999),
    [anon_sym_include] = ACTIONS(1999),
    [anon_sym_class] = ACTIONS(1999),
    [anon_sym_val] = ACTIONS(1999),
    [anon_sym_object] = ACTIONS(1999),
    [anon_sym_in] = ACTIONS(1999),
    [anon_sym_AMP] = ACTIONS(1999),
    [anon_sym_POUND] = ACTIONS(1999),
    [anon_sym_COLON_COLON] = ACTIONS(2001),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2001),
    [anon_sym_else] = ACTIONS(1999),
    [anon_sym_new] = ACTIONS(1999),
    [anon_sym_LBRACE_LT] = ACTIONS(2001),
    [anon_sym_begin] = ACTIONS(1999),
    [sym_ocamlyacc_value] = ACTIONS(2001),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2001),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1999),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1999),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2001),
    [aux_sym_number_token1] = ACTIONS(2001),
    [aux_sym_signed_number_token1] = ACTIONS(1999),
    [anon_sym_SQUOTE] = ACTIONS(2001),
    [anon_sym_DQUOTE] = ACTIONS(2001),
    [sym_prefix_operator] = ACTIONS(2001),
    [aux_sym_sign_operator_token1] = ACTIONS(1999),
    [sym_hash_operator] = ACTIONS(2001),
    [anon_sym_lsl] = ACTIONS(1999),
    [anon_sym_lsr] = ACTIONS(1999),
    [anon_sym_asr] = ACTIONS(1999),
    [aux_sym_pow_operator_token1] = ACTIONS(2001),
    [anon_sym_mod] = ACTIONS(1999),
    [anon_sym_land] = ACTIONS(1999),
    [anon_sym_lor] = ACTIONS(1999),
    [anon_sym_lxor] = ACTIONS(1999),
    [aux_sym_mult_operator_token1] = ACTIONS(1999),
    [aux_sym_add_operator_token1] = ACTIONS(1999),
    [sym_concat_operator] = ACTIONS(2001),
    [sym_rel_operator] = ACTIONS(1999),
    [anon_sym_AMP_AMP] = ACTIONS(1999),
    [anon_sym_or] = ACTIONS(1999),
    [anon_sym_PIPE_PIPE] = ACTIONS(1999),
    [sym_assign_operator] = ACTIONS(2001),
    [sym_let_operator] = ACTIONS(2001),
    [sym_let_and_operator] = ACTIONS(2001),
    [sym__uppercase_identifier] = ACTIONS(2001),
    [aux_sym_directive_token1] = ACTIONS(1999),
    [aux_sym_tag_token1] = ACTIONS(2001),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1247)] = {
    [sym_attribute] = STATE(1247),
    [ts_builtin_sym_end] = ACTIONS(2163),
    [sym__lowercase_identifier] = ACTIONS(2161),
    [anon_sym_SEMI_SEMI] = ACTIONS(2163),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_and] = ACTIONS(2161),
    [anon_sym_TILDE] = ACTIONS(2161),
    [anon_sym_QMARK] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_external] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_COMMA] = ACTIONS(2163),
    [anon_sym_PIPE] = ACTIONS(2161),
    [anon_sym_DOT] = ACTIONS(2163),
    [anon_sym_LBRACK] = ACTIONS(2161),
    [anon_sym_true] = ACTIONS(2161),
    [anon_sym_false] = ACTIONS(2161),
    [anon_sym_LBRACE] = ACTIONS(2161),
    [anon_sym_SEMI] = ACTIONS(2161),
    [anon_sym_exception] = ACTIONS(2161),
    [anon_sym_module] = ACTIONS(2161),
    [anon_sym_open] = ACTIONS(2161),
    [anon_sym_include] = ACTIONS(2161),
    [anon_sym_class] = ACTIONS(2161),
    [anon_sym_val] = ACTIONS(2161),
    [anon_sym_object] = ACTIONS(2161),
    [anon_sym_in] = ACTIONS(2161),
    [anon_sym_AMP] = ACTIONS(2161),
    [anon_sym_POUND] = ACTIONS(2161),
    [anon_sym_COLON_COLON] = ACTIONS(2163),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2163),
    [anon_sym_else] = ACTIONS(2161),
    [anon_sym_new] = ACTIONS(2161),
    [anon_sym_LBRACE_LT] = ACTIONS(2163),
    [anon_sym_begin] = ACTIONS(2161),
    [sym_ocamlyacc_value] = ACTIONS(2163),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2161),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2163),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2163),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2163),
    [aux_sym_number_token1] = ACTIONS(2163),
    [aux_sym_signed_number_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2163),
    [anon_sym_DQUOTE] = ACTIONS(2163),
    [sym_prefix_operator] = ACTIONS(2163),
    [aux_sym_sign_operator_token1] = ACTIONS(2161),
    [sym_hash_operator] = ACTIONS(2163),
    [anon_sym_lsl] = ACTIONS(2161),
    [anon_sym_lsr] = ACTIONS(2161),
    [anon_sym_asr] = ACTIONS(2161),
    [aux_sym_pow_operator_token1] = ACTIONS(2163),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_land] = ACTIONS(2161),
    [anon_sym_lor] = ACTIONS(2161),
    [anon_sym_lxor] = ACTIONS(2161),
    [aux_sym_mult_operator_token1] = ACTIONS(2161),
    [aux_sym_add_operator_token1] = ACTIONS(2161),
    [sym_concat_operator] = ACTIONS(2163),
    [sym_rel_operator] = ACTIONS(2161),
    [anon_sym_AMP_AMP] = ACTIONS(2161),
    [anon_sym_or] = ACTIONS(2161),
    [anon_sym_PIPE_PIPE] = ACTIONS(2161),
    [sym_assign_operator] = ACTIONS(2163),
    [sym_let_operator] = ACTIONS(2163),
    [sym_let_and_operator] = ACTIONS(2163),
    [sym__uppercase_identifier] = ACTIONS(2163),
    [aux_sym_directive_token1] = ACTIONS(2161),
    [aux_sym_tag_token1] = ACTIONS(2163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1248)] = {
    [sym_attribute] = STATE(1248),
    [ts_builtin_sym_end] = ACTIONS(2009),
    [sym__lowercase_identifier] = ACTIONS(2007),
    [anon_sym_SEMI_SEMI] = ACTIONS(2009),
    [anon_sym_let] = ACTIONS(2007),
    [anon_sym_and] = ACTIONS(2007),
    [anon_sym_TILDE] = ACTIONS(2007),
    [anon_sym_QMARK] = ACTIONS(2007),
    [anon_sym_LPAREN] = ACTIONS(2009),
    [anon_sym_external] = ACTIONS(2007),
    [anon_sym_type] = ACTIONS(2007),
    [anon_sym_COMMA] = ACTIONS(2009),
    [anon_sym_PIPE] = ACTIONS(2007),
    [anon_sym_DOT] = ACTIONS(2009),
    [anon_sym_LBRACK] = ACTIONS(2007),
    [anon_sym_true] = ACTIONS(2007),
    [anon_sym_false] = ACTIONS(2007),
    [anon_sym_LBRACE] = ACTIONS(2007),
    [anon_sym_SEMI] = ACTIONS(2007),
    [anon_sym_exception] = ACTIONS(2007),
    [anon_sym_module] = ACTIONS(2007),
    [anon_sym_open] = ACTIONS(2007),
    [anon_sym_include] = ACTIONS(2007),
    [anon_sym_class] = ACTIONS(2007),
    [anon_sym_val] = ACTIONS(2007),
    [anon_sym_object] = ACTIONS(2007),
    [anon_sym_in] = ACTIONS(2007),
    [anon_sym_AMP] = ACTIONS(2007),
    [anon_sym_POUND] = ACTIONS(2007),
    [anon_sym_COLON_COLON] = ACTIONS(2009),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2009),
    [anon_sym_else] = ACTIONS(2007),
    [anon_sym_new] = ACTIONS(2007),
    [anon_sym_LBRACE_LT] = ACTIONS(2009),
    [anon_sym_begin] = ACTIONS(2007),
    [sym_ocamlyacc_value] = ACTIONS(2009),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2007),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2009),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2007),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2007),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2009),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2009),
    [aux_sym_number_token1] = ACTIONS(2009),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2009),
    [sym_prefix_operator] = ACTIONS(2009),
    [aux_sym_sign_operator_token1] = ACTIONS(2007),
    [sym_hash_operator] = ACTIONS(2009),
    [anon_sym_lsl] = ACTIONS(2007),
    [anon_sym_lsr] = ACTIONS(2007),
    [anon_sym_asr] = ACTIONS(2007),
    [aux_sym_pow_operator_token1] = ACTIONS(2009),
    [anon_sym_mod] = ACTIONS(2007),
    [anon_sym_land] = ACTIONS(2007),
    [anon_sym_lor] = ACTIONS(2007),
    [anon_sym_lxor] = ACTIONS(2007),
    [aux_sym_mult_operator_token1] = ACTIONS(2007),
    [aux_sym_add_operator_token1] = ACTIONS(2007),
    [sym_concat_operator] = ACTIONS(2009),
    [sym_rel_operator] = ACTIONS(2007),
    [anon_sym_AMP_AMP] = ACTIONS(2007),
    [anon_sym_or] = ACTIONS(2007),
    [anon_sym_PIPE_PIPE] = ACTIONS(2007),
    [sym_assign_operator] = ACTIONS(2009),
    [sym_let_operator] = ACTIONS(2009),
    [sym_let_and_operator] = ACTIONS(2009),
    [sym__uppercase_identifier] = ACTIONS(2009),
    [aux_sym_directive_token1] = ACTIONS(2007),
    [aux_sym_tag_token1] = ACTIONS(2009),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1249)] = {
    [sym_attribute] = STATE(1249),
    [ts_builtin_sym_end] = ACTIONS(2013),
    [sym__lowercase_identifier] = ACTIONS(2011),
    [anon_sym_SEMI_SEMI] = ACTIONS(2013),
    [anon_sym_let] = ACTIONS(2011),
    [anon_sym_and] = ACTIONS(2011),
    [anon_sym_TILDE] = ACTIONS(2011),
    [anon_sym_QMARK] = ACTIONS(2011),
    [anon_sym_LPAREN] = ACTIONS(2013),
    [anon_sym_external] = ACTIONS(2011),
    [anon_sym_type] = ACTIONS(2011),
    [anon_sym_COMMA] = ACTIONS(2013),
    [anon_sym_PIPE] = ACTIONS(2011),
    [anon_sym_DOT] = ACTIONS(2013),
    [anon_sym_LBRACK] = ACTIONS(2011),
    [anon_sym_true] = ACTIONS(2011),
    [anon_sym_false] = ACTIONS(2011),
    [anon_sym_LBRACE] = ACTIONS(2011),
    [anon_sym_SEMI] = ACTIONS(2011),
    [anon_sym_exception] = ACTIONS(2011),
    [anon_sym_module] = ACTIONS(2011),
    [anon_sym_open] = ACTIONS(2011),
    [anon_sym_include] = ACTIONS(2011),
    [anon_sym_class] = ACTIONS(2011),
    [anon_sym_val] = ACTIONS(2011),
    [anon_sym_object] = ACTIONS(2011),
    [anon_sym_in] = ACTIONS(2011),
    [anon_sym_AMP] = ACTIONS(2011),
    [anon_sym_POUND] = ACTIONS(2011),
    [anon_sym_COLON_COLON] = ACTIONS(2013),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2013),
    [anon_sym_else] = ACTIONS(2011),
    [anon_sym_new] = ACTIONS(2011),
    [anon_sym_LBRACE_LT] = ACTIONS(2013),
    [anon_sym_begin] = ACTIONS(2011),
    [sym_ocamlyacc_value] = ACTIONS(2013),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2011),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2013),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2011),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2011),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2013),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2013),
    [aux_sym_number_token1] = ACTIONS(2013),
    [aux_sym_signed_number_token1] = ACTIONS(2011),
    [anon_sym_SQUOTE] = ACTIONS(2013),
    [anon_sym_DQUOTE] = ACTIONS(2013),
    [sym_prefix_operator] = ACTIONS(2013),
    [aux_sym_sign_operator_token1] = ACTIONS(2011),
    [sym_hash_operator] = ACTIONS(2013),
    [anon_sym_lsl] = ACTIONS(2011),
    [anon_sym_lsr] = ACTIONS(2011),
    [anon_sym_asr] = ACTIONS(2011),
    [aux_sym_pow_operator_token1] = ACTIONS(2013),
    [anon_sym_mod] = ACTIONS(2011),
    [anon_sym_land] = ACTIONS(2011),
    [anon_sym_lor] = ACTIONS(2011),
    [anon_sym_lxor] = ACTIONS(2011),
    [aux_sym_mult_operator_token1] = ACTIONS(2011),
    [aux_sym_add_operator_token1] = ACTIONS(2011),
    [sym_concat_operator] = ACTIONS(2013),
    [sym_rel_operator] = ACTIONS(2011),
    [anon_sym_AMP_AMP] = ACTIONS(2011),
    [anon_sym_or] = ACTIONS(2011),
    [anon_sym_PIPE_PIPE] = ACTIONS(2011),
    [sym_assign_operator] = ACTIONS(2013),
    [sym_let_operator] = ACTIONS(2013),
    [sym_let_and_operator] = ACTIONS(2013),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_directive_token1] = ACTIONS(2011),
    [aux_sym_tag_token1] = ACTIONS(2013),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1250)] = {
    [sym_attribute] = STATE(1250),
    [ts_builtin_sym_end] = ACTIONS(2113),
    [sym__lowercase_identifier] = ACTIONS(2111),
    [anon_sym_SEMI_SEMI] = ACTIONS(2113),
    [anon_sym_let] = ACTIONS(2111),
    [anon_sym_and] = ACTIONS(2111),
    [anon_sym_TILDE] = ACTIONS(2111),
    [anon_sym_QMARK] = ACTIONS(2111),
    [anon_sym_LPAREN] = ACTIONS(2113),
    [anon_sym_external] = ACTIONS(2111),
    [anon_sym_type] = ACTIONS(2111),
    [anon_sym_COMMA] = ACTIONS(2113),
    [anon_sym_PIPE] = ACTIONS(2111),
    [anon_sym_DOT] = ACTIONS(2177),
    [anon_sym_LBRACK] = ACTIONS(2111),
    [anon_sym_true] = ACTIONS(2111),
    [anon_sym_false] = ACTIONS(2111),
    [anon_sym_LBRACE] = ACTIONS(2111),
    [anon_sym_SEMI] = ACTIONS(2111),
    [anon_sym_exception] = ACTIONS(2111),
    [anon_sym_module] = ACTIONS(2111),
    [anon_sym_open] = ACTIONS(2111),
    [anon_sym_include] = ACTIONS(2111),
    [anon_sym_class] = ACTIONS(2111),
    [anon_sym_val] = ACTIONS(2111),
    [anon_sym_object] = ACTIONS(2111),
    [anon_sym_in] = ACTIONS(2111),
    [anon_sym_AMP] = ACTIONS(2111),
    [anon_sym_POUND] = ACTIONS(2111),
    [anon_sym_COLON_COLON] = ACTIONS(2113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2113),
    [anon_sym_else] = ACTIONS(2111),
    [anon_sym_new] = ACTIONS(2111),
    [anon_sym_LBRACE_LT] = ACTIONS(2113),
    [anon_sym_begin] = ACTIONS(2111),
    [sym_ocamlyacc_value] = ACTIONS(2113),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2113),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2113),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2113),
    [aux_sym_number_token1] = ACTIONS(2113),
    [aux_sym_signed_number_token1] = ACTIONS(2111),
    [anon_sym_SQUOTE] = ACTIONS(2113),
    [anon_sym_DQUOTE] = ACTIONS(2113),
    [sym_prefix_operator] = ACTIONS(2113),
    [aux_sym_sign_operator_token1] = ACTIONS(2111),
    [sym_hash_operator] = ACTIONS(2113),
    [anon_sym_lsl] = ACTIONS(2111),
    [anon_sym_lsr] = ACTIONS(2111),
    [anon_sym_asr] = ACTIONS(2111),
    [aux_sym_pow_operator_token1] = ACTIONS(2113),
    [anon_sym_mod] = ACTIONS(2111),
    [anon_sym_land] = ACTIONS(2111),
    [anon_sym_lor] = ACTIONS(2111),
    [anon_sym_lxor] = ACTIONS(2111),
    [aux_sym_mult_operator_token1] = ACTIONS(2111),
    [aux_sym_add_operator_token1] = ACTIONS(2111),
    [sym_concat_operator] = ACTIONS(2113),
    [sym_rel_operator] = ACTIONS(2111),
    [anon_sym_AMP_AMP] = ACTIONS(2111),
    [anon_sym_or] = ACTIONS(2111),
    [anon_sym_PIPE_PIPE] = ACTIONS(2111),
    [sym_assign_operator] = ACTIONS(2113),
    [sym_let_operator] = ACTIONS(2113),
    [sym_let_and_operator] = ACTIONS(2113),
    [sym__uppercase_identifier] = ACTIONS(2113),
    [aux_sym_directive_token1] = ACTIONS(2111),
    [aux_sym_tag_token1] = ACTIONS(2113),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1251)] = {
    [sym_attribute] = STATE(1251),
    [ts_builtin_sym_end] = ACTIONS(2017),
    [sym__lowercase_identifier] = ACTIONS(2015),
    [anon_sym_SEMI_SEMI] = ACTIONS(2017),
    [anon_sym_let] = ACTIONS(2015),
    [anon_sym_and] = ACTIONS(2015),
    [anon_sym_TILDE] = ACTIONS(2015),
    [anon_sym_QMARK] = ACTIONS(2015),
    [anon_sym_LPAREN] = ACTIONS(2017),
    [anon_sym_external] = ACTIONS(2015),
    [anon_sym_type] = ACTIONS(2015),
    [anon_sym_COMMA] = ACTIONS(2017),
    [anon_sym_PIPE] = ACTIONS(2015),
    [anon_sym_DOT] = ACTIONS(2017),
    [anon_sym_LBRACK] = ACTIONS(2015),
    [anon_sym_true] = ACTIONS(2015),
    [anon_sym_false] = ACTIONS(2015),
    [anon_sym_LBRACE] = ACTIONS(2015),
    [anon_sym_SEMI] = ACTIONS(2015),
    [anon_sym_exception] = ACTIONS(2015),
    [anon_sym_module] = ACTIONS(2015),
    [anon_sym_open] = ACTIONS(2015),
    [anon_sym_include] = ACTIONS(2015),
    [anon_sym_class] = ACTIONS(2015),
    [anon_sym_val] = ACTIONS(2015),
    [anon_sym_object] = ACTIONS(2015),
    [anon_sym_in] = ACTIONS(2015),
    [anon_sym_AMP] = ACTIONS(2015),
    [anon_sym_POUND] = ACTIONS(2015),
    [anon_sym_COLON_COLON] = ACTIONS(2017),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2017),
    [anon_sym_else] = ACTIONS(2015),
    [anon_sym_new] = ACTIONS(2015),
    [anon_sym_LBRACE_LT] = ACTIONS(2017),
    [anon_sym_begin] = ACTIONS(2015),
    [sym_ocamlyacc_value] = ACTIONS(2017),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2015),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2017),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2015),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2015),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2017),
    [aux_sym_number_token1] = ACTIONS(2017),
    [aux_sym_signed_number_token1] = ACTIONS(2015),
    [anon_sym_SQUOTE] = ACTIONS(2017),
    [anon_sym_DQUOTE] = ACTIONS(2017),
    [sym_prefix_operator] = ACTIONS(2017),
    [aux_sym_sign_operator_token1] = ACTIONS(2015),
    [sym_hash_operator] = ACTIONS(2017),
    [anon_sym_lsl] = ACTIONS(2015),
    [anon_sym_lsr] = ACTIONS(2015),
    [anon_sym_asr] = ACTIONS(2015),
    [aux_sym_pow_operator_token1] = ACTIONS(2017),
    [anon_sym_mod] = ACTIONS(2015),
    [anon_sym_land] = ACTIONS(2015),
    [anon_sym_lor] = ACTIONS(2015),
    [anon_sym_lxor] = ACTIONS(2015),
    [aux_sym_mult_operator_token1] = ACTIONS(2015),
    [aux_sym_add_operator_token1] = ACTIONS(2015),
    [sym_concat_operator] = ACTIONS(2017),
    [sym_rel_operator] = ACTIONS(2015),
    [anon_sym_AMP_AMP] = ACTIONS(2015),
    [anon_sym_or] = ACTIONS(2015),
    [anon_sym_PIPE_PIPE] = ACTIONS(2015),
    [sym_assign_operator] = ACTIONS(2017),
    [sym_let_operator] = ACTIONS(2017),
    [sym_let_and_operator] = ACTIONS(2017),
    [sym__uppercase_identifier] = ACTIONS(2017),
    [aux_sym_directive_token1] = ACTIONS(2015),
    [aux_sym_tag_token1] = ACTIONS(2017),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1252)] = {
    [sym_attribute] = STATE(1252),
    [ts_builtin_sym_end] = ACTIONS(2181),
    [sym__lowercase_identifier] = ACTIONS(2179),
    [anon_sym_SEMI_SEMI] = ACTIONS(2181),
    [anon_sym_let] = ACTIONS(2179),
    [anon_sym_and] = ACTIONS(2179),
    [anon_sym_TILDE] = ACTIONS(2179),
    [anon_sym_QMARK] = ACTIONS(2179),
    [anon_sym_LPAREN] = ACTIONS(2181),
    [anon_sym_external] = ACTIONS(2179),
    [anon_sym_type] = ACTIONS(2179),
    [anon_sym_COMMA] = ACTIONS(2181),
    [anon_sym_PIPE] = ACTIONS(2179),
    [anon_sym_DOT] = ACTIONS(2181),
    [anon_sym_LBRACK] = ACTIONS(2179),
    [anon_sym_true] = ACTIONS(2179),
    [anon_sym_false] = ACTIONS(2179),
    [anon_sym_LBRACE] = ACTIONS(2179),
    [anon_sym_SEMI] = ACTIONS(2179),
    [anon_sym_exception] = ACTIONS(2179),
    [anon_sym_module] = ACTIONS(2179),
    [anon_sym_open] = ACTIONS(2179),
    [anon_sym_include] = ACTIONS(2179),
    [anon_sym_class] = ACTIONS(2179),
    [anon_sym_val] = ACTIONS(2179),
    [anon_sym_object] = ACTIONS(2179),
    [anon_sym_in] = ACTIONS(2179),
    [anon_sym_AMP] = ACTIONS(2179),
    [anon_sym_POUND] = ACTIONS(2179),
    [anon_sym_COLON_COLON] = ACTIONS(2181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2181),
    [anon_sym_else] = ACTIONS(2179),
    [anon_sym_new] = ACTIONS(2179),
    [anon_sym_LBRACE_LT] = ACTIONS(2181),
    [anon_sym_begin] = ACTIONS(2179),
    [sym_ocamlyacc_value] = ACTIONS(2181),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2179),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2181),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2179),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2181),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2181),
    [aux_sym_number_token1] = ACTIONS(2181),
    [aux_sym_signed_number_token1] = ACTIONS(2179),
    [anon_sym_SQUOTE] = ACTIONS(2181),
    [anon_sym_DQUOTE] = ACTIONS(2181),
    [sym_prefix_operator] = ACTIONS(2181),
    [aux_sym_sign_operator_token1] = ACTIONS(2179),
    [sym_hash_operator] = ACTIONS(2181),
    [anon_sym_lsl] = ACTIONS(2179),
    [anon_sym_lsr] = ACTIONS(2179),
    [anon_sym_asr] = ACTIONS(2179),
    [aux_sym_pow_operator_token1] = ACTIONS(2181),
    [anon_sym_mod] = ACTIONS(2179),
    [anon_sym_land] = ACTIONS(2179),
    [anon_sym_lor] = ACTIONS(2179),
    [anon_sym_lxor] = ACTIONS(2179),
    [aux_sym_mult_operator_token1] = ACTIONS(2179),
    [aux_sym_add_operator_token1] = ACTIONS(2179),
    [sym_concat_operator] = ACTIONS(2181),
    [sym_rel_operator] = ACTIONS(2179),
    [anon_sym_AMP_AMP] = ACTIONS(2179),
    [anon_sym_or] = ACTIONS(2179),
    [anon_sym_PIPE_PIPE] = ACTIONS(2179),
    [sym_assign_operator] = ACTIONS(2181),
    [sym_let_operator] = ACTIONS(2181),
    [sym_let_and_operator] = ACTIONS(2181),
    [sym__uppercase_identifier] = ACTIONS(2181),
    [aux_sym_directive_token1] = ACTIONS(2179),
    [aux_sym_tag_token1] = ACTIONS(2181),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1253)] = {
    [sym_attribute] = STATE(1253),
    [ts_builtin_sym_end] = ACTIONS(2021),
    [sym__lowercase_identifier] = ACTIONS(2019),
    [anon_sym_SEMI_SEMI] = ACTIONS(2021),
    [anon_sym_let] = ACTIONS(2019),
    [anon_sym_and] = ACTIONS(2019),
    [anon_sym_TILDE] = ACTIONS(2019),
    [anon_sym_QMARK] = ACTIONS(2019),
    [anon_sym_LPAREN] = ACTIONS(2021),
    [anon_sym_external] = ACTIONS(2019),
    [anon_sym_type] = ACTIONS(2019),
    [anon_sym_COMMA] = ACTIONS(2021),
    [anon_sym_PIPE] = ACTIONS(2019),
    [anon_sym_DOT] = ACTIONS(2021),
    [anon_sym_LBRACK] = ACTIONS(2019),
    [anon_sym_true] = ACTIONS(2019),
    [anon_sym_false] = ACTIONS(2019),
    [anon_sym_LBRACE] = ACTIONS(2019),
    [anon_sym_SEMI] = ACTIONS(2019),
    [anon_sym_exception] = ACTIONS(2019),
    [anon_sym_module] = ACTIONS(2019),
    [anon_sym_open] = ACTIONS(2019),
    [anon_sym_include] = ACTIONS(2019),
    [anon_sym_class] = ACTIONS(2019),
    [anon_sym_val] = ACTIONS(2019),
    [anon_sym_object] = ACTIONS(2019),
    [anon_sym_in] = ACTIONS(2019),
    [anon_sym_AMP] = ACTIONS(2019),
    [anon_sym_POUND] = ACTIONS(2019),
    [anon_sym_COLON_COLON] = ACTIONS(2021),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2021),
    [anon_sym_else] = ACTIONS(2019),
    [anon_sym_new] = ACTIONS(2019),
    [anon_sym_LBRACE_LT] = ACTIONS(2021),
    [anon_sym_begin] = ACTIONS(2019),
    [sym_ocamlyacc_value] = ACTIONS(2021),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2019),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2021),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2021),
    [aux_sym_number_token1] = ACTIONS(2021),
    [aux_sym_signed_number_token1] = ACTIONS(2019),
    [anon_sym_SQUOTE] = ACTIONS(2021),
    [anon_sym_DQUOTE] = ACTIONS(2021),
    [sym_prefix_operator] = ACTIONS(2021),
    [aux_sym_sign_operator_token1] = ACTIONS(2019),
    [sym_hash_operator] = ACTIONS(2021),
    [anon_sym_lsl] = ACTIONS(2019),
    [anon_sym_lsr] = ACTIONS(2019),
    [anon_sym_asr] = ACTIONS(2019),
    [aux_sym_pow_operator_token1] = ACTIONS(2021),
    [anon_sym_mod] = ACTIONS(2019),
    [anon_sym_land] = ACTIONS(2019),
    [anon_sym_lor] = ACTIONS(2019),
    [anon_sym_lxor] = ACTIONS(2019),
    [aux_sym_mult_operator_token1] = ACTIONS(2019),
    [aux_sym_add_operator_token1] = ACTIONS(2019),
    [sym_concat_operator] = ACTIONS(2021),
    [sym_rel_operator] = ACTIONS(2019),
    [anon_sym_AMP_AMP] = ACTIONS(2019),
    [anon_sym_or] = ACTIONS(2019),
    [anon_sym_PIPE_PIPE] = ACTIONS(2019),
    [sym_assign_operator] = ACTIONS(2021),
    [sym_let_operator] = ACTIONS(2021),
    [sym_let_and_operator] = ACTIONS(2021),
    [sym__uppercase_identifier] = ACTIONS(2021),
    [aux_sym_directive_token1] = ACTIONS(2019),
    [aux_sym_tag_token1] = ACTIONS(2021),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1254)] = {
    [sym_attribute] = STATE(1254),
    [ts_builtin_sym_end] = ACTIONS(2025),
    [sym__lowercase_identifier] = ACTIONS(2023),
    [anon_sym_SEMI_SEMI] = ACTIONS(2025),
    [anon_sym_let] = ACTIONS(2023),
    [anon_sym_and] = ACTIONS(2023),
    [anon_sym_TILDE] = ACTIONS(2023),
    [anon_sym_QMARK] = ACTIONS(2023),
    [anon_sym_LPAREN] = ACTIONS(2025),
    [anon_sym_external] = ACTIONS(2023),
    [anon_sym_type] = ACTIONS(2023),
    [anon_sym_COMMA] = ACTIONS(2025),
    [anon_sym_PIPE] = ACTIONS(2023),
    [anon_sym_DOT] = ACTIONS(2025),
    [anon_sym_LBRACK] = ACTIONS(2023),
    [anon_sym_true] = ACTIONS(2023),
    [anon_sym_false] = ACTIONS(2023),
    [anon_sym_LBRACE] = ACTIONS(2023),
    [anon_sym_SEMI] = ACTIONS(2023),
    [anon_sym_exception] = ACTIONS(2023),
    [anon_sym_module] = ACTIONS(2023),
    [anon_sym_open] = ACTIONS(2023),
    [anon_sym_include] = ACTIONS(2023),
    [anon_sym_class] = ACTIONS(2023),
    [anon_sym_val] = ACTIONS(2023),
    [anon_sym_object] = ACTIONS(2023),
    [anon_sym_in] = ACTIONS(2023),
    [anon_sym_AMP] = ACTIONS(2023),
    [anon_sym_POUND] = ACTIONS(2023),
    [anon_sym_COLON_COLON] = ACTIONS(2025),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2025),
    [anon_sym_else] = ACTIONS(2023),
    [anon_sym_new] = ACTIONS(2023),
    [anon_sym_LBRACE_LT] = ACTIONS(2025),
    [anon_sym_begin] = ACTIONS(2023),
    [sym_ocamlyacc_value] = ACTIONS(2025),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2023),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2025),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2025),
    [aux_sym_number_token1] = ACTIONS(2025),
    [aux_sym_signed_number_token1] = ACTIONS(2023),
    [anon_sym_SQUOTE] = ACTIONS(2025),
    [anon_sym_DQUOTE] = ACTIONS(2025),
    [sym_prefix_operator] = ACTIONS(2025),
    [aux_sym_sign_operator_token1] = ACTIONS(2023),
    [sym_hash_operator] = ACTIONS(2025),
    [anon_sym_lsl] = ACTIONS(2023),
    [anon_sym_lsr] = ACTIONS(2023),
    [anon_sym_asr] = ACTIONS(2023),
    [aux_sym_pow_operator_token1] = ACTIONS(2025),
    [anon_sym_mod] = ACTIONS(2023),
    [anon_sym_land] = ACTIONS(2023),
    [anon_sym_lor] = ACTIONS(2023),
    [anon_sym_lxor] = ACTIONS(2023),
    [aux_sym_mult_operator_token1] = ACTIONS(2023),
    [aux_sym_add_operator_token1] = ACTIONS(2023),
    [sym_concat_operator] = ACTIONS(2025),
    [sym_rel_operator] = ACTIONS(2023),
    [anon_sym_AMP_AMP] = ACTIONS(2023),
    [anon_sym_or] = ACTIONS(2023),
    [anon_sym_PIPE_PIPE] = ACTIONS(2023),
    [sym_assign_operator] = ACTIONS(2025),
    [sym_let_operator] = ACTIONS(2025),
    [sym_let_and_operator] = ACTIONS(2025),
    [sym__uppercase_identifier] = ACTIONS(2025),
    [aux_sym_directive_token1] = ACTIONS(2023),
    [aux_sym_tag_token1] = ACTIONS(2025),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1255)] = {
    [sym_attribute] = STATE(1255),
    [ts_builtin_sym_end] = ACTIONS(2029),
    [sym__lowercase_identifier] = ACTIONS(2027),
    [anon_sym_SEMI_SEMI] = ACTIONS(2029),
    [anon_sym_let] = ACTIONS(2027),
    [anon_sym_and] = ACTIONS(2027),
    [anon_sym_TILDE] = ACTIONS(2027),
    [anon_sym_QMARK] = ACTIONS(2027),
    [anon_sym_LPAREN] = ACTIONS(2029),
    [anon_sym_external] = ACTIONS(2027),
    [anon_sym_type] = ACTIONS(2027),
    [anon_sym_COMMA] = ACTIONS(2029),
    [anon_sym_PIPE] = ACTIONS(2027),
    [anon_sym_DOT] = ACTIONS(2029),
    [anon_sym_LBRACK] = ACTIONS(2027),
    [anon_sym_true] = ACTIONS(2027),
    [anon_sym_false] = ACTIONS(2027),
    [anon_sym_LBRACE] = ACTIONS(2027),
    [anon_sym_SEMI] = ACTIONS(2027),
    [anon_sym_exception] = ACTIONS(2027),
    [anon_sym_module] = ACTIONS(2027),
    [anon_sym_open] = ACTIONS(2027),
    [anon_sym_include] = ACTIONS(2027),
    [anon_sym_class] = ACTIONS(2027),
    [anon_sym_val] = ACTIONS(2027),
    [anon_sym_object] = ACTIONS(2027),
    [anon_sym_in] = ACTIONS(2027),
    [anon_sym_AMP] = ACTIONS(2027),
    [anon_sym_POUND] = ACTIONS(2027),
    [anon_sym_COLON_COLON] = ACTIONS(2029),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2029),
    [anon_sym_else] = ACTIONS(2027),
    [anon_sym_new] = ACTIONS(2027),
    [anon_sym_LBRACE_LT] = ACTIONS(2029),
    [anon_sym_begin] = ACTIONS(2027),
    [sym_ocamlyacc_value] = ACTIONS(2029),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2027),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2029),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2029),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2029),
    [aux_sym_number_token1] = ACTIONS(2029),
    [aux_sym_signed_number_token1] = ACTIONS(2027),
    [anon_sym_SQUOTE] = ACTIONS(2029),
    [anon_sym_DQUOTE] = ACTIONS(2029),
    [sym_prefix_operator] = ACTIONS(2029),
    [aux_sym_sign_operator_token1] = ACTIONS(2027),
    [sym_hash_operator] = ACTIONS(2029),
    [anon_sym_lsl] = ACTIONS(2027),
    [anon_sym_lsr] = ACTIONS(2027),
    [anon_sym_asr] = ACTIONS(2027),
    [aux_sym_pow_operator_token1] = ACTIONS(2029),
    [anon_sym_mod] = ACTIONS(2027),
    [anon_sym_land] = ACTIONS(2027),
    [anon_sym_lor] = ACTIONS(2027),
    [anon_sym_lxor] = ACTIONS(2027),
    [aux_sym_mult_operator_token1] = ACTIONS(2027),
    [aux_sym_add_operator_token1] = ACTIONS(2027),
    [sym_concat_operator] = ACTIONS(2029),
    [sym_rel_operator] = ACTIONS(2027),
    [anon_sym_AMP_AMP] = ACTIONS(2027),
    [anon_sym_or] = ACTIONS(2027),
    [anon_sym_PIPE_PIPE] = ACTIONS(2027),
    [sym_assign_operator] = ACTIONS(2029),
    [sym_let_operator] = ACTIONS(2029),
    [sym_let_and_operator] = ACTIONS(2029),
    [sym__uppercase_identifier] = ACTIONS(2029),
    [aux_sym_directive_token1] = ACTIONS(2027),
    [aux_sym_tag_token1] = ACTIONS(2029),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1256)] = {
    [sym_attribute] = STATE(1256),
    [ts_builtin_sym_end] = ACTIONS(1925),
    [sym__lowercase_identifier] = ACTIONS(1923),
    [anon_sym_SEMI_SEMI] = ACTIONS(1925),
    [anon_sym_let] = ACTIONS(1923),
    [anon_sym_and] = ACTIONS(1923),
    [anon_sym_TILDE] = ACTIONS(1923),
    [anon_sym_QMARK] = ACTIONS(1923),
    [anon_sym_LPAREN] = ACTIONS(1925),
    [anon_sym_external] = ACTIONS(1923),
    [anon_sym_type] = ACTIONS(1923),
    [anon_sym_COMMA] = ACTIONS(1925),
    [anon_sym_PIPE] = ACTIONS(1923),
    [anon_sym_DOT] = ACTIONS(1925),
    [anon_sym_LBRACK] = ACTIONS(1923),
    [anon_sym_true] = ACTIONS(1923),
    [anon_sym_false] = ACTIONS(1923),
    [anon_sym_LBRACE] = ACTIONS(1923),
    [anon_sym_SEMI] = ACTIONS(1923),
    [anon_sym_exception] = ACTIONS(1923),
    [anon_sym_module] = ACTIONS(1923),
    [anon_sym_open] = ACTIONS(1923),
    [anon_sym_include] = ACTIONS(1923),
    [anon_sym_class] = ACTIONS(1923),
    [anon_sym_val] = ACTIONS(1923),
    [anon_sym_object] = ACTIONS(1923),
    [anon_sym_in] = ACTIONS(1923),
    [anon_sym_AMP] = ACTIONS(1923),
    [anon_sym_POUND] = ACTIONS(1923),
    [anon_sym_COLON_COLON] = ACTIONS(1925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1925),
    [anon_sym_else] = ACTIONS(1923),
    [anon_sym_new] = ACTIONS(1923),
    [anon_sym_LBRACE_LT] = ACTIONS(1925),
    [anon_sym_begin] = ACTIONS(1923),
    [sym_ocamlyacc_value] = ACTIONS(1925),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1923),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1925),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1923),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1923),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1925),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1925),
    [aux_sym_number_token1] = ACTIONS(1925),
    [aux_sym_signed_number_token1] = ACTIONS(1923),
    [anon_sym_SQUOTE] = ACTIONS(1925),
    [anon_sym_DQUOTE] = ACTIONS(1925),
    [sym_prefix_operator] = ACTIONS(1925),
    [aux_sym_sign_operator_token1] = ACTIONS(1923),
    [sym_hash_operator] = ACTIONS(1925),
    [anon_sym_lsl] = ACTIONS(1923),
    [anon_sym_lsr] = ACTIONS(1923),
    [anon_sym_asr] = ACTIONS(1923),
    [aux_sym_pow_operator_token1] = ACTIONS(1925),
    [anon_sym_mod] = ACTIONS(1923),
    [anon_sym_land] = ACTIONS(1923),
    [anon_sym_lor] = ACTIONS(1923),
    [anon_sym_lxor] = ACTIONS(1923),
    [aux_sym_mult_operator_token1] = ACTIONS(1923),
    [aux_sym_add_operator_token1] = ACTIONS(1923),
    [sym_concat_operator] = ACTIONS(1925),
    [sym_rel_operator] = ACTIONS(1923),
    [anon_sym_AMP_AMP] = ACTIONS(1923),
    [anon_sym_or] = ACTIONS(1923),
    [anon_sym_PIPE_PIPE] = ACTIONS(1923),
    [sym_assign_operator] = ACTIONS(1925),
    [sym_let_operator] = ACTIONS(1925),
    [sym_let_and_operator] = ACTIONS(1925),
    [sym__uppercase_identifier] = ACTIONS(1925),
    [aux_sym_directive_token1] = ACTIONS(1923),
    [aux_sym_tag_token1] = ACTIONS(1925),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1257)] = {
    [sym_attribute] = STATE(1257),
    [ts_builtin_sym_end] = ACTIONS(2037),
    [sym__lowercase_identifier] = ACTIONS(2035),
    [anon_sym_SEMI_SEMI] = ACTIONS(2037),
    [anon_sym_let] = ACTIONS(2035),
    [anon_sym_and] = ACTIONS(2035),
    [anon_sym_TILDE] = ACTIONS(2035),
    [anon_sym_QMARK] = ACTIONS(2035),
    [anon_sym_LPAREN] = ACTIONS(2037),
    [anon_sym_external] = ACTIONS(2035),
    [anon_sym_type] = ACTIONS(2035),
    [anon_sym_COMMA] = ACTIONS(2037),
    [anon_sym_PIPE] = ACTIONS(2035),
    [anon_sym_DOT] = ACTIONS(2037),
    [anon_sym_LBRACK] = ACTIONS(2035),
    [anon_sym_true] = ACTIONS(2035),
    [anon_sym_false] = ACTIONS(2035),
    [anon_sym_LBRACE] = ACTIONS(2035),
    [anon_sym_SEMI] = ACTIONS(2035),
    [anon_sym_exception] = ACTIONS(2035),
    [anon_sym_module] = ACTIONS(2035),
    [anon_sym_open] = ACTIONS(2035),
    [anon_sym_include] = ACTIONS(2035),
    [anon_sym_class] = ACTIONS(2035),
    [anon_sym_val] = ACTIONS(2035),
    [anon_sym_object] = ACTIONS(2035),
    [anon_sym_in] = ACTIONS(2035),
    [anon_sym_AMP] = ACTIONS(2035),
    [anon_sym_POUND] = ACTIONS(2035),
    [anon_sym_COLON_COLON] = ACTIONS(2037),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2037),
    [anon_sym_else] = ACTIONS(2035),
    [anon_sym_new] = ACTIONS(2035),
    [anon_sym_LBRACE_LT] = ACTIONS(2037),
    [anon_sym_begin] = ACTIONS(2035),
    [sym_ocamlyacc_value] = ACTIONS(2037),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2035),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2037),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2035),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2035),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2037),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2037),
    [aux_sym_number_token1] = ACTIONS(2037),
    [aux_sym_signed_number_token1] = ACTIONS(2035),
    [anon_sym_SQUOTE] = ACTIONS(2037),
    [anon_sym_DQUOTE] = ACTIONS(2037),
    [sym_prefix_operator] = ACTIONS(2037),
    [aux_sym_sign_operator_token1] = ACTIONS(2035),
    [sym_hash_operator] = ACTIONS(2037),
    [anon_sym_lsl] = ACTIONS(2035),
    [anon_sym_lsr] = ACTIONS(2035),
    [anon_sym_asr] = ACTIONS(2035),
    [aux_sym_pow_operator_token1] = ACTIONS(2037),
    [anon_sym_mod] = ACTIONS(2035),
    [anon_sym_land] = ACTIONS(2035),
    [anon_sym_lor] = ACTIONS(2035),
    [anon_sym_lxor] = ACTIONS(2035),
    [aux_sym_mult_operator_token1] = ACTIONS(2035),
    [aux_sym_add_operator_token1] = ACTIONS(2035),
    [sym_concat_operator] = ACTIONS(2037),
    [sym_rel_operator] = ACTIONS(2035),
    [anon_sym_AMP_AMP] = ACTIONS(2035),
    [anon_sym_or] = ACTIONS(2035),
    [anon_sym_PIPE_PIPE] = ACTIONS(2035),
    [sym_assign_operator] = ACTIONS(2037),
    [sym_let_operator] = ACTIONS(2037),
    [sym_let_and_operator] = ACTIONS(2037),
    [sym__uppercase_identifier] = ACTIONS(2037),
    [aux_sym_directive_token1] = ACTIONS(2035),
    [aux_sym_tag_token1] = ACTIONS(2037),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1258)] = {
    [sym_attribute] = STATE(1258),
    [sym__lowercase_identifier] = ACTIONS(1895),
    [anon_sym_SEMI_SEMI] = ACTIONS(1897),
    [anon_sym_let] = ACTIONS(1895),
    [anon_sym_and] = ACTIONS(1895),
    [anon_sym_TILDE] = ACTIONS(1895),
    [anon_sym_QMARK] = ACTIONS(1895),
    [anon_sym_LPAREN] = ACTIONS(1897),
    [anon_sym_external] = ACTIONS(1895),
    [anon_sym_type] = ACTIONS(1895),
    [anon_sym_COMMA] = ACTIONS(1897),
    [anon_sym_PIPE] = ACTIONS(1895),
    [anon_sym_DOT] = ACTIONS(1897),
    [anon_sym_LBRACK] = ACTIONS(1895),
    [anon_sym_RBRACK] = ACTIONS(1897),
    [anon_sym_true] = ACTIONS(1895),
    [anon_sym_false] = ACTIONS(1895),
    [anon_sym_LBRACE] = ACTIONS(1895),
    [anon_sym_SEMI] = ACTIONS(1895),
    [anon_sym_exception] = ACTIONS(1895),
    [anon_sym_module] = ACTIONS(1895),
    [anon_sym_open] = ACTIONS(1895),
    [anon_sym_include] = ACTIONS(1895),
    [anon_sym_class] = ACTIONS(1895),
    [anon_sym_val] = ACTIONS(1895),
    [anon_sym_end] = ACTIONS(1895),
    [anon_sym_object] = ACTIONS(1895),
    [anon_sym_AMP] = ACTIONS(1895),
    [anon_sym_POUND] = ACTIONS(1895),
    [anon_sym_COLON_COLON] = ACTIONS(1897),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1897),
    [anon_sym_LT_DASH] = ACTIONS(1895),
    [anon_sym_else] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1895),
    [anon_sym_LBRACE_LT] = ACTIONS(1897),
    [anon_sym_begin] = ACTIONS(1895),
    [sym_ocamlyacc_value] = ACTIONS(1897),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1895),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1897),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1895),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1895),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1897),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1897),
    [aux_sym_number_token1] = ACTIONS(1897),
    [aux_sym_signed_number_token1] = ACTIONS(1895),
    [anon_sym_SQUOTE] = ACTIONS(1897),
    [anon_sym_DQUOTE] = ACTIONS(1897),
    [sym_prefix_operator] = ACTIONS(1897),
    [aux_sym_sign_operator_token1] = ACTIONS(1895),
    [sym_hash_operator] = ACTIONS(1897),
    [anon_sym_lsl] = ACTIONS(1895),
    [anon_sym_lsr] = ACTIONS(1895),
    [anon_sym_asr] = ACTIONS(1895),
    [aux_sym_pow_operator_token1] = ACTIONS(1897),
    [anon_sym_mod] = ACTIONS(1895),
    [anon_sym_land] = ACTIONS(1895),
    [anon_sym_lor] = ACTIONS(1895),
    [anon_sym_lxor] = ACTIONS(1895),
    [aux_sym_mult_operator_token1] = ACTIONS(1895),
    [aux_sym_add_operator_token1] = ACTIONS(1895),
    [sym_concat_operator] = ACTIONS(1897),
    [sym_rel_operator] = ACTIONS(1895),
    [anon_sym_AMP_AMP] = ACTIONS(1895),
    [anon_sym_or] = ACTIONS(1895),
    [anon_sym_PIPE_PIPE] = ACTIONS(1895),
    [sym_assign_operator] = ACTIONS(1897),
    [sym_let_operator] = ACTIONS(1897),
    [sym__uppercase_identifier] = ACTIONS(1897),
    [aux_sym_directive_token1] = ACTIONS(1895),
    [aux_sym_tag_token1] = ACTIONS(1897),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1259)] = {
    [sym_attribute] = STATE(1259),
    [ts_builtin_sym_end] = ACTIONS(2053),
    [sym__lowercase_identifier] = ACTIONS(2051),
    [anon_sym_SEMI_SEMI] = ACTIONS(2053),
    [anon_sym_let] = ACTIONS(2051),
    [anon_sym_and] = ACTIONS(2051),
    [anon_sym_TILDE] = ACTIONS(2051),
    [anon_sym_QMARK] = ACTIONS(2051),
    [anon_sym_LPAREN] = ACTIONS(2053),
    [anon_sym_external] = ACTIONS(2051),
    [anon_sym_type] = ACTIONS(2051),
    [anon_sym_COMMA] = ACTIONS(2053),
    [anon_sym_PIPE] = ACTIONS(2051),
    [anon_sym_DOT] = ACTIONS(2053),
    [anon_sym_LBRACK] = ACTIONS(2051),
    [anon_sym_true] = ACTIONS(2051),
    [anon_sym_false] = ACTIONS(2051),
    [anon_sym_LBRACE] = ACTIONS(2051),
    [anon_sym_SEMI] = ACTIONS(2051),
    [anon_sym_exception] = ACTIONS(2051),
    [anon_sym_module] = ACTIONS(2051),
    [anon_sym_open] = ACTIONS(2051),
    [anon_sym_include] = ACTIONS(2051),
    [anon_sym_class] = ACTIONS(2051),
    [anon_sym_val] = ACTIONS(2051),
    [anon_sym_object] = ACTIONS(2051),
    [anon_sym_in] = ACTIONS(2051),
    [anon_sym_AMP] = ACTIONS(2051),
    [anon_sym_POUND] = ACTIONS(2051),
    [anon_sym_COLON_COLON] = ACTIONS(2053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2053),
    [anon_sym_else] = ACTIONS(2051),
    [anon_sym_new] = ACTIONS(2051),
    [anon_sym_LBRACE_LT] = ACTIONS(2053),
    [anon_sym_begin] = ACTIONS(2051),
    [sym_ocamlyacc_value] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2051),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2053),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2051),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2051),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2053),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2053),
    [aux_sym_number_token1] = ACTIONS(2053),
    [aux_sym_signed_number_token1] = ACTIONS(2051),
    [anon_sym_SQUOTE] = ACTIONS(2053),
    [anon_sym_DQUOTE] = ACTIONS(2053),
    [sym_prefix_operator] = ACTIONS(2053),
    [aux_sym_sign_operator_token1] = ACTIONS(2051),
    [sym_hash_operator] = ACTIONS(2053),
    [anon_sym_lsl] = ACTIONS(2051),
    [anon_sym_lsr] = ACTIONS(2051),
    [anon_sym_asr] = ACTIONS(2051),
    [aux_sym_pow_operator_token1] = ACTIONS(2053),
    [anon_sym_mod] = ACTIONS(2051),
    [anon_sym_land] = ACTIONS(2051),
    [anon_sym_lor] = ACTIONS(2051),
    [anon_sym_lxor] = ACTIONS(2051),
    [aux_sym_mult_operator_token1] = ACTIONS(2051),
    [aux_sym_add_operator_token1] = ACTIONS(2051),
    [sym_concat_operator] = ACTIONS(2053),
    [sym_rel_operator] = ACTIONS(2051),
    [anon_sym_AMP_AMP] = ACTIONS(2051),
    [anon_sym_or] = ACTIONS(2051),
    [anon_sym_PIPE_PIPE] = ACTIONS(2051),
    [sym_assign_operator] = ACTIONS(2053),
    [sym_let_operator] = ACTIONS(2053),
    [sym_let_and_operator] = ACTIONS(2053),
    [sym__uppercase_identifier] = ACTIONS(2053),
    [aux_sym_directive_token1] = ACTIONS(2051),
    [aux_sym_tag_token1] = ACTIONS(2053),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1260)] = {
    [sym_attribute] = STATE(1260),
    [ts_builtin_sym_end] = ACTIONS(2057),
    [sym__lowercase_identifier] = ACTIONS(2055),
    [anon_sym_SEMI_SEMI] = ACTIONS(2057),
    [anon_sym_let] = ACTIONS(2055),
    [anon_sym_and] = ACTIONS(2055),
    [anon_sym_TILDE] = ACTIONS(2055),
    [anon_sym_QMARK] = ACTIONS(2055),
    [anon_sym_LPAREN] = ACTIONS(2057),
    [anon_sym_external] = ACTIONS(2055),
    [anon_sym_type] = ACTIONS(2055),
    [anon_sym_COMMA] = ACTIONS(2057),
    [anon_sym_PIPE] = ACTIONS(2055),
    [anon_sym_DOT] = ACTIONS(2057),
    [anon_sym_LBRACK] = ACTIONS(2055),
    [anon_sym_true] = ACTIONS(2055),
    [anon_sym_false] = ACTIONS(2055),
    [anon_sym_LBRACE] = ACTIONS(2055),
    [anon_sym_SEMI] = ACTIONS(2055),
    [anon_sym_exception] = ACTIONS(2055),
    [anon_sym_module] = ACTIONS(2055),
    [anon_sym_open] = ACTIONS(2055),
    [anon_sym_include] = ACTIONS(2055),
    [anon_sym_class] = ACTIONS(2055),
    [anon_sym_val] = ACTIONS(2055),
    [anon_sym_object] = ACTIONS(2055),
    [anon_sym_in] = ACTIONS(2055),
    [anon_sym_AMP] = ACTIONS(2055),
    [anon_sym_POUND] = ACTIONS(2055),
    [anon_sym_COLON_COLON] = ACTIONS(2057),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2057),
    [anon_sym_else] = ACTIONS(2055),
    [anon_sym_new] = ACTIONS(2055),
    [anon_sym_LBRACE_LT] = ACTIONS(2057),
    [anon_sym_begin] = ACTIONS(2055),
    [sym_ocamlyacc_value] = ACTIONS(2057),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2057),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2055),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2055),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2057),
    [aux_sym_number_token1] = ACTIONS(2057),
    [aux_sym_signed_number_token1] = ACTIONS(2055),
    [anon_sym_SQUOTE] = ACTIONS(2057),
    [anon_sym_DQUOTE] = ACTIONS(2057),
    [sym_prefix_operator] = ACTIONS(2057),
    [aux_sym_sign_operator_token1] = ACTIONS(2055),
    [sym_hash_operator] = ACTIONS(2057),
    [anon_sym_lsl] = ACTIONS(2055),
    [anon_sym_lsr] = ACTIONS(2055),
    [anon_sym_asr] = ACTIONS(2055),
    [aux_sym_pow_operator_token1] = ACTIONS(2057),
    [anon_sym_mod] = ACTIONS(2055),
    [anon_sym_land] = ACTIONS(2055),
    [anon_sym_lor] = ACTIONS(2055),
    [anon_sym_lxor] = ACTIONS(2055),
    [aux_sym_mult_operator_token1] = ACTIONS(2055),
    [aux_sym_add_operator_token1] = ACTIONS(2055),
    [sym_concat_operator] = ACTIONS(2057),
    [sym_rel_operator] = ACTIONS(2055),
    [anon_sym_AMP_AMP] = ACTIONS(2055),
    [anon_sym_or] = ACTIONS(2055),
    [anon_sym_PIPE_PIPE] = ACTIONS(2055),
    [sym_assign_operator] = ACTIONS(2057),
    [sym_let_operator] = ACTIONS(2057),
    [sym_let_and_operator] = ACTIONS(2057),
    [sym__uppercase_identifier] = ACTIONS(2057),
    [aux_sym_directive_token1] = ACTIONS(2055),
    [aux_sym_tag_token1] = ACTIONS(2057),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1261)] = {
    [sym_attribute] = STATE(1261),
    [ts_builtin_sym_end] = ACTIONS(2061),
    [sym__lowercase_identifier] = ACTIONS(2059),
    [anon_sym_SEMI_SEMI] = ACTIONS(2061),
    [anon_sym_let] = ACTIONS(2059),
    [anon_sym_and] = ACTIONS(2059),
    [anon_sym_TILDE] = ACTIONS(2059),
    [anon_sym_QMARK] = ACTIONS(2059),
    [anon_sym_LPAREN] = ACTIONS(2061),
    [anon_sym_external] = ACTIONS(2059),
    [anon_sym_type] = ACTIONS(2059),
    [anon_sym_COMMA] = ACTIONS(2061),
    [anon_sym_PIPE] = ACTIONS(2059),
    [anon_sym_DOT] = ACTIONS(2061),
    [anon_sym_LBRACK] = ACTIONS(2059),
    [anon_sym_true] = ACTIONS(2059),
    [anon_sym_false] = ACTIONS(2059),
    [anon_sym_LBRACE] = ACTIONS(2059),
    [anon_sym_SEMI] = ACTIONS(2059),
    [anon_sym_exception] = ACTIONS(2059),
    [anon_sym_module] = ACTIONS(2059),
    [anon_sym_open] = ACTIONS(2059),
    [anon_sym_include] = ACTIONS(2059),
    [anon_sym_class] = ACTIONS(2059),
    [anon_sym_val] = ACTIONS(2059),
    [anon_sym_object] = ACTIONS(2059),
    [anon_sym_in] = ACTIONS(2059),
    [anon_sym_AMP] = ACTIONS(2059),
    [anon_sym_POUND] = ACTIONS(2059),
    [anon_sym_COLON_COLON] = ACTIONS(2061),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2061),
    [anon_sym_else] = ACTIONS(2059),
    [anon_sym_new] = ACTIONS(2059),
    [anon_sym_LBRACE_LT] = ACTIONS(2061),
    [anon_sym_begin] = ACTIONS(2059),
    [sym_ocamlyacc_value] = ACTIONS(2061),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2059),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2061),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2061),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2061),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2059),
    [anon_sym_SQUOTE] = ACTIONS(2061),
    [anon_sym_DQUOTE] = ACTIONS(2061),
    [sym_prefix_operator] = ACTIONS(2061),
    [aux_sym_sign_operator_token1] = ACTIONS(2059),
    [sym_hash_operator] = ACTIONS(2061),
    [anon_sym_lsl] = ACTIONS(2059),
    [anon_sym_lsr] = ACTIONS(2059),
    [anon_sym_asr] = ACTIONS(2059),
    [aux_sym_pow_operator_token1] = ACTIONS(2061),
    [anon_sym_mod] = ACTIONS(2059),
    [anon_sym_land] = ACTIONS(2059),
    [anon_sym_lor] = ACTIONS(2059),
    [anon_sym_lxor] = ACTIONS(2059),
    [aux_sym_mult_operator_token1] = ACTIONS(2059),
    [aux_sym_add_operator_token1] = ACTIONS(2059),
    [sym_concat_operator] = ACTIONS(2061),
    [sym_rel_operator] = ACTIONS(2059),
    [anon_sym_AMP_AMP] = ACTIONS(2059),
    [anon_sym_or] = ACTIONS(2059),
    [anon_sym_PIPE_PIPE] = ACTIONS(2059),
    [sym_assign_operator] = ACTIONS(2061),
    [sym_let_operator] = ACTIONS(2061),
    [sym_let_and_operator] = ACTIONS(2061),
    [sym__uppercase_identifier] = ACTIONS(2061),
    [aux_sym_directive_token1] = ACTIONS(2059),
    [aux_sym_tag_token1] = ACTIONS(2061),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1262)] = {
    [sym_attribute] = STATE(1262),
    [ts_builtin_sym_end] = ACTIONS(2065),
    [sym__lowercase_identifier] = ACTIONS(2063),
    [anon_sym_SEMI_SEMI] = ACTIONS(2065),
    [anon_sym_let] = ACTIONS(2063),
    [anon_sym_and] = ACTIONS(2063),
    [anon_sym_TILDE] = ACTIONS(2063),
    [anon_sym_QMARK] = ACTIONS(2063),
    [anon_sym_LPAREN] = ACTIONS(2065),
    [anon_sym_external] = ACTIONS(2063),
    [anon_sym_type] = ACTIONS(2063),
    [anon_sym_COMMA] = ACTIONS(2065),
    [anon_sym_PIPE] = ACTIONS(2063),
    [anon_sym_DOT] = ACTIONS(2065),
    [anon_sym_LBRACK] = ACTIONS(2063),
    [anon_sym_true] = ACTIONS(2063),
    [anon_sym_false] = ACTIONS(2063),
    [anon_sym_LBRACE] = ACTIONS(2063),
    [anon_sym_SEMI] = ACTIONS(2063),
    [anon_sym_exception] = ACTIONS(2063),
    [anon_sym_module] = ACTIONS(2063),
    [anon_sym_open] = ACTIONS(2063),
    [anon_sym_include] = ACTIONS(2063),
    [anon_sym_class] = ACTIONS(2063),
    [anon_sym_val] = ACTIONS(2063),
    [anon_sym_object] = ACTIONS(2063),
    [anon_sym_in] = ACTIONS(2063),
    [anon_sym_AMP] = ACTIONS(2063),
    [anon_sym_POUND] = ACTIONS(2063),
    [anon_sym_COLON_COLON] = ACTIONS(2065),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2065),
    [anon_sym_else] = ACTIONS(2063),
    [anon_sym_new] = ACTIONS(2063),
    [anon_sym_LBRACE_LT] = ACTIONS(2065),
    [anon_sym_begin] = ACTIONS(2063),
    [sym_ocamlyacc_value] = ACTIONS(2065),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2063),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2065),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2063),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2063),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2065),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2065),
    [aux_sym_number_token1] = ACTIONS(2065),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2065),
    [sym_prefix_operator] = ACTIONS(2065),
    [aux_sym_sign_operator_token1] = ACTIONS(2063),
    [sym_hash_operator] = ACTIONS(2065),
    [anon_sym_lsl] = ACTIONS(2063),
    [anon_sym_lsr] = ACTIONS(2063),
    [anon_sym_asr] = ACTIONS(2063),
    [aux_sym_pow_operator_token1] = ACTIONS(2065),
    [anon_sym_mod] = ACTIONS(2063),
    [anon_sym_land] = ACTIONS(2063),
    [anon_sym_lor] = ACTIONS(2063),
    [anon_sym_lxor] = ACTIONS(2063),
    [aux_sym_mult_operator_token1] = ACTIONS(2063),
    [aux_sym_add_operator_token1] = ACTIONS(2063),
    [sym_concat_operator] = ACTIONS(2065),
    [sym_rel_operator] = ACTIONS(2063),
    [anon_sym_AMP_AMP] = ACTIONS(2063),
    [anon_sym_or] = ACTIONS(2063),
    [anon_sym_PIPE_PIPE] = ACTIONS(2063),
    [sym_assign_operator] = ACTIONS(2065),
    [sym_let_operator] = ACTIONS(2065),
    [sym_let_and_operator] = ACTIONS(2065),
    [sym__uppercase_identifier] = ACTIONS(2065),
    [aux_sym_directive_token1] = ACTIONS(2063),
    [aux_sym_tag_token1] = ACTIONS(2065),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1263)] = {
    [sym_attribute] = STATE(1263),
    [ts_builtin_sym_end] = ACTIONS(2069),
    [sym__lowercase_identifier] = ACTIONS(2067),
    [anon_sym_SEMI_SEMI] = ACTIONS(2069),
    [anon_sym_let] = ACTIONS(2067),
    [anon_sym_and] = ACTIONS(2067),
    [anon_sym_TILDE] = ACTIONS(2067),
    [anon_sym_QMARK] = ACTIONS(2067),
    [anon_sym_LPAREN] = ACTIONS(2069),
    [anon_sym_external] = ACTIONS(2067),
    [anon_sym_type] = ACTIONS(2067),
    [anon_sym_COMMA] = ACTIONS(2069),
    [anon_sym_PIPE] = ACTIONS(2067),
    [anon_sym_DOT] = ACTIONS(2069),
    [anon_sym_LBRACK] = ACTIONS(2067),
    [anon_sym_true] = ACTIONS(2067),
    [anon_sym_false] = ACTIONS(2067),
    [anon_sym_LBRACE] = ACTIONS(2067),
    [anon_sym_SEMI] = ACTIONS(2067),
    [anon_sym_exception] = ACTIONS(2067),
    [anon_sym_module] = ACTIONS(2067),
    [anon_sym_open] = ACTIONS(2067),
    [anon_sym_include] = ACTIONS(2067),
    [anon_sym_class] = ACTIONS(2067),
    [anon_sym_val] = ACTIONS(2067),
    [anon_sym_object] = ACTIONS(2067),
    [anon_sym_in] = ACTIONS(2067),
    [anon_sym_AMP] = ACTIONS(2067),
    [anon_sym_POUND] = ACTIONS(2067),
    [anon_sym_COLON_COLON] = ACTIONS(2069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2069),
    [anon_sym_else] = ACTIONS(2067),
    [anon_sym_new] = ACTIONS(2067),
    [anon_sym_LBRACE_LT] = ACTIONS(2069),
    [anon_sym_begin] = ACTIONS(2067),
    [sym_ocamlyacc_value] = ACTIONS(2069),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2067),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2069),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2067),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2067),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2069),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2069),
    [aux_sym_number_token1] = ACTIONS(2069),
    [aux_sym_signed_number_token1] = ACTIONS(2067),
    [anon_sym_SQUOTE] = ACTIONS(2069),
    [anon_sym_DQUOTE] = ACTIONS(2069),
    [sym_prefix_operator] = ACTIONS(2069),
    [aux_sym_sign_operator_token1] = ACTIONS(2067),
    [sym_hash_operator] = ACTIONS(2069),
    [anon_sym_lsl] = ACTIONS(2067),
    [anon_sym_lsr] = ACTIONS(2067),
    [anon_sym_asr] = ACTIONS(2067),
    [aux_sym_pow_operator_token1] = ACTIONS(2069),
    [anon_sym_mod] = ACTIONS(2067),
    [anon_sym_land] = ACTIONS(2067),
    [anon_sym_lor] = ACTIONS(2067),
    [anon_sym_lxor] = ACTIONS(2067),
    [aux_sym_mult_operator_token1] = ACTIONS(2067),
    [aux_sym_add_operator_token1] = ACTIONS(2067),
    [sym_concat_operator] = ACTIONS(2069),
    [sym_rel_operator] = ACTIONS(2067),
    [anon_sym_AMP_AMP] = ACTIONS(2067),
    [anon_sym_or] = ACTIONS(2067),
    [anon_sym_PIPE_PIPE] = ACTIONS(2067),
    [sym_assign_operator] = ACTIONS(2069),
    [sym_let_operator] = ACTIONS(2069),
    [sym_let_and_operator] = ACTIONS(2069),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_directive_token1] = ACTIONS(2067),
    [aux_sym_tag_token1] = ACTIONS(2069),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1264)] = {
    [sym_attribute] = STATE(1264),
    [sym__lowercase_identifier] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(2415),
    [anon_sym_QMARK] = ACTIONS(2415),
    [anon_sym_LPAREN] = ACTIONS(2417),
    [anon_sym_RPAREN] = ACTIONS(2417),
    [anon_sym_COMMA] = ACTIONS(2417),
    [anon_sym_PIPE] = ACTIONS(2415),
    [anon_sym_COLON2] = ACTIONS(2415),
    [anon_sym_DASH_GT] = ACTIONS(2415),
    [anon_sym_LBRACK] = ACTIONS(2415),
    [anon_sym_RBRACK] = ACTIONS(2417),
    [anon_sym_true] = ACTIONS(2415),
    [anon_sym_false] = ACTIONS(2415),
    [anon_sym_LBRACE] = ACTIONS(2415),
    [anon_sym_SEMI] = ACTIONS(2417),
    [anon_sym_RBRACE] = ACTIONS(2417),
    [anon_sym_constraint] = ACTIONS(2415),
    [anon_sym_val] = ACTIONS(2415),
    [anon_sym_end] = ACTIONS(2415),
    [anon_sym_with] = ACTIONS(2415),
    [anon_sym_COLON_GT] = ACTIONS(2417),
    [anon_sym_object] = ACTIONS(2415),
    [anon_sym_inherit] = ACTIONS(2415),
    [anon_sym_method] = ACTIONS(2415),
    [anon_sym_initializer] = ACTIONS(2415),
    [anon_sym_AMP] = ACTIONS(2415),
    [anon_sym_COLON_COLON] = ACTIONS(2417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2417),
    [anon_sym_then] = ACTIONS(2415),
    [anon_sym_else] = ACTIONS(2415),
    [anon_sym_do] = ACTIONS(2415),
    [anon_sym_done] = ACTIONS(2415),
    [anon_sym_to] = ACTIONS(2415),
    [anon_sym_downto] = ACTIONS(2415),
    [anon_sym_new] = ACTIONS(2415),
    [anon_sym_LBRACE_LT] = ACTIONS(2417),
    [anon_sym_GT_RBRACE] = ACTIONS(2417),
    [anon_sym_begin] = ACTIONS(2415),
    [sym_ocamlyacc_value] = ACTIONS(2417),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2415),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2417),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2415),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2417),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2417),
    [aux_sym_number_token1] = ACTIONS(2417),
    [aux_sym_signed_number_token1] = ACTIONS(2415),
    [anon_sym_SQUOTE] = ACTIONS(2417),
    [anon_sym_DQUOTE] = ACTIONS(2417),
    [sym_prefix_operator] = ACTIONS(2417),
    [aux_sym_sign_operator_token1] = ACTIONS(2415),
    [anon_sym_lsl] = ACTIONS(2415),
    [anon_sym_lsr] = ACTIONS(2415),
    [anon_sym_asr] = ACTIONS(2415),
    [aux_sym_pow_operator_token1] = ACTIONS(2417),
    [anon_sym_mod] = ACTIONS(2415),
    [anon_sym_land] = ACTIONS(2415),
    [anon_sym_lor] = ACTIONS(2415),
    [anon_sym_lxor] = ACTIONS(2415),
    [aux_sym_mult_operator_token1] = ACTIONS(2415),
    [aux_sym_add_operator_token1] = ACTIONS(2415),
    [sym_concat_operator] = ACTIONS(2417),
    [sym_rel_operator] = ACTIONS(2415),
    [anon_sym_AMP_AMP] = ACTIONS(2415),
    [anon_sym_or] = ACTIONS(2415),
    [anon_sym_PIPE_PIPE] = ACTIONS(2415),
    [sym_assign_operator] = ACTIONS(2417),
    [sym__uppercase_identifier] = ACTIONS(2417),
    [aux_sym_tag_token1] = ACTIONS(2417),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1265)] = {
    [sym_attribute] = STATE(1265),
    [ts_builtin_sym_end] = ACTIONS(2105),
    [sym__lowercase_identifier] = ACTIONS(2103),
    [anon_sym_SEMI_SEMI] = ACTIONS(2105),
    [anon_sym_let] = ACTIONS(2103),
    [anon_sym_and] = ACTIONS(2103),
    [anon_sym_TILDE] = ACTIONS(2103),
    [anon_sym_QMARK] = ACTIONS(2103),
    [anon_sym_LPAREN] = ACTIONS(2105),
    [anon_sym_external] = ACTIONS(2103),
    [anon_sym_type] = ACTIONS(2103),
    [anon_sym_COMMA] = ACTIONS(2105),
    [anon_sym_PIPE] = ACTIONS(2103),
    [anon_sym_DOT] = ACTIONS(2105),
    [anon_sym_LBRACK] = ACTIONS(2103),
    [anon_sym_true] = ACTIONS(2103),
    [anon_sym_false] = ACTIONS(2103),
    [anon_sym_LBRACE] = ACTIONS(2103),
    [anon_sym_SEMI] = ACTIONS(2103),
    [anon_sym_exception] = ACTIONS(2103),
    [anon_sym_module] = ACTIONS(2103),
    [anon_sym_open] = ACTIONS(2103),
    [anon_sym_include] = ACTIONS(2103),
    [anon_sym_class] = ACTIONS(2103),
    [anon_sym_val] = ACTIONS(2103),
    [anon_sym_object] = ACTIONS(2103),
    [anon_sym_in] = ACTIONS(2103),
    [anon_sym_AMP] = ACTIONS(2103),
    [anon_sym_POUND] = ACTIONS(2103),
    [anon_sym_COLON_COLON] = ACTIONS(2105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2105),
    [anon_sym_else] = ACTIONS(2103),
    [anon_sym_new] = ACTIONS(2103),
    [anon_sym_LBRACE_LT] = ACTIONS(2105),
    [anon_sym_begin] = ACTIONS(2103),
    [sym_ocamlyacc_value] = ACTIONS(2105),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2103),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2105),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2103),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2103),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2105),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2105),
    [aux_sym_number_token1] = ACTIONS(2105),
    [aux_sym_signed_number_token1] = ACTIONS(2103),
    [anon_sym_SQUOTE] = ACTIONS(2105),
    [anon_sym_DQUOTE] = ACTIONS(2105),
    [sym_prefix_operator] = ACTIONS(2105),
    [aux_sym_sign_operator_token1] = ACTIONS(2103),
    [sym_hash_operator] = ACTIONS(2105),
    [anon_sym_lsl] = ACTIONS(2103),
    [anon_sym_lsr] = ACTIONS(2103),
    [anon_sym_asr] = ACTIONS(2103),
    [aux_sym_pow_operator_token1] = ACTIONS(2105),
    [anon_sym_mod] = ACTIONS(2103),
    [anon_sym_land] = ACTIONS(2103),
    [anon_sym_lor] = ACTIONS(2103),
    [anon_sym_lxor] = ACTIONS(2103),
    [aux_sym_mult_operator_token1] = ACTIONS(2103),
    [aux_sym_add_operator_token1] = ACTIONS(2103),
    [sym_concat_operator] = ACTIONS(2105),
    [sym_rel_operator] = ACTIONS(2103),
    [anon_sym_AMP_AMP] = ACTIONS(2103),
    [anon_sym_or] = ACTIONS(2103),
    [anon_sym_PIPE_PIPE] = ACTIONS(2103),
    [sym_assign_operator] = ACTIONS(2105),
    [sym_let_operator] = ACTIONS(2105),
    [sym_let_and_operator] = ACTIONS(2105),
    [sym__uppercase_identifier] = ACTIONS(2105),
    [aux_sym_directive_token1] = ACTIONS(2103),
    [aux_sym_tag_token1] = ACTIONS(2105),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1266)] = {
    [sym_attribute] = STATE(1266),
    [ts_builtin_sym_end] = ACTIONS(2109),
    [sym__lowercase_identifier] = ACTIONS(2107),
    [anon_sym_SEMI_SEMI] = ACTIONS(2109),
    [anon_sym_let] = ACTIONS(2107),
    [anon_sym_and] = ACTIONS(2107),
    [anon_sym_TILDE] = ACTIONS(2107),
    [anon_sym_QMARK] = ACTIONS(2107),
    [anon_sym_LPAREN] = ACTIONS(2109),
    [anon_sym_external] = ACTIONS(2107),
    [anon_sym_type] = ACTIONS(2107),
    [anon_sym_COMMA] = ACTIONS(2109),
    [anon_sym_PIPE] = ACTIONS(2107),
    [anon_sym_DOT] = ACTIONS(2109),
    [anon_sym_LBRACK] = ACTIONS(2107),
    [anon_sym_true] = ACTIONS(2107),
    [anon_sym_false] = ACTIONS(2107),
    [anon_sym_LBRACE] = ACTIONS(2107),
    [anon_sym_SEMI] = ACTIONS(2107),
    [anon_sym_exception] = ACTIONS(2107),
    [anon_sym_module] = ACTIONS(2107),
    [anon_sym_open] = ACTIONS(2107),
    [anon_sym_include] = ACTIONS(2107),
    [anon_sym_class] = ACTIONS(2107),
    [anon_sym_val] = ACTIONS(2107),
    [anon_sym_object] = ACTIONS(2107),
    [anon_sym_in] = ACTIONS(2107),
    [anon_sym_AMP] = ACTIONS(2107),
    [anon_sym_POUND] = ACTIONS(2107),
    [anon_sym_COLON_COLON] = ACTIONS(2109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2109),
    [anon_sym_else] = ACTIONS(2107),
    [anon_sym_new] = ACTIONS(2107),
    [anon_sym_LBRACE_LT] = ACTIONS(2109),
    [anon_sym_begin] = ACTIONS(2107),
    [sym_ocamlyacc_value] = ACTIONS(2109),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2107),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2109),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2109),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2109),
    [aux_sym_number_token1] = ACTIONS(2109),
    [aux_sym_signed_number_token1] = ACTIONS(2107),
    [anon_sym_SQUOTE] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(2109),
    [sym_prefix_operator] = ACTIONS(2109),
    [aux_sym_sign_operator_token1] = ACTIONS(2107),
    [sym_hash_operator] = ACTIONS(2109),
    [anon_sym_lsl] = ACTIONS(2107),
    [anon_sym_lsr] = ACTIONS(2107),
    [anon_sym_asr] = ACTIONS(2107),
    [aux_sym_pow_operator_token1] = ACTIONS(2109),
    [anon_sym_mod] = ACTIONS(2107),
    [anon_sym_land] = ACTIONS(2107),
    [anon_sym_lor] = ACTIONS(2107),
    [anon_sym_lxor] = ACTIONS(2107),
    [aux_sym_mult_operator_token1] = ACTIONS(2107),
    [aux_sym_add_operator_token1] = ACTIONS(2107),
    [sym_concat_operator] = ACTIONS(2109),
    [sym_rel_operator] = ACTIONS(2107),
    [anon_sym_AMP_AMP] = ACTIONS(2107),
    [anon_sym_or] = ACTIONS(2107),
    [anon_sym_PIPE_PIPE] = ACTIONS(2107),
    [sym_assign_operator] = ACTIONS(2109),
    [sym_let_operator] = ACTIONS(2109),
    [sym_let_and_operator] = ACTIONS(2109),
    [sym__uppercase_identifier] = ACTIONS(2109),
    [aux_sym_directive_token1] = ACTIONS(2107),
    [aux_sym_tag_token1] = ACTIONS(2109),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1267)] = {
    [sym_attribute] = STATE(1267),
    [ts_builtin_sym_end] = ACTIONS(2119),
    [sym__lowercase_identifier] = ACTIONS(2117),
    [anon_sym_SEMI_SEMI] = ACTIONS(2119),
    [anon_sym_let] = ACTIONS(2117),
    [anon_sym_and] = ACTIONS(2117),
    [anon_sym_TILDE] = ACTIONS(2117),
    [anon_sym_QMARK] = ACTIONS(2117),
    [anon_sym_LPAREN] = ACTIONS(2119),
    [anon_sym_external] = ACTIONS(2117),
    [anon_sym_type] = ACTIONS(2117),
    [anon_sym_COMMA] = ACTIONS(2119),
    [anon_sym_PIPE] = ACTIONS(2117),
    [anon_sym_DOT] = ACTIONS(2119),
    [anon_sym_LBRACK] = ACTIONS(2117),
    [anon_sym_true] = ACTIONS(2117),
    [anon_sym_false] = ACTIONS(2117),
    [anon_sym_LBRACE] = ACTIONS(2117),
    [anon_sym_SEMI] = ACTIONS(2117),
    [anon_sym_exception] = ACTIONS(2117),
    [anon_sym_module] = ACTIONS(2117),
    [anon_sym_open] = ACTIONS(2117),
    [anon_sym_include] = ACTIONS(2117),
    [anon_sym_class] = ACTIONS(2117),
    [anon_sym_val] = ACTIONS(2117),
    [anon_sym_object] = ACTIONS(2117),
    [anon_sym_in] = ACTIONS(2117),
    [anon_sym_AMP] = ACTIONS(2117),
    [anon_sym_POUND] = ACTIONS(2117),
    [anon_sym_COLON_COLON] = ACTIONS(2119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2119),
    [anon_sym_else] = ACTIONS(2117),
    [anon_sym_new] = ACTIONS(2117),
    [anon_sym_LBRACE_LT] = ACTIONS(2119),
    [anon_sym_begin] = ACTIONS(2117),
    [sym_ocamlyacc_value] = ACTIONS(2119),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2117),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2119),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2117),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2117),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2119),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2119),
    [aux_sym_number_token1] = ACTIONS(2119),
    [aux_sym_signed_number_token1] = ACTIONS(2117),
    [anon_sym_SQUOTE] = ACTIONS(2119),
    [anon_sym_DQUOTE] = ACTIONS(2119),
    [sym_prefix_operator] = ACTIONS(2119),
    [aux_sym_sign_operator_token1] = ACTIONS(2117),
    [sym_hash_operator] = ACTIONS(2119),
    [anon_sym_lsl] = ACTIONS(2117),
    [anon_sym_lsr] = ACTIONS(2117),
    [anon_sym_asr] = ACTIONS(2117),
    [aux_sym_pow_operator_token1] = ACTIONS(2119),
    [anon_sym_mod] = ACTIONS(2117),
    [anon_sym_land] = ACTIONS(2117),
    [anon_sym_lor] = ACTIONS(2117),
    [anon_sym_lxor] = ACTIONS(2117),
    [aux_sym_mult_operator_token1] = ACTIONS(2117),
    [aux_sym_add_operator_token1] = ACTIONS(2117),
    [sym_concat_operator] = ACTIONS(2119),
    [sym_rel_operator] = ACTIONS(2117),
    [anon_sym_AMP_AMP] = ACTIONS(2117),
    [anon_sym_or] = ACTIONS(2117),
    [anon_sym_PIPE_PIPE] = ACTIONS(2117),
    [sym_assign_operator] = ACTIONS(2119),
    [sym_let_operator] = ACTIONS(2119),
    [sym_let_and_operator] = ACTIONS(2119),
    [sym__uppercase_identifier] = ACTIONS(2119),
    [aux_sym_directive_token1] = ACTIONS(2117),
    [aux_sym_tag_token1] = ACTIONS(2119),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1268)] = {
    [sym_attribute] = STATE(1268),
    [ts_builtin_sym_end] = ACTIONS(2085),
    [sym__lowercase_identifier] = ACTIONS(2083),
    [anon_sym_SEMI_SEMI] = ACTIONS(2085),
    [anon_sym_let] = ACTIONS(2083),
    [anon_sym_and] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(2083),
    [anon_sym_QMARK] = ACTIONS(2083),
    [anon_sym_LPAREN] = ACTIONS(2085),
    [anon_sym_external] = ACTIONS(2083),
    [anon_sym_type] = ACTIONS(2083),
    [anon_sym_COMMA] = ACTIONS(2085),
    [anon_sym_PIPE] = ACTIONS(2083),
    [anon_sym_DOT] = ACTIONS(2085),
    [anon_sym_LBRACK] = ACTIONS(2083),
    [anon_sym_true] = ACTIONS(2083),
    [anon_sym_false] = ACTIONS(2083),
    [anon_sym_LBRACE] = ACTIONS(2083),
    [anon_sym_SEMI] = ACTIONS(2083),
    [anon_sym_exception] = ACTIONS(2083),
    [anon_sym_module] = ACTIONS(2083),
    [anon_sym_open] = ACTIONS(2083),
    [anon_sym_include] = ACTIONS(2083),
    [anon_sym_class] = ACTIONS(2083),
    [anon_sym_val] = ACTIONS(2083),
    [anon_sym_object] = ACTIONS(2083),
    [anon_sym_in] = ACTIONS(2083),
    [anon_sym_AMP] = ACTIONS(2083),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(2085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2085),
    [anon_sym_else] = ACTIONS(2083),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_LBRACE_LT] = ACTIONS(2085),
    [anon_sym_begin] = ACTIONS(2083),
    [sym_ocamlyacc_value] = ACTIONS(2085),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2083),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2085),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2083),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2083),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2085),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2085),
    [aux_sym_number_token1] = ACTIONS(2085),
    [aux_sym_signed_number_token1] = ACTIONS(2083),
    [anon_sym_SQUOTE] = ACTIONS(2085),
    [anon_sym_DQUOTE] = ACTIONS(2085),
    [sym_prefix_operator] = ACTIONS(2085),
    [aux_sym_sign_operator_token1] = ACTIONS(2083),
    [sym_hash_operator] = ACTIONS(2085),
    [anon_sym_lsl] = ACTIONS(2083),
    [anon_sym_lsr] = ACTIONS(2083),
    [anon_sym_asr] = ACTIONS(2083),
    [aux_sym_pow_operator_token1] = ACTIONS(2085),
    [anon_sym_mod] = ACTIONS(2083),
    [anon_sym_land] = ACTIONS(2083),
    [anon_sym_lor] = ACTIONS(2083),
    [anon_sym_lxor] = ACTIONS(2083),
    [aux_sym_mult_operator_token1] = ACTIONS(2083),
    [aux_sym_add_operator_token1] = ACTIONS(2083),
    [sym_concat_operator] = ACTIONS(2085),
    [sym_rel_operator] = ACTIONS(2083),
    [anon_sym_AMP_AMP] = ACTIONS(2083),
    [anon_sym_or] = ACTIONS(2083),
    [anon_sym_PIPE_PIPE] = ACTIONS(2083),
    [sym_assign_operator] = ACTIONS(2085),
    [sym_let_operator] = ACTIONS(2085),
    [sym_let_and_operator] = ACTIONS(2085),
    [sym__uppercase_identifier] = ACTIONS(2085),
    [aux_sym_directive_token1] = ACTIONS(2083),
    [aux_sym_tag_token1] = ACTIONS(2085),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1269)] = {
    [sym_attribute] = STATE(1269),
    [ts_builtin_sym_end] = ACTIONS(2089),
    [sym__lowercase_identifier] = ACTIONS(2087),
    [anon_sym_SEMI_SEMI] = ACTIONS(2089),
    [anon_sym_let] = ACTIONS(2087),
    [anon_sym_and] = ACTIONS(2087),
    [anon_sym_TILDE] = ACTIONS(2087),
    [anon_sym_QMARK] = ACTIONS(2087),
    [anon_sym_LPAREN] = ACTIONS(2089),
    [anon_sym_external] = ACTIONS(2087),
    [anon_sym_type] = ACTIONS(2087),
    [anon_sym_COMMA] = ACTIONS(2089),
    [anon_sym_PIPE] = ACTIONS(2087),
    [anon_sym_DOT] = ACTIONS(2089),
    [anon_sym_LBRACK] = ACTIONS(2087),
    [anon_sym_true] = ACTIONS(2087),
    [anon_sym_false] = ACTIONS(2087),
    [anon_sym_LBRACE] = ACTIONS(2087),
    [anon_sym_SEMI] = ACTIONS(2087),
    [anon_sym_exception] = ACTIONS(2087),
    [anon_sym_module] = ACTIONS(2087),
    [anon_sym_open] = ACTIONS(2087),
    [anon_sym_include] = ACTIONS(2087),
    [anon_sym_class] = ACTIONS(2087),
    [anon_sym_val] = ACTIONS(2087),
    [anon_sym_object] = ACTIONS(2087),
    [anon_sym_in] = ACTIONS(2087),
    [anon_sym_AMP] = ACTIONS(2087),
    [anon_sym_POUND] = ACTIONS(2087),
    [anon_sym_COLON_COLON] = ACTIONS(2089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2089),
    [anon_sym_else] = ACTIONS(2087),
    [anon_sym_new] = ACTIONS(2087),
    [anon_sym_LBRACE_LT] = ACTIONS(2089),
    [anon_sym_begin] = ACTIONS(2087),
    [sym_ocamlyacc_value] = ACTIONS(2089),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2087),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2089),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2087),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2087),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2089),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2089),
    [aux_sym_number_token1] = ACTIONS(2089),
    [aux_sym_signed_number_token1] = ACTIONS(2087),
    [anon_sym_SQUOTE] = ACTIONS(2089),
    [anon_sym_DQUOTE] = ACTIONS(2089),
    [sym_prefix_operator] = ACTIONS(2089),
    [aux_sym_sign_operator_token1] = ACTIONS(2087),
    [sym_hash_operator] = ACTIONS(2089),
    [anon_sym_lsl] = ACTIONS(2087),
    [anon_sym_lsr] = ACTIONS(2087),
    [anon_sym_asr] = ACTIONS(2087),
    [aux_sym_pow_operator_token1] = ACTIONS(2089),
    [anon_sym_mod] = ACTIONS(2087),
    [anon_sym_land] = ACTIONS(2087),
    [anon_sym_lor] = ACTIONS(2087),
    [anon_sym_lxor] = ACTIONS(2087),
    [aux_sym_mult_operator_token1] = ACTIONS(2087),
    [aux_sym_add_operator_token1] = ACTIONS(2087),
    [sym_concat_operator] = ACTIONS(2089),
    [sym_rel_operator] = ACTIONS(2087),
    [anon_sym_AMP_AMP] = ACTIONS(2087),
    [anon_sym_or] = ACTIONS(2087),
    [anon_sym_PIPE_PIPE] = ACTIONS(2087),
    [sym_assign_operator] = ACTIONS(2089),
    [sym_let_operator] = ACTIONS(2089),
    [sym_let_and_operator] = ACTIONS(2089),
    [sym__uppercase_identifier] = ACTIONS(2089),
    [aux_sym_directive_token1] = ACTIONS(2087),
    [aux_sym_tag_token1] = ACTIONS(2089),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1270)] = {
    [sym_attribute] = STATE(1270),
    [ts_builtin_sym_end] = ACTIONS(2093),
    [sym__lowercase_identifier] = ACTIONS(2091),
    [anon_sym_SEMI_SEMI] = ACTIONS(2093),
    [anon_sym_let] = ACTIONS(2091),
    [anon_sym_and] = ACTIONS(2091),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_external] = ACTIONS(2091),
    [anon_sym_type] = ACTIONS(2091),
    [anon_sym_COMMA] = ACTIONS(2093),
    [anon_sym_PIPE] = ACTIONS(2091),
    [anon_sym_DOT] = ACTIONS(2093),
    [anon_sym_LBRACK] = ACTIONS(2091),
    [anon_sym_true] = ACTIONS(2091),
    [anon_sym_false] = ACTIONS(2091),
    [anon_sym_LBRACE] = ACTIONS(2091),
    [anon_sym_SEMI] = ACTIONS(2091),
    [anon_sym_exception] = ACTIONS(2091),
    [anon_sym_module] = ACTIONS(2091),
    [anon_sym_open] = ACTIONS(2091),
    [anon_sym_include] = ACTIONS(2091),
    [anon_sym_class] = ACTIONS(2091),
    [anon_sym_val] = ACTIONS(2091),
    [anon_sym_object] = ACTIONS(2091),
    [anon_sym_in] = ACTIONS(2091),
    [anon_sym_AMP] = ACTIONS(2091),
    [anon_sym_POUND] = ACTIONS(2091),
    [anon_sym_COLON_COLON] = ACTIONS(2093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2093),
    [anon_sym_else] = ACTIONS(2091),
    [anon_sym_new] = ACTIONS(2091),
    [anon_sym_LBRACE_LT] = ACTIONS(2093),
    [anon_sym_begin] = ACTIONS(2091),
    [sym_ocamlyacc_value] = ACTIONS(2093),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2091),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2093),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2091),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2091),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2093),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2093),
    [aux_sym_number_token1] = ACTIONS(2093),
    [aux_sym_signed_number_token1] = ACTIONS(2091),
    [anon_sym_SQUOTE] = ACTIONS(2093),
    [anon_sym_DQUOTE] = ACTIONS(2093),
    [sym_prefix_operator] = ACTIONS(2093),
    [aux_sym_sign_operator_token1] = ACTIONS(2091),
    [sym_hash_operator] = ACTIONS(2093),
    [anon_sym_lsl] = ACTIONS(2091),
    [anon_sym_lsr] = ACTIONS(2091),
    [anon_sym_asr] = ACTIONS(2091),
    [aux_sym_pow_operator_token1] = ACTIONS(2093),
    [anon_sym_mod] = ACTIONS(2091),
    [anon_sym_land] = ACTIONS(2091),
    [anon_sym_lor] = ACTIONS(2091),
    [anon_sym_lxor] = ACTIONS(2091),
    [aux_sym_mult_operator_token1] = ACTIONS(2091),
    [aux_sym_add_operator_token1] = ACTIONS(2091),
    [sym_concat_operator] = ACTIONS(2093),
    [sym_rel_operator] = ACTIONS(2091),
    [anon_sym_AMP_AMP] = ACTIONS(2091),
    [anon_sym_or] = ACTIONS(2091),
    [anon_sym_PIPE_PIPE] = ACTIONS(2091),
    [sym_assign_operator] = ACTIONS(2093),
    [sym_let_operator] = ACTIONS(2093),
    [sym_let_and_operator] = ACTIONS(2093),
    [sym__uppercase_identifier] = ACTIONS(2093),
    [aux_sym_directive_token1] = ACTIONS(2091),
    [aux_sym_tag_token1] = ACTIONS(2093),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1271)] = {
    [sym_attribute] = STATE(1271),
    [ts_builtin_sym_end] = ACTIONS(2097),
    [sym__lowercase_identifier] = ACTIONS(2095),
    [anon_sym_SEMI_SEMI] = ACTIONS(2097),
    [anon_sym_let] = ACTIONS(2095),
    [anon_sym_and] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(2095),
    [anon_sym_QMARK] = ACTIONS(2095),
    [anon_sym_LPAREN] = ACTIONS(2097),
    [anon_sym_external] = ACTIONS(2095),
    [anon_sym_type] = ACTIONS(2095),
    [anon_sym_COMMA] = ACTIONS(2097),
    [anon_sym_PIPE] = ACTIONS(2095),
    [anon_sym_DOT] = ACTIONS(2097),
    [anon_sym_LBRACK] = ACTIONS(2095),
    [anon_sym_true] = ACTIONS(2095),
    [anon_sym_false] = ACTIONS(2095),
    [anon_sym_LBRACE] = ACTIONS(2095),
    [anon_sym_SEMI] = ACTIONS(2095),
    [anon_sym_exception] = ACTIONS(2095),
    [anon_sym_module] = ACTIONS(2095),
    [anon_sym_open] = ACTIONS(2095),
    [anon_sym_include] = ACTIONS(2095),
    [anon_sym_class] = ACTIONS(2095),
    [anon_sym_val] = ACTIONS(2095),
    [anon_sym_object] = ACTIONS(2095),
    [anon_sym_in] = ACTIONS(2095),
    [anon_sym_AMP] = ACTIONS(2095),
    [anon_sym_POUND] = ACTIONS(2095),
    [anon_sym_COLON_COLON] = ACTIONS(2097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2097),
    [anon_sym_else] = ACTIONS(2095),
    [anon_sym_new] = ACTIONS(2095),
    [anon_sym_LBRACE_LT] = ACTIONS(2097),
    [anon_sym_begin] = ACTIONS(2095),
    [sym_ocamlyacc_value] = ACTIONS(2097),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2095),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2097),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2095),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2095),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2097),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2097),
    [aux_sym_number_token1] = ACTIONS(2097),
    [aux_sym_signed_number_token1] = ACTIONS(2095),
    [anon_sym_SQUOTE] = ACTIONS(2097),
    [anon_sym_DQUOTE] = ACTIONS(2097),
    [sym_prefix_operator] = ACTIONS(2097),
    [aux_sym_sign_operator_token1] = ACTIONS(2095),
    [sym_hash_operator] = ACTIONS(2097),
    [anon_sym_lsl] = ACTIONS(2095),
    [anon_sym_lsr] = ACTIONS(2095),
    [anon_sym_asr] = ACTIONS(2095),
    [aux_sym_pow_operator_token1] = ACTIONS(2097),
    [anon_sym_mod] = ACTIONS(2095),
    [anon_sym_land] = ACTIONS(2095),
    [anon_sym_lor] = ACTIONS(2095),
    [anon_sym_lxor] = ACTIONS(2095),
    [aux_sym_mult_operator_token1] = ACTIONS(2095),
    [aux_sym_add_operator_token1] = ACTIONS(2095),
    [sym_concat_operator] = ACTIONS(2097),
    [sym_rel_operator] = ACTIONS(2095),
    [anon_sym_AMP_AMP] = ACTIONS(2095),
    [anon_sym_or] = ACTIONS(2095),
    [anon_sym_PIPE_PIPE] = ACTIONS(2095),
    [sym_assign_operator] = ACTIONS(2097),
    [sym_let_operator] = ACTIONS(2097),
    [sym_let_and_operator] = ACTIONS(2097),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_directive_token1] = ACTIONS(2095),
    [aux_sym_tag_token1] = ACTIONS(2097),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1272)] = {
    [sym_attribute] = STATE(1272),
    [ts_builtin_sym_end] = ACTIONS(2135),
    [sym__lowercase_identifier] = ACTIONS(2133),
    [anon_sym_SEMI_SEMI] = ACTIONS(2135),
    [anon_sym_let] = ACTIONS(2133),
    [anon_sym_and] = ACTIONS(2133),
    [anon_sym_TILDE] = ACTIONS(2133),
    [anon_sym_QMARK] = ACTIONS(2133),
    [anon_sym_LPAREN] = ACTIONS(2135),
    [anon_sym_external] = ACTIONS(2133),
    [anon_sym_type] = ACTIONS(2133),
    [anon_sym_COMMA] = ACTIONS(2135),
    [anon_sym_PIPE] = ACTIONS(2133),
    [anon_sym_DOT] = ACTIONS(2135),
    [anon_sym_LBRACK] = ACTIONS(2133),
    [anon_sym_true] = ACTIONS(2133),
    [anon_sym_false] = ACTIONS(2133),
    [anon_sym_LBRACE] = ACTIONS(2133),
    [anon_sym_SEMI] = ACTIONS(2133),
    [anon_sym_exception] = ACTIONS(2133),
    [anon_sym_module] = ACTIONS(2133),
    [anon_sym_open] = ACTIONS(2133),
    [anon_sym_include] = ACTIONS(2133),
    [anon_sym_class] = ACTIONS(2133),
    [anon_sym_val] = ACTIONS(2133),
    [anon_sym_object] = ACTIONS(2133),
    [anon_sym_in] = ACTIONS(2133),
    [anon_sym_AMP] = ACTIONS(2133),
    [anon_sym_POUND] = ACTIONS(2133),
    [anon_sym_COLON_COLON] = ACTIONS(2135),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2135),
    [anon_sym_else] = ACTIONS(2133),
    [anon_sym_new] = ACTIONS(2133),
    [anon_sym_LBRACE_LT] = ACTIONS(2135),
    [anon_sym_begin] = ACTIONS(2133),
    [sym_ocamlyacc_value] = ACTIONS(2135),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2133),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2135),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2133),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2133),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2135),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2135),
    [aux_sym_number_token1] = ACTIONS(2135),
    [aux_sym_signed_number_token1] = ACTIONS(2133),
    [anon_sym_SQUOTE] = ACTIONS(2135),
    [anon_sym_DQUOTE] = ACTIONS(2135),
    [sym_prefix_operator] = ACTIONS(2135),
    [aux_sym_sign_operator_token1] = ACTIONS(2133),
    [sym_hash_operator] = ACTIONS(2135),
    [anon_sym_lsl] = ACTIONS(2133),
    [anon_sym_lsr] = ACTIONS(2133),
    [anon_sym_asr] = ACTIONS(2133),
    [aux_sym_pow_operator_token1] = ACTIONS(2135),
    [anon_sym_mod] = ACTIONS(2133),
    [anon_sym_land] = ACTIONS(2133),
    [anon_sym_lor] = ACTIONS(2133),
    [anon_sym_lxor] = ACTIONS(2133),
    [aux_sym_mult_operator_token1] = ACTIONS(2133),
    [aux_sym_add_operator_token1] = ACTIONS(2133),
    [sym_concat_operator] = ACTIONS(2135),
    [sym_rel_operator] = ACTIONS(2133),
    [anon_sym_AMP_AMP] = ACTIONS(2133),
    [anon_sym_or] = ACTIONS(2133),
    [anon_sym_PIPE_PIPE] = ACTIONS(2133),
    [sym_assign_operator] = ACTIONS(2135),
    [sym_let_operator] = ACTIONS(2135),
    [sym_let_and_operator] = ACTIONS(2135),
    [sym__uppercase_identifier] = ACTIONS(2135),
    [aux_sym_directive_token1] = ACTIONS(2133),
    [aux_sym_tag_token1] = ACTIONS(2135),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1273)] = {
    [sym_attribute] = STATE(1273),
    [ts_builtin_sym_end] = ACTIONS(2147),
    [sym__lowercase_identifier] = ACTIONS(2145),
    [anon_sym_SEMI_SEMI] = ACTIONS(2147),
    [anon_sym_let] = ACTIONS(2145),
    [anon_sym_and] = ACTIONS(2145),
    [anon_sym_TILDE] = ACTIONS(2145),
    [anon_sym_QMARK] = ACTIONS(2145),
    [anon_sym_LPAREN] = ACTIONS(2147),
    [anon_sym_external] = ACTIONS(2145),
    [anon_sym_type] = ACTIONS(2145),
    [anon_sym_COMMA] = ACTIONS(2147),
    [anon_sym_PIPE] = ACTIONS(2145),
    [anon_sym_DOT] = ACTIONS(2147),
    [anon_sym_LBRACK] = ACTIONS(2145),
    [anon_sym_true] = ACTIONS(2145),
    [anon_sym_false] = ACTIONS(2145),
    [anon_sym_LBRACE] = ACTIONS(2145),
    [anon_sym_SEMI] = ACTIONS(2145),
    [anon_sym_exception] = ACTIONS(2145),
    [anon_sym_module] = ACTIONS(2145),
    [anon_sym_open] = ACTIONS(2145),
    [anon_sym_include] = ACTIONS(2145),
    [anon_sym_class] = ACTIONS(2145),
    [anon_sym_val] = ACTIONS(2145),
    [anon_sym_object] = ACTIONS(2145),
    [anon_sym_in] = ACTIONS(2145),
    [anon_sym_AMP] = ACTIONS(2145),
    [anon_sym_POUND] = ACTIONS(2145),
    [anon_sym_COLON_COLON] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2147),
    [anon_sym_else] = ACTIONS(2145),
    [anon_sym_new] = ACTIONS(2145),
    [anon_sym_LBRACE_LT] = ACTIONS(2147),
    [anon_sym_begin] = ACTIONS(2145),
    [sym_ocamlyacc_value] = ACTIONS(2147),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2145),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2147),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2145),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2145),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2147),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2147),
    [aux_sym_number_token1] = ACTIONS(2147),
    [aux_sym_signed_number_token1] = ACTIONS(2145),
    [anon_sym_SQUOTE] = ACTIONS(2147),
    [anon_sym_DQUOTE] = ACTIONS(2147),
    [sym_prefix_operator] = ACTIONS(2147),
    [aux_sym_sign_operator_token1] = ACTIONS(2145),
    [sym_hash_operator] = ACTIONS(2147),
    [anon_sym_lsl] = ACTIONS(2145),
    [anon_sym_lsr] = ACTIONS(2145),
    [anon_sym_asr] = ACTIONS(2145),
    [aux_sym_pow_operator_token1] = ACTIONS(2147),
    [anon_sym_mod] = ACTIONS(2145),
    [anon_sym_land] = ACTIONS(2145),
    [anon_sym_lor] = ACTIONS(2145),
    [anon_sym_lxor] = ACTIONS(2145),
    [aux_sym_mult_operator_token1] = ACTIONS(2145),
    [aux_sym_add_operator_token1] = ACTIONS(2145),
    [sym_concat_operator] = ACTIONS(2147),
    [sym_rel_operator] = ACTIONS(2145),
    [anon_sym_AMP_AMP] = ACTIONS(2145),
    [anon_sym_or] = ACTIONS(2145),
    [anon_sym_PIPE_PIPE] = ACTIONS(2145),
    [sym_assign_operator] = ACTIONS(2147),
    [sym_let_operator] = ACTIONS(2147),
    [sym_let_and_operator] = ACTIONS(2147),
    [sym__uppercase_identifier] = ACTIONS(2147),
    [aux_sym_directive_token1] = ACTIONS(2145),
    [aux_sym_tag_token1] = ACTIONS(2147),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1274)] = {
    [sym_attribute] = STATE(1274),
    [ts_builtin_sym_end] = ACTIONS(2155),
    [sym__lowercase_identifier] = ACTIONS(2153),
    [anon_sym_SEMI_SEMI] = ACTIONS(2155),
    [anon_sym_let] = ACTIONS(2153),
    [anon_sym_and] = ACTIONS(2153),
    [anon_sym_TILDE] = ACTIONS(2153),
    [anon_sym_QMARK] = ACTIONS(2153),
    [anon_sym_LPAREN] = ACTIONS(2155),
    [anon_sym_external] = ACTIONS(2153),
    [anon_sym_type] = ACTIONS(2153),
    [anon_sym_COMMA] = ACTIONS(2155),
    [anon_sym_PIPE] = ACTIONS(2153),
    [anon_sym_DOT] = ACTIONS(2155),
    [anon_sym_LBRACK] = ACTIONS(2153),
    [anon_sym_true] = ACTIONS(2153),
    [anon_sym_false] = ACTIONS(2153),
    [anon_sym_LBRACE] = ACTIONS(2153),
    [anon_sym_SEMI] = ACTIONS(2153),
    [anon_sym_exception] = ACTIONS(2153),
    [anon_sym_module] = ACTIONS(2153),
    [anon_sym_open] = ACTIONS(2153),
    [anon_sym_include] = ACTIONS(2153),
    [anon_sym_class] = ACTIONS(2153),
    [anon_sym_val] = ACTIONS(2153),
    [anon_sym_object] = ACTIONS(2153),
    [anon_sym_in] = ACTIONS(2153),
    [anon_sym_AMP] = ACTIONS(2153),
    [anon_sym_POUND] = ACTIONS(2153),
    [anon_sym_COLON_COLON] = ACTIONS(2155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2155),
    [anon_sym_else] = ACTIONS(2153),
    [anon_sym_new] = ACTIONS(2153),
    [anon_sym_LBRACE_LT] = ACTIONS(2155),
    [anon_sym_begin] = ACTIONS(2153),
    [sym_ocamlyacc_value] = ACTIONS(2155),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2153),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2155),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2153),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2153),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2155),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2155),
    [aux_sym_number_token1] = ACTIONS(2155),
    [aux_sym_signed_number_token1] = ACTIONS(2153),
    [anon_sym_SQUOTE] = ACTIONS(2155),
    [anon_sym_DQUOTE] = ACTIONS(2155),
    [sym_prefix_operator] = ACTIONS(2155),
    [aux_sym_sign_operator_token1] = ACTIONS(2153),
    [sym_hash_operator] = ACTIONS(2155),
    [anon_sym_lsl] = ACTIONS(2153),
    [anon_sym_lsr] = ACTIONS(2153),
    [anon_sym_asr] = ACTIONS(2153),
    [aux_sym_pow_operator_token1] = ACTIONS(2155),
    [anon_sym_mod] = ACTIONS(2153),
    [anon_sym_land] = ACTIONS(2153),
    [anon_sym_lor] = ACTIONS(2153),
    [anon_sym_lxor] = ACTIONS(2153),
    [aux_sym_mult_operator_token1] = ACTIONS(2153),
    [aux_sym_add_operator_token1] = ACTIONS(2153),
    [sym_concat_operator] = ACTIONS(2155),
    [sym_rel_operator] = ACTIONS(2153),
    [anon_sym_AMP_AMP] = ACTIONS(2153),
    [anon_sym_or] = ACTIONS(2153),
    [anon_sym_PIPE_PIPE] = ACTIONS(2153),
    [sym_assign_operator] = ACTIONS(2155),
    [sym_let_operator] = ACTIONS(2155),
    [sym_let_and_operator] = ACTIONS(2155),
    [sym__uppercase_identifier] = ACTIONS(2155),
    [aux_sym_directive_token1] = ACTIONS(2153),
    [aux_sym_tag_token1] = ACTIONS(2155),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1275)] = {
    [sym_attribute] = STATE(1275),
    [ts_builtin_sym_end] = ACTIONS(2101),
    [sym__lowercase_identifier] = ACTIONS(2099),
    [anon_sym_SEMI_SEMI] = ACTIONS(2101),
    [anon_sym_let] = ACTIONS(2099),
    [anon_sym_and] = ACTIONS(2099),
    [anon_sym_TILDE] = ACTIONS(2099),
    [anon_sym_QMARK] = ACTIONS(2099),
    [anon_sym_LPAREN] = ACTIONS(2101),
    [anon_sym_external] = ACTIONS(2099),
    [anon_sym_type] = ACTIONS(2099),
    [anon_sym_COMMA] = ACTIONS(2101),
    [anon_sym_PIPE] = ACTIONS(2099),
    [anon_sym_DOT] = ACTIONS(2101),
    [anon_sym_LBRACK] = ACTIONS(2099),
    [anon_sym_true] = ACTIONS(2099),
    [anon_sym_false] = ACTIONS(2099),
    [anon_sym_LBRACE] = ACTIONS(2099),
    [anon_sym_SEMI] = ACTIONS(2099),
    [anon_sym_exception] = ACTIONS(2099),
    [anon_sym_module] = ACTIONS(2099),
    [anon_sym_open] = ACTIONS(2099),
    [anon_sym_include] = ACTIONS(2099),
    [anon_sym_class] = ACTIONS(2099),
    [anon_sym_val] = ACTIONS(2099),
    [anon_sym_object] = ACTIONS(2099),
    [anon_sym_in] = ACTIONS(2099),
    [anon_sym_AMP] = ACTIONS(2099),
    [anon_sym_POUND] = ACTIONS(2099),
    [anon_sym_COLON_COLON] = ACTIONS(2101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2101),
    [anon_sym_else] = ACTIONS(2099),
    [anon_sym_new] = ACTIONS(2099),
    [anon_sym_LBRACE_LT] = ACTIONS(2101),
    [anon_sym_begin] = ACTIONS(2099),
    [sym_ocamlyacc_value] = ACTIONS(2101),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2099),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2101),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2099),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2099),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2101),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2101),
    [aux_sym_number_token1] = ACTIONS(2101),
    [aux_sym_signed_number_token1] = ACTIONS(2099),
    [anon_sym_SQUOTE] = ACTIONS(2101),
    [anon_sym_DQUOTE] = ACTIONS(2101),
    [sym_prefix_operator] = ACTIONS(2101),
    [aux_sym_sign_operator_token1] = ACTIONS(2099),
    [sym_hash_operator] = ACTIONS(2101),
    [anon_sym_lsl] = ACTIONS(2099),
    [anon_sym_lsr] = ACTIONS(2099),
    [anon_sym_asr] = ACTIONS(2099),
    [aux_sym_pow_operator_token1] = ACTIONS(2101),
    [anon_sym_mod] = ACTIONS(2099),
    [anon_sym_land] = ACTIONS(2099),
    [anon_sym_lor] = ACTIONS(2099),
    [anon_sym_lxor] = ACTIONS(2099),
    [aux_sym_mult_operator_token1] = ACTIONS(2099),
    [aux_sym_add_operator_token1] = ACTIONS(2099),
    [sym_concat_operator] = ACTIONS(2101),
    [sym_rel_operator] = ACTIONS(2099),
    [anon_sym_AMP_AMP] = ACTIONS(2099),
    [anon_sym_or] = ACTIONS(2099),
    [anon_sym_PIPE_PIPE] = ACTIONS(2099),
    [sym_assign_operator] = ACTIONS(2101),
    [sym_let_operator] = ACTIONS(2101),
    [sym_let_and_operator] = ACTIONS(2101),
    [sym__uppercase_identifier] = ACTIONS(2101),
    [aux_sym_directive_token1] = ACTIONS(2099),
    [aux_sym_tag_token1] = ACTIONS(2101),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1276)] = {
    [sym_attribute] = STATE(1276),
    [ts_builtin_sym_end] = ACTIONS(2167),
    [sym__lowercase_identifier] = ACTIONS(2165),
    [anon_sym_SEMI_SEMI] = ACTIONS(2167),
    [anon_sym_let] = ACTIONS(2165),
    [anon_sym_and] = ACTIONS(2165),
    [anon_sym_TILDE] = ACTIONS(2165),
    [anon_sym_QMARK] = ACTIONS(2165),
    [anon_sym_LPAREN] = ACTIONS(2167),
    [anon_sym_external] = ACTIONS(2165),
    [anon_sym_type] = ACTIONS(2165),
    [anon_sym_COMMA] = ACTIONS(2167),
    [anon_sym_PIPE] = ACTIONS(2165),
    [anon_sym_DOT] = ACTIONS(2167),
    [anon_sym_LBRACK] = ACTIONS(2165),
    [anon_sym_true] = ACTIONS(2165),
    [anon_sym_false] = ACTIONS(2165),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_SEMI] = ACTIONS(2165),
    [anon_sym_exception] = ACTIONS(2165),
    [anon_sym_module] = ACTIONS(2165),
    [anon_sym_open] = ACTIONS(2165),
    [anon_sym_include] = ACTIONS(2165),
    [anon_sym_class] = ACTIONS(2165),
    [anon_sym_val] = ACTIONS(2165),
    [anon_sym_object] = ACTIONS(2165),
    [anon_sym_in] = ACTIONS(2165),
    [anon_sym_AMP] = ACTIONS(2165),
    [anon_sym_POUND] = ACTIONS(2165),
    [anon_sym_COLON_COLON] = ACTIONS(2167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2167),
    [anon_sym_else] = ACTIONS(2165),
    [anon_sym_new] = ACTIONS(2165),
    [anon_sym_LBRACE_LT] = ACTIONS(2167),
    [anon_sym_begin] = ACTIONS(2165),
    [sym_ocamlyacc_value] = ACTIONS(2167),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2165),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2167),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2167),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2167),
    [aux_sym_number_token1] = ACTIONS(2167),
    [aux_sym_signed_number_token1] = ACTIONS(2165),
    [anon_sym_SQUOTE] = ACTIONS(2167),
    [anon_sym_DQUOTE] = ACTIONS(2167),
    [sym_prefix_operator] = ACTIONS(2167),
    [aux_sym_sign_operator_token1] = ACTIONS(2165),
    [sym_hash_operator] = ACTIONS(2167),
    [anon_sym_lsl] = ACTIONS(2165),
    [anon_sym_lsr] = ACTIONS(2165),
    [anon_sym_asr] = ACTIONS(2165),
    [aux_sym_pow_operator_token1] = ACTIONS(2167),
    [anon_sym_mod] = ACTIONS(2165),
    [anon_sym_land] = ACTIONS(2165),
    [anon_sym_lor] = ACTIONS(2165),
    [anon_sym_lxor] = ACTIONS(2165),
    [aux_sym_mult_operator_token1] = ACTIONS(2165),
    [aux_sym_add_operator_token1] = ACTIONS(2165),
    [sym_concat_operator] = ACTIONS(2167),
    [sym_rel_operator] = ACTIONS(2165),
    [anon_sym_AMP_AMP] = ACTIONS(2165),
    [anon_sym_or] = ACTIONS(2165),
    [anon_sym_PIPE_PIPE] = ACTIONS(2165),
    [sym_assign_operator] = ACTIONS(2167),
    [sym_let_operator] = ACTIONS(2167),
    [sym_let_and_operator] = ACTIONS(2167),
    [sym__uppercase_identifier] = ACTIONS(2167),
    [aux_sym_directive_token1] = ACTIONS(2165),
    [aux_sym_tag_token1] = ACTIONS(2167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1277)] = {
    [sym_attribute] = STATE(1277),
    [ts_builtin_sym_end] = ACTIONS(2171),
    [sym__lowercase_identifier] = ACTIONS(2169),
    [anon_sym_SEMI_SEMI] = ACTIONS(2171),
    [anon_sym_let] = ACTIONS(2169),
    [anon_sym_and] = ACTIONS(2169),
    [anon_sym_TILDE] = ACTIONS(2169),
    [anon_sym_QMARK] = ACTIONS(2169),
    [anon_sym_LPAREN] = ACTIONS(2171),
    [anon_sym_external] = ACTIONS(2169),
    [anon_sym_type] = ACTIONS(2169),
    [anon_sym_COMMA] = ACTIONS(2171),
    [anon_sym_PIPE] = ACTIONS(2169),
    [anon_sym_DOT] = ACTIONS(2171),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_true] = ACTIONS(2169),
    [anon_sym_false] = ACTIONS(2169),
    [anon_sym_LBRACE] = ACTIONS(2169),
    [anon_sym_SEMI] = ACTIONS(2169),
    [anon_sym_exception] = ACTIONS(2169),
    [anon_sym_module] = ACTIONS(2169),
    [anon_sym_open] = ACTIONS(2169),
    [anon_sym_include] = ACTIONS(2169),
    [anon_sym_class] = ACTIONS(2169),
    [anon_sym_val] = ACTIONS(2169),
    [anon_sym_object] = ACTIONS(2169),
    [anon_sym_in] = ACTIONS(2169),
    [anon_sym_AMP] = ACTIONS(2169),
    [anon_sym_POUND] = ACTIONS(2169),
    [anon_sym_COLON_COLON] = ACTIONS(2171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2171),
    [anon_sym_else] = ACTIONS(2169),
    [anon_sym_new] = ACTIONS(2169),
    [anon_sym_LBRACE_LT] = ACTIONS(2171),
    [anon_sym_begin] = ACTIONS(2169),
    [sym_ocamlyacc_value] = ACTIONS(2171),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2169),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2171),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2169),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2169),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2171),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2171),
    [aux_sym_number_token1] = ACTIONS(2171),
    [aux_sym_signed_number_token1] = ACTIONS(2169),
    [anon_sym_SQUOTE] = ACTIONS(2171),
    [anon_sym_DQUOTE] = ACTIONS(2171),
    [sym_prefix_operator] = ACTIONS(2171),
    [aux_sym_sign_operator_token1] = ACTIONS(2169),
    [sym_hash_operator] = ACTIONS(2171),
    [anon_sym_lsl] = ACTIONS(2169),
    [anon_sym_lsr] = ACTIONS(2169),
    [anon_sym_asr] = ACTIONS(2169),
    [aux_sym_pow_operator_token1] = ACTIONS(2171),
    [anon_sym_mod] = ACTIONS(2169),
    [anon_sym_land] = ACTIONS(2169),
    [anon_sym_lor] = ACTIONS(2169),
    [anon_sym_lxor] = ACTIONS(2169),
    [aux_sym_mult_operator_token1] = ACTIONS(2169),
    [aux_sym_add_operator_token1] = ACTIONS(2169),
    [sym_concat_operator] = ACTIONS(2171),
    [sym_rel_operator] = ACTIONS(2169),
    [anon_sym_AMP_AMP] = ACTIONS(2169),
    [anon_sym_or] = ACTIONS(2169),
    [anon_sym_PIPE_PIPE] = ACTIONS(2169),
    [sym_assign_operator] = ACTIONS(2171),
    [sym_let_operator] = ACTIONS(2171),
    [sym_let_and_operator] = ACTIONS(2171),
    [sym__uppercase_identifier] = ACTIONS(2171),
    [aux_sym_directive_token1] = ACTIONS(2169),
    [aux_sym_tag_token1] = ACTIONS(2171),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1278)] = {
    [sym_attribute] = STATE(1278),
    [ts_builtin_sym_end] = ACTIONS(2175),
    [sym__lowercase_identifier] = ACTIONS(2173),
    [anon_sym_SEMI_SEMI] = ACTIONS(2175),
    [anon_sym_let] = ACTIONS(2173),
    [anon_sym_and] = ACTIONS(2173),
    [anon_sym_TILDE] = ACTIONS(2173),
    [anon_sym_QMARK] = ACTIONS(2173),
    [anon_sym_LPAREN] = ACTIONS(2175),
    [anon_sym_external] = ACTIONS(2173),
    [anon_sym_type] = ACTIONS(2173),
    [anon_sym_COMMA] = ACTIONS(2175),
    [anon_sym_PIPE] = ACTIONS(2173),
    [anon_sym_DOT] = ACTIONS(2175),
    [anon_sym_LBRACK] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2173),
    [anon_sym_false] = ACTIONS(2173),
    [anon_sym_LBRACE] = ACTIONS(2173),
    [anon_sym_SEMI] = ACTIONS(2173),
    [anon_sym_exception] = ACTIONS(2173),
    [anon_sym_module] = ACTIONS(2173),
    [anon_sym_open] = ACTIONS(2173),
    [anon_sym_include] = ACTIONS(2173),
    [anon_sym_class] = ACTIONS(2173),
    [anon_sym_val] = ACTIONS(2173),
    [anon_sym_object] = ACTIONS(2173),
    [anon_sym_in] = ACTIONS(2173),
    [anon_sym_AMP] = ACTIONS(2173),
    [anon_sym_POUND] = ACTIONS(2173),
    [anon_sym_COLON_COLON] = ACTIONS(2175),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2175),
    [anon_sym_else] = ACTIONS(2173),
    [anon_sym_new] = ACTIONS(2173),
    [anon_sym_LBRACE_LT] = ACTIONS(2175),
    [anon_sym_begin] = ACTIONS(2173),
    [sym_ocamlyacc_value] = ACTIONS(2175),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2173),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2175),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2173),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2175),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2175),
    [aux_sym_number_token1] = ACTIONS(2175),
    [aux_sym_signed_number_token1] = ACTIONS(2173),
    [anon_sym_SQUOTE] = ACTIONS(2175),
    [anon_sym_DQUOTE] = ACTIONS(2175),
    [sym_prefix_operator] = ACTIONS(2175),
    [aux_sym_sign_operator_token1] = ACTIONS(2173),
    [sym_hash_operator] = ACTIONS(2175),
    [anon_sym_lsl] = ACTIONS(2173),
    [anon_sym_lsr] = ACTIONS(2173),
    [anon_sym_asr] = ACTIONS(2173),
    [aux_sym_pow_operator_token1] = ACTIONS(2175),
    [anon_sym_mod] = ACTIONS(2173),
    [anon_sym_land] = ACTIONS(2173),
    [anon_sym_lor] = ACTIONS(2173),
    [anon_sym_lxor] = ACTIONS(2173),
    [aux_sym_mult_operator_token1] = ACTIONS(2173),
    [aux_sym_add_operator_token1] = ACTIONS(2173),
    [sym_concat_operator] = ACTIONS(2175),
    [sym_rel_operator] = ACTIONS(2173),
    [anon_sym_AMP_AMP] = ACTIONS(2173),
    [anon_sym_or] = ACTIONS(2173),
    [anon_sym_PIPE_PIPE] = ACTIONS(2173),
    [sym_assign_operator] = ACTIONS(2175),
    [sym_let_operator] = ACTIONS(2175),
    [sym_let_and_operator] = ACTIONS(2175),
    [sym__uppercase_identifier] = ACTIONS(2175),
    [aux_sym_directive_token1] = ACTIONS(2173),
    [aux_sym_tag_token1] = ACTIONS(2175),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1279)] = {
    [sym_attribute] = STATE(1279),
    [ts_builtin_sym_end] = ACTIONS(2185),
    [sym__lowercase_identifier] = ACTIONS(2183),
    [anon_sym_SEMI_SEMI] = ACTIONS(2185),
    [anon_sym_let] = ACTIONS(2183),
    [anon_sym_and] = ACTIONS(2183),
    [anon_sym_TILDE] = ACTIONS(2183),
    [anon_sym_QMARK] = ACTIONS(2183),
    [anon_sym_LPAREN] = ACTIONS(2185),
    [anon_sym_external] = ACTIONS(2183),
    [anon_sym_type] = ACTIONS(2183),
    [anon_sym_COMMA] = ACTIONS(2185),
    [anon_sym_PIPE] = ACTIONS(2183),
    [anon_sym_DOT] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2183),
    [anon_sym_true] = ACTIONS(2183),
    [anon_sym_false] = ACTIONS(2183),
    [anon_sym_LBRACE] = ACTIONS(2183),
    [anon_sym_SEMI] = ACTIONS(2183),
    [anon_sym_exception] = ACTIONS(2183),
    [anon_sym_module] = ACTIONS(2183),
    [anon_sym_open] = ACTIONS(2183),
    [anon_sym_include] = ACTIONS(2183),
    [anon_sym_class] = ACTIONS(2183),
    [anon_sym_val] = ACTIONS(2183),
    [anon_sym_object] = ACTIONS(2183),
    [anon_sym_in] = ACTIONS(2183),
    [anon_sym_AMP] = ACTIONS(2183),
    [anon_sym_POUND] = ACTIONS(2183),
    [anon_sym_COLON_COLON] = ACTIONS(2185),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2185),
    [anon_sym_else] = ACTIONS(2183),
    [anon_sym_new] = ACTIONS(2183),
    [anon_sym_LBRACE_LT] = ACTIONS(2185),
    [anon_sym_begin] = ACTIONS(2183),
    [sym_ocamlyacc_value] = ACTIONS(2185),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2183),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2185),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2183),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2183),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2185),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2185),
    [aux_sym_number_token1] = ACTIONS(2185),
    [aux_sym_signed_number_token1] = ACTIONS(2183),
    [anon_sym_SQUOTE] = ACTIONS(2185),
    [anon_sym_DQUOTE] = ACTIONS(2185),
    [sym_prefix_operator] = ACTIONS(2185),
    [aux_sym_sign_operator_token1] = ACTIONS(2183),
    [sym_hash_operator] = ACTIONS(2185),
    [anon_sym_lsl] = ACTIONS(2183),
    [anon_sym_lsr] = ACTIONS(2183),
    [anon_sym_asr] = ACTIONS(2183),
    [aux_sym_pow_operator_token1] = ACTIONS(2185),
    [anon_sym_mod] = ACTIONS(2183),
    [anon_sym_land] = ACTIONS(2183),
    [anon_sym_lor] = ACTIONS(2183),
    [anon_sym_lxor] = ACTIONS(2183),
    [aux_sym_mult_operator_token1] = ACTIONS(2183),
    [aux_sym_add_operator_token1] = ACTIONS(2183),
    [sym_concat_operator] = ACTIONS(2185),
    [sym_rel_operator] = ACTIONS(2183),
    [anon_sym_AMP_AMP] = ACTIONS(2183),
    [anon_sym_or] = ACTIONS(2183),
    [anon_sym_PIPE_PIPE] = ACTIONS(2183),
    [sym_assign_operator] = ACTIONS(2185),
    [sym_let_operator] = ACTIONS(2185),
    [sym_let_and_operator] = ACTIONS(2185),
    [sym__uppercase_identifier] = ACTIONS(2185),
    [aux_sym_directive_token1] = ACTIONS(2183),
    [aux_sym_tag_token1] = ACTIONS(2185),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1280)] = {
    [sym_attribute] = STATE(1280),
    [ts_builtin_sym_end] = ACTIONS(2189),
    [sym__lowercase_identifier] = ACTIONS(2187),
    [anon_sym_SEMI_SEMI] = ACTIONS(2189),
    [anon_sym_let] = ACTIONS(2187),
    [anon_sym_and] = ACTIONS(2187),
    [anon_sym_TILDE] = ACTIONS(2187),
    [anon_sym_QMARK] = ACTIONS(2187),
    [anon_sym_LPAREN] = ACTIONS(2189),
    [anon_sym_external] = ACTIONS(2187),
    [anon_sym_type] = ACTIONS(2187),
    [anon_sym_COMMA] = ACTIONS(2189),
    [anon_sym_PIPE] = ACTIONS(2187),
    [anon_sym_DOT] = ACTIONS(2189),
    [anon_sym_LBRACK] = ACTIONS(2187),
    [anon_sym_true] = ACTIONS(2187),
    [anon_sym_false] = ACTIONS(2187),
    [anon_sym_LBRACE] = ACTIONS(2187),
    [anon_sym_SEMI] = ACTIONS(2187),
    [anon_sym_exception] = ACTIONS(2187),
    [anon_sym_module] = ACTIONS(2187),
    [anon_sym_open] = ACTIONS(2187),
    [anon_sym_include] = ACTIONS(2187),
    [anon_sym_class] = ACTIONS(2187),
    [anon_sym_val] = ACTIONS(2187),
    [anon_sym_object] = ACTIONS(2187),
    [anon_sym_in] = ACTIONS(2187),
    [anon_sym_AMP] = ACTIONS(2187),
    [anon_sym_POUND] = ACTIONS(2187),
    [anon_sym_COLON_COLON] = ACTIONS(2189),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2189),
    [anon_sym_else] = ACTIONS(2187),
    [anon_sym_new] = ACTIONS(2187),
    [anon_sym_LBRACE_LT] = ACTIONS(2189),
    [anon_sym_begin] = ACTIONS(2187),
    [sym_ocamlyacc_value] = ACTIONS(2189),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2187),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2189),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2187),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2187),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2189),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2189),
    [aux_sym_number_token1] = ACTIONS(2189),
    [aux_sym_signed_number_token1] = ACTIONS(2187),
    [anon_sym_SQUOTE] = ACTIONS(2189),
    [anon_sym_DQUOTE] = ACTIONS(2189),
    [sym_prefix_operator] = ACTIONS(2189),
    [aux_sym_sign_operator_token1] = ACTIONS(2187),
    [sym_hash_operator] = ACTIONS(2189),
    [anon_sym_lsl] = ACTIONS(2187),
    [anon_sym_lsr] = ACTIONS(2187),
    [anon_sym_asr] = ACTIONS(2187),
    [aux_sym_pow_operator_token1] = ACTIONS(2189),
    [anon_sym_mod] = ACTIONS(2187),
    [anon_sym_land] = ACTIONS(2187),
    [anon_sym_lor] = ACTIONS(2187),
    [anon_sym_lxor] = ACTIONS(2187),
    [aux_sym_mult_operator_token1] = ACTIONS(2187),
    [aux_sym_add_operator_token1] = ACTIONS(2187),
    [sym_concat_operator] = ACTIONS(2189),
    [sym_rel_operator] = ACTIONS(2187),
    [anon_sym_AMP_AMP] = ACTIONS(2187),
    [anon_sym_or] = ACTIONS(2187),
    [anon_sym_PIPE_PIPE] = ACTIONS(2187),
    [sym_assign_operator] = ACTIONS(2189),
    [sym_let_operator] = ACTIONS(2189),
    [sym_let_and_operator] = ACTIONS(2189),
    [sym__uppercase_identifier] = ACTIONS(2189),
    [aux_sym_directive_token1] = ACTIONS(2187),
    [aux_sym_tag_token1] = ACTIONS(2189),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1281)] = {
    [sym_attribute] = STATE(1281),
    [ts_builtin_sym_end] = ACTIONS(2193),
    [sym__lowercase_identifier] = ACTIONS(2191),
    [anon_sym_SEMI_SEMI] = ACTIONS(2193),
    [anon_sym_let] = ACTIONS(2191),
    [anon_sym_and] = ACTIONS(2191),
    [anon_sym_TILDE] = ACTIONS(2191),
    [anon_sym_QMARK] = ACTIONS(2191),
    [anon_sym_LPAREN] = ACTIONS(2193),
    [anon_sym_external] = ACTIONS(2191),
    [anon_sym_type] = ACTIONS(2191),
    [anon_sym_COMMA] = ACTIONS(2193),
    [anon_sym_PIPE] = ACTIONS(2191),
    [anon_sym_DOT] = ACTIONS(2193),
    [anon_sym_LBRACK] = ACTIONS(2191),
    [anon_sym_true] = ACTIONS(2191),
    [anon_sym_false] = ACTIONS(2191),
    [anon_sym_LBRACE] = ACTIONS(2191),
    [anon_sym_SEMI] = ACTIONS(2191),
    [anon_sym_exception] = ACTIONS(2191),
    [anon_sym_module] = ACTIONS(2191),
    [anon_sym_open] = ACTIONS(2191),
    [anon_sym_include] = ACTIONS(2191),
    [anon_sym_class] = ACTIONS(2191),
    [anon_sym_val] = ACTIONS(2191),
    [anon_sym_object] = ACTIONS(2191),
    [anon_sym_in] = ACTIONS(2191),
    [anon_sym_AMP] = ACTIONS(2191),
    [anon_sym_POUND] = ACTIONS(2191),
    [anon_sym_COLON_COLON] = ACTIONS(2193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2193),
    [anon_sym_else] = ACTIONS(2191),
    [anon_sym_new] = ACTIONS(2191),
    [anon_sym_LBRACE_LT] = ACTIONS(2193),
    [anon_sym_begin] = ACTIONS(2191),
    [sym_ocamlyacc_value] = ACTIONS(2193),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2191),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2193),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2191),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2191),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2193),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2193),
    [aux_sym_number_token1] = ACTIONS(2193),
    [aux_sym_signed_number_token1] = ACTIONS(2191),
    [anon_sym_SQUOTE] = ACTIONS(2193),
    [anon_sym_DQUOTE] = ACTIONS(2193),
    [sym_prefix_operator] = ACTIONS(2193),
    [aux_sym_sign_operator_token1] = ACTIONS(2191),
    [sym_hash_operator] = ACTIONS(2193),
    [anon_sym_lsl] = ACTIONS(2191),
    [anon_sym_lsr] = ACTIONS(2191),
    [anon_sym_asr] = ACTIONS(2191),
    [aux_sym_pow_operator_token1] = ACTIONS(2193),
    [anon_sym_mod] = ACTIONS(2191),
    [anon_sym_land] = ACTIONS(2191),
    [anon_sym_lor] = ACTIONS(2191),
    [anon_sym_lxor] = ACTIONS(2191),
    [aux_sym_mult_operator_token1] = ACTIONS(2191),
    [aux_sym_add_operator_token1] = ACTIONS(2191),
    [sym_concat_operator] = ACTIONS(2193),
    [sym_rel_operator] = ACTIONS(2191),
    [anon_sym_AMP_AMP] = ACTIONS(2191),
    [anon_sym_or] = ACTIONS(2191),
    [anon_sym_PIPE_PIPE] = ACTIONS(2191),
    [sym_assign_operator] = ACTIONS(2193),
    [sym_let_operator] = ACTIONS(2193),
    [sym_let_and_operator] = ACTIONS(2193),
    [sym__uppercase_identifier] = ACTIONS(2193),
    [aux_sym_directive_token1] = ACTIONS(2191),
    [aux_sym_tag_token1] = ACTIONS(2193),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1282)] = {
    [sym_attribute] = STATE(1282),
    [ts_builtin_sym_end] = ACTIONS(2113),
    [sym__lowercase_identifier] = ACTIONS(2111),
    [anon_sym_SEMI_SEMI] = ACTIONS(2113),
    [anon_sym_let] = ACTIONS(2111),
    [anon_sym_and] = ACTIONS(2111),
    [anon_sym_TILDE] = ACTIONS(2111),
    [anon_sym_QMARK] = ACTIONS(2111),
    [anon_sym_LPAREN] = ACTIONS(2113),
    [anon_sym_external] = ACTIONS(2111),
    [anon_sym_type] = ACTIONS(2111),
    [anon_sym_COMMA] = ACTIONS(2113),
    [anon_sym_PIPE] = ACTIONS(2111),
    [anon_sym_DOT] = ACTIONS(2115),
    [anon_sym_LBRACK] = ACTIONS(2111),
    [anon_sym_true] = ACTIONS(2111),
    [anon_sym_false] = ACTIONS(2111),
    [anon_sym_LBRACE] = ACTIONS(2111),
    [anon_sym_SEMI] = ACTIONS(2111),
    [anon_sym_exception] = ACTIONS(2111),
    [anon_sym_module] = ACTIONS(2111),
    [anon_sym_open] = ACTIONS(2111),
    [anon_sym_include] = ACTIONS(2111),
    [anon_sym_class] = ACTIONS(2111),
    [anon_sym_val] = ACTIONS(2111),
    [anon_sym_object] = ACTIONS(2111),
    [anon_sym_in] = ACTIONS(2111),
    [anon_sym_AMP] = ACTIONS(2111),
    [anon_sym_POUND] = ACTIONS(2111),
    [anon_sym_COLON_COLON] = ACTIONS(2113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2113),
    [anon_sym_else] = ACTIONS(2111),
    [anon_sym_new] = ACTIONS(2111),
    [anon_sym_LBRACE_LT] = ACTIONS(2113),
    [anon_sym_begin] = ACTIONS(2111),
    [sym_ocamlyacc_value] = ACTIONS(2113),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2113),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2113),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2113),
    [aux_sym_number_token1] = ACTIONS(2113),
    [aux_sym_signed_number_token1] = ACTIONS(2111),
    [anon_sym_SQUOTE] = ACTIONS(2113),
    [anon_sym_DQUOTE] = ACTIONS(2113),
    [sym_prefix_operator] = ACTIONS(2113),
    [aux_sym_sign_operator_token1] = ACTIONS(2111),
    [sym_hash_operator] = ACTIONS(2113),
    [anon_sym_lsl] = ACTIONS(2111),
    [anon_sym_lsr] = ACTIONS(2111),
    [anon_sym_asr] = ACTIONS(2111),
    [aux_sym_pow_operator_token1] = ACTIONS(2113),
    [anon_sym_mod] = ACTIONS(2111),
    [anon_sym_land] = ACTIONS(2111),
    [anon_sym_lor] = ACTIONS(2111),
    [anon_sym_lxor] = ACTIONS(2111),
    [aux_sym_mult_operator_token1] = ACTIONS(2111),
    [aux_sym_add_operator_token1] = ACTIONS(2111),
    [sym_concat_operator] = ACTIONS(2113),
    [sym_rel_operator] = ACTIONS(2111),
    [anon_sym_AMP_AMP] = ACTIONS(2111),
    [anon_sym_or] = ACTIONS(2111),
    [anon_sym_PIPE_PIPE] = ACTIONS(2111),
    [sym_assign_operator] = ACTIONS(2113),
    [sym_let_operator] = ACTIONS(2113),
    [sym_let_and_operator] = ACTIONS(2113),
    [sym__uppercase_identifier] = ACTIONS(2113),
    [aux_sym_directive_token1] = ACTIONS(2111),
    [aux_sym_tag_token1] = ACTIONS(2113),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1283)] = {
    [sym_attribute] = STATE(1283),
    [ts_builtin_sym_end] = ACTIONS(2123),
    [sym__lowercase_identifier] = ACTIONS(2121),
    [anon_sym_SEMI_SEMI] = ACTIONS(2123),
    [anon_sym_let] = ACTIONS(2121),
    [anon_sym_and] = ACTIONS(2121),
    [anon_sym_TILDE] = ACTIONS(2121),
    [anon_sym_QMARK] = ACTIONS(2121),
    [anon_sym_LPAREN] = ACTIONS(2123),
    [anon_sym_external] = ACTIONS(2121),
    [anon_sym_type] = ACTIONS(2121),
    [anon_sym_COMMA] = ACTIONS(2123),
    [anon_sym_PIPE] = ACTIONS(2121),
    [anon_sym_DOT] = ACTIONS(2123),
    [anon_sym_LBRACK] = ACTIONS(2121),
    [anon_sym_true] = ACTIONS(2121),
    [anon_sym_false] = ACTIONS(2121),
    [anon_sym_LBRACE] = ACTIONS(2121),
    [anon_sym_SEMI] = ACTIONS(2121),
    [anon_sym_exception] = ACTIONS(2121),
    [anon_sym_module] = ACTIONS(2121),
    [anon_sym_open] = ACTIONS(2121),
    [anon_sym_include] = ACTIONS(2121),
    [anon_sym_class] = ACTIONS(2121),
    [anon_sym_val] = ACTIONS(2121),
    [anon_sym_object] = ACTIONS(2121),
    [anon_sym_in] = ACTIONS(2121),
    [anon_sym_AMP] = ACTIONS(2121),
    [anon_sym_POUND] = ACTIONS(2121),
    [anon_sym_COLON_COLON] = ACTIONS(2123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2123),
    [anon_sym_else] = ACTIONS(2121),
    [anon_sym_new] = ACTIONS(2121),
    [anon_sym_LBRACE_LT] = ACTIONS(2123),
    [anon_sym_begin] = ACTIONS(2121),
    [sym_ocamlyacc_value] = ACTIONS(2123),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2121),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2123),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2121),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2121),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2123),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2123),
    [aux_sym_number_token1] = ACTIONS(2123),
    [aux_sym_signed_number_token1] = ACTIONS(2121),
    [anon_sym_SQUOTE] = ACTIONS(2123),
    [anon_sym_DQUOTE] = ACTIONS(2123),
    [sym_prefix_operator] = ACTIONS(2123),
    [aux_sym_sign_operator_token1] = ACTIONS(2121),
    [sym_hash_operator] = ACTIONS(2123),
    [anon_sym_lsl] = ACTIONS(2121),
    [anon_sym_lsr] = ACTIONS(2121),
    [anon_sym_asr] = ACTIONS(2121),
    [aux_sym_pow_operator_token1] = ACTIONS(2123),
    [anon_sym_mod] = ACTIONS(2121),
    [anon_sym_land] = ACTIONS(2121),
    [anon_sym_lor] = ACTIONS(2121),
    [anon_sym_lxor] = ACTIONS(2121),
    [aux_sym_mult_operator_token1] = ACTIONS(2121),
    [aux_sym_add_operator_token1] = ACTIONS(2121),
    [sym_concat_operator] = ACTIONS(2123),
    [sym_rel_operator] = ACTIONS(2121),
    [anon_sym_AMP_AMP] = ACTIONS(2121),
    [anon_sym_or] = ACTIONS(2121),
    [anon_sym_PIPE_PIPE] = ACTIONS(2121),
    [sym_assign_operator] = ACTIONS(2123),
    [sym_let_operator] = ACTIONS(2123),
    [sym_let_and_operator] = ACTIONS(2123),
    [sym__uppercase_identifier] = ACTIONS(2123),
    [aux_sym_directive_token1] = ACTIONS(2121),
    [aux_sym_tag_token1] = ACTIONS(2123),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1284)] = {
    [sym_attribute] = STATE(1284),
    [ts_builtin_sym_end] = ACTIONS(2159),
    [sym__lowercase_identifier] = ACTIONS(2157),
    [anon_sym_SEMI_SEMI] = ACTIONS(2159),
    [anon_sym_let] = ACTIONS(2157),
    [anon_sym_and] = ACTIONS(2157),
    [anon_sym_TILDE] = ACTIONS(2157),
    [anon_sym_QMARK] = ACTIONS(2157),
    [anon_sym_LPAREN] = ACTIONS(2159),
    [anon_sym_external] = ACTIONS(2157),
    [anon_sym_type] = ACTIONS(2157),
    [anon_sym_COMMA] = ACTIONS(2159),
    [anon_sym_PIPE] = ACTIONS(2157),
    [anon_sym_DOT] = ACTIONS(2159),
    [anon_sym_LBRACK] = ACTIONS(2157),
    [anon_sym_true] = ACTIONS(2157),
    [anon_sym_false] = ACTIONS(2157),
    [anon_sym_LBRACE] = ACTIONS(2157),
    [anon_sym_SEMI] = ACTIONS(2157),
    [anon_sym_exception] = ACTIONS(2157),
    [anon_sym_module] = ACTIONS(2157),
    [anon_sym_open] = ACTIONS(2157),
    [anon_sym_include] = ACTIONS(2157),
    [anon_sym_class] = ACTIONS(2157),
    [anon_sym_val] = ACTIONS(2157),
    [anon_sym_object] = ACTIONS(2157),
    [anon_sym_in] = ACTIONS(2157),
    [anon_sym_AMP] = ACTIONS(2157),
    [anon_sym_POUND] = ACTIONS(2157),
    [anon_sym_COLON_COLON] = ACTIONS(2159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2159),
    [anon_sym_else] = ACTIONS(2157),
    [anon_sym_new] = ACTIONS(2157),
    [anon_sym_LBRACE_LT] = ACTIONS(2159),
    [anon_sym_begin] = ACTIONS(2157),
    [sym_ocamlyacc_value] = ACTIONS(2159),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2157),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2159),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2159),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2159),
    [aux_sym_number_token1] = ACTIONS(2159),
    [aux_sym_signed_number_token1] = ACTIONS(2157),
    [anon_sym_SQUOTE] = ACTIONS(2159),
    [anon_sym_DQUOTE] = ACTIONS(2159),
    [sym_prefix_operator] = ACTIONS(2159),
    [aux_sym_sign_operator_token1] = ACTIONS(2157),
    [sym_hash_operator] = ACTIONS(2159),
    [anon_sym_lsl] = ACTIONS(2157),
    [anon_sym_lsr] = ACTIONS(2157),
    [anon_sym_asr] = ACTIONS(2157),
    [aux_sym_pow_operator_token1] = ACTIONS(2159),
    [anon_sym_mod] = ACTIONS(2157),
    [anon_sym_land] = ACTIONS(2157),
    [anon_sym_lor] = ACTIONS(2157),
    [anon_sym_lxor] = ACTIONS(2157),
    [aux_sym_mult_operator_token1] = ACTIONS(2157),
    [aux_sym_add_operator_token1] = ACTIONS(2157),
    [sym_concat_operator] = ACTIONS(2159),
    [sym_rel_operator] = ACTIONS(2157),
    [anon_sym_AMP_AMP] = ACTIONS(2157),
    [anon_sym_or] = ACTIONS(2157),
    [anon_sym_PIPE_PIPE] = ACTIONS(2157),
    [sym_assign_operator] = ACTIONS(2159),
    [sym_let_operator] = ACTIONS(2159),
    [sym_let_and_operator] = ACTIONS(2159),
    [sym__uppercase_identifier] = ACTIONS(2159),
    [aux_sym_directive_token1] = ACTIONS(2157),
    [aux_sym_tag_token1] = ACTIONS(2159),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1285)] = {
    [sym_attribute] = STATE(1285),
    [ts_builtin_sym_end] = ACTIONS(2077),
    [sym__lowercase_identifier] = ACTIONS(2075),
    [anon_sym_SEMI_SEMI] = ACTIONS(2077),
    [anon_sym_let] = ACTIONS(2075),
    [anon_sym_and] = ACTIONS(2075),
    [anon_sym_TILDE] = ACTIONS(2075),
    [anon_sym_QMARK] = ACTIONS(2075),
    [anon_sym_LPAREN] = ACTIONS(2077),
    [anon_sym_external] = ACTIONS(2075),
    [anon_sym_type] = ACTIONS(2075),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_PIPE] = ACTIONS(2075),
    [anon_sym_DOT] = ACTIONS(2077),
    [anon_sym_LBRACK] = ACTIONS(2075),
    [anon_sym_true] = ACTIONS(2075),
    [anon_sym_false] = ACTIONS(2075),
    [anon_sym_LBRACE] = ACTIONS(2075),
    [anon_sym_SEMI] = ACTIONS(2075),
    [anon_sym_exception] = ACTIONS(2075),
    [anon_sym_module] = ACTIONS(2075),
    [anon_sym_open] = ACTIONS(2075),
    [anon_sym_include] = ACTIONS(2075),
    [anon_sym_class] = ACTIONS(2075),
    [anon_sym_val] = ACTIONS(2075),
    [anon_sym_object] = ACTIONS(2075),
    [anon_sym_in] = ACTIONS(2075),
    [anon_sym_AMP] = ACTIONS(2075),
    [anon_sym_POUND] = ACTIONS(2075),
    [anon_sym_COLON_COLON] = ACTIONS(2077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2077),
    [anon_sym_else] = ACTIONS(2075),
    [anon_sym_new] = ACTIONS(2075),
    [anon_sym_LBRACE_LT] = ACTIONS(2077),
    [anon_sym_begin] = ACTIONS(2075),
    [sym_ocamlyacc_value] = ACTIONS(2077),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2075),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2077),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2075),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2075),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2077),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2077),
    [aux_sym_number_token1] = ACTIONS(2077),
    [aux_sym_signed_number_token1] = ACTIONS(2075),
    [anon_sym_SQUOTE] = ACTIONS(2077),
    [anon_sym_DQUOTE] = ACTIONS(2077),
    [sym_prefix_operator] = ACTIONS(2077),
    [aux_sym_sign_operator_token1] = ACTIONS(2075),
    [sym_hash_operator] = ACTIONS(2077),
    [anon_sym_lsl] = ACTIONS(2075),
    [anon_sym_lsr] = ACTIONS(2075),
    [anon_sym_asr] = ACTIONS(2075),
    [aux_sym_pow_operator_token1] = ACTIONS(2077),
    [anon_sym_mod] = ACTIONS(2075),
    [anon_sym_land] = ACTIONS(2075),
    [anon_sym_lor] = ACTIONS(2075),
    [anon_sym_lxor] = ACTIONS(2075),
    [aux_sym_mult_operator_token1] = ACTIONS(2075),
    [aux_sym_add_operator_token1] = ACTIONS(2075),
    [sym_concat_operator] = ACTIONS(2077),
    [sym_rel_operator] = ACTIONS(2075),
    [anon_sym_AMP_AMP] = ACTIONS(2075),
    [anon_sym_or] = ACTIONS(2075),
    [anon_sym_PIPE_PIPE] = ACTIONS(2075),
    [sym_assign_operator] = ACTIONS(2077),
    [sym_let_operator] = ACTIONS(2077),
    [sym_let_and_operator] = ACTIONS(2077),
    [sym__uppercase_identifier] = ACTIONS(2077),
    [aux_sym_directive_token1] = ACTIONS(2075),
    [aux_sym_tag_token1] = ACTIONS(2077),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1286)] = {
    [sym_attribute] = STATE(1286),
    [sym__lowercase_identifier] = ACTIONS(1887),
    [anon_sym_SEMI_SEMI] = ACTIONS(1889),
    [anon_sym_let] = ACTIONS(1887),
    [anon_sym_and] = ACTIONS(1887),
    [anon_sym_TILDE] = ACTIONS(1887),
    [anon_sym_QMARK] = ACTIONS(1887),
    [anon_sym_LPAREN] = ACTIONS(1889),
    [anon_sym_external] = ACTIONS(1887),
    [anon_sym_type] = ACTIONS(1887),
    [anon_sym_COMMA] = ACTIONS(1889),
    [anon_sym_PIPE] = ACTIONS(1887),
    [anon_sym_DOT] = ACTIONS(1889),
    [anon_sym_LBRACK] = ACTIONS(1887),
    [anon_sym_RBRACK] = ACTIONS(1889),
    [anon_sym_true] = ACTIONS(1887),
    [anon_sym_false] = ACTIONS(1887),
    [anon_sym_LBRACE] = ACTIONS(1887),
    [anon_sym_SEMI] = ACTIONS(1887),
    [anon_sym_exception] = ACTIONS(1887),
    [anon_sym_module] = ACTIONS(1887),
    [anon_sym_open] = ACTIONS(1887),
    [anon_sym_include] = ACTIONS(1887),
    [anon_sym_class] = ACTIONS(1887),
    [anon_sym_val] = ACTIONS(1887),
    [anon_sym_end] = ACTIONS(1887),
    [anon_sym_object] = ACTIONS(1887),
    [anon_sym_AMP] = ACTIONS(1887),
    [anon_sym_POUND] = ACTIONS(1887),
    [anon_sym_COLON_COLON] = ACTIONS(1889),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1889),
    [anon_sym_LT_DASH] = ACTIONS(1887),
    [anon_sym_else] = ACTIONS(1887),
    [anon_sym_new] = ACTIONS(1887),
    [anon_sym_LBRACE_LT] = ACTIONS(1889),
    [anon_sym_begin] = ACTIONS(1887),
    [sym_ocamlyacc_value] = ACTIONS(1889),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1887),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1889),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1887),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1887),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1889),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1889),
    [aux_sym_number_token1] = ACTIONS(1889),
    [aux_sym_signed_number_token1] = ACTIONS(1887),
    [anon_sym_SQUOTE] = ACTIONS(1889),
    [anon_sym_DQUOTE] = ACTIONS(1889),
    [sym_prefix_operator] = ACTIONS(1889),
    [aux_sym_sign_operator_token1] = ACTIONS(1887),
    [sym_hash_operator] = ACTIONS(1889),
    [anon_sym_lsl] = ACTIONS(1887),
    [anon_sym_lsr] = ACTIONS(1887),
    [anon_sym_asr] = ACTIONS(1887),
    [aux_sym_pow_operator_token1] = ACTIONS(1889),
    [anon_sym_mod] = ACTIONS(1887),
    [anon_sym_land] = ACTIONS(1887),
    [anon_sym_lor] = ACTIONS(1887),
    [anon_sym_lxor] = ACTIONS(1887),
    [aux_sym_mult_operator_token1] = ACTIONS(1887),
    [aux_sym_add_operator_token1] = ACTIONS(1887),
    [sym_concat_operator] = ACTIONS(1889),
    [sym_rel_operator] = ACTIONS(1887),
    [anon_sym_AMP_AMP] = ACTIONS(1887),
    [anon_sym_or] = ACTIONS(1887),
    [anon_sym_PIPE_PIPE] = ACTIONS(1887),
    [sym_assign_operator] = ACTIONS(1889),
    [sym_let_operator] = ACTIONS(1889),
    [sym__uppercase_identifier] = ACTIONS(1889),
    [aux_sym_directive_token1] = ACTIONS(1887),
    [aux_sym_tag_token1] = ACTIONS(1889),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1287)] = {
    [sym_attribute] = STATE(1287),
    [ts_builtin_sym_end] = ACTIONS(1981),
    [sym__lowercase_identifier] = ACTIONS(1979),
    [anon_sym_SEMI_SEMI] = ACTIONS(1981),
    [anon_sym_let] = ACTIONS(1979),
    [anon_sym_and] = ACTIONS(1979),
    [anon_sym_TILDE] = ACTIONS(1979),
    [anon_sym_QMARK] = ACTIONS(1979),
    [anon_sym_LPAREN] = ACTIONS(1981),
    [anon_sym_external] = ACTIONS(1979),
    [anon_sym_type] = ACTIONS(1979),
    [anon_sym_COMMA] = ACTIONS(1981),
    [anon_sym_PIPE] = ACTIONS(1979),
    [anon_sym_DOT] = ACTIONS(1981),
    [anon_sym_LBRACK] = ACTIONS(1979),
    [anon_sym_true] = ACTIONS(1979),
    [anon_sym_false] = ACTIONS(1979),
    [anon_sym_LBRACE] = ACTIONS(1979),
    [anon_sym_SEMI] = ACTIONS(1979),
    [anon_sym_exception] = ACTIONS(1979),
    [anon_sym_module] = ACTIONS(1979),
    [anon_sym_open] = ACTIONS(1979),
    [anon_sym_include] = ACTIONS(1979),
    [anon_sym_class] = ACTIONS(1979),
    [anon_sym_val] = ACTIONS(1979),
    [anon_sym_object] = ACTIONS(1979),
    [anon_sym_in] = ACTIONS(1979),
    [anon_sym_AMP] = ACTIONS(1979),
    [anon_sym_POUND] = ACTIONS(1979),
    [anon_sym_COLON_COLON] = ACTIONS(1981),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1981),
    [anon_sym_else] = ACTIONS(1979),
    [anon_sym_new] = ACTIONS(1979),
    [anon_sym_LBRACE_LT] = ACTIONS(1981),
    [anon_sym_begin] = ACTIONS(1979),
    [sym_ocamlyacc_value] = ACTIONS(1981),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1979),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1981),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1979),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1979),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1981),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1981),
    [aux_sym_number_token1] = ACTIONS(1981),
    [aux_sym_signed_number_token1] = ACTIONS(1979),
    [anon_sym_SQUOTE] = ACTIONS(1981),
    [anon_sym_DQUOTE] = ACTIONS(1981),
    [sym_prefix_operator] = ACTIONS(1981),
    [aux_sym_sign_operator_token1] = ACTIONS(1979),
    [sym_hash_operator] = ACTIONS(1981),
    [anon_sym_lsl] = ACTIONS(1979),
    [anon_sym_lsr] = ACTIONS(1979),
    [anon_sym_asr] = ACTIONS(1979),
    [aux_sym_pow_operator_token1] = ACTIONS(1981),
    [anon_sym_mod] = ACTIONS(1979),
    [anon_sym_land] = ACTIONS(1979),
    [anon_sym_lor] = ACTIONS(1979),
    [anon_sym_lxor] = ACTIONS(1979),
    [aux_sym_mult_operator_token1] = ACTIONS(1979),
    [aux_sym_add_operator_token1] = ACTIONS(1979),
    [sym_concat_operator] = ACTIONS(1981),
    [sym_rel_operator] = ACTIONS(1979),
    [anon_sym_AMP_AMP] = ACTIONS(1979),
    [anon_sym_or] = ACTIONS(1979),
    [anon_sym_PIPE_PIPE] = ACTIONS(1979),
    [sym_assign_operator] = ACTIONS(1981),
    [sym_let_operator] = ACTIONS(1981),
    [sym_let_and_operator] = ACTIONS(1981),
    [sym__uppercase_identifier] = ACTIONS(1981),
    [aux_sym_directive_token1] = ACTIONS(1979),
    [aux_sym_tag_token1] = ACTIONS(1981),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1288)] = {
    [sym_attribute] = STATE(1288),
    [ts_builtin_sym_end] = ACTIONS(2005),
    [sym__lowercase_identifier] = ACTIONS(2003),
    [anon_sym_SEMI_SEMI] = ACTIONS(2005),
    [anon_sym_let] = ACTIONS(2003),
    [anon_sym_and] = ACTIONS(2003),
    [anon_sym_TILDE] = ACTIONS(2003),
    [anon_sym_QMARK] = ACTIONS(2003),
    [anon_sym_LPAREN] = ACTIONS(2005),
    [anon_sym_external] = ACTIONS(2003),
    [anon_sym_type] = ACTIONS(2003),
    [anon_sym_COMMA] = ACTIONS(2005),
    [anon_sym_PIPE] = ACTIONS(2003),
    [anon_sym_DOT] = ACTIONS(2005),
    [anon_sym_LBRACK] = ACTIONS(2003),
    [anon_sym_true] = ACTIONS(2003),
    [anon_sym_false] = ACTIONS(2003),
    [anon_sym_LBRACE] = ACTIONS(2003),
    [anon_sym_SEMI] = ACTIONS(2003),
    [anon_sym_exception] = ACTIONS(2003),
    [anon_sym_module] = ACTIONS(2003),
    [anon_sym_open] = ACTIONS(2003),
    [anon_sym_include] = ACTIONS(2003),
    [anon_sym_class] = ACTIONS(2003),
    [anon_sym_val] = ACTIONS(2003),
    [anon_sym_object] = ACTIONS(2003),
    [anon_sym_in] = ACTIONS(2003),
    [anon_sym_AMP] = ACTIONS(2003),
    [anon_sym_POUND] = ACTIONS(2003),
    [anon_sym_COLON_COLON] = ACTIONS(2005),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2005),
    [anon_sym_else] = ACTIONS(2003),
    [anon_sym_new] = ACTIONS(2003),
    [anon_sym_LBRACE_LT] = ACTIONS(2005),
    [anon_sym_begin] = ACTIONS(2003),
    [sym_ocamlyacc_value] = ACTIONS(2005),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2003),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2005),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2005),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2005),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2003),
    [anon_sym_SQUOTE] = ACTIONS(2005),
    [anon_sym_DQUOTE] = ACTIONS(2005),
    [sym_prefix_operator] = ACTIONS(2005),
    [aux_sym_sign_operator_token1] = ACTIONS(2003),
    [sym_hash_operator] = ACTIONS(2005),
    [anon_sym_lsl] = ACTIONS(2003),
    [anon_sym_lsr] = ACTIONS(2003),
    [anon_sym_asr] = ACTIONS(2003),
    [aux_sym_pow_operator_token1] = ACTIONS(2005),
    [anon_sym_mod] = ACTIONS(2003),
    [anon_sym_land] = ACTIONS(2003),
    [anon_sym_lor] = ACTIONS(2003),
    [anon_sym_lxor] = ACTIONS(2003),
    [aux_sym_mult_operator_token1] = ACTIONS(2003),
    [aux_sym_add_operator_token1] = ACTIONS(2003),
    [sym_concat_operator] = ACTIONS(2005),
    [sym_rel_operator] = ACTIONS(2003),
    [anon_sym_AMP_AMP] = ACTIONS(2003),
    [anon_sym_or] = ACTIONS(2003),
    [anon_sym_PIPE_PIPE] = ACTIONS(2003),
    [sym_assign_operator] = ACTIONS(2005),
    [sym_let_operator] = ACTIONS(2005),
    [sym_let_and_operator] = ACTIONS(2005),
    [sym__uppercase_identifier] = ACTIONS(2005),
    [aux_sym_directive_token1] = ACTIONS(2003),
    [aux_sym_tag_token1] = ACTIONS(2005),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1289)] = {
    [sym_attribute] = STATE(1289),
    [sym__lowercase_identifier] = ACTIONS(1899),
    [anon_sym_SEMI_SEMI] = ACTIONS(1901),
    [anon_sym_let] = ACTIONS(1899),
    [anon_sym_and] = ACTIONS(1899),
    [anon_sym_TILDE] = ACTIONS(1899),
    [anon_sym_QMARK] = ACTIONS(1899),
    [anon_sym_LPAREN] = ACTIONS(1901),
    [anon_sym_external] = ACTIONS(1899),
    [anon_sym_type] = ACTIONS(1899),
    [anon_sym_COMMA] = ACTIONS(1901),
    [anon_sym_PIPE] = ACTIONS(1899),
    [anon_sym_DOT] = ACTIONS(1901),
    [anon_sym_LBRACK] = ACTIONS(1899),
    [anon_sym_RBRACK] = ACTIONS(1901),
    [anon_sym_true] = ACTIONS(1899),
    [anon_sym_false] = ACTIONS(1899),
    [anon_sym_LBRACE] = ACTIONS(1899),
    [anon_sym_SEMI] = ACTIONS(1899),
    [anon_sym_exception] = ACTIONS(1899),
    [anon_sym_module] = ACTIONS(1899),
    [anon_sym_open] = ACTIONS(1899),
    [anon_sym_include] = ACTIONS(1899),
    [anon_sym_class] = ACTIONS(1899),
    [anon_sym_val] = ACTIONS(1899),
    [anon_sym_end] = ACTIONS(1899),
    [anon_sym_object] = ACTIONS(1899),
    [anon_sym_AMP] = ACTIONS(1899),
    [anon_sym_POUND] = ACTIONS(1899),
    [anon_sym_COLON_COLON] = ACTIONS(1901),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1901),
    [anon_sym_LT_DASH] = ACTIONS(1899),
    [anon_sym_else] = ACTIONS(1899),
    [anon_sym_new] = ACTIONS(1899),
    [anon_sym_LBRACE_LT] = ACTIONS(1901),
    [anon_sym_begin] = ACTIONS(1899),
    [sym_ocamlyacc_value] = ACTIONS(1901),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1899),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1901),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1899),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1899),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1901),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1901),
    [aux_sym_number_token1] = ACTIONS(1901),
    [aux_sym_signed_number_token1] = ACTIONS(1899),
    [anon_sym_SQUOTE] = ACTIONS(1901),
    [anon_sym_DQUOTE] = ACTIONS(1901),
    [sym_prefix_operator] = ACTIONS(1901),
    [aux_sym_sign_operator_token1] = ACTIONS(1899),
    [sym_hash_operator] = ACTIONS(1901),
    [anon_sym_lsl] = ACTIONS(1899),
    [anon_sym_lsr] = ACTIONS(1899),
    [anon_sym_asr] = ACTIONS(1899),
    [aux_sym_pow_operator_token1] = ACTIONS(1901),
    [anon_sym_mod] = ACTIONS(1899),
    [anon_sym_land] = ACTIONS(1899),
    [anon_sym_lor] = ACTIONS(1899),
    [anon_sym_lxor] = ACTIONS(1899),
    [aux_sym_mult_operator_token1] = ACTIONS(1899),
    [aux_sym_add_operator_token1] = ACTIONS(1899),
    [sym_concat_operator] = ACTIONS(1901),
    [sym_rel_operator] = ACTIONS(1899),
    [anon_sym_AMP_AMP] = ACTIONS(1899),
    [anon_sym_or] = ACTIONS(1899),
    [anon_sym_PIPE_PIPE] = ACTIONS(1899),
    [sym_assign_operator] = ACTIONS(1901),
    [sym_let_operator] = ACTIONS(1901),
    [sym__uppercase_identifier] = ACTIONS(1901),
    [aux_sym_directive_token1] = ACTIONS(1899),
    [aux_sym_tag_token1] = ACTIONS(1901),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1290)] = {
    [sym_attribute] = STATE(1290),
    [sym__lowercase_identifier] = ACTIONS(1903),
    [anon_sym_SEMI_SEMI] = ACTIONS(1905),
    [anon_sym_let] = ACTIONS(1903),
    [anon_sym_and] = ACTIONS(1903),
    [anon_sym_TILDE] = ACTIONS(1903),
    [anon_sym_QMARK] = ACTIONS(1903),
    [anon_sym_LPAREN] = ACTIONS(1905),
    [anon_sym_external] = ACTIONS(1903),
    [anon_sym_type] = ACTIONS(1903),
    [anon_sym_COMMA] = ACTIONS(1905),
    [anon_sym_PIPE] = ACTIONS(1903),
    [anon_sym_DOT] = ACTIONS(1905),
    [anon_sym_LBRACK] = ACTIONS(1903),
    [anon_sym_RBRACK] = ACTIONS(1905),
    [anon_sym_true] = ACTIONS(1903),
    [anon_sym_false] = ACTIONS(1903),
    [anon_sym_LBRACE] = ACTIONS(1903),
    [anon_sym_SEMI] = ACTIONS(1903),
    [anon_sym_exception] = ACTIONS(1903),
    [anon_sym_module] = ACTIONS(1903),
    [anon_sym_open] = ACTIONS(1903),
    [anon_sym_include] = ACTIONS(1903),
    [anon_sym_class] = ACTIONS(1903),
    [anon_sym_val] = ACTIONS(1903),
    [anon_sym_end] = ACTIONS(1903),
    [anon_sym_object] = ACTIONS(1903),
    [anon_sym_AMP] = ACTIONS(1903),
    [anon_sym_POUND] = ACTIONS(1903),
    [anon_sym_COLON_COLON] = ACTIONS(1905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1905),
    [anon_sym_LT_DASH] = ACTIONS(1903),
    [anon_sym_else] = ACTIONS(1903),
    [anon_sym_new] = ACTIONS(1903),
    [anon_sym_LBRACE_LT] = ACTIONS(1905),
    [anon_sym_begin] = ACTIONS(1903),
    [sym_ocamlyacc_value] = ACTIONS(1905),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1903),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1905),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1903),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1903),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1905),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1905),
    [aux_sym_number_token1] = ACTIONS(1905),
    [aux_sym_signed_number_token1] = ACTIONS(1903),
    [anon_sym_SQUOTE] = ACTIONS(1905),
    [anon_sym_DQUOTE] = ACTIONS(1905),
    [sym_prefix_operator] = ACTIONS(1905),
    [aux_sym_sign_operator_token1] = ACTIONS(1903),
    [sym_hash_operator] = ACTIONS(1905),
    [anon_sym_lsl] = ACTIONS(1903),
    [anon_sym_lsr] = ACTIONS(1903),
    [anon_sym_asr] = ACTIONS(1903),
    [aux_sym_pow_operator_token1] = ACTIONS(1905),
    [anon_sym_mod] = ACTIONS(1903),
    [anon_sym_land] = ACTIONS(1903),
    [anon_sym_lor] = ACTIONS(1903),
    [anon_sym_lxor] = ACTIONS(1903),
    [aux_sym_mult_operator_token1] = ACTIONS(1903),
    [aux_sym_add_operator_token1] = ACTIONS(1903),
    [sym_concat_operator] = ACTIONS(1905),
    [sym_rel_operator] = ACTIONS(1903),
    [anon_sym_AMP_AMP] = ACTIONS(1903),
    [anon_sym_or] = ACTIONS(1903),
    [anon_sym_PIPE_PIPE] = ACTIONS(1903),
    [sym_assign_operator] = ACTIONS(1905),
    [sym_let_operator] = ACTIONS(1905),
    [sym__uppercase_identifier] = ACTIONS(1905),
    [aux_sym_directive_token1] = ACTIONS(1903),
    [aux_sym_tag_token1] = ACTIONS(1905),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1291)] = {
    [sym_attribute] = STATE(1291),
    [sym__lowercase_identifier] = ACTIONS(1907),
    [anon_sym_SEMI_SEMI] = ACTIONS(1909),
    [anon_sym_let] = ACTIONS(1907),
    [anon_sym_and] = ACTIONS(1907),
    [anon_sym_TILDE] = ACTIONS(1907),
    [anon_sym_QMARK] = ACTIONS(1907),
    [anon_sym_LPAREN] = ACTIONS(1909),
    [anon_sym_external] = ACTIONS(1907),
    [anon_sym_type] = ACTIONS(1907),
    [anon_sym_COMMA] = ACTIONS(1909),
    [anon_sym_PIPE] = ACTIONS(1907),
    [anon_sym_DOT] = ACTIONS(1909),
    [anon_sym_LBRACK] = ACTIONS(1907),
    [anon_sym_RBRACK] = ACTIONS(1909),
    [anon_sym_true] = ACTIONS(1907),
    [anon_sym_false] = ACTIONS(1907),
    [anon_sym_LBRACE] = ACTIONS(1907),
    [anon_sym_SEMI] = ACTIONS(1907),
    [anon_sym_exception] = ACTIONS(1907),
    [anon_sym_module] = ACTIONS(1907),
    [anon_sym_open] = ACTIONS(1907),
    [anon_sym_include] = ACTIONS(1907),
    [anon_sym_class] = ACTIONS(1907),
    [anon_sym_val] = ACTIONS(1907),
    [anon_sym_end] = ACTIONS(1907),
    [anon_sym_object] = ACTIONS(1907),
    [anon_sym_AMP] = ACTIONS(1907),
    [anon_sym_POUND] = ACTIONS(1907),
    [anon_sym_COLON_COLON] = ACTIONS(1909),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1909),
    [anon_sym_LT_DASH] = ACTIONS(1907),
    [anon_sym_else] = ACTIONS(1907),
    [anon_sym_new] = ACTIONS(1907),
    [anon_sym_LBRACE_LT] = ACTIONS(1909),
    [anon_sym_begin] = ACTIONS(1907),
    [sym_ocamlyacc_value] = ACTIONS(1909),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1907),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1909),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1907),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1907),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1909),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1909),
    [aux_sym_number_token1] = ACTIONS(1909),
    [aux_sym_signed_number_token1] = ACTIONS(1907),
    [anon_sym_SQUOTE] = ACTIONS(1909),
    [anon_sym_DQUOTE] = ACTIONS(1909),
    [sym_prefix_operator] = ACTIONS(1909),
    [aux_sym_sign_operator_token1] = ACTIONS(1907),
    [sym_hash_operator] = ACTIONS(1909),
    [anon_sym_lsl] = ACTIONS(1907),
    [anon_sym_lsr] = ACTIONS(1907),
    [anon_sym_asr] = ACTIONS(1907),
    [aux_sym_pow_operator_token1] = ACTIONS(1909),
    [anon_sym_mod] = ACTIONS(1907),
    [anon_sym_land] = ACTIONS(1907),
    [anon_sym_lor] = ACTIONS(1907),
    [anon_sym_lxor] = ACTIONS(1907),
    [aux_sym_mult_operator_token1] = ACTIONS(1907),
    [aux_sym_add_operator_token1] = ACTIONS(1907),
    [sym_concat_operator] = ACTIONS(1909),
    [sym_rel_operator] = ACTIONS(1907),
    [anon_sym_AMP_AMP] = ACTIONS(1907),
    [anon_sym_or] = ACTIONS(1907),
    [anon_sym_PIPE_PIPE] = ACTIONS(1907),
    [sym_assign_operator] = ACTIONS(1909),
    [sym_let_operator] = ACTIONS(1909),
    [sym__uppercase_identifier] = ACTIONS(1909),
    [aux_sym_directive_token1] = ACTIONS(1907),
    [aux_sym_tag_token1] = ACTIONS(1909),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1292)] = {
    [sym_attribute] = STATE(1292),
    [sym__lowercase_identifier] = ACTIONS(1911),
    [anon_sym_SEMI_SEMI] = ACTIONS(1913),
    [anon_sym_let] = ACTIONS(1911),
    [anon_sym_and] = ACTIONS(1911),
    [anon_sym_TILDE] = ACTIONS(1911),
    [anon_sym_QMARK] = ACTIONS(1911),
    [anon_sym_LPAREN] = ACTIONS(1913),
    [anon_sym_external] = ACTIONS(1911),
    [anon_sym_type] = ACTIONS(1911),
    [anon_sym_COMMA] = ACTIONS(1913),
    [anon_sym_PIPE] = ACTIONS(1911),
    [anon_sym_DOT] = ACTIONS(1913),
    [anon_sym_LBRACK] = ACTIONS(1911),
    [anon_sym_RBRACK] = ACTIONS(1913),
    [anon_sym_true] = ACTIONS(1911),
    [anon_sym_false] = ACTIONS(1911),
    [anon_sym_LBRACE] = ACTIONS(1911),
    [anon_sym_SEMI] = ACTIONS(1911),
    [anon_sym_exception] = ACTIONS(1911),
    [anon_sym_module] = ACTIONS(1911),
    [anon_sym_open] = ACTIONS(1911),
    [anon_sym_include] = ACTIONS(1911),
    [anon_sym_class] = ACTIONS(1911),
    [anon_sym_val] = ACTIONS(1911),
    [anon_sym_end] = ACTIONS(1911),
    [anon_sym_object] = ACTIONS(1911),
    [anon_sym_AMP] = ACTIONS(1911),
    [anon_sym_POUND] = ACTIONS(1911),
    [anon_sym_COLON_COLON] = ACTIONS(1913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1913),
    [anon_sym_LT_DASH] = ACTIONS(1911),
    [anon_sym_else] = ACTIONS(1911),
    [anon_sym_new] = ACTIONS(1911),
    [anon_sym_LBRACE_LT] = ACTIONS(1913),
    [anon_sym_begin] = ACTIONS(1911),
    [sym_ocamlyacc_value] = ACTIONS(1913),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1911),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1913),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1911),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1911),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1913),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1913),
    [aux_sym_number_token1] = ACTIONS(1913),
    [aux_sym_signed_number_token1] = ACTIONS(1911),
    [anon_sym_SQUOTE] = ACTIONS(1913),
    [anon_sym_DQUOTE] = ACTIONS(1913),
    [sym_prefix_operator] = ACTIONS(1913),
    [aux_sym_sign_operator_token1] = ACTIONS(1911),
    [sym_hash_operator] = ACTIONS(1913),
    [anon_sym_lsl] = ACTIONS(1911),
    [anon_sym_lsr] = ACTIONS(1911),
    [anon_sym_asr] = ACTIONS(1911),
    [aux_sym_pow_operator_token1] = ACTIONS(1913),
    [anon_sym_mod] = ACTIONS(1911),
    [anon_sym_land] = ACTIONS(1911),
    [anon_sym_lor] = ACTIONS(1911),
    [anon_sym_lxor] = ACTIONS(1911),
    [aux_sym_mult_operator_token1] = ACTIONS(1911),
    [aux_sym_add_operator_token1] = ACTIONS(1911),
    [sym_concat_operator] = ACTIONS(1913),
    [sym_rel_operator] = ACTIONS(1911),
    [anon_sym_AMP_AMP] = ACTIONS(1911),
    [anon_sym_or] = ACTIONS(1911),
    [anon_sym_PIPE_PIPE] = ACTIONS(1911),
    [sym_assign_operator] = ACTIONS(1913),
    [sym_let_operator] = ACTIONS(1913),
    [sym__uppercase_identifier] = ACTIONS(1913),
    [aux_sym_directive_token1] = ACTIONS(1911),
    [aux_sym_tag_token1] = ACTIONS(1913),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1293)] = {
    [sym_attribute] = STATE(1293),
    [sym__lowercase_identifier] = ACTIONS(1915),
    [anon_sym_SEMI_SEMI] = ACTIONS(1917),
    [anon_sym_let] = ACTIONS(1915),
    [anon_sym_and] = ACTIONS(1915),
    [anon_sym_TILDE] = ACTIONS(1915),
    [anon_sym_QMARK] = ACTIONS(1915),
    [anon_sym_LPAREN] = ACTIONS(1917),
    [anon_sym_external] = ACTIONS(1915),
    [anon_sym_type] = ACTIONS(1915),
    [anon_sym_COMMA] = ACTIONS(1917),
    [anon_sym_PIPE] = ACTIONS(1915),
    [anon_sym_DOT] = ACTIONS(1917),
    [anon_sym_LBRACK] = ACTIONS(1915),
    [anon_sym_RBRACK] = ACTIONS(1917),
    [anon_sym_true] = ACTIONS(1915),
    [anon_sym_false] = ACTIONS(1915),
    [anon_sym_LBRACE] = ACTIONS(1915),
    [anon_sym_SEMI] = ACTIONS(1915),
    [anon_sym_exception] = ACTIONS(1915),
    [anon_sym_module] = ACTIONS(1915),
    [anon_sym_open] = ACTIONS(1915),
    [anon_sym_include] = ACTIONS(1915),
    [anon_sym_class] = ACTIONS(1915),
    [anon_sym_val] = ACTIONS(1915),
    [anon_sym_end] = ACTIONS(1915),
    [anon_sym_object] = ACTIONS(1915),
    [anon_sym_AMP] = ACTIONS(1915),
    [anon_sym_POUND] = ACTIONS(1915),
    [anon_sym_COLON_COLON] = ACTIONS(1917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1917),
    [anon_sym_LT_DASH] = ACTIONS(1915),
    [anon_sym_else] = ACTIONS(1915),
    [anon_sym_new] = ACTIONS(1915),
    [anon_sym_LBRACE_LT] = ACTIONS(1917),
    [anon_sym_begin] = ACTIONS(1915),
    [sym_ocamlyacc_value] = ACTIONS(1917),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1915),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1917),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1915),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1915),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1917),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1917),
    [aux_sym_number_token1] = ACTIONS(1917),
    [aux_sym_signed_number_token1] = ACTIONS(1915),
    [anon_sym_SQUOTE] = ACTIONS(1917),
    [anon_sym_DQUOTE] = ACTIONS(1917),
    [sym_prefix_operator] = ACTIONS(1917),
    [aux_sym_sign_operator_token1] = ACTIONS(1915),
    [sym_hash_operator] = ACTIONS(1917),
    [anon_sym_lsl] = ACTIONS(1915),
    [anon_sym_lsr] = ACTIONS(1915),
    [anon_sym_asr] = ACTIONS(1915),
    [aux_sym_pow_operator_token1] = ACTIONS(1917),
    [anon_sym_mod] = ACTIONS(1915),
    [anon_sym_land] = ACTIONS(1915),
    [anon_sym_lor] = ACTIONS(1915),
    [anon_sym_lxor] = ACTIONS(1915),
    [aux_sym_mult_operator_token1] = ACTIONS(1915),
    [aux_sym_add_operator_token1] = ACTIONS(1915),
    [sym_concat_operator] = ACTIONS(1917),
    [sym_rel_operator] = ACTIONS(1915),
    [anon_sym_AMP_AMP] = ACTIONS(1915),
    [anon_sym_or] = ACTIONS(1915),
    [anon_sym_PIPE_PIPE] = ACTIONS(1915),
    [sym_assign_operator] = ACTIONS(1917),
    [sym_let_operator] = ACTIONS(1917),
    [sym__uppercase_identifier] = ACTIONS(1917),
    [aux_sym_directive_token1] = ACTIONS(1915),
    [aux_sym_tag_token1] = ACTIONS(1917),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1294)] = {
    [sym_attribute] = STATE(1294),
    [sym__lowercase_identifier] = ACTIONS(1919),
    [anon_sym_SEMI_SEMI] = ACTIONS(1921),
    [anon_sym_let] = ACTIONS(1919),
    [anon_sym_and] = ACTIONS(1919),
    [anon_sym_TILDE] = ACTIONS(1919),
    [anon_sym_QMARK] = ACTIONS(1919),
    [anon_sym_LPAREN] = ACTIONS(1921),
    [anon_sym_external] = ACTIONS(1919),
    [anon_sym_type] = ACTIONS(1919),
    [anon_sym_COMMA] = ACTIONS(1921),
    [anon_sym_PIPE] = ACTIONS(1919),
    [anon_sym_DOT] = ACTIONS(1921),
    [anon_sym_LBRACK] = ACTIONS(1919),
    [anon_sym_RBRACK] = ACTIONS(1921),
    [anon_sym_true] = ACTIONS(1919),
    [anon_sym_false] = ACTIONS(1919),
    [anon_sym_LBRACE] = ACTIONS(1919),
    [anon_sym_SEMI] = ACTIONS(1919),
    [anon_sym_exception] = ACTIONS(1919),
    [anon_sym_module] = ACTIONS(1919),
    [anon_sym_open] = ACTIONS(1919),
    [anon_sym_include] = ACTIONS(1919),
    [anon_sym_class] = ACTIONS(1919),
    [anon_sym_val] = ACTIONS(1919),
    [anon_sym_end] = ACTIONS(1919),
    [anon_sym_object] = ACTIONS(1919),
    [anon_sym_AMP] = ACTIONS(1919),
    [anon_sym_POUND] = ACTIONS(1919),
    [anon_sym_COLON_COLON] = ACTIONS(1921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1921),
    [anon_sym_LT_DASH] = ACTIONS(1919),
    [anon_sym_else] = ACTIONS(1919),
    [anon_sym_new] = ACTIONS(1919),
    [anon_sym_LBRACE_LT] = ACTIONS(1921),
    [anon_sym_begin] = ACTIONS(1919),
    [sym_ocamlyacc_value] = ACTIONS(1921),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1919),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1921),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1919),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1919),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1921),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1921),
    [aux_sym_number_token1] = ACTIONS(1921),
    [aux_sym_signed_number_token1] = ACTIONS(1919),
    [anon_sym_SQUOTE] = ACTIONS(1921),
    [anon_sym_DQUOTE] = ACTIONS(1921),
    [sym_prefix_operator] = ACTIONS(1921),
    [aux_sym_sign_operator_token1] = ACTIONS(1919),
    [sym_hash_operator] = ACTIONS(1921),
    [anon_sym_lsl] = ACTIONS(1919),
    [anon_sym_lsr] = ACTIONS(1919),
    [anon_sym_asr] = ACTIONS(1919),
    [aux_sym_pow_operator_token1] = ACTIONS(1921),
    [anon_sym_mod] = ACTIONS(1919),
    [anon_sym_land] = ACTIONS(1919),
    [anon_sym_lor] = ACTIONS(1919),
    [anon_sym_lxor] = ACTIONS(1919),
    [aux_sym_mult_operator_token1] = ACTIONS(1919),
    [aux_sym_add_operator_token1] = ACTIONS(1919),
    [sym_concat_operator] = ACTIONS(1921),
    [sym_rel_operator] = ACTIONS(1919),
    [anon_sym_AMP_AMP] = ACTIONS(1919),
    [anon_sym_or] = ACTIONS(1919),
    [anon_sym_PIPE_PIPE] = ACTIONS(1919),
    [sym_assign_operator] = ACTIONS(1921),
    [sym_let_operator] = ACTIONS(1921),
    [sym__uppercase_identifier] = ACTIONS(1921),
    [aux_sym_directive_token1] = ACTIONS(1919),
    [aux_sym_tag_token1] = ACTIONS(1921),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1295)] = {
    [sym_attribute] = STATE(1295),
    [ts_builtin_sym_end] = ACTIONS(1949),
    [sym__lowercase_identifier] = ACTIONS(1947),
    [anon_sym_SEMI_SEMI] = ACTIONS(1949),
    [anon_sym_let] = ACTIONS(1947),
    [anon_sym_and] = ACTIONS(1947),
    [anon_sym_TILDE] = ACTIONS(1947),
    [anon_sym_QMARK] = ACTIONS(1947),
    [anon_sym_LPAREN] = ACTIONS(1949),
    [anon_sym_external] = ACTIONS(1947),
    [anon_sym_type] = ACTIONS(1947),
    [anon_sym_COMMA] = ACTIONS(1949),
    [anon_sym_PIPE] = ACTIONS(1947),
    [anon_sym_DOT] = ACTIONS(1949),
    [anon_sym_LBRACK] = ACTIONS(1947),
    [anon_sym_true] = ACTIONS(1947),
    [anon_sym_false] = ACTIONS(1947),
    [anon_sym_LBRACE] = ACTIONS(1947),
    [anon_sym_SEMI] = ACTIONS(1947),
    [anon_sym_exception] = ACTIONS(1947),
    [anon_sym_module] = ACTIONS(1947),
    [anon_sym_open] = ACTIONS(1947),
    [anon_sym_include] = ACTIONS(1947),
    [anon_sym_class] = ACTIONS(1947),
    [anon_sym_val] = ACTIONS(1947),
    [anon_sym_object] = ACTIONS(1947),
    [anon_sym_in] = ACTIONS(1947),
    [anon_sym_AMP] = ACTIONS(1947),
    [anon_sym_POUND] = ACTIONS(1947),
    [anon_sym_COLON_COLON] = ACTIONS(1949),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1949),
    [anon_sym_else] = ACTIONS(1947),
    [anon_sym_new] = ACTIONS(1947),
    [anon_sym_LBRACE_LT] = ACTIONS(1949),
    [anon_sym_begin] = ACTIONS(1947),
    [sym_ocamlyacc_value] = ACTIONS(1949),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1947),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1949),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1949),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1949),
    [aux_sym_number_token1] = ACTIONS(1949),
    [aux_sym_signed_number_token1] = ACTIONS(1947),
    [anon_sym_SQUOTE] = ACTIONS(1949),
    [anon_sym_DQUOTE] = ACTIONS(1949),
    [sym_prefix_operator] = ACTIONS(1949),
    [aux_sym_sign_operator_token1] = ACTIONS(1947),
    [sym_hash_operator] = ACTIONS(1949),
    [anon_sym_lsl] = ACTIONS(1947),
    [anon_sym_lsr] = ACTIONS(1947),
    [anon_sym_asr] = ACTIONS(1947),
    [aux_sym_pow_operator_token1] = ACTIONS(1949),
    [anon_sym_mod] = ACTIONS(1947),
    [anon_sym_land] = ACTIONS(1947),
    [anon_sym_lor] = ACTIONS(1947),
    [anon_sym_lxor] = ACTIONS(1947),
    [aux_sym_mult_operator_token1] = ACTIONS(1947),
    [aux_sym_add_operator_token1] = ACTIONS(1947),
    [sym_concat_operator] = ACTIONS(1949),
    [sym_rel_operator] = ACTIONS(1947),
    [anon_sym_AMP_AMP] = ACTIONS(1947),
    [anon_sym_or] = ACTIONS(1947),
    [anon_sym_PIPE_PIPE] = ACTIONS(1947),
    [sym_assign_operator] = ACTIONS(1949),
    [sym_let_operator] = ACTIONS(1949),
    [sym_let_and_operator] = ACTIONS(1949),
    [sym__uppercase_identifier] = ACTIONS(1949),
    [aux_sym_directive_token1] = ACTIONS(1947),
    [aux_sym_tag_token1] = ACTIONS(1949),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1296)] = {
    [sym_attribute] = STATE(1296),
    [ts_builtin_sym_end] = ACTIONS(2041),
    [sym__lowercase_identifier] = ACTIONS(2039),
    [anon_sym_SEMI_SEMI] = ACTIONS(2041),
    [anon_sym_let] = ACTIONS(2039),
    [anon_sym_and] = ACTIONS(2039),
    [anon_sym_TILDE] = ACTIONS(2039),
    [anon_sym_QMARK] = ACTIONS(2039),
    [anon_sym_LPAREN] = ACTIONS(2041),
    [anon_sym_external] = ACTIONS(2039),
    [anon_sym_type] = ACTIONS(2039),
    [anon_sym_COMMA] = ACTIONS(2041),
    [anon_sym_PIPE] = ACTIONS(2039),
    [anon_sym_DOT] = ACTIONS(1017),
    [anon_sym_LBRACK] = ACTIONS(2039),
    [anon_sym_true] = ACTIONS(2039),
    [anon_sym_false] = ACTIONS(2039),
    [anon_sym_LBRACE] = ACTIONS(2039),
    [anon_sym_SEMI] = ACTIONS(2039),
    [anon_sym_exception] = ACTIONS(2039),
    [anon_sym_module] = ACTIONS(2039),
    [anon_sym_open] = ACTIONS(2039),
    [anon_sym_include] = ACTIONS(2039),
    [anon_sym_class] = ACTIONS(2039),
    [anon_sym_val] = ACTIONS(2039),
    [anon_sym_object] = ACTIONS(2039),
    [anon_sym_in] = ACTIONS(2039),
    [anon_sym_AMP] = ACTIONS(2039),
    [anon_sym_POUND] = ACTIONS(1019),
    [anon_sym_COLON_COLON] = ACTIONS(2041),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2041),
    [anon_sym_else] = ACTIONS(2039),
    [anon_sym_new] = ACTIONS(2039),
    [anon_sym_LBRACE_LT] = ACTIONS(2041),
    [anon_sym_begin] = ACTIONS(2039),
    [sym_ocamlyacc_value] = ACTIONS(2041),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2039),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2041),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2039),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2039),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2041),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2041),
    [aux_sym_number_token1] = ACTIONS(2041),
    [aux_sym_signed_number_token1] = ACTIONS(2039),
    [anon_sym_SQUOTE] = ACTIONS(2041),
    [anon_sym_DQUOTE] = ACTIONS(2041),
    [sym_prefix_operator] = ACTIONS(2041),
    [aux_sym_sign_operator_token1] = ACTIONS(2039),
    [sym_hash_operator] = ACTIONS(1021),
    [anon_sym_lsl] = ACTIONS(2039),
    [anon_sym_lsr] = ACTIONS(2039),
    [anon_sym_asr] = ACTIONS(2039),
    [aux_sym_pow_operator_token1] = ACTIONS(2041),
    [anon_sym_mod] = ACTIONS(2039),
    [anon_sym_land] = ACTIONS(2039),
    [anon_sym_lor] = ACTIONS(2039),
    [anon_sym_lxor] = ACTIONS(2039),
    [aux_sym_mult_operator_token1] = ACTIONS(2039),
    [aux_sym_add_operator_token1] = ACTIONS(2039),
    [sym_concat_operator] = ACTIONS(2041),
    [sym_rel_operator] = ACTIONS(2039),
    [anon_sym_AMP_AMP] = ACTIONS(2039),
    [anon_sym_or] = ACTIONS(2039),
    [anon_sym_PIPE_PIPE] = ACTIONS(2039),
    [sym_assign_operator] = ACTIONS(2041),
    [sym_let_operator] = ACTIONS(2041),
    [sym_let_and_operator] = ACTIONS(2041),
    [sym__uppercase_identifier] = ACTIONS(2041),
    [aux_sym_directive_token1] = ACTIONS(2039),
    [aux_sym_tag_token1] = ACTIONS(2041),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1297)] = {
    [sym__parenthesized_abstract_type] = STATE(1710),
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5287),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1297),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2558),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_PIPE] = ACTIONS(1641),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_RBRACK] = ACTIONS(1641),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_SEMI] = ACTIONS(1641),
    [anon_sym_RBRACE] = ACTIONS(1641),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_as] = ACTIONS(1645),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1641),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1298)] = {
    [sym_attribute] = STATE(1298),
    [ts_builtin_sym_end] = ACTIONS(1997),
    [sym__lowercase_identifier] = ACTIONS(1995),
    [anon_sym_SEMI_SEMI] = ACTIONS(1997),
    [anon_sym_let] = ACTIONS(1995),
    [anon_sym_and] = ACTIONS(1995),
    [anon_sym_TILDE] = ACTIONS(1995),
    [anon_sym_QMARK] = ACTIONS(1995),
    [anon_sym_LPAREN] = ACTIONS(1997),
    [anon_sym_external] = ACTIONS(1995),
    [anon_sym_type] = ACTIONS(1995),
    [anon_sym_COMMA] = ACTIONS(1997),
    [anon_sym_PIPE] = ACTIONS(1995),
    [anon_sym_DOT] = ACTIONS(1997),
    [anon_sym_LBRACK] = ACTIONS(1995),
    [anon_sym_true] = ACTIONS(1995),
    [anon_sym_false] = ACTIONS(1995),
    [anon_sym_LBRACE] = ACTIONS(1995),
    [anon_sym_SEMI] = ACTIONS(1995),
    [anon_sym_exception] = ACTIONS(1995),
    [anon_sym_module] = ACTIONS(1995),
    [anon_sym_open] = ACTIONS(1995),
    [anon_sym_include] = ACTIONS(1995),
    [anon_sym_class] = ACTIONS(1995),
    [anon_sym_val] = ACTIONS(1995),
    [anon_sym_object] = ACTIONS(1995),
    [anon_sym_in] = ACTIONS(1995),
    [anon_sym_AMP] = ACTIONS(1995),
    [anon_sym_POUND] = ACTIONS(1995),
    [anon_sym_COLON_COLON] = ACTIONS(1997),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1997),
    [anon_sym_else] = ACTIONS(1995),
    [anon_sym_new] = ACTIONS(1995),
    [anon_sym_LBRACE_LT] = ACTIONS(1997),
    [anon_sym_begin] = ACTIONS(1995),
    [sym_ocamlyacc_value] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1995),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1997),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1995),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1995),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1997),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1997),
    [aux_sym_number_token1] = ACTIONS(1997),
    [aux_sym_signed_number_token1] = ACTIONS(1995),
    [anon_sym_SQUOTE] = ACTIONS(1997),
    [anon_sym_DQUOTE] = ACTIONS(1997),
    [sym_prefix_operator] = ACTIONS(1997),
    [aux_sym_sign_operator_token1] = ACTIONS(1995),
    [sym_hash_operator] = ACTIONS(1997),
    [anon_sym_lsl] = ACTIONS(1995),
    [anon_sym_lsr] = ACTIONS(1995),
    [anon_sym_asr] = ACTIONS(1995),
    [aux_sym_pow_operator_token1] = ACTIONS(1997),
    [anon_sym_mod] = ACTIONS(1995),
    [anon_sym_land] = ACTIONS(1995),
    [anon_sym_lor] = ACTIONS(1995),
    [anon_sym_lxor] = ACTIONS(1995),
    [aux_sym_mult_operator_token1] = ACTIONS(1995),
    [aux_sym_add_operator_token1] = ACTIONS(1995),
    [sym_concat_operator] = ACTIONS(1997),
    [sym_rel_operator] = ACTIONS(1995),
    [anon_sym_AMP_AMP] = ACTIONS(1995),
    [anon_sym_or] = ACTIONS(1995),
    [anon_sym_PIPE_PIPE] = ACTIONS(1995),
    [sym_assign_operator] = ACTIONS(1997),
    [sym_let_operator] = ACTIONS(1997),
    [sym_let_and_operator] = ACTIONS(1997),
    [sym__uppercase_identifier] = ACTIONS(1997),
    [aux_sym_directive_token1] = ACTIONS(1995),
    [aux_sym_tag_token1] = ACTIONS(1997),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1299)] = {
    [sym_attribute] = STATE(1299),
    [ts_builtin_sym_end] = ACTIONS(1933),
    [sym__lowercase_identifier] = ACTIONS(1931),
    [anon_sym_SEMI_SEMI] = ACTIONS(1933),
    [anon_sym_let] = ACTIONS(1931),
    [anon_sym_and] = ACTIONS(1931),
    [anon_sym_TILDE] = ACTIONS(1931),
    [anon_sym_QMARK] = ACTIONS(1931),
    [anon_sym_LPAREN] = ACTIONS(1933),
    [anon_sym_external] = ACTIONS(1931),
    [anon_sym_type] = ACTIONS(1931),
    [anon_sym_COMMA] = ACTIONS(1933),
    [anon_sym_PIPE] = ACTIONS(1931),
    [anon_sym_DOT] = ACTIONS(1017),
    [anon_sym_LBRACK] = ACTIONS(1931),
    [anon_sym_true] = ACTIONS(1931),
    [anon_sym_false] = ACTIONS(1931),
    [anon_sym_LBRACE] = ACTIONS(1931),
    [anon_sym_SEMI] = ACTIONS(1931),
    [anon_sym_exception] = ACTIONS(1931),
    [anon_sym_module] = ACTIONS(1931),
    [anon_sym_open] = ACTIONS(1931),
    [anon_sym_include] = ACTIONS(1931),
    [anon_sym_class] = ACTIONS(1931),
    [anon_sym_val] = ACTIONS(1931),
    [anon_sym_object] = ACTIONS(1931),
    [anon_sym_in] = ACTIONS(1931),
    [anon_sym_AMP] = ACTIONS(1931),
    [anon_sym_POUND] = ACTIONS(1019),
    [anon_sym_COLON_COLON] = ACTIONS(1933),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1933),
    [anon_sym_else] = ACTIONS(1931),
    [anon_sym_new] = ACTIONS(1931),
    [anon_sym_LBRACE_LT] = ACTIONS(1933),
    [anon_sym_begin] = ACTIONS(1931),
    [sym_ocamlyacc_value] = ACTIONS(1933),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1931),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1933),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1933),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1933),
    [aux_sym_number_token1] = ACTIONS(1933),
    [aux_sym_signed_number_token1] = ACTIONS(1931),
    [anon_sym_SQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1933),
    [sym_prefix_operator] = ACTIONS(1933),
    [aux_sym_sign_operator_token1] = ACTIONS(1931),
    [sym_hash_operator] = ACTIONS(1021),
    [anon_sym_lsl] = ACTIONS(1931),
    [anon_sym_lsr] = ACTIONS(1931),
    [anon_sym_asr] = ACTIONS(1931),
    [aux_sym_pow_operator_token1] = ACTIONS(1933),
    [anon_sym_mod] = ACTIONS(1931),
    [anon_sym_land] = ACTIONS(1931),
    [anon_sym_lor] = ACTIONS(1931),
    [anon_sym_lxor] = ACTIONS(1931),
    [aux_sym_mult_operator_token1] = ACTIONS(1931),
    [aux_sym_add_operator_token1] = ACTIONS(1931),
    [sym_concat_operator] = ACTIONS(1933),
    [sym_rel_operator] = ACTIONS(1931),
    [anon_sym_AMP_AMP] = ACTIONS(1931),
    [anon_sym_or] = ACTIONS(1931),
    [anon_sym_PIPE_PIPE] = ACTIONS(1931),
    [sym_assign_operator] = ACTIONS(1933),
    [sym_let_operator] = ACTIONS(1933),
    [sym_let_and_operator] = ACTIONS(1933),
    [sym__uppercase_identifier] = ACTIONS(1933),
    [aux_sym_directive_token1] = ACTIONS(1931),
    [aux_sym_tag_token1] = ACTIONS(1933),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1300)] = {
    [sym_attribute] = STATE(1300),
    [ts_builtin_sym_end] = ACTIONS(2045),
    [sym__lowercase_identifier] = ACTIONS(2043),
    [anon_sym_SEMI_SEMI] = ACTIONS(2045),
    [anon_sym_let] = ACTIONS(2043),
    [anon_sym_and] = ACTIONS(2043),
    [anon_sym_TILDE] = ACTIONS(2043),
    [anon_sym_QMARK] = ACTIONS(2043),
    [anon_sym_LPAREN] = ACTIONS(2045),
    [anon_sym_external] = ACTIONS(2043),
    [anon_sym_type] = ACTIONS(2043),
    [anon_sym_COMMA] = ACTIONS(2045),
    [anon_sym_PIPE] = ACTIONS(2043),
    [anon_sym_DOT] = ACTIONS(2045),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_true] = ACTIONS(2043),
    [anon_sym_false] = ACTIONS(2043),
    [anon_sym_LBRACE] = ACTIONS(2043),
    [anon_sym_SEMI] = ACTIONS(2043),
    [anon_sym_exception] = ACTIONS(2043),
    [anon_sym_module] = ACTIONS(2043),
    [anon_sym_open] = ACTIONS(2043),
    [anon_sym_include] = ACTIONS(2043),
    [anon_sym_class] = ACTIONS(2043),
    [anon_sym_val] = ACTIONS(2043),
    [anon_sym_object] = ACTIONS(2043),
    [anon_sym_in] = ACTIONS(2043),
    [anon_sym_AMP] = ACTIONS(2043),
    [anon_sym_POUND] = ACTIONS(2043),
    [anon_sym_COLON_COLON] = ACTIONS(2045),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2045),
    [anon_sym_else] = ACTIONS(2043),
    [anon_sym_new] = ACTIONS(2043),
    [anon_sym_LBRACE_LT] = ACTIONS(2045),
    [anon_sym_begin] = ACTIONS(2043),
    [sym_ocamlyacc_value] = ACTIONS(2045),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2043),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2045),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2043),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2043),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2045),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2045),
    [aux_sym_number_token1] = ACTIONS(2045),
    [aux_sym_signed_number_token1] = ACTIONS(2043),
    [anon_sym_SQUOTE] = ACTIONS(2045),
    [anon_sym_DQUOTE] = ACTIONS(2045),
    [sym_prefix_operator] = ACTIONS(2045),
    [aux_sym_sign_operator_token1] = ACTIONS(2043),
    [sym_hash_operator] = ACTIONS(2045),
    [anon_sym_lsl] = ACTIONS(2043),
    [anon_sym_lsr] = ACTIONS(2043),
    [anon_sym_asr] = ACTIONS(2043),
    [aux_sym_pow_operator_token1] = ACTIONS(2045),
    [anon_sym_mod] = ACTIONS(2043),
    [anon_sym_land] = ACTIONS(2043),
    [anon_sym_lor] = ACTIONS(2043),
    [anon_sym_lxor] = ACTIONS(2043),
    [aux_sym_mult_operator_token1] = ACTIONS(2043),
    [aux_sym_add_operator_token1] = ACTIONS(2043),
    [sym_concat_operator] = ACTIONS(2045),
    [sym_rel_operator] = ACTIONS(2043),
    [anon_sym_AMP_AMP] = ACTIONS(2043),
    [anon_sym_or] = ACTIONS(2043),
    [anon_sym_PIPE_PIPE] = ACTIONS(2043),
    [sym_assign_operator] = ACTIONS(2045),
    [sym_let_operator] = ACTIONS(2045),
    [sym_let_and_operator] = ACTIONS(2045),
    [sym__uppercase_identifier] = ACTIONS(2045),
    [aux_sym_directive_token1] = ACTIONS(2043),
    [aux_sym_tag_token1] = ACTIONS(2045),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1301)] = {
    [sym_attribute] = STATE(1301),
    [ts_builtin_sym_end] = ACTIONS(2049),
    [sym__lowercase_identifier] = ACTIONS(2047),
    [anon_sym_SEMI_SEMI] = ACTIONS(2049),
    [anon_sym_let] = ACTIONS(2047),
    [anon_sym_and] = ACTIONS(2047),
    [anon_sym_TILDE] = ACTIONS(2047),
    [anon_sym_QMARK] = ACTIONS(2047),
    [anon_sym_LPAREN] = ACTIONS(2049),
    [anon_sym_external] = ACTIONS(2047),
    [anon_sym_type] = ACTIONS(2047),
    [anon_sym_COMMA] = ACTIONS(2049),
    [anon_sym_PIPE] = ACTIONS(2047),
    [anon_sym_DOT] = ACTIONS(2049),
    [anon_sym_LBRACK] = ACTIONS(2047),
    [anon_sym_true] = ACTIONS(2047),
    [anon_sym_false] = ACTIONS(2047),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_SEMI] = ACTIONS(2047),
    [anon_sym_exception] = ACTIONS(2047),
    [anon_sym_module] = ACTIONS(2047),
    [anon_sym_open] = ACTIONS(2047),
    [anon_sym_include] = ACTIONS(2047),
    [anon_sym_class] = ACTIONS(2047),
    [anon_sym_val] = ACTIONS(2047),
    [anon_sym_object] = ACTIONS(2047),
    [anon_sym_in] = ACTIONS(2047),
    [anon_sym_AMP] = ACTIONS(2047),
    [anon_sym_POUND] = ACTIONS(2047),
    [anon_sym_COLON_COLON] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2049),
    [anon_sym_else] = ACTIONS(2047),
    [anon_sym_new] = ACTIONS(2047),
    [anon_sym_LBRACE_LT] = ACTIONS(2049),
    [anon_sym_begin] = ACTIONS(2047),
    [sym_ocamlyacc_value] = ACTIONS(2049),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2047),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2049),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2047),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2047),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2049),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2049),
    [aux_sym_number_token1] = ACTIONS(2049),
    [aux_sym_signed_number_token1] = ACTIONS(2047),
    [anon_sym_SQUOTE] = ACTIONS(2049),
    [anon_sym_DQUOTE] = ACTIONS(2049),
    [sym_prefix_operator] = ACTIONS(2049),
    [aux_sym_sign_operator_token1] = ACTIONS(2047),
    [sym_hash_operator] = ACTIONS(2049),
    [anon_sym_lsl] = ACTIONS(2047),
    [anon_sym_lsr] = ACTIONS(2047),
    [anon_sym_asr] = ACTIONS(2047),
    [aux_sym_pow_operator_token1] = ACTIONS(2049),
    [anon_sym_mod] = ACTIONS(2047),
    [anon_sym_land] = ACTIONS(2047),
    [anon_sym_lor] = ACTIONS(2047),
    [anon_sym_lxor] = ACTIONS(2047),
    [aux_sym_mult_operator_token1] = ACTIONS(2047),
    [aux_sym_add_operator_token1] = ACTIONS(2047),
    [sym_concat_operator] = ACTIONS(2049),
    [sym_rel_operator] = ACTIONS(2047),
    [anon_sym_AMP_AMP] = ACTIONS(2047),
    [anon_sym_or] = ACTIONS(2047),
    [anon_sym_PIPE_PIPE] = ACTIONS(2047),
    [sym_assign_operator] = ACTIONS(2049),
    [sym_let_operator] = ACTIONS(2049),
    [sym_let_and_operator] = ACTIONS(2049),
    [sym__uppercase_identifier] = ACTIONS(2049),
    [aux_sym_directive_token1] = ACTIONS(2047),
    [aux_sym_tag_token1] = ACTIONS(2049),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1302)] = {
    [sym_attribute] = STATE(1302),
    [ts_builtin_sym_end] = ACTIONS(2073),
    [sym__lowercase_identifier] = ACTIONS(2071),
    [anon_sym_SEMI_SEMI] = ACTIONS(2073),
    [anon_sym_let] = ACTIONS(2071),
    [anon_sym_and] = ACTIONS(2071),
    [anon_sym_TILDE] = ACTIONS(2071),
    [anon_sym_QMARK] = ACTIONS(2071),
    [anon_sym_LPAREN] = ACTIONS(2073),
    [anon_sym_external] = ACTIONS(2071),
    [anon_sym_type] = ACTIONS(2071),
    [anon_sym_COMMA] = ACTIONS(2073),
    [anon_sym_PIPE] = ACTIONS(2071),
    [anon_sym_DOT] = ACTIONS(2073),
    [anon_sym_LBRACK] = ACTIONS(2071),
    [anon_sym_true] = ACTIONS(2071),
    [anon_sym_false] = ACTIONS(2071),
    [anon_sym_LBRACE] = ACTIONS(2071),
    [anon_sym_SEMI] = ACTIONS(2071),
    [anon_sym_exception] = ACTIONS(2071),
    [anon_sym_module] = ACTIONS(2071),
    [anon_sym_open] = ACTIONS(2071),
    [anon_sym_include] = ACTIONS(2071),
    [anon_sym_class] = ACTIONS(2071),
    [anon_sym_val] = ACTIONS(2071),
    [anon_sym_object] = ACTIONS(2071),
    [anon_sym_in] = ACTIONS(2071),
    [anon_sym_AMP] = ACTIONS(2071),
    [anon_sym_POUND] = ACTIONS(2071),
    [anon_sym_COLON_COLON] = ACTIONS(2073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2073),
    [anon_sym_else] = ACTIONS(2071),
    [anon_sym_new] = ACTIONS(2071),
    [anon_sym_LBRACE_LT] = ACTIONS(2073),
    [anon_sym_begin] = ACTIONS(2071),
    [sym_ocamlyacc_value] = ACTIONS(2073),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2071),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2073),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2071),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2071),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2073),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2073),
    [aux_sym_number_token1] = ACTIONS(2073),
    [aux_sym_signed_number_token1] = ACTIONS(2071),
    [anon_sym_SQUOTE] = ACTIONS(2073),
    [anon_sym_DQUOTE] = ACTIONS(2073),
    [sym_prefix_operator] = ACTIONS(2073),
    [aux_sym_sign_operator_token1] = ACTIONS(2071),
    [sym_hash_operator] = ACTIONS(2073),
    [anon_sym_lsl] = ACTIONS(2071),
    [anon_sym_lsr] = ACTIONS(2071),
    [anon_sym_asr] = ACTIONS(2071),
    [aux_sym_pow_operator_token1] = ACTIONS(2073),
    [anon_sym_mod] = ACTIONS(2071),
    [anon_sym_land] = ACTIONS(2071),
    [anon_sym_lor] = ACTIONS(2071),
    [anon_sym_lxor] = ACTIONS(2071),
    [aux_sym_mult_operator_token1] = ACTIONS(2071),
    [aux_sym_add_operator_token1] = ACTIONS(2071),
    [sym_concat_operator] = ACTIONS(2073),
    [sym_rel_operator] = ACTIONS(2071),
    [anon_sym_AMP_AMP] = ACTIONS(2071),
    [anon_sym_or] = ACTIONS(2071),
    [anon_sym_PIPE_PIPE] = ACTIONS(2071),
    [sym_assign_operator] = ACTIONS(2073),
    [sym_let_operator] = ACTIONS(2073),
    [sym_let_and_operator] = ACTIONS(2073),
    [sym__uppercase_identifier] = ACTIONS(2073),
    [aux_sym_directive_token1] = ACTIONS(2071),
    [aux_sym_tag_token1] = ACTIONS(2073),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1303)] = {
    [sym_attribute] = STATE(1303),
    [ts_builtin_sym_end] = ACTIONS(2081),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_SEMI_SEMI] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2079),
    [anon_sym_and] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_external] = ACTIONS(2079),
    [anon_sym_type] = ACTIONS(2079),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_PIPE] = ACTIONS(2079),
    [anon_sym_DOT] = ACTIONS(2081),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_SEMI] = ACTIONS(2079),
    [anon_sym_exception] = ACTIONS(2079),
    [anon_sym_module] = ACTIONS(2079),
    [anon_sym_open] = ACTIONS(2079),
    [anon_sym_include] = ACTIONS(2079),
    [anon_sym_class] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_in] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_POUND] = ACTIONS(2079),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [sym_hash_operator] = ACTIONS(2081),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym_let_operator] = ACTIONS(2081),
    [sym_let_and_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_directive_token1] = ACTIONS(2079),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1304)] = {
    [sym_attribute] = STATE(1304),
    [ts_builtin_sym_end] = ACTIONS(2143),
    [sym__lowercase_identifier] = ACTIONS(2141),
    [anon_sym_SEMI_SEMI] = ACTIONS(2143),
    [anon_sym_let] = ACTIONS(2141),
    [anon_sym_and] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_QMARK] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2143),
    [anon_sym_external] = ACTIONS(2141),
    [anon_sym_type] = ACTIONS(2141),
    [anon_sym_COMMA] = ACTIONS(2143),
    [anon_sym_PIPE] = ACTIONS(2141),
    [anon_sym_DOT] = ACTIONS(2143),
    [anon_sym_LBRACK] = ACTIONS(2141),
    [anon_sym_true] = ACTIONS(2141),
    [anon_sym_false] = ACTIONS(2141),
    [anon_sym_LBRACE] = ACTIONS(2141),
    [anon_sym_SEMI] = ACTIONS(2141),
    [anon_sym_exception] = ACTIONS(2141),
    [anon_sym_module] = ACTIONS(2141),
    [anon_sym_open] = ACTIONS(2141),
    [anon_sym_include] = ACTIONS(2141),
    [anon_sym_class] = ACTIONS(2141),
    [anon_sym_val] = ACTIONS(2141),
    [anon_sym_object] = ACTIONS(2141),
    [anon_sym_in] = ACTIONS(2141),
    [anon_sym_AMP] = ACTIONS(2141),
    [anon_sym_POUND] = ACTIONS(2141),
    [anon_sym_COLON_COLON] = ACTIONS(2143),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2143),
    [anon_sym_else] = ACTIONS(2141),
    [anon_sym_new] = ACTIONS(2141),
    [anon_sym_LBRACE_LT] = ACTIONS(2143),
    [anon_sym_begin] = ACTIONS(2141),
    [sym_ocamlyacc_value] = ACTIONS(2143),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2141),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2143),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2141),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2141),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2143),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2143),
    [aux_sym_number_token1] = ACTIONS(2143),
    [aux_sym_signed_number_token1] = ACTIONS(2141),
    [anon_sym_SQUOTE] = ACTIONS(2143),
    [anon_sym_DQUOTE] = ACTIONS(2143),
    [sym_prefix_operator] = ACTIONS(2143),
    [aux_sym_sign_operator_token1] = ACTIONS(2141),
    [sym_hash_operator] = ACTIONS(2143),
    [anon_sym_lsl] = ACTIONS(2141),
    [anon_sym_lsr] = ACTIONS(2141),
    [anon_sym_asr] = ACTIONS(2141),
    [aux_sym_pow_operator_token1] = ACTIONS(2143),
    [anon_sym_mod] = ACTIONS(2141),
    [anon_sym_land] = ACTIONS(2141),
    [anon_sym_lor] = ACTIONS(2141),
    [anon_sym_lxor] = ACTIONS(2141),
    [aux_sym_mult_operator_token1] = ACTIONS(2141),
    [aux_sym_add_operator_token1] = ACTIONS(2141),
    [sym_concat_operator] = ACTIONS(2143),
    [sym_rel_operator] = ACTIONS(2141),
    [anon_sym_AMP_AMP] = ACTIONS(2141),
    [anon_sym_or] = ACTIONS(2141),
    [anon_sym_PIPE_PIPE] = ACTIONS(2141),
    [sym_assign_operator] = ACTIONS(2143),
    [sym_let_operator] = ACTIONS(2143),
    [sym_let_and_operator] = ACTIONS(2143),
    [sym__uppercase_identifier] = ACTIONS(2143),
    [aux_sym_directive_token1] = ACTIONS(2141),
    [aux_sym_tag_token1] = ACTIONS(2143),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1305)] = {
    [sym_attribute] = STATE(1305),
    [ts_builtin_sym_end] = ACTIONS(1929),
    [sym__lowercase_identifier] = ACTIONS(1927),
    [anon_sym_SEMI_SEMI] = ACTIONS(1929),
    [anon_sym_let] = ACTIONS(1927),
    [anon_sym_and] = ACTIONS(1927),
    [anon_sym_TILDE] = ACTIONS(1927),
    [anon_sym_QMARK] = ACTIONS(1927),
    [anon_sym_LPAREN] = ACTIONS(1929),
    [anon_sym_external] = ACTIONS(1927),
    [anon_sym_type] = ACTIONS(1927),
    [anon_sym_COMMA] = ACTIONS(1929),
    [anon_sym_PIPE] = ACTIONS(1927),
    [anon_sym_DOT] = ACTIONS(1929),
    [anon_sym_LBRACK] = ACTIONS(1927),
    [anon_sym_true] = ACTIONS(1927),
    [anon_sym_false] = ACTIONS(1927),
    [anon_sym_LBRACE] = ACTIONS(1927),
    [anon_sym_SEMI] = ACTIONS(1927),
    [anon_sym_exception] = ACTIONS(1927),
    [anon_sym_module] = ACTIONS(1927),
    [anon_sym_open] = ACTIONS(1927),
    [anon_sym_include] = ACTIONS(1927),
    [anon_sym_class] = ACTIONS(1927),
    [anon_sym_val] = ACTIONS(1927),
    [anon_sym_object] = ACTIONS(1927),
    [anon_sym_in] = ACTIONS(1927),
    [anon_sym_AMP] = ACTIONS(1927),
    [anon_sym_POUND] = ACTIONS(1927),
    [anon_sym_COLON_COLON] = ACTIONS(1929),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1929),
    [anon_sym_else] = ACTIONS(1927),
    [anon_sym_new] = ACTIONS(1927),
    [anon_sym_LBRACE_LT] = ACTIONS(1929),
    [anon_sym_begin] = ACTIONS(1927),
    [sym_ocamlyacc_value] = ACTIONS(1929),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1927),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1929),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1927),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1927),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1929),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1929),
    [aux_sym_number_token1] = ACTIONS(1929),
    [aux_sym_signed_number_token1] = ACTIONS(1927),
    [anon_sym_SQUOTE] = ACTIONS(1929),
    [anon_sym_DQUOTE] = ACTIONS(1929),
    [sym_prefix_operator] = ACTIONS(1929),
    [aux_sym_sign_operator_token1] = ACTIONS(1927),
    [sym_hash_operator] = ACTIONS(1929),
    [anon_sym_lsl] = ACTIONS(1927),
    [anon_sym_lsr] = ACTIONS(1927),
    [anon_sym_asr] = ACTIONS(1927),
    [aux_sym_pow_operator_token1] = ACTIONS(1929),
    [anon_sym_mod] = ACTIONS(1927),
    [anon_sym_land] = ACTIONS(1927),
    [anon_sym_lor] = ACTIONS(1927),
    [anon_sym_lxor] = ACTIONS(1927),
    [aux_sym_mult_operator_token1] = ACTIONS(1927),
    [aux_sym_add_operator_token1] = ACTIONS(1927),
    [sym_concat_operator] = ACTIONS(1929),
    [sym_rel_operator] = ACTIONS(1927),
    [anon_sym_AMP_AMP] = ACTIONS(1927),
    [anon_sym_or] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1927),
    [sym_assign_operator] = ACTIONS(1929),
    [sym_let_operator] = ACTIONS(1929),
    [sym_let_and_operator] = ACTIONS(1929),
    [sym__uppercase_identifier] = ACTIONS(1929),
    [aux_sym_directive_token1] = ACTIONS(1927),
    [aux_sym_tag_token1] = ACTIONS(1929),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1306)] = {
    [sym_attribute] = STATE(1306),
    [sym__lowercase_identifier] = ACTIONS(1891),
    [anon_sym_SEMI_SEMI] = ACTIONS(1893),
    [anon_sym_let] = ACTIONS(1891),
    [anon_sym_and] = ACTIONS(1891),
    [anon_sym_TILDE] = ACTIONS(1891),
    [anon_sym_QMARK] = ACTIONS(1891),
    [anon_sym_LPAREN] = ACTIONS(1893),
    [anon_sym_external] = ACTIONS(1891),
    [anon_sym_type] = ACTIONS(1891),
    [anon_sym_COMMA] = ACTIONS(1893),
    [anon_sym_PIPE] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACK] = ACTIONS(1891),
    [anon_sym_RBRACK] = ACTIONS(1893),
    [anon_sym_true] = ACTIONS(1891),
    [anon_sym_false] = ACTIONS(1891),
    [anon_sym_LBRACE] = ACTIONS(1891),
    [anon_sym_SEMI] = ACTIONS(1891),
    [anon_sym_exception] = ACTIONS(1891),
    [anon_sym_module] = ACTIONS(1891),
    [anon_sym_open] = ACTIONS(1891),
    [anon_sym_include] = ACTIONS(1891),
    [anon_sym_class] = ACTIONS(1891),
    [anon_sym_val] = ACTIONS(1891),
    [anon_sym_end] = ACTIONS(1891),
    [anon_sym_object] = ACTIONS(1891),
    [anon_sym_AMP] = ACTIONS(1891),
    [anon_sym_POUND] = ACTIONS(1891),
    [anon_sym_COLON_COLON] = ACTIONS(1893),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1893),
    [anon_sym_LT_DASH] = ACTIONS(1891),
    [anon_sym_else] = ACTIONS(1891),
    [anon_sym_new] = ACTIONS(1891),
    [anon_sym_LBRACE_LT] = ACTIONS(1893),
    [anon_sym_begin] = ACTIONS(1891),
    [sym_ocamlyacc_value] = ACTIONS(1893),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1891),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1893),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1891),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1891),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1893),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1893),
    [aux_sym_number_token1] = ACTIONS(1893),
    [aux_sym_signed_number_token1] = ACTIONS(1891),
    [anon_sym_SQUOTE] = ACTIONS(1893),
    [anon_sym_DQUOTE] = ACTIONS(1893),
    [sym_prefix_operator] = ACTIONS(1893),
    [aux_sym_sign_operator_token1] = ACTIONS(1891),
    [sym_hash_operator] = ACTIONS(1893),
    [anon_sym_lsl] = ACTIONS(1891),
    [anon_sym_lsr] = ACTIONS(1891),
    [anon_sym_asr] = ACTIONS(1891),
    [aux_sym_pow_operator_token1] = ACTIONS(1893),
    [anon_sym_mod] = ACTIONS(1891),
    [anon_sym_land] = ACTIONS(1891),
    [anon_sym_lor] = ACTIONS(1891),
    [anon_sym_lxor] = ACTIONS(1891),
    [aux_sym_mult_operator_token1] = ACTIONS(1891),
    [aux_sym_add_operator_token1] = ACTIONS(1891),
    [sym_concat_operator] = ACTIONS(1893),
    [sym_rel_operator] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1891),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [sym_assign_operator] = ACTIONS(1893),
    [sym_let_operator] = ACTIONS(1893),
    [sym__uppercase_identifier] = ACTIONS(1893),
    [aux_sym_directive_token1] = ACTIONS(1891),
    [aux_sym_tag_token1] = ACTIONS(1893),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1307)] = {
    [sym_attribute] = STATE(1307),
    [ts_builtin_sym_end] = ACTIONS(2155),
    [sym__lowercase_identifier] = ACTIONS(2153),
    [anon_sym_SEMI_SEMI] = ACTIONS(2155),
    [anon_sym_let] = ACTIONS(2153),
    [anon_sym_and] = ACTIONS(2153),
    [anon_sym_TILDE] = ACTIONS(2153),
    [anon_sym_QMARK] = ACTIONS(2153),
    [anon_sym_LPAREN] = ACTIONS(2155),
    [anon_sym_external] = ACTIONS(2153),
    [anon_sym_type] = ACTIONS(2153),
    [anon_sym_COMMA] = ACTIONS(2155),
    [anon_sym_PIPE] = ACTIONS(2153),
    [anon_sym_DOT] = ACTIONS(2155),
    [anon_sym_LBRACK] = ACTIONS(2153),
    [anon_sym_RBRACK] = ACTIONS(2155),
    [anon_sym_true] = ACTIONS(2153),
    [anon_sym_false] = ACTIONS(2153),
    [anon_sym_LBRACE] = ACTIONS(2153),
    [anon_sym_SEMI] = ACTIONS(2153),
    [anon_sym_exception] = ACTIONS(2153),
    [anon_sym_module] = ACTIONS(2153),
    [anon_sym_open] = ACTIONS(2153),
    [anon_sym_include] = ACTIONS(2153),
    [anon_sym_class] = ACTIONS(2153),
    [anon_sym_val] = ACTIONS(2153),
    [anon_sym_end] = ACTIONS(2153),
    [anon_sym_object] = ACTIONS(2153),
    [anon_sym_AMP] = ACTIONS(2153),
    [anon_sym_POUND] = ACTIONS(2153),
    [anon_sym_COLON_COLON] = ACTIONS(2155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2155),
    [anon_sym_else] = ACTIONS(2153),
    [anon_sym_new] = ACTIONS(2153),
    [anon_sym_LBRACE_LT] = ACTIONS(2155),
    [anon_sym_begin] = ACTIONS(2153),
    [sym_ocamlyacc_value] = ACTIONS(2155),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2153),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2155),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2153),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2153),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2155),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2155),
    [aux_sym_number_token1] = ACTIONS(2155),
    [aux_sym_signed_number_token1] = ACTIONS(2153),
    [anon_sym_SQUOTE] = ACTIONS(2155),
    [anon_sym_DQUOTE] = ACTIONS(2155),
    [sym_prefix_operator] = ACTIONS(2155),
    [aux_sym_sign_operator_token1] = ACTIONS(2153),
    [sym_hash_operator] = ACTIONS(2155),
    [anon_sym_lsl] = ACTIONS(2153),
    [anon_sym_lsr] = ACTIONS(2153),
    [anon_sym_asr] = ACTIONS(2153),
    [aux_sym_pow_operator_token1] = ACTIONS(2155),
    [anon_sym_mod] = ACTIONS(2153),
    [anon_sym_land] = ACTIONS(2153),
    [anon_sym_lor] = ACTIONS(2153),
    [anon_sym_lxor] = ACTIONS(2153),
    [aux_sym_mult_operator_token1] = ACTIONS(2153),
    [aux_sym_add_operator_token1] = ACTIONS(2153),
    [sym_concat_operator] = ACTIONS(2155),
    [sym_rel_operator] = ACTIONS(2153),
    [anon_sym_AMP_AMP] = ACTIONS(2153),
    [anon_sym_or] = ACTIONS(2153),
    [anon_sym_PIPE_PIPE] = ACTIONS(2153),
    [sym_assign_operator] = ACTIONS(2155),
    [sym_let_operator] = ACTIONS(2155),
    [sym__uppercase_identifier] = ACTIONS(2155),
    [aux_sym_directive_token1] = ACTIONS(2153),
    [aux_sym_tag_token1] = ACTIONS(2155),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1308)] = {
    [sym_attribute] = STATE(1308),
    [ts_builtin_sym_end] = ACTIONS(2151),
    [sym__lowercase_identifier] = ACTIONS(2149),
    [anon_sym_SEMI_SEMI] = ACTIONS(2151),
    [anon_sym_let] = ACTIONS(2149),
    [anon_sym_and] = ACTIONS(2149),
    [anon_sym_TILDE] = ACTIONS(2149),
    [anon_sym_QMARK] = ACTIONS(2149),
    [anon_sym_LPAREN] = ACTIONS(2151),
    [anon_sym_external] = ACTIONS(2149),
    [anon_sym_type] = ACTIONS(2149),
    [anon_sym_COMMA] = ACTIONS(2151),
    [anon_sym_PIPE] = ACTIONS(2149),
    [anon_sym_DOT] = ACTIONS(2151),
    [anon_sym_LBRACK] = ACTIONS(2149),
    [anon_sym_true] = ACTIONS(2149),
    [anon_sym_false] = ACTIONS(2149),
    [anon_sym_LBRACE] = ACTIONS(2149),
    [anon_sym_SEMI] = ACTIONS(2149),
    [anon_sym_exception] = ACTIONS(2149),
    [anon_sym_module] = ACTIONS(2149),
    [anon_sym_open] = ACTIONS(2149),
    [anon_sym_include] = ACTIONS(2149),
    [anon_sym_class] = ACTIONS(2149),
    [anon_sym_val] = ACTIONS(2149),
    [anon_sym_object] = ACTIONS(2149),
    [anon_sym_in] = ACTIONS(2149),
    [anon_sym_AMP] = ACTIONS(2149),
    [anon_sym_POUND] = ACTIONS(2149),
    [anon_sym_COLON_COLON] = ACTIONS(2151),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2151),
    [anon_sym_else] = ACTIONS(2149),
    [anon_sym_new] = ACTIONS(2149),
    [anon_sym_LBRACE_LT] = ACTIONS(2151),
    [anon_sym_begin] = ACTIONS(2149),
    [sym_ocamlyacc_value] = ACTIONS(2151),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2149),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2151),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2149),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2151),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2151),
    [aux_sym_number_token1] = ACTIONS(2151),
    [aux_sym_signed_number_token1] = ACTIONS(2149),
    [anon_sym_SQUOTE] = ACTIONS(2151),
    [anon_sym_DQUOTE] = ACTIONS(2151),
    [sym_prefix_operator] = ACTIONS(2151),
    [aux_sym_sign_operator_token1] = ACTIONS(2149),
    [sym_hash_operator] = ACTIONS(2151),
    [anon_sym_lsl] = ACTIONS(2149),
    [anon_sym_lsr] = ACTIONS(2149),
    [anon_sym_asr] = ACTIONS(2149),
    [aux_sym_pow_operator_token1] = ACTIONS(2151),
    [anon_sym_mod] = ACTIONS(2149),
    [anon_sym_land] = ACTIONS(2149),
    [anon_sym_lor] = ACTIONS(2149),
    [anon_sym_lxor] = ACTIONS(2149),
    [aux_sym_mult_operator_token1] = ACTIONS(2149),
    [aux_sym_add_operator_token1] = ACTIONS(2149),
    [sym_concat_operator] = ACTIONS(2151),
    [sym_rel_operator] = ACTIONS(2149),
    [anon_sym_AMP_AMP] = ACTIONS(2149),
    [anon_sym_or] = ACTIONS(2149),
    [anon_sym_PIPE_PIPE] = ACTIONS(2149),
    [sym_assign_operator] = ACTIONS(2151),
    [sym_let_operator] = ACTIONS(2151),
    [sym_let_and_operator] = ACTIONS(2151),
    [sym__uppercase_identifier] = ACTIONS(2151),
    [aux_sym_directive_token1] = ACTIONS(2149),
    [aux_sym_tag_token1] = ACTIONS(2151),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1309)] = {
    [sym_attribute] = STATE(1309),
    [ts_builtin_sym_end] = ACTIONS(1937),
    [sym__lowercase_identifier] = ACTIONS(1935),
    [anon_sym_SEMI_SEMI] = ACTIONS(1937),
    [anon_sym_let] = ACTIONS(1935),
    [anon_sym_and] = ACTIONS(1935),
    [anon_sym_TILDE] = ACTIONS(1935),
    [anon_sym_QMARK] = ACTIONS(1935),
    [anon_sym_LPAREN] = ACTIONS(1937),
    [anon_sym_external] = ACTIONS(1935),
    [anon_sym_type] = ACTIONS(1935),
    [anon_sym_COMMA] = ACTIONS(1937),
    [anon_sym_PIPE] = ACTIONS(1935),
    [anon_sym_DOT] = ACTIONS(1937),
    [anon_sym_LBRACK] = ACTIONS(1935),
    [anon_sym_true] = ACTIONS(1935),
    [anon_sym_false] = ACTIONS(1935),
    [anon_sym_LBRACE] = ACTIONS(1935),
    [anon_sym_SEMI] = ACTIONS(1935),
    [anon_sym_exception] = ACTIONS(1935),
    [anon_sym_module] = ACTIONS(1935),
    [anon_sym_open] = ACTIONS(1935),
    [anon_sym_include] = ACTIONS(1935),
    [anon_sym_class] = ACTIONS(1935),
    [anon_sym_val] = ACTIONS(1935),
    [anon_sym_object] = ACTIONS(1935),
    [anon_sym_in] = ACTIONS(1935),
    [anon_sym_AMP] = ACTIONS(1935),
    [anon_sym_POUND] = ACTIONS(1935),
    [anon_sym_COLON_COLON] = ACTIONS(1937),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1937),
    [anon_sym_else] = ACTIONS(1935),
    [anon_sym_new] = ACTIONS(1935),
    [anon_sym_LBRACE_LT] = ACTIONS(1937),
    [anon_sym_begin] = ACTIONS(1935),
    [sym_ocamlyacc_value] = ACTIONS(1937),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1935),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1937),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1935),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1935),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1937),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1937),
    [aux_sym_number_token1] = ACTIONS(1937),
    [aux_sym_signed_number_token1] = ACTIONS(1935),
    [anon_sym_SQUOTE] = ACTIONS(1937),
    [anon_sym_DQUOTE] = ACTIONS(1937),
    [sym_prefix_operator] = ACTIONS(1937),
    [aux_sym_sign_operator_token1] = ACTIONS(1935),
    [sym_hash_operator] = ACTIONS(1937),
    [anon_sym_lsl] = ACTIONS(1935),
    [anon_sym_lsr] = ACTIONS(1935),
    [anon_sym_asr] = ACTIONS(1935),
    [aux_sym_pow_operator_token1] = ACTIONS(1937),
    [anon_sym_mod] = ACTIONS(1935),
    [anon_sym_land] = ACTIONS(1935),
    [anon_sym_lor] = ACTIONS(1935),
    [anon_sym_lxor] = ACTIONS(1935),
    [aux_sym_mult_operator_token1] = ACTIONS(1935),
    [aux_sym_add_operator_token1] = ACTIONS(1935),
    [sym_concat_operator] = ACTIONS(1937),
    [sym_rel_operator] = ACTIONS(1935),
    [anon_sym_AMP_AMP] = ACTIONS(1935),
    [anon_sym_or] = ACTIONS(1935),
    [anon_sym_PIPE_PIPE] = ACTIONS(1935),
    [sym_assign_operator] = ACTIONS(1937),
    [sym_let_operator] = ACTIONS(1937),
    [sym_let_and_operator] = ACTIONS(1937),
    [sym__uppercase_identifier] = ACTIONS(1937),
    [aux_sym_directive_token1] = ACTIONS(1935),
    [aux_sym_tag_token1] = ACTIONS(1937),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1310)] = {
    [sym_attribute] = STATE(1310),
    [ts_builtin_sym_end] = ACTIONS(1941),
    [sym__lowercase_identifier] = ACTIONS(1939),
    [anon_sym_SEMI_SEMI] = ACTIONS(1941),
    [anon_sym_let] = ACTIONS(1939),
    [anon_sym_and] = ACTIONS(1939),
    [anon_sym_TILDE] = ACTIONS(1939),
    [anon_sym_QMARK] = ACTIONS(1939),
    [anon_sym_LPAREN] = ACTIONS(1941),
    [anon_sym_external] = ACTIONS(1939),
    [anon_sym_type] = ACTIONS(1939),
    [anon_sym_COMMA] = ACTIONS(1941),
    [anon_sym_PIPE] = ACTIONS(1939),
    [anon_sym_DOT] = ACTIONS(1941),
    [anon_sym_LBRACK] = ACTIONS(1939),
    [anon_sym_true] = ACTIONS(1939),
    [anon_sym_false] = ACTIONS(1939),
    [anon_sym_LBRACE] = ACTIONS(1939),
    [anon_sym_SEMI] = ACTIONS(1939),
    [anon_sym_exception] = ACTIONS(1939),
    [anon_sym_module] = ACTIONS(1939),
    [anon_sym_open] = ACTIONS(1939),
    [anon_sym_include] = ACTIONS(1939),
    [anon_sym_class] = ACTIONS(1939),
    [anon_sym_val] = ACTIONS(1939),
    [anon_sym_object] = ACTIONS(1939),
    [anon_sym_in] = ACTIONS(1939),
    [anon_sym_AMP] = ACTIONS(1939),
    [anon_sym_POUND] = ACTIONS(1939),
    [anon_sym_COLON_COLON] = ACTIONS(1941),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1941),
    [anon_sym_else] = ACTIONS(1939),
    [anon_sym_new] = ACTIONS(1939),
    [anon_sym_LBRACE_LT] = ACTIONS(1941),
    [anon_sym_begin] = ACTIONS(1939),
    [sym_ocamlyacc_value] = ACTIONS(1941),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1939),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1941),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1939),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1939),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1941),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1941),
    [aux_sym_number_token1] = ACTIONS(1941),
    [aux_sym_signed_number_token1] = ACTIONS(1939),
    [anon_sym_SQUOTE] = ACTIONS(1941),
    [anon_sym_DQUOTE] = ACTIONS(1941),
    [sym_prefix_operator] = ACTIONS(1941),
    [aux_sym_sign_operator_token1] = ACTIONS(1939),
    [sym_hash_operator] = ACTIONS(1941),
    [anon_sym_lsl] = ACTIONS(1939),
    [anon_sym_lsr] = ACTIONS(1939),
    [anon_sym_asr] = ACTIONS(1939),
    [aux_sym_pow_operator_token1] = ACTIONS(1941),
    [anon_sym_mod] = ACTIONS(1939),
    [anon_sym_land] = ACTIONS(1939),
    [anon_sym_lor] = ACTIONS(1939),
    [anon_sym_lxor] = ACTIONS(1939),
    [aux_sym_mult_operator_token1] = ACTIONS(1939),
    [aux_sym_add_operator_token1] = ACTIONS(1939),
    [sym_concat_operator] = ACTIONS(1941),
    [sym_rel_operator] = ACTIONS(1939),
    [anon_sym_AMP_AMP] = ACTIONS(1939),
    [anon_sym_or] = ACTIONS(1939),
    [anon_sym_PIPE_PIPE] = ACTIONS(1939),
    [sym_assign_operator] = ACTIONS(1941),
    [sym_let_operator] = ACTIONS(1941),
    [sym_let_and_operator] = ACTIONS(1941),
    [sym__uppercase_identifier] = ACTIONS(1941),
    [aux_sym_directive_token1] = ACTIONS(1939),
    [aux_sym_tag_token1] = ACTIONS(1941),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1311)] = {
    [sym_attribute] = STATE(1311),
    [ts_builtin_sym_end] = ACTIONS(2197),
    [sym__lowercase_identifier] = ACTIONS(2195),
    [anon_sym_SEMI_SEMI] = ACTIONS(2197),
    [anon_sym_let] = ACTIONS(2195),
    [anon_sym_and] = ACTIONS(2195),
    [anon_sym_TILDE] = ACTIONS(2195),
    [anon_sym_QMARK] = ACTIONS(2195),
    [anon_sym_LPAREN] = ACTIONS(2197),
    [anon_sym_external] = ACTIONS(2195),
    [anon_sym_type] = ACTIONS(2195),
    [anon_sym_COMMA] = ACTIONS(2197),
    [anon_sym_PIPE] = ACTIONS(2195),
    [anon_sym_DOT] = ACTIONS(2197),
    [anon_sym_LBRACK] = ACTIONS(2195),
    [anon_sym_true] = ACTIONS(2195),
    [anon_sym_false] = ACTIONS(2195),
    [anon_sym_LBRACE] = ACTIONS(2195),
    [anon_sym_SEMI] = ACTIONS(2195),
    [anon_sym_exception] = ACTIONS(2195),
    [anon_sym_module] = ACTIONS(2195),
    [anon_sym_open] = ACTIONS(2195),
    [anon_sym_include] = ACTIONS(2195),
    [anon_sym_class] = ACTIONS(2195),
    [anon_sym_val] = ACTIONS(2195),
    [anon_sym_object] = ACTIONS(2195),
    [anon_sym_in] = ACTIONS(2195),
    [anon_sym_AMP] = ACTIONS(2195),
    [anon_sym_POUND] = ACTIONS(2195),
    [anon_sym_COLON_COLON] = ACTIONS(2197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2197),
    [anon_sym_else] = ACTIONS(2195),
    [anon_sym_new] = ACTIONS(2195),
    [anon_sym_LBRACE_LT] = ACTIONS(2197),
    [anon_sym_begin] = ACTIONS(2195),
    [sym_ocamlyacc_value] = ACTIONS(2197),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2195),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2197),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2195),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2195),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2197),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2197),
    [aux_sym_number_token1] = ACTIONS(2197),
    [aux_sym_signed_number_token1] = ACTIONS(2195),
    [anon_sym_SQUOTE] = ACTIONS(2197),
    [anon_sym_DQUOTE] = ACTIONS(2197),
    [sym_prefix_operator] = ACTIONS(2197),
    [aux_sym_sign_operator_token1] = ACTIONS(2195),
    [sym_hash_operator] = ACTIONS(2197),
    [anon_sym_lsl] = ACTIONS(2195),
    [anon_sym_lsr] = ACTIONS(2195),
    [anon_sym_asr] = ACTIONS(2195),
    [aux_sym_pow_operator_token1] = ACTIONS(2197),
    [anon_sym_mod] = ACTIONS(2195),
    [anon_sym_land] = ACTIONS(2195),
    [anon_sym_lor] = ACTIONS(2195),
    [anon_sym_lxor] = ACTIONS(2195),
    [aux_sym_mult_operator_token1] = ACTIONS(2195),
    [aux_sym_add_operator_token1] = ACTIONS(2195),
    [sym_concat_operator] = ACTIONS(2197),
    [sym_rel_operator] = ACTIONS(2195),
    [anon_sym_AMP_AMP] = ACTIONS(2195),
    [anon_sym_or] = ACTIONS(2195),
    [anon_sym_PIPE_PIPE] = ACTIONS(2195),
    [sym_assign_operator] = ACTIONS(2197),
    [sym_let_operator] = ACTIONS(2197),
    [sym_let_and_operator] = ACTIONS(2197),
    [sym__uppercase_identifier] = ACTIONS(2197),
    [aux_sym_directive_token1] = ACTIONS(2195),
    [aux_sym_tag_token1] = ACTIONS(2197),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1312)] = {
    [sym_attribute] = STATE(1312),
    [ts_builtin_sym_end] = ACTIONS(1945),
    [sym__lowercase_identifier] = ACTIONS(1943),
    [anon_sym_SEMI_SEMI] = ACTIONS(1945),
    [anon_sym_let] = ACTIONS(1943),
    [anon_sym_and] = ACTIONS(1943),
    [anon_sym_TILDE] = ACTIONS(1943),
    [anon_sym_QMARK] = ACTIONS(1943),
    [anon_sym_LPAREN] = ACTIONS(1945),
    [anon_sym_external] = ACTIONS(1943),
    [anon_sym_type] = ACTIONS(1943),
    [anon_sym_COMMA] = ACTIONS(1945),
    [anon_sym_PIPE] = ACTIONS(1943),
    [anon_sym_DOT] = ACTIONS(1945),
    [anon_sym_LBRACK] = ACTIONS(1943),
    [anon_sym_true] = ACTIONS(1943),
    [anon_sym_false] = ACTIONS(1943),
    [anon_sym_LBRACE] = ACTIONS(1943),
    [anon_sym_SEMI] = ACTIONS(1943),
    [anon_sym_exception] = ACTIONS(1943),
    [anon_sym_module] = ACTIONS(1943),
    [anon_sym_open] = ACTIONS(1943),
    [anon_sym_include] = ACTIONS(1943),
    [anon_sym_class] = ACTIONS(1943),
    [anon_sym_val] = ACTIONS(1943),
    [anon_sym_object] = ACTIONS(1943),
    [anon_sym_in] = ACTIONS(1943),
    [anon_sym_AMP] = ACTIONS(1943),
    [anon_sym_POUND] = ACTIONS(1943),
    [anon_sym_COLON_COLON] = ACTIONS(1945),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1945),
    [anon_sym_else] = ACTIONS(1943),
    [anon_sym_new] = ACTIONS(1943),
    [anon_sym_LBRACE_LT] = ACTIONS(1945),
    [anon_sym_begin] = ACTIONS(1943),
    [sym_ocamlyacc_value] = ACTIONS(1945),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1943),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1945),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1943),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1943),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1945),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1945),
    [aux_sym_number_token1] = ACTIONS(1945),
    [aux_sym_signed_number_token1] = ACTIONS(1943),
    [anon_sym_SQUOTE] = ACTIONS(1945),
    [anon_sym_DQUOTE] = ACTIONS(1945),
    [sym_prefix_operator] = ACTIONS(1945),
    [aux_sym_sign_operator_token1] = ACTIONS(1943),
    [sym_hash_operator] = ACTIONS(1945),
    [anon_sym_lsl] = ACTIONS(1943),
    [anon_sym_lsr] = ACTIONS(1943),
    [anon_sym_asr] = ACTIONS(1943),
    [aux_sym_pow_operator_token1] = ACTIONS(1945),
    [anon_sym_mod] = ACTIONS(1943),
    [anon_sym_land] = ACTIONS(1943),
    [anon_sym_lor] = ACTIONS(1943),
    [anon_sym_lxor] = ACTIONS(1943),
    [aux_sym_mult_operator_token1] = ACTIONS(1943),
    [aux_sym_add_operator_token1] = ACTIONS(1943),
    [sym_concat_operator] = ACTIONS(1945),
    [sym_rel_operator] = ACTIONS(1943),
    [anon_sym_AMP_AMP] = ACTIONS(1943),
    [anon_sym_or] = ACTIONS(1943),
    [anon_sym_PIPE_PIPE] = ACTIONS(1943),
    [sym_assign_operator] = ACTIONS(1945),
    [sym_let_operator] = ACTIONS(1945),
    [sym_let_and_operator] = ACTIONS(1945),
    [sym__uppercase_identifier] = ACTIONS(1945),
    [aux_sym_directive_token1] = ACTIONS(1943),
    [aux_sym_tag_token1] = ACTIONS(1945),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1313)] = {
    [sym_attribute] = STATE(1313),
    [ts_builtin_sym_end] = ACTIONS(1953),
    [sym__lowercase_identifier] = ACTIONS(1951),
    [anon_sym_SEMI_SEMI] = ACTIONS(1953),
    [anon_sym_let] = ACTIONS(1951),
    [anon_sym_and] = ACTIONS(1951),
    [anon_sym_TILDE] = ACTIONS(1951),
    [anon_sym_QMARK] = ACTIONS(1951),
    [anon_sym_LPAREN] = ACTIONS(1953),
    [anon_sym_external] = ACTIONS(1951),
    [anon_sym_type] = ACTIONS(1951),
    [anon_sym_COMMA] = ACTIONS(1953),
    [anon_sym_PIPE] = ACTIONS(1951),
    [anon_sym_DOT] = ACTIONS(1953),
    [anon_sym_LBRACK] = ACTIONS(1951),
    [anon_sym_true] = ACTIONS(1951),
    [anon_sym_false] = ACTIONS(1951),
    [anon_sym_LBRACE] = ACTIONS(1951),
    [anon_sym_SEMI] = ACTIONS(1951),
    [anon_sym_exception] = ACTIONS(1951),
    [anon_sym_module] = ACTIONS(1951),
    [anon_sym_open] = ACTIONS(1951),
    [anon_sym_include] = ACTIONS(1951),
    [anon_sym_class] = ACTIONS(1951),
    [anon_sym_val] = ACTIONS(1951),
    [anon_sym_object] = ACTIONS(1951),
    [anon_sym_in] = ACTIONS(1951),
    [anon_sym_AMP] = ACTIONS(1951),
    [anon_sym_POUND] = ACTIONS(1951),
    [anon_sym_COLON_COLON] = ACTIONS(1953),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1953),
    [anon_sym_else] = ACTIONS(1951),
    [anon_sym_new] = ACTIONS(1951),
    [anon_sym_LBRACE_LT] = ACTIONS(1953),
    [anon_sym_begin] = ACTIONS(1951),
    [sym_ocamlyacc_value] = ACTIONS(1953),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1951),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1953),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1951),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1951),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1953),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1953),
    [aux_sym_number_token1] = ACTIONS(1953),
    [aux_sym_signed_number_token1] = ACTIONS(1951),
    [anon_sym_SQUOTE] = ACTIONS(1953),
    [anon_sym_DQUOTE] = ACTIONS(1953),
    [sym_prefix_operator] = ACTIONS(1953),
    [aux_sym_sign_operator_token1] = ACTIONS(1951),
    [sym_hash_operator] = ACTIONS(1953),
    [anon_sym_lsl] = ACTIONS(1951),
    [anon_sym_lsr] = ACTIONS(1951),
    [anon_sym_asr] = ACTIONS(1951),
    [aux_sym_pow_operator_token1] = ACTIONS(1953),
    [anon_sym_mod] = ACTIONS(1951),
    [anon_sym_land] = ACTIONS(1951),
    [anon_sym_lor] = ACTIONS(1951),
    [anon_sym_lxor] = ACTIONS(1951),
    [aux_sym_mult_operator_token1] = ACTIONS(1951),
    [aux_sym_add_operator_token1] = ACTIONS(1951),
    [sym_concat_operator] = ACTIONS(1953),
    [sym_rel_operator] = ACTIONS(1951),
    [anon_sym_AMP_AMP] = ACTIONS(1951),
    [anon_sym_or] = ACTIONS(1951),
    [anon_sym_PIPE_PIPE] = ACTIONS(1951),
    [sym_assign_operator] = ACTIONS(1953),
    [sym_let_operator] = ACTIONS(1953),
    [sym_let_and_operator] = ACTIONS(1953),
    [sym__uppercase_identifier] = ACTIONS(1953),
    [aux_sym_directive_token1] = ACTIONS(1951),
    [aux_sym_tag_token1] = ACTIONS(1953),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1314)] = {
    [sym_attribute] = STATE(1314),
    [ts_builtin_sym_end] = ACTIONS(2201),
    [sym__lowercase_identifier] = ACTIONS(2199),
    [anon_sym_SEMI_SEMI] = ACTIONS(2201),
    [anon_sym_let] = ACTIONS(2199),
    [anon_sym_and] = ACTIONS(2199),
    [anon_sym_TILDE] = ACTIONS(2199),
    [anon_sym_QMARK] = ACTIONS(2199),
    [anon_sym_LPAREN] = ACTIONS(2201),
    [anon_sym_external] = ACTIONS(2199),
    [anon_sym_type] = ACTIONS(2199),
    [anon_sym_COMMA] = ACTIONS(2201),
    [anon_sym_PIPE] = ACTIONS(2199),
    [anon_sym_DOT] = ACTIONS(2201),
    [anon_sym_LBRACK] = ACTIONS(2199),
    [anon_sym_true] = ACTIONS(2199),
    [anon_sym_false] = ACTIONS(2199),
    [anon_sym_LBRACE] = ACTIONS(2199),
    [anon_sym_SEMI] = ACTIONS(2199),
    [anon_sym_exception] = ACTIONS(2199),
    [anon_sym_module] = ACTIONS(2199),
    [anon_sym_open] = ACTIONS(2199),
    [anon_sym_include] = ACTIONS(2199),
    [anon_sym_class] = ACTIONS(2199),
    [anon_sym_val] = ACTIONS(2199),
    [anon_sym_object] = ACTIONS(2199),
    [anon_sym_in] = ACTIONS(2199),
    [anon_sym_AMP] = ACTIONS(2199),
    [anon_sym_POUND] = ACTIONS(2199),
    [anon_sym_COLON_COLON] = ACTIONS(2201),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2201),
    [anon_sym_else] = ACTIONS(2199),
    [anon_sym_new] = ACTIONS(2199),
    [anon_sym_LBRACE_LT] = ACTIONS(2201),
    [anon_sym_begin] = ACTIONS(2199),
    [sym_ocamlyacc_value] = ACTIONS(2201),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2199),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2201),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2199),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2199),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2201),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2201),
    [aux_sym_number_token1] = ACTIONS(2201),
    [aux_sym_signed_number_token1] = ACTIONS(2199),
    [anon_sym_SQUOTE] = ACTIONS(2201),
    [anon_sym_DQUOTE] = ACTIONS(2201),
    [sym_prefix_operator] = ACTIONS(2201),
    [aux_sym_sign_operator_token1] = ACTIONS(2199),
    [sym_hash_operator] = ACTIONS(2201),
    [anon_sym_lsl] = ACTIONS(2199),
    [anon_sym_lsr] = ACTIONS(2199),
    [anon_sym_asr] = ACTIONS(2199),
    [aux_sym_pow_operator_token1] = ACTIONS(2201),
    [anon_sym_mod] = ACTIONS(2199),
    [anon_sym_land] = ACTIONS(2199),
    [anon_sym_lor] = ACTIONS(2199),
    [anon_sym_lxor] = ACTIONS(2199),
    [aux_sym_mult_operator_token1] = ACTIONS(2199),
    [aux_sym_add_operator_token1] = ACTIONS(2199),
    [sym_concat_operator] = ACTIONS(2201),
    [sym_rel_operator] = ACTIONS(2199),
    [anon_sym_AMP_AMP] = ACTIONS(2199),
    [anon_sym_or] = ACTIONS(2199),
    [anon_sym_PIPE_PIPE] = ACTIONS(2199),
    [sym_assign_operator] = ACTIONS(2201),
    [sym_let_operator] = ACTIONS(2201),
    [sym_let_and_operator] = ACTIONS(2201),
    [sym__uppercase_identifier] = ACTIONS(2201),
    [aux_sym_directive_token1] = ACTIONS(2199),
    [aux_sym_tag_token1] = ACTIONS(2201),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1315)] = {
    [sym_attribute] = STATE(1315),
    [ts_builtin_sym_end] = ACTIONS(1957),
    [sym__lowercase_identifier] = ACTIONS(1955),
    [anon_sym_SEMI_SEMI] = ACTIONS(1957),
    [anon_sym_let] = ACTIONS(1955),
    [anon_sym_and] = ACTIONS(1955),
    [anon_sym_TILDE] = ACTIONS(1955),
    [anon_sym_QMARK] = ACTIONS(1955),
    [anon_sym_LPAREN] = ACTIONS(1957),
    [anon_sym_external] = ACTIONS(1955),
    [anon_sym_type] = ACTIONS(1955),
    [anon_sym_COMMA] = ACTIONS(1957),
    [anon_sym_PIPE] = ACTIONS(1955),
    [anon_sym_DOT] = ACTIONS(1957),
    [anon_sym_LBRACK] = ACTIONS(1955),
    [anon_sym_true] = ACTIONS(1955),
    [anon_sym_false] = ACTIONS(1955),
    [anon_sym_LBRACE] = ACTIONS(1955),
    [anon_sym_SEMI] = ACTIONS(1955),
    [anon_sym_exception] = ACTIONS(1955),
    [anon_sym_module] = ACTIONS(1955),
    [anon_sym_open] = ACTIONS(1955),
    [anon_sym_include] = ACTIONS(1955),
    [anon_sym_class] = ACTIONS(1955),
    [anon_sym_val] = ACTIONS(1955),
    [anon_sym_object] = ACTIONS(1955),
    [anon_sym_in] = ACTIONS(1955),
    [anon_sym_AMP] = ACTIONS(1955),
    [anon_sym_POUND] = ACTIONS(1955),
    [anon_sym_COLON_COLON] = ACTIONS(1957),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1957),
    [anon_sym_else] = ACTIONS(1955),
    [anon_sym_new] = ACTIONS(1955),
    [anon_sym_LBRACE_LT] = ACTIONS(1957),
    [anon_sym_begin] = ACTIONS(1955),
    [sym_ocamlyacc_value] = ACTIONS(1957),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1955),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1957),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1955),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1955),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1957),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1957),
    [aux_sym_number_token1] = ACTIONS(1957),
    [aux_sym_signed_number_token1] = ACTIONS(1955),
    [anon_sym_SQUOTE] = ACTIONS(1957),
    [anon_sym_DQUOTE] = ACTIONS(1957),
    [sym_prefix_operator] = ACTIONS(1957),
    [aux_sym_sign_operator_token1] = ACTIONS(1955),
    [sym_hash_operator] = ACTIONS(1957),
    [anon_sym_lsl] = ACTIONS(1955),
    [anon_sym_lsr] = ACTIONS(1955),
    [anon_sym_asr] = ACTIONS(1955),
    [aux_sym_pow_operator_token1] = ACTIONS(1957),
    [anon_sym_mod] = ACTIONS(1955),
    [anon_sym_land] = ACTIONS(1955),
    [anon_sym_lor] = ACTIONS(1955),
    [anon_sym_lxor] = ACTIONS(1955),
    [aux_sym_mult_operator_token1] = ACTIONS(1955),
    [aux_sym_add_operator_token1] = ACTIONS(1955),
    [sym_concat_operator] = ACTIONS(1957),
    [sym_rel_operator] = ACTIONS(1955),
    [anon_sym_AMP_AMP] = ACTIONS(1955),
    [anon_sym_or] = ACTIONS(1955),
    [anon_sym_PIPE_PIPE] = ACTIONS(1955),
    [sym_assign_operator] = ACTIONS(1957),
    [sym_let_operator] = ACTIONS(1957),
    [sym_let_and_operator] = ACTIONS(1957),
    [sym__uppercase_identifier] = ACTIONS(1957),
    [aux_sym_directive_token1] = ACTIONS(1955),
    [aux_sym_tag_token1] = ACTIONS(1957),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1316)] = {
    [sym_attribute] = STATE(1316),
    [ts_builtin_sym_end] = ACTIONS(2205),
    [sym__lowercase_identifier] = ACTIONS(2203),
    [anon_sym_SEMI_SEMI] = ACTIONS(2205),
    [anon_sym_let] = ACTIONS(2203),
    [anon_sym_and] = ACTIONS(2203),
    [anon_sym_TILDE] = ACTIONS(2203),
    [anon_sym_QMARK] = ACTIONS(2203),
    [anon_sym_LPAREN] = ACTIONS(2205),
    [anon_sym_external] = ACTIONS(2203),
    [anon_sym_type] = ACTIONS(2203),
    [anon_sym_COMMA] = ACTIONS(2205),
    [anon_sym_PIPE] = ACTIONS(2203),
    [anon_sym_DOT] = ACTIONS(2205),
    [anon_sym_LBRACK] = ACTIONS(2203),
    [anon_sym_true] = ACTIONS(2203),
    [anon_sym_false] = ACTIONS(2203),
    [anon_sym_LBRACE] = ACTIONS(2203),
    [anon_sym_SEMI] = ACTIONS(2203),
    [anon_sym_exception] = ACTIONS(2203),
    [anon_sym_module] = ACTIONS(2203),
    [anon_sym_open] = ACTIONS(2203),
    [anon_sym_include] = ACTIONS(2203),
    [anon_sym_class] = ACTIONS(2203),
    [anon_sym_val] = ACTIONS(2203),
    [anon_sym_object] = ACTIONS(2203),
    [anon_sym_in] = ACTIONS(2203),
    [anon_sym_AMP] = ACTIONS(2203),
    [anon_sym_POUND] = ACTIONS(2203),
    [anon_sym_COLON_COLON] = ACTIONS(2205),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2205),
    [anon_sym_else] = ACTIONS(2203),
    [anon_sym_new] = ACTIONS(2203),
    [anon_sym_LBRACE_LT] = ACTIONS(2205),
    [anon_sym_begin] = ACTIONS(2203),
    [sym_ocamlyacc_value] = ACTIONS(2205),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2203),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2205),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2203),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2203),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2205),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2205),
    [aux_sym_number_token1] = ACTIONS(2205),
    [aux_sym_signed_number_token1] = ACTIONS(2203),
    [anon_sym_SQUOTE] = ACTIONS(2205),
    [anon_sym_DQUOTE] = ACTIONS(2205),
    [sym_prefix_operator] = ACTIONS(2205),
    [aux_sym_sign_operator_token1] = ACTIONS(2203),
    [sym_hash_operator] = ACTIONS(2205),
    [anon_sym_lsl] = ACTIONS(2203),
    [anon_sym_lsr] = ACTIONS(2203),
    [anon_sym_asr] = ACTIONS(2203),
    [aux_sym_pow_operator_token1] = ACTIONS(2205),
    [anon_sym_mod] = ACTIONS(2203),
    [anon_sym_land] = ACTIONS(2203),
    [anon_sym_lor] = ACTIONS(2203),
    [anon_sym_lxor] = ACTIONS(2203),
    [aux_sym_mult_operator_token1] = ACTIONS(2203),
    [aux_sym_add_operator_token1] = ACTIONS(2203),
    [sym_concat_operator] = ACTIONS(2205),
    [sym_rel_operator] = ACTIONS(2203),
    [anon_sym_AMP_AMP] = ACTIONS(2203),
    [anon_sym_or] = ACTIONS(2203),
    [anon_sym_PIPE_PIPE] = ACTIONS(2203),
    [sym_assign_operator] = ACTIONS(2205),
    [sym_let_operator] = ACTIONS(2205),
    [sym_let_and_operator] = ACTIONS(2205),
    [sym__uppercase_identifier] = ACTIONS(2205),
    [aux_sym_directive_token1] = ACTIONS(2203),
    [aux_sym_tag_token1] = ACTIONS(2205),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1317)] = {
    [sym_attribute] = STATE(1317),
    [ts_builtin_sym_end] = ACTIONS(1961),
    [sym__lowercase_identifier] = ACTIONS(1959),
    [anon_sym_SEMI_SEMI] = ACTIONS(1961),
    [anon_sym_let] = ACTIONS(1959),
    [anon_sym_and] = ACTIONS(1959),
    [anon_sym_TILDE] = ACTIONS(1959),
    [anon_sym_QMARK] = ACTIONS(1959),
    [anon_sym_LPAREN] = ACTIONS(1961),
    [anon_sym_external] = ACTIONS(1959),
    [anon_sym_type] = ACTIONS(1959),
    [anon_sym_COMMA] = ACTIONS(1961),
    [anon_sym_PIPE] = ACTIONS(1959),
    [anon_sym_DOT] = ACTIONS(1961),
    [anon_sym_LBRACK] = ACTIONS(1959),
    [anon_sym_true] = ACTIONS(1959),
    [anon_sym_false] = ACTIONS(1959),
    [anon_sym_LBRACE] = ACTIONS(1959),
    [anon_sym_SEMI] = ACTIONS(1959),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_module] = ACTIONS(1959),
    [anon_sym_open] = ACTIONS(1959),
    [anon_sym_include] = ACTIONS(1959),
    [anon_sym_class] = ACTIONS(1959),
    [anon_sym_val] = ACTIONS(1959),
    [anon_sym_object] = ACTIONS(1959),
    [anon_sym_in] = ACTIONS(1959),
    [anon_sym_AMP] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1959),
    [anon_sym_COLON_COLON] = ACTIONS(1961),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1961),
    [anon_sym_else] = ACTIONS(1959),
    [anon_sym_new] = ACTIONS(1959),
    [anon_sym_LBRACE_LT] = ACTIONS(1961),
    [anon_sym_begin] = ACTIONS(1959),
    [sym_ocamlyacc_value] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1959),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1961),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1959),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1959),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1961),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1961),
    [aux_sym_number_token1] = ACTIONS(1961),
    [aux_sym_signed_number_token1] = ACTIONS(1959),
    [anon_sym_SQUOTE] = ACTIONS(1961),
    [anon_sym_DQUOTE] = ACTIONS(1961),
    [sym_prefix_operator] = ACTIONS(1961),
    [aux_sym_sign_operator_token1] = ACTIONS(1959),
    [sym_hash_operator] = ACTIONS(1961),
    [anon_sym_lsl] = ACTIONS(1959),
    [anon_sym_lsr] = ACTIONS(1959),
    [anon_sym_asr] = ACTIONS(1959),
    [aux_sym_pow_operator_token1] = ACTIONS(1961),
    [anon_sym_mod] = ACTIONS(1959),
    [anon_sym_land] = ACTIONS(1959),
    [anon_sym_lor] = ACTIONS(1959),
    [anon_sym_lxor] = ACTIONS(1959),
    [aux_sym_mult_operator_token1] = ACTIONS(1959),
    [aux_sym_add_operator_token1] = ACTIONS(1959),
    [sym_concat_operator] = ACTIONS(1961),
    [sym_rel_operator] = ACTIONS(1959),
    [anon_sym_AMP_AMP] = ACTIONS(1959),
    [anon_sym_or] = ACTIONS(1959),
    [anon_sym_PIPE_PIPE] = ACTIONS(1959),
    [sym_assign_operator] = ACTIONS(1961),
    [sym_let_operator] = ACTIONS(1961),
    [sym_let_and_operator] = ACTIONS(1961),
    [sym__uppercase_identifier] = ACTIONS(1961),
    [aux_sym_directive_token1] = ACTIONS(1959),
    [aux_sym_tag_token1] = ACTIONS(1961),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1318)] = {
    [sym__parenthesized_abstract_type] = STATE(1739),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5036),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1318),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2437),
    [anon_sym_COMMA] = ACTIONS(2433),
    [anon_sym_PIPE] = ACTIONS(2433),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_RBRACK] = ACTIONS(2433),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_SEMI] = ACTIONS(2433),
    [anon_sym_RBRACE] = ACTIONS(2433),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_as] = ACTIONS(2439),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(2433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1319)] = {
    [sym_attribute] = STATE(1319),
    [ts_builtin_sym_end] = ACTIONS(2033),
    [sym__lowercase_identifier] = ACTIONS(2031),
    [anon_sym_SEMI_SEMI] = ACTIONS(2033),
    [anon_sym_let] = ACTIONS(2031),
    [anon_sym_and] = ACTIONS(2031),
    [anon_sym_TILDE] = ACTIONS(2031),
    [anon_sym_QMARK] = ACTIONS(2031),
    [anon_sym_LPAREN] = ACTIONS(2033),
    [anon_sym_external] = ACTIONS(2031),
    [anon_sym_type] = ACTIONS(2031),
    [anon_sym_COMMA] = ACTIONS(2033),
    [anon_sym_PIPE] = ACTIONS(2031),
    [anon_sym_DOT] = ACTIONS(2033),
    [anon_sym_LBRACK] = ACTIONS(2031),
    [anon_sym_true] = ACTIONS(2031),
    [anon_sym_false] = ACTIONS(2031),
    [anon_sym_LBRACE] = ACTIONS(2031),
    [anon_sym_SEMI] = ACTIONS(2031),
    [anon_sym_exception] = ACTIONS(2031),
    [anon_sym_module] = ACTIONS(2031),
    [anon_sym_open] = ACTIONS(2031),
    [anon_sym_include] = ACTIONS(2031),
    [anon_sym_class] = ACTIONS(2031),
    [anon_sym_val] = ACTIONS(2031),
    [anon_sym_object] = ACTIONS(2031),
    [anon_sym_in] = ACTIONS(2031),
    [anon_sym_AMP] = ACTIONS(2031),
    [anon_sym_POUND] = ACTIONS(2031),
    [anon_sym_COLON_COLON] = ACTIONS(2033),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2033),
    [anon_sym_else] = ACTIONS(2031),
    [anon_sym_new] = ACTIONS(2031),
    [anon_sym_LBRACE_LT] = ACTIONS(2033),
    [anon_sym_begin] = ACTIONS(2031),
    [sym_ocamlyacc_value] = ACTIONS(2033),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2031),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2033),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2031),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2031),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2033),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2033),
    [aux_sym_number_token1] = ACTIONS(2033),
    [aux_sym_signed_number_token1] = ACTIONS(2031),
    [anon_sym_SQUOTE] = ACTIONS(2033),
    [anon_sym_DQUOTE] = ACTIONS(2033),
    [sym_prefix_operator] = ACTIONS(2033),
    [aux_sym_sign_operator_token1] = ACTIONS(2031),
    [sym_hash_operator] = ACTIONS(2033),
    [anon_sym_lsl] = ACTIONS(2031),
    [anon_sym_lsr] = ACTIONS(2031),
    [anon_sym_asr] = ACTIONS(2031),
    [aux_sym_pow_operator_token1] = ACTIONS(2033),
    [anon_sym_mod] = ACTIONS(2031),
    [anon_sym_land] = ACTIONS(2031),
    [anon_sym_lor] = ACTIONS(2031),
    [anon_sym_lxor] = ACTIONS(2031),
    [aux_sym_mult_operator_token1] = ACTIONS(2031),
    [aux_sym_add_operator_token1] = ACTIONS(2031),
    [sym_concat_operator] = ACTIONS(2033),
    [sym_rel_operator] = ACTIONS(2031),
    [anon_sym_AMP_AMP] = ACTIONS(2031),
    [anon_sym_or] = ACTIONS(2031),
    [anon_sym_PIPE_PIPE] = ACTIONS(2031),
    [sym_assign_operator] = ACTIONS(2033),
    [sym_let_operator] = ACTIONS(2033),
    [sym_let_and_operator] = ACTIONS(2033),
    [sym__uppercase_identifier] = ACTIONS(2033),
    [aux_sym_directive_token1] = ACTIONS(2031),
    [aux_sym_tag_token1] = ACTIONS(2033),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1320)] = {
    [sym_attribute] = STATE(1320),
    [ts_builtin_sym_end] = ACTIONS(1949),
    [sym__lowercase_identifier] = ACTIONS(1947),
    [anon_sym_SEMI_SEMI] = ACTIONS(1949),
    [anon_sym_let] = ACTIONS(1947),
    [anon_sym_and] = ACTIONS(1947),
    [anon_sym_TILDE] = ACTIONS(1947),
    [anon_sym_QMARK] = ACTIONS(1947),
    [anon_sym_LPAREN] = ACTIONS(1949),
    [anon_sym_external] = ACTIONS(1947),
    [anon_sym_type] = ACTIONS(1947),
    [anon_sym_COMMA] = ACTIONS(1949),
    [anon_sym_PIPE] = ACTIONS(1947),
    [anon_sym_DOT] = ACTIONS(1949),
    [anon_sym_LBRACK] = ACTIONS(1947),
    [anon_sym_true] = ACTIONS(1947),
    [anon_sym_false] = ACTIONS(1947),
    [anon_sym_LBRACE] = ACTIONS(1947),
    [anon_sym_SEMI] = ACTIONS(1947),
    [anon_sym_exception] = ACTIONS(1947),
    [anon_sym_module] = ACTIONS(1947),
    [anon_sym_open] = ACTIONS(1947),
    [anon_sym_include] = ACTIONS(1947),
    [anon_sym_class] = ACTIONS(1947),
    [anon_sym_val] = ACTIONS(1947),
    [anon_sym_object] = ACTIONS(1947),
    [anon_sym_AMP] = ACTIONS(1947),
    [anon_sym_POUND] = ACTIONS(1947),
    [anon_sym_COLON_COLON] = ACTIONS(1949),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1949),
    [anon_sym_LT_DASH] = ACTIONS(2574),
    [anon_sym_else] = ACTIONS(1947),
    [anon_sym_new] = ACTIONS(1947),
    [anon_sym_LBRACE_LT] = ACTIONS(1949),
    [anon_sym_begin] = ACTIONS(1947),
    [sym_ocamlyacc_value] = ACTIONS(1949),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1947),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1949),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1949),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1949),
    [aux_sym_number_token1] = ACTIONS(1949),
    [aux_sym_signed_number_token1] = ACTIONS(1947),
    [anon_sym_SQUOTE] = ACTIONS(1949),
    [anon_sym_DQUOTE] = ACTIONS(1949),
    [sym_prefix_operator] = ACTIONS(1949),
    [aux_sym_sign_operator_token1] = ACTIONS(1947),
    [sym_hash_operator] = ACTIONS(1949),
    [anon_sym_lsl] = ACTIONS(1947),
    [anon_sym_lsr] = ACTIONS(1947),
    [anon_sym_asr] = ACTIONS(1947),
    [aux_sym_pow_operator_token1] = ACTIONS(1949),
    [anon_sym_mod] = ACTIONS(1947),
    [anon_sym_land] = ACTIONS(1947),
    [anon_sym_lor] = ACTIONS(1947),
    [anon_sym_lxor] = ACTIONS(1947),
    [aux_sym_mult_operator_token1] = ACTIONS(1947),
    [aux_sym_add_operator_token1] = ACTIONS(1947),
    [sym_concat_operator] = ACTIONS(1949),
    [sym_rel_operator] = ACTIONS(1947),
    [anon_sym_AMP_AMP] = ACTIONS(1947),
    [anon_sym_or] = ACTIONS(1947),
    [anon_sym_PIPE_PIPE] = ACTIONS(1947),
    [sym_assign_operator] = ACTIONS(1949),
    [sym_let_operator] = ACTIONS(1949),
    [sym_let_and_operator] = ACTIONS(1949),
    [sym__uppercase_identifier] = ACTIONS(1949),
    [aux_sym_directive_token1] = ACTIONS(1947),
    [aux_sym_tag_token1] = ACTIONS(1949),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1321)] = {
    [sym_attribute] = STATE(1321),
    [ts_builtin_sym_end] = ACTIONS(1965),
    [sym__lowercase_identifier] = ACTIONS(1963),
    [anon_sym_SEMI_SEMI] = ACTIONS(1965),
    [anon_sym_let] = ACTIONS(1963),
    [anon_sym_and] = ACTIONS(1963),
    [anon_sym_TILDE] = ACTIONS(1963),
    [anon_sym_QMARK] = ACTIONS(1963),
    [anon_sym_LPAREN] = ACTIONS(1965),
    [anon_sym_external] = ACTIONS(1963),
    [anon_sym_type] = ACTIONS(1963),
    [anon_sym_COMMA] = ACTIONS(1965),
    [anon_sym_PIPE] = ACTIONS(1963),
    [anon_sym_DOT] = ACTIONS(1965),
    [anon_sym_LBRACK] = ACTIONS(1963),
    [anon_sym_true] = ACTIONS(1963),
    [anon_sym_false] = ACTIONS(1963),
    [anon_sym_LBRACE] = ACTIONS(1963),
    [anon_sym_SEMI] = ACTIONS(1963),
    [anon_sym_exception] = ACTIONS(1963),
    [anon_sym_module] = ACTIONS(1963),
    [anon_sym_open] = ACTIONS(1963),
    [anon_sym_include] = ACTIONS(1963),
    [anon_sym_class] = ACTIONS(1963),
    [anon_sym_val] = ACTIONS(1963),
    [anon_sym_object] = ACTIONS(1963),
    [anon_sym_in] = ACTIONS(1963),
    [anon_sym_AMP] = ACTIONS(1963),
    [anon_sym_POUND] = ACTIONS(1963),
    [anon_sym_COLON_COLON] = ACTIONS(1965),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1965),
    [anon_sym_else] = ACTIONS(1963),
    [anon_sym_new] = ACTIONS(1963),
    [anon_sym_LBRACE_LT] = ACTIONS(1965),
    [anon_sym_begin] = ACTIONS(1963),
    [sym_ocamlyacc_value] = ACTIONS(1965),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1963),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1965),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1965),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1965),
    [aux_sym_signed_number_token1] = ACTIONS(1963),
    [anon_sym_SQUOTE] = ACTIONS(1965),
    [anon_sym_DQUOTE] = ACTIONS(1965),
    [sym_prefix_operator] = ACTIONS(1965),
    [aux_sym_sign_operator_token1] = ACTIONS(1963),
    [sym_hash_operator] = ACTIONS(1965),
    [anon_sym_lsl] = ACTIONS(1963),
    [anon_sym_lsr] = ACTIONS(1963),
    [anon_sym_asr] = ACTIONS(1963),
    [aux_sym_pow_operator_token1] = ACTIONS(1965),
    [anon_sym_mod] = ACTIONS(1963),
    [anon_sym_land] = ACTIONS(1963),
    [anon_sym_lor] = ACTIONS(1963),
    [anon_sym_lxor] = ACTIONS(1963),
    [aux_sym_mult_operator_token1] = ACTIONS(1963),
    [aux_sym_add_operator_token1] = ACTIONS(1963),
    [sym_concat_operator] = ACTIONS(1965),
    [sym_rel_operator] = ACTIONS(1963),
    [anon_sym_AMP_AMP] = ACTIONS(1963),
    [anon_sym_or] = ACTIONS(1963),
    [anon_sym_PIPE_PIPE] = ACTIONS(1963),
    [sym_assign_operator] = ACTIONS(1965),
    [sym_let_operator] = ACTIONS(1965),
    [sym_let_and_operator] = ACTIONS(1965),
    [sym__uppercase_identifier] = ACTIONS(1965),
    [aux_sym_directive_token1] = ACTIONS(1963),
    [aux_sym_tag_token1] = ACTIONS(1965),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1322)] = {
    [sym_attribute] = STATE(1322),
    [ts_builtin_sym_end] = ACTIONS(2209),
    [sym__lowercase_identifier] = ACTIONS(2207),
    [anon_sym_SEMI_SEMI] = ACTIONS(2209),
    [anon_sym_let] = ACTIONS(2207),
    [anon_sym_and] = ACTIONS(2207),
    [anon_sym_TILDE] = ACTIONS(2207),
    [anon_sym_QMARK] = ACTIONS(2207),
    [anon_sym_LPAREN] = ACTIONS(2209),
    [anon_sym_external] = ACTIONS(2207),
    [anon_sym_type] = ACTIONS(2207),
    [anon_sym_COMMA] = ACTIONS(2209),
    [anon_sym_PIPE] = ACTIONS(2207),
    [anon_sym_DOT] = ACTIONS(2209),
    [anon_sym_LBRACK] = ACTIONS(2207),
    [anon_sym_true] = ACTIONS(2207),
    [anon_sym_false] = ACTIONS(2207),
    [anon_sym_LBRACE] = ACTIONS(2207),
    [anon_sym_SEMI] = ACTIONS(2207),
    [anon_sym_exception] = ACTIONS(2207),
    [anon_sym_module] = ACTIONS(2207),
    [anon_sym_open] = ACTIONS(2207),
    [anon_sym_include] = ACTIONS(2207),
    [anon_sym_class] = ACTIONS(2207),
    [anon_sym_val] = ACTIONS(2207),
    [anon_sym_object] = ACTIONS(2207),
    [anon_sym_in] = ACTIONS(2207),
    [anon_sym_AMP] = ACTIONS(2207),
    [anon_sym_POUND] = ACTIONS(2207),
    [anon_sym_COLON_COLON] = ACTIONS(2209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2209),
    [anon_sym_else] = ACTIONS(2207),
    [anon_sym_new] = ACTIONS(2207),
    [anon_sym_LBRACE_LT] = ACTIONS(2209),
    [anon_sym_begin] = ACTIONS(2207),
    [sym_ocamlyacc_value] = ACTIONS(2209),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2207),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2209),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2207),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2207),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2209),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2209),
    [aux_sym_number_token1] = ACTIONS(2209),
    [aux_sym_signed_number_token1] = ACTIONS(2207),
    [anon_sym_SQUOTE] = ACTIONS(2209),
    [anon_sym_DQUOTE] = ACTIONS(2209),
    [sym_prefix_operator] = ACTIONS(2209),
    [aux_sym_sign_operator_token1] = ACTIONS(2207),
    [sym_hash_operator] = ACTIONS(2209),
    [anon_sym_lsl] = ACTIONS(2207),
    [anon_sym_lsr] = ACTIONS(2207),
    [anon_sym_asr] = ACTIONS(2207),
    [aux_sym_pow_operator_token1] = ACTIONS(2209),
    [anon_sym_mod] = ACTIONS(2207),
    [anon_sym_land] = ACTIONS(2207),
    [anon_sym_lor] = ACTIONS(2207),
    [anon_sym_lxor] = ACTIONS(2207),
    [aux_sym_mult_operator_token1] = ACTIONS(2207),
    [aux_sym_add_operator_token1] = ACTIONS(2207),
    [sym_concat_operator] = ACTIONS(2209),
    [sym_rel_operator] = ACTIONS(2207),
    [anon_sym_AMP_AMP] = ACTIONS(2207),
    [anon_sym_or] = ACTIONS(2207),
    [anon_sym_PIPE_PIPE] = ACTIONS(2207),
    [sym_assign_operator] = ACTIONS(2209),
    [sym_let_operator] = ACTIONS(2209),
    [sym_let_and_operator] = ACTIONS(2209),
    [sym__uppercase_identifier] = ACTIONS(2209),
    [aux_sym_directive_token1] = ACTIONS(2207),
    [aux_sym_tag_token1] = ACTIONS(2209),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1323)] = {
    [sym_attribute] = STATE(1323),
    [ts_builtin_sym_end] = ACTIONS(1933),
    [sym__lowercase_identifier] = ACTIONS(1931),
    [anon_sym_SEMI_SEMI] = ACTIONS(1933),
    [anon_sym_let] = ACTIONS(1931),
    [anon_sym_and] = ACTIONS(1931),
    [anon_sym_TILDE] = ACTIONS(1931),
    [anon_sym_QMARK] = ACTIONS(1931),
    [anon_sym_LPAREN] = ACTIONS(1933),
    [anon_sym_external] = ACTIONS(1931),
    [anon_sym_type] = ACTIONS(1931),
    [anon_sym_COMMA] = ACTIONS(1933),
    [anon_sym_PIPE] = ACTIONS(1931),
    [anon_sym_LBRACK] = ACTIONS(1931),
    [anon_sym_RBRACK] = ACTIONS(1933),
    [anon_sym_true] = ACTIONS(1931),
    [anon_sym_false] = ACTIONS(1931),
    [anon_sym_LBRACE] = ACTIONS(1931),
    [anon_sym_SEMI] = ACTIONS(1931),
    [anon_sym_exception] = ACTIONS(1931),
    [anon_sym_module] = ACTIONS(1931),
    [anon_sym_open] = ACTIONS(1931),
    [anon_sym_include] = ACTIONS(1931),
    [anon_sym_class] = ACTIONS(1931),
    [anon_sym_val] = ACTIONS(1931),
    [anon_sym_end] = ACTIONS(1931),
    [anon_sym_object] = ACTIONS(1931),
    [anon_sym_in] = ACTIONS(1931),
    [anon_sym_AMP] = ACTIONS(1931),
    [anon_sym_COLON_COLON] = ACTIONS(1933),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1933),
    [anon_sym_else] = ACTIONS(1931),
    [anon_sym_new] = ACTIONS(1931),
    [anon_sym_LBRACE_LT] = ACTIONS(1933),
    [anon_sym_begin] = ACTIONS(1931),
    [sym_ocamlyacc_value] = ACTIONS(1933),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1931),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1933),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1933),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1933),
    [aux_sym_number_token1] = ACTIONS(1933),
    [aux_sym_signed_number_token1] = ACTIONS(1931),
    [anon_sym_SQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1933),
    [sym_prefix_operator] = ACTIONS(1933),
    [aux_sym_sign_operator_token1] = ACTIONS(1931),
    [anon_sym_lsl] = ACTIONS(1931),
    [anon_sym_lsr] = ACTIONS(1931),
    [anon_sym_asr] = ACTIONS(1931),
    [aux_sym_pow_operator_token1] = ACTIONS(1933),
    [anon_sym_mod] = ACTIONS(1931),
    [anon_sym_land] = ACTIONS(1931),
    [anon_sym_lor] = ACTIONS(1931),
    [anon_sym_lxor] = ACTIONS(1931),
    [aux_sym_mult_operator_token1] = ACTIONS(1931),
    [aux_sym_add_operator_token1] = ACTIONS(1931),
    [sym_concat_operator] = ACTIONS(1933),
    [sym_rel_operator] = ACTIONS(1931),
    [anon_sym_AMP_AMP] = ACTIONS(1931),
    [anon_sym_or] = ACTIONS(1931),
    [anon_sym_PIPE_PIPE] = ACTIONS(1931),
    [sym_assign_operator] = ACTIONS(1933),
    [sym_let_operator] = ACTIONS(1933),
    [sym_let_and_operator] = ACTIONS(1933),
    [sym__uppercase_identifier] = ACTIONS(1933),
    [aux_sym_directive_token1] = ACTIONS(1933),
    [aux_sym_tag_token1] = ACTIONS(1933),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1324)] = {
    [sym_attribute] = STATE(1324),
    [ts_builtin_sym_end] = ACTIONS(1893),
    [sym__lowercase_identifier] = ACTIONS(1891),
    [anon_sym_SEMI_SEMI] = ACTIONS(1893),
    [anon_sym_let] = ACTIONS(1891),
    [anon_sym_and] = ACTIONS(1891),
    [anon_sym_TILDE] = ACTIONS(1891),
    [anon_sym_QMARK] = ACTIONS(1891),
    [anon_sym_LPAREN] = ACTIONS(1893),
    [anon_sym_external] = ACTIONS(1891),
    [anon_sym_type] = ACTIONS(1891),
    [anon_sym_COMMA] = ACTIONS(1893),
    [anon_sym_PIPE] = ACTIONS(1891),
    [anon_sym_DOT] = ACTIONS(1893),
    [anon_sym_LBRACK] = ACTIONS(1891),
    [anon_sym_true] = ACTIONS(1891),
    [anon_sym_false] = ACTIONS(1891),
    [anon_sym_LBRACE] = ACTIONS(1891),
    [anon_sym_SEMI] = ACTIONS(1891),
    [anon_sym_exception] = ACTIONS(1891),
    [anon_sym_module] = ACTIONS(1891),
    [anon_sym_open] = ACTIONS(1891),
    [anon_sym_include] = ACTIONS(1891),
    [anon_sym_class] = ACTIONS(1891),
    [anon_sym_val] = ACTIONS(1891),
    [anon_sym_object] = ACTIONS(1891),
    [anon_sym_AMP] = ACTIONS(1891),
    [anon_sym_POUND] = ACTIONS(1891),
    [anon_sym_COLON_COLON] = ACTIONS(1893),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1893),
    [anon_sym_LT_DASH] = ACTIONS(1891),
    [anon_sym_else] = ACTIONS(1891),
    [anon_sym_new] = ACTIONS(1891),
    [anon_sym_LBRACE_LT] = ACTIONS(1893),
    [anon_sym_begin] = ACTIONS(1891),
    [sym_ocamlyacc_value] = ACTIONS(1893),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1891),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1893),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1891),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1891),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1893),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1893),
    [aux_sym_number_token1] = ACTIONS(1893),
    [aux_sym_signed_number_token1] = ACTIONS(1891),
    [anon_sym_SQUOTE] = ACTIONS(1893),
    [anon_sym_DQUOTE] = ACTIONS(1893),
    [sym_prefix_operator] = ACTIONS(1893),
    [aux_sym_sign_operator_token1] = ACTIONS(1891),
    [sym_hash_operator] = ACTIONS(1893),
    [anon_sym_lsl] = ACTIONS(1891),
    [anon_sym_lsr] = ACTIONS(1891),
    [anon_sym_asr] = ACTIONS(1891),
    [aux_sym_pow_operator_token1] = ACTIONS(1893),
    [anon_sym_mod] = ACTIONS(1891),
    [anon_sym_land] = ACTIONS(1891),
    [anon_sym_lor] = ACTIONS(1891),
    [anon_sym_lxor] = ACTIONS(1891),
    [aux_sym_mult_operator_token1] = ACTIONS(1891),
    [aux_sym_add_operator_token1] = ACTIONS(1891),
    [sym_concat_operator] = ACTIONS(1893),
    [sym_rel_operator] = ACTIONS(1891),
    [anon_sym_AMP_AMP] = ACTIONS(1891),
    [anon_sym_or] = ACTIONS(1891),
    [anon_sym_PIPE_PIPE] = ACTIONS(1891),
    [sym_assign_operator] = ACTIONS(1893),
    [sym_let_operator] = ACTIONS(1893),
    [sym__uppercase_identifier] = ACTIONS(1893),
    [aux_sym_directive_token1] = ACTIONS(1891),
    [aux_sym_tag_token1] = ACTIONS(1893),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1325)] = {
    [sym_attribute] = STATE(1325),
    [sym__lowercase_identifier] = ACTIONS(2179),
    [anon_sym_SEMI_SEMI] = ACTIONS(2181),
    [anon_sym_let] = ACTIONS(2179),
    [anon_sym_and] = ACTIONS(2179),
    [anon_sym_TILDE] = ACTIONS(2179),
    [anon_sym_QMARK] = ACTIONS(2179),
    [anon_sym_LPAREN] = ACTIONS(2181),
    [anon_sym_external] = ACTIONS(2179),
    [anon_sym_type] = ACTIONS(2179),
    [anon_sym_COMMA] = ACTIONS(2181),
    [anon_sym_PIPE] = ACTIONS(2179),
    [anon_sym_DOT] = ACTIONS(2181),
    [anon_sym_LBRACK] = ACTIONS(2179),
    [anon_sym_RBRACK] = ACTIONS(2181),
    [anon_sym_true] = ACTIONS(2179),
    [anon_sym_false] = ACTIONS(2179),
    [anon_sym_LBRACE] = ACTIONS(2179),
    [anon_sym_SEMI] = ACTIONS(2179),
    [anon_sym_exception] = ACTIONS(2179),
    [anon_sym_module] = ACTIONS(2179),
    [anon_sym_open] = ACTIONS(2179),
    [anon_sym_include] = ACTIONS(2179),
    [anon_sym_class] = ACTIONS(2179),
    [anon_sym_val] = ACTIONS(2179),
    [anon_sym_end] = ACTIONS(2179),
    [anon_sym_object] = ACTIONS(2179),
    [anon_sym_AMP] = ACTIONS(2179),
    [anon_sym_POUND] = ACTIONS(2179),
    [anon_sym_COLON_COLON] = ACTIONS(2181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2181),
    [anon_sym_else] = ACTIONS(2179),
    [anon_sym_new] = ACTIONS(2179),
    [anon_sym_LBRACE_LT] = ACTIONS(2181),
    [anon_sym_begin] = ACTIONS(2179),
    [sym_ocamlyacc_value] = ACTIONS(2181),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2179),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2181),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2179),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2181),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2181),
    [aux_sym_number_token1] = ACTIONS(2181),
    [aux_sym_signed_number_token1] = ACTIONS(2179),
    [anon_sym_SQUOTE] = ACTIONS(2181),
    [anon_sym_DQUOTE] = ACTIONS(2181),
    [sym_prefix_operator] = ACTIONS(2181),
    [aux_sym_sign_operator_token1] = ACTIONS(2179),
    [sym_hash_operator] = ACTIONS(2181),
    [anon_sym_lsl] = ACTIONS(2179),
    [anon_sym_lsr] = ACTIONS(2179),
    [anon_sym_asr] = ACTIONS(2179),
    [aux_sym_pow_operator_token1] = ACTIONS(2181),
    [anon_sym_mod] = ACTIONS(2179),
    [anon_sym_land] = ACTIONS(2179),
    [anon_sym_lor] = ACTIONS(2179),
    [anon_sym_lxor] = ACTIONS(2179),
    [aux_sym_mult_operator_token1] = ACTIONS(2179),
    [aux_sym_add_operator_token1] = ACTIONS(2179),
    [sym_concat_operator] = ACTIONS(2181),
    [sym_rel_operator] = ACTIONS(2179),
    [anon_sym_AMP_AMP] = ACTIONS(2179),
    [anon_sym_or] = ACTIONS(2179),
    [anon_sym_PIPE_PIPE] = ACTIONS(2179),
    [sym_assign_operator] = ACTIONS(2181),
    [sym_let_operator] = ACTIONS(2181),
    [sym__uppercase_identifier] = ACTIONS(2181),
    [aux_sym_directive_token1] = ACTIONS(2179),
    [aux_sym_tag_token1] = ACTIONS(2181),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1326)] = {
    [sym_attribute] = STATE(1326),
    [ts_builtin_sym_end] = ACTIONS(1901),
    [sym__lowercase_identifier] = ACTIONS(1899),
    [anon_sym_SEMI_SEMI] = ACTIONS(1901),
    [anon_sym_let] = ACTIONS(1899),
    [anon_sym_and] = ACTIONS(1899),
    [anon_sym_TILDE] = ACTIONS(1899),
    [anon_sym_QMARK] = ACTIONS(1899),
    [anon_sym_LPAREN] = ACTIONS(1901),
    [anon_sym_external] = ACTIONS(1899),
    [anon_sym_type] = ACTIONS(1899),
    [anon_sym_COMMA] = ACTIONS(1901),
    [anon_sym_PIPE] = ACTIONS(1899),
    [anon_sym_DOT] = ACTIONS(1901),
    [anon_sym_LBRACK] = ACTIONS(1899),
    [anon_sym_true] = ACTIONS(1899),
    [anon_sym_false] = ACTIONS(1899),
    [anon_sym_LBRACE] = ACTIONS(1899),
    [anon_sym_SEMI] = ACTIONS(1899),
    [anon_sym_exception] = ACTIONS(1899),
    [anon_sym_module] = ACTIONS(1899),
    [anon_sym_open] = ACTIONS(1899),
    [anon_sym_include] = ACTIONS(1899),
    [anon_sym_class] = ACTIONS(1899),
    [anon_sym_val] = ACTIONS(1899),
    [anon_sym_object] = ACTIONS(1899),
    [anon_sym_AMP] = ACTIONS(1899),
    [anon_sym_POUND] = ACTIONS(1899),
    [anon_sym_COLON_COLON] = ACTIONS(1901),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1901),
    [anon_sym_LT_DASH] = ACTIONS(1899),
    [anon_sym_else] = ACTIONS(1899),
    [anon_sym_new] = ACTIONS(1899),
    [anon_sym_LBRACE_LT] = ACTIONS(1901),
    [anon_sym_begin] = ACTIONS(1899),
    [sym_ocamlyacc_value] = ACTIONS(1901),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1899),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1901),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1899),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1899),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1901),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1901),
    [aux_sym_number_token1] = ACTIONS(1901),
    [aux_sym_signed_number_token1] = ACTIONS(1899),
    [anon_sym_SQUOTE] = ACTIONS(1901),
    [anon_sym_DQUOTE] = ACTIONS(1901),
    [sym_prefix_operator] = ACTIONS(1901),
    [aux_sym_sign_operator_token1] = ACTIONS(1899),
    [sym_hash_operator] = ACTIONS(1901),
    [anon_sym_lsl] = ACTIONS(1899),
    [anon_sym_lsr] = ACTIONS(1899),
    [anon_sym_asr] = ACTIONS(1899),
    [aux_sym_pow_operator_token1] = ACTIONS(1901),
    [anon_sym_mod] = ACTIONS(1899),
    [anon_sym_land] = ACTIONS(1899),
    [anon_sym_lor] = ACTIONS(1899),
    [anon_sym_lxor] = ACTIONS(1899),
    [aux_sym_mult_operator_token1] = ACTIONS(1899),
    [aux_sym_add_operator_token1] = ACTIONS(1899),
    [sym_concat_operator] = ACTIONS(1901),
    [sym_rel_operator] = ACTIONS(1899),
    [anon_sym_AMP_AMP] = ACTIONS(1899),
    [anon_sym_or] = ACTIONS(1899),
    [anon_sym_PIPE_PIPE] = ACTIONS(1899),
    [sym_assign_operator] = ACTIONS(1901),
    [sym_let_operator] = ACTIONS(1901),
    [sym__uppercase_identifier] = ACTIONS(1901),
    [aux_sym_directive_token1] = ACTIONS(1899),
    [aux_sym_tag_token1] = ACTIONS(1901),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1327)] = {
    [sym_attribute] = STATE(1327),
    [ts_builtin_sym_end] = ACTIONS(1889),
    [sym__lowercase_identifier] = ACTIONS(1887),
    [anon_sym_SEMI_SEMI] = ACTIONS(1889),
    [anon_sym_let] = ACTIONS(1887),
    [anon_sym_and] = ACTIONS(1887),
    [anon_sym_TILDE] = ACTIONS(1887),
    [anon_sym_QMARK] = ACTIONS(1887),
    [anon_sym_LPAREN] = ACTIONS(1889),
    [anon_sym_external] = ACTIONS(1887),
    [anon_sym_type] = ACTIONS(1887),
    [anon_sym_COMMA] = ACTIONS(1889),
    [anon_sym_PIPE] = ACTIONS(1887),
    [anon_sym_DOT] = ACTIONS(1889),
    [anon_sym_LBRACK] = ACTIONS(1887),
    [anon_sym_true] = ACTIONS(1887),
    [anon_sym_false] = ACTIONS(1887),
    [anon_sym_LBRACE] = ACTIONS(1887),
    [anon_sym_SEMI] = ACTIONS(1887),
    [anon_sym_exception] = ACTIONS(1887),
    [anon_sym_module] = ACTIONS(1887),
    [anon_sym_open] = ACTIONS(1887),
    [anon_sym_include] = ACTIONS(1887),
    [anon_sym_class] = ACTIONS(1887),
    [anon_sym_val] = ACTIONS(1887),
    [anon_sym_object] = ACTIONS(1887),
    [anon_sym_AMP] = ACTIONS(1887),
    [anon_sym_POUND] = ACTIONS(1887),
    [anon_sym_COLON_COLON] = ACTIONS(1889),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1889),
    [anon_sym_LT_DASH] = ACTIONS(1887),
    [anon_sym_else] = ACTIONS(1887),
    [anon_sym_new] = ACTIONS(1887),
    [anon_sym_LBRACE_LT] = ACTIONS(1889),
    [anon_sym_begin] = ACTIONS(1887),
    [sym_ocamlyacc_value] = ACTIONS(1889),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1887),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1889),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1887),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1887),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1889),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1889),
    [aux_sym_number_token1] = ACTIONS(1889),
    [aux_sym_signed_number_token1] = ACTIONS(1887),
    [anon_sym_SQUOTE] = ACTIONS(1889),
    [anon_sym_DQUOTE] = ACTIONS(1889),
    [sym_prefix_operator] = ACTIONS(1889),
    [aux_sym_sign_operator_token1] = ACTIONS(1887),
    [sym_hash_operator] = ACTIONS(1889),
    [anon_sym_lsl] = ACTIONS(1887),
    [anon_sym_lsr] = ACTIONS(1887),
    [anon_sym_asr] = ACTIONS(1887),
    [aux_sym_pow_operator_token1] = ACTIONS(1889),
    [anon_sym_mod] = ACTIONS(1887),
    [anon_sym_land] = ACTIONS(1887),
    [anon_sym_lor] = ACTIONS(1887),
    [anon_sym_lxor] = ACTIONS(1887),
    [aux_sym_mult_operator_token1] = ACTIONS(1887),
    [aux_sym_add_operator_token1] = ACTIONS(1887),
    [sym_concat_operator] = ACTIONS(1889),
    [sym_rel_operator] = ACTIONS(1887),
    [anon_sym_AMP_AMP] = ACTIONS(1887),
    [anon_sym_or] = ACTIONS(1887),
    [anon_sym_PIPE_PIPE] = ACTIONS(1887),
    [sym_assign_operator] = ACTIONS(1889),
    [sym_let_operator] = ACTIONS(1889),
    [sym__uppercase_identifier] = ACTIONS(1889),
    [aux_sym_directive_token1] = ACTIONS(1887),
    [aux_sym_tag_token1] = ACTIONS(1889),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1328)] = {
    [sym_attribute] = STATE(1328),
    [sym__lowercase_identifier] = ACTIONS(2311),
    [anon_sym_SEMI_SEMI] = ACTIONS(2315),
    [anon_sym_let] = ACTIONS(2311),
    [anon_sym_and] = ACTIONS(2311),
    [anon_sym_TILDE] = ACTIONS(2311),
    [anon_sym_QMARK] = ACTIONS(2311),
    [anon_sym_COLON] = ACTIONS(2576),
    [anon_sym_LPAREN] = ACTIONS(2315),
    [anon_sym_external] = ACTIONS(2311),
    [anon_sym_type] = ACTIONS(2311),
    [anon_sym_COMMA] = ACTIONS(2315),
    [anon_sym_PIPE] = ACTIONS(2311),
    [anon_sym_LBRACK] = ACTIONS(2311),
    [anon_sym_RBRACK] = ACTIONS(2315),
    [anon_sym_true] = ACTIONS(2311),
    [anon_sym_false] = ACTIONS(2311),
    [anon_sym_LBRACE] = ACTIONS(2311),
    [anon_sym_SEMI] = ACTIONS(2311),
    [anon_sym_exception] = ACTIONS(2311),
    [anon_sym_module] = ACTIONS(2311),
    [anon_sym_open] = ACTIONS(2311),
    [anon_sym_include] = ACTIONS(2311),
    [anon_sym_class] = ACTIONS(2311),
    [anon_sym_val] = ACTIONS(2311),
    [anon_sym_end] = ACTIONS(2311),
    [anon_sym_object] = ACTIONS(2311),
    [anon_sym_in] = ACTIONS(2311),
    [anon_sym_AMP] = ACTIONS(2311),
    [anon_sym_COLON_COLON] = ACTIONS(2315),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2315),
    [anon_sym_else] = ACTIONS(2311),
    [anon_sym_new] = ACTIONS(2311),
    [anon_sym_LBRACE_LT] = ACTIONS(2315),
    [anon_sym_begin] = ACTIONS(2311),
    [sym_ocamlyacc_value] = ACTIONS(2315),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2311),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2315),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2311),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2311),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2315),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2315),
    [aux_sym_number_token1] = ACTIONS(2315),
    [aux_sym_signed_number_token1] = ACTIONS(2311),
    [anon_sym_SQUOTE] = ACTIONS(2315),
    [anon_sym_DQUOTE] = ACTIONS(2315),
    [sym_prefix_operator] = ACTIONS(2315),
    [aux_sym_sign_operator_token1] = ACTIONS(2311),
    [anon_sym_lsl] = ACTIONS(2311),
    [anon_sym_lsr] = ACTIONS(2311),
    [anon_sym_asr] = ACTIONS(2311),
    [aux_sym_pow_operator_token1] = ACTIONS(2315),
    [anon_sym_mod] = ACTIONS(2311),
    [anon_sym_land] = ACTIONS(2311),
    [anon_sym_lor] = ACTIONS(2311),
    [anon_sym_lxor] = ACTIONS(2311),
    [aux_sym_mult_operator_token1] = ACTIONS(2311),
    [aux_sym_add_operator_token1] = ACTIONS(2311),
    [sym_concat_operator] = ACTIONS(2315),
    [sym_rel_operator] = ACTIONS(2311),
    [anon_sym_AMP_AMP] = ACTIONS(2311),
    [anon_sym_or] = ACTIONS(2311),
    [anon_sym_PIPE_PIPE] = ACTIONS(2311),
    [sym_assign_operator] = ACTIONS(2315),
    [sym_let_operator] = ACTIONS(2315),
    [sym_let_and_operator] = ACTIONS(2315),
    [sym__uppercase_identifier] = ACTIONS(2315),
    [aux_sym_directive_token1] = ACTIONS(2315),
    [aux_sym_tag_token1] = ACTIONS(2315),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1329)] = {
    [sym_attribute] = STATE(1329),
    [sym__lowercase_identifier] = ACTIONS(1923),
    [anon_sym_SEMI_SEMI] = ACTIONS(1925),
    [anon_sym_let] = ACTIONS(1923),
    [anon_sym_and] = ACTIONS(1923),
    [anon_sym_TILDE] = ACTIONS(1923),
    [anon_sym_QMARK] = ACTIONS(1923),
    [anon_sym_LPAREN] = ACTIONS(1925),
    [anon_sym_external] = ACTIONS(1923),
    [anon_sym_type] = ACTIONS(1923),
    [anon_sym_COMMA] = ACTIONS(1925),
    [anon_sym_PIPE] = ACTIONS(1923),
    [anon_sym_DOT] = ACTIONS(1925),
    [anon_sym_LBRACK] = ACTIONS(1923),
    [anon_sym_RBRACK] = ACTIONS(1925),
    [anon_sym_true] = ACTIONS(1923),
    [anon_sym_false] = ACTIONS(1923),
    [anon_sym_LBRACE] = ACTIONS(1923),
    [anon_sym_SEMI] = ACTIONS(1923),
    [anon_sym_exception] = ACTIONS(1923),
    [anon_sym_module] = ACTIONS(1923),
    [anon_sym_open] = ACTIONS(1923),
    [anon_sym_include] = ACTIONS(1923),
    [anon_sym_class] = ACTIONS(1923),
    [anon_sym_val] = ACTIONS(1923),
    [anon_sym_end] = ACTIONS(1923),
    [anon_sym_object] = ACTIONS(1923),
    [anon_sym_AMP] = ACTIONS(1923),
    [anon_sym_POUND] = ACTIONS(1923),
    [anon_sym_COLON_COLON] = ACTIONS(1925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1925),
    [anon_sym_else] = ACTIONS(1923),
    [anon_sym_new] = ACTIONS(1923),
    [anon_sym_LBRACE_LT] = ACTIONS(1925),
    [anon_sym_begin] = ACTIONS(1923),
    [sym_ocamlyacc_value] = ACTIONS(1925),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1923),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1925),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1923),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1923),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1925),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1925),
    [aux_sym_number_token1] = ACTIONS(1925),
    [aux_sym_signed_number_token1] = ACTIONS(1923),
    [anon_sym_SQUOTE] = ACTIONS(1925),
    [anon_sym_DQUOTE] = ACTIONS(1925),
    [sym_prefix_operator] = ACTIONS(1925),
    [aux_sym_sign_operator_token1] = ACTIONS(1923),
    [sym_hash_operator] = ACTIONS(1925),
    [anon_sym_lsl] = ACTIONS(1923),
    [anon_sym_lsr] = ACTIONS(1923),
    [anon_sym_asr] = ACTIONS(1923),
    [aux_sym_pow_operator_token1] = ACTIONS(1925),
    [anon_sym_mod] = ACTIONS(1923),
    [anon_sym_land] = ACTIONS(1923),
    [anon_sym_lor] = ACTIONS(1923),
    [anon_sym_lxor] = ACTIONS(1923),
    [aux_sym_mult_operator_token1] = ACTIONS(1923),
    [aux_sym_add_operator_token1] = ACTIONS(1923),
    [sym_concat_operator] = ACTIONS(1925),
    [sym_rel_operator] = ACTIONS(1923),
    [anon_sym_AMP_AMP] = ACTIONS(1923),
    [anon_sym_or] = ACTIONS(1923),
    [anon_sym_PIPE_PIPE] = ACTIONS(1923),
    [sym_assign_operator] = ACTIONS(1925),
    [sym_let_operator] = ACTIONS(1925),
    [sym__uppercase_identifier] = ACTIONS(1925),
    [aux_sym_directive_token1] = ACTIONS(1923),
    [aux_sym_tag_token1] = ACTIONS(1925),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1330)] = {
    [sym_attribute] = STATE(1330),
    [sym__lowercase_identifier] = ACTIONS(1971),
    [anon_sym_SEMI_SEMI] = ACTIONS(1973),
    [anon_sym_let] = ACTIONS(1971),
    [anon_sym_and] = ACTIONS(1971),
    [anon_sym_TILDE] = ACTIONS(1971),
    [anon_sym_QMARK] = ACTIONS(1971),
    [anon_sym_LPAREN] = ACTIONS(1973),
    [anon_sym_external] = ACTIONS(1971),
    [anon_sym_type] = ACTIONS(1971),
    [anon_sym_COMMA] = ACTIONS(1973),
    [anon_sym_PIPE] = ACTIONS(1971),
    [anon_sym_DOT] = ACTIONS(1087),
    [anon_sym_LBRACK] = ACTIONS(1971),
    [anon_sym_RBRACK] = ACTIONS(1973),
    [anon_sym_true] = ACTIONS(1971),
    [anon_sym_false] = ACTIONS(1971),
    [anon_sym_LBRACE] = ACTIONS(1971),
    [anon_sym_SEMI] = ACTIONS(1971),
    [anon_sym_exception] = ACTIONS(1971),
    [anon_sym_module] = ACTIONS(1971),
    [anon_sym_open] = ACTIONS(1971),
    [anon_sym_include] = ACTIONS(1971),
    [anon_sym_class] = ACTIONS(1971),
    [anon_sym_val] = ACTIONS(1971),
    [anon_sym_end] = ACTIONS(1971),
    [anon_sym_object] = ACTIONS(1971),
    [anon_sym_AMP] = ACTIONS(1971),
    [anon_sym_POUND] = ACTIONS(1971),
    [anon_sym_COLON_COLON] = ACTIONS(1973),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1973),
    [anon_sym_else] = ACTIONS(1971),
    [anon_sym_new] = ACTIONS(1971),
    [anon_sym_LBRACE_LT] = ACTIONS(1973),
    [anon_sym_begin] = ACTIONS(1971),
    [sym_ocamlyacc_value] = ACTIONS(1973),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1971),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1973),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1973),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1973),
    [aux_sym_number_token1] = ACTIONS(1973),
    [aux_sym_signed_number_token1] = ACTIONS(1971),
    [anon_sym_SQUOTE] = ACTIONS(1973),
    [anon_sym_DQUOTE] = ACTIONS(1973),
    [sym_prefix_operator] = ACTIONS(1973),
    [aux_sym_sign_operator_token1] = ACTIONS(1971),
    [sym_hash_operator] = ACTIONS(1973),
    [anon_sym_lsl] = ACTIONS(1971),
    [anon_sym_lsr] = ACTIONS(1971),
    [anon_sym_asr] = ACTIONS(1971),
    [aux_sym_pow_operator_token1] = ACTIONS(1973),
    [anon_sym_mod] = ACTIONS(1971),
    [anon_sym_land] = ACTIONS(1971),
    [anon_sym_lor] = ACTIONS(1971),
    [anon_sym_lxor] = ACTIONS(1971),
    [aux_sym_mult_operator_token1] = ACTIONS(1971),
    [aux_sym_add_operator_token1] = ACTIONS(1971),
    [sym_concat_operator] = ACTIONS(1973),
    [sym_rel_operator] = ACTIONS(1971),
    [anon_sym_AMP_AMP] = ACTIONS(1971),
    [anon_sym_or] = ACTIONS(1971),
    [anon_sym_PIPE_PIPE] = ACTIONS(1971),
    [sym_assign_operator] = ACTIONS(1973),
    [sym_let_operator] = ACTIONS(1973),
    [sym__uppercase_identifier] = ACTIONS(1973),
    [aux_sym_directive_token1] = ACTIONS(1971),
    [aux_sym_tag_token1] = ACTIONS(1973),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1331)] = {
    [sym_attribute] = STATE(1331),
    [sym__lowercase_identifier] = ACTIONS(2035),
    [anon_sym_SEMI_SEMI] = ACTIONS(2037),
    [anon_sym_let] = ACTIONS(2035),
    [anon_sym_and] = ACTIONS(2035),
    [anon_sym_TILDE] = ACTIONS(2035),
    [anon_sym_QMARK] = ACTIONS(2035),
    [anon_sym_LPAREN] = ACTIONS(2037),
    [anon_sym_external] = ACTIONS(2035),
    [anon_sym_type] = ACTIONS(2035),
    [anon_sym_COMMA] = ACTIONS(2037),
    [anon_sym_PIPE] = ACTIONS(2035),
    [anon_sym_DOT] = ACTIONS(2037),
    [anon_sym_LBRACK] = ACTIONS(2035),
    [anon_sym_RBRACK] = ACTIONS(2037),
    [anon_sym_true] = ACTIONS(2035),
    [anon_sym_false] = ACTIONS(2035),
    [anon_sym_LBRACE] = ACTIONS(2035),
    [anon_sym_SEMI] = ACTIONS(2035),
    [anon_sym_exception] = ACTIONS(2035),
    [anon_sym_module] = ACTIONS(2035),
    [anon_sym_open] = ACTIONS(2035),
    [anon_sym_include] = ACTIONS(2035),
    [anon_sym_class] = ACTIONS(2035),
    [anon_sym_val] = ACTIONS(2035),
    [anon_sym_end] = ACTIONS(2035),
    [anon_sym_object] = ACTIONS(2035),
    [anon_sym_AMP] = ACTIONS(2035),
    [anon_sym_POUND] = ACTIONS(2035),
    [anon_sym_COLON_COLON] = ACTIONS(2037),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2037),
    [anon_sym_else] = ACTIONS(2035),
    [anon_sym_new] = ACTIONS(2035),
    [anon_sym_LBRACE_LT] = ACTIONS(2037),
    [anon_sym_begin] = ACTIONS(2035),
    [sym_ocamlyacc_value] = ACTIONS(2037),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2035),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2037),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2035),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2035),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2037),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2037),
    [aux_sym_number_token1] = ACTIONS(2037),
    [aux_sym_signed_number_token1] = ACTIONS(2035),
    [anon_sym_SQUOTE] = ACTIONS(2037),
    [anon_sym_DQUOTE] = ACTIONS(2037),
    [sym_prefix_operator] = ACTIONS(2037),
    [aux_sym_sign_operator_token1] = ACTIONS(2035),
    [sym_hash_operator] = ACTIONS(2037),
    [anon_sym_lsl] = ACTIONS(2035),
    [anon_sym_lsr] = ACTIONS(2035),
    [anon_sym_asr] = ACTIONS(2035),
    [aux_sym_pow_operator_token1] = ACTIONS(2037),
    [anon_sym_mod] = ACTIONS(2035),
    [anon_sym_land] = ACTIONS(2035),
    [anon_sym_lor] = ACTIONS(2035),
    [anon_sym_lxor] = ACTIONS(2035),
    [aux_sym_mult_operator_token1] = ACTIONS(2035),
    [aux_sym_add_operator_token1] = ACTIONS(2035),
    [sym_concat_operator] = ACTIONS(2037),
    [sym_rel_operator] = ACTIONS(2035),
    [anon_sym_AMP_AMP] = ACTIONS(2035),
    [anon_sym_or] = ACTIONS(2035),
    [anon_sym_PIPE_PIPE] = ACTIONS(2035),
    [sym_assign_operator] = ACTIONS(2037),
    [sym_let_operator] = ACTIONS(2037),
    [sym__uppercase_identifier] = ACTIONS(2037),
    [aux_sym_directive_token1] = ACTIONS(2035),
    [aux_sym_tag_token1] = ACTIONS(2037),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1332)] = {
    [sym_attribute] = STATE(1332),
    [sym__lowercase_identifier] = ACTIONS(2103),
    [anon_sym_SEMI_SEMI] = ACTIONS(2105),
    [anon_sym_let] = ACTIONS(2103),
    [anon_sym_and] = ACTIONS(2103),
    [anon_sym_TILDE] = ACTIONS(2103),
    [anon_sym_QMARK] = ACTIONS(2103),
    [anon_sym_LPAREN] = ACTIONS(2105),
    [anon_sym_external] = ACTIONS(2103),
    [anon_sym_type] = ACTIONS(2103),
    [anon_sym_COMMA] = ACTIONS(2105),
    [anon_sym_PIPE] = ACTIONS(2103),
    [anon_sym_DOT] = ACTIONS(2105),
    [anon_sym_LBRACK] = ACTIONS(2103),
    [anon_sym_RBRACK] = ACTIONS(2105),
    [anon_sym_true] = ACTIONS(2103),
    [anon_sym_false] = ACTIONS(2103),
    [anon_sym_LBRACE] = ACTIONS(2103),
    [anon_sym_SEMI] = ACTIONS(2103),
    [anon_sym_exception] = ACTIONS(2103),
    [anon_sym_module] = ACTIONS(2103),
    [anon_sym_open] = ACTIONS(2103),
    [anon_sym_include] = ACTIONS(2103),
    [anon_sym_class] = ACTIONS(2103),
    [anon_sym_val] = ACTIONS(2103),
    [anon_sym_end] = ACTIONS(2103),
    [anon_sym_object] = ACTIONS(2103),
    [anon_sym_AMP] = ACTIONS(2103),
    [anon_sym_POUND] = ACTIONS(2103),
    [anon_sym_COLON_COLON] = ACTIONS(2105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2105),
    [anon_sym_else] = ACTIONS(2103),
    [anon_sym_new] = ACTIONS(2103),
    [anon_sym_LBRACE_LT] = ACTIONS(2105),
    [anon_sym_begin] = ACTIONS(2103),
    [sym_ocamlyacc_value] = ACTIONS(2105),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2103),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2105),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2103),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2103),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2105),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2105),
    [aux_sym_number_token1] = ACTIONS(2105),
    [aux_sym_signed_number_token1] = ACTIONS(2103),
    [anon_sym_SQUOTE] = ACTIONS(2105),
    [anon_sym_DQUOTE] = ACTIONS(2105),
    [sym_prefix_operator] = ACTIONS(2105),
    [aux_sym_sign_operator_token1] = ACTIONS(2103),
    [sym_hash_operator] = ACTIONS(2105),
    [anon_sym_lsl] = ACTIONS(2103),
    [anon_sym_lsr] = ACTIONS(2103),
    [anon_sym_asr] = ACTIONS(2103),
    [aux_sym_pow_operator_token1] = ACTIONS(2105),
    [anon_sym_mod] = ACTIONS(2103),
    [anon_sym_land] = ACTIONS(2103),
    [anon_sym_lor] = ACTIONS(2103),
    [anon_sym_lxor] = ACTIONS(2103),
    [aux_sym_mult_operator_token1] = ACTIONS(2103),
    [aux_sym_add_operator_token1] = ACTIONS(2103),
    [sym_concat_operator] = ACTIONS(2105),
    [sym_rel_operator] = ACTIONS(2103),
    [anon_sym_AMP_AMP] = ACTIONS(2103),
    [anon_sym_or] = ACTIONS(2103),
    [anon_sym_PIPE_PIPE] = ACTIONS(2103),
    [sym_assign_operator] = ACTIONS(2105),
    [sym_let_operator] = ACTIONS(2105),
    [sym__uppercase_identifier] = ACTIONS(2105),
    [aux_sym_directive_token1] = ACTIONS(2103),
    [aux_sym_tag_token1] = ACTIONS(2105),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1333)] = {
    [sym_attribute] = STATE(1333),
    [sym__lowercase_identifier] = ACTIONS(2195),
    [anon_sym_SEMI_SEMI] = ACTIONS(2197),
    [anon_sym_let] = ACTIONS(2195),
    [anon_sym_and] = ACTIONS(2195),
    [anon_sym_TILDE] = ACTIONS(2195),
    [anon_sym_QMARK] = ACTIONS(2195),
    [anon_sym_LPAREN] = ACTIONS(2197),
    [anon_sym_external] = ACTIONS(2195),
    [anon_sym_type] = ACTIONS(2195),
    [anon_sym_COMMA] = ACTIONS(2197),
    [anon_sym_PIPE] = ACTIONS(2195),
    [anon_sym_DOT] = ACTIONS(2197),
    [anon_sym_LBRACK] = ACTIONS(2195),
    [anon_sym_RBRACK] = ACTIONS(2197),
    [anon_sym_true] = ACTIONS(2195),
    [anon_sym_false] = ACTIONS(2195),
    [anon_sym_LBRACE] = ACTIONS(2195),
    [anon_sym_SEMI] = ACTIONS(2195),
    [anon_sym_exception] = ACTIONS(2195),
    [anon_sym_module] = ACTIONS(2195),
    [anon_sym_open] = ACTIONS(2195),
    [anon_sym_include] = ACTIONS(2195),
    [anon_sym_class] = ACTIONS(2195),
    [anon_sym_val] = ACTIONS(2195),
    [anon_sym_end] = ACTIONS(2195),
    [anon_sym_object] = ACTIONS(2195),
    [anon_sym_AMP] = ACTIONS(2195),
    [anon_sym_POUND] = ACTIONS(2195),
    [anon_sym_COLON_COLON] = ACTIONS(2197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2197),
    [anon_sym_else] = ACTIONS(2195),
    [anon_sym_new] = ACTIONS(2195),
    [anon_sym_LBRACE_LT] = ACTIONS(2197),
    [anon_sym_begin] = ACTIONS(2195),
    [sym_ocamlyacc_value] = ACTIONS(2197),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2195),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2197),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2195),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2195),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2197),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2197),
    [aux_sym_number_token1] = ACTIONS(2197),
    [aux_sym_signed_number_token1] = ACTIONS(2195),
    [anon_sym_SQUOTE] = ACTIONS(2197),
    [anon_sym_DQUOTE] = ACTIONS(2197),
    [sym_prefix_operator] = ACTIONS(2197),
    [aux_sym_sign_operator_token1] = ACTIONS(2195),
    [sym_hash_operator] = ACTIONS(2197),
    [anon_sym_lsl] = ACTIONS(2195),
    [anon_sym_lsr] = ACTIONS(2195),
    [anon_sym_asr] = ACTIONS(2195),
    [aux_sym_pow_operator_token1] = ACTIONS(2197),
    [anon_sym_mod] = ACTIONS(2195),
    [anon_sym_land] = ACTIONS(2195),
    [anon_sym_lor] = ACTIONS(2195),
    [anon_sym_lxor] = ACTIONS(2195),
    [aux_sym_mult_operator_token1] = ACTIONS(2195),
    [aux_sym_add_operator_token1] = ACTIONS(2195),
    [sym_concat_operator] = ACTIONS(2197),
    [sym_rel_operator] = ACTIONS(2195),
    [anon_sym_AMP_AMP] = ACTIONS(2195),
    [anon_sym_or] = ACTIONS(2195),
    [anon_sym_PIPE_PIPE] = ACTIONS(2195),
    [sym_assign_operator] = ACTIONS(2197),
    [sym_let_operator] = ACTIONS(2197),
    [sym__uppercase_identifier] = ACTIONS(2197),
    [aux_sym_directive_token1] = ACTIONS(2195),
    [aux_sym_tag_token1] = ACTIONS(2197),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1334)] = {
    [sym_attribute] = STATE(1334),
    [sym__lowercase_identifier] = ACTIONS(2107),
    [anon_sym_SEMI_SEMI] = ACTIONS(2109),
    [anon_sym_let] = ACTIONS(2107),
    [anon_sym_and] = ACTIONS(2107),
    [anon_sym_TILDE] = ACTIONS(2107),
    [anon_sym_QMARK] = ACTIONS(2107),
    [anon_sym_LPAREN] = ACTIONS(2109),
    [anon_sym_external] = ACTIONS(2107),
    [anon_sym_type] = ACTIONS(2107),
    [anon_sym_COMMA] = ACTIONS(2109),
    [anon_sym_PIPE] = ACTIONS(2107),
    [anon_sym_DOT] = ACTIONS(2109),
    [anon_sym_LBRACK] = ACTIONS(2107),
    [anon_sym_RBRACK] = ACTIONS(2109),
    [anon_sym_true] = ACTIONS(2107),
    [anon_sym_false] = ACTIONS(2107),
    [anon_sym_LBRACE] = ACTIONS(2107),
    [anon_sym_SEMI] = ACTIONS(2107),
    [anon_sym_exception] = ACTIONS(2107),
    [anon_sym_module] = ACTIONS(2107),
    [anon_sym_open] = ACTIONS(2107),
    [anon_sym_include] = ACTIONS(2107),
    [anon_sym_class] = ACTIONS(2107),
    [anon_sym_val] = ACTIONS(2107),
    [anon_sym_end] = ACTIONS(2107),
    [anon_sym_object] = ACTIONS(2107),
    [anon_sym_AMP] = ACTIONS(2107),
    [anon_sym_POUND] = ACTIONS(2107),
    [anon_sym_COLON_COLON] = ACTIONS(2109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2109),
    [anon_sym_else] = ACTIONS(2107),
    [anon_sym_new] = ACTIONS(2107),
    [anon_sym_LBRACE_LT] = ACTIONS(2109),
    [anon_sym_begin] = ACTIONS(2107),
    [sym_ocamlyacc_value] = ACTIONS(2109),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2107),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2109),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2109),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2109),
    [aux_sym_number_token1] = ACTIONS(2109),
    [aux_sym_signed_number_token1] = ACTIONS(2107),
    [anon_sym_SQUOTE] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(2109),
    [sym_prefix_operator] = ACTIONS(2109),
    [aux_sym_sign_operator_token1] = ACTIONS(2107),
    [sym_hash_operator] = ACTIONS(2109),
    [anon_sym_lsl] = ACTIONS(2107),
    [anon_sym_lsr] = ACTIONS(2107),
    [anon_sym_asr] = ACTIONS(2107),
    [aux_sym_pow_operator_token1] = ACTIONS(2109),
    [anon_sym_mod] = ACTIONS(2107),
    [anon_sym_land] = ACTIONS(2107),
    [anon_sym_lor] = ACTIONS(2107),
    [anon_sym_lxor] = ACTIONS(2107),
    [aux_sym_mult_operator_token1] = ACTIONS(2107),
    [aux_sym_add_operator_token1] = ACTIONS(2107),
    [sym_concat_operator] = ACTIONS(2109),
    [sym_rel_operator] = ACTIONS(2107),
    [anon_sym_AMP_AMP] = ACTIONS(2107),
    [anon_sym_or] = ACTIONS(2107),
    [anon_sym_PIPE_PIPE] = ACTIONS(2107),
    [sym_assign_operator] = ACTIONS(2109),
    [sym_let_operator] = ACTIONS(2109),
    [sym__uppercase_identifier] = ACTIONS(2109),
    [aux_sym_directive_token1] = ACTIONS(2107),
    [aux_sym_tag_token1] = ACTIONS(2109),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1335)] = {
    [sym_attribute] = STATE(1335),
    [ts_builtin_sym_end] = ACTIONS(1913),
    [sym__lowercase_identifier] = ACTIONS(1911),
    [anon_sym_SEMI_SEMI] = ACTIONS(1913),
    [anon_sym_let] = ACTIONS(1911),
    [anon_sym_and] = ACTIONS(1911),
    [anon_sym_TILDE] = ACTIONS(1911),
    [anon_sym_QMARK] = ACTIONS(1911),
    [anon_sym_LPAREN] = ACTIONS(1913),
    [anon_sym_external] = ACTIONS(1911),
    [anon_sym_type] = ACTIONS(1911),
    [anon_sym_COMMA] = ACTIONS(1913),
    [anon_sym_PIPE] = ACTIONS(1911),
    [anon_sym_DOT] = ACTIONS(1913),
    [anon_sym_LBRACK] = ACTIONS(1911),
    [anon_sym_true] = ACTIONS(1911),
    [anon_sym_false] = ACTIONS(1911),
    [anon_sym_LBRACE] = ACTIONS(1911),
    [anon_sym_SEMI] = ACTIONS(1911),
    [anon_sym_exception] = ACTIONS(1911),
    [anon_sym_module] = ACTIONS(1911),
    [anon_sym_open] = ACTIONS(1911),
    [anon_sym_include] = ACTIONS(1911),
    [anon_sym_class] = ACTIONS(1911),
    [anon_sym_val] = ACTIONS(1911),
    [anon_sym_object] = ACTIONS(1911),
    [anon_sym_AMP] = ACTIONS(1911),
    [anon_sym_POUND] = ACTIONS(1911),
    [anon_sym_COLON_COLON] = ACTIONS(1913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1913),
    [anon_sym_LT_DASH] = ACTIONS(1911),
    [anon_sym_else] = ACTIONS(1911),
    [anon_sym_new] = ACTIONS(1911),
    [anon_sym_LBRACE_LT] = ACTIONS(1913),
    [anon_sym_begin] = ACTIONS(1911),
    [sym_ocamlyacc_value] = ACTIONS(1913),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1911),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1913),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1911),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1911),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1913),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1913),
    [aux_sym_number_token1] = ACTIONS(1913),
    [aux_sym_signed_number_token1] = ACTIONS(1911),
    [anon_sym_SQUOTE] = ACTIONS(1913),
    [anon_sym_DQUOTE] = ACTIONS(1913),
    [sym_prefix_operator] = ACTIONS(1913),
    [aux_sym_sign_operator_token1] = ACTIONS(1911),
    [sym_hash_operator] = ACTIONS(1913),
    [anon_sym_lsl] = ACTIONS(1911),
    [anon_sym_lsr] = ACTIONS(1911),
    [anon_sym_asr] = ACTIONS(1911),
    [aux_sym_pow_operator_token1] = ACTIONS(1913),
    [anon_sym_mod] = ACTIONS(1911),
    [anon_sym_land] = ACTIONS(1911),
    [anon_sym_lor] = ACTIONS(1911),
    [anon_sym_lxor] = ACTIONS(1911),
    [aux_sym_mult_operator_token1] = ACTIONS(1911),
    [aux_sym_add_operator_token1] = ACTIONS(1911),
    [sym_concat_operator] = ACTIONS(1913),
    [sym_rel_operator] = ACTIONS(1911),
    [anon_sym_AMP_AMP] = ACTIONS(1911),
    [anon_sym_or] = ACTIONS(1911),
    [anon_sym_PIPE_PIPE] = ACTIONS(1911),
    [sym_assign_operator] = ACTIONS(1913),
    [sym_let_operator] = ACTIONS(1913),
    [sym__uppercase_identifier] = ACTIONS(1913),
    [aux_sym_directive_token1] = ACTIONS(1911),
    [aux_sym_tag_token1] = ACTIONS(1913),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1336)] = {
    [sym_attribute] = STATE(1336),
    [sym__lowercase_identifier] = ACTIONS(2117),
    [anon_sym_SEMI_SEMI] = ACTIONS(2119),
    [anon_sym_let] = ACTIONS(2117),
    [anon_sym_and] = ACTIONS(2117),
    [anon_sym_TILDE] = ACTIONS(2117),
    [anon_sym_QMARK] = ACTIONS(2117),
    [anon_sym_LPAREN] = ACTIONS(2119),
    [anon_sym_external] = ACTIONS(2117),
    [anon_sym_type] = ACTIONS(2117),
    [anon_sym_COMMA] = ACTIONS(2119),
    [anon_sym_PIPE] = ACTIONS(2117),
    [anon_sym_DOT] = ACTIONS(2119),
    [anon_sym_LBRACK] = ACTIONS(2117),
    [anon_sym_RBRACK] = ACTIONS(2119),
    [anon_sym_true] = ACTIONS(2117),
    [anon_sym_false] = ACTIONS(2117),
    [anon_sym_LBRACE] = ACTIONS(2117),
    [anon_sym_SEMI] = ACTIONS(2117),
    [anon_sym_exception] = ACTIONS(2117),
    [anon_sym_module] = ACTIONS(2117),
    [anon_sym_open] = ACTIONS(2117),
    [anon_sym_include] = ACTIONS(2117),
    [anon_sym_class] = ACTIONS(2117),
    [anon_sym_val] = ACTIONS(2117),
    [anon_sym_end] = ACTIONS(2117),
    [anon_sym_object] = ACTIONS(2117),
    [anon_sym_AMP] = ACTIONS(2117),
    [anon_sym_POUND] = ACTIONS(2117),
    [anon_sym_COLON_COLON] = ACTIONS(2119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2119),
    [anon_sym_else] = ACTIONS(2117),
    [anon_sym_new] = ACTIONS(2117),
    [anon_sym_LBRACE_LT] = ACTIONS(2119),
    [anon_sym_begin] = ACTIONS(2117),
    [sym_ocamlyacc_value] = ACTIONS(2119),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2117),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2119),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2117),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2117),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2119),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2119),
    [aux_sym_number_token1] = ACTIONS(2119),
    [aux_sym_signed_number_token1] = ACTIONS(2117),
    [anon_sym_SQUOTE] = ACTIONS(2119),
    [anon_sym_DQUOTE] = ACTIONS(2119),
    [sym_prefix_operator] = ACTIONS(2119),
    [aux_sym_sign_operator_token1] = ACTIONS(2117),
    [sym_hash_operator] = ACTIONS(2119),
    [anon_sym_lsl] = ACTIONS(2117),
    [anon_sym_lsr] = ACTIONS(2117),
    [anon_sym_asr] = ACTIONS(2117),
    [aux_sym_pow_operator_token1] = ACTIONS(2119),
    [anon_sym_mod] = ACTIONS(2117),
    [anon_sym_land] = ACTIONS(2117),
    [anon_sym_lor] = ACTIONS(2117),
    [anon_sym_lxor] = ACTIONS(2117),
    [aux_sym_mult_operator_token1] = ACTIONS(2117),
    [aux_sym_add_operator_token1] = ACTIONS(2117),
    [sym_concat_operator] = ACTIONS(2119),
    [sym_rel_operator] = ACTIONS(2117),
    [anon_sym_AMP_AMP] = ACTIONS(2117),
    [anon_sym_or] = ACTIONS(2117),
    [anon_sym_PIPE_PIPE] = ACTIONS(2117),
    [sym_assign_operator] = ACTIONS(2119),
    [sym_let_operator] = ACTIONS(2119),
    [sym__uppercase_identifier] = ACTIONS(2119),
    [aux_sym_directive_token1] = ACTIONS(2117),
    [aux_sym_tag_token1] = ACTIONS(2119),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1337)] = {
    [sym_attribute] = STATE(1337),
    [sym__lowercase_identifier] = ACTIONS(2203),
    [anon_sym_SEMI_SEMI] = ACTIONS(2205),
    [anon_sym_let] = ACTIONS(2203),
    [anon_sym_and] = ACTIONS(2203),
    [anon_sym_TILDE] = ACTIONS(2203),
    [anon_sym_QMARK] = ACTIONS(2203),
    [anon_sym_LPAREN] = ACTIONS(2205),
    [anon_sym_external] = ACTIONS(2203),
    [anon_sym_type] = ACTIONS(2203),
    [anon_sym_COMMA] = ACTIONS(2205),
    [anon_sym_PIPE] = ACTIONS(2203),
    [anon_sym_DOT] = ACTIONS(2205),
    [anon_sym_LBRACK] = ACTIONS(2203),
    [anon_sym_RBRACK] = ACTIONS(2205),
    [anon_sym_true] = ACTIONS(2203),
    [anon_sym_false] = ACTIONS(2203),
    [anon_sym_LBRACE] = ACTIONS(2203),
    [anon_sym_SEMI] = ACTIONS(2203),
    [anon_sym_exception] = ACTIONS(2203),
    [anon_sym_module] = ACTIONS(2203),
    [anon_sym_open] = ACTIONS(2203),
    [anon_sym_include] = ACTIONS(2203),
    [anon_sym_class] = ACTIONS(2203),
    [anon_sym_val] = ACTIONS(2203),
    [anon_sym_end] = ACTIONS(2203),
    [anon_sym_object] = ACTIONS(2203),
    [anon_sym_AMP] = ACTIONS(2203),
    [anon_sym_POUND] = ACTIONS(2203),
    [anon_sym_COLON_COLON] = ACTIONS(2205),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2205),
    [anon_sym_else] = ACTIONS(2203),
    [anon_sym_new] = ACTIONS(2203),
    [anon_sym_LBRACE_LT] = ACTIONS(2205),
    [anon_sym_begin] = ACTIONS(2203),
    [sym_ocamlyacc_value] = ACTIONS(2205),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2203),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2205),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2203),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2203),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2205),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2205),
    [aux_sym_number_token1] = ACTIONS(2205),
    [aux_sym_signed_number_token1] = ACTIONS(2203),
    [anon_sym_SQUOTE] = ACTIONS(2205),
    [anon_sym_DQUOTE] = ACTIONS(2205),
    [sym_prefix_operator] = ACTIONS(2205),
    [aux_sym_sign_operator_token1] = ACTIONS(2203),
    [sym_hash_operator] = ACTIONS(2205),
    [anon_sym_lsl] = ACTIONS(2203),
    [anon_sym_lsr] = ACTIONS(2203),
    [anon_sym_asr] = ACTIONS(2203),
    [aux_sym_pow_operator_token1] = ACTIONS(2205),
    [anon_sym_mod] = ACTIONS(2203),
    [anon_sym_land] = ACTIONS(2203),
    [anon_sym_lor] = ACTIONS(2203),
    [anon_sym_lxor] = ACTIONS(2203),
    [aux_sym_mult_operator_token1] = ACTIONS(2203),
    [aux_sym_add_operator_token1] = ACTIONS(2203),
    [sym_concat_operator] = ACTIONS(2205),
    [sym_rel_operator] = ACTIONS(2203),
    [anon_sym_AMP_AMP] = ACTIONS(2203),
    [anon_sym_or] = ACTIONS(2203),
    [anon_sym_PIPE_PIPE] = ACTIONS(2203),
    [sym_assign_operator] = ACTIONS(2205),
    [sym_let_operator] = ACTIONS(2205),
    [sym__uppercase_identifier] = ACTIONS(2205),
    [aux_sym_directive_token1] = ACTIONS(2203),
    [aux_sym_tag_token1] = ACTIONS(2205),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1338)] = {
    [sym_attribute] = STATE(1338),
    [sym__lowercase_identifier] = ACTIONS(2051),
    [anon_sym_SEMI_SEMI] = ACTIONS(2053),
    [anon_sym_let] = ACTIONS(2051),
    [anon_sym_and] = ACTIONS(2051),
    [anon_sym_TILDE] = ACTIONS(2051),
    [anon_sym_QMARK] = ACTIONS(2051),
    [anon_sym_LPAREN] = ACTIONS(2053),
    [anon_sym_external] = ACTIONS(2051),
    [anon_sym_type] = ACTIONS(2051),
    [anon_sym_COMMA] = ACTIONS(2053),
    [anon_sym_PIPE] = ACTIONS(2051),
    [anon_sym_DOT] = ACTIONS(2053),
    [anon_sym_LBRACK] = ACTIONS(2051),
    [anon_sym_RBRACK] = ACTIONS(2053),
    [anon_sym_true] = ACTIONS(2051),
    [anon_sym_false] = ACTIONS(2051),
    [anon_sym_LBRACE] = ACTIONS(2051),
    [anon_sym_SEMI] = ACTIONS(2051),
    [anon_sym_exception] = ACTIONS(2051),
    [anon_sym_module] = ACTIONS(2051),
    [anon_sym_open] = ACTIONS(2051),
    [anon_sym_include] = ACTIONS(2051),
    [anon_sym_class] = ACTIONS(2051),
    [anon_sym_val] = ACTIONS(2051),
    [anon_sym_end] = ACTIONS(2051),
    [anon_sym_object] = ACTIONS(2051),
    [anon_sym_AMP] = ACTIONS(2051),
    [anon_sym_POUND] = ACTIONS(2051),
    [anon_sym_COLON_COLON] = ACTIONS(2053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2053),
    [anon_sym_else] = ACTIONS(2051),
    [anon_sym_new] = ACTIONS(2051),
    [anon_sym_LBRACE_LT] = ACTIONS(2053),
    [anon_sym_begin] = ACTIONS(2051),
    [sym_ocamlyacc_value] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2051),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2053),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2051),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2051),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2053),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2053),
    [aux_sym_number_token1] = ACTIONS(2053),
    [aux_sym_signed_number_token1] = ACTIONS(2051),
    [anon_sym_SQUOTE] = ACTIONS(2053),
    [anon_sym_DQUOTE] = ACTIONS(2053),
    [sym_prefix_operator] = ACTIONS(2053),
    [aux_sym_sign_operator_token1] = ACTIONS(2051),
    [sym_hash_operator] = ACTIONS(2053),
    [anon_sym_lsl] = ACTIONS(2051),
    [anon_sym_lsr] = ACTIONS(2051),
    [anon_sym_asr] = ACTIONS(2051),
    [aux_sym_pow_operator_token1] = ACTIONS(2053),
    [anon_sym_mod] = ACTIONS(2051),
    [anon_sym_land] = ACTIONS(2051),
    [anon_sym_lor] = ACTIONS(2051),
    [anon_sym_lxor] = ACTIONS(2051),
    [aux_sym_mult_operator_token1] = ACTIONS(2051),
    [aux_sym_add_operator_token1] = ACTIONS(2051),
    [sym_concat_operator] = ACTIONS(2053),
    [sym_rel_operator] = ACTIONS(2051),
    [anon_sym_AMP_AMP] = ACTIONS(2051),
    [anon_sym_or] = ACTIONS(2051),
    [anon_sym_PIPE_PIPE] = ACTIONS(2051),
    [sym_assign_operator] = ACTIONS(2053),
    [sym_let_operator] = ACTIONS(2053),
    [sym__uppercase_identifier] = ACTIONS(2053),
    [aux_sym_directive_token1] = ACTIONS(2051),
    [aux_sym_tag_token1] = ACTIONS(2053),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1339)] = {
    [sym_attribute] = STATE(1339),
    [sym__lowercase_identifier] = ACTIONS(2055),
    [anon_sym_SEMI_SEMI] = ACTIONS(2057),
    [anon_sym_let] = ACTIONS(2055),
    [anon_sym_and] = ACTIONS(2055),
    [anon_sym_TILDE] = ACTIONS(2055),
    [anon_sym_QMARK] = ACTIONS(2055),
    [anon_sym_LPAREN] = ACTIONS(2057),
    [anon_sym_external] = ACTIONS(2055),
    [anon_sym_type] = ACTIONS(2055),
    [anon_sym_COMMA] = ACTIONS(2057),
    [anon_sym_PIPE] = ACTIONS(2055),
    [anon_sym_DOT] = ACTIONS(2057),
    [anon_sym_LBRACK] = ACTIONS(2055),
    [anon_sym_RBRACK] = ACTIONS(2057),
    [anon_sym_true] = ACTIONS(2055),
    [anon_sym_false] = ACTIONS(2055),
    [anon_sym_LBRACE] = ACTIONS(2055),
    [anon_sym_SEMI] = ACTIONS(2055),
    [anon_sym_exception] = ACTIONS(2055),
    [anon_sym_module] = ACTIONS(2055),
    [anon_sym_open] = ACTIONS(2055),
    [anon_sym_include] = ACTIONS(2055),
    [anon_sym_class] = ACTIONS(2055),
    [anon_sym_val] = ACTIONS(2055),
    [anon_sym_end] = ACTIONS(2055),
    [anon_sym_object] = ACTIONS(2055),
    [anon_sym_AMP] = ACTIONS(2055),
    [anon_sym_POUND] = ACTIONS(2055),
    [anon_sym_COLON_COLON] = ACTIONS(2057),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2057),
    [anon_sym_else] = ACTIONS(2055),
    [anon_sym_new] = ACTIONS(2055),
    [anon_sym_LBRACE_LT] = ACTIONS(2057),
    [anon_sym_begin] = ACTIONS(2055),
    [sym_ocamlyacc_value] = ACTIONS(2057),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2057),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2055),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2055),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2057),
    [aux_sym_number_token1] = ACTIONS(2057),
    [aux_sym_signed_number_token1] = ACTIONS(2055),
    [anon_sym_SQUOTE] = ACTIONS(2057),
    [anon_sym_DQUOTE] = ACTIONS(2057),
    [sym_prefix_operator] = ACTIONS(2057),
    [aux_sym_sign_operator_token1] = ACTIONS(2055),
    [sym_hash_operator] = ACTIONS(2057),
    [anon_sym_lsl] = ACTIONS(2055),
    [anon_sym_lsr] = ACTIONS(2055),
    [anon_sym_asr] = ACTIONS(2055),
    [aux_sym_pow_operator_token1] = ACTIONS(2057),
    [anon_sym_mod] = ACTIONS(2055),
    [anon_sym_land] = ACTIONS(2055),
    [anon_sym_lor] = ACTIONS(2055),
    [anon_sym_lxor] = ACTIONS(2055),
    [aux_sym_mult_operator_token1] = ACTIONS(2055),
    [aux_sym_add_operator_token1] = ACTIONS(2055),
    [sym_concat_operator] = ACTIONS(2057),
    [sym_rel_operator] = ACTIONS(2055),
    [anon_sym_AMP_AMP] = ACTIONS(2055),
    [anon_sym_or] = ACTIONS(2055),
    [anon_sym_PIPE_PIPE] = ACTIONS(2055),
    [sym_assign_operator] = ACTIONS(2057),
    [sym_let_operator] = ACTIONS(2057),
    [sym__uppercase_identifier] = ACTIONS(2057),
    [aux_sym_directive_token1] = ACTIONS(2055),
    [aux_sym_tag_token1] = ACTIONS(2057),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1340)] = {
    [sym_attribute] = STATE(1340),
    [sym__lowercase_identifier] = ACTIONS(2059),
    [anon_sym_SEMI_SEMI] = ACTIONS(2061),
    [anon_sym_let] = ACTIONS(2059),
    [anon_sym_and] = ACTIONS(2059),
    [anon_sym_TILDE] = ACTIONS(2059),
    [anon_sym_QMARK] = ACTIONS(2059),
    [anon_sym_LPAREN] = ACTIONS(2061),
    [anon_sym_external] = ACTIONS(2059),
    [anon_sym_type] = ACTIONS(2059),
    [anon_sym_COMMA] = ACTIONS(2061),
    [anon_sym_PIPE] = ACTIONS(2059),
    [anon_sym_DOT] = ACTIONS(2061),
    [anon_sym_LBRACK] = ACTIONS(2059),
    [anon_sym_RBRACK] = ACTIONS(2061),
    [anon_sym_true] = ACTIONS(2059),
    [anon_sym_false] = ACTIONS(2059),
    [anon_sym_LBRACE] = ACTIONS(2059),
    [anon_sym_SEMI] = ACTIONS(2059),
    [anon_sym_exception] = ACTIONS(2059),
    [anon_sym_module] = ACTIONS(2059),
    [anon_sym_open] = ACTIONS(2059),
    [anon_sym_include] = ACTIONS(2059),
    [anon_sym_class] = ACTIONS(2059),
    [anon_sym_val] = ACTIONS(2059),
    [anon_sym_end] = ACTIONS(2059),
    [anon_sym_object] = ACTIONS(2059),
    [anon_sym_AMP] = ACTIONS(2059),
    [anon_sym_POUND] = ACTIONS(2059),
    [anon_sym_COLON_COLON] = ACTIONS(2061),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2061),
    [anon_sym_else] = ACTIONS(2059),
    [anon_sym_new] = ACTIONS(2059),
    [anon_sym_LBRACE_LT] = ACTIONS(2061),
    [anon_sym_begin] = ACTIONS(2059),
    [sym_ocamlyacc_value] = ACTIONS(2061),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2059),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2061),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2061),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2061),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2059),
    [anon_sym_SQUOTE] = ACTIONS(2061),
    [anon_sym_DQUOTE] = ACTIONS(2061),
    [sym_prefix_operator] = ACTIONS(2061),
    [aux_sym_sign_operator_token1] = ACTIONS(2059),
    [sym_hash_operator] = ACTIONS(2061),
    [anon_sym_lsl] = ACTIONS(2059),
    [anon_sym_lsr] = ACTIONS(2059),
    [anon_sym_asr] = ACTIONS(2059),
    [aux_sym_pow_operator_token1] = ACTIONS(2061),
    [anon_sym_mod] = ACTIONS(2059),
    [anon_sym_land] = ACTIONS(2059),
    [anon_sym_lor] = ACTIONS(2059),
    [anon_sym_lxor] = ACTIONS(2059),
    [aux_sym_mult_operator_token1] = ACTIONS(2059),
    [aux_sym_add_operator_token1] = ACTIONS(2059),
    [sym_concat_operator] = ACTIONS(2061),
    [sym_rel_operator] = ACTIONS(2059),
    [anon_sym_AMP_AMP] = ACTIONS(2059),
    [anon_sym_or] = ACTIONS(2059),
    [anon_sym_PIPE_PIPE] = ACTIONS(2059),
    [sym_assign_operator] = ACTIONS(2061),
    [sym_let_operator] = ACTIONS(2061),
    [sym__uppercase_identifier] = ACTIONS(2061),
    [aux_sym_directive_token1] = ACTIONS(2059),
    [aux_sym_tag_token1] = ACTIONS(2061),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1341)] = {
    [sym_attribute] = STATE(1341),
    [sym__lowercase_identifier] = ACTIONS(2133),
    [anon_sym_SEMI_SEMI] = ACTIONS(2135),
    [anon_sym_let] = ACTIONS(2133),
    [anon_sym_and] = ACTIONS(2133),
    [anon_sym_TILDE] = ACTIONS(2133),
    [anon_sym_QMARK] = ACTIONS(2133),
    [anon_sym_LPAREN] = ACTIONS(2135),
    [anon_sym_external] = ACTIONS(2133),
    [anon_sym_type] = ACTIONS(2133),
    [anon_sym_COMMA] = ACTIONS(2135),
    [anon_sym_PIPE] = ACTIONS(2133),
    [anon_sym_DOT] = ACTIONS(2135),
    [anon_sym_LBRACK] = ACTIONS(2133),
    [anon_sym_RBRACK] = ACTIONS(2135),
    [anon_sym_true] = ACTIONS(2133),
    [anon_sym_false] = ACTIONS(2133),
    [anon_sym_LBRACE] = ACTIONS(2133),
    [anon_sym_SEMI] = ACTIONS(2133),
    [anon_sym_exception] = ACTIONS(2133),
    [anon_sym_module] = ACTIONS(2133),
    [anon_sym_open] = ACTIONS(2133),
    [anon_sym_include] = ACTIONS(2133),
    [anon_sym_class] = ACTIONS(2133),
    [anon_sym_val] = ACTIONS(2133),
    [anon_sym_end] = ACTIONS(2133),
    [anon_sym_object] = ACTIONS(2133),
    [anon_sym_AMP] = ACTIONS(2133),
    [anon_sym_POUND] = ACTIONS(2133),
    [anon_sym_COLON_COLON] = ACTIONS(2135),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2135),
    [anon_sym_else] = ACTIONS(2133),
    [anon_sym_new] = ACTIONS(2133),
    [anon_sym_LBRACE_LT] = ACTIONS(2135),
    [anon_sym_begin] = ACTIONS(2133),
    [sym_ocamlyacc_value] = ACTIONS(2135),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2133),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2135),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2133),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2133),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2135),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2135),
    [aux_sym_number_token1] = ACTIONS(2135),
    [aux_sym_signed_number_token1] = ACTIONS(2133),
    [anon_sym_SQUOTE] = ACTIONS(2135),
    [anon_sym_DQUOTE] = ACTIONS(2135),
    [sym_prefix_operator] = ACTIONS(2135),
    [aux_sym_sign_operator_token1] = ACTIONS(2133),
    [sym_hash_operator] = ACTIONS(2135),
    [anon_sym_lsl] = ACTIONS(2133),
    [anon_sym_lsr] = ACTIONS(2133),
    [anon_sym_asr] = ACTIONS(2133),
    [aux_sym_pow_operator_token1] = ACTIONS(2135),
    [anon_sym_mod] = ACTIONS(2133),
    [anon_sym_land] = ACTIONS(2133),
    [anon_sym_lor] = ACTIONS(2133),
    [anon_sym_lxor] = ACTIONS(2133),
    [aux_sym_mult_operator_token1] = ACTIONS(2133),
    [aux_sym_add_operator_token1] = ACTIONS(2133),
    [sym_concat_operator] = ACTIONS(2135),
    [sym_rel_operator] = ACTIONS(2133),
    [anon_sym_AMP_AMP] = ACTIONS(2133),
    [anon_sym_or] = ACTIONS(2133),
    [anon_sym_PIPE_PIPE] = ACTIONS(2133),
    [sym_assign_operator] = ACTIONS(2135),
    [sym_let_operator] = ACTIONS(2135),
    [sym__uppercase_identifier] = ACTIONS(2135),
    [aux_sym_directive_token1] = ACTIONS(2133),
    [aux_sym_tag_token1] = ACTIONS(2135),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1342)] = {
    [sym_attribute] = STATE(1342),
    [sym__lowercase_identifier] = ACTIONS(2063),
    [anon_sym_SEMI_SEMI] = ACTIONS(2065),
    [anon_sym_let] = ACTIONS(2063),
    [anon_sym_and] = ACTIONS(2063),
    [anon_sym_TILDE] = ACTIONS(2063),
    [anon_sym_QMARK] = ACTIONS(2063),
    [anon_sym_LPAREN] = ACTIONS(2065),
    [anon_sym_external] = ACTIONS(2063),
    [anon_sym_type] = ACTIONS(2063),
    [anon_sym_COMMA] = ACTIONS(2065),
    [anon_sym_PIPE] = ACTIONS(2063),
    [anon_sym_DOT] = ACTIONS(2065),
    [anon_sym_LBRACK] = ACTIONS(2063),
    [anon_sym_RBRACK] = ACTIONS(2065),
    [anon_sym_true] = ACTIONS(2063),
    [anon_sym_false] = ACTIONS(2063),
    [anon_sym_LBRACE] = ACTIONS(2063),
    [anon_sym_SEMI] = ACTIONS(2063),
    [anon_sym_exception] = ACTIONS(2063),
    [anon_sym_module] = ACTIONS(2063),
    [anon_sym_open] = ACTIONS(2063),
    [anon_sym_include] = ACTIONS(2063),
    [anon_sym_class] = ACTIONS(2063),
    [anon_sym_val] = ACTIONS(2063),
    [anon_sym_end] = ACTIONS(2063),
    [anon_sym_object] = ACTIONS(2063),
    [anon_sym_AMP] = ACTIONS(2063),
    [anon_sym_POUND] = ACTIONS(2063),
    [anon_sym_COLON_COLON] = ACTIONS(2065),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2065),
    [anon_sym_else] = ACTIONS(2063),
    [anon_sym_new] = ACTIONS(2063),
    [anon_sym_LBRACE_LT] = ACTIONS(2065),
    [anon_sym_begin] = ACTIONS(2063),
    [sym_ocamlyacc_value] = ACTIONS(2065),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2063),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2065),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2063),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2063),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2065),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2065),
    [aux_sym_number_token1] = ACTIONS(2065),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2065),
    [sym_prefix_operator] = ACTIONS(2065),
    [aux_sym_sign_operator_token1] = ACTIONS(2063),
    [sym_hash_operator] = ACTIONS(2065),
    [anon_sym_lsl] = ACTIONS(2063),
    [anon_sym_lsr] = ACTIONS(2063),
    [anon_sym_asr] = ACTIONS(2063),
    [aux_sym_pow_operator_token1] = ACTIONS(2065),
    [anon_sym_mod] = ACTIONS(2063),
    [anon_sym_land] = ACTIONS(2063),
    [anon_sym_lor] = ACTIONS(2063),
    [anon_sym_lxor] = ACTIONS(2063),
    [aux_sym_mult_operator_token1] = ACTIONS(2063),
    [aux_sym_add_operator_token1] = ACTIONS(2063),
    [sym_concat_operator] = ACTIONS(2065),
    [sym_rel_operator] = ACTIONS(2063),
    [anon_sym_AMP_AMP] = ACTIONS(2063),
    [anon_sym_or] = ACTIONS(2063),
    [anon_sym_PIPE_PIPE] = ACTIONS(2063),
    [sym_assign_operator] = ACTIONS(2065),
    [sym_let_operator] = ACTIONS(2065),
    [sym__uppercase_identifier] = ACTIONS(2065),
    [aux_sym_directive_token1] = ACTIONS(2063),
    [aux_sym_tag_token1] = ACTIONS(2065),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1343)] = {
    [sym_attribute] = STATE(1343),
    [sym__lowercase_identifier] = ACTIONS(2067),
    [anon_sym_SEMI_SEMI] = ACTIONS(2069),
    [anon_sym_let] = ACTIONS(2067),
    [anon_sym_and] = ACTIONS(2067),
    [anon_sym_TILDE] = ACTIONS(2067),
    [anon_sym_QMARK] = ACTIONS(2067),
    [anon_sym_LPAREN] = ACTIONS(2069),
    [anon_sym_external] = ACTIONS(2067),
    [anon_sym_type] = ACTIONS(2067),
    [anon_sym_COMMA] = ACTIONS(2069),
    [anon_sym_PIPE] = ACTIONS(2067),
    [anon_sym_DOT] = ACTIONS(2069),
    [anon_sym_LBRACK] = ACTIONS(2067),
    [anon_sym_RBRACK] = ACTIONS(2069),
    [anon_sym_true] = ACTIONS(2067),
    [anon_sym_false] = ACTIONS(2067),
    [anon_sym_LBRACE] = ACTIONS(2067),
    [anon_sym_SEMI] = ACTIONS(2067),
    [anon_sym_exception] = ACTIONS(2067),
    [anon_sym_module] = ACTIONS(2067),
    [anon_sym_open] = ACTIONS(2067),
    [anon_sym_include] = ACTIONS(2067),
    [anon_sym_class] = ACTIONS(2067),
    [anon_sym_val] = ACTIONS(2067),
    [anon_sym_end] = ACTIONS(2067),
    [anon_sym_object] = ACTIONS(2067),
    [anon_sym_AMP] = ACTIONS(2067),
    [anon_sym_POUND] = ACTIONS(2067),
    [anon_sym_COLON_COLON] = ACTIONS(2069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2069),
    [anon_sym_else] = ACTIONS(2067),
    [anon_sym_new] = ACTIONS(2067),
    [anon_sym_LBRACE_LT] = ACTIONS(2069),
    [anon_sym_begin] = ACTIONS(2067),
    [sym_ocamlyacc_value] = ACTIONS(2069),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2067),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2069),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2067),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2067),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2069),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2069),
    [aux_sym_number_token1] = ACTIONS(2069),
    [aux_sym_signed_number_token1] = ACTIONS(2067),
    [anon_sym_SQUOTE] = ACTIONS(2069),
    [anon_sym_DQUOTE] = ACTIONS(2069),
    [sym_prefix_operator] = ACTIONS(2069),
    [aux_sym_sign_operator_token1] = ACTIONS(2067),
    [sym_hash_operator] = ACTIONS(2069),
    [anon_sym_lsl] = ACTIONS(2067),
    [anon_sym_lsr] = ACTIONS(2067),
    [anon_sym_asr] = ACTIONS(2067),
    [aux_sym_pow_operator_token1] = ACTIONS(2069),
    [anon_sym_mod] = ACTIONS(2067),
    [anon_sym_land] = ACTIONS(2067),
    [anon_sym_lor] = ACTIONS(2067),
    [anon_sym_lxor] = ACTIONS(2067),
    [aux_sym_mult_operator_token1] = ACTIONS(2067),
    [aux_sym_add_operator_token1] = ACTIONS(2067),
    [sym_concat_operator] = ACTIONS(2069),
    [sym_rel_operator] = ACTIONS(2067),
    [anon_sym_AMP_AMP] = ACTIONS(2067),
    [anon_sym_or] = ACTIONS(2067),
    [anon_sym_PIPE_PIPE] = ACTIONS(2067),
    [sym_assign_operator] = ACTIONS(2069),
    [sym_let_operator] = ACTIONS(2069),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_directive_token1] = ACTIONS(2067),
    [aux_sym_tag_token1] = ACTIONS(2069),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1344)] = {
    [sym_attribute] = STATE(1344),
    [sym__lowercase_identifier] = ACTIONS(2145),
    [anon_sym_SEMI_SEMI] = ACTIONS(2147),
    [anon_sym_let] = ACTIONS(2145),
    [anon_sym_and] = ACTIONS(2145),
    [anon_sym_TILDE] = ACTIONS(2145),
    [anon_sym_QMARK] = ACTIONS(2145),
    [anon_sym_LPAREN] = ACTIONS(2147),
    [anon_sym_external] = ACTIONS(2145),
    [anon_sym_type] = ACTIONS(2145),
    [anon_sym_COMMA] = ACTIONS(2147),
    [anon_sym_PIPE] = ACTIONS(2145),
    [anon_sym_DOT] = ACTIONS(2147),
    [anon_sym_LBRACK] = ACTIONS(2145),
    [anon_sym_RBRACK] = ACTIONS(2147),
    [anon_sym_true] = ACTIONS(2145),
    [anon_sym_false] = ACTIONS(2145),
    [anon_sym_LBRACE] = ACTIONS(2145),
    [anon_sym_SEMI] = ACTIONS(2145),
    [anon_sym_exception] = ACTIONS(2145),
    [anon_sym_module] = ACTIONS(2145),
    [anon_sym_open] = ACTIONS(2145),
    [anon_sym_include] = ACTIONS(2145),
    [anon_sym_class] = ACTIONS(2145),
    [anon_sym_val] = ACTIONS(2145),
    [anon_sym_end] = ACTIONS(2145),
    [anon_sym_object] = ACTIONS(2145),
    [anon_sym_AMP] = ACTIONS(2145),
    [anon_sym_POUND] = ACTIONS(2145),
    [anon_sym_COLON_COLON] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2147),
    [anon_sym_else] = ACTIONS(2145),
    [anon_sym_new] = ACTIONS(2145),
    [anon_sym_LBRACE_LT] = ACTIONS(2147),
    [anon_sym_begin] = ACTIONS(2145),
    [sym_ocamlyacc_value] = ACTIONS(2147),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2145),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2147),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2145),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2145),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2147),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2147),
    [aux_sym_number_token1] = ACTIONS(2147),
    [aux_sym_signed_number_token1] = ACTIONS(2145),
    [anon_sym_SQUOTE] = ACTIONS(2147),
    [anon_sym_DQUOTE] = ACTIONS(2147),
    [sym_prefix_operator] = ACTIONS(2147),
    [aux_sym_sign_operator_token1] = ACTIONS(2145),
    [sym_hash_operator] = ACTIONS(2147),
    [anon_sym_lsl] = ACTIONS(2145),
    [anon_sym_lsr] = ACTIONS(2145),
    [anon_sym_asr] = ACTIONS(2145),
    [aux_sym_pow_operator_token1] = ACTIONS(2147),
    [anon_sym_mod] = ACTIONS(2145),
    [anon_sym_land] = ACTIONS(2145),
    [anon_sym_lor] = ACTIONS(2145),
    [anon_sym_lxor] = ACTIONS(2145),
    [aux_sym_mult_operator_token1] = ACTIONS(2145),
    [aux_sym_add_operator_token1] = ACTIONS(2145),
    [sym_concat_operator] = ACTIONS(2147),
    [sym_rel_operator] = ACTIONS(2145),
    [anon_sym_AMP_AMP] = ACTIONS(2145),
    [anon_sym_or] = ACTIONS(2145),
    [anon_sym_PIPE_PIPE] = ACTIONS(2145),
    [sym_assign_operator] = ACTIONS(2147),
    [sym_let_operator] = ACTIONS(2147),
    [sym__uppercase_identifier] = ACTIONS(2147),
    [aux_sym_directive_token1] = ACTIONS(2145),
    [aux_sym_tag_token1] = ACTIONS(2147),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1345)] = {
    [sym_attribute] = STATE(1345),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_SEMI_SEMI] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2079),
    [anon_sym_and] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_external] = ACTIONS(2079),
    [anon_sym_type] = ACTIONS(2079),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_PIPE] = ACTIONS(2079),
    [anon_sym_DOT] = ACTIONS(2081),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_RBRACK] = ACTIONS(2081),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_SEMI] = ACTIONS(2079),
    [anon_sym_exception] = ACTIONS(2079),
    [anon_sym_module] = ACTIONS(2079),
    [anon_sym_open] = ACTIONS(2079),
    [anon_sym_include] = ACTIONS(2079),
    [anon_sym_class] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_end] = ACTIONS(2079),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_POUND] = ACTIONS(2079),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [sym_hash_operator] = ACTIONS(2081),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym_let_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_directive_token1] = ACTIONS(2079),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1346)] = {
    [sym_attribute] = STATE(1346),
    [sym__lowercase_identifier] = ACTIONS(1979),
    [anon_sym_SEMI_SEMI] = ACTIONS(1981),
    [anon_sym_let] = ACTIONS(1979),
    [anon_sym_and] = ACTIONS(1979),
    [anon_sym_TILDE] = ACTIONS(1979),
    [anon_sym_QMARK] = ACTIONS(1979),
    [anon_sym_LPAREN] = ACTIONS(1981),
    [anon_sym_external] = ACTIONS(1979),
    [anon_sym_type] = ACTIONS(1979),
    [anon_sym_COMMA] = ACTIONS(1981),
    [anon_sym_PIPE] = ACTIONS(1979),
    [anon_sym_DOT] = ACTIONS(1981),
    [anon_sym_LBRACK] = ACTIONS(1979),
    [anon_sym_RBRACK] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(1979),
    [anon_sym_false] = ACTIONS(1979),
    [anon_sym_LBRACE] = ACTIONS(1979),
    [anon_sym_SEMI] = ACTIONS(1979),
    [anon_sym_exception] = ACTIONS(1979),
    [anon_sym_module] = ACTIONS(1979),
    [anon_sym_open] = ACTIONS(1979),
    [anon_sym_include] = ACTIONS(1979),
    [anon_sym_class] = ACTIONS(1979),
    [anon_sym_val] = ACTIONS(1979),
    [anon_sym_end] = ACTIONS(1979),
    [anon_sym_object] = ACTIONS(1979),
    [anon_sym_AMP] = ACTIONS(1979),
    [anon_sym_POUND] = ACTIONS(1979),
    [anon_sym_COLON_COLON] = ACTIONS(1981),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1981),
    [anon_sym_else] = ACTIONS(1979),
    [anon_sym_new] = ACTIONS(1979),
    [anon_sym_LBRACE_LT] = ACTIONS(1981),
    [anon_sym_begin] = ACTIONS(1979),
    [sym_ocamlyacc_value] = ACTIONS(1981),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1979),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1981),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1979),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1979),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1981),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1981),
    [aux_sym_number_token1] = ACTIONS(1981),
    [aux_sym_signed_number_token1] = ACTIONS(1979),
    [anon_sym_SQUOTE] = ACTIONS(1981),
    [anon_sym_DQUOTE] = ACTIONS(1981),
    [sym_prefix_operator] = ACTIONS(1981),
    [aux_sym_sign_operator_token1] = ACTIONS(1979),
    [sym_hash_operator] = ACTIONS(1981),
    [anon_sym_lsl] = ACTIONS(1979),
    [anon_sym_lsr] = ACTIONS(1979),
    [anon_sym_asr] = ACTIONS(1979),
    [aux_sym_pow_operator_token1] = ACTIONS(1981),
    [anon_sym_mod] = ACTIONS(1979),
    [anon_sym_land] = ACTIONS(1979),
    [anon_sym_lor] = ACTIONS(1979),
    [anon_sym_lxor] = ACTIONS(1979),
    [aux_sym_mult_operator_token1] = ACTIONS(1979),
    [aux_sym_add_operator_token1] = ACTIONS(1979),
    [sym_concat_operator] = ACTIONS(1981),
    [sym_rel_operator] = ACTIONS(1979),
    [anon_sym_AMP_AMP] = ACTIONS(1979),
    [anon_sym_or] = ACTIONS(1979),
    [anon_sym_PIPE_PIPE] = ACTIONS(1979),
    [sym_assign_operator] = ACTIONS(1981),
    [sym_let_operator] = ACTIONS(1981),
    [sym__uppercase_identifier] = ACTIONS(1981),
    [aux_sym_directive_token1] = ACTIONS(1979),
    [aux_sym_tag_token1] = ACTIONS(1981),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1347)] = {
    [sym__parenthesized_abstract_type] = STATE(1696),
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5547),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1347),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2580),
    [anon_sym_COMMA] = ACTIONS(2433),
    [anon_sym_PIPE] = ACTIONS(2439),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_SEMI] = ACTIONS(2433),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_as] = ACTIONS(2439),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(2433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2433),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1348)] = {
    [sym_attribute] = STATE(1348),
    [sym__lowercase_identifier] = ACTIONS(2153),
    [anon_sym_SEMI_SEMI] = ACTIONS(2155),
    [anon_sym_let] = ACTIONS(2153),
    [anon_sym_and] = ACTIONS(2153),
    [anon_sym_TILDE] = ACTIONS(2153),
    [anon_sym_QMARK] = ACTIONS(2153),
    [anon_sym_LPAREN] = ACTIONS(2155),
    [anon_sym_external] = ACTIONS(2153),
    [anon_sym_type] = ACTIONS(2153),
    [anon_sym_COMMA] = ACTIONS(2155),
    [anon_sym_PIPE] = ACTIONS(2153),
    [anon_sym_DOT] = ACTIONS(2155),
    [anon_sym_LBRACK] = ACTIONS(2153),
    [anon_sym_RBRACK] = ACTIONS(2155),
    [anon_sym_true] = ACTIONS(2153),
    [anon_sym_false] = ACTIONS(2153),
    [anon_sym_LBRACE] = ACTIONS(2153),
    [anon_sym_SEMI] = ACTIONS(2153),
    [anon_sym_exception] = ACTIONS(2153),
    [anon_sym_module] = ACTIONS(2153),
    [anon_sym_open] = ACTIONS(2153),
    [anon_sym_include] = ACTIONS(2153),
    [anon_sym_class] = ACTIONS(2153),
    [anon_sym_val] = ACTIONS(2153),
    [anon_sym_end] = ACTIONS(2153),
    [anon_sym_object] = ACTIONS(2153),
    [anon_sym_AMP] = ACTIONS(2153),
    [anon_sym_POUND] = ACTIONS(2153),
    [anon_sym_COLON_COLON] = ACTIONS(2155),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2155),
    [anon_sym_else] = ACTIONS(2153),
    [anon_sym_new] = ACTIONS(2153),
    [anon_sym_LBRACE_LT] = ACTIONS(2155),
    [anon_sym_begin] = ACTIONS(2153),
    [sym_ocamlyacc_value] = ACTIONS(2155),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2153),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2155),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2153),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2153),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2155),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2155),
    [aux_sym_number_token1] = ACTIONS(2155),
    [aux_sym_signed_number_token1] = ACTIONS(2153),
    [anon_sym_SQUOTE] = ACTIONS(2155),
    [anon_sym_DQUOTE] = ACTIONS(2155),
    [sym_prefix_operator] = ACTIONS(2155),
    [aux_sym_sign_operator_token1] = ACTIONS(2153),
    [sym_hash_operator] = ACTIONS(2155),
    [anon_sym_lsl] = ACTIONS(2153),
    [anon_sym_lsr] = ACTIONS(2153),
    [anon_sym_asr] = ACTIONS(2153),
    [aux_sym_pow_operator_token1] = ACTIONS(2155),
    [anon_sym_mod] = ACTIONS(2153),
    [anon_sym_land] = ACTIONS(2153),
    [anon_sym_lor] = ACTIONS(2153),
    [anon_sym_lxor] = ACTIONS(2153),
    [aux_sym_mult_operator_token1] = ACTIONS(2153),
    [aux_sym_add_operator_token1] = ACTIONS(2153),
    [sym_concat_operator] = ACTIONS(2155),
    [sym_rel_operator] = ACTIONS(2153),
    [anon_sym_AMP_AMP] = ACTIONS(2153),
    [anon_sym_or] = ACTIONS(2153),
    [anon_sym_PIPE_PIPE] = ACTIONS(2153),
    [sym_assign_operator] = ACTIONS(2155),
    [sym_let_operator] = ACTIONS(2155),
    [sym__uppercase_identifier] = ACTIONS(2155),
    [aux_sym_directive_token1] = ACTIONS(2153),
    [aux_sym_tag_token1] = ACTIONS(2155),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1349)] = {
    [sym_attribute] = STATE(1349),
    [sym__lowercase_identifier] = ACTIONS(2083),
    [anon_sym_SEMI_SEMI] = ACTIONS(2085),
    [anon_sym_let] = ACTIONS(2083),
    [anon_sym_and] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(2083),
    [anon_sym_QMARK] = ACTIONS(2083),
    [anon_sym_LPAREN] = ACTIONS(2085),
    [anon_sym_external] = ACTIONS(2083),
    [anon_sym_type] = ACTIONS(2083),
    [anon_sym_COMMA] = ACTIONS(2085),
    [anon_sym_PIPE] = ACTIONS(2083),
    [anon_sym_DOT] = ACTIONS(2085),
    [anon_sym_LBRACK] = ACTIONS(2083),
    [anon_sym_RBRACK] = ACTIONS(2085),
    [anon_sym_true] = ACTIONS(2083),
    [anon_sym_false] = ACTIONS(2083),
    [anon_sym_LBRACE] = ACTIONS(2083),
    [anon_sym_SEMI] = ACTIONS(2083),
    [anon_sym_exception] = ACTIONS(2083),
    [anon_sym_module] = ACTIONS(2083),
    [anon_sym_open] = ACTIONS(2083),
    [anon_sym_include] = ACTIONS(2083),
    [anon_sym_class] = ACTIONS(2083),
    [anon_sym_val] = ACTIONS(2083),
    [anon_sym_end] = ACTIONS(2083),
    [anon_sym_object] = ACTIONS(2083),
    [anon_sym_AMP] = ACTIONS(2083),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(2085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2085),
    [anon_sym_else] = ACTIONS(2083),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_LBRACE_LT] = ACTIONS(2085),
    [anon_sym_begin] = ACTIONS(2083),
    [sym_ocamlyacc_value] = ACTIONS(2085),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2083),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2085),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2083),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2083),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2085),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2085),
    [aux_sym_number_token1] = ACTIONS(2085),
    [aux_sym_signed_number_token1] = ACTIONS(2083),
    [anon_sym_SQUOTE] = ACTIONS(2085),
    [anon_sym_DQUOTE] = ACTIONS(2085),
    [sym_prefix_operator] = ACTIONS(2085),
    [aux_sym_sign_operator_token1] = ACTIONS(2083),
    [sym_hash_operator] = ACTIONS(2085),
    [anon_sym_lsl] = ACTIONS(2083),
    [anon_sym_lsr] = ACTIONS(2083),
    [anon_sym_asr] = ACTIONS(2083),
    [aux_sym_pow_operator_token1] = ACTIONS(2085),
    [anon_sym_mod] = ACTIONS(2083),
    [anon_sym_land] = ACTIONS(2083),
    [anon_sym_lor] = ACTIONS(2083),
    [anon_sym_lxor] = ACTIONS(2083),
    [aux_sym_mult_operator_token1] = ACTIONS(2083),
    [aux_sym_add_operator_token1] = ACTIONS(2083),
    [sym_concat_operator] = ACTIONS(2085),
    [sym_rel_operator] = ACTIONS(2083),
    [anon_sym_AMP_AMP] = ACTIONS(2083),
    [anon_sym_or] = ACTIONS(2083),
    [anon_sym_PIPE_PIPE] = ACTIONS(2083),
    [sym_assign_operator] = ACTIONS(2085),
    [sym_let_operator] = ACTIONS(2085),
    [sym__uppercase_identifier] = ACTIONS(2085),
    [aux_sym_directive_token1] = ACTIONS(2083),
    [aux_sym_tag_token1] = ACTIONS(2085),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1350)] = {
    [sym_attribute] = STATE(1350),
    [sym__lowercase_identifier] = ACTIONS(2087),
    [anon_sym_SEMI_SEMI] = ACTIONS(2089),
    [anon_sym_let] = ACTIONS(2087),
    [anon_sym_and] = ACTIONS(2087),
    [anon_sym_TILDE] = ACTIONS(2087),
    [anon_sym_QMARK] = ACTIONS(2087),
    [anon_sym_LPAREN] = ACTIONS(2089),
    [anon_sym_external] = ACTIONS(2087),
    [anon_sym_type] = ACTIONS(2087),
    [anon_sym_COMMA] = ACTIONS(2089),
    [anon_sym_PIPE] = ACTIONS(2087),
    [anon_sym_DOT] = ACTIONS(2089),
    [anon_sym_LBRACK] = ACTIONS(2087),
    [anon_sym_RBRACK] = ACTIONS(2089),
    [anon_sym_true] = ACTIONS(2087),
    [anon_sym_false] = ACTIONS(2087),
    [anon_sym_LBRACE] = ACTIONS(2087),
    [anon_sym_SEMI] = ACTIONS(2087),
    [anon_sym_exception] = ACTIONS(2087),
    [anon_sym_module] = ACTIONS(2087),
    [anon_sym_open] = ACTIONS(2087),
    [anon_sym_include] = ACTIONS(2087),
    [anon_sym_class] = ACTIONS(2087),
    [anon_sym_val] = ACTIONS(2087),
    [anon_sym_end] = ACTIONS(2087),
    [anon_sym_object] = ACTIONS(2087),
    [anon_sym_AMP] = ACTIONS(2087),
    [anon_sym_POUND] = ACTIONS(2087),
    [anon_sym_COLON_COLON] = ACTIONS(2089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2089),
    [anon_sym_else] = ACTIONS(2087),
    [anon_sym_new] = ACTIONS(2087),
    [anon_sym_LBRACE_LT] = ACTIONS(2089),
    [anon_sym_begin] = ACTIONS(2087),
    [sym_ocamlyacc_value] = ACTIONS(2089),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2087),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2089),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2087),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2087),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2089),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2089),
    [aux_sym_number_token1] = ACTIONS(2089),
    [aux_sym_signed_number_token1] = ACTIONS(2087),
    [anon_sym_SQUOTE] = ACTIONS(2089),
    [anon_sym_DQUOTE] = ACTIONS(2089),
    [sym_prefix_operator] = ACTIONS(2089),
    [aux_sym_sign_operator_token1] = ACTIONS(2087),
    [sym_hash_operator] = ACTIONS(2089),
    [anon_sym_lsl] = ACTIONS(2087),
    [anon_sym_lsr] = ACTIONS(2087),
    [anon_sym_asr] = ACTIONS(2087),
    [aux_sym_pow_operator_token1] = ACTIONS(2089),
    [anon_sym_mod] = ACTIONS(2087),
    [anon_sym_land] = ACTIONS(2087),
    [anon_sym_lor] = ACTIONS(2087),
    [anon_sym_lxor] = ACTIONS(2087),
    [aux_sym_mult_operator_token1] = ACTIONS(2087),
    [aux_sym_add_operator_token1] = ACTIONS(2087),
    [sym_concat_operator] = ACTIONS(2089),
    [sym_rel_operator] = ACTIONS(2087),
    [anon_sym_AMP_AMP] = ACTIONS(2087),
    [anon_sym_or] = ACTIONS(2087),
    [anon_sym_PIPE_PIPE] = ACTIONS(2087),
    [sym_assign_operator] = ACTIONS(2089),
    [sym_let_operator] = ACTIONS(2089),
    [sym__uppercase_identifier] = ACTIONS(2089),
    [aux_sym_directive_token1] = ACTIONS(2087),
    [aux_sym_tag_token1] = ACTIONS(2089),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1351)] = {
    [sym_attribute] = STATE(1351),
    [sym__lowercase_identifier] = ACTIONS(2091),
    [anon_sym_SEMI_SEMI] = ACTIONS(2093),
    [anon_sym_let] = ACTIONS(2091),
    [anon_sym_and] = ACTIONS(2091),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_external] = ACTIONS(2091),
    [anon_sym_type] = ACTIONS(2091),
    [anon_sym_COMMA] = ACTIONS(2093),
    [anon_sym_PIPE] = ACTIONS(2091),
    [anon_sym_DOT] = ACTIONS(2093),
    [anon_sym_LBRACK] = ACTIONS(2091),
    [anon_sym_RBRACK] = ACTIONS(2093),
    [anon_sym_true] = ACTIONS(2091),
    [anon_sym_false] = ACTIONS(2091),
    [anon_sym_LBRACE] = ACTIONS(2091),
    [anon_sym_SEMI] = ACTIONS(2091),
    [anon_sym_exception] = ACTIONS(2091),
    [anon_sym_module] = ACTIONS(2091),
    [anon_sym_open] = ACTIONS(2091),
    [anon_sym_include] = ACTIONS(2091),
    [anon_sym_class] = ACTIONS(2091),
    [anon_sym_val] = ACTIONS(2091),
    [anon_sym_end] = ACTIONS(2091),
    [anon_sym_object] = ACTIONS(2091),
    [anon_sym_AMP] = ACTIONS(2091),
    [anon_sym_POUND] = ACTIONS(2091),
    [anon_sym_COLON_COLON] = ACTIONS(2093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2093),
    [anon_sym_else] = ACTIONS(2091),
    [anon_sym_new] = ACTIONS(2091),
    [anon_sym_LBRACE_LT] = ACTIONS(2093),
    [anon_sym_begin] = ACTIONS(2091),
    [sym_ocamlyacc_value] = ACTIONS(2093),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2091),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2093),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2091),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2091),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2093),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2093),
    [aux_sym_number_token1] = ACTIONS(2093),
    [aux_sym_signed_number_token1] = ACTIONS(2091),
    [anon_sym_SQUOTE] = ACTIONS(2093),
    [anon_sym_DQUOTE] = ACTIONS(2093),
    [sym_prefix_operator] = ACTIONS(2093),
    [aux_sym_sign_operator_token1] = ACTIONS(2091),
    [sym_hash_operator] = ACTIONS(2093),
    [anon_sym_lsl] = ACTIONS(2091),
    [anon_sym_lsr] = ACTIONS(2091),
    [anon_sym_asr] = ACTIONS(2091),
    [aux_sym_pow_operator_token1] = ACTIONS(2093),
    [anon_sym_mod] = ACTIONS(2091),
    [anon_sym_land] = ACTIONS(2091),
    [anon_sym_lor] = ACTIONS(2091),
    [anon_sym_lxor] = ACTIONS(2091),
    [aux_sym_mult_operator_token1] = ACTIONS(2091),
    [aux_sym_add_operator_token1] = ACTIONS(2091),
    [sym_concat_operator] = ACTIONS(2093),
    [sym_rel_operator] = ACTIONS(2091),
    [anon_sym_AMP_AMP] = ACTIONS(2091),
    [anon_sym_or] = ACTIONS(2091),
    [anon_sym_PIPE_PIPE] = ACTIONS(2091),
    [sym_assign_operator] = ACTIONS(2093),
    [sym_let_operator] = ACTIONS(2093),
    [sym__uppercase_identifier] = ACTIONS(2093),
    [aux_sym_directive_token1] = ACTIONS(2091),
    [aux_sym_tag_token1] = ACTIONS(2093),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1352)] = {
    [sym_attribute] = STATE(1352),
    [sym__lowercase_identifier] = ACTIONS(2165),
    [anon_sym_SEMI_SEMI] = ACTIONS(2167),
    [anon_sym_let] = ACTIONS(2165),
    [anon_sym_and] = ACTIONS(2165),
    [anon_sym_TILDE] = ACTIONS(2165),
    [anon_sym_QMARK] = ACTIONS(2165),
    [anon_sym_LPAREN] = ACTIONS(2167),
    [anon_sym_external] = ACTIONS(2165),
    [anon_sym_type] = ACTIONS(2165),
    [anon_sym_COMMA] = ACTIONS(2167),
    [anon_sym_PIPE] = ACTIONS(2165),
    [anon_sym_DOT] = ACTIONS(2167),
    [anon_sym_LBRACK] = ACTIONS(2165),
    [anon_sym_RBRACK] = ACTIONS(2167),
    [anon_sym_true] = ACTIONS(2165),
    [anon_sym_false] = ACTIONS(2165),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_SEMI] = ACTIONS(2165),
    [anon_sym_exception] = ACTIONS(2165),
    [anon_sym_module] = ACTIONS(2165),
    [anon_sym_open] = ACTIONS(2165),
    [anon_sym_include] = ACTIONS(2165),
    [anon_sym_class] = ACTIONS(2165),
    [anon_sym_val] = ACTIONS(2165),
    [anon_sym_end] = ACTIONS(2165),
    [anon_sym_object] = ACTIONS(2165),
    [anon_sym_AMP] = ACTIONS(2165),
    [anon_sym_POUND] = ACTIONS(2165),
    [anon_sym_COLON_COLON] = ACTIONS(2167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2167),
    [anon_sym_else] = ACTIONS(2165),
    [anon_sym_new] = ACTIONS(2165),
    [anon_sym_LBRACE_LT] = ACTIONS(2167),
    [anon_sym_begin] = ACTIONS(2165),
    [sym_ocamlyacc_value] = ACTIONS(2167),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2165),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2167),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2167),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2167),
    [aux_sym_number_token1] = ACTIONS(2167),
    [aux_sym_signed_number_token1] = ACTIONS(2165),
    [anon_sym_SQUOTE] = ACTIONS(2167),
    [anon_sym_DQUOTE] = ACTIONS(2167),
    [sym_prefix_operator] = ACTIONS(2167),
    [aux_sym_sign_operator_token1] = ACTIONS(2165),
    [sym_hash_operator] = ACTIONS(2167),
    [anon_sym_lsl] = ACTIONS(2165),
    [anon_sym_lsr] = ACTIONS(2165),
    [anon_sym_asr] = ACTIONS(2165),
    [aux_sym_pow_operator_token1] = ACTIONS(2167),
    [anon_sym_mod] = ACTIONS(2165),
    [anon_sym_land] = ACTIONS(2165),
    [anon_sym_lor] = ACTIONS(2165),
    [anon_sym_lxor] = ACTIONS(2165),
    [aux_sym_mult_operator_token1] = ACTIONS(2165),
    [aux_sym_add_operator_token1] = ACTIONS(2165),
    [sym_concat_operator] = ACTIONS(2167),
    [sym_rel_operator] = ACTIONS(2165),
    [anon_sym_AMP_AMP] = ACTIONS(2165),
    [anon_sym_or] = ACTIONS(2165),
    [anon_sym_PIPE_PIPE] = ACTIONS(2165),
    [sym_assign_operator] = ACTIONS(2167),
    [sym_let_operator] = ACTIONS(2167),
    [sym__uppercase_identifier] = ACTIONS(2167),
    [aux_sym_directive_token1] = ACTIONS(2165),
    [aux_sym_tag_token1] = ACTIONS(2167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1353)] = {
    [sym_attribute] = STATE(1353),
    [sym__lowercase_identifier] = ACTIONS(2169),
    [anon_sym_SEMI_SEMI] = ACTIONS(2171),
    [anon_sym_let] = ACTIONS(2169),
    [anon_sym_and] = ACTIONS(2169),
    [anon_sym_TILDE] = ACTIONS(2169),
    [anon_sym_QMARK] = ACTIONS(2169),
    [anon_sym_LPAREN] = ACTIONS(2171),
    [anon_sym_external] = ACTIONS(2169),
    [anon_sym_type] = ACTIONS(2169),
    [anon_sym_COMMA] = ACTIONS(2171),
    [anon_sym_PIPE] = ACTIONS(2169),
    [anon_sym_DOT] = ACTIONS(2171),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_RBRACK] = ACTIONS(2171),
    [anon_sym_true] = ACTIONS(2169),
    [anon_sym_false] = ACTIONS(2169),
    [anon_sym_LBRACE] = ACTIONS(2169),
    [anon_sym_SEMI] = ACTIONS(2169),
    [anon_sym_exception] = ACTIONS(2169),
    [anon_sym_module] = ACTIONS(2169),
    [anon_sym_open] = ACTIONS(2169),
    [anon_sym_include] = ACTIONS(2169),
    [anon_sym_class] = ACTIONS(2169),
    [anon_sym_val] = ACTIONS(2169),
    [anon_sym_end] = ACTIONS(2169),
    [anon_sym_object] = ACTIONS(2169),
    [anon_sym_AMP] = ACTIONS(2169),
    [anon_sym_POUND] = ACTIONS(2169),
    [anon_sym_COLON_COLON] = ACTIONS(2171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2171),
    [anon_sym_else] = ACTIONS(2169),
    [anon_sym_new] = ACTIONS(2169),
    [anon_sym_LBRACE_LT] = ACTIONS(2171),
    [anon_sym_begin] = ACTIONS(2169),
    [sym_ocamlyacc_value] = ACTIONS(2171),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2169),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2171),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2169),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2169),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2171),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2171),
    [aux_sym_number_token1] = ACTIONS(2171),
    [aux_sym_signed_number_token1] = ACTIONS(2169),
    [anon_sym_SQUOTE] = ACTIONS(2171),
    [anon_sym_DQUOTE] = ACTIONS(2171),
    [sym_prefix_operator] = ACTIONS(2171),
    [aux_sym_sign_operator_token1] = ACTIONS(2169),
    [sym_hash_operator] = ACTIONS(2171),
    [anon_sym_lsl] = ACTIONS(2169),
    [anon_sym_lsr] = ACTIONS(2169),
    [anon_sym_asr] = ACTIONS(2169),
    [aux_sym_pow_operator_token1] = ACTIONS(2171),
    [anon_sym_mod] = ACTIONS(2169),
    [anon_sym_land] = ACTIONS(2169),
    [anon_sym_lor] = ACTIONS(2169),
    [anon_sym_lxor] = ACTIONS(2169),
    [aux_sym_mult_operator_token1] = ACTIONS(2169),
    [aux_sym_add_operator_token1] = ACTIONS(2169),
    [sym_concat_operator] = ACTIONS(2171),
    [sym_rel_operator] = ACTIONS(2169),
    [anon_sym_AMP_AMP] = ACTIONS(2169),
    [anon_sym_or] = ACTIONS(2169),
    [anon_sym_PIPE_PIPE] = ACTIONS(2169),
    [sym_assign_operator] = ACTIONS(2171),
    [sym_let_operator] = ACTIONS(2171),
    [sym__uppercase_identifier] = ACTIONS(2171),
    [aux_sym_directive_token1] = ACTIONS(2169),
    [aux_sym_tag_token1] = ACTIONS(2171),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1354)] = {
    [sym_attribute] = STATE(1354),
    [ts_builtin_sym_end] = ACTIONS(1897),
    [sym__lowercase_identifier] = ACTIONS(1895),
    [anon_sym_SEMI_SEMI] = ACTIONS(1897),
    [anon_sym_let] = ACTIONS(1895),
    [anon_sym_and] = ACTIONS(1895),
    [anon_sym_TILDE] = ACTIONS(1895),
    [anon_sym_QMARK] = ACTIONS(1895),
    [anon_sym_LPAREN] = ACTIONS(1897),
    [anon_sym_external] = ACTIONS(1895),
    [anon_sym_type] = ACTIONS(1895),
    [anon_sym_COMMA] = ACTIONS(1897),
    [anon_sym_PIPE] = ACTIONS(1895),
    [anon_sym_DOT] = ACTIONS(1897),
    [anon_sym_LBRACK] = ACTIONS(1895),
    [anon_sym_true] = ACTIONS(1895),
    [anon_sym_false] = ACTIONS(1895),
    [anon_sym_LBRACE] = ACTIONS(1895),
    [anon_sym_SEMI] = ACTIONS(1895),
    [anon_sym_exception] = ACTIONS(1895),
    [anon_sym_module] = ACTIONS(1895),
    [anon_sym_open] = ACTIONS(1895),
    [anon_sym_include] = ACTIONS(1895),
    [anon_sym_class] = ACTIONS(1895),
    [anon_sym_val] = ACTIONS(1895),
    [anon_sym_object] = ACTIONS(1895),
    [anon_sym_AMP] = ACTIONS(1895),
    [anon_sym_POUND] = ACTIONS(1895),
    [anon_sym_COLON_COLON] = ACTIONS(1897),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1897),
    [anon_sym_LT_DASH] = ACTIONS(1895),
    [anon_sym_else] = ACTIONS(1895),
    [anon_sym_new] = ACTIONS(1895),
    [anon_sym_LBRACE_LT] = ACTIONS(1897),
    [anon_sym_begin] = ACTIONS(1895),
    [sym_ocamlyacc_value] = ACTIONS(1897),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1895),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1897),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1895),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1895),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1897),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1897),
    [aux_sym_number_token1] = ACTIONS(1897),
    [aux_sym_signed_number_token1] = ACTIONS(1895),
    [anon_sym_SQUOTE] = ACTIONS(1897),
    [anon_sym_DQUOTE] = ACTIONS(1897),
    [sym_prefix_operator] = ACTIONS(1897),
    [aux_sym_sign_operator_token1] = ACTIONS(1895),
    [sym_hash_operator] = ACTIONS(1897),
    [anon_sym_lsl] = ACTIONS(1895),
    [anon_sym_lsr] = ACTIONS(1895),
    [anon_sym_asr] = ACTIONS(1895),
    [aux_sym_pow_operator_token1] = ACTIONS(1897),
    [anon_sym_mod] = ACTIONS(1895),
    [anon_sym_land] = ACTIONS(1895),
    [anon_sym_lor] = ACTIONS(1895),
    [anon_sym_lxor] = ACTIONS(1895),
    [aux_sym_mult_operator_token1] = ACTIONS(1895),
    [aux_sym_add_operator_token1] = ACTIONS(1895),
    [sym_concat_operator] = ACTIONS(1897),
    [sym_rel_operator] = ACTIONS(1895),
    [anon_sym_AMP_AMP] = ACTIONS(1895),
    [anon_sym_or] = ACTIONS(1895),
    [anon_sym_PIPE_PIPE] = ACTIONS(1895),
    [sym_assign_operator] = ACTIONS(1897),
    [sym_let_operator] = ACTIONS(1897),
    [sym__uppercase_identifier] = ACTIONS(1897),
    [aux_sym_directive_token1] = ACTIONS(1895),
    [aux_sym_tag_token1] = ACTIONS(1897),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1355)] = {
    [sym_attribute] = STATE(1355),
    [ts_builtin_sym_end] = ACTIONS(1905),
    [sym__lowercase_identifier] = ACTIONS(1903),
    [anon_sym_SEMI_SEMI] = ACTIONS(1905),
    [anon_sym_let] = ACTIONS(1903),
    [anon_sym_and] = ACTIONS(1903),
    [anon_sym_TILDE] = ACTIONS(1903),
    [anon_sym_QMARK] = ACTIONS(1903),
    [anon_sym_LPAREN] = ACTIONS(1905),
    [anon_sym_external] = ACTIONS(1903),
    [anon_sym_type] = ACTIONS(1903),
    [anon_sym_COMMA] = ACTIONS(1905),
    [anon_sym_PIPE] = ACTIONS(1903),
    [anon_sym_DOT] = ACTIONS(1905),
    [anon_sym_LBRACK] = ACTIONS(1903),
    [anon_sym_true] = ACTIONS(1903),
    [anon_sym_false] = ACTIONS(1903),
    [anon_sym_LBRACE] = ACTIONS(1903),
    [anon_sym_SEMI] = ACTIONS(1903),
    [anon_sym_exception] = ACTIONS(1903),
    [anon_sym_module] = ACTIONS(1903),
    [anon_sym_open] = ACTIONS(1903),
    [anon_sym_include] = ACTIONS(1903),
    [anon_sym_class] = ACTIONS(1903),
    [anon_sym_val] = ACTIONS(1903),
    [anon_sym_object] = ACTIONS(1903),
    [anon_sym_AMP] = ACTIONS(1903),
    [anon_sym_POUND] = ACTIONS(1903),
    [anon_sym_COLON_COLON] = ACTIONS(1905),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1905),
    [anon_sym_LT_DASH] = ACTIONS(1903),
    [anon_sym_else] = ACTIONS(1903),
    [anon_sym_new] = ACTIONS(1903),
    [anon_sym_LBRACE_LT] = ACTIONS(1905),
    [anon_sym_begin] = ACTIONS(1903),
    [sym_ocamlyacc_value] = ACTIONS(1905),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1903),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1905),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1903),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1903),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1905),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1905),
    [aux_sym_number_token1] = ACTIONS(1905),
    [aux_sym_signed_number_token1] = ACTIONS(1903),
    [anon_sym_SQUOTE] = ACTIONS(1905),
    [anon_sym_DQUOTE] = ACTIONS(1905),
    [sym_prefix_operator] = ACTIONS(1905),
    [aux_sym_sign_operator_token1] = ACTIONS(1903),
    [sym_hash_operator] = ACTIONS(1905),
    [anon_sym_lsl] = ACTIONS(1903),
    [anon_sym_lsr] = ACTIONS(1903),
    [anon_sym_asr] = ACTIONS(1903),
    [aux_sym_pow_operator_token1] = ACTIONS(1905),
    [anon_sym_mod] = ACTIONS(1903),
    [anon_sym_land] = ACTIONS(1903),
    [anon_sym_lor] = ACTIONS(1903),
    [anon_sym_lxor] = ACTIONS(1903),
    [aux_sym_mult_operator_token1] = ACTIONS(1903),
    [aux_sym_add_operator_token1] = ACTIONS(1903),
    [sym_concat_operator] = ACTIONS(1905),
    [sym_rel_operator] = ACTIONS(1903),
    [anon_sym_AMP_AMP] = ACTIONS(1903),
    [anon_sym_or] = ACTIONS(1903),
    [anon_sym_PIPE_PIPE] = ACTIONS(1903),
    [sym_assign_operator] = ACTIONS(1905),
    [sym_let_operator] = ACTIONS(1905),
    [sym__uppercase_identifier] = ACTIONS(1905),
    [aux_sym_directive_token1] = ACTIONS(1903),
    [aux_sym_tag_token1] = ACTIONS(1905),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1356)] = {
    [sym_attribute] = STATE(1356),
    [sym__lowercase_identifier] = ACTIONS(2173),
    [anon_sym_SEMI_SEMI] = ACTIONS(2175),
    [anon_sym_let] = ACTIONS(2173),
    [anon_sym_and] = ACTIONS(2173),
    [anon_sym_TILDE] = ACTIONS(2173),
    [anon_sym_QMARK] = ACTIONS(2173),
    [anon_sym_LPAREN] = ACTIONS(2175),
    [anon_sym_external] = ACTIONS(2173),
    [anon_sym_type] = ACTIONS(2173),
    [anon_sym_COMMA] = ACTIONS(2175),
    [anon_sym_PIPE] = ACTIONS(2173),
    [anon_sym_DOT] = ACTIONS(2175),
    [anon_sym_LBRACK] = ACTIONS(2173),
    [anon_sym_RBRACK] = ACTIONS(2175),
    [anon_sym_true] = ACTIONS(2173),
    [anon_sym_false] = ACTIONS(2173),
    [anon_sym_LBRACE] = ACTIONS(2173),
    [anon_sym_SEMI] = ACTIONS(2173),
    [anon_sym_exception] = ACTIONS(2173),
    [anon_sym_module] = ACTIONS(2173),
    [anon_sym_open] = ACTIONS(2173),
    [anon_sym_include] = ACTIONS(2173),
    [anon_sym_class] = ACTIONS(2173),
    [anon_sym_val] = ACTIONS(2173),
    [anon_sym_end] = ACTIONS(2173),
    [anon_sym_object] = ACTIONS(2173),
    [anon_sym_AMP] = ACTIONS(2173),
    [anon_sym_POUND] = ACTIONS(2173),
    [anon_sym_COLON_COLON] = ACTIONS(2175),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2175),
    [anon_sym_else] = ACTIONS(2173),
    [anon_sym_new] = ACTIONS(2173),
    [anon_sym_LBRACE_LT] = ACTIONS(2175),
    [anon_sym_begin] = ACTIONS(2173),
    [sym_ocamlyacc_value] = ACTIONS(2175),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2173),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2175),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2173),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2175),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2175),
    [aux_sym_number_token1] = ACTIONS(2175),
    [aux_sym_signed_number_token1] = ACTIONS(2173),
    [anon_sym_SQUOTE] = ACTIONS(2175),
    [anon_sym_DQUOTE] = ACTIONS(2175),
    [sym_prefix_operator] = ACTIONS(2175),
    [aux_sym_sign_operator_token1] = ACTIONS(2173),
    [sym_hash_operator] = ACTIONS(2175),
    [anon_sym_lsl] = ACTIONS(2173),
    [anon_sym_lsr] = ACTIONS(2173),
    [anon_sym_asr] = ACTIONS(2173),
    [aux_sym_pow_operator_token1] = ACTIONS(2175),
    [anon_sym_mod] = ACTIONS(2173),
    [anon_sym_land] = ACTIONS(2173),
    [anon_sym_lor] = ACTIONS(2173),
    [anon_sym_lxor] = ACTIONS(2173),
    [aux_sym_mult_operator_token1] = ACTIONS(2173),
    [aux_sym_add_operator_token1] = ACTIONS(2173),
    [sym_concat_operator] = ACTIONS(2175),
    [sym_rel_operator] = ACTIONS(2173),
    [anon_sym_AMP_AMP] = ACTIONS(2173),
    [anon_sym_or] = ACTIONS(2173),
    [anon_sym_PIPE_PIPE] = ACTIONS(2173),
    [sym_assign_operator] = ACTIONS(2175),
    [sym_let_operator] = ACTIONS(2175),
    [sym__uppercase_identifier] = ACTIONS(2175),
    [aux_sym_directive_token1] = ACTIONS(2173),
    [aux_sym_tag_token1] = ACTIONS(2175),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1357)] = {
    [sym_attribute] = STATE(1357),
    [sym__lowercase_identifier] = ACTIONS(2095),
    [anon_sym_SEMI_SEMI] = ACTIONS(2097),
    [anon_sym_let] = ACTIONS(2095),
    [anon_sym_and] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(2095),
    [anon_sym_QMARK] = ACTIONS(2095),
    [anon_sym_LPAREN] = ACTIONS(2097),
    [anon_sym_external] = ACTIONS(2095),
    [anon_sym_type] = ACTIONS(2095),
    [anon_sym_COMMA] = ACTIONS(2097),
    [anon_sym_PIPE] = ACTIONS(2095),
    [anon_sym_DOT] = ACTIONS(2097),
    [anon_sym_LBRACK] = ACTIONS(2095),
    [anon_sym_RBRACK] = ACTIONS(2097),
    [anon_sym_true] = ACTIONS(2095),
    [anon_sym_false] = ACTIONS(2095),
    [anon_sym_LBRACE] = ACTIONS(2095),
    [anon_sym_SEMI] = ACTIONS(2095),
    [anon_sym_exception] = ACTIONS(2095),
    [anon_sym_module] = ACTIONS(2095),
    [anon_sym_open] = ACTIONS(2095),
    [anon_sym_include] = ACTIONS(2095),
    [anon_sym_class] = ACTIONS(2095),
    [anon_sym_val] = ACTIONS(2095),
    [anon_sym_end] = ACTIONS(2095),
    [anon_sym_object] = ACTIONS(2095),
    [anon_sym_AMP] = ACTIONS(2095),
    [anon_sym_POUND] = ACTIONS(2095),
    [anon_sym_COLON_COLON] = ACTIONS(2097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2097),
    [anon_sym_else] = ACTIONS(2095),
    [anon_sym_new] = ACTIONS(2095),
    [anon_sym_LBRACE_LT] = ACTIONS(2097),
    [anon_sym_begin] = ACTIONS(2095),
    [sym_ocamlyacc_value] = ACTIONS(2097),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2095),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2097),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2095),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2095),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2097),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2097),
    [aux_sym_number_token1] = ACTIONS(2097),
    [aux_sym_signed_number_token1] = ACTIONS(2095),
    [anon_sym_SQUOTE] = ACTIONS(2097),
    [anon_sym_DQUOTE] = ACTIONS(2097),
    [sym_prefix_operator] = ACTIONS(2097),
    [aux_sym_sign_operator_token1] = ACTIONS(2095),
    [sym_hash_operator] = ACTIONS(2097),
    [anon_sym_lsl] = ACTIONS(2095),
    [anon_sym_lsr] = ACTIONS(2095),
    [anon_sym_asr] = ACTIONS(2095),
    [aux_sym_pow_operator_token1] = ACTIONS(2097),
    [anon_sym_mod] = ACTIONS(2095),
    [anon_sym_land] = ACTIONS(2095),
    [anon_sym_lor] = ACTIONS(2095),
    [anon_sym_lxor] = ACTIONS(2095),
    [aux_sym_mult_operator_token1] = ACTIONS(2095),
    [aux_sym_add_operator_token1] = ACTIONS(2095),
    [sym_concat_operator] = ACTIONS(2097),
    [sym_rel_operator] = ACTIONS(2095),
    [anon_sym_AMP_AMP] = ACTIONS(2095),
    [anon_sym_or] = ACTIONS(2095),
    [anon_sym_PIPE_PIPE] = ACTIONS(2095),
    [sym_assign_operator] = ACTIONS(2097),
    [sym_let_operator] = ACTIONS(2097),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_directive_token1] = ACTIONS(2095),
    [aux_sym_tag_token1] = ACTIONS(2097),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1358)] = {
    [sym_attribute] = STATE(1358),
    [sym__lowercase_identifier] = ACTIONS(2099),
    [anon_sym_SEMI_SEMI] = ACTIONS(2101),
    [anon_sym_let] = ACTIONS(2099),
    [anon_sym_and] = ACTIONS(2099),
    [anon_sym_TILDE] = ACTIONS(2099),
    [anon_sym_QMARK] = ACTIONS(2099),
    [anon_sym_LPAREN] = ACTIONS(2101),
    [anon_sym_external] = ACTIONS(2099),
    [anon_sym_type] = ACTIONS(2099),
    [anon_sym_COMMA] = ACTIONS(2101),
    [anon_sym_PIPE] = ACTIONS(2099),
    [anon_sym_DOT] = ACTIONS(2101),
    [anon_sym_LBRACK] = ACTIONS(2099),
    [anon_sym_RBRACK] = ACTIONS(2101),
    [anon_sym_true] = ACTIONS(2099),
    [anon_sym_false] = ACTIONS(2099),
    [anon_sym_LBRACE] = ACTIONS(2099),
    [anon_sym_SEMI] = ACTIONS(2099),
    [anon_sym_exception] = ACTIONS(2099),
    [anon_sym_module] = ACTIONS(2099),
    [anon_sym_open] = ACTIONS(2099),
    [anon_sym_include] = ACTIONS(2099),
    [anon_sym_class] = ACTIONS(2099),
    [anon_sym_val] = ACTIONS(2099),
    [anon_sym_end] = ACTIONS(2099),
    [anon_sym_object] = ACTIONS(2099),
    [anon_sym_AMP] = ACTIONS(2099),
    [anon_sym_POUND] = ACTIONS(2099),
    [anon_sym_COLON_COLON] = ACTIONS(2101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2101),
    [anon_sym_else] = ACTIONS(2099),
    [anon_sym_new] = ACTIONS(2099),
    [anon_sym_LBRACE_LT] = ACTIONS(2101),
    [anon_sym_begin] = ACTIONS(2099),
    [sym_ocamlyacc_value] = ACTIONS(2101),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2099),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2101),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2099),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2099),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2101),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2101),
    [aux_sym_number_token1] = ACTIONS(2101),
    [aux_sym_signed_number_token1] = ACTIONS(2099),
    [anon_sym_SQUOTE] = ACTIONS(2101),
    [anon_sym_DQUOTE] = ACTIONS(2101),
    [sym_prefix_operator] = ACTIONS(2101),
    [aux_sym_sign_operator_token1] = ACTIONS(2099),
    [sym_hash_operator] = ACTIONS(2101),
    [anon_sym_lsl] = ACTIONS(2099),
    [anon_sym_lsr] = ACTIONS(2099),
    [anon_sym_asr] = ACTIONS(2099),
    [aux_sym_pow_operator_token1] = ACTIONS(2101),
    [anon_sym_mod] = ACTIONS(2099),
    [anon_sym_land] = ACTIONS(2099),
    [anon_sym_lor] = ACTIONS(2099),
    [anon_sym_lxor] = ACTIONS(2099),
    [aux_sym_mult_operator_token1] = ACTIONS(2099),
    [aux_sym_add_operator_token1] = ACTIONS(2099),
    [sym_concat_operator] = ACTIONS(2101),
    [sym_rel_operator] = ACTIONS(2099),
    [anon_sym_AMP_AMP] = ACTIONS(2099),
    [anon_sym_or] = ACTIONS(2099),
    [anon_sym_PIPE_PIPE] = ACTIONS(2099),
    [sym_assign_operator] = ACTIONS(2101),
    [sym_let_operator] = ACTIONS(2101),
    [sym__uppercase_identifier] = ACTIONS(2101),
    [aux_sym_directive_token1] = ACTIONS(2099),
    [aux_sym_tag_token1] = ACTIONS(2101),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1359)] = {
    [sym_attribute] = STATE(1359),
    [ts_builtin_sym_end] = ACTIONS(1917),
    [sym__lowercase_identifier] = ACTIONS(1915),
    [anon_sym_SEMI_SEMI] = ACTIONS(1917),
    [anon_sym_let] = ACTIONS(1915),
    [anon_sym_and] = ACTIONS(1915),
    [anon_sym_TILDE] = ACTIONS(1915),
    [anon_sym_QMARK] = ACTIONS(1915),
    [anon_sym_LPAREN] = ACTIONS(1917),
    [anon_sym_external] = ACTIONS(1915),
    [anon_sym_type] = ACTIONS(1915),
    [anon_sym_COMMA] = ACTIONS(1917),
    [anon_sym_PIPE] = ACTIONS(1915),
    [anon_sym_DOT] = ACTIONS(1917),
    [anon_sym_LBRACK] = ACTIONS(1915),
    [anon_sym_true] = ACTIONS(1915),
    [anon_sym_false] = ACTIONS(1915),
    [anon_sym_LBRACE] = ACTIONS(1915),
    [anon_sym_SEMI] = ACTIONS(1915),
    [anon_sym_exception] = ACTIONS(1915),
    [anon_sym_module] = ACTIONS(1915),
    [anon_sym_open] = ACTIONS(1915),
    [anon_sym_include] = ACTIONS(1915),
    [anon_sym_class] = ACTIONS(1915),
    [anon_sym_val] = ACTIONS(1915),
    [anon_sym_object] = ACTIONS(1915),
    [anon_sym_AMP] = ACTIONS(1915),
    [anon_sym_POUND] = ACTIONS(1915),
    [anon_sym_COLON_COLON] = ACTIONS(1917),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1917),
    [anon_sym_LT_DASH] = ACTIONS(1915),
    [anon_sym_else] = ACTIONS(1915),
    [anon_sym_new] = ACTIONS(1915),
    [anon_sym_LBRACE_LT] = ACTIONS(1917),
    [anon_sym_begin] = ACTIONS(1915),
    [sym_ocamlyacc_value] = ACTIONS(1917),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1915),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1917),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1915),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1915),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1917),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1917),
    [aux_sym_number_token1] = ACTIONS(1917),
    [aux_sym_signed_number_token1] = ACTIONS(1915),
    [anon_sym_SQUOTE] = ACTIONS(1917),
    [anon_sym_DQUOTE] = ACTIONS(1917),
    [sym_prefix_operator] = ACTIONS(1917),
    [aux_sym_sign_operator_token1] = ACTIONS(1915),
    [sym_hash_operator] = ACTIONS(1917),
    [anon_sym_lsl] = ACTIONS(1915),
    [anon_sym_lsr] = ACTIONS(1915),
    [anon_sym_asr] = ACTIONS(1915),
    [aux_sym_pow_operator_token1] = ACTIONS(1917),
    [anon_sym_mod] = ACTIONS(1915),
    [anon_sym_land] = ACTIONS(1915),
    [anon_sym_lor] = ACTIONS(1915),
    [anon_sym_lxor] = ACTIONS(1915),
    [aux_sym_mult_operator_token1] = ACTIONS(1915),
    [aux_sym_add_operator_token1] = ACTIONS(1915),
    [sym_concat_operator] = ACTIONS(1917),
    [sym_rel_operator] = ACTIONS(1915),
    [anon_sym_AMP_AMP] = ACTIONS(1915),
    [anon_sym_or] = ACTIONS(1915),
    [anon_sym_PIPE_PIPE] = ACTIONS(1915),
    [sym_assign_operator] = ACTIONS(1917),
    [sym_let_operator] = ACTIONS(1917),
    [sym__uppercase_identifier] = ACTIONS(1917),
    [aux_sym_directive_token1] = ACTIONS(1915),
    [aux_sym_tag_token1] = ACTIONS(1917),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1360)] = {
    [sym_attribute] = STATE(1360),
    [sym__lowercase_identifier] = ACTIONS(1959),
    [anon_sym_SEMI_SEMI] = ACTIONS(1961),
    [anon_sym_let] = ACTIONS(1959),
    [anon_sym_and] = ACTIONS(1959),
    [anon_sym_TILDE] = ACTIONS(1959),
    [anon_sym_QMARK] = ACTIONS(1959),
    [anon_sym_LPAREN] = ACTIONS(1961),
    [anon_sym_external] = ACTIONS(1959),
    [anon_sym_type] = ACTIONS(1959),
    [anon_sym_COMMA] = ACTIONS(1961),
    [anon_sym_PIPE] = ACTIONS(1959),
    [anon_sym_DOT] = ACTIONS(1961),
    [anon_sym_LBRACK] = ACTIONS(1959),
    [anon_sym_RBRACK] = ACTIONS(1961),
    [anon_sym_true] = ACTIONS(1959),
    [anon_sym_false] = ACTIONS(1959),
    [anon_sym_LBRACE] = ACTIONS(1959),
    [anon_sym_SEMI] = ACTIONS(1959),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_module] = ACTIONS(1959),
    [anon_sym_open] = ACTIONS(1959),
    [anon_sym_include] = ACTIONS(1959),
    [anon_sym_class] = ACTIONS(1959),
    [anon_sym_val] = ACTIONS(1959),
    [anon_sym_end] = ACTIONS(1959),
    [anon_sym_object] = ACTIONS(1959),
    [anon_sym_AMP] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1959),
    [anon_sym_COLON_COLON] = ACTIONS(1961),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1961),
    [anon_sym_else] = ACTIONS(1959),
    [anon_sym_new] = ACTIONS(1959),
    [anon_sym_LBRACE_LT] = ACTIONS(1961),
    [anon_sym_begin] = ACTIONS(1959),
    [sym_ocamlyacc_value] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1959),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1961),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1959),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1959),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1961),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1961),
    [aux_sym_number_token1] = ACTIONS(1961),
    [aux_sym_signed_number_token1] = ACTIONS(1959),
    [anon_sym_SQUOTE] = ACTIONS(1961),
    [anon_sym_DQUOTE] = ACTIONS(1961),
    [sym_prefix_operator] = ACTIONS(1961),
    [aux_sym_sign_operator_token1] = ACTIONS(1959),
    [sym_hash_operator] = ACTIONS(1961),
    [anon_sym_lsl] = ACTIONS(1959),
    [anon_sym_lsr] = ACTIONS(1959),
    [anon_sym_asr] = ACTIONS(1959),
    [aux_sym_pow_operator_token1] = ACTIONS(1961),
    [anon_sym_mod] = ACTIONS(1959),
    [anon_sym_land] = ACTIONS(1959),
    [anon_sym_lor] = ACTIONS(1959),
    [anon_sym_lxor] = ACTIONS(1959),
    [aux_sym_mult_operator_token1] = ACTIONS(1959),
    [aux_sym_add_operator_token1] = ACTIONS(1959),
    [sym_concat_operator] = ACTIONS(1961),
    [sym_rel_operator] = ACTIONS(1959),
    [anon_sym_AMP_AMP] = ACTIONS(1959),
    [anon_sym_or] = ACTIONS(1959),
    [anon_sym_PIPE_PIPE] = ACTIONS(1959),
    [sym_assign_operator] = ACTIONS(1961),
    [sym_let_operator] = ACTIONS(1961),
    [sym__uppercase_identifier] = ACTIONS(1961),
    [aux_sym_directive_token1] = ACTIONS(1959),
    [aux_sym_tag_token1] = ACTIONS(1961),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1361)] = {
    [sym_attribute] = STATE(1361),
    [sym__lowercase_identifier] = ACTIONS(2031),
    [anon_sym_SEMI_SEMI] = ACTIONS(2033),
    [anon_sym_let] = ACTIONS(2031),
    [anon_sym_and] = ACTIONS(2031),
    [anon_sym_TILDE] = ACTIONS(2031),
    [anon_sym_QMARK] = ACTIONS(2031),
    [anon_sym_LPAREN] = ACTIONS(2033),
    [anon_sym_external] = ACTIONS(2031),
    [anon_sym_type] = ACTIONS(2031),
    [anon_sym_COMMA] = ACTIONS(2033),
    [anon_sym_PIPE] = ACTIONS(2031),
    [anon_sym_DOT] = ACTIONS(2033),
    [anon_sym_LBRACK] = ACTIONS(2031),
    [anon_sym_RBRACK] = ACTIONS(2033),
    [anon_sym_true] = ACTIONS(2031),
    [anon_sym_false] = ACTIONS(2031),
    [anon_sym_LBRACE] = ACTIONS(2031),
    [anon_sym_SEMI] = ACTIONS(2031),
    [anon_sym_exception] = ACTIONS(2031),
    [anon_sym_module] = ACTIONS(2031),
    [anon_sym_open] = ACTIONS(2031),
    [anon_sym_include] = ACTIONS(2031),
    [anon_sym_class] = ACTIONS(2031),
    [anon_sym_val] = ACTIONS(2031),
    [anon_sym_end] = ACTIONS(2031),
    [anon_sym_object] = ACTIONS(2031),
    [anon_sym_AMP] = ACTIONS(2031),
    [anon_sym_POUND] = ACTIONS(2031),
    [anon_sym_COLON_COLON] = ACTIONS(2033),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2033),
    [anon_sym_else] = ACTIONS(2031),
    [anon_sym_new] = ACTIONS(2031),
    [anon_sym_LBRACE_LT] = ACTIONS(2033),
    [anon_sym_begin] = ACTIONS(2031),
    [sym_ocamlyacc_value] = ACTIONS(2033),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2031),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2033),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2031),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2031),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2033),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2033),
    [aux_sym_number_token1] = ACTIONS(2033),
    [aux_sym_signed_number_token1] = ACTIONS(2031),
    [anon_sym_SQUOTE] = ACTIONS(2033),
    [anon_sym_DQUOTE] = ACTIONS(2033),
    [sym_prefix_operator] = ACTIONS(2033),
    [aux_sym_sign_operator_token1] = ACTIONS(2031),
    [sym_hash_operator] = ACTIONS(2033),
    [anon_sym_lsl] = ACTIONS(2031),
    [anon_sym_lsr] = ACTIONS(2031),
    [anon_sym_asr] = ACTIONS(2031),
    [aux_sym_pow_operator_token1] = ACTIONS(2033),
    [anon_sym_mod] = ACTIONS(2031),
    [anon_sym_land] = ACTIONS(2031),
    [anon_sym_lor] = ACTIONS(2031),
    [anon_sym_lxor] = ACTIONS(2031),
    [aux_sym_mult_operator_token1] = ACTIONS(2031),
    [aux_sym_add_operator_token1] = ACTIONS(2031),
    [sym_concat_operator] = ACTIONS(2033),
    [sym_rel_operator] = ACTIONS(2031),
    [anon_sym_AMP_AMP] = ACTIONS(2031),
    [anon_sym_or] = ACTIONS(2031),
    [anon_sym_PIPE_PIPE] = ACTIONS(2031),
    [sym_assign_operator] = ACTIONS(2033),
    [sym_let_operator] = ACTIONS(2033),
    [sym__uppercase_identifier] = ACTIONS(2033),
    [aux_sym_directive_token1] = ACTIONS(2031),
    [aux_sym_tag_token1] = ACTIONS(2033),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1362)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5291),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1362),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_PIPE] = ACTIONS(1641),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_RBRACK] = ACTIONS(1641),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_SEMI] = ACTIONS(1641),
    [anon_sym_RBRACE] = ACTIONS(1641),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_as] = ACTIONS(1645),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1641),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1363)] = {
    [sym_attribute] = STATE(1363),
    [sym__lowercase_identifier] = ACTIONS(2111),
    [anon_sym_SEMI_SEMI] = ACTIONS(2113),
    [anon_sym_let] = ACTIONS(2111),
    [anon_sym_and] = ACTIONS(2111),
    [anon_sym_TILDE] = ACTIONS(2111),
    [anon_sym_QMARK] = ACTIONS(2111),
    [anon_sym_LPAREN] = ACTIONS(2113),
    [anon_sym_external] = ACTIONS(2111),
    [anon_sym_type] = ACTIONS(2111),
    [anon_sym_COMMA] = ACTIONS(2113),
    [anon_sym_PIPE] = ACTIONS(2111),
    [anon_sym_DOT] = ACTIONS(2115),
    [anon_sym_LBRACK] = ACTIONS(2111),
    [anon_sym_RBRACK] = ACTIONS(2113),
    [anon_sym_true] = ACTIONS(2111),
    [anon_sym_false] = ACTIONS(2111),
    [anon_sym_LBRACE] = ACTIONS(2111),
    [anon_sym_SEMI] = ACTIONS(2111),
    [anon_sym_exception] = ACTIONS(2111),
    [anon_sym_module] = ACTIONS(2111),
    [anon_sym_open] = ACTIONS(2111),
    [anon_sym_include] = ACTIONS(2111),
    [anon_sym_class] = ACTIONS(2111),
    [anon_sym_val] = ACTIONS(2111),
    [anon_sym_end] = ACTIONS(2111),
    [anon_sym_object] = ACTIONS(2111),
    [anon_sym_AMP] = ACTIONS(2111),
    [anon_sym_POUND] = ACTIONS(2111),
    [anon_sym_COLON_COLON] = ACTIONS(2113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2113),
    [anon_sym_else] = ACTIONS(2111),
    [anon_sym_new] = ACTIONS(2111),
    [anon_sym_LBRACE_LT] = ACTIONS(2113),
    [anon_sym_begin] = ACTIONS(2111),
    [sym_ocamlyacc_value] = ACTIONS(2113),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2113),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2113),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2113),
    [aux_sym_number_token1] = ACTIONS(2113),
    [aux_sym_signed_number_token1] = ACTIONS(2111),
    [anon_sym_SQUOTE] = ACTIONS(2113),
    [anon_sym_DQUOTE] = ACTIONS(2113),
    [sym_prefix_operator] = ACTIONS(2113),
    [aux_sym_sign_operator_token1] = ACTIONS(2111),
    [sym_hash_operator] = ACTIONS(2113),
    [anon_sym_lsl] = ACTIONS(2111),
    [anon_sym_lsr] = ACTIONS(2111),
    [anon_sym_asr] = ACTIONS(2111),
    [aux_sym_pow_operator_token1] = ACTIONS(2113),
    [anon_sym_mod] = ACTIONS(2111),
    [anon_sym_land] = ACTIONS(2111),
    [anon_sym_lor] = ACTIONS(2111),
    [anon_sym_lxor] = ACTIONS(2111),
    [aux_sym_mult_operator_token1] = ACTIONS(2111),
    [aux_sym_add_operator_token1] = ACTIONS(2111),
    [sym_concat_operator] = ACTIONS(2113),
    [sym_rel_operator] = ACTIONS(2111),
    [anon_sym_AMP_AMP] = ACTIONS(2111),
    [anon_sym_or] = ACTIONS(2111),
    [anon_sym_PIPE_PIPE] = ACTIONS(2111),
    [sym_assign_operator] = ACTIONS(2113),
    [sym_let_operator] = ACTIONS(2113),
    [sym__uppercase_identifier] = ACTIONS(2113),
    [aux_sym_directive_token1] = ACTIONS(2111),
    [aux_sym_tag_token1] = ACTIONS(2113),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1364)] = {
    [sym_attribute] = STATE(1364),
    [sym__lowercase_identifier] = ACTIONS(1947),
    [anon_sym_SEMI_SEMI] = ACTIONS(1949),
    [anon_sym_let] = ACTIONS(1947),
    [anon_sym_and] = ACTIONS(1947),
    [anon_sym_TILDE] = ACTIONS(1947),
    [anon_sym_QMARK] = ACTIONS(1947),
    [anon_sym_LPAREN] = ACTIONS(1949),
    [anon_sym_external] = ACTIONS(1947),
    [anon_sym_type] = ACTIONS(1947),
    [anon_sym_COMMA] = ACTIONS(1949),
    [anon_sym_PIPE] = ACTIONS(1947),
    [anon_sym_DOT] = ACTIONS(1949),
    [anon_sym_LBRACK] = ACTIONS(1947),
    [anon_sym_RBRACK] = ACTIONS(1949),
    [anon_sym_true] = ACTIONS(1947),
    [anon_sym_false] = ACTIONS(1947),
    [anon_sym_LBRACE] = ACTIONS(1947),
    [anon_sym_SEMI] = ACTIONS(1947),
    [anon_sym_exception] = ACTIONS(1947),
    [anon_sym_module] = ACTIONS(1947),
    [anon_sym_open] = ACTIONS(1947),
    [anon_sym_include] = ACTIONS(1947),
    [anon_sym_class] = ACTIONS(1947),
    [anon_sym_val] = ACTIONS(1947),
    [anon_sym_end] = ACTIONS(1947),
    [anon_sym_object] = ACTIONS(1947),
    [anon_sym_AMP] = ACTIONS(1947),
    [anon_sym_POUND] = ACTIONS(1947),
    [anon_sym_COLON_COLON] = ACTIONS(1949),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1949),
    [anon_sym_else] = ACTIONS(1947),
    [anon_sym_new] = ACTIONS(1947),
    [anon_sym_LBRACE_LT] = ACTIONS(1949),
    [anon_sym_begin] = ACTIONS(1947),
    [sym_ocamlyacc_value] = ACTIONS(1949),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1947),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1949),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1949),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1949),
    [aux_sym_number_token1] = ACTIONS(1949),
    [aux_sym_signed_number_token1] = ACTIONS(1947),
    [anon_sym_SQUOTE] = ACTIONS(1949),
    [anon_sym_DQUOTE] = ACTIONS(1949),
    [sym_prefix_operator] = ACTIONS(1949),
    [aux_sym_sign_operator_token1] = ACTIONS(1947),
    [sym_hash_operator] = ACTIONS(1949),
    [anon_sym_lsl] = ACTIONS(1947),
    [anon_sym_lsr] = ACTIONS(1947),
    [anon_sym_asr] = ACTIONS(1947),
    [aux_sym_pow_operator_token1] = ACTIONS(1949),
    [anon_sym_mod] = ACTIONS(1947),
    [anon_sym_land] = ACTIONS(1947),
    [anon_sym_lor] = ACTIONS(1947),
    [anon_sym_lxor] = ACTIONS(1947),
    [aux_sym_mult_operator_token1] = ACTIONS(1947),
    [aux_sym_add_operator_token1] = ACTIONS(1947),
    [sym_concat_operator] = ACTIONS(1949),
    [sym_rel_operator] = ACTIONS(1947),
    [anon_sym_AMP_AMP] = ACTIONS(1947),
    [anon_sym_or] = ACTIONS(1947),
    [anon_sym_PIPE_PIPE] = ACTIONS(1947),
    [sym_assign_operator] = ACTIONS(1949),
    [sym_let_operator] = ACTIONS(1949),
    [sym__uppercase_identifier] = ACTIONS(1949),
    [aux_sym_directive_token1] = ACTIONS(1947),
    [aux_sym_tag_token1] = ACTIONS(1949),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1365)] = {
    [sym_attribute] = STATE(1365),
    [sym__lowercase_identifier] = ACTIONS(1927),
    [anon_sym_SEMI_SEMI] = ACTIONS(1929),
    [anon_sym_let] = ACTIONS(1927),
    [anon_sym_and] = ACTIONS(1927),
    [anon_sym_TILDE] = ACTIONS(1927),
    [anon_sym_QMARK] = ACTIONS(1927),
    [anon_sym_LPAREN] = ACTIONS(1929),
    [anon_sym_external] = ACTIONS(1927),
    [anon_sym_type] = ACTIONS(1927),
    [anon_sym_COMMA] = ACTIONS(1929),
    [anon_sym_PIPE] = ACTIONS(1927),
    [anon_sym_DOT] = ACTIONS(1929),
    [anon_sym_LBRACK] = ACTIONS(1927),
    [anon_sym_RBRACK] = ACTIONS(1929),
    [anon_sym_true] = ACTIONS(1927),
    [anon_sym_false] = ACTIONS(1927),
    [anon_sym_LBRACE] = ACTIONS(1927),
    [anon_sym_SEMI] = ACTIONS(1927),
    [anon_sym_exception] = ACTIONS(1927),
    [anon_sym_module] = ACTIONS(1927),
    [anon_sym_open] = ACTIONS(1927),
    [anon_sym_include] = ACTIONS(1927),
    [anon_sym_class] = ACTIONS(1927),
    [anon_sym_val] = ACTIONS(1927),
    [anon_sym_end] = ACTIONS(1927),
    [anon_sym_object] = ACTIONS(1927),
    [anon_sym_AMP] = ACTIONS(1927),
    [anon_sym_POUND] = ACTIONS(1927),
    [anon_sym_COLON_COLON] = ACTIONS(1929),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1929),
    [anon_sym_else] = ACTIONS(1927),
    [anon_sym_new] = ACTIONS(1927),
    [anon_sym_LBRACE_LT] = ACTIONS(1929),
    [anon_sym_begin] = ACTIONS(1927),
    [sym_ocamlyacc_value] = ACTIONS(1929),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1927),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1929),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1927),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1927),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1929),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1929),
    [aux_sym_number_token1] = ACTIONS(1929),
    [aux_sym_signed_number_token1] = ACTIONS(1927),
    [anon_sym_SQUOTE] = ACTIONS(1929),
    [anon_sym_DQUOTE] = ACTIONS(1929),
    [sym_prefix_operator] = ACTIONS(1929),
    [aux_sym_sign_operator_token1] = ACTIONS(1927),
    [sym_hash_operator] = ACTIONS(1929),
    [anon_sym_lsl] = ACTIONS(1927),
    [anon_sym_lsr] = ACTIONS(1927),
    [anon_sym_asr] = ACTIONS(1927),
    [aux_sym_pow_operator_token1] = ACTIONS(1929),
    [anon_sym_mod] = ACTIONS(1927),
    [anon_sym_land] = ACTIONS(1927),
    [anon_sym_lor] = ACTIONS(1927),
    [anon_sym_lxor] = ACTIONS(1927),
    [aux_sym_mult_operator_token1] = ACTIONS(1927),
    [aux_sym_add_operator_token1] = ACTIONS(1927),
    [sym_concat_operator] = ACTIONS(1929),
    [sym_rel_operator] = ACTIONS(1927),
    [anon_sym_AMP_AMP] = ACTIONS(1927),
    [anon_sym_or] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1927),
    [sym_assign_operator] = ACTIONS(1929),
    [sym_let_operator] = ACTIONS(1929),
    [sym__uppercase_identifier] = ACTIONS(1929),
    [aux_sym_directive_token1] = ACTIONS(1927),
    [aux_sym_tag_token1] = ACTIONS(1929),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1366)] = {
    [sym_attribute] = STATE(1366),
    [sym__lowercase_identifier] = ACTIONS(2207),
    [anon_sym_SEMI_SEMI] = ACTIONS(2209),
    [anon_sym_let] = ACTIONS(2207),
    [anon_sym_and] = ACTIONS(2207),
    [anon_sym_TILDE] = ACTIONS(2207),
    [anon_sym_QMARK] = ACTIONS(2207),
    [anon_sym_LPAREN] = ACTIONS(2209),
    [anon_sym_external] = ACTIONS(2207),
    [anon_sym_type] = ACTIONS(2207),
    [anon_sym_COMMA] = ACTIONS(2209),
    [anon_sym_PIPE] = ACTIONS(2207),
    [anon_sym_DOT] = ACTIONS(2209),
    [anon_sym_LBRACK] = ACTIONS(2207),
    [anon_sym_RBRACK] = ACTIONS(2209),
    [anon_sym_true] = ACTIONS(2207),
    [anon_sym_false] = ACTIONS(2207),
    [anon_sym_LBRACE] = ACTIONS(2207),
    [anon_sym_SEMI] = ACTIONS(2207),
    [anon_sym_exception] = ACTIONS(2207),
    [anon_sym_module] = ACTIONS(2207),
    [anon_sym_open] = ACTIONS(2207),
    [anon_sym_include] = ACTIONS(2207),
    [anon_sym_class] = ACTIONS(2207),
    [anon_sym_val] = ACTIONS(2207),
    [anon_sym_end] = ACTIONS(2207),
    [anon_sym_object] = ACTIONS(2207),
    [anon_sym_AMP] = ACTIONS(2207),
    [anon_sym_POUND] = ACTIONS(2207),
    [anon_sym_COLON_COLON] = ACTIONS(2209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2209),
    [anon_sym_else] = ACTIONS(2207),
    [anon_sym_new] = ACTIONS(2207),
    [anon_sym_LBRACE_LT] = ACTIONS(2209),
    [anon_sym_begin] = ACTIONS(2207),
    [sym_ocamlyacc_value] = ACTIONS(2209),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2207),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2209),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2207),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2207),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2209),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2209),
    [aux_sym_number_token1] = ACTIONS(2209),
    [aux_sym_signed_number_token1] = ACTIONS(2207),
    [anon_sym_SQUOTE] = ACTIONS(2209),
    [anon_sym_DQUOTE] = ACTIONS(2209),
    [sym_prefix_operator] = ACTIONS(2209),
    [aux_sym_sign_operator_token1] = ACTIONS(2207),
    [sym_hash_operator] = ACTIONS(2209),
    [anon_sym_lsl] = ACTIONS(2207),
    [anon_sym_lsr] = ACTIONS(2207),
    [anon_sym_asr] = ACTIONS(2207),
    [aux_sym_pow_operator_token1] = ACTIONS(2209),
    [anon_sym_mod] = ACTIONS(2207),
    [anon_sym_land] = ACTIONS(2207),
    [anon_sym_lor] = ACTIONS(2207),
    [anon_sym_lxor] = ACTIONS(2207),
    [aux_sym_mult_operator_token1] = ACTIONS(2207),
    [aux_sym_add_operator_token1] = ACTIONS(2207),
    [sym_concat_operator] = ACTIONS(2209),
    [sym_rel_operator] = ACTIONS(2207),
    [anon_sym_AMP_AMP] = ACTIONS(2207),
    [anon_sym_or] = ACTIONS(2207),
    [anon_sym_PIPE_PIPE] = ACTIONS(2207),
    [sym_assign_operator] = ACTIONS(2209),
    [sym_let_operator] = ACTIONS(2209),
    [sym__uppercase_identifier] = ACTIONS(2209),
    [aux_sym_directive_token1] = ACTIONS(2207),
    [aux_sym_tag_token1] = ACTIONS(2209),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1367)] = {
    [sym_attribute] = STATE(1367),
    [sym__lowercase_identifier] = ACTIONS(2157),
    [anon_sym_SEMI_SEMI] = ACTIONS(2159),
    [anon_sym_let] = ACTIONS(2157),
    [anon_sym_and] = ACTIONS(2157),
    [anon_sym_TILDE] = ACTIONS(2157),
    [anon_sym_QMARK] = ACTIONS(2157),
    [anon_sym_LPAREN] = ACTIONS(2159),
    [anon_sym_external] = ACTIONS(2157),
    [anon_sym_type] = ACTIONS(2157),
    [anon_sym_COMMA] = ACTIONS(2159),
    [anon_sym_PIPE] = ACTIONS(2157),
    [anon_sym_DOT] = ACTIONS(2159),
    [anon_sym_LBRACK] = ACTIONS(2157),
    [anon_sym_RBRACK] = ACTIONS(2159),
    [anon_sym_true] = ACTIONS(2157),
    [anon_sym_false] = ACTIONS(2157),
    [anon_sym_LBRACE] = ACTIONS(2157),
    [anon_sym_SEMI] = ACTIONS(2157),
    [anon_sym_exception] = ACTIONS(2157),
    [anon_sym_module] = ACTIONS(2157),
    [anon_sym_open] = ACTIONS(2157),
    [anon_sym_include] = ACTIONS(2157),
    [anon_sym_class] = ACTIONS(2157),
    [anon_sym_val] = ACTIONS(2157),
    [anon_sym_end] = ACTIONS(2157),
    [anon_sym_object] = ACTIONS(2157),
    [anon_sym_AMP] = ACTIONS(2157),
    [anon_sym_POUND] = ACTIONS(2157),
    [anon_sym_COLON_COLON] = ACTIONS(2159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2159),
    [anon_sym_else] = ACTIONS(2157),
    [anon_sym_new] = ACTIONS(2157),
    [anon_sym_LBRACE_LT] = ACTIONS(2159),
    [anon_sym_begin] = ACTIONS(2157),
    [sym_ocamlyacc_value] = ACTIONS(2159),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2157),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2159),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2159),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2159),
    [aux_sym_number_token1] = ACTIONS(2159),
    [aux_sym_signed_number_token1] = ACTIONS(2157),
    [anon_sym_SQUOTE] = ACTIONS(2159),
    [anon_sym_DQUOTE] = ACTIONS(2159),
    [sym_prefix_operator] = ACTIONS(2159),
    [aux_sym_sign_operator_token1] = ACTIONS(2157),
    [sym_hash_operator] = ACTIONS(2159),
    [anon_sym_lsl] = ACTIONS(2157),
    [anon_sym_lsr] = ACTIONS(2157),
    [anon_sym_asr] = ACTIONS(2157),
    [aux_sym_pow_operator_token1] = ACTIONS(2159),
    [anon_sym_mod] = ACTIONS(2157),
    [anon_sym_land] = ACTIONS(2157),
    [anon_sym_lor] = ACTIONS(2157),
    [anon_sym_lxor] = ACTIONS(2157),
    [aux_sym_mult_operator_token1] = ACTIONS(2157),
    [aux_sym_add_operator_token1] = ACTIONS(2157),
    [sym_concat_operator] = ACTIONS(2159),
    [sym_rel_operator] = ACTIONS(2157),
    [anon_sym_AMP_AMP] = ACTIONS(2157),
    [anon_sym_or] = ACTIONS(2157),
    [anon_sym_PIPE_PIPE] = ACTIONS(2157),
    [sym_assign_operator] = ACTIONS(2159),
    [sym_let_operator] = ACTIONS(2159),
    [sym__uppercase_identifier] = ACTIONS(2159),
    [aux_sym_directive_token1] = ACTIONS(2157),
    [aux_sym_tag_token1] = ACTIONS(2159),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1368)] = {
    [sym_attribute] = STATE(1368),
    [sym__lowercase_identifier] = ACTIONS(2183),
    [anon_sym_SEMI_SEMI] = ACTIONS(2185),
    [anon_sym_let] = ACTIONS(2183),
    [anon_sym_and] = ACTIONS(2183),
    [anon_sym_TILDE] = ACTIONS(2183),
    [anon_sym_QMARK] = ACTIONS(2183),
    [anon_sym_LPAREN] = ACTIONS(2185),
    [anon_sym_external] = ACTIONS(2183),
    [anon_sym_type] = ACTIONS(2183),
    [anon_sym_COMMA] = ACTIONS(2185),
    [anon_sym_PIPE] = ACTIONS(2183),
    [anon_sym_DOT] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2183),
    [anon_sym_RBRACK] = ACTIONS(2185),
    [anon_sym_true] = ACTIONS(2183),
    [anon_sym_false] = ACTIONS(2183),
    [anon_sym_LBRACE] = ACTIONS(2183),
    [anon_sym_SEMI] = ACTIONS(2183),
    [anon_sym_exception] = ACTIONS(2183),
    [anon_sym_module] = ACTIONS(2183),
    [anon_sym_open] = ACTIONS(2183),
    [anon_sym_include] = ACTIONS(2183),
    [anon_sym_class] = ACTIONS(2183),
    [anon_sym_val] = ACTIONS(2183),
    [anon_sym_end] = ACTIONS(2183),
    [anon_sym_object] = ACTIONS(2183),
    [anon_sym_AMP] = ACTIONS(2183),
    [anon_sym_POUND] = ACTIONS(2183),
    [anon_sym_COLON_COLON] = ACTIONS(2185),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2185),
    [anon_sym_else] = ACTIONS(2183),
    [anon_sym_new] = ACTIONS(2183),
    [anon_sym_LBRACE_LT] = ACTIONS(2185),
    [anon_sym_begin] = ACTIONS(2183),
    [sym_ocamlyacc_value] = ACTIONS(2185),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2183),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2185),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2183),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2183),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2185),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2185),
    [aux_sym_number_token1] = ACTIONS(2185),
    [aux_sym_signed_number_token1] = ACTIONS(2183),
    [anon_sym_SQUOTE] = ACTIONS(2185),
    [anon_sym_DQUOTE] = ACTIONS(2185),
    [sym_prefix_operator] = ACTIONS(2185),
    [aux_sym_sign_operator_token1] = ACTIONS(2183),
    [sym_hash_operator] = ACTIONS(2185),
    [anon_sym_lsl] = ACTIONS(2183),
    [anon_sym_lsr] = ACTIONS(2183),
    [anon_sym_asr] = ACTIONS(2183),
    [aux_sym_pow_operator_token1] = ACTIONS(2185),
    [anon_sym_mod] = ACTIONS(2183),
    [anon_sym_land] = ACTIONS(2183),
    [anon_sym_lor] = ACTIONS(2183),
    [anon_sym_lxor] = ACTIONS(2183),
    [aux_sym_mult_operator_token1] = ACTIONS(2183),
    [aux_sym_add_operator_token1] = ACTIONS(2183),
    [sym_concat_operator] = ACTIONS(2185),
    [sym_rel_operator] = ACTIONS(2183),
    [anon_sym_AMP_AMP] = ACTIONS(2183),
    [anon_sym_or] = ACTIONS(2183),
    [anon_sym_PIPE_PIPE] = ACTIONS(2183),
    [sym_assign_operator] = ACTIONS(2185),
    [sym_let_operator] = ACTIONS(2185),
    [sym__uppercase_identifier] = ACTIONS(2185),
    [aux_sym_directive_token1] = ACTIONS(2183),
    [aux_sym_tag_token1] = ACTIONS(2185),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1369)] = {
    [sym_attribute] = STATE(1369),
    [sym__lowercase_identifier] = ACTIONS(2187),
    [anon_sym_SEMI_SEMI] = ACTIONS(2189),
    [anon_sym_let] = ACTIONS(2187),
    [anon_sym_and] = ACTIONS(2187),
    [anon_sym_TILDE] = ACTIONS(2187),
    [anon_sym_QMARK] = ACTIONS(2187),
    [anon_sym_LPAREN] = ACTIONS(2189),
    [anon_sym_external] = ACTIONS(2187),
    [anon_sym_type] = ACTIONS(2187),
    [anon_sym_COMMA] = ACTIONS(2189),
    [anon_sym_PIPE] = ACTIONS(2187),
    [anon_sym_DOT] = ACTIONS(2189),
    [anon_sym_LBRACK] = ACTIONS(2187),
    [anon_sym_RBRACK] = ACTIONS(2189),
    [anon_sym_true] = ACTIONS(2187),
    [anon_sym_false] = ACTIONS(2187),
    [anon_sym_LBRACE] = ACTIONS(2187),
    [anon_sym_SEMI] = ACTIONS(2187),
    [anon_sym_exception] = ACTIONS(2187),
    [anon_sym_module] = ACTIONS(2187),
    [anon_sym_open] = ACTIONS(2187),
    [anon_sym_include] = ACTIONS(2187),
    [anon_sym_class] = ACTIONS(2187),
    [anon_sym_val] = ACTIONS(2187),
    [anon_sym_end] = ACTIONS(2187),
    [anon_sym_object] = ACTIONS(2187),
    [anon_sym_AMP] = ACTIONS(2187),
    [anon_sym_POUND] = ACTIONS(2187),
    [anon_sym_COLON_COLON] = ACTIONS(2189),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2189),
    [anon_sym_else] = ACTIONS(2187),
    [anon_sym_new] = ACTIONS(2187),
    [anon_sym_LBRACE_LT] = ACTIONS(2189),
    [anon_sym_begin] = ACTIONS(2187),
    [sym_ocamlyacc_value] = ACTIONS(2189),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2187),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2189),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2187),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2187),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2189),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2189),
    [aux_sym_number_token1] = ACTIONS(2189),
    [aux_sym_signed_number_token1] = ACTIONS(2187),
    [anon_sym_SQUOTE] = ACTIONS(2189),
    [anon_sym_DQUOTE] = ACTIONS(2189),
    [sym_prefix_operator] = ACTIONS(2189),
    [aux_sym_sign_operator_token1] = ACTIONS(2187),
    [sym_hash_operator] = ACTIONS(2189),
    [anon_sym_lsl] = ACTIONS(2187),
    [anon_sym_lsr] = ACTIONS(2187),
    [anon_sym_asr] = ACTIONS(2187),
    [aux_sym_pow_operator_token1] = ACTIONS(2189),
    [anon_sym_mod] = ACTIONS(2187),
    [anon_sym_land] = ACTIONS(2187),
    [anon_sym_lor] = ACTIONS(2187),
    [anon_sym_lxor] = ACTIONS(2187),
    [aux_sym_mult_operator_token1] = ACTIONS(2187),
    [aux_sym_add_operator_token1] = ACTIONS(2187),
    [sym_concat_operator] = ACTIONS(2189),
    [sym_rel_operator] = ACTIONS(2187),
    [anon_sym_AMP_AMP] = ACTIONS(2187),
    [anon_sym_or] = ACTIONS(2187),
    [anon_sym_PIPE_PIPE] = ACTIONS(2187),
    [sym_assign_operator] = ACTIONS(2189),
    [sym_let_operator] = ACTIONS(2189),
    [sym__uppercase_identifier] = ACTIONS(2189),
    [aux_sym_directive_token1] = ACTIONS(2187),
    [aux_sym_tag_token1] = ACTIONS(2189),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1370)] = {
    [sym_attribute] = STATE(1370),
    [sym__lowercase_identifier] = ACTIONS(1975),
    [anon_sym_SEMI_SEMI] = ACTIONS(1977),
    [anon_sym_let] = ACTIONS(1975),
    [anon_sym_and] = ACTIONS(1975),
    [anon_sym_TILDE] = ACTIONS(1975),
    [anon_sym_QMARK] = ACTIONS(1975),
    [anon_sym_LPAREN] = ACTIONS(1977),
    [anon_sym_external] = ACTIONS(1975),
    [anon_sym_type] = ACTIONS(1975),
    [anon_sym_COMMA] = ACTIONS(1977),
    [anon_sym_PIPE] = ACTIONS(1975),
    [anon_sym_DOT] = ACTIONS(1977),
    [anon_sym_LBRACK] = ACTIONS(1975),
    [anon_sym_RBRACK] = ACTIONS(1977),
    [anon_sym_true] = ACTIONS(1975),
    [anon_sym_false] = ACTIONS(1975),
    [anon_sym_LBRACE] = ACTIONS(1975),
    [anon_sym_SEMI] = ACTIONS(1975),
    [anon_sym_exception] = ACTIONS(1975),
    [anon_sym_module] = ACTIONS(1975),
    [anon_sym_open] = ACTIONS(1975),
    [anon_sym_include] = ACTIONS(1975),
    [anon_sym_class] = ACTIONS(1975),
    [anon_sym_val] = ACTIONS(1975),
    [anon_sym_end] = ACTIONS(1975),
    [anon_sym_object] = ACTIONS(1975),
    [anon_sym_AMP] = ACTIONS(1975),
    [anon_sym_POUND] = ACTIONS(1975),
    [anon_sym_COLON_COLON] = ACTIONS(1977),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1977),
    [anon_sym_else] = ACTIONS(1975),
    [anon_sym_new] = ACTIONS(1975),
    [anon_sym_LBRACE_LT] = ACTIONS(1977),
    [anon_sym_begin] = ACTIONS(1975),
    [sym_ocamlyacc_value] = ACTIONS(1977),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1975),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1977),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1975),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1975),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1977),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1977),
    [aux_sym_number_token1] = ACTIONS(1977),
    [aux_sym_signed_number_token1] = ACTIONS(1975),
    [anon_sym_SQUOTE] = ACTIONS(1977),
    [anon_sym_DQUOTE] = ACTIONS(1977),
    [sym_prefix_operator] = ACTIONS(1977),
    [aux_sym_sign_operator_token1] = ACTIONS(1975),
    [sym_hash_operator] = ACTIONS(1977),
    [anon_sym_lsl] = ACTIONS(1975),
    [anon_sym_lsr] = ACTIONS(1975),
    [anon_sym_asr] = ACTIONS(1975),
    [aux_sym_pow_operator_token1] = ACTIONS(1977),
    [anon_sym_mod] = ACTIONS(1975),
    [anon_sym_land] = ACTIONS(1975),
    [anon_sym_lor] = ACTIONS(1975),
    [anon_sym_lxor] = ACTIONS(1975),
    [aux_sym_mult_operator_token1] = ACTIONS(1975),
    [aux_sym_add_operator_token1] = ACTIONS(1975),
    [sym_concat_operator] = ACTIONS(1977),
    [sym_rel_operator] = ACTIONS(1975),
    [anon_sym_AMP_AMP] = ACTIONS(1975),
    [anon_sym_or] = ACTIONS(1975),
    [anon_sym_PIPE_PIPE] = ACTIONS(1975),
    [sym_assign_operator] = ACTIONS(1977),
    [sym_let_operator] = ACTIONS(1977),
    [sym__uppercase_identifier] = ACTIONS(1977),
    [aux_sym_directive_token1] = ACTIONS(1975),
    [aux_sym_tag_token1] = ACTIONS(1977),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1371)] = {
    [sym_attribute] = STATE(1371),
    [sym__lowercase_identifier] = ACTIONS(1955),
    [anon_sym_SEMI_SEMI] = ACTIONS(1957),
    [anon_sym_let] = ACTIONS(1955),
    [anon_sym_and] = ACTIONS(1955),
    [anon_sym_TILDE] = ACTIONS(1955),
    [anon_sym_QMARK] = ACTIONS(1955),
    [anon_sym_LPAREN] = ACTIONS(1957),
    [anon_sym_external] = ACTIONS(1955),
    [anon_sym_type] = ACTIONS(1955),
    [anon_sym_COMMA] = ACTIONS(1957),
    [anon_sym_PIPE] = ACTIONS(1955),
    [anon_sym_DOT] = ACTIONS(1957),
    [anon_sym_LBRACK] = ACTIONS(1955),
    [anon_sym_RBRACK] = ACTIONS(1957),
    [anon_sym_true] = ACTIONS(1955),
    [anon_sym_false] = ACTIONS(1955),
    [anon_sym_LBRACE] = ACTIONS(1955),
    [anon_sym_SEMI] = ACTIONS(1955),
    [anon_sym_exception] = ACTIONS(1955),
    [anon_sym_module] = ACTIONS(1955),
    [anon_sym_open] = ACTIONS(1955),
    [anon_sym_include] = ACTIONS(1955),
    [anon_sym_class] = ACTIONS(1955),
    [anon_sym_val] = ACTIONS(1955),
    [anon_sym_end] = ACTIONS(1955),
    [anon_sym_object] = ACTIONS(1955),
    [anon_sym_AMP] = ACTIONS(1955),
    [anon_sym_POUND] = ACTIONS(1955),
    [anon_sym_COLON_COLON] = ACTIONS(1957),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1957),
    [anon_sym_else] = ACTIONS(1955),
    [anon_sym_new] = ACTIONS(1955),
    [anon_sym_LBRACE_LT] = ACTIONS(1957),
    [anon_sym_begin] = ACTIONS(1955),
    [sym_ocamlyacc_value] = ACTIONS(1957),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1955),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1957),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1955),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1955),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1957),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1957),
    [aux_sym_number_token1] = ACTIONS(1957),
    [aux_sym_signed_number_token1] = ACTIONS(1955),
    [anon_sym_SQUOTE] = ACTIONS(1957),
    [anon_sym_DQUOTE] = ACTIONS(1957),
    [sym_prefix_operator] = ACTIONS(1957),
    [aux_sym_sign_operator_token1] = ACTIONS(1955),
    [sym_hash_operator] = ACTIONS(1957),
    [anon_sym_lsl] = ACTIONS(1955),
    [anon_sym_lsr] = ACTIONS(1955),
    [anon_sym_asr] = ACTIONS(1955),
    [aux_sym_pow_operator_token1] = ACTIONS(1957),
    [anon_sym_mod] = ACTIONS(1955),
    [anon_sym_land] = ACTIONS(1955),
    [anon_sym_lor] = ACTIONS(1955),
    [anon_sym_lxor] = ACTIONS(1955),
    [aux_sym_mult_operator_token1] = ACTIONS(1955),
    [aux_sym_add_operator_token1] = ACTIONS(1955),
    [sym_concat_operator] = ACTIONS(1957),
    [sym_rel_operator] = ACTIONS(1955),
    [anon_sym_AMP_AMP] = ACTIONS(1955),
    [anon_sym_or] = ACTIONS(1955),
    [anon_sym_PIPE_PIPE] = ACTIONS(1955),
    [sym_assign_operator] = ACTIONS(1957),
    [sym_let_operator] = ACTIONS(1957),
    [sym__uppercase_identifier] = ACTIONS(1957),
    [aux_sym_directive_token1] = ACTIONS(1955),
    [aux_sym_tag_token1] = ACTIONS(1957),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1372)] = {
    [sym_attribute] = STATE(1372),
    [sym__lowercase_identifier] = ACTIONS(2199),
    [anon_sym_SEMI_SEMI] = ACTIONS(2201),
    [anon_sym_let] = ACTIONS(2199),
    [anon_sym_and] = ACTIONS(2199),
    [anon_sym_TILDE] = ACTIONS(2199),
    [anon_sym_QMARK] = ACTIONS(2199),
    [anon_sym_LPAREN] = ACTIONS(2201),
    [anon_sym_external] = ACTIONS(2199),
    [anon_sym_type] = ACTIONS(2199),
    [anon_sym_COMMA] = ACTIONS(2201),
    [anon_sym_PIPE] = ACTIONS(2199),
    [anon_sym_DOT] = ACTIONS(2201),
    [anon_sym_LBRACK] = ACTIONS(2199),
    [anon_sym_RBRACK] = ACTIONS(2201),
    [anon_sym_true] = ACTIONS(2199),
    [anon_sym_false] = ACTIONS(2199),
    [anon_sym_LBRACE] = ACTIONS(2199),
    [anon_sym_SEMI] = ACTIONS(2199),
    [anon_sym_exception] = ACTIONS(2199),
    [anon_sym_module] = ACTIONS(2199),
    [anon_sym_open] = ACTIONS(2199),
    [anon_sym_include] = ACTIONS(2199),
    [anon_sym_class] = ACTIONS(2199),
    [anon_sym_val] = ACTIONS(2199),
    [anon_sym_end] = ACTIONS(2199),
    [anon_sym_object] = ACTIONS(2199),
    [anon_sym_AMP] = ACTIONS(2199),
    [anon_sym_POUND] = ACTIONS(2199),
    [anon_sym_COLON_COLON] = ACTIONS(2201),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2201),
    [anon_sym_else] = ACTIONS(2199),
    [anon_sym_new] = ACTIONS(2199),
    [anon_sym_LBRACE_LT] = ACTIONS(2201),
    [anon_sym_begin] = ACTIONS(2199),
    [sym_ocamlyacc_value] = ACTIONS(2201),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2199),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2201),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2199),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2199),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2201),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2201),
    [aux_sym_number_token1] = ACTIONS(2201),
    [aux_sym_signed_number_token1] = ACTIONS(2199),
    [anon_sym_SQUOTE] = ACTIONS(2201),
    [anon_sym_DQUOTE] = ACTIONS(2201),
    [sym_prefix_operator] = ACTIONS(2201),
    [aux_sym_sign_operator_token1] = ACTIONS(2199),
    [sym_hash_operator] = ACTIONS(2201),
    [anon_sym_lsl] = ACTIONS(2199),
    [anon_sym_lsr] = ACTIONS(2199),
    [anon_sym_asr] = ACTIONS(2199),
    [aux_sym_pow_operator_token1] = ACTIONS(2201),
    [anon_sym_mod] = ACTIONS(2199),
    [anon_sym_land] = ACTIONS(2199),
    [anon_sym_lor] = ACTIONS(2199),
    [anon_sym_lxor] = ACTIONS(2199),
    [aux_sym_mult_operator_token1] = ACTIONS(2199),
    [aux_sym_add_operator_token1] = ACTIONS(2199),
    [sym_concat_operator] = ACTIONS(2201),
    [sym_rel_operator] = ACTIONS(2199),
    [anon_sym_AMP_AMP] = ACTIONS(2199),
    [anon_sym_or] = ACTIONS(2199),
    [anon_sym_PIPE_PIPE] = ACTIONS(2199),
    [sym_assign_operator] = ACTIONS(2201),
    [sym_let_operator] = ACTIONS(2201),
    [sym__uppercase_identifier] = ACTIONS(2201),
    [aux_sym_directive_token1] = ACTIONS(2199),
    [aux_sym_tag_token1] = ACTIONS(2201),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1373)] = {
    [sym_attribute] = STATE(1373),
    [ts_builtin_sym_end] = ACTIONS(1921),
    [sym__lowercase_identifier] = ACTIONS(1919),
    [anon_sym_SEMI_SEMI] = ACTIONS(1921),
    [anon_sym_let] = ACTIONS(1919),
    [anon_sym_and] = ACTIONS(1919),
    [anon_sym_TILDE] = ACTIONS(1919),
    [anon_sym_QMARK] = ACTIONS(1919),
    [anon_sym_LPAREN] = ACTIONS(1921),
    [anon_sym_external] = ACTIONS(1919),
    [anon_sym_type] = ACTIONS(1919),
    [anon_sym_COMMA] = ACTIONS(1921),
    [anon_sym_PIPE] = ACTIONS(1919),
    [anon_sym_DOT] = ACTIONS(1921),
    [anon_sym_LBRACK] = ACTIONS(1919),
    [anon_sym_true] = ACTIONS(1919),
    [anon_sym_false] = ACTIONS(1919),
    [anon_sym_LBRACE] = ACTIONS(1919),
    [anon_sym_SEMI] = ACTIONS(1919),
    [anon_sym_exception] = ACTIONS(1919),
    [anon_sym_module] = ACTIONS(1919),
    [anon_sym_open] = ACTIONS(1919),
    [anon_sym_include] = ACTIONS(1919),
    [anon_sym_class] = ACTIONS(1919),
    [anon_sym_val] = ACTIONS(1919),
    [anon_sym_object] = ACTIONS(1919),
    [anon_sym_AMP] = ACTIONS(1919),
    [anon_sym_POUND] = ACTIONS(1919),
    [anon_sym_COLON_COLON] = ACTIONS(1921),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1921),
    [anon_sym_LT_DASH] = ACTIONS(1919),
    [anon_sym_else] = ACTIONS(1919),
    [anon_sym_new] = ACTIONS(1919),
    [anon_sym_LBRACE_LT] = ACTIONS(1921),
    [anon_sym_begin] = ACTIONS(1919),
    [sym_ocamlyacc_value] = ACTIONS(1921),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1919),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1921),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1919),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1919),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1921),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1921),
    [aux_sym_number_token1] = ACTIONS(1921),
    [aux_sym_signed_number_token1] = ACTIONS(1919),
    [anon_sym_SQUOTE] = ACTIONS(1921),
    [anon_sym_DQUOTE] = ACTIONS(1921),
    [sym_prefix_operator] = ACTIONS(1921),
    [aux_sym_sign_operator_token1] = ACTIONS(1919),
    [sym_hash_operator] = ACTIONS(1921),
    [anon_sym_lsl] = ACTIONS(1919),
    [anon_sym_lsr] = ACTIONS(1919),
    [anon_sym_asr] = ACTIONS(1919),
    [aux_sym_pow_operator_token1] = ACTIONS(1921),
    [anon_sym_mod] = ACTIONS(1919),
    [anon_sym_land] = ACTIONS(1919),
    [anon_sym_lor] = ACTIONS(1919),
    [anon_sym_lxor] = ACTIONS(1919),
    [aux_sym_mult_operator_token1] = ACTIONS(1919),
    [aux_sym_add_operator_token1] = ACTIONS(1919),
    [sym_concat_operator] = ACTIONS(1921),
    [sym_rel_operator] = ACTIONS(1919),
    [anon_sym_AMP_AMP] = ACTIONS(1919),
    [anon_sym_or] = ACTIONS(1919),
    [anon_sym_PIPE_PIPE] = ACTIONS(1919),
    [sym_assign_operator] = ACTIONS(1921),
    [sym_let_operator] = ACTIONS(1921),
    [sym__uppercase_identifier] = ACTIONS(1921),
    [aux_sym_directive_token1] = ACTIONS(1919),
    [aux_sym_tag_token1] = ACTIONS(1921),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1374)] = {
    [sym_attribute] = STATE(1374),
    [sym__lowercase_identifier] = ACTIONS(2141),
    [anon_sym_SEMI_SEMI] = ACTIONS(2143),
    [anon_sym_let] = ACTIONS(2141),
    [anon_sym_and] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_QMARK] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2143),
    [anon_sym_external] = ACTIONS(2141),
    [anon_sym_type] = ACTIONS(2141),
    [anon_sym_COMMA] = ACTIONS(2143),
    [anon_sym_PIPE] = ACTIONS(2141),
    [anon_sym_DOT] = ACTIONS(2143),
    [anon_sym_LBRACK] = ACTIONS(2141),
    [anon_sym_RBRACK] = ACTIONS(2143),
    [anon_sym_true] = ACTIONS(2141),
    [anon_sym_false] = ACTIONS(2141),
    [anon_sym_LBRACE] = ACTIONS(2141),
    [anon_sym_SEMI] = ACTIONS(2141),
    [anon_sym_exception] = ACTIONS(2141),
    [anon_sym_module] = ACTIONS(2141),
    [anon_sym_open] = ACTIONS(2141),
    [anon_sym_include] = ACTIONS(2141),
    [anon_sym_class] = ACTIONS(2141),
    [anon_sym_val] = ACTIONS(2141),
    [anon_sym_end] = ACTIONS(2141),
    [anon_sym_object] = ACTIONS(2141),
    [anon_sym_AMP] = ACTIONS(2141),
    [anon_sym_POUND] = ACTIONS(2141),
    [anon_sym_COLON_COLON] = ACTIONS(2143),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2143),
    [anon_sym_else] = ACTIONS(2141),
    [anon_sym_new] = ACTIONS(2141),
    [anon_sym_LBRACE_LT] = ACTIONS(2143),
    [anon_sym_begin] = ACTIONS(2141),
    [sym_ocamlyacc_value] = ACTIONS(2143),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2141),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2143),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2141),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2141),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2143),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2143),
    [aux_sym_number_token1] = ACTIONS(2143),
    [aux_sym_signed_number_token1] = ACTIONS(2141),
    [anon_sym_SQUOTE] = ACTIONS(2143),
    [anon_sym_DQUOTE] = ACTIONS(2143),
    [sym_prefix_operator] = ACTIONS(2143),
    [aux_sym_sign_operator_token1] = ACTIONS(2141),
    [sym_hash_operator] = ACTIONS(2143),
    [anon_sym_lsl] = ACTIONS(2141),
    [anon_sym_lsr] = ACTIONS(2141),
    [anon_sym_asr] = ACTIONS(2141),
    [aux_sym_pow_operator_token1] = ACTIONS(2143),
    [anon_sym_mod] = ACTIONS(2141),
    [anon_sym_land] = ACTIONS(2141),
    [anon_sym_lor] = ACTIONS(2141),
    [anon_sym_lxor] = ACTIONS(2141),
    [aux_sym_mult_operator_token1] = ACTIONS(2141),
    [aux_sym_add_operator_token1] = ACTIONS(2141),
    [sym_concat_operator] = ACTIONS(2143),
    [sym_rel_operator] = ACTIONS(2141),
    [anon_sym_AMP_AMP] = ACTIONS(2141),
    [anon_sym_or] = ACTIONS(2141),
    [anon_sym_PIPE_PIPE] = ACTIONS(2141),
    [sym_assign_operator] = ACTIONS(2143),
    [sym_let_operator] = ACTIONS(2143),
    [sym__uppercase_identifier] = ACTIONS(2143),
    [aux_sym_directive_token1] = ACTIONS(2141),
    [aux_sym_tag_token1] = ACTIONS(2143),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1375)] = {
    [sym_attribute] = STATE(1375),
    [sym__lowercase_identifier] = ACTIONS(2125),
    [anon_sym_SEMI_SEMI] = ACTIONS(2127),
    [anon_sym_let] = ACTIONS(2125),
    [anon_sym_and] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(2125),
    [anon_sym_QMARK] = ACTIONS(2125),
    [anon_sym_LPAREN] = ACTIONS(2127),
    [anon_sym_external] = ACTIONS(2125),
    [anon_sym_type] = ACTIONS(2125),
    [anon_sym_COMMA] = ACTIONS(2127),
    [anon_sym_PIPE] = ACTIONS(2125),
    [anon_sym_DOT] = ACTIONS(2127),
    [anon_sym_LBRACK] = ACTIONS(2125),
    [anon_sym_RBRACK] = ACTIONS(2127),
    [anon_sym_true] = ACTIONS(2125),
    [anon_sym_false] = ACTIONS(2125),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_SEMI] = ACTIONS(2125),
    [anon_sym_exception] = ACTIONS(2125),
    [anon_sym_module] = ACTIONS(2125),
    [anon_sym_open] = ACTIONS(2125),
    [anon_sym_include] = ACTIONS(2125),
    [anon_sym_class] = ACTIONS(2125),
    [anon_sym_val] = ACTIONS(2125),
    [anon_sym_end] = ACTIONS(2125),
    [anon_sym_object] = ACTIONS(2125),
    [anon_sym_AMP] = ACTIONS(2125),
    [anon_sym_POUND] = ACTIONS(2125),
    [anon_sym_COLON_COLON] = ACTIONS(2127),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2127),
    [anon_sym_else] = ACTIONS(2125),
    [anon_sym_new] = ACTIONS(2125),
    [anon_sym_LBRACE_LT] = ACTIONS(2127),
    [anon_sym_begin] = ACTIONS(2125),
    [sym_ocamlyacc_value] = ACTIONS(2127),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2125),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2127),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2127),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2127),
    [aux_sym_number_token1] = ACTIONS(2127),
    [aux_sym_signed_number_token1] = ACTIONS(2125),
    [anon_sym_SQUOTE] = ACTIONS(2127),
    [anon_sym_DQUOTE] = ACTIONS(2127),
    [sym_prefix_operator] = ACTIONS(2127),
    [aux_sym_sign_operator_token1] = ACTIONS(2125),
    [sym_hash_operator] = ACTIONS(2127),
    [anon_sym_lsl] = ACTIONS(2125),
    [anon_sym_lsr] = ACTIONS(2125),
    [anon_sym_asr] = ACTIONS(2125),
    [aux_sym_pow_operator_token1] = ACTIONS(2127),
    [anon_sym_mod] = ACTIONS(2125),
    [anon_sym_land] = ACTIONS(2125),
    [anon_sym_lor] = ACTIONS(2125),
    [anon_sym_lxor] = ACTIONS(2125),
    [aux_sym_mult_operator_token1] = ACTIONS(2125),
    [aux_sym_add_operator_token1] = ACTIONS(2125),
    [sym_concat_operator] = ACTIONS(2127),
    [sym_rel_operator] = ACTIONS(2125),
    [anon_sym_AMP_AMP] = ACTIONS(2125),
    [anon_sym_or] = ACTIONS(2125),
    [anon_sym_PIPE_PIPE] = ACTIONS(2125),
    [sym_assign_operator] = ACTIONS(2127),
    [sym_let_operator] = ACTIONS(2127),
    [sym__uppercase_identifier] = ACTIONS(2127),
    [aux_sym_directive_token1] = ACTIONS(2125),
    [aux_sym_tag_token1] = ACTIONS(2127),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1376)] = {
    [sym_attribute] = STATE(1376),
    [sym__lowercase_identifier] = ACTIONS(2191),
    [anon_sym_SEMI_SEMI] = ACTIONS(2193),
    [anon_sym_let] = ACTIONS(2191),
    [anon_sym_and] = ACTIONS(2191),
    [anon_sym_TILDE] = ACTIONS(2191),
    [anon_sym_QMARK] = ACTIONS(2191),
    [anon_sym_LPAREN] = ACTIONS(2193),
    [anon_sym_external] = ACTIONS(2191),
    [anon_sym_type] = ACTIONS(2191),
    [anon_sym_COMMA] = ACTIONS(2193),
    [anon_sym_PIPE] = ACTIONS(2191),
    [anon_sym_DOT] = ACTIONS(2193),
    [anon_sym_LBRACK] = ACTIONS(2191),
    [anon_sym_RBRACK] = ACTIONS(2193),
    [anon_sym_true] = ACTIONS(2191),
    [anon_sym_false] = ACTIONS(2191),
    [anon_sym_LBRACE] = ACTIONS(2191),
    [anon_sym_SEMI] = ACTIONS(2191),
    [anon_sym_exception] = ACTIONS(2191),
    [anon_sym_module] = ACTIONS(2191),
    [anon_sym_open] = ACTIONS(2191),
    [anon_sym_include] = ACTIONS(2191),
    [anon_sym_class] = ACTIONS(2191),
    [anon_sym_val] = ACTIONS(2191),
    [anon_sym_end] = ACTIONS(2191),
    [anon_sym_object] = ACTIONS(2191),
    [anon_sym_AMP] = ACTIONS(2191),
    [anon_sym_POUND] = ACTIONS(2191),
    [anon_sym_COLON_COLON] = ACTIONS(2193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2193),
    [anon_sym_else] = ACTIONS(2191),
    [anon_sym_new] = ACTIONS(2191),
    [anon_sym_LBRACE_LT] = ACTIONS(2193),
    [anon_sym_begin] = ACTIONS(2191),
    [sym_ocamlyacc_value] = ACTIONS(2193),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2191),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2193),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2191),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2191),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2193),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2193),
    [aux_sym_number_token1] = ACTIONS(2193),
    [aux_sym_signed_number_token1] = ACTIONS(2191),
    [anon_sym_SQUOTE] = ACTIONS(2193),
    [anon_sym_DQUOTE] = ACTIONS(2193),
    [sym_prefix_operator] = ACTIONS(2193),
    [aux_sym_sign_operator_token1] = ACTIONS(2191),
    [sym_hash_operator] = ACTIONS(2193),
    [anon_sym_lsl] = ACTIONS(2191),
    [anon_sym_lsr] = ACTIONS(2191),
    [anon_sym_asr] = ACTIONS(2191),
    [aux_sym_pow_operator_token1] = ACTIONS(2193),
    [anon_sym_mod] = ACTIONS(2191),
    [anon_sym_land] = ACTIONS(2191),
    [anon_sym_lor] = ACTIONS(2191),
    [anon_sym_lxor] = ACTIONS(2191),
    [aux_sym_mult_operator_token1] = ACTIONS(2191),
    [aux_sym_add_operator_token1] = ACTIONS(2191),
    [sym_concat_operator] = ACTIONS(2193),
    [sym_rel_operator] = ACTIONS(2191),
    [anon_sym_AMP_AMP] = ACTIONS(2191),
    [anon_sym_or] = ACTIONS(2191),
    [anon_sym_PIPE_PIPE] = ACTIONS(2191),
    [sym_assign_operator] = ACTIONS(2193),
    [sym_let_operator] = ACTIONS(2193),
    [sym__uppercase_identifier] = ACTIONS(2193),
    [aux_sym_directive_token1] = ACTIONS(2191),
    [aux_sym_tag_token1] = ACTIONS(2193),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1377)] = {
    [sym_attribute] = STATE(1377),
    [sym__lowercase_identifier] = ACTIONS(1995),
    [anon_sym_SEMI_SEMI] = ACTIONS(1997),
    [anon_sym_let] = ACTIONS(1995),
    [anon_sym_and] = ACTIONS(1995),
    [anon_sym_TILDE] = ACTIONS(1995),
    [anon_sym_QMARK] = ACTIONS(1995),
    [anon_sym_LPAREN] = ACTIONS(1997),
    [anon_sym_external] = ACTIONS(1995),
    [anon_sym_type] = ACTIONS(1995),
    [anon_sym_COMMA] = ACTIONS(1997),
    [anon_sym_PIPE] = ACTIONS(1995),
    [anon_sym_DOT] = ACTIONS(1997),
    [anon_sym_LBRACK] = ACTIONS(1995),
    [anon_sym_RBRACK] = ACTIONS(1997),
    [anon_sym_true] = ACTIONS(1995),
    [anon_sym_false] = ACTIONS(1995),
    [anon_sym_LBRACE] = ACTIONS(1995),
    [anon_sym_SEMI] = ACTIONS(1995),
    [anon_sym_exception] = ACTIONS(1995),
    [anon_sym_module] = ACTIONS(1995),
    [anon_sym_open] = ACTIONS(1995),
    [anon_sym_include] = ACTIONS(1995),
    [anon_sym_class] = ACTIONS(1995),
    [anon_sym_val] = ACTIONS(1995),
    [anon_sym_end] = ACTIONS(1995),
    [anon_sym_object] = ACTIONS(1995),
    [anon_sym_AMP] = ACTIONS(1995),
    [anon_sym_POUND] = ACTIONS(1995),
    [anon_sym_COLON_COLON] = ACTIONS(1997),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1997),
    [anon_sym_else] = ACTIONS(1995),
    [anon_sym_new] = ACTIONS(1995),
    [anon_sym_LBRACE_LT] = ACTIONS(1997),
    [anon_sym_begin] = ACTIONS(1995),
    [sym_ocamlyacc_value] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1995),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1997),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1995),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1995),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1997),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1997),
    [aux_sym_number_token1] = ACTIONS(1997),
    [aux_sym_signed_number_token1] = ACTIONS(1995),
    [anon_sym_SQUOTE] = ACTIONS(1997),
    [anon_sym_DQUOTE] = ACTIONS(1997),
    [sym_prefix_operator] = ACTIONS(1997),
    [aux_sym_sign_operator_token1] = ACTIONS(1995),
    [sym_hash_operator] = ACTIONS(1997),
    [anon_sym_lsl] = ACTIONS(1995),
    [anon_sym_lsr] = ACTIONS(1995),
    [anon_sym_asr] = ACTIONS(1995),
    [aux_sym_pow_operator_token1] = ACTIONS(1997),
    [anon_sym_mod] = ACTIONS(1995),
    [anon_sym_land] = ACTIONS(1995),
    [anon_sym_lor] = ACTIONS(1995),
    [anon_sym_lxor] = ACTIONS(1995),
    [aux_sym_mult_operator_token1] = ACTIONS(1995),
    [aux_sym_add_operator_token1] = ACTIONS(1995),
    [sym_concat_operator] = ACTIONS(1997),
    [sym_rel_operator] = ACTIONS(1995),
    [anon_sym_AMP_AMP] = ACTIONS(1995),
    [anon_sym_or] = ACTIONS(1995),
    [anon_sym_PIPE_PIPE] = ACTIONS(1995),
    [sym_assign_operator] = ACTIONS(1997),
    [sym_let_operator] = ACTIONS(1997),
    [sym__uppercase_identifier] = ACTIONS(1997),
    [aux_sym_directive_token1] = ACTIONS(1995),
    [aux_sym_tag_token1] = ACTIONS(1997),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1378)] = {
    [sym_attribute] = STATE(1378),
    [sym__lowercase_identifier] = ACTIONS(2121),
    [anon_sym_SEMI_SEMI] = ACTIONS(2123),
    [anon_sym_let] = ACTIONS(2121),
    [anon_sym_and] = ACTIONS(2121),
    [anon_sym_TILDE] = ACTIONS(2121),
    [anon_sym_QMARK] = ACTIONS(2121),
    [anon_sym_LPAREN] = ACTIONS(2123),
    [anon_sym_external] = ACTIONS(2121),
    [anon_sym_type] = ACTIONS(2121),
    [anon_sym_COMMA] = ACTIONS(2123),
    [anon_sym_PIPE] = ACTIONS(2121),
    [anon_sym_DOT] = ACTIONS(2123),
    [anon_sym_LBRACK] = ACTIONS(2121),
    [anon_sym_RBRACK] = ACTIONS(2123),
    [anon_sym_true] = ACTIONS(2121),
    [anon_sym_false] = ACTIONS(2121),
    [anon_sym_LBRACE] = ACTIONS(2121),
    [anon_sym_SEMI] = ACTIONS(2121),
    [anon_sym_exception] = ACTIONS(2121),
    [anon_sym_module] = ACTIONS(2121),
    [anon_sym_open] = ACTIONS(2121),
    [anon_sym_include] = ACTIONS(2121),
    [anon_sym_class] = ACTIONS(2121),
    [anon_sym_val] = ACTIONS(2121),
    [anon_sym_end] = ACTIONS(2121),
    [anon_sym_object] = ACTIONS(2121),
    [anon_sym_AMP] = ACTIONS(2121),
    [anon_sym_POUND] = ACTIONS(2121),
    [anon_sym_COLON_COLON] = ACTIONS(2123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2123),
    [anon_sym_else] = ACTIONS(2121),
    [anon_sym_new] = ACTIONS(2121),
    [anon_sym_LBRACE_LT] = ACTIONS(2123),
    [anon_sym_begin] = ACTIONS(2121),
    [sym_ocamlyacc_value] = ACTIONS(2123),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2121),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2123),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2121),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2121),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2123),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2123),
    [aux_sym_number_token1] = ACTIONS(2123),
    [aux_sym_signed_number_token1] = ACTIONS(2121),
    [anon_sym_SQUOTE] = ACTIONS(2123),
    [anon_sym_DQUOTE] = ACTIONS(2123),
    [sym_prefix_operator] = ACTIONS(2123),
    [aux_sym_sign_operator_token1] = ACTIONS(2121),
    [sym_hash_operator] = ACTIONS(2123),
    [anon_sym_lsl] = ACTIONS(2121),
    [anon_sym_lsr] = ACTIONS(2121),
    [anon_sym_asr] = ACTIONS(2121),
    [aux_sym_pow_operator_token1] = ACTIONS(2123),
    [anon_sym_mod] = ACTIONS(2121),
    [anon_sym_land] = ACTIONS(2121),
    [anon_sym_lor] = ACTIONS(2121),
    [anon_sym_lxor] = ACTIONS(2121),
    [aux_sym_mult_operator_token1] = ACTIONS(2121),
    [aux_sym_add_operator_token1] = ACTIONS(2121),
    [sym_concat_operator] = ACTIONS(2123),
    [sym_rel_operator] = ACTIONS(2121),
    [anon_sym_AMP_AMP] = ACTIONS(2121),
    [anon_sym_or] = ACTIONS(2121),
    [anon_sym_PIPE_PIPE] = ACTIONS(2121),
    [sym_assign_operator] = ACTIONS(2123),
    [sym_let_operator] = ACTIONS(2123),
    [sym__uppercase_identifier] = ACTIONS(2123),
    [aux_sym_directive_token1] = ACTIONS(2121),
    [aux_sym_tag_token1] = ACTIONS(2123),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1379)] = {
    [sym_attribute] = STATE(1379),
    [sym__lowercase_identifier] = ACTIONS(2043),
    [anon_sym_SEMI_SEMI] = ACTIONS(2045),
    [anon_sym_let] = ACTIONS(2043),
    [anon_sym_and] = ACTIONS(2043),
    [anon_sym_TILDE] = ACTIONS(2043),
    [anon_sym_QMARK] = ACTIONS(2043),
    [anon_sym_LPAREN] = ACTIONS(2045),
    [anon_sym_external] = ACTIONS(2043),
    [anon_sym_type] = ACTIONS(2043),
    [anon_sym_COMMA] = ACTIONS(2045),
    [anon_sym_PIPE] = ACTIONS(2043),
    [anon_sym_DOT] = ACTIONS(2045),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_RBRACK] = ACTIONS(2045),
    [anon_sym_true] = ACTIONS(2043),
    [anon_sym_false] = ACTIONS(2043),
    [anon_sym_LBRACE] = ACTIONS(2043),
    [anon_sym_SEMI] = ACTIONS(2043),
    [anon_sym_exception] = ACTIONS(2043),
    [anon_sym_module] = ACTIONS(2043),
    [anon_sym_open] = ACTIONS(2043),
    [anon_sym_include] = ACTIONS(2043),
    [anon_sym_class] = ACTIONS(2043),
    [anon_sym_val] = ACTIONS(2043),
    [anon_sym_end] = ACTIONS(2043),
    [anon_sym_object] = ACTIONS(2043),
    [anon_sym_AMP] = ACTIONS(2043),
    [anon_sym_POUND] = ACTIONS(2043),
    [anon_sym_COLON_COLON] = ACTIONS(2045),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2045),
    [anon_sym_else] = ACTIONS(2043),
    [anon_sym_new] = ACTIONS(2043),
    [anon_sym_LBRACE_LT] = ACTIONS(2045),
    [anon_sym_begin] = ACTIONS(2043),
    [sym_ocamlyacc_value] = ACTIONS(2045),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2043),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2045),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2043),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2043),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2045),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2045),
    [aux_sym_number_token1] = ACTIONS(2045),
    [aux_sym_signed_number_token1] = ACTIONS(2043),
    [anon_sym_SQUOTE] = ACTIONS(2045),
    [anon_sym_DQUOTE] = ACTIONS(2045),
    [sym_prefix_operator] = ACTIONS(2045),
    [aux_sym_sign_operator_token1] = ACTIONS(2043),
    [sym_hash_operator] = ACTIONS(2045),
    [anon_sym_lsl] = ACTIONS(2043),
    [anon_sym_lsr] = ACTIONS(2043),
    [anon_sym_asr] = ACTIONS(2043),
    [aux_sym_pow_operator_token1] = ACTIONS(2045),
    [anon_sym_mod] = ACTIONS(2043),
    [anon_sym_land] = ACTIONS(2043),
    [anon_sym_lor] = ACTIONS(2043),
    [anon_sym_lxor] = ACTIONS(2043),
    [aux_sym_mult_operator_token1] = ACTIONS(2043),
    [aux_sym_add_operator_token1] = ACTIONS(2043),
    [sym_concat_operator] = ACTIONS(2045),
    [sym_rel_operator] = ACTIONS(2043),
    [anon_sym_AMP_AMP] = ACTIONS(2043),
    [anon_sym_or] = ACTIONS(2043),
    [anon_sym_PIPE_PIPE] = ACTIONS(2043),
    [sym_assign_operator] = ACTIONS(2045),
    [sym_let_operator] = ACTIONS(2045),
    [sym__uppercase_identifier] = ACTIONS(2045),
    [aux_sym_directive_token1] = ACTIONS(2043),
    [aux_sym_tag_token1] = ACTIONS(2045),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1380)] = {
    [sym_attribute] = STATE(1380),
    [sym__lowercase_identifier] = ACTIONS(2047),
    [anon_sym_SEMI_SEMI] = ACTIONS(2049),
    [anon_sym_let] = ACTIONS(2047),
    [anon_sym_and] = ACTIONS(2047),
    [anon_sym_TILDE] = ACTIONS(2047),
    [anon_sym_QMARK] = ACTIONS(2047),
    [anon_sym_LPAREN] = ACTIONS(2049),
    [anon_sym_external] = ACTIONS(2047),
    [anon_sym_type] = ACTIONS(2047),
    [anon_sym_COMMA] = ACTIONS(2049),
    [anon_sym_PIPE] = ACTIONS(2047),
    [anon_sym_DOT] = ACTIONS(2049),
    [anon_sym_LBRACK] = ACTIONS(2047),
    [anon_sym_RBRACK] = ACTIONS(2049),
    [anon_sym_true] = ACTIONS(2047),
    [anon_sym_false] = ACTIONS(2047),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_SEMI] = ACTIONS(2047),
    [anon_sym_exception] = ACTIONS(2047),
    [anon_sym_module] = ACTIONS(2047),
    [anon_sym_open] = ACTIONS(2047),
    [anon_sym_include] = ACTIONS(2047),
    [anon_sym_class] = ACTIONS(2047),
    [anon_sym_val] = ACTIONS(2047),
    [anon_sym_end] = ACTIONS(2047),
    [anon_sym_object] = ACTIONS(2047),
    [anon_sym_AMP] = ACTIONS(2047),
    [anon_sym_POUND] = ACTIONS(2047),
    [anon_sym_COLON_COLON] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2049),
    [anon_sym_else] = ACTIONS(2047),
    [anon_sym_new] = ACTIONS(2047),
    [anon_sym_LBRACE_LT] = ACTIONS(2049),
    [anon_sym_begin] = ACTIONS(2047),
    [sym_ocamlyacc_value] = ACTIONS(2049),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2047),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2049),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2047),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2047),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2049),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2049),
    [aux_sym_number_token1] = ACTIONS(2049),
    [aux_sym_signed_number_token1] = ACTIONS(2047),
    [anon_sym_SQUOTE] = ACTIONS(2049),
    [anon_sym_DQUOTE] = ACTIONS(2049),
    [sym_prefix_operator] = ACTIONS(2049),
    [aux_sym_sign_operator_token1] = ACTIONS(2047),
    [sym_hash_operator] = ACTIONS(2049),
    [anon_sym_lsl] = ACTIONS(2047),
    [anon_sym_lsr] = ACTIONS(2047),
    [anon_sym_asr] = ACTIONS(2047),
    [aux_sym_pow_operator_token1] = ACTIONS(2049),
    [anon_sym_mod] = ACTIONS(2047),
    [anon_sym_land] = ACTIONS(2047),
    [anon_sym_lor] = ACTIONS(2047),
    [anon_sym_lxor] = ACTIONS(2047),
    [aux_sym_mult_operator_token1] = ACTIONS(2047),
    [aux_sym_add_operator_token1] = ACTIONS(2047),
    [sym_concat_operator] = ACTIONS(2049),
    [sym_rel_operator] = ACTIONS(2047),
    [anon_sym_AMP_AMP] = ACTIONS(2047),
    [anon_sym_or] = ACTIONS(2047),
    [anon_sym_PIPE_PIPE] = ACTIONS(2047),
    [sym_assign_operator] = ACTIONS(2049),
    [sym_let_operator] = ACTIONS(2049),
    [sym__uppercase_identifier] = ACTIONS(2049),
    [aux_sym_directive_token1] = ACTIONS(2047),
    [aux_sym_tag_token1] = ACTIONS(2049),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1381)] = {
    [sym_attribute] = STATE(1381),
    [sym__lowercase_identifier] = ACTIONS(2071),
    [anon_sym_SEMI_SEMI] = ACTIONS(2073),
    [anon_sym_let] = ACTIONS(2071),
    [anon_sym_and] = ACTIONS(2071),
    [anon_sym_TILDE] = ACTIONS(2071),
    [anon_sym_QMARK] = ACTIONS(2071),
    [anon_sym_LPAREN] = ACTIONS(2073),
    [anon_sym_external] = ACTIONS(2071),
    [anon_sym_type] = ACTIONS(2071),
    [anon_sym_COMMA] = ACTIONS(2073),
    [anon_sym_PIPE] = ACTIONS(2071),
    [anon_sym_DOT] = ACTIONS(2073),
    [anon_sym_LBRACK] = ACTIONS(2071),
    [anon_sym_RBRACK] = ACTIONS(2073),
    [anon_sym_true] = ACTIONS(2071),
    [anon_sym_false] = ACTIONS(2071),
    [anon_sym_LBRACE] = ACTIONS(2071),
    [anon_sym_SEMI] = ACTIONS(2071),
    [anon_sym_exception] = ACTIONS(2071),
    [anon_sym_module] = ACTIONS(2071),
    [anon_sym_open] = ACTIONS(2071),
    [anon_sym_include] = ACTIONS(2071),
    [anon_sym_class] = ACTIONS(2071),
    [anon_sym_val] = ACTIONS(2071),
    [anon_sym_end] = ACTIONS(2071),
    [anon_sym_object] = ACTIONS(2071),
    [anon_sym_AMP] = ACTIONS(2071),
    [anon_sym_POUND] = ACTIONS(2071),
    [anon_sym_COLON_COLON] = ACTIONS(2073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2073),
    [anon_sym_else] = ACTIONS(2071),
    [anon_sym_new] = ACTIONS(2071),
    [anon_sym_LBRACE_LT] = ACTIONS(2073),
    [anon_sym_begin] = ACTIONS(2071),
    [sym_ocamlyacc_value] = ACTIONS(2073),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2071),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2073),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2071),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2071),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2073),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2073),
    [aux_sym_number_token1] = ACTIONS(2073),
    [aux_sym_signed_number_token1] = ACTIONS(2071),
    [anon_sym_SQUOTE] = ACTIONS(2073),
    [anon_sym_DQUOTE] = ACTIONS(2073),
    [sym_prefix_operator] = ACTIONS(2073),
    [aux_sym_sign_operator_token1] = ACTIONS(2071),
    [sym_hash_operator] = ACTIONS(2073),
    [anon_sym_lsl] = ACTIONS(2071),
    [anon_sym_lsr] = ACTIONS(2071),
    [anon_sym_asr] = ACTIONS(2071),
    [aux_sym_pow_operator_token1] = ACTIONS(2073),
    [anon_sym_mod] = ACTIONS(2071),
    [anon_sym_land] = ACTIONS(2071),
    [anon_sym_lor] = ACTIONS(2071),
    [anon_sym_lxor] = ACTIONS(2071),
    [aux_sym_mult_operator_token1] = ACTIONS(2071),
    [aux_sym_add_operator_token1] = ACTIONS(2071),
    [sym_concat_operator] = ACTIONS(2073),
    [sym_rel_operator] = ACTIONS(2071),
    [anon_sym_AMP_AMP] = ACTIONS(2071),
    [anon_sym_or] = ACTIONS(2071),
    [anon_sym_PIPE_PIPE] = ACTIONS(2071),
    [sym_assign_operator] = ACTIONS(2073),
    [sym_let_operator] = ACTIONS(2073),
    [sym__uppercase_identifier] = ACTIONS(2073),
    [aux_sym_directive_token1] = ACTIONS(2071),
    [aux_sym_tag_token1] = ACTIONS(2073),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1382)] = {
    [sym_attribute] = STATE(1382),
    [sym__lowercase_identifier] = ACTIONS(1963),
    [anon_sym_SEMI_SEMI] = ACTIONS(1965),
    [anon_sym_let] = ACTIONS(1963),
    [anon_sym_and] = ACTIONS(1963),
    [anon_sym_TILDE] = ACTIONS(1963),
    [anon_sym_QMARK] = ACTIONS(1963),
    [anon_sym_LPAREN] = ACTIONS(1965),
    [anon_sym_external] = ACTIONS(1963),
    [anon_sym_type] = ACTIONS(1963),
    [anon_sym_COMMA] = ACTIONS(1965),
    [anon_sym_PIPE] = ACTIONS(1963),
    [anon_sym_DOT] = ACTIONS(1965),
    [anon_sym_LBRACK] = ACTIONS(1963),
    [anon_sym_RBRACK] = ACTIONS(1965),
    [anon_sym_true] = ACTIONS(1963),
    [anon_sym_false] = ACTIONS(1963),
    [anon_sym_LBRACE] = ACTIONS(1963),
    [anon_sym_SEMI] = ACTIONS(1963),
    [anon_sym_exception] = ACTIONS(1963),
    [anon_sym_module] = ACTIONS(1963),
    [anon_sym_open] = ACTIONS(1963),
    [anon_sym_include] = ACTIONS(1963),
    [anon_sym_class] = ACTIONS(1963),
    [anon_sym_val] = ACTIONS(1963),
    [anon_sym_end] = ACTIONS(1963),
    [anon_sym_object] = ACTIONS(1963),
    [anon_sym_AMP] = ACTIONS(1963),
    [anon_sym_POUND] = ACTIONS(1963),
    [anon_sym_COLON_COLON] = ACTIONS(1965),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1965),
    [anon_sym_else] = ACTIONS(1963),
    [anon_sym_new] = ACTIONS(1963),
    [anon_sym_LBRACE_LT] = ACTIONS(1965),
    [anon_sym_begin] = ACTIONS(1963),
    [sym_ocamlyacc_value] = ACTIONS(1965),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1963),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1965),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1965),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1965),
    [aux_sym_signed_number_token1] = ACTIONS(1963),
    [anon_sym_SQUOTE] = ACTIONS(1965),
    [anon_sym_DQUOTE] = ACTIONS(1965),
    [sym_prefix_operator] = ACTIONS(1965),
    [aux_sym_sign_operator_token1] = ACTIONS(1963),
    [sym_hash_operator] = ACTIONS(1965),
    [anon_sym_lsl] = ACTIONS(1963),
    [anon_sym_lsr] = ACTIONS(1963),
    [anon_sym_asr] = ACTIONS(1963),
    [aux_sym_pow_operator_token1] = ACTIONS(1965),
    [anon_sym_mod] = ACTIONS(1963),
    [anon_sym_land] = ACTIONS(1963),
    [anon_sym_lor] = ACTIONS(1963),
    [anon_sym_lxor] = ACTIONS(1963),
    [aux_sym_mult_operator_token1] = ACTIONS(1963),
    [aux_sym_add_operator_token1] = ACTIONS(1963),
    [sym_concat_operator] = ACTIONS(1965),
    [sym_rel_operator] = ACTIONS(1963),
    [anon_sym_AMP_AMP] = ACTIONS(1963),
    [anon_sym_or] = ACTIONS(1963),
    [anon_sym_PIPE_PIPE] = ACTIONS(1963),
    [sym_assign_operator] = ACTIONS(1965),
    [sym_let_operator] = ACTIONS(1965),
    [sym__uppercase_identifier] = ACTIONS(1965),
    [aux_sym_directive_token1] = ACTIONS(1963),
    [aux_sym_tag_token1] = ACTIONS(1965),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1383)] = {
    [sym_attribute] = STATE(1383),
    [sym__lowercase_identifier] = ACTIONS(2157),
    [anon_sym_SEMI_SEMI] = ACTIONS(2159),
    [anon_sym_let] = ACTIONS(2157),
    [anon_sym_TILDE] = ACTIONS(2157),
    [anon_sym_QMARK] = ACTIONS(2157),
    [anon_sym_LPAREN] = ACTIONS(2159),
    [anon_sym_external] = ACTIONS(2157),
    [anon_sym_type] = ACTIONS(2157),
    [anon_sym_COMMA] = ACTIONS(2159),
    [anon_sym_PIPE] = ACTIONS(2157),
    [anon_sym_DOT] = ACTIONS(2159),
    [anon_sym_LBRACK] = ACTIONS(2157),
    [anon_sym_RBRACK] = ACTIONS(2159),
    [anon_sym_true] = ACTIONS(2157),
    [anon_sym_false] = ACTIONS(2157),
    [anon_sym_LBRACE] = ACTIONS(2157),
    [anon_sym_SEMI] = ACTIONS(2157),
    [anon_sym_exception] = ACTIONS(2157),
    [anon_sym_module] = ACTIONS(2157),
    [anon_sym_open] = ACTIONS(2157),
    [anon_sym_include] = ACTIONS(2157),
    [anon_sym_class] = ACTIONS(2157),
    [anon_sym_val] = ACTIONS(2157),
    [anon_sym_end] = ACTIONS(2157),
    [anon_sym_object] = ACTIONS(2157),
    [anon_sym_AMP] = ACTIONS(2157),
    [anon_sym_POUND] = ACTIONS(2157),
    [anon_sym_COLON_COLON] = ACTIONS(2159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2159),
    [anon_sym_LT_DASH] = ACTIONS(2317),
    [anon_sym_else] = ACTIONS(2157),
    [anon_sym_new] = ACTIONS(2157),
    [anon_sym_LBRACE_LT] = ACTIONS(2159),
    [anon_sym_begin] = ACTIONS(2157),
    [sym_ocamlyacc_value] = ACTIONS(2159),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2157),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2159),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2159),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2159),
    [aux_sym_number_token1] = ACTIONS(2159),
    [aux_sym_signed_number_token1] = ACTIONS(2157),
    [anon_sym_SQUOTE] = ACTIONS(2159),
    [anon_sym_DQUOTE] = ACTIONS(2159),
    [sym_prefix_operator] = ACTIONS(2159),
    [aux_sym_sign_operator_token1] = ACTIONS(2157),
    [sym_hash_operator] = ACTIONS(2159),
    [anon_sym_lsl] = ACTIONS(2157),
    [anon_sym_lsr] = ACTIONS(2157),
    [anon_sym_asr] = ACTIONS(2157),
    [aux_sym_pow_operator_token1] = ACTIONS(2159),
    [anon_sym_mod] = ACTIONS(2157),
    [anon_sym_land] = ACTIONS(2157),
    [anon_sym_lor] = ACTIONS(2157),
    [anon_sym_lxor] = ACTIONS(2157),
    [aux_sym_mult_operator_token1] = ACTIONS(2157),
    [aux_sym_add_operator_token1] = ACTIONS(2157),
    [sym_concat_operator] = ACTIONS(2159),
    [sym_rel_operator] = ACTIONS(2157),
    [anon_sym_AMP_AMP] = ACTIONS(2157),
    [anon_sym_or] = ACTIONS(2157),
    [anon_sym_PIPE_PIPE] = ACTIONS(2157),
    [sym_assign_operator] = ACTIONS(2159),
    [sym_let_operator] = ACTIONS(2159),
    [sym__uppercase_identifier] = ACTIONS(2159),
    [aux_sym_directive_token1] = ACTIONS(2157),
    [aux_sym_tag_token1] = ACTIONS(2159),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1384)] = {
    [sym_attribute] = STATE(1384),
    [sym__lowercase_identifier] = ACTIONS(2129),
    [anon_sym_SEMI_SEMI] = ACTIONS(2131),
    [anon_sym_let] = ACTIONS(2129),
    [anon_sym_and] = ACTIONS(2129),
    [anon_sym_TILDE] = ACTIONS(2129),
    [anon_sym_QMARK] = ACTIONS(2129),
    [anon_sym_LPAREN] = ACTIONS(2131),
    [anon_sym_external] = ACTIONS(2129),
    [anon_sym_type] = ACTIONS(2129),
    [anon_sym_COMMA] = ACTIONS(2131),
    [anon_sym_PIPE] = ACTIONS(2129),
    [anon_sym_DOT] = ACTIONS(2131),
    [anon_sym_LBRACK] = ACTIONS(2129),
    [anon_sym_RBRACK] = ACTIONS(2131),
    [anon_sym_true] = ACTIONS(2129),
    [anon_sym_false] = ACTIONS(2129),
    [anon_sym_LBRACE] = ACTIONS(2129),
    [anon_sym_SEMI] = ACTIONS(2129),
    [anon_sym_exception] = ACTIONS(2129),
    [anon_sym_module] = ACTIONS(2129),
    [anon_sym_open] = ACTIONS(2129),
    [anon_sym_include] = ACTIONS(2129),
    [anon_sym_class] = ACTIONS(2129),
    [anon_sym_val] = ACTIONS(2129),
    [anon_sym_end] = ACTIONS(2129),
    [anon_sym_object] = ACTIONS(2129),
    [anon_sym_AMP] = ACTIONS(2129),
    [anon_sym_POUND] = ACTIONS(2129),
    [anon_sym_COLON_COLON] = ACTIONS(2131),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2131),
    [anon_sym_else] = ACTIONS(2129),
    [anon_sym_new] = ACTIONS(2129),
    [anon_sym_LBRACE_LT] = ACTIONS(2131),
    [anon_sym_begin] = ACTIONS(2129),
    [sym_ocamlyacc_value] = ACTIONS(2131),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2129),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2131),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2129),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2129),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2131),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2131),
    [aux_sym_number_token1] = ACTIONS(2131),
    [aux_sym_signed_number_token1] = ACTIONS(2129),
    [anon_sym_SQUOTE] = ACTIONS(2131),
    [anon_sym_DQUOTE] = ACTIONS(2131),
    [sym_prefix_operator] = ACTIONS(2131),
    [aux_sym_sign_operator_token1] = ACTIONS(2129),
    [sym_hash_operator] = ACTIONS(2131),
    [anon_sym_lsl] = ACTIONS(2129),
    [anon_sym_lsr] = ACTIONS(2129),
    [anon_sym_asr] = ACTIONS(2129),
    [aux_sym_pow_operator_token1] = ACTIONS(2131),
    [anon_sym_mod] = ACTIONS(2129),
    [anon_sym_land] = ACTIONS(2129),
    [anon_sym_lor] = ACTIONS(2129),
    [anon_sym_lxor] = ACTIONS(2129),
    [aux_sym_mult_operator_token1] = ACTIONS(2129),
    [aux_sym_add_operator_token1] = ACTIONS(2129),
    [sym_concat_operator] = ACTIONS(2131),
    [sym_rel_operator] = ACTIONS(2129),
    [anon_sym_AMP_AMP] = ACTIONS(2129),
    [anon_sym_or] = ACTIONS(2129),
    [anon_sym_PIPE_PIPE] = ACTIONS(2129),
    [sym_assign_operator] = ACTIONS(2131),
    [sym_let_operator] = ACTIONS(2131),
    [sym__uppercase_identifier] = ACTIONS(2131),
    [aux_sym_directive_token1] = ACTIONS(2129),
    [aux_sym_tag_token1] = ACTIONS(2131),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1385)] = {
    [sym_attribute] = STATE(1385),
    [sym__lowercase_identifier] = ACTIONS(1935),
    [anon_sym_SEMI_SEMI] = ACTIONS(1937),
    [anon_sym_let] = ACTIONS(1935),
    [anon_sym_and] = ACTIONS(1935),
    [anon_sym_TILDE] = ACTIONS(1935),
    [anon_sym_QMARK] = ACTIONS(1935),
    [anon_sym_LPAREN] = ACTIONS(1937),
    [anon_sym_external] = ACTIONS(1935),
    [anon_sym_type] = ACTIONS(1935),
    [anon_sym_COMMA] = ACTIONS(1937),
    [anon_sym_PIPE] = ACTIONS(1935),
    [anon_sym_DOT] = ACTIONS(1937),
    [anon_sym_LBRACK] = ACTIONS(1935),
    [anon_sym_RBRACK] = ACTIONS(1937),
    [anon_sym_true] = ACTIONS(1935),
    [anon_sym_false] = ACTIONS(1935),
    [anon_sym_LBRACE] = ACTIONS(1935),
    [anon_sym_SEMI] = ACTIONS(1935),
    [anon_sym_exception] = ACTIONS(1935),
    [anon_sym_module] = ACTIONS(1935),
    [anon_sym_open] = ACTIONS(1935),
    [anon_sym_include] = ACTIONS(1935),
    [anon_sym_class] = ACTIONS(1935),
    [anon_sym_val] = ACTIONS(1935),
    [anon_sym_end] = ACTIONS(1935),
    [anon_sym_object] = ACTIONS(1935),
    [anon_sym_AMP] = ACTIONS(1935),
    [anon_sym_POUND] = ACTIONS(1935),
    [anon_sym_COLON_COLON] = ACTIONS(1937),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1937),
    [anon_sym_else] = ACTIONS(1935),
    [anon_sym_new] = ACTIONS(1935),
    [anon_sym_LBRACE_LT] = ACTIONS(1937),
    [anon_sym_begin] = ACTIONS(1935),
    [sym_ocamlyacc_value] = ACTIONS(1937),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1935),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1937),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1935),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1935),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1937),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1937),
    [aux_sym_number_token1] = ACTIONS(1937),
    [aux_sym_signed_number_token1] = ACTIONS(1935),
    [anon_sym_SQUOTE] = ACTIONS(1937),
    [anon_sym_DQUOTE] = ACTIONS(1937),
    [sym_prefix_operator] = ACTIONS(1937),
    [aux_sym_sign_operator_token1] = ACTIONS(1935),
    [sym_hash_operator] = ACTIONS(1937),
    [anon_sym_lsl] = ACTIONS(1935),
    [anon_sym_lsr] = ACTIONS(1935),
    [anon_sym_asr] = ACTIONS(1935),
    [aux_sym_pow_operator_token1] = ACTIONS(1937),
    [anon_sym_mod] = ACTIONS(1935),
    [anon_sym_land] = ACTIONS(1935),
    [anon_sym_lor] = ACTIONS(1935),
    [anon_sym_lxor] = ACTIONS(1935),
    [aux_sym_mult_operator_token1] = ACTIONS(1935),
    [aux_sym_add_operator_token1] = ACTIONS(1935),
    [sym_concat_operator] = ACTIONS(1937),
    [sym_rel_operator] = ACTIONS(1935),
    [anon_sym_AMP_AMP] = ACTIONS(1935),
    [anon_sym_or] = ACTIONS(1935),
    [anon_sym_PIPE_PIPE] = ACTIONS(1935),
    [sym_assign_operator] = ACTIONS(1937),
    [sym_let_operator] = ACTIONS(1937),
    [sym__uppercase_identifier] = ACTIONS(1937),
    [aux_sym_directive_token1] = ACTIONS(1935),
    [aux_sym_tag_token1] = ACTIONS(1937),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1386)] = {
    [sym_attribute] = STATE(1386),
    [sym__lowercase_identifier] = ACTIONS(1947),
    [anon_sym_SEMI_SEMI] = ACTIONS(1949),
    [anon_sym_let] = ACTIONS(1947),
    [anon_sym_TILDE] = ACTIONS(1947),
    [anon_sym_QMARK] = ACTIONS(1947),
    [anon_sym_LPAREN] = ACTIONS(1949),
    [anon_sym_external] = ACTIONS(1947),
    [anon_sym_type] = ACTIONS(1947),
    [anon_sym_COMMA] = ACTIONS(1949),
    [anon_sym_PIPE] = ACTIONS(1947),
    [anon_sym_DOT] = ACTIONS(1949),
    [anon_sym_LBRACK] = ACTIONS(1947),
    [anon_sym_RBRACK] = ACTIONS(1949),
    [anon_sym_true] = ACTIONS(1947),
    [anon_sym_false] = ACTIONS(1947),
    [anon_sym_LBRACE] = ACTIONS(1947),
    [anon_sym_SEMI] = ACTIONS(1947),
    [anon_sym_exception] = ACTIONS(1947),
    [anon_sym_module] = ACTIONS(1947),
    [anon_sym_open] = ACTIONS(1947),
    [anon_sym_include] = ACTIONS(1947),
    [anon_sym_class] = ACTIONS(1947),
    [anon_sym_val] = ACTIONS(1947),
    [anon_sym_end] = ACTIONS(1947),
    [anon_sym_object] = ACTIONS(1947),
    [anon_sym_AMP] = ACTIONS(1947),
    [anon_sym_POUND] = ACTIONS(1947),
    [anon_sym_COLON_COLON] = ACTIONS(1949),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1949),
    [anon_sym_LT_DASH] = ACTIONS(2616),
    [anon_sym_else] = ACTIONS(1947),
    [anon_sym_new] = ACTIONS(1947),
    [anon_sym_LBRACE_LT] = ACTIONS(1949),
    [anon_sym_begin] = ACTIONS(1947),
    [sym_ocamlyacc_value] = ACTIONS(1949),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1947),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1949),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1949),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1949),
    [aux_sym_number_token1] = ACTIONS(1949),
    [aux_sym_signed_number_token1] = ACTIONS(1947),
    [anon_sym_SQUOTE] = ACTIONS(1949),
    [anon_sym_DQUOTE] = ACTIONS(1949),
    [sym_prefix_operator] = ACTIONS(1949),
    [aux_sym_sign_operator_token1] = ACTIONS(1947),
    [sym_hash_operator] = ACTIONS(1949),
    [anon_sym_lsl] = ACTIONS(1947),
    [anon_sym_lsr] = ACTIONS(1947),
    [anon_sym_asr] = ACTIONS(1947),
    [aux_sym_pow_operator_token1] = ACTIONS(1949),
    [anon_sym_mod] = ACTIONS(1947),
    [anon_sym_land] = ACTIONS(1947),
    [anon_sym_lor] = ACTIONS(1947),
    [anon_sym_lxor] = ACTIONS(1947),
    [aux_sym_mult_operator_token1] = ACTIONS(1947),
    [aux_sym_add_operator_token1] = ACTIONS(1947),
    [sym_concat_operator] = ACTIONS(1949),
    [sym_rel_operator] = ACTIONS(1947),
    [anon_sym_AMP_AMP] = ACTIONS(1947),
    [anon_sym_or] = ACTIONS(1947),
    [anon_sym_PIPE_PIPE] = ACTIONS(1947),
    [sym_assign_operator] = ACTIONS(1949),
    [sym_let_operator] = ACTIONS(1949),
    [sym__uppercase_identifier] = ACTIONS(1949),
    [aux_sym_directive_token1] = ACTIONS(1947),
    [aux_sym_tag_token1] = ACTIONS(1949),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1387)] = {
    [sym_attribute] = STATE(1387),
    [sym__lowercase_identifier] = ACTIONS(1939),
    [anon_sym_SEMI_SEMI] = ACTIONS(1941),
    [anon_sym_let] = ACTIONS(1939),
    [anon_sym_and] = ACTIONS(1939),
    [anon_sym_TILDE] = ACTIONS(1939),
    [anon_sym_QMARK] = ACTIONS(1939),
    [anon_sym_LPAREN] = ACTIONS(1941),
    [anon_sym_external] = ACTIONS(1939),
    [anon_sym_type] = ACTIONS(1939),
    [anon_sym_COMMA] = ACTIONS(1941),
    [anon_sym_PIPE] = ACTIONS(1939),
    [anon_sym_DOT] = ACTIONS(1941),
    [anon_sym_LBRACK] = ACTIONS(1939),
    [anon_sym_RBRACK] = ACTIONS(1941),
    [anon_sym_true] = ACTIONS(1939),
    [anon_sym_false] = ACTIONS(1939),
    [anon_sym_LBRACE] = ACTIONS(1939),
    [anon_sym_SEMI] = ACTIONS(1939),
    [anon_sym_exception] = ACTIONS(1939),
    [anon_sym_module] = ACTIONS(1939),
    [anon_sym_open] = ACTIONS(1939),
    [anon_sym_include] = ACTIONS(1939),
    [anon_sym_class] = ACTIONS(1939),
    [anon_sym_val] = ACTIONS(1939),
    [anon_sym_end] = ACTIONS(1939),
    [anon_sym_object] = ACTIONS(1939),
    [anon_sym_AMP] = ACTIONS(1939),
    [anon_sym_POUND] = ACTIONS(1939),
    [anon_sym_COLON_COLON] = ACTIONS(1941),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1941),
    [anon_sym_else] = ACTIONS(1939),
    [anon_sym_new] = ACTIONS(1939),
    [anon_sym_LBRACE_LT] = ACTIONS(1941),
    [anon_sym_begin] = ACTIONS(1939),
    [sym_ocamlyacc_value] = ACTIONS(1941),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1939),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1941),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1939),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1939),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1941),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1941),
    [aux_sym_number_token1] = ACTIONS(1941),
    [aux_sym_signed_number_token1] = ACTIONS(1939),
    [anon_sym_SQUOTE] = ACTIONS(1941),
    [anon_sym_DQUOTE] = ACTIONS(1941),
    [sym_prefix_operator] = ACTIONS(1941),
    [aux_sym_sign_operator_token1] = ACTIONS(1939),
    [sym_hash_operator] = ACTIONS(1941),
    [anon_sym_lsl] = ACTIONS(1939),
    [anon_sym_lsr] = ACTIONS(1939),
    [anon_sym_asr] = ACTIONS(1939),
    [aux_sym_pow_operator_token1] = ACTIONS(1941),
    [anon_sym_mod] = ACTIONS(1939),
    [anon_sym_land] = ACTIONS(1939),
    [anon_sym_lor] = ACTIONS(1939),
    [anon_sym_lxor] = ACTIONS(1939),
    [aux_sym_mult_operator_token1] = ACTIONS(1939),
    [aux_sym_add_operator_token1] = ACTIONS(1939),
    [sym_concat_operator] = ACTIONS(1941),
    [sym_rel_operator] = ACTIONS(1939),
    [anon_sym_AMP_AMP] = ACTIONS(1939),
    [anon_sym_or] = ACTIONS(1939),
    [anon_sym_PIPE_PIPE] = ACTIONS(1939),
    [sym_assign_operator] = ACTIONS(1941),
    [sym_let_operator] = ACTIONS(1941),
    [sym__uppercase_identifier] = ACTIONS(1941),
    [aux_sym_directive_token1] = ACTIONS(1939),
    [aux_sym_tag_token1] = ACTIONS(1941),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1388)] = {
    [sym__parenthesized_abstract_type] = STATE(1729),
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5624),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1388),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2620),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_PIPE] = ACTIONS(1645),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_SEMI] = ACTIONS(1641),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_as] = ACTIONS(1645),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(1641),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1641),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1389)] = {
    [sym__parenthesized_abstract_type] = STATE(1689),
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5287),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1389),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2558),
    [anon_sym_RPAREN] = ACTIONS(1641),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_PIPE] = ACTIONS(1641),
    [anon_sym_COLON2] = ACTIONS(1645),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_as] = ACTIONS(1645),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1641),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1390)] = {
    [sym_attribute] = STATE(1390),
    [sym__lowercase_identifier] = ACTIONS(1943),
    [anon_sym_SEMI_SEMI] = ACTIONS(1945),
    [anon_sym_let] = ACTIONS(1943),
    [anon_sym_and] = ACTIONS(1943),
    [anon_sym_TILDE] = ACTIONS(1943),
    [anon_sym_QMARK] = ACTIONS(1943),
    [anon_sym_LPAREN] = ACTIONS(1945),
    [anon_sym_external] = ACTIONS(1943),
    [anon_sym_type] = ACTIONS(1943),
    [anon_sym_COMMA] = ACTIONS(1945),
    [anon_sym_PIPE] = ACTIONS(1943),
    [anon_sym_DOT] = ACTIONS(1945),
    [anon_sym_LBRACK] = ACTIONS(1943),
    [anon_sym_RBRACK] = ACTIONS(1945),
    [anon_sym_true] = ACTIONS(1943),
    [anon_sym_false] = ACTIONS(1943),
    [anon_sym_LBRACE] = ACTIONS(1943),
    [anon_sym_SEMI] = ACTIONS(1943),
    [anon_sym_exception] = ACTIONS(1943),
    [anon_sym_module] = ACTIONS(1943),
    [anon_sym_open] = ACTIONS(1943),
    [anon_sym_include] = ACTIONS(1943),
    [anon_sym_class] = ACTIONS(1943),
    [anon_sym_val] = ACTIONS(1943),
    [anon_sym_end] = ACTIONS(1943),
    [anon_sym_object] = ACTIONS(1943),
    [anon_sym_AMP] = ACTIONS(1943),
    [anon_sym_POUND] = ACTIONS(1943),
    [anon_sym_COLON_COLON] = ACTIONS(1945),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1945),
    [anon_sym_else] = ACTIONS(1943),
    [anon_sym_new] = ACTIONS(1943),
    [anon_sym_LBRACE_LT] = ACTIONS(1945),
    [anon_sym_begin] = ACTIONS(1943),
    [sym_ocamlyacc_value] = ACTIONS(1945),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1943),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1945),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1943),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1943),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1945),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1945),
    [aux_sym_number_token1] = ACTIONS(1945),
    [aux_sym_signed_number_token1] = ACTIONS(1943),
    [anon_sym_SQUOTE] = ACTIONS(1945),
    [anon_sym_DQUOTE] = ACTIONS(1945),
    [sym_prefix_operator] = ACTIONS(1945),
    [aux_sym_sign_operator_token1] = ACTIONS(1943),
    [sym_hash_operator] = ACTIONS(1945),
    [anon_sym_lsl] = ACTIONS(1943),
    [anon_sym_lsr] = ACTIONS(1943),
    [anon_sym_asr] = ACTIONS(1943),
    [aux_sym_pow_operator_token1] = ACTIONS(1945),
    [anon_sym_mod] = ACTIONS(1943),
    [anon_sym_land] = ACTIONS(1943),
    [anon_sym_lor] = ACTIONS(1943),
    [anon_sym_lxor] = ACTIONS(1943),
    [aux_sym_mult_operator_token1] = ACTIONS(1943),
    [aux_sym_add_operator_token1] = ACTIONS(1943),
    [sym_concat_operator] = ACTIONS(1945),
    [sym_rel_operator] = ACTIONS(1943),
    [anon_sym_AMP_AMP] = ACTIONS(1943),
    [anon_sym_or] = ACTIONS(1943),
    [anon_sym_PIPE_PIPE] = ACTIONS(1943),
    [sym_assign_operator] = ACTIONS(1945),
    [sym_let_operator] = ACTIONS(1945),
    [sym__uppercase_identifier] = ACTIONS(1945),
    [aux_sym_directive_token1] = ACTIONS(1943),
    [aux_sym_tag_token1] = ACTIONS(1945),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1391)] = {
    [sym_attribute] = STATE(1391),
    [sym__lowercase_identifier] = ACTIONS(2149),
    [anon_sym_SEMI_SEMI] = ACTIONS(2151),
    [anon_sym_let] = ACTIONS(2149),
    [anon_sym_and] = ACTIONS(2149),
    [anon_sym_TILDE] = ACTIONS(2149),
    [anon_sym_QMARK] = ACTIONS(2149),
    [anon_sym_LPAREN] = ACTIONS(2151),
    [anon_sym_external] = ACTIONS(2149),
    [anon_sym_type] = ACTIONS(2149),
    [anon_sym_COMMA] = ACTIONS(2151),
    [anon_sym_PIPE] = ACTIONS(2149),
    [anon_sym_DOT] = ACTIONS(2151),
    [anon_sym_LBRACK] = ACTIONS(2149),
    [anon_sym_RBRACK] = ACTIONS(2151),
    [anon_sym_true] = ACTIONS(2149),
    [anon_sym_false] = ACTIONS(2149),
    [anon_sym_LBRACE] = ACTIONS(2149),
    [anon_sym_SEMI] = ACTIONS(2149),
    [anon_sym_exception] = ACTIONS(2149),
    [anon_sym_module] = ACTIONS(2149),
    [anon_sym_open] = ACTIONS(2149),
    [anon_sym_include] = ACTIONS(2149),
    [anon_sym_class] = ACTIONS(2149),
    [anon_sym_val] = ACTIONS(2149),
    [anon_sym_end] = ACTIONS(2149),
    [anon_sym_object] = ACTIONS(2149),
    [anon_sym_AMP] = ACTIONS(2149),
    [anon_sym_POUND] = ACTIONS(2149),
    [anon_sym_COLON_COLON] = ACTIONS(2151),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2151),
    [anon_sym_else] = ACTIONS(2149),
    [anon_sym_new] = ACTIONS(2149),
    [anon_sym_LBRACE_LT] = ACTIONS(2151),
    [anon_sym_begin] = ACTIONS(2149),
    [sym_ocamlyacc_value] = ACTIONS(2151),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2149),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2151),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2149),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2151),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2151),
    [aux_sym_number_token1] = ACTIONS(2151),
    [aux_sym_signed_number_token1] = ACTIONS(2149),
    [anon_sym_SQUOTE] = ACTIONS(2151),
    [anon_sym_DQUOTE] = ACTIONS(2151),
    [sym_prefix_operator] = ACTIONS(2151),
    [aux_sym_sign_operator_token1] = ACTIONS(2149),
    [sym_hash_operator] = ACTIONS(2151),
    [anon_sym_lsl] = ACTIONS(2149),
    [anon_sym_lsr] = ACTIONS(2149),
    [anon_sym_asr] = ACTIONS(2149),
    [aux_sym_pow_operator_token1] = ACTIONS(2151),
    [anon_sym_mod] = ACTIONS(2149),
    [anon_sym_land] = ACTIONS(2149),
    [anon_sym_lor] = ACTIONS(2149),
    [anon_sym_lxor] = ACTIONS(2149),
    [aux_sym_mult_operator_token1] = ACTIONS(2149),
    [aux_sym_add_operator_token1] = ACTIONS(2149),
    [sym_concat_operator] = ACTIONS(2151),
    [sym_rel_operator] = ACTIONS(2149),
    [anon_sym_AMP_AMP] = ACTIONS(2149),
    [anon_sym_or] = ACTIONS(2149),
    [anon_sym_PIPE_PIPE] = ACTIONS(2149),
    [sym_assign_operator] = ACTIONS(2151),
    [sym_let_operator] = ACTIONS(2151),
    [sym__uppercase_identifier] = ACTIONS(2151),
    [aux_sym_directive_token1] = ACTIONS(2149),
    [aux_sym_tag_token1] = ACTIONS(2151),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1392)] = {
    [sym_attribute] = STATE(1392),
    [sym__lowercase_identifier] = ACTIONS(2137),
    [anon_sym_SEMI_SEMI] = ACTIONS(2139),
    [anon_sym_let] = ACTIONS(2137),
    [anon_sym_and] = ACTIONS(2137),
    [anon_sym_TILDE] = ACTIONS(2137),
    [anon_sym_QMARK] = ACTIONS(2137),
    [anon_sym_LPAREN] = ACTIONS(2139),
    [anon_sym_external] = ACTIONS(2137),
    [anon_sym_type] = ACTIONS(2137),
    [anon_sym_COMMA] = ACTIONS(2139),
    [anon_sym_PIPE] = ACTIONS(2137),
    [anon_sym_DOT] = ACTIONS(2139),
    [anon_sym_LBRACK] = ACTIONS(2137),
    [anon_sym_RBRACK] = ACTIONS(2139),
    [anon_sym_true] = ACTIONS(2137),
    [anon_sym_false] = ACTIONS(2137),
    [anon_sym_LBRACE] = ACTIONS(2137),
    [anon_sym_SEMI] = ACTIONS(2137),
    [anon_sym_exception] = ACTIONS(2137),
    [anon_sym_module] = ACTIONS(2137),
    [anon_sym_open] = ACTIONS(2137),
    [anon_sym_include] = ACTIONS(2137),
    [anon_sym_class] = ACTIONS(2137),
    [anon_sym_val] = ACTIONS(2137),
    [anon_sym_end] = ACTIONS(2137),
    [anon_sym_object] = ACTIONS(2137),
    [anon_sym_AMP] = ACTIONS(2137),
    [anon_sym_POUND] = ACTIONS(2137),
    [anon_sym_COLON_COLON] = ACTIONS(2139),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2139),
    [anon_sym_else] = ACTIONS(2137),
    [anon_sym_new] = ACTIONS(2137),
    [anon_sym_LBRACE_LT] = ACTIONS(2139),
    [anon_sym_begin] = ACTIONS(2137),
    [sym_ocamlyacc_value] = ACTIONS(2139),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2137),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2139),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2137),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2137),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2139),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2139),
    [aux_sym_number_token1] = ACTIONS(2139),
    [aux_sym_signed_number_token1] = ACTIONS(2137),
    [anon_sym_SQUOTE] = ACTIONS(2139),
    [anon_sym_DQUOTE] = ACTIONS(2139),
    [sym_prefix_operator] = ACTIONS(2139),
    [aux_sym_sign_operator_token1] = ACTIONS(2137),
    [sym_hash_operator] = ACTIONS(2139),
    [anon_sym_lsl] = ACTIONS(2137),
    [anon_sym_lsr] = ACTIONS(2137),
    [anon_sym_asr] = ACTIONS(2137),
    [aux_sym_pow_operator_token1] = ACTIONS(2139),
    [anon_sym_mod] = ACTIONS(2137),
    [anon_sym_land] = ACTIONS(2137),
    [anon_sym_lor] = ACTIONS(2137),
    [anon_sym_lxor] = ACTIONS(2137),
    [aux_sym_mult_operator_token1] = ACTIONS(2137),
    [aux_sym_add_operator_token1] = ACTIONS(2137),
    [sym_concat_operator] = ACTIONS(2139),
    [sym_rel_operator] = ACTIONS(2137),
    [anon_sym_AMP_AMP] = ACTIONS(2137),
    [anon_sym_or] = ACTIONS(2137),
    [anon_sym_PIPE_PIPE] = ACTIONS(2137),
    [sym_assign_operator] = ACTIONS(2139),
    [sym_let_operator] = ACTIONS(2139),
    [sym__uppercase_identifier] = ACTIONS(2139),
    [aux_sym_directive_token1] = ACTIONS(2137),
    [aux_sym_tag_token1] = ACTIONS(2139),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1393)] = {
    [sym_attribute] = STATE(1393),
    [sym__lowercase_identifier] = ACTIONS(1931),
    [anon_sym_SEMI_SEMI] = ACTIONS(1933),
    [anon_sym_let] = ACTIONS(1931),
    [anon_sym_and] = ACTIONS(1931),
    [anon_sym_TILDE] = ACTIONS(1931),
    [anon_sym_QMARK] = ACTIONS(1931),
    [anon_sym_LPAREN] = ACTIONS(1933),
    [anon_sym_external] = ACTIONS(1931),
    [anon_sym_type] = ACTIONS(1931),
    [anon_sym_COMMA] = ACTIONS(1933),
    [anon_sym_PIPE] = ACTIONS(1931),
    [anon_sym_DOT] = ACTIONS(1087),
    [anon_sym_LBRACK] = ACTIONS(1931),
    [anon_sym_RBRACK] = ACTIONS(1933),
    [anon_sym_true] = ACTIONS(1931),
    [anon_sym_false] = ACTIONS(1931),
    [anon_sym_LBRACE] = ACTIONS(1931),
    [anon_sym_SEMI] = ACTIONS(1931),
    [anon_sym_exception] = ACTIONS(1931),
    [anon_sym_module] = ACTIONS(1931),
    [anon_sym_open] = ACTIONS(1931),
    [anon_sym_include] = ACTIONS(1931),
    [anon_sym_class] = ACTIONS(1931),
    [anon_sym_val] = ACTIONS(1931),
    [anon_sym_end] = ACTIONS(1931),
    [anon_sym_object] = ACTIONS(1931),
    [anon_sym_AMP] = ACTIONS(1931),
    [anon_sym_POUND] = ACTIONS(1089),
    [anon_sym_COLON_COLON] = ACTIONS(1933),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1933),
    [anon_sym_else] = ACTIONS(1931),
    [anon_sym_new] = ACTIONS(1931),
    [anon_sym_LBRACE_LT] = ACTIONS(1933),
    [anon_sym_begin] = ACTIONS(1931),
    [sym_ocamlyacc_value] = ACTIONS(1933),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1931),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1933),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1933),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1933),
    [aux_sym_number_token1] = ACTIONS(1933),
    [aux_sym_signed_number_token1] = ACTIONS(1931),
    [anon_sym_SQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1933),
    [sym_prefix_operator] = ACTIONS(1933),
    [aux_sym_sign_operator_token1] = ACTIONS(1931),
    [sym_hash_operator] = ACTIONS(1091),
    [anon_sym_lsl] = ACTIONS(1931),
    [anon_sym_lsr] = ACTIONS(1931),
    [anon_sym_asr] = ACTIONS(1931),
    [aux_sym_pow_operator_token1] = ACTIONS(1933),
    [anon_sym_mod] = ACTIONS(1931),
    [anon_sym_land] = ACTIONS(1931),
    [anon_sym_lor] = ACTIONS(1931),
    [anon_sym_lxor] = ACTIONS(1931),
    [aux_sym_mult_operator_token1] = ACTIONS(1931),
    [aux_sym_add_operator_token1] = ACTIONS(1931),
    [sym_concat_operator] = ACTIONS(1933),
    [sym_rel_operator] = ACTIONS(1931),
    [anon_sym_AMP_AMP] = ACTIONS(1931),
    [anon_sym_or] = ACTIONS(1931),
    [anon_sym_PIPE_PIPE] = ACTIONS(1931),
    [sym_assign_operator] = ACTIONS(1933),
    [sym_let_operator] = ACTIONS(1933),
    [sym__uppercase_identifier] = ACTIONS(1933),
    [aux_sym_directive_token1] = ACTIONS(1931),
    [aux_sym_tag_token1] = ACTIONS(1933),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1394)] = {
    [sym_attribute] = STATE(1394),
    [sym__lowercase_identifier] = ACTIONS(1987),
    [anon_sym_SEMI_SEMI] = ACTIONS(1989),
    [anon_sym_let] = ACTIONS(1987),
    [anon_sym_and] = ACTIONS(1987),
    [anon_sym_TILDE] = ACTIONS(1987),
    [anon_sym_QMARK] = ACTIONS(1987),
    [anon_sym_LPAREN] = ACTIONS(1989),
    [anon_sym_external] = ACTIONS(1987),
    [anon_sym_type] = ACTIONS(1987),
    [anon_sym_COMMA] = ACTIONS(1989),
    [anon_sym_PIPE] = ACTIONS(1987),
    [anon_sym_DOT] = ACTIONS(1989),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_RBRACK] = ACTIONS(1989),
    [anon_sym_true] = ACTIONS(1987),
    [anon_sym_false] = ACTIONS(1987),
    [anon_sym_LBRACE] = ACTIONS(1987),
    [anon_sym_SEMI] = ACTIONS(1987),
    [anon_sym_exception] = ACTIONS(1987),
    [anon_sym_module] = ACTIONS(1987),
    [anon_sym_open] = ACTIONS(1987),
    [anon_sym_include] = ACTIONS(1987),
    [anon_sym_class] = ACTIONS(1987),
    [anon_sym_val] = ACTIONS(1987),
    [anon_sym_end] = ACTIONS(1987),
    [anon_sym_object] = ACTIONS(1987),
    [anon_sym_AMP] = ACTIONS(1987),
    [anon_sym_POUND] = ACTIONS(1987),
    [anon_sym_COLON_COLON] = ACTIONS(1989),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1989),
    [anon_sym_else] = ACTIONS(1987),
    [anon_sym_new] = ACTIONS(1987),
    [anon_sym_LBRACE_LT] = ACTIONS(1989),
    [anon_sym_begin] = ACTIONS(1987),
    [sym_ocamlyacc_value] = ACTIONS(1989),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1987),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1989),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1987),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1987),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1989),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1989),
    [aux_sym_number_token1] = ACTIONS(1989),
    [aux_sym_signed_number_token1] = ACTIONS(1987),
    [anon_sym_SQUOTE] = ACTIONS(1989),
    [anon_sym_DQUOTE] = ACTIONS(1989),
    [sym_prefix_operator] = ACTIONS(1989),
    [aux_sym_sign_operator_token1] = ACTIONS(1987),
    [sym_hash_operator] = ACTIONS(1989),
    [anon_sym_lsl] = ACTIONS(1987),
    [anon_sym_lsr] = ACTIONS(1987),
    [anon_sym_asr] = ACTIONS(1987),
    [aux_sym_pow_operator_token1] = ACTIONS(1989),
    [anon_sym_mod] = ACTIONS(1987),
    [anon_sym_land] = ACTIONS(1987),
    [anon_sym_lor] = ACTIONS(1987),
    [anon_sym_lxor] = ACTIONS(1987),
    [aux_sym_mult_operator_token1] = ACTIONS(1987),
    [aux_sym_add_operator_token1] = ACTIONS(1987),
    [sym_concat_operator] = ACTIONS(1989),
    [sym_rel_operator] = ACTIONS(1987),
    [anon_sym_AMP_AMP] = ACTIONS(1987),
    [anon_sym_or] = ACTIONS(1987),
    [anon_sym_PIPE_PIPE] = ACTIONS(1987),
    [sym_assign_operator] = ACTIONS(1989),
    [sym_let_operator] = ACTIONS(1989),
    [sym__uppercase_identifier] = ACTIONS(1989),
    [aux_sym_directive_token1] = ACTIONS(1987),
    [aux_sym_tag_token1] = ACTIONS(1989),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1395)] = {
    [sym_attribute] = STATE(1395),
    [sym__lowercase_identifier] = ACTIONS(1991),
    [anon_sym_SEMI_SEMI] = ACTIONS(1993),
    [anon_sym_let] = ACTIONS(1991),
    [anon_sym_and] = ACTIONS(1991),
    [anon_sym_TILDE] = ACTIONS(1991),
    [anon_sym_QMARK] = ACTIONS(1991),
    [anon_sym_LPAREN] = ACTIONS(1993),
    [anon_sym_external] = ACTIONS(1991),
    [anon_sym_type] = ACTIONS(1991),
    [anon_sym_COMMA] = ACTIONS(1993),
    [anon_sym_PIPE] = ACTIONS(1991),
    [anon_sym_DOT] = ACTIONS(1993),
    [anon_sym_LBRACK] = ACTIONS(1991),
    [anon_sym_RBRACK] = ACTIONS(1993),
    [anon_sym_true] = ACTIONS(1991),
    [anon_sym_false] = ACTIONS(1991),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_SEMI] = ACTIONS(1991),
    [anon_sym_exception] = ACTIONS(1991),
    [anon_sym_module] = ACTIONS(1991),
    [anon_sym_open] = ACTIONS(1991),
    [anon_sym_include] = ACTIONS(1991),
    [anon_sym_class] = ACTIONS(1991),
    [anon_sym_val] = ACTIONS(1991),
    [anon_sym_end] = ACTIONS(1991),
    [anon_sym_object] = ACTIONS(1991),
    [anon_sym_AMP] = ACTIONS(1991),
    [anon_sym_POUND] = ACTIONS(1991),
    [anon_sym_COLON_COLON] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1993),
    [anon_sym_else] = ACTIONS(1991),
    [anon_sym_new] = ACTIONS(1991),
    [anon_sym_LBRACE_LT] = ACTIONS(1993),
    [anon_sym_begin] = ACTIONS(1991),
    [sym_ocamlyacc_value] = ACTIONS(1993),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1991),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1993),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1991),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1991),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1993),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1993),
    [aux_sym_number_token1] = ACTIONS(1993),
    [aux_sym_signed_number_token1] = ACTIONS(1991),
    [anon_sym_SQUOTE] = ACTIONS(1993),
    [anon_sym_DQUOTE] = ACTIONS(1993),
    [sym_prefix_operator] = ACTIONS(1993),
    [aux_sym_sign_operator_token1] = ACTIONS(1991),
    [sym_hash_operator] = ACTIONS(1993),
    [anon_sym_lsl] = ACTIONS(1991),
    [anon_sym_lsr] = ACTIONS(1991),
    [anon_sym_asr] = ACTIONS(1991),
    [aux_sym_pow_operator_token1] = ACTIONS(1993),
    [anon_sym_mod] = ACTIONS(1991),
    [anon_sym_land] = ACTIONS(1991),
    [anon_sym_lor] = ACTIONS(1991),
    [anon_sym_lxor] = ACTIONS(1991),
    [aux_sym_mult_operator_token1] = ACTIONS(1991),
    [aux_sym_add_operator_token1] = ACTIONS(1991),
    [sym_concat_operator] = ACTIONS(1993),
    [sym_rel_operator] = ACTIONS(1991),
    [anon_sym_AMP_AMP] = ACTIONS(1991),
    [anon_sym_or] = ACTIONS(1991),
    [anon_sym_PIPE_PIPE] = ACTIONS(1991),
    [sym_assign_operator] = ACTIONS(1993),
    [sym_let_operator] = ACTIONS(1993),
    [sym__uppercase_identifier] = ACTIONS(1993),
    [aux_sym_directive_token1] = ACTIONS(1991),
    [aux_sym_tag_token1] = ACTIONS(1993),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1396)] = {
    [sym_attribute] = STATE(1396),
    [ts_builtin_sym_end] = ACTIONS(1909),
    [sym__lowercase_identifier] = ACTIONS(1907),
    [anon_sym_SEMI_SEMI] = ACTIONS(1909),
    [anon_sym_let] = ACTIONS(1907),
    [anon_sym_and] = ACTIONS(1907),
    [anon_sym_TILDE] = ACTIONS(1907),
    [anon_sym_QMARK] = ACTIONS(1907),
    [anon_sym_LPAREN] = ACTIONS(1909),
    [anon_sym_external] = ACTIONS(1907),
    [anon_sym_type] = ACTIONS(1907),
    [anon_sym_COMMA] = ACTIONS(1909),
    [anon_sym_PIPE] = ACTIONS(1907),
    [anon_sym_DOT] = ACTIONS(1909),
    [anon_sym_LBRACK] = ACTIONS(1907),
    [anon_sym_true] = ACTIONS(1907),
    [anon_sym_false] = ACTIONS(1907),
    [anon_sym_LBRACE] = ACTIONS(1907),
    [anon_sym_SEMI] = ACTIONS(1907),
    [anon_sym_exception] = ACTIONS(1907),
    [anon_sym_module] = ACTIONS(1907),
    [anon_sym_open] = ACTIONS(1907),
    [anon_sym_include] = ACTIONS(1907),
    [anon_sym_class] = ACTIONS(1907),
    [anon_sym_val] = ACTIONS(1907),
    [anon_sym_object] = ACTIONS(1907),
    [anon_sym_AMP] = ACTIONS(1907),
    [anon_sym_POUND] = ACTIONS(1907),
    [anon_sym_COLON_COLON] = ACTIONS(1909),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1909),
    [anon_sym_LT_DASH] = ACTIONS(1907),
    [anon_sym_else] = ACTIONS(1907),
    [anon_sym_new] = ACTIONS(1907),
    [anon_sym_LBRACE_LT] = ACTIONS(1909),
    [anon_sym_begin] = ACTIONS(1907),
    [sym_ocamlyacc_value] = ACTIONS(1909),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1907),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1909),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1907),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1907),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1909),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1909),
    [aux_sym_number_token1] = ACTIONS(1909),
    [aux_sym_signed_number_token1] = ACTIONS(1907),
    [anon_sym_SQUOTE] = ACTIONS(1909),
    [anon_sym_DQUOTE] = ACTIONS(1909),
    [sym_prefix_operator] = ACTIONS(1909),
    [aux_sym_sign_operator_token1] = ACTIONS(1907),
    [sym_hash_operator] = ACTIONS(1909),
    [anon_sym_lsl] = ACTIONS(1907),
    [anon_sym_lsr] = ACTIONS(1907),
    [anon_sym_asr] = ACTIONS(1907),
    [aux_sym_pow_operator_token1] = ACTIONS(1909),
    [anon_sym_mod] = ACTIONS(1907),
    [anon_sym_land] = ACTIONS(1907),
    [anon_sym_lor] = ACTIONS(1907),
    [anon_sym_lxor] = ACTIONS(1907),
    [aux_sym_mult_operator_token1] = ACTIONS(1907),
    [aux_sym_add_operator_token1] = ACTIONS(1907),
    [sym_concat_operator] = ACTIONS(1909),
    [sym_rel_operator] = ACTIONS(1907),
    [anon_sym_AMP_AMP] = ACTIONS(1907),
    [anon_sym_or] = ACTIONS(1907),
    [anon_sym_PIPE_PIPE] = ACTIONS(1907),
    [sym_assign_operator] = ACTIONS(1909),
    [sym_let_operator] = ACTIONS(1909),
    [sym__uppercase_identifier] = ACTIONS(1909),
    [aux_sym_directive_token1] = ACTIONS(1907),
    [aux_sym_tag_token1] = ACTIONS(1909),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1397)] = {
    [sym_attribute] = STATE(1397),
    [sym__lowercase_identifier] = ACTIONS(1999),
    [anon_sym_SEMI_SEMI] = ACTIONS(2001),
    [anon_sym_let] = ACTIONS(1999),
    [anon_sym_and] = ACTIONS(1999),
    [anon_sym_TILDE] = ACTIONS(1999),
    [anon_sym_QMARK] = ACTIONS(1999),
    [anon_sym_LPAREN] = ACTIONS(2001),
    [anon_sym_external] = ACTIONS(1999),
    [anon_sym_type] = ACTIONS(1999),
    [anon_sym_COMMA] = ACTIONS(2001),
    [anon_sym_PIPE] = ACTIONS(1999),
    [anon_sym_DOT] = ACTIONS(2001),
    [anon_sym_LBRACK] = ACTIONS(1999),
    [anon_sym_RBRACK] = ACTIONS(2001),
    [anon_sym_true] = ACTIONS(1999),
    [anon_sym_false] = ACTIONS(1999),
    [anon_sym_LBRACE] = ACTIONS(1999),
    [anon_sym_SEMI] = ACTIONS(1999),
    [anon_sym_exception] = ACTIONS(1999),
    [anon_sym_module] = ACTIONS(1999),
    [anon_sym_open] = ACTIONS(1999),
    [anon_sym_include] = ACTIONS(1999),
    [anon_sym_class] = ACTIONS(1999),
    [anon_sym_val] = ACTIONS(1999),
    [anon_sym_end] = ACTIONS(1999),
    [anon_sym_object] = ACTIONS(1999),
    [anon_sym_AMP] = ACTIONS(1999),
    [anon_sym_POUND] = ACTIONS(1999),
    [anon_sym_COLON_COLON] = ACTIONS(2001),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2001),
    [anon_sym_else] = ACTIONS(1999),
    [anon_sym_new] = ACTIONS(1999),
    [anon_sym_LBRACE_LT] = ACTIONS(2001),
    [anon_sym_begin] = ACTIONS(1999),
    [sym_ocamlyacc_value] = ACTIONS(2001),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2001),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1999),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1999),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2001),
    [aux_sym_number_token1] = ACTIONS(2001),
    [aux_sym_signed_number_token1] = ACTIONS(1999),
    [anon_sym_SQUOTE] = ACTIONS(2001),
    [anon_sym_DQUOTE] = ACTIONS(2001),
    [sym_prefix_operator] = ACTIONS(2001),
    [aux_sym_sign_operator_token1] = ACTIONS(1999),
    [sym_hash_operator] = ACTIONS(2001),
    [anon_sym_lsl] = ACTIONS(1999),
    [anon_sym_lsr] = ACTIONS(1999),
    [anon_sym_asr] = ACTIONS(1999),
    [aux_sym_pow_operator_token1] = ACTIONS(2001),
    [anon_sym_mod] = ACTIONS(1999),
    [anon_sym_land] = ACTIONS(1999),
    [anon_sym_lor] = ACTIONS(1999),
    [anon_sym_lxor] = ACTIONS(1999),
    [aux_sym_mult_operator_token1] = ACTIONS(1999),
    [aux_sym_add_operator_token1] = ACTIONS(1999),
    [sym_concat_operator] = ACTIONS(2001),
    [sym_rel_operator] = ACTIONS(1999),
    [anon_sym_AMP_AMP] = ACTIONS(1999),
    [anon_sym_or] = ACTIONS(1999),
    [anon_sym_PIPE_PIPE] = ACTIONS(1999),
    [sym_assign_operator] = ACTIONS(2001),
    [sym_let_operator] = ACTIONS(2001),
    [sym__uppercase_identifier] = ACTIONS(2001),
    [aux_sym_directive_token1] = ACTIONS(1999),
    [aux_sym_tag_token1] = ACTIONS(2001),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1398)] = {
    [sym_attribute] = STATE(1398),
    [sym__lowercase_identifier] = ACTIONS(2003),
    [anon_sym_SEMI_SEMI] = ACTIONS(2005),
    [anon_sym_let] = ACTIONS(2003),
    [anon_sym_and] = ACTIONS(2003),
    [anon_sym_TILDE] = ACTIONS(2003),
    [anon_sym_QMARK] = ACTIONS(2003),
    [anon_sym_LPAREN] = ACTIONS(2005),
    [anon_sym_external] = ACTIONS(2003),
    [anon_sym_type] = ACTIONS(2003),
    [anon_sym_COMMA] = ACTIONS(2005),
    [anon_sym_PIPE] = ACTIONS(2003),
    [anon_sym_DOT] = ACTIONS(2005),
    [anon_sym_LBRACK] = ACTIONS(2003),
    [anon_sym_RBRACK] = ACTIONS(2005),
    [anon_sym_true] = ACTIONS(2003),
    [anon_sym_false] = ACTIONS(2003),
    [anon_sym_LBRACE] = ACTIONS(2003),
    [anon_sym_SEMI] = ACTIONS(2003),
    [anon_sym_exception] = ACTIONS(2003),
    [anon_sym_module] = ACTIONS(2003),
    [anon_sym_open] = ACTIONS(2003),
    [anon_sym_include] = ACTIONS(2003),
    [anon_sym_class] = ACTIONS(2003),
    [anon_sym_val] = ACTIONS(2003),
    [anon_sym_end] = ACTIONS(2003),
    [anon_sym_object] = ACTIONS(2003),
    [anon_sym_AMP] = ACTIONS(2003),
    [anon_sym_POUND] = ACTIONS(2003),
    [anon_sym_COLON_COLON] = ACTIONS(2005),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2005),
    [anon_sym_else] = ACTIONS(2003),
    [anon_sym_new] = ACTIONS(2003),
    [anon_sym_LBRACE_LT] = ACTIONS(2005),
    [anon_sym_begin] = ACTIONS(2003),
    [sym_ocamlyacc_value] = ACTIONS(2005),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2003),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2005),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2005),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2005),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2003),
    [anon_sym_SQUOTE] = ACTIONS(2005),
    [anon_sym_DQUOTE] = ACTIONS(2005),
    [sym_prefix_operator] = ACTIONS(2005),
    [aux_sym_sign_operator_token1] = ACTIONS(2003),
    [sym_hash_operator] = ACTIONS(2005),
    [anon_sym_lsl] = ACTIONS(2003),
    [anon_sym_lsr] = ACTIONS(2003),
    [anon_sym_asr] = ACTIONS(2003),
    [aux_sym_pow_operator_token1] = ACTIONS(2005),
    [anon_sym_mod] = ACTIONS(2003),
    [anon_sym_land] = ACTIONS(2003),
    [anon_sym_lor] = ACTIONS(2003),
    [anon_sym_lxor] = ACTIONS(2003),
    [aux_sym_mult_operator_token1] = ACTIONS(2003),
    [aux_sym_add_operator_token1] = ACTIONS(2003),
    [sym_concat_operator] = ACTIONS(2005),
    [sym_rel_operator] = ACTIONS(2003),
    [anon_sym_AMP_AMP] = ACTIONS(2003),
    [anon_sym_or] = ACTIONS(2003),
    [anon_sym_PIPE_PIPE] = ACTIONS(2003),
    [sym_assign_operator] = ACTIONS(2005),
    [sym_let_operator] = ACTIONS(2005),
    [sym__uppercase_identifier] = ACTIONS(2005),
    [aux_sym_directive_token1] = ACTIONS(2003),
    [aux_sym_tag_token1] = ACTIONS(2005),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1399)] = {
    [sym_attribute] = STATE(1399),
    [sym__lowercase_identifier] = ACTIONS(2007),
    [anon_sym_SEMI_SEMI] = ACTIONS(2009),
    [anon_sym_let] = ACTIONS(2007),
    [anon_sym_and] = ACTIONS(2007),
    [anon_sym_TILDE] = ACTIONS(2007),
    [anon_sym_QMARK] = ACTIONS(2007),
    [anon_sym_LPAREN] = ACTIONS(2009),
    [anon_sym_external] = ACTIONS(2007),
    [anon_sym_type] = ACTIONS(2007),
    [anon_sym_COMMA] = ACTIONS(2009),
    [anon_sym_PIPE] = ACTIONS(2007),
    [anon_sym_DOT] = ACTIONS(2009),
    [anon_sym_LBRACK] = ACTIONS(2007),
    [anon_sym_RBRACK] = ACTIONS(2009),
    [anon_sym_true] = ACTIONS(2007),
    [anon_sym_false] = ACTIONS(2007),
    [anon_sym_LBRACE] = ACTIONS(2007),
    [anon_sym_SEMI] = ACTIONS(2007),
    [anon_sym_exception] = ACTIONS(2007),
    [anon_sym_module] = ACTIONS(2007),
    [anon_sym_open] = ACTIONS(2007),
    [anon_sym_include] = ACTIONS(2007),
    [anon_sym_class] = ACTIONS(2007),
    [anon_sym_val] = ACTIONS(2007),
    [anon_sym_end] = ACTIONS(2007),
    [anon_sym_object] = ACTIONS(2007),
    [anon_sym_AMP] = ACTIONS(2007),
    [anon_sym_POUND] = ACTIONS(2007),
    [anon_sym_COLON_COLON] = ACTIONS(2009),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2009),
    [anon_sym_else] = ACTIONS(2007),
    [anon_sym_new] = ACTIONS(2007),
    [anon_sym_LBRACE_LT] = ACTIONS(2009),
    [anon_sym_begin] = ACTIONS(2007),
    [sym_ocamlyacc_value] = ACTIONS(2009),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2007),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2009),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2007),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2007),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2009),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2009),
    [aux_sym_number_token1] = ACTIONS(2009),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2009),
    [sym_prefix_operator] = ACTIONS(2009),
    [aux_sym_sign_operator_token1] = ACTIONS(2007),
    [sym_hash_operator] = ACTIONS(2009),
    [anon_sym_lsl] = ACTIONS(2007),
    [anon_sym_lsr] = ACTIONS(2007),
    [anon_sym_asr] = ACTIONS(2007),
    [aux_sym_pow_operator_token1] = ACTIONS(2009),
    [anon_sym_mod] = ACTIONS(2007),
    [anon_sym_land] = ACTIONS(2007),
    [anon_sym_lor] = ACTIONS(2007),
    [anon_sym_lxor] = ACTIONS(2007),
    [aux_sym_mult_operator_token1] = ACTIONS(2007),
    [aux_sym_add_operator_token1] = ACTIONS(2007),
    [sym_concat_operator] = ACTIONS(2009),
    [sym_rel_operator] = ACTIONS(2007),
    [anon_sym_AMP_AMP] = ACTIONS(2007),
    [anon_sym_or] = ACTIONS(2007),
    [anon_sym_PIPE_PIPE] = ACTIONS(2007),
    [sym_assign_operator] = ACTIONS(2009),
    [sym_let_operator] = ACTIONS(2009),
    [sym__uppercase_identifier] = ACTIONS(2009),
    [aux_sym_directive_token1] = ACTIONS(2007),
    [aux_sym_tag_token1] = ACTIONS(2009),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1400)] = {
    [sym_attribute] = STATE(1400),
    [sym__lowercase_identifier] = ACTIONS(2011),
    [anon_sym_SEMI_SEMI] = ACTIONS(2013),
    [anon_sym_let] = ACTIONS(2011),
    [anon_sym_and] = ACTIONS(2011),
    [anon_sym_TILDE] = ACTIONS(2011),
    [anon_sym_QMARK] = ACTIONS(2011),
    [anon_sym_LPAREN] = ACTIONS(2013),
    [anon_sym_external] = ACTIONS(2011),
    [anon_sym_type] = ACTIONS(2011),
    [anon_sym_COMMA] = ACTIONS(2013),
    [anon_sym_PIPE] = ACTIONS(2011),
    [anon_sym_DOT] = ACTIONS(2013),
    [anon_sym_LBRACK] = ACTIONS(2011),
    [anon_sym_RBRACK] = ACTIONS(2013),
    [anon_sym_true] = ACTIONS(2011),
    [anon_sym_false] = ACTIONS(2011),
    [anon_sym_LBRACE] = ACTIONS(2011),
    [anon_sym_SEMI] = ACTIONS(2011),
    [anon_sym_exception] = ACTIONS(2011),
    [anon_sym_module] = ACTIONS(2011),
    [anon_sym_open] = ACTIONS(2011),
    [anon_sym_include] = ACTIONS(2011),
    [anon_sym_class] = ACTIONS(2011),
    [anon_sym_val] = ACTIONS(2011),
    [anon_sym_end] = ACTIONS(2011),
    [anon_sym_object] = ACTIONS(2011),
    [anon_sym_AMP] = ACTIONS(2011),
    [anon_sym_POUND] = ACTIONS(2011),
    [anon_sym_COLON_COLON] = ACTIONS(2013),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2013),
    [anon_sym_else] = ACTIONS(2011),
    [anon_sym_new] = ACTIONS(2011),
    [anon_sym_LBRACE_LT] = ACTIONS(2013),
    [anon_sym_begin] = ACTIONS(2011),
    [sym_ocamlyacc_value] = ACTIONS(2013),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2011),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2013),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2011),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2011),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2013),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2013),
    [aux_sym_number_token1] = ACTIONS(2013),
    [aux_sym_signed_number_token1] = ACTIONS(2011),
    [anon_sym_SQUOTE] = ACTIONS(2013),
    [anon_sym_DQUOTE] = ACTIONS(2013),
    [sym_prefix_operator] = ACTIONS(2013),
    [aux_sym_sign_operator_token1] = ACTIONS(2011),
    [sym_hash_operator] = ACTIONS(2013),
    [anon_sym_lsl] = ACTIONS(2011),
    [anon_sym_lsr] = ACTIONS(2011),
    [anon_sym_asr] = ACTIONS(2011),
    [aux_sym_pow_operator_token1] = ACTIONS(2013),
    [anon_sym_mod] = ACTIONS(2011),
    [anon_sym_land] = ACTIONS(2011),
    [anon_sym_lor] = ACTIONS(2011),
    [anon_sym_lxor] = ACTIONS(2011),
    [aux_sym_mult_operator_token1] = ACTIONS(2011),
    [aux_sym_add_operator_token1] = ACTIONS(2011),
    [sym_concat_operator] = ACTIONS(2013),
    [sym_rel_operator] = ACTIONS(2011),
    [anon_sym_AMP_AMP] = ACTIONS(2011),
    [anon_sym_or] = ACTIONS(2011),
    [anon_sym_PIPE_PIPE] = ACTIONS(2011),
    [sym_assign_operator] = ACTIONS(2013),
    [sym_let_operator] = ACTIONS(2013),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_directive_token1] = ACTIONS(2011),
    [aux_sym_tag_token1] = ACTIONS(2013),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1401)] = {
    [sym_attribute] = STATE(1401),
    [sym__lowercase_identifier] = ACTIONS(1951),
    [anon_sym_SEMI_SEMI] = ACTIONS(1953),
    [anon_sym_let] = ACTIONS(1951),
    [anon_sym_and] = ACTIONS(1951),
    [anon_sym_TILDE] = ACTIONS(1951),
    [anon_sym_QMARK] = ACTIONS(1951),
    [anon_sym_LPAREN] = ACTIONS(1953),
    [anon_sym_external] = ACTIONS(1951),
    [anon_sym_type] = ACTIONS(1951),
    [anon_sym_COMMA] = ACTIONS(1953),
    [anon_sym_PIPE] = ACTIONS(1951),
    [anon_sym_DOT] = ACTIONS(1953),
    [anon_sym_LBRACK] = ACTIONS(1951),
    [anon_sym_RBRACK] = ACTIONS(1953),
    [anon_sym_true] = ACTIONS(1951),
    [anon_sym_false] = ACTIONS(1951),
    [anon_sym_LBRACE] = ACTIONS(1951),
    [anon_sym_SEMI] = ACTIONS(1951),
    [anon_sym_exception] = ACTIONS(1951),
    [anon_sym_module] = ACTIONS(1951),
    [anon_sym_open] = ACTIONS(1951),
    [anon_sym_include] = ACTIONS(1951),
    [anon_sym_class] = ACTIONS(1951),
    [anon_sym_val] = ACTIONS(1951),
    [anon_sym_end] = ACTIONS(1951),
    [anon_sym_object] = ACTIONS(1951),
    [anon_sym_AMP] = ACTIONS(1951),
    [anon_sym_POUND] = ACTIONS(1951),
    [anon_sym_COLON_COLON] = ACTIONS(1953),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1953),
    [anon_sym_else] = ACTIONS(1951),
    [anon_sym_new] = ACTIONS(1951),
    [anon_sym_LBRACE_LT] = ACTIONS(1953),
    [anon_sym_begin] = ACTIONS(1951),
    [sym_ocamlyacc_value] = ACTIONS(1953),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1951),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1953),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1951),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1951),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1953),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1953),
    [aux_sym_number_token1] = ACTIONS(1953),
    [aux_sym_signed_number_token1] = ACTIONS(1951),
    [anon_sym_SQUOTE] = ACTIONS(1953),
    [anon_sym_DQUOTE] = ACTIONS(1953),
    [sym_prefix_operator] = ACTIONS(1953),
    [aux_sym_sign_operator_token1] = ACTIONS(1951),
    [sym_hash_operator] = ACTIONS(1953),
    [anon_sym_lsl] = ACTIONS(1951),
    [anon_sym_lsr] = ACTIONS(1951),
    [anon_sym_asr] = ACTIONS(1951),
    [aux_sym_pow_operator_token1] = ACTIONS(1953),
    [anon_sym_mod] = ACTIONS(1951),
    [anon_sym_land] = ACTIONS(1951),
    [anon_sym_lor] = ACTIONS(1951),
    [anon_sym_lxor] = ACTIONS(1951),
    [aux_sym_mult_operator_token1] = ACTIONS(1951),
    [aux_sym_add_operator_token1] = ACTIONS(1951),
    [sym_concat_operator] = ACTIONS(1953),
    [sym_rel_operator] = ACTIONS(1951),
    [anon_sym_AMP_AMP] = ACTIONS(1951),
    [anon_sym_or] = ACTIONS(1951),
    [anon_sym_PIPE_PIPE] = ACTIONS(1951),
    [sym_assign_operator] = ACTIONS(1953),
    [sym_let_operator] = ACTIONS(1953),
    [sym__uppercase_identifier] = ACTIONS(1953),
    [aux_sym_directive_token1] = ACTIONS(1951),
    [aux_sym_tag_token1] = ACTIONS(1953),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1402)] = {
    [sym_attribute] = STATE(1402),
    [sym__lowercase_identifier] = ACTIONS(2075),
    [anon_sym_SEMI_SEMI] = ACTIONS(2077),
    [anon_sym_let] = ACTIONS(2075),
    [anon_sym_and] = ACTIONS(2075),
    [anon_sym_TILDE] = ACTIONS(2075),
    [anon_sym_QMARK] = ACTIONS(2075),
    [anon_sym_LPAREN] = ACTIONS(2077),
    [anon_sym_external] = ACTIONS(2075),
    [anon_sym_type] = ACTIONS(2075),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_PIPE] = ACTIONS(2075),
    [anon_sym_DOT] = ACTIONS(2077),
    [anon_sym_LBRACK] = ACTIONS(2075),
    [anon_sym_RBRACK] = ACTIONS(2077),
    [anon_sym_true] = ACTIONS(2075),
    [anon_sym_false] = ACTIONS(2075),
    [anon_sym_LBRACE] = ACTIONS(2075),
    [anon_sym_SEMI] = ACTIONS(2075),
    [anon_sym_exception] = ACTIONS(2075),
    [anon_sym_module] = ACTIONS(2075),
    [anon_sym_open] = ACTIONS(2075),
    [anon_sym_include] = ACTIONS(2075),
    [anon_sym_class] = ACTIONS(2075),
    [anon_sym_val] = ACTIONS(2075),
    [anon_sym_end] = ACTIONS(2075),
    [anon_sym_object] = ACTIONS(2075),
    [anon_sym_AMP] = ACTIONS(2075),
    [anon_sym_POUND] = ACTIONS(2075),
    [anon_sym_COLON_COLON] = ACTIONS(2077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2077),
    [anon_sym_else] = ACTIONS(2075),
    [anon_sym_new] = ACTIONS(2075),
    [anon_sym_LBRACE_LT] = ACTIONS(2077),
    [anon_sym_begin] = ACTIONS(2075),
    [sym_ocamlyacc_value] = ACTIONS(2077),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2075),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2077),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2075),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2075),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2077),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2077),
    [aux_sym_number_token1] = ACTIONS(2077),
    [aux_sym_signed_number_token1] = ACTIONS(2075),
    [anon_sym_SQUOTE] = ACTIONS(2077),
    [anon_sym_DQUOTE] = ACTIONS(2077),
    [sym_prefix_operator] = ACTIONS(2077),
    [aux_sym_sign_operator_token1] = ACTIONS(2075),
    [sym_hash_operator] = ACTIONS(2077),
    [anon_sym_lsl] = ACTIONS(2075),
    [anon_sym_lsr] = ACTIONS(2075),
    [anon_sym_asr] = ACTIONS(2075),
    [aux_sym_pow_operator_token1] = ACTIONS(2077),
    [anon_sym_mod] = ACTIONS(2075),
    [anon_sym_land] = ACTIONS(2075),
    [anon_sym_lor] = ACTIONS(2075),
    [anon_sym_lxor] = ACTIONS(2075),
    [aux_sym_mult_operator_token1] = ACTIONS(2075),
    [aux_sym_add_operator_token1] = ACTIONS(2075),
    [sym_concat_operator] = ACTIONS(2077),
    [sym_rel_operator] = ACTIONS(2075),
    [anon_sym_AMP_AMP] = ACTIONS(2075),
    [anon_sym_or] = ACTIONS(2075),
    [anon_sym_PIPE_PIPE] = ACTIONS(2075),
    [sym_assign_operator] = ACTIONS(2077),
    [sym_let_operator] = ACTIONS(2077),
    [sym__uppercase_identifier] = ACTIONS(2077),
    [aux_sym_directive_token1] = ACTIONS(2075),
    [aux_sym_tag_token1] = ACTIONS(2077),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1403)] = {
    [sym_attribute] = STATE(1403),
    [sym__lowercase_identifier] = ACTIONS(1967),
    [anon_sym_SEMI_SEMI] = ACTIONS(1969),
    [anon_sym_let] = ACTIONS(1967),
    [anon_sym_and] = ACTIONS(1967),
    [anon_sym_TILDE] = ACTIONS(1967),
    [anon_sym_QMARK] = ACTIONS(1967),
    [anon_sym_LPAREN] = ACTIONS(1969),
    [anon_sym_external] = ACTIONS(1967),
    [anon_sym_type] = ACTIONS(1967),
    [anon_sym_COMMA] = ACTIONS(1969),
    [anon_sym_PIPE] = ACTIONS(1967),
    [anon_sym_DOT] = ACTIONS(1969),
    [anon_sym_LBRACK] = ACTIONS(1967),
    [anon_sym_RBRACK] = ACTIONS(1969),
    [anon_sym_true] = ACTIONS(1967),
    [anon_sym_false] = ACTIONS(1967),
    [anon_sym_LBRACE] = ACTIONS(1967),
    [anon_sym_SEMI] = ACTIONS(1967),
    [anon_sym_exception] = ACTIONS(1967),
    [anon_sym_module] = ACTIONS(1967),
    [anon_sym_open] = ACTIONS(1967),
    [anon_sym_include] = ACTIONS(1967),
    [anon_sym_class] = ACTIONS(1967),
    [anon_sym_val] = ACTIONS(1967),
    [anon_sym_end] = ACTIONS(1967),
    [anon_sym_object] = ACTIONS(1967),
    [anon_sym_AMP] = ACTIONS(1967),
    [anon_sym_POUND] = ACTIONS(1967),
    [anon_sym_COLON_COLON] = ACTIONS(1969),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1969),
    [anon_sym_else] = ACTIONS(1967),
    [anon_sym_new] = ACTIONS(1967),
    [anon_sym_LBRACE_LT] = ACTIONS(1969),
    [anon_sym_begin] = ACTIONS(1967),
    [sym_ocamlyacc_value] = ACTIONS(1969),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1967),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1969),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1969),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1969),
    [aux_sym_number_token1] = ACTIONS(1969),
    [aux_sym_signed_number_token1] = ACTIONS(1967),
    [anon_sym_SQUOTE] = ACTIONS(1969),
    [anon_sym_DQUOTE] = ACTIONS(1969),
    [sym_prefix_operator] = ACTIONS(1969),
    [aux_sym_sign_operator_token1] = ACTIONS(1967),
    [sym_hash_operator] = ACTIONS(1969),
    [anon_sym_lsl] = ACTIONS(1967),
    [anon_sym_lsr] = ACTIONS(1967),
    [anon_sym_asr] = ACTIONS(1967),
    [aux_sym_pow_operator_token1] = ACTIONS(1969),
    [anon_sym_mod] = ACTIONS(1967),
    [anon_sym_land] = ACTIONS(1967),
    [anon_sym_lor] = ACTIONS(1967),
    [anon_sym_lxor] = ACTIONS(1967),
    [aux_sym_mult_operator_token1] = ACTIONS(1967),
    [aux_sym_add_operator_token1] = ACTIONS(1967),
    [sym_concat_operator] = ACTIONS(1969),
    [sym_rel_operator] = ACTIONS(1967),
    [anon_sym_AMP_AMP] = ACTIONS(1967),
    [anon_sym_or] = ACTIONS(1967),
    [anon_sym_PIPE_PIPE] = ACTIONS(1967),
    [sym_assign_operator] = ACTIONS(1969),
    [sym_let_operator] = ACTIONS(1969),
    [sym__uppercase_identifier] = ACTIONS(1969),
    [aux_sym_directive_token1] = ACTIONS(1967),
    [aux_sym_tag_token1] = ACTIONS(1969),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1404)] = {
    [sym_attribute] = STATE(1404),
    [sym__lowercase_identifier] = ACTIONS(2015),
    [anon_sym_SEMI_SEMI] = ACTIONS(2017),
    [anon_sym_let] = ACTIONS(2015),
    [anon_sym_and] = ACTIONS(2015),
    [anon_sym_TILDE] = ACTIONS(2015),
    [anon_sym_QMARK] = ACTIONS(2015),
    [anon_sym_LPAREN] = ACTIONS(2017),
    [anon_sym_external] = ACTIONS(2015),
    [anon_sym_type] = ACTIONS(2015),
    [anon_sym_COMMA] = ACTIONS(2017),
    [anon_sym_PIPE] = ACTIONS(2015),
    [anon_sym_DOT] = ACTIONS(2017),
    [anon_sym_LBRACK] = ACTIONS(2015),
    [anon_sym_RBRACK] = ACTIONS(2017),
    [anon_sym_true] = ACTIONS(2015),
    [anon_sym_false] = ACTIONS(2015),
    [anon_sym_LBRACE] = ACTIONS(2015),
    [anon_sym_SEMI] = ACTIONS(2015),
    [anon_sym_exception] = ACTIONS(2015),
    [anon_sym_module] = ACTIONS(2015),
    [anon_sym_open] = ACTIONS(2015),
    [anon_sym_include] = ACTIONS(2015),
    [anon_sym_class] = ACTIONS(2015),
    [anon_sym_val] = ACTIONS(2015),
    [anon_sym_end] = ACTIONS(2015),
    [anon_sym_object] = ACTIONS(2015),
    [anon_sym_AMP] = ACTIONS(2015),
    [anon_sym_POUND] = ACTIONS(2015),
    [anon_sym_COLON_COLON] = ACTIONS(2017),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2017),
    [anon_sym_else] = ACTIONS(2015),
    [anon_sym_new] = ACTIONS(2015),
    [anon_sym_LBRACE_LT] = ACTIONS(2017),
    [anon_sym_begin] = ACTIONS(2015),
    [sym_ocamlyacc_value] = ACTIONS(2017),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2015),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2017),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2015),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2015),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2017),
    [aux_sym_number_token1] = ACTIONS(2017),
    [aux_sym_signed_number_token1] = ACTIONS(2015),
    [anon_sym_SQUOTE] = ACTIONS(2017),
    [anon_sym_DQUOTE] = ACTIONS(2017),
    [sym_prefix_operator] = ACTIONS(2017),
    [aux_sym_sign_operator_token1] = ACTIONS(2015),
    [sym_hash_operator] = ACTIONS(2017),
    [anon_sym_lsl] = ACTIONS(2015),
    [anon_sym_lsr] = ACTIONS(2015),
    [anon_sym_asr] = ACTIONS(2015),
    [aux_sym_pow_operator_token1] = ACTIONS(2017),
    [anon_sym_mod] = ACTIONS(2015),
    [anon_sym_land] = ACTIONS(2015),
    [anon_sym_lor] = ACTIONS(2015),
    [anon_sym_lxor] = ACTIONS(2015),
    [aux_sym_mult_operator_token1] = ACTIONS(2015),
    [aux_sym_add_operator_token1] = ACTIONS(2015),
    [sym_concat_operator] = ACTIONS(2017),
    [sym_rel_operator] = ACTIONS(2015),
    [anon_sym_AMP_AMP] = ACTIONS(2015),
    [anon_sym_or] = ACTIONS(2015),
    [anon_sym_PIPE_PIPE] = ACTIONS(2015),
    [sym_assign_operator] = ACTIONS(2017),
    [sym_let_operator] = ACTIONS(2017),
    [sym__uppercase_identifier] = ACTIONS(2017),
    [aux_sym_directive_token1] = ACTIONS(2015),
    [aux_sym_tag_token1] = ACTIONS(2017),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1405)] = {
    [sym_attribute] = STATE(1405),
    [sym__lowercase_identifier] = ACTIONS(2019),
    [anon_sym_SEMI_SEMI] = ACTIONS(2021),
    [anon_sym_let] = ACTIONS(2019),
    [anon_sym_and] = ACTIONS(2019),
    [anon_sym_TILDE] = ACTIONS(2019),
    [anon_sym_QMARK] = ACTIONS(2019),
    [anon_sym_LPAREN] = ACTIONS(2021),
    [anon_sym_external] = ACTIONS(2019),
    [anon_sym_type] = ACTIONS(2019),
    [anon_sym_COMMA] = ACTIONS(2021),
    [anon_sym_PIPE] = ACTIONS(2019),
    [anon_sym_DOT] = ACTIONS(2021),
    [anon_sym_LBRACK] = ACTIONS(2019),
    [anon_sym_RBRACK] = ACTIONS(2021),
    [anon_sym_true] = ACTIONS(2019),
    [anon_sym_false] = ACTIONS(2019),
    [anon_sym_LBRACE] = ACTIONS(2019),
    [anon_sym_SEMI] = ACTIONS(2019),
    [anon_sym_exception] = ACTIONS(2019),
    [anon_sym_module] = ACTIONS(2019),
    [anon_sym_open] = ACTIONS(2019),
    [anon_sym_include] = ACTIONS(2019),
    [anon_sym_class] = ACTIONS(2019),
    [anon_sym_val] = ACTIONS(2019),
    [anon_sym_end] = ACTIONS(2019),
    [anon_sym_object] = ACTIONS(2019),
    [anon_sym_AMP] = ACTIONS(2019),
    [anon_sym_POUND] = ACTIONS(2019),
    [anon_sym_COLON_COLON] = ACTIONS(2021),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2021),
    [anon_sym_else] = ACTIONS(2019),
    [anon_sym_new] = ACTIONS(2019),
    [anon_sym_LBRACE_LT] = ACTIONS(2021),
    [anon_sym_begin] = ACTIONS(2019),
    [sym_ocamlyacc_value] = ACTIONS(2021),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2019),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2021),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2021),
    [aux_sym_number_token1] = ACTIONS(2021),
    [aux_sym_signed_number_token1] = ACTIONS(2019),
    [anon_sym_SQUOTE] = ACTIONS(2021),
    [anon_sym_DQUOTE] = ACTIONS(2021),
    [sym_prefix_operator] = ACTIONS(2021),
    [aux_sym_sign_operator_token1] = ACTIONS(2019),
    [sym_hash_operator] = ACTIONS(2021),
    [anon_sym_lsl] = ACTIONS(2019),
    [anon_sym_lsr] = ACTIONS(2019),
    [anon_sym_asr] = ACTIONS(2019),
    [aux_sym_pow_operator_token1] = ACTIONS(2021),
    [anon_sym_mod] = ACTIONS(2019),
    [anon_sym_land] = ACTIONS(2019),
    [anon_sym_lor] = ACTIONS(2019),
    [anon_sym_lxor] = ACTIONS(2019),
    [aux_sym_mult_operator_token1] = ACTIONS(2019),
    [aux_sym_add_operator_token1] = ACTIONS(2019),
    [sym_concat_operator] = ACTIONS(2021),
    [sym_rel_operator] = ACTIONS(2019),
    [anon_sym_AMP_AMP] = ACTIONS(2019),
    [anon_sym_or] = ACTIONS(2019),
    [anon_sym_PIPE_PIPE] = ACTIONS(2019),
    [sym_assign_operator] = ACTIONS(2021),
    [sym_let_operator] = ACTIONS(2021),
    [sym__uppercase_identifier] = ACTIONS(2021),
    [aux_sym_directive_token1] = ACTIONS(2019),
    [aux_sym_tag_token1] = ACTIONS(2021),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1406)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5043),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1406),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_COMMA] = ACTIONS(2433),
    [anon_sym_PIPE] = ACTIONS(2433),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_RBRACK] = ACTIONS(2433),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_SEMI] = ACTIONS(2433),
    [anon_sym_RBRACE] = ACTIONS(2433),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_as] = ACTIONS(2439),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(2433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1407)] = {
    [sym_attribute] = STATE(1407),
    [sym__lowercase_identifier] = ACTIONS(2023),
    [anon_sym_SEMI_SEMI] = ACTIONS(2025),
    [anon_sym_let] = ACTIONS(2023),
    [anon_sym_and] = ACTIONS(2023),
    [anon_sym_TILDE] = ACTIONS(2023),
    [anon_sym_QMARK] = ACTIONS(2023),
    [anon_sym_LPAREN] = ACTIONS(2025),
    [anon_sym_external] = ACTIONS(2023),
    [anon_sym_type] = ACTIONS(2023),
    [anon_sym_COMMA] = ACTIONS(2025),
    [anon_sym_PIPE] = ACTIONS(2023),
    [anon_sym_DOT] = ACTIONS(2025),
    [anon_sym_LBRACK] = ACTIONS(2023),
    [anon_sym_RBRACK] = ACTIONS(2025),
    [anon_sym_true] = ACTIONS(2023),
    [anon_sym_false] = ACTIONS(2023),
    [anon_sym_LBRACE] = ACTIONS(2023),
    [anon_sym_SEMI] = ACTIONS(2023),
    [anon_sym_exception] = ACTIONS(2023),
    [anon_sym_module] = ACTIONS(2023),
    [anon_sym_open] = ACTIONS(2023),
    [anon_sym_include] = ACTIONS(2023),
    [anon_sym_class] = ACTIONS(2023),
    [anon_sym_val] = ACTIONS(2023),
    [anon_sym_end] = ACTIONS(2023),
    [anon_sym_object] = ACTIONS(2023),
    [anon_sym_AMP] = ACTIONS(2023),
    [anon_sym_POUND] = ACTIONS(2023),
    [anon_sym_COLON_COLON] = ACTIONS(2025),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2025),
    [anon_sym_else] = ACTIONS(2023),
    [anon_sym_new] = ACTIONS(2023),
    [anon_sym_LBRACE_LT] = ACTIONS(2025),
    [anon_sym_begin] = ACTIONS(2023),
    [sym_ocamlyacc_value] = ACTIONS(2025),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2023),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2025),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2025),
    [aux_sym_number_token1] = ACTIONS(2025),
    [aux_sym_signed_number_token1] = ACTIONS(2023),
    [anon_sym_SQUOTE] = ACTIONS(2025),
    [anon_sym_DQUOTE] = ACTIONS(2025),
    [sym_prefix_operator] = ACTIONS(2025),
    [aux_sym_sign_operator_token1] = ACTIONS(2023),
    [sym_hash_operator] = ACTIONS(2025),
    [anon_sym_lsl] = ACTIONS(2023),
    [anon_sym_lsr] = ACTIONS(2023),
    [anon_sym_asr] = ACTIONS(2023),
    [aux_sym_pow_operator_token1] = ACTIONS(2025),
    [anon_sym_mod] = ACTIONS(2023),
    [anon_sym_land] = ACTIONS(2023),
    [anon_sym_lor] = ACTIONS(2023),
    [anon_sym_lxor] = ACTIONS(2023),
    [aux_sym_mult_operator_token1] = ACTIONS(2023),
    [aux_sym_add_operator_token1] = ACTIONS(2023),
    [sym_concat_operator] = ACTIONS(2025),
    [sym_rel_operator] = ACTIONS(2023),
    [anon_sym_AMP_AMP] = ACTIONS(2023),
    [anon_sym_or] = ACTIONS(2023),
    [anon_sym_PIPE_PIPE] = ACTIONS(2023),
    [sym_assign_operator] = ACTIONS(2025),
    [sym_let_operator] = ACTIONS(2025),
    [sym__uppercase_identifier] = ACTIONS(2025),
    [aux_sym_directive_token1] = ACTIONS(2023),
    [aux_sym_tag_token1] = ACTIONS(2025),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1408)] = {
    [sym_attribute] = STATE(1408),
    [sym__lowercase_identifier] = ACTIONS(2161),
    [anon_sym_SEMI_SEMI] = ACTIONS(2163),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_and] = ACTIONS(2161),
    [anon_sym_TILDE] = ACTIONS(2161),
    [anon_sym_QMARK] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_external] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_COMMA] = ACTIONS(2163),
    [anon_sym_PIPE] = ACTIONS(2161),
    [anon_sym_DOT] = ACTIONS(2163),
    [anon_sym_LBRACK] = ACTIONS(2161),
    [anon_sym_RBRACK] = ACTIONS(2163),
    [anon_sym_true] = ACTIONS(2161),
    [anon_sym_false] = ACTIONS(2161),
    [anon_sym_LBRACE] = ACTIONS(2161),
    [anon_sym_SEMI] = ACTIONS(2161),
    [anon_sym_exception] = ACTIONS(2161),
    [anon_sym_module] = ACTIONS(2161),
    [anon_sym_open] = ACTIONS(2161),
    [anon_sym_include] = ACTIONS(2161),
    [anon_sym_class] = ACTIONS(2161),
    [anon_sym_val] = ACTIONS(2161),
    [anon_sym_end] = ACTIONS(2161),
    [anon_sym_object] = ACTIONS(2161),
    [anon_sym_AMP] = ACTIONS(2161),
    [anon_sym_POUND] = ACTIONS(2161),
    [anon_sym_COLON_COLON] = ACTIONS(2163),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2163),
    [anon_sym_else] = ACTIONS(2161),
    [anon_sym_new] = ACTIONS(2161),
    [anon_sym_LBRACE_LT] = ACTIONS(2163),
    [anon_sym_begin] = ACTIONS(2161),
    [sym_ocamlyacc_value] = ACTIONS(2163),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2161),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2163),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2163),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2163),
    [aux_sym_number_token1] = ACTIONS(2163),
    [aux_sym_signed_number_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2163),
    [anon_sym_DQUOTE] = ACTIONS(2163),
    [sym_prefix_operator] = ACTIONS(2163),
    [aux_sym_sign_operator_token1] = ACTIONS(2161),
    [sym_hash_operator] = ACTIONS(2163),
    [anon_sym_lsl] = ACTIONS(2161),
    [anon_sym_lsr] = ACTIONS(2161),
    [anon_sym_asr] = ACTIONS(2161),
    [aux_sym_pow_operator_token1] = ACTIONS(2163),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_land] = ACTIONS(2161),
    [anon_sym_lor] = ACTIONS(2161),
    [anon_sym_lxor] = ACTIONS(2161),
    [aux_sym_mult_operator_token1] = ACTIONS(2161),
    [aux_sym_add_operator_token1] = ACTIONS(2161),
    [sym_concat_operator] = ACTIONS(2163),
    [sym_rel_operator] = ACTIONS(2161),
    [anon_sym_AMP_AMP] = ACTIONS(2161),
    [anon_sym_or] = ACTIONS(2161),
    [anon_sym_PIPE_PIPE] = ACTIONS(2161),
    [sym_assign_operator] = ACTIONS(2163),
    [sym_let_operator] = ACTIONS(2163),
    [sym__uppercase_identifier] = ACTIONS(2163),
    [aux_sym_directive_token1] = ACTIONS(2161),
    [aux_sym_tag_token1] = ACTIONS(2163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1409)] = {
    [sym_attribute] = STATE(1409),
    [sym__lowercase_identifier] = ACTIONS(2027),
    [anon_sym_SEMI_SEMI] = ACTIONS(2029),
    [anon_sym_let] = ACTIONS(2027),
    [anon_sym_and] = ACTIONS(2027),
    [anon_sym_TILDE] = ACTIONS(2027),
    [anon_sym_QMARK] = ACTIONS(2027),
    [anon_sym_LPAREN] = ACTIONS(2029),
    [anon_sym_external] = ACTIONS(2027),
    [anon_sym_type] = ACTIONS(2027),
    [anon_sym_COMMA] = ACTIONS(2029),
    [anon_sym_PIPE] = ACTIONS(2027),
    [anon_sym_DOT] = ACTIONS(2029),
    [anon_sym_LBRACK] = ACTIONS(2027),
    [anon_sym_RBRACK] = ACTIONS(2029),
    [anon_sym_true] = ACTIONS(2027),
    [anon_sym_false] = ACTIONS(2027),
    [anon_sym_LBRACE] = ACTIONS(2027),
    [anon_sym_SEMI] = ACTIONS(2027),
    [anon_sym_exception] = ACTIONS(2027),
    [anon_sym_module] = ACTIONS(2027),
    [anon_sym_open] = ACTIONS(2027),
    [anon_sym_include] = ACTIONS(2027),
    [anon_sym_class] = ACTIONS(2027),
    [anon_sym_val] = ACTIONS(2027),
    [anon_sym_end] = ACTIONS(2027),
    [anon_sym_object] = ACTIONS(2027),
    [anon_sym_AMP] = ACTIONS(2027),
    [anon_sym_POUND] = ACTIONS(2027),
    [anon_sym_COLON_COLON] = ACTIONS(2029),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2029),
    [anon_sym_else] = ACTIONS(2027),
    [anon_sym_new] = ACTIONS(2027),
    [anon_sym_LBRACE_LT] = ACTIONS(2029),
    [anon_sym_begin] = ACTIONS(2027),
    [sym_ocamlyacc_value] = ACTIONS(2029),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2027),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2029),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2029),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2029),
    [aux_sym_number_token1] = ACTIONS(2029),
    [aux_sym_signed_number_token1] = ACTIONS(2027),
    [anon_sym_SQUOTE] = ACTIONS(2029),
    [anon_sym_DQUOTE] = ACTIONS(2029),
    [sym_prefix_operator] = ACTIONS(2029),
    [aux_sym_sign_operator_token1] = ACTIONS(2027),
    [sym_hash_operator] = ACTIONS(2029),
    [anon_sym_lsl] = ACTIONS(2027),
    [anon_sym_lsr] = ACTIONS(2027),
    [anon_sym_asr] = ACTIONS(2027),
    [aux_sym_pow_operator_token1] = ACTIONS(2029),
    [anon_sym_mod] = ACTIONS(2027),
    [anon_sym_land] = ACTIONS(2027),
    [anon_sym_lor] = ACTIONS(2027),
    [anon_sym_lxor] = ACTIONS(2027),
    [aux_sym_mult_operator_token1] = ACTIONS(2027),
    [aux_sym_add_operator_token1] = ACTIONS(2027),
    [sym_concat_operator] = ACTIONS(2029),
    [sym_rel_operator] = ACTIONS(2027),
    [anon_sym_AMP_AMP] = ACTIONS(2027),
    [anon_sym_or] = ACTIONS(2027),
    [anon_sym_PIPE_PIPE] = ACTIONS(2027),
    [sym_assign_operator] = ACTIONS(2029),
    [sym_let_operator] = ACTIONS(2029),
    [sym__uppercase_identifier] = ACTIONS(2029),
    [aux_sym_directive_token1] = ACTIONS(2027),
    [aux_sym_tag_token1] = ACTIONS(2029),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1410)] = {
    [sym_attribute] = STATE(1410),
    [sym__lowercase_identifier] = ACTIONS(2039),
    [anon_sym_SEMI_SEMI] = ACTIONS(2041),
    [anon_sym_let] = ACTIONS(2039),
    [anon_sym_and] = ACTIONS(2039),
    [anon_sym_TILDE] = ACTIONS(2039),
    [anon_sym_QMARK] = ACTIONS(2039),
    [anon_sym_LPAREN] = ACTIONS(2041),
    [anon_sym_external] = ACTIONS(2039),
    [anon_sym_type] = ACTIONS(2039),
    [anon_sym_COMMA] = ACTIONS(2041),
    [anon_sym_PIPE] = ACTIONS(2039),
    [anon_sym_DOT] = ACTIONS(1087),
    [anon_sym_LBRACK] = ACTIONS(2039),
    [anon_sym_RBRACK] = ACTIONS(2041),
    [anon_sym_true] = ACTIONS(2039),
    [anon_sym_false] = ACTIONS(2039),
    [anon_sym_LBRACE] = ACTIONS(2039),
    [anon_sym_SEMI] = ACTIONS(2039),
    [anon_sym_exception] = ACTIONS(2039),
    [anon_sym_module] = ACTIONS(2039),
    [anon_sym_open] = ACTIONS(2039),
    [anon_sym_include] = ACTIONS(2039),
    [anon_sym_class] = ACTIONS(2039),
    [anon_sym_val] = ACTIONS(2039),
    [anon_sym_end] = ACTIONS(2039),
    [anon_sym_object] = ACTIONS(2039),
    [anon_sym_AMP] = ACTIONS(2039),
    [anon_sym_POUND] = ACTIONS(1089),
    [anon_sym_COLON_COLON] = ACTIONS(2041),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2041),
    [anon_sym_else] = ACTIONS(2039),
    [anon_sym_new] = ACTIONS(2039),
    [anon_sym_LBRACE_LT] = ACTIONS(2041),
    [anon_sym_begin] = ACTIONS(2039),
    [sym_ocamlyacc_value] = ACTIONS(2041),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2039),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2041),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2039),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2039),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2041),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2041),
    [aux_sym_number_token1] = ACTIONS(2041),
    [aux_sym_signed_number_token1] = ACTIONS(2039),
    [anon_sym_SQUOTE] = ACTIONS(2041),
    [anon_sym_DQUOTE] = ACTIONS(2041),
    [sym_prefix_operator] = ACTIONS(2041),
    [aux_sym_sign_operator_token1] = ACTIONS(2039),
    [sym_hash_operator] = ACTIONS(1091),
    [anon_sym_lsl] = ACTIONS(2039),
    [anon_sym_lsr] = ACTIONS(2039),
    [anon_sym_asr] = ACTIONS(2039),
    [aux_sym_pow_operator_token1] = ACTIONS(2041),
    [anon_sym_mod] = ACTIONS(2039),
    [anon_sym_land] = ACTIONS(2039),
    [anon_sym_lor] = ACTIONS(2039),
    [anon_sym_lxor] = ACTIONS(2039),
    [aux_sym_mult_operator_token1] = ACTIONS(2039),
    [aux_sym_add_operator_token1] = ACTIONS(2039),
    [sym_concat_operator] = ACTIONS(2041),
    [sym_rel_operator] = ACTIONS(2039),
    [anon_sym_AMP_AMP] = ACTIONS(2039),
    [anon_sym_or] = ACTIONS(2039),
    [anon_sym_PIPE_PIPE] = ACTIONS(2039),
    [sym_assign_operator] = ACTIONS(2041),
    [sym_let_operator] = ACTIONS(2041),
    [sym__uppercase_identifier] = ACTIONS(2041),
    [aux_sym_directive_token1] = ACTIONS(2039),
    [aux_sym_tag_token1] = ACTIONS(2041),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1411)] = {
    [sym_attribute] = STATE(1411),
    [sym__lowercase_identifier] = ACTIONS(2111),
    [anon_sym_SEMI_SEMI] = ACTIONS(2113),
    [anon_sym_let] = ACTIONS(2111),
    [anon_sym_and] = ACTIONS(2111),
    [anon_sym_TILDE] = ACTIONS(2111),
    [anon_sym_QMARK] = ACTIONS(2111),
    [anon_sym_LPAREN] = ACTIONS(2113),
    [anon_sym_external] = ACTIONS(2111),
    [anon_sym_type] = ACTIONS(2111),
    [anon_sym_COMMA] = ACTIONS(2113),
    [anon_sym_PIPE] = ACTIONS(2111),
    [anon_sym_DOT] = ACTIONS(2177),
    [anon_sym_LBRACK] = ACTIONS(2111),
    [anon_sym_RBRACK] = ACTIONS(2113),
    [anon_sym_true] = ACTIONS(2111),
    [anon_sym_false] = ACTIONS(2111),
    [anon_sym_LBRACE] = ACTIONS(2111),
    [anon_sym_SEMI] = ACTIONS(2111),
    [anon_sym_exception] = ACTIONS(2111),
    [anon_sym_module] = ACTIONS(2111),
    [anon_sym_open] = ACTIONS(2111),
    [anon_sym_include] = ACTIONS(2111),
    [anon_sym_class] = ACTIONS(2111),
    [anon_sym_val] = ACTIONS(2111),
    [anon_sym_end] = ACTIONS(2111),
    [anon_sym_object] = ACTIONS(2111),
    [anon_sym_AMP] = ACTIONS(2111),
    [anon_sym_POUND] = ACTIONS(2111),
    [anon_sym_COLON_COLON] = ACTIONS(2113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2113),
    [anon_sym_else] = ACTIONS(2111),
    [anon_sym_new] = ACTIONS(2111),
    [anon_sym_LBRACE_LT] = ACTIONS(2113),
    [anon_sym_begin] = ACTIONS(2111),
    [sym_ocamlyacc_value] = ACTIONS(2113),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2113),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2113),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2113),
    [aux_sym_number_token1] = ACTIONS(2113),
    [aux_sym_signed_number_token1] = ACTIONS(2111),
    [anon_sym_SQUOTE] = ACTIONS(2113),
    [anon_sym_DQUOTE] = ACTIONS(2113),
    [sym_prefix_operator] = ACTIONS(2113),
    [aux_sym_sign_operator_token1] = ACTIONS(2111),
    [sym_hash_operator] = ACTIONS(2113),
    [anon_sym_lsl] = ACTIONS(2111),
    [anon_sym_lsr] = ACTIONS(2111),
    [anon_sym_asr] = ACTIONS(2111),
    [aux_sym_pow_operator_token1] = ACTIONS(2113),
    [anon_sym_mod] = ACTIONS(2111),
    [anon_sym_land] = ACTIONS(2111),
    [anon_sym_lor] = ACTIONS(2111),
    [anon_sym_lxor] = ACTIONS(2111),
    [aux_sym_mult_operator_token1] = ACTIONS(2111),
    [aux_sym_add_operator_token1] = ACTIONS(2111),
    [sym_concat_operator] = ACTIONS(2113),
    [sym_rel_operator] = ACTIONS(2111),
    [anon_sym_AMP_AMP] = ACTIONS(2111),
    [anon_sym_or] = ACTIONS(2111),
    [anon_sym_PIPE_PIPE] = ACTIONS(2111),
    [sym_assign_operator] = ACTIONS(2113),
    [sym_let_operator] = ACTIONS(2113),
    [sym__uppercase_identifier] = ACTIONS(2113),
    [aux_sym_directive_token1] = ACTIONS(2111),
    [aux_sym_tag_token1] = ACTIONS(2113),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1412)] = {
    [sym_attribute] = STATE(1412),
    [ts_builtin_sym_end] = ACTIONS(2417),
    [sym__lowercase_identifier] = ACTIONS(2415),
    [anon_sym_SEMI_SEMI] = ACTIONS(2417),
    [anon_sym_let] = ACTIONS(2415),
    [anon_sym_and] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(2415),
    [anon_sym_QMARK] = ACTIONS(2415),
    [anon_sym_LPAREN] = ACTIONS(2417),
    [anon_sym_external] = ACTIONS(2415),
    [anon_sym_type] = ACTIONS(2415),
    [anon_sym_COMMA] = ACTIONS(2417),
    [anon_sym_PIPE] = ACTIONS(2415),
    [anon_sym_LBRACK] = ACTIONS(2415),
    [anon_sym_RBRACK] = ACTIONS(2417),
    [anon_sym_true] = ACTIONS(2415),
    [anon_sym_false] = ACTIONS(2415),
    [anon_sym_LBRACE] = ACTIONS(2415),
    [anon_sym_SEMI] = ACTIONS(2415),
    [anon_sym_exception] = ACTIONS(2415),
    [anon_sym_module] = ACTIONS(2415),
    [anon_sym_open] = ACTIONS(2415),
    [anon_sym_include] = ACTIONS(2415),
    [anon_sym_class] = ACTIONS(2415),
    [anon_sym_val] = ACTIONS(2415),
    [anon_sym_end] = ACTIONS(2415),
    [anon_sym_object] = ACTIONS(2415),
    [anon_sym_in] = ACTIONS(2415),
    [anon_sym_AMP] = ACTIONS(2415),
    [anon_sym_COLON_COLON] = ACTIONS(2417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2417),
    [anon_sym_else] = ACTIONS(2415),
    [anon_sym_new] = ACTIONS(2415),
    [anon_sym_LBRACE_LT] = ACTIONS(2417),
    [anon_sym_begin] = ACTIONS(2415),
    [sym_ocamlyacc_value] = ACTIONS(2417),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2415),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2417),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2415),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2417),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2417),
    [aux_sym_number_token1] = ACTIONS(2417),
    [aux_sym_signed_number_token1] = ACTIONS(2415),
    [anon_sym_SQUOTE] = ACTIONS(2417),
    [anon_sym_DQUOTE] = ACTIONS(2417),
    [sym_prefix_operator] = ACTIONS(2417),
    [aux_sym_sign_operator_token1] = ACTIONS(2415),
    [anon_sym_lsl] = ACTIONS(2415),
    [anon_sym_lsr] = ACTIONS(2415),
    [anon_sym_asr] = ACTIONS(2415),
    [aux_sym_pow_operator_token1] = ACTIONS(2417),
    [anon_sym_mod] = ACTIONS(2415),
    [anon_sym_land] = ACTIONS(2415),
    [anon_sym_lor] = ACTIONS(2415),
    [anon_sym_lxor] = ACTIONS(2415),
    [aux_sym_mult_operator_token1] = ACTIONS(2415),
    [aux_sym_add_operator_token1] = ACTIONS(2415),
    [sym_concat_operator] = ACTIONS(2417),
    [sym_rel_operator] = ACTIONS(2415),
    [anon_sym_AMP_AMP] = ACTIONS(2415),
    [anon_sym_or] = ACTIONS(2415),
    [anon_sym_PIPE_PIPE] = ACTIONS(2415),
    [sym_assign_operator] = ACTIONS(2417),
    [sym_let_operator] = ACTIONS(2417),
    [sym_let_and_operator] = ACTIONS(2417),
    [sym__uppercase_identifier] = ACTIONS(2417),
    [aux_sym_directive_token1] = ACTIONS(2417),
    [aux_sym_tag_token1] = ACTIONS(2417),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1413)] = {
    [sym_attribute] = STATE(1413),
    [sym__lowercase_identifier] = ACTIONS(1983),
    [anon_sym_SEMI_SEMI] = ACTIONS(1985),
    [anon_sym_let] = ACTIONS(1983),
    [anon_sym_and] = ACTIONS(1983),
    [anon_sym_TILDE] = ACTIONS(1983),
    [anon_sym_QMARK] = ACTIONS(1983),
    [anon_sym_LPAREN] = ACTIONS(1985),
    [anon_sym_external] = ACTIONS(1983),
    [anon_sym_type] = ACTIONS(1983),
    [anon_sym_COMMA] = ACTIONS(1985),
    [anon_sym_PIPE] = ACTIONS(1983),
    [anon_sym_DOT] = ACTIONS(1985),
    [anon_sym_LBRACK] = ACTIONS(1983),
    [anon_sym_RBRACK] = ACTIONS(1985),
    [anon_sym_true] = ACTIONS(1983),
    [anon_sym_false] = ACTIONS(1983),
    [anon_sym_LBRACE] = ACTIONS(1983),
    [anon_sym_SEMI] = ACTIONS(1983),
    [anon_sym_exception] = ACTIONS(1983),
    [anon_sym_module] = ACTIONS(1983),
    [anon_sym_open] = ACTIONS(1983),
    [anon_sym_include] = ACTIONS(1983),
    [anon_sym_class] = ACTIONS(1983),
    [anon_sym_val] = ACTIONS(1983),
    [anon_sym_end] = ACTIONS(1983),
    [anon_sym_object] = ACTIONS(1983),
    [anon_sym_AMP] = ACTIONS(1983),
    [anon_sym_POUND] = ACTIONS(1983),
    [anon_sym_COLON_COLON] = ACTIONS(1985),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1985),
    [anon_sym_else] = ACTIONS(1983),
    [anon_sym_new] = ACTIONS(1983),
    [anon_sym_LBRACE_LT] = ACTIONS(1985),
    [anon_sym_begin] = ACTIONS(1983),
    [sym_ocamlyacc_value] = ACTIONS(1985),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1983),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1985),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1983),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1983),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1985),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1985),
    [aux_sym_number_token1] = ACTIONS(1985),
    [aux_sym_signed_number_token1] = ACTIONS(1983),
    [anon_sym_SQUOTE] = ACTIONS(1985),
    [anon_sym_DQUOTE] = ACTIONS(1985),
    [sym_prefix_operator] = ACTIONS(1985),
    [aux_sym_sign_operator_token1] = ACTIONS(1983),
    [sym_hash_operator] = ACTIONS(1985),
    [anon_sym_lsl] = ACTIONS(1983),
    [anon_sym_lsr] = ACTIONS(1983),
    [anon_sym_asr] = ACTIONS(1983),
    [aux_sym_pow_operator_token1] = ACTIONS(1985),
    [anon_sym_mod] = ACTIONS(1983),
    [anon_sym_land] = ACTIONS(1983),
    [anon_sym_lor] = ACTIONS(1983),
    [anon_sym_lxor] = ACTIONS(1983),
    [aux_sym_mult_operator_token1] = ACTIONS(1983),
    [aux_sym_add_operator_token1] = ACTIONS(1983),
    [sym_concat_operator] = ACTIONS(1985),
    [sym_rel_operator] = ACTIONS(1983),
    [anon_sym_AMP_AMP] = ACTIONS(1983),
    [anon_sym_or] = ACTIONS(1983),
    [anon_sym_PIPE_PIPE] = ACTIONS(1983),
    [sym_assign_operator] = ACTIONS(1985),
    [sym_let_operator] = ACTIONS(1985),
    [sym__uppercase_identifier] = ACTIONS(1985),
    [aux_sym_directive_token1] = ACTIONS(1983),
    [aux_sym_tag_token1] = ACTIONS(1985),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1414)] = {
    [sym_attribute] = STATE(1414),
    [ts_builtin_sym_end] = ACTIONS(1969),
    [sym__lowercase_identifier] = ACTIONS(1967),
    [anon_sym_SEMI_SEMI] = ACTIONS(1969),
    [anon_sym_let] = ACTIONS(1967),
    [anon_sym_and] = ACTIONS(1967),
    [anon_sym_TILDE] = ACTIONS(1967),
    [anon_sym_QMARK] = ACTIONS(1967),
    [anon_sym_LPAREN] = ACTIONS(1969),
    [anon_sym_external] = ACTIONS(1967),
    [anon_sym_type] = ACTIONS(1967),
    [anon_sym_COMMA] = ACTIONS(1969),
    [anon_sym_PIPE] = ACTIONS(1967),
    [anon_sym_DOT] = ACTIONS(1969),
    [anon_sym_LBRACK] = ACTIONS(1967),
    [anon_sym_true] = ACTIONS(1967),
    [anon_sym_false] = ACTIONS(1967),
    [anon_sym_LBRACE] = ACTIONS(1967),
    [anon_sym_SEMI] = ACTIONS(1967),
    [anon_sym_exception] = ACTIONS(1967),
    [anon_sym_module] = ACTIONS(1967),
    [anon_sym_open] = ACTIONS(1967),
    [anon_sym_include] = ACTIONS(1967),
    [anon_sym_class] = ACTIONS(1967),
    [anon_sym_val] = ACTIONS(1967),
    [anon_sym_object] = ACTIONS(1967),
    [anon_sym_AMP] = ACTIONS(1967),
    [anon_sym_POUND] = ACTIONS(1967),
    [anon_sym_COLON_COLON] = ACTIONS(1969),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1969),
    [anon_sym_else] = ACTIONS(1967),
    [anon_sym_new] = ACTIONS(1967),
    [anon_sym_LBRACE_LT] = ACTIONS(1969),
    [anon_sym_begin] = ACTIONS(1967),
    [sym_ocamlyacc_value] = ACTIONS(1969),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1967),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1969),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1969),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1969),
    [aux_sym_number_token1] = ACTIONS(1969),
    [aux_sym_signed_number_token1] = ACTIONS(1967),
    [anon_sym_SQUOTE] = ACTIONS(1969),
    [anon_sym_DQUOTE] = ACTIONS(1969),
    [sym_prefix_operator] = ACTIONS(1969),
    [aux_sym_sign_operator_token1] = ACTIONS(1967),
    [sym_hash_operator] = ACTIONS(1969),
    [anon_sym_lsl] = ACTIONS(1967),
    [anon_sym_lsr] = ACTIONS(1967),
    [anon_sym_asr] = ACTIONS(1967),
    [aux_sym_pow_operator_token1] = ACTIONS(1969),
    [anon_sym_mod] = ACTIONS(1967),
    [anon_sym_land] = ACTIONS(1967),
    [anon_sym_lor] = ACTIONS(1967),
    [anon_sym_lxor] = ACTIONS(1967),
    [aux_sym_mult_operator_token1] = ACTIONS(1967),
    [aux_sym_add_operator_token1] = ACTIONS(1967),
    [sym_concat_operator] = ACTIONS(1969),
    [sym_rel_operator] = ACTIONS(1967),
    [anon_sym_AMP_AMP] = ACTIONS(1967),
    [anon_sym_or] = ACTIONS(1967),
    [anon_sym_PIPE_PIPE] = ACTIONS(1967),
    [sym_assign_operator] = ACTIONS(1969),
    [sym_let_operator] = ACTIONS(1969),
    [sym__uppercase_identifier] = ACTIONS(1969),
    [aux_sym_directive_token1] = ACTIONS(1967),
    [aux_sym_tag_token1] = ACTIONS(1969),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1415)] = {
    [sym_attribute] = STATE(1415),
    [ts_builtin_sym_end] = ACTIONS(2041),
    [sym__lowercase_identifier] = ACTIONS(2039),
    [anon_sym_SEMI_SEMI] = ACTIONS(2041),
    [anon_sym_let] = ACTIONS(2039),
    [anon_sym_and] = ACTIONS(2039),
    [anon_sym_TILDE] = ACTIONS(2039),
    [anon_sym_QMARK] = ACTIONS(2039),
    [anon_sym_LPAREN] = ACTIONS(2041),
    [anon_sym_external] = ACTIONS(2039),
    [anon_sym_type] = ACTIONS(2039),
    [anon_sym_COMMA] = ACTIONS(2041),
    [anon_sym_PIPE] = ACTIONS(2039),
    [anon_sym_DOT] = ACTIONS(1157),
    [anon_sym_LBRACK] = ACTIONS(2039),
    [anon_sym_true] = ACTIONS(2039),
    [anon_sym_false] = ACTIONS(2039),
    [anon_sym_LBRACE] = ACTIONS(2039),
    [anon_sym_SEMI] = ACTIONS(2039),
    [anon_sym_exception] = ACTIONS(2039),
    [anon_sym_module] = ACTIONS(2039),
    [anon_sym_open] = ACTIONS(2039),
    [anon_sym_include] = ACTIONS(2039),
    [anon_sym_class] = ACTIONS(2039),
    [anon_sym_val] = ACTIONS(2039),
    [anon_sym_object] = ACTIONS(2039),
    [anon_sym_AMP] = ACTIONS(2039),
    [anon_sym_POUND] = ACTIONS(1159),
    [anon_sym_COLON_COLON] = ACTIONS(2041),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2041),
    [anon_sym_else] = ACTIONS(2039),
    [anon_sym_new] = ACTIONS(2039),
    [anon_sym_LBRACE_LT] = ACTIONS(2041),
    [anon_sym_begin] = ACTIONS(2039),
    [sym_ocamlyacc_value] = ACTIONS(2041),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2039),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2041),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2039),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2039),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2041),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2041),
    [aux_sym_number_token1] = ACTIONS(2041),
    [aux_sym_signed_number_token1] = ACTIONS(2039),
    [anon_sym_SQUOTE] = ACTIONS(2041),
    [anon_sym_DQUOTE] = ACTIONS(2041),
    [sym_prefix_operator] = ACTIONS(2041),
    [aux_sym_sign_operator_token1] = ACTIONS(2039),
    [sym_hash_operator] = ACTIONS(1161),
    [anon_sym_lsl] = ACTIONS(2039),
    [anon_sym_lsr] = ACTIONS(2039),
    [anon_sym_asr] = ACTIONS(2039),
    [aux_sym_pow_operator_token1] = ACTIONS(2041),
    [anon_sym_mod] = ACTIONS(2039),
    [anon_sym_land] = ACTIONS(2039),
    [anon_sym_lor] = ACTIONS(2039),
    [anon_sym_lxor] = ACTIONS(2039),
    [aux_sym_mult_operator_token1] = ACTIONS(2039),
    [aux_sym_add_operator_token1] = ACTIONS(2039),
    [sym_concat_operator] = ACTIONS(2041),
    [sym_rel_operator] = ACTIONS(2039),
    [anon_sym_AMP_AMP] = ACTIONS(2039),
    [anon_sym_or] = ACTIONS(2039),
    [anon_sym_PIPE_PIPE] = ACTIONS(2039),
    [sym_assign_operator] = ACTIONS(2041),
    [sym_let_operator] = ACTIONS(2041),
    [sym__uppercase_identifier] = ACTIONS(2041),
    [aux_sym_directive_token1] = ACTIONS(2039),
    [aux_sym_tag_token1] = ACTIONS(2041),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1416)] = {
    [sym_attribute] = STATE(1416),
    [ts_builtin_sym_end] = ACTIONS(1949),
    [sym__lowercase_identifier] = ACTIONS(1947),
    [anon_sym_SEMI_SEMI] = ACTIONS(1949),
    [anon_sym_let] = ACTIONS(1947),
    [anon_sym_TILDE] = ACTIONS(1947),
    [anon_sym_QMARK] = ACTIONS(1947),
    [anon_sym_LPAREN] = ACTIONS(1949),
    [anon_sym_external] = ACTIONS(1947),
    [anon_sym_type] = ACTIONS(1947),
    [anon_sym_COMMA] = ACTIONS(1949),
    [anon_sym_PIPE] = ACTIONS(1947),
    [anon_sym_DOT] = ACTIONS(1949),
    [anon_sym_LBRACK] = ACTIONS(1947),
    [anon_sym_true] = ACTIONS(1947),
    [anon_sym_false] = ACTIONS(1947),
    [anon_sym_LBRACE] = ACTIONS(1947),
    [anon_sym_SEMI] = ACTIONS(1947),
    [anon_sym_exception] = ACTIONS(1947),
    [anon_sym_module] = ACTIONS(1947),
    [anon_sym_open] = ACTIONS(1947),
    [anon_sym_include] = ACTIONS(1947),
    [anon_sym_class] = ACTIONS(1947),
    [anon_sym_val] = ACTIONS(1947),
    [anon_sym_object] = ACTIONS(1947),
    [anon_sym_AMP] = ACTIONS(1947),
    [anon_sym_POUND] = ACTIONS(1947),
    [anon_sym_COLON_COLON] = ACTIONS(1949),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1949),
    [anon_sym_LT_DASH] = ACTIONS(2634),
    [anon_sym_else] = ACTIONS(1947),
    [anon_sym_new] = ACTIONS(1947),
    [anon_sym_LBRACE_LT] = ACTIONS(1949),
    [anon_sym_begin] = ACTIONS(1947),
    [sym_ocamlyacc_value] = ACTIONS(1949),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1947),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1949),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1949),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1949),
    [aux_sym_number_token1] = ACTIONS(1949),
    [aux_sym_signed_number_token1] = ACTIONS(1947),
    [anon_sym_SQUOTE] = ACTIONS(1949),
    [anon_sym_DQUOTE] = ACTIONS(1949),
    [sym_prefix_operator] = ACTIONS(1949),
    [aux_sym_sign_operator_token1] = ACTIONS(1947),
    [sym_hash_operator] = ACTIONS(1949),
    [anon_sym_lsl] = ACTIONS(1947),
    [anon_sym_lsr] = ACTIONS(1947),
    [anon_sym_asr] = ACTIONS(1947),
    [aux_sym_pow_operator_token1] = ACTIONS(1949),
    [anon_sym_mod] = ACTIONS(1947),
    [anon_sym_land] = ACTIONS(1947),
    [anon_sym_lor] = ACTIONS(1947),
    [anon_sym_lxor] = ACTIONS(1947),
    [aux_sym_mult_operator_token1] = ACTIONS(1947),
    [aux_sym_add_operator_token1] = ACTIONS(1947),
    [sym_concat_operator] = ACTIONS(1949),
    [sym_rel_operator] = ACTIONS(1947),
    [anon_sym_AMP_AMP] = ACTIONS(1947),
    [anon_sym_or] = ACTIONS(1947),
    [anon_sym_PIPE_PIPE] = ACTIONS(1947),
    [sym_assign_operator] = ACTIONS(1949),
    [sym_let_operator] = ACTIONS(1949),
    [sym__uppercase_identifier] = ACTIONS(1949),
    [aux_sym_directive_token1] = ACTIONS(1947),
    [aux_sym_tag_token1] = ACTIONS(1949),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1417)] = {
    [sym_attribute] = STATE(1417),
    [ts_builtin_sym_end] = ACTIONS(2159),
    [sym__lowercase_identifier] = ACTIONS(2157),
    [anon_sym_SEMI_SEMI] = ACTIONS(2159),
    [anon_sym_let] = ACTIONS(2157),
    [anon_sym_TILDE] = ACTIONS(2157),
    [anon_sym_QMARK] = ACTIONS(2157),
    [anon_sym_LPAREN] = ACTIONS(2159),
    [anon_sym_external] = ACTIONS(2157),
    [anon_sym_type] = ACTIONS(2157),
    [anon_sym_COMMA] = ACTIONS(2159),
    [anon_sym_PIPE] = ACTIONS(2157),
    [anon_sym_DOT] = ACTIONS(2159),
    [anon_sym_LBRACK] = ACTIONS(2157),
    [anon_sym_true] = ACTIONS(2157),
    [anon_sym_false] = ACTIONS(2157),
    [anon_sym_LBRACE] = ACTIONS(2157),
    [anon_sym_SEMI] = ACTIONS(2157),
    [anon_sym_exception] = ACTIONS(2157),
    [anon_sym_module] = ACTIONS(2157),
    [anon_sym_open] = ACTIONS(2157),
    [anon_sym_include] = ACTIONS(2157),
    [anon_sym_class] = ACTIONS(2157),
    [anon_sym_val] = ACTIONS(2157),
    [anon_sym_object] = ACTIONS(2157),
    [anon_sym_AMP] = ACTIONS(2157),
    [anon_sym_POUND] = ACTIONS(2157),
    [anon_sym_COLON_COLON] = ACTIONS(2159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2159),
    [anon_sym_LT_DASH] = ACTIONS(2317),
    [anon_sym_else] = ACTIONS(2157),
    [anon_sym_new] = ACTIONS(2157),
    [anon_sym_LBRACE_LT] = ACTIONS(2159),
    [anon_sym_begin] = ACTIONS(2157),
    [sym_ocamlyacc_value] = ACTIONS(2159),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2157),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2159),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2159),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2159),
    [aux_sym_number_token1] = ACTIONS(2159),
    [aux_sym_signed_number_token1] = ACTIONS(2157),
    [anon_sym_SQUOTE] = ACTIONS(2159),
    [anon_sym_DQUOTE] = ACTIONS(2159),
    [sym_prefix_operator] = ACTIONS(2159),
    [aux_sym_sign_operator_token1] = ACTIONS(2157),
    [sym_hash_operator] = ACTIONS(2159),
    [anon_sym_lsl] = ACTIONS(2157),
    [anon_sym_lsr] = ACTIONS(2157),
    [anon_sym_asr] = ACTIONS(2157),
    [aux_sym_pow_operator_token1] = ACTIONS(2159),
    [anon_sym_mod] = ACTIONS(2157),
    [anon_sym_land] = ACTIONS(2157),
    [anon_sym_lor] = ACTIONS(2157),
    [anon_sym_lxor] = ACTIONS(2157),
    [aux_sym_mult_operator_token1] = ACTIONS(2157),
    [aux_sym_add_operator_token1] = ACTIONS(2157),
    [sym_concat_operator] = ACTIONS(2159),
    [sym_rel_operator] = ACTIONS(2157),
    [anon_sym_AMP_AMP] = ACTIONS(2157),
    [anon_sym_or] = ACTIONS(2157),
    [anon_sym_PIPE_PIPE] = ACTIONS(2157),
    [sym_assign_operator] = ACTIONS(2159),
    [sym_let_operator] = ACTIONS(2159),
    [sym__uppercase_identifier] = ACTIONS(2159),
    [aux_sym_directive_token1] = ACTIONS(2157),
    [aux_sym_tag_token1] = ACTIONS(2159),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1418)] = {
    [sym_attribute] = STATE(1418),
    [ts_builtin_sym_end] = ACTIONS(1997),
    [sym__lowercase_identifier] = ACTIONS(1995),
    [anon_sym_SEMI_SEMI] = ACTIONS(1997),
    [anon_sym_let] = ACTIONS(1995),
    [anon_sym_and] = ACTIONS(1995),
    [anon_sym_TILDE] = ACTIONS(1995),
    [anon_sym_QMARK] = ACTIONS(1995),
    [anon_sym_LPAREN] = ACTIONS(1997),
    [anon_sym_external] = ACTIONS(1995),
    [anon_sym_type] = ACTIONS(1995),
    [anon_sym_COMMA] = ACTIONS(1997),
    [anon_sym_PIPE] = ACTIONS(1995),
    [anon_sym_DOT] = ACTIONS(1997),
    [anon_sym_LBRACK] = ACTIONS(1995),
    [anon_sym_true] = ACTIONS(1995),
    [anon_sym_false] = ACTIONS(1995),
    [anon_sym_LBRACE] = ACTIONS(1995),
    [anon_sym_SEMI] = ACTIONS(1995),
    [anon_sym_exception] = ACTIONS(1995),
    [anon_sym_module] = ACTIONS(1995),
    [anon_sym_open] = ACTIONS(1995),
    [anon_sym_include] = ACTIONS(1995),
    [anon_sym_class] = ACTIONS(1995),
    [anon_sym_val] = ACTIONS(1995),
    [anon_sym_object] = ACTIONS(1995),
    [anon_sym_AMP] = ACTIONS(1995),
    [anon_sym_POUND] = ACTIONS(1995),
    [anon_sym_COLON_COLON] = ACTIONS(1997),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1997),
    [anon_sym_else] = ACTIONS(1995),
    [anon_sym_new] = ACTIONS(1995),
    [anon_sym_LBRACE_LT] = ACTIONS(1997),
    [anon_sym_begin] = ACTIONS(1995),
    [sym_ocamlyacc_value] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1995),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1997),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1995),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1995),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1997),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1997),
    [aux_sym_number_token1] = ACTIONS(1997),
    [aux_sym_signed_number_token1] = ACTIONS(1995),
    [anon_sym_SQUOTE] = ACTIONS(1997),
    [anon_sym_DQUOTE] = ACTIONS(1997),
    [sym_prefix_operator] = ACTIONS(1997),
    [aux_sym_sign_operator_token1] = ACTIONS(1995),
    [sym_hash_operator] = ACTIONS(1997),
    [anon_sym_lsl] = ACTIONS(1995),
    [anon_sym_lsr] = ACTIONS(1995),
    [anon_sym_asr] = ACTIONS(1995),
    [aux_sym_pow_operator_token1] = ACTIONS(1997),
    [anon_sym_mod] = ACTIONS(1995),
    [anon_sym_land] = ACTIONS(1995),
    [anon_sym_lor] = ACTIONS(1995),
    [anon_sym_lxor] = ACTIONS(1995),
    [aux_sym_mult_operator_token1] = ACTIONS(1995),
    [aux_sym_add_operator_token1] = ACTIONS(1995),
    [sym_concat_operator] = ACTIONS(1997),
    [sym_rel_operator] = ACTIONS(1995),
    [anon_sym_AMP_AMP] = ACTIONS(1995),
    [anon_sym_or] = ACTIONS(1995),
    [anon_sym_PIPE_PIPE] = ACTIONS(1995),
    [sym_assign_operator] = ACTIONS(1997),
    [sym_let_operator] = ACTIONS(1997),
    [sym__uppercase_identifier] = ACTIONS(1997),
    [aux_sym_directive_token1] = ACTIONS(1995),
    [aux_sym_tag_token1] = ACTIONS(1997),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1419)] = {
    [sym_attribute] = STATE(1419),
    [ts_builtin_sym_end] = ACTIONS(2167),
    [sym__lowercase_identifier] = ACTIONS(2165),
    [anon_sym_SEMI_SEMI] = ACTIONS(2167),
    [anon_sym_let] = ACTIONS(2165),
    [anon_sym_and] = ACTIONS(2165),
    [anon_sym_TILDE] = ACTIONS(2165),
    [anon_sym_QMARK] = ACTIONS(2165),
    [anon_sym_LPAREN] = ACTIONS(2167),
    [anon_sym_external] = ACTIONS(2165),
    [anon_sym_type] = ACTIONS(2165),
    [anon_sym_COMMA] = ACTIONS(2167),
    [anon_sym_PIPE] = ACTIONS(2165),
    [anon_sym_DOT] = ACTIONS(2167),
    [anon_sym_LBRACK] = ACTIONS(2165),
    [anon_sym_true] = ACTIONS(2165),
    [anon_sym_false] = ACTIONS(2165),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_SEMI] = ACTIONS(2165),
    [anon_sym_exception] = ACTIONS(2165),
    [anon_sym_module] = ACTIONS(2165),
    [anon_sym_open] = ACTIONS(2165),
    [anon_sym_include] = ACTIONS(2165),
    [anon_sym_class] = ACTIONS(2165),
    [anon_sym_val] = ACTIONS(2165),
    [anon_sym_object] = ACTIONS(2165),
    [anon_sym_AMP] = ACTIONS(2165),
    [anon_sym_POUND] = ACTIONS(2165),
    [anon_sym_COLON_COLON] = ACTIONS(2167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2167),
    [anon_sym_else] = ACTIONS(2165),
    [anon_sym_new] = ACTIONS(2165),
    [anon_sym_LBRACE_LT] = ACTIONS(2167),
    [anon_sym_begin] = ACTIONS(2165),
    [sym_ocamlyacc_value] = ACTIONS(2167),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2165),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2167),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2167),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2167),
    [aux_sym_number_token1] = ACTIONS(2167),
    [aux_sym_signed_number_token1] = ACTIONS(2165),
    [anon_sym_SQUOTE] = ACTIONS(2167),
    [anon_sym_DQUOTE] = ACTIONS(2167),
    [sym_prefix_operator] = ACTIONS(2167),
    [aux_sym_sign_operator_token1] = ACTIONS(2165),
    [sym_hash_operator] = ACTIONS(2167),
    [anon_sym_lsl] = ACTIONS(2165),
    [anon_sym_lsr] = ACTIONS(2165),
    [anon_sym_asr] = ACTIONS(2165),
    [aux_sym_pow_operator_token1] = ACTIONS(2167),
    [anon_sym_mod] = ACTIONS(2165),
    [anon_sym_land] = ACTIONS(2165),
    [anon_sym_lor] = ACTIONS(2165),
    [anon_sym_lxor] = ACTIONS(2165),
    [aux_sym_mult_operator_token1] = ACTIONS(2165),
    [aux_sym_add_operator_token1] = ACTIONS(2165),
    [sym_concat_operator] = ACTIONS(2167),
    [sym_rel_operator] = ACTIONS(2165),
    [anon_sym_AMP_AMP] = ACTIONS(2165),
    [anon_sym_or] = ACTIONS(2165),
    [anon_sym_PIPE_PIPE] = ACTIONS(2165),
    [sym_assign_operator] = ACTIONS(2167),
    [sym_let_operator] = ACTIONS(2167),
    [sym__uppercase_identifier] = ACTIONS(2167),
    [aux_sym_directive_token1] = ACTIONS(2165),
    [aux_sym_tag_token1] = ACTIONS(2167),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1420)] = {
    [sym_attribute] = STATE(1420),
    [ts_builtin_sym_end] = ACTIONS(2093),
    [sym__lowercase_identifier] = ACTIONS(2091),
    [anon_sym_SEMI_SEMI] = ACTIONS(2093),
    [anon_sym_let] = ACTIONS(2091),
    [anon_sym_and] = ACTIONS(2091),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_external] = ACTIONS(2091),
    [anon_sym_type] = ACTIONS(2091),
    [anon_sym_COMMA] = ACTIONS(2093),
    [anon_sym_PIPE] = ACTIONS(2091),
    [anon_sym_DOT] = ACTIONS(2093),
    [anon_sym_LBRACK] = ACTIONS(2091),
    [anon_sym_true] = ACTIONS(2091),
    [anon_sym_false] = ACTIONS(2091),
    [anon_sym_LBRACE] = ACTIONS(2091),
    [anon_sym_SEMI] = ACTIONS(2091),
    [anon_sym_exception] = ACTIONS(2091),
    [anon_sym_module] = ACTIONS(2091),
    [anon_sym_open] = ACTIONS(2091),
    [anon_sym_include] = ACTIONS(2091),
    [anon_sym_class] = ACTIONS(2091),
    [anon_sym_val] = ACTIONS(2091),
    [anon_sym_object] = ACTIONS(2091),
    [anon_sym_AMP] = ACTIONS(2091),
    [anon_sym_POUND] = ACTIONS(2091),
    [anon_sym_COLON_COLON] = ACTIONS(2093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2093),
    [anon_sym_else] = ACTIONS(2091),
    [anon_sym_new] = ACTIONS(2091),
    [anon_sym_LBRACE_LT] = ACTIONS(2093),
    [anon_sym_begin] = ACTIONS(2091),
    [sym_ocamlyacc_value] = ACTIONS(2093),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2091),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2093),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2091),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2091),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2093),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2093),
    [aux_sym_number_token1] = ACTIONS(2093),
    [aux_sym_signed_number_token1] = ACTIONS(2091),
    [anon_sym_SQUOTE] = ACTIONS(2093),
    [anon_sym_DQUOTE] = ACTIONS(2093),
    [sym_prefix_operator] = ACTIONS(2093),
    [aux_sym_sign_operator_token1] = ACTIONS(2091),
    [sym_hash_operator] = ACTIONS(2093),
    [anon_sym_lsl] = ACTIONS(2091),
    [anon_sym_lsr] = ACTIONS(2091),
    [anon_sym_asr] = ACTIONS(2091),
    [aux_sym_pow_operator_token1] = ACTIONS(2093),
    [anon_sym_mod] = ACTIONS(2091),
    [anon_sym_land] = ACTIONS(2091),
    [anon_sym_lor] = ACTIONS(2091),
    [anon_sym_lxor] = ACTIONS(2091),
    [aux_sym_mult_operator_token1] = ACTIONS(2091),
    [aux_sym_add_operator_token1] = ACTIONS(2091),
    [sym_concat_operator] = ACTIONS(2093),
    [sym_rel_operator] = ACTIONS(2091),
    [anon_sym_AMP_AMP] = ACTIONS(2091),
    [anon_sym_or] = ACTIONS(2091),
    [anon_sym_PIPE_PIPE] = ACTIONS(2091),
    [sym_assign_operator] = ACTIONS(2093),
    [sym_let_operator] = ACTIONS(2093),
    [sym__uppercase_identifier] = ACTIONS(2093),
    [aux_sym_directive_token1] = ACTIONS(2091),
    [aux_sym_tag_token1] = ACTIONS(2093),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1421)] = {
    [sym_attribute] = STATE(1421),
    [ts_builtin_sym_end] = ACTIONS(2097),
    [sym__lowercase_identifier] = ACTIONS(2095),
    [anon_sym_SEMI_SEMI] = ACTIONS(2097),
    [anon_sym_let] = ACTIONS(2095),
    [anon_sym_and] = ACTIONS(2095),
    [anon_sym_TILDE] = ACTIONS(2095),
    [anon_sym_QMARK] = ACTIONS(2095),
    [anon_sym_LPAREN] = ACTIONS(2097),
    [anon_sym_external] = ACTIONS(2095),
    [anon_sym_type] = ACTIONS(2095),
    [anon_sym_COMMA] = ACTIONS(2097),
    [anon_sym_PIPE] = ACTIONS(2095),
    [anon_sym_DOT] = ACTIONS(2097),
    [anon_sym_LBRACK] = ACTIONS(2095),
    [anon_sym_true] = ACTIONS(2095),
    [anon_sym_false] = ACTIONS(2095),
    [anon_sym_LBRACE] = ACTIONS(2095),
    [anon_sym_SEMI] = ACTIONS(2095),
    [anon_sym_exception] = ACTIONS(2095),
    [anon_sym_module] = ACTIONS(2095),
    [anon_sym_open] = ACTIONS(2095),
    [anon_sym_include] = ACTIONS(2095),
    [anon_sym_class] = ACTIONS(2095),
    [anon_sym_val] = ACTIONS(2095),
    [anon_sym_object] = ACTIONS(2095),
    [anon_sym_AMP] = ACTIONS(2095),
    [anon_sym_POUND] = ACTIONS(2095),
    [anon_sym_COLON_COLON] = ACTIONS(2097),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2097),
    [anon_sym_else] = ACTIONS(2095),
    [anon_sym_new] = ACTIONS(2095),
    [anon_sym_LBRACE_LT] = ACTIONS(2097),
    [anon_sym_begin] = ACTIONS(2095),
    [sym_ocamlyacc_value] = ACTIONS(2097),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2095),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2097),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2095),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2095),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2097),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2097),
    [aux_sym_number_token1] = ACTIONS(2097),
    [aux_sym_signed_number_token1] = ACTIONS(2095),
    [anon_sym_SQUOTE] = ACTIONS(2097),
    [anon_sym_DQUOTE] = ACTIONS(2097),
    [sym_prefix_operator] = ACTIONS(2097),
    [aux_sym_sign_operator_token1] = ACTIONS(2095),
    [sym_hash_operator] = ACTIONS(2097),
    [anon_sym_lsl] = ACTIONS(2095),
    [anon_sym_lsr] = ACTIONS(2095),
    [anon_sym_asr] = ACTIONS(2095),
    [aux_sym_pow_operator_token1] = ACTIONS(2097),
    [anon_sym_mod] = ACTIONS(2095),
    [anon_sym_land] = ACTIONS(2095),
    [anon_sym_lor] = ACTIONS(2095),
    [anon_sym_lxor] = ACTIONS(2095),
    [aux_sym_mult_operator_token1] = ACTIONS(2095),
    [aux_sym_add_operator_token1] = ACTIONS(2095),
    [sym_concat_operator] = ACTIONS(2097),
    [sym_rel_operator] = ACTIONS(2095),
    [anon_sym_AMP_AMP] = ACTIONS(2095),
    [anon_sym_or] = ACTIONS(2095),
    [anon_sym_PIPE_PIPE] = ACTIONS(2095),
    [sym_assign_operator] = ACTIONS(2097),
    [sym_let_operator] = ACTIONS(2097),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_directive_token1] = ACTIONS(2095),
    [aux_sym_tag_token1] = ACTIONS(2097),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1422)] = {
    [sym_attribute] = STATE(1422),
    [ts_builtin_sym_end] = ACTIONS(1965),
    [sym__lowercase_identifier] = ACTIONS(1963),
    [anon_sym_SEMI_SEMI] = ACTIONS(1965),
    [anon_sym_let] = ACTIONS(1963),
    [anon_sym_and] = ACTIONS(1963),
    [anon_sym_TILDE] = ACTIONS(1963),
    [anon_sym_QMARK] = ACTIONS(1963),
    [anon_sym_LPAREN] = ACTIONS(1965),
    [anon_sym_external] = ACTIONS(1963),
    [anon_sym_type] = ACTIONS(1963),
    [anon_sym_COMMA] = ACTIONS(1965),
    [anon_sym_PIPE] = ACTIONS(1963),
    [anon_sym_DOT] = ACTIONS(1965),
    [anon_sym_LBRACK] = ACTIONS(1963),
    [anon_sym_true] = ACTIONS(1963),
    [anon_sym_false] = ACTIONS(1963),
    [anon_sym_LBRACE] = ACTIONS(1963),
    [anon_sym_SEMI] = ACTIONS(1963),
    [anon_sym_exception] = ACTIONS(1963),
    [anon_sym_module] = ACTIONS(1963),
    [anon_sym_open] = ACTIONS(1963),
    [anon_sym_include] = ACTIONS(1963),
    [anon_sym_class] = ACTIONS(1963),
    [anon_sym_val] = ACTIONS(1963),
    [anon_sym_object] = ACTIONS(1963),
    [anon_sym_AMP] = ACTIONS(1963),
    [anon_sym_POUND] = ACTIONS(1963),
    [anon_sym_COLON_COLON] = ACTIONS(1965),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1965),
    [anon_sym_else] = ACTIONS(1963),
    [anon_sym_new] = ACTIONS(1963),
    [anon_sym_LBRACE_LT] = ACTIONS(1965),
    [anon_sym_begin] = ACTIONS(1963),
    [sym_ocamlyacc_value] = ACTIONS(1965),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1963),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1965),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1965),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1965),
    [aux_sym_signed_number_token1] = ACTIONS(1963),
    [anon_sym_SQUOTE] = ACTIONS(1965),
    [anon_sym_DQUOTE] = ACTIONS(1965),
    [sym_prefix_operator] = ACTIONS(1965),
    [aux_sym_sign_operator_token1] = ACTIONS(1963),
    [sym_hash_operator] = ACTIONS(1965),
    [anon_sym_lsl] = ACTIONS(1963),
    [anon_sym_lsr] = ACTIONS(1963),
    [anon_sym_asr] = ACTIONS(1963),
    [aux_sym_pow_operator_token1] = ACTIONS(1965),
    [anon_sym_mod] = ACTIONS(1963),
    [anon_sym_land] = ACTIONS(1963),
    [anon_sym_lor] = ACTIONS(1963),
    [anon_sym_lxor] = ACTIONS(1963),
    [aux_sym_mult_operator_token1] = ACTIONS(1963),
    [aux_sym_add_operator_token1] = ACTIONS(1963),
    [sym_concat_operator] = ACTIONS(1965),
    [sym_rel_operator] = ACTIONS(1963),
    [anon_sym_AMP_AMP] = ACTIONS(1963),
    [anon_sym_or] = ACTIONS(1963),
    [anon_sym_PIPE_PIPE] = ACTIONS(1963),
    [sym_assign_operator] = ACTIONS(1965),
    [sym_let_operator] = ACTIONS(1965),
    [sym__uppercase_identifier] = ACTIONS(1965),
    [aux_sym_directive_token1] = ACTIONS(1963),
    [aux_sym_tag_token1] = ACTIONS(1965),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1423)] = {
    [sym_attribute] = STATE(1423),
    [ts_builtin_sym_end] = ACTIONS(1929),
    [sym__lowercase_identifier] = ACTIONS(1927),
    [anon_sym_SEMI_SEMI] = ACTIONS(1929),
    [anon_sym_let] = ACTIONS(1927),
    [anon_sym_and] = ACTIONS(1927),
    [anon_sym_TILDE] = ACTIONS(1927),
    [anon_sym_QMARK] = ACTIONS(1927),
    [anon_sym_LPAREN] = ACTIONS(1929),
    [anon_sym_external] = ACTIONS(1927),
    [anon_sym_type] = ACTIONS(1927),
    [anon_sym_COMMA] = ACTIONS(1929),
    [anon_sym_PIPE] = ACTIONS(1927),
    [anon_sym_DOT] = ACTIONS(1929),
    [anon_sym_LBRACK] = ACTIONS(1927),
    [anon_sym_true] = ACTIONS(1927),
    [anon_sym_false] = ACTIONS(1927),
    [anon_sym_LBRACE] = ACTIONS(1927),
    [anon_sym_SEMI] = ACTIONS(1927),
    [anon_sym_exception] = ACTIONS(1927),
    [anon_sym_module] = ACTIONS(1927),
    [anon_sym_open] = ACTIONS(1927),
    [anon_sym_include] = ACTIONS(1927),
    [anon_sym_class] = ACTIONS(1927),
    [anon_sym_val] = ACTIONS(1927),
    [anon_sym_object] = ACTIONS(1927),
    [anon_sym_AMP] = ACTIONS(1927),
    [anon_sym_POUND] = ACTIONS(1927),
    [anon_sym_COLON_COLON] = ACTIONS(1929),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1929),
    [anon_sym_else] = ACTIONS(1927),
    [anon_sym_new] = ACTIONS(1927),
    [anon_sym_LBRACE_LT] = ACTIONS(1929),
    [anon_sym_begin] = ACTIONS(1927),
    [sym_ocamlyacc_value] = ACTIONS(1929),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1927),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1929),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1927),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1927),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1929),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1929),
    [aux_sym_number_token1] = ACTIONS(1929),
    [aux_sym_signed_number_token1] = ACTIONS(1927),
    [anon_sym_SQUOTE] = ACTIONS(1929),
    [anon_sym_DQUOTE] = ACTIONS(1929),
    [sym_prefix_operator] = ACTIONS(1929),
    [aux_sym_sign_operator_token1] = ACTIONS(1927),
    [sym_hash_operator] = ACTIONS(1929),
    [anon_sym_lsl] = ACTIONS(1927),
    [anon_sym_lsr] = ACTIONS(1927),
    [anon_sym_asr] = ACTIONS(1927),
    [aux_sym_pow_operator_token1] = ACTIONS(1929),
    [anon_sym_mod] = ACTIONS(1927),
    [anon_sym_land] = ACTIONS(1927),
    [anon_sym_lor] = ACTIONS(1927),
    [anon_sym_lxor] = ACTIONS(1927),
    [aux_sym_mult_operator_token1] = ACTIONS(1927),
    [aux_sym_add_operator_token1] = ACTIONS(1927),
    [sym_concat_operator] = ACTIONS(1929),
    [sym_rel_operator] = ACTIONS(1927),
    [anon_sym_AMP_AMP] = ACTIONS(1927),
    [anon_sym_or] = ACTIONS(1927),
    [anon_sym_PIPE_PIPE] = ACTIONS(1927),
    [sym_assign_operator] = ACTIONS(1929),
    [sym_let_operator] = ACTIONS(1929),
    [sym__uppercase_identifier] = ACTIONS(1929),
    [aux_sym_directive_token1] = ACTIONS(1927),
    [aux_sym_tag_token1] = ACTIONS(1929),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1424)] = {
    [sym_attribute] = STATE(1424),
    [ts_builtin_sym_end] = ACTIONS(2205),
    [sym__lowercase_identifier] = ACTIONS(2203),
    [anon_sym_SEMI_SEMI] = ACTIONS(2205),
    [anon_sym_let] = ACTIONS(2203),
    [anon_sym_and] = ACTIONS(2203),
    [anon_sym_TILDE] = ACTIONS(2203),
    [anon_sym_QMARK] = ACTIONS(2203),
    [anon_sym_LPAREN] = ACTIONS(2205),
    [anon_sym_external] = ACTIONS(2203),
    [anon_sym_type] = ACTIONS(2203),
    [anon_sym_COMMA] = ACTIONS(2205),
    [anon_sym_PIPE] = ACTIONS(2203),
    [anon_sym_DOT] = ACTIONS(2205),
    [anon_sym_LBRACK] = ACTIONS(2203),
    [anon_sym_true] = ACTIONS(2203),
    [anon_sym_false] = ACTIONS(2203),
    [anon_sym_LBRACE] = ACTIONS(2203),
    [anon_sym_SEMI] = ACTIONS(2203),
    [anon_sym_exception] = ACTIONS(2203),
    [anon_sym_module] = ACTIONS(2203),
    [anon_sym_open] = ACTIONS(2203),
    [anon_sym_include] = ACTIONS(2203),
    [anon_sym_class] = ACTIONS(2203),
    [anon_sym_val] = ACTIONS(2203),
    [anon_sym_object] = ACTIONS(2203),
    [anon_sym_AMP] = ACTIONS(2203),
    [anon_sym_POUND] = ACTIONS(2203),
    [anon_sym_COLON_COLON] = ACTIONS(2205),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2205),
    [anon_sym_else] = ACTIONS(2203),
    [anon_sym_new] = ACTIONS(2203),
    [anon_sym_LBRACE_LT] = ACTIONS(2205),
    [anon_sym_begin] = ACTIONS(2203),
    [sym_ocamlyacc_value] = ACTIONS(2205),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2203),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2205),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2203),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2203),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2205),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2205),
    [aux_sym_number_token1] = ACTIONS(2205),
    [aux_sym_signed_number_token1] = ACTIONS(2203),
    [anon_sym_SQUOTE] = ACTIONS(2205),
    [anon_sym_DQUOTE] = ACTIONS(2205),
    [sym_prefix_operator] = ACTIONS(2205),
    [aux_sym_sign_operator_token1] = ACTIONS(2203),
    [sym_hash_operator] = ACTIONS(2205),
    [anon_sym_lsl] = ACTIONS(2203),
    [anon_sym_lsr] = ACTIONS(2203),
    [anon_sym_asr] = ACTIONS(2203),
    [aux_sym_pow_operator_token1] = ACTIONS(2205),
    [anon_sym_mod] = ACTIONS(2203),
    [anon_sym_land] = ACTIONS(2203),
    [anon_sym_lor] = ACTIONS(2203),
    [anon_sym_lxor] = ACTIONS(2203),
    [aux_sym_mult_operator_token1] = ACTIONS(2203),
    [aux_sym_add_operator_token1] = ACTIONS(2203),
    [sym_concat_operator] = ACTIONS(2205),
    [sym_rel_operator] = ACTIONS(2203),
    [anon_sym_AMP_AMP] = ACTIONS(2203),
    [anon_sym_or] = ACTIONS(2203),
    [anon_sym_PIPE_PIPE] = ACTIONS(2203),
    [sym_assign_operator] = ACTIONS(2205),
    [sym_let_operator] = ACTIONS(2205),
    [sym__uppercase_identifier] = ACTIONS(2205),
    [aux_sym_directive_token1] = ACTIONS(2203),
    [aux_sym_tag_token1] = ACTIONS(2205),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1425)] = {
    [sym_attribute] = STATE(1425),
    [ts_builtin_sym_end] = ACTIONS(2077),
    [sym__lowercase_identifier] = ACTIONS(2075),
    [anon_sym_SEMI_SEMI] = ACTIONS(2077),
    [anon_sym_let] = ACTIONS(2075),
    [anon_sym_and] = ACTIONS(2075),
    [anon_sym_TILDE] = ACTIONS(2075),
    [anon_sym_QMARK] = ACTIONS(2075),
    [anon_sym_LPAREN] = ACTIONS(2077),
    [anon_sym_external] = ACTIONS(2075),
    [anon_sym_type] = ACTIONS(2075),
    [anon_sym_COMMA] = ACTIONS(2077),
    [anon_sym_PIPE] = ACTIONS(2075),
    [anon_sym_DOT] = ACTIONS(2077),
    [anon_sym_LBRACK] = ACTIONS(2075),
    [anon_sym_true] = ACTIONS(2075),
    [anon_sym_false] = ACTIONS(2075),
    [anon_sym_LBRACE] = ACTIONS(2075),
    [anon_sym_SEMI] = ACTIONS(2075),
    [anon_sym_exception] = ACTIONS(2075),
    [anon_sym_module] = ACTIONS(2075),
    [anon_sym_open] = ACTIONS(2075),
    [anon_sym_include] = ACTIONS(2075),
    [anon_sym_class] = ACTIONS(2075),
    [anon_sym_val] = ACTIONS(2075),
    [anon_sym_object] = ACTIONS(2075),
    [anon_sym_AMP] = ACTIONS(2075),
    [anon_sym_POUND] = ACTIONS(2075),
    [anon_sym_COLON_COLON] = ACTIONS(2077),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2077),
    [anon_sym_else] = ACTIONS(2075),
    [anon_sym_new] = ACTIONS(2075),
    [anon_sym_LBRACE_LT] = ACTIONS(2077),
    [anon_sym_begin] = ACTIONS(2075),
    [sym_ocamlyacc_value] = ACTIONS(2077),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2075),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2077),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2075),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2075),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2077),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2077),
    [aux_sym_number_token1] = ACTIONS(2077),
    [aux_sym_signed_number_token1] = ACTIONS(2075),
    [anon_sym_SQUOTE] = ACTIONS(2077),
    [anon_sym_DQUOTE] = ACTIONS(2077),
    [sym_prefix_operator] = ACTIONS(2077),
    [aux_sym_sign_operator_token1] = ACTIONS(2075),
    [sym_hash_operator] = ACTIONS(2077),
    [anon_sym_lsl] = ACTIONS(2075),
    [anon_sym_lsr] = ACTIONS(2075),
    [anon_sym_asr] = ACTIONS(2075),
    [aux_sym_pow_operator_token1] = ACTIONS(2077),
    [anon_sym_mod] = ACTIONS(2075),
    [anon_sym_land] = ACTIONS(2075),
    [anon_sym_lor] = ACTIONS(2075),
    [anon_sym_lxor] = ACTIONS(2075),
    [aux_sym_mult_operator_token1] = ACTIONS(2075),
    [aux_sym_add_operator_token1] = ACTIONS(2075),
    [sym_concat_operator] = ACTIONS(2077),
    [sym_rel_operator] = ACTIONS(2075),
    [anon_sym_AMP_AMP] = ACTIONS(2075),
    [anon_sym_or] = ACTIONS(2075),
    [anon_sym_PIPE_PIPE] = ACTIONS(2075),
    [sym_assign_operator] = ACTIONS(2077),
    [sym_let_operator] = ACTIONS(2077),
    [sym__uppercase_identifier] = ACTIONS(2077),
    [aux_sym_directive_token1] = ACTIONS(2075),
    [aux_sym_tag_token1] = ACTIONS(2077),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1426)] = {
    [sym_attribute] = STATE(1426),
    [ts_builtin_sym_end] = ACTIONS(2139),
    [sym__lowercase_identifier] = ACTIONS(2137),
    [anon_sym_SEMI_SEMI] = ACTIONS(2139),
    [anon_sym_let] = ACTIONS(2137),
    [anon_sym_and] = ACTIONS(2137),
    [anon_sym_TILDE] = ACTIONS(2137),
    [anon_sym_QMARK] = ACTIONS(2137),
    [anon_sym_LPAREN] = ACTIONS(2139),
    [anon_sym_external] = ACTIONS(2137),
    [anon_sym_type] = ACTIONS(2137),
    [anon_sym_COMMA] = ACTIONS(2139),
    [anon_sym_PIPE] = ACTIONS(2137),
    [anon_sym_DOT] = ACTIONS(2139),
    [anon_sym_LBRACK] = ACTIONS(2137),
    [anon_sym_true] = ACTIONS(2137),
    [anon_sym_false] = ACTIONS(2137),
    [anon_sym_LBRACE] = ACTIONS(2137),
    [anon_sym_SEMI] = ACTIONS(2137),
    [anon_sym_exception] = ACTIONS(2137),
    [anon_sym_module] = ACTIONS(2137),
    [anon_sym_open] = ACTIONS(2137),
    [anon_sym_include] = ACTIONS(2137),
    [anon_sym_class] = ACTIONS(2137),
    [anon_sym_val] = ACTIONS(2137),
    [anon_sym_object] = ACTIONS(2137),
    [anon_sym_AMP] = ACTIONS(2137),
    [anon_sym_POUND] = ACTIONS(2137),
    [anon_sym_COLON_COLON] = ACTIONS(2139),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2139),
    [anon_sym_else] = ACTIONS(2137),
    [anon_sym_new] = ACTIONS(2137),
    [anon_sym_LBRACE_LT] = ACTIONS(2139),
    [anon_sym_begin] = ACTIONS(2137),
    [sym_ocamlyacc_value] = ACTIONS(2139),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2137),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2139),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2137),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2137),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2139),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2139),
    [aux_sym_number_token1] = ACTIONS(2139),
    [aux_sym_signed_number_token1] = ACTIONS(2137),
    [anon_sym_SQUOTE] = ACTIONS(2139),
    [anon_sym_DQUOTE] = ACTIONS(2139),
    [sym_prefix_operator] = ACTIONS(2139),
    [aux_sym_sign_operator_token1] = ACTIONS(2137),
    [sym_hash_operator] = ACTIONS(2139),
    [anon_sym_lsl] = ACTIONS(2137),
    [anon_sym_lsr] = ACTIONS(2137),
    [anon_sym_asr] = ACTIONS(2137),
    [aux_sym_pow_operator_token1] = ACTIONS(2139),
    [anon_sym_mod] = ACTIONS(2137),
    [anon_sym_land] = ACTIONS(2137),
    [anon_sym_lor] = ACTIONS(2137),
    [anon_sym_lxor] = ACTIONS(2137),
    [aux_sym_mult_operator_token1] = ACTIONS(2137),
    [aux_sym_add_operator_token1] = ACTIONS(2137),
    [sym_concat_operator] = ACTIONS(2139),
    [sym_rel_operator] = ACTIONS(2137),
    [anon_sym_AMP_AMP] = ACTIONS(2137),
    [anon_sym_or] = ACTIONS(2137),
    [anon_sym_PIPE_PIPE] = ACTIONS(2137),
    [sym_assign_operator] = ACTIONS(2139),
    [sym_let_operator] = ACTIONS(2139),
    [sym__uppercase_identifier] = ACTIONS(2139),
    [aux_sym_directive_token1] = ACTIONS(2137),
    [aux_sym_tag_token1] = ACTIONS(2139),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1427)] = {
    [sym_attribute] = STATE(1427),
    [ts_builtin_sym_end] = ACTIONS(2101),
    [sym__lowercase_identifier] = ACTIONS(2099),
    [anon_sym_SEMI_SEMI] = ACTIONS(2101),
    [anon_sym_let] = ACTIONS(2099),
    [anon_sym_and] = ACTIONS(2099),
    [anon_sym_TILDE] = ACTIONS(2099),
    [anon_sym_QMARK] = ACTIONS(2099),
    [anon_sym_LPAREN] = ACTIONS(2101),
    [anon_sym_external] = ACTIONS(2099),
    [anon_sym_type] = ACTIONS(2099),
    [anon_sym_COMMA] = ACTIONS(2101),
    [anon_sym_PIPE] = ACTIONS(2099),
    [anon_sym_DOT] = ACTIONS(2101),
    [anon_sym_LBRACK] = ACTIONS(2099),
    [anon_sym_true] = ACTIONS(2099),
    [anon_sym_false] = ACTIONS(2099),
    [anon_sym_LBRACE] = ACTIONS(2099),
    [anon_sym_SEMI] = ACTIONS(2099),
    [anon_sym_exception] = ACTIONS(2099),
    [anon_sym_module] = ACTIONS(2099),
    [anon_sym_open] = ACTIONS(2099),
    [anon_sym_include] = ACTIONS(2099),
    [anon_sym_class] = ACTIONS(2099),
    [anon_sym_val] = ACTIONS(2099),
    [anon_sym_object] = ACTIONS(2099),
    [anon_sym_AMP] = ACTIONS(2099),
    [anon_sym_POUND] = ACTIONS(2099),
    [anon_sym_COLON_COLON] = ACTIONS(2101),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2101),
    [anon_sym_else] = ACTIONS(2099),
    [anon_sym_new] = ACTIONS(2099),
    [anon_sym_LBRACE_LT] = ACTIONS(2101),
    [anon_sym_begin] = ACTIONS(2099),
    [sym_ocamlyacc_value] = ACTIONS(2101),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2099),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2101),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2099),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2099),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2101),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2101),
    [aux_sym_number_token1] = ACTIONS(2101),
    [aux_sym_signed_number_token1] = ACTIONS(2099),
    [anon_sym_SQUOTE] = ACTIONS(2101),
    [anon_sym_DQUOTE] = ACTIONS(2101),
    [sym_prefix_operator] = ACTIONS(2101),
    [aux_sym_sign_operator_token1] = ACTIONS(2099),
    [sym_hash_operator] = ACTIONS(2101),
    [anon_sym_lsl] = ACTIONS(2099),
    [anon_sym_lsr] = ACTIONS(2099),
    [anon_sym_asr] = ACTIONS(2099),
    [aux_sym_pow_operator_token1] = ACTIONS(2101),
    [anon_sym_mod] = ACTIONS(2099),
    [anon_sym_land] = ACTIONS(2099),
    [anon_sym_lor] = ACTIONS(2099),
    [anon_sym_lxor] = ACTIONS(2099),
    [aux_sym_mult_operator_token1] = ACTIONS(2099),
    [aux_sym_add_operator_token1] = ACTIONS(2099),
    [sym_concat_operator] = ACTIONS(2101),
    [sym_rel_operator] = ACTIONS(2099),
    [anon_sym_AMP_AMP] = ACTIONS(2099),
    [anon_sym_or] = ACTIONS(2099),
    [anon_sym_PIPE_PIPE] = ACTIONS(2099),
    [sym_assign_operator] = ACTIONS(2101),
    [sym_let_operator] = ACTIONS(2101),
    [sym__uppercase_identifier] = ACTIONS(2101),
    [aux_sym_directive_token1] = ACTIONS(2099),
    [aux_sym_tag_token1] = ACTIONS(2101),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1428)] = {
    [sym_attribute] = STATE(1428),
    [ts_builtin_sym_end] = ACTIONS(2033),
    [sym__lowercase_identifier] = ACTIONS(2031),
    [anon_sym_SEMI_SEMI] = ACTIONS(2033),
    [anon_sym_let] = ACTIONS(2031),
    [anon_sym_and] = ACTIONS(2031),
    [anon_sym_TILDE] = ACTIONS(2031),
    [anon_sym_QMARK] = ACTIONS(2031),
    [anon_sym_LPAREN] = ACTIONS(2033),
    [anon_sym_external] = ACTIONS(2031),
    [anon_sym_type] = ACTIONS(2031),
    [anon_sym_COMMA] = ACTIONS(2033),
    [anon_sym_PIPE] = ACTIONS(2031),
    [anon_sym_DOT] = ACTIONS(2033),
    [anon_sym_LBRACK] = ACTIONS(2031),
    [anon_sym_true] = ACTIONS(2031),
    [anon_sym_false] = ACTIONS(2031),
    [anon_sym_LBRACE] = ACTIONS(2031),
    [anon_sym_SEMI] = ACTIONS(2031),
    [anon_sym_exception] = ACTIONS(2031),
    [anon_sym_module] = ACTIONS(2031),
    [anon_sym_open] = ACTIONS(2031),
    [anon_sym_include] = ACTIONS(2031),
    [anon_sym_class] = ACTIONS(2031),
    [anon_sym_val] = ACTIONS(2031),
    [anon_sym_object] = ACTIONS(2031),
    [anon_sym_AMP] = ACTIONS(2031),
    [anon_sym_POUND] = ACTIONS(2031),
    [anon_sym_COLON_COLON] = ACTIONS(2033),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2033),
    [anon_sym_else] = ACTIONS(2031),
    [anon_sym_new] = ACTIONS(2031),
    [anon_sym_LBRACE_LT] = ACTIONS(2033),
    [anon_sym_begin] = ACTIONS(2031),
    [sym_ocamlyacc_value] = ACTIONS(2033),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2031),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2033),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2031),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2031),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2033),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2033),
    [aux_sym_number_token1] = ACTIONS(2033),
    [aux_sym_signed_number_token1] = ACTIONS(2031),
    [anon_sym_SQUOTE] = ACTIONS(2033),
    [anon_sym_DQUOTE] = ACTIONS(2033),
    [sym_prefix_operator] = ACTIONS(2033),
    [aux_sym_sign_operator_token1] = ACTIONS(2031),
    [sym_hash_operator] = ACTIONS(2033),
    [anon_sym_lsl] = ACTIONS(2031),
    [anon_sym_lsr] = ACTIONS(2031),
    [anon_sym_asr] = ACTIONS(2031),
    [aux_sym_pow_operator_token1] = ACTIONS(2033),
    [anon_sym_mod] = ACTIONS(2031),
    [anon_sym_land] = ACTIONS(2031),
    [anon_sym_lor] = ACTIONS(2031),
    [anon_sym_lxor] = ACTIONS(2031),
    [aux_sym_mult_operator_token1] = ACTIONS(2031),
    [aux_sym_add_operator_token1] = ACTIONS(2031),
    [sym_concat_operator] = ACTIONS(2033),
    [sym_rel_operator] = ACTIONS(2031),
    [anon_sym_AMP_AMP] = ACTIONS(2031),
    [anon_sym_or] = ACTIONS(2031),
    [anon_sym_PIPE_PIPE] = ACTIONS(2031),
    [sym_assign_operator] = ACTIONS(2033),
    [sym_let_operator] = ACTIONS(2033),
    [sym__uppercase_identifier] = ACTIONS(2033),
    [aux_sym_directive_token1] = ACTIONS(2031),
    [aux_sym_tag_token1] = ACTIONS(2033),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1429)] = {
    [sym_attribute] = STATE(1429),
    [ts_builtin_sym_end] = ACTIONS(1953),
    [sym__lowercase_identifier] = ACTIONS(1951),
    [anon_sym_SEMI_SEMI] = ACTIONS(1953),
    [anon_sym_let] = ACTIONS(1951),
    [anon_sym_and] = ACTIONS(1951),
    [anon_sym_TILDE] = ACTIONS(1951),
    [anon_sym_QMARK] = ACTIONS(1951),
    [anon_sym_LPAREN] = ACTIONS(1953),
    [anon_sym_external] = ACTIONS(1951),
    [anon_sym_type] = ACTIONS(1951),
    [anon_sym_COMMA] = ACTIONS(1953),
    [anon_sym_PIPE] = ACTIONS(1951),
    [anon_sym_DOT] = ACTIONS(1953),
    [anon_sym_LBRACK] = ACTIONS(1951),
    [anon_sym_true] = ACTIONS(1951),
    [anon_sym_false] = ACTIONS(1951),
    [anon_sym_LBRACE] = ACTIONS(1951),
    [anon_sym_SEMI] = ACTIONS(1951),
    [anon_sym_exception] = ACTIONS(1951),
    [anon_sym_module] = ACTIONS(1951),
    [anon_sym_open] = ACTIONS(1951),
    [anon_sym_include] = ACTIONS(1951),
    [anon_sym_class] = ACTIONS(1951),
    [anon_sym_val] = ACTIONS(1951),
    [anon_sym_object] = ACTIONS(1951),
    [anon_sym_AMP] = ACTIONS(1951),
    [anon_sym_POUND] = ACTIONS(1951),
    [anon_sym_COLON_COLON] = ACTIONS(1953),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1953),
    [anon_sym_else] = ACTIONS(1951),
    [anon_sym_new] = ACTIONS(1951),
    [anon_sym_LBRACE_LT] = ACTIONS(1953),
    [anon_sym_begin] = ACTIONS(1951),
    [sym_ocamlyacc_value] = ACTIONS(1953),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1951),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1953),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1951),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1951),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1953),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1953),
    [aux_sym_number_token1] = ACTIONS(1953),
    [aux_sym_signed_number_token1] = ACTIONS(1951),
    [anon_sym_SQUOTE] = ACTIONS(1953),
    [anon_sym_DQUOTE] = ACTIONS(1953),
    [sym_prefix_operator] = ACTIONS(1953),
    [aux_sym_sign_operator_token1] = ACTIONS(1951),
    [sym_hash_operator] = ACTIONS(1953),
    [anon_sym_lsl] = ACTIONS(1951),
    [anon_sym_lsr] = ACTIONS(1951),
    [anon_sym_asr] = ACTIONS(1951),
    [aux_sym_pow_operator_token1] = ACTIONS(1953),
    [anon_sym_mod] = ACTIONS(1951),
    [anon_sym_land] = ACTIONS(1951),
    [anon_sym_lor] = ACTIONS(1951),
    [anon_sym_lxor] = ACTIONS(1951),
    [aux_sym_mult_operator_token1] = ACTIONS(1951),
    [aux_sym_add_operator_token1] = ACTIONS(1951),
    [sym_concat_operator] = ACTIONS(1953),
    [sym_rel_operator] = ACTIONS(1951),
    [anon_sym_AMP_AMP] = ACTIONS(1951),
    [anon_sym_or] = ACTIONS(1951),
    [anon_sym_PIPE_PIPE] = ACTIONS(1951),
    [sym_assign_operator] = ACTIONS(1953),
    [sym_let_operator] = ACTIONS(1953),
    [sym__uppercase_identifier] = ACTIONS(1953),
    [aux_sym_directive_token1] = ACTIONS(1951),
    [aux_sym_tag_token1] = ACTIONS(1953),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1430)] = {
    [sym_attribute] = STATE(1430),
    [ts_builtin_sym_end] = ACTIONS(2081),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_SEMI_SEMI] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2079),
    [anon_sym_and] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_external] = ACTIONS(2079),
    [anon_sym_type] = ACTIONS(2079),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_PIPE] = ACTIONS(2079),
    [anon_sym_DOT] = ACTIONS(2081),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_SEMI] = ACTIONS(2079),
    [anon_sym_exception] = ACTIONS(2079),
    [anon_sym_module] = ACTIONS(2079),
    [anon_sym_open] = ACTIONS(2079),
    [anon_sym_include] = ACTIONS(2079),
    [anon_sym_class] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_POUND] = ACTIONS(2079),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [sym_hash_operator] = ACTIONS(2081),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym_let_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_directive_token1] = ACTIONS(2079),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1431)] = {
    [sym_attribute] = STATE(1431),
    [ts_builtin_sym_end] = ACTIONS(2131),
    [sym__lowercase_identifier] = ACTIONS(2129),
    [anon_sym_SEMI_SEMI] = ACTIONS(2131),
    [anon_sym_let] = ACTIONS(2129),
    [anon_sym_and] = ACTIONS(2129),
    [anon_sym_TILDE] = ACTIONS(2129),
    [anon_sym_QMARK] = ACTIONS(2129),
    [anon_sym_LPAREN] = ACTIONS(2131),
    [anon_sym_external] = ACTIONS(2129),
    [anon_sym_type] = ACTIONS(2129),
    [anon_sym_COMMA] = ACTIONS(2131),
    [anon_sym_PIPE] = ACTIONS(2129),
    [anon_sym_DOT] = ACTIONS(2131),
    [anon_sym_LBRACK] = ACTIONS(2129),
    [anon_sym_true] = ACTIONS(2129),
    [anon_sym_false] = ACTIONS(2129),
    [anon_sym_LBRACE] = ACTIONS(2129),
    [anon_sym_SEMI] = ACTIONS(2129),
    [anon_sym_exception] = ACTIONS(2129),
    [anon_sym_module] = ACTIONS(2129),
    [anon_sym_open] = ACTIONS(2129),
    [anon_sym_include] = ACTIONS(2129),
    [anon_sym_class] = ACTIONS(2129),
    [anon_sym_val] = ACTIONS(2129),
    [anon_sym_object] = ACTIONS(2129),
    [anon_sym_AMP] = ACTIONS(2129),
    [anon_sym_POUND] = ACTIONS(2129),
    [anon_sym_COLON_COLON] = ACTIONS(2131),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2131),
    [anon_sym_else] = ACTIONS(2129),
    [anon_sym_new] = ACTIONS(2129),
    [anon_sym_LBRACE_LT] = ACTIONS(2131),
    [anon_sym_begin] = ACTIONS(2129),
    [sym_ocamlyacc_value] = ACTIONS(2131),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2129),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2131),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2129),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2129),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2131),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2131),
    [aux_sym_number_token1] = ACTIONS(2131),
    [aux_sym_signed_number_token1] = ACTIONS(2129),
    [anon_sym_SQUOTE] = ACTIONS(2131),
    [anon_sym_DQUOTE] = ACTIONS(2131),
    [sym_prefix_operator] = ACTIONS(2131),
    [aux_sym_sign_operator_token1] = ACTIONS(2129),
    [sym_hash_operator] = ACTIONS(2131),
    [anon_sym_lsl] = ACTIONS(2129),
    [anon_sym_lsr] = ACTIONS(2129),
    [anon_sym_asr] = ACTIONS(2129),
    [aux_sym_pow_operator_token1] = ACTIONS(2131),
    [anon_sym_mod] = ACTIONS(2129),
    [anon_sym_land] = ACTIONS(2129),
    [anon_sym_lor] = ACTIONS(2129),
    [anon_sym_lxor] = ACTIONS(2129),
    [aux_sym_mult_operator_token1] = ACTIONS(2129),
    [aux_sym_add_operator_token1] = ACTIONS(2129),
    [sym_concat_operator] = ACTIONS(2131),
    [sym_rel_operator] = ACTIONS(2129),
    [anon_sym_AMP_AMP] = ACTIONS(2129),
    [anon_sym_or] = ACTIONS(2129),
    [anon_sym_PIPE_PIPE] = ACTIONS(2129),
    [sym_assign_operator] = ACTIONS(2131),
    [sym_let_operator] = ACTIONS(2131),
    [sym__uppercase_identifier] = ACTIONS(2131),
    [aux_sym_directive_token1] = ACTIONS(2129),
    [aux_sym_tag_token1] = ACTIONS(2131),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1432)] = {
    [sym_attribute] = STATE(1432),
    [ts_builtin_sym_end] = ACTIONS(2147),
    [sym__lowercase_identifier] = ACTIONS(2145),
    [anon_sym_SEMI_SEMI] = ACTIONS(2147),
    [anon_sym_let] = ACTIONS(2145),
    [anon_sym_and] = ACTIONS(2145),
    [anon_sym_TILDE] = ACTIONS(2145),
    [anon_sym_QMARK] = ACTIONS(2145),
    [anon_sym_LPAREN] = ACTIONS(2147),
    [anon_sym_external] = ACTIONS(2145),
    [anon_sym_type] = ACTIONS(2145),
    [anon_sym_COMMA] = ACTIONS(2147),
    [anon_sym_PIPE] = ACTIONS(2145),
    [anon_sym_DOT] = ACTIONS(2147),
    [anon_sym_LBRACK] = ACTIONS(2145),
    [anon_sym_true] = ACTIONS(2145),
    [anon_sym_false] = ACTIONS(2145),
    [anon_sym_LBRACE] = ACTIONS(2145),
    [anon_sym_SEMI] = ACTIONS(2145),
    [anon_sym_exception] = ACTIONS(2145),
    [anon_sym_module] = ACTIONS(2145),
    [anon_sym_open] = ACTIONS(2145),
    [anon_sym_include] = ACTIONS(2145),
    [anon_sym_class] = ACTIONS(2145),
    [anon_sym_val] = ACTIONS(2145),
    [anon_sym_object] = ACTIONS(2145),
    [anon_sym_AMP] = ACTIONS(2145),
    [anon_sym_POUND] = ACTIONS(2145),
    [anon_sym_COLON_COLON] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2147),
    [anon_sym_else] = ACTIONS(2145),
    [anon_sym_new] = ACTIONS(2145),
    [anon_sym_LBRACE_LT] = ACTIONS(2147),
    [anon_sym_begin] = ACTIONS(2145),
    [sym_ocamlyacc_value] = ACTIONS(2147),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2145),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2147),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2145),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2145),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2147),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2147),
    [aux_sym_number_token1] = ACTIONS(2147),
    [aux_sym_signed_number_token1] = ACTIONS(2145),
    [anon_sym_SQUOTE] = ACTIONS(2147),
    [anon_sym_DQUOTE] = ACTIONS(2147),
    [sym_prefix_operator] = ACTIONS(2147),
    [aux_sym_sign_operator_token1] = ACTIONS(2145),
    [sym_hash_operator] = ACTIONS(2147),
    [anon_sym_lsl] = ACTIONS(2145),
    [anon_sym_lsr] = ACTIONS(2145),
    [anon_sym_asr] = ACTIONS(2145),
    [aux_sym_pow_operator_token1] = ACTIONS(2147),
    [anon_sym_mod] = ACTIONS(2145),
    [anon_sym_land] = ACTIONS(2145),
    [anon_sym_lor] = ACTIONS(2145),
    [anon_sym_lxor] = ACTIONS(2145),
    [aux_sym_mult_operator_token1] = ACTIONS(2145),
    [aux_sym_add_operator_token1] = ACTIONS(2145),
    [sym_concat_operator] = ACTIONS(2147),
    [sym_rel_operator] = ACTIONS(2145),
    [anon_sym_AMP_AMP] = ACTIONS(2145),
    [anon_sym_or] = ACTIONS(2145),
    [anon_sym_PIPE_PIPE] = ACTIONS(2145),
    [sym_assign_operator] = ACTIONS(2147),
    [sym_let_operator] = ACTIONS(2147),
    [sym__uppercase_identifier] = ACTIONS(2147),
    [aux_sym_directive_token1] = ACTIONS(2145),
    [aux_sym_tag_token1] = ACTIONS(2147),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1433)] = {
    [sym_attribute] = STATE(1433),
    [ts_builtin_sym_end] = ACTIONS(2151),
    [sym__lowercase_identifier] = ACTIONS(2149),
    [anon_sym_SEMI_SEMI] = ACTIONS(2151),
    [anon_sym_let] = ACTIONS(2149),
    [anon_sym_and] = ACTIONS(2149),
    [anon_sym_TILDE] = ACTIONS(2149),
    [anon_sym_QMARK] = ACTIONS(2149),
    [anon_sym_LPAREN] = ACTIONS(2151),
    [anon_sym_external] = ACTIONS(2149),
    [anon_sym_type] = ACTIONS(2149),
    [anon_sym_COMMA] = ACTIONS(2151),
    [anon_sym_PIPE] = ACTIONS(2149),
    [anon_sym_DOT] = ACTIONS(2151),
    [anon_sym_LBRACK] = ACTIONS(2149),
    [anon_sym_true] = ACTIONS(2149),
    [anon_sym_false] = ACTIONS(2149),
    [anon_sym_LBRACE] = ACTIONS(2149),
    [anon_sym_SEMI] = ACTIONS(2149),
    [anon_sym_exception] = ACTIONS(2149),
    [anon_sym_module] = ACTIONS(2149),
    [anon_sym_open] = ACTIONS(2149),
    [anon_sym_include] = ACTIONS(2149),
    [anon_sym_class] = ACTIONS(2149),
    [anon_sym_val] = ACTIONS(2149),
    [anon_sym_object] = ACTIONS(2149),
    [anon_sym_AMP] = ACTIONS(2149),
    [anon_sym_POUND] = ACTIONS(2149),
    [anon_sym_COLON_COLON] = ACTIONS(2151),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2151),
    [anon_sym_else] = ACTIONS(2149),
    [anon_sym_new] = ACTIONS(2149),
    [anon_sym_LBRACE_LT] = ACTIONS(2151),
    [anon_sym_begin] = ACTIONS(2149),
    [sym_ocamlyacc_value] = ACTIONS(2151),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2149),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2151),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2149),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2151),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2151),
    [aux_sym_number_token1] = ACTIONS(2151),
    [aux_sym_signed_number_token1] = ACTIONS(2149),
    [anon_sym_SQUOTE] = ACTIONS(2151),
    [anon_sym_DQUOTE] = ACTIONS(2151),
    [sym_prefix_operator] = ACTIONS(2151),
    [aux_sym_sign_operator_token1] = ACTIONS(2149),
    [sym_hash_operator] = ACTIONS(2151),
    [anon_sym_lsl] = ACTIONS(2149),
    [anon_sym_lsr] = ACTIONS(2149),
    [anon_sym_asr] = ACTIONS(2149),
    [aux_sym_pow_operator_token1] = ACTIONS(2151),
    [anon_sym_mod] = ACTIONS(2149),
    [anon_sym_land] = ACTIONS(2149),
    [anon_sym_lor] = ACTIONS(2149),
    [anon_sym_lxor] = ACTIONS(2149),
    [aux_sym_mult_operator_token1] = ACTIONS(2149),
    [aux_sym_add_operator_token1] = ACTIONS(2149),
    [sym_concat_operator] = ACTIONS(2151),
    [sym_rel_operator] = ACTIONS(2149),
    [anon_sym_AMP_AMP] = ACTIONS(2149),
    [anon_sym_or] = ACTIONS(2149),
    [anon_sym_PIPE_PIPE] = ACTIONS(2149),
    [sym_assign_operator] = ACTIONS(2151),
    [sym_let_operator] = ACTIONS(2151),
    [sym__uppercase_identifier] = ACTIONS(2151),
    [aux_sym_directive_token1] = ACTIONS(2149),
    [aux_sym_tag_token1] = ACTIONS(2151),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1434)] = {
    [sym_attribute] = STATE(1434),
    [ts_builtin_sym_end] = ACTIONS(2069),
    [sym__lowercase_identifier] = ACTIONS(2067),
    [anon_sym_SEMI_SEMI] = ACTIONS(2069),
    [anon_sym_let] = ACTIONS(2067),
    [anon_sym_and] = ACTIONS(2067),
    [anon_sym_TILDE] = ACTIONS(2067),
    [anon_sym_QMARK] = ACTIONS(2067),
    [anon_sym_LPAREN] = ACTIONS(2069),
    [anon_sym_external] = ACTIONS(2067),
    [anon_sym_type] = ACTIONS(2067),
    [anon_sym_COMMA] = ACTIONS(2069),
    [anon_sym_PIPE] = ACTIONS(2067),
    [anon_sym_DOT] = ACTIONS(2069),
    [anon_sym_LBRACK] = ACTIONS(2067),
    [anon_sym_true] = ACTIONS(2067),
    [anon_sym_false] = ACTIONS(2067),
    [anon_sym_LBRACE] = ACTIONS(2067),
    [anon_sym_SEMI] = ACTIONS(2067),
    [anon_sym_exception] = ACTIONS(2067),
    [anon_sym_module] = ACTIONS(2067),
    [anon_sym_open] = ACTIONS(2067),
    [anon_sym_include] = ACTIONS(2067),
    [anon_sym_class] = ACTIONS(2067),
    [anon_sym_val] = ACTIONS(2067),
    [anon_sym_object] = ACTIONS(2067),
    [anon_sym_AMP] = ACTIONS(2067),
    [anon_sym_POUND] = ACTIONS(2067),
    [anon_sym_COLON_COLON] = ACTIONS(2069),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2069),
    [anon_sym_else] = ACTIONS(2067),
    [anon_sym_new] = ACTIONS(2067),
    [anon_sym_LBRACE_LT] = ACTIONS(2069),
    [anon_sym_begin] = ACTIONS(2067),
    [sym_ocamlyacc_value] = ACTIONS(2069),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2067),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2069),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2067),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2067),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2069),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2069),
    [aux_sym_number_token1] = ACTIONS(2069),
    [aux_sym_signed_number_token1] = ACTIONS(2067),
    [anon_sym_SQUOTE] = ACTIONS(2069),
    [anon_sym_DQUOTE] = ACTIONS(2069),
    [sym_prefix_operator] = ACTIONS(2069),
    [aux_sym_sign_operator_token1] = ACTIONS(2067),
    [sym_hash_operator] = ACTIONS(2069),
    [anon_sym_lsl] = ACTIONS(2067),
    [anon_sym_lsr] = ACTIONS(2067),
    [anon_sym_asr] = ACTIONS(2067),
    [aux_sym_pow_operator_token1] = ACTIONS(2069),
    [anon_sym_mod] = ACTIONS(2067),
    [anon_sym_land] = ACTIONS(2067),
    [anon_sym_lor] = ACTIONS(2067),
    [anon_sym_lxor] = ACTIONS(2067),
    [aux_sym_mult_operator_token1] = ACTIONS(2067),
    [aux_sym_add_operator_token1] = ACTIONS(2067),
    [sym_concat_operator] = ACTIONS(2069),
    [sym_rel_operator] = ACTIONS(2067),
    [anon_sym_AMP_AMP] = ACTIONS(2067),
    [anon_sym_or] = ACTIONS(2067),
    [anon_sym_PIPE_PIPE] = ACTIONS(2067),
    [sym_assign_operator] = ACTIONS(2069),
    [sym_let_operator] = ACTIONS(2069),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_directive_token1] = ACTIONS(2067),
    [aux_sym_tag_token1] = ACTIONS(2069),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1435)] = {
    [sym_attribute] = STATE(1435),
    [ts_builtin_sym_end] = ACTIONS(2143),
    [sym__lowercase_identifier] = ACTIONS(2141),
    [anon_sym_SEMI_SEMI] = ACTIONS(2143),
    [anon_sym_let] = ACTIONS(2141),
    [anon_sym_and] = ACTIONS(2141),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_QMARK] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2143),
    [anon_sym_external] = ACTIONS(2141),
    [anon_sym_type] = ACTIONS(2141),
    [anon_sym_COMMA] = ACTIONS(2143),
    [anon_sym_PIPE] = ACTIONS(2141),
    [anon_sym_DOT] = ACTIONS(2143),
    [anon_sym_LBRACK] = ACTIONS(2141),
    [anon_sym_true] = ACTIONS(2141),
    [anon_sym_false] = ACTIONS(2141),
    [anon_sym_LBRACE] = ACTIONS(2141),
    [anon_sym_SEMI] = ACTIONS(2141),
    [anon_sym_exception] = ACTIONS(2141),
    [anon_sym_module] = ACTIONS(2141),
    [anon_sym_open] = ACTIONS(2141),
    [anon_sym_include] = ACTIONS(2141),
    [anon_sym_class] = ACTIONS(2141),
    [anon_sym_val] = ACTIONS(2141),
    [anon_sym_object] = ACTIONS(2141),
    [anon_sym_AMP] = ACTIONS(2141),
    [anon_sym_POUND] = ACTIONS(2141),
    [anon_sym_COLON_COLON] = ACTIONS(2143),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2143),
    [anon_sym_else] = ACTIONS(2141),
    [anon_sym_new] = ACTIONS(2141),
    [anon_sym_LBRACE_LT] = ACTIONS(2143),
    [anon_sym_begin] = ACTIONS(2141),
    [sym_ocamlyacc_value] = ACTIONS(2143),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2141),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2143),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2141),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2141),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2143),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2143),
    [aux_sym_number_token1] = ACTIONS(2143),
    [aux_sym_signed_number_token1] = ACTIONS(2141),
    [anon_sym_SQUOTE] = ACTIONS(2143),
    [anon_sym_DQUOTE] = ACTIONS(2143),
    [sym_prefix_operator] = ACTIONS(2143),
    [aux_sym_sign_operator_token1] = ACTIONS(2141),
    [sym_hash_operator] = ACTIONS(2143),
    [anon_sym_lsl] = ACTIONS(2141),
    [anon_sym_lsr] = ACTIONS(2141),
    [anon_sym_asr] = ACTIONS(2141),
    [aux_sym_pow_operator_token1] = ACTIONS(2143),
    [anon_sym_mod] = ACTIONS(2141),
    [anon_sym_land] = ACTIONS(2141),
    [anon_sym_lor] = ACTIONS(2141),
    [anon_sym_lxor] = ACTIONS(2141),
    [aux_sym_mult_operator_token1] = ACTIONS(2141),
    [aux_sym_add_operator_token1] = ACTIONS(2141),
    [sym_concat_operator] = ACTIONS(2143),
    [sym_rel_operator] = ACTIONS(2141),
    [anon_sym_AMP_AMP] = ACTIONS(2141),
    [anon_sym_or] = ACTIONS(2141),
    [anon_sym_PIPE_PIPE] = ACTIONS(2141),
    [sym_assign_operator] = ACTIONS(2143),
    [sym_let_operator] = ACTIONS(2143),
    [sym__uppercase_identifier] = ACTIONS(2143),
    [aux_sym_directive_token1] = ACTIONS(2141),
    [aux_sym_tag_token1] = ACTIONS(2143),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1436)] = {
    [sym_attribute] = STATE(1436),
    [ts_builtin_sym_end] = ACTIONS(2113),
    [sym__lowercase_identifier] = ACTIONS(2111),
    [anon_sym_SEMI_SEMI] = ACTIONS(2113),
    [anon_sym_let] = ACTIONS(2111),
    [anon_sym_and] = ACTIONS(2111),
    [anon_sym_TILDE] = ACTIONS(2111),
    [anon_sym_QMARK] = ACTIONS(2111),
    [anon_sym_LPAREN] = ACTIONS(2113),
    [anon_sym_external] = ACTIONS(2111),
    [anon_sym_type] = ACTIONS(2111),
    [anon_sym_COMMA] = ACTIONS(2113),
    [anon_sym_PIPE] = ACTIONS(2111),
    [anon_sym_DOT] = ACTIONS(2115),
    [anon_sym_LBRACK] = ACTIONS(2111),
    [anon_sym_true] = ACTIONS(2111),
    [anon_sym_false] = ACTIONS(2111),
    [anon_sym_LBRACE] = ACTIONS(2111),
    [anon_sym_SEMI] = ACTIONS(2111),
    [anon_sym_exception] = ACTIONS(2111),
    [anon_sym_module] = ACTIONS(2111),
    [anon_sym_open] = ACTIONS(2111),
    [anon_sym_include] = ACTIONS(2111),
    [anon_sym_class] = ACTIONS(2111),
    [anon_sym_val] = ACTIONS(2111),
    [anon_sym_object] = ACTIONS(2111),
    [anon_sym_AMP] = ACTIONS(2111),
    [anon_sym_POUND] = ACTIONS(2111),
    [anon_sym_COLON_COLON] = ACTIONS(2113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2113),
    [anon_sym_else] = ACTIONS(2111),
    [anon_sym_new] = ACTIONS(2111),
    [anon_sym_LBRACE_LT] = ACTIONS(2113),
    [anon_sym_begin] = ACTIONS(2111),
    [sym_ocamlyacc_value] = ACTIONS(2113),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2113),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2113),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2113),
    [aux_sym_number_token1] = ACTIONS(2113),
    [aux_sym_signed_number_token1] = ACTIONS(2111),
    [anon_sym_SQUOTE] = ACTIONS(2113),
    [anon_sym_DQUOTE] = ACTIONS(2113),
    [sym_prefix_operator] = ACTIONS(2113),
    [aux_sym_sign_operator_token1] = ACTIONS(2111),
    [sym_hash_operator] = ACTIONS(2113),
    [anon_sym_lsl] = ACTIONS(2111),
    [anon_sym_lsr] = ACTIONS(2111),
    [anon_sym_asr] = ACTIONS(2111),
    [aux_sym_pow_operator_token1] = ACTIONS(2113),
    [anon_sym_mod] = ACTIONS(2111),
    [anon_sym_land] = ACTIONS(2111),
    [anon_sym_lor] = ACTIONS(2111),
    [anon_sym_lxor] = ACTIONS(2111),
    [aux_sym_mult_operator_token1] = ACTIONS(2111),
    [aux_sym_add_operator_token1] = ACTIONS(2111),
    [sym_concat_operator] = ACTIONS(2113),
    [sym_rel_operator] = ACTIONS(2111),
    [anon_sym_AMP_AMP] = ACTIONS(2111),
    [anon_sym_or] = ACTIONS(2111),
    [anon_sym_PIPE_PIPE] = ACTIONS(2111),
    [sym_assign_operator] = ACTIONS(2113),
    [sym_let_operator] = ACTIONS(2113),
    [sym__uppercase_identifier] = ACTIONS(2113),
    [aux_sym_directive_token1] = ACTIONS(2111),
    [aux_sym_tag_token1] = ACTIONS(2113),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1437)] = {
    [sym_attribute] = STATE(1437),
    [ts_builtin_sym_end] = ACTIONS(1985),
    [sym__lowercase_identifier] = ACTIONS(1983),
    [anon_sym_SEMI_SEMI] = ACTIONS(1985),
    [anon_sym_let] = ACTIONS(1983),
    [anon_sym_and] = ACTIONS(1983),
    [anon_sym_TILDE] = ACTIONS(1983),
    [anon_sym_QMARK] = ACTIONS(1983),
    [anon_sym_LPAREN] = ACTIONS(1985),
    [anon_sym_external] = ACTIONS(1983),
    [anon_sym_type] = ACTIONS(1983),
    [anon_sym_COMMA] = ACTIONS(1985),
    [anon_sym_PIPE] = ACTIONS(1983),
    [anon_sym_DOT] = ACTIONS(1985),
    [anon_sym_LBRACK] = ACTIONS(1983),
    [anon_sym_true] = ACTIONS(1983),
    [anon_sym_false] = ACTIONS(1983),
    [anon_sym_LBRACE] = ACTIONS(1983),
    [anon_sym_SEMI] = ACTIONS(1983),
    [anon_sym_exception] = ACTIONS(1983),
    [anon_sym_module] = ACTIONS(1983),
    [anon_sym_open] = ACTIONS(1983),
    [anon_sym_include] = ACTIONS(1983),
    [anon_sym_class] = ACTIONS(1983),
    [anon_sym_val] = ACTIONS(1983),
    [anon_sym_object] = ACTIONS(1983),
    [anon_sym_AMP] = ACTIONS(1983),
    [anon_sym_POUND] = ACTIONS(1983),
    [anon_sym_COLON_COLON] = ACTIONS(1985),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1985),
    [anon_sym_else] = ACTIONS(1983),
    [anon_sym_new] = ACTIONS(1983),
    [anon_sym_LBRACE_LT] = ACTIONS(1985),
    [anon_sym_begin] = ACTIONS(1983),
    [sym_ocamlyacc_value] = ACTIONS(1985),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1983),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1985),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1983),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1983),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1985),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1985),
    [aux_sym_number_token1] = ACTIONS(1985),
    [aux_sym_signed_number_token1] = ACTIONS(1983),
    [anon_sym_SQUOTE] = ACTIONS(1985),
    [anon_sym_DQUOTE] = ACTIONS(1985),
    [sym_prefix_operator] = ACTIONS(1985),
    [aux_sym_sign_operator_token1] = ACTIONS(1983),
    [sym_hash_operator] = ACTIONS(1985),
    [anon_sym_lsl] = ACTIONS(1983),
    [anon_sym_lsr] = ACTIONS(1983),
    [anon_sym_asr] = ACTIONS(1983),
    [aux_sym_pow_operator_token1] = ACTIONS(1985),
    [anon_sym_mod] = ACTIONS(1983),
    [anon_sym_land] = ACTIONS(1983),
    [anon_sym_lor] = ACTIONS(1983),
    [anon_sym_lxor] = ACTIONS(1983),
    [aux_sym_mult_operator_token1] = ACTIONS(1983),
    [aux_sym_add_operator_token1] = ACTIONS(1983),
    [sym_concat_operator] = ACTIONS(1985),
    [sym_rel_operator] = ACTIONS(1983),
    [anon_sym_AMP_AMP] = ACTIONS(1983),
    [anon_sym_or] = ACTIONS(1983),
    [anon_sym_PIPE_PIPE] = ACTIONS(1983),
    [sym_assign_operator] = ACTIONS(1985),
    [sym_let_operator] = ACTIONS(1985),
    [sym__uppercase_identifier] = ACTIONS(1985),
    [aux_sym_directive_token1] = ACTIONS(1983),
    [aux_sym_tag_token1] = ACTIONS(1985),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1438)] = {
    [sym_attribute] = STATE(1438),
    [ts_builtin_sym_end] = ACTIONS(1989),
    [sym__lowercase_identifier] = ACTIONS(1987),
    [anon_sym_SEMI_SEMI] = ACTIONS(1989),
    [anon_sym_let] = ACTIONS(1987),
    [anon_sym_and] = ACTIONS(1987),
    [anon_sym_TILDE] = ACTIONS(1987),
    [anon_sym_QMARK] = ACTIONS(1987),
    [anon_sym_LPAREN] = ACTIONS(1989),
    [anon_sym_external] = ACTIONS(1987),
    [anon_sym_type] = ACTIONS(1987),
    [anon_sym_COMMA] = ACTIONS(1989),
    [anon_sym_PIPE] = ACTIONS(1987),
    [anon_sym_DOT] = ACTIONS(1989),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_true] = ACTIONS(1987),
    [anon_sym_false] = ACTIONS(1987),
    [anon_sym_LBRACE] = ACTIONS(1987),
    [anon_sym_SEMI] = ACTIONS(1987),
    [anon_sym_exception] = ACTIONS(1987),
    [anon_sym_module] = ACTIONS(1987),
    [anon_sym_open] = ACTIONS(1987),
    [anon_sym_include] = ACTIONS(1987),
    [anon_sym_class] = ACTIONS(1987),
    [anon_sym_val] = ACTIONS(1987),
    [anon_sym_object] = ACTIONS(1987),
    [anon_sym_AMP] = ACTIONS(1987),
    [anon_sym_POUND] = ACTIONS(1987),
    [anon_sym_COLON_COLON] = ACTIONS(1989),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1989),
    [anon_sym_else] = ACTIONS(1987),
    [anon_sym_new] = ACTIONS(1987),
    [anon_sym_LBRACE_LT] = ACTIONS(1989),
    [anon_sym_begin] = ACTIONS(1987),
    [sym_ocamlyacc_value] = ACTIONS(1989),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1987),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1989),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1987),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1987),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1989),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1989),
    [aux_sym_number_token1] = ACTIONS(1989),
    [aux_sym_signed_number_token1] = ACTIONS(1987),
    [anon_sym_SQUOTE] = ACTIONS(1989),
    [anon_sym_DQUOTE] = ACTIONS(1989),
    [sym_prefix_operator] = ACTIONS(1989),
    [aux_sym_sign_operator_token1] = ACTIONS(1987),
    [sym_hash_operator] = ACTIONS(1989),
    [anon_sym_lsl] = ACTIONS(1987),
    [anon_sym_lsr] = ACTIONS(1987),
    [anon_sym_asr] = ACTIONS(1987),
    [aux_sym_pow_operator_token1] = ACTIONS(1989),
    [anon_sym_mod] = ACTIONS(1987),
    [anon_sym_land] = ACTIONS(1987),
    [anon_sym_lor] = ACTIONS(1987),
    [anon_sym_lxor] = ACTIONS(1987),
    [aux_sym_mult_operator_token1] = ACTIONS(1987),
    [aux_sym_add_operator_token1] = ACTIONS(1987),
    [sym_concat_operator] = ACTIONS(1989),
    [sym_rel_operator] = ACTIONS(1987),
    [anon_sym_AMP_AMP] = ACTIONS(1987),
    [anon_sym_or] = ACTIONS(1987),
    [anon_sym_PIPE_PIPE] = ACTIONS(1987),
    [sym_assign_operator] = ACTIONS(1989),
    [sym_let_operator] = ACTIONS(1989),
    [sym__uppercase_identifier] = ACTIONS(1989),
    [aux_sym_directive_token1] = ACTIONS(1987),
    [aux_sym_tag_token1] = ACTIONS(1989),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1439)] = {
    [sym_attribute] = STATE(1439),
    [ts_builtin_sym_end] = ACTIONS(1993),
    [sym__lowercase_identifier] = ACTIONS(1991),
    [anon_sym_SEMI_SEMI] = ACTIONS(1993),
    [anon_sym_let] = ACTIONS(1991),
    [anon_sym_and] = ACTIONS(1991),
    [anon_sym_TILDE] = ACTIONS(1991),
    [anon_sym_QMARK] = ACTIONS(1991),
    [anon_sym_LPAREN] = ACTIONS(1993),
    [anon_sym_external] = ACTIONS(1991),
    [anon_sym_type] = ACTIONS(1991),
    [anon_sym_COMMA] = ACTIONS(1993),
    [anon_sym_PIPE] = ACTIONS(1991),
    [anon_sym_DOT] = ACTIONS(1993),
    [anon_sym_LBRACK] = ACTIONS(1991),
    [anon_sym_true] = ACTIONS(1991),
    [anon_sym_false] = ACTIONS(1991),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_SEMI] = ACTIONS(1991),
    [anon_sym_exception] = ACTIONS(1991),
    [anon_sym_module] = ACTIONS(1991),
    [anon_sym_open] = ACTIONS(1991),
    [anon_sym_include] = ACTIONS(1991),
    [anon_sym_class] = ACTIONS(1991),
    [anon_sym_val] = ACTIONS(1991),
    [anon_sym_object] = ACTIONS(1991),
    [anon_sym_AMP] = ACTIONS(1991),
    [anon_sym_POUND] = ACTIONS(1991),
    [anon_sym_COLON_COLON] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1993),
    [anon_sym_else] = ACTIONS(1991),
    [anon_sym_new] = ACTIONS(1991),
    [anon_sym_LBRACE_LT] = ACTIONS(1993),
    [anon_sym_begin] = ACTIONS(1991),
    [sym_ocamlyacc_value] = ACTIONS(1993),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1991),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1993),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1991),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1991),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1993),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1993),
    [aux_sym_number_token1] = ACTIONS(1993),
    [aux_sym_signed_number_token1] = ACTIONS(1991),
    [anon_sym_SQUOTE] = ACTIONS(1993),
    [anon_sym_DQUOTE] = ACTIONS(1993),
    [sym_prefix_operator] = ACTIONS(1993),
    [aux_sym_sign_operator_token1] = ACTIONS(1991),
    [sym_hash_operator] = ACTIONS(1993),
    [anon_sym_lsl] = ACTIONS(1991),
    [anon_sym_lsr] = ACTIONS(1991),
    [anon_sym_asr] = ACTIONS(1991),
    [aux_sym_pow_operator_token1] = ACTIONS(1993),
    [anon_sym_mod] = ACTIONS(1991),
    [anon_sym_land] = ACTIONS(1991),
    [anon_sym_lor] = ACTIONS(1991),
    [anon_sym_lxor] = ACTIONS(1991),
    [aux_sym_mult_operator_token1] = ACTIONS(1991),
    [aux_sym_add_operator_token1] = ACTIONS(1991),
    [sym_concat_operator] = ACTIONS(1993),
    [sym_rel_operator] = ACTIONS(1991),
    [anon_sym_AMP_AMP] = ACTIONS(1991),
    [anon_sym_or] = ACTIONS(1991),
    [anon_sym_PIPE_PIPE] = ACTIONS(1991),
    [sym_assign_operator] = ACTIONS(1993),
    [sym_let_operator] = ACTIONS(1993),
    [sym__uppercase_identifier] = ACTIONS(1993),
    [aux_sym_directive_token1] = ACTIONS(1991),
    [aux_sym_tag_token1] = ACTIONS(1993),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1440)] = {
    [sym_attribute] = STATE(1440),
    [ts_builtin_sym_end] = ACTIONS(2163),
    [sym__lowercase_identifier] = ACTIONS(2161),
    [anon_sym_SEMI_SEMI] = ACTIONS(2163),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_and] = ACTIONS(2161),
    [anon_sym_TILDE] = ACTIONS(2161),
    [anon_sym_QMARK] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_external] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_COMMA] = ACTIONS(2163),
    [anon_sym_PIPE] = ACTIONS(2161),
    [anon_sym_DOT] = ACTIONS(2163),
    [anon_sym_LBRACK] = ACTIONS(2161),
    [anon_sym_true] = ACTIONS(2161),
    [anon_sym_false] = ACTIONS(2161),
    [anon_sym_LBRACE] = ACTIONS(2161),
    [anon_sym_SEMI] = ACTIONS(2161),
    [anon_sym_exception] = ACTIONS(2161),
    [anon_sym_module] = ACTIONS(2161),
    [anon_sym_open] = ACTIONS(2161),
    [anon_sym_include] = ACTIONS(2161),
    [anon_sym_class] = ACTIONS(2161),
    [anon_sym_val] = ACTIONS(2161),
    [anon_sym_object] = ACTIONS(2161),
    [anon_sym_AMP] = ACTIONS(2161),
    [anon_sym_POUND] = ACTIONS(2161),
    [anon_sym_COLON_COLON] = ACTIONS(2163),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2163),
    [anon_sym_else] = ACTIONS(2161),
    [anon_sym_new] = ACTIONS(2161),
    [anon_sym_LBRACE_LT] = ACTIONS(2163),
    [anon_sym_begin] = ACTIONS(2161),
    [sym_ocamlyacc_value] = ACTIONS(2163),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2161),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2163),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2163),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2163),
    [aux_sym_number_token1] = ACTIONS(2163),
    [aux_sym_signed_number_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2163),
    [anon_sym_DQUOTE] = ACTIONS(2163),
    [sym_prefix_operator] = ACTIONS(2163),
    [aux_sym_sign_operator_token1] = ACTIONS(2161),
    [sym_hash_operator] = ACTIONS(2163),
    [anon_sym_lsl] = ACTIONS(2161),
    [anon_sym_lsr] = ACTIONS(2161),
    [anon_sym_asr] = ACTIONS(2161),
    [aux_sym_pow_operator_token1] = ACTIONS(2163),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_land] = ACTIONS(2161),
    [anon_sym_lor] = ACTIONS(2161),
    [anon_sym_lxor] = ACTIONS(2161),
    [aux_sym_mult_operator_token1] = ACTIONS(2161),
    [aux_sym_add_operator_token1] = ACTIONS(2161),
    [sym_concat_operator] = ACTIONS(2163),
    [sym_rel_operator] = ACTIONS(2161),
    [anon_sym_AMP_AMP] = ACTIONS(2161),
    [anon_sym_or] = ACTIONS(2161),
    [anon_sym_PIPE_PIPE] = ACTIONS(2161),
    [sym_assign_operator] = ACTIONS(2163),
    [sym_let_operator] = ACTIONS(2163),
    [sym__uppercase_identifier] = ACTIONS(2163),
    [aux_sym_directive_token1] = ACTIONS(2161),
    [aux_sym_tag_token1] = ACTIONS(2163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1441)] = {
    [sym_attribute] = STATE(1441),
    [ts_builtin_sym_end] = ACTIONS(2119),
    [sym__lowercase_identifier] = ACTIONS(2117),
    [anon_sym_SEMI_SEMI] = ACTIONS(2119),
    [anon_sym_let] = ACTIONS(2117),
    [anon_sym_and] = ACTIONS(2117),
    [anon_sym_TILDE] = ACTIONS(2117),
    [anon_sym_QMARK] = ACTIONS(2117),
    [anon_sym_LPAREN] = ACTIONS(2119),
    [anon_sym_external] = ACTIONS(2117),
    [anon_sym_type] = ACTIONS(2117),
    [anon_sym_COMMA] = ACTIONS(2119),
    [anon_sym_PIPE] = ACTIONS(2117),
    [anon_sym_DOT] = ACTIONS(2119),
    [anon_sym_LBRACK] = ACTIONS(2117),
    [anon_sym_true] = ACTIONS(2117),
    [anon_sym_false] = ACTIONS(2117),
    [anon_sym_LBRACE] = ACTIONS(2117),
    [anon_sym_SEMI] = ACTIONS(2117),
    [anon_sym_exception] = ACTIONS(2117),
    [anon_sym_module] = ACTIONS(2117),
    [anon_sym_open] = ACTIONS(2117),
    [anon_sym_include] = ACTIONS(2117),
    [anon_sym_class] = ACTIONS(2117),
    [anon_sym_val] = ACTIONS(2117),
    [anon_sym_object] = ACTIONS(2117),
    [anon_sym_AMP] = ACTIONS(2117),
    [anon_sym_POUND] = ACTIONS(2117),
    [anon_sym_COLON_COLON] = ACTIONS(2119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2119),
    [anon_sym_else] = ACTIONS(2117),
    [anon_sym_new] = ACTIONS(2117),
    [anon_sym_LBRACE_LT] = ACTIONS(2119),
    [anon_sym_begin] = ACTIONS(2117),
    [sym_ocamlyacc_value] = ACTIONS(2119),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2117),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2119),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2117),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2117),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2119),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2119),
    [aux_sym_number_token1] = ACTIONS(2119),
    [aux_sym_signed_number_token1] = ACTIONS(2117),
    [anon_sym_SQUOTE] = ACTIONS(2119),
    [anon_sym_DQUOTE] = ACTIONS(2119),
    [sym_prefix_operator] = ACTIONS(2119),
    [aux_sym_sign_operator_token1] = ACTIONS(2117),
    [sym_hash_operator] = ACTIONS(2119),
    [anon_sym_lsl] = ACTIONS(2117),
    [anon_sym_lsr] = ACTIONS(2117),
    [anon_sym_asr] = ACTIONS(2117),
    [aux_sym_pow_operator_token1] = ACTIONS(2119),
    [anon_sym_mod] = ACTIONS(2117),
    [anon_sym_land] = ACTIONS(2117),
    [anon_sym_lor] = ACTIONS(2117),
    [anon_sym_lxor] = ACTIONS(2117),
    [aux_sym_mult_operator_token1] = ACTIONS(2117),
    [aux_sym_add_operator_token1] = ACTIONS(2117),
    [sym_concat_operator] = ACTIONS(2119),
    [sym_rel_operator] = ACTIONS(2117),
    [anon_sym_AMP_AMP] = ACTIONS(2117),
    [anon_sym_or] = ACTIONS(2117),
    [anon_sym_PIPE_PIPE] = ACTIONS(2117),
    [sym_assign_operator] = ACTIONS(2119),
    [sym_let_operator] = ACTIONS(2119),
    [sym__uppercase_identifier] = ACTIONS(2119),
    [aux_sym_directive_token1] = ACTIONS(2117),
    [aux_sym_tag_token1] = ACTIONS(2119),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1442)] = {
    [sym_attribute] = STATE(1442),
    [ts_builtin_sym_end] = ACTIONS(2209),
    [sym__lowercase_identifier] = ACTIONS(2207),
    [anon_sym_SEMI_SEMI] = ACTIONS(2209),
    [anon_sym_let] = ACTIONS(2207),
    [anon_sym_and] = ACTIONS(2207),
    [anon_sym_TILDE] = ACTIONS(2207),
    [anon_sym_QMARK] = ACTIONS(2207),
    [anon_sym_LPAREN] = ACTIONS(2209),
    [anon_sym_external] = ACTIONS(2207),
    [anon_sym_type] = ACTIONS(2207),
    [anon_sym_COMMA] = ACTIONS(2209),
    [anon_sym_PIPE] = ACTIONS(2207),
    [anon_sym_DOT] = ACTIONS(2209),
    [anon_sym_LBRACK] = ACTIONS(2207),
    [anon_sym_true] = ACTIONS(2207),
    [anon_sym_false] = ACTIONS(2207),
    [anon_sym_LBRACE] = ACTIONS(2207),
    [anon_sym_SEMI] = ACTIONS(2207),
    [anon_sym_exception] = ACTIONS(2207),
    [anon_sym_module] = ACTIONS(2207),
    [anon_sym_open] = ACTIONS(2207),
    [anon_sym_include] = ACTIONS(2207),
    [anon_sym_class] = ACTIONS(2207),
    [anon_sym_val] = ACTIONS(2207),
    [anon_sym_object] = ACTIONS(2207),
    [anon_sym_AMP] = ACTIONS(2207),
    [anon_sym_POUND] = ACTIONS(2207),
    [anon_sym_COLON_COLON] = ACTIONS(2209),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2209),
    [anon_sym_else] = ACTIONS(2207),
    [anon_sym_new] = ACTIONS(2207),
    [anon_sym_LBRACE_LT] = ACTIONS(2209),
    [anon_sym_begin] = ACTIONS(2207),
    [sym_ocamlyacc_value] = ACTIONS(2209),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2207),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2209),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2207),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2207),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2209),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2209),
    [aux_sym_number_token1] = ACTIONS(2209),
    [aux_sym_signed_number_token1] = ACTIONS(2207),
    [anon_sym_SQUOTE] = ACTIONS(2209),
    [anon_sym_DQUOTE] = ACTIONS(2209),
    [sym_prefix_operator] = ACTIONS(2209),
    [aux_sym_sign_operator_token1] = ACTIONS(2207),
    [sym_hash_operator] = ACTIONS(2209),
    [anon_sym_lsl] = ACTIONS(2207),
    [anon_sym_lsr] = ACTIONS(2207),
    [anon_sym_asr] = ACTIONS(2207),
    [aux_sym_pow_operator_token1] = ACTIONS(2209),
    [anon_sym_mod] = ACTIONS(2207),
    [anon_sym_land] = ACTIONS(2207),
    [anon_sym_lor] = ACTIONS(2207),
    [anon_sym_lxor] = ACTIONS(2207),
    [aux_sym_mult_operator_token1] = ACTIONS(2207),
    [aux_sym_add_operator_token1] = ACTIONS(2207),
    [sym_concat_operator] = ACTIONS(2209),
    [sym_rel_operator] = ACTIONS(2207),
    [anon_sym_AMP_AMP] = ACTIONS(2207),
    [anon_sym_or] = ACTIONS(2207),
    [anon_sym_PIPE_PIPE] = ACTIONS(2207),
    [sym_assign_operator] = ACTIONS(2209),
    [sym_let_operator] = ACTIONS(2209),
    [sym__uppercase_identifier] = ACTIONS(2209),
    [aux_sym_directive_token1] = ACTIONS(2207),
    [aux_sym_tag_token1] = ACTIONS(2209),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1443)] = {
    [sym_attribute] = STATE(1443),
    [ts_builtin_sym_end] = ACTIONS(2045),
    [sym__lowercase_identifier] = ACTIONS(2043),
    [anon_sym_SEMI_SEMI] = ACTIONS(2045),
    [anon_sym_let] = ACTIONS(2043),
    [anon_sym_and] = ACTIONS(2043),
    [anon_sym_TILDE] = ACTIONS(2043),
    [anon_sym_QMARK] = ACTIONS(2043),
    [anon_sym_LPAREN] = ACTIONS(2045),
    [anon_sym_external] = ACTIONS(2043),
    [anon_sym_type] = ACTIONS(2043),
    [anon_sym_COMMA] = ACTIONS(2045),
    [anon_sym_PIPE] = ACTIONS(2043),
    [anon_sym_DOT] = ACTIONS(2045),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_true] = ACTIONS(2043),
    [anon_sym_false] = ACTIONS(2043),
    [anon_sym_LBRACE] = ACTIONS(2043),
    [anon_sym_SEMI] = ACTIONS(2043),
    [anon_sym_exception] = ACTIONS(2043),
    [anon_sym_module] = ACTIONS(2043),
    [anon_sym_open] = ACTIONS(2043),
    [anon_sym_include] = ACTIONS(2043),
    [anon_sym_class] = ACTIONS(2043),
    [anon_sym_val] = ACTIONS(2043),
    [anon_sym_object] = ACTIONS(2043),
    [anon_sym_AMP] = ACTIONS(2043),
    [anon_sym_POUND] = ACTIONS(2043),
    [anon_sym_COLON_COLON] = ACTIONS(2045),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2045),
    [anon_sym_else] = ACTIONS(2043),
    [anon_sym_new] = ACTIONS(2043),
    [anon_sym_LBRACE_LT] = ACTIONS(2045),
    [anon_sym_begin] = ACTIONS(2043),
    [sym_ocamlyacc_value] = ACTIONS(2045),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2043),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2045),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2043),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2043),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2045),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2045),
    [aux_sym_number_token1] = ACTIONS(2045),
    [aux_sym_signed_number_token1] = ACTIONS(2043),
    [anon_sym_SQUOTE] = ACTIONS(2045),
    [anon_sym_DQUOTE] = ACTIONS(2045),
    [sym_prefix_operator] = ACTIONS(2045),
    [aux_sym_sign_operator_token1] = ACTIONS(2043),
    [sym_hash_operator] = ACTIONS(2045),
    [anon_sym_lsl] = ACTIONS(2043),
    [anon_sym_lsr] = ACTIONS(2043),
    [anon_sym_asr] = ACTIONS(2043),
    [aux_sym_pow_operator_token1] = ACTIONS(2045),
    [anon_sym_mod] = ACTIONS(2043),
    [anon_sym_land] = ACTIONS(2043),
    [anon_sym_lor] = ACTIONS(2043),
    [anon_sym_lxor] = ACTIONS(2043),
    [aux_sym_mult_operator_token1] = ACTIONS(2043),
    [aux_sym_add_operator_token1] = ACTIONS(2043),
    [sym_concat_operator] = ACTIONS(2045),
    [sym_rel_operator] = ACTIONS(2043),
    [anon_sym_AMP_AMP] = ACTIONS(2043),
    [anon_sym_or] = ACTIONS(2043),
    [anon_sym_PIPE_PIPE] = ACTIONS(2043),
    [sym_assign_operator] = ACTIONS(2045),
    [sym_let_operator] = ACTIONS(2045),
    [sym__uppercase_identifier] = ACTIONS(2045),
    [aux_sym_directive_token1] = ACTIONS(2043),
    [aux_sym_tag_token1] = ACTIONS(2045),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1444)] = {
    [sym_attribute] = STATE(1444),
    [ts_builtin_sym_end] = ACTIONS(2001),
    [sym__lowercase_identifier] = ACTIONS(1999),
    [anon_sym_SEMI_SEMI] = ACTIONS(2001),
    [anon_sym_let] = ACTIONS(1999),
    [anon_sym_and] = ACTIONS(1999),
    [anon_sym_TILDE] = ACTIONS(1999),
    [anon_sym_QMARK] = ACTIONS(1999),
    [anon_sym_LPAREN] = ACTIONS(2001),
    [anon_sym_external] = ACTIONS(1999),
    [anon_sym_type] = ACTIONS(1999),
    [anon_sym_COMMA] = ACTIONS(2001),
    [anon_sym_PIPE] = ACTIONS(1999),
    [anon_sym_DOT] = ACTIONS(2001),
    [anon_sym_LBRACK] = ACTIONS(1999),
    [anon_sym_true] = ACTIONS(1999),
    [anon_sym_false] = ACTIONS(1999),
    [anon_sym_LBRACE] = ACTIONS(1999),
    [anon_sym_SEMI] = ACTIONS(1999),
    [anon_sym_exception] = ACTIONS(1999),
    [anon_sym_module] = ACTIONS(1999),
    [anon_sym_open] = ACTIONS(1999),
    [anon_sym_include] = ACTIONS(1999),
    [anon_sym_class] = ACTIONS(1999),
    [anon_sym_val] = ACTIONS(1999),
    [anon_sym_object] = ACTIONS(1999),
    [anon_sym_AMP] = ACTIONS(1999),
    [anon_sym_POUND] = ACTIONS(1999),
    [anon_sym_COLON_COLON] = ACTIONS(2001),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2001),
    [anon_sym_else] = ACTIONS(1999),
    [anon_sym_new] = ACTIONS(1999),
    [anon_sym_LBRACE_LT] = ACTIONS(2001),
    [anon_sym_begin] = ACTIONS(1999),
    [sym_ocamlyacc_value] = ACTIONS(2001),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2001),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1999),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1999),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2001),
    [aux_sym_number_token1] = ACTIONS(2001),
    [aux_sym_signed_number_token1] = ACTIONS(1999),
    [anon_sym_SQUOTE] = ACTIONS(2001),
    [anon_sym_DQUOTE] = ACTIONS(2001),
    [sym_prefix_operator] = ACTIONS(2001),
    [aux_sym_sign_operator_token1] = ACTIONS(1999),
    [sym_hash_operator] = ACTIONS(2001),
    [anon_sym_lsl] = ACTIONS(1999),
    [anon_sym_lsr] = ACTIONS(1999),
    [anon_sym_asr] = ACTIONS(1999),
    [aux_sym_pow_operator_token1] = ACTIONS(2001),
    [anon_sym_mod] = ACTIONS(1999),
    [anon_sym_land] = ACTIONS(1999),
    [anon_sym_lor] = ACTIONS(1999),
    [anon_sym_lxor] = ACTIONS(1999),
    [aux_sym_mult_operator_token1] = ACTIONS(1999),
    [aux_sym_add_operator_token1] = ACTIONS(1999),
    [sym_concat_operator] = ACTIONS(2001),
    [sym_rel_operator] = ACTIONS(1999),
    [anon_sym_AMP_AMP] = ACTIONS(1999),
    [anon_sym_or] = ACTIONS(1999),
    [anon_sym_PIPE_PIPE] = ACTIONS(1999),
    [sym_assign_operator] = ACTIONS(2001),
    [sym_let_operator] = ACTIONS(2001),
    [sym__uppercase_identifier] = ACTIONS(2001),
    [aux_sym_directive_token1] = ACTIONS(1999),
    [aux_sym_tag_token1] = ACTIONS(2001),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1445)] = {
    [sym_attribute] = STATE(1445),
    [ts_builtin_sym_end] = ACTIONS(2057),
    [sym__lowercase_identifier] = ACTIONS(2055),
    [anon_sym_SEMI_SEMI] = ACTIONS(2057),
    [anon_sym_let] = ACTIONS(2055),
    [anon_sym_and] = ACTIONS(2055),
    [anon_sym_TILDE] = ACTIONS(2055),
    [anon_sym_QMARK] = ACTIONS(2055),
    [anon_sym_LPAREN] = ACTIONS(2057),
    [anon_sym_external] = ACTIONS(2055),
    [anon_sym_type] = ACTIONS(2055),
    [anon_sym_COMMA] = ACTIONS(2057),
    [anon_sym_PIPE] = ACTIONS(2055),
    [anon_sym_DOT] = ACTIONS(2057),
    [anon_sym_LBRACK] = ACTIONS(2055),
    [anon_sym_true] = ACTIONS(2055),
    [anon_sym_false] = ACTIONS(2055),
    [anon_sym_LBRACE] = ACTIONS(2055),
    [anon_sym_SEMI] = ACTIONS(2055),
    [anon_sym_exception] = ACTIONS(2055),
    [anon_sym_module] = ACTIONS(2055),
    [anon_sym_open] = ACTIONS(2055),
    [anon_sym_include] = ACTIONS(2055),
    [anon_sym_class] = ACTIONS(2055),
    [anon_sym_val] = ACTIONS(2055),
    [anon_sym_object] = ACTIONS(2055),
    [anon_sym_AMP] = ACTIONS(2055),
    [anon_sym_POUND] = ACTIONS(2055),
    [anon_sym_COLON_COLON] = ACTIONS(2057),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2057),
    [anon_sym_else] = ACTIONS(2055),
    [anon_sym_new] = ACTIONS(2055),
    [anon_sym_LBRACE_LT] = ACTIONS(2057),
    [anon_sym_begin] = ACTIONS(2055),
    [sym_ocamlyacc_value] = ACTIONS(2057),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2057),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2055),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2055),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2057),
    [aux_sym_number_token1] = ACTIONS(2057),
    [aux_sym_signed_number_token1] = ACTIONS(2055),
    [anon_sym_SQUOTE] = ACTIONS(2057),
    [anon_sym_DQUOTE] = ACTIONS(2057),
    [sym_prefix_operator] = ACTIONS(2057),
    [aux_sym_sign_operator_token1] = ACTIONS(2055),
    [sym_hash_operator] = ACTIONS(2057),
    [anon_sym_lsl] = ACTIONS(2055),
    [anon_sym_lsr] = ACTIONS(2055),
    [anon_sym_asr] = ACTIONS(2055),
    [aux_sym_pow_operator_token1] = ACTIONS(2057),
    [anon_sym_mod] = ACTIONS(2055),
    [anon_sym_land] = ACTIONS(2055),
    [anon_sym_lor] = ACTIONS(2055),
    [anon_sym_lxor] = ACTIONS(2055),
    [aux_sym_mult_operator_token1] = ACTIONS(2055),
    [aux_sym_add_operator_token1] = ACTIONS(2055),
    [sym_concat_operator] = ACTIONS(2057),
    [sym_rel_operator] = ACTIONS(2055),
    [anon_sym_AMP_AMP] = ACTIONS(2055),
    [anon_sym_or] = ACTIONS(2055),
    [anon_sym_PIPE_PIPE] = ACTIONS(2055),
    [sym_assign_operator] = ACTIONS(2057),
    [sym_let_operator] = ACTIONS(2057),
    [sym__uppercase_identifier] = ACTIONS(2057),
    [aux_sym_directive_token1] = ACTIONS(2055),
    [aux_sym_tag_token1] = ACTIONS(2057),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1446)] = {
    [sym_attribute] = STATE(1446),
    [ts_builtin_sym_end] = ACTIONS(2159),
    [sym__lowercase_identifier] = ACTIONS(2157),
    [anon_sym_SEMI_SEMI] = ACTIONS(2159),
    [anon_sym_let] = ACTIONS(2157),
    [anon_sym_and] = ACTIONS(2157),
    [anon_sym_TILDE] = ACTIONS(2157),
    [anon_sym_QMARK] = ACTIONS(2157),
    [anon_sym_LPAREN] = ACTIONS(2159),
    [anon_sym_external] = ACTIONS(2157),
    [anon_sym_type] = ACTIONS(2157),
    [anon_sym_COMMA] = ACTIONS(2159),
    [anon_sym_PIPE] = ACTIONS(2157),
    [anon_sym_DOT] = ACTIONS(2159),
    [anon_sym_LBRACK] = ACTIONS(2157),
    [anon_sym_true] = ACTIONS(2157),
    [anon_sym_false] = ACTIONS(2157),
    [anon_sym_LBRACE] = ACTIONS(2157),
    [anon_sym_SEMI] = ACTIONS(2157),
    [anon_sym_exception] = ACTIONS(2157),
    [anon_sym_module] = ACTIONS(2157),
    [anon_sym_open] = ACTIONS(2157),
    [anon_sym_include] = ACTIONS(2157),
    [anon_sym_class] = ACTIONS(2157),
    [anon_sym_val] = ACTIONS(2157),
    [anon_sym_object] = ACTIONS(2157),
    [anon_sym_AMP] = ACTIONS(2157),
    [anon_sym_POUND] = ACTIONS(2157),
    [anon_sym_COLON_COLON] = ACTIONS(2159),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2159),
    [anon_sym_else] = ACTIONS(2157),
    [anon_sym_new] = ACTIONS(2157),
    [anon_sym_LBRACE_LT] = ACTIONS(2159),
    [anon_sym_begin] = ACTIONS(2157),
    [sym_ocamlyacc_value] = ACTIONS(2159),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2157),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2159),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2157),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2159),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2159),
    [aux_sym_number_token1] = ACTIONS(2159),
    [aux_sym_signed_number_token1] = ACTIONS(2157),
    [anon_sym_SQUOTE] = ACTIONS(2159),
    [anon_sym_DQUOTE] = ACTIONS(2159),
    [sym_prefix_operator] = ACTIONS(2159),
    [aux_sym_sign_operator_token1] = ACTIONS(2157),
    [sym_hash_operator] = ACTIONS(2159),
    [anon_sym_lsl] = ACTIONS(2157),
    [anon_sym_lsr] = ACTIONS(2157),
    [anon_sym_asr] = ACTIONS(2157),
    [aux_sym_pow_operator_token1] = ACTIONS(2159),
    [anon_sym_mod] = ACTIONS(2157),
    [anon_sym_land] = ACTIONS(2157),
    [anon_sym_lor] = ACTIONS(2157),
    [anon_sym_lxor] = ACTIONS(2157),
    [aux_sym_mult_operator_token1] = ACTIONS(2157),
    [aux_sym_add_operator_token1] = ACTIONS(2157),
    [sym_concat_operator] = ACTIONS(2159),
    [sym_rel_operator] = ACTIONS(2157),
    [anon_sym_AMP_AMP] = ACTIONS(2157),
    [anon_sym_or] = ACTIONS(2157),
    [anon_sym_PIPE_PIPE] = ACTIONS(2157),
    [sym_assign_operator] = ACTIONS(2159),
    [sym_let_operator] = ACTIONS(2159),
    [sym__uppercase_identifier] = ACTIONS(2159),
    [aux_sym_directive_token1] = ACTIONS(2157),
    [aux_sym_tag_token1] = ACTIONS(2159),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1447)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5291),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1447),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(1641),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_PIPE] = ACTIONS(1641),
    [anon_sym_COLON2] = ACTIONS(1645),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_as] = ACTIONS(1645),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1641),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1448)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5626),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1448),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_PIPE] = ACTIONS(1645),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_SEMI] = ACTIONS(1641),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_as] = ACTIONS(1645),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(1641),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1641),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1449)] = {
    [sym_attribute] = STATE(1449),
    [ts_builtin_sym_end] = ACTIONS(2171),
    [sym__lowercase_identifier] = ACTIONS(2169),
    [anon_sym_SEMI_SEMI] = ACTIONS(2171),
    [anon_sym_let] = ACTIONS(2169),
    [anon_sym_and] = ACTIONS(2169),
    [anon_sym_TILDE] = ACTIONS(2169),
    [anon_sym_QMARK] = ACTIONS(2169),
    [anon_sym_LPAREN] = ACTIONS(2171),
    [anon_sym_external] = ACTIONS(2169),
    [anon_sym_type] = ACTIONS(2169),
    [anon_sym_COMMA] = ACTIONS(2171),
    [anon_sym_PIPE] = ACTIONS(2169),
    [anon_sym_DOT] = ACTIONS(2171),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_true] = ACTIONS(2169),
    [anon_sym_false] = ACTIONS(2169),
    [anon_sym_LBRACE] = ACTIONS(2169),
    [anon_sym_SEMI] = ACTIONS(2169),
    [anon_sym_exception] = ACTIONS(2169),
    [anon_sym_module] = ACTIONS(2169),
    [anon_sym_open] = ACTIONS(2169),
    [anon_sym_include] = ACTIONS(2169),
    [anon_sym_class] = ACTIONS(2169),
    [anon_sym_val] = ACTIONS(2169),
    [anon_sym_object] = ACTIONS(2169),
    [anon_sym_AMP] = ACTIONS(2169),
    [anon_sym_POUND] = ACTIONS(2169),
    [anon_sym_COLON_COLON] = ACTIONS(2171),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2171),
    [anon_sym_else] = ACTIONS(2169),
    [anon_sym_new] = ACTIONS(2169),
    [anon_sym_LBRACE_LT] = ACTIONS(2171),
    [anon_sym_begin] = ACTIONS(2169),
    [sym_ocamlyacc_value] = ACTIONS(2171),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2169),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2171),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2169),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2169),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2171),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2171),
    [aux_sym_number_token1] = ACTIONS(2171),
    [aux_sym_signed_number_token1] = ACTIONS(2169),
    [anon_sym_SQUOTE] = ACTIONS(2171),
    [anon_sym_DQUOTE] = ACTIONS(2171),
    [sym_prefix_operator] = ACTIONS(2171),
    [aux_sym_sign_operator_token1] = ACTIONS(2169),
    [sym_hash_operator] = ACTIONS(2171),
    [anon_sym_lsl] = ACTIONS(2169),
    [anon_sym_lsr] = ACTIONS(2169),
    [anon_sym_asr] = ACTIONS(2169),
    [aux_sym_pow_operator_token1] = ACTIONS(2171),
    [anon_sym_mod] = ACTIONS(2169),
    [anon_sym_land] = ACTIONS(2169),
    [anon_sym_lor] = ACTIONS(2169),
    [anon_sym_lxor] = ACTIONS(2169),
    [aux_sym_mult_operator_token1] = ACTIONS(2169),
    [aux_sym_add_operator_token1] = ACTIONS(2169),
    [sym_concat_operator] = ACTIONS(2171),
    [sym_rel_operator] = ACTIONS(2169),
    [anon_sym_AMP_AMP] = ACTIONS(2169),
    [anon_sym_or] = ACTIONS(2169),
    [anon_sym_PIPE_PIPE] = ACTIONS(2169),
    [sym_assign_operator] = ACTIONS(2171),
    [sym_let_operator] = ACTIONS(2171),
    [sym__uppercase_identifier] = ACTIONS(2171),
    [aux_sym_directive_token1] = ACTIONS(2169),
    [aux_sym_tag_token1] = ACTIONS(2171),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1450)] = {
    [sym_attribute] = STATE(1450),
    [ts_builtin_sym_end] = ACTIONS(2197),
    [sym__lowercase_identifier] = ACTIONS(2195),
    [anon_sym_SEMI_SEMI] = ACTIONS(2197),
    [anon_sym_let] = ACTIONS(2195),
    [anon_sym_and] = ACTIONS(2195),
    [anon_sym_TILDE] = ACTIONS(2195),
    [anon_sym_QMARK] = ACTIONS(2195),
    [anon_sym_LPAREN] = ACTIONS(2197),
    [anon_sym_external] = ACTIONS(2195),
    [anon_sym_type] = ACTIONS(2195),
    [anon_sym_COMMA] = ACTIONS(2197),
    [anon_sym_PIPE] = ACTIONS(2195),
    [anon_sym_DOT] = ACTIONS(2197),
    [anon_sym_LBRACK] = ACTIONS(2195),
    [anon_sym_true] = ACTIONS(2195),
    [anon_sym_false] = ACTIONS(2195),
    [anon_sym_LBRACE] = ACTIONS(2195),
    [anon_sym_SEMI] = ACTIONS(2195),
    [anon_sym_exception] = ACTIONS(2195),
    [anon_sym_module] = ACTIONS(2195),
    [anon_sym_open] = ACTIONS(2195),
    [anon_sym_include] = ACTIONS(2195),
    [anon_sym_class] = ACTIONS(2195),
    [anon_sym_val] = ACTIONS(2195),
    [anon_sym_object] = ACTIONS(2195),
    [anon_sym_AMP] = ACTIONS(2195),
    [anon_sym_POUND] = ACTIONS(2195),
    [anon_sym_COLON_COLON] = ACTIONS(2197),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2197),
    [anon_sym_else] = ACTIONS(2195),
    [anon_sym_new] = ACTIONS(2195),
    [anon_sym_LBRACE_LT] = ACTIONS(2197),
    [anon_sym_begin] = ACTIONS(2195),
    [sym_ocamlyacc_value] = ACTIONS(2197),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2195),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2197),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2195),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2195),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2197),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2197),
    [aux_sym_number_token1] = ACTIONS(2197),
    [aux_sym_signed_number_token1] = ACTIONS(2195),
    [anon_sym_SQUOTE] = ACTIONS(2197),
    [anon_sym_DQUOTE] = ACTIONS(2197),
    [sym_prefix_operator] = ACTIONS(2197),
    [aux_sym_sign_operator_token1] = ACTIONS(2195),
    [sym_hash_operator] = ACTIONS(2197),
    [anon_sym_lsl] = ACTIONS(2195),
    [anon_sym_lsr] = ACTIONS(2195),
    [anon_sym_asr] = ACTIONS(2195),
    [aux_sym_pow_operator_token1] = ACTIONS(2197),
    [anon_sym_mod] = ACTIONS(2195),
    [anon_sym_land] = ACTIONS(2195),
    [anon_sym_lor] = ACTIONS(2195),
    [anon_sym_lxor] = ACTIONS(2195),
    [aux_sym_mult_operator_token1] = ACTIONS(2195),
    [aux_sym_add_operator_token1] = ACTIONS(2195),
    [sym_concat_operator] = ACTIONS(2197),
    [sym_rel_operator] = ACTIONS(2195),
    [anon_sym_AMP_AMP] = ACTIONS(2195),
    [anon_sym_or] = ACTIONS(2195),
    [anon_sym_PIPE_PIPE] = ACTIONS(2195),
    [sym_assign_operator] = ACTIONS(2197),
    [sym_let_operator] = ACTIONS(2197),
    [sym__uppercase_identifier] = ACTIONS(2197),
    [aux_sym_directive_token1] = ACTIONS(2195),
    [aux_sym_tag_token1] = ACTIONS(2197),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1451)] = {
    [sym_attribute] = STATE(1451),
    [ts_builtin_sym_end] = ACTIONS(1937),
    [sym__lowercase_identifier] = ACTIONS(1935),
    [anon_sym_SEMI_SEMI] = ACTIONS(1937),
    [anon_sym_let] = ACTIONS(1935),
    [anon_sym_and] = ACTIONS(1935),
    [anon_sym_TILDE] = ACTIONS(1935),
    [anon_sym_QMARK] = ACTIONS(1935),
    [anon_sym_LPAREN] = ACTIONS(1937),
    [anon_sym_external] = ACTIONS(1935),
    [anon_sym_type] = ACTIONS(1935),
    [anon_sym_COMMA] = ACTIONS(1937),
    [anon_sym_PIPE] = ACTIONS(1935),
    [anon_sym_DOT] = ACTIONS(1937),
    [anon_sym_LBRACK] = ACTIONS(1935),
    [anon_sym_true] = ACTIONS(1935),
    [anon_sym_false] = ACTIONS(1935),
    [anon_sym_LBRACE] = ACTIONS(1935),
    [anon_sym_SEMI] = ACTIONS(1935),
    [anon_sym_exception] = ACTIONS(1935),
    [anon_sym_module] = ACTIONS(1935),
    [anon_sym_open] = ACTIONS(1935),
    [anon_sym_include] = ACTIONS(1935),
    [anon_sym_class] = ACTIONS(1935),
    [anon_sym_val] = ACTIONS(1935),
    [anon_sym_object] = ACTIONS(1935),
    [anon_sym_AMP] = ACTIONS(1935),
    [anon_sym_POUND] = ACTIONS(1935),
    [anon_sym_COLON_COLON] = ACTIONS(1937),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1937),
    [anon_sym_else] = ACTIONS(1935),
    [anon_sym_new] = ACTIONS(1935),
    [anon_sym_LBRACE_LT] = ACTIONS(1937),
    [anon_sym_begin] = ACTIONS(1935),
    [sym_ocamlyacc_value] = ACTIONS(1937),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1935),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1937),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1935),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1935),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1937),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1937),
    [aux_sym_number_token1] = ACTIONS(1937),
    [aux_sym_signed_number_token1] = ACTIONS(1935),
    [anon_sym_SQUOTE] = ACTIONS(1937),
    [anon_sym_DQUOTE] = ACTIONS(1937),
    [sym_prefix_operator] = ACTIONS(1937),
    [aux_sym_sign_operator_token1] = ACTIONS(1935),
    [sym_hash_operator] = ACTIONS(1937),
    [anon_sym_lsl] = ACTIONS(1935),
    [anon_sym_lsr] = ACTIONS(1935),
    [anon_sym_asr] = ACTIONS(1935),
    [aux_sym_pow_operator_token1] = ACTIONS(1937),
    [anon_sym_mod] = ACTIONS(1935),
    [anon_sym_land] = ACTIONS(1935),
    [anon_sym_lor] = ACTIONS(1935),
    [anon_sym_lxor] = ACTIONS(1935),
    [aux_sym_mult_operator_token1] = ACTIONS(1935),
    [aux_sym_add_operator_token1] = ACTIONS(1935),
    [sym_concat_operator] = ACTIONS(1937),
    [sym_rel_operator] = ACTIONS(1935),
    [anon_sym_AMP_AMP] = ACTIONS(1935),
    [anon_sym_or] = ACTIONS(1935),
    [anon_sym_PIPE_PIPE] = ACTIONS(1935),
    [sym_assign_operator] = ACTIONS(1937),
    [sym_let_operator] = ACTIONS(1937),
    [sym__uppercase_identifier] = ACTIONS(1937),
    [aux_sym_directive_token1] = ACTIONS(1935),
    [aux_sym_tag_token1] = ACTIONS(1937),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1452)] = {
    [sym_attribute] = STATE(1452),
    [ts_builtin_sym_end] = ACTIONS(1945),
    [sym__lowercase_identifier] = ACTIONS(1943),
    [anon_sym_SEMI_SEMI] = ACTIONS(1945),
    [anon_sym_let] = ACTIONS(1943),
    [anon_sym_and] = ACTIONS(1943),
    [anon_sym_TILDE] = ACTIONS(1943),
    [anon_sym_QMARK] = ACTIONS(1943),
    [anon_sym_LPAREN] = ACTIONS(1945),
    [anon_sym_external] = ACTIONS(1943),
    [anon_sym_type] = ACTIONS(1943),
    [anon_sym_COMMA] = ACTIONS(1945),
    [anon_sym_PIPE] = ACTIONS(1943),
    [anon_sym_DOT] = ACTIONS(1945),
    [anon_sym_LBRACK] = ACTIONS(1943),
    [anon_sym_true] = ACTIONS(1943),
    [anon_sym_false] = ACTIONS(1943),
    [anon_sym_LBRACE] = ACTIONS(1943),
    [anon_sym_SEMI] = ACTIONS(1943),
    [anon_sym_exception] = ACTIONS(1943),
    [anon_sym_module] = ACTIONS(1943),
    [anon_sym_open] = ACTIONS(1943),
    [anon_sym_include] = ACTIONS(1943),
    [anon_sym_class] = ACTIONS(1943),
    [anon_sym_val] = ACTIONS(1943),
    [anon_sym_object] = ACTIONS(1943),
    [anon_sym_AMP] = ACTIONS(1943),
    [anon_sym_POUND] = ACTIONS(1943),
    [anon_sym_COLON_COLON] = ACTIONS(1945),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1945),
    [anon_sym_else] = ACTIONS(1943),
    [anon_sym_new] = ACTIONS(1943),
    [anon_sym_LBRACE_LT] = ACTIONS(1945),
    [anon_sym_begin] = ACTIONS(1943),
    [sym_ocamlyacc_value] = ACTIONS(1945),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1943),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1945),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1943),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1943),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1945),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1945),
    [aux_sym_number_token1] = ACTIONS(1945),
    [aux_sym_signed_number_token1] = ACTIONS(1943),
    [anon_sym_SQUOTE] = ACTIONS(1945),
    [anon_sym_DQUOTE] = ACTIONS(1945),
    [sym_prefix_operator] = ACTIONS(1945),
    [aux_sym_sign_operator_token1] = ACTIONS(1943),
    [sym_hash_operator] = ACTIONS(1945),
    [anon_sym_lsl] = ACTIONS(1943),
    [anon_sym_lsr] = ACTIONS(1943),
    [anon_sym_asr] = ACTIONS(1943),
    [aux_sym_pow_operator_token1] = ACTIONS(1945),
    [anon_sym_mod] = ACTIONS(1943),
    [anon_sym_land] = ACTIONS(1943),
    [anon_sym_lor] = ACTIONS(1943),
    [anon_sym_lxor] = ACTIONS(1943),
    [aux_sym_mult_operator_token1] = ACTIONS(1943),
    [aux_sym_add_operator_token1] = ACTIONS(1943),
    [sym_concat_operator] = ACTIONS(1945),
    [sym_rel_operator] = ACTIONS(1943),
    [anon_sym_AMP_AMP] = ACTIONS(1943),
    [anon_sym_or] = ACTIONS(1943),
    [anon_sym_PIPE_PIPE] = ACTIONS(1943),
    [sym_assign_operator] = ACTIONS(1945),
    [sym_let_operator] = ACTIONS(1945),
    [sym__uppercase_identifier] = ACTIONS(1945),
    [aux_sym_directive_token1] = ACTIONS(1943),
    [aux_sym_tag_token1] = ACTIONS(1945),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1453)] = {
    [sym_attribute] = STATE(1453),
    [ts_builtin_sym_end] = ACTIONS(1941),
    [sym__lowercase_identifier] = ACTIONS(1939),
    [anon_sym_SEMI_SEMI] = ACTIONS(1941),
    [anon_sym_let] = ACTIONS(1939),
    [anon_sym_and] = ACTIONS(1939),
    [anon_sym_TILDE] = ACTIONS(1939),
    [anon_sym_QMARK] = ACTIONS(1939),
    [anon_sym_LPAREN] = ACTIONS(1941),
    [anon_sym_external] = ACTIONS(1939),
    [anon_sym_type] = ACTIONS(1939),
    [anon_sym_COMMA] = ACTIONS(1941),
    [anon_sym_PIPE] = ACTIONS(1939),
    [anon_sym_DOT] = ACTIONS(1941),
    [anon_sym_LBRACK] = ACTIONS(1939),
    [anon_sym_true] = ACTIONS(1939),
    [anon_sym_false] = ACTIONS(1939),
    [anon_sym_LBRACE] = ACTIONS(1939),
    [anon_sym_SEMI] = ACTIONS(1939),
    [anon_sym_exception] = ACTIONS(1939),
    [anon_sym_module] = ACTIONS(1939),
    [anon_sym_open] = ACTIONS(1939),
    [anon_sym_include] = ACTIONS(1939),
    [anon_sym_class] = ACTIONS(1939),
    [anon_sym_val] = ACTIONS(1939),
    [anon_sym_object] = ACTIONS(1939),
    [anon_sym_AMP] = ACTIONS(1939),
    [anon_sym_POUND] = ACTIONS(1939),
    [anon_sym_COLON_COLON] = ACTIONS(1941),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1941),
    [anon_sym_else] = ACTIONS(1939),
    [anon_sym_new] = ACTIONS(1939),
    [anon_sym_LBRACE_LT] = ACTIONS(1941),
    [anon_sym_begin] = ACTIONS(1939),
    [sym_ocamlyacc_value] = ACTIONS(1941),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1939),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1941),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1939),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1939),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1941),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1941),
    [aux_sym_number_token1] = ACTIONS(1941),
    [aux_sym_signed_number_token1] = ACTIONS(1939),
    [anon_sym_SQUOTE] = ACTIONS(1941),
    [anon_sym_DQUOTE] = ACTIONS(1941),
    [sym_prefix_operator] = ACTIONS(1941),
    [aux_sym_sign_operator_token1] = ACTIONS(1939),
    [sym_hash_operator] = ACTIONS(1941),
    [anon_sym_lsl] = ACTIONS(1939),
    [anon_sym_lsr] = ACTIONS(1939),
    [anon_sym_asr] = ACTIONS(1939),
    [aux_sym_pow_operator_token1] = ACTIONS(1941),
    [anon_sym_mod] = ACTIONS(1939),
    [anon_sym_land] = ACTIONS(1939),
    [anon_sym_lor] = ACTIONS(1939),
    [anon_sym_lxor] = ACTIONS(1939),
    [aux_sym_mult_operator_token1] = ACTIONS(1939),
    [aux_sym_add_operator_token1] = ACTIONS(1939),
    [sym_concat_operator] = ACTIONS(1941),
    [sym_rel_operator] = ACTIONS(1939),
    [anon_sym_AMP_AMP] = ACTIONS(1939),
    [anon_sym_or] = ACTIONS(1939),
    [anon_sym_PIPE_PIPE] = ACTIONS(1939),
    [sym_assign_operator] = ACTIONS(1941),
    [sym_let_operator] = ACTIONS(1941),
    [sym__uppercase_identifier] = ACTIONS(1941),
    [aux_sym_directive_token1] = ACTIONS(1939),
    [aux_sym_tag_token1] = ACTIONS(1941),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1454)] = {
    [sym_attribute] = STATE(1454),
    [ts_builtin_sym_end] = ACTIONS(2053),
    [sym__lowercase_identifier] = ACTIONS(2051),
    [anon_sym_SEMI_SEMI] = ACTIONS(2053),
    [anon_sym_let] = ACTIONS(2051),
    [anon_sym_and] = ACTIONS(2051),
    [anon_sym_TILDE] = ACTIONS(2051),
    [anon_sym_QMARK] = ACTIONS(2051),
    [anon_sym_LPAREN] = ACTIONS(2053),
    [anon_sym_external] = ACTIONS(2051),
    [anon_sym_type] = ACTIONS(2051),
    [anon_sym_COMMA] = ACTIONS(2053),
    [anon_sym_PIPE] = ACTIONS(2051),
    [anon_sym_DOT] = ACTIONS(2053),
    [anon_sym_LBRACK] = ACTIONS(2051),
    [anon_sym_true] = ACTIONS(2051),
    [anon_sym_false] = ACTIONS(2051),
    [anon_sym_LBRACE] = ACTIONS(2051),
    [anon_sym_SEMI] = ACTIONS(2051),
    [anon_sym_exception] = ACTIONS(2051),
    [anon_sym_module] = ACTIONS(2051),
    [anon_sym_open] = ACTIONS(2051),
    [anon_sym_include] = ACTIONS(2051),
    [anon_sym_class] = ACTIONS(2051),
    [anon_sym_val] = ACTIONS(2051),
    [anon_sym_object] = ACTIONS(2051),
    [anon_sym_AMP] = ACTIONS(2051),
    [anon_sym_POUND] = ACTIONS(2051),
    [anon_sym_COLON_COLON] = ACTIONS(2053),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2053),
    [anon_sym_else] = ACTIONS(2051),
    [anon_sym_new] = ACTIONS(2051),
    [anon_sym_LBRACE_LT] = ACTIONS(2053),
    [anon_sym_begin] = ACTIONS(2051),
    [sym_ocamlyacc_value] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2051),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2053),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2051),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2051),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2053),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2053),
    [aux_sym_number_token1] = ACTIONS(2053),
    [aux_sym_signed_number_token1] = ACTIONS(2051),
    [anon_sym_SQUOTE] = ACTIONS(2053),
    [anon_sym_DQUOTE] = ACTIONS(2053),
    [sym_prefix_operator] = ACTIONS(2053),
    [aux_sym_sign_operator_token1] = ACTIONS(2051),
    [sym_hash_operator] = ACTIONS(2053),
    [anon_sym_lsl] = ACTIONS(2051),
    [anon_sym_lsr] = ACTIONS(2051),
    [anon_sym_asr] = ACTIONS(2051),
    [aux_sym_pow_operator_token1] = ACTIONS(2053),
    [anon_sym_mod] = ACTIONS(2051),
    [anon_sym_land] = ACTIONS(2051),
    [anon_sym_lor] = ACTIONS(2051),
    [anon_sym_lxor] = ACTIONS(2051),
    [aux_sym_mult_operator_token1] = ACTIONS(2051),
    [aux_sym_add_operator_token1] = ACTIONS(2051),
    [sym_concat_operator] = ACTIONS(2053),
    [sym_rel_operator] = ACTIONS(2051),
    [anon_sym_AMP_AMP] = ACTIONS(2051),
    [anon_sym_or] = ACTIONS(2051),
    [anon_sym_PIPE_PIPE] = ACTIONS(2051),
    [sym_assign_operator] = ACTIONS(2053),
    [sym_let_operator] = ACTIONS(2053),
    [sym__uppercase_identifier] = ACTIONS(2053),
    [aux_sym_directive_token1] = ACTIONS(2051),
    [aux_sym_tag_token1] = ACTIONS(2053),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1455)] = {
    [sym_attribute] = STATE(1455),
    [ts_builtin_sym_end] = ACTIONS(2201),
    [sym__lowercase_identifier] = ACTIONS(2199),
    [anon_sym_SEMI_SEMI] = ACTIONS(2201),
    [anon_sym_let] = ACTIONS(2199),
    [anon_sym_and] = ACTIONS(2199),
    [anon_sym_TILDE] = ACTIONS(2199),
    [anon_sym_QMARK] = ACTIONS(2199),
    [anon_sym_LPAREN] = ACTIONS(2201),
    [anon_sym_external] = ACTIONS(2199),
    [anon_sym_type] = ACTIONS(2199),
    [anon_sym_COMMA] = ACTIONS(2201),
    [anon_sym_PIPE] = ACTIONS(2199),
    [anon_sym_DOT] = ACTIONS(2201),
    [anon_sym_LBRACK] = ACTIONS(2199),
    [anon_sym_true] = ACTIONS(2199),
    [anon_sym_false] = ACTIONS(2199),
    [anon_sym_LBRACE] = ACTIONS(2199),
    [anon_sym_SEMI] = ACTIONS(2199),
    [anon_sym_exception] = ACTIONS(2199),
    [anon_sym_module] = ACTIONS(2199),
    [anon_sym_open] = ACTIONS(2199),
    [anon_sym_include] = ACTIONS(2199),
    [anon_sym_class] = ACTIONS(2199),
    [anon_sym_val] = ACTIONS(2199),
    [anon_sym_object] = ACTIONS(2199),
    [anon_sym_AMP] = ACTIONS(2199),
    [anon_sym_POUND] = ACTIONS(2199),
    [anon_sym_COLON_COLON] = ACTIONS(2201),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2201),
    [anon_sym_else] = ACTIONS(2199),
    [anon_sym_new] = ACTIONS(2199),
    [anon_sym_LBRACE_LT] = ACTIONS(2201),
    [anon_sym_begin] = ACTIONS(2199),
    [sym_ocamlyacc_value] = ACTIONS(2201),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2199),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2201),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2199),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2199),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2201),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2201),
    [aux_sym_number_token1] = ACTIONS(2201),
    [aux_sym_signed_number_token1] = ACTIONS(2199),
    [anon_sym_SQUOTE] = ACTIONS(2201),
    [anon_sym_DQUOTE] = ACTIONS(2201),
    [sym_prefix_operator] = ACTIONS(2201),
    [aux_sym_sign_operator_token1] = ACTIONS(2199),
    [sym_hash_operator] = ACTIONS(2201),
    [anon_sym_lsl] = ACTIONS(2199),
    [anon_sym_lsr] = ACTIONS(2199),
    [anon_sym_asr] = ACTIONS(2199),
    [aux_sym_pow_operator_token1] = ACTIONS(2201),
    [anon_sym_mod] = ACTIONS(2199),
    [anon_sym_land] = ACTIONS(2199),
    [anon_sym_lor] = ACTIONS(2199),
    [anon_sym_lxor] = ACTIONS(2199),
    [aux_sym_mult_operator_token1] = ACTIONS(2199),
    [aux_sym_add_operator_token1] = ACTIONS(2199),
    [sym_concat_operator] = ACTIONS(2201),
    [sym_rel_operator] = ACTIONS(2199),
    [anon_sym_AMP_AMP] = ACTIONS(2199),
    [anon_sym_or] = ACTIONS(2199),
    [anon_sym_PIPE_PIPE] = ACTIONS(2199),
    [sym_assign_operator] = ACTIONS(2201),
    [sym_let_operator] = ACTIONS(2201),
    [sym__uppercase_identifier] = ACTIONS(2201),
    [aux_sym_directive_token1] = ACTIONS(2199),
    [aux_sym_tag_token1] = ACTIONS(2201),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1456)] = {
    [sym_attribute] = STATE(1456),
    [ts_builtin_sym_end] = ACTIONS(2085),
    [sym__lowercase_identifier] = ACTIONS(2083),
    [anon_sym_SEMI_SEMI] = ACTIONS(2085),
    [anon_sym_let] = ACTIONS(2083),
    [anon_sym_and] = ACTIONS(2083),
    [anon_sym_TILDE] = ACTIONS(2083),
    [anon_sym_QMARK] = ACTIONS(2083),
    [anon_sym_LPAREN] = ACTIONS(2085),
    [anon_sym_external] = ACTIONS(2083),
    [anon_sym_type] = ACTIONS(2083),
    [anon_sym_COMMA] = ACTIONS(2085),
    [anon_sym_PIPE] = ACTIONS(2083),
    [anon_sym_DOT] = ACTIONS(2085),
    [anon_sym_LBRACK] = ACTIONS(2083),
    [anon_sym_true] = ACTIONS(2083),
    [anon_sym_false] = ACTIONS(2083),
    [anon_sym_LBRACE] = ACTIONS(2083),
    [anon_sym_SEMI] = ACTIONS(2083),
    [anon_sym_exception] = ACTIONS(2083),
    [anon_sym_module] = ACTIONS(2083),
    [anon_sym_open] = ACTIONS(2083),
    [anon_sym_include] = ACTIONS(2083),
    [anon_sym_class] = ACTIONS(2083),
    [anon_sym_val] = ACTIONS(2083),
    [anon_sym_object] = ACTIONS(2083),
    [anon_sym_AMP] = ACTIONS(2083),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(2085),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2085),
    [anon_sym_else] = ACTIONS(2083),
    [anon_sym_new] = ACTIONS(2083),
    [anon_sym_LBRACE_LT] = ACTIONS(2085),
    [anon_sym_begin] = ACTIONS(2083),
    [sym_ocamlyacc_value] = ACTIONS(2085),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2083),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2085),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2083),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2083),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2085),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2085),
    [aux_sym_number_token1] = ACTIONS(2085),
    [aux_sym_signed_number_token1] = ACTIONS(2083),
    [anon_sym_SQUOTE] = ACTIONS(2085),
    [anon_sym_DQUOTE] = ACTIONS(2085),
    [sym_prefix_operator] = ACTIONS(2085),
    [aux_sym_sign_operator_token1] = ACTIONS(2083),
    [sym_hash_operator] = ACTIONS(2085),
    [anon_sym_lsl] = ACTIONS(2083),
    [anon_sym_lsr] = ACTIONS(2083),
    [anon_sym_asr] = ACTIONS(2083),
    [aux_sym_pow_operator_token1] = ACTIONS(2085),
    [anon_sym_mod] = ACTIONS(2083),
    [anon_sym_land] = ACTIONS(2083),
    [anon_sym_lor] = ACTIONS(2083),
    [anon_sym_lxor] = ACTIONS(2083),
    [aux_sym_mult_operator_token1] = ACTIONS(2083),
    [aux_sym_add_operator_token1] = ACTIONS(2083),
    [sym_concat_operator] = ACTIONS(2085),
    [sym_rel_operator] = ACTIONS(2083),
    [anon_sym_AMP_AMP] = ACTIONS(2083),
    [anon_sym_or] = ACTIONS(2083),
    [anon_sym_PIPE_PIPE] = ACTIONS(2083),
    [sym_assign_operator] = ACTIONS(2085),
    [sym_let_operator] = ACTIONS(2085),
    [sym__uppercase_identifier] = ACTIONS(2085),
    [aux_sym_directive_token1] = ACTIONS(2083),
    [aux_sym_tag_token1] = ACTIONS(2085),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1457)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5548),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1457),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_COMMA] = ACTIONS(2433),
    [anon_sym_PIPE] = ACTIONS(2439),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_SEMI] = ACTIONS(2433),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_as] = ACTIONS(2439),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_COLON_COLON] = ACTIONS(2433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2433),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1458)] = {
    [sym_attribute] = STATE(1458),
    [ts_builtin_sym_end] = ACTIONS(1973),
    [sym__lowercase_identifier] = ACTIONS(1971),
    [anon_sym_SEMI_SEMI] = ACTIONS(1973),
    [anon_sym_let] = ACTIONS(1971),
    [anon_sym_and] = ACTIONS(1971),
    [anon_sym_TILDE] = ACTIONS(1971),
    [anon_sym_QMARK] = ACTIONS(1971),
    [anon_sym_LPAREN] = ACTIONS(1973),
    [anon_sym_external] = ACTIONS(1971),
    [anon_sym_type] = ACTIONS(1971),
    [anon_sym_COMMA] = ACTIONS(1973),
    [anon_sym_PIPE] = ACTIONS(1971),
    [anon_sym_DOT] = ACTIONS(1157),
    [anon_sym_LBRACK] = ACTIONS(1971),
    [anon_sym_true] = ACTIONS(1971),
    [anon_sym_false] = ACTIONS(1971),
    [anon_sym_LBRACE] = ACTIONS(1971),
    [anon_sym_SEMI] = ACTIONS(1971),
    [anon_sym_exception] = ACTIONS(1971),
    [anon_sym_module] = ACTIONS(1971),
    [anon_sym_open] = ACTIONS(1971),
    [anon_sym_include] = ACTIONS(1971),
    [anon_sym_class] = ACTIONS(1971),
    [anon_sym_val] = ACTIONS(1971),
    [anon_sym_object] = ACTIONS(1971),
    [anon_sym_AMP] = ACTIONS(1971),
    [anon_sym_POUND] = ACTIONS(1971),
    [anon_sym_COLON_COLON] = ACTIONS(1973),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1973),
    [anon_sym_else] = ACTIONS(1971),
    [anon_sym_new] = ACTIONS(1971),
    [anon_sym_LBRACE_LT] = ACTIONS(1973),
    [anon_sym_begin] = ACTIONS(1971),
    [sym_ocamlyacc_value] = ACTIONS(1973),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1971),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1973),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1973),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1973),
    [aux_sym_number_token1] = ACTIONS(1973),
    [aux_sym_signed_number_token1] = ACTIONS(1971),
    [anon_sym_SQUOTE] = ACTIONS(1973),
    [anon_sym_DQUOTE] = ACTIONS(1973),
    [sym_prefix_operator] = ACTIONS(1973),
    [aux_sym_sign_operator_token1] = ACTIONS(1971),
    [sym_hash_operator] = ACTIONS(1973),
    [anon_sym_lsl] = ACTIONS(1971),
    [anon_sym_lsr] = ACTIONS(1971),
    [anon_sym_asr] = ACTIONS(1971),
    [aux_sym_pow_operator_token1] = ACTIONS(1973),
    [anon_sym_mod] = ACTIONS(1971),
    [anon_sym_land] = ACTIONS(1971),
    [anon_sym_lor] = ACTIONS(1971),
    [anon_sym_lxor] = ACTIONS(1971),
    [aux_sym_mult_operator_token1] = ACTIONS(1971),
    [aux_sym_add_operator_token1] = ACTIONS(1971),
    [sym_concat_operator] = ACTIONS(1973),
    [sym_rel_operator] = ACTIONS(1971),
    [anon_sym_AMP_AMP] = ACTIONS(1971),
    [anon_sym_or] = ACTIONS(1971),
    [anon_sym_PIPE_PIPE] = ACTIONS(1971),
    [sym_assign_operator] = ACTIONS(1973),
    [sym_let_operator] = ACTIONS(1973),
    [sym__uppercase_identifier] = ACTIONS(1973),
    [aux_sym_directive_token1] = ACTIONS(1971),
    [aux_sym_tag_token1] = ACTIONS(1973),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1459)] = {
    [sym_attribute] = STATE(1459),
    [ts_builtin_sym_end] = ACTIONS(1957),
    [sym__lowercase_identifier] = ACTIONS(1955),
    [anon_sym_SEMI_SEMI] = ACTIONS(1957),
    [anon_sym_let] = ACTIONS(1955),
    [anon_sym_and] = ACTIONS(1955),
    [anon_sym_TILDE] = ACTIONS(1955),
    [anon_sym_QMARK] = ACTIONS(1955),
    [anon_sym_LPAREN] = ACTIONS(1957),
    [anon_sym_external] = ACTIONS(1955),
    [anon_sym_type] = ACTIONS(1955),
    [anon_sym_COMMA] = ACTIONS(1957),
    [anon_sym_PIPE] = ACTIONS(1955),
    [anon_sym_DOT] = ACTIONS(1957),
    [anon_sym_LBRACK] = ACTIONS(1955),
    [anon_sym_true] = ACTIONS(1955),
    [anon_sym_false] = ACTIONS(1955),
    [anon_sym_LBRACE] = ACTIONS(1955),
    [anon_sym_SEMI] = ACTIONS(1955),
    [anon_sym_exception] = ACTIONS(1955),
    [anon_sym_module] = ACTIONS(1955),
    [anon_sym_open] = ACTIONS(1955),
    [anon_sym_include] = ACTIONS(1955),
    [anon_sym_class] = ACTIONS(1955),
    [anon_sym_val] = ACTIONS(1955),
    [anon_sym_object] = ACTIONS(1955),
    [anon_sym_AMP] = ACTIONS(1955),
    [anon_sym_POUND] = ACTIONS(1955),
    [anon_sym_COLON_COLON] = ACTIONS(1957),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1957),
    [anon_sym_else] = ACTIONS(1955),
    [anon_sym_new] = ACTIONS(1955),
    [anon_sym_LBRACE_LT] = ACTIONS(1957),
    [anon_sym_begin] = ACTIONS(1955),
    [sym_ocamlyacc_value] = ACTIONS(1957),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1955),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1957),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1955),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1955),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1957),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1957),
    [aux_sym_number_token1] = ACTIONS(1957),
    [aux_sym_signed_number_token1] = ACTIONS(1955),
    [anon_sym_SQUOTE] = ACTIONS(1957),
    [anon_sym_DQUOTE] = ACTIONS(1957),
    [sym_prefix_operator] = ACTIONS(1957),
    [aux_sym_sign_operator_token1] = ACTIONS(1955),
    [sym_hash_operator] = ACTIONS(1957),
    [anon_sym_lsl] = ACTIONS(1955),
    [anon_sym_lsr] = ACTIONS(1955),
    [anon_sym_asr] = ACTIONS(1955),
    [aux_sym_pow_operator_token1] = ACTIONS(1957),
    [anon_sym_mod] = ACTIONS(1955),
    [anon_sym_land] = ACTIONS(1955),
    [anon_sym_lor] = ACTIONS(1955),
    [anon_sym_lxor] = ACTIONS(1955),
    [aux_sym_mult_operator_token1] = ACTIONS(1955),
    [aux_sym_add_operator_token1] = ACTIONS(1955),
    [sym_concat_operator] = ACTIONS(1957),
    [sym_rel_operator] = ACTIONS(1955),
    [anon_sym_AMP_AMP] = ACTIONS(1955),
    [anon_sym_or] = ACTIONS(1955),
    [anon_sym_PIPE_PIPE] = ACTIONS(1955),
    [sym_assign_operator] = ACTIONS(1957),
    [sym_let_operator] = ACTIONS(1957),
    [sym__uppercase_identifier] = ACTIONS(1957),
    [aux_sym_directive_token1] = ACTIONS(1955),
    [aux_sym_tag_token1] = ACTIONS(1957),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1460)] = {
    [sym_attribute] = STATE(1460),
    [ts_builtin_sym_end] = ACTIONS(2061),
    [sym__lowercase_identifier] = ACTIONS(2059),
    [anon_sym_SEMI_SEMI] = ACTIONS(2061),
    [anon_sym_let] = ACTIONS(2059),
    [anon_sym_and] = ACTIONS(2059),
    [anon_sym_TILDE] = ACTIONS(2059),
    [anon_sym_QMARK] = ACTIONS(2059),
    [anon_sym_LPAREN] = ACTIONS(2061),
    [anon_sym_external] = ACTIONS(2059),
    [anon_sym_type] = ACTIONS(2059),
    [anon_sym_COMMA] = ACTIONS(2061),
    [anon_sym_PIPE] = ACTIONS(2059),
    [anon_sym_DOT] = ACTIONS(2061),
    [anon_sym_LBRACK] = ACTIONS(2059),
    [anon_sym_true] = ACTIONS(2059),
    [anon_sym_false] = ACTIONS(2059),
    [anon_sym_LBRACE] = ACTIONS(2059),
    [anon_sym_SEMI] = ACTIONS(2059),
    [anon_sym_exception] = ACTIONS(2059),
    [anon_sym_module] = ACTIONS(2059),
    [anon_sym_open] = ACTIONS(2059),
    [anon_sym_include] = ACTIONS(2059),
    [anon_sym_class] = ACTIONS(2059),
    [anon_sym_val] = ACTIONS(2059),
    [anon_sym_object] = ACTIONS(2059),
    [anon_sym_AMP] = ACTIONS(2059),
    [anon_sym_POUND] = ACTIONS(2059),
    [anon_sym_COLON_COLON] = ACTIONS(2061),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2061),
    [anon_sym_else] = ACTIONS(2059),
    [anon_sym_new] = ACTIONS(2059),
    [anon_sym_LBRACE_LT] = ACTIONS(2061),
    [anon_sym_begin] = ACTIONS(2059),
    [sym_ocamlyacc_value] = ACTIONS(2061),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2059),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2061),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2061),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2061),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2059),
    [anon_sym_SQUOTE] = ACTIONS(2061),
    [anon_sym_DQUOTE] = ACTIONS(2061),
    [sym_prefix_operator] = ACTIONS(2061),
    [aux_sym_sign_operator_token1] = ACTIONS(2059),
    [sym_hash_operator] = ACTIONS(2061),
    [anon_sym_lsl] = ACTIONS(2059),
    [anon_sym_lsr] = ACTIONS(2059),
    [anon_sym_asr] = ACTIONS(2059),
    [aux_sym_pow_operator_token1] = ACTIONS(2061),
    [anon_sym_mod] = ACTIONS(2059),
    [anon_sym_land] = ACTIONS(2059),
    [anon_sym_lor] = ACTIONS(2059),
    [anon_sym_lxor] = ACTIONS(2059),
    [aux_sym_mult_operator_token1] = ACTIONS(2059),
    [aux_sym_add_operator_token1] = ACTIONS(2059),
    [sym_concat_operator] = ACTIONS(2061),
    [sym_rel_operator] = ACTIONS(2059),
    [anon_sym_AMP_AMP] = ACTIONS(2059),
    [anon_sym_or] = ACTIONS(2059),
    [anon_sym_PIPE_PIPE] = ACTIONS(2059),
    [sym_assign_operator] = ACTIONS(2061),
    [sym_let_operator] = ACTIONS(2061),
    [sym__uppercase_identifier] = ACTIONS(2061),
    [aux_sym_directive_token1] = ACTIONS(2059),
    [aux_sym_tag_token1] = ACTIONS(2061),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1461)] = {
    [sym_attribute] = STATE(1461),
    [ts_builtin_sym_end] = ACTIONS(2113),
    [sym__lowercase_identifier] = ACTIONS(2111),
    [anon_sym_SEMI_SEMI] = ACTIONS(2113),
    [anon_sym_let] = ACTIONS(2111),
    [anon_sym_and] = ACTIONS(2111),
    [anon_sym_TILDE] = ACTIONS(2111),
    [anon_sym_QMARK] = ACTIONS(2111),
    [anon_sym_LPAREN] = ACTIONS(2113),
    [anon_sym_external] = ACTIONS(2111),
    [anon_sym_type] = ACTIONS(2111),
    [anon_sym_COMMA] = ACTIONS(2113),
    [anon_sym_PIPE] = ACTIONS(2111),
    [anon_sym_DOT] = ACTIONS(2177),
    [anon_sym_LBRACK] = ACTIONS(2111),
    [anon_sym_true] = ACTIONS(2111),
    [anon_sym_false] = ACTIONS(2111),
    [anon_sym_LBRACE] = ACTIONS(2111),
    [anon_sym_SEMI] = ACTIONS(2111),
    [anon_sym_exception] = ACTIONS(2111),
    [anon_sym_module] = ACTIONS(2111),
    [anon_sym_open] = ACTIONS(2111),
    [anon_sym_include] = ACTIONS(2111),
    [anon_sym_class] = ACTIONS(2111),
    [anon_sym_val] = ACTIONS(2111),
    [anon_sym_object] = ACTIONS(2111),
    [anon_sym_AMP] = ACTIONS(2111),
    [anon_sym_POUND] = ACTIONS(2111),
    [anon_sym_COLON_COLON] = ACTIONS(2113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2113),
    [anon_sym_else] = ACTIONS(2111),
    [anon_sym_new] = ACTIONS(2111),
    [anon_sym_LBRACE_LT] = ACTIONS(2113),
    [anon_sym_begin] = ACTIONS(2111),
    [sym_ocamlyacc_value] = ACTIONS(2113),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2111),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2113),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2113),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2113),
    [aux_sym_number_token1] = ACTIONS(2113),
    [aux_sym_signed_number_token1] = ACTIONS(2111),
    [anon_sym_SQUOTE] = ACTIONS(2113),
    [anon_sym_DQUOTE] = ACTIONS(2113),
    [sym_prefix_operator] = ACTIONS(2113),
    [aux_sym_sign_operator_token1] = ACTIONS(2111),
    [sym_hash_operator] = ACTIONS(2113),
    [anon_sym_lsl] = ACTIONS(2111),
    [anon_sym_lsr] = ACTIONS(2111),
    [anon_sym_asr] = ACTIONS(2111),
    [aux_sym_pow_operator_token1] = ACTIONS(2113),
    [anon_sym_mod] = ACTIONS(2111),
    [anon_sym_land] = ACTIONS(2111),
    [anon_sym_lor] = ACTIONS(2111),
    [anon_sym_lxor] = ACTIONS(2111),
    [aux_sym_mult_operator_token1] = ACTIONS(2111),
    [aux_sym_add_operator_token1] = ACTIONS(2111),
    [sym_concat_operator] = ACTIONS(2113),
    [sym_rel_operator] = ACTIONS(2111),
    [anon_sym_AMP_AMP] = ACTIONS(2111),
    [anon_sym_or] = ACTIONS(2111),
    [anon_sym_PIPE_PIPE] = ACTIONS(2111),
    [sym_assign_operator] = ACTIONS(2113),
    [sym_let_operator] = ACTIONS(2113),
    [sym__uppercase_identifier] = ACTIONS(2113),
    [aux_sym_directive_token1] = ACTIONS(2111),
    [aux_sym_tag_token1] = ACTIONS(2113),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1462)] = {
    [sym_attribute] = STATE(1462),
    [ts_builtin_sym_end] = ACTIONS(2049),
    [sym__lowercase_identifier] = ACTIONS(2047),
    [anon_sym_SEMI_SEMI] = ACTIONS(2049),
    [anon_sym_let] = ACTIONS(2047),
    [anon_sym_and] = ACTIONS(2047),
    [anon_sym_TILDE] = ACTIONS(2047),
    [anon_sym_QMARK] = ACTIONS(2047),
    [anon_sym_LPAREN] = ACTIONS(2049),
    [anon_sym_external] = ACTIONS(2047),
    [anon_sym_type] = ACTIONS(2047),
    [anon_sym_COMMA] = ACTIONS(2049),
    [anon_sym_PIPE] = ACTIONS(2047),
    [anon_sym_DOT] = ACTIONS(2049),
    [anon_sym_LBRACK] = ACTIONS(2047),
    [anon_sym_true] = ACTIONS(2047),
    [anon_sym_false] = ACTIONS(2047),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_SEMI] = ACTIONS(2047),
    [anon_sym_exception] = ACTIONS(2047),
    [anon_sym_module] = ACTIONS(2047),
    [anon_sym_open] = ACTIONS(2047),
    [anon_sym_include] = ACTIONS(2047),
    [anon_sym_class] = ACTIONS(2047),
    [anon_sym_val] = ACTIONS(2047),
    [anon_sym_object] = ACTIONS(2047),
    [anon_sym_AMP] = ACTIONS(2047),
    [anon_sym_POUND] = ACTIONS(2047),
    [anon_sym_COLON_COLON] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2049),
    [anon_sym_else] = ACTIONS(2047),
    [anon_sym_new] = ACTIONS(2047),
    [anon_sym_LBRACE_LT] = ACTIONS(2049),
    [anon_sym_begin] = ACTIONS(2047),
    [sym_ocamlyacc_value] = ACTIONS(2049),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2047),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2049),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2047),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2047),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2049),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2049),
    [aux_sym_number_token1] = ACTIONS(2049),
    [aux_sym_signed_number_token1] = ACTIONS(2047),
    [anon_sym_SQUOTE] = ACTIONS(2049),
    [anon_sym_DQUOTE] = ACTIONS(2049),
    [sym_prefix_operator] = ACTIONS(2049),
    [aux_sym_sign_operator_token1] = ACTIONS(2047),
    [sym_hash_operator] = ACTIONS(2049),
    [anon_sym_lsl] = ACTIONS(2047),
    [anon_sym_lsr] = ACTIONS(2047),
    [anon_sym_asr] = ACTIONS(2047),
    [aux_sym_pow_operator_token1] = ACTIONS(2049),
    [anon_sym_mod] = ACTIONS(2047),
    [anon_sym_land] = ACTIONS(2047),
    [anon_sym_lor] = ACTIONS(2047),
    [anon_sym_lxor] = ACTIONS(2047),
    [aux_sym_mult_operator_token1] = ACTIONS(2047),
    [aux_sym_add_operator_token1] = ACTIONS(2047),
    [sym_concat_operator] = ACTIONS(2049),
    [sym_rel_operator] = ACTIONS(2047),
    [anon_sym_AMP_AMP] = ACTIONS(2047),
    [anon_sym_or] = ACTIONS(2047),
    [anon_sym_PIPE_PIPE] = ACTIONS(2047),
    [sym_assign_operator] = ACTIONS(2049),
    [sym_let_operator] = ACTIONS(2049),
    [sym__uppercase_identifier] = ACTIONS(2049),
    [aux_sym_directive_token1] = ACTIONS(2047),
    [aux_sym_tag_token1] = ACTIONS(2049),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1463)] = {
    [sym_attribute] = STATE(1463),
    [ts_builtin_sym_end] = ACTIONS(1977),
    [sym__lowercase_identifier] = ACTIONS(1975),
    [anon_sym_SEMI_SEMI] = ACTIONS(1977),
    [anon_sym_let] = ACTIONS(1975),
    [anon_sym_and] = ACTIONS(1975),
    [anon_sym_TILDE] = ACTIONS(1975),
    [anon_sym_QMARK] = ACTIONS(1975),
    [anon_sym_LPAREN] = ACTIONS(1977),
    [anon_sym_external] = ACTIONS(1975),
    [anon_sym_type] = ACTIONS(1975),
    [anon_sym_COMMA] = ACTIONS(1977),
    [anon_sym_PIPE] = ACTIONS(1975),
    [anon_sym_DOT] = ACTIONS(1977),
    [anon_sym_LBRACK] = ACTIONS(1975),
    [anon_sym_true] = ACTIONS(1975),
    [anon_sym_false] = ACTIONS(1975),
    [anon_sym_LBRACE] = ACTIONS(1975),
    [anon_sym_SEMI] = ACTIONS(1975),
    [anon_sym_exception] = ACTIONS(1975),
    [anon_sym_module] = ACTIONS(1975),
    [anon_sym_open] = ACTIONS(1975),
    [anon_sym_include] = ACTIONS(1975),
    [anon_sym_class] = ACTIONS(1975),
    [anon_sym_val] = ACTIONS(1975),
    [anon_sym_object] = ACTIONS(1975),
    [anon_sym_AMP] = ACTIONS(1975),
    [anon_sym_POUND] = ACTIONS(1975),
    [anon_sym_COLON_COLON] = ACTIONS(1977),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1977),
    [anon_sym_else] = ACTIONS(1975),
    [anon_sym_new] = ACTIONS(1975),
    [anon_sym_LBRACE_LT] = ACTIONS(1977),
    [anon_sym_begin] = ACTIONS(1975),
    [sym_ocamlyacc_value] = ACTIONS(1977),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1975),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1977),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1975),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1975),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1977),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1977),
    [aux_sym_number_token1] = ACTIONS(1977),
    [aux_sym_signed_number_token1] = ACTIONS(1975),
    [anon_sym_SQUOTE] = ACTIONS(1977),
    [anon_sym_DQUOTE] = ACTIONS(1977),
    [sym_prefix_operator] = ACTIONS(1977),
    [aux_sym_sign_operator_token1] = ACTIONS(1975),
    [sym_hash_operator] = ACTIONS(1977),
    [anon_sym_lsl] = ACTIONS(1975),
    [anon_sym_lsr] = ACTIONS(1975),
    [anon_sym_asr] = ACTIONS(1975),
    [aux_sym_pow_operator_token1] = ACTIONS(1977),
    [anon_sym_mod] = ACTIONS(1975),
    [anon_sym_land] = ACTIONS(1975),
    [anon_sym_lor] = ACTIONS(1975),
    [anon_sym_lxor] = ACTIONS(1975),
    [aux_sym_mult_operator_token1] = ACTIONS(1975),
    [aux_sym_add_operator_token1] = ACTIONS(1975),
    [sym_concat_operator] = ACTIONS(1977),
    [sym_rel_operator] = ACTIONS(1975),
    [anon_sym_AMP_AMP] = ACTIONS(1975),
    [anon_sym_or] = ACTIONS(1975),
    [anon_sym_PIPE_PIPE] = ACTIONS(1975),
    [sym_assign_operator] = ACTIONS(1977),
    [sym_let_operator] = ACTIONS(1977),
    [sym__uppercase_identifier] = ACTIONS(1977),
    [aux_sym_directive_token1] = ACTIONS(1975),
    [aux_sym_tag_token1] = ACTIONS(1977),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1464)] = {
    [sym_attribute] = STATE(1464),
    [ts_builtin_sym_end] = ACTIONS(2005),
    [sym__lowercase_identifier] = ACTIONS(2003),
    [anon_sym_SEMI_SEMI] = ACTIONS(2005),
    [anon_sym_let] = ACTIONS(2003),
    [anon_sym_and] = ACTIONS(2003),
    [anon_sym_TILDE] = ACTIONS(2003),
    [anon_sym_QMARK] = ACTIONS(2003),
    [anon_sym_LPAREN] = ACTIONS(2005),
    [anon_sym_external] = ACTIONS(2003),
    [anon_sym_type] = ACTIONS(2003),
    [anon_sym_COMMA] = ACTIONS(2005),
    [anon_sym_PIPE] = ACTIONS(2003),
    [anon_sym_DOT] = ACTIONS(2005),
    [anon_sym_LBRACK] = ACTIONS(2003),
    [anon_sym_true] = ACTIONS(2003),
    [anon_sym_false] = ACTIONS(2003),
    [anon_sym_LBRACE] = ACTIONS(2003),
    [anon_sym_SEMI] = ACTIONS(2003),
    [anon_sym_exception] = ACTIONS(2003),
    [anon_sym_module] = ACTIONS(2003),
    [anon_sym_open] = ACTIONS(2003),
    [anon_sym_include] = ACTIONS(2003),
    [anon_sym_class] = ACTIONS(2003),
    [anon_sym_val] = ACTIONS(2003),
    [anon_sym_object] = ACTIONS(2003),
    [anon_sym_AMP] = ACTIONS(2003),
    [anon_sym_POUND] = ACTIONS(2003),
    [anon_sym_COLON_COLON] = ACTIONS(2005),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2005),
    [anon_sym_else] = ACTIONS(2003),
    [anon_sym_new] = ACTIONS(2003),
    [anon_sym_LBRACE_LT] = ACTIONS(2005),
    [anon_sym_begin] = ACTIONS(2003),
    [sym_ocamlyacc_value] = ACTIONS(2005),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2003),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2005),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2005),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2005),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2003),
    [anon_sym_SQUOTE] = ACTIONS(2005),
    [anon_sym_DQUOTE] = ACTIONS(2005),
    [sym_prefix_operator] = ACTIONS(2005),
    [aux_sym_sign_operator_token1] = ACTIONS(2003),
    [sym_hash_operator] = ACTIONS(2005),
    [anon_sym_lsl] = ACTIONS(2003),
    [anon_sym_lsr] = ACTIONS(2003),
    [anon_sym_asr] = ACTIONS(2003),
    [aux_sym_pow_operator_token1] = ACTIONS(2005),
    [anon_sym_mod] = ACTIONS(2003),
    [anon_sym_land] = ACTIONS(2003),
    [anon_sym_lor] = ACTIONS(2003),
    [anon_sym_lxor] = ACTIONS(2003),
    [aux_sym_mult_operator_token1] = ACTIONS(2003),
    [aux_sym_add_operator_token1] = ACTIONS(2003),
    [sym_concat_operator] = ACTIONS(2005),
    [sym_rel_operator] = ACTIONS(2003),
    [anon_sym_AMP_AMP] = ACTIONS(2003),
    [anon_sym_or] = ACTIONS(2003),
    [anon_sym_PIPE_PIPE] = ACTIONS(2003),
    [sym_assign_operator] = ACTIONS(2005),
    [sym_let_operator] = ACTIONS(2005),
    [sym__uppercase_identifier] = ACTIONS(2005),
    [aux_sym_directive_token1] = ACTIONS(2003),
    [aux_sym_tag_token1] = ACTIONS(2005),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1465)] = {
    [sym_attribute] = STATE(1465),
    [ts_builtin_sym_end] = ACTIONS(2009),
    [sym__lowercase_identifier] = ACTIONS(2007),
    [anon_sym_SEMI_SEMI] = ACTIONS(2009),
    [anon_sym_let] = ACTIONS(2007),
    [anon_sym_and] = ACTIONS(2007),
    [anon_sym_TILDE] = ACTIONS(2007),
    [anon_sym_QMARK] = ACTIONS(2007),
    [anon_sym_LPAREN] = ACTIONS(2009),
    [anon_sym_external] = ACTIONS(2007),
    [anon_sym_type] = ACTIONS(2007),
    [anon_sym_COMMA] = ACTIONS(2009),
    [anon_sym_PIPE] = ACTIONS(2007),
    [anon_sym_DOT] = ACTIONS(2009),
    [anon_sym_LBRACK] = ACTIONS(2007),
    [anon_sym_true] = ACTIONS(2007),
    [anon_sym_false] = ACTIONS(2007),
    [anon_sym_LBRACE] = ACTIONS(2007),
    [anon_sym_SEMI] = ACTIONS(2007),
    [anon_sym_exception] = ACTIONS(2007),
    [anon_sym_module] = ACTIONS(2007),
    [anon_sym_open] = ACTIONS(2007),
    [anon_sym_include] = ACTIONS(2007),
    [anon_sym_class] = ACTIONS(2007),
    [anon_sym_val] = ACTIONS(2007),
    [anon_sym_object] = ACTIONS(2007),
    [anon_sym_AMP] = ACTIONS(2007),
    [anon_sym_POUND] = ACTIONS(2007),
    [anon_sym_COLON_COLON] = ACTIONS(2009),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2009),
    [anon_sym_else] = ACTIONS(2007),
    [anon_sym_new] = ACTIONS(2007),
    [anon_sym_LBRACE_LT] = ACTIONS(2009),
    [anon_sym_begin] = ACTIONS(2007),
    [sym_ocamlyacc_value] = ACTIONS(2009),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2007),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2009),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2007),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2007),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2009),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2009),
    [aux_sym_number_token1] = ACTIONS(2009),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2009),
    [sym_prefix_operator] = ACTIONS(2009),
    [aux_sym_sign_operator_token1] = ACTIONS(2007),
    [sym_hash_operator] = ACTIONS(2009),
    [anon_sym_lsl] = ACTIONS(2007),
    [anon_sym_lsr] = ACTIONS(2007),
    [anon_sym_asr] = ACTIONS(2007),
    [aux_sym_pow_operator_token1] = ACTIONS(2009),
    [anon_sym_mod] = ACTIONS(2007),
    [anon_sym_land] = ACTIONS(2007),
    [anon_sym_lor] = ACTIONS(2007),
    [anon_sym_lxor] = ACTIONS(2007),
    [aux_sym_mult_operator_token1] = ACTIONS(2007),
    [aux_sym_add_operator_token1] = ACTIONS(2007),
    [sym_concat_operator] = ACTIONS(2009),
    [sym_rel_operator] = ACTIONS(2007),
    [anon_sym_AMP_AMP] = ACTIONS(2007),
    [anon_sym_or] = ACTIONS(2007),
    [anon_sym_PIPE_PIPE] = ACTIONS(2007),
    [sym_assign_operator] = ACTIONS(2009),
    [sym_let_operator] = ACTIONS(2009),
    [sym__uppercase_identifier] = ACTIONS(2009),
    [aux_sym_directive_token1] = ACTIONS(2007),
    [aux_sym_tag_token1] = ACTIONS(2009),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1466)] = {
    [sym_attribute] = STATE(1466),
    [ts_builtin_sym_end] = ACTIONS(2013),
    [sym__lowercase_identifier] = ACTIONS(2011),
    [anon_sym_SEMI_SEMI] = ACTIONS(2013),
    [anon_sym_let] = ACTIONS(2011),
    [anon_sym_and] = ACTIONS(2011),
    [anon_sym_TILDE] = ACTIONS(2011),
    [anon_sym_QMARK] = ACTIONS(2011),
    [anon_sym_LPAREN] = ACTIONS(2013),
    [anon_sym_external] = ACTIONS(2011),
    [anon_sym_type] = ACTIONS(2011),
    [anon_sym_COMMA] = ACTIONS(2013),
    [anon_sym_PIPE] = ACTIONS(2011),
    [anon_sym_DOT] = ACTIONS(2013),
    [anon_sym_LBRACK] = ACTIONS(2011),
    [anon_sym_true] = ACTIONS(2011),
    [anon_sym_false] = ACTIONS(2011),
    [anon_sym_LBRACE] = ACTIONS(2011),
    [anon_sym_SEMI] = ACTIONS(2011),
    [anon_sym_exception] = ACTIONS(2011),
    [anon_sym_module] = ACTIONS(2011),
    [anon_sym_open] = ACTIONS(2011),
    [anon_sym_include] = ACTIONS(2011),
    [anon_sym_class] = ACTIONS(2011),
    [anon_sym_val] = ACTIONS(2011),
    [anon_sym_object] = ACTIONS(2011),
    [anon_sym_AMP] = ACTIONS(2011),
    [anon_sym_POUND] = ACTIONS(2011),
    [anon_sym_COLON_COLON] = ACTIONS(2013),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2013),
    [anon_sym_else] = ACTIONS(2011),
    [anon_sym_new] = ACTIONS(2011),
    [anon_sym_LBRACE_LT] = ACTIONS(2013),
    [anon_sym_begin] = ACTIONS(2011),
    [sym_ocamlyacc_value] = ACTIONS(2013),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2011),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2013),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2011),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2011),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2013),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2013),
    [aux_sym_number_token1] = ACTIONS(2013),
    [aux_sym_signed_number_token1] = ACTIONS(2011),
    [anon_sym_SQUOTE] = ACTIONS(2013),
    [anon_sym_DQUOTE] = ACTIONS(2013),
    [sym_prefix_operator] = ACTIONS(2013),
    [aux_sym_sign_operator_token1] = ACTIONS(2011),
    [sym_hash_operator] = ACTIONS(2013),
    [anon_sym_lsl] = ACTIONS(2011),
    [anon_sym_lsr] = ACTIONS(2011),
    [anon_sym_asr] = ACTIONS(2011),
    [aux_sym_pow_operator_token1] = ACTIONS(2013),
    [anon_sym_mod] = ACTIONS(2011),
    [anon_sym_land] = ACTIONS(2011),
    [anon_sym_lor] = ACTIONS(2011),
    [anon_sym_lxor] = ACTIONS(2011),
    [aux_sym_mult_operator_token1] = ACTIONS(2011),
    [aux_sym_add_operator_token1] = ACTIONS(2011),
    [sym_concat_operator] = ACTIONS(2013),
    [sym_rel_operator] = ACTIONS(2011),
    [anon_sym_AMP_AMP] = ACTIONS(2011),
    [anon_sym_or] = ACTIONS(2011),
    [anon_sym_PIPE_PIPE] = ACTIONS(2011),
    [sym_assign_operator] = ACTIONS(2013),
    [sym_let_operator] = ACTIONS(2013),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_directive_token1] = ACTIONS(2011),
    [aux_sym_tag_token1] = ACTIONS(2013),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1467)] = {
    [sym_attribute] = STATE(1467),
    [ts_builtin_sym_end] = ACTIONS(2185),
    [sym__lowercase_identifier] = ACTIONS(2183),
    [anon_sym_SEMI_SEMI] = ACTIONS(2185),
    [anon_sym_let] = ACTIONS(2183),
    [anon_sym_and] = ACTIONS(2183),
    [anon_sym_TILDE] = ACTIONS(2183),
    [anon_sym_QMARK] = ACTIONS(2183),
    [anon_sym_LPAREN] = ACTIONS(2185),
    [anon_sym_external] = ACTIONS(2183),
    [anon_sym_type] = ACTIONS(2183),
    [anon_sym_COMMA] = ACTIONS(2185),
    [anon_sym_PIPE] = ACTIONS(2183),
    [anon_sym_DOT] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2183),
    [anon_sym_true] = ACTIONS(2183),
    [anon_sym_false] = ACTIONS(2183),
    [anon_sym_LBRACE] = ACTIONS(2183),
    [anon_sym_SEMI] = ACTIONS(2183),
    [anon_sym_exception] = ACTIONS(2183),
    [anon_sym_module] = ACTIONS(2183),
    [anon_sym_open] = ACTIONS(2183),
    [anon_sym_include] = ACTIONS(2183),
    [anon_sym_class] = ACTIONS(2183),
    [anon_sym_val] = ACTIONS(2183),
    [anon_sym_object] = ACTIONS(2183),
    [anon_sym_AMP] = ACTIONS(2183),
    [anon_sym_POUND] = ACTIONS(2183),
    [anon_sym_COLON_COLON] = ACTIONS(2185),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2185),
    [anon_sym_else] = ACTIONS(2183),
    [anon_sym_new] = ACTIONS(2183),
    [anon_sym_LBRACE_LT] = ACTIONS(2185),
    [anon_sym_begin] = ACTIONS(2183),
    [sym_ocamlyacc_value] = ACTIONS(2185),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2183),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2185),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2183),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2183),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2185),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2185),
    [aux_sym_number_token1] = ACTIONS(2185),
    [aux_sym_signed_number_token1] = ACTIONS(2183),
    [anon_sym_SQUOTE] = ACTIONS(2185),
    [anon_sym_DQUOTE] = ACTIONS(2185),
    [sym_prefix_operator] = ACTIONS(2185),
    [aux_sym_sign_operator_token1] = ACTIONS(2183),
    [sym_hash_operator] = ACTIONS(2185),
    [anon_sym_lsl] = ACTIONS(2183),
    [anon_sym_lsr] = ACTIONS(2183),
    [anon_sym_asr] = ACTIONS(2183),
    [aux_sym_pow_operator_token1] = ACTIONS(2185),
    [anon_sym_mod] = ACTIONS(2183),
    [anon_sym_land] = ACTIONS(2183),
    [anon_sym_lor] = ACTIONS(2183),
    [anon_sym_lxor] = ACTIONS(2183),
    [aux_sym_mult_operator_token1] = ACTIONS(2183),
    [aux_sym_add_operator_token1] = ACTIONS(2183),
    [sym_concat_operator] = ACTIONS(2185),
    [sym_rel_operator] = ACTIONS(2183),
    [anon_sym_AMP_AMP] = ACTIONS(2183),
    [anon_sym_or] = ACTIONS(2183),
    [anon_sym_PIPE_PIPE] = ACTIONS(2183),
    [sym_assign_operator] = ACTIONS(2185),
    [sym_let_operator] = ACTIONS(2185),
    [sym__uppercase_identifier] = ACTIONS(2185),
    [aux_sym_directive_token1] = ACTIONS(2183),
    [aux_sym_tag_token1] = ACTIONS(2185),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1468)] = {
    [sym_attribute] = STATE(1468),
    [ts_builtin_sym_end] = ACTIONS(2189),
    [sym__lowercase_identifier] = ACTIONS(2187),
    [anon_sym_SEMI_SEMI] = ACTIONS(2189),
    [anon_sym_let] = ACTIONS(2187),
    [anon_sym_and] = ACTIONS(2187),
    [anon_sym_TILDE] = ACTIONS(2187),
    [anon_sym_QMARK] = ACTIONS(2187),
    [anon_sym_LPAREN] = ACTIONS(2189),
    [anon_sym_external] = ACTIONS(2187),
    [anon_sym_type] = ACTIONS(2187),
    [anon_sym_COMMA] = ACTIONS(2189),
    [anon_sym_PIPE] = ACTIONS(2187),
    [anon_sym_DOT] = ACTIONS(2189),
    [anon_sym_LBRACK] = ACTIONS(2187),
    [anon_sym_true] = ACTIONS(2187),
    [anon_sym_false] = ACTIONS(2187),
    [anon_sym_LBRACE] = ACTIONS(2187),
    [anon_sym_SEMI] = ACTIONS(2187),
    [anon_sym_exception] = ACTIONS(2187),
    [anon_sym_module] = ACTIONS(2187),
    [anon_sym_open] = ACTIONS(2187),
    [anon_sym_include] = ACTIONS(2187),
    [anon_sym_class] = ACTIONS(2187),
    [anon_sym_val] = ACTIONS(2187),
    [anon_sym_object] = ACTIONS(2187),
    [anon_sym_AMP] = ACTIONS(2187),
    [anon_sym_POUND] = ACTIONS(2187),
    [anon_sym_COLON_COLON] = ACTIONS(2189),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2189),
    [anon_sym_else] = ACTIONS(2187),
    [anon_sym_new] = ACTIONS(2187),
    [anon_sym_LBRACE_LT] = ACTIONS(2189),
    [anon_sym_begin] = ACTIONS(2187),
    [sym_ocamlyacc_value] = ACTIONS(2189),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2187),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2189),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2187),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2187),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2189),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2189),
    [aux_sym_number_token1] = ACTIONS(2189),
    [aux_sym_signed_number_token1] = ACTIONS(2187),
    [anon_sym_SQUOTE] = ACTIONS(2189),
    [anon_sym_DQUOTE] = ACTIONS(2189),
    [sym_prefix_operator] = ACTIONS(2189),
    [aux_sym_sign_operator_token1] = ACTIONS(2187),
    [sym_hash_operator] = ACTIONS(2189),
    [anon_sym_lsl] = ACTIONS(2187),
    [anon_sym_lsr] = ACTIONS(2187),
    [anon_sym_asr] = ACTIONS(2187),
    [aux_sym_pow_operator_token1] = ACTIONS(2189),
    [anon_sym_mod] = ACTIONS(2187),
    [anon_sym_land] = ACTIONS(2187),
    [anon_sym_lor] = ACTIONS(2187),
    [anon_sym_lxor] = ACTIONS(2187),
    [aux_sym_mult_operator_token1] = ACTIONS(2187),
    [aux_sym_add_operator_token1] = ACTIONS(2187),
    [sym_concat_operator] = ACTIONS(2189),
    [sym_rel_operator] = ACTIONS(2187),
    [anon_sym_AMP_AMP] = ACTIONS(2187),
    [anon_sym_or] = ACTIONS(2187),
    [anon_sym_PIPE_PIPE] = ACTIONS(2187),
    [sym_assign_operator] = ACTIONS(2189),
    [sym_let_operator] = ACTIONS(2189),
    [sym__uppercase_identifier] = ACTIONS(2189),
    [aux_sym_directive_token1] = ACTIONS(2187),
    [aux_sym_tag_token1] = ACTIONS(2189),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1469)] = {
    [sym_attribute] = STATE(1469),
    [ts_builtin_sym_end] = ACTIONS(1981),
    [sym__lowercase_identifier] = ACTIONS(1979),
    [anon_sym_SEMI_SEMI] = ACTIONS(1981),
    [anon_sym_let] = ACTIONS(1979),
    [anon_sym_and] = ACTIONS(1979),
    [anon_sym_TILDE] = ACTIONS(1979),
    [anon_sym_QMARK] = ACTIONS(1979),
    [anon_sym_LPAREN] = ACTIONS(1981),
    [anon_sym_external] = ACTIONS(1979),
    [anon_sym_type] = ACTIONS(1979),
    [anon_sym_COMMA] = ACTIONS(1981),
    [anon_sym_PIPE] = ACTIONS(1979),
    [anon_sym_DOT] = ACTIONS(1981),
    [anon_sym_LBRACK] = ACTIONS(1979),
    [anon_sym_true] = ACTIONS(1979),
    [anon_sym_false] = ACTIONS(1979),
    [anon_sym_LBRACE] = ACTIONS(1979),
    [anon_sym_SEMI] = ACTIONS(1979),
    [anon_sym_exception] = ACTIONS(1979),
    [anon_sym_module] = ACTIONS(1979),
    [anon_sym_open] = ACTIONS(1979),
    [anon_sym_include] = ACTIONS(1979),
    [anon_sym_class] = ACTIONS(1979),
    [anon_sym_val] = ACTIONS(1979),
    [anon_sym_object] = ACTIONS(1979),
    [anon_sym_AMP] = ACTIONS(1979),
    [anon_sym_POUND] = ACTIONS(1979),
    [anon_sym_COLON_COLON] = ACTIONS(1981),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1981),
    [anon_sym_else] = ACTIONS(1979),
    [anon_sym_new] = ACTIONS(1979),
    [anon_sym_LBRACE_LT] = ACTIONS(1981),
    [anon_sym_begin] = ACTIONS(1979),
    [sym_ocamlyacc_value] = ACTIONS(1981),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1979),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1981),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1979),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1979),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1981),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1981),
    [aux_sym_number_token1] = ACTIONS(1981),
    [aux_sym_signed_number_token1] = ACTIONS(1979),
    [anon_sym_SQUOTE] = ACTIONS(1981),
    [anon_sym_DQUOTE] = ACTIONS(1981),
    [sym_prefix_operator] = ACTIONS(1981),
    [aux_sym_sign_operator_token1] = ACTIONS(1979),
    [sym_hash_operator] = ACTIONS(1981),
    [anon_sym_lsl] = ACTIONS(1979),
    [anon_sym_lsr] = ACTIONS(1979),
    [anon_sym_asr] = ACTIONS(1979),
    [aux_sym_pow_operator_token1] = ACTIONS(1981),
    [anon_sym_mod] = ACTIONS(1979),
    [anon_sym_land] = ACTIONS(1979),
    [anon_sym_lor] = ACTIONS(1979),
    [anon_sym_lxor] = ACTIONS(1979),
    [aux_sym_mult_operator_token1] = ACTIONS(1979),
    [aux_sym_add_operator_token1] = ACTIONS(1979),
    [sym_concat_operator] = ACTIONS(1981),
    [sym_rel_operator] = ACTIONS(1979),
    [anon_sym_AMP_AMP] = ACTIONS(1979),
    [anon_sym_or] = ACTIONS(1979),
    [anon_sym_PIPE_PIPE] = ACTIONS(1979),
    [sym_assign_operator] = ACTIONS(1981),
    [sym_let_operator] = ACTIONS(1981),
    [sym__uppercase_identifier] = ACTIONS(1981),
    [aux_sym_directive_token1] = ACTIONS(1979),
    [aux_sym_tag_token1] = ACTIONS(1981),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1470)] = {
    [sym_attribute] = STATE(1470),
    [ts_builtin_sym_end] = ACTIONS(2105),
    [sym__lowercase_identifier] = ACTIONS(2103),
    [anon_sym_SEMI_SEMI] = ACTIONS(2105),
    [anon_sym_let] = ACTIONS(2103),
    [anon_sym_and] = ACTIONS(2103),
    [anon_sym_TILDE] = ACTIONS(2103),
    [anon_sym_QMARK] = ACTIONS(2103),
    [anon_sym_LPAREN] = ACTIONS(2105),
    [anon_sym_external] = ACTIONS(2103),
    [anon_sym_type] = ACTIONS(2103),
    [anon_sym_COMMA] = ACTIONS(2105),
    [anon_sym_PIPE] = ACTIONS(2103),
    [anon_sym_DOT] = ACTIONS(2105),
    [anon_sym_LBRACK] = ACTIONS(2103),
    [anon_sym_true] = ACTIONS(2103),
    [anon_sym_false] = ACTIONS(2103),
    [anon_sym_LBRACE] = ACTIONS(2103),
    [anon_sym_SEMI] = ACTIONS(2103),
    [anon_sym_exception] = ACTIONS(2103),
    [anon_sym_module] = ACTIONS(2103),
    [anon_sym_open] = ACTIONS(2103),
    [anon_sym_include] = ACTIONS(2103),
    [anon_sym_class] = ACTIONS(2103),
    [anon_sym_val] = ACTIONS(2103),
    [anon_sym_object] = ACTIONS(2103),
    [anon_sym_AMP] = ACTIONS(2103),
    [anon_sym_POUND] = ACTIONS(2103),
    [anon_sym_COLON_COLON] = ACTIONS(2105),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2105),
    [anon_sym_else] = ACTIONS(2103),
    [anon_sym_new] = ACTIONS(2103),
    [anon_sym_LBRACE_LT] = ACTIONS(2105),
    [anon_sym_begin] = ACTIONS(2103),
    [sym_ocamlyacc_value] = ACTIONS(2105),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2103),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2105),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2103),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2103),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2105),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2105),
    [aux_sym_number_token1] = ACTIONS(2105),
    [aux_sym_signed_number_token1] = ACTIONS(2103),
    [anon_sym_SQUOTE] = ACTIONS(2105),
    [anon_sym_DQUOTE] = ACTIONS(2105),
    [sym_prefix_operator] = ACTIONS(2105),
    [aux_sym_sign_operator_token1] = ACTIONS(2103),
    [sym_hash_operator] = ACTIONS(2105),
    [anon_sym_lsl] = ACTIONS(2103),
    [anon_sym_lsr] = ACTIONS(2103),
    [anon_sym_asr] = ACTIONS(2103),
    [aux_sym_pow_operator_token1] = ACTIONS(2105),
    [anon_sym_mod] = ACTIONS(2103),
    [anon_sym_land] = ACTIONS(2103),
    [anon_sym_lor] = ACTIONS(2103),
    [anon_sym_lxor] = ACTIONS(2103),
    [aux_sym_mult_operator_token1] = ACTIONS(2103),
    [aux_sym_add_operator_token1] = ACTIONS(2103),
    [sym_concat_operator] = ACTIONS(2105),
    [sym_rel_operator] = ACTIONS(2103),
    [anon_sym_AMP_AMP] = ACTIONS(2103),
    [anon_sym_or] = ACTIONS(2103),
    [anon_sym_PIPE_PIPE] = ACTIONS(2103),
    [sym_assign_operator] = ACTIONS(2105),
    [sym_let_operator] = ACTIONS(2105),
    [sym__uppercase_identifier] = ACTIONS(2105),
    [aux_sym_directive_token1] = ACTIONS(2103),
    [aux_sym_tag_token1] = ACTIONS(2105),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1471)] = {
    [sym_attribute] = STATE(1471),
    [ts_builtin_sym_end] = ACTIONS(1949),
    [sym__lowercase_identifier] = ACTIONS(1947),
    [anon_sym_SEMI_SEMI] = ACTIONS(1949),
    [anon_sym_let] = ACTIONS(1947),
    [anon_sym_and] = ACTIONS(1947),
    [anon_sym_TILDE] = ACTIONS(1947),
    [anon_sym_QMARK] = ACTIONS(1947),
    [anon_sym_LPAREN] = ACTIONS(1949),
    [anon_sym_external] = ACTIONS(1947),
    [anon_sym_type] = ACTIONS(1947),
    [anon_sym_COMMA] = ACTIONS(1949),
    [anon_sym_PIPE] = ACTIONS(1947),
    [anon_sym_DOT] = ACTIONS(1949),
    [anon_sym_LBRACK] = ACTIONS(1947),
    [anon_sym_true] = ACTIONS(1947),
    [anon_sym_false] = ACTIONS(1947),
    [anon_sym_LBRACE] = ACTIONS(1947),
    [anon_sym_SEMI] = ACTIONS(1947),
    [anon_sym_exception] = ACTIONS(1947),
    [anon_sym_module] = ACTIONS(1947),
    [anon_sym_open] = ACTIONS(1947),
    [anon_sym_include] = ACTIONS(1947),
    [anon_sym_class] = ACTIONS(1947),
    [anon_sym_val] = ACTIONS(1947),
    [anon_sym_object] = ACTIONS(1947),
    [anon_sym_AMP] = ACTIONS(1947),
    [anon_sym_POUND] = ACTIONS(1947),
    [anon_sym_COLON_COLON] = ACTIONS(1949),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1949),
    [anon_sym_else] = ACTIONS(1947),
    [anon_sym_new] = ACTIONS(1947),
    [anon_sym_LBRACE_LT] = ACTIONS(1949),
    [anon_sym_begin] = ACTIONS(1947),
    [sym_ocamlyacc_value] = ACTIONS(1949),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1947),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1949),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1949),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1949),
    [aux_sym_number_token1] = ACTIONS(1949),
    [aux_sym_signed_number_token1] = ACTIONS(1947),
    [anon_sym_SQUOTE] = ACTIONS(1949),
    [anon_sym_DQUOTE] = ACTIONS(1949),
    [sym_prefix_operator] = ACTIONS(1949),
    [aux_sym_sign_operator_token1] = ACTIONS(1947),
    [sym_hash_operator] = ACTIONS(1949),
    [anon_sym_lsl] = ACTIONS(1947),
    [anon_sym_lsr] = ACTIONS(1947),
    [anon_sym_asr] = ACTIONS(1947),
    [aux_sym_pow_operator_token1] = ACTIONS(1949),
    [anon_sym_mod] = ACTIONS(1947),
    [anon_sym_land] = ACTIONS(1947),
    [anon_sym_lor] = ACTIONS(1947),
    [anon_sym_lxor] = ACTIONS(1947),
    [aux_sym_mult_operator_token1] = ACTIONS(1947),
    [aux_sym_add_operator_token1] = ACTIONS(1947),
    [sym_concat_operator] = ACTIONS(1949),
    [sym_rel_operator] = ACTIONS(1947),
    [anon_sym_AMP_AMP] = ACTIONS(1947),
    [anon_sym_or] = ACTIONS(1947),
    [anon_sym_PIPE_PIPE] = ACTIONS(1947),
    [sym_assign_operator] = ACTIONS(1949),
    [sym_let_operator] = ACTIONS(1949),
    [sym__uppercase_identifier] = ACTIONS(1949),
    [aux_sym_directive_token1] = ACTIONS(1947),
    [aux_sym_tag_token1] = ACTIONS(1949),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1472)] = {
    [sym_attribute] = STATE(1472),
    [ts_builtin_sym_end] = ACTIONS(2073),
    [sym__lowercase_identifier] = ACTIONS(2071),
    [anon_sym_SEMI_SEMI] = ACTIONS(2073),
    [anon_sym_let] = ACTIONS(2071),
    [anon_sym_and] = ACTIONS(2071),
    [anon_sym_TILDE] = ACTIONS(2071),
    [anon_sym_QMARK] = ACTIONS(2071),
    [anon_sym_LPAREN] = ACTIONS(2073),
    [anon_sym_external] = ACTIONS(2071),
    [anon_sym_type] = ACTIONS(2071),
    [anon_sym_COMMA] = ACTIONS(2073),
    [anon_sym_PIPE] = ACTIONS(2071),
    [anon_sym_DOT] = ACTIONS(2073),
    [anon_sym_LBRACK] = ACTIONS(2071),
    [anon_sym_true] = ACTIONS(2071),
    [anon_sym_false] = ACTIONS(2071),
    [anon_sym_LBRACE] = ACTIONS(2071),
    [anon_sym_SEMI] = ACTIONS(2071),
    [anon_sym_exception] = ACTIONS(2071),
    [anon_sym_module] = ACTIONS(2071),
    [anon_sym_open] = ACTIONS(2071),
    [anon_sym_include] = ACTIONS(2071),
    [anon_sym_class] = ACTIONS(2071),
    [anon_sym_val] = ACTIONS(2071),
    [anon_sym_object] = ACTIONS(2071),
    [anon_sym_AMP] = ACTIONS(2071),
    [anon_sym_POUND] = ACTIONS(2071),
    [anon_sym_COLON_COLON] = ACTIONS(2073),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2073),
    [anon_sym_else] = ACTIONS(2071),
    [anon_sym_new] = ACTIONS(2071),
    [anon_sym_LBRACE_LT] = ACTIONS(2073),
    [anon_sym_begin] = ACTIONS(2071),
    [sym_ocamlyacc_value] = ACTIONS(2073),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2071),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2073),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2071),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2071),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2073),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2073),
    [aux_sym_number_token1] = ACTIONS(2073),
    [aux_sym_signed_number_token1] = ACTIONS(2071),
    [anon_sym_SQUOTE] = ACTIONS(2073),
    [anon_sym_DQUOTE] = ACTIONS(2073),
    [sym_prefix_operator] = ACTIONS(2073),
    [aux_sym_sign_operator_token1] = ACTIONS(2071),
    [sym_hash_operator] = ACTIONS(2073),
    [anon_sym_lsl] = ACTIONS(2071),
    [anon_sym_lsr] = ACTIONS(2071),
    [anon_sym_asr] = ACTIONS(2071),
    [aux_sym_pow_operator_token1] = ACTIONS(2073),
    [anon_sym_mod] = ACTIONS(2071),
    [anon_sym_land] = ACTIONS(2071),
    [anon_sym_lor] = ACTIONS(2071),
    [anon_sym_lxor] = ACTIONS(2071),
    [aux_sym_mult_operator_token1] = ACTIONS(2071),
    [aux_sym_add_operator_token1] = ACTIONS(2071),
    [sym_concat_operator] = ACTIONS(2073),
    [sym_rel_operator] = ACTIONS(2071),
    [anon_sym_AMP_AMP] = ACTIONS(2071),
    [anon_sym_or] = ACTIONS(2071),
    [anon_sym_PIPE_PIPE] = ACTIONS(2071),
    [sym_assign_operator] = ACTIONS(2073),
    [sym_let_operator] = ACTIONS(2073),
    [sym__uppercase_identifier] = ACTIONS(2073),
    [aux_sym_directive_token1] = ACTIONS(2071),
    [aux_sym_tag_token1] = ACTIONS(2073),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1473)] = {
    [sym_attribute] = STATE(1473),
    [ts_builtin_sym_end] = ACTIONS(2109),
    [sym__lowercase_identifier] = ACTIONS(2107),
    [anon_sym_SEMI_SEMI] = ACTIONS(2109),
    [anon_sym_let] = ACTIONS(2107),
    [anon_sym_and] = ACTIONS(2107),
    [anon_sym_TILDE] = ACTIONS(2107),
    [anon_sym_QMARK] = ACTIONS(2107),
    [anon_sym_LPAREN] = ACTIONS(2109),
    [anon_sym_external] = ACTIONS(2107),
    [anon_sym_type] = ACTIONS(2107),
    [anon_sym_COMMA] = ACTIONS(2109),
    [anon_sym_PIPE] = ACTIONS(2107),
    [anon_sym_DOT] = ACTIONS(2109),
    [anon_sym_LBRACK] = ACTIONS(2107),
    [anon_sym_true] = ACTIONS(2107),
    [anon_sym_false] = ACTIONS(2107),
    [anon_sym_LBRACE] = ACTIONS(2107),
    [anon_sym_SEMI] = ACTIONS(2107),
    [anon_sym_exception] = ACTIONS(2107),
    [anon_sym_module] = ACTIONS(2107),
    [anon_sym_open] = ACTIONS(2107),
    [anon_sym_include] = ACTIONS(2107),
    [anon_sym_class] = ACTIONS(2107),
    [anon_sym_val] = ACTIONS(2107),
    [anon_sym_object] = ACTIONS(2107),
    [anon_sym_AMP] = ACTIONS(2107),
    [anon_sym_POUND] = ACTIONS(2107),
    [anon_sym_COLON_COLON] = ACTIONS(2109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2109),
    [anon_sym_else] = ACTIONS(2107),
    [anon_sym_new] = ACTIONS(2107),
    [anon_sym_LBRACE_LT] = ACTIONS(2109),
    [anon_sym_begin] = ACTIONS(2107),
    [sym_ocamlyacc_value] = ACTIONS(2109),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2107),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2109),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2109),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2109),
    [aux_sym_number_token1] = ACTIONS(2109),
    [aux_sym_signed_number_token1] = ACTIONS(2107),
    [anon_sym_SQUOTE] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(2109),
    [sym_prefix_operator] = ACTIONS(2109),
    [aux_sym_sign_operator_token1] = ACTIONS(2107),
    [sym_hash_operator] = ACTIONS(2109),
    [anon_sym_lsl] = ACTIONS(2107),
    [anon_sym_lsr] = ACTIONS(2107),
    [anon_sym_asr] = ACTIONS(2107),
    [aux_sym_pow_operator_token1] = ACTIONS(2109),
    [anon_sym_mod] = ACTIONS(2107),
    [anon_sym_land] = ACTIONS(2107),
    [anon_sym_lor] = ACTIONS(2107),
    [anon_sym_lxor] = ACTIONS(2107),
    [aux_sym_mult_operator_token1] = ACTIONS(2107),
    [aux_sym_add_operator_token1] = ACTIONS(2107),
    [sym_concat_operator] = ACTIONS(2109),
    [sym_rel_operator] = ACTIONS(2107),
    [anon_sym_AMP_AMP] = ACTIONS(2107),
    [anon_sym_or] = ACTIONS(2107),
    [anon_sym_PIPE_PIPE] = ACTIONS(2107),
    [sym_assign_operator] = ACTIONS(2109),
    [sym_let_operator] = ACTIONS(2109),
    [sym__uppercase_identifier] = ACTIONS(2109),
    [aux_sym_directive_token1] = ACTIONS(2107),
    [aux_sym_tag_token1] = ACTIONS(2109),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1474)] = {
    [sym_attribute] = STATE(1474),
    [ts_builtin_sym_end] = ACTIONS(2181),
    [sym__lowercase_identifier] = ACTIONS(2179),
    [anon_sym_SEMI_SEMI] = ACTIONS(2181),
    [anon_sym_let] = ACTIONS(2179),
    [anon_sym_and] = ACTIONS(2179),
    [anon_sym_TILDE] = ACTIONS(2179),
    [anon_sym_QMARK] = ACTIONS(2179),
    [anon_sym_LPAREN] = ACTIONS(2181),
    [anon_sym_external] = ACTIONS(2179),
    [anon_sym_type] = ACTIONS(2179),
    [anon_sym_COMMA] = ACTIONS(2181),
    [anon_sym_PIPE] = ACTIONS(2179),
    [anon_sym_DOT] = ACTIONS(2181),
    [anon_sym_LBRACK] = ACTIONS(2179),
    [anon_sym_true] = ACTIONS(2179),
    [anon_sym_false] = ACTIONS(2179),
    [anon_sym_LBRACE] = ACTIONS(2179),
    [anon_sym_SEMI] = ACTIONS(2179),
    [anon_sym_exception] = ACTIONS(2179),
    [anon_sym_module] = ACTIONS(2179),
    [anon_sym_open] = ACTIONS(2179),
    [anon_sym_include] = ACTIONS(2179),
    [anon_sym_class] = ACTIONS(2179),
    [anon_sym_val] = ACTIONS(2179),
    [anon_sym_object] = ACTIONS(2179),
    [anon_sym_AMP] = ACTIONS(2179),
    [anon_sym_POUND] = ACTIONS(2179),
    [anon_sym_COLON_COLON] = ACTIONS(2181),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2181),
    [anon_sym_else] = ACTIONS(2179),
    [anon_sym_new] = ACTIONS(2179),
    [anon_sym_LBRACE_LT] = ACTIONS(2181),
    [anon_sym_begin] = ACTIONS(2179),
    [sym_ocamlyacc_value] = ACTIONS(2181),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2179),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2181),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2179),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2181),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2181),
    [aux_sym_number_token1] = ACTIONS(2181),
    [aux_sym_signed_number_token1] = ACTIONS(2179),
    [anon_sym_SQUOTE] = ACTIONS(2181),
    [anon_sym_DQUOTE] = ACTIONS(2181),
    [sym_prefix_operator] = ACTIONS(2181),
    [aux_sym_sign_operator_token1] = ACTIONS(2179),
    [sym_hash_operator] = ACTIONS(2181),
    [anon_sym_lsl] = ACTIONS(2179),
    [anon_sym_lsr] = ACTIONS(2179),
    [anon_sym_asr] = ACTIONS(2179),
    [aux_sym_pow_operator_token1] = ACTIONS(2181),
    [anon_sym_mod] = ACTIONS(2179),
    [anon_sym_land] = ACTIONS(2179),
    [anon_sym_lor] = ACTIONS(2179),
    [anon_sym_lxor] = ACTIONS(2179),
    [aux_sym_mult_operator_token1] = ACTIONS(2179),
    [aux_sym_add_operator_token1] = ACTIONS(2179),
    [sym_concat_operator] = ACTIONS(2181),
    [sym_rel_operator] = ACTIONS(2179),
    [anon_sym_AMP_AMP] = ACTIONS(2179),
    [anon_sym_or] = ACTIONS(2179),
    [anon_sym_PIPE_PIPE] = ACTIONS(2179),
    [sym_assign_operator] = ACTIONS(2181),
    [sym_let_operator] = ACTIONS(2181),
    [sym__uppercase_identifier] = ACTIONS(2181),
    [aux_sym_directive_token1] = ACTIONS(2179),
    [aux_sym_tag_token1] = ACTIONS(2181),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1475)] = {
    [sym_attribute] = STATE(1475),
    [ts_builtin_sym_end] = ACTIONS(2017),
    [sym__lowercase_identifier] = ACTIONS(2015),
    [anon_sym_SEMI_SEMI] = ACTIONS(2017),
    [anon_sym_let] = ACTIONS(2015),
    [anon_sym_and] = ACTIONS(2015),
    [anon_sym_TILDE] = ACTIONS(2015),
    [anon_sym_QMARK] = ACTIONS(2015),
    [anon_sym_LPAREN] = ACTIONS(2017),
    [anon_sym_external] = ACTIONS(2015),
    [anon_sym_type] = ACTIONS(2015),
    [anon_sym_COMMA] = ACTIONS(2017),
    [anon_sym_PIPE] = ACTIONS(2015),
    [anon_sym_DOT] = ACTIONS(2017),
    [anon_sym_LBRACK] = ACTIONS(2015),
    [anon_sym_true] = ACTIONS(2015),
    [anon_sym_false] = ACTIONS(2015),
    [anon_sym_LBRACE] = ACTIONS(2015),
    [anon_sym_SEMI] = ACTIONS(2015),
    [anon_sym_exception] = ACTIONS(2015),
    [anon_sym_module] = ACTIONS(2015),
    [anon_sym_open] = ACTIONS(2015),
    [anon_sym_include] = ACTIONS(2015),
    [anon_sym_class] = ACTIONS(2015),
    [anon_sym_val] = ACTIONS(2015),
    [anon_sym_object] = ACTIONS(2015),
    [anon_sym_AMP] = ACTIONS(2015),
    [anon_sym_POUND] = ACTIONS(2015),
    [anon_sym_COLON_COLON] = ACTIONS(2017),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2017),
    [anon_sym_else] = ACTIONS(2015),
    [anon_sym_new] = ACTIONS(2015),
    [anon_sym_LBRACE_LT] = ACTIONS(2017),
    [anon_sym_begin] = ACTIONS(2015),
    [sym_ocamlyacc_value] = ACTIONS(2017),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2015),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2017),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2015),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2015),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2017),
    [aux_sym_number_token1] = ACTIONS(2017),
    [aux_sym_signed_number_token1] = ACTIONS(2015),
    [anon_sym_SQUOTE] = ACTIONS(2017),
    [anon_sym_DQUOTE] = ACTIONS(2017),
    [sym_prefix_operator] = ACTIONS(2017),
    [aux_sym_sign_operator_token1] = ACTIONS(2015),
    [sym_hash_operator] = ACTIONS(2017),
    [anon_sym_lsl] = ACTIONS(2015),
    [anon_sym_lsr] = ACTIONS(2015),
    [anon_sym_asr] = ACTIONS(2015),
    [aux_sym_pow_operator_token1] = ACTIONS(2017),
    [anon_sym_mod] = ACTIONS(2015),
    [anon_sym_land] = ACTIONS(2015),
    [anon_sym_lor] = ACTIONS(2015),
    [anon_sym_lxor] = ACTIONS(2015),
    [aux_sym_mult_operator_token1] = ACTIONS(2015),
    [aux_sym_add_operator_token1] = ACTIONS(2015),
    [sym_concat_operator] = ACTIONS(2017),
    [sym_rel_operator] = ACTIONS(2015),
    [anon_sym_AMP_AMP] = ACTIONS(2015),
    [anon_sym_or] = ACTIONS(2015),
    [anon_sym_PIPE_PIPE] = ACTIONS(2015),
    [sym_assign_operator] = ACTIONS(2017),
    [sym_let_operator] = ACTIONS(2017),
    [sym__uppercase_identifier] = ACTIONS(2017),
    [aux_sym_directive_token1] = ACTIONS(2015),
    [aux_sym_tag_token1] = ACTIONS(2017),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1476)] = {
    [sym_attribute] = STATE(1476),
    [ts_builtin_sym_end] = ACTIONS(2021),
    [sym__lowercase_identifier] = ACTIONS(2019),
    [anon_sym_SEMI_SEMI] = ACTIONS(2021),
    [anon_sym_let] = ACTIONS(2019),
    [anon_sym_and] = ACTIONS(2019),
    [anon_sym_TILDE] = ACTIONS(2019),
    [anon_sym_QMARK] = ACTIONS(2019),
    [anon_sym_LPAREN] = ACTIONS(2021),
    [anon_sym_external] = ACTIONS(2019),
    [anon_sym_type] = ACTIONS(2019),
    [anon_sym_COMMA] = ACTIONS(2021),
    [anon_sym_PIPE] = ACTIONS(2019),
    [anon_sym_DOT] = ACTIONS(2021),
    [anon_sym_LBRACK] = ACTIONS(2019),
    [anon_sym_true] = ACTIONS(2019),
    [anon_sym_false] = ACTIONS(2019),
    [anon_sym_LBRACE] = ACTIONS(2019),
    [anon_sym_SEMI] = ACTIONS(2019),
    [anon_sym_exception] = ACTIONS(2019),
    [anon_sym_module] = ACTIONS(2019),
    [anon_sym_open] = ACTIONS(2019),
    [anon_sym_include] = ACTIONS(2019),
    [anon_sym_class] = ACTIONS(2019),
    [anon_sym_val] = ACTIONS(2019),
    [anon_sym_object] = ACTIONS(2019),
    [anon_sym_AMP] = ACTIONS(2019),
    [anon_sym_POUND] = ACTIONS(2019),
    [anon_sym_COLON_COLON] = ACTIONS(2021),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2021),
    [anon_sym_else] = ACTIONS(2019),
    [anon_sym_new] = ACTIONS(2019),
    [anon_sym_LBRACE_LT] = ACTIONS(2021),
    [anon_sym_begin] = ACTIONS(2019),
    [sym_ocamlyacc_value] = ACTIONS(2021),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2019),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2021),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2021),
    [aux_sym_number_token1] = ACTIONS(2021),
    [aux_sym_signed_number_token1] = ACTIONS(2019),
    [anon_sym_SQUOTE] = ACTIONS(2021),
    [anon_sym_DQUOTE] = ACTIONS(2021),
    [sym_prefix_operator] = ACTIONS(2021),
    [aux_sym_sign_operator_token1] = ACTIONS(2019),
    [sym_hash_operator] = ACTIONS(2021),
    [anon_sym_lsl] = ACTIONS(2019),
    [anon_sym_lsr] = ACTIONS(2019),
    [anon_sym_asr] = ACTIONS(2019),
    [aux_sym_pow_operator_token1] = ACTIONS(2021),
    [anon_sym_mod] = ACTIONS(2019),
    [anon_sym_land] = ACTIONS(2019),
    [anon_sym_lor] = ACTIONS(2019),
    [anon_sym_lxor] = ACTIONS(2019),
    [aux_sym_mult_operator_token1] = ACTIONS(2019),
    [aux_sym_add_operator_token1] = ACTIONS(2019),
    [sym_concat_operator] = ACTIONS(2021),
    [sym_rel_operator] = ACTIONS(2019),
    [anon_sym_AMP_AMP] = ACTIONS(2019),
    [anon_sym_or] = ACTIONS(2019),
    [anon_sym_PIPE_PIPE] = ACTIONS(2019),
    [sym_assign_operator] = ACTIONS(2021),
    [sym_let_operator] = ACTIONS(2021),
    [sym__uppercase_identifier] = ACTIONS(2021),
    [aux_sym_directive_token1] = ACTIONS(2019),
    [aux_sym_tag_token1] = ACTIONS(2021),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1477)] = {
    [sym_parameter] = STATE(2862),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1477),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(7381),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(3191),
    [sym_tag] = STATE(2866),
    [aux_sym_let_binding_repeat1] = STATE(1477),
    [sym__lowercase_identifier] = ACTIONS(2640),
    [anon_sym_SEMI_SEMI] = ACTIONS(2463),
    [anon_sym_and] = ACTIONS(2468),
    [anon_sym_EQ] = ACTIONS(2463),
    [anon_sym_TILDE] = ACTIONS(2643),
    [anon_sym_QMARK] = ACTIONS(2643),
    [anon_sym_LPAREN] = ACTIONS(2646),
    [anon_sym_external] = ACTIONS(2468),
    [anon_sym_type] = ACTIONS(2468),
    [anon_sym_COLON2] = ACTIONS(2463),
    [anon_sym_LBRACK] = ACTIONS(2649),
    [anon_sym_RBRACK] = ACTIONS(2463),
    [anon_sym_true] = ACTIONS(2652),
    [anon_sym_false] = ACTIONS(2652),
    [anon_sym_LBRACE] = ACTIONS(2655),
    [anon_sym_exception] = ACTIONS(2468),
    [anon_sym_module] = ACTIONS(2468),
    [anon_sym_open] = ACTIONS(2468),
    [anon_sym_include] = ACTIONS(2468),
    [anon_sym_class] = ACTIONS(2468),
    [anon_sym_val] = ACTIONS(2468),
    [anon_sym_end] = ACTIONS(2468),
    [anon_sym_POUND] = ACTIONS(2658),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2661),
    [anon_sym_begin] = ACTIONS(2664),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2468),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2463),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2667),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2670),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2463),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2463),
    [aux_sym_number_token1] = ACTIONS(2673),
    [aux_sym_signed_number_token1] = ACTIONS(2676),
    [anon_sym_SQUOTE] = ACTIONS(2679),
    [anon_sym_DQUOTE] = ACTIONS(2682),
    [sym__uppercase_identifier] = ACTIONS(2685),
    [aux_sym_tag_token1] = ACTIONS(2688),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1478)] = {
    [sym_attribute] = STATE(1478),
    [ts_builtin_sym_end] = ACTIONS(2025),
    [sym__lowercase_identifier] = ACTIONS(2023),
    [anon_sym_SEMI_SEMI] = ACTIONS(2025),
    [anon_sym_let] = ACTIONS(2023),
    [anon_sym_and] = ACTIONS(2023),
    [anon_sym_TILDE] = ACTIONS(2023),
    [anon_sym_QMARK] = ACTIONS(2023),
    [anon_sym_LPAREN] = ACTIONS(2025),
    [anon_sym_external] = ACTIONS(2023),
    [anon_sym_type] = ACTIONS(2023),
    [anon_sym_COMMA] = ACTIONS(2025),
    [anon_sym_PIPE] = ACTIONS(2023),
    [anon_sym_DOT] = ACTIONS(2025),
    [anon_sym_LBRACK] = ACTIONS(2023),
    [anon_sym_true] = ACTIONS(2023),
    [anon_sym_false] = ACTIONS(2023),
    [anon_sym_LBRACE] = ACTIONS(2023),
    [anon_sym_SEMI] = ACTIONS(2023),
    [anon_sym_exception] = ACTIONS(2023),
    [anon_sym_module] = ACTIONS(2023),
    [anon_sym_open] = ACTIONS(2023),
    [anon_sym_include] = ACTIONS(2023),
    [anon_sym_class] = ACTIONS(2023),
    [anon_sym_val] = ACTIONS(2023),
    [anon_sym_object] = ACTIONS(2023),
    [anon_sym_AMP] = ACTIONS(2023),
    [anon_sym_POUND] = ACTIONS(2023),
    [anon_sym_COLON_COLON] = ACTIONS(2025),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2025),
    [anon_sym_else] = ACTIONS(2023),
    [anon_sym_new] = ACTIONS(2023),
    [anon_sym_LBRACE_LT] = ACTIONS(2025),
    [anon_sym_begin] = ACTIONS(2023),
    [sym_ocamlyacc_value] = ACTIONS(2025),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2023),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2025),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2025),
    [aux_sym_number_token1] = ACTIONS(2025),
    [aux_sym_signed_number_token1] = ACTIONS(2023),
    [anon_sym_SQUOTE] = ACTIONS(2025),
    [anon_sym_DQUOTE] = ACTIONS(2025),
    [sym_prefix_operator] = ACTIONS(2025),
    [aux_sym_sign_operator_token1] = ACTIONS(2023),
    [sym_hash_operator] = ACTIONS(2025),
    [anon_sym_lsl] = ACTIONS(2023),
    [anon_sym_lsr] = ACTIONS(2023),
    [anon_sym_asr] = ACTIONS(2023),
    [aux_sym_pow_operator_token1] = ACTIONS(2025),
    [anon_sym_mod] = ACTIONS(2023),
    [anon_sym_land] = ACTIONS(2023),
    [anon_sym_lor] = ACTIONS(2023),
    [anon_sym_lxor] = ACTIONS(2023),
    [aux_sym_mult_operator_token1] = ACTIONS(2023),
    [aux_sym_add_operator_token1] = ACTIONS(2023),
    [sym_concat_operator] = ACTIONS(2025),
    [sym_rel_operator] = ACTIONS(2023),
    [anon_sym_AMP_AMP] = ACTIONS(2023),
    [anon_sym_or] = ACTIONS(2023),
    [anon_sym_PIPE_PIPE] = ACTIONS(2023),
    [sym_assign_operator] = ACTIONS(2025),
    [sym_let_operator] = ACTIONS(2025),
    [sym__uppercase_identifier] = ACTIONS(2025),
    [aux_sym_directive_token1] = ACTIONS(2023),
    [aux_sym_tag_token1] = ACTIONS(2025),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1479)] = {
    [sym_attribute] = STATE(1479),
    [ts_builtin_sym_end] = ACTIONS(2029),
    [sym__lowercase_identifier] = ACTIONS(2027),
    [anon_sym_SEMI_SEMI] = ACTIONS(2029),
    [anon_sym_let] = ACTIONS(2027),
    [anon_sym_and] = ACTIONS(2027),
    [anon_sym_TILDE] = ACTIONS(2027),
    [anon_sym_QMARK] = ACTIONS(2027),
    [anon_sym_LPAREN] = ACTIONS(2029),
    [anon_sym_external] = ACTIONS(2027),
    [anon_sym_type] = ACTIONS(2027),
    [anon_sym_COMMA] = ACTIONS(2029),
    [anon_sym_PIPE] = ACTIONS(2027),
    [anon_sym_DOT] = ACTIONS(2029),
    [anon_sym_LBRACK] = ACTIONS(2027),
    [anon_sym_true] = ACTIONS(2027),
    [anon_sym_false] = ACTIONS(2027),
    [anon_sym_LBRACE] = ACTIONS(2027),
    [anon_sym_SEMI] = ACTIONS(2027),
    [anon_sym_exception] = ACTIONS(2027),
    [anon_sym_module] = ACTIONS(2027),
    [anon_sym_open] = ACTIONS(2027),
    [anon_sym_include] = ACTIONS(2027),
    [anon_sym_class] = ACTIONS(2027),
    [anon_sym_val] = ACTIONS(2027),
    [anon_sym_object] = ACTIONS(2027),
    [anon_sym_AMP] = ACTIONS(2027),
    [anon_sym_POUND] = ACTIONS(2027),
    [anon_sym_COLON_COLON] = ACTIONS(2029),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2029),
    [anon_sym_else] = ACTIONS(2027),
    [anon_sym_new] = ACTIONS(2027),
    [anon_sym_LBRACE_LT] = ACTIONS(2029),
    [anon_sym_begin] = ACTIONS(2027),
    [sym_ocamlyacc_value] = ACTIONS(2029),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2027),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2029),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2029),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2029),
    [aux_sym_number_token1] = ACTIONS(2029),
    [aux_sym_signed_number_token1] = ACTIONS(2027),
    [anon_sym_SQUOTE] = ACTIONS(2029),
    [anon_sym_DQUOTE] = ACTIONS(2029),
    [sym_prefix_operator] = ACTIONS(2029),
    [aux_sym_sign_operator_token1] = ACTIONS(2027),
    [sym_hash_operator] = ACTIONS(2029),
    [anon_sym_lsl] = ACTIONS(2027),
    [anon_sym_lsr] = ACTIONS(2027),
    [anon_sym_asr] = ACTIONS(2027),
    [aux_sym_pow_operator_token1] = ACTIONS(2029),
    [anon_sym_mod] = ACTIONS(2027),
    [anon_sym_land] = ACTIONS(2027),
    [anon_sym_lor] = ACTIONS(2027),
    [anon_sym_lxor] = ACTIONS(2027),
    [aux_sym_mult_operator_token1] = ACTIONS(2027),
    [aux_sym_add_operator_token1] = ACTIONS(2027),
    [sym_concat_operator] = ACTIONS(2029),
    [sym_rel_operator] = ACTIONS(2027),
    [anon_sym_AMP_AMP] = ACTIONS(2027),
    [anon_sym_or] = ACTIONS(2027),
    [anon_sym_PIPE_PIPE] = ACTIONS(2027),
    [sym_assign_operator] = ACTIONS(2029),
    [sym_let_operator] = ACTIONS(2029),
    [sym__uppercase_identifier] = ACTIONS(2029),
    [aux_sym_directive_token1] = ACTIONS(2027),
    [aux_sym_tag_token1] = ACTIONS(2029),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1480)] = {
    [sym_attribute] = STATE(1480),
    [ts_builtin_sym_end] = ACTIONS(1925),
    [sym__lowercase_identifier] = ACTIONS(1923),
    [anon_sym_SEMI_SEMI] = ACTIONS(1925),
    [anon_sym_let] = ACTIONS(1923),
    [anon_sym_and] = ACTIONS(1923),
    [anon_sym_TILDE] = ACTIONS(1923),
    [anon_sym_QMARK] = ACTIONS(1923),
    [anon_sym_LPAREN] = ACTIONS(1925),
    [anon_sym_external] = ACTIONS(1923),
    [anon_sym_type] = ACTIONS(1923),
    [anon_sym_COMMA] = ACTIONS(1925),
    [anon_sym_PIPE] = ACTIONS(1923),
    [anon_sym_DOT] = ACTIONS(1925),
    [anon_sym_LBRACK] = ACTIONS(1923),
    [anon_sym_true] = ACTIONS(1923),
    [anon_sym_false] = ACTIONS(1923),
    [anon_sym_LBRACE] = ACTIONS(1923),
    [anon_sym_SEMI] = ACTIONS(1923),
    [anon_sym_exception] = ACTIONS(1923),
    [anon_sym_module] = ACTIONS(1923),
    [anon_sym_open] = ACTIONS(1923),
    [anon_sym_include] = ACTIONS(1923),
    [anon_sym_class] = ACTIONS(1923),
    [anon_sym_val] = ACTIONS(1923),
    [anon_sym_object] = ACTIONS(1923),
    [anon_sym_AMP] = ACTIONS(1923),
    [anon_sym_POUND] = ACTIONS(1923),
    [anon_sym_COLON_COLON] = ACTIONS(1925),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1925),
    [anon_sym_else] = ACTIONS(1923),
    [anon_sym_new] = ACTIONS(1923),
    [anon_sym_LBRACE_LT] = ACTIONS(1925),
    [anon_sym_begin] = ACTIONS(1923),
    [sym_ocamlyacc_value] = ACTIONS(1925),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1923),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1925),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1923),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1923),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1925),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1925),
    [aux_sym_number_token1] = ACTIONS(1925),
    [aux_sym_signed_number_token1] = ACTIONS(1923),
    [anon_sym_SQUOTE] = ACTIONS(1925),
    [anon_sym_DQUOTE] = ACTIONS(1925),
    [sym_prefix_operator] = ACTIONS(1925),
    [aux_sym_sign_operator_token1] = ACTIONS(1923),
    [sym_hash_operator] = ACTIONS(1925),
    [anon_sym_lsl] = ACTIONS(1923),
    [anon_sym_lsr] = ACTIONS(1923),
    [anon_sym_asr] = ACTIONS(1923),
    [aux_sym_pow_operator_token1] = ACTIONS(1925),
    [anon_sym_mod] = ACTIONS(1923),
    [anon_sym_land] = ACTIONS(1923),
    [anon_sym_lor] = ACTIONS(1923),
    [anon_sym_lxor] = ACTIONS(1923),
    [aux_sym_mult_operator_token1] = ACTIONS(1923),
    [aux_sym_add_operator_token1] = ACTIONS(1923),
    [sym_concat_operator] = ACTIONS(1925),
    [sym_rel_operator] = ACTIONS(1923),
    [anon_sym_AMP_AMP] = ACTIONS(1923),
    [anon_sym_or] = ACTIONS(1923),
    [anon_sym_PIPE_PIPE] = ACTIONS(1923),
    [sym_assign_operator] = ACTIONS(1925),
    [sym_let_operator] = ACTIONS(1925),
    [sym__uppercase_identifier] = ACTIONS(1925),
    [aux_sym_directive_token1] = ACTIONS(1923),
    [aux_sym_tag_token1] = ACTIONS(1925),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1481)] = {
    [sym_attribute] = STATE(1481),
    [ts_builtin_sym_end] = ACTIONS(2315),
    [sym__lowercase_identifier] = ACTIONS(2311),
    [anon_sym_SEMI_SEMI] = ACTIONS(2315),
    [anon_sym_let] = ACTIONS(2311),
    [anon_sym_and] = ACTIONS(2311),
    [anon_sym_TILDE] = ACTIONS(2311),
    [anon_sym_QMARK] = ACTIONS(2311),
    [anon_sym_COLON] = ACTIONS(2691),
    [anon_sym_LPAREN] = ACTIONS(2315),
    [anon_sym_external] = ACTIONS(2311),
    [anon_sym_type] = ACTIONS(2311),
    [anon_sym_COMMA] = ACTIONS(2315),
    [anon_sym_PIPE] = ACTIONS(2311),
    [anon_sym_LBRACK] = ACTIONS(2311),
    [anon_sym_true] = ACTIONS(2311),
    [anon_sym_false] = ACTIONS(2311),
    [anon_sym_LBRACE] = ACTIONS(2311),
    [anon_sym_SEMI] = ACTIONS(2311),
    [anon_sym_exception] = ACTIONS(2311),
    [anon_sym_module] = ACTIONS(2311),
    [anon_sym_open] = ACTIONS(2311),
    [anon_sym_include] = ACTIONS(2311),
    [anon_sym_class] = ACTIONS(2311),
    [anon_sym_val] = ACTIONS(2311),
    [anon_sym_object] = ACTIONS(2311),
    [anon_sym_in] = ACTIONS(2311),
    [anon_sym_AMP] = ACTIONS(2311),
    [anon_sym_COLON_COLON] = ACTIONS(2315),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2315),
    [anon_sym_else] = ACTIONS(2311),
    [anon_sym_new] = ACTIONS(2311),
    [anon_sym_LBRACE_LT] = ACTIONS(2315),
    [anon_sym_begin] = ACTIONS(2311),
    [sym_ocamlyacc_value] = ACTIONS(2315),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2311),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2315),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2311),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2311),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2315),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2315),
    [aux_sym_number_token1] = ACTIONS(2315),
    [aux_sym_signed_number_token1] = ACTIONS(2311),
    [anon_sym_SQUOTE] = ACTIONS(2315),
    [anon_sym_DQUOTE] = ACTIONS(2315),
    [sym_prefix_operator] = ACTIONS(2315),
    [aux_sym_sign_operator_token1] = ACTIONS(2311),
    [anon_sym_lsl] = ACTIONS(2311),
    [anon_sym_lsr] = ACTIONS(2311),
    [anon_sym_asr] = ACTIONS(2311),
    [aux_sym_pow_operator_token1] = ACTIONS(2315),
    [anon_sym_mod] = ACTIONS(2311),
    [anon_sym_land] = ACTIONS(2311),
    [anon_sym_lor] = ACTIONS(2311),
    [anon_sym_lxor] = ACTIONS(2311),
    [aux_sym_mult_operator_token1] = ACTIONS(2311),
    [aux_sym_add_operator_token1] = ACTIONS(2311),
    [sym_concat_operator] = ACTIONS(2315),
    [sym_rel_operator] = ACTIONS(2311),
    [anon_sym_AMP_AMP] = ACTIONS(2311),
    [anon_sym_or] = ACTIONS(2311),
    [anon_sym_PIPE_PIPE] = ACTIONS(2311),
    [sym_assign_operator] = ACTIONS(2315),
    [sym_let_operator] = ACTIONS(2315),
    [sym_let_and_operator] = ACTIONS(2315),
    [sym__uppercase_identifier] = ACTIONS(2315),
    [aux_sym_directive_token1] = ACTIONS(2315),
    [aux_sym_tag_token1] = ACTIONS(2315),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1482)] = {
    [sym_attribute] = STATE(1482),
    [ts_builtin_sym_end] = ACTIONS(2065),
    [sym__lowercase_identifier] = ACTIONS(2063),
    [anon_sym_SEMI_SEMI] = ACTIONS(2065),
    [anon_sym_let] = ACTIONS(2063),
    [anon_sym_and] = ACTIONS(2063),
    [anon_sym_TILDE] = ACTIONS(2063),
    [anon_sym_QMARK] = ACTIONS(2063),
    [anon_sym_LPAREN] = ACTIONS(2065),
    [anon_sym_external] = ACTIONS(2063),
    [anon_sym_type] = ACTIONS(2063),
    [anon_sym_COMMA] = ACTIONS(2065),
    [anon_sym_PIPE] = ACTIONS(2063),
    [anon_sym_DOT] = ACTIONS(2065),
    [anon_sym_LBRACK] = ACTIONS(2063),
    [anon_sym_true] = ACTIONS(2063),
    [anon_sym_false] = ACTIONS(2063),
    [anon_sym_LBRACE] = ACTIONS(2063),
    [anon_sym_SEMI] = ACTIONS(2063),
    [anon_sym_exception] = ACTIONS(2063),
    [anon_sym_module] = ACTIONS(2063),
    [anon_sym_open] = ACTIONS(2063),
    [anon_sym_include] = ACTIONS(2063),
    [anon_sym_class] = ACTIONS(2063),
    [anon_sym_val] = ACTIONS(2063),
    [anon_sym_object] = ACTIONS(2063),
    [anon_sym_AMP] = ACTIONS(2063),
    [anon_sym_POUND] = ACTIONS(2063),
    [anon_sym_COLON_COLON] = ACTIONS(2065),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2065),
    [anon_sym_else] = ACTIONS(2063),
    [anon_sym_new] = ACTIONS(2063),
    [anon_sym_LBRACE_LT] = ACTIONS(2065),
    [anon_sym_begin] = ACTIONS(2063),
    [sym_ocamlyacc_value] = ACTIONS(2065),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2063),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2065),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2063),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2063),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2065),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2065),
    [aux_sym_number_token1] = ACTIONS(2065),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2065),
    [sym_prefix_operator] = ACTIONS(2065),
    [aux_sym_sign_operator_token1] = ACTIONS(2063),
    [sym_hash_operator] = ACTIONS(2065),
    [anon_sym_lsl] = ACTIONS(2063),
    [anon_sym_lsr] = ACTIONS(2063),
    [anon_sym_asr] = ACTIONS(2063),
    [aux_sym_pow_operator_token1] = ACTIONS(2065),
    [anon_sym_mod] = ACTIONS(2063),
    [anon_sym_land] = ACTIONS(2063),
    [anon_sym_lor] = ACTIONS(2063),
    [anon_sym_lxor] = ACTIONS(2063),
    [aux_sym_mult_operator_token1] = ACTIONS(2063),
    [aux_sym_add_operator_token1] = ACTIONS(2063),
    [sym_concat_operator] = ACTIONS(2065),
    [sym_rel_operator] = ACTIONS(2063),
    [anon_sym_AMP_AMP] = ACTIONS(2063),
    [anon_sym_or] = ACTIONS(2063),
    [anon_sym_PIPE_PIPE] = ACTIONS(2063),
    [sym_assign_operator] = ACTIONS(2065),
    [sym_let_operator] = ACTIONS(2065),
    [sym__uppercase_identifier] = ACTIONS(2065),
    [aux_sym_directive_token1] = ACTIONS(2063),
    [aux_sym_tag_token1] = ACTIONS(2065),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1483)] = {
    [sym_attribute] = STATE(1483),
    [ts_builtin_sym_end] = ACTIONS(2123),
    [sym__lowercase_identifier] = ACTIONS(2121),
    [anon_sym_SEMI_SEMI] = ACTIONS(2123),
    [anon_sym_let] = ACTIONS(2121),
    [anon_sym_and] = ACTIONS(2121),
    [anon_sym_TILDE] = ACTIONS(2121),
    [anon_sym_QMARK] = ACTIONS(2121),
    [anon_sym_LPAREN] = ACTIONS(2123),
    [anon_sym_external] = ACTIONS(2121),
    [anon_sym_type] = ACTIONS(2121),
    [anon_sym_COMMA] = ACTIONS(2123),
    [anon_sym_PIPE] = ACTIONS(2121),
    [anon_sym_DOT] = ACTIONS(2123),
    [anon_sym_LBRACK] = ACTIONS(2121),
    [anon_sym_true] = ACTIONS(2121),
    [anon_sym_false] = ACTIONS(2121),
    [anon_sym_LBRACE] = ACTIONS(2121),
    [anon_sym_SEMI] = ACTIONS(2121),
    [anon_sym_exception] = ACTIONS(2121),
    [anon_sym_module] = ACTIONS(2121),
    [anon_sym_open] = ACTIONS(2121),
    [anon_sym_include] = ACTIONS(2121),
    [anon_sym_class] = ACTIONS(2121),
    [anon_sym_val] = ACTIONS(2121),
    [anon_sym_object] = ACTIONS(2121),
    [anon_sym_AMP] = ACTIONS(2121),
    [anon_sym_POUND] = ACTIONS(2121),
    [anon_sym_COLON_COLON] = ACTIONS(2123),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2123),
    [anon_sym_else] = ACTIONS(2121),
    [anon_sym_new] = ACTIONS(2121),
    [anon_sym_LBRACE_LT] = ACTIONS(2123),
    [anon_sym_begin] = ACTIONS(2121),
    [sym_ocamlyacc_value] = ACTIONS(2123),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2121),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2123),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2121),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2121),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2123),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2123),
    [aux_sym_number_token1] = ACTIONS(2123),
    [aux_sym_signed_number_token1] = ACTIONS(2121),
    [anon_sym_SQUOTE] = ACTIONS(2123),
    [anon_sym_DQUOTE] = ACTIONS(2123),
    [sym_prefix_operator] = ACTIONS(2123),
    [aux_sym_sign_operator_token1] = ACTIONS(2121),
    [sym_hash_operator] = ACTIONS(2123),
    [anon_sym_lsl] = ACTIONS(2121),
    [anon_sym_lsr] = ACTIONS(2121),
    [anon_sym_asr] = ACTIONS(2121),
    [aux_sym_pow_operator_token1] = ACTIONS(2123),
    [anon_sym_mod] = ACTIONS(2121),
    [anon_sym_land] = ACTIONS(2121),
    [anon_sym_lor] = ACTIONS(2121),
    [anon_sym_lxor] = ACTIONS(2121),
    [aux_sym_mult_operator_token1] = ACTIONS(2121),
    [aux_sym_add_operator_token1] = ACTIONS(2121),
    [sym_concat_operator] = ACTIONS(2123),
    [sym_rel_operator] = ACTIONS(2121),
    [anon_sym_AMP_AMP] = ACTIONS(2121),
    [anon_sym_or] = ACTIONS(2121),
    [anon_sym_PIPE_PIPE] = ACTIONS(2121),
    [sym_assign_operator] = ACTIONS(2123),
    [sym_let_operator] = ACTIONS(2123),
    [sym__uppercase_identifier] = ACTIONS(2123),
    [aux_sym_directive_token1] = ACTIONS(2121),
    [aux_sym_tag_token1] = ACTIONS(2123),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1484)] = {
    [sym_attribute] = STATE(1484),
    [ts_builtin_sym_end] = ACTIONS(2127),
    [sym__lowercase_identifier] = ACTIONS(2125),
    [anon_sym_SEMI_SEMI] = ACTIONS(2127),
    [anon_sym_let] = ACTIONS(2125),
    [anon_sym_and] = ACTIONS(2125),
    [anon_sym_TILDE] = ACTIONS(2125),
    [anon_sym_QMARK] = ACTIONS(2125),
    [anon_sym_LPAREN] = ACTIONS(2127),
    [anon_sym_external] = ACTIONS(2125),
    [anon_sym_type] = ACTIONS(2125),
    [anon_sym_COMMA] = ACTIONS(2127),
    [anon_sym_PIPE] = ACTIONS(2125),
    [anon_sym_DOT] = ACTIONS(2127),
    [anon_sym_LBRACK] = ACTIONS(2125),
    [anon_sym_true] = ACTIONS(2125),
    [anon_sym_false] = ACTIONS(2125),
    [anon_sym_LBRACE] = ACTIONS(2125),
    [anon_sym_SEMI] = ACTIONS(2125),
    [anon_sym_exception] = ACTIONS(2125),
    [anon_sym_module] = ACTIONS(2125),
    [anon_sym_open] = ACTIONS(2125),
    [anon_sym_include] = ACTIONS(2125),
    [anon_sym_class] = ACTIONS(2125),
    [anon_sym_val] = ACTIONS(2125),
    [anon_sym_object] = ACTIONS(2125),
    [anon_sym_AMP] = ACTIONS(2125),
    [anon_sym_POUND] = ACTIONS(2125),
    [anon_sym_COLON_COLON] = ACTIONS(2127),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2127),
    [anon_sym_else] = ACTIONS(2125),
    [anon_sym_new] = ACTIONS(2125),
    [anon_sym_LBRACE_LT] = ACTIONS(2127),
    [anon_sym_begin] = ACTIONS(2125),
    [sym_ocamlyacc_value] = ACTIONS(2127),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2125),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2127),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2127),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2127),
    [aux_sym_number_token1] = ACTIONS(2127),
    [aux_sym_signed_number_token1] = ACTIONS(2125),
    [anon_sym_SQUOTE] = ACTIONS(2127),
    [anon_sym_DQUOTE] = ACTIONS(2127),
    [sym_prefix_operator] = ACTIONS(2127),
    [aux_sym_sign_operator_token1] = ACTIONS(2125),
    [sym_hash_operator] = ACTIONS(2127),
    [anon_sym_lsl] = ACTIONS(2125),
    [anon_sym_lsr] = ACTIONS(2125),
    [anon_sym_asr] = ACTIONS(2125),
    [aux_sym_pow_operator_token1] = ACTIONS(2127),
    [anon_sym_mod] = ACTIONS(2125),
    [anon_sym_land] = ACTIONS(2125),
    [anon_sym_lor] = ACTIONS(2125),
    [anon_sym_lxor] = ACTIONS(2125),
    [aux_sym_mult_operator_token1] = ACTIONS(2125),
    [aux_sym_add_operator_token1] = ACTIONS(2125),
    [sym_concat_operator] = ACTIONS(2127),
    [sym_rel_operator] = ACTIONS(2125),
    [anon_sym_AMP_AMP] = ACTIONS(2125),
    [anon_sym_or] = ACTIONS(2125),
    [anon_sym_PIPE_PIPE] = ACTIONS(2125),
    [sym_assign_operator] = ACTIONS(2127),
    [sym_let_operator] = ACTIONS(2127),
    [sym__uppercase_identifier] = ACTIONS(2127),
    [aux_sym_directive_token1] = ACTIONS(2125),
    [aux_sym_tag_token1] = ACTIONS(2127),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1485)] = {
    [sym_attribute] = STATE(1485),
    [ts_builtin_sym_end] = ACTIONS(2089),
    [sym__lowercase_identifier] = ACTIONS(2087),
    [anon_sym_SEMI_SEMI] = ACTIONS(2089),
    [anon_sym_let] = ACTIONS(2087),
    [anon_sym_and] = ACTIONS(2087),
    [anon_sym_TILDE] = ACTIONS(2087),
    [anon_sym_QMARK] = ACTIONS(2087),
    [anon_sym_LPAREN] = ACTIONS(2089),
    [anon_sym_external] = ACTIONS(2087),
    [anon_sym_type] = ACTIONS(2087),
    [anon_sym_COMMA] = ACTIONS(2089),
    [anon_sym_PIPE] = ACTIONS(2087),
    [anon_sym_DOT] = ACTIONS(2089),
    [anon_sym_LBRACK] = ACTIONS(2087),
    [anon_sym_true] = ACTIONS(2087),
    [anon_sym_false] = ACTIONS(2087),
    [anon_sym_LBRACE] = ACTIONS(2087),
    [anon_sym_SEMI] = ACTIONS(2087),
    [anon_sym_exception] = ACTIONS(2087),
    [anon_sym_module] = ACTIONS(2087),
    [anon_sym_open] = ACTIONS(2087),
    [anon_sym_include] = ACTIONS(2087),
    [anon_sym_class] = ACTIONS(2087),
    [anon_sym_val] = ACTIONS(2087),
    [anon_sym_object] = ACTIONS(2087),
    [anon_sym_AMP] = ACTIONS(2087),
    [anon_sym_POUND] = ACTIONS(2087),
    [anon_sym_COLON_COLON] = ACTIONS(2089),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2089),
    [anon_sym_else] = ACTIONS(2087),
    [anon_sym_new] = ACTIONS(2087),
    [anon_sym_LBRACE_LT] = ACTIONS(2089),
    [anon_sym_begin] = ACTIONS(2087),
    [sym_ocamlyacc_value] = ACTIONS(2089),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2087),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2089),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2087),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2087),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2089),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2089),
    [aux_sym_number_token1] = ACTIONS(2089),
    [aux_sym_signed_number_token1] = ACTIONS(2087),
    [anon_sym_SQUOTE] = ACTIONS(2089),
    [anon_sym_DQUOTE] = ACTIONS(2089),
    [sym_prefix_operator] = ACTIONS(2089),
    [aux_sym_sign_operator_token1] = ACTIONS(2087),
    [sym_hash_operator] = ACTIONS(2089),
    [anon_sym_lsl] = ACTIONS(2087),
    [anon_sym_lsr] = ACTIONS(2087),
    [anon_sym_asr] = ACTIONS(2087),
    [aux_sym_pow_operator_token1] = ACTIONS(2089),
    [anon_sym_mod] = ACTIONS(2087),
    [anon_sym_land] = ACTIONS(2087),
    [anon_sym_lor] = ACTIONS(2087),
    [anon_sym_lxor] = ACTIONS(2087),
    [aux_sym_mult_operator_token1] = ACTIONS(2087),
    [aux_sym_add_operator_token1] = ACTIONS(2087),
    [sym_concat_operator] = ACTIONS(2089),
    [sym_rel_operator] = ACTIONS(2087),
    [anon_sym_AMP_AMP] = ACTIONS(2087),
    [anon_sym_or] = ACTIONS(2087),
    [anon_sym_PIPE_PIPE] = ACTIONS(2087),
    [sym_assign_operator] = ACTIONS(2089),
    [sym_let_operator] = ACTIONS(2089),
    [sym__uppercase_identifier] = ACTIONS(2089),
    [aux_sym_directive_token1] = ACTIONS(2087),
    [aux_sym_tag_token1] = ACTIONS(2089),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1486)] = {
    [sym_attribute] = STATE(1486),
    [ts_builtin_sym_end] = ACTIONS(1933),
    [sym__lowercase_identifier] = ACTIONS(1931),
    [anon_sym_SEMI_SEMI] = ACTIONS(1933),
    [anon_sym_let] = ACTIONS(1931),
    [anon_sym_and] = ACTIONS(1931),
    [anon_sym_TILDE] = ACTIONS(1931),
    [anon_sym_QMARK] = ACTIONS(1931),
    [anon_sym_LPAREN] = ACTIONS(1933),
    [anon_sym_external] = ACTIONS(1931),
    [anon_sym_type] = ACTIONS(1931),
    [anon_sym_COMMA] = ACTIONS(1933),
    [anon_sym_PIPE] = ACTIONS(1931),
    [anon_sym_DOT] = ACTIONS(1157),
    [anon_sym_LBRACK] = ACTIONS(1931),
    [anon_sym_true] = ACTIONS(1931),
    [anon_sym_false] = ACTIONS(1931),
    [anon_sym_LBRACE] = ACTIONS(1931),
    [anon_sym_SEMI] = ACTIONS(1931),
    [anon_sym_exception] = ACTIONS(1931),
    [anon_sym_module] = ACTIONS(1931),
    [anon_sym_open] = ACTIONS(1931),
    [anon_sym_include] = ACTIONS(1931),
    [anon_sym_class] = ACTIONS(1931),
    [anon_sym_val] = ACTIONS(1931),
    [anon_sym_object] = ACTIONS(1931),
    [anon_sym_AMP] = ACTIONS(1931),
    [anon_sym_POUND] = ACTIONS(1159),
    [anon_sym_COLON_COLON] = ACTIONS(1933),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1933),
    [anon_sym_else] = ACTIONS(1931),
    [anon_sym_new] = ACTIONS(1931),
    [anon_sym_LBRACE_LT] = ACTIONS(1933),
    [anon_sym_begin] = ACTIONS(1931),
    [sym_ocamlyacc_value] = ACTIONS(1933),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1931),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1933),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1933),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1933),
    [aux_sym_number_token1] = ACTIONS(1933),
    [aux_sym_signed_number_token1] = ACTIONS(1931),
    [anon_sym_SQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1933),
    [sym_prefix_operator] = ACTIONS(1933),
    [aux_sym_sign_operator_token1] = ACTIONS(1931),
    [sym_hash_operator] = ACTIONS(1161),
    [anon_sym_lsl] = ACTIONS(1931),
    [anon_sym_lsr] = ACTIONS(1931),
    [anon_sym_asr] = ACTIONS(1931),
    [aux_sym_pow_operator_token1] = ACTIONS(1933),
    [anon_sym_mod] = ACTIONS(1931),
    [anon_sym_land] = ACTIONS(1931),
    [anon_sym_lor] = ACTIONS(1931),
    [anon_sym_lxor] = ACTIONS(1931),
    [aux_sym_mult_operator_token1] = ACTIONS(1931),
    [aux_sym_add_operator_token1] = ACTIONS(1931),
    [sym_concat_operator] = ACTIONS(1933),
    [sym_rel_operator] = ACTIONS(1931),
    [anon_sym_AMP_AMP] = ACTIONS(1931),
    [anon_sym_or] = ACTIONS(1931),
    [anon_sym_PIPE_PIPE] = ACTIONS(1931),
    [sym_assign_operator] = ACTIONS(1933),
    [sym_let_operator] = ACTIONS(1933),
    [sym__uppercase_identifier] = ACTIONS(1933),
    [aux_sym_directive_token1] = ACTIONS(1931),
    [aux_sym_tag_token1] = ACTIONS(1933),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1487)] = {
    [sym_attribute] = STATE(1487),
    [ts_builtin_sym_end] = ACTIONS(2037),
    [sym__lowercase_identifier] = ACTIONS(2035),
    [anon_sym_SEMI_SEMI] = ACTIONS(2037),
    [anon_sym_let] = ACTIONS(2035),
    [anon_sym_and] = ACTIONS(2035),
    [anon_sym_TILDE] = ACTIONS(2035),
    [anon_sym_QMARK] = ACTIONS(2035),
    [anon_sym_LPAREN] = ACTIONS(2037),
    [anon_sym_external] = ACTIONS(2035),
    [anon_sym_type] = ACTIONS(2035),
    [anon_sym_COMMA] = ACTIONS(2037),
    [anon_sym_PIPE] = ACTIONS(2035),
    [anon_sym_DOT] = ACTIONS(2037),
    [anon_sym_LBRACK] = ACTIONS(2035),
    [anon_sym_true] = ACTIONS(2035),
    [anon_sym_false] = ACTIONS(2035),
    [anon_sym_LBRACE] = ACTIONS(2035),
    [anon_sym_SEMI] = ACTIONS(2035),
    [anon_sym_exception] = ACTIONS(2035),
    [anon_sym_module] = ACTIONS(2035),
    [anon_sym_open] = ACTIONS(2035),
    [anon_sym_include] = ACTIONS(2035),
    [anon_sym_class] = ACTIONS(2035),
    [anon_sym_val] = ACTIONS(2035),
    [anon_sym_object] = ACTIONS(2035),
    [anon_sym_AMP] = ACTIONS(2035),
    [anon_sym_POUND] = ACTIONS(2035),
    [anon_sym_COLON_COLON] = ACTIONS(2037),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2037),
    [anon_sym_else] = ACTIONS(2035),
    [anon_sym_new] = ACTIONS(2035),
    [anon_sym_LBRACE_LT] = ACTIONS(2037),
    [anon_sym_begin] = ACTIONS(2035),
    [sym_ocamlyacc_value] = ACTIONS(2037),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2035),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2037),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2035),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2035),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2037),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2037),
    [aux_sym_number_token1] = ACTIONS(2037),
    [aux_sym_signed_number_token1] = ACTIONS(2035),
    [anon_sym_SQUOTE] = ACTIONS(2037),
    [anon_sym_DQUOTE] = ACTIONS(2037),
    [sym_prefix_operator] = ACTIONS(2037),
    [aux_sym_sign_operator_token1] = ACTIONS(2035),
    [sym_hash_operator] = ACTIONS(2037),
    [anon_sym_lsl] = ACTIONS(2035),
    [anon_sym_lsr] = ACTIONS(2035),
    [anon_sym_asr] = ACTIONS(2035),
    [aux_sym_pow_operator_token1] = ACTIONS(2037),
    [anon_sym_mod] = ACTIONS(2035),
    [anon_sym_land] = ACTIONS(2035),
    [anon_sym_lor] = ACTIONS(2035),
    [anon_sym_lxor] = ACTIONS(2035),
    [aux_sym_mult_operator_token1] = ACTIONS(2035),
    [aux_sym_add_operator_token1] = ACTIONS(2035),
    [sym_concat_operator] = ACTIONS(2037),
    [sym_rel_operator] = ACTIONS(2035),
    [anon_sym_AMP_AMP] = ACTIONS(2035),
    [anon_sym_or] = ACTIONS(2035),
    [anon_sym_PIPE_PIPE] = ACTIONS(2035),
    [sym_assign_operator] = ACTIONS(2037),
    [sym_let_operator] = ACTIONS(2037),
    [sym__uppercase_identifier] = ACTIONS(2037),
    [aux_sym_directive_token1] = ACTIONS(2035),
    [aux_sym_tag_token1] = ACTIONS(2037),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1488)] = {
    [sym_attribute] = STATE(1488),
    [ts_builtin_sym_end] = ACTIONS(2193),
    [sym__lowercase_identifier] = ACTIONS(2191),
    [anon_sym_SEMI_SEMI] = ACTIONS(2193),
    [anon_sym_let] = ACTIONS(2191),
    [anon_sym_and] = ACTIONS(2191),
    [anon_sym_TILDE] = ACTIONS(2191),
    [anon_sym_QMARK] = ACTIONS(2191),
    [anon_sym_LPAREN] = ACTIONS(2193),
    [anon_sym_external] = ACTIONS(2191),
    [anon_sym_type] = ACTIONS(2191),
    [anon_sym_COMMA] = ACTIONS(2193),
    [anon_sym_PIPE] = ACTIONS(2191),
    [anon_sym_DOT] = ACTIONS(2193),
    [anon_sym_LBRACK] = ACTIONS(2191),
    [anon_sym_true] = ACTIONS(2191),
    [anon_sym_false] = ACTIONS(2191),
    [anon_sym_LBRACE] = ACTIONS(2191),
    [anon_sym_SEMI] = ACTIONS(2191),
    [anon_sym_exception] = ACTIONS(2191),
    [anon_sym_module] = ACTIONS(2191),
    [anon_sym_open] = ACTIONS(2191),
    [anon_sym_include] = ACTIONS(2191),
    [anon_sym_class] = ACTIONS(2191),
    [anon_sym_val] = ACTIONS(2191),
    [anon_sym_object] = ACTIONS(2191),
    [anon_sym_AMP] = ACTIONS(2191),
    [anon_sym_POUND] = ACTIONS(2191),
    [anon_sym_COLON_COLON] = ACTIONS(2193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2193),
    [anon_sym_else] = ACTIONS(2191),
    [anon_sym_new] = ACTIONS(2191),
    [anon_sym_LBRACE_LT] = ACTIONS(2193),
    [anon_sym_begin] = ACTIONS(2191),
    [sym_ocamlyacc_value] = ACTIONS(2193),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2191),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2193),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2191),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2191),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2193),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2193),
    [aux_sym_number_token1] = ACTIONS(2193),
    [aux_sym_signed_number_token1] = ACTIONS(2191),
    [anon_sym_SQUOTE] = ACTIONS(2193),
    [anon_sym_DQUOTE] = ACTIONS(2193),
    [sym_prefix_operator] = ACTIONS(2193),
    [aux_sym_sign_operator_token1] = ACTIONS(2191),
    [sym_hash_operator] = ACTIONS(2193),
    [anon_sym_lsl] = ACTIONS(2191),
    [anon_sym_lsr] = ACTIONS(2191),
    [anon_sym_asr] = ACTIONS(2191),
    [aux_sym_pow_operator_token1] = ACTIONS(2193),
    [anon_sym_mod] = ACTIONS(2191),
    [anon_sym_land] = ACTIONS(2191),
    [anon_sym_lor] = ACTIONS(2191),
    [anon_sym_lxor] = ACTIONS(2191),
    [aux_sym_mult_operator_token1] = ACTIONS(2191),
    [aux_sym_add_operator_token1] = ACTIONS(2191),
    [sym_concat_operator] = ACTIONS(2193),
    [sym_rel_operator] = ACTIONS(2191),
    [anon_sym_AMP_AMP] = ACTIONS(2191),
    [anon_sym_or] = ACTIONS(2191),
    [anon_sym_PIPE_PIPE] = ACTIONS(2191),
    [sym_assign_operator] = ACTIONS(2193),
    [sym_let_operator] = ACTIONS(2193),
    [sym__uppercase_identifier] = ACTIONS(2193),
    [aux_sym_directive_token1] = ACTIONS(2191),
    [aux_sym_tag_token1] = ACTIONS(2193),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1489)] = {
    [sym_attribute] = STATE(1489),
    [ts_builtin_sym_end] = ACTIONS(1961),
    [sym__lowercase_identifier] = ACTIONS(1959),
    [anon_sym_SEMI_SEMI] = ACTIONS(1961),
    [anon_sym_let] = ACTIONS(1959),
    [anon_sym_and] = ACTIONS(1959),
    [anon_sym_TILDE] = ACTIONS(1959),
    [anon_sym_QMARK] = ACTIONS(1959),
    [anon_sym_LPAREN] = ACTIONS(1961),
    [anon_sym_external] = ACTIONS(1959),
    [anon_sym_type] = ACTIONS(1959),
    [anon_sym_COMMA] = ACTIONS(1961),
    [anon_sym_PIPE] = ACTIONS(1959),
    [anon_sym_DOT] = ACTIONS(1961),
    [anon_sym_LBRACK] = ACTIONS(1959),
    [anon_sym_true] = ACTIONS(1959),
    [anon_sym_false] = ACTIONS(1959),
    [anon_sym_LBRACE] = ACTIONS(1959),
    [anon_sym_SEMI] = ACTIONS(1959),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_module] = ACTIONS(1959),
    [anon_sym_open] = ACTIONS(1959),
    [anon_sym_include] = ACTIONS(1959),
    [anon_sym_class] = ACTIONS(1959),
    [anon_sym_val] = ACTIONS(1959),
    [anon_sym_object] = ACTIONS(1959),
    [anon_sym_AMP] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1959),
    [anon_sym_COLON_COLON] = ACTIONS(1961),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1961),
    [anon_sym_else] = ACTIONS(1959),
    [anon_sym_new] = ACTIONS(1959),
    [anon_sym_LBRACE_LT] = ACTIONS(1961),
    [anon_sym_begin] = ACTIONS(1959),
    [sym_ocamlyacc_value] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1959),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1961),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1959),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1959),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1961),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1961),
    [aux_sym_number_token1] = ACTIONS(1961),
    [aux_sym_signed_number_token1] = ACTIONS(1959),
    [anon_sym_SQUOTE] = ACTIONS(1961),
    [anon_sym_DQUOTE] = ACTIONS(1961),
    [sym_prefix_operator] = ACTIONS(1961),
    [aux_sym_sign_operator_token1] = ACTIONS(1959),
    [sym_hash_operator] = ACTIONS(1961),
    [anon_sym_lsl] = ACTIONS(1959),
    [anon_sym_lsr] = ACTIONS(1959),
    [anon_sym_asr] = ACTIONS(1959),
    [aux_sym_pow_operator_token1] = ACTIONS(1961),
    [anon_sym_mod] = ACTIONS(1959),
    [anon_sym_land] = ACTIONS(1959),
    [anon_sym_lor] = ACTIONS(1959),
    [anon_sym_lxor] = ACTIONS(1959),
    [aux_sym_mult_operator_token1] = ACTIONS(1959),
    [aux_sym_add_operator_token1] = ACTIONS(1959),
    [sym_concat_operator] = ACTIONS(1961),
    [sym_rel_operator] = ACTIONS(1959),
    [anon_sym_AMP_AMP] = ACTIONS(1959),
    [anon_sym_or] = ACTIONS(1959),
    [anon_sym_PIPE_PIPE] = ACTIONS(1959),
    [sym_assign_operator] = ACTIONS(1961),
    [sym_let_operator] = ACTIONS(1961),
    [sym__uppercase_identifier] = ACTIONS(1961),
    [aux_sym_directive_token1] = ACTIONS(1959),
    [aux_sym_tag_token1] = ACTIONS(1961),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1490)] = {
    [sym_attribute] = STATE(1490),
    [ts_builtin_sym_end] = ACTIONS(1933),
    [sym__lowercase_identifier] = ACTIONS(1931),
    [anon_sym_SEMI_SEMI] = ACTIONS(1933),
    [anon_sym_let] = ACTIONS(1931),
    [anon_sym_and] = ACTIONS(1931),
    [anon_sym_TILDE] = ACTIONS(1931),
    [anon_sym_QMARK] = ACTIONS(1931),
    [anon_sym_LPAREN] = ACTIONS(1933),
    [anon_sym_external] = ACTIONS(1931),
    [anon_sym_type] = ACTIONS(1931),
    [anon_sym_COMMA] = ACTIONS(1933),
    [anon_sym_PIPE] = ACTIONS(1931),
    [anon_sym_LBRACK] = ACTIONS(1931),
    [anon_sym_RBRACK] = ACTIONS(1933),
    [anon_sym_true] = ACTIONS(1931),
    [anon_sym_false] = ACTIONS(1931),
    [anon_sym_LBRACE] = ACTIONS(1931),
    [anon_sym_SEMI] = ACTIONS(1931),
    [anon_sym_exception] = ACTIONS(1931),
    [anon_sym_module] = ACTIONS(1931),
    [anon_sym_open] = ACTIONS(1931),
    [anon_sym_include] = ACTIONS(1931),
    [anon_sym_class] = ACTIONS(1931),
    [anon_sym_val] = ACTIONS(1931),
    [anon_sym_end] = ACTIONS(1931),
    [anon_sym_object] = ACTIONS(1931),
    [anon_sym_AMP] = ACTIONS(1931),
    [anon_sym_COLON_COLON] = ACTIONS(1933),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1933),
    [anon_sym_else] = ACTIONS(1931),
    [anon_sym_new] = ACTIONS(1931),
    [anon_sym_LBRACE_LT] = ACTIONS(1933),
    [anon_sym_begin] = ACTIONS(1931),
    [sym_ocamlyacc_value] = ACTIONS(1933),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1931),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1933),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1933),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1933),
    [aux_sym_number_token1] = ACTIONS(1933),
    [aux_sym_signed_number_token1] = ACTIONS(1931),
    [anon_sym_SQUOTE] = ACTIONS(1933),
    [anon_sym_DQUOTE] = ACTIONS(1933),
    [sym_prefix_operator] = ACTIONS(1933),
    [aux_sym_sign_operator_token1] = ACTIONS(1931),
    [anon_sym_lsl] = ACTIONS(1931),
    [anon_sym_lsr] = ACTIONS(1931),
    [anon_sym_asr] = ACTIONS(1931),
    [aux_sym_pow_operator_token1] = ACTIONS(1933),
    [anon_sym_mod] = ACTIONS(1931),
    [anon_sym_land] = ACTIONS(1931),
    [anon_sym_lor] = ACTIONS(1931),
    [anon_sym_lxor] = ACTIONS(1931),
    [aux_sym_mult_operator_token1] = ACTIONS(1931),
    [aux_sym_add_operator_token1] = ACTIONS(1931),
    [sym_concat_operator] = ACTIONS(1933),
    [sym_rel_operator] = ACTIONS(1931),
    [anon_sym_AMP_AMP] = ACTIONS(1931),
    [anon_sym_or] = ACTIONS(1931),
    [anon_sym_PIPE_PIPE] = ACTIONS(1931),
    [sym_assign_operator] = ACTIONS(1933),
    [sym_let_operator] = ACTIONS(1933),
    [sym__uppercase_identifier] = ACTIONS(1933),
    [aux_sym_directive_token1] = ACTIONS(1933),
    [aux_sym_tag_token1] = ACTIONS(1933),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1491)] = {
    [sym_attribute] = STATE(1491),
    [sym__lowercase_identifier] = ACTIONS(2311),
    [anon_sym_SEMI_SEMI] = ACTIONS(2315),
    [anon_sym_let] = ACTIONS(2311),
    [anon_sym_and] = ACTIONS(2311),
    [anon_sym_TILDE] = ACTIONS(2311),
    [anon_sym_QMARK] = ACTIONS(2311),
    [anon_sym_COLON] = ACTIONS(2693),
    [anon_sym_LPAREN] = ACTIONS(2315),
    [anon_sym_external] = ACTIONS(2311),
    [anon_sym_type] = ACTIONS(2311),
    [anon_sym_COMMA] = ACTIONS(2315),
    [anon_sym_PIPE] = ACTIONS(2311),
    [anon_sym_LBRACK] = ACTIONS(2311),
    [anon_sym_RBRACK] = ACTIONS(2315),
    [anon_sym_true] = ACTIONS(2311),
    [anon_sym_false] = ACTIONS(2311),
    [anon_sym_LBRACE] = ACTIONS(2311),
    [anon_sym_SEMI] = ACTIONS(2311),
    [anon_sym_exception] = ACTIONS(2311),
    [anon_sym_module] = ACTIONS(2311),
    [anon_sym_open] = ACTIONS(2311),
    [anon_sym_include] = ACTIONS(2311),
    [anon_sym_class] = ACTIONS(2311),
    [anon_sym_val] = ACTIONS(2311),
    [anon_sym_end] = ACTIONS(2311),
    [anon_sym_object] = ACTIONS(2311),
    [anon_sym_AMP] = ACTIONS(2311),
    [anon_sym_COLON_COLON] = ACTIONS(2315),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2315),
    [anon_sym_else] = ACTIONS(2311),
    [anon_sym_new] = ACTIONS(2311),
    [anon_sym_LBRACE_LT] = ACTIONS(2315),
    [anon_sym_begin] = ACTIONS(2311),
    [sym_ocamlyacc_value] = ACTIONS(2315),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2311),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2315),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2311),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2311),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2315),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2315),
    [aux_sym_number_token1] = ACTIONS(2315),
    [aux_sym_signed_number_token1] = ACTIONS(2311),
    [anon_sym_SQUOTE] = ACTIONS(2315),
    [anon_sym_DQUOTE] = ACTIONS(2315),
    [sym_prefix_operator] = ACTIONS(2315),
    [aux_sym_sign_operator_token1] = ACTIONS(2311),
    [anon_sym_lsl] = ACTIONS(2311),
    [anon_sym_lsr] = ACTIONS(2311),
    [anon_sym_asr] = ACTIONS(2311),
    [aux_sym_pow_operator_token1] = ACTIONS(2315),
    [anon_sym_mod] = ACTIONS(2311),
    [anon_sym_land] = ACTIONS(2311),
    [anon_sym_lor] = ACTIONS(2311),
    [anon_sym_lxor] = ACTIONS(2311),
    [aux_sym_mult_operator_token1] = ACTIONS(2311),
    [aux_sym_add_operator_token1] = ACTIONS(2311),
    [sym_concat_operator] = ACTIONS(2315),
    [sym_rel_operator] = ACTIONS(2311),
    [anon_sym_AMP_AMP] = ACTIONS(2311),
    [anon_sym_or] = ACTIONS(2311),
    [anon_sym_PIPE_PIPE] = ACTIONS(2311),
    [sym_assign_operator] = ACTIONS(2315),
    [sym_let_operator] = ACTIONS(2315),
    [sym__uppercase_identifier] = ACTIONS(2315),
    [aux_sym_directive_token1] = ACTIONS(2315),
    [aux_sym_tag_token1] = ACTIONS(2315),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1492)] = {
    [sym__match_cases] = STATE(2716),
    [sym_match_case] = STATE(3935),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5529),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1492),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1568),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2695),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1493)] = {
    [sym__match_cases] = STATE(2716),
    [sym_match_case] = STATE(3233),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5417),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1493),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1575),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2697),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1494)] = {
    [sym__match_cases] = STATE(2635),
    [sym_match_case] = STATE(2833),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5423),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1494),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1540),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2699),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1495)] = {
    [sym__match_cases] = STATE(2716),
    [sym_match_case] = STATE(3222),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5411),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1495),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1554),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2701),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1496)] = {
    [sym_attribute] = STATE(1496),
    [ts_builtin_sym_end] = ACTIONS(2417),
    [sym__lowercase_identifier] = ACTIONS(2415),
    [anon_sym_SEMI_SEMI] = ACTIONS(2417),
    [anon_sym_let] = ACTIONS(2415),
    [anon_sym_and] = ACTIONS(2415),
    [anon_sym_TILDE] = ACTIONS(2415),
    [anon_sym_QMARK] = ACTIONS(2415),
    [anon_sym_LPAREN] = ACTIONS(2417),
    [anon_sym_external] = ACTIONS(2415),
    [anon_sym_type] = ACTIONS(2415),
    [anon_sym_COMMA] = ACTIONS(2417),
    [anon_sym_PIPE] = ACTIONS(2415),
    [anon_sym_LBRACK] = ACTIONS(2415),
    [anon_sym_RBRACK] = ACTIONS(2417),
    [anon_sym_true] = ACTIONS(2415),
    [anon_sym_false] = ACTIONS(2415),
    [anon_sym_LBRACE] = ACTIONS(2415),
    [anon_sym_SEMI] = ACTIONS(2415),
    [anon_sym_exception] = ACTIONS(2415),
    [anon_sym_module] = ACTIONS(2415),
    [anon_sym_open] = ACTIONS(2415),
    [anon_sym_include] = ACTIONS(2415),
    [anon_sym_class] = ACTIONS(2415),
    [anon_sym_val] = ACTIONS(2415),
    [anon_sym_end] = ACTIONS(2415),
    [anon_sym_object] = ACTIONS(2415),
    [anon_sym_AMP] = ACTIONS(2415),
    [anon_sym_COLON_COLON] = ACTIONS(2417),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2417),
    [anon_sym_else] = ACTIONS(2415),
    [anon_sym_new] = ACTIONS(2415),
    [anon_sym_LBRACE_LT] = ACTIONS(2417),
    [anon_sym_begin] = ACTIONS(2415),
    [sym_ocamlyacc_value] = ACTIONS(2417),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2415),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2417),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2415),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2417),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2417),
    [aux_sym_number_token1] = ACTIONS(2417),
    [aux_sym_signed_number_token1] = ACTIONS(2415),
    [anon_sym_SQUOTE] = ACTIONS(2417),
    [anon_sym_DQUOTE] = ACTIONS(2417),
    [sym_prefix_operator] = ACTIONS(2417),
    [aux_sym_sign_operator_token1] = ACTIONS(2415),
    [anon_sym_lsl] = ACTIONS(2415),
    [anon_sym_lsr] = ACTIONS(2415),
    [anon_sym_asr] = ACTIONS(2415),
    [aux_sym_pow_operator_token1] = ACTIONS(2417),
    [anon_sym_mod] = ACTIONS(2415),
    [anon_sym_land] = ACTIONS(2415),
    [anon_sym_lor] = ACTIONS(2415),
    [anon_sym_lxor] = ACTIONS(2415),
    [aux_sym_mult_operator_token1] = ACTIONS(2415),
    [aux_sym_add_operator_token1] = ACTIONS(2415),
    [sym_concat_operator] = ACTIONS(2417),
    [sym_rel_operator] = ACTIONS(2415),
    [anon_sym_AMP_AMP] = ACTIONS(2415),
    [anon_sym_or] = ACTIONS(2415),
    [anon_sym_PIPE_PIPE] = ACTIONS(2415),
    [sym_assign_operator] = ACTIONS(2417),
    [sym_let_operator] = ACTIONS(2417),
    [sym__uppercase_identifier] = ACTIONS(2417),
    [aux_sym_directive_token1] = ACTIONS(2417),
    [aux_sym_tag_token1] = ACTIONS(2417),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1497)] = {
    [sym__match_cases] = STATE(2635),
    [sym_match_case] = STATE(2735),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5489),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1497),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1560),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2703),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1498)] = {
    [sym__match_cases] = STATE(3533),
    [sym_match_case] = STATE(3811),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5534),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1498),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1567),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2705),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1499)] = {
    [sym__match_cases] = STATE(3533),
    [sym_match_case] = STATE(3789),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5407),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1499),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1514),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2707),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1500)] = {
    [sym__match_cases] = STATE(2635),
    [sym_match_case] = STATE(2724),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5528),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1500),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1548),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2709),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1501)] = {
    [sym__match_cases] = STATE(2827),
    [sym_match_case] = STATE(3233),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5417),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1501),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1522),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2697),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1502)] = {
    [sym__match_cases] = STATE(2827),
    [sym_match_case] = STATE(3222),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5411),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1502),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1533),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2701),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1503)] = {
    [sym__match_cases] = STATE(2635),
    [sym_match_case] = STATE(2691),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5469),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1503),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1549),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2711),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1504)] = {
    [sym__match_cases] = STATE(2635),
    [sym_match_case] = STATE(2879),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5522),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1504),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1519),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1505)] = {
    [sym__match_cases] = STATE(2635),
    [sym_match_case] = STATE(2896),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5450),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1505),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1532),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2715),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1506)] = {
    [sym_parameter] = STATE(2862),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1506),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(8318),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(2866),
    [aux_sym_let_binding_repeat1] = STATE(1506),
    [sym__lowercase_identifier] = ACTIONS(2640),
    [anon_sym_and] = ACTIONS(2468),
    [anon_sym_EQ] = ACTIONS(2463),
    [anon_sym_TILDE] = ACTIONS(2717),
    [anon_sym_QMARK] = ACTIONS(2717),
    [anon_sym_LPAREN] = ACTIONS(2720),
    [anon_sym_COLON2] = ACTIONS(2468),
    [anon_sym_DASH_GT] = ACTIONS(2463),
    [anon_sym_LBRACK] = ACTIONS(2649),
    [anon_sym_true] = ACTIONS(2652),
    [anon_sym_false] = ACTIONS(2652),
    [anon_sym_LBRACE] = ACTIONS(2655),
    [anon_sym_constraint] = ACTIONS(2468),
    [anon_sym_val] = ACTIONS(2468),
    [anon_sym_end] = ACTIONS(2468),
    [anon_sym_COLON_GT] = ACTIONS(2463),
    [anon_sym_inherit] = ACTIONS(2468),
    [anon_sym_method] = ACTIONS(2468),
    [anon_sym_in] = ACTIONS(2468),
    [anon_sym_initializer] = ACTIONS(2468),
    [anon_sym_POUND] = ACTIONS(2658),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2661),
    [anon_sym_begin] = ACTIONS(2664),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2468),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2463),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2667),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2670),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2463),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2463),
    [aux_sym_number_token1] = ACTIONS(2673),
    [aux_sym_signed_number_token1] = ACTIONS(2723),
    [anon_sym_SQUOTE] = ACTIONS(2679),
    [anon_sym_DQUOTE] = ACTIONS(2682),
    [sym__uppercase_identifier] = ACTIONS(2726),
    [aux_sym_tag_token1] = ACTIONS(2729),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1507)] = {
    [sym_attribute] = STATE(1507),
    [ts_builtin_sym_end] = ACTIONS(2315),
    [sym__lowercase_identifier] = ACTIONS(2311),
    [anon_sym_SEMI_SEMI] = ACTIONS(2315),
    [anon_sym_let] = ACTIONS(2311),
    [anon_sym_and] = ACTIONS(2311),
    [anon_sym_TILDE] = ACTIONS(2311),
    [anon_sym_QMARK] = ACTIONS(2311),
    [anon_sym_COLON] = ACTIONS(2732),
    [anon_sym_LPAREN] = ACTIONS(2315),
    [anon_sym_external] = ACTIONS(2311),
    [anon_sym_type] = ACTIONS(2311),
    [anon_sym_COMMA] = ACTIONS(2315),
    [anon_sym_PIPE] = ACTIONS(2311),
    [anon_sym_LBRACK] = ACTIONS(2311),
    [anon_sym_true] = ACTIONS(2311),
    [anon_sym_false] = ACTIONS(2311),
    [anon_sym_LBRACE] = ACTIONS(2311),
    [anon_sym_SEMI] = ACTIONS(2311),
    [anon_sym_exception] = ACTIONS(2311),
    [anon_sym_module] = ACTIONS(2311),
    [anon_sym_open] = ACTIONS(2311),
    [anon_sym_include] = ACTIONS(2311),
    [anon_sym_class] = ACTIONS(2311),
    [anon_sym_val] = ACTIONS(2311),
    [anon_sym_object] = ACTIONS(2311),
    [anon_sym_AMP] = ACTIONS(2311),
    [anon_sym_COLON_COLON] = ACTIONS(2315),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2315),
    [anon_sym_else] = ACTIONS(2311),
    [anon_sym_new] = ACTIONS(2311),
    [anon_sym_LBRACE_LT] = ACTIONS(2315),
    [anon_sym_begin] = ACTIONS(2311),
    [sym_ocamlyacc_value] = ACTIONS(2315),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2311),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2315),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2311),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2311),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2315),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2315),
    [aux_sym_number_token1] = ACTIONS(2315),
    [aux_sym_signed_number_token1] = ACTIONS(2311),
    [anon_sym_SQUOTE] = ACTIONS(2315),
    [anon_sym_DQUOTE] = ACTIONS(2315),
    [sym_prefix_operator] = ACTIONS(2315),
    [aux_sym_sign_operator_token1] = ACTIONS(2311),
    [anon_sym_lsl] = ACTIONS(2311),
    [anon_sym_lsr] = ACTIONS(2311),
    [anon_sym_asr] = ACTIONS(2311),
    [aux_sym_pow_operator_token1] = ACTIONS(2315),
    [anon_sym_mod] = ACTIONS(2311),
    [anon_sym_land] = ACTIONS(2311),
    [anon_sym_lor] = ACTIONS(2311),
    [anon_sym_lxor] = ACTIONS(2311),
    [aux_sym_mult_operator_token1] = ACTIONS(2311),
    [aux_sym_add_operator_token1] = ACTIONS(2311),
    [sym_concat_operator] = ACTIONS(2315),
    [sym_rel_operator] = ACTIONS(2311),
    [anon_sym_AMP_AMP] = ACTIONS(2311),
    [anon_sym_or] = ACTIONS(2311),
    [anon_sym_PIPE_PIPE] = ACTIONS(2311),
    [sym_assign_operator] = ACTIONS(2315),
    [sym_let_operator] = ACTIONS(2315),
    [sym__uppercase_identifier] = ACTIONS(2315),
    [aux_sym_directive_token1] = ACTIONS(2315),
    [aux_sym_tag_token1] = ACTIONS(2315),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1508)] = {
    [sym__match_cases] = STATE(2649),
    [sym_match_case] = STATE(2896),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5450),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1508),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2715),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1509)] = {
    [sym__match_cases] = STATE(2709),
    [sym_match_case] = STATE(3233),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5417),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1509),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2697),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1510)] = {
    [sym__match_cases] = STATE(2712),
    [sym_match_case] = STATE(3233),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5417),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1510),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2697),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1511)] = {
    [sym__match_cases] = STATE(2712),
    [sym_match_case] = STATE(3222),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5411),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1511),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2701),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1512)] = {
    [sym__match_cases] = STATE(2748),
    [sym_match_case] = STATE(3233),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5417),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1512),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2697),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1513)] = {
    [sym__match_cases] = STATE(2706),
    [sym_match_case] = STATE(3233),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5417),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1513),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2697),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1514)] = {
    [sym__match_cases] = STATE(3554),
    [sym_match_case] = STATE(3789),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5407),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1514),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2707),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1515)] = {
    [sym__match_cases] = STATE(3574),
    [sym_match_case] = STATE(3789),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5407),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1515),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2707),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1516)] = {
    [sym__match_cases] = STATE(3599),
    [sym_match_case] = STATE(3789),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5407),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1516),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2707),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1517)] = {
    [sym__match_cases] = STATE(3505),
    [sym_match_case] = STATE(3789),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5407),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1517),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2707),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1518)] = {
    [sym__match_cases] = STATE(3507),
    [sym_match_case] = STATE(3789),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5407),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1518),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2707),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1519)] = {
    [sym__match_cases] = STATE(2698),
    [sym_match_case] = STATE(2879),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5522),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1519),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1520)] = {
    [sym__match_cases] = STATE(2644),
    [sym_match_case] = STATE(2879),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5522),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1520),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1521)] = {
    [sym__match_cases] = STATE(2649),
    [sym_match_case] = STATE(2879),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5522),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1521),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1522)] = {
    [sym__match_cases] = STATE(2824),
    [sym_match_case] = STATE(3233),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5417),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1522),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2697),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1523)] = {
    [sym__match_cases] = STATE(2683),
    [sym_match_case] = STATE(2879),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5522),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1523),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1524)] = {
    [sym__match_cases] = STATE(2774),
    [sym_match_case] = STATE(3233),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5417),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1524),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2697),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1525)] = {
    [sym__match_cases] = STATE(2775),
    [sym_match_case] = STATE(3233),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5417),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1525),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2697),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1526)] = {
    [sym__match_cases] = STATE(2809),
    [sym_match_case] = STATE(3233),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5417),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1526),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2697),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1527)] = {
    [sym__match_cases] = STATE(2811),
    [sym_match_case] = STATE(3233),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5417),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1527),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2697),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1528)] = {
    [sym__match_cases] = STATE(2685),
    [sym_match_case] = STATE(2879),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5522),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1528),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2713),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1529)] = {
    [sym__match_cases] = STATE(2748),
    [sym_match_case] = STATE(3222),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5411),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1529),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2701),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1530)] = {
    [sym__match_cases] = STATE(2706),
    [sym_match_case] = STATE(3222),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5411),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1530),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2701),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1531)] = {
    [sym__match_cases] = STATE(2644),
    [sym_match_case] = STATE(2896),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5450),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1531),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2715),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1532)] = {
    [sym__match_cases] = STATE(2698),
    [sym_match_case] = STATE(2896),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5450),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1532),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2715),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1533)] = {
    [sym__match_cases] = STATE(2824),
    [sym_match_case] = STATE(3222),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5411),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1533),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2701),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1534)] = {
    [sym_parameter] = STATE(2862),
    [sym__maybe_polymorphic_typed] = STATE(4984),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1534),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(8318),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(5145),
    [aux_sym_let_binding_repeat1] = STATE(1569),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_EQ] = ACTIONS(2734),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_COLON2] = ACTIONS(2736),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_constraint] = ACTIONS(2738),
    [anon_sym_val] = ACTIONS(2738),
    [anon_sym_end] = ACTIONS(2738),
    [anon_sym_inherit] = ACTIONS(2738),
    [anon_sym_method] = ACTIONS(2738),
    [anon_sym_initializer] = ACTIONS(2738),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2740),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2740),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2740),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1535)] = {
    [sym__match_cases] = STATE(2774),
    [sym_match_case] = STATE(3222),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5411),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1535),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2701),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1536)] = {
    [sym__match_cases] = STATE(2775),
    [sym_match_case] = STATE(3222),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5411),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1536),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2701),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1537)] = {
    [sym_parameter] = STATE(2862),
    [sym__maybe_polymorphic_typed] = STATE(4976),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1537),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(8318),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(5148),
    [aux_sym_let_binding_repeat1] = STATE(1506),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_EQ] = ACTIONS(2742),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_COLON2] = ACTIONS(2736),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_constraint] = ACTIONS(2744),
    [anon_sym_val] = ACTIONS(2744),
    [anon_sym_end] = ACTIONS(2744),
    [anon_sym_inherit] = ACTIONS(2744),
    [anon_sym_method] = ACTIONS(2744),
    [anon_sym_initializer] = ACTIONS(2744),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2746),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2746),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2746),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1538)] = {
    [sym__match_cases] = STATE(2809),
    [sym_match_case] = STATE(3222),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5411),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1538),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2701),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1539)] = {
    [sym__match_cases] = STATE(2811),
    [sym_match_case] = STATE(3222),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5411),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1539),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2701),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1540)] = {
    [sym__match_cases] = STATE(2698),
    [sym_match_case] = STATE(2833),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5423),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1540),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2699),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1541)] = {
    [sym__match_cases] = STATE(2644),
    [sym_match_case] = STATE(2833),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5423),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1541),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2699),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1542)] = {
    [sym_parameter] = STATE(2862),
    [sym__maybe_polymorphic_typed] = STATE(5007),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1542),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(8318),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(5056),
    [aux_sym_let_binding_repeat1] = STATE(1537),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_EQ] = ACTIONS(2748),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_COLON2] = ACTIONS(2736),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_constraint] = ACTIONS(2750),
    [anon_sym_val] = ACTIONS(2750),
    [anon_sym_end] = ACTIONS(2750),
    [anon_sym_inherit] = ACTIONS(2750),
    [anon_sym_method] = ACTIONS(2750),
    [anon_sym_initializer] = ACTIONS(2750),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2752),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2752),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2752),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1543)] = {
    [sym__match_cases] = STATE(2683),
    [sym_match_case] = STATE(2896),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5450),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1543),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2715),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1544)] = {
    [sym__match_cases] = STATE(3507),
    [sym_match_case] = STATE(3811),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5534),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1544),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2705),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1545)] = {
    [sym__match_cases] = STATE(2683),
    [sym_match_case] = STATE(2833),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5423),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1545),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2699),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1546)] = {
    [sym__match_cases] = STATE(2685),
    [sym_match_case] = STATE(2833),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5423),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1546),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2699),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1547)] = {
    [sym_parameter] = STATE(2862),
    [sym__coerced] = STATE(7635),
    [sym__maybe_polymorphic_typed] = STATE(6344),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1547),
    [sym_item_attribute] = STATE(6240),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(8318),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(5668),
    [aux_sym_let_binding_repeat1] = STATE(1930),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_and] = ACTIONS(1821),
    [anon_sym_EQ] = ACTIONS(2754),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_COMMA] = ACTIONS(2756),
    [anon_sym_PIPE] = ACTIONS(2758),
    [anon_sym_COLON2] = ACTIONS(1833),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_COLON_GT] = ACTIONS(1841),
    [anon_sym_in] = ACTIONS(1821),
    [anon_sym_as] = ACTIONS(2760),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_COLON_COLON] = ACTIONS(2762),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2764),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2766),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2768),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym_let_and_operator] = ACTIONS(1819),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1548)] = {
    [sym__match_cases] = STATE(2698),
    [sym_match_case] = STATE(2724),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5528),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1548),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2709),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1549)] = {
    [sym__match_cases] = STATE(2698),
    [sym_match_case] = STATE(2691),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5469),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1549),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2711),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1550)] = {
    [sym__match_cases] = STATE(2644),
    [sym_match_case] = STATE(2691),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5469),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1550),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2711),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1551)] = {
    [sym__match_cases] = STATE(2649),
    [sym_match_case] = STATE(2691),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5469),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1551),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2711),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1552)] = {
    [sym__match_cases] = STATE(2683),
    [sym_match_case] = STATE(2691),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5469),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1552),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2711),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1553)] = {
    [sym__match_cases] = STATE(2685),
    [sym_match_case] = STATE(2691),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5469),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1553),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2711),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1554)] = {
    [sym__match_cases] = STATE(2740),
    [sym_match_case] = STATE(3222),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5411),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1554),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2701),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1555)] = {
    [sym__match_cases] = STATE(2644),
    [sym_match_case] = STATE(2724),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5528),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1555),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2709),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1556)] = {
    [sym__match_cases] = STATE(2685),
    [sym_match_case] = STATE(2896),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5450),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1556),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2715),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1557)] = {
    [sym__match_cases] = STATE(2649),
    [sym_match_case] = STATE(2724),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5528),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1557),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2709),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1558)] = {
    [sym__match_cases] = STATE(2683),
    [sym_match_case] = STATE(2724),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5528),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1558),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2709),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1559)] = {
    [sym__match_cases] = STATE(2685),
    [sym_match_case] = STATE(2724),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5528),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1559),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2709),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1560)] = {
    [sym__match_cases] = STATE(2698),
    [sym_match_case] = STATE(2735),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5489),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1560),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2703),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1561)] = {
    [sym_parameter] = STATE(2862),
    [sym__maybe_polymorphic_typed] = STATE(5016),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1561),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(8318),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(5093),
    [aux_sym_let_binding_repeat1] = STATE(1506),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_EQ] = ACTIONS(2770),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_COLON2] = ACTIONS(2736),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_constraint] = ACTIONS(2772),
    [anon_sym_val] = ACTIONS(2772),
    [anon_sym_end] = ACTIONS(2772),
    [anon_sym_inherit] = ACTIONS(2772),
    [anon_sym_method] = ACTIONS(2772),
    [anon_sym_initializer] = ACTIONS(2772),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2774),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2774),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2774),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1562)] = {
    [sym__match_cases] = STATE(2644),
    [sym_match_case] = STATE(2735),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5489),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1562),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2703),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1563)] = {
    [sym__match_cases] = STATE(2649),
    [sym_match_case] = STATE(2735),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5489),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1563),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2703),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1564)] = {
    [sym__match_cases] = STATE(2683),
    [sym_match_case] = STATE(2735),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5489),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1564),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2703),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1565)] = {
    [sym__match_cases] = STATE(2685),
    [sym_match_case] = STATE(2735),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5489),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1565),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2703),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1566)] = {
    [sym__match_cases] = STATE(2709),
    [sym_match_case] = STATE(3222),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5411),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1566),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2701),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1567)] = {
    [sym__match_cases] = STATE(3554),
    [sym_match_case] = STATE(3811),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5534),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1567),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2705),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1568)] = {
    [sym__match_cases] = STATE(2740),
    [sym_match_case] = STATE(3935),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5529),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1568),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2695),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1569)] = {
    [sym_parameter] = STATE(2862),
    [sym__maybe_polymorphic_typed] = STATE(5019),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1569),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(8318),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(5053),
    [aux_sym_let_binding_repeat1] = STATE(1506),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_EQ] = ACTIONS(2776),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_COLON2] = ACTIONS(2736),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_constraint] = ACTIONS(2778),
    [anon_sym_val] = ACTIONS(2778),
    [anon_sym_end] = ACTIONS(2778),
    [anon_sym_inherit] = ACTIONS(2778),
    [anon_sym_method] = ACTIONS(2778),
    [anon_sym_initializer] = ACTIONS(2778),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2780),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2780),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2780),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1570)] = {
    [sym_parameter] = STATE(2862),
    [sym__maybe_polymorphic_typed] = STATE(4978),
    [sym__parenthesized_abstract_type] = STATE(2862),
    [sym__delimited_pattern] = STATE(2866),
    [sym__simple_pattern] = STATE(2868),
    [sym_typed_pattern] = STATE(2866),
    [sym_polymorphic_variant_pattern] = STATE(2866),
    [sym_record_pattern] = STATE(2852),
    [sym_list_pattern] = STATE(2852),
    [sym_array_pattern] = STATE(2852),
    [sym_local_open_pattern] = STATE(2866),
    [sym_package_pattern] = STATE(2866),
    [sym_parenthesized_pattern] = STATE(2852),
    [sym_attribute] = STATE(1570),
    [sym_item_attribute] = STATE(4160),
    [sym_extension] = STATE(2866),
    [sym_quoted_extension] = STATE(2866),
    [sym__constant] = STATE(2840),
    [sym__signed_constant] = STATE(2866),
    [sym_number] = STATE(2785),
    [sym_signed_number] = STATE(2840),
    [sym_character] = STATE(2785),
    [sym_string] = STATE(2785),
    [sym_quoted_string] = STATE(2785),
    [sym_boolean] = STATE(2785),
    [sym_unit] = STATE(2852),
    [sym__simple_value_pattern] = STATE(2866),
    [sym_parenthesized_operator] = STATE(2866),
    [sym_module_path] = STATE(8318),
    [sym_constructor_path] = STATE(2866),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(2866),
    [aux_sym_expression_item_repeat1] = STATE(5054),
    [aux_sym_let_binding_repeat1] = STATE(1561),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_EQ] = ACTIONS(2782),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_COLON2] = ACTIONS(2736),
    [anon_sym_LBRACK] = ACTIONS(1835),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1839),
    [anon_sym_constraint] = ACTIONS(2784),
    [anon_sym_val] = ACTIONS(2784),
    [anon_sym_end] = ACTIONS(2784),
    [anon_sym_inherit] = ACTIONS(2784),
    [anon_sym_method] = ACTIONS(2784),
    [anon_sym_initializer] = ACTIONS(2784),
    [anon_sym_POUND] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1851),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2355),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2786),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1855),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1857),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2786),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2786),
    [aux_sym_number_token1] = ACTIONS(1859),
    [aux_sym_signed_number_token1] = ACTIONS(1861),
    [anon_sym_SQUOTE] = ACTIONS(1863),
    [anon_sym_DQUOTE] = ACTIONS(1865),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1571)] = {
    [sym__match_cases] = STATE(2709),
    [sym_match_case] = STATE(3935),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5529),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1571),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2695),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1572)] = {
    [sym__match_cases] = STATE(2712),
    [sym_match_case] = STATE(3935),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5529),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1572),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2695),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1573)] = {
    [sym__match_cases] = STATE(2748),
    [sym_match_case] = STATE(3935),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5529),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1573),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2695),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1574)] = {
    [sym__match_cases] = STATE(2706),
    [sym_match_case] = STATE(3935),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5529),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1574),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2695),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1575)] = {
    [sym__match_cases] = STATE(2740),
    [sym_match_case] = STATE(3233),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5417),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1575),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2697),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1576)] = {
    [sym__match_cases] = STATE(3574),
    [sym_match_case] = STATE(3811),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5534),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1576),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2705),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1577)] = {
    [sym__match_cases] = STATE(3599),
    [sym_match_case] = STATE(3811),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5534),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1577),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2705),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1578)] = {
    [sym__match_cases] = STATE(3505),
    [sym_match_case] = STATE(3811),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5534),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1578),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2705),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1579)] = {
    [sym__match_cases] = STATE(2649),
    [sym_match_case] = STATE(2833),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5423),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1579),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_PIPE] = ACTIONS(2699),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1580)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5033),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1580),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1725),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1581)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(6003),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1581),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1728),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1582)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5844),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1582),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1363),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1381),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1583)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(6050),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1583),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__attribute] = STATE(1684),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1584)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5966),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1584),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(709),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1381),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1585)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5844),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1585),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1363),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1481),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1586)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5598),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1586),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__attribute] = STATE(1772),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1587)] = {
    [sym__signature] = STATE(8321),
    [sym_external] = STATE(4942),
    [sym_type_definition] = STATE(4942),
    [sym_exception_definition] = STATE(4942),
    [sym_module_definition] = STATE(4942),
    [sym_module_type_definition] = STATE(4942),
    [sym_open_module] = STATE(4942),
    [sym_class_definition] = STATE(4942),
    [sym_class_type_definition] = STATE(4942),
    [sym__signature_item] = STATE(4945),
    [sym_value_specification] = STATE(4942),
    [sym_include_module_type] = STATE(4942),
    [sym__delimited_type] = STATE(3757),
    [sym__simple_type] = STATE(4716),
    [sym__type] = STATE(7149),
    [sym_function_type] = STATE(3246),
    [sym_labeled_argument_type] = STATE(8304),
    [sym__proper_tuple_type] = STATE(5026),
    [sym__labeled_tuple_type] = STATE(3246),
    [sym_labeled_tuple_element_type] = STATE(7194),
    [sym_constructed_type] = STATE(3757),
    [sym_aliased_type] = STATE(3246),
    [sym_local_open_type] = STATE(3757),
    [sym_polymorphic_variant_type] = STATE(3644),
    [sym_package_type] = STATE(3644),
    [sym_object_type] = STATE(3757),
    [sym_hash_type] = STATE(3757),
    [sym_parenthesized_type] = STATE(3644),
    [sym_attribute] = STATE(1587),
    [sym_floating_attribute] = STATE(4942),
    [sym_extension] = STATE(3757),
    [sym_quoted_extension] = STATE(3757),
    [sym_item_extension] = STATE(4942),
    [sym_quoted_item_extension] = STATE(4942),
    [sym_extended_module_path] = STATE(6901),
    [sym_type_constructor_path] = STATE(3757),
    [sym_type_variable] = STATE(3757),
    [aux_sym__structure_repeat1] = STATE(3790),
    [aux_sym__signature_repeat1] = STATE(3742),
    [sym__lowercase_identifier] = ACTIONS(2788),
    [anon_sym_SEMI_SEMI] = ACTIONS(2790),
    [anon_sym_QMARK] = ACTIONS(2792),
    [anon_sym_LPAREN] = ACTIONS(2794),
    [anon_sym_external] = ACTIONS(2796),
    [anon_sym_type] = ACTIONS(2798),
    [anon_sym_LBRACK] = ACTIONS(2800),
    [anon_sym_RBRACK] = ACTIONS(2802),
    [anon_sym_exception] = ACTIONS(2804),
    [anon_sym_module] = ACTIONS(2806),
    [anon_sym_open] = ACTIONS(35),
    [anon_sym_include] = ACTIONS(2808),
    [anon_sym_class] = ACTIONS(2810),
    [anon_sym_val] = ACTIONS(2812),
    [anon_sym_LBRACK_GT] = ACTIONS(2814),
    [anon_sym_LBRACK_LT] = ACTIONS(2816),
    [anon_sym_LT] = ACTIONS(2818),
    [anon_sym_POUND] = ACTIONS(2820),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(73),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2822),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2824),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(79),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(81),
    [sym__uppercase_identifier] = ACTIONS(2826),
    [aux_sym_type_variable_token1] = ACTIONS(2828),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1588)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5511),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym__sequence_pattern_content] = STATE(8325),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1588),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_RBRACK] = ACTIONS(2830),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1589)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5414),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym__sequence_pattern_content] = STATE(7709),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1589),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2832),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1590)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5993),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1590),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1724),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1591)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5744),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1591),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__attribute] = STATE(1695),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1592)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5966),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1592),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(709),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(713),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1593)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(6010),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1593),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(709),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(713),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1594)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5521),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym__sequence_binding_pattern_content] = STATE(7728),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1594),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2834),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1595)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5791),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1595),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(1461),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1467),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1596)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(6005),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1596),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1683),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1597)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5817),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1597),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(1421),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1567),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1598)] = {
    [sym_let_binding] = STATE(6036),
    [sym_open_module] = STATE(7497),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(1598),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__attribute] = STATE(1804),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1599)] = {
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3758),
    [sym__binding_pattern] = STATE(3619),
    [sym_alias_binding_pattern] = STATE(3758),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_anonymous] = STATE(3614),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7206),
    [sym__tuple_binding_pattern] = STATE(3615),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern] = STATE(3758),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3758),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_exception_binding_pattern] = STATE(3758),
    [sym_effect_binding_pattern] = STATE(3758),
    [sym_attribute] = STATE(1599),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__attribute] = STATE(1681),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(2321),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [anon_sym_effect] = ACTIONS(2323),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1600)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5801),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1600),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1421),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1567),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1601)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(6010),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1601),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(709),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1381),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1602)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5511),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym__sequence_pattern_content] = STATE(7937),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1602),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_RBRACK] = ACTIONS(2836),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1603)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5414),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym__sequence_pattern_content] = STATE(7944),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1603),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2838),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1604)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5801),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1604),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1421),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1605)] = {
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3758),
    [sym__binding_pattern] = STATE(3654),
    [sym_alias_binding_pattern] = STATE(3758),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_anonymous] = STATE(3614),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7206),
    [sym__tuple_binding_pattern] = STATE(3615),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern] = STATE(3758),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3758),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_exception_binding_pattern] = STATE(3758),
    [sym_effect_binding_pattern] = STATE(3758),
    [sym_attribute] = STATE(1605),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__attribute] = STATE(1741),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(2321),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [anon_sym_effect] = ACTIONS(2323),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1606)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5465),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym__sequence_binding_pattern_content] = STATE(8046),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1606),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_RBRACK] = ACTIONS(2840),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1607)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5521),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym__sequence_binding_pattern_content] = STATE(8062),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1607),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2842),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1608)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5465),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym__sequence_binding_pattern_content] = STATE(7707),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1608),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_RBRACK] = ACTIONS(2844),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1609)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5521),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym__sequence_binding_pattern_content] = STATE(7719),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1609),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2846),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1610)] = {
    [sym_let_binding] = STATE(6036),
    [sym_open_module] = STATE(8092),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(1610),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__attribute] = STATE(1804),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1611)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5285),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1611),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__attribute] = STATE(1686),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1612)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5849),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1612),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(1447),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1453),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1613)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5511),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym__sequence_pattern_content] = STATE(8065),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1613),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_RBRACK] = ACTIONS(2848),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1614)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5465),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym__sequence_binding_pattern_content] = STATE(8261),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1614),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_RBRACK] = ACTIONS(2850),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1615)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5521),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym__sequence_binding_pattern_content] = STATE(8264),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1615),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2852),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1616)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5285),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1616),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__attribute] = STATE(1709),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1617)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5414),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym__sequence_pattern_content] = STATE(8080),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1617),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2854),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1618)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5511),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym__sequence_pattern_content] = STATE(8331),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1618),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_RBRACK] = ACTIONS(2856),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1619)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5414),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym__sequence_pattern_content] = STATE(8334),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1619),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2858),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1620)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5545),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1620),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__attribute] = STATE(1691),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1621)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(6073),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1621),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1719),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1622)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5580),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1622),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__attribute] = STATE(1721),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1623)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5248),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1623),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1715),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1624)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5546),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1624),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__attribute] = STATE(1692),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1625)] = {
    [sym__parenthesized_abstract_type] = STATE(1720),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(6074),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1625),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2437),
    [anon_sym_COMMA] = ACTIONS(2433),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1626)] = {
    [sym_let_binding] = STATE(6036),
    [sym_open_module] = STATE(7973),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(1626),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__attribute] = STATE(1804),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1627)] = {
    [sym_let_binding] = STATE(6036),
    [sym_open_module] = STATE(7241),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(1627),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__attribute] = STATE(1804),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1628)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5817),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1628),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_RPAREN] = ACTIONS(1421),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1629)] = {
    [sym_let_binding] = STATE(6036),
    [sym_open_module] = STATE(7307),
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3673),
    [sym__binding_pattern_no_exn] = STATE(1547),
    [sym_alias_binding_pattern_no_exn] = STATE(3673),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(3673),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7360),
    [sym__tuple_binding_pattern_no_exn] = STATE(3673),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern_no_exn] = STATE(3673),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3673),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_attribute] = STATE(1629),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__attribute] = STATE(1804),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_rec] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1630)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5832),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1630),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_RPAREN] = ACTIONS(1471),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_COLON_COLON] = ACTIONS(1477),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1631)] = {
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2246),
    [sym__binding_pattern] = STATE(2335),
    [sym_alias_binding_pattern] = STATE(2246),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_anonymous] = STATE(2248),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7706),
    [sym__tuple_binding_pattern] = STATE(2249),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern] = STATE(2246),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2246),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_exception_binding_pattern] = STATE(2246),
    [sym_effect_binding_pattern] = STATE(2246),
    [sym_attribute] = STATE(1631),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__attribute] = STATE(1703),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1657),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [anon_sym_effect] = ACTIONS(1667),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2860),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2862),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1632)] = {
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2246),
    [sym__binding_pattern] = STATE(2322),
    [sym_alias_binding_pattern] = STATE(2246),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_anonymous] = STATE(2248),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7706),
    [sym__tuple_binding_pattern] = STATE(2249),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern] = STATE(2246),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2246),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_exception_binding_pattern] = STATE(2246),
    [sym_effect_binding_pattern] = STATE(2246),
    [sym_attribute] = STATE(1632),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__attribute] = STATE(1742),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1657),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [anon_sym_effect] = ACTIONS(1667),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2860),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2862),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1633)] = {
    [sym__parenthesized_abstract_type] = STATE(1685),
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(6060),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1633),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2558),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1634)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5530),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1634),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1735),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1635)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5033),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1635),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1736),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1636)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5517),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1636),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__attribute] = STATE(1698),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1637)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5999),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1637),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__attribute] = STATE(1771),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1638)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5465),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym__sequence_binding_pattern_content] = STATE(8167),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1638),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_RBRACK] = ACTIONS(2864),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1639)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(6079),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1639),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_COMMA] = ACTIONS(2433),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1640)] = {
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3758),
    [sym__binding_pattern] = STATE(3649),
    [sym_alias_binding_pattern] = STATE(3758),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_anonymous] = STATE(3614),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(3659),
    [sym__tuple_binding_pattern] = STATE(3663),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern] = STATE(3758),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3758),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_exception_binding_pattern] = STATE(3758),
    [sym_effect_binding_pattern] = STATE(3758),
    [sym_attribute] = STATE(1640),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(2866),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_DOT_DOT] = ACTIONS(2868),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(2321),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [anon_sym_effect] = ACTIONS(2323),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1641)] = {
    [sym_match_case] = STATE(3585),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5407),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1641),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1642)] = {
    [sym_match_case] = STATE(3223),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5411),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1642),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1643)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5675),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1643),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_RBRACK] = ACTIONS(2870),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1644)] = {
    [sym_match_case] = STATE(3961),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5529),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1644),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1645)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5657),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(5658),
    [sym__tuple_binding_pattern] = STATE(5319),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1645),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(2872),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_DOT_DOT] = ACTIONS(2874),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1646)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5675),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1646),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_RBRACK] = ACTIONS(2876),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1647)] = {
    [sym_match_case] = STATE(2646),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5522),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1647),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1648)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5731),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1648),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2876),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1649)] = {
    [sym_match_case] = STATE(2903),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5450),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1649),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1650)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5458),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(5459),
    [sym__tuple_binding_pattern] = STATE(5319),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1650),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(2878),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_DOT_DOT] = ACTIONS(2874),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1651)] = {
    [sym_match_case] = STATE(2692),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5469),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1651),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1652)] = {
    [sym_match_case] = STATE(2646),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5489),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1652),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1653)] = {
    [sym_match_case] = STATE(4065),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5529),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1653),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1654)] = {
    [sym_match_case] = STATE(2646),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5450),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1654),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1655)] = {
    [sym_match_case] = STATE(2851),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5411),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1655),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1656)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5723),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(5724),
    [sym__tuple_binding_pattern] = STATE(5725),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1656),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(2880),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_DOT_DOT] = ACTIONS(2882),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1657)] = {
    [sym_match_case] = STATE(2851),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5417),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1657),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1658)] = {
    [sym_match_case] = STATE(2646),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5469),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1658),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1659)] = {
    [sym_match_case] = STATE(2834),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5423),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1659),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1660)] = {
    [sym_match_case] = STATE(3252),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5417),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1660),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1661)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5710),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(5711),
    [sym__tuple_pattern] = STATE(5712),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1661),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2884),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_DOT_DOT] = ACTIONS(2886),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1662)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5568),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1662),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_RBRACK] = ACTIONS(2888),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1663)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5782),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1663),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2888),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1664)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5782),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1664),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2890),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1665)] = {
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2246),
    [sym__binding_pattern] = STATE(2378),
    [sym_alias_binding_pattern] = STATE(2246),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_anonymous] = STATE(2248),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(2380),
    [sym__tuple_binding_pattern] = STATE(2381),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern] = STATE(2246),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2246),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_exception_binding_pattern] = STATE(2246),
    [sym_effect_binding_pattern] = STATE(2246),
    [sym_attribute] = STATE(1665),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(2892),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_DOT_DOT] = ACTIONS(2894),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1657),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [anon_sym_effect] = ACTIONS(1667),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2860),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2862),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1666)] = {
    [sym_match_case] = STATE(3831),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5534),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1666),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1667)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5568),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1667),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_RBRACK] = ACTIONS(2890),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1668)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5501),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(5502),
    [sym__tuple_pattern] = STATE(5070),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1668),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2896),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_DOT_DOT] = ACTIONS(2898),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1669)] = {
    [sym_match_case] = STATE(3585),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5534),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1669),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1670)] = {
    [sym_match_case] = STATE(2749),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5528),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1670),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1671)] = {
    [sym_match_case] = STATE(2646),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5528),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1671),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1672)] = {
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3758),
    [sym__binding_pattern] = STATE(3704),
    [sym_alias_binding_pattern] = STATE(3758),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_anonymous] = STATE(3614),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(3706),
    [sym__tuple_binding_pattern] = STATE(3708),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern] = STATE(3758),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3758),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_exception_binding_pattern] = STATE(3758),
    [sym_effect_binding_pattern] = STATE(3758),
    [sym_attribute] = STATE(1672),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(2866),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_DOT_DOT] = ACTIONS(2900),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(2321),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [anon_sym_effect] = ACTIONS(2323),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1673)] = {
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2246),
    [sym__binding_pattern] = STATE(2408),
    [sym_alias_binding_pattern] = STATE(2246),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_anonymous] = STATE(2248),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(2257),
    [sym__tuple_binding_pattern] = STATE(2208),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern] = STATE(2246),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2246),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_exception_binding_pattern] = STATE(2246),
    [sym_effect_binding_pattern] = STATE(2246),
    [sym_attribute] = STATE(1673),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(2892),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_DOT_DOT] = ACTIONS(2902),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1657),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [anon_sym_effect] = ACTIONS(1667),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2860),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2862),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1674)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(6061),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1674),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1675)] = {
    [sym_match_case] = STATE(2646),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5423),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1675),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1676)] = {
    [sym_match_case] = STATE(2878),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5522),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1676),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1677)] = {
    [sym_match_case] = STATE(2705),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5489),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1677),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1678)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5217),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(5237),
    [sym__tuple_pattern] = STATE(5070),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1678),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2904),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_DOT_DOT] = ACTIONS(2898),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1679)] = {
    [sym_match_case] = STATE(3794),
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5407),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1679),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1680)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5731),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1680),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2870),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1681)] = {
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3758),
    [sym__binding_pattern] = STATE(3662),
    [sym_alias_binding_pattern] = STATE(3758),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_anonymous] = STATE(3614),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7206),
    [sym__tuple_binding_pattern] = STATE(3615),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern] = STATE(3758),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3758),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_exception_binding_pattern] = STATE(3758),
    [sym_effect_binding_pattern] = STATE(3758),
    [sym_attribute] = STATE(1681),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(2321),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [anon_sym_effect] = ACTIONS(2323),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1682)] = {
    [sym_let_binding] = STATE(4629),
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2353),
    [sym__binding_pattern_no_exn] = STATE(926),
    [sym_alias_binding_pattern_no_exn] = STATE(2353),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2353),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8145),
    [sym__tuple_binding_pattern_no_exn] = STATE(2353),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern_no_exn] = STATE(2353),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2353),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_attribute] = STATE(1682),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__attribute] = STATE(1777),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_rec] = ACTIONS(2906),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2860),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2862),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1683)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(6006),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1683),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1684)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(6286),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1684),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1685)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(6302),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1685),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1686)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5267),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1686),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1687)] = {
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3758),
    [sym__binding_pattern] = STATE(3665),
    [sym_alias_binding_pattern] = STATE(3758),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_anonymous] = STATE(3672),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7206),
    [sym__tuple_binding_pattern] = STATE(3615),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern] = STATE(3758),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3758),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_exception_binding_pattern] = STATE(3758),
    [sym_effect_binding_pattern] = STATE(3758),
    [sym_attribute] = STATE(1687),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(2321),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [anon_sym_effect] = ACTIONS(2323),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1688)] = {
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3758),
    [sym__binding_pattern] = STATE(3696),
    [sym_alias_binding_pattern] = STATE(3758),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_anonymous] = STATE(3614),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7206),
    [sym__tuple_binding_pattern] = STATE(3615),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern] = STATE(3758),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3758),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_exception_binding_pattern] = STATE(3758),
    [sym_effect_binding_pattern] = STATE(3758),
    [sym_attribute] = STATE(1688),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(2321),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [anon_sym_effect] = ACTIONS(2323),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1689)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5272),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1689),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1690)] = {
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2246),
    [sym__binding_pattern] = STATE(2269),
    [sym_alias_binding_pattern] = STATE(2246),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_anonymous] = STATE(2288),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7706),
    [sym__tuple_binding_pattern] = STATE(2249),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern] = STATE(2246),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2246),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_exception_binding_pattern] = STATE(2246),
    [sym_effect_binding_pattern] = STATE(2246),
    [sym_attribute] = STATE(1690),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1657),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [anon_sym_effect] = ACTIONS(1667),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2860),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2862),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1691)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5707),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1691),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1692)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5708),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1692),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1693)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5713),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5714),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1693),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1694)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5716),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1694),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1695)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5651),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1695),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1696)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5717),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1696),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1697)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5663),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5324),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1697),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1698)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5456),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1698),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1699)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5460),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5324),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1699),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1700)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5462),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1700),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1701)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5667),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8146),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1701),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1389),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1447),
    [sym__lowercase_identifier] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1702)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5699),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1702),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1703)] = {
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2246),
    [sym__binding_pattern] = STATE(2376),
    [sym_alias_binding_pattern] = STATE(2246),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_anonymous] = STATE(2248),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7706),
    [sym__tuple_binding_pattern] = STATE(2249),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern] = STATE(2246),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2246),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_exception_binding_pattern] = STATE(2246),
    [sym_effect_binding_pattern] = STATE(2246),
    [sym_attribute] = STATE(1703),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1657),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [anon_sym_effect] = ACTIONS(1667),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2860),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2862),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1704)] = {
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2246),
    [sym__binding_pattern] = STATE(2435),
    [sym_alias_binding_pattern] = STATE(2246),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_anonymous] = STATE(2248),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7706),
    [sym__tuple_binding_pattern] = STATE(2249),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern] = STATE(2246),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2246),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_exception_binding_pattern] = STATE(2246),
    [sym_effect_binding_pattern] = STATE(2246),
    [sym_attribute] = STATE(1704),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1657),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [anon_sym_effect] = ACTIONS(1667),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2860),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2862),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1705)] = {
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2246),
    [sym__binding_pattern] = STATE(2383),
    [sym_alias_binding_pattern] = STATE(2246),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_anonymous] = STATE(2384),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7706),
    [sym__tuple_binding_pattern] = STATE(2249),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern] = STATE(2246),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2246),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_exception_binding_pattern] = STATE(2246),
    [sym_effect_binding_pattern] = STATE(2246),
    [sym_attribute] = STATE(1705),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1657),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [anon_sym_effect] = ACTIONS(1667),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2860),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2862),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1706)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5731),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1706),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1707)] = {
    [sym__delimited_pattern] = STATE(5633),
    [sym__simple_pattern] = STATE(5584),
    [sym__effect_pattern] = STATE(5585),
    [sym__pattern] = STATE(5782),
    [sym_alias_pattern] = STATE(5585),
    [sym_typed_pattern] = STATE(5633),
    [sym__or_pattern_anonymous] = STATE(5596),
    [sym_constructor_pattern] = STATE(5584),
    [sym_tag_pattern] = STATE(5584),
    [sym_polymorphic_variant_pattern] = STATE(5633),
    [sym_labeled_tuple_element_pattern] = STATE(7526),
    [sym__tuple_pattern] = STATE(5597),
    [sym_record_pattern] = STATE(5636),
    [sym_list_pattern] = STATE(5636),
    [sym_cons_pattern] = STATE(5585),
    [sym_array_pattern] = STATE(5636),
    [sym_range_pattern] = STATE(5585),
    [sym_lazy_pattern] = STATE(5584),
    [sym_local_open_pattern] = STATE(5633),
    [sym_package_pattern] = STATE(5633),
    [sym_parenthesized_pattern] = STATE(5636),
    [sym_exception_pattern] = STATE(5585),
    [sym_effect_pattern] = STATE(5585),
    [sym_attribute] = STATE(1707),
    [sym_extension] = STATE(5633),
    [sym_quoted_extension] = STATE(5633),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5448),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5636),
    [sym__simple_value_pattern] = STATE(5633),
    [sym_parenthesized_operator] = STATE(5633),
    [sym_module_path] = STATE(7195),
    [sym_constructor_path] = STATE(1347),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1457),
    [sym__lowercase_identifier] = ACTIONS(2578),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(2638),
    [anon_sym_LBRACK] = ACTIONS(2582),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2586),
    [anon_sym_exception] = ACTIONS(2588),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2592),
    [anon_sym_lazy] = ACTIONS(2594),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2598),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1708)] = {
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2246),
    [sym__binding_pattern] = STATE(2387),
    [sym_alias_binding_pattern] = STATE(2246),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_anonymous] = STATE(2248),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7706),
    [sym__tuple_binding_pattern] = STATE(2249),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern] = STATE(2246),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2246),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_exception_binding_pattern] = STATE(2246),
    [sym_effect_binding_pattern] = STATE(2246),
    [sym_attribute] = STATE(1708),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1657),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [anon_sym_effect] = ACTIONS(1667),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2860),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2862),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1709)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5267),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1709),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1710)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5272),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1710),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1711)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5205),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1711),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1712)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5537),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1712),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1713)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5573),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1713),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1714)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5426),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1714),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1715)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5228),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1715),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1716)] = {
    [sym_attribute] = STATE(1716),
    [sym__lowercase_identifier] = ACTIONS(1947),
    [anon_sym_TILDE] = ACTIONS(1947),
    [anon_sym_QMARK] = ACTIONS(1947),
    [anon_sym_LPAREN] = ACTIONS(1949),
    [anon_sym_COMMA] = ACTIONS(1949),
    [anon_sym_PIPE] = ACTIONS(1947),
    [anon_sym_DOT] = ACTIONS(1949),
    [anon_sym_LBRACK] = ACTIONS(1947),
    [anon_sym_true] = ACTIONS(1947),
    [anon_sym_false] = ACTIONS(1947),
    [anon_sym_LBRACE] = ACTIONS(1947),
    [anon_sym_SEMI] = ACTIONS(1949),
    [anon_sym_constraint] = ACTIONS(1947),
    [anon_sym_val] = ACTIONS(1947),
    [anon_sym_end] = ACTIONS(1947),
    [anon_sym_object] = ACTIONS(1947),
    [anon_sym_inherit] = ACTIONS(1947),
    [anon_sym_method] = ACTIONS(1947),
    [anon_sym_initializer] = ACTIONS(1947),
    [anon_sym_AMP] = ACTIONS(1947),
    [anon_sym_POUND] = ACTIONS(1947),
    [anon_sym_COLON_COLON] = ACTIONS(1949),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1949),
    [anon_sym_LT_DASH] = ACTIONS(2908),
    [anon_sym_else] = ACTIONS(1947),
    [anon_sym_do] = ACTIONS(1947),
    [anon_sym_new] = ACTIONS(1947),
    [anon_sym_LBRACE_LT] = ACTIONS(1949),
    [anon_sym_begin] = ACTIONS(1947),
    [sym_ocamlyacc_value] = ACTIONS(1949),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1947),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1949),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1947),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1949),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1949),
    [aux_sym_number_token1] = ACTIONS(1949),
    [aux_sym_signed_number_token1] = ACTIONS(1947),
    [anon_sym_SQUOTE] = ACTIONS(1949),
    [anon_sym_DQUOTE] = ACTIONS(1949),
    [sym_prefix_operator] = ACTIONS(1949),
    [aux_sym_sign_operator_token1] = ACTIONS(1947),
    [sym_hash_operator] = ACTIONS(1949),
    [anon_sym_lsl] = ACTIONS(1947),
    [anon_sym_lsr] = ACTIONS(1947),
    [anon_sym_asr] = ACTIONS(1947),
    [aux_sym_pow_operator_token1] = ACTIONS(1949),
    [anon_sym_mod] = ACTIONS(1947),
    [anon_sym_land] = ACTIONS(1947),
    [anon_sym_lor] = ACTIONS(1947),
    [anon_sym_lxor] = ACTIONS(1947),
    [aux_sym_mult_operator_token1] = ACTIONS(1947),
    [aux_sym_add_operator_token1] = ACTIONS(1947),
    [sym_concat_operator] = ACTIONS(1949),
    [sym_rel_operator] = ACTIONS(1947),
    [anon_sym_AMP_AMP] = ACTIONS(1947),
    [anon_sym_or] = ACTIONS(1947),
    [anon_sym_PIPE_PIPE] = ACTIONS(1947),
    [sym_assign_operator] = ACTIONS(1949),
    [sym__uppercase_identifier] = ACTIONS(1949),
    [aux_sym_tag_token1] = ACTIONS(1949),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1717)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5680),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1717),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1718)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5075),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1718),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1719)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(6109),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1719),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1720)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(6098),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1720),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1721)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5721),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1721),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1722)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5726),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5727),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1722),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1723)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5729),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1723),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1724)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5928),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1724),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1725)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5067),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1725),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1726)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5404),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1726),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1727)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5757),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1727),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1728)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(6004),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1728),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1729)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5609),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1729),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1730)] = {
    [sym__delimited_binding_pattern] = STATE(5310),
    [sym__simple_binding_pattern] = STATE(5311),
    [sym__effect_binding_pattern] = STATE(5329),
    [sym__binding_pattern] = STATE(5675),
    [sym_alias_binding_pattern] = STATE(5329),
    [sym_typed_binding_pattern] = STATE(5310),
    [sym__or_binding_pattern_anonymous] = STATE(5327),
    [sym_constructor_binding_pattern] = STATE(5311),
    [sym_tag_binding_pattern] = STATE(5311),
    [sym_polymorphic_variant_pattern] = STATE(5310),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7619),
    [sym__tuple_binding_pattern] = STATE(5279),
    [sym_record_binding_pattern] = STATE(5312),
    [sym_list_binding_pattern] = STATE(5312),
    [sym_cons_binding_pattern] = STATE(5329),
    [sym_array_binding_pattern] = STATE(5312),
    [sym_range_pattern] = STATE(5329),
    [sym_lazy_binding_pattern] = STATE(5311),
    [sym_local_open_binding_pattern] = STATE(5310),
    [sym_package_pattern] = STATE(5310),
    [sym_parenthesized_binding_pattern] = STATE(5312),
    [sym_exception_binding_pattern] = STATE(5329),
    [sym_effect_binding_pattern] = STATE(5329),
    [sym_attribute] = STATE(1730),
    [sym_extension] = STATE(5310),
    [sym_quoted_extension] = STATE(5310),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5231),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5312),
    [sym__value_name] = STATE(5310),
    [sym_parenthesized_operator] = STATE(4997),
    [sym_module_path] = STATE(7412),
    [sym_constructor_path] = STATE(1297),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1362),
    [sym__lowercase_identifier] = ACTIONS(2556),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1419),
    [anon_sym_LBRACK] = ACTIONS(1425),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1427),
    [anon_sym_exception] = ACTIONS(2560),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(2562),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2564),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1731)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5510),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1731),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1732)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5568),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1732),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1733)] = {
    [sym_let_binding] = STATE(4505),
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2353),
    [sym__binding_pattern_no_exn] = STATE(918),
    [sym_alias_binding_pattern_no_exn] = STATE(2353),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2353),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(8145),
    [sym__tuple_binding_pattern_no_exn] = STATE(2353),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern_no_exn] = STATE(2353),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2353),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_attribute] = STATE(1733),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__attribute] = STATE(1798),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_rec] = ACTIONS(2910),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2860),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2862),
    [anon_sym_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1734)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5415),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1734),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1735)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5500),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1735),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1736)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5067),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1736),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1737)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5503),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5073),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1737),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1738)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5504),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1738),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1739)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5075),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(8220),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1739),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1318),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1406),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(2568),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(2570),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(2572),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1740)] = {
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2246),
    [sym__binding_pattern] = STATE(2290),
    [sym_alias_binding_pattern] = STATE(2246),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_anonymous] = STATE(2248),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7706),
    [sym__tuple_binding_pattern] = STATE(2249),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern] = STATE(2246),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2246),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_exception_binding_pattern] = STATE(2246),
    [sym_effect_binding_pattern] = STATE(2246),
    [sym_attribute] = STATE(1740),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1657),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [anon_sym_effect] = ACTIONS(1667),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2860),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2862),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1741)] = {
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3758),
    [sym__binding_pattern] = STATE(3701),
    [sym_alias_binding_pattern] = STATE(3758),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_anonymous] = STATE(3614),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7206),
    [sym__tuple_binding_pattern] = STATE(3615),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern] = STATE(3758),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3758),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_exception_binding_pattern] = STATE(3758),
    [sym_effect_binding_pattern] = STATE(3758),
    [sym_attribute] = STATE(1741),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(2321),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [anon_sym_effect] = ACTIONS(2323),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1742)] = {
    [sym__delimited_binding_pattern] = STATE(2336),
    [sym__simple_binding_pattern] = STATE(2238),
    [sym__effect_binding_pattern] = STATE(2246),
    [sym__binding_pattern] = STATE(2427),
    [sym_alias_binding_pattern] = STATE(2246),
    [sym_typed_binding_pattern] = STATE(2336),
    [sym__or_binding_pattern_anonymous] = STATE(2248),
    [sym_constructor_binding_pattern] = STATE(2238),
    [sym_tag_binding_pattern] = STATE(2238),
    [sym_polymorphic_variant_pattern] = STATE(2336),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7706),
    [sym__tuple_binding_pattern] = STATE(2249),
    [sym_record_binding_pattern] = STATE(2438),
    [sym_list_binding_pattern] = STATE(2438),
    [sym_cons_binding_pattern] = STATE(2246),
    [sym_array_binding_pattern] = STATE(2438),
    [sym_range_pattern] = STATE(2246),
    [sym_lazy_binding_pattern] = STATE(2238),
    [sym_local_open_binding_pattern] = STATE(2336),
    [sym_package_pattern] = STATE(2336),
    [sym_parenthesized_binding_pattern] = STATE(2438),
    [sym_exception_binding_pattern] = STATE(2246),
    [sym_effect_binding_pattern] = STATE(2246),
    [sym_attribute] = STATE(1742),
    [sym_extension] = STATE(2336),
    [sym_quoted_extension] = STATE(2336),
    [sym__constant] = STATE(2134),
    [sym__signed_constant] = STATE(2114),
    [sym_number] = STATE(2123),
    [sym_signed_number] = STATE(2134),
    [sym_character] = STATE(2123),
    [sym_string] = STATE(2123),
    [sym_quoted_string] = STATE(2123),
    [sym_boolean] = STATE(2123),
    [sym_unit] = STATE(2438),
    [sym__value_name] = STATE(2336),
    [sym_parenthesized_operator] = STATE(2272),
    [sym_module_path] = STATE(7246),
    [sym_constructor_path] = STATE(725),
    [sym__constructor_name] = STATE(2088),
    [sym_tag] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_true] = ACTIONS(1653),
    [anon_sym_false] = ACTIONS(1653),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1657),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_lazy] = ACTIONS(1663),
    [anon_sym_begin] = ACTIONS(1665),
    [anon_sym_effect] = ACTIONS(1667),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2860),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2862),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1679),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1683),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1743)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5429),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1743),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1744)] = {
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3758),
    [sym__binding_pattern] = STATE(3709),
    [sym_alias_binding_pattern] = STATE(3758),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_anonymous] = STATE(3710),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7206),
    [sym__tuple_binding_pattern] = STATE(3615),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern] = STATE(3758),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3758),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_exception_binding_pattern] = STATE(3758),
    [sym_effect_binding_pattern] = STATE(3758),
    [sym_attribute] = STATE(1744),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(2321),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [anon_sym_effect] = ACTIONS(2323),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1745)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5433),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1745),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1746)] = {
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3758),
    [sym__binding_pattern] = STATE(3717),
    [sym_alias_binding_pattern] = STATE(3758),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_anonymous] = STATE(3614),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7206),
    [sym__tuple_binding_pattern] = STATE(3615),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern] = STATE(3758),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3758),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_exception_binding_pattern] = STATE(3758),
    [sym_effect_binding_pattern] = STATE(3758),
    [sym_attribute] = STATE(1746),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(2321),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [anon_sym_effect] = ACTIONS(2323),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1747)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5442),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1747),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1748)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5443),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1748),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1749)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5446),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1749),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1750)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5447),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1750),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1751)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5400),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1751),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1752)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5452),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1752),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1753)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5454),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1753),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1754)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5455),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1754),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1755)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5461),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1755),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1756)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5463),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1756),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1757)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5468),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1757),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1758)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5470),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1758),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1759)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5473),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1759),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1760)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5474),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1760),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1761)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5477),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1761),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1762)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5478),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1762),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1763)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5480),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1763),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1764)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5481),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1764),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1765)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5483),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1765),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1766)] = {
    [sym__delimited_binding_pattern] = STATE(3678),
    [sym__simple_binding_pattern] = STATE(3679),
    [sym__effect_binding_pattern] = STATE(3758),
    [sym__binding_pattern] = STATE(3668),
    [sym_alias_binding_pattern] = STATE(3758),
    [sym_typed_binding_pattern] = STATE(3678),
    [sym__or_binding_pattern_anonymous] = STATE(3614),
    [sym_constructor_binding_pattern] = STATE(3679),
    [sym_tag_binding_pattern] = STATE(3679),
    [sym_polymorphic_variant_pattern] = STATE(3678),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7206),
    [sym__tuple_binding_pattern] = STATE(3615),
    [sym_record_binding_pattern] = STATE(3680),
    [sym_list_binding_pattern] = STATE(3680),
    [sym_cons_binding_pattern] = STATE(3758),
    [sym_array_binding_pattern] = STATE(3680),
    [sym_range_pattern] = STATE(3758),
    [sym_lazy_binding_pattern] = STATE(3679),
    [sym_local_open_binding_pattern] = STATE(3678),
    [sym_package_pattern] = STATE(3678),
    [sym_parenthesized_binding_pattern] = STATE(3680),
    [sym_exception_binding_pattern] = STATE(3758),
    [sym_effect_binding_pattern] = STATE(3758),
    [sym_attribute] = STATE(1766),
    [sym_extension] = STATE(3678),
    [sym_quoted_extension] = STATE(3678),
    [sym__constant] = STATE(3528),
    [sym__signed_constant] = STATE(3509),
    [sym_number] = STATE(3547),
    [sym_signed_number] = STATE(3528),
    [sym_character] = STATE(3547),
    [sym_string] = STATE(3547),
    [sym_quoted_string] = STATE(3547),
    [sym_boolean] = STATE(3547),
    [sym_unit] = STATE(3680),
    [sym__value_name] = STATE(3678),
    [sym_parenthesized_operator] = STATE(3689),
    [sym_module_path] = STATE(7189),
    [sym_constructor_path] = STATE(1018),
    [sym__constructor_name] = STATE(3429),
    [sym_tag] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_exception] = ACTIONS(2321),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_lazy] = ACTIONS(1785),
    [anon_sym_begin] = ACTIONS(1787),
    [anon_sym_effect] = ACTIONS(2323),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2325),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2327),
    [aux_sym_number_token1] = ACTIONS(1801),
    [aux_sym_signed_number_token1] = ACTIONS(1803),
    [anon_sym_SQUOTE] = ACTIONS(1805),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1767)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5484),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1767),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1768)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5485),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1768),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1769)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5486),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1769),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1770)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(5226),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5073),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1770),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1771)] = {
    [sym__delimited_pattern] = STATE(5132),
    [sym__simple_pattern] = STATE(5091),
    [sym__effect_pattern] = STATE(5098),
    [sym__pattern] = STATE(6000),
    [sym_alias_pattern] = STATE(5098),
    [sym_typed_pattern] = STATE(5132),
    [sym__or_pattern_anonymous] = STATE(5099),
    [sym_constructor_pattern] = STATE(5091),
    [sym_tag_pattern] = STATE(5091),
    [sym_polymorphic_variant_pattern] = STATE(5132),
    [sym_labeled_tuple_element_pattern] = STATE(7594),
    [sym__tuple_pattern] = STATE(5100),
    [sym_record_pattern] = STATE(5133),
    [sym_list_pattern] = STATE(5133),
    [sym_cons_pattern] = STATE(5098),
    [sym_array_pattern] = STATE(5133),
    [sym_range_pattern] = STATE(5098),
    [sym_lazy_pattern] = STATE(5091),
    [sym_local_open_pattern] = STATE(5132),
    [sym_package_pattern] = STATE(5132),
    [sym_parenthesized_pattern] = STATE(5133),
    [sym_exception_pattern] = STATE(5098),
    [sym_effect_pattern] = STATE(5098),
    [sym_attribute] = STATE(1771),
    [sym_extension] = STATE(5132),
    [sym_quoted_extension] = STATE(5132),
    [sym__constant] = STATE(5005),
    [sym__signed_constant] = STATE(5025),
    [sym_number] = STATE(4819),
    [sym_signed_number] = STATE(5005),
    [sym_character] = STATE(4819),
    [sym_string] = STATE(4819),
    [sym_quoted_string] = STATE(4819),
    [sym_boolean] = STATE(4819),
    [sym_unit] = STATE(5133),
    [sym__simple_value_pattern] = STATE(5132),
    [sym_parenthesized_operator] = STATE(5132),
    [sym_module_path] = STATE(7520),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(2610),
    [sym_tag] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1357),
    [anon_sym_TILDE] = ACTIONS(2435),
    [anon_sym_LPAREN] = ACTIONS(1361),
    [anon_sym_LBRACK] = ACTIONS(1369),
    [anon_sym_true] = ACTIONS(1371),
    [anon_sym_false] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1373),
    [anon_sym_exception] = ACTIONS(1375),
    [anon_sym_POUND] = ACTIONS(2441),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1383),
    [anon_sym_lazy] = ACTIONS(1385),
    [anon_sym_begin] = ACTIONS(1387),
    [anon_sym_effect] = ACTIONS(1389),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1393),
    [aux_sym_number_token1] = ACTIONS(1395),
    [aux_sym_signed_number_token1] = ACTIONS(2566),
    [anon_sym_SQUOTE] = ACTIONS(1399),
    [anon_sym_DQUOTE] = ACTIONS(1401),
    [sym__uppercase_identifier] = ACTIONS(1405),
    [aux_sym_tag_token1] = ACTIONS(1407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1772)] = {
    [sym__delimited_binding_pattern] = STATE(5557),
    [sym__simple_binding_pattern] = STATE(5558),
    [sym__effect_binding_pattern] = STATE(5638),
    [sym__binding_pattern] = STATE(5592),
    [sym_alias_binding_pattern] = STATE(5638),
    [sym_typed_binding_pattern] = STATE(5557),
    [sym__or_binding_pattern_anonymous] = STATE(5640),
    [sym_constructor_binding_pattern] = STATE(5558),
    [sym_tag_binding_pattern] = STATE(5558),
    [sym_polymorphic_variant_pattern] = STATE(5557),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7791),
    [sym__tuple_binding_pattern] = STATE(5642),
    [sym_record_binding_pattern] = STATE(5561),
    [sym_list_binding_pattern] = STATE(5561),
    [sym_cons_binding_pattern] = STATE(5638),
    [sym_array_binding_pattern] = STATE(5561),
    [sym_range_pattern] = STATE(5638),
    [sym_lazy_binding_pattern] = STATE(5558),
    [sym_local_open_binding_pattern] = STATE(5557),
    [sym_package_pattern] = STATE(5557),
    [sym_parenthesized_binding_pattern] = STATE(5561),
    [sym_exception_binding_pattern] = STATE(5638),
    [sym_effect_binding_pattern] = STATE(5638),
    [sym_attribute] = STATE(1772),
    [sym_extension] = STATE(5557),
    [sym_quoted_extension] = STATE(5557),
    [sym__constant] = STATE(5451),
    [sym__signed_constant] = STATE(5430),
    [sym_number] = STATE(5427),
    [sym_signed_number] = STATE(5451),
    [sym_character] = STATE(5427),
    [sym_string] = STATE(5427),
    [sym_quoted_string] = STATE(5427),
    [sym_boolean] = STATE(5427),
    [sym_unit] = STATE(5561),
    [sym__value_name] = STATE(5557),
    [sym_parenthesized_operator] = STATE(5617),
    [sym_module_path] = STATE(7721),
    [sym_constructor_path] = STATE(1388),
    [sym__constructor_name] = STATE(3879),
    [sym_tag] = STATE(1448),
    [sym__lowercase_identifier] = ACTIONS(2618),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(2636),
    [anon_sym_LBRACK] = ACTIONS(2622),
    [anon_sym_true] = ACTIONS(2584),
    [anon_sym_false] = ACTIONS(2584),
    [anon_sym_LBRACE] = ACTIONS(2624),
    [anon_sym_exception] = ACTIONS(2626),
    [anon_sym_POUND] = ACTIONS(2590),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2628),
    [anon_sym_lazy] = ACTIONS(2630),
    [anon_sym_begin] = ACTIONS(2596),
    [anon_sym_effect] = ACTIONS(2632),
    [aux_sym_attribute_token1] = ACTIONS(477),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2600),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2602),
    [aux_sym_number_token1] = ACTIONS(2604),
    [aux_sym_signed_number_token1] = ACTIONS(2606),
    [anon_sym_SQUOTE] = ACTIONS(2608),
    [anon_sym_DQUOTE] = ACTIONS(2610),
    [sym__uppercase_identifier] = ACTIONS(2612),
    [aux_sym_tag_token1] = ACTIONS(2614),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2912), 1,
      anon_sym_LT_DASH,
    STATE(1773), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1949), 25,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      anon_sym_GT_RBRACE,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1947), 34,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_then,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [77] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2914), 1,
      anon_sym_LT_DASH,
    STATE(1774), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1949), 25,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      anon_sym_GT_RBRACE,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1947), 33,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_then,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [153] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(1775), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6065), 1,
      sym__simple_expression,
    STATE(6442), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    STATE(8150), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [285] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1533), 1,
      anon_sym_RPAREN,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1776), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [423] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2944), 1,
      anon_sym_rec,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(926), 1,
      sym__binding_pattern_no_exn,
    STATE(1777), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4636), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [561] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1565), 1,
      anon_sym_RPAREN,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1778), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [699] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1539), 1,
      anon_sym_RPAREN,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1779), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [837] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(1780), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6179), 1,
      sym__simple_expression,
    STATE(6699), 1,
      sym_field_expression,
    STATE(7534), 1,
      sym__quoted_string,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [969] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1511), 1,
      anon_sym_RPAREN,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1781), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [1107] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(1782), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6201), 1,
      sym__simple_expression,
    STATE(6553), 1,
      sym_field_expression,
    STATE(7463), 1,
      sym__quoted_string,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [1239] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1543), 1,
      anon_sym_RPAREN,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1783), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [1377] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(1784), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6208), 1,
      sym__simple_expression,
    STATE(6617), 1,
      sym_field_expression,
    STATE(7577), 1,
      sym__quoted_string,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [1509] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1569), 1,
      anon_sym_RPAREN,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1785), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [1647] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(1786), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6179), 1,
      sym__simple_expression,
    STATE(6699), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    STATE(7879), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [1779] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(1787), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6216), 1,
      sym__simple_expression,
    STATE(6334), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    STATE(8333), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [1911] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(1788), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6026), 1,
      sym__simple_expression,
    STATE(6350), 1,
      sym_field_expression,
    STATE(7720), 1,
      sym__quoted_string,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2043] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1547), 1,
      anon_sym_RPAREN,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1789), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [2181] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(1790), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6222), 1,
      sym__simple_expression,
    STATE(6396), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    STATE(8104), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2313] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(1791), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6232), 1,
      sym__simple_expression,
    STATE(6507), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    STATE(8237), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2445] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(1792), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6235), 1,
      sym__simple_expression,
    STATE(6564), 1,
      sym_field_expression,
    STATE(7708), 1,
      sym__quoted_string,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2577] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(1793), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6236), 1,
      sym__simple_expression,
    STATE(6635), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    STATE(7929), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2709] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(1794), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6239), 1,
      sym__simple_expression,
    STATE(6668), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    STATE(8103), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2841] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(1795), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6244), 1,
      sym__simple_expression,
    STATE(6752), 1,
      sym_field_expression,
    STATE(7203), 1,
      sym__quoted_string,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2973] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1519), 1,
      anon_sym_RPAREN,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1796), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3111] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1549), 1,
      anon_sym_RPAREN,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1797), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3249] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2946), 1,
      anon_sym_rec,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(918), 1,
      sym__binding_pattern_no_exn,
    STATE(1798), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4506), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3387] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1553), 1,
      anon_sym_RPAREN,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1799), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3525] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2948), 1,
      anon_sym_LT_DASH,
    STATE(1800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1949), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1947), 37,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [3601] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1525), 1,
      anon_sym_RPAREN,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1801), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3739] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1557), 1,
      anon_sym_RPAREN,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1802), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3877] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1561), 1,
      anon_sym_RPAREN,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1803), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4015] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(1867), 1,
      anon_sym_rec,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1804), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6219), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4153] = 37,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1529), 1,
      anon_sym_RPAREN,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1805), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4291] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(1806), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6226), 1,
      sym__simple_expression,
    STATE(6449), 1,
      sym_field_expression,
    STATE(7508), 1,
      sym__quoted_string,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [4423] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(910), 1,
      sym__binding_pattern_no_exn,
    STATE(1807), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4530), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4558] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(926), 1,
      sym__binding_pattern_no_exn,
    STATE(1808), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4638), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4693] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(905), 1,
      sym__binding_pattern_no_exn,
    STATE(1809), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4439), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4828] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(905), 1,
      sym__binding_pattern_no_exn,
    STATE(1810), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4379), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4963] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1811), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6036), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [5098] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(910), 1,
      sym__binding_pattern_no_exn,
    STATE(1812), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4500), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [5233] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(918), 1,
      sym__binding_pattern_no_exn,
    STATE(1813), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4439), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [5368] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(918), 1,
      sym__binding_pattern_no_exn,
    STATE(1814), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4493), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [5503] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1815), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6219), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [5638] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(926), 1,
      sym__binding_pattern_no_exn,
    STATE(1816), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4629), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [5773] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(926), 1,
      sym__binding_pattern_no_exn,
    STATE(1817), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4636), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [5908] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1818), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6506), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [6043] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(910), 1,
      sym__binding_pattern_no_exn,
    STATE(1819), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4528), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [6178] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1785), 1,
      anon_sym_lazy,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1018), 1,
      sym_constructor_path,
    STATE(1031), 1,
      sym_tag,
    STATE(1547), 1,
      sym__binding_pattern_no_exn,
    STATE(1820), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3509), 1,
      sym__signed_constant,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(6166), 1,
      sym_let_binding,
    STATE(7189), 1,
      sym_module_path,
    STATE(7360), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3679), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3673), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3678), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [6313] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(905), 1,
      sym__binding_pattern_no_exn,
    STATE(1821), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4460), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [6448] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(905), 1,
      sym__binding_pattern_no_exn,
    STATE(1822), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4461), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [6583] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(918), 1,
      sym__binding_pattern_no_exn,
    STATE(1823), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4505), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [6718] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(918), 1,
      sym__binding_pattern_no_exn,
    STATE(1824), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4506), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [6853] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(910), 1,
      sym__binding_pattern_no_exn,
    STATE(1825), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4439), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [6988] = 36,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1647), 1,
      anon_sym_TILDE,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1663), 1,
      anon_sym_lazy,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(725), 1,
      sym_constructor_path,
    STATE(883), 1,
      sym_tag,
    STATE(926), 1,
      sym__binding_pattern_no_exn,
    STATE(1826), 1,
      sym_attribute,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2114), 1,
      sym__signed_constant,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(4439), 1,
      sym_let_binding,
    STATE(7246), 1,
      sym_module_path,
    STATE(8145), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2238), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2353), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2336), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [7123] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1827), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6065), 1,
      sym__simple_expression,
    STATE(6442), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [7249] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2950), 1,
      sym__lowercase_identifier,
    ACTIONS(2952), 1,
      anon_sym_LPAREN,
    ACTIONS(2956), 1,
      anon_sym_LBRACE,
    ACTIONS(2958), 1,
      anon_sym_begin,
    ACTIONS(2960), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2962), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2964), 1,
      anon_sym_PERCENT,
    ACTIONS(2966), 1,
      aux_sym_number_token1,
    ACTIONS(2968), 1,
      anon_sym_SQUOTE,
    ACTIONS(2970), 1,
      anon_sym_DQUOTE,
    ACTIONS(2972), 1,
      sym_prefix_operator,
    STATE(1828), 1,
      sym_attribute,
    STATE(1913), 1,
      sym__attribute,
    STATE(2327), 1,
      sym__simple_expression,
    STATE(2366), 1,
      sym_parenthesized_operator,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(7268), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2954), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2279), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [7375] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2974), 1,
      anon_sym_EQ,
    ACTIONS(2976), 1,
      anon_sym_COLON2,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(1829), 1,
      sym_attribute,
    STATE(1841), 1,
      aux_sym_let_binding_repeat1,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(5769), 1,
      sym__class_typed,
    STATE(5915), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2381), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [7509] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2976), 1,
      anon_sym_COLON2,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2980), 1,
      anon_sym_EQ,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(1830), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(5641), 1,
      sym__class_typed,
    STATE(5961), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2393), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [7643] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2976), 1,
      anon_sym_COLON2,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2982), 1,
      anon_sym_EQ,
    STATE(1831), 1,
      sym_attribute,
    STATE(1832), 1,
      aux_sym_let_binding_repeat1,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(5644), 1,
      sym__class_typed,
    STATE(5964), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2399), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [7777] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2976), 1,
      anon_sym_COLON2,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2984), 1,
      anon_sym_EQ,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(1832), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(5760), 1,
      sym__class_typed,
    STATE(5799), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2405), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [7911] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1833), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6179), 1,
      sym__simple_expression,
    STATE(6699), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8037] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1834), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6232), 1,
      sym__simple_expression,
    STATE(6507), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8163] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1835), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6235), 1,
      sym__simple_expression,
    STATE(6564), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8289] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2964), 1,
      anon_sym_PERCENT,
    ACTIONS(2986), 1,
      sym__lowercase_identifier,
    ACTIONS(2988), 1,
      anon_sym_LPAREN,
    ACTIONS(2990), 1,
      anon_sym_LBRACK,
    ACTIONS(2994), 1,
      anon_sym_LBRACE,
    ACTIONS(2996), 1,
      anon_sym_object,
    ACTIONS(2998), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3000), 1,
      anon_sym_new,
    ACTIONS(3002), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3004), 1,
      anon_sym_begin,
    ACTIONS(3006), 1,
      sym_ocamlyacc_value,
    ACTIONS(3008), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3010), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3012), 1,
      aux_sym_number_token1,
    ACTIONS(3014), 1,
      anon_sym_SQUOTE,
    ACTIONS(3016), 1,
      anon_sym_DQUOTE,
    ACTIONS(3018), 1,
      sym_prefix_operator,
    ACTIONS(3020), 1,
      sym__uppercase_identifier,
    ACTIONS(3022), 1,
      aux_sym_tag_token1,
    STATE(1836), 1,
      sym_attribute,
    STATE(1914), 1,
      sym__attribute,
    STATE(2486), 1,
      sym__value_name,
    STATE(2508), 1,
      sym_parenthesized_operator,
    STATE(2517), 1,
      sym__constructor_name,
    STATE(2557), 1,
      sym__simple_expression,
    STATE(7598), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2992), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2516), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2450), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2449), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8415] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2976), 1,
      anon_sym_COLON2,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3024), 1,
      anon_sym_EQ,
    STATE(1837), 1,
      sym_attribute,
    STATE(1852), 1,
      aux_sym_let_binding_repeat1,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(5745), 1,
      sym__class_typed,
    STATE(5805), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2427), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [8549] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2964), 1,
      anon_sym_PERCENT,
    ACTIONS(3026), 1,
      sym__lowercase_identifier,
    ACTIONS(3028), 1,
      anon_sym_LPAREN,
    ACTIONS(3030), 1,
      anon_sym_LBRACK,
    ACTIONS(3034), 1,
      anon_sym_LBRACE,
    ACTIONS(3036), 1,
      anon_sym_object,
    ACTIONS(3038), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3040), 1,
      anon_sym_new,
    ACTIONS(3042), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3044), 1,
      anon_sym_begin,
    ACTIONS(3046), 1,
      sym_ocamlyacc_value,
    ACTIONS(3048), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3050), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3052), 1,
      aux_sym_number_token1,
    ACTIONS(3054), 1,
      anon_sym_SQUOTE,
    ACTIONS(3056), 1,
      anon_sym_DQUOTE,
    ACTIONS(3058), 1,
      sym_prefix_operator,
    ACTIONS(3060), 1,
      sym__uppercase_identifier,
    ACTIONS(3062), 1,
      aux_sym_tag_token1,
    STATE(1838), 1,
      sym_attribute,
    STATE(1881), 1,
      sym__attribute,
    STATE(3296), 1,
      sym_parenthesized_operator,
    STATE(3325), 1,
      sym__simple_expression,
    STATE(3335), 1,
      sym__constructor_name,
    STATE(3360), 1,
      sym__value_name,
    STATE(7527), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3032), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3352), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3310), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(3332), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8675] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2964), 1,
      anon_sym_PERCENT,
    ACTIONS(3026), 1,
      sym__lowercase_identifier,
    ACTIONS(3028), 1,
      anon_sym_LPAREN,
    ACTIONS(3030), 1,
      anon_sym_LBRACK,
    ACTIONS(3034), 1,
      anon_sym_LBRACE,
    ACTIONS(3036), 1,
      anon_sym_object,
    ACTIONS(3038), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3040), 1,
      anon_sym_new,
    ACTIONS(3042), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3044), 1,
      anon_sym_begin,
    ACTIONS(3046), 1,
      sym_ocamlyacc_value,
    ACTIONS(3048), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3050), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3052), 1,
      aux_sym_number_token1,
    ACTIONS(3054), 1,
      anon_sym_SQUOTE,
    ACTIONS(3056), 1,
      anon_sym_DQUOTE,
    ACTIONS(3058), 1,
      sym_prefix_operator,
    ACTIONS(3060), 1,
      sym__uppercase_identifier,
    ACTIONS(3062), 1,
      aux_sym_tag_token1,
    STATE(1839), 1,
      sym_attribute,
    STATE(1882), 1,
      sym__attribute,
    STATE(3296), 1,
      sym_parenthesized_operator,
    STATE(3326), 1,
      sym__simple_expression,
    STATE(3335), 1,
      sym__constructor_name,
    STATE(3360), 1,
      sym__value_name,
    STATE(7527), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3032), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3352), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3310), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(3332), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8801] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2964), 1,
      anon_sym_PERCENT,
    ACTIONS(3064), 1,
      sym__lowercase_identifier,
    ACTIONS(3066), 1,
      anon_sym_LPAREN,
    ACTIONS(3068), 1,
      anon_sym_LBRACK,
    ACTIONS(3072), 1,
      anon_sym_LBRACE,
    ACTIONS(3074), 1,
      anon_sym_object,
    ACTIONS(3076), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3078), 1,
      anon_sym_new,
    ACTIONS(3080), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3082), 1,
      anon_sym_begin,
    ACTIONS(3084), 1,
      sym_ocamlyacc_value,
    ACTIONS(3086), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3088), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3090), 1,
      aux_sym_number_token1,
    ACTIONS(3092), 1,
      anon_sym_SQUOTE,
    ACTIONS(3094), 1,
      anon_sym_DQUOTE,
    ACTIONS(3096), 1,
      sym_prefix_operator,
    ACTIONS(3098), 1,
      sym__uppercase_identifier,
    ACTIONS(3100), 1,
      aux_sym_tag_token1,
    STATE(1840), 1,
      sym_attribute,
    STATE(1891), 1,
      sym__attribute,
    STATE(2122), 1,
      sym__constructor_name,
    STATE(2447), 1,
      sym__simple_expression,
    STATE(2451), 1,
      sym_parenthesized_operator,
    STATE(2553), 1,
      sym__value_name,
    STATE(7715), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3070), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2617), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2552), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2551), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8927] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2976), 1,
      anon_sym_COLON2,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3102), 1,
      anon_sym_EQ,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(1841), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(5684), 1,
      sym__class_typed,
    STATE(5869), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2387), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [9061] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2976), 1,
      anon_sym_COLON2,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    STATE(1830), 1,
      aux_sym_let_binding_repeat1,
    STATE(1842), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(5701), 1,
      sym__class_typed,
    STATE(5876), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2375), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [9195] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1843), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6026), 1,
      sym__simple_expression,
    STATE(6350), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9321] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1844), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6216), 1,
      sym__simple_expression,
    STATE(6334), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9447] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2976), 1,
      anon_sym_COLON2,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3106), 1,
      anon_sym_EQ,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(1845), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(5765), 1,
      sym__class_typed,
    STATE(5988), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2421), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [9581] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1846), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6208), 1,
      sym__simple_expression,
    STATE(6617), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9707] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1847), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6201), 1,
      sym__simple_expression,
    STATE(6553), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9833] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1848), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6236), 1,
      sym__simple_expression,
    STATE(6635), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9959] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2964), 1,
      anon_sym_PERCENT,
    ACTIONS(3064), 1,
      sym__lowercase_identifier,
    ACTIONS(3066), 1,
      anon_sym_LPAREN,
    ACTIONS(3068), 1,
      anon_sym_LBRACK,
    ACTIONS(3072), 1,
      anon_sym_LBRACE,
    ACTIONS(3074), 1,
      anon_sym_object,
    ACTIONS(3076), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3078), 1,
      anon_sym_new,
    ACTIONS(3080), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3082), 1,
      anon_sym_begin,
    ACTIONS(3084), 1,
      sym_ocamlyacc_value,
    ACTIONS(3086), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3088), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3090), 1,
      aux_sym_number_token1,
    ACTIONS(3092), 1,
      anon_sym_SQUOTE,
    ACTIONS(3094), 1,
      anon_sym_DQUOTE,
    ACTIONS(3096), 1,
      sym_prefix_operator,
    ACTIONS(3098), 1,
      sym__uppercase_identifier,
    ACTIONS(3100), 1,
      aux_sym_tag_token1,
    STATE(1849), 1,
      sym_attribute,
    STATE(1892), 1,
      sym__attribute,
    STATE(2122), 1,
      sym__constructor_name,
    STATE(2448), 1,
      sym__simple_expression,
    STATE(2451), 1,
      sym_parenthesized_operator,
    STATE(2553), 1,
      sym__value_name,
    STATE(7715), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3070), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2617), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2552), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2551), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10085] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2976), 1,
      anon_sym_COLON2,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3108), 1,
      anon_sym_EQ,
    STATE(1845), 1,
      aux_sym_let_binding_repeat1,
    STATE(1850), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(5693), 1,
      sym__class_typed,
    STATE(5932), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2333), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [10219] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2964), 1,
      anon_sym_PERCENT,
    ACTIONS(3110), 1,
      sym__lowercase_identifier,
    ACTIONS(3112), 1,
      anon_sym_LPAREN,
    ACTIONS(3114), 1,
      anon_sym_LBRACK,
    ACTIONS(3118), 1,
      anon_sym_LBRACE,
    ACTIONS(3120), 1,
      anon_sym_object,
    ACTIONS(3122), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3124), 1,
      anon_sym_new,
    ACTIONS(3126), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3128), 1,
      anon_sym_begin,
    ACTIONS(3130), 1,
      sym_ocamlyacc_value,
    ACTIONS(3132), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3134), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3136), 1,
      aux_sym_number_token1,
    ACTIONS(3138), 1,
      anon_sym_SQUOTE,
    ACTIONS(3140), 1,
      anon_sym_DQUOTE,
    ACTIONS(3142), 1,
      sym_prefix_operator,
    ACTIONS(3144), 1,
      sym__uppercase_identifier,
    ACTIONS(3146), 1,
      aux_sym_tag_token1,
    STATE(1851), 1,
      sym_attribute,
    STATE(1889), 1,
      sym__attribute,
    STATE(2203), 1,
      sym__simple_expression,
    STATE(2252), 1,
      sym__constructor_name,
    STATE(2297), 1,
      sym__value_name,
    STATE(2413), 1,
      sym_parenthesized_operator,
    STATE(7376), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3116), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2432), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2403), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2374), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10345] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2976), 1,
      anon_sym_COLON2,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3148), 1,
      anon_sym_EQ,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(1852), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(5696), 1,
      sym__class_typed,
    STATE(5937), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2411), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [10479] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1853), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6222), 1,
      sym__simple_expression,
    STATE(6396), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10605] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1854), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6239), 1,
      sym__simple_expression,
    STATE(6668), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10731] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2964), 1,
      anon_sym_PERCENT,
    ACTIONS(2986), 1,
      sym__lowercase_identifier,
    ACTIONS(2988), 1,
      anon_sym_LPAREN,
    ACTIONS(2990), 1,
      anon_sym_LBRACK,
    ACTIONS(2994), 1,
      anon_sym_LBRACE,
    ACTIONS(2996), 1,
      anon_sym_object,
    ACTIONS(2998), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3000), 1,
      anon_sym_new,
    ACTIONS(3002), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3004), 1,
      anon_sym_begin,
    ACTIONS(3006), 1,
      sym_ocamlyacc_value,
    ACTIONS(3008), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3010), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3012), 1,
      aux_sym_number_token1,
    ACTIONS(3014), 1,
      anon_sym_SQUOTE,
    ACTIONS(3016), 1,
      anon_sym_DQUOTE,
    ACTIONS(3018), 1,
      sym_prefix_operator,
    ACTIONS(3020), 1,
      sym__uppercase_identifier,
    ACTIONS(3022), 1,
      aux_sym_tag_token1,
    STATE(1855), 1,
      sym_attribute,
    STATE(1918), 1,
      sym__attribute,
    STATE(2486), 1,
      sym__value_name,
    STATE(2508), 1,
      sym_parenthesized_operator,
    STATE(2517), 1,
      sym__constructor_name,
    STATE(2558), 1,
      sym__simple_expression,
    STATE(7598), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2992), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2516), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2450), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2449), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10857] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1856), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6244), 1,
      sym__simple_expression,
    STATE(6752), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10983] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2916), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1857), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5342), 1,
      sym_field_path,
    STATE(6226), 1,
      sym__simple_expression,
    STATE(6449), 1,
      sym_field_expression,
    STATE(7729), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11109] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2950), 1,
      sym__lowercase_identifier,
    ACTIONS(2952), 1,
      anon_sym_LPAREN,
    ACTIONS(2956), 1,
      anon_sym_LBRACE,
    ACTIONS(2958), 1,
      anon_sym_begin,
    ACTIONS(2960), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2962), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2964), 1,
      anon_sym_PERCENT,
    ACTIONS(2966), 1,
      aux_sym_number_token1,
    ACTIONS(2968), 1,
      anon_sym_SQUOTE,
    ACTIONS(2970), 1,
      anon_sym_DQUOTE,
    ACTIONS(2972), 1,
      sym_prefix_operator,
    STATE(1858), 1,
      sym_attribute,
    STATE(1916), 1,
      sym__attribute,
    STATE(2326), 1,
      sym__simple_expression,
    STATE(2366), 1,
      sym_parenthesized_operator,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(7268), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2954), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2279), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11235] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2964), 1,
      anon_sym_PERCENT,
    ACTIONS(3110), 1,
      sym__lowercase_identifier,
    ACTIONS(3112), 1,
      anon_sym_LPAREN,
    ACTIONS(3114), 1,
      anon_sym_LBRACK,
    ACTIONS(3118), 1,
      anon_sym_LBRACE,
    ACTIONS(3120), 1,
      anon_sym_object,
    ACTIONS(3122), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3124), 1,
      anon_sym_new,
    ACTIONS(3126), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3128), 1,
      anon_sym_begin,
    ACTIONS(3130), 1,
      sym_ocamlyacc_value,
    ACTIONS(3132), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3134), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3136), 1,
      aux_sym_number_token1,
    ACTIONS(3138), 1,
      anon_sym_SQUOTE,
    ACTIONS(3140), 1,
      anon_sym_DQUOTE,
    ACTIONS(3142), 1,
      sym_prefix_operator,
    ACTIONS(3144), 1,
      sym__uppercase_identifier,
    ACTIONS(3146), 1,
      aux_sym_tag_token1,
    STATE(1859), 1,
      sym_attribute,
    STATE(1887), 1,
      sym__attribute,
    STATE(2202), 1,
      sym__simple_expression,
    STATE(2252), 1,
      sym__constructor_name,
    STATE(2297), 1,
      sym__value_name,
    STATE(2413), 1,
      sym_parenthesized_operator,
    STATE(7376), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3116), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2432), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2403), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2374), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11361] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(1547), 1,
      anon_sym_RPAREN,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1860), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11484] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(1565), 1,
      anon_sym_RPAREN,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1861), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11607] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(1539), 1,
      anon_sym_RPAREN,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1862), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11730] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(1549), 1,
      anon_sym_RPAREN,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1863), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11853] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1864), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3152), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3150), 42,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [11924] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(1511), 1,
      anon_sym_RPAREN,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1865), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12047] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(1525), 1,
      anon_sym_RPAREN,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1866), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12170] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(1519), 1,
      anon_sym_RPAREN,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1867), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12293] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(1569), 1,
      anon_sym_RPAREN,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1868), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12416] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(1557), 1,
      anon_sym_RPAREN,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1869), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12539] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(1533), 1,
      anon_sym_RPAREN,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1870), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12662] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(1529), 1,
      anon_sym_RPAREN,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1871), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12785] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(1561), 1,
      anon_sym_RPAREN,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1872), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12908] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1873), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3154), 42,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [12979] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(1553), 1,
      anon_sym_RPAREN,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1874), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13102] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1875), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3158), 42,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [13173] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(1543), 1,
      anon_sym_RPAREN,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1876), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13296] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2986), 1,
      sym__lowercase_identifier,
    ACTIONS(2988), 1,
      anon_sym_LPAREN,
    ACTIONS(2990), 1,
      anon_sym_LBRACK,
    ACTIONS(2994), 1,
      anon_sym_LBRACE,
    ACTIONS(2996), 1,
      anon_sym_object,
    ACTIONS(2998), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3000), 1,
      anon_sym_new,
    ACTIONS(3002), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3004), 1,
      anon_sym_begin,
    ACTIONS(3006), 1,
      sym_ocamlyacc_value,
    ACTIONS(3008), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3010), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3012), 1,
      aux_sym_number_token1,
    ACTIONS(3014), 1,
      anon_sym_SQUOTE,
    ACTIONS(3016), 1,
      anon_sym_DQUOTE,
    ACTIONS(3018), 1,
      sym_prefix_operator,
    ACTIONS(3020), 1,
      sym__uppercase_identifier,
    ACTIONS(3022), 1,
      aux_sym_tag_token1,
    STATE(1877), 1,
      sym_attribute,
    STATE(2486), 1,
      sym__value_name,
    STATE(2508), 1,
      sym_parenthesized_operator,
    STATE(2517), 1,
      sym__constructor_name,
    STATE(2579), 1,
      sym__simple_expression,
    STATE(7598), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2992), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2516), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2450), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2449), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13416] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3026), 1,
      sym__lowercase_identifier,
    ACTIONS(3028), 1,
      anon_sym_LPAREN,
    ACTIONS(3030), 1,
      anon_sym_LBRACK,
    ACTIONS(3034), 1,
      anon_sym_LBRACE,
    ACTIONS(3036), 1,
      anon_sym_object,
    ACTIONS(3038), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3040), 1,
      anon_sym_new,
    ACTIONS(3042), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3044), 1,
      anon_sym_begin,
    ACTIONS(3046), 1,
      sym_ocamlyacc_value,
    ACTIONS(3048), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3050), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3052), 1,
      aux_sym_number_token1,
    ACTIONS(3054), 1,
      anon_sym_SQUOTE,
    ACTIONS(3056), 1,
      anon_sym_DQUOTE,
    ACTIONS(3058), 1,
      sym_prefix_operator,
    ACTIONS(3060), 1,
      sym__uppercase_identifier,
    ACTIONS(3062), 1,
      aux_sym_tag_token1,
    STATE(1878), 1,
      sym_attribute,
    STATE(3296), 1,
      sym_parenthesized_operator,
    STATE(3313), 1,
      sym__simple_expression,
    STATE(3335), 1,
      sym__constructor_name,
    STATE(3360), 1,
      sym__value_name,
    STATE(7527), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3032), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3352), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3310), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(3332), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13536] = 30,
    ACTIONS(407), 1,
      anon_sym_LPAREN,
    ACTIONS(409), 1,
      anon_sym_LBRACK,
    ACTIONS(413), 1,
      anon_sym_LBRACE,
    ACTIONS(415), 1,
      anon_sym_object,
    ACTIONS(419), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(437), 1,
      anon_sym_new,
    ACTIONS(439), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(441), 1,
      anon_sym_begin,
    ACTIONS(443), 1,
      sym_ocamlyacc_value,
    ACTIONS(449), 1,
      aux_sym_number_token1,
    ACTIONS(451), 1,
      anon_sym_SQUOTE,
    ACTIONS(453), 1,
      anon_sym_DQUOTE,
    ACTIONS(455), 1,
      sym_prefix_operator,
    ACTIONS(459), 1,
      sym__uppercase_identifier,
    ACTIONS(461), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(479), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(481), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(989), 1,
      sym__lowercase_identifier,
    STATE(1073), 1,
      sym__value_name,
    STATE(1074), 1,
      sym_parenthesized_operator,
    STATE(1107), 1,
      sym__constructor_name,
    STATE(1125), 1,
      sym__simple_expression,
    STATE(1879), 1,
      sym_attribute,
    STATE(7491), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(411), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1106), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1046), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1045), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13656] = 30,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(25), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_LBRACE,
    ACTIONS(43), 1,
      anon_sym_object,
    ACTIONS(47), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(65), 1,
      anon_sym_new,
    ACTIONS(67), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(69), 1,
      anon_sym_begin,
    ACTIONS(71), 1,
      sym_ocamlyacc_value,
    ACTIONS(83), 1,
      aux_sym_number_token1,
    ACTIONS(87), 1,
      anon_sym_SQUOTE,
    ACTIONS(89), 1,
      anon_sym_DQUOTE,
    ACTIONS(91), 1,
      sym_prefix_operator,
    ACTIONS(99), 1,
      sym__uppercase_identifier,
    ACTIONS(103), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1153), 1,
      sym__lowercase_identifier,
    ACTIONS(1487), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1489), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1442), 1,
      sym__value_name,
    STATE(1458), 1,
      sym__simple_expression,
    STATE(1483), 1,
      sym_parenthesized_operator,
    STATE(1484), 1,
      sym__constructor_name,
    STATE(1880), 1,
      sym_attribute,
    STATE(7286), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(27), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1441), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1418), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1471), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13776] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3026), 1,
      sym__lowercase_identifier,
    ACTIONS(3028), 1,
      anon_sym_LPAREN,
    ACTIONS(3030), 1,
      anon_sym_LBRACK,
    ACTIONS(3034), 1,
      anon_sym_LBRACE,
    ACTIONS(3036), 1,
      anon_sym_object,
    ACTIONS(3038), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3040), 1,
      anon_sym_new,
    ACTIONS(3042), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3044), 1,
      anon_sym_begin,
    ACTIONS(3046), 1,
      sym_ocamlyacc_value,
    ACTIONS(3048), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3050), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3052), 1,
      aux_sym_number_token1,
    ACTIONS(3054), 1,
      anon_sym_SQUOTE,
    ACTIONS(3056), 1,
      anon_sym_DQUOTE,
    ACTIONS(3058), 1,
      sym_prefix_operator,
    ACTIONS(3060), 1,
      sym__uppercase_identifier,
    ACTIONS(3062), 1,
      aux_sym_tag_token1,
    STATE(1881), 1,
      sym_attribute,
    STATE(3296), 1,
      sym_parenthesized_operator,
    STATE(3335), 1,
      sym__constructor_name,
    STATE(3356), 1,
      sym__simple_expression,
    STATE(3360), 1,
      sym__value_name,
    STATE(7527), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3032), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3352), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3310), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(3332), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13896] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3026), 1,
      sym__lowercase_identifier,
    ACTIONS(3028), 1,
      anon_sym_LPAREN,
    ACTIONS(3030), 1,
      anon_sym_LBRACK,
    ACTIONS(3034), 1,
      anon_sym_LBRACE,
    ACTIONS(3036), 1,
      anon_sym_object,
    ACTIONS(3038), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3040), 1,
      anon_sym_new,
    ACTIONS(3042), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3044), 1,
      anon_sym_begin,
    ACTIONS(3046), 1,
      sym_ocamlyacc_value,
    ACTIONS(3048), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3050), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3052), 1,
      aux_sym_number_token1,
    ACTIONS(3054), 1,
      anon_sym_SQUOTE,
    ACTIONS(3056), 1,
      anon_sym_DQUOTE,
    ACTIONS(3058), 1,
      sym_prefix_operator,
    ACTIONS(3060), 1,
      sym__uppercase_identifier,
    ACTIONS(3062), 1,
      aux_sym_tag_token1,
    STATE(1882), 1,
      sym_attribute,
    STATE(3296), 1,
      sym_parenthesized_operator,
    STATE(3335), 1,
      sym__constructor_name,
    STATE(3359), 1,
      sym__simple_expression,
    STATE(3360), 1,
      sym__value_name,
    STATE(7527), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3032), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3352), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3310), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(3332), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14016] = 30,
    ACTIONS(407), 1,
      anon_sym_LPAREN,
    ACTIONS(409), 1,
      anon_sym_LBRACK,
    ACTIONS(413), 1,
      anon_sym_LBRACE,
    ACTIONS(415), 1,
      anon_sym_object,
    ACTIONS(419), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(437), 1,
      anon_sym_new,
    ACTIONS(439), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(441), 1,
      anon_sym_begin,
    ACTIONS(443), 1,
      sym_ocamlyacc_value,
    ACTIONS(449), 1,
      aux_sym_number_token1,
    ACTIONS(451), 1,
      anon_sym_SQUOTE,
    ACTIONS(453), 1,
      anon_sym_DQUOTE,
    ACTIONS(455), 1,
      sym_prefix_operator,
    ACTIONS(459), 1,
      sym__uppercase_identifier,
    ACTIONS(461), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(479), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(481), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(989), 1,
      sym__lowercase_identifier,
    STATE(1058), 1,
      sym__simple_expression,
    STATE(1073), 1,
      sym__value_name,
    STATE(1074), 1,
      sym_parenthesized_operator,
    STATE(1107), 1,
      sym__constructor_name,
    STATE(1883), 1,
      sym_attribute,
    STATE(7491), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(411), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1106), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1046), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1045), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14136] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3110), 1,
      sym__lowercase_identifier,
    ACTIONS(3112), 1,
      anon_sym_LPAREN,
    ACTIONS(3114), 1,
      anon_sym_LBRACK,
    ACTIONS(3118), 1,
      anon_sym_LBRACE,
    ACTIONS(3120), 1,
      anon_sym_object,
    ACTIONS(3122), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3124), 1,
      anon_sym_new,
    ACTIONS(3126), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3128), 1,
      anon_sym_begin,
    ACTIONS(3130), 1,
      sym_ocamlyacc_value,
    ACTIONS(3132), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3134), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3136), 1,
      aux_sym_number_token1,
    ACTIONS(3138), 1,
      anon_sym_SQUOTE,
    ACTIONS(3140), 1,
      anon_sym_DQUOTE,
    ACTIONS(3142), 1,
      sym_prefix_operator,
    ACTIONS(3144), 1,
      sym__uppercase_identifier,
    ACTIONS(3146), 1,
      aux_sym_tag_token1,
    STATE(1884), 1,
      sym_attribute,
    STATE(2252), 1,
      sym__constructor_name,
    STATE(2297), 1,
      sym__value_name,
    STATE(2341), 1,
      sym__simple_expression,
    STATE(2413), 1,
      sym_parenthesized_operator,
    STATE(7376), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3116), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2432), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2403), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2374), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14256] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2556), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1885), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(6136), 1,
      sym__simple_expression,
    STATE(7330), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14376] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    STATE(941), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1886), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14496] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3110), 1,
      sym__lowercase_identifier,
    ACTIONS(3112), 1,
      anon_sym_LPAREN,
    ACTIONS(3114), 1,
      anon_sym_LBRACK,
    ACTIONS(3118), 1,
      anon_sym_LBRACE,
    ACTIONS(3120), 1,
      anon_sym_object,
    ACTIONS(3122), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3124), 1,
      anon_sym_new,
    ACTIONS(3126), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3128), 1,
      anon_sym_begin,
    ACTIONS(3130), 1,
      sym_ocamlyacc_value,
    ACTIONS(3132), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3134), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3136), 1,
      aux_sym_number_token1,
    ACTIONS(3138), 1,
      anon_sym_SQUOTE,
    ACTIONS(3140), 1,
      anon_sym_DQUOTE,
    ACTIONS(3142), 1,
      sym_prefix_operator,
    ACTIONS(3144), 1,
      sym__uppercase_identifier,
    ACTIONS(3146), 1,
      aux_sym_tag_token1,
    STATE(1887), 1,
      sym_attribute,
    STATE(2252), 1,
      sym__constructor_name,
    STATE(2276), 1,
      sym__simple_expression,
    STATE(2297), 1,
      sym__value_name,
    STATE(2413), 1,
      sym_parenthesized_operator,
    STATE(7376), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3116), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2432), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2403), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2374), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14616] = 30,
    ACTIONS(187), 1,
      anon_sym_LPAREN,
    ACTIONS(189), 1,
      anon_sym_LBRACK,
    ACTIONS(193), 1,
      anon_sym_LBRACE,
    ACTIONS(195), 1,
      anon_sym_object,
    ACTIONS(199), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(209), 1,
      anon_sym_new,
    ACTIONS(211), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(213), 1,
      anon_sym_begin,
    ACTIONS(215), 1,
      sym_ocamlyacc_value,
    ACTIONS(221), 1,
      aux_sym_number_token1,
    ACTIONS(223), 1,
      anon_sym_SQUOTE,
    ACTIONS(225), 1,
      anon_sym_DQUOTE,
    ACTIONS(227), 1,
      sym_prefix_operator,
    ACTIONS(231), 1,
      sym__uppercase_identifier,
    ACTIONS(233), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1013), 1,
      sym__lowercase_identifier,
    ACTIONS(1505), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1507), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1235), 1,
      sym__constructor_name,
    STATE(1283), 1,
      sym_parenthesized_operator,
    STATE(1296), 1,
      sym__simple_expression,
    STATE(1322), 1,
      sym__value_name,
    STATE(1888), 1,
      sym_attribute,
    STATE(7415), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(191), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1267), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1298), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1295), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14736] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3110), 1,
      sym__lowercase_identifier,
    ACTIONS(3112), 1,
      anon_sym_LPAREN,
    ACTIONS(3114), 1,
      anon_sym_LBRACK,
    ACTIONS(3118), 1,
      anon_sym_LBRACE,
    ACTIONS(3120), 1,
      anon_sym_object,
    ACTIONS(3122), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3124), 1,
      anon_sym_new,
    ACTIONS(3126), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3128), 1,
      anon_sym_begin,
    ACTIONS(3130), 1,
      sym_ocamlyacc_value,
    ACTIONS(3132), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3134), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3136), 1,
      aux_sym_number_token1,
    ACTIONS(3138), 1,
      anon_sym_SQUOTE,
    ACTIONS(3140), 1,
      anon_sym_DQUOTE,
    ACTIONS(3142), 1,
      sym_prefix_operator,
    ACTIONS(3144), 1,
      sym__uppercase_identifier,
    ACTIONS(3146), 1,
      aux_sym_tag_token1,
    STATE(1889), 1,
      sym_attribute,
    STATE(2252), 1,
      sym__constructor_name,
    STATE(2278), 1,
      sym__simple_expression,
    STATE(2297), 1,
      sym__value_name,
    STATE(2413), 1,
      sym_parenthesized_operator,
    STATE(7376), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3116), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2432), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2403), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2374), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14856] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3064), 1,
      sym__lowercase_identifier,
    ACTIONS(3066), 1,
      anon_sym_LPAREN,
    ACTIONS(3068), 1,
      anon_sym_LBRACK,
    ACTIONS(3072), 1,
      anon_sym_LBRACE,
    ACTIONS(3074), 1,
      anon_sym_object,
    ACTIONS(3076), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3078), 1,
      anon_sym_new,
    ACTIONS(3080), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3082), 1,
      anon_sym_begin,
    ACTIONS(3084), 1,
      sym_ocamlyacc_value,
    ACTIONS(3086), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3088), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3090), 1,
      aux_sym_number_token1,
    ACTIONS(3092), 1,
      anon_sym_SQUOTE,
    ACTIONS(3094), 1,
      anon_sym_DQUOTE,
    ACTIONS(3096), 1,
      sym_prefix_operator,
    ACTIONS(3098), 1,
      sym__uppercase_identifier,
    ACTIONS(3100), 1,
      aux_sym_tag_token1,
    STATE(1890), 1,
      sym_attribute,
    STATE(2122), 1,
      sym__constructor_name,
    STATE(2451), 1,
      sym_parenthesized_operator,
    STATE(2502), 1,
      sym__simple_expression,
    STATE(2553), 1,
      sym__value_name,
    STATE(7715), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3070), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2617), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2552), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2551), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14976] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3064), 1,
      sym__lowercase_identifier,
    ACTIONS(3066), 1,
      anon_sym_LPAREN,
    ACTIONS(3068), 1,
      anon_sym_LBRACK,
    ACTIONS(3072), 1,
      anon_sym_LBRACE,
    ACTIONS(3074), 1,
      anon_sym_object,
    ACTIONS(3076), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3078), 1,
      anon_sym_new,
    ACTIONS(3080), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3082), 1,
      anon_sym_begin,
    ACTIONS(3084), 1,
      sym_ocamlyacc_value,
    ACTIONS(3086), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3088), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3090), 1,
      aux_sym_number_token1,
    ACTIONS(3092), 1,
      anon_sym_SQUOTE,
    ACTIONS(3094), 1,
      anon_sym_DQUOTE,
    ACTIONS(3096), 1,
      sym_prefix_operator,
    ACTIONS(3098), 1,
      sym__uppercase_identifier,
    ACTIONS(3100), 1,
      aux_sym_tag_token1,
    STATE(1891), 1,
      sym_attribute,
    STATE(2122), 1,
      sym__constructor_name,
    STATE(2451), 1,
      sym_parenthesized_operator,
    STATE(2553), 1,
      sym__value_name,
    STATE(2626), 1,
      sym__simple_expression,
    STATE(7715), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3070), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2617), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2552), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2551), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [15096] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3064), 1,
      sym__lowercase_identifier,
    ACTIONS(3066), 1,
      anon_sym_LPAREN,
    ACTIONS(3068), 1,
      anon_sym_LBRACK,
    ACTIONS(3072), 1,
      anon_sym_LBRACE,
    ACTIONS(3074), 1,
      anon_sym_object,
    ACTIONS(3076), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3078), 1,
      anon_sym_new,
    ACTIONS(3080), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3082), 1,
      anon_sym_begin,
    ACTIONS(3084), 1,
      sym_ocamlyacc_value,
    ACTIONS(3086), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3088), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3090), 1,
      aux_sym_number_token1,
    ACTIONS(3092), 1,
      anon_sym_SQUOTE,
    ACTIONS(3094), 1,
      anon_sym_DQUOTE,
    ACTIONS(3096), 1,
      sym_prefix_operator,
    ACTIONS(3098), 1,
      sym__uppercase_identifier,
    ACTIONS(3100), 1,
      aux_sym_tag_token1,
    STATE(1892), 1,
      sym_attribute,
    STATE(2122), 1,
      sym__constructor_name,
    STATE(2451), 1,
      sym_parenthesized_operator,
    STATE(2553), 1,
      sym__value_name,
    STATE(2628), 1,
      sym__simple_expression,
    STATE(7715), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3070), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2617), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2552), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2551), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [15216] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(807), 1,
      anon_sym_LPAREN,
    ACTIONS(809), 1,
      anon_sym_LBRACK,
    ACTIONS(813), 1,
      anon_sym_LBRACE,
    ACTIONS(815), 1,
      anon_sym_object,
    ACTIONS(819), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(837), 1,
      anon_sym_new,
    ACTIONS(839), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(841), 1,
      anon_sym_begin,
    ACTIONS(843), 1,
      sym_ocamlyacc_value,
    ACTIONS(845), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(847), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(849), 1,
      aux_sym_number_token1,
    ACTIONS(851), 1,
      anon_sym_SQUOTE,
    ACTIONS(853), 1,
      anon_sym_DQUOTE,
    ACTIONS(855), 1,
      sym_prefix_operator,
    ACTIONS(859), 1,
      sym__uppercase_identifier,
    ACTIONS(861), 1,
      aux_sym_tag_token1,
    ACTIONS(1687), 1,
      sym__lowercase_identifier,
    STATE(1893), 1,
      sym_attribute,
    STATE(1958), 1,
      sym__simple_expression,
    STATE(1969), 1,
      sym__constructor_name,
    STATE(2018), 1,
      sym__value_name,
    STATE(2019), 1,
      sym_parenthesized_operator,
    STATE(7454), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(811), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1968), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2002), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1977), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [15336] = 30,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(25), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_LBRACE,
    ACTIONS(43), 1,
      anon_sym_object,
    ACTIONS(47), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(65), 1,
      anon_sym_new,
    ACTIONS(67), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(69), 1,
      anon_sym_begin,
    ACTIONS(71), 1,
      sym_ocamlyacc_value,
    ACTIONS(83), 1,
      aux_sym_number_token1,
    ACTIONS(87), 1,
      anon_sym_SQUOTE,
    ACTIONS(89), 1,
      anon_sym_DQUOTE,
    ACTIONS(91), 1,
      sym_prefix_operator,
    ACTIONS(99), 1,
      sym__uppercase_identifier,
    ACTIONS(103), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1153), 1,
      sym__lowercase_identifier,
    ACTIONS(1487), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1489), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1423), 1,
      sym__simple_expression,
    STATE(1442), 1,
      sym__value_name,
    STATE(1483), 1,
      sym_parenthesized_operator,
    STATE(1484), 1,
      sym__constructor_name,
    STATE(1894), 1,
      sym_attribute,
    STATE(7286), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(27), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1441), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1418), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1471), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [15456] = 30,
    ACTIONS(253), 1,
      anon_sym_LPAREN,
    ACTIONS(255), 1,
      anon_sym_LBRACK,
    ACTIONS(259), 1,
      anon_sym_LBRACE,
    ACTIONS(261), 1,
      anon_sym_object,
    ACTIONS(265), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(275), 1,
      anon_sym_new,
    ACTIONS(277), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(279), 1,
      anon_sym_begin,
    ACTIONS(281), 1,
      sym_ocamlyacc_value,
    ACTIONS(287), 1,
      aux_sym_number_token1,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      sym_prefix_operator,
    ACTIONS(297), 1,
      sym__uppercase_identifier,
    ACTIONS(299), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1083), 1,
      sym__lowercase_identifier,
    ACTIONS(1497), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1499), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1365), 1,
      sym__simple_expression,
    STATE(1366), 1,
      sym__value_name,
    STATE(1375), 1,
      sym__constructor_name,
    STATE(1378), 1,
      sym_parenthesized_operator,
    STATE(1895), 1,
      sym_attribute,
    STATE(7186), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(257), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1336), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1377), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1364), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [15576] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2556), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1896), 1,
      sym_attribute,
    STATE(2213), 1,
      sym__simple_expression,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(7330), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [15696] = 30,
    ACTIONS(187), 1,
      anon_sym_LPAREN,
    ACTIONS(189), 1,
      anon_sym_LBRACK,
    ACTIONS(193), 1,
      anon_sym_LBRACE,
    ACTIONS(195), 1,
      anon_sym_object,
    ACTIONS(199), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(209), 1,
      anon_sym_new,
    ACTIONS(211), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(213), 1,
      anon_sym_begin,
    ACTIONS(215), 1,
      sym_ocamlyacc_value,
    ACTIONS(221), 1,
      aux_sym_number_token1,
    ACTIONS(223), 1,
      anon_sym_SQUOTE,
    ACTIONS(225), 1,
      anon_sym_DQUOTE,
    ACTIONS(227), 1,
      sym_prefix_operator,
    ACTIONS(231), 1,
      sym__uppercase_identifier,
    ACTIONS(233), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1013), 1,
      sym__lowercase_identifier,
    ACTIONS(1505), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1507), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1235), 1,
      sym__constructor_name,
    STATE(1283), 1,
      sym_parenthesized_operator,
    STATE(1305), 1,
      sym__simple_expression,
    STATE(1322), 1,
      sym__value_name,
    STATE(1897), 1,
      sym_attribute,
    STATE(7415), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(191), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1267), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1298), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1295), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [15816] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3110), 1,
      sym__lowercase_identifier,
    ACTIONS(3112), 1,
      anon_sym_LPAREN,
    ACTIONS(3114), 1,
      anon_sym_LBRACK,
    ACTIONS(3118), 1,
      anon_sym_LBRACE,
    ACTIONS(3120), 1,
      anon_sym_object,
    ACTIONS(3122), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3124), 1,
      anon_sym_new,
    ACTIONS(3126), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3128), 1,
      anon_sym_begin,
    ACTIONS(3130), 1,
      sym_ocamlyacc_value,
    ACTIONS(3132), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3134), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3136), 1,
      aux_sym_number_token1,
    ACTIONS(3138), 1,
      anon_sym_SQUOTE,
    ACTIONS(3140), 1,
      anon_sym_DQUOTE,
    ACTIONS(3142), 1,
      sym_prefix_operator,
    ACTIONS(3144), 1,
      sym__uppercase_identifier,
    ACTIONS(3146), 1,
      aux_sym_tag_token1,
    STATE(1898), 1,
      sym_attribute,
    STATE(2252), 1,
      sym__constructor_name,
    STATE(2287), 1,
      sym__simple_expression,
    STATE(2297), 1,
      sym__value_name,
    STATE(2413), 1,
      sym_parenthesized_operator,
    STATE(7376), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3116), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2432), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2403), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2374), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [15936] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2211), 1,
      sym__lowercase_identifier,
    ACTIONS(2215), 1,
      anon_sym_LPAREN,
    ACTIONS(2217), 1,
      anon_sym_LBRACK,
    ACTIONS(2221), 1,
      anon_sym_LBRACE,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(2225), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2227), 1,
      anon_sym_new,
    ACTIONS(2229), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2231), 1,
      anon_sym_begin,
    ACTIONS(2233), 1,
      sym_ocamlyacc_value,
    ACTIONS(2239), 1,
      aux_sym_number_token1,
    ACTIONS(2241), 1,
      anon_sym_SQUOTE,
    ACTIONS(2243), 1,
      anon_sym_DQUOTE,
    ACTIONS(2245), 1,
      sym_prefix_operator,
    ACTIONS(2247), 1,
      sym__uppercase_identifier,
    ACTIONS(2249), 1,
      aux_sym_tag_token1,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1899), 1,
      sym_attribute,
    STATE(2973), 1,
      sym__constructor_name,
    STATE(2993), 1,
      sym__simple_expression,
    STATE(3037), 1,
      sym__value_name,
    STATE(3038), 1,
      sym_parenthesized_operator,
    STATE(7634), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2219), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2971), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3049), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(3044), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [16056] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    STATE(969), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1900), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [16176] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2211), 1,
      sym__lowercase_identifier,
    ACTIONS(2215), 1,
      anon_sym_LPAREN,
    ACTIONS(2217), 1,
      anon_sym_LBRACK,
    ACTIONS(2221), 1,
      anon_sym_LBRACE,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(2225), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2227), 1,
      anon_sym_new,
    ACTIONS(2229), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2231), 1,
      anon_sym_begin,
    ACTIONS(2233), 1,
      sym_ocamlyacc_value,
    ACTIONS(2239), 1,
      aux_sym_number_token1,
    ACTIONS(2241), 1,
      anon_sym_SQUOTE,
    ACTIONS(2243), 1,
      anon_sym_DQUOTE,
    ACTIONS(2245), 1,
      sym_prefix_operator,
    ACTIONS(2247), 1,
      sym__uppercase_identifier,
    ACTIONS(2249), 1,
      aux_sym_tag_token1,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1901), 1,
      sym_attribute,
    STATE(2931), 1,
      sym__simple_expression,
    STATE(2973), 1,
      sym__constructor_name,
    STATE(3037), 1,
      sym__value_name,
    STATE(3038), 1,
      sym_parenthesized_operator,
    STATE(7634), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2219), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2971), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3049), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(3044), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [16296] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(489), 1,
      anon_sym_LPAREN,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_LBRACE,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(519), 1,
      anon_sym_new,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(525), 1,
      sym_ocamlyacc_value,
    ACTIONS(531), 1,
      aux_sym_number_token1,
    ACTIONS(533), 1,
      anon_sym_SQUOTE,
    ACTIONS(535), 1,
      anon_sym_DQUOTE,
    ACTIONS(537), 1,
      sym_prefix_operator,
    ACTIONS(541), 1,
      sym__uppercase_identifier,
    ACTIONS(543), 1,
      aux_sym_tag_token1,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    STATE(952), 1,
      sym__simple_expression,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(990), 1,
      sym__constructor_name,
    STATE(1011), 1,
      sym__value_name,
    STATE(1902), 1,
      sym_attribute,
    STATE(7385), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(493), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(988), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(946), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [16416] = 30,
    ACTIONS(253), 1,
      anon_sym_LPAREN,
    ACTIONS(255), 1,
      anon_sym_LBRACK,
    ACTIONS(259), 1,
      anon_sym_LBRACE,
    ACTIONS(261), 1,
      anon_sym_object,
    ACTIONS(265), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(275), 1,
      anon_sym_new,
    ACTIONS(277), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(279), 1,
      anon_sym_begin,
    ACTIONS(281), 1,
      sym_ocamlyacc_value,
    ACTIONS(287), 1,
      aux_sym_number_token1,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      sym_prefix_operator,
    ACTIONS(297), 1,
      sym__uppercase_identifier,
    ACTIONS(299), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1083), 1,
      sym__lowercase_identifier,
    ACTIONS(1497), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1499), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1366), 1,
      sym__value_name,
    STATE(1375), 1,
      sym__constructor_name,
    STATE(1378), 1,
      sym_parenthesized_operator,
    STATE(1410), 1,
      sym__simple_expression,
    STATE(1903), 1,
      sym_attribute,
    STATE(7186), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(257), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1336), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1377), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1364), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [16536] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3110), 1,
      sym__lowercase_identifier,
    ACTIONS(3112), 1,
      anon_sym_LPAREN,
    ACTIONS(3114), 1,
      anon_sym_LBRACK,
    ACTIONS(3118), 1,
      anon_sym_LBRACE,
    ACTIONS(3120), 1,
      anon_sym_object,
    ACTIONS(3122), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3124), 1,
      anon_sym_new,
    ACTIONS(3126), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3128), 1,
      anon_sym_begin,
    ACTIONS(3130), 1,
      sym_ocamlyacc_value,
    ACTIONS(3132), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3134), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3136), 1,
      aux_sym_number_token1,
    ACTIONS(3138), 1,
      anon_sym_SQUOTE,
    ACTIONS(3140), 1,
      anon_sym_DQUOTE,
    ACTIONS(3142), 1,
      sym_prefix_operator,
    ACTIONS(3144), 1,
      sym__uppercase_identifier,
    ACTIONS(3146), 1,
      aux_sym_tag_token1,
    STATE(1904), 1,
      sym_attribute,
    STATE(2252), 1,
      sym__constructor_name,
    STATE(2297), 1,
      sym__value_name,
    STATE(2347), 1,
      sym__simple_expression,
    STATE(2413), 1,
      sym_parenthesized_operator,
    STATE(7376), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3116), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2432), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2403), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2374), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [16656] = 30,
    ACTIONS(113), 1,
      anon_sym_LPAREN,
    ACTIONS(115), 1,
      anon_sym_LBRACK,
    ACTIONS(119), 1,
      anon_sym_LBRACE,
    ACTIONS(121), 1,
      anon_sym_object,
    ACTIONS(125), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(139), 1,
      anon_sym_new,
    ACTIONS(141), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(143), 1,
      anon_sym_begin,
    ACTIONS(145), 1,
      sym_ocamlyacc_value,
    ACTIONS(153), 1,
      aux_sym_number_token1,
    ACTIONS(155), 1,
      anon_sym_SQUOTE,
    ACTIONS(157), 1,
      anon_sym_DQUOTE,
    ACTIONS(159), 1,
      sym_prefix_operator,
    ACTIONS(165), 1,
      sym__uppercase_identifier,
    ACTIONS(167), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1003), 1,
      sym__lowercase_identifier,
    ACTIONS(1493), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1495), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1145), 1,
      sym__constructor_name,
    STATE(1153), 1,
      sym__value_name,
    STATE(1154), 1,
      sym_parenthesized_operator,
    STATE(1186), 1,
      sym__simple_expression,
    STATE(1905), 1,
      sym_attribute,
    STATE(7562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(117), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1141), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1190), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1189), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [16776] = 30,
    ACTIONS(113), 1,
      anon_sym_LPAREN,
    ACTIONS(115), 1,
      anon_sym_LBRACK,
    ACTIONS(119), 1,
      anon_sym_LBRACE,
    ACTIONS(121), 1,
      anon_sym_object,
    ACTIONS(125), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(139), 1,
      anon_sym_new,
    ACTIONS(141), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(143), 1,
      anon_sym_begin,
    ACTIONS(145), 1,
      sym_ocamlyacc_value,
    ACTIONS(153), 1,
      aux_sym_number_token1,
    ACTIONS(155), 1,
      anon_sym_SQUOTE,
    ACTIONS(157), 1,
      anon_sym_DQUOTE,
    ACTIONS(159), 1,
      sym_prefix_operator,
    ACTIONS(165), 1,
      sym__uppercase_identifier,
    ACTIONS(167), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1003), 1,
      sym__lowercase_identifier,
    ACTIONS(1493), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1495), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1145), 1,
      sym__constructor_name,
    STATE(1153), 1,
      sym__value_name,
    STATE(1154), 1,
      sym_parenthesized_operator,
    STATE(1204), 1,
      sym__simple_expression,
    STATE(1906), 1,
      sym_attribute,
    STATE(7562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(117), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1141), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1190), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1189), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [16896] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2950), 1,
      sym__lowercase_identifier,
    ACTIONS(2952), 1,
      anon_sym_LPAREN,
    ACTIONS(2956), 1,
      anon_sym_LBRACE,
    ACTIONS(2958), 1,
      anon_sym_begin,
    ACTIONS(2960), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2962), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2966), 1,
      aux_sym_number_token1,
    ACTIONS(2968), 1,
      anon_sym_SQUOTE,
    ACTIONS(2970), 1,
      anon_sym_DQUOTE,
    ACTIONS(2972), 1,
      sym_prefix_operator,
    STATE(1907), 1,
      sym_attribute,
    STATE(2213), 1,
      sym__simple_expression,
    STATE(2366), 1,
      sym_parenthesized_operator,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(7268), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2954), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2279), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [17016] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(807), 1,
      anon_sym_LPAREN,
    ACTIONS(809), 1,
      anon_sym_LBRACK,
    ACTIONS(813), 1,
      anon_sym_LBRACE,
    ACTIONS(815), 1,
      anon_sym_object,
    ACTIONS(819), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(837), 1,
      anon_sym_new,
    ACTIONS(839), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(841), 1,
      anon_sym_begin,
    ACTIONS(843), 1,
      sym_ocamlyacc_value,
    ACTIONS(845), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(847), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(849), 1,
      aux_sym_number_token1,
    ACTIONS(851), 1,
      anon_sym_SQUOTE,
    ACTIONS(853), 1,
      anon_sym_DQUOTE,
    ACTIONS(855), 1,
      sym_prefix_operator,
    ACTIONS(859), 1,
      sym__uppercase_identifier,
    ACTIONS(861), 1,
      aux_sym_tag_token1,
    ACTIONS(1687), 1,
      sym__lowercase_identifier,
    STATE(1908), 1,
      sym_attribute,
    STATE(1969), 1,
      sym__constructor_name,
    STATE(2017), 1,
      sym__simple_expression,
    STATE(2018), 1,
      sym__value_name,
    STATE(2019), 1,
      sym_parenthesized_operator,
    STATE(7454), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(811), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1968), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2002), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1977), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [17136] = 30,
    ACTIONS(113), 1,
      anon_sym_LPAREN,
    ACTIONS(115), 1,
      anon_sym_LBRACK,
    ACTIONS(119), 1,
      anon_sym_LBRACE,
    ACTIONS(121), 1,
      anon_sym_object,
    ACTIONS(125), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(139), 1,
      anon_sym_new,
    ACTIONS(141), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(143), 1,
      anon_sym_begin,
    ACTIONS(145), 1,
      sym_ocamlyacc_value,
    ACTIONS(153), 1,
      aux_sym_number_token1,
    ACTIONS(155), 1,
      anon_sym_SQUOTE,
    ACTIONS(157), 1,
      anon_sym_DQUOTE,
    ACTIONS(159), 1,
      sym_prefix_operator,
    ACTIONS(165), 1,
      sym__uppercase_identifier,
    ACTIONS(167), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1003), 1,
      sym__lowercase_identifier,
    ACTIONS(1493), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1495), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1145), 1,
      sym__constructor_name,
    STATE(1153), 1,
      sym__value_name,
    STATE(1154), 1,
      sym_parenthesized_operator,
    STATE(1171), 1,
      sym__simple_expression,
    STATE(1909), 1,
      sym_attribute,
    STATE(7562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(117), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1141), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1190), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1189), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [17256] = 30,
    ACTIONS(187), 1,
      anon_sym_LPAREN,
    ACTIONS(189), 1,
      anon_sym_LBRACK,
    ACTIONS(193), 1,
      anon_sym_LBRACE,
    ACTIONS(195), 1,
      anon_sym_object,
    ACTIONS(199), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(209), 1,
      anon_sym_new,
    ACTIONS(211), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(213), 1,
      anon_sym_begin,
    ACTIONS(215), 1,
      sym_ocamlyacc_value,
    ACTIONS(221), 1,
      aux_sym_number_token1,
    ACTIONS(223), 1,
      anon_sym_SQUOTE,
    ACTIONS(225), 1,
      anon_sym_DQUOTE,
    ACTIONS(227), 1,
      sym_prefix_operator,
    ACTIONS(231), 1,
      sym__uppercase_identifier,
    ACTIONS(233), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1013), 1,
      sym__lowercase_identifier,
    ACTIONS(1505), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1507), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1235), 1,
      sym__constructor_name,
    STATE(1239), 1,
      sym__simple_expression,
    STATE(1283), 1,
      sym_parenthesized_operator,
    STATE(1322), 1,
      sym__value_name,
    STATE(1910), 1,
      sym_attribute,
    STATE(7415), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(191), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1267), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1298), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1295), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [17376] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2950), 1,
      sym__lowercase_identifier,
    ACTIONS(2952), 1,
      anon_sym_LPAREN,
    ACTIONS(2956), 1,
      anon_sym_LBRACE,
    ACTIONS(2958), 1,
      anon_sym_begin,
    ACTIONS(2960), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2962), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2966), 1,
      aux_sym_number_token1,
    ACTIONS(2968), 1,
      anon_sym_SQUOTE,
    ACTIONS(2970), 1,
      anon_sym_DQUOTE,
    ACTIONS(2972), 1,
      sym_prefix_operator,
    STATE(1911), 1,
      sym_attribute,
    STATE(2304), 1,
      sym__simple_expression,
    STATE(2366), 1,
      sym_parenthesized_operator,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(7268), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2954), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2279), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [17496] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3064), 1,
      sym__lowercase_identifier,
    ACTIONS(3066), 1,
      anon_sym_LPAREN,
    ACTIONS(3068), 1,
      anon_sym_LBRACK,
    ACTIONS(3072), 1,
      anon_sym_LBRACE,
    ACTIONS(3074), 1,
      anon_sym_object,
    ACTIONS(3076), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3078), 1,
      anon_sym_new,
    ACTIONS(3080), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3082), 1,
      anon_sym_begin,
    ACTIONS(3084), 1,
      sym_ocamlyacc_value,
    ACTIONS(3086), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3088), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3090), 1,
      aux_sym_number_token1,
    ACTIONS(3092), 1,
      anon_sym_SQUOTE,
    ACTIONS(3094), 1,
      anon_sym_DQUOTE,
    ACTIONS(3096), 1,
      sym_prefix_operator,
    ACTIONS(3098), 1,
      sym__uppercase_identifier,
    ACTIONS(3100), 1,
      aux_sym_tag_token1,
    STATE(1912), 1,
      sym_attribute,
    STATE(2122), 1,
      sym__constructor_name,
    STATE(2451), 1,
      sym_parenthesized_operator,
    STATE(2553), 1,
      sym__value_name,
    STATE(2562), 1,
      sym__simple_expression,
    STATE(7715), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3070), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2617), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2552), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2551), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [17616] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2950), 1,
      sym__lowercase_identifier,
    ACTIONS(2952), 1,
      anon_sym_LPAREN,
    ACTIONS(2956), 1,
      anon_sym_LBRACE,
    ACTIONS(2958), 1,
      anon_sym_begin,
    ACTIONS(2960), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2962), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2966), 1,
      aux_sym_number_token1,
    ACTIONS(2968), 1,
      anon_sym_SQUOTE,
    ACTIONS(2970), 1,
      anon_sym_DQUOTE,
    ACTIONS(2972), 1,
      sym_prefix_operator,
    STATE(1913), 1,
      sym_attribute,
    STATE(2338), 1,
      sym__simple_expression,
    STATE(2366), 1,
      sym_parenthesized_operator,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(7268), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2954), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2279), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [17736] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2986), 1,
      sym__lowercase_identifier,
    ACTIONS(2988), 1,
      anon_sym_LPAREN,
    ACTIONS(2990), 1,
      anon_sym_LBRACK,
    ACTIONS(2994), 1,
      anon_sym_LBRACE,
    ACTIONS(2996), 1,
      anon_sym_object,
    ACTIONS(2998), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3000), 1,
      anon_sym_new,
    ACTIONS(3002), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3004), 1,
      anon_sym_begin,
    ACTIONS(3006), 1,
      sym_ocamlyacc_value,
    ACTIONS(3008), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3010), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3012), 1,
      aux_sym_number_token1,
    ACTIONS(3014), 1,
      anon_sym_SQUOTE,
    ACTIONS(3016), 1,
      anon_sym_DQUOTE,
    ACTIONS(3018), 1,
      sym_prefix_operator,
    ACTIONS(3020), 1,
      sym__uppercase_identifier,
    ACTIONS(3022), 1,
      aux_sym_tag_token1,
    STATE(1914), 1,
      sym_attribute,
    STATE(2486), 1,
      sym__value_name,
    STATE(2508), 1,
      sym_parenthesized_operator,
    STATE(2517), 1,
      sym__constructor_name,
    STATE(2570), 1,
      sym__simple_expression,
    STATE(7598), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2992), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2516), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2450), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2449), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [17856] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2556), 1,
      sym__lowercase_identifier,
    ACTIONS(2918), 1,
      anon_sym_LPAREN,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2922), 1,
      anon_sym_LBRACE,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2936), 1,
      sym_prefix_operator,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(1915), 1,
      sym_attribute,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5938), 1,
      sym__simple_expression,
    STATE(7330), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [17976] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2950), 1,
      sym__lowercase_identifier,
    ACTIONS(2952), 1,
      anon_sym_LPAREN,
    ACTIONS(2956), 1,
      anon_sym_LBRACE,
    ACTIONS(2958), 1,
      anon_sym_begin,
    ACTIONS(2960), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2962), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2966), 1,
      aux_sym_number_token1,
    ACTIONS(2968), 1,
      anon_sym_SQUOTE,
    ACTIONS(2970), 1,
      anon_sym_DQUOTE,
    ACTIONS(2972), 1,
      sym_prefix_operator,
    STATE(1916), 1,
      sym_attribute,
    STATE(2303), 1,
      sym__simple_expression,
    STATE(2366), 1,
      sym_parenthesized_operator,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(7268), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2954), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2279), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [18096] = 30,
    ACTIONS(253), 1,
      anon_sym_LPAREN,
    ACTIONS(255), 1,
      anon_sym_LBRACK,
    ACTIONS(259), 1,
      anon_sym_LBRACE,
    ACTIONS(261), 1,
      anon_sym_object,
    ACTIONS(265), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(275), 1,
      anon_sym_new,
    ACTIONS(277), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(279), 1,
      anon_sym_begin,
    ACTIONS(281), 1,
      sym_ocamlyacc_value,
    ACTIONS(287), 1,
      aux_sym_number_token1,
    ACTIONS(289), 1,
      anon_sym_SQUOTE,
    ACTIONS(291), 1,
      anon_sym_DQUOTE,
    ACTIONS(293), 1,
      sym_prefix_operator,
    ACTIONS(297), 1,
      sym__uppercase_identifier,
    ACTIONS(299), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1083), 1,
      sym__lowercase_identifier,
    ACTIONS(1497), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1499), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1330), 1,
      sym__simple_expression,
    STATE(1366), 1,
      sym__value_name,
    STATE(1375), 1,
      sym__constructor_name,
    STATE(1378), 1,
      sym_parenthesized_operator,
    STATE(1917), 1,
      sym_attribute,
    STATE(7186), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(257), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1336), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1377), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1364), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [18216] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2986), 1,
      sym__lowercase_identifier,
    ACTIONS(2988), 1,
      anon_sym_LPAREN,
    ACTIONS(2990), 1,
      anon_sym_LBRACK,
    ACTIONS(2994), 1,
      anon_sym_LBRACE,
    ACTIONS(2996), 1,
      anon_sym_object,
    ACTIONS(2998), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3000), 1,
      anon_sym_new,
    ACTIONS(3002), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3004), 1,
      anon_sym_begin,
    ACTIONS(3006), 1,
      sym_ocamlyacc_value,
    ACTIONS(3008), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3010), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3012), 1,
      aux_sym_number_token1,
    ACTIONS(3014), 1,
      anon_sym_SQUOTE,
    ACTIONS(3016), 1,
      anon_sym_DQUOTE,
    ACTIONS(3018), 1,
      sym_prefix_operator,
    ACTIONS(3020), 1,
      sym__uppercase_identifier,
    ACTIONS(3022), 1,
      aux_sym_tag_token1,
    STATE(1918), 1,
      sym_attribute,
    STATE(2486), 1,
      sym__value_name,
    STATE(2508), 1,
      sym_parenthesized_operator,
    STATE(2517), 1,
      sym__constructor_name,
    STATE(2571), 1,
      sym__simple_expression,
    STATE(7598), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2992), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2516), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2450), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2449), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [18336] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2924), 1,
      anon_sym_object,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2928), 1,
      anon_sym_new,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2934), 1,
      sym_ocamlyacc_value,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(2950), 1,
      sym__lowercase_identifier,
    ACTIONS(2952), 1,
      anon_sym_LPAREN,
    ACTIONS(2956), 1,
      anon_sym_LBRACE,
    ACTIONS(2958), 1,
      anon_sym_begin,
    ACTIONS(2960), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2962), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2966), 1,
      aux_sym_number_token1,
    ACTIONS(2968), 1,
      anon_sym_SQUOTE,
    ACTIONS(2970), 1,
      anon_sym_DQUOTE,
    ACTIONS(2972), 1,
      sym_prefix_operator,
    STATE(1919), 1,
      sym_attribute,
    STATE(2366), 1,
      sym_parenthesized_operator,
    STATE(2397), 1,
      sym__simple_expression,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2439), 1,
      sym__value_name,
    STATE(7268), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2954), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2279), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2367), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [18456] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3064), 1,
      sym__lowercase_identifier,
    ACTIONS(3066), 1,
      anon_sym_LPAREN,
    ACTIONS(3068), 1,
      anon_sym_LBRACK,
    ACTIONS(3072), 1,
      anon_sym_LBRACE,
    ACTIONS(3074), 1,
      anon_sym_object,
    ACTIONS(3076), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3078), 1,
      anon_sym_new,
    ACTIONS(3080), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3082), 1,
      anon_sym_begin,
    ACTIONS(3084), 1,
      sym_ocamlyacc_value,
    ACTIONS(3086), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3088), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3090), 1,
      aux_sym_number_token1,
    ACTIONS(3092), 1,
      anon_sym_SQUOTE,
    ACTIONS(3094), 1,
      anon_sym_DQUOTE,
    ACTIONS(3096), 1,
      sym_prefix_operator,
    ACTIONS(3098), 1,
      sym__uppercase_identifier,
    ACTIONS(3100), 1,
      aux_sym_tag_token1,
    STATE(1920), 1,
      sym_attribute,
    STATE(2122), 1,
      sym__constructor_name,
    STATE(2451), 1,
      sym_parenthesized_operator,
    STATE(2553), 1,
      sym__value_name,
    STATE(2576), 1,
      sym__simple_expression,
    STATE(7715), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3070), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2617), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2552), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2551), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [18576] = 30,
    ACTIONS(19), 1,
      anon_sym_LPAREN,
    ACTIONS(25), 1,
      anon_sym_LBRACK,
    ACTIONS(29), 1,
      anon_sym_LBRACE,
    ACTIONS(43), 1,
      anon_sym_object,
    ACTIONS(47), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(65), 1,
      anon_sym_new,
    ACTIONS(67), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(69), 1,
      anon_sym_begin,
    ACTIONS(71), 1,
      sym_ocamlyacc_value,
    ACTIONS(83), 1,
      aux_sym_number_token1,
    ACTIONS(87), 1,
      anon_sym_SQUOTE,
    ACTIONS(89), 1,
      anon_sym_DQUOTE,
    ACTIONS(91), 1,
      sym_prefix_operator,
    ACTIONS(99), 1,
      sym__uppercase_identifier,
    ACTIONS(103), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1153), 1,
      sym__lowercase_identifier,
    ACTIONS(1487), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1489), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1415), 1,
      sym__simple_expression,
    STATE(1442), 1,
      sym__value_name,
    STATE(1483), 1,
      sym_parenthesized_operator,
    STATE(1484), 1,
      sym__constructor_name,
    STATE(1921), 1,
      sym_attribute,
    STATE(7286), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(27), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1441), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1418), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1471), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [18696] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(807), 1,
      anon_sym_LPAREN,
    ACTIONS(809), 1,
      anon_sym_LBRACK,
    ACTIONS(813), 1,
      anon_sym_LBRACE,
    ACTIONS(815), 1,
      anon_sym_object,
    ACTIONS(819), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(837), 1,
      anon_sym_new,
    ACTIONS(839), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(841), 1,
      anon_sym_begin,
    ACTIONS(843), 1,
      sym_ocamlyacc_value,
    ACTIONS(845), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(847), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(849), 1,
      aux_sym_number_token1,
    ACTIONS(851), 1,
      anon_sym_SQUOTE,
    ACTIONS(853), 1,
      anon_sym_DQUOTE,
    ACTIONS(855), 1,
      sym_prefix_operator,
    ACTIONS(859), 1,
      sym__uppercase_identifier,
    ACTIONS(861), 1,
      aux_sym_tag_token1,
    ACTIONS(1687), 1,
      sym__lowercase_identifier,
    STATE(1922), 1,
      sym_attribute,
    STATE(1969), 1,
      sym__constructor_name,
    STATE(2018), 1,
      sym__value_name,
    STATE(2019), 1,
      sym_parenthesized_operator,
    STATE(2021), 1,
      sym__simple_expression,
    STATE(7454), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(811), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1968), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2002), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1977), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [18816] = 30,
    ACTIONS(407), 1,
      anon_sym_LPAREN,
    ACTIONS(409), 1,
      anon_sym_LBRACK,
    ACTIONS(413), 1,
      anon_sym_LBRACE,
    ACTIONS(415), 1,
      anon_sym_object,
    ACTIONS(419), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(437), 1,
      anon_sym_new,
    ACTIONS(439), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(441), 1,
      anon_sym_begin,
    ACTIONS(443), 1,
      sym_ocamlyacc_value,
    ACTIONS(449), 1,
      aux_sym_number_token1,
    ACTIONS(451), 1,
      anon_sym_SQUOTE,
    ACTIONS(453), 1,
      anon_sym_DQUOTE,
    ACTIONS(455), 1,
      sym_prefix_operator,
    ACTIONS(459), 1,
      sym__uppercase_identifier,
    ACTIONS(461), 1,
      aux_sym_tag_token1,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(479), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(481), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(989), 1,
      sym__lowercase_identifier,
    STATE(1044), 1,
      sym__simple_expression,
    STATE(1073), 1,
      sym__value_name,
    STATE(1074), 1,
      sym_parenthesized_operator,
    STATE(1107), 1,
      sym__constructor_name,
    STATE(1923), 1,
      sym_attribute,
    STATE(7491), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(411), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1106), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1046), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1045), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [18936] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3026), 1,
      sym__lowercase_identifier,
    ACTIONS(3028), 1,
      anon_sym_LPAREN,
    ACTIONS(3030), 1,
      anon_sym_LBRACK,
    ACTIONS(3034), 1,
      anon_sym_LBRACE,
    ACTIONS(3036), 1,
      anon_sym_object,
    ACTIONS(3038), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3040), 1,
      anon_sym_new,
    ACTIONS(3042), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3044), 1,
      anon_sym_begin,
    ACTIONS(3046), 1,
      sym_ocamlyacc_value,
    ACTIONS(3048), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3050), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3052), 1,
      aux_sym_number_token1,
    ACTIONS(3054), 1,
      anon_sym_SQUOTE,
    ACTIONS(3056), 1,
      anon_sym_DQUOTE,
    ACTIONS(3058), 1,
      sym_prefix_operator,
    ACTIONS(3060), 1,
      sym__uppercase_identifier,
    ACTIONS(3062), 1,
      aux_sym_tag_token1,
    STATE(1924), 1,
      sym_attribute,
    STATE(3286), 1,
      sym__simple_expression,
    STATE(3296), 1,
      sym_parenthesized_operator,
    STATE(3335), 1,
      sym__constructor_name,
    STATE(3360), 1,
      sym__value_name,
    STATE(7527), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3032), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3352), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3310), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(3332), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [19056] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3026), 1,
      sym__lowercase_identifier,
    ACTIONS(3028), 1,
      anon_sym_LPAREN,
    ACTIONS(3030), 1,
      anon_sym_LBRACK,
    ACTIONS(3034), 1,
      anon_sym_LBRACE,
    ACTIONS(3036), 1,
      anon_sym_object,
    ACTIONS(3038), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3040), 1,
      anon_sym_new,
    ACTIONS(3042), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3044), 1,
      anon_sym_begin,
    ACTIONS(3046), 1,
      sym_ocamlyacc_value,
    ACTIONS(3048), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3050), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3052), 1,
      aux_sym_number_token1,
    ACTIONS(3054), 1,
      anon_sym_SQUOTE,
    ACTIONS(3056), 1,
      anon_sym_DQUOTE,
    ACTIONS(3058), 1,
      sym_prefix_operator,
    ACTIONS(3060), 1,
      sym__uppercase_identifier,
    ACTIONS(3062), 1,
      aux_sym_tag_token1,
    STATE(1925), 1,
      sym_attribute,
    STATE(3296), 1,
      sym_parenthesized_operator,
    STATE(3335), 1,
      sym__constructor_name,
    STATE(3346), 1,
      sym__simple_expression,
    STATE(3360), 1,
      sym__value_name,
    STATE(7527), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3032), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3352), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3310), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(3332), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [19176] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2211), 1,
      sym__lowercase_identifier,
    ACTIONS(2215), 1,
      anon_sym_LPAREN,
    ACTIONS(2217), 1,
      anon_sym_LBRACK,
    ACTIONS(2221), 1,
      anon_sym_LBRACE,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(2225), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2227), 1,
      anon_sym_new,
    ACTIONS(2229), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2231), 1,
      anon_sym_begin,
    ACTIONS(2233), 1,
      sym_ocamlyacc_value,
    ACTIONS(2239), 1,
      aux_sym_number_token1,
    ACTIONS(2241), 1,
      anon_sym_SQUOTE,
    ACTIONS(2243), 1,
      anon_sym_DQUOTE,
    ACTIONS(2245), 1,
      sym_prefix_operator,
    ACTIONS(2247), 1,
      sym__uppercase_identifier,
    ACTIONS(2249), 1,
      aux_sym_tag_token1,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1926), 1,
      sym_attribute,
    STATE(2973), 1,
      sym__constructor_name,
    STATE(3009), 1,
      sym__simple_expression,
    STATE(3037), 1,
      sym__value_name,
    STATE(3038), 1,
      sym_parenthesized_operator,
    STATE(7634), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2219), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2971), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3049), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(3044), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [19296] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2986), 1,
      sym__lowercase_identifier,
    ACTIONS(2988), 1,
      anon_sym_LPAREN,
    ACTIONS(2990), 1,
      anon_sym_LBRACK,
    ACTIONS(2994), 1,
      anon_sym_LBRACE,
    ACTIONS(2996), 1,
      anon_sym_object,
    ACTIONS(2998), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3000), 1,
      anon_sym_new,
    ACTIONS(3002), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3004), 1,
      anon_sym_begin,
    ACTIONS(3006), 1,
      sym_ocamlyacc_value,
    ACTIONS(3008), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3010), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3012), 1,
      aux_sym_number_token1,
    ACTIONS(3014), 1,
      anon_sym_SQUOTE,
    ACTIONS(3016), 1,
      anon_sym_DQUOTE,
    ACTIONS(3018), 1,
      sym_prefix_operator,
    ACTIONS(3020), 1,
      sym__uppercase_identifier,
    ACTIONS(3022), 1,
      aux_sym_tag_token1,
    STATE(1927), 1,
      sym_attribute,
    STATE(2456), 1,
      sym__simple_expression,
    STATE(2486), 1,
      sym__value_name,
    STATE(2508), 1,
      sym_parenthesized_operator,
    STATE(2517), 1,
      sym__constructor_name,
    STATE(7598), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2992), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2516), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2450), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2449), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [19416] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2986), 1,
      sym__lowercase_identifier,
    ACTIONS(2988), 1,
      anon_sym_LPAREN,
    ACTIONS(2990), 1,
      anon_sym_LBRACK,
    ACTIONS(2994), 1,
      anon_sym_LBRACE,
    ACTIONS(2996), 1,
      anon_sym_object,
    ACTIONS(2998), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3000), 1,
      anon_sym_new,
    ACTIONS(3002), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3004), 1,
      anon_sym_begin,
    ACTIONS(3006), 1,
      sym_ocamlyacc_value,
    ACTIONS(3008), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3010), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3012), 1,
      aux_sym_number_token1,
    ACTIONS(3014), 1,
      anon_sym_SQUOTE,
    ACTIONS(3016), 1,
      anon_sym_DQUOTE,
    ACTIONS(3018), 1,
      sym_prefix_operator,
    ACTIONS(3020), 1,
      sym__uppercase_identifier,
    ACTIONS(3022), 1,
      aux_sym_tag_token1,
    STATE(1928), 1,
      sym_attribute,
    STATE(2465), 1,
      sym__simple_expression,
    STATE(2486), 1,
      sym__value_name,
    STATE(2508), 1,
      sym_parenthesized_operator,
    STATE(2517), 1,
      sym__constructor_name,
    STATE(7598), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2992), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2516), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2450), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2449), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [19536] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3170), 1,
      aux_sym_attribute_id_token1,
    STATE(1929), 1,
      sym_attribute,
    STATE(1931), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3168), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3166), 32,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [19609] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1833), 1,
      anon_sym_COLON2,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3172), 1,
      anon_sym_EQ,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(1930), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(6360), 1,
      sym__maybe_polymorphic_typed,
    STATE(7871), 1,
      sym__coerced,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [19736] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3170), 1,
      aux_sym_attribute_id_token1,
    STATE(1931), 1,
      sym_attribute,
    STATE(1939), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3176), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3174), 32,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [19809] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2317), 1,
      anon_sym_LT_DASH,
    STATE(1932), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2159), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2157), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19880] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1833), 1,
      anon_sym_COLON2,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3178), 1,
      anon_sym_EQ,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(1933), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(6469), 1,
      sym__maybe_polymorphic_typed,
    STATE(7900), 1,
      sym__coerced,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [20007] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1833), 1,
      anon_sym_COLON2,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3180), 1,
      anon_sym_EQ,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(1934), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(6501), 1,
      sym__maybe_polymorphic_typed,
    STATE(7233), 1,
      sym__coerced,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [20134] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1833), 1,
      anon_sym_COLON2,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3182), 1,
      anon_sym_EQ,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(1935), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(6355), 1,
      sym__maybe_polymorphic_typed,
    STATE(7800), 1,
      sym__coerced,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [20261] = 34,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1833), 1,
      anon_sym_COLON2,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3184), 1,
      anon_sym_EQ,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(1936), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(6515), 1,
      sym__maybe_polymorphic_typed,
    STATE(7300), 1,
      sym__coerced,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [20388] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1937), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1901), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1899), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20457] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1938), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1905), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1903), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20526] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3190), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1939), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
    ACTIONS(3188), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3186), 32,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [20597] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1909), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1907), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20666] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1941), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1913), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1911), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20735] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1917), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1915), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20804] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1943), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1921), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1919), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20873] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3193), 1,
      anon_sym_LT_DASH,
    STATE(1944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1949), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1947), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20944] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1945), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1893), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1891), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21013] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1946), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1897), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1895), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21082] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1889), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1887), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21151] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1948), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2155), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2153), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21219] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1949), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2045), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2043), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21287] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1950), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2073), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2071), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21355] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1981), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1979), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21423] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1952), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2081), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2079), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21491] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3195), 1,
      anon_sym_LT_DASH,
    STATE(1953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1949), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1947), 32,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21561] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1954), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2143), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2141), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21629] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1955), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1977), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1975), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21697] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2001), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1999), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21765] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1957), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2077), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2075), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21833] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1929), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1927), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21901] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1959), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2005), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2003), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21969] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1691), 1,
      anon_sym_DOT,
    ACTIONS(1693), 1,
      anon_sym_POUND,
    ACTIONS(1695), 1,
      sym_hash_operator,
    STATE(1960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1933), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1931), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22043] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1961), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2201), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2199), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22111] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1962), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2009), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2007), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22179] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2013), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2011), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22247] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1964), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2205), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2203), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22315] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2105), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2103), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22383] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1966), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2109), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2107), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22451] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(1967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2111), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22521] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1968), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2119), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2117), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22589] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2127), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2125), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22657] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1970), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2151), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2149), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22725] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2017), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2015), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22793] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2021), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2019), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22861] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1973), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2135), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2133), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22929] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1974), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2025), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2023), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22997] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1975), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2029), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2027), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [23065] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1976), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2033), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2031), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [23133] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1949), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1947), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [23201] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1978), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1925), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1923), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [23269] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1979), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2147), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2145), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [23337] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1980), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2197), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2195), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [23405] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1981), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2171), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2169), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [23473] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1982), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2175), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2173), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [23541] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(1983), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2111), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [23611] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2181), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2179), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [23679] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1985), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2049), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2047), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [23747] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1986), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2131), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2129), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [23815] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1987), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3188), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
      aux_sym_attribute_id_token1,
    ACTIONS(3186), 32,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [23883] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1988), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2061), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2059), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [23951] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2065), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2063), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24019] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1990), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2069), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2067), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24087] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1991), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1937), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1935), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24155] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1992), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1941), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1939), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24223] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1993), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2089), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2087), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24291] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1994), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2139), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2137), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24359] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1995), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2093), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2091), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24427] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1996), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2097), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2095), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24495] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2053), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2051), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24563] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1998), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2057), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2055), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24631] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1953), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1951), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24699] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2000), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2101), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2099), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24767] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1945), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1943), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24835] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2002), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1997), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1995), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24903] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1957), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1955), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24971] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2004), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2167), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2165), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25039] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1961), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1959), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25107] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2006), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1985), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1983), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25175] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2185), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2183), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25243] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2008), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2189), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2187), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25311] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2193), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2191), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25379] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1989), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1987), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25447] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1965), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1963), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25515] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2012), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2085), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2083), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25583] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2013), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1993), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1991), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25651] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2163), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2161), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25719] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2015), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1969), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1967), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25787] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3197), 1,
      anon_sym_LT_DASH,
    STATE(2016), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1949), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1947), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25857] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1691), 1,
      anon_sym_DOT,
    STATE(2017), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1973), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1971), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25927] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2209), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2207), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25995] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2019), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2123), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2121), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [26063] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2020), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2159), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2157), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [26131] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1691), 1,
      anon_sym_DOT,
    ACTIONS(1693), 1,
      anon_sym_POUND,
    ACTIONS(1695), 1,
      sym_hash_operator,
    STATE(2021), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2041), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2039), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [26205] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2022), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2037), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2035), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [26273] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3201), 1,
      anon_sym_DASH_GT,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2023), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7960), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26394] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3205), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2024), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    STATE(8340), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26515] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3207), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2025), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7431), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26636] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3209), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2026), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7539), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26757] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3211), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2027), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7348), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26878] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3213), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2028), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7477), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26999] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3215), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2029), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8021), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27120] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3217), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2030), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8120), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27241] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3219), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2031), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7998), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27362] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3221), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2032), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7581), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27483] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3223), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2033), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7727), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27604] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3225), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2034), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8074), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27725] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3227), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2035), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8134), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27846] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3233), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2036), 2,
      sym_attribute,
      aux_sym__structure_repeat1,
    ACTIONS(3229), 19,
      ts_builtin_sym_end,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3231), 31,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [27915] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3236), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2037), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7956), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [28036] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3238), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2038), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7905), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [28157] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3240), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2039), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7954), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [28278] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3242), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2040), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8008), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [28399] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3244), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2041), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8221), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [28520] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3246), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2042), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8195), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [28641] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3248), 1,
      anon_sym_LT_DASH,
    STATE(2043), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1949), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1947), 31,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [28710] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3250), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2044), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8262), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [28831] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3252), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2045), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7946), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [28952] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3254), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2046), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7877), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [29073] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3256), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2047), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7915), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [29194] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3258), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2048), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7769), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [29315] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3260), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2049), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7669), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [29436] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3262), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2050), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7831), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [29557] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3264), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2051), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7370), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [29678] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3266), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2052), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7545), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [29799] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3268), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2053), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7320), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [29920] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3270), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2054), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8029), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [30041] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3272), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2055), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(7478), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [30162] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3199), 1,
      anon_sym_COLON2,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3274), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2056), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8093), 1,
      sym__simple_typed,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [30283] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2056), 1,
      aux_sym_let_binding_repeat1,
    STATE(2057), 1,
      sym_attribute,
    STATE(2176), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [30401] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2031), 1,
      aux_sym_let_binding_repeat1,
    STATE(2058), 1,
      sym_attribute,
    STATE(2152), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [30519] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2033), 1,
      aux_sym_let_binding_repeat1,
    STATE(2059), 1,
      sym_attribute,
    STATE(2118), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [30637] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(2060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2111), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [30705] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2038), 1,
      aux_sym_let_binding_repeat1,
    STATE(2061), 1,
      sym_attribute,
    STATE(2185), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [30823] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2042), 1,
      aux_sym_let_binding_repeat1,
    STATE(2062), 1,
      sym_attribute,
    STATE(2108), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [30941] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2046), 1,
      aux_sym_let_binding_repeat1,
    STATE(2063), 1,
      sym_attribute,
    STATE(2167), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [31059] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2026), 1,
      aux_sym_let_binding_repeat1,
    STATE(2064), 1,
      sym_attribute,
    STATE(2166), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [31177] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3276), 1,
      anon_sym_COLON,
    STATE(2065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2315), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2311), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [31245] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2049), 1,
      aux_sym_let_binding_repeat1,
    STATE(2066), 1,
      sym_attribute,
    STATE(2171), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [31363] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2024), 1,
      aux_sym_let_binding_repeat1,
    STATE(2067), 1,
      sym_attribute,
    STATE(2119), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [31481] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2054), 1,
      aux_sym_let_binding_repeat1,
    STATE(2068), 1,
      sym_attribute,
    STATE(2117), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [31599] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2047), 1,
      aux_sym_let_binding_repeat1,
    STATE(2069), 1,
      sym_attribute,
    STATE(2142), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [31717] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2050), 1,
      aux_sym_let_binding_repeat1,
    STATE(2070), 1,
      sym_attribute,
    STATE(2175), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [31835] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2051), 1,
      aux_sym_let_binding_repeat1,
    STATE(2071), 1,
      sym_attribute,
    STATE(2151), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [31953] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(2072), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2111), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [32021] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2055), 1,
      aux_sym_let_binding_repeat1,
    STATE(2073), 1,
      sym_attribute,
    STATE(2135), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [32139] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3278), 1,
      anon_sym_LT_DASH,
    STATE(2074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1949), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1947), 29,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [32207] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2037), 1,
      aux_sym_let_binding_repeat1,
    STATE(2075), 1,
      sym_attribute,
    STATE(2144), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [32325] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2076), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3280), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3282), 31,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [32391] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1799), 1,
      anon_sym_PERCENT,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2035), 1,
      aux_sym_let_binding_repeat1,
    STATE(2077), 1,
      sym_attribute,
    STATE(2179), 1,
      sym__attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [32509] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2078), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2417), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2415), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [32574] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    ACTIONS(3302), 1,
      sym_rel_operator,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2079), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3284), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [32669] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3308), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2080), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [32784] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2081), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1933), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1931), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [32849] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    ACTIONS(3302), 1,
      sym_rel_operator,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 7,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [32940] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3310), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2083), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [33055] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    ACTIONS(3302), 1,
      sym_rel_operator,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3312), 1,
      anon_sym_COMMA,
    ACTIONS(3314), 1,
      anon_sym_SEMI,
    ACTIONS(3316), 1,
      sym_assign_operator,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2084), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3156), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3154), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [33156] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    ACTIONS(3302), 1,
      sym_rel_operator,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3312), 1,
      anon_sym_COMMA,
    ACTIONS(3314), 1,
      anon_sym_SEMI,
    ACTIONS(3316), 1,
      sym_assign_operator,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2085), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3160), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3158), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [33257] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3318), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2086), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [33372] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    ACTIONS(3302), 1,
      sym_rel_operator,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3312), 1,
      anon_sym_COMMA,
    ACTIONS(3316), 1,
      sym_assign_operator,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2087), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3322), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3320), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [33471] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2088), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2127), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2125), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [33536] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3324), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2089), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [33651] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2090), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [33726] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    ACTIONS(3302), 1,
      sym_rel_operator,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3312), 1,
      anon_sym_COMMA,
    ACTIONS(3316), 1,
      sym_assign_operator,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2091), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3332), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3330), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [33825] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2092), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 27,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [33902] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2093), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 27,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [33979] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2094), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34060] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3203), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3334), 1,
      anon_sym_DASH_GT,
    STATE(1506), 1,
      aux_sym_let_binding_repeat1,
    STATE(2095), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [34175] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    ACTIONS(3302), 1,
      sym_rel_operator,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3312), 1,
      anon_sym_COMMA,
    ACTIONS(3314), 1,
      anon_sym_SEMI,
    ACTIONS(3316), 1,
      sym_assign_operator,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3338), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3336), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34276] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    ACTIONS(3302), 1,
      sym_rel_operator,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3312), 1,
      anon_sym_COMMA,
    ACTIONS(3316), 1,
      sym_assign_operator,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2097), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3342), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3340), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34375] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    ACTIONS(3302), 1,
      sym_rel_operator,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3312), 1,
      anon_sym_COMMA,
    ACTIONS(3316), 1,
      sym_assign_operator,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2098), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3284), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34474] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    ACTIONS(3302), 1,
      sym_rel_operator,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3312), 1,
      anon_sym_COMMA,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2099), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3346), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3344), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34571] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2181), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2179), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [34636] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2101), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3350), 9,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3348), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34721] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3352), 1,
      anon_sym_LT_DASH,
    STATE(2102), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1949), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1947), 29,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_do,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [34788] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 9,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34875] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2104), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 9,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34962] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2105), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2139), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2137), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [35027] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2106), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2147), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2145), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [35092] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2107), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2137), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2139), 36,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
      aux_sym_tag_token1,
  [35156] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2044), 1,
      aux_sym_let_binding_repeat1,
    STATE(2108), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [35268] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2083), 1,
      aux_sym_let_binding_repeat1,
    STATE(2109), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [35380] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1357), 1,
      sym__lowercase_identifier,
    ACTIONS(1361), 1,
      anon_sym_LPAREN,
    ACTIONS(1369), 1,
      anon_sym_LBRACK,
    ACTIONS(1373), 1,
      anon_sym_LBRACE,
    ACTIONS(1383), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3354), 1,
      anon_sym_lazy,
    STATE(1625), 1,
      sym_constructor_path,
    STATE(1639), 1,
      sym_tag,
    STATE(2110), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(7520), 1,
      sym_module_path,
    STATE(7762), 1,
      sym__effect_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(5091), 4,
      sym__simple_pattern,
      sym_constructor_pattern,
      sym_tag_pattern,
      sym_lazy_pattern,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5133), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(5132), 10,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [35494] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3356), 1,
      anon_sym_COMMA,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    ACTIONS(3362), 1,
      sym_rel_operator,
    ACTIONS(3364), 1,
      sym_assign_operator,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3322), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3320), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [35592] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3356), 1,
      anon_sym_COMMA,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    ACTIONS(3362), 1,
      sym_rel_operator,
    ACTIONS(3364), 1,
      sym_assign_operator,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2112), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3332), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3330), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [35690] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1419), 1,
      anon_sym_LPAREN,
    ACTIONS(1425), 1,
      anon_sym_LBRACK,
    ACTIONS(1427), 1,
      anon_sym_LBRACE,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3064), 1,
      sym__lowercase_identifier,
    ACTIONS(3366), 1,
      anon_sym_lazy,
    STATE(1633), 1,
      sym_constructor_path,
    STATE(1674), 1,
      sym_tag,
    STATE(2113), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(7412), 1,
      sym_module_path,
    STATE(7835), 1,
      sym__effect_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(5311), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5312), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(5310), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [35806] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3368), 1,
      anon_sym_DOT_DOT,
    STATE(2114), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1641), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1645), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [35872] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2086), 1,
      aux_sym_let_binding_repeat1,
    STATE(2115), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [35984] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3356), 1,
      anon_sym_COMMA,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    ACTIONS(3362), 1,
      sym_rel_operator,
    ACTIONS(3364), 1,
      sym_assign_operator,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3342), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3340), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [36082] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2045), 1,
      aux_sym_let_binding_repeat1,
    STATE(2117), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [36194] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2048), 1,
      aux_sym_let_binding_repeat1,
    STATE(2118), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [36306] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2053), 1,
      aux_sym_let_binding_repeat1,
    STATE(2119), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [36418] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3370), 1,
      anon_sym_COMMA,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    ACTIONS(3376), 1,
      sym_rel_operator,
    ACTIONS(3378), 1,
      sym_assign_operator,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2120), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3284), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [36516] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3370), 1,
      anon_sym_COMMA,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    ACTIONS(3376), 1,
      sym_rel_operator,
    ACTIONS(3378), 1,
      sym_assign_operator,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2121), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3342), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3340), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [36614] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2122), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2125), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2127), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36678] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2123), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2117), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2119), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [36742] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3370), 1,
      anon_sym_COMMA,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    ACTIONS(3376), 1,
      sym_rel_operator,
    ACTIONS(3378), 1,
      sym_assign_operator,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2124), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3322), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3320), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [36840] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3370), 1,
      anon_sym_COMMA,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    ACTIONS(3376), 1,
      sym_rel_operator,
    ACTIONS(3378), 1,
      sym_assign_operator,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2125), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3332), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3330), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [36938] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(2126), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2113), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [37004] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2127), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2153), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2155), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [37068] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3370), 1,
      anon_sym_COMMA,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    ACTIONS(3376), 1,
      sym_rel_operator,
    ACTIONS(3378), 1,
      sym_assign_operator,
    ACTIONS(3380), 1,
      anon_sym_SEMI,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2128), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3156), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3154), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [37168] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2129), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [37244] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2130), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [37320] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2131), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [37400] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2132), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3350), 9,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3348), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [37484] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3386), 1,
      anon_sym_COLON,
    STATE(2133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3382), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3384), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [37550] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2134), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3390), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(3388), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [37614] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2032), 1,
      aux_sym_let_binding_repeat1,
    STATE(2135), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [37726] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    ACTIONS(3376), 1,
      sym_rel_operator,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2136), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3284), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [37820] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2137), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3154), 35,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [37884] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1419), 1,
      anon_sym_LPAREN,
    ACTIONS(1425), 1,
      anon_sym_LBRACK,
    ACTIONS(1427), 1,
      anon_sym_LBRACE,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3064), 1,
      sym__lowercase_identifier,
    ACTIONS(3366), 1,
      anon_sym_lazy,
    STATE(1633), 1,
      sym_constructor_path,
    STATE(1674), 1,
      sym_tag,
    STATE(2138), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(7412), 1,
      sym_module_path,
    STATE(8122), 1,
      sym__effect_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(5311), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5312), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(5310), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [38000] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3356), 1,
      anon_sym_COMMA,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    ACTIONS(3362), 1,
      sym_rel_operator,
    ACTIONS(3364), 1,
      sym_assign_operator,
    ACTIONS(3392), 1,
      anon_sym_SEMI,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2139), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3338), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3336), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [38100] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3356), 1,
      anon_sym_COMMA,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    ACTIONS(3362), 1,
      sym_rel_operator,
    ACTIONS(3364), 1,
      sym_assign_operator,
    ACTIONS(3392), 1,
      anon_sym_SEMI,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3156), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3154), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [38200] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2141), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [38274] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2039), 1,
      aux_sym_let_binding_repeat1,
    STATE(2142), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [38386] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2143), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2107), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2109), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [38450] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2023), 1,
      aux_sym_let_binding_repeat1,
    STATE(2144), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [38562] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1357), 1,
      sym__lowercase_identifier,
    ACTIONS(1361), 1,
      anon_sym_LPAREN,
    ACTIONS(1369), 1,
      anon_sym_LBRACK,
    ACTIONS(1373), 1,
      anon_sym_LBRACE,
    ACTIONS(1383), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3354), 1,
      anon_sym_lazy,
    STATE(1625), 1,
      sym_constructor_path,
    STATE(1639), 1,
      sym_tag,
    STATE(2145), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(7411), 1,
      sym__effect_pattern,
    STATE(7520), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(5091), 4,
      sym__simple_pattern,
      sym_constructor_pattern,
      sym_tag_pattern,
      sym_lazy_pattern,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5133), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(5132), 10,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [38676] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2146), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 27,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [38752] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 27,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [38828] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1357), 1,
      sym__lowercase_identifier,
    ACTIONS(1361), 1,
      anon_sym_LPAREN,
    ACTIONS(1369), 1,
      anon_sym_LBRACK,
    ACTIONS(1373), 1,
      anon_sym_LBRACE,
    ACTIONS(1383), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3354), 1,
      anon_sym_lazy,
    STATE(1625), 1,
      sym_constructor_path,
    STATE(1639), 1,
      sym_tag,
    STATE(2148), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(7520), 1,
      sym_module_path,
    STATE(7848), 1,
      sym__effect_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(5091), 4,
      sym__simple_pattern,
      sym_constructor_pattern,
      sym_tag_pattern,
      sym_lazy_pattern,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5133), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(5132), 10,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [38942] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1419), 1,
      anon_sym_LPAREN,
    ACTIONS(1425), 1,
      anon_sym_LBRACK,
    ACTIONS(1427), 1,
      anon_sym_LBRACE,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3064), 1,
      sym__lowercase_identifier,
    ACTIONS(3366), 1,
      anon_sym_lazy,
    STATE(1633), 1,
      sym_constructor_path,
    STATE(1674), 1,
      sym_tag,
    STATE(2149), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(7412), 1,
      sym_module_path,
    STATE(7893), 1,
      sym__effect_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(5311), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5312), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(5310), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [39058] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2080), 1,
      aux_sym_let_binding_repeat1,
    STATE(2150), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [39170] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2025), 1,
      aux_sym_let_binding_repeat1,
    STATE(2151), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [39282] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2030), 1,
      aux_sym_let_binding_repeat1,
    STATE(2152), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [39394] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2153), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2169), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2171), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [39458] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2154), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3396), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(3394), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [39522] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1419), 1,
      anon_sym_LPAREN,
    ACTIONS(1425), 1,
      anon_sym_LBRACK,
    ACTIONS(1427), 1,
      anon_sym_LBRACE,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3064), 1,
      sym__lowercase_identifier,
    ACTIONS(3366), 1,
      anon_sym_lazy,
    STATE(1633), 1,
      sym_constructor_path,
    STATE(1674), 1,
      sym_tag,
    STATE(2155), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(7230), 1,
      sym__effect_binding_pattern,
    STATE(7412), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(5311), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5312), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(5310), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [39638] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2156), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 11,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [39718] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2089), 1,
      aux_sym_let_binding_repeat1,
    STATE(2157), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [39830] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3356), 1,
      anon_sym_COMMA,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    ACTIONS(3362), 1,
      sym_rel_operator,
    ACTIONS(3364), 1,
      sym_assign_operator,
    ACTIONS(3392), 1,
      anon_sym_SEMI,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2158), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3160), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3158), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [39930] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2159), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3152), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3150), 35,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [39994] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(2160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2113), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [40060] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2173), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2175), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [40124] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2095), 1,
      aux_sym_let_binding_repeat1,
    STATE(2162), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [40236] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2163), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 9,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [40322] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2164), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3158), 35,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [40386] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2165), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 9,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [40472] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2052), 1,
      aux_sym_let_binding_repeat1,
    STATE(2166), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [40584] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2034), 1,
      aux_sym_let_binding_repeat1,
    STATE(2167), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [40696] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3370), 1,
      anon_sym_COMMA,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    ACTIONS(3376), 1,
      sym_rel_operator,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2168), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3346), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3344), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [40792] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3370), 1,
      anon_sym_COMMA,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    ACTIONS(3376), 1,
      sym_rel_operator,
    ACTIONS(3378), 1,
      sym_assign_operator,
    ACTIONS(3380), 1,
      anon_sym_SEMI,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2169), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3160), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3158), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [40892] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2170), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3350), 8,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3348), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [40976] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2040), 1,
      aux_sym_let_binding_repeat1,
    STATE(2171), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [41088] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2172), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 8,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [41174] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2173), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2103), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2105), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [41238] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2174), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2179), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2181), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [41302] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2029), 1,
      aux_sym_let_binding_repeat1,
    STATE(2175), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [41414] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2027), 1,
      aux_sym_let_binding_repeat1,
    STATE(2176), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [41526] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2177), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 8,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [41612] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2178), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2133), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2135), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [41676] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2041), 1,
      aux_sym_let_binding_repeat1,
    STATE(2179), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [41788] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    ACTIONS(3376), 1,
      sym_rel_operator,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2180), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 7,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [41878] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3356), 1,
      anon_sym_COMMA,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    ACTIONS(3362), 1,
      sym_rel_operator,
    ACTIONS(3364), 1,
      sym_assign_operator,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2181), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3284), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [41976] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    ACTIONS(3362), 1,
      sym_rel_operator,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2182), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [42066] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1419), 1,
      anon_sym_LPAREN,
    ACTIONS(1425), 1,
      anon_sym_LBRACK,
    ACTIONS(1427), 1,
      anon_sym_LBRACE,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3064), 1,
      sym__lowercase_identifier,
    ACTIONS(3366), 1,
      anon_sym_lazy,
    STATE(1633), 1,
      sym_constructor_path,
    STATE(1674), 1,
      sym_tag,
    STATE(2183), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(7412), 1,
      sym_module_path,
    STATE(8344), 1,
      sym__effect_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(5311), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5312), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(5310), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [42182] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    ACTIONS(3362), 1,
      sym_rel_operator,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2184), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3284), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [42276] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1827), 1,
      anon_sym_LPAREN,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2028), 1,
      aux_sym_let_binding_repeat1,
    STATE(2185), 1,
      sym_attribute,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2868), 1,
      sym__simple_pattern,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2862), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [42388] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3370), 1,
      anon_sym_COMMA,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    ACTIONS(3376), 1,
      sym_rel_operator,
    ACTIONS(3378), 1,
      sym_assign_operator,
    ACTIONS(3380), 1,
      anon_sym_SEMI,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2186), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3338), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3336), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [42488] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2187), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [42562] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3356), 1,
      anon_sym_COMMA,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    ACTIONS(3362), 1,
      sym_rel_operator,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2188), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3346), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3344), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [42658] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2189), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2063), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2065), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42721] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2190), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2003), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2005), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [42784] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(2191), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2113), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42849] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2192), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1923), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1925), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [42912] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2193), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2133), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2135), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42975] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2194), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2059), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2061), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [43038] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2195), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2161), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2163), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43101] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2196), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2129), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2131), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [43164] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2197), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2183), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2185), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [43227] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2198), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2043), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2045), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43290] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3398), 1,
      anon_sym_COMMA,
    ACTIONS(3400), 1,
      anon_sym_SEMI,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    ACTIONS(3406), 1,
      sym_rel_operator,
    ACTIONS(3408), 1,
      sym_assign_operator,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2199), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3156), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3154), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [43389] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2200), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2047), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2049), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43452] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2201), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2071), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2073), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43515] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3414), 1,
      anon_sym_DOT,
    ACTIONS(3416), 1,
      anon_sym_POUND,
    ACTIONS(3418), 1,
      sym_hash_operator,
    STATE(2202), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3412), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(3410), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [43584] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3414), 1,
      anon_sym_DOT,
    ACTIONS(3416), 1,
      anon_sym_POUND,
    ACTIONS(3418), 1,
      sym_hash_operator,
    STATE(2203), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3422), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(3420), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [43653] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2204), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2075), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2077), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43716] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2205), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2079), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2081), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43779] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2206), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2063), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2065), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [43842] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2207), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2067), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2069), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [43905] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2208), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3424), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3426), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [43968] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3398), 1,
      anon_sym_COMMA,
    ACTIONS(3400), 1,
      anon_sym_SEMI,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    ACTIONS(3406), 1,
      sym_rel_operator,
    ACTIONS(3408), 1,
      sym_assign_operator,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2209), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3160), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3158), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [44067] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2210), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2141), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2143), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44130] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2211), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [44203] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2212), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2087), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2089), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44266] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2213), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1927), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1929), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44329] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3428), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3430), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [44392] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2215), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3432), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3434), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [44455] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2216), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2145), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2147), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44518] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2217), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [44593] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2218), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3436), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3438), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [44656] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2219), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2091), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2093), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44719] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2220), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2149), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2151), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [44782] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2221), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [44857] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    ACTIONS(3302), 1,
      sym_rel_operator,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3312), 1,
      anon_sym_COMMA,
    ACTIONS(3316), 1,
      sym_assign_operator,
    ACTIONS(3440), 1,
      anon_sym_SEMI,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2222), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3156), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3154), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [44956] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2223), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 11,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [45035] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2095), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2097), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [45098] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2099), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2101), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [45161] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2226), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1955), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1957), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [45224] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3398), 1,
      anon_sym_COMMA,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    ACTIONS(3406), 1,
      sym_rel_operator,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3346), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3344), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [45319] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2228), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2179), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2181), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45382] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2191), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2193), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [45445] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3350), 8,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3348), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [45528] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(2231), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2113), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45593] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 8,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [45678] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 8,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [45763] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3398), 1,
      anon_sym_COMMA,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    ACTIONS(3406), 1,
      sym_rel_operator,
    ACTIONS(3408), 1,
      sym_assign_operator,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3284), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [45860] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    ACTIONS(3406), 1,
      sym_rel_operator,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2235), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [45949] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1887), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1889), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46012] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    ACTIONS(3406), 1,
      sym_rel_operator,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2237), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3284), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [46105] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2238), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3442), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3444), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [46168] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1911), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1913), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [46231] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2240), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1915), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1917), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [46294] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2241), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1919), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1921), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [46357] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(2242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2113), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [46422] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2243), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1891), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1893), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [46485] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3398), 1,
      anon_sym_COMMA,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    ACTIONS(3406), 1,
      sym_rel_operator,
    ACTIONS(3408), 1,
      sym_assign_operator,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2244), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3322), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3320), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [46582] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3398), 1,
      anon_sym_COMMA,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    ACTIONS(3406), 1,
      sym_rel_operator,
    ACTIONS(3408), 1,
      sym_assign_operator,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3332), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3330), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [46679] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3446), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3448), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [46742] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3450), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3452), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [46805] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3454), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3456), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [46868] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2249), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3458), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3460), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [46931] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2250), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2083), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2085), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [46994] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3462), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3464), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [47057] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2125), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2127), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [47120] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2149), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2151), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [47183] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2254), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2107), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2109), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47246] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3398), 1,
      anon_sym_COMMA,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    ACTIONS(3406), 1,
      sym_rel_operator,
    ACTIONS(3408), 1,
      sym_assign_operator,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2255), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3342), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3340), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [47343] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2256), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2195), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2197), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [47406] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3466), 1,
      anon_sym_COMMA,
    STATE(2257), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3424), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3426), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [47471] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2258), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1935), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1937), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47534] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2259), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1959), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1961), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [47597] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2260), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1939), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1941), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47660] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2261), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1943), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1945), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47723] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2262), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1963), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1965), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [47786] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2187), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2189), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [47849] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2264), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1951), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1953), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47912] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3470), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [47975] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3424), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3426), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [48038] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2267), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2153), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2155), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48101] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2169), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2171), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [48164] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3466), 1,
      anon_sym_COMMA,
    ACTIONS(3476), 1,
      anon_sym_PIPE,
    ACTIONS(3478), 1,
      anon_sym_COLON_COLON,
    STATE(2269), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3472), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3474), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [48233] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2270), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1955), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1957), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48296] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2271), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2161), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2163), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [48359] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2272), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2121), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2123), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [48422] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    ACTIONS(3302), 1,
      sym_rel_operator,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3312), 1,
      anon_sym_COMMA,
    ACTIONS(3316), 1,
      sym_assign_operator,
    ACTIONS(3440), 1,
      anon_sym_SEMI,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2273), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3338), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3336), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [48521] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3398), 1,
      anon_sym_COMMA,
    ACTIONS(3400), 1,
      anon_sym_SEMI,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    ACTIONS(3406), 1,
      sym_rel_operator,
    ACTIONS(3408), 1,
      sym_assign_operator,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2274), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3338), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3336), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [48620] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2275), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3480), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3482), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [48683] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3414), 1,
      anon_sym_DOT,
    ACTIONS(3416), 1,
      anon_sym_POUND,
    ACTIONS(3418), 1,
      sym_hash_operator,
    STATE(2276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3486), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(3484), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [48752] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2277), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2043), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2045), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [48815] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3414), 1,
      anon_sym_DOT,
    ACTIONS(3416), 1,
      anon_sym_POUND,
    ACTIONS(3418), 1,
      sym_hash_operator,
    STATE(2278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3490), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(3488), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [48884] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2279), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2117), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2119), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48947] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2280), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2129), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2131), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [49010] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2281), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1895), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1897), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [49073] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2282), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2173), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2175), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49136] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2283), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2157), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2159), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [49199] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2284), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1959), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1961), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49262] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1963), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1965), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49325] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2286), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1967), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1969), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [49388] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3414), 1,
      anon_sym_DOT,
    STATE(2287), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1971), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1973), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [49453] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3472), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3474), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [49516] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2289), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3492), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3494), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [49579] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3478), 1,
      anon_sym_COLON_COLON,
    STATE(2290), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3496), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3498), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [49644] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2291), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3500), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3502), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [49707] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2292), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2195), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2197), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [49770] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2293), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1975), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1977), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [49833] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2294), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2199), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2201), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [49896] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2295), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3504), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3506), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [49959] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2296), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2071), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2073), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [50022] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2207), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2209), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [50085] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2298), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1983), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1985), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [50148] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2299), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2157), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2159), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50211] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2300), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1987), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1989), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [50274] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2301), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1991), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1993), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [50337] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2302), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1887), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1889), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [50400] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3508), 1,
      anon_sym_DOT,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(3512), 1,
      sym_hash_operator,
    STATE(2303), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3486), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3484), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50469] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3508), 1,
      anon_sym_DOT,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(3512), 1,
      sym_hash_operator,
    STATE(2304), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3516), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3514), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50538] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2079), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2081), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [50601] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2306), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2007), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2009), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [50664] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2307), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2047), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2049), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [50727] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2308), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2075), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2077), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [50790] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2309), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2011), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2013), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [50853] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2310), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2141), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2143), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [50916] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2311), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2191), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2193), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [50979] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3518), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3520), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [51042] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2313), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2157), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2159), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [51105] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2015), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2017), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [51168] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2315), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2019), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2021), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [51231] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2316), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2183), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2185), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [51294] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2317), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1895), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1897), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51357] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2318), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1967), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1969), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51420] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3522), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3524), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [51483] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2199), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2201), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [51546] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2321), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2023), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2025), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [51609] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3526), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3528), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [51672] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2323), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2027), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2029), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [51735] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2324), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2187), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2189), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [51798] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2325), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1999), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2001), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [51861] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3508), 1,
      anon_sym_DOT,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(3512), 1,
      sym_hash_operator,
    STATE(2326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3412), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3410), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51930] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3508), 1,
      anon_sym_DOT,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(3512), 1,
      sym_hash_operator,
    STATE(2327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3422), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3420), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51999] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2328), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2035), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2037), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [52062] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1975), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1977), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52125] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2330), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3530), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3532), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [52188] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2195), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2197), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52251] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2332), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1979), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1981), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52314] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2199), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2201), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52377] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1899), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1901), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [52440] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3466), 1,
      anon_sym_COMMA,
    ACTIONS(3476), 1,
      anon_sym_PIPE,
    ACTIONS(3478), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3538), 1,
      anon_sym_as,
    STATE(2335), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3534), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3536), 23,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [52511] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2336), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1641), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1645), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [52574] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2337), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2203), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2205), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [52637] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3508), 1,
      anon_sym_DOT,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(3512), 1,
      sym_hash_operator,
    STATE(2338), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3490), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3488), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52706] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2339), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1903), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1905), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [52769] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2340), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1907), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1909), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [52832] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3414), 1,
      anon_sym_DOT,
    ACTIONS(3416), 1,
      anon_sym_POUND,
    ACTIONS(3418), 1,
      sym_hash_operator,
    STATE(2341), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3516), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(3514), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [52901] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2342), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1983), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1985), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52964] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2343), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2203), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2205), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [53027] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2344), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2031), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2033), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [53090] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2345), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2179), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2181), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [53153] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2346), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2133), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2135), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [53216] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1927), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1929), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [53279] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2348), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1987), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1989), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53342] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2349), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1991), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1993), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53405] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2350), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2161), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2163), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [53468] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2351), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2169), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2171), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53531] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2352), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2165), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2167), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [53594] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2353), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3540), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3542), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [53657] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2354), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1999), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2001), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53720] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2355), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2051), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2053), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [53783] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2356), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2055), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2057), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [53846] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(2357), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 23,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2111), 24,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [53911] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2358), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3544), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3546), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [53974] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2359), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2203), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2205), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54037] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2360), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3548), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3550), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54100] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2361), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2031), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2033), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [54163] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2362), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3552), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3554), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54226] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2363), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2153), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2155), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [54289] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2364), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2003), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2005), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54352] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2365), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2007), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2009), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54415] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2366), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2121), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2123), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54478] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2367), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1947), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1949), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54541] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2368), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2165), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2167), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54604] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2369), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2011), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2013), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54667] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2370), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2149), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2151), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54730] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2371), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2157), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2159), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [54793] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3290), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3300), 1,
      sym_concat_operator,
    ACTIONS(3302), 1,
      sym_rel_operator,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3312), 1,
      anon_sym_COMMA,
    ACTIONS(3316), 1,
      sym_assign_operator,
    ACTIONS(3440), 1,
      anon_sym_SEMI,
    STATE(703), 1,
      sym_pow_operator,
    STATE(704), 1,
      sym_mult_operator,
    STATE(705), 1,
      sym_add_operator,
    STATE(706), 1,
      sym_and_operator,
    STATE(707), 1,
      sym_or_operator,
    STATE(2372), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3160), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3158), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [54892] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2373), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2173), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2175), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [54955] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2374), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1947), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1949), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [55018] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2375), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2183), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2185), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [55081] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3466), 1,
      anon_sym_COMMA,
    ACTIONS(3476), 1,
      anon_sym_PIPE,
    ACTIONS(3478), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3538), 1,
      anon_sym_as,
    STATE(2376), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3556), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3558), 23,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55152] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2377), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3560), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3562), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55215] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3466), 1,
      anon_sym_COMMA,
    ACTIONS(3478), 1,
      anon_sym_COLON_COLON,
    STATE(2378), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3560), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3562), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55282] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2379), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3564), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3566), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55345] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3466), 1,
      anon_sym_COMMA,
    STATE(2380), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3560), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3562), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55410] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2381), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3560), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3562), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55473] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2382), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1891), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1893), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [55536] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3466), 1,
      anon_sym_COMMA,
    ACTIONS(3476), 1,
      anon_sym_PIPE,
    ACTIONS(3478), 1,
      anon_sym_COLON_COLON,
    STATE(2383), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3568), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3570), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55605] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2384), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3568), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3570), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55668] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2385), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3572), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3574), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55731] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2386), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2031), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2033), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [55794] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3478), 1,
      anon_sym_COLON_COLON,
    STATE(2387), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3576), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3578), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55859] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2388), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2137), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2139), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [55922] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2389), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2103), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2105), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [55985] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2015), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2017), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56048] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2391), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2019), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2021), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56111] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2392), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2023), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2025), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56174] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2393), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2129), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2131), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56237] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2394), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2027), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2029), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56300] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2395), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1923), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1925), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56363] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2396), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2145), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2147), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [56426] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3508), 1,
      anon_sym_DOT,
    STATE(2397), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1971), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1973), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56491] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2398), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2035), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2037), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56554] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2399), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2125), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2127), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56617] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2400), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2103), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2105), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56680] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(2401), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 23,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2111), 24,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [56745] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2402), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1899), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1901), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56808] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2403), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1995), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1997), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [56871] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2404), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1903), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1905), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56934] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2405), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1907), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1909), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56997] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2406), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3580), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3582), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57060] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2407), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1951), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1953), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [57123] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3466), 1,
      anon_sym_COMMA,
    ACTIONS(3478), 1,
      anon_sym_COLON_COLON,
    STATE(2408), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3424), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3426), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57190] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2409), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2051), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2053), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [57253] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2410), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1935), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1937), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [57316] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2411), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2055), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2057), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [57379] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2412), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2059), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2061), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [57442] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2413), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2121), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2123), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [57505] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2414), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2165), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2167), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [57568] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2415), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2067), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2069), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [57631] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2416), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2107), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2109), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [57694] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2417), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3584), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3586), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57757] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2418), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1911), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1913), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [57820] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2419), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1915), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1917), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [57883] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2420), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1919), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1921), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [57946] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2421), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2187), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2189), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [58009] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2422), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3588), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3590), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [58072] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2423), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2083), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2085), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [58135] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2424), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1995), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1997), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [58198] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2425), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3592), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3594), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [58261] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2426), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3596), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3598), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [58324] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2427), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3600), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3602), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [58387] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2428), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3604), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3606), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [58450] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2429), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2087), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2089), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [58513] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2430), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2091), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2093), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [58576] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2431), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1979), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1981), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [58639] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2432), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2117), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2119), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [58702] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2433), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2191), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2193), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [58765] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2434), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2095), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2097), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [58828] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2435), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3608), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3610), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [58891] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2436), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2099), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2101), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [58954] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2437), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1939), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1941), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [59017] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2438), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3612), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3614), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [59080] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2439), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2207), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2209), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [59143] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(2440), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2113), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [59208] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2441), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3616), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3618), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [59271] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2442), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3620), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3622), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [59334] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2443), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1943), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1945), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [59397] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2444), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3624), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3626), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [59460] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2445), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2075), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2077), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [59522] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2446), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2191), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2193), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [59584] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3628), 1,
      anon_sym_DOT,
    ACTIONS(3630), 1,
      anon_sym_POUND,
    ACTIONS(3632), 1,
      sym_hash_operator,
    STATE(2447), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3412), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(3410), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [59652] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3628), 1,
      anon_sym_DOT,
    ACTIONS(3630), 1,
      anon_sym_POUND,
    ACTIONS(3632), 1,
      sym_hash_operator,
    STATE(2448), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3422), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(3420), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [59720] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2449), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1947), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1949), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [59782] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2450), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1995), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1997), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [59844] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2451), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2121), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2123), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [59906] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2452), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2043), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2045), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [59968] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2453), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2071), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2073), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60030] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2454), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2079), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2081), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60092] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2455), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2141), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2143), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60154] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2456), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1927), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1929), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60216] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2457), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1935), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1937), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60278] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2458), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1939), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1941), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60340] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2459), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1943), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1945), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60402] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2460), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1955), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1957), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60464] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2461), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1959), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1961), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60526] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2462), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1963), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1965), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60588] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2463), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1895), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1897), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60650] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2464), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1967), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1969), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60712] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3634), 1,
      anon_sym_DOT,
    STATE(2465), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1971), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1973), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60776] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2466), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1975), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1977), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60838] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2467), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1983), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1985), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60900] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2468), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1987), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1989), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60962] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2469), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1991), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1993), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61024] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2470), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2007), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2009), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61086] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2011), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2013), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61148] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2472), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2015), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2017), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61210] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2473), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2019), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2021), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61272] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2474), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2023), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2025), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61334] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2475), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2027), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2029), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61396] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2476), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2035), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2037), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61458] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2477), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1899), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1901), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61520] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2478), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1903), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1905), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61582] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2479), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1907), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1909), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61644] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2480), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2051), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2053), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61706] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2481), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2055), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2057), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61768] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2482), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1911), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1913), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61830] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2483), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1915), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1917), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61892] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2484), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1919), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1921), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61954] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2485), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2083), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2085), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62016] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2486), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2207), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2209), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62078] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2487), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2047), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2049), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62140] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2488), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1891), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1893), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62202] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2489), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1951), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1953), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62264] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2490), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1979), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1981), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62326] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2491), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1999), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2001), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62388] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2492), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2003), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2005), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62450] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2493), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1887), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1889), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62512] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2494), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1923), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1925), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62574] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2495), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2059), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2061), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62636] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2496), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2063), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2065), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62698] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2497), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2067), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2069), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62760] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2498), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2087), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2089), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62822] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2499), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2091), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2093), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62884] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2500), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2095), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2097), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62946] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2501), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2099), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2101), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63008] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3628), 1,
      anon_sym_DOT,
    ACTIONS(3630), 1,
      anon_sym_POUND,
    ACTIONS(3632), 1,
      sym_hash_operator,
    STATE(2502), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3516), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(3514), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [63076] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2503), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2149), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2151), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [63138] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2504), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2195), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2197), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [63200] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2505), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2199), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2201), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [63262] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2506), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2203), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2205), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [63324] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2507), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2031), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2033), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [63386] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2508), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2121), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2123), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63448] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2509), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2157), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2159), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63510] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2510), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2129), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2131), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63572] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2511), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2137), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2139), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63634] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2512), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2161), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2163), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63696] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2513), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2103), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2105), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63758] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2514), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2107), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2109), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63820] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(2515), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2113), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63884] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2516), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2117), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2119), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63946] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2517), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2125), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2127), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64008] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2518), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2149), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2151), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64070] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2519), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2169), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2171), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64132] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(2520), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2113), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64196] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2521), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2179), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2181), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64258] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2522), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2195), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2197), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64320] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2523), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2199), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2201), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64382] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2524), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2203), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2205), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64444] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2525), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2031), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2033), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64506] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2526), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2133), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2135), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64568] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2527), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2153), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2155), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64630] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2528), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2173), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2175), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64692] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2529), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2145), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2147), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64754] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2530), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2165), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2167), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64816] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2531), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2183), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2185), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64878] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2532), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2187), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2189), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64940] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2533), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2191), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2193), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [65002] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2534), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [65074] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2535), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 26,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [65148] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2536), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 26,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [65222] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2537), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 10,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [65300] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3356), 1,
      anon_sym_COMMA,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    ACTIONS(3362), 1,
      sym_rel_operator,
    ACTIONS(3364), 1,
      sym_assign_operator,
    ACTIONS(3636), 1,
      anon_sym_SEMI,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2538), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3156), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3154), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [65398] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3356), 1,
      anon_sym_COMMA,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    ACTIONS(3362), 1,
      sym_rel_operator,
    ACTIONS(3364), 1,
      sym_assign_operator,
    ACTIONS(3636), 1,
      anon_sym_SEMI,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2539), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3160), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3158), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [65496] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3638), 1,
      anon_sym_COMMA,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    ACTIONS(3644), 1,
      sym_rel_operator,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3346), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3344), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [65590] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2541), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3350), 7,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3348), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [65672] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2542), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 7,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [65756] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 7,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [65840] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    ACTIONS(3644), 1,
      sym_rel_operator,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2544), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 5,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [65928] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    ACTIONS(3644), 1,
      sym_rel_operator,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2545), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3284), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [66020] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3638), 1,
      anon_sym_COMMA,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    ACTIONS(3644), 1,
      sym_rel_operator,
    ACTIONS(3646), 1,
      sym_assign_operator,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2546), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3322), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3320), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [66116] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3638), 1,
      anon_sym_COMMA,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    ACTIONS(3644), 1,
      sym_rel_operator,
    ACTIONS(3646), 1,
      sym_assign_operator,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2547), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3332), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3330), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [66212] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3370), 1,
      anon_sym_COMMA,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    ACTIONS(3376), 1,
      sym_rel_operator,
    ACTIONS(3378), 1,
      sym_assign_operator,
    ACTIONS(3648), 1,
      anon_sym_SEMI,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2548), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3338), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3336), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [66310] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3638), 1,
      anon_sym_COMMA,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    ACTIONS(3644), 1,
      sym_rel_operator,
    ACTIONS(3646), 1,
      sym_assign_operator,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2549), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3342), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3340), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [66406] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3638), 1,
      anon_sym_COMMA,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    ACTIONS(3644), 1,
      sym_rel_operator,
    ACTIONS(3646), 1,
      sym_assign_operator,
    ACTIONS(3650), 1,
      anon_sym_SEMI,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2550), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3338), 3,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3336), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [66504] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2551), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1947), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1949), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [66566] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2552), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1995), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1997), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [66628] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2553), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2207), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2209), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [66690] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2554), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2043), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2045), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [66752] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2555), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2047), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2049), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [66814] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2556), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2071), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2073), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [66876] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3634), 1,
      anon_sym_DOT,
    ACTIONS(3652), 1,
      anon_sym_POUND,
    ACTIONS(3654), 1,
      sym_hash_operator,
    STATE(2557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3412), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3410), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [66944] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3634), 1,
      anon_sym_DOT,
    ACTIONS(3652), 1,
      anon_sym_POUND,
    ACTIONS(3654), 1,
      sym_hash_operator,
    STATE(2558), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3422), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3420), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [67012] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2559), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2075), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2077), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [67074] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2560), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2079), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2081), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [67136] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2561), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2141), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2143), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [67198] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2562), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1927), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1929), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [67260] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2563), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1891), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1893), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [67322] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3356), 1,
      anon_sym_COMMA,
    ACTIONS(3358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3360), 1,
      sym_concat_operator,
    ACTIONS(3362), 1,
      sym_rel_operator,
    ACTIONS(3364), 1,
      sym_assign_operator,
    ACTIONS(3636), 1,
      anon_sym_SEMI,
    STATE(833), 1,
      sym_pow_operator,
    STATE(834), 1,
      sym_mult_operator,
    STATE(835), 1,
      sym_add_operator,
    STATE(836), 1,
      sym_and_operator,
    STATE(837), 1,
      sym_or_operator,
    STATE(2564), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3338), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3336), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [67420] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2565), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1935), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1937), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [67482] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2566), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1939), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1941), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [67544] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2567), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1943), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1945), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [67606] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2568), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1951), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1953), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [67668] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2569), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1955), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1957), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [67730] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3634), 1,
      anon_sym_DOT,
    ACTIONS(3652), 1,
      anon_sym_POUND,
    ACTIONS(3654), 1,
      sym_hash_operator,
    STATE(2570), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3486), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3484), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [67798] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3634), 1,
      anon_sym_DOT,
    ACTIONS(3652), 1,
      anon_sym_POUND,
    ACTIONS(3654), 1,
      sym_hash_operator,
    STATE(2571), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3490), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3488), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [67866] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2572), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1959), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1961), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [67928] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2573), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1963), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1965), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [67990] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2574), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1895), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1897), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68052] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2575), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1967), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1969), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68114] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3628), 1,
      anon_sym_DOT,
    STATE(2576), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1971), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1973), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68178] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2577), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1975), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1977), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68240] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2578), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1979), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1981), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68302] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3634), 1,
      anon_sym_DOT,
    ACTIONS(3652), 1,
      anon_sym_POUND,
    ACTIONS(3654), 1,
      sym_hash_operator,
    STATE(2579), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3516), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3514), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [68370] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2580), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1983), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1985), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68432] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2581), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1987), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1989), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68494] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1991), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1993), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68556] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2583), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1999), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2001), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68618] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2584), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2003), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2005), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68680] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2007), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2009), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68742] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2586), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2011), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2013), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68804] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2587), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1887), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1889), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68866] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2588), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2015), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2017), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68928] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2589), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2019), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2021), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [68990] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2023), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2025), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69052] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2591), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2027), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2029), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69114] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1923), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1925), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69176] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2593), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2035), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2037), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69238] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2594), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1899), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1901), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69300] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2595), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1903), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1905), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69362] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2596), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1907), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1909), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69424] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2597), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2051), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2053), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69486] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2598), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2055), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2057), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69548] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2599), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2059), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2061), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69610] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2600), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2063), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2065), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69672] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2601), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2067), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2069), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69734] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2602), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1911), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1913), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69796] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1915), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1917), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69858] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2604), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1919), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1921), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69920] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2605), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2083), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2085), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [69982] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2606), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2087), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2089), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [70044] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2607), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2091), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2093), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [70106] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2608), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2095), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2097), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [70168] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2609), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2099), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2101), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [70230] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2610), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2127), 23,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2125), 24,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [70292] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2611), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2139), 23,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2137), 24,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [70354] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3370), 1,
      anon_sym_COMMA,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    ACTIONS(3376), 1,
      sym_rel_operator,
    ACTIONS(3378), 1,
      sym_assign_operator,
    ACTIONS(3648), 1,
      anon_sym_SEMI,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2612), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3156), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3154), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [70452] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2613), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2181), 23,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2179), 24,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [70514] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3370), 1,
      anon_sym_COMMA,
    ACTIONS(3372), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3374), 1,
      sym_concat_operator,
    ACTIONS(3376), 1,
      sym_rel_operator,
    ACTIONS(3378), 1,
      sym_assign_operator,
    ACTIONS(3648), 1,
      anon_sym_SEMI,
    STATE(658), 1,
      sym_pow_operator,
    STATE(659), 1,
      sym_mult_operator,
    STATE(660), 1,
      sym_add_operator,
    STATE(661), 1,
      sym_and_operator,
    STATE(662), 1,
      sym_or_operator,
    STATE(2614), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3160), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3158), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [70612] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2615), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2103), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2105), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [70674] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2616), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2107), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2109), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [70736] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2617), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2117), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2119), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [70798] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2618), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2129), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2131), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [70860] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2619), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2133), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2135), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [70922] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2620), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2137), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2139), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [70984] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2621), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2147), 23,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2145), 24,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [71046] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2622), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2153), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2155), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [71108] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2623), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2161), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2163), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [71170] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2624), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2169), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2171), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [71232] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2625), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2173), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2175), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [71294] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3628), 1,
      anon_sym_DOT,
    ACTIONS(3630), 1,
      anon_sym_POUND,
    ACTIONS(3632), 1,
      sym_hash_operator,
    STATE(2626), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3486), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(3484), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [71362] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2627), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2145), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2147), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [71424] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3628), 1,
      anon_sym_DOT,
    ACTIONS(3630), 1,
      anon_sym_POUND,
    ACTIONS(3632), 1,
      sym_hash_operator,
    STATE(2628), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3490), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(3488), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [71492] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2629), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2165), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2167), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [71554] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3638), 1,
      anon_sym_COMMA,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    ACTIONS(3644), 1,
      sym_rel_operator,
    ACTIONS(3646), 1,
      sym_assign_operator,
    ACTIONS(3650), 1,
      anon_sym_SEMI,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2630), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3156), 3,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3154), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [71652] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2631), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2183), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2185), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [71714] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2632), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2187), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2189), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [71776] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3638), 1,
      anon_sym_COMMA,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    ACTIONS(3644), 1,
      sym_rel_operator,
    ACTIONS(3646), 1,
      sym_assign_operator,
    ACTIONS(3650), 1,
      anon_sym_SEMI,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2633), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3160), 3,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3158), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [71874] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3638), 1,
      anon_sym_COMMA,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    ACTIONS(3644), 1,
      sym_rel_operator,
    ACTIONS(3646), 1,
      sym_assign_operator,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2634), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3284), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [71970] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2635), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3658), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3656), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [72031] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3398), 1,
      anon_sym_COMMA,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    ACTIONS(3406), 1,
      sym_rel_operator,
    ACTIONS(3408), 1,
      sym_assign_operator,
    ACTIONS(3660), 1,
      anon_sym_SEMI,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2636), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3160), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3158), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [72128] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2637), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3664), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3662), 28,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [72189] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2638), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3668), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3666), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [72250] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3670), 1,
      anon_sym_COMMA,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    ACTIONS(3676), 1,
      sym_rel_operator,
    ACTIONS(3678), 1,
      sym_assign_operator,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2639), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3332), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3330), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [72345] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2640), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3682), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3680), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [72406] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2641), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3686), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3684), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [72467] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2642), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3690), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3688), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [72528] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2643), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3694), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3692), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [72589] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2644), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3698), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3696), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [72650] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3670), 1,
      anon_sym_COMMA,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    ACTIONS(3676), 1,
      sym_rel_operator,
    ACTIONS(3678), 1,
      sym_assign_operator,
    ACTIONS(3700), 1,
      anon_sym_SEMI,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2645), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3338), 3,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3336), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [72747] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2646), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3704), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [72808] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2647), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3708), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3706), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [72869] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2648), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3712), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3710), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [72930] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2649), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3716), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3714), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [72991] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3670), 1,
      anon_sym_COMMA,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    ACTIONS(3676), 1,
      sym_rel_operator,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2650), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3346), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3344), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [73084] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2651), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [73145] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2652), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(917), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(915), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [73206] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2653), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3720), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3718), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [73267] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2654), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3724), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3722), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [73328] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3730), 1,
      anon_sym_else,
    STATE(2640), 1,
      sym_else_clause,
    STATE(2655), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [73393] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2656), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3350), 7,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3348), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [73474] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2657), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 7,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [73557] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3398), 1,
      anon_sym_COMMA,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    ACTIONS(3406), 1,
      sym_rel_operator,
    ACTIONS(3408), 1,
      sym_assign_operator,
    ACTIONS(3660), 1,
      anon_sym_SEMI,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2658), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3338), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3336), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [73654] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2659), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 7,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [73737] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3670), 1,
      anon_sym_COMMA,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    ACTIONS(3676), 1,
      sym_rel_operator,
    ACTIONS(3678), 1,
      sym_assign_operator,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2660), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3284), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [73832] = 33,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3732), 1,
      anon_sym_LPAREN,
    ACTIONS(3734), 1,
      anon_sym_external,
    ACTIONS(3736), 1,
      anon_sym_private,
    ACTIONS(3738), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3740), 1,
      anon_sym_PIPE,
    ACTIONS(3742), 1,
      anon_sym_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      sym__uppercase_identifier,
    STATE(2661), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(5244), 1,
      sym__type,
    STATE(5629), 1,
      sym_constructor_declaration,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3744), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(5394), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [73949] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2662), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [74022] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [74095] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2664), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 10,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [74172] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3750), 1,
      anon_sym_COLON,
    STATE(2665), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3754), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3752), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [74235] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    ACTIONS(3676), 1,
      sym_rel_operator,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2666), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 5,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [74322] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    ACTIONS(3676), 1,
      sym_rel_operator,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2667), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3284), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [74413] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3670), 1,
      anon_sym_COMMA,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    ACTIONS(3676), 1,
      sym_rel_operator,
    ACTIONS(3678), 1,
      sym_assign_operator,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2668), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3322), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3320), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [74508] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3670), 1,
      anon_sym_COMMA,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    ACTIONS(3676), 1,
      sym_rel_operator,
    ACTIONS(3678), 1,
      sym_assign_operator,
    ACTIONS(3700), 1,
      anon_sym_SEMI,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2669), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3156), 3,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3154), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [74605] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2670), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3758), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3756), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [74666] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3398), 1,
      anon_sym_COMMA,
    ACTIONS(3402), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3404), 1,
      sym_concat_operator,
    ACTIONS(3406), 1,
      sym_rel_operator,
    ACTIONS(3408), 1,
      sym_assign_operator,
    ACTIONS(3660), 1,
      anon_sym_SEMI,
    STATE(791), 1,
      sym_pow_operator,
    STATE(792), 1,
      sym_mult_operator,
    STATE(793), 1,
      sym_add_operator,
    STATE(794), 1,
      sym_and_operator,
    STATE(795), 1,
      sym_or_operator,
    STATE(2671), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3156), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3154), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [74763] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3670), 1,
      anon_sym_COMMA,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    ACTIONS(3676), 1,
      sym_rel_operator,
    ACTIONS(3678), 1,
      sym_assign_operator,
    ACTIONS(3700), 1,
      anon_sym_SEMI,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2672), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3160), 3,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3158), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [74860] = 33,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3732), 1,
      anon_sym_LPAREN,
    ACTIONS(3734), 1,
      anon_sym_external,
    ACTIONS(3740), 1,
      anon_sym_PIPE,
    ACTIONS(3742), 1,
      anon_sym_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      sym__uppercase_identifier,
    ACTIONS(3760), 1,
      anon_sym_private,
    ACTIONS(3762), 1,
      anon_sym_DOT_DOT,
    STATE(2673), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(5221), 1,
      sym__type,
    STATE(5629), 1,
      sym_constructor_declaration,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3744), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(5380), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [74977] = 33,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(3764), 1,
      sym__lowercase_identifier,
    ACTIONS(3766), 1,
      anon_sym_LPAREN,
    ACTIONS(3768), 1,
      anon_sym_external,
    ACTIONS(3770), 1,
      anon_sym_private,
    ACTIONS(3772), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3774), 1,
      anon_sym_PIPE,
    ACTIONS(3776), 1,
      anon_sym_LBRACK,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3794), 1,
      sym__uppercase_identifier,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    STATE(2674), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3895), 1,
      sym__type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(4197), 1,
      sym_constructor_declaration,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3778), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4087), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [75094] = 33,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(3764), 1,
      sym__lowercase_identifier,
    ACTIONS(3766), 1,
      anon_sym_LPAREN,
    ACTIONS(3768), 1,
      anon_sym_external,
    ACTIONS(3774), 1,
      anon_sym_PIPE,
    ACTIONS(3776), 1,
      anon_sym_LBRACK,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3794), 1,
      sym__uppercase_identifier,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3798), 1,
      anon_sym_private,
    ACTIONS(3800), 1,
      anon_sym_DOT_DOT,
    STATE(2675), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3859), 1,
      sym__type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(4197), 1,
      sym_constructor_declaration,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3778), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4005), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [75211] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2676), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3804), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3802), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75272] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3808), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3806), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [75333] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2678), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3812), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3810), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75394] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2679), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3816), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3814), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [75455] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3670), 1,
      anon_sym_COMMA,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    ACTIONS(3676), 1,
      sym_rel_operator,
    ACTIONS(3678), 1,
      sym_assign_operator,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2680), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3342), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3340), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [75550] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2681), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3820), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3818), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [75611] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2682), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3824), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3822), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [75672] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2683), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3828), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3826), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [75733] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2684), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3832), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3830), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [75794] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2685), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3836), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3834), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [75855] = 33,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(3768), 1,
      anon_sym_external,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3794), 1,
      sym__uppercase_identifier,
    ACTIONS(3838), 1,
      sym__lowercase_identifier,
    ACTIONS(3840), 1,
      anon_sym_LPAREN,
    ACTIONS(3842), 1,
      anon_sym_private,
    ACTIONS(3844), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3846), 1,
      anon_sym_PIPE,
    ACTIONS(3848), 1,
      anon_sym_LBRACK,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    STATE(2686), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4200), 1,
      sym__type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(4646), 1,
      sym_constructor_declaration,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3778), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4414), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [75972] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3864), 1,
      anon_sym_COLON,
    STATE(2687), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3754), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3752), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [76035] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3870), 1,
      anon_sym_PIPE,
    STATE(2688), 1,
      sym_attribute,
    STATE(2690), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [76100] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3870), 1,
      anon_sym_PIPE,
    STATE(2689), 1,
      sym_attribute,
    STATE(2690), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3874), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3872), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [76165] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3876), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2690), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3704), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [76228] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3870), 1,
      anon_sym_PIPE,
    STATE(2688), 1,
      aux_sym__match_cases_repeat1,
    STATE(2691), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3881), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3879), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [76293] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3870), 1,
      anon_sym_PIPE,
    STATE(2689), 1,
      aux_sym__match_cases_repeat1,
    STATE(2692), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [76358] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2693), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3885), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3883), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [76419] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2694), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3889), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3887), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [76480] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2695), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3893), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3891), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76541] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3897), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3895), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76602] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2697), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3901), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3899), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [76663] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2698), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3905), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3903), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [76724] = 33,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(3768), 1,
      anon_sym_external,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3794), 1,
      sym__uppercase_identifier,
    ACTIONS(3838), 1,
      sym__lowercase_identifier,
    ACTIONS(3840), 1,
      anon_sym_LPAREN,
    ACTIONS(3846), 1,
      anon_sym_PIPE,
    ACTIONS(3848), 1,
      anon_sym_LBRACK,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3907), 1,
      anon_sym_private,
    ACTIONS(3909), 1,
      anon_sym_DOT_DOT,
    STATE(2699), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4172), 1,
      sym__type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(4646), 1,
      sym_constructor_declaration,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3778), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4421), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [76841] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [76912] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2701), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3913), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3911), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76973] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3730), 1,
      anon_sym_else,
    STATE(2681), 1,
      sym_else_clause,
    STATE(2702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [77038] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2703), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3921), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3919), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [77099] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2704), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3925), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3923), 32,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [77160] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3927), 1,
      anon_sym_PIPE,
    STATE(2705), 1,
      sym_attribute,
    STATE(2732), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [77224] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2706), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3836), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3834), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [77284] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2707), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3686), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3684), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [77344] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3694), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3692), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [77404] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2709), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3698), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3696), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [77464] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3929), 1,
      anon_sym_COLON,
    STATE(2710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3754), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3752), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [77526] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3690), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3688), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [77586] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3716), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3714), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [77646] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [77706] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2714), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3885), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3883), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [77766] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3638), 1,
      anon_sym_COMMA,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    ACTIONS(3644), 1,
      sym_rel_operator,
    ACTIONS(3646), 1,
      sym_assign_operator,
    ACTIONS(3931), 1,
      anon_sym_SEMI,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2715), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3158), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [77862] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2716), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3658), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3656), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [77922] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3933), 1,
      anon_sym_else,
    STATE(2681), 1,
      sym_else_clause,
    STATE(2717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [77986] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2718), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3720), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3718), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [78046] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3935), 1,
      anon_sym_PIPE,
    STATE(2719), 1,
      sym_attribute,
    STATE(2721), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [78110] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3935), 1,
      anon_sym_PIPE,
    STATE(2720), 1,
      sym_attribute,
    STATE(2721), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3874), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3872), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [78174] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3937), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2721), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3704), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [78236] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3933), 1,
      anon_sym_else,
    STATE(2640), 1,
      sym_else_clause,
    STATE(2722), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [78300] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2723), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3724), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3722), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [78360] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3935), 1,
      anon_sym_PIPE,
    STATE(2719), 1,
      aux_sym__match_cases_repeat1,
    STATE(2724), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3881), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3879), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [78424] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2725), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3921), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3919), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [78484] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3940), 1,
      anon_sym_else,
    STATE(2640), 1,
      sym_else_clause,
    STATE(2726), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [78548] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3940), 1,
      anon_sym_else,
    STATE(2681), 1,
      sym_else_clause,
    STATE(2727), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [78612] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3638), 1,
      anon_sym_COMMA,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    ACTIONS(3644), 1,
      sym_rel_operator,
    ACTIONS(3646), 1,
      sym_assign_operator,
    ACTIONS(3931), 1,
      anon_sym_SEMI,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3154), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [78708] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2729), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3758), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3756), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [78768] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2730), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(917), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(915), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [78828] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3927), 1,
      anon_sym_PIPE,
    STATE(2731), 1,
      sym_attribute,
    STATE(2733), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [78892] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3927), 1,
      anon_sym_PIPE,
    STATE(2732), 1,
      sym_attribute,
    STATE(2733), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3874), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3872), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [78956] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3942), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2733), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3704), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [79018] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2734), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3901), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3899), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [79078] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3927), 1,
      anon_sym_PIPE,
    STATE(2731), 1,
      aux_sym__match_cases_repeat1,
    STATE(2735), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3881), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3879), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [79142] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2736), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3668), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3666), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [79202] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3945), 1,
      anon_sym_COLON,
    STATE(2737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3754), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3752), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [79264] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2738), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3682), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3680), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [79324] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2739), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3889), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3887), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [79384] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2740), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3905), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3903), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [79444] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2741), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3925), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3923), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [79504] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3638), 1,
      anon_sym_COMMA,
    ACTIONS(3640), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3642), 1,
      sym_concat_operator,
    ACTIONS(3644), 1,
      sym_rel_operator,
    ACTIONS(3646), 1,
      sym_assign_operator,
    ACTIONS(3931), 1,
      anon_sym_SEMI,
    STATE(861), 1,
      sym_pow_operator,
    STATE(862), 1,
      sym_mult_operator,
    STATE(863), 1,
      sym_add_operator,
    STATE(864), 1,
      sym_and_operator,
    STATE(865), 1,
      sym_or_operator,
    STATE(2742), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3338), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3336), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [79600] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2743), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3808), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3806), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [79660] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3312), 1,
      anon_sym_COMMA,
    STATE(2744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [79722] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2745), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3816), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3814), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [79782] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3820), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3818), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [79842] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2747), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3824), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3822), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [79902] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2748), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3828), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3826), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [79962] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3935), 1,
      anon_sym_PIPE,
    STATE(2720), 1,
      aux_sym__match_cases_repeat1,
    STATE(2749), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [80026] = 27,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(1425), 1,
      anon_sym_LBRACK,
    ACTIONS(1427), 1,
      anon_sym_LBRACE,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2556), 1,
      sym__lowercase_identifier,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3947), 1,
      anon_sym_LPAREN,
    ACTIONS(3949), 1,
      sym__uppercase_identifier,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    STATE(2750), 1,
      sym_attribute,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5180), 1,
      sym__constructor_name,
    STATE(5288), 1,
      sym__simple_binding_pattern,
    STATE(7377), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5312), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(5310), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [80129] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1357), 1,
      sym__lowercase_identifier,
    ACTIONS(1369), 1,
      anon_sym_LBRACK,
    ACTIONS(1373), 1,
      anon_sym_LBRACE,
    ACTIONS(1383), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(3953), 1,
      anon_sym_LPAREN,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2751), 1,
      sym_attribute,
    STATE(5151), 1,
      sym__simple_pattern,
    STATE(7274), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5133), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(5132), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [80230] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3370), 1,
      anon_sym_COMMA,
    STATE(2752), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [80291] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1357), 1,
      sym__lowercase_identifier,
    ACTIONS(1369), 1,
      anon_sym_LBRACK,
    ACTIONS(1373), 1,
      anon_sym_LBRACE,
    ACTIONS(1383), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3949), 1,
      sym__uppercase_identifier,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(3953), 1,
      anon_sym_LPAREN,
    STATE(2753), 1,
      sym_attribute,
    STATE(5151), 1,
      sym__simple_pattern,
    STATE(5180), 1,
      sym__constructor_name,
    STATE(7420), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5133), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(5132), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [80392] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2754), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3154), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [80451] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2578), 1,
      sym__lowercase_identifier,
    ACTIONS(2582), 1,
      anon_sym_LBRACK,
    ACTIONS(2586), 1,
      anon_sym_LBRACE,
    ACTIONS(2590), 1,
      anon_sym_POUND,
    ACTIONS(2592), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2596), 1,
      anon_sym_begin,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2604), 1,
      aux_sym_number_token1,
    ACTIONS(2606), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2608), 1,
      anon_sym_SQUOTE,
    ACTIONS(2610), 1,
      anon_sym_DQUOTE,
    ACTIONS(3955), 1,
      anon_sym_LPAREN,
    ACTIONS(3957), 1,
      sym__uppercase_identifier,
    ACTIONS(3959), 1,
      aux_sym_tag_token1,
    STATE(2755), 1,
      sym_attribute,
    STATE(5611), 1,
      sym__simple_pattern,
    STATE(5631), 1,
      sym__constructor_name,
    STATE(7457), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2584), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5451), 2,
      sym__constant,
      sym_signed_number,
    STATE(5427), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5636), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(5633), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [80552] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3925), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3923), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [80611] = 27,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2590), 1,
      anon_sym_POUND,
    ACTIONS(2596), 1,
      anon_sym_begin,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2604), 1,
      aux_sym_number_token1,
    ACTIONS(2606), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2608), 1,
      anon_sym_SQUOTE,
    ACTIONS(2610), 1,
      anon_sym_DQUOTE,
    ACTIONS(2618), 1,
      sym__lowercase_identifier,
    ACTIONS(2622), 1,
      anon_sym_LBRACK,
    ACTIONS(2624), 1,
      anon_sym_LBRACE,
    ACTIONS(2628), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3957), 1,
      sym__uppercase_identifier,
    ACTIONS(3959), 1,
      aux_sym_tag_token1,
    ACTIONS(3961), 1,
      anon_sym_LPAREN,
    STATE(2757), 1,
      sym_attribute,
    STATE(5617), 1,
      sym_parenthesized_operator,
    STATE(5631), 1,
      sym__constructor_name,
    STATE(5653), 1,
      sym__simple_binding_pattern,
    STATE(7416), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2584), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5451), 2,
      sym__constant,
      sym_signed_number,
    STATE(5427), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5561), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(5557), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [80714] = 27,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(1425), 1,
      anon_sym_LBRACK,
    ACTIONS(1427), 1,
      anon_sym_LBRACE,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2938), 1,
      sym__uppercase_identifier,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(3064), 1,
      sym__lowercase_identifier,
    ACTIONS(3947), 1,
      anon_sym_LPAREN,
    STATE(2399), 1,
      sym__constructor_name,
    STATE(2758), 1,
      sym_attribute,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5315), 1,
      sym__simple_binding_pattern,
    STATE(7269), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5312), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(5310), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [80817] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3670), 1,
      anon_sym_COMMA,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    ACTIONS(3676), 1,
      sym_rel_operator,
    ACTIONS(3678), 1,
      sym_assign_operator,
    ACTIONS(3963), 1,
      anon_sym_SEMI,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2759), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3158), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [80912] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3552), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3554), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [80971] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2761), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3668), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3666), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [81030] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2762), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3470), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [81089] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3670), 1,
      anon_sym_COMMA,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    ACTIONS(3676), 1,
      sym_rel_operator,
    ACTIONS(3678), 1,
      sym_assign_operator,
    ACTIONS(3963), 1,
      anon_sym_SEMI,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2763), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3154), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [81184] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3158), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [81243] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2765), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3480), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3482), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [81302] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2766), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3394), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3396), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [81361] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2767), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3682), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3680), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [81420] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3732), 1,
      anon_sym_LPAREN,
    ACTIONS(3740), 1,
      anon_sym_PIPE,
    ACTIONS(3742), 1,
      anon_sym_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      sym__uppercase_identifier,
    ACTIONS(3760), 1,
      anon_sym_private,
    ACTIONS(3762), 1,
      anon_sym_DOT_DOT,
    STATE(2768), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(5221), 1,
      sym__type,
    STATE(5629), 1,
      sym_constructor_declaration,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3744), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(5380), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [81533] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2769), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3686), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3684), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [81592] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(3764), 1,
      sym__lowercase_identifier,
    ACTIONS(3766), 1,
      anon_sym_LPAREN,
    ACTIONS(3770), 1,
      anon_sym_private,
    ACTIONS(3772), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3774), 1,
      anon_sym_PIPE,
    ACTIONS(3776), 1,
      anon_sym_LBRACK,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3794), 1,
      sym__uppercase_identifier,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    STATE(2770), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3895), 1,
      sym__type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(4197), 1,
      sym_constructor_declaration,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3778), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4087), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [81705] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3694), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3692), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [81764] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(3764), 1,
      sym__lowercase_identifier,
    ACTIONS(3766), 1,
      anon_sym_LPAREN,
    ACTIONS(3774), 1,
      anon_sym_PIPE,
    ACTIONS(3776), 1,
      anon_sym_LBRACK,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3794), 1,
      sym__uppercase_identifier,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3798), 1,
      anon_sym_private,
    ACTIONS(3800), 1,
      anon_sym_DOT_DOT,
    STATE(2772), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3859), 1,
      sym__type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(4197), 1,
      sym_constructor_declaration,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3778), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4005), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [81877] = 27,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(2360), 1,
      sym__simple_binding_pattern,
    STATE(2773), 1,
      sym_attribute,
    STATE(7246), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2336), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [81980] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2774), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3698), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3696), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [82039] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2775), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3716), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3714), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [82098] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1357), 1,
      sym__lowercase_identifier,
    ACTIONS(1369), 1,
      anon_sym_LBRACK,
    ACTIONS(1373), 1,
      anon_sym_LBRACE,
    ACTIONS(1383), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(3953), 1,
      anon_sym_LPAREN,
    ACTIONS(3965), 1,
      sym__uppercase_identifier,
    STATE(2776), 1,
      sym_attribute,
    STATE(3688), 1,
      sym__constructor_name,
    STATE(5152), 1,
      sym__simple_pattern,
    STATE(7335), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5133), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(5132), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [82199] = 27,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(1415), 1,
      sym__lowercase_identifier,
    ACTIONS(1425), 1,
      anon_sym_LBRACK,
    ACTIONS(1427), 1,
      anon_sym_LBRACE,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3947), 1,
      anon_sym_LPAREN,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(3965), 1,
      sym__uppercase_identifier,
    STATE(2777), 1,
      sym_attribute,
    STATE(3688), 1,
      sym__constructor_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5288), 1,
      sym__simple_binding_pattern,
    STATE(7331), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5312), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(5310), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [82302] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2778), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3969), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3967), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [82361] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2779), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3973), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3971), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [82420] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2780), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3720), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3718), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [82479] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2781), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3724), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3722), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [82538] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2782), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3832), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3830), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [82597] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2783), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2105), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2103), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [82656] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2784), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2109), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2107), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [82715] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2785), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2119), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2117), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [82774] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2786), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3977), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3975), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [82833] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2787), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2131), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2129), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [82892] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2788), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2135), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2133), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [82951] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2789), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3981), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3979), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [83010] = 27,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2590), 1,
      anon_sym_POUND,
    ACTIONS(2596), 1,
      anon_sym_begin,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2604), 1,
      aux_sym_number_token1,
    ACTIONS(2606), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2608), 1,
      anon_sym_SQUOTE,
    ACTIONS(2610), 1,
      anon_sym_DQUOTE,
    ACTIONS(2618), 1,
      sym__lowercase_identifier,
    ACTIONS(2622), 1,
      anon_sym_LBRACK,
    ACTIONS(2624), 1,
      anon_sym_LBRACE,
    ACTIONS(2628), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3957), 1,
      sym__uppercase_identifier,
    ACTIONS(3959), 1,
      aux_sym_tag_token1,
    ACTIONS(3961), 1,
      anon_sym_LPAREN,
    STATE(2790), 1,
      sym_attribute,
    STATE(5617), 1,
      sym_parenthesized_operator,
    STATE(5631), 1,
      sym__constructor_name,
    STATE(5678), 1,
      sym__simple_binding_pattern,
    STATE(7416), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2584), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5451), 2,
      sym__constant,
      sym_signed_number,
    STATE(5427), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5561), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(5557), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [83113] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2791), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3985), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3983), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [83172] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2792), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3989), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3987), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [83231] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2793), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2151), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2149), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [83290] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3991), 1,
      anon_sym_else,
    STATE(2681), 1,
      sym_else_clause,
    STATE(2794), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [83353] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2795), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3995), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3993), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [83412] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2796), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3999), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3997), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [83471] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2797), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2155), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2153), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [83530] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2798), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4003), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4001), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [83589] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3808), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3806), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [83648] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3816), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3814), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [83707] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2801), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3889), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3887), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [83766] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2578), 1,
      sym__lowercase_identifier,
    ACTIONS(2582), 1,
      anon_sym_LBRACK,
    ACTIONS(2586), 1,
      anon_sym_LBRACE,
    ACTIONS(2590), 1,
      anon_sym_POUND,
    ACTIONS(2592), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2596), 1,
      anon_sym_begin,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2604), 1,
      aux_sym_number_token1,
    ACTIONS(2606), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2608), 1,
      anon_sym_SQUOTE,
    ACTIONS(2610), 1,
      anon_sym_DQUOTE,
    ACTIONS(3955), 1,
      anon_sym_LPAREN,
    ACTIONS(3957), 1,
      sym__uppercase_identifier,
    ACTIONS(3959), 1,
      aux_sym_tag_token1,
    STATE(2802), 1,
      sym_attribute,
    STATE(5631), 1,
      sym__constructor_name,
    STATE(5735), 1,
      sym__simple_pattern,
    STATE(7457), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2584), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5451), 2,
      sym__constant,
      sym_signed_number,
    STATE(5427), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5636), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(5633), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [83867] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2803), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3820), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3818), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [83926] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2804), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2163), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2161), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [83985] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2805), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2171), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2169), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [84044] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2806), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2175), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2173), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [84103] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2807), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3824), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3822), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [84162] = 27,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2808), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(3755), 1,
      sym__simple_binding_pattern,
    STATE(7189), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3678), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [84265] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2809), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3828), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3826), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [84324] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2810), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3428), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3430), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [84383] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2811), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3836), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3834), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [84442] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2812), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3432), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3434), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [84501] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3356), 1,
      anon_sym_COMMA,
    STATE(2813), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [84562] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2814), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4007), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4005), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [84621] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2815), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4011), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4009), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [84680] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2816), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4015), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4013), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [84739] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2817), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3901), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3899), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [84798] = 27,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1757), 1,
      sym__lowercase_identifier,
    ACTIONS(1761), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1771), 1,
      anon_sym_LBRACE,
    ACTIONS(1781), 1,
      anon_sym_POUND,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1787), 1,
      anon_sym_begin,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(1809), 1,
      sym__uppercase_identifier,
    ACTIONS(1811), 1,
      aux_sym_tag_token1,
    ACTIONS(2325), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2327), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2818), 1,
      sym_attribute,
    STATE(3429), 1,
      sym__constructor_name,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(3730), 1,
      sym__simple_binding_pattern,
    STATE(7189), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3678), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [84901] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2819), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3921), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3919), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [84960] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3991), 1,
      anon_sym_else,
    STATE(2640), 1,
      sym_else_clause,
    STATE(2820), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [85023] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2821), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3885), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3883), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [85082] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2197), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2195), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [85141] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2823), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2201), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2199), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [85200] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2824), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3905), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3903), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [85259] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2825), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3690), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3688), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [85318] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2826), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2205), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2203), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [85377] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3658), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3656), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [85436] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3794), 1,
      sym__uppercase_identifier,
    ACTIONS(3838), 1,
      sym__lowercase_identifier,
    ACTIONS(3840), 1,
      anon_sym_LPAREN,
    ACTIONS(3842), 1,
      anon_sym_private,
    ACTIONS(3844), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3846), 1,
      anon_sym_PIPE,
    ACTIONS(3848), 1,
      anon_sym_LBRACK,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    STATE(2828), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4200), 1,
      sym__type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(4646), 1,
      sym_constructor_declaration,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3778), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4414), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [85549] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4017), 1,
      anon_sym_PIPE,
    STATE(2829), 1,
      sym_attribute,
    STATE(2831), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [85612] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4017), 1,
      anon_sym_PIPE,
    STATE(2830), 1,
      sym_attribute,
    STATE(2831), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3874), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3872), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [85675] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4019), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2831), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3704), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [85736] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3670), 1,
      anon_sym_COMMA,
    ACTIONS(3672), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3674), 1,
      sym_concat_operator,
    ACTIONS(3676), 1,
      sym_rel_operator,
    ACTIONS(3678), 1,
      sym_assign_operator,
    ACTIONS(3963), 1,
      anon_sym_SEMI,
    STATE(763), 1,
      sym_pow_operator,
    STATE(765), 1,
      sym_mult_operator,
    STATE(767), 1,
      sym_add_operator,
    STATE(768), 1,
      sym_and_operator,
    STATE(770), 1,
      sym_or_operator,
    STATE(2832), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3296), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(3338), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3298), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3336), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [85831] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4017), 1,
      anon_sym_PIPE,
    STATE(2829), 1,
      aux_sym__match_cases_repeat1,
    STATE(2833), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3881), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3879), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [85894] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4017), 1,
      anon_sym_PIPE,
    STATE(2830), 1,
      aux_sym__match_cases_repeat1,
    STATE(2834), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [85957] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2835), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3152), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3150), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [86016] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2836), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(917), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(915), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [86075] = 27,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(1415), 1,
      sym__lowercase_identifier,
    ACTIONS(1425), 1,
      anon_sym_LBRACK,
    ACTIONS(1427), 1,
      anon_sym_LBRACE,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3947), 1,
      anon_sym_LPAREN,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(3965), 1,
      sym__uppercase_identifier,
    STATE(2837), 1,
      sym_attribute,
    STATE(3688), 1,
      sym__constructor_name,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5315), 1,
      sym__simple_binding_pattern,
    STATE(7331), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5312), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(5310), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [86178] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2838), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3758), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3756), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [86237] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3732), 1,
      anon_sym_LPAREN,
    ACTIONS(3736), 1,
      anon_sym_private,
    ACTIONS(3738), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3740), 1,
      anon_sym_PIPE,
    ACTIONS(3742), 1,
      anon_sym_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      sym__uppercase_identifier,
    STATE(2839), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(5244), 1,
      sym__type,
    STATE(5629), 1,
      sym_constructor_declaration,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3744), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(5394), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [86350] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3388), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3390), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [86409] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2841), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2167), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2165), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [86468] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2842), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [86527] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2329), 1,
      sym__lowercase_identifier,
    ACTIONS(2343), 1,
      anon_sym_LBRACK,
    ACTIONS(2347), 1,
      anon_sym_LBRACE,
    ACTIONS(2349), 1,
      anon_sym_POUND,
    ACTIONS(2351), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2353), 1,
      anon_sym_begin,
    ACTIONS(2361), 1,
      aux_sym_number_token1,
    ACTIONS(2363), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2365), 1,
      anon_sym_SQUOTE,
    ACTIONS(2367), 1,
      anon_sym_DQUOTE,
    ACTIONS(2369), 1,
      sym__uppercase_identifier,
    ACTIONS(2371), 1,
      aux_sym_tag_token1,
    ACTIONS(4022), 1,
      anon_sym_LPAREN,
    ACTIONS(4024), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4026), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2843), 1,
      sym_attribute,
    STATE(3025), 1,
      sym__simple_pattern,
    STATE(3053), 1,
      sym__constructor_name,
    STATE(7756), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2345), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3022), 2,
      sym__constant,
      sym_signed_number,
    STATE(2928), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(3036), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2927), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [86628] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2844), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2185), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2183), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [86687] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2845), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2189), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2187), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [86746] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1405), 1,
      sym__uppercase_identifier,
    ACTIONS(1407), 1,
      aux_sym_tag_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4028), 1,
      anon_sym_LPAREN,
    STATE(2610), 1,
      sym__constructor_name,
    STATE(2792), 1,
      sym__simple_pattern,
    STATE(2846), 1,
      sym_attribute,
    STATE(8318), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [86847] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2847), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2193), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2191), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [86906] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2848), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4032), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4030), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [86965] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4036), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4034), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [87024] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2850), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4040), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4038), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [87083] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2851), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3704), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [87142] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2852), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4044), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4042), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [87201] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3794), 1,
      sym__uppercase_identifier,
    ACTIONS(3838), 1,
      sym__lowercase_identifier,
    ACTIONS(3840), 1,
      anon_sym_LPAREN,
    ACTIONS(3846), 1,
      anon_sym_PIPE,
    ACTIONS(3848), 1,
      anon_sym_LBRACK,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3907), 1,
      anon_sym_private,
    ACTIONS(3909), 1,
      anon_sym_DOT_DOT,
    STATE(2853), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4172), 1,
      sym__type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(4646), 1,
      sym_constructor_declaration,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3778), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4421), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [87314] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1817), 1,
      sym__lowercase_identifier,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1839), 1,
      anon_sym_LBRACE,
    ACTIONS(1845), 1,
      anon_sym_POUND,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1851), 1,
      anon_sym_begin,
    ACTIONS(1859), 1,
      aux_sym_number_token1,
    ACTIONS(1861), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1863), 1,
      anon_sym_SQUOTE,
    ACTIONS(1865), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__uppercase_identifier,
    ACTIONS(2532), 1,
      aux_sym_tag_token1,
    ACTIONS(2766), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2768), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4046), 1,
      anon_sym_LPAREN,
    STATE(2792), 1,
      sym__simple_pattern,
    STATE(2854), 1,
      sym_attribute,
    STATE(3191), 1,
      sym__constructor_name,
    STATE(7381), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1837), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2840), 2,
      sym__constant,
      sym_signed_number,
    STATE(2785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2866), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [87415] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2855), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4050), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4048), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [87474] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1357), 1,
      sym__lowercase_identifier,
    ACTIONS(1369), 1,
      anon_sym_LBRACK,
    ACTIONS(1373), 1,
      anon_sym_LBRACE,
    ACTIONS(1383), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(3953), 1,
      anon_sym_LPAREN,
    ACTIONS(3965), 1,
      sym__uppercase_identifier,
    STATE(2856), 1,
      sym_attribute,
    STATE(3688), 1,
      sym__constructor_name,
    STATE(5151), 1,
      sym__simple_pattern,
    STATE(7335), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5133), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(5132), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [87575] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2857), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3708), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3706), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [87634] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1357), 1,
      sym__lowercase_identifier,
    ACTIONS(1369), 1,
      anon_sym_LBRACK,
    ACTIONS(1373), 1,
      anon_sym_LBRACE,
    ACTIONS(1383), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3949), 1,
      sym__uppercase_identifier,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(3953), 1,
      anon_sym_LPAREN,
    STATE(2858), 1,
      sym_attribute,
    STATE(5152), 1,
      sym__simple_pattern,
    STATE(5180), 1,
      sym__constructor_name,
    STATE(7420), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5133), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(5132), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [87735] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2859), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3712), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3710), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [87794] = 27,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1387), 1,
      anon_sym_begin,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(1425), 1,
      anon_sym_LBRACK,
    ACTIONS(1427), 1,
      anon_sym_LBRACE,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2441), 1,
      anon_sym_POUND,
    ACTIONS(2556), 1,
      sym__lowercase_identifier,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3947), 1,
      anon_sym_LPAREN,
    ACTIONS(3949), 1,
      sym__uppercase_identifier,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    STATE(2860), 1,
      sym_attribute,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5180), 1,
      sym__constructor_name,
    STATE(5315), 1,
      sym__simple_binding_pattern,
    STATE(7377), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1371), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(5312), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(5310), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [87897] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2861), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4054), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4052), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [87956] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4058), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4056), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [88015] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3518), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3520), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [88074] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2864), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4062), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4060), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [88133] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2865), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3596), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3598), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [88192] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2866), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2433), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2439), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [88251] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2867), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4066), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4064), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [88310] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4070), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4068), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [88369] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2869), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4074), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4072), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [88428] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2870), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4078), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4076), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [88487] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2871), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4082), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4080), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [88546] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4086), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4084), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [88605] = 27,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1643), 1,
      sym__lowercase_identifier,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1655), 1,
      anon_sym_LBRACE,
    ACTIONS(1659), 1,
      anon_sym_POUND,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1665), 1,
      anon_sym_begin,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(1681), 1,
      sym__uppercase_identifier,
    ACTIONS(1683), 1,
      aux_sym_tag_token1,
    ACTIONS(1685), 1,
      anon_sym_LPAREN,
    ACTIONS(2860), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2862), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2088), 1,
      sym__constructor_name,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(2406), 1,
      sym__simple_binding_pattern,
    STATE(2873), 1,
      sym_attribute,
    STATE(7246), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2336), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [88708] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2874), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2033), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2031), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [88767] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(3764), 1,
      sym__lowercase_identifier,
    ACTIONS(3766), 1,
      anon_sym_LPAREN,
    ACTIONS(3774), 1,
      anon_sym_PIPE,
    ACTIONS(3776), 1,
      anon_sym_LBRACK,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3794), 1,
      sym__uppercase_identifier,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4088), 1,
      anon_sym_DOT_DOT,
    STATE(2875), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(3966), 1,
      sym__type,
    STATE(4197), 1,
      sym_constructor_declaration,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3778), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4098), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [88877] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3732), 1,
      anon_sym_LPAREN,
    ACTIONS(3740), 1,
      anon_sym_PIPE,
    ACTIONS(3742), 1,
      anon_sym_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      sym__uppercase_identifier,
    ACTIONS(4090), 1,
      anon_sym_DOT_DOT,
    STATE(2876), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(5275), 1,
      sym__type,
    STATE(5629), 1,
      sym_constructor_declaration,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3744), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(5372), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [88987] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4092), 1,
      anon_sym_else,
    STATE(2681), 1,
      sym_else_clause,
    STATE(2877), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 29,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [89049] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4094), 1,
      anon_sym_PIPE,
    STATE(2878), 1,
      sym_attribute,
    STATE(2881), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [89111] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4094), 1,
      anon_sym_PIPE,
    STATE(2879), 1,
      sym_attribute,
    STATE(2886), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3881), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3879), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [89173] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3732), 1,
      anon_sym_LPAREN,
    ACTIONS(3740), 1,
      anon_sym_PIPE,
    ACTIONS(3742), 1,
      anon_sym_LBRACK,
    ACTIONS(3746), 1,
      anon_sym_LBRACE,
    ACTIONS(3748), 1,
      sym__uppercase_identifier,
    ACTIONS(4096), 1,
      anon_sym_DOT_DOT,
    STATE(2880), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(5323), 1,
      sym__type,
    STATE(5629), 1,
      sym_constructor_declaration,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3744), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(5391), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [89283] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4094), 1,
      anon_sym_PIPE,
    STATE(2881), 1,
      sym_attribute,
    STATE(2882), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3874), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3872), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [89345] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4098), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2882), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3704), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [89405] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(3764), 1,
      sym__lowercase_identifier,
    ACTIONS(3766), 1,
      anon_sym_LPAREN,
    ACTIONS(3774), 1,
      anon_sym_PIPE,
    ACTIONS(3776), 1,
      anon_sym_LBRACK,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3794), 1,
      sym__uppercase_identifier,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4101), 1,
      anon_sym_DOT_DOT,
    STATE(2883), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(3969), 1,
      sym__type,
    STATE(4197), 1,
      sym_constructor_declaration,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3778), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3973), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [89515] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3398), 1,
      anon_sym_COMMA,
    STATE(2884), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [89575] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3794), 1,
      sym__uppercase_identifier,
    ACTIONS(3838), 1,
      sym__lowercase_identifier,
    ACTIONS(3840), 1,
      anon_sym_LPAREN,
    ACTIONS(3846), 1,
      anon_sym_PIPE,
    ACTIONS(3848), 1,
      anon_sym_LBRACK,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4103), 1,
      anon_sym_DOT_DOT,
    STATE(2885), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4282), 1,
      sym__type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(4646), 1,
      sym_constructor_declaration,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3778), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4426), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [89685] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4094), 1,
      anon_sym_PIPE,
    STATE(2882), 1,
      aux_sym__match_cases_repeat1,
    STATE(2886), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [89747] = 31,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3794), 1,
      sym__uppercase_identifier,
    ACTIONS(3838), 1,
      sym__lowercase_identifier,
    ACTIONS(3840), 1,
      anon_sym_LPAREN,
    ACTIONS(3846), 1,
      anon_sym_PIPE,
    ACTIONS(3848), 1,
      anon_sym_LBRACK,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4105), 1,
      anon_sym_DOT_DOT,
    STATE(2887), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4273), 1,
      sym__type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(4646), 1,
      sym_constructor_declaration,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3778), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4419), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [89857] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4092), 1,
      anon_sym_else,
    STATE(2640), 1,
      sym_else_clause,
    STATE(2888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 29,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [89919] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2889), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2065), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2063), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [89976] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2890), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2189), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2187), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [90033] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2891), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3812), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3810), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [90090] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2892), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3913), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3911), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [90147] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2893), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2193), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2191), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [90204] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3804), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3802), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [90261] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4107), 1,
      anon_sym_COLON,
    STATE(2895), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4050), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4048), 22,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [90320] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4109), 1,
      anon_sym_PIPE,
    STATE(2896), 1,
      sym_attribute,
    STATE(2904), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3881), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3879), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [90381] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2897), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2001), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1999), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [90438] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4109), 1,
      anon_sym_PIPE,
    STATE(2898), 1,
      sym_attribute,
    STATE(2917), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3874), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3872), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [90499] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3893), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3891), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [90556] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2900), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3897), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3895), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [90613] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2901), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2185), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2183), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [90670] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2902), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2077), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2075), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [90727] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4109), 1,
      anon_sym_PIPE,
    STATE(2898), 1,
      aux_sym__match_cases_repeat1,
    STATE(2903), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [90788] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4109), 1,
      anon_sym_PIPE,
    STATE(2904), 1,
      sym_attribute,
    STATE(2917), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [90849] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3638), 1,
      anon_sym_COMMA,
    STATE(2905), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 29,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [90908] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2906), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2167), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2165), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [90965] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2907), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1925), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1923), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [91022] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2908), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2049), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2047), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [91079] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2909), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2061), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2059), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [91136] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(2910), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2111), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [91195] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2069), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2067), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [91252] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2912), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2089), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2087), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [91309] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2913), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2093), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2091), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [91366] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2914), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2097), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2095), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [91423] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2915), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2101), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2099), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [91480] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(2916), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2111), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [91539] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4111), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2917), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3704), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [91598] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4114), 1,
      anon_sym_else,
    STATE(2681), 1,
      sym_else_clause,
    STATE(2918), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 28,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [91659] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4114), 1,
      anon_sym_else,
    STATE(2640), 1,
      sym_else_clause,
    STATE(2919), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 28,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [91720] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2920), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2005), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2003), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [91777] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2921), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1953), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1951), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [91834] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2922), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2151), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2149), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [91890] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2923), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2197), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2195), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [91946] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2924), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2081), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2079), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [92002] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2143), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2141), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [92058] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2926), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4050), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4048), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [92114] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2433), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2439), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [92170] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2928), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4044), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4042), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [92226] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2929), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2155), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2153), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [92282] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2930), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1953), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1951), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [92338] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2931), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1929), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1927), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [92394] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2932), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2201), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2199), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [92450] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2933), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2139), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2137), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [92506] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2934), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2167), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2165), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [92562] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2935), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2205), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2203), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [92618] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2936), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2033), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2031), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [92674] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2937), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2171), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2169), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [92730] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2938), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2175), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2173), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [92786] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2939), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2131), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2129), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [92842] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1937), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1935), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [92898] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2941), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4011), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4009), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [92954] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4015), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4013), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [93010] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2943), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2139), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2137), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [93066] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1941), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1939), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [93122] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2945), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1945), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1943), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [93178] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2946), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1957), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1955), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [93234] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2163), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2161), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [93290] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2948), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1981), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1979), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [93346] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2949), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4040), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4038), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [93402] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2950), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4078), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4076), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [93458] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3969), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3967), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [93514] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2952), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3973), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3971), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [93570] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3977), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3975), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [93626] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2954), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2001), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1999), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [93682] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2955), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2005), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2003), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [93738] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1961), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1959), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [93794] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2957), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2185), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2183), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [93850] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2189), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2187), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [93906] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2959), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4058), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4056), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [93962] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3995), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3993), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [94018] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2961), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4003), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4001), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [94074] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2962), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4070), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4068), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [94130] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1925), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1923), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [94186] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2964), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1965), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1963), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [94242] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2193), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2191), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [94298] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2966), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2163), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2161), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [94354] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2105), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2103), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [94410] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2968), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2109), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2107), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [94466] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(2969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2111), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [94524] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2970), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4036), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4034), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [94580] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2119), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2117), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [94636] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4062), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4060), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [94692] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2973), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2127), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2125), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [94748] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2974), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4086), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4084), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [94804] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2975), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2147), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2145), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [94860] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2976), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2061), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2059), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [94916] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2065), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2063), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [94972] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2978), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2069), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2067), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95028] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2979), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2171), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2169), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95084] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(2980), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2111), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95142] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2981), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2181), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2179), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95198] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2982), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2089), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2087), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95254] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2983), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2093), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2091), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95310] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2097), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2095), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95366] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2985), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2135), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2133), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95422] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2986), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2155), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2153), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95478] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2987), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2175), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2173), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95534] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2988), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3664), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3662), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [95590] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2101), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2099), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95646] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2990), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1897), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1895), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95702] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2991), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2167), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2165), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95758] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2992), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1969), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1967), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95814] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4116), 1,
      anon_sym_DOT,
    STATE(2993), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1973), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1971), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95872] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2994), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2185), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2183), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95928] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2995), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2189), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2187), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [95984] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2996), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2193), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2191), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [96040] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2201), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2199), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [96096] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2998), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4066), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4064), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [96152] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1977), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1975), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [96208] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3000), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2181), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2179), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [96264] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1985), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1983), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [96320] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3002), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1989), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1987), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [96376] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1993), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1991), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [96432] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3004), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2009), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2007), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [96488] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1889), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1887), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [96544] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3006), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2013), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2011), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [96600] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2017), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2015), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [96656] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3008), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4007), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4005), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [96712] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4116), 1,
      anon_sym_DOT,
    ACTIONS(4118), 1,
      anon_sym_POUND,
    ACTIONS(4120), 1,
      sym_hash_operator,
    STATE(3009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2041), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2039), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [96774] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2021), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2019), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [96830] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2025), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2023), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [96886] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3012), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2029), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2027), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [96942] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3013), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2037), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2035), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [96998] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1901), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1899), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [97054] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3015), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1905), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1903), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [97110] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3016), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3552), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3554), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [97166] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3017), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1909), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1907), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [97222] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2053), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2051), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [97278] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3019), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2057), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2055), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [97334] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3020), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1913), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1911), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [97390] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3021), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1917), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1915), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [97446] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3022), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3388), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3390), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [97502] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3023), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1921), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1919), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [97558] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3024), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3985), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3983), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [97614] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3025), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3989), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3987), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [97670] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3026), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3999), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3997), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [97726] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3027), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2151), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2149), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [97782] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3028), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2085), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2083), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [97838] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3029), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3470), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [97894] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3480), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3482), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [97950] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3031), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2205), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2203), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [98006] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3032), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2105), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2103), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [98062] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3033), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3394), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3396), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [98118] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3034), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2109), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2107), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [98174] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3035), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2033), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2031), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [98230] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3036), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2119), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2117), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [98286] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2209), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2207), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [98342] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2123), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2121), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [98398] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2159), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2157), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [98454] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3040), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3596), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3598), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [98510] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3041), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3428), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3430), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [98566] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3042), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3432), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3434), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [98622] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3043), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3518), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3520), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [98678] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3044), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1949), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1947), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [98734] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3045), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2049), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2047), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [98790] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3046), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4032), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4030), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [98846] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3047), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2197), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2195), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [98902] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3048), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2077), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2075), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [98958] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3049), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1997), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1995), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [99014] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3670), 1,
      anon_sym_COMMA,
    STATE(3050), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 28,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [99072] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2135), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2133), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [99128] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3052), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3981), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3979), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [99184] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3053), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2127), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2125), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [99240] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3054), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2131), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2129), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [99296] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3055), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2045), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2043), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [99352] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3056), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4054), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4052), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [99408] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3057), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2147), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2145), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [99464] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4116), 1,
      anon_sym_DOT,
    ACTIONS(4118), 1,
      anon_sym_POUND,
    ACTIONS(4120), 1,
      sym_hash_operator,
    STATE(3058), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1933), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1931), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [99526] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3059), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4074), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4072), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [99582] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4082), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4080), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [99638] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3061), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2073), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2071), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [99694] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3062), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1893), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1891), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [99750] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3063), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 9,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [99817] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4122), 1,
      anon_sym_RPAREN,
    ACTIONS(4124), 1,
      anon_sym_module,
    ACTIONS(4126), 1,
      anon_sym_COLON_COLON,
    STATE(3064), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6150), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(8076), 1,
      sym_polymorphic_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [99926] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4128), 1,
      anon_sym_COMMA,
    ACTIONS(4130), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4132), 1,
      sym_concat_operator,
    ACTIONS(4134), 1,
      sym_rel_operator,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3346), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3344), 13,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
  [100013] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4130), 1,
      anon_sym_COLON_COLON,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3066), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3350), 5,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3348), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [100088] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4130), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4132), 1,
      sym_concat_operator,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3067), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 5,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [100165] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4130), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4132), 1,
      sym_concat_operator,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3068), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 5,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [100242] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4128), 1,
      anon_sym_COMMA,
    ACTIONS(4130), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4132), 1,
      sym_concat_operator,
    ACTIONS(4134), 1,
      sym_rel_operator,
    ACTIONS(4136), 1,
      sym_assign_operator,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3069), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3286), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 12,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [100331] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4130), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4132), 1,
      sym_concat_operator,
    ACTIONS(4134), 1,
      sym_rel_operator,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3286), 3,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
  [100412] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4130), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4132), 1,
      sym_concat_operator,
    ACTIONS(4134), 1,
      sym_rel_operator,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3071), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3286), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 14,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
  [100497] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4128), 1,
      anon_sym_COMMA,
    ACTIONS(4130), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4132), 1,
      sym_concat_operator,
    ACTIONS(4134), 1,
      sym_rel_operator,
    ACTIONS(4136), 1,
      sym_assign_operator,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3072), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3322), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3320), 12,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [100586] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4128), 1,
      anon_sym_COMMA,
    ACTIONS(4130), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4132), 1,
      sym_concat_operator,
    ACTIONS(4134), 1,
      sym_rel_operator,
    ACTIONS(4136), 1,
      sym_assign_operator,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3073), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3332), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3330), 12,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [100675] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4128), 1,
      anon_sym_COMMA,
    ACTIONS(4130), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4132), 1,
      sym_concat_operator,
    ACTIONS(4134), 1,
      sym_rel_operator,
    ACTIONS(4136), 1,
      sym_assign_operator,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3342), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3340), 12,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [100764] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4128), 1,
      anon_sym_COMMA,
    ACTIONS(4130), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4132), 1,
      sym_concat_operator,
    ACTIONS(4134), 1,
      sym_rel_operator,
    ACTIONS(4136), 1,
      sym_assign_operator,
    ACTIONS(4138), 1,
      anon_sym_SEMI,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3075), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3154), 11,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [100855] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4128), 1,
      anon_sym_COMMA,
    ACTIONS(4130), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4132), 1,
      sym_concat_operator,
    ACTIONS(4134), 1,
      sym_rel_operator,
    ACTIONS(4136), 1,
      sym_assign_operator,
    ACTIONS(4138), 1,
      anon_sym_SEMI,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3076), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3158), 11,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [100946] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4128), 1,
      anon_sym_COMMA,
    ACTIONS(4130), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4132), 1,
      sym_concat_operator,
    ACTIONS(4134), 1,
      sym_rel_operator,
    ACTIONS(4136), 1,
      sym_assign_operator,
    ACTIONS(4138), 1,
      anon_sym_SEMI,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3077), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3338), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3336), 11,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [101037] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(713), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4140), 1,
      anon_sym_RPAREN,
    ACTIONS(4142), 1,
      anon_sym_module,
    STATE(3078), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6253), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(7808), 1,
      sym_polymorphic_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [101146] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3079), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 12,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3158), 28,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101201] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 12,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3154), 28,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101256] = 32,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4122), 1,
      anon_sym_RPAREN,
    ACTIONS(4126), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4144), 1,
      anon_sym_module,
    STATE(3081), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6102), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(7952), 1,
      sym_polymorphic_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [101365] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 9,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 26,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [101430] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3083), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 9,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [101497] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    STATE(819), 1,
      sym_pow_operator,
    STATE(820), 1,
      sym_mult_operator,
    STATE(821), 1,
      sym_add_operator,
    STATE(822), 1,
      sym_and_operator,
    STATE(823), 1,
      sym_or_operator,
    STATE(3084), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [101568] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3085), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3152), 12,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3150), 28,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101623] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4122), 1,
      anon_sym_RBRACK,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4148), 1,
      anon_sym_PIPE,
    STATE(3086), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(6907), 1,
      sym_tag_specification,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7809), 1,
      sym__tag_spec,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [101727] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3087), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 11,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 19,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [101793] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3088), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 10,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 15,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [101863] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4150), 1,
      anon_sym_COMMA,
    ACTIONS(4152), 1,
      anon_sym_SEMI,
    ACTIONS(4154), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4156), 1,
      sym_concat_operator,
    ACTIONS(4158), 1,
      sym_rel_operator,
    ACTIONS(4160), 1,
      sym_assign_operator,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3089), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3156), 3,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3154), 9,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [101953] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3090), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 11,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 23,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [102017] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3091), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 11,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 19,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [102083] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4150), 1,
      anon_sym_COMMA,
    ACTIONS(4154), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4156), 1,
      sym_concat_operator,
    ACTIONS(4158), 1,
      sym_rel_operator,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3092), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3346), 3,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3344), 11,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_assign_operator,
  [102169] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4154), 1,
      anon_sym_COLON_COLON,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3093), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3350), 7,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3348), 14,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [102243] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4154), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4156), 1,
      sym_concat_operator,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3094), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 7,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 13,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_or,
      sym_assign_operator,
  [102319] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4154), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4156), 1,
      sym_concat_operator,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3095), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 7,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 13,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_or,
      sym_assign_operator,
  [102395] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4150), 1,
      anon_sym_COMMA,
    ACTIONS(4154), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4156), 1,
      sym_concat_operator,
    ACTIONS(4158), 1,
      sym_rel_operator,
    ACTIONS(4160), 1,
      sym_assign_operator,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3286), 3,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [102483] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4154), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4156), 1,
      sym_concat_operator,
    ACTIONS(4158), 1,
      sym_rel_operator,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3097), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3286), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 13,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_or,
      sym_assign_operator,
  [102563] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4154), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4156), 1,
      sym_concat_operator,
    ACTIONS(4158), 1,
      sym_rel_operator,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3098), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3286), 3,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 12,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_assign_operator,
  [102647] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4150), 1,
      anon_sym_COMMA,
    ACTIONS(4154), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4156), 1,
      sym_concat_operator,
    ACTIONS(4158), 1,
      sym_rel_operator,
    ACTIONS(4160), 1,
      sym_assign_operator,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3099), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3322), 3,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3320), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [102735] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4150), 1,
      anon_sym_COMMA,
    ACTIONS(4154), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4156), 1,
      sym_concat_operator,
    ACTIONS(4158), 1,
      sym_rel_operator,
    ACTIONS(4160), 1,
      sym_assign_operator,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3332), 3,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3330), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [102823] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4162), 1,
      anon_sym_PIPE,
    ACTIONS(4164), 1,
      anon_sym_RBRACK,
    STATE(3101), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6514), 1,
      sym__tag_spec,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102927] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4150), 1,
      anon_sym_COMMA,
    ACTIONS(4154), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4156), 1,
      sym_concat_operator,
    ACTIONS(4158), 1,
      sym_rel_operator,
    ACTIONS(4160), 1,
      sym_assign_operator,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3102), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3342), 3,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3340), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [103015] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4166), 1,
      anon_sym_COLON,
    STATE(3103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2315), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2311), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [103071] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4122), 1,
      anon_sym_RBRACK,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4168), 1,
      anon_sym_PIPE,
    STATE(3104), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6767), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8002), 1,
      sym__tag_spec,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103175] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4140), 1,
      anon_sym_RBRACK,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4170), 1,
      anon_sym_PIPE,
    STATE(3105), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7164), 1,
      sym_tag_specification,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8272), 1,
      sym__tag_spec,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103279] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4150), 1,
      anon_sym_COMMA,
    ACTIONS(4152), 1,
      anon_sym_SEMI,
    ACTIONS(4154), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4156), 1,
      sym_concat_operator,
    ACTIONS(4158), 1,
      sym_rel_operator,
    ACTIONS(4160), 1,
      sym_assign_operator,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3106), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3338), 3,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3336), 9,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [103369] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4150), 1,
      anon_sym_COMMA,
    ACTIONS(4152), 1,
      anon_sym_SEMI,
    ACTIONS(4154), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4156), 1,
      sym_concat_operator,
    ACTIONS(4158), 1,
      sym_rel_operator,
    ACTIONS(4160), 1,
      sym_assign_operator,
    STATE(688), 1,
      sym_pow_operator,
    STATE(689), 1,
      sym_mult_operator,
    STATE(690), 1,
      sym_add_operator,
    STATE(691), 1,
      sym_and_operator,
    STATE(692), 1,
      sym_or_operator,
    STATE(3107), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3160), 3,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3158), 9,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [103459] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4172), 1,
      anon_sym_PIPE,
    ACTIONS(4174), 1,
      anon_sym_RBRACK,
    STATE(3108), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6749), 1,
      sym__tag_spec,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103563] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4176), 1,
      anon_sym_PIPE,
    ACTIONS(4178), 1,
      anon_sym_RBRACK,
    STATE(3109), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6399), 1,
      sym__tag_spec,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103667] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    STATE(3110), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(3938), 1,
      sym__polymorphic_type,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6743), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7161), 1,
      sym_polymorphic_type,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [103770] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3286), 4,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [103845] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3112), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3286), 4,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [103920] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3286), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    ACTIONS(4184), 1,
      anon_sym_COMMA,
    ACTIONS(4186), 1,
      sym_rel_operator,
    ACTIONS(4188), 1,
      sym_assign_operator,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3113), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 11,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [104007] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3114), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3350), 4,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3348), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [104080] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3115), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 21,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [104145] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 7,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [104214] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    ACTIONS(4186), 1,
      sym_rel_operator,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3117), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3286), 2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 14,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
  [104293] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3286), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    ACTIONS(4186), 1,
      sym_rel_operator,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3118), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
  [104376] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3322), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    ACTIONS(4184), 1,
      anon_sym_COMMA,
    ACTIONS(4186), 1,
      sym_rel_operator,
    ACTIONS(4188), 1,
      sym_assign_operator,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3119), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3320), 11,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [104463] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3332), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    ACTIONS(4184), 1,
      anon_sym_COMMA,
    ACTIONS(4186), 1,
      sym_rel_operator,
    ACTIONS(4188), 1,
      sym_assign_operator,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3120), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3330), 11,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [104550] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3342), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    ACTIONS(4184), 1,
      anon_sym_COMMA,
    ACTIONS(4186), 1,
      sym_rel_operator,
    ACTIONS(4188), 1,
      sym_assign_operator,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3121), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3340), 11,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [104637] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4196), 1,
      sym_rel_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3122), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3346), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3344), 11,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_assign_operator,
  [104722] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4198), 1,
      anon_sym_RBRACK,
    STATE(3123), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6405), 1,
      sym__tag_spec,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104823] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4200), 1,
      anon_sym_PIPE,
    STATE(3124), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6298), 1,
      sym__tag_spec,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104924] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(3125), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2111), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [104979] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3156), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    ACTIONS(4184), 1,
      anon_sym_COMMA,
    ACTIONS(4186), 1,
      sym_rel_operator,
    ACTIONS(4188), 1,
      sym_assign_operator,
    ACTIONS(4202), 1,
      anon_sym_SEMI,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3126), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3154), 10,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [105068] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4196), 1,
      sym_rel_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3127), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3342), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3340), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [105155] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3128), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 23,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [105218] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3160), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    ACTIONS(4184), 1,
      anon_sym_COMMA,
    ACTIONS(4186), 1,
      sym_rel_operator,
    ACTIONS(4188), 1,
      sym_assign_operator,
    ACTIONS(4202), 1,
      anon_sym_SEMI,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3129), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3158), 10,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [105307] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4124), 1,
      anon_sym_module,
    STATE(3130), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6150), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(8076), 1,
      sym_polymorphic_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [105410] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4148), 1,
      anon_sym_PIPE,
    STATE(3131), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(6907), 1,
      sym_tag_specification,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7809), 1,
      sym__tag_spec,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105511] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3764), 1,
      sym__lowercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4206), 1,
      anon_sym_LPAREN,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    STATE(3132), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3613), 1,
      sym_type_variable,
    STATE(3936), 1,
      sym_polymorphic_type,
    STATE(3938), 1,
      sym__polymorphic_type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(4516), 1,
      sym__type,
    STATE(6083), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7199), 1,
      sym__abstract_type,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [105614] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4210), 1,
      anon_sym_PIPE,
    STATE(3133), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6156), 1,
      sym__tag_spec,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105715] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4196), 1,
      sym_rel_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4212), 1,
      anon_sym_SEMI,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3134), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3154), 9,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [105804] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3135), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2417), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2415), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [105857] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4214), 1,
      anon_sym_RBRACK,
    STATE(3136), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6416), 1,
      sym__tag_spec,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105958] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4196), 1,
      sym_rel_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3137), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3332), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3330), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [106045] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3338), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    ACTIONS(4184), 1,
      anon_sym_COMMA,
    ACTIONS(4186), 1,
      sym_rel_operator,
    ACTIONS(4188), 1,
      sym_assign_operator,
    ACTIONS(4202), 1,
      anon_sym_SEMI,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3138), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3336), 10,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [106134] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4144), 1,
      anon_sym_module,
    STATE(3139), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6102), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(7952), 1,
      sym_polymorphic_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [106237] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4142), 1,
      anon_sym_module,
    STATE(3140), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6253), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7620), 1,
      sym_polymorphic_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [106340] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4168), 1,
      anon_sym_PIPE,
    STATE(3141), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6767), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8002), 1,
      sym__tag_spec,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106441] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4216), 1,
      anon_sym_PIPE,
    STATE(3142), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6048), 1,
      sym__tag_spec,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106542] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4218), 1,
      anon_sym_COLON,
    STATE(3143), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4050), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4048), 20,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [106597] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4196), 1,
      sym_rel_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4212), 1,
      anon_sym_SEMI,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3158), 9,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [106686] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4196), 1,
      sym_rel_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3145), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3286), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 12,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_assign_operator,
  [106769] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3146), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3350), 6,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3348), 14,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [106842] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 6,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 13,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_or,
      sym_assign_operator,
  [106917] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3148), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 6,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 13,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_or,
      sym_assign_operator,
  [106992] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4170), 1,
      anon_sym_PIPE,
    STATE(3149), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(5895), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7164), 1,
      sym_tag_specification,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8272), 1,
      sym__tag_spec,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [107093] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4142), 1,
      anon_sym_module,
    STATE(3150), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6253), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(8224), 1,
      sym_polymorphic_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [107196] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4196), 1,
      sym_rel_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3151), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3286), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [107283] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4220), 1,
      anon_sym_LPAREN,
    STATE(3152), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(3936), 1,
      sym_polymorphic_type,
    STATE(3938), 1,
      sym__polymorphic_type,
    STATE(4529), 1,
      sym__simple_type,
    STATE(4944), 1,
      sym__proper_tuple_type,
    STATE(4946), 1,
      sym__type,
    STATE(6255), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(8225), 1,
      sym_labeled_argument_type,
    STATE(8227), 1,
      sym_labeled_tuple_element_type,
    STATE(8230), 1,
      sym__abstract_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [107386] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3153), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 19,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [107451] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4142), 1,
      anon_sym_module,
    STATE(3154), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6253), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(8172), 1,
      sym_polymorphic_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [107554] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4170), 1,
      anon_sym_PIPE,
    STATE(3155), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(5940), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7164), 1,
      sym_tag_specification,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8272), 1,
      sym__tag_spec,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [107655] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4222), 1,
      anon_sym_RBRACK,
    STATE(3156), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6759), 1,
      sym__tag_spec,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [107756] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3838), 1,
      sym__lowercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4224), 1,
      anon_sym_LPAREN,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    STATE(3157), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(3936), 1,
      sym_polymorphic_type,
    STATE(3938), 1,
      sym__polymorphic_type,
    STATE(3963), 1,
      sym_type_variable,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(4898), 1,
      sym__type,
    STATE(6069), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(7228), 1,
      sym__abstract_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [107859] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3158), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 19,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [107924] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4196), 1,
      sym_rel_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3159), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3322), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3320), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [108011] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(3160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2111), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [108066] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4196), 1,
      sym_rel_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3286), 3,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 13,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_or,
      sym_assign_operator,
  [108145] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    STATE(3162), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(3936), 1,
      sym_polymorphic_type,
    STATE(3938), 1,
      sym__polymorphic_type,
    STATE(4716), 1,
      sym__simple_type,
    STATE(4946), 1,
      sym__type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [108248] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4142), 1,
      anon_sym_module,
    STATE(3163), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6253), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(8117), 1,
      sym_polymorphic_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [108351] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3164), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 9,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 15,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [108420] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4142), 1,
      anon_sym_module,
    STATE(3165), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6253), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    STATE(8323), 1,
      sym_polymorphic_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [108523] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4228), 1,
      anon_sym_LPAREN,
    STATE(3166), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(6054), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6163), 1,
      sym__type,
    STATE(6513), 1,
      sym__polymorphic_type,
    STATE(6518), 1,
      sym_polymorphic_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    STATE(7845), 1,
      sym__abstract_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [108626] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4170), 1,
      anon_sym_PIPE,
    STATE(3167), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(5955), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7164), 1,
      sym_tag_specification,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8272), 1,
      sym__tag_spec,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [108727] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4142), 1,
      anon_sym_module,
    STATE(3168), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6253), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(8013), 1,
      sym_polymorphic_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [108830] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4142), 1,
      anon_sym_module,
    STATE(3169), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6253), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7250), 1,
      sym_polymorphic_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [108933] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4196), 1,
      sym_rel_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4212), 1,
      anon_sym_SEMI,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3170), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3338), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3336), 9,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [109022] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4170), 1,
      anon_sym_PIPE,
    STATE(3171), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7164), 1,
      sym_tag_specification,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8272), 1,
      sym__tag_spec,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [109123] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4142), 1,
      anon_sym_module,
    STATE(3172), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6253), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(8302), 1,
      sym_polymorphic_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [109226] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3173), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1933), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1931), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [109279] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3174), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 25,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [109342] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3346), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    ACTIONS(4184), 1,
      anon_sym_COMMA,
    ACTIONS(4186), 1,
      sym_rel_operator,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3175), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3344), 12,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
  [109427] = 30,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4142), 1,
      anon_sym_module,
    STATE(3176), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6253), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(7808), 1,
      sym_polymorphic_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [109530] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3177), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 21,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [109595] = 28,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3178), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6459), 1,
      sym__tag_spec,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [109693] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(13), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(79), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(81), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(315), 1,
      aux_sym_directive_token1,
    ACTIONS(4230), 1,
      anon_sym_let,
    ACTIONS(4232), 1,
      anon_sym_external,
    ACTIONS(4234), 1,
      anon_sym_type,
    ACTIONS(4236), 1,
      anon_sym_exception,
    ACTIONS(4238), 1,
      anon_sym_module,
    ACTIONS(4240), 1,
      anon_sym_open,
    ACTIONS(4242), 1,
      anon_sym_include,
    ACTIONS(4244), 1,
      anon_sym_class,
    ACTIONS(4246), 1,
      anon_sym_val,
    ACTIONS(4248), 1,
      sym_let_operator,
    STATE(47), 1,
      aux_sym__structure_repeat1,
    STATE(3179), 1,
      sym_attribute,
    STATE(3200), 1,
      aux_sym__structure_repeat2,
    STATE(3205), 1,
      sym_directive,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(391), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4806), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4775), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4781), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [109785] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3338), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    ACTIONS(4184), 1,
      anon_sym_COMMA,
    ACTIONS(4186), 1,
      sym_rel_operator,
    ACTIONS(4188), 1,
      sym_assign_operator,
    ACTIONS(4250), 1,
      anon_sym_SEMI,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3180), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3336), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [109873] = 28,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3181), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6607), 1,
      sym__tag_spec,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [109971] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3182), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2181), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2179), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [110023] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(13), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(79), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(81), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(315), 1,
      aux_sym_directive_token1,
    ACTIONS(4230), 1,
      anon_sym_let,
    ACTIONS(4232), 1,
      anon_sym_external,
    ACTIONS(4234), 1,
      anon_sym_type,
    ACTIONS(4236), 1,
      anon_sym_exception,
    ACTIONS(4238), 1,
      anon_sym_module,
    ACTIONS(4240), 1,
      anon_sym_open,
    ACTIONS(4242), 1,
      anon_sym_include,
    ACTIONS(4244), 1,
      anon_sym_class,
    ACTIONS(4246), 1,
      anon_sym_val,
    ACTIONS(4248), 1,
      sym_let_operator,
    STATE(45), 1,
      aux_sym__structure_repeat1,
    STATE(3183), 1,
      sym_attribute,
    STATE(3205), 1,
      sym_directive,
    STATE(3212), 1,
      aux_sym__structure_repeat2,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(395), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4806), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4775), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4781), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [110115] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4252), 1,
      sym__lowercase_identifier,
    ACTIONS(4254), 1,
      anon_sym_let,
    ACTIONS(4256), 1,
      anon_sym_LPAREN,
    ACTIONS(4258), 1,
      anon_sym_LBRACK,
    ACTIONS(4260), 1,
      anon_sym_object,
    STATE(3184), 1,
      sym_attribute,
    STATE(4307), 1,
      sym__class_type,
    STATE(5514), 1,
      sym__simple_type,
    STATE(7106), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3952), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(4215), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7989), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4081), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(3757), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [110209] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(4262), 1,
      sym__lowercase_identifier,
    ACTIONS(4264), 1,
      anon_sym_let,
    ACTIONS(4266), 1,
      anon_sym_LPAREN,
    ACTIONS(4268), 1,
      anon_sym_LBRACK,
    STATE(3185), 1,
      sym_attribute,
    STATE(4281), 1,
      sym__class_type,
    STATE(5475), 1,
      sym__simple_type,
    STATE(7125), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4135), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(4215), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7865), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4081), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(3757), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [110303] = 29,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1365), 1,
      anon_sym_type,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    STATE(3186), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4716), 1,
      sym__simple_type,
    STATE(4962), 1,
      sym__type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6031), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(7651), 1,
      sym__abstract_type,
    STATE(7896), 1,
      sym_polymorphic_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [110403] = 28,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3187), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6496), 1,
      sym__tag_spec,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [110501] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4254), 1,
      anon_sym_let,
    ACTIONS(4258), 1,
      anon_sym_LBRACK,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(4262), 1,
      sym__lowercase_identifier,
    ACTIONS(4270), 1,
      anon_sym_LPAREN,
    STATE(3188), 1,
      sym_attribute,
    STATE(4307), 1,
      sym__class_type,
    STATE(5419), 1,
      sym__simple_type,
    STATE(7125), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4135), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(4215), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7637), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4081), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(3757), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [110595] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4262), 1,
      sym__lowercase_identifier,
    ACTIONS(4272), 1,
      anon_sym_let,
    ACTIONS(4274), 1,
      anon_sym_LPAREN,
    ACTIONS(4276), 1,
      anon_sym_LBRACK,
    ACTIONS(4278), 1,
      anon_sym_object,
    STATE(3189), 1,
      sym_attribute,
    STATE(5445), 1,
      sym__simple_type,
    STATE(6135), 1,
      sym__class_type,
    STATE(7125), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4135), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(6111), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7276), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(6113), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(3757), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [110689] = 24,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4280), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(4283), 1,
      anon_sym_let,
    ACTIONS(4286), 1,
      anon_sym_external,
    ACTIONS(4289), 1,
      anon_sym_type,
    ACTIONS(4294), 1,
      anon_sym_exception,
    ACTIONS(4297), 1,
      anon_sym_module,
    ACTIONS(4300), 1,
      anon_sym_open,
    ACTIONS(4303), 1,
      anon_sym_include,
    ACTIONS(4306), 1,
      anon_sym_class,
    ACTIONS(4309), 1,
      anon_sym_val,
    ACTIONS(4312), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4315), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(4318), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4321), 1,
      sym_let_operator,
    ACTIONS(4324), 1,
      aux_sym_directive_token1,
    STATE(55), 1,
      aux_sym__structure_repeat1,
    STATE(3205), 1,
      sym_directive,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4292), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(3190), 2,
      sym_attribute,
      aux_sym__structure_repeat2,
    STATE(4806), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4775), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4781), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [110779] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3191), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2127), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2125), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [110831] = 28,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3192), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6416), 1,
      sym__tag_spec,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [110929] = 28,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3193), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6229), 1,
      sym__tag_spec,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [111027] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4262), 1,
      sym__lowercase_identifier,
    ACTIONS(4272), 1,
      anon_sym_let,
    ACTIONS(4274), 1,
      anon_sym_LPAREN,
    ACTIONS(4276), 1,
      anon_sym_LBRACK,
    ACTIONS(4278), 1,
      anon_sym_object,
    STATE(3194), 1,
      sym_attribute,
    STATE(5445), 1,
      sym__simple_type,
    STATE(6112), 1,
      sym__class_type,
    STATE(7125), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4135), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(6111), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7276), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(6113), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(3757), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [111121] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4252), 1,
      sym__lowercase_identifier,
    ACTIONS(4254), 1,
      anon_sym_let,
    ACTIONS(4256), 1,
      anon_sym_LPAREN,
    ACTIONS(4258), 1,
      anon_sym_LBRACK,
    ACTIONS(4260), 1,
      anon_sym_object,
    STATE(3195), 1,
      sym_attribute,
    STATE(4281), 1,
      sym__class_type,
    STATE(5514), 1,
      sym__simple_type,
    STATE(7106), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3952), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(4215), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7989), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4081), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(3757), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [111215] = 28,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3196), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6624), 1,
      sym__tag_spec,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [111313] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4254), 1,
      anon_sym_let,
    ACTIONS(4258), 1,
      anon_sym_LBRACK,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(4262), 1,
      sym__lowercase_identifier,
    ACTIONS(4270), 1,
      anon_sym_LPAREN,
    STATE(3197), 1,
      sym_attribute,
    STATE(4222), 1,
      sym__class_type,
    STATE(5419), 1,
      sym__simple_type,
    STATE(7125), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4135), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(4215), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7637), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4081), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(3757), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [111407] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3198), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2139), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2137), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [111459] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4262), 1,
      sym__lowercase_identifier,
    ACTIONS(4272), 1,
      anon_sym_let,
    ACTIONS(4274), 1,
      anon_sym_LPAREN,
    ACTIONS(4276), 1,
      anon_sym_LBRACK,
    ACTIONS(4278), 1,
      anon_sym_object,
    STATE(3199), 1,
      sym_attribute,
    STATE(5445), 1,
      sym__simple_type,
    STATE(6183), 1,
      sym__class_type,
    STATE(7125), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4135), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(6111), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7276), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(6113), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(3757), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [111553] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(13), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(79), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(81), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(315), 1,
      aux_sym_directive_token1,
    ACTIONS(4230), 1,
      anon_sym_let,
    ACTIONS(4232), 1,
      anon_sym_external,
    ACTIONS(4234), 1,
      anon_sym_type,
    ACTIONS(4236), 1,
      anon_sym_exception,
    ACTIONS(4238), 1,
      anon_sym_module,
    ACTIONS(4240), 1,
      anon_sym_open,
    ACTIONS(4242), 1,
      anon_sym_include,
    ACTIONS(4244), 1,
      anon_sym_class,
    ACTIONS(4246), 1,
      anon_sym_val,
    ACTIONS(4248), 1,
      sym_let_operator,
    STATE(43), 1,
      aux_sym__structure_repeat1,
    STATE(3190), 1,
      aux_sym__structure_repeat2,
    STATE(3200), 1,
      sym_attribute,
    STATE(3205), 1,
      sym_directive,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(399), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4806), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4775), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4781), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [111645] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3201), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2147), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2145), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [111697] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4254), 1,
      anon_sym_let,
    ACTIONS(4258), 1,
      anon_sym_LBRACK,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(4262), 1,
      sym__lowercase_identifier,
    ACTIONS(4270), 1,
      anon_sym_LPAREN,
    STATE(3202), 1,
      sym_attribute,
    STATE(4281), 1,
      sym__class_type,
    STATE(5419), 1,
      sym__simple_type,
    STATE(7125), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4135), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(4215), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7637), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4081), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(3757), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [111791] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3156), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    ACTIONS(4184), 1,
      anon_sym_COMMA,
    ACTIONS(4186), 1,
      sym_rel_operator,
    ACTIONS(4188), 1,
      sym_assign_operator,
    ACTIONS(4250), 1,
      anon_sym_SEMI,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3203), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3154), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [111879] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(4262), 1,
      sym__lowercase_identifier,
    ACTIONS(4264), 1,
      anon_sym_let,
    ACTIONS(4266), 1,
      anon_sym_LPAREN,
    ACTIONS(4268), 1,
      anon_sym_LBRACK,
    STATE(3204), 1,
      sym_attribute,
    STATE(4222), 1,
      sym__class_type,
    STATE(5475), 1,
      sym__simple_type,
    STATE(7125), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4135), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(4215), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7865), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4081), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(3757), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [111973] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1415), 1,
      sym__lowercase_identifier,
    ACTIONS(4331), 1,
      anon_sym_LPAREN,
    ACTIONS(4335), 1,
      anon_sym_LBRACE,
    ACTIONS(4337), 1,
      aux_sym_number_token1,
    ACTIONS(4339), 1,
      anon_sym_SQUOTE,
    ACTIONS(4341), 1,
      anon_sym_DQUOTE,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    STATE(3205), 1,
      sym_attribute,
    STATE(4700), 1,
      sym_module_path,
    STATE(4786), 1,
      sym__value_name,
    STATE(4802), 1,
      sym_parenthesized_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4333), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4773), 2,
      sym__constant,
      sym_value_path,
    STATE(4785), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    ACTIONS(4327), 8,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
    ACTIONS(4329), 10,
      anon_sym_let,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
  [112051] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3160), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4180), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4182), 1,
      sym_concat_operator,
    ACTIONS(4184), 1,
      anon_sym_COMMA,
    ACTIONS(4186), 1,
      sym_rel_operator,
    ACTIONS(4188), 1,
      sym_assign_operator,
    ACTIONS(4250), 1,
      anon_sym_SEMI,
    STATE(847), 1,
      sym_pow_operator,
    STATE(848), 1,
      sym_mult_operator,
    STATE(849), 1,
      sym_add_operator,
    STATE(850), 1,
      sym_and_operator,
    STATE(851), 1,
      sym_or_operator,
    STATE(3206), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3158), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [112139] = 28,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3207), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6759), 1,
      sym__tag_spec,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [112237] = 28,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3208), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6023), 1,
      sym__tag_spec,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [112335] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4252), 1,
      sym__lowercase_identifier,
    ACTIONS(4254), 1,
      anon_sym_let,
    ACTIONS(4256), 1,
      anon_sym_LPAREN,
    ACTIONS(4258), 1,
      anon_sym_LBRACK,
    ACTIONS(4260), 1,
      anon_sym_object,
    STATE(3209), 1,
      sym_attribute,
    STATE(4222), 1,
      sym__class_type,
    STATE(5514), 1,
      sym__simple_type,
    STATE(7106), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3952), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(4215), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7989), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4081), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(3757), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [112429] = 28,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3210), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6405), 1,
      sym__tag_spec,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [112527] = 26,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(4262), 1,
      sym__lowercase_identifier,
    ACTIONS(4264), 1,
      anon_sym_let,
    ACTIONS(4266), 1,
      anon_sym_LPAREN,
    ACTIONS(4268), 1,
      anon_sym_LBRACK,
    STATE(3211), 1,
      sym_attribute,
    STATE(4307), 1,
      sym__class_type,
    STATE(5475), 1,
      sym__simple_type,
    STATE(7125), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4135), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(4215), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7865), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4081), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(3757), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [112621] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(13), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(79), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(81), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(315), 1,
      aux_sym_directive_token1,
    ACTIONS(4230), 1,
      anon_sym_let,
    ACTIONS(4232), 1,
      anon_sym_external,
    ACTIONS(4234), 1,
      anon_sym_type,
    ACTIONS(4236), 1,
      anon_sym_exception,
    ACTIONS(4238), 1,
      anon_sym_module,
    ACTIONS(4240), 1,
      anon_sym_open,
    ACTIONS(4242), 1,
      anon_sym_include,
    ACTIONS(4244), 1,
      anon_sym_class,
    ACTIONS(4246), 1,
      anon_sym_val,
    ACTIONS(4248), 1,
      sym_let_operator,
    STATE(47), 1,
      aux_sym__structure_repeat1,
    STATE(3190), 1,
      aux_sym__structure_repeat2,
    STATE(3205), 1,
      sym_directive,
    STATE(3212), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(391), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4806), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4775), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4781), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [112713] = 28,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3951), 1,
      aux_sym_tag_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3213), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6039), 1,
      sym__tag_spec,
    STATE(6268), 1,
      sym__type,
    STATE(6276), 1,
      sym_tag,
    STATE(6750), 1,
      sym_tag_specification,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [112811] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4345), 1,
      anon_sym_PIPE,
    STATE(3214), 1,
      sym_attribute,
    STATE(3226), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3874), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3872), 26,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [112866] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(13), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(79), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(81), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(315), 1,
      aux_sym_directive_token1,
    ACTIONS(391), 1,
      ts_builtin_sym_end,
    ACTIONS(4232), 1,
      anon_sym_external,
    ACTIONS(4234), 1,
      anon_sym_type,
    ACTIONS(4236), 1,
      anon_sym_exception,
    ACTIONS(4238), 1,
      anon_sym_module,
    ACTIONS(4240), 1,
      anon_sym_open,
    ACTIONS(4242), 1,
      anon_sym_include,
    ACTIONS(4246), 1,
      anon_sym_val,
    ACTIONS(4347), 1,
      anon_sym_let,
    ACTIONS(4349), 1,
      anon_sym_class,
    ACTIONS(4351), 1,
      sym_let_operator,
    STATE(50), 1,
      aux_sym__structure_repeat1,
    STATE(3205), 1,
      sym_directive,
    STATE(3215), 1,
      sym_attribute,
    STATE(3219), 1,
      aux_sym__structure_repeat2,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4806), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4775), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4781), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [112957] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4353), 1,
      anon_sym_else,
    STATE(2803), 1,
      sym_else_clause,
    STATE(3216), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 9,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 25,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [113012] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4355), 1,
      anon_sym_COLON,
    STATE(3217), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4050), 16,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4048), 19,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [113065] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(13), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(79), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(81), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(315), 1,
      aux_sym_directive_token1,
    ACTIONS(391), 1,
      ts_builtin_sym_end,
    ACTIONS(4232), 1,
      anon_sym_external,
    ACTIONS(4234), 1,
      anon_sym_type,
    ACTIONS(4236), 1,
      anon_sym_exception,
    ACTIONS(4238), 1,
      anon_sym_module,
    ACTIONS(4240), 1,
      anon_sym_open,
    ACTIONS(4242), 1,
      anon_sym_include,
    ACTIONS(4246), 1,
      anon_sym_val,
    ACTIONS(4347), 1,
      anon_sym_let,
    ACTIONS(4349), 1,
      anon_sym_class,
    ACTIONS(4351), 1,
      sym_let_operator,
    STATE(50), 1,
      aux_sym__structure_repeat1,
    STATE(3205), 1,
      sym_directive,
    STATE(3218), 1,
      sym_attribute,
    STATE(3228), 1,
      aux_sym__structure_repeat2,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4806), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4775), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4781), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [113156] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(13), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(79), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(81), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(315), 1,
      aux_sym_directive_token1,
    ACTIONS(399), 1,
      ts_builtin_sym_end,
    ACTIONS(4232), 1,
      anon_sym_external,
    ACTIONS(4234), 1,
      anon_sym_type,
    ACTIONS(4236), 1,
      anon_sym_exception,
    ACTIONS(4238), 1,
      anon_sym_module,
    ACTIONS(4240), 1,
      anon_sym_open,
    ACTIONS(4242), 1,
      anon_sym_include,
    ACTIONS(4246), 1,
      anon_sym_val,
    ACTIONS(4347), 1,
      anon_sym_let,
    ACTIONS(4349), 1,
      anon_sym_class,
    ACTIONS(4351), 1,
      sym_let_operator,
    STATE(54), 1,
      aux_sym__structure_repeat1,
    STATE(3205), 1,
      sym_directive,
    STATE(3219), 1,
      sym_attribute,
    STATE(3228), 1,
      aux_sym__structure_repeat2,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4806), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4775), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4781), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [113247] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4361), 1,
      anon_sym_RPAREN,
    STATE(3220), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4359), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4357), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [113300] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(13), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(79), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(81), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(315), 1,
      aux_sym_directive_token1,
    ACTIONS(395), 1,
      ts_builtin_sym_end,
    ACTIONS(4232), 1,
      anon_sym_external,
    ACTIONS(4234), 1,
      anon_sym_type,
    ACTIONS(4236), 1,
      anon_sym_exception,
    ACTIONS(4238), 1,
      anon_sym_module,
    ACTIONS(4240), 1,
      anon_sym_open,
    ACTIONS(4242), 1,
      anon_sym_include,
    ACTIONS(4246), 1,
      anon_sym_val,
    ACTIONS(4347), 1,
      anon_sym_let,
    ACTIONS(4349), 1,
      anon_sym_class,
    ACTIONS(4351), 1,
      sym_let_operator,
    STATE(53), 1,
      aux_sym__structure_repeat1,
    STATE(3205), 1,
      sym_directive,
    STATE(3218), 1,
      aux_sym__structure_repeat2,
    STATE(3221), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4806), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4775), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4781), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [113391] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4345), 1,
      anon_sym_PIPE,
    STATE(3222), 1,
      sym_attribute,
    STATE(3224), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3881), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3879), 26,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [113446] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4345), 1,
      anon_sym_PIPE,
    STATE(3214), 1,
      aux_sym__match_cases_repeat1,
    STATE(3223), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 26,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [113501] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4345), 1,
      anon_sym_PIPE,
    STATE(3224), 1,
      sym_attribute,
    STATE(3226), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 26,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [113556] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(3225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4365), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4363), 21,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [113607] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4367), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3226), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3704), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 26,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [113660] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4353), 1,
      anon_sym_else,
    STATE(2767), 1,
      sym_else_clause,
    STATE(3227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 9,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 25,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [113715] = 24,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4280), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(4286), 1,
      anon_sym_external,
    ACTIONS(4289), 1,
      anon_sym_type,
    ACTIONS(4292), 1,
      ts_builtin_sym_end,
    ACTIONS(4294), 1,
      anon_sym_exception,
    ACTIONS(4297), 1,
      anon_sym_module,
    ACTIONS(4300), 1,
      anon_sym_open,
    ACTIONS(4303), 1,
      anon_sym_include,
    ACTIONS(4309), 1,
      anon_sym_val,
    ACTIONS(4312), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4315), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(4318), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4324), 1,
      aux_sym_directive_token1,
    ACTIONS(4370), 1,
      anon_sym_let,
    ACTIONS(4373), 1,
      anon_sym_class,
    ACTIONS(4376), 1,
      sym_let_operator,
    STATE(56), 1,
      aux_sym__structure_repeat1,
    STATE(3205), 1,
      sym_directive,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3228), 2,
      sym_attribute,
      aux_sym__structure_repeat2,
    STATE(4806), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4775), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4781), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [113804] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(3229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4381), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4379), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [113854] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(3230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4385), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4383), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [113904] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3156), 1,
      anon_sym_COLON2,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4387), 1,
      anon_sym_SEMI,
    ACTIONS(4389), 1,
      sym_rel_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3231), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3154), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
  [113990] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4391), 1,
      anon_sym_else,
    STATE(2803), 1,
      sym_else_clause,
    STATE(3232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 11,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 22,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [114044] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4393), 1,
      anon_sym_PIPE,
    STATE(3233), 1,
      sym_attribute,
    STATE(3254), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3881), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3879), 23,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [114098] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3338), 1,
      anon_sym_COLON2,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4387), 1,
      anon_sym_SEMI,
    ACTIONS(4389), 1,
      sym_rel_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3336), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
  [114184] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(3235), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4359), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4357), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [114234] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(4401), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4403), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(3236), 1,
      sym_attribute,
    STATE(3699), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4397), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4395), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [114294] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3237), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4407), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4405), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [114344] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4128), 1,
      anon_sym_COMMA,
    STATE(3238), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 9,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 25,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [114396] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4411), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4409), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [114446] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3240), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4415), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4413), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [114496] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(4401), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4403), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(3241), 1,
      sym_attribute,
    STATE(3699), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4419), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4417), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [114556] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4423), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4421), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [114606] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(3243), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4427), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4425), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [114656] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(3244), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4361), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4429), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [114706] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(3245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4433), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4431), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [114756] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4437), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4435), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [114806] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4393), 1,
      anon_sym_PIPE,
    STATE(3247), 1,
      sym_attribute,
    STATE(3249), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3874), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3872), 23,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [114860] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4441), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4439), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [114910] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4443), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3249), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3704), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 23,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [114962] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4391), 1,
      anon_sym_else,
    STATE(2767), 1,
      sym_else_clause,
    STATE(3250), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 11,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 22,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [115016] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4448), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4446), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [115066] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4393), 1,
      anon_sym_PIPE,
    STATE(3247), 1,
      aux_sym__match_cases_repeat1,
    STATE(3252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 23,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [115120] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3160), 1,
      anon_sym_COLON2,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4387), 1,
      anon_sym_SEMI,
    ACTIONS(4389), 1,
      sym_rel_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3158), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
  [115206] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4393), 1,
      anon_sym_PIPE,
    STATE(3249), 1,
      aux_sym__match_cases_repeat1,
    STATE(3254), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 23,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [115260] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3255), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2055), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2057), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115309] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3256), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2047), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2049), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115358] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3257), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2059), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2061), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115407] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3258), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2063), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2065), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115456] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3259), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2067), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2069), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115505] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3260), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2011), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2013), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115554] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3261), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2087), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2089), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115603] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3262), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2091), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2093), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115652] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2095), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2097), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115701] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3264), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2165), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2167), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115750] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2183), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2185), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115799] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2187), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2189), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115848] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4144), 1,
      anon_sym_module,
    STATE(3267), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6102), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [115937] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2075), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2077), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115986] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3269), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2099), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2101), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [116035] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3270), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2191), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2193), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [116084] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4144), 1,
      anon_sym_module,
    STATE(3271), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7094), 1,
      sym__type,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [116173] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4124), 1,
      anon_sym_module,
    STATE(3272), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6150), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [116262] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3273), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2169), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2171), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [116311] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3274), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2071), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2073), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [116360] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4450), 1,
      anon_sym_else,
    STATE(2746), 1,
      sym_else_clause,
    STATE(3275), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 22,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [116413] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2079), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2081), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [116462] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4452), 1,
      anon_sym_else,
    STATE(2746), 1,
      sym_else_clause,
    STATE(3277), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 24,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [116515] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2015), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2017), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [116564] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3279), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2019), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2021), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [116613] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3280), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2023), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2025), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [116662] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3281), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1983), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1985), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [116711] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3282), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2027), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2029), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [116760] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4454), 1,
      aux_sym_attribute_id_token1,
    STATE(3283), 1,
      sym_attribute,
    STATE(3292), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3168), 15,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3166), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [116813] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3284), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2141), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2143), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [116862] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2035), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2037), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [116911] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3286), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1927), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1929), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [116960] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3287), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1891), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1893), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [117009] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1899), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1901), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [117058] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3289), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1903), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1905), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [117107] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3290), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1987), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1989), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [117156] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3291), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1907), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1909), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [117205] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4454), 1,
      aux_sym_attribute_id_token1,
    STATE(3292), 1,
      sym_attribute,
    STATE(3306), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3176), 15,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3174), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [117258] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3293), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2051), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2053), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [117307] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3764), 1,
      sym__lowercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4206), 1,
      anon_sym_LPAREN,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4456), 1,
      anon_sym_private,
    STATE(3294), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3814), 1,
      sym__type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [117396] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3295), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1991), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1993), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [117445] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3296), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2121), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2123), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [117494] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2157), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2159), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [117543] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4150), 1,
      anon_sym_COMMA,
    STATE(3298), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 11,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 22,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [117594] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4458), 1,
      anon_sym_private,
    STATE(3299), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(5265), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [117683] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4460), 1,
      anon_sym_AMP,
    STATE(3300), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(5781), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [117772] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4142), 1,
      anon_sym_module,
    STATE(3301), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6253), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [117861] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3302), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1935), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1937), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [117910] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3303), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1939), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1941), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [117959] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3304), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1943), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1945), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118008] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1951), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1953), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118057] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4462), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3306), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
    ACTIONS(3188), 15,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3186), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [118108] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4465), 1,
      anon_sym_private,
    STATE(3307), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(5270), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [118197] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3308), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1911), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1913), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118246] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4124), 1,
      anon_sym_module,
    STATE(3309), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7032), 1,
      sym__type,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [118335] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3310), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1995), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1997), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118384] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3311), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2149), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2151), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118433] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1955), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1957), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118482] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4467), 1,
      anon_sym_DOT,
    ACTIONS(4469), 1,
      anon_sym_POUND,
    ACTIONS(4471), 1,
      sym_hash_operator,
    STATE(3313), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3516), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3514), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118537] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1915), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1917), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118586] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3315), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1919), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1921), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118635] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3316), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2083), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2085), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118684] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3317), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2195), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2197), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118733] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3318), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2133), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2135), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118782] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1959), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1961), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118831] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2199), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2201), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118880] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3321), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2129), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2131), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118929] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2137), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2139), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [118978] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3323), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1979), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1981), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119027] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3324), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2007), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2009), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119076] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4467), 1,
      anon_sym_DOT,
    ACTIONS(4469), 1,
      anon_sym_POUND,
    ACTIONS(4471), 1,
      sym_hash_operator,
    STATE(3325), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3412), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3410), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119131] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4467), 1,
      anon_sym_DOT,
    ACTIONS(4469), 1,
      anon_sym_POUND,
    ACTIONS(4471), 1,
      sym_hash_operator,
    STATE(3326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3422), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3420), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119186] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2203), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2205), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119235] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3328), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2153), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2155), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119284] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2173), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2175), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119333] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4142), 1,
      anon_sym_module,
    STATE(3330), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7133), 1,
      sym__type,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [119422] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1999), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2001), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119471] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3332), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1947), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1949), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119520] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1963), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1965), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119569] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(3334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2113), 23,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119620] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3335), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2125), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2127), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119669] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4450), 1,
      anon_sym_else,
    STATE(2738), 1,
      sym_else_clause,
    STATE(3336), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 22,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [119722] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3337), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2003), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2005), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119771] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3338), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1887), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1889), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119820] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3339), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2161), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2163), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119869] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3340), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1923), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1925), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119918] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3341), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2145), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2147), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [119967] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(3342), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2113), 23,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [120018] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3343), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2179), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2181), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [120067] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3344), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1895), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1897), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [120116] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3345), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1967), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1969), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [120165] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4467), 1,
      anon_sym_DOT,
    STATE(3346), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1971), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1973), 23,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [120216] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2103), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2105), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [120265] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4452), 1,
      anon_sym_else,
    STATE(2738), 1,
      sym_else_clause,
    STATE(3348), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 24,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [120318] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3838), 1,
      sym__lowercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4224), 1,
      anon_sym_LPAREN,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4473), 1,
      anon_sym_private,
    STATE(3349), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4136), 1,
      sym__type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [120407] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3838), 1,
      sym__lowercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4224), 1,
      anon_sym_LPAREN,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4475), 1,
      anon_sym_private,
    STATE(3350), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4137), 1,
      sym__type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [120496] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3351), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2107), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2109), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [120545] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3352), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2117), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2119), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [120594] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3764), 1,
      sym__lowercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4206), 1,
      anon_sym_LPAREN,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4477), 1,
      anon_sym_private,
    STATE(3353), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3853), 1,
      sym__type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [120683] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3354), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1975), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1977), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [120732] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3355), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3554), 14,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(3552), 20,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [120781] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4467), 1,
      anon_sym_DOT,
    ACTIONS(4469), 1,
      anon_sym_POUND,
    ACTIONS(4471), 1,
      sym_hash_operator,
    STATE(3356), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3486), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3484), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [120836] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4228), 1,
      anon_sym_LPAREN,
    ACTIONS(4479), 1,
      anon_sym_private,
    STATE(3357), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(5182), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [120925] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(3358), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(2113), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120976] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4467), 1,
      anon_sym_DOT,
    ACTIONS(4469), 1,
      anon_sym_POUND,
    ACTIONS(4471), 1,
      sym_hash_operator,
    STATE(3359), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3490), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3488), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [121031] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3360), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2207), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2209), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [121080] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3361), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2031), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2033), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [121129] = 25,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2788), 1,
      sym__lowercase_identifier,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4228), 1,
      anon_sym_LPAREN,
    ACTIONS(4481), 1,
      anon_sym_private,
    STATE(3362), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(5186), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [121218] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(3363), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(2113), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121269] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3364), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2043), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2045), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [121318] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4228), 1,
      anon_sym_LPAREN,
    STATE(3365), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(5186), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [121404] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3366), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(5274), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [121490] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4228), 1,
      anon_sym_LPAREN,
    STATE(3367), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(5257), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [121576] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4228), 1,
      anon_sym_LPAREN,
    STATE(3368), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(6055), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [121662] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4220), 1,
      anon_sym_LPAREN,
    STATE(3369), 1,
      sym_attribute,
    STATE(4529), 1,
      sym__simple_type,
    STATE(4944), 1,
      sym__proper_tuple_type,
    STATE(4994), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(8225), 1,
      sym_labeled_argument_type,
    STATE(8227), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [121748] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4228), 1,
      anon_sym_LPAREN,
    STATE(3370), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(5913), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [121834] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4220), 1,
      anon_sym_LPAREN,
    STATE(3371), 1,
      sym_attribute,
    STATE(4529), 1,
      sym__simple_type,
    STATE(4944), 1,
      sym__proper_tuple_type,
    STATE(5017), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(8225), 1,
      sym_labeled_argument_type,
    STATE(8227), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [121920] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4220), 1,
      anon_sym_LPAREN,
    STATE(3372), 1,
      sym_attribute,
    STATE(4529), 1,
      sym__simple_type,
    STATE(4944), 1,
      sym__proper_tuple_type,
    STATE(4962), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(8225), 1,
      sym_labeled_argument_type,
    STATE(8227), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [122006] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3373), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7093), 1,
      sym__type,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [122092] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3374), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6093), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [122178] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3375), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5017), 1,
      sym__type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [122264] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3237), 1,
      sym__type,
    STATE(3376), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [122350] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4224), 1,
      anon_sym_LPAREN,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4483), 1,
      sym__lowercase_identifier,
    STATE(3377), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(4878), 1,
      sym__type,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [122436] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4220), 1,
      anon_sym_LPAREN,
    STATE(3237), 1,
      sym__type,
    STATE(3378), 1,
      sym_attribute,
    STATE(4529), 1,
      sym__simple_type,
    STATE(4944), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(8225), 1,
      sym_labeled_argument_type,
    STATE(8227), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [122522] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4220), 1,
      anon_sym_LPAREN,
    STATE(3379), 1,
      sym_attribute,
    STATE(4529), 1,
      sym__simple_type,
    STATE(4944), 1,
      sym__proper_tuple_type,
    STATE(4964), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(8225), 1,
      sym_labeled_argument_type,
    STATE(8227), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [122608] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4220), 1,
      anon_sym_LPAREN,
    STATE(3380), 1,
      sym_attribute,
    STATE(4529), 1,
      sym__simple_type,
    STATE(4944), 1,
      sym__proper_tuple_type,
    STATE(4959), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(8225), 1,
      sym_labeled_argument_type,
    STATE(8227), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [122694] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3381), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6142), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [122780] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3382), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7158), 1,
      sym__type,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [122866] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4220), 1,
      anon_sym_LPAREN,
    STATE(3248), 1,
      sym__type,
    STATE(3383), 1,
      sym_attribute,
    STATE(4529), 1,
      sym__simple_type,
    STATE(4944), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(8225), 1,
      sym_labeled_argument_type,
    STATE(8227), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [122952] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3384), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3188), 16,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
      aux_sym_attribute_id_token1,
    ACTIONS(3186), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [123000] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3385), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(4962), 1,
      sym__type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [123086] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3386), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6263), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [123172] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4228), 1,
      anon_sym_LPAREN,
    STATE(3387), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(4986), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [123258] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4224), 1,
      anon_sym_LPAREN,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4483), 1,
      sym__lowercase_identifier,
    STATE(3388), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(4879), 1,
      sym__type,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [123344] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3389), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7054), 1,
      sym__type,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [123430] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4485), 1,
      anon_sym_COMMA,
    ACTIONS(4487), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4489), 1,
      sym_concat_operator,
    ACTIONS(4491), 1,
      sym_rel_operator,
    ACTIONS(4493), 1,
      sym_assign_operator,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3322), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3320), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [123512] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3391), 1,
      sym_attribute,
    STATE(3824), 1,
      sym__type,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [123598] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3392), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(5270), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [123684] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3393), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2145), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(2147), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [123732] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3394), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(5822), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [123818] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4485), 1,
      anon_sym_COMMA,
    ACTIONS(4487), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4489), 1,
      sym_concat_operator,
    ACTIONS(4491), 1,
      sym_rel_operator,
    ACTIONS(4493), 1,
      sym_assign_operator,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3395), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3342), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3340), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [123900] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4485), 1,
      anon_sym_COMMA,
    ACTIONS(4487), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4489), 1,
      sym_concat_operator,
    ACTIONS(4491), 1,
      sym_rel_operator,
    ACTIONS(4493), 1,
      sym_assign_operator,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3396), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3332), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3330), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [123982] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4206), 1,
      anon_sym_LPAREN,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4495), 1,
      sym__lowercase_identifier,
    STATE(3397), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(4541), 1,
      sym__type,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [124068] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3248), 1,
      sym__type,
    STATE(3398), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [124154] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3399), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6022), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [124240] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4206), 1,
      anon_sym_LPAREN,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4495), 1,
      sym__lowercase_identifier,
    STATE(3400), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3822), 1,
      sym__type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [124326] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4206), 1,
      anon_sym_LPAREN,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4495), 1,
      sym__lowercase_identifier,
    STATE(3401), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3824), 1,
      sym__type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [124412] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3402), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(5739), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [124498] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4499), 1,
      sym__lowercase_identifier,
    ACTIONS(4503), 1,
      anon_sym_POUND,
    STATE(3403), 1,
      sym_attribute,
    STATE(3525), 1,
      sym_type_constructor_path,
    STATE(6926), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4497), 12,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4501), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [124556] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4206), 1,
      anon_sym_LPAREN,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4495), 1,
      sym__lowercase_identifier,
    STATE(3404), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(4542), 1,
      sym__type,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [124642] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4228), 1,
      anon_sym_LPAREN,
    STATE(3405), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(6162), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [124728] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4228), 1,
      anon_sym_LPAREN,
    STATE(3406), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(6269), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [124814] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4485), 1,
      anon_sym_COMMA,
    ACTIONS(4487), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4489), 1,
      sym_concat_operator,
    ACTIONS(4491), 1,
      sym_rel_operator,
    ACTIONS(4493), 1,
      sym_assign_operator,
    ACTIONS(4505), 1,
      anon_sym_SEMI,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3407), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3338), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3336), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [124898] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3408), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6175), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [124984] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4487), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4489), 1,
      sym_concat_operator,
    ACTIONS(4491), 1,
      sym_rel_operator,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3409), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3286), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_assign_operator,
      sym_let_and_operator,
  [125062] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4224), 1,
      anon_sym_LPAREN,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4483), 1,
      sym__lowercase_identifier,
    STATE(3248), 1,
      sym__type,
    STATE(3410), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [125148] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3411), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4507), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(4509), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [125196] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3412), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6261), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [125282] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3413), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6058), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [125368] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4224), 1,
      anon_sym_LPAREN,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4483), 1,
      sym__lowercase_identifier,
    STATE(3237), 1,
      sym__type,
    STATE(3414), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [125454] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4485), 1,
      anon_sym_COMMA,
    ACTIONS(4487), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4489), 1,
      sym_concat_operator,
    ACTIONS(4491), 1,
      sym_rel_operator,
    ACTIONS(4493), 1,
      sym_assign_operator,
    ACTIONS(4505), 1,
      anon_sym_SEMI,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3415), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3154), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [125538] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4228), 1,
      anon_sym_LPAREN,
    STATE(3416), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4973), 1,
      sym__type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [125624] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4206), 1,
      anon_sym_LPAREN,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4495), 1,
      sym__lowercase_identifier,
    STATE(3248), 1,
      sym__type,
    STATE(3417), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [125710] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4184), 1,
      anon_sym_COMMA,
    STATE(3418), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 24,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [125760] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3419), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3470), 14,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(3468), 19,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [125808] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4224), 1,
      anon_sym_LPAREN,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4483), 1,
      sym__lowercase_identifier,
    STATE(3420), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4137), 1,
      sym__type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [125894] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4224), 1,
      anon_sym_LPAREN,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4483), 1,
      sym__lowercase_identifier,
    STATE(3421), 1,
      sym_attribute,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4138), 1,
      sym__type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [125980] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4485), 1,
      anon_sym_COMMA,
    ACTIONS(4487), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4489), 1,
      sym_concat_operator,
    ACTIONS(4491), 1,
      sym_rel_operator,
    ACTIONS(4493), 1,
      sym_assign_operator,
    ACTIONS(4505), 1,
      anon_sym_SEMI,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3422), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3158), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [126064] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4206), 1,
      anon_sym_LPAREN,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4495), 1,
      sym__lowercase_identifier,
    STATE(3237), 1,
      sym__type,
    STATE(3423), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [126150] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3424), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [126208] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3425), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6786), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [126294] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3426), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6871), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [126380] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3427), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7066), 1,
      sym__type,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [126466] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    STATE(3428), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 22,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [126516] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3429), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2125), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(2127), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [126564] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3430), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2137), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(2139), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [126612] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3431), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6898), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [126698] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3432), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [126758] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4485), 1,
      anon_sym_COMMA,
    ACTIONS(4487), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4489), 1,
      sym_concat_operator,
    ACTIONS(4491), 1,
      sym_rel_operator,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3433), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3346), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3344), 6,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_assign_operator,
      sym_let_and_operator,
  [126838] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3434), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(4964), 1,
      sym__type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [126924] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4487), 1,
      anon_sym_COLON_COLON,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3435), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3350), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3348), 10,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [126992] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4487), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4489), 1,
      sym_concat_operator,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3436), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [127062] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2794), 1,
      anon_sym_LPAREN,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    STATE(3437), 1,
      sym_attribute,
    STATE(4716), 1,
      sym__simple_type,
    STATE(4959), 1,
      sym__type,
    STATE(5026), 1,
      sym__proper_tuple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7194), 1,
      sym_labeled_tuple_element_type,
    STATE(8304), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [127148] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4487), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4489), 1,
      sym_concat_operator,
    ACTIONS(4491), 1,
      sym_rel_operator,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3438), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3286), 3,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_PIPE_PIPE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [127222] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4487), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4489), 1,
      sym_concat_operator,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3439), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [127292] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3440), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [127352] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4146), 1,
      sym__lowercase_identifier,
    ACTIONS(4228), 1,
      anon_sym_LPAREN,
    STATE(3441), 1,
      sym_attribute,
    STATE(4628), 1,
      sym__simple_type,
    STATE(4983), 1,
      sym__proper_tuple_type,
    STATE(5187), 1,
      sym__type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(7743), 1,
      sym_labeled_argument_type,
    STATE(7839), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4996), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [127438] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4485), 1,
      anon_sym_COMMA,
    ACTIONS(4487), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4489), 1,
      sym_concat_operator,
    ACTIONS(4491), 1,
      sym_rel_operator,
    ACTIONS(4493), 1,
      sym_assign_operator,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3442), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3286), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [127520] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3443), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2179), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(2181), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [127568] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    STATE(805), 1,
      sym_pow_operator,
    STATE(806), 1,
      sym_mult_operator,
    STATE(807), 1,
      sym_add_operator,
    STATE(808), 1,
      sym_and_operator,
    STATE(809), 1,
      sym_or_operator,
    STATE(3444), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 8,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [127632] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4206), 1,
      anon_sym_LPAREN,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4495), 1,
      sym__lowercase_identifier,
    STATE(3445), 1,
      sym_attribute,
    STATE(3468), 1,
      sym__simple_type,
    STATE(3814), 1,
      sym__type,
    STATE(3948), 1,
      sym__proper_tuple_type,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7747), 1,
      sym_labeled_tuple_element_type,
    STATE(8007), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [127718] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4224), 1,
      anon_sym_LPAREN,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4483), 1,
      sym__lowercase_identifier,
    STATE(3446), 1,
      sym_attribute,
    STATE(3824), 1,
      sym__type,
    STATE(3852), 1,
      sym__simple_type,
    STATE(4355), 1,
      sym__proper_tuple_type,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7217), 1,
      sym_labeled_tuple_element_type,
    STATE(8327), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3246), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [127804] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3447), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [127863] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4511), 1,
      aux_sym_attribute_id_token1,
    STATE(3448), 1,
      sym_attribute,
    STATE(3451), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3168), 9,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(3166), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [127914] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4513), 1,
      anon_sym_COLON,
    STATE(3449), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3754), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3752), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [127963] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3450), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [128020] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4511), 1,
      aux_sym_attribute_id_token1,
    STATE(3451), 1,
      sym_attribute,
    STATE(3455), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3176), 9,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(3174), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [128071] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3346), 1,
      anon_sym_PIPE,
    ACTIONS(4515), 1,
      anon_sym_COMMA,
    ACTIONS(4517), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4519), 1,
      sym_concat_operator,
    ACTIONS(4521), 1,
      sym_rel_operator,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3452), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3344), 6,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_assign_operator,
  [128150] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3453), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [128209] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3454), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 7,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [128272] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3455), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
    ACTIONS(3188), 9,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(3186), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [128321] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(4528), 1,
      anon_sym_POUND,
    STATE(3456), 1,
      sym_attribute,
    STATE(3712), 1,
      sym_type_constructor_path,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4501), 11,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4497), 16,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [128378] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3156), 1,
      anon_sym_PIPE,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4515), 1,
      anon_sym_COMMA,
    ACTIONS(4517), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4519), 1,
      sym_concat_operator,
    ACTIONS(4521), 1,
      sym_rel_operator,
    ACTIONS(4530), 1,
      anon_sym_SEMI,
    ACTIONS(4532), 1,
      sym_assign_operator,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3457), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3154), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [128461] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4517), 1,
      anon_sym_COLON_COLON,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3458), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3350), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3348), 10,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [128528] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4499), 1,
      sym__lowercase_identifier,
    ACTIONS(4538), 1,
      anon_sym_POUND,
    STATE(3459), 1,
      sym_attribute,
    STATE(3555), 1,
      sym_type_constructor_path,
    STATE(6926), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4534), 11,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4536), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [128585] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3160), 1,
      anon_sym_PIPE,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4515), 1,
      anon_sym_COMMA,
    ACTIONS(4517), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4519), 1,
      sym_concat_operator,
    ACTIONS(4521), 1,
      sym_rel_operator,
    ACTIONS(4530), 1,
      anon_sym_SEMI,
    ACTIONS(4532), 1,
      sym_assign_operator,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3460), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3158), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [128668] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3338), 1,
      anon_sym_PIPE,
    ACTIONS(4515), 1,
      anon_sym_COMMA,
    ACTIONS(4517), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4519), 1,
      sym_concat_operator,
    ACTIONS(4521), 1,
      sym_rel_operator,
    ACTIONS(4530), 1,
      anon_sym_SEMI,
    ACTIONS(4532), 1,
      sym_assign_operator,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3461), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3336), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [128751] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4517), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4519), 1,
      sym_concat_operator,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3462), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3286), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [128820] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4517), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4519), 1,
      sym_concat_operator,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3463), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3286), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [128889] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3286), 1,
      anon_sym_PIPE,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4515), 1,
      anon_sym_COMMA,
    ACTIONS(4517), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4519), 1,
      sym_concat_operator,
    ACTIONS(4521), 1,
      sym_rel_operator,
    ACTIONS(4532), 1,
      sym_assign_operator,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3464), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [128970] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4517), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4519), 1,
      sym_concat_operator,
    ACTIONS(4521), 1,
      sym_rel_operator,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3465), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3286), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_PIPE,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_or,
      sym_assign_operator,
  [129043] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3286), 1,
      anon_sym_PIPE,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4517), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4519), 1,
      sym_concat_operator,
    ACTIONS(4521), 1,
      sym_rel_operator,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3466), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_assign_operator,
  [129120] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3322), 1,
      anon_sym_PIPE,
    ACTIONS(4515), 1,
      anon_sym_COMMA,
    ACTIONS(4517), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4519), 1,
      sym_concat_operator,
    ACTIONS(4521), 1,
      sym_rel_operator,
    ACTIONS(4532), 1,
      sym_assign_operator,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3467), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3320), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [129201] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4499), 1,
      sym__lowercase_identifier,
    ACTIONS(4538), 1,
      anon_sym_POUND,
    ACTIONS(4540), 1,
      anon_sym_DASH_GT,
    ACTIONS(4542), 1,
      anon_sym_STAR,
    STATE(3468), 1,
      sym_attribute,
    STATE(3555), 1,
      sym_type_constructor_path,
    STATE(6926), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4435), 9,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4437), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [129262] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3469), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2165), 4,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2167), 28,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [129309] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3346), 1,
      anon_sym_and,
    ACTIONS(4544), 1,
      anon_sym_COMMA,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    ACTIONS(4550), 1,
      sym_rel_operator,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3470), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3344), 6,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_assign_operator,
      sym_let_and_operator,
  [129388] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3160), 1,
      anon_sym_and,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4544), 1,
      anon_sym_COMMA,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    ACTIONS(4550), 1,
      sym_rel_operator,
    ACTIONS(4552), 1,
      anon_sym_SEMI,
    ACTIONS(4554), 1,
      sym_assign_operator,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3158), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [129471] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3472), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3286), 4,
      anon_sym_and,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [129540] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3473), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4509), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4507), 20,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
  [129587] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3474), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3286), 4,
      anon_sym_and,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [129656] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3286), 1,
      anon_sym_and,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4544), 1,
      anon_sym_COMMA,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    ACTIONS(4550), 1,
      sym_rel_operator,
    ACTIONS(4554), 1,
      sym_assign_operator,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3475), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [129737] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    ACTIONS(4550), 1,
      sym_rel_operator,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3476), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3286), 2,
      anon_sym_and,
      anon_sym_PIPE_PIPE,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [129810] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3286), 1,
      anon_sym_and,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    ACTIONS(4550), 1,
      sym_rel_operator,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3477), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_assign_operator,
      sym_let_and_operator,
  [129887] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3322), 1,
      anon_sym_and,
    ACTIONS(4544), 1,
      anon_sym_COMMA,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    ACTIONS(4550), 1,
      sym_rel_operator,
    ACTIONS(4554), 1,
      sym_assign_operator,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3478), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3320), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [129968] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3332), 1,
      anon_sym_and,
    ACTIONS(4544), 1,
      anon_sym_COMMA,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    ACTIONS(4550), 1,
      sym_rel_operator,
    ACTIONS(4554), 1,
      sym_assign_operator,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3479), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3330), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [130049] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3332), 1,
      anon_sym_PIPE,
    ACTIONS(4515), 1,
      anon_sym_COMMA,
    ACTIONS(4517), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4519), 1,
      sym_concat_operator,
    ACTIONS(4521), 1,
      sym_rel_operator,
    ACTIONS(4532), 1,
      sym_assign_operator,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3480), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3330), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [130130] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3481), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2183), 4,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2185), 28,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [130177] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3482), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2187), 4,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2189), 28,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [130224] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3342), 1,
      anon_sym_and,
    ACTIONS(4544), 1,
      anon_sym_COMMA,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    ACTIONS(4550), 1,
      sym_rel_operator,
    ACTIONS(4554), 1,
      sym_assign_operator,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3483), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3340), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [130305] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3484), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2191), 4,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2193), 28,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [130352] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3485), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [130409] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3486), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [130468] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3487), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [130527] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3488), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 7,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [130590] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3342), 1,
      anon_sym_PIPE,
    ACTIONS(4515), 1,
      anon_sym_COMMA,
    ACTIONS(4517), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4519), 1,
      sym_concat_operator,
    ACTIONS(4521), 1,
      sym_rel_operator,
    ACTIONS(4532), 1,
      sym_assign_operator,
    STATE(735), 1,
      sym_pow_operator,
    STATE(736), 1,
      sym_mult_operator,
    STATE(737), 1,
      sym_add_operator,
    STATE(738), 1,
      sym_and_operator,
    STATE(739), 1,
      sym_or_operator,
    STATE(3489), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3340), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [130671] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3490), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4365), 15,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4363), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [130718] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3156), 1,
      anon_sym_and,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4544), 1,
      anon_sym_COMMA,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    ACTIONS(4550), 1,
      sym_rel_operator,
    ACTIONS(4552), 1,
      anon_sym_SEMI,
    ACTIONS(4554), 1,
      sym_assign_operator,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3491), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3154), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [130801] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4499), 1,
      sym__lowercase_identifier,
    ACTIONS(4538), 1,
      anon_sym_POUND,
    ACTIONS(4560), 1,
      anon_sym_STAR,
    STATE(3492), 1,
      sym_attribute,
    STATE(3555), 1,
      sym_type_constructor_path,
    STATE(6926), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4556), 10,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4558), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [130860] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3338), 1,
      anon_sym_and,
    ACTIONS(4544), 1,
      anon_sym_COMMA,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    ACTIONS(4550), 1,
      sym_rel_operator,
    ACTIONS(4552), 1,
      anon_sym_SEMI,
    ACTIONS(4554), 1,
      sym_assign_operator,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3493), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3336), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [130943] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    STATE(3494), 1,
      sym_attribute,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4536), 11,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4534), 16,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [131000] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3495), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3350), 4,
      anon_sym_and,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3348), 10,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [131067] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3496), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [131125] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3497), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4564), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4566), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [131171] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3498), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3720), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3718), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [131217] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3499), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3724), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3722), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [131263] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3500), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3808), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3806), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [131309] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3501), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3816), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3814), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [131355] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3502), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3820), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3818), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [131401] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3503), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3824), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3822), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [131447] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(583), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(589), 1,
      sym_rel_operator,
    ACTIONS(591), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4568), 1,
      anon_sym_DOT,
    ACTIONS(4574), 1,
      anon_sym_or,
    STATE(3504), 1,
      sym_attribute,
    STATE(7799), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(569), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(587), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(585), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(581), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4572), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(4570), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(8068), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [131515] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3505), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3828), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3826), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [131561] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3506), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3832), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3830), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [131607] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3507), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3836), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3834), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [131653] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4576), 1,
      anon_sym_COLON2,
    STATE(3508), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3470), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(3468), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [131701] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4578), 1,
      anon_sym_DOT_DOT,
    STATE(3509), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1645), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1641), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [131749] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3510), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3901), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3899), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [131795] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3511), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4580), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4582), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [131841] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3512), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4584), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4586), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [131887] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3513), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4588), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4590), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [131933] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3514), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4592), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4594), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [131979] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(4401), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4403), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4598), 1,
      anon_sym_let,
    STATE(3515), 1,
      sym_attribute,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4276), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4596), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132041] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3516), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4497), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4501), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [132087] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(4401), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4403), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4598), 1,
      anon_sym_let,
    STATE(3517), 1,
      sym_attribute,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4278), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4596), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132149] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3518), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4600), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4602), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [132195] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3519), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4604), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4606), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [132241] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3520), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4608), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4610), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [132287] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3521), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3921), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3919), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [132333] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3522), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4612), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4614), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [132379] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3523), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3885), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3883), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [132425] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(3524), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2113), 27,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132473] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3525), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4616), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4618), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [132519] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3526), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4620), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4622), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [132565] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3527), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4624), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4626), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [132611] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3528), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3390), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3388), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [132657] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3529), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4628), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4630), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [132703] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3530), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2173), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2175), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [132749] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(4401), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4403), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4634), 1,
      anon_sym_let,
    STATE(3531), 1,
      sym_attribute,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4297), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4632), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132811] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3532), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4636), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4638), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [132857] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3533), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3658), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3656), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [132903] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3534), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4640), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4642), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [132949] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(4401), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4403), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4646), 1,
      anon_sym_let,
    STATE(3535), 1,
      sym_attribute,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4280), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4644), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [133011] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3536), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4648), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2115), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [133057] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3537), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3396), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3394), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [133103] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3538), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [133159] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4576), 1,
      anon_sym_COLON2,
    STATE(3539), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 13,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3470), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [133207] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [133265] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4650), 1,
      anon_sym_COMMA,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    ACTIONS(4656), 1,
      sym_rel_operator,
    ACTIONS(4658), 1,
      sym_assign_operator,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3541), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3330), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [133343] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3542), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 6,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [133405] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3338), 1,
      anon_sym_and,
    ACTIONS(4544), 1,
      anon_sym_COMMA,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    ACTIONS(4550), 1,
      sym_rel_operator,
    ACTIONS(4554), 1,
      sym_assign_operator,
    ACTIONS(4660), 1,
      anon_sym_SEMI,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3336), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [133487] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3544), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4662), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4664), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [133533] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3545), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2103), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2105), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [133579] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3546), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4666), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4668), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [133625] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3547), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2117), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2119), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [133671] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4650), 1,
      anon_sym_COMMA,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    ACTIONS(4656), 1,
      sym_rel_operator,
    ACTIONS(4658), 1,
      sym_assign_operator,
    ACTIONS(4670), 1,
      anon_sym_SEMI,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3548), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3154), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [133751] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3549), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4672), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4674), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [133797] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3550), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2169), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2171), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [133843] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3551), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4676), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4678), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [133889] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3552), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3889), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3887), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [133935] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4680), 1,
      anon_sym_COLON,
    STATE(3553), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3384), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3382), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [133983] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3554), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3905), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3903), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [134029] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3555), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4682), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4684), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [134075] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3556), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4686), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4688), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [134121] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3758), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3756), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [134167] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3558), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2107), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2109), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [134213] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3559), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [134259] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3560), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4690), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4692), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [134305] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3561), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4694), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4696), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [134351] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4650), 1,
      anon_sym_COMMA,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    ACTIONS(4656), 1,
      sym_rel_operator,
    ACTIONS(4658), 1,
      sym_assign_operator,
    ACTIONS(4670), 1,
      anon_sym_SEMI,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3562), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3336), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [134431] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3563), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3188), 10,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
      aux_sym_attribute_id_token1,
    ACTIONS(3186), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [134477] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3564), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3690), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3688), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [134523] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3565), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3925), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3923), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [134569] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3566), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3668), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3666), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [134615] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3567), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3682), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3680), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [134661] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3568), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3686), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3684), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [134707] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3160), 1,
      anon_sym_and,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4544), 1,
      anon_sym_COMMA,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    ACTIONS(4550), 1,
      sym_rel_operator,
    ACTIONS(4554), 1,
      sym_assign_operator,
    ACTIONS(4660), 1,
      anon_sym_SEMI,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3569), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3158), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [134789] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(583), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(589), 1,
      sym_rel_operator,
    ACTIONS(591), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1449), 1,
      anon_sym_DOT,
    ACTIONS(4574), 1,
      anon_sym_or,
    STATE(3570), 1,
      sym_attribute,
    STATE(7379), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(569), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(587), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(585), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(581), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4572), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(1455), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(8068), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [134857] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3571), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3152), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3150), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [134903] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3572), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2133), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2135), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [134949] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3573), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3694), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3692), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [134995] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3574), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3698), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3696), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [135041] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(583), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(589), 1,
      sym_rel_operator,
    ACTIONS(591), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1423), 1,
      anon_sym_DOT,
    ACTIONS(4574), 1,
      anon_sym_or,
    STATE(3575), 1,
      sym_attribute,
    STATE(7244), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(569), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(587), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(585), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(581), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4572), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(1441), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(8068), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [135109] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(4401), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4403), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4700), 1,
      anon_sym_let,
    STATE(3576), 1,
      sym_attribute,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4311), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4698), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135171] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(4401), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4403), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4700), 1,
      anon_sym_let,
    STATE(3577), 1,
      sym_attribute,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4312), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4698), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135233] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3578), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2167), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(2165), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [135279] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(583), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(589), 1,
      sym_rel_operator,
    ACTIONS(591), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1463), 1,
      anon_sym_DOT,
    ACTIONS(4574), 1,
      anon_sym_or,
    STATE(3579), 1,
      sym_attribute,
    STATE(7564), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(569), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(587), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(585), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(581), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4572), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(1469), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(8068), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [135347] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4706), 1,
      anon_sym_DOT,
    STATE(3580), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4704), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4702), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135395] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4650), 1,
      anon_sym_COMMA,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    ACTIONS(4656), 1,
      sym_rel_operator,
    ACTIONS(4658), 1,
      sym_assign_operator,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3581), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3340), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [135473] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3154), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [135519] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3583), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3552), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3554), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [135565] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3584), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3158), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [135611] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3704), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [135657] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3586), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4708), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4710), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [135703] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3587), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4712), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4714), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [135749] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(583), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(589), 1,
      sym_rel_operator,
    ACTIONS(591), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(711), 1,
      anon_sym_DOT,
    ACTIONS(4574), 1,
      anon_sym_or,
    STATE(3588), 1,
      sym_attribute,
    STATE(7464), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(569), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(587), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(585), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(581), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4572), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(717), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(8068), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [135817] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4716), 1,
      anon_sym_COLON2,
    STATE(3589), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 13,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3470), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [135865] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3708), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3706), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [135911] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3591), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4718), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4720), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [135957] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3712), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3710), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [136003] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3593), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4722), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4724), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [136049] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3594), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3470), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [136095] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3595), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2153), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2155), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [136141] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3596), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2185), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(2183), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [136187] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3597), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(917), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(915), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [136233] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3598), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2189), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(2187), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [136279] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3599), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3716), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3714), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [136325] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(3600), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2113), 27,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [136373] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3601), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2193), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(2191), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [136419] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3602), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4726), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2177), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [136465] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4650), 1,
      anon_sym_COMMA,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    ACTIONS(4656), 1,
      sym_rel_operator,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3344), 6,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_assign_operator,
  [136541] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3156), 1,
      anon_sym_and,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4544), 1,
      anon_sym_COMMA,
    ACTIONS(4546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4548), 1,
      sym_concat_operator,
    ACTIONS(4550), 1,
      sym_rel_operator,
    ACTIONS(4554), 1,
      sym_assign_operator,
    ACTIONS(4660), 1,
      anon_sym_SEMI,
    STATE(672), 1,
      sym_pow_operator,
    STATE(673), 1,
      sym_mult_operator,
    STATE(674), 1,
      sym_add_operator,
    STATE(675), 1,
      sym_and_operator,
    STATE(676), 1,
      sym_or_operator,
    STATE(3604), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3154), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [136623] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3605), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3350), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3348), 10,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [136689] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3606), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3286), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [136757] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3607), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3286), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [136825] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4650), 1,
      anon_sym_COMMA,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    ACTIONS(4656), 1,
      sym_rel_operator,
    ACTIONS(4658), 1,
      sym_assign_operator,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3608), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [136903] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3286), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    ACTIONS(4656), 1,
      sym_rel_operator,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3609), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_or,
      sym_assign_operator,
  [136975] = 19,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    ACTIONS(4656), 1,
      sym_rel_operator,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3610), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_assign_operator,
  [137049] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4650), 1,
      anon_sym_COMMA,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    ACTIONS(4656), 1,
      sym_rel_operator,
    ACTIONS(4658), 1,
      sym_assign_operator,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3611), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3320), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [137127] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4650), 1,
      anon_sym_COMMA,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    ACTIONS(4656), 1,
      sym_rel_operator,
    ACTIONS(4658), 1,
      sym_assign_operator,
    ACTIONS(4670), 1,
      anon_sym_SEMI,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3612), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3158), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [137207] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3613), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4728), 2,
      anon_sym_DOT,
      aux_sym_type_variable_token1,
    ACTIONS(4662), 12,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(4664), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [137254] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3614), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3456), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3454), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [137299] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3615), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3460), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3458), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [137344] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3616), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3464), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3462), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [137389] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3617), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3626), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3624), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [137434] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3618), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3524), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3522), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [137479] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3619), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3528), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3526), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [137524] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3620), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3546), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3544), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [137569] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3621), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3502), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3500), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [137614] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3622), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3622), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3620), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [137659] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3623), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4714), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4712), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [137704] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3624), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2179), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2181), 27,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [137749] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3625), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4720), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4718), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [137794] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3626), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [137849] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3346), 1,
      anon_sym_PIPE,
    ACTIONS(4730), 1,
      anon_sym_COMMA,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4734), 1,
      sym_concat_operator,
    ACTIONS(4736), 1,
      sym_rel_operator,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3627), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3344), 4,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_assign_operator,
  [137926] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3628), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3350), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3348), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [137991] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4734), 1,
      sym_concat_operator,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3629), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3286), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [138058] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4734), 1,
      sym_concat_operator,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3630), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3286), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [138125] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3286), 1,
      anon_sym_PIPE,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4730), 1,
      anon_sym_COMMA,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4734), 1,
      sym_concat_operator,
    ACTIONS(4736), 1,
      sym_rel_operator,
    ACTIONS(4738), 1,
      sym_assign_operator,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3631), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3284), 3,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [138204] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3632), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [138261] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3633), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [138318] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3634), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 7,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [138379] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4734), 1,
      sym_concat_operator,
    ACTIONS(4736), 1,
      sym_rel_operator,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3635), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3286), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_PIPE,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_or,
      sym_assign_operator,
  [138450] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3286), 1,
      anon_sym_PIPE,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4734), 1,
      sym_concat_operator,
    ACTIONS(4736), 1,
      sym_rel_operator,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3636), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_assign_operator,
  [138525] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3322), 1,
      anon_sym_PIPE,
    ACTIONS(4730), 1,
      anon_sym_COMMA,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4734), 1,
      sym_concat_operator,
    ACTIONS(4736), 1,
      sym_rel_operator,
    ACTIONS(4738), 1,
      sym_assign_operator,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3637), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3320), 3,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [138604] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3638), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4724), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4722), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [138649] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4650), 1,
      anon_sym_COMMA,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    ACTIONS(4656), 1,
      sym_rel_operator,
    ACTIONS(4658), 1,
      sym_assign_operator,
    ACTIONS(4740), 1,
      anon_sym_SEMI,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3639), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3336), 3,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [138728] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3332), 1,
      anon_sym_PIPE,
    ACTIONS(4730), 1,
      anon_sym_COMMA,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4734), 1,
      sym_concat_operator,
    ACTIONS(4736), 1,
      sym_rel_operator,
    ACTIONS(4738), 1,
      sym_assign_operator,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3640), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3330), 3,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [138807] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(79), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(81), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(2790), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(4240), 1,
      anon_sym_open,
    ACTIONS(4742), 1,
      anon_sym_external,
    ACTIONS(4744), 1,
      anon_sym_type,
    ACTIONS(4746), 1,
      anon_sym_exception,
    ACTIONS(4748), 1,
      anon_sym_module,
    ACTIONS(4750), 1,
      anon_sym_include,
    ACTIONS(4752), 1,
      anon_sym_class,
    ACTIONS(4754), 1,
      anon_sym_val,
    ACTIONS(4756), 1,
      anon_sym_end,
    STATE(3641), 1,
      sym_attribute,
    STATE(3742), 1,
      aux_sym__signature_repeat1,
    STATE(3790), 1,
      aux_sym__structure_repeat1,
    STATE(4945), 1,
      sym__signature_item,
    STATE(7449), 1,
      sym__signature,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4942), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [138884] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3642), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2165), 9,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_POUND,
      anon_sym_when,
      sym__lowercase_identifier,
    ACTIONS(2167), 21,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym_hash_operator,
      sym__uppercase_identifier,
  [138929] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3342), 1,
      anon_sym_PIPE,
    ACTIONS(4730), 1,
      anon_sym_COMMA,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4734), 1,
      sym_concat_operator,
    ACTIONS(4736), 1,
      sym_rel_operator,
    ACTIONS(4738), 1,
      sym_assign_operator,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3643), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3340), 3,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [139008] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3644), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4668), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4666), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [139053] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3645), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3426), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3424), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [139098] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3646), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3470), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3468), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [139143] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3647), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3482), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3480), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [139188] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4499), 1,
      sym__lowercase_identifier,
    ACTIONS(4538), 1,
      anon_sym_POUND,
    ACTIONS(4762), 1,
      anon_sym_DASH_GT,
    ACTIONS(4764), 1,
      anon_sym_STAR,
    STATE(3555), 1,
      sym_type_constructor_path,
    STATE(3648), 1,
      sym_attribute,
    STATE(6545), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6926), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4758), 8,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4760), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [139249] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4766), 1,
      anon_sym_COMMA,
    ACTIONS(4768), 1,
      anon_sym_COLON_COLON,
    STATE(3649), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3426), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3424), 18,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [139298] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3650), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4622), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4620), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [139343] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3651), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4638), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4636), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [139388] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3652), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4642), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4640), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [139433] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3653), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4678), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4676), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [139478] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4766), 1,
      anon_sym_COMMA,
    ACTIONS(4768), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4770), 1,
      anon_sym_PIPE,
    ACTIONS(4772), 1,
      anon_sym_as,
    STATE(3654), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3536), 9,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3534), 17,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [139531] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3655), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4684), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4682), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [139576] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3656), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3506), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3504), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [139621] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3657), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3566), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3564), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [139666] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3658), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3586), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3584), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [139711] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4766), 1,
      anon_sym_COMMA,
    STATE(3659), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3426), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3424), 19,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [139758] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3156), 1,
      anon_sym_PIPE,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4730), 1,
      anon_sym_COMMA,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4734), 1,
      sym_concat_operator,
    ACTIONS(4736), 1,
      sym_rel_operator,
    ACTIONS(4738), 1,
      sym_assign_operator,
    ACTIONS(4774), 1,
      anon_sym_SEMI,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3660), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3154), 2,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [139839] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3661), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3590), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3588), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [139884] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3662), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3602), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3600), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [139929] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3426), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3424), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [139974] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3664), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3532), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3530), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140019] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4766), 1,
      anon_sym_COMMA,
    ACTIONS(4768), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4770), 1,
      anon_sym_PIPE,
    STATE(3665), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3474), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3472), 17,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140070] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3160), 1,
      anon_sym_PIPE,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4730), 1,
      anon_sym_COMMA,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4734), 1,
      sym_concat_operator,
    ACTIONS(4736), 1,
      sym_rel_operator,
    ACTIONS(4738), 1,
      sym_assign_operator,
    ACTIONS(4774), 1,
      anon_sym_SEMI,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3666), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3158), 2,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [140151] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3667), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3594), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3592), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140196] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3668), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3610), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3608), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140241] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3669), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3554), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3552), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140286] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3670), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4704), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4702), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [140331] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(4401), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4403), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4778), 1,
      anon_sym_let,
    STATE(3671), 1,
      sym_attribute,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4442), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4776), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [140392] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3672), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3474), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3472), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140437] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3673), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3542), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3540), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140482] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3674), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3598), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3596), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140527] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3675), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4782), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4780), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [140572] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3338), 1,
      anon_sym_PIPE,
    ACTIONS(4730), 1,
      anon_sym_COMMA,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4734), 1,
      sym_concat_operator,
    ACTIONS(4736), 1,
      sym_rel_operator,
    ACTIONS(4738), 1,
      sym_assign_operator,
    ACTIONS(4774), 1,
      anon_sym_SEMI,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3676), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3336), 2,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [140653] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2157), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2159), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140698] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3678), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1645), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1641), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140743] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3679), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3444), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3442), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140788] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3680), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3614), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3612), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140833] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3681), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3430), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3428), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140878] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3682), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3434), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3432), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140923] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3683), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2165), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2167), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [140968] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3684), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3520), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3518), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [141013] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3685), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2183), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2185), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [141058] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3686), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2187), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2189), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [141103] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(4401), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4403), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4786), 1,
      anon_sym_let,
    STATE(3687), 1,
      sym_attribute,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4410), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4784), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [141164] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3688), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2125), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2127), 27,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [141209] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3689), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2121), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2123), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [141254] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3690), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4790), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4788), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [141299] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3691), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3494), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3492), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [141344] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3692), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4626), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4624), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [141389] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3693), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4630), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4628), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [141434] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4499), 1,
      sym__lowercase_identifier,
    ACTIONS(4538), 1,
      anon_sym_POUND,
    ACTIONS(4762), 1,
      anon_sym_DASH_GT,
    ACTIONS(4764), 1,
      anon_sym_STAR,
    STATE(3555), 1,
      sym_type_constructor_path,
    STATE(3694), 1,
      sym_attribute,
    STATE(6545), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6926), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4792), 8,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4794), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [141495] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3695), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2191), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2193), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [141540] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4768), 1,
      anon_sym_COLON_COLON,
    STATE(3696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3498), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3496), 19,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [141587] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3697), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2183), 9,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_POUND,
      anon_sym_when,
      sym__lowercase_identifier,
    ACTIONS(2185), 21,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym_hash_operator,
      sym__uppercase_identifier,
  [141632] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3698), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2187), 9,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_POUND,
      anon_sym_when,
      sym__lowercase_identifier,
    ACTIONS(2189), 21,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym_hash_operator,
      sym__uppercase_identifier,
  [141677] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3699), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4798), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4796), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [141722] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4365), 9,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(4363), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [141767] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4766), 1,
      anon_sym_COMMA,
    ACTIONS(4768), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4770), 1,
      anon_sym_PIPE,
    ACTIONS(4772), 1,
      anon_sym_as,
    STATE(3701), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3558), 9,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3556), 17,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [141820] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4802), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4800), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [141865] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3703), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3562), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3560), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [141910] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4766), 1,
      anon_sym_COMMA,
    ACTIONS(4768), 1,
      anon_sym_COLON_COLON,
    STATE(3704), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3562), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3560), 18,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [141959] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(4401), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4403), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4806), 1,
      anon_sym_let,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(3705), 1,
      sym_attribute,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4464), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4804), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [142020] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4766), 1,
      anon_sym_COMMA,
    STATE(3706), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3562), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3560), 19,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [142067] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(4401), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4403), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4810), 1,
      anon_sym_let,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(3707), 1,
      sym_attribute,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4445), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4808), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [142128] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3562), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3560), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [142173] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4766), 1,
      anon_sym_COMMA,
    ACTIONS(4768), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4770), 1,
      anon_sym_PIPE,
    STATE(3709), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3570), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3568), 17,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [142224] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3570), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3568), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [142269] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4688), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4686), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [142314] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4618), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4616), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [142359] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4674), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4672), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [142404] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3714), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4692), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4690), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [142449] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3715), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4696), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4694), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [142494] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3716), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3574), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3572), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [142539] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4768), 1,
      anon_sym_COLON_COLON,
    STATE(3717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3578), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3576), 19,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [142586] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3718), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3606), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3604), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [142631] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3719), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3618), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3616), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [142676] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3720), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4814), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4812), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [142721] = 19,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4816), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(4819), 1,
      anon_sym_external,
    ACTIONS(4822), 1,
      anon_sym_type,
    ACTIONS(4827), 1,
      anon_sym_exception,
    ACTIONS(4830), 1,
      anon_sym_module,
    ACTIONS(4833), 1,
      anon_sym_open,
    ACTIONS(4836), 1,
      anon_sym_include,
    ACTIONS(4839), 1,
      anon_sym_class,
    ACTIONS(4842), 1,
      anon_sym_val,
    ACTIONS(4845), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4848), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(4851), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    STATE(3909), 1,
      aux_sym__structure_repeat1,
    STATE(4945), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4825), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(3721), 2,
      sym_attribute,
      aux_sym__signature_repeat1,
    STATE(4942), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [142794] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3722), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2149), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2151), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [142839] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(79), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(81), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(2790), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(4240), 1,
      anon_sym_open,
    ACTIONS(4742), 1,
      anon_sym_external,
    ACTIONS(4744), 1,
      anon_sym_type,
    ACTIONS(4746), 1,
      anon_sym_exception,
    ACTIONS(4748), 1,
      anon_sym_module,
    ACTIONS(4750), 1,
      anon_sym_include,
    ACTIONS(4752), 1,
      anon_sym_class,
    ACTIONS(4754), 1,
      anon_sym_val,
    ACTIONS(4854), 1,
      anon_sym_end,
    STATE(3723), 1,
      sym_attribute,
    STATE(3742), 1,
      aux_sym__signature_repeat1,
    STATE(3790), 1,
      aux_sym__structure_repeat1,
    STATE(4945), 1,
      sym__signature_item,
    STATE(7493), 1,
      sym__signature,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4942), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [142916] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3724), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4566), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4564), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [142961] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4650), 1,
      anon_sym_COMMA,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    ACTIONS(4656), 1,
      sym_rel_operator,
    ACTIONS(4658), 1,
      sym_assign_operator,
    ACTIONS(4740), 1,
      anon_sym_SEMI,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3725), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3154), 3,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [143040] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3726), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4582), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4580), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [143085] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3727), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4586), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4584), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [143130] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4590), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4588), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [143175] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3729), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4594), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4592), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [143220] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3730), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3550), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3548), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [143265] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4389), 1,
      sym_rel_operator,
    ACTIONS(4856), 1,
      anon_sym_RPAREN,
    ACTIONS(4858), 1,
      anon_sym_COLON2,
    ACTIONS(4860), 1,
      anon_sym_COLON_GT,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3731), 1,
      sym_attribute,
    STATE(7017), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [143348] = 24,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4389), 1,
      sym_rel_operator,
    ACTIONS(4858), 1,
      anon_sym_COLON2,
    ACTIONS(4862), 1,
      anon_sym_RPAREN,
    ACTIONS(4864), 1,
      anon_sym_COLON_GT,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3732), 1,
      sym_attribute,
    STATE(6953), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [143431] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3733), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4868), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4866), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143476] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3734), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4501), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4497), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [143521] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3735), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4872), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4870), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143566] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4650), 1,
      anon_sym_COMMA,
    ACTIONS(4652), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4654), 1,
      sym_concat_operator,
    ACTIONS(4656), 1,
      sym_rel_operator,
    ACTIONS(4658), 1,
      sym_assign_operator,
    ACTIONS(4740), 1,
      anon_sym_SEMI,
    STATE(718), 1,
      sym_pow_operator,
    STATE(719), 1,
      sym_mult_operator,
    STATE(720), 1,
      sym_add_operator,
    STATE(721), 1,
      sym_and_operator,
    STATE(722), 1,
      sym_or_operator,
    STATE(3736), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3158), 3,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [143645] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(4401), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4403), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4876), 1,
      anon_sym_let,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(3737), 1,
      sym_attribute,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4473), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4874), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143706] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3738), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4880), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4878), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143751] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3739), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2195), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2197), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [143796] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3740), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2199), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2201), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [143841] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3741), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4884), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4882), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143886] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(79), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(81), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(2790), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(4240), 1,
      anon_sym_open,
    ACTIONS(4742), 1,
      anon_sym_external,
    ACTIONS(4744), 1,
      anon_sym_type,
    ACTIONS(4746), 1,
      anon_sym_exception,
    ACTIONS(4748), 1,
      anon_sym_module,
    ACTIONS(4750), 1,
      anon_sym_include,
    ACTIONS(4752), 1,
      anon_sym_class,
    ACTIONS(4754), 1,
      anon_sym_val,
    STATE(3721), 1,
      aux_sym__signature_repeat1,
    STATE(3742), 1,
      sym_attribute,
    STATE(3770), 1,
      aux_sym__structure_repeat1,
    STATE(4945), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4886), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4942), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [143961] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3743), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2203), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2205), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [144006] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2031), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2033), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [144051] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3745), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2129), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2131), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [144096] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2191), 9,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_POUND,
      anon_sym_when,
      sym__lowercase_identifier,
    ACTIONS(2193), 21,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym_hash_operator,
      sym__uppercase_identifier,
  [144141] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3747), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4602), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4600), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [144186] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3748), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4606), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4604), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [144231] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3749), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4610), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4608), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [144276] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3750), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4614), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4612), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [144321] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3751), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2161), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2163), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [144366] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3752), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4890), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4888), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144411] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3753), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4894), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4892), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144456] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3754), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4710), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4708), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [144501] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3755), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3582), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3580), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [144546] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3438), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3436), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [144591] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3757), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4664), 12,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4662), 18,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [144636] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3758), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3448), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3446), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [144681] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3759), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3452), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3450), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [144726] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3286), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [144792] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4902), 1,
      anon_sym_let,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(4908), 1,
      anon_sym_COLON2,
    STATE(3761), 1,
      sym_attribute,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4161), 1,
      sym__module_typed,
    STATE(4257), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4904), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(4900), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144854] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4910), 1,
      anon_sym_PIPE,
    STATE(3762), 1,
      sym_attribute,
    STATE(3763), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3874), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3872), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [144902] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4912), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3763), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3704), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [144948] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4507), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(4509), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [144992] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4915), 1,
      sym__lowercase_identifier,
    ACTIONS(4917), 1,
      anon_sym_LPAREN,
    STATE(3765), 1,
      sym_attribute,
    STATE(3934), 1,
      sym__simple_type,
    STATE(3963), 1,
      sym_type_variable,
    STATE(4180), 1,
      sym_record_declaration,
    STATE(6237), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(7926), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [145072] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4919), 1,
      anon_sym_else,
    STATE(3567), 1,
      sym_else_clause,
    STATE(3766), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [145120] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4919), 1,
      anon_sym_else,
    STATE(3502), 1,
      sym_else_clause,
    STATE(3767), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [145168] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4921), 1,
      sym__lowercase_identifier,
    ACTIONS(4923), 1,
      anon_sym_LPAREN,
    STATE(3613), 1,
      sym_type_variable,
    STATE(3648), 1,
      sym__simple_type,
    STATE(3768), 1,
      sym_attribute,
    STATE(4180), 1,
      sym_record_declaration,
    STATE(6017), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(8109), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [145248] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4730), 1,
      anon_sym_COMMA,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4734), 1,
      sym_concat_operator,
    ACTIONS(4736), 1,
      sym_rel_operator,
    ACTIONS(4738), 1,
      sym_assign_operator,
    ACTIONS(4925), 1,
      anon_sym_SEMI,
    ACTIONS(4927), 1,
      anon_sym_PIPE_RBRACK,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3769), 1,
      sym_attribute,
    STATE(6612), 1,
      aux_sym__sequence_expression_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [145328] = 19,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(79), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(81), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(2790), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(4240), 1,
      anon_sym_open,
    ACTIONS(4742), 1,
      anon_sym_external,
    ACTIONS(4744), 1,
      anon_sym_type,
    ACTIONS(4746), 1,
      anon_sym_exception,
    ACTIONS(4748), 1,
      anon_sym_module,
    ACTIONS(4750), 1,
      anon_sym_include,
    ACTIONS(4752), 1,
      anon_sym_class,
    ACTIONS(4754), 1,
      anon_sym_val,
    STATE(3770), 1,
      sym_attribute,
    STATE(4931), 1,
      aux_sym__structure_repeat1,
    STATE(4972), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4929), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4942), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [145400] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3328), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3326), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [145454] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    ACTIONS(4931), 1,
      anon_sym_COMMA,
    ACTIONS(4933), 1,
      sym_rel_operator,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3772), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3344), 4,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_assign_operator,
  [145528] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3773), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3350), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3348), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [145592] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3774), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3286), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [145658] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(4908), 1,
      anon_sym_COLON2,
    ACTIONS(4937), 1,
      anon_sym_let,
    STATE(3761), 1,
      aux_sym_module_binding_repeat1,
    STATE(3775), 1,
      sym_attribute,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4131), 1,
      sym__module_typed,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4323), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4939), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(4935), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145720] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    STATE(3776), 1,
      sym_attribute,
    STATE(4183), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4943), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    ACTIONS(4941), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145776] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3777), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [145832] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3778), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3286), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [145888] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3779), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3286), 6,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3284), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [145948] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3286), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    ACTIONS(4933), 1,
      sym_rel_operator,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3780), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      anon_sym_or,
      sym_assign_operator,
  [146018] = 19,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    ACTIONS(4933), 1,
      sym_rel_operator,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3781), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(3284), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_assign_operator,
  [146090] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    ACTIONS(4931), 1,
      anon_sym_COMMA,
    ACTIONS(4933), 1,
      sym_rel_operator,
    ACTIONS(4953), 1,
      sym_assign_operator,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3782), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3320), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [146166] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    ACTIONS(4931), 1,
      anon_sym_COMMA,
    ACTIONS(4933), 1,
      sym_rel_operator,
    ACTIONS(4953), 1,
      sym_assign_operator,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3783), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3330), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [146242] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    ACTIONS(4931), 1,
      anon_sym_COMMA,
    ACTIONS(4933), 1,
      sym_rel_operator,
    ACTIONS(4953), 1,
      sym_assign_operator,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3784), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3340), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [146318] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4499), 1,
      sym__lowercase_identifier,
    ACTIONS(4538), 1,
      anon_sym_POUND,
    ACTIONS(4957), 1,
      anon_sym_STAR,
    STATE(3555), 1,
      sym_type_constructor_path,
    STATE(3785), 1,
      sym_attribute,
    STATE(4139), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6926), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4762), 8,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4955), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [146376] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    ACTIONS(4931), 1,
      anon_sym_COMMA,
    ACTIONS(4933), 1,
      sym_rel_operator,
    ACTIONS(4953), 1,
      sym_assign_operator,
    ACTIONS(4959), 1,
      anon_sym_SEMI,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3786), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3154), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [146454] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    ACTIONS(4931), 1,
      anon_sym_COMMA,
    ACTIONS(4933), 1,
      sym_rel_operator,
    ACTIONS(4953), 1,
      sym_assign_operator,
    ACTIONS(4959), 1,
      anon_sym_SEMI,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3787), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3158), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [146532] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    ACTIONS(4931), 1,
      anon_sym_COMMA,
    ACTIONS(4933), 1,
      sym_rel_operator,
    ACTIONS(4953), 1,
      sym_assign_operator,
    ACTIONS(4959), 1,
      anon_sym_SEMI,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3788), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3336), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [146610] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4910), 1,
      anon_sym_PIPE,
    STATE(3789), 1,
      sym_attribute,
    STATE(3795), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3881), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3879), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [146658] = 19,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(79), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(81), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(2790), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(4240), 1,
      anon_sym_open,
    ACTIONS(4742), 1,
      anon_sym_external,
    ACTIONS(4744), 1,
      anon_sym_type,
    ACTIONS(4746), 1,
      anon_sym_exception,
    ACTIONS(4748), 1,
      anon_sym_module,
    ACTIONS(4750), 1,
      anon_sym_include,
    ACTIONS(4752), 1,
      anon_sym_class,
    ACTIONS(4754), 1,
      anon_sym_val,
    STATE(3790), 1,
      sym_attribute,
    STATE(4931), 1,
      aux_sym__structure_repeat1,
    STATE(4972), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4886), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4942), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [146730] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4961), 1,
      sym__lowercase_identifier,
    ACTIONS(4963), 1,
      anon_sym_POUND,
    STATE(3791), 1,
      sym_attribute,
    STATE(3890), 1,
      sym_type_constructor_path,
    STATE(6932), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4497), 10,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4501), 14,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
  [146784] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    STATE(3792), 1,
      sym_attribute,
    STATE(3810), 1,
      sym_type_variable,
    STATE(4180), 1,
      sym_record_declaration,
    STATE(5006), 1,
      sym__simple_type,
    STATE(6076), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(8000), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [146864] = 23,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4389), 1,
      sym_rel_operator,
    ACTIONS(4927), 1,
      anon_sym_RBRACK,
    ACTIONS(4969), 1,
      anon_sym_SEMI,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3793), 1,
      sym_attribute,
    STATE(6318), 1,
      aux_sym__sequence_expression_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [146944] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4910), 1,
      anon_sym_PIPE,
    STATE(3762), 1,
      aux_sym__match_cases_repeat1,
    STATE(3794), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [146992] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4910), 1,
      anon_sym_PIPE,
    STATE(3763), 1,
      aux_sym__match_cases_repeat1,
    STATE(3795), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [147040] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    ACTIONS(4931), 1,
      anon_sym_COMMA,
    ACTIONS(4933), 1,
      sym_rel_operator,
    ACTIONS(4953), 1,
      sym_assign_operator,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3796), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3284), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [147116] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3797), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4973), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4971), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147159] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4389), 1,
      sym_rel_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3798), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(4975), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [147234] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4977), 1,
      anon_sym_else,
    STATE(3567), 1,
      sym_else_clause,
    STATE(3799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [147281] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4981), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4983), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4979), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147326] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4985), 1,
      sym__lowercase_identifier,
    ACTIONS(4987), 1,
      anon_sym_LPAREN,
    STATE(3801), 1,
      sym_attribute,
    STATE(5533), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(7750), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [147399] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4485), 1,
      anon_sym_COMMA,
    STATE(3802), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 18,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [147444] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4915), 1,
      sym__lowercase_identifier,
    ACTIONS(4917), 1,
      anon_sym_LPAREN,
    STATE(3803), 1,
      sym_attribute,
    STATE(3960), 1,
      sym__simple_type,
    STATE(4180), 1,
      sym_record_declaration,
    STATE(7174), 1,
      sym_extended_module_path,
    STATE(8053), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [147519] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(4989), 1,
      sym__lowercase_identifier,
    ACTIONS(4991), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4993), 1,
      anon_sym_GT,
    STATE(3804), 1,
      sym_attribute,
    STATE(5541), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(6941), 1,
      sym_method_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [147594] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(4989), 1,
      sym__lowercase_identifier,
    ACTIONS(4995), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4997), 1,
      anon_sym_GT,
    STATE(3805), 1,
      sym_attribute,
    STATE(5541), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(6941), 1,
      sym_method_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [147669] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4921), 1,
      sym__lowercase_identifier,
    ACTIONS(4923), 1,
      anon_sym_LPAREN,
    STATE(3785), 1,
      sym__simple_type,
    STATE(3806), 1,
      sym_attribute,
    STATE(4180), 1,
      sym_record_declaration,
    STATE(4248), 1,
      sym__constructor_argument,
    STATE(6862), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [147744] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(4989), 1,
      sym__lowercase_identifier,
    ACTIONS(4999), 1,
      anon_sym_DOT_DOT,
    ACTIONS(5001), 1,
      anon_sym_GT,
    STATE(3807), 1,
      sym_attribute,
    STATE(5541), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(6941), 1,
      sym_method_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [147819] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4389), 1,
      sym_rel_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3808), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(5003), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [147894] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(4989), 1,
      sym__lowercase_identifier,
    ACTIONS(5005), 1,
      anon_sym_DOT_DOT,
    ACTIONS(5007), 1,
      anon_sym_GT,
    STATE(3809), 1,
      sym_attribute,
    STATE(5541), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(6941), 1,
      sym_method_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [147969] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3810), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4728), 2,
      anon_sym_DOT,
      aux_sym_type_variable_token1,
    ACTIONS(4664), 11,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(4662), 15,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [148014] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5009), 1,
      anon_sym_PIPE,
    STATE(3811), 1,
      sym_attribute,
    STATE(3817), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3881), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3879), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [148061] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(4989), 1,
      sym__lowercase_identifier,
    ACTIONS(5011), 1,
      anon_sym_DOT_DOT,
    ACTIONS(5013), 1,
      anon_sym_GT,
    STATE(3812), 1,
      sym_attribute,
    STATE(5541), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(6941), 1,
      sym_method_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [148136] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4730), 1,
      anon_sym_COMMA,
    ACTIONS(4732), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4734), 1,
      sym_concat_operator,
    ACTIONS(4736), 1,
      sym_rel_operator,
    ACTIONS(4738), 1,
      sym_assign_operator,
    STATE(777), 1,
      sym_pow_operator,
    STATE(778), 1,
      sym_mult_operator,
    STATE(779), 1,
      sym_add_operator,
    STATE(780), 1,
      sym_and_operator,
    STATE(781), 1,
      sym_or_operator,
    STATE(3813), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(5015), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [148211] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5023), 1,
      anon_sym_as,
    STATE(3814), 1,
      sym_attribute,
    STATE(3866), 1,
      aux_sym_type_binding_repeat1,
    STATE(3914), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5019), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5017), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148262] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3468), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5027), 1,
      anon_sym_let,
    ACTIONS(5029), 1,
      anon_sym_EQ,
    ACTIONS(5031), 1,
      anon_sym_COLON_EQ,
    STATE(3815), 1,
      sym_attribute,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4071), 1,
      aux_sym_type_binding_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4339), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5025), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148323] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3338), 1,
      anon_sym_DASH_GT,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4389), 1,
      sym_rel_operator,
    ACTIONS(5033), 1,
      anon_sym_SEMI,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3816), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [148400] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5009), 1,
      anon_sym_PIPE,
    STATE(3817), 1,
      sym_attribute,
    STATE(3830), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [148447] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(4989), 1,
      sym__lowercase_identifier,
    ACTIONS(5035), 1,
      anon_sym_DOT_DOT,
    ACTIONS(5037), 1,
      anon_sym_GT,
    STATE(3818), 1,
      sym_attribute,
    STATE(5360), 1,
      sym__simple_type,
    STATE(6345), 1,
      sym_method_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [148522] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3154), 1,
      anon_sym_do,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    ACTIONS(4931), 1,
      anon_sym_COMMA,
    ACTIONS(4933), 1,
      sym_rel_operator,
    ACTIONS(4953), 1,
      sym_assign_operator,
    ACTIONS(5039), 1,
      anon_sym_SEMI,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3819), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [148599] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3160), 1,
      anon_sym_DASH_GT,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4389), 1,
      sym_rel_operator,
    ACTIONS(5033), 1,
      anon_sym_SEMI,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3820), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [148676] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(3821), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(2113), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [148721] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5023), 1,
      anon_sym_as,
    STATE(3822), 1,
      sym_attribute,
    STATE(3870), 1,
      aux_sym_type_binding_repeat1,
    STATE(3914), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5043), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5041), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148772] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5009), 1,
      anon_sym_PIPE,
    STATE(3823), 1,
      sym_attribute,
    STATE(3830), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3874), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3872), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [148819] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5023), 1,
      anon_sym_as,
    STATE(3824), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5047), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5045), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148864] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(4989), 1,
      sym__lowercase_identifier,
    ACTIONS(5049), 1,
      anon_sym_DOT_DOT,
    ACTIONS(5051), 1,
      anon_sym_GT,
    STATE(3825), 1,
      sym_attribute,
    STATE(5395), 1,
      sym__simple_type,
    STATE(6434), 1,
      sym_method_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [148939] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4389), 1,
      sym_rel_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3826), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(5015), 2,
      anon_sym_RBRACK,
      anon_sym_SEMI,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [149014] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4196), 1,
      sym_rel_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(5053), 2,
      anon_sym_SEMI,
      anon_sym_GT_RBRACE,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [149089] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3156), 1,
      anon_sym_DASH_GT,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    ACTIONS(4192), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4194), 1,
      sym_concat_operator,
    ACTIONS(4204), 1,
      sym_assign_operator,
    ACTIONS(4389), 1,
      sym_rel_operator,
    ACTIONS(5033), 1,
      anon_sym_SEMI,
    STATE(751), 1,
      sym_pow_operator,
    STATE(752), 1,
      sym_mult_operator,
    STATE(753), 1,
      sym_add_operator,
    STATE(754), 1,
      sym_and_operator,
    STATE(755), 1,
      sym_or_operator,
    STATE(3828), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [149166] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    STATE(3829), 1,
      sym_attribute,
    STATE(4180), 1,
      sym_record_declaration,
    STATE(5029), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(8228), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [149241] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5055), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3830), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3704), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [149286] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5009), 1,
      anon_sym_PIPE,
    STATE(3823), 1,
      aux_sym__match_cases_repeat1,
    STATE(3831), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [149333] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4499), 1,
      sym__lowercase_identifier,
    ACTIONS(4538), 1,
      anon_sym_POUND,
    STATE(3555), 1,
      sym_type_constructor_path,
    STATE(3832), 1,
      sym_attribute,
    STATE(6926), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5058), 9,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(5060), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [149386] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4985), 1,
      sym__lowercase_identifier,
    ACTIONS(4987), 1,
      anon_sym_LPAREN,
    STATE(3833), 1,
      sym_attribute,
    STATE(5403), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(7750), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [149459] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3834), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4971), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(5064), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5062), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149504] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4961), 1,
      sym__lowercase_identifier,
    ACTIONS(5066), 1,
      anon_sym_STAR,
    ACTIONS(5068), 1,
      anon_sym_POUND,
    STATE(3835), 1,
      sym_attribute,
    STATE(3927), 1,
      sym_type_constructor_path,
    STATE(6932), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4556), 8,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4558), 14,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
  [149559] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3746), 1,
      anon_sym_LBRACE,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    STATE(3836), 1,
      sym_attribute,
    STATE(5163), 1,
      sym__simple_type,
    STATE(5904), 1,
      sym__constructor_argument,
    STATE(5908), 1,
      sym_record_declaration,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [149634] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5070), 1,
      anon_sym_else,
    STATE(3502), 1,
      sym_else_clause,
    STATE(3837), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [149681] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3468), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_let,
    ACTIONS(5076), 1,
      anon_sym_EQ,
    ACTIONS(5078), 1,
      anon_sym_COLON_EQ,
    STATE(3838), 1,
      sym_attribute,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(3979), 1,
      aux_sym_type_binding_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4256), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5072), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149742] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2792), 1,
      anon_sym_QMARK,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4985), 1,
      sym__lowercase_identifier,
    ACTIONS(5080), 1,
      anon_sym_LPAREN,
    STATE(3839), 1,
      sym_attribute,
    STATE(5499), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(7536), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [149815] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5084), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5082), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149858] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(4989), 1,
      sym__lowercase_identifier,
    ACTIONS(5086), 1,
      anon_sym_DOT_DOT,
    ACTIONS(5088), 1,
      anon_sym_GT,
    STATE(3841), 1,
      sym_attribute,
    STATE(5541), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(6941), 1,
      sym_method_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [149933] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3842), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4509), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4507), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149976] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4977), 1,
      anon_sym_else,
    STATE(3502), 1,
      sym_else_clause,
    STATE(3843), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [150023] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(4989), 1,
      sym__lowercase_identifier,
    ACTIONS(5090), 1,
      anon_sym_DOT_DOT,
    ACTIONS(5092), 1,
      anon_sym_GT,
    STATE(3844), 1,
      sym_attribute,
    STATE(5377), 1,
      sym__simple_type,
    STATE(6517), 1,
      sym_method_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [150098] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(3845), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(2113), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [150143] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4915), 1,
      sym__lowercase_identifier,
    ACTIONS(4917), 1,
      anon_sym_LPAREN,
    STATE(3846), 1,
      sym_attribute,
    STATE(4039), 1,
      sym__simple_type,
    STATE(4180), 1,
      sym_record_declaration,
    STATE(4248), 1,
      sym__constructor_argument,
    STATE(7174), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [150218] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4961), 1,
      sym__lowercase_identifier,
    ACTIONS(5068), 1,
      anon_sym_POUND,
    STATE(3847), 1,
      sym_attribute,
    STATE(3927), 1,
      sym_type_constructor_path,
    STATE(6932), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4534), 9,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4536), 14,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
  [150271] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5070), 1,
      anon_sym_else,
    STATE(3567), 1,
      sym_else_clause,
    STATE(3848), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [150318] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5094), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4983), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150361] = 21,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3780), 1,
      anon_sym_LBRACE,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4921), 1,
      sym__lowercase_identifier,
    ACTIONS(4923), 1,
      anon_sym_LPAREN,
    STATE(3694), 1,
      sym__simple_type,
    STATE(3850), 1,
      sym_attribute,
    STATE(4180), 1,
      sym_record_declaration,
    STATE(6862), 1,
      sym_extended_module_path,
    STATE(7975), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [150436] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3336), 1,
      anon_sym_do,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    ACTIONS(4931), 1,
      anon_sym_COMMA,
    ACTIONS(4933), 1,
      sym_rel_operator,
    ACTIONS(4953), 1,
      sym_assign_operator,
    ACTIONS(5039), 1,
      anon_sym_SEMI,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3851), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [150513] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4961), 1,
      sym__lowercase_identifier,
    ACTIONS(5068), 1,
      anon_sym_POUND,
    ACTIONS(5096), 1,
      anon_sym_DASH_GT,
    ACTIONS(5098), 1,
      anon_sym_STAR,
    STATE(3852), 1,
      sym_attribute,
    STATE(3927), 1,
      sym_type_constructor_path,
    STATE(6932), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4435), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4437), 14,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
  [150570] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5023), 1,
      anon_sym_as,
    STATE(3853), 1,
      sym_attribute,
    STATE(3893), 1,
      aux_sym_type_binding_repeat1,
    STATE(3914), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5102), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5100), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150621] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(5110), 1,
      anon_sym_DOT,
    STATE(3854), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5106), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5104), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150668] = 22,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3158), 1,
      anon_sym_do,
    ACTIONS(3296), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3304), 1,
      anon_sym_or,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4896), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4898), 1,
      sym_concat_operator,
    ACTIONS(4931), 1,
      anon_sym_COMMA,
    ACTIONS(4933), 1,
      sym_rel_operator,
    ACTIONS(4953), 1,
      sym_assign_operator,
    ACTIONS(5039), 1,
      anon_sym_SEMI,
    STATE(873), 1,
      sym_pow_operator,
    STATE(874), 1,
      sym_mult_operator,
    STATE(875), 1,
      sym_add_operator,
    STATE(876), 1,
      sym_and_operator,
    STATE(877), 1,
      sym_or_operator,
    STATE(3855), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3288), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(3292), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(3294), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(3298), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [150745] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5116), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5114), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3856), 2,
      sym_attribute,
      aux_sym_type_binding_repeat1,
    ACTIONS(5112), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150791] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3857), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4584), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4586), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [150833] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3858), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2167), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(2165), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [150875] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5023), 1,
      anon_sym_as,
    ACTIONS(5121), 1,
      anon_sym_let,
    ACTIONS(5123), 1,
      anon_sym_EQ,
    STATE(3859), 1,
      sym_attribute,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(3976), 1,
      aux_sym_type_binding_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4216), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5119), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150933] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4923), 1,
      anon_sym_LPAREN,
    ACTIONS(5125), 1,
      sym__lowercase_identifier,
    STATE(3242), 1,
      sym__tuple_type_rhs,
    STATE(3492), 1,
      sym__simple_type,
    STATE(3860), 1,
      sym_attribute,
    STATE(3871), 1,
      sym_labeled_tuple_element_type,
    STATE(6862), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [151005] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3861), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4628), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4630), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [151047] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(5127), 1,
      sym__lowercase_identifier,
    STATE(3862), 1,
      sym_attribute,
    STATE(4666), 1,
      sym__simple_type,
    STATE(4957), 1,
      sym_labeled_tuple_element_type,
    STATE(5010), 1,
      sym__tuple_type_rhs,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [151119] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3470), 1,
      sym__lowercase_identifier,
    ACTIONS(5129), 1,
      anon_sym_COLON2,
    STATE(3863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(4640), 8,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4642), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [151167] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(5131), 1,
      sym__lowercase_identifier,
    STATE(3239), 1,
      sym__tuple_type_rhs,
    STATE(3864), 1,
      sym_attribute,
    STATE(4533), 1,
      sym__simple_type,
    STATE(4930), 1,
      sym_labeled_tuple_element_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [151239] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(5131), 1,
      sym__lowercase_identifier,
    STATE(3242), 1,
      sym__tuple_type_rhs,
    STATE(3865), 1,
      sym_attribute,
    STATE(4533), 1,
      sym__simple_type,
    STATE(4930), 1,
      sym_labeled_tuple_element_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [151311] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    STATE(3856), 1,
      aux_sym_type_binding_repeat1,
    STATE(3866), 1,
      sym_attribute,
    STATE(3914), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5135), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5133), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151359] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(5131), 1,
      sym__lowercase_identifier,
    STATE(3251), 1,
      sym__tuple_type_rhs,
    STATE(3867), 1,
      sym_attribute,
    STATE(4533), 1,
      sym__simple_type,
    STATE(4930), 1,
      sym_labeled_tuple_element_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [151431] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4666), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4668), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [151473] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3869), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4588), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4590), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [151515] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    STATE(3856), 1,
      aux_sym_type_binding_repeat1,
    STATE(3870), 1,
      sym_attribute,
    STATE(3914), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5139), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5137), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151563] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4560), 1,
      anon_sym_STAR,
    STATE(3871), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4558), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4556), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151607] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4515), 1,
      anon_sym_COMMA,
    STATE(3872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 18,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [151651] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4923), 1,
      anon_sym_LPAREN,
    ACTIONS(5125), 1,
      sym__lowercase_identifier,
    STATE(3251), 1,
      sym__tuple_type_rhs,
    STATE(3492), 1,
      sym__simple_type,
    STATE(3871), 1,
      sym_labeled_tuple_element_type,
    STATE(3873), 1,
      sym_attribute,
    STATE(6862), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [151723] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(5127), 1,
      sym__lowercase_identifier,
    STATE(3239), 1,
      sym__tuple_type_rhs,
    STATE(3874), 1,
      sym_attribute,
    STATE(4688), 1,
      sym__simple_type,
    STATE(4971), 1,
      sym_labeled_tuple_element_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [151795] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5141), 1,
      anon_sym_else,
    STATE(3567), 1,
      sym_else_clause,
    STATE(3875), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [151841] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3876), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4686), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4688), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [151883] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3877), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4971), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(2113), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151927] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(5127), 1,
      sym__lowercase_identifier,
    STATE(3878), 1,
      sym_attribute,
    STATE(4666), 1,
      sym__simple_type,
    STATE(4957), 1,
      sym_labeled_tuple_element_type,
    STATE(5002), 1,
      sym__tuple_type_rhs,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [151999] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(3879), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2125), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(2127), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [152041] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(3880), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2179), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(2181), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [152083] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(5127), 1,
      sym__lowercase_identifier,
    STATE(3881), 1,
      sym_attribute,
    STATE(4666), 1,
      sym__simple_type,
    STATE(4957), 1,
      sym_labeled_tuple_element_type,
    STATE(5018), 1,
      sym__tuple_type_rhs,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [152155] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3882), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4690), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4692), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [152197] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4917), 1,
      anon_sym_LPAREN,
    ACTIONS(5143), 1,
      sym__lowercase_identifier,
    STATE(3251), 1,
      sym__tuple_type_rhs,
    STATE(3835), 1,
      sym__simple_type,
    STATE(3883), 1,
      sym_attribute,
    STATE(4175), 1,
      sym_labeled_tuple_element_type,
    STATE(7174), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [152269] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(3884), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2137), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(2139), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [152311] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(3885), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2145), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(2147), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [152353] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5149), 1,
      anon_sym_LPAREN,
    STATE(4078), 1,
      sym_module_parameter,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3886), 2,
      sym_attribute,
      aux_sym_module_binding_repeat1,
    ACTIONS(5147), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5145), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152399] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3887), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4624), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4626), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [152441] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4608), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4610), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [152483] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3889), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2185), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(2183), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [152525] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3890), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4616), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4618), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [152567] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3891), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4592), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4594), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [152609] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3892), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4580), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4582), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [152651] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    STATE(3856), 1,
      aux_sym_type_binding_repeat1,
    STATE(3893), 1,
      sym_attribute,
    STATE(3914), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5154), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5152), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152699] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2189), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(2187), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [152741] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5023), 1,
      anon_sym_as,
    ACTIONS(5158), 1,
      anon_sym_let,
    ACTIONS(5160), 1,
      anon_sym_EQ,
    STATE(3895), 1,
      sym_attribute,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(3985), 1,
      aux_sym_type_binding_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4218), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5156), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152799] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3896), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2193), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(2191), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [152841] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3897), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4708), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4710), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [152883] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3898), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4712), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4714), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [152925] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5162), 1,
      anon_sym_COLON2,
    STATE(3899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 11,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3470), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [152969] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(5127), 1,
      sym__lowercase_identifier,
    STATE(3242), 1,
      sym__tuple_type_rhs,
    STATE(3900), 1,
      sym_attribute,
    STATE(4688), 1,
      sym__simple_type,
    STATE(4971), 1,
      sym_labeled_tuple_element_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [153041] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3901), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4640), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4642), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [153083] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3902), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4718), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4720), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [153125] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4499), 1,
      sym__lowercase_identifier,
    ACTIONS(4538), 1,
      anon_sym_POUND,
    STATE(3555), 1,
      sym_type_constructor_path,
    STATE(3903), 1,
      sym_attribute,
    STATE(6926), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5164), 8,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(5166), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [153177] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3904), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4722), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4724), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [153219] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3905), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4672), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4674), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [153261] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4499), 1,
      sym__lowercase_identifier,
    ACTIONS(4538), 1,
      anon_sym_POUND,
    STATE(3555), 1,
      sym_type_constructor_path,
    STATE(3906), 1,
      sym_attribute,
    STATE(6926), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4792), 8,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4794), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [153313] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3907), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5168), 13,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(5170), 14,
      anon_sym_let,
      anon_sym_external,
      anon_sym_type,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      sym__lowercase_identifier,
  [153355] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3908), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1881), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1879), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153397] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(79), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(81), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(2790), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(4240), 1,
      anon_sym_open,
    ACTIONS(4742), 1,
      anon_sym_external,
    ACTIONS(4744), 1,
      anon_sym_type,
    ACTIONS(4746), 1,
      anon_sym_exception,
    ACTIONS(4748), 1,
      anon_sym_module,
    ACTIONS(4750), 1,
      anon_sym_include,
    ACTIONS(4752), 1,
      anon_sym_class,
    ACTIONS(4754), 1,
      anon_sym_val,
    STATE(3909), 1,
      sym_attribute,
    STATE(4931), 1,
      aux_sym__structure_repeat1,
    STATE(4972), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4942), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [153465] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4917), 1,
      anon_sym_LPAREN,
    ACTIONS(5143), 1,
      sym__lowercase_identifier,
    STATE(3239), 1,
      sym__tuple_type_rhs,
    STATE(3835), 1,
      sym__simple_type,
    STATE(3910), 1,
      sym_attribute,
    STATE(4175), 1,
      sym_labeled_tuple_element_type,
    STATE(7174), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [153537] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4694), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4696), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [153579] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4917), 1,
      anon_sym_LPAREN,
    ACTIONS(5143), 1,
      sym__lowercase_identifier,
    STATE(3242), 1,
      sym__tuple_type_rhs,
    STATE(3835), 1,
      sym__simple_type,
    STATE(3912), 1,
      sym_attribute,
    STATE(4175), 1,
      sym_labeled_tuple_element_type,
    STATE(7174), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [153651] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3913), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3470), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [153693] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3914), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5174), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5172), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153735] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(5127), 1,
      sym__lowercase_identifier,
    STATE(3251), 1,
      sym__tuple_type_rhs,
    STATE(3915), 1,
      sym_attribute,
    STATE(4688), 1,
      sym__simple_type,
    STATE(4971), 1,
      sym_labeled_tuple_element_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [153807] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3916), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5178), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5176), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153849] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3917), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5182), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5180), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153891] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3918), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4604), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4606), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [153933] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3919), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5186), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5184), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153975] = 20,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4923), 1,
      anon_sym_LPAREN,
    ACTIONS(5125), 1,
      sym__lowercase_identifier,
    STATE(3239), 1,
      sym__tuple_type_rhs,
    STATE(3492), 1,
      sym__simple_type,
    STATE(3871), 1,
      sym_labeled_tuple_element_type,
    STATE(3920), 1,
      sym_attribute,
    STATE(6862), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [154047] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3921), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4620), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4622), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [154089] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4576), 1,
      anon_sym_COLON2,
    STATE(3922), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 11,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3470), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [154133] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3923), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4636), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4638), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [154175] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3924), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4662), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4664), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [154217] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4497), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4501), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [154259] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3926), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4676), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4678), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [154301] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4682), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4684), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [154343] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3928), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4600), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4602), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [154385] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4544), 1,
      anon_sym_COMMA,
    STATE(3929), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 18,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [154429] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3930), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3552), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3554), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [154471] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5141), 1,
      anon_sym_else,
    STATE(3502), 1,
      sym_else_clause,
    STATE(3931), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [154517] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3932), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4564), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4566), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [154559] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3933), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4612), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4614), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [154601] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4762), 1,
      anon_sym_DASH_GT,
    ACTIONS(4764), 1,
      anon_sym_STAR,
    ACTIONS(4961), 1,
      sym__lowercase_identifier,
    ACTIONS(5068), 1,
      anon_sym_POUND,
    STATE(3927), 1,
      sym_type_constructor_path,
    STATE(3934), 1,
      sym_attribute,
    STATE(6545), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6932), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4758), 6,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4760), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
  [154658] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5188), 1,
      anon_sym_PIPE,
    STATE(3935), 1,
      sym_attribute,
    STATE(3949), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3881), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3879), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [154703] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3936), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5192), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5190), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154744] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5188), 1,
      anon_sym_PIPE,
    STATE(3937), 1,
      sym_attribute,
    STATE(3942), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3874), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3872), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [154789] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3938), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5196), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5194), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154830] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5198), 1,
      anon_sym_else,
    STATE(2738), 1,
      sym_else_clause,
    STATE(3939), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [154875] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5064), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5062), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154916] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5202), 1,
      anon_sym_let,
    ACTIONS(5204), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3941), 2,
      sym_attribute,
      aux_sym_expression_item_repeat1,
    ACTIONS(5200), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154961] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5207), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3942), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3704), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [155004] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3943), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4981), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4979), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155045] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2199), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2201), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [155086] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5212), 1,
      anon_sym_let,
    ACTIONS(5214), 1,
      anon_sym_LBRACE,
    ACTIONS(5216), 1,
      anon_sym_DQUOTE,
    STATE(3945), 1,
      sym_attribute,
    STATE(4164), 1,
      aux_sym_external_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4371), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4388), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(5210), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155141] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5214), 1,
      anon_sym_LBRACE,
    ACTIONS(5216), 1,
      anon_sym_DQUOTE,
    ACTIONS(5220), 1,
      anon_sym_let,
    STATE(3946), 1,
      sym_attribute,
    STATE(4164), 1,
      aux_sym_external_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4400), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4388), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(5218), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155196] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5224), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5222), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155237] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4540), 1,
      anon_sym_DASH_GT,
    STATE(3948), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4437), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4435), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155280] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5188), 1,
      anon_sym_PIPE,
    STATE(3942), 1,
      aux_sym__match_cases_repeat1,
    STATE(3949), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [155325] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3554), 1,
      sym__lowercase_identifier,
    STATE(3950), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3552), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(4694), 8,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4696), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [155370] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5228), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5226), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155411] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4664), 1,
      sym__lowercase_identifier,
    STATE(3952), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4662), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(5230), 8,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(5232), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [155456] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4650), 1,
      anon_sym_COMMA,
    STATE(3953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 18,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [155499] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3954), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2149), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2151), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [155540] = 20,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5236), 1,
      anon_sym_let,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5240), 1,
      anon_sym_BANG,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5244), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(3955), 1,
      sym_attribute,
    STATE(4979), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7829), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [155611] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2195), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2197), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [155652] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3957), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2031), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2033), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [155693] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5248), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5246), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155734] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5198), 1,
      anon_sym_else,
    STATE(2746), 1,
      sym_else_clause,
    STATE(3959), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [155779] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4762), 1,
      anon_sym_DASH_GT,
    ACTIONS(4764), 1,
      anon_sym_STAR,
    ACTIONS(4961), 1,
      sym__lowercase_identifier,
    ACTIONS(5068), 1,
      anon_sym_POUND,
    STATE(3927), 1,
      sym_type_constructor_path,
    STATE(3960), 1,
      sym_attribute,
    STATE(6545), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6932), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4792), 6,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4794), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
  [155836] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5188), 1,
      anon_sym_PIPE,
    STATE(3937), 1,
      aux_sym__match_cases_repeat1,
    STATE(3961), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3868), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3866), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [155881] = 19,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    ACTIONS(4989), 1,
      sym__lowercase_identifier,
    STATE(3962), 1,
      sym_attribute,
    STATE(5541), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    STATE(6941), 1,
      sym_method_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [155950] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4728), 2,
      anon_sym_DOT,
      aux_sym_type_variable_token1,
    ACTIONS(4662), 10,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(4664), 14,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [155993] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3964), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5252), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5250), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156034] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5256), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5254), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156075] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5023), 1,
      anon_sym_as,
    ACTIONS(5260), 1,
      anon_sym_let,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(3966), 1,
      sym_attribute,
    STATE(4113), 1,
      aux_sym_type_binding_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4316), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5258), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156130] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2203), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2205), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [156171] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3968), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5264), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5262), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156212] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5023), 1,
      anon_sym_as,
    ACTIONS(5268), 1,
      anon_sym_let,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(3969), 1,
      sym_attribute,
    STATE(4048), 1,
      aux_sym_type_binding_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4294), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5266), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156267] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(3970), 1,
      sym_attribute,
    STATE(4313), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [156335] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(3971), 1,
      sym_attribute,
    STATE(4292), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [156403] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4730), 1,
      anon_sym_COMMA,
    STATE(3972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 16,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [156445] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5268), 1,
      anon_sym_let,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(3973), 1,
      sym_attribute,
    STATE(4048), 1,
      aux_sym_type_binding_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4294), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5266), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156497] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3974), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5288), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5286), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156537] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3975), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5292), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5290), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156577] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5296), 1,
      anon_sym_let,
    STATE(3856), 1,
      aux_sym_type_binding_repeat1,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(3976), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4251), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5294), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156629] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(3916), 1,
      sym__class_expression,
    STATE(3977), 1,
      sym_attribute,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [156697] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(3919), 1,
      sym__class_expression,
    STATE(3978), 1,
      sym_attribute,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [156765] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5300), 1,
      anon_sym_let,
    STATE(3856), 1,
      aux_sym_type_binding_repeat1,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(3979), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4243), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5298), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156817] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(3917), 1,
      sym__class_expression,
    STATE(3980), 1,
      sym_attribute,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [156885] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(3981), 1,
      sym_attribute,
    STATE(4267), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [156953] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(3982), 1,
      sym_attribute,
    STATE(4296), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [157021] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(3983), 1,
      sym_attribute,
    STATE(4304), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [157089] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3470), 1,
      sym__lowercase_identifier,
    ACTIONS(5129), 1,
      anon_sym_COLON2,
    STATE(3984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(4640), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4642), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [157135] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5308), 1,
      anon_sym_let,
    STATE(3856), 1,
      aux_sym_type_binding_repeat1,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(3985), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4319), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5306), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157187] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5236), 1,
      anon_sym_let,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5244), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(3916), 1,
      sym__class_expression,
    STATE(3986), 1,
      sym_attribute,
    STATE(7549), 1,
      sym_module_path,
    STATE(7829), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [157255] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5236), 1,
      anon_sym_let,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5244), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(3919), 1,
      sym__class_expression,
    STATE(3987), 1,
      sym_attribute,
    STATE(7549), 1,
      sym_module_path,
    STATE(7829), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [157323] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3988), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5312), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5310), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157363] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(3989), 1,
      sym_attribute,
    STATE(4290), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [157431] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5236), 1,
      anon_sym_let,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5244), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(3990), 1,
      sym_attribute,
    STATE(6383), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7829), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [157499] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5326), 1,
      anon_sym_and,
    STATE(3991), 1,
      sym_attribute,
    STATE(4052), 1,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5324), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5322), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157543] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5236), 1,
      anon_sym_let,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5244), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(3992), 1,
      sym_attribute,
    STATE(5008), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7829), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [157611] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(3993), 1,
      sym_attribute,
    STATE(5861), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [157679] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(3994), 1,
      sym_attribute,
    STATE(5882), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [157747] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(3995), 1,
      sym_attribute,
    STATE(4296), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [157815] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(3996), 1,
      sym_attribute,
    STATE(4324), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [157883] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(3997), 1,
      sym_attribute,
    STATE(4345), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [157951] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(3998), 1,
      sym_attribute,
    STATE(4334), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [158019] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(3999), 1,
      sym_attribute,
    STATE(4291), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [158087] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4000), 1,
      sym_attribute,
    STATE(6748), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [158155] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4921), 1,
      sym__lowercase_identifier,
    ACTIONS(4923), 1,
      anon_sym_LPAREN,
    STATE(3459), 1,
      sym__simple_type,
    STATE(4001), 1,
      sym_attribute,
    STATE(6862), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [158221] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4002), 1,
      sym_attribute,
    STATE(4295), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [158289] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2145), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2147), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [158329] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4004), 1,
      sym_attribute,
    STATE(4329), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [158397] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5121), 1,
      anon_sym_let,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(3976), 1,
      aux_sym_type_binding_repeat1,
    STATE(4005), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4216), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5119), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [158449] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4915), 1,
      sym__lowercase_identifier,
    ACTIONS(4917), 1,
      anon_sym_LPAREN,
    STATE(4006), 1,
      sym_attribute,
    STATE(4149), 1,
      sym__simple_type,
    STATE(7174), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [158515] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5334), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5332), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [158555] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4915), 1,
      sym__lowercase_identifier,
    ACTIONS(4917), 1,
      anon_sym_LPAREN,
    STATE(4008), 1,
      sym_attribute,
    STATE(4174), 1,
      sym__simple_type,
    STATE(7174), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [158621] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4009), 1,
      sym_attribute,
    STATE(4351), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [158689] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    STATE(4010), 1,
      sym_attribute,
    STATE(5218), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [158755] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4011), 1,
      sym_attribute,
    STATE(4292), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [158823] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4012), 1,
      sym_attribute,
    STATE(4304), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [158891] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4013), 1,
      sym_attribute,
    STATE(4226), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [158959] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4014), 1,
      sym_attribute,
    STATE(4208), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [159027] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4015), 1,
      sym_attribute,
    STATE(6490), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [159095] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4016), 1,
      sym_attribute,
    STATE(6502), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [159163] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4017), 1,
      sym_attribute,
    STATE(4334), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [159231] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4018), 1,
      sym_attribute,
    STATE(4329), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [159299] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4915), 1,
      sym__lowercase_identifier,
    ACTIONS(4917), 1,
      anon_sym_LPAREN,
    STATE(3847), 1,
      sym__simple_type,
    STATE(4019), 1,
      sym_attribute,
    STATE(7174), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [159365] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3854), 1,
      anon_sym_LT,
    ACTIONS(3856), 1,
      anon_sym_POUND,
    ACTIONS(3858), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3860), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3862), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(4915), 1,
      sym__lowercase_identifier,
    ACTIONS(4917), 1,
      anon_sym_LPAREN,
    STATE(4020), 1,
      sym_attribute,
    STATE(4187), 1,
      sym__simple_type,
    STATE(7174), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3924), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [159431] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4021), 1,
      sym_attribute,
    STATE(5951), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [159499] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4022), 1,
      sym_attribute,
    STATE(4290), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [159567] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4023), 1,
      sym_attribute,
    STATE(4291), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [159635] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4024), 1,
      sym_attribute,
    STATE(4295), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [159703] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4025), 1,
      sym_attribute,
    STATE(4351), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [159771] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4026), 1,
      sym_attribute,
    STATE(5954), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [159839] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4027), 1,
      sym_attribute,
    STATE(4321), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [159907] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4028), 1,
      sym_attribute,
    STATE(4254), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [159975] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4029), 1,
      sym_attribute,
    STATE(4313), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [160043] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4030), 1,
      sym_attribute,
    STATE(5959), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [160111] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4031), 1,
      sym_attribute,
    STATE(4322), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [160179] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    STATE(3494), 1,
      sym__simple_type,
    STATE(4032), 1,
      sym_attribute,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [160245] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(3917), 1,
      sym__class_expression,
    STATE(4033), 1,
      sym_attribute,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [160313] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4034), 1,
      sym_attribute,
    STATE(4208), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [160381] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4035), 1,
      sym_attribute,
    STATE(4246), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [160449] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(4036), 1,
      sym_attribute,
    STATE(4272), 1,
      sym__class_expression,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [160517] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4037), 1,
      sym_attribute,
    STATE(5788), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [160585] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4038), 1,
      sym_attribute,
    STATE(4267), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [160653] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4961), 1,
      sym__lowercase_identifier,
    ACTIONS(5068), 1,
      anon_sym_POUND,
    ACTIONS(5336), 1,
      anon_sym_STAR,
    STATE(3927), 1,
      sym_type_constructor_path,
    STATE(4039), 1,
      sym_attribute,
    STATE(4497), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6932), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4762), 6,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4955), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
  [160707] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4040), 1,
      sym_attribute,
    STATE(5790), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [160775] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4041), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5340), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5338), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160815] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5346), 1,
      anon_sym_EQ,
    ACTIONS(5348), 1,
      anon_sym_of,
    ACTIONS(5350), 1,
      anon_sym_COLON2,
    STATE(4042), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5344), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5342), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160861] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4043), 1,
      sym_attribute,
    STATE(5792), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [160929] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4044), 1,
      sym_attribute,
    STATE(5795), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [160997] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4045), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5354), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5352), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161037] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5358), 1,
      anon_sym_let,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4046), 1,
      sym_attribute,
    STATE(4090), 1,
      aux_sym_type_binding_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4262), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5356), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161089] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5362), 1,
      anon_sym_let,
    STATE(3856), 1,
      aux_sym_type_binding_repeat1,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4047), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4265), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5360), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161141] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5366), 1,
      anon_sym_let,
    STATE(3856), 1,
      aux_sym_type_binding_repeat1,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4048), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4266), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5364), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161193] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5370), 1,
      anon_sym_let,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4049), 1,
      sym_attribute,
    STATE(4092), 1,
      aux_sym_type_binding_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4268), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5368), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161245] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4921), 1,
      sym__lowercase_identifier,
    ACTIONS(4923), 1,
      anon_sym_LPAREN,
    STATE(3903), 1,
      sym__simple_type,
    STATE(4050), 1,
      sym_attribute,
    STATE(6862), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [161311] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2113), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161351] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5326), 1,
      anon_sym_and,
    STATE(4052), 1,
      sym_attribute,
    STATE(4096), 1,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5374), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5372), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161395] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4053), 1,
      sym_attribute,
    STATE(5815), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [161463] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5236), 1,
      anon_sym_let,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5244), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(3917), 1,
      sym__class_expression,
    STATE(4054), 1,
      sym_attribute,
    STATE(7549), 1,
      sym_module_path,
    STATE(7829), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [161531] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4055), 1,
      sym_attribute,
    STATE(4290), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [161599] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4056), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5378), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5376), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161639] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5236), 1,
      anon_sym_let,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5244), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4057), 1,
      sym_attribute,
    STATE(6519), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7829), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [161707] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(5380), 1,
      anon_sym_COLON2,
    STATE(4058), 1,
      sym_attribute,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4128), 1,
      aux_sym_module_binding_repeat1,
    STATE(4131), 1,
      sym__module_typed,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4323), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4939), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(4935), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [161763] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4059), 1,
      sym_attribute,
    STATE(5818), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [161831] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4060), 1,
      sym_attribute,
    STATE(5819), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [161899] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4061), 1,
      sym_attribute,
    STATE(4295), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [161967] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4062), 1,
      sym_attribute,
    STATE(5831), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [162035] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4921), 1,
      sym__lowercase_identifier,
    ACTIONS(4923), 1,
      anon_sym_LPAREN,
    STATE(3832), 1,
      sym__simple_type,
    STATE(4063), 1,
      sym_attribute,
    STATE(6862), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [162101] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4064), 1,
      sym_attribute,
    STATE(4272), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [162169] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3704), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3702), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [162209] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2600), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2602), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    STATE(3494), 1,
      sym__simple_type,
    STATE(4066), 1,
      sym_attribute,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [162275] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4943), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    STATE(4067), 1,
      sym_attribute,
    STATE(4652), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    ACTIONS(4941), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [162327] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4068), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3708), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3706), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [162367] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4069), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3712), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3710), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [162407] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5384), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5382), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [162447] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5388), 1,
      anon_sym_let,
    STATE(3856), 1,
      aux_sym_type_binding_repeat1,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4071), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4299), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5386), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [162499] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4072), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5392), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5390), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [162539] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4073), 1,
      sym_attribute,
    STATE(4321), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [162607] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4074), 1,
      sym_attribute,
    STATE(4226), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [162675] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    STATE(4075), 1,
      sym_attribute,
    STATE(5034), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [162741] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(3916), 1,
      sym__class_expression,
    STATE(4076), 1,
      sym_attribute,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [162809] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(3919), 1,
      sym__class_expression,
    STATE(4077), 1,
      sym_attribute,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [162877] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4078), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5396), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5394), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [162917] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4079), 1,
      sym_attribute,
    STATE(5985), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [162985] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5398), 1,
      anon_sym_else,
    STATE(2746), 1,
      sym_else_clause,
    STATE(4080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3915), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [163029] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4081), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5232), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5230), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [163069] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4082), 1,
      sym_attribute,
    STATE(5990), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [163137] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4083), 1,
      sym_attribute,
    STATE(5992), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [163205] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3786), 1,
      anon_sym_LT,
    ACTIONS(3788), 1,
      anon_sym_POUND,
    ACTIONS(3790), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3792), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(4921), 1,
      sym__lowercase_identifier,
    ACTIONS(4923), 1,
      anon_sym_LPAREN,
    STATE(3906), 1,
      sym__simple_type,
    STATE(4084), 1,
      sym_attribute,
    STATE(6862), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3544), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [163271] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4085), 1,
      sym_attribute,
    STATE(4267), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [163339] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4086), 1,
      sym_attribute,
    STATE(4246), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [163407] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5158), 1,
      anon_sym_let,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(3985), 1,
      aux_sym_type_binding_repeat1,
    STATE(4087), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4218), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5156), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [163459] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4088), 1,
      sym_attribute,
    STATE(4254), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [163527] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    STATE(4089), 1,
      sym_attribute,
    STATE(5215), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [163593] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5402), 1,
      anon_sym_let,
    STATE(3856), 1,
      aux_sym_type_binding_repeat1,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4090), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4335), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5400), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [163645] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5406), 1,
      anon_sym_let,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4091), 1,
      sym_attribute,
    STATE(4123), 1,
      aux_sym_type_binding_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4338), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5404), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [163697] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5410), 1,
      anon_sym_let,
    STATE(3856), 1,
      aux_sym_type_binding_repeat1,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4092), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4343), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5408), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [163749] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4093), 1,
      sym_attribute,
    STATE(5918), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [163817] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4094), 1,
      sym_attribute,
    STATE(4313), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [163885] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4095), 1,
      sym_attribute,
    STATE(5934), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [163953] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5412), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5256), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(4096), 2,
      sym_attribute,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(5254), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [163995] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4097), 1,
      sym_attribute,
    STATE(5922), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [164063] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5260), 1,
      anon_sym_let,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4098), 1,
      sym_attribute,
    STATE(4113), 1,
      aux_sym_type_binding_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4316), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5258), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164115] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(497), 1,
      anon_sym_object,
    ACTIONS(557), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(559), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5234), 1,
      sym__lowercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5242), 1,
      anon_sym_LBRACK,
    ACTIONS(5328), 1,
      anon_sym_let,
    ACTIONS(5330), 1,
      anon_sym_fun,
    STATE(1019), 1,
      sym__simple_class_expression,
    STATE(4099), 1,
      sym_attribute,
    STATE(5945), 1,
      sym__class_expression,
    STATE(7549), 1,
      sym_module_path,
    STATE(7841), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6444), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [164183] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4100), 1,
      sym_attribute,
    STATE(4296), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [164251] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4101), 1,
      sym_attribute,
    STATE(4324), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [164319] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4102), 1,
      sym_attribute,
    STATE(4324), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [164387] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4103), 1,
      sym_attribute,
    STATE(4345), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [164455] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4104), 1,
      sym_attribute,
    STATE(4351), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [164523] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4105), 1,
      sym_attribute,
    STATE(4321), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [164591] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4106), 1,
      sym_attribute,
    STATE(4254), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [164659] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(3917), 1,
      sym__class_expression,
    STATE(4107), 1,
      sym_attribute,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [164727] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4108), 1,
      sym_attribute,
    STATE(4329), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [164795] = 18,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2818), 1,
      anon_sym_LT,
    ACTIONS(2820), 1,
      anon_sym_POUND,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(4967), 1,
      anon_sym_LPAREN,
    STATE(4109), 1,
      sym_attribute,
    STATE(5677), 1,
      sym__simple_type,
    STATE(6901), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3757), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [164861] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4110), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5417), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5415), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164901] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5421), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5419), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164941] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4112), 1,
      sym_attribute,
    STATE(4322), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [165009] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5425), 1,
      anon_sym_let,
    STATE(3856), 1,
      aux_sym_type_binding_repeat1,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4113), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4349), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5423), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [165061] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4114), 1,
      sym_attribute,
    STATE(4322), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [165129] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4115), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3832), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3830), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [165169] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4116), 1,
      sym_attribute,
    STATE(4208), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [165237] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4117), 1,
      sym_attribute,
    STATE(4246), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [165305] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4118), 1,
      sym_attribute,
    STATE(4272), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [165373] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4119), 1,
      sym_attribute,
    STATE(4292), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [165441] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4120), 1,
      sym_attribute,
    STATE(4304), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [165509] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4121), 1,
      sym_attribute,
    STATE(4226), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [165577] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(3916), 1,
      sym__class_expression,
    STATE(4122), 1,
      sym_attribute,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [165645] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5429), 1,
      anon_sym_let,
    STATE(3856), 1,
      aux_sym_type_binding_repeat1,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4123), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4224), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5427), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [165697] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5398), 1,
      anon_sym_else,
    STATE(2738), 1,
      sym_else_clause,
    STATE(4124), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3728), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3726), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [165741] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5302), 1,
      anon_sym_let,
    ACTIONS(5304), 1,
      anon_sym_fun,
    STATE(927), 1,
      sym__simple_class_expression,
    STATE(3919), 1,
      sym__class_expression,
    STATE(4125), 1,
      sym_attribute,
    STATE(7218), 1,
      sym_value_definition,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [165809] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4126), 1,
      sym_attribute,
    STATE(4345), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [165877] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5021), 1,
      anon_sym_constraint,
    ACTIONS(5433), 1,
      anon_sym_let,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4047), 1,
      aux_sym_type_binding_repeat1,
    STATE(4127), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4285), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5431), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [165929] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(5380), 1,
      anon_sym_COLON2,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4128), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4161), 1,
      sym__module_typed,
    STATE(4257), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4904), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(4900), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [165985] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2223), 1,
      anon_sym_object,
    ACTIONS(3162), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3164), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5238), 1,
      anon_sym_LPAREN,
    ACTIONS(5314), 1,
      sym__lowercase_identifier,
    ACTIONS(5316), 1,
      anon_sym_let,
    ACTIONS(5318), 1,
      anon_sym_LBRACK,
    ACTIONS(5320), 1,
      anon_sym_fun,
    STATE(1013), 1,
      sym__simple_class_expression,
    STATE(4129), 1,
      sym_attribute,
    STATE(4334), 1,
      sym__class_expression,
    STATE(7289), 1,
      sym_value_definition,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2678), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [166053] = 19,
    ACTIONS(161), 1,
      sym_let_operator,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5270), 1,
      sym__lowercase_identifier,
    ACTIONS(5272), 1,
      anon_sym_let,
    ACTIONS(5274), 1,
      anon_sym_LPAREN,
    ACTIONS(5276), 1,
      anon_sym_LBRACK,
    ACTIONS(5278), 1,
      anon_sym_object,
    ACTIONS(5280), 1,
      anon_sym_fun,
    ACTIONS(5282), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5284), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(931), 1,
      sym__simple_class_expression,
    STATE(4130), 1,
      sym_attribute,
    STATE(4291), 1,
      sym__class_expression,
    STATE(7340), 1,
      sym_module_path,
    STATE(7754), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3908), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2891), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [166121] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5437), 1,
      anon_sym_let,
    ACTIONS(5439), 1,
      anon_sym_EQ,
    ACTIONS(5441), 1,
      anon_sym_COLON_EQ,
    STATE(4131), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4223), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5435), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [166170] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4132), 1,
      sym_attribute,
    STATE(4151), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5445), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5443), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [166215] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5451), 1,
      anon_sym_DASH_GT,
    ACTIONS(5453), 1,
      anon_sym_with,
    STATE(4133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5449), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5447), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [166258] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5451), 1,
      anon_sym_DASH_GT,
    ACTIONS(5453), 1,
      anon_sym_with,
    STATE(4134), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5457), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5455), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [166301] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4664), 1,
      sym__lowercase_identifier,
    STATE(4135), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4662), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(5230), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5232), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [166344] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5023), 1,
      anon_sym_as,
    ACTIONS(5102), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4136), 1,
      sym_attribute,
    STATE(4182), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5100), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [166391] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5019), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5023), 1,
      anon_sym_as,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4137), 1,
      sym_attribute,
    STATE(4184), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5017), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [166438] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5023), 1,
      anon_sym_as,
    ACTIONS(5043), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4138), 1,
      sym_attribute,
    STATE(4185), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5041), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [166485] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4957), 1,
      anon_sym_STAR,
    STATE(4139), 1,
      sym_attribute,
    STATE(4144), 1,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5463), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5461), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [166528] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4140), 1,
      sym_attribute,
    STATE(4141), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5467), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5465), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [166573] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4141), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5471), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5469), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [166618] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4931), 1,
      anon_sym_COMMA,
    STATE(4142), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3346), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3344), 16,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [166659] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3468), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    ACTIONS(5473), 1,
      anon_sym_EQ,
    ACTIONS(5475), 1,
      anon_sym_COLON_EQ,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4143), 1,
      sym_attribute,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4339), 1,
      aux_sym_expression_item_repeat1,
    STATE(4407), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5025), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [166714] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5477), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5060), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(4144), 2,
      sym_attribute,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(5058), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [166755] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4145), 1,
      sym_attribute,
    STATE(4150), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5482), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5480), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [166800] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4146), 1,
      sym_attribute,
    STATE(4153), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5486), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5484), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [166845] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4147), 1,
      sym_attribute,
    STATE(4155), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5490), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5488), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [166890] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5451), 1,
      anon_sym_DASH_GT,
    ACTIONS(5453), 1,
      anon_sym_with,
    STATE(4148), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5494), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5492), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [166933] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4961), 1,
      sym__lowercase_identifier,
    ACTIONS(5068), 1,
      anon_sym_POUND,
    STATE(3927), 1,
      sym_type_constructor_path,
    STATE(4149), 1,
      sym_attribute,
    STATE(6932), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5058), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5060), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
  [166982] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4150), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5498), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5496), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [167027] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4151), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5502), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5500), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [167072] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5451), 1,
      anon_sym_DASH_GT,
    ACTIONS(5453), 1,
      anon_sym_with,
    STATE(4152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5506), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5504), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [167115] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4153), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5510), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5508), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [167160] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4154), 1,
      sym_attribute,
    STATE(4167), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5514), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5512), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [167205] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4155), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5518), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5516), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [167250] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4156), 1,
      sym_attribute,
    STATE(4168), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5522), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5520), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [167295] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4157), 1,
      sym_attribute,
    STATE(4170), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5526), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5524), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [167340] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5451), 1,
      anon_sym_DASH_GT,
    ACTIONS(5453), 1,
      anon_sym_with,
    STATE(4158), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5530), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5528), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [167383] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4159), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5534), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5532), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [167422] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5538), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5536), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [167461] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5542), 1,
      anon_sym_let,
    ACTIONS(5544), 1,
      anon_sym_EQ,
    ACTIONS(5546), 1,
      anon_sym_COLON_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4161), 1,
      sym_attribute,
    STATE(4238), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5540), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [167510] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4162), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5550), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5548), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [167549] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4163), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2157), 2,
      anon_sym_let,
      anon_sym_COLON2,
    ACTIONS(2159), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [167588] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5556), 1,
      anon_sym_LBRACE,
    ACTIONS(5559), 1,
      anon_sym_DQUOTE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5554), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(4164), 2,
      sym_attribute,
      aux_sym_external_repeat1,
    STATE(4388), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(5552), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [167633] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3468), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    ACTIONS(5562), 1,
      anon_sym_EQ,
    ACTIONS(5564), 1,
      anon_sym_COLON_EQ,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4165), 1,
      sym_attribute,
    STATE(4256), 1,
      aux_sym_expression_item_repeat1,
    STATE(4417), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5072), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [167688] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3554), 1,
      sym__lowercase_identifier,
    STATE(4166), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3552), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(4694), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4696), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [167731] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4167), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5568), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5566), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [167776] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4168), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5572), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5570), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [167821] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4169), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5576), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5574), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [167866] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4170), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5580), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5578), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [167911] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5582), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4337), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4171), 2,
      sym_attribute,
      aux_sym_expression_item_repeat1,
    ACTIONS(5202), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5200), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [167954] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5023), 1,
      anon_sym_as,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    ACTIONS(5585), 1,
      anon_sym_EQ,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4172), 1,
      sym_attribute,
    STATE(4216), 1,
      aux_sym_expression_item_repeat1,
    STATE(4427), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5119), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [168006] = 19,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5591), 1,
      anon_sym_RPAREN,
    ACTIONS(5593), 1,
      anon_sym__,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5603), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4173), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6546), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [168072] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4961), 1,
      sym__lowercase_identifier,
    ACTIONS(5068), 1,
      anon_sym_POUND,
    STATE(3927), 1,
      sym_type_constructor_path,
    STATE(4174), 1,
      sym_attribute,
    STATE(6932), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4792), 6,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4794), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
  [168120] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4558), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5066), 1,
      anon_sym_STAR,
    STATE(4175), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4556), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [168160] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5114), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5605), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4176), 2,
      sym_attribute,
      aux_sym_type_binding_repeat1,
    ACTIONS(5112), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [168202] = 19,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5610), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4177), 1,
      sym_attribute,
    STATE(4440), 1,
      sym__attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6556), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [168268] = 19,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5591), 1,
      anon_sym_RPAREN,
    ACTIONS(5593), 1,
      anon_sym__,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5603), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4178), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6333), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [168334] = 19,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5612), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4179), 1,
      sym_attribute,
    STATE(4392), 1,
      sym__attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6400), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [168400] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4180), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4955), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4762), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [168438] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4181), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2191), 10,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(2193), 13,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [168476] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5154), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4176), 1,
      aux_sym_type_binding_repeat1,
    STATE(4182), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5152), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [168520] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5618), 1,
      anon_sym_PIPE,
    STATE(4183), 1,
      sym_attribute,
    STATE(4193), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5616), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5614), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [168562] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5135), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4176), 1,
      aux_sym_type_binding_repeat1,
    STATE(4184), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5133), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [168606] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5139), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4176), 1,
      aux_sym_type_binding_repeat1,
    STATE(4185), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5137), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [168650] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5620), 1,
      anon_sym_COLON2,
    STATE(4186), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3470), 6,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3468), 16,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      sym__uppercase_identifier,
  [168690] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4961), 1,
      sym__lowercase_identifier,
    ACTIONS(5068), 1,
      anon_sym_POUND,
    STATE(3927), 1,
      sym_type_constructor_path,
    STATE(4187), 1,
      sym_attribute,
    STATE(6932), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5164), 6,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5166), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
  [168738] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4188), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2187), 10,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(2189), 13,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [168776] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5451), 1,
      anon_sym_DASH_GT,
    ACTIONS(5453), 1,
      anon_sym_with,
    ACTIONS(5624), 1,
      anon_sym_let,
    STATE(4189), 1,
      sym_attribute,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4404), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5622), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [168824] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4190), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2165), 10,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(2167), 13,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [168862] = 19,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5626), 1,
      anon_sym_DASH_GT,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5630), 1,
      anon_sym_functor,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4191), 1,
      sym_attribute,
    STATE(4359), 1,
      sym__attribute,
    STATE(4689), 1,
      sym__module_type,
    STATE(6293), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [168928] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5451), 1,
      anon_sym_DASH_GT,
    ACTIONS(5453), 1,
      anon_sym_with,
    ACTIONS(5640), 1,
      anon_sym_let,
    STATE(4192), 1,
      sym_attribute,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4387), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5638), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [168976] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5618), 1,
      anon_sym_PIPE,
    STATE(4193), 1,
      sym_attribute,
    STATE(4195), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5644), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5642), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169018] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4194), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5648), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5646), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169056] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5654), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5652), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(4195), 2,
      sym_attribute,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5650), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169096] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4196), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2183), 10,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(2185), 13,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [169134] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5618), 1,
      anon_sym_PIPE,
    STATE(4197), 1,
      sym_attribute,
    STATE(4202), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4943), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4941), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169176] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5659), 1,
      anon_sym_let,
    STATE(4198), 1,
      sym_attribute,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4456), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5661), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(5657), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169222] = 19,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5663), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4199), 1,
      sym_attribute,
    STATE(4382), 1,
      sym__attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6415), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [169288] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5023), 1,
      anon_sym_as,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    ACTIONS(5665), 1,
      anon_sym_EQ,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4200), 1,
      sym_attribute,
    STATE(4218), 1,
      aux_sym_expression_item_repeat1,
    STATE(4420), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5156), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [169340] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5669), 1,
      anon_sym_let,
    STATE(4201), 1,
      sym_attribute,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4472), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5671), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(5667), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169386] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5618), 1,
      anon_sym_PIPE,
    STATE(4195), 1,
      aux_sym_variant_declaration_repeat1,
    STATE(4202), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5616), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5614), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169428] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4203), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5675), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5673), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169466] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4204), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5679), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5677), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169504] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5683), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4205), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5681), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169547] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5687), 1,
      anon_sym_let,
    ACTIONS(5689), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4206), 1,
      sym_attribute,
    STATE(4252), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5685), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169592] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5693), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4207), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5691), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169635] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5697), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4208), 1,
      sym_attribute,
    STATE(4354), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5695), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169678] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5699), 1,
      anon_sym_LPAREN,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5705), 1,
      anon_sym_end,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    STATE(4209), 1,
      sym_attribute,
    STATE(4532), 1,
      sym__attribute,
    STATE(4623), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169741] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5719), 1,
      anon_sym_let,
    ACTIONS(5721), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4210), 1,
      sym_attribute,
    STATE(4212), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5717), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169786] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5725), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4211), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5723), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169829] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5729), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4212), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5727), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169872] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5733), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4213), 1,
      sym_attribute,
    STATE(4342), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5731), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169915] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5737), 1,
      anon_sym_end,
    STATE(4214), 1,
      sym_attribute,
    STATE(4521), 1,
      sym__attribute,
    STATE(4611), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169978] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4215), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5741), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5739), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170015] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5296), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4216), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5294), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170058] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4217), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5534), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5532), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [170095] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5308), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4218), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5306), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170138] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5745), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4219), 1,
      sym_attribute,
    STATE(4275), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5743), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170181] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5749), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4220), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5747), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170224] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5751), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4221), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
    ACTIONS(3186), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3188), 12,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [170263] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4222), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5756), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5754), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170300] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5760), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4223), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5758), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170343] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5764), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5762), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170386] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5768), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5766), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170429] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5772), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4226), 1,
      sym_attribute,
    STATE(4344), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5770), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170472] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5776), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5774), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170515] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5780), 1,
      anon_sym_let,
    ACTIONS(5782), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4228), 1,
      sym_attribute,
    STATE(4314), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5778), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170560] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5550), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5548), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [170597] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5786), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5784), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170640] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5790), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4231), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5788), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170683] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5792), 1,
      anon_sym_LPAREN,
    ACTIONS(5794), 1,
      anon_sym_end,
    STATE(4232), 1,
      sym_attribute,
    STATE(4537), 1,
      sym__attribute,
    STATE(4772), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170746] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5796), 1,
      anon_sym_LPAREN,
    ACTIONS(5798), 1,
      anon_sym_end,
    STATE(4233), 1,
      sym_attribute,
    STATE(4519), 1,
      sym__attribute,
    STATE(4642), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170809] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5802), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5800), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170852] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5804), 1,
      anon_sym_LPAREN,
    ACTIONS(5806), 1,
      anon_sym_end,
    STATE(4235), 1,
      sym_attribute,
    STATE(4511), 1,
      sym__attribute,
    STATE(4721), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170915] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5810), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5808), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170958] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5814), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4237), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5812), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171001] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5818), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4238), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5816), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171044] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5820), 1,
      anon_sym_LPAREN,
    ACTIONS(5822), 1,
      anon_sym_end,
    STATE(4239), 1,
      sym_attribute,
    STATE(4498), 1,
      sym__attribute,
    STATE(4726), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171107] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5826), 1,
      anon_sym_let,
    ACTIONS(5828), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4240), 1,
      sym_attribute,
    STATE(4348), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5824), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171152] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2393), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4241), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2391), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171195] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5832), 1,
      anon_sym_let,
    STATE(4242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5830), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171232] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5836), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4243), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5834), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171275] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5838), 1,
      anon_sym_LPAREN,
    ACTIONS(5840), 1,
      anon_sym_end,
    STATE(4244), 1,
      sym_attribute,
    STATE(4526), 1,
      sym__attribute,
    STATE(4746), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171338] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5844), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5842), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171381] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5848), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4207), 1,
      aux_sym_expression_item_repeat1,
    STATE(4246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5846), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171424] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5852), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5850), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171467] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4760), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4758), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171504] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2421), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4249), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2419), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171547] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5856), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4234), 1,
      aux_sym_expression_item_repeat1,
    STATE(4250), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5854), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171590] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5860), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5858), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171633] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5864), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5862), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171676] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5868), 1,
      anon_sym_let,
    ACTIONS(5870), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4253), 1,
      sym_attribute,
    STATE(4293), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5866), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171721] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5874), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4254), 1,
      sym_attribute,
    STATE(4263), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5872), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171764] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5878), 1,
      anon_sym_let,
    ACTIONS(5880), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4255), 1,
      sym_attribute,
    STATE(4298), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5876), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171809] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5300), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4256), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5298), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171852] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5884), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4257), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5882), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171895] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5888), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4258), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5886), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171938] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5892), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4258), 1,
      aux_sym_expression_item_repeat1,
    STATE(4259), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5890), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171981] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5896), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4260), 1,
      sym_attribute,
    STATE(4318), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5894), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172024] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5900), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4261), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5898), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172067] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5402), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4262), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5400), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172110] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5904), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5902), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172153] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5908), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4264), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5906), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172196] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5912), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5910), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172239] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5916), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5914), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172282] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5920), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4237), 1,
      aux_sym_expression_item_repeat1,
    STATE(4267), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5918), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172325] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5410), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5408), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172368] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5924), 1,
      anon_sym_let,
    ACTIONS(5926), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4269), 1,
      sym_attribute,
    STATE(4306), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5922), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172413] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5928), 1,
      anon_sym_LPAREN,
    ACTIONS(5930), 1,
      anon_sym_end,
    STATE(4270), 1,
      sym_attribute,
    STATE(4514), 1,
      sym__attribute,
    STATE(4576), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [172476] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5934), 1,
      anon_sym_let,
    ACTIONS(5936), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4247), 1,
      aux_sym_expression_item_repeat1,
    STATE(4271), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5932), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172521] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5940), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4245), 1,
      aux_sym_expression_item_repeat1,
    STATE(4272), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5938), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172564] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5023), 1,
      anon_sym_as,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4273), 1,
      sym_attribute,
    STATE(4316), 1,
      aux_sym_expression_item_repeat1,
    STATE(4424), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5258), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [172613] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(89), 1,
      anon_sym_DQUOTE,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5942), 1,
      anon_sym_LBRACE,
    STATE(4274), 1,
      sym_attribute,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4400), 1,
      aux_sym_expression_item_repeat1,
    STATE(4548), 1,
      aux_sym_external_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4388), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(5218), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [172662] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5946), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4275), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5944), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172705] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5950), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5948), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172748] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5954), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4277), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5952), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172791] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5950), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5948), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172834] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5958), 1,
      anon_sym_let,
    ACTIONS(5960), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4236), 1,
      aux_sym_expression_item_repeat1,
    STATE(4279), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5956), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172879] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5964), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4280), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5962), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172922] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4281), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5968), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5966), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172959] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5023), 1,
      anon_sym_as,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4282), 1,
      sym_attribute,
    STATE(4294), 1,
      aux_sym_expression_item_repeat1,
    STATE(4431), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5266), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [173008] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5970), 1,
      anon_sym_LPAREN,
    ACTIONS(5972), 1,
      anon_sym_end,
    STATE(4283), 1,
      sym_attribute,
    STATE(4547), 1,
      sym__attribute,
    STATE(4577), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [173071] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4284), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4048), 2,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4050), 20,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [173108] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5362), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5360), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173151] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5974), 1,
      anon_sym_LPAREN,
    ACTIONS(5976), 1,
      anon_sym_end,
    STATE(4286), 1,
      sym_attribute,
    STATE(4507), 1,
      sym__attribute,
    STATE(4654), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [173214] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5980), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4287), 1,
      sym_attribute,
    STATE(4352), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5978), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173257] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5984), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5982), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173300] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5988), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4289), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5986), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173343] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5992), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4264), 1,
      aux_sym_expression_item_repeat1,
    STATE(4290), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5990), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173386] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5996), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4291), 1,
      sym_attribute,
    STATE(4301), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5994), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173429] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6000), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4292), 1,
      sym_attribute,
    STATE(4325), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5998), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173472] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6004), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4293), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6002), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173515] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5366), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4294), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5364), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173558] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6008), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4220), 1,
      aux_sym_expression_item_repeat1,
    STATE(4295), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6006), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173601] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6012), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4289), 1,
      aux_sym_expression_item_repeat1,
    STATE(4296), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6010), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173644] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6016), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6014), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173687] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6020), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4298), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6018), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173730] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6024), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4299), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6022), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173773] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6028), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4300), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6026), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173816] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6032), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4301), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6030), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173859] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2405), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4302), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2403), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173902] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6034), 1,
      anon_sym_LPAREN,
    ACTIONS(6036), 1,
      anon_sym_end,
    STATE(4303), 1,
      sym_attribute,
    STATE(4553), 1,
      sym__attribute,
    STATE(4603), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [173965] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6040), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4304), 1,
      sym_attribute,
    STATE(4326), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6038), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174008] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6044), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6042), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174051] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6048), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4306), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6046), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174094] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4307), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6052), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6050), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174131] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6054), 1,
      aux_sym_attribute_id_token1,
    STATE(4221), 1,
      aux_sym_attribute_id_repeat1,
    STATE(4308), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3174), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3176), 12,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [174172] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6058), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4309), 1,
      sym_attribute,
    STATE(4317), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6056), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174215] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(89), 1,
      anon_sym_DQUOTE,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5942), 1,
      anon_sym_LBRACE,
    STATE(4310), 1,
      sym_attribute,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4371), 1,
      aux_sym_expression_item_repeat1,
    STATE(4548), 1,
      aux_sym_external_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4388), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(5210), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [174264] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6062), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4311), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6060), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174307] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6062), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6060), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174350] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6066), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4313), 1,
      sym_attribute,
    STATE(4356), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6064), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174393] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6070), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6068), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174436] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6072), 1,
      anon_sym_LPAREN,
    ACTIONS(6074), 1,
      anon_sym_end,
    STATE(4315), 1,
      sym_attribute,
    STATE(4531), 1,
      sym__attribute,
    STATE(4558), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [174499] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5425), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4316), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5423), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174542] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6078), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4317), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6076), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174585] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6082), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4318), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6080), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174628] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6086), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6084), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174671] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6054), 1,
      aux_sym_attribute_id_token1,
    STATE(4308), 1,
      aux_sym_attribute_id_repeat1,
    STATE(4320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3166), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3168), 12,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [174712] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6090), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4321), 1,
      sym_attribute,
    STATE(4328), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6088), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174755] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6094), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4227), 1,
      aux_sym_expression_item_repeat1,
    STATE(4322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6092), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174798] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4902), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4323), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4900), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174841] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6098), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4324), 1,
      sym_attribute,
    STATE(4341), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6096), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174884] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6102), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4325), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6100), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174927] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6106), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6104), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174970] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5652), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5650), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175007] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6110), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4328), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6108), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175050] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6114), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4305), 1,
      aux_sym_expression_item_repeat1,
    STATE(4329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6112), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175093] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6116), 1,
      anon_sym_LPAREN,
    ACTIONS(6118), 1,
      anon_sym_end,
    STATE(4330), 1,
      sym_attribute,
    STATE(4496), 1,
      sym__attribute,
    STATE(4696), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [175156] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6120), 1,
      anon_sym_LPAREN,
    ACTIONS(6122), 1,
      anon_sym_end,
    STATE(4331), 1,
      sym_attribute,
    STATE(4499), 1,
      sym__attribute,
    STATE(4661), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [175219] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6126), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4332), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6124), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175262] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6130), 1,
      anon_sym_let,
    ACTIONS(6132), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4333), 1,
      sym_attribute,
    STATE(4346), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6128), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175307] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6136), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4288), 1,
      aux_sym_expression_item_repeat1,
    STATE(4334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6134), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175350] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6140), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4335), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6138), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175393] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6144), 1,
      anon_sym_let,
    STATE(4336), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6142), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175430] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4337), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5538), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5536), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [175467] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5429), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4338), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5427), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175510] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5388), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4339), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5386), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175553] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2387), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4340), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2385), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175596] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6148), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4341), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6146), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175639] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6152), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4342), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6150), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175682] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6156), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4343), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6154), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175725] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6160), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4344), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6158), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175768] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6164), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4225), 1,
      aux_sym_expression_item_repeat1,
    STATE(4345), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6162), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175811] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6168), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4346), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6166), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175854] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2411), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2409), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175897] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6172), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4348), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6170), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175940] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6176), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4349), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6174), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175983] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6180), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4261), 1,
      aux_sym_expression_item_repeat1,
    STATE(4350), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6178), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [176026] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6184), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4231), 1,
      aux_sym_expression_item_repeat1,
    STATE(4351), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6182), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [176069] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6188), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4352), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6186), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [176112] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6192), 1,
      anon_sym_let,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4277), 1,
      aux_sym_expression_item_repeat1,
    STATE(4353), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6190), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [176155] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6196), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4354), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6194), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [176198] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4437), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5096), 1,
      anon_sym_DASH_GT,
    STATE(4355), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4435), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176237] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6200), 1,
      anon_sym_let,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4356), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6198), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [176280] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5626), 1,
      anon_sym_DASH_GT,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5630), 1,
      anon_sym_functor,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4357), 1,
      sym_attribute,
    STATE(4534), 1,
      sym__module_type,
    STATE(6293), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [176340] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5626), 1,
      anon_sym_DASH_GT,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5630), 1,
      anon_sym_functor,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(3968), 1,
      sym__module_type,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4358), 1,
      sym_attribute,
    STATE(6293), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [176400] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5626), 1,
      anon_sym_DASH_GT,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5630), 1,
      anon_sym_functor,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4359), 1,
      sym_attribute,
    STATE(4653), 1,
      sym__module_type,
    STATE(6293), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [176460] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4360), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3186), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3188), 13,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
      aux_sym_attribute_id_token1,
  [176496] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6204), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4361), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6202), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [176538] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4362), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2145), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(2147), 19,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
  [176574] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6208), 1,
      anon_sym_let,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4363), 1,
      sym_attribute,
    STATE(4397), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6206), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [176616] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6208), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4364), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6206), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [176658] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6210), 1,
      sym__lowercase_identifier,
    ACTIONS(6212), 1,
      anon_sym_LPAREN,
    ACTIONS(6214), 1,
      anon_sym_DASH_GT,
    ACTIONS(6216), 1,
      anon_sym_module,
    ACTIONS(6218), 1,
      anon_sym_sig,
    ACTIONS(6220), 1,
      anon_sym_functor,
    ACTIONS(6222), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4365), 1,
      sym_attribute,
    STATE(5755), 1,
      sym__module_type,
    STATE(6169), 1,
      aux_sym_module_binding_repeat1,
    STATE(7119), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5495), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [176718] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6210), 1,
      sym__lowercase_identifier,
    ACTIONS(6212), 1,
      anon_sym_LPAREN,
    ACTIONS(6218), 1,
      anon_sym_sig,
    ACTIONS(6222), 1,
      sym__uppercase_identifier,
    ACTIONS(6224), 1,
      anon_sym_DASH_GT,
    ACTIONS(6226), 1,
      anon_sym_module,
    ACTIONS(6228), 1,
      anon_sym_functor,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4366), 1,
      sym_attribute,
    STATE(5524), 1,
      sym__module_type,
    STATE(6134), 1,
      aux_sym_module_binding_repeat1,
    STATE(7119), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5495), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [176778] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4367), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2153), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(2155), 18,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
  [176814] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4368), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2133), 3,
      anon_sym_let,
      anon_sym_LBRACE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2135), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_DQUOTE,
      sym_let_operator,
      aux_sym_directive_token1,
  [176850] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5780), 1,
      anon_sym_let,
    ACTIONS(6234), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4314), 1,
      aux_sym_expression_item_repeat1,
    STATE(4369), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5778), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [176894] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4370), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2173), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(2175), 18,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
  [176930] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6238), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4371), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6236), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [176972] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5687), 1,
      anon_sym_let,
    ACTIONS(6240), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4252), 1,
      aux_sym_expression_item_repeat1,
    STATE(4372), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5685), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [177016] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6244), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4373), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6242), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [177058] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6130), 1,
      anon_sym_let,
    ACTIONS(6246), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4346), 1,
      aux_sym_expression_item_repeat1,
    STATE(4374), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6128), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [177102] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5344), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5346), 1,
      anon_sym_EQ,
    ACTIONS(6248), 1,
      anon_sym_of,
    ACTIONS(6250), 1,
      anon_sym_COLON2,
    STATE(4375), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5342), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [177144] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6256), 1,
      anon_sym_and,
    ACTIONS(6258), 1,
      sym_let_and_operator,
    STATE(4376), 1,
      sym_attribute,
    STATE(4378), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6254), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(6252), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [177186] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6256), 1,
      anon_sym_and,
    ACTIONS(6258), 1,
      sym_let_and_operator,
    STATE(4377), 1,
      sym_attribute,
    STATE(4378), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6262), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(6260), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [177228] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6268), 1,
      anon_sym_and,
    ACTIONS(6271), 1,
      sym_let_and_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6266), 2,
      anon_sym_let,
      anon_sym_in,
    STATE(4378), 2,
      sym_attribute,
      aux_sym_value_definition_repeat1,
    ACTIONS(6264), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [177268] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6256), 1,
      anon_sym_and,
    ACTIONS(6258), 1,
      sym_let_and_operator,
    STATE(4379), 1,
      sym_attribute,
    STATE(4381), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6262), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(6260), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [177310] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6244), 1,
      anon_sym_let,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4380), 1,
      sym_attribute,
    STATE(4416), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6242), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [177352] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6256), 1,
      anon_sym_and,
    ACTIONS(6258), 1,
      sym_let_and_operator,
    STATE(4378), 1,
      aux_sym_value_definition_repeat1,
    STATE(4381), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6276), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(6274), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [177394] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(6278), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4382), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6578), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [177454] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6282), 1,
      anon_sym_let,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4383), 1,
      sym_attribute,
    STATE(4475), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6280), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [177496] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4384), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6751), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [177556] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6286), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4385), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6284), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [177598] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4386), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6338), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [177658] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6290), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4387), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6288), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [177700] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4388), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6294), 3,
      anon_sym_let,
      anon_sym_LBRACE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6292), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_DQUOTE,
      sym_let_operator,
      aux_sym_directive_token1,
  [177736] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(6296), 1,
      anon_sym_DASH_GT,
    ACTIONS(6298), 1,
      anon_sym_functor,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4158), 1,
      sym__module_type,
    STATE(4389), 1,
      sym_attribute,
    STATE(6193), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [177796] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6302), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6300), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [177832] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(6296), 1,
      anon_sym_DASH_GT,
    ACTIONS(6298), 1,
      anon_sym_functor,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4133), 1,
      sym__module_type,
    STATE(4391), 1,
      sym_attribute,
    STATE(6193), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [177892] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(6304), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4392), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6397), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [177952] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4393), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6407), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [178012] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4394), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2153), 3,
      anon_sym_let,
      anon_sym_LBRACE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2155), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_DQUOTE,
      sym_let_operator,
      aux_sym_directive_token1,
  [178048] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5826), 1,
      anon_sym_let,
    ACTIONS(6306), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4348), 1,
      aux_sym_expression_item_repeat1,
    STATE(4395), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5824), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [178092] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4396), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6458), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [178152] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6310), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4397), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6308), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [178194] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4398), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2173), 3,
      anon_sym_let,
      anon_sym_LBRACE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2175), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_DQUOTE,
      sym_let_operator,
      aux_sym_directive_token1,
  [178230] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6314), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4399), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6312), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [178272] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6318), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4400), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6316), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [178314] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6210), 1,
      sym__lowercase_identifier,
    ACTIONS(6212), 1,
      anon_sym_LPAREN,
    ACTIONS(6214), 1,
      anon_sym_DASH_GT,
    ACTIONS(6216), 1,
      anon_sym_module,
    ACTIONS(6218), 1,
      anon_sym_sig,
    ACTIONS(6220), 1,
      anon_sym_functor,
    ACTIONS(6222), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4401), 1,
      sym_attribute,
    STATE(5662), 1,
      sym__module_type,
    STATE(6169), 1,
      aux_sym_module_binding_repeat1,
    STATE(7119), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5495), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [178374] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4402), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2133), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2135), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [178410] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6210), 1,
      sym__lowercase_identifier,
    ACTIONS(6212), 1,
      anon_sym_LPAREN,
    ACTIONS(6218), 1,
      anon_sym_sig,
    ACTIONS(6222), 1,
      sym__uppercase_identifier,
    ACTIONS(6224), 1,
      anon_sym_DASH_GT,
    ACTIONS(6226), 1,
      anon_sym_module,
    ACTIONS(6228), 1,
      anon_sym_functor,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4403), 1,
      sym_attribute,
    STATE(5512), 1,
      sym__module_type,
    STATE(6134), 1,
      aux_sym_module_binding_repeat1,
    STATE(7119), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5495), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [178470] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6322), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4404), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6320), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [178512] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6210), 1,
      sym__lowercase_identifier,
    ACTIONS(6212), 1,
      anon_sym_LPAREN,
    ACTIONS(6218), 1,
      anon_sym_sig,
    ACTIONS(6222), 1,
      sym__uppercase_identifier,
    ACTIONS(6224), 1,
      anon_sym_DASH_GT,
    ACTIONS(6226), 1,
      anon_sym_module,
    ACTIONS(6228), 1,
      anon_sym_functor,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4405), 1,
      sym_attribute,
    STATE(5513), 1,
      sym__module_type,
    STATE(6134), 1,
      aux_sym_module_binding_repeat1,
    STATE(7119), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5495), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [178572] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4406), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4696), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4694), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [178608] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4176), 1,
      aux_sym_type_binding_repeat1,
    STATE(4299), 1,
      aux_sym_expression_item_repeat1,
    STATE(4407), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5386), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [178654] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4408), 1,
      sym_attribute,
    STATE(6257), 1,
      sym__module_type,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [178714] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5868), 1,
      anon_sym_let,
    ACTIONS(6324), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4293), 1,
      aux_sym_expression_item_repeat1,
    STATE(4409), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5866), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [178758] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6328), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4410), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6326), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [178800] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5878), 1,
      anon_sym_let,
    ACTIONS(6330), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4298), 1,
      aux_sym_expression_item_repeat1,
    STATE(4411), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5876), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [178844] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6334), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4412), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6332), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [178886] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4413), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2153), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2155), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [178922] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4218), 1,
      aux_sym_expression_item_repeat1,
    STATE(4414), 1,
      sym_attribute,
    STATE(4420), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5156), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [178968] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6334), 1,
      anon_sym_let,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4361), 1,
      aux_sym_expression_item_repeat1,
    STATE(4415), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6332), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [179010] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6338), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4416), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6336), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [179052] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4176), 1,
      aux_sym_type_binding_repeat1,
    STATE(4243), 1,
      aux_sym_expression_item_repeat1,
    STATE(4417), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5298), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179098] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4418), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6259), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [179158] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4316), 1,
      aux_sym_expression_item_repeat1,
    STATE(4419), 1,
      sym_attribute,
    STATE(4424), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5258), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179204] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4176), 1,
      aux_sym_type_binding_repeat1,
    STATE(4319), 1,
      aux_sym_expression_item_repeat1,
    STATE(4420), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5306), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179250] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4216), 1,
      aux_sym_expression_item_repeat1,
    STATE(4421), 1,
      sym_attribute,
    STATE(4427), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5119), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179296] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4422), 1,
      sym_attribute,
    STATE(6104), 1,
      sym__module_type,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [179356] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4423), 1,
      sym_attribute,
    STATE(6106), 1,
      sym__module_type,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [179416] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4176), 1,
      aux_sym_type_binding_repeat1,
    STATE(4349), 1,
      aux_sym_expression_item_repeat1,
    STATE(4424), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5423), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179462] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4285), 1,
      aux_sym_expression_item_repeat1,
    STATE(4425), 1,
      sym_attribute,
    STATE(4430), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5431), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179508] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4294), 1,
      aux_sym_expression_item_repeat1,
    STATE(4426), 1,
      sym_attribute,
    STATE(4431), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5266), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179554] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4176), 1,
      aux_sym_type_binding_repeat1,
    STATE(4251), 1,
      aux_sym_expression_item_repeat1,
    STATE(4427), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5294), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179600] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4428), 1,
      sym_attribute,
    STATE(6116), 1,
      sym__module_type,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [179660] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4262), 1,
      aux_sym_expression_item_repeat1,
    STATE(4429), 1,
      sym_attribute,
    STATE(4433), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5356), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179706] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4176), 1,
      aux_sym_type_binding_repeat1,
    STATE(4265), 1,
      aux_sym_expression_item_repeat1,
    STATE(4430), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5360), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179752] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4176), 1,
      aux_sym_type_binding_repeat1,
    STATE(4266), 1,
      aux_sym_expression_item_repeat1,
    STATE(4431), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5364), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179798] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5374), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6340), 1,
      anon_sym_and,
    STATE(4432), 1,
      sym_attribute,
    STATE(4436), 1,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5372), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179838] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4176), 1,
      aux_sym_type_binding_repeat1,
    STATE(4335), 1,
      aux_sym_expression_item_repeat1,
    STATE(4433), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5400), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179884] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4338), 1,
      aux_sym_expression_item_repeat1,
    STATE(4434), 1,
      sym_attribute,
    STATE(4437), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5404), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179930] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4176), 1,
      aux_sym_type_binding_repeat1,
    STATE(4343), 1,
      aux_sym_expression_item_repeat1,
    STATE(4435), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5408), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179976] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5256), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6342), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4436), 2,
      sym_attribute,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(5254), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180014] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4176), 1,
      aux_sym_type_binding_repeat1,
    STATE(4224), 1,
      aux_sym_expression_item_repeat1,
    STATE(4437), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5427), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180060] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4438), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6575), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [180120] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4439), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6266), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(6264), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [180156] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(6345), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4440), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6603), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [180216] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6349), 1,
      anon_sym_let,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4364), 1,
      aux_sym_expression_item_repeat1,
    STATE(4441), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6347), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [180258] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6353), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4442), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6351), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [180300] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4443), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2173), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2175), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [180336] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4444), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6614), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [180396] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6357), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4445), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6355), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [180438] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6210), 1,
      sym__lowercase_identifier,
    ACTIONS(6212), 1,
      anon_sym_LPAREN,
    ACTIONS(6218), 1,
      anon_sym_sig,
    ACTIONS(6222), 1,
      sym__uppercase_identifier,
    ACTIONS(6224), 1,
      anon_sym_DASH_GT,
    ACTIONS(6226), 1,
      anon_sym_module,
    ACTIONS(6228), 1,
      anon_sym_functor,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4446), 1,
      sym_attribute,
    STATE(5494), 1,
      sym__module_type,
    STATE(6134), 1,
      aux_sym_module_binding_repeat1,
    STATE(7119), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5495), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [180498] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4447), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6618), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [180558] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4448), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6659), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [180618] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4449), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6625), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [180678] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5958), 1,
      anon_sym_let,
    ACTIONS(6359), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4236), 1,
      aux_sym_expression_item_repeat1,
    STATE(4450), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5956), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [180722] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6210), 1,
      sym__lowercase_identifier,
    ACTIONS(6212), 1,
      anon_sym_LPAREN,
    ACTIONS(6214), 1,
      anon_sym_DASH_GT,
    ACTIONS(6216), 1,
      anon_sym_module,
    ACTIONS(6218), 1,
      anon_sym_sig,
    ACTIONS(6220), 1,
      anon_sym_functor,
    ACTIONS(6222), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4451), 1,
      sym_attribute,
    STATE(6153), 1,
      sym__module_type,
    STATE(6169), 1,
      aux_sym_module_binding_repeat1,
    STATE(7119), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5495), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [180782] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6210), 1,
      sym__lowercase_identifier,
    ACTIONS(6212), 1,
      anon_sym_LPAREN,
    ACTIONS(6214), 1,
      anon_sym_DASH_GT,
    ACTIONS(6216), 1,
      anon_sym_module,
    ACTIONS(6218), 1,
      anon_sym_sig,
    ACTIONS(6220), 1,
      anon_sym_functor,
    ACTIONS(6222), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4452), 1,
      sym_attribute,
    STATE(6155), 1,
      sym__module_type,
    STATE(6169), 1,
      aux_sym_module_binding_repeat1,
    STATE(7119), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5495), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [180842] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6210), 1,
      sym__lowercase_identifier,
    ACTIONS(6212), 1,
      anon_sym_LPAREN,
    ACTIONS(6214), 1,
      anon_sym_DASH_GT,
    ACTIONS(6216), 1,
      anon_sym_module,
    ACTIONS(6218), 1,
      anon_sym_sig,
    ACTIONS(6220), 1,
      anon_sym_functor,
    ACTIONS(6222), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4453), 1,
      sym_attribute,
    STATE(6157), 1,
      sym__module_type,
    STATE(6169), 1,
      aux_sym_module_binding_repeat1,
    STATE(7119), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5495), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [180902] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6210), 1,
      sym__lowercase_identifier,
    ACTIONS(6212), 1,
      anon_sym_LPAREN,
    ACTIONS(6218), 1,
      anon_sym_sig,
    ACTIONS(6222), 1,
      sym__uppercase_identifier,
    ACTIONS(6224), 1,
      anon_sym_DASH_GT,
    ACTIONS(6226), 1,
      anon_sym_module,
    ACTIONS(6228), 1,
      anon_sym_functor,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4454), 1,
      sym_attribute,
    STATE(5476), 1,
      sym__module_type,
    STATE(6134), 1,
      aux_sym_module_binding_repeat1,
    STATE(7119), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5495), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [180962] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(6296), 1,
      anon_sym_DASH_GT,
    ACTIONS(6298), 1,
      anon_sym_functor,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4189), 1,
      sym__module_type,
    STATE(4455), 1,
      sym_attribute,
    STATE(6193), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [181022] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6363), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4456), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6361), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [181064] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6210), 1,
      sym__lowercase_identifier,
    ACTIONS(6212), 1,
      anon_sym_LPAREN,
    ACTIONS(6214), 1,
      anon_sym_DASH_GT,
    ACTIONS(6216), 1,
      anon_sym_module,
    ACTIONS(6218), 1,
      anon_sym_sig,
    ACTIONS(6220), 1,
      anon_sym_functor,
    ACTIONS(6222), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4457), 1,
      sym_attribute,
    STATE(6169), 1,
      aux_sym_module_binding_repeat1,
    STATE(6265), 1,
      sym__module_type,
    STATE(7119), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5495), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [181124] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6210), 1,
      sym__lowercase_identifier,
    ACTIONS(6212), 1,
      anon_sym_LPAREN,
    ACTIONS(6218), 1,
      anon_sym_sig,
    ACTIONS(6222), 1,
      sym__uppercase_identifier,
    ACTIONS(6224), 1,
      anon_sym_DASH_GT,
    ACTIONS(6226), 1,
      anon_sym_module,
    ACTIONS(6228), 1,
      anon_sym_functor,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4458), 1,
      sym_attribute,
    STATE(5509), 1,
      sym__module_type,
    STATE(6134), 1,
      aux_sym_module_binding_repeat1,
    STATE(7119), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5495), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [181184] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6367), 1,
      anon_sym_let,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4412), 1,
      aux_sym_expression_item_repeat1,
    STATE(4459), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6365), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [181226] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6256), 1,
      anon_sym_and,
    ACTIONS(6258), 1,
      sym_let_and_operator,
    STATE(4376), 1,
      aux_sym_value_definition_repeat1,
    STATE(4460), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6371), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(6369), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [181268] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6256), 1,
      anon_sym_and,
    ACTIONS(6258), 1,
      sym_let_and_operator,
    STATE(4377), 1,
      aux_sym_value_definition_repeat1,
    STATE(4461), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6254), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(6252), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [181310] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(6296), 1,
      anon_sym_DASH_GT,
    ACTIONS(6298), 1,
      anon_sym_functor,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4148), 1,
      sym__module_type,
    STATE(4462), 1,
      sym_attribute,
    STATE(6193), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [181370] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(6296), 1,
      anon_sym_DASH_GT,
    ACTIONS(6298), 1,
      anon_sym_functor,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4152), 1,
      sym__module_type,
    STATE(4463), 1,
      sym_attribute,
    STATE(6193), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [181430] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6375), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4464), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6373), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [181472] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5719), 1,
      anon_sym_let,
    ACTIONS(6377), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4212), 1,
      aux_sym_expression_item_repeat1,
    STATE(4465), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5717), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [181516] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5924), 1,
      anon_sym_let,
    ACTIONS(6379), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4306), 1,
      aux_sym_expression_item_repeat1,
    STATE(4466), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5922), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [181560] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5934), 1,
      anon_sym_let,
    ACTIONS(6381), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4247), 1,
      aux_sym_expression_item_repeat1,
    STATE(4467), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5932), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [181604] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4468), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2133), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(2135), 18,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
  [181640] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6385), 1,
      anon_sym_let,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4399), 1,
      aux_sym_expression_item_repeat1,
    STATE(4469), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6383), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [181682] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(6296), 1,
      anon_sym_DASH_GT,
    ACTIONS(6298), 1,
      anon_sym_functor,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4192), 1,
      sym__module_type,
    STATE(4470), 1,
      sym_attribute,
    STATE(6193), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [181742] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5595), 1,
      anon_sym_DASH_GT,
    ACTIONS(5597), 1,
      anon_sym_module,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5601), 1,
      anon_sym_functor,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4471), 1,
      sym_attribute,
    STATE(6258), 1,
      aux_sym_module_binding_repeat1,
    STATE(6705), 1,
      sym__module_type,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [181802] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6389), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4472), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6387), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [181844] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6393), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4473), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6391), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [181886] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6397), 1,
      anon_sym_let,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4385), 1,
      aux_sym_expression_item_repeat1,
    STATE(4474), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6395), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [181928] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6401), 1,
      anon_sym_let,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4475), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6399), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [181970] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6405), 1,
      anon_sym_let,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4373), 1,
      aux_sym_expression_item_repeat1,
    STATE(4476), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6403), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [182012] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(6296), 1,
      anon_sym_DASH_GT,
    ACTIONS(6298), 1,
      anon_sym_functor,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4134), 1,
      sym__module_type,
    STATE(4477), 1,
      sym_attribute,
    STATE(6193), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [182072] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(6407), 1,
      anon_sym_DASH_GT,
    ACTIONS(6409), 1,
      anon_sym_functor,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4478), 1,
      sym_attribute,
    STATE(5967), 1,
      sym__module_type,
    STATE(6277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [182132] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(6407), 1,
      anon_sym_DASH_GT,
    ACTIONS(6409), 1,
      anon_sym_functor,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4479), 1,
      sym_attribute,
    STATE(5968), 1,
      sym__module_type,
    STATE(6277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [182192] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(6407), 1,
      anon_sym_DASH_GT,
    ACTIONS(6409), 1,
      anon_sym_functor,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4480), 1,
      sym_attribute,
    STATE(5969), 1,
      sym__module_type,
    STATE(6277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [182252] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5324), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6340), 1,
      anon_sym_and,
    STATE(4432), 1,
      aux_sym_module_type_constraint_repeat1,
    STATE(4481), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5322), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182292] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4482), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4642), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4640), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [182328] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(6407), 1,
      anon_sym_DASH_GT,
    ACTIONS(6409), 1,
      anon_sym_functor,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4483), 1,
      sym_attribute,
    STATE(5997), 1,
      sym__module_type,
    STATE(6277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [182388] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5626), 1,
      anon_sym_DASH_GT,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5630), 1,
      anon_sym_functor,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4484), 1,
      sym_attribute,
    STATE(4755), 1,
      sym__module_type,
    STATE(6293), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [182448] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5626), 1,
      anon_sym_DASH_GT,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5630), 1,
      anon_sym_functor,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4485), 1,
      sym_attribute,
    STATE(4536), 1,
      sym__module_type,
    STATE(6293), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [182508] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5626), 1,
      anon_sym_DASH_GT,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5630), 1,
      anon_sym_functor,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4486), 1,
      sym_attribute,
    STATE(4538), 1,
      sym__module_type,
    STATE(6293), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [182568] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(6407), 1,
      anon_sym_DASH_GT,
    ACTIONS(6409), 1,
      anon_sym_functor,
    STATE(3968), 1,
      sym__module_type,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4487), 1,
      sym_attribute,
    STATE(6277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [182628] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(6296), 1,
      anon_sym_DASH_GT,
    ACTIONS(6298), 1,
      anon_sym_functor,
    STATE(3968), 1,
      sym__module_type,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4488), 1,
      sym_attribute,
    STATE(6193), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [182688] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5626), 1,
      anon_sym_DASH_GT,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5630), 1,
      anon_sym_functor,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4489), 1,
      sym_attribute,
    STATE(4543), 1,
      sym__module_type,
    STATE(6293), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [182748] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5626), 1,
      anon_sym_DASH_GT,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5630), 1,
      anon_sym_functor,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4490), 1,
      sym_attribute,
    STATE(4764), 1,
      sym__module_type,
    STATE(6293), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [182808] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5587), 1,
      sym__lowercase_identifier,
    ACTIONS(5589), 1,
      anon_sym_LPAREN,
    ACTIONS(5599), 1,
      anon_sym_sig,
    ACTIONS(5626), 1,
      anon_sym_DASH_GT,
    ACTIONS(5628), 1,
      anon_sym_module,
    ACTIONS(5630), 1,
      anon_sym_functor,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(4491), 1,
      sym_attribute,
    STATE(4545), 1,
      sym__module_type,
    STATE(6293), 1,
      aux_sym_module_binding_repeat1,
    STATE(6925), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3964), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [182868] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4268), 1,
      aux_sym_expression_item_repeat1,
    STATE(4435), 1,
      aux_sym_type_binding_repeat1,
    STATE(4492), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5368), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182914] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6262), 1,
      anon_sym_let,
    ACTIONS(6411), 1,
      anon_sym_and,
    ACTIONS(6413), 1,
      sym_let_and_operator,
    STATE(4493), 1,
      sym_attribute,
    STATE(4555), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6260), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [182955] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6417), 1,
      anon_sym_let,
    ACTIONS(6419), 1,
      anon_sym_and,
    STATE(4494), 1,
      sym_attribute,
    STATE(4502), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6415), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [182994] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6419), 1,
      anon_sym_and,
    ACTIONS(6423), 1,
      anon_sym_let,
    STATE(4495), 1,
      sym_attribute,
    STATE(4502), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6421), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [183033] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6425), 1,
      anon_sym_LPAREN,
    ACTIONS(6427), 1,
      anon_sym_end,
    STATE(4496), 1,
      sym_attribute,
    STATE(4704), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [183090] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5336), 1,
      anon_sym_STAR,
    ACTIONS(5463), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4497), 1,
      sym_attribute,
    STATE(4509), 1,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5461), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183129] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6429), 1,
      anon_sym_LPAREN,
    ACTIONS(6431), 1,
      anon_sym_end,
    STATE(4498), 1,
      sym_attribute,
    STATE(4738), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [183186] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6433), 1,
      anon_sym_LPAREN,
    ACTIONS(6435), 1,
      anon_sym_end,
    STATE(4499), 1,
      sym_attribute,
    STATE(4664), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [183243] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6437), 1,
      anon_sym_and,
    ACTIONS(6439), 1,
      sym_let_and_operator,
    STATE(4500), 1,
      sym_attribute,
    STATE(4546), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6371), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(6369), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [183284] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6443), 1,
      anon_sym_let,
    ACTIONS(6445), 1,
      anon_sym_and,
    STATE(4501), 1,
      sym_attribute,
    STATE(4510), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6441), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [183323] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6449), 1,
      anon_sym_let,
    ACTIONS(6451), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4502), 2,
      sym_attribute,
      aux_sym_module_definition_repeat1,
    ACTIONS(6447), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [183360] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6419), 1,
      anon_sym_and,
    ACTIONS(6423), 1,
      anon_sym_let,
    STATE(4503), 1,
      sym_attribute,
    STATE(4518), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6421), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [183399] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6454), 1,
      anon_sym_LPAREN,
    ACTIONS(6456), 1,
      anon_sym_BANG,
    ACTIONS(6458), 1,
      anon_sym_functor,
    ACTIONS(6460), 1,
      anon_sym_struct,
    ACTIONS(6462), 1,
      sym__uppercase_identifier,
    STATE(4504), 1,
      sym_attribute,
    STATE(4821), 1,
      sym__attribute,
    STATE(5000), 1,
      sym__module_expression,
    STATE(5201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5249), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [183456] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6371), 1,
      anon_sym_let,
    ACTIONS(6411), 1,
      anon_sym_and,
    ACTIONS(6413), 1,
      sym_let_and_operator,
    STATE(4505), 1,
      sym_attribute,
    STATE(4513), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6369), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [183497] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6254), 1,
      anon_sym_let,
    ACTIONS(6411), 1,
      anon_sym_and,
    ACTIONS(6413), 1,
      sym_let_and_operator,
    STATE(4506), 1,
      sym_attribute,
    STATE(4515), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6252), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [183538] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6464), 1,
      anon_sym_LPAREN,
    ACTIONS(6466), 1,
      anon_sym_end,
    STATE(4507), 1,
      sym_attribute,
    STATE(4720), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [183595] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6470), 1,
      anon_sym_let,
    ACTIONS(6472), 1,
      anon_sym_and,
    STATE(4508), 1,
      sym_attribute,
    STATE(4527), 1,
      aux_sym_class_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6468), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [183634] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5060), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6474), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4509), 2,
      sym_attribute,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(5058), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183671] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6479), 1,
      anon_sym_let,
    ACTIONS(6481), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4510), 2,
      sym_attribute,
      aux_sym_type_definition_repeat1,
    ACTIONS(6477), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [183708] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6484), 1,
      anon_sym_LPAREN,
    ACTIONS(6486), 1,
      anon_sym_end,
    STATE(4511), 1,
      sym_attribute,
    STATE(4725), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [183765] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6443), 1,
      anon_sym_let,
    ACTIONS(6445), 1,
      anon_sym_and,
    STATE(4512), 1,
      sym_attribute,
    STATE(4524), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6441), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [183804] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6254), 1,
      anon_sym_let,
    ACTIONS(6411), 1,
      anon_sym_and,
    ACTIONS(6413), 1,
      sym_let_and_operator,
    STATE(4513), 1,
      sym_attribute,
    STATE(4517), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6252), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [183845] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6488), 1,
      anon_sym_LPAREN,
    ACTIONS(6490), 1,
      anon_sym_end,
    STATE(4514), 1,
      sym_attribute,
    STATE(4613), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [183902] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6262), 1,
      anon_sym_let,
    ACTIONS(6411), 1,
      anon_sym_and,
    ACTIONS(6413), 1,
      sym_let_and_operator,
    STATE(4515), 1,
      sym_attribute,
    STATE(4517), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6260), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [183943] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5023), 1,
      anon_sym_as,
    STATE(4516), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5192), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5190), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [183980] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6266), 1,
      anon_sym_let,
    ACTIONS(6492), 1,
      anon_sym_and,
    ACTIONS(6495), 1,
      sym_let_and_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4517), 2,
      sym_attribute,
      aux_sym_value_definition_repeat1,
    ACTIONS(6264), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [184019] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6419), 1,
      anon_sym_and,
    ACTIONS(6500), 1,
      anon_sym_let,
    STATE(4502), 1,
      aux_sym_module_definition_repeat1,
    STATE(4518), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6498), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [184058] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6502), 1,
      anon_sym_LPAREN,
    ACTIONS(6504), 1,
      anon_sym_end,
    STATE(4519), 1,
      sym_attribute,
    STATE(4651), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [184115] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4520), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4363), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(4365), 12,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [184150] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6506), 1,
      anon_sym_LPAREN,
    ACTIONS(6508), 1,
      anon_sym_end,
    STATE(4521), 1,
      sym_attribute,
    STATE(4620), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [184207] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6419), 1,
      anon_sym_and,
    ACTIONS(6512), 1,
      anon_sym_let,
    STATE(4494), 1,
      aux_sym_module_definition_repeat1,
    STATE(4522), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6510), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [184246] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6514), 1,
      anon_sym_and,
    ACTIONS(6517), 1,
      sym_let_and_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6266), 2,
      anon_sym_let,
      anon_sym_in,
    STATE(4523), 2,
      sym_attribute,
      aux_sym_value_definition_repeat1,
    ACTIONS(6264), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [184285] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6445), 1,
      anon_sym_and,
    ACTIONS(6522), 1,
      anon_sym_let,
    STATE(4510), 1,
      aux_sym_type_definition_repeat1,
    STATE(4524), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6520), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [184324] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6526), 1,
      anon_sym_let,
    ACTIONS(6528), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4525), 2,
      sym_attribute,
      aux_sym_class_type_definition_repeat1,
    ACTIONS(6524), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [184361] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6531), 1,
      anon_sym_LPAREN,
    ACTIONS(6533), 1,
      anon_sym_end,
    STATE(4526), 1,
      sym_attribute,
    STATE(4760), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [184418] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6472), 1,
      anon_sym_and,
    ACTIONS(6537), 1,
      anon_sym_let,
    STATE(4525), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(4527), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6535), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [184457] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6437), 1,
      anon_sym_and,
    ACTIONS(6439), 1,
      sym_let_and_operator,
    STATE(4528), 1,
      sym_attribute,
    STATE(4550), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6254), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(6252), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [184498] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(6539), 1,
      anon_sym_DASH_GT,
    ACTIONS(6541), 1,
      anon_sym_STAR,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(4529), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4435), 5,
      anon_sym_EQ,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4437), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
  [184547] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6437), 1,
      anon_sym_and,
    ACTIONS(6439), 1,
      sym_let_and_operator,
    STATE(4530), 1,
      sym_attribute,
    STATE(4535), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6262), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(6260), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [184588] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6543), 1,
      anon_sym_LPAREN,
    ACTIONS(6545), 1,
      anon_sym_end,
    STATE(4531), 1,
      sym_attribute,
    STATE(4563), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [184645] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6547), 1,
      anon_sym_LPAREN,
    ACTIONS(6549), 1,
      anon_sym_end,
    STATE(4532), 1,
      sym_attribute,
    STATE(4658), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [184702] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(6551), 1,
      anon_sym_STAR,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(4533), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4556), 6,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4558), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
  [184749] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5494), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6553), 1,
      anon_sym_DASH_GT,
    ACTIONS(6555), 1,
      anon_sym_with,
    STATE(4534), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5492), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184788] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6437), 1,
      anon_sym_and,
    ACTIONS(6439), 1,
      sym_let_and_operator,
    STATE(4523), 1,
      aux_sym_value_definition_repeat1,
    STATE(4535), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6276), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(6274), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [184829] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5506), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6553), 1,
      anon_sym_DASH_GT,
    ACTIONS(6555), 1,
      anon_sym_with,
    STATE(4536), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5504), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184868] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6557), 1,
      anon_sym_LPAREN,
    ACTIONS(6559), 1,
      anon_sym_end,
    STATE(4537), 1,
      sym_attribute,
    STATE(4687), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [184925] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5449), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6553), 1,
      anon_sym_DASH_GT,
    ACTIONS(6555), 1,
      anon_sym_with,
    STATE(4538), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5447), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184964] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6445), 1,
      anon_sym_and,
    ACTIONS(6563), 1,
      anon_sym_let,
    STATE(4510), 1,
      aux_sym_type_definition_repeat1,
    STATE(4539), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6561), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [185003] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6470), 1,
      anon_sym_let,
    ACTIONS(6472), 1,
      anon_sym_and,
    STATE(4525), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(4540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6468), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [185042] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5023), 1,
      anon_sym_as,
    STATE(4541), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6567), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6565), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [185079] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5023), 1,
      anon_sym_as,
    STATE(4542), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6571), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6569), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [185116] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5530), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6553), 1,
      anon_sym_DASH_GT,
    ACTIONS(6555), 1,
      anon_sym_with,
    STATE(4543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5528), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185155] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6445), 1,
      anon_sym_and,
    ACTIONS(6563), 1,
      anon_sym_let,
    STATE(4501), 1,
      aux_sym_type_definition_repeat1,
    STATE(4544), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6561), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [185194] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5457), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6553), 1,
      anon_sym_DASH_GT,
    ACTIONS(6555), 1,
      anon_sym_with,
    STATE(4545), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5455), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185233] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6437), 1,
      anon_sym_and,
    ACTIONS(6439), 1,
      sym_let_and_operator,
    STATE(4523), 1,
      aux_sym_value_definition_repeat1,
    STATE(4546), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6254), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(6252), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [185274] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6573), 1,
      anon_sym_LPAREN,
    ACTIONS(6575), 1,
      anon_sym_end,
    STATE(4547), 1,
      sym_attribute,
    STATE(4580), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [185331] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5554), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6577), 1,
      anon_sym_LBRACE,
    ACTIONS(6580), 1,
      anon_sym_DQUOTE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4388), 2,
      sym_string,
      sym_quoted_string,
    STATE(4548), 2,
      sym_attribute,
      aux_sym_external_repeat1,
    ACTIONS(5552), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185372] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6445), 1,
      anon_sym_and,
    ACTIONS(6585), 1,
      anon_sym_let,
    STATE(4539), 1,
      aux_sym_type_definition_repeat1,
    STATE(4549), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6583), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [185411] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6437), 1,
      anon_sym_and,
    ACTIONS(6439), 1,
      sym_let_and_operator,
    STATE(4523), 1,
      aux_sym_value_definition_repeat1,
    STATE(4550), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6262), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(6260), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [185452] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6417), 1,
      anon_sym_let,
    ACTIONS(6419), 1,
      anon_sym_and,
    STATE(4495), 1,
      aux_sym_module_definition_repeat1,
    STATE(4551), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6415), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [185491] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6589), 1,
      anon_sym_BANG,
    ACTIONS(6591), 1,
      anon_sym_functor,
    ACTIONS(6593), 1,
      anon_sym_struct,
    STATE(3580), 1,
      sym_module_path,
    STATE(3671), 1,
      sym__module_expression,
    STATE(4552), 1,
      sym_attribute,
    STATE(4815), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [185548] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6595), 1,
      anon_sym_LPAREN,
    ACTIONS(6597), 1,
      anon_sym_end,
    STATE(4553), 1,
      sym_attribute,
    STATE(4632), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [185605] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6472), 1,
      anon_sym_and,
    ACTIONS(6601), 1,
      anon_sym_let,
    STATE(4540), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(4554), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6599), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [185644] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6276), 1,
      anon_sym_let,
    ACTIONS(6411), 1,
      anon_sym_and,
    ACTIONS(6413), 1,
      sym_let_and_operator,
    STATE(4517), 1,
      aux_sym_value_definition_repeat1,
    STATE(4555), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6274), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [185685] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6603), 1,
      anon_sym_end,
    STATE(4556), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [185739] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4472), 1,
      aux_sym_expression_item_repeat1,
    STATE(4557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6605), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(5667), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185779] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6545), 1,
      anon_sym_end,
    STATE(4558), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [185833] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6607), 1,
      anon_sym_end,
    STATE(4559), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [185887] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6609), 1,
      anon_sym_end,
    STATE(4560), 1,
      sym_attribute,
    STATE(4594), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [185941] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2986), 1,
      sym__lowercase_identifier,
    ACTIONS(2990), 1,
      anon_sym_LBRACK,
    ACTIONS(2998), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3002), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3004), 1,
      anon_sym_begin,
    ACTIONS(6611), 1,
      anon_sym_LPAREN,
    ACTIONS(6613), 1,
      anon_sym_LBRACE,
    ACTIONS(6615), 1,
      sym__uppercase_identifier,
    STATE(2466), 1,
      sym__delimited_expression,
    STATE(2490), 1,
      sym__value_name,
    STATE(2508), 1,
      sym_parenthesized_operator,
    STATE(2521), 1,
      sym__constructor_name,
    STATE(4561), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2450), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [185997] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4562), 1,
      sym_attribute,
    STATE(4826), 1,
      sym__attribute,
    STATE(5060), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [186051] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6619), 1,
      anon_sym_end,
    STATE(4563), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [186105] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6454), 1,
      anon_sym_LPAREN,
    ACTIONS(6458), 1,
      anon_sym_functor,
    ACTIONS(6460), 1,
      anon_sym_struct,
    ACTIONS(6462), 1,
      sym__uppercase_identifier,
    STATE(4564), 1,
      sym_attribute,
    STATE(4827), 1,
      sym__attribute,
    STATE(4991), 1,
      sym__module_expression,
    STATE(5201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5249), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [186159] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6621), 1,
      anon_sym_end,
    STATE(4565), 1,
      sym_attribute,
    STATE(4600), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [186213] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6623), 1,
      anon_sym_end,
    STATE(4566), 1,
      sym_attribute,
    STATE(4568), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [186267] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6625), 1,
      anon_sym_end,
    STATE(4567), 1,
      sym_attribute,
    STATE(4592), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [186321] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6627), 1,
      anon_sym_end,
    STATE(4568), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [186375] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6629), 1,
      anon_sym_end,
    STATE(4569), 1,
      sym_attribute,
    STATE(4571), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [186429] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6631), 1,
      anon_sym_end,
    STATE(4570), 1,
      sym_attribute,
    STATE(4572), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [186483] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6633), 1,
      anon_sym_end,
    STATE(4571), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [186537] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6635), 1,
      anon_sym_end,
    STATE(4572), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [186591] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6637), 1,
      anon_sym_end,
    STATE(4573), 1,
      sym_attribute,
    STATE(4574), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [186645] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6639), 1,
      anon_sym_end,
    STATE(4574), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [186699] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4575), 1,
      sym_attribute,
    STATE(4817), 1,
      sym__attribute,
    STATE(5065), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [186753] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6490), 1,
      anon_sym_end,
    STATE(4576), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [186807] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6575), 1,
      anon_sym_end,
    STATE(4577), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [186861] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4578), 1,
      sym_attribute,
    STATE(4830), 1,
      sym__attribute,
    STATE(5097), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [186915] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2211), 1,
      sym__lowercase_identifier,
    ACTIONS(2217), 1,
      anon_sym_LBRACK,
    ACTIONS(2225), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2229), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2231), 1,
      anon_sym_begin,
    ACTIONS(6641), 1,
      anon_sym_LPAREN,
    ACTIONS(6643), 1,
      anon_sym_LBRACE,
    ACTIONS(6645), 1,
      sym__uppercase_identifier,
    STATE(2948), 1,
      sym__value_name,
    STATE(2981), 1,
      sym__constructor_name,
    STATE(2999), 1,
      sym__delimited_expression,
    STATE(3038), 1,
      sym_parenthesized_operator,
    STATE(4579), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3049), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [186971] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6647), 1,
      anon_sym_end,
    STATE(4580), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187025] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6649), 1,
      anon_sym_end,
    STATE(4581), 1,
      sym_attribute,
    STATE(4582), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187079] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6651), 1,
      anon_sym_end,
    STATE(4582), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187133] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6653), 1,
      anon_sym_end,
    STATE(4583), 1,
      sym_attribute,
    STATE(4586), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187187] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4584), 1,
      sym_attribute,
    STATE(4834), 1,
      sym__attribute,
    STATE(5069), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [187241] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6655), 1,
      anon_sym_end,
    STATE(4585), 1,
      sym_attribute,
    STATE(4587), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187295] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6657), 1,
      anon_sym_end,
    STATE(4586), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187349] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6659), 1,
      anon_sym_end,
    STATE(4587), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187403] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6661), 1,
      anon_sym_end,
    STATE(4588), 1,
      sym_attribute,
    STATE(4589), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187457] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6663), 1,
      anon_sym_end,
    STATE(4589), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187511] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6665), 1,
      anon_sym_end,
    STATE(4590), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187565] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4591), 1,
      sym_attribute,
    STATE(4816), 1,
      sym__attribute,
    STATE(5076), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [187619] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6667), 1,
      anon_sym_end,
    STATE(4592), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187673] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6669), 1,
      anon_sym_end,
    STATE(4593), 1,
      sym_attribute,
    STATE(4597), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187727] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6671), 1,
      anon_sym_end,
    STATE(4594), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187781] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6673), 1,
      anon_sym_end,
    STATE(4595), 1,
      sym_attribute,
    STATE(4598), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187835] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4596), 1,
      sym_attribute,
    STATE(4858), 1,
      sym__attribute,
    STATE(5089), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [187889] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6675), 1,
      anon_sym_end,
    STATE(4597), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187943] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6677), 1,
      anon_sym_end,
    STATE(4598), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [187997] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6679), 1,
      anon_sym_end,
    STATE(4599), 1,
      sym_attribute,
    STATE(4602), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [188051] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6681), 1,
      anon_sym_end,
    STATE(4600), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [188105] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6683), 1,
      anon_sym_end,
    STATE(4601), 1,
      sym_attribute,
    STATE(4614), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [188159] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6685), 1,
      anon_sym_end,
    STATE(4602), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [188213] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6597), 1,
      anon_sym_end,
    STATE(4603), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [188267] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4604), 1,
      sym_attribute,
    STATE(4848), 1,
      sym__attribute,
    STATE(5101), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [188321] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4605), 1,
      sym_attribute,
    STATE(4850), 1,
      sym__attribute,
    STATE(5149), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [188375] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4606), 1,
      sym_attribute,
    STATE(4842), 1,
      sym__attribute,
    STATE(5103), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [188429] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6687), 1,
      anon_sym_end,
    STATE(4607), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [188483] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6689), 1,
      anon_sym_end,
    STATE(4608), 1,
      sym_attribute,
    STATE(4702), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [188537] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(6691), 1,
      sym__lowercase_identifier,
    ACTIONS(6693), 1,
      anon_sym_LPAREN,
    ACTIONS(6695), 1,
      anon_sym_LBRACE,
    ACTIONS(6697), 1,
      sym__uppercase_identifier,
    STATE(2228), 1,
      sym__constructor_name,
    STATE(2329), 1,
      sym__delimited_expression,
    STATE(2332), 1,
      sym__value_name,
    STATE(4609), 1,
      sym_attribute,
    STATE(4997), 1,
      sym_parenthesized_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [188593] = 16,
    ACTIONS(409), 1,
      anon_sym_LBRACK,
    ACTIONS(419), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(439), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(441), 1,
      anon_sym_begin,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(989), 1,
      sym__lowercase_identifier,
    ACTIONS(6699), 1,
      anon_sym_LPAREN,
    ACTIONS(6701), 1,
      anon_sym_LBRACE,
    ACTIONS(6703), 1,
      sym__uppercase_identifier,
    STATE(1059), 1,
      sym__delimited_expression,
    STATE(1074), 1,
      sym_parenthesized_operator,
    STATE(1078), 1,
      sym__value_name,
    STATE(1112), 1,
      sym__constructor_name,
    STATE(4610), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1046), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [188649] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6508), 1,
      anon_sym_end,
    STATE(4611), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [188703] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6591), 1,
      anon_sym_functor,
    ACTIONS(6593), 1,
      anon_sym_struct,
    STATE(3580), 1,
      sym_module_path,
    STATE(3687), 1,
      sym__module_expression,
    STATE(4612), 1,
      sym_attribute,
    STATE(4829), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [188757] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6705), 1,
      anon_sym_end,
    STATE(4613), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [188811] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6707), 1,
      anon_sym_end,
    STATE(4614), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [188865] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4615), 1,
      sym_attribute,
    STATE(4833), 1,
      sym__attribute,
    STATE(5112), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [188919] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6254), 1,
      anon_sym_let,
    ACTIONS(6709), 1,
      anon_sym_and,
    ACTIONS(6711), 1,
      sym_let_and_operator,
    STATE(4616), 1,
      sym_attribute,
    STATE(4635), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6252), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [188959] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3064), 1,
      sym__lowercase_identifier,
    ACTIONS(3068), 1,
      anon_sym_LBRACK,
    ACTIONS(3076), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3080), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3082), 1,
      anon_sym_begin,
    ACTIONS(6713), 1,
      anon_sym_LPAREN,
    ACTIONS(6715), 1,
      anon_sym_LBRACE,
    ACTIONS(6717), 1,
      sym__uppercase_identifier,
    STATE(2174), 1,
      sym__constructor_name,
    STATE(2451), 1,
      sym_parenthesized_operator,
    STATE(2577), 1,
      sym__delimited_expression,
    STATE(2578), 1,
      sym__value_name,
    STATE(4617), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2552), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [189015] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    ACTIONS(6719), 1,
      anon_sym_RPAREN,
    ACTIONS(6721), 1,
      anon_sym_val,
    STATE(3580), 1,
      sym_module_path,
    STATE(4618), 1,
      sym_attribute,
    STATE(5131), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [189069] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4619), 1,
      sym_attribute,
    STATE(4810), 1,
      sym__attribute,
    STATE(5121), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [189123] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6723), 1,
      anon_sym_end,
    STATE(4620), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [189177] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6449), 1,
      anon_sym_let,
    STATE(4621), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6447), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [189211] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6725), 1,
      anon_sym_constraint,
    ACTIONS(6728), 1,
      anon_sym_val,
    ACTIONS(6731), 1,
      anon_sym_end,
    ACTIONS(6733), 1,
      anon_sym_inherit,
    ACTIONS(6736), 1,
      anon_sym_method,
    ACTIONS(6739), 1,
      anon_sym_initializer,
    ACTIONS(6742), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(6745), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(6748), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4622), 2,
      sym_attribute,
      aux_sym_object_expression_repeat1,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [189263] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6549), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4623), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [189317] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6751), 1,
      anon_sym_end,
    STATE(4624), 1,
      sym_attribute,
    STATE(4627), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [189371] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6755), 1,
      anon_sym_let,
    ACTIONS(6757), 1,
      anon_sym_and,
    STATE(4625), 1,
      sym_attribute,
    STATE(4649), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6753), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [189409] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6755), 1,
      anon_sym_let,
    ACTIONS(6757), 1,
      anon_sym_and,
    STATE(4626), 1,
      sym_attribute,
    STATE(4650), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6753), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [189447] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6759), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4627), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [189501] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(6761), 1,
      anon_sym_DASH_GT,
    ACTIONS(6763), 1,
      anon_sym_STAR,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(4628), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4437), 5,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
    ACTIONS(4435), 7,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [189549] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6371), 1,
      anon_sym_let,
    ACTIONS(6709), 1,
      anon_sym_and,
    ACTIONS(6711), 1,
      sym_let_and_operator,
    STATE(4616), 1,
      aux_sym_value_definition_repeat1,
    STATE(4629), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6369), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [189589] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6757), 1,
      anon_sym_and,
    ACTIONS(6767), 1,
      anon_sym_let,
    STATE(4625), 1,
      aux_sym_class_definition_repeat1,
    STATE(4630), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6765), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [189627] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(491), 1,
      anon_sym_LBRACK,
    ACTIONS(501), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(521), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(523), 1,
      anon_sym_begin,
    ACTIONS(911), 1,
      sym__lowercase_identifier,
    ACTIONS(6769), 1,
      anon_sym_LPAREN,
    ACTIONS(6771), 1,
      anon_sym_LBRACE,
    ACTIONS(6773), 1,
      sym__uppercase_identifier,
    STATE(953), 1,
      sym__delimited_expression,
    STATE(954), 1,
      sym__value_name,
    STATE(989), 1,
      sym_parenthesized_operator,
    STATE(1004), 1,
      sym__constructor_name,
    STATE(4631), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(958), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [189683] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6775), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4632), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [189737] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6479), 1,
      anon_sym_let,
    STATE(4633), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6477), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [189771] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6262), 1,
      anon_sym_let,
    ACTIONS(6709), 1,
      anon_sym_and,
    ACTIONS(6711), 1,
      sym_let_and_operator,
    STATE(4634), 1,
      sym_attribute,
    STATE(4635), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6260), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [189811] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6266), 1,
      anon_sym_let,
    ACTIONS(6777), 1,
      anon_sym_and,
    ACTIONS(6780), 1,
      sym_let_and_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4635), 2,
      sym_attribute,
      aux_sym_value_definition_repeat1,
    ACTIONS(6264), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [189849] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6254), 1,
      anon_sym_let,
    ACTIONS(6709), 1,
      anon_sym_and,
    ACTIONS(6711), 1,
      sym_let_and_operator,
    STATE(4634), 1,
      aux_sym_value_definition_repeat1,
    STATE(4636), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6252), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [189889] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6783), 1,
      anon_sym_end,
    STATE(4637), 1,
      sym_attribute,
    STATE(4641), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [189943] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6262), 1,
      anon_sym_let,
    ACTIONS(6709), 1,
      anon_sym_and,
    ACTIONS(6711), 1,
      sym_let_and_operator,
    STATE(4638), 1,
      sym_attribute,
    STATE(4655), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6260), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [189983] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4639), 1,
      sym_attribute,
    STATE(4843), 1,
      sym__attribute,
    STATE(5171), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [190037] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6785), 1,
      anon_sym_end,
    STATE(4640), 1,
      sym_attribute,
    STATE(4643), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [190091] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6787), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4641), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [190145] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6504), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4642), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [190199] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6789), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4643), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [190253] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6791), 1,
      anon_sym_end,
    STATE(4644), 1,
      sym_attribute,
    STATE(4645), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [190307] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6793), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4645), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [190361] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4943), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6795), 1,
      anon_sym_PIPE,
    STATE(4646), 1,
      sym_attribute,
    STATE(4657), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4941), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [190399] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6797), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4346), 1,
      aux_sym_expression_item_repeat1,
    STATE(4647), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6128), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [190439] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2950), 1,
      sym__lowercase_identifier,
    ACTIONS(2958), 1,
      anon_sym_begin,
    ACTIONS(6695), 1,
      anon_sym_LBRACE,
    ACTIONS(6697), 1,
      sym__uppercase_identifier,
    ACTIONS(6799), 1,
      anon_sym_LPAREN,
    STATE(2228), 1,
      sym__constructor_name,
    STATE(2329), 1,
      sym__delimited_expression,
    STATE(2332), 1,
      sym__value_name,
    STATE(2366), 1,
      sym_parenthesized_operator,
    STATE(4648), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [190495] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6803), 1,
      anon_sym_let,
    ACTIONS(6805), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4649), 2,
      sym_attribute,
      aux_sym_class_definition_repeat1,
    ACTIONS(6801), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [190531] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6757), 1,
      anon_sym_and,
    ACTIONS(6810), 1,
      anon_sym_let,
    STATE(4649), 1,
      aux_sym_class_definition_repeat1,
    STATE(4650), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6808), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [190569] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6812), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4651), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [190623] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5616), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6795), 1,
      anon_sym_PIPE,
    STATE(4652), 1,
      sym_attribute,
    STATE(4669), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5614), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [190661] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6553), 1,
      anon_sym_DASH_GT,
    ACTIONS(6555), 1,
      anon_sym_with,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4653), 1,
      sym_attribute,
    STATE(4839), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6814), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [190703] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6466), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4654), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [190757] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6276), 1,
      anon_sym_let,
    ACTIONS(6709), 1,
      anon_sym_and,
    ACTIONS(6711), 1,
      sym_let_and_operator,
    STATE(4635), 1,
      aux_sym_value_definition_repeat1,
    STATE(4655), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6274), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [190797] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(809), 1,
      anon_sym_LBRACK,
    ACTIONS(819), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(839), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(841), 1,
      anon_sym_begin,
    ACTIONS(1687), 1,
      sym__lowercase_identifier,
    ACTIONS(6816), 1,
      anon_sym_LPAREN,
    ACTIONS(6818), 1,
      anon_sym_LBRACE,
    ACTIONS(6820), 1,
      sym__uppercase_identifier,
    STATE(1951), 1,
      sym__value_name,
    STATE(1955), 1,
      sym__delimited_expression,
    STATE(1984), 1,
      sym__constructor_name,
    STATE(2019), 1,
      sym_parenthesized_operator,
    STATE(4656), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2002), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [190853] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5616), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6795), 1,
      anon_sym_PIPE,
    STATE(4657), 1,
      sym_attribute,
    STATE(4670), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5614), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [190891] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6822), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4658), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [190945] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6824), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4252), 1,
      aux_sym_expression_item_repeat1,
    STATE(4659), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5685), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [190985] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6826), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4212), 1,
      aux_sym_expression_item_repeat1,
    STATE(4660), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5717), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [191025] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6435), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4661), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [191079] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3026), 1,
      sym__lowercase_identifier,
    ACTIONS(3030), 1,
      anon_sym_LBRACK,
    ACTIONS(3038), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3042), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3044), 1,
      anon_sym_begin,
    ACTIONS(6828), 1,
      anon_sym_LPAREN,
    ACTIONS(6830), 1,
      anon_sym_LBRACE,
    ACTIONS(6832), 1,
      sym__uppercase_identifier,
    STATE(3296), 1,
      sym_parenthesized_operator,
    STATE(3323), 1,
      sym__value_name,
    STATE(3343), 1,
      sym__constructor_name,
    STATE(3354), 1,
      sym__delimited_expression,
    STATE(4662), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3310), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [191135] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6834), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4314), 1,
      aux_sym_expression_item_repeat1,
    STATE(4663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5778), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [191175] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6836), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4664), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [191229] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6591), 1,
      anon_sym_functor,
    ACTIONS(6593), 1,
      anon_sym_struct,
    STATE(3580), 1,
      sym_module_path,
    STATE(3707), 1,
      sym__module_expression,
    STATE(4665), 1,
      sym_attribute,
    STATE(4825), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [191283] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(6838), 1,
      anon_sym_STAR,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(4666), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4558), 5,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
    ACTIONS(4556), 8,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [191329] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6840), 1,
      anon_sym_end,
    STATE(4667), 1,
      sym_attribute,
    STATE(4677), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [191383] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6842), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4348), 1,
      aux_sym_expression_item_repeat1,
    STATE(4668), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5824), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [191423] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5644), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6795), 1,
      anon_sym_PIPE,
    STATE(4669), 1,
      sym_attribute,
    STATE(4670), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5642), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [191461] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5652), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6844), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4670), 2,
      sym_attribute,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5650), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [191497] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6847), 1,
      anon_sym_end,
    STATE(4671), 1,
      sym_attribute,
    STATE(4737), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [191551] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6849), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4293), 1,
      aux_sym_expression_item_repeat1,
    STATE(4672), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5866), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [191591] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6851), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4298), 1,
      aux_sym_expression_item_repeat1,
    STATE(4673), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5876), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [191631] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    ACTIONS(6853), 1,
      anon_sym_RPAREN,
    ACTIONS(6855), 1,
      anon_sym_val,
    STATE(3580), 1,
      sym_module_path,
    STATE(4674), 1,
      sym_attribute,
    STATE(5150), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [191685] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6857), 1,
      anon_sym_end,
    STATE(4559), 1,
      aux_sym_object_expression_repeat1,
    STATE(4675), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [191739] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6859), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4306), 1,
      aux_sym_expression_item_repeat1,
    STATE(4676), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5922), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [191779] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6861), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [191833] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6863), 1,
      anon_sym_end,
    STATE(4678), 1,
      sym_attribute,
    STATE(4683), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [191887] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6865), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4247), 1,
      aux_sym_expression_item_repeat1,
    STATE(4679), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5932), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [191927] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6867), 1,
      anon_sym_end,
    STATE(4680), 1,
      sym_attribute,
    STATE(4684), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [191981] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6869), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4236), 1,
      aux_sym_expression_item_repeat1,
    STATE(4681), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5956), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [192021] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6871), 1,
      anon_sym_end,
    STATE(4682), 1,
      sym_attribute,
    STATE(4705), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [192075] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6873), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4683), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [192129] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6875), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4684), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [192183] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6877), 1,
      anon_sym_end,
    STATE(4685), 1,
      sym_attribute,
    STATE(4686), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [192237] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6879), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4686), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [192291] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6881), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4687), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [192345] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(6883), 1,
      anon_sym_STAR,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(4688), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4558), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_as,
    ACTIONS(4556), 9,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
  [192391] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6553), 1,
      anon_sym_DASH_GT,
    ACTIONS(6555), 1,
      anon_sym_with,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4689), 1,
      sym_attribute,
    STATE(4840), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6885), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [192433] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6887), 1,
      anon_sym_end,
    STATE(4690), 1,
      sym_attribute,
    STATE(4692), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [192487] = 16,
    ACTIONS(255), 1,
      anon_sym_LBRACK,
    ACTIONS(265), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(277), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(279), 1,
      anon_sym_begin,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1083), 1,
      sym__lowercase_identifier,
    ACTIONS(6889), 1,
      anon_sym_LPAREN,
    ACTIONS(6891), 1,
      anon_sym_LBRACE,
    ACTIONS(6893), 1,
      sym__uppercase_identifier,
    STATE(1325), 1,
      sym__constructor_name,
    STATE(1346), 1,
      sym__value_name,
    STATE(1370), 1,
      sym__delimited_expression,
    STATE(1378), 1,
      sym_parenthesized_operator,
    STATE(4691), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1377), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [192543] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6895), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4692), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [192597] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6897), 1,
      anon_sym_end,
    STATE(4693), 1,
      sym_attribute,
    STATE(4697), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [192651] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6899), 1,
      anon_sym_end,
    STATE(4694), 1,
      sym_attribute,
    STATE(4698), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [192705] = 16,
    ACTIONS(25), 1,
      anon_sym_LBRACK,
    ACTIONS(47), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(67), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(69), 1,
      anon_sym_begin,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1153), 1,
      sym__lowercase_identifier,
    ACTIONS(6901), 1,
      anon_sym_LPAREN,
    ACTIONS(6903), 1,
      anon_sym_LBRACE,
    ACTIONS(6905), 1,
      sym__uppercase_identifier,
    STATE(1463), 1,
      sym__delimited_expression,
    STATE(1469), 1,
      sym__value_name,
    STATE(1474), 1,
      sym__constructor_name,
    STATE(1483), 1,
      sym_parenthesized_operator,
    STATE(4695), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1418), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [192761] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6427), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [192815] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6907), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4697), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [192869] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6909), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4698), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [192923] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6911), 1,
      anon_sym_end,
    STATE(4699), 1,
      sym_attribute,
    STATE(4701), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [192977] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6915), 1,
      anon_sym_let,
    ACTIONS(6917), 1,
      anon_sym_DOT,
    STATE(4700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6913), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [193013] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6919), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4701), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193067] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6921), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193121] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__lowercase_identifier,
    ACTIONS(2920), 1,
      anon_sym_LBRACK,
    ACTIONS(2926), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2930), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2932), 1,
      anon_sym_begin,
    ACTIONS(6693), 1,
      anon_sym_LPAREN,
    ACTIONS(6695), 1,
      anon_sym_LBRACE,
    ACTIONS(6697), 1,
      sym__uppercase_identifier,
    STATE(2228), 1,
      sym__constructor_name,
    STATE(2329), 1,
      sym__delimited_expression,
    STATE(2332), 1,
      sym__value_name,
    STATE(4703), 1,
      sym_attribute,
    STATE(4997), 1,
      sym_parenthesized_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2424), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [193177] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6923), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4704), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193231] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6925), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4705), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193285] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6927), 1,
      anon_sym_end,
    STATE(4706), 1,
      sym_attribute,
    STATE(4718), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193339] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6929), 1,
      anon_sym_end,
    STATE(4707), 1,
      sym_attribute,
    STATE(4708), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193393] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6931), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193447] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6933), 1,
      anon_sym_end,
    STATE(4709), 1,
      sym_attribute,
    STATE(4711), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193501] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6935), 1,
      anon_sym_end,
    STATE(4710), 1,
      sym_attribute,
    STATE(4712), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193555] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6937), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193609] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6939), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193663] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6941), 1,
      anon_sym_end,
    STATE(4713), 1,
      sym_attribute,
    STATE(4714), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193717] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6943), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4714), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193771] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6945), 1,
      anon_sym_end,
    STATE(4715), 1,
      sym_attribute,
    STATE(4722), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193825] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(6947), 1,
      anon_sym_DASH_GT,
    ACTIONS(6949), 1,
      anon_sym_STAR,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(4716), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4437), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_as,
    ACTIONS(4435), 8,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
  [193873] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6951), 1,
      anon_sym_end,
    STATE(4607), 1,
      aux_sym_object_expression_repeat1,
    STATE(4717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193927] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6953), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4718), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [193981] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6803), 1,
      anon_sym_let,
    STATE(4719), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6801), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [194015] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6955), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4720), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194069] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6486), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4721), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194123] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6957), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4722), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194177] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6959), 1,
      anon_sym_end,
    STATE(4723), 1,
      sym_attribute,
    STATE(4727), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194231] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3110), 1,
      sym__lowercase_identifier,
    ACTIONS(3114), 1,
      anon_sym_LBRACK,
    ACTIONS(3122), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3126), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(3128), 1,
      anon_sym_begin,
    ACTIONS(6961), 1,
      anon_sym_LPAREN,
    ACTIONS(6963), 1,
      anon_sym_LBRACE,
    ACTIONS(6965), 1,
      sym__uppercase_identifier,
    STATE(2293), 1,
      sym__delimited_expression,
    STATE(2345), 1,
      sym__constructor_name,
    STATE(2413), 1,
      sym_parenthesized_operator,
    STATE(2431), 1,
      sym__value_name,
    STATE(4724), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2403), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [194287] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6967), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4725), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194341] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6431), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4726), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194395] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6969), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4727), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194449] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6971), 1,
      anon_sym_end,
    STATE(4556), 1,
      aux_sym_object_expression_repeat1,
    STATE(4728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194503] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6973), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4729), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194557] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6975), 1,
      anon_sym_end,
    STATE(4730), 1,
      sym_attribute,
    STATE(4732), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194611] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6977), 1,
      anon_sym_end,
    STATE(4731), 1,
      sym_attribute,
    STATE(4733), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194665] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6979), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4732), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194719] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6981), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4733), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194773] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6983), 1,
      anon_sym_end,
    STATE(4734), 1,
      sym_attribute,
    STATE(4735), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194827] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6985), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4735), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194881] = 16,
    ACTIONS(115), 1,
      anon_sym_LBRACK,
    ACTIONS(125), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(141), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(143), 1,
      anon_sym_begin,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1003), 1,
      sym__lowercase_identifier,
    ACTIONS(6987), 1,
      anon_sym_LPAREN,
    ACTIONS(6989), 1,
      anon_sym_LBRACE,
    ACTIONS(6991), 1,
      sym__uppercase_identifier,
    STATE(1154), 1,
      sym_parenthesized_operator,
    STATE(1178), 1,
      sym__constructor_name,
    STATE(1201), 1,
      sym__value_name,
    STATE(1233), 1,
      sym__delimited_expression,
    STATE(4736), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1190), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [194937] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6993), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [194991] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6995), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4738), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [195045] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6997), 1,
      anon_sym_end,
    STATE(4739), 1,
      sym_attribute,
    STATE(4747), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [195099] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6999), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4740), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [195153] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7001), 1,
      anon_sym_end,
    STATE(4740), 1,
      aux_sym_object_expression_repeat1,
    STATE(4741), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [195207] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7003), 1,
      anon_sym_end,
    STATE(4742), 1,
      sym_attribute,
    STATE(4754), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [195261] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6526), 1,
      anon_sym_let,
    STATE(4743), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6524), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [195295] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7005), 1,
      anon_sym_end,
    STATE(4744), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [195349] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7007), 1,
      anon_sym_end,
    STATE(4590), 1,
      aux_sym_object_expression_repeat1,
    STATE(4745), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [195403] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6533), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [195457] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7009), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4747), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [195511] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4748), 1,
      sym_attribute,
    STATE(4851), 1,
      sym__attribute,
    STATE(5038), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [195565] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7011), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4749), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [195619] = 16,
    ACTIONS(189), 1,
      anon_sym_LBRACK,
    ACTIONS(199), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(211), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(213), 1,
      anon_sym_begin,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1013), 1,
      sym__lowercase_identifier,
    ACTIONS(7013), 1,
      anon_sym_LPAREN,
    ACTIONS(7015), 1,
      anon_sym_LBRACE,
    ACTIONS(7017), 1,
      sym__uppercase_identifier,
    STATE(1240), 1,
      sym__delimited_expression,
    STATE(1252), 1,
      sym__constructor_name,
    STATE(1283), 1,
      sym_parenthesized_operator,
    STATE(1287), 1,
      sym__value_name,
    STATE(4750), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1298), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [195675] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7019), 1,
      anon_sym_end,
    STATE(4751), 1,
      sym_attribute,
    STATE(4756), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [195729] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7021), 1,
      anon_sym_end,
    STATE(4752), 1,
      sym_attribute,
    STATE(4757), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [195783] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7023), 1,
      anon_sym_COLON2,
    STATE(4753), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3470), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [195819] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7025), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4754), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [195873] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6553), 1,
      anon_sym_DASH_GT,
    ACTIONS(6555), 1,
      anon_sym_with,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4404), 1,
      aux_sym_expression_item_repeat1,
    STATE(4755), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5622), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [195915] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7027), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [195969] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7029), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4757), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [196023] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7031), 1,
      anon_sym_end,
    STATE(4758), 1,
      sym_attribute,
    STATE(4762), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [196077] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7033), 1,
      anon_sym_end,
    STATE(4759), 1,
      sym_attribute,
    STATE(4761), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [196131] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7035), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [196185] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7037), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4761), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [196239] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7039), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4762), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [196293] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7041), 1,
      anon_sym_end,
    STATE(4763), 1,
      sym_attribute,
    STATE(4765), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [196347] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6553), 1,
      anon_sym_DASH_GT,
    ACTIONS(6555), 1,
      anon_sym_with,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4387), 1,
      aux_sym_expression_item_repeat1,
    STATE(4764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5638), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [196389] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7043), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4765), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [196443] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4456), 1,
      aux_sym_expression_item_repeat1,
    STATE(4766), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7045), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(5657), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [196483] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7047), 1,
      anon_sym_end,
    STATE(4767), 1,
      sym_attribute,
    STATE(4769), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [196537] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7049), 1,
      anon_sym_end,
    STATE(4768), 1,
      sym_attribute,
    STATE(4770), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [196591] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7051), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4769), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [196645] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7053), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4770), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [196699] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7055), 1,
      anon_sym_end,
    STATE(4729), 1,
      aux_sym_object_expression_repeat1,
    STATE(4771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [196753] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5703), 1,
      anon_sym_val,
    ACTIONS(5707), 1,
      anon_sym_inherit,
    ACTIONS(5709), 1,
      anon_sym_method,
    ACTIONS(5711), 1,
      anon_sym_initializer,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6559), 1,
      anon_sym_end,
    STATE(4622), 1,
      aux_sym_object_expression_repeat1,
    STATE(4772), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5328), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(5253), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [196807] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6915), 1,
      anon_sym_let,
    STATE(4773), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6913), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [196840] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2203), 1,
      anon_sym_let,
    STATE(4774), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2205), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [196873] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7059), 1,
      anon_sym_let,
    STATE(4775), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7057), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [196906] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7061), 1,
      anon_sym_LPAREN,
    ACTIONS(7063), 1,
      anon_sym_val,
    ACTIONS(7065), 1,
      anon_sym_end,
    ACTIONS(7067), 1,
      anon_sym_inherit,
    ACTIONS(7069), 1,
      anon_sym_method,
    STATE(4776), 1,
      sym_attribute,
    STATE(4844), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5356), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(5357), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [196959] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6755), 1,
      anon_sym_let,
    ACTIONS(7071), 1,
      anon_sym_and,
    STATE(4777), 1,
      sym_attribute,
    STATE(4797), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6753), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [196996] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7063), 1,
      anon_sym_val,
    ACTIONS(7067), 1,
      anon_sym_inherit,
    ACTIONS(7069), 1,
      anon_sym_method,
    ACTIONS(7073), 1,
      anon_sym_LPAREN,
    ACTIONS(7075), 1,
      anon_sym_end,
    STATE(4778), 1,
      sym_attribute,
    STATE(4846), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5356), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(5357), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [197049] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2031), 1,
      anon_sym_let,
    STATE(4779), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2033), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197082] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2169), 1,
      anon_sym_let,
    STATE(4780), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2171), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197115] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7079), 1,
      anon_sym_let,
    STATE(4781), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7077), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197148] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2103), 1,
      anon_sym_let,
    STATE(4782), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2105), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197181] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2107), 1,
      anon_sym_let,
    STATE(4783), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2109), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197214] = 19,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(7081), 1,
      sym__lowercase_identifier,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7085), 1,
      anon_sym_nonrec,
    ACTIONS(7089), 1,
      anon_sym__,
    STATE(4784), 1,
      sym_attribute,
    STATE(4900), 1,
      sym__attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6190), 1,
      sym__type_params,
    STATE(6362), 1,
      sym_type_binding,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(7435), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [197275] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2117), 1,
      anon_sym_let,
    STATE(4785), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2119), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197308] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2207), 1,
      anon_sym_let,
    STATE(4786), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2209), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197341] = 14,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    ACTIONS(6721), 1,
      anon_sym_val,
    STATE(3580), 1,
      sym_module_path,
    STATE(4787), 1,
      sym_attribute,
    STATE(5131), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [197392] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5344), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6248), 1,
      anon_sym_of,
    ACTIONS(6250), 1,
      anon_sym_COLON2,
    ACTIONS(7091), 1,
      anon_sym_EQ,
    STATE(4788), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5342), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [197431] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2199), 1,
      anon_sym_let,
    STATE(4789), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2201), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197464] = 19,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    ACTIONS(7093), 1,
      sym__lowercase_identifier,
    ACTIONS(7095), 1,
      anon_sym_nonrec,
    STATE(4790), 1,
      sym_attribute,
    STATE(4859), 1,
      sym__attribute,
    STATE(4877), 1,
      sym_type_binding,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6064), 1,
      sym__type_params,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(7746), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [197525] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7097), 1,
      anon_sym_EQ,
    ACTIONS(7099), 1,
      anon_sym_COLON2,
    ACTIONS(7103), 1,
      anon_sym_COLON_GT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4791), 1,
      sym_attribute,
    STATE(4960), 1,
      sym__coerced,
    STATE(4989), 1,
      sym__typed,
    STATE(4998), 1,
      sym__type_constrained,
    STATE(5044), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7101), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [197574] = 14,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    ACTIONS(6855), 1,
      anon_sym_val,
    STATE(3580), 1,
      sym_module_path,
    STATE(4792), 1,
      sym_attribute,
    STATE(5150), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [197625] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7099), 1,
      anon_sym_COLON2,
    ACTIONS(7103), 1,
      anon_sym_COLON_GT,
    ACTIONS(7105), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4793), 1,
      sym_attribute,
    STATE(4960), 1,
      sym__coerced,
    STATE(4981), 1,
      sym__type_constrained,
    STATE(4989), 1,
      sym__typed,
    STATE(5037), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7107), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [197674] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2149), 1,
      anon_sym_let,
    STATE(4794), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2151), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197707] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6803), 1,
      anon_sym_let,
    ACTIONS(7109), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4795), 2,
      sym_attribute,
      aux_sym_class_definition_repeat1,
    ACTIONS(6801), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197742] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2195), 1,
      anon_sym_let,
    STATE(4796), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2197), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197775] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6810), 1,
      anon_sym_let,
    ACTIONS(7071), 1,
      anon_sym_and,
    STATE(4795), 1,
      aux_sym_class_definition_repeat1,
    STATE(4797), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6808), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197812] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7059), 1,
      anon_sym_let,
    ACTIONS(7112), 1,
      anon_sym_in,
    STATE(4798), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7057), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197847] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7114), 1,
      anon_sym_let,
    STATE(4799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4292), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197880] = 19,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    ACTIONS(7116), 1,
      sym__lowercase_identifier,
    ACTIONS(7118), 1,
      anon_sym_nonrec,
    STATE(4549), 1,
      sym_type_binding,
    STATE(4800), 1,
      sym_attribute,
    STATE(4890), 1,
      sym__attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6274), 1,
      sym__type_params,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(8313), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [197941] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1979), 1,
      anon_sym_let,
    STATE(4801), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1981), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [197974] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2121), 1,
      anon_sym_let,
    STATE(4802), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2123), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [198007] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7099), 1,
      anon_sym_COLON2,
    ACTIONS(7103), 1,
      anon_sym_COLON_GT,
    ACTIONS(7120), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4803), 1,
      sym_attribute,
    STATE(4960), 1,
      sym__coerced,
    STATE(4989), 1,
      sym__typed,
    STATE(5004), 1,
      sym__type_constrained,
    STATE(5138), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7122), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [198056] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6755), 1,
      anon_sym_let,
    ACTIONS(7071), 1,
      anon_sym_and,
    STATE(4795), 1,
      aux_sym_class_definition_repeat1,
    STATE(4804), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6753), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [198093] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6767), 1,
      anon_sym_let,
    ACTIONS(7071), 1,
      anon_sym_and,
    STATE(4804), 1,
      aux_sym_class_definition_repeat1,
    STATE(4805), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6765), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [198130] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7126), 1,
      anon_sym_let,
    STATE(4806), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7124), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [198163] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6454), 1,
      anon_sym_LPAREN,
    ACTIONS(6460), 1,
      anon_sym_struct,
    ACTIONS(6462), 1,
      sym__uppercase_identifier,
    ACTIONS(7128), 1,
      anon_sym_functor,
    STATE(4807), 1,
      sym_attribute,
    STATE(4949), 1,
      sym__module_expression,
    STATE(5201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5249), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [198211] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6454), 1,
      anon_sym_LPAREN,
    ACTIONS(6460), 1,
      anon_sym_struct,
    ACTIONS(6462), 1,
      sym__uppercase_identifier,
    ACTIONS(7128), 1,
      anon_sym_functor,
    STATE(4808), 1,
      sym_attribute,
    STATE(4865), 1,
      sym__module_expression,
    STATE(5201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5249), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [198259] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6454), 1,
      anon_sym_LPAREN,
    ACTIONS(6458), 1,
      anon_sym_functor,
    ACTIONS(6460), 1,
      anon_sym_struct,
    ACTIONS(6462), 1,
      sym__uppercase_identifier,
    STATE(4809), 1,
      sym_attribute,
    STATE(5013), 1,
      sym__module_expression,
    STATE(5201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5249), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [198307] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4810), 1,
      sym_attribute,
    STATE(5122), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [198355] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6454), 1,
      anon_sym_LPAREN,
    ACTIONS(6460), 1,
      anon_sym_struct,
    ACTIONS(6462), 1,
      sym__uppercase_identifier,
    ACTIONS(7128), 1,
      anon_sym_functor,
    STATE(4811), 1,
      sym_attribute,
    STATE(4885), 1,
      sym__module_expression,
    STATE(5201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5249), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [198403] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4812), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2103), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(2105), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [198435] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4813), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2169), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(2171), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [198467] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4814), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2107), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(2109), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [198499] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6591), 1,
      anon_sym_functor,
    ACTIONS(6593), 1,
      anon_sym_struct,
    STATE(3580), 1,
      sym_module_path,
    STATE(3687), 1,
      sym__module_expression,
    STATE(4815), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [198547] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4816), 1,
      sym_attribute,
    STATE(5077), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [198595] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4817), 1,
      sym_attribute,
    STATE(5066), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [198643] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7063), 1,
      anon_sym_val,
    ACTIONS(7067), 1,
      anon_sym_inherit,
    ACTIONS(7069), 1,
      anon_sym_method,
    ACTIONS(7130), 1,
      anon_sym_end,
    STATE(4818), 1,
      sym_attribute,
    STATE(4828), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5356), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(5357), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [198693] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4819), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2117), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(2119), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [198725] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7063), 1,
      anon_sym_val,
    ACTIONS(7067), 1,
      anon_sym_inherit,
    ACTIONS(7069), 1,
      anon_sym_method,
    ACTIONS(7132), 1,
      anon_sym_end,
    STATE(4818), 1,
      aux_sym_class_body_type_repeat1,
    STATE(4820), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5356), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(5357), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [198775] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6454), 1,
      anon_sym_LPAREN,
    ACTIONS(6458), 1,
      anon_sym_functor,
    ACTIONS(6460), 1,
      anon_sym_struct,
    ACTIONS(6462), 1,
      sym__uppercase_identifier,
    STATE(4821), 1,
      sym_attribute,
    STATE(4991), 1,
      sym__module_expression,
    STATE(5201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5249), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [198823] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6591), 1,
      anon_sym_functor,
    ACTIONS(6593), 1,
      anon_sym_struct,
    STATE(3531), 1,
      sym__module_expression,
    STATE(3580), 1,
      sym_module_path,
    STATE(4822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [198871] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7059), 1,
      anon_sym_let,
    ACTIONS(7134), 1,
      anon_sym_in,
    STATE(4823), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7057), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [198905] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7063), 1,
      anon_sym_val,
    ACTIONS(7067), 1,
      anon_sym_inherit,
    ACTIONS(7069), 1,
      anon_sym_method,
    ACTIONS(7136), 1,
      anon_sym_end,
    STATE(4824), 1,
      sym_attribute,
    STATE(4838), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5356), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(5357), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [198955] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6591), 1,
      anon_sym_functor,
    ACTIONS(6593), 1,
      anon_sym_struct,
    STATE(3580), 1,
      sym_module_path,
    STATE(3705), 1,
      sym__module_expression,
    STATE(4825), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199003] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4826), 1,
      sym_attribute,
    STATE(5063), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199051] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6454), 1,
      anon_sym_LPAREN,
    ACTIONS(6458), 1,
      anon_sym_functor,
    ACTIONS(6460), 1,
      anon_sym_struct,
    ACTIONS(6462), 1,
      sym__uppercase_identifier,
    STATE(4827), 1,
      sym_attribute,
    STATE(5012), 1,
      sym__module_expression,
    STATE(5201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5249), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199099] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7138), 1,
      anon_sym_constraint,
    ACTIONS(7141), 1,
      anon_sym_val,
    ACTIONS(7144), 1,
      anon_sym_end,
    ACTIONS(7146), 1,
      anon_sym_inherit,
    ACTIONS(7149), 1,
      anon_sym_method,
    ACTIONS(7152), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(7155), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(7158), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4828), 2,
      sym_attribute,
      aux_sym_class_body_type_repeat1,
    STATE(5356), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(5357), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [199147] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6591), 1,
      anon_sym_functor,
    ACTIONS(6593), 1,
      anon_sym_struct,
    STATE(3580), 1,
      sym_module_path,
    STATE(3737), 1,
      sym__module_expression,
    STATE(4829), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199195] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4830), 1,
      sym_attribute,
    STATE(5136), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199243] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6591), 1,
      anon_sym_functor,
    ACTIONS(6593), 1,
      anon_sym_struct,
    STATE(3236), 1,
      sym__module_expression,
    STATE(3580), 1,
      sym_module_path,
    STATE(4831), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199291] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6454), 1,
      anon_sym_LPAREN,
    ACTIONS(6460), 1,
      anon_sym_struct,
    ACTIONS(6462), 1,
      sym__uppercase_identifier,
    ACTIONS(7128), 1,
      anon_sym_functor,
    STATE(4832), 1,
      sym_attribute,
    STATE(4968), 1,
      sym__module_expression,
    STATE(5201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5249), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199339] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4833), 1,
      sym_attribute,
    STATE(5116), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199387] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4834), 1,
      sym_attribute,
    STATE(5072), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199435] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6591), 1,
      anon_sym_functor,
    ACTIONS(6593), 1,
      anon_sym_struct,
    STATE(3515), 1,
      sym__module_expression,
    STATE(3580), 1,
      sym_module_path,
    STATE(4835), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199483] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6591), 1,
      anon_sym_functor,
    ACTIONS(6593), 1,
      anon_sym_struct,
    STATE(3576), 1,
      sym__module_expression,
    STATE(3580), 1,
      sym_module_path,
    STATE(4836), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199531] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6591), 1,
      anon_sym_functor,
    ACTIONS(6593), 1,
      anon_sym_struct,
    STATE(3577), 1,
      sym__module_expression,
    STATE(3580), 1,
      sym_module_path,
    STATE(4837), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199579] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7063), 1,
      anon_sym_val,
    ACTIONS(7067), 1,
      anon_sym_inherit,
    ACTIONS(7069), 1,
      anon_sym_method,
    ACTIONS(7161), 1,
      anon_sym_end,
    STATE(4828), 1,
      aux_sym_class_body_type_repeat1,
    STATE(4838), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5356), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(5357), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [199629] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4839), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7163), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [199665] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1853), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4171), 1,
      aux_sym_expression_item_repeat1,
    STATE(4337), 1,
      sym_item_attribute,
    STATE(4840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7165), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [199701] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6591), 1,
      anon_sym_functor,
    ACTIONS(6593), 1,
      anon_sym_struct,
    STATE(3241), 1,
      sym__module_expression,
    STATE(3580), 1,
      sym_module_path,
    STATE(4841), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199749] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4842), 1,
      sym_attribute,
    STATE(5105), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199797] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4843), 1,
      sym_attribute,
    STATE(5109), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199845] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7063), 1,
      anon_sym_val,
    ACTIONS(7067), 1,
      anon_sym_inherit,
    ACTIONS(7069), 1,
      anon_sym_method,
    ACTIONS(7167), 1,
      anon_sym_end,
    STATE(4828), 1,
      aux_sym_class_body_type_repeat1,
    STATE(4844), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5356), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(5357), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [199895] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4845), 1,
      sym_attribute,
    STATE(5030), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [199943] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(73), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_constraint,
    ACTIONS(5713), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5715), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(7063), 1,
      anon_sym_val,
    ACTIONS(7067), 1,
      anon_sym_inherit,
    ACTIONS(7069), 1,
      anon_sym_method,
    ACTIONS(7169), 1,
      anon_sym_end,
    STATE(4828), 1,
      aux_sym_class_body_type_repeat1,
    STATE(4846), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5356), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(5357), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [199993] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4847), 1,
      sym_attribute,
    STATE(4965), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [200041] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4848), 1,
      sym_attribute,
    STATE(5102), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [200089] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6591), 1,
      anon_sym_functor,
    ACTIONS(6593), 1,
      anon_sym_struct,
    STATE(3517), 1,
      sym__module_expression,
    STATE(3580), 1,
      sym_module_path,
    STATE(4849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [200137] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4850), 1,
      sym_attribute,
    STATE(5154), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [200185] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4851), 1,
      sym_attribute,
    STATE(5041), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [200233] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6591), 1,
      anon_sym_functor,
    ACTIONS(6593), 1,
      anon_sym_struct,
    STATE(3535), 1,
      sym__module_expression,
    STATE(3580), 1,
      sym_module_path,
    STATE(4852), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [200281] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6454), 1,
      anon_sym_LPAREN,
    ACTIONS(6460), 1,
      anon_sym_struct,
    ACTIONS(6462), 1,
      sym__uppercase_identifier,
    ACTIONS(7128), 1,
      anon_sym_functor,
    STATE(4853), 1,
      sym_attribute,
    STATE(4954), 1,
      sym__module_expression,
    STATE(5201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5249), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [200329] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6454), 1,
      anon_sym_LPAREN,
    ACTIONS(6460), 1,
      anon_sym_struct,
    ACTIONS(6462), 1,
      sym__uppercase_identifier,
    ACTIONS(7128), 1,
      anon_sym_functor,
    STATE(4854), 1,
      sym_attribute,
    STATE(4956), 1,
      sym__module_expression,
    STATE(5201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5249), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [200377] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6454), 1,
      anon_sym_LPAREN,
    ACTIONS(6460), 1,
      anon_sym_struct,
    ACTIONS(6462), 1,
      sym__uppercase_identifier,
    ACTIONS(7128), 1,
      anon_sym_functor,
    STATE(4855), 1,
      sym_attribute,
    STATE(4943), 1,
      sym__module_expression,
    STATE(5201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5249), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [200425] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6454), 1,
      anon_sym_LPAREN,
    ACTIONS(6460), 1,
      anon_sym_struct,
    ACTIONS(6462), 1,
      sym__uppercase_identifier,
    ACTIONS(7128), 1,
      anon_sym_functor,
    STATE(4856), 1,
      sym_attribute,
    STATE(4955), 1,
      sym__module_expression,
    STATE(5201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5249), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [200473] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6454), 1,
      anon_sym_LPAREN,
    ACTIONS(6458), 1,
      anon_sym_functor,
    ACTIONS(6460), 1,
      anon_sym_struct,
    ACTIONS(6462), 1,
      sym__uppercase_identifier,
    STATE(4857), 1,
      sym_attribute,
    STATE(5024), 1,
      sym__module_expression,
    STATE(5201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5249), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [200521] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(6587), 1,
      anon_sym_LPAREN,
    ACTIONS(6593), 1,
      anon_sym_struct,
    ACTIONS(6617), 1,
      anon_sym_functor,
    STATE(3580), 1,
      sym_module_path,
    STATE(4858), 1,
      sym_attribute,
    STATE(5092), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3670), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [200569] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    ACTIONS(7093), 1,
      sym__lowercase_identifier,
    ACTIONS(7171), 1,
      anon_sym_nonrec,
    STATE(4859), 1,
      sym_attribute,
    STATE(4868), 1,
      sym_type_binding,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6064), 1,
      sym__type_params,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(7746), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [200624] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7173), 1,
      anon_sym_DQUOTE,
    STATE(4860), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7456), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [200665] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7181), 1,
      anon_sym_DQUOTE,
    STATE(4861), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(8160), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [200706] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7183), 1,
      anon_sym_and,
    STATE(4862), 1,
      sym_attribute,
    STATE(4870), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6561), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [200739] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2161), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(2163), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [200770] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7185), 1,
      anon_sym_DQUOTE,
    STATE(4864), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7234), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [200811] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7187), 1,
      anon_sym_LPAREN,
    STATE(4865), 1,
      sym_attribute,
    STATE(5229), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4395), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [200850] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7183), 1,
      anon_sym_and,
    STATE(4866), 1,
      sym_attribute,
    STATE(4870), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6441), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [200883] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7183), 1,
      anon_sym_and,
    STATE(4867), 1,
      sym_attribute,
    STATE(4870), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6520), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [200916] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7183), 1,
      anon_sym_and,
    STATE(4866), 1,
      aux_sym_type_definition_repeat1,
    STATE(4868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6561), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [200949] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7189), 1,
      anon_sym_DQUOTE,
    STATE(4869), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7327), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [200990] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7191), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4870), 2,
      sym_attribute,
      aux_sym_type_definition_repeat1,
    ACTIONS(6477), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [201021] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7183), 1,
      anon_sym_and,
    STATE(4867), 1,
      aux_sym_type_definition_repeat1,
    STATE(4871), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6441), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [201054] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7194), 1,
      anon_sym_and,
    STATE(4872), 1,
      sym_attribute,
    STATE(4888), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6753), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [201087] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7196), 1,
      anon_sym_DQUOTE,
    STATE(4873), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7705), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [201128] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7194), 1,
      anon_sym_and,
    STATE(4874), 1,
      sym_attribute,
    STATE(4889), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6753), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [201161] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7198), 1,
      anon_sym_DQUOTE,
    STATE(4875), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7584), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [201202] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7200), 1,
      anon_sym_DQUOTE,
    STATE(4876), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(8091), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [201243] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7183), 1,
      anon_sym_and,
    STATE(4862), 1,
      aux_sym_type_definition_repeat1,
    STATE(4877), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6583), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [201276] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6567), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7202), 1,
      anon_sym_as,
    STATE(4878), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6565), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [201309] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6571), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7202), 1,
      anon_sym_as,
    STATE(4879), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6569), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [201342] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7204), 1,
      anon_sym_DQUOTE,
    STATE(4880), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7957), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [201383] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7206), 1,
      anon_sym_DQUOTE,
    STATE(4881), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7888), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [201424] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7208), 1,
      anon_sym_DQUOTE,
    STATE(4882), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7673), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [201465] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7210), 1,
      anon_sym_DQUOTE,
    STATE(4883), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7208), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [201506] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7212), 1,
      anon_sym_DQUOTE,
    STATE(4884), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7724), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [201547] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7187), 1,
      anon_sym_LPAREN,
    STATE(4885), 1,
      sym_attribute,
    STATE(5229), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4417), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [201586] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7214), 1,
      anon_sym_DQUOTE,
    STATE(4886), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7474), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [201627] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4887), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2129), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(2131), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [201658] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7216), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4888), 2,
      sym_attribute,
      aux_sym_class_definition_repeat1,
    ACTIONS(6801), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [201689] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7194), 1,
      anon_sym_and,
    STATE(4888), 1,
      aux_sym_class_definition_repeat1,
    STATE(4889), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6808), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [201722] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    ACTIONS(7116), 1,
      sym__lowercase_identifier,
    ACTIONS(7219), 1,
      anon_sym_nonrec,
    STATE(4544), 1,
      sym_type_binding,
    STATE(4890), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6274), 1,
      sym__type_params,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(8313), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [201777] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7221), 1,
      anon_sym_DQUOTE,
    STATE(4891), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7978), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [201818] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7223), 1,
      anon_sym_DQUOTE,
    STATE(4892), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7633), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [201859] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7194), 1,
      anon_sym_and,
    STATE(4872), 1,
      aux_sym_class_definition_repeat1,
    STATE(4893), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6765), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [201892] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7225), 1,
      anon_sym_DQUOTE,
    STATE(4894), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(8251), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [201933] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7227), 1,
      anon_sym_DQUOTE,
    STATE(4895), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7329), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [201974] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7229), 1,
      anon_sym_DQUOTE,
    STATE(4896), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7904), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [202015] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7231), 1,
      anon_sym_DQUOTE,
    STATE(4897), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(7666), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [202056] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5192), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7202), 1,
      anon_sym_as,
    STATE(4898), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5190), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [202089] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7233), 1,
      anon_sym_DQUOTE,
    STATE(4899), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(8154), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [202130] = 17,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7081), 1,
      sym__lowercase_identifier,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    ACTIONS(7235), 1,
      anon_sym_nonrec,
    STATE(4900), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6190), 1,
      sym__type_params,
    STATE(6418), 1,
      sym_type_binding,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(7435), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [202185] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7237), 1,
      anon_sym_and,
    STATE(4901), 1,
      sym_attribute,
    STATE(4903), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6415), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [202218] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7237), 1,
      anon_sym_and,
    STATE(4902), 1,
      sym_attribute,
    STATE(4903), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6421), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [202251] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7239), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4903), 2,
      sym_attribute,
      aux_sym_module_definition_repeat1,
    ACTIONS(6447), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [202282] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7237), 1,
      anon_sym_and,
    STATE(4904), 1,
      sym_attribute,
    STATE(4905), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6421), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [202315] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7237), 1,
      anon_sym_and,
    STATE(4903), 1,
      aux_sym_module_definition_repeat1,
    STATE(4905), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6498), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [202348] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7237), 1,
      anon_sym_and,
    STATE(4901), 1,
      aux_sym_module_definition_repeat1,
    STATE(4906), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6510), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [202381] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7237), 1,
      anon_sym_and,
    STATE(4902), 1,
      aux_sym_module_definition_repeat1,
    STATE(4907), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6415), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [202414] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7242), 1,
      anon_sym_DQUOTE,
    STATE(4908), 1,
      sym_attribute,
    STATE(4910), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    STATE(8335), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [202455] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4272), 1,
      anon_sym_let,
    ACTIONS(4278), 1,
      anon_sym_object,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7244), 1,
      sym__lowercase_identifier,
    ACTIONS(7246), 1,
      anon_sym_LBRACK,
    STATE(4909), 1,
      sym_attribute,
    STATE(5984), 1,
      sym__simple_class_type,
    STATE(6897), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6113), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [202501] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7248), 1,
      anon_sym_DQUOTE,
    STATE(4910), 1,
      sym_attribute,
    STATE(4933), 1,
      aux_sym_string_content_repeat1,
    STATE(5032), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7177), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7179), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7175), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [202539] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2317), 2,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7250), 13,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [202569] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4254), 1,
      anon_sym_let,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7252), 1,
      sym__lowercase_identifier,
    ACTIONS(7254), 1,
      anon_sym_LBRACK,
    STATE(4260), 1,
      sym__simple_class_type,
    STATE(4912), 1,
      sym_attribute,
    STATE(6859), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4081), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [202615] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(4264), 1,
      anon_sym_let,
    ACTIONS(7256), 1,
      sym__lowercase_identifier,
    ACTIONS(7258), 1,
      anon_sym_LBRACK,
    STATE(4070), 1,
      sym__simple_class_type,
    STATE(4913), 1,
      sym_attribute,
    STATE(7182), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4081), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [202661] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4254), 1,
      anon_sym_let,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7252), 1,
      sym__lowercase_identifier,
    ACTIONS(7254), 1,
      anon_sym_LBRACK,
    STATE(4353), 1,
      sym__simple_class_type,
    STATE(4914), 1,
      sym_attribute,
    STATE(6859), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4081), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [202707] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4254), 1,
      anon_sym_let,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7252), 1,
      sym__lowercase_identifier,
    ACTIONS(7254), 1,
      anon_sym_LBRACK,
    STATE(4070), 1,
      sym__simple_class_type,
    STATE(4915), 1,
      sym_attribute,
    STATE(6859), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4081), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [202753] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4254), 1,
      anon_sym_let,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7252), 1,
      sym__lowercase_identifier,
    ACTIONS(7254), 1,
      anon_sym_LBRACK,
    STATE(4287), 1,
      sym__simple_class_type,
    STATE(4916), 1,
      sym_attribute,
    STATE(6859), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4081), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [202799] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    ACTIONS(7116), 1,
      sym__lowercase_identifier,
    STATE(4512), 1,
      sym_type_binding,
    STATE(4917), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6274), 1,
      sym__type_params,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(8313), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [202851] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    ACTIONS(7116), 1,
      sym__lowercase_identifier,
    STATE(4633), 1,
      sym_type_binding,
    STATE(4918), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6274), 1,
      sym__type_params,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(8313), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [202903] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    ACTIONS(7093), 1,
      sym__lowercase_identifier,
    STATE(4633), 1,
      sym_type_binding,
    STATE(4919), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6064), 1,
      sym__type_params,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(7746), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [202955] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7081), 1,
      sym__lowercase_identifier,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    STATE(4920), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6190), 1,
      sym__type_params,
    STATE(6418), 1,
      sym_type_binding,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(7435), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [203007] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    ACTIONS(7116), 1,
      sym__lowercase_identifier,
    STATE(4544), 1,
      sym_type_binding,
    STATE(4921), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6274), 1,
      sym__type_params,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(8313), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [203059] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    ACTIONS(7093), 1,
      sym__lowercase_identifier,
    STATE(4871), 1,
      sym_type_binding,
    STATE(4922), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6064), 1,
      sym__type_params,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(7746), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [203111] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2604), 1,
      aux_sym_number_token1,
    ACTIONS(2606), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2608), 1,
      anon_sym_SQUOTE,
    ACTIONS(2610), 1,
      anon_sym_DQUOTE,
    ACTIONS(7262), 1,
      anon_sym_LBRACE,
    STATE(4923), 1,
      sym_attribute,
    STATE(5606), 1,
      sym__signed_constant,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7260), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5451), 2,
      sym__constant,
      sym_signed_number,
    STATE(5427), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
  [203155] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7081), 1,
      sym__lowercase_identifier,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    STATE(4924), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6190), 1,
      sym__type_params,
    STATE(6559), 1,
      sym_type_binding,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(7435), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [203207] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4254), 1,
      anon_sym_let,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7252), 1,
      sym__lowercase_identifier,
    ACTIONS(7254), 1,
      anon_sym_LBRACK,
    STATE(4309), 1,
      sym__simple_class_type,
    STATE(4925), 1,
      sym_attribute,
    STATE(6859), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4081), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [203253] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1673), 1,
      aux_sym_number_token1,
    ACTIONS(1675), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1677), 1,
      anon_sym_SQUOTE,
    ACTIONS(1679), 1,
      anon_sym_DQUOTE,
    ACTIONS(7266), 1,
      anon_sym_LBRACE,
    STATE(2330), 1,
      sym__signed_constant,
    STATE(4926), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7264), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2134), 2,
      sym__constant,
      sym_signed_number,
    STATE(2123), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
  [203297] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7256), 1,
      sym__lowercase_identifier,
    ACTIONS(7258), 1,
      anon_sym_LBRACK,
    ACTIONS(7268), 1,
      anon_sym_let,
    STATE(4927), 1,
      sym_attribute,
    STATE(5175), 1,
      sym__simple_class_type,
    STATE(7182), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4081), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [203343] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4272), 1,
      anon_sym_let,
    ACTIONS(4278), 1,
      anon_sym_object,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7244), 1,
      sym__lowercase_identifier,
    ACTIONS(7246), 1,
      anon_sym_LBRACK,
    STATE(4928), 1,
      sym_attribute,
    STATE(6002), 1,
      sym__simple_class_type,
    STATE(6897), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6113), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [203389] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7081), 1,
      sym__lowercase_identifier,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    STATE(4929), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6190), 1,
      sym__type_params,
    STATE(6947), 1,
      sym_type_binding,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(7435), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [203441] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4558), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6551), 1,
      anon_sym_STAR,
    STATE(4930), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4556), 13,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [203473] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7270), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4931), 2,
      sym_attribute,
      aux_sym__structure_repeat1,
    ACTIONS(3229), 13,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [203503] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1395), 1,
      aux_sym_number_token1,
    ACTIONS(1399), 1,
      anon_sym_SQUOTE,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(2566), 1,
      aux_sym_signed_number_token1,
    ACTIONS(7275), 1,
      anon_sym_LBRACE,
    STATE(4932), 1,
      sym_attribute,
    STATE(5155), 1,
      sym__signed_constant,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7273), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5005), 2,
      sym__constant,
      sym_signed_number,
    STATE(4819), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
  [203547] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7277), 1,
      anon_sym_DQUOTE,
    STATE(5032), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7282), 2,
      sym__null,
      aux_sym_string_content_token2,
    STATE(4933), 2,
      sym_attribute,
      aux_sym_string_content_repeat1,
    ACTIONS(7285), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(7279), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [203583] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7256), 1,
      sym__lowercase_identifier,
    ACTIONS(7258), 1,
      anon_sym_LBRACK,
    ACTIONS(7268), 1,
      anon_sym_let,
    STATE(4070), 1,
      sym__simple_class_type,
    STATE(4934), 1,
      sym_attribute,
    STATE(7182), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4081), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [203629] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4272), 1,
      anon_sym_let,
    ACTIONS(4278), 1,
      anon_sym_object,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7244), 1,
      sym__lowercase_identifier,
    ACTIONS(7246), 1,
      anon_sym_LBRACK,
    STATE(4935), 1,
      sym_attribute,
    STATE(6246), 1,
      sym__simple_class_type,
    STATE(6897), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6113), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [203675] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1801), 1,
      aux_sym_number_token1,
    ACTIONS(1803), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1805), 1,
      anon_sym_SQUOTE,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(7290), 1,
      anon_sym_LBRACE,
    STATE(3664), 1,
      sym__signed_constant,
    STATE(4936), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7288), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3528), 2,
      sym__constant,
      sym_signed_number,
    STATE(3547), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
  [203719] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4272), 1,
      anon_sym_let,
    ACTIONS(4278), 1,
      anon_sym_object,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7244), 1,
      sym__lowercase_identifier,
    ACTIONS(7246), 1,
      anon_sym_LBRACK,
    STATE(4937), 1,
      sym_attribute,
    STATE(5948), 1,
      sym__simple_class_type,
    STATE(6897), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6113), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [203765] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4254), 1,
      anon_sym_let,
    ACTIONS(4260), 1,
      anon_sym_object,
    ACTIONS(5632), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5634), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7252), 1,
      sym__lowercase_identifier,
    ACTIONS(7254), 1,
      anon_sym_LBRACK,
    STATE(4213), 1,
      sym__simple_class_type,
    STATE(4938), 1,
      sym_attribute,
    STATE(6859), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4081), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [203811] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4272), 1,
      anon_sym_let,
    ACTIONS(4278), 1,
      anon_sym_object,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7244), 1,
      sym__lowercase_identifier,
    ACTIONS(7246), 1,
      anon_sym_LBRACK,
    STATE(4939), 1,
      sym_attribute,
    STATE(5909), 1,
      sym__simple_class_type,
    STATE(6897), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6113), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [203857] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4272), 1,
      anon_sym_let,
    ACTIONS(4278), 1,
      anon_sym_object,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7244), 1,
      sym__lowercase_identifier,
    ACTIONS(7246), 1,
      anon_sym_LBRACK,
    STATE(4940), 1,
      sym_attribute,
    STATE(5811), 1,
      sym__simple_class_type,
    STATE(6897), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6113), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [203903] = 16,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    ACTIONS(7093), 1,
      sym__lowercase_identifier,
    STATE(4868), 1,
      sym_type_binding,
    STATE(4941), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6064), 1,
      sym__type_params,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(7746), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [203955] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7292), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [203982] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7187), 1,
      anon_sym_LPAREN,
    STATE(4943), 1,
      sym_attribute,
    STATE(5229), 1,
      sym__simple_module_expression,
    STATE(5905), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4644), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [204025] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4437), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6539), 1,
      anon_sym_DASH_GT,
    STATE(4944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4435), 12,
      anon_sym_EQ,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [204056] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4945), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7294), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [204083] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5192), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7296), 1,
      anon_sym_as,
    STATE(4946), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5190), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [204114] = 14,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3846), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    ACTIONS(7298), 1,
      anon_sym_private,
    ACTIONS(7300), 1,
      anon_sym_DOT_DOT,
    ACTIONS(7302), 1,
      anon_sym_LBRACE,
    STATE(4646), 1,
      sym_constructor_declaration,
    STATE(4947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4492), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [204161] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    STATE(4948), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6117), 1,
      sym__type_params,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(7169), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [204210] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7187), 1,
      anon_sym_LPAREN,
    STATE(4949), 1,
      sym_attribute,
    STATE(5229), 1,
      sym__simple_module_expression,
    STATE(5850), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4698), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [204253] = 14,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3740), 1,
      anon_sym_PIPE,
    ACTIONS(7304), 1,
      anon_sym_LPAREN,
    ACTIONS(7306), 1,
      anon_sym_private,
    ACTIONS(7308), 1,
      anon_sym_DOT_DOT,
    ACTIONS(7310), 1,
      anon_sym_LBRACK,
    ACTIONS(7314), 1,
      anon_sym_LBRACE,
    ACTIONS(7316), 1,
      sym__uppercase_identifier,
    STATE(4950), 1,
      sym_attribute,
    STATE(5629), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7312), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(5370), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [204300] = 14,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3774), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    ACTIONS(7302), 1,
      anon_sym_LBRACE,
    ACTIONS(7318), 1,
      anon_sym_private,
    ACTIONS(7320), 1,
      anon_sym_DOT_DOT,
    STATE(4197), 1,
      sym_constructor_declaration,
    STATE(4951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4127), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [204347] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    STATE(4952), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6221), 1,
      sym__type_params,
    STATE(6765), 1,
      sym_type_constructor_path,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [204396] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7324), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7322), 13,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [204425] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7187), 1,
      anon_sym_LPAREN,
    STATE(4954), 1,
      sym_attribute,
    STATE(5229), 1,
      sym__simple_module_expression,
    STATE(5898), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4596), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [204468] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7187), 1,
      anon_sym_LPAREN,
    STATE(4955), 1,
      sym_attribute,
    STATE(5229), 1,
      sym__simple_module_expression,
    STATE(5851), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4698), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [204511] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7187), 1,
      anon_sym_LPAREN,
    STATE(4956), 1,
      sym_attribute,
    STATE(5229), 1,
      sym__simple_module_expression,
    STATE(5900), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4596), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [204554] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6838), 1,
      anon_sym_STAR,
    STATE(4957), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4556), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [204583] = 14,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3846), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    ACTIONS(7302), 1,
      anon_sym_LBRACE,
    ACTIONS(7326), 1,
      anon_sym_private,
    ACTIONS(7328), 1,
      anon_sym_DOT_DOT,
    STATE(4646), 1,
      sym_constructor_declaration,
    STATE(4958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4425), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [204630] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6571), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7296), 1,
      anon_sym_as,
    STATE(4959), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6569), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [204661] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7332), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7330), 13,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [204690] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    STATE(4961), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6290), 1,
      sym__type_params,
    STATE(7113), 1,
      sym__type_param,
    STATE(7144), 1,
      sym_type_constructor_path,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [204739] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(7336), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4962), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7334), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [204770] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2137), 1,
      anon_sym_COLON2,
    STATE(4963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2139), 13,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [204799] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6567), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7296), 1,
      anon_sym_as,
    STATE(4964), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6565), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [204830] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4397), 1,
      anon_sym_COLON2,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(4965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4395), 4,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [204869] = 14,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3740), 1,
      anon_sym_PIPE,
    ACTIONS(7304), 1,
      anon_sym_LPAREN,
    ACTIONS(7310), 1,
      anon_sym_LBRACK,
    ACTIONS(7314), 1,
      anon_sym_LBRACE,
    ACTIONS(7316), 1,
      sym__uppercase_identifier,
    ACTIONS(7338), 1,
      anon_sym_private,
    ACTIONS(7340), 1,
      anon_sym_DOT_DOT,
    STATE(4966), 1,
      sym_attribute,
    STATE(5629), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7312), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(5352), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [204916] = 15,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(7083), 1,
      anon_sym_LPAREN,
    ACTIONS(7089), 1,
      anon_sym__,
    STATE(4967), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6158), 1,
      sym__type_params,
    STATE(7040), 1,
      sym_type_constructor_path,
    STATE(7113), 1,
      sym__type_param,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [204965] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(6232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7187), 1,
      anon_sym_LPAREN,
    STATE(4968), 1,
      sym_attribute,
    STATE(5229), 1,
      sym__simple_module_expression,
    STATE(5977), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4632), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [205008] = 14,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3774), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    ACTIONS(7302), 1,
      anon_sym_LBRACE,
    ACTIONS(7342), 1,
      anon_sym_private,
    ACTIONS(7344), 1,
      anon_sym_DOT_DOT,
    STATE(4197), 1,
      sym_constructor_declaration,
    STATE(4969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4049), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [205055] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4970), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3280), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [205082] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6883), 1,
      anon_sym_STAR,
    STATE(4971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4556), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_AMP,
  [205111] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4825), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [205138] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4973), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4439), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [205164] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2582), 1,
      anon_sym_LBRACK,
    ACTIONS(2592), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7346), 1,
      anon_sym_LPAREN,
    ACTIONS(7348), 1,
      anon_sym_LBRACE,
    ACTIONS(7350), 1,
      anon_sym_begin,
    ACTIONS(7352), 1,
      sym__uppercase_identifier,
    STATE(3880), 1,
      sym__constructor_name,
    STATE(4974), 1,
      sym_attribute,
    STATE(5660), 1,
      sym__delimited_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5636), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [205206] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1425), 1,
      anon_sym_LBRACK,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7354), 1,
      anon_sym_LPAREN,
    ACTIONS(7356), 1,
      anon_sym_LBRACE,
    ACTIONS(7358), 1,
      anon_sym_begin,
    ACTIONS(7360), 1,
      sym__uppercase_identifier,
    STATE(3624), 1,
      sym__constructor_name,
    STATE(4975), 1,
      sym_attribute,
    STATE(5271), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5312), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [205248] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7362), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4976), 1,
      sym_attribute,
    STATE(5061), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7364), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [205282] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1651), 1,
      anon_sym_LBRACK,
    ACTIONS(1661), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7366), 1,
      anon_sym_LPAREN,
    ACTIONS(7368), 1,
      anon_sym_LBRACE,
    ACTIONS(7370), 1,
      anon_sym_begin,
    ACTIONS(7372), 1,
      sym__uppercase_identifier,
    STATE(2100), 1,
      sym__constructor_name,
    STATE(2425), 1,
      sym__delimited_binding_pattern,
    STATE(4977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2438), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [205324] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7374), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4978), 1,
      sym_attribute,
    STATE(5090), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7376), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [205358] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7380), 1,
      anon_sym_as,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4979), 1,
      sym_attribute,
    STATE(5051), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7378), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [205392] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3846), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    ACTIONS(7302), 1,
      anon_sym_LBRACE,
    ACTIONS(7382), 1,
      anon_sym_DOT_DOT,
    STATE(4646), 1,
      sym_constructor_declaration,
    STATE(4980), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4429), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [205436] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7384), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4981), 1,
      sym_attribute,
    STATE(5140), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7386), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [205470] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1369), 1,
      anon_sym_LBRACK,
    ACTIONS(1383), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7358), 1,
      anon_sym_begin,
    ACTIONS(7360), 1,
      sym__uppercase_identifier,
    ACTIONS(7388), 1,
      anon_sym_LPAREN,
    ACTIONS(7390), 1,
      anon_sym_LBRACE,
    STATE(3624), 1,
      sym__constructor_name,
    STATE(4982), 1,
      sym_attribute,
    STATE(5170), 1,
      sym__delimited_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5133), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [205512] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6761), 1,
      anon_sym_DASH_GT,
    STATE(4983), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4435), 12,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [205540] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7392), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4984), 1,
      sym_attribute,
    STATE(5052), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7394), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [205574] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3740), 1,
      anon_sym_PIPE,
    ACTIONS(7304), 1,
      anon_sym_LPAREN,
    ACTIONS(7310), 1,
      anon_sym_LBRACK,
    ACTIONS(7314), 1,
      anon_sym_LBRACE,
    ACTIONS(7316), 1,
      sym__uppercase_identifier,
    ACTIONS(7396), 1,
      anon_sym_DOT_DOT,
    STATE(4985), 1,
      sym_attribute,
    STATE(5629), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7312), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(5378), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [205618] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4986), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4405), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [205644] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2622), 1,
      anon_sym_LBRACK,
    ACTIONS(2628), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7350), 1,
      anon_sym_begin,
    ACTIONS(7352), 1,
      sym__uppercase_identifier,
    ACTIONS(7398), 1,
      anon_sym_LPAREN,
    ACTIONS(7400), 1,
      anon_sym_LBRACE,
    STATE(3880), 1,
      sym__constructor_name,
    STATE(4987), 1,
      sym_attribute,
    STATE(5608), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5561), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [205686] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1425), 1,
      anon_sym_LBRACK,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6697), 1,
      sym__uppercase_identifier,
    ACTIONS(7354), 1,
      anon_sym_LPAREN,
    ACTIONS(7356), 1,
      anon_sym_LBRACE,
    ACTIONS(7358), 1,
      anon_sym_begin,
    STATE(2228), 1,
      sym__constructor_name,
    STATE(4988), 1,
      sym_attribute,
    STATE(5271), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5312), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [205728] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7103), 1,
      anon_sym_COLON_GT,
    ACTIONS(7404), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4953), 1,
      sym__coerced,
    STATE(4989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7402), 10,
      anon_sym_EQ,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [205760] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2582), 1,
      anon_sym_LBRACK,
    ACTIONS(2592), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7348), 1,
      anon_sym_LBRACE,
    ACTIONS(7350), 1,
      anon_sym_begin,
    ACTIONS(7406), 1,
      anon_sym_LPAREN,
    ACTIONS(7408), 1,
      sym__uppercase_identifier,
    STATE(4990), 1,
      sym_attribute,
    STATE(5637), 1,
      sym__constructor_name,
    STATE(5660), 1,
      sym__delimited_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5636), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [205802] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4784), 1,
      anon_sym_in,
    ACTIONS(7187), 1,
      anon_sym_LPAREN,
    STATE(4991), 1,
      sym_attribute,
    STATE(5229), 1,
      sym__simple_module_expression,
    STATE(6095), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [205844] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2622), 1,
      anon_sym_LBRACK,
    ACTIONS(2628), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7350), 1,
      anon_sym_begin,
    ACTIONS(7400), 1,
      anon_sym_LBRACE,
    ACTIONS(7408), 1,
      sym__uppercase_identifier,
    ACTIONS(7410), 1,
      anon_sym_LPAREN,
    STATE(4992), 1,
      sym_attribute,
    STATE(5608), 1,
      sym__delimited_binding_pattern,
    STATE(5637), 1,
      sym__constructor_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5561), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [205886] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1425), 1,
      anon_sym_LBRACK,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7356), 1,
      anon_sym_LBRACE,
    ACTIONS(7358), 1,
      anon_sym_begin,
    ACTIONS(7412), 1,
      anon_sym_LPAREN,
    ACTIONS(7414), 1,
      sym__uppercase_identifier,
    STATE(2613), 1,
      sym__constructor_name,
    STATE(4993), 1,
      sym_attribute,
    STATE(5271), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5312), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [205928] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7296), 1,
      anon_sym_as,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4994), 1,
      sym_attribute,
    STATE(5039), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7416), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [205962] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1369), 1,
      anon_sym_LBRACK,
    ACTIONS(1383), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7358), 1,
      anon_sym_begin,
    ACTIONS(7388), 1,
      anon_sym_LPAREN,
    ACTIONS(7390), 1,
      anon_sym_LBRACE,
    ACTIONS(7418), 1,
      sym__uppercase_identifier,
    STATE(4995), 1,
      sym_attribute,
    STATE(5170), 1,
      sym__delimited_pattern,
    STATE(5181), 1,
      sym__constructor_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5133), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [206004] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4996), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4435), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [206030] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2121), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(2123), 11,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      sym_hash_operator,
  [206058] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7420), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(4998), 1,
      sym_attribute,
    STATE(5081), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7422), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [206092] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1369), 1,
      anon_sym_LBRACK,
    ACTIONS(1383), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6697), 1,
      sym__uppercase_identifier,
    ACTIONS(7358), 1,
      anon_sym_begin,
    ACTIONS(7388), 1,
      anon_sym_LPAREN,
    ACTIONS(7390), 1,
      anon_sym_LBRACE,
    STATE(2228), 1,
      sym__constructor_name,
    STATE(4999), 1,
      sym_attribute,
    STATE(5170), 1,
      sym__delimited_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5133), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [206134] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4776), 1,
      anon_sym_in,
    ACTIONS(7187), 1,
      anon_sym_LPAREN,
    STATE(5000), 1,
      sym_attribute,
    STATE(5229), 1,
      sym__simple_module_expression,
    STATE(6241), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [206176] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7426), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7424), 11,
      anon_sym_DQUOTE,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [206204] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5002), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4409), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [206230] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2343), 1,
      anon_sym_LBRACK,
    ACTIONS(2351), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7428), 1,
      anon_sym_LPAREN,
    ACTIONS(7430), 1,
      anon_sym_LBRACE,
    ACTIONS(7432), 1,
      anon_sym_begin,
    ACTIONS(7434), 1,
      sym__uppercase_identifier,
    STATE(2953), 1,
      sym__delimited_pattern,
    STATE(3000), 1,
      sym__constructor_name,
    STATE(5003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2928), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [206272] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7436), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5004), 1,
      sym_attribute,
    STATE(5042), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7438), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [206306] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3390), 1,
      anon_sym_COLON2,
    STATE(5005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3388), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [206334] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4762), 1,
      anon_sym_DASH_GT,
    ACTIONS(4764), 1,
      anon_sym_STAR,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5006), 1,
      sym_attribute,
    STATE(6545), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4758), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4760), 3,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
  [206378] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7440), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5007), 1,
      sym_attribute,
    STATE(5147), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7442), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [206412] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7446), 1,
      anon_sym_as,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5008), 1,
      sym_attribute,
    STATE(5141), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7444), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [206446] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3396), 1,
      anon_sym_COLON2,
    STATE(5009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3394), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [206474] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4446), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [206500] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3740), 1,
      anon_sym_PIPE,
    ACTIONS(7304), 1,
      anon_sym_LPAREN,
    ACTIONS(7310), 1,
      anon_sym_LBRACK,
    ACTIONS(7314), 1,
      anon_sym_LBRACE,
    ACTIONS(7316), 1,
      sym__uppercase_identifier,
    ACTIONS(7448), 1,
      anon_sym_DOT_DOT,
    STATE(5011), 1,
      sym_attribute,
    STATE(5629), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7312), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(5348), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [206544] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4874), 1,
      anon_sym_in,
    ACTIONS(7187), 1,
      anon_sym_LPAREN,
    STATE(5012), 1,
      sym_attribute,
    STATE(5229), 1,
      sym__simple_module_expression,
    STATE(6231), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [206586] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7187), 1,
      anon_sym_LPAREN,
    STATE(5013), 1,
      sym_attribute,
    STATE(5229), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4417), 4,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [206622] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1767), 1,
      anon_sym_LBRACK,
    ACTIONS(1783), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7450), 1,
      anon_sym_LPAREN,
    ACTIONS(7452), 1,
      anon_sym_LBRACE,
    ACTIONS(7454), 1,
      anon_sym_begin,
    ACTIONS(7456), 1,
      sym__uppercase_identifier,
    STATE(3443), 1,
      sym__constructor_name,
    STATE(3667), 1,
      sym__delimited_binding_pattern,
    STATE(5014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3680), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [206664] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5015), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4413), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [206690] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7458), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5016), 1,
      sym_attribute,
    STATE(5117), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7460), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [206724] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(7464), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5017), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7462), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [206754] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4421), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [206780] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7466), 1,
      anon_sym_EQ,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5019), 1,
      sym_attribute,
    STATE(5087), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7468), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [206814] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1425), 1,
      anon_sym_LBRACK,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7354), 1,
      anon_sym_LPAREN,
    ACTIONS(7356), 1,
      anon_sym_LBRACE,
    ACTIONS(7358), 1,
      anon_sym_begin,
    ACTIONS(7418), 1,
      sym__uppercase_identifier,
    STATE(5020), 1,
      sym_attribute,
    STATE(5181), 1,
      sym__constructor_name,
    STATE(5271), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5312), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [206856] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7414), 1,
      sym__uppercase_identifier,
    ACTIONS(7470), 1,
      anon_sym_LPAREN,
    ACTIONS(7472), 1,
      anon_sym_LBRACE,
    ACTIONS(7474), 1,
      anon_sym_begin,
    STATE(2613), 1,
      sym__constructor_name,
    STATE(2786), 1,
      sym__delimited_pattern,
    STATE(5021), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [206898] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1369), 1,
      anon_sym_LBRACK,
    ACTIONS(1383), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7358), 1,
      anon_sym_begin,
    ACTIONS(7390), 1,
      anon_sym_LBRACE,
    ACTIONS(7414), 1,
      sym__uppercase_identifier,
    ACTIONS(7476), 1,
      anon_sym_LPAREN,
    STATE(2613), 1,
      sym__constructor_name,
    STATE(5022), 1,
      sym_attribute,
    STATE(5170), 1,
      sym__delimited_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5133), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [206940] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3846), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    ACTIONS(7302), 1,
      anon_sym_LBRACE,
    ACTIONS(7478), 1,
      anon_sym_DOT_DOT,
    STATE(4646), 1,
      sym_constructor_declaration,
    STATE(5023), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4434), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [206984] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(7187), 1,
      anon_sym_LPAREN,
    STATE(5024), 1,
      sym_attribute,
    STATE(5229), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4395), 4,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    STATE(5250), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [207020] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2439), 1,
      anon_sym_COLON2,
    ACTIONS(7480), 1,
      anon_sym_DOT_DOT,
    STATE(5025), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2433), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [207050] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6947), 1,
      anon_sym_DASH_GT,
    STATE(5026), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4435), 12,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_AMP,
  [207078] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1835), 1,
      anon_sym_LBRACK,
    ACTIONS(1849), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(7472), 1,
      anon_sym_LBRACE,
    ACTIONS(7474), 1,
      anon_sym_begin,
    ACTIONS(7482), 1,
      anon_sym_LPAREN,
    ACTIONS(7484), 1,
      sym__uppercase_identifier,
    STATE(2786), 1,
      sym__delimited_pattern,
    STATE(3182), 1,
      sym__constructor_name,
    STATE(5027), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2852), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [207120] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3774), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    ACTIONS(7302), 1,
      anon_sym_LBRACE,
    ACTIONS(7486), 1,
      anon_sym_DOT_DOT,
    STATE(4197), 1,
      sym_constructor_declaration,
    STATE(5028), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4091), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [207164] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4762), 1,
      anon_sym_DASH_GT,
    ACTIONS(4764), 1,
      anon_sym_STAR,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5029), 1,
      sym_attribute,
    STATE(6545), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4792), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4794), 3,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
  [207208] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4417), 4,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [207244] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3774), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    ACTIONS(7302), 1,
      anon_sym_LBRACE,
    ACTIONS(7488), 1,
      anon_sym_DOT_DOT,
    STATE(4197), 1,
      sym_constructor_declaration,
    STATE(5031), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4046), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [207288] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5032), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7492), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7490), 11,
      anon_sym_DQUOTE,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [207316] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7496), 1,
      anon_sym_COLON2,
    STATE(5033), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7494), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [207343] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5034), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5060), 3,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
    ACTIONS(5058), 4,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT,
  [207380] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4034), 1,
      anon_sym_COLON2,
    STATE(5035), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4036), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [207407] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7500), 1,
      anon_sym_COLON2,
    STATE(5036), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7498), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [207434] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7122), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [207465] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7502), 1,
      anon_sym_RPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5038), 1,
      sym_attribute,
    STATE(8223), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [207504] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7506), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [207535] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5040), 1,
      sym_attribute,
    STATE(5078), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7508), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [207566] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7510), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5041), 1,
      sym_attribute,
    STATE(7321), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [207605] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5042), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7512), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [207636] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7516), 1,
      anon_sym_COLON2,
    STATE(5043), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7514), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [207663] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5044), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7518), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [207694] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5045), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7520), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [207725] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5046), 1,
      sym_attribute,
    STATE(5082), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7522), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [207756] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5047), 1,
      sym_attribute,
    STATE(5083), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7524), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [207787] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4060), 1,
      anon_sym_COLON2,
    STATE(5048), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4062), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [207814] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5049), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7526), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [207845] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5050), 1,
      sym_attribute,
    STATE(5084), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7528), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [207876] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7530), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [207907] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5052), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7532), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [207938] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5053), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7534), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [207969] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5054), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2774), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208000] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4084), 1,
      anon_sym_COLON2,
    STATE(5055), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4086), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [208027] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5056), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2746), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208058] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5057), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7536), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208089] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5058), 1,
      sym_attribute,
    STATE(5094), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7538), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208120] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5059), 1,
      sym_attribute,
    STATE(5095), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7540), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208151] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7542), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5060), 1,
      sym_attribute,
    STATE(8078), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [208190] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5061), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7544), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208221] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5062), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7546), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208252] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7548), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5063), 1,
      sym_attribute,
    STATE(8303), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [208291] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7552), 1,
      anon_sym_COLON2,
    STATE(5064), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7550), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [208318] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7554), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5065), 1,
      sym_attribute,
    STATE(8049), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [208357] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7556), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5066), 1,
      sym_attribute,
    STATE(8140), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [208396] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7560), 1,
      anon_sym_COLON2,
    STATE(5067), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7558), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [208423] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7564), 1,
      anon_sym_COLON2,
    STATE(5068), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7562), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [208450] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7566), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5069), 1,
      sym_attribute,
    STATE(7890), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [208489] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7564), 1,
      anon_sym_COLON2,
    STATE(5070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7562), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [208516] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3482), 1,
      anon_sym_COLON2,
    STATE(5071), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3480), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [208543] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7568), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5072), 1,
      sym_attribute,
    STATE(8326), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [208582] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7572), 1,
      anon_sym_COLON2,
    STATE(5073), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7570), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [208609] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7576), 1,
      anon_sym_COLON2,
    STATE(5074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7574), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [208636] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7580), 1,
      anon_sym_COLON2,
    STATE(5075), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7578), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [208663] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7582), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5076), 1,
      sym_attribute,
    STATE(7979), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [208702] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7584), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5077), 1,
      sym_attribute,
    STATE(8148), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [208741] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5078), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7586), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208772] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5079), 1,
      sym_attribute,
    STATE(5106), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7588), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208803] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5080), 1,
      sym_attribute,
    STATE(5107), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7590), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208834] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5081), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7592), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208865] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7594), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208896] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5083), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7596), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208927] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5084), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7598), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208958] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5085), 1,
      sym_attribute,
    STATE(5174), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7600), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [208989] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5086), 1,
      sym_attribute,
    STATE(5110), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7602), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209020] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5087), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7604), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209051] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5088), 1,
      sym_attribute,
    STATE(5113), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7606), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209082] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7608), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5089), 1,
      sym_attribute,
    STATE(7417), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [209121] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5090), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7610), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209152] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7614), 1,
      anon_sym_COLON2,
    STATE(5091), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7612), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [209179] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7616), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5092), 1,
      sym_attribute,
    STATE(7455), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [209218] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5093), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7618), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209249] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5094), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7620), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209280] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5095), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7622), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209311] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5096), 1,
      sym_attribute,
    STATE(5118), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7624), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209342] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7626), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5097), 1,
      sym_attribute,
    STATE(7293), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [209381] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7630), 1,
      anon_sym_COLON2,
    STATE(5098), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7628), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [209408] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7634), 1,
      anon_sym_COLON2,
    STATE(5099), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7632), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [209435] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7638), 1,
      anon_sym_COLON2,
    STATE(5100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7636), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [209462] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7640), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5101), 1,
      sym_attribute,
    STATE(7759), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [209501] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7642), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5102), 1,
      sym_attribute,
    STATE(7797), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [209540] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7644), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5103), 1,
      sym_attribute,
    STATE(7943), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [209579] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3598), 1,
      anon_sym_COLON2,
    STATE(5104), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3596), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [209606] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7646), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5105), 1,
      sym_attribute,
    STATE(7972), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [209645] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5106), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7648), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209676] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5107), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7650), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209707] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5108), 1,
      sym_attribute,
    STATE(5124), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7652), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209738] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7654), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5109), 1,
      sym_attribute,
    STATE(7732), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [209777] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5110), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7656), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209808] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5111), 1,
      sym_attribute,
    STATE(5125), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7658), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209839] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7660), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5112), 1,
      sym_attribute,
    STATE(8175), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [209878] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5113), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7662), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209909] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5114), 1,
      sym_attribute,
    STATE(5126), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7664), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209940] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5115), 1,
      sym_attribute,
    STATE(5127), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7666), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [209971] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7668), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5116), 1,
      sym_attribute,
    STATE(8244), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [210010] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5117), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7670), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210041] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5118), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7672), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210072] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3430), 1,
      anon_sym_COLON2,
    STATE(5119), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3428), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [210099] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3434), 1,
      anon_sym_COLON2,
    STATE(5120), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3432), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [210126] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7674), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5121), 1,
      sym_attribute,
    STATE(7344), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [210165] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7676), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5122), 1,
      sym_attribute,
    STATE(7652), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [210204] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3520), 1,
      anon_sym_COLON2,
    STATE(5123), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3518), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [210231] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5124), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7678), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210262] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5125), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7680), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210293] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5126), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7682), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210324] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5127), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7684), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210355] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5128), 1,
      sym_attribute,
    STATE(5129), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7686), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210386] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5129), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7688), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210417] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7692), 1,
      anon_sym_LPAREN,
    ACTIONS(7694), 1,
      anon_sym_COLON2,
    STATE(5130), 1,
      sym_attribute,
    STATE(5137), 1,
      aux_sym_module_binding_repeat1,
    STATE(5405), 1,
      sym_module_parameter,
    STATE(5406), 1,
      sym__module_typed,
    STATE(5797), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4935), 2,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(7690), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [210460] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7696), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5131), 1,
      sym_attribute,
    STATE(7873), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [210499] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2439), 1,
      anon_sym_COLON2,
    STATE(5132), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2433), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [210526] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4042), 1,
      anon_sym_COLON2,
    STATE(5133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4044), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [210553] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(5134), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 11,
      anon_sym_and,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
  [210580] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(5135), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 11,
      anon_sym_and,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
  [210607] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7698), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5136), 1,
      sym_attribute,
    STATE(8255), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [210646] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7692), 1,
      anon_sym_LPAREN,
    ACTIONS(7694), 1,
      anon_sym_COLON2,
    STATE(5137), 1,
      sym_attribute,
    STATE(5307), 1,
      aux_sym_module_binding_repeat1,
    STATE(5405), 1,
      sym_module_parameter,
    STATE(5498), 1,
      sym__module_typed,
    STATE(5923), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4900), 2,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(7700), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [210689] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5138), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7101), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210720] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5045), 1,
      aux_sym_expression_item_repeat1,
    STATE(5139), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7702), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210751] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7704), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210782] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5141), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7706), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210813] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4009), 1,
      anon_sym_COLON2,
    STATE(5142), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4011), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [210840] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4013), 1,
      anon_sym_COLON2,
    STATE(5143), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4015), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [210867] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5049), 1,
      aux_sym_expression_item_repeat1,
    STATE(5144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7708), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210898] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5145), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2780), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210929] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5057), 1,
      aux_sym_expression_item_repeat1,
    STATE(5146), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7710), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210960] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7712), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [210991] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5148), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7714), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211022] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7716), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5149), 1,
      sym_attribute,
    STATE(7224), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [211061] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7718), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5150), 1,
      sym_attribute,
    STATE(7262), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [211100] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7722), 1,
      anon_sym_COLON2,
    STATE(5151), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7720), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [211127] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7726), 1,
      anon_sym_COLON2,
    STATE(5152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7724), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [211154] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5153), 1,
      sym_attribute,
    STATE(5156), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6365), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211185] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7728), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5154), 1,
      sym_attribute,
    STATE(7451), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [211224] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3532), 1,
      anon_sym_COLON2,
    STATE(5155), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3530), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [211251] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5156), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6332), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211282] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5157), 1,
      sym_attribute,
    STATE(5160), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6332), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211313] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5158), 1,
      sym_attribute,
    STATE(5162), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6347), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211344] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7304), 1,
      anon_sym_LPAREN,
    ACTIONS(7310), 1,
      anon_sym_LBRACK,
    ACTIONS(7316), 1,
      sym__uppercase_identifier,
    STATE(5159), 1,
      sym_attribute,
    STATE(5746), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7312), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    ACTIONS(4941), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [211381] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6202), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211412] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5161), 1,
      sym_attribute,
    STATE(5164), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6206), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211443] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5162), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6206), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211474] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4764), 1,
      anon_sym_STAR,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5163), 1,
      sym_attribute,
    STATE(5523), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4762), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4955), 3,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
  [211515] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5164), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6308), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211546] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5062), 1,
      aux_sym_expression_item_repeat1,
    STATE(5165), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7730), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211577] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4038), 1,
      anon_sym_COLON2,
    STATE(5166), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4040), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [211604] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4076), 1,
      anon_sym_COLON2,
    STATE(5167), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4078), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [211631] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3967), 1,
      anon_sym_COLON2,
    STATE(5168), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3969), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [211658] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3971), 1,
      anon_sym_COLON2,
    STATE(5169), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3973), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [211685] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3975), 1,
      anon_sym_COLON2,
    STATE(5170), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3977), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [211712] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1391), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1393), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4399), 1,
      anon_sym_LPAREN,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(7732), 1,
      anon_sym_RPAREN,
    STATE(3699), 1,
      sym__simple_module_expression,
    STATE(5171), 1,
      sym_attribute,
    STATE(7498), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3675), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [211751] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3993), 1,
      anon_sym_COLON2,
    STATE(5172), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3995), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [211778] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4001), 1,
      anon_sym_COLON2,
    STATE(5173), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4003), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [211805] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5174), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7734), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211836] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5175), 1,
      sym_attribute,
    STATE(5189), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7736), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211866] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5176), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7738), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211896] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5177), 1,
      sym_attribute,
    STATE(5191), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7740), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211926] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5178), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7742), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211956] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5179), 1,
      sym_attribute,
    STATE(5192), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7744), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [211986] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5180), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2127), 11,
      anon_sym_and,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
  [212010] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5181), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2181), 11,
      anon_sym_and,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
  [212034] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    ACTIONS(7748), 1,
      anon_sym_as,
    STATE(5182), 1,
      sym_attribute,
    STATE(5227), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5100), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [212066] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3468), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    ACTIONS(7750), 1,
      anon_sym_EQ,
    ACTIONS(7752), 1,
      anon_sym_COLON_EQ,
    STATE(5183), 1,
      sym_attribute,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5383), 1,
      aux_sym_type_binding_repeat1,
    STATE(5897), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5072), 2,
      anon_sym_and,
      anon_sym_in,
  [212108] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3740), 1,
      anon_sym_PIPE,
    ACTIONS(7304), 1,
      anon_sym_LPAREN,
    ACTIONS(7310), 1,
      anon_sym_LBRACK,
    ACTIONS(7316), 1,
      sym__uppercase_identifier,
    ACTIONS(7754), 1,
      anon_sym_private,
    STATE(5184), 1,
      sym_attribute,
    STATE(5629), 1,
      sym_constructor_declaration,
    STATE(5903), 1,
      sym_variant_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7312), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [212148] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3774), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    ACTIONS(7756), 1,
      anon_sym_private,
    STATE(4197), 1,
      sym_constructor_declaration,
    STATE(4250), 1,
      sym_variant_declaration,
    STATE(5185), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [212188] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    ACTIONS(7748), 1,
      anon_sym_as,
    STATE(5186), 1,
      sym_attribute,
    STATE(5212), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5017), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [212220] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    ACTIONS(7748), 1,
      anon_sym_as,
    STATE(5187), 1,
      sym_attribute,
    STATE(5213), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5041), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [212252] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5176), 1,
      aux_sym_expression_item_repeat1,
    STATE(5188), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7758), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [212282] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5189), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7760), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [212312] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5178), 1,
      aux_sym_expression_item_repeat1,
    STATE(5190), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7762), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [212342] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5191), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7764), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [212372] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2355), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3941), 1,
      aux_sym_expression_item_repeat1,
    STATE(4160), 1,
      sym_item_attribute,
    STATE(5192), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7766), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [212402] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2157), 1,
      anon_sym_POUND,
    STATE(5193), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2159), 10,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      sym_hash_operator,
  [212428] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3740), 1,
      anon_sym_PIPE,
    ACTIONS(7304), 1,
      anon_sym_LPAREN,
    ACTIONS(7310), 1,
      anon_sym_LBRACK,
    ACTIONS(7316), 1,
      sym__uppercase_identifier,
    ACTIONS(7768), 1,
      anon_sym_private,
    STATE(5194), 1,
      sym_attribute,
    STATE(5629), 1,
      sym_constructor_declaration,
    STATE(5974), 1,
      sym_variant_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7312), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [212468] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3846), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    ACTIONS(7770), 1,
      anon_sym_private,
    STATE(4219), 1,
      sym_variant_declaration,
    STATE(4646), 1,
      sym_constructor_declaration,
    STATE(5195), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [212508] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5196), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2177), 11,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [212532] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7774), 1,
      anon_sym_COLON,
    ACTIONS(7776), 1,
      anon_sym_COLON2,
    STATE(5197), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7772), 9,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [212560] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3468), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    ACTIONS(7778), 1,
      anon_sym_EQ,
    ACTIONS(7780), 1,
      anon_sym_COLON_EQ,
    STATE(5198), 1,
      sym_attribute,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5364), 1,
      aux_sym_type_binding_repeat1,
    STATE(6009), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5025), 2,
      anon_sym_and,
      anon_sym_in,
  [212602] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3846), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    ACTIONS(7782), 1,
      anon_sym_private,
    STATE(4250), 1,
      sym_variant_declaration,
    STATE(4646), 1,
      sym_constructor_declaration,
    STATE(5199), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [212642] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5200), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2115), 11,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [212666] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7784), 1,
      anon_sym_DOT,
    STATE(5201), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4702), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [212692] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3774), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    ACTIONS(7786), 1,
      anon_sym_private,
    STATE(4197), 1,
      sym_constructor_declaration,
    STATE(4219), 1,
      sym_variant_declaration,
    STATE(5202), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [212732] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5203), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4882), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [212755] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(5204), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1891), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(1893), 8,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_COLON_GT,
      sym_hash_operator,
  [212780] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7790), 1,
      anon_sym_COLON2,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    STATE(5205), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7788), 8,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_when,
  [212807] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7794), 1,
      anon_sym_EQ,
    ACTIONS(7796), 1,
      anon_sym_RPAREN,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7802), 1,
      anon_sym_COLON2,
    ACTIONS(7804), 1,
      anon_sym_as,
    STATE(5206), 1,
      sym_attribute,
    STATE(7642), 1,
      sym__typed,
    STATE(7811), 1,
      sym__polymorphic_typed,
    STATE(7867), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212848] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5207), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2167), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [212871] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5208), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2185), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [212894] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5209), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2189), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [212917] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(4476), 1,
      sym_constructor_declaration,
    STATE(5210), 1,
      sym_attribute,
    STATE(5365), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [212954] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3846), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    STATE(4350), 1,
      sym_variant_declaration,
    STATE(4646), 1,
      sym_constructor_declaration,
    STATE(5211), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [212991] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5212), 1,
      sym_attribute,
    STATE(5222), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5133), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [213020] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5213), 1,
      sym_attribute,
    STATE(5222), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5137), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [213049] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2193), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [213072] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5215), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4792), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4794), 3,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
  [213107] = 13,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7802), 1,
      anon_sym_COLON2,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(7806), 1,
      anon_sym_EQ,
    ACTIONS(7808), 1,
      anon_sym_RPAREN,
    STATE(5216), 1,
      sym_attribute,
    STATE(7328), 1,
      sym__typed,
    STATE(7548), 1,
      sym__maybe_polymorphic_typed,
    STATE(7551), 1,
      sym__polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213148] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7564), 1,
      anon_sym_COLON2,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    STATE(5217), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7562), 7,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_when,
  [213177] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4526), 1,
      sym__lowercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5218), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5164), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5166), 3,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
  [213212] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5219), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4812), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [213235] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5220), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4971), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(2113), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [213260] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    ACTIONS(7748), 1,
      anon_sym_as,
    ACTIONS(7810), 1,
      anon_sym_EQ,
    STATE(5221), 1,
      sym_attribute,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5392), 1,
      aux_sym_type_binding_repeat1,
    STATE(5970), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5156), 2,
      anon_sym_and,
      anon_sym_in,
  [213299] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7812), 1,
      anon_sym_constraint,
    STATE(5363), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5222), 2,
      sym_attribute,
      aux_sym_type_binding_repeat1,
    ACTIONS(5112), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [213326] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(7817), 1,
      anon_sym_LPAREN,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    STATE(3638), 1,
      sym__delimited_type,
    STATE(5223), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [213363] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4788), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [213386] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3740), 1,
      anon_sym_PIPE,
    ACTIONS(7304), 1,
      anon_sym_LPAREN,
    ACTIONS(7310), 1,
      anon_sym_LBRACK,
    ACTIONS(7316), 1,
      sym__uppercase_identifier,
    STATE(5225), 1,
      sym_attribute,
    STATE(5629), 1,
      sym_constructor_declaration,
    STATE(5975), 1,
      sym_variant_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7312), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [213423] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7572), 1,
      anon_sym_COLON2,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    STATE(5226), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7570), 6,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_when,
  [213454] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5222), 1,
      aux_sym_type_binding_repeat1,
    STATE(5227), 1,
      sym_attribute,
    STATE(5363), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5152), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [213483] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(7823), 1,
      anon_sym_COLON2,
    STATE(5228), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7821), 5,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_when,
  [213516] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4796), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [213539] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(4476), 1,
      sym_constructor_declaration,
    STATE(5230), 1,
      sym_attribute,
    STATE(5341), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4788), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [213576] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1645), 1,
      anon_sym_COLON2,
    ACTIONS(7480), 1,
      anon_sym_DOT_DOT,
    STATE(5231), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1641), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [213603] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4866), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [213626] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4888), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [213649] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4892), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [213672] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(7304), 1,
      anon_sym_LPAREN,
    ACTIONS(7310), 1,
      anon_sym_LBRACK,
    ACTIONS(7316), 1,
      sym__uppercase_identifier,
    STATE(5235), 1,
      sym_attribute,
    STATE(5347), 1,
      sym__attribute,
    STATE(6078), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7312), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [213709] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3782), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3784), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(4208), 1,
      anon_sym_LBRACK,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(7825), 1,
      sym__lowercase_identifier,
    ACTIONS(7827), 1,
      anon_sym_LPAREN,
    STATE(3593), 1,
      sym__delimited_type,
    STATE(5236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3546), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [213746] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7564), 1,
      anon_sym_COLON2,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    STATE(5237), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7562), 8,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [213773] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3774), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    STATE(4197), 1,
      sym_constructor_declaration,
    STATE(4259), 1,
      sym_variant_declaration,
    STATE(5238), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [213810] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4800), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [213833] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3846), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    STATE(4259), 1,
      sym_variant_declaration,
    STATE(4646), 1,
      sym_constructor_declaration,
    STATE(5240), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [213870] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(5241), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1887), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(1889), 8,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_COLON_GT,
      sym_hash_operator,
  [213895] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(7817), 1,
      anon_sym_LPAREN,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(7829), 1,
      sym__lowercase_identifier,
    STATE(3638), 1,
      sym__delimited_type,
    STATE(5242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [213932] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3852), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(4226), 1,
      anon_sym_LBRACK,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(7831), 1,
      sym__lowercase_identifier,
    ACTIONS(7833), 1,
      anon_sym_LPAREN,
    STATE(3904), 1,
      sym__delimited_type,
    STATE(5243), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3868), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [213969] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    ACTIONS(7748), 1,
      anon_sym_as,
    ACTIONS(7835), 1,
      anon_sym_EQ,
    STATE(5244), 1,
      sym_attribute,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5375), 1,
      aux_sym_type_binding_repeat1,
    STATE(5998), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5119), 2,
      anon_sym_and,
      anon_sym_in,
  [214008] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3740), 1,
      anon_sym_PIPE,
    ACTIONS(7304), 1,
      anon_sym_LPAREN,
    ACTIONS(7310), 1,
      anon_sym_LBRACK,
    ACTIONS(7316), 1,
      sym__uppercase_identifier,
    STATE(5245), 1,
      sym_attribute,
    STATE(5629), 1,
      sym_constructor_declaration,
    STATE(5800), 1,
      sym_variant_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7312), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [214045] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4870), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [214068] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4878), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [214091] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(7839), 1,
      anon_sym_COLON2,
    STATE(5248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7837), 5,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_when,
  [214124] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5249), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4702), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [214147] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5250), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4780), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [214170] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2800), 1,
      anon_sym_LBRACK,
    ACTIONS(2814), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2816), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(7817), 1,
      anon_sym_LPAREN,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(7841), 1,
      sym__lowercase_identifier,
    STATE(3638), 1,
      sym__delimited_type,
    STATE(5251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3644), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [214207] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3774), 1,
      anon_sym_PIPE,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    STATE(4197), 1,
      sym_constructor_declaration,
    STATE(4350), 1,
      sym_variant_declaration,
    STATE(5252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [214244] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7843), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [214266] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7845), 1,
      sym__lowercase_identifier,
    ACTIONS(7847), 1,
      anon_sym_LPAREN,
    ACTIONS(7849), 1,
      anon_sym_LBRACK,
    ACTIONS(7851), 1,
      anon_sym_LBRACE,
    ACTIONS(7853), 1,
      sym_indexing_operator,
    STATE(2574), 1,
      sym_field_path,
    STATE(5254), 1,
      sym_attribute,
    STATE(6663), 1,
      sym_indexing_operator_path,
    STATE(7226), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214304] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5218), 1,
      anon_sym_in,
    ACTIONS(7290), 1,
      anon_sym_LBRACE,
    STATE(5255), 1,
      sym_attribute,
    STATE(5412), 1,
      aux_sym_external_repeat1,
    STATE(6096), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6187), 2,
      sym_string,
      sym_quoted_string,
  [214340] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5256), 1,
      sym_attribute,
    STATE(7272), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [214370] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7748), 1,
      anon_sym_as,
    STATE(5257), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5045), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [214394] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7861), 1,
      anon_sym_LPAREN,
    ACTIONS(7863), 1,
      anon_sym_DOT,
    STATE(5258), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5104), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [214420] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7853), 1,
      sym_indexing_operator,
    ACTIONS(7865), 1,
      sym__lowercase_identifier,
    ACTIONS(7867), 1,
      anon_sym_LPAREN,
    ACTIONS(7869), 1,
      anon_sym_LBRACK,
    ACTIONS(7871), 1,
      anon_sym_LBRACE,
    STATE(1258), 1,
      sym_field_path,
    STATE(5259), 1,
      sym_attribute,
    STATE(6669), 1,
      sym_indexing_operator_path,
    STATE(7297), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214458] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7853), 1,
      sym_indexing_operator,
    ACTIONS(7873), 1,
      sym__lowercase_identifier,
    ACTIONS(7875), 1,
      anon_sym_LPAREN,
    ACTIONS(7877), 1,
      anon_sym_LBRACK,
    ACTIONS(7879), 1,
      anon_sym_LBRACE,
    STATE(2317), 1,
      sym_field_path,
    STATE(5260), 1,
      sym_attribute,
    STATE(6657), 1,
      sym_indexing_operator_path,
    STATE(7354), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214496] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3506), 1,
      anon_sym_COLON2,
    STATE(5261), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3504), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [214520] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5262), 1,
      sym_attribute,
    STATE(7955), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [214550] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3566), 1,
      anon_sym_COLON2,
    STATE(5263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3564), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [214574] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3586), 1,
      anon_sym_COLON2,
    STATE(5264), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3584), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [214598] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5023), 1,
      anon_sym_as,
    ACTIONS(7881), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(5265), 1,
      sym_attribute,
    STATE(5353), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5100), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [214628] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3590), 1,
      anon_sym_COLON2,
    STATE(5266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3588), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [214652] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3602), 1,
      anon_sym_COLON2,
    STATE(5267), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3600), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [214676] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5268), 1,
      sym_attribute,
    STATE(7887), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [214706] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5269), 1,
      sym_attribute,
    STATE(7665), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [214736] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5023), 1,
      anon_sym_as,
    ACTIONS(7881), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(5270), 1,
      sym_attribute,
    STATE(5354), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5017), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [214766] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3594), 1,
      anon_sym_COLON2,
    STATE(5271), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3592), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [214790] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3610), 1,
      anon_sym_COLON2,
    STATE(5272), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3608), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [214814] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5273), 1,
      sym_attribute,
    STATE(7467), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [214844] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5023), 1,
      anon_sym_as,
    ACTIONS(7881), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(5274), 1,
      sym_attribute,
    STATE(5355), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5041), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [214874] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    ACTIONS(7748), 1,
      anon_sym_as,
    STATE(5275), 1,
      sym_attribute,
    STATE(5351), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5826), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5266), 2,
      anon_sym_and,
      anon_sym_in,
  [214910] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3438), 1,
      anon_sym_COLON2,
    STATE(5276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3436), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [214934] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5277), 1,
      sym_attribute,
    STATE(7693), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [214964] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3452), 1,
      anon_sym_COLON2,
    STATE(5278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3450), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [214988] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3460), 1,
      anon_sym_COLON2,
    STATE(5279), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3458), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [215012] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7853), 1,
      sym_indexing_operator,
    ACTIONS(7883), 1,
      sym__lowercase_identifier,
    ACTIONS(7885), 1,
      anon_sym_LPAREN,
    ACTIONS(7887), 1,
      anon_sym_LBRACK,
    ACTIONS(7889), 1,
      anon_sym_LBRACE,
    STATE(2990), 1,
      sym_field_path,
    STATE(5280), 1,
      sym_attribute,
    STATE(6706), 1,
      sym_indexing_operator_path,
    STATE(7640), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215050] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5281), 1,
      sym_attribute,
    STATE(7326), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [215080] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3464), 1,
      anon_sym_COLON2,
    STATE(5282), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3462), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [215104] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3524), 1,
      anon_sym_COLON2,
    STATE(5283), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3522), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [215128] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5284), 1,
      sym_attribute,
    STATE(7631), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [215158] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3528), 1,
      anon_sym_COLON2,
    STATE(5285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3526), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [215182] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5286), 1,
      sym_attribute,
    STATE(7903), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [215212] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3546), 1,
      anon_sym_COLON2,
    STATE(5287), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3544), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [215236] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3582), 1,
      anon_sym_COLON2,
    STATE(5288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3580), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [215260] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5289), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4983), 9,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [215282] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5290), 1,
      sym_attribute,
    STATE(8083), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [215312] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3502), 1,
      anon_sym_COLON2,
    STATE(5291), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3500), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [215336] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3622), 1,
      anon_sym_COLON2,
    STATE(5292), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3620), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [215360] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7853), 1,
      sym_indexing_operator,
    ACTIONS(7891), 1,
      sym__lowercase_identifier,
    ACTIONS(7893), 1,
      anon_sym_LPAREN,
    ACTIONS(7895), 1,
      anon_sym_LBRACK,
    ACTIONS(7897), 1,
      anon_sym_LBRACE,
    STATE(1032), 1,
      sym_field_path,
    STATE(5293), 1,
      sym_attribute,
    STATE(6686), 1,
      sym_indexing_operator_path,
    STATE(7507), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215398] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3626), 1,
      anon_sym_COLON2,
    STATE(5294), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3624), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [215422] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5295), 1,
      sym_attribute,
    STATE(7193), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [215452] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7853), 1,
      sym_indexing_operator,
    ACTIONS(7899), 1,
      sym__lowercase_identifier,
    ACTIONS(7901), 1,
      anon_sym_LPAREN,
    ACTIONS(7903), 1,
      anon_sym_LBRACK,
    ACTIONS(7905), 1,
      anon_sym_LBRACE,
    STATE(1135), 1,
      sym_field_path,
    STATE(5296), 1,
      sym_attribute,
    STATE(6697), 1,
      sym_indexing_operator_path,
    STATE(7578), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215490] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1887), 1,
      anon_sym_COLON2,
    ACTIONS(2157), 1,
      anon_sym_POUND,
    STATE(5297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2159), 3,
      anon_sym_DOT,
      anon_sym_with,
      sym_hash_operator,
    ACTIONS(1889), 4,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
  [215518] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5298), 1,
      sym_attribute,
    STATE(7568), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [215548] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5299), 1,
      sym_attribute,
    STATE(8152), 1,
      sym_escape_sequence,
    STATE(8330), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [215578] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5300), 1,
      sym_attribute,
    STATE(7437), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [215608] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5301), 1,
      sym_attribute,
    STATE(7714), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [215638] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5302), 1,
      sym_attribute,
    STATE(7933), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [215668] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5303), 1,
      sym_attribute,
    STATE(8138), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [215698] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5304), 1,
      sym_attribute,
    STATE(7659), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [215728] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4971), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(5062), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [215752] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7911), 1,
      sym__right_quoted_string_delimiter,
    STATE(5306), 1,
      sym_attribute,
    STATE(5340), 1,
      aux_sym_quoted_string_content_repeat1,
    STATE(7333), 1,
      sym_quoted_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7909), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7907), 4,
      aux_sym_string_content_token1,
      aux_sym_quoted_string_content_token1,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [215782] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5147), 1,
      anon_sym_COLON2,
    ACTIONS(7913), 1,
      anon_sym_LPAREN,
    STATE(5405), 1,
      sym_module_parameter,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5307), 2,
      sym_attribute,
      aux_sym_module_binding_repeat1,
    ACTIONS(5145), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [215810] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1891), 1,
      anon_sym_COLON2,
    ACTIONS(2157), 1,
      anon_sym_POUND,
    STATE(5308), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2159), 3,
      anon_sym_DOT,
      anon_sym_with,
      sym_hash_operator,
    ACTIONS(1893), 4,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
  [215838] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7853), 1,
      sym_indexing_operator,
    ACTIONS(7916), 1,
      sym__lowercase_identifier,
    ACTIONS(7918), 1,
      anon_sym_LPAREN,
    ACTIONS(7920), 1,
      anon_sym_LBRACK,
    ACTIONS(7922), 1,
      anon_sym_LBRACE,
    STATE(2281), 1,
      sym_field_path,
    STATE(5309), 1,
      sym_attribute,
    STATE(6674), 1,
      sym_indexing_operator_path,
    STATE(7395), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215876] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1645), 1,
      anon_sym_COLON2,
    STATE(5310), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1641), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [215900] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3444), 1,
      anon_sym_COLON2,
    STATE(5311), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3442), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [215924] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3614), 1,
      anon_sym_COLON2,
    STATE(5312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3612), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [215948] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7853), 1,
      sym_indexing_operator,
    ACTIONS(7924), 1,
      sym__lowercase_identifier,
    ACTIONS(7926), 1,
      anon_sym_LPAREN,
    ACTIONS(7928), 1,
      anon_sym_LBRACK,
    ACTIONS(7930), 1,
      anon_sym_LBRACE,
    STATE(1946), 1,
      sym_field_path,
    STATE(5313), 1,
      sym_attribute,
    STATE(6683), 1,
      sym_indexing_operator_path,
    STATE(7471), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215986] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4983), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4979), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [216010] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3550), 1,
      anon_sym_COLON2,
    STATE(5315), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3548), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [216034] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5316), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4971), 9,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [216056] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7853), 1,
      sym_indexing_operator,
    ACTIONS(7932), 1,
      sym__lowercase_identifier,
    ACTIONS(7934), 1,
      anon_sym_LPAREN,
    ACTIONS(7936), 1,
      anon_sym_LBRACK,
    ACTIONS(7938), 1,
      anon_sym_LBRACE,
    STATE(1354), 1,
      sym_field_path,
    STATE(5317), 1,
      sym_attribute,
    STATE(6481), 1,
      sym_indexing_operator_path,
    STATE(8332), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216094] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3562), 1,
      anon_sym_COLON2,
    STATE(5318), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3560), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [216118] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3562), 1,
      anon_sym_COLON2,
    STATE(5319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3560), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [216142] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3606), 1,
      anon_sym_COLON2,
    STATE(5320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3604), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [216166] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3618), 1,
      anon_sym_COLON2,
    STATE(5321), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3616), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [216190] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5210), 1,
      anon_sym_in,
    ACTIONS(7290), 1,
      anon_sym_LBRACE,
    STATE(5322), 1,
      sym_attribute,
    STATE(5412), 1,
      aux_sym_external_repeat1,
    STATE(6243), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6187), 2,
      sym_string,
      sym_quoted_string,
  [216226] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    ACTIONS(7748), 1,
      anon_sym_as,
    STATE(5323), 1,
      sym_attribute,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5368), 1,
      aux_sym_type_binding_repeat1,
    STATE(5855), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5258), 2,
      anon_sym_and,
      anon_sym_in,
  [216262] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3570), 1,
      anon_sym_COLON2,
    STATE(5324), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3568), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [216286] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3574), 1,
      anon_sym_COLON2,
    STATE(5325), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3572), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [216310] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5082), 9,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [216332] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3456), 1,
      anon_sym_COLON2,
    STATE(5327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3454), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [216356] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5328), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7940), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [216378] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3448), 1,
      anon_sym_COLON2,
    STATE(5329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3446), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [216402] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7853), 1,
      sym_indexing_operator,
    ACTIONS(7942), 1,
      sym__lowercase_identifier,
    ACTIONS(7944), 1,
      anon_sym_LPAREN,
    ACTIONS(7946), 1,
      anon_sym_LBRACK,
    ACTIONS(7948), 1,
      anon_sym_LBRACE,
    STATE(3344), 1,
      sym_field_path,
    STATE(5330), 1,
      sym_attribute,
    STATE(6693), 1,
      sym_indexing_operator_path,
    STATE(7543), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216440] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4507), 9,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
  [216462] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7853), 1,
      sym_indexing_operator,
    ACTIONS(7875), 1,
      anon_sym_LPAREN,
    ACTIONS(7877), 1,
      anon_sym_LBRACK,
    ACTIONS(7879), 1,
      anon_sym_LBRACE,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(2317), 1,
      sym_field_path,
    STATE(5332), 1,
      sym_attribute,
    STATE(6657), 1,
      sym_indexing_operator_path,
    STATE(7885), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216500] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7853), 1,
      sym_indexing_operator,
    ACTIONS(7952), 1,
      sym__lowercase_identifier,
    ACTIONS(7954), 1,
      anon_sym_LPAREN,
    ACTIONS(7956), 1,
      anon_sym_LBRACK,
    ACTIONS(7958), 1,
      anon_sym_LBRACE,
    STATE(932), 1,
      sym_field_path,
    STATE(5333), 1,
      sym_attribute,
    STATE(6615), 1,
      sym_indexing_operator_path,
    STATE(7802), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216538] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7853), 1,
      sym_indexing_operator,
    ACTIONS(7960), 1,
      sym__lowercase_identifier,
    ACTIONS(7962), 1,
      anon_sym_LPAREN,
    ACTIONS(7964), 1,
      anon_sym_LBRACK,
    ACTIONS(7966), 1,
      anon_sym_LBRACE,
    STATE(1206), 1,
      sym_field_path,
    STATE(5334), 1,
      sym_attribute,
    STATE(6678), 1,
      sym_indexing_operator_path,
    STATE(7434), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216576] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5335), 1,
      sym_attribute,
    STATE(8152), 1,
      sym_escape_sequence,
    STATE(8250), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [216606] = 12,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7853), 1,
      sym_indexing_operator,
    ACTIONS(7968), 1,
      sym__lowercase_identifier,
    ACTIONS(7970), 1,
      anon_sym_LPAREN,
    ACTIONS(7972), 1,
      anon_sym_LBRACK,
    ACTIONS(7974), 1,
      anon_sym_LBRACE,
    STATE(2463), 1,
      sym_field_path,
    STATE(5336), 1,
      sym_attribute,
    STATE(6701), 1,
      sym_indexing_operator_path,
    STATE(7614), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216644] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      aux_sym_character_content_token2,
    STATE(5337), 1,
      sym_attribute,
    STATE(8149), 1,
      sym_character_content,
    STATE(8152), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7855), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7859), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [216674] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7976), 1,
      anon_sym__,
    STATE(5338), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6465), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [216705] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    STATE(4327), 1,
      sym_constructor_declaration,
    STATE(5339), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [216736] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7978), 1,
      sym__right_quoted_string_delimiter,
    STATE(5340), 1,
      sym_attribute,
    STATE(5369), 1,
      aux_sym_quoted_string_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7909), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7907), 4,
      aux_sym_string_content_token1,
      aux_sym_quoted_string_content_token1,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [216763] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    STATE(4380), 1,
      sym_constructor_declaration,
    STATE(5341), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4788), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [216794] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7980), 1,
      anon_sym_EQ,
    ACTIONS(7982), 1,
      anon_sym_COLON2,
    ACTIONS(7986), 1,
      anon_sym_COLON_GT,
    STATE(4960), 1,
      sym__coerced,
    STATE(5342), 1,
      sym_attribute,
    STATE(5834), 1,
      sym__typed,
    STATE(6329), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7984), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [216827] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7988), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5343), 2,
      sym_attribute,
      aux_sym_type_binding_repeat1,
    ACTIONS(5112), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [216852] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5344), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5352), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [216873] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5345), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [216894] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7991), 1,
      anon_sym_EQ,
    ACTIONS(7993), 1,
      anon_sym_of,
    ACTIONS(7995), 1,
      anon_sym_COLON2,
    STATE(5346), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5342), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [216921] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7304), 1,
      anon_sym_LPAREN,
    ACTIONS(7310), 1,
      anon_sym_LBRACK,
    ACTIONS(7316), 1,
      sym__uppercase_identifier,
    STATE(5347), 1,
      sym_attribute,
    STATE(6030), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7312), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [216952] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5348), 1,
      sym_attribute,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5376), 1,
      aux_sym_type_binding_repeat1,
    STATE(5926), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5356), 2,
      anon_sym_and,
      anon_sym_in,
  [216985] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    STATE(4327), 1,
      sym_constructor_declaration,
    STATE(5349), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4375), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [217016] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5222), 1,
      aux_sym_type_binding_repeat1,
    STATE(5350), 1,
      sym_attribute,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5930), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5360), 2,
      anon_sym_and,
      anon_sym_in,
  [217049] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5222), 1,
      aux_sym_type_binding_repeat1,
    STATE(5351), 1,
      sym_attribute,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5933), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5364), 2,
      anon_sym_and,
      anon_sym_in,
  [217082] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5352), 1,
      sym_attribute,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5379), 1,
      aux_sym_type_binding_repeat1,
    STATE(5941), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5368), 2,
      anon_sym_and,
      anon_sym_in,
  [217115] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7881), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(5343), 1,
      aux_sym_type_binding_repeat1,
    STATE(5353), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5152), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [217142] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7881), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(5343), 1,
      aux_sym_type_binding_repeat1,
    STATE(5354), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5133), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [217169] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7881), 1,
      anon_sym_constraint,
    STATE(3914), 1,
      sym_type_constraint,
    STATE(5343), 1,
      aux_sym_type_binding_repeat1,
    STATE(5355), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5137), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [217196] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5356), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7997), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [217217] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5357), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7999), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [217238] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7976), 1,
      anon_sym__,
    STATE(5358), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6309), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [217269] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_and,
    STATE(5359), 1,
      sym_attribute,
    STATE(5382), 1,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5372), 6,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [217294] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(8003), 1,
      anon_sym_SEMI,
    ACTIONS(8005), 1,
      anon_sym_GT,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5360), 1,
      sym_attribute,
    STATE(6414), 1,
      aux_sym_object_type_repeat1,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217329] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7976), 1,
      anon_sym__,
    STATE(5361), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6638), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [217360] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7976), 1,
      anon_sym__,
    STATE(5362), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6644), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [217391] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5363), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5172), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [217412] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5222), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5364), 1,
      sym_attribute,
    STATE(5889), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5386), 2,
      anon_sym_and,
      anon_sym_in,
  [217445] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4945), 1,
      anon_sym_LPAREN,
    ACTIONS(4947), 1,
      anon_sym_LBRACK,
    ACTIONS(4951), 1,
      sym__uppercase_identifier,
    STATE(4380), 1,
      sym_constructor_declaration,
    STATE(5365), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4949), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4042), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [217476] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7976), 1,
      anon_sym__,
    STATE(5366), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6736), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [217507] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7976), 1,
      anon_sym__,
    STATE(5367), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6628), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [217538] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5222), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5368), 1,
      sym_attribute,
    STATE(5808), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5423), 2,
      anon_sym_and,
      anon_sym_in,
  [217571] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8013), 1,
      sym__right_quoted_string_delimiter,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8010), 2,
      sym__null,
      aux_sym_string_content_token2,
    STATE(5369), 2,
      sym_attribute,
      aux_sym_quoted_string_content_repeat1,
    ACTIONS(8007), 4,
      aux_sym_string_content_token1,
      aux_sym_quoted_string_content_token1,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [217596] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5350), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5370), 1,
      sym_attribute,
    STATE(5824), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5431), 2,
      anon_sym_and,
      anon_sym_in,
  [217629] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8015), 1,
      anon_sym_COLON,
    STATE(5371), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3382), 7,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [217652] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5351), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5372), 1,
      sym_attribute,
    STATE(5826), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5266), 2,
      anon_sym_and,
      anon_sym_in,
  [217685] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8017), 1,
      anon_sym_COLON,
    STATE(5373), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7772), 7,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [217708] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7796), 1,
      anon_sym_RPAREN,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8019), 1,
      anon_sym_COLON2,
    STATE(5374), 1,
      sym_attribute,
    STATE(7334), 1,
      sym__polymorphic_typed,
    STATE(7642), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217743] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5222), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5375), 1,
      sym_attribute,
    STATE(5837), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5294), 2,
      anon_sym_and,
      anon_sym_in,
  [217776] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5222), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5376), 1,
      sym_attribute,
    STATE(5978), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5400), 2,
      anon_sym_and,
      anon_sym_in,
  [217809] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(8021), 1,
      anon_sym_SEMI,
    ACTIONS(8023), 1,
      anon_sym_GT,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5377), 1,
      sym_attribute,
    STATE(6560), 1,
      aux_sym_object_type_repeat1,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217844] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5378), 1,
      sym_attribute,
    STATE(5386), 1,
      aux_sym_type_binding_repeat1,
    STATE(5979), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5404), 2,
      anon_sym_and,
      anon_sym_in,
  [217877] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5222), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5379), 1,
      sym_attribute,
    STATE(5987), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5408), 2,
      anon_sym_and,
      anon_sym_in,
  [217910] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5380), 1,
      sym_attribute,
    STATE(5392), 1,
      aux_sym_type_binding_repeat1,
    STATE(5970), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5156), 2,
      anon_sym_and,
      anon_sym_in,
  [217943] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7976), 1,
      anon_sym__,
    STATE(5381), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6419), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [217974] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8025), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5382), 2,
      sym_attribute,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(5254), 6,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [217997] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5222), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5383), 1,
      sym_attribute,
    STATE(5973), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5298), 2,
      anon_sym_and,
      anon_sym_in,
  [218030] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7976), 1,
      anon_sym__,
    STATE(5384), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6349), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [218061] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7976), 1,
      anon_sym__,
    STATE(5385), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6368), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [218092] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5222), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5386), 1,
      sym_attribute,
    STATE(5804), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5427), 2,
      anon_sym_and,
      anon_sym_in,
  [218125] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7976), 1,
      anon_sym__,
    STATE(5387), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6353), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [218156] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(7808), 1,
      anon_sym_RPAREN,
    ACTIONS(8019), 1,
      anon_sym_COLON2,
    STATE(5388), 1,
      sym_attribute,
    STATE(7328), 1,
      sym__typed,
    STATE(7378), 1,
      sym__polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218191] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7976), 1,
      anon_sym__,
    STATE(5389), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6472), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [218222] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_and,
    STATE(5359), 1,
      aux_sym_module_type_constraint_repeat1,
    STATE(5390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5322), 6,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [218247] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5368), 1,
      aux_sym_type_binding_repeat1,
    STATE(5391), 1,
      sym_attribute,
    STATE(5855), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5258), 2,
      anon_sym_and,
      anon_sym_in,
  [218280] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5222), 1,
      aux_sym_type_binding_repeat1,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5392), 1,
      sym_attribute,
    STATE(5874), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5306), 2,
      anon_sym_and,
      anon_sym_in,
  [218313] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7304), 1,
      anon_sym_LPAREN,
    ACTIONS(7310), 1,
      anon_sym_LBRACK,
    ACTIONS(7316), 1,
      sym__uppercase_identifier,
    STATE(5393), 1,
      sym_attribute,
    STATE(5891), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7312), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(5346), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [218344] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7746), 1,
      anon_sym_constraint,
    STATE(5363), 1,
      sym_type_constraint,
    STATE(5375), 1,
      aux_sym_type_binding_repeat1,
    STATE(5394), 1,
      sym_attribute,
    STATE(5998), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5119), 2,
      anon_sym_and,
      anon_sym_in,
  [218377] = 11,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(8028), 1,
      anon_sym_SEMI,
    ACTIONS(8030), 1,
      anon_sym_GT,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5395), 1,
      sym_attribute,
    STATE(6435), 1,
      aux_sym_object_type_repeat1,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218412] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7976), 1,
      anon_sym__,
    STATE(5396), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6727), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [218443] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7976), 1,
      anon_sym__,
    STATE(5397), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6730), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [218474] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7976), 1,
      anon_sym__,
    STATE(5398), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym__type_param_repeat1,
    STATE(5868), 1,
      sym_type_variable,
    STATE(6732), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [218505] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5399), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5376), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [218526] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8032), 1,
      anon_sym_RPAREN,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    STATE(5400), 1,
      sym_attribute,
    STATE(7359), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218558] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5401), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8038), 3,
      sym__right_quoted_string_delimiter,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(8036), 4,
      aux_sym_string_content_token1,
      aux_sym_quoted_string_content_token1,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [218580] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3396), 1,
      anon_sym_PIPE,
    STATE(5402), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3394), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [218602] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(6949), 1,
      anon_sym_STAR,
    ACTIONS(8040), 1,
      anon_sym_DASH_GT,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5403), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218634] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8042), 1,
      anon_sym_RBRACK,
    ACTIONS(8044), 1,
      anon_sym_when,
    STATE(5404), 1,
      sym_attribute,
    STATE(7726), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218666] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5396), 1,
      anon_sym_COLON2,
    STATE(5405), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5394), 6,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [218688] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(8046), 1,
      anon_sym_EQ,
    ACTIONS(8048), 1,
      anon_sym_COLON_EQ,
    STATE(5406), 1,
      sym_attribute,
    STATE(5920), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5435), 2,
      anon_sym_and,
      anon_sym_in,
  [218718] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8050), 1,
      anon_sym_DASH_GT,
    ACTIONS(8052), 1,
      anon_sym_when,
    STATE(5407), 1,
      sym_attribute,
    STATE(7859), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218750] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8056), 1,
      anon_sym__,
    ACTIONS(8058), 1,
      anon_sym_RBRACE,
    STATE(5408), 1,
      sym_attribute,
    STATE(5912), 1,
      sym_field_path,
    STATE(7089), 1,
      sym_field_binding_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218782] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8060), 1,
      anon_sym_rec,
    ACTIONS(8062), 1,
      anon_sym_type,
    STATE(5409), 1,
      sym_attribute,
    STATE(6089), 1,
      sym__attribute,
    STATE(6372), 1,
      sym_module_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8064), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [218812] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8066), 1,
      sym__lowercase_identifier,
    ACTIONS(8068), 1,
      anon_sym_type,
    ACTIONS(8070), 1,
      anon_sym_LBRACK,
    ACTIONS(8072), 1,
      anon_sym_virtual,
    STATE(5410), 1,
      sym_attribute,
    STATE(6127), 1,
      sym__attribute,
    STATE(6381), 1,
      sym_class_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218844] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8052), 1,
      anon_sym_when,
    ACTIONS(8074), 1,
      anon_sym_DASH_GT,
    STATE(5411), 1,
      sym_attribute,
    STATE(8017), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218876] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8076), 1,
      anon_sym_LBRACE,
    ACTIONS(8079), 1,
      anon_sym_DQUOTE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5552), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    STATE(5412), 2,
      sym_attribute,
      aux_sym_external_repeat1,
    STATE(6187), 2,
      sym_string,
      sym_quoted_string,
  [218902] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2103), 1,
      anon_sym_PIPE,
    STATE(5413), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2105), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [218924] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8082), 1,
      anon_sym_COMMA,
    ACTIONS(8084), 1,
      anon_sym_PIPE,
    ACTIONS(8086), 1,
      anon_sym_SEMI,
    ACTIONS(8088), 1,
      anon_sym_as,
    ACTIONS(8090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(8092), 1,
      anon_sym_PIPE_RBRACK,
    STATE(5414), 1,
      sym_attribute,
    STATE(6374), 1,
      aux_sym__sequence_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218956] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8094), 1,
      anon_sym_RPAREN,
    STATE(5415), 1,
      sym_attribute,
    STATE(8199), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218988] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2107), 1,
      anon_sym_PIPE,
    STATE(5416), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2109), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [219010] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8052), 1,
      anon_sym_when,
    ACTIONS(8096), 1,
      anon_sym_DASH_GT,
    STATE(5417), 1,
      sym_attribute,
    STATE(8179), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219042] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7776), 1,
      anon_sym_PIPE,
    ACTIONS(8098), 1,
      anon_sym_COLON,
    STATE(5418), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7772), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [219066] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(6949), 1,
      anon_sym_STAR,
    ACTIONS(8100), 1,
      anon_sym_DASH_GT,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5419), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219098] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8102), 1,
      anon_sym_COLON2,
    STATE(5420), 1,
      sym_attribute,
    STATE(6781), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 5,
      anon_sym_SEMI,
      anon_sym_GT,
      anon_sym_POUND,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219122] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8104), 1,
      anon_sym_RPAREN,
    ACTIONS(8106), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    STATE(5421), 1,
      sym_attribute,
    STATE(7191), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219154] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5422), 1,
      sym_attribute,
    STATE(5912), 1,
      sym_field_path,
    STATE(6741), 1,
      sym_field_binding_pattern,
    STATE(7263), 1,
      sym__quoted_string,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219186] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8052), 1,
      anon_sym_when,
    ACTIONS(8114), 1,
      anon_sym_DASH_GT,
    STATE(5423), 1,
      sym_attribute,
    STATE(7375), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219218] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8116), 1,
      anon_sym__,
    ACTIONS(8118), 1,
      anon_sym_RBRACE,
    STATE(5424), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6866), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219250] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5425), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6753), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    STATE(8059), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219282] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8120), 1,
      anon_sym_RPAREN,
    STATE(5426), 1,
      sym_attribute,
    STATE(8089), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219314] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2117), 1,
      anon_sym_PIPE,
    STATE(5427), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2119), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [219336] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8122), 1,
      anon_sym__,
    ACTIONS(8124), 1,
      anon_sym_RBRACE,
    STATE(5428), 1,
      sym_attribute,
    STATE(5912), 1,
      sym_field_path,
    STATE(7089), 1,
      sym_field_binding_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219368] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8126), 1,
      anon_sym_RPAREN,
    STATE(5429), 1,
      sym_attribute,
    STATE(7942), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219400] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1645), 1,
      anon_sym_PIPE,
    ACTIONS(8128), 1,
      anon_sym_DOT_DOT,
    STATE(5430), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1641), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [219424] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8130), 1,
      anon_sym__,
    ACTIONS(8132), 1,
      anon_sym_RBRACE,
    STATE(5431), 1,
      sym_attribute,
    STATE(5912), 1,
      sym_field_path,
    STATE(7089), 1,
      sym_field_binding_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219456] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8134), 1,
      anon_sym__,
    ACTIONS(8136), 1,
      anon_sym_RBRACE,
    STATE(5432), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6866), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219488] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8138), 1,
      anon_sym_RPAREN,
    STATE(5433), 1,
      sym_attribute,
    STATE(8030), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219520] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8140), 1,
      anon_sym_rec,
    ACTIONS(8142), 1,
      anon_sym_type,
    STATE(4522), 1,
      sym_module_binding,
    STATE(5434), 1,
      sym_attribute,
    STATE(6042), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8144), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [219550] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    STATE(5435), 1,
      sym_attribute,
    STATE(7996), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4971), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(5062), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [219576] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8146), 1,
      anon_sym__,
    ACTIONS(8148), 1,
      anon_sym_RBRACE,
    STATE(5436), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6866), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219608] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8150), 1,
      anon_sym_RPAREN,
    STATE(5437), 1,
      sym_attribute,
    STATE(7452), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219640] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5438), 1,
      sym_attribute,
    STATE(5912), 1,
      sym_field_path,
    STATE(6666), 1,
      sym_field_binding_pattern,
    STATE(7586), 1,
      sym__quoted_string,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219672] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5439), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6306), 1,
      sym_field_pattern,
    STATE(7263), 1,
      sym__quoted_string,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219704] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8152), 1,
      anon_sym__,
    ACTIONS(8154), 1,
      anon_sym_RBRACE,
    STATE(5440), 1,
      sym_attribute,
    STATE(5912), 1,
      sym_field_path,
    STATE(7089), 1,
      sym_field_binding_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219736] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2169), 1,
      anon_sym_PIPE,
    STATE(5441), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2171), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [219758] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8156), 1,
      anon_sym_RPAREN,
    STATE(5442), 1,
      sym_attribute,
    STATE(7239), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219790] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8158), 1,
      anon_sym_RPAREN,
    STATE(5443), 1,
      sym_attribute,
    STATE(7251), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219822] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5444), 1,
      sym_attribute,
    STATE(5912), 1,
      sym_field_path,
    STATE(6455), 1,
      sym_field_binding_pattern,
    STATE(7534), 1,
      sym__quoted_string,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219854] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(6949), 1,
      anon_sym_STAR,
    ACTIONS(8160), 1,
      anon_sym_DASH_GT,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5445), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219886] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8162), 1,
      anon_sym_RPAREN,
    STATE(5446), 1,
      sym_attribute,
    STATE(7305), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219918] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8164), 1,
      anon_sym_RPAREN,
    STATE(5447), 1,
      sym_attribute,
    STATE(7315), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219950] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2439), 1,
      anon_sym_PIPE,
    ACTIONS(8128), 1,
      anon_sym_DOT_DOT,
    STATE(5448), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2433), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [219974] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8166), 1,
      anon_sym__,
    ACTIONS(8168), 1,
      anon_sym_RBRACE,
    STATE(5449), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6866), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220006] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8052), 1,
      anon_sym_when,
    ACTIONS(8170), 1,
      anon_sym_DASH_GT,
    STATE(5450), 1,
      sym_attribute,
    STATE(7803), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220038] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3390), 1,
      anon_sym_PIPE,
    STATE(5451), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3388), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [220060] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8172), 1,
      anon_sym_RPAREN,
    STATE(5452), 1,
      sym_attribute,
    STATE(7365), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220092] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8174), 1,
      anon_sym__,
    ACTIONS(8176), 1,
      anon_sym_RBRACE,
    STATE(5453), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6866), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220124] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8178), 1,
      anon_sym_RPAREN,
    STATE(5454), 1,
      sym_attribute,
    STATE(7399), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220156] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8180), 1,
      anon_sym_RPAREN,
    STATE(5455), 1,
      sym_attribute,
    STATE(7404), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220188] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8182), 1,
      anon_sym_COMMA,
    ACTIONS(8184), 1,
      anon_sym_PIPE,
    ACTIONS(8186), 1,
      anon_sym_as,
    ACTIONS(8188), 1,
      anon_sym_COLON_COLON,
    STATE(5456), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3556), 3,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [220216] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2133), 1,
      anon_sym_PIPE,
    STATE(5457), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2135), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [220238] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8182), 1,
      anon_sym_COMMA,
    ACTIONS(8188), 1,
      anon_sym_COLON_COLON,
    STATE(5458), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3560), 5,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [220262] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8182), 1,
      anon_sym_COMMA,
    STATE(5459), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3560), 6,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [220284] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8182), 1,
      anon_sym_COMMA,
    ACTIONS(8184), 1,
      anon_sym_PIPE,
    ACTIONS(8188), 1,
      anon_sym_COLON_COLON,
    STATE(5460), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3568), 4,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [220310] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8190), 1,
      anon_sym_RPAREN,
    STATE(5461), 1,
      sym_attribute,
    STATE(7438), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220342] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8188), 1,
      anon_sym_COLON_COLON,
    STATE(5462), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3576), 6,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [220364] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8192), 1,
      anon_sym_RPAREN,
    STATE(5463), 1,
      sym_attribute,
    STATE(7443), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220396] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8194), 1,
      anon_sym__,
    ACTIONS(8196), 1,
      anon_sym_RBRACE,
    STATE(5464), 1,
      sym_attribute,
    STATE(5912), 1,
      sym_field_path,
    STATE(7089), 1,
      sym_field_binding_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220428] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8182), 1,
      anon_sym_COMMA,
    ACTIONS(8184), 1,
      anon_sym_PIPE,
    ACTIONS(8186), 1,
      anon_sym_as,
    ACTIONS(8188), 1,
      anon_sym_COLON_COLON,
    ACTIONS(8198), 1,
      anon_sym_RBRACK,
    ACTIONS(8200), 1,
      anon_sym_SEMI,
    STATE(5465), 1,
      sym_attribute,
    STATE(6320), 1,
      aux_sym__sequence_binding_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220460] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2153), 1,
      anon_sym_PIPE,
    STATE(5466), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2155), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [220482] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2173), 1,
      anon_sym_PIPE,
    STATE(5467), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2175), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [220504] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8202), 1,
      anon_sym_RPAREN,
    STATE(5468), 1,
      sym_attribute,
    STATE(7475), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220536] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8052), 1,
      anon_sym_when,
    ACTIONS(8204), 1,
      anon_sym_DASH_GT,
    STATE(5469), 1,
      sym_attribute,
    STATE(7736), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220568] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8206), 1,
      anon_sym_RPAREN,
    STATE(5470), 1,
      sym_attribute,
    STATE(7480), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220600] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5254), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [220620] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8106), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    ACTIONS(8208), 1,
      anon_sym_RPAREN,
    STATE(5472), 1,
      sym_attribute,
    STATE(7966), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220652] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8210), 1,
      anon_sym_RPAREN,
    STATE(5473), 1,
      sym_attribute,
    STATE(7511), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220684] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8212), 1,
      anon_sym_RPAREN,
    STATE(5474), 1,
      sym_attribute,
    STATE(7516), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220716] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(6949), 1,
      anon_sym_STAR,
    ACTIONS(8214), 1,
      anon_sym_DASH_GT,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5475), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220748] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5476), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5262), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [220768] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8216), 1,
      anon_sym_RPAREN,
    STATE(5477), 1,
      sym_attribute,
    STATE(7547), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220800] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8218), 1,
      anon_sym_RPAREN,
    STATE(5478), 1,
      sym_attribute,
    STATE(7552), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220832] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8220), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5479), 2,
      sym_attribute,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(5058), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [220854] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8223), 1,
      anon_sym_RPAREN,
    STATE(5480), 1,
      sym_attribute,
    STATE(7582), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220886] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8225), 1,
      anon_sym_RPAREN,
    STATE(5481), 1,
      sym_attribute,
    STATE(7587), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220918] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8227), 1,
      anon_sym__,
    ACTIONS(8229), 1,
      anon_sym_RBRACE,
    STATE(5482), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6866), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220950] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8231), 1,
      anon_sym_RPAREN,
    STATE(5483), 1,
      sym_attribute,
    STATE(7618), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220982] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8233), 1,
      anon_sym_RPAREN,
    STATE(5484), 1,
      sym_attribute,
    STATE(7623), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221014] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8235), 1,
      anon_sym_RPAREN,
    STATE(5485), 1,
      sym_attribute,
    STATE(7644), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221046] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8237), 1,
      anon_sym_RPAREN,
    STATE(5486), 1,
      sym_attribute,
    STATE(7646), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221078] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8239), 1,
      sym__lowercase_identifier,
    ACTIONS(8241), 1,
      anon_sym_type,
    ACTIONS(8243), 1,
      anon_sym_LBRACK,
    ACTIONS(8245), 1,
      anon_sym_virtual,
    STATE(4805), 1,
      sym_class_binding,
    STATE(5487), 1,
      sym_attribute,
    STATE(6105), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221110] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5312), 1,
      anon_sym_COLON2,
    STATE(5488), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5310), 6,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [221132] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8052), 1,
      anon_sym_when,
    ACTIONS(8247), 1,
      anon_sym_DASH_GT,
    STATE(5489), 1,
      sym_attribute,
    STATE(7928), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221164] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8241), 1,
      anon_sym_type,
    ACTIONS(8249), 1,
      sym__lowercase_identifier,
    ACTIONS(8251), 1,
      anon_sym_LBRACK,
    ACTIONS(8253), 1,
      anon_sym_virtual,
    STATE(4630), 1,
      sym_class_binding,
    STATE(5490), 1,
      sym_attribute,
    STATE(6296), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221196] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8255), 1,
      anon_sym__,
    ACTIONS(8257), 1,
      anon_sym_RBRACE,
    STATE(5491), 1,
      sym_attribute,
    STATE(5912), 1,
      sym_field_path,
    STATE(7089), 1,
      sym_field_binding_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221228] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5288), 1,
      anon_sym_COLON2,
    STATE(5492), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5286), 6,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [221250] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5493), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5062), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [221270] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8259), 1,
      anon_sym_DASH_GT,
    ACTIONS(8261), 1,
      anon_sym_with,
    STATE(5494), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5528), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [221294] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5495), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5250), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [221314] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5496), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6482), 1,
      sym_field_pattern,
    STATE(7534), 1,
      sym__quoted_string,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221346] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5497), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5226), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [221366] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(8263), 1,
      anon_sym_EQ,
    ACTIONS(8265), 1,
      anon_sym_COLON_EQ,
    STATE(5498), 1,
      sym_attribute,
    STATE(5981), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5540), 2,
      anon_sym_and,
      anon_sym_in,
  [221396] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(6949), 1,
      anon_sym_STAR,
    ACTIONS(8267), 1,
      anon_sym_DASH_GT,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5499), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221428] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8269), 1,
      anon_sym_COMMA,
    ACTIONS(8271), 1,
      anon_sym_PIPE,
    ACTIONS(8273), 1,
      anon_sym_COLON_COLON,
    STATE(5500), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7821), 3,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [221456] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8269), 1,
      anon_sym_COMMA,
    ACTIONS(8273), 1,
      anon_sym_COLON_COLON,
    STATE(5501), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7562), 5,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [221480] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8269), 1,
      anon_sym_COMMA,
    STATE(5502), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7562), 6,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [221502] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8269), 1,
      anon_sym_COMMA,
    ACTIONS(8271), 1,
      anon_sym_PIPE,
    ACTIONS(8273), 1,
      anon_sym_COLON_COLON,
    STATE(5503), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7570), 4,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [221528] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8273), 1,
      anon_sym_COLON_COLON,
    STATE(5504), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7788), 6,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [221550] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5505), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5246), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [221570] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8275), 1,
      anon_sym__,
    ACTIONS(8277), 1,
      anon_sym_RBRACE,
    STATE(5506), 1,
      sym_attribute,
    STATE(5912), 1,
      sym_field_path,
    STATE(7089), 1,
      sym_field_binding_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221602] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3384), 1,
      anon_sym_COLON2,
    ACTIONS(8279), 1,
      anon_sym_COLON,
    STATE(5507), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3382), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [221626] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5508), 1,
      sym_attribute,
    STATE(5912), 1,
      sym_field_path,
    STATE(6527), 1,
      sym_field_binding_pattern,
    STATE(7450), 1,
      sym__quoted_string,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221658] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8259), 1,
      anon_sym_DASH_GT,
    ACTIONS(8261), 1,
      anon_sym_with,
    STATE(5509), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5447), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [221682] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8281), 1,
      anon_sym_RPAREN,
    STATE(5510), 1,
      sym_attribute,
    STATE(8070), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221714] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8092), 1,
      anon_sym_RBRACK,
    ACTIONS(8269), 1,
      anon_sym_COMMA,
    ACTIONS(8271), 1,
      anon_sym_PIPE,
    ACTIONS(8273), 1,
      anon_sym_COLON_COLON,
    ACTIONS(8283), 1,
      anon_sym_SEMI,
    STATE(5511), 1,
      sym_attribute,
    STATE(6336), 1,
      aux_sym__sequence_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221746] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8259), 1,
      anon_sym_DASH_GT,
    ACTIONS(8261), 1,
      anon_sym_with,
    STATE(5512), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5492), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [221770] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8259), 1,
      anon_sym_DASH_GT,
    ACTIONS(8261), 1,
      anon_sym_with,
    STATE(5513), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5504), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [221794] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(6949), 1,
      anon_sym_STAR,
    ACTIONS(8285), 1,
      anon_sym_DASH_GT,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5514), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221826] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5515), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6646), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    STATE(7813), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221858] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2111), 1,
      anon_sym_PIPE,
    ACTIONS(2115), 1,
      anon_sym_DOT,
    STATE(5516), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [221882] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8182), 1,
      anon_sym_COMMA,
    ACTIONS(8184), 1,
      anon_sym_PIPE,
    ACTIONS(8186), 1,
      anon_sym_as,
    ACTIONS(8188), 1,
      anon_sym_COLON_COLON,
    STATE(5517), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3534), 3,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [221910] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5518), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4979), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [221930] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2111), 1,
      anon_sym_PIPE,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    STATE(5519), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2113), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [221954] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3384), 1,
      anon_sym_PIPE,
    ACTIONS(8287), 1,
      anon_sym_COLON,
    STATE(5520), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3382), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [221978] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8198), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(8289), 1,
      anon_sym_COMMA,
    ACTIONS(8291), 1,
      anon_sym_PIPE,
    ACTIONS(8293), 1,
      anon_sym_SEMI,
    ACTIONS(8295), 1,
      anon_sym_as,
    ACTIONS(8297), 1,
      anon_sym_COLON_COLON,
    STATE(5521), 1,
      sym_attribute,
    STATE(6357), 1,
      aux_sym__sequence_binding_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222010] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8052), 1,
      anon_sym_when,
    ACTIONS(8299), 1,
      anon_sym_DASH_GT,
    STATE(5522), 1,
      sym_attribute,
    STATE(8011), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222042] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4764), 1,
      anon_sym_STAR,
    STATE(5479), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(5523), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5461), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [222066] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8259), 1,
      anon_sym_DASH_GT,
    ACTIONS(8261), 1,
      anon_sym_with,
    STATE(5524), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5455), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [222090] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5525), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5222), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [222110] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8301), 1,
      anon_sym__,
    ACTIONS(8303), 1,
      anon_sym_RBRACE,
    STATE(5526), 1,
      sym_attribute,
    STATE(5912), 1,
      sym_field_path,
    STATE(7089), 1,
      sym_field_binding_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222142] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8241), 1,
      anon_sym_type,
    ACTIONS(8305), 1,
      sym__lowercase_identifier,
    ACTIONS(8307), 1,
      anon_sym_LBRACK,
    ACTIONS(8309), 1,
      anon_sym_virtual,
    STATE(4893), 1,
      sym_class_binding,
    STATE(5527), 1,
      sym_attribute,
    STATE(6119), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222174] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8052), 1,
      anon_sym_when,
    ACTIONS(8311), 1,
      anon_sym_DASH_GT,
    STATE(5528), 1,
      sym_attribute,
    STATE(7932), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222206] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8052), 1,
      anon_sym_when,
    ACTIONS(8313), 1,
      anon_sym_DASH_GT,
    STATE(5529), 1,
      sym_attribute,
    STATE(7369), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222238] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8269), 1,
      anon_sym_COMMA,
    ACTIONS(8271), 1,
      anon_sym_PIPE,
    ACTIONS(8273), 1,
      anon_sym_COLON_COLON,
    STATE(5530), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7837), 3,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [222266] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8106), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    ACTIONS(8315), 1,
      anon_sym_RPAREN,
    STATE(5531), 1,
      sym_attribute,
    STATE(7677), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222298] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8317), 1,
      anon_sym__,
    STATE(5532), 1,
      sym_attribute,
    STATE(5685), 1,
      aux_sym__type_param_repeat1,
    STATE(5862), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [222326] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(8040), 1,
      anon_sym_DASH_GT,
    ACTIONS(8319), 1,
      anon_sym_STAR,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5533), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222358] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8052), 1,
      anon_sym_when,
    ACTIONS(8322), 1,
      anon_sym_DASH_GT,
    STATE(5534), 1,
      sym_attribute,
    STATE(7697), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222390] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8324), 1,
      anon_sym__,
    ACTIONS(8326), 1,
      anon_sym_RBRACE,
    STATE(5535), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6866), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222422] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8328), 1,
      anon_sym_RPAREN,
    STATE(5536), 1,
      sym_attribute,
    STATE(7332), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222454] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8330), 1,
      anon_sym_RPAREN,
    STATE(5537), 1,
      sym_attribute,
    STATE(7745), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222486] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8106), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    ACTIONS(8332), 1,
      anon_sym_RPAREN,
    STATE(5538), 1,
      sym_attribute,
    STATE(7298), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222518] = 10,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8054), 1,
      sym__lowercase_identifier,
    ACTIONS(8334), 1,
      anon_sym__,
    ACTIONS(8336), 1,
      anon_sym_RBRACE,
    STATE(5539), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6866), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222550] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8338), 1,
      anon_sym_rec,
    ACTIONS(8340), 1,
      anon_sym_type,
    STATE(4906), 1,
      sym_module_binding,
    STATE(5540), 1,
      sym_attribute,
    STATE(6297), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8342), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [222580] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5541), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8344), 2,
      anon_sym_SEMI,
      anon_sym_GT,
  [222610] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8346), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5542), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222639] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4084), 1,
      anon_sym_PIPE,
    STATE(5543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4086), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [222660] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8348), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5544), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222689] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8082), 1,
      anon_sym_COMMA,
    ACTIONS(8084), 1,
      anon_sym_PIPE,
    ACTIONS(8088), 1,
      anon_sym_as,
    ACTIONS(8090), 1,
      anon_sym_COLON_COLON,
    STATE(5545), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7837), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [222716] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7496), 1,
      anon_sym_PIPE,
    STATE(5546), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7494), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [222737] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7500), 1,
      anon_sym_PIPE,
    STATE(5547), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7498), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [222758] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7516), 1,
      anon_sym_PIPE,
    STATE(5548), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7514), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [222779] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8350), 1,
      anon_sym_RPAREN,
    STATE(4960), 1,
      sym__coerced,
    STATE(5549), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7351), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222808] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5445), 1,
      anon_sym_and,
    STATE(5550), 1,
      sym_attribute,
    STATE(5578), 1,
      aux_sym_expression_item_repeat1,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5443), 2,
      anon_sym_in,
      sym_let_and_operator,
  [222835] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3554), 1,
      anon_sym_PIPE,
    STATE(5551), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3552), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [222856] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8352), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5552), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222885] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5522), 1,
      anon_sym_and,
    STATE(5553), 1,
      sym_attribute,
    STATE(5732), 1,
      aux_sym_expression_item_repeat1,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5520), 2,
      anon_sym_in,
      sym_let_and_operator,
  [222912] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5202), 1,
      anon_sym_and,
    ACTIONS(8354), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5200), 2,
      anon_sym_in,
      sym_let_and_operator,
    STATE(5554), 2,
      sym_attribute,
      aux_sym_expression_item_repeat1,
  [222937] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8357), 1,
      anon_sym_RPAREN,
    STATE(4960), 1,
      sym__coerced,
    STATE(5555), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7761), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222966] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8359), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5556), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222995] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1645), 1,
      anon_sym_PIPE,
    STATE(5557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1641), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223016] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3444), 1,
      anon_sym_PIPE,
    STATE(5558), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3442), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223037] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8361), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5559), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223066] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3622), 1,
      anon_sym_PIPE,
    STATE(5560), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3620), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223087] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3614), 1,
      anon_sym_PIPE,
    STATE(5561), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3612), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223108] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8363), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5562), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223137] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8365), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5563), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223166] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3464), 1,
      anon_sym_PIPE,
    STATE(5564), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3462), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223187] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4009), 1,
      anon_sym_PIPE,
    STATE(5565), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4011), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223208] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8367), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5566), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223237] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2157), 1,
      anon_sym_PIPE,
    STATE(5567), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2159), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223258] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8269), 1,
      anon_sym_COMMA,
    ACTIONS(8271), 1,
      anon_sym_PIPE,
    ACTIONS(8273), 1,
      anon_sym_COLON_COLON,
    STATE(5568), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8369), 2,
      anon_sym_RBRACK,
      anon_sym_SEMI,
  [223285] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4013), 1,
      anon_sym_PIPE,
    STATE(5569), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4015), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223306] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2149), 1,
      anon_sym_PIPE,
    STATE(5570), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2151), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223327] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8371), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5571), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223356] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5510), 1,
      anon_sym_and,
    STATE(5554), 1,
      aux_sym_expression_item_repeat1,
    STATE(5572), 1,
      sym_attribute,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5508), 2,
      anon_sym_in,
      sym_let_and_operator,
  [223383] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8269), 1,
      anon_sym_COMMA,
    ACTIONS(8271), 1,
      anon_sym_PIPE,
    ACTIONS(8273), 1,
      anon_sym_COLON_COLON,
    STATE(5573), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8373), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [223410] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5526), 1,
      anon_sym_and,
    STATE(5574), 1,
      sym_attribute,
    STATE(5758), 1,
      aux_sym_expression_item_repeat1,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5524), 2,
      anon_sym_in,
      sym_let_and_operator,
  [223437] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8375), 1,
      sym__lowercase_identifier,
    STATE(1430), 1,
      sym_class_path,
    STATE(5575), 1,
      sym_attribute,
    STATE(6250), 1,
      sym__attribute,
    STATE(7413), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223466] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8377), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5576), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223495] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8379), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5577), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223524] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5502), 1,
      anon_sym_and,
    STATE(5554), 1,
      aux_sym_expression_item_repeat1,
    STATE(5578), 1,
      sym_attribute,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5500), 2,
      anon_sym_in,
      sym_let_and_operator,
  [223551] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8381), 1,
      sym__lowercase_identifier,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5579), 1,
      sym_attribute,
    STATE(6084), 1,
      sym__attribute,
    STATE(6883), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223580] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8289), 1,
      anon_sym_COMMA,
    ACTIONS(8291), 1,
      anon_sym_PIPE,
    ACTIONS(8295), 1,
      anon_sym_as,
    ACTIONS(8297), 1,
      anon_sym_COLON_COLON,
    STATE(5580), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3534), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [223607] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3506), 1,
      anon_sym_PIPE,
    STATE(5581), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3504), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223628] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3566), 1,
      anon_sym_PIPE,
    STATE(5582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3564), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223649] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3524), 1,
      anon_sym_PIPE,
    STATE(5583), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3522), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223670] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7614), 1,
      anon_sym_PIPE,
    STATE(5584), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7612), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223691] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7630), 1,
      anon_sym_PIPE,
    STATE(5585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7628), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223712] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3586), 1,
      anon_sym_PIPE,
    STATE(5586), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3584), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223733] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2195), 1,
      anon_sym_PIPE,
    STATE(5587), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2197), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223754] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2199), 1,
      anon_sym_PIPE,
    STATE(5588), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2201), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223775] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8385), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5589), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223804] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3590), 1,
      anon_sym_PIPE,
    STATE(5590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3588), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223825] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2203), 1,
      anon_sym_PIPE,
    STATE(5591), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2205), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223846] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3602), 1,
      anon_sym_PIPE,
    STATE(5592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3600), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223867] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8387), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5593), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223896] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5467), 1,
      anon_sym_and,
    STATE(5594), 1,
      sym_attribute,
    STATE(5733), 1,
      aux_sym_expression_item_repeat1,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5465), 2,
      anon_sym_in,
      sym_let_and_operator,
  [223923] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2031), 1,
      anon_sym_PIPE,
    STATE(5595), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2033), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223944] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7634), 1,
      anon_sym_PIPE,
    STATE(5596), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7632), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223965] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7638), 1,
      anon_sym_PIPE,
    STATE(5597), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7636), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [223986] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3528), 1,
      anon_sym_PIPE,
    STATE(5598), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3526), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224007] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8389), 1,
      sym__lowercase_identifier,
    STATE(2924), 1,
      sym_class_path,
    STATE(5599), 1,
      sym_attribute,
    STATE(6131), 1,
      sym__attribute,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224036] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8391), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5600), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224065] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8393), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5601), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224094] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8395), 1,
      anon_sym_and,
    STATE(5602), 1,
      sym_attribute,
    STATE(5605), 1,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5372), 4,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [224117] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3438), 1,
      anon_sym_PIPE,
    STATE(5603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3436), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224138] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2137), 1,
      anon_sym_PIPE,
    STATE(5604), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2139), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224159] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8397), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5605), 2,
      sym_attribute,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(5254), 4,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [224180] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3532), 1,
      anon_sym_PIPE,
    STATE(5606), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3530), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224201] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8400), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5607), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224230] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3594), 1,
      anon_sym_PIPE,
    STATE(5608), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3592), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224251] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3610), 1,
      anon_sym_PIPE,
    STATE(5609), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3608), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224272] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8402), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5610), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224301] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7726), 1,
      anon_sym_PIPE,
    STATE(5611), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7724), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224322] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8404), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5612), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224351] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8381), 1,
      sym__lowercase_identifier,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5613), 1,
      sym_attribute,
    STATE(6110), 1,
      sym__attribute,
    STATE(7070), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224380] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8406), 1,
      anon_sym_RPAREN,
    STATE(4960), 1,
      sym__coerced,
    STATE(5614), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(8055), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224409] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5518), 1,
      anon_sym_and,
    STATE(5554), 1,
      aux_sym_expression_item_repeat1,
    STATE(5615), 1,
      sym_attribute,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5516), 2,
      anon_sym_in,
      sym_let_and_operator,
  [224436] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5514), 1,
      anon_sym_and,
    STATE(5616), 1,
      sym_attribute,
    STATE(5619), 1,
      aux_sym_expression_item_repeat1,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5512), 2,
      anon_sym_in,
      sym_let_and_operator,
  [224463] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2121), 1,
      anon_sym_PIPE,
    STATE(5617), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2123), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224484] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8408), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5618), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224513] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5568), 1,
      anon_sym_and,
    STATE(5554), 1,
      aux_sym_expression_item_repeat1,
    STATE(5619), 1,
      sym_attribute,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5566), 2,
      anon_sym_in,
      sym_let_and_operator,
  [224540] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8410), 1,
      sym__lowercase_identifier,
    ACTIONS(8412), 1,
      anon_sym_LBRACK,
    ACTIONS(8414), 1,
      anon_sym_virtual,
    STATE(4554), 1,
      sym_class_type_binding,
    STATE(5620), 1,
      sym_attribute,
    STATE(6289), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224569] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8416), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5621), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224598] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8418), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5622), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224627] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8420), 1,
      sym__lowercase_identifier,
    STATE(2560), 1,
      sym_class_path,
    STATE(5623), 1,
      sym_attribute,
    STATE(6027), 1,
      sym__attribute,
    STATE(7201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224656] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3546), 1,
      anon_sym_PIPE,
    STATE(5624), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3544), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224677] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8422), 1,
      sym__lowercase_identifier,
    STATE(3276), 1,
      sym_class_path,
    STATE(5625), 1,
      sym_attribute,
    STATE(6228), 1,
      sym__attribute,
    STATE(7532), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224706] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3502), 1,
      anon_sym_PIPE,
    STATE(5626), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3500), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224727] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8424), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5627), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224756] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8426), 1,
      anon_sym_RPAREN,
    STATE(4960), 1,
      sym__coerced,
    STATE(5628), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7674), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224785] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8428), 1,
      anon_sym_PIPE,
    STATE(5629), 1,
      sym_attribute,
    STATE(5748), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4941), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [224808] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4048), 1,
      anon_sym_PIPE,
    STATE(5630), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4050), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224829] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2125), 1,
      anon_sym_PIPE,
    STATE(5631), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2127), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224850] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8430), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5632), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224879] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2439), 1,
      anon_sym_PIPE,
    STATE(5633), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2433), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224900] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    STATE(5634), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2147), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224921] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8432), 1,
      anon_sym_COLON2,
    STATE(5635), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4971), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(5062), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [224944] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4042), 1,
      anon_sym_PIPE,
    STATE(5636), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4044), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224965] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2179), 1,
      anon_sym_PIPE,
    STATE(5637), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2181), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [224986] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3448), 1,
      anon_sym_PIPE,
    STATE(5638), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3446), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [225007] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8434), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5639), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225036] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3456), 1,
      anon_sym_PIPE,
    STATE(5640), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3454), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [225057] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(8436), 1,
      anon_sym_EQ,
    STATE(5641), 1,
      sym_attribute,
    STATE(5793), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5922), 2,
      anon_sym_and,
      anon_sym_in,
  [225084] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3460), 1,
      anon_sym_PIPE,
    STATE(5642), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3458), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [225105] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8438), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5643), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225134] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(8440), 1,
      anon_sym_EQ,
    STATE(5644), 1,
      sym_attribute,
    STATE(5796), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5932), 2,
      anon_sym_and,
      anon_sym_in,
  [225161] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8442), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5645), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225190] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8444), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5646), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225219] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7552), 1,
      anon_sym_PIPE,
    STATE(5647), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7550), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [225240] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8446), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5648), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225269] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4038), 1,
      anon_sym_PIPE,
    STATE(5649), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4040), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [225290] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4076), 1,
      anon_sym_PIPE,
    STATE(5650), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4078), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [225311] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3556), 1,
      anon_sym_RPAREN,
    ACTIONS(3558), 1,
      anon_sym_COLON2,
    ACTIONS(8106), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    STATE(5651), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225340] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8448), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5652), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225369] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3582), 1,
      anon_sym_PIPE,
    STATE(5653), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3580), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [225390] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8450), 1,
      sym__lowercase_identifier,
    STATE(1198), 1,
      sym_class_path,
    STATE(5654), 1,
      sym_attribute,
    STATE(6207), 1,
      sym__attribute,
    STATE(7567), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225419] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3967), 1,
      anon_sym_PIPE,
    STATE(5655), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3969), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [225440] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3971), 1,
      anon_sym_PIPE,
    STATE(5656), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3973), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [225461] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3562), 1,
      anon_sym_COLON2,
    ACTIONS(8106), 1,
      anon_sym_COMMA,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    STATE(5657), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3560), 3,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_as,
  [225486] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3562), 1,
      anon_sym_COLON2,
    ACTIONS(8106), 1,
      anon_sym_COMMA,
    STATE(5658), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3560), 4,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [225509] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8452), 1,
      anon_sym_RPAREN,
    STATE(4960), 1,
      sym__coerced,
    STATE(5659), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(8253), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225538] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3975), 1,
      anon_sym_PIPE,
    STATE(5660), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3977), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [225559] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8454), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5661), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225588] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5622), 1,
      anon_sym_in,
    ACTIONS(8261), 1,
      anon_sym_with,
    ACTIONS(8456), 1,
      anon_sym_DASH_GT,
    STATE(5662), 1,
      sym_attribute,
    STATE(6035), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225617] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3570), 1,
      anon_sym_COLON2,
    ACTIONS(8106), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    STATE(5663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3568), 2,
      anon_sym_RPAREN,
      anon_sym_as,
  [225644] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8458), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5664), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225673] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8460), 1,
      anon_sym_RPAREN,
    STATE(4960), 1,
      sym__coerced,
    STATE(5665), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7910), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225702] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5129), 1,
      anon_sym_COLON2,
    STATE(5666), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 5,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225723] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3578), 1,
      anon_sym_COLON2,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    STATE(5667), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3576), 4,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_as,
  [225746] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5576), 1,
      anon_sym_and,
    STATE(5554), 1,
      aux_sym_expression_item_repeat1,
    STATE(5668), 1,
      sym_attribute,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5574), 2,
      anon_sym_in,
      sym_let_and_operator,
  [225773] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8462), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5669), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225802] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8381), 1,
      sym__lowercase_identifier,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5670), 1,
      sym_attribute,
    STATE(6068), 1,
      sym__attribute,
    STATE(6948), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225831] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5657), 1,
      anon_sym_in,
    STATE(5671), 1,
      sym_attribute,
    STATE(6217), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8464), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [225858] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8466), 1,
      anon_sym_RPAREN,
    STATE(4960), 1,
      sym__coerced,
    STATE(5672), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7758), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225887] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8468), 1,
      sym__lowercase_identifier,
    STATE(1952), 1,
      sym_class_path,
    STATE(5673), 1,
      sym_attribute,
    STATE(6144), 1,
      sym__attribute,
    STATE(7460), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225916] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8470), 1,
      sym__lowercase_identifier,
    STATE(2305), 1,
      sym_class_path,
    STATE(5674), 1,
      sym_attribute,
    STATE(6132), 1,
      sym__attribute,
    STATE(7384), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225945] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8182), 1,
      anon_sym_COMMA,
    ACTIONS(8184), 1,
      anon_sym_PIPE,
    ACTIONS(8186), 1,
      anon_sym_as,
    ACTIONS(8188), 1,
      anon_sym_COLON_COLON,
    STATE(5675), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8472), 2,
      anon_sym_RBRACK,
      anon_sym_SEMI,
  [225972] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8474), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5676), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226001] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4562), 1,
      anon_sym_POUND,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(8476), 1,
      anon_sym_DASH_GT,
    STATE(3655), 1,
      sym_type_constructor_path,
    STATE(5677), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226030] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3550), 1,
      anon_sym_PIPE,
    STATE(5678), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3548), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [226051] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8478), 1,
      sym__lowercase_identifier,
    STATE(1303), 1,
      sym_class_path,
    STATE(5679), 1,
      sym_attribute,
    STATE(6174), 1,
      sym__attribute,
    STATE(7423), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226080] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8182), 1,
      anon_sym_COMMA,
    ACTIONS(8184), 1,
      anon_sym_PIPE,
    ACTIONS(8186), 1,
      anon_sym_as,
    ACTIONS(8188), 1,
      anon_sym_COLON_COLON,
    STATE(5680), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8480), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [226107] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3598), 1,
      anon_sym_PIPE,
    STATE(5681), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3596), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [226128] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8482), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5682), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226157] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8484), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5683), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226186] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(8486), 1,
      anon_sym_EQ,
    STATE(5684), 1,
      sym_attribute,
    STATE(5956), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5866), 2,
      anon_sym_and,
      anon_sym_in,
  [226213] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8491), 2,
      anon_sym__,
      aux_sym_type_variable_token1,
    STATE(5685), 2,
      sym_attribute,
      aux_sym__type_param_repeat1,
    ACTIONS(8488), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [226234] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5667), 1,
      anon_sym_in,
    STATE(5686), 1,
      sym_attribute,
    STATE(6046), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8493), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [226261] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8495), 1,
      anon_sym_RPAREN,
    STATE(4960), 1,
      sym__coerced,
    STATE(5687), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7787), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226290] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8497), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5688), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226319] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8499), 1,
      anon_sym_RPAREN,
    STATE(4960), 1,
      sym__coerced,
    STATE(5689), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7576), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226348] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2129), 1,
      anon_sym_PIPE,
    STATE(5690), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2131), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [226369] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8501), 1,
      anon_sym_COLON2,
    STATE(5691), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4971), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(5062), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [226392] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8503), 1,
      sym__lowercase_identifier,
    STATE(1345), 1,
      sym_class_path,
    STATE(5692), 1,
      sym_attribute,
    STATE(6283), 1,
      sym__attribute,
    STATE(7279), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226421] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(8505), 1,
      anon_sym_EQ,
    STATE(5693), 1,
      sym_attribute,
    STATE(5986), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5685), 2,
      anon_sym_and,
      anon_sym_in,
  [226448] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8395), 1,
      anon_sym_and,
    STATE(5602), 1,
      aux_sym_module_type_constraint_repeat1,
    STATE(5694), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5322), 4,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [226471] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8507), 1,
      anon_sym_COLON2,
    STATE(5695), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4971), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(5062), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [226494] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(8509), 1,
      anon_sym_EQ,
    STATE(5696), 1,
      sym_attribute,
    STATE(5994), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5717), 2,
      anon_sym_and,
      anon_sym_in,
  [226521] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8511), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5697), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226550] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8513), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5698), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226579] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8182), 1,
      anon_sym_COMMA,
    ACTIONS(8184), 1,
      anon_sym_PIPE,
    ACTIONS(8186), 1,
      anon_sym_as,
    ACTIONS(8188), 1,
      anon_sym_COLON_COLON,
    STATE(5699), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8515), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [226606] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8517), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5700), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226635] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(8519), 1,
      anon_sym_EQ,
    STATE(5701), 1,
      sym_attribute,
    STATE(5960), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5876), 2,
      anon_sym_and,
      anon_sym_in,
  [226662] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3606), 1,
      anon_sym_PIPE,
    STATE(5702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3604), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [226683] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3618), 1,
      anon_sym_PIPE,
    STATE(5703), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3616), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [226704] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8521), 1,
      sym__lowercase_identifier,
    ACTIONS(8523), 1,
      anon_sym_BANG,
    STATE(4793), 1,
      sym__instance_variable_name,
    STATE(5704), 1,
      sym_attribute,
    STATE(5873), 1,
      aux_sym_instance_variable_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8525), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [226731] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8527), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5705), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226760] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8529), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5706), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226789] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8082), 1,
      anon_sym_COMMA,
    ACTIONS(8084), 1,
      anon_sym_PIPE,
    ACTIONS(8088), 1,
      anon_sym_as,
    ACTIONS(8090), 1,
      anon_sym_COLON_COLON,
    STATE(5707), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7821), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [226816] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7560), 1,
      anon_sym_PIPE,
    STATE(5708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7558), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [226837] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7564), 1,
      anon_sym_PIPE,
    STATE(5709), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7562), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [226858] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7564), 1,
      anon_sym_PIPE,
    ACTIONS(8082), 1,
      anon_sym_COMMA,
    ACTIONS(8090), 1,
      anon_sym_COLON_COLON,
    STATE(5710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7562), 3,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [226883] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7564), 1,
      anon_sym_PIPE,
    ACTIONS(8082), 1,
      anon_sym_COMMA,
    STATE(5711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7562), 4,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [226906] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7564), 1,
      anon_sym_PIPE,
    STATE(5712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7562), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [226927] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8082), 1,
      anon_sym_COMMA,
    ACTIONS(8084), 1,
      anon_sym_PIPE,
    ACTIONS(8090), 1,
      anon_sym_COLON_COLON,
    STATE(5713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7570), 3,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [226952] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7572), 1,
      anon_sym_PIPE,
    STATE(5714), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7570), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [226973] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7576), 1,
      anon_sym_PIPE,
    STATE(5715), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7574), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [226994] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7790), 1,
      anon_sym_PIPE,
    ACTIONS(8090), 1,
      anon_sym_COLON_COLON,
    STATE(5716), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7788), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [227017] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7580), 1,
      anon_sym_PIPE,
    STATE(5717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7578), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [227038] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8531), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5718), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227067] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3993), 1,
      anon_sym_PIPE,
    STATE(5719), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3995), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [227088] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8533), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5720), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227117] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8289), 1,
      anon_sym_COMMA,
    ACTIONS(8291), 1,
      anon_sym_PIPE,
    ACTIONS(8295), 1,
      anon_sym_as,
    ACTIONS(8297), 1,
      anon_sym_COLON_COLON,
    STATE(5721), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3556), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [227144] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3562), 1,
      anon_sym_PIPE,
    STATE(5722), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3560), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [227165] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3562), 1,
      anon_sym_PIPE,
    ACTIONS(8289), 1,
      anon_sym_COMMA,
    ACTIONS(8297), 1,
      anon_sym_COLON_COLON,
    STATE(5723), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3560), 3,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [227190] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3562), 1,
      anon_sym_PIPE,
    ACTIONS(8289), 1,
      anon_sym_COMMA,
    STATE(5724), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3560), 4,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [227213] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3562), 1,
      anon_sym_PIPE,
    STATE(5725), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3560), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [227234] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8289), 1,
      anon_sym_COMMA,
    ACTIONS(8291), 1,
      anon_sym_PIPE,
    ACTIONS(8297), 1,
      anon_sym_COLON_COLON,
    STATE(5726), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3568), 3,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [227259] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3570), 1,
      anon_sym_PIPE,
    STATE(5727), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3568), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [227280] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3574), 1,
      anon_sym_PIPE,
    STATE(5728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3572), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [227301] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3578), 1,
      anon_sym_PIPE,
    ACTIONS(8297), 1,
      anon_sym_COLON_COLON,
    STATE(5729), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3576), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [227324] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4001), 1,
      anon_sym_PIPE,
    STATE(5730), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4003), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [227345] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8289), 1,
      anon_sym_COMMA,
    ACTIONS(8291), 1,
      anon_sym_PIPE,
    ACTIONS(8295), 1,
      anon_sym_as,
    ACTIONS(8297), 1,
      anon_sym_COLON_COLON,
    STATE(5731), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8472), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [227372] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5572), 1,
      anon_sym_and,
    STATE(5554), 1,
      aux_sym_expression_item_repeat1,
    STATE(5732), 1,
      sym_attribute,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5570), 2,
      anon_sym_in,
      sym_let_and_operator,
  [227399] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5471), 1,
      anon_sym_and,
    STATE(5554), 1,
      aux_sym_expression_item_repeat1,
    STATE(5733), 1,
      sym_attribute,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5469), 2,
      anon_sym_in,
      sym_let_and_operator,
  [227426] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8535), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5734), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227455] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7722), 1,
      anon_sym_PIPE,
    STATE(5735), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7720), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [227476] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5482), 1,
      anon_sym_and,
    STATE(5736), 1,
      sym_attribute,
    STATE(5783), 1,
      aux_sym_expression_item_repeat1,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5480), 2,
      anon_sym_in,
      sym_let_and_operator,
  [227503] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5486), 1,
      anon_sym_and,
    STATE(5572), 1,
      aux_sym_expression_item_repeat1,
    STATE(5737), 1,
      sym_attribute,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5484), 2,
      anon_sym_in,
      sym_let_and_operator,
  [227530] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5490), 1,
      anon_sym_and,
    STATE(5615), 1,
      aux_sym_expression_item_repeat1,
    STATE(5738), 1,
      sym_attribute,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5488), 2,
      anon_sym_in,
      sym_let_and_operator,
  [227557] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8539), 1,
      anon_sym_AMP,
    STATE(5739), 1,
      sym_attribute,
    STATE(5821), 1,
      aux_sym_tag_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8537), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [227582] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8541), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5740), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227611] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8543), 1,
      anon_sym_COLON2,
    STATE(5741), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4971), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(5062), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [227634] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8545), 1,
      anon_sym_COLON2,
    STATE(5742), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4971), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(5062), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [227657] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2161), 1,
      anon_sym_PIPE,
    STATE(5743), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2163), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [227678] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3534), 1,
      anon_sym_RPAREN,
    ACTIONS(3536), 1,
      anon_sym_COLON2,
    ACTIONS(8106), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    STATE(5744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227707] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(8547), 1,
      anon_sym_EQ,
    STATE(5745), 1,
      sym_attribute,
    STATE(5935), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6128), 2,
      anon_sym_and,
      anon_sym_in,
  [227734] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8428), 1,
      anon_sym_PIPE,
    STATE(5746), 1,
      sym_attribute,
    STATE(5775), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5614), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [227757] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8549), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5747), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227786] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8428), 1,
      anon_sym_PIPE,
    STATE(5748), 1,
      sym_attribute,
    STATE(5776), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5614), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [227809] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8551), 1,
      sym__lowercase_identifier,
    STATE(979), 1,
      sym_class_path,
    STATE(5749), 1,
      sym_attribute,
    STATE(6304), 1,
      sym__attribute,
    STATE(7549), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227838] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8553), 1,
      sym__lowercase_identifier,
    STATE(2454), 1,
      sym_class_path,
    STATE(5750), 1,
      sym_attribute,
    STATE(6053), 1,
      sym__attribute,
    STATE(7636), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227867] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8555), 1,
      anon_sym_RPAREN,
    STATE(4960), 1,
      sym__coerced,
    STATE(5751), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(8105), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227896] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8557), 1,
      anon_sym_RPAREN,
    STATE(4960), 1,
      sym__coerced,
    STATE(5752), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7958), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227925] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3452), 1,
      anon_sym_PIPE,
    STATE(5753), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3450), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [227946] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3430), 1,
      anon_sym_PIPE,
    STATE(5754), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3428), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [227967] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5638), 1,
      anon_sym_in,
    ACTIONS(8261), 1,
      anon_sym_with,
    ACTIONS(8456), 1,
      anon_sym_DASH_GT,
    STATE(5755), 1,
      sym_attribute,
    STATE(6029), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227996] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3434), 1,
      anon_sym_PIPE,
    STATE(5756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3432), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [228017] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8269), 1,
      anon_sym_COMMA,
    ACTIONS(8271), 1,
      anon_sym_PIPE,
    ACTIONS(8273), 1,
      anon_sym_COLON_COLON,
    STATE(5757), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8559), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [228044] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5580), 1,
      anon_sym_and,
    STATE(5554), 1,
      aux_sym_expression_item_repeat1,
    STATE(5758), 1,
      sym_attribute,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5578), 2,
      anon_sym_in,
      sym_let_and_operator,
  [228071] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8561), 1,
      sym__lowercase_identifier,
    STATE(1049), 1,
      sym_class_path,
    STATE(5759), 1,
      sym_attribute,
    STATE(6199), 1,
      sym__attribute,
    STATE(7496), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228100] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(8563), 1,
      anon_sym_EQ,
    STATE(5760), 1,
      sym_attribute,
    STATE(5825), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5956), 2,
      anon_sym_and,
      anon_sym_in,
  [228127] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8565), 1,
      sym__lowercase_identifier,
    STATE(2205), 1,
      sym_class_path,
    STATE(5761), 1,
      sym_attribute,
    STATE(6191), 1,
      sym__attribute,
    STATE(7777), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228156] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3470), 1,
      anon_sym_PIPE,
    STATE(5762), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3468), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [228177] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8567), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5763), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228206] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8569), 1,
      anon_sym_RPAREN,
    STATE(4960), 1,
      sym__coerced,
    STATE(5764), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(8306), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228235] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(8571), 1,
      anon_sym_EQ,
    STATE(5765), 1,
      sym_attribute,
    STATE(5924), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5778), 2,
      anon_sym_and,
      anon_sym_in,
  [228262] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8381), 1,
      sym__lowercase_identifier,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5766), 1,
      sym_attribute,
    STATE(6248), 1,
      sym__attribute,
    STATE(7026), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228291] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    ACTIONS(8573), 1,
      anon_sym_RPAREN,
    STATE(4960), 1,
      sym__coerced,
    STATE(5767), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7814), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228320] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4034), 1,
      anon_sym_PIPE,
    STATE(5768), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4036), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [228341] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(8575), 1,
      anon_sym_EQ,
    STATE(5769), 1,
      sym_attribute,
    STATE(5865), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5824), 2,
      anon_sym_and,
      anon_sym_in,
  [228368] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8577), 1,
      sym__lowercase_identifier,
    ACTIONS(8579), 1,
      anon_sym_LBRACK,
    ACTIONS(8581), 1,
      anon_sym_virtual,
    STATE(5770), 1,
      sym_attribute,
    STATE(6242), 1,
      sym__attribute,
    STATE(6433), 1,
      sym_class_type_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228397] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    ACTIONS(8381), 1,
      sym__lowercase_identifier,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5771), 1,
      sym_attribute,
    STATE(6267), 1,
      sym__attribute,
    STATE(7103), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228426] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8583), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5772), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228455] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3520), 1,
      anon_sym_PIPE,
    STATE(5773), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3518), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [228476] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3626), 1,
      anon_sym_PIPE,
    STATE(5774), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3624), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [228497] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8428), 1,
      anon_sym_PIPE,
    STATE(5775), 1,
      sym_attribute,
    STATE(5776), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5642), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [228520] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8585), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5776), 2,
      sym_attribute,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5650), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [228541] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8588), 1,
      anon_sym_COLON2,
    STATE(5777), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4971), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(5062), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [228564] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8590), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5778), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228593] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8592), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5779), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228622] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3482), 1,
      anon_sym_PIPE,
    STATE(5780), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3480), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [228643] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8539), 1,
      anon_sym_AMP,
    STATE(5781), 1,
      sym_attribute,
    STATE(5931), 1,
      aux_sym_tag_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8594), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [228668] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8082), 1,
      anon_sym_COMMA,
    ACTIONS(8084), 1,
      anon_sym_PIPE,
    ACTIONS(8088), 1,
      anon_sym_as,
    ACTIONS(8090), 1,
      anon_sym_COLON_COLON,
    STATE(5782), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8369), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [228695] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2764), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5498), 1,
      anon_sym_and,
    STATE(5554), 1,
      aux_sym_expression_item_repeat1,
    STATE(5783), 1,
      sym_attribute,
    STATE(6240), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5496), 2,
      anon_sym_in,
      sym_let_and_operator,
  [228722] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4060), 1,
      anon_sym_PIPE,
    STATE(5784), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4062), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [228743] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8596), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5785), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228772] = 9,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    ACTIONS(8598), 1,
      anon_sym_RBRACE,
    STATE(5342), 1,
      sym_field_path,
    STATE(5786), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228801] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5216), 1,
      anon_sym_DQUOTE,
    ACTIONS(8600), 1,
      anon_sym_LBRACE,
    STATE(3945), 1,
      aux_sym_external_repeat1,
    STATE(5787), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4388), 2,
      sym_string,
      sym_quoted_string,
  [228825] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5788), 1,
      sym_attribute,
    STATE(5812), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6182), 2,
      anon_sym_and,
      anon_sym_in,
  [228849] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5789), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5747), 2,
      anon_sym_and,
      anon_sym_in,
  [228873] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5790), 1,
      sym_attribute,
    STATE(5813), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6088), 2,
      anon_sym_and,
      anon_sym_in,
  [228897] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8106), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    ACTIONS(8208), 1,
      anon_sym_RPAREN,
    STATE(5791), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228923] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5792), 1,
      sym_attribute,
    STATE(5814), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5872), 2,
      anon_sym_and,
      anon_sym_in,
  [228947] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5793), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6046), 2,
      anon_sym_and,
      anon_sym_in,
  [228971] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8604), 1,
      anon_sym_SEMI,
    ACTIONS(8606), 1,
      anon_sym_GT_RBRACE,
    STATE(5794), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6637), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228997] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5795), 1,
      sym_attribute,
    STATE(5816), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6064), 2,
      anon_sym_and,
      anon_sym_in,
  [229021] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5796), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5850), 2,
      anon_sym_and,
      anon_sym_in,
  [229045] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5797), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4900), 2,
      anon_sym_and,
      anon_sym_in,
  [229069] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8521), 1,
      sym__lowercase_identifier,
    STATE(5798), 1,
      sym_attribute,
    STATE(6227), 1,
      aux_sym_instance_variable_specification_repeat1,
    STATE(6808), 1,
      sym__instance_variable_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8525), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [229093] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5799), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5681), 2,
      anon_sym_and,
      anon_sym_in,
  [229117] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5800), 1,
      sym_attribute,
    STATE(5842), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6178), 2,
      anon_sym_and,
      anon_sym_in,
  [229141] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8150), 1,
      anon_sym_RPAREN,
    STATE(5801), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229167] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5802), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5800), 2,
      anon_sym_and,
      anon_sym_in,
  [229191] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5803), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5886), 2,
      anon_sym_and,
      anon_sym_in,
  [229215] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5804), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5762), 2,
      anon_sym_and,
      anon_sym_in,
  [229239] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5805), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2409), 2,
      anon_sym_and,
      anon_sym_in,
  [229263] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4915), 1,
      sym__lowercase_identifier,
    ACTIONS(8608), 1,
      anon_sym_POUND,
    STATE(3857), 1,
      sym_type_constructor_path,
    STATE(5806), 1,
      sym_attribute,
    STATE(6932), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229289] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8610), 1,
      anon_sym_SEMI,
    ACTIONS(8612), 1,
      anon_sym_GT_RBRACE,
    STATE(5807), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6451), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229315] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5808), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6174), 2,
      anon_sym_and,
      anon_sym_in,
  [229339] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5809), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5677), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [229357] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5810), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5952), 2,
      anon_sym_and,
      anon_sym_in,
  [229381] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5811), 1,
      sym_attribute,
    STATE(5827), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5731), 2,
      anon_sym_and,
      anon_sym_in,
  [229405] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5812), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5788), 2,
      anon_sym_and,
      anon_sym_in,
  [229429] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5813), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6108), 2,
      anon_sym_and,
      anon_sym_in,
  [229453] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5814), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5902), 2,
      anon_sym_and,
      anon_sym_in,
  [229477] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5815), 1,
      sym_attribute,
    STATE(5828), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6092), 2,
      anon_sym_and,
      anon_sym_in,
  [229501] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5816), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6198), 2,
      anon_sym_and,
      anon_sym_in,
  [229525] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8106), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    ACTIONS(8315), 1,
      anon_sym_RPAREN,
    STATE(5817), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229551] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5818), 1,
      sym_attribute,
    STATE(5829), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5695), 2,
      anon_sym_and,
      anon_sym_in,
  [229575] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5819), 1,
      sym_attribute,
    STATE(5830), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5846), 2,
      anon_sym_and,
      anon_sym_in,
  [229599] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8614), 1,
      anon_sym_SEMI,
    ACTIONS(8616), 1,
      anon_sym_GT_RBRACE,
    STATE(5820), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6673), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229625] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8539), 1,
      anon_sym_AMP,
    STATE(5821), 1,
      sym_attribute,
    STATE(5823), 1,
      aux_sym_tag_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8618), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [229647] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    STATE(5822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8620), 4,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_AMP,
  [229667] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8622), 1,
      anon_sym_AMP,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5823), 2,
      sym_attribute,
      aux_sym_tag_specification_repeat1,
    ACTIONS(8620), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [229687] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5824), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5360), 2,
      anon_sym_and,
      anon_sym_in,
  [229711] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5825), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5808), 2,
      anon_sym_and,
      anon_sym_in,
  [229735] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5826), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5364), 2,
      anon_sym_and,
      anon_sym_in,
  [229759] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5827), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6150), 2,
      anon_sym_and,
      anon_sym_in,
  [229783] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5828), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5774), 2,
      anon_sym_and,
      anon_sym_in,
  [229807] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5829), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6194), 2,
      anon_sym_and,
      anon_sym_in,
  [229831] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5830), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5691), 2,
      anon_sym_and,
      anon_sym_in,
  [229855] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5831), 1,
      sym_attribute,
    STATE(5833), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5938), 2,
      anon_sym_and,
      anon_sym_in,
  [229879] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7796), 1,
      anon_sym_RPAREN,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    STATE(5832), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229905] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5833), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5842), 2,
      anon_sym_and,
      anon_sym_in,
  [229929] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7986), 1,
      anon_sym_COLON_GT,
    STATE(4953), 1,
      sym__coerced,
    STATE(5834), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7402), 3,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [229951] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    STATE(4960), 1,
      sym__coerced,
    STATE(5835), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7468), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229977] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      sym_field_path,
    STATE(5836), 1,
      sym_attribute,
    STATE(6321), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230003] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5837), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5858), 2,
      anon_sym_and,
      anon_sym_in,
  [230027] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3949), 1,
      sym__uppercase_identifier,
    ACTIONS(8625), 1,
      anon_sym_LPAREN,
    STATE(5180), 1,
      sym__constructor_name,
    STATE(5838), 1,
      sym_attribute,
    STATE(5904), 1,
      sym_constructor_path,
    STATE(7983), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230053] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5839), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6866), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230079] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4331), 1,
      anon_sym_LPAREN,
    ACTIONS(8627), 1,
      sym__lowercase_identifier,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    STATE(4801), 1,
      sym__value_name,
    STATE(4802), 1,
      sym_parenthesized_operator,
    STATE(5840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230105] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    STATE(4960), 1,
      sym__coerced,
    STATE(5841), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(8082), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230131] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5842), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5898), 2,
      anon_sym_and,
      anon_sym_in,
  [230155] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8631), 1,
      anon_sym_SEMI,
    ACTIONS(8633), 1,
      anon_sym_GT_RBRACE,
    STATE(5843), 1,
      sym_attribute,
    STATE(6341), 1,
      sym_instance_variable_expression,
    STATE(6343), 1,
      sym__instance_variable_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230181] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(7808), 1,
      anon_sym_RPAREN,
    STATE(5844), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230207] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8635), 1,
      anon_sym_SEMI,
    ACTIONS(8637), 1,
      anon_sym_GT_RBRACE,
    STATE(5845), 1,
      sym_attribute,
    STATE(6342), 1,
      sym_instance_variable_expression,
    STATE(6343), 1,
      sym__instance_variable_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230233] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8639), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5200), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(5846), 2,
      sym_attribute,
      aux_sym_expression_item_repeat1,
  [230255] = 7,
    ACTIONS(89), 1,
      anon_sym_DQUOTE,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8642), 1,
      anon_sym_LBRACE,
    STATE(4274), 1,
      aux_sym_external_repeat1,
    STATE(5847), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4388), 2,
      sym_string,
      sym_quoted_string,
  [230279] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8644), 1,
      anon_sym_SEMI,
    ACTIONS(8646), 1,
      anon_sym_GT_RBRACE,
    STATE(5848), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6758), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230305] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8106), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    ACTIONS(8332), 1,
      anon_sym_RPAREN,
    STATE(5849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230331] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5850), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6060), 2,
      anon_sym_and,
      anon_sym_in,
  [230355] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5851), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6060), 2,
      anon_sym_and,
      anon_sym_in,
  [230379] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    STATE(4960), 1,
      sym__coerced,
    STATE(5852), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(8217), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230405] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5216), 1,
      anon_sym_DQUOTE,
    ACTIONS(8600), 1,
      anon_sym_LBRACE,
    STATE(3946), 1,
      aux_sym_external_repeat1,
    STATE(5853), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4388), 2,
      sym_string,
      sym_quoted_string,
  [230429] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      sym_field_path,
    STATE(5854), 1,
      sym_attribute,
    STATE(6393), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230455] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5855), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5423), 2,
      anon_sym_and,
      anon_sym_in,
  [230479] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5856), 1,
      sym_attribute,
    STATE(5912), 1,
      sym_field_path,
    STATE(6455), 1,
      sym_field_binding_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230505] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(5857), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8648), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [230523] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(7290), 1,
      anon_sym_LBRACE,
    STATE(5322), 1,
      aux_sym_external_repeat1,
    STATE(5858), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6187), 2,
      sym_string,
      sym_quoted_string,
  [230547] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5859), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6076), 2,
      anon_sym_and,
      anon_sym_in,
  [230571] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    STATE(4960), 1,
      sym__coerced,
    STATE(5860), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7796), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230597] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5861), 1,
      sym_attribute,
    STATE(5949), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6134), 2,
      anon_sym_and,
      anon_sym_in,
  [230621] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(5862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8650), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [230639] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8521), 1,
      sym__lowercase_identifier,
    STATE(4803), 1,
      sym__instance_variable_name,
    STATE(5863), 1,
      sym_attribute,
    STATE(5972), 1,
      aux_sym_instance_variable_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8525), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [230663] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      sym_field_path,
    STATE(5864), 1,
      sym_attribute,
    STATE(6680), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230689] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5865), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6170), 2,
      anon_sym_and,
      anon_sym_in,
  [230713] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    STATE(4960), 1,
      sym__coerced,
    STATE(5866), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7632), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230739] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      sym_field_path,
    STATE(5867), 1,
      sym_attribute,
    STATE(6370), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230765] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(5868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8652), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [230783] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5869), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5784), 2,
      anon_sym_and,
      anon_sym_in,
  [230807] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5870), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6482), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230833] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    STATE(4960), 1,
      sym__coerced,
    STATE(5871), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(8174), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230859] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5872), 1,
      sym_attribute,
    STATE(5952), 1,
      sym_field_path,
    STATE(6646), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230885] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8521), 1,
      sym__lowercase_identifier,
    STATE(4803), 1,
      sym__instance_variable_name,
    STATE(5873), 1,
      sym_attribute,
    STATE(6227), 1,
      aux_sym_instance_variable_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8525), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [230909] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5874), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6084), 2,
      anon_sym_and,
      anon_sym_in,
  [230933] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      sym_field_path,
    STATE(5875), 1,
      sym_attribute,
    STATE(6420), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230959] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5876), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2391), 2,
      anon_sym_and,
      anon_sym_in,
  [230983] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8654), 1,
      anon_sym_SEMI,
    ACTIONS(8656), 1,
      anon_sym_GT_RBRACE,
    STATE(5877), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6401), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231009] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8658), 1,
      anon_sym_type,
    ACTIONS(8660), 1,
      anon_sym_module,
    STATE(5878), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3965), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [231031] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5879), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6100), 2,
      anon_sym_and,
      anon_sym_in,
  [231055] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      sym_field_path,
    STATE(5880), 1,
      sym_attribute,
    STATE(6474), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231081] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5881), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6104), 2,
      anon_sym_and,
      anon_sym_in,
  [231105] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5882), 1,
      sym_attribute,
    STATE(5965), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6112), 2,
      anon_sym_and,
      anon_sym_in,
  [231129] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      sym_field_path,
    STATE(5883), 1,
      sym_attribute,
    STATE(6450), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231155] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      sym_field_path,
    STATE(5884), 1,
      sym_attribute,
    STATE(6833), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231181] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1807), 1,
      anon_sym_DQUOTE,
    ACTIONS(7290), 1,
      anon_sym_LBRACE,
    STATE(5255), 1,
      aux_sym_external_repeat1,
    STATE(5885), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6187), 2,
      sym_string,
      sym_quoted_string,
  [231205] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      sym_field_path,
    STATE(5886), 1,
      sym_attribute,
    STATE(6536), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231231] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5887), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6158), 2,
      anon_sym_and,
      anon_sym_in,
  [231255] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5646), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [231273] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5889), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6022), 2,
      anon_sym_and,
      anon_sym_in,
  [231297] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5890), 1,
      sym_attribute,
    STATE(5912), 1,
      sym_field_path,
    STATE(6666), 1,
      sym_field_binding_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231323] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5891), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5650), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [231341] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3965), 1,
      sym__uppercase_identifier,
    ACTIONS(8662), 1,
      anon_sym_LPAREN,
    STATE(3688), 1,
      sym__constructor_name,
    STATE(4248), 1,
      sym_constructor_path,
    STATE(5892), 1,
      sym_attribute,
    STATE(7869), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231367] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      sym_field_path,
    STATE(5893), 1,
      sym_attribute,
    STATE(6610), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231393] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8664), 1,
      anon_sym_type,
    ACTIONS(8666), 1,
      anon_sym_module,
    STATE(5894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5390), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [231415] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(8670), 1,
      anon_sym_PIPE,
    ACTIONS(8672), 1,
      anon_sym_RBRACK,
    STATE(5895), 1,
      sym_attribute,
    STATE(6602), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231441] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8658), 1,
      anon_sym_type,
    ACTIONS(8660), 1,
      anon_sym_module,
    STATE(5896), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3991), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [231463] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5897), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5298), 2,
      anon_sym_and,
      anon_sym_in,
  [231487] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5898), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5948), 2,
      anon_sym_and,
      anon_sym_in,
  [231511] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      sym_field_path,
    STATE(5899), 1,
      sym_attribute,
    STATE(6648), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231537] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5900), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5948), 2,
      anon_sym_and,
      anon_sym_in,
  [231561] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    STATE(4960), 1,
      sym__coerced,
    STATE(5901), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7223), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231587] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      sym_field_path,
    STATE(5902), 1,
      sym_attribute,
    STATE(6726), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231613] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5903), 1,
      sym_attribute,
    STATE(5976), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5743), 2,
      anon_sym_and,
      anon_sym_in,
  [231637] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5904), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4758), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [231655] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5905), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5962), 2,
      anon_sym_and,
      anon_sym_in,
  [231679] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      sym_field_path,
    STATE(5906), 1,
      sym_attribute,
    STATE(6505), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231705] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      sym_field_path,
    STATE(5907), 1,
      sym_attribute,
    STATE(6311), 1,
      sym_field_expression,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231731] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5908), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4762), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [231749] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5859), 1,
      aux_sym_expression_item_repeat1,
    STATE(5909), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6056), 2,
      anon_sym_and,
      anon_sym_in,
  [231773] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5910), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6080), 2,
      anon_sym_and,
      anon_sym_in,
  [231797] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5911), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6030), 2,
      anon_sym_and,
      anon_sym_in,
  [231821] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8674), 1,
      anon_sym_EQ,
    ACTIONS(8676), 1,
      anon_sym_COLON2,
    STATE(5912), 1,
      sym_attribute,
    STATE(6326), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8678), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [231845] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8680), 1,
      anon_sym_as,
    STATE(5913), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7334), 4,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
  [231865] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8682), 1,
      anon_sym_SEMI,
    ACTIONS(8684), 1,
      anon_sym_GT_RBRACE,
    STATE(5914), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6452), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231891] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5915), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2385), 2,
      anon_sym_and,
      anon_sym_in,
  [231915] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5916), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5986), 2,
      anon_sym_and,
      anon_sym_in,
  [231939] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4921), 1,
      sym__lowercase_identifier,
    ACTIONS(8686), 1,
      anon_sym_POUND,
    STATE(3512), 1,
      sym_type_constructor_path,
    STATE(5917), 1,
      sym_attribute,
    STATE(6926), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231965] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5879), 1,
      aux_sym_expression_item_repeat1,
    STATE(5918), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5998), 2,
      anon_sym_and,
      anon_sym_in,
  [231989] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5912), 1,
      sym_field_path,
    STATE(5919), 1,
      sym_attribute,
    STATE(7089), 1,
      sym_field_binding_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232015] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5920), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5758), 2,
      anon_sym_and,
      anon_sym_in,
  [232039] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8664), 1,
      anon_sym_type,
    ACTIONS(8666), 1,
      anon_sym_module,
    STATE(5921), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5471), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [232061] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5881), 1,
      aux_sym_expression_item_repeat1,
    STATE(5922), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6038), 2,
      anon_sym_and,
      anon_sym_in,
  [232085] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5923), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5882), 2,
      anon_sym_and,
      anon_sym_in,
  [232109] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5924), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6068), 2,
      anon_sym_and,
      anon_sym_in,
  [232133] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5673), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [232151] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5926), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5400), 2,
      anon_sym_and,
      anon_sym_in,
  [232175] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8688), 1,
      anon_sym_type,
    ACTIONS(8690), 1,
      anon_sym_module,
    STATE(5927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5694), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [232197] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8692), 1,
      anon_sym_EQ,
    STATE(5928), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232223] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    STATE(4960), 1,
      sym__coerced,
    STATE(5929), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7595), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232249] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5930), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5910), 2,
      anon_sym_and,
      anon_sym_in,
  [232273] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8539), 1,
      anon_sym_AMP,
    STATE(5823), 1,
      aux_sym_tag_specification_repeat1,
    STATE(5931), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8537), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [232295] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5932), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2419), 2,
      anon_sym_and,
      anon_sym_in,
  [232319] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5933), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5914), 2,
      anon_sym_and,
      anon_sym_in,
  [232343] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5934), 1,
      sym_attribute,
    STATE(5989), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5918), 2,
      anon_sym_and,
      anon_sym_in,
  [232367] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5935), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6166), 2,
      anon_sym_and,
      anon_sym_in,
  [232391] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8694), 1,
      anon_sym_type,
    ACTIONS(8696), 1,
      anon_sym_module,
    STATE(5936), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3965), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [232413] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5937), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6124), 2,
      anon_sym_and,
      anon_sym_in,
  [232437] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1971), 1,
      anon_sym_POUND,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    STATE(5938), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1973), 3,
      anon_sym_COMMA,
      anon_sym_with,
      sym_hash_operator,
  [232459] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5912), 1,
      sym_field_path,
    STATE(5939), 1,
      sym_attribute,
    STATE(6527), 1,
      sym_field_binding_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232485] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(8670), 1,
      anon_sym_PIPE,
    ACTIONS(8700), 1,
      anon_sym_RBRACK,
    STATE(5940), 1,
      sym_attribute,
    STATE(6391), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232511] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5941), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5408), 2,
      anon_sym_and,
      anon_sym_in,
  [232535] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8702), 1,
      anon_sym_SEMI,
    ACTIONS(8704), 1,
      anon_sym_GT_RBRACE,
    STATE(5942), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6516), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232561] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5943), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6146), 2,
      anon_sym_and,
      anon_sym_in,
  [232585] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5944), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5766), 2,
      anon_sym_and,
      anon_sym_in,
  [232609] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5887), 1,
      aux_sym_expression_item_repeat1,
    STATE(5945), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5770), 2,
      anon_sym_and,
      anon_sym_in,
  [232633] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5912), 1,
      sym_field_path,
    STATE(5946), 1,
      sym_attribute,
    STATE(6741), 1,
      sym_field_binding_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232659] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(5947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8706), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym__,
      aux_sym_type_variable_token1,
  [232677] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5948), 1,
      sym_attribute,
    STATE(6001), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5978), 2,
      anon_sym_and,
      anon_sym_in,
  [232701] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5949), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5982), 2,
      anon_sym_and,
      anon_sym_in,
  [232725] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8708), 1,
      sym__lowercase_identifier,
    ACTIONS(8712), 1,
      anon_sym_BANG,
    STATE(5950), 1,
      sym_attribute,
    STATE(6122), 1,
      aux_sym_method_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8710), 2,
      anon_sym_private,
      anon_sym_virtual,
  [232749] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5951), 1,
      sym_attribute,
    STATE(6011), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5990), 2,
      anon_sym_and,
      anon_sym_in,
  [232773] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8676), 1,
      anon_sym_COLON2,
    ACTIONS(8714), 1,
      anon_sym_EQ,
    STATE(5952), 1,
      sym_attribute,
    STATE(6359), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8716), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [232797] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8688), 1,
      anon_sym_type,
    ACTIONS(8690), 1,
      anon_sym_module,
    STATE(5953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3965), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [232819] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5911), 1,
      aux_sym_expression_item_repeat1,
    STATE(5954), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5994), 2,
      anon_sym_and,
      anon_sym_in,
  [232843] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(8670), 1,
      anon_sym_PIPE,
    ACTIONS(8718), 1,
      anon_sym_RBRACK,
    STATE(5955), 1,
      sym_attribute,
    STATE(6521), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232869] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5956), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6002), 2,
      anon_sym_and,
      anon_sym_in,
  [232893] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    STATE(4960), 1,
      sym__coerced,
    STATE(5957), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(8204), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232919] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8521), 1,
      sym__lowercase_identifier,
    STATE(5798), 1,
      aux_sym_instance_variable_specification_repeat1,
    STATE(5958), 1,
      sym_attribute,
    STATE(6761), 1,
      sym__instance_variable_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8525), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [232943] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5789), 1,
      aux_sym_expression_item_repeat1,
    STATE(5959), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6006), 2,
      anon_sym_and,
      anon_sym_in,
  [232967] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5960), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6018), 2,
      anon_sym_and,
      anon_sym_in,
  [232991] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5961), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6026), 2,
      anon_sym_and,
      anon_sym_in,
  [233015] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    ACTIONS(8720), 1,
      anon_sym_POUND,
    STATE(3727), 1,
      sym_type_constructor_path,
    STATE(5962), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233041] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(8034), 1,
      anon_sym_COLON2,
    STATE(4960), 1,
      sym__coerced,
    STATE(5963), 1,
      sym_attribute,
    STATE(6598), 1,
      sym__typed,
    STATE(7270), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233067] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5964), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2403), 2,
      anon_sym_and,
      anon_sym_in,
  [233091] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5965), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6042), 2,
      anon_sym_and,
      anon_sym_in,
  [233115] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8104), 1,
      anon_sym_RPAREN,
    ACTIONS(8106), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    STATE(5966), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233141] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8722), 1,
      anon_sym_DASH_GT,
    ACTIONS(8724), 1,
      anon_sym_with,
    STATE(5967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5492), 3,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [233163] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8722), 1,
      anon_sym_DASH_GT,
    ACTIONS(8724), 1,
      anon_sym_with,
    STATE(5968), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5504), 3,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [233185] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8722), 1,
      anon_sym_DASH_GT,
    ACTIONS(8724), 1,
      anon_sym_with,
    STATE(5969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5447), 3,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [233207] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5970), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5306), 2,
      anon_sym_and,
      anon_sym_in,
  [233231] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5952), 1,
      sym_field_path,
    STATE(5971), 1,
      sym_attribute,
    STATE(6753), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233257] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8521), 1,
      sym__lowercase_identifier,
    STATE(4791), 1,
      sym__instance_variable_name,
    STATE(5972), 1,
      sym_attribute,
    STATE(6227), 1,
      aux_sym_instance_variable_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8525), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [233281] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5973), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5834), 2,
      anon_sym_and,
      anon_sym_in,
  [233305] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5802), 1,
      aux_sym_expression_item_repeat1,
    STATE(5974), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5854), 2,
      anon_sym_and,
      anon_sym_in,
  [233329] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5803), 1,
      aux_sym_expression_item_repeat1,
    STATE(5975), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5890), 2,
      anon_sym_and,
      anon_sym_in,
  [233353] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5976), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5944), 2,
      anon_sym_and,
      anon_sym_in,
  [233377] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5977), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6014), 2,
      anon_sym_and,
      anon_sym_in,
  [233401] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5978), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6138), 2,
      anon_sym_and,
      anon_sym_in,
  [233425] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5979), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5427), 2,
      anon_sym_and,
      anon_sym_in,
  [233449] = 7,
    ACTIONS(89), 1,
      anon_sym_DQUOTE,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8642), 1,
      anon_sym_LBRACE,
    STATE(4310), 1,
      aux_sym_external_repeat1,
    STATE(5980), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4388), 2,
      sym_string,
      sym_quoted_string,
  [233473] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5981), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5816), 2,
      anon_sym_and,
      anon_sym_in,
  [233497] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8726), 1,
      anon_sym_SEMI,
    ACTIONS(8728), 1,
      anon_sym_GT_RBRACE,
    STATE(5982), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6631), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233523] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8730), 1,
      anon_sym_SEMI,
    ACTIONS(8732), 1,
      anon_sym_GT_RBRACE,
    STATE(5983), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6733), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233549] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5910), 1,
      aux_sym_expression_item_repeat1,
    STATE(5984), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5894), 2,
      anon_sym_and,
      anon_sym_in,
  [233573] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5916), 1,
      aux_sym_expression_item_repeat1,
    STATE(5985), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6010), 2,
      anon_sym_and,
      anon_sym_in,
  [233597] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5986), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5862), 2,
      anon_sym_and,
      anon_sym_in,
  [233621] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5987), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6154), 2,
      anon_sym_and,
      anon_sym_in,
  [233645] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5988), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5723), 2,
      anon_sym_and,
      anon_sym_in,
  [233669] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5989), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5812), 2,
      anon_sym_and,
      anon_sym_in,
  [233693] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5943), 1,
      aux_sym_expression_item_repeat1,
    STATE(5990), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6096), 2,
      anon_sym_and,
      anon_sym_in,
  [233717] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8734), 1,
      anon_sym_SEMI,
    ACTIONS(8736), 1,
      anon_sym_GT_RBRACE,
    STATE(5991), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6568), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233743] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5944), 1,
      aux_sym_expression_item_repeat1,
    STATE(5992), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6162), 2,
      anon_sym_and,
      anon_sym_in,
  [233767] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8738), 1,
      anon_sym_EQ,
    STATE(5993), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233793] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5994), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5727), 2,
      anon_sym_and,
      anon_sym_in,
  [233817] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8740), 1,
      anon_sym_SEMI,
    ACTIONS(8742), 1,
      anon_sym_GT_RBRACE,
    STATE(5995), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6639), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233843] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8694), 1,
      anon_sym_type,
    ACTIONS(8696), 1,
      anon_sym_module,
    STATE(5996), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4481), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [233865] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8722), 1,
      anon_sym_DASH_GT,
    ACTIONS(8724), 1,
      anon_sym_with,
    STATE(5997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5455), 3,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [233887] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(5998), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5294), 2,
      anon_sym_and,
      anon_sym_in,
  [233911] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8744), 1,
      anon_sym_EQ,
    STATE(5999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233937] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8746), 1,
      anon_sym_EQ,
    STATE(6000), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233963] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6001), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6186), 2,
      anon_sym_and,
      anon_sym_in,
  [233987] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5810), 1,
      aux_sym_expression_item_repeat1,
    STATE(6002), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6190), 2,
      anon_sym_and,
      anon_sym_in,
  [234011] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8748), 1,
      anon_sym_EQ,
    STATE(6003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234037] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8750), 1,
      anon_sym_EQ,
    STATE(6004), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234063] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8752), 1,
      anon_sym_EQ,
    STATE(6005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234089] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8754), 1,
      anon_sym_EQ,
    STATE(6006), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234115] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3098), 1,
      sym__uppercase_identifier,
    ACTIONS(8662), 1,
      anon_sym_LPAREN,
    STATE(2122), 1,
      sym__constructor_name,
    STATE(4248), 1,
      sym_constructor_path,
    STATE(6007), 1,
      sym_attribute,
    STATE(7920), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234141] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6008), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8756), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [234159] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6009), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5386), 2,
      anon_sym_and,
      anon_sym_in,
  [234183] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    ACTIONS(8328), 1,
      anon_sym_RPAREN,
    STATE(6010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234209] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6011), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5906), 2,
      anon_sym_and,
      anon_sym_in,
  [234233] = 8,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(7950), 1,
      sym__lowercase_identifier,
    STATE(5952), 1,
      sym_field_path,
    STATE(6012), 1,
      sym_attribute,
    STATE(6306), 1,
      sym_field_pattern,
    STATE(7592), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234259] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8758), 1,
      anon_sym_GT_RBRACE,
    STATE(6013), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234282] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5550), 1,
      anon_sym_and,
    STATE(6014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5548), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [234301] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8760), 1,
      anon_sym_GT_RBRACE,
    STATE(6015), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234324] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6202), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6016), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234347] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8762), 1,
      anon_sym_DOT,
    STATE(6017), 1,
      sym_attribute,
    STATE(6100), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(7117), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234370] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6242), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6018), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234393] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8764), 1,
      sym__lowercase_identifier,
    STATE(3519), 1,
      sym_class_type_path,
    STATE(6019), 1,
      sym_attribute,
    STATE(7132), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234416] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8766), 1,
      sym__lowercase_identifier,
    STATE(3653), 1,
      sym_class_type_path,
    STATE(6020), 1,
      sym_attribute,
    STATE(7182), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234439] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8768), 1,
      sym__lowercase_identifier,
    ACTIONS(8770), 1,
      anon_sym_LPAREN,
    STATE(6021), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5715), 2,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [234460] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(8772), 1,
      anon_sym_RBRACK,
    STATE(6022), 1,
      sym_attribute,
    STATE(6445), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234483] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8776), 1,
      anon_sym_RBRACK,
    ACTIONS(8778), 1,
      anon_sym_GT,
    STATE(6023), 1,
      sym_attribute,
    STATE(6070), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234506] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8776), 1,
      anon_sym_RBRACK,
    ACTIONS(8778), 1,
      anon_sym_GT,
    STATE(6024), 1,
      sym_attribute,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234529] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6025), 1,
      sym_attribute,
    STATE(7058), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8780), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [234550] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    ACTIONS(8782), 1,
      anon_sym_with,
    ACTIONS(8784), 1,
      sym_hash_operator,
    STATE(6026), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234573] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8420), 1,
      sym__lowercase_identifier,
    STATE(2572), 1,
      sym_class_path,
    STATE(6027), 1,
      sym_attribute,
    STATE(7201), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234596] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8786), 1,
      sym__lowercase_identifier,
    STATE(2900), 1,
      sym_class_path,
    STATE(6028), 1,
      sym_attribute,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234619] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6288), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6029), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234642] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6242), 1,
      anon_sym_in,
    STATE(6030), 1,
      sym_attribute,
    STATE(6045), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234665] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8788), 1,
      anon_sym_DOT,
    STATE(6031), 1,
      sym_attribute,
    STATE(6100), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(7117), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234688] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8790), 1,
      anon_sym_GT_RBRACE,
    STATE(6032), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234711] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8792), 1,
      aux_sym_attribute_id_token1,
    STATE(6033), 1,
      sym_attribute,
    STATE(6214), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3168), 2,
      sym__left_quoted_string_delimiter,
      aux_sym_quoted_extension_token1,
  [234732] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8794), 1,
      sym__lowercase_identifier,
    STATE(3030), 1,
      sym_type_constructor_path,
    STATE(6034), 1,
      sym_attribute,
    STATE(6915), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234755] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6320), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6035), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234778] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6369), 1,
      anon_sym_in,
    ACTIONS(8796), 1,
      anon_sym_and,
    ACTIONS(8798), 1,
      sym_let_and_operator,
    STATE(6036), 1,
      sym_attribute,
    STATE(6272), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234801] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8800), 1,
      anon_sym_GT_RBRACE,
    STATE(6037), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234824] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8802), 1,
      anon_sym_GT_RBRACE,
    STATE(6038), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234847] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8804), 1,
      anon_sym_RBRACK,
    ACTIONS(8806), 1,
      anon_sym_GT,
    STATE(6039), 1,
      sym_attribute,
    STATE(6178), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234870] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8808), 1,
      anon_sym_GT_RBRACE,
    STATE(6040), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234893] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8249), 1,
      sym__lowercase_identifier,
    ACTIONS(8251), 1,
      anon_sym_LBRACK,
    ACTIONS(8253), 1,
      anon_sym_virtual,
    STATE(4719), 1,
      sym_class_binding,
    STATE(6041), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234916] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8810), 1,
      anon_sym_rec,
    STATE(4551), 1,
      sym_module_binding,
    STATE(6042), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8144), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [234937] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8812), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3188), 2,
      sym__left_quoted_string_delimiter,
      aux_sym_quoted_extension_token1,
    STATE(6043), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
  [234956] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6260), 1,
      anon_sym_in,
    ACTIONS(8796), 1,
      anon_sym_and,
    ACTIONS(8798), 1,
      sym_let_and_operator,
    STATE(6044), 1,
      sym_attribute,
    STATE(6126), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234979] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6336), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6045), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235002] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6387), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6046), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235025] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8815), 1,
      sym__lowercase_identifier,
    ACTIONS(8817), 1,
      anon_sym_RBRACE,
    ACTIONS(8819), 1,
      anon_sym_mutable,
    STATE(6047), 1,
      sym_attribute,
    STATE(7130), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235048] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4214), 1,
      anon_sym_RBRACK,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8821), 1,
      anon_sym_GT,
    STATE(6048), 1,
      sym_attribute,
    STATE(6049), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235071] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8804), 1,
      anon_sym_RBRACK,
    ACTIONS(8806), 1,
      anon_sym_GT,
    STATE(6049), 1,
      sym_attribute,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235094] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3526), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    STATE(6050), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235117] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8823), 1,
      anon_sym_GT_RBRACE,
    STATE(6051), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235140] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8825), 1,
      sym__lowercase_identifier,
    ACTIONS(8827), 1,
      anon_sym_LPAREN,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(2289), 1,
      sym__value_name,
    STATE(6052), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235163] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8553), 1,
      sym__lowercase_identifier,
    STATE(2461), 1,
      sym_class_path,
    STATE(6053), 1,
      sym_attribute,
    STATE(7636), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235186] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8829), 1,
      anon_sym_DOT,
    STATE(6054), 1,
      sym_attribute,
    STATE(6100), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(7117), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235209] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8680), 1,
      anon_sym_as,
    STATE(6055), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7462), 3,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [235228] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8831), 1,
      sym__lowercase_identifier,
    STATE(4110), 1,
      sym_class_type_path,
    STATE(6056), 1,
      sym_attribute,
    STATE(6859), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235251] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8551), 1,
      sym__lowercase_identifier,
    STATE(2701), 1,
      sym_class_path,
    STATE(6057), 1,
      sym_attribute,
    STATE(7549), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235274] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(8672), 1,
      anon_sym_RBRACK,
    STATE(6058), 1,
      sym_attribute,
    STATE(6602), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235297] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6059), 1,
      sym_attribute,
    STATE(6987), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8833), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [235318] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3544), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    STATE(6060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235341] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3500), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    STATE(6061), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235364] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6206), 1,
      anon_sym_in,
    STATE(6062), 1,
      sym_attribute,
    STATE(6278), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235387] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8835), 1,
      sym__lowercase_identifier,
    STATE(6063), 1,
      sym_attribute,
    STATE(6249), 1,
      sym_class_type_path,
    STATE(6897), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235410] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8837), 1,
      sym__lowercase_identifier,
    STATE(6064), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(7775), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235433] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    ACTIONS(8784), 1,
      sym_hash_operator,
    ACTIONS(8839), 1,
      anon_sym_with,
    STATE(6065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235456] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    ACTIONS(8841), 1,
      sym__lowercase_identifier,
    STATE(6066), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5144), 2,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [235477] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(8843), 1,
      anon_sym_RBRACK,
    STATE(6067), 1,
      sym_attribute,
    STATE(6124), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235500] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8381), 1,
      sym__lowercase_identifier,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(6068), 1,
      sym_attribute,
    STATE(7022), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235523] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8845), 1,
      anon_sym_DOT,
    STATE(6069), 1,
      sym_attribute,
    STATE(6100), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(7117), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235546] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8847), 1,
      anon_sym_RBRACK,
    ACTIONS(8849), 1,
      anon_sym_GT,
    STATE(6070), 1,
      sym_attribute,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235569] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(8847), 1,
      anon_sym_RBRACK,
    STATE(6071), 1,
      sym_attribute,
    STATE(6124), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235592] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    ACTIONS(8851), 1,
      sym__lowercase_identifier,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5325), 1,
      sym__value_name,
    STATE(6072), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235615] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7494), 1,
      anon_sym_COMMA,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    STATE(6073), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235638] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7498), 1,
      anon_sym_COMMA,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    STATE(6074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235661] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8786), 1,
      sym__lowercase_identifier,
    STATE(2892), 1,
      sym_class_path,
    STATE(6075), 1,
      sym_attribute,
    STATE(7340), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235684] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8853), 1,
      anon_sym_DOT,
    STATE(6076), 1,
      sym_attribute,
    STATE(6100), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(7117), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235707] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8855), 1,
      sym__lowercase_identifier,
    STATE(3892), 1,
      sym_class_type_path,
    STATE(6077), 1,
      sym_attribute,
    STATE(6912), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235730] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6403), 1,
      anon_sym_in,
    STATE(6018), 1,
      aux_sym_expression_item_repeat1,
    STATE(6078), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235753] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7514), 1,
      anon_sym_COMMA,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    STATE(6079), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235776] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8815), 1,
      sym__lowercase_identifier,
    ACTIONS(8819), 1,
      anon_sym_mutable,
    ACTIONS(8857), 1,
      anon_sym_RBRACE,
    STATE(6080), 1,
      sym_attribute,
    STATE(7130), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235799] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8859), 1,
      sym__lowercase_identifier,
    STATE(2275), 1,
      sym_type_constructor_path,
    STATE(6081), 1,
      sym_attribute,
    STATE(6939), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235822] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6082), 1,
      sym_attribute,
    STATE(6893), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8861), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [235843] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8863), 1,
      anon_sym_DOT,
    STATE(6083), 1,
      sym_attribute,
    STATE(6100), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(7117), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235866] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8381), 1,
      sym__lowercase_identifier,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(6084), 1,
      sym_attribute,
    STATE(7033), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235889] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8825), 1,
      sym__lowercase_identifier,
    ACTIONS(8827), 1,
      anon_sym_LPAREN,
    STATE(2272), 1,
      sym_parenthesized_operator,
    STATE(2385), 1,
      sym__value_name,
    STATE(6085), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235912] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8066), 1,
      sym__lowercase_identifier,
    ACTIONS(8070), 1,
      anon_sym_LBRACK,
    ACTIONS(8072), 1,
      anon_sym_virtual,
    STATE(6086), 1,
      sym_attribute,
    STATE(7069), 1,
      sym_class_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235935] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8835), 1,
      sym__lowercase_identifier,
    STATE(6087), 1,
      sym_attribute,
    STATE(6189), 1,
      sym_class_type_path,
    STATE(6897), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235958] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(8865), 1,
      anon_sym_RBRACK,
    STATE(6088), 1,
      sym_attribute,
    STATE(6124), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235981] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8867), 1,
      anon_sym_rec,
    STATE(6089), 1,
      sym_attribute,
    STATE(6424), 1,
      sym_module_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8064), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [236002] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8869), 1,
      anon_sym_GT_RBRACE,
    STATE(6090), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236025] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8871), 1,
      anon_sym_GT_RBRACE,
    STATE(6091), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236048] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8766), 1,
      sym__lowercase_identifier,
    STATE(3748), 1,
      sym_class_type_path,
    STATE(6092), 1,
      sym_attribute,
    STATE(7182), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236071] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(8873), 1,
      anon_sym_RBRACK,
    STATE(6093), 1,
      sym_attribute,
    STATE(6387), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236094] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8875), 1,
      anon_sym_GT_RBRACE,
    STATE(6094), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236117] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6326), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6095), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236140] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6316), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6096), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236163] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(8877), 1,
      anon_sym_RBRACK,
    STATE(6097), 1,
      sym_attribute,
    STATE(6124), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236186] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7578), 1,
      anon_sym_COMMA,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    STATE(6098), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236209] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(8879), 1,
      anon_sym_RBRACK,
    STATE(6099), 1,
      sym_attribute,
    STATE(6124), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236232] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8881), 1,
      anon_sym_DOT,
    ACTIONS(8883), 1,
      aux_sym_type_variable_token1,
    STATE(7117), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6100), 2,
      sym_attribute,
      aux_sym_constructor_declaration_repeat1,
  [236253] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6101), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5290), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [236270] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(8886), 1,
      anon_sym_RPAREN,
    STATE(6102), 1,
      sym_attribute,
    STATE(6478), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236293] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1401), 1,
      anon_sym_DQUOTE,
    ACTIONS(7275), 1,
      anon_sym_LBRACE,
    STATE(6103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6194), 2,
      sym_string,
      sym_quoted_string,
  [236314] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    STATE(6104), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5492), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [236335] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8239), 1,
      sym__lowercase_identifier,
    ACTIONS(8243), 1,
      anon_sym_LBRACK,
    ACTIONS(8245), 1,
      anon_sym_virtual,
    STATE(4777), 1,
      sym_class_binding,
    STATE(6105), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236358] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    STATE(6106), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5504), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [236379] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8890), 1,
      anon_sym_GT_RBRACE,
    STATE(6107), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236402] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8892), 1,
      anon_sym_GT_RBRACE,
    STATE(6108), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236425] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7558), 1,
      anon_sym_COMMA,
    ACTIONS(7792), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7800), 1,
      anon_sym_PIPE,
    ACTIONS(7804), 1,
      anon_sym_as,
    STATE(6109), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236448] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8381), 1,
      sym__lowercase_identifier,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(6110), 1,
      sym_attribute,
    STATE(6814), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236471] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5739), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [236488] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6112), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5754), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [236505] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6113), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5230), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [236522] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8894), 1,
      sym__lowercase_identifier,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8896), 2,
      anon_sym_private,
      anon_sym_virtual,
    STATE(6114), 2,
      sym_attribute,
      aux_sym_method_specification_repeat1,
  [236541] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(8899), 1,
      anon_sym_RBRACK,
    STATE(6115), 1,
      sym_attribute,
    STATE(6124), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236564] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    STATE(6116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5447), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [236585] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    STATE(6117), 1,
      sym_attribute,
    STATE(6992), 1,
      sym_type_constructor_path,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236608] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8901), 1,
      sym__lowercase_identifier,
    STATE(2765), 1,
      sym_type_constructor_path,
    STATE(6118), 1,
      sym_attribute,
    STATE(6839), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236631] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8305), 1,
      sym__lowercase_identifier,
    ACTIONS(8307), 1,
      anon_sym_LBRACK,
    ACTIONS(8309), 1,
      anon_sym_virtual,
    STATE(4874), 1,
      sym_class_binding,
    STATE(6119), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236654] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8905), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8903), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    STATE(6120), 2,
      sym_attribute,
      aux_sym__type_params_repeat1,
  [236673] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8908), 1,
      sym__lowercase_identifier,
    STATE(6121), 1,
      sym_attribute,
    STATE(6196), 1,
      aux_sym_method_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8710), 2,
      anon_sym_private,
      anon_sym_virtual,
  [236694] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8908), 1,
      sym__lowercase_identifier,
    STATE(6114), 1,
      aux_sym_method_specification_repeat1,
    STATE(6122), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8710), 2,
      anon_sym_private,
      anon_sym_virtual,
  [236715] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8764), 1,
      sym__lowercase_identifier,
    STATE(3591), 1,
      sym_class_type_path,
    STATE(6123), 1,
      sym_attribute,
    STATE(7132), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236738] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8910), 1,
      anon_sym_RBRACK,
    ACTIONS(8912), 1,
      aux_sym_tag_token1,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6124), 2,
      sym_attribute,
      aux_sym_polymorphic_variant_type_repeat2,
  [236759] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8766), 1,
      sym__lowercase_identifier,
    STATE(4072), 1,
      sym_class_type_path,
    STATE(6125), 1,
      sym_attribute,
    STATE(7182), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236782] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6264), 1,
      anon_sym_in,
    ACTIONS(8915), 1,
      anon_sym_and,
    ACTIONS(8918), 1,
      sym_let_and_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6126), 2,
      sym_attribute,
      aux_sym_value_definition_repeat1,
  [236803] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8066), 1,
      sym__lowercase_identifier,
    ACTIONS(8070), 1,
      anon_sym_LBRACK,
    ACTIONS(8072), 1,
      anon_sym_virtual,
    STATE(6127), 1,
      sym_attribute,
    STATE(6438), 1,
      sym_class_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236826] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8305), 1,
      sym__lowercase_identifier,
    ACTIONS(8307), 1,
      anon_sym_LBRACK,
    ACTIONS(8309), 1,
      anon_sym_virtual,
    STATE(4719), 1,
      sym_class_binding,
    STATE(6128), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236849] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8921), 1,
      anon_sym_GT_RBRACE,
    STATE(6129), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236872] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(8923), 1,
      anon_sym_DASH_GT,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6130), 1,
      sym_attribute,
    STATE(6288), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236895] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8389), 1,
      sym__lowercase_identifier,
    STATE(2956), 1,
      sym_class_path,
    STATE(6131), 1,
      sym_attribute,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236918] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8470), 1,
      sym__lowercase_identifier,
    STATE(2259), 1,
      sym_class_path,
    STATE(6132), 1,
      sym_attribute,
    STATE(7384), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236941] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5332), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [236958] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(8923), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6134), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236981] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6135), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5966), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [236998] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(3514), 1,
      anon_sym_COMMA,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    ACTIONS(8784), 1,
      sym_hash_operator,
    STATE(6136), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237021] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(8925), 1,
      anon_sym_DASH_GT,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6137), 1,
      sym_attribute,
    STATE(6287), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237044] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8766), 1,
      sym__lowercase_identifier,
    STATE(3726), 1,
      sym_class_type_path,
    STATE(6138), 1,
      sym_attribute,
    STATE(7182), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237067] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8577), 1,
      sym__lowercase_identifier,
    ACTIONS(8579), 1,
      anon_sym_LBRACK,
    ACTIONS(8581), 1,
      anon_sym_virtual,
    STATE(6139), 1,
      sym_attribute,
    STATE(7139), 1,
      sym_class_type_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237090] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8927), 1,
      anon_sym_GT_RBRACE,
    STATE(6140), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237113] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8929), 1,
      anon_sym_GT_RBRACE,
    STATE(6141), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237136] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(8700), 1,
      anon_sym_RBRACK,
    STATE(6142), 1,
      sym_attribute,
    STATE(6391), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237159] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8831), 1,
      sym__lowercase_identifier,
    STATE(4072), 1,
      sym_class_type_path,
    STATE(6143), 1,
      sym_attribute,
    STATE(6859), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237182] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8468), 1,
      sym__lowercase_identifier,
    STATE(2005), 1,
      sym_class_path,
    STATE(6144), 1,
      sym_attribute,
    STATE(7460), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237205] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6145), 1,
      sym_attribute,
    STATE(6908), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8931), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [237226] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    ACTIONS(8841), 1,
      sym__lowercase_identifier,
    STATE(6146), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5047), 2,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [237247] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8815), 1,
      sym__lowercase_identifier,
    ACTIONS(8819), 1,
      anon_sym_mutable,
    ACTIONS(8933), 1,
      anon_sym_RBRACE,
    STATE(6147), 1,
      sym_attribute,
    STATE(7130), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237270] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8410), 1,
      sym__lowercase_identifier,
    ACTIONS(8412), 1,
      anon_sym_LBRACK,
    ACTIONS(8414), 1,
      anon_sym_virtual,
    STATE(4743), 1,
      sym_class_type_binding,
    STATE(6148), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237293] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8935), 1,
      anon_sym_GT_RBRACE,
    STATE(6149), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237316] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(8937), 1,
      anon_sym_RPAREN,
    STATE(6150), 1,
      sym_attribute,
    STATE(6385), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237339] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6332), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6151), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237362] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6332), 1,
      anon_sym_in,
    STATE(6016), 1,
      aux_sym_expression_item_repeat1,
    STATE(6152), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237385] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8261), 1,
      anon_sym_with,
    ACTIONS(8456), 1,
      anon_sym_DASH_GT,
    STATE(6153), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5492), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [237406] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6347), 1,
      anon_sym_in,
    STATE(6154), 1,
      sym_attribute,
    STATE(6220), 1,
      aux_sym_expression_item_repeat1,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237429] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8261), 1,
      anon_sym_with,
    ACTIONS(8456), 1,
      anon_sym_DASH_GT,
    STATE(6155), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5504), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [237450] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4222), 1,
      anon_sym_RBRACK,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8939), 1,
      anon_sym_GT,
    STATE(6024), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6156), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237473] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8261), 1,
      anon_sym_with,
    ACTIONS(8456), 1,
      anon_sym_DASH_GT,
    STATE(6157), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5447), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [237494] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    STATE(6158), 1,
      sym_attribute,
    STATE(7053), 1,
      sym_type_constructor_path,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237517] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8941), 1,
      anon_sym_GT_RBRACE,
    STATE(6159), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237540] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(8943), 1,
      sym__lowercase_identifier,
    STATE(6160), 1,
      sym_attribute,
    STATE(6925), 1,
      sym_extended_module_path,
    STATE(7055), 1,
      sym_module_type_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237563] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4694), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [237580] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8680), 1,
      anon_sym_as,
    STATE(6162), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6565), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_GT,
  [237599] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8680), 1,
      anon_sym_as,
    STATE(6163), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5190), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_GT,
  [237618] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8945), 1,
      anon_sym_EQ,
    ACTIONS(8947), 1,
      anon_sym_RPAREN,
    ACTIONS(8949), 1,
      anon_sym_COLON2,
    STATE(6164), 1,
      sym_attribute,
    STATE(6968), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237641] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(8951), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6165), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237664] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6260), 1,
      anon_sym_in,
    ACTIONS(8796), 1,
      anon_sym_and,
    ACTIONS(8798), 1,
      sym_let_and_operator,
    STATE(6166), 1,
      sym_attribute,
    STATE(6184), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237687] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8815), 1,
      sym__lowercase_identifier,
    ACTIONS(8819), 1,
      anon_sym_mutable,
    ACTIONS(8953), 1,
      anon_sym_RBRACE,
    STATE(6167), 1,
      sym_attribute,
    STATE(7130), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237710] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(8955), 1,
      anon_sym_DASH_GT,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6168), 1,
      sym_attribute,
    STATE(6170), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237733] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(8955), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6169), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237756] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(8957), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6170), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237779] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8959), 1,
      anon_sym_GT_RBRACE,
    STATE(6171), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237802] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6172), 1,
      sym_attribute,
    STATE(6919), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8961), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [237823] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8770), 1,
      anon_sym_LPAREN,
    ACTIONS(8963), 1,
      sym__lowercase_identifier,
    STATE(5617), 1,
      sym_parenthesized_operator,
    STATE(5728), 1,
      sym__value_name,
    STATE(6173), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237846] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8478), 1,
      sym__lowercase_identifier,
    STATE(1317), 1,
      sym_class_path,
    STATE(6174), 1,
      sym_attribute,
    STATE(7423), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237869] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    STATE(6175), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8965), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [237888] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8389), 1,
      sym__lowercase_identifier,
    STATE(2701), 1,
      sym_class_path,
    STATE(6176), 1,
      sym_attribute,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237911] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6177), 1,
      sym_attribute,
    STATE(6854), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8967), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [237932] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8969), 1,
      anon_sym_RBRACK,
    ACTIONS(8971), 1,
      anon_sym_GT,
    STATE(6178), 1,
      sym_attribute,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237955] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    ACTIONS(8784), 1,
      sym_hash_operator,
    ACTIONS(8973), 1,
      anon_sym_with,
    STATE(6179), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237978] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6365), 1,
      anon_sym_in,
    STATE(6151), 1,
      aux_sym_expression_item_repeat1,
    STATE(6180), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238001] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8975), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8965), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    STATE(6181), 2,
      sym_attribute,
      aux_sym_instantiated_class_type_repeat1,
  [238020] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8855), 1,
      sym__lowercase_identifier,
    STATE(3918), 1,
      sym_class_type_path,
    STATE(6182), 1,
      sym_attribute,
    STATE(6912), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238043] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6183), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6050), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [238060] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6274), 1,
      anon_sym_in,
    ACTIONS(8796), 1,
      anon_sym_and,
    ACTIONS(8798), 1,
      sym_let_and_operator,
    STATE(6126), 1,
      aux_sym_value_definition_repeat1,
    STATE(6184), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238083] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8551), 1,
      sym__lowercase_identifier,
    STATE(2696), 1,
      sym_class_path,
    STATE(6185), 1,
      sym_attribute,
    STATE(7549), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238106] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8978), 1,
      anon_sym_GT_RBRACE,
    STATE(6186), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238129] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6187), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6292), 4,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_DQUOTE,
  [238146] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8980), 1,
      sym__lowercase_identifier,
    STATE(6188), 1,
      sym_attribute,
    STATE(6238), 1,
      aux_sym_method_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8710), 2,
      anon_sym_private,
      anon_sym_virtual,
  [238167] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6189), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5415), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [238184] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8982), 1,
      sym__lowercase_identifier,
    STATE(6190), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(7832), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238207] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8565), 1,
      sym__lowercase_identifier,
    STATE(2284), 1,
      sym_class_path,
    STATE(6191), 1,
      sym_attribute,
    STATE(7777), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238230] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6192), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5419), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [238247] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(8925), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6193), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238270] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6194), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6300), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [238287] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8764), 1,
      sym__lowercase_identifier,
    STATE(3511), 1,
      sym_class_type_path,
    STATE(6195), 1,
      sym_attribute,
    STATE(7132), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238310] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8984), 1,
      sym__lowercase_identifier,
    STATE(6114), 1,
      aux_sym_method_specification_repeat1,
    STATE(6196), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8710), 2,
      anon_sym_private,
      anon_sym_virtual,
  [238331] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6197), 1,
      sym_attribute,
    STATE(6852), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8986), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [238352] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(8988), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6198), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238375] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8561), 1,
      sym__lowercase_identifier,
    STATE(1055), 1,
      sym_class_path,
    STATE(6199), 1,
      sym_attribute,
    STATE(7496), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238398] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8990), 1,
      anon_sym_GT_RBRACE,
    STATE(6200), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238421] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    ACTIONS(8784), 1,
      sym_hash_operator,
    ACTIONS(8992), 1,
      anon_sym_with,
    STATE(6201), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238444] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(8943), 1,
      sym__lowercase_identifier,
    STATE(6202), 1,
      sym_attribute,
    STATE(6925), 1,
      sym_extended_module_path,
    STATE(7180), 1,
      sym_module_type_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238467] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(8994), 1,
      anon_sym_GT_RBRACE,
    STATE(6203), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238490] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8949), 1,
      anon_sym_COLON2,
    ACTIONS(8996), 1,
      anon_sym_EQ,
    ACTIONS(8998), 1,
      anon_sym_RPAREN,
    STATE(6204), 1,
      sym_attribute,
    STATE(6845), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238513] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(9000), 1,
      anon_sym_GT_RBRACE,
    STATE(6205), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238536] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8389), 1,
      sym__lowercase_identifier,
    STATE(2696), 1,
      sym_class_path,
    STATE(6206), 1,
      sym_attribute,
    STATE(7603), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238559] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8450), 1,
      sym__lowercase_identifier,
    STATE(1176), 1,
      sym_class_path,
    STATE(6207), 1,
      sym_attribute,
    STATE(7567), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238582] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    ACTIONS(8784), 1,
      sym_hash_operator,
    ACTIONS(9002), 1,
      anon_sym_with,
    STATE(6208), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238605] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4341), 1,
      anon_sym_DQUOTE,
    ACTIONS(9004), 1,
      anon_sym_LBRACE,
    STATE(6209), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4390), 2,
      sym_string,
      sym_quoted_string,
  [238626] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(9006), 1,
      anon_sym_GT_RBRACE,
    STATE(6210), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238649] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(8969), 1,
      anon_sym_RBRACK,
    STATE(6124), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6211), 1,
      sym_attribute,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238672] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9008), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9011), 2,
      anon_sym_RBRACK,
      anon_sym_GT,
    STATE(6212), 2,
      sym_attribute,
      aux_sym_polymorphic_variant_type_repeat1,
  [238691] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(9013), 1,
      anon_sym_RBRACK,
    ACTIONS(9015), 1,
      anon_sym_GT,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6213), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238714] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8792), 1,
      aux_sym_attribute_id_token1,
    STATE(6043), 1,
      aux_sym_attribute_id_repeat1,
    STATE(6214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3176), 2,
      sym__left_quoted_string_delimiter,
      aux_sym_quoted_extension_token1,
  [238735] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6215), 1,
      sym_attribute,
    STATE(7140), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9017), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [238756] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    ACTIONS(8784), 1,
      sym_hash_operator,
    ACTIONS(9019), 1,
      anon_sym_with,
    STATE(6216), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238779] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6361), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6217), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238802] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5534), 1,
      anon_sym_and,
    STATE(6218), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5532), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [238821] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6252), 1,
      anon_sym_in,
    ACTIONS(8796), 1,
      anon_sym_and,
    ACTIONS(8798), 1,
      sym_let_and_operator,
    STATE(6044), 1,
      aux_sym_value_definition_repeat1,
    STATE(6219), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238844] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6206), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6220), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238867] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    STATE(6221), 1,
      sym_attribute,
    STATE(7051), 1,
      sym_type_constructor_path,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238890] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    ACTIONS(8784), 1,
      sym_hash_operator,
    ACTIONS(9021), 1,
      anon_sym_with,
    STATE(6222), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238913] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(8943), 1,
      sym__lowercase_identifier,
    STATE(6223), 1,
      sym_attribute,
    STATE(6925), 1,
      sym_extended_module_path,
    STATE(7057), 1,
      sym_module_type_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238936] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9023), 1,
      sym__lowercase_identifier,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9026), 2,
      anon_sym_RPAREN,
      anon_sym_DOT,
    STATE(6224), 2,
      sym_attribute,
      aux_sym__abstract_type_repeat1,
  [238955] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    ACTIONS(8627), 1,
      sym__lowercase_identifier,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(5325), 1,
      sym__value_name,
    STATE(6225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238978] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    ACTIONS(8784), 1,
      sym_hash_operator,
    ACTIONS(9028), 1,
      anon_sym_with,
    STATE(6226), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239001] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9030), 1,
      sym__lowercase_identifier,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9032), 2,
      anon_sym_mutable,
      anon_sym_virtual,
    STATE(6227), 2,
      sym_attribute,
      aux_sym_instance_variable_specification_repeat1,
  [239020] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8422), 1,
      sym__lowercase_identifier,
    STATE(3319), 1,
      sym_class_path,
    STATE(6228), 1,
      sym_attribute,
    STATE(7532), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239043] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(9035), 1,
      anon_sym_RBRACK,
    ACTIONS(9037), 1,
      anon_sym_GT,
    STATE(6213), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239066] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4640), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [239083] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6391), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6231), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239106] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    ACTIONS(8784), 1,
      sym_hash_operator,
    ACTIONS(9039), 1,
      anon_sym_with,
    STATE(6232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239129] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(9041), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239152] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    STATE(5071), 1,
      sym_type_constructor_path,
    STATE(6234), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239175] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    ACTIONS(8784), 1,
      sym_hash_operator,
    ACTIONS(9043), 1,
      anon_sym_with,
    STATE(6235), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239198] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    ACTIONS(8784), 1,
      sym_hash_operator,
    ACTIONS(9045), 1,
      anon_sym_with,
    STATE(6236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239221] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(9047), 1,
      anon_sym_DOT,
    STATE(6100), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6237), 1,
      sym_attribute,
    STATE(7117), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239244] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9049), 1,
      sym__lowercase_identifier,
    STATE(6114), 1,
      aux_sym_method_specification_repeat1,
    STATE(6238), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8710), 2,
      anon_sym_private,
      anon_sym_virtual,
  [239265] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    ACTIONS(8784), 1,
      sym_hash_operator,
    ACTIONS(9051), 1,
      anon_sym_with,
    STATE(6239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239288] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5538), 1,
      anon_sym_and,
    STATE(6240), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5536), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [239307] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6351), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6241), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239330] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8577), 1,
      sym__lowercase_identifier,
    ACTIONS(8579), 1,
      anon_sym_LBRACK,
    ACTIONS(8581), 1,
      anon_sym_virtual,
    STATE(6242), 1,
      sym_attribute,
    STATE(6627), 1,
      sym_class_type_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239353] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6236), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6243), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239376] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3510), 1,
      anon_sym_POUND,
    ACTIONS(8698), 1,
      anon_sym_DOT,
    ACTIONS(8784), 1,
      sym_hash_operator,
    ACTIONS(9053), 1,
      anon_sym_with,
    STATE(6244), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239399] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5338), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [239416] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5382), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [239433] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(9055), 1,
      anon_sym_RBRACK,
    STATE(6124), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6247), 1,
      sym_attribute,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239456] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8381), 1,
      sym__lowercase_identifier,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(6248), 1,
      sym_attribute,
    STATE(7035), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239479] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6249), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5390), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [239496] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8375), 1,
      sym__lowercase_identifier,
    STATE(1489), 1,
      sym_class_path,
    STATE(6250), 1,
      sym_attribute,
    STATE(7413), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239519] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(9057), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239542] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9059), 1,
      sym__lowercase_identifier,
    ACTIONS(9061), 1,
      anon_sym_LPAREN,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(3716), 1,
      sym__value_name,
    STATE(6252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239565] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(9063), 1,
      anon_sym_RPAREN,
    STATE(6253), 1,
      sym_attribute,
    STATE(6403), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239588] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8766), 1,
      sym__lowercase_identifier,
    STATE(4110), 1,
      sym_class_type_path,
    STATE(6254), 1,
      sym_attribute,
    STATE(7182), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239611] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    ACTIONS(9065), 1,
      anon_sym_DOT,
    STATE(6100), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6255), 1,
      sym_attribute,
    STATE(7117), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239634] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(9067), 1,
      anon_sym_DASH_GT,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6256), 1,
      sym_attribute,
    STATE(6260), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239657] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    STATE(6257), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5528), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [239678] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(9067), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6258), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239701] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    STATE(6259), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5455), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [239722] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(9069), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6260), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239745] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(9071), 1,
      anon_sym_RBRACK,
    STATE(6261), 1,
      sym_attribute,
    STATE(6480), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239768] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(9073), 1,
      sym__lowercase_identifier,
    STATE(3647), 1,
      sym_type_constructor_path,
    STATE(6262), 1,
      sym_attribute,
    STATE(6902), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239791] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(8718), 1,
      anon_sym_RBRACK,
    STATE(6263), 1,
      sym_attribute,
    STATE(6521), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239814] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8855), 1,
      sym__lowercase_identifier,
    STATE(3902), 1,
      sym_class_type_path,
    STATE(6264), 1,
      sym_attribute,
    STATE(6912), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239837] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8261), 1,
      anon_sym_with,
    ACTIONS(8456), 1,
      anon_sym_DASH_GT,
    STATE(6265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5455), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [239858] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(9035), 1,
      anon_sym_RBRACK,
    ACTIONS(9037), 1,
      anon_sym_GT,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239881] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8381), 1,
      sym__lowercase_identifier,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    STATE(4997), 1,
      sym_parenthesized_operator,
    STATE(6267), 1,
      sym_attribute,
    STATE(7105), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239904] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    STATE(6268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8670), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [239923] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8680), 1,
      anon_sym_as,
    STATE(6269), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6569), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_GT,
  [239942] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9075), 1,
      sym__lowercase_identifier,
    STATE(6224), 1,
      aux_sym__abstract_type_repeat1,
    STATE(6270), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9077), 2,
      anon_sym_RPAREN,
      anon_sym_DOT,
  [239963] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8766), 1,
      sym__lowercase_identifier,
    STATE(3625), 1,
      sym_class_type_path,
    STATE(6271), 1,
      sym_attribute,
    STATE(7182), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239986] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6252), 1,
      anon_sym_in,
    ACTIONS(8796), 1,
      anon_sym_and,
    ACTIONS(8798), 1,
      sym_let_and_operator,
    STATE(6126), 1,
      aux_sym_value_definition_repeat1,
    STATE(6272), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240009] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9059), 1,
      sym__lowercase_identifier,
    ACTIONS(9061), 1,
      anon_sym_LPAREN,
    STATE(3689), 1,
      sym_parenthesized_operator,
    STATE(3691), 1,
      sym__value_name,
    STATE(6273), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240032] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(9079), 1,
      sym__lowercase_identifier,
    STATE(6274), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_extended_module_path,
    STATE(8295), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240055] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(9081), 1,
      anon_sym_DASH_GT,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6275), 1,
      sym_attribute,
    STATE(6282), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240078] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9085), 1,
      anon_sym_of,
    STATE(6276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9083), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [240097] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(9081), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6277), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240120] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2978), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6308), 1,
      anon_sym_in,
    STATE(5846), 1,
      aux_sym_expression_item_repeat1,
    STATE(6278), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240143] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    ACTIONS(9013), 1,
      anon_sym_RBRACK,
    STATE(6124), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6279), 1,
      sym_attribute,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240166] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(9087), 1,
      sym__lowercase_identifier,
    STATE(5780), 1,
      sym_type_constructor_path,
    STATE(6280), 1,
      sym_attribute,
    STATE(6882), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240189] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8764), 1,
      sym__lowercase_identifier,
    STATE(3551), 1,
      sym_class_type_path,
    STATE(6281), 1,
      sym_attribute,
    STATE(7132), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240212] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(9089), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6282), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240235] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8503), 1,
      sym__lowercase_identifier,
    STATE(1360), 1,
      sym_class_path,
    STATE(6283), 1,
      sym_attribute,
    STATE(7279), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240258] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(9091), 1,
      anon_sym_GT_RBRACE,
    STATE(6284), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240281] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8383), 1,
      anon_sym_LPAREN,
    ACTIONS(8841), 1,
      sym__lowercase_identifier,
    STATE(6285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5074), 2,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [240302] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3600), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    STATE(6286), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240325] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(9093), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6287), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240348] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(9095), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240371] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8410), 1,
      sym__lowercase_identifier,
    ACTIONS(8412), 1,
      anon_sym_LBRACK,
    ACTIONS(8414), 1,
      anon_sym_virtual,
    STATE(4508), 1,
      sym_class_type_binding,
    STATE(6289), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240394] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(4965), 1,
      sym__lowercase_identifier,
    STATE(6290), 1,
      sym_attribute,
    STATE(7148), 1,
      sym_type_constructor_path,
    STATE(7160), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240417] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    ACTIONS(8855), 1,
      sym__lowercase_identifier,
    STATE(3926), 1,
      sym_class_type_path,
    STATE(6291), 1,
      sym_attribute,
    STATE(6912), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240440] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(9097), 1,
      anon_sym_DASH_GT,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6292), 1,
      sym_attribute,
    STATE(6294), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240463] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(9097), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6293), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240486] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    ACTIONS(9099), 1,
      anon_sym_DASH_GT,
    STATE(3886), 1,
      aux_sym_module_binding_repeat1,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6294), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240509] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(9101), 1,
      anon_sym_GT_RBRACE,
    STATE(6295), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240532] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8249), 1,
      sym__lowercase_identifier,
    ACTIONS(8251), 1,
      anon_sym_LBRACK,
    ACTIONS(8253), 1,
      anon_sym_virtual,
    STATE(4626), 1,
      sym_class_binding,
    STATE(6296), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240555] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9103), 1,
      anon_sym_rec,
    STATE(4907), 1,
      sym_module_binding,
    STATE(6297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8342), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [240576] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4198), 1,
      anon_sym_RBRACK,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(9105), 1,
      anon_sym_GT,
    STATE(6266), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6298), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240599] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8239), 1,
      sym__lowercase_identifier,
    ACTIONS(8243), 1,
      anon_sym_LBRACK,
    ACTIONS(8245), 1,
      anon_sym_virtual,
    STATE(4719), 1,
      sym_class_binding,
    STATE(6299), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240622] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    ACTIONS(9107), 1,
      anon_sym_GT_RBRACE,
    STATE(6300), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240645] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5636), 1,
      sym__uppercase_identifier,
    ACTIONS(8943), 1,
      sym__lowercase_identifier,
    STATE(6301), 1,
      sym_attribute,
    STATE(6925), 1,
      sym_extended_module_path,
    STATE(7173), 1,
      sym_module_type_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240668] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3608), 1,
      anon_sym_COMMA,
    ACTIONS(8108), 1,
      anon_sym_PIPE,
    ACTIONS(8110), 1,
      anon_sym_as,
    ACTIONS(8112), 1,
      anon_sym_COLON_COLON,
    STATE(6302), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240691] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6303), 1,
      sym_attribute,
    STATE(7177), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9109), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [240712] = 7,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(8551), 1,
      sym__lowercase_identifier,
    STATE(949), 1,
      sym_class_path,
    STATE(6304), 1,
      sym_attribute,
    STATE(7549), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240735] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(3700), 1,
      sym_attribute_id,
    STATE(6305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9111), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [240753] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9113), 1,
      anon_sym_SEMI,
    ACTIONS(9115), 1,
      anon_sym_RBRACE,
    STATE(6306), 1,
      sym_attribute,
    STATE(6412), 1,
      aux_sym_record_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240773] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8418), 1,
      anon_sym_RBRACE,
    ACTIONS(9117), 1,
      anon_sym_SEMI,
    STATE(6307), 1,
      sym_attribute,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240793] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8892), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9119), 1,
      anon_sym_SEMI,
    STATE(6308), 1,
      sym_attribute,
    STATE(6464), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240813] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6309), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8903), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [240829] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8843), 1,
      anon_sym_RBRACK,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6310), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240849] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9121), 1,
      anon_sym_SEMI,
    ACTIONS(9123), 1,
      anon_sym_RBRACE,
    STATE(6311), 1,
      sym_attribute,
    STATE(6313), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240869] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9125), 1,
      sym__lowercase_identifier,
    ACTIONS(9127), 1,
      sym_indexing_operator,
    STATE(6312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240889] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8484), 1,
      anon_sym_RBRACE,
    ACTIONS(9129), 1,
      anon_sym_SEMI,
    STATE(6313), 1,
      sym_attribute,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240909] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9131), 1,
      anon_sym_SEMI,
    ACTIONS(9133), 1,
      anon_sym_RBRACE,
    STATE(6314), 1,
      sym_attribute,
    STATE(6642), 1,
      aux_sym_record_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240929] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(9135), 1,
      anon_sym_RPAREN,
    STATE(6315), 1,
      sym_attribute,
    STATE(7846), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240949] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6316), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9137), 3,
      anon_sym_RPAREN,
      anon_sym_DOT,
      sym__lowercase_identifier,
  [240965] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9000), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9139), 1,
      anon_sym_SEMI,
    STATE(6317), 1,
      sym_attribute,
    STATE(6464), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240985] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1627), 1,
      anon_sym_RBRACK,
    ACTIONS(9141), 1,
      anon_sym_SEMI,
    STATE(6318), 1,
      sym_attribute,
    STATE(6446), 1,
      aux_sym__sequence_expression_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241005] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9127), 1,
      sym_indexing_operator,
    ACTIONS(9143), 1,
      sym__lowercase_identifier,
    STATE(6319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241025] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2870), 1,
      anon_sym_RBRACK,
    ACTIONS(9145), 1,
      anon_sym_SEMI,
    STATE(6320), 1,
      sym_attribute,
    STATE(6641), 1,
      aux_sym__sequence_binding_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241045] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9147), 1,
      anon_sym_SEMI,
    ACTIONS(9149), 1,
      anon_sym_RBRACE,
    STATE(6321), 1,
      sym_attribute,
    STATE(6369), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241065] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8393), 1,
      anon_sym_RBRACE,
    ACTIONS(9151), 1,
      anon_sym_SEMI,
    STATE(6322), 1,
      sym_attribute,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241085] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8124), 1,
      anon_sym_RBRACE,
    ACTIONS(9153), 1,
      anon_sym_SEMI,
    STATE(6323), 1,
      sym_attribute,
    STATE(6649), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241105] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1471), 1,
      anon_sym_end,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6324), 1,
      sym_attribute,
    STATE(7713), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241125] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7360), 1,
      sym__uppercase_identifier,
    ACTIONS(8662), 1,
      anon_sym_LPAREN,
    STATE(3624), 1,
      sym__constructor_name,
    STATE(6325), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241145] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9155), 1,
      anon_sym_EQ,
    STATE(6326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9157), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [241163] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6327), 1,
      sym_attribute,
    STATE(6363), 1,
      sym_attribute_id,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [241181] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8402), 1,
      anon_sym_RBRACE,
    ACTIONS(9161), 1,
      anon_sym_SEMI,
    STATE(6328), 1,
      sym_attribute,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241201] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9163), 1,
      anon_sym_EQ,
    STATE(6329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9165), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [241219] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(11), 1,
      sym_attribute_id,
    STATE(6330), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [241237] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(12), 1,
      sym_attribute_id,
    STATE(6331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [241255] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(13), 1,
      sym_attribute_id,
    STATE(6332), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [241273] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9169), 1,
      anon_sym_RPAREN,
    STATE(6333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241293] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9171), 1,
      anon_sym_SEMI,
    ACTIONS(9173), 1,
      anon_sym_RBRACE,
    STATE(6334), 1,
      sym_attribute,
    STATE(6358), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241313] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6233), 1,
      aux_sym_module_binding_repeat1,
    STATE(6335), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241333] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2888), 1,
      anon_sym_RBRACK,
    ACTIONS(9175), 1,
      anon_sym_SEMI,
    STATE(6336), 1,
      sym_attribute,
    STATE(6476), 1,
      aux_sym__sequence_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241353] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9177), 1,
      aux_sym_quoted_extension_token1,
    STATE(6337), 1,
      sym_attribute,
    STATE(8265), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241373] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9179), 1,
      anon_sym_RPAREN,
    STATE(6338), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241393] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9181), 1,
      aux_sym_quoted_extension_token1,
    STATE(6339), 1,
      sym_attribute,
    STATE(8296), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241413] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(14), 1,
      sym_attribute_id,
    STATE(6340), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [241431] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9183), 1,
      anon_sym_SEMI,
    ACTIONS(9185), 1,
      anon_sym_GT_RBRACE,
    STATE(6341), 1,
      sym_attribute,
    STATE(6365), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241451] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9187), 1,
      anon_sym_SEMI,
    ACTIONS(9189), 1,
      anon_sym_GT_RBRACE,
    STATE(6342), 1,
      sym_attribute,
    STATE(6361), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241471] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9191), 1,
      anon_sym_EQ,
    STATE(6343), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9193), 2,
      anon_sym_SEMI,
      anon_sym_GT_RBRACE,
  [241489] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(9195), 1,
      anon_sym_EQ,
    STATE(6344), 1,
      sym_attribute,
    STATE(7870), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241509] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8003), 1,
      anon_sym_SEMI,
    ACTIONS(8005), 1,
      anon_sym_GT,
    STATE(6345), 1,
      sym_attribute,
    STATE(6414), 1,
      aux_sym_object_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241529] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(9197), 1,
      anon_sym_RPAREN,
    STATE(6346), 1,
      sym_attribute,
    STATE(8329), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241549] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(3225), 1,
      sym_attribute_id,
    STATE(6347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [241567] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5591), 1,
      anon_sym_RPAREN,
    STATE(6348), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9199), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [241585] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9203), 1,
      anon_sym_RBRACK,
    STATE(6349), 1,
      sym_attribute,
    STATE(6367), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241605] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9205), 1,
      anon_sym_SEMI,
    ACTIONS(9207), 1,
      anon_sym_RBRACE,
    STATE(6328), 1,
      aux_sym_record_expression_repeat1,
    STATE(6350), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241625] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8326), 1,
      anon_sym_RBRACE,
    ACTIONS(9209), 1,
      anon_sym_SEMI,
    STATE(6351), 1,
      sym_attribute,
    STATE(6485), 1,
      aux_sym_record_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241645] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9211), 1,
      anon_sym_RBRACK,
    STATE(6120), 1,
      aux_sym__type_params_repeat1,
    STATE(6352), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241665] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9211), 1,
      anon_sym_RBRACK,
    STATE(6353), 1,
      sym_attribute,
    STATE(6738), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241685] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9213), 1,
      aux_sym_quoted_extension_token1,
    STATE(6354), 1,
      sym_attribute,
    STATE(8305), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241705] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(9215), 1,
      anon_sym_EQ,
    STATE(6355), 1,
      sym_attribute,
    STATE(8307), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241725] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4621), 1,
      sym_module_binding,
    STATE(6356), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8144), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [241743] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2870), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(9217), 1,
      anon_sym_SEMI,
    STATE(6357), 1,
      sym_attribute,
    STATE(6377), 1,
      aux_sym__sequence_binding_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241763] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_RBRACE,
    ACTIONS(9219), 1,
      anon_sym_SEMI,
    STATE(6358), 1,
      sym_attribute,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241783] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9221), 1,
      anon_sym_EQ,
    STATE(6359), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9223), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [241801] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(9225), 1,
      anon_sym_EQ,
    STATE(6360), 1,
      sym_attribute,
    STATE(8126), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241821] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8927), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9227), 1,
      anon_sym_SEMI,
    STATE(6361), 1,
      sym_attribute,
    STATE(6464), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241841] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6583), 1,
      anon_sym_in,
    ACTIONS(9229), 1,
      anon_sym_and,
    STATE(6362), 1,
      sym_attribute,
    STATE(6422), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241861] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9231), 1,
      aux_sym_quoted_extension_token1,
    STATE(6363), 1,
      sym_attribute,
    STATE(7222), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241881] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8815), 1,
      sym__lowercase_identifier,
    ACTIONS(8819), 1,
      anon_sym_mutable,
    STATE(6364), 1,
      sym_attribute,
    STATE(6541), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241901] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8935), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9233), 1,
      anon_sym_SEMI,
    STATE(6365), 1,
      sym_attribute,
    STATE(6464), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241921] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(9235), 1,
      anon_sym_RPAREN,
    STATE(6366), 1,
      sym_attribute,
    STATE(8123), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241941] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9237), 1,
      anon_sym_RBRACK,
    STATE(6120), 1,
      aux_sym__type_params_repeat1,
    STATE(6367), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241961] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9237), 1,
      anon_sym_RBRACK,
    STATE(6368), 1,
      sym_attribute,
    STATE(6384), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241981] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8529), 1,
      anon_sym_RBRACE,
    ACTIONS(9239), 1,
      anon_sym_SEMI,
    STATE(6369), 1,
      sym_attribute,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242001] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9241), 1,
      anon_sym_SEMI,
    ACTIONS(9243), 1,
      anon_sym_RBRACE,
    STATE(6370), 1,
      sym_attribute,
    STATE(6380), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242021] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8058), 1,
      anon_sym_RBRACE,
    ACTIONS(9245), 1,
      anon_sym_SEMI,
    STATE(6371), 1,
      sym_attribute,
    STATE(6649), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242041] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6510), 1,
      anon_sym_in,
    ACTIONS(9247), 1,
      anon_sym_and,
    STATE(6372), 1,
      sym_attribute,
    STATE(6431), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242061] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6717), 1,
      sym__uppercase_identifier,
    ACTIONS(8662), 1,
      anon_sym_LPAREN,
    STATE(2174), 1,
      sym__constructor_name,
    STATE(6373), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242081] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2888), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(9249), 1,
      anon_sym_SEMI,
    STATE(6374), 1,
      sym_attribute,
    STATE(6388), 1,
      aux_sym__sequence_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242101] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9251), 1,
      anon_sym_RBRACK,
    STATE(6120), 1,
      aux_sym__type_params_repeat1,
    STATE(6375), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242121] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9127), 1,
      sym_indexing_operator,
    ACTIONS(9253), 1,
      sym__lowercase_identifier,
    STATE(6376), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242141] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8472), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(9255), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6377), 2,
      sym_attribute,
      aux_sym__sequence_binding_pattern_content_repeat1,
  [242159] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4503), 1,
      sym_module_binding,
    STATE(6378), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8144), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [242177] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(9258), 1,
      anon_sym_RPAREN,
    STATE(6379), 1,
      sym_attribute,
    STATE(8209), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242197] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8462), 1,
      anon_sym_RBRACE,
    ACTIONS(9260), 1,
      anon_sym_SEMI,
    STATE(6380), 1,
      sym_attribute,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242217] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6765), 1,
      anon_sym_in,
    ACTIONS(9262), 1,
      anon_sym_and,
    STATE(6381), 1,
      sym_attribute,
    STATE(6437), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242237] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9264), 1,
      anon_sym_SEMI,
    ACTIONS(9267), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6382), 2,
      sym_attribute,
      aux_sym_record_declaration_repeat1,
  [242255] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9269), 1,
      anon_sym_RPAREN,
    ACTIONS(9271), 1,
      anon_sym_COLON2,
    STATE(6383), 1,
      sym_attribute,
    STATE(7350), 1,
      sym__class_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242275] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9273), 1,
      anon_sym_RBRACK,
    STATE(6120), 1,
      aux_sym__type_params_repeat1,
    STATE(6384), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242295] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(9275), 1,
      anon_sym_RPAREN,
    STATE(6181), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(6385), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242315] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(6071), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6386), 1,
      sym_attribute,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242335] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(9277), 1,
      anon_sym_RBRACK,
    STATE(6181), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(6387), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242355] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8369), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(9279), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6388), 2,
      sym_attribute,
      aux_sym__sequence_pattern_content_repeat1,
  [242373] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(9282), 1,
      anon_sym_type,
    STATE(6389), 1,
      sym_attribute,
    STATE(6613), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242393] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(16), 1,
      sym_attribute_id,
    STATE(6390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [242411] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(9284), 1,
      anon_sym_RBRACK,
    STATE(6181), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(6391), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242431] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(6088), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6392), 1,
      sym_attribute,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242451] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9286), 1,
      anon_sym_SEMI,
    ACTIONS(9288), 1,
      anon_sym_RBRACE,
    STATE(6393), 1,
      sym_attribute,
    STATE(6440), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242471] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(6099), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6394), 1,
      sym_attribute,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242491] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4706), 1,
      anon_sym_DOT,
    STATE(6395), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9290), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [242509] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9292), 1,
      anon_sym_SEMI,
    ACTIONS(9294), 1,
      anon_sym_RBRACE,
    STATE(6396), 1,
      sym_attribute,
    STATE(6413), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242529] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9296), 1,
      anon_sym_RPAREN,
    STATE(6397), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242549] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9298), 1,
      anon_sym_RPAREN,
    STATE(6120), 1,
      aux_sym__type_params_repeat1,
    STATE(6398), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242569] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4214), 1,
      anon_sym_RBRACK,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    STATE(6399), 1,
      sym_attribute,
    STATE(6429), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242589] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9300), 1,
      anon_sym_RPAREN,
    STATE(6400), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242609] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9302), 1,
      anon_sym_SEMI,
    ACTIONS(9304), 1,
      anon_sym_GT_RBRACE,
    STATE(6401), 1,
      sym_attribute,
    STATE(6417), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242629] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1447), 1,
      anon_sym_end,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6402), 1,
      sym_attribute,
    STATE(7617), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242649] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(9306), 1,
      anon_sym_RPAREN,
    STATE(6181), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(6403), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242669] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6404), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5536), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [242685] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(9035), 1,
      anon_sym_RBRACK,
    STATE(6405), 1,
      sym_attribute,
    STATE(6457), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242705] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(3490), 1,
      sym_attribute_id,
    STATE(6406), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9308), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [242723] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9310), 1,
      anon_sym_RPAREN,
    STATE(6407), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242743] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8815), 1,
      sym__lowercase_identifier,
    ACTIONS(8819), 1,
      anon_sym_mutable,
    STATE(6314), 1,
      sym_field_declaration,
    STATE(6408), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242763] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8303), 1,
      anon_sym_RBRACE,
    ACTIONS(9312), 1,
      anon_sym_SEMI,
    STATE(6409), 1,
      sym_attribute,
    STATE(6649), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242783] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(9035), 1,
      anon_sym_RBRACK,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6410), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242803] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(6279), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6411), 1,
      sym_attribute,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242823] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8118), 1,
      anon_sym_RBRACE,
    ACTIONS(9314), 1,
      anon_sym_SEMI,
    STATE(6412), 1,
      sym_attribute,
    STATE(6485), 1,
      aux_sym_record_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242843] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8359), 1,
      anon_sym_RBRACE,
    ACTIONS(9316), 1,
      anon_sym_SEMI,
    STATE(6413), 1,
      sym_attribute,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242863] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5013), 1,
      anon_sym_GT,
    ACTIONS(9318), 1,
      anon_sym_SEMI,
    STATE(6414), 1,
      sym_attribute,
    STATE(6658), 1,
      aux_sym_object_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242883] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9320), 1,
      anon_sym_RPAREN,
    STATE(6415), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242903] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8804), 1,
      anon_sym_RBRACK,
    STATE(6416), 1,
      sym_attribute,
    STATE(6609), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242923] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9107), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9322), 1,
      anon_sym_SEMI,
    STATE(6417), 1,
      sym_attribute,
    STATE(6464), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242943] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6561), 1,
      anon_sym_in,
    ACTIONS(9229), 1,
      anon_sym_and,
    STATE(6418), 1,
      sym_attribute,
    STATE(6532), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242963] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9324), 1,
      anon_sym_RBRACK,
    STATE(6419), 1,
      sym_attribute,
    STATE(6643), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242983] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9326), 1,
      anon_sym_SEMI,
    ACTIONS(9328), 1,
      anon_sym_RBRACE,
    STATE(6420), 1,
      sym_attribute,
    STATE(6427), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243003] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4551), 1,
      sym_module_binding,
    STATE(6421), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8144), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [243021] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6561), 1,
      anon_sym_in,
    ACTIONS(9229), 1,
      anon_sym_and,
    STATE(6422), 1,
      sym_attribute,
    STATE(6549), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243041] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9127), 1,
      sym_indexing_operator,
    ACTIONS(9330), 1,
      sym__lowercase_identifier,
    STATE(6423), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243061] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6415), 1,
      anon_sym_in,
    ACTIONS(9247), 1,
      anon_sym_and,
    STATE(6424), 1,
      sym_attribute,
    STATE(6569), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243081] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(5110), 1,
      anon_sym_DOT,
    ACTIONS(9332), 1,
      anon_sym_RPAREN,
    STATE(6425), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243101] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9127), 1,
      sym_indexing_operator,
    ACTIONS(9334), 1,
      sym__lowercase_identifier,
    STATE(6426), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243121] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8416), 1,
      anon_sym_RBRACE,
    ACTIONS(9336), 1,
      anon_sym_SEMI,
    STATE(6427), 1,
      sym_attribute,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243141] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(9338), 1,
      anon_sym_RPAREN,
    STATE(6428), 1,
      sym_attribute,
    STATE(7981), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243161] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8804), 1,
      anon_sym_RBRACK,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6429), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243181] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6430), 1,
      sym_attribute,
    STATE(7046), 1,
      sym_module_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8064), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [243199] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6415), 1,
      anon_sym_in,
    ACTIONS(9247), 1,
      anon_sym_and,
    STATE(6431), 1,
      sym_attribute,
    STATE(6622), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243219] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6432), 1,
      sym_attribute,
    STATE(6623), 1,
      sym_module_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8064), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [243237] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6599), 1,
      anon_sym_in,
    ACTIONS(9340), 1,
      anon_sym_and,
    STATE(6433), 1,
      sym_attribute,
    STATE(6626), 1,
      aux_sym_class_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243257] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8028), 1,
      anon_sym_SEMI,
    ACTIONS(8030), 1,
      anon_sym_GT,
    STATE(6434), 1,
      sym_attribute,
    STATE(6435), 1,
      aux_sym_object_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243277] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4993), 1,
      anon_sym_GT,
    ACTIONS(9342), 1,
      anon_sym_SEMI,
    STATE(6435), 1,
      sym_attribute,
    STATE(6658), 1,
      aux_sym_object_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243297] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(9344), 1,
      anon_sym_EQ,
    STATE(6436), 1,
      sym_attribute,
    STATE(7768), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243317] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6753), 1,
      anon_sym_in,
    ACTIONS(9262), 1,
      anon_sym_and,
    STATE(6437), 1,
      sym_attribute,
    STATE(6632), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243337] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6753), 1,
      anon_sym_in,
    ACTIONS(9262), 1,
      anon_sym_and,
    STATE(6438), 1,
      sym_attribute,
    STATE(6633), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243357] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(10), 1,
      sym_attribute_id,
    STATE(6439), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [243375] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8379), 1,
      anon_sym_RBRACE,
    ACTIONS(9346), 1,
      anon_sym_SEMI,
    STATE(6440), 1,
      sym_attribute,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243395] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1363), 1,
      anon_sym_end,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6441), 1,
      sym_attribute,
    STATE(7722), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243415] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9348), 1,
      anon_sym_SEMI,
    ACTIONS(9350), 1,
      anon_sym_RBRACE,
    STATE(6442), 1,
      sym_attribute,
    STATE(6466), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243435] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6424), 1,
      sym_module_binding,
    STATE(6443), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8064), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [243453] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6444), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1879), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [243469] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(9352), 1,
      anon_sym_RBRACK,
    STATE(6181), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(6445), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243489] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5015), 1,
      anon_sym_RBRACK,
    ACTIONS(9354), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6446), 2,
      sym_attribute,
      aux_sym__sequence_expression_content_repeat1,
  [243507] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9357), 1,
      sym__lowercase_identifier,
    ACTIONS(9359), 1,
      anon_sym_LPAREN,
    STATE(3008), 1,
      sym__simple_value_pattern,
    STATE(6447), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243527] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8929), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9361), 1,
      anon_sym_SEMI,
    STATE(6448), 1,
      sym_attribute,
    STATE(6464), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243547] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9363), 1,
      anon_sym_SEMI,
    ACTIONS(9365), 1,
      anon_sym_RBRACE,
    STATE(6449), 1,
      sym_attribute,
    STATE(6463), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243567] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9367), 1,
      anon_sym_SEMI,
    ACTIONS(9369), 1,
      anon_sym_RBRACE,
    STATE(6322), 1,
      aux_sym_record_expression_repeat1,
    STATE(6450), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243587] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9371), 1,
      anon_sym_SEMI,
    ACTIONS(9373), 1,
      anon_sym_GT_RBRACE,
    STATE(6451), 1,
      sym_attribute,
    STATE(6491), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243607] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9375), 1,
      anon_sym_SEMI,
    ACTIONS(9377), 1,
      anon_sym_GT_RBRACE,
    STATE(6452), 1,
      sym_attribute,
    STATE(6470), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243627] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6453), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5548), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [243643] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9379), 1,
      anon_sym_SEMI,
    ACTIONS(9382), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6454), 2,
      sym_attribute,
      aux_sym_record_expression_repeat1,
  [243661] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9384), 1,
      anon_sym_SEMI,
    ACTIONS(9386), 1,
      anon_sym_RBRACE,
    STATE(6455), 1,
      sym_attribute,
    STATE(6497), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243681] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(17), 1,
      sym_attribute_id,
    STATE(6456), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [243699] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(9013), 1,
      anon_sym_RBRACK,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6457), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243719] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9388), 1,
      anon_sym_RPAREN,
    STATE(6458), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243739] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(9013), 1,
      anon_sym_RBRACK,
    STATE(6459), 1,
      sym_attribute,
    STATE(6473), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243759] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9390), 1,
      aux_sym_quoted_extension_token1,
    STATE(6460), 1,
      sym_attribute,
    STATE(7425), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243779] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(18), 1,
      sym_attribute_id,
    STATE(6461), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [243797] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(9392), 1,
      anon_sym_EQ,
    STATE(6462), 1,
      sym_attribute,
    STATE(7899), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243817] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8531), 1,
      anon_sym_RBRACE,
    ACTIONS(9394), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6463), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243837] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9396), 1,
      anon_sym_SEMI,
    ACTIONS(9399), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6464), 2,
      sym_attribute,
      aux_sym_object_copy_expression_repeat1,
  [243855] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9401), 1,
      anon_sym_RBRACK,
    STATE(6465), 1,
      sym_attribute,
    STATE(6471), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243875] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8365), 1,
      anon_sym_RBRACE,
    ACTIONS(9403), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6466), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243895] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9407), 1,
      anon_sym_LBRACK,
    STATE(6467), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9405), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
  [243913] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9409), 1,
      anon_sym_RPAREN,
    STATE(6468), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9411), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [243931] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(9413), 1,
      anon_sym_EQ,
    STATE(6469), 1,
      sym_attribute,
    STATE(8003), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243951] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8875), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9415), 1,
      anon_sym_SEMI,
    STATE(6464), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(6470), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243971] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9417), 1,
      anon_sym_RBRACK,
    STATE(6120), 1,
      aux_sym__type_params_repeat1,
    STATE(6471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [243991] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9417), 1,
      anon_sym_RBRACK,
    STATE(6472), 1,
      sym_attribute,
    STATE(6477), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244011] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8899), 1,
      anon_sym_RBRACK,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6473), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244031] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9419), 1,
      anon_sym_SEMI,
    ACTIONS(9421), 1,
      anon_sym_RBRACE,
    STATE(6474), 1,
      sym_attribute,
    STATE(6488), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244051] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6198), 1,
      aux_sym_module_binding_repeat1,
    STATE(6475), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244071] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8369), 1,
      anon_sym_RBRACK,
    ACTIONS(9423), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6476), 2,
      sym_attribute,
      aux_sym__sequence_pattern_content_repeat1,
  [244089] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9426), 1,
      anon_sym_RBRACK,
    STATE(6120), 1,
      aux_sym__type_params_repeat1,
    STATE(6477), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244109] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(9428), 1,
      anon_sym_RPAREN,
    STATE(6181), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(6478), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244129] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(6211), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6479), 1,
      sym_attribute,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244149] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(9430), 1,
      anon_sym_RBRACK,
    STATE(6181), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(6480), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244169] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9432), 1,
      anon_sym_LPAREN,
    ACTIONS(9434), 1,
      anon_sym_LBRACK,
    ACTIONS(9436), 1,
      anon_sym_LBRACE,
    STATE(6481), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244189] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9438), 1,
      anon_sym_SEMI,
    ACTIONS(9440), 1,
      anon_sym_RBRACE,
    STATE(6482), 1,
      sym_attribute,
    STATE(6524), 1,
      aux_sym_record_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244209] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(6067), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6483), 1,
      sym_attribute,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244229] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(6247), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6484), 1,
      sym_attribute,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244249] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9442), 1,
      anon_sym_SEMI,
    ACTIONS(9445), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6485), 2,
      sym_attribute,
      aux_sym_record_pattern_repeat1,
  [244267] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(24), 1,
      sym_attribute_id,
    STATE(6486), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [244285] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9127), 1,
      sym_indexing_operator,
    ACTIONS(9447), 1,
      sym__lowercase_identifier,
    STATE(6487), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244305] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8541), 1,
      anon_sym_RBRACE,
    ACTIONS(9449), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6488), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244325] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(6097), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6489), 1,
      sym_attribute,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244345] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6490), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5176), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [244361] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8800), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9451), 1,
      anon_sym_SEMI,
    STATE(6464), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(6491), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244381] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3186), 1,
      aux_sym_attribute_id_token1,
    STATE(6492), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3188), 2,
      sym__left_quoted_string_delimiter,
      aux_sym_quoted_extension_token1,
  [244399] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(9453), 1,
      anon_sym_RPAREN,
    STATE(6493), 1,
      sym_attribute,
    STATE(7410), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244419] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(8), 1,
      sym_attribute_id,
    STATE(6494), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [244437] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8602), 1,
      sym__lowercase_identifier,
    STATE(6343), 1,
      sym__instance_variable_name,
    STATE(6495), 1,
      sym_attribute,
    STATE(6844), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244457] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6496), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9011), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [244473] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8154), 1,
      anon_sym_RBRACE,
    ACTIONS(9455), 1,
      anon_sym_SEMI,
    STATE(6497), 1,
      sym_attribute,
    STATE(6649), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244493] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9457), 1,
      aux_sym_quoted_extension_token1,
    STATE(6498), 1,
      sym_attribute,
    STATE(7207), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244513] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(20), 1,
      sym_attribute_id,
    STATE(6499), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [244531] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(9459), 1,
      anon_sym_EQ,
    STATE(6500), 1,
      sym_attribute,
    STATE(7232), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244551] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(9461), 1,
      anon_sym_EQ,
    STATE(6501), 1,
      sym_attribute,
    STATE(7248), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244571] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6502), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5184), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [244587] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      aux_sym_tag_token1,
    STATE(6115), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6503), 1,
      sym_attribute,
    STATE(6913), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244607] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(5110), 1,
      anon_sym_DOT,
    ACTIONS(9463), 1,
      anon_sym_RPAREN,
    STATE(6504), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244627] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9465), 1,
      anon_sym_SEMI,
    ACTIONS(9467), 1,
      anon_sym_RBRACE,
    STATE(6505), 1,
      sym_attribute,
    STATE(6543), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244647] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6266), 1,
      anon_sym_and,
    STATE(6506), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6264), 2,
      anon_sym_in,
      sym_let_and_operator,
  [244665] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9469), 1,
      anon_sym_SEMI,
    ACTIONS(9471), 1,
      anon_sym_RBRACE,
    STATE(6507), 1,
      sym_attribute,
    STATE(6525), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244685] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(21), 1,
      sym_attribute_id,
    STATE(6508), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [244703] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9473), 1,
      aux_sym_quoted_extension_token1,
    STATE(6509), 1,
      sym_attribute,
    STATE(8043), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244723] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(9475), 1,
      anon_sym_EQ,
    STATE(6510), 1,
      sym_attribute,
    STATE(7299), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244743] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6511), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9477), 3,
      anon_sym_mutable,
      anon_sym_virtual,
      sym__lowercase_identifier,
  [244759] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8336), 1,
      anon_sym_RBRACE,
    ACTIONS(9479), 1,
      anon_sym_SEMI,
    STATE(6485), 1,
      aux_sym_record_pattern_repeat1,
    STATE(6512), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244779] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6513), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5194), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_GT,
  [244795] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4222), 1,
      anon_sym_RBRACK,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    STATE(6514), 1,
      sym_attribute,
    STATE(6558), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244815] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(9481), 1,
      anon_sym_EQ,
    STATE(6515), 1,
      sym_attribute,
    STATE(7314), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244835] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9483), 1,
      anon_sym_SEMI,
    ACTIONS(9485), 1,
      anon_sym_GT_RBRACE,
    STATE(6516), 1,
      sym_attribute,
    STATE(6529), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244855] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8021), 1,
      anon_sym_SEMI,
    ACTIONS(8023), 1,
      anon_sym_GT,
    STATE(6517), 1,
      sym_attribute,
    STATE(6560), 1,
      aux_sym_object_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244875] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6518), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5190), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_GT,
  [244891] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9271), 1,
      anon_sym_COLON2,
    ACTIONS(9487), 1,
      anon_sym_RPAREN,
    STATE(6519), 1,
      sym_attribute,
    STATE(7563), 1,
      sym__class_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244911] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(22), 1,
      sym_attribute_id,
    STATE(6520), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [244929] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(9489), 1,
      anon_sym_RBRACK,
    STATE(6181), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(6521), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244949] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9493), 1,
      anon_sym_LBRACK,
    STATE(6522), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9491), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
  [244967] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9495), 1,
      aux_sym_quoted_extension_token1,
    STATE(6523), 1,
      sym_attribute,
    STATE(7992), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [244987] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8168), 1,
      anon_sym_RBRACE,
    ACTIONS(9497), 1,
      anon_sym_SEMI,
    STATE(6485), 1,
      aux_sym_record_pattern_repeat1,
    STATE(6524), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245007] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8377), 1,
      anon_sym_RBRACE,
    ACTIONS(9499), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6525), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245027] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(23), 1,
      sym_attribute_id,
    STATE(6526), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [245045] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9501), 1,
      anon_sym_SEMI,
    ACTIONS(9503), 1,
      anon_sym_RBRACE,
    STATE(6323), 1,
      aux_sym_record_binding_pattern_repeat1,
    STATE(6527), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245065] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9505), 1,
      aux_sym_quoted_extension_token1,
    STATE(6528), 1,
      sym_attribute,
    STATE(7857), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245085] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8978), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9507), 1,
      anon_sym_SEMI,
    STATE(6464), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(6529), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245105] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9127), 1,
      sym_indexing_operator,
    ACTIONS(9509), 1,
      sym__lowercase_identifier,
    STATE(6530), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245125] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(9511), 1,
      anon_sym_RPAREN,
    STATE(6531), 1,
      sym_attribute,
    STATE(7611), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245145] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6441), 1,
      anon_sym_in,
    ACTIONS(9229), 1,
      anon_sym_and,
    STATE(6532), 1,
      sym_attribute,
    STATE(6549), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245165] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(41), 1,
      sym_attribute_id,
    STATE(6533), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [245183] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1421), 1,
      anon_sym_end,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6534), 1,
      sym_attribute,
    STATE(8063), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245203] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9513), 1,
      aux_sym_quoted_extension_token1,
    STATE(6535), 1,
      sym_attribute,
    STATE(7970), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245223] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9515), 1,
      anon_sym_SEMI,
    ACTIONS(9517), 1,
      anon_sym_RBRACE,
    STATE(6536), 1,
      sym_attribute,
    STATE(6542), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245243] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9519), 1,
      sym__lowercase_identifier,
    ACTIONS(9521), 1,
      anon_sym_LPAREN,
    STATE(2814), 1,
      sym__simple_value_pattern,
    STATE(6537), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245263] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9127), 1,
      sym_indexing_operator,
    ACTIONS(9523), 1,
      sym__lowercase_identifier,
    STATE(6538), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245283] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(25), 1,
      sym_attribute_id,
    STATE(6539), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [245301] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9525), 1,
      aux_sym_quoted_extension_token1,
    STATE(6540), 1,
      sym_attribute,
    STATE(7409), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245321] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9527), 1,
      anon_sym_SEMI,
    ACTIONS(9529), 1,
      anon_sym_RBRACE,
    STATE(6541), 1,
      sym_attribute,
    STATE(6571), 1,
      aux_sym_record_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245341] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8391), 1,
      anon_sym_RBRACE,
    ACTIONS(9531), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6542), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245361] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8430), 1,
      anon_sym_RBRACE,
    ACTIONS(9533), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245381] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7418), 1,
      sym__uppercase_identifier,
    ACTIONS(8625), 1,
      anon_sym_LPAREN,
    STATE(5181), 1,
      sym__constructor_name,
    STATE(6544), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245401] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4764), 1,
      anon_sym_STAR,
    ACTIONS(5461), 1,
      anon_sym_DASH_GT,
    STATE(6545), 1,
      sym_attribute,
    STATE(6573), 1,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245421] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9535), 1,
      anon_sym_RPAREN,
    STATE(6546), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245441] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(26), 1,
      sym_attribute_id,
    STATE(6547), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [245459] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9537), 1,
      aux_sym_quoted_extension_token1,
    STATE(6548), 1,
      sym_attribute,
    STATE(7753), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245479] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6477), 1,
      anon_sym_in,
    ACTIONS(9539), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6549), 2,
      sym_attribute,
      aux_sym_type_definition_repeat1,
  [245497] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(9), 1,
      sym_attribute_id,
    STATE(6550), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [245515] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9542), 1,
      anon_sym_LPAREN,
    ACTIONS(9544), 1,
      anon_sym_LBRACK,
    ACTIONS(9546), 1,
      anon_sym_LBRACE,
    STATE(6551), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245535] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(27), 1,
      sym_attribute_id,
    STATE(6552), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [245553] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9548), 1,
      anon_sym_SEMI,
    ACTIONS(9550), 1,
      anon_sym_RBRACE,
    STATE(6553), 1,
      sym_attribute,
    STATE(6744), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245573] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9552), 1,
      aux_sym_quoted_extension_token1,
    STATE(6554), 1,
      sym_attribute,
    STATE(7941), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245593] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9127), 1,
      sym_indexing_operator,
    ACTIONS(9554), 1,
      sym__lowercase_identifier,
    STATE(6555), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245613] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9556), 1,
      anon_sym_RPAREN,
    STATE(6556), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245633] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(9558), 1,
      anon_sym_type,
    STATE(6395), 1,
      sym_module_path,
    STATE(6557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245653] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8776), 1,
      anon_sym_RBRACK,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6558), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245673] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6441), 1,
      anon_sym_in,
    ACTIONS(9229), 1,
      anon_sym_and,
    STATE(6559), 1,
      sym_attribute,
    STATE(6665), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245693] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5088), 1,
      anon_sym_GT,
    ACTIONS(9560), 1,
      anon_sym_SEMI,
    STATE(6560), 1,
      sym_attribute,
    STATE(6658), 1,
      aux_sym_object_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245713] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(28), 1,
      sym_attribute_id,
    STATE(6561), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [245731] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9562), 1,
      aux_sym_quoted_extension_token1,
    STATE(6562), 1,
      sym_attribute,
    STATE(8162), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245751] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(9564), 1,
      anon_sym_RPAREN,
    STATE(6563), 1,
      sym_attribute,
    STATE(8299), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245771] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9566), 1,
      anon_sym_SEMI,
    ACTIONS(9568), 1,
      anon_sym_RBRACE,
    STATE(6564), 1,
      sym_attribute,
    STATE(6585), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245791] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1461), 1,
      anon_sym_end,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6565), 1,
      sym_attribute,
    STATE(7931), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245811] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(29), 1,
      sym_attribute_id,
    STATE(6566), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [245829] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9570), 1,
      aux_sym_quoted_extension_token1,
    STATE(6567), 1,
      sym_attribute,
    STATE(7296), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245849] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9572), 1,
      anon_sym_SEMI,
    ACTIONS(9574), 1,
      anon_sym_GT_RBRACE,
    STATE(6568), 1,
      sym_attribute,
    STATE(6604), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245869] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6421), 1,
      anon_sym_in,
    ACTIONS(9247), 1,
      anon_sym_and,
    STATE(6569), 1,
      sym_attribute,
    STATE(6622), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245889] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(30), 1,
      sym_attribute_id,
    STATE(6570), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [245907] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8817), 1,
      anon_sym_RBRACE,
    ACTIONS(9576), 1,
      anon_sym_SEMI,
    STATE(6382), 1,
      aux_sym_record_declaration_repeat1,
    STATE(6571), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245927] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9578), 1,
      aux_sym_quoted_extension_token1,
    STATE(6572), 1,
      sym_attribute,
    STATE(8239), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [245947] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5058), 1,
      anon_sym_DASH_GT,
    ACTIONS(8220), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6573), 2,
      sym_attribute,
      aux_sym__constructor_argument_repeat1,
  [245965] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(31), 1,
      sym_attribute_id,
    STATE(6574), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [245983] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9580), 1,
      anon_sym_RPAREN,
    STATE(6575), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246003] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9582), 1,
      aux_sym_quoted_extension_token1,
    STATE(6576), 1,
      sym_attribute,
    STATE(7689), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246023] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(32), 1,
      sym_attribute_id,
    STATE(6577), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [246041] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9584), 1,
      anon_sym_RPAREN,
    STATE(6578), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246061] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9586), 1,
      aux_sym_quoted_extension_token1,
    STATE(6579), 1,
      sym_attribute,
    STATE(7192), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246081] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(33), 1,
      sym_attribute_id,
    STATE(6580), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [246099] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9588), 1,
      aux_sym_quoted_extension_token1,
    STATE(6581), 1,
      sym_attribute,
    STATE(7486), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246119] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(34), 1,
      sym_attribute_id,
    STATE(6582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [246137] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9590), 1,
      aux_sym_quoted_extension_token1,
    STATE(6583), 1,
      sym_attribute,
    STATE(7778), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246157] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(35), 1,
      sym_attribute_id,
    STATE(6584), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [246175] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8434), 1,
      anon_sym_RBRACE,
    ACTIONS(9592), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246195] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9594), 1,
      aux_sym_quoted_extension_token1,
    STATE(6586), 1,
      sym_attribute,
    STATE(7993), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246215] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(36), 1,
      sym_attribute_id,
    STATE(6587), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [246233] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8815), 1,
      sym__lowercase_identifier,
    ACTIONS(8819), 1,
      anon_sym_mutable,
    STATE(6588), 1,
      sym_attribute,
    STATE(7130), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246253] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9596), 1,
      aux_sym_quoted_extension_token1,
    STATE(6589), 1,
      sym_attribute,
    STATE(8252), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246273] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(37), 1,
      sym_attribute_id,
    STATE(6590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [246291] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9598), 1,
      aux_sym_quoted_extension_token1,
    STATE(6591), 1,
      sym_attribute,
    STATE(7245), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246311] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(38), 1,
      sym_attribute_id,
    STATE(6592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [246329] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9600), 1,
      aux_sym_quoted_extension_token1,
    STATE(6593), 1,
      sym_attribute,
    STATE(7266), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246349] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(39), 1,
      sym_attribute_id,
    STATE(6594), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [246367] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9602), 1,
      aux_sym_quoted_extension_token1,
    STATE(6595), 1,
      sym_attribute,
    STATE(7291), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246387] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(40), 1,
      sym_attribute_id,
    STATE(6596), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [246405] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9604), 1,
      aux_sym_quoted_extension_token1,
    STATE(6597), 1,
      sym_attribute,
    STATE(7309), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246425] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_COLON_GT,
    ACTIONS(7402), 1,
      anon_sym_RPAREN,
    STATE(4953), 1,
      sym__coerced,
    STATE(6598), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246445] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6337), 1,
      sym_attribute_id,
    STATE(6599), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [246463] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(19), 1,
      sym_attribute_id,
    STATE(6600), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [246481] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6339), 1,
      sym_attribute_id,
    STATE(6601), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [246499] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8668), 1,
      anon_sym_COMMA,
    ACTIONS(9606), 1,
      anon_sym_RBRACK,
    STATE(6181), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(6602), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246519] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9608), 1,
      anon_sym_RPAREN,
    STATE(6603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246539] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8758), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9610), 1,
      anon_sym_SEMI,
    STATE(6464), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(6604), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246559] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9521), 1,
      anon_sym_LPAREN,
    ACTIONS(9612), 1,
      sym__lowercase_identifier,
    STATE(2814), 1,
      sym__simple_value_pattern,
    STATE(6605), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246579] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8847), 1,
      anon_sym_RBRACK,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6606), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246599] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8847), 1,
      anon_sym_RBRACK,
    STATE(6607), 1,
      sym_attribute,
    STATE(6620), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246619] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(9614), 1,
      aux_sym_quoted_extension_token1,
    STATE(6608), 1,
      sym_attribute,
    STATE(7795), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246639] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8969), 1,
      anon_sym_RBRACK,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6609), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246659] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9616), 1,
      anon_sym_SEMI,
    ACTIONS(9618), 1,
      anon_sym_RBRACE,
    STATE(6610), 1,
      sym_attribute,
    STATE(6621), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246679] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9620), 1,
      anon_sym_LPAREN,
    ACTIONS(9622), 1,
      anon_sym_LBRACK,
    ACTIONS(9624), 1,
      anon_sym_LBRACE,
    STATE(6611), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246699] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1627), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(9626), 1,
      anon_sym_SEMI,
    STATE(6612), 1,
      sym_attribute,
    STATE(6619), 1,
      aux_sym__sequence_expression_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246719] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4706), 1,
      anon_sym_DOT,
    STATE(6613), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9628), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [246737] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9630), 1,
      anon_sym_RPAREN,
    STATE(6614), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246757] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9632), 1,
      anon_sym_LPAREN,
    ACTIONS(9634), 1,
      anon_sym_LBRACK,
    ACTIONS(9636), 1,
      anon_sym_LBRACE,
    STATE(6615), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246777] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9127), 1,
      sym_indexing_operator,
    ACTIONS(9638), 1,
      sym__lowercase_identifier,
    STATE(6616), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246797] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9640), 1,
      anon_sym_SEMI,
    ACTIONS(9642), 1,
      anon_sym_RBRACE,
    STATE(6617), 1,
      sym_attribute,
    STATE(6654), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246817] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9644), 1,
      anon_sym_RPAREN,
    STATE(6618), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246837] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5015), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(9646), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6619), 2,
      sym_attribute,
      aux_sym__sequence_expression_content_repeat1,
  [246855] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8865), 1,
      anon_sym_RBRACK,
    STATE(6212), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6620), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246875] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8444), 1,
      anon_sym_RBRACE,
    ACTIONS(9649), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6621), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246895] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6447), 1,
      anon_sym_in,
    ACTIONS(9651), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6622), 2,
      sym_attribute,
      aux_sym_module_definition_repeat1,
  [246913] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6421), 1,
      anon_sym_in,
    ACTIONS(9247), 1,
      anon_sym_and,
    STATE(6623), 1,
      sym_attribute,
    STATE(6681), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246933] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8969), 1,
      anon_sym_RBRACK,
    STATE(6310), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6624), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246953] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9654), 1,
      anon_sym_RPAREN,
    STATE(6625), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246973] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6468), 1,
      anon_sym_in,
    ACTIONS(9340), 1,
      anon_sym_and,
    STATE(6626), 1,
      sym_attribute,
    STATE(6689), 1,
      aux_sym_class_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [246993] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6468), 1,
      anon_sym_in,
    ACTIONS(9340), 1,
      anon_sym_and,
    STATE(6627), 1,
      sym_attribute,
    STATE(6695), 1,
      aux_sym_class_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247013] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9656), 1,
      anon_sym_RPAREN,
    STATE(6398), 1,
      aux_sym__type_params_repeat1,
    STATE(6628), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247033] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(9658), 1,
      anon_sym_RPAREN,
    STATE(6629), 1,
      sym_attribute,
    STATE(7523), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247053] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6354), 1,
      sym_attribute_id,
    STATE(6630), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [247071] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9660), 1,
      anon_sym_SEMI,
    ACTIONS(9662), 1,
      anon_sym_GT_RBRACE,
    STATE(6631), 1,
      sym_attribute,
    STATE(6661), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247091] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6801), 1,
      anon_sym_in,
    ACTIONS(9664), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6632), 2,
      sym_attribute,
      aux_sym_class_definition_repeat1,
  [247109] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6808), 1,
      anon_sym_in,
    ACTIONS(9262), 1,
      anon_sym_and,
    STATE(6632), 1,
      aux_sym_class_definition_repeat1,
    STATE(6633), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247129] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6634), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9667), 3,
      anon_sym_private,
      anon_sym_virtual,
      sym__lowercase_identifier,
  [247145] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9669), 1,
      anon_sym_SEMI,
    ACTIONS(9671), 1,
      anon_sym_RBRACE,
    STATE(6635), 1,
      sym_attribute,
    STATE(6640), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247165] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(15), 1,
      sym_attribute_id,
    STATE(6636), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9167), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [247183] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9673), 1,
      anon_sym_SEMI,
    ACTIONS(9675), 1,
      anon_sym_GT_RBRACE,
    STATE(6637), 1,
      sym_attribute,
    STATE(6645), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247203] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9677), 1,
      anon_sym_RBRACK,
    STATE(6352), 1,
      aux_sym__type_params_repeat1,
    STATE(6638), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247223] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9679), 1,
      anon_sym_SEMI,
    ACTIONS(9681), 1,
      anon_sym_GT_RBRACE,
    STATE(6317), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(6639), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247243] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8497), 1,
      anon_sym_RBRACE,
    ACTIONS(9683), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6640), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247263] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8472), 1,
      anon_sym_RBRACK,
    ACTIONS(9685), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6641), 2,
      sym_attribute,
      aux_sym__sequence_binding_pattern_content_repeat1,
  [247281] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8933), 1,
      anon_sym_RBRACE,
    ACTIONS(9688), 1,
      anon_sym_SEMI,
    STATE(6382), 1,
      aux_sym_record_declaration_repeat1,
    STATE(6642), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247301] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9690), 1,
      anon_sym_RBRACK,
    STATE(6120), 1,
      aux_sym__type_params_repeat1,
    STATE(6643), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247321] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9690), 1,
      anon_sym_RBRACK,
    STATE(6375), 1,
      aux_sym__type_params_repeat1,
    STATE(6644), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247341] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8869), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9692), 1,
      anon_sym_SEMI,
    STATE(6464), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(6645), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247361] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9694), 1,
      anon_sym_SEMI,
    ACTIONS(9696), 1,
      anon_sym_RBRACE,
    STATE(6351), 1,
      aux_sym_record_pattern_repeat1,
    STATE(6646), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247381] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6647), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5532), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [247397] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9698), 1,
      anon_sym_SEMI,
    ACTIONS(9700), 1,
      anon_sym_RBRACE,
    STATE(6648), 1,
      sym_attribute,
    STATE(6653), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247417] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9702), 1,
      anon_sym_SEMI,
    ACTIONS(9705), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6649), 2,
      sym_attribute,
      aux_sym_record_binding_pattern_repeat1,
  [247435] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9127), 1,
      sym_indexing_operator,
    ACTIONS(9707), 1,
      sym__lowercase_identifier,
    STATE(6650), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247455] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6165), 1,
      aux_sym_module_binding_repeat1,
    STATE(6651), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247475] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6460), 1,
      sym_attribute_id,
    STATE(6652), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [247493] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8517), 1,
      anon_sym_RBRACE,
    ACTIONS(9709), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6653), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247513] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8567), 1,
      anon_sym_RBRACE,
    ACTIONS(9711), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6654), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247533] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9127), 1,
      sym_indexing_operator,
    ACTIONS(9713), 1,
      sym__lowercase_identifier,
    STATE(6655), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247553] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9715), 1,
      anon_sym_LPAREN,
    ACTIONS(9717), 1,
      anon_sym_LBRACK,
    ACTIONS(9719), 1,
      anon_sym_LBRACE,
    STATE(6656), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247573] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9721), 1,
      anon_sym_LPAREN,
    ACTIONS(9723), 1,
      anon_sym_LBRACK,
    ACTIONS(9725), 1,
      anon_sym_LBRACE,
    STATE(6657), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247593] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8344), 1,
      anon_sym_GT,
    ACTIONS(9727), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6658), 2,
      sym_attribute,
      aux_sym_object_type_repeat1,
  [247611] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9730), 1,
      anon_sym_RPAREN,
    STATE(6659), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247631] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6498), 1,
      sym_attribute_id,
    STATE(6660), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [247649] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8808), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9732), 1,
      anon_sym_SEMI,
    STATE(6464), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(6661), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247669] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9734), 1,
      anon_sym_LPAREN,
    ACTIONS(9736), 1,
      anon_sym_LBRACK,
    ACTIONS(9738), 1,
      anon_sym_LBRACE,
    STATE(6662), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247689] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9740), 1,
      anon_sym_LPAREN,
    ACTIONS(9742), 1,
      anon_sym_LBRACK,
    ACTIONS(9744), 1,
      anon_sym_LBRACE,
    STATE(6663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247709] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6509), 1,
      sym_attribute_id,
    STATE(6664), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [247727] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6520), 1,
      anon_sym_in,
    ACTIONS(9229), 1,
      anon_sym_and,
    STATE(6549), 1,
      aux_sym_type_definition_repeat1,
    STATE(6665), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247747] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9746), 1,
      anon_sym_SEMI,
    ACTIONS(9748), 1,
      anon_sym_RBRACE,
    STATE(6371), 1,
      aux_sym_record_binding_pattern_repeat1,
    STATE(6666), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247767] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9750), 1,
      anon_sym_LPAREN,
    ACTIONS(9752), 1,
      anon_sym_LBRACK,
    ACTIONS(9754), 1,
      anon_sym_LBRACE,
    STATE(6667), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247787] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9756), 1,
      anon_sym_SEMI,
    ACTIONS(9758), 1,
      anon_sym_RBRACE,
    STATE(6668), 1,
      sym_attribute,
    STATE(6687), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247807] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9760), 1,
      anon_sym_LPAREN,
    ACTIONS(9762), 1,
      anon_sym_LBRACK,
    ACTIONS(9764), 1,
      anon_sym_LBRACE,
    STATE(6669), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247827] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(9766), 1,
      anon_sym_RPAREN,
    STATE(6670), 1,
      sym_attribute,
    STATE(8226), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247847] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6523), 1,
      sym_attribute_id,
    STATE(6671), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [247865] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9768), 1,
      anon_sym_LPAREN,
    ACTIONS(9770), 1,
      anon_sym_LBRACK,
    ACTIONS(9772), 1,
      anon_sym_LBRACE,
    STATE(6672), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247885] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9774), 1,
      anon_sym_SEMI,
    ACTIONS(9776), 1,
      anon_sym_GT_RBRACE,
    STATE(6673), 1,
      sym_attribute,
    STATE(6702), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247905] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9778), 1,
      anon_sym_LPAREN,
    ACTIONS(9780), 1,
      anon_sym_LBRACK,
    ACTIONS(9782), 1,
      anon_sym_LBRACE,
    STATE(6674), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247925] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(9784), 1,
      anon_sym_type,
    STATE(6395), 1,
      sym_module_path,
    STATE(6675), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247945] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6528), 1,
      sym_attribute_id,
    STATE(6676), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [247963] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9786), 1,
      anon_sym_LPAREN,
    ACTIONS(9788), 1,
      anon_sym_LBRACK,
    ACTIONS(9790), 1,
      anon_sym_LBRACE,
    STATE(6677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [247983] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9792), 1,
      anon_sym_LPAREN,
    ACTIONS(9794), 1,
      anon_sym_LBRACK,
    ACTIONS(9796), 1,
      anon_sym_LBRACE,
    STATE(6678), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248003] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6535), 1,
      sym_attribute_id,
    STATE(6679), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248021] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9798), 1,
      anon_sym_SEMI,
    ACTIONS(9800), 1,
      anon_sym_RBRACE,
    STATE(6680), 1,
      sym_attribute,
    STATE(6745), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248041] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6498), 1,
      anon_sym_in,
    ACTIONS(9247), 1,
      anon_sym_and,
    STATE(6622), 1,
      aux_sym_module_definition_repeat1,
    STATE(6681), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248061] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9802), 1,
      anon_sym_LPAREN,
    ACTIONS(9804), 1,
      anon_sym_LBRACK,
    ACTIONS(9806), 1,
      anon_sym_LBRACE,
    STATE(6682), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248081] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9808), 1,
      anon_sym_LPAREN,
    ACTIONS(9810), 1,
      anon_sym_LBRACK,
    ACTIONS(9812), 1,
      anon_sym_LBRACE,
    STATE(6683), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248101] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6540), 1,
      sym_attribute_id,
    STATE(6684), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248119] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9814), 1,
      anon_sym_LPAREN,
    ACTIONS(9816), 1,
      anon_sym_LBRACK,
    ACTIONS(9818), 1,
      anon_sym_LBRACE,
    STATE(6685), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248139] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9820), 1,
      anon_sym_LPAREN,
    ACTIONS(9822), 1,
      anon_sym_LBRACK,
    ACTIONS(9824), 1,
      anon_sym_LBRACE,
    STATE(6686), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248159] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8590), 1,
      anon_sym_RBRACE,
    ACTIONS(9826), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6687), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248179] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6548), 1,
      sym_attribute_id,
    STATE(6688), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248197] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6524), 1,
      anon_sym_in,
    ACTIONS(9828), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6689), 2,
      sym_attribute,
      aux_sym_class_type_definition_repeat1,
  [248215] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(709), 1,
      anon_sym_end,
    ACTIONS(5608), 1,
      anon_sym_PERCENT,
    STATE(6690), 1,
      sym_attribute,
    STATE(8268), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248235] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4520), 1,
      sym_attribute_id,
    STATE(6691), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9831), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248253] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9833), 1,
      anon_sym_LPAREN,
    ACTIONS(9835), 1,
      anon_sym_LBRACK,
    ACTIONS(9837), 1,
      anon_sym_LBRACE,
    STATE(6692), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248273] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9839), 1,
      anon_sym_LPAREN,
    ACTIONS(9841), 1,
      anon_sym_LBRACK,
    ACTIONS(9843), 1,
      anon_sym_LBRACE,
    STATE(6693), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248293] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6554), 1,
      sym_attribute_id,
    STATE(6694), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248311] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6535), 1,
      anon_sym_in,
    ACTIONS(9340), 1,
      anon_sym_and,
    STATE(6689), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(6695), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248331] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9845), 1,
      anon_sym_LPAREN,
    ACTIONS(9847), 1,
      anon_sym_LBRACK,
    ACTIONS(9849), 1,
      anon_sym_LBRACE,
    STATE(6696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248351] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9851), 1,
      anon_sym_LPAREN,
    ACTIONS(9853), 1,
      anon_sym_LBRACK,
    ACTIONS(9855), 1,
      anon_sym_LBRACE,
    STATE(6697), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248371] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6562), 1,
      sym_attribute_id,
    STATE(6698), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248389] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9857), 1,
      anon_sym_SEMI,
    ACTIONS(9859), 1,
      anon_sym_RBRACE,
    STATE(6699), 1,
      sym_attribute,
    STATE(6747), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248409] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9861), 1,
      anon_sym_LPAREN,
    ACTIONS(9863), 1,
      anon_sym_LBRACK,
    ACTIONS(9865), 1,
      anon_sym_LBRACE,
    STATE(6700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248429] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9867), 1,
      anon_sym_LPAREN,
    ACTIONS(9869), 1,
      anon_sym_LBRACK,
    ACTIONS(9871), 1,
      anon_sym_LBRACE,
    STATE(6701), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248449] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9006), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9873), 1,
      anon_sym_SEMI,
    STATE(6464), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(6702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248469] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6567), 1,
      sym_attribute_id,
    STATE(6703), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248487] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9875), 1,
      anon_sym_LPAREN,
    ACTIONS(9877), 1,
      anon_sym_LBRACK,
    ACTIONS(9879), 1,
      anon_sym_LBRACE,
    STATE(6704), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248507] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9881), 1,
      anon_sym_RPAREN,
    STATE(6705), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248527] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9883), 1,
      anon_sym_LPAREN,
    ACTIONS(9885), 1,
      anon_sym_LBRACK,
    ACTIONS(9887), 1,
      anon_sym_LBRACE,
    STATE(6706), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248547] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6572), 1,
      sym_attribute_id,
    STATE(6707), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248565] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9889), 1,
      anon_sym_LPAREN,
    ACTIONS(9891), 1,
      anon_sym_LBRACK,
    ACTIONS(9893), 1,
      anon_sym_LBRACE,
    STATE(6708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248585] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6576), 1,
      sym_attribute_id,
    STATE(6709), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248603] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9895), 1,
      anon_sym_LPAREN,
    ACTIONS(9897), 1,
      anon_sym_LBRACK,
    ACTIONS(9899), 1,
      anon_sym_LBRACE,
    STATE(6710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248623] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6579), 1,
      sym_attribute_id,
    STATE(6711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248641] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9901), 1,
      anon_sym_LPAREN,
    ACTIONS(9903), 1,
      anon_sym_LBRACK,
    ACTIONS(9905), 1,
      anon_sym_LBRACE,
    STATE(6712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248661] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6581), 1,
      sym_attribute_id,
    STATE(6713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248679] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9907), 1,
      anon_sym_LPAREN,
    ACTIONS(9909), 1,
      anon_sym_LBRACK,
    ACTIONS(9911), 1,
      anon_sym_LBRACE,
    STATE(6714), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248699] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6583), 1,
      sym_attribute_id,
    STATE(6715), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248717] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9913), 1,
      anon_sym_LPAREN,
    ACTIONS(9915), 1,
      anon_sym_LBRACK,
    ACTIONS(9917), 1,
      anon_sym_LBRACE,
    STATE(6716), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248737] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6586), 1,
      sym_attribute_id,
    STATE(6717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248755] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9919), 1,
      anon_sym_LPAREN,
    ACTIONS(9921), 1,
      anon_sym_LBRACK,
    ACTIONS(9923), 1,
      anon_sym_LBRACE,
    STATE(6718), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248775] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6589), 1,
      sym_attribute_id,
    STATE(6719), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248793] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9925), 1,
      anon_sym_LPAREN,
    ACTIONS(9927), 1,
      anon_sym_LBRACK,
    ACTIONS(9929), 1,
      anon_sym_LBRACE,
    STATE(6720), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248813] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6591), 1,
      sym_attribute_id,
    STATE(6721), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248831] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6593), 1,
      sym_attribute_id,
    STATE(6722), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248849] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6595), 1,
      sym_attribute_id,
    STATE(6723), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248867] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6597), 1,
      sym_attribute_id,
    STATE(6724), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248885] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6608), 1,
      sym_attribute_id,
    STATE(6725), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9159), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [248903] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9931), 1,
      anon_sym_SEMI,
    ACTIONS(9933), 1,
      anon_sym_RBRACE,
    STATE(6726), 1,
      sym_attribute,
    STATE(6740), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248923] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9935), 1,
      anon_sym_RBRACK,
    STATE(6727), 1,
      sym_attribute,
    STATE(6731), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248943] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9127), 1,
      sym_indexing_operator,
    ACTIONS(9937), 1,
      sym__lowercase_identifier,
    STATE(6728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248963] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(9939), 1,
      anon_sym_RPAREN,
    STATE(6729), 1,
      sym_attribute,
    STATE(7355), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [248983] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9941), 1,
      anon_sym_RBRACK,
    STATE(6730), 1,
      sym_attribute,
    STATE(6735), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249003] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9943), 1,
      anon_sym_RBRACK,
    STATE(6120), 1,
      aux_sym__type_params_repeat1,
    STATE(6731), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249023] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9943), 1,
      anon_sym_RBRACK,
    STATE(6732), 1,
      sym_attribute,
    STATE(6737), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249043] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9945), 1,
      anon_sym_SEMI,
    ACTIONS(9947), 1,
      anon_sym_GT_RBRACE,
    STATE(6448), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(6733), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249063] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9127), 1,
      sym_indexing_operator,
    ACTIONS(9949), 1,
      sym__lowercase_identifier,
    STATE(6734), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249083] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9951), 1,
      anon_sym_RBRACK,
    STATE(6120), 1,
      aux_sym__type_params_repeat1,
    STATE(6735), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249103] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9951), 1,
      anon_sym_RBRACK,
    STATE(6736), 1,
      sym_attribute,
    STATE(6739), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249123] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9953), 1,
      anon_sym_RBRACK,
    STATE(6120), 1,
      aux_sym__type_params_repeat1,
    STATE(6737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249143] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9955), 1,
      anon_sym_RBRACK,
    STATE(6120), 1,
      aux_sym__type_params_repeat1,
    STATE(6738), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249163] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9201), 1,
      anon_sym_COMMA,
    ACTIONS(9957), 1,
      anon_sym_RBRACK,
    STATE(6120), 1,
      aux_sym__type_params_repeat1,
    STATE(6739), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249183] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8348), 1,
      anon_sym_RBRACE,
    ACTIONS(9959), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6740), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249203] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9961), 1,
      anon_sym_SEMI,
    ACTIONS(9963), 1,
      anon_sym_RBRACE,
    STATE(6409), 1,
      aux_sym_record_binding_pattern_repeat1,
    STATE(6741), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249223] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4906), 1,
      anon_sym_LPAREN,
    STATE(4078), 1,
      sym_module_parameter,
    STATE(6251), 1,
      aux_sym_module_binding_repeat1,
    STATE(6742), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249243] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5190), 1,
      anon_sym_EQ,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(7334), 1,
      anon_sym_RPAREN,
    STATE(6743), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249263] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8482), 1,
      anon_sym_RBRACE,
    ACTIONS(9965), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249283] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8371), 1,
      anon_sym_RBRACE,
    ACTIONS(9967), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6745), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249303] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    ACTIONS(9969), 1,
      anon_sym_RPAREN,
    STATE(6746), 1,
      sym_attribute,
    STATE(7273), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249323] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8592), 1,
      anon_sym_RBRACE,
    ACTIONS(9971), 1,
      anon_sym_SEMI,
    STATE(6454), 1,
      aux_sym_record_expression_repeat1,
    STATE(6747), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249343] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(6748), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5180), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [249359] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4198), 1,
      anon_sym_RBRACK,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    STATE(6410), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6749), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249379] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6750), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8670), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [249395] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8724), 1,
      anon_sym_with,
    ACTIONS(8888), 1,
      anon_sym_DASH_GT,
    ACTIONS(9973), 1,
      anon_sym_RPAREN,
    STATE(6751), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249415] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9975), 1,
      anon_sym_SEMI,
    ACTIONS(9977), 1,
      anon_sym_RBRACE,
    STATE(6307), 1,
      aux_sym_record_expression_repeat1,
    STATE(6752), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249435] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9979), 1,
      anon_sym_SEMI,
    ACTIONS(9981), 1,
      anon_sym_RBRACE,
    STATE(6512), 1,
      aux_sym_record_pattern_repeat1,
    STATE(6753), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249455] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4621), 1,
      sym_module_binding,
    STATE(6754), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8342), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [249473] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4904), 1,
      sym_module_binding,
    STATE(6755), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8342), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [249491] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(4907), 1,
      sym_module_binding,
    STATE(6756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8342), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [249509] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4343), 1,
      sym__uppercase_identifier,
    ACTIONS(9983), 1,
      anon_sym_type,
    STATE(6395), 1,
      sym_module_path,
    STATE(6757), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249529] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9985), 1,
      anon_sym_SEMI,
    ACTIONS(9987), 1,
      anon_sym_GT_RBRACE,
    STATE(6308), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(6758), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249549] = 6,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8774), 1,
      anon_sym_PIPE,
    ACTIONS(8776), 1,
      anon_sym_RBRACK,
    STATE(6606), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(6759), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249569] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9989), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [249584] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9991), 1,
      anon_sym_COLON2,
    STATE(5188), 1,
      sym__typed,
    STATE(6761), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249601] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9993), 1,
      anon_sym_RBRACK,
    ACTIONS(9995), 1,
      anon_sym_SEMI,
    STATE(6762), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249618] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9993), 1,
      anon_sym_RBRACE,
    ACTIONS(9997), 1,
      anon_sym_SEMI,
    STATE(6763), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249635] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9999), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [249650] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6765), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10001), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [249665] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10003), 1,
      anon_sym_RPAREN,
    ACTIONS(10005), 1,
      anon_sym_LT_DASH,
    STATE(6766), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249682] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4214), 1,
      anon_sym_RBRACK,
    ACTIONS(8670), 1,
      anon_sym_PIPE,
    STATE(6767), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249699] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(10007), 1,
      sym__lowercase_identifier,
    STATE(6768), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249716] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4122), 1,
      anon_sym_RPAREN,
    ACTIONS(4126), 1,
      anon_sym_COLON_COLON,
    STATE(6769), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249733] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10009), 1,
      aux_sym_type_variable_token1,
    STATE(5015), 1,
      sym_type_variable,
    STATE(6770), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249750] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10011), 1,
      anon_sym_do,
    STATE(2703), 1,
      sym_do_clause,
    STATE(6771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249767] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6772), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10013), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [249782] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10015), 1,
      anon_sym_then,
    STATE(3250), 1,
      sym_then_clause,
    STATE(6773), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249799] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6774), 1,
      sym_attribute,
    STATE(7220), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249816] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10017), 1,
      anon_sym_then,
    STATE(3959), 1,
      sym_then_clause,
    STATE(6775), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249833] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6776), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10019), 2,
      anon_sym_to,
      anon_sym_downto,
  [249848] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6777), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10021), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [249863] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10023), 1,
      anon_sym_RPAREN,
    ACTIONS(10025), 1,
      anon_sym_LT_DASH,
    STATE(6778), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249880] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6779), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10027), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [249895] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(7825), 1,
      sym__lowercase_identifier,
    STATE(6780), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249912] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6781), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10029), 2,
      anon_sym_SEMI,
      anon_sym_GT,
  [249927] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10031), 1,
      sym__lowercase_identifier,
    ACTIONS(10033), 1,
      anon_sym_LBRACK,
    STATE(6782), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249944] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(10035), 1,
      sym__lowercase_identifier,
    STATE(6783), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249961] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10037), 1,
      anon_sym_then,
    STATE(2794), 1,
      sym_then_clause,
    STATE(6784), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249978] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(10039), 1,
      sym__lowercase_identifier,
    STATE(6785), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [249995] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(10041), 1,
      anon_sym_EQ,
    STATE(6786), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250012] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10043), 1,
      anon_sym_RPAREN,
    ACTIONS(10045), 1,
      anon_sym_LT_DASH,
    STATE(6787), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250029] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10047), 1,
      anon_sym_RPAREN,
    ACTIONS(10049), 1,
      anon_sym_LT_DASH,
    STATE(6788), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250046] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6789), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10051), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [250061] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10053), 1,
      sym__lowercase_identifier,
    ACTIONS(10055), 1,
      anon_sym_LPAREN,
    STATE(6790), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250078] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10057), 1,
      anon_sym_then,
    STATE(3767), 1,
      sym_then_clause,
    STATE(6791), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250095] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10059), 1,
      anon_sym_RPAREN,
    ACTIONS(10061), 1,
      anon_sym_LT_DASH,
    STATE(6792), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250112] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8132), 1,
      anon_sym_RBRACE,
    ACTIONS(10063), 1,
      anon_sym_SEMI,
    STATE(6793), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250129] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10065), 1,
      anon_sym_RPAREN,
    ACTIONS(10067), 1,
      anon_sym_LT_DASH,
    STATE(6794), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250146] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10069), 1,
      sym__lowercase_identifier,
    ACTIONS(10071), 1,
      anon_sym_LPAREN,
    STATE(6795), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250163] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10073), 1,
      sym__lowercase_identifier,
    ACTIONS(10075), 1,
      anon_sym_LPAREN,
    STATE(6796), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250180] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10077), 1,
      anon_sym_RPAREN,
    ACTIONS(10079), 1,
      anon_sym_LT_DASH,
    STATE(6797), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250197] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10081), 1,
      anon_sym_RPAREN,
    ACTIONS(10083), 1,
      anon_sym_LT_DASH,
    STATE(6798), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250214] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10085), 1,
      anon_sym_then,
    STATE(3216), 1,
      sym_then_clause,
    STATE(6799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250231] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10087), 1,
      anon_sym_RPAREN,
    ACTIONS(10089), 1,
      anon_sym_LT_DASH,
    STATE(6800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250248] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6801), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10091), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [250263] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6802), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10093), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [250278] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10095), 1,
      anon_sym_then,
    STATE(2919), 1,
      sym_then_clause,
    STATE(6803), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250295] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6804), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10097), 2,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
  [250310] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10099), 1,
      sym__lowercase_identifier,
    ACTIONS(10101), 1,
      anon_sym_LPAREN,
    STATE(6805), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250327] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(10103), 1,
      sym__lowercase_identifier,
    STATE(6806), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250344] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6807), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10105), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [250359] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9991), 1,
      anon_sym_COLON2,
    STATE(5177), 1,
      sym__typed,
    STATE(6808), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250376] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10107), 1,
      anon_sym_then,
    STATE(2717), 1,
      sym_then_clause,
    STATE(6809), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250393] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6810), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10109), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [250408] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(713), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4140), 1,
      anon_sym_RPAREN,
    STATE(6811), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250425] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    STATE(5015), 1,
      sym_type_variable,
    STATE(6812), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250442] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10111), 1,
      sym__lowercase_identifier,
    ACTIONS(10113), 1,
      anon_sym_LBRACK,
    STATE(6813), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250459] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10115), 1,
      anon_sym_COLON2,
    STATE(4474), 1,
      sym__maybe_polymorphic_typed,
    STATE(6814), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250476] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10017), 1,
      anon_sym_then,
    STATE(3939), 1,
      sym_then_clause,
    STATE(6815), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250493] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10117), 1,
      anon_sym_RPAREN,
    ACTIONS(10119), 1,
      anon_sym_LT_DASH,
    STATE(6816), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250510] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2736), 1,
      anon_sym_COLON2,
    STATE(5190), 1,
      sym__maybe_polymorphic_typed,
    STATE(6817), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250527] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2736), 1,
      anon_sym_COLON2,
    STATE(5179), 1,
      sym__maybe_polymorphic_typed,
    STATE(6818), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250544] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10121), 1,
      anon_sym_then,
    STATE(3336), 1,
      sym_then_clause,
    STATE(6819), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250561] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10123), 1,
      anon_sym_do,
    STATE(2739), 1,
      sym_do_clause,
    STATE(6820), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250578] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10125), 1,
      sym__lowercase_identifier,
    ACTIONS(10127), 1,
      anon_sym_LPAREN,
    STATE(6821), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250595] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10129), 1,
      anon_sym_RPAREN,
    ACTIONS(10131), 1,
      anon_sym_LT_DASH,
    STATE(6822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250612] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10011), 1,
      anon_sym_do,
    STATE(2694), 1,
      sym_do_clause,
    STATE(6823), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250629] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10133), 1,
      sym__lowercase_identifier,
    ACTIONS(10135), 1,
      anon_sym_LPAREN,
    STATE(6824), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250646] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10121), 1,
      anon_sym_then,
    STATE(3275), 1,
      sym_then_clause,
    STATE(6825), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250663] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10137), 1,
      anon_sym_then,
    STATE(2877), 1,
      sym_then_clause,
    STATE(6826), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250680] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10139), 1,
      anon_sym_then,
    STATE(2655), 1,
      sym_then_clause,
    STATE(6827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250697] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(10141), 1,
      sym__lowercase_identifier,
    STATE(6828), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250714] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6829), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10143), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [250729] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10037), 1,
      anon_sym_then,
    STATE(2820), 1,
      sym_then_clause,
    STATE(6830), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250746] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10145), 1,
      sym__lowercase_identifier,
    ACTIONS(10147), 1,
      anon_sym_LPAREN,
    STATE(6831), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250763] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8136), 1,
      anon_sym_RBRACE,
    ACTIONS(10149), 1,
      anon_sym_SEMI,
    STATE(6832), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250780] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6833), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9382), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [250795] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10057), 1,
      anon_sym_then,
    STATE(3766), 1,
      sym_then_clause,
    STATE(6834), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250812] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6835), 1,
      sym_attribute,
    STATE(7720), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250829] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10085), 1,
      anon_sym_then,
    STATE(3227), 1,
      sym_then_clause,
    STATE(6836), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250846] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10107), 1,
      anon_sym_then,
    STATE(2722), 1,
      sym_then_clause,
    STATE(6837), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250863] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6838), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10151), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [250878] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10153), 1,
      anon_sym_DOT,
    STATE(6839), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250895] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8176), 1,
      anon_sym_RBRACE,
    ACTIONS(10155), 1,
      anon_sym_SEMI,
    STATE(6840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250912] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6841), 1,
      sym_attribute,
    STATE(7534), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250929] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(7831), 1,
      sym__lowercase_identifier,
    STATE(6842), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250946] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10157), 1,
      anon_sym_RPAREN,
    ACTIONS(10159), 1,
      anon_sym_LT_DASH,
    STATE(6843), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250963] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6844), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9399), 2,
      anon_sym_SEMI,
      anon_sym_GT_RBRACE,
  [250978] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10161), 1,
      anon_sym_EQ,
    ACTIONS(10163), 1,
      anon_sym_RPAREN,
    STATE(6845), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [250995] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10165), 1,
      anon_sym_RPAREN,
    ACTIONS(10167), 1,
      anon_sym_LT_DASH,
    STATE(6846), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251012] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(10169), 1,
      sym__lowercase_identifier,
    STATE(6847), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251029] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10171), 1,
      sym__lowercase_identifier,
    ACTIONS(10173), 1,
      anon_sym_LPAREN,
    STATE(6848), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251046] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(10175), 1,
      sym__lowercase_identifier,
    STATE(6849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251063] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6850), 1,
      sym_attribute,
    STATE(7856), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251080] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6851), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10177), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [251095] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6852), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10179), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [251110] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10181), 1,
      anon_sym_COLON2,
    STATE(6853), 1,
      sym_attribute,
    STATE(8137), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251127] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6854), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10183), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [251142] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(10185), 1,
      sym__lowercase_identifier,
    STATE(6855), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251159] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10187), 1,
      anon_sym_RPAREN,
    ACTIONS(10189), 1,
      anon_sym_SEMI,
    STATE(6856), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251176] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10187), 1,
      anon_sym_RBRACK,
    ACTIONS(10191), 1,
      anon_sym_SEMI,
    STATE(6857), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251193] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10187), 1,
      anon_sym_RBRACE,
    ACTIONS(10193), 1,
      anon_sym_SEMI,
    STATE(6858), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251210] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10195), 1,
      anon_sym_DOT,
    STATE(6859), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251227] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10197), 1,
      anon_sym_RPAREN,
    ACTIONS(10199), 1,
      anon_sym_LT_DASH,
    STATE(6860), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251244] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8148), 1,
      anon_sym_RBRACE,
    ACTIONS(10201), 1,
      anon_sym_SEMI,
    STATE(6861), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251261] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10203), 1,
      anon_sym_DOT,
    STATE(6862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251278] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10205), 1,
      anon_sym_SEMI,
    ACTIONS(10207), 1,
      anon_sym_RBRACE,
    STATE(6863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251295] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10209), 1,
      anon_sym_RPAREN,
    ACTIONS(10211), 1,
      anon_sym_LT_DASH,
    STATE(6864), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251312] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6865), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10213), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [251327] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6866), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9445), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [251342] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10123), 1,
      anon_sym_do,
    STATE(2708), 1,
      sym_do_clause,
    STATE(6867), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251359] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(9713), 1,
      sym__lowercase_identifier,
    STATE(6868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251376] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10215), 1,
      anon_sym_RPAREN,
    ACTIONS(10217), 1,
      anon_sym_LT_DASH,
    STATE(6869), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251393] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10219), 1,
      anon_sym_then,
    STATE(3277), 1,
      sym_then_clause,
    STATE(6870), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251410] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(10221), 1,
      anon_sym_EQ,
    STATE(6871), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251427] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(10223), 1,
      sym__lowercase_identifier,
    STATE(6872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251444] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6873), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10225), 2,
      anon_sym_to,
      anon_sym_downto,
  [251459] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10227), 1,
      anon_sym_RPAREN,
    ACTIONS(10229), 1,
      anon_sym_LT_DASH,
    STATE(6874), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251476] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10219), 1,
      anon_sym_then,
    STATE(3348), 1,
      sym_then_clause,
    STATE(6875), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251493] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    STATE(6876), 1,
      sym_attribute,
    STATE(7337), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251510] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10231), 1,
      sym__lowercase_identifier,
    ACTIONS(10233), 1,
      anon_sym_LPAREN,
    STATE(6877), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251527] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6878), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10235), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [251542] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10237), 1,
      sym__uppercase_identifier,
    STATE(5258), 1,
      sym_extended_module_path,
    STATE(6879), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251559] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10239), 1,
      anon_sym_RPAREN,
    ACTIONS(10241), 1,
      anon_sym_LT_DASH,
    STATE(6880), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251576] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10243), 1,
      anon_sym_RPAREN,
    ACTIONS(10245), 1,
      anon_sym_LT_DASH,
    STATE(6881), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251593] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10247), 1,
      anon_sym_DOT,
    STATE(6882), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251610] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8949), 1,
      anon_sym_COLON2,
    STATE(6883), 1,
      sym_attribute,
    STATE(7773), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251627] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6884), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10249), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [251642] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6885), 1,
      sym_attribute,
    STATE(7263), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251659] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10251), 1,
      sym__lowercase_identifier,
    ACTIONS(10253), 1,
      anon_sym_LPAREN,
    STATE(6886), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251676] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10255), 1,
      anon_sym_SEMI,
    ACTIONS(10257), 1,
      anon_sym_RBRACE,
    STATE(6887), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251693] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10259), 1,
      anon_sym_do,
    STATE(2819), 1,
      sym_do_clause,
    STATE(6888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251710] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10261), 1,
      anon_sym_RPAREN,
    ACTIONS(10263), 1,
      anon_sym_LT_DASH,
    STATE(6889), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251727] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6890), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10265), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [251742] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6891), 1,
      sym_attribute,
    STATE(7875), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251759] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10181), 1,
      anon_sym_COLON2,
    STATE(6892), 1,
      sym_attribute,
    STATE(8077), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251776] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6893), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10267), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [251791] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10269), 1,
      anon_sym_RPAREN,
    ACTIONS(10271), 1,
      anon_sym_SEMI,
    STATE(6894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251808] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10269), 1,
      anon_sym_RBRACK,
    ACTIONS(10273), 1,
      anon_sym_SEMI,
    STATE(6895), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251825] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10269), 1,
      anon_sym_RBRACE,
    ACTIONS(10275), 1,
      anon_sym_SEMI,
    STATE(6896), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251842] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10277), 1,
      anon_sym_DOT,
    STATE(6897), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251859] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(10279), 1,
      anon_sym_EQ,
    STATE(6898), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251876] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10281), 1,
      anon_sym_then,
    STATE(3837), 1,
      sym_then_clause,
    STATE(6899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251893] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10259), 1,
      anon_sym_do,
    STATE(2821), 1,
      sym_do_clause,
    STATE(6900), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251910] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10283), 1,
      anon_sym_DOT,
    STATE(6901), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251927] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10285), 1,
      anon_sym_DOT,
    STATE(6902), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251944] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6903), 1,
      sym_attribute,
    STATE(7586), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251961] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10287), 1,
      anon_sym_then,
    STATE(3799), 1,
      sym_then_clause,
    STATE(6904), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251978] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10289), 1,
      anon_sym_SEMI,
    ACTIONS(10291), 1,
      anon_sym_RBRACE,
    STATE(6905), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [251995] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6906), 1,
      sym_attribute,
    STATE(8163), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252012] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4222), 1,
      anon_sym_RBRACK,
    ACTIONS(8670), 1,
      anon_sym_PIPE,
    STATE(6907), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252029] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6908), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10293), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [252044] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10295), 1,
      anon_sym_RPAREN,
    ACTIONS(10297), 1,
      anon_sym_SEMI,
    STATE(6909), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252061] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10295), 1,
      anon_sym_RBRACK,
    ACTIONS(10299), 1,
      anon_sym_SEMI,
    STATE(6910), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252078] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10295), 1,
      anon_sym_RBRACE,
    ACTIONS(10301), 1,
      anon_sym_SEMI,
    STATE(6911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252095] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10303), 1,
      anon_sym_DOT,
    STATE(6912), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252112] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6913), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10305), 2,
      anon_sym_RBRACK,
      aux_sym_tag_token1,
  [252127] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10307), 1,
      aux_sym_type_variable_token1,
    STATE(3240), 1,
      sym_type_variable,
    STATE(6914), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252144] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10309), 1,
      anon_sym_DOT,
    STATE(6915), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252161] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6916), 1,
      sym_attribute,
    STATE(7342), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252178] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10139), 1,
      anon_sym_then,
    STATE(2702), 1,
      sym_then_clause,
    STATE(6917), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252195] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6918), 1,
      sym_attribute,
    STATE(8096), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252212] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6919), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10311), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [252227] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10313), 1,
      anon_sym_RPAREN,
    ACTIONS(10315), 1,
      anon_sym_SEMI,
    STATE(6920), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252244] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10313), 1,
      anon_sym_RBRACK,
    ACTIONS(10317), 1,
      anon_sym_SEMI,
    STATE(6921), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252261] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10313), 1,
      anon_sym_RBRACE,
    ACTIONS(10319), 1,
      anon_sym_SEMI,
    STATE(6922), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252278] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(7829), 1,
      sym__lowercase_identifier,
    STATE(6923), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252295] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10321), 1,
      sym__lowercase_identifier,
    ACTIONS(10323), 1,
      anon_sym_LPAREN,
    STATE(6924), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252312] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10325), 1,
      anon_sym_DOT,
    STATE(6925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252329] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10327), 1,
      anon_sym_DOT,
    STATE(6926), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252346] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(10329), 1,
      sym__lowercase_identifier,
    STATE(6927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252363] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6928), 1,
      sym_attribute,
    STATE(8088), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252380] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10331), 1,
      anon_sym_RPAREN,
    ACTIONS(10333), 1,
      anon_sym_SEMI,
    STATE(6929), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252397] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10331), 1,
      anon_sym_RBRACK,
    ACTIONS(10335), 1,
      anon_sym_SEMI,
    STATE(6930), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252414] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10331), 1,
      anon_sym_RBRACE,
    ACTIONS(10337), 1,
      anon_sym_SEMI,
    STATE(6931), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252431] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10339), 1,
      anon_sym_DOT,
    STATE(6932), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252448] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6933), 1,
      sym_attribute,
    STATE(8106), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252465] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8196), 1,
      anon_sym_RBRACE,
    ACTIONS(10341), 1,
      anon_sym_SEMI,
    STATE(6934), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252482] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10343), 1,
      anon_sym_RPAREN,
    ACTIONS(10345), 1,
      anon_sym_SEMI,
    STATE(6935), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252499] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10343), 1,
      anon_sym_RBRACK,
    ACTIONS(10347), 1,
      anon_sym_SEMI,
    STATE(6936), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252516] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10343), 1,
      anon_sym_RBRACE,
    ACTIONS(10349), 1,
      anon_sym_SEMI,
    STATE(6937), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252533] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10351), 1,
      anon_sym_RPAREN,
    ACTIONS(10353), 1,
      anon_sym_LT_DASH,
    STATE(6938), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252550] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10355), 1,
      anon_sym_DOT,
    STATE(6939), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252567] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10357), 1,
      anon_sym_RPAREN,
    ACTIONS(10359), 1,
      anon_sym_LT_DASH,
    STATE(6940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252584] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6941), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8344), 2,
      anon_sym_SEMI,
      anon_sym_GT,
  [252599] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10361), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [252614] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6943), 1,
      sym_attribute,
    STATE(7441), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252631] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10363), 1,
      anon_sym_RPAREN,
    ACTIONS(10365), 1,
      anon_sym_SEMI,
    STATE(6944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252648] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10363), 1,
      anon_sym_RBRACK,
    ACTIONS(10367), 1,
      anon_sym_SEMI,
    STATE(6945), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252665] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10363), 1,
      anon_sym_RBRACE,
    ACTIONS(10369), 1,
      anon_sym_SEMI,
    STATE(6946), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252682] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6477), 2,
      anon_sym_and,
      anon_sym_in,
  [252697] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10371), 1,
      anon_sym_COLON2,
    STATE(4383), 1,
      sym__maybe_polymorphic_typed,
    STATE(6948), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252714] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6949), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10373), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [252729] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6950), 1,
      sym_attribute,
    STATE(7785), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252746] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10375), 1,
      anon_sym_RPAREN,
    ACTIONS(10377), 1,
      anon_sym_LT_DASH,
    STATE(6951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252763] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10379), 1,
      sym__lowercase_identifier,
    ACTIONS(10381), 1,
      sym__uppercase_identifier,
    STATE(6952), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252780] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10383), 1,
      anon_sym_RPAREN,
    ACTIONS(10385), 1,
      anon_sym_COLON_GT,
    STATE(6953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252797] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10387), 1,
      anon_sym_RPAREN,
    ACTIONS(10389), 1,
      anon_sym_SEMI,
    STATE(6954), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252814] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10387), 1,
      anon_sym_RBRACK,
    ACTIONS(10391), 1,
      anon_sym_SEMI,
    STATE(6955), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252831] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10387), 1,
      anon_sym_RBRACE,
    ACTIONS(10393), 1,
      anon_sym_SEMI,
    STATE(6956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252848] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10281), 1,
      anon_sym_then,
    STATE(3848), 1,
      sym_then_clause,
    STATE(6957), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252865] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10395), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [252880] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6959), 1,
      sym_attribute,
    STATE(7959), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252897] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10397), 1,
      sym__lowercase_identifier,
    ACTIONS(10399), 1,
      sym__uppercase_identifier,
    STATE(6960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252914] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10401), 1,
      anon_sym_RPAREN,
    ACTIONS(10403), 1,
      anon_sym_SEMI,
    STATE(6961), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252931] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10401), 1,
      anon_sym_RBRACK,
    ACTIONS(10405), 1,
      anon_sym_SEMI,
    STATE(6962), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252948] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10401), 1,
      anon_sym_RBRACE,
    ACTIONS(10407), 1,
      anon_sym_SEMI,
    STATE(6963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252965] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6964), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10409), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [252980] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(10411), 1,
      sym__lowercase_identifier,
    STATE(6965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [252997] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10071), 1,
      anon_sym_LPAREN,
    ACTIONS(10413), 1,
      sym__lowercase_identifier,
    STATE(6966), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253014] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6967), 1,
      sym_attribute,
    STATE(8231), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253031] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10415), 1,
      anon_sym_EQ,
    ACTIONS(10417), 1,
      anon_sym_RPAREN,
    STATE(6968), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253048] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10419), 1,
      sym__lowercase_identifier,
    ACTIONS(10421), 1,
      anon_sym_LBRACK,
    STATE(6969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253065] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10423), 1,
      anon_sym_RPAREN,
    ACTIONS(10425), 1,
      anon_sym_SEMI,
    STATE(6970), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253082] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10423), 1,
      anon_sym_RBRACK,
    ACTIONS(10427), 1,
      anon_sym_SEMI,
    STATE(6971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253099] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10423), 1,
      anon_sym_RBRACE,
    ACTIONS(10429), 1,
      anon_sym_SEMI,
    STATE(6972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253116] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6973), 1,
      sym_attribute,
    STATE(7450), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253133] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8229), 1,
      anon_sym_RBRACE,
    ACTIONS(10431), 1,
      anon_sym_SEMI,
    STATE(6974), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253150] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10011), 1,
      anon_sym_do,
    STATE(2693), 1,
      sym_do_clause,
    STATE(6975), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253167] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6976), 1,
      sym_attribute,
    STATE(7560), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253184] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10433), 1,
      anon_sym_RPAREN,
    ACTIONS(10435), 1,
      anon_sym_SEMI,
    STATE(6977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253201] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10433), 1,
      anon_sym_RBRACK,
    ACTIONS(10437), 1,
      anon_sym_SEMI,
    STATE(6978), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253218] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10433), 1,
      anon_sym_RBRACE,
    ACTIONS(10439), 1,
      anon_sym_SEMI,
    STATE(6979), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253235] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10441), 1,
      anon_sym_SEMI,
    ACTIONS(10443), 1,
      anon_sym_RBRACE,
    STATE(6980), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253252] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10445), 1,
      anon_sym_RPAREN,
    ACTIONS(10447), 1,
      anon_sym_LT_DASH,
    STATE(6981), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253269] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6982), 1,
      sym_attribute,
    STATE(8336), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253286] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10449), 1,
      anon_sym_RPAREN,
    ACTIONS(10451), 1,
      anon_sym_SEMI,
    STATE(6983), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253303] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10449), 1,
      anon_sym_RBRACK,
    ACTIONS(10453), 1,
      anon_sym_SEMI,
    STATE(6984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253320] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10449), 1,
      anon_sym_RBRACE,
    ACTIONS(10455), 1,
      anon_sym_SEMI,
    STATE(6985), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253337] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10457), 1,
      anon_sym_RPAREN,
    ACTIONS(10459), 1,
      anon_sym_LT_DASH,
    STATE(6986), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253354] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6987), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10461), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [253369] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6988), 1,
      sym_attribute,
    STATE(7788), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253386] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10463), 1,
      anon_sym_RPAREN,
    ACTIONS(10465), 1,
      anon_sym_SEMI,
    STATE(6989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253403] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10463), 1,
      anon_sym_RBRACK,
    ACTIONS(10467), 1,
      anon_sym_SEMI,
    STATE(6990), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253420] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10463), 1,
      anon_sym_RBRACE,
    ACTIONS(10469), 1,
      anon_sym_SEMI,
    STATE(6991), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253437] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(6992), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10471), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [253452] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10473), 1,
      anon_sym_RPAREN,
    ACTIONS(10475), 1,
      anon_sym_SEMI,
    STATE(6993), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253469] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10473), 1,
      anon_sym_RBRACK,
    ACTIONS(10477), 1,
      anon_sym_SEMI,
    STATE(6994), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253486] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10473), 1,
      anon_sym_RBRACE,
    ACTIONS(10479), 1,
      anon_sym_SEMI,
    STATE(6995), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253503] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(6996), 1,
      sym_attribute,
    STATE(7505), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253520] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10481), 1,
      anon_sym_RPAREN,
    ACTIONS(10483), 1,
      anon_sym_SEMI,
    STATE(6997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253537] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10481), 1,
      anon_sym_RBRACK,
    ACTIONS(10485), 1,
      anon_sym_SEMI,
    STATE(6998), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253554] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10481), 1,
      anon_sym_RBRACE,
    ACTIONS(10487), 1,
      anon_sym_SEMI,
    STATE(6999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253571] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8257), 1,
      anon_sym_RBRACE,
    ACTIONS(10489), 1,
      anon_sym_SEMI,
    STATE(7000), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253588] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    STATE(6504), 1,
      sym_extended_module_path,
    STATE(7001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253605] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(7002), 1,
      sym_attribute,
    STATE(7780), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253622] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10491), 1,
      anon_sym_RPAREN,
    ACTIONS(10493), 1,
      anon_sym_SEMI,
    STATE(7003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253639] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10491), 1,
      anon_sym_RBRACK,
    ACTIONS(10495), 1,
      anon_sym_SEMI,
    STATE(7004), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253656] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10491), 1,
      anon_sym_RBRACE,
    ACTIONS(10497), 1,
      anon_sym_SEMI,
    STATE(7005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253673] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10499), 1,
      sym__lowercase_identifier,
    ACTIONS(10501), 1,
      anon_sym_LPAREN,
    STATE(7006), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253690] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10503), 1,
      anon_sym_RPAREN,
    ACTIONS(10505), 1,
      anon_sym_LT_DASH,
    STATE(7007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253707] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(7008), 1,
      sym_attribute,
    STATE(7997), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253724] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10507), 1,
      anon_sym_RPAREN,
    ACTIONS(10509), 1,
      anon_sym_SEMI,
    STATE(7009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253741] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10507), 1,
      anon_sym_RBRACK,
    ACTIONS(10511), 1,
      anon_sym_SEMI,
    STATE(7010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253758] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10507), 1,
      anon_sym_RBRACE,
    ACTIONS(10513), 1,
      anon_sym_SEMI,
    STATE(7011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253775] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10181), 1,
      anon_sym_COLON2,
    STATE(7012), 1,
      sym_attribute,
    STATE(7779), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253792] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(7013), 1,
      sym_attribute,
    STATE(8256), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253809] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10515), 1,
      anon_sym_RPAREN,
    ACTIONS(10517), 1,
      anon_sym_SEMI,
    STATE(7014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253826] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10515), 1,
      anon_sym_RBRACK,
    ACTIONS(10519), 1,
      anon_sym_SEMI,
    STATE(7015), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253843] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10515), 1,
      anon_sym_RBRACE,
    ACTIONS(10521), 1,
      anon_sym_SEMI,
    STATE(7016), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253860] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10523), 1,
      anon_sym_RPAREN,
    ACTIONS(10525), 1,
      anon_sym_COLON_GT,
    STATE(7017), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253877] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(7018), 1,
      sym_attribute,
    STATE(7247), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253894] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(7019), 1,
      sym_attribute,
    STATE(7280), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253911] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(10527), 1,
      sym__lowercase_identifier,
    STATE(7020), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253928] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(7021), 1,
      sym_attribute,
    STATE(7295), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253945] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10371), 1,
      anon_sym_COLON2,
    STATE(4474), 1,
      sym__maybe_polymorphic_typed,
    STATE(7022), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253962] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(7023), 1,
      sym_attribute,
    STATE(7313), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253979] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10055), 1,
      anon_sym_LPAREN,
    ACTIONS(10529), 1,
      sym__lowercase_identifier,
    STATE(7024), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [253996] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10253), 1,
      anon_sym_LPAREN,
    ACTIONS(10531), 1,
      sym__lowercase_identifier,
    STATE(7025), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254013] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8949), 1,
      anon_sym_COLON2,
    STATE(7026), 1,
      sym_attribute,
    STATE(7184), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254030] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(10533), 1,
      sym__lowercase_identifier,
    STATE(7027), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254047] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10535), 1,
      sym__lowercase_identifier,
    ACTIONS(10537), 1,
      anon_sym_LBRACK,
    STATE(7028), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254064] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10173), 1,
      anon_sym_LPAREN,
    ACTIONS(10539), 1,
      sym__lowercase_identifier,
    STATE(7029), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254081] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8841), 1,
      sym__lowercase_identifier,
    STATE(6164), 1,
      sym__simple_value_pattern,
    STATE(7030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254098] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7031), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10541), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [254113] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(10543), 1,
      anon_sym_RPAREN,
    STATE(7032), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254130] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8949), 1,
      anon_sym_COLON2,
    STATE(7033), 1,
      sym_attribute,
    STATE(7783), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254147] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10545), 1,
      anon_sym_SEMI,
    ACTIONS(10547), 1,
      anon_sym_RBRACE,
    STATE(7034), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254164] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8949), 1,
      anon_sym_COLON2,
    STATE(7035), 1,
      sym_attribute,
    STATE(7913), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254181] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10549), 1,
      anon_sym_SEMI,
    ACTIONS(10551), 1,
      anon_sym_RBRACE,
    STATE(7036), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254198] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10553), 1,
      sym__lowercase_identifier,
    ACTIONS(10555), 1,
      anon_sym_LBRACK,
    STATE(7037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254215] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10123), 1,
      anon_sym_do,
    STATE(2725), 1,
      sym_do_clause,
    STATE(7038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254232] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10557), 1,
      anon_sym_RPAREN,
    ACTIONS(10559), 1,
      anon_sym_LT_DASH,
    STATE(7039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254249] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7040), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10561), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [254264] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(7041), 1,
      sym_attribute,
    STATE(7373), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254281] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(7042), 1,
      sym_attribute,
    STATE(8116), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254298] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(7043), 1,
      sym_attribute,
    STATE(7386), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254315] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10563), 1,
      anon_sym_RPAREN,
    ACTIONS(10565), 1,
      anon_sym_LT_DASH,
    STATE(7044), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254332] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(10567), 1,
      sym__lowercase_identifier,
    STATE(7045), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254349] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7046), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6447), 2,
      anon_sym_and,
      anon_sym_in,
  [254364] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10259), 1,
      anon_sym_do,
    STATE(2771), 1,
      sym_do_clause,
    STATE(7047), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254381] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10569), 1,
      anon_sym_open,
    STATE(7048), 1,
      sym_attribute,
    STATE(8168), 1,
      sym_open_module,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254398] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10181), 1,
      anon_sym_COLON2,
    STATE(7049), 1,
      sym_attribute,
    STATE(7408), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254415] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7050), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10571), 2,
      anon_sym_to,
      anon_sym_downto,
  [254430] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10573), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [254445] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10123), 1,
      anon_sym_do,
    STATE(2714), 1,
      sym_do_clause,
    STATE(7052), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254462] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7053), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10575), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [254477] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(10577), 1,
      anon_sym_RPAREN,
    STATE(7054), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254494] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7055), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10579), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [254509] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10581), 1,
      anon_sym_then,
    STATE(2726), 1,
      sym_then_clause,
    STATE(7056), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254526] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7057), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10583), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [254541] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7058), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10585), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [254556] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10587), 1,
      anon_sym_RPAREN,
    ACTIONS(10589), 1,
      anon_sym_LT_DASH,
    STATE(7059), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254573] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10591), 2,
      anon_sym_to,
      anon_sym_downto,
  [254588] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    STATE(3854), 1,
      sym_extended_module_path,
    STATE(7061), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254605] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10593), 1,
      anon_sym_do,
    STATE(3523), 1,
      sym_do_clause,
    STATE(7062), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254622] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10595), 1,
      anon_sym_RPAREN,
    ACTIONS(10597), 1,
      anon_sym_LT_DASH,
    STATE(7063), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254639] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10599), 1,
      anon_sym_then,
    STATE(3875), 1,
      sym_then_clause,
    STATE(7064), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254656] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10259), 1,
      anon_sym_do,
    STATE(2801), 1,
      sym_do_clause,
    STATE(7065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254673] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(10601), 1,
      anon_sym_EQ,
    STATE(7066), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254690] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7067), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10603), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [254705] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(7068), 1,
      sym_attribute,
    STATE(7579), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254722] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7069), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6801), 2,
      anon_sym_and,
      anon_sym_in,
  [254737] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10115), 1,
      anon_sym_COLON2,
    STATE(4383), 1,
      sym__maybe_polymorphic_typed,
    STATE(7070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254754] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10605), 1,
      anon_sym_RPAREN,
    ACTIONS(10607), 1,
      anon_sym_LT_DASH,
    STATE(7071), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254771] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10599), 1,
      anon_sym_then,
    STATE(3931), 1,
      sym_then_clause,
    STATE(7072), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254788] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3752), 1,
      anon_sym_COMMA,
    ACTIONS(10609), 1,
      anon_sym_COLON,
    STATE(7073), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254805] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10611), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [254820] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10613), 1,
      anon_sym_RPAREN,
    ACTIONS(10615), 1,
      anon_sym_SEMI,
    STATE(7075), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254837] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10613), 1,
      anon_sym_RBRACK,
    ACTIONS(10617), 1,
      anon_sym_SEMI,
    STATE(7076), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254854] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10619), 1,
      sym__lowercase_identifier,
    ACTIONS(10621), 1,
      anon_sym_LPAREN,
    STATE(7077), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254871] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7078), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10623), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [254886] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8102), 1,
      anon_sym_COLON2,
    STATE(7079), 1,
      sym_attribute,
    STATE(7107), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254903] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10625), 1,
      anon_sym_then,
    STATE(4080), 1,
      sym_then_clause,
    STATE(7080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254920] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7081), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10627), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [254935] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10613), 1,
      anon_sym_RBRACE,
    ACTIONS(10629), 1,
      anon_sym_SEMI,
    STATE(7082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254952] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7083), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10631), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [254967] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7084), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10633), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [254982] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8841), 1,
      sym__lowercase_identifier,
    STATE(6204), 1,
      sym__simple_value_pattern,
    STATE(7085), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [254999] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8102), 1,
      anon_sym_COLON2,
    STATE(7081), 1,
      sym__maybe_polymorphic_typed,
    STATE(7086), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255016] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9075), 1,
      sym__lowercase_identifier,
    STATE(6270), 1,
      aux_sym__abstract_type_repeat1,
    STATE(7087), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255033] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8277), 1,
      anon_sym_RBRACE,
    ACTIONS(10635), 1,
      anon_sym_SEMI,
    STATE(7088), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255050] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7089), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9705), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [255065] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(10637), 1,
      sym__lowercase_identifier,
    STATE(7090), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255082] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10569), 1,
      anon_sym_open,
    STATE(7091), 1,
      sym_attribute,
    STATE(8293), 1,
      sym_open_module,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255099] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10015), 1,
      anon_sym_then,
    STATE(3232), 1,
      sym_then_clause,
    STATE(7092), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255116] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(10639), 1,
      anon_sym_EQ,
    STATE(7093), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255133] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(10641), 1,
      anon_sym_RPAREN,
    STATE(7094), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255150] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10625), 1,
      anon_sym_then,
    STATE(4124), 1,
      sym_then_clause,
    STATE(7095), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255167] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(10643), 1,
      sym__lowercase_identifier,
    STATE(7096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255184] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10645), 1,
      anon_sym_RPAREN,
    ACTIONS(10647), 1,
      anon_sym_LT_DASH,
    STATE(7097), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255201] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10569), 1,
      anon_sym_open,
    STATE(7098), 1,
      sym_attribute,
    STATE(7963), 1,
      sym_open_module,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255218] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10055), 1,
      anon_sym_LPAREN,
    ACTIONS(10649), 1,
      sym__lowercase_identifier,
    STATE(7099), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255235] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(10651), 1,
      sym__lowercase_identifier,
    STATE(7100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255252] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10253), 1,
      anon_sym_LPAREN,
    ACTIONS(10653), 1,
      sym__lowercase_identifier,
    STATE(7101), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255269] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10655), 1,
      sym__lowercase_identifier,
    ACTIONS(10657), 1,
      anon_sym_LPAREN,
    STATE(7102), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255286] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8949), 1,
      anon_sym_COLON2,
    STATE(7103), 1,
      sym_attribute,
    STATE(7815), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255303] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10173), 1,
      anon_sym_LPAREN,
    ACTIONS(10659), 1,
      sym__lowercase_identifier,
    STATE(7104), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255320] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8949), 1,
      anon_sym_COLON2,
    STATE(7105), 1,
      sym_attribute,
    STATE(7917), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255337] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10661), 1,
      anon_sym_DOT,
    STATE(7106), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255354] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7107), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10663), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [255369] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7108), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10665), 2,
      anon_sym_to,
      anon_sym_downto,
  [255384] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10667), 1,
      anon_sym_RPAREN,
    ACTIONS(10669), 1,
      anon_sym_LT_DASH,
    STATE(7109), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255401] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7110), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10671), 2,
      anon_sym_to,
      anon_sym_downto,
  [255416] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10673), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [255431] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10287), 1,
      anon_sym_then,
    STATE(3843), 1,
      sym_then_clause,
    STATE(7112), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255448] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7113), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10675), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [255463] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_COLON2,
    STATE(7114), 1,
      sym_attribute,
    STATE(7487), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255480] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(10677), 1,
      sym__lowercase_identifier,
    STATE(7115), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255497] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10679), 1,
      anon_sym_SEMI,
    ACTIONS(10681), 1,
      anon_sym_RBRACE,
    STATE(7116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255514] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7117), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4728), 2,
      anon_sym_DOT,
      aux_sym_type_variable_token1,
  [255529] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7118), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10683), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [255544] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10685), 1,
      anon_sym_DOT,
    STATE(7119), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255561] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7120), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10687), 2,
      anon_sym_to,
      anon_sym_downto,
  [255576] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7121), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10689), 2,
      anon_sym_to,
      anon_sym_downto,
  [255591] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7122), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10691), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [255606] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3382), 1,
      anon_sym_COMMA,
    ACTIONS(10693), 1,
      anon_sym_COLON,
    STATE(7123), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255623] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10569), 1,
      anon_sym_open,
    STATE(7124), 1,
      sym_attribute,
    STATE(7521), 1,
      sym_open_module,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255640] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10695), 1,
      anon_sym_DOT,
    STATE(7125), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255657] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10593), 1,
      anon_sym_do,
    STATE(3521), 1,
      sym_do_clause,
    STATE(7126), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255674] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10697), 1,
      sym__lowercase_identifier,
    ACTIONS(10699), 1,
      anon_sym_LPAREN,
    STATE(7127), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255691] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10095), 1,
      anon_sym_then,
    STATE(2918), 1,
      sym_then_clause,
    STATE(7128), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255708] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4361), 1,
      anon_sym_RPAREN,
    ACTIONS(10701), 1,
      aux_sym_number_token1,
    STATE(7129), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255725] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7130), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9267), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [255740] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10011), 1,
      anon_sym_do,
    STATE(2643), 1,
      sym_do_clause,
    STATE(7131), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255757] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10703), 1,
      anon_sym_DOT,
    STATE(7132), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255774] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(10705), 1,
      anon_sym_RPAREN,
    STATE(7133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255791] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10707), 1,
      anon_sym_RPAREN,
    ACTIONS(10709), 1,
      anon_sym_LT_DASH,
    STATE(7134), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255808] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7135), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10711), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [255823] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3796), 1,
      aux_sym_type_variable_token1,
    STATE(3240), 1,
      sym_type_variable,
    STATE(7136), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255840] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10713), 1,
      sym__lowercase_identifier,
    ACTIONS(10715), 1,
      anon_sym_LPAREN,
    STATE(7137), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255857] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10181), 1,
      anon_sym_COLON2,
    STATE(7138), 1,
      sym_attribute,
    STATE(7591), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255874] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7139), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6524), 2,
      anon_sym_and,
      anon_sym_in,
  [255889] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10717), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [255904] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10719), 1,
      anon_sym_SEMI,
    ACTIONS(10721), 1,
      anon_sym_RBRACE,
    STATE(7141), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255921] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(10723), 1,
      sym__lowercase_identifier,
    STATE(7142), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255938] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10725), 1,
      anon_sym_RPAREN,
    ACTIONS(10727), 1,
      anon_sym_LT_DASH,
    STATE(7143), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [255955] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10729), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [255970] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7145), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10731), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [255985] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(10733), 1,
      sym__lowercase_identifier,
    STATE(7146), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256002] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10735), 1,
      sym__lowercase_identifier,
    ACTIONS(10737), 1,
      anon_sym_LBRACK,
    STATE(7147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256019] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7148), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10739), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [256034] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(8042), 1,
      anon_sym_RBRACK,
    STATE(7149), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256051] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10137), 1,
      anon_sym_then,
    STATE(2888), 1,
      sym_then_clause,
    STATE(7150), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256068] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10593), 1,
      anon_sym_do,
    STATE(3552), 1,
      sym_do_clause,
    STATE(7151), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256085] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10581), 1,
      anon_sym_then,
    STATE(2727), 1,
      sym_then_clause,
    STATE(7152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256102] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    ACTIONS(10741), 1,
      sym__lowercase_identifier,
    STATE(7153), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256119] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10743), 1,
      anon_sym_RPAREN,
    ACTIONS(10745), 1,
      anon_sym_LT_DASH,
    STATE(7154), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256136] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10747), 1,
      anon_sym_RPAREN,
    ACTIONS(10749), 1,
      anon_sym_LT_DASH,
    STATE(7155), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256153] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10135), 1,
      anon_sym_LPAREN,
    ACTIONS(10751), 1,
      sym__lowercase_identifier,
    STATE(7156), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256170] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7772), 1,
      anon_sym_COMMA,
    ACTIONS(10753), 1,
      anon_sym_COLON,
    STATE(7157), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256187] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_as,
    ACTIONS(10755), 1,
      anon_sym_RPAREN,
    STATE(7158), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256204] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7159), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10757), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [256219] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10759), 1,
      anon_sym_DOT,
    STATE(7160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256236] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5190), 1,
      anon_sym_EQ,
    ACTIONS(10761), 1,
      anon_sym_RPAREN,
    STATE(7161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256253] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(7162), 1,
      sym_attribute,
    STATE(7185), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256270] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(10763), 1,
      sym__lowercase_identifier,
    STATE(7163), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256287] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4198), 1,
      anon_sym_RBRACK,
    ACTIONS(8670), 1,
      anon_sym_PIPE,
    STATE(7164), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256304] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10765), 1,
      anon_sym_RPAREN,
    ACTIONS(10767), 1,
      anon_sym_LT_DASH,
    STATE(7165), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256321] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10769), 1,
      anon_sym_RPAREN,
    ACTIONS(10771), 1,
      anon_sym_LT_DASH,
    STATE(7166), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256338] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      aux_sym_type_variable_token1,
    STATE(3240), 1,
      sym_type_variable,
    STATE(7167), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256355] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(10773), 1,
      sym__lowercase_identifier,
    STATE(7168), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256372] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7169), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10775), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [256387] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    ACTIONS(10777), 1,
      sym__lowercase_identifier,
    STATE(7170), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256404] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7171), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10779), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [256419] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      sym__uppercase_identifier,
    STATE(6425), 1,
      sym_extended_module_path,
    STATE(7172), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256436] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7173), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10781), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [256451] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10783), 1,
      anon_sym_DOT,
    STATE(7174), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256468] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7175), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5310), 2,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
  [256483] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10181), 1,
      anon_sym_COLON2,
    STATE(7176), 1,
      sym_attribute,
    STATE(8158), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256500] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7177), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10785), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [256515] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10787), 1,
      anon_sym_RPAREN,
    ACTIONS(10789), 1,
      anon_sym_LT_DASH,
    STATE(7178), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256532] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9993), 1,
      anon_sym_RPAREN,
    ACTIONS(10791), 1,
      anon_sym_SEMI,
    STATE(7179), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256549] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7180), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(10793), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [256564] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10593), 1,
      anon_sym_do,
    STATE(3573), 1,
      sym_do_clause,
    STATE(7181), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256581] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5108), 1,
      anon_sym_LPAREN,
    ACTIONS(10795), 1,
      anon_sym_DOT,
    STATE(7182), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256598] = 5,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      sym__left_quoted_string_delimiter,
    STATE(7183), 1,
      sym_attribute,
    STATE(7196), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256615] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10797), 1,
      anon_sym_EQ,
    STATE(7184), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256629] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10799), 1,
      anon_sym_RBRACE,
    STATE(7185), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256643] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10801), 1,
      anon_sym_DOT,
    STATE(7186), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256657] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10803), 1,
      anon_sym_RPAREN,
    STATE(7187), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256671] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10805), 1,
      anon_sym_DASH_GT,
    STATE(7188), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256685] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10807), 1,
      anon_sym_DOT,
    STATE(7189), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256699] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10809), 1,
      anon_sym_RPAREN,
    STATE(7190), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256713] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10811), 1,
      anon_sym_RPAREN,
    STATE(7191), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256727] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10813), 1,
      anon_sym_RBRACE,
    STATE(7192), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256741] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10815), 1,
      anon_sym_SQUOTE2,
    STATE(7193), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256755] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10817), 1,
      anon_sym_STAR,
    STATE(7194), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256769] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10819), 1,
      anon_sym_DOT,
    STATE(7195), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256783] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10821), 1,
      anon_sym_RBRACE,
    STATE(7196), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256797] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10823), 1,
      anon_sym_with,
    STATE(7197), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256811] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10825), 1,
      anon_sym_RBRACK,
    STATE(7198), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256825] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10827), 1,
      anon_sym_DOT,
    STATE(7199), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256839] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10829), 1,
      anon_sym_with,
    STATE(7200), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256853] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10831), 1,
      anon_sym_DOT,
    STATE(7201), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256867] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10833), 1,
      anon_sym_RPAREN,
    STATE(7202), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256881] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10835), 1,
      anon_sym_RBRACE,
    STATE(7203), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256895] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10837), 1,
      anon_sym_LT_DASH,
    STATE(7204), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256909] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10257), 1,
      anon_sym_RBRACE,
    STATE(7205), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256923] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4766), 1,
      anon_sym_COMMA,
    STATE(7206), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256937] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10839), 1,
      anon_sym_RBRACE,
    STATE(7207), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256951] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10841), 1,
      anon_sym_DQUOTE,
    STATE(7208), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256965] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10843), 1,
      anon_sym_EQ,
    STATE(7209), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256979] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10845), 1,
      anon_sym_in,
    STATE(7210), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [256993] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10847), 1,
      anon_sym_type,
    STATE(7211), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257007] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10849), 1,
      anon_sym_RBRACE,
    STATE(7212), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257021] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10851), 1,
      anon_sym_RBRACE,
    STATE(7213), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257035] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10853), 1,
      anon_sym_RPAREN,
    STATE(7214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257049] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10855), 1,
      anon_sym_RBRACK,
    STATE(7215), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257063] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8030), 1,
      anon_sym_GT,
    STATE(7216), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257077] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10857), 1,
      anon_sym_STAR,
    STATE(7217), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257091] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10859), 1,
      anon_sym_in,
    STATE(7218), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257105] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10861), 1,
      anon_sym_with,
    STATE(7219), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257119] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10863), 1,
      anon_sym_RBRACE,
    STATE(7220), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257133] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10865), 1,
      anon_sym_with,
    STATE(7221), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257147] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10867), 1,
      anon_sym_RBRACE,
    STATE(7222), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257161] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10869), 1,
      anon_sym_RPAREN,
    STATE(7223), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257175] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10871), 1,
      anon_sym_RPAREN,
    STATE(7224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257189] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4515), 1,
      anon_sym_COMMA,
    STATE(7225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257203] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10873), 1,
      anon_sym_DOT,
    STATE(7226), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257217] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10875), 1,
      anon_sym_LT_DASH,
    STATE(7227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257231] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10877), 1,
      anon_sym_DOT,
    STATE(7228), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257245] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10879), 1,
      anon_sym_in,
    STATE(7229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257259] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10881), 1,
      anon_sym_COMMA,
    STATE(7230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257273] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10883), 1,
      anon_sym_in,
    STATE(7231), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257287] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10885), 1,
      anon_sym_EQ,
    STATE(7232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257301] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10887), 1,
      anon_sym_EQ,
    STATE(7233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257315] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10889), 1,
      anon_sym_DQUOTE,
    STATE(7234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257329] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10891), 1,
      anon_sym_RPAREN,
    STATE(7235), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257343] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10893), 1,
      anon_sym_DASH_GT,
    STATE(7236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257357] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10895), 1,
      anon_sym_RPAREN,
    STATE(7237), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257371] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10897), 1,
      anon_sym_RBRACK,
    STATE(7238), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257385] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10899), 1,
      anon_sym_RPAREN,
    STATE(7239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257399] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10901), 1,
      anon_sym_RPAREN,
    STATE(7240), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257413] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10903), 1,
      anon_sym_in,
    STATE(7241), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257427] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9304), 1,
      anon_sym_GT_RBRACE,
    STATE(7242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257441] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10905), 1,
      anon_sym_RBRACK,
    STATE(7243), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257455] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10907), 1,
      anon_sym_RPAREN,
    STATE(7244), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257469] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10909), 1,
      anon_sym_RBRACE,
    STATE(7245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257483] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10911), 1,
      anon_sym_DOT,
    STATE(7246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257497] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10913), 1,
      anon_sym_RBRACE,
    STATE(7247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257511] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10915), 1,
      anon_sym_EQ,
    STATE(7248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257525] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10917), 1,
      anon_sym_RPAREN,
    STATE(7249), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257539] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10919), 1,
      anon_sym_RPAREN,
    STATE(7250), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257553] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10921), 1,
      anon_sym_RPAREN,
    STATE(7251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257567] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10923), 1,
      anon_sym_RBRACE,
    STATE(7252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257581] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10925), 1,
      anon_sym_RPAREN,
    STATE(7253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257595] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10925), 1,
      anon_sym_RBRACK,
    STATE(7254), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257609] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10925), 1,
      anon_sym_RBRACE,
    STATE(7255), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257623] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10927), 1,
      anon_sym_end,
    STATE(7256), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257637] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10929), 1,
      anon_sym_in,
    STATE(7257), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257651] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10931), 1,
      anon_sym_RBRACE,
    STATE(7258), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257665] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10933), 1,
      anon_sym_RBRACK,
    STATE(7259), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257679] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10935), 1,
      anon_sym_RBRACK,
    STATE(7260), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257693] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10937), 1,
      anon_sym_RPAREN,
    STATE(7261), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257707] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10939), 1,
      anon_sym_RPAREN,
    STATE(7262), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257721] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10941), 1,
      anon_sym_RBRACE,
    STATE(7263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257735] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10943), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7264), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257749] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10945), 1,
      sym__lowercase_identifier,
    STATE(7265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257763] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10947), 1,
      anon_sym_RBRACE,
    STATE(7266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257777] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10949), 1,
      anon_sym_RPAREN,
    STATE(7267), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257791] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10951), 1,
      anon_sym_DOT,
    STATE(7268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257805] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10953), 1,
      anon_sym_DOT,
    STATE(7269), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257819] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10955), 1,
      anon_sym_RPAREN,
    STATE(7270), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257833] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10957), 1,
      anon_sym_EQ,
    STATE(7271), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257847] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10959), 1,
      anon_sym_SQUOTE2,
    STATE(7272), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257861] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10961), 1,
      anon_sym_RPAREN,
    STATE(7273), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257875] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10963), 1,
      anon_sym_DOT,
    STATE(7274), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257889] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10965), 1,
      anon_sym_with,
    STATE(7275), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257903] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8160), 1,
      anon_sym_DASH_GT,
    STATE(7276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257917] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10967), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7277), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257931] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10969), 1,
      anon_sym_with,
    STATE(7278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257945] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10971), 1,
      anon_sym_DOT,
    STATE(7279), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257959] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10973), 1,
      anon_sym_RBRACE,
    STATE(7280), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257973] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10975), 1,
      anon_sym_LT_DASH,
    STATE(7281), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [257987] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10977), 1,
      anon_sym_with,
    STATE(7282), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258001] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10979), 1,
      anon_sym_EQ,
    STATE(7283), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258015] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10981), 1,
      anon_sym_RBRACK,
    STATE(7284), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258029] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10983), 1,
      anon_sym_RBRACK,
    STATE(7285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258043] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10985), 1,
      anon_sym_DOT,
    STATE(7286), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258057] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10291), 1,
      anon_sym_RBRACE,
    STATE(7287), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258071] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10987), 1,
      anon_sym_RPAREN,
    STATE(7288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258085] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10989), 1,
      anon_sym_in,
    STATE(7289), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258099] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10991), 1,
      anon_sym_with,
    STATE(7290), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258113] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10993), 1,
      anon_sym_RBRACE,
    STATE(7291), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258127] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10995), 1,
      anon_sym_with,
    STATE(7292), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258141] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10997), 1,
      anon_sym_RPAREN,
    STATE(7293), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258155] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8406), 1,
      anon_sym_RPAREN,
    STATE(7294), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258169] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10999), 1,
      anon_sym_RBRACE,
    STATE(7295), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258183] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11001), 1,
      anon_sym_RBRACE,
    STATE(7296), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258197] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11003), 1,
      anon_sym_DOT,
    STATE(7297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258211] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11005), 1,
      anon_sym_RPAREN,
    STATE(7298), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258225] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11007), 1,
      anon_sym_EQ,
    STATE(7299), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258239] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11009), 1,
      anon_sym_EQ,
    STATE(7300), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258253] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11011), 1,
      anon_sym_RBRACK,
    STATE(7301), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258267] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11013), 1,
      anon_sym_RPAREN,
    STATE(7302), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258281] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4190), 1,
      anon_sym_COMMA,
    STATE(7303), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258295] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11015), 1,
      anon_sym_LT_DASH,
    STATE(7304), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258309] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11017), 1,
      anon_sym_RPAREN,
    STATE(7305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258323] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1549), 1,
      anon_sym_RPAREN,
    STATE(7306), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258337] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11019), 1,
      anon_sym_in,
    STATE(7307), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258351] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9485), 1,
      anon_sym_GT_RBRACE,
    STATE(7308), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258365] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11021), 1,
      anon_sym_RBRACE,
    STATE(7309), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258379] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8350), 1,
      anon_sym_end,
    STATE(7310), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258393] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8023), 1,
      anon_sym_GT,
    STATE(7311), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258407] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8629), 1,
      sym__uppercase_identifier,
    STATE(7312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258421] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11023), 1,
      anon_sym_RBRACE,
    STATE(7313), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258435] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11025), 1,
      anon_sym_EQ,
    STATE(7314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258449] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11027), 1,
      anon_sym_RPAREN,
    STATE(7315), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258463] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11029), 1,
      anon_sym_RPAREN,
    STATE(7316), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258477] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11029), 1,
      anon_sym_RBRACK,
    STATE(7317), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258491] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11029), 1,
      anon_sym_RBRACE,
    STATE(7318), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258505] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11031), 1,
      anon_sym_RPAREN,
    STATE(7319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258519] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11033), 1,
      anon_sym_DASH_GT,
    STATE(7320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258533] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11035), 1,
      anon_sym_RPAREN,
    STATE(7321), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258547] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11037), 1,
      anon_sym_RPAREN,
    STATE(7322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258561] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11039), 1,
      anon_sym_RBRACK,
    STATE(7323), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258575] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11041), 1,
      anon_sym_RBRACK,
    STATE(7324), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258589] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11043), 1,
      anon_sym_RBRACE,
    STATE(7325), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258603] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11045), 1,
      anon_sym_SQUOTE2,
    STATE(7326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258617] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11047), 1,
      anon_sym_DQUOTE,
    STATE(7327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258631] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11049), 1,
      anon_sym_RPAREN,
    STATE(7328), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258645] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11051), 1,
      anon_sym_DQUOTE,
    STATE(7329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258659] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11053), 1,
      anon_sym_DOT,
    STATE(7330), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258673] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11055), 1,
      anon_sym_DOT,
    STATE(7331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258687] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11057), 1,
      anon_sym_RPAREN,
    STATE(7332), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258701] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11059), 1,
      sym__right_quoted_string_delimiter,
    STATE(7333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258715] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11061), 1,
      anon_sym_RPAREN,
    STATE(7334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258729] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11063), 1,
      anon_sym_DOT,
    STATE(7335), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258743] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11065), 1,
      anon_sym_with,
    STATE(7336), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258757] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11067), 1,
      anon_sym_RPAREN,
    STATE(7337), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258771] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11069), 1,
      sym__lowercase_identifier,
    STATE(7338), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258785] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11071), 1,
      anon_sym_with,
    STATE(7339), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258799] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11073), 1,
      anon_sym_DOT,
    STATE(7340), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258813] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1519), 1,
      anon_sym_RPAREN,
    STATE(7341), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258827] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11075), 1,
      anon_sym_RBRACE,
    STATE(7342), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258841] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11077), 1,
      sym__lowercase_identifier,
    STATE(7343), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258855] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11079), 1,
      anon_sym_RPAREN,
    STATE(7344), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258869] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11081), 1,
      anon_sym_in,
    STATE(7345), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258883] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3370), 1,
      anon_sym_COMMA,
    STATE(7346), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258897] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11083), 1,
      anon_sym_with,
    STATE(7347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258911] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11085), 1,
      anon_sym_DASH_GT,
    STATE(7348), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258925] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11087), 1,
      anon_sym_with,
    STATE(7349), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258939] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11089), 1,
      anon_sym_RPAREN,
    STATE(7350), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258953] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11091), 1,
      anon_sym_RPAREN,
    STATE(7351), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258967] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11093), 1,
      anon_sym_LT_DASH,
    STATE(7352), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258981] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11095), 1,
      anon_sym_done,
    STATE(7353), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [258995] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11097), 1,
      anon_sym_DOT,
    STATE(7354), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259009] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11099), 1,
      anon_sym_RPAREN,
    STATE(7355), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259023] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11101), 1,
      sym__lowercase_identifier,
    STATE(7356), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259037] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11103), 1,
      aux_sym_number_token1,
    STATE(7357), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259051] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11105), 1,
      anon_sym_in,
    STATE(7358), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259065] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11107), 1,
      anon_sym_RPAREN,
    STATE(7359), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259079] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2756), 1,
      anon_sym_COMMA,
    STATE(7360), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259093] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11109), 1,
      anon_sym_RPAREN,
    STATE(7361), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259107] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11111), 1,
      anon_sym_RPAREN,
    STATE(7362), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259121] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5001), 1,
      anon_sym_GT,
    STATE(7363), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259135] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3154), 1,
      anon_sym_DASH_GT,
    STATE(7364), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259149] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11113), 1,
      anon_sym_RPAREN,
    STATE(7365), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259163] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11115), 1,
      anon_sym_RPAREN,
    STATE(7366), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259177] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11115), 1,
      anon_sym_RBRACK,
    STATE(7367), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259191] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11115), 1,
      anon_sym_RBRACE,
    STATE(7368), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259205] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11117), 1,
      anon_sym_DASH_GT,
    STATE(7369), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259219] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11119), 1,
      anon_sym_DASH_GT,
    STATE(7370), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259233] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11121), 1,
      anon_sym_RPAREN,
    STATE(7371), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259247] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11123), 1,
      anon_sym_end,
    STATE(7372), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259261] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11125), 1,
      anon_sym_RBRACE,
    STATE(7373), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259275] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11127), 1,
      sym__lowercase_identifier,
    STATE(7374), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259289] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11129), 1,
      anon_sym_DASH_GT,
    STATE(7375), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259303] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11131), 1,
      anon_sym_DOT,
    STATE(7376), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259317] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11133), 1,
      anon_sym_DOT,
    STATE(7377), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259331] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11135), 1,
      anon_sym_RPAREN,
    STATE(7378), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259345] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11137), 1,
      anon_sym_RPAREN,
    STATE(7379), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259359] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8499), 1,
      anon_sym_end,
    STATE(7380), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259373] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11139), 1,
      anon_sym_DOT,
    STATE(7381), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259387] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11141), 1,
      anon_sym_with,
    STATE(7382), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259401] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11143), 1,
      anon_sym_with,
    STATE(7383), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259415] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11145), 1,
      anon_sym_DOT,
    STATE(7384), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259429] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11147), 1,
      anon_sym_DOT,
    STATE(7385), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259443] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11149), 1,
      anon_sym_RBRACE,
    STATE(7386), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259457] = 4,
    ACTIONS(463), 1,
      ts_builtin_sym_end,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    STATE(7387), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259471] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4385), 1,
      anon_sym_RPAREN,
    STATE(7388), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259485] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11151), 1,
      anon_sym_with,
    STATE(7389), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259499] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11153), 1,
      anon_sym_with,
    STATE(7390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259513] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11155), 1,
      anon_sym_RPAREN,
    STATE(7391), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259527] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11157), 1,
      anon_sym_end,
    STATE(7392), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259541] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11159), 1,
      anon_sym_RPAREN,
    STATE(7393), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259555] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11161), 1,
      anon_sym_RBRACK,
    STATE(7394), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259569] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11163), 1,
      anon_sym_DOT,
    STATE(7395), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259583] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11165), 1,
      aux_sym_number_token1,
    STATE(7396), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259597] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11167), 1,
      anon_sym_in,
    STATE(7397), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259611] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3312), 1,
      anon_sym_COMMA,
    STATE(7398), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259625] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11169), 1,
      anon_sym_RPAREN,
    STATE(7399), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259639] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11171), 1,
      anon_sym_RBRACK,
    STATE(7400), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259653] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11173), 1,
      anon_sym_RBRACE,
    STATE(7401), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259667] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11175), 1,
      anon_sym_in,
    STATE(7402), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259681] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11177), 1,
      anon_sym_RBRACK,
    STATE(7403), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259695] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11179), 1,
      anon_sym_RPAREN,
    STATE(7404), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259709] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11181), 1,
      anon_sym_RPAREN,
    STATE(7405), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259723] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11181), 1,
      anon_sym_RBRACK,
    STATE(7406), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259737] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11181), 1,
      anon_sym_RBRACE,
    STATE(7407), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259751] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11183), 1,
      anon_sym_RPAREN,
    STATE(7408), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259765] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11185), 1,
      anon_sym_RBRACE,
    STATE(7409), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259779] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11187), 1,
      anon_sym_RPAREN,
    STATE(7410), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259793] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11189), 1,
      anon_sym_COMMA,
    STATE(7411), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259807] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11191), 1,
      anon_sym_DOT,
    STATE(7412), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259821] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11193), 1,
      anon_sym_DOT,
    STATE(7413), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259835] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11195), 1,
      anon_sym_done,
    STATE(7414), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259849] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11197), 1,
      anon_sym_DOT,
    STATE(7415), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259863] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11199), 1,
      anon_sym_DOT,
    STATE(7416), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259877] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11201), 1,
      anon_sym_RPAREN,
    STATE(7417), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259891] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11203), 1,
      anon_sym_LT_DASH,
    STATE(7418), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259905] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11205), 1,
      anon_sym_of,
    STATE(7419), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259919] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11207), 1,
      anon_sym_DOT,
    STATE(7420), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259933] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11209), 1,
      anon_sym_with,
    STATE(7421), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259947] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11211), 1,
      anon_sym_with,
    STATE(7422), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259961] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11213), 1,
      anon_sym_DOT,
    STATE(7423), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259975] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11215), 1,
      anon_sym_in,
    STATE(7424), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [259989] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11217), 1,
      anon_sym_RBRACE,
    STATE(7425), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260003] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11219), 1,
      aux_sym_number_token1,
    STATE(7426), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260017] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8426), 1,
      anon_sym_RPAREN,
    STATE(7427), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260031] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11221), 1,
      anon_sym_with,
    STATE(7428), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260045] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11223), 1,
      anon_sym_with,
    STATE(7429), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260059] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11225), 1,
      anon_sym_RPAREN,
    STATE(7430), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260073] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11227), 1,
      anon_sym_DASH_GT,
    STATE(7431), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260087] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11229), 1,
      anon_sym_RBRACK,
    STATE(7432), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260101] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11231), 1,
      anon_sym_in,
    STATE(7433), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260115] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11233), 1,
      anon_sym_DOT,
    STATE(7434), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260129] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11235), 1,
      anon_sym_PLUS_EQ,
    STATE(7435), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260143] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11237), 1,
      anon_sym_RBRACK,
    STATE(7436), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260157] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11239), 1,
      anon_sym_SQUOTE2,
    STATE(7437), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260171] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11241), 1,
      anon_sym_RPAREN,
    STATE(7438), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260185] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11243), 1,
      anon_sym_in,
    STATE(7439), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260199] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11245), 1,
      anon_sym_RPAREN,
    STATE(7440), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260213] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11247), 1,
      anon_sym_RBRACE,
    STATE(7441), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260227] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11249), 1,
      anon_sym_RPAREN,
    STATE(7442), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260241] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11251), 1,
      anon_sym_RPAREN,
    STATE(7443), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260255] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11253), 1,
      anon_sym_RPAREN,
    STATE(7444), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260269] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11253), 1,
      anon_sym_RBRACK,
    STATE(7445), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260283] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11253), 1,
      anon_sym_RBRACE,
    STATE(7446), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260297] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11255), 1,
      anon_sym_RBRACK,
    STATE(7447), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260311] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11257), 1,
      anon_sym_RBRACE,
    STATE(7448), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260325] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11259), 1,
      anon_sym_end,
    STATE(7449), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260339] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11261), 1,
      anon_sym_RBRACE,
    STATE(7450), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260353] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11263), 1,
      anon_sym_RPAREN,
    STATE(7451), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260367] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11265), 1,
      anon_sym_RPAREN,
    STATE(7452), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260381] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4140), 1,
      anon_sym_RBRACK,
    STATE(7453), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260395] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11267), 1,
      anon_sym_DOT,
    STATE(7454), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260409] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11269), 1,
      anon_sym_RPAREN,
    STATE(7455), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260423] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11271), 1,
      anon_sym_DQUOTE,
    STATE(7456), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260437] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11273), 1,
      anon_sym_DOT,
    STATE(7457), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260451] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11275), 1,
      anon_sym_with,
    STATE(7458), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260465] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11277), 1,
      anon_sym_with,
    STATE(7459), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260479] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11279), 1,
      anon_sym_DOT,
    STATE(7460), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260493] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7077), 1,
      anon_sym_in,
    STATE(7461), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260507] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11281), 1,
      anon_sym_RPAREN,
    STATE(7462), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260521] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11283), 1,
      anon_sym_RBRACE,
    STATE(7463), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260535] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1511), 1,
      anon_sym_RPAREN,
    STATE(7464), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260549] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11285), 1,
      anon_sym_with,
    STATE(7465), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260563] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11287), 1,
      anon_sym_with,
    STATE(7466), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260577] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11289), 1,
      anon_sym_SQUOTE2,
    STATE(7467), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260591] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11291), 1,
      anon_sym_RPAREN,
    STATE(7468), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260605] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11293), 1,
      sym__lowercase_identifier,
    STATE(7469), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260619] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11295), 1,
      anon_sym_RPAREN,
    STATE(7470), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260633] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11297), 1,
      anon_sym_DOT,
    STATE(7471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260647] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11299), 1,
      anon_sym_RBRACK,
    STATE(7472), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260661] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1525), 1,
      anon_sym_RPAREN,
    STATE(7473), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260675] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11301), 1,
      anon_sym_DQUOTE,
    STATE(7474), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260689] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11303), 1,
      anon_sym_RPAREN,
    STATE(7475), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260703] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11305), 1,
      anon_sym_RBRACK,
    STATE(7476), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260717] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11307), 1,
      anon_sym_DASH_GT,
    STATE(7477), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260731] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11309), 1,
      anon_sym_DASH_GT,
    STATE(7478), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260745] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11311), 1,
      anon_sym_RBRACE,
    STATE(7479), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260759] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11313), 1,
      anon_sym_RPAREN,
    STATE(7480), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260773] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11315), 1,
      anon_sym_RPAREN,
    STATE(7481), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260787] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11315), 1,
      anon_sym_RBRACK,
    STATE(7482), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260801] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11315), 1,
      anon_sym_RBRACE,
    STATE(7483), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260815] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4427), 1,
      anon_sym_RPAREN,
    STATE(7484), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260829] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11317), 1,
      anon_sym_RPAREN,
    STATE(7485), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260843] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11319), 1,
      anon_sym_RBRACE,
    STATE(7486), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260857] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11321), 1,
      anon_sym_RPAREN,
    STATE(7487), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260871] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4730), 1,
      anon_sym_COMMA,
    STATE(7488), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260885] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9189), 1,
      anon_sym_GT_RBRACE,
    STATE(7489), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260899] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11323), 1,
      anon_sym_in,
    STATE(7490), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260913] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11325), 1,
      anon_sym_DOT,
    STATE(7491), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260927] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11327), 1,
      anon_sym_of,
    STATE(7492), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260941] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11329), 1,
      anon_sym_end,
    STATE(7493), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260955] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11331), 1,
      anon_sym_with,
    STATE(7494), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260969] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11333), 1,
      anon_sym_with,
    STATE(7495), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260983] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11335), 1,
      anon_sym_DOT,
    STATE(7496), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [260997] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11337), 1,
      anon_sym_in,
    STATE(7497), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261011] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11339), 1,
      anon_sym_RPAREN,
    STATE(7498), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261025] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11341), 1,
      anon_sym_LT_DASH,
    STATE(7499), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261039] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11343), 1,
      anon_sym_RBRACK,
    STATE(7500), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261053] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11345), 1,
      anon_sym_with,
    STATE(7501), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261067] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11347), 1,
      anon_sym_with,
    STATE(7502), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261081] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1543), 1,
      anon_sym_RPAREN,
    STATE(7503), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261095] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11349), 1,
      anon_sym_in,
    STATE(7504), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261109] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11351), 1,
      anon_sym_RBRACE,
    STATE(7505), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261123] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11353), 1,
      sym__lowercase_identifier,
    STATE(7506), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261137] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11355), 1,
      anon_sym_DOT,
    STATE(7507), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261151] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11357), 1,
      anon_sym_RBRACE,
    STATE(7508), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261165] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11359), 1,
      anon_sym_EQ,
    STATE(7509), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261179] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__uppercase_identifier,
    STATE(7510), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261193] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11361), 1,
      anon_sym_RPAREN,
    STATE(7511), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261207] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11363), 1,
      anon_sym_RBRACK,
    STATE(7512), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261221] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11365), 1,
      anon_sym_RPAREN,
    STATE(7513), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261235] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11367), 1,
      sym__lowercase_identifier,
    STATE(7514), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261249] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11369), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7515), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261263] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11371), 1,
      anon_sym_RPAREN,
    STATE(7516), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261277] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11373), 1,
      anon_sym_RPAREN,
    STATE(7517), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261291] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11373), 1,
      anon_sym_RBRACK,
    STATE(7518), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261305] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11373), 1,
      anon_sym_RBRACE,
    STATE(7519), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261319] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11375), 1,
      anon_sym_DOT,
    STATE(7520), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261333] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11377), 1,
      anon_sym_in,
    STATE(7521), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261347] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11379), 1,
      anon_sym_RBRACE,
    STATE(7522), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261361] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11381), 1,
      anon_sym_RPAREN,
    STATE(7523), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261375] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11383), 1,
      anon_sym_in,
    STATE(7524), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261389] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11385), 1,
      sym__lowercase_identifier,
    STATE(7525), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261403] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8082), 1,
      anon_sym_COMMA,
    STATE(7526), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261417] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11387), 1,
      anon_sym_DOT,
    STATE(7527), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261431] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11389), 1,
      anon_sym_RPAREN,
    STATE(7528), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261445] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11391), 1,
      anon_sym_in,
    STATE(7529), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261459] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11393), 1,
      anon_sym_with,
    STATE(7530), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261473] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11395), 1,
      anon_sym_with,
    STATE(7531), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261487] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11397), 1,
      anon_sym_DOT,
    STATE(7532), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261501] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11399), 1,
      anon_sym_with,
    STATE(7533), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261515] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11401), 1,
      anon_sym_RBRACE,
    STATE(7534), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261529] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4650), 1,
      anon_sym_COMMA,
    STATE(7535), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261543] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8267), 1,
      anon_sym_DASH_GT,
    STATE(7536), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261557] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11403), 1,
      anon_sym_with,
    STATE(7537), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261571] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11405), 1,
      anon_sym_with,
    STATE(7538), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261585] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11407), 1,
      anon_sym_DASH_GT,
    STATE(7539), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261599] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11409), 1,
      anon_sym_RPAREN,
    STATE(7540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261613] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11411), 1,
      anon_sym_DASH_GT,
    STATE(7541), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261627] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11413), 1,
      anon_sym_with,
    STATE(7542), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261641] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11415), 1,
      anon_sym_DOT,
    STATE(7543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261655] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11417), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7544), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261669] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11419), 1,
      anon_sym_DASH_GT,
    STATE(7545), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261683] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11421), 1,
      anon_sym_in,
    STATE(7546), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261697] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11423), 1,
      anon_sym_RPAREN,
    STATE(7547), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261711] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11425), 1,
      anon_sym_EQ,
    STATE(7548), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261725] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11427), 1,
      anon_sym_DOT,
    STATE(7549), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261739] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11429), 1,
      anon_sym_RBRACK,
    STATE(7550), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261753] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11431), 1,
      anon_sym_RPAREN,
    STATE(7551), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261767] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11433), 1,
      anon_sym_RPAREN,
    STATE(7552), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261781] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11435), 1,
      anon_sym_RPAREN,
    STATE(7553), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261795] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11435), 1,
      anon_sym_RBRACK,
    STATE(7554), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261809] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11435), 1,
      anon_sym_RBRACE,
    STATE(7555), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261823] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11437), 1,
      anon_sym_GT,
    STATE(7556), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261837] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11439), 1,
      sym__lowercase_identifier,
    STATE(7557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261851] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11441), 1,
      anon_sym_in,
    STATE(7558), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261865] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11443), 1,
      anon_sym_RPAREN,
    STATE(7559), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261879] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11445), 1,
      anon_sym_RBRACE,
    STATE(7560), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261893] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11447), 1,
      anon_sym_LT_DASH,
    STATE(7561), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261907] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11449), 1,
      anon_sym_DOT,
    STATE(7562), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261921] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11451), 1,
      anon_sym_RPAREN,
    STATE(7563), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261935] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11453), 1,
      anon_sym_RPAREN,
    STATE(7564), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261949] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11455), 1,
      anon_sym_with,
    STATE(7565), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261963] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11457), 1,
      anon_sym_with,
    STATE(7566), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261977] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11459), 1,
      anon_sym_DOT,
    STATE(7567), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [261991] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11461), 1,
      anon_sym_SQUOTE2,
    STATE(7568), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262005] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11463), 1,
      anon_sym_RBRACE,
    STATE(7569), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262019] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11465), 1,
      anon_sym_DASH_GT,
    STATE(7570), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262033] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11467), 1,
      anon_sym_in,
    STATE(7571), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262047] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11469), 1,
      anon_sym_with,
    STATE(7572), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262061] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11471), 1,
      anon_sym_with,
    STATE(7573), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262075] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11473), 1,
      anon_sym_RBRACE,
    STATE(7574), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262089] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11475), 1,
      anon_sym_RPAREN,
    STATE(7575), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262103] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11477), 1,
      anon_sym_RPAREN,
    STATE(7576), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262117] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11479), 1,
      anon_sym_RBRACE,
    STATE(7577), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262131] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11481), 1,
      anon_sym_DOT,
    STATE(7578), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262145] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11483), 1,
      anon_sym_RBRACE,
    STATE(7579), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262159] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11485), 1,
      anon_sym_RPAREN,
    STATE(7580), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262173] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11487), 1,
      anon_sym_DASH_GT,
    STATE(7581), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262187] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11489), 1,
      anon_sym_RPAREN,
    STATE(7582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262201] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11491), 1,
      anon_sym_RBRACK,
    STATE(7583), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262215] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11493), 1,
      anon_sym_DQUOTE,
    STATE(7584), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262229] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11495), 1,
      anon_sym_in,
    STATE(7585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262243] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11497), 1,
      anon_sym_RBRACE,
    STATE(7586), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262257] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11499), 1,
      anon_sym_RPAREN,
    STATE(7587), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262271] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11501), 1,
      anon_sym_RPAREN,
    STATE(7588), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262285] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11501), 1,
      anon_sym_RBRACK,
    STATE(7589), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262299] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11501), 1,
      anon_sym_RBRACE,
    STATE(7590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262313] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11503), 1,
      anon_sym_RPAREN,
    STATE(7591), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262327] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11505), 1,
      anon_sym_DOT,
    STATE(7592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262341] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9185), 1,
      anon_sym_GT_RBRACE,
    STATE(7593), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262355] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7798), 1,
      anon_sym_COMMA,
    STATE(7594), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262369] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11507), 1,
      anon_sym_RPAREN,
    STATE(7595), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262383] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11509), 1,
      anon_sym_RBRACK,
    STATE(7596), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262397] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11511), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7597), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262411] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11513), 1,
      anon_sym_DOT,
    STATE(7598), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262425] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11515), 1,
      anon_sym_RPAREN,
    STATE(7599), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262439] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11517), 1,
      anon_sym_RBRACK,
    STATE(7600), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262453] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11519), 1,
      anon_sym_with,
    STATE(7601), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262467] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11521), 1,
      anon_sym_with,
    STATE(7602), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262481] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11523), 1,
      anon_sym_DOT,
    STATE(7603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262495] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8357), 1,
      anon_sym_end,
    STATE(7604), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262509] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11525), 1,
      anon_sym_RBRACE,
    STATE(7605), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262523] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11527), 1,
      aux_sym_number_token1,
    STATE(7606), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262537] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11529), 1,
      anon_sym_RBRACK,
    STATE(7607), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262551] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11531), 1,
      anon_sym_with,
    STATE(7608), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262565] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11533), 1,
      anon_sym_with,
    STATE(7609), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262579] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11535), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7610), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262593] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11537), 1,
      anon_sym_RPAREN,
    STATE(7611), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262607] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9574), 1,
      anon_sym_GT_RBRACE,
    STATE(7612), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262621] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10443), 1,
      anon_sym_RBRACE,
    STATE(7613), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262635] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11539), 1,
      anon_sym_DOT,
    STATE(7614), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262649] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11541), 1,
      anon_sym_RPAREN,
    STATE(7615), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262663] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9373), 1,
      anon_sym_GT_RBRACE,
    STATE(7616), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262677] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11543), 1,
      anon_sym_end,
    STATE(7617), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262691] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11545), 1,
      anon_sym_RPAREN,
    STATE(7618), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262705] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8182), 1,
      anon_sym_COMMA,
    STATE(7619), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262719] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11547), 1,
      anon_sym_RPAREN,
    STATE(7620), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262733] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8426), 1,
      anon_sym_end,
    STATE(7621), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262747] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10547), 1,
      anon_sym_RBRACE,
    STATE(7622), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262761] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11549), 1,
      anon_sym_RPAREN,
    STATE(7623), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262775] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11551), 1,
      anon_sym_RPAREN,
    STATE(7624), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262789] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11551), 1,
      anon_sym_RBRACK,
    STATE(7625), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262803] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11551), 1,
      anon_sym_RBRACE,
    STATE(7626), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262817] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8466), 1,
      anon_sym_RPAREN,
    STATE(7627), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262831] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11553), 1,
      anon_sym_RPAREN,
    STATE(7628), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262845] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11555), 1,
      anon_sym_RBRACE,
    STATE(7629), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262859] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11557), 1,
      anon_sym_RPAREN,
    STATE(7630), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262873] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11559), 1,
      anon_sym_SQUOTE2,
    STATE(7631), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262887] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11561), 1,
      anon_sym_RPAREN,
    STATE(7632), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262901] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11563), 1,
      anon_sym_DQUOTE,
    STATE(7633), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262915] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11565), 1,
      anon_sym_DOT,
    STATE(7634), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262929] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11567), 1,
      anon_sym_EQ,
    STATE(7635), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262943] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11569), 1,
      anon_sym_DOT,
    STATE(7636), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262957] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8100), 1,
      anon_sym_DASH_GT,
    STATE(7637), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262971] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11571), 1,
      anon_sym_in,
    STATE(7638), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262985] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11573), 1,
      anon_sym_DASH_GT,
    STATE(7639), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [262999] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11575), 1,
      anon_sym_DOT,
    STATE(7640), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263013] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8573), 1,
      anon_sym_end,
    STATE(7641), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263027] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11577), 1,
      anon_sym_RPAREN,
    STATE(7642), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263041] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11579), 1,
      anon_sym_done,
    STATE(7643), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263055] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11581), 1,
      anon_sym_RPAREN,
    STATE(7644), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263069] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11583), 1,
      sym__lowercase_identifier,
    STATE(7645), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263083] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11585), 1,
      anon_sym_RPAREN,
    STATE(7646), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263097] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11587), 1,
      anon_sym_RPAREN,
    STATE(7647), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263111] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11587), 1,
      anon_sym_RBRACK,
    STATE(7648), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263125] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11587), 1,
      anon_sym_RBRACE,
    STATE(7649), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263139] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11589), 1,
      anon_sym_RPAREN,
    STATE(7650), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263153] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11591), 1,
      anon_sym_DOT,
    STATE(7651), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263167] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11593), 1,
      anon_sym_RPAREN,
    STATE(7652), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263181] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11595), 1,
      anon_sym_in,
    STATE(7653), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263195] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9377), 1,
      anon_sym_GT_RBRACE,
    STATE(7654), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263209] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11597), 1,
      anon_sym_in,
    STATE(7655), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263223] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3398), 1,
      anon_sym_COMMA,
    STATE(7656), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263237] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11599), 1,
      anon_sym_LT_DASH,
    STATE(7657), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263251] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11601), 1,
      anon_sym_RPAREN,
    STATE(7658), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263265] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11603), 1,
      anon_sym_SQUOTE2,
    STATE(7659), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263279] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11605), 1,
      anon_sym_RPAREN,
    STATE(7660), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263293] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11605), 1,
      anon_sym_RBRACK,
    STATE(7661), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263307] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11605), 1,
      anon_sym_RBRACE,
    STATE(7662), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263321] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8406), 1,
      anon_sym_end,
    STATE(7663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263335] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8466), 1,
      anon_sym_end,
    STATE(7664), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263349] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11607), 1,
      anon_sym_SQUOTE2,
    STATE(7665), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263363] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11609), 1,
      anon_sym_DQUOTE,
    STATE(7666), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263377] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8495), 1,
      anon_sym_end,
    STATE(7667), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263391] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11611), 1,
      anon_sym_RBRACK,
    STATE(7668), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263405] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11613), 1,
      anon_sym_DASH_GT,
    STATE(7669), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263419] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11615), 1,
      anon_sym_RPAREN,
    STATE(7670), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263433] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11615), 1,
      anon_sym_RBRACK,
    STATE(7671), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263447] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11615), 1,
      anon_sym_RBRACE,
    STATE(7672), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263461] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11617), 1,
      anon_sym_DQUOTE,
    STATE(7673), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263475] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11619), 1,
      anon_sym_RPAREN,
    STATE(7674), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263489] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11621), 1,
      anon_sym_in,
    STATE(7675), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263503] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11623), 1,
      sym__lowercase_identifier,
    STATE(7676), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263517] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11625), 1,
      anon_sym_RPAREN,
    STATE(7677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263531] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11627), 1,
      anon_sym_RPAREN,
    STATE(7678), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263545] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4485), 1,
      anon_sym_COMMA,
    STATE(7679), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263559] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11629), 1,
      anon_sym_RPAREN,
    STATE(7680), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263573] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11629), 1,
      anon_sym_RBRACK,
    STATE(7681), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263587] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11629), 1,
      anon_sym_RBRACE,
    STATE(7682), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263601] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11631), 1,
      anon_sym_LT_DASH,
    STATE(7683), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263615] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11633), 1,
      anon_sym_RPAREN,
    STATE(7684), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263629] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11635), 1,
      anon_sym_EQ,
    STATE(7685), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263643] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11637), 1,
      anon_sym_with,
    STATE(7686), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263657] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4128), 1,
      anon_sym_COMMA,
    STATE(7687), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263671] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11639), 1,
      anon_sym_LT_DASH,
    STATE(7688), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263685] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11641), 1,
      anon_sym_RBRACE,
    STATE(7689), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263699] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11643), 1,
      anon_sym_RPAREN,
    STATE(7690), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263713] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11643), 1,
      anon_sym_RBRACK,
    STATE(7691), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263727] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11643), 1,
      anon_sym_RBRACE,
    STATE(7692), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263741] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11645), 1,
      anon_sym_SQUOTE2,
    STATE(7693), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263755] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3356), 1,
      anon_sym_COMMA,
    STATE(7694), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263769] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11647), 1,
      sym__lowercase_identifier,
    STATE(7695), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263783] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11649), 1,
      anon_sym_LT_DASH,
    STATE(7696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263797] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11651), 1,
      anon_sym_DASH_GT,
    STATE(7697), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263811] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11653), 1,
      anon_sym_end,
    STATE(7698), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263825] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1553), 1,
      anon_sym_RPAREN,
    STATE(7699), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263839] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11655), 1,
      anon_sym_RPAREN,
    STATE(7700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263853] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11655), 1,
      anon_sym_RBRACK,
    STATE(7701), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263867] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11655), 1,
      anon_sym_RBRACE,
    STATE(7702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263881] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11657), 1,
      anon_sym_RBRACE,
    STATE(7703), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263895] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11659), 1,
      anon_sym_EQ,
    STATE(7704), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263909] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11661), 1,
      anon_sym_DQUOTE,
    STATE(7705), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263923] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3466), 1,
      anon_sym_COMMA,
    STATE(7706), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263937] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11663), 1,
      anon_sym_RBRACK,
    STATE(7707), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263951] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11665), 1,
      anon_sym_RBRACE,
    STATE(7708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263965] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11667), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7709), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263979] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11669), 1,
      anon_sym_RPAREN,
    STATE(7710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [263993] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11669), 1,
      anon_sym_RBRACK,
    STATE(7711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264007] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11669), 1,
      anon_sym_RBRACE,
    STATE(7712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264021] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11671), 1,
      anon_sym_end,
    STATE(7713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264035] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11673), 1,
      anon_sym_SQUOTE2,
    STATE(7714), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264049] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11675), 1,
      anon_sym_DOT,
    STATE(7715), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264063] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11677), 1,
      anon_sym_RPAREN,
    STATE(7716), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264077] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11679), 1,
      anon_sym_in,
    STATE(7717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264091] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11681), 1,
      sym__lowercase_identifier,
    STATE(7718), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264105] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11683), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7719), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264119] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11685), 1,
      anon_sym_RBRACE,
    STATE(7720), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264133] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11687), 1,
      anon_sym_DOT,
    STATE(7721), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264147] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11689), 1,
      anon_sym_end,
    STATE(7722), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264161] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11691), 1,
      anon_sym_RPAREN,
    STATE(7723), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264175] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11693), 1,
      anon_sym_DQUOTE,
    STATE(7724), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264189] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11695), 1,
      anon_sym_in,
    STATE(7725), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264203] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11697), 1,
      anon_sym_RBRACK,
    STATE(7726), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264217] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11699), 1,
      anon_sym_DASH_GT,
    STATE(7727), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264231] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11701), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264245] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11703), 1,
      anon_sym_DOT,
    STATE(7729), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264259] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3158), 1,
      anon_sym_DASH_GT,
    STATE(7730), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264273] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11705), 1,
      sym__lowercase_identifier,
    STATE(7731), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264287] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11707), 1,
      anon_sym_RPAREN,
    STATE(7732), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264301] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11709), 1,
      sym__lowercase_identifier,
    STATE(7733), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264315] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11711), 1,
      anon_sym_RBRACK,
    STATE(7734), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264329] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8557), 1,
      anon_sym_RPAREN,
    STATE(7735), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264343] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11713), 1,
      anon_sym_DASH_GT,
    STATE(7736), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264357] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11715), 1,
      anon_sym_DASH_GT,
    STATE(7737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264371] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11717), 1,
      anon_sym_in,
    STATE(7738), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264385] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11719), 1,
      anon_sym_RBRACK,
    STATE(7739), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264399] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11721), 1,
      sym__lowercase_identifier,
    STATE(7740), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264413] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11723), 1,
      anon_sym_in,
    STATE(7741), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264427] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11725), 1,
      anon_sym_DASH_GT,
    STATE(7742), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264441] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6761), 1,
      anon_sym_DASH_GT,
    STATE(7743), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264455] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11727), 1,
      anon_sym_LT_DASH,
    STATE(7744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264469] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11729), 1,
      anon_sym_RPAREN,
    STATE(7745), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264483] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11731), 1,
      anon_sym_PLUS_EQ,
    STATE(7746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264497] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11733), 1,
      anon_sym_STAR,
    STATE(7747), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264511] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11735), 1,
      anon_sym_in,
    STATE(7748), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264525] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11737), 1,
      anon_sym_RBRACE,
    STATE(7749), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264539] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8040), 1,
      anon_sym_DASH_GT,
    STATE(7750), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264553] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11739), 1,
      anon_sym_end,
    STATE(7751), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264567] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11741), 1,
      anon_sym_RBRACK,
    STATE(7752), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264581] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11743), 1,
      anon_sym_RBRACE,
    STATE(7753), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264595] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11745), 1,
      anon_sym_in,
    STATE(7754), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264609] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11747), 1,
      anon_sym_in,
    STATE(7755), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264623] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11749), 1,
      anon_sym_DOT,
    STATE(7756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264637] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11751), 1,
      anon_sym_RPAREN,
    STATE(7757), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264651] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11753), 1,
      anon_sym_RPAREN,
    STATE(7758), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264665] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11755), 1,
      anon_sym_RPAREN,
    STATE(7759), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264679] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5007), 1,
      anon_sym_GT,
    STATE(7760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264693] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11757), 1,
      anon_sym_RPAREN,
    STATE(7761), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264707] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11759), 1,
      anon_sym_COMMA,
    STATE(7762), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264721] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11761), 1,
      anon_sym_with,
    STATE(7763), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264735] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11763), 1,
      sym__lowercase_identifier,
    STATE(7764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264749] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3150), 1,
      anon_sym_DASH_GT,
    STATE(7765), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264763] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11765), 1,
      anon_sym_with,
    STATE(7766), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264777] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11767), 1,
      anon_sym_RBRACK,
    STATE(7767), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264791] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11769), 1,
      anon_sym_EQ,
    STATE(7768), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264805] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11771), 1,
      anon_sym_DASH_GT,
    STATE(7769), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264819] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5830), 1,
      anon_sym_in,
    STATE(7770), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264833] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11773), 1,
      anon_sym_with,
    STATE(7771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264847] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11775), 1,
      anon_sym_RPAREN,
    STATE(7772), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264861] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11777), 1,
      anon_sym_EQ,
    STATE(7773), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264875] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10551), 1,
      anon_sym_RBRACE,
    STATE(7774), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264889] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11779), 1,
      anon_sym_PLUS_EQ,
    STATE(7775), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264903] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11781), 1,
      anon_sym_with,
    STATE(7776), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264917] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11783), 1,
      anon_sym_DOT,
    STATE(7777), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264931] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11785), 1,
      anon_sym_RBRACE,
    STATE(7778), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264945] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11787), 1,
      anon_sym_RPAREN,
    STATE(7779), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264959] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11789), 1,
      anon_sym_RBRACE,
    STATE(7780), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264973] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11791), 1,
      anon_sym_RBRACK,
    STATE(7781), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [264987] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11793), 1,
      anon_sym_RBRACE,
    STATE(7782), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265001] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11795), 1,
      anon_sym_EQ,
    STATE(7783), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265015] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11797), 1,
      sym__lowercase_identifier,
    STATE(7784), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265029] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11799), 1,
      anon_sym_RBRACE,
    STATE(7785), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265043] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11801), 1,
      anon_sym_RPAREN,
    STATE(7786), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265057] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11803), 1,
      anon_sym_RPAREN,
    STATE(7787), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265071] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11805), 1,
      anon_sym_RBRACE,
    STATE(7788), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265085] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11807), 1,
      anon_sym_RBRACK,
    STATE(7789), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265099] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11809), 1,
      anon_sym_RBRACK,
    STATE(7790), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265113] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8289), 1,
      anon_sym_COMMA,
    STATE(7791), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265127] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11811), 1,
      anon_sym_RBRACE,
    STATE(7792), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265141] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11813), 1,
      sym__lowercase_identifier,
    STATE(7793), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265155] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11815), 1,
      sym__lowercase_identifier,
    STATE(7794), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265169] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11817), 1,
      anon_sym_RBRACE,
    STATE(7795), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265183] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11819), 1,
      anon_sym_RPAREN,
    STATE(7796), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265197] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11821), 1,
      anon_sym_RPAREN,
    STATE(7797), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265211] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11823), 1,
      anon_sym_EQ,
    STATE(7798), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265225] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11825), 1,
      anon_sym_RPAREN,
    STATE(7799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265239] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11827), 1,
      anon_sym_EQ,
    STATE(7800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265253] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11829), 1,
      anon_sym_RPAREN,
    STATE(7801), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265267] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11831), 1,
      anon_sym_DOT,
    STATE(7802), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265281] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11833), 1,
      anon_sym_DASH_GT,
    STATE(7803), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265295] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11835), 1,
      sym__lowercase_identifier,
    STATE(7804), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265309] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3670), 1,
      anon_sym_COMMA,
    STATE(7805), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265323] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11837), 1,
      anon_sym_GT,
    STATE(7806), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265337] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11839), 1,
      anon_sym_RPAREN,
    STATE(7807), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265351] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11841), 1,
      anon_sym_RPAREN,
    STATE(7808), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265365] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11843), 1,
      anon_sym_PIPE,
    STATE(7809), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265379] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11845), 1,
      anon_sym_RPAREN,
    STATE(7810), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265393] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11847), 1,
      anon_sym_RPAREN,
    STATE(7811), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265407] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11849), 1,
      anon_sym_RBRACE,
    STATE(7812), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265421] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11851), 1,
      anon_sym_RBRACE,
    STATE(7813), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265435] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11853), 1,
      anon_sym_RPAREN,
    STATE(7814), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265449] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11855), 1,
      anon_sym_EQ,
    STATE(7815), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265463] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11857), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7816), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265477] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11859), 1,
      anon_sym_RPAREN,
    STATE(7817), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265491] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11861), 1,
      anon_sym_RBRACK,
    STATE(7818), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265505] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11863), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7819), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265519] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11865), 1,
      anon_sym_RPAREN,
    STATE(7820), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265533] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11867), 1,
      anon_sym_DOT_DOT,
    STATE(7821), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265547] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11869), 1,
      anon_sym_DOT_DOT,
    STATE(7822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265561] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11871), 1,
      anon_sym_DOT_DOT,
    STATE(7823), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265575] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11873), 1,
      anon_sym_RPAREN,
    STATE(7824), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265589] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11875), 1,
      anon_sym_RPAREN,
    STATE(7825), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265603] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4126), 1,
      anon_sym_COLON_COLON,
    STATE(7826), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265617] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11877), 1,
      anon_sym_RBRACE,
    STATE(7827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265631] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11879), 1,
      anon_sym_RPAREN,
    STATE(7828), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265645] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11881), 1,
      anon_sym_in,
    STATE(7829), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265659] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11883), 1,
      sym__lowercase_identifier,
    STATE(7830), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265673] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11885), 1,
      anon_sym_DASH_GT,
    STATE(7831), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265687] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11887), 1,
      anon_sym_PLUS_EQ,
    STATE(7832), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265701] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4150), 1,
      anon_sym_COMMA,
    STATE(7833), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265715] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11889), 1,
      sym__lowercase_identifier,
    STATE(7834), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265729] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11891), 1,
      anon_sym_COMMA,
    STATE(7835), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265743] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11893), 1,
      sym__lowercase_identifier,
    STATE(7836), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265757] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11895), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7837), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265771] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11897), 1,
      anon_sym_RBRACK,
    STATE(7838), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265785] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11899), 1,
      anon_sym_STAR,
    STATE(7839), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265799] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11901), 1,
      anon_sym_RPAREN,
    STATE(7840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265813] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11903), 1,
      anon_sym_in,
    STATE(7841), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265827] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11905), 1,
      anon_sym_RBRACE,
    STATE(7842), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265841] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11907), 1,
      anon_sym_end,
    STATE(7843), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265855] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11909), 1,
      anon_sym_RPAREN,
    STATE(7844), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265869] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11911), 1,
      anon_sym_DOT,
    STATE(7845), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265883] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11913), 1,
      anon_sym_RPAREN,
    STATE(7846), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265897] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11915), 1,
      anon_sym_EQ,
    STATE(7847), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265911] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11917), 1,
      anon_sym_COMMA,
    STATE(7848), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265925] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11919), 1,
      anon_sym_RPAREN,
    STATE(7849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265939] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11921), 1,
      anon_sym_with,
    STATE(7850), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265953] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11923), 1,
      sym__lowercase_identifier,
    STATE(7851), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265967] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11925), 1,
      sym__lowercase_identifier,
    STATE(7852), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265981] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11927), 1,
      anon_sym_RBRACK,
    STATE(7853), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [265995] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11929), 1,
      anon_sym_with,
    STATE(7854), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266009] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11931), 1,
      anon_sym_EQ,
    STATE(7855), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266023] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11933), 1,
      anon_sym_RBRACE,
    STATE(7856), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266037] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11935), 1,
      anon_sym_RBRACE,
    STATE(7857), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266051] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4997), 1,
      anon_sym_GT,
    STATE(7858), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266065] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11937), 1,
      anon_sym_DASH_GT,
    STATE(7859), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266079] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11939), 1,
      anon_sym_in,
    STATE(7860), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266093] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11941), 1,
      anon_sym_RPAREN,
    STATE(7861), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266107] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11943), 1,
      anon_sym_RPAREN,
    STATE(7862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266121] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1529), 1,
      anon_sym_RPAREN,
    STATE(7863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266135] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9662), 1,
      anon_sym_GT_RBRACE,
    STATE(7864), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266149] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8214), 1,
      anon_sym_DASH_GT,
    STATE(7865), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266163] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11945), 1,
      sym__lowercase_identifier,
    STATE(7866), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266177] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11947), 1,
      anon_sym_EQ,
    STATE(7867), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266191] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11949), 1,
      anon_sym_in,
    STATE(7868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266205] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11951), 1,
      anon_sym_DOT,
    STATE(7869), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266219] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11953), 1,
      anon_sym_EQ,
    STATE(7870), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266233] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11955), 1,
      anon_sym_EQ,
    STATE(7871), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266247] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1565), 1,
      anon_sym_RPAREN,
    STATE(7872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266261] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11957), 1,
      anon_sym_RPAREN,
    STATE(7873), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266275] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8557), 1,
      anon_sym_end,
    STATE(7874), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266289] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11959), 1,
      anon_sym_RBRACE,
    STATE(7875), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266303] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11961), 1,
      anon_sym_RBRACK,
    STATE(7876), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266317] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11963), 1,
      anon_sym_DASH_GT,
    STATE(7877), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266331] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11965), 1,
      anon_sym_RPAREN,
    STATE(7878), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266345] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11967), 1,
      anon_sym_RBRACE,
    STATE(7879), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266359] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11969), 1,
      anon_sym_RPAREN,
    STATE(7880), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266373] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11969), 1,
      anon_sym_RBRACK,
    STATE(7881), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266387] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11969), 1,
      anon_sym_RBRACE,
    STATE(7882), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266401] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11971), 1,
      anon_sym_RBRACK,
    STATE(7883), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266415] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9681), 1,
      anon_sym_GT_RBRACE,
    STATE(7884), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266429] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11973), 1,
      anon_sym_DOT,
    STATE(7885), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266443] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11975), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7886), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266457] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11977), 1,
      anon_sym_SQUOTE2,
    STATE(7887), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266471] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11979), 1,
      anon_sym_DQUOTE,
    STATE(7888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266485] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8452), 1,
      anon_sym_RPAREN,
    STATE(7889), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266499] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11981), 1,
      anon_sym_RPAREN,
    STATE(7890), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266513] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9675), 1,
      anon_sym_GT_RBRACE,
    STATE(7891), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266527] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11983), 1,
      anon_sym_RPAREN,
    STATE(7892), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266541] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11985), 1,
      anon_sym_COMMA,
    STATE(7893), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266555] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8460), 1,
      anon_sym_end,
    STATE(7894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266569] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11987), 1,
      anon_sym_RPAREN,
    STATE(7895), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266583] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10761), 1,
      anon_sym_RPAREN,
    STATE(7896), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266597] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8569), 1,
      anon_sym_RPAREN,
    STATE(7897), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266611] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11989), 1,
      anon_sym_EQ,
    STATE(7898), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266625] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11991), 1,
      anon_sym_EQ,
    STATE(7899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266639] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11993), 1,
      anon_sym_EQ,
    STATE(7900), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266653] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1557), 1,
      anon_sym_RPAREN,
    STATE(7901), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266667] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11995), 1,
      anon_sym_RPAREN,
    STATE(7902), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266681] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11997), 1,
      anon_sym_SQUOTE2,
    STATE(7903), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266695] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(11999), 1,
      anon_sym_DQUOTE,
    STATE(7904), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266709] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12001), 1,
      anon_sym_DASH_GT,
    STATE(7905), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266723] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12003), 1,
      sym__lowercase_identifier,
    STATE(7906), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266737] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12005), 1,
      sym__lowercase_identifier,
    STATE(7907), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266751] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12007), 1,
      anon_sym_in,
    STATE(7908), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266765] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12009), 1,
      sym__lowercase_identifier,
    STATE(7909), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266779] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12011), 1,
      anon_sym_RPAREN,
    STATE(7910), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266793] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8452), 1,
      anon_sym_end,
    STATE(7911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266807] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12013), 1,
      anon_sym_RPAREN,
    STATE(7912), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266821] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12015), 1,
      anon_sym_EQ,
    STATE(7913), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266835] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12017), 1,
      anon_sym_RBRACK,
    STATE(7914), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266849] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12019), 1,
      anon_sym_DASH_GT,
    STATE(7915), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266863] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12021), 1,
      sym__lowercase_identifier,
    STATE(7916), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266877] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12023), 1,
      anon_sym_EQ,
    STATE(7917), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266891] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12025), 1,
      sym_indexing_operator,
    STATE(7918), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266905] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12027), 1,
      sym__lowercase_identifier,
    STATE(7919), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266919] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12029), 1,
      anon_sym_DOT,
    STATE(7920), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266933] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12031), 1,
      sym__lowercase_identifier,
    STATE(7921), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266947] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12033), 1,
      ts_builtin_sym_end,
    STATE(7922), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266961] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12035), 1,
      anon_sym_RBRACE,
    STATE(7923), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266975] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12037), 1,
      anon_sym_RPAREN,
    STATE(7924), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [266989] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6142), 1,
      anon_sym_in,
    STATE(7925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267003] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12039), 1,
      anon_sym_DASH_GT,
    STATE(7926), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267017] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12041), 1,
      anon_sym_with,
    STATE(7927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267031] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12043), 1,
      anon_sym_DASH_GT,
    STATE(7928), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267045] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12045), 1,
      anon_sym_RBRACE,
    STATE(7929), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267059] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12047), 1,
      anon_sym_RPAREN,
    STATE(7930), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267073] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12049), 1,
      anon_sym_end,
    STATE(7931), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267087] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12051), 1,
      anon_sym_DASH_GT,
    STATE(7932), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267101] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12053), 1,
      anon_sym_SQUOTE2,
    STATE(7933), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267115] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12055), 1,
      anon_sym_end,
    STATE(7934), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267129] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12057), 1,
      anon_sym_RBRACK,
    STATE(7935), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267143] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12059), 1,
      anon_sym_end,
    STATE(7936), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267157] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12061), 1,
      anon_sym_RBRACK,
    STATE(7937), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267171] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12063), 1,
      anon_sym_RPAREN,
    STATE(7938), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267185] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12065), 1,
      anon_sym_end,
    STATE(7939), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267199] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8573), 1,
      anon_sym_RPAREN,
    STATE(7940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267213] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12067), 1,
      anon_sym_RBRACE,
    STATE(7941), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267227] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12069), 1,
      anon_sym_RPAREN,
    STATE(7942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267241] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12071), 1,
      anon_sym_RPAREN,
    STATE(7943), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267255] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12073), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267269] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12075), 1,
      anon_sym_in,
    STATE(7945), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267283] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12077), 1,
      anon_sym_DASH_GT,
    STATE(7946), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267297] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4931), 1,
      anon_sym_COMMA,
    STATE(7947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267311] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12079), 1,
      anon_sym_LT_DASH,
    STATE(7948), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267325] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12081), 1,
      anon_sym_in,
    STATE(7949), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267339] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12083), 1,
      anon_sym_RBRACE,
    STATE(7950), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267353] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12085), 1,
      sym__uppercase_identifier,
    STATE(7951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267367] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12087), 1,
      anon_sym_RPAREN,
    STATE(7952), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267381] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12089), 1,
      anon_sym_in,
    STATE(7953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267395] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12091), 1,
      anon_sym_DASH_GT,
    STATE(7954), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267409] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12093), 1,
      anon_sym_SQUOTE2,
    STATE(7955), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267423] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12095), 1,
      anon_sym_DASH_GT,
    STATE(7956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267437] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12097), 1,
      anon_sym_DQUOTE,
    STATE(7957), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267451] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12099), 1,
      anon_sym_RPAREN,
    STATE(7958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267465] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12101), 1,
      anon_sym_RBRACE,
    STATE(7959), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267479] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12103), 1,
      anon_sym_DASH_GT,
    STATE(7960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267493] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12105), 1,
      anon_sym_of,
    STATE(7961), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267507] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12107), 1,
      anon_sym_RPAREN,
    STATE(7962), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267521] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12109), 1,
      anon_sym_in,
    STATE(7963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267535] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12111), 1,
      anon_sym_RBRACK,
    STATE(7964), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267549] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12113), 1,
      anon_sym_RBRACE,
    STATE(7965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267563] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12115), 1,
      anon_sym_RPAREN,
    STATE(7966), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267577] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10721), 1,
      anon_sym_RBRACE,
    STATE(7967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267591] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12117), 1,
      anon_sym_RBRACK,
    STATE(7968), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267605] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12119), 1,
      anon_sym_type,
    STATE(7969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267619] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12121), 1,
      anon_sym_RBRACE,
    STATE(7970), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267633] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12123), 1,
      anon_sym_RBRACE,
    STATE(7971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267647] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12125), 1,
      anon_sym_RPAREN,
    STATE(7972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267661] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12127), 1,
      anon_sym_in,
    STATE(7973), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267675] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12129), 1,
      anon_sym_end,
    STATE(7974), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267689] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12131), 1,
      anon_sym_DASH_GT,
    STATE(7975), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267703] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12133), 1,
      anon_sym_RBRACK,
    STATE(7976), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267717] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(2802), 1,
      anon_sym_RBRACK,
    STATE(7977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267731] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12135), 1,
      anon_sym_DQUOTE,
    STATE(7978), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267745] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12137), 1,
      anon_sym_RPAREN,
    STATE(7979), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267759] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12139), 1,
      sym__lowercase_identifier,
    STATE(7980), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267773] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12141), 1,
      anon_sym_RPAREN,
    STATE(7981), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267787] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(713), 1,
      anon_sym_COLON_COLON,
    STATE(7982), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267801] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12143), 1,
      anon_sym_DOT,
    STATE(7983), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267815] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12145), 1,
      anon_sym_RPAREN,
    STATE(7984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267829] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12147), 1,
      anon_sym_RBRACK,
    STATE(7985), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267843] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12149), 1,
      anon_sym_RBRACK,
    STATE(7986), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267857] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12151), 1,
      anon_sym_RBRACE,
    STATE(7987), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267871] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12153), 1,
      anon_sym_RPAREN,
    STATE(7988), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267885] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8285), 1,
      anon_sym_DASH_GT,
    STATE(7989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267899] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12155), 1,
      sym__lowercase_identifier,
    STATE(7990), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267913] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12157), 1,
      sym__lowercase_identifier,
    STATE(7991), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267927] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12159), 1,
      anon_sym_RBRACE,
    STATE(7992), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267941] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12161), 1,
      anon_sym_RBRACE,
    STATE(7993), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267955] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12163), 1,
      anon_sym_RBRACK,
    STATE(7994), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267969] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12165), 1,
      anon_sym_RBRACK,
    STATE(7995), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267983] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12167), 1,
      anon_sym_RPAREN,
    STATE(7996), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [267997] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12169), 1,
      anon_sym_RBRACE,
    STATE(7997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268011] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12171), 1,
      anon_sym_DASH_GT,
    STATE(7998), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268025] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12173), 1,
      sym__lowercase_identifier,
    STATE(7999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268039] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12175), 1,
      anon_sym_DASH_GT,
    STATE(8000), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268053] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10207), 1,
      anon_sym_RBRACE,
    STATE(8001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268067] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12177), 1,
      anon_sym_PIPE,
    STATE(8002), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268081] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12179), 1,
      anon_sym_EQ,
    STATE(8003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268095] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12181), 1,
      anon_sym_DOT_DOT,
    STATE(8004), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268109] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12183), 1,
      anon_sym_DOT_DOT,
    STATE(8005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268123] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12185), 1,
      anon_sym_DOT_DOT,
    STATE(8006), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268137] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4540), 1,
      anon_sym_DASH_GT,
    STATE(8007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268151] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12187), 1,
      anon_sym_DASH_GT,
    STATE(8008), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268165] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12189), 1,
      anon_sym_in,
    STATE(8009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268179] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12191), 1,
      anon_sym_RPAREN,
    STATE(8010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268193] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12193), 1,
      anon_sym_DASH_GT,
    STATE(8011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268207] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10681), 1,
      anon_sym_RBRACE,
    STATE(8012), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268221] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12195), 1,
      anon_sym_RPAREN,
    STATE(8013), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268235] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12197), 1,
      anon_sym_DASH_GT,
    STATE(8014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268249] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12199), 1,
      anon_sym_RPAREN,
    STATE(8015), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268263] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12201), 1,
      anon_sym_RPAREN,
    STATE(8016), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268277] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12203), 1,
      anon_sym_DASH_GT,
    STATE(8017), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268291] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12205), 1,
      anon_sym_RBRACE,
    STATE(8018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268305] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12207), 1,
      anon_sym_RBRACK,
    STATE(8019), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268319] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12209), 1,
      anon_sym_type,
    STATE(8020), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268333] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12211), 1,
      anon_sym_DASH_GT,
    STATE(8021), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268347] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12213), 1,
      sym__lowercase_identifier,
    STATE(8022), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268361] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12215), 1,
      anon_sym_RPAREN,
    STATE(8023), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268375] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12217), 1,
      anon_sym_end,
    STATE(8024), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268389] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8495), 1,
      anon_sym_RPAREN,
    STATE(8025), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268403] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4381), 1,
      anon_sym_RPAREN,
    STATE(8026), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268417] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12219), 1,
      sym__lowercase_identifier,
    STATE(8027), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268431] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12221), 1,
      sym__lowercase_identifier,
    STATE(8028), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268445] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12223), 1,
      anon_sym_DASH_GT,
    STATE(8029), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268459] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12225), 1,
      anon_sym_RPAREN,
    STATE(8030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268473] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12227), 1,
      anon_sym_in,
    STATE(8031), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268487] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8357), 1,
      anon_sym_RPAREN,
    STATE(8032), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268501] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12229), 1,
      anon_sym_RPAREN,
    STATE(8033), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268515] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12231), 1,
      sym__lowercase_identifier,
    STATE(8034), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268529] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12233), 1,
      anon_sym_EQ,
    STATE(8035), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268543] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12235), 1,
      anon_sym_RBRACK,
    STATE(8036), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268557] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12237), 1,
      anon_sym_RPAREN,
    STATE(8037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268571] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12239), 1,
      anon_sym_DOT_DOT,
    STATE(8038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268585] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12241), 1,
      anon_sym_DOT_DOT,
    STATE(8039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268599] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12243), 1,
      anon_sym_DOT_DOT,
    STATE(8040), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268613] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12245), 1,
      anon_sym_RPAREN,
    STATE(8041), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268627] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12247), 1,
      anon_sym_RPAREN,
    STATE(8042), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268641] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12249), 1,
      anon_sym_RBRACE,
    STATE(8043), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268655] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12245), 1,
      anon_sym_RBRACK,
    STATE(8044), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268669] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12251), 1,
      anon_sym_in,
    STATE(8045), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268683] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12253), 1,
      anon_sym_RBRACK,
    STATE(8046), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268697] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12255), 1,
      anon_sym_in,
    STATE(8047), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268711] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12257), 1,
      sym__uppercase_identifier,
    STATE(8048), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268725] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12259), 1,
      anon_sym_RPAREN,
    STATE(8049), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268739] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12261), 1,
      anon_sym_RPAREN,
    STATE(8050), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268753] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12263), 1,
      anon_sym_RBRACK,
    STATE(8051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268767] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12245), 1,
      anon_sym_RBRACE,
    STATE(8052), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268781] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12265), 1,
      anon_sym_DASH_GT,
    STATE(8053), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268795] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12267), 1,
      sym__lowercase_identifier,
    STATE(8054), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268809] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12269), 1,
      anon_sym_RPAREN,
    STATE(8055), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268823] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12271), 1,
      anon_sym_DOT_DOT,
    STATE(8056), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268837] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12273), 1,
      anon_sym_DOT_DOT,
    STATE(8057), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268851] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12275), 1,
      anon_sym_DOT_DOT,
    STATE(8058), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268865] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12277), 1,
      anon_sym_RBRACE,
    STATE(8059), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268879] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12279), 1,
      aux_sym_number_token1,
    STATE(8060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268893] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1539), 1,
      anon_sym_RPAREN,
    STATE(8061), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268907] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12281), 1,
      anon_sym_PIPE_RBRACK,
    STATE(8062), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268921] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12283), 1,
      anon_sym_end,
    STATE(8063), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268935] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12285), 1,
      anon_sym_PIPE_RBRACK,
    STATE(8064), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268949] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12287), 1,
      anon_sym_RBRACK,
    STATE(8065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268963] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9776), 1,
      anon_sym_GT_RBRACE,
    STATE(8066), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268977] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12289), 1,
      anon_sym_RBRACK,
    STATE(8067), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [268991] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12291), 1,
      anon_sym_RPAREN,
    STATE(8068), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269005] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12293), 1,
      sym__lowercase_identifier,
    STATE(8069), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269019] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12295), 1,
      anon_sym_RPAREN,
    STATE(8070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269033] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12297), 1,
      anon_sym_DOT_DOT,
    STATE(8071), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269047] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12299), 1,
      anon_sym_DOT_DOT,
    STATE(8072), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269061] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12301), 1,
      anon_sym_DOT_DOT,
    STATE(8073), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269075] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12303), 1,
      anon_sym_DASH_GT,
    STATE(8074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269089] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8555), 1,
      anon_sym_end,
    STATE(8075), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269103] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12305), 1,
      anon_sym_RPAREN,
    STATE(8076), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269117] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12307), 1,
      anon_sym_RPAREN,
    STATE(8077), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269131] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12309), 1,
      anon_sym_RPAREN,
    STATE(8078), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269145] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12311), 1,
      anon_sym_RPAREN,
    STATE(8079), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269159] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12313), 1,
      anon_sym_PIPE_RBRACK,
    STATE(8080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269173] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12315), 1,
      anon_sym_RBRACE,
    STATE(8081), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269187] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12317), 1,
      anon_sym_RPAREN,
    STATE(8082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269201] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12319), 1,
      anon_sym_SQUOTE2,
    STATE(8083), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269215] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8460), 1,
      anon_sym_RPAREN,
    STATE(8084), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269229] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12321), 1,
      anon_sym_DOT_DOT,
    STATE(8085), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269243] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12323), 1,
      anon_sym_DOT_DOT,
    STATE(8086), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269257] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12325), 1,
      anon_sym_DOT_DOT,
    STATE(8087), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269271] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12327), 1,
      anon_sym_RBRACE,
    STATE(8088), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269285] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12329), 1,
      anon_sym_RPAREN,
    STATE(8089), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269299] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12331), 1,
      anon_sym_RPAREN,
    STATE(8090), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269313] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12333), 1,
      anon_sym_DQUOTE,
    STATE(8091), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269327] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12335), 1,
      anon_sym_in,
    STATE(8092), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269341] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12337), 1,
      anon_sym_DASH_GT,
    STATE(8093), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269355] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12339), 1,
      sym__lowercase_identifier,
    STATE(8094), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269369] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12341), 1,
      anon_sym_in,
    STATE(8095), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269383] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12343), 1,
      anon_sym_RBRACE,
    STATE(8096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269397] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12345), 1,
      anon_sym_in,
    STATE(8097), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269411] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12347), 1,
      anon_sym_RPAREN,
    STATE(8098), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269425] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12349), 1,
      anon_sym_DOT_DOT,
    STATE(8099), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269439] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12351), 1,
      anon_sym_DOT_DOT,
    STATE(8100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269453] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12353), 1,
      anon_sym_DOT_DOT,
    STATE(8101), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269467] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1561), 1,
      anon_sym_RPAREN,
    STATE(8102), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269481] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12355), 1,
      anon_sym_RBRACE,
    STATE(8103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269495] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12357), 1,
      anon_sym_RBRACE,
    STATE(8104), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269509] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12359), 1,
      anon_sym_RPAREN,
    STATE(8105), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269523] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12361), 1,
      anon_sym_RBRACE,
    STATE(8106), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269537] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(3638), 1,
      anon_sym_COMMA,
    STATE(8107), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269551] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12363), 1,
      anon_sym_LT_DASH,
    STATE(8108), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269565] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12365), 1,
      anon_sym_DASH_GT,
    STATE(8109), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269579] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12367), 1,
      anon_sym_RBRACK,
    STATE(8110), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269593] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12369), 1,
      anon_sym_COLON2,
    STATE(8111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269607] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1547), 1,
      anon_sym_RPAREN,
    STATE(8112), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269621] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12371), 1,
      anon_sym_DOT_DOT,
    STATE(8113), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269635] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12373), 1,
      anon_sym_DOT_DOT,
    STATE(8114), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269649] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12375), 1,
      anon_sym_DOT_DOT,
    STATE(8115), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269663] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12377), 1,
      anon_sym_RBRACE,
    STATE(8116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269677] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12379), 1,
      anon_sym_RPAREN,
    STATE(8117), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269691] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12381), 1,
      anon_sym_RBRACE,
    STATE(8118), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269705] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12383), 1,
      anon_sym_RPAREN,
    STATE(8119), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269719] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12385), 1,
      anon_sym_DASH_GT,
    STATE(8120), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269733] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12387), 1,
      anon_sym_RPAREN,
    STATE(8121), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269747] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12389), 1,
      anon_sym_COMMA,
    STATE(8122), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269761] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12391), 1,
      anon_sym_RPAREN,
    STATE(8123), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269775] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12393), 1,
      anon_sym_RBRACK,
    STATE(8124), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269789] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(10701), 1,
      aux_sym_number_token1,
    STATE(8125), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269803] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12395), 1,
      anon_sym_EQ,
    STATE(8126), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269817] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12397), 1,
      anon_sym_DOT_DOT,
    STATE(8127), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269831] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12399), 1,
      anon_sym_DOT_DOT,
    STATE(8128), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269845] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12401), 1,
      anon_sym_DOT_DOT,
    STATE(8129), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269859] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12403), 1,
      anon_sym_in,
    STATE(8130), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269873] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12405), 1,
      anon_sym_RPAREN,
    STATE(8131), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269887] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12407), 1,
      anon_sym_RPAREN,
    STATE(8132), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269901] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12409), 1,
      anon_sym_RBRACE,
    STATE(8133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269915] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12411), 1,
      anon_sym_DASH_GT,
    STATE(8134), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269929] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1569), 1,
      anon_sym_RPAREN,
    STATE(8135), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269943] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12413), 1,
      sym__lowercase_identifier,
    STATE(8136), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269957] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12415), 1,
      anon_sym_RPAREN,
    STATE(8137), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269971] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12417), 1,
      anon_sym_SQUOTE2,
    STATE(8138), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269985] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12419), 1,
      anon_sym_in,
    STATE(8139), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [269999] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12421), 1,
      anon_sym_RPAREN,
    STATE(8140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270013] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12423), 1,
      anon_sym_DOT_DOT,
    STATE(8141), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270027] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12425), 1,
      anon_sym_DOT_DOT,
    STATE(8142), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270041] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12427), 1,
      anon_sym_DOT_DOT,
    STATE(8143), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270055] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12429), 1,
      anon_sym_RBRACK,
    STATE(8144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270069] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1829), 1,
      anon_sym_COMMA,
    STATE(8145), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270083] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8106), 1,
      anon_sym_COMMA,
    STATE(8146), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270097] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12431), 1,
      anon_sym_RBRACK,
    STATE(8147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270111] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12433), 1,
      anon_sym_RPAREN,
    STATE(8148), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270125] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12435), 1,
      anon_sym_SQUOTE2,
    STATE(8149), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270139] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12437), 1,
      anon_sym_RBRACE,
    STATE(8150), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270153] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8350), 1,
      anon_sym_RPAREN,
    STATE(8151), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270167] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12439), 1,
      anon_sym_SQUOTE2,
    STATE(8152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270181] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8499), 1,
      anon_sym_RPAREN,
    STATE(8153), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270195] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12441), 1,
      anon_sym_DQUOTE,
    STATE(8154), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270209] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12443), 1,
      anon_sym_DOT_DOT,
    STATE(8155), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270223] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12445), 1,
      anon_sym_DOT_DOT,
    STATE(8156), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270237] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12447), 1,
      anon_sym_DOT_DOT,
    STATE(8157), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270251] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12449), 1,
      anon_sym_RPAREN,
    STATE(8158), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270265] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12451), 1,
      anon_sym_end,
    STATE(8159), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270279] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12453), 1,
      anon_sym_DQUOTE,
    STATE(8160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270293] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12455), 1,
      anon_sym_RBRACK,
    STATE(8161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270307] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12457), 1,
      anon_sym_RBRACE,
    STATE(8162), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270321] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12459), 1,
      anon_sym_RBRACE,
    STATE(8163), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270335] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12461), 1,
      anon_sym_PIPE_RBRACK,
    STATE(8164), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270349] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12463), 1,
      anon_sym_RPAREN,
    STATE(8165), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270363] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12465), 1,
      anon_sym_in,
    STATE(8166), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270377] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12467), 1,
      anon_sym_RBRACK,
    STATE(8167), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270391] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12469), 1,
      anon_sym_in,
    STATE(8168), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270405] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12471), 1,
      anon_sym_DOT_DOT,
    STATE(8169), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270419] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12473), 1,
      anon_sym_DOT_DOT,
    STATE(8170), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270433] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12475), 1,
      anon_sym_DOT_DOT,
    STATE(8171), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270447] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12477), 1,
      anon_sym_RPAREN,
    STATE(8172), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270461] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12479), 1,
      anon_sym_RBRACK,
    STATE(8173), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270475] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12481), 1,
      anon_sym_RPAREN,
    STATE(8174), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270489] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12483), 1,
      anon_sym_RPAREN,
    STATE(8175), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270503] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12485), 1,
      anon_sym_DOT_DOT,
    STATE(8176), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270517] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12487), 1,
      anon_sym_DOT_DOT,
    STATE(8177), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270531] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12489), 1,
      anon_sym_DOT_DOT,
    STATE(8178), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270545] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12491), 1,
      anon_sym_DASH_GT,
    STATE(8179), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270559] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12493), 1,
      anon_sym_DASH_GT,
    STATE(8180), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270573] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12495), 1,
      anon_sym_DOT_DOT,
    STATE(8181), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270587] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12497), 1,
      anon_sym_DOT_DOT,
    STATE(8182), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270601] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12499), 1,
      anon_sym_DOT_DOT,
    STATE(8183), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270615] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12501), 1,
      anon_sym_end,
    STATE(8184), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270629] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12503), 1,
      anon_sym_RPAREN,
    STATE(8185), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270643] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12505), 1,
      anon_sym_DOT_DOT,
    STATE(8186), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270657] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12507), 1,
      anon_sym_DOT_DOT,
    STATE(8187), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270671] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12509), 1,
      anon_sym_DOT_DOT,
    STATE(8188), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270685] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12511), 1,
      anon_sym_RBRACK,
    STATE(8189), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270699] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12513), 1,
      anon_sym_RBRACK,
    STATE(8190), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270713] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12515), 1,
      anon_sym_DOT_DOT,
    STATE(8191), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270727] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12517), 1,
      anon_sym_DOT_DOT,
    STATE(8192), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270741] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12519), 1,
      anon_sym_DOT_DOT,
    STATE(8193), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270755] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(7426), 1,
      anon_sym_SQUOTE2,
    STATE(8194), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270769] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12521), 1,
      anon_sym_DASH_GT,
    STATE(8195), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270783] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12523), 1,
      anon_sym_DOT_DOT,
    STATE(8196), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270797] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12525), 1,
      anon_sym_DOT_DOT,
    STATE(8197), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270811] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12527), 1,
      anon_sym_DOT_DOT,
    STATE(8198), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270825] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12529), 1,
      anon_sym_RPAREN,
    STATE(8199), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270839] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12531), 1,
      anon_sym_RBRACE,
    STATE(8200), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270853] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12533), 1,
      anon_sym_DOT_DOT,
    STATE(8201), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270867] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12535), 1,
      anon_sym_DOT_DOT,
    STATE(8202), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270881] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12537), 1,
      anon_sym_DOT_DOT,
    STATE(8203), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270895] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12539), 1,
      anon_sym_RPAREN,
    STATE(8204), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270909] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12541), 1,
      anon_sym_in,
    STATE(8205), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270923] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12543), 1,
      anon_sym_DOT_DOT,
    STATE(8206), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270937] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12545), 1,
      anon_sym_DOT_DOT,
    STATE(8207), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270951] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12547), 1,
      anon_sym_DOT_DOT,
    STATE(8208), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270965] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12549), 1,
      anon_sym_RPAREN,
    STATE(8209), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270979] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12551), 1,
      anon_sym_RPAREN,
    STATE(8210), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [270993] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12551), 1,
      anon_sym_RBRACK,
    STATE(8211), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271007] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12551), 1,
      anon_sym_RBRACE,
    STATE(8212), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271021] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12553), 1,
      anon_sym_RPAREN,
    STATE(8213), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271035] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12555), 1,
      anon_sym_RPAREN,
    STATE(8214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271049] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12557), 1,
      sym_indexing_operator,
    STATE(8215), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271063] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12559), 1,
      anon_sym_in,
    STATE(8216), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271077] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12561), 1,
      anon_sym_RPAREN,
    STATE(8217), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271091] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12563), 1,
      anon_sym_RBRACK,
    STATE(8218), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271105] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12565), 1,
      anon_sym_RBRACE,
    STATE(8219), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271119] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8269), 1,
      anon_sym_COMMA,
    STATE(8220), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271133] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12567), 1,
      anon_sym_DASH_GT,
    STATE(8221), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271147] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4433), 1,
      anon_sym_RPAREN,
    STATE(8222), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271161] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12569), 1,
      anon_sym_RPAREN,
    STATE(8223), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271175] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12571), 1,
      anon_sym_RPAREN,
    STATE(8224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271189] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6539), 1,
      anon_sym_DASH_GT,
    STATE(8225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271203] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12573), 1,
      anon_sym_RPAREN,
    STATE(8226), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271217] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12575), 1,
      anon_sym_STAR,
    STATE(8227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271231] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12577), 1,
      anon_sym_DASH_GT,
    STATE(8228), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271245] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9947), 1,
      anon_sym_GT_RBRACE,
    STATE(8229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271259] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12579), 1,
      anon_sym_DOT,
    STATE(8230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271273] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12581), 1,
      anon_sym_RBRACE,
    STATE(8231), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271287] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12583), 1,
      anon_sym_RPAREN,
    STATE(8232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271301] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12585), 1,
      anon_sym_RBRACK,
    STATE(8233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271315] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12587), 1,
      anon_sym_RBRACE,
    STATE(8234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271329] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12589), 1,
      anon_sym_end,
    STATE(8235), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271343] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12591), 1,
      anon_sym_RBRACK,
    STATE(8236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271357] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12593), 1,
      anon_sym_RBRACE,
    STATE(8237), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271371] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12595), 1,
      anon_sym_EQ,
    STATE(8238), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271385] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12597), 1,
      anon_sym_RBRACE,
    STATE(8239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271399] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12599), 1,
      anon_sym_of,
    STATE(8240), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271413] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8569), 1,
      anon_sym_end,
    STATE(8241), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271427] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12601), 1,
      anon_sym_RBRACE,
    STATE(8242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271441] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12603), 1,
      anon_sym_EQ,
    STATE(8243), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271455] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12605), 1,
      anon_sym_RPAREN,
    STATE(8244), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271469] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12607), 1,
      anon_sym_RPAREN,
    STATE(8245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271483] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12609), 1,
      anon_sym_RBRACK,
    STATE(8246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271497] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12611), 1,
      ts_builtin_sym_end,
    STATE(8247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271511] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12613), 1,
      anon_sym_DASH_GT,
    STATE(8248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271525] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12615), 1,
      sym__lowercase_identifier,
    STATE(8249), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271539] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12617), 1,
      anon_sym_SQUOTE2,
    STATE(8250), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271553] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12619), 1,
      anon_sym_DQUOTE,
    STATE(8251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271567] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12621), 1,
      anon_sym_RBRACE,
    STATE(8252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271581] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12623), 1,
      anon_sym_RPAREN,
    STATE(8253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271595] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12625), 1,
      sym__lowercase_identifier,
    STATE(8254), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271609] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12627), 1,
      anon_sym_RPAREN,
    STATE(8255), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271623] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12629), 1,
      anon_sym_RBRACE,
    STATE(8256), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271637] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12631), 1,
      anon_sym_end,
    STATE(8257), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271651] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12633), 1,
      anon_sym_RBRACK,
    STATE(8258), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271665] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12635), 1,
      anon_sym_RBRACK,
    STATE(8259), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271679] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12637), 1,
      anon_sym_RBRACK,
    STATE(8260), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271693] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12639), 1,
      anon_sym_RBRACK,
    STATE(8261), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271707] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12641), 1,
      anon_sym_DASH_GT,
    STATE(8262), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271721] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12643), 1,
      anon_sym_RPAREN,
    STATE(8263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271735] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12645), 1,
      anon_sym_PIPE_RBRACK,
    STATE(8264), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271749] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12647), 1,
      anon_sym_RBRACE,
    STATE(8265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271763] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12649), 1,
      sym_indexing_operator,
    STATE(8266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271777] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12651), 1,
      anon_sym_RPAREN,
    STATE(8267), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271791] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12653), 1,
      anon_sym_end,
    STATE(8268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271805] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12655), 1,
      anon_sym_RBRACK,
    STATE(8269), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271819] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12657), 1,
      anon_sym_RBRACE,
    STATE(8270), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271833] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12659), 1,
      sym_indexing_operator,
    STATE(8271), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271847] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12661), 1,
      anon_sym_PIPE,
    STATE(8272), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271861] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12663), 1,
      anon_sym_in,
    STATE(8273), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271875] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12665), 1,
      sym_indexing_operator,
    STATE(8274), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271889] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12667), 1,
      sym_indexing_operator,
    STATE(8275), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271903] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12669), 1,
      sym_indexing_operator,
    STATE(8276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271917] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12671), 1,
      sym_indexing_operator,
    STATE(8277), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271931] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12673), 1,
      sym_indexing_operator,
    STATE(8278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271945] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12675), 1,
      sym_indexing_operator,
    STATE(8279), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271959] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12677), 1,
      sym_indexing_operator,
    STATE(8280), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271973] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12679), 1,
      sym_indexing_operator,
    STATE(8281), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [271987] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12681), 1,
      sym_indexing_operator,
    STATE(8282), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272001] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12683), 1,
      sym_indexing_operator,
    STATE(8283), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272015] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12685), 1,
      sym_indexing_operator,
    STATE(8284), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272029] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12687), 1,
      sym_indexing_operator,
    STATE(8285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272043] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12689), 1,
      sym_indexing_operator,
    STATE(8286), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272057] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12691), 1,
      sym_indexing_operator,
    STATE(8287), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272071] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12693), 1,
      sym_indexing_operator,
    STATE(8288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272085] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12695), 1,
      sym_indexing_operator,
    STATE(8289), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272099] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12697), 1,
      anon_sym_RPAREN,
    STATE(8290), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272113] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12699), 1,
      anon_sym_RPAREN,
    STATE(8291), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272127] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12701), 1,
      anon_sym_RBRACK,
    STATE(8292), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272141] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12703), 1,
      anon_sym_in,
    STATE(8293), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272155] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12705), 1,
      anon_sym_with,
    STATE(8294), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272169] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12707), 1,
      anon_sym_PLUS_EQ,
    STATE(8295), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272183] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12709), 1,
      anon_sym_RBRACE,
    STATE(8296), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272197] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12711), 1,
      anon_sym_type,
    STATE(8297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272211] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4544), 1,
      anon_sym_COMMA,
    STATE(8298), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272225] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12713), 1,
      anon_sym_RPAREN,
    STATE(8299), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272239] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12715), 1,
      anon_sym_RPAREN,
    STATE(8300), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272253] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12717), 1,
      anon_sym_RBRACK,
    STATE(8301), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272267] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12719), 1,
      anon_sym_RPAREN,
    STATE(8302), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272281] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12721), 1,
      anon_sym_RPAREN,
    STATE(8303), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272295] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(6947), 1,
      anon_sym_DASH_GT,
    STATE(8304), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272309] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12723), 1,
      anon_sym_RBRACE,
    STATE(8305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272323] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12725), 1,
      anon_sym_RPAREN,
    STATE(8306), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272337] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12727), 1,
      anon_sym_EQ,
    STATE(8307), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272351] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4122), 1,
      anon_sym_RBRACK,
    STATE(8308), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272365] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12729), 1,
      anon_sym_RBRACE,
    STATE(8309), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272379] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8555), 1,
      anon_sym_RPAREN,
    STATE(8310), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272393] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12731), 1,
      anon_sym_RPAREN,
    STATE(8311), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272407] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12733), 1,
      anon_sym_RPAREN,
    STATE(8312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272421] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12735), 1,
      anon_sym_PLUS_EQ,
    STATE(8313), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272435] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4184), 1,
      anon_sym_COMMA,
    STATE(8314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272449] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12737), 1,
      anon_sym_in,
    STATE(8315), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272463] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12739), 1,
      sym__lowercase_identifier,
    STATE(8316), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272477] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12741), 1,
      anon_sym_done,
    STATE(8317), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272491] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12743), 1,
      anon_sym_DOT,
    STATE(8318), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272505] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12745), 1,
      anon_sym_EQ,
    STATE(8319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272519] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12747), 1,
      anon_sym_RPAREN,
    STATE(8320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272533] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8042), 1,
      anon_sym_RBRACK,
    STATE(8321), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272547] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(1533), 1,
      anon_sym_RPAREN,
    STATE(8322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272561] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12749), 1,
      anon_sym_RPAREN,
    STATE(8323), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272575] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12751), 1,
      anon_sym_LT_DASH,
    STATE(8324), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272589] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12753), 1,
      anon_sym_RBRACK,
    STATE(8325), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272603] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12755), 1,
      anon_sym_RPAREN,
    STATE(8326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272617] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(5096), 1,
      anon_sym_DASH_GT,
    STATE(8327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272631] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12757), 1,
      anon_sym_RBRACK,
    STATE(8328), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272645] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12759), 1,
      anon_sym_RPAREN,
    STATE(8329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272659] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12761), 1,
      anon_sym_SQUOTE2,
    STATE(8330), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272673] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12763), 1,
      anon_sym_RBRACK,
    STATE(8331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272687] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12765), 1,
      anon_sym_DOT,
    STATE(8332), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272701] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12767), 1,
      anon_sym_RBRACE,
    STATE(8333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272715] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12769), 1,
      anon_sym_PIPE_RBRACK,
    STATE(8334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272729] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12771), 1,
      anon_sym_DQUOTE,
    STATE(8335), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272743] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12773), 1,
      anon_sym_RBRACE,
    STATE(8336), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272757] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12775), 1,
      anon_sym_RBRACK,
    STATE(8337), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272771] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12777), 1,
      anon_sym_RPAREN,
    STATE(8338), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272785] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(8005), 1,
      anon_sym_GT,
    STATE(8339), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272799] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12779), 1,
      anon_sym_DASH_GT,
    STATE(8340), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272813] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12781), 1,
      anon_sym_DOT_DOT,
    STATE(8341), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272827] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12783), 1,
      anon_sym_DOT_DOT,
    STATE(8342), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272841] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12785), 1,
      anon_sym_PIPE_RBRACK,
    STATE(8343), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272855] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12787), 1,
      anon_sym_COMMA,
    STATE(8344), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272869] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12789), 1,
      anon_sym_DOT_DOT,
    STATE(8345), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272883] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12791), 1,
      anon_sym_RBRACE,
    STATE(8346), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272897] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(9987), 1,
      anon_sym_GT_RBRACE,
    STATE(8347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272911] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12793), 1,
      anon_sym_GT,
    STATE(8348), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272925] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12795), 1,
      anon_sym_DASH_GT,
    STATE(8349), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272939] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12797), 1,
      anon_sym_RPAREN,
    STATE(8350), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272953] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12799), 1,
      anon_sym_RPAREN,
    STATE(8351), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272967] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(12801), 1,
      anon_sym_RBRACK,
    STATE(8352), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272981] = 4,
    ACTIONS(477), 1,
      aux_sym_attribute_token1,
    ACTIONS(4361), 1,
      anon_sym_RPAREN,
    STATE(8353), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [272995] = 1,
    ACTIONS(12803), 1,
      ts_builtin_sym_end,
  [272999] = 1,
    ACTIONS(12805), 1,
      ts_builtin_sym_end,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(1773)] = 0,
  [SMALL_STATE(1774)] = 77,
  [SMALL_STATE(1775)] = 153,
  [SMALL_STATE(1776)] = 285,
  [SMALL_STATE(1777)] = 423,
  [SMALL_STATE(1778)] = 561,
  [SMALL_STATE(1779)] = 699,
  [SMALL_STATE(1780)] = 837,
  [SMALL_STATE(1781)] = 969,
  [SMALL_STATE(1782)] = 1107,
  [SMALL_STATE(1783)] = 1239,
  [SMALL_STATE(1784)] = 1377,
  [SMALL_STATE(1785)] = 1509,
  [SMALL_STATE(1786)] = 1647,
  [SMALL_STATE(1787)] = 1779,
  [SMALL_STATE(1788)] = 1911,
  [SMALL_STATE(1789)] = 2043,
  [SMALL_STATE(1790)] = 2181,
  [SMALL_STATE(1791)] = 2313,
  [SMALL_STATE(1792)] = 2445,
  [SMALL_STATE(1793)] = 2577,
  [SMALL_STATE(1794)] = 2709,
  [SMALL_STATE(1795)] = 2841,
  [SMALL_STATE(1796)] = 2973,
  [SMALL_STATE(1797)] = 3111,
  [SMALL_STATE(1798)] = 3249,
  [SMALL_STATE(1799)] = 3387,
  [SMALL_STATE(1800)] = 3525,
  [SMALL_STATE(1801)] = 3601,
  [SMALL_STATE(1802)] = 3739,
  [SMALL_STATE(1803)] = 3877,
  [SMALL_STATE(1804)] = 4015,
  [SMALL_STATE(1805)] = 4153,
  [SMALL_STATE(1806)] = 4291,
  [SMALL_STATE(1807)] = 4423,
  [SMALL_STATE(1808)] = 4558,
  [SMALL_STATE(1809)] = 4693,
  [SMALL_STATE(1810)] = 4828,
  [SMALL_STATE(1811)] = 4963,
  [SMALL_STATE(1812)] = 5098,
  [SMALL_STATE(1813)] = 5233,
  [SMALL_STATE(1814)] = 5368,
  [SMALL_STATE(1815)] = 5503,
  [SMALL_STATE(1816)] = 5638,
  [SMALL_STATE(1817)] = 5773,
  [SMALL_STATE(1818)] = 5908,
  [SMALL_STATE(1819)] = 6043,
  [SMALL_STATE(1820)] = 6178,
  [SMALL_STATE(1821)] = 6313,
  [SMALL_STATE(1822)] = 6448,
  [SMALL_STATE(1823)] = 6583,
  [SMALL_STATE(1824)] = 6718,
  [SMALL_STATE(1825)] = 6853,
  [SMALL_STATE(1826)] = 6988,
  [SMALL_STATE(1827)] = 7123,
  [SMALL_STATE(1828)] = 7249,
  [SMALL_STATE(1829)] = 7375,
  [SMALL_STATE(1830)] = 7509,
  [SMALL_STATE(1831)] = 7643,
  [SMALL_STATE(1832)] = 7777,
  [SMALL_STATE(1833)] = 7911,
  [SMALL_STATE(1834)] = 8037,
  [SMALL_STATE(1835)] = 8163,
  [SMALL_STATE(1836)] = 8289,
  [SMALL_STATE(1837)] = 8415,
  [SMALL_STATE(1838)] = 8549,
  [SMALL_STATE(1839)] = 8675,
  [SMALL_STATE(1840)] = 8801,
  [SMALL_STATE(1841)] = 8927,
  [SMALL_STATE(1842)] = 9061,
  [SMALL_STATE(1843)] = 9195,
  [SMALL_STATE(1844)] = 9321,
  [SMALL_STATE(1845)] = 9447,
  [SMALL_STATE(1846)] = 9581,
  [SMALL_STATE(1847)] = 9707,
  [SMALL_STATE(1848)] = 9833,
  [SMALL_STATE(1849)] = 9959,
  [SMALL_STATE(1850)] = 10085,
  [SMALL_STATE(1851)] = 10219,
  [SMALL_STATE(1852)] = 10345,
  [SMALL_STATE(1853)] = 10479,
  [SMALL_STATE(1854)] = 10605,
  [SMALL_STATE(1855)] = 10731,
  [SMALL_STATE(1856)] = 10857,
  [SMALL_STATE(1857)] = 10983,
  [SMALL_STATE(1858)] = 11109,
  [SMALL_STATE(1859)] = 11235,
  [SMALL_STATE(1860)] = 11361,
  [SMALL_STATE(1861)] = 11484,
  [SMALL_STATE(1862)] = 11607,
  [SMALL_STATE(1863)] = 11730,
  [SMALL_STATE(1864)] = 11853,
  [SMALL_STATE(1865)] = 11924,
  [SMALL_STATE(1866)] = 12047,
  [SMALL_STATE(1867)] = 12170,
  [SMALL_STATE(1868)] = 12293,
  [SMALL_STATE(1869)] = 12416,
  [SMALL_STATE(1870)] = 12539,
  [SMALL_STATE(1871)] = 12662,
  [SMALL_STATE(1872)] = 12785,
  [SMALL_STATE(1873)] = 12908,
  [SMALL_STATE(1874)] = 12979,
  [SMALL_STATE(1875)] = 13102,
  [SMALL_STATE(1876)] = 13173,
  [SMALL_STATE(1877)] = 13296,
  [SMALL_STATE(1878)] = 13416,
  [SMALL_STATE(1879)] = 13536,
  [SMALL_STATE(1880)] = 13656,
  [SMALL_STATE(1881)] = 13776,
  [SMALL_STATE(1882)] = 13896,
  [SMALL_STATE(1883)] = 14016,
  [SMALL_STATE(1884)] = 14136,
  [SMALL_STATE(1885)] = 14256,
  [SMALL_STATE(1886)] = 14376,
  [SMALL_STATE(1887)] = 14496,
  [SMALL_STATE(1888)] = 14616,
  [SMALL_STATE(1889)] = 14736,
  [SMALL_STATE(1890)] = 14856,
  [SMALL_STATE(1891)] = 14976,
  [SMALL_STATE(1892)] = 15096,
  [SMALL_STATE(1893)] = 15216,
  [SMALL_STATE(1894)] = 15336,
  [SMALL_STATE(1895)] = 15456,
  [SMALL_STATE(1896)] = 15576,
  [SMALL_STATE(1897)] = 15696,
  [SMALL_STATE(1898)] = 15816,
  [SMALL_STATE(1899)] = 15936,
  [SMALL_STATE(1900)] = 16056,
  [SMALL_STATE(1901)] = 16176,
  [SMALL_STATE(1902)] = 16296,
  [SMALL_STATE(1903)] = 16416,
  [SMALL_STATE(1904)] = 16536,
  [SMALL_STATE(1905)] = 16656,
  [SMALL_STATE(1906)] = 16776,
  [SMALL_STATE(1907)] = 16896,
  [SMALL_STATE(1908)] = 17016,
  [SMALL_STATE(1909)] = 17136,
  [SMALL_STATE(1910)] = 17256,
  [SMALL_STATE(1911)] = 17376,
  [SMALL_STATE(1912)] = 17496,
  [SMALL_STATE(1913)] = 17616,
  [SMALL_STATE(1914)] = 17736,
  [SMALL_STATE(1915)] = 17856,
  [SMALL_STATE(1916)] = 17976,
  [SMALL_STATE(1917)] = 18096,
  [SMALL_STATE(1918)] = 18216,
  [SMALL_STATE(1919)] = 18336,
  [SMALL_STATE(1920)] = 18456,
  [SMALL_STATE(1921)] = 18576,
  [SMALL_STATE(1922)] = 18696,
  [SMALL_STATE(1923)] = 18816,
  [SMALL_STATE(1924)] = 18936,
  [SMALL_STATE(1925)] = 19056,
  [SMALL_STATE(1926)] = 19176,
  [SMALL_STATE(1927)] = 19296,
  [SMALL_STATE(1928)] = 19416,
  [SMALL_STATE(1929)] = 19536,
  [SMALL_STATE(1930)] = 19609,
  [SMALL_STATE(1931)] = 19736,
  [SMALL_STATE(1932)] = 19809,
  [SMALL_STATE(1933)] = 19880,
  [SMALL_STATE(1934)] = 20007,
  [SMALL_STATE(1935)] = 20134,
  [SMALL_STATE(1936)] = 20261,
  [SMALL_STATE(1937)] = 20388,
  [SMALL_STATE(1938)] = 20457,
  [SMALL_STATE(1939)] = 20526,
  [SMALL_STATE(1940)] = 20597,
  [SMALL_STATE(1941)] = 20666,
  [SMALL_STATE(1942)] = 20735,
  [SMALL_STATE(1943)] = 20804,
  [SMALL_STATE(1944)] = 20873,
  [SMALL_STATE(1945)] = 20944,
  [SMALL_STATE(1946)] = 21013,
  [SMALL_STATE(1947)] = 21082,
  [SMALL_STATE(1948)] = 21151,
  [SMALL_STATE(1949)] = 21219,
  [SMALL_STATE(1950)] = 21287,
  [SMALL_STATE(1951)] = 21355,
  [SMALL_STATE(1952)] = 21423,
  [SMALL_STATE(1953)] = 21491,
  [SMALL_STATE(1954)] = 21561,
  [SMALL_STATE(1955)] = 21629,
  [SMALL_STATE(1956)] = 21697,
  [SMALL_STATE(1957)] = 21765,
  [SMALL_STATE(1958)] = 21833,
  [SMALL_STATE(1959)] = 21901,
  [SMALL_STATE(1960)] = 21969,
  [SMALL_STATE(1961)] = 22043,
  [SMALL_STATE(1962)] = 22111,
  [SMALL_STATE(1963)] = 22179,
  [SMALL_STATE(1964)] = 22247,
  [SMALL_STATE(1965)] = 22315,
  [SMALL_STATE(1966)] = 22383,
  [SMALL_STATE(1967)] = 22451,
  [SMALL_STATE(1968)] = 22521,
  [SMALL_STATE(1969)] = 22589,
  [SMALL_STATE(1970)] = 22657,
  [SMALL_STATE(1971)] = 22725,
  [SMALL_STATE(1972)] = 22793,
  [SMALL_STATE(1973)] = 22861,
  [SMALL_STATE(1974)] = 22929,
  [SMALL_STATE(1975)] = 22997,
  [SMALL_STATE(1976)] = 23065,
  [SMALL_STATE(1977)] = 23133,
  [SMALL_STATE(1978)] = 23201,
  [SMALL_STATE(1979)] = 23269,
  [SMALL_STATE(1980)] = 23337,
  [SMALL_STATE(1981)] = 23405,
  [SMALL_STATE(1982)] = 23473,
  [SMALL_STATE(1983)] = 23541,
  [SMALL_STATE(1984)] = 23611,
  [SMALL_STATE(1985)] = 23679,
  [SMALL_STATE(1986)] = 23747,
  [SMALL_STATE(1987)] = 23815,
  [SMALL_STATE(1988)] = 23883,
  [SMALL_STATE(1989)] = 23951,
  [SMALL_STATE(1990)] = 24019,
  [SMALL_STATE(1991)] = 24087,
  [SMALL_STATE(1992)] = 24155,
  [SMALL_STATE(1993)] = 24223,
  [SMALL_STATE(1994)] = 24291,
  [SMALL_STATE(1995)] = 24359,
  [SMALL_STATE(1996)] = 24427,
  [SMALL_STATE(1997)] = 24495,
  [SMALL_STATE(1998)] = 24563,
  [SMALL_STATE(1999)] = 24631,
  [SMALL_STATE(2000)] = 24699,
  [SMALL_STATE(2001)] = 24767,
  [SMALL_STATE(2002)] = 24835,
  [SMALL_STATE(2003)] = 24903,
  [SMALL_STATE(2004)] = 24971,
  [SMALL_STATE(2005)] = 25039,
  [SMALL_STATE(2006)] = 25107,
  [SMALL_STATE(2007)] = 25175,
  [SMALL_STATE(2008)] = 25243,
  [SMALL_STATE(2009)] = 25311,
  [SMALL_STATE(2010)] = 25379,
  [SMALL_STATE(2011)] = 25447,
  [SMALL_STATE(2012)] = 25515,
  [SMALL_STATE(2013)] = 25583,
  [SMALL_STATE(2014)] = 25651,
  [SMALL_STATE(2015)] = 25719,
  [SMALL_STATE(2016)] = 25787,
  [SMALL_STATE(2017)] = 25857,
  [SMALL_STATE(2018)] = 25927,
  [SMALL_STATE(2019)] = 25995,
  [SMALL_STATE(2020)] = 26063,
  [SMALL_STATE(2021)] = 26131,
  [SMALL_STATE(2022)] = 26205,
  [SMALL_STATE(2023)] = 26273,
  [SMALL_STATE(2024)] = 26394,
  [SMALL_STATE(2025)] = 26515,
  [SMALL_STATE(2026)] = 26636,
  [SMALL_STATE(2027)] = 26757,
  [SMALL_STATE(2028)] = 26878,
  [SMALL_STATE(2029)] = 26999,
  [SMALL_STATE(2030)] = 27120,
  [SMALL_STATE(2031)] = 27241,
  [SMALL_STATE(2032)] = 27362,
  [SMALL_STATE(2033)] = 27483,
  [SMALL_STATE(2034)] = 27604,
  [SMALL_STATE(2035)] = 27725,
  [SMALL_STATE(2036)] = 27846,
  [SMALL_STATE(2037)] = 27915,
  [SMALL_STATE(2038)] = 28036,
  [SMALL_STATE(2039)] = 28157,
  [SMALL_STATE(2040)] = 28278,
  [SMALL_STATE(2041)] = 28399,
  [SMALL_STATE(2042)] = 28520,
  [SMALL_STATE(2043)] = 28641,
  [SMALL_STATE(2044)] = 28710,
  [SMALL_STATE(2045)] = 28831,
  [SMALL_STATE(2046)] = 28952,
  [SMALL_STATE(2047)] = 29073,
  [SMALL_STATE(2048)] = 29194,
  [SMALL_STATE(2049)] = 29315,
  [SMALL_STATE(2050)] = 29436,
  [SMALL_STATE(2051)] = 29557,
  [SMALL_STATE(2052)] = 29678,
  [SMALL_STATE(2053)] = 29799,
  [SMALL_STATE(2054)] = 29920,
  [SMALL_STATE(2055)] = 30041,
  [SMALL_STATE(2056)] = 30162,
  [SMALL_STATE(2057)] = 30283,
  [SMALL_STATE(2058)] = 30401,
  [SMALL_STATE(2059)] = 30519,
  [SMALL_STATE(2060)] = 30637,
  [SMALL_STATE(2061)] = 30705,
  [SMALL_STATE(2062)] = 30823,
  [SMALL_STATE(2063)] = 30941,
  [SMALL_STATE(2064)] = 31059,
  [SMALL_STATE(2065)] = 31177,
  [SMALL_STATE(2066)] = 31245,
  [SMALL_STATE(2067)] = 31363,
  [SMALL_STATE(2068)] = 31481,
  [SMALL_STATE(2069)] = 31599,
  [SMALL_STATE(2070)] = 31717,
  [SMALL_STATE(2071)] = 31835,
  [SMALL_STATE(2072)] = 31953,
  [SMALL_STATE(2073)] = 32021,
  [SMALL_STATE(2074)] = 32139,
  [SMALL_STATE(2075)] = 32207,
  [SMALL_STATE(2076)] = 32325,
  [SMALL_STATE(2077)] = 32391,
  [SMALL_STATE(2078)] = 32509,
  [SMALL_STATE(2079)] = 32574,
  [SMALL_STATE(2080)] = 32669,
  [SMALL_STATE(2081)] = 32784,
  [SMALL_STATE(2082)] = 32849,
  [SMALL_STATE(2083)] = 32940,
  [SMALL_STATE(2084)] = 33055,
  [SMALL_STATE(2085)] = 33156,
  [SMALL_STATE(2086)] = 33257,
  [SMALL_STATE(2087)] = 33372,
  [SMALL_STATE(2088)] = 33471,
  [SMALL_STATE(2089)] = 33536,
  [SMALL_STATE(2090)] = 33651,
  [SMALL_STATE(2091)] = 33726,
  [SMALL_STATE(2092)] = 33825,
  [SMALL_STATE(2093)] = 33902,
  [SMALL_STATE(2094)] = 33979,
  [SMALL_STATE(2095)] = 34060,
  [SMALL_STATE(2096)] = 34175,
  [SMALL_STATE(2097)] = 34276,
  [SMALL_STATE(2098)] = 34375,
  [SMALL_STATE(2099)] = 34474,
  [SMALL_STATE(2100)] = 34571,
  [SMALL_STATE(2101)] = 34636,
  [SMALL_STATE(2102)] = 34721,
  [SMALL_STATE(2103)] = 34788,
  [SMALL_STATE(2104)] = 34875,
  [SMALL_STATE(2105)] = 34962,
  [SMALL_STATE(2106)] = 35027,
  [SMALL_STATE(2107)] = 35092,
  [SMALL_STATE(2108)] = 35156,
  [SMALL_STATE(2109)] = 35268,
  [SMALL_STATE(2110)] = 35380,
  [SMALL_STATE(2111)] = 35494,
  [SMALL_STATE(2112)] = 35592,
  [SMALL_STATE(2113)] = 35690,
  [SMALL_STATE(2114)] = 35806,
  [SMALL_STATE(2115)] = 35872,
  [SMALL_STATE(2116)] = 35984,
  [SMALL_STATE(2117)] = 36082,
  [SMALL_STATE(2118)] = 36194,
  [SMALL_STATE(2119)] = 36306,
  [SMALL_STATE(2120)] = 36418,
  [SMALL_STATE(2121)] = 36516,
  [SMALL_STATE(2122)] = 36614,
  [SMALL_STATE(2123)] = 36678,
  [SMALL_STATE(2124)] = 36742,
  [SMALL_STATE(2125)] = 36840,
  [SMALL_STATE(2126)] = 36938,
  [SMALL_STATE(2127)] = 37004,
  [SMALL_STATE(2128)] = 37068,
  [SMALL_STATE(2129)] = 37168,
  [SMALL_STATE(2130)] = 37244,
  [SMALL_STATE(2131)] = 37320,
  [SMALL_STATE(2132)] = 37400,
  [SMALL_STATE(2133)] = 37484,
  [SMALL_STATE(2134)] = 37550,
  [SMALL_STATE(2135)] = 37614,
  [SMALL_STATE(2136)] = 37726,
  [SMALL_STATE(2137)] = 37820,
  [SMALL_STATE(2138)] = 37884,
  [SMALL_STATE(2139)] = 38000,
  [SMALL_STATE(2140)] = 38100,
  [SMALL_STATE(2141)] = 38200,
  [SMALL_STATE(2142)] = 38274,
  [SMALL_STATE(2143)] = 38386,
  [SMALL_STATE(2144)] = 38450,
  [SMALL_STATE(2145)] = 38562,
  [SMALL_STATE(2146)] = 38676,
  [SMALL_STATE(2147)] = 38752,
  [SMALL_STATE(2148)] = 38828,
  [SMALL_STATE(2149)] = 38942,
  [SMALL_STATE(2150)] = 39058,
  [SMALL_STATE(2151)] = 39170,
  [SMALL_STATE(2152)] = 39282,
  [SMALL_STATE(2153)] = 39394,
  [SMALL_STATE(2154)] = 39458,
  [SMALL_STATE(2155)] = 39522,
  [SMALL_STATE(2156)] = 39638,
  [SMALL_STATE(2157)] = 39718,
  [SMALL_STATE(2158)] = 39830,
  [SMALL_STATE(2159)] = 39930,
  [SMALL_STATE(2160)] = 39994,
  [SMALL_STATE(2161)] = 40060,
  [SMALL_STATE(2162)] = 40124,
  [SMALL_STATE(2163)] = 40236,
  [SMALL_STATE(2164)] = 40322,
  [SMALL_STATE(2165)] = 40386,
  [SMALL_STATE(2166)] = 40472,
  [SMALL_STATE(2167)] = 40584,
  [SMALL_STATE(2168)] = 40696,
  [SMALL_STATE(2169)] = 40792,
  [SMALL_STATE(2170)] = 40892,
  [SMALL_STATE(2171)] = 40976,
  [SMALL_STATE(2172)] = 41088,
  [SMALL_STATE(2173)] = 41174,
  [SMALL_STATE(2174)] = 41238,
  [SMALL_STATE(2175)] = 41302,
  [SMALL_STATE(2176)] = 41414,
  [SMALL_STATE(2177)] = 41526,
  [SMALL_STATE(2178)] = 41612,
  [SMALL_STATE(2179)] = 41676,
  [SMALL_STATE(2180)] = 41788,
  [SMALL_STATE(2181)] = 41878,
  [SMALL_STATE(2182)] = 41976,
  [SMALL_STATE(2183)] = 42066,
  [SMALL_STATE(2184)] = 42182,
  [SMALL_STATE(2185)] = 42276,
  [SMALL_STATE(2186)] = 42388,
  [SMALL_STATE(2187)] = 42488,
  [SMALL_STATE(2188)] = 42562,
  [SMALL_STATE(2189)] = 42658,
  [SMALL_STATE(2190)] = 42721,
  [SMALL_STATE(2191)] = 42784,
  [SMALL_STATE(2192)] = 42849,
  [SMALL_STATE(2193)] = 42912,
  [SMALL_STATE(2194)] = 42975,
  [SMALL_STATE(2195)] = 43038,
  [SMALL_STATE(2196)] = 43101,
  [SMALL_STATE(2197)] = 43164,
  [SMALL_STATE(2198)] = 43227,
  [SMALL_STATE(2199)] = 43290,
  [SMALL_STATE(2200)] = 43389,
  [SMALL_STATE(2201)] = 43452,
  [SMALL_STATE(2202)] = 43515,
  [SMALL_STATE(2203)] = 43584,
  [SMALL_STATE(2204)] = 43653,
  [SMALL_STATE(2205)] = 43716,
  [SMALL_STATE(2206)] = 43779,
  [SMALL_STATE(2207)] = 43842,
  [SMALL_STATE(2208)] = 43905,
  [SMALL_STATE(2209)] = 43968,
  [SMALL_STATE(2210)] = 44067,
  [SMALL_STATE(2211)] = 44130,
  [SMALL_STATE(2212)] = 44203,
  [SMALL_STATE(2213)] = 44266,
  [SMALL_STATE(2214)] = 44329,
  [SMALL_STATE(2215)] = 44392,
  [SMALL_STATE(2216)] = 44455,
  [SMALL_STATE(2217)] = 44518,
  [SMALL_STATE(2218)] = 44593,
  [SMALL_STATE(2219)] = 44656,
  [SMALL_STATE(2220)] = 44719,
  [SMALL_STATE(2221)] = 44782,
  [SMALL_STATE(2222)] = 44857,
  [SMALL_STATE(2223)] = 44956,
  [SMALL_STATE(2224)] = 45035,
  [SMALL_STATE(2225)] = 45098,
  [SMALL_STATE(2226)] = 45161,
  [SMALL_STATE(2227)] = 45224,
  [SMALL_STATE(2228)] = 45319,
  [SMALL_STATE(2229)] = 45382,
  [SMALL_STATE(2230)] = 45445,
  [SMALL_STATE(2231)] = 45528,
  [SMALL_STATE(2232)] = 45593,
  [SMALL_STATE(2233)] = 45678,
  [SMALL_STATE(2234)] = 45763,
  [SMALL_STATE(2235)] = 45860,
  [SMALL_STATE(2236)] = 45949,
  [SMALL_STATE(2237)] = 46012,
  [SMALL_STATE(2238)] = 46105,
  [SMALL_STATE(2239)] = 46168,
  [SMALL_STATE(2240)] = 46231,
  [SMALL_STATE(2241)] = 46294,
  [SMALL_STATE(2242)] = 46357,
  [SMALL_STATE(2243)] = 46422,
  [SMALL_STATE(2244)] = 46485,
  [SMALL_STATE(2245)] = 46582,
  [SMALL_STATE(2246)] = 46679,
  [SMALL_STATE(2247)] = 46742,
  [SMALL_STATE(2248)] = 46805,
  [SMALL_STATE(2249)] = 46868,
  [SMALL_STATE(2250)] = 46931,
  [SMALL_STATE(2251)] = 46994,
  [SMALL_STATE(2252)] = 47057,
  [SMALL_STATE(2253)] = 47120,
  [SMALL_STATE(2254)] = 47183,
  [SMALL_STATE(2255)] = 47246,
  [SMALL_STATE(2256)] = 47343,
  [SMALL_STATE(2257)] = 47406,
  [SMALL_STATE(2258)] = 47471,
  [SMALL_STATE(2259)] = 47534,
  [SMALL_STATE(2260)] = 47597,
  [SMALL_STATE(2261)] = 47660,
  [SMALL_STATE(2262)] = 47723,
  [SMALL_STATE(2263)] = 47786,
  [SMALL_STATE(2264)] = 47849,
  [SMALL_STATE(2265)] = 47912,
  [SMALL_STATE(2266)] = 47975,
  [SMALL_STATE(2267)] = 48038,
  [SMALL_STATE(2268)] = 48101,
  [SMALL_STATE(2269)] = 48164,
  [SMALL_STATE(2270)] = 48233,
  [SMALL_STATE(2271)] = 48296,
  [SMALL_STATE(2272)] = 48359,
  [SMALL_STATE(2273)] = 48422,
  [SMALL_STATE(2274)] = 48521,
  [SMALL_STATE(2275)] = 48620,
  [SMALL_STATE(2276)] = 48683,
  [SMALL_STATE(2277)] = 48752,
  [SMALL_STATE(2278)] = 48815,
  [SMALL_STATE(2279)] = 48884,
  [SMALL_STATE(2280)] = 48947,
  [SMALL_STATE(2281)] = 49010,
  [SMALL_STATE(2282)] = 49073,
  [SMALL_STATE(2283)] = 49136,
  [SMALL_STATE(2284)] = 49199,
  [SMALL_STATE(2285)] = 49262,
  [SMALL_STATE(2286)] = 49325,
  [SMALL_STATE(2287)] = 49388,
  [SMALL_STATE(2288)] = 49453,
  [SMALL_STATE(2289)] = 49516,
  [SMALL_STATE(2290)] = 49579,
  [SMALL_STATE(2291)] = 49644,
  [SMALL_STATE(2292)] = 49707,
  [SMALL_STATE(2293)] = 49770,
  [SMALL_STATE(2294)] = 49833,
  [SMALL_STATE(2295)] = 49896,
  [SMALL_STATE(2296)] = 49959,
  [SMALL_STATE(2297)] = 50022,
  [SMALL_STATE(2298)] = 50085,
  [SMALL_STATE(2299)] = 50148,
  [SMALL_STATE(2300)] = 50211,
  [SMALL_STATE(2301)] = 50274,
  [SMALL_STATE(2302)] = 50337,
  [SMALL_STATE(2303)] = 50400,
  [SMALL_STATE(2304)] = 50469,
  [SMALL_STATE(2305)] = 50538,
  [SMALL_STATE(2306)] = 50601,
  [SMALL_STATE(2307)] = 50664,
  [SMALL_STATE(2308)] = 50727,
  [SMALL_STATE(2309)] = 50790,
  [SMALL_STATE(2310)] = 50853,
  [SMALL_STATE(2311)] = 50916,
  [SMALL_STATE(2312)] = 50979,
  [SMALL_STATE(2313)] = 51042,
  [SMALL_STATE(2314)] = 51105,
  [SMALL_STATE(2315)] = 51168,
  [SMALL_STATE(2316)] = 51231,
  [SMALL_STATE(2317)] = 51294,
  [SMALL_STATE(2318)] = 51357,
  [SMALL_STATE(2319)] = 51420,
  [SMALL_STATE(2320)] = 51483,
  [SMALL_STATE(2321)] = 51546,
  [SMALL_STATE(2322)] = 51609,
  [SMALL_STATE(2323)] = 51672,
  [SMALL_STATE(2324)] = 51735,
  [SMALL_STATE(2325)] = 51798,
  [SMALL_STATE(2326)] = 51861,
  [SMALL_STATE(2327)] = 51930,
  [SMALL_STATE(2328)] = 51999,
  [SMALL_STATE(2329)] = 52062,
  [SMALL_STATE(2330)] = 52125,
  [SMALL_STATE(2331)] = 52188,
  [SMALL_STATE(2332)] = 52251,
  [SMALL_STATE(2333)] = 52314,
  [SMALL_STATE(2334)] = 52377,
  [SMALL_STATE(2335)] = 52440,
  [SMALL_STATE(2336)] = 52511,
  [SMALL_STATE(2337)] = 52574,
  [SMALL_STATE(2338)] = 52637,
  [SMALL_STATE(2339)] = 52706,
  [SMALL_STATE(2340)] = 52769,
  [SMALL_STATE(2341)] = 52832,
  [SMALL_STATE(2342)] = 52901,
  [SMALL_STATE(2343)] = 52964,
  [SMALL_STATE(2344)] = 53027,
  [SMALL_STATE(2345)] = 53090,
  [SMALL_STATE(2346)] = 53153,
  [SMALL_STATE(2347)] = 53216,
  [SMALL_STATE(2348)] = 53279,
  [SMALL_STATE(2349)] = 53342,
  [SMALL_STATE(2350)] = 53405,
  [SMALL_STATE(2351)] = 53468,
  [SMALL_STATE(2352)] = 53531,
  [SMALL_STATE(2353)] = 53594,
  [SMALL_STATE(2354)] = 53657,
  [SMALL_STATE(2355)] = 53720,
  [SMALL_STATE(2356)] = 53783,
  [SMALL_STATE(2357)] = 53846,
  [SMALL_STATE(2358)] = 53911,
  [SMALL_STATE(2359)] = 53974,
  [SMALL_STATE(2360)] = 54037,
  [SMALL_STATE(2361)] = 54100,
  [SMALL_STATE(2362)] = 54163,
  [SMALL_STATE(2363)] = 54226,
  [SMALL_STATE(2364)] = 54289,
  [SMALL_STATE(2365)] = 54352,
  [SMALL_STATE(2366)] = 54415,
  [SMALL_STATE(2367)] = 54478,
  [SMALL_STATE(2368)] = 54541,
  [SMALL_STATE(2369)] = 54604,
  [SMALL_STATE(2370)] = 54667,
  [SMALL_STATE(2371)] = 54730,
  [SMALL_STATE(2372)] = 54793,
  [SMALL_STATE(2373)] = 54892,
  [SMALL_STATE(2374)] = 54955,
  [SMALL_STATE(2375)] = 55018,
  [SMALL_STATE(2376)] = 55081,
  [SMALL_STATE(2377)] = 55152,
  [SMALL_STATE(2378)] = 55215,
  [SMALL_STATE(2379)] = 55282,
  [SMALL_STATE(2380)] = 55345,
  [SMALL_STATE(2381)] = 55410,
  [SMALL_STATE(2382)] = 55473,
  [SMALL_STATE(2383)] = 55536,
  [SMALL_STATE(2384)] = 55605,
  [SMALL_STATE(2385)] = 55668,
  [SMALL_STATE(2386)] = 55731,
  [SMALL_STATE(2387)] = 55794,
  [SMALL_STATE(2388)] = 55859,
  [SMALL_STATE(2389)] = 55922,
  [SMALL_STATE(2390)] = 55985,
  [SMALL_STATE(2391)] = 56048,
  [SMALL_STATE(2392)] = 56111,
  [SMALL_STATE(2393)] = 56174,
  [SMALL_STATE(2394)] = 56237,
  [SMALL_STATE(2395)] = 56300,
  [SMALL_STATE(2396)] = 56363,
  [SMALL_STATE(2397)] = 56426,
  [SMALL_STATE(2398)] = 56491,
  [SMALL_STATE(2399)] = 56554,
  [SMALL_STATE(2400)] = 56617,
  [SMALL_STATE(2401)] = 56680,
  [SMALL_STATE(2402)] = 56745,
  [SMALL_STATE(2403)] = 56808,
  [SMALL_STATE(2404)] = 56871,
  [SMALL_STATE(2405)] = 56934,
  [SMALL_STATE(2406)] = 56997,
  [SMALL_STATE(2407)] = 57060,
  [SMALL_STATE(2408)] = 57123,
  [SMALL_STATE(2409)] = 57190,
  [SMALL_STATE(2410)] = 57253,
  [SMALL_STATE(2411)] = 57316,
  [SMALL_STATE(2412)] = 57379,
  [SMALL_STATE(2413)] = 57442,
  [SMALL_STATE(2414)] = 57505,
  [SMALL_STATE(2415)] = 57568,
  [SMALL_STATE(2416)] = 57631,
  [SMALL_STATE(2417)] = 57694,
  [SMALL_STATE(2418)] = 57757,
  [SMALL_STATE(2419)] = 57820,
  [SMALL_STATE(2420)] = 57883,
  [SMALL_STATE(2421)] = 57946,
  [SMALL_STATE(2422)] = 58009,
  [SMALL_STATE(2423)] = 58072,
  [SMALL_STATE(2424)] = 58135,
  [SMALL_STATE(2425)] = 58198,
  [SMALL_STATE(2426)] = 58261,
  [SMALL_STATE(2427)] = 58324,
  [SMALL_STATE(2428)] = 58387,
  [SMALL_STATE(2429)] = 58450,
  [SMALL_STATE(2430)] = 58513,
  [SMALL_STATE(2431)] = 58576,
  [SMALL_STATE(2432)] = 58639,
  [SMALL_STATE(2433)] = 58702,
  [SMALL_STATE(2434)] = 58765,
  [SMALL_STATE(2435)] = 58828,
  [SMALL_STATE(2436)] = 58891,
  [SMALL_STATE(2437)] = 58954,
  [SMALL_STATE(2438)] = 59017,
  [SMALL_STATE(2439)] = 59080,
  [SMALL_STATE(2440)] = 59143,
  [SMALL_STATE(2441)] = 59208,
  [SMALL_STATE(2442)] = 59271,
  [SMALL_STATE(2443)] = 59334,
  [SMALL_STATE(2444)] = 59397,
  [SMALL_STATE(2445)] = 59460,
  [SMALL_STATE(2446)] = 59522,
  [SMALL_STATE(2447)] = 59584,
  [SMALL_STATE(2448)] = 59652,
  [SMALL_STATE(2449)] = 59720,
  [SMALL_STATE(2450)] = 59782,
  [SMALL_STATE(2451)] = 59844,
  [SMALL_STATE(2452)] = 59906,
  [SMALL_STATE(2453)] = 59968,
  [SMALL_STATE(2454)] = 60030,
  [SMALL_STATE(2455)] = 60092,
  [SMALL_STATE(2456)] = 60154,
  [SMALL_STATE(2457)] = 60216,
  [SMALL_STATE(2458)] = 60278,
  [SMALL_STATE(2459)] = 60340,
  [SMALL_STATE(2460)] = 60402,
  [SMALL_STATE(2461)] = 60464,
  [SMALL_STATE(2462)] = 60526,
  [SMALL_STATE(2463)] = 60588,
  [SMALL_STATE(2464)] = 60650,
  [SMALL_STATE(2465)] = 60712,
  [SMALL_STATE(2466)] = 60776,
  [SMALL_STATE(2467)] = 60838,
  [SMALL_STATE(2468)] = 60900,
  [SMALL_STATE(2469)] = 60962,
  [SMALL_STATE(2470)] = 61024,
  [SMALL_STATE(2471)] = 61086,
  [SMALL_STATE(2472)] = 61148,
  [SMALL_STATE(2473)] = 61210,
  [SMALL_STATE(2474)] = 61272,
  [SMALL_STATE(2475)] = 61334,
  [SMALL_STATE(2476)] = 61396,
  [SMALL_STATE(2477)] = 61458,
  [SMALL_STATE(2478)] = 61520,
  [SMALL_STATE(2479)] = 61582,
  [SMALL_STATE(2480)] = 61644,
  [SMALL_STATE(2481)] = 61706,
  [SMALL_STATE(2482)] = 61768,
  [SMALL_STATE(2483)] = 61830,
  [SMALL_STATE(2484)] = 61892,
  [SMALL_STATE(2485)] = 61954,
  [SMALL_STATE(2486)] = 62016,
  [SMALL_STATE(2487)] = 62078,
  [SMALL_STATE(2488)] = 62140,
  [SMALL_STATE(2489)] = 62202,
  [SMALL_STATE(2490)] = 62264,
  [SMALL_STATE(2491)] = 62326,
  [SMALL_STATE(2492)] = 62388,
  [SMALL_STATE(2493)] = 62450,
  [SMALL_STATE(2494)] = 62512,
  [SMALL_STATE(2495)] = 62574,
  [SMALL_STATE(2496)] = 62636,
  [SMALL_STATE(2497)] = 62698,
  [SMALL_STATE(2498)] = 62760,
  [SMALL_STATE(2499)] = 62822,
  [SMALL_STATE(2500)] = 62884,
  [SMALL_STATE(2501)] = 62946,
  [SMALL_STATE(2502)] = 63008,
  [SMALL_STATE(2503)] = 63076,
  [SMALL_STATE(2504)] = 63138,
  [SMALL_STATE(2505)] = 63200,
  [SMALL_STATE(2506)] = 63262,
  [SMALL_STATE(2507)] = 63324,
  [SMALL_STATE(2508)] = 63386,
  [SMALL_STATE(2509)] = 63448,
  [SMALL_STATE(2510)] = 63510,
  [SMALL_STATE(2511)] = 63572,
  [SMALL_STATE(2512)] = 63634,
  [SMALL_STATE(2513)] = 63696,
  [SMALL_STATE(2514)] = 63758,
  [SMALL_STATE(2515)] = 63820,
  [SMALL_STATE(2516)] = 63884,
  [SMALL_STATE(2517)] = 63946,
  [SMALL_STATE(2518)] = 64008,
  [SMALL_STATE(2519)] = 64070,
  [SMALL_STATE(2520)] = 64132,
  [SMALL_STATE(2521)] = 64196,
  [SMALL_STATE(2522)] = 64258,
  [SMALL_STATE(2523)] = 64320,
  [SMALL_STATE(2524)] = 64382,
  [SMALL_STATE(2525)] = 64444,
  [SMALL_STATE(2526)] = 64506,
  [SMALL_STATE(2527)] = 64568,
  [SMALL_STATE(2528)] = 64630,
  [SMALL_STATE(2529)] = 64692,
  [SMALL_STATE(2530)] = 64754,
  [SMALL_STATE(2531)] = 64816,
  [SMALL_STATE(2532)] = 64878,
  [SMALL_STATE(2533)] = 64940,
  [SMALL_STATE(2534)] = 65002,
  [SMALL_STATE(2535)] = 65074,
  [SMALL_STATE(2536)] = 65148,
  [SMALL_STATE(2537)] = 65222,
  [SMALL_STATE(2538)] = 65300,
  [SMALL_STATE(2539)] = 65398,
  [SMALL_STATE(2540)] = 65496,
  [SMALL_STATE(2541)] = 65590,
  [SMALL_STATE(2542)] = 65672,
  [SMALL_STATE(2543)] = 65756,
  [SMALL_STATE(2544)] = 65840,
  [SMALL_STATE(2545)] = 65928,
  [SMALL_STATE(2546)] = 66020,
  [SMALL_STATE(2547)] = 66116,
  [SMALL_STATE(2548)] = 66212,
  [SMALL_STATE(2549)] = 66310,
  [SMALL_STATE(2550)] = 66406,
  [SMALL_STATE(2551)] = 66504,
  [SMALL_STATE(2552)] = 66566,
  [SMALL_STATE(2553)] = 66628,
  [SMALL_STATE(2554)] = 66690,
  [SMALL_STATE(2555)] = 66752,
  [SMALL_STATE(2556)] = 66814,
  [SMALL_STATE(2557)] = 66876,
  [SMALL_STATE(2558)] = 66944,
  [SMALL_STATE(2559)] = 67012,
  [SMALL_STATE(2560)] = 67074,
  [SMALL_STATE(2561)] = 67136,
  [SMALL_STATE(2562)] = 67198,
  [SMALL_STATE(2563)] = 67260,
  [SMALL_STATE(2564)] = 67322,
  [SMALL_STATE(2565)] = 67420,
  [SMALL_STATE(2566)] = 67482,
  [SMALL_STATE(2567)] = 67544,
  [SMALL_STATE(2568)] = 67606,
  [SMALL_STATE(2569)] = 67668,
  [SMALL_STATE(2570)] = 67730,
  [SMALL_STATE(2571)] = 67798,
  [SMALL_STATE(2572)] = 67866,
  [SMALL_STATE(2573)] = 67928,
  [SMALL_STATE(2574)] = 67990,
  [SMALL_STATE(2575)] = 68052,
  [SMALL_STATE(2576)] = 68114,
  [SMALL_STATE(2577)] = 68178,
  [SMALL_STATE(2578)] = 68240,
  [SMALL_STATE(2579)] = 68302,
  [SMALL_STATE(2580)] = 68370,
  [SMALL_STATE(2581)] = 68432,
  [SMALL_STATE(2582)] = 68494,
  [SMALL_STATE(2583)] = 68556,
  [SMALL_STATE(2584)] = 68618,
  [SMALL_STATE(2585)] = 68680,
  [SMALL_STATE(2586)] = 68742,
  [SMALL_STATE(2587)] = 68804,
  [SMALL_STATE(2588)] = 68866,
  [SMALL_STATE(2589)] = 68928,
  [SMALL_STATE(2590)] = 68990,
  [SMALL_STATE(2591)] = 69052,
  [SMALL_STATE(2592)] = 69114,
  [SMALL_STATE(2593)] = 69176,
  [SMALL_STATE(2594)] = 69238,
  [SMALL_STATE(2595)] = 69300,
  [SMALL_STATE(2596)] = 69362,
  [SMALL_STATE(2597)] = 69424,
  [SMALL_STATE(2598)] = 69486,
  [SMALL_STATE(2599)] = 69548,
  [SMALL_STATE(2600)] = 69610,
  [SMALL_STATE(2601)] = 69672,
  [SMALL_STATE(2602)] = 69734,
  [SMALL_STATE(2603)] = 69796,
  [SMALL_STATE(2604)] = 69858,
  [SMALL_STATE(2605)] = 69920,
  [SMALL_STATE(2606)] = 69982,
  [SMALL_STATE(2607)] = 70044,
  [SMALL_STATE(2608)] = 70106,
  [SMALL_STATE(2609)] = 70168,
  [SMALL_STATE(2610)] = 70230,
  [SMALL_STATE(2611)] = 70292,
  [SMALL_STATE(2612)] = 70354,
  [SMALL_STATE(2613)] = 70452,
  [SMALL_STATE(2614)] = 70514,
  [SMALL_STATE(2615)] = 70612,
  [SMALL_STATE(2616)] = 70674,
  [SMALL_STATE(2617)] = 70736,
  [SMALL_STATE(2618)] = 70798,
  [SMALL_STATE(2619)] = 70860,
  [SMALL_STATE(2620)] = 70922,
  [SMALL_STATE(2621)] = 70984,
  [SMALL_STATE(2622)] = 71046,
  [SMALL_STATE(2623)] = 71108,
  [SMALL_STATE(2624)] = 71170,
  [SMALL_STATE(2625)] = 71232,
  [SMALL_STATE(2626)] = 71294,
  [SMALL_STATE(2627)] = 71362,
  [SMALL_STATE(2628)] = 71424,
  [SMALL_STATE(2629)] = 71492,
  [SMALL_STATE(2630)] = 71554,
  [SMALL_STATE(2631)] = 71652,
  [SMALL_STATE(2632)] = 71714,
  [SMALL_STATE(2633)] = 71776,
  [SMALL_STATE(2634)] = 71874,
  [SMALL_STATE(2635)] = 71970,
  [SMALL_STATE(2636)] = 72031,
  [SMALL_STATE(2637)] = 72128,
  [SMALL_STATE(2638)] = 72189,
  [SMALL_STATE(2639)] = 72250,
  [SMALL_STATE(2640)] = 72345,
  [SMALL_STATE(2641)] = 72406,
  [SMALL_STATE(2642)] = 72467,
  [SMALL_STATE(2643)] = 72528,
  [SMALL_STATE(2644)] = 72589,
  [SMALL_STATE(2645)] = 72650,
  [SMALL_STATE(2646)] = 72747,
  [SMALL_STATE(2647)] = 72808,
  [SMALL_STATE(2648)] = 72869,
  [SMALL_STATE(2649)] = 72930,
  [SMALL_STATE(2650)] = 72991,
  [SMALL_STATE(2651)] = 73084,
  [SMALL_STATE(2652)] = 73145,
  [SMALL_STATE(2653)] = 73206,
  [SMALL_STATE(2654)] = 73267,
  [SMALL_STATE(2655)] = 73328,
  [SMALL_STATE(2656)] = 73393,
  [SMALL_STATE(2657)] = 73474,
  [SMALL_STATE(2658)] = 73557,
  [SMALL_STATE(2659)] = 73654,
  [SMALL_STATE(2660)] = 73737,
  [SMALL_STATE(2661)] = 73832,
  [SMALL_STATE(2662)] = 73949,
  [SMALL_STATE(2663)] = 74022,
  [SMALL_STATE(2664)] = 74095,
  [SMALL_STATE(2665)] = 74172,
  [SMALL_STATE(2666)] = 74235,
  [SMALL_STATE(2667)] = 74322,
  [SMALL_STATE(2668)] = 74413,
  [SMALL_STATE(2669)] = 74508,
  [SMALL_STATE(2670)] = 74605,
  [SMALL_STATE(2671)] = 74666,
  [SMALL_STATE(2672)] = 74763,
  [SMALL_STATE(2673)] = 74860,
  [SMALL_STATE(2674)] = 74977,
  [SMALL_STATE(2675)] = 75094,
  [SMALL_STATE(2676)] = 75211,
  [SMALL_STATE(2677)] = 75272,
  [SMALL_STATE(2678)] = 75333,
  [SMALL_STATE(2679)] = 75394,
  [SMALL_STATE(2680)] = 75455,
  [SMALL_STATE(2681)] = 75550,
  [SMALL_STATE(2682)] = 75611,
  [SMALL_STATE(2683)] = 75672,
  [SMALL_STATE(2684)] = 75733,
  [SMALL_STATE(2685)] = 75794,
  [SMALL_STATE(2686)] = 75855,
  [SMALL_STATE(2687)] = 75972,
  [SMALL_STATE(2688)] = 76035,
  [SMALL_STATE(2689)] = 76100,
  [SMALL_STATE(2690)] = 76165,
  [SMALL_STATE(2691)] = 76228,
  [SMALL_STATE(2692)] = 76293,
  [SMALL_STATE(2693)] = 76358,
  [SMALL_STATE(2694)] = 76419,
  [SMALL_STATE(2695)] = 76480,
  [SMALL_STATE(2696)] = 76541,
  [SMALL_STATE(2697)] = 76602,
  [SMALL_STATE(2698)] = 76663,
  [SMALL_STATE(2699)] = 76724,
  [SMALL_STATE(2700)] = 76841,
  [SMALL_STATE(2701)] = 76912,
  [SMALL_STATE(2702)] = 76973,
  [SMALL_STATE(2703)] = 77038,
  [SMALL_STATE(2704)] = 77099,
  [SMALL_STATE(2705)] = 77160,
  [SMALL_STATE(2706)] = 77224,
  [SMALL_STATE(2707)] = 77284,
  [SMALL_STATE(2708)] = 77344,
  [SMALL_STATE(2709)] = 77404,
  [SMALL_STATE(2710)] = 77464,
  [SMALL_STATE(2711)] = 77526,
  [SMALL_STATE(2712)] = 77586,
  [SMALL_STATE(2713)] = 77646,
  [SMALL_STATE(2714)] = 77706,
  [SMALL_STATE(2715)] = 77766,
  [SMALL_STATE(2716)] = 77862,
  [SMALL_STATE(2717)] = 77922,
  [SMALL_STATE(2718)] = 77986,
  [SMALL_STATE(2719)] = 78046,
  [SMALL_STATE(2720)] = 78110,
  [SMALL_STATE(2721)] = 78174,
  [SMALL_STATE(2722)] = 78236,
  [SMALL_STATE(2723)] = 78300,
  [SMALL_STATE(2724)] = 78360,
  [SMALL_STATE(2725)] = 78424,
  [SMALL_STATE(2726)] = 78484,
  [SMALL_STATE(2727)] = 78548,
  [SMALL_STATE(2728)] = 78612,
  [SMALL_STATE(2729)] = 78708,
  [SMALL_STATE(2730)] = 78768,
  [SMALL_STATE(2731)] = 78828,
  [SMALL_STATE(2732)] = 78892,
  [SMALL_STATE(2733)] = 78956,
  [SMALL_STATE(2734)] = 79018,
  [SMALL_STATE(2735)] = 79078,
  [SMALL_STATE(2736)] = 79142,
  [SMALL_STATE(2737)] = 79202,
  [SMALL_STATE(2738)] = 79264,
  [SMALL_STATE(2739)] = 79324,
  [SMALL_STATE(2740)] = 79384,
  [SMALL_STATE(2741)] = 79444,
  [SMALL_STATE(2742)] = 79504,
  [SMALL_STATE(2743)] = 79600,
  [SMALL_STATE(2744)] = 79660,
  [SMALL_STATE(2745)] = 79722,
  [SMALL_STATE(2746)] = 79782,
  [SMALL_STATE(2747)] = 79842,
  [SMALL_STATE(2748)] = 79902,
  [SMALL_STATE(2749)] = 79962,
  [SMALL_STATE(2750)] = 80026,
  [SMALL_STATE(2751)] = 80129,
  [SMALL_STATE(2752)] = 80230,
  [SMALL_STATE(2753)] = 80291,
  [SMALL_STATE(2754)] = 80392,
  [SMALL_STATE(2755)] = 80451,
  [SMALL_STATE(2756)] = 80552,
  [SMALL_STATE(2757)] = 80611,
  [SMALL_STATE(2758)] = 80714,
  [SMALL_STATE(2759)] = 80817,
  [SMALL_STATE(2760)] = 80912,
  [SMALL_STATE(2761)] = 80971,
  [SMALL_STATE(2762)] = 81030,
  [SMALL_STATE(2763)] = 81089,
  [SMALL_STATE(2764)] = 81184,
  [SMALL_STATE(2765)] = 81243,
  [SMALL_STATE(2766)] = 81302,
  [SMALL_STATE(2767)] = 81361,
  [SMALL_STATE(2768)] = 81420,
  [SMALL_STATE(2769)] = 81533,
  [SMALL_STATE(2770)] = 81592,
  [SMALL_STATE(2771)] = 81705,
  [SMALL_STATE(2772)] = 81764,
  [SMALL_STATE(2773)] = 81877,
  [SMALL_STATE(2774)] = 81980,
  [SMALL_STATE(2775)] = 82039,
  [SMALL_STATE(2776)] = 82098,
  [SMALL_STATE(2777)] = 82199,
  [SMALL_STATE(2778)] = 82302,
  [SMALL_STATE(2779)] = 82361,
  [SMALL_STATE(2780)] = 82420,
  [SMALL_STATE(2781)] = 82479,
  [SMALL_STATE(2782)] = 82538,
  [SMALL_STATE(2783)] = 82597,
  [SMALL_STATE(2784)] = 82656,
  [SMALL_STATE(2785)] = 82715,
  [SMALL_STATE(2786)] = 82774,
  [SMALL_STATE(2787)] = 82833,
  [SMALL_STATE(2788)] = 82892,
  [SMALL_STATE(2789)] = 82951,
  [SMALL_STATE(2790)] = 83010,
  [SMALL_STATE(2791)] = 83113,
  [SMALL_STATE(2792)] = 83172,
  [SMALL_STATE(2793)] = 83231,
  [SMALL_STATE(2794)] = 83290,
  [SMALL_STATE(2795)] = 83353,
  [SMALL_STATE(2796)] = 83412,
  [SMALL_STATE(2797)] = 83471,
  [SMALL_STATE(2798)] = 83530,
  [SMALL_STATE(2799)] = 83589,
  [SMALL_STATE(2800)] = 83648,
  [SMALL_STATE(2801)] = 83707,
  [SMALL_STATE(2802)] = 83766,
  [SMALL_STATE(2803)] = 83867,
  [SMALL_STATE(2804)] = 83926,
  [SMALL_STATE(2805)] = 83985,
  [SMALL_STATE(2806)] = 84044,
  [SMALL_STATE(2807)] = 84103,
  [SMALL_STATE(2808)] = 84162,
  [SMALL_STATE(2809)] = 84265,
  [SMALL_STATE(2810)] = 84324,
  [SMALL_STATE(2811)] = 84383,
  [SMALL_STATE(2812)] = 84442,
  [SMALL_STATE(2813)] = 84501,
  [SMALL_STATE(2814)] = 84562,
  [SMALL_STATE(2815)] = 84621,
  [SMALL_STATE(2816)] = 84680,
  [SMALL_STATE(2817)] = 84739,
  [SMALL_STATE(2818)] = 84798,
  [SMALL_STATE(2819)] = 84901,
  [SMALL_STATE(2820)] = 84960,
  [SMALL_STATE(2821)] = 85023,
  [SMALL_STATE(2822)] = 85082,
  [SMALL_STATE(2823)] = 85141,
  [SMALL_STATE(2824)] = 85200,
  [SMALL_STATE(2825)] = 85259,
  [SMALL_STATE(2826)] = 85318,
  [SMALL_STATE(2827)] = 85377,
  [SMALL_STATE(2828)] = 85436,
  [SMALL_STATE(2829)] = 85549,
  [SMALL_STATE(2830)] = 85612,
  [SMALL_STATE(2831)] = 85675,
  [SMALL_STATE(2832)] = 85736,
  [SMALL_STATE(2833)] = 85831,
  [SMALL_STATE(2834)] = 85894,
  [SMALL_STATE(2835)] = 85957,
  [SMALL_STATE(2836)] = 86016,
  [SMALL_STATE(2837)] = 86075,
  [SMALL_STATE(2838)] = 86178,
  [SMALL_STATE(2839)] = 86237,
  [SMALL_STATE(2840)] = 86350,
  [SMALL_STATE(2841)] = 86409,
  [SMALL_STATE(2842)] = 86468,
  [SMALL_STATE(2843)] = 86527,
  [SMALL_STATE(2844)] = 86628,
  [SMALL_STATE(2845)] = 86687,
  [SMALL_STATE(2846)] = 86746,
  [SMALL_STATE(2847)] = 86847,
  [SMALL_STATE(2848)] = 86906,
  [SMALL_STATE(2849)] = 86965,
  [SMALL_STATE(2850)] = 87024,
  [SMALL_STATE(2851)] = 87083,
  [SMALL_STATE(2852)] = 87142,
  [SMALL_STATE(2853)] = 87201,
  [SMALL_STATE(2854)] = 87314,
  [SMALL_STATE(2855)] = 87415,
  [SMALL_STATE(2856)] = 87474,
  [SMALL_STATE(2857)] = 87575,
  [SMALL_STATE(2858)] = 87634,
  [SMALL_STATE(2859)] = 87735,
  [SMALL_STATE(2860)] = 87794,
  [SMALL_STATE(2861)] = 87897,
  [SMALL_STATE(2862)] = 87956,
  [SMALL_STATE(2863)] = 88015,
  [SMALL_STATE(2864)] = 88074,
  [SMALL_STATE(2865)] = 88133,
  [SMALL_STATE(2866)] = 88192,
  [SMALL_STATE(2867)] = 88251,
  [SMALL_STATE(2868)] = 88310,
  [SMALL_STATE(2869)] = 88369,
  [SMALL_STATE(2870)] = 88428,
  [SMALL_STATE(2871)] = 88487,
  [SMALL_STATE(2872)] = 88546,
  [SMALL_STATE(2873)] = 88605,
  [SMALL_STATE(2874)] = 88708,
  [SMALL_STATE(2875)] = 88767,
  [SMALL_STATE(2876)] = 88877,
  [SMALL_STATE(2877)] = 88987,
  [SMALL_STATE(2878)] = 89049,
  [SMALL_STATE(2879)] = 89111,
  [SMALL_STATE(2880)] = 89173,
  [SMALL_STATE(2881)] = 89283,
  [SMALL_STATE(2882)] = 89345,
  [SMALL_STATE(2883)] = 89405,
  [SMALL_STATE(2884)] = 89515,
  [SMALL_STATE(2885)] = 89575,
  [SMALL_STATE(2886)] = 89685,
  [SMALL_STATE(2887)] = 89747,
  [SMALL_STATE(2888)] = 89857,
  [SMALL_STATE(2889)] = 89919,
  [SMALL_STATE(2890)] = 89976,
  [SMALL_STATE(2891)] = 90033,
  [SMALL_STATE(2892)] = 90090,
  [SMALL_STATE(2893)] = 90147,
  [SMALL_STATE(2894)] = 90204,
  [SMALL_STATE(2895)] = 90261,
  [SMALL_STATE(2896)] = 90320,
  [SMALL_STATE(2897)] = 90381,
  [SMALL_STATE(2898)] = 90438,
  [SMALL_STATE(2899)] = 90499,
  [SMALL_STATE(2900)] = 90556,
  [SMALL_STATE(2901)] = 90613,
  [SMALL_STATE(2902)] = 90670,
  [SMALL_STATE(2903)] = 90727,
  [SMALL_STATE(2904)] = 90788,
  [SMALL_STATE(2905)] = 90849,
  [SMALL_STATE(2906)] = 90908,
  [SMALL_STATE(2907)] = 90965,
  [SMALL_STATE(2908)] = 91022,
  [SMALL_STATE(2909)] = 91079,
  [SMALL_STATE(2910)] = 91136,
  [SMALL_STATE(2911)] = 91195,
  [SMALL_STATE(2912)] = 91252,
  [SMALL_STATE(2913)] = 91309,
  [SMALL_STATE(2914)] = 91366,
  [SMALL_STATE(2915)] = 91423,
  [SMALL_STATE(2916)] = 91480,
  [SMALL_STATE(2917)] = 91539,
  [SMALL_STATE(2918)] = 91598,
  [SMALL_STATE(2919)] = 91659,
  [SMALL_STATE(2920)] = 91720,
  [SMALL_STATE(2921)] = 91777,
  [SMALL_STATE(2922)] = 91834,
  [SMALL_STATE(2923)] = 91890,
  [SMALL_STATE(2924)] = 91946,
  [SMALL_STATE(2925)] = 92002,
  [SMALL_STATE(2926)] = 92058,
  [SMALL_STATE(2927)] = 92114,
  [SMALL_STATE(2928)] = 92170,
  [SMALL_STATE(2929)] = 92226,
  [SMALL_STATE(2930)] = 92282,
  [SMALL_STATE(2931)] = 92338,
  [SMALL_STATE(2932)] = 92394,
  [SMALL_STATE(2933)] = 92450,
  [SMALL_STATE(2934)] = 92506,
  [SMALL_STATE(2935)] = 92562,
  [SMALL_STATE(2936)] = 92618,
  [SMALL_STATE(2937)] = 92674,
  [SMALL_STATE(2938)] = 92730,
  [SMALL_STATE(2939)] = 92786,
  [SMALL_STATE(2940)] = 92842,
  [SMALL_STATE(2941)] = 92898,
  [SMALL_STATE(2942)] = 92954,
  [SMALL_STATE(2943)] = 93010,
  [SMALL_STATE(2944)] = 93066,
  [SMALL_STATE(2945)] = 93122,
  [SMALL_STATE(2946)] = 93178,
  [SMALL_STATE(2947)] = 93234,
  [SMALL_STATE(2948)] = 93290,
  [SMALL_STATE(2949)] = 93346,
  [SMALL_STATE(2950)] = 93402,
  [SMALL_STATE(2951)] = 93458,
  [SMALL_STATE(2952)] = 93514,
  [SMALL_STATE(2953)] = 93570,
  [SMALL_STATE(2954)] = 93626,
  [SMALL_STATE(2955)] = 93682,
  [SMALL_STATE(2956)] = 93738,
  [SMALL_STATE(2957)] = 93794,
  [SMALL_STATE(2958)] = 93850,
  [SMALL_STATE(2959)] = 93906,
  [SMALL_STATE(2960)] = 93962,
  [SMALL_STATE(2961)] = 94018,
  [SMALL_STATE(2962)] = 94074,
  [SMALL_STATE(2963)] = 94130,
  [SMALL_STATE(2964)] = 94186,
  [SMALL_STATE(2965)] = 94242,
  [SMALL_STATE(2966)] = 94298,
  [SMALL_STATE(2967)] = 94354,
  [SMALL_STATE(2968)] = 94410,
  [SMALL_STATE(2969)] = 94466,
  [SMALL_STATE(2970)] = 94524,
  [SMALL_STATE(2971)] = 94580,
  [SMALL_STATE(2972)] = 94636,
  [SMALL_STATE(2973)] = 94692,
  [SMALL_STATE(2974)] = 94748,
  [SMALL_STATE(2975)] = 94804,
  [SMALL_STATE(2976)] = 94860,
  [SMALL_STATE(2977)] = 94916,
  [SMALL_STATE(2978)] = 94972,
  [SMALL_STATE(2979)] = 95028,
  [SMALL_STATE(2980)] = 95084,
  [SMALL_STATE(2981)] = 95142,
  [SMALL_STATE(2982)] = 95198,
  [SMALL_STATE(2983)] = 95254,
  [SMALL_STATE(2984)] = 95310,
  [SMALL_STATE(2985)] = 95366,
  [SMALL_STATE(2986)] = 95422,
  [SMALL_STATE(2987)] = 95478,
  [SMALL_STATE(2988)] = 95534,
  [SMALL_STATE(2989)] = 95590,
  [SMALL_STATE(2990)] = 95646,
  [SMALL_STATE(2991)] = 95702,
  [SMALL_STATE(2992)] = 95758,
  [SMALL_STATE(2993)] = 95814,
  [SMALL_STATE(2994)] = 95872,
  [SMALL_STATE(2995)] = 95928,
  [SMALL_STATE(2996)] = 95984,
  [SMALL_STATE(2997)] = 96040,
  [SMALL_STATE(2998)] = 96096,
  [SMALL_STATE(2999)] = 96152,
  [SMALL_STATE(3000)] = 96208,
  [SMALL_STATE(3001)] = 96264,
  [SMALL_STATE(3002)] = 96320,
  [SMALL_STATE(3003)] = 96376,
  [SMALL_STATE(3004)] = 96432,
  [SMALL_STATE(3005)] = 96488,
  [SMALL_STATE(3006)] = 96544,
  [SMALL_STATE(3007)] = 96600,
  [SMALL_STATE(3008)] = 96656,
  [SMALL_STATE(3009)] = 96712,
  [SMALL_STATE(3010)] = 96774,
  [SMALL_STATE(3011)] = 96830,
  [SMALL_STATE(3012)] = 96886,
  [SMALL_STATE(3013)] = 96942,
  [SMALL_STATE(3014)] = 96998,
  [SMALL_STATE(3015)] = 97054,
  [SMALL_STATE(3016)] = 97110,
  [SMALL_STATE(3017)] = 97166,
  [SMALL_STATE(3018)] = 97222,
  [SMALL_STATE(3019)] = 97278,
  [SMALL_STATE(3020)] = 97334,
  [SMALL_STATE(3021)] = 97390,
  [SMALL_STATE(3022)] = 97446,
  [SMALL_STATE(3023)] = 97502,
  [SMALL_STATE(3024)] = 97558,
  [SMALL_STATE(3025)] = 97614,
  [SMALL_STATE(3026)] = 97670,
  [SMALL_STATE(3027)] = 97726,
  [SMALL_STATE(3028)] = 97782,
  [SMALL_STATE(3029)] = 97838,
  [SMALL_STATE(3030)] = 97894,
  [SMALL_STATE(3031)] = 97950,
  [SMALL_STATE(3032)] = 98006,
  [SMALL_STATE(3033)] = 98062,
  [SMALL_STATE(3034)] = 98118,
  [SMALL_STATE(3035)] = 98174,
  [SMALL_STATE(3036)] = 98230,
  [SMALL_STATE(3037)] = 98286,
  [SMALL_STATE(3038)] = 98342,
  [SMALL_STATE(3039)] = 98398,
  [SMALL_STATE(3040)] = 98454,
  [SMALL_STATE(3041)] = 98510,
  [SMALL_STATE(3042)] = 98566,
  [SMALL_STATE(3043)] = 98622,
  [SMALL_STATE(3044)] = 98678,
  [SMALL_STATE(3045)] = 98734,
  [SMALL_STATE(3046)] = 98790,
  [SMALL_STATE(3047)] = 98846,
  [SMALL_STATE(3048)] = 98902,
  [SMALL_STATE(3049)] = 98958,
  [SMALL_STATE(3050)] = 99014,
  [SMALL_STATE(3051)] = 99072,
  [SMALL_STATE(3052)] = 99128,
  [SMALL_STATE(3053)] = 99184,
  [SMALL_STATE(3054)] = 99240,
  [SMALL_STATE(3055)] = 99296,
  [SMALL_STATE(3056)] = 99352,
  [SMALL_STATE(3057)] = 99408,
  [SMALL_STATE(3058)] = 99464,
  [SMALL_STATE(3059)] = 99526,
  [SMALL_STATE(3060)] = 99582,
  [SMALL_STATE(3061)] = 99638,
  [SMALL_STATE(3062)] = 99694,
  [SMALL_STATE(3063)] = 99750,
  [SMALL_STATE(3064)] = 99817,
  [SMALL_STATE(3065)] = 99926,
  [SMALL_STATE(3066)] = 100013,
  [SMALL_STATE(3067)] = 100088,
  [SMALL_STATE(3068)] = 100165,
  [SMALL_STATE(3069)] = 100242,
  [SMALL_STATE(3070)] = 100331,
  [SMALL_STATE(3071)] = 100412,
  [SMALL_STATE(3072)] = 100497,
  [SMALL_STATE(3073)] = 100586,
  [SMALL_STATE(3074)] = 100675,
  [SMALL_STATE(3075)] = 100764,
  [SMALL_STATE(3076)] = 100855,
  [SMALL_STATE(3077)] = 100946,
  [SMALL_STATE(3078)] = 101037,
  [SMALL_STATE(3079)] = 101146,
  [SMALL_STATE(3080)] = 101201,
  [SMALL_STATE(3081)] = 101256,
  [SMALL_STATE(3082)] = 101365,
  [SMALL_STATE(3083)] = 101430,
  [SMALL_STATE(3084)] = 101497,
  [SMALL_STATE(3085)] = 101568,
  [SMALL_STATE(3086)] = 101623,
  [SMALL_STATE(3087)] = 101727,
  [SMALL_STATE(3088)] = 101793,
  [SMALL_STATE(3089)] = 101863,
  [SMALL_STATE(3090)] = 101953,
  [SMALL_STATE(3091)] = 102017,
  [SMALL_STATE(3092)] = 102083,
  [SMALL_STATE(3093)] = 102169,
  [SMALL_STATE(3094)] = 102243,
  [SMALL_STATE(3095)] = 102319,
  [SMALL_STATE(3096)] = 102395,
  [SMALL_STATE(3097)] = 102483,
  [SMALL_STATE(3098)] = 102563,
  [SMALL_STATE(3099)] = 102647,
  [SMALL_STATE(3100)] = 102735,
  [SMALL_STATE(3101)] = 102823,
  [SMALL_STATE(3102)] = 102927,
  [SMALL_STATE(3103)] = 103015,
  [SMALL_STATE(3104)] = 103071,
  [SMALL_STATE(3105)] = 103175,
  [SMALL_STATE(3106)] = 103279,
  [SMALL_STATE(3107)] = 103369,
  [SMALL_STATE(3108)] = 103459,
  [SMALL_STATE(3109)] = 103563,
  [SMALL_STATE(3110)] = 103667,
  [SMALL_STATE(3111)] = 103770,
  [SMALL_STATE(3112)] = 103845,
  [SMALL_STATE(3113)] = 103920,
  [SMALL_STATE(3114)] = 104007,
  [SMALL_STATE(3115)] = 104080,
  [SMALL_STATE(3116)] = 104145,
  [SMALL_STATE(3117)] = 104214,
  [SMALL_STATE(3118)] = 104293,
  [SMALL_STATE(3119)] = 104376,
  [SMALL_STATE(3120)] = 104463,
  [SMALL_STATE(3121)] = 104550,
  [SMALL_STATE(3122)] = 104637,
  [SMALL_STATE(3123)] = 104722,
  [SMALL_STATE(3124)] = 104823,
  [SMALL_STATE(3125)] = 104924,
  [SMALL_STATE(3126)] = 104979,
  [SMALL_STATE(3127)] = 105068,
  [SMALL_STATE(3128)] = 105155,
  [SMALL_STATE(3129)] = 105218,
  [SMALL_STATE(3130)] = 105307,
  [SMALL_STATE(3131)] = 105410,
  [SMALL_STATE(3132)] = 105511,
  [SMALL_STATE(3133)] = 105614,
  [SMALL_STATE(3134)] = 105715,
  [SMALL_STATE(3135)] = 105804,
  [SMALL_STATE(3136)] = 105857,
  [SMALL_STATE(3137)] = 105958,
  [SMALL_STATE(3138)] = 106045,
  [SMALL_STATE(3139)] = 106134,
  [SMALL_STATE(3140)] = 106237,
  [SMALL_STATE(3141)] = 106340,
  [SMALL_STATE(3142)] = 106441,
  [SMALL_STATE(3143)] = 106542,
  [SMALL_STATE(3144)] = 106597,
  [SMALL_STATE(3145)] = 106686,
  [SMALL_STATE(3146)] = 106769,
  [SMALL_STATE(3147)] = 106842,
  [SMALL_STATE(3148)] = 106917,
  [SMALL_STATE(3149)] = 106992,
  [SMALL_STATE(3150)] = 107093,
  [SMALL_STATE(3151)] = 107196,
  [SMALL_STATE(3152)] = 107283,
  [SMALL_STATE(3153)] = 107386,
  [SMALL_STATE(3154)] = 107451,
  [SMALL_STATE(3155)] = 107554,
  [SMALL_STATE(3156)] = 107655,
  [SMALL_STATE(3157)] = 107756,
  [SMALL_STATE(3158)] = 107859,
  [SMALL_STATE(3159)] = 107924,
  [SMALL_STATE(3160)] = 108011,
  [SMALL_STATE(3161)] = 108066,
  [SMALL_STATE(3162)] = 108145,
  [SMALL_STATE(3163)] = 108248,
  [SMALL_STATE(3164)] = 108351,
  [SMALL_STATE(3165)] = 108420,
  [SMALL_STATE(3166)] = 108523,
  [SMALL_STATE(3167)] = 108626,
  [SMALL_STATE(3168)] = 108727,
  [SMALL_STATE(3169)] = 108830,
  [SMALL_STATE(3170)] = 108933,
  [SMALL_STATE(3171)] = 109022,
  [SMALL_STATE(3172)] = 109123,
  [SMALL_STATE(3173)] = 109226,
  [SMALL_STATE(3174)] = 109279,
  [SMALL_STATE(3175)] = 109342,
  [SMALL_STATE(3176)] = 109427,
  [SMALL_STATE(3177)] = 109530,
  [SMALL_STATE(3178)] = 109595,
  [SMALL_STATE(3179)] = 109693,
  [SMALL_STATE(3180)] = 109785,
  [SMALL_STATE(3181)] = 109873,
  [SMALL_STATE(3182)] = 109971,
  [SMALL_STATE(3183)] = 110023,
  [SMALL_STATE(3184)] = 110115,
  [SMALL_STATE(3185)] = 110209,
  [SMALL_STATE(3186)] = 110303,
  [SMALL_STATE(3187)] = 110403,
  [SMALL_STATE(3188)] = 110501,
  [SMALL_STATE(3189)] = 110595,
  [SMALL_STATE(3190)] = 110689,
  [SMALL_STATE(3191)] = 110779,
  [SMALL_STATE(3192)] = 110831,
  [SMALL_STATE(3193)] = 110929,
  [SMALL_STATE(3194)] = 111027,
  [SMALL_STATE(3195)] = 111121,
  [SMALL_STATE(3196)] = 111215,
  [SMALL_STATE(3197)] = 111313,
  [SMALL_STATE(3198)] = 111407,
  [SMALL_STATE(3199)] = 111459,
  [SMALL_STATE(3200)] = 111553,
  [SMALL_STATE(3201)] = 111645,
  [SMALL_STATE(3202)] = 111697,
  [SMALL_STATE(3203)] = 111791,
  [SMALL_STATE(3204)] = 111879,
  [SMALL_STATE(3205)] = 111973,
  [SMALL_STATE(3206)] = 112051,
  [SMALL_STATE(3207)] = 112139,
  [SMALL_STATE(3208)] = 112237,
  [SMALL_STATE(3209)] = 112335,
  [SMALL_STATE(3210)] = 112429,
  [SMALL_STATE(3211)] = 112527,
  [SMALL_STATE(3212)] = 112621,
  [SMALL_STATE(3213)] = 112713,
  [SMALL_STATE(3214)] = 112811,
  [SMALL_STATE(3215)] = 112866,
  [SMALL_STATE(3216)] = 112957,
  [SMALL_STATE(3217)] = 113012,
  [SMALL_STATE(3218)] = 113065,
  [SMALL_STATE(3219)] = 113156,
  [SMALL_STATE(3220)] = 113247,
  [SMALL_STATE(3221)] = 113300,
  [SMALL_STATE(3222)] = 113391,
  [SMALL_STATE(3223)] = 113446,
  [SMALL_STATE(3224)] = 113501,
  [SMALL_STATE(3225)] = 113556,
  [SMALL_STATE(3226)] = 113607,
  [SMALL_STATE(3227)] = 113660,
  [SMALL_STATE(3228)] = 113715,
  [SMALL_STATE(3229)] = 113804,
  [SMALL_STATE(3230)] = 113854,
  [SMALL_STATE(3231)] = 113904,
  [SMALL_STATE(3232)] = 113990,
  [SMALL_STATE(3233)] = 114044,
  [SMALL_STATE(3234)] = 114098,
  [SMALL_STATE(3235)] = 114184,
  [SMALL_STATE(3236)] = 114234,
  [SMALL_STATE(3237)] = 114294,
  [SMALL_STATE(3238)] = 114344,
  [SMALL_STATE(3239)] = 114396,
  [SMALL_STATE(3240)] = 114446,
  [SMALL_STATE(3241)] = 114496,
  [SMALL_STATE(3242)] = 114556,
  [SMALL_STATE(3243)] = 114606,
  [SMALL_STATE(3244)] = 114656,
  [SMALL_STATE(3245)] = 114706,
  [SMALL_STATE(3246)] = 114756,
  [SMALL_STATE(3247)] = 114806,
  [SMALL_STATE(3248)] = 114860,
  [SMALL_STATE(3249)] = 114910,
  [SMALL_STATE(3250)] = 114962,
  [SMALL_STATE(3251)] = 115016,
  [SMALL_STATE(3252)] = 115066,
  [SMALL_STATE(3253)] = 115120,
  [SMALL_STATE(3254)] = 115206,
  [SMALL_STATE(3255)] = 115260,
  [SMALL_STATE(3256)] = 115309,
  [SMALL_STATE(3257)] = 115358,
  [SMALL_STATE(3258)] = 115407,
  [SMALL_STATE(3259)] = 115456,
  [SMALL_STATE(3260)] = 115505,
  [SMALL_STATE(3261)] = 115554,
  [SMALL_STATE(3262)] = 115603,
  [SMALL_STATE(3263)] = 115652,
  [SMALL_STATE(3264)] = 115701,
  [SMALL_STATE(3265)] = 115750,
  [SMALL_STATE(3266)] = 115799,
  [SMALL_STATE(3267)] = 115848,
  [SMALL_STATE(3268)] = 115937,
  [SMALL_STATE(3269)] = 115986,
  [SMALL_STATE(3270)] = 116035,
  [SMALL_STATE(3271)] = 116084,
  [SMALL_STATE(3272)] = 116173,
  [SMALL_STATE(3273)] = 116262,
  [SMALL_STATE(3274)] = 116311,
  [SMALL_STATE(3275)] = 116360,
  [SMALL_STATE(3276)] = 116413,
  [SMALL_STATE(3277)] = 116462,
  [SMALL_STATE(3278)] = 116515,
  [SMALL_STATE(3279)] = 116564,
  [SMALL_STATE(3280)] = 116613,
  [SMALL_STATE(3281)] = 116662,
  [SMALL_STATE(3282)] = 116711,
  [SMALL_STATE(3283)] = 116760,
  [SMALL_STATE(3284)] = 116813,
  [SMALL_STATE(3285)] = 116862,
  [SMALL_STATE(3286)] = 116911,
  [SMALL_STATE(3287)] = 116960,
  [SMALL_STATE(3288)] = 117009,
  [SMALL_STATE(3289)] = 117058,
  [SMALL_STATE(3290)] = 117107,
  [SMALL_STATE(3291)] = 117156,
  [SMALL_STATE(3292)] = 117205,
  [SMALL_STATE(3293)] = 117258,
  [SMALL_STATE(3294)] = 117307,
  [SMALL_STATE(3295)] = 117396,
  [SMALL_STATE(3296)] = 117445,
  [SMALL_STATE(3297)] = 117494,
  [SMALL_STATE(3298)] = 117543,
  [SMALL_STATE(3299)] = 117594,
  [SMALL_STATE(3300)] = 117683,
  [SMALL_STATE(3301)] = 117772,
  [SMALL_STATE(3302)] = 117861,
  [SMALL_STATE(3303)] = 117910,
  [SMALL_STATE(3304)] = 117959,
  [SMALL_STATE(3305)] = 118008,
  [SMALL_STATE(3306)] = 118057,
  [SMALL_STATE(3307)] = 118108,
  [SMALL_STATE(3308)] = 118197,
  [SMALL_STATE(3309)] = 118246,
  [SMALL_STATE(3310)] = 118335,
  [SMALL_STATE(3311)] = 118384,
  [SMALL_STATE(3312)] = 118433,
  [SMALL_STATE(3313)] = 118482,
  [SMALL_STATE(3314)] = 118537,
  [SMALL_STATE(3315)] = 118586,
  [SMALL_STATE(3316)] = 118635,
  [SMALL_STATE(3317)] = 118684,
  [SMALL_STATE(3318)] = 118733,
  [SMALL_STATE(3319)] = 118782,
  [SMALL_STATE(3320)] = 118831,
  [SMALL_STATE(3321)] = 118880,
  [SMALL_STATE(3322)] = 118929,
  [SMALL_STATE(3323)] = 118978,
  [SMALL_STATE(3324)] = 119027,
  [SMALL_STATE(3325)] = 119076,
  [SMALL_STATE(3326)] = 119131,
  [SMALL_STATE(3327)] = 119186,
  [SMALL_STATE(3328)] = 119235,
  [SMALL_STATE(3329)] = 119284,
  [SMALL_STATE(3330)] = 119333,
  [SMALL_STATE(3331)] = 119422,
  [SMALL_STATE(3332)] = 119471,
  [SMALL_STATE(3333)] = 119520,
  [SMALL_STATE(3334)] = 119569,
  [SMALL_STATE(3335)] = 119620,
  [SMALL_STATE(3336)] = 119669,
  [SMALL_STATE(3337)] = 119722,
  [SMALL_STATE(3338)] = 119771,
  [SMALL_STATE(3339)] = 119820,
  [SMALL_STATE(3340)] = 119869,
  [SMALL_STATE(3341)] = 119918,
  [SMALL_STATE(3342)] = 119967,
  [SMALL_STATE(3343)] = 120018,
  [SMALL_STATE(3344)] = 120067,
  [SMALL_STATE(3345)] = 120116,
  [SMALL_STATE(3346)] = 120165,
  [SMALL_STATE(3347)] = 120216,
  [SMALL_STATE(3348)] = 120265,
  [SMALL_STATE(3349)] = 120318,
  [SMALL_STATE(3350)] = 120407,
  [SMALL_STATE(3351)] = 120496,
  [SMALL_STATE(3352)] = 120545,
  [SMALL_STATE(3353)] = 120594,
  [SMALL_STATE(3354)] = 120683,
  [SMALL_STATE(3355)] = 120732,
  [SMALL_STATE(3356)] = 120781,
  [SMALL_STATE(3357)] = 120836,
  [SMALL_STATE(3358)] = 120925,
  [SMALL_STATE(3359)] = 120976,
  [SMALL_STATE(3360)] = 121031,
  [SMALL_STATE(3361)] = 121080,
  [SMALL_STATE(3362)] = 121129,
  [SMALL_STATE(3363)] = 121218,
  [SMALL_STATE(3364)] = 121269,
  [SMALL_STATE(3365)] = 121318,
  [SMALL_STATE(3366)] = 121404,
  [SMALL_STATE(3367)] = 121490,
  [SMALL_STATE(3368)] = 121576,
  [SMALL_STATE(3369)] = 121662,
  [SMALL_STATE(3370)] = 121748,
  [SMALL_STATE(3371)] = 121834,
  [SMALL_STATE(3372)] = 121920,
  [SMALL_STATE(3373)] = 122006,
  [SMALL_STATE(3374)] = 122092,
  [SMALL_STATE(3375)] = 122178,
  [SMALL_STATE(3376)] = 122264,
  [SMALL_STATE(3377)] = 122350,
  [SMALL_STATE(3378)] = 122436,
  [SMALL_STATE(3379)] = 122522,
  [SMALL_STATE(3380)] = 122608,
  [SMALL_STATE(3381)] = 122694,
  [SMALL_STATE(3382)] = 122780,
  [SMALL_STATE(3383)] = 122866,
  [SMALL_STATE(3384)] = 122952,
  [SMALL_STATE(3385)] = 123000,
  [SMALL_STATE(3386)] = 123086,
  [SMALL_STATE(3387)] = 123172,
  [SMALL_STATE(3388)] = 123258,
  [SMALL_STATE(3389)] = 123344,
  [SMALL_STATE(3390)] = 123430,
  [SMALL_STATE(3391)] = 123512,
  [SMALL_STATE(3392)] = 123598,
  [SMALL_STATE(3393)] = 123684,
  [SMALL_STATE(3394)] = 123732,
  [SMALL_STATE(3395)] = 123818,
  [SMALL_STATE(3396)] = 123900,
  [SMALL_STATE(3397)] = 123982,
  [SMALL_STATE(3398)] = 124068,
  [SMALL_STATE(3399)] = 124154,
  [SMALL_STATE(3400)] = 124240,
  [SMALL_STATE(3401)] = 124326,
  [SMALL_STATE(3402)] = 124412,
  [SMALL_STATE(3403)] = 124498,
  [SMALL_STATE(3404)] = 124556,
  [SMALL_STATE(3405)] = 124642,
  [SMALL_STATE(3406)] = 124728,
  [SMALL_STATE(3407)] = 124814,
  [SMALL_STATE(3408)] = 124898,
  [SMALL_STATE(3409)] = 124984,
  [SMALL_STATE(3410)] = 125062,
  [SMALL_STATE(3411)] = 125148,
  [SMALL_STATE(3412)] = 125196,
  [SMALL_STATE(3413)] = 125282,
  [SMALL_STATE(3414)] = 125368,
  [SMALL_STATE(3415)] = 125454,
  [SMALL_STATE(3416)] = 125538,
  [SMALL_STATE(3417)] = 125624,
  [SMALL_STATE(3418)] = 125710,
  [SMALL_STATE(3419)] = 125760,
  [SMALL_STATE(3420)] = 125808,
  [SMALL_STATE(3421)] = 125894,
  [SMALL_STATE(3422)] = 125980,
  [SMALL_STATE(3423)] = 126064,
  [SMALL_STATE(3424)] = 126150,
  [SMALL_STATE(3425)] = 126208,
  [SMALL_STATE(3426)] = 126294,
  [SMALL_STATE(3427)] = 126380,
  [SMALL_STATE(3428)] = 126466,
  [SMALL_STATE(3429)] = 126516,
  [SMALL_STATE(3430)] = 126564,
  [SMALL_STATE(3431)] = 126612,
  [SMALL_STATE(3432)] = 126698,
  [SMALL_STATE(3433)] = 126758,
  [SMALL_STATE(3434)] = 126838,
  [SMALL_STATE(3435)] = 126924,
  [SMALL_STATE(3436)] = 126992,
  [SMALL_STATE(3437)] = 127062,
  [SMALL_STATE(3438)] = 127148,
  [SMALL_STATE(3439)] = 127222,
  [SMALL_STATE(3440)] = 127292,
  [SMALL_STATE(3441)] = 127352,
  [SMALL_STATE(3442)] = 127438,
  [SMALL_STATE(3443)] = 127520,
  [SMALL_STATE(3444)] = 127568,
  [SMALL_STATE(3445)] = 127632,
  [SMALL_STATE(3446)] = 127718,
  [SMALL_STATE(3447)] = 127804,
  [SMALL_STATE(3448)] = 127863,
  [SMALL_STATE(3449)] = 127914,
  [SMALL_STATE(3450)] = 127963,
  [SMALL_STATE(3451)] = 128020,
  [SMALL_STATE(3452)] = 128071,
  [SMALL_STATE(3453)] = 128150,
  [SMALL_STATE(3454)] = 128209,
  [SMALL_STATE(3455)] = 128272,
  [SMALL_STATE(3456)] = 128321,
  [SMALL_STATE(3457)] = 128378,
  [SMALL_STATE(3458)] = 128461,
  [SMALL_STATE(3459)] = 128528,
  [SMALL_STATE(3460)] = 128585,
  [SMALL_STATE(3461)] = 128668,
  [SMALL_STATE(3462)] = 128751,
  [SMALL_STATE(3463)] = 128820,
  [SMALL_STATE(3464)] = 128889,
  [SMALL_STATE(3465)] = 128970,
  [SMALL_STATE(3466)] = 129043,
  [SMALL_STATE(3467)] = 129120,
  [SMALL_STATE(3468)] = 129201,
  [SMALL_STATE(3469)] = 129262,
  [SMALL_STATE(3470)] = 129309,
  [SMALL_STATE(3471)] = 129388,
  [SMALL_STATE(3472)] = 129471,
  [SMALL_STATE(3473)] = 129540,
  [SMALL_STATE(3474)] = 129587,
  [SMALL_STATE(3475)] = 129656,
  [SMALL_STATE(3476)] = 129737,
  [SMALL_STATE(3477)] = 129810,
  [SMALL_STATE(3478)] = 129887,
  [SMALL_STATE(3479)] = 129968,
  [SMALL_STATE(3480)] = 130049,
  [SMALL_STATE(3481)] = 130130,
  [SMALL_STATE(3482)] = 130177,
  [SMALL_STATE(3483)] = 130224,
  [SMALL_STATE(3484)] = 130305,
  [SMALL_STATE(3485)] = 130352,
  [SMALL_STATE(3486)] = 130409,
  [SMALL_STATE(3487)] = 130468,
  [SMALL_STATE(3488)] = 130527,
  [SMALL_STATE(3489)] = 130590,
  [SMALL_STATE(3490)] = 130671,
  [SMALL_STATE(3491)] = 130718,
  [SMALL_STATE(3492)] = 130801,
  [SMALL_STATE(3493)] = 130860,
  [SMALL_STATE(3494)] = 130943,
  [SMALL_STATE(3495)] = 131000,
  [SMALL_STATE(3496)] = 131067,
  [SMALL_STATE(3497)] = 131125,
  [SMALL_STATE(3498)] = 131171,
  [SMALL_STATE(3499)] = 131217,
  [SMALL_STATE(3500)] = 131263,
  [SMALL_STATE(3501)] = 131309,
  [SMALL_STATE(3502)] = 131355,
  [SMALL_STATE(3503)] = 131401,
  [SMALL_STATE(3504)] = 131447,
  [SMALL_STATE(3505)] = 131515,
  [SMALL_STATE(3506)] = 131561,
  [SMALL_STATE(3507)] = 131607,
  [SMALL_STATE(3508)] = 131653,
  [SMALL_STATE(3509)] = 131701,
  [SMALL_STATE(3510)] = 131749,
  [SMALL_STATE(3511)] = 131795,
  [SMALL_STATE(3512)] = 131841,
  [SMALL_STATE(3513)] = 131887,
  [SMALL_STATE(3514)] = 131933,
  [SMALL_STATE(3515)] = 131979,
  [SMALL_STATE(3516)] = 132041,
  [SMALL_STATE(3517)] = 132087,
  [SMALL_STATE(3518)] = 132149,
  [SMALL_STATE(3519)] = 132195,
  [SMALL_STATE(3520)] = 132241,
  [SMALL_STATE(3521)] = 132287,
  [SMALL_STATE(3522)] = 132333,
  [SMALL_STATE(3523)] = 132379,
  [SMALL_STATE(3524)] = 132425,
  [SMALL_STATE(3525)] = 132473,
  [SMALL_STATE(3526)] = 132519,
  [SMALL_STATE(3527)] = 132565,
  [SMALL_STATE(3528)] = 132611,
  [SMALL_STATE(3529)] = 132657,
  [SMALL_STATE(3530)] = 132703,
  [SMALL_STATE(3531)] = 132749,
  [SMALL_STATE(3532)] = 132811,
  [SMALL_STATE(3533)] = 132857,
  [SMALL_STATE(3534)] = 132903,
  [SMALL_STATE(3535)] = 132949,
  [SMALL_STATE(3536)] = 133011,
  [SMALL_STATE(3537)] = 133057,
  [SMALL_STATE(3538)] = 133103,
  [SMALL_STATE(3539)] = 133159,
  [SMALL_STATE(3540)] = 133207,
  [SMALL_STATE(3541)] = 133265,
  [SMALL_STATE(3542)] = 133343,
  [SMALL_STATE(3543)] = 133405,
  [SMALL_STATE(3544)] = 133487,
  [SMALL_STATE(3545)] = 133533,
  [SMALL_STATE(3546)] = 133579,
  [SMALL_STATE(3547)] = 133625,
  [SMALL_STATE(3548)] = 133671,
  [SMALL_STATE(3549)] = 133751,
  [SMALL_STATE(3550)] = 133797,
  [SMALL_STATE(3551)] = 133843,
  [SMALL_STATE(3552)] = 133889,
  [SMALL_STATE(3553)] = 133935,
  [SMALL_STATE(3554)] = 133983,
  [SMALL_STATE(3555)] = 134029,
  [SMALL_STATE(3556)] = 134075,
  [SMALL_STATE(3557)] = 134121,
  [SMALL_STATE(3558)] = 134167,
  [SMALL_STATE(3559)] = 134213,
  [SMALL_STATE(3560)] = 134259,
  [SMALL_STATE(3561)] = 134305,
  [SMALL_STATE(3562)] = 134351,
  [SMALL_STATE(3563)] = 134431,
  [SMALL_STATE(3564)] = 134477,
  [SMALL_STATE(3565)] = 134523,
  [SMALL_STATE(3566)] = 134569,
  [SMALL_STATE(3567)] = 134615,
  [SMALL_STATE(3568)] = 134661,
  [SMALL_STATE(3569)] = 134707,
  [SMALL_STATE(3570)] = 134789,
  [SMALL_STATE(3571)] = 134857,
  [SMALL_STATE(3572)] = 134903,
  [SMALL_STATE(3573)] = 134949,
  [SMALL_STATE(3574)] = 134995,
  [SMALL_STATE(3575)] = 135041,
  [SMALL_STATE(3576)] = 135109,
  [SMALL_STATE(3577)] = 135171,
  [SMALL_STATE(3578)] = 135233,
  [SMALL_STATE(3579)] = 135279,
  [SMALL_STATE(3580)] = 135347,
  [SMALL_STATE(3581)] = 135395,
  [SMALL_STATE(3582)] = 135473,
  [SMALL_STATE(3583)] = 135519,
  [SMALL_STATE(3584)] = 135565,
  [SMALL_STATE(3585)] = 135611,
  [SMALL_STATE(3586)] = 135657,
  [SMALL_STATE(3587)] = 135703,
  [SMALL_STATE(3588)] = 135749,
  [SMALL_STATE(3589)] = 135817,
  [SMALL_STATE(3590)] = 135865,
  [SMALL_STATE(3591)] = 135911,
  [SMALL_STATE(3592)] = 135957,
  [SMALL_STATE(3593)] = 136003,
  [SMALL_STATE(3594)] = 136049,
  [SMALL_STATE(3595)] = 136095,
  [SMALL_STATE(3596)] = 136141,
  [SMALL_STATE(3597)] = 136187,
  [SMALL_STATE(3598)] = 136233,
  [SMALL_STATE(3599)] = 136279,
  [SMALL_STATE(3600)] = 136325,
  [SMALL_STATE(3601)] = 136373,
  [SMALL_STATE(3602)] = 136419,
  [SMALL_STATE(3603)] = 136465,
  [SMALL_STATE(3604)] = 136541,
  [SMALL_STATE(3605)] = 136623,
  [SMALL_STATE(3606)] = 136689,
  [SMALL_STATE(3607)] = 136757,
  [SMALL_STATE(3608)] = 136825,
  [SMALL_STATE(3609)] = 136903,
  [SMALL_STATE(3610)] = 136975,
  [SMALL_STATE(3611)] = 137049,
  [SMALL_STATE(3612)] = 137127,
  [SMALL_STATE(3613)] = 137207,
  [SMALL_STATE(3614)] = 137254,
  [SMALL_STATE(3615)] = 137299,
  [SMALL_STATE(3616)] = 137344,
  [SMALL_STATE(3617)] = 137389,
  [SMALL_STATE(3618)] = 137434,
  [SMALL_STATE(3619)] = 137479,
  [SMALL_STATE(3620)] = 137524,
  [SMALL_STATE(3621)] = 137569,
  [SMALL_STATE(3622)] = 137614,
  [SMALL_STATE(3623)] = 137659,
  [SMALL_STATE(3624)] = 137704,
  [SMALL_STATE(3625)] = 137749,
  [SMALL_STATE(3626)] = 137794,
  [SMALL_STATE(3627)] = 137849,
  [SMALL_STATE(3628)] = 137926,
  [SMALL_STATE(3629)] = 137991,
  [SMALL_STATE(3630)] = 138058,
  [SMALL_STATE(3631)] = 138125,
  [SMALL_STATE(3632)] = 138204,
  [SMALL_STATE(3633)] = 138261,
  [SMALL_STATE(3634)] = 138318,
  [SMALL_STATE(3635)] = 138379,
  [SMALL_STATE(3636)] = 138450,
  [SMALL_STATE(3637)] = 138525,
  [SMALL_STATE(3638)] = 138604,
  [SMALL_STATE(3639)] = 138649,
  [SMALL_STATE(3640)] = 138728,
  [SMALL_STATE(3641)] = 138807,
  [SMALL_STATE(3642)] = 138884,
  [SMALL_STATE(3643)] = 138929,
  [SMALL_STATE(3644)] = 139008,
  [SMALL_STATE(3645)] = 139053,
  [SMALL_STATE(3646)] = 139098,
  [SMALL_STATE(3647)] = 139143,
  [SMALL_STATE(3648)] = 139188,
  [SMALL_STATE(3649)] = 139249,
  [SMALL_STATE(3650)] = 139298,
  [SMALL_STATE(3651)] = 139343,
  [SMALL_STATE(3652)] = 139388,
  [SMALL_STATE(3653)] = 139433,
  [SMALL_STATE(3654)] = 139478,
  [SMALL_STATE(3655)] = 139531,
  [SMALL_STATE(3656)] = 139576,
  [SMALL_STATE(3657)] = 139621,
  [SMALL_STATE(3658)] = 139666,
  [SMALL_STATE(3659)] = 139711,
  [SMALL_STATE(3660)] = 139758,
  [SMALL_STATE(3661)] = 139839,
  [SMALL_STATE(3662)] = 139884,
  [SMALL_STATE(3663)] = 139929,
  [SMALL_STATE(3664)] = 139974,
  [SMALL_STATE(3665)] = 140019,
  [SMALL_STATE(3666)] = 140070,
  [SMALL_STATE(3667)] = 140151,
  [SMALL_STATE(3668)] = 140196,
  [SMALL_STATE(3669)] = 140241,
  [SMALL_STATE(3670)] = 140286,
  [SMALL_STATE(3671)] = 140331,
  [SMALL_STATE(3672)] = 140392,
  [SMALL_STATE(3673)] = 140437,
  [SMALL_STATE(3674)] = 140482,
  [SMALL_STATE(3675)] = 140527,
  [SMALL_STATE(3676)] = 140572,
  [SMALL_STATE(3677)] = 140653,
  [SMALL_STATE(3678)] = 140698,
  [SMALL_STATE(3679)] = 140743,
  [SMALL_STATE(3680)] = 140788,
  [SMALL_STATE(3681)] = 140833,
  [SMALL_STATE(3682)] = 140878,
  [SMALL_STATE(3683)] = 140923,
  [SMALL_STATE(3684)] = 140968,
  [SMALL_STATE(3685)] = 141013,
  [SMALL_STATE(3686)] = 141058,
  [SMALL_STATE(3687)] = 141103,
  [SMALL_STATE(3688)] = 141164,
  [SMALL_STATE(3689)] = 141209,
  [SMALL_STATE(3690)] = 141254,
  [SMALL_STATE(3691)] = 141299,
  [SMALL_STATE(3692)] = 141344,
  [SMALL_STATE(3693)] = 141389,
  [SMALL_STATE(3694)] = 141434,
  [SMALL_STATE(3695)] = 141495,
  [SMALL_STATE(3696)] = 141540,
  [SMALL_STATE(3697)] = 141587,
  [SMALL_STATE(3698)] = 141632,
  [SMALL_STATE(3699)] = 141677,
  [SMALL_STATE(3700)] = 141722,
  [SMALL_STATE(3701)] = 141767,
  [SMALL_STATE(3702)] = 141820,
  [SMALL_STATE(3703)] = 141865,
  [SMALL_STATE(3704)] = 141910,
  [SMALL_STATE(3705)] = 141959,
  [SMALL_STATE(3706)] = 142020,
  [SMALL_STATE(3707)] = 142067,
  [SMALL_STATE(3708)] = 142128,
  [SMALL_STATE(3709)] = 142173,
  [SMALL_STATE(3710)] = 142224,
  [SMALL_STATE(3711)] = 142269,
  [SMALL_STATE(3712)] = 142314,
  [SMALL_STATE(3713)] = 142359,
  [SMALL_STATE(3714)] = 142404,
  [SMALL_STATE(3715)] = 142449,
  [SMALL_STATE(3716)] = 142494,
  [SMALL_STATE(3717)] = 142539,
  [SMALL_STATE(3718)] = 142586,
  [SMALL_STATE(3719)] = 142631,
  [SMALL_STATE(3720)] = 142676,
  [SMALL_STATE(3721)] = 142721,
  [SMALL_STATE(3722)] = 142794,
  [SMALL_STATE(3723)] = 142839,
  [SMALL_STATE(3724)] = 142916,
  [SMALL_STATE(3725)] = 142961,
  [SMALL_STATE(3726)] = 143040,
  [SMALL_STATE(3727)] = 143085,
  [SMALL_STATE(3728)] = 143130,
  [SMALL_STATE(3729)] = 143175,
  [SMALL_STATE(3730)] = 143220,
  [SMALL_STATE(3731)] = 143265,
  [SMALL_STATE(3732)] = 143348,
  [SMALL_STATE(3733)] = 143431,
  [SMALL_STATE(3734)] = 143476,
  [SMALL_STATE(3735)] = 143521,
  [SMALL_STATE(3736)] = 143566,
  [SMALL_STATE(3737)] = 143645,
  [SMALL_STATE(3738)] = 143706,
  [SMALL_STATE(3739)] = 143751,
  [SMALL_STATE(3740)] = 143796,
  [SMALL_STATE(3741)] = 143841,
  [SMALL_STATE(3742)] = 143886,
  [SMALL_STATE(3743)] = 143961,
  [SMALL_STATE(3744)] = 144006,
  [SMALL_STATE(3745)] = 144051,
  [SMALL_STATE(3746)] = 144096,
  [SMALL_STATE(3747)] = 144141,
  [SMALL_STATE(3748)] = 144186,
  [SMALL_STATE(3749)] = 144231,
  [SMALL_STATE(3750)] = 144276,
  [SMALL_STATE(3751)] = 144321,
  [SMALL_STATE(3752)] = 144366,
  [SMALL_STATE(3753)] = 144411,
  [SMALL_STATE(3754)] = 144456,
  [SMALL_STATE(3755)] = 144501,
  [SMALL_STATE(3756)] = 144546,
  [SMALL_STATE(3757)] = 144591,
  [SMALL_STATE(3758)] = 144636,
  [SMALL_STATE(3759)] = 144681,
  [SMALL_STATE(3760)] = 144726,
  [SMALL_STATE(3761)] = 144792,
  [SMALL_STATE(3762)] = 144854,
  [SMALL_STATE(3763)] = 144902,
  [SMALL_STATE(3764)] = 144948,
  [SMALL_STATE(3765)] = 144992,
  [SMALL_STATE(3766)] = 145072,
  [SMALL_STATE(3767)] = 145120,
  [SMALL_STATE(3768)] = 145168,
  [SMALL_STATE(3769)] = 145248,
  [SMALL_STATE(3770)] = 145328,
  [SMALL_STATE(3771)] = 145400,
  [SMALL_STATE(3772)] = 145454,
  [SMALL_STATE(3773)] = 145528,
  [SMALL_STATE(3774)] = 145592,
  [SMALL_STATE(3775)] = 145658,
  [SMALL_STATE(3776)] = 145720,
  [SMALL_STATE(3777)] = 145776,
  [SMALL_STATE(3778)] = 145832,
  [SMALL_STATE(3779)] = 145888,
  [SMALL_STATE(3780)] = 145948,
  [SMALL_STATE(3781)] = 146018,
  [SMALL_STATE(3782)] = 146090,
  [SMALL_STATE(3783)] = 146166,
  [SMALL_STATE(3784)] = 146242,
  [SMALL_STATE(3785)] = 146318,
  [SMALL_STATE(3786)] = 146376,
  [SMALL_STATE(3787)] = 146454,
  [SMALL_STATE(3788)] = 146532,
  [SMALL_STATE(3789)] = 146610,
  [SMALL_STATE(3790)] = 146658,
  [SMALL_STATE(3791)] = 146730,
  [SMALL_STATE(3792)] = 146784,
  [SMALL_STATE(3793)] = 146864,
  [SMALL_STATE(3794)] = 146944,
  [SMALL_STATE(3795)] = 146992,
  [SMALL_STATE(3796)] = 147040,
  [SMALL_STATE(3797)] = 147116,
  [SMALL_STATE(3798)] = 147159,
  [SMALL_STATE(3799)] = 147234,
  [SMALL_STATE(3800)] = 147281,
  [SMALL_STATE(3801)] = 147326,
  [SMALL_STATE(3802)] = 147399,
  [SMALL_STATE(3803)] = 147444,
  [SMALL_STATE(3804)] = 147519,
  [SMALL_STATE(3805)] = 147594,
  [SMALL_STATE(3806)] = 147669,
  [SMALL_STATE(3807)] = 147744,
  [SMALL_STATE(3808)] = 147819,
  [SMALL_STATE(3809)] = 147894,
  [SMALL_STATE(3810)] = 147969,
  [SMALL_STATE(3811)] = 148014,
  [SMALL_STATE(3812)] = 148061,
  [SMALL_STATE(3813)] = 148136,
  [SMALL_STATE(3814)] = 148211,
  [SMALL_STATE(3815)] = 148262,
  [SMALL_STATE(3816)] = 148323,
  [SMALL_STATE(3817)] = 148400,
  [SMALL_STATE(3818)] = 148447,
  [SMALL_STATE(3819)] = 148522,
  [SMALL_STATE(3820)] = 148599,
  [SMALL_STATE(3821)] = 148676,
  [SMALL_STATE(3822)] = 148721,
  [SMALL_STATE(3823)] = 148772,
  [SMALL_STATE(3824)] = 148819,
  [SMALL_STATE(3825)] = 148864,
  [SMALL_STATE(3826)] = 148939,
  [SMALL_STATE(3827)] = 149014,
  [SMALL_STATE(3828)] = 149089,
  [SMALL_STATE(3829)] = 149166,
  [SMALL_STATE(3830)] = 149241,
  [SMALL_STATE(3831)] = 149286,
  [SMALL_STATE(3832)] = 149333,
  [SMALL_STATE(3833)] = 149386,
  [SMALL_STATE(3834)] = 149459,
  [SMALL_STATE(3835)] = 149504,
  [SMALL_STATE(3836)] = 149559,
  [SMALL_STATE(3837)] = 149634,
  [SMALL_STATE(3838)] = 149681,
  [SMALL_STATE(3839)] = 149742,
  [SMALL_STATE(3840)] = 149815,
  [SMALL_STATE(3841)] = 149858,
  [SMALL_STATE(3842)] = 149933,
  [SMALL_STATE(3843)] = 149976,
  [SMALL_STATE(3844)] = 150023,
  [SMALL_STATE(3845)] = 150098,
  [SMALL_STATE(3846)] = 150143,
  [SMALL_STATE(3847)] = 150218,
  [SMALL_STATE(3848)] = 150271,
  [SMALL_STATE(3849)] = 150318,
  [SMALL_STATE(3850)] = 150361,
  [SMALL_STATE(3851)] = 150436,
  [SMALL_STATE(3852)] = 150513,
  [SMALL_STATE(3853)] = 150570,
  [SMALL_STATE(3854)] = 150621,
  [SMALL_STATE(3855)] = 150668,
  [SMALL_STATE(3856)] = 150745,
  [SMALL_STATE(3857)] = 150791,
  [SMALL_STATE(3858)] = 150833,
  [SMALL_STATE(3859)] = 150875,
  [SMALL_STATE(3860)] = 150933,
  [SMALL_STATE(3861)] = 151005,
  [SMALL_STATE(3862)] = 151047,
  [SMALL_STATE(3863)] = 151119,
  [SMALL_STATE(3864)] = 151167,
  [SMALL_STATE(3865)] = 151239,
  [SMALL_STATE(3866)] = 151311,
  [SMALL_STATE(3867)] = 151359,
  [SMALL_STATE(3868)] = 151431,
  [SMALL_STATE(3869)] = 151473,
  [SMALL_STATE(3870)] = 151515,
  [SMALL_STATE(3871)] = 151563,
  [SMALL_STATE(3872)] = 151607,
  [SMALL_STATE(3873)] = 151651,
  [SMALL_STATE(3874)] = 151723,
  [SMALL_STATE(3875)] = 151795,
  [SMALL_STATE(3876)] = 151841,
  [SMALL_STATE(3877)] = 151883,
  [SMALL_STATE(3878)] = 151927,
  [SMALL_STATE(3879)] = 151999,
  [SMALL_STATE(3880)] = 152041,
  [SMALL_STATE(3881)] = 152083,
  [SMALL_STATE(3882)] = 152155,
  [SMALL_STATE(3883)] = 152197,
  [SMALL_STATE(3884)] = 152269,
  [SMALL_STATE(3885)] = 152311,
  [SMALL_STATE(3886)] = 152353,
  [SMALL_STATE(3887)] = 152399,
  [SMALL_STATE(3888)] = 152441,
  [SMALL_STATE(3889)] = 152483,
  [SMALL_STATE(3890)] = 152525,
  [SMALL_STATE(3891)] = 152567,
  [SMALL_STATE(3892)] = 152609,
  [SMALL_STATE(3893)] = 152651,
  [SMALL_STATE(3894)] = 152699,
  [SMALL_STATE(3895)] = 152741,
  [SMALL_STATE(3896)] = 152799,
  [SMALL_STATE(3897)] = 152841,
  [SMALL_STATE(3898)] = 152883,
  [SMALL_STATE(3899)] = 152925,
  [SMALL_STATE(3900)] = 152969,
  [SMALL_STATE(3901)] = 153041,
  [SMALL_STATE(3902)] = 153083,
  [SMALL_STATE(3903)] = 153125,
  [SMALL_STATE(3904)] = 153177,
  [SMALL_STATE(3905)] = 153219,
  [SMALL_STATE(3906)] = 153261,
  [SMALL_STATE(3907)] = 153313,
  [SMALL_STATE(3908)] = 153355,
  [SMALL_STATE(3909)] = 153397,
  [SMALL_STATE(3910)] = 153465,
  [SMALL_STATE(3911)] = 153537,
  [SMALL_STATE(3912)] = 153579,
  [SMALL_STATE(3913)] = 153651,
  [SMALL_STATE(3914)] = 153693,
  [SMALL_STATE(3915)] = 153735,
  [SMALL_STATE(3916)] = 153807,
  [SMALL_STATE(3917)] = 153849,
  [SMALL_STATE(3918)] = 153891,
  [SMALL_STATE(3919)] = 153933,
  [SMALL_STATE(3920)] = 153975,
  [SMALL_STATE(3921)] = 154047,
  [SMALL_STATE(3922)] = 154089,
  [SMALL_STATE(3923)] = 154133,
  [SMALL_STATE(3924)] = 154175,
  [SMALL_STATE(3925)] = 154217,
  [SMALL_STATE(3926)] = 154259,
  [SMALL_STATE(3927)] = 154301,
  [SMALL_STATE(3928)] = 154343,
  [SMALL_STATE(3929)] = 154385,
  [SMALL_STATE(3930)] = 154429,
  [SMALL_STATE(3931)] = 154471,
  [SMALL_STATE(3932)] = 154517,
  [SMALL_STATE(3933)] = 154559,
  [SMALL_STATE(3934)] = 154601,
  [SMALL_STATE(3935)] = 154658,
  [SMALL_STATE(3936)] = 154703,
  [SMALL_STATE(3937)] = 154744,
  [SMALL_STATE(3938)] = 154789,
  [SMALL_STATE(3939)] = 154830,
  [SMALL_STATE(3940)] = 154875,
  [SMALL_STATE(3941)] = 154916,
  [SMALL_STATE(3942)] = 154961,
  [SMALL_STATE(3943)] = 155004,
  [SMALL_STATE(3944)] = 155045,
  [SMALL_STATE(3945)] = 155086,
  [SMALL_STATE(3946)] = 155141,
  [SMALL_STATE(3947)] = 155196,
  [SMALL_STATE(3948)] = 155237,
  [SMALL_STATE(3949)] = 155280,
  [SMALL_STATE(3950)] = 155325,
  [SMALL_STATE(3951)] = 155370,
  [SMALL_STATE(3952)] = 155411,
  [SMALL_STATE(3953)] = 155456,
  [SMALL_STATE(3954)] = 155499,
  [SMALL_STATE(3955)] = 155540,
  [SMALL_STATE(3956)] = 155611,
  [SMALL_STATE(3957)] = 155652,
  [SMALL_STATE(3958)] = 155693,
  [SMALL_STATE(3959)] = 155734,
  [SMALL_STATE(3960)] = 155779,
  [SMALL_STATE(3961)] = 155836,
  [SMALL_STATE(3962)] = 155881,
  [SMALL_STATE(3963)] = 155950,
  [SMALL_STATE(3964)] = 155993,
  [SMALL_STATE(3965)] = 156034,
  [SMALL_STATE(3966)] = 156075,
  [SMALL_STATE(3967)] = 156130,
  [SMALL_STATE(3968)] = 156171,
  [SMALL_STATE(3969)] = 156212,
  [SMALL_STATE(3970)] = 156267,
  [SMALL_STATE(3971)] = 156335,
  [SMALL_STATE(3972)] = 156403,
  [SMALL_STATE(3973)] = 156445,
  [SMALL_STATE(3974)] = 156497,
  [SMALL_STATE(3975)] = 156537,
  [SMALL_STATE(3976)] = 156577,
  [SMALL_STATE(3977)] = 156629,
  [SMALL_STATE(3978)] = 156697,
  [SMALL_STATE(3979)] = 156765,
  [SMALL_STATE(3980)] = 156817,
  [SMALL_STATE(3981)] = 156885,
  [SMALL_STATE(3982)] = 156953,
  [SMALL_STATE(3983)] = 157021,
  [SMALL_STATE(3984)] = 157089,
  [SMALL_STATE(3985)] = 157135,
  [SMALL_STATE(3986)] = 157187,
  [SMALL_STATE(3987)] = 157255,
  [SMALL_STATE(3988)] = 157323,
  [SMALL_STATE(3989)] = 157363,
  [SMALL_STATE(3990)] = 157431,
  [SMALL_STATE(3991)] = 157499,
  [SMALL_STATE(3992)] = 157543,
  [SMALL_STATE(3993)] = 157611,
  [SMALL_STATE(3994)] = 157679,
  [SMALL_STATE(3995)] = 157747,
  [SMALL_STATE(3996)] = 157815,
  [SMALL_STATE(3997)] = 157883,
  [SMALL_STATE(3998)] = 157951,
  [SMALL_STATE(3999)] = 158019,
  [SMALL_STATE(4000)] = 158087,
  [SMALL_STATE(4001)] = 158155,
  [SMALL_STATE(4002)] = 158221,
  [SMALL_STATE(4003)] = 158289,
  [SMALL_STATE(4004)] = 158329,
  [SMALL_STATE(4005)] = 158397,
  [SMALL_STATE(4006)] = 158449,
  [SMALL_STATE(4007)] = 158515,
  [SMALL_STATE(4008)] = 158555,
  [SMALL_STATE(4009)] = 158621,
  [SMALL_STATE(4010)] = 158689,
  [SMALL_STATE(4011)] = 158755,
  [SMALL_STATE(4012)] = 158823,
  [SMALL_STATE(4013)] = 158891,
  [SMALL_STATE(4014)] = 158959,
  [SMALL_STATE(4015)] = 159027,
  [SMALL_STATE(4016)] = 159095,
  [SMALL_STATE(4017)] = 159163,
  [SMALL_STATE(4018)] = 159231,
  [SMALL_STATE(4019)] = 159299,
  [SMALL_STATE(4020)] = 159365,
  [SMALL_STATE(4021)] = 159431,
  [SMALL_STATE(4022)] = 159499,
  [SMALL_STATE(4023)] = 159567,
  [SMALL_STATE(4024)] = 159635,
  [SMALL_STATE(4025)] = 159703,
  [SMALL_STATE(4026)] = 159771,
  [SMALL_STATE(4027)] = 159839,
  [SMALL_STATE(4028)] = 159907,
  [SMALL_STATE(4029)] = 159975,
  [SMALL_STATE(4030)] = 160043,
  [SMALL_STATE(4031)] = 160111,
  [SMALL_STATE(4032)] = 160179,
  [SMALL_STATE(4033)] = 160245,
  [SMALL_STATE(4034)] = 160313,
  [SMALL_STATE(4035)] = 160381,
  [SMALL_STATE(4036)] = 160449,
  [SMALL_STATE(4037)] = 160517,
  [SMALL_STATE(4038)] = 160585,
  [SMALL_STATE(4039)] = 160653,
  [SMALL_STATE(4040)] = 160707,
  [SMALL_STATE(4041)] = 160775,
  [SMALL_STATE(4042)] = 160815,
  [SMALL_STATE(4043)] = 160861,
  [SMALL_STATE(4044)] = 160929,
  [SMALL_STATE(4045)] = 160997,
  [SMALL_STATE(4046)] = 161037,
  [SMALL_STATE(4047)] = 161089,
  [SMALL_STATE(4048)] = 161141,
  [SMALL_STATE(4049)] = 161193,
  [SMALL_STATE(4050)] = 161245,
  [SMALL_STATE(4051)] = 161311,
  [SMALL_STATE(4052)] = 161351,
  [SMALL_STATE(4053)] = 161395,
  [SMALL_STATE(4054)] = 161463,
  [SMALL_STATE(4055)] = 161531,
  [SMALL_STATE(4056)] = 161599,
  [SMALL_STATE(4057)] = 161639,
  [SMALL_STATE(4058)] = 161707,
  [SMALL_STATE(4059)] = 161763,
  [SMALL_STATE(4060)] = 161831,
  [SMALL_STATE(4061)] = 161899,
  [SMALL_STATE(4062)] = 161967,
  [SMALL_STATE(4063)] = 162035,
  [SMALL_STATE(4064)] = 162101,
  [SMALL_STATE(4065)] = 162169,
  [SMALL_STATE(4066)] = 162209,
  [SMALL_STATE(4067)] = 162275,
  [SMALL_STATE(4068)] = 162327,
  [SMALL_STATE(4069)] = 162367,
  [SMALL_STATE(4070)] = 162407,
  [SMALL_STATE(4071)] = 162447,
  [SMALL_STATE(4072)] = 162499,
  [SMALL_STATE(4073)] = 162539,
  [SMALL_STATE(4074)] = 162607,
  [SMALL_STATE(4075)] = 162675,
  [SMALL_STATE(4076)] = 162741,
  [SMALL_STATE(4077)] = 162809,
  [SMALL_STATE(4078)] = 162877,
  [SMALL_STATE(4079)] = 162917,
  [SMALL_STATE(4080)] = 162985,
  [SMALL_STATE(4081)] = 163029,
  [SMALL_STATE(4082)] = 163069,
  [SMALL_STATE(4083)] = 163137,
  [SMALL_STATE(4084)] = 163205,
  [SMALL_STATE(4085)] = 163271,
  [SMALL_STATE(4086)] = 163339,
  [SMALL_STATE(4087)] = 163407,
  [SMALL_STATE(4088)] = 163459,
  [SMALL_STATE(4089)] = 163527,
  [SMALL_STATE(4090)] = 163593,
  [SMALL_STATE(4091)] = 163645,
  [SMALL_STATE(4092)] = 163697,
  [SMALL_STATE(4093)] = 163749,
  [SMALL_STATE(4094)] = 163817,
  [SMALL_STATE(4095)] = 163885,
  [SMALL_STATE(4096)] = 163953,
  [SMALL_STATE(4097)] = 163995,
  [SMALL_STATE(4098)] = 164063,
  [SMALL_STATE(4099)] = 164115,
  [SMALL_STATE(4100)] = 164183,
  [SMALL_STATE(4101)] = 164251,
  [SMALL_STATE(4102)] = 164319,
  [SMALL_STATE(4103)] = 164387,
  [SMALL_STATE(4104)] = 164455,
  [SMALL_STATE(4105)] = 164523,
  [SMALL_STATE(4106)] = 164591,
  [SMALL_STATE(4107)] = 164659,
  [SMALL_STATE(4108)] = 164727,
  [SMALL_STATE(4109)] = 164795,
  [SMALL_STATE(4110)] = 164861,
  [SMALL_STATE(4111)] = 164901,
  [SMALL_STATE(4112)] = 164941,
  [SMALL_STATE(4113)] = 165009,
  [SMALL_STATE(4114)] = 165061,
  [SMALL_STATE(4115)] = 165129,
  [SMALL_STATE(4116)] = 165169,
  [SMALL_STATE(4117)] = 165237,
  [SMALL_STATE(4118)] = 165305,
  [SMALL_STATE(4119)] = 165373,
  [SMALL_STATE(4120)] = 165441,
  [SMALL_STATE(4121)] = 165509,
  [SMALL_STATE(4122)] = 165577,
  [SMALL_STATE(4123)] = 165645,
  [SMALL_STATE(4124)] = 165697,
  [SMALL_STATE(4125)] = 165741,
  [SMALL_STATE(4126)] = 165809,
  [SMALL_STATE(4127)] = 165877,
  [SMALL_STATE(4128)] = 165929,
  [SMALL_STATE(4129)] = 165985,
  [SMALL_STATE(4130)] = 166053,
  [SMALL_STATE(4131)] = 166121,
  [SMALL_STATE(4132)] = 166170,
  [SMALL_STATE(4133)] = 166215,
  [SMALL_STATE(4134)] = 166258,
  [SMALL_STATE(4135)] = 166301,
  [SMALL_STATE(4136)] = 166344,
  [SMALL_STATE(4137)] = 166391,
  [SMALL_STATE(4138)] = 166438,
  [SMALL_STATE(4139)] = 166485,
  [SMALL_STATE(4140)] = 166528,
  [SMALL_STATE(4141)] = 166573,
  [SMALL_STATE(4142)] = 166618,
  [SMALL_STATE(4143)] = 166659,
  [SMALL_STATE(4144)] = 166714,
  [SMALL_STATE(4145)] = 166755,
  [SMALL_STATE(4146)] = 166800,
  [SMALL_STATE(4147)] = 166845,
  [SMALL_STATE(4148)] = 166890,
  [SMALL_STATE(4149)] = 166933,
  [SMALL_STATE(4150)] = 166982,
  [SMALL_STATE(4151)] = 167027,
  [SMALL_STATE(4152)] = 167072,
  [SMALL_STATE(4153)] = 167115,
  [SMALL_STATE(4154)] = 167160,
  [SMALL_STATE(4155)] = 167205,
  [SMALL_STATE(4156)] = 167250,
  [SMALL_STATE(4157)] = 167295,
  [SMALL_STATE(4158)] = 167340,
  [SMALL_STATE(4159)] = 167383,
  [SMALL_STATE(4160)] = 167422,
  [SMALL_STATE(4161)] = 167461,
  [SMALL_STATE(4162)] = 167510,
  [SMALL_STATE(4163)] = 167549,
  [SMALL_STATE(4164)] = 167588,
  [SMALL_STATE(4165)] = 167633,
  [SMALL_STATE(4166)] = 167688,
  [SMALL_STATE(4167)] = 167731,
  [SMALL_STATE(4168)] = 167776,
  [SMALL_STATE(4169)] = 167821,
  [SMALL_STATE(4170)] = 167866,
  [SMALL_STATE(4171)] = 167911,
  [SMALL_STATE(4172)] = 167954,
  [SMALL_STATE(4173)] = 168006,
  [SMALL_STATE(4174)] = 168072,
  [SMALL_STATE(4175)] = 168120,
  [SMALL_STATE(4176)] = 168160,
  [SMALL_STATE(4177)] = 168202,
  [SMALL_STATE(4178)] = 168268,
  [SMALL_STATE(4179)] = 168334,
  [SMALL_STATE(4180)] = 168400,
  [SMALL_STATE(4181)] = 168438,
  [SMALL_STATE(4182)] = 168476,
  [SMALL_STATE(4183)] = 168520,
  [SMALL_STATE(4184)] = 168562,
  [SMALL_STATE(4185)] = 168606,
  [SMALL_STATE(4186)] = 168650,
  [SMALL_STATE(4187)] = 168690,
  [SMALL_STATE(4188)] = 168738,
  [SMALL_STATE(4189)] = 168776,
  [SMALL_STATE(4190)] = 168824,
  [SMALL_STATE(4191)] = 168862,
  [SMALL_STATE(4192)] = 168928,
  [SMALL_STATE(4193)] = 168976,
  [SMALL_STATE(4194)] = 169018,
  [SMALL_STATE(4195)] = 169056,
  [SMALL_STATE(4196)] = 169096,
  [SMALL_STATE(4197)] = 169134,
  [SMALL_STATE(4198)] = 169176,
  [SMALL_STATE(4199)] = 169222,
  [SMALL_STATE(4200)] = 169288,
  [SMALL_STATE(4201)] = 169340,
  [SMALL_STATE(4202)] = 169386,
  [SMALL_STATE(4203)] = 169428,
  [SMALL_STATE(4204)] = 169466,
  [SMALL_STATE(4205)] = 169504,
  [SMALL_STATE(4206)] = 169547,
  [SMALL_STATE(4207)] = 169592,
  [SMALL_STATE(4208)] = 169635,
  [SMALL_STATE(4209)] = 169678,
  [SMALL_STATE(4210)] = 169741,
  [SMALL_STATE(4211)] = 169786,
  [SMALL_STATE(4212)] = 169829,
  [SMALL_STATE(4213)] = 169872,
  [SMALL_STATE(4214)] = 169915,
  [SMALL_STATE(4215)] = 169978,
  [SMALL_STATE(4216)] = 170015,
  [SMALL_STATE(4217)] = 170058,
  [SMALL_STATE(4218)] = 170095,
  [SMALL_STATE(4219)] = 170138,
  [SMALL_STATE(4220)] = 170181,
  [SMALL_STATE(4221)] = 170224,
  [SMALL_STATE(4222)] = 170263,
  [SMALL_STATE(4223)] = 170300,
  [SMALL_STATE(4224)] = 170343,
  [SMALL_STATE(4225)] = 170386,
  [SMALL_STATE(4226)] = 170429,
  [SMALL_STATE(4227)] = 170472,
  [SMALL_STATE(4228)] = 170515,
  [SMALL_STATE(4229)] = 170560,
  [SMALL_STATE(4230)] = 170597,
  [SMALL_STATE(4231)] = 170640,
  [SMALL_STATE(4232)] = 170683,
  [SMALL_STATE(4233)] = 170746,
  [SMALL_STATE(4234)] = 170809,
  [SMALL_STATE(4235)] = 170852,
  [SMALL_STATE(4236)] = 170915,
  [SMALL_STATE(4237)] = 170958,
  [SMALL_STATE(4238)] = 171001,
  [SMALL_STATE(4239)] = 171044,
  [SMALL_STATE(4240)] = 171107,
  [SMALL_STATE(4241)] = 171152,
  [SMALL_STATE(4242)] = 171195,
  [SMALL_STATE(4243)] = 171232,
  [SMALL_STATE(4244)] = 171275,
  [SMALL_STATE(4245)] = 171338,
  [SMALL_STATE(4246)] = 171381,
  [SMALL_STATE(4247)] = 171424,
  [SMALL_STATE(4248)] = 171467,
  [SMALL_STATE(4249)] = 171504,
  [SMALL_STATE(4250)] = 171547,
  [SMALL_STATE(4251)] = 171590,
  [SMALL_STATE(4252)] = 171633,
  [SMALL_STATE(4253)] = 171676,
  [SMALL_STATE(4254)] = 171721,
  [SMALL_STATE(4255)] = 171764,
  [SMALL_STATE(4256)] = 171809,
  [SMALL_STATE(4257)] = 171852,
  [SMALL_STATE(4258)] = 171895,
  [SMALL_STATE(4259)] = 171938,
  [SMALL_STATE(4260)] = 171981,
  [SMALL_STATE(4261)] = 172024,
  [SMALL_STATE(4262)] = 172067,
  [SMALL_STATE(4263)] = 172110,
  [SMALL_STATE(4264)] = 172153,
  [SMALL_STATE(4265)] = 172196,
  [SMALL_STATE(4266)] = 172239,
  [SMALL_STATE(4267)] = 172282,
  [SMALL_STATE(4268)] = 172325,
  [SMALL_STATE(4269)] = 172368,
  [SMALL_STATE(4270)] = 172413,
  [SMALL_STATE(4271)] = 172476,
  [SMALL_STATE(4272)] = 172521,
  [SMALL_STATE(4273)] = 172564,
  [SMALL_STATE(4274)] = 172613,
  [SMALL_STATE(4275)] = 172662,
  [SMALL_STATE(4276)] = 172705,
  [SMALL_STATE(4277)] = 172748,
  [SMALL_STATE(4278)] = 172791,
  [SMALL_STATE(4279)] = 172834,
  [SMALL_STATE(4280)] = 172879,
  [SMALL_STATE(4281)] = 172922,
  [SMALL_STATE(4282)] = 172959,
  [SMALL_STATE(4283)] = 173008,
  [SMALL_STATE(4284)] = 173071,
  [SMALL_STATE(4285)] = 173108,
  [SMALL_STATE(4286)] = 173151,
  [SMALL_STATE(4287)] = 173214,
  [SMALL_STATE(4288)] = 173257,
  [SMALL_STATE(4289)] = 173300,
  [SMALL_STATE(4290)] = 173343,
  [SMALL_STATE(4291)] = 173386,
  [SMALL_STATE(4292)] = 173429,
  [SMALL_STATE(4293)] = 173472,
  [SMALL_STATE(4294)] = 173515,
  [SMALL_STATE(4295)] = 173558,
  [SMALL_STATE(4296)] = 173601,
  [SMALL_STATE(4297)] = 173644,
  [SMALL_STATE(4298)] = 173687,
  [SMALL_STATE(4299)] = 173730,
  [SMALL_STATE(4300)] = 173773,
  [SMALL_STATE(4301)] = 173816,
  [SMALL_STATE(4302)] = 173859,
  [SMALL_STATE(4303)] = 173902,
  [SMALL_STATE(4304)] = 173965,
  [SMALL_STATE(4305)] = 174008,
  [SMALL_STATE(4306)] = 174051,
  [SMALL_STATE(4307)] = 174094,
  [SMALL_STATE(4308)] = 174131,
  [SMALL_STATE(4309)] = 174172,
  [SMALL_STATE(4310)] = 174215,
  [SMALL_STATE(4311)] = 174264,
  [SMALL_STATE(4312)] = 174307,
  [SMALL_STATE(4313)] = 174350,
  [SMALL_STATE(4314)] = 174393,
  [SMALL_STATE(4315)] = 174436,
  [SMALL_STATE(4316)] = 174499,
  [SMALL_STATE(4317)] = 174542,
  [SMALL_STATE(4318)] = 174585,
  [SMALL_STATE(4319)] = 174628,
  [SMALL_STATE(4320)] = 174671,
  [SMALL_STATE(4321)] = 174712,
  [SMALL_STATE(4322)] = 174755,
  [SMALL_STATE(4323)] = 174798,
  [SMALL_STATE(4324)] = 174841,
  [SMALL_STATE(4325)] = 174884,
  [SMALL_STATE(4326)] = 174927,
  [SMALL_STATE(4327)] = 174970,
  [SMALL_STATE(4328)] = 175007,
  [SMALL_STATE(4329)] = 175050,
  [SMALL_STATE(4330)] = 175093,
  [SMALL_STATE(4331)] = 175156,
  [SMALL_STATE(4332)] = 175219,
  [SMALL_STATE(4333)] = 175262,
  [SMALL_STATE(4334)] = 175307,
  [SMALL_STATE(4335)] = 175350,
  [SMALL_STATE(4336)] = 175393,
  [SMALL_STATE(4337)] = 175430,
  [SMALL_STATE(4338)] = 175467,
  [SMALL_STATE(4339)] = 175510,
  [SMALL_STATE(4340)] = 175553,
  [SMALL_STATE(4341)] = 175596,
  [SMALL_STATE(4342)] = 175639,
  [SMALL_STATE(4343)] = 175682,
  [SMALL_STATE(4344)] = 175725,
  [SMALL_STATE(4345)] = 175768,
  [SMALL_STATE(4346)] = 175811,
  [SMALL_STATE(4347)] = 175854,
  [SMALL_STATE(4348)] = 175897,
  [SMALL_STATE(4349)] = 175940,
  [SMALL_STATE(4350)] = 175983,
  [SMALL_STATE(4351)] = 176026,
  [SMALL_STATE(4352)] = 176069,
  [SMALL_STATE(4353)] = 176112,
  [SMALL_STATE(4354)] = 176155,
  [SMALL_STATE(4355)] = 176198,
  [SMALL_STATE(4356)] = 176237,
  [SMALL_STATE(4357)] = 176280,
  [SMALL_STATE(4358)] = 176340,
  [SMALL_STATE(4359)] = 176400,
  [SMALL_STATE(4360)] = 176460,
  [SMALL_STATE(4361)] = 176496,
  [SMALL_STATE(4362)] = 176538,
  [SMALL_STATE(4363)] = 176574,
  [SMALL_STATE(4364)] = 176616,
  [SMALL_STATE(4365)] = 176658,
  [SMALL_STATE(4366)] = 176718,
  [SMALL_STATE(4367)] = 176778,
  [SMALL_STATE(4368)] = 176814,
  [SMALL_STATE(4369)] = 176850,
  [SMALL_STATE(4370)] = 176894,
  [SMALL_STATE(4371)] = 176930,
  [SMALL_STATE(4372)] = 176972,
  [SMALL_STATE(4373)] = 177016,
  [SMALL_STATE(4374)] = 177058,
  [SMALL_STATE(4375)] = 177102,
  [SMALL_STATE(4376)] = 177144,
  [SMALL_STATE(4377)] = 177186,
  [SMALL_STATE(4378)] = 177228,
  [SMALL_STATE(4379)] = 177268,
  [SMALL_STATE(4380)] = 177310,
  [SMALL_STATE(4381)] = 177352,
  [SMALL_STATE(4382)] = 177394,
  [SMALL_STATE(4383)] = 177454,
  [SMALL_STATE(4384)] = 177496,
  [SMALL_STATE(4385)] = 177556,
  [SMALL_STATE(4386)] = 177598,
  [SMALL_STATE(4387)] = 177658,
  [SMALL_STATE(4388)] = 177700,
  [SMALL_STATE(4389)] = 177736,
  [SMALL_STATE(4390)] = 177796,
  [SMALL_STATE(4391)] = 177832,
  [SMALL_STATE(4392)] = 177892,
  [SMALL_STATE(4393)] = 177952,
  [SMALL_STATE(4394)] = 178012,
  [SMALL_STATE(4395)] = 178048,
  [SMALL_STATE(4396)] = 178092,
  [SMALL_STATE(4397)] = 178152,
  [SMALL_STATE(4398)] = 178194,
  [SMALL_STATE(4399)] = 178230,
  [SMALL_STATE(4400)] = 178272,
  [SMALL_STATE(4401)] = 178314,
  [SMALL_STATE(4402)] = 178374,
  [SMALL_STATE(4403)] = 178410,
  [SMALL_STATE(4404)] = 178470,
  [SMALL_STATE(4405)] = 178512,
  [SMALL_STATE(4406)] = 178572,
  [SMALL_STATE(4407)] = 178608,
  [SMALL_STATE(4408)] = 178654,
  [SMALL_STATE(4409)] = 178714,
  [SMALL_STATE(4410)] = 178758,
  [SMALL_STATE(4411)] = 178800,
  [SMALL_STATE(4412)] = 178844,
  [SMALL_STATE(4413)] = 178886,
  [SMALL_STATE(4414)] = 178922,
  [SMALL_STATE(4415)] = 178968,
  [SMALL_STATE(4416)] = 179010,
  [SMALL_STATE(4417)] = 179052,
  [SMALL_STATE(4418)] = 179098,
  [SMALL_STATE(4419)] = 179158,
  [SMALL_STATE(4420)] = 179204,
  [SMALL_STATE(4421)] = 179250,
  [SMALL_STATE(4422)] = 179296,
  [SMALL_STATE(4423)] = 179356,
  [SMALL_STATE(4424)] = 179416,
  [SMALL_STATE(4425)] = 179462,
  [SMALL_STATE(4426)] = 179508,
  [SMALL_STATE(4427)] = 179554,
  [SMALL_STATE(4428)] = 179600,
  [SMALL_STATE(4429)] = 179660,
  [SMALL_STATE(4430)] = 179706,
  [SMALL_STATE(4431)] = 179752,
  [SMALL_STATE(4432)] = 179798,
  [SMALL_STATE(4433)] = 179838,
  [SMALL_STATE(4434)] = 179884,
  [SMALL_STATE(4435)] = 179930,
  [SMALL_STATE(4436)] = 179976,
  [SMALL_STATE(4437)] = 180014,
  [SMALL_STATE(4438)] = 180060,
  [SMALL_STATE(4439)] = 180120,
  [SMALL_STATE(4440)] = 180156,
  [SMALL_STATE(4441)] = 180216,
  [SMALL_STATE(4442)] = 180258,
  [SMALL_STATE(4443)] = 180300,
  [SMALL_STATE(4444)] = 180336,
  [SMALL_STATE(4445)] = 180396,
  [SMALL_STATE(4446)] = 180438,
  [SMALL_STATE(4447)] = 180498,
  [SMALL_STATE(4448)] = 180558,
  [SMALL_STATE(4449)] = 180618,
  [SMALL_STATE(4450)] = 180678,
  [SMALL_STATE(4451)] = 180722,
  [SMALL_STATE(4452)] = 180782,
  [SMALL_STATE(4453)] = 180842,
  [SMALL_STATE(4454)] = 180902,
  [SMALL_STATE(4455)] = 180962,
  [SMALL_STATE(4456)] = 181022,
  [SMALL_STATE(4457)] = 181064,
  [SMALL_STATE(4458)] = 181124,
  [SMALL_STATE(4459)] = 181184,
  [SMALL_STATE(4460)] = 181226,
  [SMALL_STATE(4461)] = 181268,
  [SMALL_STATE(4462)] = 181310,
  [SMALL_STATE(4463)] = 181370,
  [SMALL_STATE(4464)] = 181430,
  [SMALL_STATE(4465)] = 181472,
  [SMALL_STATE(4466)] = 181516,
  [SMALL_STATE(4467)] = 181560,
  [SMALL_STATE(4468)] = 181604,
  [SMALL_STATE(4469)] = 181640,
  [SMALL_STATE(4470)] = 181682,
  [SMALL_STATE(4471)] = 181742,
  [SMALL_STATE(4472)] = 181802,
  [SMALL_STATE(4473)] = 181844,
  [SMALL_STATE(4474)] = 181886,
  [SMALL_STATE(4475)] = 181928,
  [SMALL_STATE(4476)] = 181970,
  [SMALL_STATE(4477)] = 182012,
  [SMALL_STATE(4478)] = 182072,
  [SMALL_STATE(4479)] = 182132,
  [SMALL_STATE(4480)] = 182192,
  [SMALL_STATE(4481)] = 182252,
  [SMALL_STATE(4482)] = 182292,
  [SMALL_STATE(4483)] = 182328,
  [SMALL_STATE(4484)] = 182388,
  [SMALL_STATE(4485)] = 182448,
  [SMALL_STATE(4486)] = 182508,
  [SMALL_STATE(4487)] = 182568,
  [SMALL_STATE(4488)] = 182628,
  [SMALL_STATE(4489)] = 182688,
  [SMALL_STATE(4490)] = 182748,
  [SMALL_STATE(4491)] = 182808,
  [SMALL_STATE(4492)] = 182868,
  [SMALL_STATE(4493)] = 182914,
  [SMALL_STATE(4494)] = 182955,
  [SMALL_STATE(4495)] = 182994,
  [SMALL_STATE(4496)] = 183033,
  [SMALL_STATE(4497)] = 183090,
  [SMALL_STATE(4498)] = 183129,
  [SMALL_STATE(4499)] = 183186,
  [SMALL_STATE(4500)] = 183243,
  [SMALL_STATE(4501)] = 183284,
  [SMALL_STATE(4502)] = 183323,
  [SMALL_STATE(4503)] = 183360,
  [SMALL_STATE(4504)] = 183399,
  [SMALL_STATE(4505)] = 183456,
  [SMALL_STATE(4506)] = 183497,
  [SMALL_STATE(4507)] = 183538,
  [SMALL_STATE(4508)] = 183595,
  [SMALL_STATE(4509)] = 183634,
  [SMALL_STATE(4510)] = 183671,
  [SMALL_STATE(4511)] = 183708,
  [SMALL_STATE(4512)] = 183765,
  [SMALL_STATE(4513)] = 183804,
  [SMALL_STATE(4514)] = 183845,
  [SMALL_STATE(4515)] = 183902,
  [SMALL_STATE(4516)] = 183943,
  [SMALL_STATE(4517)] = 183980,
  [SMALL_STATE(4518)] = 184019,
  [SMALL_STATE(4519)] = 184058,
  [SMALL_STATE(4520)] = 184115,
  [SMALL_STATE(4521)] = 184150,
  [SMALL_STATE(4522)] = 184207,
  [SMALL_STATE(4523)] = 184246,
  [SMALL_STATE(4524)] = 184285,
  [SMALL_STATE(4525)] = 184324,
  [SMALL_STATE(4526)] = 184361,
  [SMALL_STATE(4527)] = 184418,
  [SMALL_STATE(4528)] = 184457,
  [SMALL_STATE(4529)] = 184498,
  [SMALL_STATE(4530)] = 184547,
  [SMALL_STATE(4531)] = 184588,
  [SMALL_STATE(4532)] = 184645,
  [SMALL_STATE(4533)] = 184702,
  [SMALL_STATE(4534)] = 184749,
  [SMALL_STATE(4535)] = 184788,
  [SMALL_STATE(4536)] = 184829,
  [SMALL_STATE(4537)] = 184868,
  [SMALL_STATE(4538)] = 184925,
  [SMALL_STATE(4539)] = 184964,
  [SMALL_STATE(4540)] = 185003,
  [SMALL_STATE(4541)] = 185042,
  [SMALL_STATE(4542)] = 185079,
  [SMALL_STATE(4543)] = 185116,
  [SMALL_STATE(4544)] = 185155,
  [SMALL_STATE(4545)] = 185194,
  [SMALL_STATE(4546)] = 185233,
  [SMALL_STATE(4547)] = 185274,
  [SMALL_STATE(4548)] = 185331,
  [SMALL_STATE(4549)] = 185372,
  [SMALL_STATE(4550)] = 185411,
  [SMALL_STATE(4551)] = 185452,
  [SMALL_STATE(4552)] = 185491,
  [SMALL_STATE(4553)] = 185548,
  [SMALL_STATE(4554)] = 185605,
  [SMALL_STATE(4555)] = 185644,
  [SMALL_STATE(4556)] = 185685,
  [SMALL_STATE(4557)] = 185739,
  [SMALL_STATE(4558)] = 185779,
  [SMALL_STATE(4559)] = 185833,
  [SMALL_STATE(4560)] = 185887,
  [SMALL_STATE(4561)] = 185941,
  [SMALL_STATE(4562)] = 185997,
  [SMALL_STATE(4563)] = 186051,
  [SMALL_STATE(4564)] = 186105,
  [SMALL_STATE(4565)] = 186159,
  [SMALL_STATE(4566)] = 186213,
  [SMALL_STATE(4567)] = 186267,
  [SMALL_STATE(4568)] = 186321,
  [SMALL_STATE(4569)] = 186375,
  [SMALL_STATE(4570)] = 186429,
  [SMALL_STATE(4571)] = 186483,
  [SMALL_STATE(4572)] = 186537,
  [SMALL_STATE(4573)] = 186591,
  [SMALL_STATE(4574)] = 186645,
  [SMALL_STATE(4575)] = 186699,
  [SMALL_STATE(4576)] = 186753,
  [SMALL_STATE(4577)] = 186807,
  [SMALL_STATE(4578)] = 186861,
  [SMALL_STATE(4579)] = 186915,
  [SMALL_STATE(4580)] = 186971,
  [SMALL_STATE(4581)] = 187025,
  [SMALL_STATE(4582)] = 187079,
  [SMALL_STATE(4583)] = 187133,
  [SMALL_STATE(4584)] = 187187,
  [SMALL_STATE(4585)] = 187241,
  [SMALL_STATE(4586)] = 187295,
  [SMALL_STATE(4587)] = 187349,
  [SMALL_STATE(4588)] = 187403,
  [SMALL_STATE(4589)] = 187457,
  [SMALL_STATE(4590)] = 187511,
  [SMALL_STATE(4591)] = 187565,
  [SMALL_STATE(4592)] = 187619,
  [SMALL_STATE(4593)] = 187673,
  [SMALL_STATE(4594)] = 187727,
  [SMALL_STATE(4595)] = 187781,
  [SMALL_STATE(4596)] = 187835,
  [SMALL_STATE(4597)] = 187889,
  [SMALL_STATE(4598)] = 187943,
  [SMALL_STATE(4599)] = 187997,
  [SMALL_STATE(4600)] = 188051,
  [SMALL_STATE(4601)] = 188105,
  [SMALL_STATE(4602)] = 188159,
  [SMALL_STATE(4603)] = 188213,
  [SMALL_STATE(4604)] = 188267,
  [SMALL_STATE(4605)] = 188321,
  [SMALL_STATE(4606)] = 188375,
  [SMALL_STATE(4607)] = 188429,
  [SMALL_STATE(4608)] = 188483,
  [SMALL_STATE(4609)] = 188537,
  [SMALL_STATE(4610)] = 188593,
  [SMALL_STATE(4611)] = 188649,
  [SMALL_STATE(4612)] = 188703,
  [SMALL_STATE(4613)] = 188757,
  [SMALL_STATE(4614)] = 188811,
  [SMALL_STATE(4615)] = 188865,
  [SMALL_STATE(4616)] = 188919,
  [SMALL_STATE(4617)] = 188959,
  [SMALL_STATE(4618)] = 189015,
  [SMALL_STATE(4619)] = 189069,
  [SMALL_STATE(4620)] = 189123,
  [SMALL_STATE(4621)] = 189177,
  [SMALL_STATE(4622)] = 189211,
  [SMALL_STATE(4623)] = 189263,
  [SMALL_STATE(4624)] = 189317,
  [SMALL_STATE(4625)] = 189371,
  [SMALL_STATE(4626)] = 189409,
  [SMALL_STATE(4627)] = 189447,
  [SMALL_STATE(4628)] = 189501,
  [SMALL_STATE(4629)] = 189549,
  [SMALL_STATE(4630)] = 189589,
  [SMALL_STATE(4631)] = 189627,
  [SMALL_STATE(4632)] = 189683,
  [SMALL_STATE(4633)] = 189737,
  [SMALL_STATE(4634)] = 189771,
  [SMALL_STATE(4635)] = 189811,
  [SMALL_STATE(4636)] = 189849,
  [SMALL_STATE(4637)] = 189889,
  [SMALL_STATE(4638)] = 189943,
  [SMALL_STATE(4639)] = 189983,
  [SMALL_STATE(4640)] = 190037,
  [SMALL_STATE(4641)] = 190091,
  [SMALL_STATE(4642)] = 190145,
  [SMALL_STATE(4643)] = 190199,
  [SMALL_STATE(4644)] = 190253,
  [SMALL_STATE(4645)] = 190307,
  [SMALL_STATE(4646)] = 190361,
  [SMALL_STATE(4647)] = 190399,
  [SMALL_STATE(4648)] = 190439,
  [SMALL_STATE(4649)] = 190495,
  [SMALL_STATE(4650)] = 190531,
  [SMALL_STATE(4651)] = 190569,
  [SMALL_STATE(4652)] = 190623,
  [SMALL_STATE(4653)] = 190661,
  [SMALL_STATE(4654)] = 190703,
  [SMALL_STATE(4655)] = 190757,
  [SMALL_STATE(4656)] = 190797,
  [SMALL_STATE(4657)] = 190853,
  [SMALL_STATE(4658)] = 190891,
  [SMALL_STATE(4659)] = 190945,
  [SMALL_STATE(4660)] = 190985,
  [SMALL_STATE(4661)] = 191025,
  [SMALL_STATE(4662)] = 191079,
  [SMALL_STATE(4663)] = 191135,
  [SMALL_STATE(4664)] = 191175,
  [SMALL_STATE(4665)] = 191229,
  [SMALL_STATE(4666)] = 191283,
  [SMALL_STATE(4667)] = 191329,
  [SMALL_STATE(4668)] = 191383,
  [SMALL_STATE(4669)] = 191423,
  [SMALL_STATE(4670)] = 191461,
  [SMALL_STATE(4671)] = 191497,
  [SMALL_STATE(4672)] = 191551,
  [SMALL_STATE(4673)] = 191591,
  [SMALL_STATE(4674)] = 191631,
  [SMALL_STATE(4675)] = 191685,
  [SMALL_STATE(4676)] = 191739,
  [SMALL_STATE(4677)] = 191779,
  [SMALL_STATE(4678)] = 191833,
  [SMALL_STATE(4679)] = 191887,
  [SMALL_STATE(4680)] = 191927,
  [SMALL_STATE(4681)] = 191981,
  [SMALL_STATE(4682)] = 192021,
  [SMALL_STATE(4683)] = 192075,
  [SMALL_STATE(4684)] = 192129,
  [SMALL_STATE(4685)] = 192183,
  [SMALL_STATE(4686)] = 192237,
  [SMALL_STATE(4687)] = 192291,
  [SMALL_STATE(4688)] = 192345,
  [SMALL_STATE(4689)] = 192391,
  [SMALL_STATE(4690)] = 192433,
  [SMALL_STATE(4691)] = 192487,
  [SMALL_STATE(4692)] = 192543,
  [SMALL_STATE(4693)] = 192597,
  [SMALL_STATE(4694)] = 192651,
  [SMALL_STATE(4695)] = 192705,
  [SMALL_STATE(4696)] = 192761,
  [SMALL_STATE(4697)] = 192815,
  [SMALL_STATE(4698)] = 192869,
  [SMALL_STATE(4699)] = 192923,
  [SMALL_STATE(4700)] = 192977,
  [SMALL_STATE(4701)] = 193013,
  [SMALL_STATE(4702)] = 193067,
  [SMALL_STATE(4703)] = 193121,
  [SMALL_STATE(4704)] = 193177,
  [SMALL_STATE(4705)] = 193231,
  [SMALL_STATE(4706)] = 193285,
  [SMALL_STATE(4707)] = 193339,
  [SMALL_STATE(4708)] = 193393,
  [SMALL_STATE(4709)] = 193447,
  [SMALL_STATE(4710)] = 193501,
  [SMALL_STATE(4711)] = 193555,
  [SMALL_STATE(4712)] = 193609,
  [SMALL_STATE(4713)] = 193663,
  [SMALL_STATE(4714)] = 193717,
  [SMALL_STATE(4715)] = 193771,
  [SMALL_STATE(4716)] = 193825,
  [SMALL_STATE(4717)] = 193873,
  [SMALL_STATE(4718)] = 193927,
  [SMALL_STATE(4719)] = 193981,
  [SMALL_STATE(4720)] = 194015,
  [SMALL_STATE(4721)] = 194069,
  [SMALL_STATE(4722)] = 194123,
  [SMALL_STATE(4723)] = 194177,
  [SMALL_STATE(4724)] = 194231,
  [SMALL_STATE(4725)] = 194287,
  [SMALL_STATE(4726)] = 194341,
  [SMALL_STATE(4727)] = 194395,
  [SMALL_STATE(4728)] = 194449,
  [SMALL_STATE(4729)] = 194503,
  [SMALL_STATE(4730)] = 194557,
  [SMALL_STATE(4731)] = 194611,
  [SMALL_STATE(4732)] = 194665,
  [SMALL_STATE(4733)] = 194719,
  [SMALL_STATE(4734)] = 194773,
  [SMALL_STATE(4735)] = 194827,
  [SMALL_STATE(4736)] = 194881,
  [SMALL_STATE(4737)] = 194937,
  [SMALL_STATE(4738)] = 194991,
  [SMALL_STATE(4739)] = 195045,
  [SMALL_STATE(4740)] = 195099,
  [SMALL_STATE(4741)] = 195153,
  [SMALL_STATE(4742)] = 195207,
  [SMALL_STATE(4743)] = 195261,
  [SMALL_STATE(4744)] = 195295,
  [SMALL_STATE(4745)] = 195349,
  [SMALL_STATE(4746)] = 195403,
  [SMALL_STATE(4747)] = 195457,
  [SMALL_STATE(4748)] = 195511,
  [SMALL_STATE(4749)] = 195565,
  [SMALL_STATE(4750)] = 195619,
  [SMALL_STATE(4751)] = 195675,
  [SMALL_STATE(4752)] = 195729,
  [SMALL_STATE(4753)] = 195783,
  [SMALL_STATE(4754)] = 195819,
  [SMALL_STATE(4755)] = 195873,
  [SMALL_STATE(4756)] = 195915,
  [SMALL_STATE(4757)] = 195969,
  [SMALL_STATE(4758)] = 196023,
  [SMALL_STATE(4759)] = 196077,
  [SMALL_STATE(4760)] = 196131,
  [SMALL_STATE(4761)] = 196185,
  [SMALL_STATE(4762)] = 196239,
  [SMALL_STATE(4763)] = 196293,
  [SMALL_STATE(4764)] = 196347,
  [SMALL_STATE(4765)] = 196389,
  [SMALL_STATE(4766)] = 196443,
  [SMALL_STATE(4767)] = 196483,
  [SMALL_STATE(4768)] = 196537,
  [SMALL_STATE(4769)] = 196591,
  [SMALL_STATE(4770)] = 196645,
  [SMALL_STATE(4771)] = 196699,
  [SMALL_STATE(4772)] = 196753,
  [SMALL_STATE(4773)] = 196807,
  [SMALL_STATE(4774)] = 196840,
  [SMALL_STATE(4775)] = 196873,
  [SMALL_STATE(4776)] = 196906,
  [SMALL_STATE(4777)] = 196959,
  [SMALL_STATE(4778)] = 196996,
  [SMALL_STATE(4779)] = 197049,
  [SMALL_STATE(4780)] = 197082,
  [SMALL_STATE(4781)] = 197115,
  [SMALL_STATE(4782)] = 197148,
  [SMALL_STATE(4783)] = 197181,
  [SMALL_STATE(4784)] = 197214,
  [SMALL_STATE(4785)] = 197275,
  [SMALL_STATE(4786)] = 197308,
  [SMALL_STATE(4787)] = 197341,
  [SMALL_STATE(4788)] = 197392,
  [SMALL_STATE(4789)] = 197431,
  [SMALL_STATE(4790)] = 197464,
  [SMALL_STATE(4791)] = 197525,
  [SMALL_STATE(4792)] = 197574,
  [SMALL_STATE(4793)] = 197625,
  [SMALL_STATE(4794)] = 197674,
  [SMALL_STATE(4795)] = 197707,
  [SMALL_STATE(4796)] = 197742,
  [SMALL_STATE(4797)] = 197775,
  [SMALL_STATE(4798)] = 197812,
  [SMALL_STATE(4799)] = 197847,
  [SMALL_STATE(4800)] = 197880,
  [SMALL_STATE(4801)] = 197941,
  [SMALL_STATE(4802)] = 197974,
  [SMALL_STATE(4803)] = 198007,
  [SMALL_STATE(4804)] = 198056,
  [SMALL_STATE(4805)] = 198093,
  [SMALL_STATE(4806)] = 198130,
  [SMALL_STATE(4807)] = 198163,
  [SMALL_STATE(4808)] = 198211,
  [SMALL_STATE(4809)] = 198259,
  [SMALL_STATE(4810)] = 198307,
  [SMALL_STATE(4811)] = 198355,
  [SMALL_STATE(4812)] = 198403,
  [SMALL_STATE(4813)] = 198435,
  [SMALL_STATE(4814)] = 198467,
  [SMALL_STATE(4815)] = 198499,
  [SMALL_STATE(4816)] = 198547,
  [SMALL_STATE(4817)] = 198595,
  [SMALL_STATE(4818)] = 198643,
  [SMALL_STATE(4819)] = 198693,
  [SMALL_STATE(4820)] = 198725,
  [SMALL_STATE(4821)] = 198775,
  [SMALL_STATE(4822)] = 198823,
  [SMALL_STATE(4823)] = 198871,
  [SMALL_STATE(4824)] = 198905,
  [SMALL_STATE(4825)] = 198955,
  [SMALL_STATE(4826)] = 199003,
  [SMALL_STATE(4827)] = 199051,
  [SMALL_STATE(4828)] = 199099,
  [SMALL_STATE(4829)] = 199147,
  [SMALL_STATE(4830)] = 199195,
  [SMALL_STATE(4831)] = 199243,
  [SMALL_STATE(4832)] = 199291,
  [SMALL_STATE(4833)] = 199339,
  [SMALL_STATE(4834)] = 199387,
  [SMALL_STATE(4835)] = 199435,
  [SMALL_STATE(4836)] = 199483,
  [SMALL_STATE(4837)] = 199531,
  [SMALL_STATE(4838)] = 199579,
  [SMALL_STATE(4839)] = 199629,
  [SMALL_STATE(4840)] = 199665,
  [SMALL_STATE(4841)] = 199701,
  [SMALL_STATE(4842)] = 199749,
  [SMALL_STATE(4843)] = 199797,
  [SMALL_STATE(4844)] = 199845,
  [SMALL_STATE(4845)] = 199895,
  [SMALL_STATE(4846)] = 199943,
  [SMALL_STATE(4847)] = 199993,
  [SMALL_STATE(4848)] = 200041,
  [SMALL_STATE(4849)] = 200089,
  [SMALL_STATE(4850)] = 200137,
  [SMALL_STATE(4851)] = 200185,
  [SMALL_STATE(4852)] = 200233,
  [SMALL_STATE(4853)] = 200281,
  [SMALL_STATE(4854)] = 200329,
  [SMALL_STATE(4855)] = 200377,
  [SMALL_STATE(4856)] = 200425,
  [SMALL_STATE(4857)] = 200473,
  [SMALL_STATE(4858)] = 200521,
  [SMALL_STATE(4859)] = 200569,
  [SMALL_STATE(4860)] = 200624,
  [SMALL_STATE(4861)] = 200665,
  [SMALL_STATE(4862)] = 200706,
  [SMALL_STATE(4863)] = 200739,
  [SMALL_STATE(4864)] = 200770,
  [SMALL_STATE(4865)] = 200811,
  [SMALL_STATE(4866)] = 200850,
  [SMALL_STATE(4867)] = 200883,
  [SMALL_STATE(4868)] = 200916,
  [SMALL_STATE(4869)] = 200949,
  [SMALL_STATE(4870)] = 200990,
  [SMALL_STATE(4871)] = 201021,
  [SMALL_STATE(4872)] = 201054,
  [SMALL_STATE(4873)] = 201087,
  [SMALL_STATE(4874)] = 201128,
  [SMALL_STATE(4875)] = 201161,
  [SMALL_STATE(4876)] = 201202,
  [SMALL_STATE(4877)] = 201243,
  [SMALL_STATE(4878)] = 201276,
  [SMALL_STATE(4879)] = 201309,
  [SMALL_STATE(4880)] = 201342,
  [SMALL_STATE(4881)] = 201383,
  [SMALL_STATE(4882)] = 201424,
  [SMALL_STATE(4883)] = 201465,
  [SMALL_STATE(4884)] = 201506,
  [SMALL_STATE(4885)] = 201547,
  [SMALL_STATE(4886)] = 201586,
  [SMALL_STATE(4887)] = 201627,
  [SMALL_STATE(4888)] = 201658,
  [SMALL_STATE(4889)] = 201689,
  [SMALL_STATE(4890)] = 201722,
  [SMALL_STATE(4891)] = 201777,
  [SMALL_STATE(4892)] = 201818,
  [SMALL_STATE(4893)] = 201859,
  [SMALL_STATE(4894)] = 201892,
  [SMALL_STATE(4895)] = 201933,
  [SMALL_STATE(4896)] = 201974,
  [SMALL_STATE(4897)] = 202015,
  [SMALL_STATE(4898)] = 202056,
  [SMALL_STATE(4899)] = 202089,
  [SMALL_STATE(4900)] = 202130,
  [SMALL_STATE(4901)] = 202185,
  [SMALL_STATE(4902)] = 202218,
  [SMALL_STATE(4903)] = 202251,
  [SMALL_STATE(4904)] = 202282,
  [SMALL_STATE(4905)] = 202315,
  [SMALL_STATE(4906)] = 202348,
  [SMALL_STATE(4907)] = 202381,
  [SMALL_STATE(4908)] = 202414,
  [SMALL_STATE(4909)] = 202455,
  [SMALL_STATE(4910)] = 202501,
  [SMALL_STATE(4911)] = 202539,
  [SMALL_STATE(4912)] = 202569,
  [SMALL_STATE(4913)] = 202615,
  [SMALL_STATE(4914)] = 202661,
  [SMALL_STATE(4915)] = 202707,
  [SMALL_STATE(4916)] = 202753,
  [SMALL_STATE(4917)] = 202799,
  [SMALL_STATE(4918)] = 202851,
  [SMALL_STATE(4919)] = 202903,
  [SMALL_STATE(4920)] = 202955,
  [SMALL_STATE(4921)] = 203007,
  [SMALL_STATE(4922)] = 203059,
  [SMALL_STATE(4923)] = 203111,
  [SMALL_STATE(4924)] = 203155,
  [SMALL_STATE(4925)] = 203207,
  [SMALL_STATE(4926)] = 203253,
  [SMALL_STATE(4927)] = 203297,
  [SMALL_STATE(4928)] = 203343,
  [SMALL_STATE(4929)] = 203389,
  [SMALL_STATE(4930)] = 203441,
  [SMALL_STATE(4931)] = 203473,
  [SMALL_STATE(4932)] = 203503,
  [SMALL_STATE(4933)] = 203547,
  [SMALL_STATE(4934)] = 203583,
  [SMALL_STATE(4935)] = 203629,
  [SMALL_STATE(4936)] = 203675,
  [SMALL_STATE(4937)] = 203719,
  [SMALL_STATE(4938)] = 203765,
  [SMALL_STATE(4939)] = 203811,
  [SMALL_STATE(4940)] = 203857,
  [SMALL_STATE(4941)] = 203903,
  [SMALL_STATE(4942)] = 203955,
  [SMALL_STATE(4943)] = 203982,
  [SMALL_STATE(4944)] = 204025,
  [SMALL_STATE(4945)] = 204056,
  [SMALL_STATE(4946)] = 204083,
  [SMALL_STATE(4947)] = 204114,
  [SMALL_STATE(4948)] = 204161,
  [SMALL_STATE(4949)] = 204210,
  [SMALL_STATE(4950)] = 204253,
  [SMALL_STATE(4951)] = 204300,
  [SMALL_STATE(4952)] = 204347,
  [SMALL_STATE(4953)] = 204396,
  [SMALL_STATE(4954)] = 204425,
  [SMALL_STATE(4955)] = 204468,
  [SMALL_STATE(4956)] = 204511,
  [SMALL_STATE(4957)] = 204554,
  [SMALL_STATE(4958)] = 204583,
  [SMALL_STATE(4959)] = 204630,
  [SMALL_STATE(4960)] = 204661,
  [SMALL_STATE(4961)] = 204690,
  [SMALL_STATE(4962)] = 204739,
  [SMALL_STATE(4963)] = 204770,
  [SMALL_STATE(4964)] = 204799,
  [SMALL_STATE(4965)] = 204830,
  [SMALL_STATE(4966)] = 204869,
  [SMALL_STATE(4967)] = 204916,
  [SMALL_STATE(4968)] = 204965,
  [SMALL_STATE(4969)] = 205008,
  [SMALL_STATE(4970)] = 205055,
  [SMALL_STATE(4971)] = 205082,
  [SMALL_STATE(4972)] = 205111,
  [SMALL_STATE(4973)] = 205138,
  [SMALL_STATE(4974)] = 205164,
  [SMALL_STATE(4975)] = 205206,
  [SMALL_STATE(4976)] = 205248,
  [SMALL_STATE(4977)] = 205282,
  [SMALL_STATE(4978)] = 205324,
  [SMALL_STATE(4979)] = 205358,
  [SMALL_STATE(4980)] = 205392,
  [SMALL_STATE(4981)] = 205436,
  [SMALL_STATE(4982)] = 205470,
  [SMALL_STATE(4983)] = 205512,
  [SMALL_STATE(4984)] = 205540,
  [SMALL_STATE(4985)] = 205574,
  [SMALL_STATE(4986)] = 205618,
  [SMALL_STATE(4987)] = 205644,
  [SMALL_STATE(4988)] = 205686,
  [SMALL_STATE(4989)] = 205728,
  [SMALL_STATE(4990)] = 205760,
  [SMALL_STATE(4991)] = 205802,
  [SMALL_STATE(4992)] = 205844,
  [SMALL_STATE(4993)] = 205886,
  [SMALL_STATE(4994)] = 205928,
  [SMALL_STATE(4995)] = 205962,
  [SMALL_STATE(4996)] = 206004,
  [SMALL_STATE(4997)] = 206030,
  [SMALL_STATE(4998)] = 206058,
  [SMALL_STATE(4999)] = 206092,
  [SMALL_STATE(5000)] = 206134,
  [SMALL_STATE(5001)] = 206176,
  [SMALL_STATE(5002)] = 206204,
  [SMALL_STATE(5003)] = 206230,
  [SMALL_STATE(5004)] = 206272,
  [SMALL_STATE(5005)] = 206306,
  [SMALL_STATE(5006)] = 206334,
  [SMALL_STATE(5007)] = 206378,
  [SMALL_STATE(5008)] = 206412,
  [SMALL_STATE(5009)] = 206446,
  [SMALL_STATE(5010)] = 206474,
  [SMALL_STATE(5011)] = 206500,
  [SMALL_STATE(5012)] = 206544,
  [SMALL_STATE(5013)] = 206586,
  [SMALL_STATE(5014)] = 206622,
  [SMALL_STATE(5015)] = 206664,
  [SMALL_STATE(5016)] = 206690,
  [SMALL_STATE(5017)] = 206724,
  [SMALL_STATE(5018)] = 206754,
  [SMALL_STATE(5019)] = 206780,
  [SMALL_STATE(5020)] = 206814,
  [SMALL_STATE(5021)] = 206856,
  [SMALL_STATE(5022)] = 206898,
  [SMALL_STATE(5023)] = 206940,
  [SMALL_STATE(5024)] = 206984,
  [SMALL_STATE(5025)] = 207020,
  [SMALL_STATE(5026)] = 207050,
  [SMALL_STATE(5027)] = 207078,
  [SMALL_STATE(5028)] = 207120,
  [SMALL_STATE(5029)] = 207164,
  [SMALL_STATE(5030)] = 207208,
  [SMALL_STATE(5031)] = 207244,
  [SMALL_STATE(5032)] = 207288,
  [SMALL_STATE(5033)] = 207316,
  [SMALL_STATE(5034)] = 207343,
  [SMALL_STATE(5035)] = 207380,
  [SMALL_STATE(5036)] = 207407,
  [SMALL_STATE(5037)] = 207434,
  [SMALL_STATE(5038)] = 207465,
  [SMALL_STATE(5039)] = 207504,
  [SMALL_STATE(5040)] = 207535,
  [SMALL_STATE(5041)] = 207566,
  [SMALL_STATE(5042)] = 207605,
  [SMALL_STATE(5043)] = 207636,
  [SMALL_STATE(5044)] = 207663,
  [SMALL_STATE(5045)] = 207694,
  [SMALL_STATE(5046)] = 207725,
  [SMALL_STATE(5047)] = 207756,
  [SMALL_STATE(5048)] = 207787,
  [SMALL_STATE(5049)] = 207814,
  [SMALL_STATE(5050)] = 207845,
  [SMALL_STATE(5051)] = 207876,
  [SMALL_STATE(5052)] = 207907,
  [SMALL_STATE(5053)] = 207938,
  [SMALL_STATE(5054)] = 207969,
  [SMALL_STATE(5055)] = 208000,
  [SMALL_STATE(5056)] = 208027,
  [SMALL_STATE(5057)] = 208058,
  [SMALL_STATE(5058)] = 208089,
  [SMALL_STATE(5059)] = 208120,
  [SMALL_STATE(5060)] = 208151,
  [SMALL_STATE(5061)] = 208190,
  [SMALL_STATE(5062)] = 208221,
  [SMALL_STATE(5063)] = 208252,
  [SMALL_STATE(5064)] = 208291,
  [SMALL_STATE(5065)] = 208318,
  [SMALL_STATE(5066)] = 208357,
  [SMALL_STATE(5067)] = 208396,
  [SMALL_STATE(5068)] = 208423,
  [SMALL_STATE(5069)] = 208450,
  [SMALL_STATE(5070)] = 208489,
  [SMALL_STATE(5071)] = 208516,
  [SMALL_STATE(5072)] = 208543,
  [SMALL_STATE(5073)] = 208582,
  [SMALL_STATE(5074)] = 208609,
  [SMALL_STATE(5075)] = 208636,
  [SMALL_STATE(5076)] = 208663,
  [SMALL_STATE(5077)] = 208702,
  [SMALL_STATE(5078)] = 208741,
  [SMALL_STATE(5079)] = 208772,
  [SMALL_STATE(5080)] = 208803,
  [SMALL_STATE(5081)] = 208834,
  [SMALL_STATE(5082)] = 208865,
  [SMALL_STATE(5083)] = 208896,
  [SMALL_STATE(5084)] = 208927,
  [SMALL_STATE(5085)] = 208958,
  [SMALL_STATE(5086)] = 208989,
  [SMALL_STATE(5087)] = 209020,
  [SMALL_STATE(5088)] = 209051,
  [SMALL_STATE(5089)] = 209082,
  [SMALL_STATE(5090)] = 209121,
  [SMALL_STATE(5091)] = 209152,
  [SMALL_STATE(5092)] = 209179,
  [SMALL_STATE(5093)] = 209218,
  [SMALL_STATE(5094)] = 209249,
  [SMALL_STATE(5095)] = 209280,
  [SMALL_STATE(5096)] = 209311,
  [SMALL_STATE(5097)] = 209342,
  [SMALL_STATE(5098)] = 209381,
  [SMALL_STATE(5099)] = 209408,
  [SMALL_STATE(5100)] = 209435,
  [SMALL_STATE(5101)] = 209462,
  [SMALL_STATE(5102)] = 209501,
  [SMALL_STATE(5103)] = 209540,
  [SMALL_STATE(5104)] = 209579,
  [SMALL_STATE(5105)] = 209606,
  [SMALL_STATE(5106)] = 209645,
  [SMALL_STATE(5107)] = 209676,
  [SMALL_STATE(5108)] = 209707,
  [SMALL_STATE(5109)] = 209738,
  [SMALL_STATE(5110)] = 209777,
  [SMALL_STATE(5111)] = 209808,
  [SMALL_STATE(5112)] = 209839,
  [SMALL_STATE(5113)] = 209878,
  [SMALL_STATE(5114)] = 209909,
  [SMALL_STATE(5115)] = 209940,
  [SMALL_STATE(5116)] = 209971,
  [SMALL_STATE(5117)] = 210010,
  [SMALL_STATE(5118)] = 210041,
  [SMALL_STATE(5119)] = 210072,
  [SMALL_STATE(5120)] = 210099,
  [SMALL_STATE(5121)] = 210126,
  [SMALL_STATE(5122)] = 210165,
  [SMALL_STATE(5123)] = 210204,
  [SMALL_STATE(5124)] = 210231,
  [SMALL_STATE(5125)] = 210262,
  [SMALL_STATE(5126)] = 210293,
  [SMALL_STATE(5127)] = 210324,
  [SMALL_STATE(5128)] = 210355,
  [SMALL_STATE(5129)] = 210386,
  [SMALL_STATE(5130)] = 210417,
  [SMALL_STATE(5131)] = 210460,
  [SMALL_STATE(5132)] = 210499,
  [SMALL_STATE(5133)] = 210526,
  [SMALL_STATE(5134)] = 210553,
  [SMALL_STATE(5135)] = 210580,
  [SMALL_STATE(5136)] = 210607,
  [SMALL_STATE(5137)] = 210646,
  [SMALL_STATE(5138)] = 210689,
  [SMALL_STATE(5139)] = 210720,
  [SMALL_STATE(5140)] = 210751,
  [SMALL_STATE(5141)] = 210782,
  [SMALL_STATE(5142)] = 210813,
  [SMALL_STATE(5143)] = 210840,
  [SMALL_STATE(5144)] = 210867,
  [SMALL_STATE(5145)] = 210898,
  [SMALL_STATE(5146)] = 210929,
  [SMALL_STATE(5147)] = 210960,
  [SMALL_STATE(5148)] = 210991,
  [SMALL_STATE(5149)] = 211022,
  [SMALL_STATE(5150)] = 211061,
  [SMALL_STATE(5151)] = 211100,
  [SMALL_STATE(5152)] = 211127,
  [SMALL_STATE(5153)] = 211154,
  [SMALL_STATE(5154)] = 211185,
  [SMALL_STATE(5155)] = 211224,
  [SMALL_STATE(5156)] = 211251,
  [SMALL_STATE(5157)] = 211282,
  [SMALL_STATE(5158)] = 211313,
  [SMALL_STATE(5159)] = 211344,
  [SMALL_STATE(5160)] = 211381,
  [SMALL_STATE(5161)] = 211412,
  [SMALL_STATE(5162)] = 211443,
  [SMALL_STATE(5163)] = 211474,
  [SMALL_STATE(5164)] = 211515,
  [SMALL_STATE(5165)] = 211546,
  [SMALL_STATE(5166)] = 211577,
  [SMALL_STATE(5167)] = 211604,
  [SMALL_STATE(5168)] = 211631,
  [SMALL_STATE(5169)] = 211658,
  [SMALL_STATE(5170)] = 211685,
  [SMALL_STATE(5171)] = 211712,
  [SMALL_STATE(5172)] = 211751,
  [SMALL_STATE(5173)] = 211778,
  [SMALL_STATE(5174)] = 211805,
  [SMALL_STATE(5175)] = 211836,
  [SMALL_STATE(5176)] = 211866,
  [SMALL_STATE(5177)] = 211896,
  [SMALL_STATE(5178)] = 211926,
  [SMALL_STATE(5179)] = 211956,
  [SMALL_STATE(5180)] = 211986,
  [SMALL_STATE(5181)] = 212010,
  [SMALL_STATE(5182)] = 212034,
  [SMALL_STATE(5183)] = 212066,
  [SMALL_STATE(5184)] = 212108,
  [SMALL_STATE(5185)] = 212148,
  [SMALL_STATE(5186)] = 212188,
  [SMALL_STATE(5187)] = 212220,
  [SMALL_STATE(5188)] = 212252,
  [SMALL_STATE(5189)] = 212282,
  [SMALL_STATE(5190)] = 212312,
  [SMALL_STATE(5191)] = 212342,
  [SMALL_STATE(5192)] = 212372,
  [SMALL_STATE(5193)] = 212402,
  [SMALL_STATE(5194)] = 212428,
  [SMALL_STATE(5195)] = 212468,
  [SMALL_STATE(5196)] = 212508,
  [SMALL_STATE(5197)] = 212532,
  [SMALL_STATE(5198)] = 212560,
  [SMALL_STATE(5199)] = 212602,
  [SMALL_STATE(5200)] = 212642,
  [SMALL_STATE(5201)] = 212666,
  [SMALL_STATE(5202)] = 212692,
  [SMALL_STATE(5203)] = 212732,
  [SMALL_STATE(5204)] = 212755,
  [SMALL_STATE(5205)] = 212780,
  [SMALL_STATE(5206)] = 212807,
  [SMALL_STATE(5207)] = 212848,
  [SMALL_STATE(5208)] = 212871,
  [SMALL_STATE(5209)] = 212894,
  [SMALL_STATE(5210)] = 212917,
  [SMALL_STATE(5211)] = 212954,
  [SMALL_STATE(5212)] = 212991,
  [SMALL_STATE(5213)] = 213020,
  [SMALL_STATE(5214)] = 213049,
  [SMALL_STATE(5215)] = 213072,
  [SMALL_STATE(5216)] = 213107,
  [SMALL_STATE(5217)] = 213148,
  [SMALL_STATE(5218)] = 213177,
  [SMALL_STATE(5219)] = 213212,
  [SMALL_STATE(5220)] = 213235,
  [SMALL_STATE(5221)] = 213260,
  [SMALL_STATE(5222)] = 213299,
  [SMALL_STATE(5223)] = 213326,
  [SMALL_STATE(5224)] = 213363,
  [SMALL_STATE(5225)] = 213386,
  [SMALL_STATE(5226)] = 213423,
  [SMALL_STATE(5227)] = 213454,
  [SMALL_STATE(5228)] = 213483,
  [SMALL_STATE(5229)] = 213516,
  [SMALL_STATE(5230)] = 213539,
  [SMALL_STATE(5231)] = 213576,
  [SMALL_STATE(5232)] = 213603,
  [SMALL_STATE(5233)] = 213626,
  [SMALL_STATE(5234)] = 213649,
  [SMALL_STATE(5235)] = 213672,
  [SMALL_STATE(5236)] = 213709,
  [SMALL_STATE(5237)] = 213746,
  [SMALL_STATE(5238)] = 213773,
  [SMALL_STATE(5239)] = 213810,
  [SMALL_STATE(5240)] = 213833,
  [SMALL_STATE(5241)] = 213870,
  [SMALL_STATE(5242)] = 213895,
  [SMALL_STATE(5243)] = 213932,
  [SMALL_STATE(5244)] = 213969,
  [SMALL_STATE(5245)] = 214008,
  [SMALL_STATE(5246)] = 214045,
  [SMALL_STATE(5247)] = 214068,
  [SMALL_STATE(5248)] = 214091,
  [SMALL_STATE(5249)] = 214124,
  [SMALL_STATE(5250)] = 214147,
  [SMALL_STATE(5251)] = 214170,
  [SMALL_STATE(5252)] = 214207,
  [SMALL_STATE(5253)] = 214244,
  [SMALL_STATE(5254)] = 214266,
  [SMALL_STATE(5255)] = 214304,
  [SMALL_STATE(5256)] = 214340,
  [SMALL_STATE(5257)] = 214370,
  [SMALL_STATE(5258)] = 214394,
  [SMALL_STATE(5259)] = 214420,
  [SMALL_STATE(5260)] = 214458,
  [SMALL_STATE(5261)] = 214496,
  [SMALL_STATE(5262)] = 214520,
  [SMALL_STATE(5263)] = 214550,
  [SMALL_STATE(5264)] = 214574,
  [SMALL_STATE(5265)] = 214598,
  [SMALL_STATE(5266)] = 214628,
  [SMALL_STATE(5267)] = 214652,
  [SMALL_STATE(5268)] = 214676,
  [SMALL_STATE(5269)] = 214706,
  [SMALL_STATE(5270)] = 214736,
  [SMALL_STATE(5271)] = 214766,
  [SMALL_STATE(5272)] = 214790,
  [SMALL_STATE(5273)] = 214814,
  [SMALL_STATE(5274)] = 214844,
  [SMALL_STATE(5275)] = 214874,
  [SMALL_STATE(5276)] = 214910,
  [SMALL_STATE(5277)] = 214934,
  [SMALL_STATE(5278)] = 214964,
  [SMALL_STATE(5279)] = 214988,
  [SMALL_STATE(5280)] = 215012,
  [SMALL_STATE(5281)] = 215050,
  [SMALL_STATE(5282)] = 215080,
  [SMALL_STATE(5283)] = 215104,
  [SMALL_STATE(5284)] = 215128,
  [SMALL_STATE(5285)] = 215158,
  [SMALL_STATE(5286)] = 215182,
  [SMALL_STATE(5287)] = 215212,
  [SMALL_STATE(5288)] = 215236,
  [SMALL_STATE(5289)] = 215260,
  [SMALL_STATE(5290)] = 215282,
  [SMALL_STATE(5291)] = 215312,
  [SMALL_STATE(5292)] = 215336,
  [SMALL_STATE(5293)] = 215360,
  [SMALL_STATE(5294)] = 215398,
  [SMALL_STATE(5295)] = 215422,
  [SMALL_STATE(5296)] = 215452,
  [SMALL_STATE(5297)] = 215490,
  [SMALL_STATE(5298)] = 215518,
  [SMALL_STATE(5299)] = 215548,
  [SMALL_STATE(5300)] = 215578,
  [SMALL_STATE(5301)] = 215608,
  [SMALL_STATE(5302)] = 215638,
  [SMALL_STATE(5303)] = 215668,
  [SMALL_STATE(5304)] = 215698,
  [SMALL_STATE(5305)] = 215728,
  [SMALL_STATE(5306)] = 215752,
  [SMALL_STATE(5307)] = 215782,
  [SMALL_STATE(5308)] = 215810,
  [SMALL_STATE(5309)] = 215838,
  [SMALL_STATE(5310)] = 215876,
  [SMALL_STATE(5311)] = 215900,
  [SMALL_STATE(5312)] = 215924,
  [SMALL_STATE(5313)] = 215948,
  [SMALL_STATE(5314)] = 215986,
  [SMALL_STATE(5315)] = 216010,
  [SMALL_STATE(5316)] = 216034,
  [SMALL_STATE(5317)] = 216056,
  [SMALL_STATE(5318)] = 216094,
  [SMALL_STATE(5319)] = 216118,
  [SMALL_STATE(5320)] = 216142,
  [SMALL_STATE(5321)] = 216166,
  [SMALL_STATE(5322)] = 216190,
  [SMALL_STATE(5323)] = 216226,
  [SMALL_STATE(5324)] = 216262,
  [SMALL_STATE(5325)] = 216286,
  [SMALL_STATE(5326)] = 216310,
  [SMALL_STATE(5327)] = 216332,
  [SMALL_STATE(5328)] = 216356,
  [SMALL_STATE(5329)] = 216378,
  [SMALL_STATE(5330)] = 216402,
  [SMALL_STATE(5331)] = 216440,
  [SMALL_STATE(5332)] = 216462,
  [SMALL_STATE(5333)] = 216500,
  [SMALL_STATE(5334)] = 216538,
  [SMALL_STATE(5335)] = 216576,
  [SMALL_STATE(5336)] = 216606,
  [SMALL_STATE(5337)] = 216644,
  [SMALL_STATE(5338)] = 216674,
  [SMALL_STATE(5339)] = 216705,
  [SMALL_STATE(5340)] = 216736,
  [SMALL_STATE(5341)] = 216763,
  [SMALL_STATE(5342)] = 216794,
  [SMALL_STATE(5343)] = 216827,
  [SMALL_STATE(5344)] = 216852,
  [SMALL_STATE(5345)] = 216873,
  [SMALL_STATE(5346)] = 216894,
  [SMALL_STATE(5347)] = 216921,
  [SMALL_STATE(5348)] = 216952,
  [SMALL_STATE(5349)] = 216985,
  [SMALL_STATE(5350)] = 217016,
  [SMALL_STATE(5351)] = 217049,
  [SMALL_STATE(5352)] = 217082,
  [SMALL_STATE(5353)] = 217115,
  [SMALL_STATE(5354)] = 217142,
  [SMALL_STATE(5355)] = 217169,
  [SMALL_STATE(5356)] = 217196,
  [SMALL_STATE(5357)] = 217217,
  [SMALL_STATE(5358)] = 217238,
  [SMALL_STATE(5359)] = 217269,
  [SMALL_STATE(5360)] = 217294,
  [SMALL_STATE(5361)] = 217329,
  [SMALL_STATE(5362)] = 217360,
  [SMALL_STATE(5363)] = 217391,
  [SMALL_STATE(5364)] = 217412,
  [SMALL_STATE(5365)] = 217445,
  [SMALL_STATE(5366)] = 217476,
  [SMALL_STATE(5367)] = 217507,
  [SMALL_STATE(5368)] = 217538,
  [SMALL_STATE(5369)] = 217571,
  [SMALL_STATE(5370)] = 217596,
  [SMALL_STATE(5371)] = 217629,
  [SMALL_STATE(5372)] = 217652,
  [SMALL_STATE(5373)] = 217685,
  [SMALL_STATE(5374)] = 217708,
  [SMALL_STATE(5375)] = 217743,
  [SMALL_STATE(5376)] = 217776,
  [SMALL_STATE(5377)] = 217809,
  [SMALL_STATE(5378)] = 217844,
  [SMALL_STATE(5379)] = 217877,
  [SMALL_STATE(5380)] = 217910,
  [SMALL_STATE(5381)] = 217943,
  [SMALL_STATE(5382)] = 217974,
  [SMALL_STATE(5383)] = 217997,
  [SMALL_STATE(5384)] = 218030,
  [SMALL_STATE(5385)] = 218061,
  [SMALL_STATE(5386)] = 218092,
  [SMALL_STATE(5387)] = 218125,
  [SMALL_STATE(5388)] = 218156,
  [SMALL_STATE(5389)] = 218191,
  [SMALL_STATE(5390)] = 218222,
  [SMALL_STATE(5391)] = 218247,
  [SMALL_STATE(5392)] = 218280,
  [SMALL_STATE(5393)] = 218313,
  [SMALL_STATE(5394)] = 218344,
  [SMALL_STATE(5395)] = 218377,
  [SMALL_STATE(5396)] = 218412,
  [SMALL_STATE(5397)] = 218443,
  [SMALL_STATE(5398)] = 218474,
  [SMALL_STATE(5399)] = 218505,
  [SMALL_STATE(5400)] = 218526,
  [SMALL_STATE(5401)] = 218558,
  [SMALL_STATE(5402)] = 218580,
  [SMALL_STATE(5403)] = 218602,
  [SMALL_STATE(5404)] = 218634,
  [SMALL_STATE(5405)] = 218666,
  [SMALL_STATE(5406)] = 218688,
  [SMALL_STATE(5407)] = 218718,
  [SMALL_STATE(5408)] = 218750,
  [SMALL_STATE(5409)] = 218782,
  [SMALL_STATE(5410)] = 218812,
  [SMALL_STATE(5411)] = 218844,
  [SMALL_STATE(5412)] = 218876,
  [SMALL_STATE(5413)] = 218902,
  [SMALL_STATE(5414)] = 218924,
  [SMALL_STATE(5415)] = 218956,
  [SMALL_STATE(5416)] = 218988,
  [SMALL_STATE(5417)] = 219010,
  [SMALL_STATE(5418)] = 219042,
  [SMALL_STATE(5419)] = 219066,
  [SMALL_STATE(5420)] = 219098,
  [SMALL_STATE(5421)] = 219122,
  [SMALL_STATE(5422)] = 219154,
  [SMALL_STATE(5423)] = 219186,
  [SMALL_STATE(5424)] = 219218,
  [SMALL_STATE(5425)] = 219250,
  [SMALL_STATE(5426)] = 219282,
  [SMALL_STATE(5427)] = 219314,
  [SMALL_STATE(5428)] = 219336,
  [SMALL_STATE(5429)] = 219368,
  [SMALL_STATE(5430)] = 219400,
  [SMALL_STATE(5431)] = 219424,
  [SMALL_STATE(5432)] = 219456,
  [SMALL_STATE(5433)] = 219488,
  [SMALL_STATE(5434)] = 219520,
  [SMALL_STATE(5435)] = 219550,
  [SMALL_STATE(5436)] = 219576,
  [SMALL_STATE(5437)] = 219608,
  [SMALL_STATE(5438)] = 219640,
  [SMALL_STATE(5439)] = 219672,
  [SMALL_STATE(5440)] = 219704,
  [SMALL_STATE(5441)] = 219736,
  [SMALL_STATE(5442)] = 219758,
  [SMALL_STATE(5443)] = 219790,
  [SMALL_STATE(5444)] = 219822,
  [SMALL_STATE(5445)] = 219854,
  [SMALL_STATE(5446)] = 219886,
  [SMALL_STATE(5447)] = 219918,
  [SMALL_STATE(5448)] = 219950,
  [SMALL_STATE(5449)] = 219974,
  [SMALL_STATE(5450)] = 220006,
  [SMALL_STATE(5451)] = 220038,
  [SMALL_STATE(5452)] = 220060,
  [SMALL_STATE(5453)] = 220092,
  [SMALL_STATE(5454)] = 220124,
  [SMALL_STATE(5455)] = 220156,
  [SMALL_STATE(5456)] = 220188,
  [SMALL_STATE(5457)] = 220216,
  [SMALL_STATE(5458)] = 220238,
  [SMALL_STATE(5459)] = 220262,
  [SMALL_STATE(5460)] = 220284,
  [SMALL_STATE(5461)] = 220310,
  [SMALL_STATE(5462)] = 220342,
  [SMALL_STATE(5463)] = 220364,
  [SMALL_STATE(5464)] = 220396,
  [SMALL_STATE(5465)] = 220428,
  [SMALL_STATE(5466)] = 220460,
  [SMALL_STATE(5467)] = 220482,
  [SMALL_STATE(5468)] = 220504,
  [SMALL_STATE(5469)] = 220536,
  [SMALL_STATE(5470)] = 220568,
  [SMALL_STATE(5471)] = 220600,
  [SMALL_STATE(5472)] = 220620,
  [SMALL_STATE(5473)] = 220652,
  [SMALL_STATE(5474)] = 220684,
  [SMALL_STATE(5475)] = 220716,
  [SMALL_STATE(5476)] = 220748,
  [SMALL_STATE(5477)] = 220768,
  [SMALL_STATE(5478)] = 220800,
  [SMALL_STATE(5479)] = 220832,
  [SMALL_STATE(5480)] = 220854,
  [SMALL_STATE(5481)] = 220886,
  [SMALL_STATE(5482)] = 220918,
  [SMALL_STATE(5483)] = 220950,
  [SMALL_STATE(5484)] = 220982,
  [SMALL_STATE(5485)] = 221014,
  [SMALL_STATE(5486)] = 221046,
  [SMALL_STATE(5487)] = 221078,
  [SMALL_STATE(5488)] = 221110,
  [SMALL_STATE(5489)] = 221132,
  [SMALL_STATE(5490)] = 221164,
  [SMALL_STATE(5491)] = 221196,
  [SMALL_STATE(5492)] = 221228,
  [SMALL_STATE(5493)] = 221250,
  [SMALL_STATE(5494)] = 221270,
  [SMALL_STATE(5495)] = 221294,
  [SMALL_STATE(5496)] = 221314,
  [SMALL_STATE(5497)] = 221346,
  [SMALL_STATE(5498)] = 221366,
  [SMALL_STATE(5499)] = 221396,
  [SMALL_STATE(5500)] = 221428,
  [SMALL_STATE(5501)] = 221456,
  [SMALL_STATE(5502)] = 221480,
  [SMALL_STATE(5503)] = 221502,
  [SMALL_STATE(5504)] = 221528,
  [SMALL_STATE(5505)] = 221550,
  [SMALL_STATE(5506)] = 221570,
  [SMALL_STATE(5507)] = 221602,
  [SMALL_STATE(5508)] = 221626,
  [SMALL_STATE(5509)] = 221658,
  [SMALL_STATE(5510)] = 221682,
  [SMALL_STATE(5511)] = 221714,
  [SMALL_STATE(5512)] = 221746,
  [SMALL_STATE(5513)] = 221770,
  [SMALL_STATE(5514)] = 221794,
  [SMALL_STATE(5515)] = 221826,
  [SMALL_STATE(5516)] = 221858,
  [SMALL_STATE(5517)] = 221882,
  [SMALL_STATE(5518)] = 221910,
  [SMALL_STATE(5519)] = 221930,
  [SMALL_STATE(5520)] = 221954,
  [SMALL_STATE(5521)] = 221978,
  [SMALL_STATE(5522)] = 222010,
  [SMALL_STATE(5523)] = 222042,
  [SMALL_STATE(5524)] = 222066,
  [SMALL_STATE(5525)] = 222090,
  [SMALL_STATE(5526)] = 222110,
  [SMALL_STATE(5527)] = 222142,
  [SMALL_STATE(5528)] = 222174,
  [SMALL_STATE(5529)] = 222206,
  [SMALL_STATE(5530)] = 222238,
  [SMALL_STATE(5531)] = 222266,
  [SMALL_STATE(5532)] = 222298,
  [SMALL_STATE(5533)] = 222326,
  [SMALL_STATE(5534)] = 222358,
  [SMALL_STATE(5535)] = 222390,
  [SMALL_STATE(5536)] = 222422,
  [SMALL_STATE(5537)] = 222454,
  [SMALL_STATE(5538)] = 222486,
  [SMALL_STATE(5539)] = 222518,
  [SMALL_STATE(5540)] = 222550,
  [SMALL_STATE(5541)] = 222580,
  [SMALL_STATE(5542)] = 222610,
  [SMALL_STATE(5543)] = 222639,
  [SMALL_STATE(5544)] = 222660,
  [SMALL_STATE(5545)] = 222689,
  [SMALL_STATE(5546)] = 222716,
  [SMALL_STATE(5547)] = 222737,
  [SMALL_STATE(5548)] = 222758,
  [SMALL_STATE(5549)] = 222779,
  [SMALL_STATE(5550)] = 222808,
  [SMALL_STATE(5551)] = 222835,
  [SMALL_STATE(5552)] = 222856,
  [SMALL_STATE(5553)] = 222885,
  [SMALL_STATE(5554)] = 222912,
  [SMALL_STATE(5555)] = 222937,
  [SMALL_STATE(5556)] = 222966,
  [SMALL_STATE(5557)] = 222995,
  [SMALL_STATE(5558)] = 223016,
  [SMALL_STATE(5559)] = 223037,
  [SMALL_STATE(5560)] = 223066,
  [SMALL_STATE(5561)] = 223087,
  [SMALL_STATE(5562)] = 223108,
  [SMALL_STATE(5563)] = 223137,
  [SMALL_STATE(5564)] = 223166,
  [SMALL_STATE(5565)] = 223187,
  [SMALL_STATE(5566)] = 223208,
  [SMALL_STATE(5567)] = 223237,
  [SMALL_STATE(5568)] = 223258,
  [SMALL_STATE(5569)] = 223285,
  [SMALL_STATE(5570)] = 223306,
  [SMALL_STATE(5571)] = 223327,
  [SMALL_STATE(5572)] = 223356,
  [SMALL_STATE(5573)] = 223383,
  [SMALL_STATE(5574)] = 223410,
  [SMALL_STATE(5575)] = 223437,
  [SMALL_STATE(5576)] = 223466,
  [SMALL_STATE(5577)] = 223495,
  [SMALL_STATE(5578)] = 223524,
  [SMALL_STATE(5579)] = 223551,
  [SMALL_STATE(5580)] = 223580,
  [SMALL_STATE(5581)] = 223607,
  [SMALL_STATE(5582)] = 223628,
  [SMALL_STATE(5583)] = 223649,
  [SMALL_STATE(5584)] = 223670,
  [SMALL_STATE(5585)] = 223691,
  [SMALL_STATE(5586)] = 223712,
  [SMALL_STATE(5587)] = 223733,
  [SMALL_STATE(5588)] = 223754,
  [SMALL_STATE(5589)] = 223775,
  [SMALL_STATE(5590)] = 223804,
  [SMALL_STATE(5591)] = 223825,
  [SMALL_STATE(5592)] = 223846,
  [SMALL_STATE(5593)] = 223867,
  [SMALL_STATE(5594)] = 223896,
  [SMALL_STATE(5595)] = 223923,
  [SMALL_STATE(5596)] = 223944,
  [SMALL_STATE(5597)] = 223965,
  [SMALL_STATE(5598)] = 223986,
  [SMALL_STATE(5599)] = 224007,
  [SMALL_STATE(5600)] = 224036,
  [SMALL_STATE(5601)] = 224065,
  [SMALL_STATE(5602)] = 224094,
  [SMALL_STATE(5603)] = 224117,
  [SMALL_STATE(5604)] = 224138,
  [SMALL_STATE(5605)] = 224159,
  [SMALL_STATE(5606)] = 224180,
  [SMALL_STATE(5607)] = 224201,
  [SMALL_STATE(5608)] = 224230,
  [SMALL_STATE(5609)] = 224251,
  [SMALL_STATE(5610)] = 224272,
  [SMALL_STATE(5611)] = 224301,
  [SMALL_STATE(5612)] = 224322,
  [SMALL_STATE(5613)] = 224351,
  [SMALL_STATE(5614)] = 224380,
  [SMALL_STATE(5615)] = 224409,
  [SMALL_STATE(5616)] = 224436,
  [SMALL_STATE(5617)] = 224463,
  [SMALL_STATE(5618)] = 224484,
  [SMALL_STATE(5619)] = 224513,
  [SMALL_STATE(5620)] = 224540,
  [SMALL_STATE(5621)] = 224569,
  [SMALL_STATE(5622)] = 224598,
  [SMALL_STATE(5623)] = 224627,
  [SMALL_STATE(5624)] = 224656,
  [SMALL_STATE(5625)] = 224677,
  [SMALL_STATE(5626)] = 224706,
  [SMALL_STATE(5627)] = 224727,
  [SMALL_STATE(5628)] = 224756,
  [SMALL_STATE(5629)] = 224785,
  [SMALL_STATE(5630)] = 224808,
  [SMALL_STATE(5631)] = 224829,
  [SMALL_STATE(5632)] = 224850,
  [SMALL_STATE(5633)] = 224879,
  [SMALL_STATE(5634)] = 224900,
  [SMALL_STATE(5635)] = 224921,
  [SMALL_STATE(5636)] = 224944,
  [SMALL_STATE(5637)] = 224965,
  [SMALL_STATE(5638)] = 224986,
  [SMALL_STATE(5639)] = 225007,
  [SMALL_STATE(5640)] = 225036,
  [SMALL_STATE(5641)] = 225057,
  [SMALL_STATE(5642)] = 225084,
  [SMALL_STATE(5643)] = 225105,
  [SMALL_STATE(5644)] = 225134,
  [SMALL_STATE(5645)] = 225161,
  [SMALL_STATE(5646)] = 225190,
  [SMALL_STATE(5647)] = 225219,
  [SMALL_STATE(5648)] = 225240,
  [SMALL_STATE(5649)] = 225269,
  [SMALL_STATE(5650)] = 225290,
  [SMALL_STATE(5651)] = 225311,
  [SMALL_STATE(5652)] = 225340,
  [SMALL_STATE(5653)] = 225369,
  [SMALL_STATE(5654)] = 225390,
  [SMALL_STATE(5655)] = 225419,
  [SMALL_STATE(5656)] = 225440,
  [SMALL_STATE(5657)] = 225461,
  [SMALL_STATE(5658)] = 225486,
  [SMALL_STATE(5659)] = 225509,
  [SMALL_STATE(5660)] = 225538,
  [SMALL_STATE(5661)] = 225559,
  [SMALL_STATE(5662)] = 225588,
  [SMALL_STATE(5663)] = 225617,
  [SMALL_STATE(5664)] = 225644,
  [SMALL_STATE(5665)] = 225673,
  [SMALL_STATE(5666)] = 225702,
  [SMALL_STATE(5667)] = 225723,
  [SMALL_STATE(5668)] = 225746,
  [SMALL_STATE(5669)] = 225773,
  [SMALL_STATE(5670)] = 225802,
  [SMALL_STATE(5671)] = 225831,
  [SMALL_STATE(5672)] = 225858,
  [SMALL_STATE(5673)] = 225887,
  [SMALL_STATE(5674)] = 225916,
  [SMALL_STATE(5675)] = 225945,
  [SMALL_STATE(5676)] = 225972,
  [SMALL_STATE(5677)] = 226001,
  [SMALL_STATE(5678)] = 226030,
  [SMALL_STATE(5679)] = 226051,
  [SMALL_STATE(5680)] = 226080,
  [SMALL_STATE(5681)] = 226107,
  [SMALL_STATE(5682)] = 226128,
  [SMALL_STATE(5683)] = 226157,
  [SMALL_STATE(5684)] = 226186,
  [SMALL_STATE(5685)] = 226213,
  [SMALL_STATE(5686)] = 226234,
  [SMALL_STATE(5687)] = 226261,
  [SMALL_STATE(5688)] = 226290,
  [SMALL_STATE(5689)] = 226319,
  [SMALL_STATE(5690)] = 226348,
  [SMALL_STATE(5691)] = 226369,
  [SMALL_STATE(5692)] = 226392,
  [SMALL_STATE(5693)] = 226421,
  [SMALL_STATE(5694)] = 226448,
  [SMALL_STATE(5695)] = 226471,
  [SMALL_STATE(5696)] = 226494,
  [SMALL_STATE(5697)] = 226521,
  [SMALL_STATE(5698)] = 226550,
  [SMALL_STATE(5699)] = 226579,
  [SMALL_STATE(5700)] = 226606,
  [SMALL_STATE(5701)] = 226635,
  [SMALL_STATE(5702)] = 226662,
  [SMALL_STATE(5703)] = 226683,
  [SMALL_STATE(5704)] = 226704,
  [SMALL_STATE(5705)] = 226731,
  [SMALL_STATE(5706)] = 226760,
  [SMALL_STATE(5707)] = 226789,
  [SMALL_STATE(5708)] = 226816,
  [SMALL_STATE(5709)] = 226837,
  [SMALL_STATE(5710)] = 226858,
  [SMALL_STATE(5711)] = 226883,
  [SMALL_STATE(5712)] = 226906,
  [SMALL_STATE(5713)] = 226927,
  [SMALL_STATE(5714)] = 226952,
  [SMALL_STATE(5715)] = 226973,
  [SMALL_STATE(5716)] = 226994,
  [SMALL_STATE(5717)] = 227017,
  [SMALL_STATE(5718)] = 227038,
  [SMALL_STATE(5719)] = 227067,
  [SMALL_STATE(5720)] = 227088,
  [SMALL_STATE(5721)] = 227117,
  [SMALL_STATE(5722)] = 227144,
  [SMALL_STATE(5723)] = 227165,
  [SMALL_STATE(5724)] = 227190,
  [SMALL_STATE(5725)] = 227213,
  [SMALL_STATE(5726)] = 227234,
  [SMALL_STATE(5727)] = 227259,
  [SMALL_STATE(5728)] = 227280,
  [SMALL_STATE(5729)] = 227301,
  [SMALL_STATE(5730)] = 227324,
  [SMALL_STATE(5731)] = 227345,
  [SMALL_STATE(5732)] = 227372,
  [SMALL_STATE(5733)] = 227399,
  [SMALL_STATE(5734)] = 227426,
  [SMALL_STATE(5735)] = 227455,
  [SMALL_STATE(5736)] = 227476,
  [SMALL_STATE(5737)] = 227503,
  [SMALL_STATE(5738)] = 227530,
  [SMALL_STATE(5739)] = 227557,
  [SMALL_STATE(5740)] = 227582,
  [SMALL_STATE(5741)] = 227611,
  [SMALL_STATE(5742)] = 227634,
  [SMALL_STATE(5743)] = 227657,
  [SMALL_STATE(5744)] = 227678,
  [SMALL_STATE(5745)] = 227707,
  [SMALL_STATE(5746)] = 227734,
  [SMALL_STATE(5747)] = 227757,
  [SMALL_STATE(5748)] = 227786,
  [SMALL_STATE(5749)] = 227809,
  [SMALL_STATE(5750)] = 227838,
  [SMALL_STATE(5751)] = 227867,
  [SMALL_STATE(5752)] = 227896,
  [SMALL_STATE(5753)] = 227925,
  [SMALL_STATE(5754)] = 227946,
  [SMALL_STATE(5755)] = 227967,
  [SMALL_STATE(5756)] = 227996,
  [SMALL_STATE(5757)] = 228017,
  [SMALL_STATE(5758)] = 228044,
  [SMALL_STATE(5759)] = 228071,
  [SMALL_STATE(5760)] = 228100,
  [SMALL_STATE(5761)] = 228127,
  [SMALL_STATE(5762)] = 228156,
  [SMALL_STATE(5763)] = 228177,
  [SMALL_STATE(5764)] = 228206,
  [SMALL_STATE(5765)] = 228235,
  [SMALL_STATE(5766)] = 228262,
  [SMALL_STATE(5767)] = 228291,
  [SMALL_STATE(5768)] = 228320,
  [SMALL_STATE(5769)] = 228341,
  [SMALL_STATE(5770)] = 228368,
  [SMALL_STATE(5771)] = 228397,
  [SMALL_STATE(5772)] = 228426,
  [SMALL_STATE(5773)] = 228455,
  [SMALL_STATE(5774)] = 228476,
  [SMALL_STATE(5775)] = 228497,
  [SMALL_STATE(5776)] = 228520,
  [SMALL_STATE(5777)] = 228541,
  [SMALL_STATE(5778)] = 228564,
  [SMALL_STATE(5779)] = 228593,
  [SMALL_STATE(5780)] = 228622,
  [SMALL_STATE(5781)] = 228643,
  [SMALL_STATE(5782)] = 228668,
  [SMALL_STATE(5783)] = 228695,
  [SMALL_STATE(5784)] = 228722,
  [SMALL_STATE(5785)] = 228743,
  [SMALL_STATE(5786)] = 228772,
  [SMALL_STATE(5787)] = 228801,
  [SMALL_STATE(5788)] = 228825,
  [SMALL_STATE(5789)] = 228849,
  [SMALL_STATE(5790)] = 228873,
  [SMALL_STATE(5791)] = 228897,
  [SMALL_STATE(5792)] = 228923,
  [SMALL_STATE(5793)] = 228947,
  [SMALL_STATE(5794)] = 228971,
  [SMALL_STATE(5795)] = 228997,
  [SMALL_STATE(5796)] = 229021,
  [SMALL_STATE(5797)] = 229045,
  [SMALL_STATE(5798)] = 229069,
  [SMALL_STATE(5799)] = 229093,
  [SMALL_STATE(5800)] = 229117,
  [SMALL_STATE(5801)] = 229141,
  [SMALL_STATE(5802)] = 229167,
  [SMALL_STATE(5803)] = 229191,
  [SMALL_STATE(5804)] = 229215,
  [SMALL_STATE(5805)] = 229239,
  [SMALL_STATE(5806)] = 229263,
  [SMALL_STATE(5807)] = 229289,
  [SMALL_STATE(5808)] = 229315,
  [SMALL_STATE(5809)] = 229339,
  [SMALL_STATE(5810)] = 229357,
  [SMALL_STATE(5811)] = 229381,
  [SMALL_STATE(5812)] = 229405,
  [SMALL_STATE(5813)] = 229429,
  [SMALL_STATE(5814)] = 229453,
  [SMALL_STATE(5815)] = 229477,
  [SMALL_STATE(5816)] = 229501,
  [SMALL_STATE(5817)] = 229525,
  [SMALL_STATE(5818)] = 229551,
  [SMALL_STATE(5819)] = 229575,
  [SMALL_STATE(5820)] = 229599,
  [SMALL_STATE(5821)] = 229625,
  [SMALL_STATE(5822)] = 229647,
  [SMALL_STATE(5823)] = 229667,
  [SMALL_STATE(5824)] = 229687,
  [SMALL_STATE(5825)] = 229711,
  [SMALL_STATE(5826)] = 229735,
  [SMALL_STATE(5827)] = 229759,
  [SMALL_STATE(5828)] = 229783,
  [SMALL_STATE(5829)] = 229807,
  [SMALL_STATE(5830)] = 229831,
  [SMALL_STATE(5831)] = 229855,
  [SMALL_STATE(5832)] = 229879,
  [SMALL_STATE(5833)] = 229905,
  [SMALL_STATE(5834)] = 229929,
  [SMALL_STATE(5835)] = 229951,
  [SMALL_STATE(5836)] = 229977,
  [SMALL_STATE(5837)] = 230003,
  [SMALL_STATE(5838)] = 230027,
  [SMALL_STATE(5839)] = 230053,
  [SMALL_STATE(5840)] = 230079,
  [SMALL_STATE(5841)] = 230105,
  [SMALL_STATE(5842)] = 230131,
  [SMALL_STATE(5843)] = 230155,
  [SMALL_STATE(5844)] = 230181,
  [SMALL_STATE(5845)] = 230207,
  [SMALL_STATE(5846)] = 230233,
  [SMALL_STATE(5847)] = 230255,
  [SMALL_STATE(5848)] = 230279,
  [SMALL_STATE(5849)] = 230305,
  [SMALL_STATE(5850)] = 230331,
  [SMALL_STATE(5851)] = 230355,
  [SMALL_STATE(5852)] = 230379,
  [SMALL_STATE(5853)] = 230405,
  [SMALL_STATE(5854)] = 230429,
  [SMALL_STATE(5855)] = 230455,
  [SMALL_STATE(5856)] = 230479,
  [SMALL_STATE(5857)] = 230505,
  [SMALL_STATE(5858)] = 230523,
  [SMALL_STATE(5859)] = 230547,
  [SMALL_STATE(5860)] = 230571,
  [SMALL_STATE(5861)] = 230597,
  [SMALL_STATE(5862)] = 230621,
  [SMALL_STATE(5863)] = 230639,
  [SMALL_STATE(5864)] = 230663,
  [SMALL_STATE(5865)] = 230689,
  [SMALL_STATE(5866)] = 230713,
  [SMALL_STATE(5867)] = 230739,
  [SMALL_STATE(5868)] = 230765,
  [SMALL_STATE(5869)] = 230783,
  [SMALL_STATE(5870)] = 230807,
  [SMALL_STATE(5871)] = 230833,
  [SMALL_STATE(5872)] = 230859,
  [SMALL_STATE(5873)] = 230885,
  [SMALL_STATE(5874)] = 230909,
  [SMALL_STATE(5875)] = 230933,
  [SMALL_STATE(5876)] = 230959,
  [SMALL_STATE(5877)] = 230983,
  [SMALL_STATE(5878)] = 231009,
  [SMALL_STATE(5879)] = 231031,
  [SMALL_STATE(5880)] = 231055,
  [SMALL_STATE(5881)] = 231081,
  [SMALL_STATE(5882)] = 231105,
  [SMALL_STATE(5883)] = 231129,
  [SMALL_STATE(5884)] = 231155,
  [SMALL_STATE(5885)] = 231181,
  [SMALL_STATE(5886)] = 231205,
  [SMALL_STATE(5887)] = 231231,
  [SMALL_STATE(5888)] = 231255,
  [SMALL_STATE(5889)] = 231273,
  [SMALL_STATE(5890)] = 231297,
  [SMALL_STATE(5891)] = 231323,
  [SMALL_STATE(5892)] = 231341,
  [SMALL_STATE(5893)] = 231367,
  [SMALL_STATE(5894)] = 231393,
  [SMALL_STATE(5895)] = 231415,
  [SMALL_STATE(5896)] = 231441,
  [SMALL_STATE(5897)] = 231463,
  [SMALL_STATE(5898)] = 231487,
  [SMALL_STATE(5899)] = 231511,
  [SMALL_STATE(5900)] = 231537,
  [SMALL_STATE(5901)] = 231561,
  [SMALL_STATE(5902)] = 231587,
  [SMALL_STATE(5903)] = 231613,
  [SMALL_STATE(5904)] = 231637,
  [SMALL_STATE(5905)] = 231655,
  [SMALL_STATE(5906)] = 231679,
  [SMALL_STATE(5907)] = 231705,
  [SMALL_STATE(5908)] = 231731,
  [SMALL_STATE(5909)] = 231749,
  [SMALL_STATE(5910)] = 231773,
  [SMALL_STATE(5911)] = 231797,
  [SMALL_STATE(5912)] = 231821,
  [SMALL_STATE(5913)] = 231845,
  [SMALL_STATE(5914)] = 231865,
  [SMALL_STATE(5915)] = 231891,
  [SMALL_STATE(5916)] = 231915,
  [SMALL_STATE(5917)] = 231939,
  [SMALL_STATE(5918)] = 231965,
  [SMALL_STATE(5919)] = 231989,
  [SMALL_STATE(5920)] = 232015,
  [SMALL_STATE(5921)] = 232039,
  [SMALL_STATE(5922)] = 232061,
  [SMALL_STATE(5923)] = 232085,
  [SMALL_STATE(5924)] = 232109,
  [SMALL_STATE(5925)] = 232133,
  [SMALL_STATE(5926)] = 232151,
  [SMALL_STATE(5927)] = 232175,
  [SMALL_STATE(5928)] = 232197,
  [SMALL_STATE(5929)] = 232223,
  [SMALL_STATE(5930)] = 232249,
  [SMALL_STATE(5931)] = 232273,
  [SMALL_STATE(5932)] = 232295,
  [SMALL_STATE(5933)] = 232319,
  [SMALL_STATE(5934)] = 232343,
  [SMALL_STATE(5935)] = 232367,
  [SMALL_STATE(5936)] = 232391,
  [SMALL_STATE(5937)] = 232413,
  [SMALL_STATE(5938)] = 232437,
  [SMALL_STATE(5939)] = 232459,
  [SMALL_STATE(5940)] = 232485,
  [SMALL_STATE(5941)] = 232511,
  [SMALL_STATE(5942)] = 232535,
  [SMALL_STATE(5943)] = 232561,
  [SMALL_STATE(5944)] = 232585,
  [SMALL_STATE(5945)] = 232609,
  [SMALL_STATE(5946)] = 232633,
  [SMALL_STATE(5947)] = 232659,
  [SMALL_STATE(5948)] = 232677,
  [SMALL_STATE(5949)] = 232701,
  [SMALL_STATE(5950)] = 232725,
  [SMALL_STATE(5951)] = 232749,
  [SMALL_STATE(5952)] = 232773,
  [SMALL_STATE(5953)] = 232797,
  [SMALL_STATE(5954)] = 232819,
  [SMALL_STATE(5955)] = 232843,
  [SMALL_STATE(5956)] = 232869,
  [SMALL_STATE(5957)] = 232893,
  [SMALL_STATE(5958)] = 232919,
  [SMALL_STATE(5959)] = 232943,
  [SMALL_STATE(5960)] = 232967,
  [SMALL_STATE(5961)] = 232991,
  [SMALL_STATE(5962)] = 233015,
  [SMALL_STATE(5963)] = 233041,
  [SMALL_STATE(5964)] = 233067,
  [SMALL_STATE(5965)] = 233091,
  [SMALL_STATE(5966)] = 233115,
  [SMALL_STATE(5967)] = 233141,
  [SMALL_STATE(5968)] = 233163,
  [SMALL_STATE(5969)] = 233185,
  [SMALL_STATE(5970)] = 233207,
  [SMALL_STATE(5971)] = 233231,
  [SMALL_STATE(5972)] = 233257,
  [SMALL_STATE(5973)] = 233281,
  [SMALL_STATE(5974)] = 233305,
  [SMALL_STATE(5975)] = 233329,
  [SMALL_STATE(5976)] = 233353,
  [SMALL_STATE(5977)] = 233377,
  [SMALL_STATE(5978)] = 233401,
  [SMALL_STATE(5979)] = 233425,
  [SMALL_STATE(5980)] = 233449,
  [SMALL_STATE(5981)] = 233473,
  [SMALL_STATE(5982)] = 233497,
  [SMALL_STATE(5983)] = 233523,
  [SMALL_STATE(5984)] = 233549,
  [SMALL_STATE(5985)] = 233573,
  [SMALL_STATE(5986)] = 233597,
  [SMALL_STATE(5987)] = 233621,
  [SMALL_STATE(5988)] = 233645,
  [SMALL_STATE(5989)] = 233669,
  [SMALL_STATE(5990)] = 233693,
  [SMALL_STATE(5991)] = 233717,
  [SMALL_STATE(5992)] = 233743,
  [SMALL_STATE(5993)] = 233767,
  [SMALL_STATE(5994)] = 233793,
  [SMALL_STATE(5995)] = 233817,
  [SMALL_STATE(5996)] = 233843,
  [SMALL_STATE(5997)] = 233865,
  [SMALL_STATE(5998)] = 233887,
  [SMALL_STATE(5999)] = 233911,
  [SMALL_STATE(6000)] = 233937,
  [SMALL_STATE(6001)] = 233963,
  [SMALL_STATE(6002)] = 233987,
  [SMALL_STATE(6003)] = 234011,
  [SMALL_STATE(6004)] = 234037,
  [SMALL_STATE(6005)] = 234063,
  [SMALL_STATE(6006)] = 234089,
  [SMALL_STATE(6007)] = 234115,
  [SMALL_STATE(6008)] = 234141,
  [SMALL_STATE(6009)] = 234159,
  [SMALL_STATE(6010)] = 234183,
  [SMALL_STATE(6011)] = 234209,
  [SMALL_STATE(6012)] = 234233,
  [SMALL_STATE(6013)] = 234259,
  [SMALL_STATE(6014)] = 234282,
  [SMALL_STATE(6015)] = 234301,
  [SMALL_STATE(6016)] = 234324,
  [SMALL_STATE(6017)] = 234347,
  [SMALL_STATE(6018)] = 234370,
  [SMALL_STATE(6019)] = 234393,
  [SMALL_STATE(6020)] = 234416,
  [SMALL_STATE(6021)] = 234439,
  [SMALL_STATE(6022)] = 234460,
  [SMALL_STATE(6023)] = 234483,
  [SMALL_STATE(6024)] = 234506,
  [SMALL_STATE(6025)] = 234529,
  [SMALL_STATE(6026)] = 234550,
  [SMALL_STATE(6027)] = 234573,
  [SMALL_STATE(6028)] = 234596,
  [SMALL_STATE(6029)] = 234619,
  [SMALL_STATE(6030)] = 234642,
  [SMALL_STATE(6031)] = 234665,
  [SMALL_STATE(6032)] = 234688,
  [SMALL_STATE(6033)] = 234711,
  [SMALL_STATE(6034)] = 234732,
  [SMALL_STATE(6035)] = 234755,
  [SMALL_STATE(6036)] = 234778,
  [SMALL_STATE(6037)] = 234801,
  [SMALL_STATE(6038)] = 234824,
  [SMALL_STATE(6039)] = 234847,
  [SMALL_STATE(6040)] = 234870,
  [SMALL_STATE(6041)] = 234893,
  [SMALL_STATE(6042)] = 234916,
  [SMALL_STATE(6043)] = 234937,
  [SMALL_STATE(6044)] = 234956,
  [SMALL_STATE(6045)] = 234979,
  [SMALL_STATE(6046)] = 235002,
  [SMALL_STATE(6047)] = 235025,
  [SMALL_STATE(6048)] = 235048,
  [SMALL_STATE(6049)] = 235071,
  [SMALL_STATE(6050)] = 235094,
  [SMALL_STATE(6051)] = 235117,
  [SMALL_STATE(6052)] = 235140,
  [SMALL_STATE(6053)] = 235163,
  [SMALL_STATE(6054)] = 235186,
  [SMALL_STATE(6055)] = 235209,
  [SMALL_STATE(6056)] = 235228,
  [SMALL_STATE(6057)] = 235251,
  [SMALL_STATE(6058)] = 235274,
  [SMALL_STATE(6059)] = 235297,
  [SMALL_STATE(6060)] = 235318,
  [SMALL_STATE(6061)] = 235341,
  [SMALL_STATE(6062)] = 235364,
  [SMALL_STATE(6063)] = 235387,
  [SMALL_STATE(6064)] = 235410,
  [SMALL_STATE(6065)] = 235433,
  [SMALL_STATE(6066)] = 235456,
  [SMALL_STATE(6067)] = 235477,
  [SMALL_STATE(6068)] = 235500,
  [SMALL_STATE(6069)] = 235523,
  [SMALL_STATE(6070)] = 235546,
  [SMALL_STATE(6071)] = 235569,
  [SMALL_STATE(6072)] = 235592,
  [SMALL_STATE(6073)] = 235615,
  [SMALL_STATE(6074)] = 235638,
  [SMALL_STATE(6075)] = 235661,
  [SMALL_STATE(6076)] = 235684,
  [SMALL_STATE(6077)] = 235707,
  [SMALL_STATE(6078)] = 235730,
  [SMALL_STATE(6079)] = 235753,
  [SMALL_STATE(6080)] = 235776,
  [SMALL_STATE(6081)] = 235799,
  [SMALL_STATE(6082)] = 235822,
  [SMALL_STATE(6083)] = 235843,
  [SMALL_STATE(6084)] = 235866,
  [SMALL_STATE(6085)] = 235889,
  [SMALL_STATE(6086)] = 235912,
  [SMALL_STATE(6087)] = 235935,
  [SMALL_STATE(6088)] = 235958,
  [SMALL_STATE(6089)] = 235981,
  [SMALL_STATE(6090)] = 236002,
  [SMALL_STATE(6091)] = 236025,
  [SMALL_STATE(6092)] = 236048,
  [SMALL_STATE(6093)] = 236071,
  [SMALL_STATE(6094)] = 236094,
  [SMALL_STATE(6095)] = 236117,
  [SMALL_STATE(6096)] = 236140,
  [SMALL_STATE(6097)] = 236163,
  [SMALL_STATE(6098)] = 236186,
  [SMALL_STATE(6099)] = 236209,
  [SMALL_STATE(6100)] = 236232,
  [SMALL_STATE(6101)] = 236253,
  [SMALL_STATE(6102)] = 236270,
  [SMALL_STATE(6103)] = 236293,
  [SMALL_STATE(6104)] = 236314,
  [SMALL_STATE(6105)] = 236335,
  [SMALL_STATE(6106)] = 236358,
  [SMALL_STATE(6107)] = 236379,
  [SMALL_STATE(6108)] = 236402,
  [SMALL_STATE(6109)] = 236425,
  [SMALL_STATE(6110)] = 236448,
  [SMALL_STATE(6111)] = 236471,
  [SMALL_STATE(6112)] = 236488,
  [SMALL_STATE(6113)] = 236505,
  [SMALL_STATE(6114)] = 236522,
  [SMALL_STATE(6115)] = 236541,
  [SMALL_STATE(6116)] = 236564,
  [SMALL_STATE(6117)] = 236585,
  [SMALL_STATE(6118)] = 236608,
  [SMALL_STATE(6119)] = 236631,
  [SMALL_STATE(6120)] = 236654,
  [SMALL_STATE(6121)] = 236673,
  [SMALL_STATE(6122)] = 236694,
  [SMALL_STATE(6123)] = 236715,
  [SMALL_STATE(6124)] = 236738,
  [SMALL_STATE(6125)] = 236759,
  [SMALL_STATE(6126)] = 236782,
  [SMALL_STATE(6127)] = 236803,
  [SMALL_STATE(6128)] = 236826,
  [SMALL_STATE(6129)] = 236849,
  [SMALL_STATE(6130)] = 236872,
  [SMALL_STATE(6131)] = 236895,
  [SMALL_STATE(6132)] = 236918,
  [SMALL_STATE(6133)] = 236941,
  [SMALL_STATE(6134)] = 236958,
  [SMALL_STATE(6135)] = 236981,
  [SMALL_STATE(6136)] = 236998,
  [SMALL_STATE(6137)] = 237021,
  [SMALL_STATE(6138)] = 237044,
  [SMALL_STATE(6139)] = 237067,
  [SMALL_STATE(6140)] = 237090,
  [SMALL_STATE(6141)] = 237113,
  [SMALL_STATE(6142)] = 237136,
  [SMALL_STATE(6143)] = 237159,
  [SMALL_STATE(6144)] = 237182,
  [SMALL_STATE(6145)] = 237205,
  [SMALL_STATE(6146)] = 237226,
  [SMALL_STATE(6147)] = 237247,
  [SMALL_STATE(6148)] = 237270,
  [SMALL_STATE(6149)] = 237293,
  [SMALL_STATE(6150)] = 237316,
  [SMALL_STATE(6151)] = 237339,
  [SMALL_STATE(6152)] = 237362,
  [SMALL_STATE(6153)] = 237385,
  [SMALL_STATE(6154)] = 237406,
  [SMALL_STATE(6155)] = 237429,
  [SMALL_STATE(6156)] = 237450,
  [SMALL_STATE(6157)] = 237473,
  [SMALL_STATE(6158)] = 237494,
  [SMALL_STATE(6159)] = 237517,
  [SMALL_STATE(6160)] = 237540,
  [SMALL_STATE(6161)] = 237563,
  [SMALL_STATE(6162)] = 237580,
  [SMALL_STATE(6163)] = 237599,
  [SMALL_STATE(6164)] = 237618,
  [SMALL_STATE(6165)] = 237641,
  [SMALL_STATE(6166)] = 237664,
  [SMALL_STATE(6167)] = 237687,
  [SMALL_STATE(6168)] = 237710,
  [SMALL_STATE(6169)] = 237733,
  [SMALL_STATE(6170)] = 237756,
  [SMALL_STATE(6171)] = 237779,
  [SMALL_STATE(6172)] = 237802,
  [SMALL_STATE(6173)] = 237823,
  [SMALL_STATE(6174)] = 237846,
  [SMALL_STATE(6175)] = 237869,
  [SMALL_STATE(6176)] = 237888,
  [SMALL_STATE(6177)] = 237911,
  [SMALL_STATE(6178)] = 237932,
  [SMALL_STATE(6179)] = 237955,
  [SMALL_STATE(6180)] = 237978,
  [SMALL_STATE(6181)] = 238001,
  [SMALL_STATE(6182)] = 238020,
  [SMALL_STATE(6183)] = 238043,
  [SMALL_STATE(6184)] = 238060,
  [SMALL_STATE(6185)] = 238083,
  [SMALL_STATE(6186)] = 238106,
  [SMALL_STATE(6187)] = 238129,
  [SMALL_STATE(6188)] = 238146,
  [SMALL_STATE(6189)] = 238167,
  [SMALL_STATE(6190)] = 238184,
  [SMALL_STATE(6191)] = 238207,
  [SMALL_STATE(6192)] = 238230,
  [SMALL_STATE(6193)] = 238247,
  [SMALL_STATE(6194)] = 238270,
  [SMALL_STATE(6195)] = 238287,
  [SMALL_STATE(6196)] = 238310,
  [SMALL_STATE(6197)] = 238331,
  [SMALL_STATE(6198)] = 238352,
  [SMALL_STATE(6199)] = 238375,
  [SMALL_STATE(6200)] = 238398,
  [SMALL_STATE(6201)] = 238421,
  [SMALL_STATE(6202)] = 238444,
  [SMALL_STATE(6203)] = 238467,
  [SMALL_STATE(6204)] = 238490,
  [SMALL_STATE(6205)] = 238513,
  [SMALL_STATE(6206)] = 238536,
  [SMALL_STATE(6207)] = 238559,
  [SMALL_STATE(6208)] = 238582,
  [SMALL_STATE(6209)] = 238605,
  [SMALL_STATE(6210)] = 238626,
  [SMALL_STATE(6211)] = 238649,
  [SMALL_STATE(6212)] = 238672,
  [SMALL_STATE(6213)] = 238691,
  [SMALL_STATE(6214)] = 238714,
  [SMALL_STATE(6215)] = 238735,
  [SMALL_STATE(6216)] = 238756,
  [SMALL_STATE(6217)] = 238779,
  [SMALL_STATE(6218)] = 238802,
  [SMALL_STATE(6219)] = 238821,
  [SMALL_STATE(6220)] = 238844,
  [SMALL_STATE(6221)] = 238867,
  [SMALL_STATE(6222)] = 238890,
  [SMALL_STATE(6223)] = 238913,
  [SMALL_STATE(6224)] = 238936,
  [SMALL_STATE(6225)] = 238955,
  [SMALL_STATE(6226)] = 238978,
  [SMALL_STATE(6227)] = 239001,
  [SMALL_STATE(6228)] = 239020,
  [SMALL_STATE(6229)] = 239043,
  [SMALL_STATE(6230)] = 239066,
  [SMALL_STATE(6231)] = 239083,
  [SMALL_STATE(6232)] = 239106,
  [SMALL_STATE(6233)] = 239129,
  [SMALL_STATE(6234)] = 239152,
  [SMALL_STATE(6235)] = 239175,
  [SMALL_STATE(6236)] = 239198,
  [SMALL_STATE(6237)] = 239221,
  [SMALL_STATE(6238)] = 239244,
  [SMALL_STATE(6239)] = 239265,
  [SMALL_STATE(6240)] = 239288,
  [SMALL_STATE(6241)] = 239307,
  [SMALL_STATE(6242)] = 239330,
  [SMALL_STATE(6243)] = 239353,
  [SMALL_STATE(6244)] = 239376,
  [SMALL_STATE(6245)] = 239399,
  [SMALL_STATE(6246)] = 239416,
  [SMALL_STATE(6247)] = 239433,
  [SMALL_STATE(6248)] = 239456,
  [SMALL_STATE(6249)] = 239479,
  [SMALL_STATE(6250)] = 239496,
  [SMALL_STATE(6251)] = 239519,
  [SMALL_STATE(6252)] = 239542,
  [SMALL_STATE(6253)] = 239565,
  [SMALL_STATE(6254)] = 239588,
  [SMALL_STATE(6255)] = 239611,
  [SMALL_STATE(6256)] = 239634,
  [SMALL_STATE(6257)] = 239657,
  [SMALL_STATE(6258)] = 239678,
  [SMALL_STATE(6259)] = 239701,
  [SMALL_STATE(6260)] = 239722,
  [SMALL_STATE(6261)] = 239745,
  [SMALL_STATE(6262)] = 239768,
  [SMALL_STATE(6263)] = 239791,
  [SMALL_STATE(6264)] = 239814,
  [SMALL_STATE(6265)] = 239837,
  [SMALL_STATE(6266)] = 239858,
  [SMALL_STATE(6267)] = 239881,
  [SMALL_STATE(6268)] = 239904,
  [SMALL_STATE(6269)] = 239923,
  [SMALL_STATE(6270)] = 239942,
  [SMALL_STATE(6271)] = 239963,
  [SMALL_STATE(6272)] = 239986,
  [SMALL_STATE(6273)] = 240009,
  [SMALL_STATE(6274)] = 240032,
  [SMALL_STATE(6275)] = 240055,
  [SMALL_STATE(6276)] = 240078,
  [SMALL_STATE(6277)] = 240097,
  [SMALL_STATE(6278)] = 240120,
  [SMALL_STATE(6279)] = 240143,
  [SMALL_STATE(6280)] = 240166,
  [SMALL_STATE(6281)] = 240189,
  [SMALL_STATE(6282)] = 240212,
  [SMALL_STATE(6283)] = 240235,
  [SMALL_STATE(6284)] = 240258,
  [SMALL_STATE(6285)] = 240281,
  [SMALL_STATE(6286)] = 240302,
  [SMALL_STATE(6287)] = 240325,
  [SMALL_STATE(6288)] = 240348,
  [SMALL_STATE(6289)] = 240371,
  [SMALL_STATE(6290)] = 240394,
  [SMALL_STATE(6291)] = 240417,
  [SMALL_STATE(6292)] = 240440,
  [SMALL_STATE(6293)] = 240463,
  [SMALL_STATE(6294)] = 240486,
  [SMALL_STATE(6295)] = 240509,
  [SMALL_STATE(6296)] = 240532,
  [SMALL_STATE(6297)] = 240555,
  [SMALL_STATE(6298)] = 240576,
  [SMALL_STATE(6299)] = 240599,
  [SMALL_STATE(6300)] = 240622,
  [SMALL_STATE(6301)] = 240645,
  [SMALL_STATE(6302)] = 240668,
  [SMALL_STATE(6303)] = 240691,
  [SMALL_STATE(6304)] = 240712,
  [SMALL_STATE(6305)] = 240735,
  [SMALL_STATE(6306)] = 240753,
  [SMALL_STATE(6307)] = 240773,
  [SMALL_STATE(6308)] = 240793,
  [SMALL_STATE(6309)] = 240813,
  [SMALL_STATE(6310)] = 240829,
  [SMALL_STATE(6311)] = 240849,
  [SMALL_STATE(6312)] = 240869,
  [SMALL_STATE(6313)] = 240889,
  [SMALL_STATE(6314)] = 240909,
  [SMALL_STATE(6315)] = 240929,
  [SMALL_STATE(6316)] = 240949,
  [SMALL_STATE(6317)] = 240965,
  [SMALL_STATE(6318)] = 240985,
  [SMALL_STATE(6319)] = 241005,
  [SMALL_STATE(6320)] = 241025,
  [SMALL_STATE(6321)] = 241045,
  [SMALL_STATE(6322)] = 241065,
  [SMALL_STATE(6323)] = 241085,
  [SMALL_STATE(6324)] = 241105,
  [SMALL_STATE(6325)] = 241125,
  [SMALL_STATE(6326)] = 241145,
  [SMALL_STATE(6327)] = 241163,
  [SMALL_STATE(6328)] = 241181,
  [SMALL_STATE(6329)] = 241201,
  [SMALL_STATE(6330)] = 241219,
  [SMALL_STATE(6331)] = 241237,
  [SMALL_STATE(6332)] = 241255,
  [SMALL_STATE(6333)] = 241273,
  [SMALL_STATE(6334)] = 241293,
  [SMALL_STATE(6335)] = 241313,
  [SMALL_STATE(6336)] = 241333,
  [SMALL_STATE(6337)] = 241353,
  [SMALL_STATE(6338)] = 241373,
  [SMALL_STATE(6339)] = 241393,
  [SMALL_STATE(6340)] = 241413,
  [SMALL_STATE(6341)] = 241431,
  [SMALL_STATE(6342)] = 241451,
  [SMALL_STATE(6343)] = 241471,
  [SMALL_STATE(6344)] = 241489,
  [SMALL_STATE(6345)] = 241509,
  [SMALL_STATE(6346)] = 241529,
  [SMALL_STATE(6347)] = 241549,
  [SMALL_STATE(6348)] = 241567,
  [SMALL_STATE(6349)] = 241585,
  [SMALL_STATE(6350)] = 241605,
  [SMALL_STATE(6351)] = 241625,
  [SMALL_STATE(6352)] = 241645,
  [SMALL_STATE(6353)] = 241665,
  [SMALL_STATE(6354)] = 241685,
  [SMALL_STATE(6355)] = 241705,
  [SMALL_STATE(6356)] = 241725,
  [SMALL_STATE(6357)] = 241743,
  [SMALL_STATE(6358)] = 241763,
  [SMALL_STATE(6359)] = 241783,
  [SMALL_STATE(6360)] = 241801,
  [SMALL_STATE(6361)] = 241821,
  [SMALL_STATE(6362)] = 241841,
  [SMALL_STATE(6363)] = 241861,
  [SMALL_STATE(6364)] = 241881,
  [SMALL_STATE(6365)] = 241901,
  [SMALL_STATE(6366)] = 241921,
  [SMALL_STATE(6367)] = 241941,
  [SMALL_STATE(6368)] = 241961,
  [SMALL_STATE(6369)] = 241981,
  [SMALL_STATE(6370)] = 242001,
  [SMALL_STATE(6371)] = 242021,
  [SMALL_STATE(6372)] = 242041,
  [SMALL_STATE(6373)] = 242061,
  [SMALL_STATE(6374)] = 242081,
  [SMALL_STATE(6375)] = 242101,
  [SMALL_STATE(6376)] = 242121,
  [SMALL_STATE(6377)] = 242141,
  [SMALL_STATE(6378)] = 242159,
  [SMALL_STATE(6379)] = 242177,
  [SMALL_STATE(6380)] = 242197,
  [SMALL_STATE(6381)] = 242217,
  [SMALL_STATE(6382)] = 242237,
  [SMALL_STATE(6383)] = 242255,
  [SMALL_STATE(6384)] = 242275,
  [SMALL_STATE(6385)] = 242295,
  [SMALL_STATE(6386)] = 242315,
  [SMALL_STATE(6387)] = 242335,
  [SMALL_STATE(6388)] = 242355,
  [SMALL_STATE(6389)] = 242373,
  [SMALL_STATE(6390)] = 242393,
  [SMALL_STATE(6391)] = 242411,
  [SMALL_STATE(6392)] = 242431,
  [SMALL_STATE(6393)] = 242451,
  [SMALL_STATE(6394)] = 242471,
  [SMALL_STATE(6395)] = 242491,
  [SMALL_STATE(6396)] = 242509,
  [SMALL_STATE(6397)] = 242529,
  [SMALL_STATE(6398)] = 242549,
  [SMALL_STATE(6399)] = 242569,
  [SMALL_STATE(6400)] = 242589,
  [SMALL_STATE(6401)] = 242609,
  [SMALL_STATE(6402)] = 242629,
  [SMALL_STATE(6403)] = 242649,
  [SMALL_STATE(6404)] = 242669,
  [SMALL_STATE(6405)] = 242685,
  [SMALL_STATE(6406)] = 242705,
  [SMALL_STATE(6407)] = 242723,
  [SMALL_STATE(6408)] = 242743,
  [SMALL_STATE(6409)] = 242763,
  [SMALL_STATE(6410)] = 242783,
  [SMALL_STATE(6411)] = 242803,
  [SMALL_STATE(6412)] = 242823,
  [SMALL_STATE(6413)] = 242843,
  [SMALL_STATE(6414)] = 242863,
  [SMALL_STATE(6415)] = 242883,
  [SMALL_STATE(6416)] = 242903,
  [SMALL_STATE(6417)] = 242923,
  [SMALL_STATE(6418)] = 242943,
  [SMALL_STATE(6419)] = 242963,
  [SMALL_STATE(6420)] = 242983,
  [SMALL_STATE(6421)] = 243003,
  [SMALL_STATE(6422)] = 243021,
  [SMALL_STATE(6423)] = 243041,
  [SMALL_STATE(6424)] = 243061,
  [SMALL_STATE(6425)] = 243081,
  [SMALL_STATE(6426)] = 243101,
  [SMALL_STATE(6427)] = 243121,
  [SMALL_STATE(6428)] = 243141,
  [SMALL_STATE(6429)] = 243161,
  [SMALL_STATE(6430)] = 243181,
  [SMALL_STATE(6431)] = 243199,
  [SMALL_STATE(6432)] = 243219,
  [SMALL_STATE(6433)] = 243237,
  [SMALL_STATE(6434)] = 243257,
  [SMALL_STATE(6435)] = 243277,
  [SMALL_STATE(6436)] = 243297,
  [SMALL_STATE(6437)] = 243317,
  [SMALL_STATE(6438)] = 243337,
  [SMALL_STATE(6439)] = 243357,
  [SMALL_STATE(6440)] = 243375,
  [SMALL_STATE(6441)] = 243395,
  [SMALL_STATE(6442)] = 243415,
  [SMALL_STATE(6443)] = 243435,
  [SMALL_STATE(6444)] = 243453,
  [SMALL_STATE(6445)] = 243469,
  [SMALL_STATE(6446)] = 243489,
  [SMALL_STATE(6447)] = 243507,
  [SMALL_STATE(6448)] = 243527,
  [SMALL_STATE(6449)] = 243547,
  [SMALL_STATE(6450)] = 243567,
  [SMALL_STATE(6451)] = 243587,
  [SMALL_STATE(6452)] = 243607,
  [SMALL_STATE(6453)] = 243627,
  [SMALL_STATE(6454)] = 243643,
  [SMALL_STATE(6455)] = 243661,
  [SMALL_STATE(6456)] = 243681,
  [SMALL_STATE(6457)] = 243699,
  [SMALL_STATE(6458)] = 243719,
  [SMALL_STATE(6459)] = 243739,
  [SMALL_STATE(6460)] = 243759,
  [SMALL_STATE(6461)] = 243779,
  [SMALL_STATE(6462)] = 243797,
  [SMALL_STATE(6463)] = 243817,
  [SMALL_STATE(6464)] = 243837,
  [SMALL_STATE(6465)] = 243855,
  [SMALL_STATE(6466)] = 243875,
  [SMALL_STATE(6467)] = 243895,
  [SMALL_STATE(6468)] = 243913,
  [SMALL_STATE(6469)] = 243931,
  [SMALL_STATE(6470)] = 243951,
  [SMALL_STATE(6471)] = 243971,
  [SMALL_STATE(6472)] = 243991,
  [SMALL_STATE(6473)] = 244011,
  [SMALL_STATE(6474)] = 244031,
  [SMALL_STATE(6475)] = 244051,
  [SMALL_STATE(6476)] = 244071,
  [SMALL_STATE(6477)] = 244089,
  [SMALL_STATE(6478)] = 244109,
  [SMALL_STATE(6479)] = 244129,
  [SMALL_STATE(6480)] = 244149,
  [SMALL_STATE(6481)] = 244169,
  [SMALL_STATE(6482)] = 244189,
  [SMALL_STATE(6483)] = 244209,
  [SMALL_STATE(6484)] = 244229,
  [SMALL_STATE(6485)] = 244249,
  [SMALL_STATE(6486)] = 244267,
  [SMALL_STATE(6487)] = 244285,
  [SMALL_STATE(6488)] = 244305,
  [SMALL_STATE(6489)] = 244325,
  [SMALL_STATE(6490)] = 244345,
  [SMALL_STATE(6491)] = 244361,
  [SMALL_STATE(6492)] = 244381,
  [SMALL_STATE(6493)] = 244399,
  [SMALL_STATE(6494)] = 244419,
  [SMALL_STATE(6495)] = 244437,
  [SMALL_STATE(6496)] = 244457,
  [SMALL_STATE(6497)] = 244473,
  [SMALL_STATE(6498)] = 244493,
  [SMALL_STATE(6499)] = 244513,
  [SMALL_STATE(6500)] = 244531,
  [SMALL_STATE(6501)] = 244551,
  [SMALL_STATE(6502)] = 244571,
  [SMALL_STATE(6503)] = 244587,
  [SMALL_STATE(6504)] = 244607,
  [SMALL_STATE(6505)] = 244627,
  [SMALL_STATE(6506)] = 244647,
  [SMALL_STATE(6507)] = 244665,
  [SMALL_STATE(6508)] = 244685,
  [SMALL_STATE(6509)] = 244703,
  [SMALL_STATE(6510)] = 244723,
  [SMALL_STATE(6511)] = 244743,
  [SMALL_STATE(6512)] = 244759,
  [SMALL_STATE(6513)] = 244779,
  [SMALL_STATE(6514)] = 244795,
  [SMALL_STATE(6515)] = 244815,
  [SMALL_STATE(6516)] = 244835,
  [SMALL_STATE(6517)] = 244855,
  [SMALL_STATE(6518)] = 244875,
  [SMALL_STATE(6519)] = 244891,
  [SMALL_STATE(6520)] = 244911,
  [SMALL_STATE(6521)] = 244929,
  [SMALL_STATE(6522)] = 244949,
  [SMALL_STATE(6523)] = 244967,
  [SMALL_STATE(6524)] = 244987,
  [SMALL_STATE(6525)] = 245007,
  [SMALL_STATE(6526)] = 245027,
  [SMALL_STATE(6527)] = 245045,
  [SMALL_STATE(6528)] = 245065,
  [SMALL_STATE(6529)] = 245085,
  [SMALL_STATE(6530)] = 245105,
  [SMALL_STATE(6531)] = 245125,
  [SMALL_STATE(6532)] = 245145,
  [SMALL_STATE(6533)] = 245165,
  [SMALL_STATE(6534)] = 245183,
  [SMALL_STATE(6535)] = 245203,
  [SMALL_STATE(6536)] = 245223,
  [SMALL_STATE(6537)] = 245243,
  [SMALL_STATE(6538)] = 245263,
  [SMALL_STATE(6539)] = 245283,
  [SMALL_STATE(6540)] = 245301,
  [SMALL_STATE(6541)] = 245321,
  [SMALL_STATE(6542)] = 245341,
  [SMALL_STATE(6543)] = 245361,
  [SMALL_STATE(6544)] = 245381,
  [SMALL_STATE(6545)] = 245401,
  [SMALL_STATE(6546)] = 245421,
  [SMALL_STATE(6547)] = 245441,
  [SMALL_STATE(6548)] = 245459,
  [SMALL_STATE(6549)] = 245479,
  [SMALL_STATE(6550)] = 245497,
  [SMALL_STATE(6551)] = 245515,
  [SMALL_STATE(6552)] = 245535,
  [SMALL_STATE(6553)] = 245553,
  [SMALL_STATE(6554)] = 245573,
  [SMALL_STATE(6555)] = 245593,
  [SMALL_STATE(6556)] = 245613,
  [SMALL_STATE(6557)] = 245633,
  [SMALL_STATE(6558)] = 245653,
  [SMALL_STATE(6559)] = 245673,
  [SMALL_STATE(6560)] = 245693,
  [SMALL_STATE(6561)] = 245713,
  [SMALL_STATE(6562)] = 245731,
  [SMALL_STATE(6563)] = 245751,
  [SMALL_STATE(6564)] = 245771,
  [SMALL_STATE(6565)] = 245791,
  [SMALL_STATE(6566)] = 245811,
  [SMALL_STATE(6567)] = 245829,
  [SMALL_STATE(6568)] = 245849,
  [SMALL_STATE(6569)] = 245869,
  [SMALL_STATE(6570)] = 245889,
  [SMALL_STATE(6571)] = 245907,
  [SMALL_STATE(6572)] = 245927,
  [SMALL_STATE(6573)] = 245947,
  [SMALL_STATE(6574)] = 245965,
  [SMALL_STATE(6575)] = 245983,
  [SMALL_STATE(6576)] = 246003,
  [SMALL_STATE(6577)] = 246023,
  [SMALL_STATE(6578)] = 246041,
  [SMALL_STATE(6579)] = 246061,
  [SMALL_STATE(6580)] = 246081,
  [SMALL_STATE(6581)] = 246099,
  [SMALL_STATE(6582)] = 246119,
  [SMALL_STATE(6583)] = 246137,
  [SMALL_STATE(6584)] = 246157,
  [SMALL_STATE(6585)] = 246175,
  [SMALL_STATE(6586)] = 246195,
  [SMALL_STATE(6587)] = 246215,
  [SMALL_STATE(6588)] = 246233,
  [SMALL_STATE(6589)] = 246253,
  [SMALL_STATE(6590)] = 246273,
  [SMALL_STATE(6591)] = 246291,
  [SMALL_STATE(6592)] = 246311,
  [SMALL_STATE(6593)] = 246329,
  [SMALL_STATE(6594)] = 246349,
  [SMALL_STATE(6595)] = 246367,
  [SMALL_STATE(6596)] = 246387,
  [SMALL_STATE(6597)] = 246405,
  [SMALL_STATE(6598)] = 246425,
  [SMALL_STATE(6599)] = 246445,
  [SMALL_STATE(6600)] = 246463,
  [SMALL_STATE(6601)] = 246481,
  [SMALL_STATE(6602)] = 246499,
  [SMALL_STATE(6603)] = 246519,
  [SMALL_STATE(6604)] = 246539,
  [SMALL_STATE(6605)] = 246559,
  [SMALL_STATE(6606)] = 246579,
  [SMALL_STATE(6607)] = 246599,
  [SMALL_STATE(6608)] = 246619,
  [SMALL_STATE(6609)] = 246639,
  [SMALL_STATE(6610)] = 246659,
  [SMALL_STATE(6611)] = 246679,
  [SMALL_STATE(6612)] = 246699,
  [SMALL_STATE(6613)] = 246719,
  [SMALL_STATE(6614)] = 246737,
  [SMALL_STATE(6615)] = 246757,
  [SMALL_STATE(6616)] = 246777,
  [SMALL_STATE(6617)] = 246797,
  [SMALL_STATE(6618)] = 246817,
  [SMALL_STATE(6619)] = 246837,
  [SMALL_STATE(6620)] = 246855,
  [SMALL_STATE(6621)] = 246875,
  [SMALL_STATE(6622)] = 246895,
  [SMALL_STATE(6623)] = 246913,
  [SMALL_STATE(6624)] = 246933,
  [SMALL_STATE(6625)] = 246953,
  [SMALL_STATE(6626)] = 246973,
  [SMALL_STATE(6627)] = 246993,
  [SMALL_STATE(6628)] = 247013,
  [SMALL_STATE(6629)] = 247033,
  [SMALL_STATE(6630)] = 247053,
  [SMALL_STATE(6631)] = 247071,
  [SMALL_STATE(6632)] = 247091,
  [SMALL_STATE(6633)] = 247109,
  [SMALL_STATE(6634)] = 247129,
  [SMALL_STATE(6635)] = 247145,
  [SMALL_STATE(6636)] = 247165,
  [SMALL_STATE(6637)] = 247183,
  [SMALL_STATE(6638)] = 247203,
  [SMALL_STATE(6639)] = 247223,
  [SMALL_STATE(6640)] = 247243,
  [SMALL_STATE(6641)] = 247263,
  [SMALL_STATE(6642)] = 247281,
  [SMALL_STATE(6643)] = 247301,
  [SMALL_STATE(6644)] = 247321,
  [SMALL_STATE(6645)] = 247341,
  [SMALL_STATE(6646)] = 247361,
  [SMALL_STATE(6647)] = 247381,
  [SMALL_STATE(6648)] = 247397,
  [SMALL_STATE(6649)] = 247417,
  [SMALL_STATE(6650)] = 247435,
  [SMALL_STATE(6651)] = 247455,
  [SMALL_STATE(6652)] = 247475,
  [SMALL_STATE(6653)] = 247493,
  [SMALL_STATE(6654)] = 247513,
  [SMALL_STATE(6655)] = 247533,
  [SMALL_STATE(6656)] = 247553,
  [SMALL_STATE(6657)] = 247573,
  [SMALL_STATE(6658)] = 247593,
  [SMALL_STATE(6659)] = 247611,
  [SMALL_STATE(6660)] = 247631,
  [SMALL_STATE(6661)] = 247649,
  [SMALL_STATE(6662)] = 247669,
  [SMALL_STATE(6663)] = 247689,
  [SMALL_STATE(6664)] = 247709,
  [SMALL_STATE(6665)] = 247727,
  [SMALL_STATE(6666)] = 247747,
  [SMALL_STATE(6667)] = 247767,
  [SMALL_STATE(6668)] = 247787,
  [SMALL_STATE(6669)] = 247807,
  [SMALL_STATE(6670)] = 247827,
  [SMALL_STATE(6671)] = 247847,
  [SMALL_STATE(6672)] = 247865,
  [SMALL_STATE(6673)] = 247885,
  [SMALL_STATE(6674)] = 247905,
  [SMALL_STATE(6675)] = 247925,
  [SMALL_STATE(6676)] = 247945,
  [SMALL_STATE(6677)] = 247963,
  [SMALL_STATE(6678)] = 247983,
  [SMALL_STATE(6679)] = 248003,
  [SMALL_STATE(6680)] = 248021,
  [SMALL_STATE(6681)] = 248041,
  [SMALL_STATE(6682)] = 248061,
  [SMALL_STATE(6683)] = 248081,
  [SMALL_STATE(6684)] = 248101,
  [SMALL_STATE(6685)] = 248119,
  [SMALL_STATE(6686)] = 248139,
  [SMALL_STATE(6687)] = 248159,
  [SMALL_STATE(6688)] = 248179,
  [SMALL_STATE(6689)] = 248197,
  [SMALL_STATE(6690)] = 248215,
  [SMALL_STATE(6691)] = 248235,
  [SMALL_STATE(6692)] = 248253,
  [SMALL_STATE(6693)] = 248273,
  [SMALL_STATE(6694)] = 248293,
  [SMALL_STATE(6695)] = 248311,
  [SMALL_STATE(6696)] = 248331,
  [SMALL_STATE(6697)] = 248351,
  [SMALL_STATE(6698)] = 248371,
  [SMALL_STATE(6699)] = 248389,
  [SMALL_STATE(6700)] = 248409,
  [SMALL_STATE(6701)] = 248429,
  [SMALL_STATE(6702)] = 248449,
  [SMALL_STATE(6703)] = 248469,
  [SMALL_STATE(6704)] = 248487,
  [SMALL_STATE(6705)] = 248507,
  [SMALL_STATE(6706)] = 248527,
  [SMALL_STATE(6707)] = 248547,
  [SMALL_STATE(6708)] = 248565,
  [SMALL_STATE(6709)] = 248585,
  [SMALL_STATE(6710)] = 248603,
  [SMALL_STATE(6711)] = 248623,
  [SMALL_STATE(6712)] = 248641,
  [SMALL_STATE(6713)] = 248661,
  [SMALL_STATE(6714)] = 248679,
  [SMALL_STATE(6715)] = 248699,
  [SMALL_STATE(6716)] = 248717,
  [SMALL_STATE(6717)] = 248737,
  [SMALL_STATE(6718)] = 248755,
  [SMALL_STATE(6719)] = 248775,
  [SMALL_STATE(6720)] = 248793,
  [SMALL_STATE(6721)] = 248813,
  [SMALL_STATE(6722)] = 248831,
  [SMALL_STATE(6723)] = 248849,
  [SMALL_STATE(6724)] = 248867,
  [SMALL_STATE(6725)] = 248885,
  [SMALL_STATE(6726)] = 248903,
  [SMALL_STATE(6727)] = 248923,
  [SMALL_STATE(6728)] = 248943,
  [SMALL_STATE(6729)] = 248963,
  [SMALL_STATE(6730)] = 248983,
  [SMALL_STATE(6731)] = 249003,
  [SMALL_STATE(6732)] = 249023,
  [SMALL_STATE(6733)] = 249043,
  [SMALL_STATE(6734)] = 249063,
  [SMALL_STATE(6735)] = 249083,
  [SMALL_STATE(6736)] = 249103,
  [SMALL_STATE(6737)] = 249123,
  [SMALL_STATE(6738)] = 249143,
  [SMALL_STATE(6739)] = 249163,
  [SMALL_STATE(6740)] = 249183,
  [SMALL_STATE(6741)] = 249203,
  [SMALL_STATE(6742)] = 249223,
  [SMALL_STATE(6743)] = 249243,
  [SMALL_STATE(6744)] = 249263,
  [SMALL_STATE(6745)] = 249283,
  [SMALL_STATE(6746)] = 249303,
  [SMALL_STATE(6747)] = 249323,
  [SMALL_STATE(6748)] = 249343,
  [SMALL_STATE(6749)] = 249359,
  [SMALL_STATE(6750)] = 249379,
  [SMALL_STATE(6751)] = 249395,
  [SMALL_STATE(6752)] = 249415,
  [SMALL_STATE(6753)] = 249435,
  [SMALL_STATE(6754)] = 249455,
  [SMALL_STATE(6755)] = 249473,
  [SMALL_STATE(6756)] = 249491,
  [SMALL_STATE(6757)] = 249509,
  [SMALL_STATE(6758)] = 249529,
  [SMALL_STATE(6759)] = 249549,
  [SMALL_STATE(6760)] = 249569,
  [SMALL_STATE(6761)] = 249584,
  [SMALL_STATE(6762)] = 249601,
  [SMALL_STATE(6763)] = 249618,
  [SMALL_STATE(6764)] = 249635,
  [SMALL_STATE(6765)] = 249650,
  [SMALL_STATE(6766)] = 249665,
  [SMALL_STATE(6767)] = 249682,
  [SMALL_STATE(6768)] = 249699,
  [SMALL_STATE(6769)] = 249716,
  [SMALL_STATE(6770)] = 249733,
  [SMALL_STATE(6771)] = 249750,
  [SMALL_STATE(6772)] = 249767,
  [SMALL_STATE(6773)] = 249782,
  [SMALL_STATE(6774)] = 249799,
  [SMALL_STATE(6775)] = 249816,
  [SMALL_STATE(6776)] = 249833,
  [SMALL_STATE(6777)] = 249848,
  [SMALL_STATE(6778)] = 249863,
  [SMALL_STATE(6779)] = 249880,
  [SMALL_STATE(6780)] = 249895,
  [SMALL_STATE(6781)] = 249912,
  [SMALL_STATE(6782)] = 249927,
  [SMALL_STATE(6783)] = 249944,
  [SMALL_STATE(6784)] = 249961,
  [SMALL_STATE(6785)] = 249978,
  [SMALL_STATE(6786)] = 249995,
  [SMALL_STATE(6787)] = 250012,
  [SMALL_STATE(6788)] = 250029,
  [SMALL_STATE(6789)] = 250046,
  [SMALL_STATE(6790)] = 250061,
  [SMALL_STATE(6791)] = 250078,
  [SMALL_STATE(6792)] = 250095,
  [SMALL_STATE(6793)] = 250112,
  [SMALL_STATE(6794)] = 250129,
  [SMALL_STATE(6795)] = 250146,
  [SMALL_STATE(6796)] = 250163,
  [SMALL_STATE(6797)] = 250180,
  [SMALL_STATE(6798)] = 250197,
  [SMALL_STATE(6799)] = 250214,
  [SMALL_STATE(6800)] = 250231,
  [SMALL_STATE(6801)] = 250248,
  [SMALL_STATE(6802)] = 250263,
  [SMALL_STATE(6803)] = 250278,
  [SMALL_STATE(6804)] = 250295,
  [SMALL_STATE(6805)] = 250310,
  [SMALL_STATE(6806)] = 250327,
  [SMALL_STATE(6807)] = 250344,
  [SMALL_STATE(6808)] = 250359,
  [SMALL_STATE(6809)] = 250376,
  [SMALL_STATE(6810)] = 250393,
  [SMALL_STATE(6811)] = 250408,
  [SMALL_STATE(6812)] = 250425,
  [SMALL_STATE(6813)] = 250442,
  [SMALL_STATE(6814)] = 250459,
  [SMALL_STATE(6815)] = 250476,
  [SMALL_STATE(6816)] = 250493,
  [SMALL_STATE(6817)] = 250510,
  [SMALL_STATE(6818)] = 250527,
  [SMALL_STATE(6819)] = 250544,
  [SMALL_STATE(6820)] = 250561,
  [SMALL_STATE(6821)] = 250578,
  [SMALL_STATE(6822)] = 250595,
  [SMALL_STATE(6823)] = 250612,
  [SMALL_STATE(6824)] = 250629,
  [SMALL_STATE(6825)] = 250646,
  [SMALL_STATE(6826)] = 250663,
  [SMALL_STATE(6827)] = 250680,
  [SMALL_STATE(6828)] = 250697,
  [SMALL_STATE(6829)] = 250714,
  [SMALL_STATE(6830)] = 250729,
  [SMALL_STATE(6831)] = 250746,
  [SMALL_STATE(6832)] = 250763,
  [SMALL_STATE(6833)] = 250780,
  [SMALL_STATE(6834)] = 250795,
  [SMALL_STATE(6835)] = 250812,
  [SMALL_STATE(6836)] = 250829,
  [SMALL_STATE(6837)] = 250846,
  [SMALL_STATE(6838)] = 250863,
  [SMALL_STATE(6839)] = 250878,
  [SMALL_STATE(6840)] = 250895,
  [SMALL_STATE(6841)] = 250912,
  [SMALL_STATE(6842)] = 250929,
  [SMALL_STATE(6843)] = 250946,
  [SMALL_STATE(6844)] = 250963,
  [SMALL_STATE(6845)] = 250978,
  [SMALL_STATE(6846)] = 250995,
  [SMALL_STATE(6847)] = 251012,
  [SMALL_STATE(6848)] = 251029,
  [SMALL_STATE(6849)] = 251046,
  [SMALL_STATE(6850)] = 251063,
  [SMALL_STATE(6851)] = 251080,
  [SMALL_STATE(6852)] = 251095,
  [SMALL_STATE(6853)] = 251110,
  [SMALL_STATE(6854)] = 251127,
  [SMALL_STATE(6855)] = 251142,
  [SMALL_STATE(6856)] = 251159,
  [SMALL_STATE(6857)] = 251176,
  [SMALL_STATE(6858)] = 251193,
  [SMALL_STATE(6859)] = 251210,
  [SMALL_STATE(6860)] = 251227,
  [SMALL_STATE(6861)] = 251244,
  [SMALL_STATE(6862)] = 251261,
  [SMALL_STATE(6863)] = 251278,
  [SMALL_STATE(6864)] = 251295,
  [SMALL_STATE(6865)] = 251312,
  [SMALL_STATE(6866)] = 251327,
  [SMALL_STATE(6867)] = 251342,
  [SMALL_STATE(6868)] = 251359,
  [SMALL_STATE(6869)] = 251376,
  [SMALL_STATE(6870)] = 251393,
  [SMALL_STATE(6871)] = 251410,
  [SMALL_STATE(6872)] = 251427,
  [SMALL_STATE(6873)] = 251444,
  [SMALL_STATE(6874)] = 251459,
  [SMALL_STATE(6875)] = 251476,
  [SMALL_STATE(6876)] = 251493,
  [SMALL_STATE(6877)] = 251510,
  [SMALL_STATE(6878)] = 251527,
  [SMALL_STATE(6879)] = 251542,
  [SMALL_STATE(6880)] = 251559,
  [SMALL_STATE(6881)] = 251576,
  [SMALL_STATE(6882)] = 251593,
  [SMALL_STATE(6883)] = 251610,
  [SMALL_STATE(6884)] = 251627,
  [SMALL_STATE(6885)] = 251642,
  [SMALL_STATE(6886)] = 251659,
  [SMALL_STATE(6887)] = 251676,
  [SMALL_STATE(6888)] = 251693,
  [SMALL_STATE(6889)] = 251710,
  [SMALL_STATE(6890)] = 251727,
  [SMALL_STATE(6891)] = 251742,
  [SMALL_STATE(6892)] = 251759,
  [SMALL_STATE(6893)] = 251776,
  [SMALL_STATE(6894)] = 251791,
  [SMALL_STATE(6895)] = 251808,
  [SMALL_STATE(6896)] = 251825,
  [SMALL_STATE(6897)] = 251842,
  [SMALL_STATE(6898)] = 251859,
  [SMALL_STATE(6899)] = 251876,
  [SMALL_STATE(6900)] = 251893,
  [SMALL_STATE(6901)] = 251910,
  [SMALL_STATE(6902)] = 251927,
  [SMALL_STATE(6903)] = 251944,
  [SMALL_STATE(6904)] = 251961,
  [SMALL_STATE(6905)] = 251978,
  [SMALL_STATE(6906)] = 251995,
  [SMALL_STATE(6907)] = 252012,
  [SMALL_STATE(6908)] = 252029,
  [SMALL_STATE(6909)] = 252044,
  [SMALL_STATE(6910)] = 252061,
  [SMALL_STATE(6911)] = 252078,
  [SMALL_STATE(6912)] = 252095,
  [SMALL_STATE(6913)] = 252112,
  [SMALL_STATE(6914)] = 252127,
  [SMALL_STATE(6915)] = 252144,
  [SMALL_STATE(6916)] = 252161,
  [SMALL_STATE(6917)] = 252178,
  [SMALL_STATE(6918)] = 252195,
  [SMALL_STATE(6919)] = 252212,
  [SMALL_STATE(6920)] = 252227,
  [SMALL_STATE(6921)] = 252244,
  [SMALL_STATE(6922)] = 252261,
  [SMALL_STATE(6923)] = 252278,
  [SMALL_STATE(6924)] = 252295,
  [SMALL_STATE(6925)] = 252312,
  [SMALL_STATE(6926)] = 252329,
  [SMALL_STATE(6927)] = 252346,
  [SMALL_STATE(6928)] = 252363,
  [SMALL_STATE(6929)] = 252380,
  [SMALL_STATE(6930)] = 252397,
  [SMALL_STATE(6931)] = 252414,
  [SMALL_STATE(6932)] = 252431,
  [SMALL_STATE(6933)] = 252448,
  [SMALL_STATE(6934)] = 252465,
  [SMALL_STATE(6935)] = 252482,
  [SMALL_STATE(6936)] = 252499,
  [SMALL_STATE(6937)] = 252516,
  [SMALL_STATE(6938)] = 252533,
  [SMALL_STATE(6939)] = 252550,
  [SMALL_STATE(6940)] = 252567,
  [SMALL_STATE(6941)] = 252584,
  [SMALL_STATE(6942)] = 252599,
  [SMALL_STATE(6943)] = 252614,
  [SMALL_STATE(6944)] = 252631,
  [SMALL_STATE(6945)] = 252648,
  [SMALL_STATE(6946)] = 252665,
  [SMALL_STATE(6947)] = 252682,
  [SMALL_STATE(6948)] = 252697,
  [SMALL_STATE(6949)] = 252714,
  [SMALL_STATE(6950)] = 252729,
  [SMALL_STATE(6951)] = 252746,
  [SMALL_STATE(6952)] = 252763,
  [SMALL_STATE(6953)] = 252780,
  [SMALL_STATE(6954)] = 252797,
  [SMALL_STATE(6955)] = 252814,
  [SMALL_STATE(6956)] = 252831,
  [SMALL_STATE(6957)] = 252848,
  [SMALL_STATE(6958)] = 252865,
  [SMALL_STATE(6959)] = 252880,
  [SMALL_STATE(6960)] = 252897,
  [SMALL_STATE(6961)] = 252914,
  [SMALL_STATE(6962)] = 252931,
  [SMALL_STATE(6963)] = 252948,
  [SMALL_STATE(6964)] = 252965,
  [SMALL_STATE(6965)] = 252980,
  [SMALL_STATE(6966)] = 252997,
  [SMALL_STATE(6967)] = 253014,
  [SMALL_STATE(6968)] = 253031,
  [SMALL_STATE(6969)] = 253048,
  [SMALL_STATE(6970)] = 253065,
  [SMALL_STATE(6971)] = 253082,
  [SMALL_STATE(6972)] = 253099,
  [SMALL_STATE(6973)] = 253116,
  [SMALL_STATE(6974)] = 253133,
  [SMALL_STATE(6975)] = 253150,
  [SMALL_STATE(6976)] = 253167,
  [SMALL_STATE(6977)] = 253184,
  [SMALL_STATE(6978)] = 253201,
  [SMALL_STATE(6979)] = 253218,
  [SMALL_STATE(6980)] = 253235,
  [SMALL_STATE(6981)] = 253252,
  [SMALL_STATE(6982)] = 253269,
  [SMALL_STATE(6983)] = 253286,
  [SMALL_STATE(6984)] = 253303,
  [SMALL_STATE(6985)] = 253320,
  [SMALL_STATE(6986)] = 253337,
  [SMALL_STATE(6987)] = 253354,
  [SMALL_STATE(6988)] = 253369,
  [SMALL_STATE(6989)] = 253386,
  [SMALL_STATE(6990)] = 253403,
  [SMALL_STATE(6991)] = 253420,
  [SMALL_STATE(6992)] = 253437,
  [SMALL_STATE(6993)] = 253452,
  [SMALL_STATE(6994)] = 253469,
  [SMALL_STATE(6995)] = 253486,
  [SMALL_STATE(6996)] = 253503,
  [SMALL_STATE(6997)] = 253520,
  [SMALL_STATE(6998)] = 253537,
  [SMALL_STATE(6999)] = 253554,
  [SMALL_STATE(7000)] = 253571,
  [SMALL_STATE(7001)] = 253588,
  [SMALL_STATE(7002)] = 253605,
  [SMALL_STATE(7003)] = 253622,
  [SMALL_STATE(7004)] = 253639,
  [SMALL_STATE(7005)] = 253656,
  [SMALL_STATE(7006)] = 253673,
  [SMALL_STATE(7007)] = 253690,
  [SMALL_STATE(7008)] = 253707,
  [SMALL_STATE(7009)] = 253724,
  [SMALL_STATE(7010)] = 253741,
  [SMALL_STATE(7011)] = 253758,
  [SMALL_STATE(7012)] = 253775,
  [SMALL_STATE(7013)] = 253792,
  [SMALL_STATE(7014)] = 253809,
  [SMALL_STATE(7015)] = 253826,
  [SMALL_STATE(7016)] = 253843,
  [SMALL_STATE(7017)] = 253860,
  [SMALL_STATE(7018)] = 253877,
  [SMALL_STATE(7019)] = 253894,
  [SMALL_STATE(7020)] = 253911,
  [SMALL_STATE(7021)] = 253928,
  [SMALL_STATE(7022)] = 253945,
  [SMALL_STATE(7023)] = 253962,
  [SMALL_STATE(7024)] = 253979,
  [SMALL_STATE(7025)] = 253996,
  [SMALL_STATE(7026)] = 254013,
  [SMALL_STATE(7027)] = 254030,
  [SMALL_STATE(7028)] = 254047,
  [SMALL_STATE(7029)] = 254064,
  [SMALL_STATE(7030)] = 254081,
  [SMALL_STATE(7031)] = 254098,
  [SMALL_STATE(7032)] = 254113,
  [SMALL_STATE(7033)] = 254130,
  [SMALL_STATE(7034)] = 254147,
  [SMALL_STATE(7035)] = 254164,
  [SMALL_STATE(7036)] = 254181,
  [SMALL_STATE(7037)] = 254198,
  [SMALL_STATE(7038)] = 254215,
  [SMALL_STATE(7039)] = 254232,
  [SMALL_STATE(7040)] = 254249,
  [SMALL_STATE(7041)] = 254264,
  [SMALL_STATE(7042)] = 254281,
  [SMALL_STATE(7043)] = 254298,
  [SMALL_STATE(7044)] = 254315,
  [SMALL_STATE(7045)] = 254332,
  [SMALL_STATE(7046)] = 254349,
  [SMALL_STATE(7047)] = 254364,
  [SMALL_STATE(7048)] = 254381,
  [SMALL_STATE(7049)] = 254398,
  [SMALL_STATE(7050)] = 254415,
  [SMALL_STATE(7051)] = 254430,
  [SMALL_STATE(7052)] = 254445,
  [SMALL_STATE(7053)] = 254462,
  [SMALL_STATE(7054)] = 254477,
  [SMALL_STATE(7055)] = 254494,
  [SMALL_STATE(7056)] = 254509,
  [SMALL_STATE(7057)] = 254526,
  [SMALL_STATE(7058)] = 254541,
  [SMALL_STATE(7059)] = 254556,
  [SMALL_STATE(7060)] = 254573,
  [SMALL_STATE(7061)] = 254588,
  [SMALL_STATE(7062)] = 254605,
  [SMALL_STATE(7063)] = 254622,
  [SMALL_STATE(7064)] = 254639,
  [SMALL_STATE(7065)] = 254656,
  [SMALL_STATE(7066)] = 254673,
  [SMALL_STATE(7067)] = 254690,
  [SMALL_STATE(7068)] = 254705,
  [SMALL_STATE(7069)] = 254722,
  [SMALL_STATE(7070)] = 254737,
  [SMALL_STATE(7071)] = 254754,
  [SMALL_STATE(7072)] = 254771,
  [SMALL_STATE(7073)] = 254788,
  [SMALL_STATE(7074)] = 254805,
  [SMALL_STATE(7075)] = 254820,
  [SMALL_STATE(7076)] = 254837,
  [SMALL_STATE(7077)] = 254854,
  [SMALL_STATE(7078)] = 254871,
  [SMALL_STATE(7079)] = 254886,
  [SMALL_STATE(7080)] = 254903,
  [SMALL_STATE(7081)] = 254920,
  [SMALL_STATE(7082)] = 254935,
  [SMALL_STATE(7083)] = 254952,
  [SMALL_STATE(7084)] = 254967,
  [SMALL_STATE(7085)] = 254982,
  [SMALL_STATE(7086)] = 254999,
  [SMALL_STATE(7087)] = 255016,
  [SMALL_STATE(7088)] = 255033,
  [SMALL_STATE(7089)] = 255050,
  [SMALL_STATE(7090)] = 255065,
  [SMALL_STATE(7091)] = 255082,
  [SMALL_STATE(7092)] = 255099,
  [SMALL_STATE(7093)] = 255116,
  [SMALL_STATE(7094)] = 255133,
  [SMALL_STATE(7095)] = 255150,
  [SMALL_STATE(7096)] = 255167,
  [SMALL_STATE(7097)] = 255184,
  [SMALL_STATE(7098)] = 255201,
  [SMALL_STATE(7099)] = 255218,
  [SMALL_STATE(7100)] = 255235,
  [SMALL_STATE(7101)] = 255252,
  [SMALL_STATE(7102)] = 255269,
  [SMALL_STATE(7103)] = 255286,
  [SMALL_STATE(7104)] = 255303,
  [SMALL_STATE(7105)] = 255320,
  [SMALL_STATE(7106)] = 255337,
  [SMALL_STATE(7107)] = 255354,
  [SMALL_STATE(7108)] = 255369,
  [SMALL_STATE(7109)] = 255384,
  [SMALL_STATE(7110)] = 255401,
  [SMALL_STATE(7111)] = 255416,
  [SMALL_STATE(7112)] = 255431,
  [SMALL_STATE(7113)] = 255448,
  [SMALL_STATE(7114)] = 255463,
  [SMALL_STATE(7115)] = 255480,
  [SMALL_STATE(7116)] = 255497,
  [SMALL_STATE(7117)] = 255514,
  [SMALL_STATE(7118)] = 255529,
  [SMALL_STATE(7119)] = 255544,
  [SMALL_STATE(7120)] = 255561,
  [SMALL_STATE(7121)] = 255576,
  [SMALL_STATE(7122)] = 255591,
  [SMALL_STATE(7123)] = 255606,
  [SMALL_STATE(7124)] = 255623,
  [SMALL_STATE(7125)] = 255640,
  [SMALL_STATE(7126)] = 255657,
  [SMALL_STATE(7127)] = 255674,
  [SMALL_STATE(7128)] = 255691,
  [SMALL_STATE(7129)] = 255708,
  [SMALL_STATE(7130)] = 255725,
  [SMALL_STATE(7131)] = 255740,
  [SMALL_STATE(7132)] = 255757,
  [SMALL_STATE(7133)] = 255774,
  [SMALL_STATE(7134)] = 255791,
  [SMALL_STATE(7135)] = 255808,
  [SMALL_STATE(7136)] = 255823,
  [SMALL_STATE(7137)] = 255840,
  [SMALL_STATE(7138)] = 255857,
  [SMALL_STATE(7139)] = 255874,
  [SMALL_STATE(7140)] = 255889,
  [SMALL_STATE(7141)] = 255904,
  [SMALL_STATE(7142)] = 255921,
  [SMALL_STATE(7143)] = 255938,
  [SMALL_STATE(7144)] = 255955,
  [SMALL_STATE(7145)] = 255970,
  [SMALL_STATE(7146)] = 255985,
  [SMALL_STATE(7147)] = 256002,
  [SMALL_STATE(7148)] = 256019,
  [SMALL_STATE(7149)] = 256034,
  [SMALL_STATE(7150)] = 256051,
  [SMALL_STATE(7151)] = 256068,
  [SMALL_STATE(7152)] = 256085,
  [SMALL_STATE(7153)] = 256102,
  [SMALL_STATE(7154)] = 256119,
  [SMALL_STATE(7155)] = 256136,
  [SMALL_STATE(7156)] = 256153,
  [SMALL_STATE(7157)] = 256170,
  [SMALL_STATE(7158)] = 256187,
  [SMALL_STATE(7159)] = 256204,
  [SMALL_STATE(7160)] = 256219,
  [SMALL_STATE(7161)] = 256236,
  [SMALL_STATE(7162)] = 256253,
  [SMALL_STATE(7163)] = 256270,
  [SMALL_STATE(7164)] = 256287,
  [SMALL_STATE(7165)] = 256304,
  [SMALL_STATE(7166)] = 256321,
  [SMALL_STATE(7167)] = 256338,
  [SMALL_STATE(7168)] = 256355,
  [SMALL_STATE(7169)] = 256372,
  [SMALL_STATE(7170)] = 256387,
  [SMALL_STATE(7171)] = 256404,
  [SMALL_STATE(7172)] = 256419,
  [SMALL_STATE(7173)] = 256436,
  [SMALL_STATE(7174)] = 256451,
  [SMALL_STATE(7175)] = 256468,
  [SMALL_STATE(7176)] = 256483,
  [SMALL_STATE(7177)] = 256500,
  [SMALL_STATE(7178)] = 256515,
  [SMALL_STATE(7179)] = 256532,
  [SMALL_STATE(7180)] = 256549,
  [SMALL_STATE(7181)] = 256564,
  [SMALL_STATE(7182)] = 256581,
  [SMALL_STATE(7183)] = 256598,
  [SMALL_STATE(7184)] = 256615,
  [SMALL_STATE(7185)] = 256629,
  [SMALL_STATE(7186)] = 256643,
  [SMALL_STATE(7187)] = 256657,
  [SMALL_STATE(7188)] = 256671,
  [SMALL_STATE(7189)] = 256685,
  [SMALL_STATE(7190)] = 256699,
  [SMALL_STATE(7191)] = 256713,
  [SMALL_STATE(7192)] = 256727,
  [SMALL_STATE(7193)] = 256741,
  [SMALL_STATE(7194)] = 256755,
  [SMALL_STATE(7195)] = 256769,
  [SMALL_STATE(7196)] = 256783,
  [SMALL_STATE(7197)] = 256797,
  [SMALL_STATE(7198)] = 256811,
  [SMALL_STATE(7199)] = 256825,
  [SMALL_STATE(7200)] = 256839,
  [SMALL_STATE(7201)] = 256853,
  [SMALL_STATE(7202)] = 256867,
  [SMALL_STATE(7203)] = 256881,
  [SMALL_STATE(7204)] = 256895,
  [SMALL_STATE(7205)] = 256909,
  [SMALL_STATE(7206)] = 256923,
  [SMALL_STATE(7207)] = 256937,
  [SMALL_STATE(7208)] = 256951,
  [SMALL_STATE(7209)] = 256965,
  [SMALL_STATE(7210)] = 256979,
  [SMALL_STATE(7211)] = 256993,
  [SMALL_STATE(7212)] = 257007,
  [SMALL_STATE(7213)] = 257021,
  [SMALL_STATE(7214)] = 257035,
  [SMALL_STATE(7215)] = 257049,
  [SMALL_STATE(7216)] = 257063,
  [SMALL_STATE(7217)] = 257077,
  [SMALL_STATE(7218)] = 257091,
  [SMALL_STATE(7219)] = 257105,
  [SMALL_STATE(7220)] = 257119,
  [SMALL_STATE(7221)] = 257133,
  [SMALL_STATE(7222)] = 257147,
  [SMALL_STATE(7223)] = 257161,
  [SMALL_STATE(7224)] = 257175,
  [SMALL_STATE(7225)] = 257189,
  [SMALL_STATE(7226)] = 257203,
  [SMALL_STATE(7227)] = 257217,
  [SMALL_STATE(7228)] = 257231,
  [SMALL_STATE(7229)] = 257245,
  [SMALL_STATE(7230)] = 257259,
  [SMALL_STATE(7231)] = 257273,
  [SMALL_STATE(7232)] = 257287,
  [SMALL_STATE(7233)] = 257301,
  [SMALL_STATE(7234)] = 257315,
  [SMALL_STATE(7235)] = 257329,
  [SMALL_STATE(7236)] = 257343,
  [SMALL_STATE(7237)] = 257357,
  [SMALL_STATE(7238)] = 257371,
  [SMALL_STATE(7239)] = 257385,
  [SMALL_STATE(7240)] = 257399,
  [SMALL_STATE(7241)] = 257413,
  [SMALL_STATE(7242)] = 257427,
  [SMALL_STATE(7243)] = 257441,
  [SMALL_STATE(7244)] = 257455,
  [SMALL_STATE(7245)] = 257469,
  [SMALL_STATE(7246)] = 257483,
  [SMALL_STATE(7247)] = 257497,
  [SMALL_STATE(7248)] = 257511,
  [SMALL_STATE(7249)] = 257525,
  [SMALL_STATE(7250)] = 257539,
  [SMALL_STATE(7251)] = 257553,
  [SMALL_STATE(7252)] = 257567,
  [SMALL_STATE(7253)] = 257581,
  [SMALL_STATE(7254)] = 257595,
  [SMALL_STATE(7255)] = 257609,
  [SMALL_STATE(7256)] = 257623,
  [SMALL_STATE(7257)] = 257637,
  [SMALL_STATE(7258)] = 257651,
  [SMALL_STATE(7259)] = 257665,
  [SMALL_STATE(7260)] = 257679,
  [SMALL_STATE(7261)] = 257693,
  [SMALL_STATE(7262)] = 257707,
  [SMALL_STATE(7263)] = 257721,
  [SMALL_STATE(7264)] = 257735,
  [SMALL_STATE(7265)] = 257749,
  [SMALL_STATE(7266)] = 257763,
  [SMALL_STATE(7267)] = 257777,
  [SMALL_STATE(7268)] = 257791,
  [SMALL_STATE(7269)] = 257805,
  [SMALL_STATE(7270)] = 257819,
  [SMALL_STATE(7271)] = 257833,
  [SMALL_STATE(7272)] = 257847,
  [SMALL_STATE(7273)] = 257861,
  [SMALL_STATE(7274)] = 257875,
  [SMALL_STATE(7275)] = 257889,
  [SMALL_STATE(7276)] = 257903,
  [SMALL_STATE(7277)] = 257917,
  [SMALL_STATE(7278)] = 257931,
  [SMALL_STATE(7279)] = 257945,
  [SMALL_STATE(7280)] = 257959,
  [SMALL_STATE(7281)] = 257973,
  [SMALL_STATE(7282)] = 257987,
  [SMALL_STATE(7283)] = 258001,
  [SMALL_STATE(7284)] = 258015,
  [SMALL_STATE(7285)] = 258029,
  [SMALL_STATE(7286)] = 258043,
  [SMALL_STATE(7287)] = 258057,
  [SMALL_STATE(7288)] = 258071,
  [SMALL_STATE(7289)] = 258085,
  [SMALL_STATE(7290)] = 258099,
  [SMALL_STATE(7291)] = 258113,
  [SMALL_STATE(7292)] = 258127,
  [SMALL_STATE(7293)] = 258141,
  [SMALL_STATE(7294)] = 258155,
  [SMALL_STATE(7295)] = 258169,
  [SMALL_STATE(7296)] = 258183,
  [SMALL_STATE(7297)] = 258197,
  [SMALL_STATE(7298)] = 258211,
  [SMALL_STATE(7299)] = 258225,
  [SMALL_STATE(7300)] = 258239,
  [SMALL_STATE(7301)] = 258253,
  [SMALL_STATE(7302)] = 258267,
  [SMALL_STATE(7303)] = 258281,
  [SMALL_STATE(7304)] = 258295,
  [SMALL_STATE(7305)] = 258309,
  [SMALL_STATE(7306)] = 258323,
  [SMALL_STATE(7307)] = 258337,
  [SMALL_STATE(7308)] = 258351,
  [SMALL_STATE(7309)] = 258365,
  [SMALL_STATE(7310)] = 258379,
  [SMALL_STATE(7311)] = 258393,
  [SMALL_STATE(7312)] = 258407,
  [SMALL_STATE(7313)] = 258421,
  [SMALL_STATE(7314)] = 258435,
  [SMALL_STATE(7315)] = 258449,
  [SMALL_STATE(7316)] = 258463,
  [SMALL_STATE(7317)] = 258477,
  [SMALL_STATE(7318)] = 258491,
  [SMALL_STATE(7319)] = 258505,
  [SMALL_STATE(7320)] = 258519,
  [SMALL_STATE(7321)] = 258533,
  [SMALL_STATE(7322)] = 258547,
  [SMALL_STATE(7323)] = 258561,
  [SMALL_STATE(7324)] = 258575,
  [SMALL_STATE(7325)] = 258589,
  [SMALL_STATE(7326)] = 258603,
  [SMALL_STATE(7327)] = 258617,
  [SMALL_STATE(7328)] = 258631,
  [SMALL_STATE(7329)] = 258645,
  [SMALL_STATE(7330)] = 258659,
  [SMALL_STATE(7331)] = 258673,
  [SMALL_STATE(7332)] = 258687,
  [SMALL_STATE(7333)] = 258701,
  [SMALL_STATE(7334)] = 258715,
  [SMALL_STATE(7335)] = 258729,
  [SMALL_STATE(7336)] = 258743,
  [SMALL_STATE(7337)] = 258757,
  [SMALL_STATE(7338)] = 258771,
  [SMALL_STATE(7339)] = 258785,
  [SMALL_STATE(7340)] = 258799,
  [SMALL_STATE(7341)] = 258813,
  [SMALL_STATE(7342)] = 258827,
  [SMALL_STATE(7343)] = 258841,
  [SMALL_STATE(7344)] = 258855,
  [SMALL_STATE(7345)] = 258869,
  [SMALL_STATE(7346)] = 258883,
  [SMALL_STATE(7347)] = 258897,
  [SMALL_STATE(7348)] = 258911,
  [SMALL_STATE(7349)] = 258925,
  [SMALL_STATE(7350)] = 258939,
  [SMALL_STATE(7351)] = 258953,
  [SMALL_STATE(7352)] = 258967,
  [SMALL_STATE(7353)] = 258981,
  [SMALL_STATE(7354)] = 258995,
  [SMALL_STATE(7355)] = 259009,
  [SMALL_STATE(7356)] = 259023,
  [SMALL_STATE(7357)] = 259037,
  [SMALL_STATE(7358)] = 259051,
  [SMALL_STATE(7359)] = 259065,
  [SMALL_STATE(7360)] = 259079,
  [SMALL_STATE(7361)] = 259093,
  [SMALL_STATE(7362)] = 259107,
  [SMALL_STATE(7363)] = 259121,
  [SMALL_STATE(7364)] = 259135,
  [SMALL_STATE(7365)] = 259149,
  [SMALL_STATE(7366)] = 259163,
  [SMALL_STATE(7367)] = 259177,
  [SMALL_STATE(7368)] = 259191,
  [SMALL_STATE(7369)] = 259205,
  [SMALL_STATE(7370)] = 259219,
  [SMALL_STATE(7371)] = 259233,
  [SMALL_STATE(7372)] = 259247,
  [SMALL_STATE(7373)] = 259261,
  [SMALL_STATE(7374)] = 259275,
  [SMALL_STATE(7375)] = 259289,
  [SMALL_STATE(7376)] = 259303,
  [SMALL_STATE(7377)] = 259317,
  [SMALL_STATE(7378)] = 259331,
  [SMALL_STATE(7379)] = 259345,
  [SMALL_STATE(7380)] = 259359,
  [SMALL_STATE(7381)] = 259373,
  [SMALL_STATE(7382)] = 259387,
  [SMALL_STATE(7383)] = 259401,
  [SMALL_STATE(7384)] = 259415,
  [SMALL_STATE(7385)] = 259429,
  [SMALL_STATE(7386)] = 259443,
  [SMALL_STATE(7387)] = 259457,
  [SMALL_STATE(7388)] = 259471,
  [SMALL_STATE(7389)] = 259485,
  [SMALL_STATE(7390)] = 259499,
  [SMALL_STATE(7391)] = 259513,
  [SMALL_STATE(7392)] = 259527,
  [SMALL_STATE(7393)] = 259541,
  [SMALL_STATE(7394)] = 259555,
  [SMALL_STATE(7395)] = 259569,
  [SMALL_STATE(7396)] = 259583,
  [SMALL_STATE(7397)] = 259597,
  [SMALL_STATE(7398)] = 259611,
  [SMALL_STATE(7399)] = 259625,
  [SMALL_STATE(7400)] = 259639,
  [SMALL_STATE(7401)] = 259653,
  [SMALL_STATE(7402)] = 259667,
  [SMALL_STATE(7403)] = 259681,
  [SMALL_STATE(7404)] = 259695,
  [SMALL_STATE(7405)] = 259709,
  [SMALL_STATE(7406)] = 259723,
  [SMALL_STATE(7407)] = 259737,
  [SMALL_STATE(7408)] = 259751,
  [SMALL_STATE(7409)] = 259765,
  [SMALL_STATE(7410)] = 259779,
  [SMALL_STATE(7411)] = 259793,
  [SMALL_STATE(7412)] = 259807,
  [SMALL_STATE(7413)] = 259821,
  [SMALL_STATE(7414)] = 259835,
  [SMALL_STATE(7415)] = 259849,
  [SMALL_STATE(7416)] = 259863,
  [SMALL_STATE(7417)] = 259877,
  [SMALL_STATE(7418)] = 259891,
  [SMALL_STATE(7419)] = 259905,
  [SMALL_STATE(7420)] = 259919,
  [SMALL_STATE(7421)] = 259933,
  [SMALL_STATE(7422)] = 259947,
  [SMALL_STATE(7423)] = 259961,
  [SMALL_STATE(7424)] = 259975,
  [SMALL_STATE(7425)] = 259989,
  [SMALL_STATE(7426)] = 260003,
  [SMALL_STATE(7427)] = 260017,
  [SMALL_STATE(7428)] = 260031,
  [SMALL_STATE(7429)] = 260045,
  [SMALL_STATE(7430)] = 260059,
  [SMALL_STATE(7431)] = 260073,
  [SMALL_STATE(7432)] = 260087,
  [SMALL_STATE(7433)] = 260101,
  [SMALL_STATE(7434)] = 260115,
  [SMALL_STATE(7435)] = 260129,
  [SMALL_STATE(7436)] = 260143,
  [SMALL_STATE(7437)] = 260157,
  [SMALL_STATE(7438)] = 260171,
  [SMALL_STATE(7439)] = 260185,
  [SMALL_STATE(7440)] = 260199,
  [SMALL_STATE(7441)] = 260213,
  [SMALL_STATE(7442)] = 260227,
  [SMALL_STATE(7443)] = 260241,
  [SMALL_STATE(7444)] = 260255,
  [SMALL_STATE(7445)] = 260269,
  [SMALL_STATE(7446)] = 260283,
  [SMALL_STATE(7447)] = 260297,
  [SMALL_STATE(7448)] = 260311,
  [SMALL_STATE(7449)] = 260325,
  [SMALL_STATE(7450)] = 260339,
  [SMALL_STATE(7451)] = 260353,
  [SMALL_STATE(7452)] = 260367,
  [SMALL_STATE(7453)] = 260381,
  [SMALL_STATE(7454)] = 260395,
  [SMALL_STATE(7455)] = 260409,
  [SMALL_STATE(7456)] = 260423,
  [SMALL_STATE(7457)] = 260437,
  [SMALL_STATE(7458)] = 260451,
  [SMALL_STATE(7459)] = 260465,
  [SMALL_STATE(7460)] = 260479,
  [SMALL_STATE(7461)] = 260493,
  [SMALL_STATE(7462)] = 260507,
  [SMALL_STATE(7463)] = 260521,
  [SMALL_STATE(7464)] = 260535,
  [SMALL_STATE(7465)] = 260549,
  [SMALL_STATE(7466)] = 260563,
  [SMALL_STATE(7467)] = 260577,
  [SMALL_STATE(7468)] = 260591,
  [SMALL_STATE(7469)] = 260605,
  [SMALL_STATE(7470)] = 260619,
  [SMALL_STATE(7471)] = 260633,
  [SMALL_STATE(7472)] = 260647,
  [SMALL_STATE(7473)] = 260661,
  [SMALL_STATE(7474)] = 260675,
  [SMALL_STATE(7475)] = 260689,
  [SMALL_STATE(7476)] = 260703,
  [SMALL_STATE(7477)] = 260717,
  [SMALL_STATE(7478)] = 260731,
  [SMALL_STATE(7479)] = 260745,
  [SMALL_STATE(7480)] = 260759,
  [SMALL_STATE(7481)] = 260773,
  [SMALL_STATE(7482)] = 260787,
  [SMALL_STATE(7483)] = 260801,
  [SMALL_STATE(7484)] = 260815,
  [SMALL_STATE(7485)] = 260829,
  [SMALL_STATE(7486)] = 260843,
  [SMALL_STATE(7487)] = 260857,
  [SMALL_STATE(7488)] = 260871,
  [SMALL_STATE(7489)] = 260885,
  [SMALL_STATE(7490)] = 260899,
  [SMALL_STATE(7491)] = 260913,
  [SMALL_STATE(7492)] = 260927,
  [SMALL_STATE(7493)] = 260941,
  [SMALL_STATE(7494)] = 260955,
  [SMALL_STATE(7495)] = 260969,
  [SMALL_STATE(7496)] = 260983,
  [SMALL_STATE(7497)] = 260997,
  [SMALL_STATE(7498)] = 261011,
  [SMALL_STATE(7499)] = 261025,
  [SMALL_STATE(7500)] = 261039,
  [SMALL_STATE(7501)] = 261053,
  [SMALL_STATE(7502)] = 261067,
  [SMALL_STATE(7503)] = 261081,
  [SMALL_STATE(7504)] = 261095,
  [SMALL_STATE(7505)] = 261109,
  [SMALL_STATE(7506)] = 261123,
  [SMALL_STATE(7507)] = 261137,
  [SMALL_STATE(7508)] = 261151,
  [SMALL_STATE(7509)] = 261165,
  [SMALL_STATE(7510)] = 261179,
  [SMALL_STATE(7511)] = 261193,
  [SMALL_STATE(7512)] = 261207,
  [SMALL_STATE(7513)] = 261221,
  [SMALL_STATE(7514)] = 261235,
  [SMALL_STATE(7515)] = 261249,
  [SMALL_STATE(7516)] = 261263,
  [SMALL_STATE(7517)] = 261277,
  [SMALL_STATE(7518)] = 261291,
  [SMALL_STATE(7519)] = 261305,
  [SMALL_STATE(7520)] = 261319,
  [SMALL_STATE(7521)] = 261333,
  [SMALL_STATE(7522)] = 261347,
  [SMALL_STATE(7523)] = 261361,
  [SMALL_STATE(7524)] = 261375,
  [SMALL_STATE(7525)] = 261389,
  [SMALL_STATE(7526)] = 261403,
  [SMALL_STATE(7527)] = 261417,
  [SMALL_STATE(7528)] = 261431,
  [SMALL_STATE(7529)] = 261445,
  [SMALL_STATE(7530)] = 261459,
  [SMALL_STATE(7531)] = 261473,
  [SMALL_STATE(7532)] = 261487,
  [SMALL_STATE(7533)] = 261501,
  [SMALL_STATE(7534)] = 261515,
  [SMALL_STATE(7535)] = 261529,
  [SMALL_STATE(7536)] = 261543,
  [SMALL_STATE(7537)] = 261557,
  [SMALL_STATE(7538)] = 261571,
  [SMALL_STATE(7539)] = 261585,
  [SMALL_STATE(7540)] = 261599,
  [SMALL_STATE(7541)] = 261613,
  [SMALL_STATE(7542)] = 261627,
  [SMALL_STATE(7543)] = 261641,
  [SMALL_STATE(7544)] = 261655,
  [SMALL_STATE(7545)] = 261669,
  [SMALL_STATE(7546)] = 261683,
  [SMALL_STATE(7547)] = 261697,
  [SMALL_STATE(7548)] = 261711,
  [SMALL_STATE(7549)] = 261725,
  [SMALL_STATE(7550)] = 261739,
  [SMALL_STATE(7551)] = 261753,
  [SMALL_STATE(7552)] = 261767,
  [SMALL_STATE(7553)] = 261781,
  [SMALL_STATE(7554)] = 261795,
  [SMALL_STATE(7555)] = 261809,
  [SMALL_STATE(7556)] = 261823,
  [SMALL_STATE(7557)] = 261837,
  [SMALL_STATE(7558)] = 261851,
  [SMALL_STATE(7559)] = 261865,
  [SMALL_STATE(7560)] = 261879,
  [SMALL_STATE(7561)] = 261893,
  [SMALL_STATE(7562)] = 261907,
  [SMALL_STATE(7563)] = 261921,
  [SMALL_STATE(7564)] = 261935,
  [SMALL_STATE(7565)] = 261949,
  [SMALL_STATE(7566)] = 261963,
  [SMALL_STATE(7567)] = 261977,
  [SMALL_STATE(7568)] = 261991,
  [SMALL_STATE(7569)] = 262005,
  [SMALL_STATE(7570)] = 262019,
  [SMALL_STATE(7571)] = 262033,
  [SMALL_STATE(7572)] = 262047,
  [SMALL_STATE(7573)] = 262061,
  [SMALL_STATE(7574)] = 262075,
  [SMALL_STATE(7575)] = 262089,
  [SMALL_STATE(7576)] = 262103,
  [SMALL_STATE(7577)] = 262117,
  [SMALL_STATE(7578)] = 262131,
  [SMALL_STATE(7579)] = 262145,
  [SMALL_STATE(7580)] = 262159,
  [SMALL_STATE(7581)] = 262173,
  [SMALL_STATE(7582)] = 262187,
  [SMALL_STATE(7583)] = 262201,
  [SMALL_STATE(7584)] = 262215,
  [SMALL_STATE(7585)] = 262229,
  [SMALL_STATE(7586)] = 262243,
  [SMALL_STATE(7587)] = 262257,
  [SMALL_STATE(7588)] = 262271,
  [SMALL_STATE(7589)] = 262285,
  [SMALL_STATE(7590)] = 262299,
  [SMALL_STATE(7591)] = 262313,
  [SMALL_STATE(7592)] = 262327,
  [SMALL_STATE(7593)] = 262341,
  [SMALL_STATE(7594)] = 262355,
  [SMALL_STATE(7595)] = 262369,
  [SMALL_STATE(7596)] = 262383,
  [SMALL_STATE(7597)] = 262397,
  [SMALL_STATE(7598)] = 262411,
  [SMALL_STATE(7599)] = 262425,
  [SMALL_STATE(7600)] = 262439,
  [SMALL_STATE(7601)] = 262453,
  [SMALL_STATE(7602)] = 262467,
  [SMALL_STATE(7603)] = 262481,
  [SMALL_STATE(7604)] = 262495,
  [SMALL_STATE(7605)] = 262509,
  [SMALL_STATE(7606)] = 262523,
  [SMALL_STATE(7607)] = 262537,
  [SMALL_STATE(7608)] = 262551,
  [SMALL_STATE(7609)] = 262565,
  [SMALL_STATE(7610)] = 262579,
  [SMALL_STATE(7611)] = 262593,
  [SMALL_STATE(7612)] = 262607,
  [SMALL_STATE(7613)] = 262621,
  [SMALL_STATE(7614)] = 262635,
  [SMALL_STATE(7615)] = 262649,
  [SMALL_STATE(7616)] = 262663,
  [SMALL_STATE(7617)] = 262677,
  [SMALL_STATE(7618)] = 262691,
  [SMALL_STATE(7619)] = 262705,
  [SMALL_STATE(7620)] = 262719,
  [SMALL_STATE(7621)] = 262733,
  [SMALL_STATE(7622)] = 262747,
  [SMALL_STATE(7623)] = 262761,
  [SMALL_STATE(7624)] = 262775,
  [SMALL_STATE(7625)] = 262789,
  [SMALL_STATE(7626)] = 262803,
  [SMALL_STATE(7627)] = 262817,
  [SMALL_STATE(7628)] = 262831,
  [SMALL_STATE(7629)] = 262845,
  [SMALL_STATE(7630)] = 262859,
  [SMALL_STATE(7631)] = 262873,
  [SMALL_STATE(7632)] = 262887,
  [SMALL_STATE(7633)] = 262901,
  [SMALL_STATE(7634)] = 262915,
  [SMALL_STATE(7635)] = 262929,
  [SMALL_STATE(7636)] = 262943,
  [SMALL_STATE(7637)] = 262957,
  [SMALL_STATE(7638)] = 262971,
  [SMALL_STATE(7639)] = 262985,
  [SMALL_STATE(7640)] = 262999,
  [SMALL_STATE(7641)] = 263013,
  [SMALL_STATE(7642)] = 263027,
  [SMALL_STATE(7643)] = 263041,
  [SMALL_STATE(7644)] = 263055,
  [SMALL_STATE(7645)] = 263069,
  [SMALL_STATE(7646)] = 263083,
  [SMALL_STATE(7647)] = 263097,
  [SMALL_STATE(7648)] = 263111,
  [SMALL_STATE(7649)] = 263125,
  [SMALL_STATE(7650)] = 263139,
  [SMALL_STATE(7651)] = 263153,
  [SMALL_STATE(7652)] = 263167,
  [SMALL_STATE(7653)] = 263181,
  [SMALL_STATE(7654)] = 263195,
  [SMALL_STATE(7655)] = 263209,
  [SMALL_STATE(7656)] = 263223,
  [SMALL_STATE(7657)] = 263237,
  [SMALL_STATE(7658)] = 263251,
  [SMALL_STATE(7659)] = 263265,
  [SMALL_STATE(7660)] = 263279,
  [SMALL_STATE(7661)] = 263293,
  [SMALL_STATE(7662)] = 263307,
  [SMALL_STATE(7663)] = 263321,
  [SMALL_STATE(7664)] = 263335,
  [SMALL_STATE(7665)] = 263349,
  [SMALL_STATE(7666)] = 263363,
  [SMALL_STATE(7667)] = 263377,
  [SMALL_STATE(7668)] = 263391,
  [SMALL_STATE(7669)] = 263405,
  [SMALL_STATE(7670)] = 263419,
  [SMALL_STATE(7671)] = 263433,
  [SMALL_STATE(7672)] = 263447,
  [SMALL_STATE(7673)] = 263461,
  [SMALL_STATE(7674)] = 263475,
  [SMALL_STATE(7675)] = 263489,
  [SMALL_STATE(7676)] = 263503,
  [SMALL_STATE(7677)] = 263517,
  [SMALL_STATE(7678)] = 263531,
  [SMALL_STATE(7679)] = 263545,
  [SMALL_STATE(7680)] = 263559,
  [SMALL_STATE(7681)] = 263573,
  [SMALL_STATE(7682)] = 263587,
  [SMALL_STATE(7683)] = 263601,
  [SMALL_STATE(7684)] = 263615,
  [SMALL_STATE(7685)] = 263629,
  [SMALL_STATE(7686)] = 263643,
  [SMALL_STATE(7687)] = 263657,
  [SMALL_STATE(7688)] = 263671,
  [SMALL_STATE(7689)] = 263685,
  [SMALL_STATE(7690)] = 263699,
  [SMALL_STATE(7691)] = 263713,
  [SMALL_STATE(7692)] = 263727,
  [SMALL_STATE(7693)] = 263741,
  [SMALL_STATE(7694)] = 263755,
  [SMALL_STATE(7695)] = 263769,
  [SMALL_STATE(7696)] = 263783,
  [SMALL_STATE(7697)] = 263797,
  [SMALL_STATE(7698)] = 263811,
  [SMALL_STATE(7699)] = 263825,
  [SMALL_STATE(7700)] = 263839,
  [SMALL_STATE(7701)] = 263853,
  [SMALL_STATE(7702)] = 263867,
  [SMALL_STATE(7703)] = 263881,
  [SMALL_STATE(7704)] = 263895,
  [SMALL_STATE(7705)] = 263909,
  [SMALL_STATE(7706)] = 263923,
  [SMALL_STATE(7707)] = 263937,
  [SMALL_STATE(7708)] = 263951,
  [SMALL_STATE(7709)] = 263965,
  [SMALL_STATE(7710)] = 263979,
  [SMALL_STATE(7711)] = 263993,
  [SMALL_STATE(7712)] = 264007,
  [SMALL_STATE(7713)] = 264021,
  [SMALL_STATE(7714)] = 264035,
  [SMALL_STATE(7715)] = 264049,
  [SMALL_STATE(7716)] = 264063,
  [SMALL_STATE(7717)] = 264077,
  [SMALL_STATE(7718)] = 264091,
  [SMALL_STATE(7719)] = 264105,
  [SMALL_STATE(7720)] = 264119,
  [SMALL_STATE(7721)] = 264133,
  [SMALL_STATE(7722)] = 264147,
  [SMALL_STATE(7723)] = 264161,
  [SMALL_STATE(7724)] = 264175,
  [SMALL_STATE(7725)] = 264189,
  [SMALL_STATE(7726)] = 264203,
  [SMALL_STATE(7727)] = 264217,
  [SMALL_STATE(7728)] = 264231,
  [SMALL_STATE(7729)] = 264245,
  [SMALL_STATE(7730)] = 264259,
  [SMALL_STATE(7731)] = 264273,
  [SMALL_STATE(7732)] = 264287,
  [SMALL_STATE(7733)] = 264301,
  [SMALL_STATE(7734)] = 264315,
  [SMALL_STATE(7735)] = 264329,
  [SMALL_STATE(7736)] = 264343,
  [SMALL_STATE(7737)] = 264357,
  [SMALL_STATE(7738)] = 264371,
  [SMALL_STATE(7739)] = 264385,
  [SMALL_STATE(7740)] = 264399,
  [SMALL_STATE(7741)] = 264413,
  [SMALL_STATE(7742)] = 264427,
  [SMALL_STATE(7743)] = 264441,
  [SMALL_STATE(7744)] = 264455,
  [SMALL_STATE(7745)] = 264469,
  [SMALL_STATE(7746)] = 264483,
  [SMALL_STATE(7747)] = 264497,
  [SMALL_STATE(7748)] = 264511,
  [SMALL_STATE(7749)] = 264525,
  [SMALL_STATE(7750)] = 264539,
  [SMALL_STATE(7751)] = 264553,
  [SMALL_STATE(7752)] = 264567,
  [SMALL_STATE(7753)] = 264581,
  [SMALL_STATE(7754)] = 264595,
  [SMALL_STATE(7755)] = 264609,
  [SMALL_STATE(7756)] = 264623,
  [SMALL_STATE(7757)] = 264637,
  [SMALL_STATE(7758)] = 264651,
  [SMALL_STATE(7759)] = 264665,
  [SMALL_STATE(7760)] = 264679,
  [SMALL_STATE(7761)] = 264693,
  [SMALL_STATE(7762)] = 264707,
  [SMALL_STATE(7763)] = 264721,
  [SMALL_STATE(7764)] = 264735,
  [SMALL_STATE(7765)] = 264749,
  [SMALL_STATE(7766)] = 264763,
  [SMALL_STATE(7767)] = 264777,
  [SMALL_STATE(7768)] = 264791,
  [SMALL_STATE(7769)] = 264805,
  [SMALL_STATE(7770)] = 264819,
  [SMALL_STATE(7771)] = 264833,
  [SMALL_STATE(7772)] = 264847,
  [SMALL_STATE(7773)] = 264861,
  [SMALL_STATE(7774)] = 264875,
  [SMALL_STATE(7775)] = 264889,
  [SMALL_STATE(7776)] = 264903,
  [SMALL_STATE(7777)] = 264917,
  [SMALL_STATE(7778)] = 264931,
  [SMALL_STATE(7779)] = 264945,
  [SMALL_STATE(7780)] = 264959,
  [SMALL_STATE(7781)] = 264973,
  [SMALL_STATE(7782)] = 264987,
  [SMALL_STATE(7783)] = 265001,
  [SMALL_STATE(7784)] = 265015,
  [SMALL_STATE(7785)] = 265029,
  [SMALL_STATE(7786)] = 265043,
  [SMALL_STATE(7787)] = 265057,
  [SMALL_STATE(7788)] = 265071,
  [SMALL_STATE(7789)] = 265085,
  [SMALL_STATE(7790)] = 265099,
  [SMALL_STATE(7791)] = 265113,
  [SMALL_STATE(7792)] = 265127,
  [SMALL_STATE(7793)] = 265141,
  [SMALL_STATE(7794)] = 265155,
  [SMALL_STATE(7795)] = 265169,
  [SMALL_STATE(7796)] = 265183,
  [SMALL_STATE(7797)] = 265197,
  [SMALL_STATE(7798)] = 265211,
  [SMALL_STATE(7799)] = 265225,
  [SMALL_STATE(7800)] = 265239,
  [SMALL_STATE(7801)] = 265253,
  [SMALL_STATE(7802)] = 265267,
  [SMALL_STATE(7803)] = 265281,
  [SMALL_STATE(7804)] = 265295,
  [SMALL_STATE(7805)] = 265309,
  [SMALL_STATE(7806)] = 265323,
  [SMALL_STATE(7807)] = 265337,
  [SMALL_STATE(7808)] = 265351,
  [SMALL_STATE(7809)] = 265365,
  [SMALL_STATE(7810)] = 265379,
  [SMALL_STATE(7811)] = 265393,
  [SMALL_STATE(7812)] = 265407,
  [SMALL_STATE(7813)] = 265421,
  [SMALL_STATE(7814)] = 265435,
  [SMALL_STATE(7815)] = 265449,
  [SMALL_STATE(7816)] = 265463,
  [SMALL_STATE(7817)] = 265477,
  [SMALL_STATE(7818)] = 265491,
  [SMALL_STATE(7819)] = 265505,
  [SMALL_STATE(7820)] = 265519,
  [SMALL_STATE(7821)] = 265533,
  [SMALL_STATE(7822)] = 265547,
  [SMALL_STATE(7823)] = 265561,
  [SMALL_STATE(7824)] = 265575,
  [SMALL_STATE(7825)] = 265589,
  [SMALL_STATE(7826)] = 265603,
  [SMALL_STATE(7827)] = 265617,
  [SMALL_STATE(7828)] = 265631,
  [SMALL_STATE(7829)] = 265645,
  [SMALL_STATE(7830)] = 265659,
  [SMALL_STATE(7831)] = 265673,
  [SMALL_STATE(7832)] = 265687,
  [SMALL_STATE(7833)] = 265701,
  [SMALL_STATE(7834)] = 265715,
  [SMALL_STATE(7835)] = 265729,
  [SMALL_STATE(7836)] = 265743,
  [SMALL_STATE(7837)] = 265757,
  [SMALL_STATE(7838)] = 265771,
  [SMALL_STATE(7839)] = 265785,
  [SMALL_STATE(7840)] = 265799,
  [SMALL_STATE(7841)] = 265813,
  [SMALL_STATE(7842)] = 265827,
  [SMALL_STATE(7843)] = 265841,
  [SMALL_STATE(7844)] = 265855,
  [SMALL_STATE(7845)] = 265869,
  [SMALL_STATE(7846)] = 265883,
  [SMALL_STATE(7847)] = 265897,
  [SMALL_STATE(7848)] = 265911,
  [SMALL_STATE(7849)] = 265925,
  [SMALL_STATE(7850)] = 265939,
  [SMALL_STATE(7851)] = 265953,
  [SMALL_STATE(7852)] = 265967,
  [SMALL_STATE(7853)] = 265981,
  [SMALL_STATE(7854)] = 265995,
  [SMALL_STATE(7855)] = 266009,
  [SMALL_STATE(7856)] = 266023,
  [SMALL_STATE(7857)] = 266037,
  [SMALL_STATE(7858)] = 266051,
  [SMALL_STATE(7859)] = 266065,
  [SMALL_STATE(7860)] = 266079,
  [SMALL_STATE(7861)] = 266093,
  [SMALL_STATE(7862)] = 266107,
  [SMALL_STATE(7863)] = 266121,
  [SMALL_STATE(7864)] = 266135,
  [SMALL_STATE(7865)] = 266149,
  [SMALL_STATE(7866)] = 266163,
  [SMALL_STATE(7867)] = 266177,
  [SMALL_STATE(7868)] = 266191,
  [SMALL_STATE(7869)] = 266205,
  [SMALL_STATE(7870)] = 266219,
  [SMALL_STATE(7871)] = 266233,
  [SMALL_STATE(7872)] = 266247,
  [SMALL_STATE(7873)] = 266261,
  [SMALL_STATE(7874)] = 266275,
  [SMALL_STATE(7875)] = 266289,
  [SMALL_STATE(7876)] = 266303,
  [SMALL_STATE(7877)] = 266317,
  [SMALL_STATE(7878)] = 266331,
  [SMALL_STATE(7879)] = 266345,
  [SMALL_STATE(7880)] = 266359,
  [SMALL_STATE(7881)] = 266373,
  [SMALL_STATE(7882)] = 266387,
  [SMALL_STATE(7883)] = 266401,
  [SMALL_STATE(7884)] = 266415,
  [SMALL_STATE(7885)] = 266429,
  [SMALL_STATE(7886)] = 266443,
  [SMALL_STATE(7887)] = 266457,
  [SMALL_STATE(7888)] = 266471,
  [SMALL_STATE(7889)] = 266485,
  [SMALL_STATE(7890)] = 266499,
  [SMALL_STATE(7891)] = 266513,
  [SMALL_STATE(7892)] = 266527,
  [SMALL_STATE(7893)] = 266541,
  [SMALL_STATE(7894)] = 266555,
  [SMALL_STATE(7895)] = 266569,
  [SMALL_STATE(7896)] = 266583,
  [SMALL_STATE(7897)] = 266597,
  [SMALL_STATE(7898)] = 266611,
  [SMALL_STATE(7899)] = 266625,
  [SMALL_STATE(7900)] = 266639,
  [SMALL_STATE(7901)] = 266653,
  [SMALL_STATE(7902)] = 266667,
  [SMALL_STATE(7903)] = 266681,
  [SMALL_STATE(7904)] = 266695,
  [SMALL_STATE(7905)] = 266709,
  [SMALL_STATE(7906)] = 266723,
  [SMALL_STATE(7907)] = 266737,
  [SMALL_STATE(7908)] = 266751,
  [SMALL_STATE(7909)] = 266765,
  [SMALL_STATE(7910)] = 266779,
  [SMALL_STATE(7911)] = 266793,
  [SMALL_STATE(7912)] = 266807,
  [SMALL_STATE(7913)] = 266821,
  [SMALL_STATE(7914)] = 266835,
  [SMALL_STATE(7915)] = 266849,
  [SMALL_STATE(7916)] = 266863,
  [SMALL_STATE(7917)] = 266877,
  [SMALL_STATE(7918)] = 266891,
  [SMALL_STATE(7919)] = 266905,
  [SMALL_STATE(7920)] = 266919,
  [SMALL_STATE(7921)] = 266933,
  [SMALL_STATE(7922)] = 266947,
  [SMALL_STATE(7923)] = 266961,
  [SMALL_STATE(7924)] = 266975,
  [SMALL_STATE(7925)] = 266989,
  [SMALL_STATE(7926)] = 267003,
  [SMALL_STATE(7927)] = 267017,
  [SMALL_STATE(7928)] = 267031,
  [SMALL_STATE(7929)] = 267045,
  [SMALL_STATE(7930)] = 267059,
  [SMALL_STATE(7931)] = 267073,
  [SMALL_STATE(7932)] = 267087,
  [SMALL_STATE(7933)] = 267101,
  [SMALL_STATE(7934)] = 267115,
  [SMALL_STATE(7935)] = 267129,
  [SMALL_STATE(7936)] = 267143,
  [SMALL_STATE(7937)] = 267157,
  [SMALL_STATE(7938)] = 267171,
  [SMALL_STATE(7939)] = 267185,
  [SMALL_STATE(7940)] = 267199,
  [SMALL_STATE(7941)] = 267213,
  [SMALL_STATE(7942)] = 267227,
  [SMALL_STATE(7943)] = 267241,
  [SMALL_STATE(7944)] = 267255,
  [SMALL_STATE(7945)] = 267269,
  [SMALL_STATE(7946)] = 267283,
  [SMALL_STATE(7947)] = 267297,
  [SMALL_STATE(7948)] = 267311,
  [SMALL_STATE(7949)] = 267325,
  [SMALL_STATE(7950)] = 267339,
  [SMALL_STATE(7951)] = 267353,
  [SMALL_STATE(7952)] = 267367,
  [SMALL_STATE(7953)] = 267381,
  [SMALL_STATE(7954)] = 267395,
  [SMALL_STATE(7955)] = 267409,
  [SMALL_STATE(7956)] = 267423,
  [SMALL_STATE(7957)] = 267437,
  [SMALL_STATE(7958)] = 267451,
  [SMALL_STATE(7959)] = 267465,
  [SMALL_STATE(7960)] = 267479,
  [SMALL_STATE(7961)] = 267493,
  [SMALL_STATE(7962)] = 267507,
  [SMALL_STATE(7963)] = 267521,
  [SMALL_STATE(7964)] = 267535,
  [SMALL_STATE(7965)] = 267549,
  [SMALL_STATE(7966)] = 267563,
  [SMALL_STATE(7967)] = 267577,
  [SMALL_STATE(7968)] = 267591,
  [SMALL_STATE(7969)] = 267605,
  [SMALL_STATE(7970)] = 267619,
  [SMALL_STATE(7971)] = 267633,
  [SMALL_STATE(7972)] = 267647,
  [SMALL_STATE(7973)] = 267661,
  [SMALL_STATE(7974)] = 267675,
  [SMALL_STATE(7975)] = 267689,
  [SMALL_STATE(7976)] = 267703,
  [SMALL_STATE(7977)] = 267717,
  [SMALL_STATE(7978)] = 267731,
  [SMALL_STATE(7979)] = 267745,
  [SMALL_STATE(7980)] = 267759,
  [SMALL_STATE(7981)] = 267773,
  [SMALL_STATE(7982)] = 267787,
  [SMALL_STATE(7983)] = 267801,
  [SMALL_STATE(7984)] = 267815,
  [SMALL_STATE(7985)] = 267829,
  [SMALL_STATE(7986)] = 267843,
  [SMALL_STATE(7987)] = 267857,
  [SMALL_STATE(7988)] = 267871,
  [SMALL_STATE(7989)] = 267885,
  [SMALL_STATE(7990)] = 267899,
  [SMALL_STATE(7991)] = 267913,
  [SMALL_STATE(7992)] = 267927,
  [SMALL_STATE(7993)] = 267941,
  [SMALL_STATE(7994)] = 267955,
  [SMALL_STATE(7995)] = 267969,
  [SMALL_STATE(7996)] = 267983,
  [SMALL_STATE(7997)] = 267997,
  [SMALL_STATE(7998)] = 268011,
  [SMALL_STATE(7999)] = 268025,
  [SMALL_STATE(8000)] = 268039,
  [SMALL_STATE(8001)] = 268053,
  [SMALL_STATE(8002)] = 268067,
  [SMALL_STATE(8003)] = 268081,
  [SMALL_STATE(8004)] = 268095,
  [SMALL_STATE(8005)] = 268109,
  [SMALL_STATE(8006)] = 268123,
  [SMALL_STATE(8007)] = 268137,
  [SMALL_STATE(8008)] = 268151,
  [SMALL_STATE(8009)] = 268165,
  [SMALL_STATE(8010)] = 268179,
  [SMALL_STATE(8011)] = 268193,
  [SMALL_STATE(8012)] = 268207,
  [SMALL_STATE(8013)] = 268221,
  [SMALL_STATE(8014)] = 268235,
  [SMALL_STATE(8015)] = 268249,
  [SMALL_STATE(8016)] = 268263,
  [SMALL_STATE(8017)] = 268277,
  [SMALL_STATE(8018)] = 268291,
  [SMALL_STATE(8019)] = 268305,
  [SMALL_STATE(8020)] = 268319,
  [SMALL_STATE(8021)] = 268333,
  [SMALL_STATE(8022)] = 268347,
  [SMALL_STATE(8023)] = 268361,
  [SMALL_STATE(8024)] = 268375,
  [SMALL_STATE(8025)] = 268389,
  [SMALL_STATE(8026)] = 268403,
  [SMALL_STATE(8027)] = 268417,
  [SMALL_STATE(8028)] = 268431,
  [SMALL_STATE(8029)] = 268445,
  [SMALL_STATE(8030)] = 268459,
  [SMALL_STATE(8031)] = 268473,
  [SMALL_STATE(8032)] = 268487,
  [SMALL_STATE(8033)] = 268501,
  [SMALL_STATE(8034)] = 268515,
  [SMALL_STATE(8035)] = 268529,
  [SMALL_STATE(8036)] = 268543,
  [SMALL_STATE(8037)] = 268557,
  [SMALL_STATE(8038)] = 268571,
  [SMALL_STATE(8039)] = 268585,
  [SMALL_STATE(8040)] = 268599,
  [SMALL_STATE(8041)] = 268613,
  [SMALL_STATE(8042)] = 268627,
  [SMALL_STATE(8043)] = 268641,
  [SMALL_STATE(8044)] = 268655,
  [SMALL_STATE(8045)] = 268669,
  [SMALL_STATE(8046)] = 268683,
  [SMALL_STATE(8047)] = 268697,
  [SMALL_STATE(8048)] = 268711,
  [SMALL_STATE(8049)] = 268725,
  [SMALL_STATE(8050)] = 268739,
  [SMALL_STATE(8051)] = 268753,
  [SMALL_STATE(8052)] = 268767,
  [SMALL_STATE(8053)] = 268781,
  [SMALL_STATE(8054)] = 268795,
  [SMALL_STATE(8055)] = 268809,
  [SMALL_STATE(8056)] = 268823,
  [SMALL_STATE(8057)] = 268837,
  [SMALL_STATE(8058)] = 268851,
  [SMALL_STATE(8059)] = 268865,
  [SMALL_STATE(8060)] = 268879,
  [SMALL_STATE(8061)] = 268893,
  [SMALL_STATE(8062)] = 268907,
  [SMALL_STATE(8063)] = 268921,
  [SMALL_STATE(8064)] = 268935,
  [SMALL_STATE(8065)] = 268949,
  [SMALL_STATE(8066)] = 268963,
  [SMALL_STATE(8067)] = 268977,
  [SMALL_STATE(8068)] = 268991,
  [SMALL_STATE(8069)] = 269005,
  [SMALL_STATE(8070)] = 269019,
  [SMALL_STATE(8071)] = 269033,
  [SMALL_STATE(8072)] = 269047,
  [SMALL_STATE(8073)] = 269061,
  [SMALL_STATE(8074)] = 269075,
  [SMALL_STATE(8075)] = 269089,
  [SMALL_STATE(8076)] = 269103,
  [SMALL_STATE(8077)] = 269117,
  [SMALL_STATE(8078)] = 269131,
  [SMALL_STATE(8079)] = 269145,
  [SMALL_STATE(8080)] = 269159,
  [SMALL_STATE(8081)] = 269173,
  [SMALL_STATE(8082)] = 269187,
  [SMALL_STATE(8083)] = 269201,
  [SMALL_STATE(8084)] = 269215,
  [SMALL_STATE(8085)] = 269229,
  [SMALL_STATE(8086)] = 269243,
  [SMALL_STATE(8087)] = 269257,
  [SMALL_STATE(8088)] = 269271,
  [SMALL_STATE(8089)] = 269285,
  [SMALL_STATE(8090)] = 269299,
  [SMALL_STATE(8091)] = 269313,
  [SMALL_STATE(8092)] = 269327,
  [SMALL_STATE(8093)] = 269341,
  [SMALL_STATE(8094)] = 269355,
  [SMALL_STATE(8095)] = 269369,
  [SMALL_STATE(8096)] = 269383,
  [SMALL_STATE(8097)] = 269397,
  [SMALL_STATE(8098)] = 269411,
  [SMALL_STATE(8099)] = 269425,
  [SMALL_STATE(8100)] = 269439,
  [SMALL_STATE(8101)] = 269453,
  [SMALL_STATE(8102)] = 269467,
  [SMALL_STATE(8103)] = 269481,
  [SMALL_STATE(8104)] = 269495,
  [SMALL_STATE(8105)] = 269509,
  [SMALL_STATE(8106)] = 269523,
  [SMALL_STATE(8107)] = 269537,
  [SMALL_STATE(8108)] = 269551,
  [SMALL_STATE(8109)] = 269565,
  [SMALL_STATE(8110)] = 269579,
  [SMALL_STATE(8111)] = 269593,
  [SMALL_STATE(8112)] = 269607,
  [SMALL_STATE(8113)] = 269621,
  [SMALL_STATE(8114)] = 269635,
  [SMALL_STATE(8115)] = 269649,
  [SMALL_STATE(8116)] = 269663,
  [SMALL_STATE(8117)] = 269677,
  [SMALL_STATE(8118)] = 269691,
  [SMALL_STATE(8119)] = 269705,
  [SMALL_STATE(8120)] = 269719,
  [SMALL_STATE(8121)] = 269733,
  [SMALL_STATE(8122)] = 269747,
  [SMALL_STATE(8123)] = 269761,
  [SMALL_STATE(8124)] = 269775,
  [SMALL_STATE(8125)] = 269789,
  [SMALL_STATE(8126)] = 269803,
  [SMALL_STATE(8127)] = 269817,
  [SMALL_STATE(8128)] = 269831,
  [SMALL_STATE(8129)] = 269845,
  [SMALL_STATE(8130)] = 269859,
  [SMALL_STATE(8131)] = 269873,
  [SMALL_STATE(8132)] = 269887,
  [SMALL_STATE(8133)] = 269901,
  [SMALL_STATE(8134)] = 269915,
  [SMALL_STATE(8135)] = 269929,
  [SMALL_STATE(8136)] = 269943,
  [SMALL_STATE(8137)] = 269957,
  [SMALL_STATE(8138)] = 269971,
  [SMALL_STATE(8139)] = 269985,
  [SMALL_STATE(8140)] = 269999,
  [SMALL_STATE(8141)] = 270013,
  [SMALL_STATE(8142)] = 270027,
  [SMALL_STATE(8143)] = 270041,
  [SMALL_STATE(8144)] = 270055,
  [SMALL_STATE(8145)] = 270069,
  [SMALL_STATE(8146)] = 270083,
  [SMALL_STATE(8147)] = 270097,
  [SMALL_STATE(8148)] = 270111,
  [SMALL_STATE(8149)] = 270125,
  [SMALL_STATE(8150)] = 270139,
  [SMALL_STATE(8151)] = 270153,
  [SMALL_STATE(8152)] = 270167,
  [SMALL_STATE(8153)] = 270181,
  [SMALL_STATE(8154)] = 270195,
  [SMALL_STATE(8155)] = 270209,
  [SMALL_STATE(8156)] = 270223,
  [SMALL_STATE(8157)] = 270237,
  [SMALL_STATE(8158)] = 270251,
  [SMALL_STATE(8159)] = 270265,
  [SMALL_STATE(8160)] = 270279,
  [SMALL_STATE(8161)] = 270293,
  [SMALL_STATE(8162)] = 270307,
  [SMALL_STATE(8163)] = 270321,
  [SMALL_STATE(8164)] = 270335,
  [SMALL_STATE(8165)] = 270349,
  [SMALL_STATE(8166)] = 270363,
  [SMALL_STATE(8167)] = 270377,
  [SMALL_STATE(8168)] = 270391,
  [SMALL_STATE(8169)] = 270405,
  [SMALL_STATE(8170)] = 270419,
  [SMALL_STATE(8171)] = 270433,
  [SMALL_STATE(8172)] = 270447,
  [SMALL_STATE(8173)] = 270461,
  [SMALL_STATE(8174)] = 270475,
  [SMALL_STATE(8175)] = 270489,
  [SMALL_STATE(8176)] = 270503,
  [SMALL_STATE(8177)] = 270517,
  [SMALL_STATE(8178)] = 270531,
  [SMALL_STATE(8179)] = 270545,
  [SMALL_STATE(8180)] = 270559,
  [SMALL_STATE(8181)] = 270573,
  [SMALL_STATE(8182)] = 270587,
  [SMALL_STATE(8183)] = 270601,
  [SMALL_STATE(8184)] = 270615,
  [SMALL_STATE(8185)] = 270629,
  [SMALL_STATE(8186)] = 270643,
  [SMALL_STATE(8187)] = 270657,
  [SMALL_STATE(8188)] = 270671,
  [SMALL_STATE(8189)] = 270685,
  [SMALL_STATE(8190)] = 270699,
  [SMALL_STATE(8191)] = 270713,
  [SMALL_STATE(8192)] = 270727,
  [SMALL_STATE(8193)] = 270741,
  [SMALL_STATE(8194)] = 270755,
  [SMALL_STATE(8195)] = 270769,
  [SMALL_STATE(8196)] = 270783,
  [SMALL_STATE(8197)] = 270797,
  [SMALL_STATE(8198)] = 270811,
  [SMALL_STATE(8199)] = 270825,
  [SMALL_STATE(8200)] = 270839,
  [SMALL_STATE(8201)] = 270853,
  [SMALL_STATE(8202)] = 270867,
  [SMALL_STATE(8203)] = 270881,
  [SMALL_STATE(8204)] = 270895,
  [SMALL_STATE(8205)] = 270909,
  [SMALL_STATE(8206)] = 270923,
  [SMALL_STATE(8207)] = 270937,
  [SMALL_STATE(8208)] = 270951,
  [SMALL_STATE(8209)] = 270965,
  [SMALL_STATE(8210)] = 270979,
  [SMALL_STATE(8211)] = 270993,
  [SMALL_STATE(8212)] = 271007,
  [SMALL_STATE(8213)] = 271021,
  [SMALL_STATE(8214)] = 271035,
  [SMALL_STATE(8215)] = 271049,
  [SMALL_STATE(8216)] = 271063,
  [SMALL_STATE(8217)] = 271077,
  [SMALL_STATE(8218)] = 271091,
  [SMALL_STATE(8219)] = 271105,
  [SMALL_STATE(8220)] = 271119,
  [SMALL_STATE(8221)] = 271133,
  [SMALL_STATE(8222)] = 271147,
  [SMALL_STATE(8223)] = 271161,
  [SMALL_STATE(8224)] = 271175,
  [SMALL_STATE(8225)] = 271189,
  [SMALL_STATE(8226)] = 271203,
  [SMALL_STATE(8227)] = 271217,
  [SMALL_STATE(8228)] = 271231,
  [SMALL_STATE(8229)] = 271245,
  [SMALL_STATE(8230)] = 271259,
  [SMALL_STATE(8231)] = 271273,
  [SMALL_STATE(8232)] = 271287,
  [SMALL_STATE(8233)] = 271301,
  [SMALL_STATE(8234)] = 271315,
  [SMALL_STATE(8235)] = 271329,
  [SMALL_STATE(8236)] = 271343,
  [SMALL_STATE(8237)] = 271357,
  [SMALL_STATE(8238)] = 271371,
  [SMALL_STATE(8239)] = 271385,
  [SMALL_STATE(8240)] = 271399,
  [SMALL_STATE(8241)] = 271413,
  [SMALL_STATE(8242)] = 271427,
  [SMALL_STATE(8243)] = 271441,
  [SMALL_STATE(8244)] = 271455,
  [SMALL_STATE(8245)] = 271469,
  [SMALL_STATE(8246)] = 271483,
  [SMALL_STATE(8247)] = 271497,
  [SMALL_STATE(8248)] = 271511,
  [SMALL_STATE(8249)] = 271525,
  [SMALL_STATE(8250)] = 271539,
  [SMALL_STATE(8251)] = 271553,
  [SMALL_STATE(8252)] = 271567,
  [SMALL_STATE(8253)] = 271581,
  [SMALL_STATE(8254)] = 271595,
  [SMALL_STATE(8255)] = 271609,
  [SMALL_STATE(8256)] = 271623,
  [SMALL_STATE(8257)] = 271637,
  [SMALL_STATE(8258)] = 271651,
  [SMALL_STATE(8259)] = 271665,
  [SMALL_STATE(8260)] = 271679,
  [SMALL_STATE(8261)] = 271693,
  [SMALL_STATE(8262)] = 271707,
  [SMALL_STATE(8263)] = 271721,
  [SMALL_STATE(8264)] = 271735,
  [SMALL_STATE(8265)] = 271749,
  [SMALL_STATE(8266)] = 271763,
  [SMALL_STATE(8267)] = 271777,
  [SMALL_STATE(8268)] = 271791,
  [SMALL_STATE(8269)] = 271805,
  [SMALL_STATE(8270)] = 271819,
  [SMALL_STATE(8271)] = 271833,
  [SMALL_STATE(8272)] = 271847,
  [SMALL_STATE(8273)] = 271861,
  [SMALL_STATE(8274)] = 271875,
  [SMALL_STATE(8275)] = 271889,
  [SMALL_STATE(8276)] = 271903,
  [SMALL_STATE(8277)] = 271917,
  [SMALL_STATE(8278)] = 271931,
  [SMALL_STATE(8279)] = 271945,
  [SMALL_STATE(8280)] = 271959,
  [SMALL_STATE(8281)] = 271973,
  [SMALL_STATE(8282)] = 271987,
  [SMALL_STATE(8283)] = 272001,
  [SMALL_STATE(8284)] = 272015,
  [SMALL_STATE(8285)] = 272029,
  [SMALL_STATE(8286)] = 272043,
  [SMALL_STATE(8287)] = 272057,
  [SMALL_STATE(8288)] = 272071,
  [SMALL_STATE(8289)] = 272085,
  [SMALL_STATE(8290)] = 272099,
  [SMALL_STATE(8291)] = 272113,
  [SMALL_STATE(8292)] = 272127,
  [SMALL_STATE(8293)] = 272141,
  [SMALL_STATE(8294)] = 272155,
  [SMALL_STATE(8295)] = 272169,
  [SMALL_STATE(8296)] = 272183,
  [SMALL_STATE(8297)] = 272197,
  [SMALL_STATE(8298)] = 272211,
  [SMALL_STATE(8299)] = 272225,
  [SMALL_STATE(8300)] = 272239,
  [SMALL_STATE(8301)] = 272253,
  [SMALL_STATE(8302)] = 272267,
  [SMALL_STATE(8303)] = 272281,
  [SMALL_STATE(8304)] = 272295,
  [SMALL_STATE(8305)] = 272309,
  [SMALL_STATE(8306)] = 272323,
  [SMALL_STATE(8307)] = 272337,
  [SMALL_STATE(8308)] = 272351,
  [SMALL_STATE(8309)] = 272365,
  [SMALL_STATE(8310)] = 272379,
  [SMALL_STATE(8311)] = 272393,
  [SMALL_STATE(8312)] = 272407,
  [SMALL_STATE(8313)] = 272421,
  [SMALL_STATE(8314)] = 272435,
  [SMALL_STATE(8315)] = 272449,
  [SMALL_STATE(8316)] = 272463,
  [SMALL_STATE(8317)] = 272477,
  [SMALL_STATE(8318)] = 272491,
  [SMALL_STATE(8319)] = 272505,
  [SMALL_STATE(8320)] = 272519,
  [SMALL_STATE(8321)] = 272533,
  [SMALL_STATE(8322)] = 272547,
  [SMALL_STATE(8323)] = 272561,
  [SMALL_STATE(8324)] = 272575,
  [SMALL_STATE(8325)] = 272589,
  [SMALL_STATE(8326)] = 272603,
  [SMALL_STATE(8327)] = 272617,
  [SMALL_STATE(8328)] = 272631,
  [SMALL_STATE(8329)] = 272645,
  [SMALL_STATE(8330)] = 272659,
  [SMALL_STATE(8331)] = 272673,
  [SMALL_STATE(8332)] = 272687,
  [SMALL_STATE(8333)] = 272701,
  [SMALL_STATE(8334)] = 272715,
  [SMALL_STATE(8335)] = 272729,
  [SMALL_STATE(8336)] = 272743,
  [SMALL_STATE(8337)] = 272757,
  [SMALL_STATE(8338)] = 272771,
  [SMALL_STATE(8339)] = 272785,
  [SMALL_STATE(8340)] = 272799,
  [SMALL_STATE(8341)] = 272813,
  [SMALL_STATE(8342)] = 272827,
  [SMALL_STATE(8343)] = 272841,
  [SMALL_STATE(8344)] = 272855,
  [SMALL_STATE(8345)] = 272869,
  [SMALL_STATE(8346)] = 272883,
  [SMALL_STATE(8347)] = 272897,
  [SMALL_STATE(8348)] = 272911,
  [SMALL_STATE(8349)] = 272925,
  [SMALL_STATE(8350)] = 272939,
  [SMALL_STATE(8351)] = 272953,
  [SMALL_STATE(8352)] = 272967,
  [SMALL_STATE(8353)] = 272981,
  [SMALL_STATE(8354)] = 272995,
  [SMALL_STATE(8355)] = 272999,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6486),
  [5] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [7] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compilation_unit, 0, 0, 0),
  [9] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1417),
  [11] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [13] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2076),
  [15] = {.entry = {.count = 1, .reusable = false}}, SHIFT(896),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6790),
  [19] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [21] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5579),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4800),
  [25] = {.entry = {.count = 1, .reusable = false}}, SHIFT(266),
  [27] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1470),
  [29] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1788),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5210),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5434),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4552),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4665),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5487),
  [41] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5613),
  [43] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4214),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2068),
  [47] = {.entry = {.count = 1, .reusable = true}}, SHIFT(419),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(202),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(167),
  [53] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1590),
  [55] = {.entry = {.count = 1, .reusable = false}}, SHIFT(162),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1505),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(155),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1840),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1849),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5575),
  [67] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5845),
  [69] = {.entry = {.count = 1, .reusable = false}}, SHIFT(125),
  [71] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1471),
  [73] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6390),
  [75] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6439),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6630),
  [79] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6636),
  [81] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6327),
  [83] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1473),
  [85] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3235),
  [87] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5256),
  [89] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4883),
  [91] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1894),
  [93] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3235),
  [95] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1812),
  [97] = {.entry = {.count = 1, .reusable = true}}, SHIFT(366),
  [99] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1436),
  [101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6878),
  [103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7074),
  [105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1138),
  [107] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_anonymous, 2, 0, 0),
  [109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(892),
  [111] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression_anonymous, 2, 0, 0),
  [113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(385),
  [117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1169),
  [119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1793),
  [121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4239),
  [123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2069),
  [125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(181),
  [129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(209),
  [131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1503),
  [133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(210),
  [135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1859),
  [137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1851),
  [139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5654),
  [141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5794),
  [143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(130),
  [145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1189),
  [147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6590),
  [149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6721),
  [151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6347),
  [153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1139),
  [155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5301),
  [157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4891),
  [159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1906),
  [161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1811),
  [163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(605),
  [165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1148),
  [167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6801),
  [169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(893),
  [171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2077),
  [173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(182),
  [175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(211),
  [177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1497),
  [179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(212),
  [181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(610),
  [183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1225),
  [185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(902),
  [187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(353),
  [191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1265),
  [193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1790),
  [195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4235),
  [197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2058),
  [199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(354),
  [201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(175),
  [203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(197),
  [205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1500),
  [207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(198),
  [209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5679),
  [211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5877),
  [213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(124),
  [215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1295),
  [217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6570),
  [219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6707),
  [221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1266),
  [223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5337),
  [225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4899),
  [227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1897),
  [229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(576),
  [231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1282),
  [233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7118),
  [235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(888),
  [237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2059),
  [239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(213),
  [241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(207),
  [243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1494),
  [245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(208),
  [247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(600),
  [249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1383),
  [251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(901),
  [253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(329),
  [257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1332),
  [259] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1784),
  [261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4233),
  [263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2067),
  [265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(172),
  [269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(195),
  [271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1504),
  [273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(196),
  [275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5692),
  [277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5982),
  [279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(121),
  [281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1364),
  [283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6533),
  [285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6679),
  [287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1334),
  [289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5277),
  [291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4873),
  [293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1895),
  [295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1363),
  [299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6764),
  [301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(899),
  [303] = {.entry = {.count = 1, .reusable = false}}, SHIFT(900),
  [305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1726),
  [307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1587),
  [309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3642),
  [311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5490),
  [313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1821),
  [315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6878),
  [317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4217),
  [319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1419),
  [321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7770),
  [323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2352),
  [325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6180),
  [327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4159),
  [329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4459),
  [331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4242),
  [333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1000),
  [335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6647),
  [337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5153),
  [339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6218),
  [341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3469),
  [343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2841),
  [345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2629),
  [347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8354),
  [349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4190),
  [351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3683),
  [353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2368),
  [355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2934),
  [357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2414),
  [359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1276),
  [361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3578),
  [363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2906),
  [365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5207),
  [367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2004),
  [369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1118),
  [371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3264),
  [373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1166),
  [375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2530),
  [377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3858),
  [379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2991),
  [381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1352),
  [383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3690),
  [385] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure, 4, 0, 0),
  [387] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure, 4, 0, 0),
  [389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5224),
  [391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure, 2, 0, 0),
  [393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure, 2, 0, 0),
  [395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure, 1, 0, 0),
  [397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure, 1, 0, 0),
  [399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure, 3, 0, 0),
  [401] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure, 3, 0, 0),
  [403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1026),
  [405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(890),
  [407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(369),
  [411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1101),
  [413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1791),
  [415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4270),
  [417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2071),
  [419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(179),
  [423] = {.entry = {.count = 1, .reusable = false}}, SHIFT(173),
  [425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1596),
  [427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(205),
  [429] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1502),
  [431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(206),
  [433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1836),
  [435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1855),
  [437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5759),
  [439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5942),
  [441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(127),
  [443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1045),
  [445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6584),
  [447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6717),
  [449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1103),
  [451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5299),
  [453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4908),
  [455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1879),
  [457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(595),
  [459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1104),
  [461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6779),
  [463] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compilation_unit, 1, 0, 0),
  [465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(891),
  [467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2070),
  [469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(178),
  [471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(203),
  [473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1501),
  [475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(204),
  [477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6486),
  [479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6584),
  [481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6717),
  [483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(590),
  [485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1017),
  [487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(889),
  [489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(62),
  [491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(543),
  [493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(985),
  [495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1775),
  [497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4232),
  [499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2073),
  [501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(544),
  [503] = {.entry = {.count = 1, .reusable = false}}, SHIFT(174),
  [505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(164),
  [507] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1637),
  [509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(193),
  [511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1495),
  [513] = {.entry = {.count = 1, .reusable = false}}, SHIFT(194),
  [515] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1858),
  [517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1828),
  [519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5749),
  [521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5807),
  [523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(136),
  [525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(946),
  [527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6456),
  [529] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6652),
  [531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(986),
  [533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5335),
  [535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4894),
  [537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1886),
  [539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(987),
  [543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7135),
  [545] = {.entry = {.count = 1, .reusable = false}}, SHIFT(897),
  [547] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2057),
  [549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(163),
  [551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(185),
  [553] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1493),
  [555] = {.entry = {.count = 1, .reusable = false}}, SHIFT(186),
  [557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6456),
  [559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6652),
  [561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(545),
  [563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1160),
  [565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8287),
  [567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4606),
  [569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8026),
  [571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7678),
  [573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3220),
  [575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1874),
  [577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7699),
  [579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7388),
  [581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7388),
  [583] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7484),
  [585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8353),
  [587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8068),
  [589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8068),
  [591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8222),
  [593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1799),
  [595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(240),
  [597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1384),
  [599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8277),
  [601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4562),
  [603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7462),
  [605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1866),
  [607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7473),
  [609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1801),
  [611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1431),
  [615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7918),
  [617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4578),
  [619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7615),
  [621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1861),
  [623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7872),
  [625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1778),
  [627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(991),
  [631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8266),
  [633] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4605),
  [635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8131),
  [637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1868),
  [639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8135),
  [641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1785),
  [643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2393),
  [647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8280),
  [649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4639),
  [651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7861),
  [653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1871),
  [655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7863),
  [657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1805),
  [659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2280),
  [663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8282),
  [665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4575),
  [667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8320),
  [669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1870),
  [671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8322),
  [673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1776),
  [675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1236),
  [679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8283),
  [681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4584),
  [683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8050),
  [685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1862),
  [687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8061),
  [689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1779),
  [691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1986),
  [695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8284),
  [697] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4591),
  [699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7430),
  [701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1876),
  [703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7503),
  [705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1783),
  [707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4887),
  [711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8271),
  [713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7440),
  [715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1865),
  [717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7464),
  [719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1781),
  [721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1095),
  [725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8285),
  [727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4596),
  [729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8015),
  [731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1860),
  [733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8112),
  [735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1789),
  [737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(237),
  [739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3321),
  [741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8286),
  [743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4604),
  [745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7288),
  [747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1863),
  [749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7306),
  [751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1797),
  [753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2510),
  [757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8288),
  [759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4615),
  [761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7202),
  [763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1869),
  [765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7901),
  [767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1802),
  [769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2939),
  [773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8289),
  [775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4619),
  [777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7249),
  [779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1872),
  [781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8102),
  [783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1803),
  [785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2618),
  [789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8275),
  [791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4748),
  [793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7658),
  [795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1867),
  [797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7341),
  [799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1796),
  [801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1932),
  [805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(894),
  [807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(70),
  [809] = {.entry = {.count = 1, .reusable = false}}, SHIFT(360),
  [811] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1965),
  [813] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1806),
  [815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4209),
  [817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2062),
  [819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(184),
  [823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(171),
  [825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1581),
  [827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(201),
  [829] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1499),
  [831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(144),
  [833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1838),
  [835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1839),
  [837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5673),
  [839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5914),
  [841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(126),
  [843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1977),
  [845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6582),
  [847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6715),
  [849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1966),
  [851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5298),
  [853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4875),
  [855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1893),
  [857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(586),
  [859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1967),
  [861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7122),
  [863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(887),
  [865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2066),
  [867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(177),
  [869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(895),
  [871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2061),
  [873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(183),
  [875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(199),
  [877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1498),
  [879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(200),
  [881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(581),
  [883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(903),
  [885] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2063),
  [887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(176),
  [889] = {.entry = {.count = 1, .reusable = false}}, SHIFT(904),
  [891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2064),
  [893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(170),
  [895] = {.entry = {.count = 1, .reusable = false}}, SHIFT(191),
  [897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1492),
  [899] = {.entry = {.count = 1, .reusable = false}}, SHIFT(192),
  [901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(561),
  [903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(898),
  [905] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2075),
  [907] = {.entry = {.count = 1, .reusable = false}}, SHIFT(190),
  [909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6347),
  [911] = {.entry = {.count = 1, .reusable = false}}, SHIFT(998),
  [913] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6831),
  [915] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0),
  [917] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0),
  [919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5333),
  [921] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7676),
  [923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1902),
  [925] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(998),
  [928] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45),
  [930] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6831),
  [933] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(62),
  [936] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45),
  [938] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(543),
  [941] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(985),
  [944] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1775),
  [947] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4232),
  [950] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(544),
  [953] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5749),
  [956] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5807),
  [959] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(136),
  [962] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(946),
  [965] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6456),
  [968] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6652),
  [971] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(986),
  [974] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5335),
  [977] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4894),
  [980] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1886),
  [983] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(987),
  [986] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(7135),
  [989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1114),
  [991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7077),
  [993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5293),
  [995] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7338),
  [997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1883),
  [999] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_application_expression, 2, 0, 21),
  [1001] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_application_expression, 2, 0, 21),
  [1003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1159),
  [1005] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6824),
  [1007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5296),
  [1009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7906),
  [1011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1909),
  [1013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1284),
  [1015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7156),
  [1017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5334),
  [1019] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7506),
  [1021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1910),
  [1023] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1114),
  [1026] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(7077),
  [1029] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(74),
  [1032] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(369),
  [1035] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1101),
  [1038] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1791),
  [1041] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4270),
  [1044] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(370),
  [1047] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5759),
  [1050] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5942),
  [1053] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(127),
  [1056] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1045),
  [1059] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6584),
  [1062] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6717),
  [1065] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1103),
  [1068] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5299),
  [1071] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4908),
  [1074] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1879),
  [1077] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1104),
  [1080] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6779),
  [1083] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1367),
  [1085] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6966),
  [1087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5259),
  [1089] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7909),
  [1091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1917),
  [1093] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1159),
  [1096] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6824),
  [1099] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(78),
  [1102] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(385),
  [1105] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1169),
  [1108] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1793),
  [1111] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4239),
  [1114] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(386),
  [1117] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5654),
  [1120] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5794),
  [1123] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(130),
  [1126] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1189),
  [1129] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6590),
  [1132] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6721),
  [1135] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1139),
  [1138] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5301),
  [1141] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4891),
  [1144] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1906),
  [1147] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1148),
  [1150] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6801),
  [1153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1446),
  [1155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6795),
  [1157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5317),
  [1159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7343),
  [1161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1880),
  [1163] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1284),
  [1166] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(7156),
  [1169] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(69),
  [1172] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(353),
  [1175] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1265),
  [1178] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1790),
  [1181] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4235),
  [1184] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(354),
  [1187] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5679),
  [1190] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5877),
  [1193] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(124),
  [1196] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1295),
  [1199] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6570),
  [1202] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6707),
  [1205] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1266),
  [1208] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5337),
  [1211] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4899),
  [1214] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1897),
  [1217] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1282),
  [1220] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(7118),
  [1223] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1367),
  [1226] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6966),
  [1229] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(60),
  [1232] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(329),
  [1235] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1332),
  [1238] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1784),
  [1241] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4233),
  [1244] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(330),
  [1247] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5692),
  [1250] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5982),
  [1253] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(121),
  [1256] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1364),
  [1259] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6533),
  [1262] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6679),
  [1265] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1334),
  [1268] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5277),
  [1271] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4873),
  [1274] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1895),
  [1277] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1363),
  [1280] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6764),
  [1283] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1446),
  [1286] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6795),
  [1289] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(61),
  [1292] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(266),
  [1295] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1470),
  [1298] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1788),
  [1301] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4214),
  [1304] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(419),
  [1307] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5575),
  [1310] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5845),
  [1313] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(125),
  [1316] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1471),
  [1319] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6439),
  [1322] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6630),
  [1325] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1473),
  [1328] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5256),
  [1331] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4883),
  [1334] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1894),
  [1337] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1436),
  [1340] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(7074),
  [1343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1384),
  [1345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2393),
  [1347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2280),
  [1349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1236),
  [1351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1431),
  [1353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1986),
  [1355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1095),
  [1357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4284),
  [1359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6848),
  [1361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [1363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2787),
  [1365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7087),
  [1367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8274),
  [1369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1613),
  [1371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4812),
  [1373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5496),
  [1375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1623),
  [1377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6025),
  [1379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6234),
  [1381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7319),
  [1383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1617),
  [1385] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1580),
  [1387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6690),
  [1389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2145),
  [1391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6494),
  [1393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6660),
  [1395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4814),
  [1397] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7129),
  [1399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5262),
  [1401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4880),
  [1403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7684),
  [1405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2357),
  [1407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6958),
  [1409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3321),
  [1411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1160),
  [1413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4887),
  [1415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4163),
  [1417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6886),
  [1419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [1421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5690),
  [1423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8278),
  [1425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1608),
  [1427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5444),
  [1429] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1591),
  [1431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6082),
  [1433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7235),
  [1435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1609),
  [1437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1611),
  [1439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2138),
  [1441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7244),
  [1443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2510),
  [1445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2939),
  [1447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2196),
  [1449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8215),
  [1451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6215),
  [1453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7361),
  [1455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7379),
  [1457] = {.entry = {.count = 1, .reusable = false}}, SHIFT(991),
  [1459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2618),
  [1461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3745),
  [1463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8279),
  [1465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6145),
  [1467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7559),
  [1469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7564),
  [1471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3054),
  [1473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8281),
  [1475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6172),
  [1477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8033),
  [1479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8042),
  [1481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7895),
  [1483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6177),
  [1485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2647),
  [1487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6439),
  [1489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6630),
  [1491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4068),
  [1493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6590),
  [1495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6721),
  [1497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6533),
  [1499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6679),
  [1501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3590),
  [1503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2857),
  [1505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6570),
  [1507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6707),
  [1509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1099),
  [1511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3954),
  [1513] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4863),
  [1515] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3568),
  [1517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2641),
  [1519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2503),
  [1521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2623),
  [1523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2769),
  [1525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1391),
  [1527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1408),
  [1529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2370),
  [1531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2195),
  [1533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2253),
  [1535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2271),
  [1537] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1440),
  [1539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1308),
  [1541] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1247),
  [1543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1970),
  [1545] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2014),
  [1547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1090),
  [1549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3311),
  [1551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3339),
  [1553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1199),
  [1555] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1163),
  [1557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2518),
  [1559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2512),
  [1561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3027),
  [1563] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2947),
  [1565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1433),
  [1567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8079),
  [1569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(996),
  [1571] = {.entry = {.count = 1, .reusable = false}}, SHIFT(999),
  [1573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2707),
  [1575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1443),
  [1577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2198),
  [1579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2201),
  [1581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2554),
  [1583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2556),
  [1585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1379),
  [1587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1381),
  [1589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2277),
  [1591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2296),
  [1593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1300),
  [1595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1302),
  [1597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1949),
  [1599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1950),
  [1601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1047),
  [1603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1048),
  [1605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3364),
  [1607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3274),
  [1609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1191),
  [1611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1196),
  [1613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2452),
  [1615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2453),
  [1617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3055),
  [1619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3061),
  [1621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1472),
  [1623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(970),
  [1625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(977),
  [1627] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_content, 2, 0, 0),
  [1629] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_content, 3, 0, 0),
  [1631] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6796),
  [1633] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7024),
  [1635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7127),
  [1637] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6877),
  [1639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7099),
  [1641] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_binding_pattern, 1, 0, 0),
  [1643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2313),
  [1645] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_binding_pattern, 1, 0, 0),
  [1647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6886),
  [1649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(135),
  [1651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1638),
  [1653] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2173),
  [1655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5508),
  [1657] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1631),
  [1659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6081),
  [1661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1594),
  [1663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1632),
  [1665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6402),
  [1667] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2183),
  [1669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6331),
  [1671] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6599),
  [1673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2143),
  [1675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7396),
  [1677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5304),
  [1679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4897),
  [1681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2060),
  [1683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6802),
  [1685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [1687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2020),
  [1689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7102),
  [1691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5313),
  [1693] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7718),
  [1695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1908),
  [1697] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(2020),
  [1700] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(7102),
  [1703] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(70),
  [1706] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(360),
  [1709] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1965),
  [1712] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1806),
  [1715] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4209),
  [1718] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(361),
  [1721] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5673),
  [1724] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5914),
  [1727] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(126),
  [1730] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1977),
  [1733] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6582),
  [1736] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6715),
  [1739] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1966),
  [1742] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5298),
  [1745] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4875),
  [1748] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1893),
  [1751] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1967),
  [1754] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(7122),
  [1757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3677),
  [1759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1815),
  [1761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [1763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5766),
  [1765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4784),
  [1767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1606),
  [1769] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3545),
  [1771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5438),
  [1773] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5235),
  [1775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5409),
  [1777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4504),
  [1779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5410),
  [1781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6262),
  [1783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1607),
  [1785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1599),
  [1787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6565),
  [1789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6330),
  [1791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6547),
  [1793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6688),
  [1795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6332),
  [1797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6601),
  [1799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6406),
  [1801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3558),
  [1803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7606),
  [1805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5284),
  [1807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4892),
  [1809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3358),
  [1811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6890),
  [1813] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1819),
  [1815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1822),
  [1817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2855),
  [1819] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 1, 0, 7),
  [1821] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 1, 0, 7),
  [1823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [1825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6537),
  [1827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [1829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1673),
  [1831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1690),
  [1833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3162),
  [1835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1602),
  [1837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2783),
  [1839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5515),
  [1841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3375),
  [1843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6052),
  [1845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6118),
  [1847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1740),
  [1849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1603),
  [1851] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6441),
  [1853] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6550),
  [1855] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6520),
  [1857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6671),
  [1859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2784),
  [1861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7357),
  [1863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5268),
  [1865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4881),
  [1867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1820),
  [1869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1807),
  [1871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [1873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(335),
  [1875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1810),
  [1877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [1879] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_expression, 1, 0, 0),
  [1881] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_expression, 1, 0, 0),
  [1883] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_application, 2, 0, 77),
  [1885] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_application, 2, 0, 77),
  [1887] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_path, 3, 0, 61),
  [1889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_path, 3, 0, 61),
  [1891] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_path, 1, 0, 12),
  [1893] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_path, 1, 0, 12),
  [1895] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_get_expression, 3, 0, 42),
  [1897] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_get_expression, 3, 0, 42),
  [1899] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_get_expression, 5, 0, 123),
  [1901] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_get_expression, 5, 0, 123),
  [1903] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_get_expression, 5, 0, 124),
  [1905] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_get_expression, 5, 0, 124),
  [1907] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bigarray_get_expression, 5, 0, 123),
  [1909] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bigarray_get_expression, 5, 0, 123),
  [1911] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_get_expression, 6, 0, 166),
  [1913] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_get_expression, 6, 0, 166),
  [1915] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_get_expression, 6, 0, 167),
  [1917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_get_expression, 6, 0, 167),
  [1919] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bigarray_get_expression, 6, 0, 166),
  [1921] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bigarray_get_expression, 6, 0, 166),
  [1923] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 5, 0, 108),
  [1925] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 5, 0, 108),
  [1927] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_prefix_expression, 2, 0, 19),
  [1929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_prefix_expression, 2, 0, 19),
  [1931] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 1, 0, 20),
  [1933] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 1, 0, 20),
  [1935] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 3, 0, 18),
  [1937] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 3, 0, 18),
  [1939] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_expression, 3, 0, 0),
  [1941] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_expression, 3, 0, 0),
  [1943] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 3, 0, 0),
  [1945] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 3, 0, 0),
  [1947] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_expression, 1, 0, 0),
  [1949] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_expression, 1, 0, 0),
  [1951] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 3, 0, 0),
  [1953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 3, 0, 0),
  [1955] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 3, 0, 0),
  [1957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 3, 0, 0),
  [1959] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_new_expression, 3, 0, 0),
  [1961] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_new_expression, 3, 0, 0),
  [1963] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_copy_expression, 3, 0, 0),
  [1965] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_copy_expression, 3, 0, 0),
  [1967] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_invocation, 3, 0, 43),
  [1969] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_invocation, 3, 0, 43),
  [1971] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_expression, 3, 0, 44),
  [1973] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_expression, 3, 0, 44),
  [1975] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_local_open_expression, 3, 0, 40),
  [1977] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_local_open_expression, 3, 0, 40),
  [1979] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_path, 3, 0, 0),
  [1981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_path, 3, 0, 0),
  [1983] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_expression, 4, 0, 31),
  [1985] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_expression, 4, 0, 31),
  [1987] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_expression, 4, 0, 57),
  [1989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_expression, 4, 0, 57),
  [1991] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 4, 0, 0),
  [1993] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 4, 0, 0),
  [1995] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__delimited_expression, 1, 0, 0),
  [1997] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__delimited_expression, 1, 0, 0),
  [1999] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 4, 0, 0),
  [2001] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 4, 0, 0),
  [2003] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_path, 3, 0, 81),
  [2005] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_path, 3, 0, 81),
  [2007] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_copy_expression, 4, 0, 0),
  [2009] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_copy_expression, 4, 0, 0),
  [2011] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 4, 0, 40),
  [2013] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 4, 0, 40),
  [2015] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_expression, 5, 0, 89),
  [2017] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_expression, 5, 0, 89),
  [2019] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_expression, 5, 0, 67),
  [2021] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_expression, 5, 0, 67),
  [2023] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 5, 0, 94),
  [2025] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 5, 0, 94),
  [2027] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 5, 0, 0),
  [2029] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 5, 0, 0),
  [2031] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 9, 0, 0),
  [2033] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 9, 0, 0),
  [2035] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_copy_expression, 5, 0, 0),
  [2037] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_copy_expression, 5, 0, 0),
  [2039] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_argument, 4, 0, 55),
  [2041] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument, 4, 0, 55),
  [2043] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_expression, 2, 0, 0),
  [2045] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_expression, 2, 0, 0),
  [2047] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 2, 0, 0),
  [2049] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 2, 0, 0),
  [2051] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_expression, 6, 0, 134),
  [2053] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_expression, 6, 0, 134),
  [2055] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 6, 0, 94),
  [2057] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 6, 0, 94),
  [2059] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 6, 0, 108),
  [2061] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 6, 0, 108),
  [2063] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 6, 0, 152),
  [2065] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 6, 0, 152),
  [2067] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 6, 0, 163),
  [2069] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 6, 0, 163),
  [2071] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 2, 0, 0),
  [2073] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 2, 0, 0),
  [2075] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_path, 1, 0, 14),
  [2077] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_path, 1, 0, 14),
  [2079] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_new_expression, 2, 0, 0),
  [2081] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_new_expression, 2, 0, 0),
  [2083] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 7, 0, 94),
  [2085] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 7, 0, 94),
  [2087] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 7, 0, 152),
  [2089] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 7, 0, 152),
  [2091] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 7, 0, 163),
  [2093] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 7, 0, 163),
  [2095] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 7, 0, 190),
  [2097] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 7, 0, 190),
  [2099] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 8, 0, 190),
  [2101] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 8, 0, 190),
  [2103] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_boolean, 1, 0, 0),
  [2105] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean, 1, 0, 0),
  [2107] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_number, 1, 0, 0),
  [2109] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_number, 1, 0, 0),
  [2111] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_name, 1, 0, 4),
  [2113] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_name, 1, 0, 4),
  [2115] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_path, 1, 0, 3),
  [2117] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constant, 1, 0, 0),
  [2119] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constant, 1, 0, 0),
  [2121] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__value_name, 1, 0, 0),
  [2123] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__value_name, 1, 0, 0),
  [2125] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_path, 1, 0, 0),
  [2127] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_path, 1, 0, 0),
  [2129] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unit, 2, 0, 0),
  [2131] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unit, 2, 0, 0),
  [2133] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 2, 0, 0),
  [2135] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 2, 0, 0),
  [2137] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tag, 2, 0, 0),
  [2139] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag, 2, 0, 0),
  [2141] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_copy_expression, 2, 0, 0),
  [2143] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_copy_expression, 2, 0, 0),
  [2145] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_name, 3, 0, 23),
  [2147] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_name, 3, 0, 23),
  [2149] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 3, 0, 0),
  [2151] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 3, 0, 0),
  [2153] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_string, 3, 0, 0),
  [2155] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_string, 3, 0, 0),
  [2157] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__value_name, 1, 0, 1),
  [2159] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__value_name, 1, 0, 1),
  [2161] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unit, 3, 0, 0),
  [2163] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unit, 3, 0, 0),
  [2165] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extension, 3, 0, 0),
  [2167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extension, 3, 0, 0),
  [2169] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_character, 3, 0, 0),
  [2171] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_character, 3, 0, 0),
  [2173] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 3, 0, 0),
  [2175] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 3, 0, 0),
  [2177] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_path, 3, 0, 47),
  [2179] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_path, 3, 0, 0),
  [2181] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_path, 3, 0, 0),
  [2183] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extension, 4, 0, 0),
  [2185] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extension, 4, 0, 0),
  [2187] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_extension, 4, 0, 0),
  [2189] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_extension, 4, 0, 0),
  [2191] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_extension, 5, 0, 0),
  [2193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_extension, 5, 0, 0),
  [2195] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 6, 0, 0),
  [2197] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 6, 0, 0),
  [2199] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 7, 0, 0),
  [2201] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 7, 0, 0),
  [2203] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 8, 0, 0),
  [2205] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 8, 0, 0),
  [2207] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_path, 1, 0, 0),
  [2209] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_path, 1, 0, 0),
  [2211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3039),
  [2213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7006),
  [2215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [2217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(398),
  [2219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2967),
  [2221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1795),
  [2223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4283),
  [2225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [2227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5599),
  [2229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5848),
  [2231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(134),
  [2233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3044),
  [2235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6596),
  [2237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6724),
  [2239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2968),
  [2241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5303),
  [2243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4864),
  [2245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1901),
  [2247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2969),
  [2249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6884),
  [2251] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(3039),
  [2254] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(7006),
  [2257] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(82),
  [2260] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(398),
  [2263] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(2967),
  [2266] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1795),
  [2269] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4283),
  [2272] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(399),
  [2275] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5599),
  [2278] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5848),
  [2281] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(134),
  [2284] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(3044),
  [2287] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6596),
  [2290] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6724),
  [2293] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(2968),
  [2296] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(5303),
  [2299] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(4864),
  [2302] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(1901),
  [2305] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(2969),
  [2308] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 45), SHIFT_REPEAT(6884),
  [2311] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_argument, 2, 0, 8),
  [2313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1900),
  [2315] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument, 2, 0, 8),
  [2317] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__instance_variable_name, 1, 0, 2),
  [2319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [2321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1605),
  [2323] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2155),
  [2325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6547),
  [2327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6688),
  [2329] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2926),
  [2331] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 2, 0, 33),
  [2333] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 2, 0, 33),
  [2335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3995),
  [2337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6447),
  [2339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(139),
  [2341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3209),
  [2343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1618),
  [2345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3032),
  [2347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5425),
  [2349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6034),
  [2351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1619),
  [2353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6324),
  [2355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6340),
  [2357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6561),
  [2359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6698),
  [2361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3034),
  [2363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8060),
  [2365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5290),
  [2367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4876),
  [2369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2916),
  [2371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6964),
  [2373] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 147),
  [2375] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 147),
  [2377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4024),
  [2379] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 103),
  [2381] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 103),
  [2383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4017),
  [2385] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 103),
  [2387] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 103),
  [2389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4023),
  [2391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 147),
  [2393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 147),
  [2395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4028),
  [2397] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 178),
  [2399] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 178),
  [2401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4029),
  [2403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 178),
  [2405] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 178),
  [2407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4035),
  [2409] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 2, 0, 14),
  [2411] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 2, 0, 14),
  [2413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3997),
  [2415] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_argument, 5, 0, 88),
  [2417] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument, 5, 0, 88),
  [2419] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 33),
  [2421] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 33),
  [2423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4012),
  [2425] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 1, 0, 14),
  [2427] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 1, 0, 14),
  [2429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3981),
  [2431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3998),
  [2433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_pattern, 1, 0, 0),
  [2435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6848),
  [2437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [2439] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_pattern, 1, 0, 0),
  [2441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6234),
  [2443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8125),
  [2445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4130),
  [2447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4061),
  [2449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4106),
  [2451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3970),
  [2453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4086),
  [2455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3983),
  [2457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4038),
  [2459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3982),
  [2461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4126),
  [2463] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0),
  [2465] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2926),
  [2468] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0),
  [2470] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6447),
  [2473] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(139),
  [2476] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(1618),
  [2479] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(3032),
  [2482] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5425),
  [2485] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6034),
  [2488] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(1619),
  [2491] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6324),
  [2494] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6561),
  [2497] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6698),
  [2500] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(3034),
  [2503] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(8060),
  [2506] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5290),
  [2509] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(4876),
  [2512] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2916),
  [2515] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6964),
  [2518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(708),
  [2520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(663),
  [2522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4085),
  [2524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6605),
  [2526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [2528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3197),
  [2530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3160),
  [2532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6829),
  [2534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1923),
  [2536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4100),
  [2538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4103),
  [2540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4117),
  [2542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4120),
  [2544] = {.entry = {.count = 1, .reusable = false}}, SHIFT(838),
  [2546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4129),
  [2548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3999),
  [2550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4002),
  [2552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4088),
  [2554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4094),
  [2556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5193),
  [2558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [2560] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1636),
  [2562] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1616),
  [2564] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2113),
  [2566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8125),
  [2568] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1634),
  [2570] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1635),
  [2572] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2110),
  [2574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(796),
  [2576] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1905),
  [2578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5630),
  [2580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(142),
  [2582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1588),
  [2584] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5413),
  [2586] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5439),
  [2588] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1620),
  [2590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6280),
  [2592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1589),
  [2594] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1624),
  [2596] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6534),
  [2598] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2148),
  [2600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6539),
  [2602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6684),
  [2604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5416),
  [2606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7426),
  [2608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5281),
  [2610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4869),
  [2612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3845),
  [2614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6807),
  [2616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(866),
  [2618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5567),
  [2620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [2622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1614),
  [2624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5422),
  [2626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1622),
  [2628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1615),
  [2630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1586),
  [2632] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2149),
  [2634] = {.entry = {.count = 1, .reusable = false}}, SHIFT(784),
  [2636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [2638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [2640] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2855),
  [2643] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6605),
  [2646] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(140),
  [2649] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(1602),
  [2652] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2783),
  [2655] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5515),
  [2658] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6118),
  [2661] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(1603),
  [2664] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6441),
  [2667] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6520),
  [2670] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6671),
  [2673] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2784),
  [2676] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(7357),
  [2679] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5268),
  [2682] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(4881),
  [2685] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(3160),
  [2688] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6829),
  [2691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1888),
  [2693] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1903),
  [2695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1644),
  [2697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1660),
  [2699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1659),
  [2701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1642),
  [2703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1677),
  [2705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1666),
  [2707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1679),
  [2709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1670),
  [2711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1651),
  [2713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1676),
  [2715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1649),
  [2717] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6537),
  [2720] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(128),
  [2723] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(7357),
  [2726] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2357),
  [2729] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6958),
  [2732] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1921),
  [2734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(534),
  [2736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3152),
  [2738] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 3, 0, 78),
  [2740] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 3, 0, 78),
  [2742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(542),
  [2744] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 3, 0, 37),
  [2746] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 3, 0, 37),
  [2748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [2750] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 2, 0, 37),
  [2752] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 2, 0, 37),
  [2754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(575),
  [2756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1640),
  [2758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1687),
  [2760] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6273),
  [2762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1688),
  [2764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6499),
  [2766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6520),
  [2768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6671),
  [2770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [2772] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 5, 0, 118),
  [2774] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 118),
  [2776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(589),
  [2778] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 4, 0, 78),
  [2780] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 78),
  [2782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(448),
  [2784] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 4, 0, 118),
  [2786] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 118),
  [2788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3508),
  [2790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4970),
  [2792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7794),
  [2794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3150),
  [2796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5771),
  [2798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4790),
  [2800] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3171),
  [2802] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_payload, 1, 0, 0),
  [2804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5230),
  [2806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5540),
  [2808] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4191),
  [2810] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5527),
  [2812] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5670),
  [2814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3108),
  [2816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3124),
  [2818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3818),
  [2820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6020),
  [2822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6494),
  [2824] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6660),
  [2826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3797),
  [2828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7083),
  [2830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5565),
  [2832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5569),
  [2834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2319),
  [2836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2815),
  [2838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2816),
  [2840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3616),
  [2842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3618),
  [2844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5282),
  [2846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5283),
  [2848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5142),
  [2850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5564),
  [2852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5583),
  [2854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5143),
  [2856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2941),
  [2858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2942),
  [2860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6331),
  [2862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6599),
  [2864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2251),
  [2866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6924),
  [2868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3645),
  [2870] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_binding_pattern_content, 2, 0, 0),
  [2872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7025),
  [2874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5318),
  [2876] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_binding_pattern_content, 3, 0, 0),
  [2878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7101),
  [2880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7137),
  [2882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5722),
  [2884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6821),
  [2886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5709),
  [2888] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_pattern_content, 2, 0, 0),
  [2890] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_pattern_content, 3, 0, 0),
  [2892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6805),
  [2894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2377),
  [2896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7104),
  [2898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5068),
  [2900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3703),
  [2902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2266),
  [2904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7029),
  [2906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1817),
  [2908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(824),
  [2910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1824),
  [2912] = {.entry = {.count = 1, .reusable = false}}, SHIFT(693),
  [2914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(756),
  [2916] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5308),
  [2918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [2920] = {.entry = {.count = 1, .reusable = false}}, SHIFT(272),
  [2922] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1780),
  [2924] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4303),
  [2926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [2928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5761),
  [2930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5983),
  [2932] = {.entry = {.count = 1, .reusable = false}}, SHIFT(131),
  [2934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2367),
  [2936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1896),
  [2938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2231),
  [2940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6760),
  [2942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5306),
  [2944] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1808),
  [2946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1814),
  [2948] = {.entry = {.count = 1, .reusable = false}}, SHIFT(852),
  [2950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2299),
  [2952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [2954] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2400),
  [2956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1786),
  [2958] = {.entry = {.count = 1, .reusable = false}}, SHIFT(122),
  [2960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6552),
  [2962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6694),
  [2964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6691),
  [2966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2254),
  [2968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5286),
  [2970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4896),
  [2972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1907),
  [2974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3993),
  [2976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3194),
  [2978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6461),
  [2980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4043),
  [2982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4044),
  [2984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4060),
  [2986] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2509),
  [2988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(80),
  [2990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(393),
  [2992] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2513),
  [2994] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1794),
  [2996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4315),
  [2998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(394),
  [3000] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5750),
  [3002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5820),
  [3004] = {.entry = {.count = 1, .reusable = false}}, SHIFT(133),
  [3006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2449),
  [3008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6592),
  [3010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6722),
  [3012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2514),
  [3014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5302),
  [3016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4861),
  [3018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1927),
  [3020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2515),
  [3022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6865),
  [3024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4095),
  [3026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3297),
  [3028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [3030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(377),
  [3032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3347),
  [3034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1792),
  [3036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4331),
  [3038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [3040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5625),
  [3042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5991),
  [3044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(129),
  [3046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3332),
  [3048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6587),
  [3050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6719),
  [3052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3351),
  [3054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5300),
  [3056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4884),
  [3058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1924),
  [3060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3334),
  [3062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6789),
  [3064] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2283),
  [3066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(85),
  [3068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(307),
  [3070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2615),
  [3072] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1782),
  [3074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4286),
  [3076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [3078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5623),
  [3080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5995),
  [3082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(137),
  [3084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2551),
  [3086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6526),
  [3088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6676),
  [3090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2616),
  [3092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5269),
  [3094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4882),
  [3096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1912),
  [3098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2126),
  [3100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6777),
  [3102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4026),
  [3104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4030),
  [3106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4097),
  [3108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4079),
  [3110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2371),
  [3112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [3114] = {.entry = {.count = 1, .reusable = false}}, SHIFT(345),
  [3116] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2389),
  [3118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1787),
  [3120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4330),
  [3122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(346),
  [3124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5674),
  [3126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5843),
  [3128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(123),
  [3130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2374),
  [3132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6566),
  [3134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6703),
  [3136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2416),
  [3138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5295),
  [3140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4895),
  [3142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1904),
  [3144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2440),
  [3146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7067),
  [3148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4083),
  [3150] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression, 1, 0, 6),
  [3152] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression, 1, 0, 6),
  [3154] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_anonymous, 3, 0, 0),
  [3156] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression_anonymous, 3, 0, 0),
  [3158] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_anonymous, 4, 0, 0),
  [3160] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression_anonymous, 4, 0, 0),
  [3162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6596),
  [3164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6724),
  [3166] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_id, 1, 0, 0),
  [3168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_id, 1, 0, 0),
  [3170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7078),
  [3172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(598),
  [3174] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_id, 2, 0, 0),
  [3176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_id, 2, 0, 0),
  [3178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [3180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [3182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(384),
  [3184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [3186] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0),
  [3188] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0),
  [3190] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(7078),
  [3193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(810),
  [3195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(740),
  [3197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(677),
  [3199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4109),
  [3201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(553),
  [3203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7357),
  [3205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(343),
  [3207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(558),
  [3209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(441),
  [3211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [3213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [3215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [3217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(382),
  [3219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [3221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [3223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [3225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [3227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [3229] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__structure_repeat1, 2, 0, 0),
  [3231] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__structure_repeat1, 2, 0, 0),
  [3233] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat1, 2, 0, 0), SHIFT_REPEAT(2076),
  [3236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(551),
  [3238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [3240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(467),
  [3242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(400),
  [3244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [3246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [3248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(723),
  [3250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(402),
  [3252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(410),
  [3254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [3256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(447),
  [3258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [3260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(395),
  [3262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(413),
  [3264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(450),
  [3266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(443),
  [3268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [3270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [3272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [3274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [3276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1922),
  [3278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(782),
  [3280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__structure_repeat1, 1, 0, 0),
  [3282] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__structure_repeat1, 1, 0, 0),
  [3284] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_infix_expression, 3, 0, 44),
  [3286] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_infix_expression, 3, 0, 44),
  [3288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3229),
  [3290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(699),
  [3292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3244),
  [3294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3230),
  [3296] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3243),
  [3298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3243),
  [3300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(700),
  [3302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(701),
  [3304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3245),
  [3306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3245),
  [3308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4016),
  [3310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3987),
  [3312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(698),
  [3314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2),
  [3316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(702),
  [3318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3978),
  [3320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_expression, 3, 0, 41),
  [3322] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_expression, 3, 0, 41),
  [3324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4125),
  [3326] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sign_expression, 2, 0, 19),
  [3328] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sign_expression, 2, 0, 19),
  [3330] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_then_clause, 2, 0, 18),
  [3332] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_then_clause, 2, 0, 18),
  [3334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4077),
  [3336] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression, 1, 0, 0),
  [3338] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression, 1, 0, 0),
  [3340] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_clause, 2, 0, 18),
  [3342] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_clause, 2, 0, 18),
  [3344] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_expression, 3, 0, 0),
  [3346] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_expression, 3, 0, 0),
  [3348] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cons_expression, 3, 0, 46),
  [3350] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cons_expression, 3, 0, 46),
  [3352] = {.entry = {.count = 1, .reusable = false}}, SHIFT(878),
  [3354] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1621),
  [3356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(828),
  [3358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(829),
  [3360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(830),
  [3362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(831),
  [3364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(832),
  [3366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1583),
  [3368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4926),
  [3370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(653),
  [3372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(654),
  [3374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(655),
  [3376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(656),
  [3378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(657),
  [3380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4),
  [3382] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 2, 0, 8),
  [3384] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 2, 0, 8),
  [3386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2773),
  [3388] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signed_constant, 1, 0, 0),
  [3390] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__signed_constant, 1, 0, 0),
  [3392] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3),
  [3394] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signed_number, 2, 0, 0),
  [3396] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signed_number, 2, 0, 0),
  [3398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(787),
  [3400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5),
  [3402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(882),
  [3404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [3406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(789),
  [3408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(790),
  [3410] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assert_expression, 2, 0, 18),
  [3412] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assert_expression, 2, 0, 18),
  [3414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5309),
  [3416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7695),
  [3418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1898),
  [3420] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_expression, 2, 0, 18),
  [3422] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_expression, 2, 0, 18),
  [3424] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_binding_pattern_no_exn, 3, 0, 0),
  [3426] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_binding_pattern_no_exn, 3, 0, 0),
  [3428] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_pattern, 5, 0, 127),
  [3430] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_pattern, 5, 0, 127),
  [3432] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_pattern, 5, 0, 128),
  [3434] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_pattern, 5, 0, 128),
  [3436] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 5, 0, 0),
  [3438] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 5, 0, 0),
  [3440] = {.entry = {.count = 1, .reusable = false}}, SHIFT(93),
  [3442] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__effect_binding_pattern, 1, 0, 0),
  [3444] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__effect_binding_pattern, 1, 0, 0),
  [3446] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binding_pattern, 1, 0, 0),
  [3448] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__binding_pattern, 1, 0, 0),
  [3450] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 5, 0, 126),
  [3452] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 5, 0, 126),
  [3454] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binding_pattern, 1, 0, 16),
  [3456] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__binding_pattern, 1, 0, 16),
  [3458] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binding_pattern, 1, 0, 17),
  [3460] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__binding_pattern, 1, 0, 17),
  [3462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_binding_pattern, 2, 0, 0),
  [3464] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_binding_pattern, 2, 0, 0),
  [3466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1665),
  [3468] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_constructor_path, 1, 0, 11),
  [3470] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_constructor_path, 1, 0, 11),
  [3472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__or_binding_pattern_no_exn_anonymous, 3, 0, 0),
  [3474] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__or_binding_pattern_no_exn_anonymous, 3, 0, 0),
  [3476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1705),
  [3478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1708),
  [3480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_pattern, 2, 0, 0),
  [3482] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_pattern, 2, 0, 0),
  [3484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assert_expression, 3, 0, 40),
  [3486] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assert_expression, 3, 0, 40),
  [3488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_expression, 3, 0, 40),
  [3490] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_expression, 3, 0, 40),
  [3492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alias_binding_pattern_no_exn, 3, 0, 54),
  [3494] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alias_binding_pattern_no_exn, 3, 0, 54),
  [3496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cons_binding_pattern_no_exn, 3, 0, 46),
  [3498] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cons_binding_pattern_no_exn, 3, 0, 46),
  [3500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_binding_pattern, 2, 0, 22),
  [3502] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tag_binding_pattern, 2, 0, 22),
  [3504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_binding_pattern, 3, 0, 0),
  [3506] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_binding_pattern, 3, 0, 0),
  [3508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5260),
  [3510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8254),
  [3512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1919),
  [3514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element, 4, 0, 55),
  [3516] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element, 4, 0, 55),
  [3518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_pattern, 6, 0, 168),
  [3520] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_pattern, 6, 0, 168),
  [3522] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_binding_pattern, 2, 0, 0),
  [3524] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_binding_pattern, 2, 0, 0),
  [3526] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_binding_pattern, 2, 0, 22),
  [3528] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_binding_pattern, 2, 0, 22),
  [3530] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_pattern, 3, 0, 46),
  [3532] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_pattern, 3, 0, 46),
  [3534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_binding_pattern, 2, 0, 22),
  [3536] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_binding_pattern, 2, 0, 22),
  [3538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6085),
  [3540] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binding_pattern_no_exn, 1, 0, 0),
  [3542] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__binding_pattern_no_exn, 1, 0, 0),
  [3544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_binding_pattern, 2, 0, 22),
  [3546] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_binding_pattern, 2, 0, 22),
  [3548] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 4, 0, 82),
  [3550] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 4, 0, 82),
  [3552] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_constructor_path, 3, 0, 59),
  [3554] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_constructor_path, 3, 0, 59),
  [3556] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_binding_pattern, 3, 0, 49),
  [3558] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_binding_pattern, 3, 0, 49),
  [3560] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_binding_pattern, 3, 0, 0),
  [3562] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_binding_pattern, 3, 0, 0),
  [3564] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_binding_pattern, 3, 0, 0),
  [3566] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_binding_pattern, 3, 0, 0),
  [3568] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__or_binding_pattern_anonymous, 3, 0, 0),
  [3570] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__or_binding_pattern_anonymous, 3, 0, 0),
  [3572] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alias_binding_pattern, 3, 0, 54),
  [3574] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alias_binding_pattern, 3, 0, 54),
  [3576] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cons_binding_pattern, 3, 0, 46),
  [3578] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cons_binding_pattern, 3, 0, 46),
  [3580] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_effect_binding_pattern, 4, 0, 122),
  [3582] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_effect_binding_pattern, 4, 0, 122),
  [3584] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 3, 0, 0),
  [3586] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 3, 0, 0),
  [3588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_binding_pattern, 3, 0, 0),
  [3590] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_binding_pattern, 3, 0, 0),
  [3592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_local_open_binding_pattern, 3, 0, 49),
  [3594] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_local_open_binding_pattern, 3, 0, 49),
  [3596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_pattern, 4, 0, 47),
  [3598] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_pattern, 4, 0, 47),
  [3600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_binding_pattern, 3, 0, 49),
  [3602] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_binding_pattern, 3, 0, 49),
  [3604] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_binding_pattern, 4, 0, 83),
  [3606] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_binding_pattern, 4, 0, 83),
  [3608] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_binding_pattern, 3, 0, 49),
  [3610] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_binding_pattern, 3, 0, 49),
  [3612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__delimited_binding_pattern, 1, 0, 0),
  [3614] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__delimited_binding_pattern, 1, 0, 0),
  [3616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 4, 0, 0),
  [3618] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 4, 0, 0),
  [3620] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 7, 0, 0),
  [3622] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 7, 0, 0),
  [3624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 6, 0, 0),
  [3626] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 6, 0, 0),
  [3628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5254),
  [3630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7991),
  [3632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1920),
  [3634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5336),
  [3636] = {.entry = {.count = 1, .reusable = false}}, SHIFT(94),
  [3638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(856),
  [3640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(857),
  [3642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(858),
  [3644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [3646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(860),
  [3648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(95),
  [3650] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6),
  [3652] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8094),
  [3654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1928),
  [3656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_expression, 2, 0, 0),
  [3658] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_expression, 2, 0, 0),
  [3660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(96),
  [3662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__parenthesized_abstract_type, 3, 0, 0),
  [3664] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__parenthesized_abstract_type, 3, 0, 0),
  [3666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fun_expression, 4, 0, 50),
  [3668] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fun_expression, 4, 0, 50),
  [3670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(757),
  [3672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(759),
  [3674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(760),
  [3676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(761),
  [3678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(762),
  [3680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 4, 0, 39),
  [3682] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 4, 0, 39),
  [3684] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_clause, 2, 0, 0),
  [3686] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_clause, 2, 0, 0),
  [3688] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1, 0, 5),
  [3690] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1, 0, 5),
  [3692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_expression, 4, 0, 80),
  [3694] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_expression, 4, 0, 80),
  [3696] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_expression, 4, 0, 18),
  [3698] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_expression, 4, 0, 18),
  [3700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7),
  [3702] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0),
  [3704] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0),
  [3706] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_refutation_case, 1, 0, 0),
  [3708] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_refutation_case, 1, 0, 0),
  [3710] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_case, 3, 0, 51),
  [3712] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_case, 3, 0, 51),
  [3714] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_expression, 4, 0, 18),
  [3716] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_expression, 4, 0, 18),
  [3718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_expression, 5, 0, 87),
  [3720] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_expression, 5, 0, 87),
  [3722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element, 5, 0, 88),
  [3724] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element, 5, 0, 88),
  [3726] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 3, 0, 39),
  [3728] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 3, 0, 39),
  [3730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(710),
  [3732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3078),
  [3734] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6103),
  [3736] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2876),
  [3738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5394),
  [3740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5159),
  [3742] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3105),
  [3744] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5344),
  [3746] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6364),
  [3748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5220),
  [3750] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1911),
  [3752] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element, 2, 0, 8),
  [3754] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element, 2, 0, 8),
  [3756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_expression, 3, 0, 41),
  [3758] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_expression, 3, 0, 41),
  [3760] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2880),
  [3762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5380),
  [3764] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3539),
  [3766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3064),
  [3768] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6209),
  [3770] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2875),
  [3772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4087),
  [3774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3776),
  [3776] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3086),
  [3778] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4045),
  [3780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6408),
  [3782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3101),
  [3784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3133),
  [3786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3844),
  [3788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6281),
  [3790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6574),
  [3792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6709),
  [3794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3877),
  [3796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6810),
  [3798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2883),
  [3800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4005),
  [3802] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_class_expression, 3, 0, 0),
  [3804] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_class_expression, 3, 0, 0),
  [3806] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fun_expression, 5, 0, 87),
  [3808] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fun_expression, 5, 0, 87),
  [3810] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_class_expression, 1, 0, 0),
  [3812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_class_expression, 1, 0, 0),
  [3814] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fun_expression, 5, 0, 121),
  [3816] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fun_expression, 5, 0, 121),
  [3818] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 5, 0, 80),
  [3820] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 5, 0, 80),
  [3822] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_clause, 3, 0, 0),
  [3824] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_clause, 3, 0, 0),
  [3826] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_expression, 5, 0, 40),
  [3828] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_expression, 5, 0, 40),
  [3830] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_case, 4, 0, 86),
  [3832] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_case, 4, 0, 86),
  [3834] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_expression, 5, 0, 40),
  [3836] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_expression, 5, 0, 40),
  [3838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3922),
  [3840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3081),
  [3842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2887),
  [3844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4414),
  [3846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4067),
  [3848] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3104),
  [3850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3109),
  [3852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3142),
  [3854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3825),
  [3856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6291),
  [3858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6594),
  [3860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6723),
  [3862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6949),
  [3864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1884),
  [3866] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__match_cases, 2, 0, 0),
  [3868] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__match_cases, 2, 0, 0),
  [3870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1658),
  [3872] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__match_cases, 3, 0, 0),
  [3874] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__match_cases, 3, 0, 0),
  [3876] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1658),
  [3879] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__match_cases, 1, 0, 0),
  [3881] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__match_cases, 1, 0, 0),
  [3883] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_expression, 8, 0, 210),
  [3885] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_expression, 8, 0, 210),
  [3887] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_expression, 3, 0, 39),
  [3889] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_expression, 3, 0, 39),
  [3891] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_class_expression, 4, 0, 156),
  [3893] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_class_expression, 4, 0, 156),
  [3895] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instantiated_class, 4, 0, 0),
  [3897] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_class, 4, 0, 0),
  [3899] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fun_expression, 6, 0, 165),
  [3901] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fun_expression, 6, 0, 165),
  [3903] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_expression, 3, 0, 0),
  [3905] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_expression, 3, 0, 0),
  [3907] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2885),
  [3909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4421),
  [3911] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instantiated_class, 5, 0, 0),
  [3913] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_class, 5, 0, 0),
  [3915] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 4, 0, 80),
  [3917] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 4, 0, 80),
  [3919] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_expression, 7, 0, 192),
  [3921] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_expression, 7, 0, 192),
  [3923] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_expression, 4, 0, 50),
  [3925] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_expression, 4, 0, 50),
  [3927] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1652),
  [3929] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1877),
  [3931] = {.entry = {.count = 1, .reusable = false}}, SHIFT(98),
  [3933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(840),
  [3935] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1671),
  [3937] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1671),
  [3940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(665),
  [3942] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1652),
  [3945] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1890),
  [3947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [3949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5134),
  [3951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7171),
  [3953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [3955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [3957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5516),
  [3959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6851),
  [3961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [3963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(100),
  [3965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3600),
  [3967] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 3, 0, 0),
  [3969] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 3, 0, 0),
  [3971] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_pattern, 3, 0, 0),
  [3973] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_pattern, 3, 0, 0),
  [3975] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_local_open_pattern, 3, 0, 49),
  [3977] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_local_open_pattern, 3, 0, 49),
  [3979] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 9, 0, 228),
  [3981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 9, 0, 228),
  [3983] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 4, 0, 49),
  [3985] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 4, 0, 49),
  [3987] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 4, 0, 82),
  [3989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 4, 0, 82),
  [3991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(798),
  [3993] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_pattern, 4, 0, 83),
  [3995] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_pattern, 4, 0, 83),
  [3997] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 4, 0, 83),
  [3999] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 4, 0, 83),
  [4001] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 4, 0, 0),
  [4003] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 4, 0, 0),
  [4005] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 2, 0, 22),
  [4007] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 2, 0, 22),
  [4009] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_pattern, 2, 0, 0),
  [4011] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 2, 0, 0),
  [4013] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_pattern, 2, 0, 0),
  [4015] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_pattern, 2, 0, 0),
  [4017] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1675),
  [4019] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1675),
  [4022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [4024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6561),
  [4026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6698),
  [4028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [4030] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 5, 0, 164),
  [4032] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 5, 0, 164),
  [4034] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 5, 0, 0),
  [4036] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 5, 0, 0),
  [4038] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_pattern, 3, 0, 0),
  [4040] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_pattern, 3, 0, 0),
  [4042] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__delimited_pattern, 1, 0, 0),
  [4044] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__delimited_pattern, 1, 0, 0),
  [4046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [4048] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_value_pattern, 1, 0, 15),
  [4050] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_value_pattern, 1, 0, 15),
  [4052] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 6, 0, 191),
  [4054] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 6, 0, 191),
  [4056] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 1, 0, 0),
  [4058] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 1, 0, 0),
  [4060] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 6, 0, 0),
  [4062] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 6, 0, 0),
  [4064] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 8, 0, 221),
  [4066] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 8, 0, 221),
  [4068] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 1, 0, 7),
  [4070] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 1, 0, 7),
  [4072] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 7, 0, 208),
  [4074] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 7, 0, 208),
  [4076] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_pattern, 3, 0, 0),
  [4078] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 3, 0, 0),
  [4080] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 7, 0, 209),
  [4082] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 7, 0, 209),
  [4084] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 7, 0, 0),
  [4086] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 7, 0, 0),
  [4088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4098),
  [4090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5372),
  [4092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(881),
  [4094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1647),
  [4096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5391),
  [4098] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1647),
  [4101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3973),
  [4103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4426),
  [4105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4419),
  [4107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2843),
  [4109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1654),
  [4111] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1654),
  [4114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(728),
  [4116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5280),
  [4118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7374),
  [4120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1899),
  [4122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4056),
  [4124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4177),
  [4126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7650),
  [4128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(814),
  [4130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(815),
  [4132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(816),
  [4134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(817),
  [4136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(818),
  [4138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(48),
  [4140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5399),
  [4142] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4179),
  [4144] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4199),
  [4146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3508),
  [4148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3207),
  [4150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(683),
  [4152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [4154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(684),
  [4156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(685),
  [4158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(686),
  [4160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(687),
  [4162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3156),
  [4164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3526),
  [4166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1926),
  [4168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3192),
  [4170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3210),
  [4172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3123),
  [4174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3650),
  [4176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3136),
  [4178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3921),
  [4180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(843),
  [4182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [4184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [4186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(845),
  [4188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(846),
  [4190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(745),
  [4192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(747),
  [4194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(748),
  [4196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(749),
  [4198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3754),
  [4200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3193),
  [4202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(57),
  [4204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(750),
  [4206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3130),
  [4208] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3131),
  [4210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3208),
  [4212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [4214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3897),
  [4216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3213),
  [4218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2854),
  [4220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3172),
  [4222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3586),
  [4224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3139),
  [4226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3141),
  [4228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3176),
  [4230] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1733),
  [4232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5579),
  [4234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4800),
  [4236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5210),
  [4238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5434),
  [4240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4552),
  [4242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4665),
  [4244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5490),
  [4246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5613),
  [4248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1823),
  [4250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(109),
  [4252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3863),
  [4254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7091),
  [4256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3165),
  [4258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3167),
  [4260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4776),
  [4262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3984),
  [4264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7098),
  [4266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3168),
  [4268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3149),
  [4270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3169),
  [4272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7048),
  [4274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3154),
  [4276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3155),
  [4278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4778),
  [4280] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(2076),
  [4283] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(1733),
  [4286] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(5579),
  [4289] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(4800),
  [4292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0),
  [4294] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(5210),
  [4297] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(5434),
  [4300] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(4552),
  [4303] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(4665),
  [4306] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(5490),
  [4309] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(5613),
  [4312] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(6390),
  [4315] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(6636),
  [4318] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(6327),
  [4321] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(1823),
  [4324] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(6878),
  [4327] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_toplevel_directive, 1, 0, 0),
  [4329] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_toplevel_directive, 1, 0, 0),
  [4331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3504),
  [4333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4782),
  [4335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7068),
  [4337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4783),
  [4339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5273),
  [4341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4886),
  [4343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3536),
  [4345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1655),
  [4347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1682),
  [4349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5487),
  [4351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1816),
  [4353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [4355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2846),
  [4357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sign_operator, 1, 0, 0),
  [4359] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sign_operator, 1, 0, 0),
  [4361] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_add_operator, 1, 0, 0),
  [4363] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__attribute, 2, 0, 0),
  [4365] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__attribute, 2, 0, 0),
  [4367] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1655),
  [4370] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(1682),
  [4373] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(5487),
  [4376] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(1816),
  [4379] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_and_operator, 1, 0, 0),
  [4381] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_and_operator, 1, 0, 0),
  [4383] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pow_operator, 1, 0, 0),
  [4385] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pow_operator, 1, 0, 0),
  [4387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(114),
  [4389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(749),
  [4391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(696),
  [4393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1657),
  [4395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor, 4, 0, 50),
  [4397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor, 4, 0, 50),
  [4399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4618),
  [4401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6508),
  [4403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6664),
  [4405] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 3, 0, 111),
  [4407] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 3, 0, 111),
  [4409] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__proper_tuple_type, 3, 0, 0),
  [4411] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__proper_tuple_type, 3, 0, 0),
  [4413] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_aliased_type, 3, 0, 113),
  [4415] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_aliased_type, 3, 0, 113),
  [4417] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_of, 4, 0, 67),
  [4419] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_of, 4, 0, 67),
  [4421] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__labeled_tuple_type, 3, 0, 0),
  [4423] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__labeled_tuple_type, 3, 0, 0),
  [4425] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mult_operator, 1, 0, 0),
  [4427] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mult_operator, 1, 0, 0),
  [4429] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_add_operator, 1, 0, 0),
  [4431] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_or_operator, 1, 0, 0),
  [4433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_or_operator, 1, 0, 0),
  [4435] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type, 1, 0, 0),
  [4437] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type, 1, 0, 0),
  [4439] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 5, 0, 181),
  [4441] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 5, 0, 181),
  [4443] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1657),
  [4446] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_type_rhs, 3, 0, 0),
  [4448] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_type_rhs, 3, 0, 0),
  [4450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(769),
  [4452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(854),
  [4454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7159),
  [4456] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3400),
  [4458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3392),
  [4460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3402),
  [4462] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(7159),
  [4465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3366),
  [4467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5330),
  [4469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7645),
  [4471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1925),
  [4473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3420),
  [4475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3421),
  [4477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3445),
  [4479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3365),
  [4481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3441),
  [4483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3922),
  [4485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [4487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(801),
  [4489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(802),
  [4491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(803),
  [4493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(804),
  [4495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3539),
  [4497] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_type, 3, 0, 0),
  [4499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3594),
  [4501] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_type, 3, 0, 0),
  [4503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6195),
  [4505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [4507] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_variable, 2, 0, 0),
  [4509] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_variable, 2, 0, 0),
  [4511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7111),
  [4513] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1878),
  [4515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(730),
  [4517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(731),
  [4519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(732),
  [4521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(733),
  [4523] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(7111),
  [4526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3419),
  [4528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6138),
  [4530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [4532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(734),
  [4534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_type, 3, 0, 110),
  [4536] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_type, 3, 0, 110),
  [4538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6123),
  [4540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3423),
  [4542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3920),
  [4544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(667),
  [4546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(668),
  [4548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(669),
  [4550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(670),
  [4552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [4554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(671),
  [4556] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_type_rhs, 1, 0, 0),
  [4558] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_type_rhs, 1, 0, 0),
  [4560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3873),
  [4562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6271),
  [4564] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_type, 5, 0, 180),
  [4566] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_type, 5, 0, 180),
  [4568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8276),
  [4570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7799),
  [4572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7484),
  [4574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8222),
  [4576] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3801),
  [4578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4936),
  [4580] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_type, 5, 0, 0),
  [4582] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_type, 5, 0, 0),
  [4584] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructed_type, 5, 0, 0),
  [4586] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructed_type, 5, 0, 0),
  [4588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 5, 0, 0),
  [4590] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 5, 0, 0),
  [4592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 5, 0, 0),
  [4594] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 5, 0, 0),
  [4596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 4, 0, 98),
  [4598] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 4, 0, 98),
  [4600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_type, 6, 0, 203),
  [4602] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_type, 6, 0, 203),
  [4604] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_type, 6, 0, 0),
  [4606] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_type, 6, 0, 0),
  [4608] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 6, 0, 0),
  [4610] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 6, 0, 0),
  [4612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 6, 0, 0),
  [4614] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 6, 0, 0),
  [4616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructed_type, 4, 0, 0),
  [4618] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructed_type, 4, 0, 0),
  [4620] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 2, 0, 0),
  [4622] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 2, 0, 0),
  [4624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_type, 7, 0, 219),
  [4626] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_type, 7, 0, 219),
  [4628] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 7, 0, 0),
  [4630] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 7, 0, 0),
  [4632] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 3, 0, 63),
  [4634] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 3, 0, 63),
  [4636] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 2, 0, 0),
  [4638] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 2, 0, 0),
  [4640] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_path, 1, 0, 71),
  [4642] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_path, 1, 0, 71),
  [4644] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 4, 0, 99),
  [4646] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 4, 0, 99),
  [4648] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_path, 1, 0, 3),
  [4650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(713),
  [4652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(714),
  [4654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(715),
  [4656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(716),
  [4658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(717),
  [4660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [4662] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_type, 1, 0, 0),
  [4664] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_type, 1, 0, 0),
  [4666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__delimited_type, 1, 0, 0),
  [4668] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__delimited_type, 1, 0, 0),
  [4670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(90),
  [4672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 4, 0, 0),
  [4674] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 4, 0, 0),
  [4676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_type, 2, 0, 0),
  [4678] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_type, 2, 0, 0),
  [4680] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2818),
  [4682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructed_type, 2, 0, 0),
  [4684] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructed_type, 2, 0, 0),
  [4686] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_type, 4, 0, 125),
  [4688] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_type, 4, 0, 125),
  [4690] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 4, 0, 0),
  [4692] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 4, 0, 0),
  [4694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_path, 3, 0, 150),
  [4696] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_path, 3, 0, 150),
  [4698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 5, 0, 143),
  [4700] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 5, 0, 143),
  [4702] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__module_expression, 1, 0, 0),
  [4704] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__module_expression, 1, 0, 0),
  [4706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7312),
  [4708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 3, 0, 0),
  [4710] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 3, 0, 0),
  [4712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 3, 0, 0),
  [4714] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 3, 0, 0),
  [4716] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4001),
  [4718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_type, 3, 0, 0),
  [4720] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_type, 3, 0, 0),
  [4722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_local_open_type, 3, 0, 114),
  [4724] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_local_open_type, 3, 0, 114),
  [4726] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_path, 3, 0, 47),
  [4728] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_constructor_declaration_repeat1, 1, 0, 0),
  [4730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(772),
  [4732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(773),
  [4734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(774),
  [4736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(775),
  [4738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(776),
  [4740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(120),
  [4742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5771),
  [4744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4790),
  [4746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5230),
  [4748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5540),
  [4750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4191),
  [4752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5527),
  [4754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5670),
  [4756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3947),
  [4758] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_declaration, 3, 0, 0),
  [4760] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_declaration, 3, 0, 0),
  [4762] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_argument, 1, 0, 0),
  [4764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4075),
  [4766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1672),
  [4768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1746),
  [4770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1744),
  [4772] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6252),
  [4774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [4776] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 2, 0, 13),
  [4778] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 2, 0, 13),
  [4780] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_module_expression, 1, 0, 0),
  [4782] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_module_expression, 1, 0, 0),
  [4784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 3, 0, 31),
  [4786] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 3, 0, 31),
  [4788] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_structure, 2, 0, 0),
  [4790] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_structure, 2, 0, 0),
  [4792] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_declaration, 5, 0, 0),
  [4794] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_declaration, 5, 0, 0),
  [4796] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_application, 2, 0, 32),
  [4798] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_application, 2, 0, 32),
  [4800] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_packed_module, 5, 0, 144),
  [4802] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_packed_module, 5, 0, 144),
  [4804] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module, 3, 0, 31),
  [4806] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_include_module, 3, 0, 31),
  [4808] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module, 2, 0, 13),
  [4810] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_include_module, 2, 0, 13),
  [4812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_packed_module, 6, 0, 174),
  [4814] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_packed_module, 6, 0, 174),
  [4816] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4970),
  [4819] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5771),
  [4822] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4790),
  [4825] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0),
  [4827] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5230),
  [4830] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5540),
  [4833] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4552),
  [4836] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4191),
  [4839] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5527),
  [4842] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5670),
  [4845] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(6390),
  [4848] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(6636),
  [4851] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(6327),
  [4854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5525),
  [4856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3735),
  [4858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4408),
  [4860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4471),
  [4862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5246),
  [4864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4438),
  [4866] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_module_expression, 3, 0, 0),
  [4868] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_module_expression, 3, 0, 0),
  [4870] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_packed_module, 4, 0, 100),
  [4872] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_packed_module, 4, 0, 100),
  [4874] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 4, 0, 67),
  [4876] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 4, 0, 67),
  [4878] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_module_expression, 4, 0, 101),
  [4880] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_module_expression, 4, 0, 101),
  [4882] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_packed_module, 7, 0, 194),
  [4884] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_packed_module, 7, 0, 194),
  [4886] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signature, 1, 0, 0),
  [4888] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_structure, 3, 0, 0),
  [4890] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_structure, 3, 0, 0),
  [4892] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_application, 3, 0, 68),
  [4894] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_application, 3, 0, 68),
  [4896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(869),
  [4898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(870),
  [4900] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 2, 0, 3),
  [4902] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 2, 0, 3),
  [4904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4852),
  [4906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6348),
  [4908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4389),
  [4910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1641),
  [4912] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1641),
  [4915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3913),
  [4917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3267),
  [4919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(812),
  [4921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3594),
  [4923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3272),
  [4925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(626),
  [4927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_content, 1, 0, 0),
  [4929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signature, 2, 0, 0),
  [4931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(651),
  [4933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [4935] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 1, 0, 3),
  [4937] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 1, 0, 3),
  [4939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4822),
  [4941] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variant_declaration, 1, 0, 0),
  [4943] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variant_declaration, 1, 0, 0),
  [4945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6769),
  [4947] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8308),
  [4949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4045),
  [4951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4051),
  [4953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(872),
  [4955] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_argument, 1, 0, 0),
  [4957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4063),
  [4959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [4961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3913),
  [4963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6077),
  [4965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3419),
  [4967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3301),
  [4969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(632),
  [4971] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extended_module_path, 1, 0, 3),
  [4973] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extended_module_path, 1, 0, 3),
  [4975] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 4, 0, 95),
  [4977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(742),
  [4979] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_path, 3, 0, 29),
  [4981] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_path, 3, 0, 29),
  [4983] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extended_module_path, 3, 0, 47),
  [4985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5666),
  [4987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3163),
  [4989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5420),
  [4991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7858),
  [4993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3882),
  [4995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7556),
  [4997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3891),
  [4999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8348),
  [5001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3729),
  [5003] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 3, 0, 41),
  [5005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7806),
  [5007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3514),
  [5009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1669),
  [5011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7363),
  [5013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3714),
  [5015] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__sequence_expression_content_repeat1, 2, 0, 0),
  [5017] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 5, 0, 222),
  [5019] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 5, 0, 222),
  [5021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3425),
  [5023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6914),
  [5025] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 1, 0, 10),
  [5027] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 1, 0, 10),
  [5029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2674),
  [5031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2770),
  [5033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [5035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8339),
  [5037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3651),
  [5039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [5041] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 6, 0, 229),
  [5043] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 6, 0, 229),
  [5045] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_constraint, 4, 0, 112),
  [5047] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_constraint, 4, 0, 112),
  [5049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7216),
  [5051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3923),
  [5053] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_expression, 3, 0, 40),
  [5055] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1669),
  [5058] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__constructor_argument_repeat1, 2, 0, 0),
  [5060] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__constructor_argument_repeat1, 2, 0, 0),
  [5062] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_path, 1, 0, 64),
  [5064] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_path, 1, 0, 64),
  [5066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3883),
  [5068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6264),
  [5070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(681),
  [5072] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 2, 0, 26),
  [5074] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 2, 0, 26),
  [5076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2675),
  [5078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2772),
  [5080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3140),
  [5082] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extended_module_path, 4, 0, 0),
  [5084] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extended_module_path, 4, 0, 0),
  [5086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7760),
  [5088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3560),
  [5090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7311),
  [5092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3532),
  [5094] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extended_module_path, 3, 0, 47),
  [5096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3414),
  [5098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3910),
  [5100] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 4, 0, 211),
  [5102] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 4, 0, 211),
  [5104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_module, 4, 0, 212),
  [5106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_module, 4, 0, 212),
  [5108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7172),
  [5110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7510),
  [5112] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0),
  [5114] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0),
  [5116] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(3425),
  [5119] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 91),
  [5121] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 91),
  [5123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4969),
  [5125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3589),
  [5127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4186),
  [5129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3833),
  [5131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4753),
  [5133] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 6, 0, 222),
  [5135] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 6, 0, 222),
  [5137] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 7, 0, 229),
  [5139] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 7, 0, 229),
  [5141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(726),
  [5143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3899),
  [5145] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_binding_repeat1, 2, 0, 0),
  [5147] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_binding_repeat1, 2, 0, 0),
  [5149] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6348),
  [5152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 5, 0, 211),
  [5154] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 5, 0, 211),
  [5156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 3, 0, 58),
  [5158] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 3, 0, 58),
  [5160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4951),
  [5162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4019),
  [5164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_declaration, 7, 0, 0),
  [5166] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_declaration, 7, 0, 0),
  [5168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_directive, 2, 0, 0),
  [5170] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_directive, 2, 0, 0),
  [5172] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 1, 0, 0),
  [5174] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_type_binding_repeat1, 1, 0, 0),
  [5176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_open_class_expression, 4, 0, 50),
  [5178] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_open_class_expression, 4, 0, 50),
  [5180] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_class_expression, 3, 0, 41),
  [5182] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_class_expression, 3, 0, 41),
  [5184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_function, 4, 0, 50),
  [5186] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_function, 4, 0, 50),
  [5188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1653),
  [5190] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__polymorphic_type, 1, 0, 0),
  [5192] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__polymorphic_type, 1, 0, 0),
  [5194] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__maybe_polymorphic_typed, 2, 0, 52),
  [5196] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__maybe_polymorphic_typed, 2, 0, 52),
  [5198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(785),
  [5200] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0),
  [5202] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0),
  [5204] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0), SHIFT_REPEAT(6340),
  [5207] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1653),
  [5210] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external, 5, 0, 35),
  [5212] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external, 5, 0, 35),
  [5214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6916),
  [5216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4860),
  [5218] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external, 6, 0, 74),
  [5220] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external, 6, 0, 74),
  [5222] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signature, 2, 0, 0),
  [5224] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signature, 2, 0, 0),
  [5226] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_module_type, 3, 0, 0),
  [5228] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_module_type, 3, 0, 0),
  [5230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_class_type, 1, 0, 0),
  [5232] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_class_type, 1, 0, 0),
  [5234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(978),
  [5236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1598),
  [5238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3990),
  [5240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3992),
  [5242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3399),
  [5244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2109),
  [5246] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signature, 3, 0, 0),
  [5248] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signature, 3, 0, 0),
  [5250] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__module_type, 1, 0, 0),
  [5252] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__module_type, 1, 0, 0),
  [5254] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0),
  [5256] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0),
  [5258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 90),
  [5260] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 90),
  [5262] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_module_type, 5, 0, 223),
  [5264] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_module_type, 5, 0, 223),
  [5266] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 137),
  [5268] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 137),
  [5270] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2902),
  [5272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1610),
  [5274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4057),
  [5276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3374),
  [5278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4244),
  [5280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2115),
  [5282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6577),
  [5284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6711),
  [5286] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_parameter, 2, 0, 0),
  [5288] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_parameter, 2, 0, 0),
  [5290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_body_type, 6, 0, 201),
  [5292] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_body_type, 6, 0, 201),
  [5294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 91),
  [5296] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 91),
  [5298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 3, 0, 26),
  [5300] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 3, 0, 26),
  [5302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1627),
  [5304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2157),
  [5306] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 58),
  [5308] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 58),
  [5310] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_parameter, 4, 0, 140),
  [5312] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_parameter, 4, 0, 140),
  [5314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3048),
  [5316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1629),
  [5318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3412),
  [5320] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2162),
  [5322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_constraint, 3, 0, 142),
  [5324] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_constraint, 3, 0, 142),
  [5326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5878),
  [5328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1626),
  [5330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2150),
  [5332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_body_type, 3, 0, 0),
  [5334] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_body_type, 3, 0, 0),
  [5336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4006),
  [5338] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_body_type, 2, 0, 0),
  [5340] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_body_type, 2, 0, 0),
  [5342] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_declaration, 1, 0, 0),
  [5344] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_declaration, 1, 0, 0),
  [5346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5892),
  [5348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3806),
  [5350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3768),
  [5352] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_declaration_name, 1, 0, 0),
  [5354] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_declaration_name, 1, 0, 0),
  [5356] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 171),
  [5358] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 171),
  [5360] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 136),
  [5362] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 136),
  [5364] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 137),
  [5366] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 137),
  [5368] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 172),
  [5370] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 172),
  [5372] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_constraint, 4, 0, 142),
  [5374] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_constraint, 4, 0, 142),
  [5376] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_declaration_name, 2, 0, 0),
  [5378] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_declaration_name, 2, 0, 0),
  [5380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4489),
  [5382] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_open_class_type, 4, 0, 50),
  [5384] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_open_class_type, 4, 0, 50),
  [5386] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 2, 0, 10),
  [5388] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 2, 0, 10),
  [5390] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_class_type, 4, 0, 0),
  [5392] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instantiated_class_type, 4, 0, 0),
  [5394] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_binding_repeat1, 1, 0, 0),
  [5396] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_binding_repeat1, 1, 0, 0),
  [5398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [5400] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 171),
  [5402] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 171),
  [5404] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 193),
  [5406] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 193),
  [5408] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 172),
  [5410] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 172),
  [5412] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0), SHIFT_REPEAT(5878),
  [5415] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_class_type, 5, 0, 0),
  [5417] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instantiated_class_type, 5, 0, 0),
  [5419] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_body_type, 5, 0, 201),
  [5421] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_body_type, 5, 0, 201),
  [5423] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 90),
  [5425] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 90),
  [5427] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 8, 0, 193),
  [5429] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 8, 0, 193),
  [5431] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 136),
  [5433] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 136),
  [5435] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 2, 0, 30),
  [5437] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 2, 0, 30),
  [5439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4835),
  [5441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4849),
  [5443] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 6, 0, 169),
  [5445] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 6, 0, 169),
  [5447] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor_type, 4, 0, 173),
  [5449] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor_type, 4, 0, 173),
  [5451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4463),
  [5453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5896),
  [5455] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor_type, 2, 0, 97),
  [5457] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor_type, 2, 0, 97),
  [5459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3373),
  [5461] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_argument, 2, 0, 0),
  [5463] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_argument, 2, 0, 0),
  [5465] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 3, 0, 51),
  [5467] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 3, 0, 51),
  [5469] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 4, 0, 51),
  [5471] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 4, 0, 51),
  [5473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2686),
  [5475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2828),
  [5477] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__constructor_argument_repeat1, 2, 0, 0), SHIFT_REPEAT(4063),
  [5480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 4, 0, 84),
  [5482] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 4, 0, 84),
  [5484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 4, 0, 85),
  [5486] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 4, 0, 85),
  [5488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 4, 0, 86),
  [5490] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 4, 0, 86),
  [5492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor_type, 3, 0, 141),
  [5494] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor_type, 3, 0, 141),
  [5496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 84),
  [5498] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 84),
  [5500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 7, 0, 169),
  [5502] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 7, 0, 169),
  [5504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor_type, 3, 0, 111),
  [5506] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor_type, 3, 0, 111),
  [5508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 85),
  [5510] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 85),
  [5512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 131),
  [5514] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 131),
  [5516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 86),
  [5518] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 86),
  [5520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 132),
  [5522] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 132),
  [5524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 133),
  [5526] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 133),
  [5528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__module_typed, 2, 0, 65),
  [5530] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__module_typed, 2, 0, 65),
  [5532] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_attribute, 3, 0, 0),
  [5534] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_attribute, 3, 0, 0),
  [5536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_expression_item_repeat1, 1, 0, 0),
  [5538] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_expression_item_repeat1, 1, 0, 0),
  [5540] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 3, 0, 66),
  [5542] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 3, 0, 66),
  [5544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4836),
  [5546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4837),
  [5548] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_attribute, 4, 0, 0),
  [5550] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_attribute, 4, 0, 0),
  [5552] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0),
  [5554] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0),
  [5556] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(6916),
  [5559] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(4860),
  [5562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2699),
  [5564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2853),
  [5566] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 6, 0, 131),
  [5568] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 6, 0, 131),
  [5570] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 6, 0, 132),
  [5572] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 6, 0, 132),
  [5574] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 2, 0, 7),
  [5576] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 2, 0, 7),
  [5578] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 6, 0, 133),
  [5580] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 6, 0, 133),
  [5582] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0), SHIFT_REPEAT(6550),
  [5585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4947),
  [5587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3940),
  [5589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4178),
  [5591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3974),
  [5593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6876),
  [5595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4418),
  [5597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8297),
  [5599] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3641),
  [5601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6256),
  [5603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5435),
  [5605] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(3373),
  [5608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6305),
  [5610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5691),
  [5612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5635),
  [5614] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variant_declaration, 2, 0, 0),
  [5616] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variant_declaration, 2, 0, 0),
  [5618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5339),
  [5620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4032),
  [5622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 6, 0, 139),
  [5624] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 6, 0, 139),
  [5626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4491),
  [5628] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7211),
  [5630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6292),
  [5632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6508),
  [5634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6664),
  [5636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3834),
  [5638] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 5, 0, 96),
  [5640] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 5, 0, 96),
  [5642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variant_declaration, 3, 0, 0),
  [5644] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variant_declaration, 3, 0, 0),
  [5646] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_declaration, 3, 0, 0),
  [5648] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_declaration, 3, 0, 0),
  [5650] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0),
  [5652] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0),
  [5654] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(5339),
  [5657] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 4, 0, 62),
  [5659] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 4, 0, 62),
  [5661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4455),
  [5663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5741),
  [5665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4958),
  [5667] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 3, 0, 29),
  [5669] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 3, 0, 29),
  [5671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4470),
  [5673] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_declaration, 5, 0, 0),
  [5675] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_declaration, 5, 0, 0),
  [5677] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_declaration, 4, 0, 0),
  [5679] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_declaration, 4, 0, 0),
  [5681] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 178),
  [5683] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 178),
  [5685] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 69),
  [5687] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 69),
  [5689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4011),
  [5691] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 10, 0, 227),
  [5693] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 10, 0, 227),
  [5695] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 226),
  [5697] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 226),
  [5699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1757),
  [5701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3426),
  [5703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5704),
  [5705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1985),
  [5707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3955),
  [5709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5950),
  [5711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [5713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6600),
  [5715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6725),
  [5717] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 73),
  [5719] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 73),
  [5721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4013),
  [5723] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 33),
  [5725] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 33),
  [5727] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 73),
  [5729] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 73),
  [5731] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 8, 0, 224),
  [5733] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 8, 0, 224),
  [5735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1714),
  [5737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1462),
  [5739] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_type, 1, 0, 0),
  [5741] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_type, 1, 0, 0),
  [5743] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 3, 0, 60),
  [5745] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 3, 0, 60),
  [5747] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 198),
  [5749] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 198),
  [5751] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(7084),
  [5754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_typed, 2, 0, 72),
  [5756] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_typed, 2, 0, 72),
  [5758] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 3, 0, 30),
  [5760] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 3, 0, 30),
  [5762] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 9, 0, 193),
  [5764] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 9, 0, 193),
  [5766] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 107),
  [5768] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 107),
  [5770] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 151),
  [5772] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 151),
  [5774] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 10, 0, 225),
  [5776] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 10, 0, 225),
  [5778] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 105),
  [5780] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 105),
  [5782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4018),
  [5784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 103),
  [5786] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 103),
  [5788] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 214),
  [5790] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 214),
  [5792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1731),
  [5794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(971),
  [5796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1749),
  [5798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1380),
  [5800] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 92),
  [5802] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 92),
  [5804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1753),
  [5806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1301),
  [5808] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 218),
  [5810] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 218),
  [5812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 70),
  [5814] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 70),
  [5816] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 4, 0, 66),
  [5818] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 4, 0, 66),
  [5820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1763),
  [5822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1165),
  [5824] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 146),
  [5826] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 146),
  [5828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4022),
  [5830] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_floating_attribute, 3, 0, 0),
  [5832] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_floating_attribute, 3, 0, 0),
  [5834] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 26),
  [5836] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 26),
  [5838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1755),
  [5840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2908),
  [5842] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 11, 0, 230),
  [5844] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 11, 0, 230),
  [5846] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 227),
  [5848] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 227),
  [5850] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 200),
  [5852] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 200),
  [5854] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 92),
  [5856] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 92),
  [5858] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 91),
  [5860] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 91),
  [5862] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 69),
  [5864] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 69),
  [5866] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 176),
  [5868] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 176),
  [5870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4025),
  [5872] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 216),
  [5874] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 216),
  [5876] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 177),
  [5878] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 177),
  [5880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4027),
  [5882] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 3, 0, 3),
  [5884] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 3, 0, 3),
  [5886] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 93),
  [5888] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 93),
  [5890] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 93),
  [5892] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 93),
  [5894] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 3, 0, 102),
  [5896] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 3, 0, 102),
  [5898] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 138),
  [5900] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 138),
  [5902] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 216),
  [5904] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 216),
  [5906] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 196),
  [5908] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 196),
  [5910] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 136),
  [5912] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 136),
  [5914] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 137),
  [5916] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 137),
  [5918] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 70),
  [5920] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 70),
  [5922] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 199),
  [5924] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 199),
  [5926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4031),
  [5928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1759),
  [5930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1075),
  [5932] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 200),
  [5934] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 200),
  [5936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4034),
  [5938] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 10, 0, 230),
  [5940] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 10, 0, 230),
  [5942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6835),
  [5944] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 60),
  [5946] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 60),
  [5948] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 5, 0, 98),
  [5950] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 5, 0, 98),
  [5952] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 8, 0, 213),
  [5954] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 8, 0, 213),
  [5956] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 218),
  [5958] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 218),
  [5960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4036),
  [5962] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 5, 0, 99),
  [5964] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 5, 0, 99),
  [5966] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_function_type, 3, 0, 111),
  [5968] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_function_type, 3, 0, 111),
  [5970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1768),
  [5972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3045),
  [5974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1747),
  [5976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2555),
  [5978] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 6, 0, 195),
  [5980] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 6, 0, 195),
  [5982] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 175),
  [5984] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 175),
  [5986] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 104),
  [5988] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 104),
  [5990] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 196),
  [5992] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 196),
  [5994] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 197),
  [5996] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 197),
  [5998] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 148),
  [6000] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 148),
  [6002] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 176),
  [6004] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 176),
  [6006] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 198),
  [6008] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 198),
  [6010] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 104),
  [6012] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 104),
  [6014] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 4, 0, 63),
  [6016] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 4, 0, 63),
  [6018] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 177),
  [6020] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 177),
  [6022] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 3, 0, 10),
  [6024] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 3, 0, 10),
  [6026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 147),
  [6028] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 147),
  [6030] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 197),
  [6032] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 197),
  [6034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1743),
  [6036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2200),
  [6038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 149),
  [6040] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 149),
  [6042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 179),
  [6044] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 179),
  [6046] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 199),
  [6048] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 199),
  [6050] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_function_type, 5, 0, 181),
  [6052] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_function_type, 5, 0, 181),
  [6054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7084),
  [6056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 4, 0, 145),
  [6058] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 4, 0, 145),
  [6060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 6, 0, 143),
  [6062] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 6, 0, 143),
  [6064] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 217),
  [6066] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 217),
  [6068] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 105),
  [6070] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 105),
  [6072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1765),
  [6074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2487),
  [6076] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 5, 0, 145),
  [6078] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 5, 0, 145),
  [6080] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 4, 0, 102),
  [6082] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 4, 0, 102),
  [6084] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 58),
  [6086] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 58),
  [6088] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 215),
  [6090] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 215),
  [6092] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 225),
  [6094] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 225),
  [6096] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 106),
  [6098] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 106),
  [6100] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 148),
  [6102] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 148),
  [6104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 149),
  [6106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 149),
  [6108] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 215),
  [6110] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 215),
  [6112] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 179),
  [6114] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 179),
  [6116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1751),
  [6118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2307),
  [6120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1761),
  [6122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3256),
  [6124] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 14),
  [6126] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 14),
  [6128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 2, 0, 34),
  [6130] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 2, 0, 34),
  [6132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3996),
  [6134] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 175),
  [6136] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 175),
  [6138] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 8, 0, 171),
  [6140] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 8, 0, 171),
  [6142] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_floating_attribute, 4, 0, 0),
  [6144] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_floating_attribute, 4, 0, 0),
  [6146] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 106),
  [6148] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 106),
  [6150] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 9, 0, 224),
  [6152] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 9, 0, 224),
  [6154] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 8, 0, 172),
  [6156] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 8, 0, 172),
  [6158] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 151),
  [6160] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 151),
  [6162] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 107),
  [6164] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 107),
  [6166] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 34),
  [6168] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 34),
  [6170] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 146),
  [6172] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 146),
  [6174] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 90),
  [6176] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 90),
  [6178] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 138),
  [6180] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 138),
  [6182] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 214),
  [6184] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 214),
  [6186] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 7, 0, 195),
  [6188] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 7, 0, 195),
  [6190] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 7, 0, 213),
  [6192] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 7, 0, 213),
  [6194] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 10, 0, 226),
  [6196] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 10, 0, 226),
  [6198] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 217),
  [6200] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 217),
  [6202] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_extension, 5, 0, 0),
  [6204] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_extension, 5, 0, 0),
  [6206] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_item_extension, 5, 0, 0),
  [6208] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_item_extension, 5, 0, 0),
  [6210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5493),
  [6212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4173),
  [6214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4457),
  [6216] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7969),
  [6218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3723),
  [6220] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6168),
  [6222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5305),
  [6224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4366),
  [6226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8020),
  [6228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6130),
  [6230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6580),
  [6232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6713),
  [6234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4004),
  [6236] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external, 6, 0, 35),
  [6238] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external, 6, 0, 35),
  [6240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3971),
  [6242] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_definition, 3, 0, 0),
  [6244] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_definition, 3, 0, 0),
  [6246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4101),
  [6248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3846),
  [6250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3765),
  [6252] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_definition, 3, 0, 0),
  [6254] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_definition, 3, 0, 0),
  [6256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1809),
  [6258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1809),
  [6260] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_definition, 4, 0, 0),
  [6262] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_definition, 4, 0, 0),
  [6264] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0),
  [6266] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0),
  [6268] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1809),
  [6271] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1809),
  [6274] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_definition, 5, 0, 0),
  [6276] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_definition, 5, 0, 0),
  [6278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5742),
  [6280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_specification, 3, 0, 35),
  [6282] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_specification, 3, 0, 35),
  [6284] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_specification, 5, 0, 74),
  [6286] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_specification, 5, 0, 74),
  [6288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 6, 0, 96),
  [6290] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 6, 0, 96),
  [6292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 1, 0, 0),
  [6294] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_external_repeat1, 1, 0, 0),
  [6296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4477),
  [6298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6137),
  [6300] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_declaration, 2, 0, 0),
  [6302] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external_declaration, 2, 0, 0),
  [6304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5777),
  [6306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4055),
  [6308] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_item_extension, 6, 0, 0),
  [6310] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_item_extension, 6, 0, 0),
  [6312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_item, 2, 0, 0),
  [6314] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_item, 2, 0, 0),
  [6316] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external, 7, 0, 74),
  [6318] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external, 7, 0, 74),
  [6320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 7, 0, 139),
  [6322] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 7, 0, 139),
  [6324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4104),
  [6326] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 4, 0, 31),
  [6328] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 4, 0, 31),
  [6330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4105),
  [6332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_extension, 4, 0, 0),
  [6334] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_extension, 4, 0, 0),
  [6336] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_definition, 4, 0, 0),
  [6338] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_definition, 4, 0, 0),
  [6340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5936),
  [6342] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0), SHIFT_REPEAT(5936),
  [6345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5695),
  [6347] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_item_extension, 4, 0, 0),
  [6349] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_item_extension, 4, 0, 0),
  [6351] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 3, 0, 13),
  [6353] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 3, 0, 13),
  [6355] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module, 3, 0, 13),
  [6357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_include_module, 3, 0, 13),
  [6359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4064),
  [6361] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 5, 0, 62),
  [6363] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 5, 0, 62),
  [6365] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_extension, 3, 0, 0),
  [6367] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_extension, 3, 0, 0),
  [6369] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_definition, 2, 0, 0),
  [6371] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_definition, 2, 0, 0),
  [6373] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module, 4, 0, 31),
  [6375] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_include_module, 4, 0, 31),
  [6377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4074),
  [6379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4114),
  [6381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4014),
  [6383] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_item, 1, 0, 0),
  [6385] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_item, 1, 0, 0),
  [6387] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 4, 0, 29),
  [6389] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 4, 0, 29),
  [6391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 5, 0, 67),
  [6393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 5, 0, 67),
  [6395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_specification, 4, 0, 74),
  [6397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_specification, 4, 0, 74),
  [6399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_specification, 4, 0, 35),
  [6401] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_specification, 4, 0, 35),
  [6403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_definition, 2, 0, 0),
  [6405] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_definition, 2, 0, 0),
  [6407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4483),
  [6409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6275),
  [6411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1813),
  [6413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1813),
  [6415] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_definition, 3, 0, 0),
  [6417] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_definition, 3, 0, 0),
  [6419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6356),
  [6421] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_definition, 4, 0, 0),
  [6423] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_definition, 4, 0, 0),
  [6425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1752),
  [6427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2407),
  [6429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1764),
  [6431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1194),
  [6433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1762),
  [6435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3305),
  [6437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1825),
  [6439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1825),
  [6441] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 4, 0, 0),
  [6443] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 4, 0, 0),
  [6445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4918),
  [6447] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0),
  [6449] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0),
  [6451] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6356),
  [6454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4792),
  [6456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4564),
  [6458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6335),
  [6460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [6462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5200),
  [6464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1748),
  [6466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2568),
  [6468] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_definition, 4, 0, 0),
  [6470] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_definition, 4, 0, 0),
  [6472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6148),
  [6474] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__constructor_argument_repeat1, 2, 0, 0), SHIFT_REPEAT(4006),
  [6477] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0),
  [6479] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0),
  [6481] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4918),
  [6484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1754),
  [6486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1313),
  [6488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1760),
  [6490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1077),
  [6492] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1813),
  [6495] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1813),
  [6498] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_definition, 5, 0, 0),
  [6500] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_definition, 5, 0, 0),
  [6502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1750),
  [6504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1401),
  [6506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1712),
  [6508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1429),
  [6510] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_definition, 2, 0, 0),
  [6512] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_definition, 2, 0, 0),
  [6514] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1825),
  [6517] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1825),
  [6520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 5, 0, 0),
  [6522] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 5, 0, 0),
  [6524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_type_definition_repeat1, 2, 0, 0),
  [6526] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_type_definition_repeat1, 2, 0, 0),
  [6528] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6148),
  [6531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1756),
  [6533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2921),
  [6535] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_definition, 5, 0, 0),
  [6537] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_definition, 5, 0, 0),
  [6539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3378),
  [6541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3864),
  [6543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1767),
  [6545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2489),
  [6547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1758),
  [6549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1999),
  [6551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3867),
  [6553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4485),
  [6555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5996),
  [6557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1734),
  [6559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(947),
  [6561] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 3, 0, 0),
  [6563] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 3, 0, 0),
  [6565] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_type, 3, 0, 130),
  [6567] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_type, 3, 0, 130),
  [6569] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_type, 3, 0, 114),
  [6571] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_type, 3, 0, 114),
  [6573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1769),
  [6575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2930),
  [6577] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(6835),
  [6580] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(4883),
  [6583] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 2, 0, 0),
  [6585] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 2, 0, 0),
  [6587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4787),
  [6589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4612),
  [6591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6475),
  [6593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(42),
  [6595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1745),
  [6597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2264),
  [6599] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_definition, 3, 0, 0),
  [6601] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_definition, 3, 0, 0),
  [6603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1261),
  [6605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4490),
  [6607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(974),
  [6609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(975),
  [6611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(79),
  [6613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1854),
  [6615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2520),
  [6617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6742),
  [6619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2491),
  [6621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(976),
  [6623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2494),
  [6625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2592),
  [6627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2495),
  [6629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2496),
  [6631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2497),
  [6633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2498),
  [6635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2499),
  [6637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2500),
  [6639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2501),
  [6641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [6643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1856),
  [6645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2980),
  [6647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2954),
  [6649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2963),
  [6651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2976),
  [6653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2977),
  [6655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2978),
  [6657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2982),
  [6659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2983),
  [6661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2984),
  [6663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2989),
  [6665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1485),
  [6667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2599),
  [6669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2600),
  [6671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(981),
  [6673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2601),
  [6675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2606),
  [6677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2607),
  [6679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2608),
  [6681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(982),
  [6683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(983),
  [6685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2609),
  [6687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1420),
  [6689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1421),
  [6691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5297),
  [6693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [6695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1833),
  [6697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2191),
  [6699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [6701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1834),
  [6703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1111),
  [6705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1079),
  [6707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(984),
  [6709] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1826),
  [6711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1826),
  [6713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [6715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1847),
  [6717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2160),
  [6719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3753),
  [6721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(680),
  [6723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1444),
  [6725] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(3426),
  [6728] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5704),
  [6731] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0),
  [6733] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(3955),
  [6736] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5950),
  [6739] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(265),
  [6742] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(6390),
  [6745] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(6600),
  [6748] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(6725),
  [6751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1081),
  [6753] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 3, 0, 0),
  [6755] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 3, 0, 0),
  [6757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6041),
  [6759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1082),
  [6761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3387),
  [6763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3878),
  [6765] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 2, 0, 0),
  [6767] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 2, 0, 0),
  [6769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [6771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1827),
  [6773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1003),
  [6775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2354),
  [6777] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1826),
  [6780] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1826),
  [6783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1043),
  [6785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1083),
  [6787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1084),
  [6789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1085),
  [6791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1086),
  [6793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1087),
  [6795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5349),
  [6797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4102),
  [6799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [6801] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0),
  [6803] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0),
  [6805] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6041),
  [6808] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 4, 0, 0),
  [6810] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 4, 0, 0),
  [6812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1397),
  [6814] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module_type, 3, 0, 125),
  [6816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(71),
  [6818] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1857),
  [6820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1983),
  [6822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1956),
  [6824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4119),
  [6826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4121),
  [6828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(75),
  [6830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1835),
  [6832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3342),
  [6834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4108),
  [6836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3331),
  [6838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3862),
  [6840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1329),
  [6842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3989),
  [6844] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(5349),
  [6847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1978),
  [6849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4009),
  [6851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4073),
  [6853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5234),
  [6855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(764),
  [6857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(940),
  [6859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4112),
  [6861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1340),
  [6863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1342),
  [6865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4116),
  [6867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1343),
  [6869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4118),
  [6871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2395),
  [6873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1350),
  [6875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1351),
  [6877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1357),
  [6879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1358),
  [6881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(959),
  [6883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3915),
  [6885] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module_type, 2, 0, 65),
  [6887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3340),
  [6889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [6891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1846),
  [6893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1411),
  [6895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3257),
  [6897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3258),
  [6899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3259),
  [6901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [6903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1843),
  [6905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1461),
  [6907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3261),
  [6909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3262),
  [6911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3263),
  [6913] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_toplevel_directive, 2, 0, 0),
  [6915] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_toplevel_directive, 2, 0, 0),
  [6917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5840),
  [6919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3269),
  [6921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1427),
  [6923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2325),
  [6925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2412),
  [6927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2189),
  [6929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2192),
  [6931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2194),
  [6933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2206),
  [6935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2207),
  [6937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2212),
  [6939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2219),
  [6941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2224),
  [6943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2225),
  [6945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2415),
  [6947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3376),
  [6949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3874),
  [6951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1434),
  [6953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2429),
  [6955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2583),
  [6957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2430),
  [6959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2434),
  [6961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [6963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1844),
  [6965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2242),
  [6967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1246),
  [6969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2436),
  [6971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1256),
  [6973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2915),
  [6975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1262),
  [6977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1263),
  [6979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1269),
  [6981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1270),
  [6983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1271),
  [6985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1275),
  [6987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [6989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1848),
  [6991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1167),
  [6993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1988),
  [6995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1202),
  [6997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1989),
  [6999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1460),
  [7001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1480),
  [7003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1990),
  [7005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1218),
  [7007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1482),
  [7009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1993),
  [7011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1220),
  [7013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [7015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1853),
  [7017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1250),
  [7019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1224),
  [7021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1227),
  [7023] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4066),
  [7025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1995),
  [7027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1228),
  [7029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1231),
  [7031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1996),
  [7033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1232),
  [7035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2897),
  [7037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1181),
  [7039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2000),
  [7041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2907),
  [7043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2909),
  [7045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4484),
  [7047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2889),
  [7049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2911),
  [7051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2912),
  [7053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2913),
  [7055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2914),
  [7057] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure_item, 1, 0, 0),
  [7059] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure_item, 1, 0, 0),
  [7061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3389),
  [7063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5958),
  [7065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4041),
  [7067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4927),
  [7069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6188),
  [7071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6299),
  [7073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3382),
  [7075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6245),
  [7077] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__local_structure_item, 1, 0, 0),
  [7079] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__local_structure_item, 1, 0, 0),
  [7081] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5198),
  [7083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5367),
  [7085] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4920),
  [7087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5947),
  [7089] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6008),
  [7091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6007),
  [7093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4143),
  [7095] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4941),
  [7097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(579),
  [7099] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3372),
  [7101] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 4, 0, 0),
  [7103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3371),
  [7105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [7107] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 2, 0, 0),
  [7109] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6299),
  [7112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(327),
  [7114] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0),
  [7116] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3815),
  [7118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4921),
  [7120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(520),
  [7122] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 3, 0, 0),
  [7124] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 1, 0, 0),
  [7126] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__structure_repeat2, 1, 0, 0),
  [7128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6651),
  [7130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6101),
  [7132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6192),
  [7134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(492),
  [7136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4111),
  [7138] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(3426),
  [7141] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(5958),
  [7144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0),
  [7146] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(4927),
  [7149] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(6188),
  [7152] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(6390),
  [7155] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(6600),
  [7158] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(6725),
  [7161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3975),
  [7163] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module_type, 4, 0, 125),
  [7165] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module_type, 3, 0, 65),
  [7167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4007),
  [7169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6133),
  [7171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4922),
  [7173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4368),
  [7175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5032),
  [7177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5032),
  [7179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5001),
  [7181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2526),
  [7183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4919),
  [7185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2985),
  [7187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4674),
  [7189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5457),
  [7191] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4919),
  [7194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6128),
  [7196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1341),
  [7198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1973),
  [7200] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3051),
  [7202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7136),
  [7204] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4468),
  [7206] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2788),
  [7208] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2619),
  [7210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1237),
  [7212] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3318),
  [7214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4402),
  [7216] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6128),
  [7219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4917),
  [7221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1140),
  [7223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3572),
  [7225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(992),
  [7227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2346),
  [7229] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2193),
  [7231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2178),
  [7233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1272),
  [7235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4924),
  [7237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6754),
  [7239] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6754),
  [7242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1115),
  [7244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6230),
  [7246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3381),
  [7248] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_content, 1, 0, 0),
  [7250] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__instance_variable_name, 1, 0, 2),
  [7252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4482),
  [7254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3386),
  [7256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3652),
  [7258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3413),
  [7260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5413),
  [7262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6885),
  [7264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2173),
  [7266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6973),
  [7268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7124),
  [7270] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat1, 2, 0, 0), SHIFT_REPEAT(4970),
  [7273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4812),
  [7275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6841),
  [7277] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0),
  [7279] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(5032),
  [7282] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(5032),
  [7285] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(5001),
  [7288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3545),
  [7290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6903),
  [7292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signature_item, 1, 0, 0),
  [7294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 1, 0, 0),
  [7296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7167),
  [7298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5023),
  [7300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4492),
  [7302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6408),
  [7304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6811),
  [7306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5011),
  [7308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5370),
  [7310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7453),
  [7312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5344),
  [7314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6364),
  [7316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5345),
  [7318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5031),
  [7320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4127),
  [7322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_constrained, 2, 0, 56),
  [7324] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_constrained, 2, 0, 56),
  [7326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4980),
  [7328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4425),
  [7330] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_constrained, 1, 0, 25),
  [7332] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_constrained, 1, 0, 25),
  [7334] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__typed, 2, 0, 52),
  [7336] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__typed, 2, 0, 52),
  [7338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4985),
  [7340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5352),
  [7342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5028),
  [7344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4049),
  [7346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1604),
  [7348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6012),
  [7350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6534),
  [7352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3821),
  [7354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1592),
  [7356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5856),
  [7358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6690),
  [7360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3524),
  [7362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [7364] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 120),
  [7366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1612),
  [7368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5939),
  [7370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6402),
  [7372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2072),
  [7374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [7376] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 160),
  [7378] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 2, 0, 36),
  [7380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6066),
  [7382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4429),
  [7384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(527),
  [7386] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 3, 0, 75),
  [7388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1593),
  [7390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5870),
  [7392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(584),
  [7394] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 117),
  [7396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5378),
  [7398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1628),
  [7400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5946),
  [7402] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_constrained, 1, 0, 24),
  [7404] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_constrained, 1, 0, 24),
  [7406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1600),
  [7408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5519),
  [7410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1597),
  [7412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1584),
  [7414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2401),
  [7416] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_constraint, 4, 0, 112),
  [7418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5135),
  [7420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [7422] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 154),
  [7424] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_escape_sequence, 1, 0, 0),
  [7426] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_escape_sequence, 1, 0, 0),
  [7428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1630),
  [7430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5971),
  [7432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6324),
  [7434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2910),
  [7436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(574),
  [7438] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 4, 0, 115),
  [7440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(541),
  [7442] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 3, 0, 79),
  [7444] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 3, 0, 76),
  [7446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6146),
  [7448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5348),
  [7450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1595),
  [7452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5890),
  [7454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6565),
  [7456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3363),
  [7458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(432),
  [7460] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 188),
  [7462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__coerced, 2, 0, 53),
  [7464] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__coerced, 2, 0, 53),
  [7466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(358),
  [7468] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 159),
  [7470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1582),
  [7472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5872),
  [7474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6441),
  [7476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1601),
  [7478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4434),
  [7480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4932),
  [7482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1585),
  [7484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3125),
  [7486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4091),
  [7488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4046),
  [7490] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 1, 0, 0),
  [7492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_string_content_repeat1, 1, 0, 0),
  [7494] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_pattern, 2, 0, 22),
  [7496] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_pattern, 2, 0, 22),
  [7498] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_pattern, 2, 0, 22),
  [7500] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_pattern, 2, 0, 22),
  [7502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2580),
  [7504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4408),
  [7506] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_constraint, 5, 0, 112),
  [7508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 87),
  [7510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2589),
  [7512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 115),
  [7514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_pattern, 2, 0, 22),
  [7516] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tag_pattern, 2, 0, 22),
  [7518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 0),
  [7520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 50),
  [7522] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 155),
  [7524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 5, 0, 157),
  [7526] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 5, 0, 116),
  [7528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 158),
  [7530] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 3, 0, 36),
  [7532] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 117),
  [7534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 78),
  [7536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 119),
  [7538] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 161),
  [7540] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 162),
  [7542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1413),
  [7544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 120),
  [7546] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_initializer, 3, 0, 38),
  [7548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1405),
  [7550] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_pattern, 5, 0, 126),
  [7552] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_pattern, 5, 0, 126),
  [7554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2298),
  [7556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2315),
  [7558] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_pattern, 3, 0, 49),
  [7560] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_pattern, 3, 0, 49),
  [7562] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_pattern, 3, 0, 0),
  [7564] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_pattern, 3, 0, 0),
  [7566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1243),
  [7568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1253),
  [7570] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__or_pattern_anonymous, 3, 0, 0),
  [7572] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__or_pattern_anonymous, 3, 0, 0),
  [7574] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alias_pattern, 3, 0, 54),
  [7576] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alias_pattern, 3, 0, 54),
  [7578] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_pattern, 3, 0, 49),
  [7580] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_pattern, 3, 0, 49),
  [7582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2006),
  [7584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1972),
  [7586] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 87),
  [7588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 183),
  [7590] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 184),
  [7592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 154),
  [7594] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 155),
  [7596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 6, 0, 157),
  [7598] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 158),
  [7600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 185),
  [7602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 186),
  [7604] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 159),
  [7606] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 187),
  [7608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1060),
  [7610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 160),
  [7612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__effect_pattern, 1, 0, 0),
  [7614] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__effect_pattern, 1, 0, 0),
  [7616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1066),
  [7618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 118),
  [7620] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 161),
  [7622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 162),
  [7624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 189),
  [7626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1437),
  [7628] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern, 1, 0, 0),
  [7630] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pattern, 1, 0, 0),
  [7632] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern, 1, 0, 16),
  [7634] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pattern, 1, 0, 16),
  [7636] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern, 1, 0, 17),
  [7638] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pattern, 1, 0, 17),
  [7640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3281),
  [7642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3279),
  [7644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1223),
  [7646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1180),
  [7648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 7, 0, 183),
  [7650] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 7, 0, 184),
  [7652] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 7, 0, 204),
  [7654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2391),
  [7656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 186),
  [7658] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 205),
  [7660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2467),
  [7662] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 187),
  [7664] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 206),
  [7666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 207),
  [7668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2473),
  [7670] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 188),
  [7672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 189),
  [7674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3001),
  [7676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3010),
  [7678] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 8, 0, 204),
  [7680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 8, 0, 205),
  [7682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 8, 0, 206),
  [7684] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 8, 0, 207),
  [7686] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 8, 0, 220),
  [7688] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 9, 0, 220),
  [7690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4832),
  [7692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6468),
  [7694] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4446),
  [7696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3733),
  [7698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1476),
  [7700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4855),
  [7702] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 4, 0, 50),
  [7704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 4, 0, 75),
  [7706] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 4, 0, 76),
  [7708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 4, 0, 116),
  [7710] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 119),
  [7712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 79),
  [7714] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 37),
  [7716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(955),
  [7718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5232),
  [7720] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_pattern, 4, 0, 82),
  [7722] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_pattern, 4, 0, 82),
  [7724] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_effect_pattern, 4, 0, 122),
  [7726] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_effect_pattern, 4, 0, 122),
  [7728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(964),
  [7730] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_initializer, 2, 0, 38),
  [7732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2342),
  [7734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 185),
  [7736] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_specification, 2, 0, 72),
  [7738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_specification, 4, 0, 35),
  [7740] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_specification, 4, 0, 74),
  [7742] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_specification, 4, 0, 79),
  [7744] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_specification, 4, 0, 117),
  [7746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3431),
  [7748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6770),
  [7750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2661),
  [7752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2839),
  [7754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5225),
  [7756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5252),
  [7758] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_specification, 3, 0, 35),
  [7760] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_specification, 3, 0, 72),
  [7762] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_specification, 3, 0, 79),
  [7764] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_specification, 5, 0, 74),
  [7766] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_specification, 5, 0, 117),
  [7768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5245),
  [7770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5240),
  [7772] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_pattern, 2, 0, 8),
  [7774] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2856),
  [7776] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_pattern, 2, 0, 8),
  [7778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2673),
  [7780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2768),
  [7782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5211),
  [7784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7951),
  [7786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5238),
  [7788] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cons_pattern, 3, 0, 46),
  [7790] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cons_pattern, 3, 0, 46),
  [7792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1711),
  [7794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [7796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2949),
  [7798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1678),
  [7800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1770),
  [7802] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3110),
  [7804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6285),
  [7806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [7808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2850),
  [7810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4950),
  [7812] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(3431),
  [7815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4166),
  [7817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3330),
  [7819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3849),
  [7821] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_pattern, 3, 0, 49),
  [7823] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_pattern, 3, 0, 49),
  [7825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3583),
  [7827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3309),
  [7829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3355),
  [7831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3930),
  [7833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3271),
  [7835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4966),
  [7837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_pattern, 2, 0, 22),
  [7839] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_pattern, 2, 0, 22),
  [7841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3950),
  [7843] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_field, 1, 0, 0),
  [7845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2563),
  [7847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [7849] = {.entry = {.count = 1, .reusable = false}}, SHIFT(471),
  [7851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [7853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6467),
  [7855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8152),
  [7857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8152),
  [7859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8194),
  [7861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7001),
  [7863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8048),
  [7865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1306),
  [7867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(479),
  [7869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(480),
  [7871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(481),
  [7873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2382),
  [7875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(458),
  [7877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(459),
  [7879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(460),
  [7881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3427),
  [7883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3062),
  [7885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(535),
  [7887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(536),
  [7889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(537),
  [7891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1041),
  [7893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(507),
  [7895] = {.entry = {.count = 1, .reusable = false}}, SHIFT(508),
  [7897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(509),
  [7899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1127),
  [7901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(521),
  [7903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(522),
  [7905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(523),
  [7907] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5401),
  [7909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5401),
  [7911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8018),
  [7913] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6468),
  [7916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2243),
  [7918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(486),
  [7920] = {.entry = {.count = 1, .reusable = false}}, SHIFT(487),
  [7922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(488),
  [7924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1945),
  [7926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(500),
  [7928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(501),
  [7930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(502),
  [7932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1324),
  [7934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [7936] = {.entry = {.count = 1, .reusable = false}}, SHIFT(344),
  [7938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [7940] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 1, 0, 0),
  [7942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3287),
  [7944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(514),
  [7946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(515),
  [7948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(516),
  [7950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5204),
  [7952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(930),
  [7954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [7956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(426),
  [7958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(427),
  [7960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1214),
  [7962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(493),
  [7964] = {.entry = {.count = 1, .reusable = false}}, SHIFT(494),
  [7966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(495),
  [7968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2488),
  [7970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(528),
  [7972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(529),
  [7974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(530),
  [7976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6008),
  [7978] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_string_content, 1, 0, 0),
  [7980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(678),
  [7982] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3370),
  [7984] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 1, 0, 0),
  [7986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3368),
  [7988] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(3427),
  [7991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5838),
  [7993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3836),
  [7995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3792),
  [7997] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 1, 0, 0),
  [7999] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_field_specification, 1, 0, 0),
  [8001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5921),
  [8003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3812),
  [8005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3623),
  [8007] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_quoted_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(5401),
  [8010] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_quoted_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(5401),
  [8013] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_quoted_string_content_repeat1, 2, 0, 0),
  [8015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2860),
  [8017] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2753),
  [8019] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3186),
  [8021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3841),
  [8023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3587),
  [8025] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0), SHIFT_REPEAT(5921),
  [8028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3804),
  [8030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3898),
  [8032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4707),
  [8034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3385),
  [8036] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_quoted_string_content_repeat1, 1, 0, 0),
  [8038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_quoted_string_content_repeat1, 1, 0, 0),
  [8040] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument_type, 3, 0, 110),
  [8042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_payload, 2, 0, 0),
  [8044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(585),
  [8046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4853),
  [8048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4854),
  [8050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [8052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [8054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5204),
  [8056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6793),
  [8058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3719),
  [8060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6443),
  [8062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6197),
  [8064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5130),
  [8066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1837),
  [8068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5770),
  [8070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5384),
  [8072] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6813),
  [8074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [8076] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(6903),
  [8079] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(4892),
  [8082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1661),
  [8084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1693),
  [8086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1663),
  [8088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6021),
  [8090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1694),
  [8092] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_pattern_content, 1, 0, 0),
  [8094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4565),
  [8096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [8098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2802),
  [8100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3202),
  [8102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3166),
  [8104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5261),
  [8106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1645),
  [8108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1697),
  [8110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6225),
  [8112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1701),
  [8114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [8116] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6840),
  [8118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5730),
  [8120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4741),
  [8122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7088),
  [8124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2441),
  [8126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4682),
  [8128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4923),
  [8130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6863),
  [8132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3756),
  [8134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7116),
  [8136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2970),
  [8138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4715),
  [8140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6421),
  [8142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6059),
  [8144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3775),
  [8146] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7141),
  [8148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2849),
  [8150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5649),
  [8152] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6934),
  [8154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5321),
  [8156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4567),
  [8158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4595),
  [8160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3189),
  [8162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4667),
  [8164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4680),
  [8166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6974),
  [8168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5173),
  [8170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [8172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4710),
  [8174] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6887),
  [8176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5768),
  [8178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4728),
  [8180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4731),
  [8182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1650),
  [8184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1699),
  [8186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6072),
  [8188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1700),
  [8190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4763),
  [8192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4768),
  [8194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6980),
  [8196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5276),
  [8198] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_binding_pattern_content, 1, 0, 0),
  [8200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1643),
  [8202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4671),
  [8204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [8206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4742),
  [8208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3656),
  [8210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4624),
  [8212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4640),
  [8214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3185),
  [8216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4690),
  [8218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4694),
  [8220] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__constructor_argument_repeat1, 2, 0, 0), SHIFT_REPEAT(4075),
  [8223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4744),
  [8225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4752),
  [8227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7036),
  [8229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5035),
  [8231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4566),
  [8233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4570),
  [8235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4581),
  [8237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4585),
  [8239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1113),
  [8241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5620),
  [8243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5381),
  [8245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6969),
  [8247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(147),
  [8249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1040),
  [8251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5396),
  [8253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7028),
  [8255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6905),
  [8257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5603),
  [8259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4405),
  [8261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5894),
  [8263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4807),
  [8265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4856),
  [8267] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument_type, 4, 0, 153),
  [8269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1668),
  [8271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1737),
  [8273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1738),
  [8275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7034),
  [8277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2218),
  [8279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2837),
  [8281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4675),
  [8283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1662),
  [8285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3195),
  [8287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2790),
  [8289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1656),
  [8291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1722),
  [8293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1680),
  [8295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6173),
  [8297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1723),
  [8299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [8301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7000),
  [8303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5703),
  [8305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1193),
  [8307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5338),
  [8309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7147),
  [8311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [8313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [8315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5581),
  [8317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5857),
  [8319] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_labeled_tuple_element_type, 3, 0, 110), SHIFT(3874),
  [8322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(150),
  [8324] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6861),
  [8326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2798),
  [8328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5166),
  [8330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4717),
  [8332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2295),
  [8334] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6832),
  [8336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2961),
  [8338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6756),
  [8340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6303),
  [8342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4058),
  [8344] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_type_repeat1, 2, 0, 0),
  [8346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2605),
  [8348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2481),
  [8350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1051),
  [8352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1479),
  [8354] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0), SHIFT_REPEAT(6499),
  [8357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2410),
  [8359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1245),
  [8361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2485),
  [8363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2475),
  [8365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(957),
  [8367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2394),
  [8369] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__sequence_pattern_content_repeat1, 2, 0, 0),
  [8371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1339),
  [8373] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 3, 0, 49),
  [8375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1425),
  [8377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1062),
  [8379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2598),
  [8381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2283),
  [8383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3588),
  [8385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1068),
  [8387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(966),
  [8389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3048),
  [8391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1071),
  [8393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1445),
  [8395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5953),
  [8397] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0), SHIFT_REPEAT(5953),
  [8400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1072),
  [8402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1439),
  [8404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1255),
  [8406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2940),
  [8408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1349),
  [8410] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7685),
  [8412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5361),
  [8414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6782),
  [8416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1260),
  [8418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3003),
  [8420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2559),
  [8422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3268),
  [8424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2423),
  [8426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3302),
  [8428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5393),
  [8430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(973),
  [8432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4396),
  [8434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3295),
  [8436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4053),
  [8438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2301),
  [8440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4059),
  [8442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3282),
  [8444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3255),
  [8446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1268),
  [8448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3316),
  [8450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1188),
  [8452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2565),
  [8454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(980),
  [8456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4452),
  [8458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2323),
  [8460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1215),
  [8462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2356),
  [8464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4401),
  [8466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1451),
  [8468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1957),
  [8470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2308),
  [8472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__sequence_binding_pattern_content_repeat1, 2, 0, 0),
  [8474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3012),
  [8476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_typed, 2, 0, 52),
  [8478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1285),
  [8480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_binding_pattern, 4, 0, 129),
  [8482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2582),
  [8484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3019),
  [8486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4037),
  [8488] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__type_param_repeat1, 2, 0, 0), SHIFT_REPEAT(5947),
  [8491] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_param_repeat1, 2, 0, 0),
  [8493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4365),
  [8495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1991),
  [8497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1150),
  [8499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1309),
  [8501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4447),
  [8503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1402),
  [8505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4093),
  [8507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4449),
  [8509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4099),
  [8511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1144),
  [8513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3028),
  [8515] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_binding_pattern, 3, 0, 49),
  [8517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1200),
  [8519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4040),
  [8521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4911),
  [8523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5863),
  [8525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6511),
  [8527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1149),
  [8529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2411),
  [8531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2013),
  [8533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1456),
  [8535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1975),
  [8537] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_specification, 4, 0, 0),
  [8539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3394),
  [8541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1998),
  [8543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4384),
  [8545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4386),
  [8547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4082),
  [8549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2012),
  [8551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(978),
  [8553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2445),
  [8555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2457),
  [8557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1385),
  [8559] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 4, 0, 129),
  [8561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1076),
  [8563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4062),
  [8565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2204),
  [8567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1395),
  [8569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2258),
  [8571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3994),
  [8573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(943),
  [8575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4021),
  [8577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7704),
  [8579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5397),
  [8581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7037),
  [8583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2591),
  [8585] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(5393),
  [8588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4448),
  [8590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2469),
  [8592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2349),
  [8594] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_specification, 3, 0, 0),
  [8596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1409),
  [8598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2250),
  [8600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6916),
  [8602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4911),
  [8604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7891),
  [8606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1203),
  [8608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6182),
  [8610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7616),
  [8612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(994),
  [8614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8066),
  [8616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2455),
  [8618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_specification, 5, 0, 0),
  [8620] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_tag_specification_repeat1, 2, 0, 0),
  [8622] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tag_specification_repeat1, 2, 0, 0), SHIFT_REPEAT(3394),
  [8625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7982),
  [8627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4163),
  [8629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3602),
  [8631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7593),
  [8633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2310),
  [8635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7489),
  [8637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1435),
  [8639] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0), SHIFT_REPEAT(6461),
  [8642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6835),
  [8644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8347),
  [8646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2925),
  [8648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_param, 2, 0, 27),
  [8650] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_param, 2, 0, 0),
  [8652] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_param, 1, 0, 0),
  [8654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7242),
  [8656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1304),
  [8658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4952),
  [8660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6675),
  [8662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7826),
  [8664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4967),
  [8666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6389),
  [8668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3408),
  [8670] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tag_spec, 1, 0, 0),
  [8672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6125),
  [8674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1702),
  [8676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3370),
  [8678] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_binding_pattern, 1, 0, 0),
  [8680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6812),
  [8682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7654),
  [8684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1954),
  [8686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6019),
  [8688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4961),
  [8690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6757),
  [8692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(414),
  [8694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4948),
  [8696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6557),
  [8698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5332),
  [8700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6063),
  [8702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7308),
  [8704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1050),
  [8706] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_param_repeat1, 1, 0, 0),
  [8708] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1542),
  [8710] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6634),
  [8712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6121),
  [8714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1713),
  [8716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 1, 0, 0),
  [8718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6143),
  [8720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6092),
  [8722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4479),
  [8724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5927),
  [8726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7864),
  [8728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1374),
  [8730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8229),
  [8732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2210),
  [8734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7612),
  [8736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3284),
  [8738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [8740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7884),
  [8742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2561),
  [8744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(613),
  [8746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(614),
  [8748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(615),
  [8750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(616),
  [8752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(617),
  [8754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [8756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_param, 1, 0, 9),
  [8758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3324),
  [8760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1331),
  [8762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3850),
  [8764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3534),
  [8766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3652),
  [8768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5630),
  [8770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3575),
  [8772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6185),
  [8774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3187),
  [8776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3549),
  [8778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6392),
  [8780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6493),
  [8782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5883),
  [8784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1915),
  [8786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2902),
  [8788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3437),
  [8790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2398),
  [8792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6838),
  [8794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3029),
  [8796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1818),
  [8798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1818),
  [8800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(961),
  [8802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1257),
  [8804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3905),
  [8806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6483),
  [8808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1399),
  [8810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6378),
  [8812] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(6838),
  [8815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7086),
  [8817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5809),
  [8819] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7916),
  [8821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6479),
  [8823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2593),
  [8825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2313),
  [8827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3570),
  [8829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3406),
  [8831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4482),
  [8833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4201),
  [8835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6230),
  [8837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4165),
  [8839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5906),
  [8841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4284),
  [8843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3888),
  [8845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3388),
  [8847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3513),
  [8849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6394),
  [8851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5193),
  [8853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3829),
  [8855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3901),
  [8857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5925),
  [8859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2265),
  [8861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6346),
  [8863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3404),
  [8865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3520),
  [8867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6432),
  [8869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1152),
  [8871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1487),
  [8873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6028),
  [8875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1962),
  [8877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3693),
  [8879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3529),
  [8881] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_constructor_declaration_repeat1, 2, 0, 0),
  [8883] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_constructor_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(7083),
  [8886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3791),
  [8888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4423),
  [8890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2022),
  [8892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3004),
  [8894] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_specification_repeat1, 2, 0, 0),
  [8896] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_method_specification_repeat1, 2, 0, 0), SHIFT_REPEAT(6634),
  [8899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3749),
  [8901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2762),
  [8903] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_params_repeat1, 2, 0, 0),
  [8905] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__type_params_repeat1, 2, 0, 0), SHIFT_REPEAT(5358),
  [8908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1534),
  [8910] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat2, 2, 0, 0),
  [8912] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat2, 2, 0, 0), SHIFT_REPEAT(6760),
  [8915] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1818),
  [8918] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1818),
  [8921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3013),
  [8923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4403),
  [8925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4462),
  [8927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1465),
  [8929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2365),
  [8931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6670),
  [8933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4204),
  [8935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2306),
  [8937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3403),
  [8939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6386),
  [8941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1187),
  [8943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3940),
  [8945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [8947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3024),
  [8949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3162),
  [8951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4808),
  [8953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4203),
  [8955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4451),
  [8957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4453),
  [8959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2328),
  [8961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6366),
  [8963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5567),
  [8965] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_instantiated_class_type_repeat1, 2, 0, 0),
  [8967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6563),
  [8969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3869),
  [8971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6484),
  [8973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5836),
  [8975] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_instantiated_class_type_repeat1, 2, 0, 0), SHIFT_REPEAT(3408),
  [8978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1063),
  [8980] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6817),
  [8982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5183),
  [8984] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1570),
  [8986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5686),
  [8988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4831),
  [8990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(968),
  [8992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5854),
  [8994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1069),
  [8996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(549),
  [8998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2791),
  [9000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2585),
  [9002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5864),
  [9004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7068),
  [9006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2470),
  [9008] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat1, 2, 0, 0), SHIFT_REPEAT(3187),
  [9011] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat1, 2, 0, 0),
  [9013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3728),
  [9015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6489),
  [9017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6315),
  [9019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5867),
  [9021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5875),
  [9023] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__abstract_type_repeat1, 2, 0, 0), SHIFT_REPEAT(6316),
  [9026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__abstract_type_repeat1, 2, 0, 0),
  [9028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5880),
  [9030] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_instance_variable_specification_repeat1, 2, 0, 0),
  [9032] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variable_specification_repeat1, 2, 0, 0), SHIFT_REPEAT(6511),
  [9035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3713),
  [9037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6503),
  [9039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5886),
  [9041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4857),
  [9043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5893),
  [9045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5899),
  [9047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3803),
  [9049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6818),
  [9051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5902),
  [9053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5907),
  [9055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3861),
  [9057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4847),
  [9059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3677),
  [9061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3579),
  [9063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3456),
  [9065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3380),
  [9067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4422),
  [9069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4428),
  [9071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6206),
  [9073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3646),
  [9075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6316),
  [9077] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__abstract_type, 2, 0, 0),
  [9079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3838),
  [9081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4478),
  [9083] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_specification, 1, 0, 0),
  [9085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3300),
  [9087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5762),
  [9089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4480),
  [9091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2476),
  [9093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4391),
  [9095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4458),
  [9097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4357),
  [9099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4486),
  [9101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3285),
  [9103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6755),
  [9105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6411),
  [9107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1248),
  [9109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4557),
  [9111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3448),
  [9113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5424),
  [9115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5655),
  [9117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5676),
  [9119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6129),
  [9121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5683),
  [9123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3011),
  [9125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2493),
  [9127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6522),
  [9129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5698),
  [9131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6147),
  [9133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4194),
  [9135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2426),
  [9137] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__abstract_type_repeat1, 1, 0, 11),
  [9139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6051),
  [9141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(647),
  [9143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3005),
  [9145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1646),
  [9147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5706),
  [9149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2392),
  [9151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5720),
  [9153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5506),
  [9155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1717),
  [9157] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_binding_pattern, 2, 0, 48),
  [9159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6033),
  [9161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5552),
  [9163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(727),
  [9165] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 2, 0, 28),
  [9167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1929),
  [9169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3951),
  [9171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5643),
  [9173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2443),
  [9175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1667),
  [9177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7041),
  [9179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3887),
  [9181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7043),
  [9183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6149),
  [9185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2262),
  [9187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6140),
  [9189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1422),
  [9191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(711),
  [9193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_expression, 1, 0, 0),
  [9195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(594),
  [9197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5681),
  [9199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6876),
  [9201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5358),
  [9203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7265),
  [9205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5610),
  [9207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1452),
  [9209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5436),
  [9211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8249),
  [9213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7162),
  [9215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(513),
  [9217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1648),
  [9219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5664),
  [9221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1727),
  [9223] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 2, 0, 48),
  [9225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(604),
  [9227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6091),
  [9229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4929),
  [9231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6774),
  [9233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6171),
  [9235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3040),
  [9237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7514),
  [9239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5627),
  [9241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5669),
  [9243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2321),
  [9245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5431),
  [9247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6430),
  [9249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1664),
  [9251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8316),
  [9253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2236),
  [9255] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_binding_pattern_content_repeat1, 2, 0, 0), SHIFT_REPEAT(1706),
  [9258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3042),
  [9260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5786),
  [9262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6086),
  [9264] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6588),
  [9267] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_declaration_repeat1, 2, 0, 0),
  [9269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2676),
  [9271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3204),
  [9273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7731),
  [9275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5917),
  [9277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6075),
  [9279] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_pattern_content_repeat1, 2, 0, 0), SHIFT_REPEAT(1707),
  [9282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6160),
  [9284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6087),
  [9286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5577),
  [9288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2590),
  [9290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7061),
  [9292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5556),
  [9294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1312),
  [9296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3724),
  [9298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6942),
  [9300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3711),
  [9302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6300),
  [9304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1321),
  [9306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5962),
  [9308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3283),
  [9310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3741),
  [9312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5491),
  [9314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5453),
  [9316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5612),
  [9318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3807),
  [9320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3876),
  [9322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6038),
  [9324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7919),
  [9326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5621),
  [9328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1254),
  [9330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2587),
  [9332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3840),
  [9334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2302),
  [9336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5648),
  [9338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5756),
  [9340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6139),
  [9342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3805),
  [9344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(376),
  [9346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5542),
  [9348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5563),
  [9350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(945),
  [9352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6057),
  [9354] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_expression_content_repeat1, 2, 0, 0), SHIFT_REPEAT(743),
  [9357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2895),
  [9359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7030),
  [9361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6032),
  [9363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5718),
  [9365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2001),
  [9367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5601),
  [9369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1478),
  [9371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6037),
  [9373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(950),
  [9375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6094),
  [9377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2011),
  [9379] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5884),
  [9382] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_expression_repeat1, 2, 0, 0),
  [9384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5440),
  [9386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5264),
  [9388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3747),
  [9390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6850),
  [9392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [9394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5734),
  [9396] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_copy_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(6495),
  [9399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_copy_expression_repeat1, 2, 0, 0),
  [9401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7525),
  [9403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5593),
  [9405] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_indexing_operator_path, 1, 0, 0),
  [9407] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_indexing_operator_path, 1, 0, 0),
  [9409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5492),
  [9411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7114),
  [9413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [9415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6107),
  [9417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7907),
  [9419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5740),
  [9421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1974),
  [9423] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_pattern_content_repeat1, 2, 0, 0), SHIFT_REPEAT(1732),
  [9426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7469),
  [9428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5806),
  [9430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6176),
  [9432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(421),
  [9434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(439),
  [9436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(478),
  [9438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5449),
  [9440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5168),
  [9442] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_pattern_repeat1, 2, 0, 0), SHIFT_REPEAT(5839),
  [9445] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_pattern_repeat1, 2, 0, 0),
  [9447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1157),
  [9449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5747),
  [9451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6200),
  [9453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2865),
  [9455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5464),
  [9457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6891),
  [9459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [9461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [9463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5326),
  [9465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5632),
  [9467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(965),
  [9469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5576),
  [9471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1053),
  [9473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6906),
  [9475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [9477] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_instance_variable_specification_repeat1, 1, 0, 0),
  [9479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5432),
  [9481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [9483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6186),
  [9485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1056),
  [9487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2894),
  [9489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6056),
  [9491] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_indexing_operator_path, 3, 0, 0),
  [9493] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_indexing_operator_path, 3, 0, 0),
  [9495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6918),
  [9497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5482),
  [9499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5589),
  [9501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5428),
  [9503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2417),
  [9505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6928),
  [9507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6203),
  [9509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(929),
  [9511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2812),
  [9513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6933),
  [9515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5600),
  [9517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1067),
  [9519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3217),
  [9521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7085),
  [9523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1947),
  [9525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6943),
  [9527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6047),
  [9529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5888),
  [9531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5607),
  [9533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5661),
  [9535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5497),
  [9537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6950),
  [9539] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4929),
  [9542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7179),
  [9544] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6762),
  [9546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6763),
  [9548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5682),
  [9550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2567),
  [9552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6959),
  [9554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1327),
  [9556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3556),
  [9558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6202),
  [9560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3809),
  [9562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6967),
  [9564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5104),
  [9566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5639),
  [9568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3304),
  [9570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6976),
  [9572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6013),
  [9574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3333),
  [9576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6080),
  [9578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6982),
  [9580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5219),
  [9582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6988),
  [9584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3932),
  [9586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7183),
  [9588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6996),
  [9590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7002),
  [9592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5645),
  [9594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7008),
  [9596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7013),
  [9598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7018),
  [9600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7019),
  [9602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7021),
  [9604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7023),
  [9606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6254),
  [9608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3497),
  [9610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6295),
  [9612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3143),
  [9614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7042),
  [9616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5646),
  [9618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3280),
  [9620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7075),
  [9622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7076),
  [9624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7082),
  [9626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(631),
  [9628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6879),
  [9630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5203),
  [9632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(428),
  [9634] = {.entry = {.count = 1, .reusable = false}}, SHIFT(429),
  [9636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [9638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1042),
  [9640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5763),
  [9642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1390),
  [9644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3518),
  [9646] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_expression_content_repeat1, 2, 0, 0), SHIFT_REPEAT(766),
  [9649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5652),
  [9651] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6430),
  [9654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3527),
  [9656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6772),
  [9658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2215),
  [9660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6040),
  [9662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1382),
  [9664] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6086),
  [9667] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_specification_repeat1, 1, 0, 0),
  [9669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5688),
  [9671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1170),
  [9673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6090),
  [9675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1177),
  [9677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7733),
  [9679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6205),
  [9681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2573),
  [9683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5697),
  [9685] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_binding_pattern_content_repeat1, 2, 0, 0), SHIFT_REPEAT(1730),
  [9688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6167),
  [9690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7921),
  [9692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6159),
  [9694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5535),
  [9696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2778),
  [9698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5700),
  [9700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1192),
  [9702] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_binding_pattern_repeat1, 2, 0, 0), SHIFT_REPEAT(5919),
  [9705] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_binding_pattern_repeat1, 2, 0, 0),
  [9707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3338),
  [9709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5705),
  [9711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5785),
  [9713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5241),
  [9715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6856),
  [9717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6857),
  [9719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6858),
  [9721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [9723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(462),
  [9725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(463),
  [9727] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_type_repeat1, 2, 0, 0), SHIFT_REPEAT(3962),
  [9730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3692),
  [9732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6015),
  [9734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6894),
  [9736] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6895),
  [9738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6896),
  [9740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(473),
  [9742] = {.entry = {.count = 1, .reusable = false}}, SHIFT(474),
  [9744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(475),
  [9746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5408),
  [9748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3658),
  [9750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6909),
  [9752] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6910),
  [9754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6911),
  [9756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5778),
  [9758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2459),
  [9760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(482),
  [9762] = {.entry = {.count = 1, .reusable = false}}, SHIFT(483),
  [9764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(484),
  [9766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3674),
  [9768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6920),
  [9770] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6921),
  [9772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6922),
  [9774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6210),
  [9776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2462),
  [9778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(489),
  [9780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(490),
  [9782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(491),
  [9784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6223),
  [9786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6929),
  [9788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6930),
  [9790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6931),
  [9792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(496),
  [9794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(497),
  [9796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(498),
  [9798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5571),
  [9800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1407),
  [9802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6935),
  [9804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6936),
  [9806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6937),
  [9808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(503),
  [9810] = {.entry = {.count = 1, .reusable = false}}, SHIFT(504),
  [9812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(505),
  [9814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6944),
  [9816] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6945),
  [9818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6946),
  [9820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(510),
  [9822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(511),
  [9824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(512),
  [9826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5562),
  [9828] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(6139),
  [9831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4320),
  [9833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6954),
  [9835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6955),
  [9837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6956),
  [9839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(517),
  [9841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(518),
  [9843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(519),
  [9845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6961),
  [9847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6962),
  [9849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6963),
  [9851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(524),
  [9853] = {.entry = {.count = 1, .reusable = false}}, SHIFT(525),
  [9855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(526),
  [9857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5779),
  [9859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2261),
  [9861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6970),
  [9863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6971),
  [9865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6972),
  [9867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(531),
  [9869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(532),
  [9871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(533),
  [9873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6284),
  [9875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6977),
  [9877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6978),
  [9879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6979),
  [9881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3720),
  [9883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(538),
  [9885] = {.entry = {.count = 1, .reusable = false}}, SHIFT(539),
  [9887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(540),
  [9889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6983),
  [9891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6984),
  [9893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6985),
  [9895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6989),
  [9897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6990),
  [9899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6991),
  [9901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6993),
  [9903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6994),
  [9905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6995),
  [9907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6997),
  [9909] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6998),
  [9911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6999),
  [9913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7003),
  [9915] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7004),
  [9917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7005),
  [9919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7009),
  [9921] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7010),
  [9923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7011),
  [9925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7014),
  [9927] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7015),
  [9929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7016),
  [9931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5544),
  [9933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2474),
  [9935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7804),
  [9937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1286),
  [9939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3682),
  [9941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7834),
  [9943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7836),
  [9945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6141),
  [9947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2285),
  [9949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1128),
  [9951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7851),
  [9953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7852),
  [9955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7557),
  [9957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7866),
  [9959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5559),
  [9961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5526),
  [9963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5586),
  [9965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5772),
  [9967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5618),
  [9969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5120),
  [9971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5566),
  [9973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3928),
  [9975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5622),
  [9977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2945),
  [9979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5539),
  [9981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2951),
  [9983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6301),
  [9985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6108),
  [9987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2964),
  [9989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2107),
  [9991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3372),
  [9993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7178),
  [9995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8342),
  [9997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8345),
  [9999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1392),
  [10001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3353),
  [10003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1311),
  [10005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8350),
  [10007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2364),
  [10009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7031),
  [10011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [10013] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_params, 3, 0, 0),
  [10015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(694),
  [10017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(783),
  [10019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(569),
  [10021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2620),
  [10023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1316),
  [10025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7302),
  [10027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1096),
  [10029] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_type, 2, 0, 109),
  [10031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7898),
  [10033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5387),
  [10035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3715),
  [10037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(797),
  [10039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2190),
  [10041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3401),
  [10043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1980),
  [10045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7824),
  [10047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1964),
  [10049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7902),
  [10051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3322),
  [10053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7073),
  [10055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7830),
  [10057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(811),
  [10059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4796),
  [10061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7214),
  [10063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8001),
  [10065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1091),
  [10067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7237),
  [10069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1507),
  [10071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7356),
  [10073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2665),
  [10075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7740),
  [10077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4774),
  [10079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7240),
  [10081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1093),
  [10083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7261),
  [10085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(825),
  [10087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3956),
  [10089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8090),
  [10091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1161),
  [10093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2105),
  [10095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(695),
  [10097] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_guard, 2, 0, 18),
  [10099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2133),
  [10101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7764),
  [10103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4406),
  [10105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3884),
  [10107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(839),
  [10109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3411),
  [10111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1850),
  [10113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5385),
  [10115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3132),
  [10117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3317),
  [10119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7540),
  [10121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(758),
  [10123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [10125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5418),
  [10127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7784),
  [10129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3327),
  [10131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7580),
  [10133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1328),
  [10135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7999),
  [10137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(867),
  [10139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(709),
  [10141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6161),
  [10143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3198),
  [10145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1016),
  [10147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7793),
  [10149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8012),
  [10151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6492),
  [10153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7168),
  [10155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7205),
  [10157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1208),
  [10159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7801),
  [10161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [10163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2848),
  [10165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1156),
  [10167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7825),
  [10169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2362),
  [10171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7157),
  [10173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8136),
  [10175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2955),
  [10177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5604),
  [10179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5671),
  [10181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3385),
  [10183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6746),
  [10185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1288),
  [10187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7007),
  [10189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8004),
  [10191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8005),
  [10193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8006),
  [10195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6806),
  [10197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2522),
  [10199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8016),
  [10201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7967),
  [10203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5236),
  [10205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7213),
  [10207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3617),
  [10209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2524),
  [10211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8023),
  [10213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2511),
  [10215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1333),
  [10217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8312),
  [10219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(853),
  [10221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3369),
  [10223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(960),
  [10225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [10227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2923),
  [10229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8338),
  [10231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2710),
  [10233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8022),
  [10235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3907),
  [10237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5316),
  [10239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1337),
  [10241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7187),
  [10243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2935),
  [10245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7190),
  [10247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7163),
  [10249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2943),
  [10251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7123),
  [10253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8027),
  [10255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7258),
  [10257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5784),
  [10259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [10261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3967),
  [10263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7267),
  [10265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3430),
  [10267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6428),
  [10269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6800),
  [10271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8038),
  [10273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8039),
  [10275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8040),
  [10277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6828),
  [10279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3367),
  [10281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(679),
  [10283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5242),
  [10285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7100),
  [10287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(741),
  [10289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8081),
  [10291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5774),
  [10293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6729),
  [10295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7166),
  [10297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8056),
  [10299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8057),
  [10301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8058),
  [10303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7170),
  [10305] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat2, 1, 0, 0),
  [10307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7145),
  [10309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7045),
  [10311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6379),
  [10313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7143),
  [10315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8071),
  [10317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8072),
  [10319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8073),
  [10321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3553),
  [10323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8028),
  [10325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6960),
  [10327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6780),
  [10329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1959),
  [10331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6792),
  [10333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8085),
  [10335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8086),
  [10337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8087),
  [10339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6842),
  [10341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7613),
  [10343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6869),
  [10345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8099),
  [10347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8100),
  [10349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8101),
  [10351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2256),
  [10353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7628),
  [10355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6847),
  [10357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5587),
  [10359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7485),
  [10361] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_params, 4, 0, 0),
  [10363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6940),
  [10365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8113),
  [10367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8114),
  [10369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8115),
  [10371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3157),
  [10373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3764),
  [10375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5591),
  [10377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7513),
  [10379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5518),
  [10381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5314),
  [10383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5239),
  [10385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4444),
  [10387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7044),
  [10389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8127),
  [10391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8128),
  [10393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8129),
  [10395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2611),
  [10397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3943),
  [10399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3800),
  [10401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7097),
  [10403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8141),
  [10405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8142),
  [10407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8143),
  [10409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2933),
  [10411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3561),
  [10413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1491),
  [10415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [10417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3046),
  [10419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1123),
  [10421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5362),
  [10423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7155),
  [10425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8155),
  [10427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8156),
  [10429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8157),
  [10431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7774),
  [10433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6981),
  [10435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8169),
  [10437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8170),
  [10439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8171),
  [10441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7782),
  [10443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5294),
  [10445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2292),
  [10447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7757),
  [10449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6766),
  [10451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8176),
  [10453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8177),
  [10455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8178),
  [10457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2826),
  [10459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7716),
  [10461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4198),
  [10463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6787),
  [10465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8181),
  [10467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8182),
  [10469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8183),
  [10471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3350),
  [10473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6794),
  [10475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8186),
  [10477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8187),
  [10479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8188),
  [10481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6816),
  [10483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8191),
  [10485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8192),
  [10487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8193),
  [10489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7287),
  [10491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6843),
  [10493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8196),
  [10495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8197),
  [10497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8198),
  [10499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3103),
  [10501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8069),
  [10503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1008),
  [10505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7362),
  [10507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6860),
  [10509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8201),
  [10511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8202),
  [10513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8203),
  [10515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6874),
  [10517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8206),
  [10519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8207),
  [10521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8208),
  [10523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3702),
  [10525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4393),
  [10527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1464),
  [10529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2737),
  [10531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5507),
  [10533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1080),
  [10535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1020),
  [10537] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5398),
  [10539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5197),
  [10541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5331),
  [10543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3516),
  [10545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7703),
  [10547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2444),
  [10549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7812),
  [10551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5048),
  [10553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8035),
  [10555] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5366),
  [10557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2343),
  [10559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7820),
  [10561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3357),
  [10563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3739),
  [10565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7828),
  [10567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3016),
  [10569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4504),
  [10571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(433),
  [10573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3294),
  [10575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3362),
  [10577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4824),
  [10579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4454),
  [10581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(664),
  [10583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4488),
  [10585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6531),
  [10587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3743),
  [10589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7844),
  [10591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [10593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [10595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2337),
  [10597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7912),
  [10599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(724),
  [10601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3391),
  [10603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2388),
  [10605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1424),
  [10607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7630),
  [10609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1885),
  [10611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1426),
  [10613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6938),
  [10615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7821),
  [10617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7822),
  [10619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1205),
  [10621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8054),
  [10623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1987),
  [10625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(879),
  [10627] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 2, 0, 135),
  [10629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7823),
  [10631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3473),
  [10633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4360),
  [10635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7622),
  [10637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3337),
  [10639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3446),
  [10641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3925),
  [10643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2920),
  [10645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2331),
  [10647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7988),
  [10649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2687),
  [10651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3669),
  [10653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5371),
  [10655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2065),
  [10657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8034),
  [10659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5373),
  [10661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5251),
  [10663] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 3, 0, 170),
  [10665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(464),
  [10667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2359),
  [10669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8010),
  [10671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [10673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3563),
  [10675] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_params, 1, 0, 0),
  [10677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1398),
  [10679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7842),
  [10681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2972),
  [10683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1241),
  [10685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6952),
  [10687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(476),
  [10689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(477),
  [10691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1994),
  [10693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2758),
  [10695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5223),
  [10697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3449),
  [10699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7980),
  [10701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5009),
  [10703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6965),
  [10705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3734),
  [10707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1010),
  [10709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8132),
  [10711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(993),
  [10713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5520),
  [10715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7990),
  [10717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6629),
  [10719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7569),
  [10721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2864),
  [10723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2584),
  [10725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2504),
  [10727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8214),
  [10729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3299),
  [10731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3842),
  [10733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1207),
  [10735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1209),
  [10737] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5389),
  [10739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3307),
  [10741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2492),
  [10743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2506),
  [10745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8245),
  [10747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3047),
  [10749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8291),
  [10751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1481),
  [10753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2751),
  [10755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4820),
  [10757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3384),
  [10759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6923),
  [10761] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__polymorphic_typed, 2, 0, 52),
  [10763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5551),
  [10765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3031),
  [10767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8300),
  [10769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2822),
  [10771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7930),
  [10773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2760),
  [10775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3349),
  [10777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3911),
  [10779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4963),
  [10781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4487),
  [10783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5243),
  [10785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4766),
  [10787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1450),
  [10789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7878),
  [10791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8341),
  [10793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4358),
  [10795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6783),
  [10797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5858),
  [10799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1488),
  [10801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4691),
  [10803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1361),
  [10805] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument_type, 5, 0, 182),
  [10807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5014),
  [10809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2936),
  [10811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5320),
  [10813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2890),
  [10815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2268),
  [10817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3900),
  [10819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4974),
  [10821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2893),
  [10823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1566),
  [10825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1184),
  [10827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3397),
  [10829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1511),
  [10831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7142),
  [10833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2529),
  [10835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2986),
  [10837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(693),
  [10839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3698),
  [10841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1242),
  [10843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [10845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [10847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7419),
  [10849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1185),
  [10851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3622),
  [10853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4789),
  [10855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2419),
  [10857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3912),
  [10859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4107),
  [10861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1529),
  [10863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4363),
  [10865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1530),
  [10867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4441),
  [10869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2723),
  [10871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(963),
  [10873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6423),
  [10875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(740),
  [10877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3377),
  [10879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [10881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2808),
  [10883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(618),
  [10885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [10887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [10889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2987),
  [10891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3885),
  [10893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3398),
  [10895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1092),
  [10897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1310),
  [10899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4593),
  [10901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4779),
  [10903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4122),
  [10905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1172),
  [10907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5570),
  [10909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1173),
  [10911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4977),
  [10913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1174),
  [10915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [10917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2975),
  [10919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7639),
  [10921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4599),
  [10923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2420),
  [10925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6889),
  [10927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3006),
  [10929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(424),
  [10931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5543),
  [10933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1052),
  [10935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2531),
  [10937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1094),
  [10939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5247),
  [10941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5466),
  [10943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1459),
  [10945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1829),
  [10947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2532),
  [10949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3957),
  [10951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4648),
  [10953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4988),
  [10955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1496),
  [10957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(375),
  [10959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1449),
  [10961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5123),
  [10963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4999),
  [10965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1520),
  [10967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1054),
  [10969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1521),
  [10971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7115),
  [10973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2533),
  [10975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(677),
  [10977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1556),
  [10979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [10981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2197),
  [10983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3889),
  [10985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4695),
  [10987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3341),
  [10989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4033),
  [10991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1523),
  [10993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3894),
  [10995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1528),
  [10997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1475),
  [10999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3896),
  [11001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2263),
  [11003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6728),
  [11005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2428),
  [11007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [11009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(340),
  [11011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2994),
  [11013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1319),
  [11015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(756),
  [11017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4678),
  [11019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4076),
  [11021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2995),
  [11023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2996),
  [11025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [11027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4685),
  [11029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6986),
  [11031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2621),
  [11033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(359),
  [11035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2597),
  [11037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1335),
  [11039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1453),
  [11041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1359),
  [11043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1373),
  [11045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5441),
  [11047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5467),
  [11049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2795),
  [11051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2373),
  [11053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4703),
  [11055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4975),
  [11057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5172),
  [11059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7574),
  [11061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3026),
  [11063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4982),
  [11065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1555),
  [11067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3988),
  [11069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1057),
  [11071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1557),
  [11073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7096),
  [11075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4394),
  [11077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1414),
  [11079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3007),
  [11081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(556),
  [11083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1558),
  [11085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [11087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1559),
  [11089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2695),
  [11091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1061),
  [11093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(663),
  [11095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2747),
  [11097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6376),
  [11099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3684),
  [11101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5963),
  [11103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2766),
  [11105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [11107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4709),
  [11109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2106),
  [11111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1009),
  [11113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4713),
  [11115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7154),
  [11117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [11119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(564),
  [11121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1326),
  [11123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2369),
  [11125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2229),
  [11127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2992),
  [11129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [11131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4724),
  [11133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5020),
  [11135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2796),
  [11137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2220),
  [11139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5027),
  [11141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1576),
  [11143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1577),
  [11145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6785),
  [11147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4631),
  [11149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6062),
  [11151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1578),
  [11153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1544),
  [11155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2789),
  [11157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1064),
  [11159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(934),
  [11161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(935),
  [11163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6426),
  [11165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2154),
  [11167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(492),
  [11169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4730),
  [11171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4196),
  [11173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(936),
  [11175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [11177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4162),
  [11179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4734),
  [11181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6797),
  [11183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2247),
  [11185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4188),
  [11187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2810),
  [11189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2776),
  [11191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4993),
  [11193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7020),
  [11195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3503),
  [11197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4750),
  [11199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4992),
  [11201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1065),
  [11203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(708),
  [11205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4841),
  [11207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4995),
  [11209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1515),
  [11211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1516),
  [11213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6855),
  [11215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(446),
  [11217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1006),
  [11219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5402),
  [11221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1517),
  [11223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1518),
  [11225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1979),
  [11227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(565),
  [11229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2437),
  [11231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [11233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6487),
  [11235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5184),
  [11237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(944),
  [11239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3273),
  [11241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4767),
  [11243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [11245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4362),
  [11247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4181),
  [11249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1034),
  [11251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4771),
  [11253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6880),
  [11255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1033),
  [11257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1035),
  [11259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3958),
  [11261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2127),
  [11263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(972),
  [11265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5719),
  [11267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4656),
  [11269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1070),
  [11271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4398),
  [11273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4990),
  [11275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1524),
  [11277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1525),
  [11279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6927),
  [11281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1344),
  [11283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2622),
  [11285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1526),
  [11287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1527),
  [11289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4780),
  [11291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3499),
  [11293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1230),
  [11295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1036),
  [11297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6538),
  [11299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5208),
  [11301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4443),
  [11303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4739),
  [11305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1037),
  [11307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [11309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [11311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1038),
  [11313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4758),
  [11315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6951),
  [11317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5588),
  [11319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5209),
  [11321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5488),
  [11323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(362),
  [11325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4610),
  [11327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4811),
  [11329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5505),
  [11331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1535),
  [11333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1536),
  [11335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7027),
  [11337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3986),
  [11339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2390),
  [11341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [11343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2566),
  [11345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1538),
  [11347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1539),
  [11349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [11351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5214),
  [11353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1238),
  [11355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6616),
  [11357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1948),
  [11359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4914),
  [11361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4637),
  [11363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1355),
  [11365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5595),
  [11367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1842),
  [11369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2226),
  [11371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4644),
  [11373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7059),
  [11375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5022),
  [11377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4934),
  [11379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1396),
  [11381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2312),
  [11383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [11385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1219),
  [11387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4662),
  [11389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2869),
  [11391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [11393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1541),
  [11395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1579),
  [11397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7090),
  [11399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1509),
  [11401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4367),
  [11403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1545),
  [11405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1546),
  [11407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(444),
  [11409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3320),
  [11411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3417),
  [11413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1510),
  [11415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6650),
  [11417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(948),
  [11419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(445),
  [11421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(397),
  [11423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4693),
  [11425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(451),
  [11427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6872),
  [11429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1992),
  [11431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2871),
  [11433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4699),
  [11435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7109),
  [11437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3933),
  [11439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7855),
  [11441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [11443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3393),
  [11445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2311),
  [11447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(723),
  [11449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4736),
  [11451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2899),
  [11453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3722),
  [11455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1550),
  [11457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1551),
  [11459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7146),
  [11461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1981),
  [11463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2872),
  [11465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3211),
  [11467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(580),
  [11469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1552),
  [11471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1553),
  [11473] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_string, 3, 0, 0),
  [11475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3014),
  [11477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1244),
  [11479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1348),
  [11481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6734),
  [11483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4413),
  [11485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3361),
  [11487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [11489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4751),
  [11491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3303),
  [11493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1982),
  [11495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [11497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3595),
  [11499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4759),
  [11501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7165),
  [11503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5647),
  [11505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6868),
  [11507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1030),
  [11509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3015),
  [11511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3312),
  [11513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4561),
  [11515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(937),
  [11517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(938),
  [11519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1562),
  [11521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1563),
  [11523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6849),
  [11525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(939),
  [11527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3537),
  [11529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4229),
  [11531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1564),
  [11533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1565),
  [11535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2003),
  [11537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2863),
  [11539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6312),
  [11541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1432),
  [11543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2350),
  [11545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4569),
  [11547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8014),
  [11549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4573),
  [11551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7063),
  [11553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2320),
  [11555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3017),
  [11557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1428),
  [11559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3550),
  [11561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1264),
  [11563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3530),
  [11565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4579),
  [11567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(593),
  [11569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7153),
  [11571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [11573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3188),
  [11575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6319),
  [11577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2960),
  [11579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2807),
  [11581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4583),
  [11583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3345),
  [11585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4588),
  [11587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6778),
  [11589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4003),
  [11591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3434),
  [11593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3018),
  [11595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [11597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(392),
  [11599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(796),
  [11601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2627),
  [11603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2153),
  [11605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6788),
  [11607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2624),
  [11609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2161),
  [11611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3596),
  [11613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
  [11615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6798),
  [11617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2625),
  [11619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3290),
  [11621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [11623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(951),
  [11625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5702),
  [11627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1147),
  [11629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6822),
  [11631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(810),
  [11633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2793),
  [11635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4912),
  [11637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1543),
  [11639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(824),
  [11641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3598),
  [11643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6846),
  [11645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1353),
  [11647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2286),
  [11649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(838),
  [11651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [11653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3752),
  [11655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6864),
  [11657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2442),
  [11659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4909),
  [11661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1356),
  [11663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5263),
  [11665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3328),
  [11667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5656),
  [11669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6881),
  [11671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2966),
  [11673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1162),
  [11675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4617),
  [11677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2874),
  [11679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [11681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2015),
  [11683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5266),
  [11685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1307),
  [11687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4987),
  [11689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2804),
  [11691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3020),
  [11693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3329),
  [11695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [11697] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_payload, 3, 0, 0),
  [11699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(367),
  [11701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2422),
  [11703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4609),
  [11705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1831),
  [11707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2409),
  [11709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8243),
  [11711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3021),
  [11713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [11715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3199),
  [11717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [11719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1005),
  [11721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5901),
  [11723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [11725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3410),
  [11727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(784),
  [11729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4608),
  [11731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5195),
  [11733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3860),
  [11735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(390),
  [11737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3023),
  [11739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3260),
  [11741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3685),
  [11743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3686),
  [11745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3980),
  [11747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [11749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5003),
  [11751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2294),
  [11753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1438),
  [11755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3278),
  [11757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2300),
  [11759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2858),
  [11761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1512),
  [11763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7049),
  [11765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1571),
  [11767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2007),
  [11769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(499),
  [11771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [11773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1572),
  [11775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3056),
  [11777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5787),
  [11779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5199),
  [11781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1513),
  [11783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6768),
  [11785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2008),
  [11787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5278),
  [11789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2009),
  [11791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1387),
  [11793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5292),
  [11795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5853),
  [11797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7138),
  [11799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3695),
  [11801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3288),
  [11803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2010),
  [11805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3601),
  [11807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3289),
  [11809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5157),
  [11811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3291),
  [11813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5929),
  [11815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8111),
  [11817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5158),
  [11819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2078),
  [11821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3293),
  [11823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [11825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4794),
  [11827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(506),
  [11829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1146),
  [11831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6530),
  [11833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [11835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1022),
  [11837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3522),
  [11839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2602),
  [11841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8180),
  [11843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3181),
  [11845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3059),
  [11847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3060),
  [11849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5055),
  [11851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2797),
  [11853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(956),
  [11855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5980),
  [11857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2569),
  [11859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3308),
  [11861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3314),
  [11863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1315),
  [11865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2361),
  [11867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8210),
  [11869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8211),
  [11871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8212),
  [11873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1961),
  [11875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1158),
  [11877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3315),
  [11879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3740),
  [11881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4054),
  [11883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5957),
  [11885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [11887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5194),
  [11889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8238),
  [11891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2750),
  [11893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1021),
  [11895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1371),
  [11897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2603),
  [11899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3881),
  [11901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2998),
  [11903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4000),
  [11905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2974),
  [11907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1249),
  [11909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3744),
  [11911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3405),
  [11913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2214),
  [11915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4940),
  [11917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2755),
  [11919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3052),
  [11921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1573),
  [11923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8319),
  [11925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1027),
  [11927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2631),
  [11929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1574),
  [11931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4938),
  [11933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1007),
  [11935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2632),
  [11937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [11939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [11941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2216),
  [11943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2402),
  [11945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7847),
  [11947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [11949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [11951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6325),
  [11953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(599),
  [11955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(603),
  [11957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3738),
  [11959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3746),
  [11961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1168),
  [11963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [11965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1455),
  [11967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2267),
  [11969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7071),
  [11971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2404),
  [11973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6655),
  [11975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1175),
  [11977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2805),
  [11979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2806),
  [11981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1251),
  [11983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2594),
  [11985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2757),
  [11987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3201),
  [11989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4925),
  [11991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [11993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [11995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1976),
  [11997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2351),
  [11999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2282),
  [12001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [12003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1197),
  [12005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1222),
  [12007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(456),
  [12009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1403),
  [12011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1226),
  [12013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2344),
  [12015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5885),
  [12017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2595),
  [12019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(468),
  [12021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7079),
  [12023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5847),
  [12025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6551),
  [12027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1088),
  [12029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6373),
  [12031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1098),
  [12033] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [12035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2604),
  [12037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2637),
  [12039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4008),
  [12041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1531),
  [12043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [12045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1142),
  [12047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2823),
  [12049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3751),
  [12051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [12053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2519),
  [12055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1155),
  [12057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2375),
  [12059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5233),
  [12061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2870),
  [12063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1292),
  [12065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1963),
  [12067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2421),
  [12069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4706),
  [12071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1179),
  [12073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2779),
  [12075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [12077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(559),
  [12079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(878),
  [12081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(550),
  [12083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2405),
  [12085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5196),
  [12087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7742),
  [12089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(552),
  [12091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [12093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4813),
  [12095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(554),
  [12097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4370),
  [12099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1394),
  [12101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2433),
  [12103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(555),
  [12105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4809),
  [12107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1132),
  [12109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4913),
  [12111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1133),
  [12113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1136),
  [12115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3718),
  [12117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1368),
  [12119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7961),
  [12121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1369),
  [12123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2596),
  [12125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1182),
  [12127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4015),
  [12129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2309),
  [12131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4050),
  [12133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2844),
  [12135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1151),
  [12137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1971),
  [12139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5835),
  [12141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5773),
  [12143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6544),
  [12145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1129),
  [12147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1119),
  [12149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1130),
  [12151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1131),
  [12153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2333),
  [12155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6853),
  [12157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2575),
  [12159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2845),
  [12161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1120),
  [12163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6453),
  [12165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8355),
  [12167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7175),
  [12169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1121),
  [12171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [12173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5841),
  [12175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4089),
  [12177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3196),
  [12179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [12181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8041),
  [12183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8044),
  [12185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8052),
  [12187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [12189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [12191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2386),
  [12193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(153),
  [12195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7570),
  [12197] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument_type, 6, 0, 202),
  [12199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1108),
  [12201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2523),
  [12203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [12205] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_string, 2, 0, 0),
  [12207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6014),
  [12209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7492),
  [12211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(420),
  [12213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5852),
  [12215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2525),
  [12217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1400),
  [12219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7012),
  [12221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6892),
  [12223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [12225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4723),
  [12227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(608),
  [12229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3057),
  [12231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5860),
  [12233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4939),
  [12235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3481),
  [12237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2988),
  [12239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7253),
  [12241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7254),
  [12243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7255),
  [12245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7134),
  [12247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2922),
  [12249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3482),
  [12251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [12253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3657),
  [12255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [12257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5289),
  [12259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2314),
  [12261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1273),
  [12263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2458),
  [12265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4020),
  [12267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5866),
  [12269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3002),
  [12271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7316),
  [12273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7317),
  [12275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7318),
  [12277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2929),
  [12279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3033),
  [12281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3661),
  [12283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5743),
  [12285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2460),
  [12287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5167),
  [12289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1293),
  [12291] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__infix_operator, 1, 0, 0),
  [12293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5871),
  [12295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4560),
  [12297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7366),
  [12299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7367),
  [12301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7368),
  [12303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [12305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7541),
  [12307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3759),
  [12309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1404),
  [12311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5634),
  [12313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5169),
  [12315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5560),
  [12317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1412),
  [12319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2937),
  [12321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7405),
  [12323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7406),
  [12325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7407),
  [12327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2446),
  [12329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4745),
  [12331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3944),
  [12333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2938),
  [12335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3977),
  [12337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [12339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2464),
  [12341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [12343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2847),
  [12345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [12347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2334),
  [12349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7444),
  [12351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7445),
  [12353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7446),
  [12355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2527),
  [12357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1274),
  [12359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2468),
  [12361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1376),
  [12363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(866),
  [12365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4084),
  [12367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2339),
  [12369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3839),
  [12371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7481),
  [12373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7482),
  [12375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7483),
  [12377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5161),
  [12379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7188),
  [12381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2340),
  [12383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1937),
  [12385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(387),
  [12387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1212),
  [12389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2777),
  [12391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3041),
  [12393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1938),
  [12395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(609),
  [12397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7517),
  [12399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7518),
  [12401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7519),
  [12403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(334),
  [12405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(995),
  [12407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(967),
  [12409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1940),
  [12411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [12413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7176),
  [12415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5753),
  [12417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2979),
  [12419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [12421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2355),
  [12423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7553),
  [12425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7554),
  [12427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7555),
  [12429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1216),
  [12431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2260),
  [12433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1997),
  [12435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1277),
  [12437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(997),
  [12439] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_character_content, 1, 0, 0),
  [12441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1278),
  [12443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7588),
  [12445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7589),
  [12447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7590),
  [12449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5064),
  [12451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2471),
  [12453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2528),
  [12455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2957),
  [12457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2958),
  [12459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3484),
  [12461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2270),
  [12463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2418),
  [12465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(401),
  [12467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2379),
  [12469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4935),
  [12471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7624),
  [12473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7625),
  [12475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7626),
  [12477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7737),
  [12479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3265),
  [12481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3135),
  [12483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2472),
  [12485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7647),
  [12487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7648),
  [12489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7649),
  [12491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [12493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3416),
  [12495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7660),
  [12497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7661),
  [12499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7662),
  [12501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2586),
  [12503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2239),
  [12505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7670),
  [12507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7671),
  [12509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7672),
  [12511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3697),
  [12513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2240),
  [12515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7680),
  [12517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7681),
  [12519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7682),
  [12521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(407),
  [12523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7690),
  [12525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7691),
  [12527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7692),
  [12529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4601),
  [12531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2241),
  [12533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7700),
  [12535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7701),
  [12537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7702),
  [12539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2654),
  [12541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [12543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7710),
  [12545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7711),
  [12547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7712),
  [12549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3043),
  [12551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7039),
  [12553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1289),
  [12555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2505),
  [12557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6611),
  [12559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(548),
  [12561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2781),
  [12563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1290),
  [12565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1291),
  [12567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [12569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2588),
  [12571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7236),
  [12573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3681),
  [12575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3865),
  [12577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4010),
  [12579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3379),
  [12581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2965),
  [12583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2477),
  [12585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2478),
  [12587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2479),
  [12589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1466),
  [12591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1279),
  [12593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1116),
  [12595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4937),
  [12597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1280),
  [12599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4845),
  [12601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1164),
  [12603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4916),
  [12605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2480),
  [12607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2507),
  [12609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4336),
  [12611] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compilation_unit, 2, 0, 0),
  [12613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3383),
  [12615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7509),
  [12617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1001),
  [12619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1002),
  [12621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3266),
  [12623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2581),
  [12625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2318),
  [12627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1454),
  [12629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3270),
  [12631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(962),
  [12633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7925),
  [12635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1467),
  [12637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2316),
  [12639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5582),
  [12641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [12643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2867),
  [12645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5590),
  [12647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2324),
  [12649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6656),
  [12651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2482),
  [12653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4863),
  [12655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2483),
  [12657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2484),
  [12659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6662),
  [12661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3178),
  [12663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [12665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6667),
  [12667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6672),
  [12669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6677),
  [12671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6682),
  [12673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6685),
  [12675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6692),
  [12677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6696),
  [12679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6700),
  [12681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6704),
  [12683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6708),
  [12685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6710),
  [12687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6712),
  [12689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6714),
  [12691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6716),
  [12693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6718),
  [12695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6720),
  [12697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1941),
  [12699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2997),
  [12701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6152),
  [12703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4915),
  [12705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1508),
  [12707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5185),
  [12709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6154),
  [12711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8240),
  [12713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5119),
  [12715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3035),
  [12717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1942),
  [12719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8248),
  [12721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1338),
  [12723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1468),
  [12725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2348),
  [12727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [12729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1943),
  [12731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2861),
  [12733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1372),
  [12735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5202),
  [12737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [12739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1102),
  [12741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2682),
  [12743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5021),
  [12745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4928),
  [12747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2396),
  [12749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8349),
  [12751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(852),
  [12753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5650),
  [12755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1259),
  [12757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2944),
  [12759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5754),
  [12761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1110),
  [12763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2950),
  [12765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6555),
  [12767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2363),
  [12769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2952),
  [12771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1117),
  [12773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1281),
  [12775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2901),
  [12777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2932),
  [12779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [12781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7880),
  [12783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7881),
  [12785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2946),
  [12787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2873),
  [12789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7882),
  [12791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1294),
  [12793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3750),
  [12795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3184),
  [12797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1314),
  [12799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1183),
  [12801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4415),
  [12803] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 3, 0, 0),
  [12805] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 4, 0, 0),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token_comment = 0,
  ts_external_token__left_quoted_string_delimiter = 1,
  ts_external_token__right_quoted_string_delimiter = 2,
  ts_external_token_DQUOTE = 3,
  ts_external_token_line_number_directive = 4,
  ts_external_token__null = 5,
  ts_external_token__error_sentinel = 6,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token_comment] = sym_comment,
  [ts_external_token__left_quoted_string_delimiter] = sym__left_quoted_string_delimiter,
  [ts_external_token__right_quoted_string_delimiter] = sym__right_quoted_string_delimiter,
  [ts_external_token_DQUOTE] = anon_sym_DQUOTE,
  [ts_external_token_line_number_directive] = sym_line_number_directive,
  [ts_external_token__null] = sym__null,
  [ts_external_token__error_sentinel] = sym__error_sentinel,
};

static const bool ts_external_scanner_states[10][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token_comment] = true,
    [ts_external_token__left_quoted_string_delimiter] = true,
    [ts_external_token__right_quoted_string_delimiter] = true,
    [ts_external_token_DQUOTE] = true,
    [ts_external_token_line_number_directive] = true,
    [ts_external_token__null] = true,
    [ts_external_token__error_sentinel] = true,
  },
  [2] = {
    [ts_external_token_comment] = true,
    [ts_external_token_DQUOTE] = true,
    [ts_external_token_line_number_directive] = true,
  },
  [3] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_number_directive] = true,
  },
  [4] = {
    [ts_external_token_comment] = true,
    [ts_external_token__left_quoted_string_delimiter] = true,
    [ts_external_token_DQUOTE] = true,
    [ts_external_token_line_number_directive] = true,
  },
  [5] = {
    [ts_external_token_comment] = true,
    [ts_external_token_DQUOTE] = true,
    [ts_external_token_line_number_directive] = true,
    [ts_external_token__null] = true,
  },
  [6] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_number_directive] = true,
    [ts_external_token__null] = true,
  },
  [7] = {
    [ts_external_token_comment] = true,
    [ts_external_token__right_quoted_string_delimiter] = true,
    [ts_external_token_line_number_directive] = true,
    [ts_external_token__null] = true,
  },
  [8] = {
    [ts_external_token_comment] = true,
    [ts_external_token__left_quoted_string_delimiter] = true,
    [ts_external_token_line_number_directive] = true,
  },
  [9] = {
    [ts_external_token_comment] = true,
    [ts_external_token__right_quoted_string_delimiter] = true,
    [ts_external_token_line_number_directive] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_ocaml_external_scanner_create(void);
void tree_sitter_ocaml_external_scanner_destroy(void *);
bool tree_sitter_ocaml_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_ocaml_external_scanner_serialize(void *, char *);
void tree_sitter_ocaml_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_ocaml(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .supertype_map_slices = ts_supertype_map_slices,
    .supertype_map_entries = ts_supertype_map_entries,
    .supertype_symbols = ts_supertype_symbols,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym__lowercase_identifier,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_ocaml_external_scanner_create,
      tree_sitter_ocaml_external_scanner_destroy,
      tree_sitter_ocaml_external_scanner_scan,
      tree_sitter_ocaml_external_scanner_serialize,
      tree_sitter_ocaml_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "ocaml",
    .reserved_words = &ts_reserved_words[0][0],
    .max_reserved_word_set_size = 55,
    .metadata = {
      .major_version = 0,
      .minor_version = 25,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
